#
# this module looks for PETSC
#  ( Portable, Extensible Toolkit for
#     Scientific Computation)
# http://www-unix.mcs.anl.gov/petsc/petsc-2/
#
# it will search for the following values
#
# PETSC_ARCH        = Type of system optimization. It takes $PETSC_ARCH 
#                     environnement variable if defined.
# BOPT              = Specify the PETSC's build type.
# PETSC_DIR         = Directory where PETSC is installed
# PETSC_VARS_FILE   = Makefile to include to determine compilation and link flags
#                       ex: ...../bmake/common/variables
#
# and set those variables :
#
# PETSC_EXE_LINKER_FLAGS = to set the linker flags for exe
#   ex : SET(CMAKE_EXE_LINKER_FLAGS 
#            ${CMAKE_EXE_LINKER_FLAGS} 
#            ${PETSC_EXE_LINKER_FLAGS})
#
# PETSC_CCPPFLAGS   = C and CCC flags to compile with PETSC
#    ex : CMAKE_C_FLAGS("${CMAKE_C_FLAGS} ${PETSC_CCPPFLAGS}")
#         CMAKE_CXX_FLAGS("${CMAKE_CXX_FLAGS} ${PETSC_CCPPFLAGS}")
#
# PETSC_SLES_LIB    = Link flags to use Linear Solvers
# PETSC_SYS_LIB     = Link flags to use PETSC System routines
# PETSC_VEC_LIB     = Link flags to use Vectors operations
# PETSC_MAT_LIB     = Link flags to use Matrices operations
# PETSC_SLES_LIB    = Link flags to use Linear Solvers
# PETSC_SNES_LIB    = Link flags to use Nonlinear Solvers
# PETSC_TS_LIB      = Link flags to use Time Stepping ODE solvers
# PETSC_MESH_LIB    = Link flags to use Unstructured Meshes
# PETSC_GRID_LIB    = Link flags to use Computational Grids
# PETSC_GSOLVER_LIB = Link flags to use Solvers on Grids
#    ex :  LINK_DIRECTORIES ("${PETSC_SLES_LIB}")

# Search for findPETSC-Vars.make file used to get PETSC build variables
FIND_FILE(PETSC_GET_VARS_FILE findPETSC-Vars.make
    ${CMAKE_MODULE_PATH}
    DOC "Makefile that get formated results from variables Makefile" 
)
MARK_AS_ADVANCED(PETSC_GET_VARS_FILE)


FIND_PATH(PETSC_DIR include/petsc.h
	/usr/local
	/usr 
	/usr/local/petsc
)

SET(PETSC_ARCH "$ENV{PETSC_ARCH}" CACHE STRING "System for which PETSC is optimized linux hpux alpha ...")
IF(NOT PETSC_ARCH)
  SET(PETSC_ARCH "linux")
ENDIF(NOT PETSC_ARCH)

FIND_FILE(PETSC_VARS_FILE variables
    ${PETSC_DIR}/bmake/common
    DOC "Makefile that defines build and link flags (PETSC_DIR/bmake/common/variables generaly)" 
)

#SET(PETSC_BOPT "$ENV{BOPT}" CACHE STRING "PETSC's build type : g,O,O_c++,O_complex,...")
#IF(NOT PETSC_BOPT)
#  SET(PETSC_BOPT "O_c++")
#ENDIF(NOT PETSC_BOPT)

# SET(PETSC_EXE_LINKER_FLAGS "-Wl,-rpath,${PETSC_DIR}/lib/lib${PETSC_BOPT}/${PETSC_ARCH}")

IF(NOT APPLE)
  SET(PETSC_EXE_LINKER_FLAGS "-Wl,-rpath,${PETSC_DIR}/lib/${PETSC_ARCH}")
ELSE(NOT APPLE)
   SET(PETSC_EXE_LINKER_FLAGS "-Wl,-L${PETSC_DIR}/externalpackages/mpich2-1.0.5p4/${PETSC_ARCH}/lib/,-lmpich")
ENDIF(NOT APPLE)

SET(PETSC_VARS
	CCPPFLAGS 
	PETSC_KSP_LIB 
	PETSC_SYS_LIB	
	PETSC_VEC_LIB
	PETSC_MAT_LIB 
	PETSC_SNES_LIB 
	PETSC_TS_LIB 
	PETSC_DM_LIB
	PETSC_LIB
	PETSC_FORTRAN_LIB
	PETSC_CONTRIB
)


FOREACH (PETSC_VAR ${PETSC_VARS})
  EXEC_PROGRAM(make 
	  ARGS 'PETSC_DIR=${PETSC_DIR}' 'PETSC_ARCH=${PETSC_ARCH}' 'PETSC_VARS_FILE=${PETSC_VARS_FILE}' 'WANTED_VARIABLE=${PETSC_VAR}' -s -f ${PETSC_GET_VARS_FILE}
	  OUTPUT_VARIABLE PESTC_TMP)
  IF(PESTC_TMP)
    SET("${PETSC_VAR}" ${PESTC_TMP} CACHE INTERNAL "This variable is set automaticaly by PETSC bmake build process" FORCE)
  ENDIF(PESTC_TMP)
ENDFOREACH (PETSC_VAR ${PETSC_VARS})
