#
# this module looks for XML-2.0 
#
# it will search for the following values
#
# PKGCONFIG      = Package configuration program  
#                  (http://www.freedesktop.org/Software/pkgconfig)
#
# and set those variables :
#
# XML_CFLAGS      = Compilation flags and include directories. 
#                     Ex: ADD_DEFINITIONS(${XML_CFLAGS})
# XML_INCLUDE_DIRS= Include directories. 
#                     Ex: INCLUDE_DIRECTORIES(${XML_INCLUDE_DIRS})
# XML_LINK_LIBS   = Libraries to link with. 
#                     Ex: LINK_LIBRARIES(${XML_LINK_LIBS})
# XML_LINK_DIRS   = Path to find the libs to link with. 
#                     Ex: LINK_DIRECTORIES(${XML_LINK_DIRS})
#
#
FIND_PROGRAM(PKGCONFIG pkg-config DOC "Full path to the ${PKGCONFIG} tool")

IF (PKGCONFIG)

  EXEC_PROGRAM (${PKGCONFIG} ARGS --cflags libxml-2.0
	OUTPUT_VARIABLE PKG_CFLAGS)

  EXEC_PROGRAM (${PKGCONFIG} ARGS --cflags-only-I libxml-2.0
	OUTPUT_VARIABLE PKG_INCLUDE_DIRS)

  EXEC_PROGRAM (${PKGCONFIG} ARGS --libs-only-L libxml-2.0
	OUTPUT_VARIABLE PKG_LINK_DIRS)
  EXEC_PROGRAM (${PKGCONFIG} ARGS --libs-only-l libxml-2.0
	OUTPUT_VARIABLE PKG_LINK_LIBS)
  
  STRING(REGEX REPLACE "^-I[^ ]*| -I[^ ]*" "" REX_CFLAGS "${PKG_CFLAGS}")

  STRING(REGEX REPLACE " -I" ";" REX "${PKG_INCLUDE_DIRS}")
  STRING(REGEX REPLACE "^-I" "" REX_INCLUDE_DIRS "${REX}")

  STRING(REGEX REPLACE " -L" ";" REX "${PKG_LINK_DIRS}")
  STRING(REGEX REPLACE "^-L" "" REX_LINK_DIRS "${REX}")
  
  STRING(REGEX REPLACE " -l" ";" REX "${PKG_LINK_LIBS}")
  STRING(REGEX REPLACE "^-l" "" REX_LINK_LIBS "${REX}")
  
  #MESSAGE(PKG_LINK_LIBS= "${PKG_LINK_LIBS}")
  #MESSAGE(REX= "${REX}")
  #MESSAGE(REX_LINK_LIBS= "${REX_LINK_LIBS}")

  SET(XML_CFLAGS "${REX_CFLAGS}" CACHE STRING 
	"XML compilation flags" )
  SET(XML_INCLUDE_DIRS "${REX_INCLUDE_DIRS}" CACHE STRING 
	"Path to find the XML headers" )
  SET(XML_LINK_DIRS "${REX_LINK_DIRS}" CACHE STRING 
	"XML path to find the libs to link against" )
  SET(XML_LINK_LIBS "${REX_LINK_LIBS}" CACHE STRING 
	"XML libraries to link with" )

  MARK_AS_ADVANCED(XML_CFLAGS XML_INCLUDE_DIRS XML_LINK_DIRS XML_LINK_LIBS)
  
  #MESSAGE (XML_CFLAGS="${XML_CFLAGS}")
  #MESSAGE (XML_LINK_DIRS="${XML_LINK_DIRS}")
  #MESSAGE (XML_LINK_LIBS="${XML_LINK_LIBS}")

  SET(XML-2_FOUND ON CACHE INTERNAL "XML-2 found" FORCE)
ELSE (PKGCONFIG)
  SET(XML-2_FOUND OFF CACHE INTERNAL "XML-2 not found" FORCE)
ENDIF (PKGCONFIG)


