#
# Macros for the MIPS Package/Module framework
# these handle the local/prebuilt mechanism
#


MACRO(USEMODULES relative_path_to_root)
  SET(CONFIGFILE ModuleConfig.cmake)
  SET(PREBUILT_SRCDIR $ENV{MIPS_PREBUILT_SRCDIR}/${relative_path_to_root})
  SET(PREBUILT_BUILDDIR $ENV{MIPS_PREBUILT_BUILDDIR}/${relative_path_to_root})

  # load module list
  INCLUDE (${PROJECT_SOURCE_DIR}/ModuleList.cmake)

  SET (PACKAGE_INCLUDE_DIR ${PROJECT_SOURCE_DIR})

  FOREACH(MOD ${ALLMODULES})
    # test if module is present locally
    IF (EXISTS ${PROJECT_SOURCE_DIR}/${MOD}/${CONFIGFILE})

       SET (CURRENT_MODULE_CONFIG_SOURCE_DIR ${PROJECT_SOURCE_DIR}/${MOD})
       SET (CURRENT_MODULE_CONFIG_BINARY_DIR ${PROJECT_BINARY_DIR}/${MOD})
       INCLUDE(${PROJECT_SOURCE_DIR}/${MOD}/${CONFIGFILE})
       SUBDIRS(${MOD})
       # collect include/link dirs
       SET (PACKAGE_INCLUDE_DIR ${PACKAGE_INCLUDE_DIR};
		 ${PROJECT_SOURCE_DIR}/${MOD})
	   IF(NOT LIBRARY_OUTPUT_PATH) #if already set see after the loop
		 SET (PACKAGE_LINK_DIR ${PACKAGE_LINK_DIR};
		   ${PROJECT_BINARY_DIR}/${MOD})
	   ENDIF(NOT LIBRARY_OUTPUT_PATH)

    ELSE (EXISTS ${PROJECT_SOURCE_DIR}/${MOD}/${CONFIGFILE})
       IF (USE_PREBUILT)
          # test if module is available prebuilt
          IF (EXISTS ${PREBUILT_SRCDIR}/${MOD}/${CONFIGFILE})

              SET (CURRENT_MODULE_CONFIG_SOURCE_DIR ${PREBUILT_SRCDIR}/${MOD})
              SET (CURRENT_MODULE_CONFIG_BINARY_DIR ${PREBUILT_BUILDDIR}/${MOD})
              INCLUDE(${PREBUILT_SRCDIR}/${MOD}/${CONFIGFILE})
              # collect include/link dirs
              SET (PACKAGE_INCLUDE_DIR ${PACKAGE_INCLUDE_DIR};
				${PREBUILT_SRCDIR}/${MOD})
              SET (PACKAGE_PREBUILD_LINK_DIR ${PACKAGE_PREBUILD_LINK_DIR};
				${PREBUILT_BUILDDIR}/${MOD})

          ELSE (EXISTS ${PREBUILT_SRCDIR}/${MOD}/${CONFIGFILE})
             MESSAGE ("Module ${MOD} not found in ${PROJECT_SOURCE_DIR} and ${PREBUILT_SRCDIR}.")
          ENDIF (EXISTS ${PREBUILT_SRCDIR}/${MOD}/${CONFIGFILE})
       ENDIF (USE_PREBUILT)
    ENDIF (EXISTS  ${PROJECT_SOURCE_DIR}/${MOD}/${CONFIGFILE})
  ENDFOREACH(MOD)

  # Concatenate PACKAGE_LINK_DIR and PACKAGE_PREBUILD_LINK_DIR
  # if LIBRARY_OUTPUT_PATH si already defined PACKAGE_LINK_DIR contains it only.
  IF(LIBRARY_OUTPUT_PATH)
	SET (PACKAGE_LINK_DIR ${LIBRARY_OUTPUT_PATH};
	  ${PACKAGE_PREBUILD_LINK_DIR})
  ELSE(LIBRARY_OUTPUT_PATH)
	SET (PACKAGE_LINK_DIR ${PACKAGE_LINK_DIR};
	  ${PACKAGE_PREBUILD_LINK_DIR})
  ENDIF(LIBRARY_OUTPUT_PATH)

  # use PACKAGE_INCLUDE_DIR and PACKAGE_LINK_DIR for this package
  INCLUDE_DIRECTORIES(${PACKAGE_INCLUDE_DIR})
  LINK_DIRECTORIES(${PACKAGE_LINK_DIR})

ENDMACRO(USEMODULES)

#
# Copy the given file in the given directory (absolute path). The directory is created 
# if not already done.
MACRO(MIPS_COPY_IN_DIRECTORY _dir _file)
  IF(EXISTS _dir)
  ELSE(EXISTS _dir)
    MAKE_DIRECTORY(_dir)
  ENDIF(EXISTS _dir)
  ADD_CUSTOM_TARGET(
    ${_file}
    ALL
    ${CMAKE_COMMAND}
    -E copy ${CMAKE_CURRENT_SOURCE_DIR}/${_file} ${_dir}/${_file}
    DEPENDS  ${CMAKE_CURRENT_SOURCE_DIR}/${_file}
    )
ENDMACRO(MIPS_COPY_IN_DIRECTORY)

#
# Copy the given file from the source directory to 
# the EXECUTABLE_OUTPUT_PATH if it is defined.
#
MACRO(MIPS_COPY_AS_EXECUTABLE _file)
  IF(SHELLS_OUTPUT_PATH)
    MIPS_COPY_IN_DIRECTORY(${SHELLS_OUTPUT_PATH} ${_file})
  ENDIF(SHELLS_OUTPUT_PATH)
ENDMACRO(MIPS_COPY_AS_EXECUTABLE)




# 
# Macro to wrap c++ classes into tcl.
# It uses VTK wrapping, so you need to do :
# 
# FIND_PACKAGE(VTK)
# IF(VTK_FOUND)
#	INCLUDE(${VTK_USE_FILE})
# ELSE(VTK_FOUND)
#	MESSAGE(FATAL_ERROR "VTK not found. Please set VTK_DIR.")
# ENDIF(VTK_FOUND)
# 
# before using it.
# 
#

MACRO(MIPS_WRAP_TCL target src_list_name sources commands)

  IF(VTK_MAJOR_VERSION AND VTK_MAJOR_VERSION LESS 5)
      
    # Attempt at supporting 4.4. 
    # This won't work anyway, as the lexer in VTK 4.x was updated later on
    # to support export macros not starting with VTK_ 
    VTK_WRAP_TCL2(
      ${target}
      SOURCES ${src_list_name} ${sources}
      COMMANDS ${commands})
    
  ELSE(VTK_MAJOR_VERSION AND VTK_MAJOR_VERSION LESS 5)
    
    IF(NOT VTK_CMAKE_DIR)
      SET(VTK_CMAKE_DIR "${VTK_SOURCE_DIR}/CMake")
    ENDIF(NOT VTK_CMAKE_DIR)
    INCLUDE("${VTK_CMAKE_DIR}/vtkWrapTcl.cmake")
    
    # VTK 5.0 really need some help. The evil VTK_WRAP_TCL3_INIT_DIR hack was
    # fixed on the 5.0 branch but people have downloaded earlier 5.0 as well.
    # Furthermore, old 5.0 can not be used once it has been installed, since
    # vtkWrapperInit.data.in is not installed properly: report that sad fact.

    IF("${VTK_MAJOR_VERSION}.${VTK_MINOR_VERSION}" EQUAL "5.0")
      IF(VTK_INSTALL_PREFIX)
        IF(NOT EXISTS "${VTK_CMAKE_DIR}/vtkWrapperInit.data.in")
          MESSAGE("Sorry, you are using a VTK 5.0 that can not be used properly once it has been installed. You can either download a more recent VTK 5.0 snapshot from the CVS repository, or simply point your project to your VTK build directory instead of your VTK install directory.")
        ENDIF(NOT EXISTS "${VTK_CMAKE_DIR}/vtkWrapperInit.data.in")
      ELSE(VTK_INSTALL_PREFIX)
        SET(VTK_WRAP_TCL3_INIT_DIR "${VTK_SOURCE_DIR}/Wrapping")
        SET(VTK_WRAP_PYTHON3_INIT_DIR "${VTK_SOURCE_DIR}/Wrapping")
      ENDIF(VTK_INSTALL_PREFIX)
    ENDIF("${VTK_MAJOR_VERSION}.${VTK_MINOR_VERSION}" EQUAL "5.0")

    VTK_WRAP_TCL3(
      ${target}
      ${src_list_name} "${sources}" 
      "${commands}" 
      ${ARGN})
    
  ENDIF(VTK_MAJOR_VERSION AND VTK_MAJOR_VERSION LESS 5)

ENDMACRO(MIPS_WRAP_TCL)
