
/*
* Written by: Thomas Yeo
* 
* Example usage:
* CreateTensorImages dim1 dim2 dim3 filename
* 
*/


#include "itkImageFileWriter.h"
#include "itkImage.h"
#include "itkTensor.h"
#include "itkImageRegionIterator.h"
#include "itkTensorImageIO.h"

#include <getopt.h>
#include <iostream>
#include <string>
#include "Debug.h"
#include "UsefulFunctions.h"
#include "UsefulTensorFunctions.h"
#include "itkNumericTraitsTensorPixel2.h" 

#define tensor_dim 3

int main( int argc, char *argv[] )
{   

  if(argc < 5)
    {
    std::cout<<"Usage: CreateTensorImages dim1 dim2 dim3 DT1a DT2a DT3a DT1b DT2b DT3b LB1 LB2 LB3 UB1 UB2 UB3 PADDING filename"<<std::endl;
    exit(EXIT_FAILURE);
    }


  //Read in Arguments
  int dim1   = atoi(argv[1]);
  int dim2   = atoi(argv[2]);
  int dim3   = atoi(argv[3]);
  

  const std::string filename = argv[17];

  std::cout<<"image dimension: "<<dim1<<"x"<<dim2<<"x"<<dim3<<std::endl;
  std::cout<<"Output file: "<<filename<<std::endl;

  //Allocate memory for image
  typedef itk::Tensor<float, tensor_dim> TensorType;
  typedef itk::Image<TensorType, 3> ImageType;
  ImageType::Pointer image = ImageType::New();

  ImageType::IndexType start;
  start[0] = 0;
  start[1] = 0;
  start[2] = 0;

  ImageType::SizeType size;
  size[0] = dim1;
  size[1] = dim2;
  size[2] = dim3;

  ImageType::RegionType region;
  region.SetSize(size);
  region.SetIndex(start);

  image->SetRegions(region);
  image->Allocate();

  //Set up 2 kinds of tensors
  TensorType DT; //center tensor

  DT[0] = atof(argv[4]);//0.2;
  DT[3] = 0;
  DT[5] = atof(argv[6]);//1;
  DT[1] = 0;
  DT[2] = atof(argv[5]);//0.2;
  DT[4] = 0;
  
  TensorType DTr; //edge tensor
  DTr[0] = atof(argv[7]);//0.5;
  DTr[3] = 0;
  DTr[5] = atof(argv[9]);//0.5;
  DTr[1] = 0;
  DTr[2] = atof(argv[8]);//0.5;
  DTr[4] = 0;

  std::cout<<"DT: "<<DT<<std::endl;
  std::cout<<"DTr: "<<DTr<<std::endl;


  typedef ImageType::IndexType IndexType;

  IndexType FirstIndex = image->GetLargestPossibleRegion().GetIndex();
  IndexType LastIndex = image->GetLargestPossibleRegion().GetIndex() + 
    image->GetLargestPossibleRegion().GetSize();

  std::cout<<"FirstIndex: "<<FirstIndex<<std::endl;
  std::cout<<"LastIndex: "<<LastIndex<<std::endl;
  
  IndexType index;

  //Fill the image!
  typedef itk::ImageRegionIterator< ImageType> IteratorType;
  IteratorType It = IteratorType(image, image->GetLargestPossibleRegion());
  for(It.GoToBegin(); !It.IsAtEnd(); ++It)
    {
    index = It.GetIndex();

    if( (index[0] <= FirstIndex[0]+ atoi(argv[16])) || 
	(index[1] <= FirstIndex[1]+ atoi(argv[16])) || 
	(index[2] <= FirstIndex[2]+ atoi(argv[16])) || 
	(index[0] >= LastIndex[0] - atoi(argv[16])-1)  || 
	(index[1] >= LastIndex[1] - atoi(argv[16])-1)  || 
	(index[2] >= LastIndex[2] - atoi(argv[16])-1)  )
      {//Padding 
      It.Set(itk::NumericTraits<TensorType>::Zero); 
      //It.Set(DT); 
      }
    else if( (index[0] > FirstIndex[0] + atoi(argv[10])) && 
	(index[1] > FirstIndex[1] + atoi(argv[11])) &&
	(index[2] > FirstIndex[2] + atoi(argv[12])) && 
	(index[0] < FirstIndex[0] + atoi(argv[13]))  && 
	(index[1] < FirstIndex[1] + atoi(argv[14]))  && 
	(index[2] < FirstIndex[2] + atoi(argv[15]))) 
      {
      It.Set(DT); //middle
      }
    else
      {
      It.Set(DTr); //edge

      }
    }

  yeoWriteTensorImage<ImageType>(image, filename);
  
  

  //This part just checks whether we can re-read the image and it's
  //the same as before.
  ImageType::Pointer image2 = ImageType::New();
  yeoReadTensorImage<ImageType>(image2, filename);
  
  IteratorType It2 = IteratorType(image2, image2->GetLargestPossibleRegion());

  TensorType DT2;
  TensorType DT3;

  for(It.GoToBegin(), It2.GoToBegin(); !It.IsAtEnd(); ++It, ++It2)
    {
    DT = It.Get();
    DT2 = It2.Get();
    DT3 = DT2 - DT;
    if(DT3.GetNorm() > 0)
      {
      std::cerr<<"Difference is Greater than 0: "<<DT3.GetNorm()<<std::endl;
      exit(EXIT_FAILURE);
      }
    }


  std::cout<<"Image creation success!"<<std::endl;
  exit(EXIT_SUCCESS);
}

