#include "itkMultiResolutionPDEDeformableRegistration2Tensor.h"
#include "itkDiffeomorphicDemonsRegistrationFilterTensor.h"
#include "itkFastSymmetricForcesDemonsRegistrationFilter.h"
#include "itkWarpImageFilter.h"
#include "itkWarpTensorImageFilter.h"
#include "itkCommand.h"
/**
   Since ITK 3.8, this filter has been added in global ITK code,
   and it has changed of name,
   please switch the lines if any problems.
*/
// #include "itkWarpJacobianDeterminantFilter.h"
#include "itkDisplacementFieldJacobianDeterminantFilterOld.h"
#include "itkMinimumMaximumImageCalculator.h"
#include "itkWarpSmoothnessCalculator.h"
#include "itkGridForwardWarpImageFilter.h"
#include "itkVectorCentralDifferenceImageFunction.h"
#include "itkNumericTraitsTensorPixel2.h"

#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"
#include "Debug.h"
#include "UsefulFunctions.h"
#include "UsefulTensorFunctions.h"
#include "itkHistogramMatchingImageFilter.h"


#include <getopt.h>
#include <iostream>
#include <fstream>
#include <time.h>
//#define DEBUG

//#define NUM_ITER_PER_UPDATE 2

struct arguments
{
   std::string  fixedImageFile;  /* -f option */
   std::string  movingImageFile; /* -m option */
   std::string  outputImageFile; /* -o option */
   std::string  outputFieldFile; /* -O option */
   std::string  trueFieldFile;   /* -r option */
  std::string   resultsFile; /*-k option */
   unsigned int numLevels;       /* -n option */
   std::vector<unsigned int> numIterations;   /* -i option */
   float sigmaDef;               /* -s option */
   float sigmaUp;                /* -g option */
   float maxStepLength;          /* -l option */
   bool useVanillaSym;           /* -a option */
   unsigned int gradientType;    /* -t option */
  unsigned int gradientType2;    /* -j option */
  unsigned int gradientType3;    /* -c option */
   bool useHistogramMatching;    /* -e option */
   unsigned int verbosity;       /* -v option */
  bool eucBool;            /* -p option */
  std::string atlasVarFile;      /* -b option */
  bool useAtlas;                 /* -b option */
  int dim;                       /* -d option */

   arguments () :
      fixedImageFile(""),
      movingImageFile(""),
      outputImageFile("output.vtk"),
      outputFieldFile(""),
      trueFieldFile(""),
      resultsFile(""),
      numLevels(3u),
      sigmaDef(3.0f),
      sigmaUp(0.0f),
      maxStepLength(2.0f),
      useVanillaSym(false),
      gradientType(2u),
      gradientType2(0u),
      gradientType3(0u),
      useHistogramMatching(false),
      verbosity(0u),
      eucBool(false),
      atlasVarFile(""),
      useAtlas(0u),
      //maxOptIterations(100u),
      //minAccurarcy(0.01f),
      dim(3u)
   {
      numIterations = std::vector<unsigned int>(numLevels, 10u);
   }

   friend std::ostream& operator<< (std::ostream& o, const arguments& args)
   {
      std::ostringstream osstr;
      for (unsigned int i=0; i<args.numIterations.size(); ++i)
         osstr<<args.numIterations[i]<<" ";
      std::string iterstr = "[ " + osstr.str() + "]";

      std::string gtypeStr;
      switch (args.gradientType)
      {
      case 0:
	gtypeStr = "Symmetrized";
	break;
      case 1:
	gtypeStr = "Fixed";
	break;
      case 2:
	gtypeStr = "Moving";
	break;
      default:
	gtypeStr = "unsupported";
      }

      std::string gtypeStr2;
      switch (args.gradientType2)
      {
      case 0:
	gtypeStr2 = "Rotation";
	break;
      case 1:
	gtypeStr2 = "Without Rotation";
	break;
      case 2:
	gtypeStr2 = "Rotation Approximation";
	break;
      default:
	gtypeStr2 = "unsupported";
      }

      std::string gtypeStr3;
      switch (args.gradientType3)
      {
      case 0:
	gtypeStr3 = "Tom";
	break;
      case 1:
	gtypeStr3 = "Xavier";
	break;
      case 2:
	gtypeStr3 = "Conjugate Gradient Line Search";
	break;
      default:
	gtypeStr3 = "unsupported";
      }
        
      return o
         <<"Arguments structure:"<<std::endl
         <<"  Fixed image file: "<<args.fixedImageFile<<std::endl
         <<"  Moving image file: "<<args.movingImageFile<<std::endl
         <<"  Output image file: "<<args.outputImageFile<<std::endl
         <<"  Output field file: "<<args.outputFieldFile<<std::endl
         <<"  True field file: "<<args.trueFieldFile<<std::endl
	 <<"  Results file: "<<args.resultsFile<<std::endl
         <<"  Number of multiresolution levels: "<<args.numLevels<<std::endl
         <<"  Number of demons iterations: "<<iterstr<<std::endl
         <<"  Deformation field sigma: "<<args.sigmaDef<<std::endl
         <<"  Update field sigma: "<<args.sigmaUp<<std::endl
         <<"  Maximum update step length: "<<args.maxStepLength<<std::endl
         <<"  Use vanilla symmetric demons: "<<args.useVanillaSym<<std::endl
         <<"  Type of gradient: "<<gtypeStr<<std::endl
	 <<"  Type of gradient2: "<<gtypeStr2<<std::endl
	 <<"  Type of gradient3: "<<gtypeStr3<<std::endl
	//<<"  Max Linear Solver Iterations: "<<args.maxOptIterations<<std::endl
         <<"  Use histogram matching: "<<args.useHistogramMatching<<std::endl
         <<"  Verbosity: "<<args.verbosity<<std::endl
	 <<"  Euclidean Bool: "<<args.eucBool<<std::endl
	 <<"  Var image file: "<<args.atlasVarFile<<std::endl
	 <<"  Use Atlas Bool: "<<args.useAtlas<<std::endl
         <<" Image Dimension: "<<args.dim<<std::endl;
   }
};

static const char *optString = "f:m:o:O:r:k:n:i:s:g:l:at:j:c:epv::b:d:h?";

static const struct option longOpts[] = {
   { "fixed-image", required_argument, NULL, 'f' },
   { "moving-image", required_argument, NULL, 'm' },
   { "output-image", required_argument, NULL, 'o' },
   { "output-field", optional_argument, NULL, 'O' },
   { "true-field", required_argument, NULL, 'r' },
   { "result-file", required_argument, NULL, 'k' },
   { "num-levels", required_argument, NULL, 'n' },
   { "num-iterations", required_argument, NULL, 'i' },
   { "def-field-sigma", required_argument, NULL, 's' },
   { "up-field-sigma", required_argument, NULL, 'g' },
   { "max-step-length", required_argument, NULL, 'l' },
   { "use-vanilla-sym", no_argument, NULL, 'a' },
   { "gradient-type", required_argument, NULL, 't' },
   { "gradient-type2", required_argument, NULL, 'j' },
   { "gradient-type3", required_argument, NULL, 'c' },
   //{ "max-optimization-iterations", required_argument, NULL, 'k'},
   { "use-histogram-matching", no_argument, NULL, 'e' },
   { "verbose", optional_argument, NULL, 'v' },
   { "variance-image", required_argument, NULL, 'b'},
   { "image-dimension", required_argument, NULL, 'd'},
   { "help", no_argument, NULL, 'h' },
   { NULL, no_argument, NULL, 0 }
};

/* Display program usage, and exit.
 */
void display_usage( const std::string progname )
{   
   struct arguments defargs = arguments();

   std::ostringstream osstr;
   for (unsigned int i=0; i<defargs.numIterations.size(); ++i)
      osstr<<defargs.numIterations[i]<<" ";
   std::string iterstr = "[ " + osstr.str() + "]";
   
   std::cout<<std::endl;
   std::cout<<progname<<" - register an image and an atlas using demons algorithm."<<std::endl; 
   std::cout<<"If no variance-image set, then program will treat this as pair-wise registration."<<std::endl;
   std::cout<<"Note that fixed-image is assumed to be atlas mean image, so that interpolation is not done on the atlas."<<std::endl;
   std::cout<<"Usage: "<<progname<<" [OPTION...]"<<std::endl;
   
   std::cout<<"  -f/--fixed-image=STRING    Fixed image filename - mandatory"<<std::endl;
   std::cout<<"  -m/--moving-image=STRING   Moving image filename - mandatory"<<std::endl;
   std::cout<<"  -o/--output-image=STRING   Output image filename - default: "<<defargs.outputImageFile<<std::endl;
   std::cout<<"  -O/--output-field(=STRING) Output field filename - default: OUTPUTIMAGENAME-field.mha"<<std::endl;
   std::cout<<"  -r/--true-field=STRING     True field filename - default: not used"<<std::endl;
   std::cout<<"  -k/--results-file=STRING     Results filename - default: not used"<<std::endl;								  
   std::cout<<"  -n/--num-levels=UINT       number of multi-scale levels - default: "<<defargs.numLevels<<std::endl;
   std::cout<<"  -i/--num-iterations=UINTx...xUINT   number of demons iterations - default: "<<iterstr<<std::endl;
   std::cout<<"  -s/--def-field-sigma=FLOAT Smoothing sigma for the deformation field"<<std::endl
            <<"                             at each iteration - default: "<<defargs.sigmaDef<<std::endl;
   std::cout<<"  -g/--up-field-sigma=FLOAT  Smoothing sigma for the update field"<<std::endl
            <<"                             at each iteration - default: "<<defargs.sigmaUp<<std::endl;
   std::cout<<"  -l/--max-step-length=FLOAT Maximum length of an update vector"<<std::endl
            <<"                             (0: no restriction) - default: "<<defargs.maxStepLength<<std::endl;
   std::cout<<"  -a/--use-vanilla-sym       Use the vanilla symmetric forces demons"<<std::endl
            <<"                             instead of the exponential one"<<std::endl;
   std::cout<<"  -t/--gradient-type=UINT    type of gradient used for computing the demons force"<<std::endl
            <<"                             (0 is symmetrized, 1 is Fixed, 2 is Moving (2 default)"<<std::endl; 
   std::cout<<"  -j/--gradient-type=UINT    type of gradient used for computing the demons force"<<std::endl
            <<"                             (0 is rotation, 1 is no-rotation, 2 is approximate rotation (0 default)"<<std::endl; 
   std::cout<<"  -c/--gradient-type=UINT    Hessian Type"<<std::endl
            <<"                             (0 is Tom, 1 is Xavier, 2 is Conjugate Gradient Line Search (0 default)"<<std::endl; 

   std::cout<<"  -e/--use-histogram-matching  Use histogram matching (e.g. for different MRs)"<<std::endl;
   std::cout<<"  -p/--Euc bool                Use Euclidean instead of Log-Euclidean Metric"<<std::endl;
   std::cout<<"  -v/--verbose(=UINT)        Verbosity - default: "<<defargs.verbosity<<"; without argurment: 1"<<std::endl;
   std::cout<<"  -b/--variance-image=STRING  Atlas variance image - default: not used"<<std::endl;
   std::cout<<"  -d/--image-dimension=UINT   Dimension of Tensor Image - default: "<<defargs.dim<<std::endl;

   std::cout<<"  -h/--help                  Display this message and exit"<<std::endl;

   std::cout<<std::endl;
   std::cout<<"Copyright (c) 2006 Mauna Kea Technologies."<<std::endl;
   std::cout<<"Code: Tom Vercauteren."<<std::endl;
   std::cout<<"Report bugs to <tom@maunakeatech.com>."<<std::endl;
   std::cout<<"Copyright (c) 2006 Mauna Kea Technologies."<<std::endl;
   std::cout<<"Code: Adapted from Tom Vercauteren's code on Diffeomorphic Demons by:"<<std::endl;
   std::cout<<"                                                           Thomas Yeo, CSAIL, MIT @ Asclepios, INRIA"<<std::endl;
   std::cout<<"Report bugs to <ythomas@mit.edu>."<<std::endl;
   
	exit( EXIT_FAILURE );
}



std::vector<unsigned int> parseUIntVector( const std::string & str)
{
   std::vector<unsigned int> vect;
   
   std::string::size_type crosspos = str.find('x',0);

   if (crosspos == std::string::npos)
   {
      // only one uint
      vect.push_back( static_cast<unsigned int>( atoi(str.c_str()) ));
      return vect;
   }

   // first uint
   vect.push_back( static_cast<unsigned int>(
                      atoi( (str.substr(0,crosspos)).c_str()  ) ));

   while(true)
   {
      std::string::size_type crossposfrom = crosspos;
      crosspos =  str.find('x',crossposfrom+1);

      if (crosspos == std::string::npos)
      {
         vect.push_back( static_cast<unsigned int>(
                            atoi( (str.substr(crossposfrom+1,str.length()-crossposfrom-1)).c_str()  ) ));
         return vect;
      }

      vect.push_back( static_cast<unsigned int>(
                         atoi( (str.substr(crossposfrom+1,crosspos)).c_str()  ) ));
   }
}



void parseOpts (int argc, char **argv, struct arguments & args)
{
  //itk::Systemfilesystem::path progpath(argv[0]);
   const std::string progname( "DemonsRegistration" );
   
   // Default values.
   args = arguments();

   std::vector<unsigned int> defiter = args.numIterations;
   args.numIterations.clear();

   if (argc == 1)
   {
      display_usage(progname);
   }
   
   int opt = 0; /* it's actually going to hold a char */
   int longIndex = 0;

   while ( (opt = getopt_long(argc, argv, optString, longOpts, &longIndex)) != -1 )
   {
      switch( opt ) {
      case 'f':
         if (! optarg) display_usage(progname);
         args.fixedImageFile = optarg;
         break;
			
      case 'm':
         if (! optarg) display_usage(progname);
         args.movingImageFile = optarg;
         break;

      case 'o':
         if (! optarg) display_usage(progname);
         args.outputImageFile = optarg;
         break;

      case 'O':
         if (! optarg) args.outputFieldFile = "CHANGETHISSTRING";
         else args.outputFieldFile = optarg;
         break;

      case 'r':
         if (! optarg) display_usage(progname);
         else args.trueFieldFile = optarg;
         break;

      case 'k':
	if (! optarg) display_usage(progname);
	else args.resultsFile = optarg;
	break;

      case 'n':
         if (! optarg) display_usage(progname);
         args.numLevels = static_cast<unsigned int>( atoi(optarg) );
         break;

      case 'i':
         if (! optarg) display_usage(progname);
         args.numIterations = parseUIntVector(std::string(optarg));
         break;
         
      case 's':
         if (! optarg) display_usage(progname);
         args.sigmaDef = atof(optarg);
         if ( args.sigmaDef<0.1 && args.sigmaDef>0.0 )
         {
            std::cout<<"Sigma is too small (min=0.1). We set it to 0.0 (no smoothing)."
                     <<std::endl<<std::endl;
            args.sigmaDef = 0.0;
         }
         break;

      case 'g':
         if (! optarg) display_usage(progname);
         args.sigmaUp = atof(optarg);
         if ( args.sigmaUp<0.5 && args.sigmaUp>0.0 )
         {
             std::cout<<"Sigma is too small (min=0.5). We set it to 0.0 (no smoothing)."
                 <<std::endl<<std::endl;
             args.sigmaUp = 0.0;
         }
         break;

      case 'l':
         if (! optarg) display_usage(progname);
         args.maxStepLength = atof(optarg);
         break;

      case 'a':
         args.useVanillaSym = true;
         break;

      case 't':
         if (! optarg) display_usage(progname);
         args.gradientType = static_cast<unsigned int>( atoi(optarg) );
         break;

      case 'j':
	if (! optarg) display_usage(progname);
	args.gradientType2 = static_cast<unsigned int>( atoi(optarg) );
	break;

      case 'c':
	if (! optarg) display_usage(progname);
	args.gradientType3 = static_cast<unsigned int>( atoi(optarg) );
	break;

      case 'e':
         args.useHistogramMatching = true;
         break;

      case 'p':
	args.eucBool = true;
	break;
		
      case 'b':
	args.useAtlas = true;
	if (! optarg) display_usage(progname);
	args.atlasVarFile = optarg;
	break;

      case 'd':
	args.dim = static_cast<unsigned int>( atoi(optarg) );
	break;

      case 'v':
         if (! optarg) args.verbosity++;
         else args.verbosity = static_cast<unsigned int>( atoi(optarg) );
         break;
			
      case 'h':	/* fall-through is intentional */
      case '?':   /* fall-through is intentional */
      default:
         display_usage(progname);
         break;
		}
   }

   if ( args.outputFieldFile=="CHANGETHISSTRING" )
   {
      
      unsigned int pos = args.outputImageFile.find(".");
      if ( pos < args.outputFieldFile.size() )
      {
         args.outputFieldFile = args.outputImageFile;
         args.outputFieldFile.replace(pos, args.outputFieldFile.size(), "-field.mha");
      }
      else
      {
         args.outputFieldFile = args.outputImageFile + "-field.mha";
      }
         
   }

   if ( args.numLevels ==0)
   {
      std::cout<<"The number of levels should be at least one."<<std::endl;
      display_usage(progname);
   }
   if ( args.numIterations.empty() )
   {
      // set a default number of iterations per level
      args.numIterations = std::vector<unsigned int>(args.numLevels, defiter[0]);
   }
   else if ( args.numLevels != args.numIterations.size() )
   {
      std::cout<<"The number of levels and the number of iterations do not match."<<std::endl;
      display_usage(progname);
   }

   if ( args.gradientType > 2 )
   {
      std::cout<<"The gradient type should be 0 to 2."<<std::endl;
      display_usage(progname);
   }

   if ( args.gradientType2 > 2 )
     {
     std::cout<<"The gradient type should be 0 to 2."<<std::endl;
     display_usage(progname);
     }
   
   if ( args.gradientType3 > 2 )
     {
     std::cout<<"The gradient type should be 0 to 2."<<std::endl;
     display_usage(progname);
     }
}



//  The following section of code implements a Command observer
//  that will monitor the evolution of the registration process.
//

template <class TPixel=float, unsigned int VImageDimension=3, unsigned int TensorDim = 3>
class CommandIterationUpdate : public itk::Command 
{
public:
   typedef  CommandIterationUpdate   Self;
   typedef  itk::Command             Superclass;
   typedef  itk::SmartPointer<Self>  Pointer;

  typedef itk::Image< itk::Tensor<TPixel, TensorDim>, VImageDimension > TensorImageType;
   typedef itk::Image< float, VImageDimension > InternalImageType;
   typedef itk::Vector< TPixel, VImageDimension >    VectorPixelType;
   typedef itk::Image<  VectorPixelType, VImageDimension > DeformationFieldType;

    typedef itk::DiffeomorphicDemonsRegistrationFilterTensor<
       TensorImageType,
       TensorImageType,
       DeformationFieldType>   DiffeomorphicDemonsRegistrationFilterType;


    typedef itk::FastSymmetricForcesDemonsRegistrationFilter<
       TensorImageType,
       TensorImageType,
       DeformationFieldType>   FastSymmetricForcesDemonsRegistrationFilterType;

    typedef itk::MultiResolutionPDEDeformableRegistration2Tensor<
       TensorImageType, TensorImageType,
       DeformationFieldType, TPixel>   MultiResRegistrationFilterType;

//     typedef itk::WarpJacobianDeterminantFilter<
    typedef itk::DisplacementFieldJacobianDeterminantFilterOld<
       DeformationFieldType, InternalImageType> JacobianFilterType;

   typedef itk::MinimumMaximumImageCalculator<InternalImageType> MinMaxFilterType;

    typedef itk::WarpSmoothnessCalculator<DeformationFieldType>
       SmoothnessCalculatorType;

    typedef itk::VectorCentralDifferenceImageFunction<DeformationFieldType>
       WarpGradientCalculatorType;

    typedef typename WarpGradientCalculatorType::OutputType WarpGradientType;

    itkNewMacro( Self );

 private:
    std::ofstream m_Fid;
    bool m_headerwritten;
  typename JacobianFilterType::Pointer m_JacobianFilter;
  typename MinMaxFilterType::Pointer m_Minmaxfilter;
   typename SmoothnessCalculatorType::Pointer m_SmothnessCalculator;
   typename DeformationFieldType::ConstPointer m_TrueField;
   typename WarpGradientCalculatorType::Pointer m_TrueWarpGradientCalculator;
   typename WarpGradientCalculatorType::Pointer m_CompWarpGradientCalculator;

public:
   void SetTrueField(const DeformationFieldType * truefield)
   {
      m_TrueField = truefield;

      m_TrueWarpGradientCalculator = WarpGradientCalculatorType::New();
      m_TrueWarpGradientCalculator->SetInputImage( m_TrueField );

      m_CompWarpGradientCalculator =  WarpGradientCalculatorType::New();
   }
   
   void Execute(itk::Object *caller, const itk::EventObject & event)
   {
#ifdef DEBUG
      std::cout<<"UnConst Observer"<<std::endl;
#endif
      Execute( (const itk::Object *)caller, event);
   }

   void Execute(const itk::Object * object, const itk::EventObject & event)
   {
      if( !(itk::IterationEvent().CheckEvent( &event )) )
      {
         return;
      }
#ifdef DEBUG
      std::cout<<"Const Observer"<<std::endl;
#endif
      typename DeformationFieldType::ConstPointer deffield = 0;
      unsigned int iter = -1;
      double metricbefore = -1.0;
      double normalizedmetric = -1.0;
      unsigned long numberOfPixelsProcessed = 0;

      if ( const DiffeomorphicDemonsRegistrationFilterType * filter = 
           dynamic_cast< const DiffeomorphicDemonsRegistrationFilterType * >( object ) )
      {
         iter = filter->GetElapsedIterations() - 1;

         metricbefore = filter->GetMetric();
         deffield = const_cast<DiffeomorphicDemonsRegistrationFilterType *>
            (filter)->GetDeformationField();
	 normalizedmetric = filter->GetNormalizedMetric();
	 numberOfPixelsProcessed = filter->GetNumberOfPixelsProcessed();

	 
      }
      else if ( const FastSymmetricForcesDemonsRegistrationFilterType * filter = 
           dynamic_cast< const FastSymmetricForcesDemonsRegistrationFilterType * >( object ) )
      {
      std::cerr<<"Not implemented for Fast Symmetric Demons"<<std::endl;
      exit(EXIT_FAILURE);
      std::cerr<<filter<<std::endl;
      }
      else if ( const MultiResRegistrationFilterType * multiresfilter = 
           dynamic_cast< const MultiResRegistrationFilterType * >( object ) )
      {
         std::cout<<"Finished Multi-resolution iteration :"<<multiresfilter->GetCurrentLevel()-1<<std::endl;
         std::cout<<"=============================="<<std::endl<<std::endl;
      }
      else
      {
         return;
      }

	  
      if (deffield)
	{
	
	
	std::cout<<iter<<": MSE "<<metricbefore<<" - ";
	
	
	std::cout<<": NormalizedMSE "<<normalizedmetric<<" - ";
	std::cout<<": NumPixels "<<numberOfPixelsProcessed<<" - ";
	
	double fieldDist = -1.0;
	double fieldGradDist = -1.0;
	double tmp;
	if (m_TrueField)
	  {
	  typedef itk::ImageRegionConstIteratorWithIndex<DeformationFieldType>
	    FieldIteratorType;
	  FieldIteratorType currIter(
	    deffield, deffield->GetLargestPossibleRegion() );
	  FieldIteratorType trueIter(
	    m_TrueField, deffield->GetLargestPossibleRegion() );
	  
	  m_CompWarpGradientCalculator->SetInputImage( deffield );
	  
	  fieldDist = 0.0;
	  fieldGradDist = 0.0;
	  for ( currIter.GoToBegin(), trueIter.GoToBegin();
		! currIter.IsAtEnd(); ++currIter, ++trueIter )
	    {
	    fieldDist += (currIter.Value() - trueIter.Value()).GetSquaredNorm();
	    
	    // No need to add Id matrix here as we do a substraction
	    tmp = (
	      ( m_CompWarpGradientCalculator->EvaluateAtIndex(currIter.GetIndex())
		-m_TrueWarpGradientCalculator->EvaluateAtIndex(trueIter.GetIndex())
		).GetVnlMatrix() ).frobenius_norm();
	    fieldGradDist += tmp*tmp;
	    }
	  fieldDist = sqrt( fieldDist/ (double)(
			      deffield->GetLargestPossibleRegion().GetNumberOfPixels()) );
	  fieldGradDist = sqrt( fieldGradDist/ (double)(
				  deffield->GetLargestPossibleRegion().GetNumberOfPixels()) );
	  
	  std::cout<<"d(.,true) "<<fieldDist<<" - ";
	  std::cout<<"d(.,Jac(true)) "<<fieldGradDist<<" - ";
	  }
	
	m_SmothnessCalculator->SetImage( deffield );
	m_SmothnessCalculator->Compute();
	const double harmonicEnergy = m_SmothnessCalculator->GetSmoothness();
	std::cout<<"harmo. "<<harmonicEnergy<<" - ";
	
	
	m_JacobianFilter->SetInput( deffield );
	m_JacobianFilter->UpdateLargestPossibleRegion();
	
	
	const unsigned int numPix = m_JacobianFilter->
	  GetOutput()->GetLargestPossibleRegion().GetNumberOfPixels();
	
	TPixel* pix_start = m_JacobianFilter->GetOutput()->GetBufferPointer();
	TPixel* pix_end = pix_start + numPix;
	
	TPixel* jac_ptr;
	
	// Get percentage of det(Jac) below 0
	unsigned int jacBelowZero(0u);
	for (jac_ptr=pix_start; jac_ptr!=pix_end; ++jac_ptr)
	  {
	  if ( *jac_ptr<=0.0 ) ++jacBelowZero;
	  }
	const double jacBelowZeroPrc = static_cast<double>(jacBelowZero)
	  / static_cast<double>(numPix);
	
	
	// Get min an max jac
	
	//std::pair<TPixel*, TPixel*> minmax_res =
	//   boost::minmax_element(pix_start, pix_end);
	
	
	//const double minJac = *(minmax_res.first);
	//const double maxJac = *(minmax_res.second);
	
	const double minJac = *(std::min_element (pix_start, pix_end));
	const double maxJac = *(std::max_element (pix_start, pix_end));
	
	// Get some quantiles
	// We don't need the jacobian image
	// we can modify/sort it in place
	jac_ptr = pix_start + static_cast<unsigned int>(0.002*numPix);
	std::nth_element(pix_start, jac_ptr, pix_end);
	const double Q002 = *jac_ptr;
	
	jac_ptr = pix_start + static_cast<unsigned int>(0.01*numPix);
	std::nth_element(pix_start, jac_ptr, pix_end);
	const double Q01 = *jac_ptr;
	
	jac_ptr = pix_start + static_cast<unsigned int>(0.99*numPix);
	std::nth_element(pix_start, jac_ptr, pix_end);
	const double Q99 = *jac_ptr;
	
	jac_ptr = pix_start + static_cast<unsigned int>(0.998*numPix);
	std::nth_element(pix_start, jac_ptr, pix_end);
	const double Q998 = *jac_ptr;
	
	
	std::cout<<"max|Jac| "<<maxJac<<" - "
		 <<"min|Jac| "<<minJac<<" - "
		 <<"ratio(|Jac|<=0) "<<jacBelowZeroPrc<<std::endl;
	
	


	ofstream myfile;
	myfile.open("synthetic_stats.txt", ios::app);
	myfile << metricbefore;
	myfile << "\n";
	myfile << harmonicEnergy;
	myfile << "\n";
	myfile << minJac;
	myfile << "\n";
	myfile << maxJac;
	myfile << "\n";	      
	myfile << fieldDist;
	myfile << "\n";
	myfile << fieldGradDist;
	myfile << "\n";
	myfile.close();


	if (this->m_Fid.is_open())
	  {
	  if (! m_headerwritten)
	    {
	    this->m_Fid<<"Iteration"
		       <<", MSE before"
		       <<", Harmonic energy"
		       <<", min|Jac|"
		       <<", 0.2% |Jac|"
		       <<", 01% |Jac|"
		       <<", 99% |Jac|"
		       <<", 99.8% |Jac|"
		       <<", max|Jac|"
		       <<", ratio(|Jac|<=0)";
	    
	    if (m_TrueField)
	      {
	      this->m_Fid<<", dist(warp,true warp)"
			 <<", dist(Jac,true Jac)";
	      }
	    
	    this->m_Fid<<std::endl;
	    
	    m_headerwritten = true;
	    }
	  
	  this->m_Fid<<iter
		     <<", "<<metricbefore
		     <<", "<<harmonicEnergy
		     <<", "<<minJac
		     <<", "<<Q002
		     <<", "<<Q01
		     <<", "<<Q99
		     <<", "<<Q998
		     <<", "<<maxJac
		     <<", "<<jacBelowZeroPrc;
	  
	  if (m_TrueField)
	    {
	    this->m_Fid<<", "<<fieldDist
		       <<", "<<fieldGradDist;
	    }
	  
	  this->m_Fid<<std::endl;
	  }
	}//end if(defField)
      
   } //end of command execute
  
protected:   
   CommandIterationUpdate() :
      m_Fid( "metricvalues.csv" ),
      m_headerwritten(false)
   {
      m_JacobianFilter = JacobianFilterType::New();
      m_JacobianFilter->SetUseImageSpacing( true );
      m_JacobianFilter->ReleaseDataFlagOn();
      
      m_Minmaxfilter = MinMaxFilterType::New();

      m_SmothnessCalculator = SmoothnessCalculatorType::New();

      m_TrueField = 0;
      m_TrueWarpGradientCalculator = 0;
      m_CompWarpGradientCalculator = 0;
   };

   ~CommandIterationUpdate()
   {
      this->m_Fid.close();
   }
};



template <unsigned int Dimension>
void DemonsRegistrationFunction( arguments args )
{
// Declare the types of the images (float or double only)   
  typedef itk::Tensor<float, 3> TensorType;
  typedef itk::Image<TensorType, Dimension> TensorImageType;
  typedef float PixelType;

   // Images we use
   typename TensorImageType::Pointer fixedImage = TensorImageType::New();
   typename TensorImageType::Pointer movingImage = TensorImageType::New();
   typename TensorImageType::Pointer varImage = TensorImageType::New();


#ifdef DEBUG
  typename TensorImageType::IndexType pixelIndex;
  pixelIndex[0] = 64;
  pixelIndex[1] = 64;
  pixelIndex[2] = 15;    
#endif

   //read variance
   if(args.useAtlas)
     {
     typename TensorImageType::Pointer varImage2 = TensorImageType::New();
     if(args.verbosity)
       std::cout<<"Reading variance file: "<<args.atlasVarFile.c_str()<<std::endl;
     
     yeoReadTensorImage<TensorImageType>(varImage2, args.atlasVarFile.c_str());
     yeoCopyImage<TensorImageType>(varImage2, varImage);
     }
  

   //Histogram matching
   if (!args.useHistogramMatching)
     {
     if(args.verbosity)
       {
       std::cout<<"No Histogram Matching"<<std::endl;
       std::cout<<"Read Fixed Image"<<std::endl;
       }
     yeoReadTensorImage<TensorImageType>(fixedImage, args.fixedImageFile.c_str());
     if(!args.eucBool)
       {
       if(args.verbosity)
	 std::cout<<"Taking Log of Fixed Image"<<std::endl;
       yeoLogTensorImageInPlace<TensorImageType>(fixedImage);
       }
     else
       {
       if(args.verbosity)
	 std::cout<<"Do not take log of Fixed Image"<<std::endl;
       }


     if(args.verbosity)
       {
       std::cout<<"Read Moving Image"<<std::endl;
       }
     yeoReadTensorImage<TensorImageType>(movingImage,  args.movingImageFile.c_str());
     if(!args.eucBool)
       {
       if(args.verbosity)
	 std::cout<<"Taking Log of Moving Image"<<std::endl;
       yeoLogTensorImageInPlace<TensorImageType>(movingImage);
       }
     else
       {
       if(args.verbosity)
	 std::cout<<"Do not take log of Moving Image"<<std::endl;
       }
     }
   else
     {   
     if(args.verbosity)
       {
       std::cout<<"Histogram Matching"<<std::endl;
       }
     std::cerr<<"Histogram Matching Not Implemented"<<std::endl;
     }
   
//Print Initial Objective Function Value
   // warp the result
   typedef itk::Vector< PixelType, Dimension >    VectorPixelType;
   typedef typename itk::Image<  VectorPixelType, Dimension > DeformationFieldType;
   typedef typename itk::Image<PixelType, Dimension> MaskImageType; 

   typename MaskImageType::Pointer SSDimage = MaskImageType::New();
   typename MaskImageType::Pointer nSSDimage = MaskImageType::New();
   if(args.verbosity)
     {
     
     
     //create empty deformation field
     typename DeformationFieldType::Pointer defField = DeformationFieldType::New();
     yeoCreateZeroImage<TensorImageType, DeformationFieldType>(fixedImage, defField);


     //warp moving image
     std::cout<<"Warping Image"<<std::endl;
     typename TensorImageType::Pointer warpedMovingImage = TensorImageType::New();
     yeoWarpTensorImageNewImage<TensorImageType, DeformationFieldType>(movingImage, defField, warpedMovingImage);

     //computing mask
     typename MaskImageType::Pointer fixedImage_fg = MaskImageType::New();
     yeoGetForegroundMask<TensorImageType, MaskImageType>(fixedImage, fixedImage_fg, itk::NumericTraits<typename TensorImageType::PixelType>::Zero, 0, 1);
                                                                                     
     typename MaskImageType::Pointer movingImage_fg = MaskImageType::New();
     yeoGetForegroundMask<TensorImageType, MaskImageType>(warpedMovingImage, movingImage_fg, itk::NumericTraits<typename TensorImageType::PixelType>::Zero, 0, 1);
      
     yeoPowerCombine2ImagesInPlace<MaskImageType>(fixedImage_fg, movingImage_fg, 1, 1);

     //compute SSD
     std::cout<<"Computing SSD"<<std::endl;
     yeoLinearCombine2ImagesInPlace<TensorImageType>(warpedMovingImage, fixedImage, 1.0, -1.0);
     yeoPowerCombine2TensorImagesInPlace<TensorImageType>(warpedMovingImage, warpedMovingImage, 1.0, 1.0);


     //apply mask
     yeoMultiplyTensorImageWithImageInPlace<TensorImageType, MaskImageType>(warpedMovingImage, fixedImage_fg);


     double numPixels = yeoSumImage<MaskImageType>(fixedImage_fg);
     if(args.eucBool)
       std::cout<<"Initial Euc MSE fixed image vs warped moving image averaged over "<<numPixels<<" pixels: "<<yeoSumTensorImage<TensorImageType>(warpedMovingImage)/numPixels<<std::endl;
     else
       std::cout<<"Initial Log MSE fixed image vs warped moving image averaged over "<<numPixels<<" pixels: "<<yeoSumTensorImage<TensorImageType>(warpedMovingImage)/numPixels<<std::endl;


     /*
     //write SSD
     yeoSumTensorImageNewImage<TensorImageType, MaskImageType>(warpedMovingImage, SSDimage);
     yeoWriteImage<MaskImageType>(SSDimage, "iniSSD.nhdr");
            
     //normalized SSD
     if(args.useAtlas)
       {
       yeoPowerCombine2TensorImagesInPlace<TensorImageType>(warpedMovingImage, varImage, 1.0, -1.0);
       std::cout<<"Initial Normalized MSE Atlas vs warped moving image averaged over "<<numPixels<<" pixels: "<<yeoSumTensorImage<TensorImageType>(warpedMovingImage)/numPixels<<std::endl;  
       
       //Write SSD     
       yeoSumTensorImageNewImage<TensorImageType, MaskImageType>(warpedMovingImage, nSSDimage);
       yeoWriteImage<MaskImageType>(nSSDimage, "ininSSD.nhdr");
       }
*/
     }
 
   
   // Set up the demons filter output
   typename DeformationFieldType::Pointer defField = 0;

   {//for mem allocations
   
   // Set up the demons filter
   typedef typename itk::PDEDeformableRegistrationFilter
      < TensorImageType, TensorImageType, DeformationFieldType>   BaseRegistrationFilterType;
   typename BaseRegistrationFilterType::Pointer filter;
   
   //my little hack because I could not upcast a template to one with
   //more variables
/*   typedef typename itk::DiffeomorphicDemonsRegistrationFilterTensor
         < TensorImageType, TensorImageType, DeformationFieldType> DiffeomorphicDemonsRegistrationFilterTensorType;
   typename DiffeomorphicDemonsRegistrationFilterTensorType::Pointer myactualfilter = 0;
*/

   if ( !args.useVanillaSym )
   {
   if(args.verbosity)
     {
     std::cout<<"No Vanilla Symmetric Demons"<<std::endl;
     }

      typedef typename itk::DiffeomorphicDemonsRegistrationFilterTensor
         < TensorImageType, TensorImageType, DeformationFieldType>   ActualRegistrationFilterType;
      typedef typename ActualRegistrationFilterType::GradientType GradientType;
      typedef typename ActualRegistrationFilterType::GradientType2 GradientType2;
      typedef typename ActualRegistrationFilterType::GradientType3 GradientType3;
      
      typename ActualRegistrationFilterType::Pointer actualfilter = ActualRegistrationFilterType::New();
/*  
    myactualfilter = actualfilter;
*/
      actualfilter->SetMaximumUpdateStepLength( args.maxStepLength );
      actualfilter->SetUseGradientType( static_cast<GradientType>(args.gradientType) );
      actualfilter->SetUseGradientType2( static_cast<GradientType2>(args.gradientType2) );
      actualfilter->SetUseGradientType3( static_cast<GradientType3>(args.gradientType3) );
      actualfilter->SetUseAtlas(args.useAtlas);
      filter = actualfilter;
   }
   else
   {
   std::cerr<<"Not implemented for Fast Symmetric Vanilla Demons"<<std::endl;
   exit(EXIT_FAILURE);
   }

   if ( args.sigmaDef > 0.1 )
   {
      filter->SmoothDeformationFieldOn();
      filter->SetStandardDeviations( args.sigmaDef );
   }
   else
      filter->SmoothDeformationFieldOff();

   if ( args.sigmaUp > 0.1 )
   {
      filter->SmoothUpdateFieldOn();
      filter->SetUpdateFieldStandardDeviations( args.sigmaUp );
   }
   else
      filter->SmoothUpdateFieldOff();

   //filter->SetIntensityDifferenceThreshold( 0.001 );

 


   if ( args.verbosity > 0 )
   {
   std::cout<<"Set up Observer for verbose"<<std::endl;
      
      // Create the Command observer and register it with the registration filter.
      typename CommandIterationUpdate<PixelType, Dimension, TensorType::Dimension>::Pointer observer =
         CommandIterationUpdate<PixelType, Dimension, TensorType::Dimension>::New();

      if (!args.trueFieldFile.empty())
      {
         if (args.numLevels>1)
         {
            std::cout << "You cannot compare the results with a true filed in a multiresolution setting yet." << std::endl;
            exit( EXIT_FAILURE );
         }
         
         // Set up the file readers
         typedef itk::ImageFileReader< DeformationFieldType > FieldReaderType;
         typename FieldReaderType::Pointer fieldReader = FieldReaderType::New();
         fieldReader->SetFileName(  args.trueFieldFile.c_str() );

         // Update the reader
         try
         {
            fieldReader->Update();
         }
         catch( itk::ExceptionObject& err )
         {
            std::cout << "Could not read the true field." << std::endl;
            std::cout << err << std::endl;
            exit( EXIT_FAILURE );
         }

         observer->SetTrueField( fieldReader->GetOutput() );
      }
      
      filter->AddObserver( itk::IterationEvent(), observer );
   }

 
   // Set up the multi-resolution filter
   if ( args.verbosity > 0 )
     {
     std::cout<<"Set up Multi-resolution Filter"<<std::endl;
     }
   
   
   typedef typename itk::MultiResolutionPDEDeformableRegistration2Tensor<
     TensorImageType, TensorImageType, DeformationFieldType, PixelType >   MultiResRegistrationFilterType;
   typename MultiResRegistrationFilterType::Pointer multires = MultiResRegistrationFilterType::New();
   
   multires->SetRegistrationFilter( filter );
   
   
   //my little hack because I could not upcast a template to one with
   //more variables
  //if( !myactualfilter.IsNull() )
  //   {
  //   multires->SetRegistrationFilter2(myactualfilter);
  //   }


   multires->SetNumberOfLevels( args.numLevels );

   multires->SetNumberOfIterations( &args.numIterations[0] );
   /*
   if(args.gradientType != 0)
     multires->SetNumberOfIterations( &args.numIterations[0] );
   else
     {
     for(unsigned int k = 0; k < args.numLevels; k++)
       args.numIterations[k] *= NUM_ITER_PER_UPDATE;
     
     multires->SetNumberOfIterations( &args.numIterations[0] );

     }
     */

   multires->SetFixedImage( fixedImage );
   multires->SetMovingImage( movingImage );
   
   if(args.useAtlas)
     {       
     multires->SetVarImage(varImage);
     }
   else
     multires->SetVarImage(NULL);
   
   
   
   if ( args.verbosity > 0 )
     {
     std::cout<<"Connect Observer to Multi-Res Filter"<<std::endl;
     // Create the Command observer and register it with the registration filter.
     typename CommandIterationUpdate<PixelType, Dimension, TensorType::Dimension>::Pointer multiresobserver =
       CommandIterationUpdate<PixelType, Dimension, TensorType::Dimension>::New();
     multires->AddObserver( itk::IterationEvent(), multiresobserver );
     }
   
   
   // Compute the deformation field
   try
     {
     std::cout<<"Update MultiRes Filter"<<std::endl;
     multires->UpdateLargestPossibleRegion();
     }
   catch( itk::ExceptionObject& err )
     {
     std::cout << "Unexpected error." << std::endl;
     std::cout << err << std::endl;
     exit( EXIT_FAILURE );
     }
   
   
   // The outputs
   defField = multires->GetOutput();
   defField->DisconnectPipeline();
   
   }//end for mem allocations
   
 

  
   // warp the result
   typedef itk::WarpTensorImageFilter<TensorImageType, TensorImageType, DeformationFieldType> WarperType;
   typename WarperType::Pointer warper = WarperType::New();
   warper->SetInput( movingImage );
   warper->SetOutputSpacing( fixedImage->GetSpacing() );
   warper->SetOutputOrigin( fixedImage->GetOrigin() );
   warper->SetDeformationField( defField );
   warper->SetEdgePaddingValue( itk::NumericTraits<typename TensorImageType::PixelType>::Zero );

   if ( args.verbosity > 0 )
     {
     std::cout<<"Warping final image"<<std::endl;
     }
   try
     {
     warper->Update();
     }
   catch( itk::ExceptionObject& err )
     {
     std::cout << "Warper fails." << std::endl;
     std::cout << err << std::endl;
     exit( EXIT_FAILURE );
     }

   if(args.verbosity)
     {  
     double eucMSE;
     double logMSE;

     typename TensorImageType::Pointer warpedMovingImage = TensorImageType::New();
     yeoCopyImage<TensorImageType>(warper->GetOutput(), warpedMovingImage);
     
     //Computing foreground mask
     typename MaskImageType::Pointer fixedImage_fg = MaskImageType::New();
     yeoGetForegroundMask<TensorImageType, MaskImageType>(fixedImage, fixedImage_fg, itk::NumericTraits<typename TensorImageType::PixelType>::Zero, 0, 1);
     
     //Compute Statistics on Deformation Field
     double numPixels = yeoSumImage<MaskImageType>(fixedImage_fg);
     typename MaskImageType::Pointer DefFieldNormImage = MaskImageType::New();
     yeoNormTensorImageNewImage<DeformationFieldType, MaskImageType>(defField, DefFieldNormImage);
     
     //yeoWriteImage<MaskImageType>(DefFieldNormImage, "DefFieldNorm.nhdr");
     
     yeoPowerCombine2ImagesInPlace<MaskImageType>(DefFieldNormImage, fixedImage_fg, 1, 1);
     std::cout<<"Mean Warp: "<<yeoSumImage<MaskImageType>(DefFieldNormImage)/numPixels<<std::endl;
     std::cout<<"Max Warp: "<<yeoMaxValue<MaskImageType>(DefFieldNormImage)<<std::endl;
                            
     //Compute Background Mask
     typename MaskImageType::Pointer movingImage_fg = MaskImageType::New();
     yeoGetForegroundMask<TensorImageType, MaskImageType>(warpedMovingImage, movingImage_fg, itk::NumericTraits<typename TensorImageType::PixelType>::Zero, 0, 1);
      
     yeoPowerCombine2ImagesInPlace<MaskImageType>(fixedImage_fg, movingImage_fg, 1, 1);
 
     //compute SSD
     std::cout<<"Computing SSD"<<std::endl;
     yeoLinearCombine2ImagesInPlace<TensorImageType>(warpedMovingImage, fixedImage, 1.0, -1.0);
     yeoPowerCombine2TensorImagesInPlace<TensorImageType>(warpedMovingImage, warpedMovingImage, 1.0, 1.0);
      
     //apply mask
     yeoMultiplyTensorImageWithImageInPlace<TensorImageType, MaskImageType>(warpedMovingImage, fixedImage_fg);

     numPixels = yeoSumImage<MaskImageType>(fixedImage_fg);
     if(args.eucBool)
       {
       eucMSE = yeoSumTensorImage<TensorImageType>(warpedMovingImage)/numPixels;
       std::cout<<"Final Euc MSE fixed image vs warped moving image averaged over "<<numPixels<<" pixels: "<<eucMSE<<std::endl;
       }
     else
       {
       logMSE = yeoSumTensorImage<TensorImageType>(warpedMovingImage)/numPixels;
       std::cout<<"Final Log MSE fixed image vs warped moving image averaged over "<<numPixels<<" pixels: "<<logMSE<<std::endl;
       }

     /* 
      * Now compute Euclidean MSE (Or Log)
      */
 
     //Reread in fixed and moving image to compute euclidean MSE
     typename TensorImageType::Pointer fixedImageNew = TensorImageType::New();
     typename TensorImageType::Pointer movingImageNew = TensorImageType::New();

     yeoReadTensorImage<TensorImageType>(fixedImageNew, args.fixedImageFile.c_str());
     if(args.eucBool)
       {
       std::cout<<"Taking Log of Fixed Image"<<std::endl;
       yeoLogTensorImageInPlace<TensorImageType>(fixedImageNew);
       }
     else
       {
       std::cout<<"Do not take log of Fixed Image"<<std::endl;
       }

     yeoReadTensorImage<TensorImageType>(movingImageNew, args.movingImageFile.c_str());
     if(!args.eucBool)
       {
       std::cout<<"Taking Log of Warped Moving Image"<<std::endl;
       yeoLogTensorImageInPlace<TensorImageType>(movingImageNew);
       }
     else
       {
       std::cout<<"Do not take log of warped Moving Image"<<std::endl;
       }

     //warp tensor image
     typename WarperType::Pointer warperNew = WarperType::New();
     warperNew->SetInput( movingImageNew );
     warperNew->SetOutputSpacing( fixedImageNew->GetSpacing() );
     warperNew->SetOutputOrigin( fixedImageNew->GetOrigin() );
     warperNew->SetDeformationField( defField );
     warperNew->SetEdgePaddingValue( itk::NumericTraits<typename TensorImageType::PixelType>::Zero );
     
     if ( args.verbosity > 0 )
       {
       std::cout<<"Warping new moving image"<<std::endl;
       }
     try
       {
       warperNew->Update();
       }
     catch( itk::ExceptionObject& err )
       {
       std::cout << "Warper fails." << std::endl;
       std::cout << err << std::endl;
       exit( EXIT_FAILURE );
       }

     typename TensorImageType::Pointer warpedMovingImageNew = TensorImageType::New();
     yeoCopyImage<TensorImageType>(warperNew->GetOutput(), warpedMovingImageNew);

     if(args.eucBool)
       {
       std::cout<<"Taking Log of Warped Moving Image"<<std::endl;
       yeoLogTensorImageInPlace<TensorImageType>(warpedMovingImageNew);
       }
     else
       {
       std::cout<<"Exponentiate warped Moving Image"<<std::endl;
       yeoExpTensorImageInPlace<TensorImageType>(warpedMovingImageNew);
       }
     
     //Compute Euclidean MSE (or Log)
     yeoLinearCombine2ImagesInPlace<TensorImageType>(warpedMovingImageNew, fixedImageNew, 1.0, -1.0);
     yeoPowerCombine2TensorImagesInPlace<TensorImageType>(warpedMovingImageNew, warpedMovingImageNew, 1.0, 1.0);
     
     //apply Mask
     yeoMultiplyTensorImageWithImageInPlace<TensorImageType, MaskImageType>(warpedMovingImageNew, fixedImage_fg);

     if(args.eucBool)
       {
       logMSE = yeoSumTensorImage<TensorImageType>(warpedMovingImageNew)/numPixels;
       std::cout<<"Final Log MSE fixed image vs warped moving image averaged over "<<numPixels<<" pixels: "<<logMSE<<std::endl;
       }
     else
       {
       eucMSE = yeoSumTensorImage<TensorImageType>(warpedMovingImageNew)/numPixels;
       std::cout<<"Final Euc MSE fixed image vs warped moving image averaged over "<<numPixels<<" pixels: "<<eucMSE<<std::endl;
       }
     
     //Write out results
     if(!args.resultsFile.empty())
       {
       typedef itk::WarpSmoothnessCalculator<DeformationFieldType> SmoothnessCalculatorType;
       typename SmoothnessCalculatorType::Pointer m_SmothnessCalculator;
       m_SmothnessCalculator = SmoothnessCalculatorType::New();
       m_SmothnessCalculator->SetImage( defField );
       m_SmothnessCalculator->Compute();
       const double harmonicEnergy = m_SmothnessCalculator->GetSmoothness();

       ofstream myfile;
       myfile.open(args.resultsFile.c_str(), ios::app);
       myfile << logMSE;
       myfile << "\n";
       myfile << eucMSE;
       myfile << "\n";
       myfile << harmonicEnergy;
       myfile << "\n";
       myfile.close();
       }
     
     /*
     //Write SSD
     typename MaskImageType::Pointer SSDimage2 = MaskImageType::New();
     yeoSumTensorImageNewImage<TensorImageType, MaskImageType>(warpedMovingImage, SSDimage2);
     yeoWriteImage<MaskImageType>(SSDimage2, "finalSSD.nhdr");
     
     //Write SSD improvement
     typename MaskImageType::Pointer diffImage = MaskImageType::New();
     yeoLinearCombine2ImagesNewImage<MaskImageType>(SSDimage, SSDimage2, diffImage, 1, -1);
     yeoWriteImage<MaskImageType>(diffImage, "SSDImprovement.nhdr");

     //normalized nSSD
     if(args.useAtlas)
       {
       yeoPowerCombine2TensorImagesInPlace<TensorImageType>(warpedMovingImage, varImage, 1.0, -1.0);
       std::cout<<"Final Normalized MSE Atlas vs warped moving image averaged over "<<numPixels<<" pixels: "<<yeoSumTensorImage<TensorImageType>(warpedMovingImage)/numPixels<<std::endl;  
       
       //Write nSSD
       typename MaskImageType::Pointer nSSDimage2 = MaskImageType::New();
       yeoSumTensorImageNewImage<TensorImageType, MaskImageType>(warpedMovingImage, nSSDimage2);
       yeoWriteImage<MaskImageType>(nSSDimage2, "finalnSSD.nhdr");

       //Write nSSD improvement
       typename MaskImageType::Pointer diffnImage = MaskImageType::New();
       yeoLinearCombine2ImagesNewImage<MaskImageType>(nSSDimage, nSSDimage2, diffnImage, 1, -1);
       yeoWriteImage<MaskImageType>(diffnImage, "nSSDImprovement.nhdr");
       }
    */
     }


   //write the result
   if ( args.verbosity > 0 )
     {
     std::cout<<"Exponentiating final image"<<std::endl;
     }
   typename TensorImageType::Pointer finalImage = warper->GetOutput();
   yeoExpTensorImageInPlace<TensorImageType>(finalImage);
   
   if ( args.verbosity > 0 )
     {
     std::cout<<"Writing final image"<<std::endl;
     }
   yeoWriteTensorImage<TensorImageType>(finalImage, args.outputImageFile.c_str());

       
   // Write deformation field
   if (!args.outputFieldFile.empty())
     {
     // Write the deformation field as an image of vectors.
     // Note that the file format used for writing the deformation field must be
     // capable of representing multiple components per pixel. This is the case
     // for the MetaImage and VTK file formats for example.
     typedef itk::ImageFileWriter< DeformationFieldType > FieldWriterType;
     typename FieldWriterType::Pointer fieldWriter = FieldWriterType::New();
     fieldWriter->SetFileName(  args.outputFieldFile.c_str() );
     fieldWriter->SetInput( defField );
     //fieldWriter->SetUseCompression( true );
     
     try
       {
       fieldWriter->Update();
       }
     catch( itk::ExceptionObject& err )
       {
       std::cout << "Unexpected error." << std::endl;
       std::cout << err << std::endl;
       exit( EXIT_FAILURE );
       }
     }
   
   
   // Create and write warped grid image
   if ( args.verbosity > 0 )
   {
      typedef itk::Image< unsigned char, Dimension > GridImageType;
      typename GridImageType::Pointer gridImage = GridImageType::New();
      gridImage->SetRegions( movingImage->GetRequestedRegion() );
      gridImage->SetOrigin( movingImage->GetOrigin() );
      gridImage->SetSpacing( movingImage->GetSpacing() );
      gridImage->Allocate();
      gridImage->FillBuffer(0);
      
      typedef itk::ImageRegionIteratorWithIndex<GridImageType> GridImageIteratorWithIndex;
      GridImageIteratorWithIndex itergrid = GridImageIteratorWithIndex(
         gridImage, gridImage->GetRequestedRegion() );

      std::cout<<"Create Grid Image"<<std::endl;
      const int gridspacing(8);
      for (itergrid.GoToBegin(); !itergrid.IsAtEnd(); ++itergrid)
      {
         itk::Index<Dimension> index = itergrid.GetIndex();

         if (Dimension==2 || Dimension==3)
         {
            // produce an xy grid for all z
            if ( (index[0]%gridspacing)==0 ||
                 (index[1]%gridspacing)==0 )
            {
               itergrid.Set( itk::NumericTraits<unsigned char>::max() );
            }
         }
         else
         {
            unsigned int numGridIntersect = 0;
            for( unsigned int dim = 0; dim < Dimension; dim++ )
               numGridIntersect += ( (index[dim]%gridspacing)==0 );
            if (numGridIntersect >= (Dimension-1))
               itergrid.Set( itk::NumericTraits<unsigned char>::max() );
         }
      }

      std::cout<<"Warp Grid Image"<<std::endl;
      typedef itk::WarpImageFilter
         < GridImageType, GridImageType, DeformationFieldType >  GridWarperType;
      typename GridWarperType::Pointer gridwarper = GridWarperType::New();
      gridwarper->SetInput( gridImage );
      gridwarper->SetOutputSpacing( fixedImage->GetSpacing() );
      gridwarper->SetOutputOrigin( fixedImage->GetOrigin() );
      gridwarper->SetDeformationField( defField );
      
      // Write warped grid to file
      typedef itk::ImageFileWriter< GridImageType >  GridWriterType;
   
      typename GridWriterType::Pointer      gridwriter =  GridWriterType::New();
      gridwriter->SetFileName( "WarpedGridImage.mha" );
      gridwriter->SetInput( gridwarper->GetOutput()   );
      gridwriter->SetUseCompression( true );
   
      try
      {
         gridwriter->Update();
      }
      catch( itk::ExceptionObject& err )
      {
         std::cout << "Unexpected error." << std::endl;
         std::cout << err << std::endl;
         exit( EXIT_FAILURE );
      }
   }


   // Create and write forewardwarped grid image
   if ( args.verbosity > 0 )
   {
      typedef itk::Image< unsigned char, Dimension > GridImageType;
      typedef itk::GridForwardWarpImageFilter<DeformationFieldType, GridImageType> GridForwardWarperType;

      typename GridForwardWarperType::Pointer fwWarper = GridForwardWarperType::New();
      fwWarper->SetInput(defField);
      fwWarper->SetForegroundValue( itk::NumericTraits<unsigned char>::max() );

      // Write warped grid to file
      typedef itk::ImageFileWriter< GridImageType >  GridWriterType;
   
      typename GridWriterType::Pointer      gridwriter =  GridWriterType::New();
      gridwriter->SetFileName( "ForwardWarpedGridImage.mha" );
      gridwriter->SetInput( fwWarper->GetOutput()   );
      gridwriter->SetUseCompression( true );
   
      std::cout<<"Forward Warp Grid Image"<<std::endl;
      try
      {
         gridwriter->Update();
      }
      catch( itk::ExceptionObject& err )
      {
         std::cout << "Unexpected error." << std::endl;
         std::cout << err << std::endl;
         exit( EXIT_FAILURE );
      }
   }


/*   
   // Create and write jacobian of the deformation field
   if ( args.verbosity > 0 )
   {
      typedef itk::WarpJacobianDeterminantFilter
         <DeformationFieldType, ImageType> JacobianFilterType;
      typename JacobianFilterType::Pointer jacobianFilter = JacobianFilterType::New();
      jacobianFilter->SetInput( defField );
      jacobianFilter->SetUseImageSpacing( true );

      writer->SetFileName( "TransformJacobianDeteminant.mha" );
      caster->SetInput( jacobianFilter->GetOutput() );
      writer->SetInput( caster->GetOutput()   );
      writer->SetUseCompression( true );
      
      try
      {
         writer->Update();
      }
      catch( itk::ExceptionObject& err )
      {
         std::cout << "Unexpected error." << std::endl;
         std::cout << err << std::endl;
         exit( EXIT_FAILURE );
      }

      typedef itk::MinimumMaximumImageCalculator<ImageType> MinMaxFilterType;
      typename MinMaxFilterType::Pointer minmaxfilter = MinMaxFilterType::New();
      minmaxfilter->SetImage( jacobianFilter->GetOutput() );
      minmaxfilter->Compute();
      std::cout<<"Minimum of the determinant of the Jacobian of the warp: "
               <<minmaxfilter->GetMinimum()<<std::endl;
      std::cout<<"Maximum of the determinant of the Jacobian of the warp: "
               <<minmaxfilter->GetMaximum()<<std::endl;
   }

   */
}


int main( int argc, char *argv[] )
{
   struct arguments args;

   parseOpts (argc, argv, args);
   
   std::cout<<"Starting demons registration with the following arguments:"<<std::endl;
   std::cout<<args<<std::endl<<std::endl;

   // FIXME uncomment for debug only
   //itk::MultiThreader::SetGlobalDefaultNumberOfThreads(1);

   time_t start;
   time (&start);

   switch ( args.dim )
   {
   case 2:
     std::cerr << "Unsuported dimension" << std::endl;
     exit( EXIT_FAILURE );
     //DemonsRegistrationFunction<2>(args);
     break;
   case 3:
      DemonsRegistrationFunction<3>(args);
      break;
   default:
      std::cerr << "Unsuported dimension" << std::endl;
      exit( EXIT_FAILURE );
   }
	
   time_t final;
   final = time(&final);
   double dif;
   dif = difftime(final, start);
   std::cout<<"Total Time for registration: "<<dif<<" seconds"<<std::endl;
	
   return EXIT_SUCCESS;
}
