
/*
* Written by: Thomas Yeo
* 
* Example usage:
*  LinearCombine2Images im1 im2 output_file binary_file a b
* 
*/


#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"
#include "itkImage.h"

#include <getopt.h>
#include <iostream>
#include <string>
#include "Debug.h"
#include "UsefulFunctions.h"

#define ImageDimension 3

int main( int argc, char *argv[] )
{   
  
  if(argc < 7)
    {
    std::cout<<"LinearCombine2Images im1 im2 output_file binary_file a b"<<std::endl;
    exit(EXIT_FAILURE);
    }

  //Read in Arguments
  const std::string input1_filename = argv[1];
  const std::string input2_filename = argv[2];
  const std::string output_filename = argv[3];
  const std::string binary_filename = argv[4];

  const float a = atof(argv[5]);
  const float b = atof(argv[6]);
  
  typedef float PixelType;
  typedef itk::Image< PixelType, ImageDimension>   ImageType;

  std::cout<<"Read Image 1: "<<input1_filename<<std::endl;
  ImageType::Pointer input1 = ImageType::New();
  yeoReadImage<ImageType>(input1, input1_filename);

  std::cout<<"Read Image 2: "<<input2_filename<<std::endl;
  ImageType::Pointer input2 = ImageType::New();
  yeoReadImage<ImageType>(input2, input2_filename);

  ImageType::Pointer output = ImageType::New();

  std::cout<<"Linear Combine Images"<<std::endl;
  yeoLinearCombine2ImagesNewImage<ImageType>(input1, input2, output, a, b);

  std::cout<<"Write output: "<<output_filename<<std::endl;
  yeoWriteImage<ImageType>(output, output_filename);

  std::cout<<"Compute binary image"<<std::endl;
  int neg_pixels = yeoReplaceValueInImageInPlace<ImageType>(output, 0, -1, 1);
  int pos_pixels = yeoReplaceValueInImageInPlace<ImageType>(output, 0, 1, 2);
  
  std::cout<<"Number of -pixels: "<<neg_pixels<<std::endl;
  std::cout<<"Number of +pixels: "<<pos_pixels<<std::endl;
  std::cout<<"Number of +pixels minus -pixels: "<<yeoSumImage<ImageType>(output)<<std::endl;

  std::cout<<"Write binary image: "<<binary_filename<<std::endl;
  yeoWriteImage<ImageType>(output, binary_filename);

  std::cout<<"Linear Combine success!"<<std::endl;
  exit(EXIT_SUCCESS);
}

