#ifndef _itkDiffeomorphicDemonsRegistrationFilterAtlas_h_
#define _itkDiffeomorphicDemonsRegistrationFilterAtlas_h_

#include "itkDiffeomorphicDemonsRegistrationFilter.h"
#include "itkESMDemonsRegistrationFunctionAtlas.h"

#include "itkMultiplyByConstantImageFilter.h"
#include "itkExponentialDeformationFieldImageFilter.h"
#include "itkWarpVectorImageFilter.h"
#include "itkVectorLinearInterpolateNearestNeighborExtrapolateImageFunction.h"
#include "itkAddImageFilter.h"

#include "Debug.h"

namespace itk {

template<class TFixedImage, class TMovingImage, class TDeformationField>
class ITK_EXPORT DiffeomorphicDemonsRegistrationFilterAtlas : 
    public DiffeomorphicDemonsRegistrationFilter< TFixedImage, TMovingImage,
                                            TDeformationField>
{

public:

 /** Standard class typedefs. */
  typedef DiffeomorphicDemonsRegistrationFilterAtlas    Self;
  typedef DiffeomorphicDemonsRegistrationFilter<
    TFixedImage, TMovingImage,TDeformationField>    Superclass;
  typedef SmartPointer<Self> Pointer;
  typedef SmartPointer<const Self> ConstPointer;

  /** Method for creation through the object factory. */
  itkNewMacro(Self);

  /** Run-time type information (and related methods). */
  itkTypeMacro( DiffeomorphicDemonsRegistrationFilterAtlas, 
		DiffeomorphicDemonsRegistrationFilter);



  /** FixedImage image type. */
  typedef typename Superclass::FixedImageType   FixedImageType;
  
/** MovingImage image type. */
  typedef typename Superclass::MovingImageType    MovingImageType;
  
  /** Deformation field type. */
  typedef typename Superclass::DeformationFieldType DeformationFieldType;
  

  /** FiniteDifferenceFunction type. */
  typedef typename Superclass::FiniteDifferenceFunctionType
  FiniteDifferenceFunctionType;

  /** Take timestep type from the FiniteDifferenceFunction. */
  typedef typename Superclass::TimeStepType TimeStepType;


  /* Define Variance Image Type */
  typedef FixedImageType VarImageType;
  typedef typename VarImageType::Pointer VarImagePointer;
  typedef typename VarImageType::ConstPointer VarImageConstPointer;

  typedef ESMDemonsRegistrationFunctionAtlas<FixedImageType,MovingImageType,
			   DeformationFieldType>  DemonsRegistrationFunctionType;

  
  virtual void SetUseAtlas(bool);

  /** Set the fixed image. */
  virtual void SetVarImage( const VarImageType * ptr );

  /** Get the fixed image. */
  virtual const VarImageType * GetVarImage(void) const;

/** Initialize the state of filter and equation before each iteration.
   * Progress feeback is implemented as part of this method. */
  virtual void InitializeIteration();

  /** Apply update. */
  virtual void ApplyUpdate(TimeStepType dt);

protected:
  DiffeomorphicDemonsRegistrationFilterAtlas();
  ~DiffeomorphicDemonsRegistrationFilterAtlas() {}
  void PrintSelf(std::ostream& os, Indent indent) const;


private:
  DiffeomorphicDemonsRegistrationFilterAtlas(const Self&); //purposely not implemented
  void operator=(const Self&); //purposely not implemented

};


} // end namespace itk

#ifndef ITK_MANUAL_INSTANTIATION
#include "itkDiffeomorphicDemonsRegistrationFilterAtlas.txx"
#endif

#endif
