#ifndef _itkDiffeomorphicDemonsRegistrationFilterAtlas_txx
#define _itkDiffeomorphicDemonsRegistrationFilterAtlas_txx
#include "itkDiffeomorphicDemonsRegistrationFilterAtlas.h"

#include "itkSmoothingRecursiveGaussianImageFilter.h"

#include "UsefulFunctions.h"
#include "UsefulTensorFunctions.h"

//#define DEBUG
#include "Debug.h"
namespace itk {

/* Usage: 
*          ComputeStatisticsDeformationField<DeformationFieldType, FixedImageType>(df, filename1, filename2 );
*/
template <class DeformationFieldType, class FixedImageType> void ComputeStatisticsDeformationField(const DeformationFieldType *defField, const FixedImageType *fixedImage, std::string filename1, std::string filename2)
{
  typedef typename itk::Image<float, DeformationFieldType::ImageDimension> MaskImageType;

  typename MaskImageType::Pointer DefFieldNormImage = MaskImageType::New();
  yeoNormTensorImageNewImage<DeformationFieldType, MaskImageType>(defField, DefFieldNormImage);
     
  typename MaskImageType::Pointer fixedImage_fg = MaskImageType::New();
  yeoGetForegroundMask<FixedImageType, MaskImageType>(fixedImage, fixedImage_fg, itk::NumericTraits<typename FixedImageType::PixelType>::Zero, 0, 1);
                                                                                     
  double numPixels = yeoSumImage<MaskImageType>(fixedImage_fg);
     
  yeoPowerCombine2ImagesInPlace<MaskImageType>(DefFieldNormImage, fixedImage_fg, 1, 1);
  std::cout<<"Mean Warp: "<<yeoSumImage<MaskImageType>(DefFieldNormImage)/numPixels<<", ";
  std::cout<<"Max Warp: "<<yeoMaxValue<MaskImageType>(DefFieldNormImage)<<", "; 


  if(!filename1.length()==0)
    {
    yeoWriteImage<DeformationFieldType>(defField, filename1);
    }
  
  if(!filename2.length()==0)
    {
    yeoWriteImage<MaskImageType>(DefFieldNormImage, filename2);
    }
}


/*
 * Default constructor
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
DiffeomorphicDemonsRegistrationFilterAtlas<TFixedImage,TMovingImage,TDeformationField>
::DiffeomorphicDemonsRegistrationFilterAtlas()
{

  typename DemonsRegistrationFunctionType::Pointer drfp;
  drfp = DemonsRegistrationFunctionType::New();
  
  this->SetDifferenceFunction( static_cast<FiniteDifferenceFunctionType *>(
                                 drfp.GetPointer() ) );
  
#ifdef DEBUG
//  this->PrintSelf(std::cout, 3);
  std::cout<<"Calling Constructor of DiffeomorphicDemonsRegistrationFilterAtlas"<<std::endl;
#endif
}


template <class TFixedImage, class TMovingImage, class TDeformationField>
void
DiffeomorphicDemonsRegistrationFilterAtlas<TFixedImage,TMovingImage,TDeformationField>
::PrintSelf(std::ostream& os, Indent indent) const
{ 
  Superclass::PrintSelf( os, indent );
}




/*
 * 
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
void
DiffeomorphicDemonsRegistrationFilterAtlas<TFixedImage,TMovingImage,TDeformationField>
::SetUseAtlas(bool useAtlas) 
{
  DemonsRegistrationFunctionType *drfp = 
    dynamic_cast<DemonsRegistrationFunctionType *>
    (this->GetDifferenceFunction().GetPointer());
  
  if( !drfp )
   {
   itkExceptionMacro( << 
     "Could not cast difference function to SymmetricDemonsRegistrationFunction" );
   }
  
  drfp->SetUseAtlas(useAtlas);
}


/*
 * Set the Var image.
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
void
DiffeomorphicDemonsRegistrationFilterAtlas<TFixedImage,TMovingImage,TDeformationField>
::SetVarImage(
  const VarImageType * ptr )
{
  this->ProcessObject::SetNthInput( 3, const_cast< VarImageType * >( ptr ) );
}


/*
 * Get the Var image.
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
const typename DiffeomorphicDemonsRegistrationFilterAtlas<TFixedImage,TMovingImage,TDeformationField>
::VarImageType *
DiffeomorphicDemonsRegistrationFilterAtlas<TFixedImage,TMovingImage,TDeformationField>
::GetVarImage() const
{
  return dynamic_cast< const VarImageType * >
    ( this->ProcessObject::GetInput( 3 ) );
}


/*
 * Set the function state values before each iteration
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
void
DiffeomorphicDemonsRegistrationFilterAtlas<TFixedImage,TMovingImage,TDeformationField>
::InitializeIteration()
{
  // update variables in the equation object
  DemonsRegistrationFunctionType *f = 
    dynamic_cast<DemonsRegistrationFunctionType *>
    (this->GetDifferenceFunction().GetPointer());

  if ( !f )
    {
    itkExceptionMacro(<<"FiniteDifferenceFunction not of type DemonsRegistrationFunctionType");
    }

  VarImageConstPointer varPtr = this->GetVarImage();
  f->SetVarImage(varPtr);

  
  if(!f->GetVarImage() && f->GetUseAtlas())
    {
    std::cerr<<"*************************"<<std::endl;
    std::cerr<<"Diffeomorphic Demons: UseAtlas Boolean is true but VarImage does not exists"<<std::endl;
    std::cerr<<"*************************"<<std::endl;
    exit(EXIT_FAILURE);
    } 


  // call the superclass  implementation ( initializes f )
  Superclass::InitializeIteration();
}


/*
 * Apply Update
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
void
DiffeomorphicDemonsRegistrationFilterAtlas<TFixedImage,TMovingImage,TDeformationField>
::ApplyUpdate(TimeStepType dt)
{
  DemonsRegistrationFunctionType *drfp = 
    dynamic_cast<DemonsRegistrationFunctionType *>
    (this->GetDifferenceFunction().GetPointer());
  
  if( !drfp )
    {
    itkExceptionMacro( << 
		       "Could not cast difference function to DiffeomorphicDemonsRegistrationFunction" );
    }

/*
  std::cout<<"========Before Update==========="<<std::endl;
  drfp->UpdateVarianceImage();
  drfp->PrintMSE("MSE1.nhdr");
  drfp->PrintWMSE("WMSE1.nhdr", "VarImage1.nhdr", "VarGradMag1.nhdr");
*/
  std::cout<<"----------------------------"<<std::endl;
  ComputeStatisticsDeformationField<DeformationFieldType, FixedImageType>(this->GetUpdateBuffer(), drfp->GetFixedImage(), "updatefield.nhdr", "updatefieldnorm.nhdr");
  Superclass::ApplyUpdate(dt);
  ComputeStatisticsDeformationField<DeformationFieldType, FixedImageType>(this->GetDeformationField(), drfp->GetFixedImage(), "", "");
  std::cout<<std::endl;
/*
  std::cout<<"==========After Update==========="<<std::endl;
  drfp->PrintMSE("MSE4.nhdr");
  drfp->PrintWMSE("WMSE4.nhdr", "VarImage4.nhdr", "VarGradMag4.nhdr");

  //Compute Diff Image between MSE4-MSE1 
  typename FixedImageType::Pointer MSE1 = FixedImageType::New();
  typename FixedImageType::Pointer MSE4 = FixedImageType::New();
  yeoReadImage<FixedImageType>(MSE1, "MSE1.nhdr");
  yeoReadImage<FixedImageType>(MSE4, "MSE4.nhdr");

  yeoLinearCombine2ImagesInPlace<FixedImageType>(MSE4, MSE1, 1, -1);
  yeoWriteImage<FixedImageType>(MSE4, "MSE4mMSE1.nhdr");

  //Compute Diff Image between WMSE4-WMSE1
  typename FixedImageType::Pointer WMSE1 = FixedImageType::New();
  typename FixedImageType::Pointer WMSE4 = FixedImageType::New();
  yeoReadImage<FixedImageType>(WMSE1, "WMSE1.nhdr");
  yeoReadImage<FixedImageType>(WMSE4, "WMSE4.nhdr");

  yeoLinearCombine2ImagesInPlace<FixedImageType>(WMSE4, WMSE1, 1, -1);
  yeoWriteImage<FixedImageType>(WMSE4, "WMSE4mWMSE1.nhdr");

  //Compute diff between VarImage1 and VarImage4 (just to show
  //it's 0.
  typename FixedImageType::Pointer VarImage1 = FixedImageType::New();
  typename FixedImageType::Pointer VarImage4 = FixedImageType::New();
  yeoReadImage<FixedImageType>(VarImage1, "VarImage1.nhdr");
  yeoReadImage<FixedImageType>(VarImage4, "VarImage4.nhdr");
  
  yeoLinearCombine2ImagesInPlace<FixedImageType>(VarImage1, VarImage4, 1, -1);
  yeoAbsImageInPlace<FixedImageType>(VarImage1);
  std::cout<<"Total Difference: "<<yeoSumImage<FixedImageType>(VarImage1)<<std::endl;

  //Compute diff between VarImage1 and VarImage4 (just to show
  //it's 0.
  typename FixedImageType::Pointer VarGradMag1 = FixedImageType::New();
  typename FixedImageType::Pointer VarGradMag4 = FixedImageType::New();
  yeoReadImage<FixedImageType>(VarGradMag1, "VarGradMag1.nhdr");
  yeoReadImage<FixedImageType>(VarGradMag4, "VarGradMag4.nhdr");
  
  yeoLinearCombine2ImagesInPlace<FixedImageType>(VarGradMag1, VarGradMag4, 1, -1);
  yeoAbsImageInPlace<FixedImageType>(VarGradMag1);
  std::cout<<"Total Difference: "<<yeoSumImage<FixedImageType>(VarGradMag1)<<std::endl;

  exit(EXIT_FAILURE);
*/



}




} // end namespace itk

#endif
