#ifndef _itkESMDemonsRegistrationFunctionAtlas_txx_
#define _itkESMDemonsRegistrationFunctionAtlas_txx_

#include "itkESMDemonsRegistrationFunctionAtlas.h"
#include "itkExceptionObject.h"
#include "vnl/vnl_math.h"
#include "UsefulFunctions.h"
#include <math.h>
#include "itkImageRegionIterator.h"
#include "itkImageRegionConstIterator.h"

#include "Debug.h"
namespace itk {

/*
 * Default constructor
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
ESMDemonsRegistrationFunctionAtlas<TFixedImage,TMovingImage,TDeformationField>
::ESMDemonsRegistrationFunctionAtlas()
{
  
  m_useAtlas = 0;
  m_VarImage = NULL;
  m_CacheVarianceImage = VarImageType::New();
#ifdef DEBUG
//  this->PrintSelf(std::cout, 3);
 std::cout<<"Calling Constructor of ESMDemonsRegistrationFunctionAtlas"<<std::endl;
#endif
}

/*
 * Standard "PrintSelf" method.
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
void
ESMDemonsRegistrationFunctionAtlas<TFixedImage,TMovingImage,TDeformationField>
::PrintSelf(std::ostream& os, Indent indent) const
{
  Superclass::PrintSelf(os, indent);
  
  os<< indent << "useAtlas: ";
  os<< m_useAtlas << std::endl; 
}

/*
 * Compute update at a non boundary neighbourhood
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
typename ESMDemonsRegistrationFunctionAtlas<TFixedImage,TMovingImage,TDeformationField>
::PixelType
ESMDemonsRegistrationFunctionAtlas<TFixedImage,TMovingImage,TDeformationField>
::ComputeUpdate(const NeighborhoodType &it, void * gd,
                const FloatOffsetType& itkNotUsed(offset))
{


  GlobalDataStruct *globalData = (GlobalDataStruct *)gd;
  PixelType update;
  IndexType FirstIndex = this->GetFixedImage()->GetLargestPossibleRegion().GetIndex();
  IndexType LastIndex = this->GetFixedImage()->GetLargestPossibleRegion().GetIndex() + 
     this->GetFixedImage()->GetLargestPossibleRegion().GetSize();
  
  const IndexType index = it.GetIndex();
  SpacingType m_FixedImageSpacing = this->GetFixedImage()->GetSpacing();



  if(!GetVarImage() && m_useAtlas) //Just a check
    {
    std::cerr<<"ESM: UseAtlas Boolean is true but VarImage does not exists"<<std::endl;
    exit(EXIT_FAILURE);
    }

  
#ifdef DEBUG
  if(GetVarImage())
    {
    typename VarImageType::IndexType referenceIndex2 = GetVarImage()->GetLargestPossibleRegion().GetIndex() 
      +  GetVarImage()->GetLargestPossibleRegion().GetSize();
    
    referenceIndex2[0] = referenceIndex2[0]/2;
    referenceIndex2[1] = referenceIndex2[1]/2;
    referenceIndex2[2] = referenceIndex2[2]/2;

    if(index == referenceIndex2)
      {
      std::cout<<"***********"<<std::endl;
      std::cout<<"VarImage["<<referenceIndex2[0]<<","<<referenceIndex2[1]<<","<<referenceIndex2[2]<<"]: "<<GetVarImage()->GetPixel(referenceIndex2)<<std::endl;
      std::cout<<"***********"<<std::endl;
      }
    }
#endif

#ifdef DEBUG
  IndexType ReferenceIndex = LastIndex;
  ReferenceIndex[0] = ReferenceIndex[0]/2;
  ReferenceIndex[1] = ReferenceIndex[1]/2;
  ReferenceIndex[2] = ReferenceIndex[2]/2;
#endif



  // Get fixed image related information
  // Note: no need to check the index is within
  // fixed image buffer. This is done by the external filter.
  const double fixedValue = static_cast<double>( this->GetFixedImage()->GetPixel( index ) );
  const CovariantVectorType fixedGradient = this->GetFixedImageGradientCalculator()->EvaluateAtIndex( index );
  

#ifdef DEBUG 
  if(index == ReferenceIndex)
    {
    std::cout<<"Fixed Gradient["<<ReferenceIndex[0]<<","<<ReferenceIndex[1]<<","<<ReferenceIndex[2]<<"]: "<<fixedGradient<<std::endl;
    std::cout<<"Fixed Value [ "<<ReferenceIndex[0]<<","<<ReferenceIndex[1]<<","<<ReferenceIndex[2]<<"]: "<<fixedValue<<std::endl;
    PrintRadius1FixedImageNeighborhood(index);
    }
#endif
  // Get moving image related information
  // check if the point was mapped outside of the moving image using
  // the "special value" NumericTraits<MovingPixelType>::max()
  typename MovingImageType::PixelType movingPixValue = this->m_MovingImageWarper->GetOutput()->GetPixel( index );
  if ( movingPixValue == NumericTraits<typename MovingImageType::PixelType>::max() )
  {
#ifdef DEBUG
  if(index == ReferenceIndex)
    {
    std::cout<<"Out of Bound "<<std::endl;
    }
#endif
     update.Fill( 0.0 );
     return update;

     // TEST
     // movingPixValue =  NumericTraits<typename MovingImageType::PixelType>::Zero;
  }


  
  const double movingValue = static_cast<double>( movingPixValue );
#ifdef DEBUG
  if(index == ReferenceIndex)
    {
    std::cout<<"First Index: Moving Value is "<<movingValue<<std::endl;
    }
#endif

  CovariantVectorType usedDoubleGradient;
  
  if ( (this->GetUseGradientType()==Superclass::Symmetric) || (this->GetUseGradientType()==Superclass::Moving) )
  {

#ifdef DEBUG
  if(index == ReferenceIndex)
    {
    std::cout<<"Computing Moving Gradient"<<std::endl;
    }
#endif


     // we don't use a CentralDifferenceImageFunction here to be able to
     // check for NumericTraits<MovingPixelType>::max()
     CovariantVectorType movingGradient;
     IndexType tmpIndex = index;
     for( unsigned int dim = 0; dim < ImageDimension; dim++ )
     {
        // bounds checking
        if ( FirstIndex[dim]==LastIndex[dim]  || index[dim]<FirstIndex[dim] || index[dim]>=LastIndex[dim] )
        {
#ifdef DEBUG
	std::cout<<"***This should never happen***"<<std::endl;
#endif
           movingGradient[dim] = 0.0;
           continue;
        }
        else if ( index[dim] == FirstIndex[dim] )
        {
           // compute derivative
           tmpIndex[dim] += 1;
           movingPixValue = this->m_MovingImageWarper->GetOutput()->GetPixel( tmpIndex );
           if ( movingPixValue == NumericTraits<typename MovingImageType::PixelType>::max() )
           {
              // weird crunched border case
              movingGradient[dim] = 0.0;
              //std::cout<<"A weird crunched border case at index: "<<index<<std::endl;
           }
           else
           {
              // forward difference
              movingGradient[dim] = static_cast<double>( movingPixValue ) - movingValue;
              
              movingGradient[dim] /= m_FixedImageSpacing[dim]; 
           }
           tmpIndex[dim] -= 1;
           continue;
        }
        else if ( index[dim] == (LastIndex[dim]-1) )
        {
           // compute derivative
           tmpIndex[dim] -= 1;
           movingPixValue = this->m_MovingImageWarper->GetOutput()->GetPixel( tmpIndex );
           if ( movingPixValue == NumericTraits<typename MovingImageType::PixelType>::max() )
           {
              // weird crunched border case
              movingGradient[dim] = 0.0;
              //std::cout<<"B weird crunched border case at index: "<<index<<std::endl;
           }
           else
           {
              // backward difference
              movingGradient[dim] = movingValue - static_cast<double>( movingPixValue );
              
              movingGradient[dim] /= m_FixedImageSpacing[dim]; 
           }
           tmpIndex[dim] += 1;
           continue;
        }
        
        
        
        // compute derivative
        tmpIndex[dim] += 1;
        movingPixValue = this->m_MovingImageWarper->GetOutput()->GetPixel( tmpIndex );
        if ( movingPixValue == NumericTraits<typename MovingImageType::PixelType>::max() )
        {
           // backward difference
           movingGradient[dim] = movingValue;
           
           tmpIndex[dim] -= 2;
           movingPixValue = this->m_MovingImageWarper->GetOutput()->GetPixel( tmpIndex );
           if ( movingPixValue == NumericTraits<typename MovingImageType::PixelType>::max() )
           {
              // weird crunched border case
              movingGradient[dim] = 0.0;
              
              //bool onborder = false;
              //for( unsigned int tmpdim = 0; tmpdim < ImageDimension; tmpdim++ )
              //{
              //   onborder |= ( index[tmpdim]==FirstIndex[tmpdim] or index[tmpdim]==(LastIndex[tmpdim]-1) );
              //}
              //if (onborder)
              //   std::cout<<"C Weird crunched border case at index: "<<index<<std::endl;
              //else
              //   std::cout<<"D Weird crunched border case at index: "<<index<<std::endl;
           }
           else
           {
              // backward difference
              movingGradient[dim] -= static_cast<double>( this->m_MovingImageWarper->GetOutput()->GetPixel( tmpIndex ) );
              
              movingGradient[dim] /= m_FixedImageSpacing[dim];
           }
           
        }
        else
        {
           movingGradient[dim] = static_cast<double>( movingPixValue );
           
           tmpIndex[dim] -= 2;
           movingPixValue = this->m_MovingImageWarper->GetOutput()->GetPixel( tmpIndex );
           if ( movingPixValue == NumericTraits<typename MovingImageType::PixelType>::max() )
           {
              // forward difference
              movingGradient[dim] -= movingValue;
              
              movingGradient[dim] /= m_FixedImageSpacing[dim];
           }
           else
           {
              // normal case, central difference
              movingGradient[dim] -= static_cast<double>( movingPixValue );
              
              movingGradient[dim] *= 0.5 / m_FixedImageSpacing[dim];
           }
        }
        tmpIndex[dim] += 1;
        
        
        // // test without max()
        // tmpIndex[dim] += 1;
        // movingGradient[dim] = static_cast<double>( m_MovingImageWarper->GetOutput()->GetPixel( tmpIndex ) );
        
        // tmpIndex[dim] -= 2;
        // movingGradient[dim] -= static_cast<double>( m_MovingImageWarper->GetOutput()->GetPixel( tmpIndex ) );
        
        // tmpIndex[dim] += 1;
        // movingGradient[dim] *= 0.5 / m_FixedImageSpacing[dim];
     }

     if (this->GetUseGradientType()==Superclass::Symmetric)
     {
        usedDoubleGradient = fixedGradient + movingGradient;
#ifdef DEBUG
	if(index == ReferenceIndex)
	  {
	  std::cout<<"Symmetric Gradient"<<usedDoubleGradient<<std::endl;
	  }
#endif
     }
     else if (this->GetUseGradientType()==Superclass::Moving)
     {
        usedDoubleGradient = movingGradient + movingGradient;
     }
     else
     {
        itkExceptionMacro(<<"Unknown gradient type");
     }
  }
  else if (this->GetUseGradientType()==Superclass::Fixed)
  {
     usedDoubleGradient = fixedGradient + fixedGradient;
  }
  else
  {
     itkExceptionMacro(<<"Unknown gradient type");
  }


  
  
  // Compute Update.
 //  We avoid the mismatch in units between the two terms. 
  // and avoid large step using a normalization term.
   
  
  const double usedDoubleGradientSquaredMagnitude =
     usedDoubleGradient.GetSquaredNorm();
     
  const double speedValue = fixedValue - movingValue;

  if ( vnl_math_abs(speedValue) < this->GetIntensityDifferenceThreshold() )
  {
     update.Fill( 0.0 );
  }
  else
  {

    double m_Normalizer = 0.0;
    // compute the normalizer
    if ( this->GetMaximumUpdateStepLength() > 0.0 )
    {
      m_Normalizer = 0.0;
      for( unsigned int k = 0; k < ImageDimension; k++ )
      {
        m_Normalizer += m_FixedImageSpacing[k] * m_FixedImageSpacing[k];
      }
      m_Normalizer *= this->GetMaximumUpdateStepLength() * this->GetMaximumUpdateStepLength() /
        static_cast<double>( ImageDimension );
  }
  else
  {
     // set it to minus one to denote a special case
     // ( unrestricted update length )
     m_Normalizer = -1.0;
  }
    double denom;
     if (  m_Normalizer > 0.0 )
     {
#ifdef DEBUG
	if(index == ReferenceIndex)
	  {
	  std::cout<<"Thirion Normalization"<<usedDoubleGradient<<std::endl;
	  }
#endif    


        // "ITK-Thirion" normalization
	if(m_useAtlas)
	  {
#ifdef DEBUG
	if(index == ReferenceIndex)
	  {
	  std::cout<<"Using My Normalization"<<std::endl;
	  }
#endif
	const double varValue =  static_cast<double>(GetVarImage()->GetPixel(index));
	denom =  usedDoubleGradientSquaredMagnitude +
	    (varValue/m_Normalizer);
	//denom =  usedDoubleGradientSquaredMagnitude+
	//  varValue*(1/m_Normalizer + 0.2 );
	//denom = varValue;
	//denom =  usedDoubleGradientSquaredMagnitude +
	//  (vnl_math_sqr(speedValue)/m_Normalizer);
	  }
	else
	  {
	  denom =  usedDoubleGradientSquaredMagnitude +
	    (vnl_math_sqr(speedValue)/m_Normalizer);
	  }
     }
     else
     {
#ifdef DEBUG
     if(index == ReferenceIndex)
       {
       std::cout<<"Least Square Normalization"<<usedDoubleGradient<<std::endl;
       }     
#endif
     std::cerr<<"Should always use ITK-Thirion!!"<<std::endl;
     exit(EXIT_FAILURE);
        // least square solution of the system
        denom =  usedDoubleGradientSquaredMagnitude;
     }
        
   
     if ( denom < 1e-9 )
     {
        update.Fill( 0.0 );

#ifdef DEBUG
	std::cout<<"Denom Below Threshold"<<1e-9<<std::endl;
#endif
     }
     else
     {
        const double factor = 2.0 * speedValue / denom;           
    
	for( unsigned int j = 0; j < ImageDimension; j++ )
        {
           update[j] = factor * usedDoubleGradient[j];
        }
     }
  }

#ifdef DEBUG
  if(index == ReferenceIndex)
    {
    std::cout<<"Final Update: "<<update<<std::endl;
    }
#endif


  // WARNING!! We compute the global data without taking into account the current update step.
  // There are several reasons for that: If an exponential, a smoothing or any other operation
  // is applied on the update field, we cannot compute the newMappedCenterPoint here; and even
  // if we could, this would be an often unnecessary time-consuming task.
  if ( globalData )
  {
     globalData->m_SumOfSquaredDifference += vnl_math_sqr( speedValue );
     globalData->m_NumberOfPixelsProcessed += 1;
     globalData->m_SumOfSquaredChange += update.GetSquaredNorm();
     if(m_useAtlas)
       {
       const double varValue =  static_cast<double>(GetVarImage()->GetPixel(index));
       globalData->m_NormalizedSumOfSquaredDifference += vnl_math_sqr(speedValue)/varValue;
       }
     else
       {
       globalData->m_NormalizedSumOfSquaredDifference += vnl_math_sqr( speedValue );
       }
  }

  return update;
}


template <class TFixedImage, class TMovingImage, class TDeformationField>
void
ESMDemonsRegistrationFunctionAtlas<TFixedImage,TMovingImage,TDeformationField>
::PrintMSE(std::string filename) const
{
  typedef FixedImageType GenericImageType;
  typename GenericImageType::ConstPointer fixedImage  
                      =  static_cast<typename GenericImageType::ConstPointer> (this->GetFixedImage());
  this->m_MovingImageWarper->Update();
  typename GenericImageType::ConstPointer movingImage2 
                      =  static_cast<typename GenericImageType::ConstPointer> (this->m_MovingImageWarper->GetOutput());

  if(fixedImage->GetLargestPossibleRegion().GetSize() != movingImage2->GetLargestPossibleRegion().GetSize())
    {
    std::cerr<<"Not the same size!!"<<std::endl;
    }
  
  float numPixels = (float) fixedImage->GetRequestedRegion().GetNumberOfPixels();

  typename GenericImageType::Pointer movingImage = GenericImageType::New();
  yeoCopyImage<GenericImageType>(movingImage2, movingImage);

  typedef typename itk::ImageRegionIterator< GenericImageType> IteratorType;
  IteratorType It1 = IteratorType(movingImage, movingImage->GetLargestPossibleRegion());
  
  typedef typename itk::ImageRegionConstIterator< GenericImageType> ConstIteratorType;
  ConstIteratorType It2 = ConstIteratorType(fixedImage, fixedImage->GetLargestPossibleRegion());
  
  for(It1.GoToBegin(), It2.GoToBegin(); !It1.IsAtEnd(); ++It1, ++It2)
    {
    if(It1.Get() == NumericTraits<typename MovingImageType::PixelType>::max())
      {
      It1.Set(It2.Get());
      numPixels--;
      }
    }

  typename GenericImageType::Pointer diffImage = GenericImageType::New();
  //difference
  yeoLinearCombine2ImagesNewImage<GenericImageType>(fixedImage, movingImage, diffImage, 1, -1);
  //square
  yeoPowerCombine2ImagesInPlace<GenericImageType>(diffImage, diffImage, 1, 1);
  //divide by numPixels
  yeoMultiplyImageByConstantInPlace<GenericImageType>(diffImage, 1/numPixels);
  //sum
  std::cout<<"MSE: "<<yeoSumImage<GenericImageType>(diffImage)<<std::endl;

  if(!filename.length()==0)
    {
    yeoWriteImage<GenericImageType>(diffImage, filename);
    }
  

}

template <class TFixedImage, class TMovingImage, class TDeformationField>
void
ESMDemonsRegistrationFunctionAtlas<TFixedImage,TMovingImage,TDeformationField>
::PrintWMSE(std::string filename1, std::string filename2, std::string filename3) const
{
  typedef FixedImageType GenericImageType;
  typename GenericImageType::ConstPointer fixedImage  
    =  static_cast<typename GenericImageType::ConstPointer> (this->GetFixedImage());
  this->m_MovingImageWarper->Update();
  typename GenericImageType::ConstPointer movingImage2 
    =  static_cast<typename GenericImageType::ConstPointer> (this->m_MovingImageWarper->GetOutput());
  
  if(fixedImage->GetLargestPossibleRegion().GetSize() != movingImage2->GetLargestPossibleRegion().GetSize())
    {
    std::cerr<<"Not the same size!!"<<std::endl;
    }
  
  float numPixels = (float) fixedImage->GetRequestedRegion().GetNumberOfPixels();
  
  typename GenericImageType::Pointer movingImage = GenericImageType::New();
  yeoCopyImage<GenericImageType>(movingImage2, movingImage);
  
  typedef typename itk::ImageRegionIterator< GenericImageType> IteratorType;
  IteratorType It1 = IteratorType(movingImage, movingImage->GetLargestPossibleRegion());
  
  typedef typename itk::ImageRegionConstIterator< GenericImageType> ConstIteratorType;
  ConstIteratorType It2 = ConstIteratorType(fixedImage, fixedImage->GetLargestPossibleRegion());
  
  for(It1.GoToBegin(), It2.GoToBegin(); !It1.IsAtEnd(); ++It1, ++It2)
    {
    if(It1.Get() == NumericTraits<typename MovingImageType::PixelType>::max())
      {
      It1.Set(It2.Get());
      numPixels--;
      }
    }
  
  typename GenericImageType::Pointer diffImage = GenericImageType::New();
  //difference
  yeoLinearCombine2ImagesNewImage<GenericImageType>(fixedImage, movingImage, diffImage, 1, -1);
  //square
  yeoPowerCombine2ImagesInPlace<GenericImageType>(diffImage, diffImage, 1, 1);
  //divide by var image
  yeoPowerCombine2ImagesInPlace<GenericImageType>(diffImage, m_CacheVarianceImage, 1, -1);
 
  if(!filename1.length()==0)
    {
    yeoWriteImage<GenericImageType>(diffImage, filename1);
    }

  //divide by numPixels
  yeoMultiplyImageByConstantInPlace<GenericImageType>(diffImage, 1/numPixels);
  //sum
  std::cout<<"WMSE: "<<yeoSumImage<GenericImageType>(diffImage)<<std::endl;
  
 
  if(!filename2.length()==0)
    {
    yeoWriteImage<GenericImageType>(m_CacheVarianceImage, filename2);
    }
  if(!filename3.length()==0)
    {
    typename VarImageType::Pointer GradMagImage = VarImageType::New();
    yeoGradientMagnitudeNewImage<VarImageType,itkGetStaticConstMacro(ImageDimension) >(m_CacheVarianceImage, GradMagImage);
    yeoWriteImage<GenericImageType>(GradMagImage, filename3);
    }
}

template <class TFixedImage, class TMovingImage, class TDeformationField>
void
ESMDemonsRegistrationFunctionAtlas<TFixedImage,TMovingImage,TDeformationField>
::UpdateVarianceImage(void) 
{
  if(m_useAtlas)
    {
    yeoCopyImage<VarImageType>(m_VarImage, m_CacheVarianceImage);
    }
  else
    {
    //Read fixed image and moving image
    typedef FixedImageType GenericImageType;
    typename GenericImageType::ConstPointer fixedImage  
      =  static_cast<typename GenericImageType::ConstPointer> (this->GetFixedImage());
    
    this->m_MovingImageWarper->Update(); //need to update warper!
    typename GenericImageType::ConstPointer movingImage2 
      =  static_cast<typename GenericImageType::ConstPointer> (this->m_MovingImageWarper->GetOutput());
    
    if(fixedImage->GetLargestPossibleRegion().GetSize() != movingImage2->GetLargestPossibleRegion().GetSize())
      {
      std::cerr<<"Not the same size!!"<<std::endl;
      }

    //Need to reset points of warped moving image which are undefined
    typename GenericImageType::Pointer movingImage = GenericImageType::New();
    yeoCopyImage<GenericImageType>(movingImage2, movingImage);
    
    typedef typename itk::ImageRegionIterator< GenericImageType> IteratorType;
    IteratorType It1 = IteratorType(movingImage, movingImage->GetLargestPossibleRegion());
    
    typedef typename itk::ImageRegionConstIterator< GenericImageType> ConstIteratorType;
    ConstIteratorType It2 = ConstIteratorType(fixedImage, fixedImage->GetLargestPossibleRegion());
    
    for(It1.GoToBegin(), It2.GoToBegin(); !It1.IsAtEnd(); ++It1, ++It2)
      {
      if(It1.Get() == NumericTraits<typename MovingImageType::PixelType>::max())
	{
	It1.Set(It2.Get()+1); //0 Variance is a bad idea, since later
			      //on you need to divide by the variance.
	}
      if( fabs(It1.Get() - It2.Get()) <= this->m_IntensityDifferenceThreshold)
	{
	It1.Set(It2.Get()+this->m_IntensityDifferenceThreshold); //Set
								 //limit of variance
	}
      }

    //difference
    yeoLinearCombine2ImagesNewImage<GenericImageType>(fixedImage, movingImage, m_CacheVarianceImage, 1, -1);
    //square
    yeoPowerCombine2ImagesInPlace<GenericImageType>(m_CacheVarianceImage, m_CacheVarianceImage, 1, 1);
    
    }
}




} // end namespace itk

#endif
