#ifndef __itkMultiResolutionPDEDeformableRegistration2Atlas_h
#define __itkMultiResolutionPDEDeformableRegistration2Atlas_h

#include "itkImage.h"
#include "itkImageToImageFilter.h"
#include "itkPDEDeformableRegistrationFilter.h"
#include "itkMultiResolutionPDEDeformableRegistration2.h"
#include "itkDiffeomorphicDemonsRegistrationFilterAtlas.h"
#include "itkDemonsRegistrationFilter.h"
#include "itkMultiResolutionPyramidImageFilter.h"
#include "itkVectorResampleImageFilter.h"

#include <vector>
#include "Debug.h"

namespace itk
{

template <class TFixedImage, class TMovingImage, class TDeformationField, class  TRealType = float>
class ITK_EXPORT MultiResolutionPDEDeformableRegistration2Atlas :
    public MultiResolutionPDEDeformableRegistration2<TFixedImage, TMovingImage, TDeformationField, TRealType>
{
public:
  /** Standard class typedefs */
  typedef MultiResolutionPDEDeformableRegistration2Atlas Self;
  typedef MultiResolutionPDEDeformableRegistration2<TFixedImage, TMovingImage, TDeformationField, TRealType>
    Superclass;
  typedef SmartPointer<Self>  Pointer;
  typedef SmartPointer<const Self>  ConstPointer;

  /** Method for creation through the object factory. */
  itkNewMacro(Self);

  /** Run-time type information (and related methods). */
  itkTypeMacro( MultiResolutionPDEDeformableRegistration2Atlas, 
                MultiResolutionPDEDeformableRegistration2 );

  /** Fixed image type. */
  typedef TFixedImage FixedImageType;
  typedef typename Superclass::FixedImagePointer FixedImagePointer;
  typedef typename Superclass::FixedImageConstPointer FixedImageConstPointer;
  
  /** Moving image type. */
  typedef TMovingImage MovingImageType;
  typedef typename Superclass::MovingImagePointer MovingImagePointer;
  typedef typename Superclass::MovingImageConstPointer MovingImageConstPointer;
  
  /** Deformation field image type. */
  typedef TDeformationField DeformationFieldType;
  typedef typename Superclass::DeformationFieldPointer DeformationFieldPointer;

  /** ImageDimension. */
  itkStaticConstMacro(ImageDimension, unsigned int,
                      Superclass::FixedImageType::ImageDimension);

  /** Internal float image type. */
  typedef typename Superclass::FloatImageType FloatImageType;

  /** The internal registration type. */
  typedef DiffeomorphicDemonsRegistrationFilterAtlas
  <TFixedImage,TMovingImage,TDeformationField> DiffeomorphicRegistrationType;

  typedef typename DiffeomorphicRegistrationType::Pointer DiffeomorphicRegistrationPointer;
  
  /** The internal Registration Function Type */
  typedef FixedImageType VarImageType;
  typedef ESMDemonsRegistrationFunctionAtlas<FixedImageType,MovingImageType,
					     DeformationFieldType>  DemonsRegistrationFunctionType;


  /** TVar image type. */
  typedef typename VarImageType::Pointer VarImagePointer;
  typedef typename VarImageType::ConstPointer VarImageConstPointer;

  /** The Var multi-resolution image pyramid type. */
  typedef MultiResolutionPyramidImageFilter<
    VarImageType, VarImageType > VarImagePyramidType;
  typedef typename VarImagePyramidType::Pointer VarImagePyramidPointer;

  /** Set number of multi-resolution levels. */
  virtual void SetNumberOfLevels( unsigned int num );

  /** Set the Var image. */
  virtual void SetVarImage( const VarImageType * ptr );

  /** Get the Var image. */
  const VarImageType * GetVarImage(void) const;

  /** Set the fixed image pyramid. */
  itkSetObjectMacro( VarImagePyramid, VarImagePyramidType );

  /** Get the fixed image pyramid. */
  itkGetObjectMacro( VarImagePyramid, VarImagePyramidType );

/*
  void SetRegistrationFilter2(DiffeomorphicRegistrationType* filter)
  {
    m_RegistrationFilter2 = filter;
  }
*/

protected:
  MultiResolutionPDEDeformableRegistration2Atlas();
  ~MultiResolutionPDEDeformableRegistration2Atlas() {}
  void PrintSelf(std::ostream& os, Indent indent) const;

 /** Generate output data by performing the registration
   * at each resolution level. */
  virtual void GenerateData();

  /*DiffeomorphicRegistrationPointer m_RegistrationFilter2;*/

private:
  MultiResolutionPDEDeformableRegistration2Atlas(const Self&); //purposely not implemented
  void operator=(const Self&); //purposely not implemented

  VarImagePyramidPointer   m_VarImagePyramid;

};


} // end namespace itk

#ifndef ITK_MANUAL_INSTANTIATION
#include "itkMultiResolutionPDEDeformableRegistration2Atlas.txx"
#endif

#endif
