#include <iostream>

#include <itkImageFileReader.h>
#include <itkImageFileWriter.h>
#include <itkImageRegionConstIterator.h>
#include <itkImageRegionIterator.h>
#include <itkImage.h>

#include "GetPot.h"

void PrintHelp( const char* exec )
{
	std::cout << "Usage: " << std::endl;
	std::cout << exec << " <-i input> <-m mask image> <-o output>" << std::endl;
	exit (0);
}

using namespace itk;

int main (int narg, char* arg[])
{

	itk::Object::GlobalWarningDisplayOff();

	GetPot cl(narg, arg); // argument parser
  if( cl.size() == 1 || cl.search(2, "--help", "-h") )
  {
    PrintHelp(cl[0]);
  }

  const char* input = cl.follow("NoFile",2,"-I","-i");
  const char* output = cl.follow("NoFile",2,"-O","-o");
  const char* mask = cl.follow( "NoFile", 2, "-m", "-M");


  typedef double                        ScalarType;
  typedef unsigned short                IntegerType;
  typedef Image<IntegerType, 3>         ImageType;
  typedef Image<IntegerType, 3>         ImageMaskType;

  typedef itk::ImageFileReader<ImageType> ImageReaderType;
  typedef itk::ImageFileReader<ImageMaskType> ImageMaskReaderType;
  typedef itk::ImageFileWriter<ImageType> WriterType;

  ImageReaderType::Pointer reader = ImageReaderType::New();
  reader->SetFileName( input );

  std::cout << "Reading: " << input << std::flush;
  try
  {
	reader->Update();
  }
  catch( itk::ExceptionObject &e)
  {
	  std::cerr << e;
	  exit (-1);
  }
  std::cout << " Done." << std::endl;

  ImageType::Pointer image = reader->GetOutput();



  ImageMaskReaderType::Pointer reader1 = ImageMaskReaderType::New();
  reader1->SetFileName( mask );

  std::cout << "Reading: " << mask << std::flush;
  try
  {
	reader1->Update();
  }
  catch( itk::ExceptionObject &e)
  {
	  std::cerr << e;
	  exit (-1);
  }
  std::cout << " Done." << std::endl;

  ImageType::Pointer maskImage = reader1->GetOutput();


  itk::ImageRegionConstIterator<ImageType> itIn (image, image->GetLargestPossibleRegion() );
  itk::ImageRegionConstIterator<ImageMaskType> itMask (maskImage, maskImage->GetLargestPossibleRegion() );

  ImageType::Pointer finalImage = ImageType::New();
  finalImage->SetRegions( image->GetLargestPossibleRegion() );
  finalImage->SetOrigin( image->GetOrigin() );
  finalImage->SetSpacing( image->GetSpacing() );
  finalImage->Allocate();

  itk::ImageRegionIterator<ImageType> itOut (finalImage, finalImage->GetLargestPossibleRegion() );
  while( !itOut.IsAtEnd() )
  {
	  if( itMask.Value() )
	  {
		itOut.Set( itIn.Get() );
	  }
	  else
	  {
		itOut.Set( static_cast<IntegerType>(0.0) );
	  }
	
	  ++itOut;
	  ++itIn;
	  ++itMask;
  }


  
  // write the image
  WriterType::Pointer writer = WriterType::New();
  writer->SetFileName( output );
  writer->SetInput( finalImage );

  std::cout << "Writing: " << output << std::flush;
  try
  {
	writer->Update();
  }
  catch( itk::ExceptionObject &e)
		{
			std::cerr << e;
			exit (-1);
		}

  std::cout << " Done." << std::endl;



	return 0;
}
