#ifndef _itk_NormalizedGaussianInterpolationTensorImageFilter_txx_
#define _itk_NormalizedGaussianInterpolationTensorImageFilter_txx_
#include "itkNormalizedGaussianInterpolationTensorImageFilter.h"

#include <itkImageRegionConstIteratorWithIndex.h>
#include <itkImageRegionIterator.h>


namespace itk
{


  template< class TInputImage, class TOutputImage >
  void
  NormalizedGaussianInterpolationTensorImageFilter< TInputImage, TOutputImage >
  ::ThreadedGenerateData(const OutputImageRegionType &outputRegionForThread, int threadId)
  {

    typedef ImageRegionIterator<OutputImageType>               OutputImageIteratorType;
    typedef ImageRegionConstIteratorWithIndex<InputImageType>  InputImageIteratorType;
    typedef typename InputImageType::IndexType                 IndexType;
    
    InputImageIteratorType itIn(this->GetInput(),outputRegionForThread);
    OutputImageIteratorType itOut(this->GetOutput(),outputRegionForThread);

    // get the length of the list == number of seed points
    unsigned int numPoints  = m_Tensors.size();
    double SIGMA2 = m_Sigma*m_Sigma;    
    
    while(!itOut.IsAtEnd())
      {

        OutputPixelType out( static_cast<ScalarType>(0.0) );
        // threshold the input image that must be scalar, in order
        // to not interpolate in the full brain
	if( itIn.Value() > static_cast<InputPixelType>(0.0)  )
	  {
           
            // w will contain the weights of the different seed tensors
            // that have an influence on the current voxel
	    double *w = new double[numPoints];
            // sum is the sum of the weights
            double sum = 0.0;
            // current index
            IndexType index = itIn.GetIndex();
	    
	    for(unsigned int n=0; n<numPoints; n++)
	      {
		PointType ptn = m_Points[n];
		double dist = 0.0;

                // transform the index (coordinates in the image) into
                // world coordinates
		PointType pt;                
		this->GetInput()->TransformIndexToPhysicalPoint(index,pt);

                // Euclidean distance to the considered point
                for(unsigned int m=0; m<ImageDimension; m++)
		  dist += (ptn[m]-pt[m])*(ptn[m]-pt[m]);
		
		// coeff gaussien: as we normalize, one does not need
                // the 1/sqrt(2*pi)^(N/2) thing. 
		double G = exp(-0.5*dist/SIGMA2);
		w[n] = G;
		sum += G;
	      }

            // Tensor weighted mean: we simply compute the mean in the
            // logarithmic domain
            OutputPixelType Mean ( static_cast<ScalarType>(0.0) );
            
            for(unsigned int n=0; n<numPoints; n++)
            {

              if( m_DoNotDiffuseTangents )
              {

                PointType tn = m_Tangents[n];
                OutputPixelType W = m_Tensors[n];
                // compute tn*tn' : need a tensor of dimensions [NxN], where N is the
                // image dimension. I will take an OutputPixelType, but this is not
                // the CORRECT solution...what happen if we take a 6x6 tensor on a 3D
                // grid? 
                OutputPixelType X( static_cast<ScalarType>(0.0) );
                for( unsigned int nc=0; nc < ImageDimension; nc++ )
                {
                  for( unsigned int nl=0; nl <= nc; nl++ )
                    X.SetComponent( nl, nc, static_cast<ScalarType>( tn[nl]*tn[nc] ) );
                }
                
                ScalarType dotprod = X.ScalarProductWith(W);
                ScalarType norm2   = X.GetSquaredNorm();
                
                W = W - X * dotprod/norm2;
                
                Mean += W * static_cast<ScalarType>( w[n] );
                
              }              
              else
                Mean += m_Tensors[n] * static_cast<ScalarType>( w[n] );
              
            }
            

            Mean /= static_cast<ScalarType>(sum);
            out = Mean.Exp();
            delete [] w;
                                    
	  }

        
	itOut.Set(out);
	
	++itIn;
	++itOut;
      }
    
  }

} // end of namespace itk

#endif
