#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "basics/GeneralFormat.h"
#pragma implementation "basics/ListElement.h"
#pragma implementation "basics/ObjectsList.h"

#include <stdio.h>
#include <basics/GeneralFormat.h>
#include <basics/Constants.h>

#ifdef WIN32
#if ((!defined( _SGI_) || (defined(BASIC_IOSTREAM))) && (!defined(OLD_OSTREAM)))
using std::ofstream;
using std::ifstream;
using std::setprecision;
using std::ios;
#define OFSTREAM_PATCH ofstream
#define IFSTREAM_PATCH ifstream
#endif
#else
#define OFSTREAM_PATCH OFSTREAM
#define IFSTREAM_PATCH IFSTREAM
#endif

namespace yav {
void GeneralFormat::addModuleVersion(const TclString modulename,
                                     const bool mandatory,
                                     const double moduleVersionNumber,
                                     const GeneralFormatModuleVersion::INPUTREADFUNCTION readFunction,
                                     const GeneralFormatModuleVersion::INPUTSKIPFUNCTION skipFunction,
                                     const GeneralFormatModuleVersion::OUTPUTWRITEFUNCTION writeFunction)
{
  GeneralFormatModule *module;

  /* First seach if the module exists */
  for(module=moduleList.getFirst();
      (module!=0) && (module->getName()!=modulename);
      module=module->following()) ;

  if (module==0)
    {
      /* Create a new module */
      GeneralFormatModule *newModule =
        new GeneralFormatModule(modulename,mandatory);

      /* add module in list */
      moduleList.append(newModule);
      module = newModule;
    }
  /* Then search in the module if there is a version corresponding with this version */
   GeneralFormatModuleVersion *moduleVersion;

  /* First search if the module exists */
  for(moduleVersion=module->getFirstVersion();(moduleVersion!=0) && (moduleVersion->getVersion()!=moduleVersionNumber);
      moduleVersion=moduleVersion->following());

  if (moduleVersion==0)
    {
      /* Create a new module */
      GeneralFormatModuleVersion *newModuleVersion =
        new GeneralFormatModuleVersion(moduleVersionNumber,
                                       readFunction,skipFunction,writeFunction);
      /* add version in list */
      moduleVersion= newModuleVersion;
      (*module).addModuleVersion(moduleVersion);
    }
  else
    {
      moduleVersion->setReadFunction(readFunction);
      moduleVersion->setSkipFunction(skipFunction);
      moduleVersion->setWriteFunction(writeFunction);
    }

  /* updates the overall version number */
  if (GeneralFormat::version<moduleVersionNumber)
    GeneralFormat::version=moduleVersionNumber;

}


void GeneralFormat::setClassName(const char *k, const char *h)
{
  key = k;
  head = h;
}

TclString GeneralFormat::getMagic() const {
  TclString n='#';
  n+=key;
  return n;
}

void GeneralFormat::writeMagic(GFOutput &output, const double outputVersion)
#ifdef WITHOUT_STREAM_EXCEPTIONS

#else
  throw (ios::failure)
#endif
{
  /* Write General header */
  ((OFSTREAM& ) output) << "# " << key << " version " << outputVersion << ENDL
                        << "CONTENT" << ENDL;
}

void GeneralFormat::writeData(const double outputVersion, GFOutput &output)
#ifdef WITHOUT_STREAM_EXCEPTIONS
 throw (GeneralFormatModule::InvalidVersion,GeneralFormatModuleVersion::EmptyVersion)
#else
  throw (GeneralFormatModule::InvalidVersion,GeneralFormatModuleVersion::EmptyVersion,ios::failure)
#endif
{
  GeneralFormatModule *module;
  GeneralFormatModuleVersion *moduleVersion;

  for(unsigned int i = 0; i < numberOfElements; i++) {
    /* Write object header */
    ((OFSTREAM&) output) << head << ' ' << filename << ENDL;
    /* Write the name of each module */
    for(module=moduleList.getFirst();(module!=0);module=module->following())
      {
        if (module->isSelectedWriting()==true)
          ((OFSTREAM&) output) << "  " << module->getName() << ENDL;
      }
    ((OFSTREAM&) output) << "END" << ENDL;

    /* write the content of each module */
    for(module=moduleList.getFirst();(module!=0);module=module->following())
      {
        if (module->isSelectedWriting()==true)
          {
            /* Get the right version */
            moduleVersion= module->getCorrespondingVersion(outputVersion);

            /* test if the write function exists */
            if (moduleVersion->write ==
                (GeneralFormatModuleVersion::OUTPUTWRITEFUNCTION) 0)
              throw (GeneralFormatModuleVersion::EmptyVersion("GeneralFormat::writeData(ostream &,const double )"));
            ( this->*(moduleVersion->write))(output);
          }
      }
  }
}

void GeneralFormat::write(const char *filename, unsigned int n)
#ifdef WITHOUT_STREAM_EXCEPTIONS
  throw (GeneralFormatModule::InvalidVersion,UnknownModule,GeneralFormatModuleVersion::EmptyVersion)
#else
  throw (GeneralFormatModule::InvalidVersion,UnknownModule,GeneralFormatModuleVersion::EmptyVersion,ios::failure)
#endif
{
  GeneralFormat::write(DEFAULT_GF_MODE,GeneralFormat::version, filename,
                       n, (char *) 0);
}

void GeneralFormat::write(const double outputVersion, const char  *filename,
                          unsigned int n)
#ifdef WITHOUT_STREAM_EXCEPTIONS
  throw (GeneralFormatModule::InvalidVersion,UnknownModule,GeneralFormatModuleVersion::EmptyVersion)
#else
  throw (GeneralFormatModule::InvalidVersion,UnknownModule,GeneralFormatModuleVersion::EmptyVersion,ios::failure)
#endif
{
  GeneralFormat::write(DEFAULT_GF_MODE,outputVersion, filename, n, (char *) 0);
}


void GeneralFormat::write(const GFIO::GFMode mode,const char  *filename,
                          unsigned int number)
#ifdef WITHOUT_STREAM_EXCEPTIONS
  throw (GeneralFormatModule::InvalidVersion,UnknownModule,GeneralFormatModuleVersion::EmptyVersion)
#else
  throw (GeneralFormatModule::InvalidVersion,UnknownModule,GeneralFormatModuleVersion::EmptyVersion,ios::failure)
#endif
{
  GeneralFormat::write(mode,GeneralFormat::version, filename, number,
                       (char *) 0);
}

void GeneralFormat::write(const GFIO::GFMode mode, const double outputVersion,
                          const char *filename, unsigned int number...)
#ifdef WITHOUT_STREAM_EXCEPTIONS
  throw (GeneralFormatModule::InvalidVersion,UnknownModule,GeneralFormatModuleVersion::EmptyVersion)
#else
  throw (GeneralFormatModule::InvalidVersion,UnknownModule,GeneralFormatModuleVersion::EmptyVersion,ios::failure)
#endif
{
  /* Initialize variable list */
  va_list args;
  va_start(args,number);
  /* Open filename for output */
  if (strlen(filename)>0) {
	  GFOutput output(filename);
	  output.setMode(mode);
	  write(mode,outputVersion,output,number,va_arg(args,char*));
  }
  va_end(args);

}

void GeneralFormat::write(GFOutput &output, unsigned int n)
#ifdef WITHOUT_STREAM_EXCEPTIONS
  throw (GeneralFormatModule::InvalidVersion,UnknownModule,GeneralFormatModuleVersion::EmptyVersion)
#else
  throw (GeneralFormatModule::InvalidVersion,UnknownModule,GeneralFormatModuleVersion::EmptyVersion,ios::failure)
#endif
{
  GeneralFormat::write(DEFAULT_GF_MODE,GeneralFormat::version, output,
                       n, (char *) 0);
}

void GeneralFormat::write(const double outputVersion, GFOutput &output,
                          unsigned int n)
#ifdef WITHOUT_STREAM_EXCEPTIONS
  throw (GeneralFormatModule::InvalidVersion,UnknownModule,GeneralFormatModuleVersion::EmptyVersion)
#else
  throw (GeneralFormatModule::InvalidVersion,UnknownModule,GeneralFormatModuleVersion::EmptyVersion,ios::failure)
#endif
{
  GeneralFormat::write(DEFAULT_GF_MODE,outputVersion, output, n, (char *) 0);
}


void GeneralFormat::write(const GFIO::GFMode mode, GFOutput &output,
                          unsigned int number)
#ifdef WITHOUT_STREAM_EXCEPTIONS
  throw (GeneralFormatModule::InvalidVersion,UnknownModule,GeneralFormatModuleVersion::EmptyVersion)
#else
  throw (GeneralFormatModule::InvalidVersion,UnknownModule,GeneralFormatModuleVersion::EmptyVersion,ios::failure)
#endif
{
  GeneralFormat::write(mode,GeneralFormat::version, output, number,
                       (char *) 0);
}


void GeneralFormat::write(const GFIO::GFMode mode, const double outputVersion,
                          GFOutput &output, unsigned int number...)
#ifdef WITHOUT_STREAM_EXCEPTIONS
  throw (GeneralFormatModule::InvalidVersion,UnknownModule,GeneralFormatModuleVersion::EmptyVersion)
#else
  throw (GeneralFormatModule::InvalidVersion,UnknownModule,GeneralFormatModuleVersion::EmptyVersion,ios::failure)
#endif
{
  char *moduleName;
  int nbargs=0;
  numberOfElements = number;

  /* Initialize variable list */
  va_list args;
  va_start(args, number);
  /* reset all modules */
  while ((moduleName=va_arg(args,char*))!= (char *) 0){
    ++nbargs;
    /* select the module whose name is stored in the list */
    selectModuleWriting(moduleName);
    }
  va_end(args);

  writeMagic(output, outputVersion);

  ((OFSTREAM& ) output) << "  Number of " << head << " = " << number << ENDL;

  switch(mode) {
  case GFIO::GF_BINARY :
    ((OFSTREAM& ) output) << "  Binary" << ENDL;
    break;
  case GFIO::GF_ASCII :
    ((OFSTREAM& ) output) << "  ASCII" << ENDL;
    break;
  }
#if (defined(_ALPHA_)||defined(_LINUX_))
  ((OFSTREAM& ) output) << "  Big endian" << ENDL;
#else
  ((OFSTREAM& ) output) << "  Little endian" << ENDL;
#endif
  ((OFSTREAM& ) output) << "END" << ENDL;

  writeData(outputVersion, output);
}


void GeneralFormat::readMagic(STD_ISTREAM &input, double &inputVersion)
#ifdef WITHOUT_STREAM_EXCEPTIONS
  throw (UnknownFormat)
#else
  throw (UnknownFormat,ios::failure)
#endif
{
  TclString head;
  const char from[] =
    "void GeneralFormat::readMagic(ISTREAM &,double &)";
  unsigned int l = key.getLength();

  // Read General header
   do {
    input >> head;
    } while ((head.getLength()==0)||(head[0]==' '));

  if (strncmp((char *) head, "# ", 2) ||
      strncmp((char *) head + 2, (char *) key, l) ||
      strncmp((char *) head + 3 + l, "version", 7))
    throw UnknownFormat(from);
  if(sscanf((char *) head + 11 + l, "%lf", &inputVersion) != 1)
    throw UnknownFormat(from);
}

void GeneralFormat::readData(const double currentVersion, GFInput &input,
                             GFIO::GFMode mode, bool swap)
#ifdef WITHOUT_STREAM_EXCEPTIONS
throw (GeneralFormatModule::InvalidVersion, UnknownModule,
       GeneralFormatModuleVersion::EmptyVersion, Exception)
#else
throw (GeneralFormatModule::InvalidVersion, UnknownModule,
       GeneralFormatModuleVersion::EmptyVersion,ios::failure, Exception)
#endif
{
  TclString dummy;
  GeneralFormatModule *module;
  GeneralFormatModuleVersion *moduleVersion;

  const TclString from="GeneralFormat::readData";
 /* read the content of each module */

  for(unsigned int i = 0; i < numberOfElements; i++) {
    /* ignore the first line */
    input >> dummy;

    /* allocate a list of module with the same order as in the input file */
    std::list<GeneralFormatModule *> fileModuleList;

    input >> dummy;
    while(dummy != (const char *) "END") {
      /*  seach if the module exists */
      for(module=moduleList.getFirst();
          (module!=0) && (module->getName() != dummy);
          module=module->following());
      /* if the name of the module is unknown throw an exception */
      if (module==0)
        throw(UnknownModule("GeneralFormat::readData"));
      /* append the module in the list */
      fileModuleList.push_back(module);

      input >> dummy;
    }

    input.setMode(mode, swap);

    /* Now scan the module list to either read or skip the data */
    std::list<GeneralFormatModule *>::iterator moduleIterator;

    for(moduleIterator = fileModuleList.begin();
        moduleIterator != fileModuleList.end();moduleIterator++) {
        /* Get the right version */
        moduleVersion= (*moduleIterator)->getCorrespondingVersion(currentVersion);

        if ((*moduleIterator)->isSelectedReading())
          {
            /* test if the read function exists */
            if (moduleVersion->read==0)
              throw (GeneralFormatModuleVersion::EmptyVersion("GeneralFormat::readData"));
            (this->*(moduleVersion->read))(input);
            (*moduleIterator)->selectWriting();
          }
        else
          {
            /* test if the skip function exists */
            if (moduleVersion->skip==(GeneralFormatModuleVersion::INPUTSKIPFUNCTION) 0)
              throw (GeneralFormatModuleVersion::EmptyVersion("GeneralFormat::readData"));
            (this->*(moduleVersion->skip))(input);
          }
      }
  }

}



void GFIO::setMode(GFMode m) {
  mode = m;
}


GFOutput::GFOutput(const char *f) : OFSTREAM_PATCH(f), filename(f) {
#ifdef WITHOUT_STREAM_EXCEPTIONS
#else
  exceptions(ios_base::badbit|iosbase::failbit|iosbase::eofbit);
#endif
  setf(IOS::fixed, IOS::floatfield);
}


GFInput::GFInput(const char *f) : IFSTREAM_PATCH(f), filename(f) {
#ifdef WITHOUT_STREAM_EXCEPTIONS
#else
  exceptions(ios::badbit|ios::failbit|ios::eofbit);
#endif
  swap = false;
}


void GFInput::setMode(GFMode m, bool s) {
  mode = m;
  swap = s;
}

void GeneralFormat::read(char *filen...)
#ifdef WITHOUT_STREAM_EXCEPTIONS
      throw (GeneralFormatModule::InvalidVersion, UnknownModule,
             GeneralFormatModuleVersion::EmptyVersion,
             UnknownFormat, FileNotFound, Exception)
#else
      throw (GeneralFormatModule::InvalidVersion, UnknownModule,
             GeneralFormatModuleVersion::EmptyVersion, ios::failure,
             UnknownFormat, FileNotFound, Exception)
#endif
{
  char from[255];
  sprintf(from,"GeneralFormat::read(%s)",filen);

  filename = filen;

  /* Open filename for output */
  GFInput input(filen);
  if(!input) throw(FileNotFound(from, filen));

  /* Call read(GFInput) */
  va_list args;
  va_start(args,filen);
  read(&input,va_arg(args,char*));
  va_end(args);
}

void GeneralFormat::read(GFInput *input...)
#ifdef WITHOUT_STREAM_EXCEPTIONS
      throw (GeneralFormatModule::InvalidVersion, UnknownModule,
             GeneralFormatModuleVersion::EmptyVersion,
             UnknownFormat, FileNotFound, Exception)
#else
      throw (GeneralFormatModule::InvalidVersion, UnknownModule,
             GeneralFormatModuleVersion::EmptyVersion, ios::failure,
             UnknownFormat, FileNotFound, Exception)
#endif
{
  char *name;
  int nbargs=0;
  bool swap;
  GFIO::GFMode mode;
  double currentVersion;
  unsigned int i;
  TclString line;
  char from[] = "GeneralFormat::read";
  char context[255];

  /* Initialize variable list */
  va_list args;
  va_start(args,input);
  /* reset all modules */
  while ((name=va_arg(args,char*))!= (char *) 0)
    {
      CERR << "arg " << nbargs << " = " << (int)name[0] << ENDL;
      ++nbargs;
      /* select the module whose name is stored in the list */
      selectModuleReading(name);
    }
  va_end(args);

  readMagic(*input, currentVersion);

  /* Initialize default variable */
  mode = DEFAULT_GF_MODE;
  swap = DEFAULT_GF_READ_SWAP;

  /* read the module name */
  *input >> line;
  if(line != (const char *) "CONTENT") {
    sprintf(context,"%s expecting CONTENT got {%s}",from,(char *)line);
     throw UnknownFormat(context);
  }
  *input >> line;
  if(strncmp((char *) line, "Number of ", 10)) {
    sprintf(context,"%s expecting Number of got {%s}",from,(char *)line);
    throw UnknownFormat(context);
  }
  for(i = 10; i < line.getLength() && line[i] != '='; i++) ;
  if(line[i] != '=')  {
    sprintf(context,"%s expecting = got {%s}",from,(char *)line);
    throw UnknownFormat(context);
  }
  if(sscanf((char *) line + i + 1, "%u", &numberOfElements) != 1)
  {
    sprintf(context,"%s expecting integer got {%s}",from,(char *)line);
    throw UnknownFormat(context);
  }

  *input >> line;
  while(line != (const char *) "END") {
    if (line == (const char *) "Binary")
      mode=GFIO::GF_BINARY;
    else if (line == (const char *) "ASCII")
      mode=GFIO::GF_ASCII;
#if (defined(_ALPHA_)||defined(_LINUX_)|| defined(_WINDOWS_) )
    else if (line == (const char *) "Big endian")
      swap=false;
    else if (line == (const char *) "Little endian")
      swap=true;
#else
    else if (line == (const char *) "Big endian")
      swap=true;
    else if (line == (const char *) "Little endian")
      swap=false;
#endif
    else  {
    sprintf(context,"%s expecting END, endiannes or format got {%s}",from,(char *)line);
    throw UnknownFormat(context);
  }

    *input >> line;
  }

  readData(currentVersion, *input, mode, swap);
}


GeneralFormatModule& GeneralFormat::moduleByName(const TclString& name)
  throw(UnknownModule) {
  char from[255];
  sprintf(from,"GeneralFormat::moduleByName(const TclString& \"%s\")",(char*)name);
  GeneralFormatModule *module = moduleList.getFirst();
  for(; module; module = module->following())
    if(module->getName() == name) return *module;
  throw UnknownModule(from);
}


void GeneralFormat::unselectModuleWriting(const TclString& moduleName)
  throw(UnknownModule)
{
  GeneralFormatModule &module = moduleByName(moduleName);
  module.unselectWriting();
}

void GeneralFormat::unselectModuleReading(const TclString& moduleName)
  throw(UnknownModule)
{
  GeneralFormatModule &module = moduleByName(moduleName);
  module.unselectReading();
}


void GeneralFormat::selectModuleWriting(const TclString& moduleName)
  throw(UnknownModule)
{
  GeneralFormatModule &module = moduleByName(moduleName);
  module.selectWriting();
}

void GeneralFormat::selectModuleReading(const TclString& moduleName)
  throw(UnknownModule)
{
  GeneralFormatModule &module = moduleByName(moduleName);
  module.selectReading();
}


void GeneralFormat::selectAllModulesReading()
{
  GeneralFormatModule *module;
  /*  seach if the module exists */
  for(module=moduleList.getFirst();(module!=0);module=module->following())
    module->selectReading();
}
void GeneralFormat::selectAllModulesWriting()
{
  GeneralFormatModule *module;
  /*  seach if the module exists */
  for(module=moduleList.getFirst();(module!=0);module=module->following())
    module->selectWriting();
}

void GeneralFormat::unselectAllWritingModules()
{
  GeneralFormatModule *module;
  /*  seach if the module exists */
  for(module=moduleList.getFirst();(module!=0);module=module->following())
    {
      if (module->isMandatoryForWriting())
        module->selectReading();
      else
        module->unselectWriting();
    }

}

void GeneralFormat::unselectAllReadingModules()
{
  GeneralFormatModule *module;
  /*  seach if the module exists */
  for(module=moduleList.getFirst();(module!=0);module=module->following())
    {
      module->unselectReading();
    }

}

GeneralFormatModule::InvalidVersion::InvalidVersion(const std::string& from) :
  Exception(from,
            "Attempt to find a version that is smaller that all existing versions") {
}

void GeneralFormatModule::addModuleVersion(GeneralFormatModuleVersion *moduleVersion)
{
  versionList.append(moduleVersion);
}

GeneralFormatModuleVersion *GeneralFormatModule::getFirstVersion()
{
  return(versionList.getFirst());
}

GeneralFormatModuleVersion *GeneralFormatModule::getCorrespondingVersion(const double currentVersion) throw(InvalidVersion)
{
  GeneralFormatModuleVersion *moduleVersion,*theVersion;

  /* First search if the module exists */
  theVersion=0;
  /* scan all version of a given module */
  for(moduleVersion=versionList.getFirst();(moduleVersion!=0);
      moduleVersion=moduleVersion->following())
    {
      if (moduleVersion->getVersion()<=currentVersion)
        {
          if ((theVersion==0) || (theVersion->getVersion()<moduleVersion->getVersion()))
            theVersion=moduleVersion;
        }
    }
  if (theVersion==0)
    throw(InvalidVersion("GeneralFormatModuleVersion *getCorrespondingVersion(const double)"));

  return(theVersion);
}

TclString GeneralFormatModule::getName() const
{
  return(GeneralFormatModule::moduleName);
}
void GeneralFormatModule::selectWriting()
{
  GeneralFormatModule::selectedWriting=true;
}

void GeneralFormatModule::unselectWriting()
{
  if (mandatoryWriting==false)
    GeneralFormatModule::selectedWriting=false;
}
void GeneralFormatModule::selectReading()
{
  GeneralFormatModule::selectedReading=true;
}

void GeneralFormatModule::unselectReading()
{
  GeneralFormatModule::selectedReading=false;
}

bool GeneralFormatModule::isMandatoryForWriting() const
{
  return(GeneralFormatModule::mandatoryWriting);
}
bool GeneralFormatModule::isSelectedWriting() const
{
  return(GeneralFormatModule::selectedWriting);
}
bool GeneralFormatModule::isSelectedReading() const
{
  return(GeneralFormatModule::selectedReading);
}

double GeneralFormatModuleVersion::getVersion() const
{
  return(GeneralFormatModuleVersion::version);
}

void GeneralFormatModuleVersion::setReadFunction(const INPUTREADFUNCTION function){
  read=function;
}

void GeneralFormatModuleVersion::setSkipFunction(const INPUTSKIPFUNCTION function){
  skip=function;
}

void GeneralFormatModuleVersion::setWriteFunction(const OUTPUTWRITEFUNCTION function){
  write=function;
}


void GFIO::Swap(void *ptr, unsigned int size) {
  unsigned int i, mid;
  unsigned char c;

  if((size % 2)==0) {
    mid = size / 2;
    for(i=0; i<mid; i++) {
      c = *((unsigned char *)ptr + mid + i);
      *((unsigned char *)ptr + mid + i) =
        *((unsigned char *)ptr + mid - i - 1);
      *((unsigned char *)ptr + mid - i - 1) = c;
    }
  }
}

GFInput& operator>>(GFInput& input, bool &val) {
  if(input.getMode() == GFIO::GF_ASCII)
  {
        IFSTREAM &f=input;
        unsigned int v;
        f >> v;
                if (v==0)
                                val=false;
                        else
                                val=true;
  }
  else input.read((char *) &val, sizeof(bool));
  return input;
}

GFInput& operator>>(GFInput& input, char &val) {
  if(input.getMode() == GFIO::GF_ASCII) ((IFSTREAM&) input) >> val;
  else input.read(&val, sizeof(char));
  return input;
}

GFInput& operator>>(GFInput& input, char *val) {
  char c;
  int i = 0;

  if(input.getMode() == GFIO::GF_ASCII) {
    c = ' ';
    while(c == ' ' || c == '\t' || c == '\n' || c == '\0') c = input.get();
    val[i++] = c;
    while(((c = input.get()) != '\0') && (c != '\n')) val[i++] = c;
    val[i++] = '\0';
  }
  else {
    while((c = input.get()) != '\0') val[i++] = c;
    val[i++] = '\0';
  }
  return input;
}


GFInput& operator>>(GFInput& input, TclString& val) {
  char c;

  if(input.getMode() == GFIO::GF_ASCII) {
    c = ' ';
    while(c == ' ' || c == '\t' || c == '\n' || c == '\0') c = input.get();
    ((IFSTREAM&) input) >> val;
    val.prepend(c);
  }
  else {
    val = "";
    while((c = input.get()) != '\0') val += c;
  }
  return input;
}


GFInput& operator>>(GFInput& input, int &val) {
  if(input.getMode() == GFIO::GF_ASCII) ((IFSTREAM&) input) >> val;
  else {
    input.read((char *) &val, sizeof(int));
    if(input.getSwap()) GFIO::Swap((void *) &val, sizeof(int));
  }
  return input;
}


GFInput& operator>>(GFInput& input, unsigned int &val) {
  if(input.getMode() == GFIO::GF_ASCII) ((IFSTREAM&) input) >> val;
  else {
	  input.read((char *) &val, sizeof(unsigned int));
	  if(input.getSwap()) GFIO::Swap((void *) &val, sizeof(unsigned int));
  }
  return input;
}


GFInput& operator>>(GFInput& input, short int &val) {
  if(input.getMode() == GFIO::GF_ASCII) ((IFSTREAM&) input) >> val;
  else {
    input.read((char *) &val, sizeof(short int));
    if(input.getSwap()) GFIO::Swap((void *) &val, sizeof(short int));
  }
  return input;
}


GFInput& operator>>(GFInput& input, unsigned short int &val) {
  if(input.getMode() == GFIO::GF_ASCII) ((IFSTREAM&) input) >> val;
  else {
    input.read((char *) &val, sizeof(unsigned short int));
    if(input.getSwap()) GFIO::Swap((void *) &val, sizeof(unsigned short int));
  }
  return input;
}


GFInput& operator>>(GFInput& input, long int &val) {
  if(input.getMode() == GFIO::GF_ASCII) ((IFSTREAM&) input) >> val;
  else {
    input.read((char *) &val, sizeof(long int));
    if(input.getSwap()) GFIO::Swap((void *) &val, sizeof(long int));
  }
  return input;
}


GFInput& operator>>(GFInput& input, unsigned long int &val) {
  if(input.getMode() == GFIO::GF_ASCII) ((IFSTREAM&) input) >> val;
  else {
    input.read((char *) &val, sizeof(unsigned long int));
    if(input.getSwap()) GFIO::Swap((void *) &val, sizeof(unsigned long int));
  }
  return input;
}


GFInput& operator>>(GFInput& input, float &val) {
  if(input.getMode() == GFIO::GF_ASCII) ((IFSTREAM&) input) >> val;
  else {
    input.read((char *) &val, sizeof(float));
    if(input.getSwap()) GFIO::Swap((void *) &val, sizeof(float));
  }
  return input;
}


GFInput& operator>>(GFInput& input, double &val) {
  if(input.getMode() == GFIO::GF_ASCII) ((IFSTREAM&) input) >> val;
  else {
    //input.read((void *) &val, sizeof(double));
    //if(input.getSwap()) GFIO::Swap((void *) &val, sizeof(double));
    float f;
    input.read((char *) &f, sizeof(float));
    if(input.getSwap()) GFIO::Swap((void *) &f, sizeof(float));
    val = (double) f;
  }
  return input;
}


GFInput& operator>>(GFInput& input, Vec3<double> &val) {
  int i;
  if(input.getMode() == GFIO::GF_ASCII)
    for(i = 0; i < 3; i++) ((IFSTREAM&) input) >> val[i];
  else {
	  float f;
	  for(i = 0; i < 3; i++) {
		   input.read((char *) &f, sizeof(float));
		  if(input.getSwap()) 
			  GFIO::Swap((void *) &f, sizeof(float));
		  val[i] = (double) f;
	  }
  }
  return input;
}
GFInput& operator>>(GFInput& input, Vec3<unsigned int> &val) {
  int i;
  if(input.getMode() == GFIO::GF_ASCII)
    for(i = 0; i < 3; i++) ((IFSTREAM&) input) >> val[i];
  else {
    /*for(i = 0; i < 3; i++) {
      input.read((void *) &val[i], sizeof(double));
      if(input.getSwap()) GFIO::Swap((void *) &val[i], sizeof(double));
      }*/
    unsigned int f;
    for(i = 0; i < 3; i++) {
      input.read((char *) &f, sizeof(unsigned int));
      if(input.getSwap()) GFIO::Swap((void *) &f, sizeof(unsigned int));
      val[i] = (unsigned int) f;
    }
  }
  return input;
}

GFInput& operator>>(GFInput& input, Rotation3D &val) {
  Matrix3x3<double> mat;
  int i,j;
  if(input.getMode() == GFIO::GF_ASCII)
    {
      for(i = 0; i < 3; i++)
        {
          for(j = 0; j < 3; j++)
            {
              ((IFSTREAM&) input) >> mat(i,j);
            }
        }
    }
  else {
    float f;
    for(i = 0; i < 3; i++) {
      for(j = 0; j < 3; j++) {
        input.read((char *) &f, sizeof(float));
        if(input.getSwap()) GFIO::Swap((void *) &f, sizeof(float));
        mat(i,j)=(double) f;
      }
    }
  }
  val=mat;
  return input;
}
GFInput& operator>>(GFInput& input, Rotation2D &val) {
  double mat[2][2];
  int i,j;
  if(input.getMode() == GFIO::GF_ASCII)
    {
      for(i = 0; i < 2; i++)
        {
          for(j = 0; j < 2; j++)
            {
              ((IFSTREAM&) input) >> mat[i][j];
            }
        }
    }
  else {
    float f;
    for(i = 0; i < 2; i++) {
      for(j = 0; j < 2; j++) {
        input.read((char *) &f, sizeof(float));
        if(input.getSwap()) GFIO::Swap((void *) &f, sizeof(float));
        mat[i][j]=(double) f;
      }
    }
  }
  val=mat;
  return input;
}


GFInput& operator>>(GFInput& input, Vec2<double> &val) {
  int i;
  if(input.getMode() == GFIO::GF_ASCII)
    for(i = 0; i < 2; i++) ((IFSTREAM&) input) >> val[i];
  else {
    /*for(i = 0; i < 2; i++) {
      input.read((void *) &val[i], sizeof(double));
      if(input.getSwap()) GFIO::Swap((void *) &val[i], sizeof(double));
      }*/
    float f;
    for(i = 0; i < 2; i++) {
      input.read((char *) &f, sizeof(float));
      if(input.getSwap()) GFIO::Swap((void *) &f, sizeof(float));
      val[i] = (double) f;
    }
  }
  return input;
}


void GFInput::readArray(double *ptr, unsigned int arraySize)
{
  unsigned int i;
  if(mode == GF_ASCII) {
    for(i = 0; i < arraySize; i++) *this >> ptr[i];
  }
  else {
    read((char *) ptr, sizeof(double)*arraySize);
    if(swap) {
      double *q;
      unsigned int i;
      for(q = ptr, i = 0; i < arraySize; ++i, ++q)
        GFIO::Swap((void *) q, sizeof(double));
    }
  }
}


void GFInput::readArray(float *ptr, unsigned int arraySize)
{
  unsigned int i;
  if(mode == GF_ASCII) {
    for(i = 0; i < arraySize; i++) *this >> ptr[i];
  }
  else {
    read((char *) ptr, sizeof(float)*arraySize);
    if(swap) {
      float *q;
      unsigned int i;
      for(q = ptr, i = 0; i < arraySize; ++i, ++q)
        GFIO::Swap((void *) q, sizeof(float));
    }
  }
}


void GFInput::readArray(int *ptr, unsigned int arraySize)
{
  unsigned int i;
  if(mode == GF_ASCII) {
    for(i = 0; i < arraySize; i++) *this >> ptr[i];
  }
  else {
    read((char *) ptr, sizeof(int)*arraySize);
    if(swap) {
      int *q;
      unsigned int i;
      for(q = ptr, i = 0; i < arraySize; ++i, ++q)
        GFIO::Swap((void *)q, sizeof(int));
    }
  }
}


void GFInput::readArray(short *ptr, unsigned int arraySize)
{
  unsigned int i;
  if(mode == GF_ASCII) {
    for(i = 0; i < arraySize; i++) *this >> ptr[i];
  }
  else {
    read((char *) ptr, sizeof(short)*arraySize);
    if(swap) {
      short *q;
      unsigned int i;
      for(q = ptr, i = 0; i < arraySize; ++i, ++q)
        GFIO::Swap((void *)q, sizeof(short));
    }
  }
}


void GFInput::readArray(Vec3<double> *ptr, unsigned int arraySize)
{
  double *p, *q;
  unsigned int i;
  Vec3<double> *v;

  if(mode == GF_ASCII) {
    for(i = 0; i < arraySize; i++) *this >> ptr[i];
  }
  else {
    p = new double[arraySize * 3];
    read((char *) ptr, sizeof(double) * 3 * arraySize);

    if(swap) {
      for(q = p, i = 0; i < 3 * arraySize; ++i, ++q)
        GFIO::Swap((void *) q, sizeof(double));
    }
    for(i = 0, q = p, v = ptr; i < arraySize; ++i, ++v, q += 3)
      v->setValue(q[0], q[1], q[2]);
    delete p;
  }
}


void GFInput::readArray(Vec2<double> *ptr, unsigned int arraySize)
{
  double *p, *q;
  unsigned int i;
  Vec2<double> *v;

  if(mode == GF_ASCII) {
    for(i = 0; i < arraySize; i++) *this >> ptr[i];
  }
  else {
    p = new double[arraySize * 2];
    read((char *) ptr, sizeof(double) * 2 * arraySize);

    if(swap) {
      for(q = p, i = 0; i < 2 * arraySize; ++i, ++q)
        GFIO::Swap((void *) q, sizeof(double));
    }
    for(i = 0, q = p, v = ptr; i < arraySize; ++i, ++v, q += 2)
      v->setValue(q[0], q[1]);
    delete p;
  }
}


GFOutput& operator<<(GFOutput& output, const bool &val) {
  if(output.getMode() == GFIO::GF_ASCII) ((OFSTREAM&) output) << val << ' ';
  else output.write((char *) &val, sizeof(bool));
  return output;
}

GFOutput& operator<<(GFOutput& output, const char &val) {
  if(output.getMode() == GFIO::GF_ASCII) ((OFSTREAM&) output) << val << ' ';
  else if(val != ' ' && val != '\t' && val != '\n')
    output.write((char *) &val, sizeof(char));
  return output;
}


GFOutput& operator<<(GFOutput& output, const char *val) {
  if(strlen(val) == 1) return output << val[0] << ' ';
  if(output.getMode() == GFIO::GF_ASCII) ((OFSTREAM&) output) << val;
  else
    output.write((char *) &val, strlen(val));
  output.put('\0');
  return output;
}


GFOutput& operator<<(GFOutput& output, const TclString& val) {
  const char c=val[0U];
  if(val.getLength() == 1) return output << c << ' ';
  if(output.getMode() == GFIO::GF_ASCII) ((OFSTREAM&) output) << val;
  else
    output.write((const char *) val, val.getLength());
  output.put('\0');
  return output;
}


GFOutput& operator<<(GFOutput& output, const int &val) {
  if(output.getMode() == GFIO::GF_ASCII) ((OFSTREAM&) output) << val << ' ';
  else output.write((char *) &val, sizeof(int));
  return output;
}


GFOutput& operator<<(GFOutput& output, const unsigned int &val) {
  if(output.getMode() == GFIO::GF_ASCII) ((OFSTREAM&) output) << val << ' ';
  else output.write((char *) &val, sizeof(unsigned int));
  return output;
}


GFOutput& operator<<(GFOutput& output, const short &val) {
  if(output.getMode() == GFIO::GF_ASCII) ((OFSTREAM&) output) << val << ' ';
  else output.write((char *) &val, sizeof(short int));
  return output;
}


GFOutput& operator<<(GFOutput& output, const unsigned short &val) {
  if(output.getMode() == GFIO::GF_ASCII) ((OFSTREAM&) output) << val << ' ';
  else output.write((char *) &val, sizeof(unsigned short int));
  return output;
}


GFOutput& operator<<(GFOutput& output, const long &val) {
  if(output.getMode() == GFIO::GF_ASCII) ((OFSTREAM&) output) << val << ' ';
  else output.write((char *) &val, sizeof(long));
  return output;
}


GFOutput& operator<<(GFOutput& output, const unsigned long &val) {
  if(output.getMode() == GFIO::GF_ASCII) ((OFSTREAM&) output) << val << ' ';
  else output.write((char *) &val, sizeof(unsigned long));
  return output;
}


GFOutput& operator<<(GFOutput& output, const float &val) {
  if(output.getMode() == GFIO::GF_ASCII)
    ((OFSTREAM &) output) << SETPRECISION(10) << val << ' ';
  else output.write((char *) &val, sizeof(float));
  return output;
}


GFOutput& operator<<(GFOutput& output, const double &val) {
  if(output.getMode() == GFIO::GF_ASCII)
    ((OFSTREAM&) output) << SETPRECISION(10) << val << ' ';
  else {
    //output.write((char *) &val, sizeof(double));
    float f = (float) val;
    output.write((char *) &f, sizeof(float));
  }
  return output;
}


GFOutput& operator<<(GFOutput& output, const Vec3<double> &val) {
  if(output.getMode() == GFIO::GF_ASCII)
    ((OFSTREAM&) output) << SETPRECISION(10) << val.getX() << ' '
                         << SETPRECISION(10) << val.getY() << ' '
                         << SETPRECISION(10) << val.getZ() << ' ';
  else ((OFSTREAM&) output) << val.getX() << val.getY() << val.getZ();
  return output;
}
GFOutput& operator<<(GFOutput& output, const Vec3<unsigned int> &val) {
  if(output.getMode() == GFIO::GF_ASCII)
    ((OFSTREAM&) output) <<  val.getX() << ' '
                         <<  val.getY() << ' '
                         <<  val.getZ() << ' ';
  else ((OFSTREAM&) output) << val.getX() << val.getY() << val.getZ();
  return output;
}

GFOutput& operator<<(GFOutput& output, const Rotation3D &val) {
  if(output.getMode() == GFIO::GF_ASCII)
    ((OFSTREAM&) output) << SETPRECISION(10) << val(0,0) << ' '
                         << SETPRECISION(10) << val(0,1) << ' '
                         << SETPRECISION(10) << val(0,2) << ' '
                         << SETPRECISION(10) << val(1,0) << ' '
                         << SETPRECISION(10) << val(1,1) << ' '
                         << SETPRECISION(10) << val(1,2) << ' '
                         << SETPRECISION(10) << val(2,0) << ' '
                         << SETPRECISION(10) << val(2,1) << ' '
                         << SETPRECISION(10) << val(2,2) << ' ';
  else ((OFSTREAM&) output) << val(0,0) << val(0,1) <<
         val(0,2)<< val(1,0) << val(1,1) <<
         val(1,2)<< val(2,0) << val(2,1) <<
         val(2,2) ;
  return output;
}
GFOutput& operator<<(GFOutput& output, const Rotation2D &val) {
  if(output.getMode() == GFIO::GF_ASCII)
    ((OFSTREAM&) output) << SETPRECISION(10) << val(0,0) << ' '
                         << SETPRECISION(10) << val(0,1) << ' '
                         << SETPRECISION(10) << val(1,0) << ' '
                         << SETPRECISION(10) << val(1,1) << ' ';
  else ((OFSTREAM&) output) << val(0,0) << val(0,1) <<
         val(1,0) << val(1,1);
  return output;
}


GFOutput& operator<<(GFOutput& output, const Vec2<double> &val) {
  if(output.getMode() == GFIO::GF_ASCII)
    ((OFSTREAM&) output) << SETPRECISION(10) << val.getX() << ' '
                         << SETPRECISION(10) << val.getY() << ' ';
  else ((OFSTREAM&) output) << val.getX() << val.getY();
  return output;
}




void GFOutput::writeArray(const int *val, unsigned int arraySize) {
  if(mode == GF_ASCII)
    for(unsigned int i = 0; i < arraySize; i++) *this << val[i] << ' ';
  else write((char *) val, sizeof(int) * arraySize);
}

void GFOutput::writeArray(const float *val, unsigned int arraySize) {
  if(mode == GF_ASCII)
    for(unsigned int i = 0; i < arraySize; i++)
      ((OFSTREAM&) *this) << SETPRECISION(10) <<  val[i] << ' ';
  else write((char *) val, sizeof(float) * arraySize);
}

void GFOutput::writeArray(const double *val, unsigned int arraySize) {
  if(mode == GF_ASCII)
    for(unsigned int i = 0; i < arraySize; i++)
      ((OFSTREAM&) *this) << SETPRECISION(10) << val[i] << ' ';
  else write((char *) val, sizeof(double) * arraySize);
}

void GFOutput::writeArray(const Vec2<double> *val,
                          unsigned int arraySize)
{
  if(mode == GF_ASCII)
    for(unsigned int i = 0; i < arraySize; i++) *this << val[i] << ' ';
  else for(unsigned int i = 0; i < arraySize; ++i, val++) *this << (*val);
}

void GFOutput::writeArray(const Vec3<double> *val,
                          unsigned int arraySize)
{
  if(mode == GF_ASCII)
    for(unsigned int i = 0; i < arraySize; i++) *this << val[i] << ' ';
  else for(unsigned int i = 0; i < arraySize; ++i, val++) *this << (*val);
}




#if (defined(_SGI_)||defined(WIN32))
}
yav::GFOutput& endl(yav::GFOutput& output) {
        if(output.getMode() == yav::GFIO::GF_ASCII) ((OFSTREAM&) output) << ENDL;
  return output;
}
#else

GFOutput& endl(GFOutput& output) {
  if(output.getMode() == GFIO::GF_ASCII) ((OFSTREAM&) output) << ENDL;
  return output;
}

}
#endif
