#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation  "basics/GenericHistogram.h"
#include <basics/GenericHistogram.h>
#include <basics/Constants.h>
#include <cstring>

namespace yav {

GenericHistogram::GenericHistogram(const std::map< double, unsigned long int> & h) : data(h)
{ 
  histogram=0;
  minValue=MAX_DOUBLE;
  maxValue=MIN_DOUBLE; 
  if (data.size()>0)
    {
      std::map< double, unsigned long int>::const_iterator it;
      for(it=data.begin();it!=data.end();it++)
	{
	  if (it->first<minValue)
	    minValue=it->first;
	  if (it->first>maxValue)
	    maxValue=it->first;
	}
    }
  guessHistogramParameters();
  buildHistogram();
  computeStatistics(minValue, maxValue);

}
GenericHistogram::GenericHistogram(const std::list<double> & h) 
{ 
  histogram=0;
  minValue=MAX_DOUBLE;
  maxValue=MIN_DOUBLE; 

  if (h.size()>0) {
    std::list<double>::const_iterator it;
    std::pair< std::map< double, unsigned long int>::iterator , bool > res;
    for(it=h.begin();it!=h.end();it++) {
      if ((*it)<minValue)
	minValue=(*it);
      if ((*it)>maxValue)
	maxValue=(*it);
      res=data.insert(std::pair<double,unsigned long int>((*it),1));
      if (res.second==false) {
	data[(*it)]=(*res.first).second+1;
      }
    }
  }

  guessHistogramParameters();
  buildHistogram();
  computeStatistics(minValue, maxValue);

}
GenericHistogram::GenericHistogram(const unsigned long int *v,
				   const unsigned int arraySize,
				   double minVal,  double maxVal) {
  
  unsigned int i,minIndex,maxIndex=arraySize-1;
  double val,scale;

  scale=(maxVal-minVal)/(arraySize-1);
  histogram=0;
  nbBuckets=0;
  const unsigned long int *p=v;
  for(i=0;(i<arraySize) && (*p==0);++i,++p);
  minIndex=i;

  for(;(i<arraySize) ;++i,++p)
    {
      if ((*p)>0)
	{
	  maxIndex=i;
	  val=minVal+ i*scale;
	  data[val]=(*p);
	}
    }

  minValue=minVal+ minIndex*scale;
  maxValue=minVal+ maxIndex*scale;

  guessHistogramParameters();
  buildHistogram();
  computeStatistics(minValue,maxValue);
}
GenericHistogram::GenericHistogram(const GenericHistogram& h) : Pointed<GenericHistogram>(),data(h.data),
  nbBuckets(h.nbBuckets),bucketSize(h.bucketSize),
  minValue(h.minValue),maxValue(h.maxValue){
  histogram=(unsigned long int *) malloc(nbBuckets*sizeof(unsigned long int));
  memcpy ((void *)histogram,(void *)h.histogram,nbBuckets*sizeof(unsigned long int));
  computeStatistics(minValue,maxValue);
}
GenericHistogram::GenericHistogram(char * name) {
  IFSTREAM file(name);
  unsigned int nData;
  double bSize;
  file>>nData;
  file>>bSize;
  double  x;
  unsigned long int y;
  minValue=MAX_DOUBLE;
  maxValue=MIN_DOUBLE;
  for(;nData>0;nData--){
    file>> x >> y;
    data[x]=y;
    if (x<minValue)
      minValue=x;
    if (x>maxValue)
      maxValue=x;
  }
  file.close();
  histogram=0;
  buildHistogram(bSize);
  computeStatistics(minValue,maxValue);
}
void GenericHistogram::computeStatistics() {
  double meanSqr;
  nbSamples=0;
  meanValue=meanSqr=0.0;
  double q;
  std::map< double, unsigned long int>::const_iterator it;
  for(it=data.begin();it!=data.end();it++)
    {
      q=nbSamples/it->second;
      meanSqr = (meanSqr*q)/(1+q)+(it->first*it->first)/(1+q); 
      meanValue = (meanValue*q)/(1+q)+(it->first)/(1+q); 
      nbSamples+= it->second;
    }
  standardDeviation= meanSqr -meanValue*meanValue;
  standardDeviation= ::sqrt(standardDeviation);
  medianValue=getHistogramQuantile(0.5);
}
void GenericHistogram::computeStatistics(double min, double max) {
  double meanSqr;
  nbSamples=0;
  meanValue=meanSqr=0.0;
  double q;
  std::map< double, unsigned long int>::const_iterator it;
  for(it=data.begin();it!=data.end();it++)
    {
      if ((it->first>=min) && (it->first<=max)) {
	q=nbSamples/it->second;
	meanSqr = (meanSqr*q)/(1+q)+(it->first*it->first)/(1+q); 
	meanValue = (meanValue*q)/(1+q)+(it->first)/(1+q); 
	nbSamples+= it->second;
      }
    }
  standardDeviation= meanSqr -meanValue*meanValue;
  standardDeviation= ::sqrt(standardDeviation);
  medianValue=getHistogramQuantile(0.5, min, max);
}
double GenericHistogram::getHistogramQuantile(const double q) const {
  if ((q>=0) && (q<=1)) {
    unsigned long int n=0;
    unsigned long int target=(unsigned long int)(nbSamples*q);
    std::map< double, unsigned long int>::const_iterator it;
    for(it=data.begin();it!=data.end();it++)
      {
	n+=it->second;
	if (n>target)
	  return it->first;
      }
    return maxValue;
  }
  else return minValue;
}
double GenericHistogram::getHistogramQuantile(const double q, double min, double max) const {
  std::map< double, unsigned long int>::const_iterator it;
    unsigned long int sample =0;
  if ((q>=0) && (q<=1)) {
    unsigned long int n=0;
    
    for (it=data.begin();it!=data.end();it++) {
      if ((it->first>=min) && (it->first<=max)) {
	sample+=it->second;
      }
    }
    unsigned long int target=(unsigned long int)(sample*q);

    std::map< double, unsigned long int>::const_iterator it;
    for(it=data.begin();it!=data.end();it++)
      {
	if ((it->first>=min) && (it->first<=max)) {
	  n+=it->second;
	  if (n>target)
	    return it->first;
	}
      }
    return maxValue;
  }
  else return minValue;
}
void GenericHistogram::guessHistogramParameters() {
  if ((data.size()<=256) && ((maxValue-minValue)<256))
    {
      if   (((maxValue-minValue)>50))
	nbBuckets=(unsigned int)(maxValue-minValue+1);
      else
      nbBuckets=data.size();
    }
  else
    nbBuckets=256;
  bucketSize=(maxValue-minValue)/nbBuckets;
}
unsigned long int  GenericHistogram::getMaxNbSamples () const {
  unsigned long int max=0;
  unsigned int i;
  for(i=0;i<nbBuckets;++i)
    {

      if (histogram[i]>max)
	{
	  max=histogram[i];
	}
    }
  return max;

}
unsigned long int  GenericHistogram::getNbSamples (const double  x) {
  if ( (data.size()>0) && (bucketSize>0) && (x>=minValue) && (x<=maxValue))
    {
      if (histogram==0) 
	buildHistogram();
      return histogram[(unsigned int )((x-minValue)/bucketSize +.5)];
    }
  else return 0;
}
void GenericHistogram::buildHistogram() {
  if (data.size()>0)
    {
      if (histogram)
	histogram=(unsigned long int *) realloc((void *)histogram,nbBuckets*sizeof(unsigned long int));
      else
	histogram=(unsigned long int *) malloc(nbBuckets*sizeof(unsigned long int));

      memset((void *)histogram,0,nbBuckets*sizeof(unsigned long int));

      std::map< double, unsigned long int>::const_iterator it;
      unsigned int i;
      COUT << ENDL<< ENDL;
      for(it=data.begin();it!=data.end();it++)
	{
	  i= (unsigned int ) ((((it->first)-minValue)/bucketSize) +.5);
	  if (i==255) {
	    COUT << it->first << "    " << it->second << ENDL;
	    COUT << minValue << ENDL;
	    COUT << bucketSize << ENDL;
	  }
	  histogram[i]+=it->second;
	}
      
    }
}
void GenericHistogram::write(char* name) const {
  OFSTREAM file(name);
  unsigned nData=data.size();
  file<<nData<<ENDL;
  file<<bucketSize<<ENDL;

  std::map< double, unsigned long int>::const_iterator it;
  for(it=data.begin();it!=data.end();it++)
    {
      file<<it->first<<" "<<it->second<<ENDL;
    }
  file.close();
}

void GenericHistogram::smooth(double sigma,SmoothingFilter f)
{
	if (f==GAUSSIAN_SMOOTHING)
	{
		double newval,weight,totalWeight,x;
		unsigned int i,k,delta;
		int j;

		unsigned long int *hcopy;
		hcopy=(unsigned long int *) malloc(nbBuckets*sizeof(unsigned long int));
		memcpy ((void *)hcopy,(void *)histogram,nbBuckets*sizeof(unsigned long int));

		delta=(unsigned int )((3.0*sigma/bucketSize)+.5);
		for(i=0;i<nbBuckets;++i)
		{
			newval=hcopy[i];
			totalWeight=1.0;
			for(j=i-1;(j>=0) && ((i-j)<delta);--j)
			{
				x=(i-j)*bucketSize;
				weight=exp(-0.5*x*x/(sigma*sigma));
				newval+=weight*hcopy[j];
				totalWeight+=weight;
			}
			for(k=i+1;(k<nbBuckets) && ((k-i)<delta);++k)
			{
				x=(k-i)*bucketSize;
				weight=exp(-0.5*x*x/(sigma*sigma));
				newval+=weight*hcopy[k];
				totalWeight+=weight;
			}
			newval/=totalWeight;
			histogram[i]= (unsigned long int) newval;
		}
		free(hcopy);
	}
}

} // namespace yav

