#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 4503)
#endif

#ifndef WIN32
#pragma implementation "basics/IniFile.h"
#endif

#include "basics/Architecture.h"
#include "basics/IniFile.h"

#include <string.h>
#include <stdio.h>

#include <iostream>
#include <fstream>
#include <iomanip>



#ifdef WIN32
#include "basics/export.h"
#endif

#if ((!defined(WIN32))&&(!defined(_BCPLUSPLUS_)))
int GetPrivateProfileString (const char * AppName, const char * KeyName, const char * Default, char * Buffer, int MaxLength, const char * FileName)
{
  char line[255];
  IFSTREAM input(FileName);

  if (KeyName) {
    if (input) {
      bool OK=false;
      while (!input.eof()) {
        input >> line;
        if (strncmp(line,"//",2)==0) continue;
        char entry[255];
        if (sscanf(line,"[%[^]]",entry)==1) {
          OK=(strcmp(entry,AppName)==0);
          continue;
        }
        if (OK) {
          char key[255];
          char value[255];
          if ((sscanf(line,"%[^=]=%s",key,value)==2)&&(strcmp(key,KeyName)==0)) {
            strcpy(Buffer,value);
            return strlen(Buffer);
          }
          if ((sscanf(line,"%[^=]=",key)==1)&&(strcmp(key,KeyName)==0)) {
            strcpy(Buffer,"");
            return strlen(Buffer);
          }
        }
      }
    }
    strcpy(Buffer,Default);
    return strlen(Buffer);
  } else {
    bool sectionexists=false;
    Buffer[0]=0;
    int nbchar=0;

    if (input) {
      bool OK=false;
      while (!input.eof()) {
        input >> line;
        char entry[255];
        if (sscanf(line,"[%[^]]",entry)==1) {
          OK=(strcmp(entry,AppName)==0);
          if (OK) sectionexists=true;
          continue;
        }
        if (OK) {
          char key[255];
          if ((sscanf(line,"%[^=]=",key)==1)) {
            strcpy(Buffer+nbchar,key);
            nbchar+=strlen(key)+1;
            continue;
          }
        }
      }
    }
    if (!sectionexists) {
      strcpy(Buffer,Default);
      nbchar=strlen(Default)+1;
    }
    Buffer[nbchar]=0;
    return nbchar;
  }
}
#endif


namespace yav {


bool GetInitFileString (const char * FileName, const char * EntryName, const char * KeyName, char * Buffer, const char * Default, bool warning)
{
  GetPrivateProfileString (EntryName, KeyName, Default, Buffer, 255, FileName);

  bool result=(strcmp(Buffer, Default)!=0);
  if ((!result)&&(warning)) CERR << "Erreur  la lecture de la cl " << EntryName << "-" << KeyName << " dans le fichier " << FileName << ENDL;

  return result;
}

bool GetInitFileStringND (const char * FileName, const char * EntryName, const char * KeyName, char * Buffer, bool warning)
{
  return GetInitFileString (FileName, EntryName, KeyName, Buffer, "Error", warning);
}


bool GetInitFileInt (const char * FileName, const char * EntryName, const char * KeyName, int& Buffer, int Default, bool warning)
{
  char Def[] = "Error";
  char Buf[255];
  GetPrivateProfileString (EntryName, KeyName, Def, Buf, 255, FileName);

  bool result=(sscanf(Buf,"%d",&Buffer)==1);
  if (!result) {
    Buffer=Default;
    if (warning)
      CERR << "Erreur  la lecture de la cl " << EntryName << "-" << KeyName << " dans le fichier " << FileName << ENDL;
  }
  return result;
}

bool GetInitFileIntND (const char * FileName, const char * EntryName, const char * KeyName, int& Buffer, bool warning)
{
  return GetInitFileInt (FileName, EntryName, KeyName, Buffer, 0, warning);
}


bool GetInitFileFloat (const char * FileName, const char * EntryName, const char * KeyName, float& Buffer, float Default, bool warning)
{
  char Def[] = "Error";
  char Buf[255];
  GetPrivateProfileString (EntryName, KeyName, Def, Buf, 255, FileName);

  bool result=(sscanf(Buf,"%f",&Buffer)==1);
  if (!result) {
    Buffer=Default;
    if (warning)
      CERR << "Erreur  la lecture de la cl " << EntryName << "-" << KeyName << " dans le fichier " << FileName << ENDL;
  }
  return result;
}

bool GetInitFileFloatND (const char * FileName, const char * EntryName, const char * KeyName, float& Buffer, bool warning)
{
  return GetInitFileFloat (FileName, EntryName, KeyName, Buffer, 0.0, warning);
}


bool GetInitFileDouble (const char * FileName, const char * EntryName, const char * KeyName, double& Buffer, double Default, bool warning)
{
  char Def[] = "Error";
  char Buf[255];
  GetPrivateProfileString (EntryName, KeyName, Def, Buf, 255, FileName);

  bool result=(sscanf(Buf,"%lf",&Buffer)==1);
  if (!result) {
    Buffer=Default;
    if (warning)
      CERR << "Erreur  la lecture de la cl " << EntryName << "-" << KeyName << " dans le fichier " << FileName << ENDL;
  }
  return result;
}

bool GetInitFileDoubleND (const char * FileName, const char * EntryName, const char * KeyName, double& Buffer, bool warning)
{
  return GetInitFileDouble (FileName, EntryName, KeyName, Buffer, 0.0, warning);
}


bool GetInitFileBool (const char * FileName, const char * EntryName, const char * KeyName, bool& Buffer, bool Default, bool warning)
{
  char Def[] = "Error";
  char Buf[255];
  GetPrivateProfileString (EntryName, KeyName, Def, Buf, 255, FileName);

  try {
    Buffer=atobe(Buf);return true;
  }
  catch (...) {
    Buffer=Default;
    return false;
  }
}

bool GetInitFileBoolND (const char * FileName, const char * EntryName, const char * KeyName, bool& Buffer, bool warning)
{
  return GetInitFileBool (FileName, EntryName, KeyName, Buffer, false, warning);
}

}


IniFile::IniFile(const char* filename)
{
  Read(filename);
}

IniFile::~IniFile()
{
  Clear();
}

void IniFile::Read(const char* filename)
{
  char line[255];
  IFSTREAM input(filename);

  if (input) {
    char entry[255];
    KeyMap* keyMap=0;
    strcpy(entry,"");
    while (!input.eof()) {
      input >> line;
      if (strncmp(line,"//",2)==0) continue;
      if (sscanf(line,"[%[^]]",entry)==1) {
        EntryMap::iterator ie=entryMap.find(entry);
        if (ie==entryMap.end())
          ie=entryMap.insert(std::pair<std::string,KeyMap>(entry,KeyMap())).first;
        keyMap=&(ie->second);
        continue;
      }
      if (keyMap) {
        char key[255];
        char value[255];
        if ((sscanf(line,"%[^=]=%s",key,value)==2)) {
          KeyMap::iterator ie=keyMap->find(key);
          if (ie==keyMap->end())
            keyMap->insert(std::pair<std::string,std::string>(key,value));
          else {
            CERR << "Trying to set a new value for " << key << ENDL;
          }
          continue;
        }
        if ((sscanf(line,"%[^=]=",key)==1)) {
          KeyMap::iterator ie=keyMap->find(key);
          if (ie==keyMap->end())
            keyMap->insert(std::pair<std::string,std::string>(key,""));
          else {
            CERR << "Trying to set a new value for " << key << ENDL;
          }
          continue;
        }
      }
    }
  }
}

void IniFile::Clear()
{
  entryMap.clear();
}

int IniFile::GetString(const char* EntryName, const char* KeyName, char* Buffer, const char* Default) const
{
  EntryMap::const_iterator ie;
  KeyMap::const_iterator is;
  if (((ie=entryMap.find(EntryName))!=entryMap.end())
      &&((is=ie->second.find(KeyName))!=ie->second.end())) {
    strcpy(Buffer,is->second.c_str());return 1;
  } else {
    strcpy(Buffer,Default);return 0;
  }
}
