#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#ifndef WIN32
#pragma implementation "basics/Normal3D.h"
#endif
#include "basics/Normal3D.h"
#include <vector>

#ifdef WIN32
#include "basics/export.h"
#endif

namespace yav {

const double TH_TINY = 1e-5;

void Normal3D::normalise(const Vec3<double>& v) throw(Normal3D::ErrNonNormal)
{
  double norm2=v.norm2();
  if (norm2<TH_TINY) throw(ErrNonNormal("Normal3D::normalise"));
  double norm= ::sqrt(norm2);
  val[0]=v[0]/norm;
  val[1]=v[1]/norm;
  val[2]=v[2]/norm;
}

void Normal3D::rnd() throw()
{
  Vec3<double>& n=*this;
  double norm2;
  do {
    n.Rnd();
    norm2 = n.norm2();
  } while ((norm2<0.1)||(norm2>1));
  n*=1/::sqrt(norm2);
}

void Normal3D::rndgaussian(const Normal3D& n, double ecarttype) throw()
{
  Vec3<double> noise;
  noise.rndgaussian(ecarttype*M_SQRT1_2);
  // TODO: Est ce vraiment necessaire ???
  noise-=n*DotProd(((const Vec3<double>&)n),noise);
  this->Apply(RotMat3D((const RotVec3D&) noise),n);
}

double DistAngle(const Normal3D& n1, const Normal3D& n2) throw()
{
  double cos=DotProd((const Vec3<double>&)n1,(const Vec3<double>&)n2);
  if (cos>1) return 0;
  if (cos<-1) return M_PI;
  return acos(cos);
}

double DistVect(const Normal3D& n1, const Normal3D& n2) throw()
{
  return dist((const Vec3<double>&)n1,(const Vec3<double>&)n2);
}

double DistAngleNO(const Normal3D& n1, const Normal3D& n2) throw()
{
  double cos=DotProd((const Vec3<double>&)n1,(const Vec3<double>&)n2);
  if (cos<0) cos=-cos;
  if (cos>1) return 0;
  return acos(cos);
}

double DistVectNO(const Normal3D& n1, const Normal3D& n2) throw()
{
  double cos=DotProd((const Vec3<double>&)n1,(const Vec3<double>&)n2);
  if (cos>0)
    return dist((const Vec3<double>&)n1,(const Vec3<double>&)n2);
  else
    return dist(-(const Vec3<double>&)n1,(const Vec3<double>&)n2);
}

double DistTens(const Normal3D& n1, const Normal3D& n2) throw()
{
  double cos=DotProd((const Vec3<double>&)n1,(const Vec3<double>&)n2);
  if ( ::fabs(cos)>1 ) return 0;
  return ::sqrt(1-cos*cos);
}

// Return canonical rotation R so that R*(*this)=m
void Normal3D::m2Rnm(const Normal3D& m, RotVect& f) const throw()
{
  Vec3<double> r;
  double sin,cos;
  r=cross(*this,m);
  sin=r.Norm();
  cos=DotProd(*this,m);
  if (sin<0.000001) {
    // Cas ou les deux normales sont alignees
    if (cos>0) {
      // elles sont identiques
      r.resetToNull();
    } else {
      // elles sont opposees : on cherche un vecteur perpendiculaire
      r=cross(*this,Vec3<double>(0,0,1));
      sin=r.Norm();
      if (sin<0.000001) {
	r=cross(*this,Vec3<double>(0,1,0));
	sin=r.Norm();
      }
      r*=M_PI/sin;
    }
  } else {
    // Cas ou les deux normales ne sont pas alignees
    if (cos>0) {
      r*=asin(sin)/sin;
    } else {
      r*=(M_PI-asin(sin))/sin;
    }
  }
  f=r;
}

void Normal3D::N2mnTens(const Normal3D& _m, Vec3<double>& mn) const throw()
{
  const Vec3<double>& m=(const Vec3<double>&)_m;
  const Vec3<double>& n=(const Vec3<double>&)*this;
  if (DotProd(m,n)>0) mn=n;else mn=-n;
  mn-=m*DotProd(n,m);
}

void Normal3D::N2mnVect(const Normal3D& _m, Vec3<double>& mn) const throw()
{
  const Vec3<double>& m=(const Vec3<double>&)_m;
  const Vec3<double>& n=(const Vec3<double>&)*this;
  mn.sub(n,m*DotProd(n,m));
  double norm=mn.norm();
  if (norm>0) mn*=DistVect(m,n)/norm;
}

void Normal3D::N2mnAngl(const Normal3D& _m, Vec3<double>& mn) const throw()
{
  const Vec3<double>& m=(const Vec3<double>&)_m;
  const Vec3<double>& n=(const Vec3<double>&)*this;
  mn.sub(n,m*DotProd(n,m));
  double norm=mn.norm();
  if (norm>0) mn*=DistAngle(m,n)/norm;
}

void Normal3D::mnAngl2N(const Normal3D& _m, const Vec3<double>& mn) throw()
{
  const Vec3<double>& m=(const Vec3<double>&)_m;
  this->Apply(RotMat3D((const RotVect&) cross(m,mn)),m);
}

void Normal3D::mnTens2N(const Normal3D& m, const Vec3<double>& mn) throw()
{
  Vec3<double> mnAngle(mn);
  double sin=mn.norm();
  if (sin>0.00001) mnAngle*=asin(sin)/sin;
  this->mnAngl2N(m,mnAngle);
}

void Normal3D::mnVect2N(const Normal3D& m, const Vec3<double>& mn) throw()
{
  Vec3<double> mnAngle(mn);
  double sin=mn.norm()/2;
  if (sin>0.00001) mnAngle*=asin(sin)/sin;
  this->mnAngl2N(m,mnAngle);
}


void Normal3D::Test()
{
  // TEST des fonctions N2mn.. et mn..2N
  int i;
  double maxerrort=0,maxerrora=0,maxerrorv=0,d;
  Normal3D m,n1,n2;
  Vec3<double> mn;
  for(i=0;i<100000;i++) {
    m.rnd();
    n1.rnd();
    //    if (DotProd(m,n1)<0) n1=-n1;

    n1.N2mnTens(m,mn);
    n2.mnTens2N(m,mn);
    d=DistAngle(n1,n2);
    if (maxerrort<d) maxerrort=d;

    n1.N2mnAngl(m,mn);
    n2.mnAngl2N(m,mn);
    d=DistAngle(n1,n2);
    if (maxerrora<d) maxerrora=d;

    n1.N2mnVect(m,mn);
    n2.mnVect2N(m,mn);
    d=DistAngle(n1,n2);
    if (maxerrorv<d) maxerrorv=d;
  }
  COUT << "Max error with nmAngl : " << maxerrora << ENDL;
  COUT << "Max error with nmVect : " << maxerrorv << ENDL;
  COUT << "Max error with nmTens : " << maxerrort << ENDL;

  // TEST de MEAN
  int nbnormals=100;double ecarttype=0.2;
  std::vector<Normal3D> nlist;
  Normal3D realmean,vectmean,tensmean,anglmean,tmp;
  // Calcul des normales aleatoires
  for (i=0; i<nbnormals; i++) nlist.push_back(tmp.RndGaussian(realmean,ecarttype));
  vectmean.meanVect(nlist.begin(),nlist.end());
  tensmean.meanTens(nlist.begin(),nlist.end());
  anglmean.meanAngl(nlist.begin(),nlist.end());

  COUT << "" <<  DistAngle(anglmean,realmean);
  COUT << "\t" <<  DistAngle(vectmean,realmean);
  COUT << "\t" << DistAngle(vectmean,anglmean);
  COUT << "\t" <<  DistAngle(tensmean,realmean);
  COUT << "\t" << DistAngle(tensmean,anglmean);
  COUT << ENDL;
}

} // namespace yav
