#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 4355)
#endif

#pragma implementation "basics/SharedMemory.h"

#define _XOPEN_SOURCE 500

#ifdef _SGI_
#include <sys/types.h>
#endif


#include <basics/SharedMemory.h>

#ifdef _SGI_
#include <sys/types.h>
#endif 

#ifndef WIN32
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#endif

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <basics/Architecture.h>
#include <basics/TclString.h>


namespace yav {

SharedMemory::SharedMemory(int _key ,int _size){
  if(_size==0){
    TclString from;
    from.sprintf("SharedMemory::SharedMemory(int %d ,int %d)",_key,_size);
    throw IllegalData((char*)from);
  }
  createSegment(_key,_size);
}

SharedMemory::SharedMemory(int data,CreateWith mode){
  switch(mode){
  case KEY:
    getSegment(data);
    break;
  case SIZE:
    createSegment(data);
    break;
  }
}

SharedMemory::~SharedMemory(){
#ifndef WIN32
#ifndef _SOLARIS_
  shmdt(shmaddr);
  if(main){
    shmctl(shmid,IPC_RMID,0);
    semctl(semid,0,IPC_RMID);
  }
#endif
#endif
}

void SharedMemory::createSegment(int _key, int _size) {
#ifndef WIN32
  shmid = shmget(_key,_size,IPC_CREAT|IPC_EXCL|0644);
  if(shmid==-1){
    TclString from;
    from.sprintf("SharedMemory::createSegment(int %d,int %d)",_key,_size);
    throw CanNotCreateSegment((char*)from,strerror(errno));
  }
  semid = semget(_key,1,IPC_CREAT|IPC_EXCL|0644);
  if(semid==-1){
    TclString from;
    from.sprintf("SharedMemory::createSegment(int %d,int %d)",_key,_size);
    shmctl(shmid,IPC_RMID,0); // remove shm if semaphore creation fails
    throw CanNotCreateSemaphore((char*)from,strerror(errno));
  }
  shmkey = _key;
  shmsz = _size;
  main = true;
  shmaddr = shmat(shmid,0,0);
  semctl(semid,0,SETVAL,1);
#endif
}

void SharedMemory::createSegment(int _size) {
#ifndef WIN32
  int _key = 0;
  do{
    do{
      ++_key;
      shmid = shmget(_key,_size,IPC_CREAT|IPC_EXCL|0644);
    }while((shmid==-1)&&(errno==EEXIST));
    if(shmid==-1){
      TclString from;
      from.sprintf("SharedMemory::createSegment(int %d)",_size);
      throw CanNotCreateSegment((char*)from,strerror(errno));
    }
    semid =  semget(_key,1,IPC_CREAT|IPC_EXCL|0644);
    if(semid==-1) shmctl(shmid,IPC_RMID,0); // remove shm if semaphore creation fails
  }while((semid==-1)&&(errno==EEXIST));
  if(semid==-1){
    TclString from;
    from.sprintf("SharedMemory::createSegment(int %d)",_size);
    throw CanNotCreateSemaphore((char*)from,strerror(errno));
  }
  shmkey = _key;
  shmsz = _size;
  main = true;
  //ushort_t val;
#ifdef _SGI_
  semun val;
  val.val = 1;
#else
  int val = 1;
#endif
  shmaddr = shmat(shmid,0,0);
  semctl(semid,0,SETVAL,val);
#endif
}

void SharedMemory::getSegment(int _key) {
#ifndef WIN32
  shmid = shmget(_key,0,0644);
  if(shmid==-1){
    TclString from;
    from.sprintf("SharedMemory::getSegment(int %d)",_key);
    std::string f = (char*)from;
    std::string reason = strerror(errno);
    throw CanNotCreateSegment(f,reason);
  }
  semid = semget(_key,0,0644);
  if(semid==-1){
    TclString from;
    from.sprintf("SharedMemory::getSegment(int %d)",_key);
    throw CanNotCreateSemaphore((char*)from,strerror(errno));
  }
  struct shmid_ds shmds;
  shmctl(shmid,IPC_STAT,&shmds);
  shmkey = _key;
  shmsz = shmds.shm_segsz;
  main = false;
  shmaddr = shmat(shmid,0,0);
#endif
}
bool SharedMemory::checkSemaphore(){
#ifndef WIN32
  if(semctl(semid,0,GETVAL)!=0) return false;
  if(semctl(semid,0,GETPID)!=getpid()) return false;
  return true;
#else
  return true;
#endif
}

int SharedMemory::getSemaphore(){
#ifndef WIN32
  struct sembuf buf;
  buf.sem_num=0;
  buf.sem_op=-1;
  buf.sem_flg=0;
  return semop(semid,&buf,1);
#else
  return 0;
#endif
}
bool SharedMemory::getSemaphoreNoWait(){
#ifndef WIN32
  struct sembuf buf;
  buf.sem_num=0;
  buf.sem_op=-1;
  buf.sem_flg=IPC_NOWAIT;
  return semop(semid,&buf,1);
#else
  return true;
#endif
}
int SharedMemory::releaseSemaphore(){
#ifndef WIN32
  struct sembuf buf;
  buf.sem_num=0;
  buf.sem_op=1;
  buf.sem_flg=IPC_NOWAIT;
  return semop(semid,&buf,1);
#else
  return 0;
#endif

}

short SharedMemory::getNbOfAttaches(){
#ifndef WIN32
  struct shmid_ds shmds;
  shmctl(shmid,IPC_STAT,&shmds);
  return shmds.shm_nattch;
#else
  return 0;
#endif

}


} // namespace yav
