#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "basics/TclString.h"
#include <basics/TclString.h>


namespace yav {

#ifndef WIN32
  const TclString EMPTY_STRING;
  const TclString SPACE_STRING(" ");
#else
  /*static const TclString EMPTY_STRING;
  static const TclString SPACE_STRING(" ");*/
#endif

// Destructor
TclString::~TclString() {
  delete[] str;
}


// Constructors
TclString::TclString() : length (1) {
  str = new char[1];
  str[0] = '\0';
}

TclString::TclString(const char *init) {
  if(init) {
    length = strlen(init) + 1;
    copy(init);
  }
  else {
    length = 1;
    str = new char[1];
    str[0] = '\0';
  }
}

TclString::TclString(const TclString &init) {
  length = init.length;
  copy(init.str);
}


TclString::TclString(const char c) {
  length = 2;
  str = new char[2];
  str[0] = c;
  str[1] = '\0';
}


TclString::TclString(const char *init, const unsigned int start,
               const unsigned int stop)
  throw(TclString::InvalidIndex, TclString::OutOfString) {

  char from[] = "TclString::TclString(const char *, const unsigned int, const unsigned int)";

  if(!init) throw OutOfString(from);
  if(stop < start) throw InvalidIndex(from);
  if(stop >= strlen(init)) throw OutOfString(from);

  length = stop - start + 2;
  str = new char[length];
  strncpy(str, init + start, stop - start + 1);
  str[length - 1] = '\0';
}


TclString::TclString(const TclString &init, const unsigned int start,
               const unsigned int stop)
  throw(TclString::InvalidIndex, TclString::OutOfString) {

  char from[] = "TclString::TclString(const TclString&, const unsigned int, const unsigned int)";

  if(stop < start) throw InvalidIndex(from);
  if(stop >= init.getLength()) throw OutOfString(from);

  length = stop - start + 2;
  str = new char[length];
  strncpy(str, (char *) init + start, stop - start + 1);
  str[length - 1] = '\0';
}


TclString::TclString(const size_t l) : length(l+1) {
  str = new char[length];
  str[0] = '\0';
}


// private initalization routine
void TclString::copy(const char *init) {
  str = new char[length];
  strcpy(str, init);
}


// Assignement operators
TclString& TclString::operator=(const char *init) {
  delete[] str;
  if(init) {
    length = strlen(init) + 1;
    copy(init);
  }
  else {
    length = 1;
    str = new char[1];
    str[0] = '\0';
  }
  return *this;
}

TclString& TclString::operator=(const TclString &init) {
  delete[] str;
  length = init.length;
  copy(init.str);
  return *this;
}


// Concatenation operators
TclString& TclString::operator+=(const TclString &apstr) {
  char *tmp;

  if(apstr.length > 0) {
    length += apstr.length - 1;
    tmp = new char[length];
    strcpy(tmp, str);
    delete[] str;
    str = tmp;
    strcat(str, apstr.str);
  }
  return *this;
}

TclString& TclString::operator+=(const char *apstr) {
  char *tmp;

  if(apstr) {
    length += strlen(apstr);
    tmp = new char[length];
    strcpy(tmp, str);
    delete[] str;
    str = tmp;
    strcat(str, apstr);
  }
  return *this;
}


TclString& TclString::prepend(const TclString &prepstr) {
  char *tmp;

  if(prepstr.length > 0) {
    length += prepstr.length - 1;
    tmp = new char[length];
    strcpy(tmp, prepstr.str);
    strcat(tmp, str);
    delete[] str;
    str = tmp;
  }
  return *this;
}


TclString& TclString::prepend(const char *prepstr) {
  char *tmp;

  if(prepstr) {
    length += strlen(prepstr);
    tmp = new char[length];
    strcpy(tmp, prepstr);
    strcat(tmp, str);
    delete[] str;
    str = tmp;
  }
  return *this;
}



TclString TclString::operator+(const TclString &sc) const {
  TclString sr(length-1 + sc.getLength());
  strcpy(sr.str, str);
  strcat(sr.str, sc.str);
  return sr;
}


TclString TclString::operator+(const char *sc) const {
  if(sc) {
    TclString sr(length-1 + strlen(sc));
    strcpy(sr.str, str);
    strcat(sr.str, sc);
    return sr;
  }
  else return *this;
}


#ifndef WIN32
// index operator
char& TclString::operator[](int n) const throw(OutOfString) {
  if(n >= ((int)length) - 1) throw OutOfString("TclString::operator[]");
  return str[n];
}
#endif


TclString& TclString::truncAtBegining(unsigned int i) {
  if(i > length - 1) i = length - 1;
  length -= i;
  char *tmp = new char[length];
  strcpy(tmp, str + i);
  delete[] str;
  str = tmp;
  return *this;
}


TclString& TclString::truncAtEnd(unsigned int i) {
  if(i > length - 1) i = length - 1;
  length -= i;
  char *tmp = new char[length];
  strncpy(tmp, str, length - 1);
  delete[] str;
  tmp[length - 1] = '\0';
  str = tmp;
  return *this;
}



void TclString::toUpper(void) {
  if(str) {
    for(unsigned int i = 0; i < length; i++)
      str[i] = toupper(str[i]);
  }
}

void TclString::toLower(void) {
  if(str) {
    for(unsigned int i = 0; i < length; i++)
      str[i] = tolower(str[i]);
  }
}

void TclString::toAscii(void) {
  if(str) {
    for(unsigned int i = 0; i < length; i++)
      str[i] &= 0x7F;
  }
}

void TclString::toUpper(unsigned int n) throw(TclString::OutOfString) {
 if(n < length) str[n] = toupper(str[n]);
 else throw OutOfString("TclString::toUpper(unsigned int)");
}

void TclString::toLower(unsigned int n) throw(TclString::OutOfString) {
 if(n < length) str[n] = tolower(str[n]);
 else throw OutOfString("TclString::toLower(unsigned int)");
}

void TclString::toAscii(unsigned int n) throw(TclString::OutOfString) {
 if(n < length) str[n] &= 0x7F;
 else throw OutOfString("TclString::toAscii(unsigned int)");
}



// output operator
STD_OSTREAM & operator<<(STD_OSTREAM & os, const TclString &s) {
  os << s.str;
  return os;
}


// input operator
STD_ISTREAM & operator>>(STD_ISTREAM & is, TclString &s) {
  char buf[100], c;
#if (defined __GNUC__) && (__GNUC__ >= 3) // bug ? when line starts with \n
  while(is.peek()=='\n') is.get();        // get() does not seem
#endif                                    // to extract anything from is
  is.get(buf, 100);
  s = buf;
  while(is.get(c) && c != '\n') {
  is.get(buf+1, 99);
    buf[0] = c;
    s += buf;
  }
  return is;
}


// char * + TclString
TclString operator+(const char *s1, const TclString& s2) {
  TclString s3(strlen(s1) + s2.length-1);
  strcpy(s3.str, s1);
  strcat(s3.str, s2.str);
  return s3;
}



void TclString::realloc(size_t n) {
  char *realloced;
  size_t l;

  if(n == 0) {
    l = strlen(str) + 1;
    if(length != l) {
      length = l;
      realloced = new char[l];
      strcpy(realloced, str);
      delete[] str;
      str = realloced;
    }
  }
  else {
    if(length != n) {
      length = n;
      realloced = new char[n];
      strncpy(realloced, str, n - 1);
      realloced[n-1] = '\0';
      delete[] str;
      str = realloced;
    }
  }
}


void TclString::concat(const char *conc) {
  size_t l, nl;

  l = strlen(conc) + strlen(str) + 1;
  if(l > length) {
    nl = 2 * length;
    if(l > nl) nl = l;
    realloc(nl);
  }
  strcat(str, conc);
}



void TclString::concat(const char c) {
  size_t l, nl;

  l = strlen(str) + 2;
  if(l > length) {
    nl = 2 * length;
    if(l > nl) nl = l;
    realloc(nl);
  }
  str[l - 2] = c;
  str[l - 1] = '\0';
}



void TclString::sprintf(const char *s...) {
  va_list args;

  delete[] str;

  va_start(args, s);
  isprintf(s, args);
  va_end(args);
}


void TclString::isprintf(const char *s, va_list args) {
  char format[20]="", tmp[400], *as;
  unsigned int i, k;
  bool flg;

  length = 20;
  str = new char[20];
  str[0] = '\0';

  /* for each char of the format string ... */
  for(i = 0; i < strlen(s);) {
    /* if character is not '%' copy it */
    if(s[i] != '%') {
      concat(s[i]);
      i++;
    }

    /* if char is '%' ... */
    else {

      /* %x where x != '%', format... */
      if(s[i+1] != '%') {

        /* get format string */
        k = 0;
        flg = true;
        while(flg && k < 19) {
          format[k++] = s[i++];
          if(strchr("dioxXueEfgGcsSpn", s[i-1])) {
            flg = false;
            format[k] = '\0';
          }
        }

        /* process format and append to argument */
        switch(s[i-1]) {
          case 'd' :
          case 'i' :
          case 'o' :
          case 'x' :
          case 'X' :
            ::sprintf(tmp, format, va_arg(args, int));
            concat(tmp);
            break;
          case 'u' :
            ::sprintf(tmp, format, va_arg(args, unsigned int));
            concat(tmp);
            break;
          case 'p' :
            ::sprintf(tmp, format, va_arg(args, void *));
            concat(tmp);
            break;
          case 'e' :
          case 'f' :
          case 'g' :
          case 'E' :
          case 'G' :
            ::sprintf(tmp, format, va_arg(args, double));
            concat(tmp);
            break;
          case 'c' :
	    #if ( __GNUC__ <3 )
            concat(va_arg(args, char));
	    #else
            concat(va_arg(args, int));
	    #endif
            break;
          case 'C' :
	    #if ( __GNUC__ <3 )
            ::sprintf(tmp, format, va_arg(args, wchar_t));
	    #else
            ::sprintf(tmp, format, va_arg(args, int));
	    #endif
	  concat(tmp);
            break;
          case 's' :
          case 'S' :
            as = va_arg(args, char *);
            if(as)
              concat(as);
            else
              concat("(null)");
            break;
          case 'n' :
            *va_arg(args, int *) = strlen(str);
            break;
        }
      }

      /* %% means '%' */
      else {
        concat('%');
        i += 2;
      }
    }
  }
#ifndef WIN32
  // adapt string length
  realloc(0);
#endif

}

TclString TclString::baseName(char separator) const {
	unsigned int sep=0;
	while ((sep<(length-1)) && (str[sep]!=separator))
		sep++;
	sep=length-1-sep;
	TclString out(*this);
	out.truncAtEnd(sep);
	return out;
}
const char *TclString::extension(char separator) const {
  const char *ext=strrchr(str,separator);
  return ext;
}

int TclString::split(TclString**& array, char *separators) {
  unsigned int i = 0, j, k, n = 0;
  do {
    while(i < length - 1 && strchr(separators, str[i])) i++;
    if(i < length - 1) {
      n++;
      while(i < length - 1 && !strchr(separators, str[i])) i++;
    }
  }
  while(i < length - 1);

  array = new TclString*[n];
  for(k = 0, i = 0, j = 0; k < n; k++) {
    while(i < length - 1 && strchr(separators, str[i])) i++;
    j = i;
    while(j < length - 1 && !strchr(separators, str[j])) j++;
    array[k] = new TclString(*this, i, j - 1);
    i = j;
  }
  return n;
}


void TclString::expand(void) {
 TclString tmp;
  unsigned int i = 0;
  while(i < getLength()) {
    unsigned int j = i;
    while(j < getLength() && (*this)[j] != '$') j++;
    if(j > i) tmp += TclString((*this), i, j - 1);
    if(j < getLength()) {
      unsigned int k = j + 1;
      if(k < getLength()) {
        TclString var;
        if((*this)[k] == '{') {
          unsigned int l = ++k;
          while(l < getLength() && (*this)[l] != '}') l++;
          if((*this)[l] != '}') {
            CERR << "warning: parsing .yavrc: missing closing brace in \'"
                 << (*this) << "\'" << ENDL;
            k = l;
          }
          else {
            var = TclString((*this), k, l - 1);
            k = l;
          }
        }
        else {
          unsigned int l = k;
          while((l < getLength()) && ((*this)[l] != '/') &&
                ((*this)[l] != '\\') && ((*this)[l] != ' ') &&
                ((*this)[l] != '.') && ((*this)[l] != '$')) l++;
          var = TclString((*this), k, l - 1);
          k = l - 1;
        }
        if(var != EMPTY_STRING) {
          char *value = getenv((char *) var);
          if(!value) {
            CERR << "warning: parsing .yavrc: varible \'" << var
                 << "\' does not exist" << ENDL;
          }
          else tmp += value;
        }
      }
      else {
        CERR << "warning: parsing .yavrc: no variable after $ name in \'"
             << (*this) << "\'" << ENDL;
      }
      j = k + 1;
    }
    i = j;
  }
  (*this) = tmp;
}


}
