#include <basics/Timer.h>

namespace yav {

Timer::Timer()
{
  elapsedTime=0;
  reset();
  stepTime = getTime();
}

Timer::~Timer()
{
}
void Timer::pause()
{
        elapsedTime=Timer::getTime();
}

void Timer::restart()
{
#ifndef WIN32
  gettimeofday(&initialTime, 0);
#else
  time(&initialTime);
#endif
}


void Timer::reset()
{
  elapsedTime=0;
#ifndef WIN32
  gettimeofday(&initialTime, 0);
#else
  time(&initialTime);
#endif
}

void Timer::reset(double delay){
  reset();
  elapsedTime = delay;
}

double Timer::getTime()
{
#ifndef WIN32
  struct timeval current;
  gettimeofday(&current, 0);
  double t = (current.tv_sec - initialTime.tv_sec)
         + 1e-6 * (current.tv_usec - initialTime.tv_usec);
  return (t+elapsedTime);
#else

  return(1000*(time(0)-initialTime+elapsedTime));
#endif
}

void Timer::step(){
  stepTime = getTime();
}

double Timer::getElapsedTimeFromLastStep(){
  return (getTime() - stepTime);
}

double Timer::getElapsedTimeFromLastStepAndSetStep(){
  double t = getTime() - stepTime;
  stepTime = getTime();
  return (t);
}

}
#ifdef MAIN
#include <iostream.h>
#include <basics/Architecture.h>

using namespace yav;
int main()
{
        Timer t1;

        t1.reset();
        for (int i=0; i<10; i++) {
          CERR<<"Elapsed time is: "<<t1.getTime()<<" seconds\n"<<ENDL;
                t1.reset();
        }
}
#endif







/*]**************************************************************************/
