#ifndef CONSTANTS_H
#define CONSTANTS_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#ifndef WIN32
#pragma interface "basics/Constants.h"
#endif
#include <math.h>
#include <basics/Architecture.h>
#include <basics/Bool.h>

#ifdef WIN32
#include "basics/export.h"
#endif

namespace yav {

/* e */
#ifndef M_E
#  define M_E        2.7182818284590452354E0  /*Hex  2^ 0 * 1.5bf0a8b145769 */
#endif
/* log2(e) */
#ifndef M_LOG2E
#  define M_LOG2E    1.4426950408889634074E0  /*Hex  2^ 0 * 1.71547652B82FE */
#endif
/* log10(e)*/
#ifndef M_LOG10E
#  define M_LOG10E   4.3429448190325182765E-1 /*Hex  2^-2 * 1.BCB7B1526E50E */
#endif
/* ln(2)*/
#ifndef M_LN2
#  define M_LN2      6.9314718055994530942E-1 /*Hex  2^-1 * 1.62E42FEFA39EF */
#endif
/* ln(10)*/
#ifndef M_LN10
#  define M_LN10     2.3025850929940456840E0  /*Hex  2^ 1 * 1.26bb1bbb55516 */
#endif
/* pi*/
#ifndef M_PI
#  define M_PI       3.1415926535897932385E0  /*Hex  2^ 1 * 1.921FB54442D18 */
#endif
/* pi / 2*/
#ifndef M_PI_2
#  define M_PI_2     1.5707963267948966192E0  /*Hex  2^ 0 * 1.921FB54442D18 */
#endif
/* pi / 4*/
#ifndef M_PI_4
#  define M_PI_4     7.8539816339744830962E-1 /*Hex  2^-1 * 1.921FB54442D18 */
#endif
/* 1 / pi*/
#ifndef M_1_PI
#  define M_1_PI     3.1830988618379067154E-1 /*Hex  2^-2 * 1.45f306dc9c883 */
#endif
/* 2 / pi*/
#ifndef M_2_PI
#  define M_2_PI     6.3661977236758134308E-1 /*Hex  2^-1 * 1.45f306dc9c883 */
#endif
/* 2 / ::sqrt(pi)*/
#ifndef M_2_SQRTPI
#  define M_2_SQRTPI 1.1283791670955125739E0  /*Hex  2^ 0 * 1.20dd750429b6d */
#endif
/* ::sqrt(2)*/
#ifndef M_SQRT2
#  define M_SQRT2    1.4142135623730950488E0  /*Hex  2^ 0 * 1.6A09E667F3BCD */
#endif
/* 1 / ::sqrt(2)*/
#ifndef M_SQRT1_2
#  define M_SQRT1_2  7.0710678118654752440E-1 /*Hex  2^-1 * 1.6a09e667f3bcd */
#endif


/* signed 8-bits min value*/
#define MIN_SIGNED_CHAR -128
/* signed 8-bits max value*/
#define MAX_SIGNED_CHAR 127

/* unsigned 8-bits min value*/
#define MIN_UNSIGNED_CHAR 0
/* unsigned 8-bits max value*/
#define MAX_UNSIGNED_CHAR 255

/* signed 16-bits min value*/
#define MIN_SIGNED_SHORT -32768
/* signed 16-bits max value*/
#define MAX_SIGNED_SHORT 32767

/* unsigned 16-bits min value*/
#define MIN_UNSIGNED_SHORT 0
/* unsigned 16-bits max value*/
#define MAX_UNSIGNED_SHORT 65535

/* signed 32-bits max value*/
#define MAX_SIGNED_INT 2147483647
/* signed 32-bits min value*/
#define MIN_SIGNED_INT (- MAX_SIGNED_INT - 1)

/* unsigned 32-bits min value*/
#define MIN_UNSIGNED_INT 0U
/* unsigned 32-bits max value*/
#define MAX_UNSIGNED_INT 4294967295U

#ifdef __GNUC__
/* signed 64-bits max value*/
#define MAX_SIGNED_LONG 9223372036854775807LL
/* signed 64-bits min value*/
#define MIN_SIGNED_LONG (- MAX_SIGNED_LONG - 1)
#else
/* signed 64-bits max value*/
#define MAX_SIGNED_LONG 9223372036854775807L
/* signed 64-bits min value*/
#define MIN_SIGNED_LONG (- MAX_SIGNED_LONG - 1)
#endif

#ifdef __GNUC__
/* unsigned 64-bits min value*/
#define MIN_UNSIGNED_LONG 0
/* unsigned 64-bits max value*/
#define MAX_UNSIGNED_LONG 18446744073709551615ULL
#else
/* signed 64-bits max value*/
#define MIN_UNSIGNED_LONG 0UL
/* signed 64-bits min value*/
#define MAX_UNSIGNED_LONG 18446744073709551615UL
#endif

#ifndef MAXFLOAT
#  define _MAXFLOAT
#  define MAXFLOAT    ((float) 3.40282346638528860e+38)
#endif
/* floatting 32-bits max value*/
#define MAX_FLOAT MAXFLOAT
/* floatting 32-bits min value*/
#define MIN_FLOAT -MAXFLOAT

#ifndef HUGE_VAL
#  if defined(_IEEE_FP)
#    define HUGE_VAL 1.797693134862315708e308
#  else
#    define HUGE_VAL 1.797693134862315708e308
#  endif
#endif
/* floatting 64-bits max value*/
//#define MAX_DOUBLE HUGE_VAL
#define MAX_DOUBLE 1.797693134862315708e308
/*  floatting 64-bits min value*/
//#define MIN_DOUBLE -HUGE_VAL
#define MIN_DOUBLE -1.797693134862315708e308


/** min type value for template type
    @return min value coded for type T */
template <class T> T minTypeValue(T);
/** max type value for template type
    @return max value coded for type T */
template <class T> T maxTypeValue(T);
/** cast a double value into type T
    @return  value as the closest val from the input double value (eventually apply round) */
template <class T> T castFromDouble(double v) {
	if (v>(double) maxTypeValue((T)0)) 
		return maxTypeValue((T)0);
	else if (v<(double) minTypeValue((T)0)) 
		return minTypeValue((T)0);
	else 
		return (T) (v+0.5);
}

#ifdef __GNUC__
/** min argument
    @return min value between a and b */
template <class T> T min(T a, T b) {
  return (a < b) ? a : b;
}
/** max argument
    @return max value between a and b */
template <class T> T max(T a, T b) {
  return (a > b) ? a : b;
}
#endif
#ifdef WIN32
#undef min
#undef max
#endif

template<> char minTypeValue<char>(char);
template<> char maxTypeValue<char>(char);
template<> unsigned char minTypeValue<unsigned char>(unsigned char);
template<> unsigned char maxTypeValue<unsigned char>(unsigned char);
template<> short int minTypeValue<short int>(short int);
template<> short int maxTypeValue<short int>(short int);
template<> unsigned short int minTypeValue<unsigned short int>(unsigned short int);
template<> unsigned short int maxTypeValue<unsigned short int>(unsigned short int);
template<> int minTypeValue<int>(int);
template<> int maxTypeValue<int>(int);
template<> unsigned int minTypeValue<unsigned int>(unsigned int);
template<> unsigned int maxTypeValue<unsigned int>(unsigned int);
template<> float minTypeValue<float>(float);
template<> float maxTypeValue<float>(float);
template<> double minTypeValue<double>(double);
template<> double maxTypeValue<double>(double);
template<> LONGINT minTypeValue<LONGINT>(LONGINT);
template<> LONGINT maxTypeValue<LONGINT>(LONGINT);
template<> unsigned LONGINT minTypeValue<unsigned LONGINT>(unsigned LONGINT);
template<> unsigned LONGINT maxTypeValue<unsigned LONGINT>(unsigned LONGINT);

}

#endif
