namespace yav {

// build a new DPList from a list
template <class T> DPList<T>::DPList(const std::list<T *>& lst,
                                     DPOwner<T> *_owner)
  : owner(_owner) {
  typename std::list<T *>::const_iterator it;
  for(it = lst.begin(); it != lst.end(); it++) append(*it);
}


template <class T> DPList<T>::~DPList() {
    typename DPList<T>::iterator it;
    for(it = this->begin(); it != this->end(); it++)
        (*it)->DPListElement<T>::outOfList(this);
}


template <class T> void DPList<T>::dying(T *dyingElt) {
  // remove element from list
  DPList<T>::remove(dyingElt);
  // warns list owner
  owner->dying(dyingElt);
}


template <class T> void DPList<T>::append(T *t)
  throw (AlreadyInList, NullPointer) {

  if(!t) throw NullPointer("DPList<T>::append");
  typename std::list<DPList<T> *>::iterator it;
  for(it = ((DPListElement<T> *) t)->owners.begin();
      it != ((DPListElement<T> *) t)->owners.end(); it++)
    if(*it == this) throw AlreadyInList("DPList<T>::append(T *)");
  this->push_back(t);
  t->DPListElement<T>::inList(this);
}


template <class T> void DPList<T>::prepend(T *t)
  throw (AlreadyInList, NullPointer) {

  if(!t) throw NullPointer("DPList<T>::prepend");
  typename std::list<DPList<T> *>::iterator it;
  for(it = ((DPListElement<T> *) t)->owners.begin();
      it != ((DPListElement<T> *) t)->owners.end(); it++)
    if(*it == this) throw AlreadyInList("DPList<T>::prepend(T *)");
  this->push_front(t);
  t->DPListElement<T>::inList(this);
}


template <class T> void DPList<T>::remove(T *t)
  throw( NotFound) {

  iterator it;
  bool found = false;
  for(it = this->begin(); it != this->end(); it++) {
    if(*it == t) {
      found = true;
      break;
    }
  }
  if(found) {
    t->DPListElement<T>::outOfList(this);
    erase(it);
  }
  else throw typename DPList<T>::NotFound("DPList<T>::remove(T *t)");
}


// removes the first list element
template <class T> T *DPList<T>::removeFirst(void)
  throw( Empty) {

  if(this->empty()) throw typename DPList<T>::Empty("DPList<T>::removeFirst(void)");
  T *p = this->front();
  this->pop_front();
  p->DPListElement<T>::outOfList(this);
  return p;
}
template <class T> const DPOwner<T> *DPList<T>::getOwner(void) const { 
	return owner; 
}
template <class T>  DPOwner<T> *DPList<T>::getOwner(void)  { 
	return owner; 
}
// removes last list element
template <class T> T *DPList<T>::removeLast(void)
  throw( Empty) {

  if(this->empty()) throw typename DPList<T>::Empty("DPList<T>::removeLast(void)");
  T *p = this->back();
  this->pop_back();
  p->DPListElement<T>::outOfList(this);
  return p;
}


// returns first list element
template <class T> T *DPList<T>::getFirst(void)
  throw( Empty) {

  if(this->empty()) throw typename DPList<T>::Empty("DPList<T>::getFirst(void)");
  return this->front();
}


// returns last list element
template <class T> T *DPList<T>::getLast(void)
  throw( Empty) {

  if(this->empty()) throw typename DPList<T>::Empty("DPList<T>::getLast(void)");
  return this->back();
}


}
