#ifndef DPLISTELEMENT_H
#define DPLISTELEMENT_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "basics/DPListElement.h"
#include <list>
#include <basics/DPList.h>

namespace yav {


template <class T> class DPOwner;
  /*! \class DPListElement DPListElement.h basics/DPListElement.h
    \ingroup libBasics
    element of a DPList
  */
template <class T> class DPListElement {
 private:
  /// a list of DPList that owns this DPListElement object
  std::list<DPList<T> *> owners;

  /** called by DPList<T> when this object is added to list
      @param list list this object is added in */
  void inList(DPList<T> * const lst) { owners.push_front(lst); }
  /** called by DPList<T> when this object is removed from list
      @param list list this object is removed from in
      @exception List<DPList<T> *>::NotFound this object was not found in list */
  void outOfList(DPList<T> * const lst) { owners.remove(lst); }

 public:
  /// destructor: call DPList::dying method for each DPList in owners
  virtual ~DPListElement();

  /// returns an iterator on owners list
  const std::list<DPList<T> *>& getOwners(void) const {
    return owners;
  }

  friend class DPList<T>;
};

} // namespace yav

#include "DPListElement.C"

#endif
