#ifndef FILEINFO_H
#define FILEINFO_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "basics/FileInfo.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <basics/TclString.h>


namespace yav {


  /*! \class FileInfo FileInfo.h basics/FileInfo.h
    \ingroup libBasics
    Gives info on file named
  */
class FileInfo {
private:
  /// file name
  TclString name;

public:
  /** @name Exception class */
  //@{
  /*! \class InvalidFileName FileInfo.h basics/FileInfo.h
    \ingroup libBasics
    try to build a FileInfo object with a non file name
  */
  class InvalidFileName : public Exception {
  public:
    /** constructor
        @param from method throwing exception name
        @param name invalid file name string */
    InvalidFileName(const std::string& from, const std::string& name)
                : Exception(from,std::string(std::string("not a valid file name: ") +
                name)) {}
  };
  //@}

  /** @name Constructors */
  //@{
  /** construction from a TclString
      @param _name file name (file is not supposed to exist)
      @exception InvalidFileName file name cannot end with '/' */
  FileInfo(const TclString& _name) throw(InvalidFileName) : name(_name) {
    if(name[name.getLength() - 1] == '/')
      throw new InvalidFileName("FileInfo(const TclString&)", name.tostring());
  }

  /** construction from a C string
      @param _name file name (file is not supposed to exist)
      @exception InvalidFileName file name cannot end with '/' */
  FileInfo(const char *_name) throw(InvalidFileName) : name(_name) {
    if(name[name.getLength() - 1] == '/')
      throw new InvalidFileName("FileInfo(const char *)", name.tostring());
  }
  //@}

  /// returns object's file name
  const TclString& getName(void) const { return name; }

  /// returns directory part from file name
  TclString directory(void) const;
  /// returns non directory part from file name
  TclString file(void) const;
  /** returns directory part after removing given extension
      @param ext extension to remove */
  TclString noextension(const char *ext) const;

  /// indicates whether file exists on disk
  bool exists(void) const;
  /// returns file size in bytes
  off_t size(void) const;
};


} // namespace yav

#endif
