#ifndef __IniFile_h
#define __IniFile_h

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#ifndef WIN32
#pragma interface "basics/IniFile.h"
#endif

#include <stdio.h>
#include <iostream>
#include <string>
#include <cstring>
#include <map>

#include <basics/Architecture.h>
#include <basics/Bool.h>

#ifdef WIN32
#include "basics/export.h"
#endif

#if ((!defined(WIN32))&&(!defined(_BCPLUSPLUS_)))
int GetPrivateProfileString (const char *AppName, const char * KeyName, const char * Default, char * Buffer, int MaxLength, const char * FileName);
#else
#include <windows.h>
#endif


namespace yav {

bool GetInitFileString (const char * FileName, const char * EntryName, const char * KeyName, char * Buffer, const char * Default, bool warning=false);
bool GetInitFileStringND (const char * FileName, const char * EntryName, const char * KeyName, char * Buffer, bool warning=true);

bool GetInitFileInt (const char * FileName, const char * EntryName, const char * KeyName, int & Buffer, int Default, bool warning=false);
bool GetInitFileIntND (const char * FileName, const char * EntryName, const char * KeyName, int & Buffer, bool warning=true);

bool GetInitFileFloat (const char * FileName, const char * EntryName, const char * KeyName, float & Buffer, float Default, bool warning=false);
bool GetInitFileFloatND (const char * FileName, const char * EntryName, const char * KeyName, float & Buffer, bool warning=true);

bool GetInitFileDouble (const char * FileName, const char * EntryName, const char * KeyName, double & Buffer, double Default, bool warning=false);
bool GetInitFileDoubleND (const char * FileName, const char * EntryName, const char * KeyName, double & Buffer, bool warning=true);

bool GetInitFileBool (const char * FileName, const char * EntryName, const char * KeyName, bool & Buffer, bool Default, bool warning=false);
bool GetInitFileBoolND (const char * FileName, const char * EntryName, const char * KeyName, bool & Buffer, bool warning=true);

}

/*! \class IniFile IniFile.h basics/IniFile.h
  \ingroup libBasics
  more detailled description
*/
class IniFile {
protected:
  typedef std::map<std::string,std::string> KeyMap;
  typedef std::map<std::string,KeyMap> EntryMap;
  EntryMap entryMap;

public:
  /// Constructor : Read IniFile
  IniFile(const char* filename);
  /// Destructor
  ~IniFile();

public:
  /// Read IniFile
  void Read(const char* filename);
  /// Clear content
  void Clear();

public:
  /// Get a string - return : 0 if not found, 1 if found
  int GetString(const char* EntryName, const char* KeyName, char* Buffer, const char* Default="") const;

  /// Get a int - return : 0 if not found, 1 if found
  template <class T>
  int GetInt(const char* EntryName, const char* KeyName, T& Buffer) const
    {return GetInt(EntryName, KeyName, Buffer, 0);}
  /// Get a int - return : 0 if not found, 1 if found
  template <class T,class T2>
  int GetInt(const char* EntryName, const char* KeyName, T& Buffer, T2 Default) const
    {
      EntryMap::const_iterator ie;
      KeyMap::const_iterator is;
      T tmp;
      if (((ie=entryMap.find(EntryName))!=entryMap.end())
          &&((is=ie->second.find(KeyName))!=ie->second.end())
          &&(sscanf(is->second.c_str(),"%d",&tmp)==1)) {
        Buffer=tmp;
        return 1;
      }
      Buffer=Default;
      return 0;
    }

  /// Get a float - return : 0 if not found, 1 if found
  template <class T>
  int GetFloat(const char* EntryName, const char* KeyName, T& Buffer) const
    {return GetFloat(EntryName, KeyName, Buffer,0);}
  /// Get a float - return : 0 if not found, 1 if found
  template <class T, class T2>
  int GetFloat(const char* EntryName, const char* KeyName, T& Buffer, T2 Default) const
    {
      EntryMap::const_iterator ie;
      KeyMap::const_iterator is;
      T tmp;
      if (((ie=entryMap.find(EntryName))!=entryMap.end())
          &&((is=ie->second.find(KeyName))!=ie->second.end())
          &&(sscanf(is->second.c_str(),"%f",&tmp)==1)) {
        Buffer=tmp;
        return 1;
      }
      Buffer=Default;
      return 0;
    }

  /// Get a double - return : 0 if not found, 1 if found
  template <class T>
  int GetDouble(const char* EntryName, const char* KeyName, T& Buffer) const
    {return GetDouble(EntryName, KeyName, Buffer, 0);}
  /// Get a double - return : 0 if not found, 1 if found
  template <class T,class T2>
  int GetDouble(const char* EntryName, const char* KeyName, T& Buffer, T2 Default) const
    {
      EntryMap::const_iterator ie;
      KeyMap::const_iterator is;
      T tmp;
      if (((ie=entryMap.find(EntryName))!=entryMap.end())
          &&((is=ie->second.find(KeyName))!=ie->second.end())
          &&(sscanf(is->second.c_str(),"%lf",&tmp)==1)) {
        Buffer=tmp;
        return 1;
      }
      Buffer=Default;
      return 0;
    }

  /// Get a boolean - return : 0 if not found, 1 if found
  template <class T>
  int GetBool(const char* EntryName, const char* KeyName, T& Buffer) const
    {return GetBool(EntryName, KeyName, Buffer, 0);}
  /// Get a boolean - return : 0 if not found, 1 if found
  template <class T,class T2>
  int GetBool(const char* EntryName, const char* KeyName, T& Buffer, T2 Default) const
    {
      EntryMap::const_iterator ie;
      KeyMap::const_iterator is;
      int tmp;
      if (((ie=entryMap.find(EntryName))!=entryMap.end())
          &&((is=ie->second.find(KeyName))!=ie->second.end())) {
        if (sscanf(is->second.c_str(),"%d",&tmp)==1){
          Buffer=(tmp!=0);
          return 1;
        }
        if (strcmp(is->second.c_str(),"false")==0){
          Buffer=0;
          return 1;
        }
        if (strcmp(is->second.c_str(),"true")==0){
          Buffer=1;
          return 1;
        }
      }
      Buffer=Default;
      return 0;
    }

};

#endif // __IniFile_h
