#ifndef LISTELEMENT_H
#define LISTELEMENT_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "basics/ListElement.h"


namespace yav {


template <class T> class ObjectsList;

/*! \class ListElement ListElement.h basics/ListElement.h
  \ingroup libBasics
  An object that is linkable in a list 
*/
template <class T> class ListElement {

 private:
  /// pointer to previous element in list
  ListElement<T> *prev;
  /// pointer to following element in list
  ListElement<T> *next;
  /// pointer to list object
  ObjectsList<T> *list;

 public:
  /** @name Constructors */
  //@{
  /// default constructor: object is not linked in any list
  ListElement() : prev(0), next(0), list(0) {}
  /** copy constructor: object is bound to same neighbourg has original
      @param  le  object to copy */
  ListElement(const ListElement<T>& le) { prev = le.prev; next = le.next; }
  //@}

  /** @name Destructor */
  //@{
  /// remove object from list
  virtual ~ListElement() { if(list) remove(); }
  //@}

  /// return pointer to following object
  T* following(void) const { return (T*) next; }
  /// return pointer to previous object
  T* previous(void) const { return (T*) prev; }

  /// return pointer to list containing that object
  ObjectsList<T> *getList(void) const { return list; }
  /// return pointer to following object
  ListElement<T> *getNext(void) const { return next; }
  /// return pointer to previous object
  ListElement<T> *getPrevious(void) const { return prev; }
  /// insert a list Element after a given element
  void insert(ListElement<T> *le);
  /// disconnect a list Element
  void disconnect(void);
  /// remove object from list
  void remove(void);
  /** returns the rank of the element in the list
      @return the rank of the element in the list : if the element does not
      belong to any list then return -1 */
  int getRank() const;


  friend class ObjectsList<T>;
};

} // namespace yav

#include <basics/ObjectsList.h>

#include "ListElement.C"

#endif
