#ifndef MATH_H
#define MATH_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#ifndef WIN32
#pragma interface "basics/Math.h"
#endif

#include <math.h>
#include <stdlib.h>     // for rand() and srand() declaration
#include <time.h>       // for time() declaration
#include <basics/Exception.h>
#include <basics/Bool.h>
#include <basics/Constants.h>
#include <basics/Architecture.h>
#ifdef WIN32
#include <float.h>
#include "basics/export.h"
#endif

#ifdef WIN32
#ifndef __MINGW32__
/**
       rint is not defined on VC++.
       We declare it outside of namespace yav to make it accessible
       exactly like on Unix.
     */
int rint(double a);
#define isnan(X) _isnan(X)
#endif
#endif

namespace yav {

/** allocate with malloc a 2D array of type T
    @param nr number of rows
    @param nc number of columns
    @return 2D array of type T */
template <class T> T **create2DArray(unsigned int nr,unsigned int nc)
{
  T **fbuf, *p;
  unsigned int i,size;

  size=nr*sizeof(T *);
  // align bytes
  if (size%8)
    size+= 8-(size%8);

  fbuf = (T **) new char [size+nr*nc*sizeof(T)];
  p=(T* )fbuf+size;
  for (i = 0; i <nr; ++i)
    {
      fbuf[i] = p;
      p += nc;
    }
  return fbuf;
}


///  returns -1 +1 or 0 depending on the sign of x
double signd(double x);
int round(double x);
int signi(double a);
int signedFloor(double a);
double drand(void);

}

#endif

