namespace yav {

template <class T> ObjectsList<T>::ObjectsList(const ObjectsList<T>& list) {
  for(ListElement<T>* obj = list.getFirst(); obj; obj = obj->following())
    *this += new ListElement<T>(*obj);
}


template <class T> void ObjectsList<T>::empty(void) {
  typename std::list<EmptyObjectsList<T> *>::const_iterator it;
  for(it = warn.begin(); it != warn.end(); it++)
    (*it)->empty();
}


template <class T> void ObjectsList<T>::operator=(const ObjectsList<T>& list) {
  while(head) delete head;
  for(ListElement<T>* obj = list.getFirst(); obj; obj = obj->following())
    *this += new ListElement<T>(*obj);
}


template <class T> ObjectsList<T> ObjectsList<T>::operator+(const ObjectsList<T>& list) {
  ObjectsList<T> nl;
  ListElement<T>* obj;

  for(obj = head; obj; obj = obj->following())
    nl += new ListElement<T>(*obj);
  for(obj = list.getFirst(); obj; obj = obj->following())
    nl += new ListElement<T>(*obj);

  return nl;
}


template <class T> void ObjectsList<T>::operator+=(const ObjectsList<T>& list) {
  for(ListElement<T>* obj = list.getFirst(); obj; obj = obj->following())
    *this += new ListElement<T>(*obj);
}


template <class T> void ObjectsList<T>::append(ListElement<T> *obj)
  throw (NullPointer,AlreadyInList){
  if (obj==0)
    throw NullPointer("ObjectsList<T>::append(ListElement<T> *obj)");
  if (obj->list!=0)
    throw AlreadyInList("ObjectsList<T>::append(ListElement<T> *obj)");
  if(!head) head = obj;
  obj->prev = tail;
  if(tail) tail->next = obj;
  obj->next = 0;
  tail = obj;
  count++;
  obj->list=this;
}

template <class T> void ObjectsList<T>::prepend(ListElement<T> *obj)
  throw (NullPointer,AlreadyInList){
   if (obj==0)
     throw NullPointer("ObjectsList<T>::prepend(ListElement<T> *obj)");
  if (obj->list!=0)
    throw AlreadyInList("ObjectsList<T>::prepend(ListElement<T> *obj)");
  if(!tail) tail = obj;
  obj->next = head;
  if(head) head->prev = obj;
  obj->prev = 0;
  head = obj;
  count++;
  obj->list=this;
}



template <class T> void ObjectsList<T>::remove(ListElement<T> *obj)
  throw(NotFound) {

  if(obj->list != this) throw NotFound("ObjectsList<T>::remove(ListElement<T> *)");
  obj->remove();
}


template <class T> T* ObjectsList<T>::removeFirst(void)
  throw(Empty) {

  if(!head) throw Empty("ObjectsList<T>::removeFirst(void)");
  ListElement<T> *t = head;
  head = t->next;
  head->prev = 0;
  t->next = t->prev = 0;
  count--;
  if(!count) empty();
  return (T *) t;
}


template <class T> T* ObjectsList<T>::removeLast(void)
  throw(Empty) {

  if(!tail) throw Empty("ObjectsList<T>::removeLast(void)");
  ListElement<T> *t = tail;
  tail = t->prev;
  tail->next = 0;
  t->next = t->prev = 0;
  count--;
  if(!count) empty();
  return (T *) t;
}


template <class T> T* ObjectsList<T>::getFirst(void) const
  {
    return (T *) head;
}


template <class T> T* ObjectsList<T>::getLast(void) const
  throw(Empty) {

  if(!tail) throw Empty("ObjectsList<T>::getLast(void)");
  return (T *) tail;
}


template <class T> T* ObjectsList<T>::operator[](unsigned int n) const
  throw(NotFound) {

  ListElement<T> *obj;
  unsigned int i;

  if(n >= count) throw NotFound("ObjectsList<T>::operator[](unsigned int)");
  for(i = 0, obj = head; i < n; i++, obj = obj->next) ;
  return (T *) obj;
}

template <class T> ObjectsListSTLiterator<T>::ObjectsListSTLiterator(ListElement<T> *e)
{
  current=e;
  if (e)
    list=e->getList();
  else
    list=0;
}

template <class T> ObjectsListSTLiterator<T>::ObjectsListSTLiterator(const ObjectsListSTLiterator<T> &t)
{
  current=t.current;
  list=t.list;
}
template <class T> ObjectsListSTLiterator<T> ObjectsListSTLiterator<T>::operator++(int)  throw(AccessBeyondList) {
  if(!current) throw AccessBeyondList("ObjectsListSTLIterator<T>::operator++(int)");
  typename ObjectsList<T>::iterator it= (*this);
  current = current->getNext();
  return it;
}
template <class T> ObjectsListSTLiterator<T> &ObjectsListSTLiterator<T>::operator++(void) throw(AccessBeyondList) {
  if(!current) throw AccessBeyondList("ObjectsListSTLIterator<T>::operator++(void)");
  current = current->getNext();
  return *this;
}
template <class T> ObjectsListSTLiterator<T> ObjectsListSTLiterator<T>::operator--(int)  throw(AccessBeyondList) {
    if(!current) throw AccessBeyondList("ObjectsListSTLIterator<T>::operator--(int)");
  ObjectsListSTLiterator<T> it= (*this);
  current = current->getPrevious();
  return it;
}
template <class T> ObjectsListSTLiterator<T> &ObjectsListSTLiterator<T>::operator--(void)  throw(AccessBeyondList) {
  if(!current) throw AccessBeyondList("ObjectsListSTLIterator<T>::operator--(void)");
  current = current->getPrevious();
  return *this;
}

template <class T> T *ObjectsListSTLiterator<T>::operator*(void) const{
  return((T *) current);
}
template <class T> bool ObjectsListSTLiterator<T>::operator==(const ObjectsListSTLiterator<T> &t) const{
  return(current==t.current);
}
template <class T> bool ObjectsListSTLiterator<T>::operator!=(const ObjectsListSTLiterator<T> &t) const{
  return(current!=t.current);
}

template <class T> ObjectsListIterator<T>::ObjectsListIterator(const ObjectsListIterator<T> &it)
{
  current=it.current;
  list=it.list;
}




template <class T> T *ObjectsListIterator<T>::operator++(int)
  throw(AccessBeyondList) {

  if(!current) throw AccessBeyondList("ObjectsListIterator<T>::operator++(int)");
  T *t = (T *)current;
  current = current->getNext();
  return t;
}


template <class T> T *ObjectsListIterator<T>::operator--(int)
  throw(AccessBeyondList) {

  if(!current) throw AccessBeyondList("ObjectsListIterator<T>::operator--(int)");
  T *t = (T *)current;
  current = current->getPrevious();
  return t;
}


template <class T> T *ObjectsListIterator<T>::operator++(void)
  throw(AccessBeyondList) {

  if(!current || !current->getNext())
    throw AccessBeyondList("ObjectsListIterator<T>::operator++(void)");
  current = current->getNext();
  return ((T*) current);
}


template <class T> T *ObjectsListIterator<T>::operator--(void)
  throw(AccessBeyondList) {

  if(!current || !current->getPrevious())
    throw AccessBeyondList("List<T>::Iterator::operator--(void)");
  current = current->getPrevious();
  return ((T *)current);
}


template <class T> T *ObjectsListIterator<T>::operator()(void) const
  throw(AccessBeyondList) {

  if(!current) throw AccessBeyondList("ObjectsListIterator<T>::operator()(void)");
  return ((T*) current);
}


template <class T> T *ObjectsListIterator<T>::operator[](unsigned int i)
  throw(AccessBeyondList) {

  unsigned int n;

  if(i >= list->count)
    throw AccessBeyondList("ObjectsListIterator<T>::operator[](unsigned int)");
  for(n = 0, current = list->head; n < i; n++, current = current->following()) ;
  return ((T *) current);
}

template <class T> ObjectsListSTLiterator<T>  ObjectsList<T>::begin() const {
  return(ObjectsListSTLiterator < T > (getHead()));
}
template <class T> ObjectsListSTLiterator<T>  ObjectsList<T>::end() const {
  return(ObjectsListSTLiterator<T>(0));
}

template <class T,class V>
CastIterator<T,V>::CastIterator(const ObjectsList<T> &l)
  : list(&l), current(l.getFirst()) {}
template <class T,class V>
CastIterator<T,V>::CastIterator(const CastIterator &t)
  : list(t.list), current(t.current) {}

template <class T,class V> CastIterator<T,V>::AccessBeyondList::AccessBeyondList(const std::string& from) :
  Exception(from,
            "Attempt to access an element beyond list bounds") {
}


template <class T,class V> ListElement<T> * CastIterator<T,V>::getNext(ListElement<T> *elt)throw(AccessBeyondList)
{
  if(!elt) throw AccessBeyondList("CastIterator<T,V>::getNext(ListElement<T> *)");
  return( elt->ListElement<T>::getNext());
}

template <class T,class V> ListElement<T> * CastIterator<T,V>::getPrevious(ListElement<T> *elt)throw(AccessBeyondList)
{
   if(!elt) throw AccessBeyondList("CastIterator<T,V>::getPrevious(ListElement<T> *)");
   return(elt->ListElement<T>::getPrevious());
}


template <class T,class V> V *CastIterator<T,V>::operator++(int)
  throw(AccessBeyondList) {

  if(!current) throw AccessBeyondList("ObjectsList<T,V>::Iterator::operator++(int)");
  V *t = (V *)current;
  current = getNext(current);
  return t;
}


template <class T,class V> V *CastIterator<T,V>::operator--(int)
  throw(AccessBeyondList) {

  if(!current) throw AccessBeyondList("CastIterator<T,V>::operator--(int)");
  V *t = (V *)current;
  current = getPrevious(current);
  return t;
}


template <class T,class V> V *CastIterator<T,V>::operator++(void)
  throw(AccessBeyondList) {

  if(!current || !getNext(current))
    throw AccessBeyondList("CastIterator<T,V>::operator++(void)");
  current = getNext(current);
  return ((V*) current);
}


template <class T,class V> V *CastIterator<T,V>::operator--(void)
  throw(AccessBeyondList) {

  if(!current || !getPrevious(current))
    throw AccessBeyondList("List<T>::Iterator::operator--(void)");
  current = getPrevious(current);
  return ((V *)current);
}


template <class T,class V> V *CastIterator<T,V>::operator()(void) const
  throw(AccessBeyondList) {

  if(!current) throw AccessBeyondList("CastIterator<T,V>::operator()(void)");
  return ((V*) current);
}


template <class T,class V> V *CastIterator<T,V>::operator[](unsigned int i)
  throw(AccessBeyondList) {

  unsigned int n;

  if(i >= list->count)
    throw AccessBeyondList("CastIterator<T,V>::operator[](unsigned int)");
  for(n = 0, current = list->head; n < i; n++, current = current->following) ;
  return ((V *) current);
}

}
