#ifndef OWNER_H
#define OWNER_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 4355)
#endif

#pragma interface "basics/Owner.h"


namespace yav {


template <class T> class ObjectPointer;


/*! \class Owner Owner.h basics/Owner.h
  \ingroup libBasics
  object that owns a smart pointer
*/
template <class T> class Owner {
 public:
  /** @name Constructor */
  //@{
  /** constructs object and initialize pointer value
      @param _pointer initial pointer value */
  Owner(T *_pointer=0) : pointer(this, _pointer) {}
  //@}

  /** @name Destructor */
  //@{
  /** reset pointer to 0 (thus calling loosing) and destroy object */
  virtual ~Owner() { 
  }
  //@}

 protected:
  /// pointer owned
  ObjectPointer<T> pointer;

  /// returns the pointer value
  T *getPointer() const {
    return(pointer()); }
  /// cannot copy an owner
  Owner(const Owner&) : pointer(this) {}

  /** an element owned by this object is being re-assigned
      @param obj element before reassignement */
  virtual void loosing(const ObjectPointer<T>& obj) {}

  /** an element owned by this object is being destroyed
      @param obj element before destruction */
  virtual void dying(const ObjectPointer<T>& obj) = 0;

  friend class ObjectPointer<T>;
};


} // namespace yav


#endif
