#ifndef POINTED_H
#define POINTED_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 4355)
#endif

#pragma interface "basics/Pointed.h"
#include <list>


namespace yav {


template <class T> class ObjectPointer;

/*! \class Pointed Pointed.h basics/Pointed.h
  \ingroup libBasics
  an object being referenced by smart pointers that have to know when
  this object is destroyed 
*/
template <class T> class Pointed {
 protected:
  /// list of pointers referecing this object
  std::list<ObjectPointer<T> *> pointers;
  /** adds a new pointer referecing this object
      @param pointer pointer on this */
  void pointedFrom(ObjectPointer<T>* pointer) { 
	  pointers.push_front(pointer); 
  }
  /** given pointer is not referencing this any more
      @param pointer pointer releasing this */
  void pointedAway(ObjectPointer<T>* pointer) { 
	  pointers.remove(pointer); 
  }

 public:
  /// empty default constructor
  Pointed() {}
  /// default destructors: warns pointers of object deletion
  virtual ~Pointed();
  /// returns the first list item of pointers referecing this object
  typename std::list<ObjectPointer<T> *>::iterator getFirstPointer(void);
  /// returns the last list item of pointers referecing this object
  typename std::list<ObjectPointer<T> *>::iterator getLastPointer(void);

  friend class ObjectPointer<T>;
};

/**
 * comparison operator for slist operations
 */
template <class T> int operator==(const Pointed<T>& p1, const Pointed<T>& p2) {
  if(&p1 == &p2) return 1;
  return 0;
}
/**
 * comparison operator for slist operations
 */
template <class T> int operator<(const Pointed<T>& p1, const Pointed<T>& p2) {
  if(&p1 < &p2) return 1;
  return 0;
}


} // namespace yav

#include <basics/Pointer.h>

#include "Pointed.C"

#endif
