#ifndef TIMER_H
#define TIMER_H

#define _BSD_COMPAT
#ifndef WIN32
#include <sys/time.h>
#else
#include <time.h>
#endif


namespace yav {

  /*! \class Timer Timer.h basics/Timer.h
    \ingroup libBasics
   */
  class Timer
{
public:
  /** @name constructor */
  //@{
  /// create a new Timer
  Timer();
   //@}

  /** @name destructor */
  //@{
  /// destructor
  virtual ~Timer();
  //@}

  /// reset the elapsed time
  void reset();
  /// reset the elapsed time to - delay
  void reset(double delay);

  /// pause the timer
  void pause();

  /// restart the timer
  void restart();
  
  /// set the time step
  void step();
  
  /// get elapsed time from last time step
  double getElapsedTimeFromLastStep();
    
  /// get elapsed time from last time step and set a new one
  double getElapsedTimeFromLastStepAndSetStep();

  /** return the elapsed time since last reset()
      time is given as a double number of seconds */
  double getTime();

 private:
#ifndef WIN32
         /// the initial time spent
  struct timeval initialTime;
#else
  time_t initialTime;
#endif
  /// elapsed Time when paused
  double elapsedTime;
  /// step Time
  double stepTime;
};


}


#endif
/*]**************************************************************************/
