#include <classification/BestNumberClustersEM.h>
#include <classification/ExpectationMaximization.h>
#include <basics/Math.h>
#include <iostream>

namespace yav{

BestNumberClustersEM::BestNumberClustersEM(double** data, unsigned int minModes, unsigned int maxModes, unsigned int trials, unsigned int nData, unsigned int nDim, unsigned int* dataRange, unsigned int minDim, unsigned int maxIte, double minImpro)
{
	this->data = data;
	this->minModes = minModes;
	this->maxModes = maxModes;
	this->trials = trials;
	this->nData = nData;
	this->nDim = nDim;
	this->dataRange = dataRange;
	this->minDim = minDim;
	this->maxIte = maxIte;
	this->minImpro = minImpro;
}

ExpectationMaximization* BestNumberClustersEM::findEM()
{
	unsigned int i, j, range, ite, bestVaic, bestVaicc, bestVbic, bestVos, bestVpi, bestVfvq, bestVpc, bestVpe;
	double vaic, vaicc, vbic, vpi, vfvq, vpc, vpe, vos, vaic2, vaicc2, vbic2, vpi2, vfvq2, vpc2, vpe2, vos2, ove, sep;

	ExpectationMaximization* myEM1 = NULL;
	ExpectationMaximization* myEM2 = NULL;

	range = (maxModes-minModes)+1;
	ite = 0;

	for (i=minModes; i<=maxModes; i++)
	{
		std::cout << "===== NUMBER OF MODES: " << i << " =====" << std::endl;

		for (j=0; j<trials; j++)
		{
			if (ite==0)
			{
				myEM1 = new ExpectationMaximization(data, i, nData, nDim);
				myEM1->em();
				vaic = myEM1->getVaic();
				vaicc= myEM1->getVaicc();
				vbic = myEM1->getVbic();
				ove = myEM1->getVo();
				sep = myEM1->getVs();
				sep = 1.0-sep;
				vos = ove/sep;				
				vpi = myEM1->getVpi();
				vfvq= myEM1->getVfvq();
				vpc = myEM1->getVpc();
				vpe = myEM1->getVpe();

				bestVaic = 0;
				bestVaicc= 0;
				bestVbic= 0;
				bestVos  = 0;
				bestVpi  = 0;
				bestVfvq = 0;
				bestVpc  = 0;
				bestVpe  = 0;
		
				std::cout << "[" << ite << "]\tAIC:  " << vaic << "\tAICc: " << vaicc << "\tBIC:  " << vbic << "\tVOS:  " << vos << "\tVPI:  " << vpi << "\tFVQ:  " << vfvq << "\tVPC:  " << vpc << "\tVPE:  " << vpe << std::endl;
			}
			else
			{
				myEM2 = new ExpectationMaximization(data, i, nData, nDim);
				myEM2->em();
				vaic2 = myEM2->getVaic();
				vaicc2= myEM2->getVaicc();
				vbic2 = myEM2->getVbic();

				ove  = myEM1->getVo();
				sep  = myEM1->getVs();
				sep  = 1.0-sep;
				vos2 = ove/sep;				
				vpi2 = myEM2->getVpi();
				vfvq2= myEM2->getVfvq();
				vpc2 = myEM2->getVpc();
				vpe2 = myEM2->getVpe();

				std::cout << "[" << ite << "]\tAIC:  " << vaic2 << "\tAICc: " << vaicc2 << "\tBIC:  " << vbic2 << "\tVOS:  " << vos2 << "\tVPI:  " << vpi2 << "\tFVQ:  " << vfvq2 << "\tVPC:  " << vpc2 << "\tVPE:  " << vpe2 << std::endl;

				// Find best Vpe
				if (vpe2<vpe)
				{
					vpe = vpe2;
					bestVpe = ite;
				}

				// Find best Vpc
				if (vpc2>vpc)
				{
					vpc = vpc2;
					bestVpc = ite;
				}

				// Find best Vfvq
				if (vfvq2<vfvq)
				{
					vfvq = vfvq2;
					bestVfvq = ite;
				}

				// Find best Vpi
				if (vpi2<vpi)
				{
					vpi = vpi2;
					bestVpi = ite;
				}

				// Find best Vos
				if (vos2<vos)
				{
					vos = vos2;
					bestVos = ite;
				}

				// Find best Vbic
				if (vbic2<vbic)
				{
					vbic = vbic2;
					bestVbic = ite;
				}

				// Find best Vaicc
				if (vaicc2<vaicc)
				{
					vaicc = vaicc2;
					bestVaicc = ite;
				}

				// Find best Vaic
				if (vaic2<vaic)
				{
					delete myEM1;
					vaic = vaic2;
					myEM1 = myEM2;
					bestVaic = ite;
				}
				else
					delete myEM2;
			}

			ite++;
		}
	}

	std::cout << std::endl << " ===> AIC :     " << round(bestVaic/trials)+2 << " (" << bestVaic << ")";
	std::cout << std::endl << " ===> AICc:     " << round(bestVaicc/trials)+2 << " (" << bestVaicc << ")";
	std::cout << std::endl << " ===> BIC :     " << round(bestVbic/trials)+2 << " (" << bestVbic << ")";
	std::cout << std::endl << " ===> VOS :     " << round(bestVos/trials)+2 << " (" << bestVos << ")";
	std::cout << std::endl << " ===> VPI :     " << round(bestVpi/trials)+2 << " (" << bestVpi << ")";
	std::cout << std::endl << " ===> FVQ :     " << round(bestVfvq/trials)+2 << " (" << bestVfvq << ")";
	std::cout << std::endl << " ===> VPC :     " << round(bestVpc/trials)+2 << " (" << bestVpc << ")";
	std::cout << std::endl << " ===> VPE :     " << round(bestVpe/trials)+2 << " (" << bestVpe << ")";
	std::cout << std::endl << std::endl;

	return myEM1;
}
}
