#include <classification/BestNumberClustersFCM.h>
#include <classification/FuzzyCMeans.h>
#include <basics/Math.h>
#include <iostream>

namespace yav{

BestNumberClustersFCM::BestNumberClustersFCM(double** data, unsigned int minClust, unsigned int maxClust, unsigned int trials, unsigned int nData, unsigned int nDim, double expo, unsigned int* dataRange, unsigned int minDim, unsigned int maxIte, double minImpro, double* fuzzyPartition, double* initCluster)
{
	this->data = data;
	this->minClust = minClust;
	this->maxClust = maxClust;
	this->trials = trials;
	this->nData = nData;
	this->nDim = nDim;
	this->expo = expo;
	this->dataRange = dataRange;
	this->minDim = minDim;
	this->maxIte = maxIte;
	this->minImpro = minImpro;
	this->fuzzyPartition = fuzzyPartition;
	this->initCluster = initCluster;
}

FuzzyCMeans* BestNumberClustersFCM::findFCM()
{
	unsigned int i, j, range, ite;
	unsigned int bestVpi, bestVpc, bestVpe, bestVfvq, bestVos;
	double vpi, vpi2, vpc, vos, vpe, vfvq, vpc2, vos2, vpe2, vfvq2, sep, ove;

	FuzzyCMeans* myFCM1 = NULL;
	FuzzyCMeans* myFCM2 = NULL;

	range = (maxClust-minClust)+1;
	ite = 0;

	// Time stuff for pseudo-random numbers generation used by U as a first cluster assignation
	srand((unsigned int) time(NULL));

	for (i=minClust; i<=maxClust; i++)
	{
		std::cout << "NBRCLUSTERS: " << i << std::endl;

		for (j=0; j<trials; j++)
		{
			if (ite==0)
			{
				myFCM1 = new FuzzyCMeans(data, i, nData, nDim, dataRange, minDim, false, expo);
				myFCM1->fcm();
				vpi = myFCM1->getVpi();
				ove = myFCM1->getVo();
				sep = myFCM1->getVs();
				sep = 1.0-sep;
				vos = ove/sep;
				vpc = myFCM1->getVpc();
				vpe = myFCM1->getVpe();
				vfvq= myFCM1->getVfvq();
				bestVos = 0;
				bestVpc = 0;
				bestVpi  = 0;
				bestVpe = 0;
				bestVfvq = 0;
		
				std::cout <<  ite << " => Vpc:  " << vpc << ", Vpe: " << vpe << ", Vpi: " << vpi << ", Vos: " << vos << ", Vfvq: " << vfvq << std::endl;
			}
			else
			{
				myFCM2 = new FuzzyCMeans(data, i, nData, nDim, dataRange, minDim, false, expo);
				myFCM2->fcm();
				vpi2 = myFCM2->getVpi();
				ove = myFCM2->getVo();
				sep = myFCM2->getVs();
				sep = 1.0-sep;
				vos2 = ove/sep;
				vpc2 = myFCM2->getVpc();
				vpe2 = myFCM2->getVpe();
				vfvq2= myFCM2->getVfvq();

				std::cout << ite << " => Vpc:  " << vpc2 << ", Vpe: " << vpe2 << ", Vpi: " << vpi2 << ", Vos:  " << vos2 << ", Vfvq:  " << vfvq2 << std::endl;

				// Find best VOS
				if (vos2<vos)
				{
					vos = vos2;
					bestVos = ite;
				}
				
				// Find best VPC
				if (vpc2>vpc)
				{
					vpc = vpc2;
					bestVpc = ite;
				}

				// Find best VPE
				if (vpe2<vpe)
				{
					vpe = vpe2;
					bestVpe = ite;
				}

				// Find best VFVQ
				if (vfvq2<vfvq)
				{
					vfvq = vfvq2;
					bestVfvq = ite;
				}

				// Find best VPI
				if (vpi2<vpi)
				{
					delete myFCM1;
					vpi = vpi2;
					myFCM1 = myFCM2;
					bestVpi = ite;
				}
				else
					delete myFCM2;
			}

			ite++;
		}
	}

	std::cout << std::endl << " ===> bestNbrClusters by Vpc :       " << round(bestVpc/trials)+2 << " (" << bestVpc << ")";
	std::cout << std::endl << " ===> bestNbrClusters by Vpe :       " << round(bestVpe/trials)+2 << " (" << bestVpe << ")";
	std::cout << std::endl << " ===> bestNbrClusters by Vpi :       " << round(bestVpi/trials)+2 << " (" << bestVpi << ")";
	std::cout << std::endl << " ===> bestNbrClusters by Vos :       " << round(bestVos/trials)+2 << " (" << bestVos << ")";
	std::cout << std::endl << " ===> bestNbrClusters by Vfvq:       " << round(bestVfvq/trials)+2 << " (" << bestVfvq << ")" << std::endl << std::endl;

	return myFCM1;
}
}
