#include <classification/BestNumberClustersOrientedPoints2EM.h>
#include <classification/OrientedPoints2EM.h>
#include <basics/Math.h>
#include <iostream>

namespace yav{

BestNumberClustersOrientedPoints2EM::BestNumberClustersOrientedPoints2EM(double** data, unsigned int nbrMinClust, unsigned int nbrMaxClust, unsigned int trialsByCluster, unsigned int nbrDataPoints, unsigned int nbrCoordinates, double weightCov, double weightPosition, double weightOrientation, double weightIntensity, Metric metricToUse, unsigned int* dataRange, unsigned int minSizeData, unsigned int max_iter, double log_like)
{
	this->data = data;
	this->nbrMinClust = nbrMinClust;
	this->nbrMaxClust = nbrMaxClust;
	this->trialsByCluster = trialsByCluster;
	this->nbrDataPoints = nbrDataPoints;
	this->nbrCoordinates = nbrCoordinates;
	this->weightCov = weightCov;
	this->dataRange = dataRange;
	this->minSizeData = minSizeData;
	this->max_iter = max_iter;
	this->log_like = log_like;
	this->weightPosition = weightPosition;
	this->weightOrientation = weightOrientation;
	this->weightIntensity = weightIntensity;
	this->metricToUse = metricToUse;
}

ExpectationMaximization* BestNumberClustersOrientedPoints2EM::findOP2EM()
{
	unsigned int i, j, nbrIterClusters, arrayIterator, bestNbrClustersAIC, bestNbrClustersAICc, bestNbrClustersBIC, bestNbrClustersVOS, bestNbrClustersPI, bestNbrClustersFVQ, bestNbrClustersPC, bestNbrClustersPE;
	double aic, aicc, bic, pi, fvq, pc, pe, vos, aicTemp, aiccTemp, bicTemp, piTemp, fvqTemp, pcTemp, peTemp, vosTemp, overlap, separation;

	ExpectationMaximization* myEM1 = NULL;
	ExpectationMaximization* myEM2 = NULL;

	nbrIterClusters = (nbrMaxClust-nbrMinClust)+1;
	arrayIterator = 0;

	// Time stuff for pseudo-random numbers generation used by U as a first cluster assignation
	srand((unsigned int) time(NULL));

	for (i=nbrMinClust; i<=nbrMaxClust; i++)
	{
		std::cout << i << " "; std::cout.flush();

		for (j=0; j<trialsByCluster; j++)
		{
			std::cout << "*"; std::cout.flush();

			if (arrayIterator==0)
			{
				myEM1 = new OrientedPoints2EM(data, i, nbrDataPoints, nbrCoordinates, false, weightCov, weightPosition, weightOrientation, weightIntensity, metricToUse, dataRange, minSizeData);
				myEM1->em();

				aic = myEM1->getVaic();
				aicc= myEM1->getVaicc();
				bic = myEM1->getVbic();

				overlap = myEM1->getVo();
				separation = myEM1->getVs();
				separation = 1.0-separation;
				vos = overlap/separation;				
				pi = myEM1->getVpi();
				fvq = myEM1->getVfvq();
				pc = myEM1->getVpc();
				pe = myEM1->getVpe();

				bestNbrClustersAIC = 0;
				bestNbrClustersAICc= 0;
				bestNbrClustersBIC = 0;

				bestNbrClustersVOS = 0;
				bestNbrClustersPI  = 0;
				bestNbrClustersFVQ = 0;
				bestNbrClustersPC  = 0;
				bestNbrClustersPE  = 0;
		
				std::cout <<  arrayIterator << " => AIC:  " << aic << " => AICc: " << aicc << " => BIC:  " << bic << " => VOS:  " << vos << " => VPI:  " << pi << " => FVQ:  " << fvq << " => VPC:  " << pc << " => VPE:  " << pe << std::endl;
			}
			else
			{
				myEM2 = new OrientedPoints2EM(data, i, nbrDataPoints, nbrCoordinates, false, weightCov, weightPosition, weightOrientation, weightIntensity, metricToUse, dataRange, minSizeData);
				myEM2->em();
				aicTemp = myEM2->getVaic();
				aiccTemp= myEM2->getVaicc();
				bicTemp = myEM2->getVbic();

				overlap = myEM1->getVo();
				separation = myEM1->getVs();
				separation = 1.0-separation;
				vosTemp = overlap/separation;				
				piTemp = myEM2->getVpi();
				fvqTemp = myEM2->getVfvq();
				pcTemp = myEM2->getVpc();
				peTemp = myEM2->getVpe();

				std::cout << arrayIterator << " => AIC:  " << aicTemp << " => AICc: " << aiccTemp << " => BIC:  " << bicTemp << " => VOS:  " << vosTemp << " => VPI:  " << piTemp << " => FVQ:  " << fvqTemp << " => VPC:  " << pcTemp << " => VPE:  " << peTemp << std::endl;

				// Find best PE
				if (peTemp<pe)
				{
					pe = peTemp;
					bestNbrClustersPE = arrayIterator;
				}

				// Find best PC
				if (pcTemp>pc)
				{
					pc = pcTemp;
					bestNbrClustersPC = arrayIterator;
				}

				// Find best FVQ
				if (fvqTemp<fvq)
				{
					fvq = fvqTemp;
					bestNbrClustersFVQ = arrayIterator;
				}

				// Find best PI
				if (piTemp<pi)
				{
					pi = piTemp;
					bestNbrClustersPI = arrayIterator;
				}

				// Find best VOS
				if (vosTemp<vos)
				{
					vos = vosTemp;
					bestNbrClustersVOS = arrayIterator;
				}

				// Find best BIC
				if (bicTemp<bic)
				{
					bic = bicTemp;
					bestNbrClustersBIC = arrayIterator;
				}

				// Find best AICc
				if (aiccTemp<aicc)
				{
					aicc = aiccTemp;
					bestNbrClustersAICc = arrayIterator;
				}

				// Find best AIC
				if (aicTemp<aic)
				{
					delete myEM1;
					aic = aicTemp;
					myEM1 = myEM2;
					bestNbrClustersAIC = arrayIterator;
				}
				else
					delete myEM2;
			}
			arrayIterator++;
		}

		std::cout << " "; std::cout.flush();
	}

	// Result
	std::cout << std::endl << " ===> bestNbrClusters by AIC :       " << round(bestNbrClustersAIC/trialsByCluster)+2 << " (" << bestNbrClustersAIC << ")";
	std::cout << std::endl << " ===> bestNbrClusters by AICc:      " << 	round(bestNbrClustersAICc/trialsByCluster)+2 << " (" << bestNbrClustersAICc << ")";
	std::cout << std::endl << " ===> bestNbrClusters by BIC :       " << round(bestNbrClustersBIC/trialsByCluster)+2 << " (" << bestNbrClustersBIC << ")";
	std::cout << std::endl << " ===> bestNbrClusters by VOS :       " << round(bestNbrClustersVOS/trialsByCluster)+2 << " (" << bestNbrClustersVOS << ")";
	std::cout << std::endl << " ===> bestNbrClusters by VPI :      " << 	round(bestNbrClustersPI/trialsByCluster)+2 << " (" << bestNbrClustersPI << ")";
	std::cout << std::endl << " ===> bestNbrClusters by FVQ :       " << round(bestNbrClustersFVQ/trialsByCluster)+2 << " (" << bestNbrClustersFVQ << ")";
	std::cout << std::endl << " ===> bestNbrClusters by VPC :      " << 	round(bestNbrClustersPC/trialsByCluster)+2 << " (" << bestNbrClustersPC << ")";
	std::cout << std::endl << " ===> bestNbrClusters by VPE :       " << round(bestNbrClustersPE/trialsByCluster)+2 << " (" << bestNbrClustersPE << ")" << std::endl << std::endl;

	return myEM1;
}
}
