#include <classification/BestNumberClustersOrientedPoints2FCM.h>
#include <classification/OrientedPoints2FCM.h>
#include <basics/Math.h>
#include <iostream>

namespace yav{

BestNumberClustersOrientedPoints2FCM::BestNumberClustersOrientedPoints2FCM(double** data, unsigned int minClust, unsigned int maxClust, unsigned int trials, unsigned int nData, unsigned int nCoord, double wPos, double wOri, double wInt, Metric myMetric, double expo, unsigned int* dataRange, unsigned int minDim, unsigned int maxIte, double minImpro, double* fuzzyPartition, double* initCluster)
{
	this->data = data;
	this->minClust = minClust;
	this->maxClust = maxClust;
	this->trials = trials;
	this->nData = nData;
	this->nCoord = nCoord;
	this->expo = expo;
	this->dataRange = dataRange;
	this->minDim = minDim;
	this->maxIte = maxIte;
	this->minImpro = minImpro;
	this->fuzzyPartition = fuzzyPartition;
	this->initCluster = initCluster;
	this->wPos = wPos;
	this->wOri = wOri;
	this->wInt = wInt;
	this->myMetric = myMetric;
}

FuzzyCMeans* BestNumberClustersOrientedPoints2FCM::findOP2FCM()
{
	unsigned int i, j, range, ite, bestVpc, bestVpi, bestVpe, bestVos, bestVfvq;
	double vpi, vpi2, vpc, vos, vpe, vfvq, vpc2, vos2, vpe2, vfvq2, sep, ove;

	FuzzyCMeans* myFuzzyCMeans1 = NULL;
	FuzzyCMeans* myFuzzyCMeans2 = NULL;

	range = (maxClust-minClust)+1;
	ite = 0;

	// Time stuff for pseudo-random numbers generation used by U as a first cluster assignation
	srand((unsigned int) time(NULL));

	for (i=minClust; i<=maxClust; i++)
	{
		std::cout << i << " "; std::cout.flush();

		for (j=0; j<trials; j++)
		{
			std::cout << "*"; std::cout.flush();

			if (ite==0)
			{
				myFuzzyCMeans1 = new OrientedPoints2FCM(data, i, nData, nCoord, wPos, wOri, wInt, myMetric, false, expo, dataRange, minDim);
				myFuzzyCMeans1->fcm();
				vpi = myFuzzyCMeans1->getVpi();
				ove = myFuzzyCMeans1->getVo();
				sep = myFuzzyCMeans1->getVs();
				vos = ove/sep;
				vfvq= myFuzzyCMeans1->getVfvq();
				vpc = myFuzzyCMeans1->getVpc();
				vpe = myFuzzyCMeans1->getVpe();
				bestVos = 0;
				bestVpc = 0;
				bestVpi  = 0;
				bestVpe = 0;
				bestVfvq= 0;
			}
			else
			{
				myFuzzyCMeans2 = new OrientedPoints2FCM(data, i, nData, nCoord, wPos, wOri, wInt, myMetric, false, expo, dataRange, minDim);
				myFuzzyCMeans2->fcm();
				vpi2 = myFuzzyCMeans2->getVpi();
				ove = myFuzzyCMeans2->getVo();
				sep = myFuzzyCMeans2->getVs();
				vos2 = ove/sep;
				vfvq2= myFuzzyCMeans2->getVfvq();
				vpc2 = myFuzzyCMeans2->getVpc();
				vpe2 = myFuzzyCMeans2->getVpe();

				// Find best VOS
				if (vos2<vos)
				{
					vos = vos2;
					bestVos = ite;
				}
				// Find best VFVQ
				if (vfvq2<vfvq)
				{
					vfvq = vfvq2;
					bestVfvq = ite;
				}
				// Find best VPC
				if (vpc2>vpc)
				{
					vpc = vpc2;
					bestVpc = ite;
				}
				// Find best VPE
				if (vpe2<vpe)
				{
					vpe = vpe2;
					bestVpe = ite;
				}
				// Find best VPI
				if (vpi2<vpi)
				{
					delete myFuzzyCMeans1;
					vpi = vpi2;
					myFuzzyCMeans1 = myFuzzyCMeans2;
					bestVpi = ite;
				}
				else
					delete myFuzzyCMeans2;
			}
			ite++;
		}

		std::cout << " "; std::cout.flush();
	}

	// Result
	std::cout << std::endl << " ===> bestNbrClusters by Vpc :       " << round(bestVpc/trials)+2 << " (" << bestVpc << ") " << vpc;
	std::cout << std::endl << " ===> bestNbrClusters by Vpe :       " << round(bestVpe/trials)+2 << " (" << bestVpe << ") " << vpe;
	std::cout << std::endl << " ===> bestNbrClusters by Vpi :       " << round(bestVpi/trials)+2 << " (" << bestVpi << ") " << vpi;
	std::cout << std::endl << " ===> bestNbrClusters by Vos :       " << round(bestVos/trials)+2 << " (" << bestVos << ") " << vos;
	std::cout << std::endl << " ===> bestNbrClusters by Vfvq:       " << round(bestVfvq/trials)+2 << " (" << bestVfvq << ") " << vfvq << std::endl << std::endl;

	return myFuzzyCMeans1;
}
}
