#include <classification/OrientedPoints2EM.h>
#include <limits>

#define eps 2.2204*(10^-16)

namespace yav{

OrientedPoints2EM::OrientedPoints2EM(double** data, unsigned int nModes, unsigned int nData, unsigned int nDim, bool once, double wCov, double wPos, double wOri, double wInt, Metric myMetric, unsigned int* dataRange, unsigned int minDim, unsigned int maxIte, double minImpro)
: ExpectationMaximization(data, nModes, nData, nDim, wCov, dataRange, minDim, maxIte, minImpro)
{
	unsigned int i, j, sum;
	double delta, stdVarPos, stdVarOri, stdVarInt;

	this->wPos = wPos;
	this->wOri = wOri;
	this->wInt = wInt;
	this->once = once;
	this->myMetric = myMetric;
	stdVarPos = stdVarOri = stdVarInt = 0.0;
	sum = 0;

	mean = new double[nDim];
	var = new double[nDim];
	stdVar = new double[nDim];
	nInt = new unsigned int[nDim];

	for (i=0; i<nDim; i++)
		mean[i] = var[i] = stdVar[i] = 0.0;

	for (i=0; i<nDim; i++)
		nInt[i] = 0;

	// Compute the position, orientation and intensity var through the different dimensions
	for (i=0; i<nData; i++)
	{
		if(isReliable[i])
		{
			sum++;

			for (j=0; j<6; j++)
			{
				delta = data[i][j]-mean[j];
				mean[j] = mean[j]+(delta/(double)sum);
				var[j] = var[j]+(delta*(data[i][j]-mean[j]));
			}
	
			for (j=6+minRange[i]; j<=6+maxRange[i]; j++)
			{
				nInt[j-6]++;
				delta = data[i][j]-mean[j];
				mean[j] = mean[j]+(delta/(double)nInt[j-6]);
				var[j] = var[j]+(delta*(data[i][j]-mean[j]));
			}
		}
	}

	// Compute the standard deviation
	for (i=0; i<nDim; i++)
		stdVar[i] = sqrt(var[i]);

	// Compute the average position and orientation standard deviation
	for (i=0; i<3; i++)
	{
		stdVarPos += stdVar[i];
		stdVarOri += stdVar[i+3];
	}

	for (i=6; i<nDim; i++)
		stdVarInt += stdVar[i]/(double)(nInt[i-6]);

	stdVarPos = stdVarPos/(3.0*((double)sum));
	stdVarOri = stdVarOri/(3.0*((double)sum));
	stdVarInt = stdVarInt/((double)(nDim-6));
	wPosNorm = wPos/stdVarPos;
	wOriNorm = wOri/stdVarOri;
	wIntNorm = wInt/stdVarInt;
}

OrientedPoints2EM::~OrientedPoints2EM()
{
	if (nInt) delete [] nInt;
	if (stdVar) delete [] stdVar;
	if (var) delete [] var;
	if (mean) delete [] mean;
}

/*void OrientedPoints2EM::initEM()
{
	unsigned int i, j;

	OrientedPoints2FCM* myOP2FCM = new OrientedPoints2FCM(data,nModes,nData,nDim,wPos,wOri,wInt,myMetric,once,2.0,dataRange,minDim);

	// Initialize centers by Fuzzy C-Means
	myOP2FCM->fcm();

	for (i=0; i<sizeCen; i++)
		centers[i] = myOP2FCM->getCenters()[i];

	for (i=0; i<nModes; i++)
	{
		// Initialize covariance matrix
		getCovMatrix(myOP2FCM,i);

		// Initialize mode weights
		weightMode[i] = 0.0;

		for (j=0; j<nData; j++)
			if(isReliable[j])
				weightMode[i] += myOP2FCM->getMembership()[(j*nModes)+i];

		weightMode[i] /= nDataUsed;
	}

	delete myOP2FCM;
}*/

}
