#ifndef BESTNUMBERCLUSTERSEM_H
#define BESTNUMBERCLUSTERSEM_H

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <classification/ExpectationMaximization.h>

namespace yav{

class BestNumberClustersEM
{
 protected:
	/// Maximum number of iterations
	unsigned int maxIte;
	/// Minimum number of modes
	unsigned int minModes;
	/// Maximum number of modes
	unsigned int maxModes;
	/// Number of trials
	unsigned int trials;
	/// Number of data
	unsigned int nData;
	/// Data dimension
	unsigned int nDim;
	/// Data minimum dimension
	unsigned int minDim;
	/// Minimum amount of improvment
	double minImpro;

	/// For each data, data range to be considered as reliable [min1 max1 min2 max2...]
	unsigned int* dataRange;

	/// Data
	double** data;
 public:
	/** Initializes the Expectation-Maximization algorithm
	@param data data
	@param minModes minimum numbers of modes
	@param maxModes maximum numbers of modes
	@param trials number of trials for each mode between NBRMINCLUST and NBRMAXCLUST
	@param nData number of data
	@param nDim data dimension
	@param dataRange data range to be considered as reliable [min1 max1 min2 max2...]
	@param minDim data minimum dimension to be considered as reliable (default=1)
	@param maxIte maximum number of iteration (default=1000)
	@param minImpro minimum amount of improvisation between 2 iterations (default=0.00001)*/
	BestNumberClustersEM(double** data, unsigned int minModes, unsigned int maxModes, unsigned int trials, unsigned int nData, unsigned int nDim, unsigned int* dataRange=NULL, unsigned int minDim=1, unsigned int maxIte=1000, double minImpro=0.00001);

	/** @name Destructor */
	/** Deallocates the different arrays used in memory.*/
	~BestNumberClustersEM(){}

	/** @name Other functions */
	/** Finds the best number of modes launching the Expectation-Maximization algorithm using from NBRMINCLUST to NBRMAXCLUST numbers of modes. From each number of modes in this interval, this function launches TRIALSBYCLUSTER times the Expectation-Maximization clustering algorithm.
      @return pointer on ExpectationMaximization*/
	ExpectationMaximization* findEM();
};

}

#endif
