#ifndef BESTNUMBERCLUSTERSFCM_H
#define BESTNUMBERCLUSTERSFCM_H

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <classification/FuzzyCMeans.h>

namespace yav{

class BestNumberClustersFCM
{
 protected:
	/// Maximum number of iterations allowed
	unsigned int maxIte;
	/// Minimum number of clusters
	unsigned int minClust;
	/// Maximum number of clusters
	unsigned int maxClust;
	/// Number of trials for each number of clusters between minClust and maxClust
	unsigned int trials;
	/// Number of data
	unsigned int nData;
	/// Data dimension
	unsigned int nDim;
	/// Minimum data dimension (to be defined as reliable)
	unsigned int minDim;
	/// Fuzziness exponent
	double expo;
	/// Minimum amount of improvement
	double minImpro;

	/// For each data, range to be taken into account [min1 max1 min2 max2...]
	unsigned int* dataRange;
	/// Initial membership
	double* fuzzyPartition;
	/// Initial centers
	double* initCluster;

	/// Data
	double** data;
 public:
	/** Initializes the Fuzzy C-Means
	@param data data
	@param minClust minimum number of clusters
	@param maxClust maximum number of clusters
	@param trials number of trials for each number of clusters between minClust and maxClust
	@param nData number of data
	@param nDim data dimension
	@param expo fuzziness exponent (default = 2.0)
	@param dataRange for each data, range to be taken into account [min1 max1 min2 max2...]
	@param minDim data minimum dimension to be considered as reliable (default=1)
	@param maxIte maximum number of iteration allowed (default=100)
	@param minImpro minimum amount of improvement (default=0.00001)
	@param fuzzyPartition initial membership (default = NULL).
	@param initCluster initial cluster centers (default = NULL).*/
	BestNumberClustersFCM(double** data, unsigned int minClust, unsigned int maxClust, unsigned int trials, unsigned int nData, unsigned int nDim, double expo=2.0, unsigned int* dataRange=NULL, unsigned int minDim=1, unsigned int maxIte=100, double minImpro=0.00001, double* fuzzyPartition=NULL, double* initCluster=NULL);

	/** @name Destructor */
	/** Deallocates the different arrays used in memory.*/
	~BestNumberClustersFCM(){}

	/** @name Other functions */
	/** Finds the best number of clusters.
      @return pointer on FuzzyCMeans*/
	FuzzyCMeans* findFCM();
};

}

#endif
