#ifndef BESTNUMBERCLUSTERSORIENTEDPOINTS2EM_H
#define BESTNUMBERCLUSTERSORIENTEDPOINTS2EM_H

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <classification/OrientedPoints2EM.h>

namespace yav{

class BestNumberClustersOrientedPoints2EM
{
 protected:
	/// Maximum number of iterations
	unsigned int max_iter;
	/// Actual number of clusters
	unsigned int nbrClusters;
	/// Minimum number of clusters
	unsigned int nbrMinClust;
	/// Maximum number of clusters
	unsigned int nbrMaxClust;
	/// Number of trials for each number of clusters between NBRMINCLUST and NBRMAXCLUST
	unsigned int trialsByCluster;
	/// Number of data points
	unsigned int nbrDataPoints;
	/// Number of coordinates for each data point
	unsigned int nbrCoordinates;
	/// Minimum size of data to be defined as reliable
	unsigned int minSizeData;
	/// Percentage of the log likelihood difference between 2 iterations (default=0.1)
	double log_like;
	/** Weight for position normalized */
	double weightPosition;
	/** Weight for orientation normalized */	
	double weightOrientation;
	/** Weight for intensity normalized */	
	double weightIntensity;
	/** Weight to regularize the covariance matrix */
	double weightCov;
	/** Metric to use with intensities. EUCLIDIAN_POSITION uses only the position, EUCLIDIAN_POSITION_ORIENTATION uses both position and orientation and GEODESIC uses a geodesic metric */
	Metric metricToUse;

	/// For each data, range of the data to be taken into account [min1 max1 min2 max2...]
	unsigned int* dataRange;

	/// Array containing data to classify (in x-direction: data size, in y-direction: number of data)
	double** data;
 public:
	/** @name Constructors */
	/** Initializes the Expectation-Maximization algorithm in order to use ORIENTEDPOINTS2FCM
      @param data pointer on 2-dimension array containing data to classify (in x-direction: data size, in y-direction: number of data)
      @param nbrMinClust number minimum of clusters
      @param nbrMaxClust number maximum of clusters
      @param trialsByCluster number of trials for each number of clusters between NBRMINCLUST and NBRMAXCLUST
      @param nbrDataPoints number of data points (size in y-direction of data)
      @param nbrCoordinates number of coordinates for each data point (size in x-direction of data)
	@param weightCov weight to regularize the covariance matrix
	@param dataRange for each data, range of intensities to be taken into account [min1 max1 min2 max2...]
      @param minSizeData minimum intensity profile size to be considered as reliable (default=5)
      @param max_iter maximum number of iteration (default=100)
      @param log_like percentage of the log likelihood difference between 2 iterations (default=0.1)
      @param weightPosition weight for position [0 1] used as a trade-off for classification between position, orientation and intensity
	@param weightOrientation weight for orientation [0 1] used as a trade-off for classification between position, orientation and intensity
	@param weightIntensity weight for orientation [0 1] used as a trade-off for classification between position, orientation and intensity
	@param metricToUse metric to use with intensities */
	BestNumberClustersOrientedPoints2EM(double** data, unsigned int nbrMinClust, unsigned int nbrMaxClust, unsigned int trialsByClusters, unsigned int nbrDataPoints, unsigned int nbrCoordinates, double weightCov=0.0, double weightPosition=1.0, double weightOrientation=1.0, double weightIntensity=1.0, Metric metricToUse=EUCLIDIAN_POSITION, unsigned int* dataRange=NULL, unsigned int minSizeData=1, unsigned int max_iter=100, double log_like=0.1);

	/** @name Destructor */
	/** Deallocates the different arrays used in memory.*/
	~BestNumberClustersOrientedPoints2EM(){}

	/** @name Other functions */
	/** Finds the best number of clusters launching the OrientedPoints2EM class using from NBRMINCLUST to NBRMAXCLUST numbers of clusters. From each number of clusters in this interval, this function launches TRIALSBYCLUSTER times the Expectation-Maximization algorithm.
      @return pointer on ExpectationMaximization*/
	ExpectationMaximization* findOP2EM();
};

}

#endif
