#ifndef BESTNUMBERCLUSTERSORIENTEDPOINTS2FCM_H
#define BESTNUMBERCLUSTERSORIENTEDPOINTS2FCM_H

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <classification/OrientedPoints2FCM.h>

namespace yav{

class BestNumberClustersOrientedPoints2FCM
{
 protected:
	/// Maximum number of iterations allowed
	unsigned int maxIte;
	/// Minimum number of clusters
	unsigned int minClust;
	/// Maximum number of clusters
	unsigned int maxClust;
	/// Number of trials for each number of clusters between minClust and maxClust
	unsigned int trials;
	/// Number of data
	unsigned int nData;
	/// Data dimension
	unsigned int nCoord;
	/// Data minimum dimension (to be defined as reliable)
	unsigned int minDim;
	/// Fuzziness exponent
	double expo;
	/// Minimum amount of improvement
	double minImpro;
	/// Normalized position weight
	double wPos;
	/// Normalized orientation weight
	double wOri;
	/// Normalized intensity weight
	double wInt;
	/// Metric to use with intensities
	Metric myMetric;

	/// For each data, range to be taken into account [min1 max1 min2 max2...]
	unsigned int* dataRange;
	/// Initial membership
	double* fuzzyPartition;
	/// Initial cluster centers
	double* initCluster;

	/// Data
	double** data;
 public:
	/** @name Constructors */
	/** Initializes the Fuzzy C-Means algorithm
      @param data data
      @param minClust minimum number of clusters
      @param maxClust maximum number of clusters
      @param trials number of trials for each number of clusters between minClust and maxClust
      @param nData number of data
      @param nDim data dimension
      @param expo fuzziness exponent (default=2.0)
	@param dataRange for each data, range to be taken into account [min1 max1 min2 max2...]
      @param minDim data minimum dimension to be considered as reliable (default=1)
      @param maxIte maximum number of iteration allowed (default=100)
      @param minImpro minimum amount of improvement (default=0.00001)
      @param fuzzyPartition initial membership (default=NULL)
      @param initCluster initial cluster centers (default=NULL)
      @param wPos position weight [0 1] used as a trade-off between position, orientation and intensity
	@param wOri orientation weight [0 1] used as a trade-off between position, orientation and intensity
	@param wInt orientation weight [0 1] used as a trade-off between position, orientation and intensity
	@param myMetric metric to use with intensities */
	BestNumberClustersOrientedPoints2FCM(double** data, unsigned int minClust, unsigned int maxClust, unsigned int trials, unsigned int nData, unsigned int nDim, double wPos=1.0, double wOri=1.0, double wInt=1.0, Metric myMetric=EUCLIDIAN_POSITION, double expo=2.0, unsigned int* dataRange=NULL, unsigned int minDim=1, unsigned int maxIte=100, double minImpro=0.00001, double* fuzzyPartition=NULL, double* initCluster=NULL);

	/** @name Destructor */
	/** Deallocates the different arrays used in memory.*/
	~BestNumberClustersOrientedPoints2FCM(){}

	/** @name Other functions */
	/** Finds the best number of clusters.
      @return pointer on FuzzyCMeans*/
	FuzzyCMeans* findOP2FCM();
};

}

#endif
