#ifndef ORIENTEDPOINTS2EM_H
#define ORIENTEDPOINTS2EM_H

#include <classification/ExpectationMaximization.h>
#include <classification/OrientedPoints2FCM.h>
#include <basics/TclString.h>
#include <linearAlgebra/Vec3.h>
#include <linearAlgebra/Matrix3x3.h>

namespace yav{

/** ORIENTEDPOINTS2EM class inherits from EXPECTATIONMAXIMIZATION and is used to take the position and orientation of points into account. Position should take the 3 first coordinates of DATA ([0 1 2]) and orientation the 3 following coordinates ([3 4 5]). The next coordinates are voxel intensities */
class OrientedPoints2EM : public ExpectationMaximization
{
 protected:
	/** Position weight */
	double wPos;
	/** Orientation weight */
	double wOri;
	/** Intensity weight */
	double wInt;
	/** Normalized position weight */
	double wPosNorm;
	/** Normalized orientation weight */
	double wOriNorm;
	/** Normalized intensity weight */
	double wIntNorm;
	/** Boolean for the pseudo-random numbers generation */
	bool once;
	/** Metric used with intensities */
	Metric myMetric;

	/** Number of reliable data */
	unsigned int* nInt;
	/** Data mean */
	double* mean;
	/** Data variance */
	double* var;
	/** Data standard deviation */
	double* stdVar;
 public:
	/** @name Constructor */
	/** Initializes the class
      @param data data (data[number_of_data][number_of_coordinates])
      @param nModes number of modes
      @param nData number of data
      @param nDim data dimension
	@param once boolean for the pseudo-random numbers generation used by U. Should be always true (default = true). This boolean is supposed to be false only when used by BESTNUMBERCLUSTERSEM class.
	@param wCov weight to regularize the covar matrix
      @param wPos position weight [0 1] used as a trade-off between position, orientation and intensity
	@param wOri orientation weight [0 1] used as a trade-off between position, orientation and intensity
	@param wInt intensity weight [0 1] used as a trade-off between position, orientation and intensity
	@param myMetric metric used with intensities
	@param dataRange for each data, range of intensities to be taken into account [min1 max1 min2 max2...]
      @param minDim data minimum dimension to be considered as reliable (default=1)
      @param maxIte maximum number of iteration (default=1000)
      @param minImpro minimum improvment between 2 iterations (default=0.00001) */
	OrientedPoints2EM(double** data, unsigned int nModes, unsigned int nData, unsigned int nDim, bool once=true, double wCov=0.0, double wPos=1.0, double wOri=1.0, double wInt=1.0, Metric myMetric=EUCLIDIAN_POSITION, unsigned int* dataRange=NULL, unsigned int minDim=1, unsigned int maxIte=1000, double minImpro=0.00001);

	/** @name Destructor */
	/** Deallocates the different arrays used in memory.*/
	~OrientedPoints2EM();

	/** @name Other functions */
	/** Overwritten function INITEM from the EXPECTATIONMAXIMIZATION class */
	//void initEM();
};

}
#endif
