#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "framework/IteratedActiveObjectManager.h"
#include <framework/IteratedActiveObjectManager.h>
#include <basics/Timer.h>

namespace yav {
	void IteratedActiveObjectManager::activation(){
#ifndef WIN32
		struct timeval tv;
#endif
		unsigned LONGINT startIteration;

		ObjectsList< IteratedActiveObject >::iterator it =
			activeObjectList.begin();
		// tests if there is at list one deformable model to iterate
		bool stopIteration=true;

		for(; it!=activeObjectList.end(); it++) 
		{
			if ((*it)->isActive()==true)
			{
				(*it)->iterationNumber+=nbIterationsPerCall;
				// record the time before deformation
#ifndef WIN32
				gettimeofday(&tv, 0);
				startIteration=tv.tv_sec * 1000000 + tv.tv_usec;
#else
				startIteration=time(0);
#endif
				try{
					(*it)->iterate(nbIterationsPerCall);
				} catch(Exception e){
					CERR << "Problem iterating on object" << ENDL;
					off();
				}
				// get the time after deformation
#ifndef WIN32
				gettimeofday(&tv, 0);
				// updates the duration of deformation
				(*it)->computationTime+=tv.tv_sec * 1000000 + tv.tv_usec -
					startIteration;
#else
				(*it)->computationTime+=time(0)-startIteration;
#endif
				stopIteration=false;
			}
		}
		numberOfIterations--;
		if(stopIteration || (numberOfIterations == 1))
			off();
	}

	void IteratedActiveObjectManager::on() {
		// store the current time in startIteration
		ObjectsList< IteratedActiveObject >::iterator it =
			activeObjectList.begin();
		for(; it!=activeObjectList.end(); it++) 
		{
			(*it)->setActivation(true);
			(*it)->computationTime=0UL;
			(*it)->iterationNumber=0;
		}
		Activable::on();
	}

	void IteratedActiveObjectManager::off(void)
	{
		Activable::off();

	}
	

} // namespace yav

