#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "framework/PCATransformation.h"

#include <linearAlgebra/Vec3.h>
#include <linearAlgebra/Matrix4x4.h>
#include <framework/PCATransformation.h>
#include <framework/Transformation.h>


namespace yav {


  // default constructor:
  PCAData::PCAData() : _reference(NULL) {}

  // read PCA data from a file:
  PCAData::PCAData(PointSetIterator* reference,
		     TclString filename,
		     double* limit) 
    : _reference(reference) {
    read_from_file(reference, filename);
	unsigned int i;

    _limit = new double[_nbModes];
    if (limit)
      for (i = 0; i < _nbModes; i++)
	_limit[i] = limit[i];
    else
      for (i = 0; i < _nbModes; i++)
	_limit[i] = -1.0f;

  }
  // read PCA data from arrays:
  PCAData::PCAData(PointSetIterator* reference,
		     unsigned int nbModes,unsigned  int nbPoints,
		     double* vectors, double* values,
		     double* limit)
    : _reference(reference) {
    read_from_arrays(reference, nbModes, nbPoints, vectors, values);
	unsigned int i;

    _limit = new double[_nbModes];
    if (limit)
      for (i = 0; i < _nbModes; i++)
	_limit[i] = limit[i];
    else
      for (i = 0; i < _nbModes; i++)
	_limit[i] = -1.0f;
  }

  // default destructor:
  PCAData::~PCAData() {
    delete _limit;
  }

  
  // read PCA data from a file:
  void PCAData::read_from_file(PointSetIterator* reference, TclString filename) {
    unsigned int i,j;
    _reference = reference;

    FILE* fmode = fopen((char *)filename, "r");
    if (!fmode) {
      CERR << "unable to open mode file\n";
      // error when opening the file.
    }

    // read the modes: (get the number of points from the reference iterator ?):
    fscanf(fmode, "%i %i\n", &_nbModes, &_nbPoints);
    
    // read the eigen values:
    _eigenValues = new double[_nbModes];
	float eigen;
    for (i = 0; i < _nbModes; i++) {
     fscanf(fmode, "%f", &eigen);
      _eigenValues[i] = eigen;
    }
    
    // read the eigen vectors:
    _eigenVectors = new double[_nbModes * _nbPoints * 3];
    
	
    for (i = 0; i < _nbPoints*3; i++)
      for (j = 0; j < _nbModes; j++) {
	 fscanf(fmode, "%f", &eigen);
	_eigenVectors[i +j*_nbPoints*3] = eigen;
      }
    
    fclose(fmode);


  }
  void PCAData::setModeLimit( const double *limit){
	  unsigned int i;
	  for (i=0;i<_nbModes;++i) {
		  _limit[i]=limit[i];
	  }
  }
  // read PCA data from arrays:
  void PCAData::read_from_arrays(PointSetIterator* reference,
				  unsigned int nbModes, unsigned int nbPoints,
				  double* vectors, double* values) {
    unsigned int i,j;
    _reference = reference;

    _nbModes = nbModes;
    _nbPoints = nbPoints;

    // read the eigen values:
    _eigenValues = new double[_nbModes];
    for (i = 0; i < _nbModes; i++)
      _eigenValues[i] = values[i];
    
    // read the eigen vectors:
    _eigenVectors = new double[_nbModes * _nbPoints * 3];

    for (i = 0; i < _nbPoints*3; i++)
      for ( j = 0; j < _nbModes; j++)
	_eigenVectors[i +j*_nbPoints*3] = vectors[i +j*_nbPoints*3];
  }
  
  // returns the reference point set iterator:
  PointSetIterator* PCAData::getReferenceShape() const {
    return _reference;
  }

  // sets the reference point set iterator:
  void PCAData::setReferenceShape(PointSetIterator* reference) {
    _reference = reference;
  }
  
  // returns the number of modes:
  unsigned int PCAData::getNbModes() const {
    return _nbModes;
  }

  // returns the number of points:
  unsigned int PCAData::getNbPoints() const {
    return _nbPoints;
  }

  // read PCA data from a yav++ GFInput:
  void PCAData::read_from_GFInput(GFInput& input) {
  }

  // write PCA data to a yav++ GFoutput:
  void PCAData::write_to_GFOutput(GFOutput& output) {
  }

  // -------------------------------------------------------------------------


  // default constructor:
  PCATransformation::PCATransformation() { _type = "PCA"; db=0;_mat.Id();}
  PCATransformation::PCATransformation(PointSetIterator* current,
					 PointSetIterator* deformed,
					 PCAData* pcaData)
    : MatrixTransformation(current, deformed) {
    _type = "PCA";
    _pcaData = pcaData;
	if (pcaData) {
	db = new double[pcaData->getNbModes()];
	nbUsefulModes=pcaData->getNbModes();
	for (unsigned int i=0;i<pcaData->getNbModes();++i)
		_pcaData->_eigenValues[i]=sqrt(_pcaData->_eigenValues[i]);
	} else db=0;
	
	_mat.Id();

  }

  // default destructor:
  PCATransformation::~PCATransformation() {
	  if (db)
		  delete [] db;
	  if (_pcaData)
		  delete _pcaData;

  }

  // returns the associated PCAData:
  PCAData* PCATransformation::getPCAData() {
    return _pcaData;
  }

  // sets the current point set iterator:
  void PCATransformation::setPCAData(PCAData* pcaData) {
	  _pcaData = pcaData;
	  if (pcaData) {
		  nbUsefulModes=pcaData->getNbModes();
		  db = new double[pcaData->getNbModes()];
		  for (unsigned int i=0;i<pcaData->getNbModes();++i)
		_pcaData->_eigenValues[i]=sqrt(_pcaData->_eigenValues[i]);
	  } else db=0;
  }

  // computes the PCA transformation:
  void PCATransformation::compute(void) {

    // compute the similarity transformation between the deformed point set
    // and the reference point set:
    SimilarityTransformation transfDefCur(_current,_deformed);
    transfDefCur.compute();
	_mat*=transfDefCur.getMatrix();
    Matrix4x4<double> defcur;
	defcur.Inv(_mat);
	
	unsigned int i,j;
	double x,y,z;
    // project onto the shape space:
    unsigned int _nbPoints = _pcaData->getNbPoints();
    unsigned int _nbModes = _pcaData->getNbModes();

	// if necessary creates the mode amplitude array
	if (!db)
		db= new double [_nbModes];
	for(i = 0; i < nbUsefulModes; i ++)
		db[i] = 0.0;


    // compute the displacement from the mean in the reference frame 
    for(j=0,_deformed->first(), _pcaData->_reference->first();

	!_deformed->isAtEnd();
	(*_deformed)++, (*_pcaData->_reference)++,j++) {
      
      x = defcur[0][0]*_deformed->getX() + 
	defcur[0][1]*_deformed->getY() +
	defcur[0][2]*_deformed->getZ() + 
	defcur[0][3] - _pcaData->_reference->getX();
      y = defcur[1][0]*_deformed->getX() + 
	defcur[1][1]*_deformed->getY() +
	defcur[1][2]*_deformed->getZ() + 
	defcur[1][3] - _pcaData->_reference->getY();
      z = defcur[2][0]*_deformed->getX() + 
	defcur[2][1]*_deformed->getY() +
	defcur[2][2]*_deformed->getZ() + 
	defcur[2][3] - _pcaData->_reference->getZ();

	  /// compute the dot product with each eigen vector and store it in db
	  for(i = 0; i < nbUsefulModes; i ++) {
		  db[i] += _pcaData->_eigenVectors[j*3 +i*_nbPoints*3] *x + 
			  _pcaData->_eigenVectors[j*3+1 +i*_nbPoints*3] * y + 
			  _pcaData->_eigenVectors[j*3+2 +i*_nbPoints*3] * z; 
	  }
    }

	// limit the projections:
	for(i = 0; i < nbUsefulModes; i ++)

		if (_pcaData->_limit[i] > 0) {
			if(db[i] > _pcaData->_limit[i] * _pcaData->_eigenValues[i])
				db[i] = _pcaData->_eigenValues[i];
			if(db[i] < - _pcaData->_limit[i] * _pcaData->_eigenValues[i])
				db[i] = -_pcaData->_eigenValues[i];
		}

  }

  // apply the PCA transformation:
  void PCATransformation::apply(PointSetIterator* set) {


	  _deformed->first();
	  Vec3<double> V;
	  double x,y,z;
	  unsigned int i,j;
      unsigned int _nbPoints = _pcaData->getNbPoints();


	  for(i = 0,_pcaData->_reference->first(); !_deformed->isAtEnd(); i++, (*_deformed)++,(*_pcaData->_reference)++) {

		  x = 0.0, y = 0.0, z = 0.0;

		  for( j = 0; j < nbUsefulModes; j ++) {
			  x += _pcaData->_eigenVectors[i*3 +j*_nbPoints*3] * db[j];
			  y += _pcaData->_eigenVectors[i*3+1 +j*_nbPoints*3] * db[j];
			  z += _pcaData->_eigenVectors[i*3+2 +j*_nbPoints*3] * db[j];
		  }
		  x += _pcaData->_reference->getX();
		  y += _pcaData->_reference->getY();
		  z += _pcaData->_reference->getZ();

		  V=Vec3<double>(x,y,z);

		  _deformed->set(_mat*V);
	  }


  }




} // namespace yav
