#ifndef ACTIVABLE_H
#define ACTIVABLE_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 )
#endif

#pragma interface "framework/Activable.h"

#ifdef WIN32
#include "basics/export.h"
#endif

#include <basics/AutoLink.h>


namespace yav {
  
  //! Prototype for an activable object. An activable object can be activated in the TK events loop. 
  /*! \class Activable Activable.h basics/Activable.h
    \ingroup libBasics
    More detailled description ...
  */
  
  class Activable : public AutoLink<Activable> {
    
  private:

    //! total number of activated objects
    /*! More detailled description ...  */
    static unsigned int activatedObjects;
    
    //! activation flag
    /*! More detailled description ...  */
    bool activated;
    
  protected:
    
    //! activation callback method
    /*! More detailled description ... 
      \sa
    */
    virtual void activation(void) = 0;
    
  public:

    //! A constructor
    /*!
      new object is unactivated
      \sa
    */
    Activable();

    //! A destructor
    /*!
      unactivates object
      \sa
    */
    virtual ~Activable();
    
    //! request activation
    /*! More detailled description ... 
      \sa
    */
    virtual void on(void);

    //! request deactivation
    /*! More detailled description ...
      \sa
    */
    virtual void off(void);

    //! returns activation flag
    /*! More detailled description ... 
      \return true if object is activated and false otherwise
      \sa
    */
    bool active(void) { return activated; }
    
    //! get total number of active objects
    /*! More detailled description ... 
      \return total number of active objects
      \sa
    */    
    static unsigned int getActivatedObjectsCount(void) {
      return activatedObjects;
    }
    
    //! call activation callback method for all active objects
    /*! More detailled description ... 
      \sa
    */    
    static void activate(void);
  };

  
  
} // namespace yav


#endif
