#ifndef INTERFACE_H
#define INTERFACE_H

#pragma interface "framework/Interface.h"
#include <basics/Exception.h>


namespace yav {


class Client;

/*! \class Interface Interface.h basics/Interface.h
  \ingroup libBasics
  interface object that offer interface services to its client 
*/
class Interface {
 public:
  /** @name exception */
  //@{
  /*! \class NullClient  Interface.h basics/Interface.h
    \ingroup libBasics
    null client transmitted to constructor
  */
  class NullClient : public Exception {
  public:
    NullClient(const std::string& from);
  };
  //@}

 protected:
  /// pointer on client object
  Client *client;
  /// destruction flag
  bool destroyClient;

 public:
  /** @name constructor */
  //@{
  /** only constructor from a client object
      @param client non null pointer on client object
      @exception NullClient null pointer transmitted */
#ifndef WIN32
   Interface(Client *client) throw(Interface::NullClient);
#else
   Interface(Client *client) throw(NullClient);
#endif
  /** copy constructor: do not copy pointer on client */
  Interface(const Interface&) : client(0) {}
  //@}

  /** @name destructor */
  //@{
  /// destroy client and this
  virtual ~Interface();
  //@}

  /// returns client
  Client *getData(void);
  /// returns client
  Client *getClient(void) const;
  /** set new client
      @param client new client */
  void setClient(Client *client);

  /// destroy object upon client destruction
  virtual void clientDying(void);
};



} // namespace yav


#include <framework/Client.h>


#endif // ifndef INTERFACE_H
