#ifndef ITERATEDACTIVEOBJECTMANAGER_H
#define ITERATEDACTIVEOBJECTMANAGER_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 )
#endif

#pragma interface "framework/IteratedActiveObjectManager.h"

#include <framework/IteratedActiveObject.h>
#include <framework/Activable.h>


namespace yav {
  
  
//! Prototype for a manager of ActiveObject. The manager has a list of ActiveObjects that are	actived in a synchronous manner. It is mainly used for managing ActivableObjects in a Scene
	/*! \class ActiveGroupManager ActiveGroupManager.h basics/ActiveGroupManager.h
    \ingroup libBasics
    More detailled description ...
  */
class IteratedActiveObjectManager : public Activable {
protected:
	/// list of active objects managed by this object
	ObjectsList< IteratedActiveObject > activeObjectList;
	/// The number of iterations on each activeA objects when the activation function is called
	unsigned int nbIterationsPerCall;
	/// number of iterations remaining (iteration stops when equal to zero)
	unsigned int numberOfIterations;
protected:
	/// increase the number of iteration in a IteratedActiveObject
	void incrementIteration(unsigned int iter, IteratedActiveObject *o) {
		o->iterationNumber+=iter;
	}
	/// increase the number of iteration in a IteratedActiveObject
	void incrementComputationTime(unsigned LONGINT time, IteratedActiveObject *o) {
		o->computationTime+=time;
	}
public :
	//! The constructor
	/*!
	Create a new IteratedActiveObjectManager
	\sa
	*/
	IteratedActiveObjectManager() {
		nbIterationsPerCall=1;
		numberOfIterations=0;
	}

	virtual ~IteratedActiveObjectManager() {}

      /// get the iterator on the first item of the list
	ObjectsListSTLiterator<IteratedActiveObject> firstActiveObject() {
		return activeObjectList.begin();
	}
      /// get the iterator on the last item of the list
	ObjectsListSTLiterator<IteratedActiveObject> lastActiveObject() {
		return activeObjectList.begin();
	}
	/// returns the number of iterations per call to the scene activation routine
  unsigned int getNbIterationsPerCall() const {
	  return nbIterationsPerCall; }
  /// set the number of iterations per call to the scene activation routine
  void setNbIterationsPerCall(const unsigned int n) {
	nbIterationsPerCall=n;
  }
  /// returns the number of iterations 
  unsigned int getNumberOfIterations() const {
	  return numberOfIterations; }
  /// set the number of iterations 
  void setNumberOfIterations(const unsigned int n) {
	numberOfIterations=n;
  }
	/** the callback function that iteratively call the iterate() function of all
	objects in the list */
	virtual void activation();

	/// Initialize iterations by resetting the number of iterations in each objects
    virtual void on(void);

    /// Stop iterations on each object
    virtual void off(void);

	friend class IteratedActiveObject;

};

  
  
} // namespace yav


#endif
