#ifndef POINTSET_H
#define POINTSET_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "framework/PointSet.h"

#include <linearAlgebra/Vec4.h>
#include <linearAlgebra/Matrix4x4.h>


namespace yav {


  /*! \class PointSetIterator PointSet.h basics/PointSet.h 
    \ingroup libBasics
    PointSetIterator base class.
    Iterator over a set of points 
  */
  class PointSetIterator {
  public:

    // constructor:
    PointSetIterator();
    // destructor:
    virtual ~PointSetIterator();

    // reset iterator to first point
    virtual void first();
    // increments iterator
    virtual void operator++();
    // increments iterator
    virtual void operator++(int);
    // returns true if end of list was reached
    virtual bool isAtEnd() const;
    
    // returns current point coordinates
    virtual const Vec3<double> get() const;
    // returns current point weight
    virtual double getWeight() const;
    
    // returns current point X coordinate
    virtual double getX() const;
    // returns current point Y coordinate
    virtual double getY() const;
    // returns current point Z coordinate
    virtual double getZ() const;
    
    // changes the coordinates of the current point
    virtual void set(const Vec3<double>& Coordinates);

    // computes the barycenter of the point set:
    virtual unsigned int barycenter(Vec3<double>& G);

  };



} // namespace yav

#endif
