#ifndef RANGEINFORMATION_H
#define RANGEINFORMATION_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#pragma interface "framework/RangeInformation.h"
#include <basics/Exception.h>
#include <basics/TclString.h>
#include <basics/ListElement.h>
#include <basics/GeneralFormat.h>
#include <map>

namespace yav {

/** \class RangeInformation RangeInformation.h basics/RangeInformation.h
    \ingroup libBasics
    \brief Base class where some a priori information about the properties of the closest point can be stored */
class RangeInformation {
 protected:
   /// name of the boundary finding algorithm
  const TclString name;
 public:
  /** defines the prototype of function that creates a RangeInformation object from a file pointer
      @param input The input stream */
  typedef  RangeInformation * (*RIREADFUNCTION)(GFInput &input); 
  /** a hash table that associates a read function given a keyword */
  static std::map<char *,RIREADFUNCTION,CStringCompare> readFunctionMap; 
  /** @name constructor */
  //@{
  /// empty constructor
  RangeInformation(TclString _name) : name(_name) {}
  //@}
  
  /** @name destructor */
  //@{
  /// empty
  virtual ~RangeInformation() {}
  //@}
  /** create a copy of this RangeInformation object. This is used to apply 
      the same algorithm to all vertices of a mesh */
  virtual RangeInformation* duplicate() {
    return new RangeInformation (*this); }
 
  /// return the name of the boundary information algorithm
  TclString getName() const {
    return name; }
  /** return a RangeInformation given a keyword and a pointer on an input file
      @param key a keyword that indicates the type of RangeInformation
      @param input a pointer on an input file
      @return a range information object corresponding to the input file */
  static  RangeInformation * createFromFile(const char *key,GFInput &input) {
    std::map<char *,RIREADFUNCTION,CStringCompare>::iterator it;
    it=readFunctionMap.find((char *)key);
    if (it==readFunctionMap.end())
      return 0;
    else
      {
	return (*(it->second))(input);
      }
  } 
  /** write in a file the data stored in a Range Information object
      @param output a pointeur on a file */
  virtual void writeInFile(GFOutput &output){}
 };

}
#endif
