// ************************************************************************
//
//   TestError.cc  ------  Test of Error handling (Error.hh / Error.cc)
//
//   C++ translation of a part of K_linalg (c) X. Pennec INRIA 1994 / 1995
//
//	(c) Xavier Pennec -- INRIA / MIT -- Ferbuary 1997
//
// ************************************************************************
//
//
// ************************************************************************


#include <basics/Exception.h>

using namespace yav;


void foo() throw(ErrBug) {
  std::cout << "**** Throw ErrBug in function foo\n";
  throw ErrBug("foo");
}

void bar(){
  std::cout << "**** Calling foo from bar\n";
  try{
    foo();
  }
  catch( ErrBug& e) {
    e.push("bar");
    std::cout << "**** ErrBug trapped in bar, raising exception to higher level\n";
    throw e;
  }
}


int BasicsErrorTest(int argc, char *argv[]){
  try{
    try{
      foo();
    }
    catch( ErrBug& e) {
      std::cout << "**** ErrBug trapped in main:\n";
      std::cout << e << std::endl;
      std::cout << "**** Ok, exception delt with, continue\n";
    }
    
    try{
      bar();
    }
    catch( ErrBug& e) {
      std::cout << "**** ErrBug trapped in main:\n";
      std::cout << e << std::endl;
      std::cout << "**** Ok, exception delt with, continue\n";
    }
    
    try{
      std::cout << "**** Throw ErrUnknown in main\n";
      throw ErrUnknown("main");
    }
    catch( ErrUnknown& e ){
      e.push( "ErrUnknown trapped in main, rethrow it");
      std::cout << "**** ErrUnknown trapped, rethrow it\n";
      throw e;
    }
  }
  catch( Exception& e){
    std::cout << "**** ErrUnknown trapped in main:end, exiting normally\n";
    e.push( "main:end");
    std::cout << e << std::endl;
    return 0;
    }
   catch( std::exception& e){
     std::cerr << "Trapped an unexpected stdlib exception in main:end, exiting\n";
     std::cerr << e.what() << std::endl;
     return( 1);
   }
   catch( std::exception *e){
     std::cerr << "Trapped an unexpected stdlib exception in main:end, exiting\n";
     std::cerr << e->what() << std::endl;
     return ( 1);
   }
  catch(...){
    std::cerr << "Trapped an unexpected exception in main:end, exiting\n";
    return ( 1);
  }

  std::cerr << "**** ERROR: ErrUnknown not trapped in main:end\n";
  return (1);
}

