// ************************************************************************
//
//   TestInv.cc  ------  Test of static matrices classes 
//
//   C++ translation of a part of K_linalg (c) X. Pennec INRIA 1994 / 1995
//
//	(c) Xavier Pennec -- INRIA -- Ferbuary 1997
//
// ************************************************************************
//

#include <linearAlgebra/Mat.h>
#include "BasicsTest.h"

using namespace yav;

const Index N = 100000;



template <class Real>
int TestPseudoInv(void){
  try{
    Real maxdiff = 0;
    int trial =0;
  
    {
      for (Index i =0 ; i< N; i++){
	Matrix<4,Real> M, Minv, MMinv, MinvM, Id4;
	Id4.resetToIdentity();
	Real tmp;
      
	M.Rnd();
	int singular = Minv.PseudoInv(M);
	// since M is random it can be singular
	// in which case its pseudo-inv will also be singular
	// and cannot be tested against Id().
	
	if (singular == 0){
	  MMinv.Mul( M, Minv);
	  MinvM =  Minv * M;
	  if ( (tmp = DistLinf(MMinv, Id4)) > maxdiff ) maxdiff = tmp; 
	  if ( (tmp = DistLinf(MinvM, Id4)) > maxdiff ) maxdiff = tmp; 
	  trial++;

	  if ( maxdiff > EPS(maxdiff) * 1e6 )
	    throw ErrBug("Mat::PseudoInv( Mat& )");
	}
      }
    }
    std::cout << "Functions Mat<n,n>::PeudoInv( Mat& ) appear OK" << std::endl;
    std::cout << "  maximum distance between M.Minv, Minv.M and Id is " << maxdiff 
	 << " ( " << trial << " trials)" << std::endl << std::endl;
    
    
    maxdiff = 0;
    trial =0;
  
    {
      for (Index i =0 ; i< N; i++){
	Mat<4,8,Real> M;
	Mat<8,4,Real> Minv;
	Matrix<4,Real> MMinv, Id4;
	Id4.resetToIdentity();
	Real tmp;
      
	M.Rnd();
	int singular = Minv.PseudoInv(M);
	// since M is random it can be singular
	// if an exception is trown, this matrix is ignored
	
	if (singular == 0){
	  MMinv.Mul( M, Minv);
	  if ( (tmp = DistLinf(MMinv, Id4)) > maxdiff ) maxdiff = tmp; 
	  trial++;

	  if ( maxdiff > EPS(maxdiff) * 1e6 )
	    throw ErrBug("Mat::PseudoInv( Mat& )");
	}
      }
    }
    std::cout << "Functions Mat<n,m,"<<classname<Real>()
	      <<">::PseudoInv( Mat<m,n,"<<classname<Real>()
	      <<">& ) appear OK" << std::endl;
    std::cout << "  maximum distance between M.Minv and Id is " << maxdiff 
	 << " ( " << trial << " trials)" << std::endl << std::endl;
    
    
    {
      maxdiff = 0;
      trial=0;
      for (Index i =0 ; i< N; i++){
	Matrix<3,Real> M, Minv, MMinv, MinvM;
	Real tmp;
	Matrix<3,Real> Id3;
	Id3.resetToIdentity();
      
	M.Rnd();
	int singular = Minv.PseudoInv(M);
	// since M is random it can be singular
	// if an exception is trown, this matrix is ignored
	
	if (singular == 0){
	  MMinv.Mul( M, Minv);
	  MinvM = Minv * M;
	  if ( (tmp = DistLinf(MMinv, Id3)) > maxdiff ) maxdiff = tmp; 
	  if ( (tmp = DistLinf(MinvM, Id3)) > maxdiff ) maxdiff = tmp; 
	  trial++;
	  if ( maxdiff > EPS(maxdiff) *1e6 )
	    throw ErrBug("Matrix<3>::Inv( Matrix<3>& ) or Matrix<3>::Mul(Matrix<3>&, Matrix<3>& )");
	}
      }
    }
    std::cout << "Functions Matrix<3," <<classname<Real>()
	      <<">::Inv( Matrix<3," <<classname<Real>()
	      <<">) appear OK" << std::endl;
    std::cout << "  maximum distance between M.Minv, Minv.M and Id is " << maxdiff 
	 << " ( " << N << " trials)" << std::endl << std::endl;
  
  }
  catch( Exception& e){
    e.push( "main:end");
    std::cerr << e << std::endl;
    return( 1);
  }
  catch( std::exception& e){
    std::cerr << "Trapped an unexpected stdlib exception in main:end, exiting\n";
    std::cerr << e.what() << std::endl;
    return(1);
  }
  catch( std::exception *e){
    std::cerr << "Trapped an unexpected stdlib exception in main:end, exiting\n";
    std::cerr << e->what() << std::endl;
    return(1);
  }
  catch(...){
    std::cerr << "Trapped an unexpected exception in main:end, exiting\n";
    return(1);
  }

  std::cout << "*******************************************************************";
  std::cout << std::endl; 
  std::cout << "Pseudo-Inverse for TEMPLATE CLASS Mat<dim,dim,"
	    <<classname<Real>()<< "> APPEAR TO BE OK" ;
  std::cout << std::endl ;
  std::cout << "*******************************************************************";
  std::cout << std::endl;
  std::cout << std::endl;

  return 0;
}

template <class Real>
int TestInv(void){
  try{
    Real maxdiff = 0;
    int trial =0;

    {
      for (Index i =0 ; i< N; i++){
	Matrix<6,Real> M, Minv, MMinv, MinvM, Id6;
	Id6.resetToIdentity();
	Real tmp;
      
	try { 
	  M.Rnd();
	  Minv.Inv(M);
	  // since M is random it can be singular
	  // if an exception is trown, this matrix is ignored

	  MMinv.Mul( M, Minv);
	  MinvM =  Minv * M;
	  if ( (tmp = DistLinf(MMinv, Id6)) > maxdiff ) maxdiff = tmp; 
	  if ( (tmp = DistLinf(MinvM, Id6)) > maxdiff ) maxdiff = tmp; 
	  trial++;
	  if ( maxdiff > EPS(maxdiff) * 1e6 )
	    throw ErrBug("Mat::Inv( Mat& ) or Mat::Mul(Mat&, Mat&)");
	}
	// since M is random it can be singular
	catch( ErrMatSingular& e ){trial--;}
      }
    }
    std::cout << "Functions Matrix<n,"<<classname<Real>()
	      <<">::Inv( Mat& ) appear OK" << std::endl;
    std::cout << "  maximum distance between M.Minv, Minv.M and Id is " << maxdiff 
	      << " ( " << trial << " trials)" << std::endl << std::endl;
    
    trial = 0;
    {
      for (Index i =0 ; i< N; i++){
	Matrix<4,Real> M, Minv, MMinv, MinvM, Id4;
	Id4.resetToIdentity();
	Real tmp;
      
	try { 
	  M.Rnd();
	  Minv.Inv(M);
	  // since M is random it can be singular
	  // if an exception is trown, this matrix is ignored

	  MMinv.Mul( M, Minv);
	  MinvM =  Minv * M;
	  if ( (tmp = DistLinf(MMinv, Id4)) > maxdiff ) maxdiff = tmp; 
	  if ( (tmp = DistLinf(MinvM, Id4)) > maxdiff ) maxdiff = tmp; 
	  trial++;
	  if ( maxdiff > EPS(maxdiff) * 1e6 )
	    throw ErrBug("Mat::Inv( Mat& ) or Mat::Mul(Mat&, Mat&)");
	}
	// since M is random it can be singular
	catch( ErrMatSingular& e ){trial--;}
      }
    }
    std::cout << "Functions Matrix<4,"<<classname<Real>()
	      <<">::Inv( Mat& ) appear OK" << std::endl;
    std::cout << "  maximum distance between M.Minv, Minv.M and Id is " << maxdiff 
	      << " ( " << trial << " trials)" << std::endl << std::endl;
    

    trial = 0;
    {
      maxdiff = 0;
      for (Index i =0 ; i< N; i++){
	Matrix<3,Real> M, Minv, MMinv, MinvM;
	Real tmp;
	Matrix<3,Real> Id3;
	Id3.resetToIdentity();
      
	try{
	  M.Rnd();
	  Minv.Inv(M);
	  // since M is random it can be singular
	  // exception makes it ignored
	  MMinv.Mul( M, Minv);
	  MinvM = Minv * M;
	  if ( (tmp = DistLinf(MMinv, Id3)) > maxdiff ) maxdiff = tmp; 
	  if ( (tmp = DistLinf(MinvM, Id3)) > maxdiff ) maxdiff = tmp; 
	  trial++;
	  if ( maxdiff > EPS(maxdiff) *1e6 )
	    throw ErrBug("Matrix<3>::Inv( Matrix<3>& ) or Matrix<3>::Mul(Matrix<3>&, Matrix<3>& )");
	}
	catch( ErrMatSingular& e) {trial--;};
      }
    }
    std::cout << "Functions Matrix<3,"<<classname<Real>()
	      <<">::Inv( Matrix<3,"<<classname<Real>()
	      <<">) appear OK" << std::endl;
    std::cout << "  maximum distance between M.Minv, Minv.M and Id is " << maxdiff 
	 << " ( " << N << " trials)" << std::endl << std::endl;
  
  }
  catch( Exception& e){
    e.push( "main:end");
    std::cerr << e << std::endl;
    return(1);
  }
  catch( std::exception& e){
    std::cerr << "Trapped an unexpected stdlib exception in main:end, exiting\n";
    std::cerr << e.what() << std::endl;
    return(1);
  }
  catch( std::exception *e){
    std::cerr << "Trapped an unexpected stdlib exception in main:end, exiting\n";
    std::cerr << e->what() << std::endl;
    return(1);
  }
  catch(...){
    std::cerr << "Trapped an unexpected exception in main:end, exiting\n";
    return(1);
  }

  std::cout << "*******************************************************************";
  std::cout << std::endl; 
  std::cout << "Inverse for TEMPLATE CLASS Matrix<dim,"
	    <<classname<Real>()<< "> APPEAR TO BE OK" ;
  std::cout << std::endl ;
  std::cout << "*******************************************************************";
  std::cout << std::endl;
  std::cout << std::endl;

  return 0;
}


int BasicsInvTest(int argc, char *argv[]){
  clock_t begin=0;
  clock_t end=0;
  
//     TestInv<float>();
//     TestInv<double>();


  begin=clock();
  int test_double = TestPseudoInv<double>();;
  end=clock();
  std::cout << "Clock time: " << (end-begin)/ ( (double) CLOCKS_PER_SEC)  << std::endl;
  
  begin=clock();
  int test_float= TestPseudoInv<float>();;
  end=clock();
  std::cout << "Clock time: " << (end-begin)/ ( (double) CLOCKS_PER_SEC) << std::endl;


  begin=clock();
  int test2_double = TestInv<double>();;
  end=clock();
  std::cout << "Clock time: " << (end-begin)/ ( (double) CLOCKS_PER_SEC)  << std::endl;
  
  begin=clock();
  int test2_float= TestInv<float>();;
  end=clock();
  std::cout << "Clock time: " << (end-begin)/ ( (double) CLOCKS_PER_SEC) << std::endl;
  
  return( test_double && test_float && test2_double && test2_float );
}
