// ************************************************************************
//
//   TestMat.cc  ------  Test of static matrices classes 
//
//   C++ translation of a part of K_linalg (c) X. Pennec INRIA 1994 / 1995
//
//	(c) Xavier Pennec -- INRIA -- Ferbuary 1997
//
// ************************************************************************
//

#include <linearAlgebra/Vec.h>
#include <linearAlgebra/Mat.h>
#include "BasicsTest.h"


using namespace yav;

const Index N = 100000;



template <class Real>
int TestMat(void){
  try{
    Real n;
    Mat<10,5,Real> M;
    M.Null();
    
    
    std::cout << std::endl;
    std::cout << "*******************************************************************";
    std::cout << std::endl; 
    std::cout << "Verifying TEMPLATE CLASS Mat<dim,dim,"<<classname<Real>()<<
      ">, and Sym<dim,"<<classname<Real>()<<"> " ;
    std::cout << std::endl ;
    std::cout << "*******************************************************************";
    std::cout << std::endl << std::endl;


    std::cout << "M set to null:\n" << M ;
    std::cout << std::endl << "Norm(M) = " << (n = M.Norm()) << std::endl;
    
    if ( n != 0 ){
      throw ErrBug("Mat<row,col,Real>::Norm()");
    }
    std::cout << "Functions Mat<row,col,"<<classname<Real>()<<">::Norm() appear OK" << std::endl;
    
    RndInit();
    
    std::cout << std::endl << "Test of transpose and multiply" << std::endl;
    {
      for (Index i =0 ; i< N; i++){
	Mat<3,4,Real> M1,M2trp;
	Mat<4,3,Real> M2,M1trp;
	Mat<3,3,Real> M,Mtrp;
	M2.Rnd();M1.Rnd();
	
	M.Mul( M1, M2);
	M1trp.Trp( M1);
	M2trp.Trp( M2);
	Mtrp = M2trp * M1trp;
	Mtrp.Trp( Mtrp);
	if ( DistLinf(M, Mtrp) != 0 )
	  throw ErrBug("Mat::Trp( Mat& ) or Mat::Mul(Mat&, Mat&)");
      }
    }
    std::cout << "Functions Mat::Trp( Mat& ) and Mat::Mul(Mat&, Mat&) appear OK" << std::endl;
    
    // Test of operator +, - and unary -
    // -----------------------------------------------------
    {
      RndInit();
      std::cout << std::endl;
      
      Real maxdiff =0;
      for(Index i=0; i< N; i++){
	Mat<4,6, Real> u, v, upv, upv_func, umv, umv_func, upmv;
	u.Rnd(); v.Rnd();
	upv_func.Add(u,v);
	upv = u + v;
	umv = u - v;
	umv_func.Sub(u,v);
	upmv.Add(u, -v);

	maxdiff = max( maxdiff, 
		       max( DistLinf( upv, upv_func),
			    max( DistLinf( umv, umv_func),
		  		 DistLinf( umv, upmv) )));
	
	if ( maxdiff > EPS( maxdiff)  )
	  throw ErrBug("Operator + (Mat&, Mat&) or operator - (Mat&, Mat&) or unary operator  (Mat&) is bugged");
      }
      std::cout << "Operators +, - (Mat&, Mat&) and - (Mat&) appears OK (maxdiff = " 
	   << maxdiff << ")" << std::endl;
    }

  
    std::cout << std::endl << "Test of square matrices: identity, inversion and determinant" << std::endl;
    std::cout << "-------------------------------------------" << std::endl << std::endl;
    Matrix<4,Real> Id4, Id4inv;
    Id4.resetToIdentity();
  
    Id4inv.Inv( Id4);
    if ( DistLinf(Id4, Id4inv) != 0 )
      throw ErrBug("Mat::resetToIdentity() or Mat& Mat::Inv(Mat&)");
  
    Real maxdiff = 0;
  
    {
      for (Index i =0 ; i< N; i++){
	Matrix<4,Real> M, Minv, MMinv, MinvM;
	Real tmp;
      
	try { 
	  M.Rnd();
	  Minv.Inv(M);
	  // since M is random it can be singular
	  // if an exception is trown, this matrix is ignored

	  MMinv.Mul( M, Minv);
	  MinvM =  Minv * M;
	  if ( (tmp = DistLinf(MMinv, Id4)) > maxdiff ) maxdiff = tmp; 
	  if ( (tmp = DistLinf(MinvM, Id4)) > maxdiff ) maxdiff = tmp; 
	
	  if ( maxdiff > EPS( maxdiff) * 1e6 )
	    throw ErrBug("Mat::Inv( Mat& ) or Mat::Mul(Mat&, Mat&)");
	}
	// since M is random it can be singular
	catch( ErrMatSingular& e ){}
      }
    }
    std::cout << "Functions Mat::resetToIdentity() and Mat::Inv( Mat& ) appear OK" << std::endl;
    std::cout << "  maximum distance between M.Minv, Minv.M and Id is " << maxdiff 
	 << " ( " << N << " trials)" << std::endl << std::endl;
    {
      maxdiff = 0;
      for (Index i =0 ; i< N; i++){
	Matrix<3,Real> M, Minv, MMinv, MinvM;
	Real tmp;
	Matrix<3,Real> Id3;
	Id3.resetToIdentity();
      
	try{
	  M.Rnd();
	  Minv.Inv(M);
	  // since M is random it can be singular
	  // exception makes it ignored
	  MMinv.Mul( M, Minv);
	  MinvM = Minv * M;
	  if ( (tmp = DistLinf(MMinv, Id3)) > maxdiff ) maxdiff = tmp; 
	  if ( (tmp = DistLinf(MinvM, Id3)) > maxdiff ) maxdiff = tmp; 
	
	  if ( maxdiff > EPS( maxdiff) *1e6 )
	    throw ErrBug("Matrix<3>::Inv( Matrix<3>& ) or Matrix<3>::Mul(Matrix<3>&, Matrix<3>& )");
	}
	catch( ErrMatSingular& e) {};
      }
    }
    std::cout << "Functions Matrix::resetToIdentity() and Matrix<3,"<<classname<Real>()<<">::Inv( Matrix<3,"
	 <<classname<Real>()<<">) appear OK" << std::endl;
    std::cout << "  maximum distance between M.Minv, Minv.M and Id is " << maxdiff 
	 << " ( " << N << " trials)" << std::endl << std::endl;
  
    // Test of determinant
    {
      maxdiff = 0;
      for (Index i =0 ; i< N; i++){
	Matrix<4,Real> M1, M2, M;
	Real det1,det2, det12,det21,tmp;

	M1.Rnd(); M2.Rnd();
	det1 = M1.Det();
	det2 = M2.Det();
	M.Mul(M1,M2);
	det12 = M.Det();
	M = M2 * M1;
	det21 = M.Det();
	if ( (tmp = fabs(det21-det12)) > maxdiff ) maxdiff = tmp; 
	if ( (tmp = fabs(det1*det2-det21)) > maxdiff ) maxdiff = tmp; 
	if ( (tmp = fabs(det1*det2-det12)) > maxdiff ) maxdiff = tmp; 
	
	if ( maxdiff > EPS( maxdiff) *1e6 )
	  throw ErrBug("Matrix<dim,Real>::Det()");
      }
    }
    std::cout << "Functions Matrix<dim,"<<classname<Real>()<<">::Det() appear OK" << std::endl;
    std::cout << "  maximum distance between |M1|.|M2|, |M1.M2| and |M2.M1| is " << maxdiff 
	 << " ( " << N << " trials)" << std::endl << std::endl;
  

    // Test of determinant (specialisation dim 2)
    {
      maxdiff = 0;
      for (Index i =0 ; i< N; i++){
	Matrix<2,Real> M1, M2, M;
	Real det1,det2, det12,det21,tmp;

	M1.Rnd(); M2.Rnd();
	det1 = M1.Det();
	det2 = M2.Det();
	M.Mul(M1,M2);
	det12 = M.Det();
	M = M2 *M1;
	det21 = M.Det();
	if ( (tmp = fabs(det21-det12)) > maxdiff ) maxdiff = tmp; 
	if ( (tmp = fabs(det1*det2-det21)) > maxdiff ) maxdiff = tmp; 
	if ( (tmp = fabs(det1*det2-det12)) > maxdiff ) maxdiff = tmp; 
	
	if ( maxdiff > EPS( maxdiff) *1e6 )
	  throw ErrBug("Matrix<2,Real>::Det()");
      }
    }
    std::cout << "Functions Matrix<2,"<<classname<Real>()<<">::Det() appear OK" << std::endl;
    std::cout << "  maximum distance between |M1|.|M2|, |M1.M2| and |M2.M1| is " << maxdiff 
	 << " ( " << N << " trials)" << std::endl << std::endl;
  

    // Test of determinant (specialisation dim 3)
    {
      maxdiff = 0;
      for (Index i =0 ; i< N; i++){
	Matrix<3,Real> M1, M2, M;
	Real det1,det2, det12,det21,tmp;

	M1.Rnd(); M2.Rnd();
	det1 = M1.Det();
	det2 = M2.Det();
	M.Mul(M1,M2);
	det12 = M.Det();
	M.Mul(M2,M1);
	det21 = M.Det();
	if ( (tmp = fabs(det21-det12)) > maxdiff ) maxdiff = tmp; 
	if ( (tmp = fabs(det1*det2-det21)) > maxdiff ) maxdiff = tmp; 
	if ( (tmp = fabs(det1*det2-det12)) > maxdiff ) maxdiff = tmp; 
	
	if ( maxdiff > EPS( maxdiff) *1e6 )
	  throw ErrBug("Matrix<3,Real>::Det()");
      }
    }
    std::cout << "Functions Matrix<3,"<<classname<Real>()<<">::Det() appear OK" 
	 << std::endl;
    std::cout << "  maximum distance between |M1|.|M2|, |M1.M2| and |M2.M1| is " 
	 << maxdiff 
	 << " ( " << N << " trials)" << std::endl << std::endl;


    // Test of operator +, - and unary -
    // -----------------------------------------------------
    {
      RndInit();
      std::cout << std::endl;
      
      Real maxdiff =0;
      for(Index i=0; i< N; i++){
	Matrix<4, Real> u, v, upv, upv_func, umv, umv_func, upmv;
	u.Rnd(); v.Rnd();
	upv_func.Add(u,v);
	upv = u+v;
	umv = u - v;
	umv_func.Sub(u,v);
	upmv.Add(u, -v);
	
	maxdiff = max( maxdiff, 
		       max( DistLinf( upv, upv_func),
			    max( DistLinf( umv, umv_func),
				 DistLinf( umv, upmv) )));
	
	if ( maxdiff > EPS( maxdiff)  )
	  throw ErrBug("Operator + (Matrix<n>&, Matrix<n>&) or operator - (Matrix<n>&, Matrix<n>&) or unary operator - (Matrix<n>&) is bugged");
      }
      std::cout << "Operators +, - (Matrix<n>&, Matrix<n>&) and - (Matrix<n>&) appears OK (maxdiff = " 
	   << maxdiff << ")" << std::endl;
    }
    
  
    std::cout << std::endl << "Test of symmetric matrices" << std::endl;
    std::cout << "-------------------------------------------" << std::endl << std::endl;
    Sym<6,Real> W;
    W.Null();

    {
      for(Index i=0; i<6; i++)
	for(Index j=0; j<6; j++)
	  W(i,j) = i+j;
    }

    {
      for(Index i=0; i<6; i++)
	for(Index j=0; j<6; j++){
	  if ( W(i,j) != i+j )
	    throw ErrBug("operator W(i,j)");
	}
    }
    std::cout << "Operators Sym::(i,j) appear ok" << std::endl;
    std::cout << std::endl;
    std::cout << "Printing test: W(i,j) = i+j:" << std::endl;
    std::cout << W << std::endl;


    Vect<6,Real> v;
  
    {
      for(Index i=0; i<6; i++)
	v(i) = i;
    }

    W.Diag(v);
    if ( fabs( W.Trace() - 15.) > EPS( maxdiff) )
      throw ErrBug("Sym::Diag( const Vect& ) or Sym::Trace()");
    std::cout << "Functions Sym::Diag( const Vect& ) and Sym::Trace() appear ok" << std::endl;


    maxdiff = 0;
    {
      for (Index i =0 ; i< N; i++){
      
	// Fast action
	Mat<4,5,Real> J; J.Rnd();
	Sym<5,Real> Win; Win.Rnd();
	Sym<4,Real> Wout; J.SymAction(Win, Wout);
      
	// Standard chain operation 
	Matrix<5,Real> Min; Sym2Mat( Win, (Mat<5,5,Real>&) Min);
	Mat<5,4,Real> Jtrp; Jtrp.Trp( J);
	J.Mul( J, Min);
	Matrix<4,Real> Mout; 
	Mout.Mul( J, Jtrp);
	Sym<4,Real> W2out( Mout);
      
	Real tmp;
	if ( (tmp = DistLinf(Wout, W2out)) > maxdiff ) maxdiff = tmp; 
      
	if ( maxdiff > EPS( maxdiff)*10. )
	  throw ErrBug("Sym2Mat(Sym, Mat) or Mat2Sym( Mat, Sym) or Mat::SymAction(Sym, Sym)");
      }
    }
    std::cout << std::endl;
    std::cout << "Functions Sym2Mat(Sym, Mat) and Mat2Sym( Mat, Sym)"
	 << " and  Mat::SymAction(Sym, Sym) appear OK" << std::endl;
    std::cout << "  maximum distance between fast and standard action is " 
	 << maxdiff << " ( " << N << " trials)" << std::endl;


    maxdiff = 0;
    {
      for (Index i =0 ; i< N; i++){
      
	// Fast Autocorrelation
	Vect<3,Real> v; v.Rnd();
	Sym<3,Real> W; 
	W.AutoCorrel(v);
	      
	// Standard chain operation 
	Mat<3,3,Real> M; 
	M.Mul( (Mat<3,1,Real>&) v, (Mat<1,3,Real>&) v );
	Sym<3,Real> Wb(M); 
	      
	Real tmp = DistLinf( W, Wb);
	if ( tmp > maxdiff ) maxdiff = tmp; 
      
	if ( maxdiff > EPS( maxdiff)*10.0 )
	  throw ErrBug("Sym::AutoCorrel( Vect )");
      }
    }
    std::cout << std::endl;
    std::cout << "Functions Sym::AutoCorrel( Vect ) appears OK" << std::endl;
    std::cout << "  maximum distance between fast and standard chain is " 
	 << maxdiff << " ( " << N << " trials)" << std::endl;
  
     maxdiff = 0;
    {
      for (Index i =0 ; i< N; i++){
      
	// Fast action
	Vect<3,Real> v; v.Rnd();
	Sym<3,Real> W; W.Rnd();
	Real SqNorm = W.SqNorm( v );
      
	// Standard chain operation 
	Matrix<3,Real> M(W); //Sym2Mat( W, M);
	Vect<3,Real> Mv; 
	//Mv.Mul( M, v);
	Mv = M * v;

	Real SqNorm2 = DotProd(v, Mv);
      
	Real tmp = fabs( SqNorm - SqNorm2);
	if ( tmp > maxdiff ) maxdiff = tmp; 
      
	if ( maxdiff > EPS( maxdiff)*10.0 )
	  throw ErrBug("Sym::SqNorm( Vect ) or DotProd( Vect, Vect)");
      }
    }
    std::cout << std::endl;
    std::cout << "Functions Sym::SqNorm( Vect ) appears OK" << std::endl;
    std::cout << "  maximum distance between fast and standard action is " 
	 << maxdiff << " ( " << N << " trials)" << std::endl;
  
 
  
    maxdiff = 0;
    Real itermean =0, iter, itermax = 0;
    {
      for (Index i =0 ; i< N; i++){
      
	// Diagonalizition
	Sym<5,Real> W; W.Rnd(); 
	Vect<5,Real> d; Matrix<5,Real> V;
	try{
	  iter = W.Diagonalize( V, d);
	}
	catch( ErrConvergence& e){
	  // If the algorithm does not converge, print this curious matrix
	  std::cerr << "Jacobi failed for the matrix:\n" << W << std::endl;
	  throw e;
	}
	itermean += iter;
	if (iter > itermax) itermax = iter;
      
      
	// Reconstitution
	Matrix<5,Real> Vtrp; Vtrp.Trp(V);
	Matrix<5,Real> Diag(d);
      
	Diag.Mul( V, Diag);
	Vtrp.Mul( Diag, Vtrp);
	Sym<5,Real> Wout(Vtrp);
      
	Real tmp;
	if ( (tmp = DistLinf(Wout, W)) > maxdiff ) maxdiff = tmp; 
      
	if ( maxdiff > EPS( maxdiff) *10.0 ){
	  std::cout << " maxdiff = " << maxdiff <<std::endl;
	  throw ErrBug("Sym::Diagonalize( Mat, Vect )");
	}
      }
    }
    std::cout << std::endl;
    std::cout << "Function Sym::Diagonalize( Mat, Vect ) appears OK" << std::endl;
    std::cout << "  maximum distance between W and V.Diag(D).Vtrp is " 
	 << maxdiff << " ( " << N << " trials)" << std::endl;
    std::cout << "  mean number of Jacobi iterations " << itermean / N << 
      " and maximum " << itermax << std::endl;


    maxdiff = 0;
    Real detmaxdiff=0;
    {
      for (Index i =0 ; i< N; i++){
      
	// Diagonalizition
	Sym<5,Real> W; W.Rnd(); 
	// Vect<5,Real> d; Matrix<5,Real> V;
	Sym<5,Real> Winv; 

	try{ 
	  // Index singular = Inv( W, Winv);
	  Index singular = Winv.inv(W);
	  Real dd = W.Det() * Winv.Det(); 
       
	  // see how close W.Winv.W is from W
	  Mat<5,5,Real> M; Sym2Mat(W, M);
	  Sym<5,Real> W2; M.SymAction(Winv, W2);
	
	  Real tmp;
	  if ( (tmp = DistLinf(W2, W)) > maxdiff ) maxdiff = tmp; 
	
	  if ( maxdiff > EPS( maxdiff)*1e5 )
	    throw ErrBug("Sym::Inv( Sym, Sym )");
	
	  // Test determinant
	  if ( singular == 0 ){
	    if ( (tmp = fabs(dd - 1.0)) > detmaxdiff ) 
	      detmaxdiff = tmp;
	  } 
	  else
	    if ( (tmp = fabs(dd)) > detmaxdiff ) 
	      detmaxdiff = tmp; 
	
	  if ( detmaxdiff > EPS( maxdiff)*1e5 )
	    throw ErrBug("Sym::Det()");
	}
	catch(ErrConvergence& e){
	  std::cerr << "Jacobi did not converge for matrix: \n" << W <<std::endl;
	  throw e;
	}

      }
    }
    std::cout << std::endl;
    std::cout << "Functions Sym::Inv( Sym, Sym ) and Dym::Det() appear OK" << std::endl;
    std::cout << "  maximum distance between det(W) and det(Winv) is " 
	 << detmaxdiff << " ( " << N << " trials)" << std::endl;
    std::cout << "  maximum distance between W and W.Winv.W is " 
	 << maxdiff << " ( " << N << " trials)" << std::endl;


        // Test of operator +, - and unary -
    // -----------------------------------------------------
    {
      RndInit();
      std::cout << std::endl;
      
      Real maxdiff =0;
      for(Index i=0; i< N; i++){
	Sym<4, Real> u, v, upv, upv_func, umv, umv_func, upmv;
	u.Rnd(); v.Rnd();
	upv_func.Add(u,v);
	upv = u+v;
	umv = u - v;
	umv_func.Sub(u,v);
	upmv.Add(u, -v);
	
	maxdiff = max( maxdiff, 
		       max( DistLinf( upv, upv_func),
			    max( DistLinf( umv, umv_func),
				 DistLinf( umv, upmv) )));
	
	if ( maxdiff > EPS( maxdiff)  )
	  throw ErrBug("Operator + (Sym&, Sym&) or operator - (Sym&, Sym&) or unary operator - (Sym&) is bugged");
      }
      std::cout << "Operators +, - (Sym&, Sym&) and - (Sym&) appears OK (maxdiff = " 
	   << maxdiff << ")" << std::endl;
    }
  

  }
  catch( Exception& e){
    e.push( "main:end");
    std::cerr << e << std::endl;
    return( 1);
  }
  catch( std::exception& e){
    std::cerr << "Trapped an unexpected stdlib exception in main:end, exiting\n";
    std::cerr << e.what() << std::endl;
    return(1);
  }
  catch( std::exception *e){
    std::cerr << "Trapped an unexpected stdlib exception in main:end, exiting\n";
    std::cerr << e->what() << std::endl;
    return(1);
  }
  catch(...){
    std::cerr << "Trapped an unexpected exception in main:end, exiting\n";
    return(1);
  }

  std::cout << std::endl;
  std::cout << "*******************************************************************";
  std::cout << std::endl; 
  std::cout << "TEMPLATE CLASS Mat<dim,dim,"<<classname<Real>()<<
    ">, and Sym<dim,"<<classname<Real>()<<"> APPEAR TO BE OK" ;
  std::cout << std::endl ;
  std::cout << "*******************************************************************";
  std::cout << std::endl;

  return 0;
}


#include <time.h>

int BasicsMatTest(int argc, char *argv[]){
  clock_t begin=0;
  clock_t end=0;
  
  begin=clock();
  int test_double = TestMat<double>();
  end=clock();
  std::cout << "Clock time: " << (end-begin)/ ( (double) CLOCKS_PER_SEC)  << std::endl;
  
  begin=clock();
  int test_float= TestMat<float>();
  end=clock();
  std::cout << "Clock time: " << (end-begin)/ ( (double) CLOCKS_PER_SEC) << std::endl;
  
  return( test_double && test_float );

}

