// ************************************************************************
//
//   TestQuat.cc  ------  Test of the Vect<4> specialization
//
//   C++ translation of a part of K_linalg (c) X. Pennec INRIA 1994 / 1995
//
//	(c) Xavier Pennec -- INRIA -- Ferbuary 1997
//
// ************************************************************************
//


#include <linearAlgebra/Vec4.h>
#include "BasicsTest.h"

using namespace yav;


template<class Real>
int TestQuat(){
  try{
    Index ITER = 100000;
    {
      std::cout << std::endl << "******************************************" <<std::endl; 
      std::cout << "Verifying CLASS Vec4<"<<classname<Real>()<<"> "<< std::endl ;
      std::cout << "******************************************" << std::endl << std::endl;


      std::cout << "Verifying initialisation:" << std::endl;
      Vec4<Real> e0(0), e1(1), e2(2), e3(3);
      std::cout << "e0 : " << e0 ;
      std::cout << "e1 : " << e1 ;
      std::cout << "e2 : " << e2 ;
      std::cout << "e3 : " << e3 << std::endl;
    
      Vec4<Real> q(1,2,3,4);
      for(Index i=0;i<4;i++)
	if ( q(i) != i+1){
	  throw ErrBug("Vec4<Real>::Vect( Real, Real, Real, Real) ");
	}
      {
	Vec3<Real> v(1,2,3);
	Vec4<Real> qq(0,v);
	for(Index i=0;i<4;i++)
	  if ( qq(i) != i)
	    throw ErrBug("Vec4<Real>::Vect( Real, Vect<3>&) ");
	std::cout << "Vec4<"<<classname<Real>()<<">(Index), Vec4<"<<
	  classname<Real>()<<">("<<classname<Real>()<<", "<<
	  classname<Real>()<<", "<<classname<Real>()<<", "<<
	  classname<Real>()<<") and Vec4<"<<classname<Real>()<<
	  ">(Real, Vect<3,"<<classname<Real>()<<">&) appear OK" << std::endl;
      }

      std::cout << "Verifying access to data (RealPart / VectPart)" << std::endl;
      {
	Vec4<Real> q; q.Rnd();
	Real& a = q.RealPart();
	Vect<3,Real>& v = q.VectPart();
	v.Rnd(); a = 2;
	Vec4<Real> q1( a, v);
	if( DistLinf( q1, q) > 0)
	  throw ErrBug("Vec4<Real>( q.RealPart(), q.VectPart() ) != q");
	std::cout << "Real& Vec4<"<<classname<Real>()<<
	  ">::RealPart() and Vect<3,"<<classname<Real>()<<
	  ">& Vec4<"<<classname<Real>()<<">::VectPart() appear OK" << std::endl;
      }

      Real tmp=0;

      std::cout << std::endl << "Verifying the Quaternion product on basis vectors:" << std::endl;
      Vec4<Real> res; 
      QuatProd(e0,e0, res);
      if( DistLinf( res, e0) > EPS(tmp))
	throw ErrBug("QuatProd: e0 * e0 should be e0");
    
      QuatProd(e1,e1, res);
      res *= -1;
      if( DistLinf( res, e0) > EPS(tmp))
	throw ErrBug("QuatProd: e1 * e1 should be -e0");

      QuatProd(e2,e2, res);
      res *= -1;
      if( DistLinf( res, e0) > EPS(tmp))
	throw ErrBug("QuatProd: e2 * e2 should be -e0");

      QuatProd(e3,e3, res);
      res *= -1;
      if( DistLinf( res, e0) > EPS(tmp))
	throw ErrBug("QuatProd: e3 * e3 should be -e0");
    
      QuatProd(e1,e2, res);
      if( DistLinf( res, e3) > EPS(tmp))
	throw ErrBug("QuatProd: e1 * e2 should be e3");

      QuatProd(e2,e1, res);
      res *= -1;
      if( DistLinf( res, e3) > EPS(tmp))
	throw ErrBug("QuatProd: e2 * e1 should be -e3");

      QuatProd(e2,e3, res);
      if( DistLinf( res, e1) > EPS(tmp))
	throw ErrBug("QuatProd: e2 * e3 should be e1");

      QuatProd(e3,e2, res);
      res *= -1;
      if( DistLinf( res, e1) > EPS(tmp))
	throw ErrBug("QuatProd: e3 * e2 should be -e1");

      QuatProd(e3,e1, res);
      if( DistLinf( res, e2) > EPS(tmp))
	throw ErrBug("QuatProd: e3 * e1 should be e2");

      QuatProd(e1,e3, res);
      res *= -1;
      if( DistLinf( res, e2) > EPS(tmp))
	throw ErrBug("QuatProd: e1 * e3 should be -e2");

      std::cout << "QuatProd(const Vec4<"<<classname<Real>()<<
	">&, const Vec4<"<<classname<Real>()<<">&, Vec4<"<<
	classname<Real>()<<">&) is OK on basis vectors" << std::endl;
    }
  
    {
      Real maxdiff = 0, d;
      Vec4<Real> barq, res;
      Vec4<Real>  Qid(0), tmp, q;
      std::cout << std::endl << "Verifying Conjugate Quaternion" << std::endl;
      for(Index iter=0; iter < ITER; iter++){
	q.Rnd();
	tmp.Mul( DotProd( q,q), Qid);
	QuatConj(q, barq);
	QuatProd(q, barq, res);
	if( (d = DistLinf( res, tmp)) > maxdiff) 
	  maxdiff = d;
	QuatProd(barq, q, res);
	if( (d = DistLinf( res, tmp)) > maxdiff) 
	  maxdiff = d;
      }
    
      if ( maxdiff > EPS(maxdiff) )
	throw ErrBug("QuatConj or QuatProd");
      std::cout << "QuatConj( const Vec4<"<<classname<Real>()<<
	">&, Vec4<"<<classname<Real>()<<">&) appears OK" << std::endl;
    }
  
    {
      Real maxdiff = 0, d;
      Vec4<Real> qinv, res;
      Vec4<Real>  Qid(0), q;
      std::cout << std::endl << "Verifying Inverse Quaternion" << std::endl;
      for(Index iter=0; iter < ITER; iter++)
	try {
	  q.Rnd();
	  QuatInv( q, qinv);
	  QuatProd(q, qinv, res);
	  if( (d = DistLinf( res, Qid)) > maxdiff) 
	    maxdiff = d;
	  QuatProd(qinv, q, res);
	  if( (d = DistLinf( res, Qid)) > maxdiff) 
	    maxdiff = d;
	}
      catch( ErrDivZero ) {};
    
      if ( maxdiff > EPS(maxdiff) )
	throw ErrBug("InvConj or QuatProd");
      std::cout << "QuatInv( const Quat&, Vec4<"<<classname<Real>()<<
	">&) appears OK" << std::endl;
      std::cout << "  maximum distance between q*qinv and ID is " << maxdiff 
	   << " ( " << ITER << " trials)" << std::endl;
    }

    {
      Real maxdiff = 0, d;
      Vec4<Real> q1, q2, q, qq;
      Matrix<4,Real> Q1;
      std::cout << std::endl << "Verifying TildeLeft operation" << std::endl;
      for(Index iter=0; iter < ITER; iter++){
	q1.Rnd(); q2.Rnd();
	TildeLeft(q1,Q1);
	QuatProd(q1,q2,q);
	qq.Mul(Q1,q2);
	if( (d = DistLinf( q, qq)) > maxdiff) 
	  maxdiff = d;
      }
      if ( maxdiff > EPS(maxdiff) )
	throw ErrBug("TildeLeft");
      std::cout << "TildeLeft( const Vec4<"<<classname<Real>()<<
	">&, Matrix<4,"<<classname<Real>()<<">&) appears OK" << std::endl;
      std::cout << "  maximum distance between q*p and Q.p is " << maxdiff 
	   << " ( " << ITER << " trials)" << std::endl;
    }
  
    {
      Real maxdiff = 0, d;
      Vec4<Real> q1, q2, q, qq;
      Matrix<4,Real> Q1;
      std::cout << std::endl << "Verifying TildeRight operation" << std::endl;
      for(Index iter=0; iter < ITER; iter++){
	q1.Rnd(); q2.Rnd();
	TildeRight(q1,Q1);
	QuatProd(q2,q1,q);
	qq.Mul(Q1,q2);
	if( (d = DistLinf( q, qq)) > maxdiff) 
	  maxdiff = d;
      }
      if ( maxdiff > EPS(maxdiff) )
	throw ErrBug("TildeRight");
      std::cout << "TildeRight( const Vec4<"<<classname<Real>()<<
	">&, Matrix<4,"<<classname<Real>()<<">&) appears OK" << std::endl;
      std::cout << "  maximum distance between q*p and P.q is " << maxdiff 
	   << " ( " << ITER << " trials)" << std::endl;
    }

    {
      Real maxdiff = 0, d;
      Vec4<Real> q;
      Matrix<4,Real> Q, Qanti;
      std::cout << std::endl << "Verifying Anti operation" << std::endl;
      for(Index iter=0; iter < ITER; iter++){
	q.Rnd();
	TildeRight(q,Q);
	Anti(q, Qanti);
	Qanti.Trp(Qanti);
	if( (d = DistLinf( Q, Qanti)) > maxdiff) 
	  maxdiff = d;
      }
      if ( maxdiff > EPS(maxdiff) )
	throw ErrBug("Anti");
      std::cout << "Anti( const Vec4<"<<classname<Real>()<<
	">&, Matrix<4,"<<classname<Real>()<<">&) appears OK" << std::endl;
      std::cout << "  maximum distance between TildeRight and trp(Anti) is " << maxdiff 
	   << " ( " << ITER << " trials)" << std::endl;
    }

  
  
  }
  catch( Exception& e){
    e.push( "main:end");
    std::cerr << e << std::endl;
    return(1);
  }
  catch( std::exception& e){
    std::cerr << "Trapped an unexpected stdlib exception in main:end, exiting\n";
    std::cerr << e.what() << std::endl;
    return(1);
  }
  catch( std::exception *e){
    std::cerr << "Trapped an unexpected stdlib exception in main:end, exiting\n";
    std::cerr << e->what() << std::endl;
    return(1);
  }
  catch(...){
    std::cerr << "Trapped an unexpected exception in main:end, exiting\n";
    return(1);
  }


  std::cout << std::endl << "******************************************" <<std::endl; 
  std::cout << "CLASS Vec4<"<<classname<Real>()<<"> APPEARS TO BE OK" << std::endl ;
  std::cout << "******************************************" << std::endl;

  return 0;
}



#include <time.h>

int BasicsQuatTest(int argc, char *argv[]){
  clock_t begin=0;
  clock_t end=0;
  
  begin=clock();
  int test_double = TestQuat<double>();
  end=clock();
  std::cout << "Clock time: " << (end-begin)/ ( (double) CLOCKS_PER_SEC)  << std::endl;
  
  begin=clock();
  int test_float= TestQuat<float>();
  end=clock();
  std::cout << "Clock time: " << (end-begin)/ ( (double) CLOCKS_PER_SEC) << std::endl;
  
  return( test_double && test_float );

}

