// **********************************************************************
//
//   BasicsRotationTest  ------  Test of 3D Rotation group class 
//
//   C++ translation of a part of PRigid.c (c) X. Pennec INRIA 1994 / 1995
//
//	(c) Xavier Pennec -- INRIA -- November 1997 / July 2003
//
// **********************************************************************

#include <linearAlgebra/Rotation3D.h>
#include "BasicsTest.h"


using namespace yav;

const Index Niter = 100000;
const double PRECISION = 1e-9;
const double PRECISION_ACOS = 1e-7; // Absolute accuracy of around 0 is acos is 1.49012e-08

//double min( double a, double b){ if (a<b) return a; else return b;}
//double max( double a, double b){ if (a>b) return a; else return b;}


int BasicsRotationTest(int argc, char *argv[]){
  int count_errors =0;

  try{
    
    // RndInit();
    std::cout.precision(15);
    
    
    {// Test 1 *****************************************************
      Index count, countEuc;
      std::cout <<("Simple Quat/RV translations (Quat -> RV -> Quat)\n");
      std::cout << "-------------------------------------------" << std::endl ;
      
      count = countEuc = 0;
      double maxdiff =0.0, maxdiffEuc =0.0;
      for (Index i=0;i<Niter;i++) {
	UnitQuat q, qp;
	RotVect r;
	
	qp = r = q.Rnd();
	double normEuc = DistEuc( q, qp);
	// Correction because this is not a distance on rotation quaternions
	normEuc = min( normEuc, 2.0-normEuc);
      maxdiffEuc = max( maxdiffEuc, normEuc);
      
      if (normEuc > PRECISION){ 
	printf("Quaternion error (extrinsinc Euclidean distance): %g\n", normEuc);
	std::cout << "q : "<< q<< std::endl;
	std::cout << "r : "<< r<< std::endl;
	std::cout << "qp: "<< qp <<std::endl;
	countEuc ++;
      }
      
      double norm = DistRot( q, qp);
      maxdiff = max( maxdiff, norm);

      if (norm > PRECISION_ACOS ) { // Absolute accuracy of acos is 1.49012e-08
	printf("Quaternion error (intrinsic distance on UniQuat): %g\n", norm);
	std::cout << "q : "<< q << std::endl;
	std::cout << "r : "<< r<< std::endl;
	std::cout << "qp: "<< qp <<std::endl;
	count ++;
      }
      
    }
    if (count >0)
      std::cout << "Test Quat <-> RV ended : "
	   << count << " errors detected in intrinsinc metric (" <<
	((double) count)/((double) Niter)*100. << "%%)" << std::endl;
    else
      std::cout << "   No error detected in intrinsinc metric " << std::endl;

    count_errors += count;

    if (countEuc >0)
      std::cout << "Test Quat <-> RV ended : "
	   << countEuc << " errors detected in Extrinsinc Euclidean metric (" <<
	((double) countEuc)/((double) Niter)*100. << "%)" << std::endl;
    else
      std::cout << "   No error detected in Extrinsinc Euclidean metric " << std::endl;
    count_errors += countEuc;
    

    std::cout << "   Max differences:\n\t    Dist : " << maxdiff
	 << "\n\t    DistEuc = " << maxdiffEuc << std::endl;
    std::cout << "-------------------------------------------" << std::endl << std::endl;
  } 
   

  { // Test 2 *****************************************************
    std::cout <<("Simple RV/RM translations ( [Quat ->] RV -> RM -> RV)\n");
    std::cout << "-------------------------------------------" << std::endl ;
    
    Index count = 0;
    double maxdiff = 0.0;
    for (Index i=0;i<Niter;i++) {
      UnitQuat q;
      RotMat3D rot;
      RotVect r,rp;
      
      rp = rot = r = q.Rnd();
      double norm = Dist( r, rp);
      maxdiff = max( maxdiff, norm);
      if (norm > PRECISION){
	printf("Rotation Vector error : %g\n", norm);
	std::cout << "r: "<< r << std::endl;
	std::cout << "rot: "<< rot<< std::endl;
	std::cout << "rp: "<< rp <<std::endl;
	count ++;
      }
    }
    if (count >0)
      std::cout << "Test RotVect <-> RotMat ended : "
	   << count << " errors detected in intrinsinc metric (" <<
	((double) count)/((double) Niter)*100. << "%)" << std::endl;
    else
      std::cout << "   No error detected in intrinsinc metric " << std::endl;
    
    count_errors += count;

    std::cout << "   Max differences:\n\t    Dist : " << maxdiff << std::endl;
    std::cout << "-------------------------------------------" << std::endl << std::endl;
  }
 
  {// Test 3 *****************************************************
    Index count, countEuc;
    std::cout <<("Simple Quat/RM translations ( [Quat -> RV ->]  RM -> Quat -> RM)\n");
    std::cout << "-------------------------------------------" << std::endl ;
 
    count = countEuc = 0;
    double maxdiff =0.0, maxdiffEuc =0.0;
    for (Index i=0;i<Niter;i++) {
      UnitQuat q;
      RotVect r;
      RotMat3D rot, rotp;
      
      rot = r = q.Rnd();
      rotp = q = rot;

      double normEuc = DistEuc( rot, rotp);
      maxdiffEuc = max( maxdiffEuc, normEuc);

      if (normEuc > PRECISION){ 
	printf("Error (extrinsinc Euclidean distance on Rotations): %g\n", normEuc);
	std::cout << "rot : "<< std::endl << rot<<std::endl;
	std::cout << "q : "<< q <<std::endl;
	std::cout << "rotp: " << rotp <<std::endl;
	countEuc ++;
      }
      
      double norm = Dist( rot, rotp);
     maxdiff = max( maxdiff, norm);

      if (norm > PRECISION_ACOS ) { // Absolute accuracy of around 0 is acos is 1.49012e-08
	printf("Error (intrinsic Rotation distance): %g\n", norm);
	std::cout << "rot : "<< std::endl << rot<<std::endl;
	std::cout << "q : "<< q<<std::endl;
	std::cout << "rotp: "<< std::endl << rotp <<std::endl;
	count ++;
      }
      
    }
    if (count >0)
      std::cout << "Test Quat <-> RV ended : "
	   << count << " errors detected in intrinsinc metric (" <<
	((double) count)/((double) Niter)*100. << "%)" << std::endl;
    else
      std::cout << "   No error detected in intrinsinc metric " << std::endl;

    count_errors += count + countEuc;

    if (countEuc >0)
      std::cout << "Test Quat <-> RV ended : "
	   << countEuc << " errors detected in Extrinsinc Euclidean metric (" <<
	((double) countEuc)/((double) Niter)*100. << "%)" << std::endl;
    else
      std::cout << "   No error detected in Extrinsinc Euclidean metric " << std::endl;
    std::cout << "   Max differences:\n\t    Dist : " << maxdiff
	 << "\n\t    DistEuc = " << maxdiffEuc << std::endl;
    std::cout << "-------------------------------------------" << std::endl << std::endl;
  } 


  { // Test 4 *****************************************************
    std::cout <<("Homogeneity of intrinsic distances (RM, RV, Quat)\n");
    std::cout << "-------------------------------------------" << std::endl ;
    
    Index count = 0;
    double maxdiff_q_rot =0, maxdiff_r_rot =0, maxdiff_q_r =0;

    for (Index i=0;i<Niter;i++) {
      UnitQuat q1,q2;
      RotMat3D rot1,rot2;
      RotVect r1,r2;
      
      r1 = q1.Rnd();
      rot1 = q1;
      r2 = q2.Rnd();
      rot2 = q2;
      
      double dq = DistRot( q1, q2);
      double dr = Dist( r1, r2);
      double drot = Dist( rot1, rot2);

      maxdiff_q_rot = max( maxdiff_q_rot , fabs(dq -drot));
      maxdiff_q_r = max( maxdiff_q_r , fabs(dq -dr));
      maxdiff_r_rot = max( maxdiff_r_rot , fabs(dr -drot));

      if ( (fabs(dq -dr) > PRECISION_ACOS) ||
	   (fabs(drot -dr) > PRECISION_ACOS) ||
	   (fabs(dq -drot) > PRECISION_ACOS) ){
	std::cout << "Distance error : \n\t    Quat / RV = " << fabs(dq -dr)
	     << "\n\t    Quat / RM = " << fabs(dq -drot)
	     << "\n\t    RM / RV = " << fabs(drot -dr) << std::endl;
 	std::cout << "dr : "<< dr <<std::endl;
 	std::cout << "drot : "<< drot <<std::endl;
 	std::cout << "dq: "<< dq <<std::endl;
	count ++;
      }
    }
    
    if (count >0)
      std::cout << "Test distances ended : "
	   << count << " errors detected in intrinsinc distances (" <<
	((double) count)/((double) Niter)*100. << "%)" << std::endl;
    else
      std::cout << "   No error detected in intrinsinc distances " << std::endl;
    
    count_errors += count;


    std::cout << "   Max differences:\n\t    Quat / RV = " << maxdiff_q_r
	 << "\n\t    Quat / RM = " << maxdiff_q_rot
	 << "\n\t    RM / RV = " << maxdiff_r_rot << std::endl;
    std::cout << "-------------------------------------------" << std::endl << std::endl;
    }

   { // Test 4 *****************************************************
    std::cout <<("Verification of Comp and Inv (RM, RV, Quat)\n");
    std::cout << "-------------------------------------------" << std::endl ;
    
    Index count = 0;
    double maxdiff_q =0, maxdiff_r =0, maxdiff_rot =0;

    for (Index i=0;i<Niter;i++) {
      UnitQuat q,qinv, qqinv,qinvq, qid;
      RotMat3D rot,rotinv, rotrotinv, rotinvrot, rotid;
      RotVect r, rinv, rrinv, rinvr, rid;
      
      r = q.Rnd();
      rot = q;
      
      qinv.Inv(q);
      qqinv.Comp(q,qinv);
      qinvq.Comp(qinv,q);
      double dq = max( DistRot( qid, qqinv), DistRot( qid, qinvq));
      maxdiff_q = max( maxdiff_q , dq);

      rinv.Inv(r);
      rrinv.Comp(r,rinv);
      rinvr.Comp(rinv,r);
      double dr = max( Dist( rid, rrinv), Dist( rid, rinvr));
      maxdiff_r = max( maxdiff_r , dr);

      rotinv.Inv(rot);
      rotrotinv.Comp(rot,rotinv);
      rotinvrot.Comp(rotinv,rot);
      double drot = max( Dist( rotid, rotrotinv), Dist( rotid, rotinvrot));
      maxdiff_rot = max( maxdiff_rot , drot);

      if ( (dq > PRECISION_ACOS) ||
	   (dr > PRECISION_ACOS) ||
	   (drot > PRECISION_ACOS) ){
	std::cout << "Error\t q * qinv / qinv * q / Id = " << dq
	     << "\n\t r * rinv / rinv * r / Id = " << dr
	     << "\n\t R * Rinv / Rinv * R / Id = " << drot << std::endl;
	count ++;
      }
    }
    
    if (count >0)
      std::cout << "Test Comp and Inv (1) ended : "
	   << count << " errors detected in Comp or Inv (" <<
	((double) count)/((double) Niter)*100. << "%)" << std::endl;
    else
      std::cout << "   No error detected in Comp or Inv " << std::endl;

    count_errors += count;

    std::cout << "   Max differences:\n\t    q * qinv / qinv * q / Id = " << maxdiff_q
	 << "\n\t    r * rinv / rinv * r / Id = " << maxdiff_r
	 << "\n\t    R * Rinv / Rinv * R / Id = " << maxdiff_rot << std::endl;
    std::cout << "-------------------------------------------" << std::endl << std::endl;
    }

    { // Test 5 *****************************************************
    std::cout <<("Homogeneity of composition and inversion (RM, RV, Quat)\n");
    std::cout << "-------------------------------------------" << std::endl ;
    
    Index count = 0;
    double maxdiff_comp =0, maxdiff_inv =0;

    for (Index i=0;i<Niter;i++) {
      UnitQuat q1,q2,q,qinv;
      RotMat3D rot1,rot2,rot,rotinv;
      RotVect r1,r2,r,rinv;
      
      r1 = q1.Rnd();
      rot1 = q1;
      r2 = q2.Rnd();
      rot2 = q2;

      q.Comp(q1,q2);
      qinv.Inv(q);

      r.Comp(r1,r2);
      rinv.Inv(r);

      rot.Comp(rot1,rot2);
      rotinv.Inv(rot);

      double dcomp =  max( Dist( r, r2 =q ),  Dist( r, r2 =rot));
      double dinv = max( Dist( rinv, r2 =qinv ), Dist( rinv, r2 = rotinv));

      maxdiff_comp = max( maxdiff_comp, dcomp);
      maxdiff_inv = max( maxdiff_inv, dinv);
 
      if ( (dcomp > PRECISION_ACOS) ||
	   (dinv > PRECISION_ACOS) ){
	std::cout << "Composition or inversion error : " 
	     << "\n\t    Comp = " <<  maxdiff_comp
	     << "\n\t    Inv = " << maxdiff_inv <<std::endl;
	count ++;
      }
    }
    
    if (count >0)
      std::cout << "Test distances ended : "
	   << count << " errors detected Comp and Inv (RM, RV, Quat) (" <<
	((double) count)/((double) Niter)*100. << "%)" << std::endl;
    else
      std::cout << "   No error detected in Composition and Inv (RM, RV, Quat) " << std::endl;
    
    count_errors += count;

    std::cout << "Maximum differences: " 
	 << "\n\t    Comp = " <<  maxdiff_comp
	 << "\n\t    Inv = " << maxdiff_inv <<std::endl;
    std::cout << "-------------------------------------------" << std::endl << std::endl;
    }
   
   
   { // Test of LSQRot
     std::cout <<("Test of LSQ rotation estimation \n");
     std::cout << "-------------------------------------------" << std::endl ;
     
     double maxdiff = 0;
     for(Index i=0; i<Niter;i++){
       RotMat3D R;
       R.Rnd();
       
       Matrix3x3<double> K(R);
       RotMat3D Restim(K);
       
       double d = DistLinf( (const Matrix3x3<double>) Restim, K);
       maxdiff = max( maxdiff,  d);
       
       if ( maxdiff > EPS(maxdiff) *100.0 ){
	 std::cout << "Original rotation : K"<< std::endl << K;
	 std::cout << "Estimated matrix : "<< std::endl << Restim << std::endl;
	 
	 throw ErrBug("LSQRot( Mat<m,n>&, Mat<n,n>&)");
       }
     }
     
     std::cout << std::endl;
     std::cout << "    No error detected in RotMat3D::LSQ( Mat<3>& )" << std::endl;
     std::cout << "  maxdiff R / LSQ(R) is " << maxdiff << " ( " << Niter << " trials)" << std::endl;
     std::cout << "-------------------------------------------" << std::endl << std::endl;
   }

  }
  catch( Exception& e){
    e.push( "main:end");
    std::cerr << e << std::endl;
    return(1);
  }
  catch( std::exception& e){
    std::cerr << "Trapped an unexpected stdlib exception in main:end, exiting\n";
    std::cerr << e.what() << std::endl;
    return(1);
  }
  catch( std::exception *e){
    std::cerr << "Trapped an unexpected stdlib exception in main:end, exiting\n";
    std::cerr << e->what() << std::endl;
    return(1);
  }
  catch(...){
    std::cerr << "Trapped an unexpected exception in main:end, exiting\n";
    return(1);
  }
  return count_errors;
}

