// ************************************************************************
//
//   TestVect3.cc  ------  Test of the Vect<3> specialization
//
//   C++ translation of a part of K_linalg (c) X. Pennec INRIA 1994 / 1995
//
//	(c) Xavier Pennec -- INRIA -- Ferbuary 1997
//
// ************************************************************************
//

#include <linearAlgebra/Vec3.h>
#include "BasicsTest.h"

using namespace yav;


template<class Real>
int TestVect3(void){
  try{

   Vec3<Real> e1(1,0,0), e2(0,1,0), e3 (0,0,1);
   Vec3<Real> res;
  
    std::cout << std::endl << "******************************************" <<std::endl; 
    std::cout << "Verifying CLASSVec3<"<<classname<Real>()<<"> "<< std::endl ;
    std::cout << "******************************************" << std::endl << std::endl;

    std::cout << "Verify initialisation:" << std::endl;
    {
      std::cout << "e1 : " << e1 ;
      std::cout << "e2 : " << e2 ;
      std::cout << "e3 : " << e3 << std::endl;
    
      res.Null();
      std::cout << "Verify cross product:" << std::endl;
      res.Cross(e1,e2);
      std::cout << "e1 x e2 = (should be e3) " << res ;
      res.Cross(e2,e3);
      std::cout << "e2 x e3 = (should be e1) " << res ;
      res.Cross(e1,e3);
      std::cout << "e1 x e3 = (should be -e2)" << res << std::endl;
    }

    std::cout << "Verify Tilde (associated antisymmetric matrix):\n";
    {
      Mat<3,3,Real> Utilde;
     Vec3<Real> u, v, uxv, Utilde_v; 
      RndInit();
      u.Rnd(); v.Rnd();
      std::cout << "u = " << u;
      std::cout << "v = " << v;
    
      uxv.Cross(u, v);
      Tilde(u, Utilde);
       Utilde_v.Mul(Utilde, v);
      Real d = Dist(  Utilde_v, uxv);
      std::cout << "Dist( U.v, u x v ) = " << d << std::endl;
      if ( d != 0 )
	throw ErrBug("Function Mul( Mat, Vect, Vect ) or Tilde(Vec3<Real>, Mat<3,3,Real>) is bugged");
      std::cout << "Functions Mul( Mat, Vect, Vect ) and Tilde(Vec3<"<<
	classname<Real>()<<">, Mat<3,3,"<<classname<Real>()<<">) appear OK" << std::endl;

      Utilde_v.Mul( v, Utilde);

      Utilde_v *= -1;
      d = Dist(  Utilde_v, uxv);
      if ( d != 0 )
	throw ErrBug("Function Mul( Vect, Mat, Vect ) or Tilde(Vec3<Real>, Mat<3,3,Real>) is bugged");
      std::cout << "Functions Mul( Vect, Mat, Vect ) and Tilde(Vec3<"
	   <<classname<Real>()<<">, Mat<3,3,"<<classname<Real>()<<">) appear OK"
	   << std::endl;
    }

  }
  catch( Exception& e){
    e.push( "main:end");
    std::cerr << e << std::endl;
    return( 1);
    }
  catch( std::exception& e){
    std::cerr << "Trapped an unexpected stdlib exception in main:end, exiting\n";
    std::cerr << e.what() << std::endl;
    return(1);
  }
  catch( std::exception *e){
    std::cerr << "Trapped an unexpected stdlib exception in main:end, exiting\n";
    std::cerr << e->what() << std::endl;
    return(1);
  }
  catch(...){
    std::cerr << "Trapped an unexpected exception in main:end, exiting\n";
    return(1);
  }


  std::cout << std::endl << "******************************************" <<std::endl; 
  std::cout << "CLASSVec3<"<<classname<Real>()<<"> APPEARS TO BE OK" << std::endl ;
  std::cout << "******************************************" << std::endl;
  return 0;
}


int BasicsVect3Test(int argc, char *argv[]){
  clock_t begin=0;
  clock_t end=0;
  
  begin=clock();
  int test_double = TestVect3<double>();
  end=clock();
  std::cout << "Clock time: " << (end-begin)/ ( (double) CLOCKS_PER_SEC)  << std::endl;
  
  begin=clock();
  int test_float= TestVect3<float>();
  end=clock();
  std::cout << "Clock time: " << (end-begin)/ ( (double) CLOCKS_PER_SEC) << std::endl;
  
  return( test_double && test_float );
}


