#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#ifdef USE_MTL

#pragma implementation "linearAlgebra/MTLSparseMatrix.h"
#include <linearAlgebra/MTLSparseMatrix.h>
#include <string>
using namespace std;
#include <itl/interface/mtl.h>
#include <mtl/utils.h>
#include <itl/itl.h>
#include <itl/krylov/bicgstab.h>
#include <itl/preconditioner/cholesky.h>
#include <itl/preconditioner/ssor.h>
#include <itl/preconditioner/ilu.h>
#include <itl/preconditioner/ilut.h>
#include <itl/krylov/gmres.h>
#include <itl/modified_gram_schmidt.h>
#include <itl/krylov/qmr.h>
#include <itl/krylov/cg.h>


namespace yav {

	MTLSparseVector::MTLSparseVector(const unsigned int nbRows) : vector(0){
		 vector= new mtl::dense1D<double>(nbRows);
	}

	MTLSparseVector::~MTLSparseVector(){
		if (vector)
			delete vector;
	}


	void MTLSparseVector::addVectorElement(const unsigned int row,
		const  double value){
		(*vector)[row]+=value;
	}

		/// set a value to a vector
	void MTLSparseVector::setVectorElement(const unsigned int row,
		const  double value){
					(*vector)[row]=value;
	}
		/// set vector to zero
	void MTLSparseVector::resetVector(){
		double *buf=vector->data();
		memset((void *)buf,0,vector->size()*sizeof(double));
	}
	double MTLSparseVector::getVectorElement(const unsigned int row) const{
		return (*vector)[row];
	}

	const double *MTLSparseVector::getVectorAsArray(){
		return (const double *) &((*vector)[0]);
	}


	/// get vector size
	unsigned int MTLSparseVector::getSize() const{
		return (*vector).size();
	}

	void MTLSparseVector::saveInFile(const std::string filename) const{
		OFSTREAM output((const char *) filename.c_str());
		unsigned int i;
		for(i=0;i<vector->size();++i)
		  output << (*vector)[i] << " ";
		output << ENDL;
	}




	/** Create a matrix of a given number of rows and columns */ 
	MTLSparseMatrix::MTLSparseMatrix(const unsigned int sizeRows, 
		const unsigned int sizeColumns) : matrix(0) {
			 matrix= new MTLMatrix(sizeRows,sizeColumns);
	}

	MTLSparseMatrix::~MTLSparseMatrix(){
		if (matrix)
			delete matrix;
	}
	void MTLSparseMatrix::resetMatrix(){
	zero_matrix(*matrix);
	}
	void MTLSparseMatrix::addMatrixElement(const unsigned int row,
		const  unsigned int column,const double val){
			(*matrix)(row,column)+=val;
	}
	void MTLSparseMatrix::addSymmetricMatrixElement(const unsigned int row,
		const  unsigned int column,const double val){
			(*matrix)(row,column)+=val;
			if (row!=column)
				(*matrix)(column,row)+=val;
	}


	void MTLSparseMatrix::saveInFile(const std::string filename) const{
		OFSTREAM output((const char *) filename.c_str());
		 int i, j;
		for(i=0;i<matrix->nrows();++i) {
			for(j=0;j<matrix->ncols();++j) {
				output << (*matrix)(i,j) << " ";
			}
			output << ENDL;
		}
	}

	/// set a given value to an entry in the matrix
	void MTLSparseMatrix::setMatrixElement(const unsigned int row,
		const  unsigned int column,const double val){
			(*matrix)(row,column)=val;
	}
	/// set a given value to an entry in the matrix
	void MTLSparseMatrix::setSymmetricMatrixElement(const unsigned int row,
		const  unsigned int column,const double val){
			(*matrix)(row,column)=val;
			if (row!=column)
				(*matrix)(column,row)=val;
	}
		/// get a given value to an entry in the matrix
	double MTLSparseMatrix::getMatrixElement(const unsigned int row,
		const  unsigned int column) const{
			return (*matrix)(row,column);
	}
	/// multiply all matrix elements with a scalar  
	void MTLSparseMatrix::matrixAddMatrix(const GenericSparseMatrix *m){
		mtl::add(*matrix,*(((MTLSparseMatrix*)(m))->matrix));
	}

	/// add multiply all matrix elements with a scalar  
	void MTLSparseMatrix::scalarMultiplyMatrix(const double s){
		mtl::scale(*matrix,s);
	}

	void MTLSparseMatrix::solveIterativeMethod(GenericSparseVector *rightHandSide,GenericSparseVector *result,
			const unsigned int maximumNumberIterations,
			const double tolerance,double &residual)
	{

		mtl::dense1D<double> *res=((MTLSparseVector *) result)->vector;
		mtl::dense1D<double> *rhs=((MTLSparseVector *) rightHandSide)->vector;

		itl::basic_iteration<double> iter(*rhs, maximumNumberIterations, tolerance);

		if (st==GenericSparseMatrix::BICONJUGATED_GRADIENT) {
			if (pt==GenericSparseMatrix::ILU) {
				itl::ILU<MTLMatrix> ilu(*matrix);
				itl::bicgstab((*matrix),*res,*rhs,ilu(),iter);
			}
			else if (pt==GenericSparseMatrix::ILUT) {
				itl::ILUT<MTLMatrix> ilut(*matrix,5,1e-3);
				itl::bicgstab((*matrix),*res,*rhs,ilut(),iter);
			}
			else if (pt==GenericSparseMatrix::SSOR) {
				itl::SSOR<MTLMatrix> ssor(*matrix);
				itl::bicgstab((*matrix),*res,*rhs,ssor(),iter);
			}
			else {
			  std::cerr << "[WARNING] Unknown preconditioner. Resolution skipped\n";
			}
			residual=iter.resid();
		} 
		else if (st==GenericSparseMatrix::GMRES){
			unsigned int restart=10;
			mtl::dense1D<double> b2((*matrix).ncols());
			itl::modified_gram_schmidt<mtl::dense1D<double> > orth(restart, (*res).size());

			if (pt==GenericSparseMatrix::ILU) {
				itl::ILU<MTLMatrix> ilu(*matrix);
			//	ilu.solve(*rhs,b2);
				itl::gmres((*matrix),*res,*rhs,ilu(),restart,iter,orth);
			}					
			else  if (pt==GenericSparseMatrix::ILUT) {
				itl::ILUT<MTLMatrix> ilut(*matrix,5,1e-3);
	//			ilut.solve(*rhs,b2);
				itl::gmres((*matrix),*res,*rhs,ilut(),restart,iter,orth);
			}			  					
			else if (pt==GenericSparseMatrix::SSOR) {
				itl::SSOR<MTLMatrix> ssor(*matrix);
	//			ssor.solve(*rhs,b2);
				itl::gmres((*matrix),*res,*rhs,ssor(),restart,iter,orth);
			}
			else {
			  std::cerr << "[WARNING] Unknown preconditioner. Resolution skipped\n";
			}
			residual=iter.resid();
		}
		else {
		  std::cerr << "[WARNING] Unknown solver type. Resolution skipped\n";
		}
	}

	/** Create a matrix of a given number of rows and columns */ 
	MTLSymmetricSparseMatrix::MTLSymmetricSparseMatrix(const unsigned int sizeRows, 
		const unsigned int sizeColumns) : matrix(0) {
			 matrix= new MTLSymmetricMatrix(sizeRows,sizeColumns);
	}

	MTLSymmetricSparseMatrix::~MTLSymmetricSparseMatrix(){
		if (matrix)
			delete matrix;
	}
	void MTLSymmetricSparseMatrix::resetMatrix(){
	zero_matrix(*matrix);
	}
	void MTLSymmetricSparseMatrix::addMatrixElement(const unsigned int row,
		const  unsigned int column,const double val){
			(*matrix)(row,column)+=val;
	}

	/// set a given value to an entry in the matrix
	void MTLSymmetricSparseMatrix::setMatrixElement(const unsigned int row,
		const  unsigned int column,const double val){
			(*matrix)(row,column)=val;
	}
	/// get a given value to an entry in the matrix
	double MTLSymmetricSparseMatrix::getMatrixElement(const unsigned int row,
		const  unsigned int column) const{
			return (*matrix)(row,column);
	}

	void MTLSymmetricSparseMatrix::saveInFile(const std::string filename) const{
		OFSTREAM output((const char *) filename.c_str());
		unsigned int i, j;
		for(i=0;i<matrix->nrows();++i) {
			for(j=i;j<matrix->ncols();++j) {
				output << (*matrix)(i,j) << " ";
			}
			output << ENDL;
		}
	}
	/// multiply all matrix elements with a scalar  
	void MTLSymmetricSparseMatrix::matrixAddMatrix(const GenericSparseMatrix *m){
		mtl::add(*matrix,*(((MTLSymmetricSparseMatrix*)(m))->matrix));
	}

	/// add multiply all matrix elements with a scalar  
	void MTLSymmetricSparseMatrix::scalarMultiplyMatrix(const double s){
		mtl::scaled(*matrix,s);
	}

	void MTLSymmetricSparseMatrix::solveIterativeMethod(GenericSparseVector *rightHandSide,GenericSparseVector *result,
			const unsigned int maximumNumberIterations,
			const double tolerance,double &residual)
	{
		mtl::dense1D<double> *res=((MTLSparseVector *) result)->vector;
		mtl::dense1D<double> *rhs=((MTLSparseVector *) rightHandSide)->vector;
	//	CERR<<"Matrix=";
		itl::basic_iteration<double> iter(*rhs, maximumNumberIterations, tolerance);
		CERR<<ENDL;
		if (st==GenericSparseMatrix::CONJUGATED_GRADIENT) {
			if (pt==GenericSparseMatrix::CHOLESKY) {
				itl::cholesky<MTLSymmetricMatrix> cholesky(*matrix);
//				mtl::print_all_matrix(*matrix);
				itl::cg((*matrix),*res,*rhs,cholesky(),iter);
			}
			else {
			  std::cerr << "[WARNING] Unknown preconditioner. Resolution skipped\n";
			}
			residual=iter.resid();
		} 
		else if (st==GenericSparseMatrix::GMRES){
			unsigned int restart=10;
			mtl::dense1D<double> b2((*matrix).ncols());
			itl::modified_gram_schmidt<mtl::dense1D<double> > orth(restart, (*res).size());

			if (pt==GenericSparseMatrix::CHOLESKY) {
				itl::cholesky<MTLSymmetricMatrix> cholesky(*matrix);
				//cholesky.solve(*rhs,b2);
				itl::gmres((*matrix),*res,*rhs,cholesky(),restart,iter,orth);
			}
			else {
			  std::cerr << "[WARNING] Unknown preconditioner. Resolution skipped\n";
			}
			residual=iter.resid();
		}
		else {
		  std::cerr << "[WARNING] Unknown solver type. Resolution skipped\n";
		}
	}


} // namespace yav

#endif
