#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#ifndef WIN32
#pragma implementation "linearAlgebra/Rigid3D.h"
#endif
#include "basics/Constants.h"
#include "linearAlgebra/Rigid3D.h"


#ifdef WIN32
#include "basics/export.h"
#endif

namespace yav {



#define TEST_CONSTRAINTS


// value of PI up to the machine precision.
// ----------------------------------------------------------------------
#include <math.h>
// const double M_PI = 4.*atan(1.0);
const double TH_TINY = 1e-5;
const double PI_TH_TINY = M_PI - TH_TINY;
const double TH_SMALL = 1e-1;







/// set rotation part
RigidMat3D& RigidMat3D::operator= (const RotMat3D& rot) throw()
{
  for (Index i=0; i<3; i++)
    for (Index j=0; j<3; j++)
      this->setValue(i,j,rot.getValue(i,j));
  return *this;
}

/// set translation part
RigidMat3D& RigidMat3D::operator= (const Vec3<double>& trans) throw()
{
  for (Index i=0; i<3; i++)
    this->setValue(i,3,trans(i));
  return *this;
}


// Rigid Matrix <- Rigid Vector
// ----------------------------------------------------------------------
void RigidMat3DToRigidVec3D( const RigidMat3D& rig, RigidVec3D& r ) throw()
{
  {
    Matrix3x3<double> m;
    for (Index i=0; i<3; i++) {
      for (Index j=0; j<3; j++) {
        m(i,j)=rig(i,j);
      }
    }
#ifndef WIN32
    RotMat3DToRotVec3D( (const RotMat3D&) m, (RotVec3D&) r);
#else
    RotMat3DToRotVec3D( RotMat3D(m), (RotVec3D&) r);
#endif
  }
  {
    for (Index i=0; i<3; i++) {
      r.setVal(i+3,rig(i,3));
    }
  }
}


// Rigid Vector -> Rigid Matrix
// ----------------------------------------------------------------------
void RigidVec3DToRigidMat3D( const RigidVec3D& r, RigidMat3D& rig) throw()
{
  {
    RotMat3D rot;
    RotVec3DToRotMat3D((const RotVec3D&) r, rot);
    for (Index i=0; i<3; i++)
      for (Index j=0; j<3; j++)
        rig.setValue(i,j,rot.getValue(i,j));
  }
  {
    for (Index i=0; i<3; i++)
      rig.setValue(i,3,r(i+3));
  }
  {
    for (Index i=0; i<4; i++)
      rig.setValue(3,i,(i==3));
  }
}








// **********************************************************************
// Operations specific to 3D rotations
// **********************************************************************

RigidMat3D& RigidMat3D::LSQ(const Matrix3x3<double>& K, const Vec3<double>& ModelMean, const Vec3<double>& SceneMean)
  throw(NonUniqueOptimum,RigidMat3D::ErrNonRigidMat,ErrConvergence)
{
  RotMat3D rot;
  rot.LSQ(K);
  Index i;
  Vec3<double> trans=ModelMean-rot*SceneMean;
  for (i=0; i<3; i++) {
    for (Index j=0; j<3; j++)
      this->setValue(i,j,rot.getValue(i,j));
    this->setValue(i,3,trans(i));
  }
  for (i=0; i<4; i++) {
    this->setValue(3,i,(i==3));
  }
  return *this;
}

void RigidMat3D::inv( const RigidMat3D& R) throw()
{
  if (&R==this) {RigidMat3D tmp(R);this->inv(tmp);return;}
  Index i,j;
  // On calcule l'inverse de la rotation
  for (i=0; i<3; i++) for (j=0; j<3; j++) {val[i*4+j]=R.val[j*4+i];}
  // On calcule la nouvelle translation
  val[3] =-(val[0]*R.val[3]+val[1]*R.val[7]+val[2]*R.val[11]);
  val[7] =-(val[4]*R.val[3]+val[5]*R.val[7]+val[6]*R.val[11]);
  val[11]=-(val[8]*R.val[3]+val[9]*R.val[7]+val[10]*R.val[11]);
}

//  LSQRigid : input matrix K, output the rotation matrix R maximizing
//              C = trace( R. K\trp ) (K and R are (n,n) matrices).
//           residue is the value at the maximum
//----------------------------------------------------------------------
double RigidMat3D::LSQ(const Matrix4x4<double>& K) throw(NonUniqueOptimum,RigidMat3D::ErrNonRigidMat,ErrConvergence)
{
  int i,j;
  Matrix3x3<double> Kr;

  for (i=0; i<3; i++)
    for (j=0; j<3; j++)
      Kr(i,j)=K(i,j);

  RotMat3D rot;
  double res=rot.LSQ(Kr);
  Vec3<double> trans(K(0,3),K(1,3),K(2,3));

  for (i=0; i<3; i++) {
    for (j=0; j<3; j++)
      val[i*4+j]=rot(i,j);
    val[i*4+j]=trans(i);
  }

  //val[12]=val[13]=val[14]=0;val[15]=1;Commented because already set

 return res;
}

#ifdef TODO
RigidMat3D& RigidMat3D::Rnd(void) throw()
{
  UnitQuat q;
  *this=q.Rnd();

  return *this;
};

// Intrinsic Distance between rotation matrices : angle( R1trp.R2)
// ----------------------------------------------------------------------
double Dist( const RigidMat3D& R1, const RigidMat3D& R2) throw()
{
  RigidMat3D R;
  R.Trp(R1);
  R.Mul( R, R2);
  double tmp = (R.Trace() -1.0)*0.5;
  if (tmp > 1.) tmp = 1.;
  if (tmp < -1.) tmp = -1.;
  return acos( tmp);

  return 0;
}
#endif



void RigidMat3D::readO8(const char* FileName) throw(ErrUnknown)
{
  try {
    IFSTREAM fin(FileName);
    readO8(fin);
  }
  catch (...) {
    throw(ErrUnknown("RigidMat3D::readO8"));
  }
}
void RigidMat3D::readO8(IFSTREAM& fin) throw(ErrUnknown)
{
  try {
    Matrix4x4<double> m;
    char Buffer[255];
    fin >> Buffer;
    fin >> Buffer;
    for(int i=0 ; i<4;i++) {
      for(int j=0 ; j<4;j++) {
        double e;
        fin >> e;
        m(i, j)=e;
      }
    }
    fin >> Buffer;
    *this=m;
  }
  catch (...) {
    throw(ErrUnknown("RigidMat3D::readO8"));
  }
}
void RigidMat3D::writeO8(const char* FileName) const throw(ErrUnknown)
{
  try {
    OFSTREAM fout(FileName);
    writeO8(fout);
  }
  catch (...) {
    throw(ErrUnknown("RigidMat3D::readO8"));
  }
}
void RigidMat3D::writeO8(OFSTREAM& fout) const throw(ErrUnknown)
{
  try {
    fout << "(" << ENDL;
    fout << "O8" << ENDL;
    for(int i=0 ; i<4;i++) {
      for(int j=0 ; j<4;j++)
        fout << (*this)(i, j) << " ";
      fout << ENDL;
    }
    fout << ")" << ENDL;
  }
  catch (...) {
    throw(ErrUnknown("RigidMat3D::readO8"));
  }
}




// Intrinsic Distance between rotation vectors : angle( r1\inv o r2)
// ----------------------------------------------------------------------
#ifdef TODO
double Dist( const RigidVec3D& r1, const RigidVec3D& r2) throw()
{
  UnitQuat q1(r1), q2(r2);
  QuatInv(q1, q1);
  QuatProd(q1,q2,q1);
  RigidVec3D r(q1);
  return( r.Norm() );

  return 0;
}

RigidVec3D& RigidVec3D::Rnd(void) throw()
{
  UnitQuat q;
  *this=q.Rnd();

  return *this;
}
#endif

// Compose rotation vectors
// ----------------------------------------------------------------------
void RigidVec3D::comp( const RigidVec3D& r1, const RigidVec3D& r2)  throw()
{
  *this=RigidMat3D(r1)*RigidMat3D(r2);
}

// Inv rigid vectors
// ----------------------------------------------------------------------
void RigidVec3D::inv( const RigidVec3D& r)  throw()
{
  this->Mul(-1.,r);
  this->TransPart()=RotMat3D(this->RotPart())*this->TransPart();
}

// **********************************************************************
void RigidVec3D::J_Rt_r(const Vec3<double>& x, Mat<3,6,double>& dRx_dr) const throw()
{
  Matrix3x3<double> dRx_dR;
  ((const RotVec3D*)this)->J_Rt_r(x,dRx_dR);
  for (Index i=0; i<3; i++) {
    for (Index j=0; j<3; j++) {
      dRx_dr(i,j)=dRx_dR(i,j);
      dRx_dr(i,j+3)=(i==j);
    }
  }
}


void RigidVec3D::readV6(const char* FileName) throw(ErrUnknown)
{
  try {
    IFSTREAM fin(FileName);
    readV6(fin);
  }
  catch (...) {
    throw(ErrUnknown("RigidVec3D::readV6"));
  }
}
void RigidVec3D::readV6(IFSTREAM& fin) throw(ErrUnknown)
{
  try {
    char buffer[1000];
    fin.getline(buffer,1000);
    Vect<6,double> v;
    if (sscanf(buffer,"%lf%lf%lf%lf%lf%lf",&v[0],&v[1],&v[2],&v[3],&v[4],&v[5])!=6)
      throw(ErrUnknown("RigidVec3D::readV6"));;
    *this=v;
  }
  catch (...) {
    throw(ErrUnknown("RigidVec3D::readV6"));
  }
}
void RigidVec3D::writeV6(const char* FileName) const throw(ErrUnknown)
{
  try {
    OFSTREAM fout(FileName);
    writeV6(fout);
  }
  catch (...) {
    throw(ErrUnknown("RigidVec3D::readV6"));
  }
}
void RigidVec3D::writeV6(OFSTREAM& fout) const throw(ErrUnknown)
{
  try {
    for(int i=0 ; i<6; i++) {
      fout << (*this)(i) << " ";
    }
    fout << ENDL;
  }
  catch (...) {
    throw(ErrUnknown("RigidVec3D::readV6"));
  }
}





} // namespace
