#ifndef _MATRIX2X2_H
#define _MATRIX2X2_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#ifndef WIN32
#pragma interface "linearAlgebra/Matrix2x2.h"
#endif
#include "linearAlgebra/Vec.h"
#include "linearAlgebra/Mat.h"
#include "linearAlgebra/Vec2.h"

#ifdef WIN32
#include "basics/export.h"
#endif

namespace yav {

class Rotation2D;

/** @name Vectorial redefined friend functions */
//@{
/** unary +
    @param p vector
    @return +p */
template < class T> vec_Inline Matrix2x2< T> operator+(const Matrix2x2< T> &p) throw()
  {return p;}
/** unary -
    @param p vector
    @return -p */
template < class T> vec_Inline Matrix2x2< T> operator-(const Matrix2x2< T> &p) throw()
  {return Matrix2x2< T>(VIM_NOINIT).Minus(p);}
/** vector sum
    @param p first vector
    @param q second vector
    @return p + q */
template < class T> vec_Inline Matrix2x2< T> operator+(const Matrix2x2< T> &p, const Matrix2x2 < T> &q) throw()
  {return Matrix2x2< T>(VIM_NOINIT).Add(p,q);}
/** vector substraction
    @param p first vector
    @param q second vector
    @return p - q */
template < class T> vec_Inline Matrix2x2< T> operator-(const Matrix2x2< T> &p, const Matrix2x2 < T> &q) throw()
  {return Matrix2x2< T>(VIM_NOINIT).Sub(p,q);}
/** vector multiplication by scalar
    @param p vector
    @param a scalar
    @return ap */
template < class T> vec_Inline Matrix2x2< T> operator*(const Matrix2x2< T> &p, const T a) throw()
  {return Matrix2x2< T>(VIM_NOINIT).Mul(p,a);}
/** vector scalar multiplication
    @param a scalar
    @param p vector
    @return ap */
template < class T> vec_Inline Matrix2x2< T> operator*(const T a, const Matrix2x2< T> &p) throw()
  {return Matrix2x2< T>(VIM_NOINIT).Mul(p,a);}
/** vector scalar division
    @param p vector
    @param a scalar
    @return p / a */
template < class T> vec_Inline Matrix2x2< T> operator/(const Matrix2x2< T> &p, const T a) throw(ErrDivZero)
  {return Matrix2x2< T>(VIM_NOINIT).Div(p,a);}
/** dot product
    @param u first vector
    @param v second vector
    @return u.v */
template < class T> vec_Inline T dot(const Matrix2x2< T>& u, const Matrix2x2< T> &v) throw()
  {return Matrix2x2< T>::dot(u,v);}
/// Square Euclidian Distance
template < class T> vec_Inline T dist2( const Matrix2x2< T>& u, const Matrix2x2< T>& v) throw()
  {return Matrix2x2< T>::dist2(u,v);}
/// Euclidian Distance
template < class T> vec_Inline T dist( const Matrix2x2< T>& u, const Matrix2x2< T>& v) throw()
  {return Matrix2x2< T>::dist(u,v);}
/// L1 Distance
template < class T> vec_Inline T distL1( const Matrix2x2< T>& u, const Matrix2x2< T>& v) throw()
  {return Matrix2x2< T>::distL1(u,v);}
/// Linf Distance
template < class T> vec_Inline T distLinf( const Matrix2x2< T>& u, const Matrix2x2< T>& v) throw()
  {return Matrix2x2< T>::distLinf(u,v);}
/** dot product
    @param p first vector
    @param q second vector
    @return p.q */
template < class T> vec_Inline T DotProd(const Matrix2x2< T> &p, const Matrix2x2< T> &q) throw()
  {return Matrix2x2< T>::dot(p,q);}
/// Square Euclidian Distance
template < class T> vec_Inline T Dist2( const Matrix2x2< T>& u, const Matrix2x2< T>& v) throw()
  {return Matrix2x2< T>::dist2(u,v);}
/// Euclidian Distance
template < class T> vec_Inline T Dist( const Matrix2x2< T>& u, const Matrix2x2< T>& v) throw()
  {return Matrix2x2< T>::dist(u,v);}
/// L1 Distance
template < class T> vec_Inline T DistL1( const Matrix2x2< T>& u, const Matrix2x2< T>& v) throw()
  {return Matrix2x2< T>::distL1(u,v);}
/// Linf Distance
template < class T> vec_Inline T DistLinf( const Matrix2x2< T>& u, const Matrix2x2< T>& v) throw()
  {return Matrix2x2< T>::distLinf(u,v);}
/** operator == */
template < class T> vec_Inline bool operator == (const Matrix2x2< T>& u, const Matrix2x2< T> &v) throw()
  {return Matrix2x2< T>::isEqual(u,v);}
/** operator != */
template < class T> vec_Inline bool operator != (const Matrix2x2< T>& u, const Matrix2x2< T> &v) throw()
  {return Matrix2x2< T>::isDifferent(u,v);}
/** operator < */
template < class T> vec_Inline bool operator < (const Matrix2x2< T>& u, const Matrix2x2< T> &v) throw()
  {return Matrix2x2< T>::isSmaller(u,v);}
/** operator > */
template < class T> vec_Inline bool operator > (const Matrix2x2< T>& u, const Matrix2x2< T> &v) throw()
  {return Matrix2x2< T>::isGreater(u,v);}
/** operator <= */
template < class T> vec_Inline bool operator <= (const Matrix2x2< T>& u, const Matrix2x2< T> &v) throw()
  {return Matrix2x2< T>::isSmallerOrEqual(u,v);}
/** operator >= */
template < class T> vec_Inline bool operator >= (const Matrix2x2< T>& u, const Matrix2x2< T> &v) throw()
  {return Matrix2x2< T>::isGreaterOrEqual(u,v);}
//@}
/** @name Square matrix redefined friend functions */
//@{
template < class T> vec_Inline Matrix2x2< T> operator*(const Matrix2x2< T> &m1, const Matrix2x2< T> &m2) throw()
  {return Matrix2x2< T>(VIM_NOINIT).Mul(m1,m2);}
//@}

/*! \class Matrix2x2 Matrix2x2.h linearAlgebra/Matrix2x2.h
  \ingroup liblinearAlgebra
  template class for a square matrix of dimension 2 and type T 
*/
template <class T> class Matrix2x2 : public Matrix<2,T> {
public:
  /** @name Constructors */
  //@{
  /** @name Vectorial redefined constructors */
  //@{
  /** default constructor - initializes to 0 if UNINIT_ZERO not defined */
  vec_Inline Matrix2x2() throw() : Matrix<2,T>()
    {}
  /** constructor - initializes to 0 if b==VIM_INIT */
  explicit vec_Inline Matrix2x2(VectInitMode b) throw() : Matrix<2,T>(b)
    {}
  //@}
  /** @name Square matrix redefined constructors */
  //@{
  /// Set to diag( v )
  explicit vec_Inline Matrix2x2(const Vect<2,T>& v) throw() : Matrix<2,T>(VIM_NOINIT)
    {diag(v);}
  
#ifndef WIN32
  /** copy constructor
      @param v vector to copy */
  vec_Inline Matrix2x2(const Mat<2,2,T>& v) throw() : Matrix<2,T>(VIM_NOINIT)
    {copy(v);}
  /** operator =
      @param v vector to copy
      @return this */
  vec_Inline Matrix2x2< T>& operator=(const Mat<2,2,T>& v) throw()
    {copy(v);return *this;}
#endif
  /** copy constructor with cast
      @param v vector to copy */
  template<class T2> vec_Inline Matrix2x2(const Mat<2,2,T2>& v) throw() : Matrix<2,T>(VIM_NOINIT)
    {copy(v);}
  /** operator = with cast
      @param v vector to copy
      @return this */
  template<class T2> vec_Inline Matrix2x2< T>& operator=(const Mat<2,2,T2>& v) throw()
    {copy(v);return *this;}
  
#if (!defined( WIN32) || (defined(__CYGWIN__)))
  /// Constructor : set to matrix coord
  vec_Inline Matrix2x2(const T coord[2][2]) throw() : Matrix<2,T>(VIM_NOINIT)
    {setValues(coord);}
#endif
  /// Constructor : set to matrix coord
  template <class T2> vec_Inline Matrix2x2(const T2 coord[2][2]) throw() : Matrix<2,T>(VIM_NOINIT)
    {setValues(coord);}
  
  /** Creates a square matrix from a symetric matrix
      @param W symetric matrix */
  explicit vec_Inline Matrix2x2(const Sym<2,T>& W) throw() : Matrix<2,T>(VIM_NOINIT)
    {Sym2Mat<2,T>(W,*this);}
  /** Set from a symetric matrix
      @param W symetric matrix */
  vec_Inline Matrix2x2< T>& operator=(const Sym<2,T>& W) throw()
    {Sym2Mat<2,T>(W,*this);return *this;}
  //@}

  /** builds the matrix (x_ij)
      @param x00 first line, first column element
      @param x01 first line, second column element
      @param x10 second line, first column element
      @param x11 second line, second column element */
  Matrix2x2(T x00, T x01,T x10, T x11) throw();

  /** creates a 2x2 matrix from 2 column vectors
      @param v1 first column vector
      @param v2 second column vector */
  Matrix2x2(const Vect<2,T> &v1, const Vect<2,T> &v2) throw();

  /** @name Data members access */
  //@{
  /// setValues
  void setValues(T x00, T x01,
                 T x10, T x11) throw();

  /// setValues with a T[L][C] array
  inline void setValues(const T v[2][2]) throw()
    {((Matrix<2,T>&)*this).setValues(v);}

  /** access ithline vector
      @param i line number
      @return line vector */
  Vec2<T> operator[](int i) const throw(ErrRange);
  //@}

  /** @name mathematical operators */
  //@{
  /** @name Vectorial redefined menber functions */
  //@{
  /** set this vector to p+q
      @return this = p + q */
  vec_Inline Matrix2x2< T>& Minus(const Matrix2x2< T> &p) throw()
    {vec_opposite(this->val,4,p.val);return *this;}
  /** set this vector to p+q
      @return this = p + q */
  vec_Inline Matrix2x2< T>& Add(const Matrix2x2< T> &p, const Matrix2x2< T> &q) throw()
    {vec_add(this->val,4,p.val,q.val);return *this;}
  /** set this vector to p-q
      @return this = p - q */
  vec_Inline Matrix2x2< T>& Sub(const Matrix2x2< T> &p, const Matrix2x2< T> &q) throw()
    {vec_sub(this->val,4,p.val,q.val);return *this;}
  /** set each coordinate of this vector to product of this and p same coordinate
      @return this = this.*q */
  vec_Inline Matrix2x2< T>& CMul(const Matrix2x2< T> &q) throw()
    {vec_cmul(this->val,4,q.val);return *this;}
  /** set each coordinate of this vector to product of q and p same coordinate
      @return this = p.*q */
  vec_Inline Matrix2x2< T>& CMul(const Matrix2x2< T> &p, const Matrix2x2< T> &q) throw()
    {vec_cmul(this->val,4,p.val,q.val);return *this;}
  /** set this vector to p*a
      @return this = p*a */
  vec_Inline Matrix2x2< T>& Mul(const Matrix2x2< T> &p, const T a) throw()
    {vec_mul(this->val,4,p.val,a);return *this;}
  /** set this vector to p*a
      @return this = p*a */
  vec_Inline Matrix2x2< T>& Mul(const T a, const Matrix2x2< T> &p) throw()
    {vec_mul(this->val,4,p.val,a);return *this;}
  /** set this vector to p/a
      @return this = p/a */
  vec_Inline Matrix2x2< T>& Div(const Matrix2x2< T> &p, const T a) throw(ErrDivZero)
    {vec_div(this->val,4,p.val,a);return *this;}
  /** adds vector q to vector p
      @param q vector to add
      @return p + q */
  vec_Inline Matrix2x2< T>& operator+=(const Matrix2x2< T> &q) throw()
    {vec_add(this->val,4,q.val);return *this;}
  /** substracts vector q from vector p
      @param q vector to substract
      @return p - q */
  vec_Inline Matrix2x2< T>& operator-=(const Matrix2x2< T> &q) throw()
    {vec_sub(this->val,4,q.val);return *this;}
  /** multiplies vector p by scalar a
      @param a scalar
      @return ap */
  vec_Inline Matrix2x2< T>& operator*=(const T a) throw()
    {vec_mul(this->val,4,a);return *this;}
  /** divides vector p by scalar a
      @param a scalar
      @return p / a */
  vec_Inline Matrix2x2< T>& operator/=(const T a) throw(ErrDivZero)
    {vec_div(this->val,4,a);return *this;}
  /** randomize vector
      @return this */
  vec_Inline Matrix2x2< T>& Rnd(T min = -1., T max = 1.) throw()
    {vec_rnd(this->val,4,min,max);return *this;}
  /** randomize vector : isotropic gaussian centred noise
      @return this */
  vec_Inline Matrix2x2< T>& RndGaussian(T standard_deviation = 1.) throw()
    {vec_rndgaussian(this->val,4,standard_deviation);return *this;}
  /** takes the X,Y or Z value of p if they are less than its own X,Y,Z coordinates
      @return _Vect<4,T>(min(x,p.x),min(y,p.y),...) */
  vec_Inline Matrix2x2< T>& Min(const Matrix2x2< T>& p) throw()
    {vec_min(this->val,4,p.val);return *this;}
  /** takes the X,Y or Z value of p if they are greater than its own X,Y,Z coordinates
      @return _Vect<4,T>(max(x,p.x),max(y,p.y),...) */
  vec_Inline Matrix2x2< T>& Max(const Matrix2x2< T>& p) throw()
    {vec_max(this->val,4,p.val);return *this;}
  /** set each element of this vector to value
      @param v value
      @return this */
  vec_Inline Matrix2x2< T>& Reset(T v) throw()
    {resetTo(v);return *this;}
  /** set this vector to null vector
      @return this */
  vec_Inline Matrix2x2< T>& Null() throw()
  {resetTo((T)0);return *this;}
  /** set this vector to mean of a given list of vectors
      @return this */
  template<class InputIterator>
  vec_Inline Matrix2x2< T>& Mean(InputIterator first, InputIterator last) throw()
    {mean(first,last);return *this;}
  //@}

  /** set this matrix to inverse of a given matrix
    @return this */
  void inv(const Matrix2x2<T> &m) throw(ErrMatSingular);
#ifdef TODO
invert rend matrice nulle en cas de singularite
#endif

  /** computes the determninant
      @return det matrix determinant */
  T det(void) const throw();

  /** @name Square matrix redefined menber functions */
  //@{
  /** set this matrix to inverse of a given matrix
      @return this */
  vec_Inline Matrix2x2< T>& Inv(const Matrix2x2< T>& m) throw(ErrMatSingular)
    {this->inv(m);return *this;}
  /** get this matrix inverse matrix (this matrix is NOT modifyed)
      @return inverse matrix */
  vec_Inline Matrix2x2< T> getInverse(void) const throw(ErrMatSingular)
    {return Matrix2x2< T>(VIM_NOINIT).Inv(*this);}
  /// invert this matrix (this matrix is modified)
  vec_Inline void invert(void) throw(ErrMatSingular)
    {this->Inv(*this);}
  
  /** set this matrix to transposed of a given matrix
      @return this */
  vec_Inline Matrix2x2< T>& Trp(const Matrix2x2< T>& m) throw()
    {this->trp(m);return *this;}
  /** get this matrix transposed matrix (this matrix is NOT modifyed)
      @return transpose matrix */
  vec_Inline Matrix2x2< T> getTransposed(void) const throw()
    {return Matrix2x2< T>(VIM_NOINIT).Trp(*this);}
  /// transpose this matrix (this matrix is modified)
  vec_Inline void transpose(void) throw()
    {this->Trp(*this);}
  
  /// Matrix multiplication
  vec_Inline Matrix2x2< T>& operator*=(const Matrix2x2< T>& m) throw()
    {mulmat(*this,m);return *this;}
  /// Matrix multiplication
  Matrix2x2< T>& Mul(const Matrix2x2< T> &m1, const Matrix2x2< T> &m2) throw()
    {mulmat(m1,m2);return *this;}
  
  /** computes the determninant
      @return det matrix determinant */
  vec_Inline T Det(void) const throw()
    {return this->det();}
  /** computes the trace
      @return trace matrix trace */
  vec_Inline T Trace(void) const throw()
    {return this->trace();}
  
  /// set this matrix to the identity matrix
  vec_Inline Matrix2x2< T>& Id(const T v=1.0) throw()
    {this->resetToIdentity(v);return *this;}
  /// Set to diag( v )
  vec_Inline Matrix2x2< T>& Diag(const Vect<2,T>& v) throw()
    {this->diag(v);return *this;}
  /// Set to tensor product v1*v2
  vec_Inline Matrix2x2< T>& TensorProduct(const Vect<2,T>& v1,const Vect<2,T>& v2) throw()
    {this->tensorProduct(v1,v2);return *this;}
  //@}

  static void Test() throw();
};


} // namespace yav


/*][** NODOC ****************************************************************/
/****************************************************************************/

/*]]**************************************************************************/

namespace yav {

// constructors
template <class T> vec_Inline
Matrix2x2<T>::Matrix2x2(const Vect<2,T> &v1, const Vect<2,T> &v2) throw()
  : Matrix<2,T>(VIM_NOINIT)
{
  this->val[0]=v1[0];this->val[1]=v2[0];
  this->val[2]=v1[1];this->val[3]=v2[1];
}

template <class T> vec_Inline
Matrix2x2<T>::Matrix2x2(T x00, T x01,T x10, T x11) throw()
  : Matrix<2,T>(VIM_NOINIT)
{
  this->val[0]=x00;this->val[1]=x01;
  this->val[2]=x10;this->val[3]=x11;
}

template <class T> vec_Inline
void Matrix2x2<T>::setValues(T x00, T x01,T x10, T x11) throw()
{
  this->val[0]=x00;this->val[1]=x01;
  this->val[2]=x10;this->val[3]=x11;
}


// access to data
template <class T> vec_Inline
Vec2<T> Matrix2x2<T>::operator[](int i) const throw(ErrRange)
{
#ifdef RANGE_CHECK
  if ( (i<0) || (i >= 2) )
    throw  ErrRange("Mat::getValue(Index,Index)");
#endif
  return Vec2<T>(this->val+i*2);
}

// mathematicl operations
template <class T> vec_Inline
void Matrix2x2<T>::inv(const Matrix2x2<T> &m) throw(ErrMatSingular)
{
  if ((&m)==this) {
    Matrix2x2<T> M(m);
    this->inv(M);
  } else {
    this->val[0]=m.val[3];
    this->val[1]=-m.val[1];
    this->val[2]=-m.val[2];
    this->val[3]=m.val[0];
    try{
      *this/=m.Det();
    }
    catch( ErrDivZero& exc ){
      throw ErrMatSingular("Matrix2x2::Inv(const Matrix2x2 & )");
    }
  }
}


template <class T> vec_Inline
T Matrix2x2<T>::det() const throw()
{
  return this->val[0]*this->val[3]-this->val[1]*this->val[2];
}


template <class T>
void Matrix2x2<T>::Test() throw()
{
  float Af[2][2];double Ad[2][2];int Ai[2][2];
  T A[2][2];for (Index k=0; k<2; k++) for (Index l=0; l<2; l++) {A[k][l]=k*2+l+1;Af[k][l]=k*2+l+1;Ad[k][l]=k*2+l+1;Ai[k][l]=k*2+l+1;}
  Matrix2x2<float> df;Matrix2x2<double> dd;Matrix2x2<int> di;
  Vec2<T> bb(0);

  Matrix2x2<T> a(A),aaa(1,2,3,4),aab(bb,bb),b(bb),c(VIM_INIT),d(a),/*af(Af),ad(Ad),ai(Ai),*/ef(df),ed(dd),ei(di);
  T v;char TName[100];Xlinalg_Class_Name(v,TName);
  COUT << "Matrix2x2<"<<TName<<">" << ENDL;
  COUT << "Constructors" << ENDL;
  COUT << "a must be 1,2,... : " << a << ENDL;
  COUT << "aaa must be 1,2,... : " << aaa << ENDL;
  COUT << "aab must be 1,0,1,0 : " << aab << ENDL;
//   COUT << "af must be 1,2,... : " << af << ENDL;
//   COUT << "ad must be 1,2,... : " << ad << ENDL;
//   COUT << "ai must be 1,2,... : " << ai << ENDL;
  COUT << "b must be diag(1,0,...) : " << b << ENDL;
  COUT << "c must be 0,0,... : " << c << ENDL;
  COUT << "d must be as a (copy) : " << d << ENDL;
  COUT << "ef must be 0 : " << ef << ENDL;
  COUT << "ed must be 0 : " << ed << ENDL;
  COUT << "ei must be 0 : " << ei << ENDL;

  COUT << "operator=" << ENDL;
  d=c;COUT << "d=c : " << d << ENDL;
  d=df;COUT << "d=df : " << d << ENDL;
  d=dd;COUT << "d=dd : " << d << ENDL;
  d=di;COUT << "d=di : " << d << ENDL;

  COUT << "Access to Data" << ENDL;
  d.copy(b);COUT << "d.copy(b)  : " << d << ENDL;
  COUT << "d[0] : " << d[0] << ENDL;
  d[0][0]=25;COUT << "d[0][0]=25;  d[0][0] : " << d[0][0] << ENDL;
  d(0,0)=26;COUT << "d(0,0)=26;  d(0,0) : " << d(0,0) << ENDL;
  d.setValue(0,0,11);COUT << "d.setValue(0,0,11);  d.getValue(0,0) : " << d.getValue(0,0) << ENDL;
  d.Null();COUT << "d.Null() : " << d << ENDL;
  d.resetTo(2);COUT << "d.resetTo(2) : " << d << ENDL;
  T*Dd=d.getValues();
  COUT << "getValues() : " << Dd[0] << " " << Dd[1] << " " << Dd[2] << ENDL;
  const T*E=d;
  COUT << "(const T*) d : " << E[0] << " " << E[1] << " " << E[2] << ENDL;
  d.setValues(A);COUT << "d.setValues(T[1,2,3...]) : " << d << ENDL;
//   d.setValues(Af);COUT << "d.setValues(float[1,2,3...]) : " << d << ENDL;
//   d.setValues(Ad);COUT << "d.setValues(double[1,2,3...]) : " << d << ENDL;
//   d.setValues(Ai);COUT << "d.setValues(int[1,2,3...]) : " << d << ENDL;

  COUT << "Math" << ENDL;
  COUT << "d.Minus(a) : " << d.Minus(a) << ENDL;
  COUT << "d.Add(a,b) : " << d.Add(a,b) << ENDL;
  COUT << "d.Sub(a,b) : " << d.Sub(a,b) << ENDL;
  COUT << "d.Mul(a,2) : " << d.Mul(a,(T)2) << ENDL;
  COUT << "d.Mul(2,a) : " << d.Mul((T)2,a) << ENDL;
  COUT << "d.Div(a,2) : " << d.Div(a,(T)2) << ENDL;
  COUT << "d.Div(a,2) : " << d.Div(a,(T)2) << ENDL;

  COUT << "+a : " << +a << ENDL;
  COUT << "-a : " << -a << ENDL;
  COUT << "a+b : " << a+b << ENDL;
  COUT << "a-b : " << a-b << ENDL;
  COUT << "a*2 : " << a*((T)2) << ENDL;
  COUT << "2*a : " << ((T)2)*a << ENDL;
  COUT << "a/2 : " << a/((T)2) << ENDL;

  d.copy(b);d+=a;COUT << "d=b; d+=a : " << d << ENDL;
  d.copy(b);d-=a;COUT << "d=b; d-=a : " << d << ENDL;
  d.copy(b);d*=(T)2;COUT << "d=b; d*=2 : " << d << ENDL;
  d.copy(b);d/=(T)2;COUT << "d=b; d/=2 : " << d << ENDL;

  d.copy(b);COUT << "d=b; d==b : " << (d==b) << ENDL;
  d.copy(b);COUT << "d=b; d!=b : " << (d!=b) << ENDL;

  COUT << "dot(a,b) : " << dot(a,b) << ENDL;
  COUT << "DotProd(a,b) : " << DotProd(a,b) << ENDL;
  COUT << "norm2(a) : " << a.norm2() << ENDL;
  COUT << "norm(a) : " << a.norm() << ENDL;
  COUT << "Norm2(a) : " << a.Norm2() << ENDL;
  COUT << "Norm(a) : " << a.Norm() << ENDL;
  COUT << "NormL1(a) : " << a.NormL1() << ENDL;
  COUT << "NormLinf(a) : " << a.NormLinf() << ENDL;
  COUT << "Dist2(a,b) : " << Dist2(a,b) << ENDL;
  COUT << "Dist(a,b) : " << Dist(a,b) << ENDL;
  COUT << "DistL1(a,b) : " << DistL1(a,b) << ENDL;
  COUT << "DistLinf(a,b) : " << DistLinf(a,b) << ENDL;

  d.resetTo(2); d.Min(a); COUT << "d.resetTo(2); d.Min(a) : " << d << ENDL;
  d.resetTo(2); d.Max(a); COUT << "d.resetTo(2); d.Max(a) : " << d << ENDL;

  d.Rnd(2); COUT << "d.Rnd(2) : " << d << ENDL;

  Matrix2x2<T> m;COUT << "m.Trp(d) : " << m.Trp(d) << ENDL;
  d.transpose();COUT << "d.transpose() : " << d << ENDL;
  COUT << "d.getTransposed() : " << d.getTransposed() << ENDL;

  {Matrix2x2<T> m;Matrix2x2<T> n;COUT << "Matrix2x2<T>*Matrix2x2<T> : " << m*n << "  Mul : " << d.Mul(m,n) << ENDL;}

  {d.Rnd();Sym<2,T> S1;Sym<2,T> S2;d.SymAction(S1,S2);COUT << "SymAction" << ENDL;}
  {d.Rnd();Matrix2x2<T> S1;Vec2<T>W;Matrix2x2<T> S2;d.SVD(S1,W,S2);COUT << "SVD" << ENDL;}
  {d.Rnd();Matrix2x2<T> S;d.PseudoInv(S);COUT << "PseudoInv" << ENDL;}

  {d.Rnd();d.Inv(d);COUT << "Inv" << ENDL;}
  {d.Rnd();COUT << "getInverse" << d.getInverse() << ENDL;}
  {d.Rnd();d.invert();COUT << "invert" << ENDL;}

  {d.Rnd();COUT << "Det" << d.Det() << ENDL;}
  {d.Rnd();COUT << "Trace" << d.Trace() << ENDL;}

  {d.Rnd();COUT << "Id" << d.Id() << ENDL;}
  {d.Rnd();Vec2<T>v;COUT << "Diag" << d.Diag(v) << ENDL;}
  {d.Rnd();Vec2<T>v;COUT << "TensorProduct" << d.TensorProduct(v,v) << ENDL;}

  {Sym<2,T> m;m.Rnd();d=m;COUT << "operator=Sym<2,T>" << d << "=" << m << ENDL;}
  {Sym<2,T> m;m.Rnd();Matrix2x2<T> d(m);COUT << "Constructeur(Sym<2,T>)" << d << "=" << m << ENDL;}

  COUT << ENDL;
}


} // namespace yav


#endif
