// General comments : see Vec.h

#ifndef _MATRIX3X3_H
#define _MATRIX3X3_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#ifndef WIN32
#pragma interface "linearAlgebra/Matrix3x3.h"
#endif
#include "linearAlgebra/Vec.h"
#include "linearAlgebra/Mat.h"
#include "linearAlgebra/Vec3.h"

#ifdef WIN32
#include "basics/export.h"
#endif

namespace yav {


class RotMat3D;
typedef RotMat3D Rotation3D;


/** @name Vectorial redefined friend functions */
//@{
/** unary +
    @param p vector
    @return +p */
template < class T> vec_Inline Matrix3x3< T> operator+(const Matrix3x3< T> &p) throw()
  {return p;}
/** unary -
    @param p vector
    @return -p */
template < class T> vec_Inline Matrix3x3< T> operator-(const Matrix3x3< T> &p) throw()
  {return Matrix3x3< T>(VIM_NOINIT).Minus(p);}
/** vector sum
    @param p first vector
    @param q second vector
    @return p + q */
template < class T> vec_Inline Matrix3x3< T> operator+(const Matrix3x3< T> &p, const Matrix3x3 < T> &q) throw()
  {return Matrix3x3< T>(VIM_NOINIT).Add(p,q);}
/** vector substraction
    @param p first vector
    @param q second vector
    @return p - q */
template < class T> vec_Inline Matrix3x3< T> operator-(const Matrix3x3< T> &p, const Matrix3x3 < T> &q) throw()
  {return Matrix3x3< T>(VIM_NOINIT).Sub(p,q);}
/** vector multiplication by scalar
    @param p vector
    @param a scalar
    @return ap */
template < class T> vec_Inline Matrix3x3< T> operator*(const Matrix3x3< T> &p, const T a) throw()
  {return Matrix3x3< T>(VIM_NOINIT).Mul(p,a);}
/** vector scalar multiplication
    @param a scalar
    @param p vector
    @return ap */
template < class T> vec_Inline Matrix3x3< T> operator*(const T a, const Matrix3x3< T> &p) throw()
  {return Matrix3x3< T>(VIM_NOINIT).Mul(p,a);}
/** vector scalar division
    @param p vector
    @param a scalar
    @return p / a */
template < class T> vec_Inline Matrix3x3< T> operator/(const Matrix3x3< T> &p, const T a) throw(ErrDivZero)
  {return Matrix3x3< T>(VIM_NOINIT).Div(p,a);}
/** dot product
    @param u first vector
    @param v second vector
    @return u.v */
template < class T> vec_Inline T dot(const Matrix3x3< T>& u, const Matrix3x3< T> &v) throw()
  {return Matrix3x3< T>::dot(u,v);}
/// Square Euclidian Distance
template < class T> vec_Inline T dist2( const Matrix3x3< T>& u, const Matrix3x3< T>& v) throw()
  {return Matrix3x3< T>::dist2(u,v);}
/// Euclidian Distance
template < class T> vec_Inline T dist( const Matrix3x3< T>& u, const Matrix3x3< T>& v) throw()
  {return Matrix3x3< T>::dist(u,v);}
/// L1 Distance
template < class T> vec_Inline T distL1( const Matrix3x3< T>& u, const Matrix3x3< T>& v) throw()
  {return Matrix3x3< T>::distL1(u,v);}
/// Linf Distance
template < class T> vec_Inline T distLinf( const Matrix3x3< T>& u, const Matrix3x3< T>& v) throw()
  {return Matrix3x3< T>::distLinf(u,v);}
/** dot product
    @param p first vector
    @param q second vector
    @return p.q */
template < class T> vec_Inline T DotProd(const Matrix3x3< T> &p, const Matrix3x3< T> &q) throw()
  {return Matrix3x3< T>::dot(p,q);}
/// Square Euclidian Distance
template < class T> vec_Inline T Dist2( const Matrix3x3< T>& u, const Matrix3x3< T>& v) throw()
  {return Matrix3x3< T>::dist2(u,v);}
/// Euclidian Distance
template < class T> vec_Inline T Dist( const Matrix3x3< T>& u, const Matrix3x3< T>& v) throw()
  {return Matrix3x3< T>::dist(u,v);}
/// L1 Distance
template < class T> vec_Inline T DistL1( const Matrix3x3< T>& u, const Matrix3x3< T>& v) throw()
  {return Matrix3x3< T>::distL1(u,v);}
/// Linf Distance
template < class T> vec_Inline T DistLinf( const Matrix3x3< T>& u, const Matrix3x3< T>& v) throw()
  {return Matrix3x3< T>::distLinf(u,v);}
/** operator == */
template < class T> vec_Inline bool operator == (const Matrix3x3< T>& u, const Matrix3x3< T> &v) throw()
  {return Matrix3x3< T>::isEqual(u,v);}
/** operator != */
template < class T> vec_Inline bool operator != (const Matrix3x3< T>& u, const Matrix3x3< T> &v) throw()
  {return Matrix3x3< T>::isDifferent(u,v);}
/** operator < */
template < class T> vec_Inline bool operator < (const Matrix3x3< T>& u, const Matrix3x3< T> &v) throw()
  {return Matrix3x3< T>::isSmaller(u,v);}
/** operator > */
template < class T> vec_Inline bool operator > (const Matrix3x3< T>& u, const Matrix3x3< T> &v) throw()
  {return Matrix3x3< T>::isGreater(u,v);}
/** operator <= */
template < class T> vec_Inline bool operator <= (const Matrix3x3< T>& u, const Matrix3x3< T> &v) throw()
  {return Matrix3x3< T>::isSmallerOrEqual(u,v);}
/** operator >= */
template < class T> vec_Inline bool operator >= (const Matrix3x3< T>& u, const Matrix3x3< T> &v) throw()
  {return Matrix3x3< T>::isGreaterOrEqual(u,v);}
//@}
/** @name Square matrix redefined friend functions */
//@{
template < class T> vec_Inline Matrix3x3< T> operator*(const Matrix3x3< T> &m1, const Matrix3x3< T> &m2) throw()
  {return Matrix3x3< T>(VIM_NOINIT).Mul(m1,m2);}
//@}


/*! \class Matrix3x3 Matrix3x3.h linearAlgebra/Matrix3x3.h
  \ingroup liblinearAlgebra
  template class for a square matrix of dimension 3 and type T 
*/
template <class T> class Matrix3x3 : public Matrix<3,T> {
public:
  /** @name Constructors */
  //@{
  /** @name Vectorial redefined constructors */
  //@{
  /** default constructor - initializes to 0 if UNINIT_ZERO not defined */
  vec_Inline Matrix3x3() throw() : Matrix<3,T>()
    {}
  /** constructor - initializes to 0 if b==VIM_INIT */
  explicit vec_Inline Matrix3x3(VectInitMode b) throw() : Matrix<3,T>(b)
    {}
  //@}
  /** @name Square matrix redefined constructors */
  //@{
  /// Set to diag( v )
  explicit vec_Inline Matrix3x3(const Vect<3,T>& v) throw() : Matrix<3,T>(VIM_NOINIT)
    {diag(v);}
  
#ifndef WIN32
  /** copy constructor
      @param v vector to copy */
  vec_Inline Matrix3x3(const Mat<3,3,T>& v) throw() : Matrix<3,T>(VIM_NOINIT)
    {copy(v);}
  /** operator =
      @param v vector to copy
      @return this */
  vec_Inline Matrix3x3< T>& operator=(const Mat<3,3,T>& v) throw()
    {copy(v);return *this;}
#endif
  /** copy constructor with cast
      @param v vector to copy */
  template<class T2> vec_Inline Matrix3x3(const Mat<3,3,T2>& v) throw() : Matrix<3,T>(VIM_NOINIT)
    {copy(v);}
  /** operator = with cast
      @param v vector to copy
      @return this */
  template<class T2> vec_Inline Matrix3x3< T>& operator=(const Mat<3,3,T2>& v) throw()
    {copy(v);return *this;}
  
#ifndef WIN32
  /// Constructor : set to matrix coord
  vec_Inline Matrix3x3(const T coord[3][3]) throw() : Matrix<3,T>(VIM_NOINIT)
    {setValues(coord);}
#endif
  /// Constructor : set to matrix coord
  template <class T2> vec_Inline Matrix3x3(const T2 coord[3][3]) throw() : Matrix<3,T>(VIM_NOINIT)
    {setValues(coord);}
  
  /** Creates a square matrix from a symetric matrix
      @param W symetric matrix */
  explicit vec_Inline Matrix3x3(const Sym<3,T>& W) throw() : Matrix<3,T>(VIM_NOINIT)
    {Sym2Mat<3,T>(W,*this);}
  /** Set from a symetric matrix
      @param W symetric matrix */
  vec_Inline Matrix3x3< T>& operator=(const Sym<3,T>& W) throw()
    {Sym2Mat<3,T>(W,*this);return *this;}
  //@}

  /** builds the matrix (x_ij)
      @param x00 first line, first column element
      @param x01 first line, second column element
      @param x02 first line, third column element
      @param x10 second line, first column element
      @param x11 second line, second column element
      @param x12 second line, third column element
      @param x20 third line, first column element
      @param x21 third line, second column element
      @param x22 third line, third column element */
  Matrix3x3(T x00, T x01, T x02,
            T x10, T x11, T x12,
            T x20, T x21, T x22) throw();
  /** builds the matrix (x_ij)
      @param value for i,j xij = value */
  Matrix3x3(T value) throw();



  /** creates a 3x3 matrix from 3 column vectors
      @param v1 first column vector
      @param v2 first column vector
      @param v3 second column vector */
  Matrix3x3(const Vect<3,T> &v1, const Vect<3,T> &v2, const Vect<3,T> &v3) throw();

  /** creates a 3x3 matrix as the tensor product of two vector
      ie its elements  (i,j) are equal to v1[i]*v2[j]
      @param v1 a 3D vector
      @param v2 a 3D vector */
  Matrix3x3(const Vect<3,T> &v1, const Vect<3,T> &v2) throw() : Matrix<3,T>(VIM_NOINIT)
    {tensorProduct(v1,v2);}

  /** creates a 3x3 matrix from a rotation matrix
      @param r the rotation matrix to copy */
  explicit Matrix3x3(const Rotation3D& r) throw() : Matrix<3,T>((const Matrix<3,double>&)r)
    {}
  /// copy a 3x3 matrix from a rotation matrix
  /// @param r
  Matrix3x3& operator=(const Rotation3D & r) throw()
    {this->copy((const Matrix3x3<double>&)r);return *this;}

  /** @name Data members access */
  //@{
  void setValues(T x00, T x01, T x02,
                 T x10, T x11, T x12,
                 T x20, T x21, T x22) throw();

  /** setValues with a T[L][C] array */
  inline void setValues(const T v[3][3]) throw()
    {((Matrix<3,T>&)*this).setValues(v);}

  /** access ithline vector
      @param i line number
      @return line vector */
  Vec3<T> operator[](int i) const throw(ErrRange);
  //@}

  /** @name mathematical operators */
  //@{
  /** @name Vectorial redefined menber functions */
  //@{
  /** set this vector to p+q
      @return this = p + q */
  vec_Inline Matrix3x3< T>& Minus(const Matrix3x3< T> &p) throw()
    {vec_opposite(this->val,9,p.val);return *this;}
  /** set this vector to p+q
      @return this = p + q */
  vec_Inline Matrix3x3< T>& Add(const Matrix3x3< T> &p, const Matrix3x3< T> &q) throw()
    {vec_add(this->val,9,p.val,q.val);return *this;}
  /** set this vector to p-q
      @return this = p - q */
  vec_Inline Matrix3x3< T>& Sub(const Matrix3x3< T> &p, const Matrix3x3< T> &q) throw()
    {vec_sub(this->val,9,p.val,q.val);return *this;}
  /** set each coordinate of this vector to product of this and p same coordinate
      @return this = this.*q */
  vec_Inline Matrix3x3< T>& CMul(const Matrix3x3< T> &q) throw()
    {vec_cmul(this->val,9,q.val);return *this;}
  /** set each coordinate of this vector to product of q and p same coordinate
      @return this = p.*q */
  vec_Inline Matrix3x3< T>& CMul(const Matrix3x3< T> &p, const Matrix3x3< T> &q) throw()
    {vec_cmul(this->val,9,p.val,q.val);return *this;}
  /** set this vector to p*a
      @return this = p*a */
  vec_Inline Matrix3x3< T>& Mul(const Matrix3x3< T> &p, const T a) throw()
    {vec_mul(this->val,9,p.val,a);return *this;}
  /** set this vector to p*a
      @return this = p*a */
  vec_Inline Matrix3x3< T>& Mul(const T a, const Matrix3x3< T> &p) throw()
    {vec_mul(this->val,9,p.val,a);return *this;}
  /** set this vector to p/a
      @return this = p/a */
  vec_Inline Matrix3x3< T>& Div(const Matrix3x3< T> &p, const T a) throw(ErrDivZero)
    {vec_div(this->val,9,p.val,a);return *this;}
  /** adds vector q to vector p
      @param q vector to add
      @return p + q */
  vec_Inline Matrix3x3< T>& operator+=(const Matrix3x3< T> &q) throw()
    {vec_add(this->val,9,q.val);return *this;}
  /** substracts vector q from vector p
      @param q vector to substract
      @return p - q */
  vec_Inline Matrix3x3< T>& operator-=(const Matrix3x3< T> &q) throw()
    {vec_sub(this->val,9,q.val);return *this;}
  /** multiplies vector p by scalar a
      @param a scalar
      @return ap */
  vec_Inline Matrix3x3< T>& operator*=(const T a) throw()
    {vec_mul(this->val,9,a);return *this;}
  /** divides vector p by scalar a
      @param a scalar
      @return p / a */
  vec_Inline Matrix3x3< T>& operator/=(const T a) throw(ErrDivZero)
    {vec_div(this->val,9,a);return *this;}
  /** randomize vector
      @return this */
  vec_Inline Matrix3x3< T>& Rnd(T min = -1., T max = 1.) throw()
    {vec_rnd(this->val,9,min,max);return *this;}
  /** randomize vector : isotropic gaussian centred noise
      @return this */
  vec_Inline Matrix3x3< T>& RndGaussian(T standard_deviation = 1.) throw()
    {vec_rndgaussian(this->val,9,standard_deviation);return *this;}
  /** takes the X,Y or Z value of p if they are less than its own X,Y,Z coordinates
      @return _Vect<9,T>(min(x,p.x),min(y,p.y),...) */
  vec_Inline Matrix3x3< T>& Min(const Matrix3x3< T>& p) throw()
    {vec_min(this->val,9,p.val);return *this;}
  /** takes the X,Y or Z value of p if they are greater than its own X,Y,Z coordinates
      @return _Vect<9,T>(max(x,p.x),max(y,p.y),...) */
  vec_Inline Matrix3x3< T>& Max(const Matrix3x3< T>& p) throw()
    {vec_max(this->val,9,p.val);return *this;}
  /** set each element of this vector to value
      @param v value
      @return this */
  vec_Inline Matrix3x3< T>& Reset(T v) throw()
    {resetTo(v);return *this;}
  /** set this vector to null vector
      @return this */
  vec_Inline Matrix3x3< T>& Null() throw()
  {resetTo((T)0);return *this;}
  /** set this vector to mean of a given list of vectors
      @return this */
  template<class InputIterator>
  vec_Inline Matrix3x3< T>& Mean(InputIterator first, InputIterator last) throw()
    {mean(first,last);return *this;}
  //@}

  /** set this matrix to inverse of a given matrix
    @return this */
  void inv(const Matrix3x3<T> &m) throw(ErrMatSingular);
#ifdef TODO
invert rend matrice nulle en cas de singularite
#endif

  /** computes the determninant
      @return det matrix determinant */
  T det(void) const throw();

  /** @name Square matrix redefined menber functions */
  //@{
  /** set this matrix to inverse of a given matrix
      @return this */
  vec_Inline Matrix3x3< T>& Inv(const Matrix3x3< T>& m) throw(ErrMatSingular)
    {this->inv(m);return *this;}
  /** get this matrix inverse matrix (this matrix is NOT modifyed)
      @return inverse matrix */
  vec_Inline Matrix3x3< T> getInverse(void) const throw(ErrMatSingular)
    {return Matrix3x3< T>(VIM_NOINIT).Inv(*this);}
  /// invert this matrix (this matrix is modified)
  vec_Inline void invert(void) throw(ErrMatSingular)
    {this->Inv(*this);}
  
  /** set this matrix to transposed of a given matrix
      @return this */
  vec_Inline Matrix3x3< T>& Trp(const Matrix3x3< T>& m) throw()
    {this->trp(m);return *this;}
  /** get this matrix transposed matrix (this matrix is NOT modifyed)
      @return transpose matrix */
  vec_Inline Matrix3x3< T> getTransposed(void) const throw()
    {return Matrix3x3< T>(VIM_NOINIT).Trp(*this);}
  /// transpose this matrix (this matrix is modified)
  vec_Inline void transpose(void) throw()
    {this->Trp(*this);}
  
  /// Matrix multiplication
  vec_Inline Matrix3x3< T>& operator*=(const Matrix3x3< T>& m) throw()
    {mulmat(*this,m);return *this;}
  /// Matrix multiplication
  Matrix3x3< T>& Mul(const Matrix3x3< T> &m1, const Matrix3x3< T> &m2) throw()
    {mulmat(m1,m2);return *this;}
  
  /** computes the determninant
      @return det matrix determinant */
  vec_Inline T Det(void) const throw()
    {return this->det();}
  /** computes the trace
      @return trace matrix trace */
  vec_Inline T Trace(void) const throw()
    {return this->trace();}
  
  /// set this matrix to the identity matrix
  vec_Inline Matrix3x3< T>& Id(const T v=1.0) throw()
    {this->resetToIdentity(v);return *this;}
  /// Set to diag( v )
  vec_Inline Matrix3x3< T>& Diag(const Vect<3,T>& v) throw()
    {this->diag(v);return *this;}
  /// Set to tensor product v1*v2
  vec_Inline Matrix3x3< T>& TensorProduct(const Vect<3,T>& v1,const Vect<3,T>& v2) throw()
    {this->tensorProduct(v1,v2);return *this;}
  //@}

  static void Test() throw();
};

} // namespace yav


/*][** NODOC ****************************************************************/
/****************************************************************************/

/*]]**************************************************************************/

namespace yav {

// constructors
template <class T> vec_Inline
Matrix3x3<T>::Matrix3x3(T x00, T x01, T x02,
                        T x10, T x11, T x12,
                        T x20, T x21, T x22) throw()
  : Matrix<3,T>(VIM_NOINIT)
{
  this->val[0]=x00;this->val[1]=x01;this->val[2]=x02;
  this->val[3]=x10;this->val[4]=x11;this->val[5]=x12;
  this->val[6]=x20;this->val[7]=x21;this->val[8]=x22;
}


template <class T> vec_Inline
Matrix3x3<T>::Matrix3x3(T value) throw()
  : Matrix<3,T>(VIM_NOINIT)
{
  unsigned int i;
  for (i=0;i<9;++i)
    this->val[i]=value;
}



template <class T> vec_Inline
Matrix3x3<T>::Matrix3x3(const Vect<3,T> &v1, const Vect<3,T> &v2, const Vect<3,T> &v3) throw()
  : Matrix<3,T>(VIM_NOINIT)
{
  this->val[0]=v1[0]; this->val[1]=v2[0]; this->val[2]=v3[0];
  this->val[3]=v1[1]; this->val[4]=v2[1]; this->val[5]=v3[1];
  this->val[6]=v1[2]; this->val[7]=v2[2]; this->val[8]=v3[2];
}

template <class T> vec_Inline
void Matrix3x3<T>::setValues(T x00, T x01, T x02,
                             T x10, T x11, T x12,
                             T x20, T x21, T x22) throw()
{
  this->val[0]=x00;this->val[1]=x01;this->val[2]=x02;
  this->val[3]=x10;this->val[4]=x11;this->val[5]=x12;
  this->val[6]=x20;this->val[7]=x21;this->val[8]=x22;
}

// access to data
template <class T> vec_Inline
Vec3<T> Matrix3x3<T>::operator[](int i) const throw(ErrRange)
{
#ifdef RANGE_CHECK
  if ( (i<0) || (i >= 3) )
    throw  ErrRange("Mat::getValue(Index,Index)");
#endif
  return Vec3<T>(this->val+i*3);
}

// operators
template <class T> vec_Inline
void Matrix3x3<T>::inv(const Matrix3x3<T> &m) throw(ErrMatSingular)
{
  if ((&m)==this) {
    Matrix3x3<T> M(m);
    this->inv(M);
  } else {
    this->val[0]=m.val[4]*m.val[8]-m.val[7]*m.val[5];
    this->val[1]=m.val[7]*m.val[2]-m.val[1]*m.val[8];
    this->val[2]=m.val[1]*m.val[5]-m.val[4]*m.val[2];
    this->val[3]=m.val[6]*m.val[5]-m.val[3]*m.val[8];
    this->val[4]=m.val[0]*m.val[8]-m.val[6]*m.val[2];
    this->val[5]=m.val[3]*m.val[2]-m.val[0]*m.val[5];
    this->val[6]=m.val[3]*m.val[7]-m.val[6]*m.val[4];
    this->val[7]=m.val[6]*m.val[1]-m.val[0]*m.val[7];
    this->val[8]=m.val[0]*m.val[4]-m.val[3]*m.val[1];
    try{
      *this/=m.Det();
    }
    catch( ErrDivZero exc ){
      throw ErrMatSingular("Matrix3x3::Inv(const Matrix3x3 & )");
    }
  }
}

template <class T> vec_Inline
T Matrix3x3<T>::det() const throw()
{
  return (this->val[0] * this->val[4] * this->val[8] +
          this->val[1] * this->val[5] * this->val[6] +
          this->val[2] * this->val[3] * this->val[7] -
          this->val[6] * this->val[4] * this->val[2] -
          this->val[3] * this->val[1] * this->val[8] -
          this->val[0] * this->val[7] * this->val[5]);
}


template <class T>
void Matrix3x3<T>::Test() throw()
{
  float Af[3][3];double Ad[3][3];int Ai[3][3];
  T A[3][3];for (Index k=0; k<3; k++) for (Index l=0; l<3; l++) {A[k][l]=k*3+l+1;Af[k][l]=k*3+l+1;Ad[k][l]=k*3+l+1;Ai[k][l]=k*3+l+1;}
  Matrix3x3<float> df;Matrix3x3<double> dd;Matrix3x3<int> di;
  Vec3<T> bb(0);

  Matrix3x3<T> a(A),aaa(1,2,3,4,5,6,7,8,9),aab(bb,bb),aac(bb,bb,bb),b(bb),c(VIM_INIT),d(a),/*af(Af),ad(Ad),ai(Ai),*/ef(df),ed(dd),ei(di);
  T v;char TName[100];Xlinalg_Class_Name(v,TName);
  COUT << "Matrix3x3<"<<TName<<">" << ENDL;
  COUT << "Constructors" << ENDL;
  COUT << "a must be 1,2,... : " << a << ENDL;
  COUT << "aaa must be 1,2,... : " << aaa << ENDL;
  COUT << "aab must be 1,0,0*1,0,0 : " << aab << ENDL;
  COUT << "aac must be 1,0,0,1,0 : " << aac << ENDL;
//   COUT << "af must be 1,2,... : " << af << ENDL;
//   COUT << "ad must be 1,2,... : " << ad << ENDL;
//   COUT << "ai must be 1,2,... : " << ai << ENDL;
  COUT << "b must be diag(1,0,...) : " << b << ENDL;
  COUT << "c must be 0,0,... : " << c << ENDL;
  COUT << "d must be as a (copy) : " << d << ENDL;
  COUT << "ef must be 0 : " << ef << ENDL;
  COUT << "ed must be 0 : " << ed << ENDL;
  COUT << "ei must be 0 : " << ei << ENDL;

  COUT << "operator=" << ENDL;
  d=c;COUT << "d=c : " << d << ENDL;
  d=df;COUT << "d=df : " << d << ENDL;
  d=dd;COUT << "d=dd : " << d << ENDL;
  d=di;COUT << "d=di : " << d << ENDL;

  COUT << "Access to Data" << ENDL;
  d.copy(b);COUT << "d.copy(b)  : " << d << ENDL;
  COUT << "d[0] : " << d[0] << ENDL;
  d[0][0]=25;COUT << "d[0][0]=25;  d[0][0] : " << d[0][0] << ENDL;
  d(0,0)=26;COUT << "d(0,0)=26;  d(0,0) : " << d(0,0) << ENDL;
  d.setValue(0,0,11);COUT << "d.setValue(0,0,11);  d.getValue(0,0) : " << d.getValue(0,0) << ENDL;
  d.Null();COUT << "d.Null() : " << d << ENDL;
  d.resetTo(2);COUT << "d.resetTo(2) : " << d << ENDL;
  T*Dd=d.getValues();
  COUT << "getValues() : " << Dd[0] << " " << Dd[1] << " " << Dd[2] << ENDL;
  const T*E=d;
  COUT << "(const T*) d : " << E[0] << " " << E[1] << " " << E[2] << ENDL;
  d.setValues(A);COUT << "d.setValues(T[1,2,3...]) : " << d << ENDL;
//   d.setValues(Af);COUT << "d.setValues(float[1,2,3...]) : " << d << ENDL;
//   d.setValues(Ad);COUT << "d.setValues(double[1,2,3...]) : " << d << ENDL;
//   d.setValues(Ai);COUT << "d.setValues(int[1,2,3...]) : " << d << ENDL;

  COUT << "Math" << ENDL;
  COUT << "d.Minus(a) : " << d.Minus(a) << ENDL;
  COUT << "d.Add(a,b) : " << d.Add(a,b) << ENDL;
  COUT << "d.Sub(a,b) : " << d.Sub(a,b) << ENDL;
  COUT << "d.Mul(a,2) : " << d.Mul(a,(T)2) << ENDL;
  COUT << "d.Mul(2,a) : " << d.Mul((T)2,a) << ENDL;
  COUT << "d.Div(a,2) : " << d.Div(a,(T)2) << ENDL;
  COUT << "d.Div(a,2) : " << d.Div(a,(T)2) << ENDL;

  COUT << "+a : " << +a << ENDL;
  COUT << "-a : " << -a << ENDL;
  COUT << "a+b : " << a+b << ENDL;
  COUT << "a-b : " << a-b << ENDL;
  COUT << "a*2 : " << a*((T)2) << ENDL;
  COUT << "2*a : " << ((T)2)*a << ENDL;
  COUT << "a/2 : " << a/((T)2) << ENDL;

  d.copy(b);d+=a;COUT << "d=b; d+=a : " << d << ENDL;
  d.copy(b);d-=a;COUT << "d=b; d-=a : " << d << ENDL;
  d.copy(b);d*=(T)2;COUT << "d=b; d*=2 : " << d << ENDL;
  d.copy(b);d/=(T)2;COUT << "d=b; d/=2 : " << d << ENDL;

  d.copy(b);COUT << "d=b; d==b : " << (d==b) << ENDL;
  d.copy(b);COUT << "d=b; d!=b : " << (d!=b) << ENDL;

  COUT << "dot(a,b) : " << dot(a,b) << ENDL;
  COUT << "DotProd(a,b) : " << DotProd(a,b) << ENDL;
  COUT << "norm2(a) : " << a.norm2() << ENDL;
  COUT << "norm(a) : " << a.norm() << ENDL;
  COUT << "Norm2(a) : " << a.Norm2() << ENDL;
  COUT << "Norm(a) : " << a.Norm() << ENDL;
  COUT << "NormL1(a) : " << a.NormL1() << ENDL;
  COUT << "NormLinf(a) : " << a.NormLinf() << ENDL;
  COUT << "Dist2(a,b) : " << Dist2(a,b) << ENDL;
  COUT << "Dist(a,b) : " << Dist(a,b) << ENDL;
  COUT << "DistL1(a,b) : " << DistL1(a,b) << ENDL;
  COUT << "DistLinf(a,b) : " << DistLinf(a,b) << ENDL;

  d.resetTo(2); d.min(a); COUT << "d.resetTo(2); d.min(a) : " << d << ENDL;
  d.resetTo(2); d.max(a); COUT << "d.resetTo(2); d.max(a) : " << d << ENDL;

  d.Rnd(2); COUT << "d.Rnd(2) : " << d << ENDL;

  Matrix3x3<T> m;COUT << "m.Trp(d) : " << m.Trp(d) << ENDL;
  d.transpose();COUT << "d.transpose() : " << d << ENDL;
  COUT << "d.getTransposed() : " << d.getTransposed() << ENDL;

  {Matrix3x3<T> m;Matrix3x3<T> n;COUT << "Matrix3x3<T>*Matrix3x3<T> : " << m*n << "  Mul : " << d.Mul(m,n) << ENDL;}

  {d.Rnd();Sym<3,T> S1;Sym<3,T> S2;d.SymAction(S1,S2);COUT << "SymAction" << ENDL;}
  {d.Rnd();Matrix3x3<T> S1;Vec3<T>W;Matrix3x3<T> S2;d.SVD(S1,W,S2);COUT << "SVD" << ENDL;}
  {d.Rnd();Matrix3x3<T> S;d.PseudoInv(S);COUT << "PseudoInv" << ENDL;}

  {d.Rnd();d.Inv(d);COUT << "Inv" << ENDL;}
  {d.Rnd();COUT << "getInverse" << d.getInverse() << ENDL;}
  {d.Rnd();d.invert();COUT << "invert" << ENDL;}

  {d.Rnd();COUT << "Det" << d.Det() << ENDL;}
  {d.Rnd();COUT << "Trace" << d.Trace() << ENDL;}

  {d.Rnd();COUT << "Id" << d.Id() << ENDL;}
  {d.Rnd();Vec3<T>v;COUT << "Diag" << d.Diag(v) << ENDL;}
  {d.Rnd();Vec3<T>v;COUT << "TensorProduct" << d.TensorProduct(v,v) << ENDL;}

  {Sym<3,T> m;m.Rnd();d=m;COUT << "operator=Sym<3,T>" << d << "=" << m << ENDL;}
  {Sym<3,T> m;m.Rnd();Matrix3x3<T> d(m);COUT << "Constructeur(Sym<3,T>)" << d << "=" << m << ENDL;}

  COUT << ENDL;
}


} // namespace yav

#endif
