
// General comments : see Vec.h

#ifndef _MATRIX4X4_H
#define _MATRIX4X4_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#ifndef WIN32
#pragma interface "linearAlgebra/Matrix4x4.h"
#endif
#include "linearAlgebra/Vec.h"
#include "linearAlgebra/Mat.h"
#include "linearAlgebra/Vec3.h"
#include "linearAlgebra/Matrix3x3.h"
#include "linearAlgebra/Vec4.h"
#include "linearAlgebra/HVec4.h"
#include "linearAlgebra/Rotation3D.h"

#ifdef WIN32
#include "basics/export.h"
#endif


namespace yav {


// forward declarations
class RigidMat3D;


template <class T> Vec3<T> operator*(const Matrix4x4<T> &m, const Vec3<T>& v);

/** @name Vectorial redefined friend functions */
//@{
/** unary +
    @param p vector
    @return +p */
template < class T> vec_Inline Matrix4x4< T> operator+(const Matrix4x4< T> &p) throw()
  {return p;}
/** unary -
    @param p vector
    @return -p */
template < class T> vec_Inline Matrix4x4< T> operator-(const Matrix4x4< T> &p) throw()
  {return Matrix4x4< T>(VIM_NOINIT).Minus(p);}
/** vector sum
    @param p first vector
    @param q second vector
    @return p + q */
template < class T> vec_Inline Matrix4x4< T> operator+(const Matrix4x4< T> &p, const Matrix4x4 < T> &q) throw()
  {return Matrix4x4< T>(VIM_NOINIT).Add(p,q);}
/** vector substraction
    @param p first vector
    @param q second vector
    @return p - q */
template < class T> vec_Inline Matrix4x4< T> operator-(const Matrix4x4< T> &p, const Matrix4x4 < T> &q) throw()
  {return Matrix4x4< T>(VIM_NOINIT).Sub(p,q);}
/** vector multiplication by scalar
    @param p vector
    @param a scalar
    @return ap */
template < class T> vec_Inline Matrix4x4< T> operator*(const Matrix4x4< T> &p, const T a) throw()
  {return Matrix4x4< T>(VIM_NOINIT).Mul(p,a);}
/** vector scalar multiplication
    @param a scalar
    @param p vector
    @return ap */
template < class T> vec_Inline Matrix4x4< T> operator*(const T a, const Matrix4x4< T> &p) throw()
  {return Matrix4x4< T>(VIM_NOINIT).Mul(p,a);}
/** vector scalar division
    @param p vector
    @param a scalar
    @return p / a */
template < class T> vec_Inline Matrix4x4< T> operator/(const Matrix4x4< T> &p, const T a) throw(ErrDivZero)
  {return Matrix4x4< T>(VIM_NOINIT).Div(p,a);}
/** dot product
    @param u first vector
    @param v second vector
    @return u.v */
template < class T> vec_Inline T dot(const Matrix4x4< T>& u, const Matrix4x4< T> &v) throw()
  {return Matrix4x4< T>::dot(u,v);}
/// Square Euclidian Distance
template < class T> vec_Inline T dist2( const Matrix4x4< T>& u, const Matrix4x4< T>& v) throw()
  {return Matrix4x4< T>::dist2(u,v);}
/// Euclidian Distance
template < class T> vec_Inline T dist( const Matrix4x4< T>& u, const Matrix4x4< T>& v) throw()
  {return Matrix4x4< T>::dist(u,v);}
/// L1 Distance
template < class T> vec_Inline T distL1( const Matrix4x4< T>& u, const Matrix4x4< T>& v) throw()
  {return Matrix4x4< T>::distL1(u,v);}
/// Linf Distance
template < class T> vec_Inline T distLinf( const Matrix4x4< T>& u, const Matrix4x4< T>& v) throw()
  {return Matrix4x4< T>::distLinf(u,v);}
/** dot product
    @param p first vector
    @param q second vector
    @return p.q */
template < class T> vec_Inline T DotProd(const Matrix4x4< T> &p, const Matrix4x4< T> &q) throw()
  {return Matrix4x4< T>::dot(p,q);}
/// Square Euclidian Distance
template < class T> vec_Inline T Dist2( const Matrix4x4< T>& u, const Matrix4x4< T>& v) throw()
  {return Matrix4x4< T>::dist2(u,v);}
/// Euclidian Distance
template < class T> vec_Inline T Dist( const Matrix4x4< T>& u, const Matrix4x4< T>& v) throw()
  {return Matrix4x4< T>::dist(u,v);}
/// L1 Distance
template < class T> vec_Inline T DistL1( const Matrix4x4< T>& u, const Matrix4x4< T>& v) throw()
  {return Matrix4x4< T>::distL1(u,v);}
/// Linf Distance
template < class T> vec_Inline T DistLinf( const Matrix4x4< T>& u, const Matrix4x4< T>& v) throw()
  {return Matrix4x4< T>::distLinf(u,v);}
/** operator == */
template < class T> vec_Inline bool operator == (const Matrix4x4< T>& u, const Matrix4x4< T> &v) throw()
  {return Matrix4x4< T>::isEqual(u,v);}
/** operator != */
template < class T> vec_Inline bool operator != (const Matrix4x4< T>& u, const Matrix4x4< T> &v) throw()
  {return Matrix4x4< T>::isDifferent(u,v);}
/** operator < */
template < class T> vec_Inline bool operator < (const Matrix4x4< T>& u, const Matrix4x4< T> &v) throw()
  {return Matrix4x4< T>::isSmaller(u,v);}
/** operator > */
template < class T> vec_Inline bool operator > (const Matrix4x4< T>& u, const Matrix4x4< T> &v) throw()
  {return Matrix4x4< T>::isGreater(u,v);}
/** operator <= */
template < class T> vec_Inline bool operator <= (const Matrix4x4< T>& u, const Matrix4x4< T> &v) throw()
  {return Matrix4x4< T>::isSmallerOrEqual(u,v);}
/** operator >= */
template < class T> vec_Inline bool operator >= (const Matrix4x4< T>& u, const Matrix4x4< T> &v) throw()
  {return Matrix4x4< T>::isGreaterOrEqual(u,v);}
//@}
/** @name Square matrix redefined friend functions */
//@{
template < class T> vec_Inline Matrix4x4< T> operator*(const Matrix4x4< T> &m1, const Matrix4x4< T> &m2) throw()
  {return Matrix4x4< T>(VIM_NOINIT).Mul(m1,m2);}
//@}

/*! \class Matrix4x4 Matrix4x4.h linearAlgebra/Matrix4x4.h
  \ingroup liblinearAlgebra
  template class for a square matrix of dimension 4 and type T 
*/
template <class T> class Matrix4x4 : public Matrix<4,T> {
public:
  /** @name Constructors */
  //@{
  /** @name Vectorial redefined constructors */
  //@{
  /** default constructor - initializes to 0 if UNINIT_ZERO not defined */
  vec_Inline Matrix4x4() throw() : Matrix<4,T>()
    {}
  /** constructor - initializes to 0 if b==VIM_INIT */
  explicit vec_Inline Matrix4x4(VectInitMode b) throw() : Matrix<4,T>(b)
    {}
  //@}
  /** @name Square matrix redefined constructors */
  //@{
  /// Set to diag( v )
  explicit vec_Inline Matrix4x4(const Vect<4,T>& v) throw() : Matrix<4,T>(VIM_NOINIT)
    {diag(v);}
  
#ifndef WIN32
  /** copy constructor
      @param v vector to copy */
  vec_Inline Matrix4x4(const Mat<4,4,T>& v) throw() : Matrix<4,T>(VIM_NOINIT)
    {copy(v);}
  /** operator =
      @param v vector to copy
      @return this */
  vec_Inline Matrix4x4< T>& operator=(const Mat<4,4,T>& v) throw()
    {copy(v);return *this;}
#endif
  /** copy constructor with cast
      @param v vector to copy */
  template<class T2> vec_Inline Matrix4x4(const Mat<4,4,T2>& v) throw() : Matrix<4,T>(VIM_NOINIT)
    {copy(v);}
  /** operator = with cast
      @param v vector to copy
      @return this */
  template<class T2> vec_Inline Matrix4x4< T>& operator=(const Mat<4,4,T2>& v) throw()
    {copy(v);return *this;}
  
#if (!defined( WIN32) || (defined(__CYGWIN__)))
  /// Constructor : set to matrix coord
  vec_Inline Matrix4x4(const T coord[4][4]) throw() : Matrix<4,T>(VIM_NOINIT)
    {setValues(coord);}
#endif
  /// Constructor : set to matrix coord
  template <class T2> vec_Inline Matrix4x4(const T2 coord[4][4]) throw() : Matrix<4,T>(VIM_NOINIT)
    {setValues(coord);}
  
  /** Creates a square matrix from a symetric matrix
      @param W symetric matrix */
  explicit vec_Inline Matrix4x4(const Sym<4,T>& W) throw() : Matrix<4,T>(VIM_NOINIT)
    {Sym2Mat<4,T>(W,*this);}
  /** Set from a symetric matrix
      @param W symetric matrix */
  vec_Inline Matrix4x4< T>& operator=(const Sym<4,T>& W) throw()
    {Sym2Mat<4,T>(W,*this);return *this;}
  //@}

  /** builds the matrix (x_ij)
      @param x00 first line, first column element
      @param x01 first line, second column element
      @param x02 first line, third column element
      @param x03 first line, fourth column element
      @param x10 second line, first column element
      @param x11 second line, second column element
      @param x12 second line, third column element
      @param x13 second line, fourth column element
      @param x20 third line, first column element
      @param x21 third line, second column element
      @param x22 third line, third column element
      @param x23 third line, fourth column element
      @param x30 fourth line, first column element
      @param x31 fourth line, second column element
      @param x32 fourth line, third column element
      @param x33 fourth line, fourth column element */
  Matrix4x4(T x00, T x01, T x02, T x03,
            T x10, T x11, T x12, T x13,
            T x20, T x21, T x22, T x23,
            T x30, T x31, T x32, T x33) throw();

  /** creates a 4x4 matrix from a rotation matrix and
      a translation vector
      @param Rot rotation matrix
      @param Tr translation vector */
  Matrix4x4(const Rotation3D& Rot, const Vec3<T>& Tr) throw();

  /** creates a 3x3 matrix from a rotation matrix
      @param r the rotation matrix to copy */
  explicit Matrix4x4(const RigidMat3D& r) throw() : Matrix<4,T>((const Matrix<4,double>&)r)
    {}
  /// copy a 3x3 matrix from a rotation matrix
  Matrix4x4& operator=(const RigidMat3D & r) throw()
    {this->copy((const Matrix4x4<double>&)r);return *this;}

  /** @name Data members access */
  //@{
  void setValues(T x00, T x01, T x02, T x03,
                 T x10, T x11, T x12, T x13,
                 T x20, T x21, T x22, T x23,
                 T x30, T x31, T x32, T x33) throw();

  /** setValues with a T[L][C] array */
  inline void setValues(const T v[4][4]) throw()
    {((Matrix<4,T>&)*this).setValues(v);}

  /** returns a COPY of ithline vector
      @param i line number
      @return line vector */
  HVec4<T> operator[](int i) const throw(ErrRange);

  /** set rotation part of this matrix using m values
      @param m matrix value used to reset this */
  void operator=(const Matrix3x3<T> &m) throw()
    {
      for(Index i = 0; i < 3; i++)
        for(Index j = 0; j < 3; j++)
          this->val[i*4+j] = m.getValue(i, j);
    };

  /** reset rotation part of this matrix with m matrix values
      last line and last column of this matrix are unchanged
      @param m new rotation matrix */
  void operator=(const Rotation3D& m) throw()
    {
      for(Index i = 0; i < 3; i++)
        for(Index j = 0; j < 3; j++)
          this->val[i*4+j] = m.getValue(i, j);
    };

  /** reset translation part of this matrix, rest of matrix is unchanged
      @param v translation vector */
  Matrix4x4<T>&  operator=(const Vect<3,T>& v) throw()
    {
      for(Index i = 0; i < 3; i++)
        this->val[i*4+3] = v[i];
      return *this;
    };
  //@}

  /** @name mathematical operators */
  //@{
  /** @name Vectorial redefined menber functions */
  //@{
  /** set this vector to p+q
      @return this = p + q */
  vec_Inline Matrix4x4< T>& Minus(const Matrix4x4< T> &p) throw()
    {vec_opposite(this->val,16,p.val);return *this;}
  /** set this vector to p+q
      @return this = p + q */
  vec_Inline Matrix4x4< T>& Add(const Matrix4x4< T> &p, const Matrix4x4< T> &q) throw()
    {vec_add(this->val,16,p.val,q.val);return *this;}
  /** set this vector to p-q
      @return this = p - q */
  vec_Inline Matrix4x4< T>& Sub(const Matrix4x4< T> &p, const Matrix4x4< T> &q) throw()
    {vec_sub(this->val,16,p.val,q.val);return *this;}
  /** set each coordinate of this vector to product of this and p same coordinate
      @return this = this.*q */
  vec_Inline Matrix4x4< T>& CMul(const Matrix4x4< T> &q) throw()
    {vec_cmul(this->val,16,q.val);return *this;}
  /** set each coordinate of this vector to product of q and p same coordinate
      @return this = p.*q */
  vec_Inline Matrix4x4< T>& CMul(const Matrix4x4< T> &p, const Matrix4x4< T> &q) throw()
    {vec_cmul(this->val,16,p.val,q.val);return *this;}
  /** set this vector to p*a
      @return this = p*a */
  vec_Inline Matrix4x4< T>& Mul(const Matrix4x4< T> &p, const T a) throw()
    {vec_mul(this->val,16,p.val,a);return *this;}
  /** set this vector to p*a
      @return this = p*a */
  vec_Inline Matrix4x4< T>& Mul(const T a, const Matrix4x4< T> &p) throw()
    {vec_mul(this->val,16,p.val,a);return *this;}
  /** set this vector to p/a
      @return this = p/a */
  vec_Inline Matrix4x4< T>& Div(const Matrix4x4< T> &p, const T a) throw(ErrDivZero)
    {vec_div(this->val,16,p.val,a);return *this;}
  /** adds vector q to vector p
      @param q vector to add
      @return p + q */
  vec_Inline Matrix4x4< T>& operator+=(const Matrix4x4< T> &q) throw()
    {vec_add(this->val,16,q.val);return *this;}
  /** substracts vector q from vector p
      @param q vector to substract
      @return p - q */
  vec_Inline Matrix4x4< T>& operator-=(const Matrix4x4< T> &q) throw()
    {vec_sub(this->val,16,q.val);return *this;}
  /** multiplies vector p by scalar a
      @param a scalar
      @return ap */
  vec_Inline Matrix4x4< T>& operator*=(const T a) throw()
    {vec_mul(this->val,16,a);return *this;}
  /** divides vector p by scalar a
      @param a scalar
      @return p / a */
  vec_Inline Matrix4x4< T>& operator/=(const T a) throw(ErrDivZero)
    {vec_div(this->val,16,a);return *this;}
  /** randomize vector
      @return this */
  vec_Inline Matrix4x4< T>& Rnd(T min = (T)-1., T max = (T)1.) throw()
    {vec_rnd(this->val,16,min,max);return *this;}
  /** randomize vector : isotropic gaussian centred noise
      @return this */
  vec_Inline Matrix4x4< T>& RndGaussian(T standard_deviation = (T)1.) throw()
    {vec_rndgaussian(this->val,16,standard_deviation);return *this;}
  /** takes the X,Y or Z value of p if they are less than its own X,Y,Z coordinates
      @return _Vect<16,T>(min(x,p.x),min(y,p.y),...) */
  vec_Inline Matrix4x4< T>& Min(const Matrix4x4< T>& p) throw()
    {vec_min(this->val,16,p.val);return *this;}
  /** takes the X,Y or Z value of p if they are greater than its own X,Y,Z coordinates
      @return _Vect<16,T>(max(x,p.x),max(y,p.y),...) */
  vec_Inline Matrix4x4< T>& Max(const Matrix4x4< T>& p) throw()
    {vec_max(this->val,16,p.val);return *this;}
  /** set each element of this vector to value
      @param v value
      @return this */
  vec_Inline Matrix4x4< T>& Reset(T v) throw()
    {resetTo(v);return *this;}
  /** set this vector to null vector
      @return this */
  vec_Inline Matrix4x4< T>& Null() throw()
  {resetTo((T)0);return *this;}
  /** set this vector to mean of a given list of vectors
      @return this */
  template<class InputIterator>
  vec_Inline Matrix4x4< T>& Mean(InputIterator first, InputIterator last) throw()
    {mean(first,last);return *this;}
  //@}

  /** set this matrix to inverse of a given matrix
    @return this */
  void inv(const Matrix4x4<T> &m) throw(ErrMatSingular);
#ifdef TODO
invert rend matrice nulle en cas de singularite
#endif

  /** computes the determninant
      @return det matrix determinant */
  T det(void) const throw();

  /** @name Square matrix redefined menber functions */
  //@{
  /** set this matrix to inverse of a given matrix
      @return this */
  vec_Inline Matrix4x4< T>& Inv(const Matrix4x4< T>& m) throw(ErrMatSingular)
    {this->inv(m);return *this;}
  /** get this matrix inverse matrix (this matrix is NOT modifyed)
      @return inverse matrix */
  vec_Inline Matrix4x4< T> getInverse(void) const throw(ErrMatSingular)
    {return Matrix4x4< T>(VIM_NOINIT).Inv(*this);}
  /// invert this matrix (this matrix is modified)
  vec_Inline void invert(void) throw(ErrMatSingular)
    {this->Inv(*this);}
  
  /** set this matrix to transposed of a given matrix
      @return this */
  vec_Inline Matrix4x4< T>& Trp(const Matrix4x4< T>& m) throw()
    {this->trp(m);return *this;}
  /** get this matrix transposed matrix (this matrix is NOT modifyed)
      @return transpose matrix */
  vec_Inline Matrix4x4< T> getTransposed(void) const throw()
    {return Matrix4x4< T>(VIM_NOINIT).Trp(*this);}
  /// transpose this matrix (this matrix is modified)
  vec_Inline void transpose(void) throw()
    {this->Trp(*this);}
  
  /// Matrix multiplication
  vec_Inline Matrix4x4< T>& operator*=(const Matrix4x4< T>& m) throw()
    {mulmat(*this,m);return *this;}
  /// Matrix multiplication
  Matrix4x4< T>& Mul(const Matrix4x4< T> &m1, const Matrix4x4< T> &m2) throw()
    {mulmat(m1,m2);return *this;}
  
  /** computes the determninant
      @return det matrix determinant */
  vec_Inline T Det(void) const throw()
    {return this->det();}
  /** computes the trace
      @return trace matrix trace */
  vec_Inline T Trace(void) const throw()
    {return this->trace();}
  
  /// set this matrix to the identity matrix
  vec_Inline Matrix4x4< T>& Id(const T v=1.0) throw()
    {this->resetToIdentity(v);return *this;}
  /// Set to diag( v )
  vec_Inline Matrix4x4< T>& Diag(const Vect<4,T>& v) throw()
    {this->diag(v);return *this;}
  /// Set to tensor product v1*v2
  vec_Inline Matrix4x4< T>& TensorProduct(const Vect<4,T>& v1,const Vect<4,T>& v2) throw()
    {this->tensorProduct(v1,v2);return *this;}
  //@}

  //friend Vec3<T> operator* FRIEND_TEMPLATE_ARG(const Matrix4x4<T> &m, const Vec3<T>& v);

  static void Test() throw();

protected:
  /** computes one matrix element cofactor
      @param nl element line number
      @param nc element column number
      @return cofactor of matrix[nl][nc] */
  T cofactor(Index nl, Index nc) const throw();
};



} // namespace yav


/*][** NODOC ****************************************************************/
/****************************************************************************/

/*]]**************************************************************************/

namespace yav {

// constructors
template <class T> vec_Inline
Matrix4x4<T>::Matrix4x4(const Rotation3D& Rot, const Vec3<T>& Tr) throw()
  : Matrix<4,T>(VIM_NOINIT)
{
  this->val[0]=Rot(0,0);this->val[1]=Rot(0,1);this->val[2]=Rot(0,2); this->val[3]=Tr(0);
  this->val[4]=Rot(1,0);this->val[5]=Rot(1,1);this->val[6]=Rot(1,2); this->val[7]=Tr(1);
  this->val[8]=Rot(2,0);this->val[9]=Rot(2,1);this->val[10]=Rot(2,2);this->val[11]=Tr(2);
  this->val[12]=0;      this->val[13]=0;      this->val[14]=0;       this->val[15]=1;
}

// access to data
template <class T> vec_Inline
Matrix4x4<T>::Matrix4x4(T x00, T x01, T x02, T x03,
                        T x10, T x11, T x12, T x13,
                        T x20, T x21, T x22, T x23,
                        T x30, T x31, T x32, T x33) throw()
  : Matrix<4,T>(VIM_NOINIT)
{
  this->val[0]=x00; this->val[1]=x01; this->val[2]=x02; this->val[3]=x03;
  this->val[4]=x10; this->val[5]=x11; this->val[6]=x12; this->val[7]=x13;
  this->val[8]=x20; this->val[9]=x21; this->val[10]=x22;this->val[11]=x23;
  this->val[12]=x30;this->val[13]=x31;this->val[14]=x32;this->val[15]=x33;
}

template <class T> vec_Inline
void Matrix4x4<T>::setValues(T x00, T x01, T x02, T x03,
                             T x10, T x11, T x12, T x13,
                             T x20, T x21, T x22, T x23,
                             T x30, T x31, T x32, T x33) throw()
{
  this->val[0]=x00; this->val[1]=x01; this->val[2]=x02; this->val[3]=x03;
  this->val[4]=x10; this->val[5]=x11; this->val[6]=x12; this->val[7]=x13;
  this->val[8]=x20; this->val[9]=x21; this->val[10]=x22;this->val[11]=x23;
  this->val[12]=x30;this->val[13]=x31;this->val[14]=x32;this->val[15]=x33;
}


template <class T> vec_Inline
HVec4<T> Matrix4x4<T>::operator[](int i) const throw(ErrRange)
{
#ifdef RANGE_CHECK
  if ( (i<0) || (i >= 4) )
    throw  ErrRange("Mat::getValue(Index,Index)");
#endif
  return HVec4<T>(this->val[i*4],this->val[i*4+1],this->val[i*4+2],this->val[i*4+3]);
}

// operations
template <class T> vec_Inline
Vec3<T> operator*(const Matrix4x4<T> &m, const Vec3<T>& v)
{
  return Vec3<T>(v(0)*m(0,0) + v(1)*m(0,1) + v(2)*m(0,2) + m(0,3),
                 v(0)*m(1,0) + v(1)*m(1,1) + v(2)*m(1,2) + m(1,3),
                 v(0)*m(2,0) + v(1)*m(2,1) + v(2)*m(2,2) + m(2,3));
}

template <class T>
T Matrix4x4<T>::cofactor(Index nl, Index nc) const throw() {
  unsigned int i, j, l[3], c[3];

  for(i = 0, (nl == 0) ? j = 1 : j = 0; i < 3;
      i++, (j + 1 == nl) ? j += 2 : j++) l[i] = j*4;
  for(i = 0, (nc == 0) ? j = 1 : j = 0; i < 3;
      i++, (j + 1 == nc) ? j += 2 : j++) c[i] = j;

  return this->val[l[0]+c[0]] * this->val[l[1]+c[1]] * this->val[l[2]+c[2]] +
      this->val[l[0]+c[1]] * this->val[l[1]+c[2]] * this->val[l[2]+c[0]] +
      this->val[l[1]+c[0]] * this->val[l[2]+c[1]] * this->val[l[0]+c[2]] -
      this->val[l[2]+c[0]] * this->val[l[1]+c[1]] * this->val[l[0]+c[2]] -
      this->val[l[1]+c[0]] * this->val[l[0]+c[1]] * this->val[l[2]+c[2]] -
      this->val[l[0]+c[0]] * this->val[l[2]+c[1]] * this->val[l[1]+c[2]];
}

template <class T>
void Matrix4x4<T>::inv(const Matrix4x4<T> &m) throw(ErrMatSingular)
{
#ifdef TODO
  verifier cas matrice singuliere
#endif //TODO
  T det, cof[16];
  Index i, j;
  int sgn;
  cof[0] = m.cofactor(0, 0);
  cof[1] = m.cofactor(0, 1);
  cof[2] = m.cofactor(0, 2);
  cof[3] = m.cofactor(0, 3);
  cof[4] = m.cofactor(1, 0);
  cof[5] = m.cofactor(1, 1);
  cof[6] = m.cofactor(1, 2);
  cof[7] = m.cofactor(1, 3);
  cof[8] = m.cofactor(2, 0);
  cof[9] = m.cofactor(2, 1);
  cof[10] = m.cofactor(2, 2);
  cof[11] = m.cofactor(2, 3);
  cof[12] = m.cofactor(3, 0);
  cof[13] = m.cofactor(3, 1);
  cof[14] = m.cofactor(3, 2);
  cof[15] = m.cofactor(3, 3);
  det = ( m.val[0]*cof[0] - m.val[1]*cof[1]
          + m.val[2]*cof[2] - m.val[3]*cof[3] );

  for(i = 0; i < 4; i++) {
    if(i % 2) sgn = -1;
    else sgn = 1;
    this->val[i*4] = sgn * cof[i];
    for(j = 1; j < 4; j++) {
      if((i + j ) % 2) sgn = -1;
      else sgn = 1;
      this->val[i*4+j] = sgn * cof[4*j+i];
    }
  }

  try{
    *this/=det;
  }
  catch( ErrDivZero exc ){
    throw ErrMatSingular("Matrix4x4::Inv(const Matrix4x4 & )");
  }
}

template <class T> vec_Inline
T Matrix4x4<T>::det() const throw()
{
  return ( this->val[0]*cofactor(0, 0) - this->val[1]*cofactor(0, 1)
           + this->val[2]*cofactor(0, 2) - this->val[3]*cofactor(0, 3) );
}

template <class T>
void Matrix4x4<T>::Test() throw()
{
  float Af[4][4];double Ad[4][4];int Ai[4][4];
  T A[4][4];for (Index k=0; k<4; k++) for (Index l=0; l<4; l++) {A[k][l]=k*4+l+1;Af[k][l]=k*4+l+1;Ad[k][l]=k*4+l+1;Ai[k][l]=k*4+l+1;}
  Matrix4x4<float> df;Matrix4x4<double> dd;Matrix4x4<int> di;
  Vec4<T> bb(0);Vec3<T> bbb(0);Rotation3D rot;

  Matrix4x4<T> a(A),aaa(1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16),aab(rot,bbb),b(bb),c(VIM_INIT),d(a),/*af(Af),ad(Ad),ai(Ai),*/ef(df),ed(dd),ei(di);
  T v;char TName[100];Xlinalg_Class_Name(v,TName);
  COUT << "Matrix4x4<"<<TName<<">" << ENDL;
  COUT << "Constructors" << ENDL;
  COUT << "a must be 1,2,... : " << a << ENDL;
  COUT << "aaa must be 1,2,... : " << aaa << ENDL;
  COUT << "aab must be rot=Id Trans=(1,0,0) : " << aab << ENDL;
//   COUT << "af must be 1,2,... : " << af << ENDL;
//   COUT << "ad must be 1,2,... : " << ad << ENDL;
//   COUT << "ai must be 1,2,... : " << ai << ENDL;
  COUT << "b must be diag(1,0,...) : " << b << ENDL;
  COUT << "c must be 0,0,... : " << c << ENDL;
  COUT << "d must be as a (copy) : " << d << ENDL;
  COUT << "ef must be 0 : " << ef << ENDL;
  COUT << "ed must be 0 : " << ed << ENDL;
  COUT << "ei must be 0 : " << ei << ENDL;

  COUT << "operator=" << ENDL;
  d=c;COUT << "d=c : " << d << ENDL;
  d=df;COUT << "d=df : " << d << ENDL;
  d=dd;COUT << "d=dd : " << d << ENDL;
  d=di;COUT << "d=di : " << d << ENDL;

  COUT << "Access to Data" << ENDL;
  d.copy(b);COUT << "d.copy(b)  : " << d << ENDL;
  COUT << "d[0] : " << d[0] << ENDL;
  d[0][0]=25;COUT << "d[0][0]=25;  d[0][0] : " << d[0][0] << ENDL;
  d(0,0)=26;COUT << "d(0,0)=26;  d(0,0) : " << d(0,0) << ENDL;
  d.setValue(0,0,11);COUT << "d.setValue(0,0,11);  d.getValue(0,0) : " << d.getValue(0,0) << ENDL;
  d.Null();COUT << "d.Null() : " << d << ENDL;
  d.resetTo(2);COUT << "d.resetTo(2) : " << d << ENDL;
  T*Dd=d.getValues();
  COUT << "getValues() : " << Dd[0] << " " << Dd[1] << " " << Dd[2] << ENDL;
  const T*E=d;
  COUT << "(const T*) d : " << E[0] << " " << E[1] << " " << E[2] << ENDL;
  d.setValues(A);COUT << "d.setValues(T[1,2,3...]) : " << d << ENDL;
//   d.setValues(Af);COUT << "d.setValues(float[1,2,3...]) : " << d << ENDL;
//   d.setValues(Ad);COUT << "d.setValues(double[1,2,3...]) : " << d << ENDL;
//   d.setValues(Ai);COUT << "d.setValues(int[1,2,3...]) : " << d << ENDL;

  COUT << "Math" << ENDL;
  COUT << "d.Minus(a) : " << d.Minus(a) << ENDL;
  COUT << "d.Add(a,b) : " << d.Add(a,b) << ENDL;
  COUT << "d.Sub(a,b) : " << d.Sub(a,b) << ENDL;
  COUT << "d.Mul(a,2) : " << d.Mul(a,(T)2) << ENDL;
  COUT << "d.Mul(2,a) : " << d.Mul((T)2,a) << ENDL;
  COUT << "d.Div(a,2) : " << d.Div(a,(T)2) << ENDL;
  COUT << "d.Div(a,2) : " << d.Div(a,(T)2) << ENDL;

  COUT << "+a : " << +a << ENDL;
  COUT << "-a : " << -a << ENDL;
  COUT << "a+b : " << a+b << ENDL;
  COUT << "a-b : " << a-b << ENDL;
  COUT << "a*2 : " << a*((T)2) << ENDL;
  COUT << "2*a : " << ((T)2)*a << ENDL;
  COUT << "a/2 : " << a/((T)2) << ENDL;

  d.copy(b);d+=a;COUT << "d=b; d+=a : " << d << ENDL;
  d.copy(b);d-=a;COUT << "d=b; d-=a : " << d << ENDL;
  d.copy(b);d*=(T)2;COUT << "d=b; d*=2 : " << d << ENDL;
  d.copy(b);d/=(T)2;COUT << "d=b; d/=2 : " << d << ENDL;

  d.copy(b);COUT << "d=b; d==b : " << (d==b) << ENDL;
  d.copy(b);COUT << "d=b; d!=b : " << (d!=b) << ENDL;

  COUT << "dot(a,b) : " << dot(a,b) << ENDL;
  COUT << "DotProd(a,b) : " << DotProd(a,b) << ENDL;
  COUT << "norm2(a) : " << a.norm2() << ENDL;
  COUT << "norm(a) : " << a.norm() << ENDL;
  COUT << "Norm2(a) : " << a.Norm2() << ENDL;
  COUT << "Norm(a) : " << a.Norm() << ENDL;
  COUT << "NormL1(a) : " << a.NormL1() << ENDL;
  COUT << "NormLinf(a) : " << a.NormLinf() << ENDL;
  COUT << "Dist2(a,b) : " << Dist2(a,b) << ENDL;
  COUT << "Dist(a,b) : " << Dist(a,b) << ENDL;
  COUT << "DistL1(a,b) : " << DistL1(a,b) << ENDL;
  COUT << "DistLinf(a,b) : " << DistLinf(a,b) << ENDL;

  d.resetTo(2); d.min(a); COUT << "d.resetTo(2); d.min(a) : " << d << ENDL;
  d.resetTo(2); d.max(a); COUT << "d.resetTo(2); d.max(a) : " << d << ENDL;

  d.Rnd(2); COUT << "d.Rnd(2) : " << d << ENDL;

  Matrix4x4<T> m;COUT << "m.Trp(d) : " << m.Trp(d) << ENDL;
  d.transpose();COUT << "d.transpose() : " << d << ENDL;
  COUT << "d.getTransposed() : " << d.getTransposed() << ENDL;

  {Matrix4x4<T> m;Matrix4x4<T> n;COUT << "Matrix4x4<T>*Matrix4x4<T> : " << m*n << "  Mul : " << d.Mul(m,n) << ENDL;}

  {d.Rnd();Sym<4,T> S1;Sym<4,T> S2;d.SymAction(S1,S2);COUT << "SymAction" << ENDL;}
  {d.Rnd();Matrix4x4<T> S1;Vec4<T>W;Matrix4x4<T> S2;d.SVD(S1,W,S2);COUT << "SVD" << ENDL;}
  {d.Rnd();Matrix4x4<T> S;d.PseudoInv(S);COUT << "PseudoInv" << ENDL;}

  {d.Rnd();d.Inv(d);COUT << "Inv" << ENDL;}
  {d.Rnd();COUT << "getInverse" << d.getInverse() << ENDL;}
  {d.Rnd();d.invert();COUT << "invert" << ENDL;}

  {d.Rnd();COUT << "Det" << d.Det() << ENDL;}
  {d.Rnd();COUT << "Trace" << d.Trace() << ENDL;}

  {d.Rnd();COUT << "Id" << d.Id() << ENDL;}
  {d.Rnd();Vec4<T>v;COUT << "Diag" << d.Diag(v) << ENDL;}
  {d.Rnd();Vec4<T>v;COUT << "TensorProduct" << d.TensorProduct(v,v) << ENDL;}

  {Sym<4,T> m;m.Rnd();d=m;COUT << "operator=Sym<4,T>" << d << "=" << m << ENDL;}
  {Sym<4,T> m;m.Rnd();Matrix4x4<T> d(m);COUT << "Constructeur(Sym<4,T>)" << d << "=" << m << ENDL;}

  d=Matrix3x3<T>(VIM_INIT);COUT << "operator=Matrix3x3<T>" << d << ENDL;
  d=rot;COUT << "operator=Rotation3D" << d << ENDL;
  d=Vec3<T>(VIM_INIT);COUT << "operator=Vec3<T>" << d << ENDL;

  COUT << "Matrix4x4<T>*Vec3<T>" << d*Vec3<T>(VIM_INIT) << ENDL;

  COUT << ENDL;
}


} // namespace yav

#endif
