#ifndef ROTATION2D_H
#define ROTATION2D_H


#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "linearAlgebra/Rotation2D.h"
#ifndef __GNUC__
#ifndef determinantThreshold
#define determinantThreshold 1e-6
#endif
#endif
#include <math.h>
#include <basics/Exception.h>
#include <basics/TclString.h>
#include <linearAlgebra/Matrix2x2.h>

#ifdef WIN32
#include "basics/export.h"
#endif


namespace yav {


/*! \class Rotation2D Rotation2D.h linearAlgebra/Rotation2D.h
  \ingroup liblinearAlgebra
  template for a 2x2 rotation matrix 
*/
class Rotation2D : public Matrix2x2<double> {
 public:
  /** @name Exception class */
  //@{
  /*! \class NonRotationMatrix Rotation2D.h linearAlgebra/Rotation2D.h
    \ingroup liblinearAlgebra
    attempt to initialize a rotation from a non rotation element
  */
  class NonRotationMatrix : public Exception {
  public:
    NonRotationMatrix(const std::string& from) :
      Exception(from,
                "Trying to define a rotation matrix with a determinant different from 1"){}
  };
  //@}

  public:
#ifdef __GNUC__
  /// constant that determines if a 2x2 matrix if a rotation matrix
  static const double determinantThreshold;
#endif

  /** @name Constructors */
  //@{
  /// Default constructor creates the identity matrix
  Rotation2D() : Matrix2x2<double>(VIM_NOINIT)
    { Matrix2x2<double>::resetToIdentity(); }
  /** Initialization from an 2x2 array
      @param matrix values' array
      @exception NonRotationMatrix the determinant is not close to 1 enough */
  Rotation2D(double matrix[2][2]) throw(NonRotationMatrix);
  /** Initialization from 2 column vectors
      @param v1 first column vector
      @param v2 second column vector
      @exception NonRotationMatrix the determinant is not close to 1 enough */
  Rotation2D(const Vec2<double> &v1,const Vec2<double> &v2)
    throw(NonRotationMatrix);
  /** copy constructor
      @param r matrix to copy */
  Rotation2D(const Rotation2D &r) : Matrix2x2<double>(VIM_NOINIT)
    {  Matrix2x2<double>::copy(r); }
  /** creates a rotation matrix from rotation angle
      @param angle rotation angle */
  Rotation2D(double angle);
  //@}

  /** @name Destructor */
  //@{
  /// destructor
  virtual ~Rotation2D(){}
  //@}

  /// returns rotation angle
  double getRotationAngle() const;
};


} // namespace yav


#endif
