// General comments : see Vec.h

#ifndef _SYM_H
#define _SYM_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#ifndef WIN32
#pragma interface "linearAlgebra/Sym.h"
#endif
#include "linearAlgebra/Mat.h"



namespace yav {


/** @name Vectorial redefined friend functions */
//@{
/** unary +
    @param p vector
    @return +p */
template <Index D, class T> vec_Inline Sym<D, T> operator+(const Sym<D, T> &p) throw()
  {return p;}
/** unary -
    @param p vector
    @return -p */
template <Index D, class T> vec_Inline Sym<D, T> operator-(const Sym<D, T> &p) throw()
  {return Sym<D, T>(VIM_NOINIT).Minus(p);}
/** vector sum
    @param p first vector
    @param q second vector
    @return p + q */
template <Index D, class T> vec_Inline Sym<D, T> operator+(const Sym<D, T> &p, const Sym <D, T> &q) throw()
  {return Sym<D, T>(VIM_NOINIT).Add(p,q);}
/** vector substraction
    @param p first vector
    @param q second vector
    @return p - q */
template <Index D, class T> vec_Inline Sym<D, T> operator-(const Sym<D, T> &p, const Sym <D, T> &q) throw()
  {return Sym<D, T>(VIM_NOINIT).Sub(p,q);}
/** vector multiplication by scalar
    @param p vector
    @param a scalar
    @return ap */
template <Index D, class T> vec_Inline Sym<D, T> operator*(const Sym<D, T> &p, const T a) throw()
  {return Sym<D, T>(VIM_NOINIT).Mul(p,a);}
/** vector scalar multiplication
    @param a scalar
    @param p vector
    @return ap */
template <Index D, class T> vec_Inline Sym<D, T> operator*(const T a, const Sym<D, T> &p) throw()
  {return Sym<D, T>(VIM_NOINIT).Mul(p,a);}
/** vector scalar division
    @param p vector
    @param a scalar
    @return p / a */
template <Index D, class T> vec_Inline Sym<D, T> operator/(const Sym<D, T> &p, const T a) throw(ErrDivZero)
  {return Sym<D, T>(VIM_NOINIT).Div(p,a);}
/** dot product
    @param u first vector
    @param v second vector
    @return u.v */
template <Index D, class T> vec_Inline T dot(const Sym<D, T>& u, const Sym<D, T> &v) throw()
  {return Sym<D, T>::dot(u,v);}
/// Square Euclidian Distance
template <Index D, class T> vec_Inline T dist2( const Sym<D, T>& u, const Sym<D, T>& v) throw()
  {return Sym<D, T>::dist2(u,v);}
/// Euclidian Distance
template <Index D, class T> vec_Inline T dist( const Sym<D, T>& u, const Sym<D, T>& v) throw()
  {return Sym<D, T>::dist(u,v);}
/// L1 Distance
template <Index D, class T> vec_Inline T distL1( const Sym<D, T>& u, const Sym<D, T>& v) throw()
  {return Sym<D, T>::distL1(u,v);}
/// Linf Distance
template <Index D, class T> vec_Inline T distLinf( const Sym<D, T>& u, const Sym<D, T>& v) throw()
  {return Sym<D, T>::distLinf(u,v);}
/** dot product
    @param p first vector
    @param q second vector
    @return p.q */
template <Index D, class T> vec_Inline T DotProd(const Sym<D, T> &p, const Sym<D, T> &q) throw()
  {return Sym<D, T>::dot(p,q);}
/// Square Euclidian Distance
template <Index D, class T> vec_Inline T Dist2( const Sym<D, T>& u, const Sym<D, T>& v) throw()
  {return Sym<D, T>::dist2(u,v);}
/// Euclidian Distance
template <Index D, class T> vec_Inline T Dist( const Sym<D, T>& u, const Sym<D, T>& v) throw()
  {return Sym<D, T>::dist(u,v);}
/// L1 Distance
template <Index D, class T> vec_Inline T DistL1( const Sym<D, T>& u, const Sym<D, T>& v) throw()
  {return Sym<D, T>::distL1(u,v);}
/// Linf Distance
template <Index D, class T> vec_Inline T DistLinf( const Sym<D, T>& u, const Sym<D, T>& v) throw()
  {return Sym<D, T>::distLinf(u,v);}
/** operator == */
template <Index D, class T> vec_Inline bool operator == (const Sym<D, T>& u, const Sym<D, T> &v) throw()
  {return Sym<D, T>::isEqual(u,v);}
/** operator != */
template <Index D, class T> vec_Inline bool operator != (const Sym<D, T>& u, const Sym<D, T> &v) throw()
  {return Sym<D, T>::isDifferent(u,v);}
/** operator < */
template <Index D, class T> vec_Inline bool operator < (const Sym<D, T>& u, const Sym<D, T> &v) throw()
  {return Sym<D, T>::isSmaller(u,v);}
/** operator > */
template <Index D, class T> vec_Inline bool operator > (const Sym<D, T>& u, const Sym<D, T> &v) throw()
  {return Sym<D, T>::isGreater(u,v);}
/** operator <= */
template <Index D, class T> vec_Inline bool operator <= (const Sym<D, T>& u, const Sym<D, T> &v) throw()
  {return Sym<D, T>::isSmallerOrEqual(u,v);}
/** operator >= */
template <Index D, class T> vec_Inline bool operator >= (const Sym<D, T>& u, const Sym<D, T> &v) throw()
  {return Sym<D, T>::isGreaterOrEqual(u,v);}
//@}

template <Index D, class T> vec_Inline Sym<D,T> operator*(const Sym<D,T> &m1, const Sym<D,T> &m2) throw()
  {return Sym<D,T>(VIM_NOINIT).Mul(m1,m2);}

/** operator << for symetric matrix of dimension D and type T */
template<Index D, class T> STD_OSTREAM& operator<<(STD_OSTREAM& os, const Sym<D,T>& W) throw();

/** \class Sym
    \ingroup liblinearAlgebra
    template for a symetric matrix of dimension D and type T */
template<Index D, class T>
class Sym {
 public:
  /** @name Constructors */
  //@{
  //  /** @name Vectorial redefined constructors */
  //  //@{
  //  /** default constructor - initializes to 0 if UNINIT_ZERO not defined */
  //  vec_Inline Vect() throw() : _Vect<D,T>()
  //    {}
  //  /** constructor - initializes to 0 if b==VIM_INIT */
  //  explicit vec_Inline Vect(VectInitMode b) throw() : _Vect<D,T>(b)
  //    {}
  //  //@}
  vec_Inline Sym() throw()
    {
#ifndef UNINIT_ZERO
      resetTo(0);
#endif
    }
  /// Constructor : initializes to 0 if b==VIM_INIT
  vec_Inline Sym(VectInitMode b) throw()
    {if (b==VIM_INIT) resetTo(0);}

#ifndef WIN32
  /** copy constructor
      @param v vector to copy */
  vec_Inline Sym(const Sym<D,T>& v) throw()
    {copy(v);}
  /** operator =
      @param v vector to copy
      @return this */
  vec_Inline Sym<D, T>& operator=(const Sym<D,T>& v) throw()
    {copy(v);return *this;}
#endif
  /** copy constructor with cast
      @param v vector to copy */
  template<class T2> vec_Inline Sym(const Sym<D,T2>& v) throw()
    {copy(v);}
  /** operator = with cast
      @param v vector to copy
      @return this */
  template<class T2> vec_Inline Sym<D, T>& operator=(const Sym<D,T2>& v) throw()
    {copy(v);return *this;}

  /// Intialize with variance sig2 on the diagonal
  explicit Sym(const T sig2) throw()
    {resetToIdentity(sig2);}

  /** creates a diagonal matrix using a vector for diagonal terms
      @param v vector */
  explicit Sym(const Vect<D,T>& v) throw()
    {diag(v);}

  /** creates a diagonal matrix using a vector x as x.xtrp
      @param v vector
      @param a fake parameter */
  explicit Sym(const Vect<D,T>& v, int a) throw()
    {autoCorrel(v);}

  /** creates a diagonal matrix using a matrix m as m.mtrp
      @param m vector
      @param a fake parameter */
  template <Index K>
  explicit Sym(const Mat<D,K,T>& m, int a) throw()
    {autoCorrel(m);}

  /** Creates a symetric matrix from a square matrice
      @param m square matrix */
  explicit Sym(const Mat<D,D,T>& m) throw()
    {Mat2Sym<D,T>(m,*this);}

  /** Set a symetric matrix from a square matrice
      @param m square matrix */
  Sym<D,T>& operator=(const Mat<D,D,T>& m) throw()
    {Mat2Sym<D,T>(m,*this);return *this;}
  //@}

  vec_Inline T getVal(int i) const
    {return val[i];}
  vec_Inline void setVal(int i, T v)
    {val[i]=v;}

  /** @name Data members access */
  //@{
  //   /** copy as a member function
  //       @param v vector to copy */
  //  void copy(const Sym<D,T>& v) throw()
  //     {for (Index i=0; i<(D*(D+1))/2; i++) val[i] = v.getVal(i);}
  /** copy with cast as a member function
      @param v vector to copy */
  template<class T2> vec_Inline void copy(const Sym<D,T2>& v) throw()
    {for (Index i=0; i<(D*(D+1))/2; i++) val[i] = (T) v.getVal(i);}

  /** set each element of this vector to value
    @param value elements new value */
  void resetTo(T value) throw()
    {for (Index i=0; i<(D*(D+1))/2; i++) val[i]=value;}
  /// set this vector to the null vector
  vec_Inline void resetToNull(void) throw()
    {resetTo(0);}

  //   /** setValues with a T[(D*(D+1))/2] array */
  //   vec_Inline void setValues(const T v[(D*(D+1))/2]) throw()
  //     { for (Index i=0; i<(D*(D+1))/2; i++) val[i]=v[i]; }
  /** setValues with a T2[(D*(D+1))/2] array */
  template<class T2> vec_Inline void setValues(const T2 v[(D*(D+1))/2]) throw()
    {for (Index i=0; i<(D*(D+1))/2; i++) val[i]=(T) v[i];}

  /** returns a T* (usefull for some extern functions, like OGL ones) */
  vec_Inline T* getValues(void) throw()
    {return val;}
  /** returns a const T* (usefull for some extern functions, like OGL ones) */
  vec_Inline const T* getValues(void) const throw()
    {return val;}

  /** cast vector in a T*
   */
  vec_Inline operator const T*() const throw()
    {return (const T*) val;}

  /** access a matrix element
      @param i element line
      @param j element column
      @return this[i][j] */
  T& operator()(int i, int j) throw(ErrRange);
  /** access a matrix element (const version)
      @param i element line
      @param j element column
      @return this[i][j] */
  T operator()(int i, int j) const throw(ErrRange);

  /// set this matrix to the identity matrix
  void resetToIdentity(const T v=1.0) throw();

  /// set this matrix to the identity matrix
  Sym<D,T>& Id(const T v=1.0) throw()
    {resetToIdentity(v);return *this;}

  /// Set to Diag( vec )
  void diag( const Vect<D,T>& v) throw();

  /// Set to Diag( vec )
  Sym<D,T>& Diag( const Vect<D,T>& v) throw()
    {diag(v);return *this;}

  //friend void Sym2Mat FRIEND_TEMPLATE_ARG( const Sym<D,T>& W, Mat<D,D,T>& M) throw();
  //friend void Mat2Sym FRIEND_TEMPLATE_ARG( const Mat<D,D,T>& M, Sym<D,T>& W) throw();
  //@}

  /** @name mathematical operators */
  //@{
  /** squared norm
      @return ||this||^2 */
  vec_Inline T norm2(void) const throw()
    {return vec_norm2(val,(D*(D+1))/2);}
  /** norm
    @return ||this|| */
  vec_Inline T norm(void) const throw()
    {return vec_norm(val,(D*(D+1))/2);}
  /** L1 norm
      @return sum abs(v_i) */
  vec_Inline T normL1(void) const throw()
    {return vec_normL1(val,(D*(D+1))/2);}
  /** Linf norm
    @return max abs(v_i) */
  vec_Inline T normLinf(void) const throw()
    {return vec_normLinf(val,(D*(D+1))/2);}
  /** squared norm
      @return ||this||^2 */
  vec_Inline T Norm2(void) const throw()
    {return vec_norm2(val,(D*(D+1))/2);}
  /** norm
    @return ||this|| */
  vec_Inline T Norm(void) const throw()
    {return vec_norm(val,(D*(D+1))/2);}
  /** L1 norm
      @return sum abs(v_i) */
  vec_Inline T NormL1(void) const throw()
    {return vec_normL1(val,(D*(D+1))/2);}
  /** Linf norm
    @return max abs(v_i) */
  vec_Inline T NormLinf(void) const throw()
    {return vec_normLinf(val,(D*(D+1))/2);}
  
  /** add q to this vector */
  vec_Inline void add(const Sym<D, T> &q) throw()
    {vec_add(val,(D*(D+1))/2,q.val);}
  /** substract q to this vector */
  vec_Inline void sub(const Sym<D, T> &q) throw()
    {vec_sub(val,(D*(D+1))/2,q.val);}
  /** set each coordinate of this vector to product of this and q same coordinate */
  vec_Inline void cmul(const Sym<D, T> &q) throw()
    {vec_cmul(val,(D*(D+1))/2,q.val);}
  /** multiply this vector with a */
  vec_Inline void mul(const T a) throw()
    {vec_mul(val,(D*(D+1))/2,a);}
  /** divide this vector with a */
  vec_Inline void div(const T a) throw(ErrDivZero)
    {vec_div(val,(D*(D+1))/2,a);}
  
  /** set this vector to -p */
  vec_Inline void opposite(const Sym<D, T> &p) throw()
    {vec_opposite(val,(D*(D+1))/2,p.val);}
  /** set this vector to p+q */
  vec_Inline void add(const Sym<D, T> &p, const Sym<D, T> &q) throw()
    {vec_add(val,(D*(D+1))/2,p.val,q.val);}
  /** set this vector to p-q */
  vec_Inline void sub(const Sym<D, T> &p, const Sym<D, T> &q) throw()
    {vec_sub(val,(D*(D+1))/2,p.val,q.val);}
  /** set each coordinate of this vector to product of p and q same coordinate */
  vec_Inline void cmul(const Sym<D, T> &p, const Sym<D, T> &q) throw()
    {vec_cmul(val,(D*(D+1))/2,p.val,q.val);}
  /** set this vector to p*a */
  vec_Inline void mul(const Sym<D, T> &p, const T a) throw()
    {vec_mul(val,(D*(D+1))/2,p.val,a);}
  /** set this vector to p/a */
  void div(const Sym<D, T> &p, const T a) throw(ErrDivZero)
    {vec_div(val,(D*(D+1))/2,p.val,a);}
  
  /** randomize vector */
  vec_Inline void rnd(T min = -1., T max = 1.) throw()
    {vec_rnd(val,(D*(D+1))/2,min,max);}
  /** randomize vector : isotropic gaussian centred noise */
  vec_Inline void rndgaussian(T standard_deviation = 1.) throw()
    {vec_rndgaussian(val,(D*(D+1))/2,standard_deviation);}
  
  /** takes the X,Y or Z value of p if they are less than its own X,Y,Z coordinates */
  vec_Inline void min(const Sym<D, T> &p) throw()
    {vec_min(val,(D*(D+1))/2,p.val);}
  /** takes the X,Y or Z value of p if they are greater than its own X,Y,Z coordinates */
  vec_Inline void max(const Sym<D, T> &p) throw()
    {vec_max(val,(D*(D+1))/2,p.val);}
  
  /** dot product
      @param p first vector
      @param q second vector
      @return p.q */
  vec_Inline static T dot(const Sym<D, T> &p, const Sym<D, T> &q) throw()
    {return vec_dot((D*(D+1))/2,p.val,q.val);}
  /// Square Euclidian Distance
  vec_Inline static T dist2( const Sym<D, T>& p, const Sym<D, T>& q) throw()
    {return vec_dist2((D*(D+1))/2,p.val,q.val);}
  /// Euclidian Distance
  vec_Inline static T dist( const Sym<D, T>& p, const Sym<D, T>& q) throw()
    {return vec_dist((D*(D+1))/2,p.val,q.val);}
  /// L1 Distance
  vec_Inline static T distL1( const Sym<D, T>& p, const Sym<D, T>& q) throw()
    {return vec_distL1((D*(D+1))/2,p.val,q.val);}
  /// Linf Distance
  vec_Inline static T distLinf( const Sym<D, T>& p, const Sym<D, T>& q) throw()
    {return vec_distLinf((D*(D+1))/2,p.val,q.val);}
  
  /// operator ==
  vec_Inline static bool isEqual(const Sym<D, T> &p, const Sym<D, T> &q) throw()
    {return vec_isEqual((D*(D+1))/2,p.val,q.val);}
  /// operator !=
  vec_Inline static bool isDifferent(const Sym<D, T> &p, const Sym<D, T> &q) throw()
    {return vec_isDifferent((D*(D+1))/2,p.val,q.val);}
  /// operator <
  vec_Inline static bool isSmaller(const Sym<D, T> &p, const Sym<D, T> &q) throw()
    {return vec_isSmaller((D*(D+1))/2,p.val,q.val);}
  /// operator >
  vec_Inline static bool isGreater(const Sym<D, T> &p, const Sym<D, T> &q) throw()
    {return vec_isSmaller((D*(D+1))/2,q.val,p.val);}
  /// operator <=
  vec_Inline static bool isSmallerOrEqual(const Sym<D, T> &p, const Sym<D, T> &q) throw()
    {return !vec_isSmaller((D*(D+1))/2,q.val,p.val);}
  /// operator >=
  vec_Inline static bool isGreaterOrEqual(const Sym<D, T> &p, const Sym<D, T> &q) throw()
    {return !vec_isSmaller((D*(D+1))/2,p.val,q.val);}
  /** @name Vectorial redefined menber functions */
  //@{
  /** set this vector to p+q
      @return this = p + q */
  vec_Inline Sym<D, T>& Minus(const Sym<D, T> &p) throw()
    {vec_opposite(val,(D*(D+1))/2,p.val);return *this;}
  /** set this vector to p+q
      @return this = p + q */
  vec_Inline Sym<D, T>& Add(const Sym<D, T> &p, const Sym<D, T> &q) throw()
    {vec_add(val,(D*(D+1))/2,p.val,q.val);return *this;}
  /** set this vector to p-q
      @return this = p - q */
  vec_Inline Sym<D, T>& Sub(const Sym<D, T> &p, const Sym<D, T> &q) throw()
    {vec_sub(val,(D*(D+1))/2,p.val,q.val);return *this;}
  /** set each coordinate of this vector to product of this and p same coordinate
      @return this = this.*q */
  vec_Inline Sym<D, T>& CMul(const Sym<D, T> &q) throw()
    {vec_cmul(val,(D*(D+1))/2,q.val);return *this;}
  /** set each coordinate of this vector to product of q and p same coordinate
      @return this = p.*q */
  vec_Inline Sym<D, T>& CMul(const Sym<D, T> &p, const Sym<D, T> &q) throw()
    {vec_cmul(val,(D*(D+1))/2,p.val,q.val);return *this;}
  /** set this vector to p*a
      @return this = p*a */
  vec_Inline Sym<D, T>& Mul(const Sym<D, T> &p, const T a) throw()
    {vec_mul(val,(D*(D+1))/2,p.val,a);return *this;}
  /** set this vector to p*a
      @return this = p*a */
  vec_Inline Sym<D, T>& Mul(const T a, const Sym<D, T> &p) throw()
    {vec_mul(val,(D*(D+1))/2,p.val,a);return *this;}
  /** set this vector to p/a
      @return this = p/a */
  vec_Inline Sym<D, T>& Div(const Sym<D, T> &p, const T a) throw(ErrDivZero)
    {vec_div(val,(D*(D+1))/2,p.val,a);return *this;}
  /** adds vector q to vector p
      @param q vector to add
      @return p + q */
  vec_Inline Sym<D, T>& operator+=(const Sym<D, T> &q) throw()
    {vec_add(val,(D*(D+1))/2,q.val);return *this;}
  /** substracts vector q from vector p
      @param q vector to substract
      @return p - q */
  vec_Inline Sym<D, T>& operator-=(const Sym<D, T> &q) throw()
    {vec_sub(val,(D*(D+1))/2,q.val);return *this;}
  /** multiplies vector p by scalar a
      @param a scalar
      @return ap */
  vec_Inline Sym<D, T>& operator*=(const T a) throw()
    {vec_mul(val,(D*(D+1))/2,a);return *this;}
  /** divides vector p by scalar a
      @param a scalar
      @return p / a */
  vec_Inline Sym<D, T>& operator/=(const T a) throw(ErrDivZero)
    {vec_div(val,(D*(D+1))/2,a);return *this;}
  /** randomize vector
      @return this */
  vec_Inline Sym<D, T>& Rnd(T min = -1., T max = 1.) throw()
    {vec_rnd(val,(D*(D+1))/2,min,max);return *this;}
  /** randomize vector : isotropic gaussian centred noise
      @return this */
  vec_Inline Sym<D, T>& RndGaussian(T standard_deviation = 1.) throw()
    {vec_rndgaussian(val,(D*(D+1))/2,standard_deviation);return *this;}
  /** takes the X,Y or Z value of p if they are less than its own X,Y,Z coordinates
      @return _Vect<(D*(D+1))/2,T>(min(x,p.x),min(y,p.y),...) */
  vec_Inline Sym<D, T>& Min(const Sym<D, T>& p) throw()
    {vec_min(val,(D*(D+1))/2,p.val);return *this;}
  /** takes the X,Y or Z value of p if they are greater than its own X,Y,Z coordinates
      @return _Vect<(D*(D+1))/2,T>(max(x,p.x),max(y,p.y),...) */
  vec_Inline Sym<D, T>& Max(const Sym<D, T>& p) throw()
    {vec_max(val,(D*(D+1))/2,p.val);return *this;}
  /** set each element of this vector to value
      @param v value
      @return this */
  vec_Inline Sym<D, T>& Reset(T v) throw()
    {resetTo(v);return *this;}
  /** set this vector to null vector
      @return this */
  vec_Inline Sym<D, T>& Null() throw()
    {resetTo(0);return *this;}
  /** set this vector to mean of a given list of vectors
      @return this */
  template<class InputIterator>
  vec_Inline Sym<D, T>& Mean(InputIterator first, InputIterator last) throw()
    {mean(first,last);return *this;}
  //@}

  /** diagonalisation,  eigen-values and -vectors extraction (Jacobi method - NRC p364)
      @param EVect eigenVecttors eigen vectors returned
      @param EVal eigen values returned
      @return number of iterations, throw ErrConvergence if did not converge */
  int Diagonalize(Matrix<D,T>& EVect, Vect<D,T>& EVal) const throw(ErrConvergence);
  /** diagonalisation,  eigen-values and -vectors extraction (Jacobi method - NRC p364)
      Eigen values and vectors are sorted accordingly to the decreasing eigen
      values.
      @param EVect eigenVecttors eigen vectors returned
      @param EVal eigen values returned
      @return number of iterations, throw ErrConvergence if the method does not converge */
  int DiagonalizeSort(Matrix<D,T>& EVect, Vect<D,T>& EVal) const throw(ErrConvergence);

  /// Sym - Vecttor multiplication function. Returns vtrp.W.v
  T SqNorm( const Vect<D,T>& v) const throw();

  /// Trace of matrix
  T Trace() const throw();

  /// Determinant
  T Det() const throw(ErrConvergence);

  /// Autocorrelation : set *this = x.xtrp
  void autoCorrel( const Vect<D,T>& x) throw();
  /// Autocorrelation : set *this = m.mtrp
  template <Index K>
  void autoCorrel( const Mat<D,K,T>& m) throw()
    {//TODO:optimize
      for(Index i=0; i< D ; i++) {
	for(Index j=0; j <= i; j++) {
	  T v=0;
	  for (Index k=0; k<K; k++) v+=m(i,k) * m(j,k);
	  val[ (i*(i+1))/2 + j ] = v;
	}
      }
    }


  /// Autocorrelation : set *this = x.xtrp
  Sym<D,T>& AutoCorrel( const Vect<D,T>& x) throw()
    {autoCorrel(x);return *this;}
  /// Autocorrelation : set *this = m.mtrp
  template <Index K>
  Sym<D,T>& AutoCorrel( const Mat<D,K,T>& m) throw()
    {autoCorrel(m);return *this;}

  /// DiProd : return x.*this.xtrp
  T DiProd(const Vect<D,T>& x) throw()
    {T res;((const Mat<1,D,T>&)x).SymAction(*this,(Sym<1,T>&)res);return res;}

  /// Matrix multiplication
  void mulsym(const Sym<D,T> &m1, const Sym<D,T> &m2) throw();

  /** set matrix to Pseudo inverse of W (menber method)
      @param W matrix to invert
      @return singular number of singular values */
  int inv( const Sym<D,T>&  W ) throw(ErrConvergence);
  /** set matrix to Pseudo inverse of W (menber method)
      @param W matrix to invert
      @return *this */
  Sym<D,T>& Inv( const Sym<D,T>&  W ) throw(ErrConvergence)
    {inv(W);return *this;}

  /** get this matrix inverse matrix (this matrix is NOT modifyed)
      @return inverse matrix */
  vec_Inline Sym<D,T> getInverse(void) const throw(ErrMatSingular)
    {return Sym<D,T>(VIM_NOINIT).Inv(*this);}
  /// invert this matrix (this matrix is modified)
  vec_Inline void invert(void) throw(ErrMatSingular)
    {this->Inv(*this);}

  /// Matrix multiplication
  vec_Inline Sym<D,T>& operator*=(const Sym<D,T>& m) throw()
    {mulsym(*this,m);return *this;}
  /// Matrix multiplication
  Sym<D,T>& Mul(const Sym<D,T> &m1, const Sym<D,T> &m2) throw()
    {mulsym(m1,m2);return *this;}

  /** @name I/O operations */
  //@{
  /** writes matrix on outstream as "((x00 x01 x02) (x10 x11 x12) (x20 x21 x22))"
      param s outstream
      param m matrix
      return s */
  //friend STD_OSTREAM& operator<< FRIEND_TEMPLATE_ARG(STD_OSTREAM &s, const Sym<D,T> &m) throw();
  //@}

  static void Test() throw();
 private:
  /// Matrix multiplication
  void private_mulsym(const Sym<D,T> &m1, const Sym<D,T> &m2) throw();
protected:
  /// vector data
  T val[(D*(D+1))/2];
};

} // namespace yav


/*][** NODOC ****************************************************************/
/****************************************************************************/

/*]]**************************************************************************/



namespace yav {

template<Index D, class T> vec_Inline void Sym2Mat( const Sym<D,T>& W, Mat<D,D,T>& M) throw()
{
  for (Index j=0; j<D; j++)
    for (Index i=0; i<=j; i++)
      M(i,j) = M(j,i) = W(i,j);
}

template<Index D, class T> vec_Inline void Mat2Sym( const Mat<D,D,T>& M, Sym<D,T>& W) throw()
{
  for (Index j=0; j<D; j++)
    for (Index i=0; i <= j; i++)
      W(i,j) = (M(i,j) + M(j,i))/2.;
}

template<Index D, class T> vec_Inline T& Sym<D,T>::operator()(int i, int j) throw(ErrRange)
{
#ifdef RANGE_CHECK
  if ( (i<0) || (j<0) || (i >= (int)D ) || (j >= (int)D) ){
    throw ErrRange("T& Sym::operator()");
  }
#endif
  if (i >= j)
    return val[ (i*(i+1))/2 + j ];
  else
    return val[ (j*(j+1))/2 + i ];
}

template<Index D, class T> vec_Inline T Sym<D,T>::operator()(int i, int j) const throw(ErrRange)
{
#ifdef RANGE_CHECK
  if ( (i<0) || (j<0) || (i >= (int)D ) || (j >= (int)D) ){
    throw ErrRange("T Sym::operator() const");
  }
#endif
  if (i >= j)
    return val[ (i*(i+1))/2 + j ];
  else
    return val[ (j*(j+1))/2 + i ];
}

template <Index D, class T> vec_Inline void Sym<D,T>::resetToIdentity(const T v) throw()
{
  T*vij=val;
  for ( Index i=0; i< D; i++) {
    for ( Index j=0; j<i; j++)
      *vij++=0;
    *vij++=v;
  }
}

template <Index D, class T> vec_Inline void Sym<D,T>::diag(const Vect<D,T>& v) throw()
{
  T*vij=val;
  for ( Index i=0; i< D; i++) {
    for ( Index j=0; j<i; j++)
      *vij++=0;
    *vij++=v[i];
  }
}

// mathematical operations
// Adapted from NRC
template<class T> vec_Inline void ROTATE(T*val, Index D, Index i, Index j, Index k, Index l, T s, T tau) throw()
{
  T g = val[i*D + j];
  T h = val[k*D + l];

  val[i*D + j] = g-s*(h+g*tau);
  val[k*D + l] = h+s*(g-h*tau);
}

template<class T> int private_Diagonalize(T*A, Index D, T* EVect, T*  EVal, T*B, T*Z) throw(ErrConvergence)
{
  T tresh,theta,tau,t,sm,s,h,g,c;

  const T Precision_Factor = 1e2;
  const Index MAXIT = 50;
  T Small_Val;Xlinalg_Small_Val<T>(Small_Val);
  // const T tresh_min = Small_Val*Small_Val;

  {
    for (Index ip=0;ip<D;ip++) {
      B[ip]=EVal[ip]=A[ip*D+ip];
      Z[ip]=0.0;
    }
  }
  int nrot=0;
  for (Index i=0;i<MAXIT;i++) {
    sm=0.0;
    {
      for (Index ip=0; ip+1 < D; ip++) {
        for (Index iq=ip+1; iq<D; iq++)
          sm += (T) ::fabs( A[ip*D+iq] );
      }
    }
    if ( sm < Small_Val ) // Addon. Was: if ( (T) sm == (T) 0.0)
      return nrot;

    if (i < 4)
      tresh=0.2*sm/(D*D);
    else
      tresh= 0.0; //tresh_min; // Addon. Was tresh = 0.0;
    {
      for (Index ip=0;ip+1 < D;ip++) {
        T& Dip = EVal[ip];
        for (Index iq=ip+1;iq<D;iq++) {
          T& Diq = EVal[iq];
          T& Apq = A[ip*D+iq];

          g= Precision_Factor * ((T) ::fabs( Apq ));
          // COUT << "g: " << g << "    Apq: " << Apq <<
          //  "    Dip: " << Dip << "    Diq: " << Diq << ENDL;
          if ((i >= 4)
              && (g < ::fabs( Dip )* Small_Val )
              && (g < ::fabs( Diq )* Small_Val )
//            && ( ((T) ::fabs( Dip ) + g) == ((T) ::fabs( Dip )) )
//            && ( ((T) ::fabs( Diq ) + g) == ((T) ::fabs( Diq )) )
              )
            Apq = 0.0;
          else
            if ( ::fabs( Apq ) > tresh) {
              h= Diq - Dip;
              // if (  ((T) ::fabs(h)) + g  == ((T) ::fabs(h)) )
              if( g < ::fabs(h) * Small_Val )
                t=( Apq )/h;
              else {
                theta=0.5*h/( Apq );
                t=1.0/(::fabs(theta)+ ::sqrt(1.0+theta*theta));
                if (theta < 0.0) t = -t;
              }
              c=1.0/ ::sqrt(1+t*t);
              s= t*c;
              tau= s/(1.0+c);

              // COUT << "s: " << s << "  tau: " << tau << ENDL;
              h=t*Apq;
              Z[ip] -= h;
              Z[iq] += h;
              Dip -= h;
              Diq += h;
              Apq=0.0;
              {
                Index j;
                for (j=0;j<ip;j++)
                  ROTATE(A,D,j,ip,j,iq,s,tau);
                for (j=ip+1;j<iq;j++)
                  ROTATE(A,D,ip,j,j,iq,s,tau);
                for (j=iq+1;j<D;j++)
                  ROTATE(A,D,ip,j,iq,j,s,tau);
                for (j=0;j<D;j++)
                  ROTATE(EVect,D,j,ip,j,iq,s,tau);
                nrot++;
              }
            }
        }
      }
    }
    {
      for (Index ip=0;ip<D;ip++) {
        B[ip] += Z[ip];
        EVal[ip] = B[ip];
        Z[ip] = 0.0;
      }
    }
  }

  throw ErrConvergence("Sym<D,T>::Diagonalize(...) [Jacobi]");
}

template<Index D, class T> int Sym<D,T>::Diagonalize(Matrix<D,T>& EVect, Vect<D,T>&  EVal) const  throw(ErrConvergence)
{
  Matrix<D,T> A(*this);
  EVect.Id();

  Vect<D,T> B,Z;

  return private_Diagonalize(A.getValues(), D, EVect.getValues(), EVal.getValues(),B.getValues(),Z.getValues());
}

template<Index D, class T> int Sym<D,T>::DiagonalizeSort(Matrix<D,T>& EVect, Vect<D,T>&  EVal) const  throw(ErrConvergence)
{
  int n=this->Diagonalize(EVect,EVal);
  Index k,i,j;
  T p;
  for (i=0;i<D;i++) {
    p=EVal(k=i);
    for (j=i+1;j<D;j++) {
      if (EVal(j)>p) p=EVal(k=j);
    }
    if (k!=i) {
      EVal(k)=EVal(i);
      EVal(i)=p;
      for (j=0; j<D; j++) {
        p=EVect(j,k);
        EVect(j,k)=EVect(j,i);
        EVect(j,i)=p;
      }
    }
  }
  return n;
}

// TODO : vec_Inline ???
template<Index D, class T> T Sym<D,T>::SqNorm( const Vect<D,T>& v) const throw()
{
  T res_val = 0.0;
  const T *W = this->val;
  for(Index i=0; i< D ; i++){
    const T Vi = v(i);
    for(Index j=0; j < i; j++)
      res_val += 2 * (*W++) *  v(j) * Vi;
    res_val += (*W++)* Vi * Vi;
  }
  return res_val;
}

template <Index D, class T> vec_Inline T Sym<D,T>::Trace() const throw()
{
  T t=0;
  for (Index i=0; i< D; i++)
    t+=val[i*(i+3)/2];
  return t;
}

template<Index D, class T> vec_Inline T Sym<D,T>::Det() const throw(ErrConvergence)
{
  Matrix<D,T> EVect(VIM_NOINIT);
  Vect<D,T> EVal(VIM_NOINIT);

  try{
    Diagonalize( EVect, EVal);
    T det = EVal(0);
    for( Index i = 1; i<D; i++)
      det *= EVal(i);
    return det;
  }
  catch(ErrConvergence& exc){
    exc.push("T Sym<D,T>::Det() const");
    throw exc;
  }
}

template<Index D, class T> vec_Inline void Sym<D,T>::autoCorrel( const Vect<D,T>& x) throw()
{//TODO:optimize
  for(Index i=0; i< D ; i++){
    T xi = x(i);
    for(Index j=0; j <= i; j++)
      val[ (i*(i+1))/2 + j ] = xi * x(j);
  }
}

template<Index D, class T> int Sym<D,T>::inv(const Sym<D,T>&  W) throw(ErrConvergence)
{
  Vect<D,T> EVal(VIM_NOINIT);
  Matrix<D,T> V(VIM_NOINIT), Vt(VIM_NOINIT);

  try{ W.Diagonalize( V, EVal); }
  catch(ErrConvergence& exc){
    exc.push("int Sym<D,T>::Inv<D>( const Sym<D,T>&  W)");
    throw exc;
  }
  Vt.Trp(V);

  const T Precision_Factor = 1e1;//was 1e3
  T Small_Val;Xlinalg_Small_Val<T>(Small_Val);

  /* compute max and min eigenvalues */
  T max = ::fabs( EVal(0) ), min = max;
  {
    for(Index i=1; i<D; i++){
      T tmp;
      if ( (tmp= ::fabs( EVal(i) ) ) > max) max = tmp;
      if ( tmp < min) min = tmp;
    }
  }

  /* Exeption for the null matrix */
  if ( ((T) max) == ((T) 0.0)) {
    this->resetTo(0);
    return D;
  }

  /* Ok, now we can divide by max */
  int singular = 0; /* count of singular values */
  {
    for(Index i=0; i<D; i++){
      T& Di = EVal(i);
      T g = ::fabs( Di )/Precision_Factor;
      //if ( ((T) (g + max))  == ((T) max) )
      if ( g <  max * Small_Val ) {
        // if (K_debug) printf(" Very small eigenvalue : %g\n", (double) D[i]);
        Di =0.0;
        singular ++;
      }
      else Di = 1.0/Di;
    }
  }

  // Winv = V.Dinv.Vtrp
  {
    for (Index i=0; i<D; i++)
      for(Index j=0; j<D; j++)
        V(i,j) *= EVal(j);
  }
  *this=V*Vt;

  return singular;
}


// I/O operations
template <Index D, class T> STD_OSTREAM& operator<<(STD_OSTREAM& s, const Sym<D,T>& W) throw() {
  s<<"(";
  for (Index i=0; i<D; i++) {
    s<<"(";
    for (Index j=0; j<D; j++)
      s<<W(i,j)<<" ";
    s<<")\n";
  }
  s<<")";
  return s;
}

template <Index D, class T> void Sym<D,T>::private_mulsym(const Sym<D,T> &m1, const Sym<D,T> &m2) throw()
{
  T *res = val;
  for(Index i=0; i<D; i++)
    for (Index j=0; j<=i; j++) {
      T cur_val = 0. ;
      for(Index k=0; k<D; k++)
        cur_val += m1(i,k) * m2(k,j);
      *res++ = cur_val;
    }
}

template <Index D, class T> void Sym<D,T>::mulsym(const Sym<D,T> &m1, const Sym<D,T> &m2) throw()
{
  if ( (&m1 == this) || (&m2 == this) ) {
    Sym<D,T> tmp(VIM_NOINIT);
    tmp.private_mulsym(m1,m2);
    this->copy(tmp);
  } else {
    this->private_mulsym(m1,m2);
  }
}


template <Index D, class T> void Sym<D,T>::Test() throw()
{
  Sym<D,float> df;Sym<D,double> dd;Sym<D,int> di;
  Vect<D,T> bb(0);

  Sym<D,T> a(0.1),b(bb),c(VIM_INIT),d(a),/*af(Af),ad(Ad),ai(Ai),*/ef(df),ed(dd),ei(di);
  T v;char TName[100];Xlinalg_Class_Name(v,TName);
  COUT << "Sym<"<<D<<","<<TName<<">" << ENDL;
  COUT << "Constructors" << ENDL;
  COUT << "a must be Id*0.1 : " << a << ENDL;
  COUT << "b must be diag(1,0,...) : " << b << ENDL;
  COUT << "c must be 0,0,... : " << c << ENDL;
  COUT << "d must be as a (copy) : " << d << ENDL;
  COUT << "ef must be 0 : " << ef << ENDL;
  COUT << "ed must be 0 : " << ed << ENDL;
  COUT << "ei must be 0 : " << ei << ENDL;

  COUT << "operator=" << ENDL;
  d=c;COUT << "d=c : " << d << ENDL;
  d=df;COUT << "d=df : " << d << ENDL;
  d=dd;COUT << "d=dd : " << d << ENDL;
  d=di;COUT << "d=di : " << d << ENDL;

  COUT << "Access to Data" << ENDL;
  d.copy(b);COUT << "d.copy(b)  : " << d << ENDL;
  d(0,0)=26;COUT << "d(0,0)=26;  d(0,0) : " << d(0,0) << ENDL;
  d.Null();COUT << "d.Null() : " << d << ENDL;
  d.resetTo(2);COUT << "d.resetTo(2) : " << d << ENDL;
  T*Dd=d.getValues();
  COUT << "getValues() : " << Dd[0] << " " << Dd[1] << " " << Dd[2] << ENDL;
  const T*E=d;
  COUT << "(const T*) d : " << E[0] << " " << E[1] << " " << E[2] << ENDL;

  COUT << "Math" << ENDL;
  COUT << "d.Minus(a) : " << d.Minus(a) << ENDL;
  COUT << "d.Add(a,b) : " << d.Add(a,b) << ENDL;
  COUT << "d.Sub(a,b) : " << d.Sub(a,b) << ENDL;
  COUT << "d.Mul(a,2) : " << d.Mul(a,(T)2) << ENDL;
  COUT << "d.Mul(2,a) : " << d.Mul((T)2,a) << ENDL;
  COUT << "d.Div(a,2) : " << d.Div(a,(T)2) << ENDL;
  COUT << "d.Div(a,2) : " << d.Div(a,(T)2) << ENDL;

  COUT << "+a : " << +a << ENDL;
  COUT << "-a : " << -a << ENDL;
  COUT << "a+b : " << a+b << ENDL;
  COUT << "a-b : " << a-b << ENDL;
  COUT << "a*2 : " << a*((T)2) << ENDL;
  COUT << "2*a : " << ((T)2)*a << ENDL;
  COUT << "a/2 : " << a/((T)2) << ENDL;

  d.copy(b);d+=a;COUT << "d=b; d+=a : " << d << ENDL;
  d.copy(b);d-=a;COUT << "d=b; d-=a : " << d << ENDL;
  d.copy(b);d*=(T)2;COUT << "d=b; d*=2 : " << d << ENDL;
  d.copy(b);d/=(T)2;COUT << "d=b; d/=2 : " << d << ENDL;

  d.copy(b);COUT << "d=b; d==b : " << (d==b) << ENDL;
  d.copy(b);COUT << "d=b; d!=b : " << (d!=b) << ENDL;

  COUT << "dot(a,b) : " << dot(a,b) << ENDL;
  COUT << "DotProd(a,b) : " << DotProd(a,b) << ENDL;
  COUT << "norm2(a) : " << a.norm2() << ENDL;
  COUT << "norm(a) : " << a.norm() << ENDL;
  COUT << "Norm2(a) : " << a.Norm2() << ENDL;
  COUT << "Norm(a) : " << a.Norm() << ENDL;
  COUT << "NormL1(a) : " << a.NormL1() << ENDL;
  COUT << "NormLinf(a) : " << a.NormLinf() << ENDL;
  COUT << "Dist2(a,b) : " << Dist2(a,b) << ENDL;
  COUT << "Dist(a,b) : " << Dist(a,b) << ENDL;
  COUT << "DistL1(a,b) : " << DistL1(a,b) << ENDL;
  COUT << "DistLinf(a,b) : " << DistLinf(a,b) << ENDL;

  d.resetTo(2); d.min(a); COUT << "d.resetTo(2); d.min(a) : " << d << ENDL;
  d.resetTo(2); d.max(a); COUT << "d.resetTo(2); d.max(a) : " << d << ENDL;

  d.Rnd(2); COUT << "d.Rnd(2) : " << d << ENDL;

  {Matrix<D,T> m;m.Rnd();d=m;COUT << "operator=Matrix<D,T>" << d << "=" << m << ENDL;}
  {Matrix<D,T> m;m.Rnd();Sym<D,T> d(m);COUT << "Constructeur(Matrix<D,T>)" << d << "=" << m << ENDL;}

  {d.Rnd();Vect<D,T>v;COUT << "Diag" << d.Diag(v) << ENDL;}

  {d.Rnd();COUT << "Det" << d.Det() << ENDL;}
  {d.Rnd();COUT << "Trace" << d.Trace() << ENDL;}

  {Vect<D,T>v;COUT << "SqNorm" << d.SqNorm(v) << ENDL;}
  {Vect<D,T>v;COUT << "AutoCorrel" << d.AutoCorrel(v) << ENDL;}

  {d.Rnd();d.Inv(d);COUT << "Inv" << ENDL;}
  {d.Rnd();Matrix<D,T>m;Vect<D,T>v;d.DiagonalizeSort(m,v);COUT << "Diagonalise" << ENDL;}

  COUT << ENDL;
}



} // namespace yav

#endif
