// General comments : see Vec.h

#ifndef _VEC2_H
#define _VEC2_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#ifndef WIN32
#pragma interface "linearAlgebra/Vec2.h"
#endif
#include "linearAlgebra/Vec.h"
//#include "linearAlgebra/Mat.h"

#ifdef WIN32
#include "basics/export.h"
#endif


namespace yav {


template <class T> class Vec2;
template <class T> class Matrix2x2;

/** @name Vectorial redefined friend functions */
//@{
/** unary +
    @param p vector
    @return +p */
template < class T> vec_Inline Vec2< T> operator+(const Vec2< T> &p) throw()
  {return p;}
/** unary -
    @param p vector
    @return -p */
template < class T> vec_Inline Vec2< T> operator-(const Vec2< T> &p) throw()
  {return Vec2< T>(VIM_NOINIT).Minus(p);}
/** vector sum
    @param p first vector
    @param q second vector
    @return p + q */
template < class T> vec_Inline Vec2< T> operator+(const Vec2< T> &p, const Vec2 < T> &q) throw()
  {return Vec2< T>(VIM_NOINIT).Add(p,q);}
/** vector substraction
    @param p first vector
    @param q second vector
    @return p - q */
template < class T> vec_Inline Vec2< T> operator-(const Vec2< T> &p, const Vec2 < T> &q) throw()
  {return Vec2< T>(VIM_NOINIT).Sub(p,q);}
/** vector multiplication by scalar
    @param p vector
    @param a scalar
    @return ap */
template < class T> vec_Inline Vec2< T> operator*(const Vec2< T> &p, const T a) throw()
  {return Vec2< T>(VIM_NOINIT).Mul(p,a);}
/** vector scalar multiplication
    @param a scalar
    @param p vector
    @return ap */
template < class T> vec_Inline Vec2< T> operator*(const T a, const Vec2< T> &p) throw()
  {return Vec2< T>(VIM_NOINIT).Mul(p,a);}
/** vector scalar division
    @param p vector
    @param a scalar
    @return p / a */
template < class T> vec_Inline Vec2< T> operator/(const Vec2< T> &p, const T a) throw(ErrDivZero)
  {return Vec2< T>(VIM_NOINIT).Div(p,a);}
/** dot product
    @param u first vector
    @param v second vector
    @return u.v */
template < class T> vec_Inline T dot(const Vec2< T>& u, const Vec2< T> &v) throw()
  {return Vec2< T>::dot(u,v);}
/// Square Euclidian Distance
template < class T> vec_Inline T dist2( const Vec2< T>& u, const Vec2< T>& v) throw()
  {return Vec2< T>::dist2(u,v);}
/// Euclidian Distance
template < class T> vec_Inline T dist( const Vec2< T>& u, const Vec2< T>& v) throw()
  {return Vec2< T>::dist(u,v);}
/// L1 Distance
template < class T> vec_Inline T distL1( const Vec2< T>& u, const Vec2< T>& v) throw()
  {return Vec2< T>::distL1(u,v);}
/// Linf Distance
template < class T> vec_Inline T distLinf( const Vec2< T>& u, const Vec2< T>& v) throw()
  {return Vec2< T>::distLinf(u,v);}
/** dot product
    @param p first vector
    @param q second vector
    @return p.q */
template < class T> vec_Inline T DotProd(const Vec2< T> &p, const Vec2< T> &q) throw()
  {return Vec2< T>::dot(p,q);}
/// Square Euclidian Distance
template < class T> vec_Inline T Dist2( const Vec2< T>& u, const Vec2< T>& v) throw()
  {return Vec2< T>::dist2(u,v);}
/// Euclidian Distance
template < class T> vec_Inline T Dist( const Vec2< T>& u, const Vec2< T>& v) throw()
  {return Vec2< T>::dist(u,v);}
/// L1 Distance
template < class T> vec_Inline T DistL1( const Vec2< T>& u, const Vec2< T>& v) throw()
  {return Vec2< T>::distL1(u,v);}
/// Linf Distance
template < class T> vec_Inline T DistLinf( const Vec2< T>& u, const Vec2< T>& v) throw()
  {return Vec2< T>::distLinf(u,v);}
/** operator == */
template < class T> vec_Inline bool operator == (const Vec2< T>& u, const Vec2< T> &v) throw()
  {return Vec2< T>::isEqual(u,v);}
/** operator != */
template < class T> vec_Inline bool operator != (const Vec2< T>& u, const Vec2< T> &v) throw()
  {return Vec2< T>::isDifferent(u,v);}
/** operator < */
template < class T> vec_Inline bool operator < (const Vec2< T>& u, const Vec2< T> &v) throw()
  {return Vec2< T>::isSmaller(u,v);}
/** operator > */
template < class T> vec_Inline bool operator > (const Vec2< T>& u, const Vec2< T> &v) throw()
  {return Vec2< T>::isGreater(u,v);}
/** operator <= */
template < class T> vec_Inline bool operator <= (const Vec2< T>& u, const Vec2< T> &v) throw()
  {return Vec2< T>::isSmallerOrEqual(u,v);}
/** operator >= */
template < class T> vec_Inline bool operator >= (const Vec2< T>& u, const Vec2< T> &v) throw()
  {return Vec2< T>::isGreaterOrEqual(u,v);}
//@}

/** @name Vector n redefined friend functions */
//@{
template <class T> vec_Inline Vec2<T> operator*(const Matrix<2,T> &m, const Vec2<T> &v) throw()
  {return Vec2<T>(VIM_NOINIT).Mul(m,v);}
template <class T> vec_Inline Vec2<T> operator*(const Vec2<T> &v, const Matrix<2,T> &m) throw()
  {return Vec2<T>(VIM_NOINIT).Mul(v,m);}
//@}

template <class T> T det(const Vec2<T>&p, const Vec2<T>&q) throw();

/*! \class Vec2 Vec2.h linearAlgebra/Vec2.h
  \ingroup liblinearAlgebra
  template for a vector of dimension 2 and type T 
*/
template <class T> class Vec2 : public Vect<2,T> {
public:
  /** @name Constructors */
  //@{
  /** @name Vectorial redefined constructors */
  //@{
  /** default constructor - initializes to 0 if UNINIT_ZERO not defined */
  vec_Inline Vec2() throw() : Vect<2,T>()
    {}
  /** constructor - initializes to 0 if b==VIM_INIT */
  explicit vec_Inline Vec2(VectInitMode b) throw() : Vect<2,T>(b)
    {}
  //@}
  /** @name Vector redefined constructors */
  //@{
#ifdef  __GNUC__
  /** copy constructor
      @param v vector to copy */
  vec_Inline Vec2(const Vect<2,T>& v) throw() : Vect<2,T>(v)
    {}
  /** operator =
      @param v vector to copy
      @return this */
  vec_Inline Vec2< T>& operator=(const Vect<2,T>& v) throw()
    {copy(v);return *this;}
#endif
  /** copy constructor with cast
      @param v vector to copy */
  template<class T2> vec_Inline Vec2(const Vect<2,T2>& v) throw() : Vect<2,T>(v)
    {}
  /** operator = with cast
      @param v vector to copy
      @return this */
  template<class T2> vec_Inline Vec2< T>& operator=(const Vect<2,T2>& v) throw()
    {copy(v);return *this;}
  
#ifndef WIN32
  /** constructor: initialize to vector (coord[0], coord[1], ...)
      @param coord vector coordinates array */
  vec_Inline Vec2(const T coord[2]) throw() : Vect<2,T>(coord)
    {}
#endif
  /** constructor: initialize to vector (coord[0], coord[1], ...)
      @param coord vector coordinates array */
  template<class T2> vec_Inline Vec2(const T2 coord[2]) throw() : Vect<2,T>(coord)
    {}
  
  /** constructor: initializes element i with 1, other with 0 (canonical basis vector i) */
  explicit vec_Inline Vec2(int i) throw(ErrRange) : Vect<2,T>(i)
    {}
  //@}

  /** constructor: initialize to vector (x,y)
      @param x first vector coordinate
      @param y second vector coordinate
  */
  Vec2(T x, T y) throw();

  /** @name Data members access */
  //@{
  /** set this vector to vector (x,y)
      @param x first vector coordinate
      @param y second vector coordinate */
  void setValue(T x, T y) throw();

  /// returns X coordinate
  T getX(void) const throw();

  /// returns Y coordinate
  T getY(void) const throw();
  //@}

  /** @name mathematical operators */
  //@{
  /** @name Vectorial redefined menber functions */
  //@{
  /** set this vector to p+q
      @return this = p + q */
  vec_Inline Vec2< T>& Minus(const Vec2< T> &p) throw()
    {vec_opposite(this->val,2,p.val);return *this;}
  /** set this vector to p+q
      @return this = p + q */
  vec_Inline Vec2< T>& Add(const Vec2< T> &p, const Vec2< T> &q) throw()
    {vec_add(this->val,2,p.val,q.val);return *this;}
  /** set this vector to p-q
      @return this = p - q */
  vec_Inline Vec2< T>& Sub(const Vec2< T> &p, const Vec2< T> &q) throw()
    {vec_sub(this->val,2,p.val,q.val);return *this;}
  /** set each coordinate of this vector to product of this and p same coordinate
      @return this = this.*q */
  vec_Inline Vec2< T>& CMul(const Vec2< T> &q) throw()
    {vec_cmul(this->val,2,q.val);return *this;}
  /** set each coordinate of this vector to product of q and p same coordinate
      @return this = p.*q */
  vec_Inline Vec2< T>& CMul(const Vec2< T> &p, const Vec2< T> &q) throw()
    {vec_cmul(this->val,2,p.val,q.val);return *this;}
  /** set this vector to p*a
      @return this = p*a */
  vec_Inline Vec2< T>& Mul(const Vec2< T> &p, const T a) throw()
    {vec_mul(this->val,2,p.val,a);return *this;}
  /** set this vector to p*a
      @return this = p*a */
  vec_Inline Vec2< T>& Mul(const T a, const Vec2< T> &p) throw()
    {vec_mul(this->val,2,p.val,a);return *this;}
  /** set this vector to p/a
      @return this = p/a */
  vec_Inline Vec2< T>& Div(const Vec2< T> &p, const T a) throw(ErrDivZero)
    {vec_div(this->val,2,p.val,a);return *this;}
  /** adds vector q to vector p
      @param q vector to add
      @return p + q */
  vec_Inline Vec2< T>& operator+=(const Vec2< T> &q) throw()
    {vec_add(this->val,2,q.val);return *this;}
  /** substracts vector q from vector p
      @param q vector to substract
      @return p - q */
  vec_Inline Vec2< T>& operator-=(const Vec2< T> &q) throw()
    {vec_sub(this->val,2,q.val);return *this;}
  /** multiplies vector p by scalar a
      @param a scalar
      @return ap */
  vec_Inline Vec2< T>& operator*=(const T a) throw()
    {vec_mul(this->val,2,a);return *this;}
  /** divides vector p by scalar a
      @param a scalar
      @return p / a */
  vec_Inline Vec2< T>& operator/=(const T a) throw(ErrDivZero)
    {vec_div(this->val,2,a);return *this;}
  /** randomize vector
      @return this */
  vec_Inline Vec2< T>& Rnd(T min = -1., T max = 1.) throw()
    {vec_rnd(this->val,2,min,max);return *this;}
  /** randomize vector : isotropic gaussian centred noise
      @return this */
  vec_Inline Vec2< T>& RndGaussian(T standard_deviation = 1.) throw()
    {vec_rndgaussian(this->val,2,standard_deviation);return *this;}
  /** takes the X,Y or Z value of p if they are less than its own X,Y,Z coordinates
      @return _Vect<2,T>(min(x,p.x),min(y,p.y),...) */
  vec_Inline Vec2< T>& Min(const Vec2< T>& p) throw()
    {vec_min(this->val,2,p.val);return *this;}
  /** takes the X,Y or Z value of p if they are greater than its own X,Y,Z coordinates
      @return _Vect<2,T>(max(x,p.x),max(y,p.y),...) */
  vec_Inline Vec2< T>& Max(const Vec2< T>& p) throw()
    {vec_max(this->val,2,p.val);return *this;}
  /** set each element of this vector to value
      @param v value
      @return this */
  vec_Inline Vec2< T>& Reset(T v) throw()
    {resetTo(v);return *this;}
  /** set this vector to null vector
      @return this */
  vec_Inline Vec2< T>& Null() throw()
  {resetTo((T)0);return *this;}
  /** set this vector to mean of a given list of vectors
      @return this */
  template<class InputIterator>
  vec_Inline Vec2< T>& Mean(InputIterator first, InputIterator last) throw()
    {mean(first,last);return *this;}
  //@}

  /** determinant
      @param p first vector
      @param q second vector
      @return determinant of (p,q) */
  //friend T det FRIEND_TEMPLATE_ARG(const Vec2<T> &p, const Vec2<T> &q) throw();

  /** orthogonal vector
      @return (-y,x) */
  Vec2<T> orthogonal(void) const throw();

  /// x=a et y=a
  void MinMaxInit(T a) throw();

  /// x=min(x,a) et y=max(y,a)
  void MinMax(T a) throw();

  /** @name Matrix operations */
  //@{
  /** @name Vector n redefined menber functions */
  //@{
  vec_Inline Vec2<T>& Mul(const Matrix<2,T> &m, const Vec2<T> &v) throw()
    {((Mat<2,1,T>*)this)->Mul(m,(const Mat<2,1,T>&)v);return *this;}
  vec_Inline Vec2<T>& Mul(const Vec2<T> &v, const Matrix<2,T> &m) throw()
    {((Mat<1,2,T>*)this)->Mul((const Mat<1,2,T>&)v,m);return *this;}
  //@}

  static void Test() throw();
};

} // namespace yav

/*][** NODOC ****************************************************************/
/****************************************************************************/

/*]]**************************************************************************/

namespace yav {

/****************************************************************************/
/* Vec2 Specialisation                                                      */
/****************************************************************************/
// constructors
template <class T> vec_Inline Vec2<T>::Vec2(T x, T y) throw()
  : Vect<2,T>(VIM_NOINIT)
{
    this->val[0]=x; this->val[1]=y;
}

// access to data
template <class T> vec_Inline void Vec2<T>::setValue(T x, T y) throw()
{
    this->val[0]=x; this->val[1]=y;
}
template <class T> vec_Inline T Vec2<T>::getX(void) const throw() {return this->val[0];}
template <class T> vec_Inline T Vec2<T>::getY(void) const throw() {return this->val[1];}

// mathematical operations
template <class T> vec_Inline T det(const Vec2<T> &p, const Vec2<T> &q) throw()
{
  return p(0)*q(1) - p(1)*q(0) ;
}

template <class T> vec_Inline Vec2<T> Vec2<T>::orthogonal(void) const throw()
{
  return (Vec2<T>(-this->val[1],this->val[0]));
}

template <class T> vec_Inline void Vec2<T>::MinMaxInit(T a) throw()
{
  this->val[0]=a;
  this->val[1]=a;
}
template <class T> vec_Inline void Vec2<T>::MinMax(T a) throw()
{
  if (this->val[0]>a) this->val[0]=a;
  if (this->val[1]<a) this->val[1]=a;
}

template <class T> void Vec2<T>::Test() throw()
{
  float Af[2];double Ad[2];int Ai[2];
  T A[2];for (Index k=0; k<2; k++) {A[k]=k+1;Af[k]=k+1;Ad[k]=k+1;Ai[k]=k+1;}
  Vec2<float> df;Vec2<double> dd;Vec2<int> di;

  Vec2<T> a(A),b(1),c(VIM_INIT),d(a),e(1,2),af(Af),ad(Ad),ai(Ai),ef(df),ed(dd),ei(di);
  T v;char TName[100];Xlinalg_Class_Name(v,TName);
  COUT << "Vec2<" << TName << ">" << ENDL;
  COUT << "Constructors" << ENDL;
  COUT << "a must be 1,2,... : " << a << ENDL;
  COUT << "af must be 1,2,... : " << af << ENDL;
  COUT << "ad must be 1,2,... : " << ad << ENDL;
  COUT << "ai must be 1,2,... : " << ai << ENDL;
  COUT << "b must be 0,1,... : " << b << ENDL;
  COUT << "c must be 0,0,... : " << c << ENDL;
  COUT << "d must be as a (copy) : " << d << ENDL;
  COUT << "e must be 1,2,... : " << e << ENDL;
  COUT << "ef must be 0 : " << ef << ENDL;
  COUT << "ed must be 0 : " << ed << ENDL;
  COUT << "ei must be 0 : " << ei << ENDL;

  COUT << "operator=" << ENDL;
  d=c;COUT << "d=c : " << d << ENDL;
  d=df;COUT << "d=df : " << d << ENDL;
  d=dd;COUT << "d=dd : " << d << ENDL;
  d=di;COUT << "d=di : " << d << ENDL;

  COUT << "Access to Data" << ENDL;
  d.copy(b);COUT << "d.copy(b)  : " << d << ENDL;
  d[0]=1;COUT << "d[0]=1 : d[0] " << d[0] << ENDL;
  d(0)=1;COUT << "d(0)=1 : d(0) " << d(0) << ENDL;
  d.setValue(2,3);COUT << "d.setValue(2,3); : " << d << ENDL;
  COUT << "d.get.() : " << d.getX() << " " << d.getY() << ENDL;
  d.Null();COUT << "d.Null() : " << d << ENDL;
  d.resetTo(2);COUT << "d.resetTo(2) : " << d << ENDL;
  T*D=d.getValues();
  COUT << "getValues() : " << D[0] << " " << D[1] << " " << D[2] << ENDL;
  const T*E=d;
  COUT << "(const T*) d : " << E[0] << " " << E[1] << " " << E[2] << ENDL;
  COUT << "a.ColMat : " << a.ColMat() << ENDL;
  COUT << "a.RowMat : " << a.RowMat() << ENDL;

  COUT << "Math" << ENDL;
  COUT << "d.Minus(a) : " << d.Minus(a) << ENDL;
  COUT << "d.Add(a,b) : " << d.Add(a,b) << ENDL;
  COUT << "d.Sub(a,b) : " << d.Sub(a,b) << ENDL;
  COUT << "d.Mul(a,2) : " << d.Mul(a,(T)2) << ENDL;
  COUT << "d.Mul(2,a) : " << d.Mul((T)2,a) << ENDL;
  COUT << "d.Div(a,2) : " << d.Div(a,(T)2) << ENDL;
  COUT << "d.Div(a,2) : " << d.Div(a,(T)2) << ENDL;

  COUT << "+a : " << +a << ENDL;
  COUT << "-a : " << -a << ENDL;
  COUT << "a+b : " << a+b << ENDL;
  COUT << "a-b : " << a-b << ENDL;
  COUT << "a*2 : " << a*((T)2) << ENDL;
  COUT << "2*a : " << ((T)2)*a << ENDL;
  COUT << "a/2 : " << a/((T)2) << ENDL;

  d.copy(b);d+=a;COUT << "d=b; d+=a : " << d << ENDL;
  d.copy(b);d-=a;COUT << "d=b; d-=a : " << d << ENDL;
  d.copy(b);d*=(T)2;COUT << "d=b; d*=2 : " << d << ENDL;
  d.copy(b);d/=(T)2;COUT << "d=b; d/=2 : " << d << ENDL;

  d.copy(b);COUT << "d=b; d==b : " << (d==b) << ENDL;
  d.copy(b);COUT << "d=b; d!=b : " << (d!=b) << ENDL;

  COUT << "dot(a,b) : " << dot(a,b) << ENDL;
  COUT << "DotProd(a,b) : " << DotProd(a,b) << ENDL;
  COUT << "norm2(a) : " << a.norm2() << ENDL;
  COUT << "norm(a) : " << a.norm() << ENDL;
  COUT << "Norm2(a) : " << a.Norm2() << ENDL;
  COUT << "Norm(a) : " << a.Norm() << ENDL;
  COUT << "NormL1(a) : " << a.NormL1() << ENDL;
  COUT << "NormLinf(a) : " << a.NormLinf() << ENDL;
  COUT << "Dist2(a,b) : " << Dist2(a,b) << ENDL;
  COUT << "Dist(a,b) : " << Dist(a,b) << ENDL;
  COUT << "DistL1(a,b) : " << DistL1(a,b) << ENDL;
  COUT << "DistLinf(a,b) : " << DistLinf(a,b) << ENDL;

  d.resetTo(2); d.min(a); COUT << "d.resetTo(2); d.min(a) : " << d << ENDL;
  d.resetTo(2); d.max(a); COUT << "d.resetTo(2); d.max(a) : " << d << ENDL;

  d.Rnd(2); COUT << "d.Rnd(2) : " << d << ENDL;
  d.RndGaussian(2); COUT << "d.RndGaussian(2) : " << d << ENDL;

  COUT << "det(a,b) : " << det(a,b) << ENDL;
  COUT << "a.orthogonal() : " << a.orthogonal() << ENDL;

  {Matrix<2,T> m; COUT << "Matrix<2,T> m*d : " << m*d << "  Mul : " << c.Mul(m,d) << ENDL;}
  {Matrix<2,T> m; COUT << "d*Matrix<2,T> m : " << d*m << "  Mul : " << c.Mul(d,m) << ENDL;}

  COUT << ENDL;
}



} // namespace yav


#endif
