// General comments : see Vec.h

#ifndef _VEC3_H
#define _VEC3_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#ifndef WIN32
#pragma interface "linearAlgebra/Vec3.h"
#endif
#include "linearAlgebra/Vec.h"
#include "linearAlgebra/Mat.h"

#ifdef WIN32
#include "basics/export.h"
#endif


namespace yav {


template <class T> class Vec3;
template <class T> class Matrix3x3;
template <class T> class HVec4;
template <class T> class Matrix4x4;

/** @name Vectorial redefined friend functions */
//@{
/** unary +
    @param p vector
    @return +p */
template < class T> vec_Inline Vec3< T> operator+(const Vec3< T> &p) throw()
  {return p;}
/** unary -
    @param p vector
    @return -p */
template < class T> vec_Inline Vec3< T> operator-(const Vec3< T> &p) throw()
  {return Vec3< T>(VIM_NOINIT).Minus(p);}
/** vector sum
    @param p first vector
    @param q second vector
    @return p + q */
template < class T> vec_Inline Vec3< T> operator+(const Vec3< T> &p, const Vec3 < T> &q) throw()
  {return Vec3< T>(VIM_NOINIT).Add(p,q);}
/** vector substraction
    @param p first vector
    @param q second vector
    @return p - q */
template < class T> vec_Inline Vec3< T> operator-(const Vec3< T> &p, const Vec3 < T> &q) throw()
  {return Vec3< T>(VIM_NOINIT).Sub(p,q);}
/** vector multiplication by scalar
    @param p vector
    @param a scalar
    @return ap */
template < class T> vec_Inline Vec3< T> operator*(const Vec3< T> &p, const T a) throw()
  {return Vec3< T>(VIM_NOINIT).Mul(p,a);}
/** vector scalar multiplication
    @param a scalar
    @param p vector
    @return ap */
template < class T> vec_Inline Vec3< T> operator*(const T a, const Vec3< T> &p) throw()
  {return Vec3< T>(VIM_NOINIT).Mul(p,a);}
/** vector scalar division
    @param p vector
    @param a scalar
    @return p / a */
template < class T> vec_Inline Vec3< T> operator/(const Vec3< T> &p, const T a) throw(ErrDivZero)
  {return Vec3< T>(VIM_NOINIT).Div(p,a);}
/** dot product
    @param u first vector
    @param v second vector
    @return u.v */
template < class T> vec_Inline T dot(const Vec3< T>& u, const Vec3< T> &v) throw()
  {return Vec3< T>::dot(u,v);}
/// Square Euclidian Distance
template < class T> vec_Inline T dist2( const Vec3< T>& u, const Vec3< T>& v) throw()
  {return Vec3< T>::dist2(u,v);}
/// Euclidian Distance
template < class T> vec_Inline T dist( const Vec3< T>& u, const Vec3< T>& v) throw()
  {return Vec3< T>::dist(u,v);}
/// L1 Distance
template < class T> vec_Inline T distL1( const Vec3< T>& u, const Vec3< T>& v) throw()
  {return Vec3< T>::distL1(u,v);}
/// Linf Distance
template < class T> vec_Inline T distLinf( const Vec3< T>& u, const Vec3< T>& v) throw()
  {return Vec3< T>::distLinf(u,v);}
/** dot product
    @param p first vector
    @param q second vector
    @return p.q */
template < class T> vec_Inline T DotProd(const Vec3< T> &p, const Vec3< T> &q) throw()
  {return Vec3< T>::dot(p,q);}
/// Square Euclidian Distance
template < class T> vec_Inline T Dist2( const Vec3< T>& u, const Vec3< T>& v) throw()
  {return Vec3< T>::dist2(u,v);}
/// Euclidian Distance
template < class T> vec_Inline T Dist( const Vec3< T>& u, const Vec3< T>& v) throw()
  {return Vec3< T>::dist(u,v);}
/// L1 Distance
template < class T> vec_Inline T DistL1( const Vec3< T>& u, const Vec3< T>& v) throw()
  {return Vec3< T>::distL1(u,v);}
/// Linf Distance
template < class T> vec_Inline T DistLinf( const Vec3< T>& u, const Vec3< T>& v) throw()
  {return Vec3< T>::distLinf(u,v);}
/** operator == */
template < class T> vec_Inline bool operator == (const Vec3< T>& u, const Vec3< T> &v) throw()
  {return Vec3< T>::isEqual(u,v);}
/** operator != */
template < class T> vec_Inline bool operator != (const Vec3< T>& u, const Vec3< T> &v) throw()
  {return Vec3< T>::isDifferent(u,v);}
/** operator < */
template < class T> vec_Inline bool operator < (const Vec3< T>& u, const Vec3< T> &v) throw()
  {return Vec3< T>::isSmaller(u,v);}
/** operator > */
template < class T> vec_Inline bool operator > (const Vec3< T>& u, const Vec3< T> &v) throw()
  {return Vec3< T>::isGreater(u,v);}
/** operator <= */
template < class T> vec_Inline bool operator <= (const Vec3< T>& u, const Vec3< T> &v) throw()
  {return Vec3< T>::isSmallerOrEqual(u,v);}
/** operator >= */
template < class T> vec_Inline bool operator >= (const Vec3< T>& u, const Vec3< T> &v) throw()
  {return Vec3< T>::isGreaterOrEqual(u,v);}
//@}

/** @name Vector n redefined friend functions */
//@{
template <class T> vec_Inline Vec3<T> operator*(const Matrix<3,T> &m, const Vec3<T> &v) throw()
  {return Vec3<T>(VIM_NOINIT).Mul(m,v);}
template <class T> vec_Inline Vec3<T> operator*(const Vec3<T> &v, const Matrix<3,T> &m) throw()
  {return Vec3<T>(VIM_NOINIT).Mul(v,m);}
//@}

//template <class T> Vec3<T> operator*(const Matrix4x4<T> &m, const Vec3<T>& v);

template <class T> Vec3<T> cross(const Vec3<T> &p, const Vec3<T> &q) throw();
template <class T> T triple(const Vec3<T> &p, const Vec3<T> &q,const Vec3<T> &r) throw();
template <class T> void Tilde(const Vect<3,T>& v,_Mat<3,3,T>& M) throw();

/*! \class Vec3 Vec3.h linearAlgebra/Vec3.h
  \ingroup liblinearAlgebra
  template for a vector of dimension 3 and type T
*/
template <class T> class Vec3 : public Vect<3,T> {
public:
  /** @name Constructors */
  //@{
  /** @name Vectorial redefined constructors */
  //@{
  /** default constructor - initializes to 0 if UNINIT_ZERO not defined */
  vec_Inline Vec3() throw() : Vect<3,T>()
    {}
  /** constructor - initializes to 0 if b==VIM_INIT */
  explicit vec_Inline Vec3(VectInitMode b) throw() : Vect<3,T>(b)
    {}
  //@}
  /** @name Vector redefined constructors */
  //@{
#ifdef  __GNUC__
  /** copy constructor
      @param v vector to copy */
  vec_Inline Vec3(const Vect<3,T>& v) throw() : Vect<3,T>(v)
    {}
  /** operator =
      @param v vector to copy
      @return this */
  vec_Inline Vec3< T>& operator=(const Vect<3,T>& v) throw()
    {copy(v);return *this;}
#endif
  /** copy constructor with cast
      @param v vector to copy */
  template<class T2> vec_Inline Vec3(const Vect<3,T2>& v) throw() : Vect<3,T>(v)
    {}
  /** operator = with cast
      @param v vector to copy
      @return this */
  template<class T2> vec_Inline Vec3< T>& operator=(const Vect<3,T2>& v) throw()
    {copy(v);return *this;}

#ifndef WIN32
  /** constructor: initialize to vector (coord[0], coord[1], ...)
      @param coord vector coordinates array */
  vec_Inline Vec3(const T coord[3]) throw() : Vect<3,T>(coord)
    {}
#endif
  /** constructor: initialize to vector (coord[0], coord[1], ...)
      @param coord vector coordinates array */
  template<class T2> vec_Inline Vec3(const T2 coord[3]) throw() : Vect<3,T>(coord)
    {}

  /** constructor: initializes element i with 1, other with 0 (canonical basis vector i) */
  explicit vec_Inline Vec3(int i) throw(ErrRange) : Vect<3,T>(i)
    {}
  //@}

  /** constructor: initialize to vector (x,y)
      @param x first vector coordinate
      @param y second vector coordinate
      @param z third vector coordinate */
  Vec3(T x, T y, T z) throw();

  /** @name Data members access */
  //@{
  /** set this vector to vector (x,y,z)
      @param x first vector coordinate
      @param y second vector coordinate
      @param z third vector coordinate */
  void setValue(T x, T y, T z) throw();

  /// returns X coordinate
  T getX(void) const throw();

  /// returns Y coordinate
  T getY(void) const throw();

  /// returns Z coordinate
  T getZ(void) const throw();

#ifdef WIN32
#ifndef __MINGW32__
  inline T &operator[](unsigned int i) throw()
    {return vec_getRefVal(val,3,i);}
  inline T &operator[](int i) throw()
    {return vec_getRefVal(val,3,i);}
#endif
#endif
  //@}

  /** @name mathematical operators */
  //@{
  /** @name Vectorial redefined menber functions */
  //@{
  /** set this vector to p+q
      @return this = p + q */
  vec_Inline Vec3< T>& Minus(const Vec3< T> &p) throw()
    {vec_opposite(this->val,3,p.val);return *this;}
  /** set this vector to p+q
      @return this = p + q */
  vec_Inline Vec3< T>& Add(const Vec3< T> &p, const Vec3< T> &q) throw()
    {vec_add(this->val,3,p.val,q.val);return *this;}
  /** set this vector p = {px,py,pz} to {px+q,py+q,pz+q}
      @return this = {px+q,py+q,pz+q} */
  vec_Inline Vec3< T>& Add(const Vec3< T> &p, const T &q) throw()
    {vec_add(this->val,3,p.val,q);return *this;}


  /** set this vector to p-q
      @return this = p - q */
  vec_Inline Vec3< T>& Sub(const Vec3< T> &p, const Vec3< T> &q) throw()
    {vec_sub(this->val,3,p.val,q.val);return *this;}
  /** set each coordinate of this vector to product of this and p same coordinate
      @return this = this.*q */
  vec_Inline Vec3< T>& CMul(const Vec3< T> &q) throw()
    {vec_cmul(this->val,3,q.val);return *this;}
  /** set each coordinate of this vector to product of q and p same coordinate
      @return this = p.*q */
  vec_Inline Vec3< T>& CMul(const Vec3< T> &p, const Vec3< T> &q) throw()
    {vec_cmul(this->val,3,p.val,q.val);return *this;}
  /** set this vector to p*a
      @return this = p*a */
  vec_Inline Vec3< T>& Mul(const Vec3< T> &p, const T a) throw()
    {vec_mul(this->val,3,p.val,a);return *this;}
  /** set this vector to p*a
      @return this = p*a */
  vec_Inline Vec3< T>& Mul(const T a, const Vec3< T> &p) throw()
    {vec_mul(this->val,3,p.val,a);return *this;}
  /** set this vector to p/a
      @return this = p/a */
  vec_Inline Vec3< T>& Div(const Vec3< T> &p, const T a) throw(ErrDivZero)
    {vec_div(this->val,3,p.val,a);return *this;}
  /** adds vector q to vector p
      @param q vector to add
      @return p + q */
  vec_Inline Vec3< T>& operator+=(const Vec3< T> &q) throw()
    {vec_add(this->val,3,q.val);return *this;}
  /** substracts vector q from vector p
      @param q vector to substract
      @return p - q */
  vec_Inline Vec3< T>& operator-=(const Vec3< T> &q) throw()
    {vec_sub(this->val,3,q.val);return *this;}
  /** multiplies vector p by scalar a
      @param a scalar
      @return ap */
  vec_Inline Vec3< T>& operator*=(const T a) throw()
    {vec_mul(this->val,3,a);return *this;}
  /** divides vector p by scalar a
      @param a scalar
      @return p / a */
  vec_Inline Vec3< T>& operator/=(const T a) throw(ErrDivZero)
    {vec_div(this->val,3,a);return *this;}
  /** randomize vector
      @return this */
  vec_Inline Vec3< T>& Rnd(T min = (T)-1., T max = (T)1.) throw()
    {vec_rnd(this->val,3,min,max);return *this;}
  /** randomize vector : isotropic gaussian centred noise
      @return this */
  vec_Inline Vec3< T>& RndGaussian(T standard_deviation = (T)1.) throw()
    {vec_rndgaussian(this->val,3,standard_deviation);return *this;}
  /** takes the X,Y or Z value of p if they are less than its own X,Y,Z coordinates
      @return _Vect<3,T>(min(x,p.x),min(y,p.y),...) */
  vec_Inline Vec3< T>& Min(const Vec3< T>& p) throw()
    {vec_min(this->val,3,p.val);return *this;}
  /** takes the X,Y or Z value of p if they are greater than its own X,Y,Z coordinates
      @return _Vect<3,T>(max(x,p.x),max(y,p.y),...) */
  vec_Inline Vec3< T>& Max(const Vec3< T>& p) throw()
    {vec_max(this->val,3,p.val);return *this;}
  /** set each element of this vector to value
      @param v value
      @return this */
  vec_Inline Vec3< T>& Reset(T v) throw()
    {resetTo(v);return *this;}
  /** set this vector to null vector
      @return this */
  vec_Inline Vec3< T>& Null() throw()
  {resetTo((T)0);return *this;}
  /** set this vector to mean of a given list of vectors
      @return this */
  template<class InputIterator>
  vec_Inline Vec3< T>& Mean(InputIterator first, InputIterator last) throw()
    {mean(first,last);return *this;}
  //@}

  /** set this to result of cross product
      @param u first vector
      @param v second vector
      @return this = u ^ v */
  Vec3<T>& Cross(const Vec3<T>& u, const Vec3<T>& v) throw();
  /** set matrix to antisymetric matrix associated with cross product : M.u=v^u
      @param v vector
      @param M matrix */
  //friend void Tilde FRIEND_TEMPLATE_ARG(const Vect<3,T>& v,_Mat<3,3,T>& M) throw();
  /** cross product
      @param p first vector
      @param q second vector
      @return p ^ q */
  //friend Vec3<T> cross FRIEND_TEMPLATE_ARG(const Vec3<T> &p, const Vec3<T> &q) throw();
  /** triple product of 3 vectors
      @param p first vector
      @param q second vector
      @param r second vector
      @return [p,q,r] */
  //friend T triple FRIEND_TEMPLATE_ARG(const Vec3<T> &p, const Vec3<T> &q,const Vec3<T> &r) throw();

  /** @name Matrix operations */
  //@{
  /** @name Vector n redefined menber functions */
  //@{
  vec_Inline Vec3<T>& Mul(const Matrix<3,T> &m, const Vec3<T> &v) throw()
    {((Mat<3,1,T>*)this)->Mul(m,(const Mat<3,1,T>&)v);return *this;}
  vec_Inline Vec3<T>& Mul(const Vec3<T> &v, const Matrix<3,T> &m) throw()
    {((Mat<1,3,T>*)this)->Mul((const Mat<1,3,T>&)v,m);return *this;}
  //@}

  //  friend class HVec4<T>;
  /// cast to an homogeneous vector (x, y, z, 1)
  operator HVec4<T>() throw();
  /** assign vector from translation part of an homogeneous matrix
      @param m  homogeneous matrix */
  void operator=(const Matrix4x4<T> &m) throw();

  static void Test() throw();
};

} // namespace yav

/*][** NODOC ****************************************************************/
/****************************************************************************/

/*]]**************************************************************************/

namespace yav {

// constructors
template <class T> vec_Inline Vec3<T>::Vec3(T x, T y, T z) throw()
  : Vect<3,T>(VIM_NOINIT)
{
    this->val[0]=x; this->val[1]=y; this->val[2]=z;
}

// access to data
template <class T> vec_Inline void Vec3<T>::setValue(T x, T y, T z) throw()
{
    this->val[0]=x; this->val[1]=y; this->val[2]=z;
}
template <class T> vec_Inline T Vec3<T>::getX(void) const throw() {return this->val[0];}
template <class T> vec_Inline T Vec3<T>::getY(void) const throw() {return this->val[1];}
template <class T> vec_Inline T Vec3<T>::getZ(void) const throw() {return this->val[2];}

// mathematical operations
// cross product, triple product
template <class T> vec_Inline Vec3<T>& Vec3<T>::Cross(const Vec3<T> &p, const Vec3<T> &q) throw()
{
  this->val[0]=p.val[1]*q.val[2] - p.val[2]*q.val[1];
  this->val[1]=p.val[2]*q.val[0] - p.val[0]*q.val[2];
  this->val[2]=p.val[0]*q.val[1] - p.val[1]*q.val[0];
  return *this;
}
template<class T> void Tilde(const Vect<3,T>& v, _Mat<3,3,T>& M) throw()
{
  M(0,0) = M(1,1) = M(2,2) = 0.0;
  M(1,2) = - ( M(2,1) = v[0] );
  M(2,0) = - ( M(0,2) = v[1] );
  M(0,1) = - ( M(1,0) = v[2] );
};

template <class T> vec_Inline Vec3<T> cross(const Vec3<T> &p, const Vec3<T> &q) throw()
{  return Vec3<T>(VIM_NOINIT).Cross(p,q);}

template <class T> vec_Inline T triple(const Vec3<T> &p, const Vec3<T> &q, const Vec3<T> &r) throw()
{
  return (dot(cross(p,q),r));
}

template <class T> vec_Inline void Vec3<T>::operator=(const Matrix4x4<T> &m) throw() {
  for(unsigned int i = 0; i < 3; i++)
    this->val[i] = m.getValue(i, 3);
}

template <class T> void Vec3<T>::Test() throw()
{
  float Af[3];double Ad[3];int Ai[3];
  T A[3];for (Index k=0; k<3; k++) {A[k]=k+1;Af[k]=k+1;Ad[k]=k+1;Ai[k]=k+1;}
  Vec3<float> df;Vec3<double> dd;Vec3<int> di;

  Vec3<T> a(A),b(1),c(VIM_INIT),d(a),e(1,2,3),af(Af),ad(Ad),ai(Ai),ef(df),ed(dd),ei(di);
  T v;char TName[100];Xlinalg_Class_Name(v,TName);
  COUT << "Vec3<" << TName << ">" << ENDL;
  COUT << "Constructors" << ENDL;
  COUT << "a must be 1,2,... : " << a << ENDL;
  COUT << "af must be 1,2,... : " << af << ENDL;
  COUT << "ad must be 1,2,... : " << ad << ENDL;
  COUT << "ai must be 1,2,... : " << ai << ENDL;
  COUT << "b must be 0,1,... : " << b << ENDL;
  COUT << "c must be 0,0,... : " << c << ENDL;
  COUT << "d must be as a (copy) : " << d << ENDL;
  COUT << "e must be 1,2,3 : " << e << ENDL;
  COUT << "ef must be 0 : " << ef << ENDL;
  COUT << "ed must be 0 : " << ed << ENDL;
  COUT << "ei must be 0 : " << ei << ENDL;

  COUT << "operator=" << ENDL;
  d=c;COUT << "d=c : " << d << ENDL;
  d=df;COUT << "d=df : " << d << ENDL;
  d=dd;COUT << "d=dd : " << d << ENDL;
  d=di;COUT << "d=di : " << d << ENDL;

  COUT << "Access to Data" << ENDL;
  d.copy(b);COUT << "d.copy(b)  : " << d << ENDL;
  d[0]=1;COUT << "d[0]=1 : d[0] " << d[0] << ENDL;
  d(0)=1;COUT << "d(0)=1 : d(0) " << d(0) << ENDL;
  d.setValue(2,3,4);COUT << "d.setValue(2,3,4); : " << d << ENDL;
  COUT << "d.get.() : " << d.getX() << " " << d.getY() << " " << d.getZ() << ENDL;
  d.Null();COUT << "d.Null() : " << d << ENDL;
  d.resetTo(2);COUT << "d.resetTo(2) : " << d << ENDL;
  T*D=d.getValues();
  COUT << "getValues() : " << D[0] << " " << D[1] << " " << D[2] << ENDL;
  const T*E=d;
  COUT << "(const T*) d : " << E[0] << " " << E[1] << " " << E[2] << ENDL;
  COUT << "a.ColMat : " << a.ColMat() << ENDL;
  COUT << "a.RowMat : " << a.RowMat() << ENDL;

  COUT << "Math" << ENDL;
  COUT << "d.Minus(a) : " << d.Minus(a) << ENDL;
  COUT << "d.Add(a,b) : " << d.Add(a,b) << ENDL;
  COUT << "d.Sub(a,b) : " << d.Sub(a,b) << ENDL;
  COUT << "d.Mul(a,2) : " << d.Mul(a,(T)2) << ENDL;
  COUT << "d.Mul(2,a) : " << d.Mul((T)2,a) << ENDL;
  COUT << "d.Div(a,2) : " << d.Div(a,(T)2) << ENDL;
  COUT << "d.Div(a,2) : " << d.Div(a,(T)2) << ENDL;

  COUT << "+a : " << +a << ENDL;
  COUT << "-a : " << -a << ENDL;
  COUT << "a+b : " << a+b << ENDL;
  COUT << "a-b : " << a-b << ENDL;
  COUT << "a*2 : " << a*((T)2) << ENDL;
  COUT << "2*a : " << ((T)2)*a << ENDL;
  COUT << "a/2 : " << a/((T)2) << ENDL;

  d.copy(b);d+=a;COUT << "d=b; d+=a : " << d << ENDL;
  d.copy(b);d-=a;COUT << "d=b; d-=a : " << d << ENDL;
  d.copy(b);d*=(T)2;COUT << "d=b; d*=2 : " << d << ENDL;
  d.copy(b);d/=(T)2;COUT << "d=b; d/=2 : " << d << ENDL;

  d.copy(b);COUT << "d=b; d==b : " << (d==b) << ENDL;
  d.copy(b);COUT << "d=b; d!=b : " << (d!=b) << ENDL;

  COUT << "dot(a,b) : " << dot(a,b) << ENDL;
  COUT << "DotProd(a,b) : " << DotProd(a,b) << ENDL;
  COUT << "norm2(a) : " << a.norm2() << ENDL;
  COUT << "norm(a) : " << a.norm() << ENDL;
  COUT << "Norm2(a) : " << a.Norm2() << ENDL;
  COUT << "Norm(a) : " << a.Norm() << ENDL;
  COUT << "NormL1(a) : " << a.NormL1() << ENDL;
  COUT << "NormLinf(a) : " << a.NormLinf() << ENDL;
  COUT << "Dist2(a,b) : " << Dist2(a,b) << ENDL;
  COUT << "Dist(a,b) : " << Dist(a,b) << ENDL;
  COUT << "DistL1(a,b) : " << DistL1(a,b) << ENDL;
  COUT << "DistLinf(a,b) : " << DistLinf(a,b) << ENDL;

  d.resetTo(2); d.min(a); COUT << "d.resetTo(2); d.min(a) : " << d << ENDL;
  d.resetTo(2); d.max(a); COUT << "d.resetTo(2); d.max(a) : " << d << ENDL;

  d.Rnd(2); COUT << "d.Rnd(2) : " << d << ENDL;
  d.RndGaussian(2); COUT << "d.RndGaussian(2) : " << d << ENDL;

  // TODO ???
  {Matrix4x4<T> m; COUT << "Matrix4x4<T> m*d : " << m*d << ENDL;}
  {Matrix<3,T> m; COUT << "Matrix<3,T> m*d : " << m*d << "  Mul : " << c.Mul(m,d) << ENDL;}
  {Matrix<3,T> m; COUT << "d*Matrix<3,T> m : " << d*m << "  Mul : " << c.Mul(d,m) << ENDL;}

  COUT << "cross(a,b) : " << cross(a,b) << ENDL;
  COUT << "d.Cross(a,b) : " << d.Cross(a,b) << ENDL;
  Matrix3x3<T>Z(VIM_INIT);Tilde(a,Z);COUT << "Tilde(a,Z); Z : " << Z << " Z.b " << Z*b << ENDL;
  COUT << "triple(a,b,c) : " << triple(a,b,c) << ENDL;
  Matrix4x4<T> m(VIM_INIT);
  d=m;COUT << "operator=Matrix4x4 : " << d << ENDL;
  COUT << "operator HVec4 : " << (HVec4<T>) d << ENDL;

  COUT << ENDL;
}


} // namespace yav


#endif
