// General comments : see Vec.h

#ifndef _VEC4_H
#define _VEC4_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#ifndef WIN32
#pragma interface "linearAlgebra/Vec4.h"
#endif
#include "linearAlgebra/Vec.h"
#include "linearAlgebra/Mat.h"
#include "linearAlgebra/Vec3.h"
#include "linearAlgebra/Matrix3x3.h"

#ifdef WIN32
#include "basics/export.h"
#endif


namespace yav {


template <class T> class Vec4;
template <class T> class Matrix4x4;

/** @name Vectorial redefined friend functions */
//@{
/** unary +
    @param p vector
    @return +p */
template < class T> vec_Inline Vec4< T> operator+(const Vec4< T> &p) throw()
  {return p;}
/** unary -
    @param p vector
    @return -p */
template < class T> vec_Inline Vec4< T> operator-(const Vec4< T> &p) throw()
  {return Vec4< T>(VIM_NOINIT).Minus(p);}
/** vector sum
    @param p first vector
    @param q second vector
    @return p + q */
template < class T> vec_Inline Vec4< T> operator+(const Vec4< T> &p, const Vec4 < T> &q) throw()
  {return Vec4< T>(VIM_NOINIT).Add(p,q);}
/** vector substraction
    @param p first vector
    @param q second vector
    @return p - q */
template < class T> vec_Inline Vec4< T> operator-(const Vec4< T> &p, const Vec4 < T> &q) throw()
  {return Vec4< T>(VIM_NOINIT).Sub(p,q);}
/** vector multiplication by scalar
    @param p vector
    @param a scalar
    @return ap */
template < class T> vec_Inline Vec4< T> operator*(const Vec4< T> &p, const T a) throw()
  {return Vec4< T>(VIM_NOINIT).Mul(p,a);}
/** vector scalar multiplication
    @param a scalar
    @param p vector
    @return ap */
template < class T> vec_Inline Vec4< T> operator*(const T a, const Vec4< T> &p) throw()
  {return Vec4< T>(VIM_NOINIT).Mul(p,a);}
/** vector scalar division
    @param p vector
    @param a scalar
    @return p / a */
template < class T> vec_Inline Vec4< T> operator/(const Vec4< T> &p, const T a) throw(ErrDivZero)
  {return Vec4< T>(VIM_NOINIT).Div(p,a);}
/** dot product
    @param u first vector
    @param v second vector
    @return u.v */
template < class T> vec_Inline T dot(const Vec4< T>& u, const Vec4< T> &v) throw()
  {return Vec4< T>::dot(u,v);}
/// Square Euclidian Distance
template < class T> vec_Inline T dist2( const Vec4< T>& u, const Vec4< T>& v) throw()
  {return Vec4< T>::dist2(u,v);}
/// Euclidian Distance
template < class T> vec_Inline T dist( const Vec4< T>& u, const Vec4< T>& v) throw()
  {return Vec4< T>::dist(u,v);}
/// L1 Distance
template < class T> vec_Inline T distL1( const Vec4< T>& u, const Vec4< T>& v) throw()
  {return Vec4< T>::distL1(u,v);}
/// Linf Distance
template < class T> vec_Inline T distLinf( const Vec4< T>& u, const Vec4< T>& v) throw()
  {return Vec4< T>::distLinf(u,v);}
/** dot product
    @param p first vector
    @param q second vector
    @return p.q */
template < class T> vec_Inline T DotProd(const Vec4< T> &p, const Vec4< T> &q) throw()
  {return Vec4< T>::dot(p,q);}
/// Square Euclidian Distance
template < class T> vec_Inline T Dist2( const Vec4< T>& u, const Vec4< T>& v) throw()
  {return Vec4< T>::dist2(u,v);}
/// Euclidian Distance
template < class T> vec_Inline T Dist( const Vec4< T>& u, const Vec4< T>& v) throw()
  {return Vec4< T>::dist(u,v);}
/// L1 Distance
template < class T> vec_Inline T DistL1( const Vec4< T>& u, const Vec4< T>& v) throw()
  {return Vec4< T>::distL1(u,v);}
/// Linf Distance
template < class T> vec_Inline T DistLinf( const Vec4< T>& u, const Vec4< T>& v) throw()
  {return Vec4< T>::distLinf(u,v);}
/** operator == */
template < class T> vec_Inline bool operator == (const Vec4< T>& u, const Vec4< T> &v) throw()
  {return Vec4< T>::isEqual(u,v);}
/** operator != */
template < class T> vec_Inline bool operator != (const Vec4< T>& u, const Vec4< T> &v) throw()
  {return Vec4< T>::isDifferent(u,v);}
/** operator < */
template < class T> vec_Inline bool operator < (const Vec4< T>& u, const Vec4< T> &v) throw()
  {return Vec4< T>::isSmaller(u,v);}
/** operator > */
template < class T> vec_Inline bool operator > (const Vec4< T>& u, const Vec4< T> &v) throw()
  {return Vec4< T>::isGreater(u,v);}
/** operator <= */
template < class T> vec_Inline bool operator <= (const Vec4< T>& u, const Vec4< T> &v) throw()
  {return Vec4< T>::isSmallerOrEqual(u,v);}
/** operator >= */
template < class T> vec_Inline bool operator >= (const Vec4< T>& u, const Vec4< T> &v) throw()
  {return Vec4< T>::isGreaterOrEqual(u,v);}
//@}

/** @name Vector n redefined friend functions */
//@{
template <class T> vec_Inline Vec4<T> operator*(const Matrix<4,T> &m, const Vec4<T> &v) throw()
  {return Vec4<T>(VIM_NOINIT).Mul(m,v);}
template <class T> vec_Inline Vec4<T> operator*(const Vec4<T> &v, const Matrix<4,T> &m) throw()
  {return Vec4<T>(VIM_NOINIT).Mul(v,m);}
//@}

template <class T> void QuatProd(const Vec4<T>& q1, const Vec4<T>& q2, Vec4<T>& q) throw();
template <class T> void QuatConj(const Vec4<T>& q, Vec4<T>& barq) throw();
template <class T> void QuatInv(const Vec4<T>& q, Vec4<T>& invq) throw(ErrDivZero);
template <class T> void TildeLeft(const Vec4<T>& q, Matrix<4,T>& Q) throw();
template <class T> void Anti(const Vec4<T>& q, Matrix<4,T>& Q) throw();
template <class T> void TildeRight(const Vec4<T>& q, Matrix<4,T>& Q) throw();

/*! \class Vec4 Vec4.h linearAlgebra/Vec4.h
  \ingroup liblinearAlgebra
  template for a vector of dimension 4 and type T 
*/
template <class T> class Vec4 : public Vect<4,T> {
public:
  /** @name Constructors */
  //@{
  /** @name Vectorial redefined constructors */
  //@{
  /** default constructor - initializes to 0 if UNINIT_ZERO not defined */
  vec_Inline Vec4() throw() : Vect<4,T>()
    {}
  /** constructor - initializes to 0 if b==VIM_INIT */
  explicit vec_Inline Vec4(VectInitMode b) throw() : Vect<4,T>(b)
    {}
  //@}
  /** @name Vector redefined constructors */
  //@{
#ifdef  __GNUC__
  /** copy constructor
      @param v vector to copy */
  vec_Inline Vec4(const Vect<4,T>& v) throw() : Vect<4,T>(v)
    {}
  /** operator =
      @param v vector to copy
      @return this */
  vec_Inline Vec4< T>& operator=(const Vect<4,T>& v) throw()
    {copy(v);return *this;}
#endif
  /** copy constructor with cast
      @param v vector to copy */
  template<class T2> vec_Inline Vec4(const Vect<4,T2>& v) throw() : Vect<4,T>(v)
    {}
  /** operator = with cast
      @param v vector to copy
      @return this */
  template<class T2> vec_Inline Vec4< T>& operator=(const Vect<4,T2>& v) throw()
    {copy(v);return *this;}
  
#ifndef WIN32
  /** constructor: initialize to vector (coord[0], coord[1], ...)
      @param coord vector coordinates array */
  vec_Inline Vec4(const T coord[4]) throw() : Vect<4,T>(coord)
    {}
#endif
  /** constructor: initialize to vector (coord[0], coord[1], ...)
      @param coord vector coordinates array */
  template<class T2> vec_Inline Vec4(const T2 coord[4]) throw() : Vect<4,T>(coord)
    {}
  
  /** constructor: initializes element i with 1, other with 0 (canonical basis vector i) */
  explicit vec_Inline Vec4(int i) throw(ErrRange) : Vect<4,T>(i)
    {}
  //@}

  /** constructor: initialize to vector (x,y)
      @param a first vector coordinate
      @param b second vector coordinate
      @param c third vector coordinate 
      @param d fourth vector coordinate */
  Vec4(T a, T b, T c, T d) throw();

  /** constructor: initializes to quaternion (alpha,beta) */
  Vec4(const T alpha, const Vect<3,T>& beta) throw();

  /** @name Data members access */
  //@{
 public:
  /// Access to real part of a quat
  T RealPart() const throw();
  /// Access to vect part of a quat
  const Vec3<T>& VectPart() const throw() ;

  /// Ref to real part of a quat
  T& RealPart() throw();
  /// Ref to vect part of a quat
  Vec3<T>& VectPart() throw();

  /** set this vector to vector (x,y,z)
      @param x first vector coordinate
      @param y second vector coordinate
      @param z third vector coordinate
      @param t fourth vector coordinate */
  void setValue(T x, T y, T z, T t) throw();

  /// returns X coordinate
  T getX(void) const throw();

  /// returns Y coordinate
  T getY(void) const throw();

  /// returns Z coordinate
  T getZ(void) const throw();

  /// returns T coordinate
  T getT(void) const throw();
  //@}

  /** @name mathematical operators */
  //@{
  /** @name Vectorial redefined menber functions */
  //@{
  /** set this vector to p+q
      @return this = p + q */
  vec_Inline Vec4< T>& Minus(const Vec4< T> &p) throw()
    {vec_opposite(this->val,4,p.val);return *this;}
  /** set this vector to p+q
      @return this = p + q */
  vec_Inline Vec4< T>& Add(const Vec4< T> &p, const Vec4< T> &q) throw()
    {vec_add(this->val,4,p.val,q.val);return *this;}
  /** set this vector to p-q
      @return this = p - q */
  vec_Inline Vec4< T>& Sub(const Vec4< T> &p, const Vec4< T> &q) throw()
    {vec_sub(this->val,4,p.val,q.val);return *this;}
  /** set each coordinate of this vector to product of this and p same coordinate
      @return this = this.*q */
  vec_Inline Vec4< T>& CMul(const Vec4< T> &q) throw()
    {vec_cmul(this->val,4,q.val);return *this;}
  /** set each coordinate of this vector to product of q and p same coordinate
      @return this = p.*q */
  vec_Inline Vec4< T>& CMul(const Vec4< T> &p, const Vec4< T> &q) throw()
    {vec_cmul(this->val,4,p.val,q.val);return *this;}
  /** set this vector to p*a
      @return this = p*a */
  vec_Inline Vec4< T>& Mul(const Vec4< T> &p, const T a) throw()
    {vec_mul(this->val,4,p.val,a);return *this;}
  /** set this vector to p*a
      @return this = p*a */
  vec_Inline Vec4< T>& Mul(const T a, const Vec4< T> &p) throw()
    {vec_mul(this->val,4,p.val,a);return *this;}
  /** set this vector to p/a
      @return this = p/a */
  vec_Inline Vec4< T>& Div(const Vec4< T> &p, const T a) throw(ErrDivZero)
    {vec_div(this->val,4,p.val,a);return *this;}
  /** adds vector q to vector p
      @param q vector to add
      @return p + q */
  vec_Inline Vec4< T>& operator+=(const Vec4< T> &q) throw()
    {vec_add(this->val,4,q.val);return *this;}
  /** substracts vector q from vector p
      @param q vector to substract
      @return p - q */
  vec_Inline Vec4< T>& operator-=(const Vec4< T> &q) throw()
    {vec_sub(this->val,4,q.val);return *this;}
  /** multiplies vector p by scalar a
      @param a scalar
      @return ap */
  vec_Inline Vec4< T>& operator*=(const T a) throw()
    {vec_mul(this->val,4,a);return *this;}
  /** divides vector p by scalar a
      @param a scalar
      @return p / a */
  vec_Inline Vec4< T>& operator/=(const T a) throw(ErrDivZero)
    {vec_div(this->val,4,a);return *this;}
  /** randomize vector
      @return this */
  vec_Inline Vec4< T>& Rnd(T min = -1., T max = 1.) throw()
    {vec_rnd(this->val,4,min,max);return *this;}
  /** randomize vector : isotropic gaussian centred noise
      @return this */
  vec_Inline Vec4< T>& RndGaussian(T standard_deviation = 1.) throw()
    {vec_rndgaussian(this->val,4,standard_deviation);return *this;}
  /** takes the X,Y or Z value of p if they are less than its own X,Y,Z coordinates
      @return _Vect<4,T>(min(x,p.x),min(y,p.y),...) */
  vec_Inline Vec4< T>& Min(const Vec4< T>& p) throw()
    {vec_min(this->val,4,p.val);return *this;}
  /** takes the X,Y or Z value of p if they are greater than its own X,Y,Z coordinates
      @return _Vect<4,T>(max(x,p.x),max(y,p.y),...) */
  vec_Inline Vec4< T>& Max(const Vec4< T>& p) throw()
    {vec_max(this->val,4,p.val);return *this;}
  /** set each element of this vector to value
      @param v value
      @return this */
  vec_Inline Vec4< T>& Reset(T v) throw()
    {resetTo(v);return *this;}
  /** set this vector to null vector
      @return this */
  vec_Inline Vec4< T>& Null() throw()
  {resetTo((T)0);return *this;}
  /** set this vector to mean of a given list of vectors
      @return this */
  template<class InputIterator>
  vec_Inline Vec4< T>& Mean(InputIterator first, InputIterator last) throw()
    {mean(first,last);return *this;}
  //@}

  /// converts quaternion to tilde (4x4 matrix) representation
  void toTilde(Matrix<4,T> &mat) const throw();
  /// converts quaternion to anti (4x4 matrix) representation
  void toAnti(Matrix<4,T> &mat) const throw();
  /// converts quaternion to tilde right (4x4 matrix) representation
  void toTildeRight(Matrix<4,T> &mat) const throw();

  /** Quaternion product: q = q1 * q2
      @param q1 first quaternion
      @param q1 second quaternion
      @param q result */
  //friend void QuatProd FRIEND_TEMPLATE_ARG(const Vec4<T>& q1, const Vec4<T>& q2, Vec4<T>& q) throw();

  /** Conjugate quaternion: barq = (alpha, -beta)
      @param q quaternion
      @param barq result */
  //friend void QuatConj FRIEND_TEMPLATE_ARG(const Vec4<T>& q, Vec4<T>& barq) throw();

  /** Inverse quaternion: invq = barq/|q|^2
      @param q quaternion
      @param invq result */
  //friend void QuatInv FRIEND_TEMPLATE_ARG(const Vec4<T>& q, Vec4<T>& invq) throw(ErrDivZero);

  /** Antisymmetric matrix associated with the left Quaternion product:
                                                     |  0    -trp(beta)  |
     Q.p = q * p   with    Q = Tilde(q) = alpha I4 + |                   |
                                                     | beta  tilde(beta) |
      @param q quaternion
      @param Q result */
  //friend void TildeLeft FRIEND_TEMPLATE_ARG(const Vec4<T>& q, Matrix<4,T>& Q) throw();

  /** Antisymmetric matrix associated with the right Quaternion product by transposed q :
                                                       |  0     trp(beta)  |
     P.q = q * bar(p)   with  P = Anti(q) = alpha I4 + |                   |
                                                       | -beta  tilde(beta)|
      @param q quaternion
      @param Q result */
  //friend void Anti FRIEND_TEMPLATE_ARG(const Vec4<T>& q, Matrix<4,T>& Q) throw();

  /** Antisymmetric matrix associated with the Right Quaternion product:
                                                  |  0    -trp(beta)  |
     P.q = q * p   with  P = Anti(q) = alpha I4 + |                   |
                                                  | beta  -tilde(beta)|
      @param q quaternion
      @param Q result */
  //friend void TildeRight FRIEND_TEMPLATE_ARG(const Vec4<T>& q, Matrix<4,T>& Q) throw();

  /** @name Matrix operations */
  //@{
  /** @name Vector n redefined menber functions */
  //@{
  vec_Inline Vec4<T>& Mul(const Matrix<4,T> &m, const Vec4<T> &v) throw()
    {((Mat<4,1,T>*)this)->Mul(m,(const Mat<4,1,T>&)v);return *this;}
  vec_Inline Vec4<T>& Mul(const Vec4<T> &v, const Matrix<4,T> &m) throw()
    {((Mat<1,4,T>*)this)->Mul((const Mat<1,4,T>&)v,m);return *this;}
  //@}

  static void Test() throw();
};

typedef Vec4<double> Quaternion;

} // namespace yav

/*][** NODOC ****************************************************************/
/****************************************************************************/

/*]]**************************************************************************/



namespace yav {

/****************************************************************************/
/* Vec4 Specialisation                                                      */
/****************************************************************************/
// constructors
template <class T> vec_Inline Vec4<T>::Vec4(T a, T b, T c, T d) throw()
  : Vect<4,T>(VIM_NOINIT)
{
    this->val[0]=a; this->val[1]=b; this->val[2]=c; this->val[3]=d;
}

template <class T> vec_Inline Vec4<T>::Vec4(const T alpha, const Vect<3,T>& beta) throw()
{
  this->val[0] = alpha;
  this->val[1] = beta(0); this->val[2] = beta(1); this->val[3] = beta(2);
}

template<class T> T Vec4<T>::RealPart() const throw()
{ return this->val[0];}

template<class T> const Vec3<T>& Vec4<T>::VectPart() const throw()
{ return (const Vec3<T>&) this->val[1]; }


template<class T> T& Vec4<T>::RealPart() throw()
{ return this->val[0];}

template<class T> Vec3<T>& Vec4<T>::VectPart() throw()
{  return (Vec3<T>&) this->val[1]; }

template <class T> vec_Inline void Vec4<T>::setValue(T x, T y, T z, T t) throw()
{
    this->val[0]=x; this->val[1]=y; this->val[2]=z;
    this->val[3]=t;
}
template <class T> vec_Inline T Vec4<T>::getX(void) const throw() {return this->val[0];}
template <class T> vec_Inline T Vec4<T>::getY(void) const throw() {return this->val[1];}
template <class T> vec_Inline T Vec4<T>::getZ(void) const throw() {return this->val[2];}
template <class T> vec_Inline T Vec4<T>::getT(void) const throw() {return this->val[3];}

// Quaternion product: q = q1 * q2
// ----------------------------------------------------------------------
template<class T>
vec_Inline void QuatProd(const Vec4<T>& q1,
                     const Vec4<T>& q2,
                     Vec4<T>& q) throw()
{
  const T &a1 = q1.RealPart(), &a2 = q2.RealPart();
  const Vec3<T> &b1 = q1.VectPart();
  const Vec3<T> &b2 = q2.VectPart();
  Vec3<T> tmp,b;

  // alpha = a1.a2 - <b1|b2>
  T a = a1 * a2 - DotProd( b1, b2);
  // beta = a1.beta2 + a2.beta1 + beta1 x beta2
  b.Mul( a1, b2);
  b += tmp.Mul( a2, b1);
  b += tmp.Cross( b1, b2);
  // Assign result
  q.RealPart() = a;
  q.VectPart() = b;
};

// Conjugate quaternion: barq = (alpha, -beta)
// ----------------------------------------------------------------------
template<class T>
vec_Inline void QuatConj(const Vec4<T>& q,
                     Vec4<T>& barq) throw()
{
  barq.Mul( (T) -1., q);
  barq.RealPart() *= -1.;
}

// Inverse quaternion: invq = barq/|q|^2
// ----------------------------------------------------------------------
template<class T>
vec_Inline void QuatInv(const Vec4<T>& q,
                    Vec4<T>& invq) throw(ErrDivZero)
{
  T small_val;Xlinalg_Small_Val<T>(small_val);
  T nq2 = DotProd( q,q);
  if ( ::fabs(nq2) < small_val )
    throw ErrDivZero("void QuatInv(const Vect<4>& q, Vect<4>& invq)");
  invq.Mul( (T) -1./nq2, q);
  invq.RealPart() *= -1.;
}


template <class T>
void Vec4<T>::toTilde(Matrix<4,T> &mat) const throw()
{
  mat(0,0) = mat(1,1) = mat(2,2) = mat(3,3) = this->val[0];
  for(Index i=1; i < 4; i++) mat(0,i) = - ( mat(i,0) = this->val[i]);
  mat(2,3) = -( mat(3,2) = this->val[1] );
  mat(3,1) = -( mat(1,3) = this->val[2] );
  mat(1,2) = -( mat(2,1) = this->val[3] );
}

template <class T>
void Vec4<T>::toAnti(Matrix<4,T> &mat) const throw()
{
  mat(0,0) = mat(1,1) = mat(2,2) = mat(3,3) = this->val[0];
  for(Index i=1; i < 4; i++) mat(i,0) = - ( mat(0,i) = this->val[i]);
  mat(2,3) = -( mat(3,2) = this->val[1] );
  mat(3,1) = -( mat(1,3) = this->val[2] );
  mat(1,2) = -( mat(2,1) = this->val[3] );
}

template <class T>
void Vec4<T>::toTildeRight(Matrix<4,T> &mat) const throw()
{
  mat(0,0) = mat(1,1) = mat(2,2) = mat(3,3) = this->val[0];
  for(Index i=1; i < 4; i++) mat(0,i) = - ( mat(i,0) = this->val[i]);
  mat(3,2) = -( mat(2,3) = this->val[1] );
  mat(1,3) = -( mat(3,1) = this->val[2] );
  mat(2,1) = -( mat(1,2) = this->val[3] );
}

template<class T>
vec_Inline void TildeLeft(const Vec4<T>& q,Matrix<4,T>& Q) throw()
{q.toTilde(Q);}
template<class T>
vec_Inline void Anti(const Vec4<T>& q,Matrix<4,T>& Q) throw()
{q.toAnti(Q);}
template<class T>
vec_Inline void TildeRight(const Vec4<T>& q,Matrix<4,T>& Q) throw()
{q.toTildeRight(Q);}





template <class T> void Vec4<T>::Test() throw()
{
  float Af[4];double Ad[4];int Ai[4];
  T A[4];for (Index k=0; k<4; k++) {A[k]=(T)(k+1);Af[k]=(float)(k+1);Ad[k]=(double)(k+1);Ai[k]=int(k+1);}
  Vec4<float> df;Vec4<double> dd;Vec4<int> di;

  Vec4<T> a(A),b(1),c(VIM_INIT),d(a),e(1,2,3,4),ee(1,Vec3<T>(2,3,4)),af(Af),ad(Ad),ai(Ai),ef(df),ed(dd),ei(di);
  T v;char TName[100];Xlinalg_Class_Name(v,TName);
  COUT << "Vec4" << TName << ">" << ENDL;
  COUT << "Constructors" << ENDL;
  COUT << "a must be 1,2,... : " << a << ENDL;
  COUT << "af must be 1,2,... : " << af << ENDL;
  COUT << "ad must be 1,2,... : " << ad << ENDL;
  COUT << "ai must be 1,2,... : " << ai << ENDL;
  COUT << "b must be 0,1,... : " << b << ENDL;
  COUT << "c must be 0,0,... : " << c << ENDL;
  COUT << "d must be as a (copy) : " << d << ENDL;
  COUT << "e must be 1,2,3,4 : " << e << ENDL;
  COUT << "ee must be 1,2,3,4 : " << ee << ENDL;
  COUT << "ef must be 0 : " << ef << ENDL;
  COUT << "ed must be 0 : " << ed << ENDL;
  COUT << "ei must be 0 : " << ei << ENDL;

  COUT << "operator=" << ENDL;
  d=c;COUT << "d=c : " << d << ENDL;
  d=df;COUT << "d=df : " << d << ENDL;
  d=dd;COUT << "d=dd : " << d << ENDL;
  d=di;COUT << "d=di : " << d << ENDL;

  COUT << "Access to Data" << ENDL;
  d.copy(b);COUT << "d.copy(b)  : " << d << ENDL;
  d[0]=1;COUT << "d[0]=1 : d[0] " << d[0] << ENDL;
  d(0)=1;COUT << "d(0)=1 : d(0) " << d(0) << ENDL;
  d.setValue(2,3,4,5);COUT << "d.setValue(2,3,4,5); : " << d << ENDL;
  COUT << "d.get.() : " << d.getX() << " " << d.getY() << " " << d.getZ() << ENDL;
  d.Null();COUT << "d.Null() : " << d << ENDL;
  d.resetTo(2);COUT << "d.resetTo(2) : " << d << ENDL;
  T*D=d.getValues();
  COUT << "getValues() : " << D[0] << " " << D[1] << " " << D[2] << ENDL;
  const T*E=d;
  COUT << "(const T*) d : " << E[0] << " " << E[1] << " " << E[2] << ENDL;
  COUT << "a.ColMat : " << a.ColMat() << ENDL;
  COUT << "a.RowMat : " << a.RowMat() << ENDL;

  COUT << "Math" << ENDL;
  COUT << "d.Minus(a) : " << d.Minus(a) << ENDL;
  COUT << "d.Add(a,b) : " << d.Add(a,b) << ENDL;
  COUT << "d.Sub(a,b) : " << d.Sub(a,b) << ENDL;
  COUT << "d.Mul(a,2) : " << d.Mul(a,(T)2) << ENDL;
  COUT << "d.Mul(2,a) : " << d.Mul((T)2,a) << ENDL;
  COUT << "d.Div(a,2) : " << d.Div(a,(T)2) << ENDL;
  COUT << "d.Div(a,2) : " << d.Div(a,(T)2) << ENDL;

  COUT << "+a : " << +a << ENDL;
  COUT << "-a : " << -a << ENDL;
  COUT << "a+b : " << a+b << ENDL;
  COUT << "a-b : " << a-b << ENDL;
  COUT << "a*2 : " << a*((T)2) << ENDL;
  COUT << "2*a : " << ((T)2)*a << ENDL;
  COUT << "a/2 : " << a/((T)2) << ENDL;

  d.copy(b);d+=a;COUT << "d=b; d+=a : " << d << ENDL;
  d.copy(b);d-=a;COUT << "d=b; d-=a : " << d << ENDL;
  d.copy(b);d*=(T)2;COUT << "d=b; d*=2 : " << d << ENDL;
  d.copy(b);d/=(T)2;COUT << "d=b; d/=2 : " << d << ENDL;

  d.copy(b);COUT << "d=b; d==b : " << (d==b) << ENDL;
  d.copy(b);COUT << "d=b; d!=b : " << (d!=b) << ENDL;

  COUT << "dot(a,b) : " << dot(a,b) << ENDL;
  COUT << "DotProd(a,b) : " << DotProd(a,b) << ENDL;
  COUT << "norm2(a) : " << a.norm2() << ENDL;
  COUT << "norm(a) : " << a.norm() << ENDL;
  COUT << "Norm2(a) : " << a.Norm2() << ENDL;
  COUT << "Norm(a) : " << a.Norm() << ENDL;
  COUT << "NormL1(a) : " << a.NormL1() << ENDL;
  COUT << "NormLinf(a) : " << a.NormLinf() << ENDL;
  COUT << "Dist2(a,b) : " << Dist2(a,b) << ENDL;
  COUT << "Dist(a,b) : " << Dist(a,b) << ENDL;
  COUT << "DistL1(a,b) : " << DistL1(a,b) << ENDL;
  COUT << "DistLinf(a,b) : " << DistLinf(a,b) << ENDL;

  d.resetTo(2); d.min(a); COUT << "d.resetTo(2); d.min(a) : " << d << ENDL;
  d.resetTo(2); d.max(a); COUT << "d.resetTo(2); d.max(a) : " << d << ENDL;

  d.Rnd(2); COUT << "d.Rnd(2) : " << d << ENDL;
  d.RndGaussian(2); COUT << "d.RndGaussian(2) : " << d << ENDL;


  {Matrix<4,T> m; COUT << "Matrix<4,T> m*d : " << m*d << "  Mul : " << c.Mul(m,d) << ENDL;}
  {Matrix<4,T> m; COUT << "d*Matrix<4,T> m : " << d*m << "  Mul : " << c.Mul(d,m) << ENDL;}

  d.RealPart()=1;COUT << "RealPart=1 : RealPart= " << d.RealPart() << ENDL;
  d.VectPart()=Vec3<T>(1,2,3);COUT << "VectPart=(1,2,3) : VectPart= " << d.VectPart() << ENDL;

  QuatProd(a,b,c);COUT << "QuatProd(a,b,c)=" << c << ENDL;
  QuatConj(a,c);COUT << "QuatConj(a,c)=" << c << ENDL;
  QuatInv(a,c);COUT << "QuatInv(a,c)=" << c << ENDL;
  Matrix<4,T> m;
  TildeLeft(a,m);COUT << "TildeLeft(a,m)=" << m << ENDL;
  Anti(a,m);COUT << "Anti(a,m)=" << m << ENDL;
  TildeRight(a,m);COUT << "TildeRight(a,m)=" << m << ENDL;

  COUT << ENDL;
}



} // namespace yav


#endif
