#ifdef GRAPHICS


namespace yav {


template <class VEC, class ROT, class SCENE, class DATA, class CAMERA> 
Camera<VEC, ROT, SCENE, DATA, CAMERA>::Camera(SCENE *scn, unsigned int w,
					 unsigned int h, double _scale) 
  : scene(scn), scale(_scale > 0.0 ? _scale : DEFAULT_CAMERA_SCALE),
    width(w), height(h) {
  radius = DEFAULT_CAMERA_RADIUS;
  margin = DEFAULT_MARGIN;
  motionMode = MOVE_CAMERA;
  selectionMode = SELECT_OBJECT;

  translationFunction = NULL;
  rotationFunction = NULL;
  scaleFunction = NULL;

  autoFocus = true;
  focusedYet = false;
  showBoundingBox=false;
  backgroundColor[0]=0.0; backgroundColor[1]=0.0; backgroundColor[2]=0.0;
  boundingBoxColor[0]=1.0;boundingBoxColor[1]=1.0;boundingBoxColor[2]=1.0;
  boundingBoxWidth=1.0;
}

template <class VEC, class ROT, class SCENE, class DATA, class CAMERA> 
Camera<VEC, ROT, SCENE, DATA, CAMERA>::~Camera() {}

template <class VEC, class ROT, class SCENE, class DATA, class CAMERA> 
void Camera<VEC, ROT, SCENE, DATA, CAMERA>::displayBoundingBox()
{
  if (showBoundingBox == false)
    {
      showBoundingBox = true;
   //   redraw();
    }
}
template <class VEC, class ROT, class SCENE, class DATA, class CAMERA> 
void Camera<VEC, ROT, SCENE, DATA, CAMERA>::hideBoundingBox()
{
  if (showBoundingBox == true)
    {
      showBoundingBox = false;
    //  redraw();
    }
}
  
/* focus camera field of view if auto focus is on or no focus has not yet
   been done */
template <class VEC, class ROT, class SCENE, class DATA, class CAMERA> 
void Camera<VEC, ROT, SCENE, DATA, CAMERA>::focus(void) {
  if(autoFocus || !focusedYet) {
    doFocus();
 //   redraw();
  }
}


template <class VEC, class ROT, class SCENE, class DATA, class CAMERA> 
void Camera<VEC, ROT, SCENE, DATA, CAMERA>::resize(void) {
}

template <class VEC, class ROT, class SCENE, class DATA, class CAMERA> 
void Camera<VEC, ROT, SCENE, DATA, CAMERA>::resize(unsigned int w, unsigned int h) {
  if(w == 0) width = 1;
  else  width = w;
  if(h == 0) height = 1;
  else height = h;
  resize();
}

template <class VEC, class ROT, class SCENE, class DATA, class CAMERA> 
void Camera<VEC, ROT, SCENE, DATA, CAMERA>::setBackgroundColor(float r, float g, float b) {
  backgroundColor[BASE_RED] = r > 1.0f ? 1.0f : (r < 0.0f ? 0.0f : r);
  backgroundColor[BASE_GREEN] = g > 1.0f ? 1.0f : (g < 0.0f ? 0.0f : g);
  backgroundColor[BASE_BLUE] = b > 1.0f ? 1.0f : (b < 0.0f ? 0.0f : b);
//  redraw();
}


template <class VEC, class ROT, class SCENE, class DATA, class CAMERA> 
void Camera<VEC, ROT, SCENE, DATA, CAMERA>::setSelectedObject(DATA *obj)
{
  // test if the object exists and if it is different from the object allread selected
  if ((obj!=0) && (obj!=selectedObject))
    {
      // set the selected object
     selectedObject=obj; 
    }
}


template <class VEC, class ROT, class SCENE, class DATA, class CAMERA> 
void Camera<VEC, ROT, SCENE, DATA, CAMERA>::unsetSelectedObject(DATA *obj)
{
  // test if the object exists and if it is different from the object allread selected
  if ((obj==selectedObject))
    {
      // set the selected object
      selectedObject=0; 
    }
}

template <class VEC, class ROT, class SCENE, class DATA, class CAMERA> 
void  Camera<VEC, ROT, SCENE, DATA, CAMERA>::translateObject(VEC &t) {
  if (selectedObject)  {
    selectedObject->getData()->translate(t);
    selectedObject->getData()->translateBoundingBox(t);
    selectedObject->getData()->redrawScene();
  }
}

template <class VEC, class ROT, class SCENE, class DATA, class CAMERA> 
void  Camera<VEC, ROT, SCENE, DATA, CAMERA>::scaleObject(double s) {
  if (selectedObject) {
    selectedObject->getData()->scale(s);
    selectedObject->getData()->scaleBoundingBox(s);
    selectedObject->getData()->redrawScene();
  }
}

template <class VEC, class ROT, class SCENE, class DATA, class CAMERA> 
void  Camera<VEC, ROT, SCENE, DATA, CAMERA>::rotateObject(const ROT &rotation) {
  if (selectedObject) {
    selectedObject->getData()->rotate(rotation);
    selectedObject->getData()->rotateBoundingBox(rotation);
    selectedObject->getData()->redrawScene();
  }
}




template <class VEC, class ROT, class SCENE, class DATA, class CAMERA> 
void Camera<VEC, ROT, SCENE, DATA, CAMERA>::buttonPressed(WHICH_BUTTON w,
							  KEY_MODIFIER k,
							  int x, int y) {
#ifndef WIN32
	struct timeval tv;
#endif
  pressPoint = lastPoint = getTrackballVector(x, y);
  switch(k) {
  case NO_MODIFIER:
    B[w] = BUTTON_PRESSED;
    break;
  case CTRL_KEY:
    B[w] = CTRL_BUTTON_PRESSED;
    break;
  case SHIFT_KEY:
    B[w] = SHIFT_BUTTON_PRESSED;
    break;
  }
#ifndef WIN32
  gettimeofday(&tv, 0);
  clickTime = tv.tv_sec * 1000000 + tv.tv_usec;
#else
  clickTime=(LONGINT)time(0);
#endif
  renderDuration = 0;

  // send event to the seclected object
  if (selectedObject) {
    selectedObject->mouseButtonPressed(w,k,(CAMERA *) this,x,y);
  }

}

template <class VEC, class ROT, class SCENE, class DATA, class CAMERA> 
void Camera<VEC, ROT, SCENE, DATA, CAMERA>::buttonReleased(WHICH_BUTTON w,
							   int x, int y) {
#ifndef WIN32
	struct timeval tv;
#endif
  LONGINT releaseTime;
  VEC p = getTrackballVector(x, y);
#ifndef WIN32
  gettimeofday(&tv, 0);
  releaseTime = tv.tv_sec * 1000000 + tv.tv_usec;
#else
  releaseTime=(LONGINT) time(0);
#endif
  clickDuration = releaseTime - clickTime;

  switch(w) {

    // button 1
  case LEFT_BUTTON:
    switch(B[LEFT_BUTTON]) {

    // selection
    case BUTTON_PRESSED:
      if (selectedObject) {
	select(Vec2<double>(x, y), getType());
      }
      if (selectedObject)
	selectedObject->mouseButtonReleased(w,NO_MODIFIER,(CAMERA *) this,x,y);
      break;
    // world scale
    case CTRL_BUTTON_PRESSED:
      moveMode = ROTATION_MODE;
      if (selectedObject)
	selectedObject->mouseButtonReleased(w,CTRL_KEY,(CAMERA *) this,x,y);
      redraw();
     break;
    case SHIFT_BUTTON_PRESSED:
      // send event to the selected object
      if (selectedObject)
	selectedObject->mouseButtonReleased(w,SHIFT_KEY,(CAMERA *) this,x,y);
      break;
    default:

      break;
    }
    break;
    
  // button 2
  case MIDDLE_BUTTON:
    
    switch(B[MIDDLE_BUTTON]) {
    case SHIFT_BUTTON_PRESSED:
      // send event to the selected object
      if (selectedObject)
	selectedObject->mouseButtonReleased(w,SHIFT_KEY,(CAMERA *) this,x,y);
      break;
	
    default:
      break;
    }
    moveMode = ROTATION_MODE;
    redraw();
    break;
      
  case RIGHT_BUTTON:
  default:
    break;
  }

  B[w] = BUTTON_RELEASED;
  // bind the mouse motion to call mouseMoved()
}


template <class VEC, class ROT, class SCENE, class DATA, class CAMERA> 
void Camera<VEC, ROT, SCENE, DATA, CAMERA>::mouseMoved(int x, int y) {
  double s;
  VEC p = getTrackballVector(x, y);

  if(B[MIDDLE_BUTTON] != BUTTON_RELEASED ||
     B[LEFT_BUTTON] != BUTTON_RELEASED) {
    moveMode = ROTATION_MODE;

    // button 2
    switch(B[MIDDLE_BUTTON]) {

    case BUTTON_PRESSED:
      {
	ROT autoRotation = getTrackballRotation(lastPoint, p);
	lastPoint = p;
	switch(motionMode) {
	case CameraTypes::MOVE_CAMERA:
	  rotateCamera(autoRotation);
	  break;
	case CameraTypes::MOVE_CAMERA_OBJECT:
	  rotateCameraObject(autoRotation);
	  break;
	case CameraTypes::MOVE_SELECTED_OBJECT:
	  rotateObject(autoRotation);
	  break;
	}
      }
      break;
     
    case CTRL_BUTTON_PRESSED:
      {
	VEC autoTranslation = getTrackballTranslation(p, lastPoint);
	lastPoint = p;
     
	switch(motionMode) {
	case CameraTypes::MOVE_CAMERA:
	  translateCamera(autoTranslation);
	  break;
	case CameraTypes::MOVE_CAMERA_OBJECT:
	  translateCameraObject(autoTranslation);
	  break;
	case CameraTypes::MOVE_SELECTED_OBJECT:
	  translateObject(autoTranslation);
	  break;
	}
      }
      break;

    case SHIFT_BUTTON_PRESSED:
      {
	
	VEC autoTranslation= getTrackballZTranslation( p.getY() - lastPoint.getY());
	lastPoint = p;
	switch(motionMode) {
	case CameraTypes::MOVE_CAMERA:
	  translateCamera(autoTranslation);
	  break;
	case CameraTypes::MOVE_CAMERA_OBJECT:
	  translateCameraObject(autoTranslation);
	  break;
	case CameraTypes::MOVE_SELECTED_OBJECT:
	  translateObject(autoTranslation);
	  break;
	}
      }
      break;
    
    case BUTTON_RELEASED:
      break;
    }

    /* button 1 */
    switch(B[LEFT_BUTTON]) {
    case BUTTON_PRESSED:
      if (selectedObject)
	selectedObject->mouseButtonMoved((CAMERA *) this,x,y);
     break;

    case CTRL_BUTTON_PRESSED:
      s = getTrackballScale(p, lastPoint);
      lastPoint = p;
      switch(motionMode) {
      case CameraTypes::MOVE_CAMERA:
	scaleCamera(s);
	break;
      case CameraTypes::MOVE_CAMERA_OBJECT:
	scaleCameraObject(s);
	break;
      case CameraTypes::MOVE_SELECTED_OBJECT:
	scaleObject(s);
	if (selectedObject)
	  selectedObject->mouseButtonMoved((CAMERA *) this,x,y);
	break;
      }
      break;
    case SHIFT_BUTTON_PRESSED:
      // send event to the selected object
      if (selectedObject)
	selectedObject->mouseButtonMoved((CAMERA *) this,x,y);
      break;
    case BUTTON_RELEASED:
      break;
    }
  }
}


template <class VEC, class ROT, class SCENE, class DATA, class CAMERA> 
void Camera<VEC, ROT, SCENE, DATA, CAMERA>::redraw(void) {
  typename ObjectsList< DATA >::iterator it(scene->firstData());
  for(; it!=scene->lastData(); it++) {
    if (selectedObject && ((*it)==dynamic_cast<DATA *>(selectedObject->getData())) && (showBoundingBox==true))
      drawBoundingBox((*it));
    (*it)->redrawAll(getType());
  }
  //  grab();
}
} // namespace yav

#endif // ifdef GRAPHICS
