#ifndef CAMERASELECTION_H
#define CAMERASELECTION_H


#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#ifdef GRAPHICS


/// cameras parent class
template <class DATA, class CAMERA>
class Camera Selection {
 public:
  /// the type of function used to select sub-items of a Data
   typedef void (*SelectionFunction)(DATA *,const Vec2<double> pixel,
				     CAMERA *cam);
   /// selection mode
  typedef enum {
    SELECT_OBJECT,
    OBJECT_DEFINED
  } SELECTION_MODE;  

 protected:
  /// the camera that is being used for selection
  CAMERA camera;
   /// what to do when  the left button is pushed
   SELECTION_MODE selectionMode;
   /// a callback function for user defined selection
   SelectionFunction selectionFunction;
   /// whether record camera at each redraw
   bool record;
   /// movie frame number
   unsigned int frameNumber;
   /// movie base file name
   TclString movieName;
 public:
 /** @name Constructor */
  //@{
  /** creates a new camera
      @param h height in pixel size
      @param w width in pixel size
      @param scale default scale value */
   CameraSelection(CAMERA *cam) :camera(cam){
     selectionMode = SELECT_OBJECT;
     selectionFunction =NULL;
   }

  //@}

  /** @name Destructor */
  //@{
   virtual ~CameraSelection(){}
  //@}

   /** set selection mode
       @param mm Camera selection mode */
   void setSelectionMode(SELECTION_MODE mm) { selectionMode = mm; }
 /** select a new object
      @param obj object to select */
   virtual void setSelectedObject(DATA *obj) {
     // test if the object exists and if it is different from the object allread selected
     if ((obj!=0) && (obj!=camera->getClient()->getSelectedObject()))
       {
	 //  Test if there is a selected object
	 if(camera->getClient()->getSelectedObject ()!= 0)
	   // if so unselect it
	   applyObjectUnselection(selectedObject);
	 // set the selected object
	 camera->getClient()->setSelectedObject(obj); 
	 // call the select() function
	 applyObjectSelection(obj);
	 //redraw to update  the bounding box
	 camera->getClient()->redraw();
       }
   }
  /** unselect an object
      @param obj object to unselect */
   virtual void unsetSelectedObject (DATA *obj) {
     // test if the object exists and if it is different from the object allread selected
     if  (obj==camera->getClient()->getSelectedObject())
       {
	 applyObjectUnselection(obj);
	 // set the selected object
	 camera->getClient()->unsetSelectedObject(obj);
       }

   }
     /** Performes some task before activating the scene 
      pointed by the camera */
  virtual void startComputation (){}
  /** Performes some task before desactivating the scene 
      pointed by the camera */
  virtual void stopComputation (){}
  /** virtual object selection callback */
  virtual void applyObjectSelection(DATA *obj) { obj->selected(); }
  /** virtual object unselection callback */
  virtual void applyObjectUnselection(DATA *obj) { selectionFunction=0;
  obj->unselected(); }
  /** find and select closest object from a pixel
      @param pixel selection pixel */
  virtual void pickObject(const Vec2<double> &pixel) = 0;
  /** set scale function on subobject
      @param sf new scale function */  
  void setSelectionFunction(SelectionFunction sf) {
    selectionFunction = sf;
  }
  /** subobject selection method
      @param vec mouse click point
      @param cam clicked camera */
  virtual void select(const Vec2<double>& vec, CAMERA *cam) {
    if (selectionFunction)
      (*selectionFunction) (selectedObject,vec,cam);
    else
      selectedObject->selectObject(vec, cam);
  }
  /// is the camera recording  each frame ?
  bool isRecording() const {
    return record;
  }
  /** start camera recording
      @param name base file name */
  void startMovie(const char *name){
    movieName = name;
    frameNumber = 0;
    record = true;
  }
  /** stop camera recording */
  void stopMovie(void) {
    record=false;
  }
  /** dump camera window content in an xwd file
      @param filename dump file name */
  virtual void grab(const TclString& filename) {}

  /// call grab(const TclString&) to record camera image in a file if requested
    void grab(void) {
      if(record) {
	TclString filename;
	filename.sprintf("%s%04u.xwd", (char *) movieName, frameNumber++);
	grab(filename);
      }
    }

  /// when you want to record a complete widget owning the camera
  TclString widgetRecordingName;
  void setwidgetRecordingName(const TclString& _name) {
    if(_name != EMPTY_STRING) widgetRecordingName = _name;
  }
  /** pause camera recording
      @param state pause if state = true, record if state = false */
  void pauseMovie(bool state) {
    record = !b;
  }


#endif

#endif
