#ifndef CAMERATYPES_H
#define CAMERATYPES_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#ifdef GRAPHICS

#include <framework/Client.h>
namespace yav {

/** \class CameraTypes
    \brief cameras base class
    \ingroup Graphics
*/
class CameraTypes : public Client {
 public:
  /// RGB color index
  typedef enum {
    /// red
    BASE_RED = 0,
    /// green
    BASE_GREEN = 1,
    /// blue
    BASE_BLUE = 2
  } BASE_COLOR;
  
  /// mouse button state
  typedef enum {
    /// button was released
    BUTTON_RELEASED,
    /// button was pressed
    BUTTON_PRESSED,
    /// button was pressed with CTRL key modifier
    CTRL_BUTTON_PRESSED,
    /// button was pressed with SHIFT key modifier
    SHIFT_BUTTON_PRESSED
  } BUTTON_STATE;
  
  /// selection mode
  typedef enum {
    SELECT_OBJECT,
    OBJECT_DEFINED
  } SELECTION_MODE;  

  /// key modifier (with mouse button)
  typedef enum {
    /// no key pressed
    NO_MODIFIER,
    /// CTRL key pressed
    CTRL_KEY,
    /// SHIFT key pressed
    SHIFT_KEY,
  } KEY_MODIFIER;

  /// mouse button
  typedef enum {
    /// left mouse button
    LEFT_BUTTON = 0,
    /// middle (or both) mouse button
    MIDDLE_BUTTON = 1,
    /// right mouse button
    RIGHT_BUTTON = 2
  } WHICH_BUTTON;

  /// scene or camera move
  typedef enum {
    /// move camera location
    MOVE_CAMERA,
    /// call back camera to handle move
    MOVE_CAMERA_OBJECT,
    /// move selected object
    MOVE_SELECTED_OBJECT
  } MOTION_MODE;


  /// camera move mode
  typedef enum {
    /// camera navigates (translations)
    NAVIGATION_MODE,
    /// camera rotates from its center point
    ROTATION_MODE
  } MOVE_MODE;
  
  /// camera permanent mode
  typedef enum {
    /// no move
    NO_AUTO_MOVE,
    /// rotate camera
    AUTO_ROTATION_MOVE,
    /// translate camera
    AUTO_TRANSLATION_MOVE,
    /// scale camera
    AUTO_SCALE_MOVE
  } AUTO_MOVE_MODE;
  
  /// redraw kind
  typedef enum {
    /// redraw at creation time
    CREATION,
    /// redraw at destruction time
    DESTRUCTION,
    /// updates objects and redraw
    UPDATE,
    /// redraw
    REDRAW
  } RedrawOrigin; 

};

}


#endif

#endif
