#ifndef DATA_C
#define DATA_C

#include <basics/AutoLink.h>
#include <graphics/Camera.h>


namespace yav {

template <class SCENE, class DATA>
Data<SCENE, DATA>::Data(SCENE *scn, bool excl)
  :  scene(0), exclusive(excl) {

  if(!scn) {
    // If there are no existing scene
    if (AutoLink< SCENE >::getCount() == 0) {
      // creates a new scene
      scn = new SCENE();
    }
    else {
      // Otherwise gets the first in list
      scn = AutoLink< SCENE >::getHead();
    }
  }
  
  if(exclusive) {
    if(scn->getExclusivity()) scn = new SCENE();
  }
  scene = scn;

  scene->addData((DATA*) this);
}

template <class SCENE,  class DATA>
void Data<SCENE, DATA>::setScene(SCENE *scn) {
  if(scene && scene != scn) scn->addData((DATA *) this);
}

#if (defined GRAPHICS || defined OPENGL)
template <class VEC, class ROT, class SCENE, class CAMERA, class DATA>
GraphicsData<VEC, ROT, SCENE, CAMERA, DATA>::GraphicsData(SCENE *scn, bool excl)
  : Data<SCENE,DATA>(scn,excl),center(), radius(1.0) {

  boundingBox[0] = VEC();
  boundingBox[1] = VEC();
}


template <class VEC, class ROT, class SCENE, class CAMERA, class DATA>
GraphicsData<VEC, ROT, SCENE, CAMERA, DATA>::~GraphicsData() {
}


  /*
template <class VEC, class ROT, class SCENE, class CAMERA, class DATA>
void GraphicsData<VEC, ROT, SCENE, CAMERA, DATA>::selectScene() {
  scene->selectCameraObject(getData());
}

template <class VEC, class ROT, class SCENE, class CAMERA, class DATA>
void GraphicsData<VEC, ROT, SCENE, CAMERA, DATA>::unselectScene() {
  scene->unselectCameraObject(getData());
}
  */

template <class VEC, class ROT, class SCENE, class CAMERA, class DATA>
void GraphicsData<VEC, ROT, SCENE, CAMERA, DATA>::redrawScene() {
  this->scene->redraw();
}

template <class VEC, class ROT, class SCENE, class CAMERA, class DATA>
void GraphicsData<VEC, ROT, SCENE, CAMERA, DATA>::focusScene() {
  this->scene->focus();
}

template <class VEC, class ROT, class SCENE, class CAMERA, class DATA>
void GraphicsData<VEC, ROT, SCENE, CAMERA, DATA>::translateBoundingBox(const VEC &t) {  
  boundingBox[0] += t;
  boundingBox[1] += t;
}

template <class VEC, class ROT, class SCENE, class CAMERA, class DATA>
void GraphicsData<VEC, ROT, SCENE, CAMERA, DATA>::rotateBoundingBox(const ROT& rotation) {
  ROT tmp(orientation);
  orientation = rotation;
  orientation *= tmp;
}

template <class VEC, class ROT, class SCENE, class CAMERA, class DATA>
void GraphicsData<VEC, ROT, SCENE, CAMERA, DATA>::scaleBoundingBox(double scale) {
  boundingBox[0] = center + (boundingBox[0] - center) * scale;
  boundingBox[1] = center + (boundingBox[1] - center) * scale;
}



template <class VEC, class ROT, class SCENE, class CAMERA, class DATA>
void GraphicsData<VEC, ROT, SCENE, CAMERA, DATA>::redrawAll(CAMERA *cam) {
  redraw(cam);
  if (callbackList.size()>0) {
    typename std::list<RedrawCallback *>::iterator it=callbackList.begin();
    for(;it!=callbackList.end();it++) {
      (*(*it)->redrawFunction)(this->getData(),cam,(*it)->clientData);
    }
  }
}

template <class VEC, class ROT, class SCENE, class CAMERA, class DATA>
void GraphicsData<VEC, ROT, SCENE, CAMERA, DATA>:: addRedrawCallback(DATA_REDRAW_FUNCTION f,void *c) {
  if (f) {
    RedrawCallback *rc=new RedrawCallback;
    rc->redrawFunction=f;
    rc->clientData=c;
    callbackList.push_back(rc);
  }
}
template <class VEC, class ROT, class SCENE, class CAMERA, class DATA>
void GraphicsData<VEC, ROT, SCENE, CAMERA, DATA>:: removeRedrawCallback(DATA_REDRAW_FUNCTION f) {

  if (callbackList.size()>0) {
    typename std::list<RedrawCallback *>::iterator it=callbackList.begin();
    for(;(it!=callbackList.end())&& ((*it)->redrawFunction!=f);it++);
    if ((*it)->redrawFunction==f) 
      callbackList.erase(it);
  }
}

#endif   // if (defined GRAPHICS || defined OPENGL)

} // namespace yav

#endif   // ifndef DATA_C
