#ifndef DEFORMABLEMODEL_H
#define DEFORMABLEMODEL_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#pragma interface "graphics/DeformableModel.h"
#include <basics/TclString.h>
#include <basics/ListElement.h>

namespace yav {

/** \class DeformableModel
    \ingroup Graphics
    defines a deformable as an object with an iterate function, a set 
    of position  and force constraint and some function to manage its
    number of iterations and the amount of time spent in the iterate function */
template <class SCENE, class DEFORMABLE >
class DeformableModel :  public ListElement<DEFORMABLE > {
 public:
   /** @name constructor */
  //@{
  /** builds a new DeformableModel
      @param scene  scene where the range data object should be added (if scene is 0, the first existing scene is used or a scene is created== 0) */
  DeformableModel(SCENE *scene=0);
  //@}

  /** @name destructor */
  //@{
  /// empty
  virtual ~DeformableModel(){}
  //@}
  protected:
  /** a boolean indicated whether the deformable should be deformed by its scene. As default it is set to true */
  bool active;
 public:
  /// the total number of iteration
  unsigned int nbIterations;
  /// the time spent for deforming the model
  unsigned LONGINT deformationDuration;
  
 public:
  /** routine that performs some deformation iterations on a deformable model
      @param n the number of iterations to performed. Can be set to zero */
  virtual void iterate(const unsigned int n)=0;
  /// return the value of the active boolean value
  bool getActivation() const {
    return(active);}
  /// set the value of the active boolean value
  void setActivation(const bool activation)  {
    active=activation;}

  /// whether this deformable model should still be deformed
  virtual bool continueDeformation() {
    return(true);}
  /// returns the time spent at deforming the deformable model
  unsigned LONGINT getDeformationDuration() const {
    return(deformationDuration);}
  /// sets the time spent at deforming the deformable model
  void setDeformationDuration(const unsigned LONGINT duration)  {
     deformationDuration=duration;}
  /// get the total number of iterations used to deform the deformable model
   unsigned int  getNbIterations() const  {
     return(nbIterations);}
   
  /** change object' scene
      @param scene new object scene */
  void setScene(SCENE *scene);


};

}
#include "DeformableModel.txx"

#endif
