
#ifndef RANGEDATA_H
#define RANGEDATA_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics/RangeData.h"
#include <basics/TclString.h>
#include <basics/ListElement.h>
#include <framework/RangeInformation.h>

namespace yav {
  

 
/** \class RangeData
    \ingroup Graphics
    generic range data that exists in a Scene. Those objects
    provide a member function that computes the closest data point 
    from any point */
template <class VEC, class SCENE, class RANGE >
class RangeData : public ListElement<RANGE >
{

 protected:
  /// range data scene
  SCENE *scene;

 public:
  /** @name constructor */
  //@{
  /** builds a new RangeData
      @param scene  scene where the range data object should be added (if scene is 0, the first existing scene is used or a scene is created== 0) */
  RangeData(SCENE *scene = 0);
  //@}

  /** @name destructor */
  //@{
  /// empty
  virtual ~RangeData() {}
  //@}

 public:
  /** routine that computes the closest point on a dataset from a given point and normal direction 
      @param position the position of the point from which the closest point should be computed 
      @param normalDirection the normal direction of the point from which the closest point should be computed 
      @param rangeSearchDistance the maximum distance (= cut-off distance) along the normal direction at which we should search for the closest point
      @param closestPoint the closestPoint of the dataset from a given position and normal direction. This output value is only set if a closest point were found. Otherwise it is undetermined
      @param clientData a data structure that can be used to constrain the search of the closest point. This pointer may be 0
      @return a boolean equal to true if a closest Point were found and false otherwise */
  virtual bool getClosestPoint(const VEC &position,
			       const VEC &normalDirection,
			       const double rangeSearchDistance,
			       VEC &closestPoint,
			       const RangeInformation *clientData) const = 0;

  /** compute level set speed as 1/nabla I| truncated function
      @param position the position of a 3D point in the image
      @param mul returned multiplicative speed term
      @param add returned additive speed term 
      @param clientData a data structure that can be used to constrain the search of the closest point. This pointer may be 0*/
  virtual void getSpeed(const VEC &position, double &mul, double &add,
			const RangeInformation *clientData)
    const = 0;

  /// returns object scene
  SCENE *getScene(void) const { return scene; }

  /** change object' scene
      @param scene new object scene */
  void setScene(SCENE *scene);


};

}
#include "RangeData.txx"


#endif
