#ifdef WIN32
#include <time.h>
#endif

namespace yav {

template <class DATA,  class USERINTERFACE, class SCENE>
bool Scene<DATA,  USERINTERFACE, SCENE>::getExclusivity(void) const {
  typename ObjectsList< DATA >::iterator it = dataList.begin();
  for(; it!=dataList.end(); it++)
    if((*it)->getExclusivity()) return true;
  return false;
}


template <class DATA,  class USERINTERFACE, class SCENE>
Scene< DATA,  USERINTERFACE, SCENE>::~Scene() {
  while(!dataList.isEmpty()) delete dataList.getHead();
}


template <class DATA,  class USERINTERFACE, class SCENE>
void Scene<DATA,  USERINTERFACE, SCENE>::addData(DATA * d)
  /*throw (typename ObjectsList< DATA >::AlreadyInList,
    typename ObjectsList< DATA >::NullPointer)*/ {
  
  if(d->getScene()) d->disconnect();

  try{
    /* add in list of elements */
    dataList.append(d);
  }
  catch (typename ObjectsList< DATA >::AlreadyInList e) {
    e.push("Scene<DATA,  USERINTERFACE, SCENE>::addData(const DATA *)");
    throw(e);
  }
  catch(typename ObjectsList< DATA >::NullPointer e) {
    e.push("Scene<DATA,  USERINTERFACE, SCENE>::addData(const DATA *)");
    throw(e);
  }
  d->setScene( dynamic_cast<SCENE *>(this));
}







template <class DATA,  class USERINTERFACE, class SCENE>
void Scene<DATA,  USERINTERFACE, SCENE>::removeData(DATA *d)
  /*throw(typename ObjectsList< DATA >::NotFound)*/ {
  try{
    /* remove from the list of elements */
    dataList.remove(d);
  }
  catch (typename ObjectsList< DATA >::NotFound e) {
    e.push("Scene<DATA,  USERINTERFACE, SCENE>::removeData(const DATA *)");
    throw(e);
  }
  /* add the scene in the list of */
  delete d;
}


template <class DATA,  class USERINTERFACE, class SCENE>
void  Scene<DATA,  USERINTERFACE, SCENE>::addUserInterface(USERINTERFACE *d) 
  /*throw (typename ObjectsList< USERINTERFACE >::AlreadyInList,
    typename ObjectsList< USERINTERFACE >::NullPointer)*/ {
  
  d->disconnect();
  try{
    userInterfaceList.append(d);
  }
  catch(typename ObjectsList< USERINTERFACE >::AlreadyInList e) {
    e.push("GraphicsScene<VEC, DATA,  USERINTERFACE, CAMERA, SCENE>::addUserInterface(const USERINTERFACE *)");
    throw(e);
  }
  catch(typename ObjectsList< USERINTERFACE >::NullPointer e) {
    e.push("GraphicsScene<VEC, DATA,  USERINTERFACE, CAMERA, SCENE>::addUserInterface(const USERINTERFACE *)");
    throw(e);
  }
}

template <class DATA,  class USERINTERFACE, class SCENE>
void Scene <DATA,  USERINTERFACE, SCENE>::removeUserInterface(USERINTERFACE *d)
  /*throw (typename ObjectsList< USERINTERFACE >::NotFound)*/ {
  try{
    userInterfaceList.remove(d);
  }
  catch (typename ObjectsList< USERINTERFACE >::NotFound e) {
    e.push("GraphicsScene<VEC, DATA,  USERINTERFACE, CAMERA, SCENE>::removeUserInterface(const USERINTERFACE *)");
    throw(e);
  }
  catch (typename ObjectsList< DATA >::NotFound e) {
    e.push("GraphicsScene<VEC, DATA,  USERINTERFACE, CAMERA, SCENE>::removeUserInterface(const USERINTERFACE *)");
    throw(e);
  }
  delete d;
}

#if (defined GRAPHICS)

template <class VEC, class DATA,  class USERINTERFACE, class CAMERA, class SCENE>
void GraphicsScene<VEC, DATA,  USERINTERFACE, CAMERA, SCENE>::redraw(void) {
  typename ObjectsList< CAMERA >::iterator it = firstCamera();
  for(; it!=lastCamera(); it++) (*it)->redraw();
}



template <class VEC, class DATA,  class USERINTERFACE, class CAMERA, class SCENE>
void GraphicsScene<VEC, DATA,  USERINTERFACE, CAMERA, SCENE>::focus(void){


  // refocus all cameras
  typename ObjectsList< CAMERA >::iterator it = this->firstCamera();
  for(; it!=this->lastCamera(); it++) (*it)->focus();
  
}

template <class VEC, class DATA,  class USERINTERFACE, class CAMERA, class SCENE>
void GraphicsScene<VEC, DATA,  USERINTERFACE, CAMERA, SCENE>::getCenterRadius(VEC& center, double& radius) {
  double r;
  unsigned int nTotal,nObject;
  typename ObjectsList< DATA >::iterator it = this->firstData();
  VEC objectCenter;

  center = VEC();
  radius = 0.0;
  nTotal = 0;

  for(; it!= this->lastData(); it++) {
    (*it)->updateCenter(objectCenter, nObject); 
    center+= objectCenter*((double)nObject);
    nTotal+=nObject;
  }   
  if(nTotal > 0) {
    center /= (double) nTotal;
    for(it.first(); it!= this->lastData(); it++) {
      r = (*it)->updateRadius(center);
      if(r > radius) radius = r;
    }
  }
}
#endif
} // namespace yav
