#ifndef SCENEACTIVATION_H
#define SCENEACTIVATION_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics/SceneActivation.h"
#include <graphics/Scene.h>
#include <basics/IteratedActiveObjectManager.h>

namespace yav {

#define SAFETY_MARGIN_TO_GET_PROPER_FRAME_RATE 0.8


/// world description
template <class SCENE>
class SceneActivation :  public Activable {
 protected:

  /// value of the iteration step
  unsigned int iterationStep;
  /// the scene which is being activated
  SCENE scene;
  public:
  /** @name constructor */
  //@{
  /// initializes scene with nothing inside
  SceneActivation() :  iterationStep(1) {
  }
  //@}

  /** @name destructor */
  //@{
  /// destroy each object in list
    virtual ~SceneActivation(){}
  //@}
 /** iterates over all active objects and iterate each deformable
      model if it is activated.  The number of iteration each time the routine 
      is called is  nbIterationsPerCall. If there are no deformable model to 
      iterate on then call off() which stops the activation of the scene */
  virtual void activation(void);

  /** iterates over all deformable models and iterate each deformable
      model if it is activated. The total number of iteration is 
      nbIterationsPerCall*iterationStep */
  virtual void stepIteration(void);

  /** Makes the scene activated which allows to iterate over deformable models */
  virtual void on(void);

  /** Stops the iteration over deformable models */
  virtual void off(void);
 
  
  /// returns the number iteration step
  unsigned int getIterationStep() const {
	  return iterationStep; }
   /// set the number of iteration step
  void setIterationStep(const unsigned int n) {
	iterationStep=n;
  }
 };

} // namespace yav

#include "SceneActivation.txx"

#endif // ifndef SCENE_H


