#ifndef TCLSCENE_H
#define TCLSCENE_H

#pragma interface "graphics/TclScene.h"
#include <modules/TclObject.h>
#include <framework/Interface.h>
#include <graphics/SceneItem.h>
#include <graphics/TclSceneActivation.h>
#include <basics/Constants.h>

namespace yav {


  template <class S, class TCLS,class DATA, class CAM, class TCLCAM> class TclScene
 : public TclObject, public Interface {
protected:
  /// an object that activates all deformable models in the scene
  TclSceneActivation<DATA,CAM,TCLCAM,S,TCLS> *sceneActivation;
 public:
  /** @name Exception */
  //@{
  /// using a 0 pointer on a Data object
  class EmptyGraphicsObject : public Exception {
  public:
    EmptyGraphicsObject(const std::string& from) : 
      Exception(from,
		"Using a null pointer on a Data object") {}
  };
  //@}
  /// this scene objects construction type
  typedef enum {
    /// object can be constructed standalone (new OBJ)
    NONE = 1,
    /// object can be constructed from scene (new OBJ -scene scn...)
    SCENE = 2,
    /// object can be constructed from camera (new OBJ -camera cam...)
    CAMERA = 4,
    /// object can be constructed from a file (new OBJ filename)
    FILE = 8,
    /// object can be constructed from a Tcl script
    SCRIPT = 16,
    /// object has no regular construction mode
    OTHER = 32
  } ObjectMode;

  /** constructor
      @param type dynamic object type
      @param scene
      @param m TCL objects'module */
  TclScene(const std::type_info& type, Client *scene, TclModule *m)
    : TclObject(type, m), Interface(scene)  {
    sceneActivation=new TclSceneActivation<DATA,CAM,TCLCAM,S,TCLS> (getClient());
  }

  
    /** removes a range data from this scene
      @param rd data to remove
      @exception ObjectsList<RANGE>::NotFound rd does not exist in scene */
  virtual void removeData(DATA *rd) /*throw (typename ObjectsList< RANGE >::NotFound)*/;

 

  /// redraw all scene cameras : call the redraw function of the tcl camera and not the camera
  void redraw(void);

  /** selects a new object in all cameras
      @param d selected object
      @exception EmptyGraphicsObject d cannot be 0 */
  virtual void selectCameraObject(DATA *d) const;
  /** unselects an object from all cameras
      @param obj selected object
      @exception EmptyGraphicsObject d cannot be 0 */
  virtual void unselectCameraObject(DATA *obj) const;
  /// returns a pointer on a data structure describing the way a scene is activated
    TclSceneActivation<DATA,CAM,TCLCAM,S,TCLS> *getSceneActivation() const {
    return sceneActivation;
  }

  /// -which option
  static int whichCmd(TclModule *m, int argc, char **argv);

  /// -computation TCL command option
  int computationCmd(int argc, char **argv);
  /// -maxFrameRate TCL command option
  int maxFrameRateCmd(int argc, char **argv);
  /// -iteration TCL command option
  int iterationCmd(int argc, char **argv);
  /// -objects TCL command option
  int objectsCmd(int argc, char **argv);
  /// -deformable TCL command option
  int deformableCmd(int argc, char **argv);
  

  /// returns client scene
  S *getClient(void) const { return (S *) client; }
};


}



#include "TclScene.txx"

#endif
