namespace yav {

  template <class S, class TCLS, class DATA,  class CAM, class TCLCAM >
int TclScene<S,TCLS, DATA,CAM,TCLCAM>::whichCmd(TclModule *m, int argc,
					    char **argv) {
  if(argc < 2) return COMMAND_SYNOPSIS;

  TclString res;
  if(!strcmp(argv[1], "-which")) {
    int filter = 0;
    if(argc > 2) {
      for(int i = 2; i < argc; i++) {
	if(!strcmp(argv[i], "none")) filter |= NONE;
	else if(!strcmp(argv[i], "scene")) filter |= SCENE;
	else if(!strcmp(argv[i], "camera")) filter |= CAMERA;
	else if(!strcmp(argv[i], "file")) filter |= FILE;
	else if(!strcmp(argv[i], "script")) filter |= SCRIPT;
	else if(!strcmp(argv[i], "other")) filter |= OTHER;
	else return COMMAND_SYNOPSIS;
      }
    }
    else filter = NONE | SCENE | CAMERA | FILE | SCRIPT | OTHER;
    
    TclObjectKind *any;
    for(any = AutoLink<TclObjectKind>::getHead(); any;
	any = any->following()) {
      if(typeid(*any) == typeid(SceneItem<S>)) {
	SceneItem<S> *item = (SceneItem<S> *)(any);
	if(item->getMode() & filter) res += item->getName() + SPACE_STRING;
      }
    }
  }
  else if(!strcmp(argv[1], "-realName")) {
    if(argc != 3) return COMMAND_SYNOPSIS;
    res="none";
    TclObjectKind *any;
    for(any = AutoLink<TclObjectKind>::getHead(); any;
	any = any->following()) {
      if (!strcmp((any)->getName(),argv[2])) {
	if ((any)->getRealName())
	  res=(any)->getRealName();
      }
    }
  }
  else if(!strcmp(argv[1], "-suffix")) {
    if(argc != 3) return COMMAND_SYNOPSIS;
    TclObjectKind *any;
    for(any = AutoLink<TclObjectKind>::getHead(); any;
	any = any->following()) {
      if(typeid(*any) == typeid(SceneItem<S>)) {
	SceneItem<S> *item = (SceneItem<S> *)(any);
	if((item->getMode() & FILE) && !strcmp(item->getName(), argv[2])) {
	  std::list<const char *>::const_iterator it;
	  for(it = item->getExtensions().begin();
	      it != item->getExtensions().end(); it++)
	    res += (*it) + SPACE_STRING;
	}
      }
    }
    
  }
  else if(!strcmp(argv[1], "-script")) {
    if(argc != 3) return COMMAND_SYNOPSIS;
    TclObjectKind *any;
    for(any = AutoLink<TclObjectKind>::getHead(); any;
	any = any->following()) {
      if(typeid(*any) == typeid(SceneItem<S>)) {
	SceneItem<S> *item = (SceneItem<S> *)(any);
	if((item->getMode() & SCRIPT) && !strcmp(item->getName(), argv[2])) {
	  res += (item)->getCreationScript();
	}
      }
    }
  }

  else return COMMAND_SYNOPSIS;

  m->setResult(res);
  return TCL_OK;
}
template <class S, class TCLS, class DATA,  class CAM, class TCLCAM >
int TclScene<S,TCLS, DATA,CAM,TCLCAM>::maxFrameRateCmd(int argc, char **argv) {
  if (argc==0) {
    setResult("%f",sceneActivation->getMaximumFrameRate());
    return TCL_OK;
  } else if (argc==1) {
    double f=(double) atof(argv[0]);
    sceneActivation->setMaximumFrameRate(f);
    setResult("%f",sceneActivation->getMaximumFrameRate());
    return TCL_OK;
  }
  else
    return COMMAND_SYNOPSIS;
}

  template <class S, class TCLS, class DATA,  class CAM, class TCLCAM>
int TclScene<S,TCLS, DATA,CAM,TCLCAM>::computationCmd(int argc, char **argv) {
  if (argc==0) {
    setResult(TclModule::boolString(sceneActivation->active()));
    return TCL_OK;
  }
  else if (argc==1) {
    if (!strcmp(argv[0],"true")) {
      sceneActivation->setNumberOfIterations(maxTypeValue<unsigned int>(0));
      sceneActivation->on();
      setResult("%s%u started computation", getObjectName(), getRef());
      return TCL_OK;
    }
    else if (!strcmp(argv[0],"false")) {
      sceneActivation->setNumberOfIterations(1);
      sceneActivation->off();
      TclString output;
      output.sprintf("%s%u stopped computation\n", getObjectName(), getRef());
      ObjectsListSTLiterator<IteratedActiveObject> 
	it=sceneActivation->firstActiveObject();
      for(;it!=sceneActivation->lastActiveObject();it++) {
	if ((*it)->isActive() == true) {
	  DATA *data = dynamic_cast<DATA *>((*it));
	  if (data) {
	    Interface *i = data->getInterface();
	    if(i) {
	      TclString time;
	      time.sprintf("%s deformed over %u iterations during %.3f second(s)\n",
			   (char*)dynamic_cast<TclObject *>(i)->name(),
			   (*it)->getNbIterations(),
			   (float) ((*it)->getComputationTime()*1e-6));
	      output += time;
	    }
	  }
	}
      }
      getModule()->setResult(output);
      return TCL_OK;
    }
    else if (!strcmp(argv[0],"loop")) {
      unsigned int i=0;
      sceneActivation->on();
      do {
	i++;
	sceneActivation->activation();
	while(Tk_DoOneEvent(TK_ALL_EVENTS | TK_DONT_WAIT));
      } while (sceneActivation->active());
      setResult("%s%u performed %u iterations", getObjectName(), getRef(), i);
      return TCL_OK;
    }
    else {
      unsigned int u;
      if(sscanf(argv[0], "%u", &u) != 1) return COMMAND_SYNOPSIS;
      sceneActivation->on();
      for(unsigned int i = 0; i < u; i++) {
	sceneActivation->activation();
	while(Tk_DoOneEvent(TK_ALL_EVENTS | TK_DONT_WAIT));
      }
      setResult("%s%u started %u iterations", getObjectName(), getRef(), u);
      return TCL_OK;
    }
  }
  else
    return COMMAND_SYNOPSIS;
}


template <class S, class TCLS, class DATA,  class CAM, class TCLCAM>
int TclScene<S,TCLS, DATA,CAM,TCLCAM>::iterationCmd(int argc, char **argv) {
  if((argc == 1) || (argc==2)) {
    if (!strcmp(argv[0],"step")) {
      if (argc==2)
	sceneActivation->setIterationStep((unsigned int) atoi(argv[1]));
      setResult("%u",sceneActivation->getIterationStep());
      return TCL_OK;
    }
    else if (!strcmp(argv[0],"call")) {
      if (argc==2)
	sceneActivation->setNbIterationsPerCall((unsigned int) atoi(argv[1]));
      setResult("%u",sceneActivation->getNbIterationsPerCall());
      return TCL_OK;
    }
    else {
      setResult("Error with parameter \"%s\" : Must be \"step\" or \"call\"", argv[0]);
      return TCL_ERROR;
    }
  }
  else if (argc==0) {
    sceneActivation->stepIteration();
    setResult("Iterated %u time(s)",sceneActivation->getIterationStep());
    TclString output;
    output.sprintf("Iterated %u time(s)",sceneActivation->getIterationStep());
    
    ObjectsListSTLiterator<IteratedActiveObject> 
      it=sceneActivation->firstActiveObject();
    for(;it!=sceneActivation->lastActiveObject();it++) {
      if ((*it)->isActive()==true) {
	DATA *data=dynamic_cast<DATA *>((*it));
	if (data) {
	  Interface *i = data->getInterface();
	  if(i) {
	    TclString time;
	    time.sprintf("%s deformed during %.3f second(s)\n",
			 (char*)dynamic_cast<TclObject *>(i)->name(),
			 (float) ((*it)->getComputationTime()*1e-6));
	    output+=time; 
	  }
	}
      }
    }
    getModule()->setResult(output);
      
    return TCL_OK;
  }
  else
    return COMMAND_SYNOPSIS;
}


  template <class S, class TCLS, class DATA,  class CAM, class TCLCAM>
int TclScene<S,TCLS, DATA,CAM,TCLCAM>::objectsCmd(int argc, char **argv) {
  if(argc > 0) return COMMAND_SYNOPSIS;
  TclString output;
  typename ObjectsList<DATA>::iterator it=getClient()->firstData();
  for(; it!=getClient()->lastData(); it++) {
    Interface *i = (*it)->getInterface();
    if(i) {
      output += dynamic_cast<TclObject *>(i)->name();
      output += SPACE_STRING;
    }
  }
  setResult(output);
  return TCL_OK;
}

template <class S, class TCLS, class DATA,  class CAM, class TCLCAM>
int TclScene<S,TCLS, DATA,CAM,TCLCAM>::deformableCmd(int argc, char **argv) {
  if(argc > 0) return COMMAND_SYNOPSIS;
  TclString output;
  Client  *c;
  ObjectsListSTLiterator<IteratedActiveObject> 
      it=sceneActivation->firstActiveObject();
    for(;it!=sceneActivation->lastActiveObject();it++) {
    c= dynamic_cast<Client *>(*it);
    if (c)
      {
	Interface *i = c->getInterface();
	if(i) {
	  output += dynamic_cast<TclObject *>(i)->name();
	  output += SPACE_STRING;
	}
      }
  }
  setResult(output);
  return TCL_OK;
}


template <class S, class TCLS, class DATA,  class CAM, class TCLCAM>
void TclScene<S,TCLS, DATA,CAM,TCLCAM>::removeData(DATA *d)
  {
    // First unselect the corresponding graphical object
    unselectCameraObject(d);
    getClient()->removeData(d);
    getClient()->redraw();
}
template <class S, class TCLS, class DATA,  class CAM, class TCLCAM>
void TclScene<S,TCLS, DATA,CAM,TCLCAM>::redraw(void)
  {
 typename ObjectsList< CAM >::iterator it = getClient()->firstCamera();
 for(; it!=getClient()->lastCamera(); it++) {
   TCLCAM *tclcam=dynamic_cast<TCLCAM *>((*it)->getInterface());
       if (tclcam) 
	 tclcam->redraw();
 }


}




template <class S, class TCLS, class DATA,  class CAM, class TCLCAM>
void TclScene<S,TCLS, DATA,CAM,TCLCAM>::selectCameraObject(DATA *obj) const 
{
  // Test if the object exists
  if (obj==0)
    throw(EmptyGraphicsObject("void TclScene<S,TCLS, DATA,CAM,TCLCAM>::selectCameraObject(DATA *)"));
  typename ObjectsList< CAM >::iterator it = getClient()->firstCamera();
  for(; it!=getClient()->lastCamera(); it++) {
    TCLCAM *tclcam=dynamic_cast<TCLCAM *>((*it)->getInterface());
    if (tclcam) 
      tclcam->applyObjectSelection(obj);
  }
}


template <class S, class TCLS, class DATA,  class CAM, class TCLCAM>
void TclScene<S,TCLS, DATA,CAM,TCLCAM>::unselectCameraObject(DATA *obj) const 
{
  // Test if the object exists
  if (obj==0)
    throw(EmptyGraphicsObject("void TclScene<S,TCLS, DATA,CAM,TCLCAM>::unselectCameraObject(const DATA *)"));
  typename ObjectsList< CAM >::iterator it = getClient()->firstCamera();
  for(; it!=getClient()->lastCamera(); it++) {
    TCLCAM *tclcam=dynamic_cast<TCLCAM *>((*it)->getInterface());
    if (tclcam) 
      tclcam->applyObjectUnselection(obj);
  }
}

}
