#ifndef TCLSCENEACTIVATION_H
#define TCLSCENEACTIVATION_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics/TclSceneActivation.h"
#include <graphics/Scene.h>
#include <framework/IteratedActiveObjectManager.h>
#include <basics/Timer.h>

namespace yav {

#define SAFETY_MARGIN_TO_GET_PROPER_FRAME_RATE 0.8
#define DEFAULT_MAX_FRAME_RATE 30.0

/** An object that handles the active objects (ie deformable objects) that
    are stored within a scene. The activation() routine describes the events
    that occur when the scene is active. It performs mainly 3 tasks : call the iterate
    function of all deformable models, eventually grabs a camera display to make 
    a movie and redraw the scene in cameras */
  template <class DATA,class CAMERA,class TCLCAM, class S, class TCLS>
class TclSceneActivation :  public IteratedActiveObjectManager {
 protected:
  /// camera to extract movie from
  CAMERA *movieCamera;
  /// movie frame number
  unsigned int frameNumber;
  /// movie base file name
  TclString movieName;
  /// timer that measures time between two frames in order to control frame rate
  Timer redrawTimer;
  /// minimum time between two redraws to control maximum frame rate in seconds
  double minTimeBetweenRedraw;
  /// value of the iteration step
  unsigned int iterationStep;
  /// the scene which is being activated
  S *scene;
  public:
  /** @name constructor */
  //@{
  /// initializes scene with nothing inside
  TclSceneActivation(S *s) : 
    movieCamera(0),frameNumber(0),movieName("default"),iterationStep(1),  scene(s){
    setMaximumFrameRate(DEFAULT_MAX_FRAME_RATE);
  }
  //@}

  /** @name destructor */
  //@{
  /// destroy each object in list
    virtual ~TclSceneActivation(){}
  //@}
 /** iterates over all deformable models and iterate each deformable
      model if it is activated.  The number of iteration each time the routine 
      is called is  nbIterationsPerCall. If there are no deformable model to 
      iterate on then call off() which stops the activation of the scene */
  virtual void activation(void);

  /** iterates over all deformable models and iterate each deformable
      model if it is activated. The total number of iteration is 
      nbIterationsPerCall*iterationStep */
  virtual void stepIteration(void);

  /** Makes the scene activated which allows to iterate over deformable models */
  virtual void on(void);

  /** Stops the iteration over deformable models */
  virtual void off(void);


  /// get maximum frame rate (in Hz)
  double getMaximumFrameRate() const {
    return (SAFETY_MARGIN_TO_GET_PROPER_FRAME_RATE/minTimeBetweenRedraw); }
  /// set maximum frame rate (in Hz)
  void setMaximumFrameRate(const double f)  {
    minTimeBetweenRedraw=SAFETY_MARGIN_TO_GET_PROPER_FRAME_RATE/f;
  }
 

  /** start camera recording
      @param cam camera to create movie from
      @param name base file name */
  void startMovie(CAMERA *cam, const char *name){
  frameNumber = 0;
  movieCamera = cam;
  movieName = name;
  }
  /** stop camera recording */
  void stopMovie(void){
    movieCamera=0;
  }
  /// returns the number iteration step
  unsigned int getIterationStep() const {
	  return iterationStep; }
   /// set the number of iteration step
  void setIterationStep(const unsigned int n) {
	iterationStep=n;
  }
 };
#include "TclSceneActivation.txx"
} // namespace yav



#endif // ifndef SCENE_H


