

template <class DATA,    class CAMERA, class TCLCAM, class SCENE, class TCLSCENE>
void TclSceneActivation<DATA, CAMERA, TCLCAM,SCENE,TCLSCENE>::stepIteration(void) 
{
  /// gets the current time
#ifndef WIN32
  struct timeval tv;
#endif
  unsigned LONGINT startIteration;

  for(unsigned int i=0; i<iterationStep;++i)
    {
    ObjectsList< IteratedActiveObject >::iterator it =
			activeObjectList.begin();

		for(; it!=activeObjectList.end(); it++) 
		{
      
	  if ((*it)->isActive() )
	    {
#ifndef WIN32
		  // record the time before deformation
	      gettimeofday(&tv, 0);
	      startIteration=tv.tv_sec * 1000000 + tv.tv_usec;
#else
		  startIteration=(unsigned int)time(0);
#endif
	      (*it)->iterate(nbIterationsPerCall);
	       incrementIteration(nbIterationsPerCall,(*it));
	      // get the time after deformation
#ifndef WIN32
	      gettimeofday(&tv, 0);
	      // updates the duration of deformation
	      if (i==0)
	      	    incrementComputationTime(tv.tv_sec * 1000000 + tv.tv_usec -startIteration-(*it)->getComputationTime(), (*it));
	      else
	      incrementComputationTime(tv.tv_sec * 1000000 + tv.tv_usec -startIteration, (*it));
#else
    if (i==0)
        incrementComputationTime((unsigned LONGINT)time(0)-startIteration-(*it)->getComputationTime(),
        (*it));
    else
    incrementComputationTime((unsigned LONGINT)time(0)-startIteration,(*it));
#endif
	    }
	}
    }
  
  if(movieCamera) {
    TclString filename;
    filename.sprintf("%s%04u.xwd", (char *) movieName,
		     frameNumber++);
   TCLCAM *tclcam=dynamic_cast<TCLCAM *>(movieCamera->Client::getInterface());
       if (tclcam) 
            tclcam->dumpToFile(filename);
  }

  ((TCLSCENE *)(scene->getInterface()))->redraw();

}

template <class DATA,  class CAMERA, class TCLCAM,class SCENE, class TCLSCENE>
void TclSceneActivation< DATA,   CAMERA, TCLCAM,SCENE,TCLSCENE>::activation(void) 
{
#ifndef WIN32
  struct timeval tv;
#endif
  unsigned LONGINT startIteration;

  ObjectsList< IteratedActiveObject >::iterator it =
			activeObjectList.begin();
// tests if there is at list one deformable model to iterate
  bool stopIteration=true;
		for(; it!=activeObjectList.end(); it++) 
		{
      if ((*it)->isActive()==true)
	{
	  incrementIteration(nbIterationsPerCall,(*it));
	  // record the time before deformation
#ifndef WIN32
	  gettimeofday(&tv, 0);
	  startIteration=tv.tv_sec * 1000000 + tv.tv_usec;
#else
	  startIteration=(unsigned int)time(0);
#endif
	  try{
	    (*it)->iterate(nbIterationsPerCall);
	  } catch(Exception e){
	    CERR << "Problem iterating on object" << ENDL;
	    off();
	  }
	  // get the time after deformation
#ifndef WIN32
	  gettimeofday(&tv, 0);
	  // updates the duration of deformation
	  incrementComputationTime(tv.tv_sec * 1000000 + tv.tv_usec -startIteration, (*it));
#else
incrementComputationTime((unsigned LONGINT)time(0)-startIteration,(*it));
#endif
	  stopIteration=false;
	}
    }

  if(movieCamera) {
    TclString filename;
    filename.sprintf("%s%04u.xwd", (char *) movieName,
		     frameNumber++);
   TCLCAM *tclcam=dynamic_cast<TCLCAM *>(movieCamera->Client::getInterface());
       if (tclcam) 
            tclcam->dumpToFile(filename);
  }

  numberOfIterations--;
  if  (redrawTimer.getTime()>minTimeBetweenRedraw)
    {
  ((TCLSCENE *)(scene->getInterface()))->redraw();
      redrawTimer.reset();
    }
  if(stopIteration || (numberOfIterations == 1))
    off();

}

template <class DATA,  class CAMERA, class TCLCAM, class SCENE, class TCLSCENE>
void TclSceneActivation<DATA,  CAMERA, TCLCAM,SCENE,TCLSCENE>::on(void) 
{
  typename ObjectsList< CAMERA >::iterator it =
    scene->firstCamera();
  // tests if there is at list one deformable model to iterate
  for(; it!=scene->lastCamera(); it++) 
    {
      TCLCAM *tclcam=dynamic_cast<TCLCAM *>((*it)->Client::getInterface());
       if (tclcam) 
            tclcam->startComputation();
    }
    IteratedActiveObjectManager::on();
}

template <class DATA,  class CAMERA, class TCLCAM,class SCENE, class TCLSCENE>
void TclSceneActivation<DATA,  CAMERA, TCLCAM,SCENE,TCLSCENE>::off(void)
{
  typename ObjectsList< CAMERA >::iterator it =
    scene->firstCamera();
  // tests if there is at list one deformable model to iterate
  for(; it!=scene->lastCamera(); it++) 
    {
      TCLCAM *tclcam=dynamic_cast<TCLCAM *>((*it)->Client::getInterface());
       if (tclcam) 
            tclcam->stopComputation();
    }
  Activable::off();
}


