#ifndef TCLSCENEOBJECT_H
#define TCLSCENEOBJECT_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics/TclSceneObject.h"
#include <basics/ListElement.h>
#include <modules/TclObject.h>
#include <framework/Interface.h>


namespace yav {

/** TCL Object offering a TCL interface to a client */
class TclSceneObject
  : public TclObject,
    public ListElement<TclSceneObject>,
    public Interface {

 private:
  /// forbids construction copy
  TclSceneObject(const TclSceneObject& so) :
    TclObject(so), Interface(0) {}

 protected:
  /// default string for installation of menus of selected objects
  TclString installMenuCommandName;
  /// default string for uninstallation of menus of selected objects
  TclString uninstallMenuCommandName;
  /// default string for installation of popup menus of selected objects
  TclString installPopupMenuCommandName;

 public:
  /** @name constructor */
  //@{
  /** only constructor from a client object
      @param type object type
      @param client non null pointer on client object
      @param m TCL module to build object in
      @exception NullClient null pointer transmitted */
  TclSceneObject(const std::type_info& type, Client *client, TclModule *m = 0)
    throw(Interface::NullClient)
    : TclObject(type, m), Interface(client) {
    client->setInterface(this);
#if (defined GRAPHICS)
    installMenuCommandName = EMPTY_STRING;
    installPopupMenuCommandName = EMPTY_STRING;
    uninstallMenuCommandName = EMPTY_STRING;
#endif
  }
  //@}

  /** @name destructor */
  //@{
  /// destroy client and this
  virtual ~TclSceneObject() { 
  }
  //@}


  /// function called when this object is selected
  virtual void selected(void) {}
  /// function called when this object is unselected
  virtual void unselected(void) {}
  
  /// returns client
  Client *getData(void) const { return client; }
  /// returns client
  Client *getClient(void) const { return client; }

  /** gets the Tcl command name for installing the menu of
      selected/unselected objects */
  TclString getInstallMenuCommandName() const {
    return(installMenuCommandName); 
  }
  /** gets the Tcl command name for installing the popup menu of
      selected/unselected objects */
  TclString getInstallPopupMenuCommandName() const {
    return(installPopupMenuCommandName); 
  }
  /** gets the Tcl command name for uninstalling the menu of
      selected/unselected objects */
  TclString getUninstallMenuCommandName() const {
    return(uninstallMenuCommandName); 
  }
  /** sets Tcl command name for installing the menu of
      selected/unselected objects */
  void setInstallMenuCommandName(const TclString name) {
    installMenuCommandName = name; 
  }
  /** sets Tcl command name for installing the popup menu of
      selected/unselected objects */
  void setInstallPopupMenuCommandName(const TclString name) {
    installPopupMenuCommandName = name; 
  }
  /** sets Tcl command name for uninstalling the menu of
      selected/unselected objects */
  void setUninstallMenuCommandName(const TclString name) {
    uninstallMenuCommandName = name;  
  }

};


} // namespace yav

#endif // ifndef TCLSCENEOBJECT_H
