#ifndef USERINTERFACE_H
#define USERINTERFACE_H

#pragma interface "graphics/UserInterface.h"

namespace yav {


/** \class UserInterface
    \ingroup Graphics
    A Generic User Interface class that provides a position 
    and that can output a force (if it is a force feedback device) */
template <class SCENE, class VEC, class INTERFACE > class UserInterface :  
public ListElement<INTERFACE > {
 public:
  /** @name constructor */
  //@{
  /** builds a new UserInterface
      @param scene  scene where the user interface object should be added (if scene is 0, the first existing scene is used or a scene is created== 0) */
  UserInterface(SCENE *scene);
  //@}

  /** @name destructor */
  //@{
  /// empty
  virtual ~UserInterface(){}
  //@}
  
  /// returns the current position of the device
  virtual VEC getPosition()=0;
  /** send a force to the device
      @param force the force to be applied on the device */
  virtual void setForce(const VEC &force)=0;

};

} // namespace yav

#include "UserInterface.txx"

#endif

