#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#include <graphics2D/BasicCameraBackground.h>

namespace yav {

// static TclString BasicCameraBackground::sliceName="2D Image";
	
typedef struct _ImageRec {
    unsigned short imagic;
    unsigned short type;
    unsigned short dim;
    unsigned short xsize, ysize, zsize;
    unsigned int min, max;
    unsigned int wasteBytes;
    char name[80];
    unsigned long colorMap;
    FILE *file;
    unsigned char *tmp, *tmpR, *tmpG, *tmpB;
    unsigned long rleEnd;
    unsigned int *rowStart;
    int *rowSize;
} ImageRec;


static void bwtorgba(unsigned char *b,unsigned char *l,int n) {
  while(n--) {
    l[0] = *b;
    l[1] = *b;
    l[2] = *b;
    l[3] = 0xff;
    l += 4; b++;
  }
}

static void latorgba(unsigned char *b, unsigned char *a,unsigned char *l,int n) {
  while(n--) {
    l[0] = *b;
    l[1] = *b;
    l[2] = *b;
    l[3] = *a;
    l += 4; b++; a++;
  }
}

static void rgbtorgba(unsigned char *r,unsigned char *g,unsigned char *b,unsigned char *l,int n) {
  while(n--) {
    l[0] = r[0];
    l[1] = g[0];
    l[2] = b[0];
    l[3] = 0xff;
    l += 4; r++; g++; b++;
  }
}

static void rgbatorgba(unsigned char *r,unsigned char *g,unsigned char *b,unsigned char *a,unsigned char *l,int n) {
  while(n--) {
    l[0] = r[0];
    l[1] = g[0];
    l[2] = b[0];
    l[3] = a[0];
    l += 4; r++; g++; b++; a++;
  }
}

static void ConvertShort(unsigned short *array, long length) {
  unsigned b1, b2;
  unsigned char *ptr;
  
  ptr = (unsigned char *)array;
  while (length--) {
    b1 = *ptr++;
    b2 = *ptr++;
    *array++ = (b1 << 8) | (b2);
  }
}

static void ConvertLong(unsigned *array, long length) {
  unsigned b1, b2, b3, b4;
  unsigned char *ptr;
  
  ptr = (unsigned char *)array;
  while (length--) {
    b1 = *ptr++;
    b2 = *ptr++;
    b3 = *ptr++;
    b4 = *ptr++;
    *array++ = (b1 << 24) | (b2 << 16) | (b3 << 8) | (b4);
  }
}

static ImageRec *ImageOpen(const char *fileName)
{
  union {
    int testWord;
    char testByte[4];
  } endianTest;
  ImageRec *image;
  int swapFlag;
  int x;
  
  endianTest.testWord = 1;
  if (endianTest.testByte[0] == 1) {
    swapFlag = 1;
  } else {
    swapFlag = 0;
  }
  
  image = (ImageRec *) new ImageRec;;
  if (image == NULL) {
    fprintf(stderr, "Out of memory!\n");
    exit(1);
  }
  if ((image->file = fopen(fileName, "rb")) == NULL) {
    perror(fileName);
    exit(1);
  }
  
  fread(image, 1, 12, image->file);
  
  if (swapFlag) {
    ConvertShort(&image->imagic, 6);
  }
  
  image->tmp = (unsigned char *) new unsigned char[image->xsize*256];
  image->tmpR = (unsigned char *) new unsigned char[image->xsize*256];
  image->tmpG = (unsigned char *) new unsigned char[image->xsize*256];
  image->tmpB = (unsigned char *) new unsigned char[image->xsize*256];
  if (image->tmp == NULL || image->tmpR == NULL || image->tmpG == NULL ||
      image->tmpB == NULL) {
    fprintf(stderr, "Out of memory!\n");
    exit(1);
    }
  
  if ((image->type & 0xFF00) == 0x0100) {
    x = image->ysize * image->zsize * sizeof(unsigned);
    image->rowStart = (unsigned *)new unsigned char[x];
    image->rowSize = (int *) new unsigned char[x];
    if (image->rowStart == NULL || image->rowSize == NULL) {
      fprintf(stderr, "Out of memory!\n");
      exit(1);
    }
    image->rleEnd = 512 + (2 * x);
    fseek(image->file, 512, SEEK_SET);
    fread(image->rowStart, 1, x, image->file);
    fread(image->rowSize, 1, x, image->file);
    if (swapFlag) {
      ConvertLong(image->rowStart, x/(int)sizeof(unsigned));
      ConvertLong((unsigned *)image->rowSize, x/(int)sizeof(int));
    }
  } else {
    image->rowStart = NULL;
    image->rowSize = NULL;
  }
  return image;
}

static void ImageClose(ImageRec *image) {
  fclose(image->file);
  delete [] image->tmp;
  delete [] image->tmpR;
  delete [] image->tmpG;
  delete [] image->tmpB;
  delete [] image->rowSize;
  delete [] image->rowStart;
  delete image;
}

static void ImageGetRow(ImageRec *image, unsigned char *buf, int y, int z) {
  unsigned char *iPtr, *oPtr, pixel;
  int count;
  
  if ((image->type & 0xFF00) == 0x0100) {
    fseek(image->file, (long)image->rowStart[y+z*image->ysize], SEEK_SET);
    fread(image->tmp, 1, (unsigned int)image->rowSize[y+z*image->ysize],
	  image->file);
    
    iPtr = image->tmp;
    oPtr = buf;
    for (;;) {
      pixel = *iPtr++;
      count = (int)(pixel & 0x7F);
      if (!count) {
	return;
      }
      if (pixel & 0x80) {
	while (count--) {
	  *oPtr++ = *iPtr++;
	}
      } else {
	pixel = *iPtr++;
	while (count--) {
	  *oPtr++ = pixel;
	}
      }
    }
  } else {
    fseek(image->file, 512+(y*image->xsize)+(z*image->xsize*image->ysize),
	  SEEK_SET);
    fread(buf, 1, image->xsize, image->file);
  }
}

static unsigned *read_texture(const char *name, unsigned int *width, unsigned int *height, unsigned int *components) {
  unsigned *base, *lptr;
  unsigned char *rbuf, *gbuf, *bbuf, *abuf;
  ImageRec *image;
  int y;
  
  image = ImageOpen(name);
  
  if(!image)
    return NULL;
  (*width)=image->xsize;
  (*height)=image->ysize;
  (*components)=image->zsize;
  base = (unsigned *) new unsigned int [image->xsize*image->ysize];
  rbuf = (unsigned char *) new unsigned char[image->xsize*sizeof(unsigned char)];
  gbuf = (unsigned char *) new unsigned char[image->xsize*sizeof(unsigned char)];
  bbuf = (unsigned char *) new unsigned char[image->xsize*sizeof(unsigned char)];
  abuf = (unsigned char *) new unsigned char[image->xsize*sizeof(unsigned char)];
  if(!base || !rbuf || !gbuf || !bbuf)
    return NULL;
  lptr = base;
  for(y=0; y<image->ysize; y++) {
    if(image->zsize>=4) {
      ImageGetRow(image,rbuf,y,0);
      ImageGetRow(image,gbuf,y,1);
      ImageGetRow(image,bbuf,y,2);
      ImageGetRow(image,abuf,y,3);
      rgbatorgba(rbuf,gbuf,bbuf,abuf,(unsigned char *)lptr,image->xsize);
      lptr += image->xsize;
    } else if(image->zsize==3) {
      ImageGetRow(image,rbuf,y,0);
      ImageGetRow(image,gbuf,y,1);
      ImageGetRow(image,bbuf,y,2);
	    rgbtorgba(rbuf,gbuf,bbuf,(unsigned char *)lptr,image->xsize);
	    lptr += image->xsize;
	} else if(image->zsize==2) {
	  ImageGetRow(image,rbuf,y,0);
	  ImageGetRow(image,abuf,y,1);
	  latorgba(rbuf,abuf,(unsigned char *)lptr,image->xsize);
	  lptr += image->xsize;
	} else {
	  ImageGetRow(image,rbuf,y,0);
	  bwtorgba(rbuf,(unsigned char *)lptr,image->xsize);
	  lptr += image->xsize;
	}
  }
  ImageClose(image);
  delete [] rbuf;
  delete [] gbuf;
  delete [] bbuf;
  delete [] abuf;
  
  return (unsigned *) base;
}

BasicCameraBackground::BasicCameraBackground(Camera2D *cam,const char *name ) : Camera2DBackground(cam),
sliceName("2D Image"),binded(false){
  unsigned int nr,nc;
  unsigned int components;
  unsigned int *b=read_texture(name,&nc,&nr,&components); 
  format=TEXTURE_RGBA;
  enlargeTexture(b,nr,nc);
  delete [] b;
  init();
}


BasicCameraBackground::BasicCameraBackground(Camera2D *cam,unsigned int *b, const unsigned int nr, 
		     const unsigned int nc,   BasicCameraBackground::Format f) : Camera2DBackground(cam),
			 sliceName("2D Image"),binded(false), format(f) {
  enlargeTexture(b,nr,nc);
  init();
}

void BasicCameraBackground::updateTexture() const {
  if (format==TEXTURE_RGBA)
    glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
  else if (format==TEXTURE_LUMINANCE)
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  
  if (useMipmapping)
    {
      if (format==TEXTURE_RGBA)
	gluBuild2DMipmaps(GL_TEXTURE_2D, mipmapLevel, ncols, nrows, GL_RGBA,
			  GL_RGBA, buffer);
      else if (format==TEXTURE_LUMINANCE)
	gluBuild2DMipmaps(GL_TEXTURE_2D, mipmapLevel, ncols, nrows, GL_LUMINANCE,
			  GL_LUMINANCE, buffer);
    }
  else
    {
      if (format==TEXTURE_RGBA)
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, ncols,nrows, 0,
		     GL_RGBA, GL_UNSIGNED_BYTE, buffer);
      else if (format==TEXTURE_LUMINANCE)
	glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE, ncols,nrows, 0,
		     GL_LUMINANCE, GL_UNSIGNED_BYTE, buffer);
	
    }
  glBindTexture(GL_TEXTURE_2D, 0);
}
void BasicCameraBackground::createTexture() {
  glGenTextures(1,&id);
  binded=true;
  glBindTexture(GL_TEXTURE_2D, id);
  updateTexture();
}
void BasicCameraBackground::init() {
  useMipmapping=false;
  mipmapLevel=2;
  updateDisplayList=true;
}
void BasicCameraBackground::redraw(){
	   glDisable(GL_LIGHTING);
	enableTexture();
	glColor3f(1,1,1);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	  double s=getNbRealRows()/(double) getNbRows();
      double t=getNbRealColumns()/(double) getNbColumns();


	  glBegin(GL_QUADS);
 glTexCoord2d(0,0);
  glVertex2d(-0.5, -0.5); 
  glTexCoord2d(t,0);
  glVertex2d(getNbRealColumns()+.5,-0.5);
  glTexCoord2d(t,s);
  glVertex2d(getNbRealColumns()+.5,getNbRealRows()+.5);
  glTexCoord2d(0,s);
  glVertex2d(-0.5,getNbRealRows()+.5); 
  
  glEnd();
  
disableTexture();


}
  
void BasicCameraBackground::getViewportSize(Vec2<double> &vmin,Vec2<double> &vmax){
vmin=Vec2<double>(-0.5,-0.5);
vmax=Vec2<double>(getNbRealColumns()+.5,getNbRealRows()+.5);
}

void BasicCameraBackground::getImagePlane(Vec3<double> &org, Vec3<double> &normal){
	org=Vec3<double>();
	normal=Vec3<double>(0,0,1);
}

bool BasicCameraBackground::map3DWorldTo2DWorldCoordinates(const Vec3<double> &p, Vec2<double> &pt){
	if (p.getZ()!=0.) 
		return false;
	else {
		pt=Vec2<double>(p.getX(),p.getY());
		return true;
	}

}
  
void BasicCameraBackground::map2DWorldToImageCoordinates(const Vec2<double> &pt, Vec4<double> &pixel){
	pixel=Vec4<double>(pt.getX(),pt.getY(),0,0);
}
void BasicCameraBackground::map2DWorldTo3DWorldCoordinates(const Vec2<double> &pt, Vec3<double> &p){
p=Vec3<double>(pt.getX(),pt.getY(),0);
}

const char *BasicCameraBackground::getSliceName() const{
	return sliceName;
}
int BasicCameraBackground::getSliceNumber() const{
return 1;}

const char *BasicCameraBackground::getLabel(const Vec2<double> &pt){
	return (0);
}

void BasicCameraBackground::enlargeTexture(const void *bb, const unsigned int nr, 
			       const unsigned int nc) {
  unsigned int i;
  for(i=1;i<nr;i*=2);
  nrows=i;
  nbRealRows=nr;
  for(i=1;i<nc;i*=2);
  ncols=i;
  nbRealColumns=nc;

  if (format==TEXTURE_RGBA)
    {
      const unsigned int *b=( unsigned int *) bb;
      buffer=new unsigned int[nrows*ncols];
      /// put 0 in the image
      memset((void *) buffer,0,nrows*ncols*sizeof(unsigned int));
      // copy buffer
      const unsigned int *p;
      unsigned int *q;
      for(p=b,q=(unsigned int *)buffer,i=0;i<nr;++i)
	{
	  memcpy((void *)q,(void *) p,nc*sizeof(unsigned int));
	  q+=ncols;
	  p+=nc;
	}
    }
  else if (format==TEXTURE_LUMINANCE) {
    const unsigned char *b=( unsigned char *) bb;
    buffer=new unsigned char[nrows*ncols];
    /// put 0 in the image
    memset((void *) buffer,0,nrows*ncols*sizeof(unsigned char));
    // copy buffer
    const unsigned char *p;
    unsigned char *q;
    for(p=b,q=(unsigned char *)buffer,i=0;i<nr;++i)
      {
	memcpy((void *)q,(void *) p,nc*sizeof(unsigned char));
	q+=ncols;
	p+=nc;
      }
  }
}


}
