#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "graphics2D/Camera2D.h"
#pragma implementation "graphics/Camera.h"
#include <graphics2D/Camera2D.h>
#include <graphics2D/Camera2DActivation.h>


namespace yav {

#define ALPHA_VAL 1.0



Camera2DBackground::~Camera2DBackground() {
		camera->background=0;
		if (camera->beingDestroyed==false)
			camera->doFocus();
	}
	Camera2DBackground::Camera2DBackground(Camera2D *_camera) {
		camera=_camera;
		camera->background=this;
		camera->doFocus();
	}
	// static members initialization
template <> CameraTypes::BUTTON_STATE Camera<class Vec2<double>,Rotation2D,Scene2D,Data2D,Camera2D>::B[3] = {
		CameraTypes::BUTTON_RELEASED,
			CameraTypes::BUTTON_RELEASED,
			CameraTypes::BUTTON_RELEASED
	};

template <> 	Vec2<double> Camera<Vec2<double>,Rotation2D,Scene2D,Data2D,Camera2D>::pressPoint=Vec2<double>();
template <> 	Vec2<double> Camera<Vec2<double>,Rotation2D,Scene2D,Data2D,Camera2D>::lastPoint=Vec2<double>();
template <> 	LONGINT Camera<Vec2<double>,Rotation2D,Scene2D,Data2D,Camera2D>::clickTime=0;
template <> 	LONGINT Camera<Vec2<double>,Rotation2D,Scene2D,Data2D,Camera2D>::renderDuration=0;
template <> 	LONGINT Camera<Vec2<double>,Rotation2D,Scene2D,Data2D,Camera2D>::clickDuration=0;

	Camera2D::Camera2D(Camera2D *cam) :  ListElement<Camera2D>(),
		Camera<Vec2<double>,Rotation2D,Scene2D,Data2D,Camera2D>() 
	{
		initCamera();
	}

	Camera2D::Camera2D(Scene2D *scn, unsigned int w, unsigned int h, double scale) 
		:  ListElement<Camera2D>(), Camera<Vec2<double>,Rotation2D,Scene2D,Data2D,Camera2D>(scn, w, h, scale)
	{
		initCamera();
	}

	void Camera2D::initCamera(void) {
		selectionFunction = NULL;
		center = DEFAULT_CAMERA2D_CENTER;
		background=0; 
		// A priori no object is selected
		selectedObject=0;
		// Initialize the camera Center
		center.setValue(width/2.0,height/2.0);
		refCenter=center;
		cameraActivation=0;
		beingDestroyed=false;
		/// whether to use stencil Buffer for display
		stencilBuffer=false;
		/// the display list associated with a stencil
		stencilBufferDisplayList=0;

		globalTimer.reset();
		count = 0;
		showCopyright=true;
		showFrameRate=false;
		sprintf(stringFrameRate,"Frame Rate   :");
		sprintf(stringCopyright,"%cINRIA - Asclepios project",169);
		/// the field depth ratio multiplied by the radius
		scale=1.0;  
		/* Projection matrix */
		matTransform.resetToIdentity();

		/* If a null pointer was passed */
		if(!scene) {
			/* First check that there exists allready one graphics scene*/
			if(AutoLink<Scene2D>::getCount() > 0)  
				scene = AutoLink<Scene2D>::getHead(); 
			else 
				/*Otherwise  creates a new scene */
				scene = new Scene2D();
		}
		scene->addCamera(this);
	}


	Camera2D::~Camera2D() 
	{
		beingDestroyed=true;
		if (background) 
			delete background;
		if (cameraActivation)
			delete cameraActivation;
	}

	/// return the object dealing with the camera activation
	Camera2DActivation *Camera2D::getCameraActivation() const{
		return cameraActivation;
	}
	/// set the object dealing with the camera activation
	void Camera2D::setCameraActivation(Camera2DActivation *activation) {
		cameraActivation=activation;
	}

	void Camera2D::toggleFrameRate() {
		if (showFrameRate)
			showFrameRate=false;
		else
			showFrameRate=true;
	}

	void Camera2D::toggleCopyright() {
		if (showCopyright)
			showCopyright=false;
		else
			showCopyright=true;
	}

	bool Camera2D::isCameraBeingDestroyed() const {
		return beingDestroyed;
	}


	void Camera2D::destroyCamera() {
		beingDestroyed=true;

		if(stencilBufferDisplayList) {
			glDeleteLists(stencilBufferDisplayList,1);
		}
	}

	/* --------------------------------------------------------*/
	Rotation2D Camera2D::getTrackballRotation(const Vec2<double>& b1,
		const Vec2<double>& b2) 
	{
		double angle;
		Vec2<double> dp1, dp2;

		dp1=b1-refCenter;
		dp2=b2-refCenter;
		double norm1=dp1.norm();
		double norm2=dp1.norm();
		if ((norm1==0) || (norm2==0))
			angle=0;
		else
		{
			dp1/=norm1;
			dp2/=norm2;
			angle=atan2(det(dp1,dp2),dot(dp1,dp2));
		}
		Rotation2D rot(-angle);
		return(rot);
	}

	/* --------------------------------------------------------*/
	double Camera2D::getTrackballScale(const Vec2<double>& b1,
		const Vec2<double>& b2)
	{
		double theScale= (b2.getY()-b1.getY()) / height;
		if (theScale > 0.0)
			return(1.0 + theScale);
		else
			return(1.0 / (1.0 - theScale));
	}

	/* --------------------------------------------------------*/
	Vec2<double>  Camera2D::getTrackballTranslation(const Vec2<double>& b1,
		const Vec2<double>& b2)
	{

		return(Vec2<double>((b1.getX()-b2.getX())/scale,(b2.getY()-b1.getY())/scale));  // (radius * scale)));
	}


	void Camera2D::rotateCameraObject(Rotation2D &rot) {
		if ((selectedObject) && (rotationFunction))
			(selectedObject->getData()->*rotationFunction)(rot);

	}


	/*-----------------------------------*/ 
	void Camera2D::initViewport(void) {
	  //GLfloat light_model_ambient[] = {0.1f, 0.1f, 0.1f, 1.0f};

		//  glDrawBuffer(GL_FRONT_AND_BACK); //jcl
		glClearColor(0, 0, 0, ALPHA_VAL); 
		glViewport(margin, margin, width-2*margin,
			height-2*margin);

		glEnable(GL_DEPTH_TEST);  
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); 

		// transparent drawing
		glDepthFunc((GLenum) GL_ALWAYS); 

		// enable blending (remove font background)

		glEnable(GL_BLEND);
		glBlendFunc(GL_ONE, GL_ZERO); 

		//   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		//   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
		//   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		//   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		//   glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);

		///  glEnable(GL_NORMALIZE); 
		//  glCullFace(GL_BACK);
		// glBlendFunc(GL_ONE, GL_ONE);

		setViewingParameters();
		glFlush();
	}

	void Camera2D::translateCameraObject(Vec2<double> &t) {
		if ((selectedObject) && (translationFunction))
			(selectedObject->getData()->*translationFunction)(t);


	}
	void Camera2D::buttonReleased(WHICH_BUTTON w,  int x, int y) {
#ifndef WIN32
		struct timeval tv;
#endif
		LONGINT releaseTime;
		Vec2<double> p = getTrackballVector(x, y);
#ifndef WIN32
		gettimeofday(&tv, 0);
		releaseTime = tv.tv_sec * 1000000 + tv.tv_usec;
#else
		releaseTime=time(0);
#endif
		clickDuration = releaseTime - clickTime;

		switch(w) {

			// button 1
  case LEFT_BUTTON:
	  switch(B[LEFT_BUTTON]) 
	  {

		  // selection
	  case BUTTON_PRESSED:
		  if (selectedObject) {
			  Vec2<double> mouse=Vec2<double>(x, y);
			  select(mouse, getType());
			  selectedObject->mouseButtonReleased(w,NO_MODIFIER,this,x,y);
		  } else if (background) {
			  background->mouseButtonReleased(w,NO_MODIFIER,this,x,y);
		  }
		  break;
		  // world scale
	  case CTRL_BUTTON_PRESSED:
		  moveMode = ROTATION_MODE;
		  if (selectedObject)
			  selectedObject->mouseButtonReleased(w,CTRL_KEY,this,x,y);
		  else if (background) {
			  background->mouseButtonReleased(w,CTRL_KEY,this,x,y);
		  }
		  redraw();
		  break;
  case SHIFT_BUTTON_PRESSED:
	  // send event to the selected object
	  if (selectedObject)
		  selectedObject->mouseButtonReleased(w,SHIFT_KEY,this,x,y);
	  else if (background) {
		  background->mouseButtonReleased(w,SHIFT_KEY,this,x,y);
	  }
	  break;
  default:

	  break;
	  }
	  break;

	  // button 2
  case MIDDLE_BUTTON:


	  switch(B[MIDDLE_BUTTON]) 
	  {
	  case BUTTON_PRESSED:
	  default:
		  // send event to the selected object
		  if (selectedObject)
			  selectedObject->mouseButtonReleased(w,SHIFT_KEY,this,x,y);
		  else if (background) {
			  background->mouseButtonReleased(w,SHIFT_KEY,this,x,y);
		  }
		  break;
	  }

	  moveMode = ROTATION_MODE;
	  redraw();
	  break;

  case RIGHT_BUTTON:
  default:
	  break;
		}

		B[w] = BUTTON_RELEASED;
		// bind the mouse motion to call mouseMoved()

	}


	void Camera2D::buttonPressed(WHICH_BUTTON w,
		KEY_MODIFIER k,
		int x, int y) 
	{
#ifndef WIN32
		struct timeval tv;
#endif
		pressPoint = lastPoint = getTrackballVector(x, y);
		switch(k) {
  case NO_MODIFIER:
	  B[w] = BUTTON_PRESSED;
	  break;
  case CTRL_KEY:
	  B[w] = CTRL_BUTTON_PRESSED;
	  break;
  case SHIFT_KEY:
	  B[w] = SHIFT_BUTTON_PRESSED;
	  break;
		}
#ifndef WIN32
		gettimeofday(&tv, 0);
		clickTime = tv.tv_sec * 1000000 + tv.tv_usec;
#else
		clickTime=time(0);
#endif
		renderDuration = 0;
		// send event to the seclected object
		if (selectedObject) {
			selectedObject->mouseButtonPressed(w,k,this,x,y);
		}
		else if (background) {
			background->mouseButtonPressed(w,k,this,x,y);
		}

	}

	void Camera2D::rotateCamera(Rotation2D &rot) {
		int i, j;
		Matrix3x3<double> r;

		for(i = 0; i < 2; ++i) {
			r.setValue(i,2,matTransform.getValue(i,2));
			r.setValue(2,i,matTransform.getValue(2,i));
			for(j = 0; j < 2; ++j) {
				r.setValue(i, j,(double)
					matTransform.getValue(i, 0) * rot.getValue(j, 0) +
					matTransform.getValue(i, 1) * rot.getValue(j, 1));
			}
		}
		matTransform = r;


	}
	/*----------------------------*/ 
	void Camera2D::setBackgroundColor(float r, float g, float b) {
		float f4[4];
		glClearColor(r, g, b, ALPHA_VAL);
		f4[0] = r; f4[1] = g; f4[2] = b; f4[3] = ALPHA_VAL;
	}

	void Camera2D::printString(unsigned int posx,unsigned int posy, 
		char *str, GLuint font) 
	{
		if (font==0)
			font=defaultFont;
		unsigned int l=strlen(str);
		unsigned int i;
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(-0.5, width-0.5, -0.5, height - .5, 0.0, 1.0);
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		glDisable(GL_LIGHTING);
		glColor3f(1.0, 1.0, 1.0);
		glRasterPos2i(posx,posy);

		for(i=0;i<l;++i)
			if (str[i] == '')
				glCallList(font+169);
			else
				glCallList(font+(GLuint)str[i]);

		glPopMatrix();
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
	}


	void Camera2D::translateCamera(Vec2<double> &t) {
		center -= t;
	}

	void Camera2D::scaleCamera(double s) {
		scale *= s;
	}
	void Camera2D::mouseMoved(int x, int y) {
		double s,autoScale;
		Vec2<double> p = getTrackballVector(x, y);

		if (background)
			background->mouseButtonMoved(this,x,y);
		if(B[MIDDLE_BUTTON] != BUTTON_RELEASED ||
			B[LEFT_BUTTON] != BUTTON_RELEASED) {
				moveMode = ROTATION_MODE;

				// button 2
				switch(B[MIDDLE_BUTTON]) {

		  case BUTTON_PRESSED:
			  autoRotation = getTrackballRotation(lastPoint, p);
			  lastPoint = p;
			  switch(motionMode) {
		  case CameraTypes::MOVE_CAMERA:
			  if (!background)
				  rotateCamera(autoRotation);
			  break;
		  case CameraTypes::MOVE_CAMERA_OBJECT:
			  rotateCameraObject(autoRotation);
			  break;
		  case CameraTypes::MOVE_SELECTED_OBJECT:
			  rotateObject(autoRotation);
			  break;
			  }
			  break;

		  case CTRL_BUTTON_PRESSED:
			  {
				  Vec2<double> autoTranslation = getTrackballTranslation(p, lastPoint);
				  lastPoint = p;

				  switch(motionMode) {
		  case CameraTypes::MOVE_CAMERA:
			  if (!background)
				  translateCamera(autoTranslation);
			  break;
		  case CameraTypes::MOVE_CAMERA_OBJECT:
			  translateCameraObject(autoTranslation);
			  break;
		  case CameraTypes::MOVE_SELECTED_OBJECT:
			  translateObject(autoTranslation);
			  break;
				  }
			  }
			  break;

		  case SHIFT_BUTTON_PRESSED:
			  {

				  Vec2<double> autoTranslation= getTrackballZTranslation( p.getY() - lastPoint.getY());
				  lastPoint = p;
				  switch(motionMode) {
		  case CameraTypes::MOVE_CAMERA:
			  if (!background)
				  translateCamera(autoTranslation);
			  break;
		  case CameraTypes::MOVE_CAMERA_OBJECT:
			  translateCameraObject(autoTranslation);
			  break;
		  case CameraTypes::MOVE_SELECTED_OBJECT:
			  translateObject(autoTranslation);
			  break;
				  }
			  }
			  break;

		  case BUTTON_RELEASED:
			  break;
				}

				/* button 1 */
				switch(B[LEFT_BUTTON]) {
		  case BUTTON_PRESSED:
			  if (selectedObject)
				  selectedObject->mouseButtonMoved(this,x,y);
			  break;

		  case CTRL_BUTTON_PRESSED:
			  autoScale = s = getTrackballScale(p, lastPoint);
			  lastPoint = p;
			  switch(motionMode) {
		  case CameraTypes::MOVE_CAMERA:
			  if (!background)
				  scaleCamera(s);
			  break;
		  case CameraTypes::MOVE_CAMERA_OBJECT:
			  scaleCameraObject(s);
			  break;
		  case CameraTypes::MOVE_SELECTED_OBJECT:
			  scaleObject(s);
			  if (selectedObject)
				  selectedObject->mouseButtonMoved(this,x,y);
			  break;
			  }
			  break;
		  case SHIFT_BUTTON_PRESSED:
			  // send event to the selected object
			  if (selectedObject)
				  selectedObject->mouseButtonMoved(this,x,y);
			  break;
		  case BUTTON_RELEASED:
			  break;
				}
			}


//			if (cameraActivation) {
//				cameraActivation->redraw();
		//	}

	}

	/*---------------------------------*/
	void Camera2D::resize(void) {
		glViewport(margin, margin, width-2*margin,
			height-2*margin);
//				if (background)
//		CERR << background->getSliceName() <<" width=" << width << "height=" << height << "margin=" << margin << ENDL;

		setViewingParameters();
		if(stencilBuffer) buildStencilBufferDisplayList();
	}

	/*-----------------------------------------------*/
	void Camera2D::applyTransformation() {
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); 
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		//  glLoadIdentity();
		glScaled(scale, scale, 1);
		glTranslated(-refCenter.getX(), -refCenter.getY(),0); 
		glMultMatrixd((const double *) matTransform);
		//glTranslated(center.getX(), center.getY(), 0);
	} 
	/*------------------------------------------*/
	void Camera2D::setProjectionParameters(bool reset)
	{
		/* Set the Projection Matrix */
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		if (background==0) {
			lowerCorner=Vec2<double>(-1.0*width/2, -1.0*height/2);
			upperCorner=Vec2<double>( width/2,height/2);
			//    glScaled(scale, scale, 1);
			gluOrtho2D(-1.0*width/2, width/2,-1.0*height/2, height/2);
		} else {
			Vec2<double> vmin,vmax;
			background->getViewportSize(vmin,vmax);
			/// preserves the aspect ratio of the image
			double ratioH,ratioW;
			double realAspectRatio=(double)height/(double)width;
			double theoreticAspectRatio=(vmax.getY()-vmin.getY())/(vmax.getX()-vmin.getX());
			if (realAspectRatio>theoreticAspectRatio) {
				ratioH=realAspectRatio/theoreticAspectRatio;
				ratioW=1.0;
			} else {
				ratioW=theoreticAspectRatio/realAspectRatio;
				ratioH=1.0;
			}
			this->lowerCorner=Vec2<double>(vmin.getX(),vmin.getY());
			this->upperCorner=Vec2<double>(vmax.getX()*ratioW,vmax.getY()*ratioH);
			// CERR << "lowerCorner = "<< this->lowerCorner << ENDL;
			// CERR << "upperCorner = "<< this->upperCorner << ENDL;
			//CERR << "slice= " << background->getSliceName() << " ratioW="<< ratioW << " ratioH= " << ratioH << "width= "<<getWidth()<<"height= "<<getHeight()<<ENDL;

			gluOrtho2D(vmin.getX(), vmax.getX()*ratioW, vmin.getY(), vmax.getY()*ratioH);
			center=refCenter=Vec2<double>();
			scale=1.0;	   
		}
		glMatrixMode(GL_MODELVIEW);
	}
	/*------------------------------------------*/
	void Camera2D::setViewingParameters(bool reset) {

		/* Set the Projection Matrix */
		setProjectionParameters(reset);
		/* Set the Viewing Matrix */
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity(); 
	}
	void Camera2D::init(void) {
		glDrawBuffer(GL_BACK);
	}



	void Camera2D::redraw(void) {
		Vec2<double> oldRefCenter;
		Vec2<double> delta;
		// compute model view matrix
		applyTransformation();
		/// set projection matrix
		setProjectionParameters();

		glDisable (GL_DEPTH_TEST);

		if(stencilBuffer){
			glCallList(stencilBufferDisplayList);
			glEnable(GL_STENCIL_TEST);
		}

		/// redraw background (draw image)
		if (background)
			background->redraw();

		// draw all scene objects
		ObjectsList< Data2D >::iterator it(getScene()->firstData());
		for(; it!=getScene()->lastData(); it++) {
			if((*it)->getRedrawOrder() == 1) {
				(*it)->redrawAll(this);
			}
		}
		for(it = getScene()->firstData(); it!=getScene()->lastData(); it++) {
			if((*it)->getRedrawOrder() == 2) {
				(*it)->redrawAll(this);
			}
		}



		if(stencilBuffer)
			glDisable(GL_STENCIL_TEST);



		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();

		count++; 
		if (showFrameRate) {
			printString(5,30,stringFrameRate,smallFont);
		}
		if (showCopyright)
			printString(5,5,stringCopyright,largeFont);

		// Compute display frequency every seconds
		if (globalTimer.getTime() > 1.0) {
			sprintf(stringFrameRate,"Frame Rate   : %.1f Hz (%.4f s/iter.)",
				count/globalTimer.getTime(),globalTimer.getTime()/count);
			count = 0;
			globalTimer.reset();
		}

	}
	Camera2DBackground *Camera2D::getBackground() const {
	  return background;
  }
	void Camera2D::doFocus(void) {
		Vec2<double> sceneCenter;
		double sceneRadius;
		/// only changes the radius and scene center if there are no backgrounds
		if (background==0) {
			scene->getCenterRadius(sceneCenter, sceneRadius);
			if(sceneRadius > 0.0) {
				focusedYet = true;
				center = refCenter=sceneCenter;
				radius = sceneRadius;
				scale = (width+height)/(4.0*radius);
				if (getCameraActivation())
					getCameraActivation()->setContext();
				setViewingParameters();
			} 
		}
	}


	void Camera2D::mapScreenTo2DWorldCoordinates(const Vec2<double> &pixel,Vec2<double> &position)
	{
		/// actually pixel is not the right frame : one must be adjust it according to the
		/// camera size
		double rx=(upperCorner.getX()-lowerCorner.getX())/width;
		double ry=(upperCorner.getY()-lowerCorner.getY())/height;
//		CERR << "lower=" << lowerCorner << "upper=" << upperCorner << ENDL;

//		Vec2<double> p=Vec2<double>(pixel.getX()*rx+lowerCorner.getX(),
//    -pixel.getY()*ry+upperCorner.getY());
		Vec2<double> p=Vec2<double>(pixel.getX()*rx+lowerCorner.getX(),
			pixel.getY()*ry+lowerCorner.getY());
		/*
		position=  p/scale+refCenter; */
		position=p/(radius*scale) + center;
		//CERR << "pixel= "<< pixel << " height= "<< height << "p= "<< p<< ENDL;
		//CERR << "position=" << position << "radius=" << radius<< ENDL;
	}
	void Camera2D::map2DWorldToScreenCoordinates(const Vec2<double> &position,Vec2<double> &pixel)
	{
		pixel=refCenter + (position-center)*(radius*scale);
	}


	void Camera2D::setStencilBuffer(bool nstencil) {
		if(stencilBuffer!=nstencil){
			stencilBuffer=nstencil;
		}
	}

	void Camera2D::buildStencilBufferDisplayList() {
		if(stencilBufferDisplayList!=0){
			glDeleteLists(stencilBufferDisplayList,1);
		}
		const double scale = ((double)height)/width;
		stencilBufferDisplayList=glGenLists(1);
		glNewList(stencilBufferDisplayList,GL_COMPILE);

		glClearStencil(0x0);
		glEnable(GL_STENCIL_TEST);
		glClear(GL_STENCIL_BUFFER_BIT);
		glStencilFunc(GL_ALWAYS,0x1,0x1);
		glStencilOp(GL_REPLACE,GL_REPLACE,GL_REPLACE);
		glColorMask(0,0,0,0);
		glDepthMask(0);

		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		glScaled(scale,1.0,0.0);

		GLUquadricObj * disk = gluNewQuadric();
		gluDisk(disk,0,1,40,1);
		gluDeleteQuadric(disk);

		glColorMask(1,1,1,1);
		glDepthMask(1); 

		glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
		glStencilFunc(GL_NOTEQUAL,0x1,0x1);

		glDisable(GL_LIGHTING);

		glColor3f(0.0,0.0,0.0);
		glBegin(GL_QUADS);
		glVertex3d(1/scale,1.0,0.0);
		glVertex3d(-1/scale,1.0,0.0);
		glVertex3d(-1/scale,-1.0,0.0);
		glVertex3d(1/scale,-1.0,0.0);
		glEnd();

		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();

		glEnable(GL_LIGHTING);

		glStencilFunc(GL_EQUAL,0x1,0x1);
		glDisable(GL_STENCIL_TEST);

		glEndList();
	}




} // namespace yav


