#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#pragma implementation "graphics2D/SliceCamera.h"
#include <graphics2D/SliceCamera.h>



namespace yav {


// display a 2D scene
SliceCamera::SliceCamera(Scene2D *scene,
			 double _slice,
			 SLICE_ORIENTATION _geometry,
			 double _zoom, ScaleCamera *sc)
  :  Camera2D(scene, 90U, 90U, _zoom),Owner<ScaleCamera>(sc),
    slice(_slice), thickness(1.0), geometry(_geometry), 
    dieOnBackground(true),background(0)  {
}


SliceCamera::~SliceCamera() {
  ObjectsList<Data2D>::iterator it= scene->firstData();
  Sliceable *sc;
  for(; it!=scene->lastData(); it++)
    {
      sc=dynamic_cast<Sliceable *>((*it));
      if (sc)
	sc->destroyedCamera(*this);
    }
#ifdef TCL_INTERFACE
 // while(Tcl_DoOneEvent(TCL_ALL_EVENTS | TCL_DONT_WAIT));
#endif
} TclString SliceCamera::objectName(void) const {
	TclString _name="slicecam";
	return _name;
}
void SliceCamera::setScaleCamera(ScaleCamera *sc) {
  Owner<ScaleCamera>::pointer =sc;
  sc->reset(this,sliceName(geometry));
}
void SliceCamera::resetScaleCamera() {
  getScaleCamera()->reset(this,sliceName(geometry));
}

// destroy viewer upon inrimage destruction
void SliceCamera::backgroundDying(void) {
  if(dieOnBackground) delete this;
  else
    background=0;
}
ScaleCamera * SliceCamera::getScaleCamera() const
{
  ScaleCamera *sc=pointer();
  return (sc);
}

const char *SliceCamera::sliceName(SliceCamera::SLICE_ORIENTATION geom) const{
  switch(geom) {
  case SliceCamera::SLICE_X:
    return "X";
  case SliceCamera::SLICE_Y:
    return "Y";
  case SliceCamera::SLICE_Z:
    return "Z";
  case SliceCamera::SLICE_ALPHA:
    return "alpha";
  }
  return 0;
}

SliceCamera::SLICE_ORIENTATION SliceCamera::geomByName(const char *name) const{
  if(!strcmp(name, "X")) return SliceCamera::SLICE_X;
  else if(!strcmp(name, "Y")) return SliceCamera::SLICE_Y;
  else if(!strcmp(name, "Z")) return SliceCamera::SLICE_Z;
  else if(!strcmp(name, "alpha")) return SliceCamera::SLICE_ALPHA;
  else return (SliceCamera::SLICE_ORIENTATION) -1;
}
void SliceCamera::scaleValue(const TclScale *scale, double val) {
  setSlice(val);
}
// Change slice geometry. Cause recompute method to be called.
void SliceCamera::setGeometry(SLICE_ORIENTATION _geom) {
  if(_geom != geometry) {
    geometry = _geom;
    slice = 0.0;
    getScaleCamera()->reset(this,EMPTY_STRING);
    update();
  }
}


void SliceCamera::setSlice(double val) {
  if(val != slice) {
    slice = val;
    if (getScaleCamera())
      getScaleCamera()->set(val);
    update();
  }
}


void SliceCamera::focusOnBackground(void) {
  if(background) {
    double x=0.0, y=0.0,xmin=0.0,ymin=0.0;
    const unsigned int *va = background->getVisibleArea();
    const double *sc = background->getScaleVector();
    switch(geometry) {
    case SLICE_X:
      x = (va[5] - va[2]+1)*sc[2];
      y = (va[4] - va[1]+1)*sc[1];
	  xmin=(va[2]+va[5]+1)*sc[2]/2.0;
	  ymin=(va[1]+va[4]+1)*sc[1]/2.0;
	  offset[0]=(va[2]*sc[2]);
	  offset[1]=(va[1]*sc[1]);
      break;
    case  SLICE_Y:
	  x = (va[3] - va[0]+1)*sc[0];
      y = (va[5] - va[2]+1)*sc[2];
	  xmin=(va[0]+va[3]+1)*sc[0]/2.0;
	  ymin=(va[5]+va[2]+1)*sc[2]/2.0;
	  offset[0]=(va[0]*sc[0]);
	  offset[1]=(va[2]*sc[2]);
	  break;
    case SLICE_Z:
    case SLICE_ALPHA:
      x = (va[3] - va[0]+1)*sc[0] ;
      y = (va[4] - va[1]+1)*sc[1] ;
	  xmin=(va[0]+va[3]+1)*sc[0]/2.0;
	  ymin=(va[1]+va[4]+1)*sc[1]/2.0;
	  offset[0]=(va[0]*sc[0]);
	  offset[1]=(va[1]*sc[1]);
      break;
    }
    setCenter (Vec2<double>(xmin,ymin));
    resize((unsigned int)(x * scale),(unsigned int)(y * scale));
    updateDisplay();
  }  
}




void SliceCamera::changeBackground(SliceCameraBackground *b) {
  if(background) delete background;
  background=b;
}



void SliceCamera::translateCamera(Vec2<double> &t) {
  Camera2D::translateCamera(t);
  update();
}
void SliceCamera::update() {
  updateSliceables();
}
void SliceCamera::updateSliceables() {
  ObjectsList<Data2D>::iterator it= scene->firstData();
  Sliceable *sc;
  for(; it!=scene->lastData(); it++)
    {
      sc=dynamic_cast<Sliceable *>((*it));
      if (sc)
	sc->recompute(this);
    }
  updateDisplay();
}

void SliceCamera::scaleCamera(double s) {
  Camera2D::scaleCamera(s);
  update();
}



void SliceCamera::doFocus(void) {
  if(!background) Camera2D::doFocus();
}

void SliceCamera::setBackground(SliceCameraBackground *_background) {
  background=_background;
  if (pointer())
    pointer()->reset(this,EMPTY_STRING);
}
double  SliceCamera::getMin(const TclScale *scale) const {
  if (background)
    {  
      const unsigned int *v = background->getVisibleArea();
      switch(geometry) {
      case SliceCamera::SLICE_X:
	return v[0];
	break;
      case SliceCamera::SLICE_Y:
	return v[1];
	break;
      case SliceCamera::SLICE_ALPHA:
      case SliceCamera::SLICE_Z:
	return v[2];
	break;
      default:
	return 0;
	break;
      }
    }
  else
    return minTypeValue<float>(0);
}
double  SliceCamera::getMax(const TclScale *scale) const {
  if (background)
    {
      const unsigned int *v = background->getVisibleArea();
      switch(geometry) {
      case SliceCamera::SLICE_X:
	return v[3];
	break;
      case SliceCamera::SLICE_Y:
	return v[4];
	break;
      case SliceCamera::SLICE_ALPHA:
      case SliceCamera::SLICE_Z:
	return v[5];
	break;
      default:
	return 100;
	break;
      }
    }
  else
    return maxTypeValue<float>(0);
}


}


