#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "graphics2D/SliceScale.h"
#include <basics/Constants.h>
#include <graphics2D/SliceScale.h>

namespace yav {


SliceScale::SliceScale(TclModule *module, const TclString& scaleName,
		       Scalable *sc,  const std::type_info &id)
  : TclScale(id,scaleName,module,sc) {
  init();
}
SliceScale::SliceScale(TclModule *module, const TclString& scaleName,
		       const TclString& label, const TclString color,
		       Scalable *sc,  const std::type_info &id)
  : TclScale(id,scaleName,module,sc) {
#ifndef WIN32
  tickle("scale %s -orient horizontal -width 10 -font fixed -label %s -bg %s", (char *) scaleName, (char *)label, (char *)color);
#else
    tickle("scale %s -orient horizontal -width 10 -label %s -bg %s", (char *) scaleName, (char *)label, (char *)color);
#endif
  init();
}


SliceScale::SliceScale(TclModule *module, const TclString& scaleName,
		       const TclString& label, const TclString color,
		       std::list<Scalable *>& list, const std::type_info &id)
  : TclScale(id, scaleName, module,list) {
#ifndef WIN32
  tickle("scale %s -orient horizontal -width 10 -font fixed -label %s -bg %s", (char *) scaleName, (char *)label, (char *)color);
#else
  tickle("scale %s -orient horizontal -label %s -bg %s", (char *) scaleName, (char *)label, (char *)color);
#endif
  init();
}


SliceScale::~SliceScale() {
}


void SliceScale::init(void) {

  tickle("%s configure -command {%s%u -command}",
	 (char *) TCLwindow, TclScale::getObjectName(),
	 getRef());

  tickle("bind %s <Up> \"%s set \\[expr \\[%s get\\] + 2\\]\"",
	 (char *) TCLwindow, (char *) TCLwindow,
	 (char *) TCLwindow);
  tickle("bind %s <Down> \"%s set \\[expr \\[%s get\\] - 2\\]\"",
	 (char *) TCLwindow, (char *) TCLwindow,
	 (char *) TCLwindow);
}


void SliceScale::reset(Scalable *sa, const TclString& lab) {
  min = sa->getMin(this);
  max = sa->getMax(this);
  value = sa->getSlice(this);
  firstupdate=true;
  tickle("%s configure -from %u -to %u -resolution %f; %s set %f",
	 (char *) TCLwindow, (unsigned int) min,
	 (unsigned int) max, sa->getStep(this),
	 (char *) TCLwindow, value);

  if(lab != EMPTY_STRING) setLabel(lab);
  firstupdate=false;

  // command(get());
}


void SliceScale::operator+=(Scalable *cam)
  throw(DPList<Scalable>::NullPointer, DPList<Scalable>::AlreadyInList) {
  TclScale::operator+=(cam);
  configureToMax();
}


void SliceScale::operator-=(Scalable *cam) throw(DPList<Scalable>::NotFound) {
  TclScale::operator-=(cam);
  configureToMax();
}


void SliceScale::dying(const Scalable *elt) {
  configureToMax();
  TclScale::dying(elt);
}




/* "create a new slice scale object associated to given TCL scale (-in argument) and add any given slicecam to scale."*/
int SliceScale::create(TclModule *m, int argc, char *argv[]) {
  SliceScale *sc;
  char *name = 0;
  int i;

  for(i = 1; i < argc; i++) {
    // -in %w
    if(!strcmp(argv[i], "-in")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(!m->windowExists(argv[i])) return COMMAND_SYNOPSIS;
      name = argv[i];
    }
    else return COMMAND_SYNOPSIS;
  }

  if(!name) return COMMAND_SYNOPSIS;
  sc = new SliceScale(m, name);
  m->setResult("%s%u", (char *) sc->getObjectName(), sc->getRef());
  return TCL_OK;
}


} // namespace yav



