#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "graphics2D/SurfaceView.h"
#include <graphics2D/SurfaceView.h>
#include <graphics2D/Camera2DActivation.h>


namespace yav {

double SurfaceView::selectThreshold=13.0;

static float zoneColor[6][3]= { { 1,0,0}, {0,1,0},{0,0,1},{0,1,1},{1,0,1},{1,1,0}};
/*
static char *zoneColor[20] = {
  "red", "green", "blue", "yellow", "orange", "purple", "cyan", "ivory",
  "aquamarine", "gold", "lightsalmon", "darkorchid", "skyblue", "seagreen",
  "chocolate", "pink", "turquoise", "coral", "royalblue", "springgreen",
};
*/



SurfaceView::SurfaceView(Scene2D *_scene,SurfaceSliceable *_mesh) : Data2D(_scene), Owner<SurfaceSliceable>(_mesh), useZonesColor(false)  {
  traceColor[0]=1.0; traceColor[1]=0.0; traceColor[2]=0.0;
  traceWidth = 3.0;
  interactionMode=GRAB_MODE;
  selectedConstraint=0;
  drawSegment=false;

}

SurfaceView::~SurfaceView(){
  std::list<PointPosition *>::iterator it=constraintList.begin();
  for(;it!=constraintList.end();it++) {
    delete (*it);
  }
}

void  SurfaceView::dying(ObjectPointer<SurfaceSliceable> const &) {
    delete this;
  }

void SurfaceView::redraw(Camera2D *cam) {
	if (cam->getBackground()) {
		double crossLength=5.0;
		//double crossWidth=2.0;
		std::list<SliceContour *> contours;
		std::list<SliceContour *>::iterator it;
		Camera2DBackground *b=cam->getBackground();
		Vec3<double> org,normal,pos;
		Vec2<double> p;
		 int i;
		float *oldX,*oldY;
		 std::list<PointPosition *>::iterator itp;

		b->getImagePlane(org,normal);
		glLineWidth(traceWidth);
		glBegin(GL_LINES);
		if (useZonesColor==false) 
			glColor3f(traceColor[0],traceColor[1],traceColor[2]);

		contours =
			getSurface()->intersectWithPlane(normal,org);

		for(it = contours.begin(); it != contours.end(); it++) {
					oldX=oldY=0;
			for (i = 0; i < (*it)->nbvertices; i++) {
				pos=Vec3<double>((double)(*it)->X[i],(double)(*it)->Y[i],(double)(*it)->Z[i]);
				b->map3DWorldTo2DWorldCoordinates(pos,p);
				(*it)->X[i] = (float) p.getX();
				(*it)->Y[i] = (float) p.getY();
				if (oldX) {
					if ((useZonesColor) &&((*it)->zone[i-1]>=0)) {
						unsigned int zoneIndex=(*it)->zone[i-1];
						glColor3f(zoneColor[zoneIndex][0],
							zoneColor[zoneIndex][1],
							zoneColor[zoneIndex][2]);
					} 

					glVertex2f(*oldX,*oldY);
					glVertex2f((*it)->X[i],(*it)->Y[i]);

					oldX++;
					oldY++;

				} else {
					oldX=(*it)->X;
					oldY=(*it)->Y;
				}
		
			}

			if ((useZonesColor) &&((*it)->zone[(*it)->nbvertices-1]>=0)) {
				unsigned int zoneIndex=(*it)->zone[(*it)->nbvertices-1];
				glColor3f(zoneColor[zoneIndex][0],
					zoneColor[zoneIndex][1],
					zoneColor[zoneIndex][2]);
			} 

			glVertex2f(*oldX,*oldY);
			glVertex2f((*it)->X[0],(*it)->Y[0]);
		}
	glEnd();
     
  
	itp=constraintList.begin();
	for(;itp!=constraintList.end();itp++) {
		if (b->map3DWorldTo2DWorldCoordinates((*itp)->position,p)) {
			glColor3f(0,1.0,1.0);
			glBegin(GL_LINES);
			glVertex2d(p.getX()-crossLength,p.getY());
			glVertex2d(p.getX()+crossLength,p.getY());
			glVertex2d(p.getX(),p.getY()-crossLength);
			glVertex2d(p.getX(),p.getY()+crossLength);
			glEnd();
		}
	 
	}

	if (drawSegment) 
	{
		glBegin(GL_LINES);
		glColor3f(1.0,0,1.0);
		glVertex2dv((const double *)segmentPoint[0]);
		glVertex2dv((const double *)segmentPoint[1]);
		glEnd();
		drawSegment=false;
	}
	}

}
void SurfaceView::mouseButtonPressed(CameraTypes::WHICH_BUTTON button,
				     CameraTypes::KEY_MODIFIER modifier,
				     Camera2D *cam,int mx,int my)
{
	if (button==CameraTypes::LEFT_BUTTON) {
	if (cam->getBackground()) {
		Vec3<double> normal,p,org;
		Vec2<double> pixel=Vec2<double>(mx,my);
		Vec2<double> posPlane;

		cam->mapScreenTo2DWorldCoordinates(pixel,posPlane);
		cam->getBackground()->map2DWorldTo3DWorldCoordinates(posPlane,p);
		cam->getBackground()->getImagePlane(org,normal);

		if (interactionMode==GRAB_MODE) {
			Vec3<double> *v= getSurface()->grab2D(normal,p);
			if (v)
			{
				Vec3<double> projectedPoint= *v + normal * dot<double>((p- *v),normal);
				segmentPoint[0]=posPlane;
				cam->getBackground()->map3DWorldTo2DWorldCoordinates(projectedPoint,segmentPoint[1]);
				drawSegment=true;
				if (cam->getCameraActivation())
					cam->getCameraActivation()->redraw();
			}
			} else if (interactionMode==POSITION_CONSTRAINT_MODE) {

				unsigned int _ref=getSurface()->addSurfacePointConstraint(p);
				SurfaceView::PointPosition *pp= new SurfaceView::PointPosition();
				pp->ref=_ref;
				pp->position=p;
				constraintList.push_back(pp);
				redrawScene();


			} else if  (interactionMode==MOVE_POSITION_CONSTRAINT_MODE) {
				/* set to null to find the closest position constraint */
				selectedConstraint=0;
				PointPosition *bestPoint=0;
				double dist2,bestDist2= MAX_DOUBLE;
				std::list<PointPosition *>::iterator itp=constraintList.begin();
				for (;itp!=constraintList.end();itp++) {
					dist2=((*itp)->position-p).norm2();
					if ((dist2 <SurfaceView::selectThreshold) && (dist2<bestDist2)) {
						bestDist2=dist2;
						bestPoint=(*itp);
					}
				}
				if (bestPoint) {
					selectedConstraint=bestPoint;
				}
			} else if  (interactionMode==REMOVE_POSITION_CONSTRAINT_MODE) {
				/* set to null to find the closest position constraint */
				PointPosition *bestPoint=0;
				double dist2,bestDist2= MAX_DOUBLE;
				std::list<PointPosition *>::iterator itp=constraintList.begin();
				std::list<PointPosition *>::iterator itBest;
				for (;itp!=constraintList.end();itp++) {
					dist2=((*itp)->position-p).norm2();
					if ((dist2 <SurfaceView::selectThreshold) && (dist2<bestDist2)) {
						bestDist2=dist2;
						bestPoint=(*itp);
						itBest=itp;
					}
				}
				if (bestPoint) {
					constraintList.erase(itBest);
					getSurface()->deleteSurfacePointConstraint(bestPoint->ref);
					redrawScene();
				}
			}
		}
	}

}

void SurfaceView::mouseButtonMoved(Camera2D *cam,int mx,int my) {
	if ((interactionMode==MOVE_POSITION_CONSTRAINT_MODE) && (selectedConstraint)){


		if ((cam) && (cam->getBackground()))
		{
			Vec3<double> p;
			Vec2<double> pixel=Vec2<double>(mx,my);
			Vec2<double> posPlane;

			cam->mapScreenTo2DWorldCoordinates(pixel,posPlane);
			cam->getBackground()->map2DWorldTo3DWorldCoordinates(posPlane,p);

			selectedConstraint->position=p;
			getSurface()->moveSurfacePointConstraint(p,selectedConstraint->ref);
			redrawScene();
		}
	}
}
} // namespace yav

