#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#pragma implementation "graphics2D/TclBoundingBox3D.h"
#include <graphics2D/TclBoundingBox3D.h>


namespace yav {


TclBoundingBox3D::TclBoundingBox3D(TclModule *m,
			 TclScene2D *scene, 
			 unsigned int x1, unsigned int y1, unsigned int z1,
			 unsigned int x2, unsigned int y2, unsigned int z2,
			 const TclString& _color, const TclString& _menu)
  :  Sliceable(scene->getClient()),TclObject(typeid(TclBoundingBox3D),m), color(_color), menu(_menu), face(NONE)  {
  done(typeid(TclBoundingBox3D));
  
  if(x1 < x2) {
    box[0] = x1;
    box[3] = x2;
  }
  else {
    box[0] = x2;
    box[3] = x1;
  }
  if(y1 < y2) {
    box[1] = y1;
    box[4] = y2;
  }
  else {
    box[1] = y2;
    box[4] = y1;
  }
  if(z1 < z2) {
    box[2] = z1;
    box[5] = z2;
  }
  else {
    box[2] = z2;
    box[5] = z1;
  }

  ObjectsList<Camera2D>::iterator it = scene->getClient()->firstCamera();
  for(;it!=scene->getClient()->lastCamera(); it++) {
    TclSliceCamera *cam = (TclSliceCamera *) (*it)->getInterface();
    if (cam)
      {
	cam->applyObjectSelection(this);
	if(menu != EMPTY_STRING) {
	  cam->tickle("%s %s%u %s%u", (char *) menu,
		      TclObjectKind::getObjectName(typeid(TclSliceCamera)),
		      cam->getRef(),TclObjectKind::getObjectName(typeid(TclBoundingBox3D)),TclObject::getRef());
	}
      }

  }  

  updateDisplay();
}


TclBoundingBox3D::~TclBoundingBox3D() {
  TclSliceCamera *cam;
  //  unselectScene();
  ObjectsList<Camera2D>::iterator it = scene->firstCamera();
  for(;it!=scene->lastCamera(); it++) {
    cam = (TclSliceCamera *) (*it)->getInterface();
    cam->tickle("%s delete bbox3D", (char *) cam->getCanvas());
    cam->installPopupMenuCommand();
    //    cam->installDefaultPopupMenuCommand();
  }
  TclScene2D *tclscene=  (TclScene2D *)( this->getScene()->getInterface());
  tclscene->unselectCameraObject((Data2D *) this);
  deleted(typeid(TclBoundingBox3D));

}


void TclBoundingBox3D::setClosestFace(SliceCamera::SLICE_ORIENTATION geom, unsigned int x,
				 unsigned int y, unsigned int z) {
  unsigned int opt, v;

  switch(geom) {
  case SliceCamera::SLICE_X:
    opt = abs((int)(box[1] - y));
    face = YMIN;
    v = abs((int)(box[2] - z));
    if(v < opt) { 
      opt = v; 
      face = ZMIN;
    }
    v = abs((int)(box[4] - y));
    if(v < opt) { 
      opt = v;
      face = YMAX;
    }
    v = abs((int)(box[5] - z));
    if(v < opt) {
      opt = v;
      face = ZMAX;
    }
    break;
  case SliceCamera::SLICE_Y:
    opt = abs((int)(box[0] - x));
    face = XMIN;
    v = abs((int)(box[2] - z));
    if(v < opt) { 
      opt = v; 
      face = ZMIN;
    }
    v = abs((int)(box[3] - x));
    if(v < opt) { 
      opt = v; 
      face = XMAX;
    }
    v = abs((int)(box[5] - z));
    if(v < opt) {
      opt = v;
      face = ZMAX;
    }
    break;
  case SliceCamera::SLICE_Z:
  case SliceCamera::SLICE_ALPHA:
    opt = abs((int)(box[0] - x));
    face = XMIN;
    v = abs((int)(box[1] - y));
    if(v < opt) {
      opt = v; 
      face = YMIN;
    }
    v = abs((int)(box[3] - x));
    if(v < opt) { 
      opt = v; 
      face = XMAX;
    }
    v = abs((int)(box[4] - y));
    if(v < opt) { 
      opt = v;
      face = YMAX;
    }
    break;
  }

  moveFace(x, y, z);
}


void TclBoundingBox3D::moveFace(unsigned int x, unsigned int y, unsigned int z) {
  if(face != NONE) {
    if(face == XMIN || face == XMAX) box[(int) face] = x;
    else if(face == YMIN || face == YMAX) box[(int) face] = y;
    else if(face == ZMIN || face == ZMAX) box[(int) face] = z;
    updateDisplay();
  }
}


void TclBoundingBox3D::recompute(SliceCamera *cam) {
  TclSliceCamera *tcam;

  int mx1, mx2, my1, my2;
  if((cam) && (cam->getBackground())) {
    SliceCamera::SLICE_ORIENTATION geom = cam->getGeometry();
    
    cam->getBackground()->getPixelPositionFromImageCoordinates(box[0],box[1],box[2],mx1,my1);
    cam->getBackground()->getPixelPositionFromImageCoordinates(box[3],box[4],box[5],mx2,my2);
    tcam = cam->getInterface();

    tcam->tickle("%s delete bbox3D", (char *) tcam->getCanvas());
    
    if((geom == SliceCamera::SLICE_X && cam->getSlice() >= box[0] &&
	cam->getSlice() <= box[3]) ||
       (geom == SliceCamera::SLICE_Y && cam->getSlice() >= box[1] &&
	cam->getSlice() <= box[4]) ||
       (geom == SliceCamera::SLICE_Z && cam->getSlice() >= box[2] &&
	cam->getSlice() <= box[5])) {
      tcam->getModule()->tickle("%s create rectangle %u %u %u %u -outline %s -width 3 -tags bbox3D; %s raise bbox3D",
				(char *) tcam->getCanvas(),
				mx1, my1, mx2, my2, (char *) color,
				(char *) tcam->getCanvas());

      
    }
  }
}

void TclBoundingBox3D::redraw(Camera2D *cam) {
  switch(redrawFlag) {
  case Data2D::UPDATE:
    recompute(dynamic_cast<SliceCamera *>(cam));
    break;
  case Data2D::CREATION:
  case Data2D::REDRAW:
    break;
  case Data2D::DESTRUCTION:
    break;
  }
}

void TclBoundingBox3D::mouseButtonPressed(CameraTypes::WHICH_BUTTON button,
					  CameraTypes::KEY_MODIFIER modifier,
					  Camera2D *cam,int mx,int my) {
  SliceCamera *slicecam=dynamic_cast<SliceCamera *> (cam);
  if ((slicecam) && (slicecam->getBackground()))
    {
      unsigned int vx,vy,vz;
      slicecam->getBackground()->getImageCoordinates(mx,my,vx,vy,vz);
      setClosestFace(slicecam->getGeometry(), vx, vy, vz);      
    }
}
void TclBoundingBox3D::mouseButtonMoved(Camera2D *cam, 
					int  mx, int my) {
  SliceCamera *slicecam=dynamic_cast<SliceCamera *> (cam);
  if ((slicecam) && (slicecam->getBackground()))
    {
      unsigned int vx,vy,vz;
      slicecam->getBackground()->getImageCoordinates(mx,my,vx,vy,vz);
      moveFace(vx, vy, vz);      
    }
}
void TclBoundingBox3D::mouseButtonReleased(CameraTypes::WHICH_BUTTON button,
					   CameraTypes::KEY_MODIFIER modifier,
					   Camera2D *cam,int mx,int my) {
  SliceCamera *slicecam=dynamic_cast<SliceCamera *> (cam);
  if ((slicecam) && (slicecam->getBackground()))
    {
      unsigned int vx,vy,vz;
      slicecam->getBackground()->getImageCoordinates(mx,my,vx,vy,vz);
      releaseFace();      
    }
}
int TclBoundingBox3D::destroyCmd(int argc, char **argv) {
  TclModule *m;
  unsigned int ref;

  if(argc != 0) return COMMAND_SYNOPSIS;
  m = getModule();
  ref = TclObject::getRef();
  TclString oname=TclBoundingBox3D::getObjectName();
  delete this;
  m->setResult("%s%u", (char *) oname, ref);
  return TCL_OK;
}
int TclBoundingBox3D::boxSizeCmd(int argc, char **argv) {

  if(argc != 1) return COMMAND_SYNOPSIS;
  if (!strcmp(argv[0],"-org"))
    {
      setResult("%u %u %u",box[0],box[1],box[2]);
      return TCL_OK;
    }
  else if (!strcmp(argv[0],"-end"))
    {
      setResult("%u %u %u",box[3],box[4],box[5]);
      return TCL_OK;
    }
  else return COMMAND_SYNOPSIS;
}

int TclBoundingBox3D::create(TclModule *m, int argc, char **argv) {
  int i;

  if (argc<2)
    return COMMAND_SYNOPSIS;

  TclSliceCamera *slicecam=(TclSliceCamera *)
    TclObjectKind::getObject(typeid(TclSliceCamera), argc, argv, 1);
  if (!slicecam)
    {
      m->setResult("%s is not a slice camera",argv[1]);
      return TCL_ERROR;
    }


  unsigned int x1,y1,z1;
  x1=y1=z1=0;
  unsigned int x2,y2,z2=0;
  if (slicecam->getClient()->getBackground())
    {
      const unsigned int *v = slicecam->getClient()->getBackground()->getVisibleArea();
      x1=v[0];
      y1=v[1];
      z1=v[2];
      x2=v[3];
      y2=v[4];
      z2=v[5];
    }
  else
    {
      x2=y2=z1=1;
    }
  TclString color="green";
  TclString menu="graphics2D::viewerVisibleMenu";
  

  for(i=2; i < argc; i++) {
    if(!strcmp(argv[i], "-area")) {
      if( (i+6) >= argc) return COMMAND_SYNOPSIS;
      ++i;
      x1=(unsigned int) atoi(argv[i++]);
      y1=(unsigned int) atoi(argv[i++]);
      z1=(unsigned int) atoi(argv[i++]);
      x2=(unsigned int) atoi(argv[i++]);
      y2=(unsigned int) atoi(argv[i++]);
      z2=(unsigned int) atoi(argv[i++]);
    }
    else if(!strcmp(argv[i], "-color")) {
      if( ++i >= argc) return COMMAND_SYNOPSIS;
      color=argv[i];
    }
    else if(!strcmp(argv[i], "-menu")) {
      if( ++i >= argc) return COMMAND_SYNOPSIS;
      menu=argv[i];
    }
    else return COMMAND_SYNOPSIS;
  }

  TclBoundingBox3D *bbox = new TclBoundingBox3D(m,slicecam->getScene(),x1,y1,z1,x2,y2,z2,color,menu);
  m->setResult("%s%u", bbox->getObjectName(), bbox->TclObject::getRef());
  return TCL_OK;
}


} // namespace yav


