#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "graphics2D/TclCamera2D.h"
#include <graphics/SceneItem.h>
#include <graphics2D/TclCamera2D.h>
#include <modules/TclMatrix4D.h>
#include <graphics2D/TclScene2D.h>
#include <graphics2D/TclCamera2DActivation.h>
#include <graphics2D/TclScene2DObject.h>
#include <set>
#ifndef WIN32
#include <modules/WindowDump.h>
#endif
namespace yav {
	


static TclCamera2D *widgetBeingCreated;


 int createCamera2D(ClientData clientData, Tcl_Interp *interp, int objc,
        Tcl_Obj *const *objv)
{
    struct Togl   *togl;

    if (objc != 2) {
        Tcl_WrongNumArgs(interp, 1, objv, "pathName");
        return TCL_ERROR;
    }

    if (Togl_GetToglFromObj(interp, objv[1], &togl) != TCL_OK) {
        return TCL_ERROR;
    }

	if (widgetBeingCreated==NULL) {
		Tcl_AppendResult(interp, "createCamera2D : ", "Couldn't find a TclCamera2D being created \n", NULL);
        return TCL_ERROR;

	} else {
		widgetBeingCreated->setWidget(togl);
		Togl_SetClientData(togl,(ClientData) widgetBeingCreated);
		widgetBeingCreated->getClient()->setWidth(Togl_Width(togl));
		widgetBeingCreated->getClient()->setHeight(Togl_Height(togl));
		widgetBeingCreated->getClient()->init();
		widgetBeingCreated->getClient()->initViewport();

		widgetBeingCreated->getClient()->setCenter( widgetBeingCreated->getClient()->getRefCenter());
		glFlush();
		/// set fonts 
		widgetBeingCreated->getClient()->setDefaultFont(Togl_FontDisplayList(togl,Togl_LoadBitmapFont(togl,TOGL_BITMAP_8_BY_13)));
		widgetBeingCreated->getClient()->setSmallFont(Togl_FontDisplayList(togl,Togl_LoadBitmapFont(togl,TOGL_BITMAP_HELVETICA_12)));
		widgetBeingCreated->getClient()->setLargeFont(Togl_FontDisplayList(togl,Togl_LoadBitmapFont(togl,TOGL_BITMAP_TIMES_ROMAN_24)));

		widgetBeingCreated->getClient()->resize();
		widgetBeingCreated->getClient()->focus();

		// getModule()->tickle("tkwait visibility %s",(char *) parentFrame);
		/// set camera activation object to handle auto motion of the camera or selected object
		widgetBeingCreated->getClient()->setCameraActivation((Camera2DActivation *)new TclCamera2DActivation(widgetBeingCreated->getClient()));

		/** select the first data in the scene */
		ObjectsList<Data2D>::iterator it=widgetBeingCreated->getScene()->getClient()->firstData();
		if (it!=widgetBeingCreated->getScene()->getClient()->lastData()) {
			widgetBeingCreated->applyObjectSelection(*it);
		}

		widgetBeingCreated=NULL;
	}
 
    return TCL_OK;
}


 int displayCamera2D(ClientData clientData, Tcl_Interp *interp, int objc,
        Tcl_Obj *const *objv)
{
    Togl   *togl;

    if (objc != 2) {
        Tcl_WrongNumArgs(interp, 1, objv, "pathName");
        return TCL_ERROR;
    }

	if (Togl_GetToglFromObj(interp, objv[1], &togl) != TCL_OK)
		return TCL_ERROR;

//	CERR << "display callback from" << Tk_PathName(Togl_TkWin(togl)) << ENDL;	
	TclCamera2D *cam= (TclCamera2D *) Togl_GetClientData(togl);
	if (!cam) {
		Tcl_AppendResult(interp, "displayCamera2D : ", "Couldn't find a TclCamera2D as clientdata \n", NULL);
		return TCL_ERROR;
	}
	Togl_MakeCurrent(togl);

	cam->getClient()->redraw();
	if(cam->isRecording()) {
		TclString filename;
		filename.sprintf("%s%04u.ppm", (char *) cam->getMovieName(), cam->getFrameNumberAndIncrement());
		cam->dumpToFile(filename);
	}
	Togl_SwapBuffers(togl);
	return TCL_OK;
}

 int resizeCamera2D(ClientData clientData, Tcl_Interp *interp, int objc,
        Tcl_Obj *const *objv)
{
    Togl   *togl;

    if (objc != 2) {
        Tcl_WrongNumArgs(interp, 1, objv, "pathName");
        return TCL_ERROR;
    }

	if (Togl_GetToglFromObj(interp, objv[1], &togl) != TCL_OK)
		return TCL_ERROR;

	TclCamera2D *cam= (TclCamera2D *) Togl_GetClientData(togl);
	if (!cam) {
		Tcl_AppendResult(interp, "resizeCamera2D : ", "Couldn't find a TclCamera2D as clientdata \n", NULL);
		return TCL_ERROR;
	}

	Togl_MakeCurrent(togl);
	cam->getClient()->setWidth(Togl_Width(togl));
    cam->getClient()->setHeight(Togl_Height(togl));

	cam->getClient()->resize();
	return TCL_OK;
}
 int destroyCamera2D(ClientData clientData, Tcl_Interp *interp, int objc,
        Tcl_Obj *const *objv)
{
    Togl   *togl;

    if (objc != 2) {
        Tcl_WrongNumArgs(interp, 1, objv, "pathName");
        return TCL_ERROR;
    }


	if (Togl_GetToglFromObj(interp, objv[1], &togl) != TCL_OK)
		return TCL_ERROR;

//	CERR << "destroy callback from" << Tk_PathName(Togl_TkWin(togl)) << ENDL;	

	TclCamera2D *cam= (TclCamera2D *) Togl_GetClientData(togl);
	if (cam) {
		delete cam;
	}
	return TCL_OK;
}

/************************************************************************/
TclCamera2D::TclCamera2D(TclModule* m, const TclString& parent, TclScene2D *scn,
						 const unsigned int wInit,
						 const unsigned int hInit,
						 const std::type_info& id, Camera2D *data)
						 : TclSceneObject(id, data ? data : new Camera2D( scn ? scn->getClient():0), m) {
	record=false;
	alwaysTrackMouseMotion=false;
	frameNumber=0;

	getClient()->width = wInit;
	getClient()->height = hInit;

    widgetBeingCreated = this;
    if(parent == EMPTY_STRING) {
      parentFrame.sprintf(".viewer2D%s%u", getObjectName(), getRef());
      menuFrame = parentFrame + ".menu";
      submenuFrame = parentFrame + ".submenu";
      buttonFrame = parentFrame + ".buttons";
      getModule()->tickle("graphics2D::topViewer2D %s%u",
			  getObjectName(), getRef());
    }
    else {
      parentFrame = parent;
      getModule()->tickle("graphics2D::viewer2D %s%u %s %u %u",
			  getObjectName(), getRef(), (char *) parent, wInit, hInit);
    }
    cameraFrame = parentFrame + ".view";
  

  
	done(typeid(TclCamera2D));
}
/************************************************************************/	
void TclCamera2D::makeCurrent(void) const {
	Togl_MakeCurrent(widget);
}
/************************************************************************/
void TclCamera2D::swapBuffers(void) const {
	Togl_SwapBuffers(widget);
}

/************************************************************************/
TclCamera2D::~TclCamera2D() {
	if (widget)
		Togl_SetClientData(widget,(ClientData) 0);

	getClient()->destroyCamera();
	deleted(typeid(TclCamera2D));
  }
	

void TclCamera2D::startComputation() {
  getModule()->tickle("graphics2D::startViewer2D  %s %s", (char *) buttonFrame, (char *) name());
}


void TclCamera2D::stopComputation() {
  getModule()->tickle("graphics2D::stopViewer2D  %s %s", (char *) buttonFrame,(char *) name() );
}

void TclCamera2D::alwaysTrackMouseMotionOn() 
{
	alwaysTrackMouseMotion=true;
}
void TclCamera2D::alwaysTrackMouseMotionOff() 
{
	alwaysTrackMouseMotion=false;
}
bool TclCamera2D::getAlwaysTrackMouseMotion() const 
{
	return(alwaysTrackMouseMotion);
}
/*
 * Init
 *
 *   Called upon system startup to create TclCamera2D command.
 */

int TclCamera2D::recordCmd(int argc, char **argv) {
  if(argc < 1) return COMMAND_SYNOPSIS;
  if(!strcmp(argv[0], "frame")) {
    if(argc != 2) return COMMAND_SYNOPSIS;
    getScene()->getSceneActivation()->stopMovie();
    startMovie(argv[1]);
    setResult("movie record started at frame rate");
  }
  else if(!strcmp(argv[0], "simulation")) {
    if(argc != 2) return COMMAND_SYNOPSIS;
    stopMovie();
    getScene()->getSceneActivation()->startMovie(getClient(), argv[1]);
    setResult("movie record started at simulation rate");
  }
  else if(!strcmp(argv[0], "pause")) {
    if(argc != 1) return COMMAND_SYNOPSIS;
    pauseMovie(record);
    if(record) setResult("movie record restarted");
    else setResult("movie record paused");
  }
  else if(!strcmp(argv[0], "stop")) {
    if(argc != 1) return COMMAND_SYNOPSIS;
    if(record) stopMovie();
    else getScene()->getSceneActivation()->stopMovie();
    setResult("end of movie record");
  }
  else return COMMAND_SYNOPSIS;
  return TCL_OK;
}

/************************************************************************/
TclScene2D *TclCamera2D::getScene(void) const {
	return (TclScene2D *)(getClient()->getScene()->getInterface());
}
/************************************************************************/
void TclCamera2D::startMovie(const char *name) {
	movieName = name;
	frameNumber = 0;
	record = true;
}
/************************************************************************/
void TclCamera2D::stopMovie(void) {
	record = false;
}
/************************************************************************/
void TclCamera2D::pauseMovie(bool b) {
	record = !b;
}

/************************************************************************/
void TclCamera2D::setSelectedObject(Data2D *data) {
	if(!data) {
		if(getClient()->getSelectedObject())
			applyObjectUnselection(getClient()->getSelectedObject());
		getClient()->selectedObject = 0;
		if(getModule()->windowExists((char *) submenuFrame)) {
			getModule()->tickle("%s.name configure -text \"no selected object\"",
				(char *) submenuFrame);
		}
	}
}
/************************************************************************/
void TclCamera2D::applyObjectUnselection(Data2D *obj) {
	if (obj!=0) {
		getClient()->unsetSelectedObject(obj);
		if(obj->getInterface()) {
			if(getModule()->windowExists((char *) submenuFrame)) {
				getModule()->tickle("%s.name configure -text \"no selected object\"",
					(char *) submenuFrame);
				TclScene2DObject *tclsobj =
					dynamic_cast<TclScene2DObject *>(obj->getInterface());
				if (tclsobj)
					/// call callback function for unselection
					tclsobj->unselectionCallback(this);
			}
		}
	}
}
/************************************************************************/
void TclCamera2D::applyObjectSelection (Data2D *obj){
	if (obj!=getClient()->getSelectedObject()) {
		applyObjectUnselection(getClient()->getSelectedObject());
		getClient()->setSelectedObject(obj);
		if(obj->getInterface()) {
			TclObject *tclobj = dynamic_cast<TclObject *>(obj->getInterface());
			TclScene2DObject *tclsobj =
				dynamic_cast<TclScene2DObject *>(obj->getInterface());
			if(getModule()->windowExists((char *) submenuFrame)) {
				getModule()->tickle("%s.name configure -text \"%s\"",
					(char *) submenuFrame, (char *) tclobj->name());
			}
			if (tclsobj)
				// call the callback function to warn the object it is being selected
				tclsobj->selectionCallback(this);
		}
	}
}

/*
 * Called when the widget's contents must be redrawn.  Basically, we
 * just call the user's render callback function.
 *
 * Note that the parameter type is ClientData so this function can be
 * passed to Tk_DoWhenIdle().
 */

void TclCamera2D::redraw(void) {
	if (!Togl_UpdatePending(widget))
		Togl_PostRedisplay(widget);
}


void TclCamera2D::bindMouseMotion(bool on){
  if (on==true)
    {
      getModule()->tickle("bind %s <Motion> \"%s%u -mouseMotion %%x %%y\"",
			  (char *) cameraFrame,
			  TclCamera2D::getObjectName(),getRef());
      getModule()->setResult("%s%u mouse motion bind to move camera or object",
			     TclCamera2D::getObjectName(),getRef());
    }
  else
    {
      getModule()->tickle("bind %s <Motion> \"\"",
			  (char *) cameraFrame);
      getModule()->setResult("%s%u mouse motion unbind",
			     TclCamera2D::getObjectName(),getRef());
    }

}


/*  cam3D [-scene scn3D%u] [-in %w] [-stereo] */
int TclCamera2D::create(TclModule *m, int argc, char **argv) {
  TclString in;
  TclScene2D *scn = 0;
  TclCamera2D *togl;
  int i,w,h;
  //bool stereo = false;

  if(argc > 10) return COMMAND_SYNOPSIS;

  w=h=-1;
  for(i = 1; i < argc; i++) {
    if(!strcmp(argv[i], "-scene")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      scn = (TclScene2D *)
	TclObjectKind::getObject(typeid(TclScene2D), argc, argv, i);
      if(!scn) return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "-in")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(!m->windowExists(argv[i])) return COMMAND_SYNOPSIS;
      in = argv[i];
    }
    else if(!strcmp(argv[i], "-width")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      w=(int) atoi(argv[i]);
    }
    else if(!strcmp(argv[i], "-height")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      h=(int) atoi(argv[i]);
    }
 
    else return COMMAND_SYNOPSIS;
  }
  if (!scn) 
	  scn= TclScene2D::getDefaultScene(m);
  try {
	  if ((w>0) && (h>0)) {
    togl = new TclCamera2D(m, in, scn,w,h);
	  } else {
    togl = new TclCamera2D(m, in, scn);
	  }
  }
  catch (Exception &) {
    m->setResult("Unable to create camera");
    return TCL_ERROR;
  }
  
  m->setResult("%s%u", TclObjectKind::getObjectName(typeid(TclCamera2D)),
	       togl->getRef());

  return TCL_OK;
}




/* '+' key pressed: move camera forward */
int TclCamera2D::keyPlusCmd(int argc, char **argv) {
  if (argc>0)  return COMMAND_SYNOPSIS;

  Vec2<double> t( 0.0, getClient()->getRadius() / 2000.0);
  t *= getClient()->getRadius() / getClient()->getScale();
//  t = getClient()->mapToScreen(t);
  getClient()->setCenter(getClient()->getCenter() - t);
  redraw();
  return TCL_OK;
}


/* '-' key pressed: move camera backward */
int TclCamera2D::keyMinusCmd(int argc, char **argv) {
  if (argc>0)  return COMMAND_SYNOPSIS;
  
  Vec2<double> t( 0.0, -getClient()->getRadius() / 2000.0);
  t *= getClient()->getRadius() / getClient()->getScale();
  //t = getClient()->mapToScreen(t);
  getClient()->setCenter(getClient()->getCenter() - t);
  redraw();

  return TCL_OK;
}


/* up pressed: turn objects down */
int TclCamera2D::keyUpCmd(int argc, char **argv) {
  if (argc>0)  return COMMAND_SYNOPSIS;

//  getClient()->setMoveMode(Camera2D::NAVIGATION_MODE);
//  getClient()->rotate(Vec3<double>(1.0, 0.0, 0.0), -0.03);
 // redraw();

  return TCL_OK;
}


/* down pressed: turn objects up */
int TclCamera2D::keyDownCmd(int argc, char **argv) {
  if (argc>0)  return COMMAND_SYNOPSIS;

 // getClient()->setMoveMode(Camera2D::NAVIGATION_MODE);
 // getClient()->rotate(Vec3<double>(1.0, 0.0, 0.0), 0.03);
  redraw();
  return TCL_OK;
}


/* left pressed: turn objects right */
int TclCamera2D::keyLeftCmd(int argc, char **argv) {
  if (argc>0)  return COMMAND_SYNOPSIS;

 // getClient()->setMoveMode(Camera2D::NAVIGATION_MODE);
  //getClient()->rotate(Vec3<double>(0.0, 1.0, 0.0), -0.03);
  redraw();
  return TCL_OK;
}


/* right pressed: turn objects left */
int TclCamera2D::keyRightCmd(int argc, char **argv) {
  if (argc>0)  return COMMAND_SYNOPSIS;

  //getClient()->setMoveMode(Camera2D::NAVIGATION_MODE);
  //getClient()->rotate(Vec3<double>(0.0, 1.0, 0.0), 0.03);
  redraw();
  return TCL_OK;
}
/* left pressed: turn objects right */
int TclCamera2D::keyCmd(int argc, char **argv) {
  if (argc!=1)  return COMMAND_SYNOPSIS;
  
  if (argv[0][0]=='c')
    getClient()->toggleCopyright();
  if (argv[0][0]=='f')
    getClient()->toggleFrameRate();

  redraw();
  return TCL_OK;
}



/*][************************************************************************/
/**********************************************************************/
/**************** rotate camera trackball-wise ************************/
/**********************************************************************/




/* Button1 pressed: begin model rotation */
int TclCamera2D::button1PressedCmd(int argc, char **argv) {
  if (argc!=2)  return COMMAND_SYNOPSIS;

  int x, y;

  if(sscanf(argv[0], "%d", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%d", &y) != 1) return COMMAND_SYNOPSIS;
  y=getClient()->getHeight()-y;
  getClient()->buttonPressed(Camera2D::LEFT_BUTTON, Camera2D::NO_MODIFIER, x, y);
  bindMouseMotion(true); 
  return TCL_OK;
}
/* Button2 pressed: begin model rotation */
int TclCamera2D::button2PressedCmd(int argc, char **argv) {
  if (argc!=2)  return COMMAND_SYNOPSIS;

  int x,y;

  if(sscanf(argv[0], "%d", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%d", &y) != 1) return COMMAND_SYNOPSIS;
  y=getClient()->getHeight()-y;
  getClient()->buttonPressed(Camera2D::MIDDLE_BUTTON, Camera2D::NO_MODIFIER, x, y);
  bindMouseMotion(true); 
  return TCL_OK;
}
/* Button2 pressed: begin model rotation */
int TclCamera2D::button3PressedCmd(int argc, char **argv) {
  if (argc!=2)  return COMMAND_SYNOPSIS;

  int x, y;

  if(sscanf(argv[0], "%d", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%d", &y) != 1) return COMMAND_SYNOPSIS;
  y=getClient()->getHeight()-y;
  getClient()->buttonPressed(Camera2D::RIGHT_BUTTON, Camera2D::NO_MODIFIER, x, y);
  bindMouseMotion(true); 
  return TCL_OK;
}
/* Shift + Button1 pressed: begin model rotation */
int TclCamera2D::shiftButton1PressedCmd(int argc, char **argv) {
  if (argc!=2)  return COMMAND_SYNOPSIS;

  int x, y;

  if(sscanf(argv[0], "%d", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%d", &y) != 1) return COMMAND_SYNOPSIS;
  y=getClient()->getHeight()-y;
  getClient()->buttonPressed(Camera2D::LEFT_BUTTON, Camera2D::SHIFT_KEY, x, y);
  bindMouseMotion(true); 
  return TCL_OK;
}
/* Shift + Button2 pressed: begin model rotation */
int TclCamera2D::shiftButton2PressedCmd(int argc, char **argv) {
  if (argc!=2)  return COMMAND_SYNOPSIS;

  int x, y;

  if(sscanf(argv[0], "%d", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%d", &y) != 1) return COMMAND_SYNOPSIS;
  y=getClient()->getHeight()-y;
  getClient()->buttonPressed(Camera2D::MIDDLE_BUTTON, Camera2D::SHIFT_KEY, x, y);
  bindMouseMotion(true); 
  return TCL_OK;
}
/* Shift + Button3 pressed: begin model rotation */
int TclCamera2D::shiftButton3PressedCmd(int argc, char **argv) {
  if (argc!=2)  return COMMAND_SYNOPSIS;

  int x, y;

  if(sscanf(argv[0], "%d", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%d", &y) != 1) return COMMAND_SYNOPSIS;
  y=getClient()->getHeight()-y;
  getClient()->buttonPressed(Camera2D::RIGHT_BUTTON, Camera2D::SHIFT_KEY, x, y);
  bindMouseMotion(true); 
  return TCL_OK;
}
/* Ctrl + Button1 pressed: begin model rotation */
int TclCamera2D::ctrlButton1PressedCmd(int argc, char **argv) {
  if (argc!=2)  return COMMAND_SYNOPSIS;

  int x,y;

  if(sscanf(argv[0], "%d", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%d", &y) != 1) return COMMAND_SYNOPSIS;
  y=getClient()->getHeight()-y;
  getClient()->buttonPressed(Camera2D::LEFT_BUTTON, Camera2D::CTRL_KEY, x, y);
  bindMouseMotion(true); 
  return TCL_OK;
}
/* Ctrl + Button2 pressed: begin model rotation */
int TclCamera2D::ctrlButton2PressedCmd(int argc, char **argv) {
  if (argc!=2)  return COMMAND_SYNOPSIS;

  int x,y;

  if(sscanf(argv[0], "%d", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%d", &y) != 1) return COMMAND_SYNOPSIS;
  y=getClient()->getHeight()-y;
  getClient()->buttonPressed(Camera2D::MIDDLE_BUTTON, Camera2D::CTRL_KEY, x, y);
  bindMouseMotion(true); 
  return TCL_OK;
}
/* Ctrl + Button3 pressed: begin model rotation */
int TclCamera2D::ctrlButton3PressedCmd(int argc, char **argv) {
  if (argc!=2)  return COMMAND_SYNOPSIS;

  int x,y;

  if(sscanf(argv[0], "%d", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%d", &y) != 1) return COMMAND_SYNOPSIS;
  y=getClient()->getHeight()-y;
  getClient()->buttonPressed(Camera2D::RIGHT_BUTTON, Camera2D::CTRL_KEY, x, y);
  bindMouseMotion(true); 
  return TCL_OK;
}



int TclCamera2D::button1ReleasedCmd(int argc, char **argv) {
  if (argc!=2)  return COMMAND_SYNOPSIS;
  int x,y;


  if(sscanf(argv[0], "%d", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%d", &y) != 1) return COMMAND_SYNOPSIS;
  y=getClient()->getHeight()-y;
  getClient()->buttonReleased(Camera2D::LEFT_BUTTON, x, y);
  if (alwaysTrackMouseMotion==false)
	  bindMouseMotion(false);
  return TCL_OK;
}
int TclCamera2D::button2ReleasedCmd(int argc, char **argv) {
  if (argc!=2)  return COMMAND_SYNOPSIS;
  int x,y;

  if(sscanf(argv[0], "%d", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%d", &y) != 1) return COMMAND_SYNOPSIS;
  y=getClient()->getHeight()-y;
  getClient()->buttonReleased(Camera2D::MIDDLE_BUTTON, x, y);
 if (alwaysTrackMouseMotion==false)
	  bindMouseMotion(false);
  return TCL_OK;
}
int TclCamera2D::button3ReleasedCmd(int argc, char **argv) {
  if (argc!=2)  return COMMAND_SYNOPSIS;
  int x,y;

  if(sscanf(argv[0], "%d", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%d", &y) != 1) return COMMAND_SYNOPSIS;
  y=getClient()->getHeight()-y;
  getClient()->buttonReleased(Camera2D::RIGHT_BUTTON, x, y);
 if (alwaysTrackMouseMotion==false)
	  bindMouseMotion(false);
  return TCL_OK;
}

void TclCamera2D::dumpToFile(const TclString& filename) {
	dumpToPPM(filename);

}
void TclCamera2D::dumpToPPM(const TclString& filename) {
  unsigned char *image;

   /* Allocate our buffer for the image */
   if ((image = new unsigned char[3*getClient()->getWidth()*getClient()->getHeight()])==0) {
      CERR << "Failed to allocate memory for image" << ENDL;
      throw (Exception("void TclCamera2D::dumpToFile(const TclString& filename)",
		  "no more memory"));
   }
   makeCurrent();
   glFlush();
   glPixelStorei(GL_PACK_ALIGNMENT,1);
 /* Copy the image into our buffer */
   glReadBuffer(GL_FRONT);
   glReadPixels(0,0,getClient()->getWidth(),getClient()->getHeight(),GL_RGB,GL_UNSIGNED_BYTE,image);

   FILE *fptr = fopen(filename,"wb");
   fprintf(fptr,"P6\n%d %d\n255\n",getClient()->getWidth(),
	   getClient()->getHeight()); 
   int i,j;
   for (j=getClient()->getHeight()-1;j>=0;j--) {
	   for (i=0;i<(int)getClient()->getWidth();i++) {
		   fputc(image[3*j*getClient()->getWidth()+3*i+0],fptr);
		   fputc(image[3*j*getClient()->getWidth()+3*i+1],fptr);
		   fputc(image[3*j*getClient()->getWidth()+3*i+2],fptr);
	   }
   }
   fclose(fptr);
}


/*][************************************************************************/

int TclCamera2D::list(TclModule *m, int argc, char **argv) {
  TclScene2D *scn;
  TclString list, one;
  TclObjectKind *kind = TclObjectKind::findByKind(typeid(TclCamera2D));
    
  if(argc != 1 && argc != 2) return COMMAND_SYNOPSIS;

  // all cameras
  if(argc == 1) {
    ObjectsList<TclObject>::iterator it(kind->firstObject());
    for(; it!=kind->lastObject(); it++) {
      one.sprintf("%s%u ", TclObjectKind::getObjectName(typeid(TclCamera2D)),
		  ((TclCamera2D *) (*it))->getRef()); 
      list += one;
    }
    m->setResult(list);
  }

  // one scene cameras
  else {
    scn = 
      (TclScene2D *) TclObjectKind::getObject(typeid(TclScene2D), argc, argv);
    if(scn) {
      for(ObjectsList<Camera2D>::iterator it = scn->getClient()->firstCamera();
	  it!= scn->getClient()->lastCamera(); it++) {
	one.sprintf("%s%i ", TclObjectKind::getObjectName(typeid(TclCamera2D)), ((TclCamera2D *)(*it))->getRef());
	list += one;
      }
      m->setResult(list);
    }
  }

  return TCL_OK;
}

/*][************************************************************************/
int TclCamera2D::redrawCmd(int argc, char **argv) {
  if (argc!=0)  return COMMAND_SYNOPSIS;
  redraw();
  getModule()->setResult("%s%u redisplayed", getObjectName(), 
			 getRef());
  return TCL_OK;
}

/*][************************************************************************/
int TclCamera2D::selectCmd(int argc, char **argv) {
  TclObjectKind *kind;
  bool found = false;
 
  if(argc > 1) return COMMAND_SYNOPSIS;
  
  if(argc == 1) {
	  if(!strcmp(argv[0], "none")) {
		  applyObjectSelection(0);
		  getClient()->setMotionMode(Camera2D::MOVE_CAMERA);
		  setResult("no selected object");
		  return TCL_OK;
	  }
	  else {
		  ObjectsList<Data2D>::iterator it = getScene()->getClient()->firstData();
		  for(;  it!=getScene()->getClient()->lastData() && !found; it++) {
			  Interface *interf = (*it)->getInterface();
			  if(interf &&
				  dynamic_cast<TclObject *>(interf)->name() == (const char *)argv[0]) {
				  applyObjectSelection(*it);
				  found = true;
			  }
		  }
		  
		  if(!found) {
			  kind = TclObjectKind::findByName(argv[0]);
			  if(!kind) {
				  setResult("object %s does not exists", argv[0]);
				  return TCL_ERROR;
			  }
			  setResult("unknown object kind %s", argv[0]);
			  return TCL_ERROR;
		  }
		  
		  else {
			  setResult(argv[0]);
			  return TCL_OK;
		  }
	  }
  }
  
  TclString output("none");
  if(getClient()->getSelectedObject()) {
	  Interface *interf = getClient()->getSelectedObject()->getInterface();
	  if(interf)
		  output = dynamic_cast<TclObject *>(interf)->name();
   }
  getModule()->setResult(output);
  return TCL_OK;  
}

/************************************************************************/
int TclCamera2D::destroyCmd(int argc, char **argv) {
	if(argc > 0) return COMMAND_SYNOPSIS;
	TclModule *m = getModule();
	TclString n=name();	

	if(getModule()->windowExists((char *) getParent()))
		getModule()->tickle("destroy %s", (char *) getParent());
	else
		getModule()->tickle("destroy %s", (char *) getCameraFrame());

	m->setResult("%s does not exist any longer",
		(char *) n);
	return TCL_OK;
}
/************************************************************************/
int TclCamera2D::transformationMatrixCmd(int argc, char **argv) {
	if((argc!=0) && (argc!=16) && (argc!=1)) return COMMAND_SYNOPSIS;
	if(argc==0){
		Matrix4x4<double> mat = getClient()->getMatTransform();
		TclMatrix4D *m= new TclMatrix4D(getModule(),mat);
		setResult(m->TclObject::name());
		return TCL_OK;
	} else if (argc==16){
		int i;
		double coeff[16];
		for(i=0;i<16;i++)
			if(sscanf(argv[i],"%lf",&coeff[i])!=1){
				return COMMAND_SYNOPSIS;
			}
			Matrix4x4<double> mat(coeff[0],coeff[1],coeff[2],coeff[3],coeff[4],coeff[5],coeff[6],coeff[7],coeff[8],coeff[9],coeff[10],coeff[11],coeff[12],coeff[13],coeff[14],coeff[15]);
			getClient()->setMatTransform(mat);
			redraw();
			return TCL_OK;
	} else {
		TclMatrix4D *m= (TclMatrix4D *)  TclMatrix4D::getObject(argc, argv, 0);
		if (m==0) {
			setResult("%s must be a 4D matrix",argv[0]);
			return TCL_ERROR;
		}
		else {
			getClient()->setMatTransform(*m);
			redraw();
			return TCL_OK;
		}
	}
}
/************************************************************************/
int TclCamera2D::viewPointCmd(int argc, char **argv) {
	if((argc!=0) && (argc!=21)) return COMMAND_SYNOPSIS;
	if(argc==0){
		Vec2<double> center = getClient()->getCenter();
		Vec2<double> refCenter = getClient()->getRefCenter();
		double scale = getClient()->getScale();
		Matrix4x4<double> mat = getClient()->getMatTransform();
		setResult("%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f",center.getX(),center.getY(),refCenter.getX(),refCenter.getY(),scale,mat[0][0],mat[0][1],mat[0][2],mat[0][3],mat[1][0],mat[1][1],mat[1][2],mat[1][3],mat[2][0],mat[2][1],mat[2][2],mat[2][3],mat[3][0],mat[3][1],mat[3][2],mat[3][3]);
		return TCL_OK;
	}
	else {
		int i;
		double c[2];
		for(i=0;i<2;i++)
			if(sscanf(argv[i],"%lf",&c[i])!=1){
				return COMMAND_SYNOPSIS;
			}
			Vec2<double> center(c[0],c[1]);
			getClient()->setCenter(center);
			for(i=0;i<2;i++)
				if(sscanf(argv[i+2],"%lf",&c[i])!=1){
					return COMMAND_SYNOPSIS;
				}
				center.setValue(c[0],c[1]);
				getClient()->refCenter=center;
				double scale;
				if(sscanf(argv[4],"%lf",&scale)!=1)
					return COMMAND_SYNOPSIS;
				getClient()->setScale(scale);
				double coeff[16];
				for(i=0;i<16;i++)
					if(sscanf(argv[i+5],"%lf",&coeff[i])!=1){
						return COMMAND_SYNOPSIS;
					}
					Matrix4x4<double> mat(coeff[0],coeff[1],coeff[2],coeff[3],coeff[4],coeff[5],coeff[6],coeff[7],coeff[8],coeff[9],coeff[10],coeff[11],coeff[12],coeff[13],coeff[14],coeff[15]);
					getClient()->setMatTransform(mat);
					redraw();
					return TCL_OK;
	}
}
/************************************************************************/
int TclCamera2D::centerCmd(int argc, char **argv) {
	if((argc!=0) && (argc!=2)) return COMMAND_SYNOPSIS;
	if(argc==0){
		Vec2<double> center = getClient()->getCenter();
		setResult("%f %f ",center.getX(),center.getY());
		return TCL_OK;
	}
	int i;
	double c[2];
	for(i=0;i<2;i++)
		if(sscanf(argv[i],"%lf",&c[i])!=1){
			return COMMAND_SYNOPSIS;
		}
		Vec2<double> center(c[0],c[1]);
		getClient()->setCenter(center);
		redraw();
		return TCL_OK;
}

/************************************************************************/
int TclCamera2D::refCenterCmd(int argc, char **argv) {
	if((argc!=0) && (argc!=2)) return COMMAND_SYNOPSIS;
	if(argc==0){
		Vec2<double> refCenter = getClient()->getRefCenter();
		setResult("%f %f ",refCenter.getX(),refCenter.getY());
		return TCL_OK;
	}
	int i;
	double c[2];
	for(i=0;i<2;i++)
		if(sscanf(argv[i],"%lf",&c[i])!=1){
			return COMMAND_SYNOPSIS;
		}
		Vec2<double> _refCenter(c[0],c[1]);
		getClient()->refCenter=_refCenter;
		redraw();
		return TCL_OK;
		
}
/************************************************************************/
int TclCamera2D::scaleFactorCmd(int argc, char **argv) {
	if((argc!=0) && (argc!=1)) return COMMAND_SYNOPSIS;
	if(argc==0){
		double scale = getClient()->getScale();
		setResult("%f",scale);
		return TCL_OK;
	}
	double scale;
	if(sscanf(argv[0],"%lf",&scale)!=1)
		return COMMAND_SYNOPSIS;
	getClient()->setScale(scale);
	redraw();
	return TCL_OK;
}



/*************************************************************************/
int TclCamera2D::bgColorCmd(int argc, char **argv) {
	if ((argc!=0)&& (argc!=3))  return COMMAND_SYNOPSIS;
	if (argc==0)
    {
		float r,g,b;
		getClient()->getBackgroundColor(r,g,b);
		getModule()->setResult("%f %f %f",r,g,b);
		return TCL_OK;
    }
	else
    {
		float r,g,b;
		if(sscanf(argv[0], "%f", &r) == -1 || r < 0.0 || r > 1.0)
			return COMMAND_SYNOPSIS;
		if(sscanf(argv[1], "%f", &g) == -1 || g < 0.0 || g > 1.0)
			return COMMAND_SYNOPSIS;
		if(sscanf(argv[2], "%f", &b) == -1 || b < 0.0 || b > 1.0) 
			return COMMAND_SYNOPSIS;
		getClient()->setBackgroundColor(r,g,b);
		redraw();
		getModule()->setResult("%f %f %f",r,g,b);
		return TCL_OK;
	}
}
/*************************************************************************/
int TclCamera2D::sceneCmd(int argc, char **argv) {
	if (argc!=0) return COMMAND_SYNOPSIS;
	getModule()->setResult("%s%u",
		TclObjectKind::getObjectName(typeid(TclScene2D)),
		getScene()->getRef());
	return TCL_OK;
}
/*************************************************************************/

int TclCamera2D::widthCmd(int argc, char **argv) {
	if ((argc!=0)&&(argc!=1)) return COMMAND_SYNOPSIS;
	if (argc==0)
    {
		getModule()->setResult("%u", getClient()->getWidth());
		return TCL_OK;
    }
	else
    {
		unsigned int u;
		if(sscanf(argv[0], "%u", &u) != 1) return COMMAND_SYNOPSIS;
		resize(u, getClient()->getHeight());
		getModule()->setResult("%u",getClient()->getWidth()); 
		return TCL_OK;
    }
}
/*************************************************************************/
void TclCamera2D::resize(unsigned int w, unsigned int h) {
	getClient()->resize(w, h);
	Tk_GeometryRequest(Togl_TkWin(widget), getClient()->width, getClient()->height);  
}

int TclCamera2D::heightCmd(int argc, char **argv) {
	if ((argc!=0)&&(argc!=1)) return COMMAND_SYNOPSIS;
	if (argc==0)
    {
		getModule()->setResult("%u", getClient()->getHeight());
		return TCL_OK;
    }
	else
    {
		unsigned int u;
		if(sscanf(argv[0], "%u", &u) != 1)  return COMMAND_SYNOPSIS;
		resize(getClient()->getWidth(),u);
		getModule()->setResult("%u",getClient()->getHeight()); 
		return TCL_OK;
    }
}
/*************************************************************************/
/* indicates that the scene should be moved instead of the selected object */
int TclCamera2D::motionModeCmd(int argc, char **argv) {
	if(argc > 1) return COMMAND_SYNOPSIS;
	if(argc == 1) {
		if(!strcmp(argv[0], "camera")) {
			getClient()->setMotionMode(Camera2D::MOVE_CAMERA);
		}
		else if(!strcmp(argv[0], "cameraObject")) {
			getClient()->setMotionMode(Camera2D::MOVE_CAMERA_OBJECT);
		}
		else if(!strcmp(argv[0], "object")) {
			getClient()->setMotionMode(Camera2D::MOVE_SELECTED_OBJECT);
		}
		else return COMMAND_SYNOPSIS;
	}
	
	switch(getClient()->getMotionMode()) {
	case Camera2D::MOVE_CAMERA:
		setResult("camera");      
		break;
	case Camera2D::MOVE_CAMERA_OBJECT:
		setResult("cameraObject");
		break;
	case Camera2D::MOVE_SELECTED_OBJECT:
		setResult("object");
		break;
	}    
	
	return TCL_OK;
}


/*************************************************************************/
/* mouse motion: move model */
int TclCamera2D::mouseMotionCmd(int argc, char **argv) {
	if (argc!=2)  return COMMAND_SYNOPSIS;
	int x,y;
	
	if(sscanf(argv[0], "%d", &x) != 1) return COMMAND_SYNOPSIS;
	if(sscanf(argv[1], "%d", &y) != 1) return COMMAND_SYNOPSIS;
	y=getClient()->getHeight()-y;
	getClient()->mouseMoved(x, y);
	
	return TCL_OK;
}
/*************************************************************************/
/**********************************************************************/
/**************** navigate with camera ********************************/
/**********************************************************************/


//*************************************************************************/
int TclCamera2D::focusCmd(int argc, char **argv) {
	if (argc!=0)  return COMMAND_SYNOPSIS;
	getClient()->doFocus();
	redraw();
	getModule()->setResult("%s%u focused on scene", getObjectName(), 
		getRef());
	return TCL_OK;
}

/*************************************************************************/
int TclCamera2D::widgetCmd(int argc, char **argv) {
	if (argc>=1) {
		if (!strcmp(argv[0],"menu")) {
			if (argc==1) {
				setResult(menuFrame);
				return TCL_OK;
			}
			else {
				menuFrame=argv[1];
				return TCL_OK;
			}
		}
		else if (!strcmp(argv[0],"submenu")) {
			if (argc==1) {
				setResult(submenuFrame);
				return TCL_OK;
			}
			else {
				submenuFrame=argv[1];
				return TCL_OK;
			}
		}
		else if (!strcmp(argv[0],"parent")) {
			if (argc==1) {
				setResult(parentFrame);
				return TCL_OK;
			}
			else 
				return TCL_ERROR;
		}
		else if (!strcmp(argv[0],"button")) {
			if (argc==1) {
				setResult(buttonFrame);
				return TCL_OK;
			}
			else {
				buttonFrame=argv[1];
				return TCL_OK;
			}
		}
		else {
			setResult("%s must be either menu, submenu or button",argv[0]);
			return TCL_ERROR;
		}
	}
	else
    {
		getModule()->setResult("%s",(char *) cameraFrame);
		return TCL_OK;
    }
}
/*************************************************************************/
int TclCamera2D::menuCmd(int argc, char **argv) {
	if (argc>1)  return COMMAND_SYNOPSIS;
	else
    {
		if (argc==1) {
			submenuFrame=argv[0];
			if (getClient()->getSelectedObject()) {
				Data2D *obj=getClient()->getSelectedObject();
				if(obj->getInterface()) {
					TclObject *tclobj = dynamic_cast<TclObject *>(obj->getInterface());
					TclScene2DObject *tclsobj =
						dynamic_cast<TclScene2DObject *>(obj->getInterface());
					if(getModule()->windowExists((char *) submenuFrame)) {
						getModule()->tickle("%s.name configure -text \"%s\"",
							(char *) submenuFrame, (char *) tclobj->name());
					}
					if (tclsobj)
						// call the callback function to warn the object it is being selected
						tclsobj->selectionCallback(this);
				}

			}
		}
		getModule()->setResult("%s",(char *) submenuFrame);
		return TCL_OK;
    }
}

/*************************************************************************/
int TclCamera2D::objectsCmd(int argc, char **argv) {
	if (argc!=0)  return COMMAND_SYNOPSIS;
	TclString output;
	ObjectsList<Data2D>::iterator it = getScene()->getClient()->firstData();
	for(;   it!= getScene()->getClient()->lastData(); it++)
    {
		Interface *interf = (*it)->getInterface();
		if(interf) {
			output+= (dynamic_cast<TclObject *>(interf))->name();
			output+= SPACE_STRING;
		}
    }
	getModule()->setResult(output);
	return TCL_OK;
}
int TclCamera2D::alwaysTrackMouseMotionCmd(int argc, char **argv) {
	if(argc > 1) return COMMAND_SYNOPSIS;
	if(argc == 1) {
		if(!TclModule::boolValue(argv[0], alwaysTrackMouseMotion)) return COMMAND_SYNOPSIS;
		redraw();
	}
	setResult(TclModule::boolString(alwaysTrackMouseMotion));
	return TCL_OK;
}
int TclCamera2D::dumpToPPMCmd(int argc, char **argv) {
	if(argc != 1) return COMMAND_SYNOPSIS;
	dumpToFile(argv[0]);
	setResult("file dumped as %s",argv[0]);
	return TCL_OK;
}
int TclCamera2D::stencilBufferCmd(int argc, char **argv) {
	bool usestencil;
	
	if(argc > 1) return COMMAND_SYNOPSIS;
	if(argc == 1) {
		if(!TclModule::boolValue(argv[0], usestencil)) return COMMAND_SYNOPSIS;
		getClient()->setStencilBuffer(usestencil);
		if(usestencil)  getClient()->buildStencilBufferDisplayList();
		redraw();
	}
	setResult(TclModule::boolString(getClient()->stencilBuffer));
	return TCL_OK;
}
int TclCamera2D::resizeCmd(int argc, char **argv) {
  int w, h;
  if(argc != 2) return COMMAND_SYNOPSIS;
  
  if(sscanf(argv[0], "%i", &w) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%i", &h) != 1) return COMMAND_SYNOPSIS;
  if(w < 0 || h < 0) return COMMAND_SYNOPSIS;
  makeCurrent();
  resize((unsigned int) w, (unsigned int) h);
  setResult("%u %u", getClient()->getWidth(), getClient()->getHeight());

  return TCL_OK;
}

void TclCamera2D::createCommands(TclModule *m) {
   TclObjectKind *kind;
   
   Tcl_CreateObjCommand(m->getInterp(), "createCamera2D", createCamera2D, NULL, NULL);
   Tcl_CreateObjCommand(m->getInterp(), "displayCamera2D", displayCamera2D, NULL, NULL);
   Tcl_CreateObjCommand(m->getInterp(), "resizeCamera2D", resizeCamera2D, NULL, NULL);
   Tcl_CreateObjCommand(m->getInterp(), "destroyCamera2D", destroyCamera2D, NULL, NULL);



  kind = new Scene2DItem(m, typeid(TclCamera2D), "cam2D","2D Camera",
			 &TclCamera2D::create, &TclCamera2D::list,
			 TclScene2D::NONE | TclScene2D::SCENE, "[-scene scn2D%u] [-in widget=%s]");
  m->newCommand("cam2D%u", kind->getOptions(), "2D camera management command");

  kind->newOption("-destroy", (TclObject::OPTION) &TclCamera2D::destroyCmd,
		  0, "Delete 2D camera");
  kind->newOption("-redraw", (TclObject::OPTION) &TclCamera2D::redrawCmd,
		  0, "Cause 2D camera to redraw itself");
  kind->newOption("-focus", (TclObject::OPTION) &TclCamera2D::focusCmd,
		  0, "Reset 2D camera focus to view all scene");
   kind->newOption("-alwaysTrackMouseMotion", (TclObject::OPTION) &TclCamera2D::alwaysTrackMouseMotionCmd,
		  "[%b]", "Get/Set the boolean that indicates if the mouse motion is always tracked when the mouse enters the camera widget. If false the mouse is tracked only when a mouse button is pressed");
    kind->newOption("-bgColor", (TclObject::OPTION) &TclCamera2D::bgColorCmd,
		  "[%f %f %f]", "Get/Set RGB color of background (each color is between 0.0 and 1.0)");
  kind->newOption("-resize", (TclObject::OPTION) &TclCamera2D::resizeCmd, "%u %u",
		  "Associated widget is asked to be resized given width and height");
  kind->newOption("-select", (TclObject::OPTION) &TclCamera2D::selectCmd,
		  0, "Get/Set the selected object");
  kind->newOption("-widget", (TclObject::OPTION) &TclCamera2D::widgetCmd,
		  0, "Return associated TCL widget name");
  kind->newOption("-menu", (TclObject::OPTION) &TclCamera2D::menuCmd,
		  "[%s]", "set or get the name of the Tk widget where the menu associated with the selected object is updated");
  kind->newOption("-scene", (TclObject::OPTION) &TclCamera2D::sceneCmd,
		  0, "Return scene shown");
  kind->newOption("-width", (TclObject::OPTION) &TclCamera2D::widthCmd,
		  "[%u [-height %u]]", "Get width or set width and possibly height of 2D camera");
  kind->newOption("-height", (TclObject::OPTION) &TclCamera2D::heightCmd,
		  "[%u [-width %u]]", "Get height or set height and possibly width of 2D camera");
//  kind->newOption("-scale", (TclObject::OPTION) &TclCamera2D::scaleCmd,
//		  "[%f | * %f | / %f]", "Get or set 2D camera scale (can set either an absolute scale or a relative one)");
  kind->newOption("-center", (TclObject::OPTION) &TclCamera2D::centerCmd,
		  "[%f %f]","Get/Set the center in world coordinate where the camera is looking for(important when scaling camera)");
//  kind->newOption("-cameraCenter", (TclObject::OPTION) &TclCamera2D::cameraCenterCmd,
//		  "[%f %f]","Get/Set the center in canvas coordinate where the camera is looking for(important when scaling camera)");

   kind->newOption("-button1Released", (TclObject::OPTION) &TclCamera2D::button1ReleasedCmd,
		  "%x %y", "Left button released callback");
  kind->newOption("-button2Released", (TclObject::OPTION) &TclCamera2D::button2ReleasedCmd,
		  "%x %y", "Middle button released callback");
  kind->newOption("-button3Released", (TclObject::OPTION) &TclCamera2D::button3ReleasedCmd,
		  "%x %y", "Right button released callback");
  kind->newOption("-button1Pressed", (TclObject::OPTION) &TclCamera2D::button1PressedCmd,
		  "%x %y", "Left button pressed callback");
  kind->newOption("-button2Pressed", (TclObject::OPTION) &TclCamera2D::button2PressedCmd,
		  "%x %y", "Middle button pressed callback");
  kind->newOption("-button3Pressed", (TclObject::OPTION) &TclCamera2D::button3PressedCmd,
		  "%x %y", "Right button pressed callback");

kind->newOption("-ctrlButton1Pressed", (TclObject::OPTION) &TclCamera2D::ctrlButton1PressedCmd,
		  "%x %y", "Ctrl-left button pressed callback");
  kind->newOption("-ctrlButton2Pressed", (TclObject::OPTION) &TclCamera2D::ctrlButton2PressedCmd,
		  "%x %y", "Ctrl-middle button pressed callback");
  kind->newOption("-ctrlButton3Pressed", (TclObject::OPTION) &TclCamera2D::ctrlButton3PressedCmd,
		  "%x %y", "Ctrl-right button pressed callback");
  kind->newOption("-shiftButton1Pressed", (TclObject::OPTION) &TclCamera2D::shiftButton1PressedCmd,
		  "%x %y", "Shift-left button pressed callback");
  kind->newOption("-shiftButton2Pressed", (TclObject::OPTION) &TclCamera2D::shiftButton2PressedCmd,
		  "%x %y", "Shift-left middle pressed callback");
  kind->newOption("-shiftButton3Pressed", (TclObject::OPTION) &TclCamera2D::shiftButton3PressedCmd,
		  "%x %y", "Shift-left right pressed callback");
   kind->newOption("-key",(TclObject::OPTION) &TclCamera2D::keyCmd,0,
		  "Handles Any other key pressed (c,f)"); 

#ifdef OLD_CODE
  kind->newOption("-translate", (TclObject::OPTION) &TclCamera2D::translateCmd,
		  "[%f %f]", "Get/Set camera center translation");
  kind->newOption("-ctrlButton1Released", (TclObject::OPTION) &TclCamera2D::ctrlButton1ReleasedCmd,
		  "%x %y", "Ctrl-left button released callback");
  kind->newOption("-ctrlButton2Released", (TclObject::OPTION) &TclCamera2D::ctrlButton2ReleasedCmd,
		  "%x %y", "Ctrl-middle button released callback");
  kind->newOption("-ctrlButton3Released", (TclObject::OPTION) &TclCamera2D::ctrlButton3ReleasedCmd,
		  "%x %y", "Ctrl-right button released callback");
  kind->newOption("-shiftButton1Released", (TclObject::OPTION) &TclCamera2D::shiftButton1ReleasedCmd,
		  "%x %y", "Shift-left button released callback");
  kind->newOption("-shiftButton2Released", (TclObject::OPTION) &TclCamera2D::shiftButton2ReleasedCmd,
		  "%x %y", "Shift-middle button released callback");
  kind->newOption("-shiftButton3Released", (TclObject::OPTION) &TclCamera2D::shiftButton3ReleasedCmd,
		  "%x %y", "Shift-right button released callback");

#endif
  kind->newOption("-mouseMotion", (TclObject::OPTION) &TclCamera2D::mouseMotionCmd,
		  "%x %y", "Mouse moved callback");
    kind->newOption("-motionMode",(TclObject::OPTION) &TclCamera2D::motionModeCmd,"[object|camera]",
		  "get/set the motion mode either to move the camera or to move the selected object");

  kind->newOption("-keyUp", (TclObject::OPTION) &TclCamera2D::keyUpCmd);
  kind->newOption("-keyDown", (TclObject::OPTION) &TclCamera2D::keyDownCmd);
  kind->newOption("-keyRight", (TclObject::OPTION) &TclCamera2D::keyRightCmd);
  kind->newOption("-keyLeft", (TclObject::OPTION) &TclCamera2D::keyLeftCmd);
  kind->newOption("-record", (TclObject::OPTION) &TclCamera2D::recordCmd,
		  "<<frame [%s] | simulation [%s] > file=%f | pause | stop>",
		  "Start/stop camera movie dumps in file named file0000.ppm, file0001.ppm...It is possible to record an other widget by giving its name %s");
  kind->newOption("-dumpToPPM", (TclObject::OPTION) &TclCamera2D::dumpToPPMCmd,
		  "<filename=%s>",
		  "Save the camera widget to a ppm file");
}

}

