#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#pragma implementation "graphics2D/Scalable.h"
#pragma implementation "graphics2D/ScaleCamera.h"
#pragma implementation "graphics2D/TclScale.h"

#include <graphics2D/TclScale.h>
#include <graphics2D/TclScene2D.h>

namespace yav {

AUTOLINK_STATIC_VARIABLES(Scalable);


void TclScale::init(const std::type_info& type, TclString menu)
{
  getModule()->tickle("bind %s <Enter> \"focus %s\"",
		      (char *) TCLwindow, (char *) TCLwindow);
  getModule()->tickle("%s %s%u",(char *) menu,
		      TclObjectKind::getObjectName(type), getRef());
  getModule()->tickle("%s get", (char *) TCLwindow);
  sscanf(getModule()->getInterp()->result, "%lf", &value);
  getModule()->tickle("%s cget -from", (char *) TCLwindow);
  sscanf(getModule()->getInterp()->result, "%lf", &min);
  getModule()->tickle("%s cget -to", (char *) TCLwindow);
  sscanf(getModule()->getInterp()->result, "%lf", &max);
  maxSteps=100;
}

TclScale::TclScale(const std::type_info& type,
		   const TclString& scaleName,
		   TclModule *module,TclString menu)
  : TclSon(module,scaleName), TclObject(type, module), ScaleCamera(),
   dieWhenEmpty(true),firstupdate(true),   activationDir(1)   {
  init(type,menu);
}
TclScale::TclScale(const std::type_info& type,
		   const TclString& scaleName,
		   TclModule *module, std::list<Scalable *>& list,
		   TclString menu)
  : TclSon(module, scaleName), TclObject(type, module), ScaleCamera(list),
     dieWhenEmpty(true),firstupdate(true),   activationDir(1)  {
  init(type,menu);
}
TclScale::TclScale(const std::type_info& type,
		   const TclString& scaleName,
		   TclModule *module, Scalable * elt,TclString menu)
  : TclSon(module, scaleName), TclObject(type, module), ScaleCamera(elt),
     dieWhenEmpty(true),firstupdate(true),   activationDir(1)  {
  init(type,menu);
}

TclScale::~TclScale() {
}

void TclScale::dying(const Scalable *elt) {
   if(elements.size() == 0 && dieWhenEmpty) delete this;
}

void TclScale::command(double val) {
  if(firstupdate) { 
    firstupdate = false;
    DPList<Scalable>::iterator it;
    for(it = elements.begin(); it != elements.end(); it++)
      (*it)->scaleTo(this, val);
    update(val);
  }
  else {
    DPList<Scalable>::const_iterator it;
    const DPList<Scalable>& lst = getObjects();
    for(it = lst.begin(); it != lst.end(); it++) {
      (*it)->scaleValue(this, val);
      update(val);
    }
  }
}

// add a new viewer to scale object
void TclScale::operator+=(Scalable *sa)
  throw(DPList<Scalable>::NullPointer, DPList<Scalable>::AlreadyInList) {

  if(elements.empty()) reset(sa,EMPTY_STRING);
  else sa->scaleValue(this, get());
  DPOwner<Scalable>::operator+=(sa);
}



// remove a viewer from a scale object
void TclScale::operator-=(Scalable *sa)
  throw(DPList<Scalable>::NotFound) {

  DPOwner<Scalable>::operator-=(sa);
}

void TclScale::configureToMax(void) {
  min = maxTypeValue<float>(0);
  max = minTypeValue<float>(0);

  DPList<Scalable>::iterator it;
  for(it = elements.begin(); it != elements.end(); it++) {
    if((*it)->getMin(this) < min) min = (*it)->getMin(this);
    if((*it)->getMax(this) > max) max = (*it)->getMax(this);
  }

  if(max < min) {
    min = 0;
    max = 255;
  }

  tickle("%s configure -from %f -to %f", (char *) TCLwindow, min, max);
}


void TclScale::setLabel(const TclString& lab) {
  getModule()->tickle("%s configure -label %s", (char *) TCLwindow, (char *) lab);
}


void TclScale::setBounds(double m, double M) {
  if(M < m) {
    double t = M;
    M = m;
    m = t;
  }
  min = m;
  max = M;
  tickle("%s configure -from %f -to %f", (char *) TCLwindow, min, max);
}

void TclScale::set(double val) {
  value = val;
  getModule()->tickle("%s set %f", (char *) TCLwindow, val);
  getModule()->tickle("event generate %s <Motion>",(char *)TCLwindow);
}

void TclScale::to(double val) {
  value = val;
  firstupdate = true;
  getModule()->tickle("%s set %f",(char *) TCLwindow, val);
  getModule()->tickle("event generate %s <Motion>",(char *)TCLwindow);

}


double TclScale::get(void) const {
  double val;
  getModule()->tickle("%s get", (char *) TCLwindow);
  sscanf(getModule()->getInterp()->result, "%lf", &val);
  return val;
}
  

void TclScale::on(void) {
  value = get();
  Activable::on();
}

void TclScale::checkValue(void) {
  if ((value<min) || (value>max)) {
    set((min+max)/2);
    command((min+max)/2);
  }
}
void TclScale::activation(void) {
  //  if(activationDir > 0) {
  value += (max-min)/(double)maxSteps;
  if(value > max) {
    //       activationDir = -activationDir;
    //       value -= 2.0*(max-min)/(double)maxSteps;
    value = min;
  }
  //  }
  //   else {
  //     value -= (max-min)/(double)maxSteps;
  //     if(value < min) {
  //       activationDir = -activationDir;
  //       value += 2*(max-min)/(double)maxSteps;
  //     }    
  //   }
  getModule()->tickle("%s set %f", (char *) TCLwindow, value);
}



int TclScale::destroyCmd(int argc, char **argv) {
  TclModule *m;
  unsigned int ref;

  if(argc != 0) return COMMAND_SYNOPSIS;
  m = getModule();
  ref = getRef();
  TclString n=getObjectName();
  delete this;
  m->setResult("%s%u does not exist any longer", (char *)n, ref);
  return TCL_OK;
}

int TclScale::widgetCmd(int argc, char **argv) {
  if(argc != 0) return COMMAND_SYNOPSIS;
  setResult(windowName());
  return TCL_OK;
}

int TclScale::fromCmd(int argc, char **argv) {
  if(argc > 1) return COMMAND_SYNOPSIS;
  if(argc == 1) {
    if(sscanf(argv[0], "%lf", &min) != 1) return COMMAND_SYNOPSIS;
    bool redraw = min > value;
    if(tickle("%s configure -from %f", (char *) windowName(), min) == TCL_ERROR)
      return TCL_ERROR;
    if(redraw) command(min);
  }
  setResult("%f", min);
  return TCL_OK;
}

int TclScale::toCmd(int argc, char **argv) {
  if(argc > 1) return COMMAND_SYNOPSIS;
  if(argc == 1) {
    if(sscanf(argv[0], "%lf", &max) != 1) return COMMAND_SYNOPSIS;
    bool redraw = max < value;
    if(tickle("%s configure -to %f", (char *) windowName(), max) == TCL_ERROR)
      return TCL_ERROR;
    if(redraw) command(max);
  }
  setResult("%f", max);
  return TCL_OK;
}

int TclScale::resolutionCmd(int argc, char **argv) {
  double resolution;

  if(argc > 1) return COMMAND_SYNOPSIS;
  if(argc == 1) {
    if(sscanf(argv[0], "%lf", &resolution) != 1) return COMMAND_SYNOPSIS;
    if(tickle("%s configure -resolution %f", (char *) windowName(), resolution)
       == TCL_ERROR) return TCL_ERROR;
  }
  tickle("%s cget -resolution", (char *) windowName());
  setResult(getModule()->getInterp()->result);
  return TCL_OK;
}

int TclScale::getCmd(int argc, char **argv) {
  if(argc != 0) return COMMAND_SYNOPSIS;
  setResult("%f", get());
  return TCL_OK;
}

int TclScale::setCmd(int argc, char **argv) {
  double val;

  if(argc != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[0], "%lf", &val) != 1) return COMMAND_SYNOPSIS;
  set(val);
  setResult("%f", get());
  return TCL_OK;
}
int TclScale::maxStepsCmd(int argc, char **argv) {
  unsigned int val;

  if (argc==1)
    {
      if(sscanf(argv[0], "%ud", &val) != 1) return COMMAND_SYNOPSIS;
      maxSteps=val;
    }
  if(argc <2) 
    {
      setResult("%u", maxSteps);
      return TCL_OK;

    }
  else
    return COMMAND_SYNOPSIS;
}
int TclScale::commandCmd(int argc, char **argv) {
  double val;

  if(argc != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[0], "%lf", &val) != 1) return COMMAND_SYNOPSIS;
  command(val);
  setResult("%f", get());
  return TCL_OK;
}

int TclScale::activateCmd(int argc, char **argv) {
  bool bval;

  if(argc > 1) return COMMAND_SYNOPSIS;
  if(argc == 1) {
    if(!TclModule::boolValue(argv[0], bval)) return COMMAND_SYNOPSIS;
    if(bval) on();
    else off();
  }
  setResult(TclModule::boolString(active()));
  return TCL_OK;
}


int TclScale::scalablesCmd(TclModule *m, int argc, char **argv) {
  TclString res;

  if(argc > 1) return COMMAND_SYNOPSIS;
  Scalable *sc=AutoLink<Scalable>::getHead();
  for(;sc;sc=sc->AutoLink<Scalable>::following())
    res += sc->objectName() + SPACE_STRING;
  m->setResult(res);
  return TCL_OK;
}


int TclScale::objectsCmd(int argc, char **argv) {
  TclString res, one;

  if(argc > 0) return COMMAND_SYNOPSIS;
  DPList<Scalable>::const_iterator it;
  const DPList<Scalable>& lst = getObjects();
  for(it = lst.begin(); it != lst.end(); it++) {
    one.sprintf("%s ", (char *) (*it)->objectName());
    res += one;
  }
  setResult(res);
  return TCL_OK;
}


int TclScale::addCmd(int argc, char **argv) {
  if(argc < 1) return COMMAND_SYNOPSIS;

  int i = 0;
  TclString res;
  while(i < argc) {
    Scalable *scalable=0;
    Scalable *sc=AutoLink<Scalable>::getHead();
    for(;sc;sc=sc->AutoLink<Scalable>::following())
      {
	if (sc->objectName()==(const char *)argv[i])
	  scalable=sc;
      }
    if (scalable==0)
      return COMMAND_SYNOPSIS;
    try {
      *this += scalable;
      res += TclString(argv[i]) + SPACE_STRING;
    }
    catch(DPList<Scalable>::AlreadyInList&) {
      setResult("%s already handled by %s%u", (char *) scalable->objectName(),
		getObjectName(), getRef());
      return TCL_ERROR;
    }
    i++;
  }
  res += "added";
  setResult(res);
  return TCL_OK;
}


int TclScale::removeCmd(int argc, char **argv) {
  if(argc < 1) return COMMAND_SYNOPSIS;

  int i = 0;
  TclString res;
  Scalable *sc=0;
  while(i < argc) {
    DPList<Scalable>::const_iterator it;
    const DPList<Scalable>& lst = getObjects();
    for(it = lst.begin(); it != lst.end(); it++) {
      if ((*it)->objectName()==(const char *)argv[i])
	sc=(*it);
    }
    if (!sc) {
      setResult("%s not handled by %s%u", (char *) sc->objectName(),
		getObjectName(), getRef());
      return TCL_ERROR;
    }
    *this -= sc;
    res += TclString(argv[i]) + SPACE_STRING;
    i++;
  }
  res += "removed";
  setResult(res);
  return TCL_OK;
}

int TclScale::dieCmd(int argc, char **argv) {
  if(argc>1) return COMMAND_SYNOPSIS;
  if(argc==1){
     if(!TclModule::boolValue(argv[0],dieWhenEmpty)) return  COMMAND_SYNOPSIS;
  }
  setResult(TclModule::boolString(dieWhenEmpty));
  return TCL_OK;
}


} // namespace yav

