#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#pragma implementation "graphics2D/TclScene2DObject.h"
#include <graphics2D/TclScene2DObject.h>
#include <graphics2D/TclScene2D.h>

namespace yav {

TclScene2DObject::~TclScene2DObject() {
 }
void TclScene2DObject::done(const std::type_info& id,
			    const char *install, const char *uninstall,
			    Camera2D::SelectionFunctionType fct) {
  TclObject::done(id);
  if (fct)
    selectionFunction=fct;
  setInstallMenuCommandName(install);
  setUninstallMenuCommandName(uninstall);
  TclScene2D *tclscene=  (TclScene2D *)(((Data2D *) client)->getScene()->getInterface());
  tclscene->selectCameraObject((Data2D *) client);
  ((Data2D *) client)->updateCenterAndRadius();
  tclscene->redraw();

}

void TclScene2DObject::deleted(const std::type_info& id) {
  if(client) {
    getTclScene()->unselectCameraObject((Data2D *) client);
  }
  TclObject::deleted(id);
}

/// add a selection to the map of selection modes
void TclScene2DObject::addSelectionMode(SelectionMode *m) {
  selectionModeMap[(char *)m->name]=m;
}
/// remove a selection given its name
void TclScene2DObject::removeSelectionMode(char *name) {
  std::map<char *,SelectionMode *,CStringCompare>::iterator it=selectionModeMap.find(name); 
  if (it!=selectionModeMap.end()) {
    delete  (*it).second;
    selectionModeMap.erase(it);
  }
}
/// get a selection given from its  name
const TclScene2DObject::SelectionMode *TclScene2DObject::getSelectionMode(char *name) const {
  std::map<char *,SelectionMode *,CStringCompare>::const_iterator it=selectionModeMap.find(name); 
  if (it!=selectionModeMap.end()) {
    return (*it).second;
  }
  else 
    return  0;
}
void TclScene2DObject::selectionCallback(TclCamera2D *camera){
  /// eventually install menu
  if((getInstallMenuCommandName() != EMPTY_STRING)&&(getModule()->windowExists((char *)camera->getSubMenuFrame() ))) 
  {
    getModule()->tickle("%s %s %s %s #90C090",
			(char *)getInstallMenuCommandName(),
			(char *) camera->getSubMenuFrame(),
			(char *) camera->name(),
			(char *)(name()));
  }
  if (selectionFunction) {
    camera->getClient()->setSelectionFunction(selectionFunction);    
  }
  if (getInstallPopupMenuCommandName() != EMPTY_STRING) {
    getModule()->tickle("%s  %s %s #90C090",
			(char *)getInstallPopupMenuCommandName(),
			(char *) camera->name(),
			(char *)(name()));
  }
}
void TclScene2DObject::unselectionCallback(TclCamera2D *camera){
  /// eventually install menu
  if(getUninstallMenuCommandName() != EMPTY_STRING) {
    getModule()->tickle("%s %s %s %s",
			(char *)getUninstallMenuCommandName(),
			(char *) camera->getSubMenuFrame(),
			(char *) camera->name(),
			(char *)(name()));
  }
  /// remove selection function
  camera->getClient()->setSelectionFunction(0);    
  
}
int TclScene2DObject::selectionModeCmd( int argc, char **argv) 
{
	if (argc==0)
	{  
		if (selectedMode) 
			setResult(selectedMode->name);
		else {
			setResult("none");
		}
		return TCL_OK;
	}
  else if (argc==1) {
	  if (!strcmp(argv[0],"list")) {
		  std::map<char *,SelectionMode *,CStringCompare>::iterator it=selectionModeMap.begin();
		  TclString out;
		  for(;it!=selectionModeMap.end();++it) {
			  TclString res;
			  res.sprintf("{ {%s} %s } ",(char *)(*it).second->label,(char *)(*it).second->name);
			  out+=res;
		  }
		  setResult(out);
		  return TCL_OK;
	  } else
      return COMMAND_SYNOPSIS;
  }
  else if (argc==2)
    {
		TclCamera2D *cam= (TclCamera2D *) TclObjectKind::getObject(typeid(TclCamera2D), argc,argv,1);
		if (cam!=NULL)
		{
			const  TclScene2DObject::SelectionMode *newMode=getSelectionMode(argv[0]);
			if (newMode) {
				if (newMode!=selectedMode) {
					/// eventually removes the submenu installed by the previous mode
					if ((selectedMode) &&(selectedMode->TclUnInstallProc!=EMPTY_STRING) &&(getModule()->windowExists((char *)cam->getSubMenuFrame())))
						getModule()->tickle("%s %s %s",(char *)selectedMode->TclUnInstallProc,cam->getSubMenuFrame(),
						(char *)name());
					/// decides the global selection mode of the camera
					if (newMode->name=="Object")
						cam->getClient()->setSelectionMode(Camera2D::SELECT_OBJECT);
					else 
						cam->getClient()->setSelectionMode(Camera2D::OBJECT_DEFINED);
					/// sets the new selection Function
					cam->getClient()->setSelectionFunction(newMode->selectionFunction);
					/// eventually calls an additional function to perform additional tasks
					if (newMode->installProc)
						(*newMode->installProc)(this,cam);
					/// eventually sets the new submenu 
					if ((newMode->TclInstallProc!=EMPTY_STRING) &&(getModule()->windowExists((char *)cam->getSubMenuFrame())))
						getModule()->tickle("%s %s %s %s",(char *)newMode->TclInstallProc,cam->getSubMenuFrame(),
						(char *)cam->name(),(char *)name());
					/// replace the pointer
					setSelectedMode(newMode);
				}
			}
	  else 
	    {
	      setResult("unknown selection mode %s",argv[0]);
	      return TCL_ERROR;
	    }
	}
      else {
	setResult("Unknown camera name %s",argv[1]);
	return TCL_ERROR;
      }

      setResult(selectedMode->name);
      return TCL_OK;
    }
  else
    return COMMAND_SYNOPSIS;
}
int TclScene2DObject::sceneCmd(int argc, char **argv) 
{
  if (argc==0)
    {
      TclScene2D *tclscene=  (TclScene2D *)(((Data2D *) client)->getScene()->getInterface());
      setResult("%s%u", TclObjectKind::getObjectName(typeid(TclScene2D)),
		tclscene->getRef());
      return TCL_OK;
    }
    else
    return COMMAND_SYNOPSIS;
}

int TclScene2DObject::translateCmd( int argc, char **argv) 
{
 double tx,ty;
 if (argc!=2) return COMMAND_SYNOPSIS;
 if((sscanf(argv[0],"%lf",&tx)!=1)||(sscanf(argv[1],"%lf",&ty)!=1)) return COMMAND_SYNOPSIS;

  Vec2<double> t(tx,ty);
  getClient()->translate(t);
  getClient()->redrawScene();
  return TCL_OK;
}
int TclScene2DObject::centerAndRadiusCmd( int argc, char **argv) 
{
  if (argc!=0) 
    return COMMAND_SYNOPSIS;
  setResult("%f %f %f",getClient()->getCenter().getX(),getClient()->getCenter().getY(),
	  getClient()->getRadius());
  return TCL_OK;
}

int TclScene2DObject::scaleCmd( int argc, char **argv) 
{
  if (argc!=1) 
    return COMMAND_SYNOPSIS;

  getClient()->scale((double) (atof(argv[0])));
 getClient()->redrawScene();
  return TCL_OK;
}

int TclScene2DObject::rotateCmd( int argc, char **argv) 
{
  if (argc!=1) 
    return COMMAND_SYNOPSIS;
  else
    {
      Rotation2D m(atof(argv[0]));
     getClient()->rotate(m);
     getClient()->redrawScene();
      return TCL_OK;
    }
}

void TclScene2DObject::createOptionCommands( TclObjectKind *kind) {
  kind->newOption("-translate",(TclObject::OPTION) &TclScene2DObject::translateCmd,"%f %f %f",
		  "translate the 2D object");
  kind->newOption("-scale",(TclObject::OPTION) &TclScene2DObject::scaleCmd,"%f",
		  "scale a 2D object around its center");
  kind->newOption("-rotate",(TclObject::OPTION) &TclScene2DObject::rotateCmd,"%f %f %f",
		  "rotate a 2D object around its center given a rotation vector");
  kind->newOption("-scene",(TclObject::OPTION) &TclScene2DObject::sceneCmd,0,
		  "return the 2D scene where is the 2D object");
  kind->newOption("-centerAndRadius",(TclObject::OPTION) &TclScene2DObject::centerAndRadiusCmd,0,
		  "returns the center and radius of the object");
  kind->newOption("-selectionMode",(TclObject::OPTION)  &TclScene2DObject::selectionModeCmd,
			  "[ list | key=%s cam2D%u]",
			  "with no argument this function returns the keyword associated with the selection mode. With the keyword \"list\", it returns a list of keywords and labels corresponding to all selection modes available. With a keyword and camera, it selects a selection mode");


}

} // namespace yav


