#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif



#pragma implementation "graphics2D/TclSurfaceView.h"
#include <graphics2D/TclSurfaceView.h>


namespace yav {


TclSurfaceView::TclSurfaceView(TclModule *m,TclScene2D *scene,
			       SurfaceSliceable  *_mesh) 
  : TclScene2DObject(typeid(TclSurfaceView),
		   new SurfaceView(scene->getClient(), _mesh), m)
{
  done(typeid(TclSurfaceView),"graphics2D::installSurfaceViewMenu","graphics2D::uninstallSurfaceViewMenu");
}

TclSurfaceView::~TclSurfaceView(){
  deleted(typeid(TclSurfaceView));
}

int TclSurfaceView::colorCmd(int argc, char **argv) {
  if ((argc!=0) && (argc!=3))
    return COMMAND_SYNOPSIS;
  if(argc ==3) {
    float col[3];
    col[0]=(float)atof(argv[0]);
    col[1]=(float)atof(argv[1]);
    col[2]=(float)atof(argv[2]);
    getClient()->setTraceColor(col);
	getClient()->redrawScene();
  }
  setResult("%f %f %f", getClient()->getTraceColor(0),
	    getClient()->getTraceColor(1),
	    getClient()->getTraceColor(2));
  return TCL_OK;
}

int TclSurfaceView::destroyCmd(int argc, char **argv) {
  if(argc > 0) return COMMAND_SYNOPSIS;
  TclModule *m = getModule();
  unsigned int ref = getRef();
  const char *n= TclSurfaceView::getObjectName();
  delete this;
  m->setResult("%s%u does not exist any longer",
	      n, ref);
  return TCL_OK;
}
int TclSurfaceView::sceneCmd(int argc, char **argv) {
  if(argc > 0) return COMMAND_SYNOPSIS;
  TclScene2D *scene = (TclScene2D *)getClient()->getScene()->getInterface();
  setResult("%s%u", scene->getObjectName(), scene->getRef());
  return TCL_OK;
}
int TclSurfaceView::widthCmd(int argc, char **argv) {
  if(argc > 1) return COMMAND_SYNOPSIS;
  if (argc > 0) 
    getClient()->setTraceWidth( (float) atof(argv[0]));

  getClient()->updateSliceDisplay();
  
  setResult("%f", getClient()->getTraceWidth());
  return TCL_OK;
}
int TclSurfaceView::interactionModeCmd(int argc, char **argv) {
  if(argc > 1) return COMMAND_SYNOPSIS;
  if (argc > 0) {
    SurfaceView::INTERACTION_MODE im;

    if (!strcmp(argv[0],"grab")) 
      im=SurfaceView::GRAB_MODE;
    else if (!strcmp(argv[0],"position")) 
      im=SurfaceView::POSITION_CONSTRAINT_MODE;
    else if (!strcmp(argv[0],"movePosition")) 
      im=SurfaceView::MOVE_POSITION_CONSTRAINT_MODE;
    else if (!strcmp(argv[0],"removePosition") )
      im=SurfaceView::REMOVE_POSITION_CONSTRAINT_MODE;
    else {
      setResult("%s is not a valid interaction mode. It must be either grab, position, movePosition or removePosition",argv[0]);
      return TCL_ERROR;
    }
    getClient()->setInteractionMode(im);
  }
  switch(getClient()->getInteractionMode()) {
  case SurfaceView::GRAB_MODE :
    setResult("grab");
    break;
  case SurfaceView::POSITION_CONSTRAINT_MODE:
    setResult("position");
    break;
  case SurfaceView::MOVE_POSITION_CONSTRAINT_MODE:
    setResult("movePosition");
    break;
  case SurfaceView::REMOVE_POSITION_CONSTRAINT_MODE:
    setResult("removePosition");
    break;
  }
  return TCL_OK;
}

int TclSurfaceView::useZonesColorCmd(int argc, char **argv) {
  if (argc==0) {
    setResult(TclModule::boolString(getClient()->getUseZonesColor()));
    return TCL_OK;
  }
  else if (argc==1) {
    bool b;
    if (!TclModule::boolValue(argv[0],b))
      return COMMAND_SYNOPSIS;
    getClient()->setUseZonesColor(b);
    getClient()->redrawScene();
    setResult(TclModule::boolString(getClient()->getUseZonesColor()));
    return TCL_OK;
  }
  else 
    return COMMAND_SYNOPSIS;
}
int TclSurfaceView::meshCmd(int argc, char **argv) {
  if(argc > 0) return COMMAND_SYNOPSIS;
  setResult("%s",(char *) getClient()->getSurface()->getSurfaceSliceableName());
  return TCL_OK;
}

int TclSurfaceView::surfaceSliceablesCmd(TclModule *m, int argc, char **argv) {
  TclString res;

  if(argc > 1) return COMMAND_SYNOPSIS;
  SurfaceSliceable *sc=AutoLink<SurfaceSliceable>::getHead();
  for(;sc;sc=sc->AutoLink<SurfaceSliceable>::following())
    res += sc->getSurfaceSliceableName() + SPACE_STRING;
  m->setResult(res);
  return TCL_OK;
}

int TclSurfaceView::create(TclModule *m, int argc, char **argv) {
  TclScene2D *scene = 0;
  SurfaceSliceable *s=0;
  if(argc != 2 && argc != 4) return COMMAND_SYNOPSIS;

  s=AutoLink<SurfaceSliceable>::getHead();
  for(;s && (s->getSurfaceSliceableName()!=(const char *)argv[1]);
      s=s->AutoLink<SurfaceSliceable>::following());
  if (!s)
    {
      m->setResult("can not read surface sliceable %s",argv[1]);
      return TCL_ERROR;
    }
  
  if(argc == 4) {
    if(strcmp(argv[2], "-scene")) return COMMAND_SYNOPSIS;
    scene = (TclScene2D *)
      TclObjectKind::getObject(typeid(TclScene2D), argc, argv, 3);
  } else {
    scene = TclScene2D::getDefaultScene(m);
  }
  if(!scene) return TCL_ERROR;

  TclSurfaceView *smv = new TclSurfaceView(m,scene,s);
  m->setResult("%s%u", smv->TclSurfaceView::getObjectName(), smv->getRef());
  return TCL_OK;
}

} // namespace yav

