#ifndef DATA2D_H
#define DATA2D_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 )
#pragma warning ( disable : 4786 )
#pragma warning ( disable : 4081 )
#endif

#pragma interface "graphics2D/Data2D.h"
#include <basics/ListElement.h>
#include <linearAlgebra/Vec2.h>
#include <linearAlgebra/Rotation2D.h>
#include <graphics/Data.h>
#include <graphics/UserInterface.h>

namespace yav {

class Scene2D;
#ifdef GRAPHICS
class Camera2D;
#endif


/** \class Data2D
    \ingroup Graphics2D
    \brief 2D data that can exist in a Scene2D */
class Data2D : 
#ifdef GRAPHICS
public GraphicsData< Vec2 < double > , Rotation2D, Scene2D, Camera2D, Data2D> 
#else
public Data< Scene2D,Data2D> 
#endif
{
 public:
  /** @name constructor */
  //@{
  /** builds a new Data2D
      @param scene scene to create data in (one is created if scene == 0)
      @param exclusive wether this data is exclusive in scene */
  Data2D(Scene2D *scene = 0, bool exclusive = false
       , unsigned int redrawOrder = 1
);
  //@}
  
  /** @name destructor */
  //@{
  /// empty
  virtual ~Data2D();
  //@}
  
 protected:
  /** redraw order (objects with redraw order 1 are rendered before
      objects with redraw order 2) */
  unsigned int redrawOrder;
public:
  /// retrieve object redraw order (1 or 2)
  unsigned int getRedrawOrder(void) const { return redrawOrder; }
  /** set new object redraw order
      @param ro order, should be 1 or 2 */
  void setRedrawOrder(unsigned int ro) {
    if(ro > 0 && ro < 3) redrawOrder = ro;
  }
  
  /// update object center and radius and re-focus
  void updateCenterAndRadius(void);


};


/** \class UserInterface2D
    \ingroup Graphics2D
    2D user Interface that exists in a Scene2D. These objects
    have an "getPosition" and "setForce" routines that are  suitable 
    for interaction with a Scene2D */
class UserInterface2D : public UserInterface<Scene2D, Vec2<double>,
                                             UserInterface2D> {
 public:
  /// constructor
  UserInterface2D(Scene2D *scene);
  /// destructor
  ~UserInterface2D();
};



} // namespace yav

#ifdef GRAPHICS
#include <graphics2D/Camera2D.h>
#else
#include <graphics2D/Scene2D.h>
#endif

#endif
