#ifndef SCALABLE_H
#define SCALABLE_H

#pragma interface "graphics2D/Scalable.h"
#include <basics/TclString.h>
#include <basics/DPListElement.h>
#include <framework/Activable.h>
#include <basics/AutoLink.h>

namespace yav {

class TclScale;

/** \class Scalable
    \ingroup Graphics2D
    \brief any object managed through a TclScale
*/
class Scalable :  public AutoLink<Scalable>,
  public DPListElement<Scalable> {
 public:
  /** Dummy Constructor */
  Scalable() : DPListElement<Scalable>() {}
  /** Dummy Destructor */
  virtual ~Scalable(){}
  /** get object string identifier
      @return object name */
  virtual TclString objectName(void) const = 0;

  /** scale value changed
      @param scale calling scale object
      @param val new scale value */
  virtual void scaleValue(const TclScale *scale, double val) = 0;

  /** scale was required to move without updating scalable objects: warn them anyway
      @param scale calling scale object
      @param val new scale value */
  virtual void scaleTo(const TclScale *scale, double val) {}

  /** get min possible scale value for this scalable
      @param scale calling scale object
      @return min possible value */
  virtual double getMin(const TclScale *scale) const=0;

  /** get max possible scale value for this scalable
      @param scale calling scale object
      @return max possible value */
  virtual double getMax(const TclScale *scale) const=0;

  /** get required scale step for this scalable
      @param scale calling scale object
      @return step value */
  virtual double getStep(const TclScale *scale) const { return 1.0; }

  /** get current object slice
      @param scale calling scale object
      @return slice number */
  virtual double getSlice(const TclScale *scale) const { return 0.0; }

};


} // namespace yav


#endif

