#ifndef SCENE2D_H
#define SCENE2D_H


#pragma interface "graphics2D/Scene2D.h"
#include <basics/AutoLink.h>
#include <linearAlgebra/Vec2.h>
#include <linearAlgebra/Rotation2D.h>
#ifdef GRAPHICS
#include <graphics2D/Camera2D.h>
#else
#include <graphics/Scene.h>
#endif
#include <graphics2D/Data2D.h>
#include <graphics/FileScene.h>
#ifdef GRAPHICS
#include <graphics/SceneItem.h>
#endif

namespace yav {


#ifdef TCL_INTERFACE
class TclScene2D;
#endif


/** \class Scene2D
    \ingroup Graphics2D
    \brief 2D scene */
class Scene2D : 
#ifdef GRAPHICS
public GraphicsScene<Vec2<double>, Data2D,
  UserInterface2D,  Camera2D, Scene2D>,
#else
public Scene<Data2D, UserInterface2D, Scene2D>,
#endif
  public AutoLink<Scene2D> {
 public:
  /** @name constructor */
  //@{
  /// initializes scene with nothing inside
  Scene2D() : 
#ifdef GRAPHICS
GraphicsScene<Vec2<double>, Data2D,  UserInterface2D, Camera2D, Scene2D>() 
#else
Scene<Data2D,  UserInterface2D,Scene2D>() 
#endif
{}
  //@}

  /** @name destructor */
  //@{
  /// destroy each object in list
  virtual ~Scene2D() {}
  //@}
  /// redraw all scene cameras
  virtual void redraw(void);

  // read a file into a scene : returns true if it worked 
  virtual bool read(const TclString &filename) 
    { return(FileScene<Scene2D>::readScene(filename, this)); }

  /// returns scene reference
  virtual unsigned int getRef(void) const {
    return AutoLink<Scene2D>::getRef();
  }


};


} // namespace yav


#endif
