#ifndef SLICESCALE_H
#define SLICESCALE_H

#pragma interface "graphics2D/SliceScale.h"
#include <graphics2D/TclScale.h>


namespace yav {


/** \class SliceScale
    \ingroup Graphics2D
    \brief TCL scale for inrimage slices */
class SliceScale : public TclScale {
 public:
  /** build a slice scale object from an existing TCL scale
      @param module module this SliceScale is created in
      @param scaleName existing TCL scale name 
      @param sc a pointer on a scalable attached to that scale
      @param id the identifier of the scale */
  SliceScale(TclModule *module, const TclString& scaleName,
	     Scalable *sc=0,
	     const std::type_info& id = typeid(SliceScale));
  /** build a slice scale object from an existing TCL scale
      @param module module this SliceScale is created in
      @param scaleName existing TCL scale name 
      @param label the scale label describing the geometry of the slice (X,Y,Z Or Alpha) 
      @param color a string storing the background scale color 
      @param sc a pointer on a scalable attached to that scale
      @param id the identifier of the scale */
  SliceScale(TclModule *module, const TclString& scaleName,
	     const TclString& label, const TclString color,
	     Scalable *sc=0,
	     const std::type_info& id = typeid(SliceScale));

  /** builds a slice scale object from an existing TCL scale
      manage given list of inrimage views
      @param module this SliceScale is created in
      @param scaleName
      @param label the scale label describing the geometry of the slice (X,Y,Z Or Alpha) 
      @param color a string storing the background scale color 
      @param list a list of scalable attached to the slice scale 
      @param id the identifier of the scale */
  SliceScale(TclModule *module, const TclString& scaleName,
	     const TclString& label, const TclString color,
	     std::list<Scalable *>& list,
	     const std::type_info& id = typeid(SliceScale));

  /// empty
  virtual ~SliceScale();

  /// dying camera callback
  virtual void dying(const Scalable *);

  /** reset bounds and label
      @param sc scalable giving new scale bounds
      @param lab new scale label */
  virtual void reset(Scalable *sc, const TclString& lab = EMPTY_STRING);
  
  /** manage new scalable
      @param sc new scalable */
  virtual void operator+=(Scalable *sc)
    throw(DPList<Scalable>::NullPointer, DPList<Scalable>::AlreadyInList);

  /** get rid of one scalable management
      @param sc scalable to release */
  virtual void operator-=(Scalable *sc) throw(DPList<Scalable>::NotFound);

  /// returns object by ref or 0 if it is not found
  static SliceScale *getObject(unsigned int n) {
    return (SliceScale *) TclObjectKind::getObject(typeid(SliceScale), n);
  }

  /** returns object by name (argv[i]) or 0 with an error message
      in module interpretor if it is not found */
  static SliceScale *getObject(int argc, char **argv, int i = 1) {
    return (SliceScale *) TclObjectKind::getObject(typeid(SliceScale),
						   argc, argv, i);
  }

  /// TCL new operator callback
  static int create(TclModule *m, int argc, char **argv);

  /// returns scale type
  virtual const std::type_info& getType(void) const { return typeid(SliceScale); }

 private:
  /// constructors common code
  void init(void);
};


} // namespace yav


#endif
