#ifndef SLICEABLE_H
#define SLICEABLE_H

#pragma interface "graphics2D/Sliceable.h"
#include <basics/AutoLink.h>
#include <basics/ListElement.h>
#include <linearAlgebra/Vec3.h>
#include <basics/Pointed.h>
#include <graphics2D/Scene2D.h>
#include <map>


namespace yav {


class SliceCamera;

/** \class Sliceable
    \ingroup Graphics2D
    \brief any 2D element to put in one slice of a 3D image */
class Sliceable : public Data2D {
 public:
 /// cameras related information
  struct CamInfo {
    /// is redraw needed in this camera ?
    bool needRedraw;
    /// default constructor
    CamInfo() : needRedraw(true) {}
    /// empty destructor
    virtual ~CamInfo() {}
  };
 protected:
  /// pairs of cameras and need recompute flag
  std::map<SliceCamera *, CamInfo *> cameras;
  /// object creation method
  void init(void);
  /// deleted camera callback
  virtual void deletedCamera(SliceCamera& cam) {}
 public:
  /** @name constructor */
  //@{
  /** build a new sliceable
      @param scene the 2D where the object is located 
      @param _exclusive wether the object should be the only one of his kind in a 2D scene */
  Sliceable(Scene2D *scene, bool _exclusive=false);
  //@}

  /** @name destructor */
  //@{
  /// empty
  virtual ~Sliceable();
  //@}

  /** slice changed callback
      @param cam camera whose slice changed */
  virtual void recompute(SliceCamera *cam);

  /** callback function for dying slice camera
      @param cam dying slice camera pointing on this object */
  virtual void dying(SliceCamera *cam);
  /// new camera callback
  virtual void createdCamera(SliceCamera& cam);
  /// deleted camera callback
  virtual void destroyedCamera(SliceCamera& cam);
};



} // namespace yav


#endif
