#ifndef _SURFACESLICEABLE_H 
#define _SURFACESLICEABLE_H 
 
#ifdef _MSC_VER 
#pragma warning ( disable : 4068 4786 4081 ) 
#endif 

#pragma interface "graphics2D/SurfaceSliceable.h"
#include <linearAlgebra/Vec2.h>
#include <linearAlgebra/Vec3.h>
#include <basics/TclString.h>
#include <basics/Pointed.h>
#include <basics/AutoLink.h>

namespace yav { 

/// contour produced when a mesh is intersected by a plane
struct SliceContour {
  /// contour number of vertices
  int nbvertices;
  /// X coordinates of each contour vertex
  float *X;
  /// Y coordinates of each contour vertex
  float *Y;
  /// Y coordinates of each contour vertex
   float *Z;

  /** zone[i] is the rank of the zone ( -1 if none)  that the segment between vertex i and i+1 crosses */
  int *zone;
  /// was the contour built from selected model ?
  bool fromSelected;
  /// should we draw segment between vertex i and i+1 ?
  bool *draw;
  /// Sagital, coronal or transversal slice?
  int geometry;

  /// default constructor
  SliceContour(): nbvertices(0), X(0), Y(0),Z(0), zone(0), fromSelected(false),
    draw(0), geometry(0) {
  }
  /// destructor: delete arrays
  ~SliceContour() {
    if(X) delete[] X;
    if(Y) delete[] Y;
    if(Z) delete[] Z;
    if(zone) delete[] zone;
    if(draw) delete[] draw;
  }
};

/** \class SurfaceSliceable
    \ingroup Graphics2D
    class allowing to create region of interest and to display surfaces in
    slice cameras */
class SurfaceSliceable :   public AutoLink<SurfaceSliceable>,  public Pointed<SurfaceSliceable> {
  public:
  /** @name constructor */
  //@{
  /** build a new surface sliceable*/
  SurfaceSliceable() {
  }
  //@}

  /** @name destructor */
  //@{
  /// empty
  virtual ~SurfaceSliceable(){}
  //@}

  /** compute surface intersection with a plane
      @param normal plane normal
      @param origin one plane point
      @param vx plane x axis direction
      @param vy plane y axis direction */
  virtual std::list<SliceContour *> intersectWithPlane(const Vec3<double>& normal,
	  const Vec3<double>& origin)=0;
  

  /** returns the vertex position pointer that is closest from a point
      @param dir the normal of slice plane
      @param pos the point in the slice plane that has been selected */
  virtual Vec3<double> *grab2D(const Vec3<double>& dir, 
			       const Vec3<double>& pos) {
    return 0; }
  /** tells the surface mesh that a position constraint should be added at a given position 
      @param position the 3D position where the surface should go through
      @return retuns a reference that identifies the position constraint  implemented on the surface */
  virtual unsigned int addSurfacePointConstraint(const Vec3<double>& position) {
    return 0;
  }
  /** modifies the position of a given position constraint 
      @param position the 3D position where the surface should go through
      @param ref  a reference that identifies the position constraint  implemented on the surface */
  virtual void moveSurfacePointConstraint(const Vec3<double>& position, unsigned int ref){}

  /** removes a given position constraint 
      @param ref  a reference that identifies the position constraint  implemented on the surface */
  virtual void deleteSurfacePointConstraint(unsigned int ref) {}

  /** modifies the display of all sliceables (usually SurfaceView) that 
      owns this surface sliceable object */
  virtual void updateSliceDisplay(); 

   /// returns the string identifying the Tcl name of the object
  virtual TclString getSurfaceSliceableName() const =0;
  
};

} // namespace yav 

#endif 
