#ifndef TCLBOUNDINGBOX3D_H
#define TCLBOUNDINGBOX3D_H

#pragma interface "graphics2D/TclBoundingBox3D.h"
#include <basics/DPOwner.h>
#include <graphics2D/Data2D.h>
#include <graphics2D/Sliceable.h>
#include <graphics2D/TclSliceCamera.h>



namespace yav {



/** \class TclBoundingBox3D
    \ingroup Graphics2D
    \brief 3D box drawable in a slicecamera for ROI selection purpose */
class TclBoundingBox3D : public Sliceable, public TclObject  {
 private:
  /// box ccordinates
  unsigned int box[6];
  /// drawing color
  TclString color;
  /// right mouse button popup menu
  TclString menu;

  /// cube faces enumeration
  enum FACE {
    XMIN = 0,
    XMAX = 3,
    YMIN = 1,
    YMAX = 4,
    ZMIN = 2,
    ZMAX = 5,
    NONE = 6
  };

  /// selected box face
  FACE face;

 public:
  /** @name constructor */
  //@{
  /** creates a box
      @param m module to create TCL object
      @param scene scene that object lies in
      @param x1 box first x coordinate
      @param y1 box first y coordinate
      @param z1 box first z coordinate
      @param x2 box second x coordinate
      @param y2 box second y coordinate
      @param z2 box second z coordinate
      @param color box color
      @param menu mouse button popup menu */
  TclBoundingBox3D(TclModule *m,
		 TclScene2D *scene,
		 unsigned int x1, unsigned int y1, unsigned int z1,
		 unsigned int x2, unsigned int y2, unsigned int z2,
		 const TclString& color = "cyan",
		 const TclString& menu = EMPTY_STRING);
  //@}

  /** @name destructor */
  //@{
  /// destroys this box
  virtual ~TclBoundingBox3D();
  //@}

  /// Sliceable recompute virtual method
  virtual void recompute(SliceCamera *cam);

  /// returns box coordinates
  const unsigned int *getBox(void) const { return box; }

  /// selects closest box face to given voxel
  void setClosestFace(SliceCamera::SLICE_ORIENTATION geom, unsigned int x,
		      unsigned int y, unsigned int z);

  /// moves selected box face to given voxel
  void moveFace(unsigned int x, unsigned int y, unsigned int z);

  /// unselect box face
  void releaseFace(void) { face = NONE; }
/** callback called when a button is pressed in a camera
     @param button indicates which button is pressed
     @param modifier indicates if a modifier (shift or control key) is pressed
     @param cam the 2D camera where occured the click button
     @param mx first coordinate of the pixel clicked 
     @param my second coordinate of the pixel clicked */ 
  virtual void mouseButtonPressed(CameraTypes::WHICH_BUTTON button,
				  CameraTypes::KEY_MODIFIER modifier,
				  Camera2D *cam,int mx,int my);
 /** callback called when a button is released in a camera
     @param button indicates which button is released
     @param modifier indicates if a modifier (shift or control key) is pressed
     @param cam the 2D camera where occured the click button
     @param mx first coordinate of the pixel clicked 
     @param my second coordinate of the pixel clicked */ 
  virtual void mouseButtonReleased(CameraTypes::WHICH_BUTTON button,
				   CameraTypes::KEY_MODIFIER modifier,
				     Camera2D *cam,int mx,int my);
 /** callback called when a button is moved after having been pressed
     @param cam the 2D camera where occured the click button
     @param x first coordinate of the pixel clicked 
     @param y second coordinate of the pixel clicked */ 
  virtual void mouseButtonMoved(Camera2D *cam,int x,int y);


  /** Translates the 2D object @param t translation vector */
  virtual void translate(const Vec2<double>& t) {}

  /** Rotates the 2D object: no rotation available for images
      @param r rotation matrix */
  virtual void rotate(const Rotation2D &r) {}

  /** Scales the 2D object
      @param s scale factor */
  virtual void scale(const double s) {}

  /// Returns the graphic object center and weight.
  virtual void updateCenter(Vec2<double>& center, unsigned int& n)  {
    center = Vec2<double>(0.0, 0.0);
    n = 0;
  }

  /// Returns the graphic object radius given its center
  virtual double updateRadius(const Vec2<double>& center)  { return 0.0; }

  /** Return MAXDOUBLE, an image is considered as background
      and may not be selected */
  virtual double computeClosestDistance(DataTypes::MouseClick& mc) const {
    return maxTypeValue<double>(0);
  }

  /// nothing done when an image is selected
  virtual void select(Camera2D *) {}

  /// nothing done when an image is unselected
  virtual void unselect(Camera2D *) {}

  /** default redraw function for sliceable objects
      @param cam camera to redraw */
  virtual void redraw(Camera2D *cam);

  /// new TCL command callback
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }
  /// new TCL command callback
  static int create(TclModule *m, int argc, char **argv);
  /// -destroy option
  int destroyCmd(int argc, char **argv);
  /// -boxSize option
  int boxSizeCmd(int argc, char **argv);
};


} // namespace yav


#endif
