#ifndef TCLSCENE2D_H
#define TCLSCENE2D_H

#pragma interface "graphics2D/TclScene2D.h"
#include <graphics/TclScene.h>
#include <graphics2D/Scene2D.h>
#include <graphics2D/TclCamera2D.h>
#include <modules/TclModule.h>
#include <modules/TclObject.h>


namespace yav {


/** \class TclScene2D
    \ingroup Graphics2D
    \brief 2D scene for TclData2D objects */
 class TclScene2D : public TclScene<Scene2D, TclScene2D, Data2D, Camera2D,TclCamera2D> {

protected:
#ifndef WIN32
  /** a display that is necessary to destroy context */
  Display *aDisplay;

#endif
 public:

  /** @name constructor */
  //@{
  /** builds a new 2D TclScene
      @param m module to create scene in */
  TclScene2D(TclModule *m);
  //@}

  /** @name destructor */
  //@{
  /// empty destructor
  virtual ~TclScene2D();
  //@}

  /// returns scene reference
  virtual unsigned int getRef(void) const {
    return TclObject::getRef();
  }
  
 /** returns the default  TclScene2D : if there are several existing TclScene2D then returns the first TclScene2D; otherwise creates a new  TclScene2D
      @param m module used to eventually create a new TclScene2D
      @return the default TclScene2D */
  static TclScene2D *getDefaultScene(TclModule *m);

  /** reads a data file into a scene
      @param filename data file name
      @return true if data was succesfully read, false otherwise */
  bool read(const TclString &filename) {
    return FileScene<Scene2D>::readScene(filename, getClient());
  }
  
  /// fix ambigous AutoLink::getHead method call
  static TclScene2D *getHead(void) {
    return (TclScene2D *) TclObject::getHead(typeid(TclScene2D));
  }

  /// fix ambigous AutoLink::following method call
  virtual TclScene2D *following(void) const {
    return (TclScene2D *) TclObject::following();
  }  

  /// TCL -destroy command option
  int destroyCmd(int argc, char **argv);
  /// TCL -load command option
  int loadCmd(int argc, char **argv);
  /// TCL -iteration command option
  int iterationCmd(int argc, char **argv);
  /// TCL -start/stop computation command option
  int computationCmd(int argc, char **argv);
  /// TCL -redraw command option
  int redrawCmd(int argc, char **argv);
  /// TCL -focus command option
  int focusCmd(int argc, char **argv);
  /// -object TCL command option
  int objectsCmd(int argc, char **);
  /// -camera TCL command option
  int cameraCmd(int argc, char **);
  /// -deformable TCL command option
  int deformableCmd(int argc, char **);
 

  /** creates a new scene
      @param m module to create scene in
      @param argc number of command line arguments
      @param argv command line arguments */
  static int create(TclModule *m, int argc, char **argv);

  /// returns client
  Scene2D *getClient(void) const { return (Scene2D *) client; }

  friend class TclCamera2D;

};

/// TclObjectKind for 2D scenes short-hand
typedef SceneItem< Scene2D > Scene2DItem;



} // namespace yav


#endif
