#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#pragma implementation "graphics3D/Camera3D.h"
#pragma implementation "graphics/Camera.h"
#include <graphics3D/Camera3D.h>
#include <graphics3D/Camera3DActivation.h>
#include <graphics3D/ClippingPlane.h>
#include <graphics3D/Light.h>

namespace yav {


#define ALPHA_VAL 1.0

#ifndef M_PI
#define M_PI 3.1415926535897932
#endif

// static members initialization
template <> CameraTypes::BUTTON_STATE Camera<Vec3<double>, Rotation3D, Scene3D, Data3D, Camera3D>::B[3] = {
  CameraTypes::BUTTON_RELEASED,
  CameraTypes::BUTTON_RELEASED,
  CameraTypes::BUTTON_RELEASED
};

template <> Vec3<double> Camera<Vec3<double>, Rotation3D, Scene3D, Data3D, Camera3D>::pressPoint=Vec3<double>();
template <> Vec3<double> Camera<Vec3<double>, Rotation3D, Scene3D, Data3D, Camera3D>::lastPoint=Vec3<double>();
template <> LONGINT Camera<Vec3<double>, Rotation3D, Scene3D, Data3D, Camera3D>::clickTime=0;
template <> LONGINT Camera<Vec3<double>, Rotation3D, Scene3D, Data3D, Camera3D>::renderDuration=0;
template <> LONGINT Camera<Vec3<double>, Rotation3D, Scene3D, Data3D, Camera3D>::clickDuration=0;

void Camera3D::setEyeOffset(const Vec3<double> &v)
{
eyeOffset=v;
}

Camera3D::Camera3D(Camera3D *cam) : ListElement<Camera3D>(),
Camera<Vec3<double>, Rotation3D, Scene3D, Data3D, Camera3D>()   {
  initCamera();
}

  Camera3D::Camera3D(Scene3D *scn) :  ListElement<Camera3D>(),
	  Camera<Vec3<double>, Rotation3D, Scene3D, Data3D, Camera3D>(scn)  {
  initCamera();
}

void Camera3D::initCamera(void) {

  selectionFunction = 0;

  center = refCenter = DEFAULT_CAMERA3D_CENTER;
  fieldOfView = DEFAULT_FIELD_OF_VIEW;
  fieldDepthRatio = DEFAULT_FIELD_DEPTH_RATIO;
  // A priori no object is selected
  selectedObject=0;
  cameraActivation=0;
  matTransform.resetToIdentity();

  moveMode = ROTATION_MODE;
  autoMoveMode = NO_AUTO_MOVE;
  autoRotation.resetToIdentity();
  autoScale = 0.0;

  /* If a null pointer was passed */
  if(!scene) {
    /* First check that there exists allready one graphics scene*/
    if(AutoLink<Scene3D>::getCount() > 0)  
      scene = AutoLink<Scene3D>::getHead(); 
    else 
    /*Otherwise  creates a new scene */
    scene = new Scene3D();
  }
  scene->addCamera(this);

 stencilBuffer=false;
 stencilBufferDisplayList=0;
 StereoFlag=false;

 eyeOffset=Vec3<double>(0.035, 0.0, 0.0);
  fog=false;
  fogRange = DEFAULT_FOG_RANGE;
  useTwoSides = false;
  useTrueSpecularity = false;
  beingDestroyed=false; 
  globalTimer.reset();
  count = 0;
  showCopyright=true;
  showFrameRate=false;
  sprintf(stringFrameRate,"Frame Rate   :");
  sprintf(stringCopyright,"%cINRIA - Asclepios project",169);


}


Camera3D::~Camera3D() {
  beingDestroyed=true;
	if (cameraActivation)
    delete cameraActivation;
}

void Camera3D::toggleFrameRate() {
    if (showFrameRate)
      showFrameRate=false;
    else
      showFrameRate=true;
  }

void Camera3D::toggleCopyright() {
    if (showCopyright)
      showCopyright=false;
    else
      showCopyright=true;
  }
 
  bool Camera3D::isCameraBeingDestroyed() const {
		return beingDestroyed;
  }

void Camera3D::destroyCamera() {
  beingDestroyed=true;

  std::list<ClippingPlane *>::iterator itp;
  std::list<ClippingPlane *>::iterator itp2;
  for(itp = clipPlanes.begin(); itp != clipPlanes.end(); itp=itp2) {
	  itp2=itp;
	  itp2++;
    delete (*itp);
  }
  clipPlanes.erase(clipPlanes.begin(),clipPlanes.end()); 

  std::list<Light *>::iterator itl;
  std::list<Light *>::iterator itl2;
  for(itl = lights.begin(); itl != lights.end(); itl=itl2)
  {	
    // save the iterator because Light::~Light manipulates lights list
    itl2=itl;
    itl2++;
    delete (*itl);
  }
  lights.erase(lights.begin(),lights.end()); 	

  if(stencilBufferDisplayList) {
    glDeleteLists(stencilBufferDisplayList,1);
  }
}

/*-----------------------------------------*/
Vec3<double> Camera3D::computeEye() {
  eye = Vec3<double>(matTransform.getValue(0, 2) * dl / scale,
		     matTransform.getValue(1, 2) * dl / scale,
		     matTransform.getValue(2, 2) * dl / scale);
  eye += center;

  return eye;
}

/* ---------------------------------------------------------------*/
void Camera3D::setParameters() {
  expansionX = fieldDepthRatio * radius * 
    tan(fieldOfView * M_PI / 360.0) / (width / 2.0);
  expansionY = fieldDepthRatio * radius *
    tan(fieldOfView * M_PI / 360.0) / (height / 2.0); 
  dl = fieldDepthRatio * radius;
}

/*-----------------------------------------------------------------------------*/
void Camera3D::mapworld2(const Vec2<double> &pixel, double *wx, double *wy)
{
  double mx = (pixel.getX() - (width / 2.0));
  double my = (-pixel.getY() + (height / 2.0));

  *wx = (float) (mx * expansionX / scale);
  *wy = (float) (my * expansionY / scale);
} 

/*-----------------------------------------------------------------*/
Vec3<double> Camera3D::getMousePosition(const Vec2<double> &pixel) {
  double wx, wy;
  Vec3<double> v;

  mapworld2(pixel, &wx, &wy);

  v = Vec3<double>(matTransform.getValue(0, 0) * wx +
		   matTransform.getValue(0, 1) * wy,
		   matTransform.getValue(1, 0) * wx +
		   matTransform.getValue(1, 1) * wy,
		   matTransform.getValue(2, 0) * wx +
		   matTransform.getValue(2, 1) * wy);
  v += center;

  return v;
}

/*---------------------------------------------------------------*/
void Camera3D::getEyeDirectionVector(const Vec2<double> &pixel, 
				     Vec3<double>& eye,
				     Vec3<double>& dir) {
  Vec3<double> mouse;

  mouse = getMousePosition(pixel);
  eye = computeEye();
  dir = mouse - eye;
  dir /= dir.norm();
}

/*-----------------------------------------------------------*/
/* get trackball location function of mouse pointer location */
Vec3<double> Camera3D::getTrackballVector(int x, int y) const {
  double dx, dy, dist, z;

  dx = 2.0 * (double) x / width - 1.0;
  dy = -2.0 * (double) y / height + 1.0;
  dist = dx*dx + dy*dy;

  if(dist < (SQR_TRACKBALL / 2.0)) z = sqrt(SQR_TRACKBALL - dist);
  else z = SQR_TRACKBALL / (2.0 * sqrt(dist));

  return Vec3<double>(dx, dy, z);
}


/* --------------------------------------------------------*/
Rotation3D Camera3D::getTrackballRotation(const Vec3<double>& b1,
					  const Vec3<double>& b2) {
  Vec3<double> db, r, vi1, vi2, vi3, v[3], vi;
  double dist, phi, cphi, sphi, nr;

  db = b1 - b2;
  dist = db.norm2();

  if (dist > 0) {
    r = cross(b2, b1);
    if( (nr = r.norm()) != 0.0) r /= nr;

    phi = sin (sqrt(dist)) / (2.*TRACKBALL);
    cphi = cos(phi);
    sphi = sin(phi);

    vi1 = r * r.getX();
    vi2 = Vec3<double>(1. - vi1.getX(), -vi1.getY(), -vi1.getZ());
    vi3 = cross(vi2, r);
    vi = (vi2 * cphi) + (vi3 * sphi);
    v[0] = vi1 + vi;
    
    vi1 = r * r.getY();
    vi2 = Vec3<double>(-vi1.getX(), 1 - vi1.getY(), -vi1.getZ());
    vi3 = cross(vi2, r);
    vi = (vi2 * cphi) + (vi3 * sphi);
    v[1] = vi1 + vi;

    vi1 = r * r.getZ();
    vi2 = Vec3<double>(-vi1.getX(), -vi1.getY(), 1 - vi1.getZ());
    vi3 = cross(vi2, r);
    vi = (vi2 * cphi) + (vi3 * sphi);
    v[2] = vi1 + vi;

    return Rotation3D(v[0], v[1], v[2]);
  }
  
  Rotation3D rot;
  rot.resetToIdentity();
  return rot;
}

double Camera3D::getTrackballScale(const Vec3<double>& b1,
				   const Vec3<double>& b2) {
  double theScale= (b1.getY() - b2.getY()) / 2.0;
  if(theScale > 0.0)
    return(1.0 + theScale);
  else
    return(1.0 / (1.0 - theScale));
}

Vec3<double> Camera3D::getTrackballTranslation(const Vec3<double>& b1,
					       const Vec3<double>& b2) {
  return mapToScreen(Vec3<double>(b1.getX() - b2.getX(), b1.getY() - b2.getY(), 0.0) * (radius / scale));
}
Vec3<double> Camera3D::getTrackballZTranslation(const double a) {
  return mapToScreen(Vec3<double>(0.0,0.0,a) * (radius / scale));
}
/*---------------------------------*/
void Camera3D::defineLights(void) {}
void Camera3D::useLights(void) const {}
void Camera3D::init(void) {}

void Camera3D::focusOnRegion(Vec3<double> &c,double _radius)
{
  if (_radius>0)
    {
      focusedYet = true;
      center = refCenter = c; 
      radius = _radius;
      scale = 1.0;
	  if (getCameraActivation())
		getCameraActivation()->setContext();
      setViewingParameters();
    }
}

void Camera3D::doFocus(void) {
  Vec3<double> sceneCenter;
  double sceneRadius;

  scene->getCenterRadius(sceneCenter, sceneRadius);
  if(sceneRadius > 0.0) {
    focusedYet = true;
    center = refCenter = sceneCenter; 
    radius = sceneRadius;
    scale = 1.0;
	 if (getCameraActivation())
		getCameraActivation()->setContext();
    setViewingParameters();
  } 
}



/* compose rotation matrix of camera with the
   rotation (axis, angle) */
void Camera3D::rotate(const Vec3<double>& axis, double angle) {
  int i, j, k;
  double r;

  Matrix3x3<double> S(0.0, -axis.getZ(), axis.getY(),
		      axis.getZ(), 0.0, -axis.getX(),
		      -axis.getY(), axis.getX(), 0.0);
  Matrix3x3<double> S2 = S * S;
  S *= sin(angle);
  S2 *= 1.0 - cos(angle);
  Matrix3x3<double> M(1.0, 0.0, 0.0,
		      0.0, 1.0, 0.0,
		      0.0, 0.0, 1.0);
  M += S;
  M += S2;

  for(i = 0; i < 3; i++)
    for(j = 0; j < 3; j++) {
      r = 0.0;
      for(k = 0; k < 3; k++)
	r += M.getValue(i, k) * matTransform.getValue(j, k);
      S.setValue(j, i, r);
    }
  matTransform = S;

}


void Camera3D::translateCamera(Vec3<double> &t) {
  center -= t;
}

void Camera3D::scaleCamera(double s) {
  scale *= s;
}

void Camera3D::rotateCamera(Rotation3D &rot) {
  int i, j;
  Matrix3x3<double> r;
  for(i = 0; i < 3; ++i)
    for(j = 0; j < 3; ++j)
      r.setValue(i, j,
		 matTransform.getValue(i, 0) * autoRotation.getValue(j, 0) +
		 matTransform.getValue(i, 1) * autoRotation.getValue(j, 1) +
		 matTransform.getValue(i, 2) * autoRotation.getValue(j, 2));
  matTransform = r;
}


Vec3<double> Camera3D::mapToScreen(const Vec3<double>& t) {
  Vec3<double> v1((double) matTransform.getValue(0, 0),
		  (double) matTransform.getValue(0, 1),
		  (double) matTransform.getValue(0, 2));
  Vec3<double> v2((double) matTransform.getValue(1, 0),
		  (double) matTransform.getValue(1, 1),
		  (double) matTransform.getValue(1, 2));
  Vec3<double> v3((double) matTransform.getValue(2, 0),
		  (double) matTransform.getValue(2, 1),
		  (double) matTransform.getValue(2, 2));
  return Vec3<double>(dot(t, v1), dot(t, v2), dot(t, v3));
}



void Camera3D::startAutoMove(AUTO_MOVE_MODE amm) {
  if(autoMoveMode == NO_AUTO_MOVE) {
    autoMoveMode = amm;
    if (cameraActivation)
      cameraActivation->startAutoMotion();
  }
}



void Camera3D::buttonPressed(WHICH_BUTTON w,
			     KEY_MODIFIER k,
			     int x, int y) {
#ifndef WIN32
	struct timeval tv;
#endif
  pressPoint = lastPoint = getTrackballVector(x, y);
  switch(k) {
  case NO_MODIFIER:
    B[w] = BUTTON_PRESSED;
    break;
  case CTRL_KEY:
    B[w] = CTRL_BUTTON_PRESSED;
    break;
  case SHIFT_KEY:
    B[w] = SHIFT_BUTTON_PRESSED;
    break;
  }
#ifndef WIN32
  gettimeofday(&tv, 0);
  clickTime = tv.tv_sec * 1000000 + tv.tv_usec;
#else
  clickTime=time(0);
#endif
  renderDuration = 0;
  stopAutoMove();
  // send event to the seclected object
  if (selectedObject) {
    selectedObject->mouseButtonPressed(w,k,this,x,y);
  }

}
void Camera3D::setSelectionFunction(Selection3DFunctionType sf){
	  selectionFunction=sf;
  }
  
  /** subobject selection method
      @param vec mouse click point
      @param cam clicked camera */
 void Camera3D::select(const Vec2<double> vec, Camera3D *cam){
  if (selectionFunction)
      (*selectionFunction) (selectedObject,vec,cam);
  }
void Camera3D::buttonReleased(WHICH_BUTTON w,  int x, int y) {
#ifndef WIN32
  struct timeval tv;
#endif
  LONGINT releaseTime;
  Vec3<double> p = getTrackballVector(x, y);
#ifndef WIN32
  gettimeofday(&tv, 0);
  releaseTime = tv.tv_sec * 1000000 + tv.tv_usec;
#else
  releaseTime=time(0);
#endif
  clickDuration = releaseTime - clickTime;

  switch(w) {

    // button 1
  case LEFT_BUTTON:
    switch(B[LEFT_BUTTON]) {

    // selection
    case BUTTON_PRESSED:
      if (selectedObject) {
		  Vec2<double> mouse=Vec2<double>(x, y);
          select(mouse, getType());
         selectedObject->mouseButtonReleased(w,NO_MODIFIER,this,x,y);
      }
      break;
    // world scale
    case CTRL_BUTTON_PRESSED:
      autoScale = getTrackballScale(lastPoint, p);
      if(clickDuration - renderDuration > AUTO_TIMING ||
	 pressPoint == p) {
	stopAutoMove();
      }
      else {
	startAutoMove(AUTO_SCALE_MOVE);
      }
      
      moveMode = ROTATION_MODE;
      if (selectedObject)
	selectedObject->mouseButtonReleased(w,CTRL_KEY,this,x,y);
      redraw();
     break;
    case SHIFT_BUTTON_PRESSED:
      // send event to the selected object
      if (selectedObject)
	selectedObject->mouseButtonReleased(w,SHIFT_KEY,this,x,y);
      break;
    default:

      break;
    }
    break;
    
  // button 2
  case MIDDLE_BUTTON:
    
    if(clickDuration - renderDuration > AUTO_TIMING ||
       pressPoint == p)
      stopAutoMove();
    
    else {
      switch(B[MIDDLE_BUTTON]) {
      case BUTTON_PRESSED:
	startAutoMove(AUTO_ROTATION_MOVE);
	break;
      case CTRL_BUTTON_PRESSED:
	startAutoMove(AUTO_TRANSLATION_MOVE);
	break;
      case SHIFT_BUTTON_PRESSED:
	// send event to the selected object
	if (selectedObject)
	  selectedObject->mouseButtonReleased(w,SHIFT_KEY,this,x,y);
	break;
	
      default:
	break;
      }
    }
    moveMode = ROTATION_MODE;
    redraw();
    break;
      
  case RIGHT_BUTTON:
  default:
    break;
  }

  B[w] = BUTTON_RELEASED;
  // bind the mouse motion to call mouseMoved()

}

 /// return the object dealing with the camera activation
  Camera3DActivation *Camera3D::getCameraActivation() const{
	return cameraActivation;
  }
  /// set the object dealing with the camera activation
void Camera3D::setCameraActivation(Camera3DActivation *activation) {
    cameraActivation=activation;
  }
void Camera3D::mouseMoved(int x, int y) {
	  double s;
	  Vec3<double> p = getTrackballVector(x, y);
	  
	  if(B[MIDDLE_BUTTON] != BUTTON_RELEASED ||
		  B[LEFT_BUTTON] != BUTTON_RELEASED) {
		  moveMode = ROTATION_MODE;
		  
		  // button 2
		  switch(B[MIDDLE_BUTTON]) {
			  
		  case BUTTON_PRESSED:
			  autoRotation = getTrackballRotation(lastPoint, p);
			  lastPoint = p;
			  switch(motionMode) {
			  case CameraTypes::MOVE_CAMERA:
				  rotateCamera(autoRotation);
				  break;
			  case CameraTypes::MOVE_CAMERA_OBJECT:
				  rotateCameraObject(autoRotation);
				  break;
			  case CameraTypes::MOVE_SELECTED_OBJECT:
				  rotateObject(autoRotation);
				  break;
			  }
			  break;
			  
			  case CTRL_BUTTON_PRESSED:
				  {
					  Vec3<double> autoTranslation = getTrackballTranslation(p, lastPoint);
					  lastPoint = p;
					  
					  switch(motionMode) {
					  case CameraTypes::MOVE_CAMERA:
						  translateCamera(autoTranslation);
						  break;
					  case CameraTypes::MOVE_CAMERA_OBJECT:
						  translateCameraObject(autoTranslation);
						  break;
					  case CameraTypes::MOVE_SELECTED_OBJECT:
						  translateObject(autoTranslation);
						  break;
					  }
				  }
				  break;
				  
			  case SHIFT_BUTTON_PRESSED:
				  {
					  
					  Vec3<double> autoTranslation= getTrackballZTranslation( p.getY() - lastPoint.getY());
					  lastPoint = p;
					  switch(motionMode) {
					  case CameraTypes::MOVE_CAMERA:
						  translateCamera(autoTranslation);
						  break;
					  case CameraTypes::MOVE_CAMERA_OBJECT:
						  translateCameraObject(autoTranslation);
						  break;
					  case CameraTypes::MOVE_SELECTED_OBJECT:
						  translateObject(autoTranslation);
						  break;
					  }
				  }
				  break;
				  
			  case BUTTON_RELEASED:
				  break;
		  }
		  
		  /* button 1 */
		  switch(B[LEFT_BUTTON]) {
		  case BUTTON_PRESSED:
			  if (selectedObject)
				  selectedObject->mouseButtonMoved(this,x,y);
			  break;
			  
		  case CTRL_BUTTON_PRESSED:
			  autoScale = s = getTrackballScale(p, lastPoint);
			  lastPoint = p;
			  switch(motionMode) {
			  case CameraTypes::MOVE_CAMERA:
				  scaleCamera(s);
				  break;
			  case CameraTypes::MOVE_CAMERA_OBJECT:
				  scaleCameraObject(s);
				  break;
			  case CameraTypes::MOVE_SELECTED_OBJECT:
				  scaleObject(s);
				  if (selectedObject)
					  selectedObject->mouseButtonMoved(this,x,y);
				  break;
			  }
			  break;
			  case SHIFT_BUTTON_PRESSED:
				  // send event to the selected object
				  if (selectedObject)
					  selectedObject->mouseButtonMoved(this,x,y);
				  break;
			  case BUTTON_RELEASED:
				  break;
		  }
	  }
	  
	  
	  if (cameraActivation) {
		  cameraActivation->redraw();
	  }
	  
}



void Camera3D::applyAutoMove(void) {
  switch(autoMoveMode) {
  case NO_AUTO_MOVE:
    break;
  case AUTO_ROTATION_MOVE:
    switch(motionMode) {
    case CameraTypes::MOVE_CAMERA:
      rotateCamera(autoRotation);
      break;
    case CameraTypes::MOVE_CAMERA_OBJECT:
      rotateCameraObject(autoRotation);
      break;
    case CameraTypes::MOVE_SELECTED_OBJECT:
      if(selectedObject) rotateObject(autoRotation);
      break;
    }
    break;
  case AUTO_TRANSLATION_MOVE:
    switch(motionMode) {
    case CameraTypes::MOVE_CAMERA:
      translateCamera(autoTranslation);
      break;
    case CameraTypes::MOVE_CAMERA_OBJECT:
      translateCameraObject(autoTranslation);
      break;
    case CameraTypes::MOVE_SELECTED_OBJECT:
      if(selectedObject) translateObject(autoTranslation);
      break;
    }
    break;
  case AUTO_SCALE_MOVE:
    switch(motionMode) {
    case CameraTypes::MOVE_CAMERA:
      scaleCamera(autoScale);
      break;
    case CameraTypes::MOVE_CAMERA_OBJECT:
      scaleCameraObject(autoScale);
      break;
    case CameraTypes::MOVE_SELECTED_OBJECT:
      if(selectedObject) scaleObject(autoScale);
      break;
    }
    break;
  }
}


void Camera3D::stopAutoMove(void) {
  if(autoMoveMode != NO_AUTO_MOVE) {
    autoMoveMode = NO_AUTO_MOVE;
    if (cameraActivation) 
      cameraActivation->stopAutoMotion();
  }
}

void Camera3D::rotateObject(const Rotation3D &r) {
  Matrix3x3<double> r1, r12;
  unsigned int i, j;

  for(i = 0; i < 3; ++i)
    for(j = 0; j < 3; ++j)
      r1.setValue(i, j, matTransform.getValue(i, 0) * r.getValue(0, j)
		  + matTransform.getValue(i, 1) * r.getValue(1, j)
		  + matTransform.getValue(i, 2) * r.getValue(2, j));

  for(i = 0; i < 3; ++i)
    for(j = 0; j < 3; ++j)
      r12.setValue(i, j, r1.getValue(i, 0) * matTransform.getValue(j, 0)
		   + r1.getValue(i, 1) * matTransform.getValue(j, 1)
		   + r1.getValue(i, 2) * matTransform.getValue(j, 2));

  Rotation3D r2(r12);
  if (selectedObject) {
    selectedObject->getData()->rotate(r2);
    selectedObject->getData()->rotateBoundingBox(r2);
    selectedObject->getData()->redrawScene();
  }  
}

/*------------------------------*/
void Camera3D::drawBoundingBox(Data3D *obj) const
{
  glDisable(GL_LIGHTING);
  glColor3f(boundingBoxColor[0],boundingBoxColor[1],boundingBoxColor[2]);
  glLineWidth(boundingBoxWidth);

  const Vec3<double> *bbox = obj->getBoundingBox();
  Vec3<double> c((bbox[0] + bbox[1]) / 2.0);
  Vec3<double> b0(bbox[0] - c);
  Vec3<double> b1(bbox[1] - c);
  Vec3<double> d0(obj->getOrientation() * b0 + c);
  Vec3<double> d1(b1.getX(), b0.getY(), b0.getZ());
  d1 = obj->getOrientation() * d1 + c;
  Vec3<double> d2(b1.getX(), b1.getY(), b0.getZ());
  d2 = obj->getOrientation() * d2 + c;
  Vec3<double> d3(b0.getX(), b1.getY(), b0.getZ());
  d3 = obj->getOrientation() * d3 + c;
  Vec3<double> d4(b0.getX(), b0.getY(), b1.getZ());
  d4 = obj->getOrientation() * d4 + c;
  Vec3<double> d5(b1.getX(), b0.getY(), b1.getZ());
  d5 = obj->getOrientation() * d5 + c;
  Vec3<double> d6(obj->getOrientation() * b1 + c);
  Vec3<double> d7(b0.getX(), b1.getY(), b1.getZ());
  d7 = obj->getOrientation() * d7 + c;

  glBegin(GL_LINE_LOOP);
  glVertex3dv((const double *) d0);
  glVertex3dv((const double *) d1);
  glVertex3dv((const double *) d2);
  glVertex3dv((const double *) d3);
  glEnd();
  glBegin(GL_LINE_LOOP);
  glVertex3dv((const double *) d4);
  glVertex3dv((const double *) d5);
  glVertex3dv((const double *) d6);
  glVertex3dv((const double *) d7);
  glEnd();
  glBegin(GL_LINES);
  glVertex3dv((const double *) d0);
  glVertex3dv((const double *) d4);
  glVertex3dv((const double *) d1);
  glVertex3dv((const double *) d5);
  glVertex3dv((const double *) d2);
  glVertex3dv((const double *) d6);
  glVertex3dv((const double *) d3);
  glVertex3dv((const double *) d7);
  glEnd();
  glEnable(GL_LIGHTING);
}

/*-----------------------------------------------*/
void Camera3D::applyTransformation() {
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); 
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glTranslated(refCenter.getX(), refCenter.getY(), refCenter.getZ()); 
  //  glScaled(scale, scale, scale);
  glMultMatrixd((const double *) matTransform);
  glTranslated(-center.getX(), -center.getY(), -center.getZ());
} 
/*------------------------------------------*/
void Camera3D::setProjectionParameters(bool reset)
{
  /* Set the Projection Matrix */
  glMatrixMode(GL_PROJECTION);
  if (reset) glLoadIdentity();
  glScaled(scale, scale, 1);
#ifdef _ALPHA_
  gluPerspective(fieldOfView/scale, (GLfloat) width/(GLfloat) height, 
		 1e-3 * radius, 2 * fieldDepthRatio * radius);
#else
  gluPerspective(fieldOfView/scale, (GLfloat) width/(GLfloat) height, 
		 fabs((fieldDepthRatio-2.0)) * radius, 2 * fieldDepthRatio * radius);
#endif
  glMatrixMode(GL_MODELVIEW);
}

/*------------------------------------------*/
void Camera3D::setViewingParameters(bool reset) {
  double cx, cy, cz;

  cx = refCenter.getX();
  cy = refCenter.getY();
  cz = refCenter.getZ();

  /* Set the Projection Matrix */
  setProjectionParameters(reset);
  /* Set the Viewing Matrix */
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(cx, cy, cz + fieldDepthRatio * radius, cx, cy, cz, 0.0, 1.0, 0.0);

  /* Set fog model */
  glFogi(GL_FOG_MODE,GL_LINEAR); 
  glFogf(GL_FOG_START,(float) ((fieldDepthRatio - fogRange) * radius));
  glFogf(GL_FOG_END,(float) ((fieldDepthRatio + fogRange) * radius));
  
  setParameters();
}


/*-----------------------------------*/ 
void Camera3D::resize(void) {
  glViewport(margin, margin, width-2*margin,
	     height-2*margin);
   if (getCameraActivation())
		getCameraActivation()->setContext();
  setViewingParameters();
  if(stencilBuffer) buildStencilBufferDisplayList();
}

/*-----------------------------------*/ 
void Camera3D::resize(unsigned int w, unsigned int h) {
  glViewport(margin, margin, width-2*margin,
	     height-2*margin);
   if (getCameraActivation())
		getCameraActivation()->setContext();
  setViewingParameters();
  if(stencilBuffer) buildStencilBufferDisplayList();
}

/*-----------------------------------*/ 
void Camera3D::initViewport(void) {
  GLfloat light_model_ambient[] = {0.1f, 0.1f, 0.1f, 1.0f};

//  glDrawBuffer(GL_FRONT_AND_BACK); //jcl
  glClearColor(0, 0, 0, ALPHA_VAL); 
  glViewport(margin, margin, width-2*margin,
	     height-2*margin);

  glEnable(GL_DEPTH_TEST);  
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); 

//   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
//   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
//   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,GL_LINEAR);
//   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,GL_LINEAR);
//   glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);

  ///  glEnable(GL_NORMALIZE); 
  glCullFace(GL_BACK);
  glBlendFunc(GL_ONE, GL_ONE);
  
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, light_model_ambient);

  if(useTwoSides)
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
  else
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);

  if(useTrueSpecularity)
    glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
  else
    glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_FALSE);

  setViewingParameters();
  glFlush();
}
bool Camera3D::getFog() const 
  { 
	  return fog;
  }

/*-----------------------------------*/ 
void Camera3D::setBackgroundColor(float r, float g, float b) {
  float f4[4];
  glClearColor(r, g, b, ALPHA_VAL);
  f4[0] = r; f4[1] = g; f4[2] = b; f4[3] = ALPHA_VAL;
  glFogfv(GL_FOG_COLOR, f4);
}

void Camera3D::printString(double posx,double posy, 
			    char *str, GLuint font) {
  if (font==0)
    font=defaultFont;
  unsigned int l=strlen(str);
  unsigned int i;
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glOrtho(-0.5, width-0.5, -0.5, height - .5, 0.0, 1.0);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glDisable(GL_LIGHTING);
  glColor3f(1.0, 1.0, 1.0);
  glRasterPos2i(posx,posy);

  glListBase(font);
  glCallLists(l, GL_UNSIGNED_BYTE, str);
	/*
  for(i=0;i<l;++i)
    if (str[i] == '')
      glCallList(font+169);
    else
      glCallList(font+(GLuint)str[i]); */
  
  glPopMatrix();
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
}



void Camera3D::redraw(void) {
  Vec3<double> oldRefCenter;
  Vec3<double> delta;
  if(moveMode == NAVIGATION_MODE)
    refCenter += Vec3<double>(0.0, 0.0, 5.0 * radius);

  if (StereoFlag) { // prepare things for right eye image
    // move the point of view
    oldRefCenter = refCenter;
    delta = eyeOffset*radius;
    refCenter = oldRefCenter - delta;
    // and select the right buffer to draw in
    glDrawBuffer(GL_BACK_RIGHT);
  }
  
  // compute model view matrix
  applyTransformation();
  /// set projection matrix
  setProjectionParameters();
  // set camera clipping planes
  std::list<ClippingPlane *>::iterator itp;
  for(itp = clipPlanes.begin(); itp != clipPlanes.end(); itp++)
    (*itp)->putInCamera();

  // set spot lights 
  std::list<Light *>::iterator itl;
  for(itl = lights.begin(); itl != lights.end(); itl++)
    (*itl)->putInCamera();

  if(fog) glEnable(GL_FOG);

  if(stencilBuffer){
    glCallList(stencilBufferDisplayList);
    glEnable(GL_STENCIL_TEST);
  }

 // draw all scene objects
ObjectsList< Data3D >::iterator it(getScene()->firstData());
  for(; it!=getScene()->lastData(); it++) {
    if((*it)->getRedrawOrder() == 1) {
      (*it)->redrawAll(this);
    }
  }
  for(it = getScene()->firstData(); it!=getScene()->lastData(); it++) {
    if((*it)->getRedrawOrder() == 2) {
      (*it)->redrawAll(this);
    }
  }

  if(stencilBuffer)
  glDisable(GL_STENCIL_TEST);
 
  if (StereoFlag) {
    // if in stereo, draw left eye image
    
    // first, need to pop previously pushed transform matrix
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();
    
    // select the left buffer to draw in
    glDrawBuffer(GL_BACK_LEFT);
    
    // move the point of view to the left eye (it was at the right one)
    refCenter = oldRefCenter + delta;
    applyTransformation();
    /// set projection matrix
    setProjectionParameters();

    // reset the point of view back to the center
    refCenter = oldRefCenter;
    
    // and redraw the scene
ObjectsList< Data3D >::iterator it(getScene()->firstData());
  for(; it!=getScene()->lastData(); it++) {
    if((*it)->getRedrawOrder() == 1) {
      (*it)->redrawAll(this);
    }
  }
  for(it = getScene()->firstData(); it!=getScene()->lastData(); it++) {
    if((*it)->getRedrawOrder() == 2) {
      (*it)->redrawAll(this);
    }
  }
  }
  
  if(fog) glDisable(GL_FOG);

  if(moveMode == NAVIGATION_MODE)
    refCenter -= Vec3<double>(0.0, 0.0, 5.0 * radius);
  
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();

  count++; 
  if (showFrameRate) {
    printString(5,30,stringFrameRate,smallFont);
  }
  if (showCopyright)
    printString(5,5,stringCopyright,largeFont);

  // Compute display frequency every seconds
  if (globalTimer.getTime() > 1.0) {
    sprintf(stringFrameRate,"Frame Rate   : %.1f Hz (%.4f s/iter.)",
	    count/globalTimer.getTime(),globalTimer.getTime()/count);
    count = 0;
    globalTimer.reset();
  }

}
Light *Camera3D::newLight() {
	Light *l=new Light(getFreeLight());
	lights.push_back(l);
	return l;
}
Light *Camera3D::getLight(unsigned int n) {
	if (n<=lights.size()) {
		unsigned int i;
		std::list<Light *>::iterator itl;
		for (i=1,itl=getFirstLight();i!=n;++i,++itl);
		return (*itl);
	} else return 0;
}
void Camera3D::addLight(Light *l) {
  lights.push_back(l);
}
void Camera3D::removeLight(Light *l) {
	  lights.remove(l);
	  delete l;
}

ClippingPlane *Camera3D::newClippingPlane() {
	ClippingPlane *l=new ClippingPlane(getFreeClippingPlane(),refCenter.getZ());
	clipPlanes.push_back(l);
	return l;
}
ClippingPlane *Camera3D::getClippingPlane(unsigned int n) {
	if (n<=clipPlanes.size()) {
		unsigned int i;
		std::list<ClippingPlane *>::iterator itl;
		for (i=1,itl=getFirstClippingPlane();i!=n;++i,++itl);
		return (*itl);
	} else return 0;
}

void Camera3D::removeClippingPlane(ClippingPlane *l) {
	  clipPlanes.remove(l);
	  delete l;
}

GLenum Camera3D::getFreeClippingPlane(void) const {
  GLenum newplane;
  if(clipPlanes.empty()) newplane = GL_CLIP_PLANE0;
  else {
    ClippingPlane *clip = clipPlanes.back();
    newplane = GLenum (clip->getIndex() + 1);
  }
  return newplane;
}


GLenum Camera3D::getFreeLight(void) const {
  GLenum newlight;
  if(lights.empty()) newlight = GL_LIGHT0;
  else {
    Light *light = lights.back();
    newlight = (GLenum) (light->getIndex() + 1);
  }
  return newlight;
}





void Camera3D::rotateCameraObject(Rotation3D &rot) {
 if ((selectedObject) && (rotationFunction))
      (selectedObject->getData()->*rotationFunction)(rot);
  std::list<ClippingPlane *>::iterator itc=getFirstClippingPlane();
  for(;itc!=getLastClippingPlane();itc++) {
    (*itc)->rotate(rot, refCenter);
  }
}


void Camera3D::translateCameraObject(Vec3<double> &t) {
  if ((selectedObject) && (translationFunction))
      (selectedObject->getData()->*translationFunction)(t);

  std::list<ClippingPlane *>::iterator itc=getFirstClippingPlane();
  for(;itc!=getLastClippingPlane();itc++) {
    (*itc)->translate(t);
  }
}



void Camera3D::setFog(bool nfog) {
	if(this->fog != nfog) {
    this->fog = nfog;
  }
}

void Camera3D::setStencilBuffer(bool nstencil) {
  if(stencilBuffer!=nstencil){
    stencilBuffer=nstencil;
  }
}

void Camera3D::buildStencilBufferDisplayList() {
  if(stencilBufferDisplayList!=0){
    glDeleteLists(stencilBufferDisplayList,1);
  }
  const double scale = ((double)height)/width;
  stencilBufferDisplayList=glGenLists(1);
  glNewList(stencilBufferDisplayList,GL_COMPILE);

  glClearStencil(0x0);
  glEnable(GL_STENCIL_TEST);
  glClear(GL_STENCIL_BUFFER_BIT);
  glStencilFunc(GL_ALWAYS,0x1,0x1);
  glStencilOp(GL_REPLACE,GL_REPLACE,GL_REPLACE);
  glColorMask(0,0,0,0);
  glDepthMask(0);

  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glScaled(scale,1.0,0.0);

  GLUquadricObj * disk = gluNewQuadric();
  gluDisk(disk,0,1,40,1);
  gluDeleteQuadric(disk);

  glColorMask(1,1,1,1);
  glDepthMask(1); 

  glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
  glStencilFunc(GL_NOTEQUAL,0x1,0x1);

  glDisable(GL_LIGHTING);

  glColor3f(0.0,0.0,0.0);
  glBegin(GL_QUADS);
  glVertex3d(1/scale,1.0,0.0);
  glVertex3d(-1/scale,1.0,0.0);
  glVertex3d(-1/scale,-1.0,0.0);
  glVertex3d(1/scale,-1.0,0.0);
  glEnd();

  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();

  glEnable(GL_LIGHTING);

  glStencilFunc(GL_EQUAL,0x1,0x1);
  glDisable(GL_STENCIL_TEST);

  glEndList();
}



void Camera3D::setSpecularity(bool specularity) {
  if(useTrueSpecularity != specularity) {
    useTrueSpecularity = specularity;
    if(useTrueSpecularity) glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
    else glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_FALSE);
  }
}
bool Camera3D::getSpecularity() const {
	return useTrueSpecularity;
}

void Camera3D::setRenderFrontAndBack(bool flag) {
  if(useTwoSides != flag) {
    useTwoSides = flag;
    if(useTwoSides) glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
    else glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);
  }
}
std::list<Light *>::iterator Camera3D::getFirstLight() {
		return lights.begin();
}
std::list<Light *>::iterator Camera3D::getLastLight() {
		return lights.end();
}
std::list<ClippingPlane *>::iterator Camera3D::getFirstClippingPlane() {
		return clipPlanes.begin();
}
std::list<ClippingPlane *>::iterator Camera3D::getLastClippingPlane() {
		return clipPlanes.end();
}

bool Camera3D::getUseTwoSides() const 
{
	return useTwoSides;
}
Vec3<double> Camera3D::getEyeOffset(void) {
	return eyeOffset;}




} // namespace yav

