#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "graphics3D/Color.h"
#include <graphics3D/Color.h>


namespace yav {



const char *Color::synopsis = "[#RGB | #RRGGBB | #RRRGGGBBB | #RRRRGGGGBBBB | %f %f %f]";


Color::InvalidString::InvalidString(const std::string& from)
  : Exception(from,
	      "color string does not match #RGB, #RRGGBB, #RRRGGGBBB, #RRRRGGGGBBBB") {
}


Color::Color(const Material& mat) : str(0), mode(NONE) {
  const GLfloat *diffuse = mat.getDiffuse();

  color[0] = diffuse[0];
  color[1] = diffuse[1];
  color[2] = diffuse[2];
  color[3] = 1.0f;
}


Color::Color(const char *src) throw(Color::InvalidString)
  : str(0), mode(NONE) {
  if(!scan(src)) throw InvalidString("Color::Color(const char *)");
}


Color::Color(const Color& c) {
  color[0] = c.color[0];
  color[1] = c.color[1];
  color[2] = c.color[2];
  color[3] = c.color[3];
  mode = NONE;
  str = 0;
}


Color& Color::operator=(const Color& c) {
  color[0] = c.color[0];
  color[1] = c.color[1];
  color[2] = c.color[2];
  color[3] = c.color[3];
  mode = NONE;
  str = 0;
  return *this;
}


bool Color::scan(const char *src) {
  unsigned int r, g, b;
  int l;

  if(src[0] != '#') return false;
  l = strlen(src);
  switch(l) {
  case 4:
    mode = THREE_DIGITS;
    break;
  case 7:
    mode = SIX_DIGITS;
    break;
  case 10:
    mode = NINE_DIGITS;
    break;
  case 13:
    mode = TWELVE_DIGITS;
    break;
  default:
    return false;
  }
  switch(mode) {
    case NONE:
      break;
    case THREE_DIGITS:
      if(sscanf(src+1, "%1x", &r) != 1) return false;
      if(sscanf(src+2, "%1x", &g) != 1) return false;
      if(sscanf(src+3, "%1x", &b) != 1) return false;
      color[0] = (GLfloat)(r / 15.0);
      color[1] = (GLfloat)(g / 15.0);
      color[2] = (GLfloat)(b / 15.0);
      break;
    case SIX_DIGITS:
      if(sscanf(src+1, "%02x", &r) != 1) return false;
      if(sscanf(src+3, "%02x", &g) != 1) return false;
      if(sscanf(src+5, "%02x", &b) != 1) return false;
      color[0] = (GLfloat)(r / 255.0);
      color[1] = (GLfloat)(g / 255.0);
      color[2] = (GLfloat)(b / 255.0);
      break;
    case NINE_DIGITS:
      if(sscanf(src+1, "%03x", &r) != 1) return false;
      if(sscanf(src+4, "%03x", &g) != 1) return false;
      if(sscanf(src+7, "%03x", &b) != 1) return false;
      color[0] = (GLfloat)(r / 4095.0);
      color[1] = (GLfloat)(g / 4095.0);
      color[2] = (GLfloat)(b / 4095.0);
      break;
    case TWELVE_DIGITS:
      if(sscanf(src+1, "%04x", &r) != 1) return false;
      if(sscanf(src+5, "%04x", &g) != 1) return false;
      if(sscanf(src+9, "%04x", &b) != 1) return false;
      color[0] = (GLfloat)(r / 65535.0);
      color[1] = (GLfloat)(g / 65535.0);
      color[2] = (GLfloat)(b / 65535.0);
      break;
  }
  color[3] = 1.0f;
  if(!str) {
    str = new char[14];
    str[0] = '#';
  }
  strcpy(str, src);
  return true;
}


const char *Color::print(Color::StringMode _mode) {
  if(!str) {
    str = new char[14];
    str[0] = '#';
  }
  if(_mode != mode) {
    mode = _mode;
    switch(mode) {
    case NONE:
      break;
    case THREE_DIGITS:
      sprintf(str+1, "%X", (unsigned int) (color[0] * 15.0f + 0.5f));
      sprintf(str+2, "%X", (unsigned int) (color[1] * 15.0f + 0.5f));
      sprintf(str+3, "%X", (unsigned int) (color[2] * 15.0f + 0.5f));
      str[4] = '\0';
      break;
    case SIX_DIGITS:
      sprintf(str+1, "%02X", (unsigned int) (color[0] * 255.0f + 0.5f));
      sprintf(str+3, "%02X", (unsigned int) (color[1] * 255.0f + 0.5f));
      sprintf(str+5, "%02X", (unsigned int) (color[2] * 255.0f + 0.5f));
      str[7] = '\0';
      break;
    case NINE_DIGITS:
      sprintf(str+1, "%03X", (unsigned int) (color[0] * 4095.0f + 0.5f));
      sprintf(str+4, "%03X", (unsigned int) (color[1] * 4095.0f + 0.5f));
      sprintf(str+7, "%03X", (unsigned int) (color[2] * 4095.0f + 0.5f));
      str[11] = '\0';
      break;
    case TWELVE_DIGITS:
      sprintf(str+1, "%04X", (unsigned int) (color[0] * 65535.0f + 0.5f));
      sprintf(str+5, "%04X", (unsigned int) (color[1] * 65535.0f + 0.5f));
      sprintf(str+9, "%04X", (unsigned int) (color[2] * 65535.0f + 0.5f));
      str[13] = '\0';
      break;
    }
  }
  return str;
}


int Color::colorCmd(TclModule *m, Data3D *data, int argc, char **argv, int i) {
  float r, g, b;

  if(i < argc) {
    if(argv[i][0] == '#') {
      if(!scan(argv[i])) {
	m->setResult("invalid color \'%s\'", argv[i]);
	return TCL_ERROR;
      }
    }
    else {
      if(i + 2 >= argc) return COMMAND_SYNOPSIS;
      else {
	if(sscanf(argv[i], "%f", &r) != 1) return COMMAND_SYNOPSIS;
	if(sscanf(argv[i + 1], "%f", &g) != 1) return COMMAND_SYNOPSIS;
	if(sscanf(argv[i + 2], "%f", &b) != 1) return COMMAND_SYNOPSIS;
	setColor(r, g, b);
      }
    }

    i++;
    data->redrawScene();
  }

  print(SIX_DIGITS);
  m->setResult(str);
  return TCL_OK;
}

void Color::RGBtoHSV(Color & col){
  GLfloat r = getRed();
  GLfloat g = getGreen();
  GLfloat b = getBlue();
  float h, s, v ,Min, Max, delta;
  Min = min(min(r, g),b );
  Max = max(max(r, g),b );
  v = Max;				// v
  delta = Max - Min;
  if( Max != 0 )
    s = delta / Max;		// s
  else {
    // r = g = b = 0		// s = 0, v is undefined
    s = 0;
    h = -1;
    return;
  }
  if( r == Max )
    h = ( g - b ) / delta;		// between yellow & magenta
  else if( g == Max )
    h = 2 + ( b - r ) / delta;	// between cyan & yellow
  else
    h = 4 + ( r - g ) / delta;	// between magenta & cyan

  h *= 60;				// degrees
  if( h < 0 )
    h += 360;
  col.setColor(h,s,v, 1.0f);
}

void Color::HSVtoRGB(Color & col){
  GLfloat h = 360.0f*getRed();
  GLfloat s = getGreen();
  GLfloat v = getBlue();  
  int i;
  float f, p, q, t, r, g, b;

  if( s == 0 ) {
    // achromatic (grey)
    r = g = b = v;
    return;
  }

  h /= 60;			// sector 0 to 5
  i = (int)floor( h );
  f = h - i;			// factorial part of h
  p = v * ( 1 - s );
  q = v * ( 1 - s * f );
  t = v * ( 1 - s * ( 1 - f ) );
  switch( i ) {
  case 0:
    r = v;
    g = t;
    b = p;
    break;
  case 1:
    r = q;
    g = v;
    b = p;
    break;
  case 2:
    r = p;
    g = v;
    b = t;
    break;
  case 3:
    r = p;
    g = q;
    b = v;
    break;
  case 4:
    r = t;
    g = p;
    b = v;
    break;
  default:		// case 5:
    r = v;
    g = p;
    b = q;
    break;
  }
  col.setColor(r,g,b, 1.0f);
}


} // namespace yav

