#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "graphics3D/Geometry.h"
#include <graphics3D/Geometry.h>



namespace yav {


double Distance::epsilon = 1e-10;
double VectorialBase::epsilon = 1e-15;


VectorialBase::VectorialBase() {
  vx.setValue(1.0, 0.0, 0.0);
  vy.setValue(0.0, 1.0, 0.0);
  vz.setValue(0.0, 0.0, 1.0);
}


VectorialBase::VectorialBase(const Vec3<double>& v) {
  vz = v;
  if(fabs(vz.getZ()) >= epsilon) {
    vx[0] = 1.0;
    vx[1] = 1.0;
    vx[2] = (fabs(vz.getX()) < epsilon && fabs(vz.getY()) < epsilon)
      ? 1.0
      : -(vz.getX() * vx.getX() + vz.getY() * vx.getY()) / vz.getZ();
  }
  else {
    if(fabs(vz.getX()) < epsilon)
      vx = Vec3<double>(1.0, 0.0, 1.0);
    else {
      vx[1] = 1.0;
      vx[2] = 1.0;
      vx[0] = -(vz.getY() * vx.getY() - vz.getZ() * vx.getZ()) / vz.getX();
    }
  }
  vx /= vx.norm();
  vy = cross<double>(vz, vx);
}


void VectorialBase::normalize(void) {
  vx /= vx.norm();
  vy /= vy.norm();
  vz /= vz.norm();
}


double Distance::pointToLine(const Vec3<double>& org,
			     const Vec3<double>& dir,
			     const Vec3<double>& point) {
  double k = dot<double>(point - org, dir) / dir.norm2();
  return (point - org - k * dir).norm();
}


double Distance::segmentToLine(const Vec3<double>& org,
			       const Vec3<double>& dir,
			       const Vec3<double>& segA,
			       const Vec3<double>& segB) {
  if(segA == segB) return Distance::pointToLine(org, dir, segA);
  Vec3<double> OA(segA - org);
  Vec3<double> AB(segB - segA);
  double k = dir.norm2() - dot<double>(AB, dir - OA)
    * dot<double>(AB, dir) / AB.norm2();
  if(k < epsilon) return OA.norm();
  k = dot<double>(OA, dir) / k;
  Vec3<double> M(org + k * dir);
  double l = (k * dot<double>(AB, dir) - dot<double>(OA, AB)) / AB.norm2();
  if(l < 0.0) return (M - segA).norm();
  if(l > 1.0) return (M - segB).norm();
  return (M - segA - l * AB).norm();
}


double Distance::triangleToLine(const Vec3<double>& org,
				const Vec3<double>& dir,
				const Vec3<double>& trA,
				const Vec3<double>& trB,
				const Vec3<double>& trC) {
  return 0.0;
}




}


