#include <modules/TclObjectKind.h>
#include <modules/TclModule.h>
#include <graphics/SceneItem.h>
#include <graphics3D/TclScene3D.h>
#include <graphics3D/SceneCmd.h>
#include <graphics3D/TclCamera3D.h>
#include <graphics3D/TclClippingPlane.h>
#include <graphics3D/TclMaterialSphere.h>
#include <graphics3D/TclLightFantom.h>
#include <graphics3D/TclRectangle3D.h>
#include <graphics3D/TclTexture2D.h>
#include <graphics3D/TclTexture1D.h>
#include <graphics3D/Graphics3DCmd.h>


#ifdef EMBEDED_TCL_SCRIPTS
#include "tclGraphics3D.tcl.h" // generated by tcl2h
#endif


namespace yav {


const char *TclGraphics3DInit::tclGraphics3DName="tclGraphics3D";
static TclGraphics3DInit tclGraphics3DInit;
/*][************************************************************************/

void Tclgraphics3d_Init(TclModule* m) {
  TclObjectKind *kind;

  TclScene3D::createCommands(m);
#ifdef OPENGL
  TclCamera3D::createCommands(m);
  TclTexture1D::createCommands(m);
  TclTexture2D::createCommands(m);


  kind = new Scene3DItem(m, typeid(TclMaterialSphere), "sphere","Material Sphere",
			 &TclMaterialSphere::create, &TclMaterialSphere::list,
			 TclScene3D::NONE | TclScene3D::SCENE,
			 "[-scene scn3D%u]");
  m->newCommand("sphere%u", kind->getOptions(),
		"surface sphere management command");
  kind->newOption("-destroy", (TclObject::OPTION) &TclMaterialSphere::destroyCmd, 0,
		  "destroy a sphere");
  kind->newOption("-radius", (TclObject::OPTION) &TclMaterialSphere::radiusCmd, 0,
		  "get/set sphere radius");
  kind->newOption("-color", (TclObject::OPTION) &TclMaterialSphere::colorCmd,
		  Color::synopsis,
		  "get/set sphere wireframe color");
  kind->newOption("-material", (TclObject::OPTION) &TclMaterialSphere::materialCmd,
		  Material::synopsis,
		  "get/set sphere material");
  kind->newOption("-render", (TclObject::OPTION) &TclMaterialSphere::renderCmd,
		  Surface::synopsis,
		  "get/set sphere render mode");
  kind->newOption("-texture2D", (TclObject::OPTION) &TclMaterialSphere::texture2DCmd,
		  "[text2D%u]","get/set a given 2D texture");
  kind->newOption("-texture1D", (TclObject::OPTION) &TclMaterialSphere::texture1DCmd,
		  "[text2D%u]","get/set a given 1D texture");
  kind->newOption("-texture2DOption", (TclObject::OPTION) &TclMaterialSphere::texture2DOptionCmd,
		  Surface::texture2DSynopsis,
		  "get/set texture display properties");
  kind->newOption("-texture1DOption", (TclObject::OPTION) &TclMaterialSphere::texture1DOptionCmd,
		  Surface::texture1DSynopsis,
		  "get/set 1D texture display properties");
  kind->newOption("-highlightTexture", (TclObject::OPTION) &TclMaterialSphere::highlightTextureCmd,
		  "[text2D%u]",
		  "get/set the highlight texture");
  kind->newOption("-renderOption", (TclObject::OPTION) &TclMaterialSphere::renderOptionCmd,0,
		  " list sphere rendering modes");

  kind = new TclObjectKind(m, typeid(TclLightFantom), "fantom","Light Fantom",0,
			   &TclLightFantom::create, &TclLightFantom::list);
  kind->newOption("-destroy", (TclObject::OPTION) &TclLightFantom::destroyCmd,
		  0, "destroy a fantom");
  kind = new Scene3DItem(m, typeid(TclRectangle3D), "rect3D","3D Rectangle",
			 &TclRectangle3D::create, &TclRectangle3D::list,
			 TclScene3D::CAMERA,
			 "[cam3D%u | scn3D%u %f %f %f %f %f %f %f %f %f %f %f %f]");
  m->newCommand("rect3D%u", kind->getOptions(),
		"3D rectangle management command");
  kind->newOption("-destroy", (TclObject::OPTION) &TclRectangle3D::destroyCmd, 0,
		  "destroy a 3D rectangle");
  kind->newOption("-point", (TclObject::OPTION) &TclRectangle3D::pointCmd, "%u",
		  "outputs the coordinates of each of the 4 points");
  kind->newOption("-direction", (TclObject::OPTION) &TclRectangle3D::directionCmd, "X | Y",
		  "outputs the 4 line parameters of the 2 orthogonal directions of the rectangle");
  kind->newOption("-coordinate", (TclObject::OPTION) &TclRectangle3D::coordinateCmd, "X | Y",
		  "outputs the 4 line parameters of the 2 orthogonal directions of the rectangle");
  kind->newOption("-render", (TclObject::OPTION) &TclRectangle3D::renderCmd,
		  Surface::synopsis,
		  "get/set rectangle render mode"); 
  kind->newOption("-renderOption", (TclObject::OPTION) &TclRectangle3D::renderOptionCmd,0,
		  "get the list of possible rendering modes for a rectangle"); 
  kind->newOption("-scale", (TclObject::OPTION) &TclRectangle3D::scaleCmd,
		  "%f",
		  "scale the rectangle");
  kind->newOption("-color", (TclObject::OPTION) &TclRectangle3D::colorCmd,
		  Color::synopsis,
		  "get/set rectangle  wireframe color");
  kind->newOption("-flipNormals", (TclObject::OPTION) &TclRectangle3D::flipNormalsCmd,
		  0,"flip the surface normals for display");
  kind->newOption("-texture2D", (TclObject::OPTION) &TclRectangle3D::texture2DCmd,
		  "[text2D%u]","get/set a given 2D texture");
  kind->newOption("-texture1D", (TclObject::OPTION) &TclRectangle3D::texture1DCmd,
		  "[text2D%u]","get/set a given 1D texture");
  kind->newOption("-material", (TclObject::OPTION) &TclRectangle3D::materialCmd,
		  Material::synopsis,
		  "get/set rectangle material");
  kind->newOption("-textureScale", (TclObject::OPTION) &TclRectangle3D::textureScaleCmd,
		  "[%f %f]",
		  "get/set texture coordinates scale");

  kind->newOption("-texture2DOption", (TclObject::OPTION) &TclRectangle3D::texture2DOptionCmd,
		  Surface::texture2DSynopsis,
		  "get/set 2D texture display properties");
  kind->newOption("-texture1DOption", (TclObject::OPTION) &TclRectangle3D::texture1DOptionCmd,
		  Surface::texture1DSynopsis,
		  "get/set 1D texture display properties");
kind->newOption("-highlightTexture", (TclObject::OPTION) &TclRectangle3D::highlightTextureCmd,
		  "[text2D%u]",
		  "get/set the highlight texture");

 


#endif


#ifdef EMBEDED_TCL_SCRIPTS
 /* Statically loads the tcl script from tclModules.tcl.h (generated by tcl2h) */
 Tcl_Eval(m->getInterp(), tclGraphics3D_tcl);
 /* Tell the TclModule constructor not to load the script again */
 m->tclScriptLoaded();
#endif
 /* Otherwise, the tcl script will be loaded dynamically from tclModules.tcl by the TclModule constructor */


}

} // namespace yav
