#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "graphics3D/Light.h"
#include <graphics3D/Light.h>
#include <graphics3D/Camera3D.h>
#include <graphics3D/LightFantom.h>


namespace yav {


Light::Light( GLenum _ind)
  : fantom(0),  index(_ind), active(true) {
  ambient[0] = 0.1f;
  ambient[1] = 0.1f;
  ambient[2] = 0.1f;
  ambient[3] = 0.0f;
  diffuse[0] = 1.0f;
  diffuse[1] = 1.0f;
  diffuse[2] = 1.0f;
  diffuse[3] = 1.0f;
  specular[0] = 1.0f;
  specular[1] = 1.0f;
  specular[2] = 1.0f;
  specular[3] = 1.0f;

  position[0] = 0.0;
  position[1] = 0.0;
  position[2] = 1.0;
  position[3] = 0.0;
  
  direction[0] = 0.0;
  direction[1] = 0.0;
  direction[2] = -1.0;
  exponent = 0.0;
  cutoff = 180.0;
  attenuation[0] = 1.0;
  attenuation[1] = 0.0;
  attenuation[2] = 0.0;
  //cam->lights.push_back(this);
  glSet();
  glEnable(index);
}


Light::Light(GLenum ind, const GLfloat *ambient, const GLfloat *diffuse,
	     const GLfloat *specular, const GLfloat *position,
	     const GLfloat *direction, GLfloat exponent,
	     GLfloat cutoff, const GLfloat *attenuation)
  : fantom(0), index(ind), active(true) {
  setLight(ambient, diffuse, specular, position, direction,
	   exponent, cutoff, attenuation);
  glSet();
  glEnable(index);
}


void Light::setLight(const GLfloat *_ambient, const GLfloat *_diffuse,
		     const GLfloat *_specular, const GLfloat *_position,
		     const GLfloat *_direction, GLfloat _exponent,
		     GLfloat _cutoff,
		     const GLfloat *_attenuation) {
  ambient[0] = _ambient[0];
  ambient[1] = _ambient[1];
  ambient[2] = _ambient[2];
  ambient[3] = _ambient[3];
  diffuse[0] = _diffuse[0];
  diffuse[1] = _diffuse[1];
  diffuse[2] = _diffuse[2];
  diffuse[3] = _diffuse[3];
  specular[0] = _specular[0];
  specular[1] = _specular[1];
  specular[2] = _specular[2];
  specular[3] = _specular[3];

  position[0] = _position[0];
  position[1] = _position[1];
  position[2] = _position[2];
  position[3] = _position[3];
  
  if(_direction) {
    direction[0] = _direction[0];
    direction[1] = _direction[1];
    direction[2] = _direction[2];
  }
  else {
    direction[0] = 0.0;
    direction[1] = 0.0;
    direction[2] = 1.0;
  }

  exponent = _exponent;
  cutoff = _cutoff;
  if(_attenuation) {
    attenuation[0] = _attenuation[0];
    attenuation[1] = _attenuation[1];
    attenuation[2] = _attenuation[2];
  }
  else {
    attenuation[0] = 1.0;
    attenuation[1] = 0.0;
    attenuation[2] = 0.0;
  }

  glSet();
}


Light::~Light() {
 glDisable(getIndex()); 
}

/// returns light ambient coefficients
const GLfloat *Light::getAmbient(void) const { return ambient; }
/// returns light diffuse coefficients
const GLfloat *Light::getDiffuse(void) const { return diffuse; }
/// returns light specular coefficients
const GLfloat *Light::getSpecular(void) const { return specular; }
/// returns light position
const GLfloat *Light::getPosition(void) const { return position; }
/// returns spotlight direction
const GLfloat *Light::getDirection(void) const { return direction; }
/// returns spotlight exponent
GLfloat Light::getExponent(void) const { return exponent; }
/// returns spotlight cutoff (in degrees)
GLfloat Light::getCutoff(void) const { return cutoff; }
/// returns light attenuation factors (constant, linear and quadratic)
const GLfloat *Light::getAttenuation(void) const { return attenuation; }
void Light::setAmbient(const GLfloat *_ambient) {
	ambient[0] = _ambient[0];
  ambient[1] = _ambient[1];
  ambient[2] = _ambient[2];
  ambient[3] = _ambient[3];
  glSet();
}


void Light::setDiffuse(const GLfloat *_diffuse) {
  diffuse[0] = _diffuse[0];
  diffuse[1] = _diffuse[1];
  diffuse[2] = _diffuse[2];
  diffuse[3] = _diffuse[3];
  glSet();
  if(fantom) fantom->setColor();
}



void Light::setSpecular(const GLfloat *_specular) {
  specular[0] = _specular[0];
  specular[1] = _specular[1];
  specular[2] = _specular[2];
  specular[3] = _specular[3];
  glSet();
}



void Light::setPosition(const GLfloat *_position) {
  if(fantom && _position[3] == 0) delete fantom;
  position[0] = _position[0];
  position[1] = _position[1];
  position[2] = _position[2];
  position[3] = _position[3];
  glSet();
  if(fantom) fantom->move();
}



void Light::setDirection(const GLfloat *_direction) {
  direction[0] = _direction[0];
  direction[1] = _direction[1];
  direction[2] = _direction[2];
  glSet();
  if(fantom) fantom->move();
}



void Light::setExponent(const GLfloat _exponent) {
  exponent = _exponent;
  glSet();
}



void Light::setCutoff(const GLfloat _cutoff) {
  if((_cutoff >= 0.0 && _cutoff <= 90.0) || _cutoff == 180.0) {
    cutoff = _cutoff;
    glSet();
  }
}



void Light::setAttenuation(const GLfloat *_attenuation) {
  attenuation[0] = _attenuation[0];
  attenuation[1] = _attenuation[1];
  attenuation[2] = _attenuation[2];
  glSet();
}


void Light::enable(void) {
  if(!active) {
    active = true;
    glEnable(index);
  }
}


void Light::disable(void) {
  if(active) {
    active = false;
    glDisable(index);
  }
}


void Light::glSet(void) {
  glLightfv(index, GL_AMBIENT, ambient);
  glLightfv(index, GL_DIFFUSE, diffuse);
  glLightfv(index, GL_SPECULAR, specular);
  glLightfv(index, GL_POSITION, position);
  glLightfv(index, GL_SPOT_DIRECTION, direction);
  glLightf(index, GL_SPOT_EXPONENT, exponent);
  glLightf(index, GL_SPOT_CUTOFF, cutoff);
  glLightf(index, GL_CONSTANT_ATTENUATION, attenuation[0]);
  glLightf(index, GL_LINEAR_ATTENUATION, attenuation[1]);
  glLightf(index, GL_QUADRATIC_ATTENUATION, attenuation[2]);
}



void Light::putInCamera(void) {
  if(position[3] > 0.0) {
    glLightfv(index, GL_POSITION, position);
    glLightfv(index, GL_SPOT_DIRECTION, direction);
  }
}
void  Light::setFantom(LightFantom *f) {
	fantom=f;
}

/*void Light::displayFantom(bool disp, Camera3D *cam) {
  if(disp) {
    if(!fantom && position[3] > 0.0) fantom = new LightFantom(this,cam);
  }
  else {
    if(fantom) delete fantom;
  }
}
*/



} // namespace yav

