#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "graphics3D/LightFantom.h"
#include <graphics3D/LightFantom.h>

namespace yav {


LightFantom::LightFantom(Light *_light, Camera3D *cam)
  : MaterialSphere(Vec3<double>(_light->position[0], _light->position[1],
				_light->position[2]),
		   cam->getRadius() / 10.0,
		   cam->getScene()), light(_light), camera(cam) {
  setColor();
  setRendering("Gouraud");
}


LightFantom::~LightFantom() {
  light->fantom = 0;
}


void LightFantom::buildWireframe(void) {
  buildSolid();
}


void LightFantom::buildSolid(void) {
  Vec3<double> dir(light->direction[0], light->direction[1],
		   light->direction[2]);
  dir /= dir.norm();
  dir *= 2 * Sphere::radius;
  dir += Data3D::center;

  MaterialSphere::buildSolid();
  color.glSet();
  glLineWidth(3.0);
  glBegin(GL_LINES);
  glVertex3dv((const double *) Data3D::center);
  glVertex3dv((const double *) dir);
  glEnd();
}


void LightFantom::move(void) {
  setCenter(Vec3<double>(light->position[0], light->position[1],
			 light->position[2]));
  if(camera->getSelectedObject() == (Data3D *) this) {
    initBoundingBox();
  }
}


void LightFantom::setColor() {
  const GLfloat *emission = light->getDiffuse();
  color.setColor(emission[0], emission[1], emission[2]);
  mat.setAmbient(emission);
  mat.setDiffuse(emission);
  mat.setSpecular(emission);
  mat.setEmission(emission);
  mat.setLight(true);
  surfaceChanged();
}


void LightFantom::translate(const Vec3<double>& tr) {
  MaterialSphere::translate(tr);
  GLfloat pos[4];
  pos[0] = (GLfloat)(light->position[0] + tr.getX());
  pos[1] = (GLfloat)(light->position[1] + tr.getY());
  pos[2] = (GLfloat)(light->position[2] + tr.getZ());
  pos[3] = (GLfloat)light->position[3];
  light->setPosition(pos);
}


void LightFantom::rotate(const Rotation3D& rotation) {
  Vec3<double> n = rotation * Vec3<double>(light->direction[0],
					   light->direction[1],
					   light->direction[2]);
  GLfloat dir[3];
  dir[0] = (GLfloat) n.getX();
  dir[1] = (GLfloat)n.getY();
  dir[2] = (GLfloat) n.getZ();
  light->setDirection(dir);
  surfaceChanged();
}



} // namespace yav

