#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "graphics3D/Material.h"
#include <graphics3D/Material.h>
#include <graphics3D/TclScene3D.h>


namespace yav {


const char *Material::synopsis = "[%f %f %f %f %f %f %f %f %f %f %f %f %f | %f %f %f %f %f %f %f %f %f %f | #RGB | #RRGGBB | #RRRGGGBBB | #RRRRGGGGBBBB | [ambient %f %f %f [%f]] [diffuse %f %f %f [%f]] [specular %f %f %f [%f]] [shininess %f]] [emission %b %f %f %f [%f]]]";


Material::Material() : list(0), lighted(false), transparent(false) {
  ambient[0] = 0.3f;
  ambient[1] = 0.3f;
  ambient[2] = 0.3f;
  ambient[3] = 1.0f;
  diffuse[0] = 1.0f;
  diffuse[1] = 1.0f;
  diffuse[2] = 1.0f;
  diffuse[3] = 1.0f;
  specular[0] = 0.2f;
  specular[1] = 0.2f;
  specular[2] = 0.2f;
  specular[3] = 1.0f;
  shininess = 30.0f;
  emission[0] = 0.0;
  emission[1] = 0.0;
  emission[2] = 0.0;
  emission[3] = 1.0;

  ambientBack[0] = 0.3f;
  ambientBack[1] = 0.3f;
  ambientBack[2] = 0.3f;
  ambientBack[3] = 1.0f;
  diffuseBack[0] = 1.0f;
  diffuseBack[1] = 1.0f;
  diffuseBack[2] = 1.0f;
  diffuseBack[3] = 1.0f;
  specularBack[0] = 0.2f;
  specularBack[1] = 0.2f;
  specularBack[2] = 0.2f;
  specularBack[3] = 1.0f;
  shininessBack = 30.0f;
  emissionBack[0] = 0.0;
  emissionBack[1] = 0.0;
  emissionBack[2] = 0.0;
  emissionBack[3] = 1.0;

}


Material::Material(const GLfloat *ambient, const GLfloat *diffuse,
		   const GLfloat *specular, GLfloat shininess) : list(0), transparent(false) {
  setMaterial(ambient, diffuse, specular, shininess);
  setBackMaterial(ambient, diffuse, specular, shininess);
}


Material::Material(const Color& color) : list(0), transparent(false) {
  setMaterial(color);
  setBackMaterial(color);
}


Material::~Material() {
  if(list > 0) glDeleteLists(list, 1);
}


void Material::setMaterial(const GLfloat *_ambient, const GLfloat *_diffuse,
			   const GLfloat *_specular, GLfloat _shininess) {
  for(int i = 0; i < 4; i++) {
    ambient[i] = _ambient[i] < 0.0f ? 0.0f : (_ambient[i] > 1.0f ? 1.0f : _ambient[i]);
    diffuse[i] = _diffuse[i] < 0.0f ? 0.0f : (_diffuse[i] > 1.0f ? 1.0f : _diffuse[i]);
    specular[i] = _specular[i] < 0.0f ? 0.0f : (_specular[i] > 1.0f ? 1.0f : _specular[i]);
    shininess = _shininess < 0.0f ? 0.0f : _shininess;
  }
  if ((ambient[3]<1.0)|| (diffuse[3]<1.0) || (specular[3]<1.0)|| (emission[3]<1.0))
	  transparent=true;
  else
	  transparent=false;
  if(list > 0) {
    glDeleteLists(list, 1);
    list = 0;
  }
}
void Material::setBackMaterial(const GLfloat *_ambient, const GLfloat *_diffuse,
			   const GLfloat *_specular, GLfloat _shininess) {
  for(int i = 0; i < 4; i++) {
    ambientBack[i] = _ambient[i] < 0.0f ? 0.0f : (_ambient[i] > 1.0f ? 1.0f : _ambient[i]);
    diffuseBack[i] = _diffuse[i] < 0.0f ? 0.0f : (_diffuse[i] > 1.0f ? 1.0f : _diffuse[i]);
    specularBack[i] = _specular[i] < 0.0f ? 0.0f : (_specular[i] > 1.0f ? 1.0f : _specular[i]);
    shininessBack = _shininess < 0.0f ? 0.0f : _shininess;
  }
  if ((ambient[3]<1.0)|| (diffuse[3]<1.0) || (specular[3]<1.0)|| (emission[3]<1.0))
	  transparent=true;
  else
	  transparent=false;
  if(list > 0) {
    glDeleteLists(list, 1);
    list = 0;
  }
}


void Material::setAmbient(const GLfloat *_ambient) {
  for(int i = 0; i < 4; i++) {
    ambient[i] = _ambient[i] < 0.0f ? 0.0f : (_ambient[i] > 1.0f ? 1.0f : _ambient[i]);
  }
 if ((ambient[3]<1.0)|| (diffuse[3]<1.0) || (specular[3]<1.0)|| (emission[3]<1.0))
	  transparent=true;
  else
	  transparent=false;
 
  if(list > 0) {
    glDeleteLists(list, 1);
    list = 0;
  }
}


void Material::setBackAmbient(const GLfloat *_ambient) {
  for(int i = 0; i < 4; i++) {
    ambientBack[i] = _ambient[i] < 0.0f ? 0.0f : (_ambient[i] > 1.0f ? 1.0f : _ambient[i]);
  }
   if ((ambient[3]<1.0)|| (diffuse[3]<1.0) || (specular[3]<1.0)|| (emission[3]<1.0))
	  transparent=true;
  else
	  transparent=false;
  if(list > 0) {
    glDeleteLists(list, 1);
    list = 0;
  }
}


void Material::setDiffuse(const GLfloat *_diffuse) {
  for(int i = 0; i < 4; i++) {
    diffuse[i] = _diffuse[i] < 0.0f ? 0.0f : (_diffuse[i] > 1.0f ? 1.0f : _diffuse[i]);
  }
  if ((ambient[3]<1.0)|| (diffuse[3]<1.0) || (specular[3]<1.0)|| (emission[3]<1.0))
	  transparent=true;
  else
	  transparent=false;
  if(list > 0) {
    glDeleteLists(list, 1);
    list = 0;
  }
}

void Material::setBackDiffuse(const GLfloat *_diffuse) {
  for(int i = 0; i < 4; i++) {
    diffuseBack[i] = _diffuse[i] < 0.0f ? 0.0f : (_diffuse[i] > 1.0f ? 1.0f : _diffuse[i]);
  }
  if ((ambient[3]<1.0)|| (diffuse[3]<1.0) || (specular[3]<1.0)|| (emission[3]<1.0))
	  transparent=true;
  else
	  transparent=false;  if(list > 0) {
    glDeleteLists(list, 1);
    list = 0;
  }
}


void Material::setSpecular(const GLfloat *_specular) {
  for(int i = 0; i < 4; i++) {
    specular[i] = _specular[i] < 0.0f ? 0.0f : (_specular[i] > 1.0f ? 1.0f : _specular[i]);
  }
  if ((ambient[3]<1.0)|| (diffuse[3]<1.0) || (specular[3]<1.0)|| (emission[3]<1.0))
	  transparent=true;
  else
	  transparent=false;
  if(list > 0) {
    glDeleteLists(list, 1);
    list = 0;
  }
}

void Material::setBackSpecular(const GLfloat *_specular) {
  for(int i = 0; i < 4; i++) {
    specularBack[i] = _specular[i] < 0.0f ? 0.0f : (_specular[i] > 1.0f ? 1.0f : _specular[i]);
  }
  if ((ambient[3]<1.0)|| (diffuse[3]<1.0) || (specular[3]<1.0)|| (emission[3]<1.0))
	  transparent=true;
  else
	  transparent=false;
  if(list > 0) {
    glDeleteLists(list, 1);
    list = 0;
  }
}


void Material::setShininess(const GLfloat _shininess) {
  shininess = _shininess < 0.0f ? 0.0f : _shininess;
  if(list > 0) {
    glDeleteLists(list, 1);
    list = 0;
  }
}

void Material::setBackShininess(const GLfloat _shininess) {
  shininessBack = _shininess < 0.0f ? 0.0f : _shininess;
  if(list > 0) {
    glDeleteLists(list, 1);
    list = 0;
  }
}


void Material::setEmission(const GLfloat *_emission) {
  for(int i = 0; i < 4; i++) {
    emission[i] = _emission[i] < 0.0f ? 0.0f : (_emission[i] > 1.0f ? 1.0f : _emission[i]);
  }
  if ((ambient[3]<1.0)|| (diffuse[3]<1.0) || (specular[3]<1.0) || (emission[3]<1.0))
	  transparent=true;
  else
	  transparent=false;
  if(list > 0) {
    glDeleteLists(list, 1);
    list = 0;
  }
}


void Material::setBackEmission(const GLfloat *_emission) {
  for(int i = 0; i < 4; i++) {
    emissionBack[i] = _emission[i] < 0.0f ? 0.0f : (_emission[i] > 1.0f ? 1.0f : _emission[i]);
  }
   if ((ambient[3]<1.0)|| (diffuse[3]<1.0) || (specular[3]<1.0) || (emission[3]<1.0))
	  transparent=true;
  else
	  transparent=false;
  if(list > 0) {
    glDeleteLists(list, 1);
    list = 0;
  }
}


void Material::setLight(bool l) {
  if(l != lighted) {
    lighted = l;
    if(list > 0) {
      glDeleteLists(list, 1);
      list = 0;
    }
  }
}


void Material::setMaterial(const Color& color) {
	const GLfloat *col = color.getColor();
	
	ambient[0] = ambient[1] = ambient[2] = 0.2f;
	ambient[3] = 1.0f;
	diffuse[0] = col[0];
	diffuse[1] = col[1];
	diffuse[2] = col[2];
	diffuse[3] = 1.0f;  
	specular[0] = 0.3f;
	specular[1] = 0.3f;
	specular[2] = 0.3f;
	specular[3] = 1.0f;
	shininess = 30.0f;
	transparent=false;
}

void Material::setBackMaterial(const Color& color) {
  const GLfloat *col = color.getColor();

  ambientBack[0] = ambientBack[1] = ambientBack[2] = 0.2f;
  ambientBack[3] = 1.0f;
  diffuseBack[0] = col[0];
  diffuseBack[1] = col[1];
  diffuseBack[2] = col[2];
  diffuseBack[1] = 1.0f;  
  specularBack[0] = 0.3f;
  specularBack[1] = 0.3f;
  specularBack[2] = 0.3f;
  specularBack[1] = 1.0f;
  shininessBack = 30.0f;
  transparent=false;
}


void Material::glSet(void) {
  if(list == 0) {
    float nullEmission[4] = { 0.f, 0.f, 0.f, 0.f };

    list = glGenLists(1);
    glNewList(list, GL_COMPILE);
    glMaterialfv(GL_FRONT, GL_AMBIENT, ambient);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, diffuse);
    glMaterialfv(GL_FRONT, GL_SPECULAR, specular);
    glMaterialfv(GL_FRONT, GL_SHININESS, &shininess);
    if(lighted) glMaterialfv(GL_FRONT, GL_EMISSION, emission);
    else glMaterialfv(GL_FRONT, GL_EMISSION, nullEmission);

    glMaterialfv(GL_BACK, GL_AMBIENT, ambientBack);
    glMaterialfv(GL_BACK, GL_DIFFUSE, diffuseBack);
    glMaterialfv(GL_BACK, GL_SPECULAR, specularBack);
    glMaterialfv(GL_BACK, GL_SHININESS, &shininessBack);
    if(lighted) glMaterialfv(GL_BACK, GL_EMISSION, emissionBack);
    else glMaterialfv(GL_BACK, GL_EMISSION, nullEmission);
    glEndList();
  }
  glCallList(list);
}



void Material::setAlpha(double alpha) {
  if(alpha >= 0.0 && alpha <= 1.0) {
    ambient[3] = (float)alpha;
    diffuse[3] = (float)alpha;
    specular[3] = (float)alpha;
    emission[3] = (float)alpha;
	if (alpha==1.0)
		transparent=false;
	else
		transparent=true;
    if(list > 0) {
      glDeleteLists(list, 1);
      list = 0;
    }
  }
}

void Material::setBackAlpha(double alpha) {
  if(alpha >= 0.0 && alpha <= 1.0) {
    ambientBack[3] = (float)alpha;
    diffuseBack[3] = (float)alpha;
    specularBack[3] = (float)alpha;
    emissionBack[3] = (float)alpha;
   if (alpha==1.0)
		transparent=false;
	else
		transparent=true;
	if(list > 0) {
      glDeleteLists(list, 1);
      list = 0;
    }
  }
}


int Material::materialCmd(TclModule *m, Data3D *data, int argc, char **argv,
			  int i) {
	
	if(i < argc) {
		if(argv[0][0] == '#') {
			try {
				Color color(argv[0]);
				setMaterial(color);
			}
			catch(Color::InvalidString& e) {
				m->setResult(e.getError().c_str());
				return TCL_ERROR;
			}
		}
		else if(!strcmp(argv[0], "ambient") || !strcmp(argv[0], "diffuse")
			|| !strcmp(argv[0], "specular") || !strcmp(argv[0], "shininess")
			|| !strcmp(argv[0], "emission")) {
			GLfloat coef[4];
			bool b;
			
			for(; i < argc; i++) {
				if(!strcmp(argv[i], "ambient")) {
					if(i+3 >= argc) return COMMAND_SYNOPSIS;
					if(sscanf(argv[++i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[++i], "%f", &coef[1]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[++i], "%f", &coef[2]) != 1) return COMMAND_SYNOPSIS;
					if(i+1 >= argc || !strcmp(argv[i+1], "diffuse")
						|| !strcmp(argv[i+1], "specular")
						|| !strcmp(argv[i+1], "shininess")
						|| !strcmp(argv[i+1], "emission")) coef[3] = 1.0;
					else {
						if(sscanf(argv[++i], "%f", &coef[3]) != 1) return COMMAND_SYNOPSIS;
					}
					setAmbient(coef);
				}
				else if(!strcmp(argv[i], "diffuse")) {
					if(i+3 >= argc) return COMMAND_SYNOPSIS;
					if(sscanf(argv[++i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[++i], "%f", &coef[1]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[++i], "%f", &coef[2]) != 1) return COMMAND_SYNOPSIS;
					if(i+1 >= argc || !strcmp(argv[i+1], "ambient")
						|| !strcmp(argv[i+1], "specular")
						|| !strcmp(argv[i+1], "shininess")
						|| !strcmp(argv[i+1], "emission")) coef[3] = 1.0;
					else {
						if(sscanf(argv[++i], "%f", &coef[3]) != 1) return COMMAND_SYNOPSIS;
					}
					setDiffuse(coef);
				}
				
				else if(!strcmp(argv[i], "specular")) {
					if(i+3 >= argc) return COMMAND_SYNOPSIS;
					if(sscanf(argv[++i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[++i], "%f", &coef[1]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[++i], "%f", &coef[2]) != 1) return COMMAND_SYNOPSIS;
					if(i+1 >= argc || !strcmp(argv[i+1], "ambient")
						|| !strcmp(argv[i+1], "diffuse")
						|| !strcmp(argv[i+1], "shininess")
						|| !strcmp(argv[i+1], "emission")) coef[3] = 1.0;
					else {
						if(sscanf(argv[++i], "%f", &coef[3]) != 1) return COMMAND_SYNOPSIS;
					}
					setSpecular(coef);
				}
				
				else if(!strcmp(argv[i], "shininess")) {
					if(sscanf(argv[++i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
					setShininess(coef[0]);
				}
				
				else if(!strcmp(argv[i], "emission")) {
					if(i+4 >= argc) return COMMAND_SYNOPSIS;
					if(!TclModule::boolValue(argv[++i], b)) return COMMAND_SYNOPSIS;
					if(sscanf(argv[++i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[++i], "%f", &coef[1]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[++i], "%f", &coef[2]) != 1) return COMMAND_SYNOPSIS;
					if(i+1 >= argc || !strcmp(argv[i+1], "ambient")
						|| !strcmp(argv[i+1], "diffuse")
						|| !strcmp(argv[i+1], "specular")
						|| !strcmp(argv[i+1], "shininess")) coef[3] = 1.0;
					else {
						if(sscanf(argv[++i], "%f", &coef[3]) != 1) return COMMAND_SYNOPSIS;
					}
					setEmission(coef);
					setLight(b);
				}
				
			}
		}
		
		else {
			GLfloat _ambient[4], _diffuse[4], _specular[4], _shininess;
			
			if(i + 13 < argc) {
				if(sscanf(argv[0], "%f", &_ambient[0]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[1], "%f", &_ambient[1]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[2], "%f", &_ambient[2]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[3], "%f", &_ambient[3]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[4], "%f", &_diffuse[0]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[5], "%f", &_diffuse[1]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[6], "%f", &_diffuse[2]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[7], "%f", &_diffuse[3]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[8], "%f", &_specular[0]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[9], "%f", &_specular[1]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[10], "%f", &_specular[2]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[11], "%f", &_specular[3]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[12], "%f", &_shininess) != 1) return COMMAND_SYNOPSIS;
				setMaterial(_ambient, _diffuse, _specular, _shininess);
			}
			else if(i + 10 < argc) {
				if(sscanf(argv[0], "%f", &_ambient[0]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[1], "%f", &_ambient[1]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[2], "%f", &_ambient[2]) != 1) return COMMAND_SYNOPSIS;
				ambient[3] = 1.0f;
				if(sscanf(argv[4], "%f", &_diffuse[0]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[5], "%f", &_diffuse[1]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[6], "%f", &_diffuse[2]) != 1) return COMMAND_SYNOPSIS;
				diffuse[3] = 1.0f;
				if(sscanf(argv[8], "%f", &_specular[0]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[9], "%f", &_specular[1]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[10], "%f", &_specular[2]) != 1) return COMMAND_SYNOPSIS;
				specular[3] = 1.0f;
				if(sscanf(argv[12], "%f", &_shininess) != 1) return COMMAND_SYNOPSIS;
				setMaterial(_ambient, _diffuse, _specular, _shininess);
			}
			else return COMMAND_SYNOPSIS;
		}
	
  }
  	((TclScene3D *)(data->getScene()->getInterface()))->redraw();
  m->setResult("ambient %f %f %f %f diffuse %f %f %f %f specular %f %f %f %f shininess %f emission %s %f %f %f %f",
	  ambient[0], ambient[1], ambient[2], ambient[3],
	  diffuse[0], diffuse[1], diffuse[2], diffuse[3],
	  specular[0], specular[1], specular[2], specular[3],
	  shininess, TclModule::boolString(lighted),
	  emission[0], emission[1], emission[2], emission[3]);
  return TCL_OK;
}



int Material::materialBackCmd(TclModule *m, Data3D *data, int argc, char **argv,
							  int i) {
	if(i < argc) {
		if(argv[0][0] == '#') {
			try {
				Color color(argv[0]);
				setBackMaterial(color);
			}
			catch(Color::InvalidString& e) {
				m->setResult(e.getError().c_str());
				return TCL_ERROR;
			}
		}
		else if(!strcmp(argv[0], "ambient") || !strcmp(argv[0], "diffuse")
			|| !strcmp(argv[0], "specular") || !strcmp(argv[0], "shininess")
			|| !strcmp(argv[0], "emission")) {
			GLfloat coef[4];
			bool b;
			
			for(; i < argc; i++) {
				if(!strcmp(argv[i], "ambient")) {
					if(i+3 >= argc) return COMMAND_SYNOPSIS;
					if(sscanf(argv[++i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[++i], "%f", &coef[1]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[++i], "%f", &coef[2]) != 1) return COMMAND_SYNOPSIS;
					if(i+1 >= argc || !strcmp(argv[i+1], "diffuse")
						|| !strcmp(argv[i+1], "specular")
						|| !strcmp(argv[i+1], "shininess")
						|| !strcmp(argv[i+1], "emission")) coef[3] = 1.0;
					else {
						if(sscanf(argv[++i], "%f", &coef[3]) != 1) return COMMAND_SYNOPSIS;
					}
					setBackAmbient(coef);
				}
				else if(!strcmp(argv[i], "diffuse")) {
					if(i+3 >= argc) return COMMAND_SYNOPSIS;
					if(sscanf(argv[++i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[++i], "%f", &coef[1]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[++i], "%f", &coef[2]) != 1) return COMMAND_SYNOPSIS;
					if(i+1 >= argc || !strcmp(argv[i+1], "ambient")
						|| !strcmp(argv[i+1], "specular")
						|| !strcmp(argv[i+1], "shininess")
						|| !strcmp(argv[i+1], "emission")) coef[3] = 1.0;
					else {
						if(sscanf(argv[++i], "%f", &coef[3]) != 1) return COMMAND_SYNOPSIS;
					}
					setBackDiffuse(coef);
				}
				
				else if(!strcmp(argv[i], "specular")) {
					if(i+3 >= argc) return COMMAND_SYNOPSIS;
					if(sscanf(argv[++i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[++i], "%f", &coef[1]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[++i], "%f", &coef[2]) != 1) return COMMAND_SYNOPSIS;
					if(i+1 >= argc || !strcmp(argv[i+1], "ambient")
						|| !strcmp(argv[i+1], "diffuse")
						|| !strcmp(argv[i+1], "shininess")
						|| !strcmp(argv[i+1], "emission")) coef[3] = 1.0;
					else {
						if(sscanf(argv[++i], "%f", &coef[3]) != 1) return COMMAND_SYNOPSIS;
					}
					setBackSpecular(coef);
				}
				
				else if(!strcmp(argv[i], "shininess")) {
					if(sscanf(argv[++i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
					setBackShininess(coef[0]);
				}
				
				else if(!strcmp(argv[i], "emission")) {
					if(i+4 >= argc) return COMMAND_SYNOPSIS;
					if(!TclModule::boolValue(argv[++i], b)) return COMMAND_SYNOPSIS;
					if(sscanf(argv[++i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[++i], "%f", &coef[1]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[++i], "%f", &coef[2]) != 1) return COMMAND_SYNOPSIS;
					if(i+1 >= argc || !strcmp(argv[i+1], "ambient")
						|| !strcmp(argv[i+1], "diffuse")
						|| !strcmp(argv[i+1], "specular")
						|| !strcmp(argv[i+1], "shininess")) coef[3] = 1.0;
					else {
						if(sscanf(argv[++i], "%f", &coef[3]) != 1) return COMMAND_SYNOPSIS;
					}
					setBackEmission(coef);
					setLight(b);
				}
				
			}
		}
		
		else {
			GLfloat _ambient[4], _diffuse[4], _specular[4], _shininess;
			
			if(i + 13 < argc) {
				if(sscanf(argv[0], "%f", &_ambient[0]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[1], "%f", &_ambient[1]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[2], "%f", &_ambient[2]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[3], "%f", &_ambient[3]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[4], "%f", &_diffuse[0]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[5], "%f", &_diffuse[1]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[6], "%f", &_diffuse[2]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[7], "%f", &_diffuse[3]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[8], "%f", &_specular[0]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[9], "%f", &_specular[1]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[10], "%f", &_specular[2]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[11], "%f", &_specular[3]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[12], "%f", &_shininess) != 1) return COMMAND_SYNOPSIS;
				setBackMaterial(_ambient, _diffuse, _specular, _shininess);
			}
			else if(i + 10 < argc) {
				if(sscanf(argv[0], "%f", &_ambient[0]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[1], "%f", &_ambient[1]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[2], "%f", &_ambient[2]) != 1) return COMMAND_SYNOPSIS;
				ambient[3] = 1.0f;
				if(sscanf(argv[4], "%f", &_diffuse[0]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[5], "%f", &_diffuse[1]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[6], "%f", &_diffuse[2]) != 1) return COMMAND_SYNOPSIS;
				diffuse[3] = 1.0f;
				if(sscanf(argv[8], "%f", &_specular[0]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[9], "%f", &_specular[1]) != 1) return COMMAND_SYNOPSIS;
				if(sscanf(argv[10], "%f", &_specular[2]) != 1) return COMMAND_SYNOPSIS;
				specular[3] = 1.0f;
				if(sscanf(argv[12], "%f", &_shininess) != 1) return COMMAND_SYNOPSIS;
				setBackMaterial(_ambient, _diffuse, _specular, _shininess);
			}
			else return COMMAND_SYNOPSIS;
		}

	
  }
  	((TclScene3D *)(data->getScene()->getInterface()))->redraw();
  m->setResult("ambient %f %f %f %f diffuse %f %f %f %f specular %f %f %f %f shininess %f emission %s %f %f %f %f",
	  ambientBack[0], ambientBack[1], ambientBack[2], ambientBack[3],
	  diffuseBack[0], diffuseBack[1], diffuseBack[2], diffuseBack[3],
	  specularBack[0], specularBack[1], specularBack[2], specularBack[3],
	  shininessBack, TclModule::boolString(lighted),
	  emissionBack[0], emissionBack[1], emissionBack[2], emissionBack[3]);
  return TCL_OK;
}



} // namespace yav

