#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "graphics3D/MaterialSphere.h"
#pragma implementation "graphics3D/Sphere.h"
#include <graphics3D/MaterialSphere.h>
#include <graphics3D/Geometry.h>


namespace yav {




MaterialSphere::MaterialSphere(Scene3D *scene)
  : Sphere(), Surface(), Data3D(scene) {
  init();
}


MaterialSphere::MaterialSphere(const Vec3<double>& c, double r, Scene3D *scene)
  : Sphere(c, r), Surface(), Data3D(scene) {
  init();
}


MaterialSphere::~MaterialSphere() {
  gluDeleteQuadric(sphere);
  redrawScene();
}


void MaterialSphere::init(void) {
  unsigned int n = 1;

  sphere = gluNewQuadric();
  updateCenter(Data3D::center, n);
  Data3D::radius = updateRadius(Data3D::center);

  initBoundingBox();
}


void MaterialSphere::initBoundingBox(void) {
  double s;
  Vec3<double> diag(0.577350269189625, 0.577350269189625, 0.577350269189625);
  s = sqrt(3 * Data3D::radius * Data3D::radius);
  boundingBox[0] = Data3D::center + s * diag;
  boundingBox[1] = Data3D::center - s * diag;
}


void MaterialSphere::redraw(Camera3D *cam) {
  Surface::redraw(cam);
}


void MaterialSphere::setRadius(double r) {
  Sphere::setRadius(r);
  Data3D::radius = Sphere::radius;
  surfaceChanged();
  redrawScene();
}


void MaterialSphere::setCenter(const Vec3<double>& c) {
  Sphere::setCenter(c);
  Data3D::center = Sphere::center;
  surfaceChanged();
  redrawScene();
}


void MaterialSphere::updateCenter(Vec3<double>& c, unsigned int& n) {
  c = Sphere::center;
  n = 1;
}


double MaterialSphere::updateRadius(const Vec3<double>& center) {
  return Sphere::radius;
}

  
double MaterialSphere::computeClosestDistance(DataTypes::MouseClick& mc)
  const {
  Data3D::MouseClick3D& mc3D = (Data3D::MouseClick3D&) mc;
  
  double d = Distance::pointToLine(mc3D.eye, mc3D.direction, Sphere::center);
  if(d < Sphere::radius) return 0.0;
  else return d - Sphere::radius;
}


void MaterialSphere::buildWireframe(void) {
  glLineWidth(1.0);
  gluQuadricDrawStyle(sphere, (GLenum) GLU_LINE);
  glTranslated(Sphere::center.getX(), Sphere::center.getY(),
	       Sphere::center.getZ());
  gluSphere(sphere, Sphere::radius, 10, 10);
  glTranslated(-Sphere::center.getX(), -Sphere::center.getY(),
	       -Sphere::center.getZ());
}


void MaterialSphere::buildSolid(void) {
  gluQuadricDrawStyle(sphere, (GLenum) GLU_FILL);
  glTranslated(Sphere::center.getX(), Sphere::center.getY(),
	       Sphere::center.getZ());
  gluSphere(sphere, Sphere::radius, 10, 10);
  glTranslated(-Sphere::center.getX(), -Sphere::center.getY(),
	       -Sphere::center.getZ());
}


void MaterialSphere::translate(const Vec3<double>& tr) {
  Sphere::translate(tr);
  Data3D::center = Sphere::center;
  surfaceChanged();
}

void MaterialSphere::scale(const double scale) {
  Sphere::radius *= scale;
  Data3D::radius = Sphere::radius;
  surfaceChanged();
}


} // namespace yav

