#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "graphics3D/Plane.h"
#include <graphics3D/Plane.h>


namespace yav {



Plane::Plane() {
  equation[0] = 0.0;
  equation[1] = 0.0;
  equation[2] = 1.0;
  equation[3] = 0.0;
}


Plane::Plane(const double *eqn) {
  for(int i = 0; i < 4; i++) equation[i] = eqn[i];
}


Plane::Plane(const Vec3<double>& P, const Vec3<double>& n) {
  setEquation(P, n);
}


void Plane::setEquation(const double *eqn) {
  for(int i = 0; i < 4; i++) equation[i] = eqn[i];
}


void Plane::setEquation(const Vec3<double>& P, const Vec3<double>& n) {
  equation[0] = n.getX();
  equation[1] = n.getY();
  equation[2] = n.getZ();
  equation[3] = -dot(P, n);
}

Vec3<double> Plane::getPoint(void) const {
  double M[3];
  int i;

  if(fabs(equation[0]) > fabs(equation[1])) i = 0;
  else i = 1;
  if(fabs(equation[2]) > fabs(equation[i])) i = 2;

  M[i] = -equation[3] / equation[i];
  M[(i+1) % 3] = 0.0;
  M[(i+2) % 3] = 0.0;

  return Vec3<double>(M);
}


Vec3<double> Plane::getPoint(const Vec3<double> &P) const {
  Vec3<double> n(equation[0], equation[1], equation[2]);
  double a;

  a = -(equation[3] + dot<double>(n, P)) / n.norm2();
  return P + (a * n);
}


Vec3<double> Plane::getNormal(void) const {
  Vec3<double> n(equation[0], equation[1], equation[2]);
  return n / n.norm();
}



void Plane::translate(const Vec3<double>& t) {
  Vec3<double> M = getPoint();
  M += t;
  equation[3] = -equation[0] * M.getX() - equation[1] * M.getY()
    - equation[2] * M.getZ();
}


void Plane::rotate(const Rotation3D& r) {
  Vec3<double> M = getPoint();
  Vec3<double> n = getNormal();
  n = r * n;
  equation[0] = n.getX();
  equation[1] = n.getY();
  equation[2] = n.getZ();
  equation[3] = -n.getX() * M.getX() - n.getY() * M.getY()
    - n.getZ() * M.getZ();
}



void Plane::rotate(const Rotation3D& r, const Vec3<double>& P) {
  translate(-P);
  rotate(r);
  translate(P);
}



} // namespace yav

