
#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "graphics3D/Rectangle3D.h"
#include <graphics3D/Rectangle3D.h>



namespace yav {

double Rectangle3D::computeClosestDistance(DataTypes::MouseClick& mc) const {
  double axialDistance, radialDistance, minRadialDistance;
  Data3D::MouseClick3D& mc3D = (Data3D::MouseClick3D&) mc;
  Vec3<double> dp;

  // distance of the center from the center of the camera
  minRadialDistance= -1;
  for(unsigned int i=0;i<4;++i)
    { 
      axialDistance = dot(point[i]- mc3D.eye, mc3D.direction);
      if ( (axialDistance>0) && (axialDistance<mc3D.threshold))
	{
	  dp= point[i]-(mc3D.eye + mc3D.direction * axialDistance);
	  radialDistance = dp.norm2();
	  if ((minRadialDistance < 0)|| (radialDistance < minRadialDistance))
	    minRadialDistance = radialDistance;
	}
    }
  return minRadialDistance;
}
void Rectangle3D::init() {
  useFlat=true;
  redrawOrder = 2;
  flipNormal=false;
  tscale=sscale=1.0;

}

Rectangle3D::Rectangle3D(Scene3D *scene, const Vec3<double> &pp1,
			 const Vec3<double> &pp2,const Vec3<double> &vv1,
			 const Vec3<double> &vv2) : Data3D(scene), Surface(){
  Vec3<double> p1,p2,v1,v2;
  p1=pp1; p2=pp2; v1=vv1; v2=vv2;
  if (p2==p1)
    p2=p1+Vec3<double>(1,1,0);
  Vec3<double> dp=(p2-p1);
  v1/=v1.norm();
  /// if v1 is not orthogonal to v2 then orthogonalize it
  if (fabs(dot<double>(v1,v2))>1e-5)
    {
      v2-= dot<double>(v1,v2)* v1;
    }
  v2/=v2.norm();

  point[0]=p1;
  point[1]=p1+dot<double>(dp,v1) *v1;
  point[2]=p2;
  point[3]=p1+dot<double>(dp,v2) *v2;
  init();
}
Rectangle3D::~Rectangle3D() {
  //  glDeleteLists(displayList,1);
}
void Rectangle3D::buildWireframe(void) {
  glBegin(GL_LINE_LOOP);
  glVertex3dv((const double *) point[0]);
  glVertex3dv((const double *) point[1]);
  glVertex3dv((const double *) point[2]);
  glVertex3dv((const double *) point[3]);
  glEnd();
}
void Rectangle3D::buildSolid(void) {
  buildFlat();
}
void Rectangle3D::buildFlat(void) {
  double s,t;
  s=t=1.0;
  
  glShadeModel(GL_FLAT);
  if (texture2D)
    {
      s=sscale*texture2D->getNbRealRows()/(double) texture2D->getNbRows();
      t=tscale*texture2D->getNbRealColumns()/(double) texture2D->getNbColumns();
    }
  glBegin(GL_QUADS);
  if (flipNormal)
    glNormal3dv((const double *) -getNormal());
  else
    glNormal3dv((const double *) getNormal());
  glTexCoord2d(0.0,0.0);
  glVertex3dv((const double *) point[0]);
  glTexCoord2d(t,0.0);
  glVertex3dv((const double *) point[1]);
  glTexCoord2d(t,s);
  glVertex3dv((const double *) point[2]);
  glTexCoord2d(0.0,s);
  glVertex3dv((const double *) point[3]);
  glEnd();
  glShadeModel(GL_SMOOTH);
}

void Rectangle3D::getSize(double &length, double &width) const {
  length=(point[1]-point[0]).norm();
  width=(point[2]-point[1]).norm();
}
void Rectangle3D::surfaceHasChanged() {
  Surface::surfaceChanged();
}
void Rectangle3D::redraw(Camera3D *cam) {
  
  if (updateDisplayList)
    surfaceHasChanged();
  Surface::redraw(cam);
}


}


