#include <graphics3D/SceneCmd.h>
#include <graphics3D/TclCamera3D.h>


namespace yav {


	/*
std::map<TclString,RangeInformation::RIREADFUNCTION> RangeInformation::readFunctionMap;
*/


void Scene3D_Init(TclModule *m) {
  m->newCommand("scn3D", TclScene3D::whichCmd,
		"<-which [none] [scene] [camera] [file] [other] | -suffix %s | -realName %s>",
		"returns objects handled by 3D scenes (filtered by construction mode) or filename suffixes used for a given object type or the expanded name associated with a given object type");
  
  TclObjectKind *kind = new TclObjectKind(m, typeid(TclScene3D), "scn3D","3D Scene",
					  &TclScene3D::create,
					  &TclScene3D::list, 0);
  m->newCommand("scn3D%u", kind->getOptions(), "3D scene management command");

  kind->newOption("-destroy", (TclObject::OPTION) &TclScene3D::destroyCmd, 0,
		  "delete a 3D scene");
  kind->newOption("-maxFrameRate",(TclObject::OPTION) &TclScene3D::maxFrameRateCmd,"[%f]",
		  "get/set the maximum frame rate of display in Hz");
  kind->newOption("-objects", (TclObject::OPTION) &TclScene3D::objectsCmd,
		  0, "return scene objects list");
  kind->newOption("-deformable", (TclObject::OPTION) &TclScene3D::deformableCmd,
		  0, "return scene deformable models  list");
  kind->newOption("-camera", (TclObject::OPTION) &TclScene3D::cameraCmd,
		  0, "return the list of camera in a scene");
  kind->newOption("-load", (TclObject::OPTION) &TclScene3D::loadCmd, "%s",
		  "load 3D scene from given file name");
  kind->newOption("-iteration", (TclObject::OPTION) &TclScene3D::iterationCmd, "[step|call] [%u]",
		  "If no parameter is provided then all activated deformable models of the scene are deformed. Otherwise get or set the number of iterations per step or the number of iterations per background call");
  kind->newOption("-computation",
		  (TclObject::OPTION) &TclScene3D::computationCmd,
		  "[%b|%u|loop]","get/set the boolean corresponding to the activation status of the scene; if true it means that the scene is activating all deformable models. An integer value indicates the number of step to iterate before a complete stop.\"loop\" indicates that the scene will iterate until no deformable models are active");
  
  kind->newOption("-redraw", (TclObject::OPTION) &TclScene3D::redrawCmd, 0,
		  "redraw all cameras viewing this 3D scene");
  kind->newOption("-focus", (TclObject::OPTION) &TclScene3D::focusCmd, 0,
		  "reset all cameras viewing this 3D scene focus");

}

#ifdef _SOLARIS_
int main() {
}
#endif


} // namespace yav

