#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "graphics3D/Surface.h"
#include <graphics3D/Surface.h>
#include <graphics3D/TclTexture2D.h>
#include <graphics3D/TclTexture1D.h>
#include <graphics3D/TclScene3D.h>
#include <graphics3D/Camera3D.h>

namespace yav {



const char *Surface::synopsis = "[hide | wireframe | flat | solid | wireframesolid | hiddenlines | transparency | wireframetransparency]";

const char *Surface::texture2DSynopsis = "[ replaceLighting [true | false] | use [true | false] | useHighlightTexture [true | false] | interpolation [nearest|linear] | extrapolation [ clamp | repeat ] | mode [ spherical | coordinates | planar | camera ] | sparam [ %f %f %f %f ] | tparam [ %f %f %f %f ] ]";

const char *Surface::texture1DSynopsis = "[ use [true | false] | | interpolation [nearest|linear] | extrapolation [ clamp | repeat ] | mode [ spherical | coordinates | planar | camera ] | sparam [ %f %f %f %f ] | tparam [ %f %f %f %f ] ]";

const TclString Surface::NONE("Hide");
const TclString Surface::WIREFRAME("Wireframe");
const TclString Surface::FLAT("Flat");
const TclString Surface::SOLID("Gouraud");
const TclString Surface::WIREFRAME_SOLID("Gouraud and Wireframe");
const TclString Surface::HIDDEN_LINES("Hidden Lines");

void  Surface::initRenderingList(void)
{
  texture2DInterpolation=Texture2D::NEAREST;
  texture2DExtrapolation=Texture2D::REPEAT;
  texture2DMode=Texture2D::PLANAR_MAPPING;
  texture2D=0;
  texture2D.setOwner(this);

  texture1DInterpolation=Texture1D::NEAREST;
  texture1DExtrapolation=Texture1D::REPEAT;
  texture1DMode=Texture1D::PLANAR_MAPPING;
  texture1D=0;
  texture1D.setOwner(this);
  
  useTexture2D=false;
  useTexture1D=false;
  useTextureHighlight=false;
  textureReplaceLighting=false;
  useDisplayLists=true;
  sTextureParameters[0]=1.0;
  sTextureParameters[1]=sTextureParameters[2]=sTextureParameters[3]=0.0;
  tTextureParameters[1]=1.0;
  tTextureParameters[0]=tTextureParameters[2]=tTextureParameters[3]=0.0;

  Rendering *rHide= new Rendering;
  rHide->renderingName=Surface::NONE;
  rHide->optionName="hide";
  rHide->redraw= &dummySurfaceRedraw;
  rHide->clientData=0;
  renderingList.push_back(rHide);
 
  Rendering *rWire= new Rendering;
  rWire->renderingName=Surface::WIREFRAME;
  rWire->optionName="wireframe";
  rWire->redraw= &wireframeSurfaceRedraw;
  rWire->clientData=0;
  renderingList.push_back(rWire);
 
  Rendering *rFlat= new Rendering;
  rFlat->renderingName=Surface::FLAT;
  rFlat->optionName="flat";
  rFlat->redraw= &flatSurfaceRedraw;
  rFlat->clientData=0;
  renderingList.push_back(rFlat);
  
  Rendering *rGouraud= new Rendering;
  rGouraud->renderingName=Surface::SOLID;
  rGouraud->optionName="solid";
  rGouraud->redraw= &gouraudSurfaceRedraw;
  rGouraud->clientData=0;
  renderingList.push_back(rGouraud);
  
  Rendering *rWireGouraud= new Rendering;
  rWireGouraud->renderingName=Surface::WIREFRAME_SOLID;
  rWireGouraud->optionName="wireframesolid";
  rWireGouraud->redraw= &wireGouraudSurfaceRedraw;
  rWireGouraud->clientData=0;
  renderingList.push_back(rWireGouraud);
  
  Rendering *rHidden= new Rendering;
  rHidden->renderingName=Surface::HIDDEN_LINES;
  rHidden->optionName="hiddenlines";
  rHidden->redraw= &hiddenSurfaceRedraw;
  rHidden->clientData=0;
  renderingList.push_back(rHidden);
  
  selectedRendering=rWire;
}

void Surface::destroyRenderingList(void)
{
  std::list<Rendering *>::iterator itr=renderingList.begin();
  for(;itr!=renderingList.end();itr++)
    delete (*itr);

}
  
Surface::Surface()
  : Owner<Texture2D>(0),Owner<Texture1D>(0),useFlat(false),mat(), color(), width(1.0f),
    wireframeList(0), solidList(0),flatList(0),
    texture2D(this,0),highlightTexture(this,0),useDisplayLists(true) {
  initRenderingList();
}


Surface::Surface(const GLfloat *ambient, const GLfloat *diffuse,
		 const GLfloat *specular, GLfloat shininess)
  :  Owner<Texture2D>(0),Owner<Texture1D>(0),useFlat(false),
     mat(ambient, diffuse, specular, shininess), color(mat),
    width(1.0f), wireframeList(0), solidList(0),flatList(0),
    texture2D(this,0),highlightTexture(this,0),useDisplayLists(true) {
  initRenderingList();
}


Surface::Surface(const Surface& org)
  :   Owner<Texture2D>(0),Owner<Texture1D>(0),
      useFlat(false),mat(org.mat), color(org.color), 
    width(1.0f), wireframeList(0), solidList(0),flatList(0),
    texture2D(this,0),highlightTexture(this,0),useDisplayLists(true) {
  initRenderingList();
}


Surface::~Surface() {
  destroyWireframeDisplayList();
  destroySolidDisplayList();
  destroyFlatDisplayList();

  destroyRenderingList();
}
void Surface::dying(const ObjectPointer<Texture2D>& obj) {
  if (obj()==texture2D())
    {
      setTexture2DUsage(false);
    }
  else if  (obj()==highlightTexture())
    setTextureHighlightUsage(false);
}
void Surface::dying(const ObjectPointer<Texture1D>& obj) {
  if (obj()==texture1D())
    {
      setTexture1DUsage(false);
    }
}
void Surface::surfaceChanged(void) {
  destroyWireframeDisplayList();
  destroySolidDisplayList();
  destroyFlatDisplayList();
}

void Surface::redraw(Camera3D *cam) {
  if (selectedRendering)
    (selectedRendering)->redraw(cam,this,selectedRendering->clientData);
}

void Surface::removeRenderingMode(Rendering *r) {
  if (r) {
    std::list<Rendering *>::iterator itr=renderingList.begin();
    for(;((*itr)!=r) && (itr!=renderingList.end());itr++);
    if  (itr!=renderingList.end())
      renderingList.erase(itr);
    if (selectedRendering==r)
      selectedRendering=0;
    delete r;
  }
}


void Surface::addRenderingMode(TclString _renderingName,TclString _optionName,
			       RedrawRenderingMode _redrawFn, void *ClientData) {
  Rendering *r=new Rendering;
  r->renderingName= _renderingName;
  r->optionName= _optionName;
  r->redraw=_redrawFn;
  r->clientData=ClientData;
  renderingList.push_back(r);
}
void Surface::redrawBasic(Camera3D *cam,RenderMode rm) {
	switch(rm) {

  case RM_NONE:
	  break;

  case RM_WIREFRAME:
	  glDisable(GL_LIGHTING);
	  color.glSet();
	  glLineWidth(width);
	  callWireframeDisplayList();
	  glEnable(GL_LIGHTING);
	  break;

  case RM_FLAT:
	  glShadeModel(GL_FLAT);
	  glEnable(GL_LIGHTING);
	  mat.glSet();
	  if (useTexture2D && texture2D())
	  {
		  if (textureReplaceLighting)
			  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		  else
			  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		  texture2D->enableTexture(texture2DMode);
		  texture2D->interpolateTexture(texture2DInterpolation);
		  texture2D->extrapolateTexture(texture2DExtrapolation);
		  texture2D->applyTextureMode(texture2DMode,sTextureParameters,tTextureParameters);
	  }
	  else if (useTexture1D && texture1D())
	  {
		  if (textureReplaceLighting)
			  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		  else
			  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		  texture1D->enableTexture(texture1DMode);
		  texture1D->interpolateTexture(texture1DInterpolation);
		  texture1D->extrapolateTexture(texture1DExtrapolation);
		  texture1D->applyTextureMode(texture1DMode,sTextureParameters,tTextureParameters);
	  }
	  callFlatDisplayList();
	  if (useTexture2D && texture2D())
	  {
		  texture2D->disableTexture(texture2DMode);
	  }
	  else if (useTexture1D && texture1D())
	  {
		  texture1D->disableTexture(texture1DMode);
	  }
	  glShadeModel(GL_SMOOTH);
	  break;

  case RM_SOLID:
	  if(mat.isTransparent()==true) {
		  glEnable(GL_BLEND);
		  glDepthMask(GL_FALSE);
		  glBlendFunc(GL_SRC_ALPHA, GL_ONE);
		  /*      glBlendEquationEXT(GL_FUNC_ADD_EXT);
		  glBlendFunc(GL_CONSTANT_ALPHA_EXT, GL_ONE);
		  glBlendColorEXT(1.f, 1.f, 1.f, 1.f / 4.f);*/
	  }
	  glShadeModel(GL_SMOOTH);
	  glEnable(GL_LIGHTING);
	  mat.glSet();

	  if (useTexture2D && texture2D())
	  {
		  if (textureReplaceLighting)
			  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		  else
			  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		  texture2D->enableTexture(texture2DMode);
		  texture2D->interpolateTexture(texture2DInterpolation);
		  texture2D->extrapolateTexture(texture2DExtrapolation);
		  texture2D->applyTextureMode(texture2DMode,sTextureParameters,tTextureParameters);
	  }
	  else if (useTexture1D && texture1D())
	  {
		  if (textureReplaceLighting)
			  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		  else
			  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		  texture1D->enableTexture(texture1DMode);
		  texture1D->interpolateTexture(texture1DInterpolation);
		  texture1D->extrapolateTexture(texture1DExtrapolation);
		  texture1D->applyTextureMode(texture1DMode,sTextureParameters,tTextureParameters);
	  }
	  callSolidDisplayList();
	  if (useTexture2D && texture2D())
	  {
		  texture2D->disableTexture(texture2DMode);
	  }
	  else if (useTexture1D && texture1D())
	  {
		  texture1D->disableTexture(texture1DMode);
	  }
	  if (useTextureHighlight && highlightTexture())
	  {
		  glDepthFunc(GL_LEQUAL);
		  glBlendFunc(GL_ONE, GL_ONE);
		  glEnable(GL_BLEND);  
		  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		  highlightTexture->enableTexture(Texture2D::SPHERICAL_MAPPING);
		  highlightTexture->interpolateTexture(Texture2D::LINEAR);
		  highlightTexture->extrapolateTexture(Texture2D::REPEAT);
		  highlightTexture->applyTextureMode(Texture2D::SPHERICAL_MAPPING,sTextureParameters,tTextureParameters);
		  callSolidDisplayList();
		  glDisable(GL_BLEND);
		  glDepthFunc(GL_LESS);
		  highlightTexture->disableTexture(Texture2D::SPHERICAL_MAPPING);
	  }
	  if(mat.isTransparent()==true) {
		  glDepthMask(GL_TRUE);
		  glDisable(GL_BLEND);
	  }
	  break;

  case RM_WIREFRAME_SOLID:
  case RM_HIDDEN_LINES:
	  if(rm == RM_HIDDEN_LINES) {
		  glDisable(GL_LIGHTING);
		  glColor3f(cam->getBackgroundColor(CameraTypes::BASE_RED),
			  cam->getBackgroundColor(CameraTypes::BASE_GREEN),
			  cam->getBackgroundColor(CameraTypes::BASE_BLUE));
	  }
	  else {
		  if(mat.isTransparent()==true) {
			  glEnable(GL_BLEND);
			  glDepthMask(GL_FALSE);
			  glBlendFunc(GL_SRC_ALPHA, GL_ONE);
		  }
		  glShadeModel(GL_SMOOTH);
		  glEnable(GL_LIGHTING);
		  mat.glSet();
	  }
	  if (useTexture2D && texture2D())
	  {
		  if (textureReplaceLighting)
			  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		  else
			  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		  texture2D->enableTexture(texture2DMode);
		  texture2D->interpolateTexture(texture2DInterpolation);
		  texture2D->extrapolateTexture(texture2DExtrapolation);
		  texture2D->applyTextureMode(texture2DMode,sTextureParameters,tTextureParameters);
	  }
	  else if (useTexture1D && texture1D())
	  {
		  if (textureReplaceLighting)
			  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		  else
			  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		  texture1D->enableTexture(texture1DMode);
		  texture1D->interpolateTexture(texture1DInterpolation);
		  texture1D->extrapolateTexture(texture1DExtrapolation);
		  texture1D->applyTextureMode(texture1DMode,sTextureParameters,tTextureParameters);
	  }
	  callSolidDisplayList();
	  if(mat.isTransparent()==true) {
		  glDepthMask(GL_TRUE);
		  glDisable(GL_BLEND);
	  }
	  if (useTexture2D && texture2D())
	  {
		  texture2D->disableTexture(texture2DMode);
	  }
	  else if (useTexture1D && texture1D())
	  {
		  texture1D->disableTexture(texture1DMode);
	  }
	  glDisable(GL_LIGHTING);
	  color.glSet();
	  glMatrixMode(GL_PROJECTION);
	  glTranslatef (0.0f, 0.0f, (float)(lineDisplacementOffset * cam->getRadius()));
	  glMatrixMode(GL_MODELVIEW);
	  glLineWidth(width);
	  callWireframeDisplayList();
	  glMatrixMode(GL_PROJECTION);
	  glTranslatef (0.0f, 0.0f, (float)(-lineDisplacementOffset * cam->getRadius()));
	  glMatrixMode(GL_MODELVIEW);
	  glEnable(GL_LIGHTING);
	  break;

	}
}
void Surface::drawColorbar(Camera3D *cam,
						   double colorbarHeight,double colorbarWidth,
		  double posx,double posy,char *minLabel,char *maxLabel)
{
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glOrtho(-0.5, cam->getWidth()-0.5, -0.5, cam->getHeight() - 0.5, 0.0, 1.0);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glDisable(GL_LIGHTING);
  glColor3f(1.0, 1.0, 1.0);

  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
  texture1D->enableTexture(Texture1D::COORDINATES);
  texture1D->interpolateTexture(Texture1D::NEAREST);
  texture1D->extrapolateTexture(Texture1D::CLAMP);

  double coordTexture=(double)(texture1D->getNbRealColumns())/texture1D->getNbColumns();
  glBegin(GL_POLYGON);
  glTexCoord1d(0.0);
  glVertex3d(posx,posy,0);
  glTexCoord1d(coordTexture);
  glVertex3d(posx,posy+colorbarHeight,0);
  glTexCoord1d(coordTexture);
  glVertex3d(posx+colorbarWidth,posy+colorbarHeight,0);
  glTexCoord1d(0.0);
  glVertex3d(posx+colorbarWidth,posy,0);
  glEnd();
  texture1D->disableTexture(Texture1D::COORDINATES);
  glPopMatrix();
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();	  
  glMatrixMode(GL_MODELVIEW);
  if (minLabel) {
	  unsigned int i,length;
	  length=strlen(minLabel);
	  char lab[2];
	  lab[1]='\0';
	  for(i=0;i<length;++i) {
		  lab[0]=minLabel[length-1-i];
		  cam->printString((posx-10.0-i*7.0),posy,lab);
	  }
  }
 if (maxLabel) {
	  unsigned int i,length;
	  length=strlen(maxLabel);
	  char lab[2];
	  lab[1]='\0';
	  for(i=0;i<length;++i) {
		  lab[0]=maxLabel[length-1-i];
		  cam->printString((posx-10.0-i*7.0),posy+colorbarHeight,lab);
	  }
 }

}
void Surface::wireframeLinesRedraw(Camera3D *cam) {
  glDisable(GL_LIGHTING);
  color.glSet();
  glLineWidth(width);
  callWireframeDisplayList();
  glEnable(GL_LIGHTING);
}

void Surface::wireframeHiddenRedraw(Camera3D *cam) {
  glDisable(GL_LIGHTING);
  color.glSet();
  glMatrixMode(GL_PROJECTION);
  glTranslatef (0.0f, 0.0f, (float)(lineDisplacementOffset * cam->getRadius()));
  glMatrixMode(GL_MODELVIEW);
  glLineWidth(width);
  callWireframeDisplayList();
  glMatrixMode(GL_PROJECTION);
  glTranslatef (0.0f, 0.0f, (float)(-lineDisplacementOffset * cam->getRadius()));
  glMatrixMode(GL_MODELVIEW);
  glEnable(GL_LIGHTING);
}

void Surface::solidFlatRedraw(Camera3D *cam) {
  glShadeModel(GL_FLAT);
  glEnable(GL_LIGHTING);
  mat.glSet();
  if (useFlat)
    callFlatDisplayList();
  else
    callSolidDisplayList();
  glDisable(GL_LIGHTING);
  glShadeModel(GL_SMOOTH);
}

void Surface::solidGouraudRedraw(Camera3D *cam) {
  if(mat.isTransparent()==true) {
 glEnable(GL_BLEND);
  glDepthMask(GL_FALSE);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE);
  }
  glShadeModel(GL_SMOOTH);
  mat.glSet();
  glEnable(GL_LIGHTING);
  callSolidDisplayList();
   if(mat.isTransparent()==true) {
  glDepthMask(GL_TRUE);
  glDisable(GL_BLEND);
   }
}



void Surface::buildWireframeDisplayList(void) {
  wireframeList = glGenLists(1);
  glNewList(wireframeList, GL_COMPILE_AND_EXECUTE);
  buildWireframe();
  glEndList();
}


void Surface::buildSolidDisplayList(void) {
  solidList = glGenLists(1);
  glNewList(solidList, GL_COMPILE_AND_EXECUTE);
  buildSolid();
  glEndList();
}


void Surface::buildFlatDisplayList(void) {
  flatList = glGenLists(1);
  glNewList(flatList, GL_COMPILE_AND_EXECUTE);
  buildFlat();
  glEndList();
}


void Surface::callWireframeDisplayList(void) {
  if(!useDisplayLists)
    buildWireframe();
  else if(wireframeList == 0) {
    wireframeList = glGenLists(1);
    glNewList(wireframeList, GL_COMPILE_AND_EXECUTE);
    buildWireframe();
    glEndList();
  }
  else glCallList(wireframeList);
}

void Surface::callSolidDisplayList(void) {
	if(!useDisplayLists)
    buildSolid();
  else if(solidList == 0) {
    solidList = glGenLists(1);
    glNewList(solidList, GL_COMPILE_AND_EXECUTE);
    buildSolid();
    glEndList();
  }
  else glCallList(solidList);
}

void Surface::callFlatDisplayList(void) {
  if (useFlat) {
    if(!useDisplayLists)
      buildFlat();
    else if(flatList == 0) {    
      flatList = glGenLists(1);
      glNewList(flatList, GL_COMPILE_AND_EXECUTE);
      buildFlat();
      glEndList();
    }
    else glCallList(flatList);
  } else {
    callSolidDisplayList();
  }
}


void Surface::destroyWireframeDisplayList(void) {
  if(wireframeList > 0) {
    glDeleteLists(wireframeList, 1);
    wireframeList=0;
  }
}

void Surface::destroySolidDisplayList(void) {
  if(solidList > 0) {
    glDeleteLists(solidList, 1);
    solidList=0;
  }
}

void Surface::destroyFlatDisplayList(void) {
  if(flatList > 0) {
    glDeleteLists(flatList, 1);
    flatList=0;
  }
}

void Surface::setRendering(const TclString& n) {
  setRendering(getRenderingFromName(n));
}
void Surface::setRendering(Rendering *r) {
  if (r && (r!=selectedRendering)){
 
    selectedRendering=r;
  
  }
}


Surface::Rendering *Surface::getRenderingFromName(TclString n)
{
  Rendering *r=0;
  std::list<Rendering *>::iterator itr=renderingList.begin();
  for(;itr!=renderingList.end();itr++)
    {
      if ((*itr)->renderingName==n)
	r=(*itr);
    }
  return(r);
}
Surface::Rendering *Surface::getRenderingFromOptionName(TclString n)
{
  Rendering *r=0;
  std::list<Rendering *>::iterator itr=renderingList.begin();
  for(;itr!=renderingList.end();itr++)
    {
      if ((*itr)->optionName==n)
	r=(*itr);
    }
  return(r);
}


void Surface::setLineWidth(GLfloat w) {
  if(w > 0.0) width = w;
}

int Surface::renderCmd(TclModule *m, Data3D *data,
		       int argc, char **argv, int i) {
	if(i < argc) 
		{
		if (!strcmp(argv[i],"checkOpacity")) 
			{
			if (mat.isTransparent()) 
				{
				setRedrawOrder(2);
				} else 
				{
				setRedrawOrder(1);
					}
				((TclScene3D *)(data->getScene()->getInterface()))->redraw();
				return TCL_OK;
			} else 
			{
			Rendering *r=getRenderingFromOptionName(argv[i]);
			if (r==0)
				{
				m->setResult("invalid render mode");
				return TCL_ERROR;
				}
			i++;
			if (selectedRendering!=r)
				{
				setRendering(r);
				((TclScene3D *)(data->getScene()->getInterface()))->redraw();
				
				}
			
			}
		}
		m->setResult(selectedRendering->optionName);
		return TCL_OK;
		
}
int Surface::renderOptionCmd(TclModule *m, Data3D *data,
			     int argc, char **argv, int i) {
  TclString output;
  std::list<Rendering *>::iterator itr=renderingList.begin();
  for(;itr!=renderingList.end();itr++)
    {
      TclString option;
	option.sprintf("{ {%s} {%s} } ",(char *)((*itr)->renderingName),
		       (char *)((*itr)->optionName));
      output +=option;
    }
  m->setResult(output );
  return TCL_OK;
}


int Surface::texture2DOptionCmd(TclModule *m, Data3D *data,
				 int argc, char **argv, int i)
{
  if (argc==1)
    {
      if (!strcmp(argv[0],"use"))
	{
	  m->setResult(TclModule::boolString(useTexture2D));
	  return TCL_OK;
	}
      else if (!strcmp(argv[0],"replaceLighting"))
	{
	  m->setResult(TclModule::boolString(textureReplaceLighting));
	  return TCL_OK;
	}
      else if (!strcmp(argv[0],"useHighlightTexture"))
	{
	  m->setResult(TclModule::boolString(useTextureHighlight));
	  return TCL_OK;
	}
      else if (!strcmp(argv[0],"interpolation"))
	{
	  if (texture2DInterpolation==Texture2D::LINEAR)
	    m->setResult("linear");
	  else if (texture2DInterpolation==Texture2D::NEAREST)
	    m->setResult("nearest");
	  return TCL_OK;
	}
      else if (!strcmp(argv[0],"extrapolation"))
	{
	  if (texture2DExtrapolation==Texture2D::CLAMP)
	    m->setResult("clamp");
	  else if (texture2DExtrapolation==Texture2D::REPEAT)
	    m->setResult("repeat");
	  return TCL_OK;
	}
      else if (!strcmp(argv[0],"mode"))
	{
	  if (texture2DMode==Texture2D::COORDINATES)
	    m->setResult("coordinates");
	  else if (texture2DMode==Texture2D::SPHERICAL_MAPPING)
	    m->setResult("spherical");
	  else if (texture2DMode==Texture2D::PLANAR_MAPPING)
	    m->setResult("planar");
	  else if (texture2DMode==Texture2D::CAMERA_MAPPING)
	    m->setResult("camera");
	  return TCL_OK;
	}
      else if (!strcmp(argv[0],"sparam"))
	{
	  m->setResult("%f %f %f %f",sTextureParameters[0],sTextureParameters[1],sTextureParameters[2],sTextureParameters[3]);
	  return TCL_OK;
	}
      else if (!strcmp(argv[0],"tparam"))
	{
	  m->setResult("%f %f %f %f",tTextureParameters[0],tTextureParameters[1],tTextureParameters[2],tTextureParameters[3]);
	  return TCL_OK;
	}
      else
	{
	  m->setResult("%s: invalid texture query : ",argv[0]);
	  return TCL_ERROR;
	}
    }
  else if (argc==2)
    {
      if (!strcmp(argv[0],"interpolation"))
	{
	  if (!strcmp(argv[1],"linear"))
	    texture2DInterpolation=Texture2D::LINEAR;
	  else 	  if (!strcmp(argv[1],"nearest"))
	    texture2DInterpolation=Texture2D::NEAREST;
	  else
	    {
	      m->setResult("%s must be linear or nearest",argv[1]);
	      return TCL_OK;
	    }
 ((TclScene3D *)(data->getScene()->getInterface()))->redraw();
	  m->setResult("%s",argv[1]);
	  return TCL_OK;
	}
      else if (!strcmp(argv[0],"use"))
	{
	  bool uset;
	  if(!TclModule::boolValue(argv[1], uset)) 
	    return COMMAND_SYNOPSIS;
	  setTexture2DUsage(uset);
	  ((TclScene3D *)(data->getScene()->getInterface()))->redraw();
	  m->setResult(TclModule::boolString(useTexture2D));
	  return TCL_OK;
	}
      else if (!strcmp(argv[0],"replaceLighting"))
	{
	  bool uset;
	  if(!TclModule::boolValue(argv[1], uset)) 
	    return COMMAND_SYNOPSIS;
	  setLightingTextureUsage(uset);
	  ((TclScene3D *)(data->getScene()->getInterface()))->redraw();
	  m->setResult(TclModule::boolString(textureReplaceLighting));
	  return TCL_OK;
	}
      else if (!strcmp(argv[0],"useHighlightTexture"))
	{
	  bool uset;
	  if(!TclModule::boolValue(argv[1], uset)) 
	    return COMMAND_SYNOPSIS;
	  setTextureHighlightUsage(uset);
	  ((TclScene3D *)(data->getScene()->getInterface()))->redraw();
	  m->setResult(TclModule::boolString(useTextureHighlight));
	  return TCL_OK;
	}
      else if (!strcmp(argv[0],"extrapolation"))
	{
	  if (!strcmp(argv[1],"clamp"))
	    texture2DExtrapolation=Texture2D::CLAMP;
	  else 	  if (!strcmp(argv[1],"repeat"))
	    texture2DExtrapolation=Texture2D::REPEAT;
	  else
	    {
	      m->setResult("%s must be clamp or repeat",argv[1]);
	      return TCL_OK;
	    }
	  ((TclScene3D *)(data->getScene()->getInterface()))->redraw();
	  m->setResult("%s",argv[1]);
	  return TCL_OK;
	}
      else if (!strcmp(argv[0],"mode"))
	{
	  if (!strcmp(argv[1],"coordinates"))
	    {
	      texture2DMode=Texture2D::COORDINATES;
	      surfaceChanged();
	    }
	  else 	  if (!strcmp(argv[1],"planar"))
	    texture2DMode=Texture2D::PLANAR_MAPPING;
	  else 	  if (!strcmp(argv[1],"camera"))
	    texture2DMode=Texture2D::CAMERA_MAPPING;
	  else 	  if (!strcmp(argv[1],"spherical"))
	    texture2DMode=Texture2D::SPHERICAL_MAPPING;
	  else
	    {
	      m->setResult("%s must be coordinates or planar or spherical or camera",argv[1]);
	      return TCL_OK;
	    }
 ((TclScene3D *)(data->getScene()->getInterface()))->redraw();
	  m->setResult("%s",argv[1]);
	  return TCL_OK;
	}
      else
	{
	  m->setResult("%s: invalid texture query : ",argv[0]);
	  return TCL_ERROR;
	}
    }
  else if (argc==5) {
    if (!strcmp(argv[0],"sparam"))
      {  
	sTextureParameters[0]=(double) atof(argv[1]);
	sTextureParameters[1]=(double) atof(argv[2]);
	sTextureParameters[2]=(double) atof(argv[3]);
	sTextureParameters[3]=(double) atof(argv[4]);
	surfaceChanged();
 ((TclScene3D *)(data->getScene()->getInterface()))->redraw();
	m->setResult("%f %f %f %f",sTextureParameters[0],sTextureParameters[1],sTextureParameters[2],sTextureParameters[3]);
	return TCL_OK;
      }
    else if (!strcmp(argv[0],"tparam"))
      {  
	tTextureParameters[0]=(double) atof(argv[1]);
	tTextureParameters[1]=(double) atof(argv[2]);
	tTextureParameters[2]=(double) atof(argv[3]);
	tTextureParameters[3]=(double) atof(argv[4]);
	surfaceChanged();
	 ((TclScene3D *)(data->getScene()->getInterface()))->redraw();
	m->setResult("%f %f %f %f",tTextureParameters[0],tTextureParameters[1],tTextureParameters[2],tTextureParameters[3]);
	return TCL_OK;
      } 
    else
      {
	m->setResult("%s: invalid texture query : ",argv[0]);
	return TCL_ERROR;
      }
  }
  else
    return COMMAND_SYNOPSIS;

}

int Surface::texture1DOptionCmd(TclModule *m, Data3D *data,
				 int argc, char **argv, int i)
{
  if (argc==1)
    {
      if (!strcmp(argv[0],"use"))
	{
	  m->setResult(TclModule::boolString(useTexture1D));
	  return TCL_OK;
	}
      else if (!strcmp(argv[0],"interpolation"))
	{
	  if (texture1DInterpolation==Texture1D::LINEAR)
	    m->setResult("linear");
	  else if (texture1DInterpolation==Texture1D::NEAREST)
	    m->setResult("nearest");
	  return TCL_OK;
	}
      else if (!strcmp(argv[0],"extrapolation"))
	{
	  if (texture1DExtrapolation==Texture1D::CLAMP)
	    m->setResult("clamp");
	  else if (texture1DExtrapolation==Texture1D::REPEAT)
	    m->setResult("repeat");
	  return TCL_OK;
	}
      else if (!strcmp(argv[0],"mode"))
	{
	  if (texture1DMode==Texture1D::COORDINATES)
	    m->setResult("coordinates");
	  else if (texture1DMode==Texture1D::SPHERICAL_MAPPING)
	    m->setResult("spherical");
	  else if (texture1DMode==Texture1D::PLANAR_MAPPING)
	    m->setResult("planar");
	  else if (texture1DMode==Texture1D::CAMERA_MAPPING)
	    m->setResult("camera");
	  return TCL_OK;
	}
      else if (!strcmp(argv[0],"sparam"))
	{
	  m->setResult("%f %f %f %f",sTextureParameters[0],sTextureParameters[1],sTextureParameters[2],sTextureParameters[3]);
	  return TCL_OK;
	}
      else if (!strcmp(argv[0],"tparam"))
	{
	  m->setResult("%f %f %f %f",tTextureParameters[0],tTextureParameters[1],tTextureParameters[2],tTextureParameters[3]);
	  return TCL_OK;
	}
      else
	{
	  m->setResult("%s: invalid texture query : ",argv[0]);
	  return TCL_ERROR;
	}
    }
  else if (argc==2)
    {
      if (!strcmp(argv[0],"interpolation"))
	{
	  if (!strcmp(argv[1],"linear"))
	    texture1DInterpolation=Texture1D::LINEAR;
	  else 	  if (!strcmp(argv[1],"nearest"))
	    texture1DInterpolation=Texture1D::NEAREST;
	  else
	    {
	      m->setResult("%s must be linear or nearest",argv[1]);
	      return TCL_OK;
	    }
	  ((TclScene3D *)(data->getScene()->getInterface()))->redraw();
	  m->setResult("%s",argv[1]);
	  return TCL_OK;
	}
      else if (!strcmp(argv[0],"use"))
	{
	  bool uset;
	  if(!TclModule::boolValue(argv[1], uset)) 
	    return COMMAND_SYNOPSIS;
	  setTexture1DUsage(uset);
 ((TclScene3D *)(data->getScene()->getInterface()))->redraw();
	  m->setResult(TclModule::boolString(useTexture2D));
	  return TCL_OK;
	}
      else if (!strcmp(argv[0],"extrapolation"))
	{
	  if (!strcmp(argv[1],"clamp"))
	    texture1DExtrapolation=Texture1D::CLAMP;
	  else 	  if (!strcmp(argv[1],"repeat"))
	    texture1DExtrapolation=Texture1D::REPEAT;
	  else
	    {
	      m->setResult("%s must be clamp or repeat",argv[1]);
	      return TCL_OK;
	    }
	  ((TclScene3D *)(data->getScene()->getInterface()))->redraw();
	  m->setResult("%s",argv[1]);
	  return TCL_OK;
	}
      else if (!strcmp(argv[0],"mode"))
	{
	  if (!strcmp(argv[1],"coordinates"))
	    {
	      texture1DMode=Texture1D::COORDINATES;
	      surfaceChanged();
	    }
	  else 	  if (!strcmp(argv[1],"planar"))
	    texture1DMode=Texture1D::PLANAR_MAPPING;
	  else 	  if (!strcmp(argv[1],"camera"))
	    texture1DMode=Texture1D::CAMERA_MAPPING;
	  else 	  if (!strcmp(argv[1],"spherical"))
	    texture1DMode=Texture1D::SPHERICAL_MAPPING;
	  else
	    {
	      m->setResult("%s must be coordinates or planar or spherical or camera",argv[1]);
	      return TCL_OK;
	    }
	 ((TclScene3D *)(data->getScene()->getInterface()))->redraw();
	  m->setResult("%s",argv[1]);
	  return TCL_OK;
	}
      else
	{
	  m->setResult("%s: invalid texture query : ",argv[0]);
	  return TCL_ERROR;
	}
    }
  else if (argc==5) {
    if (!strcmp(argv[0],"sparam"))
      {  
	sTextureParameters[0]=(double) atof(argv[1]);
	sTextureParameters[1]=(double) atof(argv[2]);
	sTextureParameters[2]=(double) atof(argv[3]);
	sTextureParameters[3]=(double) atof(argv[4]);
	surfaceChanged();
 ((TclScene3D *)(data->getScene()->getInterface()))->redraw();
	m->setResult("%f %f %f %f",sTextureParameters[0],sTextureParameters[1],sTextureParameters[2],sTextureParameters[3]);
	return TCL_OK;
      }
    else if (!strcmp(argv[0],"tparam"))
      {  
	tTextureParameters[0]=(double) atof(argv[1]);
	tTextureParameters[1]=(double) atof(argv[2]);
	tTextureParameters[2]=(double) atof(argv[3]);
	tTextureParameters[3]=(double) atof(argv[4]);
	surfaceChanged();
 ((TclScene3D *)(data->getScene()->getInterface()))->redraw();
	m->setResult("%f %f %f %f",tTextureParameters[0],tTextureParameters[1],tTextureParameters[2],tTextureParameters[3]);
	return TCL_OK;
      } 
    else
      {
	m->setResult("%s: invalid texture query : ",argv[0]);
	return TCL_ERROR;
      }
  }
  else
    return COMMAND_SYNOPSIS;

}

int Surface::texture1DCmd(TclModule *m, Data3D *data,
			  int argc, char **argv, int i)
{
  if (argc==0)
    {
      if (texture1D)
	m->setResult("%s",((char *)((TclTexture1D *)texture1D())->name()));
      return TCL_OK;
    }
  else if (argc==1)
    {
      TclTexture1D *t= (TclTexture1D *)
	TclObjectKind::getObject(typeid(TclTexture1D), argc, argv, 0);
      if (!t)
	{
	  m->setResult("%s must be a 1D texture",argv[0]);
	  return TCL_ERROR;
	}
      texture1D=(Texture1D *)t;
      setTexture1DUsage(true);
      surfaceChanged();
      m->setResult("%s",((char *)((TclTexture1D *)texture1D())->name()));
      return TCL_OK;
    }
  else
    return COMMAND_SYNOPSIS;

}

int Surface::texture2DCmd(TclModule *m, Data3D *data,
				 int argc, char **argv, int i)
{
  if (argc==0)
    {
      if (texture2D)
	m->setResult("%s",((char *)((TclTexture2D *)texture2D())->name()));
      return TCL_OK;
    }
  else if (argc==1)
    {
      TclTexture2D *t= (TclTexture2D *)
	TclObjectKind::getObject(typeid(TclTexture2D), argc, argv, 0);
      if (!t)
	{
	  m->setResult("%s must be a 2D texture",argv[0]);
	  return TCL_ERROR;
	}
      texture2D=(Texture2D *)t;
      setTexture2DUsage(true);
      surfaceChanged();
      m->setResult("%s",((char *)((TclTexture2D *)texture2D())->name()));
      return TCL_OK;
    }
  else
    return COMMAND_SYNOPSIS;

}

int Surface::highlightTextureCmd(TclModule *m, Data3D *data,
				   int argc, char **argv, int i)
{
  if (argc==0)
    {
      if (highlightTexture)
	m->setResult("%s",((char *)((TclTexture2D *)highlightTexture())->name()));
      return TCL_OK;
    }
  else if (argc==1)
    {
      TclTexture2D *t= (TclTexture2D *)
	TclObjectKind::getObject(typeid(TclTexture2D), argc, argv, 0);
      if (!t)
	{
	  m->setResult("%s must be a 2D texture",argv[0]);
	  return TCL_ERROR;
	}
      highlightTexture=(Texture2D *)t;
      setTextureHighlightUsage(true);
      surfaceChanged();
      m->setResult("%s",((char *)((TclTexture2D *)highlightTexture())->name()));
      return TCL_OK;
    }
  else
    return COMMAND_SYNOPSIS;

}

int Surface::useDisplayListsCmd(TclModule *m, Data3D *data, int argc, char **argv, int i){
  bool usedl;
  if(argc > 1) return COMMAND_SYNOPSIS;
  if(argc == 1) {
    if(!TclModule::boolValue(argv[0], usedl)) return COMMAND_SYNOPSIS;
    if(useDisplayLists&&!usedl)
      surfaceChanged();
    useDisplayLists=usedl;
  }
  m->setResult(TclModule::boolString(useDisplayLists));
  return TCL_OK;
}

void Surface::setTexture2DMode(Texture2D::TextureMode tm){
  if((texture2DMode==Texture2D::COORDINATES)||(tm==Texture2D::COORDINATES))
    surfaceChanged();
  texture2DMode=tm;
}


  /// redraw function for the "hide" rendering
  void dummySurfaceRedraw(Camera3D *cam,Surface *s,void *) {}
  /// redraw function for the "wireframe" rendering
  void wireframeSurfaceRedraw(Camera3D *cam,Surface *s,void *) {
    s->redrawBasic(cam,Surface::RM_WIREFRAME);}
  /// redraw function for the "flat" rendering
  void flatSurfaceRedraw(Camera3D *cam,Surface *s,void *) {
    s->redrawBasic(cam,Surface::RM_FLAT);}
  /// redraw function for the "solid" rendering
  void gouraudSurfaceRedraw(Camera3D *cam,Surface *s,void *) {
    s->redrawBasic(cam,Surface::RM_SOLID);}
  /// redraw function for the "wireframesolid" rendering
  void wireGouraudSurfaceRedraw(Camera3D *cam,Surface *s,void *) {
    s->redrawBasic(cam,Surface::RM_WIREFRAME_SOLID);}
  /// redraw function for the "hidden" rendering
  void hiddenSurfaceRedraw(Camera3D *cam,Surface *s,void *) {
    s->redrawBasic(cam,Surface::RM_HIDDEN_LINES);}
  /// redraw function for the "transparency" rendering
 
} // namespace yav


