#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "graphics3D/TclCamera3D.h"
#include <graphics/SceneItem.h>
#include <graphics3D/TclCamera3D.h>
#include <modules/TclMatrix4D.h>
#include <graphics3D/TclScene3D.h>
#include <graphics3D/Scene3D.h>
#include <graphics3D/TclScene3DObject.h>
#include <graphics3D/ClippingPlane.h>
#include <graphics3D/TclCamera3DActivation.h>
#include <graphics3D/TclLightFantom.h>
#include <set>
#ifndef WIN32
#include <modules/WindowDump.h>
#endif



namespace yav {
	
#define DEFAULT_WIDTH		400
#define DEFAULT_HEIGHT		400




static TclCamera3D *widgetBeingCreated=NULL;


 int createCamera3D(ClientData clientData, Tcl_Interp *interp, int objc,
        Tcl_Obj *const *objv)
{
    struct Togl   *togl;

    if (objc != 2) {
        Tcl_WrongNumArgs(interp, 1, objv, "pathName");
        return TCL_ERROR;
    }

    if (Togl_GetToglFromObj(interp, objv[1], &togl) != TCL_OK) {
        return TCL_ERROR;
    }

	if (widgetBeingCreated==NULL) {
		Tcl_AppendResult(interp, "createCamera3D : ", "Couldn't find a TclCamera3D being created \n", NULL);
        return TCL_ERROR;

	} else {
		widgetBeingCreated->setWidget(togl);
		Togl_SetClientData(togl,(ClientData) widgetBeingCreated);
		widgetBeingCreated->getClient()->setWidth(Togl_Width(togl));
		widgetBeingCreated->getClient()->setHeight(Togl_Height(togl));
		widgetBeingCreated->getClient()->init();
		widgetBeingCreated->getClient()->initViewport();

		widgetBeingCreated->getClient()->newLight();
		widgetBeingCreated->getClient()->setCenter( widgetBeingCreated->getClient()->getRefCenter());
		glFlush();
		/// set fonts 
		widgetBeingCreated->getClient()->setDefaultFont(Togl_FontDisplayList(togl,Togl_LoadBitmapFont(togl,TOGL_BITMAP_8_BY_13)));
		widgetBeingCreated->getClient()->setSmallFont(Togl_FontDisplayList(togl,Togl_LoadBitmapFont(togl,TOGL_BITMAP_HELVETICA_12)));
		widgetBeingCreated->getClient()->setLargeFont(Togl_FontDisplayList(togl,Togl_LoadBitmapFont(togl,TOGL_BITMAP_TIMES_ROMAN_24)));

		widgetBeingCreated->getClient()->resize();
		widgetBeingCreated->getClient()->focus();

		// getModule()->tickle("tkwait visibility %s",(char *) parentFrame);
		/// set camera activation object to handle auto motion of the camera or selected object
		widgetBeingCreated->getClient()->setCameraActivation((Camera3DActivation *)new TclCamera3DActivation(widgetBeingCreated->getClient()));

		/** select the first data in the scene */
		ObjectsList<Data3D>::iterator it=widgetBeingCreated->getScene()->getClient()->firstData();
		if (it!=widgetBeingCreated->getScene()->getClient()->lastData()) {
			widgetBeingCreated->applyObjectSelection(*it);
		}

		widgetBeingCreated=NULL;
	}
 
    return TCL_OK;
}


 int displayCamera3D(ClientData clientData, Tcl_Interp *interp, int objc,
        Tcl_Obj *const *objv)
{
    Togl   *togl;

    if (objc != 2) {
        Tcl_WrongNumArgs(interp, 1, objv, "pathName");
        return TCL_ERROR;
    }

	if (Togl_GetToglFromObj(interp, objv[1], &togl) != TCL_OK)
		return TCL_ERROR;

//	CERR << "display callback from" << Tk_PathName(Togl_TkWin(togl)) << ENDL;	
	TclCamera3D *cam= (TclCamera3D *) Togl_GetClientData(togl);
	if (!cam) {
		Tcl_AppendResult(interp, "displayCamera3D : ", "Couldn't find a TclCamera3D as clientdata \n", NULL);
		return TCL_ERROR;
	}
	Togl_MakeCurrent(togl);

	cam->getClient()->redraw();
	if(cam->isRecording()) {
		TclString filename;
		filename.sprintf("%s%04u.ppm", (char *) cam->getMovieName(), cam->getFrameNumberAndIncrement());
		cam->dumpToFile(filename);
	}
	Togl_SwapBuffers(togl);
	return TCL_OK;
}

 int resizeCamera3D(ClientData clientData, Tcl_Interp *interp, int objc,
        Tcl_Obj *const *objv)
{
    Togl   *togl;

    if (objc != 2) {
        Tcl_WrongNumArgs(interp, 1, objv, "pathName");
        return TCL_ERROR;
    }

	if (Togl_GetToglFromObj(interp, objv[1], &togl) != TCL_OK)
		return TCL_ERROR;

	TclCamera3D *cam= (TclCamera3D *) Togl_GetClientData(togl);
	if (!cam) {
		Tcl_AppendResult(interp, "resizeCamera3D : ", "Couldn't find a TclCamera3D as clientdata \n", NULL);
		return TCL_ERROR;
	}

	Togl_MakeCurrent(togl);
	cam->getClient()->setWidth(Togl_Width(togl));
    cam->getClient()->setHeight(Togl_Height(togl));

	cam->getClient()->resize();
	return TCL_OK;
}
 int destroyCamera3D(ClientData clientData, Tcl_Interp *interp, int objc,
        Tcl_Obj *const *objv)
{
    Togl   *togl;

    if (objc != 2) {
        Tcl_WrongNumArgs(interp, 1, objv, "pathName");
        return TCL_ERROR;
    }


	if (Togl_GetToglFromObj(interp, objv[1], &togl) != TCL_OK)
		return TCL_ERROR;

//	CERR << "destroy callback from" << Tk_PathName(Togl_TkWin(togl)) << ENDL;	

	TclCamera3D *cam= (TclCamera3D *) Togl_GetClientData(togl);
	if (cam) {
		delete cam;
	}
	return TCL_OK;
}



/************************************************************************/
TclCamera3D::TclCamera3D(TclModule* m, const TclString& parent, TclScene3D *scn,
						 bool stereo ,const std::type_info& id, Camera3D *data)
						 : TclSceneObject(typeid(TclCamera3D), data ? data : new Camera3D( scn->getClient()), m) {
	record=false;
	frameNumber=0;



	getClient()->setWidth( DEFAULT_WIDTH);
	getClient()->setHeight(DEFAULT_HEIGHT);
#ifdef X11
	//		masterDisplayHandle=(getScene()->aDisplay);
#endif
	widgetBeingCreated = this;
	if(parent == EMPTY_STRING) {
		parentFrame.sprintf(".viewer3D%s%u", getObjectName(), getRef());
		menuFrame = parentFrame + ".menu";
		submenuFrame = parentFrame + ".submenu";
		buttonFrame = parentFrame + ".buttons";
		cameraFrame = parentFrame + ".view";

		getModule()->tickle("graphics3D::topViewer3D %s%u ",
			getObjectName(), getRef());
	}
	else {
		parentFrame = parent;

		getModule()->tickle("graphics3D::viewer3D %s%u %s",
			getObjectName(), getRef(), (char *) parent);
	}


	done(typeid(TclCamera3D));
						 }
/************************************************************************/	
void TclCamera3D::makeCurrent(void) const {
	Togl_MakeCurrent(widget);
}
/************************************************************************/
void TclCamera3D::swapBuffers(void) const {
Togl_SwapBuffers(widget);
}

/************************************************************************/
TclCamera3D::~TclCamera3D() 
{
//	CERR << "cam destroyed"<<ENDL;
	if (widget)
		Togl_SetClientData(widget,(ClientData) 0);

	getClient()->destroyCamera();
	deleted(typeid(TclCamera3D));
}

/************************************************************************/
TclScene3D *TclCamera3D::getScene(void) const {
	return (TclScene3D *)(getClient()->getScene()->getInterface());
}
/************************************************************************/
void TclCamera3D::startMovie(const char *name) {
	movieName = name;
	frameNumber = 0;
	record = true;
}
/************************************************************************/
void TclCamera3D::stopMovie(void) {
	record = false;
}
/************************************************************************/
void TclCamera3D::pauseMovie(bool b) {
	record = !b;
}

/************************************************************************/
void TclCamera3D::setSelectedObject(Data3D *data) {
	if(!data) {
		if(getClient()->getSelectedObject())
			applyObjectUnselection(getClient()->getSelectedObject());
		getClient()->selectedObject = 0;
		if(getModule()->windowExists((char *) submenuFrame)) {
			getModule()->tickle("%s.name configure -text \"no selected object\"",
				(char *) submenuFrame);
		}
	}
}
/************************************************************************/
void TclCamera3D::applyObjectUnselection(Data3D *obj) {
	if (obj!=0) {
		getClient()->unsetSelectedObject(obj);
		if(obj->getInterface()) {
			if(getModule()->windowExists((char *) submenuFrame)) {
				getModule()->tickle("%s.name configure -text \"no selected object\"",
					(char *) submenuFrame);
				TclScene3DObject *tclsobj =
					dynamic_cast<TclScene3DObject *>(obj->getInterface());
				if (tclsobj)
					/// call callback function for unselection
					tclsobj->unselectionCallback(this);
			}
		}
	}
}
/************************************************************************/
void TclCamera3D::applyObjectSelection (Data3D *obj){
	if (obj!=getClient()->getSelectedObject()) {
		applyObjectUnselection(getClient()->getSelectedObject());
		getClient()->setSelectedObject(obj);
		if(obj->getInterface()) {
			if(getModule()->windowExists((char *) submenuFrame)) {
				TclObject *tclobj = dynamic_cast<TclObject *>(obj->getInterface());
				TclScene3DObject *tclsobj =
					dynamic_cast<TclScene3DObject *>(obj->getInterface());
				
				getModule()->tickle("%s.name configure -text \"%s\"",
					(char *) submenuFrame, (char *) tclobj->name());
				if (tclsobj)
					// call the callback function to warn the object it is being selected
					tclsobj->selectionCallback(this);
			}
		}
	}
}
/************************************************************************/
int TclCamera3D::destroyCmd(int argc, char **argv) {
	if(argc > 0) return COMMAND_SYNOPSIS;
	TclModule *m = getModule();
	TclString n=name();
	if(getModule()->windowExists((char *) getParent()))
		getModule()->tickle("destroy %s", (char *) getParent());
	else
		getModule()->tickle("destroy %s", (char *) getCameraFrame());

	m->setResult("%s does not exist any longer",
		(char *) n);
	return TCL_OK;
}
/************************************************************************/
int TclCamera3D::transformationMatrixCmd(int argc, char **argv) {
	if((argc!=0) && (argc!=16) && (argc!=1)) return COMMAND_SYNOPSIS;
	if(argc==0){
		Matrix4x4<double> mat = getClient()->getMatTransform();
		TclMatrix4D *m= new TclMatrix4D(getModule(),mat);
		setResult(m->TclObject::name());
		return TCL_OK;
	} else if (argc==16){
		int i;
		double coeff[16];
		for(i=0;i<16;i++)
			if(sscanf(argv[i],"%lf",&coeff[i])!=1){
				return COMMAND_SYNOPSIS;
			}
			Matrix4x4<double> mat(coeff[0],coeff[1],coeff[2],coeff[3],coeff[4],coeff[5],coeff[6],coeff[7],coeff[8],coeff[9],coeff[10],coeff[11],coeff[12],coeff[13],coeff[14],coeff[15]);
			getClient()->setMatTransform(mat);
			redraw();
			return TCL_OK;
	} else {
		TclMatrix4D *m= (TclMatrix4D *)  TclMatrix4D::getObject(argc, argv, 0);
		if (m==0) {
			setResult("%s must be a 4D matrix",argv[0]);
			return TCL_ERROR;
		}
		else {
			getClient()->setMatTransform(*m);
			redraw();
			return TCL_OK;
		}
	}
}
/************************************************************************/
int TclCamera3D::viewPointCmd(int argc, char **argv) {
	if((argc!=0) && (argc!=23)) return COMMAND_SYNOPSIS;
	if(argc==0){
		Vec3<double> center = getClient()->getCenter();
		Vec3<double> refCenter = getClient()->getRefCenter();
		double scale = getClient()->getScale();
		Matrix4x4<double> mat = getClient()->getMatTransform();
		setResult("%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f",center.getX(),center.getY(),center.getZ(),refCenter.getX(),refCenter.getY(),refCenter.getZ(),scale,mat[0][0],mat[0][1],mat[0][2],mat[0][3],mat[1][0],mat[1][1],mat[1][2],mat[1][3],mat[2][0],mat[2][1],mat[2][2],mat[2][3],mat[3][0],mat[3][1],mat[3][2],mat[3][3]);
		return TCL_OK;
	}
	else {
		int i;
		double c[3];
		for(i=0;i<3;i++)
			if(sscanf(argv[i],"%lf",&c[i])!=1){
				return COMMAND_SYNOPSIS;
			}
			Vec3<double> center(c[0],c[1],c[2]);
			getClient()->setCenter(center);
			for(i=0;i<3;i++)
				if(sscanf(argv[i+3],"%lf",&c[i])!=1){
					return COMMAND_SYNOPSIS;
				}
				center.setValue(c[0],c[1],c[2]);
				getClient()->setRefCenter(center);
				double scale;
				if(sscanf(argv[6],"%lf",&scale)!=1)
					return COMMAND_SYNOPSIS;
				getClient()->setScale(scale);
				double coeff[16];
				for(i=0;i<16;i++)
					if(sscanf(argv[i+7],"%lf",&coeff[i])!=1){
						return COMMAND_SYNOPSIS;
					}
					Matrix4x4<double> mat(coeff[0],coeff[1],coeff[2],coeff[3],coeff[4],coeff[5],coeff[6],coeff[7],coeff[8],coeff[9],coeff[10],coeff[11],coeff[12],coeff[13],coeff[14],coeff[15]);
					getClient()->setMatTransform(mat);
					redraw();
					return TCL_OK;
	}
}
/************************************************************************/
int TclCamera3D::centerCmd(int argc, char **argv) {
	if((argc!=0) && (argc!=3)) return COMMAND_SYNOPSIS;
	if(argc==0){
		Vec3<double> center = getClient()->getCenter();
		setResult("%f %f %f",center.getX(),center.getY(),center.getZ());
		return TCL_OK;
	}
	int i;
	double c[3];
	for(i=0;i<3;i++)
		if(sscanf(argv[i],"%lf",&c[i])!=1){
			return COMMAND_SYNOPSIS;
		}
		Vec3<double> center(c[0],c[1],c[2]);
		getClient()->setCenter(center);
		redraw();
		return TCL_OK;
}

/************************************************************************/
int TclCamera3D::refCenterCmd(int argc, char **argv) {
	if((argc!=0) && (argc!=3)) return COMMAND_SYNOPSIS;
	if(argc==0){
		Vec3<double> refCenter = getClient()->getRefCenter();
		setResult("%f %f %f",refCenter.getX(),refCenter.getY(),refCenter.getZ());
		return TCL_OK;
	}
	int i;
	double c[3];
	for(i=0;i<3;i++)
		if(sscanf(argv[i],"%lf",&c[i])!=1){
			return COMMAND_SYNOPSIS;
		}
		Vec3<double> refCenter(c[0],c[1],c[2]);
		getClient()->setRefCenter(refCenter);
		redraw();
		return TCL_OK;
		
}
/************************************************************************/
int TclCamera3D::scaleFactorCmd(int argc, char **argv) {
	if((argc!=0) && (argc!=1)) return COMMAND_SYNOPSIS;
	if(argc==0){
		double scale = getClient()->getScale();
		setResult("%f",scale);
		return TCL_OK;
	}
	double scale;
	if(sscanf(argv[0],"%lf",&scale)!=1)
		return COMMAND_SYNOPSIS;
	getClient()->setScale(scale);
	redraw();
	return TCL_OK;
}

/************************************************************************/
int TclCamera3D::clippingPlanesCmd(int argc, char **argv) {

	if (argc != 0) return COMMAND_SYNOPSIS;
	setResult("%u",getClient()->getNumberOfClippingPlanes());
	return TCL_OK;
}
/*************************************************************************/
int TclCamera3D::clippingPlaneCreateCmd(int argc, char **argv) {
	if (argc != 0) return COMMAND_SYNOPSIS;
	Camera3D *cam=getClient();
	cam->newClippingPlane();
	redraw();
	setResult("Clipping Plane Created");
	return TCL_OK;
}
/*************************************************************************/
int TclCamera3D::clippingPlaneRemoveCmd(int argc, char **argv) {
	
	if (argc != 1) return COMMAND_SYNOPSIS;
	ClippingPlane *l=getClient()->getClippingPlane((unsigned int)atoi(argv[0]));
	if (l==0) {
		setResult("%s is not a correct clipping plane index",argv[0]);
		return TCL_ERROR;
	} 
	else 
	{
		getClient()->removeClippingPlane(l);
		redraw();
	}
	setResult("Clipping Plane Removed");
	return TCL_OK;
}
/*************************************************************************/
int TclCamera3D::clippingPlaneEditCmd(int argc, char **argv) {
	
	if (argc <2) return COMMAND_SYNOPSIS;
	ClippingPlane *l=getClient()->getClippingPlane((unsigned int)atoi(argv[0]));
	if (l==0) {
		setResult("%s is not a correct clipping plane index",argv[0]);
		return TCL_ERROR;
	}
	else 
	{
		if (!strcmp(argv[1],"active")) {
			if (argc>3) 
				return COMMAND_SYNOPSIS;
			if (argc==3) {
				bool val;
				if(!TclModule::boolValue(argv[2], val)) 
					return COMMAND_SYNOPSIS;
				if(val) l->enable();
				else l->disable();
				redraw();
			}
			setResult(TclModule::boolString(l->isActive()));
			return TCL_OK;
		}
		else if (!strcmp(argv[1],"changeActivation")) {
			if (argc!=2) 
				return COMMAND_SYNOPSIS;
			if (l->isActive())
				l->disable();
			else
				l->enable();
			redraw();
			return TCL_OK;
		}
		else if (!strcmp(argv[1],"swap")) {
			const double *equation=l->getEquation();
			GLdouble _equation[4];
			unsigned int i;
			for(i = 0; i < 4; i++) _equation[i] = -equation[i];
			l->setEquation(_equation);
			redraw();
			setResult("swapped");
			return TCL_OK;
		}
		else if (!strcmp(argv[1],"moving")) {
			if (argc!=2) 
				return COMMAND_SYNOPSIS;
			
			setResult(TclModule::boolString(l->isMoving()));
			return TCL_OK;	
		}
		else if (!strcmp(argv[1],"changeMotion")) {
			if (argc!=2) 
				return COMMAND_SYNOPSIS;
			if (l->isMoving()) {
				l->fixPlane();
				getClient()->setMotionMode(Camera3D::MOVE_CAMERA);
			} else {
				l->movePlane();
				getClient()->setMotionMode(Camera3D::MOVE_CAMERA_OBJECT);	}
			setResult(TclModule::boolString(l->isMoving()));
			return TCL_OK;	
		}		else
			return COMMAND_SYNOPSIS;
	}
}
/*************************************************************************/
int TclCamera3D::lightCreateCmd(int argc, char **argv) {
	if (argc != 0) return COMMAND_SYNOPSIS;
	Camera3D *cam=getClient();
	cam->newLight();
	redraw();
	setResult("Light Created");
	return TCL_OK;
}
/*************************************************************************/
int TclCamera3D::lightsCmd(int argc, char **argv) {

	if (argc != 0) return COMMAND_SYNOPSIS;
	setResult("%u",getClient()->getNumberOfLights());
	return TCL_OK;
}
/*************************************************************************/
int TclCamera3D::lightRemoveCmd(int argc, char **argv) {
	
	if (argc != 1) return COMMAND_SYNOPSIS;
	Light *l=getClient()->getLight((unsigned int)atoi(argv[0]));
	if (l==0) {
		setResult("%s is not a correct light index",argv[0]);
		return TCL_ERROR;
	} 
	else 
	{
		getClient()->removeLight(l);
		redraw();
	}
	setResult("Light Removed");
	return TCL_OK;
}
/*************************************************************************/
int TclCamera3D::lightEditCmd(int argc, char **argv)
{
	if (argc <2) return COMMAND_SYNOPSIS;
	Light *l=getClient()->getLight((unsigned int)atoi(argv[0]));
	if (l==0) {
		setResult("%s is not a correct light index",argv[0]);
		return TCL_ERROR;
	}
	else 
	{
		if (!strcmp(argv[1],"active")) {
			if (argc>3) 
				return COMMAND_SYNOPSIS;
			if (argc==3) {
				bool val;
				if(!TclModule::boolValue(argv[2], val)) 
					return COMMAND_SYNOPSIS;
				if(val) l->enable();
				else l->disable();
				redraw();
			}
			setResult(TclModule::boolString(l->isActive()));
			return TCL_OK;
		}
		else if (!strcmp(argv[1],"changeActivation")) {
			if (argc!=2) 
				return COMMAND_SYNOPSIS;
			if (l->isActive())
				l->disable();
			else
				l->enable();
			return TCL_OK;
		}
		else if (!strcmp(argv[1],"show")) {
			if (argc>3) 
				return COMMAND_SYNOPSIS;
			if (argc==3) {
				bool val;
				
				if(!TclModule::boolValue(argv[2], val)) return COMMAND_SYNOPSIS;
				if(val) {
					if((l->getFantom()==0) && (l->getPosition()[3] > 0.0)) 
					{
						TclLightFantom *interf = new TclLightFantom(l,this);
						l->setFantom(interf->getClient());
						setResult("light%s is displayed",argv[0]);
					}
				}
				else {
					if(l->getFantom()) {
						delete l->getFantom();
						l->setFantom(0);				
					}
					setResult("light%s is hidden",argv[0]);
				}
				redraw();
				return TCL_OK;
			} else {
				if (l->getFantom()==0) {
					setResult("false");
				} else {
					setResult("true");
				}
				return TCL_OK;
			}
		}
		else if (!strcmp(argv[1],"edit")){
			
			if ((argc>2) && ((!strcmp(argv[2], "ambient") || !strcmp(argv[2], "diffuse")
				|| !strcmp(argv[2], "specular") || !strcmp(argv[2], "position")
				|| !strcmp(argv[2], "direction") || !strcmp(argv[2], "exponent")
				|| !strcmp(argv[2], "cutoff") || !strcmp(argv[2], "attenuation")))) {
				GLfloat coef[4];
				unsigned int i;
				
				for(i = 2; i < (unsigned int) argc; i++) {
					if(!strcmp(argv[i], "ambient")) {
						if(i+3 >= (unsigned int)argc) return COMMAND_SYNOPSIS;
						if(sscanf(argv[++i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
						if(sscanf(argv[++i], "%f", &coef[1]) != 1) return COMMAND_SYNOPSIS;
						if(sscanf(argv[++i], "%f", &coef[2]) != 1) return COMMAND_SYNOPSIS;
						if(i+1 >= (unsigned int)argc || !strcmp(argv[i+1], "diffuse")
							|| !strcmp(argv[i+1], "specular") || !strcmp(argv[i+1], "position")
							|| !strcmp(argv[i+1], "direction") || !strcmp(argv[i+1], "exponent")
							|| !strcmp(argv[i+1], "cutoff") || !strcmp(argv[i+1], "attenuation"))
							coef[3] = 1.0;
						else {
							if(sscanf(argv[++i], "%f", &coef[3]) != 1) return COMMAND_SYNOPSIS;
						}
						l->setAmbient(coef);
					}
					
					else if(!strcmp(argv[i], "diffuse")) {
						if(i+3 >= (unsigned int)argc) return COMMAND_SYNOPSIS;
						if(sscanf(argv[++i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
						if(sscanf(argv[++i], "%f", &coef[1]) != 1) return COMMAND_SYNOPSIS;
						if(sscanf(argv[++i], "%f", &coef[2]) != 1) return COMMAND_SYNOPSIS;
						if(i+1 >= (unsigned int)argc || !strcmp(argv[i+1], "ambient")
							|| !strcmp(argv[i+1], "specular") || !strcmp(argv[i+1], "position")
							|| !strcmp(argv[i+1], "direction") || !strcmp(argv[i+1], "exponent")
							|| !strcmp(argv[i+1], "cutoff") || !strcmp(argv[i+1], "attenuation"))
							coef[3] = 1.0;
						else {
							if(sscanf(argv[++i], "%f", &coef[3]) != 1) return COMMAND_SYNOPSIS;
						}
						l->setDiffuse(coef);
					}
					
					else if(!strcmp(argv[i], "specular")) {
						if(i+3 >= (unsigned int)argc) return COMMAND_SYNOPSIS;
						if(sscanf(argv[++i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
						if(sscanf(argv[++i], "%f", &coef[1]) != 1) return COMMAND_SYNOPSIS;
						if(sscanf(argv[++i], "%f", &coef[2]) != 1) return COMMAND_SYNOPSIS;
						if(i+1 >= (unsigned int)argc || !strcmp(argv[i+1], "ambient")
							|| !strcmp(argv[i+1], "diffuse") || !strcmp(argv[i+1], "position")
							|| !strcmp(argv[i+1], "direction") || !strcmp(argv[i+1], "exponent")
							|| !strcmp(argv[i+1], "cutoff") || !strcmp(argv[i+1], "attenuation"))
							coef[3] = 1.0;
						else {
							if(sscanf(argv[++i], "%f", &coef[3]) != 1) return COMMAND_SYNOPSIS;
						}
						l->setSpecular(coef);
					}
					
					else if(!strcmp(argv[i], "position")) {
						if(i+4 >= (unsigned int)argc) return COMMAND_SYNOPSIS;
						if(sscanf(argv[++i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
						if(sscanf(argv[++i], "%f", &coef[1]) != 1) return COMMAND_SYNOPSIS;
						if(sscanf(argv[++i], "%f", &coef[2]) != 1) return COMMAND_SYNOPSIS;
						if(sscanf(argv[++i], "%f", &coef[3]) != 1) return COMMAND_SYNOPSIS;
						l->setPosition(coef);
					}
					
					else if(!strcmp(argv[i], "direction")) {
						if(i+3 >= (unsigned int) argc) return COMMAND_SYNOPSIS;
						if(sscanf(argv[++i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
						if(sscanf(argv[++i], "%f", &coef[1]) != 1) return COMMAND_SYNOPSIS;
						if(sscanf(argv[++i], "%f", &coef[2]) != 1) return COMMAND_SYNOPSIS;
						l->setDirection(coef);
					}
					
					else if(!strcmp(argv[i], "exponent")) {
						if(++i >= (unsigned int)argc) return COMMAND_SYNOPSIS;
						if(sscanf(argv[i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
						l->setExponent(coef[0]);
					}
					
					else if(!strcmp(argv[i], "cutoff")) {
						if(++i >= (unsigned int)argc) return COMMAND_SYNOPSIS;
						if(sscanf(argv[i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
						l->setCutoff(coef[0]);
					}
					
					else if(!strcmp(argv[i], "attenuation")) {
						if(i+3 >= (unsigned int)argc) return COMMAND_SYNOPSIS;
						if(sscanf(argv[++i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
						if(sscanf(argv[++i], "%f", &coef[1]) != 1) return COMMAND_SYNOPSIS;
						if(sscanf(argv[++i], "%f", &coef[2]) != 1) return COMMAND_SYNOPSIS;
						l->setAttenuation(coef);
					}
				}
			
			}
			
			else if (argc>2) {
				GLfloat ambient[4], diffuse[4], specular[4], shininess;
				if(2 + 12 < argc) {
					if(sscanf(argv[0], "%f", &ambient[0]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[1], "%f", &ambient[1]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[2], "%f", &ambient[2]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[3], "%f", &ambient[3]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[4], "%f", &diffuse[0]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[5], "%f", &diffuse[1]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[6], "%f", &diffuse[2]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[7], "%f", &diffuse[3]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[8], "%f", &specular[0]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[9], "%f", &specular[1]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[10], "%f", &specular[2]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[11], "%f", &specular[3]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[12], "%f", &shininess) != 1) return COMMAND_SYNOPSIS;
					l->setAmbient(ambient);
					l->setDiffuse(diffuse);
					l->setSpecular(specular);
				}
				else if(2 + 9 < argc) {
					if(sscanf(argv[0], "%f", &ambient[0]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[1], "%f", &ambient[1]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[2], "%f", &ambient[2]) != 1) return COMMAND_SYNOPSIS;
					ambient[3] = 1.0f;
					if(sscanf(argv[4], "%f", &diffuse[0]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[5], "%f", &diffuse[1]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[6], "%f", &diffuse[2]) != 1) return COMMAND_SYNOPSIS;
					diffuse[3] = 1.0f;
					if(sscanf(argv[8], "%f", &specular[0]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[9], "%f", &specular[1]) != 1) return COMMAND_SYNOPSIS;
					if(sscanf(argv[10], "%f", &specular[2]) != 1) return COMMAND_SYNOPSIS;
					specular[3] = 1.0f;
					l->setAmbient(ambient);
					l->setDiffuse(diffuse);
					l->setSpecular(specular);
				}
				else 
					return COMMAND_SYNOPSIS;
			}
			redraw();
			
			const GLfloat *_ambient=l->getAmbient();
			const GLfloat *_diffuse=l->getDiffuse();
			const GLfloat *_specular=l->getSpecular();
			const GLfloat *_position=l->getPosition();
			const GLfloat *_direction=l->getDirection();
			const GLfloat *_attenuation=l->getAttenuation();
			const GLfloat _exponent=l->getExponent();
			const GLfloat _cutoff=l->getCutoff();
			setResult("ambient %f %f %f %f diffuse %f %f %f %f specular %f %f %f %f position %f %f %f %f direction %f %f %f exponent %f cutoff %f attenuation %f %f %f",
				_ambient[0], _ambient[1], _ambient[2], _ambient[3],
				_diffuse[0], _diffuse[1], _diffuse[2], _diffuse[3],
				_specular[0], _specular[1], _specular[2], _specular[3],
				_position[0], _position[1], _position[2], _position[3],
				_direction[0], _direction[1], _direction[2],
				_exponent, _cutoff,
				_attenuation[0], _attenuation[1], _attenuation[2]);
			return TCL_OK;
	} else
return TCL_OK;
	}
}

/*************************************************************************/
int TclCamera3D::boundingBoxColorCmd(int argc, char **argv) {
	if ((argc!=0)&& (argc!=3))  return COMMAND_SYNOPSIS;
	if (argc==0)
    {
		float r,g,b;
		getClient()->getBoundingBoxColor(r,g,b);
		getModule()->setResult("%f %f %f",r,g,b);
		return TCL_OK;
    }
	else
    {
		float r,g,b;
		if(sscanf(argv[0], "%f", &r) == -1 || r < 0.0 || r > 1.0)
			return COMMAND_SYNOPSIS;
		if(sscanf(argv[1], "%f", &g) == -1 || g < 0.0 || g > 1.0)
			return COMMAND_SYNOPSIS;
		if(sscanf(argv[2], "%f", &b) == -1 || b < 0.0 || b > 1.0) 
			return COMMAND_SYNOPSIS;
		getClient()->setBoundingBoxColor(r,g,b);
		redraw();
		getModule()->setResult("%f %f %f",r,g,b);
		return TCL_OK;
	}
}
/*************************************************************************/
int TclCamera3D::boundingBoxWidthCmd(int argc, char **argv) {
	if ((argc!=0)&& (argc!=1))  return COMMAND_SYNOPSIS;
	
	if (argc==0)
    {
		float w;
		getClient()->getBoundingBoxWidth(w);
		getModule()->setResult("%f ",w);
		return TCL_OK;
    }
	else
    {
		float w;
		w=(float)atof(argv[0]);
		getClient()->setBoundingBoxWidth(w);
		redraw();
		getModule()->setResult("%f ",w);
		return TCL_OK;
    }
}
/*************************************************************************/
int TclCamera3D::boundingBoxDisplayCmd(int argc, char **argv) {
	if ((argc!=0)&& (argc!=1))  return COMMAND_SYNOPSIS;
	if (argc==0)
    {
		getModule()->setResult(TclModule::boolString(getClient()->getShowBoundingBox()));
		return TCL_OK;
    }
	else
    {
		if (!strcmp(argv[0],"true"))
			getClient()->displayBoundingBox();
		else if (!strcmp(argv[0],"false"))
			getClient()->hideBoundingBox();
		else  return COMMAND_SYNOPSIS;
		redraw();
		getModule()->setResult(TclModule::boolString(getClient()->getShowBoundingBox()));
		return TCL_OK;
    }
}

/*************************************************************************/
int TclCamera3D::bgColorCmd(int argc, char **argv) {
	if ((argc!=0)&& (argc!=3))  return COMMAND_SYNOPSIS;
	if (argc==0)
    {
		float r,g,b;
		getClient()->getBackgroundColor(r,g,b);
		getModule()->setResult("%f %f %f",r,g,b);
		return TCL_OK;
    }
	else
    {
		float r,g,b;
		if(sscanf(argv[0], "%f", &r) == -1 || r < 0.0 || r > 1.0)
			return COMMAND_SYNOPSIS;
		if(sscanf(argv[1], "%f", &g) == -1 || g < 0.0 || g > 1.0)
			return COMMAND_SYNOPSIS;
		if(sscanf(argv[2], "%f", &b) == -1 || b < 0.0 || b > 1.0) 
			return COMMAND_SYNOPSIS;
		getClient()->setBackgroundColor(r,g,b);
		redraw();
		getModule()->setResult("%f %f %f",r,g,b);
		return TCL_OK;
	}
}
/*************************************************************************/
int TclCamera3D::sceneCmd(int argc, char **argv) {
	if (argc!=0) return COMMAND_SYNOPSIS;
	getModule()->setResult("%s%u",
		TclObjectKind::getObjectName(typeid(TclScene3D)),
		getScene()->getRef());
	return TCL_OK;
}
/*************************************************************************/

int TclCamera3D::widthCmd(int argc, char **argv) {
	if ((argc!=0)&&(argc!=1)) return COMMAND_SYNOPSIS;
	if (argc==0)
    {
		getModule()->setResult("%u", getClient()->getWidth());
		return TCL_OK;
    }
	else
    {
		unsigned int u;
		if(sscanf(argv[0], "%u", &u) != 1) return COMMAND_SYNOPSIS;
		resize(u, getClient()->getHeight());
		getModule()->setResult("%u",getClient()->getWidth()); 
		return TCL_OK;
    }
}
/*************************************************************************/
void TclCamera3D::resize(unsigned int w, unsigned int h) {
	getClient()->resize(w, h);
	Tk_GeometryRequest(Togl_TkWin(widget), getClient()->width, getClient()->height);  
}

int TclCamera3D::heightCmd(int argc, char **argv) {
	if ((argc!=0)&&(argc!=1)) return COMMAND_SYNOPSIS;
	if (argc==0)
    {
		getModule()->setResult("%u", getClient()->getHeight());
		return TCL_OK;
    }
	else
    {
		unsigned int u;
		if(sscanf(argv[0], "%u", &u) != 1)  return COMMAND_SYNOPSIS;
		resize(getClient()->getWidth(),u);
		getModule()->setResult("%u",getClient()->getHeight()); 
		return TCL_OK;
    }
}
/*************************************************************************/
int TclCamera3D::fieldOfViewCmd(int argc, char **argv) {
	if (argc==0)
    {
		getModule()->setResult("%f",getClient()->getFieldOfView());
		return TCL_OK;
    }
	else if (argc==1)
    {
		getClient()->setFieldOfView((double)(atof(argv[0])));
		getClient()->setViewingParameters();
		getClient()->resize();
		redraw();
		getModule()->setResult("%f",getClient()->getFieldOfView());
		return TCL_OK;
    }
	else return COMMAND_SYNOPSIS;
    
}
int TclCamera3D::eyeOffsetCmd(int argc, char **argv) {
	if ((argc!=0)&&(argc!=3)) return COMMAND_SYNOPSIS;
	if (argc==0)
    {
		getModule()->setResult("%g %g %g",
			getClient()->getEyeOffset()[0],
			getClient()->getEyeOffset()[1],
			getClient()->getEyeOffset()[2]);
		return TCL_OK;
    }
	else
    {
		double x,y,z;
		if(sscanf(argv[0], "%lf", &x) != 1)  return COMMAND_SYNOPSIS;
		if(sscanf(argv[1], "%lf", &y) != 1)  return COMMAND_SYNOPSIS;
		if(sscanf(argv[2], "%lf", &z) != 1)  return COMMAND_SYNOPSIS;
		
		getClient()->setEyeOffset(Vec3<double>(x,y,z));
		return TCL_OK;
    }
}
/*************************************************************************/
/* indicates that the scene should be moved instead of the selected object */
int TclCamera3D::motionModeCmd(int argc, char **argv) {
	if(argc > 1) return COMMAND_SYNOPSIS;
	if(argc == 1) {
		if(!strcmp(argv[0], "camera")) {
			getClient()->setMotionMode(Camera3D::MOVE_CAMERA);
		}
		else if(!strcmp(argv[0], "cameraObject")) {
			getClient()->setMotionMode(Camera3D::MOVE_CAMERA_OBJECT);
		}
		else if(!strcmp(argv[0], "object")) {
			getClient()->setMotionMode(Camera3D::MOVE_SELECTED_OBJECT);
		}
		else return COMMAND_SYNOPSIS;
	}
	
	switch(getClient()->getMotionMode()) {
	case Camera3D::MOVE_CAMERA:
		setResult("camera");      
		break;
	case Camera3D::MOVE_CAMERA_OBJECT:
		setResult("cameraObject");
		break;
	case Camera3D::MOVE_SELECTED_OBJECT:
		setResult("object");
		break;
	}    
	
	return TCL_OK;
}


/*************************************************************************/

/* mouse motion: move model */
int TclCamera3D::mouseMotionCmd(int argc, char **argv) {
	if (argc!=2)  return COMMAND_SYNOPSIS;
	unsigned int x, y;
	
	if(sscanf(argv[0], "%u", &x) != 1) return COMMAND_SYNOPSIS;
	if(sscanf(argv[1], "%u", &y) != 1) return COMMAND_SYNOPSIS;
	getClient()->mouseMoved(x, y);
	
	return TCL_OK;
}
/*************************************************************************/
/**********************************************************************/
/**************** navigate with camera ********************************/
/**********************************************************************/



int TclCamera3D::useTwoSidesCmd(int argc, char **argv) {
	if (argc==0)
    {  
		setResult(TclModule::boolString(getClient()->useTwoSides));
		return TCL_OK;
    }
	else if (argc==1) {
		bool twoSides;
		if(!TclModule::boolValue(argv[0], twoSides)) return COMMAND_SYNOPSIS;
		getClient()->useTwoSides=twoSides;
		makeCurrent();
		getClient()->setRenderFrontAndBack(twoSides);
  	    getClient()->initViewport();
		redraw();
		return TCL_OK;
	}
	else
		return COMMAND_SYNOPSIS;
}

//*************************************************************************/
int TclCamera3D::focusCmd(int argc, char **argv) {
	if (argc!=0)  return COMMAND_SYNOPSIS;
	makeCurrent();
	getClient()->doFocus();
	redraw();
	getModule()->setResult("%s%u focused on scene", getObjectName(), 
		getRef());
	return TCL_OK;
}
/*************************************************************************/
int TclCamera3D::focusOnRegionCmd(int argc, char **argv) {
	if (argc!=4)  return COMMAND_SYNOPSIS;
	Vec3<double> c= Vec3<double>((double) atof(argv[0]),
		(double) atof(argv[1]),
		(double) atof(argv[2]));
	makeCurrent();
	getClient()->focusOnRegion(c,(double)atof(argv[3]));
	redraw();
	getModule()->setResult("%s%u focused on region", getObjectName(), 
		getRef());
	return TCL_OK;
}
/*************************************************************************/
int TclCamera3D::autofocusCmd(int argc, char **argv) {
	if (argc>1)  return COMMAND_SYNOPSIS;
	if (argc==0) {
		getModule()->setResult(TclModule::boolString(getClient()->getAutoFocus()));
		return TCL_OK;
	}
	else {
		bool bval;
		if(TclModule::boolValue(argv[0], bval)) {
			getClient()->setAutoFocus(bval);
			getModule()->setResult(TclModule::boolString(getClient()->
				getAutoFocus()));
			return TCL_OK;
		}
		else
			return COMMAND_SYNOPSIS;
	}
}
/*************************************************************************/
int TclCamera3D::widgetCmd(int argc, char **argv) {
	if (argc>1) {
		if (!strcmp(argv[0],"menu")) {
			if (argc==1) {
				setResult(menuFrame);
				return TCL_OK;
			}
			else {
				menuFrame=argv[1];
				return TCL_OK;
			}
		}
		else if (!strcmp(argv[0],"submenu")) {
			if (argc==1) {
				setResult(submenuFrame);
				return TCL_OK;
			}
			else {
				submenuFrame=argv[1];
				return TCL_OK;
			}
		}
		else if (!strcmp(argv[0],"button")) {
			if (argc==1) {
				setResult(buttonFrame);
				return TCL_OK;
			}
			else {
				buttonFrame=argv[1];
				return TCL_OK;
			}
		}
		else {
			setResult("%s must be either menu, submenu or button",argv[0]);
			return TCL_ERROR;
		}
	}
	else
    {
		getModule()->setResult("%s",(char *) cameraFrame);
		return TCL_OK;
    }
}
/*************************************************************************/
int TclCamera3D::menuCmd(int argc, char **argv) {
	if (argc>1)  return COMMAND_SYNOPSIS;
	else
    {
		getModule()->setResult("%s",(char *) submenuFrame);
		return TCL_OK;
    }
}

/*************************************************************************/
int TclCamera3D::objectsCmd(int argc, char **argv) {
	if (argc!=0)  return COMMAND_SYNOPSIS;
	TclString output;
	ObjectsList<Data3D>::iterator it = getScene()->getClient()->firstData();
	for(;   it!= getScene()->getClient()->lastData(); it++)
    {
		Interface *interf = (*it)->getInterface();
		if(interf) {
			output+= (dynamic_cast<TclObject *>(interf))->name();
			output+= SPACE_STRING;
		}
    }
	getModule()->setResult(output);
	return TCL_OK;
}

int TclCamera3D::fogCmd(int argc, char **argv) {
	bool usefog;
	
	if(argc > 1) return COMMAND_SYNOPSIS;
	if(argc == 1) {
		if(!TclModule::boolValue(argv[0], usefog)) return COMMAND_SYNOPSIS;
		getClient()->setFog(usefog);
		redraw();
	}
	setResult(TclModule::boolString(getClient()->getFog()));
	return TCL_OK;
}

int TclCamera3D::fogRangeCmd(int argc, char **argv) {
	double range;
	
	if(argc > 1) return COMMAND_SYNOPSIS;
	if(argc == 1) {
		if(sscanf(argv[0],"%lf",&range)!=1) return COMMAND_SYNOPSIS;
		getClient()->fogRange=range;
		getClient()->setViewingParameters();
		redraw();
	}
	setResult("%f",getClient()->fogRange);
	return TCL_OK;
}

int TclCamera3D::dumpToPPMCmd(int argc, char **argv) {
    if(argc != 1) return COMMAND_SYNOPSIS;
    dumpToFile(argv[0]);
    setResult("file dumped as %s",argv[0]);
    return TCL_OK;
}

int TclCamera3D::stencilBufferCmd(int argc, char **argv) {
	bool usestencil;
	
	if(argc > 1) return COMMAND_SYNOPSIS;
	if(argc == 1) {
		if(!TclModule::boolValue(argv[0], usestencil)) return COMMAND_SYNOPSIS;
		getClient()->setStencilBuffer(usestencil);
		if(usestencil)  getClient()->buildStencilBufferDisplayList();
		redraw();
	}
	setResult(TclModule::boolString(getClient()->stencilBuffer));
	return TCL_OK;
}

int TclCamera3D::specularityCmd(int argc, char **argv) {
	bool flag;
	
	if(argc > 1) return COMMAND_SYNOPSIS;
	if(argc == 1) {
		if(!TclModule::boolValue(argv[0], flag)) return COMMAND_SYNOPSIS;
		getClient()->setSpecularity(flag);
		redraw();
	}
	setResult(TclModule::boolString(getClient()->getSpecularity()));
	return TCL_OK;
}

int TclCamera3D::renderFrontAndBackCmd(int argc, char **argv) {
	bool flag;
	
	if(argc > 1) return COMMAND_SYNOPSIS;
	if(argc == 1) {
		if(!TclModule::boolValue(argv[0], flag)) return COMMAND_SYNOPSIS;
		getClient()->setRenderFrontAndBack(flag);
	}
	setResult(TclModule::boolString(getClient()->getUseTwoSides()));
	return TCL_OK;
}


void TclCamera3D::startComputation() {
  getModule()->tickle("graphics3D::startViewer3D  %s %s", (char *) buttonFrame, (char *) name());
}

void TclCamera3D::stopComputation() {
  getModule()->tickle("graphics3D::stopViewer3D  %s %s", (char *) buttonFrame,(char *) name() );
}





/*
 * Called when the widget's contents must be redrawn.  Basically, we
 * just call the user's render callback function.
 *
 * Note that the parameter type is ClientData so this function can be
 * passed to Tk_DoWhenIdle().
 */

void TclCamera3D::redraw(void) {
	if (!Togl_UpdatePending(widget))
		Togl_PostRedisplay(widget);
}


void TclCamera3D::bindMouseMotion(bool on){
  if (on==true)
    {
      getModule()->tickle("bind %s <Motion> \"%s%u -mouseMotion %%x %%y\"",
			  (char *) cameraFrame,
			  TclCamera3D::getObjectName(),getRef());
      getModule()->setResult("%s%u mouse motion bind to move camera or object",
			     TclCamera3D::getObjectName(),getRef());
    }
  else
    {
      getModule()->tickle("bind %s <Motion> \"\"",
			  (char *) cameraFrame);
      getModule()->setResult("%s%u mouse motion unbind",
			     TclCamera3D::getObjectName(),getRef());
    }

}





/*][************************************************************************/


/*  cam3D [-scene scn3D%u] [-in %w] [-stereo] */
int TclCamera3D::create(TclModule *m, int argc, char **argv) {
  TclString in;
  TclScene3D *scn = 0;
  TclCamera3D *togl;
  int i;
  bool stereo = false;

  if(argc > 6) return COMMAND_SYNOPSIS;

  for(i = 1; i < argc; i++) {
    if(!strcmp(argv[i], "-scene")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      scn = (TclScene3D *)
	TclObjectKind::getObject(typeid(TclScene3D), argc, argv, i);
      if(!scn) return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "-in")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(!m->windowExists(argv[i])) return COMMAND_SYNOPSIS;
      in = argv[i];
    }
    else if(!strcmp(argv[i], "-stereo")) {
      stereo = true;
    }
    else return COMMAND_SYNOPSIS;
  }
 if (!scn) 
	  scn= TclScene3D::getDefaultScene(m);
  try {
    togl = new TclCamera3D(m, in, scn, stereo);
  }
  catch (Exception &) {
    m->setResult("Unable to create camera");
    return TCL_ERROR;
  }
  
  m->setResult("%s%u", TclObjectKind::getObjectName(typeid(TclCamera3D)),
	       togl->getRef());

  return TCL_OK;
}

/*][************************************************************************/

int TclCamera3D::list(TclModule *m, int argc, char **argv) {
  TclScene3D *scn;
  TclString list, one;
  TclObjectKind *kind = TclObjectKind::findByKind(typeid(TclCamera3D));
    
  if(argc != 1 && argc != 2) return COMMAND_SYNOPSIS;

  // all cameras
  if(argc == 1) {
    ObjectsList<TclObject>::iterator it(kind->firstObject());
    for(; it!=kind->lastObject(); it++) {
      one.sprintf("%s%u ", TclObjectKind::getObjectName(typeid(TclCamera3D)),
		  ((TclCamera3D *) (*it))->getRef()); 
      list += one;
    }
    m->setResult(list);
  }

  // one scene cameras
  else {
    scn = 
      (TclScene3D *) TclObjectKind::getObject(typeid(TclScene3D), argc, argv);
    if(scn) {
      for(ObjectsList<Camera3D>::iterator it = scn->getClient()->firstCamera();
	  it!= scn->getClient()->lastCamera(); it++) {
	one.sprintf("%s%i ", TclObjectKind::getObjectName(typeid(TclCamera3D)), ((TclCamera3D *)(*it))->getRef());
	list += one;
      }
      m->setResult(list);
    }
  }

  return TCL_OK;
}

/*][************************************************************************/
int TclCamera3D::redrawCmd(int argc, char **argv) {
  if (argc!=0)  return COMMAND_SYNOPSIS;
  redraw();
  getModule()->setResult("%s%u redisplayed", getObjectName(), 
			 getRef());
  return TCL_OK;
}

/*][************************************************************************/
int TclCamera3D::selectCmd(int argc, char **argv) {
  TclObjectKind *kind;
  bool found = false;
 
  if(argc > 1) return COMMAND_SYNOPSIS;
  
  if(argc == 1) {
	  if(!strcmp(argv[0], "none")) {
		  applyObjectSelection(0);
		  getClient()->setMotionMode(Camera3D::MOVE_CAMERA);
		  setResult("no selected object");
		  return TCL_OK;
	  }
	  
/*	  std::list<ClippingPlane *>::iterator it;
	  else if(!strncmp(argv[0], "clip", 4)) {
		  if(sscanf(argv[0], "clip%u", &ref) != 1) return COMMAND_SYNOPSIS;
		  
		  for(it = getClient()->getFirstClippingPlane();
		  it != getClient()->getLastClippingPlane(); it++) {
			  cp = dynamic_cast<TclClippingPlane *>(*it);
			  if(cp && cp->getRef() == ref) {
				  applyObjectUnselection(getClient()->getSelectedObject());
				  setSelectedClippingPlane(cp);
				  setResult(argv[0]);
				  return TCL_OK;
			  }
		  }
		  setResult("no clipping plane %s", argv[0]);
		  return TCL_ERROR;
	  }
	  */
/*	  
	  else if(!strncmp(argv[0], "light", 5)) {
		  if(sscanf(argv[0], "light%u", &ref) != 1) return COMMAND_SYNOPSIS;
		  std::list<Light *>::iterator it;
		  
		  for(it = getClient()->lights.begin(); it != getClient()->lights.end(); it++) {
			  cl = dynamic_cast<TclLight *>(*it);
			  if(cl && cl->getRef() == ref) {
				  setSelectedLight(cl);
				  setResult(argv[0]);
				  return TCL_OK;
			  }
		  } 
		  setResult("no light %s", argv[0]);
		  return TCL_ERROR;
	  }
*/	  
	  else {
		  ObjectsList<Data3D>::iterator it = getScene()->getClient()->firstData();
		  for(;  it!=getScene()->getClient()->lastData() && !found; it++) {
			  Interface *interf = (*it)->getInterface();
			  if(interf &&
				  dynamic_cast<TclObject *>(interf)->name() == (const char *)argv[0]) {
				  applyObjectSelection(*it);
				  found = true;
			  }
		  }
		  
		  if(!found) {
			  kind = TclObjectKind::findByName(argv[0]);
			  if(!kind) {
				  setResult("object %s does not exists", argv[0]);
				  return TCL_ERROR;
			  }
			  setResult("unknown object kind %s", argv[0]);
			  return TCL_ERROR;
		  }
		  
		  else {
			  setResult(argv[0]);
			  return TCL_OK;
		  }
	  }
  }
  
  TclString output("none");
  if(getClient()->getSelectedObject()) {
	  Interface *interf = getClient()->getSelectedObject()->getInterface();
	  if(interf)
		  output = dynamic_cast<TclObject *>(interf)->name();
/*  } else if (selectedLight) {
	  output= ((TclLight *) (selectedLight))->name();
	  */
/*  } else if (selectedClippingPlane) {
	  output= ((TclClippingPlane *) (selectedClippingPlane))->name();
 */
   }
  getModule()->setResult(output);
  return TCL_OK;  
}



/* '+' key pressed: move camera forward */
int TclCamera3D::keyPlusCmd(int argc, char **argv) {
  if (argc>0)  return COMMAND_SYNOPSIS;

  Vec3<double> t(0.0, 0.0, getClient()->getRadius() / 2000.0);
  t *= getClient()->getRadius() / getClient()->getScale();
  t = getClient()->mapToScreen(t);
  getClient()->setCenter(getClient()->getCenter() - t);
  redraw();
  return TCL_OK;
}


/* '-' key pressed: move camera backward */
int TclCamera3D::keyMinusCmd(int argc, char **argv) {
  if (argc>0)  return COMMAND_SYNOPSIS;
  
  Vec3<double> t(0.0, 0.0, -getClient()->getRadius() / 2000.0);
  t *= getClient()->getRadius() / getClient()->getScale();
  t = getClient()->mapToScreen(t);
  getClient()->setCenter(getClient()->getCenter() - t);
  redraw();

  return TCL_OK;
}


/* up pressed: turn objects down */
int TclCamera3D::keyUpCmd(int argc, char **argv) {
  if (argc>0)  return COMMAND_SYNOPSIS;

  getClient()->setMoveMode(Camera3D::NAVIGATION_MODE);
  getClient()->rotate(Vec3<double>(1.0, 0.0, 0.0), -0.03);
  redraw();

  return TCL_OK;
}


/* down pressed: turn objects up */
int TclCamera3D::keyDownCmd(int argc, char **argv) {
  if (argc>0)  return COMMAND_SYNOPSIS;

  getClient()->setMoveMode(Camera3D::NAVIGATION_MODE);
  getClient()->rotate(Vec3<double>(1.0, 0.0, 0.0), 0.03);
  redraw();
  return TCL_OK;
}


/* left pressed: turn objects right */
int TclCamera3D::keyLeftCmd(int argc, char **argv) {
  if (argc>0)  return COMMAND_SYNOPSIS;

  getClient()->setMoveMode(Camera3D::NAVIGATION_MODE);
  getClient()->rotate(Vec3<double>(0.0, 1.0, 0.0), -0.03);
  redraw();
  return TCL_OK;
}


/* right pressed: turn objects left */
int TclCamera3D::keyRightCmd(int argc, char **argv) {
  if (argc>0)  return COMMAND_SYNOPSIS;

  getClient()->setMoveMode(Camera3D::NAVIGATION_MODE);
  getClient()->rotate(Vec3<double>(0.0, 1.0, 0.0), 0.03);
  redraw();
  return TCL_OK;
}
/* left pressed: turn objects right */
int TclCamera3D::keyCmd(int argc, char **argv) {
  if (argc!=1)  return COMMAND_SYNOPSIS;
  
  if (argv[0][0]=='c')
    getClient()->toggleCopyright();
  if (argv[0][0]=='f')
    getClient()->toggleFrameRate();

  redraw();
  return TCL_OK;
}



/*][************************************************************************/
/**********************************************************************/
/**************** rotate camera trackball-wise ************************/
/**********************************************************************/




/* Button1 pressed: begin model rotation */
int TclCamera3D::button1PressedCmd(int argc, char **argv) {
  if (argc!=2)  return COMMAND_SYNOPSIS;

  unsigned int x, y;

  if(sscanf(argv[0], "%u", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%u", &y) != 1) return COMMAND_SYNOPSIS;
  getClient()->buttonPressed(Camera3D::LEFT_BUTTON, Camera3D::NO_MODIFIER, x, y);
  bindMouseMotion(true); 
  return TCL_OK;
}
/* Button2 pressed: begin model rotation */
int TclCamera3D::button2PressedCmd(int argc, char **argv) {
  if (argc!=2)  return COMMAND_SYNOPSIS;

  unsigned int x, y;

  if(sscanf(argv[0], "%u", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%u", &y) != 1) return COMMAND_SYNOPSIS;
  getClient()->buttonPressed(Camera3D::MIDDLE_BUTTON, Camera3D::NO_MODIFIER, x, y);
  bindMouseMotion(true); 
  return TCL_OK;
}
/* Button2 pressed: begin model rotation */
int TclCamera3D::button3PressedCmd(int argc, char **argv) {
  if (argc!=2)  return COMMAND_SYNOPSIS;

  unsigned int x, y;

  if(sscanf(argv[0], "%u", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%u", &y) != 1) return COMMAND_SYNOPSIS;
  getClient()->buttonPressed(Camera3D::RIGHT_BUTTON, Camera3D::NO_MODIFIER, x, y);
  bindMouseMotion(true); 
  return TCL_OK;
}
/* Shift + Button1 pressed: begin model rotation */
int TclCamera3D::shiftButton1PressedCmd(int argc, char **argv) {
  if (argc!=2)  return COMMAND_SYNOPSIS;

  unsigned int x, y;

  if(sscanf(argv[0], "%u", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%u", &y) != 1) return COMMAND_SYNOPSIS;
  getClient()->buttonPressed(Camera3D::LEFT_BUTTON, Camera3D::SHIFT_KEY, x, y);
  bindMouseMotion(true); 
  return TCL_OK;
}
/* Shift + Button2 pressed: begin model rotation */
int TclCamera3D::shiftButton2PressedCmd(int argc, char **argv) {
  if (argc!=2)  return COMMAND_SYNOPSIS;

  unsigned int x, y;

  if(sscanf(argv[0], "%u", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%u", &y) != 1) return COMMAND_SYNOPSIS;
  getClient()->buttonPressed(Camera3D::MIDDLE_BUTTON, Camera3D::SHIFT_KEY, x, y);
  bindMouseMotion(true); 
  return TCL_OK;
}
/* Shift + Button3 pressed: begin model rotation */
int TclCamera3D::shiftButton3PressedCmd(int argc, char **argv) {
  if (argc!=2)  return COMMAND_SYNOPSIS;

  unsigned int x, y;

  if(sscanf(argv[0], "%u", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%u", &y) != 1) return COMMAND_SYNOPSIS;
  getClient()->buttonPressed(Camera3D::RIGHT_BUTTON, Camera3D::SHIFT_KEY, x, y);
  bindMouseMotion(true); 
  return TCL_OK;
}
/* Ctrl + Button1 pressed: begin model rotation */
int TclCamera3D::ctrlButton1PressedCmd(int argc, char **argv) {
  if (argc!=2)  return COMMAND_SYNOPSIS;

  unsigned int x, y;

  if(sscanf(argv[0], "%u", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%u", &y) != 1) return COMMAND_SYNOPSIS;
  getClient()->buttonPressed(Camera3D::LEFT_BUTTON, Camera3D::CTRL_KEY, x, y);
  bindMouseMotion(true); 
  return TCL_OK;
}
/* Ctrl + Button2 pressed: begin model rotation */
int TclCamera3D::ctrlButton2PressedCmd(int argc, char **argv) {
  if (argc!=2)  return COMMAND_SYNOPSIS;

  unsigned int x, y;

  if(sscanf(argv[0], "%u", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%u", &y) != 1) return COMMAND_SYNOPSIS;
  getClient()->buttonPressed(Camera3D::MIDDLE_BUTTON, Camera3D::CTRL_KEY, x, y);
  bindMouseMotion(true); 
  return TCL_OK;
}
/* Ctrl + Button3 pressed: begin model rotation */
int TclCamera3D::ctrlButton3PressedCmd(int argc, char **argv) {
  if (argc!=2)  return COMMAND_SYNOPSIS;

  unsigned int x, y;

  if(sscanf(argv[0], "%u", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%u", &y) != 1) return COMMAND_SYNOPSIS;
  getClient()->buttonPressed(Camera3D::RIGHT_BUTTON, Camera3D::CTRL_KEY, x, y);
  bindMouseMotion(true); 
  return TCL_OK;
}



int TclCamera3D::button1ReleasedCmd(int argc, char **argv) {
  if (argc!=2)  return COMMAND_SYNOPSIS;
  unsigned int x, y;


  if(sscanf(argv[0], "%u", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%u", &y) != 1) return COMMAND_SYNOPSIS;

  getClient()->buttonReleased(Camera3D::LEFT_BUTTON, x, y);
  bindMouseMotion(false);
  return TCL_OK;
}
int TclCamera3D::button2ReleasedCmd(int argc, char **argv) {
  if (argc!=2)  return COMMAND_SYNOPSIS;
  unsigned int x, y;

  if(sscanf(argv[0], "%u", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%u", &y) != 1) return COMMAND_SYNOPSIS;
  getClient()->buttonReleased(Camera3D::MIDDLE_BUTTON, x, y);
  bindMouseMotion(false);
  return TCL_OK;
}
int TclCamera3D::button3ReleasedCmd(int argc, char **argv) {
  if (argc!=2)  return COMMAND_SYNOPSIS;
  unsigned int x, y;

  if(sscanf(argv[0], "%u", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%u", &y) != 1) return COMMAND_SYNOPSIS;
  getClient()->buttonReleased(Camera3D::RIGHT_BUTTON, x, y);
  bindMouseMotion(false);
  return TCL_OK;
}


void TclCamera3D::dumpToFile(const TclString& filename) {
    dumpToPPM(filename);
/*#ifdef X11
	WindowDump(Tk_Display(TkWin), Tk_WindowId(TkWin), (char *) filename);
#endif*/
}

void TclCamera3D::dumpToPPM(const TclString& filename) {
    unsigned char *image;

    /* Allocate our buffer for the image */
    if ((image = new unsigned char[3*getClient()->getWidth()*getClient()->getHeight()])==0) {
        CERR << "Failed to allocate memory for image" << ENDL;
        throw (Exception("void TclCam2D::dumpToFile(const TclString& filename)",
               "no more memory"));
    }
    makeCurrent();
    glFlush();
    glPixelStorei(GL_PACK_ALIGNMENT,1);
    /* Copy the image into our buffer */
    glReadBuffer(GL_FRONT);
    glReadPixels(0,0,getClient()->getWidth(),getClient()->getHeight(),GL_RGB,GL_UNSIGNED_BYTE,image);

    FILE *fptr = fopen(filename,"wb");
    fprintf(fptr,"P6\n%d %d\n255\n",getClient()->getWidth(),
            getClient()->getHeight()); 
    int i,j;
    for (j=getClient()->getHeight()-1;j>=0;j--) {
        for (i=0;i<(int)getClient()->getWidth();i++) {
            fputc(image[3*j*getClient()->getWidth()+3*i+0],fptr);
            fputc(image[3*j*getClient()->getWidth()+3*i+1],fptr);
            fputc(image[3*j*getClient()->getWidth()+3*i+2],fptr);
        }
    }
    fclose(fptr);
}


int TclCamera3D::recordCmd(int argc, char **argv) {
  if(argc < 1) return COMMAND_SYNOPSIS;
  if(!strcmp(argv[0], "frame")) {
    if(argc != 2) return COMMAND_SYNOPSIS;
    getScene()->getSceneActivation()->stopMovie();
    startMovie(argv[1]);
    setResult("movie record started at frame rate");
  }
  else if(!strcmp(argv[0], "simulation")) {
    if(argc != 2) return COMMAND_SYNOPSIS;
    stopMovie();
    getScene()->getSceneActivation()->startMovie(getClient(), argv[1]);
    setResult("movie record started at simulation rate");
  }
  else if(!strcmp(argv[0], "pause")) {
    if(argc != 1) return COMMAND_SYNOPSIS;
    pauseMovie(record);
    if(record) setResult("movie record restarted");
    else setResult("movie record paused");
  }
  else if(!strcmp(argv[0], "stop")) {
    if(argc != 1) return COMMAND_SYNOPSIS;
    if(record) stopMovie();
    else getScene()->getSceneActivation()->stopMovie();
    setResult("end of movie record");
  }
  else return COMMAND_SYNOPSIS;
  return TCL_OK;
}


int TclCamera3D::mouselineCmd(int argc, char **argv) {
  if(argc != 2) return COMMAND_SYNOPSIS;
  unsigned int x, y;

  if(sscanf(argv[0], "%u", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%u", &y) != 1) return COMMAND_SYNOPSIS;
  Vec2<double> pixel(x, y);
  Vec3<double> eye, dir;
  getClient()->getEyeDirectionVector(pixel, eye, dir);
  setResult("%f %f %f %f %f %f", eye.getX(), eye.getY(), eye.getZ(),
	    dir.getX(), dir.getY(), dir.getZ());
  return TCL_OK;
}


void TclCamera3D::createCommands(TclModule *m) {
   TclObjectKind *kind;

   Tcl_CreateObjCommand(m->getInterp(), "createCamera3D", createCamera3D, NULL, NULL);
   Tcl_CreateObjCommand(m->getInterp(), "displayCamera3D", displayCamera3D, NULL, NULL);
   Tcl_CreateObjCommand(m->getInterp(), "resizeCamera3D", resizeCamera3D, NULL, NULL);
   Tcl_CreateObjCommand(m->getInterp(), "destroyCamera3D", destroyCamera3D, NULL, NULL);


  kind = new Scene3DItem(m, typeid(TclCamera3D), "cam3D","3D Camera",
			 &TclCamera3D::create, &TclCamera3D::list,
			 TclScene3D::NONE | TclScene3D::SCENE |
			 TclScene3D::OTHER,
			 "[-scene scn3D%u] [-in %w] [-stereo]");
  m->newCommand("cam3D%u", kind->getOptions(), "3D camera management command");
 
  kind->newOption("-destroy",(TclObject::OPTION) &TclCamera3D::destroyCmd,0,
		  "destroy a 3D camera");
  kind->newOption("-redraw",(TclObject::OPTION) &TclCamera3D::redrawCmd,0,
		  "redraw the scene seen by a 3D camera");
  kind->newOption("-transformationMatrix",(TclObject::OPTION) &TclCamera3D::transformationMatrixCmd, "[%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f]",
		  "get/set the projection matrix of the 3D camera");
  kind->newOption("-center",(TclObject::OPTION) &TclCamera3D::centerCmd, "[%f %f %f]",
		  "get/set the center of the 3D camera");
  kind->newOption("-refCenter",(TclObject::OPTION) &TclCamera3D::refCenterCmd, "[%f %f %f]",
		  "get/set the reference center of the 3D camera");
  kind->newOption("-scaleFactor",(TclObject::OPTION) &TclCamera3D::scaleFactorCmd, "[%f]",
		  "get/set the scale factor of the 3D camera");
  kind->newOption("-viewPoint",(TclObject::OPTION) &TclCamera3D::viewPointCmd, "[%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f]",
		  "get/set the center, reference center, scale factor and projection matrix of the 3D camera");
  kind->newOption("-focus",(TclObject::OPTION) &TclCamera3D::focusCmd,0,
		  "focus the camera on the 3D such that the scene appears centered in the 3D camera");
  kind->newOption("-focusOnRegion",(TclObject::OPTION) &TclCamera3D::focusOnRegionCmd,"[%f %f %f %f]", "focus the camera on a  3D sphere given its center (the first 3 parameters) and its radius (the last parameter)");
  kind->newOption("-fog", (TclObject::OPTION) &TclCamera3D::fogCmd, "[%b]",
		  "get/set fog in image");
  kind->newOption("-fogRange", (TclObject::OPTION) &TclCamera3D::fogRangeCmd, "[%f]",
		  "get/set fog range in image");
  kind->newOption("-stencilBuffer", (TclObject::OPTION) &TclCamera3D::stencilBufferCmd, "[%b]",
		  "get/use stencilBuffer on image");
  kind->newOption("-specularity", (TclObject::OPTION) &TclCamera3D::specularityCmd,
		  "[%b]", "get/set true specularity computation flag");
  kind->newOption("-frontAndBack",
		  (TclObject::OPTION) &TclCamera3D::renderFrontAndBackCmd,
		  "[%b]", "get/set front and back sides rendering flag");
  kind->newOption("-autofocus",(TclObject::OPTION) &TclCamera3D::autofocusCmd,"[%b]",
		  "get/set the autofocus mode of a camera");
  kind->newOption("-widget",(TclObject::OPTION) &TclCamera3D::widgetCmd,0,
		  "get the name of TclCamera3D widget where the scene is drawn");
  kind->newOption("-useTwoSides",(TclObject::OPTION) &TclCamera3D::useTwoSidesCmd,
		  "[%b]","get/set the boolean controlling if two sides or one side of each surface element");
  kind->newOption("-menu",(TclObject::OPTION) &TclCamera3D::menuCmd,0,
		  "get the name of menu linked to the camera where submenus can be inserted or deleted");
  kind->newOption("-objects",(TclObject::OPTION) &TclCamera3D::objectsCmd,0,
		  "returns the list of object names that are in the 3D scene seen by a 3D camera");
  kind->newOption("-select", (TclObject::OPTION) &TclCamera3D::selectCmd, "[objectName=%s]",
		  "get/set selected object in camera");
  kind->newOption("-clippingPlanes", (TclObject::OPTION) &TclCamera3D::clippingPlanesCmd, 0,
		  "returns lights list");
 kind->newOption("-clippingPlaneCreate", (TclObject::OPTION) &TclCamera3D::clippingPlaneCreateCmd, 0,
		  "add a new clipping plane");
  kind->newOption("-clippingPlaneRemove", (TclObject::OPTION) &TclCamera3D::clippingPlaneRemoveCmd, "%u",
		  "remove a clipping plane given its index");
   kind->newOption("-clippingPlaneEdit", (TclObject::OPTION) &TclCamera3D::clippingPlaneEditCmd, "%u active | changeActivation | swap",
		  "edit the characteristics of a clipping plane given its index");
  kind->newOption("-lightCreate", (TclObject::OPTION) &TclCamera3D::lightCreateCmd, 0,
		  "add a new light");
  kind->newOption("-lightRemove", (TclObject::OPTION) &TclCamera3D::lightRemoveCmd, "%u",
		  "remove a light given its rank");
   kind->newOption("-lights", (TclObject::OPTION) &TclCamera3D::lightsCmd, 0,
		  "returns the  number of lights");
  kind->newOption("-lightEdit", (TclObject::OPTION) &TclCamera3D::lightEditCmd, "changeActivation|active [%b]|show [%b]|edit %f %f %f %f %f %f %f %f %f %f %f %f",
		  "edit the characteristics of a light");
 kind->newOption("-boundingBoxColor",(TclObject::OPTION) &TclCamera3D::boundingBoxColorCmd,"[%f %f %f]",
		  "get/set the color of the bounding box around selected object");
  kind->newOption("-boundingBoxWidth",(TclObject::OPTION) &TclCamera3D::boundingBoxWidthCmd,"[%f]",
		  "get/set the width of the bounding box around selected object");
  kind->newOption("-boundingBoxDisplay",(TclObject::OPTION) &TclCamera3D::boundingBoxDisplayCmd,"[%b]",
		  "get/set the bounding box display mode");
  kind->newOption("-bgcolor",(TclObject::OPTION) &TclCamera3D::bgColorCmd,"[%f %f %f]",
		  "get/set the color of the background of a 3D camera");
  kind->newOption("-scene",(TclObject::OPTION) &TclCamera3D::sceneCmd,0,
		  "get the name of the scene observed by a 3D camera");
  kind->newOption("-width",(TclObject::OPTION) &TclCamera3D::widthCmd,"[%u]",
		  "get/set the width of  a 3D camera");
  kind->newOption("-height",(TclObject::OPTION) &TclCamera3D::heightCmd,"[%u]",
		  "get/set the height of  a 3D camera");
  kind->newOption("-fieldOfView",(TclObject::OPTION) &TclCamera3D::fieldOfViewCmd,"[%f ]"," get/ set the focal length of the camera (in degrees)");
  kind->newOption("-eyeOffset",(TclObject::OPTION) &TclCamera3D::eyeOffsetCmd,"[%f %f %f]",
		  "get/set the eye offset for a 3D stereo camera");
  
  /* motion mode */
  kind->newOption("-motionMode",(TclObject::OPTION) &TclCamera3D::motionModeCmd,"[object|camera]",
		  "get/set the motion mode either to move the camera or to move the selected object");
  kind->newOption("-mouseMotion",(TclObject::OPTION) &TclCamera3D::mouseMotionCmd,0,
		  "call the mouseMove() function of the camera to update the display according to the mouse motion");

  /* key bindings */
  kind->newOption("-keyPlus",(TclObject::OPTION) &TclCamera3D::keyPlusCmd,0,
		  "Key ""+"" pressed"); 
  kind->newOption("-keyMinus",(TclObject::OPTION) &TclCamera3D::keyMinusCmd,0,
		  "Key ""-"" pressed"); 
  kind->newOption("-keyUp",(TclObject::OPTION) &TclCamera3D::keyUpCmd,0,
		  "Key ""arrow up"" pressed"); 
  kind->newOption("-keyDown",(TclObject::OPTION) &TclCamera3D::keyDownCmd,0,
		  "Key ""arrow down"" pressed"); 
  kind->newOption("-keyRight",(TclObject::OPTION) &TclCamera3D::keyRightCmd,0,
		  "Key ""arrow right"" pressed"); 
  kind->newOption("-keyLeft",(TclObject::OPTION) &TclCamera3D::keyLeftCmd,0,
		  "Key ""arrow left"" pressed"); 
   kind->newOption("-key",(TclObject::OPTION) &TclCamera3D::keyCmd,0,
		  "Any other key  pressed"); 

  /* mouse button bindings */
  kind->newOption("-button1Released",(TclObject::OPTION) &TclCamera3D::button1ReleasedCmd,"%x %y",
		  "left mouse button released");
  kind->newOption("-button2Released",(TclObject::OPTION) &TclCamera3D::button2ReleasedCmd,"%x %y",
		  "middle mouse button released");
  kind->newOption("-button3Released",(TclObject::OPTION) &TclCamera3D::button3ReleasedCmd,"%x %y",
		  "right mouse button released");
  kind->newOption("-button1Pressed",(TclObject::OPTION) &TclCamera3D::button1PressedCmd,"%x %y",
		  "left mouse button pressed");
  kind->newOption("-button2Pressed",(TclObject::OPTION) &TclCamera3D::button2PressedCmd,"%x %y",
		  "middle mouse button pressed");
  kind->newOption("-button3Pressed",(TclObject::OPTION) &TclCamera3D::button3PressedCmd,"%x %y",
		  "right mouse button pressed");
  kind->newOption("-shiftButton1Pressed",(TclObject::OPTION) &TclCamera3D::shiftButton1PressedCmd,"%x %y",
		  "shift + left mouse button pressed");
  kind->newOption("-shiftButton2Pressed",(TclObject::OPTION) &TclCamera3D::shiftButton2PressedCmd,"%x %y",
		  "shift + middle mouse button pressed");
  kind->newOption("-shiftButton3Pressed",(TclObject::OPTION) &TclCamera3D::shiftButton3PressedCmd,"%x %y",
		  "shift + right mouse button pressed");
  kind->newOption("-ctrlButton1Pressed",(TclObject::OPTION) &TclCamera3D::ctrlButton1PressedCmd,"%x %y",
		  "ctrl + left mouse button pressed");
  kind->newOption("-ctrlButton2Pressed",(TclObject::OPTION) &TclCamera3D::ctrlButton2PressedCmd,"%x %y",
		  "ctrl + middle mouse button pressed");
  kind->newOption("-ctrlButton3Pressed",(TclObject::OPTION) &TclCamera3D::ctrlButton3PressedCmd,"%x %y",
		  "ctrl + right mouse button pressed");

  kind->newOption("-record", (TclObject::OPTION) &TclCamera3D::recordCmd,
		  "<<frame | simulation> file=%f | pause | stop>",
		  "start/stop camera movie dumps in file named file0000.ppm, file0001.ppm...");
  kind->newOption("-dumpToPPM", (TclObject::OPTION) &TclCamera3D::dumpToPPMCmd,
                  "<filename=%s>",
                  "Save the camera widget to a ppm file");
  kind->newOption("-mouseline", (TclObject::OPTION) &TclCamera3D::mouselineCmd,
		  "%x %y",
		  "Returns 3D line defined bye the camera eye point and the (x,y) coordinate in the camera plane. Six coordinates (a 3D point and a unit vector) are returned.");
}
}

