#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "graphics3D/TclClippingPlane.h"
#include <graphics3D/TclClippingPlane.h>
#include <graphics3D/TclCamera3D.h>


namespace yav {



TclClippingPlane::TclClippingPlane(TclModule *m, Camera3D *cam)
  : ClippingPlane(cam), TclObject(typeid(TclClippingPlane), m) {
  done(typeid(TclClippingPlane));
}


TclClippingPlane::TclClippingPlane(TclModule *m, Camera3D *cam,
				   const GLdouble *equation)
  : ClippingPlane(cam, equation), TclObject(typeid(TclClippingPlane), m) {
  done(typeid(TclClippingPlane));
}


TclClippingPlane::TclClippingPlane(TclModule *m, Camera3D *cam,
				   const Vec3<double>& P,
				   const Vec3<double>& n)
  : ClippingPlane(cam, P, n), TclObject(typeid(TclClippingPlane), m) {
  done(typeid(TclClippingPlane));
}

  
TclClippingPlane::~TclClippingPlane() {
  deleted(typeid(TclClippingPlane));
}


int TclClippingPlane::create(TclModule *m, int argc, char **argv) {
  TclClippingPlane *clip;

  if(argc != 2) return COMMAND_SYNOPSIS;
  Togl *cam = (Togl *) TclObjectKind::getObject(typeid(Togl), argc, argv, 1);
  if(!cam) return TCL_ERROR;
  clip = new TclClippingPlane(m,cam->getClient());
  cam->setSelectedClippingPlane(clip);
  m->setResult(clip->name());
  return TCL_OK;
}


int TclClippingPlane::destroyCmd(int argc, char **argv) {
  if(argc != 0) return COMMAND_SYNOPSIS;
  TclModule *m = getModule();
  unsigned int u = getRef();
  m->setResult("%s%u does not exist any longer",
	       TclClippingPlane::getObjectName(), u);
  Togl *cam=(Togl *)getCamera()->getInterface();
  if (cam->getSelectedClippingPlane()==this) {
    cam->unsetSelectedClippingPlane();
  }
 delete this;
  return TCL_OK;
}


int TclClippingPlane::activeCmd(int argc, char **argv) {
  bool val;
  
  if(argc > 1) return COMMAND_SYNOPSIS;
  if(argc > 0) {
    if(!TclModule::boolValue(argv[0], val)) return COMMAND_SYNOPSIS;
    if(val) enable();
    else disable();
  }

  setResult(TclModule::boolString(isActive()));
  return TCL_OK;
}


int TclClippingPlane::translateCmd(int argc, char **argv) {
  if(argc != 3) return COMMAND_SYNOPSIS;
  Vec3<double> t;
  if(sscanf(argv[0], "%lf", &t[0]) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%lf", &t[1]) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[2], "%lf", &t[2]) != 1) return COMMAND_SYNOPSIS;
  translate(t);
  setResult("%f %f %f", t.getX(), t.getY(), t.getZ());
  return TCL_OK;
}


int TclClippingPlane::swapCmd(int argc, char **argv) {
  int i;

  if(argc != 0) return COMMAND_SYNOPSIS;
  for(i = 0; i < 4; i++) equation[i] = -equation[i];
  cam->redraw();
  setResult("swaped");
  return TCL_OK;
}


} // namespace yav


