#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "graphics3D/TclLight.h"
#include <graphics3D/TclLight.h>
#include <graphics3D/Togl.h>

namespace yav {


TclLight::TclLight(TclModule *m, Camera3D *cam)
  : TclObject(typeid(TclLight), m), Light(cam->getFreeLight()) {
	camera=cam;
	done(typeid(TclLight));
}


TclLight::TclLight(TclModule *m, Camera3D *cam,
		   const GLfloat *ambient, const GLfloat *diffuse,
		   const GLfloat *specular, const GLfloat *position,
		   const GLfloat *direction,
		   GLfloat exponent, GLfloat cutoff,
		   const GLfloat *attenuation)
  : TclObject(typeid(TclLight), m),
    Light(cam->getFreeLight(), ambient, diffuse, specular, position, direction,
	  exponent, cutoff, attenuation) {
  camera=cam;
  done(typeid(TclLight));
}

TclLight::~TclLight() {
  deleted(typeid(TclLight));
}


int TclLight::create(TclModule *m, int argc, char **argv) {
  TclLight *light;

  if(argc != 2) return COMMAND_SYNOPSIS;
  Togl *cam =  (Togl *) TclObjectKind::getObject(typeid(Togl), argc, argv, 1);
  if(!cam) return TCL_ERROR;
  light = new TclLight(m,cam->getClient());
  m->setResult(light->name());
  return TCL_OK;
}


int TclLight::destroyCmd(int argc, char **argv) {
  if(argc != 0) return COMMAND_SYNOPSIS;
  TclModule *m = getModule();
  unsigned int u = getRef();
  Togl *cam=(Togl *)camera->getInterface();
  if (cam->getSelectedLight()==this) {
    cam->unsetSelectedLight();
  }
  char *n=name();
  delete this;
  m->setResult("%s does not exist any longer",n);
  return TCL_OK;
}


int TclLight::activeCmd(int argc, char **argv) {
  bool val;
  
  if(argc > 1) return COMMAND_SYNOPSIS;
  if(argc > 0) {
    if(!TclModule::boolValue(argv[0], val)) return COMMAND_SYNOPSIS;
    if(val) enable();
    else disable();
  }

  setResult(TclModule::boolString(isActive()));
  return TCL_OK;
}


int TclLight::paramsCmd(int argc, char **argv) {
  int i=0;

  if(argc > 0) {
    if(!strcmp(argv[0], "ambient") || !strcmp(argv[0], "diffuse")
       || !strcmp(argv[0], "specular") || !strcmp(argv[0], "position")
       || !strcmp(argv[0], "direction") || !strcmp(argv[0], "exponent")
       || !strcmp(argv[0], "cutoff") || !strcmp(argv[0], "attenuation")) {
      GLfloat coef[4];

      for(i = 0; i < argc; i++) {
	if(!strcmp(argv[i], "ambient")) {
	  if(i+3 >= argc) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[++i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[++i], "%f", &coef[1]) != 1) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[++i], "%f", &coef[2]) != 1) return COMMAND_SYNOPSIS;
	  if(i+1 >= argc || !strcmp(argv[i+1], "diffuse")
	     || !strcmp(argv[i+1], "specular") || !strcmp(argv[0], "position")
	     || !strcmp(argv[0], "direction") || !strcmp(argv[0], "exponent")
	     || !strcmp(argv[0], "cutoff") || !strcmp(argv[0], "attenuation"))
	    coef[3] = 1.0;
	  else {
	    if(sscanf(argv[++i], "%f", &coef[3]) != 1) return COMMAND_SYNOPSIS;
	  }
	  setAmbient(coef);
	}

	else if(!strcmp(argv[i], "diffuse")) {
	  if(i+3 >= argc) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[++i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[++i], "%f", &coef[1]) != 1) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[++i], "%f", &coef[2]) != 1) return COMMAND_SYNOPSIS;
	  if(i+1 >= argc || !strcmp(argv[i+1], "ambient")
	     || !strcmp(argv[i+1], "specular") || !strcmp(argv[0], "position")
	     || !strcmp(argv[0], "direction") || !strcmp(argv[0], "exponent")
	     || !strcmp(argv[0], "cutoff") || !strcmp(argv[0], "attenuation"))
	    coef[3] = 1.0;
	  else {
	    if(sscanf(argv[++i], "%f", &coef[3]) != 1) return COMMAND_SYNOPSIS;
	  }
	  setDiffuse(coef);
	}

	else if(!strcmp(argv[i], "specular")) {
	  if(i+3 >= argc) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[++i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[++i], "%f", &coef[1]) != 1) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[++i], "%f", &coef[2]) != 1) return COMMAND_SYNOPSIS;
	  if(i+1 >= argc || !strcmp(argv[i+1], "ambient")
	     || !strcmp(argv[i+1], "diffuse") || !strcmp(argv[0], "position")
	     || !strcmp(argv[0], "direction") || !strcmp(argv[0], "exponent")
	     || !strcmp(argv[0], "cutoff") || !strcmp(argv[0], "attenuation"))
	    coef[3] = 1.0;
	  else {
	    if(sscanf(argv[++i], "%f", &coef[3]) != 1) return COMMAND_SYNOPSIS;
	  }
	  setSpecular(coef);
	}

	else if(!strcmp(argv[i], "position")) {
	  if(i+4 >= argc) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[++i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[++i], "%f", &coef[1]) != 1) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[++i], "%f", &coef[2]) != 1) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[++i], "%f", &coef[3]) != 1) return COMMAND_SYNOPSIS;
	  setPosition(coef);
	}

	else if(!strcmp(argv[i], "direction")) {
	  if(i+3 >= argc) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[++i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[++i], "%f", &coef[1]) != 1) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[++i], "%f", &coef[2]) != 1) return COMMAND_SYNOPSIS;
	  setDirection(coef);
	}

	else if(!strcmp(argv[i], "exponent")) {
	  if(++i >= argc) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
	  setExponent(coef[0]);
	}

	else if(!strcmp(argv[i], "cutoff")) {
	  if(++i >= argc) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
	  setCutoff(coef[0]);
	}

	else if(!strcmp(argv[i], "attenuation")) {
	  if(i+3 >= argc) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[++i], "%f", &coef[0]) != 1) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[++i], "%f", &coef[1]) != 1) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[++i], "%f", &coef[2]) != 1) return COMMAND_SYNOPSIS;
	  setAttenuation(coef);
	}
      }
    }

    else {
      GLfloat ambient[4], diffuse[4], specular[4], shininess;

      if(i + 12 < argc) {
	if(sscanf(argv[0], "%f", &ambient[0]) != 1) return COMMAND_SYNOPSIS;
	if(sscanf(argv[1], "%f", &ambient[1]) != 1) return COMMAND_SYNOPSIS;
	if(sscanf(argv[2], "%f", &ambient[2]) != 1) return COMMAND_SYNOPSIS;
	if(sscanf(argv[3], "%f", &ambient[3]) != 1) return COMMAND_SYNOPSIS;
	if(sscanf(argv[4], "%f", &diffuse[0]) != 1) return COMMAND_SYNOPSIS;
	if(sscanf(argv[5], "%f", &diffuse[1]) != 1) return COMMAND_SYNOPSIS;
	if(sscanf(argv[6], "%f", &diffuse[2]) != 1) return COMMAND_SYNOPSIS;
	if(sscanf(argv[7], "%f", &diffuse[3]) != 1) return COMMAND_SYNOPSIS;
	if(sscanf(argv[8], "%f", &specular[0]) != 1) return COMMAND_SYNOPSIS;
	if(sscanf(argv[9], "%f", &specular[1]) != 1) return COMMAND_SYNOPSIS;
	if(sscanf(argv[10], "%f", &specular[2]) != 1) return COMMAND_SYNOPSIS;
	if(sscanf(argv[11], "%f", &specular[3]) != 1) return COMMAND_SYNOPSIS;
	if(sscanf(argv[12], "%f", &shininess) != 1) return COMMAND_SYNOPSIS;
	setAmbient(ambient);
	setDiffuse(diffuse);
	setSpecular(specular);
      }
      else if(i + 9 < argc) {
	if(sscanf(argv[0], "%f", &ambient[0]) != 1) return COMMAND_SYNOPSIS;
	if(sscanf(argv[1], "%f", &ambient[1]) != 1) return COMMAND_SYNOPSIS;
	if(sscanf(argv[2], "%f", &ambient[2]) != 1) return COMMAND_SYNOPSIS;
	ambient[3] = 1.0f;
	if(sscanf(argv[4], "%f", &diffuse[0]) != 1) return COMMAND_SYNOPSIS;
	if(sscanf(argv[5], "%f", &diffuse[1]) != 1) return COMMAND_SYNOPSIS;
	if(sscanf(argv[6], "%f", &diffuse[2]) != 1) return COMMAND_SYNOPSIS;
	diffuse[3] = 1.0f;
	if(sscanf(argv[8], "%f", &specular[0]) != 1) return COMMAND_SYNOPSIS;
	if(sscanf(argv[9], "%f", &specular[1]) != 1) return COMMAND_SYNOPSIS;
	if(sscanf(argv[10], "%f", &specular[2]) != 1) return COMMAND_SYNOPSIS;
	specular[3] = 1.0f;
	setAmbient(ambient);
	setDiffuse(diffuse);
	setSpecular(specular);
      }
      else return COMMAND_SYNOPSIS;
    }

    camera->redraw();
  }

  setResult("ambient %f %f %f %f diffuse %f %f %f %f specular %f %f %f %f position %f %f %f %f direction %f %f %f exponent %f cutoff %f attenuation %f %f %f",
	    ambient[0], ambient[1], ambient[2], ambient[3],
	    diffuse[0], diffuse[1], diffuse[2], diffuse[3],
	    specular[0], specular[1], specular[2], specular[3],
	    position[0], position[1], position[2], position[3],
	    direction[0], direction[1], direction[2],
	    exponent, cutoff,
	    attenuation[0], attenuation[1], attenuation[2]);
  return TCL_OK;
}


int TclLight::showCmd(int argc, char **argv) {
  bool val;
  
  if(argc != 2) return COMMAND_SYNOPSIS;
  if(argc == 2) {
    Togl *cam =  (Togl *) TclObjectKind::getObject(typeid(Togl), argc, argv, 0);
    if(!TclModule::boolValue(argv[1], val)) return COMMAND_SYNOPSIS;
    if(val) {
      if(!fantom && position[3] > 0.0) {
	TclLightFantom *interf = new TclLightFantom(this,cam->getClient());
	fantom = interf->getClient();
      }
    }
    else {
      if(fantom) delete fantom;
    }
  }

  if(fantom) setResult("fantom displayed");
  else setResult("no fantom");
  return TCL_OK;
}



} // namespace yav

