#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "graphics3D/TclScene3D.h"
#pragma implementation "graphics/TclSceneObject.h"
#pragma implementation "graphics/TclScene.h"
#include <basics/Constants.h>
#include <graphics/TclSceneObject.h>
#include <graphics3D/TclScene3D.h>


namespace yav {

typedef SceneItem <Scene3D> SceneItem3D;
AUTOLINK_STATIC_VARIABLES( SceneItem3D);

TclScene3D::TclScene3D(TclModule *m, Scene3D *scene)
  : TclScene<Scene3D, TclScene3D, Data3D,  Camera3D,TclCamera3D>(typeid(TclScene3D), scene ? scene : new Scene3D(), m) {
  getClient()->setInterface(this);
  masterContext=0;
#ifndef WIN32
  aDisplay=0;
#endif
  done(typeid(TclScene3D));
}

TclScene3D::~TclScene3D() {
#ifndef WIN32
  /* Destroy GL context associated to widget */
  if (masterContext){
    glXDestroyContext(aDisplay, masterContext);
    /* Set current context to none since it might still be
       set to this togl window */
    glXMakeCurrent(aDisplay, None, 0);
  }
#else 
  if (masterContext) {
      wglDeleteContext(masterContext);
  }
#endif
  deleted(typeid(TclScene3D));
}

TclScene3D *TclScene3D::getDefaultScene(TclModule *m) {
  TclObjectKind *kind = TclObjectKind::findByKind(typeid(TclScene3D));
  if(kind->getFirstObject()) return (TclScene3D *) kind->getFirstObject();
  else return new TclScene3D(m);
}


int TclScene3D::destroyCmd(int argc, char **argv) {
  unsigned int u = getRef();
  TclModule *m = getModule();
  if(argc > 0) return COMMAND_SYNOPSIS;
  TclString n=getObjectName();
  delete this;
  m->setResult("%s%u does not exist any longer",(char *)n,  u);
  return TCL_OK;
}

int TclScene3D::loadCmd(int argc, char **argv) {
  if(argc != 1) return COMMAND_SYNOPSIS;
  if(TclScene3D::read(argv[0])) {
    setResult("one object added in scene %s%u",
	      TclScene3D::getObjectName(), getRef());
    getClient()->redraw();
  }
  else
    setResult("load error");
  return TCL_OK;
}

int TclScene3D::iterationCmd(int argc, char **argv) {
  return TclScene<Scene3D,TclScene3D, Data3D, Camera3D, TclCamera3D>::iterationCmd(argc, argv);
}


int TclScene3D::computationCmd(int argc, char **argv) {
  return TclScene<Scene3D,TclScene3D, Data3D,  Camera3D, TclCamera3D>::computationCmd(argc, argv);
}


int TclScene3D::maxFrameRateCmd(int argc, char **argv) {
  return TclScene<Scene3D,TclScene3D, Data3D,  Camera3D, TclCamera3D>::maxFrameRateCmd(argc, argv);
}
int TclScene3D::redrawCmd(int argc, char **argv) {
  if(argc > 0) return COMMAND_SYNOPSIS;
  getClient()->redraw();
  setResult("%s%u redrawn", TclScene3D::getObjectName(), getRef());
  return TCL_OK;
}

int TclScene3D::focusCmd(int argc, char **argv) {
  if(argc > 0) return COMMAND_SYNOPSIS;
  getClient()->focus();
  setResult("focused on %s%u", TclScene3D::getObjectName(), getRef());
  return TCL_OK;
}


int TclScene3D::create(TclModule *m, int argc, char **argv) {
  TclScene3D *scn;

  if(argc != 1) return COMMAND_SYNOPSIS;

  scn = new TclScene3D(m);
  m->setResult("%s%u", TclObjectKind::getObjectName(typeid(TclScene3D)),
	       scn->getRef());
  return TCL_OK;
}

int TclScene3D::objectsCmd(int argc, char **argv) {
  return TclScene<Scene3D,TclScene3D, Data3D,  Camera3D, TclCamera3D>::objectsCmd(argc, argv);
}
int TclScene3D::deformableCmd(int argc, char **argv) {
  return TclScene<Scene3D,TclScene3D, Data3D,  Camera3D, TclCamera3D>::deformableCmd(argc, argv);
}


  /*  void TclScene3D::selectCameraObject(Data3D *d) const{
    CERR << " TclScene3D::selectCameraObject(Data3D *d) " << ENDL;
    }*/
int TclScene3D::cameraCmd(int argc, char **argv) {
  if(argc > 0) return COMMAND_SYNOPSIS;
  TclString output;
  Client  *c;
  ObjectsList<Camera3D>::iterator it=getClient()->firstCamera();
  for(; it!=getClient()->lastCamera(); it++) {
    c= dynamic_cast<Client *>(*it);
    if (c)
      {
	Interface *i = c->getInterface();
	if(i) {
	  output += dynamic_cast<TclObject *>(i)->name();
	  output += SPACE_STRING;
	}
      }
  }
  setResult(output);
  return TCL_OK;
}

void TclScene3D::createCommands(TclModule *m) {
  m->newCommand("scn3D", TclScene3D::whichCmd,
		"<-which [none] [scene] [camera] [file] [other] | -suffix %s | -realName %s>",
		"returns objects handled by 3D scenes (filtered by construction mode) or filename suffixes used for a given object type or the expanded name associated with a given object type");
  
  TclObjectKind *kind = new TclObjectKind(m, typeid(TclScene3D), "scn3D","3D Scene",
					  &TclScene3D::create,
					  &TclScene3D::list, 0);
  m->newCommand("scn3D%u", kind->getOptions(), "3D scene management command");

  kind->newOption("-destroy", (TclObject::OPTION) &TclScene3D::destroyCmd, 0,
		  "delete a 3D scene");
  kind->newOption("-maxFrameRate",(TclObject::OPTION) &TclScene3D::maxFrameRateCmd,"[%f]",
		  "get/set the maximum frame rate of display in Hz");
  kind->newOption("-objects", (TclObject::OPTION) &TclScene3D::objectsCmd,
		  0, "return scene objects list");
  kind->newOption("-deformable", (TclObject::OPTION) &TclScene3D::deformableCmd,
		  0, "return scene deformable models  list");
  kind->newOption("-camera", (TclObject::OPTION) &TclScene3D::cameraCmd,
		  0, "return the list of camera in a scene");
  kind->newOption("-load", (TclObject::OPTION) &TclScene3D::loadCmd, "%s",
		  "load 3D scene from given file name");
  kind->newOption("-iteration", (TclObject::OPTION) &TclScene3D::iterationCmd, "[step|call] [%u]",
		  "If no parameter is provided then all activated deformable models of the scene are deformed. Otherwise get or set the number of iterations per step or the number of iterations per background call");
  kind->newOption("-computation",
		  (TclObject::OPTION) &TclScene3D::computationCmd,
		  "[%b|%u|loop]","get/set the boolean corresponding to the activation status of the scene; if true it means that the scene is activating all deformable models. An integer value indicates the number of step to iterate before a complete stop.\"loop\" indicates that the scene will iterate until no deformable models are active");
  
  kind->newOption("-redraw", (TclObject::OPTION) &TclScene3D::redrawCmd, 0,
		  "redraw all cameras viewing this 3D scene");
  kind->newOption("-focus", (TclObject::OPTION) &TclScene3D::focusCmd, 0,
		  "reset all cameras viewing this 3D scene focus");

}
} // namespace yav

