#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#pragma implementation "graphics3D/TclScene3DObject.h"
#include <graphics3D/TclScene3DObject.h>
#include <graphics3D/TclScene3D.h>
#include <graphics3D/TclCamera3D.h>

namespace yav {

TclScene3DObject::~TclScene3DObject() {
 }
void TclScene3DObject::done(const std::type_info& id,
			    const char *install, const char *uninstall,
			    Camera3D::Selection3DFunctionType fct) {
  TclObject::done(id);
  if (fct)
    selectionFunction=fct;
  setInstallMenuCommandName(install);
  setUninstallMenuCommandName(uninstall);
  TclScene3D *tclscene=  (TclScene3D *)(((Data3D *) client)->getScene()->getInterface());
  tclscene->selectCameraObject((Data3D *) client);
  ((Data3D *) client)->updateCenterAndRadius();
  tclscene->redraw();
}

void TclScene3DObject::deleted(const std::type_info& id) {
  if(client) {
    getTclScene()->unselectCameraObject((Data3D *) client);
  }
  TclObject::deleted(id);
}

/// add a selection to the map of selection modes
void TclScene3DObject::addSelectionMode(SelectionMode *m) {
  selectionModeMap[(char *)m->name]=m;
}
/// remove a selection given its name
void TclScene3DObject::removeSelectionMode(char *name) {
  std::map<char *,SelectionMode *,CStringCompare>::iterator it=selectionModeMap.find(name); 
  if (it!=selectionModeMap.end()) {
    delete  (*it).second;
    selectionModeMap.erase(it);
  }
}
/// get a selection given from its  name
const TclScene3DObject::SelectionMode *TclScene3DObject::getSelectionMode(char *name) const {
  std::map<char *,SelectionMode *,CStringCompare>::const_iterator it=selectionModeMap.find(name); 
  if (it!=selectionModeMap.end()) {
    return (*it).second;
  }
  else 
    return  0;
}
void TclScene3DObject::selectionCallback(TclCamera3D *camera){
  /// eventually install menu
  if(getInstallMenuCommandName() != EMPTY_STRING) {
    getModule()->tickle("%s %s %s %s #90C090",
			(char *)getInstallMenuCommandName(),
			(char *) camera->getSubMenuFrame(),
			(char *) camera->name(),
			(char *)(name()));
  }
  if (selectionFunction) {
    camera->getClient()->setSelectionFunction(selectionFunction);    
  }
}
void TclScene3DObject::unselectionCallback(TclCamera3D *camera){
  /// eventually install menu
  if(getUninstallMenuCommandName() != EMPTY_STRING) {
    getModule()->tickle("%s %s %s %s",
			(char *)getUninstallMenuCommandName(),
			(char *) camera->getSubMenuFrame(),
			(char *) camera->name(),
			(char *)(name()));
  }
  /// remove selection function
  camera->getClient()->setSelectionFunction(0);    
  
}
int TclScene3DObject::selectionModeCmd( int argc, char **argv) 
{
  if (argc==0)
    {  
      if (selectedMode) 
	setResult(selectedMode->name);
      else {
	setResult("none");
      }
      return TCL_OK;
    }
  else if (argc==1) {
    if (!strcmp(argv[0],"list")) {
      std::map<char *,SelectionMode *,CStringCompare>::iterator it=selectionModeMap.begin();
      TclString out;
      for(;it!=selectionModeMap.end();++it) {
	TclString res;
	res.sprintf("{ {%s} %s } ",(char *)(*it).second->label,(char *)(*it).second->name);
	out+=res;
      }
      setResult(out);
      return TCL_OK;
    } else
      return COMMAND_SYNOPSIS;
  }
  else if (argc==2)
    {
      TclCamera3D *cam= (TclCamera3D *) TclObjectKind::getObject(typeid(TclCamera3D), argc,argv,1);
      if (cam!=NULL)
	{
	  const  TclScene3DObject::SelectionMode *newMode=getSelectionMode(argv[0]);
	  if (newMode) {
	    if (newMode!=selectedMode) {
	      /// eventually removes the submenu installed by the previous mode
	      if (selectedMode->TclUnInstallProc!=EMPTY_STRING)
		getModule()->tickle("%s %s %s",(char *)selectedMode->TclUnInstallProc,cam->getSubMenuFrame(),
				    (char *)name());
	      /// decides the global selection mode of the camera
	      if (newMode->name=="Object")
		cam->getClient()->setSelectionMode(Camera3D::SELECT_OBJECT);
	      else 
		cam->getClient()->setSelectionMode(Camera3D::OBJECT_DEFINED);
	      /// sets the new selection Function
	      cam->getClient()->setSelectionFunction(newMode->selectionFunction);
	      /// eventually calls an additional function to perform additional tasks
	      if (newMode->installProc)
		(*newMode->installProc)(this,cam);
	      /// eventually sets the new submenu 
	      if (newMode->TclInstallProc!=EMPTY_STRING)
		getModule()->tickle("%s %s %s %s",(char *)newMode->TclInstallProc,cam->getSubMenuFrame(),
				    (char *)cam->name(),(char *)name());
	      /// replace the pointer
	      setSelectedMode(newMode);
	    }
	  }
	  else 
	    {
	      setResult("unknown selection mode %s",argv[0]);
	      return TCL_ERROR;
	    }
	}
      else {
	setResult("Unknown camera name %s",argv[1]);
	return TCL_ERROR;
      }

      setResult(selectedMode->name);
      return TCL_OK;
    }
  else
    return COMMAND_SYNOPSIS;
}
int TclScene3DObject::sceneCmd(int argc, char **argv) 
{
  if (argc==0)
    {
      TclScene3D *tclscene=  (TclScene3D *)(((Data3D *) client)->getScene()->getInterface());
      setResult("%s%u", TclObjectKind::getObjectName(typeid(TclScene3D)),
		tclscene->getRef());
      return TCL_OK;
    }
    else
    return COMMAND_SYNOPSIS;
}

int TclScene3DObject::translateCmd( int argc, char **argv) 
{
 double tx,ty,tz;
 if (argc!=3) return COMMAND_SYNOPSIS;
 if((sscanf(argv[0],"%lf",&tx)!=1)||(sscanf(argv[1],"%lf",&ty)!=1)||(sscanf(argv[2],"%lf",&tz)!=1)) return COMMAND_SYNOPSIS;

  Vec3<double> t(tx,ty,tz);
  getClient()->translate(t);
  getClient()->redrawScene();
  return TCL_OK;
}

int TclScene3DObject::scaleCmd( int argc, char **argv) 
{
  if (argc!=1) 
    return COMMAND_SYNOPSIS;

  getClient()->scale((double) (atof(argv[0])));
 getClient()->redrawScene();
  return TCL_OK;
}

int TclScene3DObject::rotateCmd( int argc, char **argv) 
{
  if (argc!=3) 
    return COMMAND_SYNOPSIS;
  else
    {
      RotVect t((double) atof(argv[0]),(double) atof(argv[1]),
		     (double) atof(argv[2]));
      Rotation3D m(t);
     getClient()->rotate(m);
     getClient()->redrawScene();
      return TCL_OK;
    }
}

void TclScene3DObject::createOptionCommands( TclObjectKind *kind) {
  kind->newOption("-translate",(TclObject::OPTION) &TclScene3DObject::translateCmd,"%f %f %f",
		  "translate the 3D object");
  kind->newOption("-scale",(TclObject::OPTION) &TclScene3DObject::scaleCmd,"%f",
		  "scale a 3D object around its center");
  kind->newOption("-rotate",(TclObject::OPTION) &TclScene3DObject::rotateCmd,"%f %f %f",
		  "rotate a 3D object around its center given a rotation vector");
  kind->newOption("-scene",(TclObject::OPTION) &TclScene3DObject::sceneCmd,0,
		  "return the 3D scene where is the 3D object");
  kind->newOption("-selectionMode",(TclObject::OPTION)  &TclScene3DObject::selectionModeCmd,
			  "[ list | key=%s cam3D%u]",
			  "with no argument this function returns the keyword associated with the selection mode. With the keyword \"list\", it returns a list of keywords and labels corresponding to all selection modes available. With a keyword and camera, it selects a selection mode");


}

} // namespace yav


