#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "graphics3D/TclTexture1D.h"
#include <graphics3D/TclTexture1D.h>


namespace yav {


TclTexture1D::TclTexture1D(TclModule *m,unsigned int *b, 
			   const unsigned int nc,
			   Format f) : Texture1D(b,nc,f), 
  TclObject(typeid(TclTexture1D),m) {
  done(typeid(TclTexture1D));
}
TclTexture1D::TclTexture1D(TclModule *m,char *name) : Texture1D(name), 
  TclObject(typeid(TclTexture1D),m) {
  done(typeid(TclTexture1D));
}
TclTexture1D::TclTexture1D(TclModule *m,Texture1D::GenericTexture type,
			   const unsigned int size): Texture1D(type,size), 
  TclObject(typeid(TclTexture1D),m) {
  done(typeid(TclTexture1D));
}
TclTexture1D::~TclTexture1D() {
  deleted(typeid(TclTexture1D));
}

int TclTexture1D::destroyCmd(int argc, char **argv) {
  if(argc != 0) return COMMAND_SYNOPSIS;
  TclModule *m = getModule();
  unsigned int u = getRef();
  m->setResult("%s%u does not exist any longer",
	       TclTexture1D::getObjectName(), u);
  delete this;
  return TCL_OK;
}
int TclTexture1D::colorCmd(int argc, char **argv) {
	if (argc==0) {
	  setResult("%u",getNbRealColumns());
	  return TCL_OK;
  }
  if  (argc != 1) return COMMAND_SYNOPSIS;
  unsigned int index=(unsigned int) atoi(argv[0]);
  if (index>=getNbRealColumns()) {
	  setResult("%s must be less than %u",argv[0],getNbRealColumns());
	  return TCL_ERROR;
  }
  unsigned char color[4];
  getColor(index,color);
  TclString str;
  str.sprintf("#%02x%02x%02x",(unsigned int)color[0],
	  (unsigned int)color[1],(unsigned int)color[2]);
  setResult(str);
  return TCL_OK;
}

int TclTexture1D::create(TclModule *m, int argc, char **argv) {
  TclTexture1D *t;
  if(argc == 2) {
    t= new TclTexture1D(m,argv[1]);
  }
  else if ((argc == 3)){
    GenericTexture type;
  
    if (!strcmp(argv[1],"temperature"))
      type=TEMPERATURE;
    else if (!strcmp(argv[1],"temperature_reverse"))
      type=TEMPERATURE_REVERSE;
    else if (!strcmp(argv[1],"blue2yellow"))
      type=BLUE2YELLOW;
    else if (!strcmp(argv[1],"random"))
      type=RANDOM;
    else if (!strcmp(argv[1],"jet"))
      type=JET;
    else if (!strcmp(argv[1],"jet_reverse"))
      type=JET_REVERSE;
    else if (!strcmp(argv[1],"heart"))
      type=HEART;
    else if (!strcmp(argv[1],"heart_reverse"))
      type=HEART_REVERSE;
    else if (!strcmp(argv[1],"activation"))
      type=ACTIVATION;
    else if (!strcmp(argv[1],"brb"))
      type=BRB;
    else
      {
	m->setResult("%s must be either temperature, temperature_reverse, random, jet, jet_reverse or activation",argv[1]);
	return TCL_ERROR;
      }
      t= new TclTexture1D(m,type,(unsigned int) atoi(argv[2]));
  }
  else
    return COMMAND_SYNOPSIS;
  m->setResult(t->name());
  return TCL_OK;
}

void TclTexture1D::createCommands(TclModule *m) 
{
	 TclObjectKind *kind;

  kind = new TclObjectKind(m, typeid(TclTexture1D), "text1D","1D Texture. It is a 1d array of colors",0,
			   &TclTexture1D::create, &TclTexture1D::list);
  kind->newOption("-destroy", (TclObject::OPTION) &TclTexture1D::destroyCmd, 0,
		  "destroy a 1D texture");
  kind->newOption("-color", (TclObject::OPTION) &TclTexture1D::colorCmd, 
	  " [colorIndex=%u ]",
		  "with no parameters returns the number of color in the 1D texture. When a color index is provided it returns the hexadecimal value of the RGB component of that colorin the 1D texture");


}

}


