#ifndef BOXCOLLISION_H
#define BOXCOLLISION_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 4355)
#endif
#pragma interface "graphics3D/BoxCollision.h"
#ifdef WIN32
#include <windows.h>
#endif
#include <list>
#include <linearAlgebra/Vec3.h>
#include <basics/Constants.h>

namespace yav {

/** \class BoxCollision
    \ingroup Graphics3D
    a virtual class that is used to detect collision between a box and an
    geometrical object defined in some way */
template <class T> class BoxCollision {
 public:
  /* The data computed through the collision detection */
  std::list<T> collisionList;
 public:
  /** @name constructor */
  //@{
  /** builds a BoxCollision object   */
  BoxCollision(){}
  //@}

  /** @name destructor */
  //@{
  virtual ~BoxCollision(){}
  //@}
  /** detect collision with a 3D box
      @param p1 the first point of box's axis
      @param p2 the second point of box's axis
      @param r1 the square width 
      @param p3 the up direction (optional = 1,0,0)
      @param r2 the square lenght in p3 direction (optional = r1)
      @return a boolean that is true if there were a collision
      and build the collision list */
  virtual bool detectCollisionWithBox(Vec3<double> p1,Vec3<double> p2, double r1, Vec3<double> p3 = Vec3<double>(1,0,0),double r2=-1){return false;}
  /** detect collision with a 3D triangle
      @param org the top of triangle
      @param dest the middle of triangle's base
      @param up the width direction
      @param base the triangle's base length
      @param width the triangle's width
      @return a boolean that is true if there were a collision
      and buid the collision list*/
  virtual bool detectCollisionWithTriangle(Vec3<double> org,Vec3<double> dest, Vec3<double> up, double base,double width){return false;}
  /** inform that the geometry has changed */
  virtual void geometryHasChanged()=0;
  /// return the first item of the collision list
  typename std::list<T>::iterator getFirstCollision(){return collisionList.begin();}
  /// return the last item of the collision list
  typename std::list<T>::iterator getLastCollision() {return collisionList.end();}

};

}

#endif // BOXCOLLISION_H
