#ifndef CAVEINTERFACE3D_H
#define CAVEINTERFACE3D_H

#pragma interface "graphics3D/CAVEInterface3D.h"

#include <linearAlgebra/Vec3.h>
#include <basics/Pointed.h>
#include <graphics3D/UserInterface3D.h>

namespace yav {
/** \class CAVEInterface3D
    \ingroup Graphics3D
    3D CAVE Stick Interface that correspond to a specific user interface.
    It provides a member fonction to get the fixed point of the device
*/
class CAVEInterface3D : public UserInterface3D, public Pointed <CAVEInterface3D> {
 public:
  /** @name constructor */
  //@{
  /** builds a new CAVE Interface
      @param scene  scene where the CAVE interface object should be added (if scene is 0, the first existing scene is used or a scene is created== 0)
      @param l the tool length
      @param r the tool radius
 */
  CAVEInterface3D(Scene3D *scene,double l=1.0, double r=-1);
  //@}
  
  /** @name destructor */
  //@{
  /// empty
  virtual ~CAVEInterface3D(){}
  //@}

  /// returns the length
  virtual double getLength() const
    { return m_length; }
  /// returns the radius 
  virtual double getRadius() const
    { return m_radius; }
  /// sets the tool length
  virtual void setLength(const double &l)
    { m_length = l; }
  /// sets the tool radius
  virtual void setRadius(const double &r)
    { m_radius = r; }

  /// returns the position of the origin of the tool
  virtual Vec3<double> getPosition()=0; // why UserInterface3D does not use const on this one ?
  /// returns the position of end of the tool
  virtual Vec3<double> getEndPosition() const =0;
  /// returns the orientation of the tool in the form of EulerAnglesYXZ
  virtual Vec3<double> getOrientation() const =0;
  /// returns the ZAxis of the tool (along the tool)
  virtual Vec3<double> getZAxis() const =0;

  friend class Scene3D;

protected:
  /// length of the Stick3D
  double m_length;
  /// radius of the Stick3D
  double m_radius;
};

typedef BasicUserInterface3DIterator<CAVEInterface3D> CAVEInterface3DIterator;
} // namespace yav


#endif // CAVEINTERFACE3D_H
