#ifndef DATA3D_H
#define DATA3D_H

#pragma interface "graphics3D/Data3D.h"
#include <basics/AutoLink.h>
#include <basics/ListElement.h>
#include <linearAlgebra/Vec3.h>
#include <linearAlgebra/Rotation3D.h>
#include <graphics/Data.h>
#include <graphics/UserInterface.h>
#include <graphics/FileScene.h>

namespace yav {

class Scene3D;
#ifdef GRAPHICS
class Camera3D;
#endif

/** \class Data3D
    \ingroup Graphics3D
    \brief Data available in a 3D scene */
class Data3D : 
#ifdef GRAPHICS
 public GraphicsData<Vec3<double>, Rotation3D, Scene3D, Camera3D,
  Data3D>
#else
public Data<Scene3D, Data3D>
#endif
 {
 public:
  /** @name constructor */
  //@{
  /** builds a new Data3D
      @param scene scene to create data in (one is created if scene == 0)
      @param exclusive wether this data is exclusive in scene
      @param redrawOrder
  */
  Data3D(Scene3D *scene = 0, bool exclusive = false
#ifdef GRAPHICS
       , unsigned int redrawOrder = 1
#endif
	 );
  //@}
  
  /** @name destructor */
  //@{
  /// empty
  virtual ~Data3D();
  //@}

#ifdef GRAPHICS
 protected:
  /** redraw order (objects with redraw order 1 are rendered before
      objects with redraw order 2) */
  unsigned int redrawOrder;

 public:
  /// 3D mouse click information
  struct MouseClick3D : public DataTypes::MouseClick {
    MouseClick3D(const Vec3<double> &e,
		 const Vec3<double> &d,
		 double t) : eye(e), direction(d), threshold(t) {}
    Vec3<double> eye;
    Vec3<double> direction;
    double threshold;
  };

  /// retrieve object redraw order (1 or 2)
  unsigned int getRedrawOrder(void) const { return redrawOrder; }
  /** set new object redraw order
      @param ro order, should be 1 or 2 */
  void setRedrawOrder(unsigned int ro) {
    if(ro > 0 && ro < 3) redrawOrder = ro;
  }
#endif

  /// update object center and radius and re-focus
  void updateCenterAndRadius(void);

};





} // namespace yav


#include <graphics3D/UserInterface3D.h>
#include <graphics3D/Scene3D.h>

#endif
