#ifndef LIGHTFANTOM_H
#define LIGHTFANTOM_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics3D/LightFantom.h"
#include <graphics3D/MaterialSphere.h>

namespace yav {

class Light;


/** \class LightFantom
    \ingroup Graphics3D
    \brief spot light graphics representation */
class LightFantom : public MaterialSphere {
 protected:
  /// represented light
  Light *light;
  /// where to draw the Fantom
  Camera3D *camera;
 public:
  /** @name constructor */
  //@{
  /** builds a new sphere
      @param light represented light
  */
  LightFantom(Light *light,Camera3D *cam);
  //@}

  /** @name destructor */
  //@{
  /// empty
  virtual ~LightFantom();
  //@}

  /// wireframe display list creation callback
  virtual void buildWireframe(void);

  /// wireframe display list creation callback
  virtual void buildSolid(void);

  /** translates the object
      @param tr translation vector */
  virtual void translate(const Vec3<double>& tr);
  /** rotates the object
      @param rotation rotation matrix */
  virtual void rotate(const Rotation3D& rotation);  

  /// light moved: reset fantom location
  void move(void);

  /// light color changed: reset fantom color
  void setColor(void);
};


} // namespace yav

#include <graphics3D/Light.h>



#endif  // ifndef LIGHTFANTOM_H
