#ifndef MATERIALSPHERE_H
#define MATERIALSPHERE_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics3D/MaterialSphere.h"
#ifdef WIN32
#include <windows.h>
#endif
#include <GL/gl.h>
#include <GL/glu.h>
#include <graphics3D/Sphere.h>
#include <graphics3D/Data3D.h>
#include <graphics3D/Scene3D.h>
#include <graphics3D/Surface.h>

namespace yav {

/** \class MaterialSphere
    \ingroup Graphics3D
    \brief Sphere surface */
class MaterialSphere : public Sphere,
		       public Surface,
		       public Data3D {
 protected:
  /// rendered object
  GLUquadricObj *sphere;

  /// computes bounding box
  void initBoundingBox(void);

 public:
  /** @name constructors */
  //@{
  /** default sphere centered at 0 with 1.0 radius
      @param scene scene to build sphere in */
  MaterialSphere(Scene3D *scene = 0);
  /** builds a new sphere
      @param c new sphere center
      @param r new sphere radius
      @param scene scene to build sphere in */
  MaterialSphere(const Vec3<double>& c, double r, Scene3D *scene = 0);
  //@}

  /** @name destructor */
  //@{
  /// empty
  virtual ~MaterialSphere();
  //@}

  /** set new sphere radius
      @param r new radius (should be > 0) */
  virtual void setRadius(double r);
  /** set new sphere center
      @param c new center */
  virtual void setCenter(const Vec3<double>& c);

  /// raise getRadius() ambigous method call
  double getRadius(void) const { return Sphere::getRadius(); }

  /** translates the object
      @param tr translation vector */
  virtual void translate(const Vec3<double>& tr);
  /** rotates the object
      @param rotation rotation matrix */
  virtual void rotate(const Rotation3D& rotation) {}
  /** scales the object
      @param scale scale facto */
  virtual void scale(const double scale);

  /// all graphics objects can be drawn
  virtual void redraw(Camera3D *);

  /** Returns the graphic object center and weight.
      The weight is a measure of the number of primitives to allow
      barycentric computation of the whole scene center.
      @param center returned center 
      @param n returned barycentric weight */
  virtual void updateCenter(Vec3<double>& center, unsigned int& n);

  /// Returns the graphic object radius given its center
  virtual double updateRadius(const Vec3<double>& center);
  
  /** The routine that allows the camera to select an object
      by computing the distance of an object from its radial line
      @param mc 3D line descriptor
      @return distance to line */
  virtual double computeClosestDistance(DataTypes::MouseClick& mc) const;

  /// wireframe display list creation callback
  virtual void buildWireframe(void);

  /// wireframe display list creation callback
  virtual void buildSolid(void);

  /** set object redraw order for transparency sorting
      @param order redraw order (should be 1 or 2) */
  virtual void setRedrawOrder(unsigned int order) {
    Data3D::setRedrawOrder(order);
  }

  friend class TclMaterialSphere;

 private:
  /// common constructors code
  void init(void);
};


} // namespace yav


#endif // ifndef MATERIALSPHERE_H
