
namespace yav {


template <class OBJ, class DATA> OGLBoxCollision<OBJ,DATA>::OGLBoxCollision(OBJ* object,void (*updateG)(OBJ*),void (*buildCL)(OGLBoxCollision<OBJ,DATA>*,OBJ*)){
  useCollisionDisplayList=true;
  collisionDisplayList=0;
  objectCollided=object;
  updateGeometry=updateG;
  buildCollisionList=buildCL;
}

template <class OBJ, class DATA> OGLBoxCollision<OBJ,DATA>::~OGLBoxCollision(){
  destroyCollisionDisplayList();
  typename std::list<CollisionData*>::iterator it=collisionDataList.begin();
  for(;it!=collisionDataList.end();it++)
    delete (*it);
  collisionDataList.clear();
}

template <class OBJ, class DATA> bool OGLBoxCollision<OBJ,DATA>::detectCollisionWithBox(Vec3<double> org,Vec3<double> dest, 
					  double radius, Vec3<double> up, double upradius)
{
  if (upradius==-1) upradius=radius;

  GLint hits;
#ifndef WIN32  
  glSelectBuffer (bufferSize, selectionBuffer);
#else
glSelectBuffer (600, selectionBuffer);
#endif
  (void) glRenderMode (GL_SELECT);
  
  glMatrixMode (GL_PROJECTION);
  glPushMatrix ();
  /// defines the camera position
  glLoadIdentity ();
  double zz = (dest-org).norm();
  glOrtho(-radius,radius, -upradius,upradius, 0, zz);
  glMatrixMode (GL_MODELVIEW);
  glPushMatrix ();
  glLoadIdentity (); // This call seems to cause an error somewhere.... jcl
  gluLookAt(org[0], org[1], org[2], dest[0], dest[1], dest[2], up[0] , up[1], up[2]);
  /// draw object
  callCollisionDisplayList();
  /// pop all matrices
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
  /// get the hits 
  hits = glRenderMode(GL_RENDER);
  if (hits <= 0) 
    return false;
  buildCollisionDataList(hits);
  return true;

}


template <class OBJ, class DATA> bool OGLBoxCollision<OBJ,DATA>::detectCollisionWithTriangle(Vec3<double> org,Vec3<double> dest, Vec3<double> up, double base, double width){

  GLint hits;
#ifndef WIN32  
  glSelectBuffer (bufferSize, selectionBuffer);
#else
glSelectBuffer (600, selectionBuffer);
#endif
  (void) glRenderMode (GL_SELECT);
  
  glMatrixMode (GL_PROJECTION);
  glPushMatrix ();
  /// defines the camera position
  glLoadIdentity ();
  double zz = (dest-org).norm();
  double alpha = 2 * acos(width/(2*zz)*180/M_PI);
  glMatrixMode (GL_MODELVIEW);
  glPushMatrix ();
  glLoadIdentity ();
  gluPerspective(alpha,base/width,0.001,zz);

  gluLookAt(org[0],org[1],org[2],dest[0],dest[1],dest[2],up[0],up[1],up[2]);


  callCollisionDisplayList();

  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();

  hits = glRenderMode(GL_RENDER);
  if (hits <= 0) 
    return false;
  buildCollisionDataList(hits);
  return true;
}

template <class OBJ, class DATA> void OGLBoxCollision<OBJ,DATA>::destroyCollisionDisplayList(void){
  if(collisionDisplayList!=0){
    glDeleteLists(collisionDisplayList,1);
    collisionDisplayList=0;
  }
}

template <class OBJ, class DATA> void OGLBoxCollision<OBJ,DATA>::geometryHasChanged(){
  this->collisionList.clear();
  destroyCollisionDisplayList();
}

template <class OBJ, class DATA> void OGLBoxCollision<OBJ,DATA>::callCollisionDisplayList(void){
  if(!useCollisionDisplayList){
    updateGeometry(objectCollided);
  }else{
    if(collisionDisplayList == 0){
      collisionDisplayList = glGenLists(1);
      glNewList(collisionDisplayList,GL_COMPILE_AND_EXECUTE);
      //updateCollisionList();
      updateGeometry(objectCollided);
      glEndList();
    }
    else glCallList(collisionDisplayList);
  }
}

template <class OBJ, class DATA> void OGLBoxCollision<OBJ,DATA>::buildCollisionDataList(int hits){
  typename std::list<CollisionData*>::iterator it=collisionDataList.begin();
  for(;it!=collisionDataList.end();it++)
    delete (*it);
  collisionDataList.clear();
  this->collisionList.clear();
  GLuint *ptr = (GLuint *) selectionBuffer;
  GLuint names;
  for( int i=0;i<hits;++i){
    /*  for each hit  */
    names = *ptr; ptr++;
    CollisionData *cD = new CollisionData(names);
    cD->namesDepth = names;
    
    cD->zMin = *ptr++; // zMin
    cD->zMax = *ptr++; // zMax
    if (names<255) {
      for (unsigned int j = 0; j < names; j++)
	  cD->names[j]=*ptr++; // skipping names
    }
    collisionDataList.push_back(cD);
  }
  buildCollisionList(this,objectCollided);
}
#ifdef TCL_INTERFACE
template <class OBJ, class DATA> int OGLBoxCollision<OBJ,DATA>::useCollisionDisplayListCmd(TclModule *m, int argc, char **argv){
  bool usedl;
  if(argc > 1) return COMMAND_SYNOPSIS;
  if(argc == 1) {
    if(!TclModule::boolValue(argv[0], usedl)) return COMMAND_SYNOPSIS;
    if(useCollisionDisplayList&&!usedl)
      destroyCollisionDisplayList();
    useCollisionDisplayList=usedl;
  }
  m->setResult(TclModule::boolString(useCollisionDisplayList));
  return TCL_OK;
}
#endif

}
