#ifndef PICKINGOBJECT_H
#define PICKINGOBJECT_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics3D/PickingObject.h"
#include <graphics3D/Camera3D.h>
#include <basics/Constants.h>
#include <GL/glu.h>

namespace yav {

 /// the size of the pickingBuffer
#define _pickingBufferSize 512

/** \class PickingObject
    \ingroup Graphics3D
    class that implements a basic routine for picking one item from 
    a given OpenGL display list */
class PickingObject {
 protected:
 
 
  /// the selection buffer used to pick items
  GLuint pickingBuffer[_pickingBufferSize];
 protected:
    /** returns the reference of a selected item after OpenGL picking
      @param displayList an OpenGL display list using the name stack
      @param pixel the position in screen coordinate where the mouse button was pressed
      @param cam the camera where the selection is performed
      @param areaSize the half size in pixels of the area around the pixel where any item is searched 
      @return -1 if no item has been selected or returns the reference of the selected item */
  int doPicking(GLuint displayList, const Vec2<double> &pixel, 
		Camera3D *cam, double areaSize);
 public:
  /** @name constructors */
  //@{
  /** build a picking Object */
  PickingObject(){}
  //@}

  /** @name destructor */
  //@{
  /// delete a PickingObject
  virtual ~PickingObject(){}
  //@}
};


}

#endif
