#ifndef SCENE3D_H
#define SCENE3D_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics3D/Scene3D.h"
#include <basics/AutoLink.h>
#include <basics/ListElement.h>
#include <linearAlgebra/Vec3.h>
#include <linearAlgebra/Rotation3D.h>
#include <framework/Activable.h>
#ifdef GRAPHICS
#include <graphics3D/Camera3D.h>
#else
#include <graphics/Scene.h>
#endif
#include <graphics3D/Data3D.h>
#include <graphics/FileScene.h>


namespace yav {


/** \class Scene3D
    \ingroup Graphics3D
    \brief a 3D scene including Graphics, Deformable and RangeData Objects */
 class Scene3D : 
#ifdef GRAPHICS
public GraphicsScene<Vec3<double>, Data3D,  UserInterface3D, Camera3D, Scene3D>,
#else
public Scene<Data3D, UserInterface3D, Scene3D>,
#endif
   public AutoLink<Scene3D> 
{
   
 public:
  /** @name constructor */
  //@{
  /// initializes scene with nothing inside
  Scene3D() : 
#ifdef GRAPHICS
GraphicsScene<Vec3<double>, Data3D,  UserInterface3D, Camera3D, Scene3D>() 
#else
Scene<Data3D,  UserInterface3D, Scene3D>() 
#endif
{}
  //@}

  /** @name destructor */
  //@{
  /// destroy each object in list
  virtual ~Scene3D();
  //@}
  /// redraw all scene cameras
  virtual void redraw(void);
  /** read a file into a scene 
      @param filename the name of the file to be read
      @return  returns true if it the file could have been read and false otherwise */
  virtual bool read(const TclString &filename) 
    { return(FileScene<Scene3D>::readScene(filename,this)); }
  /// returns the index of the scene in the list of scenes
  virtual unsigned int getRef(void) const {
   return AutoLink<Scene3D>::getRef();
  }

};


} // namespace yav

#ifdef GRAPHICS
#include <graphics3D/Camera3D.h>
#include <graphics3D/UserInterface3D.txx>
#endif

#endif
