#ifndef SPHERE_H
#define SPHERE_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#pragma interface "graphics3D/Sphere.h"
#include <linearAlgebra/Vec3.h>
#include <linearAlgebra/Rotation3D.h>

namespace yav {

/** \class Sphere
    \ingroup Graphics3D
    \brief a geometric sphere */
class Sphere {
 protected:
  /// sphere center
  Vec3<double> center;
  /// sphere radius
  double radius;

 public:
  /** @name constructors */
  //@{
  /// default sphere centered at 0 with 1.0 radius
  Sphere() : center(0.0, 0.0, 0.0), radius(1.0) {}
  /** builds a new sphere
      @param c new sphere center
      @param r new sphere radius */
  Sphere(const Vec3<double>& c, double r)
    : center(c), radius(r <= 0.0 ? 1.0 : r) {}
  //@}

  /** @name destructor */
  //@{
  /// empty
  virtual ~Sphere() {}
  //@}

  /// returns sphere radius
  double getRadius(void) const { return radius; }
  /** set new sphere radius
      @param r new radius (should be > 0) */
  virtual void setRadius(double r) { radius = r <= 0.0 ? 1.0 : r; }
  /// returns sphere center
  Vec3<double> getCenter(void) const { return center; }
  /** set new sphere center
      @param c new center */
  virtual void setCenter(const Vec3<double>& c) { center = c; }

  /** translate sphere
      @param t translation vector */
  virtual void translate(const Vec3<double>& t) { center += t; }

  /** rotates sphere around P
      @param r rotation matrix
      @param P rotation center point */
  virtual void rotate(const Rotation3D& r, const Vec3<double>& P) {
    center = r * (center - P) + P;
  }
};


} // namespace yav


#endif // ifndef SPHERE_H
