#ifndef TCLCLIPPINGPLANE_H
#define TCLCLIPPINGPLANE_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics3D/TclClippingPlane.h"
#include <modules/TclObject.h>
#include <modules/TclModule.h>
#include <graphics3D/ClippingPlane.h>

namespace yav {

class Camera3D;

/** \class TclClippingPlane
    \ingroup Graphics3D
    \brief TCL interface for a clipping plane */
class TclClippingPlane : public ClippingPlane,
			 public TclObject {
 public:
  /** @name constructors */
  //@{
  /** construct a new clipping plane passing through scene center and
      orthogonal to (0, 0, 1)
      @param m module to create TCL object
      @param cam camera to apply clipping plane in */
  TclClippingPlane(TclModule *m, Camera3D *cam);
  /** construct a new clipping plane with plane equation
      equation[0] x + equation[1] y + equation[2] z + equation[3] = 0
      @param m module to create TCL object
      @param cam camera to apply clipping plane in
      @param equation four plane equation coefficients */
  TclClippingPlane(TclModule *m, Camera3D *cam, const GLdouble *equation);
  /** construct a new clipping plane with passing through P and
      orthogonal to n
      @param m module to create TCL object
      @param cam camera to apply clipping plane in
      @param P one clipping plane point
      @param n clipping plane normal */
  TclClippingPlane(TclModule *m, Camera3D *cam, const Vec3<double>& P,
		   const Vec3<double>& n);
  //@}
  
  /** @name destructor */
  //@{
  ~TclClippingPlane();
  //@}

  /// -destroy option
  int destroyCmd(int argc, char **argv);
  /// -active option
  int activeCmd(int argc, char **argv);
  /// -translate option
  int translateCmd(int argc, char **argv);
  /// -swap option
  int swapCmd(int argc, char **argv);

  /// new TCL command callback
  static int create(TclModule *m, int argc, char **argv);
};


} // namespace yav


#endif // ifdef TCLCLIPPINGPLANE_H


