#ifndef TCLLIGHT_H
#define TCLLIGHT_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics3D/TclLight.h"
#include <modules/TclObject.h>
#include <graphics3D/Light.h>
#include <graphics3D/TclLightFantom.h>


namespace yav {


/** \class TclLight
    \ingroup Graphics3D
    \brief TCL interface for a light */
class TclLight : public TclObject,
		 public Light {
protected:
	Camera3D *camera;
 public:
  /** @name constructors */
  //@{
  /** construct a new directional light
      @param m module to create TCL object
      @param cam camera to apply light  in */
  TclLight(TclModule *m, Camera3D *cam);
  /** construct a new directional light
      @param m module to create TCL object
      @param cam camera to apply light  in
      @param ambient ambient light 4 coefficients
      @param diffuse diffuse light 4 coefficients
      @param specular specular light 4 coefficients
      @param position light position 4 coefficients
      @param direction spotlight direction 3 coefficients
      @param exponent spotlight exponent
      @param cutoff spotlight cutoff (in degrees)
      @param attenuation spotlight 3 attenuation factors */
  TclLight(TclModule *m, Camera3D *cam,
	   const GLfloat *ambient, const GLfloat *diffuse,
	   const GLfloat *specular, const GLfloat *position,
	   const GLfloat *direction = 0,
	   GLfloat exponent = 0.0, GLfloat cutoff = 180.0,
	   const GLfloat *attenuation = 0);
  //@}
  
  /** @name destructor */
  //@{
  virtual ~TclLight();
  //@}

  /// -destroy option
  int destroyCmd(int argc, char **argv);
  /// -active option
  int activeCmd(int argc, char **argv);
  /// -params TCL command
  int paramsCmd(int argc, char **argv);
  /// -show TCL command
  int showCmd(int argc, char **argv);

  /// new TCL command callback
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }
  /// new TCL command callback
  static int create(TclModule *m, int argc, char **argv);
};


} // namespace yav


#endif // ifndef TCLLIGHT_H
