#ifndef TCLMATERIALSPHERE_H
#define TCLMATERIALSPHERE_H


#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics3D/TclMaterialSphere.h"
#include <graphics3D/TclScene3DObject.h>
#include <graphics3D/TclScene3D.h>
#include <graphics3D/MaterialSphere.h>


namespace yav {


/** \class TclMaterialSphere
    \ingroup Graphics3D
    \brief TCL implementation for a spherical surface */
class TclMaterialSphere : public TclScene3DObject {
 protected:
  /** @name constructor */
  //@{
  /** builds a TCL interface for a child
      @param id dynamic type
      @param ms client
      @param m TCL module to build object in */
  TclMaterialSphere(const std::type_info& id, MaterialSphere *ms,
		    TclModule *m);
  //@}

 public:
  /** @name constructor */
  //@{
  /** builds a TCL interface for a sphere
      @param m TCL module to build object in
      @param scene scene to build sphere in */
  TclMaterialSphere(TclModule *m, TclScene3D *scene);
  //@}

  /** @name destructor */
  //@{
  /** destroy a TclMaterialSphere */
  ~TclMaterialSphere();
  //@}

  /// returns client
  MaterialSphere *getClient(void) { return (MaterialSphere *) client; }
  
  /// -destroy TCL command
  int destroyCmd(int argc, char **argv);

  /// -radius TCL command
  int radiusCmd(int argc, char **argv);

  /// -color TCL command
  int colorCmd(int argc, char **argv) {
    return getClient()->color.colorCmd(getModule(),
				       (Data3D *) client, argc, argv);
  }

  /// -material TCL command
  int materialCmd(int argc, char **argv) {
    return getClient()->mat.materialCmd(getModule(),
					(Data3D *) client, argc, argv);
  }

  /// -render TCL command
  int renderCmd(int argc, char **argv) {
    return getClient()->renderCmd(getModule(), (Data3D *) client, argc, argv);
  }
  /// -renderOption TCL command
  int renderOptionCmd(int argc, char **argv) {
    return getClient()->renderOptionCmd(getModule(), (Data3D *) client, argc, argv);
  }
  /// -texture2D TCL command
  int texture2DCmd(int argc, char **argv) {
    return getClient()->texture2DCmd(getModule(), (Data3D *) client, argc, argv);
  }
  /// -texture1D TCL command
  int texture1DCmd(int argc, char **argv) {
    return getClient()->texture1DCmd(getModule(), (Data3D *) client, argc, argv);
  }
  /// -highLightTexture TCL command
  int highlightTextureCmd(int argc, char **argv) {
    return getClient()->highlightTextureCmd(getModule(), (Data3D *) client, argc, argv);
  }
  /// -texture2DOption TCL command
  int texture2DOptionCmd(int argc, char **argv) {
    return getClient()->texture2DOptionCmd(getModule(), (Data3D *) client, argc, argv);
  }
 /// -texture1DOption TCL command
  int texture1DOptionCmd(int argc, char **argv) {
    return getClient()->texture1DOptionCmd(getModule(), (Data3D *) client, argc, argv);
  }
  /// create function
  static int create(TclModule *m, int argc, char **argv);

  /// list function
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }

};


} // namespace yav


#endif // ifndef TCLMATERIALSPHERE_H
