#ifndef TCLSCENE3D_H
#define TCLSCENE3D_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#pragma interface "graphics3D/TclScene3D.h"
#include <graphics/TclScene.h>
#include <graphics3D/Scene3D.h>
#include <modules/TclModule.h>
#include <modules/TclObject.h>
#include <graphics3D/TclCamera3D.h>

#ifdef GRAPHICS 
#include <GL/gl.h>
#ifndef WIN32
#include <GL/glx.h>
#endif
#endif

namespace yav {


/** \class TclScene3D
    \ingroup Graphics3D
    \brief a Tcl object representing a 3D scene */
 class TclScene3D : public TclScene<Scene3D, TclScene3D, Data3D, Camera3D,TclCamera3D> {
 public:
  /** @name constructor */
  //@{
  /** builds a TclScene3D object given a module
      @param m pointer on a module
      @param scene
  */
  TclScene3D(TclModule *m, Scene3D *scene = 0);
  //@}

  /** @name destructor */
  //@{
  /// empty 
  virtual ~TclScene3D();
  //@}
 protected:
#ifndef WIN32
  /** The GL Context that share display lists with contexts with all contexts in all cameras */ 
  GLXContext masterContext;
  /** a display that is necessary to destroy context */
  Display *aDisplay;
#else
  /** The GL Context that share display lists with contexts with all contexts in all cameras */ 
  HGLRC masterContext;
#endif
  
 public:
#ifndef WIN32
  GLXContext getMasterGLContext() const {
    return masterContext;
  }
#else
  HGLRC getMasterGLContext() const {
    return masterContext;
  }
#endif
  /** read a file into a scene 
      @param filename the name of the file to be read
      @return  returns true if it the file could have been read and false otherwise */
  bool read(const TclString &filename) {
    return FileScene<Scene3D>::readScene(filename,getClient());
  }
  /// returns the index of the scene in the list of scenes
  unsigned int getRef(void) const { return TclObject::getRef(); }
  /** returns the default  TclScene3D : if there are several existing TclScene3D then returns the first TclScene3D; otherwise creates a new  TclScene3D
      @param m module used to eventually create a new TclScene3D
      @return the default TclScene3D */
  static TclScene3D *getDefaultScene(TclModule *m);
  /// -destroy TCL command option
  int destroyCmd(int argc, char **argv);
  /// -load TCL command option
  int loadCmd(int argc, char **argv);
  /// -iteration TCL command option
  int iterationCmd(int argc, char **argv);
  /// -computation TCL command option
  int computationCmd(int argc, char **argv);
  /// -objects TCL command option
  int objectsCmd(int argc, char **argv);
  /// -deformable TCL command option
  int deformableCmd(int argc, char **);
  /// -camera TCL command option
  int cameraCmd(int argc, char **);
  

#ifdef GRAPHICS
  /// -redraw TCL command option
  int redrawCmd(int argc, char **argv);
  /// -focus TCL command option
  int focusCmd(int argc, char **argv);
  /// -maxFrameRate TCL command option
  int maxFrameRateCmd(int argc, char **argv);
#endif

  /// returns this interface's client
  Scene3D *getClient(void) const { return (Scene3D *) client; }

 /** creates a new TclScene3D
      @param m module to create the scene3D  in
      @param argc command line number of arguments 
      @param argv command line arguments */
  static int create(TclModule *m, int argc, char **argv);
  /// create the tcl objectkind for this object 
  static void createCommands(TclModule *m);

};

/// TclObjectKind for 3D scenes short-hand
typedef SceneItem< Scene3D > Scene3DItem;


} // namespace yav



#endif
