#ifndef TCLSCENE3DOBJECT_H
#define TCLSCENE3DOBJECT_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics3D/TclScene3DObject.h"
#include <graphics/TclSceneObject.h>
#include <graphics3D/Data3D.h>
#include <graphics3D/TclScene3D.h>


namespace yav {

class TclCamera3D;

/** \class TclScene3DObject
    \ingroup Graphics3D
    \brief TCL Object offering a TCL interface to a client */
class TclScene3DObject : public TclSceneObject {
 public:
  /// a prototype of a function that is called when a selection mode is selected
  typedef void (*SELECTION_INSTALL_PROC)(TclScene3DObject *,TclCamera3D *);

  /** a selection mode defines what is performed when the left mouse is clicked
      and when the camera selection mode is itself \"user defined\" */
  class SelectionMode {
  public:
    /** this routine is called when this mode is selected */
    SELECTION_INSTALL_PROC installProc;
    /** the selection function that will be provided to the camera */
    Camera3D::Selection3DFunctionType selectionFunction;
    /** the name of the mode : one word that is used for the radio buttons */
    TclString name;
    /** the label that will appear in the radio buttons */
    TclString label;
    /** the Tcl procedure called when the mode is selected */
    TclString TclInstallProc;
    /** the Tcl procedure called when the mode stops being selected */
    TclString TclUnInstallProc;
    SelectionMode() {
    selectionFunction=0;
    TclInstallProc=0;
    TclUnInstallProc=0;
    installProc=0;
    }
  };
 protected:
  /// the selected Mode
  const SelectionMode *selectedMode;
  /// the list of available selection Modes
  std::map<char *,SelectionMode *,CStringCompare> selectionModeMap;
  /** a single selectionFunction that can be used instead of the list of 
      selection modes */
  Camera3D::Selection3DFunctionType selectionFunction;  

 public:
  /** @name constructor */
  //@{
  /** only constructor from a client object
      @param type object type
      @param client non null pointer on client object
      @param m TCL module to build object in
      @exception NullClient null pointer transmitted */
  TclScene3DObject(const std::type_info& type, Data3D *client,
		   TclModule *m = 0)
    throw(Interface::NullClient)
    : TclSceneObject(type, client, m) {
    selectedMode=0;
    selectionFunction=0;
  }
  //@}

  /** @name destructor */
  //@{
  /// destroy client and this
    virtual ~TclScene3DObject();
  //@}

  /** method that should be called at the end of any
      TclObject constructor to call all WarnedObjects callback
      @param id built object type
      @param install install menu TCL command name
      @param uninstall uninstall menu TCL command name
      @param fct
  */
  void done(const std::type_info& id, const char *install,
	    const char *uninstall,Camera3D::Selection3DFunctionType fct=0);

  /** method that should be called at the end of any
      TclObject destructor to call all WarnedObjects callback
      @param id built object type */
  void deleted(const std::type_info& id);  
  /// a function called whenever the object is no longer selected in a given camera
  virtual void unselectionCallback(TclCamera3D *camera);
  /// a function called whenever the object is  selected in a given camera
  virtual void selectionCallback(TclCamera3D *camera);
  /// returns client
  Data3D *getClient(void) const { return (Data3D *) client; }
    /// return the associated Tcl Scene
  TclScene3D *getTclScene() const {
    return (TclScene3D *)getClient()->getScene()->getInterface();
  }
  /// add a selection to the map of selection modes
  virtual void addSelectionMode(SelectionMode *m);
  /// remove a selection given its name
  virtual void removeSelectionMode(char *name);
  /// get a selection given from its  name
  virtual const SelectionMode *getSelectionMode(char *name) const;
  /// set the selected selection mode
  void setSelectedMode(const SelectionMode *m) {
    selectedMode=m;
  }
  
  /// -selectionMode TCL command
  int selectionModeCmd(int argc, char **argv);
  /// -scene TCL command
  int sceneCmd(int argc, char **argv);
  /// -translate TCL command
  int translateCmd(int argc, char **argv);
  /// -rotate TCL command
  int rotateCmd(int argc, char **argv);
  /// -scale TCL command
  int scaleCmd(int argc, char **argv);
  /// add the option commands to the object kind
  static void createOptionCommands(TclObjectKind *);

};


} // namespace yav


#endif // ifndef TCLSCENE3DOBJECT_H
