#ifndef TCLTEXTURE2D_H
#define TCLTEXTURE2D_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics3D/TclTexture2D.h"
#include <graphics3D/Texture2D.h>
#include <modules/TclObject.h>



namespace yav {


/** \class TclTexture2D
    \ingroup Graphics3D
    \brief a Tcl interface of a 2D texture based on Opengl standard */
class TclTexture2D : public Texture2D, public TclObject {
 public:
  /** @name constructors */
  //@{
  /** building a 2D texture from aRGBA or ARGB  buffer 
      @param  m module to create TCL object
      @param b a buffer where each pixel is each coded with RGBA or ABGR
      @param nr the number of rows (may not be a power of 2) 
      @param nc the number of columns (may not be a power of 2) 
      @param f the format of each pixel */ 
  TclTexture2D(TclModule *m,unsigned int *b, 
	       const unsigned int nr, const unsigned int nc,
	       Format f=TEXTURE_RGBA);
  /** build a 2D texture interface from a file
      @param  m module to create TCL object
      @param filename file name of a GIF file */
  TclTexture2D(TclModule *m,char *filename);
  /** build a 2D texture interface of a given type
      @param m module to create TCL object
      @param type the type of generic texture
      @param size the number of rows and columns of the texture 
      @param exponent for the gaussian and halo texture controls the spread of the gaussian
      @param radius for the halo texture controls the radius of the halo (as default it is a quarter of the size )
  */
  TclTexture2D(TclModule *m,Texture2D::GenericTexture type,
	       const unsigned int size, const double exponent=1.0, 
	       const double radius=1.0);
  //@}
  
  /** @name destructor */
  //@{
  /// delete a 2D Texture 
  virtual ~TclTexture2D();
  //@}
  
  /// -destroy option
  int destroyCmd(int argc, char **argv);
  /// new TCL command callback
  static int create(TclModule *m, int argc, char **argv);
  /// create Tcl commands
  static void createCommands(TclModule *m);

};


}


#endif
