
#include "vtkActor.h"
#include "vtkConeSource.h"
#include "vtkGlyph3D.h"
#include "vtkPolyData.h"
#include "vtkPolyDataMapper.h"
#include "vtkSphereSource.h"
#include "vtkRenderWindow.h"
#include "mipsVtkView.h"

#include <gtk/gtk.h>
#include "gtk/gtkgl.h"

#include <assert.h>

mips::VtkView * view;
bool view_attached;


static void
build_vtk_ui (mips::VtkView * p_view)
{
  vtkActor *sphereActor1, *spikeActor1;
  vtkSphereSource *sphere;
  vtkConeSource *cone;
  vtkGlyph3D *glyph;
  vtkPolyDataMapper *sphereMapper, *spikeMapper;
 
 

  sphere = vtkSphereSource::New();
  sphereMapper = vtkPolyDataMapper::New();
  sphereMapper->SetInput(sphere->GetOutput());
  sphereActor1 = vtkActor::New();
  sphereActor1->SetMapper(sphereMapper);
  cone = vtkConeSource::New();
  glyph = vtkGlyph3D::New();
  glyph->SetInput(sphere->GetOutput());
  glyph->SetSource(cone->GetOutput());
  glyph->SetVectorModeToUseNormal();
  glyph->SetScaleModeToScaleByVector();
  glyph->SetScaleFactor(0.25);
  spikeMapper = vtkPolyDataMapper::New();
  spikeMapper->SetInput(glyph->GetOutput());
  spikeActor1 = vtkActor::New();
  spikeActor1->SetMapper(spikeMapper);
  
  p_view->AddActor(sphereActor1);
  p_view->AddActor(spikeActor1);
}

static void
detach_attach_cb (GtkButton *btn, gpointer data)
{
  assert (data != 0);
  if (view_attached) {    
    view_attached = false;  
    gtk_container_remove(GTK_CONTAINER(data),view->GetGtkWidget());
    gtk_button_set_label (GTK_BUTTON(btn),"Attach");
  }
  else {
        view_attached = true;
        gtk_box_pack_start (GTK_BOX (data),view->GetGtkWidget() , TRUE, TRUE, 0);
        gtk_button_set_label (GTK_BUTTON(btn),"Detach");
  }
}


static void
offscreen_snapshot_cb (GtkButton *btn, gpointer data)
{
  assert (data != 0);

  GtkWidget* image = gtk_image_new_from_pixbuf(view->CreatePixbuf (300,300));
   
   /* Create the widgets */
   GtkWidget* dialog = gtk_dialog_new_with_buttons ("Offscreen rendered snapshot",
                                         GTK_WINDOW(data),
                                         GTK_DIALOG_DESTROY_WITH_PARENT,
                                         GTK_STOCK_OK,
                                         GTK_RESPONSE_NONE,
                                         NULL);
   gtk_window_set_position(GTK_WINDOW(dialog),GTK_WIN_POS_NONE);
   
   /* Ensure that the dialog box is destroyed when the user responds. */
   g_signal_connect_swapped (GTK_OBJECT (dialog), 
                             "response", 
                             G_CALLBACK (gtk_widget_destroy),
                             GTK_OBJECT (dialog));
   /* Add the label, and show everything we've added to the dialog. */
   gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox),
                      image);
   gtk_widget_show_all (dialog);
}

static void
clicked_cb (GtkButton *btn, gpointer data)
{
  assert (data == 0);
  /* trigger end of main loop */
  gtk_main_quit ();
}

static gboolean
delete_cb (GtkWidget *widget, GdkEvent *event, gpointer data)
{
  assert (data == 0);
  /* trigger end of main loop */
  gtk_main_quit ();
  return TRUE;
}

static GtkWidget *
build_gtk_ui (GtkWidget *area)
{
  GtkWidget *win, *button,  *vbox;
  win = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  g_signal_connect (G_OBJECT (win), "delete-event",
		    G_CALLBACK (delete_cb), 0);
  vbox = gtk_vbox_new (FALSE, 5);
  gtk_container_add (GTK_CONTAINER (win), vbox);

  gtk_widget_set_size_request (GTK_WIDGET (area), 300, 300);
  view_attached=false;
  
  button = gtk_button_new_with_label (" exit ");
  gtk_box_pack_end (GTK_BOX (vbox), button, FALSE, FALSE, 0);
  g_signal_connect (G_OBJECT (button), "clicked", 
		    G_CALLBACK (clicked_cb), 0);

  button = gtk_button_new_with_label (" Attach View ");
  gtk_box_pack_end (GTK_BOX (vbox), button, FALSE, FALSE, 0);
  g_signal_connect (G_OBJECT (button), "clicked", 
		    G_CALLBACK (detach_attach_cb), vbox);
  detach_attach_cb (GTK_BUTTON(button), vbox);

  button = gtk_button_new_with_label (" Get offscreen snapshot ");
  gtk_box_pack_end (GTK_BOX (vbox), button, FALSE, FALSE, 0);
  g_signal_connect (G_OBJECT (button), "clicked", 
		    G_CALLBACK (offscreen_snapshot_cb),
                    gtk_widget_get_toplevel(vbox));

  gtk_widget_show_all (GTK_WIDGET (win));
  return area;
}

int main (int argc, char *argv[])
{
  gtk_init(&argc, &argv);
  gtk_gl_init (&argc, &argv);
  
  view = new mips::VtkView() ;
  
  build_vtk_ui (view);

  build_gtk_ui (view->GetGtkWidget());
  gtk_main ();

  return 0;
}
