
#include "vtkRenderWindow.h"
#include "mipsVtkViewImage2D.h"
#include "MIPS_GUIPKG_config.h"

#include <gtk/gtk.h>
#include "gtk/gtkgl.h"

#include <assert.h>
#include <string>


#include<vtkStructuredPointsReader.h>
#include<vtkStructuredPoints.h>
#include<vtkImageData.h>

mips::VtkViewImage2D * view;
bool view_attached;


static void
build_vtk_ui (mips::VtkViewImage2D * p_view, char * p_filename)
{
  std::string filename;
  if (p_filename) {
    filename = p_filename;
  }
  else {
    filename = std::string(TESTDATA_DIR) +
      std::string("/Images/MRI16bitsFull.vtk");
  }  

  vtkStructuredPointsReader* reader = vtkStructuredPointsReader::New();
  reader->SetFileName(filename.c_str());
  if(! reader->IsFileStructuredPoints ()) {
    std::cout << filename <<" is not a valid vtk StructuredPoints file !\n";
    exit(1);    
  }
  
  vtkStructuredPoints* image = (reader->GetOutput());
  image->Update();
  p_view->SetImage(image);
  p_view->ResetCurrentPoint();
  p_view->ResetWindowLevel();  
}

static void
detach_attach_cb (GtkButton *btn, gpointer data)
{
  assert (data != 0);
  if (view_attached) {    
    view_attached = false;  
    gtk_container_remove(GTK_CONTAINER(data),view->GetGtkWidget());
    gtk_button_set_label (GTK_BUTTON(btn),"Attach");
  }
  else {
        view_attached = true;
        gtk_box_pack_start (GTK_BOX (data),view->GetGtkWidget() ,
                            TRUE, TRUE, 0);
        gtk_button_set_label (GTK_BUTTON(btn),"Detach");
  }
}


static void
offscreen_snapshot_cb (GtkButton *btn, gpointer data)
{
  assert (data != 0);

  GtkWidget* image = gtk_image_new_from_pixbuf(view->CreatePixbuf (300,300));
   
   /* Create the widgets */
   GtkWidget* dialog = gtk_dialog_new_with_buttons ("Offscreen rendered snapshot",
                                         GTK_WINDOW(data),
                                         GTK_DIALOG_DESTROY_WITH_PARENT,
                                         GTK_STOCK_OK,
                                         GTK_RESPONSE_NONE,
                                         NULL);
   gtk_window_set_position(GTK_WINDOW(dialog),GTK_WIN_POS_NONE);
   
   /* Ensure that the dialog box is destroyed when the user responds. */
   g_signal_connect_swapped (GTK_OBJECT (dialog), 
                             "response", 
                             G_CALLBACK (gtk_widget_destroy),
                             GTK_OBJECT (dialog));
   /* Add the label, and show everything we've added to the dialog. */
   gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox),
                      image);
   gtk_widget_show_all (dialog);
}


static void
clicked_cb (GtkButton *btn, gpointer data)
{
  assert (data == 0);
  /* trigger end of main loop */
  gtk_main_quit ();
}

static gboolean
delete_cb (GtkWidget *widget, GdkEvent *event, gpointer data)
{
  assert (data == 0);
  /* trigger end of main loop */
  gtk_main_quit ();
  return TRUE;
}


static GtkWidget *
build_gtk_ui (GtkWidget *area)
{
  GtkWidget *win, *button,  *vbox;
  win = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  g_signal_connect (G_OBJECT (win), "delete-event",
		    G_CALLBACK (delete_cb), 0);
  vbox = gtk_vbox_new (FALSE, 5);
  gtk_container_add (GTK_CONTAINER (win), vbox);

  gtk_widget_set_size_request (GTK_WIDGET (area), 300, 300);
  view_attached=false;
  
  button = gtk_button_new_with_label (" exit ");
  gtk_box_pack_end (GTK_BOX (vbox), button, FALSE, FALSE, 0);
  g_signal_connect (G_OBJECT (button), "clicked", 
		    G_CALLBACK (clicked_cb), 0);

  button = gtk_button_new_with_label (" Attach View ");
  gtk_box_pack_end (GTK_BOX (vbox), button, FALSE, FALSE, 0);
  g_signal_connect (G_OBJECT (button), "clicked", 
		    G_CALLBACK (detach_attach_cb), vbox);
  detach_attach_cb (GTK_BUTTON(button), vbox);

  button = gtk_button_new_with_label (" Get offscreen snapshot ");
  gtk_box_pack_end (GTK_BOX (vbox), button, FALSE, FALSE, 0);
  g_signal_connect (G_OBJECT (button), "clicked", 
		    G_CALLBACK (offscreen_snapshot_cb),
                    gtk_widget_get_toplevel(vbox));

  gtk_widget_show_all (GTK_WIDGET (win));
  return area;
}

int main (int argc, char *argv[])
{
  gtk_init(&argc, &argv);
  gtk_gl_init (&argc, &argv);

  char * filename = 0;  
  if (argc == 2) {
    filename = argv[1];
  }
  
  view = new mips::VtkViewImage2D() ;
  view->SetFullInteraction();
  view->SetName("Example_View");  
  
  build_vtk_ui (view, filename);
  
  build_gtk_ui (view->GetGtkWidget());
  gtk_main ();

  return 0;
}
