/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */

#include "vtkRendererCollection.h"
#include "vtkRenderer.h"
#include "vtkCommand.h"
#include "vtkMapper.h"
#include "vtkActor.h"
#include "vtkPropCollection.h"
#include "GtkRenderWindow.h"
#include <vector>


#include <GL/gl.h>

#include <assert.h>

// instantiate static New method.
#include "vtkObjectFactory.h"
vtkStandardNewMacro(GtkRenderWindow);

GtkRenderWindow::GtkRenderWindow ()
        : m_area (0),
        m_start_has_run (FALSE),
        m_renderOffscreen (FALSE)
{
        Initialize ();
}
GtkRenderWindow::~GtkRenderWindow ()
{}

#ifdef O        
static void
print_gl_config_attrib (GdkGLConfig *glconfig,
                        const gchar *attrib_str,
                        int          attrib,
                        gboolean     is_boolean)
{
        int value;

        g_print ("%s = ", attrib_str);
        if (gdk_gl_config_get_attrib (glconfig, attrib, &value)) {
                if (is_boolean)
                        g_print ("%s\n", value == TRUE ? "TRUE" : "FALSE");
                else
                        g_print ("%d\n", value);
        } else {
                g_print ("*** Cannot get %s attribute value\n", attrib_str);
        }
}

static void
examine_gl_config_attrib (GdkGLConfig *glconfig)
{
        g_print ("\nOpenGL visual configurations :\n\n");
        
        g_print ("gdk_gl_config_is_rgba (glconfig) = %s\n",
                 gdk_gl_config_is_rgba (glconfig) ? "TRUE" : "FALSE");
        g_print ("gdk_gl_config_is_double_buffered (glconfig) = %s\n",
                 gdk_gl_config_is_double_buffered (glconfig) ? "TRUE" : "FALSE");
        g_print ("gdk_gl_config_is_stereo (glconfig) = %s\n",
                 gdk_gl_config_is_stereo (glconfig) ? "TRUE" : "FALSE");
        g_print ("gdk_gl_config_has_alpha (glconfig) = %s\n",
                 gdk_gl_config_has_alpha (glconfig) ? "TRUE" : "FALSE");
        g_print ("gdk_gl_config_has_depth_buffer (glconfig) = %s\n",
                 gdk_gl_config_has_depth_buffer (glconfig) ? "TRUE" : "FALSE");
        g_print ("gdk_gl_config_has_stencil_buffer (glconfig) = %s\n",
                 gdk_gl_config_has_stencil_buffer (glconfig) ? "TRUE" : "FALSE");
        g_print ("gdk_gl_config_has_accum_buffer (glconfig) = %s\n",
                 gdk_gl_config_has_accum_buffer (glconfig) ? "TRUE" : "FALSE");
        
        g_print ("\n");
        
        print_gl_config_attrib (glconfig, "GDK_GL_USE_GL",           GDK_GL_USE_GL,           TRUE);
        print_gl_config_attrib (glconfig, "GDK_GL_BUFFER_SIZE",      GDK_GL_BUFFER_SIZE,      FALSE);
        print_gl_config_attrib (glconfig, "GDK_GL_LEVEL",            GDK_GL_LEVEL,            FALSE);
        print_gl_config_attrib (glconfig, "GDK_GL_RGBA",             GDK_GL_RGBA,             TRUE);
        print_gl_config_attrib (glconfig, "GDK_GL_DOUBLEBUFFER",     GDK_GL_DOUBLEBUFFER,     TRUE);
        print_gl_config_attrib (glconfig, "GDK_GL_STEREO",           GDK_GL_STEREO,           TRUE);
        print_gl_config_attrib (glconfig, "GDK_GL_AUX_BUFFERS",      GDK_GL_AUX_BUFFERS,      FALSE);
        print_gl_config_attrib (glconfig, "GDK_GL_RED_SIZE",         GDK_GL_RED_SIZE,         FALSE);
        print_gl_config_attrib (glconfig, "GDK_GL_GREEN_SIZE",       GDK_GL_GREEN_SIZE,       FALSE);
        print_gl_config_attrib (glconfig, "GDK_GL_BLUE_SIZE",        GDK_GL_BLUE_SIZE,        FALSE);
        print_gl_config_attrib (glconfig, "GDK_GL_ALPHA_SIZE",       GDK_GL_ALPHA_SIZE,       FALSE);
        print_gl_config_attrib (glconfig, "GDK_GL_DEPTH_SIZE",       GDK_GL_DEPTH_SIZE,       FALSE);
        print_gl_config_attrib (glconfig, "GDK_GL_STENCIL_SIZE",     GDK_GL_STENCIL_SIZE,     FALSE);
        print_gl_config_attrib (glconfig, "GDK_GL_ACCUM_RED_SIZE",   GDK_GL_ACCUM_RED_SIZE,   FALSE);
        print_gl_config_attrib (glconfig, "GDK_GL_ACCUM_GREEN_SIZE", GDK_GL_ACCUM_GREEN_SIZE, FALSE);
        print_gl_config_attrib (glconfig, "GDK_GL_ACCUM_BLUE_SIZE",  GDK_GL_ACCUM_BLUE_SIZE,  FALSE);
        print_gl_config_attrib (glconfig, "GDK_GL_ACCUM_ALPHA_SIZE", GDK_GL_ACCUM_ALPHA_SIZE, FALSE);

        g_print ("\n");
}
#endif

void
GtkRenderWindow::MarkDirty (void)
{
        vtkRendererCollection* ren_collection = 0;
        ren_collection = this->GetRenderers();
        ren_collection->InitTraversal();
        for (vtkRenderer* ren = ren_collection->GetNextItem();
             ren != NULL;
             ren = ren_collection->GetNextItem()) {
                vtkPropCollection* prop_col = ren->GetProps();
                prop_col->InitTraversal();
                for(vtkProp* prop = prop_col->GetNextProp();
                    prop != NULL;
                    prop = prop_col->GetNextProp()) {
                        vtkActor* actor = vtkActor::SafeDownCast(prop);
                        if (actor != NULL) {
                                actor->GetMapper()->Modified ();
                        }
                        else {
                                prop->ReleaseGraphicsResources(this);
                                prop->Modified ();
                        }                        
                }
        }
}


gboolean
GtkRenderWindow::OnSizeAllocate (GtkWidget *widget,
                                 GtkAllocation *allocation,
                                 gpointer user_data)
{
        GtkRenderWindow * self = (GtkRenderWindow *)user_data;
        //cout<< "GtkRenderWindow::OnSizeAllocate ->OScreen="<< self->OffScreenRendering<<endl;
        self->SetSize (allocation->width, allocation->height);
        if (widget->window != NULL) {
                GdkGLDrawable *gldrawable = gdk_window_get_gl_drawable (GDK_WINDOW (widget->window));
                // this ensures the OpenGL rendering pipeline will use the correct window size 
                // when executing the following OpenGL rendering/compositing operations.
                gdk_gl_drawable_wait_gdk (GDK_GL_DRAWABLE (gldrawable));
        }
        return TRUE;
}

gboolean
GtkRenderWindow::OnMap (GtkWidget *widget,
                        GdkEvent *event,
                        gpointer user_data)
{
        //cout<< "GtkRenderWindow::OnMap\n";        
        GtkRenderWindow * self = (GtkRenderWindow *)user_data;
        self->SetMapped (1);
        return TRUE;
}

gboolean
GtkRenderWindow::OnUnmap (GtkWidget *widget,
                          GdkEvent *event,
                          gpointer user_data)
{
        //cout<< "GtkRenderWindow::OnUnmap\n";        
        GtkRenderWindow * self = (GtkRenderWindow *)user_data;
        self->SetMapped (0);
        return TRUE;
}
void
GtkRenderWindow::OnRealize (GtkWidget *widget,
                            gpointer user_data)
{
        //cout<< "GtkRenderWindow::OnRealize\n";
        GtkRenderWindow * self = (GtkRenderWindow *)user_data;
        gint width , height;
        // we mark the system as dirty to make sure that the model will
        // trigger a complete redraw after a detach/attach
        self->MarkDirty ();
        // the code below fixes the position and size of 
        // the drawing area after a detach/attach.
        gdk_drawable_get_size (GDK_DRAWABLE (widget->window),&width, &height);
        self->SetSize (width, height);
        // Used to restore the opengl state machine after dettach/attach.
        self->m_start_has_run = FALSE;
}

gboolean
GtkRenderWindow::OnExpose (GtkWidget *widget,
                           GdkEvent *event,
                           gpointer user_data)
{
        //cout<< "GtkRenderWindow::OnExpose\n";
        GdkGLDrawable *gldrawable = gdk_window_get_gl_drawable (GDK_WINDOW (widget->window));
        gdk_gl_drawable_wait_gdk (gldrawable);
        // propagate the event further
        return FALSE;
}

void
GtkRenderWindow::Initialize (void)
{
        gint major, minor;
        GdkGLConfig *glconfig;
        gdk_gl_query_version (&major, &minor);
        //g_print ("\nOpenGL extension version - %d.%d\n",
        //         major, minor);
        
        /* Try double-buffered visual */
        glconfig = gdk_gl_config_new_by_mode ((GdkGLConfigMode)
                                              (GDK_GL_MODE_RGB   |
                                               GDK_GL_MODE_DEPTH  |
                                               GDK_GL_MODE_DOUBLE));
        if (glconfig == NULL) {
                g_print ("*** Cannot find the double-buffered visual.\n");
                g_print ("*** Trying single-buffered visual.\n");
                        
                /* Try single-buffered visual */
                glconfig = gdk_gl_config_new_by_mode ((GdkGLConfigMode)
                                                      (GDK_GL_MODE_RGB  |
                                                       GDK_GL_MODE_DEPTH));
                if (glconfig == NULL) {
                        g_print ("*** No appropriate OpenGL-capable visual found.\n");
                        exit (1);
                }
        }
        
        // examine_gl_config_attrib (glconfig);

        m_area = gtk_drawing_area_new ();
        GTK_WIDGET_SET_FLAGS (GTK_WIDGET (m_area), GTK_CAN_FOCUS);
        g_signal_connect (G_OBJECT (m_area), "size-allocate",
                          G_CALLBACK (GtkRenderWindow::OnSizeAllocate), this);
        g_signal_connect (G_OBJECT (m_area), "map-event",
                          G_CALLBACK (GtkRenderWindow::OnMap), this);
        g_signal_connect (G_OBJECT (m_area), "realize", 
                          G_CALLBACK (GtkRenderWindow::OnRealize), this);
        g_signal_connect (G_OBJECT (m_area), "unmap-event",
                          G_CALLBACK (GtkRenderWindow::OnUnmap), this);
        g_signal_connect (G_OBJECT (m_area), "expose-event",
                          G_CALLBACK (GtkRenderWindow::OnExpose), this);
        // Used to remove black redraw after detach/attach.
        gtk_widget_set_double_buffered (GTK_WIDGET (m_area), FALSE);

        /* Set OpenGL-capability to the widget. */
        gtk_widget_set_gl_capability (GTK_WIDGET (m_area),
                                      glconfig,
                                      NULL,
                                      TRUE,
                                      GDK_GL_RGBA_TYPE);
        /* We can complete initialization only later. We do it in FirstStart. */
}

void GtkRenderWindow::Render (void)
{
        // chain up.
        vtkOpenGLRenderWindow::Render ();
}

void GtkRenderWindow::FirstStart (void)
{
        if (!m_start_has_run) {
                m_start_has_run = TRUE;
                MakeCurrent ();
                OpenGLInit ();
                glAlphaFunc(GL_GREATER,0);
        }
}



void GtkRenderWindow::Start()
{
        FirstStart ();
        MakeCurrent ();
}
void GtkRenderWindow::MakeCurrent()
{
        if (!m_area || !GTK_WIDGET_REALIZED (m_area)
            || m_renderOffscreen) {
                return;
        }
                    
        GdkGLDrawable *gldrawable;
        GdkGLContext *context;
        gldrawable =  gtk_widget_get_gl_drawable (GTK_WIDGET (m_area));
        context = gtk_widget_get_gl_context (GTK_WIDGET (m_area));
        if  (context != gdk_gl_context_get_current ()) {
                gdk_gl_drawable_make_current (GDK_GL_DRAWABLE (gldrawable), 
                                              context);
        }
}
void GtkRenderWindow::Frame()
{
        MakeCurrent ();
        glFlush ();
        if (!GetAbortRender () && GetSwapBuffers () && this->DoubleBuffer
            && !m_renderOffscreen) {
                assert (m_area && GTK_WIDGET_REALIZED (m_area));
                GdkGLDrawable *drawable;
                drawable =  gtk_widget_get_gl_drawable (GTK_WIDGET (m_area));
                gdk_gl_drawable_swap_buffers (GDK_GL_DRAWABLE (drawable));
        }
}
void GtkRenderWindow::HideCursor()
{
        if (!m_area || !GTK_WIDGET_REALIZED (m_area)) {
                return;
        }

        // create a blank cursor and set it on the window of the widget.
        static char blank_bits[] = {
                0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
 
        GdkCursor *cursor;
        GdkPixmap *source, *mask;
        GdkColor fg = { 0, 0, 0, 0 };
        GdkColor bg = { 0, 0, 0, 0 };
 
 
        source = gdk_bitmap_create_from_data (NULL, blank_bits,
                                              16, 16);
        mask = gdk_bitmap_create_from_data (NULL, blank_bits,
                                            16, 16);
        cursor = gdk_cursor_new_from_pixmap (source, mask, &fg, &bg, 8, 8);
        gdk_pixmap_unref (source);
        gdk_pixmap_unref (mask);
        gdk_window_set_cursor (GDK_WINDOW (GTK_WIDGET (m_area)->window), cursor);
}
void GtkRenderWindow::ShowCursor()
{
        if (!m_area || !GTK_WIDGET_REALIZED (m_area)) {
                return;
        }

        SetCurrentCursor (GetCurrentCursor ());
}
void GtkRenderWindow::SetCurrentCursor(int shape)
{
        if (!m_area || !GTK_WIDGET_REALIZED (m_area)) {
                goto out;
        }

        GdkCursor *cursor;

        switch (shape) {
        case  VTK_CURSOR_DEFAULT:
                // window now uses the cursor of its parent window.
                cursor = NULL;
                break;
        case VTK_CURSOR_ARROW:
                cursor = gdk_cursor_new (GDK_ARROW);
                break;
        case VTK_CURSOR_SIZEALL:
                cursor = gdk_cursor_new (GDK_FLEUR);
                break;
        case VTK_CURSOR_SIZENS:
                cursor = gdk_cursor_new (GDK_SB_V_DOUBLE_ARROW);
                break;
        case VTK_CURSOR_SIZEWE:
                cursor = gdk_cursor_new (GDK_SB_H_DOUBLE_ARROW);
                break;
        case VTK_CURSOR_SIZENE:
                cursor = gdk_cursor_new (GDK_TOP_RIGHT_CORNER);
                break;
        case VTK_CURSOR_SIZENW:
                cursor = gdk_cursor_new (GDK_TOP_LEFT_CORNER);
                break;
        case VTK_CURSOR_SIZESE:
                cursor = gdk_cursor_new (GDK_BOTTOM_RIGHT_CORNER);
                break;
        case VTK_CURSOR_SIZESW:
                cursor = gdk_cursor_new (GDK_BOTTOM_LEFT_CORNER);
                break;
        case VTK_CURSOR_HAND:
                cursor = gdk_cursor_new (GDK_HAND1);
                break;
        }

        gdk_window_set_cursor (GDK_WINDOW (GTK_WIDGET (m_area)->window), cursor);
 out:
        return;
}
void GtkRenderWindow::SetFullScreen(int)
{
        // we can't implement this since we are dealing with a single widget.
        // this request would need to be forwarded to the parent toplevel window of the widget.
        // can we do this ?
        // XXX
}
void GtkRenderWindow::WindowRemap()
{
        /**
           This function is useful only if we provide a new "windowid" to use next after the remap.
           basically, here, it would mean using a new widget provided by a SetNextWidget method.
           However, this is pretty hard because we would need to notify the RenderWindowInteractor
           that it must use the new widget for its signal connections and give up on its
           previous widget.
           XXX
         */
#if 0
        gtk_widget_unmap (GTK_WIDGET (m_area));
        gtk_widget_map (GTK_WIDGET (m_area));
#endif
}
int GtkRenderWindow::GetEventPending()
{
        if (!m_area || !GTK_WIDGET_REALIZED (m_area)) {
                return 0;
        }

        /**
         * We must return 1 if only if there are mouse button 
         * events pending.
         */
        int ok;
        GdkEvent *next_event;
        next_event = gdk_event_peek ();
        if (next_event) {
                // there is an event pending.
                switch (next_event->type) {
                case GDK_BUTTON_PRESS:
                case GDK_2BUTTON_PRESS:
                case GDK_3BUTTON_PRESS:
                case GDK_BUTTON_RELEASE:
                        ok = 1;
                        break;
                default:
                        ok = 0;
                        break;
                }

                // free the copy of the event we peeked into the event queue.
                gdk_event_free (next_event);
        }
        return ok;
}

/* I think none of the Set functions below are applicable. */
void GtkRenderWindow::SetDisplayId(void *)
{}
void GtkRenderWindow::SetWindowId(void *id)
{}
void GtkRenderWindow::SetParentId(void *)
{}
void GtkRenderWindow::SetWindowInfo(char *)
{}
void GtkRenderWindow::SetParentInfo(char *)
{}
/* */
void *GtkRenderWindow::GetGenericDisplayId()
{
        return 0;
}
void *GtkRenderWindow::GetGenericWindowId()
{
        return m_area;
}
void *GtkRenderWindow::GetGenericParentId()
{
        return 0;
}
void *GtkRenderWindow::GetGenericContext()
{
        return 0;
}
void *GtkRenderWindow::GetGenericDrawable()
{        return 0;
}
void GtkRenderWindow::SetNextWindowId(void *arg)
{
}
void GtkRenderWindow::SetNextWindowInfo(char *info)
{
}

GdkPixmap *
GtkRenderWindow::GetPixmap(int p_width, int p_height)
{
        GdkGLConfig* config;
        GdkPixmap *pixmap;
        GdkGLPixmap* glpixmap;
        GdkGLContext *context;
        
        // Try single-buffered visual
        // XXX: free config ?
        config = gdk_gl_config_new_by_mode((GdkGLConfigMode) (GDK_GL_MODE_RGB   |
                                                              GDK_GL_MODE_DEPTH |
                                                              GDK_GL_MODE_SINGLE));
        if (!config) {
                std::cerr << "*** Cannot find any OpenGL-capable visual.\n";
                exit(1);
        }        
        pixmap = gdk_pixmap_new(NULL, p_width, p_height,
                                gdk_gl_config_get_depth (config));
        glpixmap = gdk_pixmap_set_gl_capability (pixmap, config, NULL);
        context = gdk_gl_context_new(GDK_GL_DRAWABLE (glpixmap),
                                     NULL, FALSE, GDK_GL_RGBA_TYPE);

        // ensure the rendering pipeline will perform a full redraw
        MarkDirty ();

        gdk_gl_drawable_make_current (GDK_GL_DRAWABLE (glpixmap),
                                      context);

        // perform the actual rendering into the pixmap.
        this->DoubleBuffer = FALSE;
        this->SetSize (p_width, p_height);
        gdk_gl_drawable_wait_gdk (GDK_GL_DRAWABLE (glpixmap));

        m_renderOffscreen = TRUE;
        this->OpenGLInit (); 
        glAlphaFunc(GL_GREATER,0);
        this->Render();
        m_renderOffscreen = FALSE;
        glFlush ();
        // make sure the previously-posted rendering operations are completed
        // before destroying the context and returning the pixmap.
        gdk_gl_drawable_wait_gl (GDK_GL_DRAWABLE (glpixmap));
        gdk_gl_drawable_wait_gdk (GDK_GL_DRAWABLE (glpixmap));

        // rendering is done. Try now to restore the previous state.
        gdk_gl_context_destroy (context);
        this->DoubleBuffer = TRUE;

        if (m_area && GTK_WIDGET_REALIZED (m_area)) {
                gint width , height;
                gdk_drawable_get_size (GDK_DRAWABLE (m_area->window),&width, &height);
                this->SetSize (width, height);
                GdkGLDrawable *gldrawable = gdk_window_get_gl_drawable (GDK_WINDOW (m_area->window));
                gdk_gl_drawable_wait_gdk (gldrawable);
                MakeCurrent ();
        }
        MarkDirty ();

        return pixmap;
}

/*******
 * RenderWindow factory
 *******/

#include "vtkVersion.h"
vtkStandardNewMacro(GtkRenderWindowFactory);
//VTK_FACTORY_INTERFACE_IMPLEMENT(GtkRenderWindowFactory);
GtkRenderWindowFactory::GtkRenderWindowFactory ()
{}
GtkRenderWindowFactory::~GtkRenderWindowFactory ()
{}
const char* GtkRenderWindowFactory::GetVTKSourceVersion(void)
{
        return VTK_SOURCE_VERSION;
}
const char* GtkRenderWindowFactory::GetDescription(void)
{
        return "Factory for GtkRenderWindows";
}
vtkObject* GtkRenderWindowFactory::CreateObject(const char* vtkclassname)
{
        // hook in place of vtkRenderWindow.
        if (strcmp (vtkclassname, "vtkRenderWindow") == 0) {
                return GtkRenderWindow::New ();
        } else {
                return 0;
        }
}
