/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */

#include "vtkObjectFactory.h"

#include "gtk/gtk.h"
#include "gtk/gtkgl.h"

#ifndef GTK_RENDER_WINDOW_H
#define GTK_RENDER_WINDOW_H

#ifdef _WIN32_
#include "vtkWin32OpenGLRenderWindow.h"
#else
#include "vtkXOpenGLRenderWindow.h"
#endif

/** \class GtkRenderWindow
    \ingroup MipsGUI
*/
class GtkRenderWindow
#ifdef _WIN32_
 : public vtkWin32OpenGLRenderWindow
#else 
 : public vtkXOpenGLRenderWindow
#endif

{
 public:
        static GtkRenderWindow *New (void);

        /* Methods inherited from vtkRenderWindow which must be implemented here.*/
        virtual void Start();
        virtual void Frame();
        virtual void MakeCurrent();
        virtual void HideCursor();
        virtual void ShowCursor();
        virtual void SetFullScreen(int);
        virtual void WindowRemap();
        virtual int GetEventPending();
        virtual void SetDisplayId(void *);
        virtual void SetWindowId(void *);
        virtual void SetParentId(void *);
        virtual void *GetGenericDisplayId();
        virtual void *GetGenericWindowId();
        virtual void *GetGenericParentId();
        virtual void *GetGenericContext();
        virtual void *GetGenericDrawable();
        virtual void SetWindowInfo(char *);
        virtual void SetParentInfo(char *);
        virtual void SetNextWindowId(void *);
        virtual void SetNextWindowInfo(char *);


        // XXX 
        virtual void Render (void);

        // methods inherited from vtkWindow
        virtual void SetCurrentCursor(int);        

        // Offscreen additional methods
        // The returned pixmap must be unrefed by the caller.
        virtual GdkPixmap* GetPixmap(int p_width, int p_height);


 protected:
        GtkRenderWindow ();
        ~GtkRenderWindow ();

 private:
        void MarkDirty (void);
        static gboolean OnSizeAllocate (GtkWidget *widget,
                                        GtkAllocation *allocation,
                                        gpointer user_data);
        static gboolean OnMap (GtkWidget *widget,
                               GdkEvent *event,
                               gpointer user_data);
        static gboolean OnUnmap (GtkWidget *widget,
                                 GdkEvent *event,
                                 gpointer user_data);
        static void OnRealize (GtkWidget *widget,
                               gpointer user_data);
        static void OnUnrealize (GtkWidget *widget,
                                 gpointer user_data);
        static gboolean OnExpose (GtkWidget *widget,
                                  GdkEvent *event,
                                  gpointer user_data);

        void FirstStart (void);
        void Initialize (void);
        GtkWidget *m_area;
        gboolean m_start_has_run;
        gboolean m_renderOffscreen;
};



/** \class GtkRenderWindowFactory
    \ingroup MipsViewer
*/
class GtkRenderWindowFactory : public vtkObjectFactory
{
 public:
        static GtkRenderWindowFactory *New (void);
        // methods inherited from ObjectFactory
        GtkRenderWindowFactory ();
        virtual ~GtkRenderWindowFactory ();
        virtual vtkObject* CreateObject(const char* vtkclassname);
        virtual const char* GetVTKSourceVersion(void);
        virtual const char* GetDescription(void);
};

#endif /* GTK_RENDER_WINDOW_H */
