/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */
#include "GtkRenderWindowInteractor.h"
#include "GtkRenderWindow.h"


#include <gtk/gtk.h>

#include "vtkCommand.h"

// instantiate static New method.
#include "vtkObjectFactory.h"
vtkStandardNewMacro(GtkRenderWindowInteractor);

GtkRenderWindowInteractor::GtkRenderWindowInteractor ()
        : m_timeout (0)
{
}

GtkRenderWindowInteractor::~GtkRenderWindowInteractor ()
{
        DestroyTimer ();
}

gboolean GtkRenderWindowInteractor::TimerTimeout (gpointer data)
{
        GtkRenderWindowInteractor *self = (GtkRenderWindowInteractor *)data;

        if (!self->GetEnabled ()) {
                return TRUE;
        }

        self->InvokeEvent(vtkCommand::TimerEvent, NULL);

        self->m_timeout = 0;

        return FALSE;
}

int GtkRenderWindowInteractor::CreateTimer(int vtkNotUsed(timertype))
{
        m_timeout = gtk_timeout_add (10, GtkRenderWindowInteractor::TimerTimeout, this);
        return 1;
}

int GtkRenderWindowInteractor::DestroyTimer (void)
{
        // XXX: I _think_ the timeout id cannot be 0
        if (m_timeout) {
                gtk_timeout_remove (m_timeout);
        }
        return 1;
}

gboolean GtkRenderWindowInteractor::OnRealize (GtkWidget *widget, gpointer data)
{
        //cout<< "GtkRenderWindowInteractor::OnRealize\n";
        GtkRenderWindowInteractor *self = (GtkRenderWindowInteractor *)data;
        // chain up.
        self->vtkRenderWindowInteractor::Initialize ();
        return FALSE;
}
gboolean GtkRenderWindowInteractor::OnConfigure (GtkWidget *widget, 
                                                 GdkEventConfigure *event,
                                                 gpointer data)
{
        //cout << "GtkRenderWindowInteractor::OnConfigure\n";
        //cout << "width: " << event->width << " height: " << event->height << "\n";
        GtkRenderWindowInteractor *self = (GtkRenderWindowInteractor *)data;
        self->UpdateSize (event->width, event->height);
        self->InvokeEvent(vtkCommand::ConfigureEvent, NULL);
        return TRUE;
}

gboolean GtkRenderWindowInteractor::OnExpose (GtkWidget *widget,
                                              GdkEventExpose *event,
                                              gpointer user_data)
{
        // cout<< "GtkRenderWindowInteractor::OnExpose\n";        
        GtkRenderWindowInteractor *self = (GtkRenderWindowInteractor *) user_data;
        self->InvokeEvent(vtkCommand::ExposeEvent, NULL);
        self->Render ();
        // We propagate the event further
        return FALSE;
}

gboolean GtkRenderWindowInteractor::OnMotionNotify (GtkWidget *widget,
                                                    GdkEventMotion *event,
                                                    gpointer user_data)
{
        GtkRenderWindowInteractor *self = (GtkRenderWindowInteractor *) user_data;
        int x, y;
        GdkModifierType state;
        int ctrl = (event->state & GDK_CONTROL_MASK)?1:0;
        int shift =  (event->state & GDK_SHIFT_MASK)?1:0;

        if (!self->GetEnabled ()) {
                goto out;
        }

        if (event->is_hint) {
                gdk_window_get_pointer (event->window, &x, &y, &state);
        } else {
                x = (int)event->x;
                y = (int)event->y;
                state = (GdkModifierType)event->state;
        }

        self->SetEventInformationFlipY (x, y, ctrl, shift);
        self->InvokeEvent(vtkCommand::MouseMoveEvent, NULL);
 out:
        // We stop other handlers from being invoked for the event
        return TRUE;
}

gboolean GtkRenderWindowInteractor::OnEnterNotify (GtkWidget *widget,
                                                   GdkEventCrossing *event,
                                                   gpointer user_data)
{
        GtkRenderWindowInteractor *self = (GtkRenderWindowInteractor *) user_data;

        if (!self->GetEnabled ()) {
                goto out;
        }

        self->SetEventPositionFlipY ((int)event->x, (int)event->y);
        self->InvokeEvent(vtkCommand::EnterEvent, NULL);
 out:
        // We propagate the event further
        return FALSE;
}

gboolean GtkRenderWindowInteractor::OnLeaveNotify (GtkWidget *widget,
                                                   GdkEventCrossing *event,
                                                   gpointer user_data)
{
        GtkRenderWindowInteractor *self = (GtkRenderWindowInteractor *) user_data;

        if (!self->GetEnabled ()) {
                goto out;
        }

        self->SetEventPositionFlipY ((int)event->x, (int)event->y);
        self->InvokeEvent(vtkCommand::LeaveEvent, NULL);
 out:
        // We propagate the event further
        return FALSE;
}

gboolean GtkRenderWindowInteractor::OnKeyPress (GtkWidget *widget,
                                                GdkEventKey *event,
                                                gpointer user_data)
{
        GtkRenderWindowInteractor *self = (GtkRenderWindowInteractor *) user_data;
        int ctrl = (event->state & GDK_CONTROL_MASK)?1:0;
        int shift =  (event->state & GDK_SHIFT_MASK)?1:0;
        int x;
        int y;
        char c;
        guint32 ucs4;

        if (!self->GetEnabled ()) {
                goto out;
        }
        // this is a ucs4-encoded unicode character
        ucs4 = gdk_keyval_to_unicode (event->keyval);
        // We have a 7 bit ascii character for the keycode.
        c = (char) ucs4;
        gtk_widget_get_pointer (GTK_WIDGET (widget), &x, &y);
        // KeySym field is used to store the name of the key
        // For instance : Up, Down, KP_Up, KP_Page_Up, a, A, b, B ...
        self->SetEventInformationFlipY (x, y, ctrl, shift, c, 0,
                                        gdk_keyval_name(event->keyval));
        self->InvokeEvent(vtkCommand::KeyPressEvent, NULL);
        self->InvokeEvent(vtkCommand::CharEvent, NULL);

 out:
        // We don't propagate the event further
        // because of the GTK focus management not
        // compatible with use of direction keys (Up Down...)
        return TRUE;
}

gboolean GtkRenderWindowInteractor::OnKeyRelease (GtkWidget *widget,
                                                  GdkEventKey *event,
                                                  gpointer user_data)
{
        GtkRenderWindowInteractor *self = (GtkRenderWindowInteractor *) user_data;

        if (!self->GetEnabled ()) {
                goto out;
        }

        self->InvokeEvent(vtkCommand::KeyReleaseEvent, NULL);
 out:
        // We propagate the event further
        return FALSE;
}


gboolean GtkRenderWindowInteractor::OnButtonPress (GtkWidget *widget,
                                                   GdkEventButton *event,
                                                   gpointer user_data)
{
        GtkRenderWindowInteractor *self = (GtkRenderWindowInteractor *) user_data;
        int ctrl = (event->state & GDK_CONTROL_MASK)?1:0;
        int shift =  (event->state & GDK_SHIFT_MASK)?1:0;
        int x = (int)event->x;
        int y = (int)event->y;

        if (!self->GetEnabled ()) {
                goto out;
        }

        self->SetEventInformationFlipY (x, y, ctrl, shift);

        switch (event->button) {
        case 1:
                self->InvokeEvent(vtkCommand::LeftButtonPressEvent, NULL);
                gtk_widget_grab_focus(GTK_WIDGET(self->GetRenderWindow()->GetGenericWindowId()));
                break;
        case 2:
                self->InvokeEvent(vtkCommand::MiddleButtonPressEvent, NULL);
                gtk_widget_grab_focus(GTK_WIDGET(self->GetRenderWindow()->GetGenericWindowId()));
                break;
        case 3:
                self->InvokeEvent(vtkCommand::RightButtonPressEvent, NULL);
                gtk_widget_grab_focus(GTK_WIDGET(self->GetRenderWindow()->GetGenericWindowId()));
                break;
        default:
                // likely to be the mouse wheel
                break;
        }
 out:
        // We propagate the event further
        return FALSE;
}

gboolean GtkRenderWindowInteractor::OnButtonRelease (GtkWidget *widget,
                                                     GdkEventButton *event,
                                                     gpointer user_data)
{
        GtkRenderWindowInteractor *self = (GtkRenderWindowInteractor *) user_data;
        int ctrl = (event->state & GDK_CONTROL_MASK)?1:0;
        int shift =  (event->state & GDK_SHIFT_MASK)?1:0;
        int x = (int)event->x;
        int y = (int)event->y;

        if (!self->GetEnabled ()) {
                goto out;
        }

        self->SetEventInformationFlipY (x, y, ctrl, shift);

        switch (event->button) {
        case 1:
                self->InvokeEvent(vtkCommand::LeftButtonReleaseEvent, NULL);
                break;
        case 2:
                self->InvokeEvent(vtkCommand::MiddleButtonReleaseEvent, NULL);
                break;
        case 3:
                self->InvokeEvent(vtkCommand::RightButtonReleaseEvent, NULL);
                break;
        default:
                // likely to be the mouse wheel
                break;
        }
 out:
        // We propagate the event further
        return FALSE;
}

void GtkRenderWindowInteractor::Initialize(void)
{
        GtkRenderWindow *vtk_win = (GtkRenderWindow *)GetRenderWindow ();
        GtkWidget *widget = GTK_WIDGET (vtk_win->GetGenericWindowId ());
        gtk_widget_add_events (GTK_WIDGET (widget), 
                               GDK_EXPOSURE_MASK | GDK_POINTER_MOTION_MASK |
                               GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
                               GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK |
                               GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK |
                               GDK_POINTER_MOTION_HINT_MASK);
        g_signal_connect (G_OBJECT (widget), "realize", 
                          G_CALLBACK (GtkRenderWindowInteractor::OnRealize), this);
        g_signal_connect (G_OBJECT (widget), "configure-event", 
                          G_CALLBACK (GtkRenderWindowInteractor::OnConfigure), this);
        g_signal_connect (G_OBJECT (widget), "expose-event", 
                          G_CALLBACK (GtkRenderWindowInteractor::OnExpose), this);
        g_signal_connect (G_OBJECT (widget), "button-press-event", 
                          G_CALLBACK (GtkRenderWindowInteractor::OnButtonPress), this);
        g_signal_connect (G_OBJECT (widget), "button-release-event", 
                          G_CALLBACK (GtkRenderWindowInteractor::OnButtonRelease), this);
        g_signal_connect (G_OBJECT (widget), "motion-notify-event", 
                          G_CALLBACK (GtkRenderWindowInteractor::OnMotionNotify), this);
        g_signal_connect (G_OBJECT (widget), "key-press-event", 
                          G_CALLBACK (GtkRenderWindowInteractor::OnKeyPress), this);
        g_signal_connect (G_OBJECT (widget), "key-release-event", 
                          G_CALLBACK (GtkRenderWindowInteractor::OnKeyRelease), this);
        g_signal_connect (G_OBJECT (widget), "leave-notify-event", 
                          G_CALLBACK (GtkRenderWindowInteractor::OnLeaveNotify), this);
        g_signal_connect (G_OBJECT (widget), "enter-notify-event", 
                          G_CALLBACK (GtkRenderWindowInteractor::OnEnterNotify), this);
}
void GtkRenderWindowInteractor::Start(void)
{}
void GtkRenderWindowInteractor::TerminateApp(void)
{}



#include "vtkVersion.h"

vtkStandardNewMacro(GtkInteractorFactory);

//VTK_FACTORY_INTERFACE_IMPLEMENT(GtkInteractorFactory);

GtkInteractorFactory::GtkInteractorFactory ()
{}

GtkInteractorFactory::~GtkInteractorFactory ()
{}

const char* GtkInteractorFactory::GetVTKSourceVersion(void)
{
        return VTK_SOURCE_VERSION;
}
const char* GtkInteractorFactory::GetDescription(void)
{
        return "Factory for GtkInteractors";
}


vtkObject* GtkInteractorFactory::CreateObject(const char* vtkclassname)
{
        // hook in place of vtkRenderWindowInteractor.
        if (strcmp (vtkclassname, "vtkRenderWindowInteractor") == 0) {
                return GtkRenderWindowInteractor::New ();
        } else {
                return 0;
        }
}
