/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */
#include "vtkGenericRenderWindowInteractor.h"
#include "vtkObjectFactory.h"

#include <gtk/gtk.h>

#ifndef GTK_RENDER_WINDOW_INTERACTOR_H
#define GTK_RENDER_WINDOW_INTERACTOR_H

/** \class GtkRenderWindowInteractor
    \ingroup MipsGUI
*/
class GtkRenderWindowInteractor : public vtkRenderWindowInteractor
{
 public:
        // methods implemented by all implementations of a
        // RenderWindowInteractor.
        static GtkRenderWindowInteractor *New (void);

        virtual int CreateTimer(int vtkNotUsed(timertype));
        virtual int DestroyTimer (void);
        
        virtual void Initialize (void);
        virtual void Start (void);
        virtual void TerminateApp (void);

 protected:
        GtkRenderWindowInteractor ();
        virtual ~GtkRenderWindowInteractor ();

 private:
        static gboolean TimerTimeout (gpointer data);
        guint m_timeout;
        // Event list : triggered...

        // Basically when created
        static gboolean  OnRealize (GtkWidget *widget, 
                                    gpointer data);
        // For example, when resized
        static gboolean OnConfigure (GtkWidget *widget, 
                                     GdkEventConfigure *event,
                                     gpointer data);
        // For example, when the window is hidden then shown again
        static gboolean OnExpose (GtkWidget *widget,
                                  GdkEventExpose *event,
                                  gpointer user_data);
        // When the mouse moves over the window
        static gboolean OnMotionNotify (GtkWidget *widget,
                                        GdkEventMotion *event,
                                        gpointer user_data);
        // When the mouse enters the window. If the user clicks and drag
        // outside the window, and then enters the window again, the event is
        // triggered again
        static gboolean OnEnterNotify (GtkWidget *widget,
                                       GdkEventCrossing *event,
                                       gpointer user_data);
        // When the mouse leaves the window
        static gboolean OnLeaveNotify (GtkWidget *widget,
                                       GdkEventCrossing *event,
                                       gpointer user_data);
        // When a key is pressed (pressed, not released)
        static gboolean OnKeyPress (GtkWidget *widget,
                                    GdkEventKey *event,
                                    gpointer user_data);
        // When a key is released
        static gboolean OnKeyRelease (GtkWidget *widget,
                                      GdkEventKey *event,
                                      gpointer user_data);
        // When a button is pressed
        static gboolean OnButtonPress (GtkWidget *widget,
                                       GdkEventButton *event,
                                       gpointer user_data);
        // When a button is released
        static gboolean OnButtonRelease (GtkWidget *widget,
                                         GdkEventButton *event,
                                         gpointer user_data);
};


/** \class GtkInteractorFactory
    \ingroup MipsViewer
*/
class GtkInteractorFactory : public vtkObjectFactory
{
 public:
        static GtkInteractorFactory *New (void);
        // methods inherited from ObjectFactory
        GtkInteractorFactory ();
        virtual ~GtkInteractorFactory ();
        virtual vtkObject* CreateObject(const char* vtkclassname);
        virtual const char* GetVTKSourceVersion(void);
        virtual const char* GetDescription(void);
};


#endif /* GTK_RENDER_WINDOW_INTERACTOR_H */
