#ifndef MIPS_DATASET_FACTORY_H
#define MIPS_DATASET_FACTORY_H

#include <vector>

namespace mips {

class DataSet;  
  
/** \class DataSetFactory
    \ingroup MipsGUI
*/
class DataSetFactory 
{
 public:
  virtual ~DataSetFactory();
  
  /** Create a new DataSet. Return NULL if th given dataSet type name do not
      correspond to any of the registered DataSet type or if there is no
      memory left.*/  
  virtual DataSet* NewDataSet(const char * p_datatSetTypeName)=0;

  /** Return the list of classe names the factory is able to create. */
  const std::vector<const char *>& GetTypeNames();  

 protected:
  /** Add a classe name that the factory is able to create. */
  void AddTypeName(const char *);  
  
  //Attributes
 private:
  /// List of classes the factory is able to create 
  std::vector<const char *> m_typeNames;  
  
};

} //namespace mips

#endif /* MIPS_DATASET_FACTORY_H */
