#include "mipsDataSetManager.h"

#include "mipsReader.h"
#include "mipsDataSet.h"
#include "mipsDataSetFactory.h"
#include <iostream>

namespace mips {

std::vector<DataSetFactory*>  DataSetManager::m_dataSetFactories;  
std::vector<DataSet*>         DataSetManager::m_dataSets;  
std::vector<Reader*>          DataSetManager::m_readers;  

const std::vector<DataSet*>&  
DataSetManager::GetDataSets() 
{
  return m_dataSets;  
}

DataSet*
DataSetManager::GetDataSetByFileName(const char* p_filename)
{
  DataSet* ds = NULL;
  // First : try to find the filename in the datasets already loaded
  for (unsigned int i=0; i < m_dataSets.size(); i++) {
    if (strcmp(m_dataSets[i]->GetFileName(), p_filename) ==0 ) {
      ds = m_dataSets[i];      
      break;
    }    
  }
  
  if(ds) {
    return ds;
  }
  
  // Second : we try to load the file ...
  if(ds == NULL) {
    DataSetManager::OpenDataSet(p_filename);    
  }

  // Thrid : No way.. :-(
  return NULL;  
}

DataSet*
DataSetManager::OpenDataSet(const char* p_filename)
{
  DataSet* ds=0;
  
  for (unsigned int i_reader=0; i_reader < m_readers.size(); i_reader++){
    ds = m_readers[i_reader]->Read(p_filename);      
    if(ds) break;          
  }
  return ds;
}

DataSet*
DataSetManager::OpenDataSet(const Reader* p_reader, const char* p_filename)
{
  DataSet* ds=0;
  if (p_reader) {    
    ds = p_reader->Read(p_filename);      
  }
  return ds;
}


void
DataSetManager::AddReader(Reader* p_reader)
{
  m_readers.push_back(p_reader);  
}

void
DataSetManager::AddDataSetFactory(DataSetFactory* p_factory)
{
  m_dataSetFactories.push_back(p_factory);  
}

DataSet*
DataSetManager::NewDataSet(const char * p_datatSetTypeName)
{
  DataSet * ds = 0;
  
  for (unsigned int i=0; i<m_dataSetFactories.size(); i++) {
    ds = m_dataSetFactories[i]->NewDataSet(p_datatSetTypeName);    
    if (ds) {
      m_dataSets.push_back(ds);
      break;
    }    
  }

  return ds;  
}

} //namespace mips
