#ifndef MIPS_DATASET_MANAGER_H
#define MIPS_DATASET_MANAGER_H

#include <vector>

namespace mips {

class Reader;
class DataSet;
class DataSetFactory;
 
/** \class DataSetManager
    \ingroup MipsGUI
*/
class DataSetManager 
{
 public:
  /** Create a new DataSet. Return NULL if th given dataSet type name do not
      correspond to any of the registered DataSet type or if there is no
      memory left.*/
  static DataSet* NewDataSet(const char * p_datatSetTypeName);  

  /** Get the list of DataSets loaded in memory */
  static const std::vector<DataSet*>& GetDataSets();

  /** Return the DataSet associated to the specified filename.
      If the DataSet have not been loaded yet, the given file in read with
      the OpenDataSet method.*/
  static DataSet* GetDataSetByFileName(const char* p_filename);
  
  /** Load the DataSet associated to the specified filename with the first
      capable registered reader. If no reader capable is found, return 0. */
  static DataSet* OpenDataSet(const char* p_filename);
  
  /** Load the DataSet associated to the specified filename with the given reader
      then return it if the operation is succesfull. Else return 0.*/
  static DataSet* OpenDataSet(const Reader* p_reader, const char* p_filename);
  
  /** Add a reader to try when loading a DataSet from a file.*/
  static void AddReader(Reader* p_reader);

  /** Register a DataSetFactory to the manager.*/
  static void AddDataSetFactory(DataSetFactory* p_factory);

 private:
  /** The Factory have not to be instanciated.  */
  DataSetManager(){};

  // Attributes
 private:  
  static std::vector<DataSetFactory*>  m_dataSetFactories;  
  static std::vector<DataSet*>         m_dataSets;  
  static std::vector<Reader*>          m_readers;
};

} //namespace mips

#endif /* MIPS_DATASET_MANAGER_H */
