#ifndef MIPS_GLADE_GUI_H
#define MIPS_GLADE_GUI_H

#include <glade/glade.h>

namespace mips {

/** \class GladeGui
    \ingroup MipsGUI
*/
class GladeGui
{
 public:
  /** Create a GladeGui from the specified glade project xml file.  */
  GladeGui(const char * p_filename);

  /** Destroy GladeGui.  */
  ~GladeGui();

  /** Return the widget searched with the specified name in the glade project.*/
  GtkWidget* GetWidget(const char * p_widgetName);

  /** Do a g_signal_connect on the widget corresponding to the widget_name 
	  located in the glade project managed by this object.
	  @param widget_name : the name of the widget from which the signal is 
	  comming.
	  @param detailed_signal : a string of the form "signal-name::detail".
	  @param cb_handler : the callback to connect.
	  @param user_data : data to pass to c_handler calls.
	  @return : the handler id.
  */
  gulong SignalConnect(const char* widget_name,const  char* detailed_signal, 
                       GCallback cb_handler, gpointer user_data);

  /** Do a gtk_widget_reparent on the widget corresponding to the given 
	  widget name */
	void Reparent(const char* p_widgetName, GtkWidget* p_newParent); 

 private: 
  /** Default constructor is not allowed  */
  GladeGui();

  static const char * GLADE_FILES_DIRECTORY;
  GladeXML* m_gladeXml;

};

} //namespace mips

#endif /* MIPS_GLADE_GUI_H */
