#include "mipsInteractorStyleImage2D.h"
#include "mipsVtkViewImage2D.h"

#include "vtkAbstractPropPicker.h"
#include "vtkAssemblyPath.h"
#include "vtkMath.h"
#include "vtkObjectFactory.h"
#include "vtkRenderWindowInteractor.h"
#include "mipsVtkViewImage2DCommand.h"

namespace mips
{

//----------------------------------------------------------------------------
InteractorStyleImage2D::InteractorStyleImage2D():
  vtkInteractorStyleImage(), m_view(0),  m_ZSliceMoveStep(0)                
{
}

//----------------------------------------------------------------------------
InteractorStyleImage2D::~InteractorStyleImage2D() 
{
}

//----------------------------------------------------------------------------
void InteractorStyleImage2D::OnMouseMove() 
{
  int x = this->Interactor->GetEventPosition()[0];
  int y = this->Interactor->GetEventPosition()[1];

  switch (this->State) 
    {
    case VTKIS_WINDOW_LEVEL:
      this->FindPokedRenderer(x, y);
      this->WindowLevel();
      this->InvokeEvent(vtkCommand::InteractionEvent, NULL);
      break;

    case VTKIS_PICK:
      // No Drag is allowed for picking. Toggle to ZSliceMove state
      // We force StartZSliceMove without ending the picking.
      // We don't want to pick at all
      this->StopState();
      this->StartZSliceMove();
      // There we don't break to let the hand to ZSliceMove      
    case MIPSIS_ZSLICE_MOVE:
      this->FindPokedRenderer(x, y);
      this->ZSliceMove();
      this->InvokeEvent(vtkCommand::InteractionEvent, NULL);
      break;
    case VTKIS_DOLLY:
      this->FindPokedRenderer(x, y);
      this->Dolly();
      this->InvokeEvent(vtkCommand::InteractionEvent, NULL);
      break;
    }
}

//----------------------------------------------------------------------------
void InteractorStyleImage2D::OnLeftButtonDown() 
{
  int x = this->Interactor->GetEventPosition()[0];
  int y = this->Interactor->GetEventPosition()[1];

  this->FindPokedRenderer(x, y);
  if (this->CurrentRenderer == NULL) {
    return;
  }
  
  switch (m_view->GetInteractionStyle()) {
  case View::WINDOW_LEVEL_INTERACTION :
    this->WindowLevelStartPosition[0] = x;
    this->WindowLevelStartPosition[1] = y;      
    this->StartWindowLevel();
    break;
  case View::SELECT_INTERACTION :
    this->StartPick();
    break;    
  case View::FULL_PAGE_INTERACTION :
    this->FullPage();
    break;
  case View::MEASURE_INTERACTION :
    this->StartMeasure();
    break;
  case View::ZOOM_INTERACTION :
    this->StartDolly();// continuous zoom
    break;
  }
}

//----------------------------------------------------------------------------
void InteractorStyleImage2D::OnLeftButtonUp()
{
  switch (this->State) 
    {
    case MIPSIS_ZSLICE_MOVE:
      this->EndZSliceMove();    
      break;    
    case VTKIS_WINDOW_LEVEL:
      this->EndWindowLevel();
      break;
    case VTKIS_PICK:
      this->EndPick();    
      break;    
    case MIPSIS_MEASURE:
      this->EndMeasure();    
      break;    
    case VTKIS_DOLLY: // continuous zoom
      this->EndDolly();    
      break;    
    }
}

//----------------------------------------------------------------------------
void InteractorStyleImage2D::OnRightButtonDown() 
{
}

//----------------------------------------------------------------------------
void InteractorStyleImage2D::OnRightButtonUp() 
{
}

//----------------------------------------------------------------------------
void InteractorStyleImage2D::OnChar() 
{
  vtkRenderWindowInteractor *rwi = this->Interactor;

  std::string key_sym(rwi->GetKeySym());
  
  if ((key_sym.compare("Up")==0) || (key_sym.compare("KP_Up")==0)){
    m_ZSliceMoveStep = 1;    
    this->InvokeEvent(VtkViewImage2DCommand::ZSliceMoveEvent, this);
    m_ZSliceMoveStep = 1;    
    return;    
  }
  if ((key_sym.compare("Down")==0) || (key_sym.compare("KP_Down")==0)){
    m_ZSliceMoveStep = -1;    
    this->InvokeEvent(VtkViewImage2DCommand::ZSliceMoveEvent, this);
    return;    
  }
  if ((key_sym.compare("Page_Down")==0) || (key_sym.compare("KP_Page_Down")==0)){
    m_ZSliceMoveStep = -10;    
    this->InvokeEvent(VtkViewImage2DCommand::ZSliceMoveEvent, this);
    return;    
  }
  if ((key_sym.compare("Page_Up")==0) || (key_sym.compare("KP_Page_Up")==0)){
    m_ZSliceMoveStep = 10;    
    this->InvokeEvent(VtkViewImage2DCommand::ZSliceMoveEvent, this);
    return;    
  }
  
  switch (rwi->GetKeyCode()) 
  {
  case 'f' :      
  case 'F' :
  {
    this->AnimState = VTKIS_ANIM_ON;
    vtkAssemblyPath *path=NULL;
    this->FindPokedRenderer(rwi->GetEventPosition()[0],
                            rwi->GetEventPosition()[1]);
    rwi->GetPicker()->Pick(rwi->GetEventPosition()[0],
                           rwi->GetEventPosition()[1], 0.0, 
                           this->CurrentRenderer);
    vtkAbstractPropPicker *picker;
    if ( (picker=vtkAbstractPropPicker::SafeDownCast(rwi->GetPicker())) )
    {
      path = picker->GetPath();
    }
    if ( path != NULL )
    {
      rwi->FlyToImage(this->CurrentRenderer,picker->GetPickPosition());
    }
    this->AnimState = VTKIS_ANIM_OFF;
    break;
  }

  case 'r' :      
  case 'R' :
    if (m_view->GetInteractionStyle() == View::WINDOW_LEVEL_INTERACTION)
    {
      this->InvokeEvent(vtkCommand::ResetWindowLevelEvent, this);
    }
    break;
  }
  
}

//----------------------------------------------------------------------------
void
InteractorStyleImage2D::StartZSliceMove()
{
  if ((this->State != VTKIS_NONE) && (this->State != VTKIS_PICK)) {
    return;
  }
  this->StartState(MIPSIS_ZSLICE_MOVE);
  this->InvokeEvent(VtkViewImage2DCommand::StartZSliceMoveEvent, this);
}
void
InteractorStyleImage2D::ZSliceMove()
{
  vtkRenderWindowInteractor *rwi = this->Interactor;
  int dy = rwi->GetEventPosition()[1] - rwi->GetLastEventPosition()[1];
  m_ZSliceMoveStep = dy;  
  this->InvokeEvent(VtkViewImage2DCommand::ZSliceMoveEvent, this);
}
void
InteractorStyleImage2D::EndZSliceMove()
{
  if (this->State != MIPSIS_ZSLICE_MOVE) {
    return;
  }
  this->InvokeEvent(VtkViewImage2DCommand::EndZSliceMoveEvent, this);
  this->StopState();
}

//----------------------------------------------------------------------------
void
InteractorStyleImage2D::FullPage()
{
  this->InvokeEvent(VtkViewImage2DCommand::FullPageEvent, this);
}

//----------------------------------------------------------------------------
void
InteractorStyleImage2D::StartMeasure()
{
  if (this->State != VTKIS_NONE) {
    return;
  }
  this->StartState(MIPSIS_MEASURE);
  this->InvokeEvent(VtkViewImage2DCommand::StartMeasureEvent, this);
}
void
InteractorStyleImage2D::Measure()
{
  this->InvokeEvent(VtkViewImage2DCommand::MeasureEvent, this);
}
void
InteractorStyleImage2D::EndMeasure()
{
  if (this->State != MIPSIS_MEASURE) {
    return;
  }
  this->InvokeEvent(VtkViewImage2DCommand::EndMeasureEvent, this);
  this->StopState();
}

//----------------------------------------------------------------------------
void
InteractorStyleImage2D::SetView(VtkViewImage2D* p_view)
{
  m_view = p_view;  
}

int
InteractorStyleImage2D::GetZSliceStep()
{
  return m_ZSliceMoveStep;  
}


} // end of namespace


