#ifdef _MSC_VER
#  pragma warning (disable : 4786) 
#endif

#include <sstream>
#include <iostream>
#include "mipsGladeGui.h"
#include "mipsViewSpace.h"
#include "mipsView.h"
#include <math.h>

#include "mipsNewPageDialog.h"

namespace mips {

  enum {
    LIST_VIEW_PIXMAP_COLUMN,
    LIST_VIEW_NAME_COLUMN,
    LIST_VIEW_N_COLUMNS
  };

NewPageDialog::NewPageDialog (GtkWidget* p_parentTopWidget)
  :m_parentTopWidget(p_parentTopWidget)
{  
  // Load the main viewspace interface
  m_gladeGui = new mips::GladeGui("viewspace.glade");

  // Get and initialize the pageLayoutEditorDialog (Dialog box to add new layouts)
  m_layoutEditorDialog = m_gladeGui->GetWidget("layout_editor_dialog");
  gtk_widget_hide_all(m_layoutEditorDialog);

  // Initialize the layout list
  GtkWidget* layout_combo = m_gladeGui->GetWidget("layout_combo");
  GList *glist = NULL;
  glist = g_list_append (glist,
                         const_cast<void *>(static_cast<const void*>("Auto")));
  glist = g_list_append (glist,
                         const_cast<void *>(static_cast<const void*>("H6")));
  glist = g_list_append (glist,
                         const_cast<void*>(static_cast<const void*>("H8"))); 
  gtk_combo_set_popdown_strings (GTK_COMBO (layout_combo), glist);
  g_list_free(glist);
    
  // Initialize ViewList
  GtkCellRenderer   *renderer=0;
  GtkTreeViewColumn *column=0;
  GtkWidget* view_list_treeview = m_gladeGui->GetWidget("view_list_treeview");
  //       VIEW_NAME column
  renderer = gtk_cell_renderer_text_new ();
  column = gtk_tree_view_column_new_with_attributes ("View",
                                                     renderer,
                                                     "text",
                                                     LIST_VIEW_NAME_COLUMN,
                                                     NULL);
  gtk_tree_view_column_set_sizing (column, GTK_TREE_VIEW_COLUMN_AUTOSIZE);    
  gtk_tree_view_insert_column (GTK_TREE_VIEW(view_list_treeview), column,
                               LIST_VIEW_NAME_COLUMN);
  //       VIEW_PIX column
  renderer = gtk_cell_renderer_pixbuf_new ();
  column = gtk_tree_view_column_new_with_attributes ("Icon",
                                                     renderer,
                                                     "pixbuf",
                                                     LIST_VIEW_PIXMAP_COLUMN,
                                                     NULL);
  gtk_tree_view_column_set_sizing (column, GTK_TREE_VIEW_COLUMN_AUTOSIZE);    
  gtk_tree_view_insert_column (GTK_TREE_VIEW(view_list_treeview), column,
                               LIST_VIEW_PIXMAP_COLUMN);
  //Signal connections
  m_gladeGui->SignalConnect("layout_editor_add_button", "clicked",
                            G_CALLBACK (NewPageDialog::OnLayoutDialogAddCallback),
                            this);
  m_gladeGui->SignalConnect("layout_editor_remove_button", "clicked",
                            G_CALLBACK (NewPageDialog::OnLayoutDialogRemoveCallback),
                            this);
  //m_gladeGui->SignalConnect("layout_combo", "clicked",
  //                          G_CALLBACK(NewPageDialog::OnLayoutComboChangedCallback),
  //                          this);
}

NewPageDialog::~NewPageDialog ()
{
  if (m_gladeGui) delete m_gladeGui;  
}
  
gint
NewPageDialog::Run(const ViewSpace::ViewArrayType&    p_views,
                         ViewSpace::Page&             p_page)
{
  // Initialization of th Views list Model
  GtkWidget* view_list_treeview = m_gladeGui->GetWidget("view_list_treeview");
  GtkListStore *list_store = gtk_list_store_new (LIST_VIEW_N_COLUMNS, 
                                                 GDK_TYPE_PIXBUF, 
                                                 G_TYPE_STRING);
  GtkTreeIter iter;
  std::ostringstream os;
  unsigned int i =0;
  for (i=0; i<p_views.size() ; i++) {
    // VIEW_NAME_COLUMN
    os.str("");
    os << "View " << i ;
    // VIEW_PIXMAP_COLUMN
    GdkPixbuf* pixbuf = p_views[i]->CreatePixbuf(50,50);
    // Add line in model
    gtk_list_store_append(list_store, &iter);
    gtk_list_store_set (list_store, &iter,
                        LIST_VIEW_PIXMAP_COLUMN, pixbuf,
                        LIST_VIEW_NAME_COLUMN, os.str().c_str(),
                        -1);
  }
  // Associate the model the the gtk tree view
  gtk_tree_view_set_model(GTK_TREE_VIEW(view_list_treeview), GTK_TREE_MODEL(list_store));
  
  // Set transient now to take care of reparent operations on the ViewSpace
  gtk_window_set_transient_for(GTK_WINDOW(m_layoutEditorDialog),
                               GTK_WINDOW(gtk_widget_get_toplevel(m_parentTopWidget)));
  // Show dialog
  gtk_widget_show_all(m_layoutEditorDialog);
  gint response = gtk_dialog_run(GTK_DIALOG(m_layoutEditorDialog));  
  if (response == GTK_RESPONSE_OK){
    // Setting the p_page values
    for (int i = 0 ; i < (int)p_views.size(); i++) {
      if (p_views[i]->IsSelected()) {
        p_page.viewIds.push_back(i);
      }
    }
    unsigned int dimension = (unsigned int)ceil(sqrt(p_page.viewIds.size()));
    p_page.layout = ViewSpaceLayout::GetGridLayout(dimension, dimension);
  }
  
  gtk_widget_hide(m_layoutEditorDialog);

  return response;  
}


void 
NewPageDialog::OnLayoutDialogAdd()
{
}
void 
NewPageDialog::OnLayoutDialogAddCallback(GtkButton *p_button,
                                     gpointer p_page)
{
  static_cast<NewPageDialog*>(p_page)->OnLayoutDialogAdd();
}

void 
NewPageDialog::OnLayoutDialogRemove()
{
}
void 
NewPageDialog::OnLayoutDialogRemoveCallback(GtkButton *p_button,
                                        gpointer p_page)
{
  static_cast<NewPageDialog*>(p_page)->OnLayoutDialogRemove();
}

void 
NewPageDialog::OnLayoutComboChanged()
{
}
void 
NewPageDialog::OnLayoutComboChangedCallback(GtkCombo *p_combo,
                                            gpointer p_page)
{
  static_cast<NewPageDialog*>(p_page)->OnLayoutComboChanged();
}




} //namespace mips
