#include "mipsViewSpace.h"
#include "assert.h"
#include <iostream>
#include <sstream>

#include "mipsView.h"

namespace mips {

View::View ():m_name(""), m_topWidget(0), m_selected(FALSE),
              m_interactionStyle(SELECT_INTERACTION)
{
  this->Initialize();
}

View::~View ()
{
  g_object_unref(this->GetGtkWidget());
}

void
View::viewport_unrealize_cb (GtkWidget *widget,
			     gpointer user_data)
{
#if 1
	/*
	 * This is a workaround for a bug in the Gtk+ viewport.
	 */
	gtk_container_forall (GTK_CONTAINER (widget),
			      (GtkCallback) gtk_widget_unrealize,
			      NULL);
#endif
}

void
View::Initialize()
{
  // members
  m_currentPoint[0] = 0;  
  m_currentPoint[1] = 0;
  m_currentPoint[2] = 0;
  
  // Setting the Top Widget  
  m_topWidget = gtk_viewport_new (NULL, NULL);
  
  g_signal_connect (G_OBJECT (m_topWidget), "unrealize",
                    G_CALLBACK (viewport_unrealize_cb), NULL);
  gtk_widget_show (m_topWidget);
  gtk_container_set_border_width (GTK_CONTAINER (m_topWidget), 1);
  this->SetSize (100, 100);

  /* Add a reference to the GtkWidget embeded in the GtkRenderWindow
	 to preserve it to be finalize when it will be removed from a container */
  g_object_ref(this->GetGtkWidget());

  // Add an alignement
  m_alignment = gtk_alignment_new (0.5, 0.5, 1.0, 1.0);
  gtk_container_add (GTK_CONTAINER (m_topWidget), m_alignment);
  
  gtk_alignment_set_padding (GTK_ALIGNMENT (m_alignment), 2, 2, 2, 2);

  // Define the colors used for the border of the view
  m_unselectedColor = gdk_color_copy(gtk_widget_get_style(m_topWidget)->bg);
  GdkColor color = {
    0,
  };
  color.red = 0xffff;  
  color.green = 0;
  color.blue = 0;
  m_selectedColor = gdk_color_copy(&color);

  // Signal Connection
  g_signal_connect (G_OBJECT(m_topWidget), "button-press-event", 
                    G_CALLBACK (View::OnMouseButtonPressCbk), this);
 
}

GdkPixbuf* 
View::CreatePixbuf (unsigned int p_width, unsigned int p_height)
{
  if (!GTK_WIDGET_REALIZED (this->GetGtkWidget())) {
    return NULL;
  }  
  return gdk_pixbuf_scale_simple (this->GetPixbufFromWidget(this->GetGtkWidget()),
                                  p_width, p_height,GDK_INTERP_BILINEAR);
}

GdkPixbuf* 
View::GetPixbufFromWidget (GtkWidget* p_widget)
{
  assert (GTK_WIDGET_REALIZED (p_widget));
  gint width, height;
  gdk_drawable_get_size (p_widget->window,
                         &width, &height);
  return gdk_pixbuf_get_from_drawable (NULL, p_widget->window, NULL,
                                       0, 0, 0, 0,
                                       width, height);
}

void
View::AddGtkWidget (GtkWidget *widget)
{

  gtk_container_add (GTK_CONTAINER (m_alignment),
                     GTK_WIDGET (widget));  
}

GtkWidget* 
View::GetGtkWidget()
{
  return m_topWidget;
}

void
View::SetName(const char* p_name)
{
  m_name = p_name;
  this->SetUpLeftAnnotation(p_name);  
}

const char *
View::GetName()
{
  return m_name.c_str();  
}


void
View::SetSelected()
{
  if (!m_selected) {
	m_selected=true;
       gtk_widget_modify_bg (GTK_WIDGET (m_topWidget), GTK_STATE_NORMAL,
                             m_selectedColor);       
  }
}

void
View::SetUnselected()
{
  if (m_selected) {
	m_selected=false;
        gtk_widget_modify_bg (GTK_WIDGET (m_topWidget), GTK_STATE_NORMAL,
                              m_unselectedColor);            
  }
}

bool
View::IsSelected()
{
  return m_selected;
}

bool 
View::IsContaining(GtkWidget* p_widget)
{
  return (p_widget == m_topWidget);
}



void
View::SetUpLeftAnnotation(const char *p_annotation)
{
  m_upLeftAnnotation = p_annotation;  
  this->UpdateAnnotations();  
}
const char *
View::GetUpLeftAnnotation()
{
  return m_upLeftAnnotation.c_str();  
}
void
View::SetUpRightAnnotation(const char *p_annotation)
{
  m_upRightAnnotation = p_annotation;  
  this->UpdateAnnotations();  
}
const char *
View::GetUpRightAnnotation()
{
  return m_upRightAnnotation.c_str();  
}
void
View::SetNorthAnnotation(const char *p_annotation)
{
  m_northAnnotation = p_annotation;  
  this->UpdateAnnotations();  
}
const char *
View::GetNorthAnnotation()
{
  return m_northAnnotation.c_str();  
}
void
View::SetSouthAnnotation(const char *p_annotation)
{
  m_southAnnotation = p_annotation;  
  this->UpdateAnnotations();  
}
const char *
View::GetSouthAnnotation()
{
  return m_southAnnotation.c_str();  
}
void
View::SetEastAnnotation(const char *p_annotation)
{
  m_eastAnnotation = p_annotation;  
  this->UpdateAnnotations();  
}
const char *
View::GetEastAnnotation()
{
  return m_eastAnnotation.c_str();  
}
void
View::SetOuestAnnotation(const char *p_annotation)
{
  m_ouestAnnotation = p_annotation;
  this->UpdateAnnotations();  
}
const char *
View::GetOuestAnnotation()
{
  return m_ouestAnnotation.c_str();  
}


void
View::SetSize(unsigned int p_width,unsigned int p_height)
{
  gtk_widget_set_size_request (m_topWidget, p_width, p_height);
}

void
View::SetInteractionStyle(unsigned int p_style)
{
  m_interactionStyle = p_style;  
}

unsigned int
View::GetInteractionStyle() const
{
  return m_interactionStyle;  
}

void
View::SetCurrentPointFrom(const View* p_view)
{
  this->SetCurrentPoint(p_view->GetCurrentPoint());  
}

void
View::SetCurrentPoint(const double p_point[3])
{
  if (memcmp ( static_cast<const void*>(m_currentPoint),
               static_cast<const void*>(p_point),
               sizeof(m_currentPoint)) == 0)
  {    
    return;
  }
  
  memcpy(static_cast<      void*>(m_currentPoint),
         static_cast<const void*>(p_point),
         sizeof(m_currentPoint));
  this->UpdatePosition();
  
  // Send CurrentPoint signal to the observers
  void (*callback) (View*, void*);
  for (unsigned int i = 0; i < m_currentPointObservers.size() ; i++ ) {
    callback = (void (*) (View*, void*))
      (m_currentPointObservers[i].first);
   (*callback) (this, m_currentPointObservers[i].second);    
  }  
}

const double*
View::GetCurrentPoint() const
{
  return m_currentPoint;
}
void
View::GetCurrentPoint(double p_point[3]) const
{
  memcpy(static_cast<      void*>(p_point),
         static_cast<const void*>(m_currentPoint),
         sizeof(m_currentPoint));
  return;
}

const char*
View::GetCurrentPointStringValue() 
{
  static std::string  str;
  std::ostringstream os;
  os.setf(std::ios::fixed, std::ios::floatfield);
  os.precision(1);
  os << (this->GetCurrentPointDoubleValue());
  str = os.str();  
  return str.c_str();  
}


void
View::AddMouseButtonPressObserver(GCallback p_callback,  void *p_observer)
{
  m_mouseButtonPressObservers.push_back(
    std::pair<GCallback, void *>(p_callback,p_observer));  
}
void
View::AddFullPageObserver(GCallback p_callback,  void *p_observer)
{
  m_fullPageObservers.push_back(
    std::pair<GCallback, void *>(p_callback,p_observer));  
}
void
View::AddCurrentPointObserver(GCallback p_callback,  void *p_observer)
{
  m_currentPointObservers.push_back(
    std::pair<GCallback, void *>(p_callback,p_observer));  
}
void
View::AddWindowLevelObserver(GCallback p_callback,  void *p_observer)
{
  m_windowLevelObservers.push_back(
    std::pair<GCallback, void *>(p_callback,p_observer));  
}

// CALLBACKS
void
View::OnMouseButtonPress(GtkWidget *p_viewWidget,
                         GdkEventButton *p_event)
{
  void (*callback) (View*, GdkEventButton*, void*);
  
  // Forward the signal to the observers 
  for (unsigned int i = 0; i < m_mouseButtonPressObservers.size() ; i++ ) {
    callback = (void (*) (View*, GdkEventButton*, void*))
      (m_mouseButtonPressObservers[i].first);
   (*callback) (this,p_event, m_mouseButtonPressObservers[i].second);    
  }
}
gboolean 
View::OnMouseButtonPressCbk (GtkWidget *p_viewWidget,
                             GdkEventButton *p_event,
                             gpointer p_view)
{
  static_cast<View*>(p_view)->OnMouseButtonPress(p_viewWidget, p_event);
  return false;
}


} //namespace mips
