#ifdef _MSC_VER
#  pragma warning (disable : 4786)
#endif

#include "MIPS_GUIPKG_config.h"
#include "mipsGladeGui.h"
#include "mipsView.h"
#include "mipsDataSet.h"
#include "mipsViewSpaceLayout.h"
#include "assert.h"
#include <sstream>
#include <iostream>
#include <string>
#include <math.h>
#include <algorithm>

#include "mipsViewSpace.h"

namespace mips {

// CONSTRUCTORS / DESTRUCTOR
ViewSpace::ViewSpace ():
  m_previousPage(0), m_isFullPageOn(false), m_isOnWindowLevelChange(false),
  m_lastSelectedView(0)
{
  // Load the main viewspace interface
  m_gladeGui = new mips::GladeGui("viewspace.glade");
  
  // Building the top_widget from the glade main_window
  m_topWidget = gtk_viewport_new (NULL, NULL);
  m_gladeGui->Reparent("top_vbox", m_topWidget);
  m_table = m_gladeGui->GetWidget("main_table");

  // Building the Toolbar in the top_hbox
  GtkToolbar* toolbar = GTK_TOOLBAR(gtk_toolbar_new ());
  gtk_toolbar_set_orientation (GTK_TOOLBAR (toolbar), GTK_ORIENTATION_HORIZONTAL);
  gtk_toolbar_set_style (GTK_TOOLBAR (toolbar), GTK_TOOLBAR_ICONS);
  gtk_toolbar_set_show_arrow(GTK_TOOLBAR (toolbar),FALSE);  
  gtk_container_set_border_width (GTK_CONTAINER (toolbar), 0);
  GtkWidget* toolbar_viewport = m_gladeGui->GetWidget("toolbar_viewport");  
  gtk_container_add(GTK_CONTAINER (toolbar_viewport), GTK_WIDGET(toolbar));

  // Add toolbar buttons
  GSList      *radio_group;  
  GtkWidget   *iconw;
  GtkTooltips *toolbar_tips = gtk_tooltips_new ();
  gchar       *iconname;  
  
  m_selectorButton = gtk_radio_tool_button_new(NULL);  
  radio_group = gtk_radio_tool_button_get_group(
    GTK_RADIO_TOOL_BUTTON(m_selectorButton));
  gtk_toolbar_insert (toolbar, GTK_TOOL_ITEM (m_selectorButton),-1);
  g_signal_connect (G_OBJECT (m_selectorButton), "toggled",
		    G_CALLBACK (ViewSpace::OnToolbarButtonToggledCbk), this);
  iconname = g_strjoin(G_DIR_SEPARATOR_S,MIPS_GLADE_FILES_DIR,"selector.xpm",NULL);  
  iconw = gtk_image_new_from_file (iconname);
  g_free(iconname);  
  gtk_tool_button_set_icon_widget (GTK_TOOL_BUTTON(m_selectorButton), iconw);
  gtk_tooltips_set_tip (toolbar_tips, GTK_WIDGET(m_selectorButton),
                        "Select current point / slice", "");
  
  m_windowLevelButton = gtk_radio_tool_button_new(radio_group);  
  gtk_toolbar_insert (toolbar, GTK_TOOL_ITEM (m_windowLevelButton),-1);
  g_signal_connect (G_OBJECT (m_windowLevelButton), "toggled",
		    G_CALLBACK (ViewSpace::OnToolbarButtonToggledCbk), this);
  iconname = g_strjoin(G_DIR_SEPARATOR_S,MIPS_GLADE_FILES_DIR,"greyscale.xpm",NULL);  
  iconw = gtk_image_new_from_file (iconname);
  g_free(iconname);  
  gtk_tool_button_set_icon_widget (GTK_TOOL_BUTTON(m_windowLevelButton), iconw);
  gtk_tooltips_set_tip (toolbar_tips, GTK_WIDGET(m_windowLevelButton),
                        "Window Level", "");

  m_fullPageButton =
    gtk_radio_tool_button_new_from_widget(GTK_RADIO_TOOL_BUTTON(m_selectorButton));  
  gtk_toolbar_insert (toolbar, GTK_TOOL_ITEM (m_fullPageButton),-1);
  g_signal_connect (G_OBJECT (m_fullPageButton), "toggled",
		    G_CALLBACK (ViewSpace::OnToolbarButtonToggledCbk), this);
  iconw = gtk_image_new_from_stock (GTK_STOCK_ZOOM_FIT,GTK_ICON_SIZE_BUTTON); 
  gtk_tool_button_set_icon_widget (GTK_TOOL_BUTTON(m_fullPageButton), iconw);
  gtk_tooltips_set_tip (toolbar_tips, GTK_WIDGET(m_fullPageButton),
                        "Full Page", "");

  m_measureButton =
    gtk_radio_tool_button_new_from_widget(GTK_RADIO_TOOL_BUTTON(m_selectorButton));  
  gtk_toolbar_insert (toolbar, GTK_TOOL_ITEM (m_measureButton),-1);
  g_signal_connect (G_OBJECT (m_measureButton), "toggled",
		    G_CALLBACK (ViewSpace::OnToolbarButtonToggledCbk), this);
  iconname = g_strjoin(G_DIR_SEPARATOR_S,MIPS_GLADE_FILES_DIR,"hruler.xpm",NULL);  
  iconw = gtk_image_new_from_file (iconname);
  g_free(iconname);  
  gtk_tool_button_set_icon_widget (GTK_TOOL_BUTTON(m_measureButton), iconw);
  gtk_tooltips_set_tip (toolbar_tips, GTK_WIDGET(m_measureButton),
                        "Measure", "");
 
  m_zoomOutButton = 
    gtk_radio_tool_button_new_from_widget(GTK_RADIO_TOOL_BUTTON(m_selectorButton));  
  gtk_toolbar_insert (toolbar, GTK_TOOL_ITEM (m_zoomOutButton),-1);
  g_signal_connect (G_OBJECT (m_zoomOutButton), "toggled",
		    G_CALLBACK (ViewSpace::OnToolbarButtonToggledCbk), this);
  iconname = g_strjoin(G_DIR_SEPARATOR_S,MIPS_GLADE_FILES_DIR,"zoom.png",NULL);  
  iconw = gtk_image_new_from_file (iconname); 
  gtk_tool_button_set_icon_widget (GTK_TOOL_BUTTON(m_zoomOutButton), iconw);
  gtk_tooltips_set_tip (toolbar_tips, GTK_WIDGET(m_zoomOutButton),
                        "Zoom Out", "");


  this->SetInteractionTool(View::SELECT_INTERACTION);

  gtk_widget_show_all (m_topWidget);

  this->DisplayAllViews();
}

ViewSpace::~ViewSpace ()
{
  delete m_gladeGui;
}



// PUBLIC METHODS

GtkWidget* 
ViewSpace::GetTopWidget()
{
  return m_topWidget;
}

void
ViewSpace::DisplayOneView (View *p_view)
{
  if (!p_view) {
    return;
  }

  // Search the corresponding id of the view in the m_views array
  ViewId view_id = -1;  
  for (unsigned int i = 0; i < m_views.size(); i++) {
    if (m_views[i] == p_view) {
      view_id = i;
      break;
    }
  }

  // If we have found one view,
  // Create a ViewIdArray and a layout corresponding to our needs
  // Then we display
  if (view_id != -1) {    
    this->SelectOneView(p_view);    
    ViewIdArrayType view_id_array;
    view_id_array.push_back(view_id);
    this->DisplayTable(view_id_array,ViewSpaceLayout::GetGridLayout(1,1));
  }
  
}



void 
ViewSpace::DisplayAllViews ()
{
  // Take care from where we are comming from to manage the views selected
  if (m_views.size() && (m_previousPage != -1) ) {
    this->SelectOneView(m_views[0]);
  }
  m_previousPage = -1;

  // Initialize the list of views to display
  ViewIdArrayType view_id_array;
  for (unsigned int i=0; i<m_views.size(); i++) {
    view_id_array.push_back(i);
  }

  // Display
  if (m_views.size()) {
    unsigned int dimension = (unsigned int)ceil(sqrt( double(m_views.size()) ));
    this->DisplayTable(view_id_array,
                       ViewSpaceLayout::GetGridLayout(dimension, dimension));
  }
}

int 
ViewSpace::AddView (View * p_view)
{
  if (!p_view) {
    return -1;
  }  
  m_views.push_back(p_view);
  p_view->SetInteractionStyle(this->GetInteractionTool());
  p_view->AddMouseButtonPressObserver(G_CALLBACK(ViewSpace::OnViewMouseButtonPressCbk), this);
  p_view->AddFullPageObserver(G_CALLBACK(ViewSpace::OnFullPageEventCbk), this);
  p_view->AddCurrentPointObserver(G_CALLBACK(ViewSpace::OnCurrentPointChangeCbk), this);
  p_view->AddWindowLevelObserver(G_CALLBACK(ViewSpace::OnWindowLevelChangeCbk), this);

  return m_views.size()-1;  
}

void
ViewSpace::DisplayPage (unsigned int p_pageNumber)
{
  assert(p_pageNumber < m_pages.size());
  
  // Take care from where we are comming from to manage the views selected
  // If we are comming from an other page, select the first non empty view
  if (m_pages.size() && (m_previousPage != (int)p_pageNumber) ) {
    for (unsigned int i = 0; i < m_pages[p_pageNumber].viewIds.size() ; i++) {
      if (m_pages[p_pageNumber].viewIds[i] != -1) {        
        this->SelectOneView(m_views[m_pages[p_pageNumber].viewIds[0]]);
      }
    }    
  }
  m_previousPage = p_pageNumber;
  
  // Display the page selected in the table
  this->DisplayTable(m_pages[p_pageNumber].viewIds,
                     m_pages[p_pageNumber].layout);
}


void 
ViewSpace::AddPage (const Page* p_page)
{
  if ((!p_page)||(p_page->viewIds.size()==0)|| (p_page->layout.GetNbCells()==0)) return;  
  
  // Copy the Page in a new page
  m_pages.push_back(*p_page);
}

void 
ViewSpace::SelectOneView(View* p_view){
  // Unselect all views 
  for (unsigned int i = 0 ; i < m_views.size(); i++) {
    m_views[i]->SetUnselected();
  }
  // Select the one to be selected
  p_view->SetSelected(); 
}

// DATASET PAGE PUBLIC METHODS

void
ViewSpace::AddDataSetPage (DataSet* p_dataSet)
{
  // Create an empty Page and associated its id to the dataset
  m_pages.push_back(Page());  
  m_dataSetPageIds[p_dataSet] = m_pages.size() -1;
}

void
ViewSpace::DisplayDataSetPage (DataSet* p_dataSet)
{
  // A page have to be defined for that DataSet
  if (m_dataSetPageIds.find(p_dataSet) == m_dataSetPageIds.end()) return;

  // Update views
  ViewIdArrayType view_ids = m_pages[m_dataSetPageIds[p_dataSet]].viewIds;
  for (unsigned int i=0 ; i < view_ids.size() ; i++) {
    if (view_ids[i] != -1) {      
      m_views[view_ids[i]]->Update();
    }    
  }
  
  // Display the page
  this->DisplayPage(m_dataSetPageIds[p_dataSet]);
}

void
ViewSpace::SetDataSetPageLayout(DataSet* p_dataSet, ViewSpaceLayout p_layout)
{
  // A page have to be defined for that DataSet
  if (m_dataSetPageIds.find(p_dataSet) == m_dataSetPageIds.end()) return;

  unsigned int page_id = m_dataSetPageIds[p_dataSet];
  assert ((page_id >=0) && (page_id <m_pages.size()));

  m_pages[page_id].layout = p_layout;  
}

void
ViewSpace::AddViewInDataSetPage(DataSet* p_dataSet, View * p_view, int p_position)
{
  // A page have to be defined for that DataSet
  if (m_dataSetPageIds.find(p_dataSet) == m_dataSetPageIds.end()) return;

  unsigned int page_id = m_dataSetPageIds[p_dataSet];
  assert ((page_id >=0) && (page_id <m_pages.size()));

  // test special values of p_position
  if (p_position < -1 ) {
    return;
  }  
  if ((p_position == -1 ) ){
    p_position = m_pages[page_id].viewIds.size();    
    m_pages[page_id].viewIds.resize(p_position+1);
  }
  
  // Add empty(-1) layout cells if necessary
  if (p_position >= (int)(m_pages[page_id].viewIds.size())) {
    unsigned int old_size = m_pages[page_id].viewIds.size();    
    m_pages[page_id].viewIds.resize(p_position+1);
    for (unsigned int i = old_size ; i<m_pages[page_id].viewIds.size() -1 ; i++) {
      m_pages[page_id].viewIds[i] = -1;      
    }    
  }

  // Add the view in the global list and set the correct id in the page list
  m_pages[page_id].viewIds[p_position] = this->AddView(p_view);  
}



// PRIVATE METHODS

void 
ViewSpace::DisplayTable (ViewIdArrayType& p_viewIdArray, ViewSpaceLayout& p_layout)
{
  if ((p_viewIdArray.size() == 0) || (p_layout.GetNbCells() == 0)) return;
  
  // Create a list of GtkWidgets
  std::vector<GtkWidget*> view_widgets_array;
  for (unsigned int i=0; i<p_viewIdArray.size(); i++) {
    assert ((p_viewIdArray[i] >= -1) && (p_viewIdArray[i] <(int) m_views.size()));    
    if ((p_viewIdArray[i] != -1) && (m_views[p_viewIdArray[i]])) {      
      view_widgets_array.push_back(m_views[p_viewIdArray[i]]->GetGtkWidget());
    }
    else {
      view_widgets_array.push_back(0);
    }    
  }
  this->DisplayWidgetsOnTable (GTK_TABLE(m_table), view_widgets_array, p_layout);
}

void 
ViewSpace::DisplayWidgetsOnTable (GtkTable* p_table, 
                                  std::vector<GtkWidget*> p_widgets,
                                  ViewSpaceLayout& p_layout)
{
  gtk_widget_hide_all(GTK_WIDGET(p_table));

  // Detach all widgets from the table
  GList* list_element = gtk_container_get_children(GTK_CONTAINER(p_table));
  while(list_element) {
    gtk_container_remove(GTK_CONTAINER(p_table), GTK_WIDGET(list_element->data));
    list_element = g_list_next(list_element);
  }

  // Define and set the table size
  unsigned int left=0, right=0 ,top=0, bottom=0;    
  unsigned int columns = 0;
  unsigned int rows = 0;
  unsigned int i = 0;
  for (i=0; i<p_layout.GetNbCells(); i++) {
    p_layout.GetCellAttachement(i, left, right, top, bottom);
    if (right  > columns) columns = right;
    if (bottom > rows)    rows    = bottom;
  }
  gtk_table_resize(p_table, columns, rows);

  // Attach views to the table
  for (i=0; (i<p_layout.GetNbCells()) && (i<p_widgets.size()); i++) {
    if (p_widgets[i]) {      
      p_layout.GetCellAttachement(i, left, right, top, bottom);      
      gtk_table_attach_defaults (p_table,
                                 p_widgets[i],
                                 left,
                                 right,
                                 top,
                                 bottom);
    }    
  }

  gtk_widget_show_all(GTK_WIDGET(p_table));
  
}

unsigned int
ViewSpace::GetInteractionTool()
{
  if (gtk_toggle_tool_button_get_active(
        GTK_TOGGLE_TOOL_BUTTON(m_selectorButton))){
    return View::SELECT_INTERACTION;    
  }
  if (gtk_toggle_tool_button_get_active(
        GTK_TOGGLE_TOOL_BUTTON(m_windowLevelButton))){
    return View::WINDOW_LEVEL_INTERACTION;    
  }
  if (gtk_toggle_tool_button_get_active(
        GTK_TOGGLE_TOOL_BUTTON(m_fullPageButton))){
    return View::FULL_PAGE_INTERACTION;    
  }
  if (gtk_toggle_tool_button_get_active(
        GTK_TOGGLE_TOOL_BUTTON(m_measureButton))){
    return View::MEASURE_INTERACTION;    
  }
  if (gtk_toggle_tool_button_get_active(
        GTK_TOGGLE_TOOL_BUTTON(m_zoomOutButton))){
    return View::ZOOM_INTERACTION;    
  }
  return View::NO_INTERACTION;  
}

void
ViewSpace::SetInteractionTool(unsigned int p_tool)
{
  GtkToolItem* button = 0;  
  switch (p_tool) {
  case View::SELECT_INTERACTION:
    button = m_selectorButton;    
    break;
  case View::WINDOW_LEVEL_INTERACTION:
    button = m_windowLevelButton;    
    break;
  case View::FULL_PAGE_INTERACTION:
    button = m_fullPageButton;    
    break;
  case View::MEASURE_INTERACTION:
    button = m_measureButton;    
    break;
  case View::ZOOM_INTERACTION:
    button = m_zoomOutButton;    
    break;
  }

  if (!gtk_toggle_tool_button_get_active (GTK_TOGGLE_TOOL_BUTTON(button))) {
    gtk_toggle_tool_button_set_active(GTK_TOGGLE_TOOL_BUTTON(button), TRUE);
    // We do nothing more because activating the button will emit a signal
    // And finally we will be back to really do the job.
    return;    
  }
}

void
ViewSpace::SetViewsInteractionStyle(unsigned int p_tool)
{
  for (unsigned int i=0; i < m_views.size(); i++){
    m_views[i]->SetInteractionStyle((View::InteractionStyleIds)p_tool);
  }    
}

void
ViewSpace::SetViewGroup(View* p_view, int p_groupId)
{
  if (p_groupId >= (int)m_groups.size()){
    return;
  }
 
  this->RemoveViewFromGroup(p_view);  
  if (p_groupId == -1) {
    return;    
  }
  m_groups[p_groupId].push_back(p_view);  
}

void
ViewSpace::RemoveViewFromGroup(View* p_view)
{
  int old_group = GetViewGroup(p_view);
  if (old_group == -1) {
    return;
  }

  ViewArrayType::iterator it = std::find(m_groups[old_group].begin(),
                                         m_groups[old_group].end(),
                                         p_view);
  m_groups[old_group].erase(it);  
}

int
ViewSpace::GetViewGroup(View* p_view)
{
  int result = -1;  
  for (unsigned int g_id = 0 ; g_id < m_groups.size() ; g_id++) {
    for (unsigned int i = 0 ; i < m_groups[g_id].size() ; i++) {
      if (p_view == m_groups[g_id][i]) {
        result = g_id;
        break;        
      }      
    }    
  }
  return result;  
}

unsigned int
ViewSpace::CreateViewGroup ()
{
  m_groups.push_back(ViewArrayType());
  return m_groups.size()-1;  
}


void
ViewSpace::RefreshPositionDisplay()
{
  GtkWidget* label;
  std::ostringstream os;  
  double pos[3];
  m_lastSelectedView->GetCurrentPoint(pos);

  os.setf(std::ios::fixed, std::ios::floatfield);
  os.precision(2);

  label = m_gladeGui->GetWidget("x_position_label");
  os.str("");  
  os << "x: ";
  os<< pos[0];  
  gtk_label_set_text (GTK_LABEL(label), os.str().c_str());
  
  label = m_gladeGui->GetWidget("y_position_label");
  os.str("");  
  os << "y: " << pos[1];  
  gtk_label_set_text (GTK_LABEL(label), os.str().c_str());
  
  label = m_gladeGui->GetWidget("z_position_label");
  os.str("");  
  os << "z: " << pos[2];  
  gtk_label_set_text (GTK_LABEL(label), os.str().c_str());

  label = m_gladeGui->GetWidget("voxel_value_label");
  os.str("");  
  os << "value: " << m_lastSelectedView->GetCurrentPointStringValue();  
  gtk_label_set_text (GTK_LABEL(label), os.str().c_str());

}


//
// CALLBACKS
//
void 
ViewSpace::OnToolbarButtonToggled(GtkToggleToolButton *p_toggleToolButton)
{
  // That signal is catch both for the activated button and for the disactivated one.
  if (!gtk_toggle_tool_button_get_active(p_toggleToolButton))
    return;
  
  if (GTK_TOOL_ITEM(p_toggleToolButton) == m_selectorButton) {
    this->SetViewsInteractionStyle(View::SELECT_INTERACTION);    
  }
  if (GTK_TOOL_ITEM(p_toggleToolButton) == m_windowLevelButton) {
    this->SetViewsInteractionStyle(View::WINDOW_LEVEL_INTERACTION);    
  }
  if (GTK_TOOL_ITEM(p_toggleToolButton) == m_fullPageButton) {
    this->SetViewsInteractionStyle(View::FULL_PAGE_INTERACTION);    
  }
  if (GTK_TOOL_ITEM(p_toggleToolButton) == m_measureButton) {
    this->SetViewsInteractionStyle(View::MEASURE_INTERACTION);    
  }
  if (GTK_TOOL_ITEM(p_toggleToolButton) == m_zoomOutButton) {
    this->SetViewsInteractionStyle(View::ZOOM_INTERACTION);    
  }
}
void 
ViewSpace::OnToolbarButtonToggledCbk(GtkToggleToolButton *p_toggleToolButton,
                                     gpointer p_viewSpace)
{
  static_cast<ViewSpace*>(p_viewSpace)->OnToolbarButtonToggled(p_toggleToolButton);
}


// CALLBACKS FROM THE VIEWS
void 
ViewSpace::OnViewMouseButtonPress(View* p_view, GdkEventButton *p_event) {
  assert(p_view != 0);
  
  // Select the view
  unsigned int ctrl = (p_event->state & GDK_CONTROL_MASK);
  unsigned int shift = (p_event->state & GDK_SHIFT_MASK);
  if ((shift) || (ctrl)) {
    // Select the one to be selected in addition to the views previously selected
    // Do the same than ctrl for the moment...
    p_view->SetSelected();
  }
  else {
    this->SelectOneView(p_view);
  }
  m_lastSelectedView = p_view;
  this->RefreshPositionDisplay();  
}
void 
ViewSpace::OnViewMouseButtonPressCbk (View* p_view, GdkEventButton *p_event,
                                      void * p_viewSpace)
{
   static_cast<ViewSpace*>(p_viewSpace)->OnViewMouseButtonPress(p_view, p_event);
}

void
ViewSpace::OnFullPageEvent(View* p_view)
{
  if (m_isFullPageOn) {
    m_isFullPageOn=false;
    this->DisplayPage(m_previousPage);    
  }
  else {    
    m_isFullPageOn=true;
    this->DisplayOneView(p_view);
  }
}
void
ViewSpace::OnFullPageEventCbk (View* p_view, void * p_viewSpace)
{
  static_cast<ViewSpace*>(p_viewSpace)->OnFullPageEvent(p_view);  
}

void
ViewSpace::OnCurrentPointChange    (View* p_view)
{
  if ((p_view == 0) || (p_view != m_lastSelectedView)) {
    return;
  }

  this->RefreshPositionDisplay();

  int g_id = GetViewGroup(p_view);
  if (g_id == -1) {
    return;
  }
  for (unsigned int i = 0 ; i < m_groups[g_id].size() ; i++) {
    if (m_groups[g_id][i] != p_view) {      
      m_groups[g_id][i]->SetCurrentPointFrom(p_view);
      m_groups[g_id][i]->Render();      
    }    
  }   
}
void
ViewSpace::OnCurrentPointChangeCbk (View* p_view, void * p_viewSpace)
{
  static_cast<ViewSpace*>(p_viewSpace)->OnCurrentPointChange(p_view);  
}

void
ViewSpace::OnWindowLevelChange    (View* p_view)
{
  // to avoid infinite loop on changing color that change color...
  if (m_isOnWindowLevelChange) {
    return;
  }
  m_isOnWindowLevelChange = true;
  
  int g_id = GetViewGroup(p_view);
  if (g_id == -1) {
    return;
  }
  for (unsigned int i = 0 ; i < m_groups[g_id].size() ; i++) {
    if (m_groups[g_id][i] != p_view) {      
      m_groups[g_id][i]->SetWindowLevelFrom(p_view);
      m_groups[g_id][i]->Render();      
    }    
  }  
  m_isOnWindowLevelChange = false;
}
void
ViewSpace::OnWindowLevelChangeCbk (View* p_view, void * p_viewSpace)
{
  static_cast<ViewSpace*>(p_viewSpace)->OnWindowLevelChange(p_view);  
}

} //namespace mips
