#include <iostream>
#include "mipsViewSpaceLayout.h"

namespace mips {
  ViewSpaceLayout ViewSpaceLayout::m_staticLayout = ViewSpaceLayout();  

// Clear and reconstruct the given layout to a square homogeneous layout 
ViewSpaceLayout&
ViewSpaceLayout::GetGridLayout(unsigned int p_col,
                               unsigned int p_row)
{
  m_staticLayout.m_cells.clear();
  for (unsigned int row=0; row < p_row; row++) {
    for(unsigned int col=0; col < p_col; col++){
      m_staticLayout.m_cells.push_back(Attachment(col,col+1,row,row+1));
    }
  }
  return m_staticLayout;  
}

void
ViewSpaceLayout::PrintSelf()
{
  unsigned int left=0, right=0 ,top=0, bottom=0;    
  std::cout<<"+++++++++++++++++++++++++++++++++++++++++\n";
  std::cout<<"ViewSpaceLayout::PrintSelf() -> "<<this<<std::endl;
  
  for (unsigned int i=0; i < m_cells.size(); i++){
    this->GetCellAttachement(i, left, right, top, bottom);
    std::cout<<"GetCellAttachement("<<i<<") left="<<left<<", right="<<right<<", top"<<top<<", bottom"<<bottom<<"\n";    
  }  
  std::cout<<"-----------------------------------------\n";
}

void
ViewSpaceLayout::GetCellAttachement(unsigned int p_cell,
                                    unsigned int& p_left,
                                    unsigned int& p_right,
                                    unsigned int& p_top,
                                    unsigned int& p_bottom)
{
  if (p_cell >= m_cells.size()){
    p_left   = 0;
    p_right  = 0;
    p_top    = 0;
    p_bottom = 0;
    return;
  }
  
  p_left   = m_cells[p_cell].left;  
  p_right  = m_cells[p_cell].right;  
  p_top    = m_cells[p_cell].top;  
  p_bottom = m_cells[p_cell].bottom;
  return;  
}

unsigned int
ViewSpaceLayout::GetNbCells() const
{
  return m_cells.size();  
}

// SUB_CLASS 

ViewSpaceLayout::Attachment::Attachment()
{
  left=0; right=0; top=0; bottom=0;
}

ViewSpaceLayout::Attachment::Attachment(unsigned int p_left, unsigned int p_right,
                                        unsigned int p_top,  unsigned int p_bottom)
{
  left=p_left; right=p_right; top=p_top; bottom=p_bottom;
}

} //namespace mips
