#ifndef MIPS_VIEWSPACE_LAYOUT_H
#define MIPS_VIEWSPACE_LAYOUT_H

#include <vector>

namespace mips {

/** \class ViewSpaceLayout
    \ingroup MipsGUI
*/
class ViewSpaceLayout
{
 public:
  /** Build a layout as an honogeneous grid. The returned layout is shared for
      all the static ViewSpace methods. So, the returned layout have to be
      copied before the next call of one of those methods.  */
  static ViewSpaceLayout& GetGridLayout(unsigned int p_col,unsigned int p_row);

  /** Get how a cell of the layout have to be attached on the viewspace grid. */
  void GetCellAttachement(unsigned int p_cell,
                          unsigned int& p_left, unsigned int& p_right,
                          unsigned int& p_top,  unsigned int& p_bottom);

  /** Returns the number of cells that compose the layout. */
  unsigned int GetNbCells() const;  

  /** Prints the layout values under the std output */
  void PrintSelf();  

  private:
  
  // SUB-CLASSE
  /**An Attachement is the description of how a view is attached to a 
     GtkTableWidget */
  class Attachment
    {
    public:
      unsigned int left;
      unsigned int right;
      unsigned int top;
      unsigned int bottom;
      Attachment();
      Attachment(unsigned int p_left, unsigned int p_right,
                 unsigned int p_top,  unsigned int p_bottom);
    };

  // ATTRIBUTES
 private:
  std::vector<Attachment> m_cells;
  static ViewSpaceLayout  m_staticLayout;  
    
  
};

} //namespace mips

#endif /* MIPS_VIEWSPACE_LAYOUT_H */
