#include "mipsVtkView.h"
#include "GtkRenderWindowInteractor.h"
#include "GtkRenderWindow.h"
#include "vtkInteractorObserver.h"
#include "vtkInteractorStyleSwitch.h"
#include "vtkRenderer.h"
#include "vtkRendererCollection.h"
#include "vtkProp.h"
#include "vtkTextActor.h"
#include "vtkCoordinate.h"
#include "vtkProperty2D.h"
#include "vtkTextProperty.h"
#include <gtk/gtk.h>
#include "assert.h"

namespace mips {

VtkView::VtkView ():View()
{
  this->Initialize (NULL);
}

VtkView::VtkView(vtkRendererCollection* p_rendererCollection):View()
{
  this->Initialize(p_rendererCollection);
}

VtkView::~VtkView()
{
  m_renderWindow->Delete();
  m_interactorStyle->Delete();
  m_renderWindow->Delete();  
  m_upLeftAnnotationActor->Delete();
  m_upRightAnnotationActor->Delete();
  m_northAnnotationActor->Delete();
  m_southAnnotationActor->Delete();
  m_eastAnnotationActor->Delete();
  m_ouestAnnotationActor->Delete();
};

vtkRendererCollection*
VtkView::CreateDefaultRendererCollection()
{
  vtkRenderer *ren = vtkRenderer::New();
  vtkRendererCollection *ren_collection = vtkRendererCollection::New ();
  ren->SetBackground(0.9,0.9,0.9);
  ren_collection->AddItem (ren);
  return ren_collection;
}

void
VtkView::Initialize(vtkRendererCollection* p_rendererCollection)
{
  m_interactionOn=true;
  m_interactorStyle= vtkInteractorStyleSwitch::New();

  // Set the m_renderWindow and attach it to the top widget
  m_renderWindow= GtkRenderWindow::New();
  GtkRenderWindowInteractor *iren=0;
  vtkRendererCollection* ren_collection = p_rendererCollection;
  
  iren = GtkRenderWindowInteractor::New();
  iren->SetRenderWindow(m_renderWindow);
  iren->Initialize ();

  if(!ren_collection) {
    ren_collection = this->CreateDefaultRendererCollection();
  }
  ren_collection->InitTraversal();
  for (vtkRenderer* ren = ren_collection->GetNextItem();
       ren != NULL;
       ren = ren_collection->GetNextItem()) {
    m_renderWindow->AddRenderer(ren);
  }

  this->AddGtkWidget (GTK_WIDGET (m_renderWindow->GetGenericWindowId ()));
  
  // VtkView selection initialisation
  this->SetUnselected();

  // Initilize Annotation stuff
  m_upLeftAnnotationActor = vtkTextActor::New();
  m_upLeftAnnotationActor->GetProperty()->SetColor(0,1,0);
  m_upLeftAnnotationActor->GetTextProperty()->SetJustificationToLeft();
  m_upLeftAnnotationActor->GetTextProperty()->SetVerticalJustificationToTop();
  m_upLeftAnnotationActor->GetPositionCoordinate()->SetCoordinateSystem(VTK_NORMALIZED_VIEWPORT);
  m_upLeftAnnotationActor->SetPosition(0,1);
  this->AddActor(m_upLeftAnnotationActor);
  m_upRightAnnotationActor = vtkTextActor::New();
  m_upRightAnnotationActor->GetProperty()->SetColor(0,1,0);
  m_upRightAnnotationActor->GetTextProperty()->SetJustificationToRight();
  m_upRightAnnotationActor->GetTextProperty()->SetVerticalJustificationToTop();
  m_upRightAnnotationActor->GetPositionCoordinate()->SetCoordinateSystem(VTK_NORMALIZED_VIEWPORT);
  m_upRightAnnotationActor->SetPosition(1,1);
  this->AddActor(m_upRightAnnotationActor);
  m_northAnnotationActor = vtkTextActor::New();
  m_northAnnotationActor->GetProperty()->SetColor(0,1,0);
  m_northAnnotationActor->GetTextProperty()->SetJustificationToCentered();
  m_northAnnotationActor->GetTextProperty()->SetVerticalJustificationToTop();
  m_northAnnotationActor->GetPositionCoordinate()->SetCoordinateSystem(VTK_NORMALIZED_VIEWPORT);
  m_northAnnotationActor->SetPosition(0.5,1);
  this->AddActor(m_northAnnotationActor);
  m_southAnnotationActor = vtkTextActor::New();
  m_southAnnotationActor->GetProperty()->SetColor(0,1,0);
  m_southAnnotationActor->GetTextProperty()->SetJustificationToCentered();
  m_southAnnotationActor->GetTextProperty()->SetVerticalJustificationToBottom();
  m_southAnnotationActor->GetPositionCoordinate()->SetCoordinateSystem(VTK_NORMALIZED_VIEWPORT);
  m_southAnnotationActor->SetPosition(0.5,0);
  this->AddActor(m_southAnnotationActor);
  m_eastAnnotationActor = vtkTextActor::New();
  m_eastAnnotationActor->GetProperty()->SetColor(0,1,0);
  m_eastAnnotationActor->GetTextProperty()->SetJustificationToLeft();
  m_eastAnnotationActor->GetTextProperty()->SetVerticalJustificationToCentered();
  m_eastAnnotationActor->GetPositionCoordinate()->SetCoordinateSystem(VTK_NORMALIZED_VIEWPORT);
  m_eastAnnotationActor->SetPosition(0,0.5);
  this->AddActor(m_eastAnnotationActor);
  m_ouestAnnotationActor = vtkTextActor::New();
  m_ouestAnnotationActor->GetProperty()->SetColor(0,1,0);
  m_ouestAnnotationActor->GetTextProperty()->SetJustificationToRight();
  m_ouestAnnotationActor->GetTextProperty()->SetVerticalJustificationToCentered();
  m_ouestAnnotationActor->GetPositionCoordinate()->SetCoordinateSystem(VTK_NORMALIZED_VIEWPORT);
  m_ouestAnnotationActor->SetPosition(1,0.5);
  this->AddActor(m_ouestAnnotationActor);
}

void 
VtkView::Render ()
{
  gtk_widget_queue_draw(GTK_WIDGET (m_renderWindow->GetGenericWindowId ()));
}

void
VtkView::Update()
{
}

VtkView*
VtkView::CreateClone ()
{
  return new VtkView(m_renderWindow->GetRenderers());
}

GdkPixbuf* 
VtkView::CreatePixbuf (unsigned int p_width, unsigned int p_height)
{
  GdkPixmap* pixmap=m_renderWindow->GetPixmap (p_width,p_height);
  return gdk_pixbuf_get_from_drawable (NULL,
                                       pixmap, NULL,
                                       0, 0, 0, 0,
                                       p_width, p_height);
}

vtkRenderWindow* 
VtkView::GetVtkRenderWindow ()
{
  return m_renderWindow;
}

bool 
VtkView::IsContaining(GtkWidget* p_widget)
{
  return (this->View::IsContaining (p_widget) || 
		  (p_widget == m_renderWindow->GetGenericWindowId ()));
}

void 
VtkView::AddActor (vtkProp* actor)
{
  m_renderWindow->GetRenderers()->InitTraversal();
  vtkRenderer* first_renderer = m_renderWindow->GetRenderers()->GetNextItem();
  if (first_renderer) {
	first_renderer->AddActor(actor);
  }
}


void 
VtkView::RemoveActor (vtkProp* actor)
{
  m_renderWindow->GetRenderers()->InitTraversal();
  vtkRenderer* first_renderer = m_renderWindow->GetRenderers()->GetNextItem();
  if (first_renderer) {
	first_renderer->RemoveActor(actor);
  }
}

void
VtkView::SetInteractionOff()
{
  m_interactionOn = false;
  this->SetInteraction();  
}

void
VtkView::SetInteractionOn()
{
  m_interactionOn = true;
  this->SetInteraction();  
}

void
VtkView::SetInteraction()
{
  //cout<<"VtkView::SetInteraction()\n";
  
  if (!m_interactionOn) {
    m_interactorStyle = dynamic_cast<vtkInteractorStyle*>(this->GetVtkRenderWindow()->GetInteractor()->GetInteractorStyle());
    this->GetVtkRenderWindow()->GetInteractor()->SetInteractorStyle(NULL);
  }
  else {
    this->GetVtkRenderWindow()->GetInteractor()->SetInteractorStyle(m_interactorStyle);
  }
}

void
VtkView::SetInteractorStyle(vtkInteractorStyle* style)
{
  if(m_interactorStyle)
    m_interactorStyle->Delete();
  
  m_interactorStyle = style;
  this->SetInteraction();
}


void
VtkView::UpdateAnnotations()
{
  m_upLeftAnnotationActor->SetInput(this->GetUpLeftAnnotation());
  m_upRightAnnotationActor->SetInput(this->GetUpRightAnnotation());
  m_northAnnotationActor->SetInput(this->GetNorthAnnotation());
  m_southAnnotationActor->SetInput(this->GetSouthAnnotation());
  m_eastAnnotationActor->SetInput(this->GetEastAnnotation());
  m_ouestAnnotationActor->SetInput(this->GetOuestAnnotation());
}


void
VtkView::SetBackgroundColor(float r, float g, float b)
{
  vtkRendererCollection* ren_collec = m_renderWindow->GetRenderers();
  if(ren_collec == NULL)
    return;
  
  ren_collec->InitTraversal();
  for (vtkRenderer* ren = ren_collec->GetNextItem();
       ren != NULL;
       ren = ren_collec->GetNextItem()) 
    {
      ren->SetBackground(r,g,b);
    }
}

} //namespace mips
