#ifndef MIPS_VTKVIEW_H
#define MIPS_VTKVIEW_H

#include "mipsView.h"
#include "GtkRenderWindow.h"

class vtkProp;
class vtkRendererCollection;
class vtkInteractorStyle;
class vtkTextActor;

namespace mips {

/** \class VtkView
    \ingroup MipsGUI
*/
class VtkView : public View
{
 public:
  /** Create a VtkView.  */
  VtkView();
  VtkView(vtkRendererCollection* p_rendererCollection);

   /** Destroy the View and unreference the GtkWidget.  */
  virtual ~VtkView();

  /** Send a render message to the vtkRenderWindow associated */
  virtual void Render();

  /** Update the the display parameters based on the data information. */
  virtual void Update();

  /** Render OffScreen and return a GdkPixbuf containing the result. */
  virtual GdkPixbuf*  CreatePixbuf (unsigned int p_width, unsigned int p_height); 

  /** Create a new VtkView instance using the same vtkRendererCollection.
    example code :
        VtkView model_view();
        VtkView* clone = model_view.CreateClone();
        clone->SetInteractionOff();    
   */
  virtual VtkView* CreateClone ();

  /** Get the vtkRenderWindow associated */
  virtual vtkRenderWindow* GetVtkRenderWindow();

  /** Add the actor to the first renderer of the render window if exist. 
      Do nothing otherwise.*/
  virtual void AddActor(vtkProp* actor);

  /** remove the actor from the first renderer of the render window if exist. 
      Do nothing otherwise.*/
  virtual void RemoveActor(vtkProp* actor);

  /** Return true if the view contains the given widget in it's widget tree. */
  virtual bool IsContaining(GtkWidget* p_widget);

 /** Enable or Disable interaction on the view. The Interaction mode is store
      internaly and set at each time the widget is showed. The interaction
      mode cannot be set before the GtkRenderWindowInteractor is initialized. */
  virtual void SetInteractionOff();
  virtual void SetInteractionOn();
  virtual void SetInteractorStyle(vtkInteractorStyle*);

  /** Set the background color. Format is RGB, 0 <= R,G,B <=1
      Example: SetBackgroundColor(0.9,0.9,0.9) for grey-white. */
  virtual void SetBackgroundColor(float r, float g, float b);
    
 private: 

  GtkRenderWindow        *m_renderWindow;
  bool                    m_interactionOn;
  vtkInteractorStyle     *m_interactorStyle;
  vtkTextActor           *m_upLeftAnnotationActor;
  vtkTextActor           *m_upRightAnnotationActor;
  vtkTextActor           *m_northAnnotationActor;
  vtkTextActor           *m_southAnnotationActor;
  vtkTextActor           *m_eastAnnotationActor;
  vtkTextActor           *m_ouestAnnotationActor;

 private:
  /** Create a collection containing one renderer */
  virtual vtkRendererCollection* CreateDefaultRendererCollection();

  /** Class initialisation. If renderer is given it is used, othewise it is 
      created. */
  void Initialize(vtkRendererCollection* p_renderer=0);
  
  virtual void SetInteraction();

  /** Overriden Methods*/
  virtual void UpdateAnnotations();  

};

} //namespace mips

#endif /* MIPS_VTKVIEW3D_H */
