#ifdef _MSC_VER
#  pragma warning (disable : 4018) 
#endif

#include "mipsVtkViewImage2D.h"
#include "mipsInteractorStyleImage2D.h"
#include "vtkAbstractPropPicker.h"
#include "vtkRenderWindowInteractor.h"
#include "vtkImageData.h"
#include <string.h> 

#include "mipsVtkViewImage2DCommand.h"

namespace mips
{

VtkViewImage2DCommand::VtkViewImage2DCommand() :
  m_view(NULL), m_initialWindow(0), m_initialLevel(0),
  m_windowEventStatus(false),  m_signalEmitTest(NULL)
{
}

void
VtkViewImage2DCommand::Execute(vtkObject*    caller, 
                               unsigned long event, 
                               void*         callData)
{
  if (m_view->GetInput() == NULL)
    return;
     
  if (event == vtkCommand::KeyPressEvent)
  {    
    vtkRenderWindowInteractor *rwi = m_view->GetVtkRenderWindow()->GetInteractor();
    std::string key_sym(rwi->GetKeySym());
    
    if ((key_sym.compare("i")==0) || (key_sym.compare("I")==0))
    {
      int mode;      
      mode = m_view->GetInterpolationMode();
      m_view->SetInterpolationMode((++mode)%2);
      m_view->Render();
      return;      
    }
    
    return;
  }

  if (event == vtkCommand::StartWindowLevelEvent) {
    this->StartWindowing();
    return;    
  }
  if (event == vtkCommand::EndWindowLevelEvent) {
    this->EndWindowing();
    return;    
  }
  if (event == vtkCommand::WindowLevelEvent) {
    this->Windowing(static_cast<vtkInteractorStyleImage *>(caller));
    return;    
  }
  if (event == vtkCommand::ResetWindowLevelEvent) {
    m_view->ResetWindowLevel();
    return;    
  }
  if (event == vtkCommand::StartPickEvent) {
    this->StartPicking();
    return;    
  }
  if (event == vtkCommand::EndPickEvent) {
    this->EndPicking(static_cast<vtkInteractorStyleImage *>(caller));
    return;    
  }
  if (event == VtkViewImage2DCommand::StartMeasureEvent) {
    std::cout<<"StartMeasureEvent\n";    
    return;    
  }
  if (event == VtkViewImage2DCommand::MeasureEvent) {
    std::cout<<"MeasureEvent\n";    
    return;    
  }
  if (event == VtkViewImage2DCommand::EndMeasureEvent) {
    std::cout<<"EndMeasureEvent\n";    
    return;    
  }
  if (event == VtkViewImage2DCommand::StartZSliceMoveEvent) {
    return;    
  }
  if (event == VtkViewImage2DCommand::ZSliceMoveEvent) {
    this->ChangeZSlice(static_cast<InteractorStyleImage2D *>(caller)->GetZSliceStep());
    return;    
  }
  if (event == VtkViewImage2DCommand::EndZSliceMoveEvent) {
    return;    
  }
  if (event == VtkViewImage2DCommand::FullPageEvent) {
    m_view->DisplayInFullPage();    
    return;    
  }

}

void
VtkViewImage2DCommand::StartWindowing()
{
  m_initialWindow = m_view->GetColorWindow();
  m_initialLevel = m_view->GetColorLevel();
}
  
void
VtkViewImage2DCommand::Windowing(vtkInteractorStyleImage* p_isi)
{
  int *size;
  size = m_view->GetVtkRenderWindow()->GetSize();
  float window = m_initialWindow;
  float level = m_initialLevel;
    
  // Compute normalized delta
  float dx = 4.0 * 
    (p_isi->GetWindowLevelCurrentPosition()[0] - 
     p_isi->GetWindowLevelStartPosition()[0]) / size[0];
  float dy = 4.0 * 
    (p_isi->GetWindowLevelStartPosition()[1] - 
     p_isi->GetWindowLevelCurrentPosition()[1]) / size[1];
    
  // Scale by current values
  if (fabs(window) > 0.01)
    dx = dx * window;
  else
    dx = dx * (window < 0 ? -0.01 : 0.01);

  if (fabs(level) > 0.01)
    dy = dy * level;
  else
    dy = dy * (level < 0 ? -0.01 : 0.01);
  
  // Abs so that direction does not flip
  if (window < 0.0) 
    dx = -1*dx;
  if (level < 0.0) 
    dy = -1*dy;
    
  // Compute new window level
  float newWindow = dx + window;
  float newLevel;
  newLevel = level - dy;
    
  // Stay away from zero and really
  if (fabs(newWindow) < 0.01)
    newWindow = 0.01 * (newWindow < 0 ? -1 : 1);

  if (fabs(newLevel) < 0.01)
    newLevel = 0.01 * (newLevel < 0 ? -1 : 1);
    
  m_view->SetColorWindow(newWindow);
  m_view->SetColorLevel(newLevel);
  m_view->Render();
}

void
VtkViewImage2DCommand::EndWindowing()
{
}

void
VtkViewImage2DCommand::StartPicking()
{
}

void
VtkViewImage2DCommand::EndPicking(vtkInteractorStyleImage* p_isi)
{
  vtkRenderWindowInteractor *rwi = p_isi->GetInteractor();
  vtkAssemblyPath *path=NULL;
  p_isi->FindPokedRenderer(rwi->GetEventPosition()[0],
                          rwi->GetEventPosition()[1]);
  rwi->GetPicker()->Pick(rwi->GetEventPosition()[0],
                         rwi->GetEventPosition()[1], 0.0, 
                         p_isi->GetCurrentRenderer());
  vtkAbstractPropPicker *picker;
  if ( (picker=vtkAbstractPropPicker::SafeDownCast(rwi->GetPicker())) ){
    path = picker->GetPath();
  }
  if ( path != NULL ) {
    double* world = picker->GetPickPosition();
    double  pos[3];
    m_view->GetCurrentPoint(pos);    
    switch(m_view->GetOrientation()) {
    case VtkViewImage2D::SAGITTAL_ID:
      pos [1] = world[0] * -1;
      pos [2] = world[1];
      break;
    case VtkViewImage2D::CORONAL_ID:
      pos [0] = world[0];
      pos [2] = world[1];
      break;
    case VtkViewImage2D::AXIAL_ID:
      pos [0] = world[0];
      pos [1] = world[1] * -1;
      break;
      
    }
    // Treat extrem positions
    for (unsigned int i=0; i<3; i++) {
      if (pos[i] < m_view->GetWholeMinPosition(i)) pos[i] = m_view->GetWholeMinPosition(i)+0.0005;
      if (pos[i] > m_view->GetWholeMaxPosition(i)) pos[i] = m_view->GetWholeMaxPosition(i)-0.0005;
    }
    // Set the position
    m_view->SetCurrentPoint(pos);    
  }
}

void
VtkViewImage2DCommand::ChangeZSlice(int p_nbSlices)
{
    int current_slice = m_view->GetZSlice();
    int dest_slice    = current_slice + p_nbSlices;
    int min_slice     = m_view->GetWholeZMin();
    int max_slice     = m_view->GetWholeZMax();
     
    if ((dest_slice > min_slice) && (dest_slice < max_slice)) {        
      m_view->SetZSlice(dest_slice);
    }
    if ((dest_slice >= max_slice) && (current_slice != max_slice)) {        
      m_view->SetZSlice(max_slice);
    }
    if ((dest_slice <= min_slice) && (current_slice != min_slice)) {        
      m_view->SetZSlice(min_slice);
    }
    m_view->Render();
}

void
VtkViewImage2DCommand::SetView(VtkViewImage2D *p_view)
{
  m_view = p_view;  
}

}// end of namespace
