#include "mipsTestMain.h"

#include "mipsIterationScheduler.h"
#include "mipsIterationSchedulerGlib.h"
#include "mipsIterationUnitC.h"
#include "mipsIterationUnitCpp.h"

using namespace mips;
using namespace std;
#include <iostream>
#include <iomanip>
//------------------------------------------------------------------
static int _counter=0;   


class Sleeper 
{
public:
  Sleeper(double seconds_to_sleep) {
	m_secondsToSleep=seconds_to_sleep;
  };
  void Wait(){
	Timer t;
	t.Start();
	while(t.GetTime()<m_secondsToSleep){}
  };
protected:
  double m_secondsToSleep;
private:
  Sleeper(){};
};


void Wait90msAndCount(){
	Timer t;
	t.Start();
	while(t.GetTime()<0.09){}
	_counter++;
}

void Wait50ms(){
	Timer t;
	t.Start();
	while(t.GetTime()<0.05){}
}


int BaseSchedulerTest(IterationScheduler& scheduler){
  int ret_val=0;
  Sleeper sleeper(0.01);
  IterationUnitCpp<Sleeper> unit10(&sleeper, &Sleeper::Wait);
  IterationUnitC unit50(Wait50ms);
  IterationUnitC unit90(Wait90msAndCount);

  scheduler.ActivatePrintIterationRate();
  scheduler.AddUnit(&unit10);
  scheduler.AddUnit(&unit50);
  scheduler.AddUnit(&unit90);
  unit10.SetLabel("My 10ms unit");
  unit50.SetLabel("My 50ms unit");
  unit90.SetLabel("My 90ms unit");
  // unit10.SetMethodToRun(&sleeper, &Sleeper::Wait);
  // unit50.SetFunctionToRun(Wait50ms);
  // unit90.SetFunctionToRun(Wait90msAndCount);
  unit50.SetActivationPeriod(2);
  unit50.SetNumberOfNonActivatedRuns(3);
  unit10.PrintSelf(cout);
  unit50.PrintSelf(cout);
  unit90.PrintSelf(cout);



  // Default Iteration = 1
  _counter=0;
  cout << endl << "Default Iteration" << endl;
  cout << "Counter before " << _counter << endl;
  scheduler.Iterate();
  cout << "Counter after " << _counter << endl;
  if (_counter == 1) {
	cout << "------>  Success " << endl;
  }
  else {
	cout << "scheduler.GetIterationsDone()="<< scheduler.GetIterationsDone()<< endl;
	cout << "effective counts = " << _counter << endl;
	cout << "------>  Failed " << endl;
	ret_val=1;
  }

  // multiple Iterations
  //
  // Set again that value it have been decremented during previous run  
  unit50.SetNumberOfNonActivatedRuns(3);  
  _counter=0;
  cout << endl << "Multiple Iterations" << endl;
  cout << "Counter before " << _counter << endl;
  scheduler.Iterate(10);
  cout << "Counter after " << _counter << endl;
  if (_counter == 10) {
	cout << " Counter --> OK " << endl;
  }
  else {
	cout << " Counter --> Failed " << endl;
	ret_val=2;
  }

  // unit50 activated 4 times => 4*150ms+6*100ms = 1.2 seconds
  double global_time_expected = (1.2 / 10);
  if ((scheduler.GetGlobalTimePerIteration() - global_time_expected) < 1e-3) {
	cout << " GlobalTimePerIteration --> OK " << endl;
  }
  else {
	cout << " GlobalTimePerIteration expected ("<<global_time_expected
		 << ") --> Failed " 
		 << scheduler.GetGlobalTimePerIteration() << endl;
	ret_val=2;
  }

  if ((scheduler.GetGlobalIterationFrequency() - 
	   (1/global_time_expected)) < 1e-3 ) {
	cout << " GlobalIterationFrequency--> OK " << endl;
  }
  else {
	cout << " GlobalterationFrequency expected ("<< 1/global_time_expected
		 << ") --> Failed " 
		 <<scheduler.GetGlobalIterationFrequency()<< endl;
	ret_val=2;
  }

  //
  //Test the collected results over 9 iterations (1.05 seconds)
  //
  if (unit50.GetRunningTime() == 0.15) {
	cout << " Unit-RunningTime--> OK " << endl;
  }
  else {
	cout << " Unit-RunningTime --> Failed " 
		 << unit50.GetRunningTime() << endl;
	ret_val=2;
  }

  if (unit50.GetNumberOfActivations() ==3) {
	cout << " Unit-NumberOfActivations--> OK " << endl;
  }
  else {
	cout << " Unit-NumberOfActivations --> Failed " 
		 << unit50.GetNumberOfActivations() << endl;
	ret_val=2;
  }

  cout << "ret_val = "<<ret_val<< endl;
  return ret_val;

}




int SchedulerTest(int argc, char *argv[]){
  IterationScheduler scheduler;
  return BaseSchedulerTest(scheduler);
}
int SchedulerGlibTest(int argc, char *argv[]){
  IterationSchedulerGlib scheduler;
  return BaseSchedulerTest(scheduler);
}


//////////////////////////////////////////////////////////////////
//             Register all tests to be done
//////////////////////////////////////////////////////////////////

void RegisterTests() {
  REGISTER_TEST(SchedulerTest);
  REGISTER_TEST(SchedulerGlibTest);
}
