#include "gtk/gtk.h"
#include "gtk/gtkgl.h"
#include "mipsDataSetManager.h"
#include "mipsPlugInFactory.h"
#include "mipsViewer.h"
#include "mipsViewSpace.h"
#include "mipsVtkView.h"
#include "mipsView3D.h"
#include "vtkActor.h"
#include "vtkPolyData.h"
#include "vtkCellArray.h"
#include "vtkPolyDataMapper.h"
#include "MIPS_GUIPKG_config.h"

vtkActor* CreateSampleCubeActor()
{
  // Static model initialisation
  cout<<"Init Cube"<<endl;
  vtkPolyData *cube = vtkPolyData::New();
  vtkActor *cubeActor=0;

  int i;
  static float x[8][3]={{0,0,0}, {1,0,0}, {1,1,0}, {0,1,0},
                        {0,0,1}, {1,0,1}, {1,1,1}, {0,1,1}};
  static vtkIdType pts[6][4]={{0,1,2,3}, {4,5,6,7}, {0,1,5,4},
                              {1,2,6,5}, {2,3,7,6}, {3,0,4,7}};
	
  vtkPoints *points = vtkPoints::New();
  vtkCellArray *polys = vtkCellArray::New();
  // vtkFloatArray *scalars = vtkFloatArray::New();
  // Load the point, cell, and data attributes.
  for (i=0; i<8; i++) points->InsertPoint(i,x[i]);
  for (i=0; i<6; i++) polys->InsertNextCell(4,pts[i]);
  //  for (i=0; i<8; i++) scalars->InsertTuple1(i,5);

  // We now assign the pieces to the vtkPolyData.
  cube->SetPoints(points);
  points->Delete();
  cube->SetPolys(polys);
  polys->Delete();
  //  cube->GetPointData()->SetScalars(scalars);
  // scalars->Delete();
	
  // Now we'll look at it.
  vtkPolyDataMapper *cubeMapper = vtkPolyDataMapper::New();
  cubeMapper->SetInput(cube);
  cubeMapper->SetScalarRange(0,7);
  cubeActor = vtkActor::New();
  cubeActor->SetMapper(cubeMapper);

  return cubeActor;
}

int
main(int argc, char **argv)
{
    gtk_init(&argc, &argv);
    gtk_gl_init (&argc, &argv);
    
    mips::Viewer::Initialize();

    // Chargement des plug-ins
    mips::PlugInFactory::SetPath(MIPS_PLUGIN_PATH);
    mips::PlugInFactory::LoadPlugIns();    

    // Demo cubes
    vtkActor* cube_actor;
    mips::VtkView* view=0;
    for (int i=1; i<=1; i++) {
      /* Create a cube, put the cube in a new view and add 
         the view in the viewspace */
      cube_actor = CreateSampleCubeActor();
      view = new mips::VtkView();	
      view->AddActor(cube_actor);
      mips::Viewer::GetViewSpace()->AddView(view);
    }

    // Camera Yav
    std::string filename = std::string(TESTDATA_DIR)+
      std::string("/Meshes/dodecahedre_sz_fix.atet3D");    
    if (mips::DataSetManager::GetDataSetByFileName(filename.c_str()) == NULL) {
      std::cout<<"DataSet loading failed for file : "<<filename<<endl;
    }
    
    mips::Viewer::GetViewSpace()->AddView(new mips::View3D);    
    
    gtk_main();
    
    return 0;
}
