#include "mipsTestMain.h"

#include "mipsTimer.h"
#include <time.h>
#include <iostream>

using namespace std;

//------------------------------------------------------------------
int TimerTest(int argc, char *argv[]){
 
  const double TEST_TIME=1;
  const double TEST_TIME_TOLERANCE=0.001;

  clock_t begin, end;
  
  mips::Timer timer;

  if (timer.GetTime() != 0 ) {
	cout << "Error : Bad timer initialisation" << endl;
	return 1;
  }
  else {
	cout << "Test Timer Initialisation  OK. "<<endl;
  }

  begin=clock();
  timer.Start();
  bool isOK=true;
  double last_time, tested_time;
  while (((last_time=timer.GetTime()) < TEST_TIME) && isOK) {
	// loosing time ...
	int j=0;
	for (int i=1;i<100000;i++){j++;}
// 	cout << "          Loosing time ; j="<<j<<"  "<< endl;
	tested_time = (double(clock()-begin)/CLOCKS_PER_SEC);
	isOK=(tested_time < (TEST_TIME+TEST_TIME_TOLERANCE) );
// 	cout << "           last_time="<<last_time<<" sec." << endl;
// 	cout << "           tested_time="<<tested_time<<" sec." << endl;
  }
  if (!isOK) {
	cout << "Error : GetTime is not OK "<<endl;
	cout << "           last_time="<<last_time<<" sec." << endl;
	cout << "           tested_time="<<tested_time<<" sec." << endl;
	return 1;
  }
  else {
	cout << "Test GetTime time OK. "<<endl;
  }


  timer.Reset();
  begin=clock();
  isOK=true;
  while (isOK) {
	tested_time = (double(clock()-begin)/CLOCKS_PER_SEC);
	isOK=(tested_time < TEST_TIME );
  }
  end=clock();
  double stop_time=timer.Stop();
  tested_time=(double(end-begin)/CLOCKS_PER_SEC);
  if ((tested_time>stop_time) || (stop_time > tested_time+ TEST_TIME_TOLERANCE)){
	cout << "Error : Stop returned time is not OK "<<endl;
	cout << "           stop_time="<<stop_time<<" sec." << endl;
	cout << "           tested_time="<<tested_time<<" sec." << endl;
	return 1;
  }
  else {
	cout << "Test Stop returned time OK. "<<endl;
	cout << "           stop_time="<<stop_time<<" sec." << endl;
	cout << "           tested_time="<<tested_time<<" sec." << endl;
  }
  
  if (timer.IsRunning()) {
	cout << "Error : timer is still running after stop "<<endl;
	return 1;
  }
  else {
	cout << "Test IsRunning OK. "<<endl;
  }

  return 0;

}




//////////////////////////////////////////////////////////////////
//             Register all tests to be done
//////////////////////////////////////////////////////////////////

void RegisterTests() {
  REGISTER_TEST(TimerTest);
}
