#include "mipsIterationUnit.h"
#include <iostream>


namespace mips {

IterationUnit::IterationUnit ()
{
  m_label = "IterationUnit default label";
  m_activationPeriod=1;
  m_numberOfNonActivatedRuns=m_activationPeriod-1;
  this->ResetIterationRateData();
}

IterationUnit::~IterationUnit ()
{
}
void IterationUnit::Run ()
{
  if (m_numberOfNonActivatedRuns == 0) {
	if(m_activationPeriod>1) m_numberOfNonActivatedRuns=m_activationPeriod-1;
	m_currentNumberOfActivations++;
	m_runningTimer.Start();
	CallFunction();
	m_runningTimer.Stop();
  }
  else {
	m_numberOfNonActivatedRuns--;
  }
}

const std::string& IterationUnit::GetLabel()
{
  return m_label;
}

void IterationUnit::SetLabel(const std::string& label)
{
  m_label = label;
}

void IterationUnit::SetActivationPeriod(unsigned int activationPeriod) {
  m_activationPeriod=activationPeriod;
}

void IterationUnit::SetNumberOfNonActivatedRuns(unsigned int nbOfNonActivRuns) {
  m_numberOfNonActivatedRuns=nbOfNonActivRuns;
}

double IterationUnit::GetRunningTime() {
  return m_runningTime;
}

unsigned long IterationUnit::GetNumberOfActivations() {
  return m_numberOfActivations;
}

void IterationUnit::CollectIterationRateData() {
  m_runningTime=m_runningTimer.GetTime();
  m_runningTimer.Reset();
  m_numberOfActivations=m_currentNumberOfActivations;
  m_currentNumberOfActivations=0;
}

void IterationUnit::ResetIterationRateData() {
  m_runningTime=0;
  m_runningTimer.Reset();
  m_numberOfActivations=0;
  m_currentNumberOfActivations=0;
}

void IterationUnit::PrintSelf(std::ostream& os)
{
  os << "IterationUnit : "<< m_label << std::endl;
  os << "       collected running time =  "<< this->GetRunningTime() << " second(s)" << std::endl;
  os << "  collected nb of activations =  "<< this->GetNumberOfActivations() << std::endl;
  os << "            activation period =  "<< m_activationPeriod << " time(s) per iter." << std::endl;
  os << "number of non activation runs =  "<< m_numberOfNonActivatedRuns << " run(s)" << std::endl;
}

} //namespace mips
