#ifndef MIPS_PLUGIN_H
#define MIPS_PLUGIN_H

#include <gmodule.h>

#define MIPS_PLUGIN(name)                                                   \
extern "C" G_MODULE_EXPORT  void * mips_get_plug_in(void)                   \
{                                                                           \
  return static_cast<void*>(reinterpret_cast<mips::PlugIn *>(new name ())); \
}


namespace mips {

/** \class PlugIn
    \ingroup MipsViewer
*/
class PlugIn 
{
 public:
  /** The callback called when the plug-in is loaded.  */
  virtual void Init(GModule *) = 0;

  /** The callback called before the plug-in is unloaded*/
  virtual void Finalize() = 0;

  /** The plug-in's name */
  virtual const char* GetName() = 0;  

  /** Return the GModule from where the plugin is comming from. */
  GModule* GetModule();  

 protected:
  /** Sets the GModule from where the plugin is comming from.
      Normaly set by the Init method. */
  void SetModule(GModule* );  

 private:
  GModule* m_module;  
  
};

} //namespace mips

#endif /* MIPS_PLUGIN_H */
