#include "mipsPlugInFactory.h"
#include "mipsPlugIn.h"

#include "glib.h"

#include "gmodule.h"
#include <iostream>
#include <string.h>

namespace mips {

std::vector<PlugIn*> PlugInFactory::m_plugIns;  
std::string           PlugInFactory::m_path;  

void   
PlugInFactory::LoadPlugIns()
{
  if (!g_module_supported()) {
    g_warning ("dynamic loading is not supported.\n LoadPlugIns Fails \n");
    return;
  }
      
  unsigned int pos = 0;  
  std::string path = m_path;
  while (pos < std::string::size_type(-1)) {
    pos = path.find(";",0);
    if (pos < std::string::size_type(-1)) {
      PlugInFactory::LoadPlugInsInDirectory(path.substr(0, pos).c_str());
      path = path.erase(0,pos+1);      
    }
  }
  PlugInFactory::LoadPlugInsInDirectory(path.c_str());
}

void
PlugInFactory::LoadPlugIn(const char* p_name)
{
  if (!g_module_supported()) {
    g_warning ("dynamic loading is not supported.\n LoadPlugIns Fails \n");
    return;
  }
      
  unsigned int pos = 0;  
  std::string path = m_path;
  while (pos < std::string::size_type(-1)) {
    pos = path.find(";",0);
    if (pos < std::string::size_type(-1)) {
      PlugInFactory::LoadPlugInInDirectory(path.substr(0, pos).c_str(), p_name);
      path = path.erase(0,pos+1);      
    }
  }
  PlugInFactory::LoadPlugInInDirectory(path.c_str(), p_name);
}
 
bool
_select(const char * p_filename)
{
  std::string suffix = "PlugIn.";
  suffix += G_MODULE_SUFFIX;  
  unsigned int entry_len = strlen(p_filename);
  const char* entry_suffix = p_filename + entry_len - suffix.size();
  
  return (strcmp(entry_suffix,suffix.c_str()) == 0);
}
  
  
void   
PlugInFactory::LoadPlugInsInDirectory(const char* p_dir)
{
  GDir *dp=0;
  const char *entry=0;
  GError * error=0;  
  
  if((dp = g_dir_open(p_dir,0,&error)) == NULL)
    std::cerr<<"Open directory ("<<p_dir<<") : "<<error->message<<std::endl;  
  else {
    while((entry = g_dir_read_name(dp)) != NULL)
      if (_select(entry)){
        PlugInFactory::LoadPlugInInDirectory(p_dir, entry, false);
      }        
    g_dir_close(dp);
  }
}

void   
PlugInFactory::LoadPlugInInDirectory(const char* p_dir, const char* p_name, bool p_addPrefixAndSuffixes)
{
  std::string filename = p_dir;
  if (p_addPrefixAndSuffixes) {    
    filename += "/lib";            
    filename += p_name;
    filename += ".";
    filename += G_MODULE_SUFFIX;
  }
  else {
    filename += "/";            
    filename += p_name;
  }
  
  GModule* module = g_module_open(filename.c_str(),G_MODULE_BIND_LAZY);
  PlugIn* plug_in = NULL;
  std::cout<< "Trying "<<filename.c_str()<< " loading ...";
  if (module) {
    std::cout<< "ok"<<std::endl;
    void* (* get_plugin_function)(void) = NULL;              
    if (g_module_symbol(module, "mips_get_plug_in",(gpointer *)&get_plugin_function)) {
      plug_in = static_cast<PlugIn*>(get_plugin_function());
      plug_in->Init(module);                
      m_plugIns.push_back(plug_in);              
    }
  }
  if (plug_in == NULL) {
    std::cout<< "failed"<<std::endl<<g_module_error ()<<std::endl ;    
  }
}        
  
void
PlugInFactory::SetPath(const char* p_path)
{
  m_path = p_path;    
}

const char*
PlugInFactory::GetPath()
{
  return m_path.c_str();  
}

  
} //namespace mips
