/*
 * Based on Togl - a Tk OpenGL widget
 * Version 1.4
 * Copyright (C) 1996-1997  Brian Paul and Ben Bederson
 * See the LICENSE file for copyright details.
 */
#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#pragma implementation "graphics3D/BasicTogl.h"
#include <modules/WindowDump.h>
#include <modules/BasicTogl.h>


extern "C" {

/*** Tcl/Tk headers ***/
#ifdef X11
#ifndef _TKPORT
#define _TKPORT  /* This eliminates need to include a bunch of Tk baggage */
#endif /* X11 */
#endif
#include <tcl.h>
#include <tk.h>
#if defined(X11)
#if TK_MAJOR_VERSION==4 && TK_MINOR_VERSION==0
#  include <graphics3D/tkInt4.0.h>
#  define NO_TK_CURSOR
#elif TK_MAJOR_VERSION==4 && TK_MINOR_VERSION==1
#  include <graphics3D/tkInt4.1.h>
#elif TK_MAJOR_VERSION==4 && TK_MINOR_VERSION==2
#  include <graphics3D/tkInt4.2.h>
#elif TK_MAJOR_VERSION==8 && TK_MINOR_VERSION==0
#  include <graphics3D/tkInt8.0.h>
#elif TK_MAJOR_VERSION==8 && TK_MINOR_VERSION==1
#  include <graphics3D/tkInt8.1.h>
#elif TK_MAJOR_VERSION==8 && TK_MINOR_VERSION==3
#  include <graphics3D/tkInt8.3.h>
#elif TK_MAJOR_VERSION==8 && TK_MINOR_VERSION==4
#  include <graphics3D/tkInt8.4.h>
#elif TK_MAJOR_VERSION==8 && TK_MINOR_VERSION==5
#  include <graphics3D/tkInt8.5.h>
#else
#error   Sorry, you will have to edit togl.c to include the right tkInt.h file
#endif
#elif defined(WIN32)
#if TK_MAJOR_VERSION<8
   Sorry Windows version requires Tcl/Tk ver 8.0 or higher.
#endif
#if TK_MAJOR_VERSION==8 && TK_MINOR_VERSION==3
#include <graphics3D/tkInt8.3.h>
#endif
#if TK_MAJOR_VERSION==8 && TK_MINOR_VERSION==4
#include <graphics3D/tkInt8.4.h>
#endif
#include <graphics3D/tkWinInt.h>
#else
#endif /* X11 */

}

namespace yav {



#ifdef WIN32
/* maximum size of a logical palette corresponding to a colormap in color index mode */
#define MAX_CI_COLORMAP_SIZE 4096
#endif /* WIN32 */

#ifndef MAX
#define MAX(a,b)	(((a)>(b))?(a):(b))
#endif

#define TCL_ERR(interp, string) {Tcl_ResetResult(interp);		\
			         Tcl_AppendResult(interp, string, 0);\
				 return(TCL_ERROR);}

#define ALL_EVENTS_MASK \
    KeyPressMask|KeyReleaseMask|ButtonPressMask|ButtonReleaseMask|	\
    EnterWindowMask|LeaveWindowMask|PointerMotionMask|ExposureMask|	\
    VisibilityChangeMask|FocusChangeMask|PropertyChangeMask|ColormapChangeMask


/* NTNTNT need to change to handle Windows Data Types */
/*
 * Prototypes for functions local to this file
 */
#ifdef MESA_COLOR_HACK
static int get_free_color_cells( Display *display, int screen,
                                 Colormap colormap);
static void free_default_color_cells( Display *display, Colormap colormap);
#endif


  ClientData BasicTogl::DefaultClientData = 0;

#ifdef MESA_COLOR_HACK
  unsigned long *BasicTogl::ToglMesaUsedPixelCells;
  int BasicTogl::ToglMesaUsedFreeCells;
#endif


/* Defaults */
#define DEFAULT_WIDTH		400
#define DEFAULT_HEIGHT		400
#define DEFAULT_WIDTH_STRING		"400"
#define DEFAULT_HEIGHT_STRING		"400"
#define DEFAULT_IDENT		""
#define DEFAULT_SCENE		0
#define DEFAULT_FONTNAME	"fixed"
#define DEFAULT_TIME		"1"

// Warning here: (int)(&Togl::width)-1 is supposed to give width
// offset in Togl class. I am not sure it is valid with any
// C++ compiler (eg++ &Togl::width gives width's offset + 1...).
#if (defined __GNUC__) && (__GNUC__ >= 3)
#undef Tk_Offset
#define Tk_Offset(T,m) ((size_t)(&((T *)1)->m) - 1)
#endif
  Tk_ConfigSpec BasicTogl::configSpecs[] = { 
       {TK_CONFIG_INT, "-height", "height", "Height",
    DEFAULT_HEIGHT_STRING, Tk_Offset(BasicTogl, requiredHeight), 0, 0},

     {TK_CONFIG_INT, "-width", "width", "Width",
     DEFAULT_WIDTH_STRING, Tk_Offset(BasicTogl, requiredWidth), 0, 0},

    {TK_CONFIG_STRING, "-ident", "ident", "Ident",
	DEFAULT_IDENT, Tk_Offset(BasicTogl, Ident), 0, 0},

    {TK_CONFIG_BOOLEAN, "-rgba", "rgba", "Rgba",
     "true", Tk_Offset(BasicTogl, RgbaFlag), 0, 0},

    {TK_CONFIG_INT, "-redsize", "redsize", "RedSize",
     "1", Tk_Offset(BasicTogl, RgbaRed), 0, 0},

    {TK_CONFIG_INT, "-greensize", "greensize", "GreenSize",
     "1", Tk_Offset(BasicTogl, RgbaGreen), 0, 0},

    {TK_CONFIG_INT, "-bluesize", "bluesize", "BlueSize",
     "1", Tk_Offset(BasicTogl, RgbaBlue), 0, 0},

    {TK_CONFIG_BOOLEAN, "-double", "double", "Double",
     "false", Tk_Offset(BasicTogl, requiredDoubleFlag), 0, 0},

    {TK_CONFIG_BOOLEAN, "-depth", "depth", "Depth",
     "false", Tk_Offset(BasicTogl, DepthFlag), 0, 0},

    {TK_CONFIG_INT, "-depthsize", "depthsize", "DepthSize",
     "1", Tk_Offset(BasicTogl, DepthSize), 0, 0},

    {TK_CONFIG_BOOLEAN, "-accum", "accum", "Accum",
     "false", Tk_Offset(BasicTogl, AccumFlag), 0, 0},

    {TK_CONFIG_INT, "-accumredsize", "accumredsize", "AccumRedSize",
     "1", Tk_Offset(BasicTogl, AccumRed), 0, 0},

    {TK_CONFIG_INT, "-accumgreensize", "accumgreensize", "AccumGreenSize",
     "1", Tk_Offset(BasicTogl, AccumGreen), 0, 0},

    {TK_CONFIG_INT, "-accumbluesize", "accumbluesize", "AccumBlueSize",
     "1", Tk_Offset(BasicTogl, AccumBlue), 0, 0},

    {TK_CONFIG_INT, "-accumalphasize", "accumalphasize", "AccumAlphaSize",
     "1", Tk_Offset(BasicTogl, AccumAlpha), 0, 0},

    {TK_CONFIG_BOOLEAN, "-alpha", "alpha", "Alpha",
     "false", Tk_Offset(BasicTogl, AlphaFlag), 0, 0},

    {TK_CONFIG_INT, "-alphasize", "alphasize", "AlphaSize",
     "1", Tk_Offset(BasicTogl, AlphaSize), 0, 0},

    {TK_CONFIG_BOOLEAN, "-stencil", "stencil", "Stencil",
     "false", Tk_Offset(BasicTogl, StencilFlag), 0, 0},

    {TK_CONFIG_INT, "-stencilsize", "stencilsize", "StencilSize",
     "1", Tk_Offset(BasicTogl, StencilSize), 0, 0},

    {TK_CONFIG_INT, "-auxbuffers", "auxbuffers", "AuxBuffers",
     "0", Tk_Offset(BasicTogl, AuxNumber), 0, 0},

    {TK_CONFIG_BOOLEAN, "-privatecmap", "privateCmap", "PrivateCmap",
     "false", Tk_Offset(BasicTogl, PrivateCmapFlag), 0, 0},

    {TK_CONFIG_BOOLEAN, "-overlay", "overlay", "Overlay",
    "false", Tk_Offset(BasicTogl, OverlayFlag), 0, 0},

#ifndef NO_TK_CURSOR
    { TK_CONFIG_ACTIVE_CURSOR, "-cursor", "cursor", "Cursor",
     "", Tk_Offset(BasicTogl, Cursor), TK_CONFIG_NULL_OK },
#endif

    {TK_CONFIG_INT, "-time", "time", "Time",
     DEFAULT_TIME, Tk_Offset(BasicTogl, Time), 0, 0},

    {TK_CONFIG_END, (char *) 0, (char *) 0, (char *) 0,
     (char *) 0, 0, 0, 0}
};

  //static char *toglUsage = "usage: togl %w [-scene scn%i] [-height %i] [-width %i] [-ident %s] [-rgba %b] [-redsize %i] [-greensize %i] [-bluesize %i] [-double %b] [-depth %b] [-depthsize %i] [-accum %b] [-accumredsize %i] [-accumgreensize %i] [-accumbluesize %i] [-accumalphasize %i] [-alpha %b] [-alphasize %i] [-stencil %b] [-stencilsize %i] [-auxbuffers %i] [-privatecmap %b] [-overlay %b] [-stereo %b] [-cursor %cur] [-time %i]";

#if defined(X11)
   /// the General OpenGL context of the camera
    GLXContext BasicTogl::masterCtxtHandle=0;
 /// the display of the application
  Display *BasicTogl::masterDisplayHandle=0;
#elif defined(WIN32)
    /// the General OpenGL context of the camera
 HGLRC BasicTogl::masterCtxtHandle=0;
#endif




#if defined(X11)
/*
 * Return an X colormap to use for OpenGL RGB-mode rendering.
 * Input:  dpy - the X display
 *         scrnum - the X screen number
 *         visinfo - the XVisualInfo as returned by glXChooseVisual()
 * Return:  an X Colormap or 0 if there's a _serious_ error.
 */
static Colormap
get_rgb_colormap(Display *dpy, int scrnum, XVisualInfo *visinfo)
{
   Atom hp_cr_maps;
   Status status;
   int numCmaps;
   int i;
   XStandardColormap *standardCmaps;
   Window root = RootWindow(dpy,scrnum);
   int using_mesa;

   /*
    * First check if visinfo's visual matches the default/root visual.
    */
   if (visinfo->visual==DefaultVisual(dpy,scrnum)) {
      /* use the default/root colormap */
      Colormap cmap;
      cmap = DefaultColormap( dpy, scrnum );
#ifdef MESA_COLOR_HACK
      (void) get_free_color_cells( dpy, scrnum, cmap);
#endif
      return cmap;
   }

   /*
    * Check if we're using Mesa.
    */
   if (strstr(glXQueryServerString( dpy, scrnum, GLX_VERSION ), "Mesa")) {
      using_mesa = 1;
   }
   else {
      using_mesa = 0;
   }

   /*
    * Next, if we're using Mesa and displaying on an HP with the "Color
    * Recovery" feature and the visual is 8-bit TrueColor, search for a
    * special colormap initialized for dithering.  Mesa will know how to
    * dither using this colormap.
    */
   if (using_mesa) {
      hp_cr_maps = XInternAtom( dpy, "_HP_RGB_SMOOTH_MAP_LIST", True );
      if (hp_cr_maps
#ifdef __cplusplus
	  && visinfo->visual->c_class==TrueColor
#else
	  && visinfo->visual->class==TrueColor
#endif
	  && visinfo->depth==8) {
	 status = XGetRGBColormaps( dpy, root, &standardCmaps,
				    &numCmaps, hp_cr_maps );
	 if (status) {
	    for (i=0; i<numCmaps; i++) {
	       if (standardCmaps[i].visualid == visinfo->visual->visualid) {
                  Colormap cmap = standardCmaps[i].colormap;
                  XFree( standardCmaps );
		  return cmap;
	       }
	    }
            XFree(standardCmaps);
	 }
      }
   }

   /*
    * Next, try to find a standard X colormap.
    */
#ifndef SOLARIS_BUG
   status = XmuLookupStandardColormap( dpy, visinfo->screen,
				       visinfo->visualid, visinfo->depth,
				       XA_RGB_DEFAULT_MAP,
				       /* replace */ False, /* retain */ True);
   if (status == 1) {
      status = XGetRGBColormaps( dpy, root, &standardCmaps,
				 &numCmaps, XA_RGB_DEFAULT_MAP);
      if (status == 1) {
         for (i = 0; i < numCmaps; i++) {
	    if (standardCmaps[i].visualid == visinfo->visualid) {
               Colormap cmap = standardCmaps[i].colormap;
	       XFree(standardCmaps);
	       return cmap;
	    }
	 }
         XFree(standardCmaps);
      }
   }
#endif

   /*
    * If we get here, give up and just allocate a new colormap.
    */
   return XCreateColormap( dpy, root, visinfo->visual, AllocNone );
}
#elif defined(WIN32)

/* Code to create RGB palette is taken from the GENGL sample program
   of Win32 SDK */

static unsigned char threeto8[8] = {
    0, 0111>>1, 0222>>1, 0333>>1, 0444>>1, 0555>>1, 0666>>1, 0377
};

static unsigned char twoto8[4] = {
    0, 0x55, 0xaa, 0xff
};

static unsigned char oneto8[2] = {
    0, 255
};

static int defaultOverride[13] = {
    0, 3, 24, 27, 64, 67, 88, 173, 181, 236, 247, 164, 91
};

static PALETTEENTRY defaultPalEntry[20] = {
    { 0,   0,   0,    0 },
    { 0x80,0,   0,    0 },
    { 0,   0x80,0,    0 },
    { 0x80,0x80,0,    0 },
    { 0,   0,   0x80, 0 },
    { 0x80,0,   0x80, 0 },
    { 0,   0x80,0x80, 0 },
    { 0xC0,0xC0,0xC0, 0 },

    { 192, 220, 192,  0 },
    { 166, 202, 240,  0 },
    { 255, 251, 240,  0 },
    { 160, 160, 164,  0 },

    { 0x80,0x80,0x80, 0 },
    { 0xFF,0,   0,    0 },
    { 0,   0xFF,0,    0 },
    { 0xFF,0xFF,0,    0 },
    { 0,   0,   0xFF, 0 },
    { 0xFF,0,   0xFF, 0 },
    { 0,   0xFF,0xFF, 0 },
    { 0xFF,0xFF,0xFF, 0 }
};

static unsigned char
ComponentFromIndex(int i, UINT nbits, UINT shift)
{
    unsigned char val;

    val = (unsigned char) (i >> shift);
    switch (nbits) {

    case 1:
        val &= 0x1;
        return oneto8[val];

    case 2:
        val &= 0x3;
        return twoto8[val];

    case 3:
        val &= 0x7;
        return threeto8[val];

    default:
        return 0;
    }
}

static Colormap Win32CreateRgbColormap(PIXELFORMATDESCRIPTOR pfd)
{
    TkWinColormap *cmap = (TkWinColormap *) ckalloc(sizeof(TkWinColormap));
    LOGPALETTE *pPal;
    int n, i;

    n = 1 << pfd.cColorBits;
    pPal = (PLOGPALETTE)LocalAlloc(LMEM_FIXED, sizeof(LOGPALETTE) +
            n * sizeof(PALETTEENTRY));
    pPal->palVersion = 0x300;
    pPal->palNumEntries = n;
    for (i=0; i<n; i++) {
        pPal->palPalEntry[i].peRed =
                ComponentFromIndex(i, pfd.cRedBits, pfd.cRedShift);
        pPal->palPalEntry[i].peGreen =
                ComponentFromIndex(i, pfd.cGreenBits, pfd.cGreenShift);
        pPal->palPalEntry[i].peBlue =
                ComponentFromIndex(i, pfd.cBlueBits, pfd.cBlueShift);
        pPal->palPalEntry[i].peFlags = 0;
    }

    /* fix up the palette to include the default GDI palette */
    if ((pfd.cColorBits == 8)                           &&
        (pfd.cRedBits   == 3) && (pfd.cRedShift   == 0) &&
        (pfd.cGreenBits == 3) && (pfd.cGreenShift == 3) &&
        (pfd.cBlueBits  == 2) && (pfd.cBlueShift  == 6)
       ) {
        for (i = 1 ; i <= 12 ; i++)
            pPal->palPalEntry[defaultOverride[i]] = defaultPalEntry[i];
    }

    cmap->palette = CreatePalette(pPal);
    LocalFree(pPal);
    cmap->size = n;
    cmap->stale = 0;

    /* Since this is a private colormap of a fix size, we do not need
       a valid hash table, but a dummy one */

    Tcl_InitHashTable(&cmap->refCounts, TCL_ONE_WORD_KEYS);
    return (Colormap)cmap;
}

Colormap BasicTogl::Win32CreateCiColormap() {
    /* Create a colormap with size of CiColormapSize and set all
       entries to black */

    LOGPALETTE logPalette;
    TkWinColormap *cmap = (TkWinColormap *) ckalloc(sizeof(TkWinColormap));

    logPalette.palVersion = 0x300;
    logPalette.palNumEntries = 1;
    logPalette.palPalEntry[0].peRed = 0;
    logPalette.palPalEntry[0].peGreen = 0;
    logPalette.palPalEntry[0].peBlue = 0;
    logPalette.palPalEntry[0].peFlags = 0;

    cmap->palette = CreatePalette(&logPalette);
    cmap->size = CiColormapSize;
    ResizePalette(cmap->palette, cmap->size);  /* sets new entries to black */
    cmap->stale = 0;

    /* Since this is a private colormap of a fix size, we do not need
       a valid hash table, but a dummy one */

    Tcl_InitHashTable(&cmap->refCounts, TCL_ONE_WORD_KEYS);
    return (Colormap)cmap;
}
#endif /*X11*/



/*
 * MakeCurrent
 *
 *   Bind the OpenGL rendering context to the specified
 *   Togl widget.
 */





/* create a TOGL object
   argv[1] is the window file name 
   argc must be >= 1 */
BasicTogl::BasicTogl(TclModule *m) {
interp=m->getInterp();		
}


BasicTogl::~BasicTogl() {
	/*
 getClient()->destroyCamera();
	deleted(typeid(Togl));  
  if(getModule()->windowExists((char *) getParent()))
      getModule()->tickle("destroy %s", (char *) getParent());
*/
  }






#ifdef WIN32
#define TOGL_CLASS_NAME "Togl Class"
static int ToglClassInitialized = 0;
static LRESULT (CALLBACK *tkWinChildProc)(HWND hwnd, UINT message,
				 WPARAM wParam, LPARAM lParam) = NULL;

static LRESULT CALLBACK Win32WinProc( HWND hwnd, UINT message,
				      WPARAM wParam, LPARAM lParam)
{
    LRESULT result;
    BasicTogl *togl = (BasicTogl*) GetWindowLong(hwnd, 0);

    switch( message ){
    case WM_WINDOWPOSCHANGED:
        /* Should be processed by DefWindowProc, otherwise a double buffered
        context is not properly resized when the corresponding window is resized.*/
        break;
    case WM_DESTROY:
      if (togl->getRenderingContext()) {
	if ((togl->getMasterCtxtHandle()) !=togl->getRenderingContext())
	  wglDeleteContext(togl->getRenderingContext());
      }
      if (togl->getDeviceContext()) {
	ReleaseDC(hwnd, togl->getDeviceContext());
      }
      if (togl->isCameraBeingDestroyed()==false)
	delete togl;
      break;
    default:
      if (tkWinChildProc == NULL) {
	WNDCLASS childClass;
	GetClassInfo(Tk_GetHINSTANCE(), TK_WIN_CHILD_CLASS_NAME, &childClass);
	tkWinChildProc = childClass.lpfnWndProc;
      }
      return tkWinChildProc(hwnd, message, wParam, lParam);
    }
    result = DefWindowProc(hwnd, message, wParam, lParam);
    Tcl_ServiceAll();
    return result;
}
#endif /* WIN32 */

/*
 * Togl::MakeWindowExist
 *
 *   Modified version of Tk_MakeWindowExist.
 *   Creates an OpenGL window for the Togl widget.
 */
int BasicTogl::MakeWindowExist() {
   XVisualInfo *visinfo;
   Display *dpy;
   Tk_Window TkWin=getTkWindow();
   TkWindow *winPtr = (TkWindow *)TkWin;
   TkWindow *winPtr2;
   Window parent;
   Colormap cmap;
   Tcl_HashEntry *hPtr;
   int new_flag;
   int scrnum;

#if defined(X11)
   int dummy;
   int attrib_list[1000];
   int attrib_count;
   XSetWindowAttributes swa;
   int attempt;

#define MAX_ATTEMPTS 12
   static int ci_depths[MAX_ATTEMPTS] = {
      8, 4, 2, 1, 12, 16, 8, 4, 2, 1, 12, 16
   };
   static int dbl_flags[MAX_ATTEMPTS] = {
      0, 0, 0, 0,  0,  0, 1, 1, 1, 1,  1,  1
   };
#elif defined(WIN32)
   HWND hwnd, parentWin;
   int pixelformat;
   HINSTANCE hInstance;
   WNDCLASS ToglClass;
   PIXELFORMATDESCRIPTOR pfd;
   XVisualInfo VisInf;
#endif /* X11 */

   dpy = Tk_Display(TkWin);


   if (winPtr->window != None) {
      XDestroyWindow(dpy, winPtr->window);
   } 

#if defined(X11)
   /* Make sure OpenGL's GLX extension supported */
   if (!glXQueryExtension(dpy, &dummy, &dummy)) {
      TCL_ERR(interp, "Togl: X server has no OpenGL GLX extension");
   }

   /* It may take a few tries to get a visual */
   for (attempt=0; attempt<MAX_ATTEMPTS; attempt++) {
      attrib_count = 0;
      attrib_list[attrib_count++] = GLX_USE_GL;
      if (RgbaFlag) {
         /* RGB[A] mode */
         attrib_list[attrib_count++] = GLX_RGBA;
         attrib_list[attrib_count++] = GLX_RED_SIZE;
         attrib_list[attrib_count++] = RgbaRed;
         attrib_list[attrib_count++] = GLX_GREEN_SIZE;
         attrib_list[attrib_count++] = RgbaGreen;
         attrib_list[attrib_count++] = GLX_BLUE_SIZE;
         attrib_list[attrib_count++] = RgbaBlue;
         if (AlphaFlag) {
            attrib_list[attrib_count++] = GLX_ALPHA_SIZE;
            attrib_list[attrib_count++] = AlphaSize;
         }

	 /* for EPS Output */
	 if ( EpsRedMap) free( ( char *)EpsRedMap);
	 if ( EpsGreenMap) free( ( char *)EpsGreenMap);
	 if ( EpsBlueMap) free( ( char *)EpsBlueMap);
	 EpsRedMap = EpsGreenMap = EpsBlueMap = 0;
	 EpsMapSize = 0;
      }
      else {
         /* Color index mode */
         int depth;
         attrib_list[attrib_count++] = GLX_BUFFER_SIZE;
         depth = ci_depths[attempt];
         attrib_list[attrib_count++] = depth;
      }
      if (DepthFlag) {
         attrib_list[attrib_count++] = GLX_DEPTH_SIZE;
         attrib_list[attrib_count++] = DepthSize;
      }
      if (requiredDoubleFlag || dbl_flags[attempt]) {
         attrib_list[attrib_count++] = GLX_DOUBLEBUFFER;
      }
      if (StencilFlag) {
         attrib_list[attrib_count++] = GLX_STENCIL_SIZE;
         attrib_list[attrib_count++] = StencilSize;
      }
      if (AccumFlag) {
         attrib_list[attrib_count++] = GLX_ACCUM_RED_SIZE;
         attrib_list[attrib_count++] = AccumRed;
         attrib_list[attrib_count++] = GLX_ACCUM_GREEN_SIZE;
         attrib_list[attrib_count++] = AccumGreen;
         attrib_list[attrib_count++] = GLX_ACCUM_BLUE_SIZE;
         attrib_list[attrib_count++] = AccumBlue;
         if (AlphaFlag) {
            attrib_list[attrib_count++] = GLX_ACCUM_ALPHA_SIZE;
            attrib_list[attrib_count++] = AccumAlpha;
         }
      }
      if (AuxNumber != 0) {
         attrib_list[attrib_count++] = GLX_AUX_BUFFERS;
         attrib_list[attrib_count++] = AuxNumber;
      }
      

      attrib_list[attrib_count++] = None;

      visinfo = glXChooseVisual( dpy, DefaultScreen(dpy), attrib_list );
      if (visinfo) {
         /* found a GLX visual! */
         break;
      }
   }
   if (visinfo==0) {
      TCL_ERR(interp, "Togl: couldn't get visual");
   }
   /* Create an OpenGL rendering context */
   /* Direct rendering if possible */
   if((masterCtxtHandle)==0) {
	   (masterCtxtHandle)= glXCreateContext(dpy, visinfo, None, GL_TRUE);
	   if ((masterCtxtHandle) == 0)
		   TCL_ERR(interp, "could not create master rendering context");
	   (masterDisplayHandle)=dpy;
	   setGLXContext(masterCtxtHandle);
   } else {
     setGLXContext( glXCreateContext(dpy, visinfo, (masterCtxtHandle), GL_TRUE));
     if (getGLXContext() == 0) {
		   TCL_ERR(interp, "could not create camera rendering context");
	   }
   }

#endif /* X11 */

   /* Find parent of window */
   /* Necessary for creation */
   if ((winPtr->parentPtr == 0) || (winPtr->flags & TK_TOP_LEVEL)) {
      parent = XRootWindow(winPtr->display, winPtr->screenNum);
   }
   else {
      if (winPtr->parentPtr->window == None) {
         Tk_MakeWindowExist((Tk_Window) winPtr->parentPtr);
      }
      parent = winPtr->parentPtr->window;
   }
#ifdef WIN32
   parentWin = Tk_GetHWND(parent);
   hInstance = Tk_GetHINSTANCE();
   if (ToglClassInitialized == 0) {
       ToglClassInitialized = 1;
       ToglClass.style = CS_HREDRAW | CS_VREDRAW;
       ToglClass.cbClsExtra = 0;
       ToglClass.cbWndExtra = 4;   /* to save Togl* */
       ToglClass.hInstance = hInstance;
       ToglClass.hbrBackground = 0;
       ToglClass.lpszMenuName = 0;
       ToglClass.lpszClassName = TOGL_CLASS_NAME;
       ToglClass.lpfnWndProc = Win32WinProc;
       ToglClass.hIcon = 0;
       ToglClass.hCursor = 0;
       if (!RegisterClass(&ToglClass)){
           TCL_ERR(interp, "unable register Togl window class");
       }
   }

   hwnd = CreateWindow(TOGL_CLASS_NAME, 0, WS_CHILD | WS_CLIPCHILDREN
                       | WS_CLIPSIBLINGS, 0, 0, requiredWidth,requiredHeight,
                       parentWin, 0, hInstance, 0);
   SetWindowLong(hwnd, 0, (LRESULT) this);
   SetWindowPos(hwnd, HWND_TOP, 0, 0, 0, 0,
  	            SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE);

   setDeviceContext(GetDC(hwnd));

   pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR);
   pfd.nVersion = 1;
   pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL;
   if (requiredDoubleFlag) {
        pfd.dwFlags |= PFD_DOUBLEBUFFER;
   }
/* The stereo flag is not supported in the current generic OpenGL implementation,
   but may be supported by specific hardware devices */
  /* if (requiredStereoFlag) {
        pfd.dwFlags |= PFD_STEREO;
   } */

   pfd.cColorBits = RgbaRed + RgbaGreen + RgbaBlue;
   pfd.iPixelType = RgbaFlag ? PFD_TYPE_RGBA : PFD_TYPE_COLORINDEX;
/* Alpha bitplanes are not supported in the current generic OpenGL implementation,
   but may be supported by specific hardware devices */
   pfd.cAlphaBits = AlphaFlag ? AlphaSize : 0;
   pfd.cAccumBits = AccumFlag ? (AccumRed + AccumGreen +
                                       AccumBlue +AccumAlpha) : 0;
   pfd.cDepthBits = DepthFlag ? DepthSize : 0;
   pfd.cStencilBits = StencilFlag ? StencilSize : 0;
/* Auxiliary buffers are not supported in the current generic OpenGL implementation,
   but may be supported by specific hardware devices */
   pfd.cAuxBuffers = AuxNumber;
   pfd.iLayerType = PFD_MAIN_PLANE;

   if ( (pixelformat = ChoosePixelFormat(getDeviceContext(), &pfd)) == 0 ) {
        TCL_ERR(interp, "Togl: couldn't choose pixel format");
   }
   if (SetPixelFormat(getDeviceContext(), pixelformat, &pfd) == FALSE) {
        TCL_ERR(interp, "Togl: couldn't choose pixel format");
   }

/* Get the actual pixel format */
   DescribePixelFormat(getDeviceContext(), pixelformat, sizeof(pfd), &pfd);

/* Create an OpenGL rendering context */
   setRenderingContext( wglCreateContext(getDeviceContext()));
   if (!getRenderingContext()) {
        TCL_ERR(interp, "could not create rendering context");
   }
    if ((masterCtxtHandle)==0) {
  (masterCtxtHandle)=getRenderingContext();
	//	(*masterCtxtHandle)=wglCreateContext(getDeviceContext());
   }
   else
     wglShareLists((masterCtxtHandle),getRenderingContext()); 

/* Just for portability, define the simplest visinfo */
   visinfo = &VisInf;
   visinfo->visual = DefaultVisual(dpy, DefaultScreen(dpy));
   visinfo->depth = visinfo->visual->bits_per_rgb;
#endif /*WIN32 */

   /*
    * find a colormap
    */
   scrnum = DefaultScreen(dpy);
   if (RgbaFlag) {
      /* Colormap for RGB mode */
#if defined(X11)
      cmap = get_rgb_colormap( dpy, scrnum, visinfo );
#elif defined(WIN32)
   if (pfd.dwFlags & PFD_NEED_PALETTE) {
          cmap = Win32CreateRgbColormap(pfd);
      }
      else {
          cmap = DefaultColormap(dpy,scrnum);
      }

	  /* for EPS Output */
	  if ( EpsRedMap) free( ( char *)EpsRedMap);
	  if ( EpsGreenMap) free( ( char *)EpsGreenMap);
	  if ( EpsBlueMap) free( ( char *)EpsBlueMap);
	  EpsRedMap = EpsGreenMap = EpsBlueMap = 0;
	  EpsMapSize = 0;
#endif /* X11 */
   }
   else {
      /* Colormap for CI mode */
#ifdef WIN32
      CiColormapSize = 1 << pfd.cColorBits;
      CiColormapSize = CiColormapSize < MAX_CI_COLORMAP_SIZE ?
                             CiColormapSize : MAX_CI_COLORMAP_SIZE;

#endif /* WIN32 */
      if (PrivateCmapFlag) {
         /* need read/write colormap so user can store own color entries */
#if defined(X11)
         cmap = XCreateColormap(dpy, RootWindow(dpy, visinfo->screen),
                                visinfo->visual, AllocAll);
#elif defined(WIN32)
         cmap = Win32CreateCiColormap();
#endif /* X11 */
      }
      else {
         if (visinfo->visual==DefaultVisual(dpy, scrnum)) {
            /* share default/root colormap */
            cmap = DefaultColormap(dpy,scrnum);
         }
         else {
            /* make a new read-only colormap */
            cmap = XCreateColormap(dpy, RootWindow(dpy, visinfo->screen),
                                   visinfo->visual, AllocNone);
         }
      }
   }

   /* Make sure Tk knows to switch to the new colormap when the cursor
    * is over this window when running in color index mode.
    */
   Tk_SetWindowVisual(TkWin, visinfo->visual, visinfo->depth, cmap);
#ifdef WIN32
    /* Install the colormap */
    SelectPalette(getDeviceContext(), ((TkWinColormap *)cmap)->palette, TRUE);
	RealizePalette(getDeviceContext());
#endif /* WIN32 */

   /* for DumpToEpsFile */
   EpsVisual = visinfo;

#if defined(X11)
   swa.colormap = cmap;
   swa.border_pixel = 0;
   swa.event_mask = ALL_EVENTS_MASK;
   winPtr->window = XCreateWindow(dpy, parent,
                                  0, 0,
				  requiredWidth,requiredHeight,
                                  0, visinfo->depth,
                                  InputOutput, visinfo->visual,
                                  CWBorderPixel | CWColormap | CWEventMask,
                                  &swa);

   /* Make sure window manager installs our colormap */
   XSetWMColormapWindows( dpy, winPtr->window, &winPtr->window, 1 );
#elif defined(WIN32)
   winPtr->window = Tk_AttachHWND((Tk_Window)winPtr, hwnd);
#endif /* X11 */
   hPtr = Tcl_CreateHashEntry(&winPtr->dispPtr->winTable,
			      (char *) winPtr->window, &new_flag);
   Tcl_SetHashValue(hPtr, winPtr);

   winPtr->dirtyAtts = 0;
   winPtr->dirtyChanges = 0;
#ifdef TK_USE_INPUT_METHODS
   winPtr->inputContext = 0;
#endif /* TK_USE_INPUT_METHODS */

   if (!(winPtr->flags & TK_TOP_LEVEL)) {
      /*
       * If any siblings higher up in the stacking order have already
       * been created then move this window to its rightful position
       * in the stacking order.
       *
       * NOTE: this code ignores any changes anyone might have made
       * to the sibling and stack_mode field of the window's attributes,
       * so it really isn't safe for these to be manipulated except
       * by calling Tk_RestackWindow.
       */

      for (winPtr2 = winPtr->nextPtr; winPtr2 != 0;
           winPtr2 = winPtr2->nextPtr) {
         if ((winPtr2->window != None) && !(winPtr2->flags & TK_TOP_LEVEL)) {
            XWindowChanges changes;
            changes.sibling = winPtr2->window;
            changes.stack_mode = Below;
            XConfigureWindow(winPtr->display, winPtr->window,
                             CWSibling|CWStackMode, &changes);
            break;
         }
      }

      /*
       * If this window has a different colormap than its parent, add
       * the window to the WM_COLORMAP_WINDOWS property for its top-level.
       */

      if ((winPtr->parentPtr != 0) &&
          (winPtr->atts.colormap != winPtr->parentPtr->atts.colormap)) {
         TkWmAddToColormapWindows(winPtr);
      }
   }

   if (OverlayFlag) {
      if (SetupOverlay()==TCL_ERROR) {
         fprintf(stderr,"Warning: couldn't setup overlay.\n");
         OverlayFlag = 0;
      }
   }

   /*
    * Issue a ConfigureNotify event if there were deferred configuration
    * changes (but skip it if the window is being deleted;  the
    * ConfigureNotify event could cause problems if we're being called
    * from Tk_DestroyWindow under some conditions).
    */
   if ((winPtr->flags & TK_NEED_CONFIG_NOTIFY)
       && !(winPtr->flags & TK_ALREADY_DEAD)){
      XEvent event;

      winPtr->flags &= ~TK_NEED_CONFIG_NOTIFY;

      event.type = ConfigureNotify;
      event.xconfigure.serial = LastKnownRequestProcessed(winPtr->display);
      event.xconfigure.send_event = False;
      event.xconfigure.display = winPtr->display;
      event.xconfigure.event = winPtr->window;
      event.xconfigure.window = winPtr->window;
      event.xconfigure.x = winPtr->changes.x;
      event.xconfigure.y = winPtr->changes.y;
      event.xconfigure.width = winPtr->changes.width;
      event.xconfigure.height = winPtr->changes.height;
      event.xconfigure.border_width = winPtr->changes.border_width;
      if (winPtr->changes.stack_mode == Above) {
         event.xconfigure.above = winPtr->changes.sibling;
      }
      else {
         event.xconfigure.above = None;
      }
      event.xconfigure.override_redirect = winPtr->atts.override_redirect;
      Tk_HandleEvent(&event);
   }

   /* Request the X window to be displayed */
   XMapWindow(dpy, Tk_WindowId(TkWin));
   /* Bind the context to the window and make it the current context. */
   makeCurrent();

#if defined(X11)
   /* Check for a single/double buffering snafu */
   {
      int dbl_flag;
      if (glXGetConfig( dpy, visinfo, GLX_DOUBLEBUFFER, &dbl_flag )) {
         if (requiredDoubleFlag==false && dbl_flag) {
            /* We requested single buffering but had to accept a */
            /* double buffered visual.  Set the GL draw buffer to */
            /* be the front buffer to simulate single buffering. */
            glDrawBuffer( GL_FRONT );
         }
      }
   }
#endif /* X11 */

   /* for EPS Output */
   if ( !RgbaFlag) {
      int index_size;
#if defined(X11)
      int index_bits;
      glGetIntegerv( GL_INDEX_BITS, &index_bits);
      index_size = 1<<index_bits;
#elif defined(WIN32)
      index_size = CiColormapSize;
#endif /* X11 */
      if ( EpsMapSize != index_size) {
         if ( EpsRedMap) free( ( char *)EpsRedMap);
         if ( EpsGreenMap) free( ( char *)EpsGreenMap);
         if ( EpsBlueMap) free( ( char *)EpsBlueMap);
         EpsMapSize = index_size;
         EpsRedMap = ( GLfloat *)calloc( index_size, sizeof( GLfloat));
         EpsGreenMap = ( GLfloat *)calloc( index_size, sizeof( GLfloat));
         EpsBlueMap = ( GLfloat *)calloc( index_size, sizeof( GLfloat));
      }
   }
   return TCL_OK;
}




/*char *Togl::getParent(void) const {
  return Tk_PathName(((TkWindow *) TkWin)->parentPtr);
}*/





#if defined(X11)
/*
 * A replacement for XAllocColor.  This function should never
 * fail to allocate a color.  When XAllocColor fails, we return
 * the nearest matching color.  If we have to allocate many colors
 * this function isn't too efficient; the XQueryColors() could be
 * done just once.
 * Written by Michael Pichler, Brian Paul, Mark Kilgard
 * Input:  dpy - X display
 *         cmap - X colormap
 *         cmapSize - size of colormap
 * In/Out: color - the XColor struct
 * Output:  exact - 1=exact color match, 0=closest match
 */
static void
noFaultXAllocColor( Display *dpy, Colormap cmap, int cmapSize,
                    XColor *color, int *exact )
{
   XColor *ctable, subColor;
   int i, bestmatch;
   double mindist;       /* 3*2^16^2 exceeds long int precision.
                          */

   /* First try just using XAllocColor. */
   if (XAllocColor(dpy, cmap, color)) {
      *exact = 1;
      return;
   }

   /* Retrieve color table entries. */
   /* XXX alloca candidate. */
   ctable = new XColor[cmapSize];
   for (i = 0; i < cmapSize; i++) {
      ctable[i].pixel = i;
   }
   XQueryColors(dpy, cmap, ctable, cmapSize);

   /* Find best match. */
   bestmatch = -1;
   mindist = 0.0;
   for (i = 0; i < cmapSize; i++) {
      double dr = (double) color->red - (double) ctable[i].red;
      double dg = (double) color->green - (double) ctable[i].green;
      double db = (double) color->blue - (double) ctable[i].blue;
      double dist = dr * dr + dg * dg + db * db;
      if (bestmatch < 0 || dist < mindist) {
         bestmatch = i;
         mindist = dist;
      }
   }

   /* Return result. */
   subColor.red = ctable[bestmatch].red;
   subColor.green = ctable[bestmatch].green;
   subColor.blue = ctable[bestmatch].blue;
   delete[] ctable;
   /* Try to allocate the closest match color.  This should only
    * fail if the cell is read/write.  Otherwise, we're incrementing
    * the cell's reference count.
    */
   if (!XAllocColor(dpy, cmap, &subColor)) {
      /* do this to work around a problem reported by Frank Ortega */
      subColor.pixel = (unsigned long) bestmatch;
      subColor.red   = ctable[bestmatch].red;
      subColor.green = ctable[bestmatch].green;
      subColor.blue  = ctable[bestmatch].blue;
      subColor.flags = DoRed | DoGreen | DoBlue;
   }
   *color = subColor;
}

#elif defined(WIN32)



static void Win32FreeColor( const BasicTogl *togl, unsigned long index )
{
    TkWinColormap *cmap = (TkWinColormap *) Tk_Colormap(togl->getTkWindow());
    COLORREF cref;
    UINT count, refCount;
    PALETTEENTRY entry, *entries;
    Tcl_HashEntry *entryPtr;

	if (index >= cmap->size ) {
		panic("Tried to free a color that isn't allocated.");
	}
	GetPaletteEntries(cmap->palette, index, 1, &entry);
	cref = PALETTERGB(entry.peRed, entry.peGreen, entry.peBlue);
	entryPtr = Tcl_FindHashEntry(&cmap->refCounts, (char *) cref);
	if (!entryPtr) {
		panic("Tried to free a color that isn't allocated.");
	}
	refCount = (int) Tcl_GetHashValue(entryPtr) - 1;
	if (refCount == 0) {
		count = cmap->size - index;
		entries = (PALETTEENTRY *) ckalloc(sizeof(PALETTEENTRY)* count);
		GetPaletteEntries(cmap->palette, index+1, count, entries);
		SetPaletteEntries(cmap->palette, index, count, entries);
		SelectPalette(togl->getDeviceContext(), cmap->palette, TRUE);
		RealizePalette(togl->getDeviceContext());
		ckfree((char *) entries);
		cmap->size--;
		Tcl_DeleteHashEntry(entryPtr);
	} else {
		Tcl_SetHashValue(entryPtr, (ClientData)refCount);
	}
}

static void Win32SetColor( const BasicTogl *togl,
                    unsigned long index, float red, float green, float blue )
{
    TkWinColormap *cmap = (TkWinColormap *) Tk_Colormap(togl->getTkWindow());
    PALETTEENTRY entry;

    entry.peRed = (BYTE)( red*255 + .5);
    entry.peGreen = (BYTE)(green*255 + .5);
    entry.peBlue = (BYTE)(blue*255 + .5);
    entry.peFlags = 0;
    SetPaletteEntries(cmap->palette, index, 1, &entry);
	SelectPalette(togl->getDeviceContext(), cmap->palette, TRUE);
	RealizePalette(togl->getDeviceContext());

	/* for EPS output */
    togl->EpsRedMap[index] = (GLfloat)(entry.peRed / 255.0);
    togl->EpsGreenMap[index] = (GLfloat)(entry.peGreen / 255.0);
    togl->EpsBlueMap[index] = (GLfloat)(entry.peBlue / 255.0);
}
#endif /* X11 */



unsigned long BasicTogl::AllocColor(float red, float green, float blue) const {
#if defined(X11)
   XColor xcol;
   int exact;
#endif
   if (RgbaFlag) {
      fprintf(stderr,"Error: Togl_AllocColor illegal in RGBA mode.\n");
      return 0;
   }
   /* TODO: maybe not... */
   if (PrivateCmapFlag) {
      fprintf(stderr,"Error: Togl_FreeColor illegal with private colormap\n");
      return 0;
   }
Tk_Window TkWin=getTkWindow(); 
#if defined(X11)
   xcol.red   = (short) (red   * 65535.0);
   xcol.green = (short) (green * 65535.0);
   xcol.blue  = (short) (blue  * 65535.0);

   noFaultXAllocColor( Tk_Display(TkWin), Tk_Colormap(TkWin),
                       Tk_Visual(TkWin)->map_entries, &xcol, &exact );


   /* for EPS output */
   EpsRedMap[ xcol.pixel] = xcol.red / 65535.0;
   EpsGreenMap[ xcol.pixel] = xcol.green / 65535.0;
   EpsBlueMap[ xcol.pixel] = xcol.blue / 65535.0;

   return xcol.pixel;
#elif defined(WIN32)
   {   
    TkWinColormap *cmap = (TkWinColormap *) Tk_Colormap(TkWin);
    UINT index;
    COLORREF newColor, closeColor;
    PALETTEENTRY entry, closeEntry;
    int newc, refCount;
    Tcl_HashEntry *entryPtr;

    entry.peRed =  (BYTE) (red*255 + .5);
    entry.peGreen = (BYTE) (green*255 + .5);
    entry.peBlue = (BYTE) (blue*255 + .5);
    entry.peFlags = 0;

	/*
	 * Find the nearest existing palette entry.
	 */

    newColor = RGB(entry.peRed, entry.peGreen, entry.peBlue);
    index = GetNearestPaletteIndex(cmap->palette, newColor);
    GetPaletteEntries(cmap->palette, index, 1, &closeEntry);
    closeColor = RGB(closeEntry.peRed, closeEntry.peGreen,  closeEntry.peBlue);

     /*
	 * If this is not a duplicate and colormap is not full, allocate a new entry.
	 */

	if (newColor != closeColor) {
        if (cmap->size == (unsigned) CiColormapSize) {
            entry = closeEntry;
        }
        else {
            cmap->size++;
		    ResizePalette(cmap->palette, cmap->size);
		    index = cmap->size -1;
		    SetPaletteEntries(cmap->palette, index, 1, &entry);
		    SelectPalette(getDeviceContext(), cmap->palette, TRUE);
		    RealizePalette(getDeviceContext());
		}
	}
	newColor = PALETTERGB(entry.peRed, entry.peGreen, entry.peBlue);
	entryPtr = Tcl_CreateHashEntry(&cmap->refCounts, (char *) newColor, &newc);
	if (newc) {
	    refCount = 1;
	} else {
	    refCount = ((int) Tcl_GetHashValue(entryPtr)) + 1;
	}
	Tcl_SetHashValue(entryPtr, (ClientData)refCount);

   /* for EPS output */
    EpsRedMap[index] = (GLfloat)(entry.peRed / 255.0);
    EpsGreenMap[index] = (GLfloat)(entry.peGreen / 255.0);
    EpsBlueMap[index] = (GLfloat)(entry.peBlue / 255.0);
   
   return index;
   }
#endif /* X11 */
}



void BasicTogl::FreeColor(unsigned long pixel) const {
   if (RgbaFlag) {
      fprintf(stderr,"Error: Togl_AllocColor illegal in RGBA mode.\n");
      return;
   }
   /* TODO: maybe not... */
   if (PrivateCmapFlag) {
      fprintf(stderr,"Error: Togl_FreeColor illegal with private colormap\n");
      return;
   }
Tk_Window TkWin= getTkWindow(); 
#if defined(X11)
   XFreeColors( Tk_Display(TkWin), Tk_Colormap(TkWin),
                &pixel, 1, 0 );
#elif defined(WIN32)
   Win32FreeColor(this,pixel);
#endif /* X11 */
}



void BasicTogl::SetColor(unsigned long index, float red, float green, float blue)
  const {
#if defined(X11)
   XColor xcol;
#endif

   if (RgbaFlag) {
      fprintf(stderr,"Error: Togl_AllocColor illegal in RGBA mode.\n");
      return;
   }
   if (!PrivateCmapFlag) {
      fprintf(stderr,"Error: Togl_SetColor requires a private colormap\n");
      return;
   }
Tk_Window TkWin= getTkWindow(); 
#if defined(X11)
   xcol.pixel = index;
   xcol.red   = (short) (red   * 65535.0);
   xcol.green = (short) (green * 65535.0);
   xcol.blue  = (short) (blue  * 65535.0);
   xcol.flags = DoRed | DoGreen | DoBlue;

   XStoreColor( Tk_Display(TkWin), Tk_Colormap(TkWin), &xcol );

   /* for EPS output */
   EpsRedMap[ xcol.pixel] = xcol.red / 65535.0;
   EpsGreenMap[ xcol.pixel] = xcol.green / 65535.0;
   EpsBlueMap[ xcol.pixel] = xcol.blue / 65535.0;

#elif defined(WIN32)
   Win32SetColor( this, index, red, green, blue );
#endif /* X11 */
}



#if defined(WIN32)
#include <graphics3D/tkFont8.3.h>

/*
 * The following structure represents Windows' implementation of a font.
 */

typedef struct WinFont {
    TkFont font;		/* Stuff used by generic font package.  Must
				 * be first in structure. */
    HFONT hFont;		/* Windows information about font. */
    HWND hwnd;			/* Toplevel window of application that owns
				 * this font, used for getting HDC. */
    int widths[256];		/* widths of first 256 chars in this font. */
} WinFont;
#endif /* WIN32 */


#define MAX_FONTS 1000
static GLuint ListBase[MAX_FONTS];
static GLuint ListCount[MAX_FONTS];



/*
 * Load the named bitmap font as a sequence of bitmaps in a display list.
 * fontname may be one of the predefined fonts like TOGL_BITMAP_8_BY_13
 * or an X font name, or a Windows font name, etc.
 */
GLuint BasicTogl::LoadBitmapFont(const char *fontname) const {
   int FirstTime = 1;
#if defined(X11)
   XFontStruct *fontinfo;
#elif defined(WIN32)
   WinFont *winfont;
   void *oldFont;
   TEXTMETRIC tm;
#endif /* X11 */
   int first, last, count;
   GLuint fontbase;
   const char *name;

   /* Initialize the ListBase and ListCount arrays */
   if (FirstTime) {
      int i;
      for (i=0;i<MAX_FONTS;i++) {
         ListBase[i] = ListCount[i] = 0;
      }
      FirstTime = 0;
   }

   /*
    * This method of selecting X fonts according to a TOGL_ font name
    * is a kludge.  To be fixed when I find time...
    */
   if (fontname==TOGL_BITMAP_8_BY_13) {
      name = "8x13";
   }
   else if (fontname==TOGL_BITMAP_9_BY_15) {
      name = "9x15";
   }
   else if (fontname==TOGL_BITMAP_TIMES_ROMAN_10) {
      name = "-adobe-times-medium-r-normal--10-100-75-75-p-54-iso8859-1";
   }
   else if (fontname==TOGL_BITMAP_TIMES_ROMAN_24) {
      name = "-adobe-times-medium-r-normal--24-240-75-75-p-124-iso8859-1";
   }
   else if (fontname==TOGL_BITMAP_HELVETICA_10) {
      name = "-adobe-helvetica-medium-r-normal--10-100-75-75-p-57-iso8859-1";
   }
   else if (fontname==TOGL_BITMAP_HELVETICA_12) {
      name = "-adobe-helvetica-medium-r-normal--12-120-75-75-p-67-iso8859-1";
   }
   else if (fontname==TOGL_BITMAP_HELVETICA_18) {
      name = "-adobe-helvetica-medium-r-normal--18-180-75-75-p-98-iso8859-1";
   }
   else if (!fontname) {
      name = DEFAULT_FONTNAME;
   }
   else {
      name = (const char *) fontname;
   }

   assert( (long int) name );
Tk_Window TkWin= getTkWindow(); 

#if defined(X11)
   fontinfo = XLoadQueryFont( Tk_Display(TkWin), name );
   if (!fontinfo) {
      return 0;
   }

   first = fontinfo->min_char_or_byte2;
   last = fontinfo->max_char_or_byte2;
#elif defined(WIN32)
   winfont = (WinFont*) Tk_GetFont(interp, TkWin, name);
   if (!winfont) {
      return 0;
   }
HDC tglGLHdc=getDeviceContext();
   oldFont = SelectObject(tglGLHdc, winfont->hFont);
   GetTextMetrics(tglGLHdc, &tm);
   first = tm.tmFirstChar;
   last = tm.tmLastChar;
#endif /* X11 */
   count = last-first+1;

   fontbase = glGenLists( (GLuint) (last+1) );
   if (fontbase==0) {
#ifdef WIN32
      SelectObject(tglGLHdc, oldFont);
      Tk_FreeFont((Tk_Font) winfont);
#endif /* WIN32 */
      return 0;
   }
#if defined(WIN32)
   wglUseFontBitmaps(tglGLHdc, first, count, (int) fontbase+first );
   SelectObject(tglGLHdc, oldFont);
   Tk_FreeFont((Tk_Font) winfont);
#elif defined(X11)
   glXUseXFont( fontinfo->fid, first, count, (int) fontbase+first );
#endif

   /* Record the list base and number of display lists
    * for UnloadBitmapFont().
    */
   {
      int i;
      for (i=0;i<MAX_FONTS;i++) {
         if (ListBase[i]==0) {
            ListBase[i] = fontbase;
            ListCount[i] = last+1;
            break;
         }
      }
   }

   return fontbase;
}



/*
 * Release the display lists which were generated by LoadBitmapFont().
 */
void BasicTogl::UnloadBitmapFont(GLuint fontbase) const {
  int i;

  for (i=0;i<MAX_FONTS;i++) {
    if (ListBase[i]==fontbase) {
      glDeleteLists(ListBase[i], ListCount[i]);
      ListBase[i] = ListCount[i] = 0;
      return;
    }
  }
}


/*
 * Overlay functions
 */


void BasicTogl::UseLayer(int layer) {
   if (OverlayWindow) {
      if (layer==TOGL_OVERLAY) {
#if defined(WIN32)
		  HDC tglGLHdc=getDeviceContext();
         int res = wglMakeCurrent(tglGLHdc, tglGLOverlayHglrc);
         assert(res == TRUE);
#elif defined(X11)
	Tk_Window TkWin=getTkWindow(); 
         glXMakeCurrent( Tk_Display(TkWin), OverlayWindow, OverlayCtx );
#endif /*WIN32 */
      }
      else if (layer==TOGL_NORMAL) {
#if defined(WIN32)
	int res = wglMakeCurrent(getDeviceContext(), getRenderingContext());
	assert(res == TRUE);
#elif defined(X11)
	Tk_Window TkWin=getTkWindow(); 
	glXMakeCurrent( Tk_Display(TkWin), Tk_WindowId(TkWin), getGLXContext());
#endif /* WIN32 */
      }
      else {
         /* error */
      }
   }
}



/*
 * Do all the setup for overlay planes
 * Return:   TCL_OK or TCL_ERROR
 */
int BasicTogl::SetupOverlay() {
#if defined(X11)

#ifdef GLX_TRANSPARENT_TYPE_EXT
   static int ovAttributeList[] = {
      GLX_BUFFER_SIZE, 2,
      GLX_LEVEL, 1,
      GLX_TRANSPARENT_TYPE_EXT, GLX_TRANSPARENT_INDEX_EXT,
      None
   };
#else
   static int ovAttributeList[] = {
      GLX_BUFFER_SIZE, 2,
      GLX_LEVEL, 1,
      None
   };
#endif

   Display *dpy;
   XVisualInfo *visinfo;
   Tk_Window TkWin=getTkWindow(); 
   TkWindow *winPtr = (TkWindow *) TkWin;
   XSetWindowAttributes swa;
   Tcl_HashEntry *hPtr;
   int new_flag;

   dpy = Tk_Display(TkWin);

   visinfo = glXChooseVisual( dpy, DefaultScreen(dpy), ovAttributeList );
   if (!visinfo){
      Tcl_AppendResult(interp,Tk_PathName(winPtr),
                       ": No suitable overlay index visual available",
                       (char *) 0);
      OverlayCtx = 0;
      OverlayWindow = 0;
      OverlayCmap = 0;
      return TCL_ERROR;
   }

#ifdef GLX_TRANSPARENT_INDEX_EXT
   {
      int fail = glXGetConfig(dpy, visinfo,GLX_TRANSPARENT_INDEX_VALUE_EXT,
                              &OverlayTransparentPixel);
      if (fail)
         OverlayTransparentPixel=0; /* maybe, maybe ... */
   }
#else
   OverlayTransparentPixel=0; /* maybe, maybe ... */
#endif

   OverlayCtx = glXCreateContext( dpy, visinfo, None, GL_TRUE );

   swa.colormap = XCreateColormap( dpy, RootWindow(dpy, visinfo->screen),
                                   visinfo->visual, AllocNone );
   OverlayCmap = swa.colormap;

   swa.border_pixel = 0;
   swa.event_mask = ALL_EVENTS_MASK;
   OverlayWindow = XCreateWindow( dpy, Tk_WindowId(TkWin), 0, 0,
				 requiredWidth,requiredHeight, 0,
				  visinfo->depth, InputOutput,
				  visinfo->visual,
				  CWBorderPixel|CWColormap|CWEventMask,
				  &swa );

   hPtr = Tcl_CreateHashEntry( &winPtr->dispPtr->winTable,
                               (char *) OverlayWindow, &new_flag );
   Tcl_SetHashValue( hPtr, winPtr );

/*   XMapWindow( dpy, OverlayWindow );*/
   OverlayIsMapped = 0;

   /* Make sure window manager installs our colormap */
   XSetWMColormapWindows( dpy, OverlayWindow, &OverlayWindow, 1 );

   return TCL_OK;

#elif defined(WIN32)       /* not yet implemented on Windows*/

   return TCL_ERROR;

#endif /* X11 */
}



#if defined(X11)  /* not yet implemented on Windows*/
void BasicTogl::ShowOverlay(void) {
Tk_Window TkWin=getTkWindow(); 
  if (OverlayWindow) {
    XMapWindow( Tk_Display(TkWin), OverlayWindow );
    XInstallColormap(Tk_Display(TkWin), OverlayCmap);
    OverlayIsMapped = 1;
  }
}
#endif /* X11 */



void BasicTogl::HideOverlay(void) {
  if (OverlayWindow && OverlayIsMapped) {
	  Tk_Window TkWin= getTkWindow(); 
    XUnmapWindow( Tk_Display(TkWin), OverlayWindow );
    OverlayIsMapped=0;
  }
}




#if defined(X11) /* not yet implemented on Windows*/
unsigned long BasicTogl::AllocColorOverlay(float red, float green, float blue)
  const {
   if (OverlayFlag && OverlayCmap) {
	   Tk_Window TkWin=getTkWindow(); 
      XColor xcol;
      xcol.red   = (short) (red* 65535.0);
      xcol.green = (short) (green* 65535.0);
      xcol.blue  = (short) (blue* 65535.0);
      if (!XAllocColor(Tk_Display(TkWin), OverlayCmap, &xcol))
         return (unsigned long) -1;
      return xcol.pixel;
   }
   else {
      return (unsigned long) -1;
   }
}


void BasicTogl::FreeColorOverlay(unsigned long pixel) const {
   if (OverlayFlag && OverlayCmap) {
	   Tk_Window TkWin=getTkWindow(); 
      XFreeColors( Tk_Display(TkWin), OverlayCmap, &pixel, 1, 0 );
   }
}
#endif /* X11 */



/*
 * User client data
 */

#ifdef MESA_COLOR_HACK
/*
 * Let's know how many free colors do we have
 */
#if 0
static unsigned char rojo[] = { 4, 39, 74, 110, 145, 181, 216, 251},
                     verde[] = { 4, 39, 74, 110, 145, 181, 216, 251},
		     azul[] = { 4, 39, 74, 110, 145, 181, 216, 251};

unsigned char rojo[] = { 4, 36, 72, 109, 145, 182, 218, 251},
              verde[] = { 4, 36, 72, 109, 145, 182, 218, 251},
              azul[] = { 4, 36, 72, 109, 145, 182, 218, 251};
              azul[] = { 0, 85, 170, 255};
#endif

#define RLEVELS     5
#define GLEVELS     9
#define BLEVELS     5

/* to free dithered_rgb_colormap pixels allocated by Mesa */
BasicTogl::BasicToglMesaUsedPixelCells = 0;
BasicTogl::ToglMesaUsedFreeCells = 0;

static int get_free_color_cells(Display *display, int screen,
				Colormap colormap)
{
   if ( !ToglMesaUsedPixelCells) {
      XColor xcol;
      int i;
      int colorsfailed, ncolors = XDisplayCells( display, screen);

      long r, g, b;

      ToglMesaUsedPixelCells = ( unsigned long *)calloc( ncolors, sizeof( unsigned long));

      /* Allocate X colors and initialize color_table[], red_table[], etc */
      /* de Mesa 2.1: xmesa1.c setup_dithered_(...) */
      i = colorsfailed = 0;
      for (r = 0; r < RLEVELS; r++)
         for (g = 0; g < GLEVELS; g++)
            for (b = 0; b < BLEVELS; b++) {
               int exact;
               xcol.red   = ( r*65535)/(RLEVELS-1);
               xcol.green = ( g*65535)/(GLEVELS-1);
               xcol.blue  = ( b*65535)/(BLEVELS-1);
               noFaultXAllocColor( display, colormap, ncolors,
                                   &xcol, &exact );
               ToglMesaUsedPixelCells[ i++] = xcol.pixel;
               if (!exact) {
                  colorsfailed++;
               }
            }
      ToglMesaUsedFreeCells = i;

      XFreeColors( display, colormap, ToglMesaUsedPixelCells,
                   ToglMesaUsedFreeCells, 0x00000000);
   }
   return ToglMesaUsedFreeCells;
}


static void free_default_color_cells( Display *display, Colormap colormap)
{
   if ( ToglMesaUsedPixelCells) {
      XFreeColors( display, colormap, ToglMesaUsedPixelCells,
                   ToglMesaUsedFreeCells, 0x00000000);
      free( ( char *)ToglMesaUsedPixelCells);
      ToglMesaUsedPixelCells = 0;
      ToglMesaUsedFreeCells = 0;
   }
}
#endif


/*
 * Generate EPS file.
 * Contributed by Miguel A. De Riera Pasenau (miguel@DALILA.UPC.ES)
 */

/* Function that creates a EPS File from a created pixmap on the current
 * context.
 * Based on the code from Copyright (c) Mark J. Kilgard, 1996.
 * Parameters: name_file, b&w / Color flag, redraw function.
 * The redraw function is needed in order to draw things into the new
 * created pixmap.
 */

/* Copyright (c) Mark J. Kilgard, 1996. */

static GLvoid *grabPixels(int inColor, unsigned int width, unsigned int height)
{
   GLvoid *buffer;
   GLint swapbytes, lsbfirst, rowlength;
   GLint skiprows, skippixels, alignment;
   GLenum format;
   unsigned int size;

   if (inColor) {
      format = GL_RGB;
      size = width * height * 3;
   } else {
      format = GL_LUMINANCE;
      size = width * height * 1;
   }

   buffer = (GLvoid *) new char[size];
   if (buffer == 0)
      return 0;

   /* Save current modes. */
   glGetIntegerv(GL_PACK_SWAP_BYTES, &swapbytes);
   glGetIntegerv(GL_PACK_LSB_FIRST, &lsbfirst);
   glGetIntegerv(GL_PACK_ROW_LENGTH, &rowlength);
   glGetIntegerv(GL_PACK_SKIP_ROWS, &skiprows);
   glGetIntegerv(GL_PACK_SKIP_PIXELS, &skippixels);
   glGetIntegerv(GL_PACK_ALIGNMENT, &alignment);
   /* Little endian machines (DEC Alpha for example) could
      benefit from setting GL_PACK_LSB_FIRST to GL_TRUE
      instead of GL_FALSE, but this would require changing the
      generated bitmaps too. */
   glPixelStorei(GL_PACK_SWAP_BYTES, GL_FALSE);
   glPixelStorei(GL_PACK_LSB_FIRST, GL_FALSE);
   glPixelStorei(GL_PACK_ROW_LENGTH, 0);
   glPixelStorei(GL_PACK_SKIP_ROWS, 0);
   glPixelStorei(GL_PACK_SKIP_PIXELS, 0);
   glPixelStorei(GL_PACK_ALIGNMENT, 1);

   /* Actually read the pixels. */
   glReadPixels(0, 0, width, height, format,
                GL_UNSIGNED_BYTE, (GLvoid *) buffer);

   /* Restore saved modes. */
   glPixelStorei(GL_PACK_SWAP_BYTES, swapbytes);
   glPixelStorei(GL_PACK_LSB_FIRST, lsbfirst);
   glPixelStorei(GL_PACK_ROW_LENGTH, rowlength);
   glPixelStorei(GL_PACK_SKIP_ROWS, skiprows);
   glPixelStorei(GL_PACK_SKIP_PIXELS, skippixels);
   glPixelStorei(GL_PACK_ALIGNMENT, alignment);
   return buffer;
}


static int generateEPS(const char *filename, int inColor,
                       unsigned int width, unsigned int height)
{
   FILE *fp;
   GLvoid *pixels;
   unsigned char *curpix;
   unsigned int components, i;
   int pos;
   unsigned char bitpixel;

   pixels = grabPixels(inColor, width, height);
   if (pixels == 0)
      return 1;
   if (inColor)
      components = 3;     /* Red, green, blue. */
   else
      components = 1;     /* Luminance. */

   fp = fopen(filename, "w");
   if (fp == 0) {
      return 2;
   }
   fprintf(fp, "%%!PS-Adobe-2.0 EPSF-1.2\n");
   fprintf(fp, "%%%%Creator: OpenGL pixmap render output\n");
   fprintf(fp, "%%%%BoundingBox: 0 0 %d %d\n", width, height);
   fprintf(fp, "%%%%EndComments\n");

   i = ((( width * height) + 7) / 8 ) / 40; /* # of lines, 40 bytes per line */
   fprintf(fp, "%%%%BeginPreview: %d %d %d %d\n%%", width, height, 1, i);
   pos = 0;
   curpix = ( unsigned char *)pixels;
   for ( i = 0; i < width * height * components; ) {
      bitpixel = 0;
      if ( inColor) {
         double pix = 0.0;
         pix = 0.30 * ( double)curpix[ i++] + 0.59 * ( double)curpix[ i++] + 0.11 * ( double)curpix[ i++];
         if ( pix > 127.0) bitpixel |= 0x80;
         pix = 0.30 * ( double)curpix[ i++] + 0.59 * ( double)curpix[ i++] + 0.11 * ( double)curpix[ i++];
         if ( pix > 127.0) bitpixel |= 0x40;
         pix = 0.30 * ( double)curpix[ i++] + 0.59 * ( double)curpix[ i++] + 0.11 * ( double)curpix[ i++];
         if ( pix > 127.0) bitpixel |= 0x20;
         pix = 0.30 * ( double)curpix[ i++] + 0.59 * ( double)curpix[ i++] + 0.11 * ( double)curpix[ i++];
         if ( pix > 127.0) bitpixel |= 0x10;
         pix = 0.30 * ( double)curpix[ i++] + 0.59 * ( double)curpix[ i++] + 0.11 * ( double)curpix[ i++];
         if ( pix > 127.0) bitpixel |= 0x08;
         pix = 0.30 * ( double)curpix[ i++] + 0.59 * ( double)curpix[ i++] + 0.11 * ( double)curpix[ i++];
         if ( pix > 127.0) bitpixel |= 0x04;
         pix = 0.30 * ( double)curpix[ i++] + 0.59 * ( double)curpix[ i++] + 0.11 * ( double)curpix[ i++];
         if ( pix > 127.0) bitpixel |= 0x02;
         pix = 0.30 * ( double)curpix[ i++] + 0.59 * ( double)curpix[ i++] + 0.11 * ( double)curpix[ i++];
         if ( pix > 127.0) bitpixel |= 0x01;
      }
      else {
         if ( curpix[ i++] > 0x7f) bitpixel |= 0x80;
         if ( curpix[ i++] > 0x7f) bitpixel |= 0x40;
         if ( curpix[ i++] > 0x7f) bitpixel |= 0x20;
         if ( curpix[ i++] > 0x7f) bitpixel |= 0x10;
         if ( curpix[ i++] > 0x7f) bitpixel |= 0x08;
         if ( curpix[ i++] > 0x7f) bitpixel |= 0x04;
         if ( curpix[ i++] > 0x7f) bitpixel |= 0x02;
         if ( curpix[ i++] > 0x7f) bitpixel |= 0x01;
      }
      fprintf(fp, "%02hx", bitpixel);
      if (++pos >= 40) {
         fprintf(fp, "\n%%");
         pos = 0;
      }
   }
   if (pos)
      fprintf(fp, "\n%%%%EndPreview\n");
   else
      fprintf(fp, "%%EndPreview\n");

   fprintf(fp, "gsave\n");
   fprintf(fp, "/bwproc {\n");
   fprintf(fp, "    rgbproc\n");
   fprintf(fp, "    dup length 3 idiv string 0 3 0\n");
   fprintf(fp, "    5 -1 roll {\n");
   fprintf(fp, "    add 2 1 roll 1 sub dup 0 eq\n");
   fprintf(fp, "    { pop 3 idiv 3 -1 roll dup 4 -1 roll dup\n");
   fprintf(fp, "        3 1 roll 5 -1 roll put 1 add 3 0 }\n");
   fprintf(fp, "    { 2 1 roll } ifelse\n");
   fprintf(fp, "    } forall\n");
   fprintf(fp, "    pop pop pop\n");
   fprintf(fp, "} def\n");
   fprintf(fp, "systemdict /colorimage known not {\n");
   fprintf(fp, "    /colorimage {\n");
   fprintf(fp, "        pop\n");
   fprintf(fp, "        pop\n");
   fprintf(fp, "        /rgbproc exch def\n");
   fprintf(fp, "        { bwproc } image\n");
   fprintf(fp, "    } def\n");
   fprintf(fp, "} if\n");
   fprintf(fp, "/picstr %d string def\n", width * components);
   fprintf(fp, "%d %d scale\n", width, height);
   fprintf(fp, "%d %d %d\n", width, height, 8);
   fprintf(fp, "[%d 0 0 %d 0 0]\n", width, height);
   fprintf(fp, "{currentfile picstr readhexstring pop}\n");
   fprintf(fp, "false %d\n", components);
   fprintf(fp, "colorimage\n");

   curpix = (unsigned char *) pixels;
   pos = 0;
   for (i = width * height * components; i > 0; i--) {
      fprintf(fp, "%02hx", *curpix++);
      if (++pos >= 40) {
	 fprintf(fp, "\n");
	 pos = 0;
      }
   }
   if (pos)
      fprintf(fp, "\n");

   fprintf(fp, "grestore\n");
   delete[] (unsigned char *) pixels;
   fclose(fp);
   return 0;
}


/* int DumpToEpsFile( const Togl *togl, const char *filename,
                      int inColor, void (*user_redraw)(void)) */
/* changed by GG */
int BasicTogl::DumpToEpsFile(const char *filename, int inColor,
			void (*user_redraw)( const BasicTogl *)) const {
   int using_mesa = 0;
#if 0
   Pixmap eps_pixmap;
   GLXPixmap eps_glxpixmap;
   XVisualInfo *vi = EpsVisual;
   Window win = Tk_WindowId(TkWin);
#endif
   int retval;
#if defined X11
 Tk_Window TkWin=getTkWindow();
   Display *dpy = Tk_Display(TkWin);
   int scrnum = Tk_ScreenNumber(TkWin);
#endif
   unsigned int wth = getWidth() , hght = getHeight(); 

#if defined(X11)
   if (strstr(glXQueryServerString( dpy, scrnum, GLX_VERSION ), "Mesa"))
      using_mesa = 1;
   else
#endif /* X11 */
      using_mesa = 0;
   /* I don't use Pixmap do drawn into, because the code should link
    * with Mesa libraries and OpenGL libraries, and the which library
    * we use at run time should not matter, but the name of the calls
    * differs one from another:
    * MesaGl: glXCreateGLXPixmapMESA( dpy, vi, eps_pixmap, Tk_Colormap(TkWin))
    * OpenGl: glXCreateGLXPixmap( dpy, vi, eps_pixmap);
    *
    * instead of this I read direct from back buffer of the screeen.
    */
#if 0
   eps_pixmap = XCreatePixmap( dpy, win, wth, hght, vi->depth);
   if ( using_mesa)
      eps_glxpixmap = glXCreateGLXPixmapMESA( dpy, vi, eps_pixmap, 
					      Tk_Colormap(TkWin));
   else
      eps_glxpixmap = glXCreateGLXPixmap( dpy, vi, eps_pixmap);

   glXMakeCurrent( dpy, eps_glxpixmap, GlCtx);
   user_redraw();
#endif
   if ( !RgbaFlag) {

#if defined(WIN32)
/* Due to the lack of a unique inverse mapping from the frame buffer to
   the logical palette we need a translation map from the complete
   logical palette. */
       {
           int n, i;
		   Tk_Window TkWin=getTkWindow();
           TkWinColormap *cmap = (TkWinColormap *)Tk_Colormap(TkWin);
           LPPALETTEENTRY entry = new PALETTEENTRY[EpsMapSize];
           n = GetPaletteEntries(cmap->palette, 0, EpsMapSize, entry);
           for (i=0; i<n; i++) {
               EpsRedMap[i] = (GLfloat)(entry[i].peRed / 255.0);
               EpsGreenMap[i] = (GLfloat)(entry[i].peGreen / 255.0);
               EpsBlueMap[i] = (GLfloat)(entry[i].peBlue / 255.0);
           }
           delete[] entry;
       }
#endif /* WIN32 */

      glPixelMapfv( GL_PIXEL_MAP_I_TO_R, EpsMapSize, EpsRedMap);
      glPixelMapfv( GL_PIXEL_MAP_I_TO_G, EpsMapSize, EpsGreenMap);
      glPixelMapfv( GL_PIXEL_MAP_I_TO_B, EpsMapSize, EpsBlueMap);
   }
   /*  user_redraw(); */
   user_redraw(this);  /* changed by GG */
   /* glReadBuffer( GL_FRONT); */
   /* by default it read GL_BACK in double buffer mode*/
   glFlush();
   retval = generateEPS( filename, inColor, wth, hght);
#if 0
   glXMakeCurrent( dpy, win, GlCtx );
   glXDestroyGLXPixmap( dpy, eps_glxpixmap);
   XFreePixmap( dpy, eps_pixmap);
#endif
   return retval;
}




/*
 * It's possible to change with this function or in a script some
 * options like RGBA - ColorIndex ; Z-buffer and so on
 */
int BasicTogl::configure(int argc, char *argv[], int flags) {
  
  int oldRgbaFlag    = RgbaFlag;
  int oldRgbaRed     = RgbaRed;
  int oldRgbaGreen   = RgbaGreen;
  int oldRgbaBlue    = RgbaBlue;
  bool oldDoubleFlag  = requiredDoubleFlag;
  int oldDepthFlag   = DepthFlag;
  int oldDepthSize   = DepthSize;
  int oldAccumFlag   = AccumFlag;
  int oldAccumRed    = AccumRed;
  int oldAccumGreen  = AccumGreen;
  int oldAccumBlue   = AccumBlue;
  int oldAccumAlpha  = AccumAlpha;
  int oldAlphaFlag   = AlphaFlag;
  int oldAlphaSize   = AlphaSize;
  int oldStencilFlag = StencilFlag;
  int oldStencilSize = StencilSize;
  int oldAuxNumber   = AuxNumber;


Tk_Window TkWin=getTkWindow();
#if TK_MAJOR_VERSION==8 && TK_MINOR_VERSION>=4

 if (Tk_ConfigureWidget(interp, TkWin, BasicTogl::configSpecs,
			 argc, (const char **)argv, (char *) this, flags) == TCL_ERROR) {
    return(TCL_ERROR);
  }
#else
  if (Tk_ConfigureWidget(interp, TkWin, BasicTogl::configSpecs,
			 argc, (char **)argv, (char *) this, flags) == TCL_ERROR) {
    return(TCL_ERROR);
  }
#endif
 
  
  Tk_GeometryRequest(TkWin, requiredWidth,requiredHeight);
  
  if (RgbaFlag != oldRgbaFlag
      || RgbaRed != oldRgbaRed
      || RgbaGreen != oldRgbaGreen
      || RgbaBlue != oldRgbaBlue
      || requiredDoubleFlag != oldDoubleFlag
      || DepthFlag != oldDepthFlag
      || DepthSize != oldDepthSize
      || AccumFlag != oldAccumFlag
      || AccumRed != oldAccumRed
      || AccumGreen != oldAccumGreen
      || AccumBlue != oldAccumBlue
      || AccumAlpha != oldAccumAlpha
      || AlphaFlag != oldAlphaFlag
      || AlphaSize != oldAlphaSize
      || StencilFlag != oldStencilFlag
      || StencilSize != oldStencilSize
      || AuxNumber != oldAuxNumber) {
#ifdef MESA_COLOR_HACK
    free_default_color_cells( Tk_Display(TkWin),
			      Tk_Colormap(TkWin) );
#endif
    /* Have to recreate the window and GLX context */
    if (MakeWindowExist()==TCL_ERROR) {
      return TCL_ERROR;
    }
  }
  return TCL_OK;
}





#ifndef WIN32
struct GLContextData {
  GLXContext theContext;
  Display *theDisplay; 
};
#endif
/*
 *
 destroy togl when all 3D rendering has been performed 
 *
 */
void BasicTogl::destroyContext(ClientData clientData ) {

#ifndef WIN32
  struct GLContextData *c = (struct GLContextData *) clientData;
  
  glXDestroyContext(c->theDisplay, c->theContext);
  
  /* Set current context to none since it might still be
     set to this togl window */
  glXMakeCurrent(c->theDisplay, None, 0);
  delete c;
  
#endif
}





/*][************************************************************************/


} // namespace yav

/*]*************************************************************************/
