#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "modules/TclRangeInformation.h"
#pragma implementation "modules/TclFuzzyCMeans.h"
#pragma implementation "modules/TclExpectationMaximization.h"

#include <modules/TclModule.h>
#include <modules/TclObjectKind.h>
#include <modules/TclGenericHistogram.h>
#include <modules/TclMatrix4D.h>
#include <modules/TclRangeInformation.h>
#include <modules/TclConstraint.h>
#include <modules/TclFuzzyCMeans.h>
#include <modules/TclExpectationMaximization.h>
#include <modules/TclProbabilityDistribution.h>
#include <modules/togl.h>

#ifdef EMBEDED_TCL_SCRIPTS
#include "tclModules.tcl.h" //generated by tcl2h
#endif


namespace yav {


void Tclmodule_Init(TclModule *m) {
  
  TclObjectKind *kind;
  m->newCommand("module", TclModule::Module, "<-list | -load name=%s | -delete name=%s | -path >",
		"Get loaded modules list / dynamicaly load a module if the associated .so file can be found in TCL_MODULES or LD_LIBRARY_PATH / delete a dynamically loaded module / return the module search path");
  m->newCommand("shell", TclModule::Shell, "-path",
		"Return TCL shell search path");
  m->newCommand("isModuleLoaded", TclModule::IsModuleLoaded, "name=%s",
		"Check whether the module name has been loaded");
  m->newCommand("isModuleAvail", TclModule::IsModuleAvailable, "name=%s",
		"Check whether the library associated to a module is available and the module can be loaded");
  m->newCommand("help", TclModule::Help,
		"[module=%s | function=%s [synopsis | description | option=%s <synopsis | description>]]",
		"Return a list of modules or a list of functions of the given module or the description of the given function. For a function, only synopsis, description, or a specific option can be selected.");
  m->newCommand("limitedVersion", TclModule::LimitedVersion,
		"", "Return true (resp. false) if LIMITED_VERSION flag is on (resp. off)");
  m->newCommand("functions", TclModule::Functions, "module=%s",
		"Return the given module's function list");
  m->newCommand("options", TclModule::Options, "function=%s",
		"Return given function options");
  m->newCommand("loadSource", TclModule::LoadSource, "name=%s",
		"Look for file name (using the YAV_SHELLS variable) and evaluate it");
  m->newCommand("script", TclModule::StartScript,
		"name=%s [arg1=%s [arg2=%s [...]]]",
		"Open file name and evaluate all command lines. Arguments can be transmitted in the script command line and are stored as arg1, arg2, ..., argn variables.");
  m->newCommand("end", TclModule::EndScript, 0, "end a running script");
  m->newCommand("setDisplayFunction", TclModule::SetDisplayMessageFunction,
		"<default|name=%s>",
		"Display message function used for description. Set the tcl function to call when 'display' is invoked. A default function is used if no function name is provided.");
  m->newCommand("setEvalFunction", TclModule::SetEvalFunction, "name=%s",
		"Set the tcl function to call when 'evaluate' is invoked. A default function is used if no function name is provided.");
  m->newCommand("flushEvents", TclModule::FlushEvents, 0, "Flush all TCL events");
  m->newCommand("mainLoop", TclModule::MainLoop, 0, "Loop on TCL events");
  m->newCommand("activateActivables", TclModule::ActivateActivables, 0, "Loop on TCL events until no activable objects are activated");
  m->newCommand("evaluate", TclModule::Evaluate, "%s", "Evaluate given string");
  m->newCommand("display", TclModule::Display, "%s", "Display given string");
  m->newCommand("exit", TclModule::Exit, 0,
		"Call any module cleanup procedures and exit");

  m->newCommand("new", TclObjectKind::newCmd, "kind=%s [%s]*",
		"Creates a new object of given kind");
  m->newCommand("which", TclObjectKind::whichCmd, "[kind=%s]",
		"Return a list of item kind, or a list of objects of a given kind name");
  m->newCommand("xgrab", TclModule::xgrabCmd, "%w file=%s",
		"Grab a TK window and stores it in a file");

  TclConstraint::createCommands(m);
  TclProbabilityDistribution::createCommands(m);
  TclGaussianDistribution::createCommands(m);
  TclGaussianMixtureDistribution::createCommands(m);
  TclHistogramDistribution::createCommands(m);
  TclGenericHistogram::createCommands(m);

  
  Togl_Init(m->getInterp());

  kind = new TclObjectKind(m, typeid(TclMatrix4D), "mat4D", 
			   "4D Matrice", &TclMatrix4D::create,
			   &TclMatrix4D::list, "-diag %f %f %f %f ] [-val %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f] [-mat %s] [-name %s]");
  
  m->newCommand("mat4D%u", kind->getOptions(),
		"Matrix management command");
  
  kind->newOption("-val", (TclObject::OPTION) &TclMatrix4D::valCmd, "%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f ", "Get/Set resample matrix values" );
  kind->newOption("-name", (TclObject::OPTION) &TclMatrix4D::nameCmd, "[%s]",
		  "get/set name of the matrix");
  kind->newOption("-id", (TclObject::OPTION) &TclMatrix4D::idCmd, "", "Set resample matrix to identity" );
  kind->newOption("-raz", (TclObject::OPTION) &TclMatrix4D::razCmd, "", "Set resample matrix to zero" );
  kind->newOption("-inv", (TclObject::OPTION) &TclMatrix4D::invCmd, "", "Inverse matrix" );
  kind->newOption("-set", (TclObject::OPTION) &TclMatrix4D::setCmd, "mat4D%u", "set the current matrix from another matrix");
  kind->newOption("-mul", (TclObject::OPTION) &TclMatrix4D::mulCmd, "mat4D%u+", "Multiply (on the right) the current matrix with a set of matrices");
  kind->newOption("-leftMul", (TclObject::OPTION) &TclMatrix4D::leftMulCmd, "mat4D%u+", "Multiply (on the left) the current matrix with a set of matrices");
  kind->newOption("-destroy", (TclObject::OPTION) &TclMatrix4D::destroyCmd, "", "destroy the object");
  kind->newOption("-saveAstransfo", (TclObject::OPTION) &TclMatrix4D::saveAstransfoCmd, "%s", "save the matrix in %s file");
  kind->newOption("-saveRowMajor", (TclObject::OPTION) &TclMatrix4D::saveRowMajorCmd, "%s", "save the matrix in %s file each row after another");
  kind->newOption("-saveColumnMajor", (TclObject::OPTION) &TclMatrix4D::saveColumnMajorCmd, "%s", "save the matrix in %s file each column after another");
  kind->newOption("-loadAstransfo", (TclObject::OPTION) &TclMatrix4D::loadAstransfoCmd, "%s", "load the matrix from file");
  kind->newOption("-loadFromMedINRIA", (TclObject::OPTION) &TclMatrix4D::loadFromMedINRIACmd, "%s", "load the matrix from a MedINRia matrix file");

  kind = new TclObjectKind(m, typeid(TclFuzzyCMeans), "fcm","Fuzzy C Means Classification", &TclFuzzyCMeans::create, &TclFuzzyCMeans::list,0);
  m->newCommand("fcm%u", kind->getOptions(),
		"Fuzzy C Means management command");
  
  kind->newOption("-destroy", (TclObject::OPTION) &TclFuzzyCMeans::destroyCmd, 0, "Destroy range information");
  kind->newOption("-clusterCenter", (TclObject::OPTION) &TclFuzzyCMeans::clusterCenterCmd, "clusterIndex=%u", "Returns the coordinates of the cluster center given the cluster index");
  kind->newOption("-onlyIntensity", (TclObject::OPTION) &TclFuzzyCMeans::onlyIntensityCmd, 0, "Returns a boolean whether the FCM includes only intensity or position+orientation+intensity");
  kind->newOption("-nbrClusters", (TclObject::OPTION) &TclFuzzyCMeans::nbrClustersCmd, 0, "Returns the number of clusters");
  kind->newOption("-nbrData", (TclObject::OPTION) &TclFuzzyCMeans::nbrDataCmd, 0, "Returns the number of data points");
  kind->newOption("-nbrCoord", (TclObject::OPTION) &TclFuzzyCMeans::nbrCoordCmd, 0, "Returns the number of coordinates for each data point");
  kind->newOption("-expo", (TclObject::OPTION) &TclFuzzyCMeans::expoCmd, 0, "Returns the exponent used for the partition matrix");
  kind->newOption("-maxIter", (TclObject::OPTION) &TclFuzzyCMeans::maxIterCmd, 0, "Returns the maximum number of iterations");
  kind->newOption("-nbrIter", (TclObject::OPTION) &TclFuzzyCMeans::nbrIterCmd, 0, "Returns the number of iterations used until stability");
  kind->newOption("-minImpro", (TclObject::OPTION) &TclFuzzyCMeans::minImproCmd, 0, "Returns the minimum amount of improvement");
  kind->newOption("-vpc", (TclObject::OPTION) &TclFuzzyCMeans::vpcCmd, 0, "Returns partition coefficient (Vpc)");
  kind->newOption("-vpe", (TclObject::OPTION) &TclFuzzyCMeans::vpeCmd, 0, "Returns partition entropy (Vpe)");
  kind->newOption("-overlap", (TclObject::OPTION) &TclFuzzyCMeans::overlapCmd, 0, "Returns the overlap");
  kind->newOption("-separation", (TclObject::OPTION) &TclFuzzyCMeans::separationCmd, 0, "Returns the separation");
  kind->newOption("-data", (TclObject::OPTION) &TclFuzzyCMeans::dataCmd, "v%u", "returns the FCM data associated to vertex n");
  kind->newOption("-membership", (TclObject::OPTION) &TclFuzzyCMeans::membershipCmd, "v%u",
		  "returns the FCM membership associated to vertex n. For display purposes, the order of the membership is from the last cluster (cluster n-1) to the first one (cluster 0).");
  kind->newOption("-metric", (TclObject::OPTION) &TclFuzzyCMeans::metricCmd, "v%u", "returns the metric used by FCM");

  kind = new TclObjectKind(m, typeid(TclExpectationMaximization), "em","Expectation-Maximization Classification", &TclExpectationMaximization::create, &TclExpectationMaximization::list,0);
  m->newCommand("em%u", kind->getOptions(),
		"Expectation-Maximization management command");
  
  kind->newOption("-destroy", (TclObject::OPTION) &TclExpectationMaximization::destroyCmd, 0, "Destroy range information");
  kind->newOption("-modeCenter", (TclObject::OPTION) &TclExpectationMaximization::modeCenterCmd, "clusterIndex=%u", "Returns the coordinates of the mode center given the mode index");
 kind->newOption("-modeStdDeviation", (TclObject::OPTION) &TclExpectationMaximization::modeStdDeviationCmd, "clusterIndex=%u", "Returns the std deviation of a given mode from the estimated covariance matrix");

  kind->newOption("-nbrModes", (TclObject::OPTION) &TclExpectationMaximization::nbrModesCmd, 0, "Returns the number of modes");
  kind->newOption("-nbrData", (TclObject::OPTION) &TclExpectationMaximization::nbrDataCmd, 0, "Returns the number of data");
  kind->newOption("-nbrCoord", (TclObject::OPTION) &TclExpectationMaximization::nbrCoordCmd, 0, "Returns the number of coordinates");
  kind->newOption("-maxIter", (TclObject::OPTION) &TclExpectationMaximization::maxIterCmd, 0, "Returns the maximum number of iterations");
  kind->newOption("-nbrIter", (TclObject::OPTION) &TclExpectationMaximization::nbrIterCmd, 0, "Returns the number of iterations used until stability");
  kind->newOption("-minImpro", (TclObject::OPTION) &TclExpectationMaximization::minImproCmd, 0, "Returns the minimum amount of improvement");
  kind->newOption("-vpc", (TclObject::OPTION) &TclExpectationMaximization::vpcCmd, 0, "Returns partition coefficient (Vpc)");
  kind->newOption("-vpe", (TclObject::OPTION) &TclExpectationMaximization::vpeCmd, 0, "Returns partition entropy (Vpe)");
  kind->newOption("-overlap", (TclObject::OPTION) &TclExpectationMaximization::overlapCmd, 0, "Returns the overlap");
  kind->newOption("-separation", (TclObject::OPTION) &TclExpectationMaximization::separationCmd, 0, "Returns the separation");
  kind->newOption("-data", (TclObject::OPTION) &TclExpectationMaximization::dataCmd, "v%u", "returns the EM data associated to vertex n");
  kind->newOption("-proba", (TclObject::OPTION) &TclExpectationMaximization::probaCmd, "v%u",
		  "returns the EM probability associated to vertex n. For display purposes, the order of the probability is from the last cluster (cluster n-1) to the first one (cluster 0).");


  kind = new TclObjectKind(m, typeid(TclRangeInformation), "rinfo","Range Information", &TclRangeInformation::create, &TclRangeInformation::list,0);
  m->newCommand("rinfo%u", kind->getOptions(),
		"Range Information management command");
  
  kind->newOption("-destroy", (TclObject::OPTION) &TclRangeInformation::destroyCmd, 0, "Destroy range information");
  kind->newOption("-name", (TclObject::OPTION) &TclRangeInformation::nameCmd, 0, "Get the name of the range information");
  kind->newOption("-cast", (TclObject::OPTION) &TclRangeInformation::castCmd, 0, "Cast the current object into  a new TclRangeInformation object associated with the same range information. Delete the old object");
  
  kind = new TclObjectKind(m, typeid(TclRangeInformationSet), "rinfos","Range Information Set", &TclRangeInformationSet::create, &TclRangeInformationSet::list,0);
  m->newCommand("rinfos%u", kind->getOptions(),
		"Range Information Set management command");
  
  kind->newOption("-destroy", (TclObject::OPTION) &TclRangeInformationSet::destroyCmd, 0, "Destroy range information");
  kind->newOption("-name", (TclObject::OPTION) &TclRangeInformationSet::nameCmd, 0, "Get the name of the range information set");

#ifdef EMBEDED_TCL_SCRIPTS
 /* Statically loads the tcl script from tclModules.tcl.h (generated by tcl2h) */
 Tcl_Eval(m->getInterp(), tclModules_tcl);
 /* Tell the TclModule constructor not to load the script again */
 m->tclScriptLoaded();
#endif
 /* Otherwise, the tcl script will be loaded dynamically from tclModules.tcl by the TclModule constructor */
}

}
