
#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#pragma implementation "modules/TclConstraint.h"

#include <modules/TclConstraint.h>
#include <framework/Constraint.h>

namespace yav {

TclConstraint::TclConstraint(TclModule *m,
			     Constraint *c) : TclObject(typeid(TclConstraint),m),constraint(c) {
  done(typeid(TclConstraint));
}

TclConstraint::TclConstraint(const std::type_info& type, TclModule *m,
			     Constraint *c) : TclObject(type,m),constraint(c) {
}
TclConstraint::~TclConstraint() {}


int TclConstraint::create(TclModule *m, int argc, char **argv) {
  if (argc!=1)
    return COMMAND_SYNOPSIS; 
  TclConstraint *c=new TclConstraint (m,0);
  m->setResult("%s%u", c->getObjectName(), c->TclObject::getRef());
  return TCL_OK;
}
int TclConstraint::destroyCmd(int argc, char **argv) {
  TclModule *m;
  
  if(argc != 0) return COMMAND_SYNOPSIS;
  m = getModule();
  TclString res=name();
  delete this;
  m->setResult(res);
  return TCL_OK;
}

  /// -deleteConstraint option
int TclConstraint::sameConstraintCmd(int argc, char **argv) {
if (argc!=1) 
  return COMMAND_SYNOPSIS;
 else {
   TclConstraint *tc=(TclConstraint *)  TclObjectKind::getObject(typeid(TclConstraint), argc, argv, 0);
   if (tc==0) {
     setResult("Illegal constraint name : %s",argv[1]);
     return TCL_ERROR;
   }
   if (tc->getConstraint()==getConstraint()){
     setResult("true");
   } else {
     setResult("false");
   }
    return TCL_OK; 
 }
}
  /// -deleteConstraint option
int TclConstraint::deleteConstraintCmd(int argc, char **argv) {
  TclModule *m;
  unsigned int ref;
  
  if(argc != 0) return COMMAND_SYNOPSIS;
  m = getModule();
  ref = TclObject::getRef();
  TclString res=name();
  delete constraint;
  delete this;
  m->setResult(res);
  return TCL_OK;
}
int TclConstraint::frequencyCmd(int argc, char **argv) {
  if (argc>1) 
    return COMMAND_SYNOPSIS;
  else if(argc == 1) {
    unsigned int freq= (unsigned int) atoi(argv[0]);
    if (getConstraint())
      getConstraint()->setFrequency(freq);
  }
  if (getConstraint())
    setResult("%u",getConstraint()->getFrequency());
  return TCL_OK;
}
int TclConstraint::constraintNameCmd(int argc, char **argv) {
  TclModule *m;
  
  if(argc != 0) return COMMAND_SYNOPSIS;
  m = getModule();
  m->setResult("%s", constraint->getName());
  return TCL_OK;
  
}

/// -cast option
int TclConstraint::castCmd(int argc, char **argv) {
  if(argc != 0) return COMMAND_SYNOPSIS;
  TclModule *m;
  m = getModule();
  TclConstraint *c=new TclConstraint(m,constraint);
  
  delete this;
  m->setResult(c->name());
  return TCL_OK;
}
Constraint *TclConstraint::getConstraint() const {
  return constraint; }
void TclConstraint::createCommands(TclModule *m) {
  TclObjectKind *kind = new TclObjectKind(m, typeid(TclConstraint), "constr","Constraint on Deformable Models", &TclConstraint::create, &TclConstraint::list,0);
  m->newCommand("constr%u", kind->getOptions(),
		"Constraint management command");
  
  kind->newOption("-destroy", (TclObject::OPTION) &TclConstraint::destroyCmd, 0, "destroy constraint interface but without deleting the constraint"); 
  kind->newOption("-deleteConstraint", (TclObject::OPTION) &TclConstraint::deleteConstraintCmd, 0, "delete constraint and its Tcl interface"); 
  kind->newOption("-sameConstraint", (TclObject::OPTION) &TclConstraint::sameConstraintCmd, "constr%u", "returns true or false depending if both constraints points towards the same constraint"); 
  kind->newOption("-frequency", (TclObject::OPTION) &TclConstraint::frequencyCmd, "[freq=%u]", "get or set the frequency at which the constraint is applied"); 
  kind->newOption("-cast", (TclObject::OPTION) &TclConstraint::castCmd, 0, "cast the current object into  a new TclConstraint object associated with the same constraint. Delete the old object");
  kind->newOption("-constraintName", (TclObject::OPTION) &TclConstraint::constraintNameCmd, 0, "returns the  expanded name of the constraint pointed by this Tcl interface");
}
void TclConstraint::registerConstraintEditProcedure(TclModule *m,
						    const char *constraintName,
						    const char *procedureName,
						    const char *tclname) {
  if (procedureName)
    m->tickle("lappend availableConstraintEditProcedure { {%s} %s %s}",
	      constraintName,procedureName,tclname);
}

}
