#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#include <modules/TclFuzzyCMeans.h>
#include <modules/TclGenericHistogram.h>

namespace yav
{
/// create Tcl command to create an object and to handle its option commands
void TclFuzzyCMeans::createCommands(TclModule *m)
{
	TclObjectKind *kind = new TclObjectKind(m, typeid(TclFuzzyCMeans), "proba","Probability Density Function", &TclFuzzyCMeans::create, &TclFuzzyCMeans::list,0);
	m->newCommand("proba%u", kind->getOptions(), "Probability management command");
	kind->newOption("-destroy", (TclObject::OPTION) &TclFuzzyCMeans::destroyCmd, 0, "destroy probability interface but without deleting the probabity density function pointing to it"); 
	kind->newOption("-deleteDistribution", (TclObject::OPTION) &TclFuzzyCMeans::deleteDistributionCmd, 0, "delete distribution and its Tcl interface"); 
	kind->newOption("-sameDistribution", (TclObject::OPTION) &TclFuzzyCMeans::sameDistributionCmd, "proba%u", "returns true or false depending if both constraints points towards the same constraint"); 
	kind->newOption("-cast", (TclObject::OPTION) &TclFuzzyCMeans::castCmd, 0, "cast the current object into  a new TclFuzzyCMeans object associated with the same probability function. Delete the old object");
	kind->newOption("-range", (TclObject::OPTION) &TclFuzzyCMeans::rangeCmd, 0, "returns the minimum value the maximum value and the number of significant samples");
	kind->newOption("-statistics", (TclObject::OPTION) &TclFuzzyCMeans::statisticsCmd, 0, "returns the mean and standard deviation of the distribution");
	kind->newOption("-probabilityValue", (TclObject::OPTION) &TclFuzzyCMeans::probabilityValueCmd, "%f", "returns the probability density function associated with that value ");
	kind->newOption("-histogramValue", (TclObject::OPTION) &TclFuzzyCMeans::histogramValueCmd, "%f", "returns the histogram value associated with that value ");
	kind->newOption("-histogram", (TclObject::OPTION) &TclFuzzyCMeans::histogramCmd, "%f", "returns a Tcl Histogram");
}

int TclFuzzyCMeans::create(TclModule *m, int argc, char **argv)
{
	if (argc!=1)
		return COMMAND_SYNOPSIS; 

	TclFuzzyCMeans *ri=new TclFuzzyCMeans (m,0);
	m->setResult("%s%u", (char *)ri->getObjectName(), ri->TclObject::getRef());

	return TCL_OK;
}

 /// -destroy option
int TclFuzzyCMeans::destroyCmd(int argc, char **argv)
{
	TclModule *m;

	if(argc != 0) return COMMAND_SYNOPSIS;
		m = getModule();

	TclString n=TclObject::name();
	delete this;
	m->setResult(n);

	return TCL_OK;
}
}
