#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation  "modules/TclGenericHistogram.h"
#pragma implementation  "modules/TclConstraintType.h"

#include <modules/TclGenericHistogram.h>

namespace yav {


TclGenericHistogram::TclGenericHistogram(TclModule *m, GenericHistogram *h) :
  TclObject(typeid(TclGenericHistogram),m),Owner<GenericHistogram>(h){
  done(typeid(TclGenericHistogram));
}
int TclGenericHistogram::create(TclModule *m, int argc, char **argv) {
  TclGenericHistogram * h;    
  if(argc != 2) return COMMAND_SYNOPSIS;
  else h = new TclGenericHistogram(m,new GenericHistogram(argv[1]));
  m->setResult("%s%u", h->getObjectName(), h->getRef());
  return TCL_OK;
}
int TclGenericHistogram::destroyCmd(int argc, char **argv) {
  if(argc != 0) return COMMAND_SYNOPSIS;
  TclModule *m = getModule();
  TclString n=name();
  delete this;
  m->setResult("%s does not exist any longer",
	       (char *) n);
  return TCL_OK;
}
int TclGenericHistogram::statCmd(int argc, char **argv) {
  double min=0.0, max=0.0;
  if ((argc==0) || (argc>6)) return COMMAND_SYNOPSIS;
  else if (((argc==1)||(argc==5)) && (!strcmp(argv[0],"mean"))) {
      if (argc==5) {
	  if (!strcmp(argv[1],"min"))
      if(sscanf(argv[2], "%lf", &min) != 1) return COMMAND_SYNOPSIS;
    if (!strcmp(argv[3],"max"))
      if(sscanf(argv[4], "%lf", &max) != 1) return COMMAND_SYNOPSIS;
    getClient()->computeStatistics(min, max);
	  }
    setResult("%f",getClient()->getMeanValue());
    return TCL_OK;
  }
  else if (((argc==1)||(argc==5)) && (!strcmp(argv[0],"median"))) {
	  if (argc==5) {
	  if (!strcmp(argv[1],"min"))
      if(sscanf(argv[2], "%lf", &min) != 1) return COMMAND_SYNOPSIS;
    if (!strcmp(argv[3],"max"))
      if(sscanf(argv[4], "%lf", &max) != 1) return COMMAND_SYNOPSIS;
    getClient()->computeStatistics(min, max);
	  }
    setResult("%f",getClient()->getMedianValue());
    return TCL_OK;
  }
  else if (((argc==1)||(argc==5)) && (!strcmp(argv[0],"stdDeviation"))) {
	  if (argc==5) {
	  if (!strcmp(argv[1],"min"))
      if(sscanf(argv[2], "%lf", &min) != 1) return COMMAND_SYNOPSIS;
    if (!strcmp(argv[3],"max"))
      if(sscanf(argv[4], "%lf", &max) != 1) return COMMAND_SYNOPSIS;
    getClient()->computeStatistics(min, max);
	  }
    setResult("%f",getClient()->getStandardDeviationValue());
    return TCL_OK;
  }
  else if (((argc==2)||(argc==6)) && (!strcmp(argv[0],"quantile"))) {
	  if (argc==6) {
	  if (!strcmp(argv[2],"min"))
      if(sscanf(argv[3], "%lf", &min) != 1) return COMMAND_SYNOPSIS;
    if (!strcmp(argv[4],"max"))
      if(sscanf(argv[5], "%lf", &max) != 1) return COMMAND_SYNOPSIS;
    getClient()->computeStatistics(min, max);
	  }
    setResult("%f",getClient()->getHistogramQuantile((double) atof(argv[1])));
    return TCL_OK;
  }
  else if (((argc==1)||(argc==5)) && (!strcmp(argv[0],"maxNbSamples"))) {
	if (argc==5) {
    if (!strcmp(argv[1],"min"))
      if(sscanf(argv[2], "%lf", &min) != 1) return COMMAND_SYNOPSIS;
    if (!strcmp(argv[3],"max"))
      if(sscanf(argv[4], "%lf", &max) != 1) return COMMAND_SYNOPSIS;
    getClient()->computeStatistics(min, max);
	}
    setResult("%lu",getClient()->getMaxNbSamples());
    return TCL_OK;
  }
  else
    return COMMAND_SYNOPSIS;
    
}
int TclGenericHistogram::valueCmd(int argc, char **argv) {
  if(argc!=1) return COMMAND_SYNOPSIS;
  setResult("%lu",getClient()->getNbSamples((double) atof(argv[0])));
  return TCL_OK;
}
int TclGenericHistogram::minValueCmd(int argc, char **argv) {
  if (argc!=0)
    return COMMAND_SYNOPSIS;
  else
    {
      setResult("%f",getClient()->getMinimumValue());
      return TCL_OK;
    }
}
int TclGenericHistogram::maxValueCmd(int argc, char **argv) {
  if (argc!=0)
    return COMMAND_SYNOPSIS;
  else
    {
      setResult("%f",getClient()->getMaximumValue());
      return TCL_OK;
    }
}

int TclGenericHistogram::nbCmd(int argc, char **argv) {
  double abs=0.0;
  if (argc!=1)
    return COMMAND_SYNOPSIS;
  else {
    if(sscanf(argv[0], "%lf", &abs) != 1) return COMMAND_SYNOPSIS;
    abs = ((double) getClient()->nbBuckets)*(abs-getClient()->minValue)/(getClient()->maxValue-getClient()->minValue);
    setResult("%u", getClient()->histogram[(unsigned int)abs]);
    return TCL_OK;
  }
}
int TclGenericHistogram::plotCmd(int argc, char **argv) {
  if (argc!=0)
    return COMMAND_SYNOPSIS;
  else {
    getModule()->tickle("modules::plotHistogram %s",(char *)name());
    return TCL_OK;
  }
}
int TclGenericHistogram::smoothCmd(int argc, char **argv) {
  if (argc!=1)
    return COMMAND_SYNOPSIS;
  else {
    double filter= (double) atof(argv[0]);
    getClient()->smooth(filter);
    setResult("%s smoothed",(char *) name());
    return TCL_OK;
  }
}

int TclGenericHistogram::addHistogramCmd(int argc, char **argv) {
  if (argc!=1)
    return COMMAND_SYNOPSIS;
  else {
    TclGenericHistogram *h=(TclGenericHistogram *)	TclObjectKind::getObject(typeid(TclGenericHistogram),
		argc, argv, 0);
	if (h==0) {
		setResult("%s is not an histogram",argv[0]);
		return TCL_ERROR;
	}
	(*(getClient()))+= *(h->getClient());
    setResult("added %s",argv[0]);
    return TCL_OK;
  }
}
int TclGenericHistogram::bucketCmd(int argc, char **argv) {
  if ((argc==0) || (argc>2))
    return COMMAND_SYNOPSIS;
  if (!strcmp(argv[0],"size"))
    {
      if (argc==1)
	{
	  setResult("%f",getClient()->getBucketSize());
	  return TCL_OK;
	}
      else
	{
	  double s=(double) atof(argv[1]);
	  getClient()->buildHistogram(s);
	  setResult("%s changed bucket size to %f",(char *) name(),s);
	  return TCL_OK;
	}
    }
  else   if (!strcmp(argv[0],"number"))
    {
      if (argc==1)
	{
	  setResult("%u",getClient()->getNumberBuckets());
	  return TCL_OK;
	}
      else
	{
	  unsigned int n=(unsigned int) atoi(argv[1]);
	  getClient()->buildHistogram(n);
	  setResult("%s changed number of buckets to %u",(char *) name(),n);
	  return TCL_OK;
	}
    }
  else {
    setResult("%s must be either size or number",argv[0]);
    return TCL_ERROR;
  }
}
int TclGenericHistogram::saveCmd(int argc, char **argv) {
  if(argc!=1) return COMMAND_SYNOPSIS;
  getClient()->write(argv[0]);
  setResult("Histogram saved as %s",argv[0]);
  return TCL_OK;
}
TclGenericHistogram::~TclGenericHistogram() {
  if (getClient()) {
    GenericHistogram *h=getClient();
    pointer=0;
    delete h;
  }
}	

void TclGenericHistogram::createCommands(TclModule *m) {
 TclObjectKind *kind = new TclObjectKind(m, typeid(TclGenericHistogram), "histo","Histogram","%s",
			   &TclGenericHistogram::create, 
			   &TclGenericHistogram::list);
  kind->newOption("-nb", (TclObject::OPTION) &TclGenericHistogram::nbCmd,"%f" ,
		  "Get the value of the histogram for a given value");
  kind->newOption("-value", (TclObject::OPTION) &TclGenericHistogram::valueCmd,"%f" ,
		  "Get number of sample having the value (depending of the bucket size) than a given value");
  kind->newOption("-save", (TclObject::OPTION) &TclGenericHistogram::saveCmd,"%s" ,
		  "Saves an histogram in a file");
  kind->newOption("-bucket", (TclObject::OPTION) &TclGenericHistogram::bucketCmd," size [%f] | number [%u]" ,
		  "Get or set the bucket size or get/set the number of buckets");
  kind->newOption("-plot", (TclObject::OPTION) &TclGenericHistogram::plotCmd,0,
		  "Plot the histogram");
  kind->newOption("-smooth", (TclObject::OPTION) &TclGenericHistogram::smoothCmd,"%f",
		  "Smooth the histogram with a Gaussian filter whose size is given");
kind->newOption("-addHistogram", (TclObject::OPTION) &TclGenericHistogram::addHistogramCmd,"histo%u",
		  "Add a given histogram to the current histogram and updates the statistics");
kind->newOption("-minValue", (TclObject::OPTION) &TclGenericHistogram::minValueCmd,0,
		  "Get the minimum value of the histogram");
  kind->newOption("-maxValue", (TclObject::OPTION) &TclGenericHistogram::maxValueCmd,0,
		  "Get the maximum value of the histogram");
  kind->newOption("-stat", (TclObject::OPTION) &TclGenericHistogram::statCmd,"mean | median | stdDeviation | quantile %f",
		  "Get some basic statistics on the histogram. It includes the computation of the mean or median value, the standard deviation and the quantile for any value between 0.0 and 1.0");
  kind->newOption("-destroy", (TclObject::OPTION) &TclGenericHistogram::destroyCmd,0,
		  "Delete an histogram");
} 

}
