#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "modules/TclMatrix4D.h"

#include <modules/TclMatrix4D.h>

namespace yav {

TclMatrix4D::TclMatrix4D(const std::type_info& info, TclModule *m)
  : TclObject(info, m), Matrix4x4<double>(1,0,0,0,0,1,0,0,0,0,1,0,0,0,0,1), name("noname")
{
  done(typeid(TclMatrix4D));
}

TclMatrix4D::TclMatrix4D(TclModule *m, double t[4][4])
  : TclObject(typeid(TclMatrix4D), m), Matrix4x4<double>(t[0][0],t[0][1],t[0][2],t[0][3],
							 t[1][0],t[1][1],t[1][2],t[1][3],
							 t[2][0],t[2][1],t[2][2],t[2][3],
							 t[3][0],t[3][1],t[3][2],t[3][3]), name("noname")
{
  done(typeid(TclMatrix4D));
}

TclMatrix4D::TclMatrix4D(TclModule *m, double t[16])
  : TclObject(typeid(TclMatrix4D), m), Matrix4x4<double>(t[0],t[1],t[2],t[3],
							 t[4],t[5],t[6],t[7],
							 t[8],t[9],t[10],t[11],
							 t[12],t[13],t[14],t[15]), name("noname")
{
  done(typeid(TclMatrix4D));
}

TclMatrix4D::TclMatrix4D(TclModule *m, Matrix4x4<double> &mat)
  : TclObject(typeid(TclMatrix4D), m), Matrix4x4<double>(mat), name("noname")
{
  done(typeid(TclMatrix4D));
}




int TclMatrix4D::create(TclModule *m, int argc, char **argv) {

  double t[16];
  TclMatrix4D *tmat;
  int i,j;
  TclString nom="noname";

  t[0]=1.0; t[1]=0.0; t[2]=0.0; t[3]=0.0;
  t[4]=0.0; t[5]=1.0; t[6]=0.0; t[7]=0.0;
  t[8]=0.0; t[9]=0.0; t[10]=1.0; t[11]=0.0;
  t[12]=0.0; t[13]=0.0; t[14]=0.0; t[15]=1.0;
  
  if (argc>=2)
    for (i=1; i<argc; i++) {
      if (!strcmp(argv[i], "-name"))
	if(++i >= argc) return COMMAND_SYNOPSIS;
	else nom = TclString(argv[i]);
      else if (!strcmp(argv[i], "-fileRow")) {
	if(++i >= argc) return COMMAND_SYNOPSIS;
	else {
	  try {
	    IFSTREAM f(argv[i]);
	    
	    f>>t[0] >> t[1] >> t[2] >> t[3];
	    f>>t[4] >> t[5] >> t[6] >> t[7];
	    f>>t[8] >> t[9] >> t[10] >> t[11];
	    f>>t[12] >> t[13] >> t[14] >> t[15];
	  }
	  catch(...) {
	    CERR << " Unable to open  " << argv[i] << " as input" << ENDL;
	    return TCL_ERROR;
	  }
	}
      }
      else if (!strcmp(argv[i], "-fileCol")) {
	if(++i >= argc) return COMMAND_SYNOPSIS;
	else {

	  try {
	    IFSTREAM f(argv[i]);
	    f>>t[0] >> t[4] >> t[8] >> t[12];
	    f>>t[1] >> t[5] >> t[9] >> t[13];
	    f>>t[2] >> t[6] >> t[10] >> t[14];
	    f>>t[3] >> t[7] >> t[11] >> t[15];
	  }
	  catch(...) {
	    CERR << " Unable to open  " << argv[i] << " as input" << ENDL;
	    return TCL_ERROR;
	  }
	}
      }
      else if (!strcmp(argv[i], "-mat"))
	if(++i >= argc) return COMMAND_SYNOPSIS;
	else {
	  TclMatrix4D *mget = TclMatrix4D::getObject(argc, argv, i);
	  if (!mget) return COMMAND_SYNOPSIS;
	  t[0]=(*mget)[0][0];t[1]=(*mget)[0][1]; t[2]=(*mget)[0][2]; t[3]=(*mget)[0][3];
	  t[4]=(*mget)[1][0]; t[5]=(*mget)[1][1]; t[6]=(*mget)[1][2]; t[7]=(*mget)[1][3];
	  t[8]=(*mget)[2][0]; t[9]=(*mget)[2][1]; t[10]=(*mget)[2][2]; t[11]=(*mget)[2][3];
	  t[12]=(*mget)[3][0]; t[13]=(*mget)[3][1]; t[14]=(*mget)[3][2]; t[15]=(*mget)[3][3];
	}
      else if (!strcmp(argv[i], "-val"))
	for (j=0; j<16; j++) {
	  if(++i >= argc) return COMMAND_SYNOPSIS;
	  else if ( !(sscanf(argv[i], "%lf", &(t[j]))) ) return COMMAND_SYNOPSIS;
	}
      else if (!strcmp(argv[i], "-diag"))
	for (j=0; j<4; j++) {
	  if(++i >= argc) return COMMAND_SYNOPSIS;
	  else if ( !(sscanf(argv[i], "%lf", &(t[5*j]))) ) return COMMAND_SYNOPSIS;
	}
      else return COMMAND_SYNOPSIS;
    }
  
  tmat = new TclMatrix4D(m, t);
  tmat->name=nom;

  m->setResult("%s%u", tmat->getObjectName(), tmat->getRef());
  
  return TCL_OK;
  
}




TclMatrix4D::~TclMatrix4D() {}

int TclMatrix4D::idCmd(int argc, char **argv) {

  if (argc >0 ) return COMMAND_SYNOPSIS;

  (*this).Id();

  setResult("%s%u is identity", getObjectName(), getRef());
  return TCL_OK;
  
}

int TclMatrix4D::razCmd(int argc, char **argv) {

  if (argc >0 ) return COMMAND_SYNOPSIS;

  for (int i=0; i<4; ++i)
    for (int j=0; j<4; ++j)
      (*this)(i,j)=0.0;

  setResult("%s%u is a zero matrix", getObjectName(), getRef());
  return TCL_OK;
   
}

int TclMatrix4D::invCmd(int argc, char **argv) {

  if (argc >0 ) return COMMAND_SYNOPSIS;

  inv(*this);

  setResult("%s%u inversed", getObjectName(), getRef());
  return TCL_OK;
   
}

int TclMatrix4D::nameCmd(int argc, char **argv) {
  if(argc > 1) return COMMAND_SYNOPSIS;
  if(argc == 1) name=argv[0];
  setResult("%s", (char *) name);
  return TCL_OK;
}


int TclMatrix4D::mulCmd(int argc, char **argv) {
  if (argc<1) return COMMAND_SYNOPSIS;
  int i=0;
  TclString res;
  res.sprintf("%s%u ", getTCLObjectName(), getRef());
  res += "multiplied by";
  
    while(i < argc) {
    TclMatrix4D *mat;
    mat = TclMatrix4D::getObject(argc, argv, i);
    if(!mat) return TCL_ERROR;
    (Matrix4x4<double>)*this = Mul(*this, (*mat));
    res += " " + TclString(argv[i]);
    i++;
  }

  setResult(res);
  return TCL_OK;


}

int TclMatrix4D::setCmd(int argc, char **argv) {
  if (argc!=1) return COMMAND_SYNOPSIS;
  //int i=0;
  TclString res;
  res.sprintf("%s%u ", getTCLObjectName(), getRef());
  res += "set as ";
  
  TclMatrix4D *mat;
  mat = TclMatrix4D::getObject(argc, argv, 0);
  if(!mat) return TCL_ERROR;
  (Matrix4x4<double>)*this = (*mat);
  res += " " + TclString(argv[0]);

  setResult(res);
  return TCL_OK;

}

int TclMatrix4D::leftMulCmd(int argc, char **argv) {
  if (argc<1) return COMMAND_SYNOPSIS;
  int i=0;
  TclString res;
  res.sprintf("%s%u ", getTCLObjectName(), getRef());
  res += "left multiplied by";
  
    while(i < argc) {
    TclMatrix4D *mat;
    mat = TclMatrix4D::getObject(argc, argv, i);
    if(!mat) return TCL_ERROR;
    (Matrix4x4<double>)*this = Mul((*mat),*this);
    res += " " + TclString(argv[i]);
    i++;
  }

  setResult(res);
  return TCL_OK;


}

int TclMatrix4D::valCmd(int argc, char **argv) {


  if ( (argc != 16 ) && (argc!= 0) )
    return COMMAND_SYNOPSIS;
  
  if ( argc == 16 )
  for (int i=0; i<4; ++i)
    for (int j=0; j<4; ++j)
      if(sscanf(argv[4*i+j], "%lf", &((*this)(i,j))) != 1) return COMMAND_SYNOPSIS;    

  setResult("%f %f %f %f \n %f %f %f %f \n %f %f %f %f \n %f %f %f %f \n ", 
	    (*this)(0,0), (*this)(0,1), (*this)(0,2),(*this)(0,3),
	    (*this)(1,0), (*this)(1,1), (*this)(1,2),(*this)(1,3),
	    (*this)(2,0), (*this)(2,1), (*this)(2,2),(*this)(2,3),
	    (*this)(3,0), (*this)(3,1), (*this)(3,2),(*this)(3,3));  
  
  return TCL_OK;

}

int TclMatrix4D::destroyCmd(int argc, char **argv){
  delete this;
  return TCL_OK;
}

TclMatrix4D& TclMatrix4D::operator= (const Matrix4x4<double> & m){
  setValues(m.getValue(0,0), m.getValue(0,1), m.getValue(0,2), m.getValue(0,3),
	    m.getValue(1,0), m.getValue(1,1), m.getValue(1,2), m.getValue(1,3),
	    m.getValue(2,0), m.getValue(2,1), m.getValue(2,2), m.getValue(2,3),
	    m.getValue(3,0), m.getValue(3,1), m.getValue(3,2), m.getValue(3,3));
  return *this;
}

int TclMatrix4D::saveRowMajorCmd(int argc, char **argv){
  if (argc!=1) return COMMAND_SYNOPSIS;
  OFSTREAM f(argv[0]);
  if (f) {
    for (int i=0; i<4; ++i)
      f << (*this)(i,0) << " " << (*this)(i,1) << " " << (*this)(i,2) << " " << (*this)(i,3) << ENDL;
    f << ENDL;
  } else {
    CERR << " Unable to write " << getObjectName() << getRef() << " in " << argv[0] << ENDL;
    return TCL_ERROR;
  }
  setResult("%s%u saved in %s", getObjectName(), getRef(), argv[0]);
  return TCL_OK;
}
int TclMatrix4D::saveColumnMajorCmd(int argc, char **argv){
  if (argc!=1) return COMMAND_SYNOPSIS;
  OFSTREAM f(argv[0]);
  if (f) {
    for (int i=0; i<4; ++i)
      f << (*this)(0,i) << " " << (*this)(1,i) << " " << (*this)(2,i) << " " << (*this)(3,i) << ENDL;
    f << ENDL;
  } else {
    CERR << " Unable to write " << getObjectName() << getRef() << " in " << argv[0] << ENDL;
    return TCL_ERROR;
  }
  setResult("%s%u saved in %s", getObjectName(), getRef(), argv[0]);
  return TCL_OK;
}

int TclMatrix4D::saveAstransfoCmd(int argc, char **argv){
  if (argc!=1) return COMMAND_SYNOPSIS;
  OFSTREAM f(argv[0]);
  if (f) {
    f << "(\nO8\n";
    for (int i=0; i<4; ++i)
      f << (*this)(i,0) << " " << (*this)(i,1) << " " << (*this)(i,2) << " " << (*this)(i,3) << ENDL;
    f << ")\n";
  } else {
    CERR << " Impossible d'ecrire " << getObjectName() << getRef() << " dans " << argv[0] << ENDL;
    return TCL_ERROR;
  }
  setResult("%s%u saved in %s", getObjectName(), getRef(), argv[0]);
  return TCL_OK;
}

int TclMatrix4D::loadAstransfoCmd(int argc, char **argv){
  if (argc!=1) return COMMAND_SYNOPSIS;
  IFSTREAM f(argv[0]);
  double m00, m01, m02, m03, m10, m11, m12, m13, m20, m21, m22, m23, m30, m31, m32, m33;
  char tmp[3];
  if (f) {
    f >> tmp; 
    f >> tmp; 
    f >> m00 >> m01 >> m02 >> m03;
    f >> m10 >> m11 >> m12 >> m13;
    f >> m20 >> m21 >> m22 >> m23;
    f >> m30 >> m31 >> m32 >> m33;
  } else {
    CERR << " Impossible d'ouvrir " << argv[0] << " en lecture" << ENDL;
    return TCL_ERROR;
  }

  this->setValues(m00, m01, m02, m03, m10, m11, m12, m13, m20, m21, m22, m23, m30, m31, m32, m33);
  this->name = argv[0];

  setResult("%s%u loaded from %s", getObjectName(), getRef(), argv[0]);
  return TCL_OK;
}

int TclMatrix4D::loadFromMedINRIACmd(int argc, char **argv){
  if (argc!=1) return COMMAND_SYNOPSIS;
  IFSTREAM f(argv[0]);
  double m00, m01, m02, m03, m10, m11, m12, m13, m20, m21, m22, m23;
  char tmp[250];
  if (f) {
    f.getline(tmp,250);
    f.getline(tmp,250);
   f.getline(tmp,250);
    f >> tmp >> m00 >> m10 >> m20 >> m01 >> m11 >> m21 >> m02 >> m12 >> m22 >> m03 >> m13 >> m23;
  } else {
    CERR << " Impossible d'ouvrir " << argv[0] << " en lecture" << ENDL;
    return TCL_ERROR;
  }

  this->setValues(m00, m01, m02, m03, m10, m11, m12, m13, m20, m21, m22, m23,0,0,0,1);
  this->name = argv[0];

  setResult("%s%u loaded from %s", getObjectName(), getRef(), argv[0]);
  return TCL_OK;
}


}
