#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 4503 )
#endif

#pragma implementation "modules/TclModule.h"
#include <basics/Constants.h>
#include <framework/Activable.h>
#include <modules/TclModule.h>
#include <modules/TclObjectKind.h>
#include <modules/WindowDump.h>



namespace yav {

// static members initialization
AUTOLINK_STATIC_VARIABLES(TclModule);

const char *TclModule::tclModuleName="tclModules";

std::map<char *,INIT_MODULE_FUNCTION,CStringCompare> &TclModule::getModuleNameMap() {
static std::map<char *,INIT_MODULE_FUNCTION,CStringCompare> moduleNameMap;
return moduleNameMap;
}

/*TclModuleInit tclModuleInit;*/
/** this static object automatically includes Tclmodule_Init in the hash table initMap when 
	it is initialized */
static TclModuleInit tclModuleInit;

LOOP_FUNCTION TclModule::loopFct = TclModule::DefaultMainLoop;
DISPLAY_FUNCTION TclModule::displayFct = TclModule::DefaultDisplay;
EVAL_FUNCTION TclModule::evalFct = TclModule::DefaultEval;
TclString TclModule::TCLevalCommandName;
TclString TclModule::TCLdisplayCommandName;
int TclModule::initargc = 0;
char **TclModule::initargv = 0;
unsigned int TclModule::nmodules = 0;
bool TclModule::loadDynamically=true;
char **TclModule::modulenames = 0;
int (*TclModule::mainProc)(TclModule *, int, char **) = 0;
TclModule *TclModule::coremodule = 0;

#ifndef DEFAULT_MODULES_PATH
#define DEFAULT_MODULES_PATH "DEFAULT_MODULES_PATH_NOT_DEFINE"
#endif
char *TclModule::DefaultModulesPath = DEFAULT_MODULES_PATH;
#ifndef DEFAULT_SHELLS_PATH
#if (!defined( WIN32)) || (defined(__CYGWIN__))
#define DEFAULT_SHELLS_PATH "/proj/epidaure/home/shells/tcl"
#else
#define DEFAULT_SHELLS_PATH "C:\Program Files\yav++\tcl"
#endif
#endif
#ifndef DEFAULT_RESSOURCE_FILE_LOCATION
#define DEFAULT_RESSOURCE_FILE_LOCATION "/proj/epidaure/home/software/yav++/yav++"
#endif
char *TclModule::DefaultShellsPath = DEFAULT_SHELLS_PATH;
char *TclModule::DefaultRessourceFileLocation =
  DEFAULT_RESSOURCE_FILE_LOCATION;
char *TclModule::RessourceModulesPath = 0;
char *TclModule::RessourceShellsPath = 0;
bool TclModule::verbose = false;



TclModule::DlopenFailed::DlopenFailed(const std::string& mname,
				      const std::string& from) :
#if (defined( __GNUC__)) && (!defined(__CYGWIN__))
  Exception(from,
	    "dlopen failed while attempting to load module " + mname +
	    " (" + dlerror() + ")") {
#else
 Exception(from,
	    "dlopen failed while attempting to load module " + mname +
	    " ()") {
#endif
}

TclModule::ModuleNotFound::ModuleNotFound(const std::string& mname, 
					  const std::string& from) :
  Exception(from,
	    "module " + mname + " not found") {
}

TclModule::ModuleAlreadyExists::ModuleAlreadyExists(const std::string& mname, 
						    const std::string& from) :
  Exception(from,
	    "module " + mname + " already loaded") {
}

TclModule::NoInitFunction::NoInitFunction(const std::string& mname,
					  const std::string& fname,
					  const std::string& from) :
#if (defined( __GNUC__)) && (!defined(__CYGWIN__))
  Exception(from,
	    "No module initialization routine \'" + fname +
	    "\' found for module " + mname + " (" + dlerror() +
	    "). If " + fname + " routine exists, you might have compiled with an unhandled C++ compiler (handled compilers are g++/eg++, DEC cxx, SGI CC, SUN CC). Under LINUX, check that you compiled using the -rdynamic option.") {
#else
  Exception(from,
	    "No module initialization routine \'" + fname +
	    "\' found for module " + mname + " (). If " + 
		fname + " routine exists, you might have compiled with an unhandled C++ compiler (handled compilers are g++/eg++, DEC cxx, SGI CC, SUN CC). Under LINUX, check that you compiled using the -rdynamic option.") {
#endif
}


TclModule::ShellNotFound::ShellNotFound(const std::string& sname,
					const std::string& from) :
  Exception(from,
	    "Unable to find shell file \'" + sname + "\'.\nSearch path includes YAV_SHELLS variable, " + TclModule::DefaultShellsPath + ", and current directory.") {
}


CommandInformation::CommandInformation(TclModule *m, const char *_name,
		 int(*fct)(TclModule *m, int argc, char **argv),
		 const char *_synopsis, const char *_description)
  : name(_name), synopsis(_synopsis), description(_description), options(0),
    globalCommand(fct), module(m) {
  Tcl_CreateCommand(m->getInterp(), (char *) name, (Tcl_CmdProc *) CommandInformation::genericCmd,
		    (ClientData) this, (Tcl_CmdDeleteProc *) 0);
}

CommandInformation::CommandInformation(const char *_name,
		 std::map < char *, CommandOption *, CStringCompare > *_options,
		 const char *_description)
  : name(_name), synopsis(0), description(_description), options(_options),
    globalCommand(0), module(0) {
}

int CommandInformation::exec(int argc, char **argv) const {
  if(globalCommand) {
    int res = (*globalCommand)(module, argc, argv);
    if(res == COMMAND_SYNOPSIS) {
      TclString *syn = getSynopsis();
      module->setResult(*syn);
      delete syn;
      return TCL_ERROR;
    }
    return res;
  }
  return TCL_OK;
}

int CommandInformation::genericCmd(ClientData cd, Tcl_Interp *interp,
			int argc, char **argv) {
  CommandInformation *cmd = (CommandInformation *) cd;
  return cmd->exec(argc, argv);
}


TclString *CommandInformation::getSynopsis(void) const {
  TclString *res = new TclString("usage: ");
  *res += name + SPACE_STRING;
  if(options) {
    *res += "< ";
    std::map < char *, CommandOption *, CStringCompare >::iterator it;
    for(it = options->begin(); it != options->end(); it++) {
      *res += it->first;
      *res += " | ";
    }
    *res += ">";
  }
  else *res += synopsis;
  return res;
}



int TclModule::init(MODULEMAINPROC initProc,
		    bool v, int argc, char **argv...) {
  initargc = argc;
  initargv = argv;
  va_list args;
  char *str;
  unsigned int n;
  if(!initProc) {
    CERR << argv[0]
	 << ": fatal: null main procedure transmitted to TclModule::init" << ENDL;
    ::exit(-1);
  }
  else mainProc = initProc;
  TclModule::verbose = v;
  nmodules = 0;
  va_start(args, argv);
  while((str = va_arg(args, char *)) != 0) nmodules++;
  va_end(args);
  modulenames = new char*[nmodules];
  va_start(args, argv);
  n = 0;
  while((str = va_arg(args, char *)) != 0)
    modulenames[n++] = str;
  va_end(args);

  n = TCL_ERROR;
  try {
    Tk_Main(argc, argv, TclModule::main);
    delete[] modulenames;
    n = TCL_OK;
  }
  catch(Exception &e) {
    e.what();
  }
  return n;
}


int TclModule::main(Tcl_Interp *interp) {
  unsigned int i;
  // initializes Tcl/Tk


  if(Tcl_Init(interp) == TCL_ERROR) return TCL_ERROR;
  if(Tk_Init(interp) == TCL_ERROR) return TCL_ERROR;
  coremodule = new TclModule(interp, "tclModules");
  
  try {
    // initializes user modules
    for(i = 0; i < nmodules; i++) {
		if (TclModule::loadDynamically==true)
      (void) loadNewModule(interp, modulenames[i]);
		else
		(void) new TclModule(coremodule->getInterp(), modulenames[i]);	
    }

    // enters application main procedure
    return (*mainProc)(coremodule, initargc, initargv);
  }
  catch(Exception& e) {
    CERR << e;
  }

  Tcl_Exit(-1);
  return TCL_OK;
}


void TclModule::newCommand(const char *name,
			   int (*fct)(TclModule *, int, char **),
			   const char *synopsis, const char *description) {
  commands[(char *) name] = new CommandInformation(this, name, fct, synopsis, description);
}

void TclModule::newCommand(const char *name,
			   std::map < char *, CommandOption *, CStringCompare > *options,
			   const char *description) {
  commands[(char *) name] = new CommandInformation(name, options, description);
}


int TclModule::eval(const char *str) const {
  if(evalFct) return (*evalFct)(*this, str);
  else {
    tickle("%s {%s}", (char *) TCLevalCommandName, str);
    return TCL_OK;
  }
}


void TclModule::display(const char *str) const {
  if(displayFct) (*displayFct)(*this, str);
  else
    tickle("%s {%s}", (char *) TCLdisplayCommandName, str);
}

void TclModule::moduleSearchPath(TclString& path) {
  char *lpath;
  TclString **paths;

  if(TclModule::RessourceModulesPath) path = TclModule::RessourceModulesPath;
  else path = "";
  lpath = getenv("TCL_MODULES");
  if(lpath) {
    if(path.getLength() > 0) path += ":";
    path += lpath;
  }
  if(TclModule::DefaultModulesPath) {
    int n = TclString(TclModule::DefaultModulesPath).split(paths, ":");
    for(int i = 0; i < n; i++) {
      // the default modules path should be right from the strat and so 
      // there's no need for anything to be added to it
      // *paths[i] += "/lib/";
      // *paths[i] += getenv("BASEDIR");
      if(path.getLength() > 0) path += ":";
      path += *paths[i];
      delete paths[i];
    }
    delete [] paths;
  }
  lpath = getenv("LD_LIBRARY_PATH");
  if(lpath) {
    if(path.getLength() > 0) path += ":";
    path += lpath;
  }
}

/* check wether a dynamic module exists and return its library path */
TclString *TclModule::findLibrary(const TclString& moduleName) {
  struct stat finfo;
  TclString **paths, *libname = 0;

  TclString path;
  TclModule::moduleSearchPath(path);

  if(verbose) COUT << "searching module " << moduleName << ENDL;
  int n = path.split(paths, ":");
  int i;
  for(i = 0; i < n && !libname; i++) {
    libname = new TclString(*paths[i]);
    while((*libname)[libname->getLength() - 1] == '/') libname->truncAtEnd(1);
    *libname += "/lib" + moduleName + ".so";
    if(verbose) COUT << "  trying " << *libname << ENDL;
    if(stat((char *) *libname, &finfo) != 0) {
      delete libname;
      libname = 0;
    }
    else {
      if(verbose) COUT << "  found " << *libname << ENDL;
    }
  }

  for(i = 0; i < n; i++) delete paths[i];
  delete [] paths;
  return libname;
}


/* return a TclModule struct from module list */
TclModule *TclModule::findModuleByName(const TclString& moduleName) {
  TclModule *m = TclModule::getHead();
  while(m && (m->getName() != moduleName)) m = m->following();
  return m;
}


void (*TclModule::resolveSymbol(const TclString& initFct))(TclModule*) {
#ifndef WIN32
  void *init, *hdl;
#else
  HANDLE_TYPE hdl;
  FARPROC  init;
#endif
  // static module
  if(!handle) hdl = dlopen(0, RTLD_NOW);
  // dynamic module
  else hdl = handle;
#ifdef WIN32
TclString s(initFct);
#else
  // try g++, eg++, or SGI CC -o32 symbol resolution
//#ifdef NAMESPACE
  TclString s(initFct + "__FPQ22sm9TclModule");
  //#else
  //TclString s(initFct + "__FP9TclModule");
  //#endif
#endif
  init = dlsym(hdl, (char *) s);
  if(init) return (void(*)(TclModule*)) init;

  // try SGI CC -n32 or -64 symbol resolution
  //#ifdef NAMESPACE
  //namespace not supported by sgi CC today...
  //#else
  //s = initFct + "__GP9TclModule";
  //#endif
  init = dlsym(hdl, (char *) s);
  if(init) return (void(*)(TclModule*)) init;

  // try DEC cxx symbol resolution
  //#ifdef NAMESPACE
  s = initFct + "__XPQ12sm9TclModule";
  //#else
  //s = initFct + "__XP9TclModule";
  //#endif
  init = dlsym(hdl, (char *) s);
  if(init) return (void(*)(TclModule*)) init;
  
  // try SUN CC symbol resolution
  //#ifdef NAMESPACE
  //namespace not supported sun CC today...
  //#else
  // size_t l;

//   s = "__0F";
//   l = initFct.getLength();
//   while(l >= 52) {
//     s += "0";
//     l -= 52;
//   }
//   if(l <= 25) s += (char) ((unsigned int) 'A' + l);
//   else s += (char) ((unsigned int) 'a' + l - 26);
//   s += initFct;
//   s += "P6JTclModule";
  //#endif
  init = dlsym(hdl, (char *) initFct);
  if(init) return (void(*)(TclModule*)) init;

  // last ressort: try symbol itself (should not work with any
  // C++ compiler anyway)
  init = dlsym(hdl, (char *) initFct);
  return (void(*)(TclModule*)) init;
}



TclModule::TclModule(Tcl_Interp *inter, const TclString& nm,
		     HANDLE_TYPE hdl) 
  throw(TclModule::NoInitFunction)
  : interp(inter), name(nm), handle(hdl), embededTclScript(false) {

  //void (*init)(TclModule*);
  // TclString initFct(name);
  
  if(verbose) COUT << "creating module " << name << ENDL;    

  // build module init function name
  // initFct.toLower();
  // initFct.toUpper(0);
  //  initFct += "_Init";

  char *moduleName=(char *) nm;
  std::map<char *,INIT_MODULE_FUNCTION,CStringCompare>::iterator itMap;
  itMap=TclModule::getModuleNameMap().find(moduleName);
  if (itMap!=TclModule::getModuleNameMap().end())
  {
    if (verbose) COUT << "got handle of module  "<< moduleName <<ENDL;
    (*(*itMap).second)(this);
  }

  /* If embededTclScript is set to true, then the script have been loaded
     in the init function.
     Otherwise, it has to be loaded now.*/
  if (!embededTclScript) {
    if (verbose) COUT << "loading tcl script dynamically for " << name << ENDL;
      loadShellScript();
  }
  else if (verbose)
    COUT << "tcl script already loaded for " << name << ENDL;
  
}


 void TclModule::loadShellScript() {
 // load shell script
  try {
    loadSource(name + ".tcl");
  }
  catch(ShellNotFound&) {
  }
 }


/* load a dynamic module */
TclModule *TclModule::loadNewModule(Tcl_Interp *interp,
				    const TclString& moduleName)
  throw(TclModule::DlopenFailed, TclModule::ModuleNotFound,
	TclModule::ModuleAlreadyExists, TclModule::NoInitFunction) {
  TclString *path = 0;
  TclModule *m = 0;
  HANDLE_TYPE handle = 0;
  char from[] = "TclModule::loadNewModule(const Tcl_Interp *, const TclString&)";
  /* check wether module was already loaded */
  m = findModuleByName(moduleName);
  if(!m) {
    /* find dynamic library path */
    path = findLibrary(moduleName);

    if(path) {
      /* load dynamic library and resolve all symbol */
      if(verbose) COUT << "loading " << *path << "...";
      handle = dlopen((char *) *path, RTLD_NOW
#if defined _SOLARIS_ || defined _LINUX_ || defined _SGI_
		      | RTLD_GLOBAL
#endif
		      );
      delete path;

      if(handle) {
	if(verbose) COUT << " succeded!" << ENDL;
	// create new module
	m = new TclModule(interp, moduleName, handle);
	return m;
      }
      else {
	if(verbose) COUT << " failed!" << ENDL;
	throw DlopenFailed(moduleName.tostring(), from);
      }
    }

    /* module not found */
    throw ModuleNotFound(moduleName.tostring(), from);
  }

  /* module already loaded */
  throw ModuleAlreadyExists(moduleName.tostring(), from);
}



TclModule::~TclModule() {
  if(verbose) COUT << "deleting module " << name << ENDL;
  // destroy objects created by this module
  std::list<TclObjectKind *>::reverse_iterator itm;
  for(itm = descriptor.rbegin(); itm != descriptor.rend(); itm++) {
    bool b = (*itm)->getFirstObject() != 0;
    if(verbose) {
      COUT << "  deleting ObjectKind " << (*itm)->getName();
      if(b) COUT << " (deleting: ";
	  STD_FLUSH(COUT);
    }
    delete *itm;
    if(verbose) {
      if(b) COUT << ")";
      COUT << ENDL;
    }
  }
  // delete all associated tcl functions
  std::map< char *, CommandInformation *, CStringCompare >::iterator it;
  for(it = commands.begin(); it != commands.end(); it++) {
    if(it->second->getCommand()) Tcl_DeleteCommand(interp, it->first);
  }
  /** remove entry from module name map */
  char *moduleName=(char *) name;
  TclModule::getModuleNameMap().erase(moduleName);
  
  // close dynamic library is any
  if(handle) dlclose(handle);
}

/** This function must be called by the init function 
    if it statically loads the tcl script corresponding
    to the module. Otherwise, the tcl script will be
    dynamically loaded. */
void TclModule::tclScriptLoaded()
{ 
  embededTclScript = true;
  if (verbose)
  {
    COUT << "tcl script for " << name << " statically loaded." << ENDL;
  }
  return; 
}

/* display function list of a module */
void TclModule::displayFunctionList(void) {
  TclString msg;
  std::map< char *, CommandInformation *, CStringCompare >::iterator it;

  it = commands.begin();
  if(it != commands.end()) {
    msg = it->first;
    it++;
    for(; it != commands.end(); it++) {
      if(msg.getLength() + strlen(it->first) < 75)
	msg += SPACE_STRING + it->first;
      else {
	display(msg);
	msg = "";
      }
    }
    display(msg);
  }
}



void TclModule::setResult(char *s...) const {
  va_list args;

  // format string
  va_start(args, s);
  TclString str(s, args);
  va_end(args);
  
  // eval command
  Tcl_SetResult(interp, (char *) str, TCL_VOLATILE);
}



int TclModule::tickle(const char *s...) const {
  va_list args;

  // format string
  va_start(args, s);
  TclString str(s, args);
  va_end(args);
  
  // evaluation
  return tickle((const TclString) str);
}


int TclModule::tickle(const TclString &str) const {
  if(Tcl_GlobalEval(interp, (char *) str) == TCL_ERROR) {
	  TclString strError;
	  strError.sprintf("tk_messageBox -type ok -message {Tickle error: %s while executing %s\n stack trace: %s } ",
           interp->result, (char *) str,  Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY));
    Tcl_GlobalEval(interp,(char *)strError);

    CERR << "Tickle error: " << interp->result << ENDL
	 << "     while executing: " << (char *) str << ENDL
	 << "     in module: " << name << ENDL
	 << "     stack trace: " << ENDL
	 << Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY) << ENDL;
    return TCL_ERROR;
  }
  return TCL_OK;
}
 


void TclModule::shellSearchPath(TclString& path) {
  char *lpath;

  if(TclModule::RessourceShellsPath) path = TclModule::RessourceShellsPath;
  else path = "";
  lpath = getenv("YAV_SHELLS");
  if(lpath) {
	  if(path.getLength() > 0) path += "::";
    path += lpath;
  }
  if(TclModule::DefaultShellsPath) {
	  if(path.getLength() > 0) path += "::";
    path += TclModule::DefaultShellsPath;
  }
}


/* check wether a script file exists and executes it */
void TclModule::loadSource(const TclString& sname)
  throw(TclModule::ShellNotFound) {

  char from[] = "TclModule::loadSource(TclInterp *, const TclString&)";
  TclString **paths, *shellName = 0;
  struct stat finfo;

  if(verbose) COUT << "loading TCL script " << sname << ENDL;
  if(sname[0U] == '/') {
    if(verbose) COUT << "  trying " << sname << ENDL;
    if(stat((char *) sname, &finfo) == 0) {
      if(verbose) COUT << "  found " << sname << ENDL;
      Tcl_EvalFile(interp, (char *) sname);
      return;
    }
    else {
      if(verbose) COUT << "  did not found " << sname << ENDL;
      throw ShellNotFound(sname.tostring(), from);
    }
  }

  TclString path;
  TclModule::shellSearchPath(path);

  // path.split(paths, ",");
  //TclString**& paths;
  unsigned int n = 0;
  unsigned int i = 0, j, k;

  // First, determine the number of groups once splited.
  do {
    if(i < path.getLength() - 1 && strchr(":", ((char*)path)[i]) ) {
      j = i + 1;
      while (j < path.getLength() - 1 && strchr(":", ((char *)path)[j] ) ) 
	++j;
      // if we didn't move at least once, them it wasn't a "::" string, 
      // but just a single ":"
      // so we don't need to create another group
      if (j == i + 1) 
	--n;
      // skip to first character following the ":"s
      i = j;
    }

    if(i < path.getLength() - 1) {
      n++;
      while(i < path.getLength() - 1 && !strchr(":", ((char*)path)[i])) 
	++i;
    }
  }
  while(i < path.getLength() - 1);
  

  // Now create the groups.
  paths = new TclString*[n];
  for(k = 0, i = 0; k < n; k++) {
    if (i < path.getLength() - 1 && strchr(":", ((char*)path)[i])) {
      j = i +1;
      // skip leading "::"
      if (j < path.getLength() - 1 && strchr(":", ((char*)path)[i])) {    
	do {
	  ++j;
	}
	while (j < path.getLength() - 1 && strchr(":", ((char*)path)[j]));
	i = j;
      }
    }
    j = i + 1;

    // add to current group while we do not encounter "::"
    while (j < path.getLength() - 1) {
      if (strchr(":", ((char*)path)[j]) && j < path.getLength() - 2 && strchr(":", ((char*)path)[j+1]))
	break;
      else
	++j;
    }
    if (j < path.getLength() - 1)
      paths[k] = new TclString(path, i, j - 1); //do not include the ":"
    else
      paths[k] = new TclString(path, i, j);     //but include the last character of the string
    i = j;
  }


  //int i;
  for(i = 0; i < n && !shellName; i++) {
    shellName = new TclString(*paths[i]);
    while((*shellName)[shellName->getLength() - 1] == '/')
      shellName->truncAtEnd(1);
    *shellName += TclString("/") + sname;
    if(verbose) COUT << "  trying " << *shellName << ENDL;
    if(stat((char *) *shellName, &finfo) != 0) {
      delete shellName;
      shellName = 0;
    }
    else {
      if(verbose) COUT << "  found " << *shellName << ENDL;
    }
  }

  for(i = 0; i < n; i++) delete paths[i];
  delete [] paths;
  if(shellName) {
    if(verbose) COUT << "  loading " << *shellName << ENDL;
    Tcl_EvalFile(interp, (char *) *shellName);
    delete shellName;
    return;
  }

  if(verbose) COUT << "  did not found " << sname << ENDL;
  throw ShellNotFound(sname.tostring(), from);
}



const char *TclModule::booleans[2] = { "false", "true" };

const char *TclModule::boolString(bool b) {
  if(b) return booleans[1];
  else return booleans[0];
}


bool TclModule::boolValue(char *str, bool& value) {
  if(!strcmp(str, booleans[0])) {
    value = false;
    return true;
  }
  if(!strcmp(str, booleans[1])) {
    value = true;
    return true;
  }
  return false;
}



// default TCL evaluation command
int TclModule::Evaluate(TclModule *m, int argc, char **argv) {
  if(argc >= 2) return m->eval(argv[1]);
  return TCL_OK;
}

// default TCL display command
int TclModule::Display(TclModule *m, int argc, char **argv) {
  TclString s;
  if(argc >= 2){
    int i;
    for(i=1;i<argc;++i){
      s+=argv[i];
      s+=" ";
    }
  }
   //m->display(argv[1]);
  m->setResult(s);
  return TCL_OK;
}

int TclModule::Functions(TclModule *m, int argc, char **argv) {
  TclString res;
  TclModule *l;

  if(argc != 2) return COMMAND_SYNOPSIS;
  l = TclModule::getHead();
  while(l && (l->getName() != (const char *) argv[1])) l = l->following();
  if(!l) {
    m->setResult("not a module \'%s\'", argv[1]);
    return TCL_ERROR;
  }

  std::map < char *, CommandInformation *, CStringCompare >::iterator it;
  for(it = l->getCommands().begin(); it != l->getCommands().end(); it++) {
    res += it->first + SPACE_STRING;
  }
  m->setResult(res);
  return TCL_OK;
}


int TclModule::Options(TclModule *m, int argc, char **argv) {
  TclModule *l;
  TclString res;
  std::map < char *, CommandInformation *, CStringCompare >::iterator it;

  if(argc != 2) return COMMAND_SYNOPSIS;
  l = TclModule::getHead();
  while(l) {
    it = l->getCommands().find(argv[1]);
    if(it != l->getCommands().end()) {
      std::map < char *, CommandOption *, CStringCompare > *options
	= it->second->getOptions();
      if(options) {
	std::map < char *, CommandOption *, CStringCompare >::iterator it2;
	for(it2 = options->begin(); it2 != options->end();) {
	  res += it2->first;
	  it2++;
	  if(it2 != options->end()) res += SPACE_STRING;
	}
      }
      m->setResult(res);
      return TCL_OK;
    }
    l = l->following();
  }

  m->setResult("not a function \'%s\'", argv[1]);
  return TCL_OK;
}


int TclModule::Module(TclModule *m, int argc, char **argv) {
  if(argc < 2) return COMMAND_SYNOPSIS;

  if(!strcmp(argv[1], "-list")) {
    if(argc != 2) return COMMAND_SYNOPSIS;
    TclString res;
    TclModule *l;

    for(l = TclModule::getHead(); l; l = l->following()) {
      res += l->getName();
      if(l->following()) res += SPACE_STRING;
    }
    m->setResult(res);
  }

  else if(!strcmp(argv[1], "-load")) {
    if(argc != 3) return COMMAND_SYNOPSIS;
    
    try {
      TclModule::loadNewModule(m->getInterp(), argv[2]);
    }
    catch(TclModule::DlopenFailed& e) {
      m->setResult("load failed: %s", (const char *) e);
      return TCL_ERROR;
    }
    catch(TclModule::ModuleNotFound&) {
      m->setResult("unable to find library lib%s.so", argv[2]);
      return TCL_ERROR;
    }
    catch(TclModule::ModuleAlreadyExists&) {
      m->setResult("module %s was already loaded", argv[2]);
      return TCL_OK;
    }
    catch(TclModule::NoInitFunction) {
      m->setResult("module init function not found");
      return TCL_OK;
    }
    m->setResult("module %s loaded", argv[2]);
  }

  else if(!strcmp(argv[1], "-delete")) {
    if(argc != 3) return COMMAND_SYNOPSIS;
    TclModule *l = TclModule::findModuleByName(argv[2]);
    if(!l) {
      m->setResult("no module %s", argv[2]);
      return TCL_ERROR;
    }
    else {
      if(l == m) {
	m->setResult("module %s refuses to kill itself!", argv[2]);
	return TCL_ERROR;
      }
      delete l;
      m->tickle("if [winfo exists .yav] {yav::updateYavMenu .yav.menu}");
      m->setResult("module %s deleted", argv[2]);
      return TCL_OK;
    }
  }

  else if(!strcmp(argv[1], "-path")) {
    TclString str;
    TclModule::moduleSearchPath(str);
    m->setResult(str);
    return TCL_OK;
  }

  else return COMMAND_SYNOPSIS;

  return TCL_OK;
}


// check wether a module (= its dynamic library) exists
int TclModule::IsModuleAvailable(TclModule *m, int argc, char **argv) {
  TclString *str;
  char usage[] = "usage: isModuleAvail name";

  if(argc != 2) {
    m->setResult(usage);
    return TCL_ERROR;
  }

  str = TclModule::findLibrary(argv[1]);
  if(!str) m->setResult("no");
  else {
    delete str;
    m->setResult("yes");
  }
  return TCL_OK;
}


// check wether a module is loaded
int TclModule::IsModuleLoaded(TclModule *n, int argc, char **argv) {
  char usage[] = "usage: isModuleLoaded name";
  TclModule *m;

  if(argc != 2) {
    n->setResult(usage);
    return TCL_ERROR;
  }

  m = TclModule::findModuleByName(argv[1]);
  if(!m) n->setResult("no");
  else n->setResult("yes");

  return TCL_OK;
}


/* display help */
int TclModule::Help(TclModule *m, int argc, char **argv) {
  char res[] = "hope that helps";
  TclString *syn = 0;
  std::map< char *, CommandOption *, CStringCompare >::iterator it2;
  
  if(argc > 4) return COMMAND_SYNOPSIS;

  

  /* help on help */
  if(argc < 2) {
    TclString s("usage: help < function | [all");
    for(TclModule *n = TclModule::getHead(); n; n = n->following())
      s += "|" + n->getName();
    s += "] >";
    m->setResult(s);
    return TCL_OK;
  }

  /* help on module / function */

  /* list all functions of all modules */
  if(!strcmp(argv[1], "all")) {
    for(TclModule *n = TclModule::getHead(); n; n = n->following()) {
      m->display(n->getName() + ":");
      n->displayFunctionList();
    }
    m->setResult(res);
  }
  
  /* display one module or function help */
  else {
    TclModule *n = TclModule::findModuleByName(argv[1]);
    
    /* list module functions */
    if(n) {
      n->displayFunctionList();
      m->setResult(res);
    }
    
    /* display function help */
    else {
      for(n = TclModule::getHead(); n; n = n->following()) {
	std::map < char *, CommandInformation *, CStringCompare >::iterator it;
	it = n->getCommands().find(argv[1]);
	if(it != n->getCommands().end()) {
	  switch(argc) {
	  case 2:
	    m->display(TclString("name: ") + it->first);
	    syn = it->second->getSynopsis();
	    m->display("synopsis: " + *syn);
	    delete syn;
	    m->display(TclString("description: ") + it->second->getDescription());
	    m->setResult(res);
	    return TCL_OK;
	    break;
	    
	  case 3:
	    if(!strcmp(argv[2], "synopsis")) {
	      TclString *syn = it->second->getSynopsis();
	      m->setResult(*syn);
	      delete syn;
	      return TCL_OK;
	    }
	    else if(!strcmp(argv[2], "description")) {
	      m->setResult(it->second->getDescription());
	      return TCL_OK;
	    }
	    else if(!strcmp(argv[2], "creation")) {
 	      TclString res = argv[1];
 	      if ((res[res.getLength()-2] == '%') && (res[res.getLength()-1] == 'u')){
  		res.truncAtEnd(2);
 		TclObjectKind *ok = TclObjectKind::findByName(res);
		m->setResult("Creation synospis: new %s %s", (char *)res, ok->getCreationSynopsis());
 	      } return TCL_OK;
	    }
	    else return COMMAND_SYNOPSIS;
	    break;

	  case 4:
	    it2 = it->second->getOptions()->find(argv[2]);
	    if(it2 != it->second->getOptions()->end()) {
	      if(!strcmp(argv[3], "synopsis")) {
		TclString str;
		str.sprintf("usage: %s %s", it->first, it2->first);
		if(it2->second->getOptionSynopsis()) {
		  str += SPACE_STRING;
		  str += it2->second->getOptionSynopsis();
		}
		m->setResult(str);
	      }
	      else if(!strcmp(argv[3], "description")) {
		TclString str(it2->second->getOptionDescription());
		m->setResult(str);
	      }
	      else return COMMAND_SYNOPSIS;
	      return TCL_OK;
	    }
	  
	    else if(!strcmp(argv[2], "-kind")) {
	      char *descr = "returns object kind (object name whithout numerical id)";
	      if(!strcmp(argv[3], "synopsis")) {
		TclString str;
		str.sprintf("usage: %s -kind", it->first);
		m->setResult(str);
	      }
	      else if(!strcmp(argv[3], "description")) {
		m->setResult(descr);
	      }
	      else return COMMAND_SYNOPSIS;
	      return TCL_OK;
	    }

	    else {
	      m->setResult("%s is not a valid option of %s", argv[2], it->first);
	      return TCL_ERROR;
	    }
	    break;

	  default:
	    return COMMAND_SYNOPSIS;
	    break;
	  }
	}
      }
      
      m->setResult(TclString("function \'") + argv[1] + "\' help not found");
      return TCL_ERROR;
    }
  }

  return TCL_OK;
}





int TclModule::LoadSource(TclModule *m, int argc, char **argv) {
  char usage[] = "loadSource %s";

  if(argc != 2) {
    m->setResult(usage);
    return TCL_ERROR;
  }

  try {
    m->loadSource(argv[1]);
  }
  catch(TclModule::ShellNotFound&) {
	  m->setResult(TclString("\'") + argv[1] + 
		 "\' not found, set YAV_SHELLS variable to proper value");
    return TCL_ERROR;
  }

  m->setResult(TclString(argv[1]) + " evaluated");
  return TCL_OK;
}



/* set a new display message function */
int TclModule::SetDisplayMessageFunction(TclModule *m, int argc, char **argv) {
  if(argc > 2) {
    m->setResult("setDisplayFunction [%%s]");
    return TCL_ERROR;
  }

  if(argc < 2) {
    TclModule::setDisplayFunction(TclModule::DefaultDisplay);
    m->setResult("default display message function used");
  }
  else {
    TclModule::setTCLDisplayFunction(argv[1]);
    m->setResult(TclString(argv[1]) + " display message function used");
  }

  return TCL_OK;
}



/* set a new eval function */
int TclModule::SetEvalFunction(TclModule *m, int argc, char **argv) {
  if(argc > 2) {
    m->setResult("setEvalFunction [%%s]");
    return TCL_ERROR;
  }

  if(argc < 2) {
    TclModule::setEvalFunction(TclModule::DefaultEval);
    m->setResult("default eval function used");
  }
  else {
    TclModule::setTCLEvalFunction(argv[1]);
    m->setResult(TclString(argv[1]) + " eval function used");
  }

  return TCL_OK;
}


int TclModule::FlushEvents(TclModule *m, int argc, char **argv) {
  while(Tcl_DoOneEvent(TCL_ALL_EVENTS | TCL_DONT_WAIT));
  m->setResult("events flushed");
  return TCL_OK;
}

int TclModule::MainLoop(TclModule *m, int argc, char **argv) {
  TclModule::run();
  return TCL_OK;
}
int TclModule::ActivateActivables(TclModule *m, int argc, char **argv) {
  do {
    while(Tcl_DoOneEvent(TCL_ALL_EVENTS | TCL_DONT_WAIT));
    Activable::activate();
  } while (Activable::getActivatedObjectsCount()>0);
  return TCL_OK;
}

int TclModule::EndScript(TclModule *m, int argc, char **argv) {
  m->setResult("script terminated");
  return TCL_OK;
}

int TclModule::Shell(TclModule *m, int argc, char **argv) {
  if(argc != 2 || strcmp(argv[1], "-path")) return COMMAND_SYNOPSIS;
  TclString str;
  TclModule::shellSearchPath(str);
  m->setResult(str);
  return TCL_OK;
}

void TclModule::activateWhenIdle(ClientData clientData) {
    if(Activable::getActivatedObjectsCount()) 
    Activable::activate();
  //  Tk_DoWhenIdle(activateWhenIdle, (ClientData) 0);
  CERR << "activateWhenIdle" << ENDL;

}

/* script: read a script file and call the eval function for each line
   Script is evaluated asynchronously: a new thread is created for each
   command.*/
int TclModule::StartScript(TclModule *m, int argc, char **argv) {
  int i;
  char usage[] = "usage: script %%s [%%s [%%s...]]";
  TclString line, command;
  bool firstLine = true, cont = true;
  IFSTREAM f;

  if(argc < 2) goto error;
#ifdef WIN32
  f.open(argv[1],IOS::binary);
#else
  f.open(argv[1]);
#endif
  if(!f) {
    m->setResult(TclString("unable to open \'") + argv[1] + "\'");
    return TCL_ERROR;
  }

  /* transmit global arguments */
  for(i = 1; i < argc; i++) m->tickle("set arg%i %s", i - 1, argv[i]);


  /* eval script file line by line */
  while(cont && !f.eof()) {
    f >> line;
    
    /* Special first lines management; if script file begins with
           #!...
           exec...
       both lines are skipped. Thus we can write executable scripts
       beginning with:  #!/bin/sh
                        exec /my-home-dir/my-interpretor $0
       When script file ($0) is transmitted to my-interpretor, the two 
       shell relative first lines are discarded. */
    if(firstLine) {
      firstLine = false;
      if(!line.compare("#!", 2)) {
	f >> line;
	if(!line.compare("exec", 4)) f >> line;
      }
    }
    
    command += line + "\n";
    if(Tcl_CommandComplete((char *) command)) {
      /* suppress white characters */
      i = 0;
      while((unsigned int) i < command.getLength() &&
	    (command[(unsigned int) i] == ' ' ||
	     command[(unsigned int) i] == '\t')) i++;
      
      /* the command is neither empty nor a comment */
      if(command[(unsigned int) i] != '\n' &&
	 command[(unsigned int) i] != '#') {
	m->resetResult();
	if(m->eval((char *) command + i) == TCL_ERROR) {
	  m->display("Error: while executing " + TclString(command, i, command.getLength() - 1));
	  m->display(m->getInterp()->result);
	  CERR << "Error: while executing " << TclString(command, i, command.getLength() - 1) << ENDL;
	  CERR << m->getInterp()->result << ENDL;
	}
	if(!strcmp(m->getInterp()->result, "script terminated")) cont = false;
	while(Tcl_DoOneEvent(TCL_ALL_EVENTS | TCL_DONT_WAIT)) ;
      }
      
      command = EMPTY_STRING;
    }
  }

  f.close();
  m->resetResult();
  if(cont) m->setResult("script evaluated");
  else m->setResult("script terminated");

  return TCL_OK;

 error:
  m->setResult(usage);
  return TCL_ERROR;
}
  
// default display function
void TclModule::DefaultDisplay(const TclModule &, const char *str) {
  COUT << str << ENDL;
}


// default eval function
int TclModule::DefaultEval(const TclModule &m, const char *str) {
  return Tcl_RecordAndEval(m.getInterp(), (char *) str, TCL_EVAL_GLOBAL);  
}


// default main loop function
void TclModule::DefaultMainLoop(unsigned int n) {
  for(unsigned int i = 0; i < n; i++) {
    // if no module requires to handle X events, loop on TK events only
    if(!Activable::getActivatedObjectsCount()) Tcl_DoOneEvent(TCL_ALL_EVENTS) ;
    
    // else ...
    else {
      // ... handle TK events and don't block ...
      while(Tcl_DoOneEvent(TCL_ALL_EVENTS | TCL_DONT_WAIT));
      // ... call modules handler
      Activable::activate();
    }
  }
}


void TclModule::exit(int ret) {
  TclString cleanupFct;
  //void (*cleanup)(TclModule*);

  // call user cleanup functions
  for(TclModule *mod = TclModule::getTail(); mod; mod = mod->previous()) {
    cleanupFct = mod->getName();
    cleanupFct.toLower();
    cleanupFct.toUpper(0);
    cleanupFct += "_Cleanup";
    //    cleanup = mod->resolveSymbol(cleanupFct);
    //    if(cleanup) (*cleanup)(mod);
  }
  // destroy any object created
  while(AutoLink<TclModule>::getHead()) delete AutoLink<TclModule>::getTail();

  Tcl_Exit(ret);
}


int TclModule::Exit(TclModule *m, int argc, char **argv) {
  TclModule::exit();
  return TCL_OK;
}


int TclModule::xgrabCmd(TclModule *m, int argc, char **argv) {
  if(argc != 3) return COMMAND_SYNOPSIS;
  if(!m->windowExists(argv[1])) {
    m->setResult("invalid window name");
    return TCL_ERROR;
  }

  Tk_Window tkwin = Tk_NameToWindow(m->getInterp(), argv[1],
				    Tk_MainWindow(m->getInterp()));
  #ifndef WIN32
  WindowDump(Tk_Display(tkwin), Tk_WindowId(tkwin), argv[2]);
#else
  WindowDump(tkwin, Tk_WindowId(tkwin), argv[2]);
#endif
  m->setResult(argv[2]);

  return TCL_OK;
}


bool TclModule::readrc(char *rfile, TclString& init,
		       unsigned int& nbModules, TclString*& modules, bool _verbose) {
  TclString rc, line;
  IFSTREAM ifile;
  
  if(_verbose) COUT << "Searching for ressource file " << rfile << ENDL;
  if(!rfile) {
    if(getenv("HOME")) {
      rc.sprintf("%s/.yavrc", getenv("HOME"));
      if(_verbose) COUT << "  trying " << rc << "... ";
      ifile.clear();
      ifile.open((char *) rc);
      if(_verbose) {
	if(!ifile) COUT << "failed" << ENDL;
	else COUT << "succeded" << ENDL;
      }
    }
    if(!ifile && getenv("YAV_SHELLS")) {
      rc.sprintf("%s/.yavrc", getenv("YAV_SHELLS"));
      if(_verbose) COUT << "  trying " << rc << "... ";
      ifile.clear();
      ifile.open((char *) rc);
      if(_verbose) {
	if(!ifile) COUT << "failed" << ENDL;
	else COUT << "succeded" << ENDL;
      }
    }
    if(!ifile) {
      rc.sprintf("%s/.yavrc", DefaultRessourceFileLocation);
      if(_verbose) COUT << "  trying " << rc << "... ";
      ifile.clear();
      ifile.open((char *) rc);
      if(_verbose) {
	if(!ifile) COUT << "failed" << ENDL;
	else COUT << "succeded" << ENDL;
      }
    }
  }
  else {
    rc = rfile;
    if(_verbose) COUT << "  trying " << rc << "... ";
    ifile.clear();
    ifile.open((char *) rc);
    if(_verbose) {
      if(!ifile) COUT << "failed" << ENDL;
      else COUT << "succeded" << ENDL;
    }
    if(!ifile && getenv("HOME")) {
      rc.sprintf("%s/%s", getenv("HOME"), rfile);
      if(_verbose) COUT << "  trying " << rc << "... ";
      ifile.clear();
      ifile.open((char *) rc);
      if(_verbose) {
	if(!ifile) COUT << "failed" << ENDL;
	else COUT << "succeded" << ENDL;
      }
    }
    if(!ifile && getenv("YAV_SHELLS")) {
      rc.sprintf("%s/%s", getenv("YAV_SHELLS"), rfile);
      if(_verbose) COUT << "  trying " << rc << "... ";
      ifile.clear();
      ifile.open((char *) rc);
      if(_verbose) {
	if(!ifile) COUT << "failed" << ENDL;
	else COUT << "succeded" << ENDL;
      }
    }
    if(!ifile) {
      rc.sprintf("%s/%s", DefaultRessourceFileLocation, rfile);
      if(_verbose) COUT << "  trying " << rc << "... ";
      ifile.clear();
      ifile.open((char *) rc);
      if(_verbose) {
	if(!ifile) COUT << "failed" << ENDL;
	else COUT << "succeded" << ENDL;
      }
    }
  }

  if(!ifile) {
    if(_verbose) COUT << "unable to find " << rfile << " ressource file" << ENDL;
    return false;
  }

  while(!ifile.eof()) {
    ifile >> line;

    if(line.getLength() >= 1 && line[0U] != '#') {

      if(!line.compare("init:", 5)) {
	try {
	  init = TclString(line, 6, line.getLength() - 1);
	}
	catch(TclString::OutOfString&) {
	}
	catch(TclString::InvalidIndex&) {
	}
      }
      
      else {
	TclString **strings;
	int n = line.split(strings, ": ");
	if(!line.compare("libraries:", 10)) {
	  if(n >= 2) {
	    strings[1]->expand();
	    TclString path(*strings[1]);
	      for(int i = 2; i < n; i++) {
		strings[i]->expand();
		path += ":" + *strings[i];
	      }
	      TclModule::RessourceModulesPath = new char[path.getLength()+1];
	      strcpy(TclModule::RessourceModulesPath, (char *) path);
	  }
	  else
	    COUT << "TclModule::readrc: warning: invalid libraries path: "
		 << line << ENDL;
	}
	else if(!line.compare("shells:", 7)) {
	  if(n >= 2) {
	    strings[1]->expand();
	    TclString path(*strings[1]);
	    for(int i = 2; i < n; i++) {
	      strings[i]->expand();
		path += ":" + *strings[i];
	    }
	    TclModule::RessourceShellsPath = new char[path.getLength()+1];
	    strcpy(TclModule::RessourceShellsPath, (char *) path);
	  }
	  else
	    COUT << "TclModule::readrc: warning: invalid TCL shells path: "
		 << line << ENDL;
	}
	else if(!line.compare("modules:", 8)) {
	  if(n > 1) {
	    nbModules = n - 1;
	    modules = new TclString[nbModules];
	    for(int i = 1; i < n; i++) {
	      modules[i-1] = *strings[i];
	    }
	  }
	}
	else {
	  COUT << "TclModule::readrc: warning: unknown ressource in .yavrc: "
	       << line << ENDL;
	}
	for(int i = 0; i < n; i++) delete *strings[i];
	if(n >= 0) delete[] strings;
      }   
    }
  }
  ifile.close();
  return true;
}


/* print a prompt in text interface */
void TclModule::Prompt(Tcl_Interp *interp, int partial) {
  Tcl_Channel outChannel = Tcl_GetChannel(interp, "stdout", NULL);

  if(outChannel) {
    if(partial) Tcl_Write(outChannel, "...> ", 5);
    else Tcl_Write(outChannel, "> ", 2);
    Tcl_Flush(outChannel);
  }
}

 int TclModule::LimitedVersion(TclModule *m, int argc, char **argv) {

#ifdef LIMITED_VERSION
   m->setResult("true");
#else
   m->setResult("false");
#endif
    return TCL_OK;
 }


} // namespace yav


