#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "modules/TclObjectKind.h"
#include <basics/Constants.h>
#include <modules/TclObjectKind.h>
#ifdef __GNUC__
#if ( __GNUC__ < 3 )
#include <strstream>
#else
#include <sstream>
#endif // __GNUC__ < 3
#endif

namespace yav {

AUTOLINK_STATIC_VARIABLES(TclObjectKind);


WarnedObject::WarnedObject(const std::type_info& t1) {
  warnMe(t1);
}

WarnedObject::WarnedObject(const std::type_info& t1, const std::type_info& t2) {
  warnMe(t1);
  warnMe(t2);
}

WarnedObject::WarnedObject(const std::type_info& t1, const std::type_info& t2,
			   const std::type_info& t3) {
  warnMe(t1);
  warnMe(t2);
  warnMe(t3);
}

WarnedObject::~WarnedObject() {
  std::list<const std::type_info *>::const_iterator it;
  for(it = type.begin(); it != type.end(); it++) {
    TclObjectKind *kind = TclObjectKind::findByKind(**it);
    if(kind) kind->doNotWarnMe(this);
  }
}

void WarnedObject::warnMe(const std::type_info& t) {
  TclObjectKind *kind;
  kind = TclObjectKind::findByKind(t);
  if(kind) {
    std::list<const std::type_info *>::iterator it;
    for(it = type.begin(); it != type.end(); it++)
      if(**it == t) return;
    kind->warnMe(this);
    type.push_front(&t);
  }
}

void WarnedObject::doNotWarnMe(const std::type_info& t) {
  TclObjectKind *kind = TclObjectKind::findByKind(t);
  if(kind) kind->doNotWarnMe(this);

  std::list<const std::type_info *>::iterator it;
  for(it = type.begin(); it != type.end(); it++) {
    if(**it == t) {
      type.erase(it);
      return;
    }
  }
}

TclObjectKind::TclObjectKind(TclModule *m,
			     const std::type_info& type,
			     const char *n,
			     const char *real,
			     const char *synopsis,
			     int (*args)(TclModule *, int, char **),
			     int (*list)(TclModule *, int, char **))
  throw(TclObjectKind::InvalidParameters)
  : EmptyObjectsList<TclObject>(), name(n), realName(real),
    createFromArgs(args), objectsList(list), globalCommand(0), module(m),
    objectsType(type) {
  if(TclModule::verbose) COUT << "  creating ObjectKind " << name << ENDL;    
  m->newDescriptor(this);
  objects.warnMe(this);
  if(!name || !createFromArgs || !objectsList)
    throw InvalidParameters("TclObjectKind::TclObjectKind");
  if(synopsis) newSynopsis = synopsis;
  else newSynopsis.sprintf("(no synopsis provided for 'new %s')", n);
  newOption("-kind", (TclObject::OPTION) &TclObject::kindCmd, 0,
	    "returns the name of the object kind associated with a given object");
  newOption("-realName", (TclObject::OPTION) &TclObject::realNameCmd, 0,
	    "returns the expanded name of the type of object");
  char *commandName=new char[200];
  sprintf(commandName,"%s%%u",n);
  char *expandedName=new char[200];
  sprintf(expandedName,"%s management command",real);
  m->newCommand((char *)commandName,getOptions(),(char *) expandedName);
}
TclObjectKind::TclObjectKind(TclModule *m,
			     const std::type_info& type,
			     const std::type_info& motherType,
			     const char *n,
			     const char *real,
			     const char *synopsis,
			     int (*args)(TclModule *, int, char **),
			     int (*list)(TclModule *, int, char **))
  throw(TclObjectKind::InvalidParameters)
  : EmptyObjectsList<TclObject>(), name(n), realName(real),
    createFromArgs(args), objectsList(list), globalCommand(0), module(m),
    objectsType(type) {
  if(TclModule::verbose) COUT << "  creating ObjectKind " << name << ENDL;    
  m->newDescriptor(this);
  objects.warnMe(this);
  if(!name || !createFromArgs || !objectsList)
    throw InvalidParameters("TclObjectKind::TclObjectKind");
  if(synopsis) newSynopsis = synopsis;
  else newSynopsis.sprintf("(no synopsis provided for 'new %s')", n);
  newOption("-kind", (TclObject::OPTION) &TclObject::kindCmd, 0,
	    "returns the name of the object kind associated with a given object");
  newOption("-realName", (TclObject::OPTION) &TclObject::realNameCmd, 0,
	    "returns the expanded name of the type of object");
  addOptions(motherType);

  char *commandName=new char[200];
  sprintf(commandName,"%s%%u",n);
  char *expandedName=new char[200];
  sprintf(expandedName,"%s management command",real);
  m->newCommand((char *)commandName,getOptions(),(char *) expandedName);
}


TclObjectKind::TclObjectKind(TclModule *m,
			     const std::type_info& type,
			     const char *n,
			     const char *r,
			     int (*args)(TclModule *, int, char **),
			     int (*list)(TclModule *, int, char **),
			     const char *synopsis)
  throw(TclObjectKind::InvalidParameters)
  : EmptyObjectsList<TclObject>(), name(n), realName(r),
    createFromArgs(args), objectsList(list), globalCommand(0), module(m),
    objectsType(type) {
  if(TclModule::verbose) COUT << "  creating ObjectKind " << name << ENDL;    
  m->newDescriptor(this);
  objects.warnMe(this);
  if(!name || !createFromArgs || !objectsList)
    throw InvalidParameters("TclObjectKind::TclObjectKind");
  if(synopsis) newSynopsis = synopsis;
  else newSynopsis.sprintf("(no synopsis provided for 'new %s')", n);
  newOption("-kind", (TclObject::OPTION) &TclObject::kindCmd, 0,
	    "returns the name of the object kind associated with a given object");
  newOption("-realName", (TclObject::OPTION) &TclObject::realNameCmd, 0,
	    "returns the expanded name of the type of object");
  /*  newOption("-nickname", (TclObject::OPTION) &TclObject::nicknameCmd, 0,
      "returns the nickname of the object");*/
}


TclObjectKind::~TclObjectKind() {
  objects.doNotWarnMe(this);
  deleteAllObjects();
}


void TclObjectKind::deleteAllObjects(void) {
  while(!objects.isEmpty()) {
    if(TclModule::verbose) {
      COUT << ((TclObject *) objects.getTail())->name();
      if(objects.getTail()->getPrevious()) COUT << " ";
    }
    bool b = TclModule::verbose;
    TclModule::verbose = false;
    delete objects.getTail();
    TclModule::verbose = b;
  }
}

 
void TclObjectKind::newOption(char *name, int(TclObject::*cmd)(int, char **),
			      const char *synopsis, const char *description) {
  options[name] = new CommandOption(cmd, synopsis, description);
}
void TclObjectKind::newOption(char *name, int(*cmd)(TclObject *,int, char **),
			      const char *synopsis, const char *description) {
  options[name] = new CommandOption(cmd, synopsis, description);
}
void TclObjectKind::newOption(char *name, CommandOption *opt) {
  options[name] = opt;
}


int TclObjectKind::newCmd(TclModule *m, int argc, char **argv) {
  TclObjectKind *any;
  int ret;

  if(argc > 1) {
    for(any = AutoLink<TclObjectKind>::getHead(); any; any = any->following()) {
      if(!strcmp(any->getName(), argv[1])) {
	 try{
	    ret = (*any->createFromArgs)(m, argc - 1, argv + 1);
	 }
	 catch(Exception& e){
#ifdef __GNUC__
#if ( __GNUC__ < 3 )
	   std::ostrstream os;
	   os << e << ENDS;
	   m->setResult("%s",os.str());
#else
	   std::ostringstream os;
	   os << e << ENDS;
	   m->setResult("%s",os.str().c_str());
#endif // __GNUC__ < 3
#else
	   m->setResult("Exception %s thrown from %s\n%s", e.what(),
			e.getStack().top(), e.getError()); 
#endif
	   return TCL_ERROR;
	 }
	 catch(std::exception& e){
	   m->setResult("%s",e.what());
	   return TCL_ERROR;
	 }
	 catch(...){
	   m->setResult("Unknown error occured.");
	   return TCL_ERROR;
	 }
	 if(ret == COMMAND_SYNOPSIS) {
	    TclString res;
	    res.sprintf("usage: new %s %s",
			(char *)any->name, (char *) any->newSynopsis);
	    m->setResult(res);
	    return TCL_ERROR;
	 }
	 return ret;
      }
    }
  }

  TclString res("usage: new < ");
  for(any = AutoLink<TclObjectKind>::getHead(); any; any = any->following()) {
    res += any->getName();
    if(any->following()) res += " | ";
    else res += " > [%s]*";
  }
  m->setResult(res);
  return TCL_ERROR;
}

TclString TclObjectKind::getOptionSynopsis(void) const{
	TclString res;
  std::map< char *, CommandOption *, CStringCompare >::const_iterator it;
  res.sprintf("usage: %s%u < ", getName(), getRef());
  for(it = options.begin(); it != options.end();
      it++) {
    res += it->first;
    res += " | ";
  }
  res = TclString(res, 0, res.getLength() - 3);
  res += ">";
  return res;
}

int TclObjectKind::whichCmd(TclModule *m, int argc, char **argv) {
  TclObjectKind *any;

  if(argc > 2) return COMMAND_SYNOPSIS;

  if(argc == 1) {
    TclString res;
    for(any = AutoLink<TclObjectKind>::getHead(); any; any = any->following())
      {
	res += any->getName();
	if(any->following()) res += SPACE_STRING;
      }
    m->setResult(res);
    return TCL_OK;
  }
  
  else {
    for(any = AutoLink<TclObjectKind>::getHead(); any; any = any->following()) 
      {
	if(!strcmp(any->getName(), argv[1])) {
	  return (*any->objectsList)(m, argc - 1, argv + 1);
	}
      }
    m->setResult("unknown object kind : %s", argv[1]);
    return TCL_ERROR;
  }
}


TclObjectKind *TclObjectKind::findByKind(const std::type_info& type) {
  for(TclObjectKind *kind = AutoLink<TclObjectKind>::getHead(); kind;
      kind = kind->following()) {
    if(kind->objectsType == type) return kind;
  }
  return 0;
}


TclObjectKind *TclObjectKind::findByName(const char *name) {
  for(TclObjectKind *kind = AutoLink<TclObjectKind>::getHead(); kind;
      kind = kind->following()) {
    if(!strcmp(kind->name, name)) return kind;
  }
  return 0;
}


unsigned int TclObjectKind::getNextRef(void) const {
  if(objects.isEmpty()) return 1;
  else return objects.getLast()->getRef() + 1;
}

void TclObjectKind::operator+=(TclObject *obj) {
  obj->objectKind = this;
  obj->ref = getNextRef();
  objects.append(obj);
  if(module) obj->buildTCLCommand();
}


void TclObjectKind::operator-=(TclObject *obj)
  throw(ObjectsList<TclObject>::NotFound) {
  objects.remove(obj);
}


void TclObjectKind::created(TclObject *obj) {
  std::list<WarnedObject *>::iterator it;
  for(it = warn.begin(); it != warn.end(); it++)
    (*it)->createdObject(obj, objectsType);
}


void TclObjectKind::deleted(TclObject *obj) {
  std::list<WarnedObject *>::iterator it;
  for(it = warn.begin(); it != warn.end(); it++)
    (*it)->deletedObject(obj, objectsType);
}


void TclObjectKind::empty(void) const {
  std::list<WarnedObject *>::const_iterator it;
  for(it = warn.begin(); it != warn.end(); it++)
    (*it)->emptyObjectsList();
}


void TclObjectKind::addOptions(const std::type_info& src) {
  std::map<char *, CommandOption*, CStringCompare>::iterator it;
  TclObjectKind *skind = TclObjectKind::findByKind(src);

  if(skind) {
    skind->addChild(name);
    for(it = skind->getOptions()->begin(); it != skind->getOptions()->end();
	it++) {
      if (it->second->getOptionName()) {
	newOption(it->first,
		  it->second->getOptionName(),
		  it->second->getOptionSynopsis(),
		  it->second->getOptionDescription());
      } else {
	newOption(it->first,
		  it->second->getGlobalOptionName(),
		  it->second->getOptionSynopsis(),
		  it->second->getOptionDescription());
      }
      if(!globalCommand)
	globalCommand=skind->globalCommand;
    }
  }
}


TclObject *TclObjectKind::getObject(const std::type_info &type, unsigned int n) {
  TclObjectKind *kind;

  kind = TclObjectKind::findByKind(type);
  if(kind) {
    ObjectsList<TclObject>::iterator it = kind->objects.begin();
      
    for(; it!=kind->objects.end(); it++)
      if((*it)->getRef() == n) return (*it);
  }
  return 0;
}

TclObject *TclObjectKind::getObject(TclString name){
  for(TclObjectKind *kind = AutoLink<TclObjectKind>::getHead(); kind;
      kind = kind->following()) {
    ObjectsListSTLiterator<yav::TclObject> it = kind->firstObject();
    for(;it!=kind->lastObject();++it)
      if((*it)->name()==name) return (*it);
  }
  return NULL;
}
TclObject *TclObjectKind::getChildObject(char *objectName) {
  if(!strncmp(name, objectName, strlen(name))) {
    unsigned int n;

    if(sscanf(objectName + strlen(name), "%u", &n) != 1) {

      if(module) 
	module->setResult("No %s \'%s\'",(char *) name,(char *)objectName);
      return 0;
    }
    else {
      ObjectsList<TclObject>::iterator it = objects.begin();
      
      for(; it!=objects.end(); it++)
	if((*it)->getRef() == n) return (*it);
	  
      if(module)
	module->setResult("No %s \'%s\'",(char *)name,(char *)objectName);
      return 0;
    }
 
  } else {
    std::list<const char *>::iterator itk=childList.begin();
    TclObject *res=0;
    TclObjectKind *kind;
    for (;(itk!=childList.end())&&(res==0);++itk) {
      kind=TclObjectKind::findByName(*itk);
      if (kind) {
	res=kind->getChildObject(objectName);
      }
    }
    return res;
  }
}


TclObject *TclObjectKind::getObject(const std::type_info &type,
				    int argc, char **argv, int i) {
  TclObjectKind *kind;

  kind = TclObjectKind::findByKind(type);
  if(kind) {
    if(argc > (int) i) {
      TclObject *res=kind->getChildObject(argv[i]);
      if (res)
	return res;
      else {
	if(kind->module)
	  kind->module->setResult("Not a %s \'%s\'", kind->getName(), argv[i]);
	return 0;
      }
    }
    else {
      if(kind->module)
	kind->module->setResult("%s expected as %uth argument",
				kind->getName(), i);
      return 0;
    }
  }
  else return 0;
}


const char *TclObjectKind::getObjectName(const std::type_info& type) {
  TclObjectKind *kind = TclObjectKind::findByKind(type);
  if(kind) return kind->name;
  return 0;
}


} // namespace yav


