#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#include <modules/TclProbabilityDistribution.h>
#include <modules/TclGenericHistogram.h>

namespace yav {
	/// create Tcl command to create an object and to handle its option commands
	void TclProbabilityDistribution::createCommands(TclModule *m)
	{
		TclObjectKind *kind = new TclObjectKind(m, typeid(TclProbabilityDistribution), "proba","Probability Density Function",
			&TclProbabilityDistribution::create, &TclProbabilityDistribution::list,0);
		m->newCommand("proba%u", kind->getOptions(),
			"Probability management command");

		kind->newOption("-destroy", (TclObject::OPTION) &TclProbabilityDistribution::destroyCmd, 0, "destroy probability interface but without deleting the probabity density function pointing to it"); 
		kind->newOption("-deleteDistribution", (TclObject::OPTION) &TclProbabilityDistribution::deleteDistributionCmd, 
			0, "delete distribution and its Tcl interface"); 
		kind->newOption("-sameDistribution", (TclObject::OPTION) &TclProbabilityDistribution::sameDistributionCmd, 
			"proba%u", "returns true or false depending if both constraints points towards the same constraint"); 
		kind->newOption("-cast", (TclObject::OPTION) &TclProbabilityDistribution::castCmd, 
			0, "cast the current object into  a new TclProbabilityDistribution object associated with the same probability function. Delete the old object");
		kind->newOption("-range", (TclObject::OPTION) &TclProbabilityDistribution::rangeCmd, 
			0, "returns the minimum value the maximum value and the number of significant samples");
		kind->newOption("-statistics", (TclObject::OPTION) &TclProbabilityDistribution::statisticsCmd, 
			0, "returns the mean and standard deviation of the distribution");
		kind->newOption("-probabilityValue", (TclObject::OPTION) &TclProbabilityDistribution::probabilityValueCmd, 
			"%f", "returns the probability density function associated with that value ");
		kind->newOption("-histogramValue", (TclObject::OPTION) &TclProbabilityDistribution::histogramValueCmd, 
			"%f", "returns the histogram value associated with that value ");
		kind->newOption("-histogram", (TclObject::OPTION) &TclProbabilityDistribution::histogramCmd, 
			"%f", "returns a Tcl Histogram");

	}



int TclProbabilityDistribution::sameDistributionCmd(int argc, char **argv) {
if (argc!=1) 
  return COMMAND_SYNOPSIS;
 else {
   TclProbabilityDistribution *tc=(TclProbabilityDistribution *)  TclObjectKind::getObject(typeid(TclProbabilityDistribution), argc, argv, 0);
   if (tc==0) {
     setResult("Illegal distribution name : %s",argv[1]);
     return TCL_ERROR;
   }
   if (tc->getProbabilityFunction()==getProbabilityFunction()){
     setResult("true");
   } else {
     setResult("false");
   }
    return TCL_OK; 
 }
}
  /// -deleteConstaint option
int TclProbabilityDistribution::deleteDistributionCmd(int argc, char **argv) {
  TclModule *m;
  unsigned int ref;
  
  if(argc != 0) return COMMAND_SYNOPSIS;
  m = getModule();
  ref = TclObject::getRef();
  TclString res=name();
  delete probabilityDensityFunction;
  delete this;
  m->setResult(res);
  return TCL_OK;
}
int TclProbabilityDistribution::create(TclModule *m, int argc, char **argv) {
     if (argc!=1)
       return COMMAND_SYNOPSIS; 
     TclProbabilityDistribution *ri=new TclProbabilityDistribution (m,0);
     m->setResult("%s%u", (char *)ri->getObjectName(), ri->TclObject::getRef());
     return TCL_OK;
   }
 /// -destroy option
	int TclProbabilityDistribution::destroyCmd(int argc, char **argv) {
     TclModule *m;
    
     if(argc != 0) return COMMAND_SYNOPSIS;
     m = getModule();
	 TclString n=TclObject::name();
     delete this;
     m->setResult(n);
     return TCL_OK;
   }

int TclProbabilityDistribution::histogramCmd(int argc, char **argv)
   {
	   if (argc!=0)return COMMAND_SYNOPSIS;
	   ProbabilityDensityFunction *g=getProbabilityFunction();
	   unsigned long int *array= new unsigned long int [g->getNumberSamples()];
	   unsigned int i;
	   double x;
	   double minV=g->getMinValue();
	   double maxV=g->getMaxValue();
	   double a=(maxV-minV)/(g->getNumberSamples()-1);
	   for (i=0;i<g->getNumberSamples();++i) {
		   x=minV+a*i;
		   array[i]=(unsigned long int) g->getHistogramValue(x);
	   }
	   GenericHistogram *h= new GenericHistogram(array,g->getNumberSamples(),
		   minV,maxV);
	   delete [] array;
		TclGenericHistogram *tclh= new TclGenericHistogram(getModule(),h);
	   setResult(tclh->name());
	   return TCL_OK;
   }
   
int TclProbabilityDistribution::statisticsCmd(int argc, char **argv)
   {
	   if (argc!=0)return COMMAND_SYNOPSIS;
	   ProbabilityDensityFunction *g=getProbabilityFunction();
	
	double meanPDF=0;
	double sigmaPDF=0;
	double val;
	double step=(g->getMaxValue()-g->getMinValue())/(g->getNumberSamples()-1);
	   unsigned int i;
	   for (i=0;i< g->getNumberSamples();++i) {
		   val=g->getProbability(g->getMinValue()+i*step);
		   meanPDF+=val*val;
		   sigmaPDF+=val*val*val;
	   }
	   sigmaPDF-=meanPDF*meanPDF;
	   sigmaPDF=sqrt(sigmaPDF);

	   setResult("%f %f %f %f",g->getMeanValue(),g->getStandardDeviation(),meanPDF,sigmaPDF);
	   return TCL_OK;
   }
	/// -probabilty option
   int TclProbabilityDistribution::probabilityValueCmd(int argc, char **argv) {
     if(argc != 1) return COMMAND_SYNOPSIS;
	 double x=atof(argv[0]);
	setResult("%f",probabilityDensityFunction->getProbability(x));
     return TCL_OK;
   }
    /// -histogram option
   int TclProbabilityDistribution::histogramValueCmd(int argc, char **argv) {
     if(argc != 1) return COMMAND_SYNOPSIS;
	 double x=atof(argv[0]);
	setResult("%f",probabilityDensityFunction->getHistogramValue(x));
     return TCL_OK;
   }
   /// -histogram option
   int TclProbabilityDistribution::rangeCmd(int argc, char **argv) {
	   if(argc != 0) return COMMAND_SYNOPSIS;
	   setResult("%f %f %u",probabilityDensityFunction->getMinValue(),
		   probabilityDensityFunction->getMaxValue(),
		   probabilityDensityFunction->getNumberSamples());
	   return TCL_OK;
   }
   /// -cast option
   int TclProbabilityDistribution::castCmd(int argc, char **argv) {
     if(argc != 0) return COMMAND_SYNOPSIS;
     TclModule *m;
     m = getModule();
     TclProbabilityDistribution *ri=new TclProbabilityDistribution(m,probabilityDensityFunction);

     delete this;
     m->setResult(ri->name());
     return TCL_OK;
   }
   int TclGaussianDistribution::parametersCmd(int argc, char **argv)
   {
	   if (argc!=0)return COMMAND_SYNOPSIS;
	   GaussianDistribution *g=(GaussianDistribution *)getProbabilityFunction();
	   setResult("%f %f %f",g->getMeanValue(),g->getStandardDeviation(),
		   g->getMaxHistogram());
	   return TCL_OK;
   }

int TclGaussianDistribution::create(TclModule *m, int argc, char **argv) {
	if (argc!=4)
		return COMMAND_SYNOPSIS; 
	double mean=atof(argv[1]);
	double sigma=atof(argv[2]);
	double h=atof(argv[3]);
	GaussianDistribution *g=new GaussianDistribution(mean,sigma,h);
	TclGaussianDistribution *ri=new TclGaussianDistribution (m,g);
	m->setResult(ri->name());
	return TCL_OK;
}
	/// create Tcl command to create an object and to handle its option commands
void TclGaussianDistribution::createCommands(TclModule *m)
	{
		TclObjectKind *kind = new TclObjectKind(m, typeid(TclGaussianDistribution), "gauss","Gaussian Probability Density Function",
			&TclGaussianDistribution::create, &TclGaussianDistribution::list,"mean=%f stdDeviation=%f maxHistogram=%f");
		m->newCommand("gauss%u", kind->getOptions(),
			"Gaussian Probability management command");
		kind->addOptions(typeid(TclProbabilityDistribution));
		kind->newOption("-parameters", (TclObject::OPTION) &TclGaussianDistribution::parametersCmd, 
			0, "returns the meanValue, standardDeviation and maxHistogram values of that Gaussian distribution"); 
	}

int TclGaussianMixtureDistribution::parametersCmd(int argc, char **argv)
   {
	   if (argc!=0)return COMMAND_SYNOPSIS;
	   GaussianMixtureDistribution *g=(GaussianMixtureDistribution *)getProbabilityFunction();
	   TclString result,r;
	   unsigned int i;
	   for (i=0;i<g->getNumberOfDistributions();++i) {
		   r.sprintf("{%f %f %f } ",g->getMeanValue(i),g->getStandardDeviation(i),
			   g->getMaxHistogram(i));
		   result+=r;
	   }
	   setResult(result);
	   return TCL_OK;
   }

int TclGaussianMixtureDistribution::create(TclModule *m, int argc, char **argv) {
	if ((argc-1)%3!=0)
		return COMMAND_SYNOPSIS;
	unsigned int nb=(argc-1)/3;
	std::vector<double> mean;
	std::vector<double> sigma;
	std::vector<double> max;

	mean.reserve(nb+1);
	sigma.reserve(nb+1);
	max.reserve(nb+1);

	unsigned int i;
	for (i=0;i<nb;++i) 
	{
		mean[i]=atof(argv[3*i+1]);
		sigma[i]=atof(argv[3*i+2]);
		max[i]=atof(argv[3*i+3]);
	}
GaussianMixtureDistribution *g=new GaussianMixtureDistribution(nb,mean,sigma,max);
	TclGaussianMixtureDistribution *ri=new TclGaussianMixtureDistribution (m,g);
	m->setResult(ri->name());
	return TCL_OK;
}
	/// create Tcl command to create an object and to handle its option commands
void TclGaussianMixtureDistribution::createCommands(TclModule *m)
	{
		TclObjectKind *kind = new TclObjectKind(m, typeid(TclGaussianMixtureDistribution), "gaussMix","Gaussian Mixture Probability Density Function",
			&TclGaussianMixtureDistribution::create, &TclGaussianMixtureDistribution::list,"list of mean=%f stdDeviation=%f maxHistogram=%f");
		m->newCommand("gaussMix%u", kind->getOptions(),
			"Gaussian Mixture Probability management command");
		kind->addOptions(typeid(TclProbabilityDistribution));
		kind->newOption("-parameters", (TclObject::OPTION) &TclGaussianMixtureDistribution::parametersCmd, 
			0, "returns the set of meanValue, standardDeviation and maxHistogram values of that Gaussian mixture distribution"); 
	}

int TclHistogramDistribution::parametersCmd(int argc, char **argv)
   {
	   if (argc!=0)return COMMAND_SYNOPSIS;
	   HistogramDistribution *g=(HistogramDistribution *)getProbabilityFunction();
	   setResult("%f",g->getScaleDistribution());
	   return TCL_OK;
   }

int TclHistogramDistribution::create(TclModule *m, int argc, char **argv) {
	if (argc!=4)
		return COMMAND_SYNOPSIS;
	TclGenericHistogram *h= (TclGenericHistogram *)	TclObjectKind::getObject(typeid(TclGenericHistogram),
		argc, argv, 1);
	if (!h) {
		m->setResult("%s is not an histogram",argv[1]);
		return TCL_ERROR;
	}
	double window=atof(argv[2]);
	double quantile=atof(argv[3]);
	HistogramDistribution *g=new HistogramDistribution(h->getClient(),window,quantile);
	TclHistogramDistribution *ri=new TclHistogramDistribution (m,g);
	m->setResult(ri->name());
	return TCL_OK;
}
	/// create Tcl command to create an object and to handle its option commands
void TclHistogramDistribution::createCommands(TclModule *m)
	{
		TclObjectKind *kind = new TclObjectKind(m, typeid(TclHistogramDistribution), "pdf","Histogram Probability Density Function",
			&TclHistogramDistribution::create, &TclHistogramDistribution::list,"histo%u parzenWindow=%f quantile=%f");
		m->newCommand("pdf%u", kind->getOptions(),
			"Histogram Probability management command");
		kind->addOptions(typeid(TclProbabilityDistribution));
		kind->newOption("-parameters", (TclObject::OPTION) &TclHistogramDistribution::parametersCmd, 
			0, "returns the scale value used to insure that the distributions sums to one"); 
	}
}
