#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <modules/WindowDump.h>

#ifndef WIN32
#ifdef X_NOT_STDC_ENV
extern int errno;
#endif

#define XWD_FILE_VERSION 7
#ifdef WORD64
#define sz_XWDheader 104
#else
#define sz_XWDheader 100
#endif

#ifdef PRE_R6_ICCCM
#define sz_XWDColor sizeof(XColor)
#else
#define sz_XWDColor 12
#endif

typedef CARD32 xwdval;          /* for old broken programs */

/* Values in the file are most significant byte first. */

typedef struct _xwd_file_header {
        /* header_size = SIZEOF(XWDheader) + length of null-terminated
         * window name. */
        CARD32 header_size B32;

        CARD32 file_version B32;        /* = XWD_FILE_VERSION above */
        CARD32 pixmap_format B32;       /* ZPixmap or XYPixmap */
        CARD32 pixmap_depth B32;        /* Pixmap depth */
        CARD32 pixmap_width B32;        /* Pixmap width */
        CARD32 pixmap_height B32;       /* Pixmap height */
        CARD32 xoffset B32;             /* Bitmap x offset, normally 0 */
        CARD32 byte_order B32;          /* of image data: MSBFirst, LSBFirst */

        /* bitmap_unit applies to bitmaps (depth 1 format XY) only.
         * It is the number of bits that each scanline is padded to. */
        CARD32 bitmap_unit B32;

        CARD32 bitmap_bit_order B32;    /* bitmaps only: MSBFirst, LSBFirst */

        /* bitmap_pad applies to pixmaps (non-bitmaps) only.
         * It is the number of bits that each scanline is padded to. */
        CARD32 bitmap_pad B32;

        CARD32 bits_per_pixel B32;      /* Bits per pixel */

        /* bytes_per_line is pixmap_width padded to bitmap_unit (bitmaps)
         * or bitmap_pad (pixmaps).  It is the delta (in bytes) to get
         * to the same x position on an adjacent row. */
        CARD32 bytes_per_line B32;
        CARD32 visual_class B32;        /* Class of colormap */
        CARD32 red_mask B32;            /* Z red mask */
        CARD32 green_mask B32;          /* Z green mask */
        CARD32 blue_mask B32;           /* Z blue mask */
        CARD32 bits_per_rgb B32;        /* Log2 of distinct color values */
        CARD32 colormap_entries B32;    /* Number of entries in colormap; not used? */
        CARD32 ncolors B32;             /* Number of XWDColor structures */
        CARD32 window_width B32;        /* Window width */
        CARD32 window_height B32;       /* Window height */
        CARD32 window_x B32;            /* Window upper left X coordinate */
        CARD32 window_y B32;            /* Window upper left Y coordinate */
        CARD32 window_bdrwidth B32;     /* Window border width */
#ifdef WORD64
        CARD32 header_end B32;          /* Pad to fill out word */
#endif
} XWDFileHeader;

/* Null-terminated window name follows the above structure. */

/* Next comes XWDColor structures, at offset XWDFileHeader.header_size in
 * the file.  XWDFileHeader.ncolors tells how many XWDColor structures
 * there are.
 */

#ifdef PRE_R6_ICCCM
typedef XColor XWDColor;
#else
typedef struct {
        CARD32  pixel B32;
        CARD16  red B16;
        CARD16  green B16;
        CARD16  blue B16;
        CARD8   flags;
        CARD8   pad;
} XWDColor;
#endif


#define FEEP_VOLUME 0

#define STATIC_GRAY	0x01
#define GRAY_SCALE	0x02
#define PSEUDO_COLOR	0x04
#define TRUE_COLOR	0x10
#define DIRECT_COLOR	0x11
#define	RED_SHIFT         0
#define GREEN_SHIFT       8
#define BLUE_SHIFT       16

#define lowbit(x) ((x) & (~(x) + 1))
#define Extent(string)  ((int) strlen(string))
#define DirectClass  1
#define PseudoClass  2

static int    screen;
static int    ImageClass;
static int    format = ZPixmap;
static Bool   nobdrs = False;
static Bool   debug = False;
static Bool   silent = True;
static Bool   use_installed = False;
static long   add_pixel_value = 0;


void initFakeVisual(Visual *Vis)
{
  Vis->ext_data=NULL;
  Vis->class = DirectColor ;
  Vis->red_mask   = 0x00FF0000;
  Vis->green_mask = 0x0000FF00 ;
  Vis->blue_mask  = 0x000000FF ;
  Vis->map_entries = 256 ;
  Vis->bits_per_rgb = 8 ;
}


/*
 * Determine the pixmap size.
 */

int Pixmap_Size(XImage *image)
{
  if (image->format != ZPixmap)
    return(image->bytes_per_line * image->height * image->depth);
  else
    return(image->bytes_per_line * image->height);
}


static int ReadColors(Display *display, Visual *vis, Colormap cmap,
		      XColor **colors)
{
  int i, ncolors ;
  
  ncolors = vis->map_entries;
  
  if (!(*colors = (XColor *) malloc (sizeof(XColor) * ncolors)))
    {
      fprintf(stderr, "Out of memory!");
      exit(-1);
    }
  
  if (vis->class == DirectColor || vis->class == TrueColor) 
    {
      unsigned short red, green, blue, red1, green1, blue1;
      
      if(debug) printf("'DirectColor' or 'TrueColor' visual detected.\n");
      ImageClass = DirectClass;
      red = green = blue = 0;
      red1 = lowbit(vis->red_mask);
      green1 = lowbit(vis->green_mask);
      blue1 = lowbit(vis->blue_mask);

      for (i=0; i<ncolors; i++) 
	{
	  (*colors)[i].pixel = red|green|blue;
	  (*colors)[i].pad = 0;
	  red += red1;
	  if (red > vis->red_mask)
	    red = 0;
	  green += green1;
	  if (green > vis->green_mask)
	    green = 0;
	  blue += blue1;
	  if (blue > vis->blue_mask)
	    blue = 0;
	}
    } 
  else 
    {
      ImageClass = PseudoClass;
      for (i=0; i<ncolors; i++) 
	{
	  (*colors)[i].pixel = i;
	  (*colors)[i].pad = 0;
	}
    }
  
  XQueryColors(display, cmap, *colors, ncolors);
  
  return(ncolors);
}

/*
 * Get the XColors of all pixels in image - returns # of colors
 */
int Get_XColors(Display *display, XWindowAttributes *win_info, XColor **colors)
{
  int       i, ncolors;
  Colormap  cmap = win_info->colormap;

  if (use_installed)
    /* assume the visual will be OK ... */
    cmap = XListInstalledColormaps(display, win_info->root, &i)[0];
  if (!cmap)
    return(0);
  ncolors = ReadColors(display, win_info->visual, cmap, colors) ;
  
  return ncolors ;
}


/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%  M S B F i r s t O r d e r L o n g                                          %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function MSBFirstOrderLong converts a least-significant byte first buffer
%  of integers to most-significant byte first.
%
%  The format of the MSBFirstOrderLong routine is:
%
%       MSBFirstOrderLong(p,length);
%
%  A description of each parameter follows.
%
%   o  p:  Specifies a vertexer to a buffer of integers.
%
%   o  length:  Specifies the length of the buffer.
%
%
*/
void MSBFirstOrderLong(register char *p,const unsigned int length)
{
  register char
    c,
    *q,
    *sp;

  q=p+length;
  while (p < q)
  {
    sp=p+3;
    c=(*sp);
    *sp=(*p);
    *p++=c;
    sp=p+1;
    c=(*sp);
    *sp=(*p);
    *p++=c;
    p+=2;
  }
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%  M S B F i r s t O r d e r S h o r t                                        %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function MSBFirstOrderShort converts a least-significant byte first buffer
%  of integers to most-significant byte first.
%
%  The format of the MSBFirstOrderShort routine is:
%
%       MSBFirstOrderLongShort(p,length);
%
%  A description of each parameter follows.
%
%   o  p:  Specifies a vertexer to a buffer of integers.
%
%   o  length:  Specifies the length of the buffer.
%
%
*/
void MSBFirstOrderShort(register char *p,const unsigned int length)
{
  register char
    c,
    *q;

  q=p+length;
  while (p < q)
  {
    c=(*p);
    *p=(*(p+1));
    p++;
    *p++=c;
  }
}



/*  Window_Dump: dump a window to a file which must already 
    be open for writting.  */

void WindowDump(Display *display, Window window, char *filename)
{
  unsigned long       swaptest = 1;
  XColor             *colors;
  unsigned            buffer_size;
  int                 win_name_size;
  int                 header_size;
  int                 ncolors, i;
  char               *win_name;
  Bool                got_win_name;
  XWindowAttributes   win_info;
  XImage             *image;
  int                 absx, absy, x, y;
  unsigned            width, height;
  int                 dwidth, dheight;
  int                 bw;
  Window              dummywin;
  XWDFileHeader       header;
  XWDColor            xwdcolor;   
  Visual	      *vis ;

  FILE *out = fopen(filename, "w");
  if(!out) return;

  /* Inform the user not to alter the screen. */
  if (!silent) 
    {
#ifdef XKB
      XkbStdBell(display,None,50,XkbBI_Wait);
#else
      XBell(display,FEEP_VOLUME);
#endif
      XFlush(display);
    }

  /* Get the parameters of the window being dumped.   */
  if (debug) 
    printf("windump: Getting target window information.\n");

  if(!XGetWindowAttributes(display, window, &win_info)) 
    {
      fprintf(stderr,"Can't get target window attributes.");
      exit(-1);
    }
  else
    {
      if(debug) {
	printf("Depth=%d", win_info.depth);
	if (win_info.map_installed)
	  printf(", colormap installed.\n");
      }
    }
  
  /* handle any frame window */
  if (!XTranslateCoordinates (display, window, RootWindow (display, screen),
			      0, 0, &absx, &absy, &dummywin)) 
    {
      fprintf (stderr, "Unable to translate window coordinates (%d,%d)\n", absx, absy);
      exit (1);
    }
  win_info.x = absx;
  win_info.y = absy;
  width = win_info.width;
  height = win_info.height;
  bw = 0;
  
  if (!nobdrs) 
    {
      absx -= win_info.border_width;
      absy -= win_info.border_width;
      bw = win_info.border_width;
      width += (2 * bw);
      height += (2 * bw);
    }
  dwidth = DisplayWidth (display, screen);
  dheight = DisplayHeight (display, screen);
  
  /* clip to window */
  if (absx < 0) width += absx, absx = 0;
  if (absy < 0) height += absy, absy = 0;
  if (absx + width > dwidth) width = dwidth - absx;
  if (absy + height > dheight) height = dheight - absy;
  
  XFetchName(display, window, &win_name);
  if (!win_name || !win_name[0]) 
    {
      win_name = "win_dump";
      got_win_name = False;
    } 
  else 
    {
      got_win_name = True;
    }
  if(debug) printf("win_name is: %s\n", win_name);
  
  /* sizeof(char) is included for the null string terminator. */
  win_name_size = strlen(win_name) + sizeof(char);
  
  /*
   * Snarf the pixmap with XGetImage.
   */
  
  x = absx - win_info.x;
  y = absy - win_info.y;
  
  image = XGetImage (display, window, x, y, width, height, AllPlanes, format);

  if (!image) 
    {
      fprintf (stderr, "Unable to get image at %dx%d+%d+%d\n",
	       width, height, x, y);
      exit (1);
    }
  
  if (add_pixel_value != 0) XAddPixel(image, add_pixel_value);
  
  /* Determine the pixmap size.  */
  buffer_size = Pixmap_Size(image);
  
  if (debug) printf("windump: Getting Colors.\n");
  
  ncolors = Get_XColors(display, &win_info, &colors);
  vis = win_info.visual ;
  if(debug) printf("%d colors found.\n", ncolors);
  
  /* ncolors = Get24bitDirectColors(&colors) ;
     initFakeVisual(&vis_h) ;
     vis = &vis_h ;
     printf("%d colors found.\n", ncolors); */

  /*
   * Inform the user that the image has been retrieved.
   */
  if (!silent) 
    {
#ifdef XKB
      XkbStdBell(display,window,FEEP_VOLUME,XkbBI_Proceed);
      XkbStdBell(display,window,FEEP_VOLUME,XkbBI_RepeatingLastBell);
#else
      XBell(display, FEEP_VOLUME);
      XBell(display, FEEP_VOLUME);
#endif
      XFlush(display);
    }
  
  /* Calculate header size */
  if (debug) printf("windump: Calculating header size.\n");
  header_size = sz_XWDheader+Extent(win_name)+1;
  
  /* Write out header information */
  if (debug) 
    printf("windump: Constructing and dumping file header.\n");
  
  header.header_size= header_size;
  header.file_version=(CARD32) XWD_FILE_VERSION;
  header.pixmap_format=(CARD32) format;
  header.pixmap_depth=(CARD32) (ImageClass == DirectClass ? 24 : 8);
  header.pixmap_width=(CARD32) image->width;
  header.pixmap_height=(CARD32) image->height;

  header.xoffset = (CARD32) image->xoffset;
  header.byte_order = (CARD32) image->byte_order;

  header.bitmap_unit=(CARD32) (ImageClass == DirectClass ? 32 : 8);

  header.bitmap_bit_order=(CARD32) image->bitmap_bit_order;
  header.bitmap_unit = (CARD32) image->bitmap_unit;
  header.bitmap_bit_order = (CARD32) image->bitmap_bit_order;
  header.bitmap_pad = (CARD32) image->bitmap_pad;
  header.bits_per_pixel = (CARD32) image->bits_per_pixel;
  header.bytes_per_line = (CARD32) image->bytes_per_line; 

  header.visual_class=(CARD32) (ImageClass == DirectClass ? DirectColor : PseudoColor);
  header.red_mask=(CARD32) (ImageClass == DirectClass ? 0xff0000 : 0);
  header.green_mask=(CARD32) (ImageClass == DirectClass ? 0xff00 : 0);
  header.blue_mask=(CARD32) (ImageClass == DirectClass ? 0xff : 0);
  header.bits_per_rgb=(CARD32) (ImageClass == DirectClass ? 24 : 8);
  header.colormap_entries=(CARD32) (ImageClass == DirectClass ? 256 : ncolors);
  header.ncolors=(ImageClass == DirectClass ? 0 : ncolors);
  header.window_width=(CARD32) win_info.width;
  header.window_height=(CARD32) win_info.height;
  header.window_x = absx; 
  header.window_y = absy; 
  header.window_bdrwidth=(CARD32) win_info.border_width;

  if (debug)
    {
      printf("Header Informations\n\n");
      printf("     header_size = %d\n", header_size);
      printf("     file_version = %li\n",header.file_version);
      printf("     pixmap_format = %li\n",header.pixmap_format);
      printf("     pixmap_depth = %li\n", header.pixmap_depth);
      printf("     pixmap_width = %li\n", header.pixmap_width);
      printf("     pixmap_height = %li\n", header.pixmap_height);
      printf("     xoffset = %li\n", header.xoffset);
      if (header.byte_order == MSBFirst)
	printf("     byte_order = MSBFirst\n");
      if (header.byte_order == LSBFirst)
	printf("     byte_order = LSBFirst\n");
      printf("     bitmap_unit = %li\n", header.bitmap_unit);
      if (header.bitmap_bit_order == MSBFirst)
	printf("     bitmap_bit_order = MSBFirst\n");
      if (header.bitmap_bit_order == LSBFirst)
	printf("     bitmap_bit_order = LSBFirst\n");
      printf("     bitmap_pad = %li\n", header.bitmap_pad);
      printf("     bits_per_pixel = %li\n", header.bits_per_pixel);
      printf("     bytes_per_line = %li\n", header.bytes_per_line);
      printf("     visual_class = %li\n", header.visual_class);
      printf("     bits_per_rgb = %li\n", header.bits_per_rgb);
      printf("     colormap_entries = %li\n", header.colormap_entries);
      printf("     ncolors = %li\n", header.ncolors);
      printf("     window_width = %li\n", header.window_width);
      printf("     window_height = %li\n", header.window_height);
      printf("     window_x = %li\n", header.window_x);
      printf("     window_y = %li\n", header.window_y);
      printf("     window_bdrwidth = %li\n", header.window_bdrwidth);
    }

  if (*(char *) &swaptest) 
    {
      if(debug) printf("Need to swap some bytes...\n");
      MSBFirstOrderLong((char *) &header,sizeof(header));
      for (i = 0; i < ncolors; i++) 
	{
	  MSBFirstOrderLong((char *) &colors[i].pixel,sizeof(long));
	  MSBFirstOrderShort((char *) &colors[i].red,3*sizeof(short));
	}
    }

  (void) fwrite((char *) &header, sz_XWDheader, 1, out);
  (void) fwrite((char *) win_name, 1, Extent(win_name)+1, out);
  
  /* Write out the color maps, if any */
  
  if (debug) 
    printf("windump: Dumping %d colors.\n", ncolors);

  if (ImageClass == PseudoClass)
    {
      for (i = 0; i < ncolors; i++) 
	{
	  xwdcolor.pixel = colors[i].pixel;
	  xwdcolor.red   = colors[i].red;
	  xwdcolor.green = colors[i].green;
	  xwdcolor.blue  = colors[i].blue; 
	  xwdcolor.flags = colors[i].flags;
	  xwdcolor.pad   = 0;
	  
	  if (fwrite((char *) &xwdcolor, sz_XWDColor, 1, out) != 1) 
	    {
	      perror("xwd");
	      exit(1);
	    }
	} 
    }
  
  /* Write out the buffer */
  if (debug) 
    printf("windump: Dumping pixmap.  bufsize=%d\n",buffer_size);
  
  /*  This copying of the bit stream (data) to a file is to be replaced
   *  by an Xlib call which hasn't been written yet.  It is not clear
   *  what other functions of xwd will be taken over by this (as yet)
   *  non-existant X function.
   */

  if (fwrite(image->data, (int) buffer_size, 1, out) != 1) 
    {
      perror("xwd");
      exit(1);
    }
  
  /* free the color buffer */

  if(debug && ncolors > 0) 
    printf("windump: Freeing colors.\n");
  if(ncolors > 0) 
    free(colors);
  
  /* Free window name string */
  if (debug) 
    printf("windump: Freeing window name string.\n");
  if (got_win_name) 
    XFree(win_name);
  
  /*Free image */
  XDestroyImage(image);

  fclose(out);
}
#else

#define GAMMA 1.0

/*
 *----------------------------------------------------------------------
 *
 * Blt_ColorImage --
 *
 *      The structure below represents a color image.  Each pixel
 *	occupies a 32-bit word of memory: one byte for each of the
 *	red, green, and blue color intensities, and another for
 *	alpha-channel image compositing (e.g. transparency).
 *
 *----------------------------------------------------------------------
 */
/*
 * The TkWinDrawable is the internal implementation of an X Drawable (either
 * a Window or a Pixmap).  The following constants define the valid Drawable
 * types.
 */

#define TWD_BITMAP	1
#define TWD_WINDOW	2
#define TWD_WINDC	3

typedef struct TkWindowStruct TkWindow;

typedef struct {
    int type;
    HWND handle;
    TkWindow *winPtr;
} TkWinWindow;

typedef struct {
    int type;
    HBITMAP handle;
    Colormap colormap;
    int depth;
} TkWinBitmap;

typedef struct {
    int type;
    HDC hdc;
} TkWinDC;

typedef union {
    int type;
    TkWinWindow window;
    TkWinBitmap bitmap;
    TkWinDC winDC;
} TkWinDrawable;

/*
 * The TkWinDCState is used to save the state of a device context
 * so that it can be restored later.
 */

typedef struct {
    HPALETTE palette;
    int bkmode;			/* This field was added in Tk
				 * 8.3.1. Be careful that you don't 
				 * use this structure in a context
				 * where its size is important.  */
} TkWinDCState;

typedef union {
    unsigned int value;		/* Lookup table address */
    struct RGBA {
	unsigned char red;	/* Red intensity 0..255 */
	unsigned char green;	/* Green intensity 0.255 */
	unsigned char blue;	/* Blue intensity 0..255 */
	unsigned char alpha;	/* Alpha-channel for compositing. 0..255 */
    } rgba;
    unsigned char channel[4];
} Pix32;

#define Red	rgba.red
#define Blue	rgba.blue
#define Green	rgba.green
#define Alpha	rgba.alpha

typedef struct ColorImage {
    int width, height;		/* Dimensions of the image */
    Pix32 *bits;		/* Array of pixels representing the image. */
} *Blt_ColorImage;

#define CLAMP(c)	((((c) < 0.0) ? 0.0 : ((c) > 255.0) ? 255.0 : (c)))
#define Blt_ColorImageHeight(c)	((c)->height)
#define Blt_ColorImageWidth(c)	((c)->width)

extern HDC TkWinGetDrawableDC(Display *display, Drawable drawable,
    TkWinDCState * state);

extern HDC TkWinReleaseDrawableDC(Drawable drawable, HDC dc,
    TkWinDCState * state);

#ifndef SelectBitmap
#define SelectBitmap(hdc, hbm)  ((HBITMAP)SelectObject((hdc), (HGDIOBJ)(HBITMAP)(hbm)))
#endif

#ifndef DeleteBitmap
#define DeleteBitmap(hbm)       DeleteObject((HGDIOBJ)(HBITMAP)(hbm))
#endif

#ifndef DeletePalette
#define DeletePalette(hpal)     DeleteObject((HGDIOBJ)(HPALETTE)(hpal))
#endif

#define Blt_ColorImageBits(c)	((c)->bits)

Blt_ColorImage
Blt_CreateColorImage(width, height)
    int width, height;		/* Dimensions of new image */
{
    struct ColorImage *imagePtr;
    size_t size;

    size = width * height;
    imagePtr = malloc(sizeof(struct ColorImage));

    imagePtr->bits = malloc(sizeof(Pix32) * size);


    imagePtr->width = width;
    imagePtr->height = height;
    return imagePtr;
}


HPALETTE
Blt_GetSystemPalette(void)
{
    HDC hDC;
    HPALETTE hPalette;
    DWORD flags;

    hPalette = NULL;
    hDC = GetDC(NULL);		/* Get the desktop device context */
    flags = GetDeviceCaps(hDC, RASTERCAPS);
    if (flags & RC_PALETTE) {
	LOGPALETTE *palettePtr;

	palettePtr = (LOGPALETTE *)
	    GlobalAlloc(GPTR, sizeof(LOGPALETTE) + 256 * sizeof(PALETTEENTRY));
	palettePtr->palVersion = 0x300;
	palettePtr->palNumEntries = 256;
	GetSystemPaletteEntries(hDC, 0, 256, palettePtr->palPalEntry);
	hPalette = CreatePalette(palettePtr);
	GlobalFree(palettePtr);
    }
    ReleaseDC(NULL, hDC);
    return hPalette;
}

Blt_ColorImage
Blt_DrawableToColorImage(
    Tk_Window tkwin,
    Drawable drawable,
    int x, int y,
    int width, int height,	/* Dimension of the drawable. */
    double inputGamma)
{
    void *data;
    BITMAPINFO info;
    DIBSECTION ds;
    HBITMAP hBitmap, oldBitmap;
    HPALETTE hPalette;
    HDC memDC;
    unsigned char *srcArr;
    register unsigned char *srcPtr;
    HDC hDC;
    TkWinDCState state;
    register Pix32 *destPtr;
    Blt_ColorImage image;
    unsigned char lut[256];

    hDC = TkWinGetDrawableDC(Tk_Display(tkwin), drawable, &state);

    /* Create the intermediate drawing surface at window resolution. */
    ZeroMemory(&info, sizeof(info));
    info.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    info.bmiHeader.biWidth = width;
    info.bmiHeader.biHeight = height;
    info.bmiHeader.biPlanes = 1;
    info.bmiHeader.biBitCount = 32;
    info.bmiHeader.biCompression = BI_RGB;
    hBitmap = CreateDIBSection(hDC, &info, DIB_RGB_COLORS, &data, NULL, 0);
    memDC = CreateCompatibleDC(hDC);
    oldBitmap = SelectBitmap(memDC, hBitmap);

    hPalette = Blt_GetSystemPalette();
    if (hPalette != NULL) {
	SelectPalette(hDC, hPalette, FALSE);
	RealizePalette(hDC);
	SelectPalette(memDC, hPalette, FALSE);
	RealizePalette(memDC);
    }
    image = NULL;
    /* Copy the window contents to the memory surface. */
    if (!BitBlt(memDC, 0, 0, width, height, hDC, x, y, SRCCOPY)) {
#ifdef notdef
	PurifyPrintf("can't blit: %s\n", Blt_LastError());
#endif
	goto done;
    }
    if (GetObject(hBitmap, sizeof(DIBSECTION), &ds) == 0) {
#ifdef notdef
	PurifyPrintf("can't get object: %s\n", Blt_LastError());
#endif
	goto done;
    }
    srcArr = (unsigned char *)ds.dsBm.bmBits;
    image = Blt_CreateColorImage(width, height);
    destPtr = Blt_ColorImageBits(image);

    {
	register int i;
	double value;

	for (i = 0; i < 256; i++) {
	    value = pow(i / 255.0, inputGamma) * 255.0 + 0.5;
	    lut[i] = (unsigned char)CLAMP(value);
	}
    }

    /* 
     * Copy the DIB RGB data into the color image. The DIB scanlines
     * are stored bottom-to-top and the order of the RGB color
     * components is BGR. Who says Win32 GDI programming isn't
     * backwards?  
     */

    for (y = height - 1; y >= 0; y--) {
	srcPtr = srcArr + (y * ds.dsBm.bmWidthBytes);
	for (x = 0; x < width; x++) {
	    destPtr->Blue = lut[*srcPtr++];
	    destPtr->Green = lut[*srcPtr++];
	    destPtr->Red = lut[*srcPtr++];
	    destPtr->Alpha = (unsigned char)-1;
	    destPtr++;
	    srcPtr++;
	}
    }
  done:
    DeleteBitmap(SelectBitmap(memDC, oldBitmap));
    DeleteDC(memDC);
    TkWinReleaseDrawableDC(drawable, hDC, &state);
    if (hPalette != NULL) {
	DeletePalette(hPalette);
    }
    return image;
}

void WindowDump( Tk_Window tkwin,
    Drawable drawable, char *filename)
{
	int i,j;
	Pix32 pixel;
	/// get the image
	Blt_ColorImage image=
		Blt_DrawableToColorImage(
		tkwin,drawable,0,0,Tk_Width(tkwin),Tk_Height(tkwin), GAMMA);
	///dump as ppm file

	FILE *fptr = fopen(filename,"wb");
	int width,height;

	width=Blt_ColorImageWidth(image);
	height=Blt_ColorImageHeight(image);
	fprintf(fptr,"P6\n%d %d\n255\n",width,height);

	for (j=0; j<height;j++) {
		for (i=0;i<(int)width;i++) {
			pixel=image->bits[j*width+i];
			fputc(pixel.Red,fptr);
			fputc(pixel.Green,fptr);
			fputc(pixel.Blue,fptr);
		}
	}
	fclose(fptr);
}
#endif

