#ifndef BASICTOGL_H
#define BASICTOGL_H

#ifndef WIN32
#define X11
#endif

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#pragma interface "graphics3D/BasicTogl.h"
#include <assert.h>

extern "C" {

/*** Windows headers ***/
#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#undef WIN32_LEAN_AND_MEAN
#include <winnt.h>
#endif /*WIN32*/

/*** X Window System headers ***/
#ifdef X11
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>  /* for XA_RGB_DEFAULT_MAP atom */
#if defined(__vms)
#include <X11/StdCmap.h>  /* for XmuLookupStandardColormap */
#else
#include <X11/Xmu/StdCmap.h>  /* for XmuLookupStandardColormap */
#endif
#include <GL/glx.h>
#endif /*X11*/


#include <tcl.h>
#include <tk.h>
#include <GL/gl.h>
#ifdef TOGL_X11
#include <X11/Xlib.h>
#endif

}


#include <basics/Exception.h>
#include <modules/TclModule.h>
#include <modules/TclObject.h>


namespace yav {


/** TOgl Widget Version */
#define TOGL_VERSION "2.0"
#define TOGL_MAJOR_VERSION 2
#define TOGL_MINOR_VERSION 0



/*
 * "Standard" fonts which can be specified to Togl_LoadBitmapFont()
 */
#define TOGL_BITMAP_8_BY_13		((char *) 1)
#define TOGL_BITMAP_9_BY_15		((char *) 2)
#define TOGL_BITMAP_TIMES_ROMAN_10	((char *) 3)
#define TOGL_BITMAP_TIMES_ROMAN_24	((char *) 4)
#define TOGL_BITMAP_HELVETICA_10	((char *) 5)
#define TOGL_BITMAP_HELVETICA_12	((char *) 6)
#define TOGL_BITMAP_HELVETICA_18	((char *) 7)
 

/*
 * Normal and overlay plane constants
 */
#define TOGL_NORMAL	1
#define TOGL_OVERLAY	2



class BasicTogl;



/************************************************************************/
/************************************************************************/
/**                            CLASS Basic Togl                              **/
/************************************************************************/
/************************************************************************/

/** \class BasicTogl
    \ingroup Graphics3D
    \brief a Basic TCL OpenGL widget  */
class BasicTogl  {
    // instance members
protected:
  /** @name Togl Widget variables */
  //@{
   /** Tcl interpreter */
   Tcl_Interp *interp;	
#ifndef NO_TK_CURSOR
   /** The widget's cursor */
   Tk_Cursor Cursor;
#endif
   /** Time value for timer */
   int Time;

   /* configuration flags (ala GLX parameters) */
   /// Rgba flag
   int RgbaFlag;
   /// Rgba flag
   int RgbaRed;
   /// Rgba green flag
   int RgbaGreen;
   /// Rgba blue flag
   int RgbaBlue;
   /// depth flag
   int DepthFlag;
   /// Z-buffer depth value
   int DepthSize;
   /// Accumulation buffer flag
   int AccumFlag;
   /// Accumulation buffer red-component size
   int AccumRed;
   /// Accumulation buffer green-component size
   int AccumGreen;
   /// Accumulation buffer blue-component size
   int AccumBlue;
   /// Accumulation buffer alpha-component size
   int AccumAlpha;
   /// Alpha flag
   int AlphaFlag;
   /// Alpha bit size
   int AlphaSize;
   /// Stencil buffer flag
   int StencilFlag;
   /// Stencil buffer depth size
   int StencilSize;
   /// Private Colormap flag
   int PrivateCmapFlag;
   /// Overlay  flag
   int OverlayFlag;
   /// Auxiliairy number
   int AuxNumber;
  /// the widget height
   int requiredHeight;
   /// the widget width
   int requiredWidth;
/// double buffering flag
	bool requiredDoubleFlag;

   /** User's identification string */
   char *Ident;
   /** Pointer to user data */
   ClientData Client_Data;
   /// should the camera be redrawn at next idle moment
   GLboolean updatePending;	
   /// should the camera be resized at next idle moment
   GLboolean resizePending;	
   /// should the camera be redrawn at next idle moment
   GLboolean exposePending;

   /* Overlay stuff */
#if defined(X11)
   /** Overlay planes OpenGL context */
   GLXContext OverlayCtx;	
#elif defined(WIN32)
   /** Overlay planes OpenGL context */
   HGLRC tglGLOverlayHglrc;
   /** (Maximum) size of colormap in color index mode */
	int CiColormapSize;
#endif /* X11 */
   /** The overlay window, or 0 */
   Window OverlayWindow;
   /** Should overlay be redrawn? */
   GLboolean OverlayUpdatePending;
   /** colormap for overlay is created */
   Colormap OverlayCmap;
   /** transparent pixel */
   int OverlayTransparentPixel;
   /** whether the overlay is mapped  */
   int OverlayIsMapped;

   /* for DumpToEpsFile: Added by Miguel A. de Riera Pasenau 10.01.1997 */
   /** Visual info of the current */
   XVisualInfo *EpsVisual;	
					/* context needed for DumpToEpsFile */
public:
   /** Index2RGB Maps for Color index modes */
   /** Index2RGB red Map for Color index modes */
   GLfloat *EpsRedMap;	
   /** Index2RGB green Map for Color index modes */
   GLfloat *EpsGreenMap;
   /** Index2RGB blue Map for Color index modes */
   GLfloat *EpsBlueMap;
   /**  Number of indices in our Togl */
   GLint EpsMapSize; 
  // class members
protected:
#if defined(X11)
   /// the General OpenGL context of the camera
 static GLXContext masterCtxtHandle;
 /// the display of the application
  static  Display *masterDisplayHandle;
#elif defined(WIN32)
    /// the General OpenGL context of the camera
 static HGLRC masterCtxtHandle;
#endif
  /// the default client data transmitted at the initialization of the widget
  static ClientData DefaultClientData;
#ifdef MESA_COLOR_HACK
  /// array where is stored the used color cells when using MESAGL
  static unsigned long *ToglMesaUsedPixelCells;
  /// the number of used color cells when using MESAGL
  static int ToglMesaUsedFreeCells;
#endif
  /// the array describing the syntax of the original "togl" TCL command
  static Tk_ConfigSpec configSpecs[];

  // class functions
public:
  ///Called upon system startup to create Togl command.
  static int Init(TclModule *m);
  /// Changes the default Client Data
  static void SetDefaultClientData(ClientData clientData) {
    DefaultClientData = clientData;
  }

  // constructors
protected:
  /// parse the TCL command line and eventually recreate the window and GLX context
  int configure(int argc, char *argv[], int flags);

public:
  /** @name constructor */
  //@{
  /** builds a new 3D OpenGL/TCL camera
      @param m module to build camera in
      @param parent parent window to create camera in (if EMPTY_STRING, camera is created in a new toplevel window)
      @param scene visualized scene interface
      @param stereo if true, allocates a new stereo buffer
      @param type dynamic object type
      @param cam client (if already created)
      @exception InvalidCommandLine invalid command line argument transmitted to Togl command
      @exception CannotMakeWindowExist unable to create TCL window */
  BasicTogl(TclModule *m);
  //@}


  /** @name destructor */
  //@{
  /** destroys a Togl */
  virtual ~BasicTogl();
  //@}



  /// function that makes the current context active
virtual void makeCurrent(void) const=0;

  // Query functions
  /// returns the user's identification string
  char *GetIdent(void) const { return Ident; }
  /// returns the associated TCL interpretor
  Tcl_Interp *GetInterp(void) const { return interp; }
 
  virtual bool isCameraBeingDestroyed() const=0;
  virtual unsigned int getWidth() const=0;
  virtual unsigned int getHeight() const=0;

  virtual Tk_Window getTkWindow() const=0;
#ifdef WIN32
  virtual HDC getDeviceContext() const=0;
  virtual HGLRC getRenderingContext() const=0;
  static HGLRC getMasterCtxtHandle()  {
	  return masterCtxtHandle;
  }
  protected:
virtual void setRenderingContext(const HGLRC c)=0;
virtual void setDeviceContext(const HDC c)=0;

	public:
  #elif defined(X11)
	/** Normal planes GLX context */
	virtual GLXContext getGLXContext()const=0;
protected:
virtual void setGLXContext(const GLXContext c)=0;
	public:
#endif /* WIN32 */

  // Color Index mode
  /** When using color index mode, try to get a color cell as close 
      as possible to a given (red,green,blue) color
      @param red the red component (between 0.0 and 1.0)
      @param green the green component (between 0.0 and 1.0)
      @param blue the blue component (between 0.0 and 1.0) */
  unsigned long AllocColor(float red, float green, float blue) const;
  /** When using color index mode, remove a color cell previously allocated
      by AllocColor 
      @param index the index of the color in the colormap */
  void FreeColor(unsigned long index) const;
  /** When using color index mode, allocate a color cell as close 
      as possible to a given (red,green,blue) color
      @param index
      @param red the red component (between 0.0 and 1.0)
      @param green the green component (between 0.0 and 1.0)
      @param blue the blue component (between 0.0 and 1.0) */  
  void SetColor(unsigned long index, float red, float green, float blue) const;

  // Bitmap fonts
  /** Load the named bitmap font as a sequence of bitmaps in a display list.
      fontname may be one of the predefined fonts like TOGL_BITMAP_8_BY_13
      or an X font name, or a Windows font name, etc. */
  GLuint LoadBitmapFont(const char *fontname) const;
  /** Release the display lists which were generated by LoadBitmapFont().*/
  void UnloadBitmapFont(GLuint fontbase) const;


  // Overlay functions
  /// redraw as overlay the scene
  virtual void redraw_overlay(void) const {}
  /** defines the layer where to draw the overlay
      @param layer either TOGL_OVERLAY or TOGL_NORMAL */
  void UseLayer(int layer);
  /** map the overlay window */
  void ShowOverlay(void);
  /** unmap the overlay window */
  void HideOverlay(void);
  /// query a redisplay of the overlay window
  void PostOverlayRedisplay(void);
  /** whether there is an overlay window
      @return boolean */
  int ExistsOverlay(void) const { return OverlayFlag; }
  /** returns the transparent pixel value 
   @return the index in the colormap of the transparent pixel*/
  int GetOverlayTransparentValue() const {
    return OverlayTransparentPixel;
  }
  /** returns if the overlay has been mapped */
  int IsMappedOverlay(void) const {
    return OverlayFlag && OverlayIsMapped;
  }
  /** When using color index mode for overlay, try to get a color cell 
      as close as possible to a given (red,green,blue) color
      @param red the red component (between 0.0 and 1.0)
      @param green the green component (between 0.0 and 1.0)
      @param blue the blue component (between 0.0 and 1.0) */  
  unsigned long AllocColorOverlay(float red, float green, 
				       float blue) const;
  /** When using color index mode for overlay, remove a color cell 
      previously allocated  by AllocColorOverlay 
      @param index the index of the color in the colormap */
  void FreeColorOverlay(unsigned long index) const;

  // User client data
  /** returns the current client_data */
  ClientData GetClientData(void) const { return Client_Data; }
  /** sets the client data */
  void SetClientData(ClientData clientData) {
    Client_Data = clientData;
  }


  // X11-only commands.
#ifdef TOGL_X11
  /** return the X11 Display associated with the widget 
      @return the X11 Display */
  Display *GetDisplay(void) const { return Tk_Display(TkWin); }
  /** return the X11 Screen associated with the widget 
      @return the X11 Screen */
  Screen *GetScreen(void) const { return Tk_Screen(TkWin); }
  /** return the X11 Screen number associated with the widget 
      @return the X11 Screen number */
  int GetScreenNumber(void) const { return Tk_ScreenNumber(TkWin); }
  /** return the X11 colormap  associated with the widget 
      @return the X11 colormap */
  Colormap GetColormap(void) const { return Tk_Colormap(TkWin); }
#endif


  // Generate EPS file.
  /** save an OpenGL scene in a encapsulated postcript file 
      @param filename the name of the file where to write the Postcript
      @param inColor boolean indicating if the image should be considered as a color or grey level image
      @param user_redraw a routine that should draw the scene by using OpenGL calls 
      @return 0 if no problems were encountered, 1 if there is a color allocation problem, 2 if there is a problem when opening the file*/
  int DumpToEpsFile(const char *filename, int inColor,
		    void (*user_redraw)(const BasicTogl *)) const;
  
protected:
  /** Modified version of Tk_MakeWindowExist.Creates an OpenGL window for 
      the Togl widget.
      @return TCL_OK or TCL_ERROR */
  int MakeWindowExist(void);
  /** Do all the setup for overlay planes
      @return  TCL_OK or TCL_ERROR */
  int SetupOverlay(void);
#if defined(WIN32)
  /** creates a colormap for Windows
      @return colormap */
  Colormap Win32CreateCiColormap(void);
#endif


  // friends
  /** destroy the widget and the rendering context when possible 
      @param clientData pointer to a Togl casted as ClientData */
  static void destroyContext(ClientData clientData);
  /** render the scene by calling Camera3D::redraw() 
      @param clientData pointer to a Togl casted as ClientData */
  static void redrawWhenIdle(ClientData clientData);
  /** render in the overlay planes by calling redraw_overlay()
      @param clientData pointer to a Togl casted as ClientData */  
  static void renderOverlayWhenIdle(ClientData clientData);
  /** resize camera
      @param clientData pointer to a Togl casted as ClientData */
  static void resizeWhenIdle(ClientData clientData);
  /** destroy the widget 
      @param clientData pointer to a Togl casted as ClientData */    
  static void Destroy(char *clientData);
 
  /**  Gets called from Tk_CreateTimerHandler.
       @param clientData pointer to a Togl casted as ClientData   */
  static void Timer(ClientData);
 
 
};



} // namespace yav


#endif

