#ifndef TCLCONSTRAINT_H
#define TCLCONSTRAINT_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#pragma interface "modules/TclConstraint.h"
#include <framework/Constraint.h>
#include <modules/TclObject.h>



namespace yav {

class Constraint;

/** \class TclConstraint TclConstraint.h modules/TclConstraint.h
    \ingroup Modules
    base class for Tcl interface of a Constraint object : used by Mesh obhect 
    to manipulate Constraint (base class) without knowing the explicit nature of
    the inherited classes */
 class TclConstraint : public TclObject {
   protected :
     /// the constraint pointed by the Tcl interface
     Constraint *constraint;
   public :
   /** @name constructor */
   //@{
   /** creates a Tcl Interface of a Constraint object
       @param m module to create TCL object
       @param c a pointer to a Constraint object 
   */
   TclConstraint(TclModule *m,
		 Constraint *c);
   /** creates a Tcl Interface of a Constraint object : called by an inherited class
       @param type the type of the inherited object that is  really created
       @param m module to create TCL object
       @param c a pointer to a Constraint object */
   TclConstraint(const std::type_info& type, TclModule *m,
		 Constraint *c);
  //@}

  /** @name destructor */
  //@{
  /// destroys this interface
   virtual ~TclConstraint();
  //@}

   /// new TCL command callback
   static int create(TclModule *m, int argc, char **argv);

   /// -destroy option
   int destroyCmd(int argc, char **argv);
   /// -deleteConstraint option
   int deleteConstraintCmd(int argc, char **argv);
   /// -frequency option
   int frequencyCmd(int argc, char **argv);
   /// -constraintName option
   int constraintNameCmd(int argc, char **argv);
   /// -cast option
   int castCmd(int argc, char **argv);
   /// -sameConstraint option
   int sameConstraintCmd(int argc, char **argv);

   /// returns the Constraint object associated with a TclConstraint
   Constraint *getConstraint() const;

   static void createCommands(TclModule *m);
 /** updates the database that associated with each constraint name, the name of
     a Tcl procedure that can be used to edit a TclConstraint of that type */
 static void registerConstraintEditProcedure(TclModule *m,
					    const char *constraintName,
					    const char *procedureName,
					    const char *tclname);
 };
}

#endif
