#ifndef TCLCONSTRAINTTYPE_H
#define TCLCONSTRAINTTYPE_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#pragma interface "modules/TclConstraintType.h"
#include <modules/TclConstraint.h>


/** a macro used to automatically generate an option command of a TclConstraintType<CLIENT> object.
    This option command simply set and returns the value of a float. The user must provide the name
    of the setValue() and getValue() as well as the name of the procedure */
#define CREATE_FLOAT_OPTION_COMMAND(PROCNAME,CLIENT,SETROUTINE,GETROUTINE) int PROCNAME(TclObject *obj,int argc, char **argv) {\
  TclConstraintType<CLIENT> *c=(TclConstraintType<CLIENT> *) obj;   if (argc>1) return COMMAND_SYNOPSIS; \
  if (argc==1) { double d=(double) atof(argv[0]);  c->getConstraint()->SETROUTINE(d);  } \
  c->setResult("%f", c->getConstraint()->GETROUTINE());   return TCL_OK; }

namespace yav {

  /// defines a callback function that can be called for in the constructor or destructor
  typedef void (*TCLCONSTRAINTTYPE_CALLBACK)(TclObject *obj);

/** template class that is useful to quickly generate a Tcl Interface of a constraint class. The template
    argument is a class that derives from the Constraint class. The user still has to define the set of option
    commands for that class and to write a creation routine and a command definition routine */
template <class CLIENT> class TclConstraintType : public TclConstraint {
 protected:
  /// callback called in the constructor
  TCLCONSTRAINTTYPE_CALLBACK initRoutine;
  /// callback called in the destructor
  TCLCONSTRAINTTYPE_CALLBACK destroyRoutine;
   public :
/** @name exceptions */
  //@{
     /*** This exception is thrown when trying to create a TclConstraintType object while pointing on a 
	  Constraint class different from CLIENT */
  class NotTheRightConstraintType : public Exception {
    public :
    NotTheRightConstraintType(const std::string& from) :
      Exception(from, "Trying to create a  TclConstraintType from a constraint of wrong type") {}
  };
  //@}


   /** @name constructor */
   //@{
   /** creates a Tcl Interface of a Constraint object
       @param m module to create TCL object
       @param c a pointer to a Constraint object
       @param init
       @param destroy
   */
   TclConstraintType(TclModule *m,
		     CLIENT *c,TCLCONSTRAINTTYPE_CALLBACK init=0,TCLCONSTRAINTTYPE_CALLBACK destroy=0) 
     throw(NotTheRightConstraintType): 
     TclConstraint(typeid(TclConstraintType<CLIENT>),m,c) {
     if (c==0)  throw(NotTheRightConstraintType(" TclConstraintType(TclModule *m, CLIENT *c)"));
     initRoutine=init;
     destroyRoutine=destroy;
     if (init) 
       (*init)(this);
     done(typeid(TclConstraintType<CLIENT>));
   }
   /** creates a Tcl Interface of a Constraint object from a TclConstraint : it
       checks if the constraint pointed by TclConstraint is of type CLIENT
       @param c a TclConstraint pointing on a Constraint object of type
       CLIENT
       @param init
       @param destroy
   */
   TclConstraintType(TclConstraint *c,TCLCONSTRAINTTYPE_CALLBACK init=0,
                     TCLCONSTRAINTTYPE_CALLBACK destroy=0)
     throw(NotTheRightConstraintType): TclConstraint(typeid(TclConstraintType<CLIENT>),c->getModule(),c->getConstraint()) {
     if ((c->getConstraint()==0) || (dynamic_cast<CLIENT *>(c->getConstraint())==0)) throw(NotTheRightConstraintType(" TclConstraintType(TclModule *m, CLIENT *c)"));
     initRoutine=init;
     destroyRoutine=destroy;
     if (init) 
       (*init)(this);

     done(typeid(TclConstraintType<CLIENT>));     
   }
  //@}



  /** @name destructor */
  //@{
  /// destroys this interface
   virtual ~TclConstraintType(){
     if (destroyRoutine)
       (*destroyRoutine)(this);
   }
  //@}

 public:
   /// a pointer on a structure
   void *clientData;
 public:

   /** This is a basic creation routine that handles the creation from a TclConstraint. It should be called by
    the more general creation routine */
   static int create(TclModule *m, int argc, char **argv,TCLCONSTRAINTTYPE_CALLBACK init=0,
		     TCLCONSTRAINTTYPE_CALLBACK destroy=0) {
     if (argc==2) {
       TclConstraint *tc=(TclConstraint *)  TclObjectKind::getObject(typeid(TclConstraint), argc, argv, 1);
       if (tc==0) {
	 m->setResult("Illegal constraint name : %s",argv[1]);
	 return TCL_ERROR;
       }
       TclConstraintType<CLIENT> *c= new TclConstraintType<CLIENT>(tc,init,destroy);
       m->setResult("%s",(char *)c->name());
       return TCL_OK;
     } else
       return COMMAND_SYNOPSIS; 
   }
   /** This is a basic command definition routine that handles the creation from a TclObjectKind
       @param m a module
       @param objName the name that will be used to name TclConstraintType<CLIENT> in a Tcl interface. The commands will be objName%u 
       @param realName a sentence that describes the TclConstraintType<CLIENT> object 
       @param synopsis the synopsis for the creation of the object 
       @param createFromArgs the global creation routine
   */
   static TclObjectKind *createCommands(TclModule *m,char *objName,char *realName,char *synopsis, 
					int (*createFromArgs)(TclModule *, int, char **)) {
     
    return( new TclObjectKind(m, typeid(TclConstraintType<CLIENT>), typeid(TclConstraint),objName,
			      realName,synopsis, createFromArgs,&TclObject::list));

   }
   /// returns the constraint associated with this object
   CLIENT *getConstraint() const {
     return (CLIENT *)TclConstraint::constraint; }
 };
}

#endif
