#ifndef TCLEXPECTATIONMAXIMIZATION_H
#define TCLEXPECTATIONMAXIMIZATION_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "modules/TclExpectationMaximization.h"
#include <classification/ExpectationMaximization.h>
#include <modules/TclObject.h>
#include <framework/Interface.h>

namespace yav {

/** \class TclExpectationMaximization TclExpectationMaximization.h modules/TclExpectationMaximization.h
    \ingroup Modules
    base class for Tcl interface of a range Information object : used for
    exchanging information between deformable models and range datasets */
 class TclExpectationMaximization : public TclObject {
   protected :
     /// the rangeInformation pointed by the Tcl interface
     ExpectationMaximization *em;
   public :
   /** @name constructor */
   //@{
   /** creates a Tcl Interface of a ExpectationMaximization object
       @param m module to create TCL object
       @param ri a pointer to a ExpectationMaximization object */
   TclExpectationMaximization(TclModule *m, ExpectationMaximization *_em) : TclObject(typeid(TclExpectationMaximization),m),em(_em){
     done(typeid(TclExpectationMaximization));
   }

  //@}

  /** @name destructor */
  //@{
  /// destroys this interface
   virtual ~TclExpectationMaximization(){
   deleted(typeid(TclExpectationMaximization));
   }
  //@}

   /// new TCL command callback
   static int list(TclModule *m, int argc, char **argv) {
     return TclObject::list(m, argc, argv);
   }
   /// new TCL command callback
   static int create(TclModule *m, int argc, char **argv) {
     if (argc!=1)
       return COMMAND_SYNOPSIS; 
     TclExpectationMaximization *em=new TclExpectationMaximization (m,0);
     m->setResult("%s%u", (char *)em->getObjectName(), em->TclObject::getRef());
     return TCL_OK;
   }
   int nbrModesCmd(int argc, char **argv) {     
     if(argc != 0) return COMMAND_SYNOPSIS;
     setResult("%u",em->getNbrModes());
     return TCL_OK;
   }
   int nbrDataCmd(int argc, char **argv) {     
	   if(argc != 0) return COMMAND_SYNOPSIS;
	   setResult("%u",em->getNbrData());
	   return TCL_OK;
   }
   int nbrCoordCmd(int argc, char **argv) {     
	   if(argc != 0) return COMMAND_SYNOPSIS;

	unsigned int nbrCoord;
	nbrCoord = em->getNbrDim();	

	   setResult("%u",nbrCoord);
	   return TCL_OK;
   }
   int maxIterCmd(int argc, char **argv) {     
	   if(argc != 0) return COMMAND_SYNOPSIS;
	   setResult("%u",em->getMaxIte());
	   return TCL_OK;
   }
   int nbrIterCmd(int argc, char **argv) {     
	   if(argc != 0) return COMMAND_SYNOPSIS;
	   setResult("%u",em->getNbrIteUsed());
	   return TCL_OK;
   }
   int minImproCmd(int argc, char **argv) {     
	   if(argc != 0) return COMMAND_SYNOPSIS;
	   setResult("%1.5f",em->getMinImpro());
	   return TCL_OK;
   }
   int vpcCmd(int argc, char **argv) {     
	   if(argc != 0) return COMMAND_SYNOPSIS;
	   setResult("%f",em->getVpc());
	   return TCL_OK;
   }
   int vpeCmd(int argc, char **argv) {     
	   if(argc != 0) return COMMAND_SYNOPSIS;
	   setResult("%f",em->getVpe());
	   return TCL_OK;
   }
   int overlapCmd(int argc, char **argv) {     
	   if(argc != 0) return COMMAND_SYNOPSIS;
	   setResult("%f",em->getVo());
	   return TCL_OK;
   }
   int separationCmd(int argc, char **argv) {     
	   if(argc != 0) return COMMAND_SYNOPSIS;
	   setResult("%f",em->getVs());
	   return TCL_OK;
   }
   int modeCenterCmd(int argc, char **argv) {     
	   if(argc != 1) return COMMAND_SYNOPSIS;
	   unsigned int idMode, i, nbrDim;
	   TclString res, tmp;

	   idMode = (unsigned int) atoi(argv[0]);
	   double* centers = em->getCenters();
	   nbrDim = em->getNbrDim();

	   for (i=0; i<nbrDim; i++)
	   {
		   tmp.sprintf("{%6f} ",centers[(idMode*nbrDim)+i]);
		   res+=tmp;
	   }

	   setResult(res);
	   return TCL_OK;
   }
int modeStdDeviationCmd(int argc, char **argv) {     
	   if(argc != 1) return COMMAND_SYNOPSIS;
	   unsigned int idMode, i, nbrDim;
	   TclString res, tmp;
	   std::vector<std::vector<double> > covMatrices;
	   double stdDev;

	   idMode = (unsigned int) atoi(argv[0]);
	   covMatrices = em->getCovMatrices();
	   nbrDim = em->getNbrDim();

	   for (i=0; i<nbrDim; i++)
	   {
		stdDev=sqrt(covMatrices[idMode][(i*em->getNbrDim())+i]);
		tmp.sprintf("{%6f} ",stdDev);
		res+=tmp;
	   }

	   setResult(res);
	   return TCL_OK;
   }

  /// Command to display data associated to a vertex 
  int dataCmd(int argc, char **argv)
  {
	if(argc != 1) return COMMAND_SYNOPSIS;

	unsigned int i, dataIndex, nbrDim;
	TclString res, tmp;

	if (argv[0][0]=='v')
	{
		// If we receive a "v%u"-kind of data, we should retrieve the "v" to extract the number of the vertex
		TclString* check = new TclString ((const char *) argv[0]);
		const char* correct = check->truncAtBegining(1);
		dataIndex = (unsigned int) atoi(correct);
	}
	else
		// If not, if we receive a "%u"-kind of data, we can directly use the input as the number of the vertex
		dataIndex = (unsigned int) atoi(argv[0]);

	// Take the data corresponding to the vertex
	double** data = em->getData();
	nbrDim = em->getNbrDim();

	for (i=0; i<nbrDim; i++)
	{
		tmp.sprintf("{%6f} ",data[dataIndex][i]);
		res+=tmp;
	}

	setResult(res);

	return TCL_OK;
  }
  /// Command to display membership associated to a vertex
  int probaCmd(int argc, char** argv)
  {
	if(argc!=1)
		return COMMAND_SYNOPSIS;

	unsigned int idData, nModes;
	int i;
	TclString res, tmp;

	if (argv[0][0]=='v')
	{
		// If we receive a "v%u"-kind of data, we should retrieve the "v" to extract the number of the vertex
		TclString* check = new TclString ((const char*) argv[0]);
		const char* correct = check->truncAtBegining(1);
		idData = (unsigned int) atoi(correct);
	}
	else
		// If not, if we receive a "%u"-kind of data, we can directly use the input as the number of the vertex
		idData = (unsigned int) atoi(argv[0]);

	// Take the data corresponding to the vertex
	double* proba = em->getProba();
	nModes = em->getNbrModes();

	for (i=(nModes-1); i>=0; i--)
	{
		tmp.sprintf("{%6f} ",proba[(idData*nModes)+i]);
		res += tmp;
	}

	setResult(res);

	return TCL_OK;
  }
   /// -destroy option
   int destroyCmd(int argc, char **argv) {
	   TclModule* m;
	   unsigned int ref;

	   if(argc != 0) return COMMAND_SYNOPSIS;
	   m = getModule();
	   ref = TclObject::getRef();
	   TclString name=TclExpectationMaximization::getObjectName();
	   delete this;
	   m->setResult("%s%u", (char*) name, ref);
	   return TCL_OK;
   }
   /// returns the ExpectationMaximization associated with a TclExpectationMaximization
   ExpectationMaximization *getExpectationMaximization() const {
     return em; }
 };

}

#endif
