#ifndef TCLFUZZYCMEANS_H
#define TCLFUZZYCMEANS_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#pragma interface "modules/TclFuzzyCMeans.h"
#include <classification/FuzzyCMeans.h>
#include <modules/TclObject.h>
#include <framework/Interface.h>
#include <classification/OrientedPoints2FCM.h>

namespace yav {

/** \class TclFuzzyCMeans TclFuzzyCMeans.h modules/TclFuzzyCMeans.h
    \ingroup Modules
    base class for Tcl interface of a range Information object : used for
    exchanging information between deformable models and range datasets */
 class TclFuzzyCMeans : public TclObject {
   protected :
     /// the rangeInformation pointed by the Tcl interface
     FuzzyCMeans *fcm;
     bool onlyIntensity;
     Metric metricUsed;
   public :
   /** @name constructor */
   //@{
   /** creates a Tcl Interface of a FuzzyCMeans object
       @param m module to create TCL object
       @param ri a pointer to a FuzzyCMeans object */
   TclFuzzyCMeans(TclModule *m, FuzzyCMeans *_fcm, bool _onlyIntensity=true, Metric _metricUsed=EUCLIDIAN_POSITION) : TclObject(typeid(TclFuzzyCMeans),m),fcm(_fcm),onlyIntensity(_onlyIntensity),metricUsed(_metricUsed){
     done(typeid(TclFuzzyCMeans));
   }

  //@}

  /** @name destructor */
  //@{
  /// destroys this interface
   virtual ~TclFuzzyCMeans(){
   deleted(typeid(TclFuzzyCMeans));
   }
  //@}

   /// new TCL command callback
   static int list(TclModule *m, int argc, char **argv) {
     return TclObject::list(m, argc, argv);
   }
   /// new TCL command callback
   static int create(TclModule *m, int argc, char **argv) {
     if (argc!=1)
       return COMMAND_SYNOPSIS; 
     TclFuzzyCMeans *fcm=new TclFuzzyCMeans (m,0);
     m->setResult("%s%u", (char *)fcm->getObjectName(), fcm->TclObject::getRef());
     return TCL_OK;
   }
   int nbrClustersCmd(int argc, char **argv) {     
     if(argc != 0) return COMMAND_SYNOPSIS;
     setResult("%u",fcm->getNbrClust());
     return TCL_OK;
   }
   int nbrDataCmd(int argc, char **argv) {     
	   if(argc != 0) return COMMAND_SYNOPSIS;
	   setResult("%u",fcm->getNbrData());
	   return TCL_OK;
   }
   int nbrCoordCmd(int argc, char **argv) {     
	   if(argc != 0) return COMMAND_SYNOPSIS;

	unsigned int nbrCoord;

	if (metricUsed==EUCLIDIAN_POSITION && !onlyIntensity)
		nbrCoord = (fcm->getNbrDim())-3;
	else
		nbrCoord = fcm->getNbrDim();	

	   setResult("%u",nbrCoord);
	   return TCL_OK;
   }
   int expoCmd(int argc, char **argv) {     
	   if(argc != 0) return COMMAND_SYNOPSIS;
	   setResult("%1.1f",fcm->getExpo());
	   return TCL_OK;
   }
   int maxIterCmd(int argc, char **argv) {     
	   if(argc != 0) return COMMAND_SYNOPSIS;
	   setResult("%u",fcm->getMaxIte());
	   return TCL_OK;
   }
   int nbrIterCmd(int argc, char **argv) {     
	   if(argc != 0) return COMMAND_SYNOPSIS;
	   setResult("%u",fcm->getNbrIteUsed());
	   return TCL_OK;
   }
   int minImproCmd(int argc, char **argv) {     
	   if(argc != 0) return COMMAND_SYNOPSIS;
	   setResult("%1.5f",fcm->getMinImpro());
	   return TCL_OK;
   }
   int vpcCmd(int argc, char **argv) {     
	   if(argc != 0) return COMMAND_SYNOPSIS;
	   setResult("%f",fcm->getVpc());
	   return TCL_OK;
   }
   int vpeCmd(int argc, char **argv) {     
	   if(argc != 0) return COMMAND_SYNOPSIS;
	   setResult("%f",fcm->getVpe());
	   return TCL_OK;
   }
   int overlapCmd(int argc, char **argv) {     
	   if(argc != 0) return COMMAND_SYNOPSIS;
	   setResult("%f",fcm->getVo());
	   return TCL_OK;
   }
   int separationCmd(int argc, char **argv) {     
	   if(argc != 0) return COMMAND_SYNOPSIS;
	   setResult("%f",fcm->getVs());
	   return TCL_OK;
   }
   int onlyIntensityCmd(int argc, char **argv) {     
     if(argc != 0) return COMMAND_SYNOPSIS;
     setResult("%u",onlyIntensity);
     return TCL_OK;
   }
   int clusterCenterCmd(int argc, char **argv) {     
	   if(argc != 1) return COMMAND_SYNOPSIS;
	   unsigned int clusterIndex, i, nbrCoordinates, countCoord;
	   TclString res, tmp;
	   double * membershipFunction;

	   clusterIndex=(unsigned int) atoi(argv[0]);
	   membershipFunction = fcm->getCenters();
	   nbrCoordinates=fcm->getNbrDim();

	   // Check if fcm takes only intensities or position and orientation as well
	   if (onlyIntensity)
		   countCoord = 0;
	   else
		   countCoord = 6;

	   for (i=countCoord; i<nbrCoordinates; i++)
	   {
		   tmp.sprintf("{%6f} ",membershipFunction[(clusterIndex*nbrCoordinates)+i]);
		   res+=tmp;
	   }
	   //	res+= "";

	   setResult(res);
	   return TCL_OK;
   }

  /// Command to display data associated to a vertex 
  int dataCmd(int argc, char **argv)
  {
	if(argc != 1) return COMMAND_SYNOPSIS;

	unsigned int i, dataIndex, nbrCoordinates;
	double ** data;
	TclString * check, res, tmp;
	const char * correct;

	if (argv[0][0]=='v')
	{
		// If we receive a "v%u"-kind of data, we should retrieve the "v" to extract the number of the vertex
		check = new TclString ((const char *) argv[0]);
		correct = check->truncAtBegining(1);
		dataIndex = (unsigned int) atoi(correct);
	}
	else
		// If not, if we receive a "%u"-kind of data, we can directly use the input as the number of the vertex
		dataIndex = (unsigned int) atoi(argv[0]);

	// Take the data corresponding to the vertex
	data = fcm->getData();
	nbrCoordinates=fcm->getNbrDim();

	// If the metric used is EUCLIDIAN_POSITION, the orientation is not taken into account	
	if (metricUsed==EUCLIDIAN_POSITION && !onlyIntensity)
	{
		for (i=0; i<3; i++)
		{
			tmp.sprintf("{%6f} ",data[dataIndex][i]);
			res+=tmp;
		}
		for (i=6; i<nbrCoordinates; i++)
		{
			tmp.sprintf("{%6f} ",data[dataIndex][i]);
			res+=tmp;
		}
	}
	else
	{
		for (i=0; i<nbrCoordinates; i++)
		{
			tmp.sprintf("{%6f} ",data[dataIndex][i]);
			res+=tmp;
		}
	}

	setResult(res);

	return TCL_OK;
  }
  /// Command to display membership associated to a vertex
  int membershipCmd(int argc, char **argv)
  {
	if(argc != 1) return COMMAND_SYNOPSIS;

	unsigned int membershipIndex, nbrClusters, nbrData;
	int i;
	double * U;
	TclString * check, res, tmp;
	const char * correct;

	if (argv[0][0]=='v')
	{
		// If we receive a "v%u"-kind of data, we should retrieve the "v" to extract the number of the vertex
		check = new TclString ((const char *) argv[0]);
		correct = check->truncAtBegining(1);
		membershipIndex = (unsigned int) atoi(correct);
	}
	else
		// If not, if we receive a "%u"-kind of data, we can directly use the input as the number of the vertex
		membershipIndex = (unsigned int) atoi(argv[0]);

	// Take the data corresponding to the vertex
	U = fcm->getMembership();
	nbrClusters=fcm->getNbrClust();
	nbrData=fcm->getNbrData();

	for (i=(nbrClusters-1); i>=0; i--)
	{
		tmp.sprintf("{%6f} ",U[(i*nbrData)+membershipIndex]);
		res+=tmp;
	}

	setResult(res);

	return TCL_OK;
  }
  /// Command to display metric used by FCM
  int metricCmd(int argc, char **argv)
  {
	if(argc != 0) return COMMAND_SYNOPSIS;

	TclString res;

	if (onlyIntensity)
		res = "Metric using only intensity";
	else
	{
		if (metricUsed==EUCLIDIAN_POSITION)
			res = "Euclidian metric using position";
		else if (metricUsed==EUCLIDIAN_POSITION_ORIENTATION)
			res = "Euclidian metric using position & orientation";
		else
			res = "No metric used";
	}

	setResult(res);

	return TCL_OK;
  }
   /// -destroy option
   int destroyCmd(int argc, char **argv) {
	   TclModule *m;
	   unsigned int ref;

	   if(argc != 0) return COMMAND_SYNOPSIS;
	   m = getModule();
	   ref = TclObject::getRef();
	   TclString name=TclFuzzyCMeans::getObjectName();
	   delete this;
	   m->setResult("%s%u", (char *) name, ref);
	   return TCL_OK;
   }
   /// returns the fuzzyCMeans associated with a TclFuzzyCMeans
   FuzzyCMeans *getFuzzyCMeans() const {
     return fcm; }
 };

}

#endif
