#ifndef TCLMATRIX4D_H
#define TCLMATRIX4D_H

#pragma interface "modules/TclMatrix4D.h"
#include <modules/TclObject.h>
#include <linearAlgebra/Matrix4x4.h>

namespace yav {

  
  
/** \class TclMatrix4D
    \ingroup Modules
    \brief Tcl interface for any double 4x4  matrix  */
class TclMatrix4D :  public TclObject, public Matrix4x4<double>
  {
  protected:
    
    /// name of the matrix
    TclString name;
  public:
    /** @name constructors */
    //@{
    /** Creates an interface object 
	@param info dynamic object type
	@param m module this matrix was created in
    */
      TclMatrix4D(const std::type_info& info, TclModule *m);
      /** Creates an interface object for an existing mat. 
	  @param m module this matrix was created in
	  @param t value of the matrix
      */
      TclMatrix4D(TclModule *m, double t[16]);
      /** Creates an interface object for an existing mat. 
	  @param m module this matrix was created in
	  @param t value of the matrix [i][j]
      */
      TclMatrix4D(TclModule *m, double t[4][4]);
      /** Creates an interface object for an existing mat. 
	  @param m module this matrix was created in
	  @param mat the matrix
      */
      TclMatrix4D(TclModule *m, Matrix4x4<double> &mat);
      //@}

      /** @name destructor */
      //@{
      // deallocates image buffers and destroy object
      ~TclMatrix4D();
      //@}

      //!
      TclMatrix4D & operator=(const Matrix4x4<double> &);



      // return object by ref or 0 if it is not found
      static TclMatrix4D *getObject(unsigned int n) {
	return dynamic_cast<TclMatrix4D*>(TclObjectKind::getObject(typeid(TclMatrix4D),
								    n));
      }

      // return object by name (argv[i]) or 0 with an error message
      // in module interpretor if it is not found
      static TclMatrix4D *getObject(int argc, char **argv, int i = 1) {
	return
	  dynamic_cast<TclMatrix4D *>(TclObjectKind::getObject(typeid(TclMatrix4D),
							       argc, argv, i));
      }
      
      // return object id by command line arguments
      static const char *getTCLObjectName(void) {
	return TclObjectKind::getObjectName(typeid(TclMatrix4D));
      }
      
      // return object id by command line arguments
      static int list(TclModule *m, int argc, char **argv) {
	return TclObject::list(m, argc, argv);
      }

      /// createMatrix command callback
      static int create(TclModule *m, int argc, char **argv);



      /// -val command option
      int valCmd(int argc, char **argv);
      /// -name command option
      int nameCmd(int argc, char **argv);
      /// -id command option
      int idCmd(int argc, char **argv);
      /// -raz command option
      int razCmd(int argc, char **argv);
      /// -inv command option
      int invCmd(int argc, char **argv);
      /// -mul command option
      int mulCmd(int argc, char **argv);
      /// -leftMul command option
      int leftMulCmd(int argc, char **argv);
      /// -set command option
      int setCmd(int argc, char **argv);
      /// -destroy command option
      int destroyCmd(int argc, char **argv);
      /// -saveRowMajorCmd command option
      int saveRowMajorCmd(int argc, char **argv);
      /// -saveColumnMajorCmd command option
      int saveColumnMajorCmd(int argc, char **argv);
      /// -saveAstransfoCmd command option
      int saveAstransfoCmd(int argc, char **argv);
      /// -loadAstransfoCmd command option
      int loadAstransfoCmd(int argc, char **argv);
 /// -loadFromMEDINRIACmd command option
      int loadFromMedINRIACmd(int argc, char **argv);

    };

} // namespace yav

#endif




