#ifndef TCLPARENT_H
#define TCLPARENT_H

#pragma interface "modules/TclParent.h"
#include <map>
#include <string>
#include <modules/TclObject.h>

namespace yav {

/** \class TclParent
    \ingroup Modules
    \brief Class to keep track of objects derived from T with different tcl names. */
template<class T> class TclParent : public TclObject {
   public:
      /** @name Constructors */
      //@{
      /** Add this to child of parent. The list of existing
	  children is available in the <b>children</b> variable.
	  @param t object kind
	  @param m module this object is created in */
      TclParent(const type_info& t,TclModule* m=0) : TclObject(t,m) {
	 children[static_cast<char*>(name())]=static_cast<T*>(this);
      }
      //@}
      /** @name Destructor */
      //@{
      /// destroy associated name in the <b>children</b> variable.
      ~TclParent() {children.erase(static_cast<char*>(name()));}
      //@}
      /** find a children.
	  @param k the child's key
	  @return the associated pointer or 0 if not found */
      static T* find(string const& k){
	 map<string,T*>::iterator i=children.find(k);
	 if(i==children.end()) return 0;
	 return i->second;
      }
      /** add a children to list
	  @param k the child's key
          @param t
	  @return the associated pointer */
      static T* add(string const& k,T* t){
	 return children[k]=t;
      }
      /** add a children to list
	  @param k the child's key
	  @return the associated pointer */
      static void erase(string const& k){
	 children.erase(k);
      }
      /** returns list with name of children */
      static int childrenCmd(TclModule* m,int argc,char** argv);
   protected:
      /// lookup map of available children of T by TCL name.
      static map<string,T*> children;
};

template<class T> int TclParent<T>::childrenCmd(TclModule* m,int argc,char** argv){
   if(argc!=1) return COMMAND_SYNOPSIS;
   string s="";
   bool first=true;
   for(map<string,T*>::const_iterator i=children.begin();i!=children.end();++i){
      if(!first) s+=" "; else first=false;
      s+=i->first;
      if(dynamic_cast<T*>(i->second)==0) s+=" (error: not defined!)";
   }
   m->setResult("%s",s.c_str());
   return TCL_OK;
}

/// Build a TclParent, i.e. the TclObject and the static children map.
/// AutoLink<T> class static variables declaration macro
#define TCL_PARENT_STATIC_VARIABLES(type) \
map<string,type*> TclParent< type >::children;
#endif


}

