#ifndef TCLPROBABILITYDISTRIBUTION_H
#define TCLPROBABILITYDISTRIBUTION_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#include <basics/Probability.h>
#include <modules/TclObject.h>
#include <framework/Interface.h>

namespace yav {

/** \class TclProbabilityDistribution TclProbabilityDistribution.cpp modules/TclProbabilityDistribution.h
    \ingroup Modules
    base class for Tcl interface of a probability density function : used for
    exchanging information between deformable models and images */
 class TclProbabilityDistribution : public TclObject {
   protected :
     /// the probability function pointed by the Tcl interface
     ProbabilityDensityFunction *probabilityDensityFunction;
   public :
   /** @name constructor */
   //@{
   /** creates a Tcl Interface of a ProbabilityDensityFunction object
       @param m module to create TCL object
       @param pdf a pointer to a ProbabilityDensityFunction object */
   TclProbabilityDistribution(TclModule *m,
		       ProbabilityDensityFunction *pdf) : TclObject(typeid(TclProbabilityDistribution),m),probabilityDensityFunction(pdf) {
     done(typeid(TclProbabilityDistribution));
   }
   /** creates a Tcl Interface of a ProbabilityDensityFunction object : called by an inherited class
       @param type the type of the inherited object that is  really created
       @param m module to create TCL object
       @param pdf a pointer to a ProbabilityDensityFunction object */
   TclProbabilityDistribution(const std::type_info& type, TclModule *m,
		       ProbabilityDensityFunction *pdf) : TclObject(type,m),probabilityDensityFunction(pdf) {
   }
  //@}

  /** @name destructor */
  //@{
  /// destroys this interface
   virtual ~TclProbabilityDistribution(){
   }
  //@}

   /// new TCL command callback
   static int list(TclModule *m, int argc, char **argv) {
     return TclObject::list(m, argc, argv);
   }
   /// new TCL command callback
   static int create(TclModule *m, int argc, char **argv) ;
   /// create Tcl command to create an object and to handle its option commands
   static void createCommands(TclModule *m);
   /// -destroy option
   int destroyCmd(int argc, char **argv) ;
   /// -probabilty option
   int probabilityValueCmd(int argc, char **argv);
   /// -range option
   int rangeCmd(int argc, char **argv);
    /// -histogram option
   int histogramCmd(int argc, char **argv) ;
    /// -histogram option
   int histogramValueCmd(int argc, char **argv) ;
   /// -statistics option
   int statisticsCmd(int argc, char **argv) ;
   /// -sameDistribution option
   int sameDistributionCmd(int argc, char **argv) ;
   /// -deleteDistribution option
   int deleteDistributionCmd(int argc, char **argv) ;
   /// -cast option
   int castCmd(int argc, char **argv);
   /// returns the rangeInformation associated with a TclRangeInformation
   ProbabilityDensityFunction *getProbabilityFunction() const {
     return probabilityDensityFunction; }
 };


/** \class TclGaussianDistribution TclProbabilityDistribution.cpp modules/TclProbabilityDistribution.h
    \ingroup Modules
    Tcl interface for a Gaussian distribution */
 class TclGaussianDistribution : public TclProbabilityDistribution {
 public:
 /** creates a Tcl Interface of a ProbabilityDensityFunction object
       @param m module to create TCL object
       @param pdf a pointer to a ProbabilityDensityFunction object */
   TclGaussianDistribution(TclModule *m,
		       GaussianDistribution *pdf) : TclProbabilityDistribution(typeid(TclGaussianDistribution),m,pdf) 
   {
     done(typeid(TclGaussianDistribution));
   }
    /// new TCL command callback
   static int create(TclModule *m, int argc, char **argv) ;
   /// -parameters option
   int parametersCmd(int argc, char **argv);
   /// create Tcl command to create an object and to handle its option commands
   static void createCommands(TclModule *m);
 };


/** \class TclGaussianDistribution TclProbabilityDistribution.cpp modules/TclProbabilityDistribution.h
    \ingroup Modules
    Tcl interface for a Gaussian distribution */
 class TclGaussianMixtureDistribution : public TclProbabilityDistribution {
 public:
 /** creates a Tcl Interface of a ProbabilityDensityFunction object
       @param m module to create TCL object
       @param pdf a pointer to a ProbabilityDensityFunction object */
   TclGaussianMixtureDistribution(TclModule *m,
		       GaussianMixtureDistribution *pdf) : TclProbabilityDistribution(typeid(TclGaussianMixtureDistribution),m,pdf) 
   {
     done(typeid(TclGaussianMixtureDistribution));
   }
    /// new TCL command callback
   static int create(TclModule *m, int argc, char **argv) ;
   /// -parameters option
   int parametersCmd(int argc, char **argv);
   /// create Tcl command to create an object and to handle its option commands
   static void createCommands(TclModule *m);
 };

/** \class TclHistogramDistribution TclProbabilityDistribution.cpp modules/TclProbabilityDistribution.h
    \ingroup Modules
    Tcl interface for a Gaussian distribution */
 class TclHistogramDistribution : public TclProbabilityDistribution {
 public:
 /** creates a Tcl Interface of a ProbabilityDensityFunction object
       @param m module to create TCL object
       @param pdf a pointer to a ProbabilityDensityFunction object */
   TclHistogramDistribution(TclModule *m,
		       HistogramDistribution *pdf) : TclProbabilityDistribution(typeid(TclHistogramDistribution),m,pdf) 
   {
     done(typeid(TclHistogramDistribution));
   }
    /// new TCL command callback
   static int create(TclModule *m, int argc, char **argv) ;
   /// -parameters option
   int parametersCmd(int argc, char **argv);
   /// create Tcl command to create an object and to handle its option commands
   static void createCommands(TclModule *m);
 };
}

#endif
