#ifndef TCLRANGEINFORMATION_H
#define TCLRANGEINFORMATION_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#pragma interface "modules/TclRangeInformation.h"
#include <framework/RangeInformation.h>
#include <modules/TclObject.h>
#include <framework/Interface.h>

namespace yav {

/** \class TclRangeInformation TclRangeInformation.h modules/TclRangeInformation.h
    \ingroup Modules
    base class for Tcl interface of a range Information object : used for
    exchanging information between deformable models and range datasets */
 class TclRangeInformation : public TclObject {
   protected :
     /// the rangeInformation pointed by the Tcl interface
     RangeInformation *rangeInformation;
   public :
   /** @name constructor */
   //@{
   /** creates a Tcl Interface of a RangeInformation object
       @param m module to create TCL object
       @param ri a pointer to a RangeInformation object */
   TclRangeInformation(TclModule *m,
		       RangeInformation *ri) : TclObject(typeid(TclRangeInformation),m),rangeInformation(ri) {
     done(typeid(TclRangeInformation));
   }
   /** creates a Tcl Interface of a RangeInformation object : called by an inherited class
       @param type the type of the inherited object that is  really created
       @param m module to create TCL object
       @param ri a pointer to a RangeInformation object */
   TclRangeInformation(const std::type_info& type, TclModule *m,
		       RangeInformation *ri) : TclObject(type,m),rangeInformation(ri) {
   }
  //@}

  /** @name destructor */
  //@{
  /// destroys this interface
   virtual ~TclRangeInformation(){}
  //@}

   /// new TCL command callback
   static int list(TclModule *m, int argc, char **argv) {
     return TclObject::list(m, argc, argv);
   }
   /// new TCL command callback
   static int create(TclModule *m, int argc, char **argv) {
     if (argc!=1)
       return COMMAND_SYNOPSIS; 
     TclRangeInformation *ri=new TclRangeInformation (m,0);
     m->setResult("%s%u", (char *)ri->getObjectName(), ri->TclObject::getRef());
     return TCL_OK;
   }
   /// -destroy option
   int destroyCmd(int argc, char **argv) {
     TclModule *m;
     unsigned int ref;
     
     if(argc != 0) return COMMAND_SYNOPSIS;
     m = getModule();
     ref = TclObject::getRef();
     delete this;
     m->setResult("%s%u", TclRangeInformation::getObjectName(), ref);
     return TCL_OK;
   }
   /// -name option
   int nameCmd(int argc, char **argv) {
     if(argc != 0) return COMMAND_SYNOPSIS;
     if (rangeInformation)
       getModule()->setResult(rangeInformation->getName());
     else 
       getModule()->setResult("none");
     return TCL_OK;
   }
   /// -cast option
   int castCmd(int argc, char **argv) {
     if(argc != 0) return COMMAND_SYNOPSIS;
     TclModule *m;
     m = getModule();
     TclRangeInformation *ri=new TclRangeInformation(m,rangeInformation);

     delete this;
     m->setResult(ri->name());
     return TCL_OK;
   }
   /// returns the rangeInformation associated with a TclRangeInformation
   RangeInformation *getRangeInformation() const {
     return rangeInformation; }
 };


  /** \class TclRangeInformationSet TclRangeInformation.h modules/TclRangeInformation.h
      \ingroup Modules
      base class for Tcl interface to  a set of range Information object  
      used for exchanging information between deformable models and 
      range datasets */
 class TclRangeInformationSet : public TclObject {
   protected :
     /// the list of rangeInformation pointer stored by the Tcl interface
     std::list<RangeInformation *> rangeInformationList;
   public :
   /** @name constructor */
   //@{
   /** creates a Tcl Interface of a RangeInformationSet object
       @param m module to create TCL object
       @param ril a list of RangeInformation object pointers*/
     TclRangeInformationSet(TclModule *m,
			    std::list<RangeInformation *> ril) : TclObject(typeid(TclRangeInformationSet),m),rangeInformationList(ril) {
     done(typeid(TclRangeInformationSet));
   }
   /** creates a Tcl Interface of a RangeInformationSet object
       @param m module to create TCL object */
     TclRangeInformationSet(TclModule *m) : TclObject(typeid(TclRangeInformationSet),m) {
     done(typeid(TclRangeInformationSet));
   }
   /** creates a Tcl Interface of a RangeInformation object : called by an inherited class
       @param type the type of the inherited object that is  really created
       @param m module to create TCL object
       @param ril a pointer to a RangeInformation object */
   TclRangeInformationSet(const std::type_info& type, TclModule *m,
			  std::list<RangeInformation *> ril) : TclObject(type,m),rangeInformationList(ril) {
   }
  //@}

  /** @name destructor */
  //@{
  /// destroys this interface
   virtual ~TclRangeInformationSet() {}
   //@}

   /// new TCL command callback
   static int list(TclModule *m, int argc, char **argv) {
     return TclObject::list(m, argc, argv);
   }
   /// new TCL command callback
   static int create(TclModule *m, int argc, char **argv) {
     if (argc!=1)
       return COMMAND_SYNOPSIS; 
     TclRangeInformationSet *ri=new TclRangeInformationSet (m);
     m->setResult("%s%u", ri->getObjectName(), ri->TclObject::getRef());
     return TCL_OK;
   }
   /// -destroy option
   int destroyCmd(int argc, char **argv) {
     TclModule *m;
     unsigned int ref;
     
     if(argc != 0) return COMMAND_SYNOPSIS;
     m = getModule();
     ref = TclObject::getRef();
     delete this;
     m->setResult("%s%u", TclRangeInformationSet::getObjectName(), ref);
     return TCL_OK;
   }
   /// -name option
   int nameCmd(int argc, char **argv) {
     if(argc != 0) return COMMAND_SYNOPSIS;
     if (rangeInformationList.size()>0)
       getModule()->setResult((*(rangeInformationList.begin()))->getName());
     else
       getModule()->setResult("none");
     return TCL_OK;
   }
   /// returns the rangeInformation associated with a TclRangeInformation
   std::list<RangeInformation *> getRangeInformationList() const {
     return rangeInformationList; }
 };


}

#endif
