#ifndef TCLSON_H
#define TCLSON_H

#pragma interface "modules/TclSon.h"
#include <tcl.h>
#include <basics/TclString.h>
#include <modules/TclObject.h>


namespace yav {


/** \class TclSon
    \ingroup Modules
    \brief Parent class for an object attached to a TCL window.
    When the window is destroyed, the object dies */
class TclSon {

 protected:
  TclModule *module;
  /// TCL window name
  TclString TCLwindow;

 private:
  /// TCL window delete procedure
  Tcl_CmdDeleteProc *deleteProc;
  /// TCL window delete client data
  ClientData deleteData;

  /** procedure called when the TCL window is deleted
      @param cd parameter to transmit to TCL delete procedure */
  static void deleted(ClientData cd);

 public:
  /** @name Constructor */
  //@{
  /** builds a TclSon
      @param m module this object is created in
      @param name associated TCL window */
  TclSon(TclModule *m,
	 const TclString& name = EMPTY_STRING);
  //@}

  /** @name destructor */
  //@{
  /// destroy associated TCL window
  virtual ~TclSon();
  //@}

  /** associates object to a new TCL window
      @param name TCL window name */
  void setTCLwindow(const TclString& name);

  /// returns associated window name
  const TclString& windowName(void) { return TCLwindow; }
};


} // namespace yav

#endif
