namespace eval completion {

}

# Find the longest common prefix
proc completion::commonPrefix { names } {
    set l 0
    set miss 0
    set prefix [lindex $names 0]
    set max [string length $prefix]

    while {!$miss} {
	if { $l == $max } {
	    set miss 1
	} else {
	    incr l
	    set new [string range $prefix 0 $l]
	    foreach f $names {
		if ![string match $new* $f] {
		    set miss 1
		    incr l -1
		    break
		}
	    }
	}
    }
    return [string range $prefix 0 $l]
}


# perform file name completion on last str word
proc completion::fileCompletion { str } {
    set l [string length $str]
    set c [string index $str [expr $l - 1]]

    if { [string compare $c " "] } {
	set i [string last " " $str]
	set bstr [string range $str 0 $i]
	set path [string range $str [expr $i + 1] $l]

	if { [string match ~* $path] } {           # ~username
	    regexp (~\[^/\]*)(/*) $path dir username eop
	    if [catch {file dirname $username} dir] {
		return $str
	    }
	    if { [string compare $eop ""] } {
		set dir [file dirname $path]
	    } else {
		set dir $path
	    }
	    set tail [file tail $path]
	} else {                                   # other path
	    set dir [file dirname $path]
	    set tail [file tail $path]
	}
	if { ![string compare $dir "."] } {
	    set cdir true
	} else {
	    set cdir false
	}

	# See what files are there
	if { $cdir } {
	    set files [glob -nocomplain $tail*]
	} else {
	    set files [glob -nocomplain $dir/$tail*]
	}

	set alone 1
	switch [llength [split $files]] {
	    0 {}                                       # no file match
	    1 {                                        # matched a single file
		if { $cdir } {
		    set path [file tail $files]
		} else {
		    if { [string compare $dir "/"] } {
			set path $dir/[file tail $files]
		    } else {
			set path /[file tail $files]
		    }
		}
	    }
	    default {                                  # matched several files
		set alone 0
		set path [commonPrefix $files]
	    }
	}

	if { [file isdirectory $path] } {
	    if { ![string compare $path "/"] } {
		set rstr "$bstr/"
	    } else {
		set rstr "$bstr$path/"
	    }
	} elseif { [file exists $path] && $alone } {
	    set rstr "$bstr$path "
	} else {
	    set rstr "$bstr$path"
	}

	return $rstr
    }

    return $str
}


# perform command name completion on first str word
proc completion::commandCompletion { str } {
    set cmds [info commands [lindex $str 0]*]

    if {[llength $cmds] == 0} {
	return "$str"
    } elseif {[llength $cmds] == 1} {
	return "$cmds "
    } else {
	return [commonPrefix $cmds]
    }
}


# perform command name completion if str has only one word or file name
# completion on last word otherwise
proc completion::completion { str } {
    switch [llength $str] {
	0 {
	    return {}
	}
	1 {
	    return [commandCompletion $str]
	}
	default {
	    return [fileCompletion $str]
	}
    }
}



namespace eval modules {

    set yavcrtcmd 0
    set yavncmds 0
    set yavcrt ""

    set parent .

    set helpFct ""

    set constraintBg #A0D0A0
    set helpConstraintBg lightyellow
    
}
#end of interface 

proc modules::interface { parent color } {
    set modules::parent $parent

    setDisplayFunction "modules::displayMessage"
    setEvalFunction "modules::evaluate"

    # main frame
    frame $parent.cmd -bg $color
    frame $parent.cmd.r -bg $color
    frame $parent.cmd.r.b -bg $color
    label $parent.cmd.lbl -text "Command:" -bg $color
    pack $parent.cmd.lbl -side left
    pack $parent.cmd.r -side right -fill both -expand true -padx 8 -pady 5
    
    entry $parent.cmd.r.txt -bg white -relief sunken -width 60 -textvariable compname
    #eliminate binding tag to do our own focus management
    bindtags $parent.cmd.r.txt [list $parent.cmd.r.txt [winfo class $parent.cmd.r.txt] .]
    # exit on Control-d or Control-c
    #	bind $parent.cmd.r.txt <Control-d> { exit }
    #	bind $parent.cmd.r.txt <Control-D> { exit }
    #	bind $parent.cmd.r.txt <Control-c> { exit }
    #	bind $parent.cmd.r.txt <Control-C> { exit }
    
    bind $parent.cmd.r.txt <Return> "modules::evaluate \[$parent.cmd.r.txt get\]"
    bind $parent.cmd.r.txt <Up> "modules::historyUp $parent"
    bind $parent.cmd.r.txt <Down> "modules::historyDown $parent"
    #paste management
    bind $parent.cmd.r.txt <ButtonRelease-2> {
	if { ![catch {set str [selection get]}] } {
	    set i [string length $str];
	    set c [string index $str [expr $i - 1]];
	    if { ![string compare $c "\n"] } { 
		modules::evaluate [string range $str 0 [expr $i - 2]];
		selection clear
	    }
	}
    }
    #focus management
    bind $parent.cmd.r.txt <Tab> "set str \[completion::completion \[$parent.cmd.r.txt get\]\]; $parent.cmd.r.txt delete 0 1000; $parent.cmd.r.txt insert 0 \$str"
    bind $parent <Enter> "focus $parent.cmd.r.txt"
    bind $parent.cmd <Tab> "focus $parent.cmd.r.txt"
    
    listbox $parent.cmd.r.b.list -height 3  -bg $color -yscrollcommand [list $parent.cmd.r.b.sy set] -xscrollcommand [list $parent.cmd.r.b.sx set] 
    bind $parent.cmd.r.b.list <ButtonPress-1> "set str \[$parent.cmd.r.b.list get @%x,%y\]; if { \[string compare str \"\" \] != 0 } { $parent.cmd.r.txt delete 0 1000; $parent.cmd.r.txt insert 0 \$str }"
    scrollbar $parent.cmd.r.b.sy -orient vertical -bg $color -command [list $parent.cmd.r.b.list yview]
    scrollbar $parent.cmd.r.b.sx -orient horizontal -bg $color -command [list $parent.cmd.r.b.list xview]
    pack $parent.cmd.r.b.sx -fill x -side bottom
    pack $parent.cmd.r.b.list -fill both -expand true -side left
    pack $parent.cmd.r.b.sy -fill y -side right
    
    pack $parent.cmd.r.b -fill both -expand true
    pack $parent.cmd.r.txt -fill x
    
    frame $parent.msg -bg $color
    frame $parent.msg.r -bg $color
    frame $parent.msg.r.b -bg $color
    label $parent.msg.lbl -text "Message: " -bg $color
    pack $parent.msg.lbl -side left
    pack $parent.msg.r -side right -fill both -expand true -padx 8 -pady 5
    
    label $parent.msg.r.txt -text "Ready" -relief sunken -width 60 -anchor w -bg gray
    listbox $parent.msg.r.b.list -height 5 -bg $color -yscrollcommand [list $parent.msg.r.b.sy set] -xscrollcommand [list $parent.msg.r.b.sx set]
    scrollbar $parent.msg.r.b.sy -orient vertical -bg $color -command [list $parent.msg.r.b.list yview]
    scrollbar $parent.msg.r.b.sx -orient horizontal -bg $color -command [list $parent.msg.r.b.list xview]
    pack $parent.msg.r.b.sx -fill x -side bottom
    pack $parent.msg.r.b.list -fill both -expand true -side left
    pack $parent.msg.r.b.sy -fill y -side right
    
    pack $parent.msg.r.b -fill both -expand true
    pack $parent.msg.r.txt -fill x
    bind $parent.msg <Tab> "focus $parent.cmd.r.txt"
    

    frame $parent.buttons -bg $color
    button $parent.buttons.quit -text "Exit" -command exit -bg $color
    button $parent.buttons.help -text "Help" -command "modules::help $color" -bg $color

    pack $parent.buttons.quit $parent.buttons.help -side left -padx 20 -pady 10 -expand true

    pack $parent.cmd $parent.msg -fill both -expand true
    pack $parent.buttons -fill x -expand true
    focus $parent.cmd.r.txt
    
}

proc modules::flushCommands { } {
    while { [$modules::parent.cmd.r.b.list size] > 0 } {
	$modules::parent.cmd.r.b.list delete 0
    }
    set modules::yavcrtcmd 0
    set modules::yavncmds 0
    set modules::yavcrt "" 
}


proc modules::flushResults { } {
    while { [$modules::parent.msg.r.b.list size] > 0 } {
	$modules::parent.msg.r.b.list delete 0
    }
    $modules::parent.msg.r.txt configure -text "Ready" 
}


proc modules::setPartialCommand { cmd } {
    $modules::parent.cmd.r.txt delete 0 1000
    $modules::parent.cmd.r.txt insert 0 $cmd
}


proc modules::setCommand { cmd } {
    $modules::parent.cmd.r.b.list insert $modules::yavncmds $cmd
    $modules::parent.cmd.r.b.list see end
    incr modules::yavncmds
    set modules::yavcrtcmd $modules::yavncmds
}

proc modules::setResult { res } {
    modules::displayMessage $res	
}

proc modules::removeObjectMenu { var def object} {
    set list [lindex [array get $var  $object] 1]
    set counter 0
    set index -1
    foreach entry $list {
	if [ expr [lsearch $entry $def] != -1 ] {
	    set index $counter
	}
	incr counter
    }
    if [expr $index >= 0 ] {
	set list [lreplace $list $index $index]
	if [string match $list {}] {
	    array set $var "$object {}"
	} else {
	    array set $var  "$object [ list $list]"
	}
    }

}

proc modules::addObjectMenu { var def object objectMenu} {
    set val "{$objectMenu $def } "
    set list [lindex [array get $var $object] 1]
    if ![ string match $list {}] {
	set list [concat  $list $val]
    } else {
	set list $val
    }
    array set $var "$object [list $list ]"
}

proc modules::historyUp { parent } {
    if { $modules::yavcrtcmd == $modules::yavncmds } {
	set modules::yavcrt [$parent.cmd.r.txt get]
    }
    if {$modules::yavcrtcmd > 0} { 
	set modules::yavcrtcmd [expr $modules::yavcrtcmd - 1]
	$parent.cmd.r.txt delete 0 1000
	$parent.cmd.r.txt insert 0 [$parent.cmd.r.b.list get $modules::yavcrtcmd]
    }
}


proc modules::historyDown { parent } {
    if {$modules::yavcrtcmd < $modules::yavncmds} {
	set modules::yavcrtcmd [expr $modules::yavcrtcmd + 1]
	$parent.cmd.r.txt delete 0 1000
	if {$modules::yavcrtcmd == $modules::yavncmds} {
	    $parent.cmd.r.txt insert 0 $modules::yavcrt
	} else {
	    $parent.cmd.r.txt insert 0 [$parent.cmd.r.b.list get $modules::yavcrtcmd]
	}
    }
}
proc modules::menusubstitute { str tag elt } {
    set index [string first $tag $str]
    if { $index == -1 } {
	return "eval {$str $elt}"
    } else {
	return "eval {[string range $str 0 [expr $index - 1]]$elt[string range $str [expr $index + [string length $tag]] end]}"
    }
}

proc modules::createForceConstraintRadioButtons { menu constraintList object} {
    modules::createConstraintRadioButtons "Add Force Constraints" $menu $constraintList force $object
}
proc modules::createPositionConstraintRadioButtons { menu constraintList object} {
    modules::createConstraintRadioButtons "Add Position Constraints" $menu $constraintList position $object
}
proc modules::createConstraintRadioButtons { title menu constraintList id object} {
    $menu add cascade -label $title  -menu   $menu.cascadeConstraint$id
    menu $menu.cascadeConstraint$id -tearoff 0 
    foreach {r} $constraintList {
	set label [lindex $r 0]
	set cmd "[lindex $r 1] $object"
	$menu.cascadeConstraint$id add command -label $label -command $cmd
    }
}
proc modules::editForceConstraint { menu cmd object creationList} {
    $menu add command -label "Force Constraint..." -command  "modules::editConstraint \"$cmd\" \"List of Force Constraint on $object\"  $object \"$creationList\" force"
}
proc modules::editPositionConstraint { menu cmd object creationList} {
    $menu add command -label "Position Constraint..." -command  "modules::editConstraint \"$cmd\" \"List of Force Constraint on $object \"  $object \"$creationList\" position"
}

proc modules::createConstraintFrame {root title constraint} {
    foreach w [winfo children $root.process] {
	destroy $w 
    }
    
    if ![winfo exists $root.process.label] {
	set bg $modules::constraintBg
	label $root.process.label  -text $title -bg $bg 
	pack $root.process.label -fill x
    }
    
    frame $root.process.f -bg $bg
    pack $root.process.f -expand true 
    frame $root.process.buttons -bg $bg
    pack $root.process.buttons -expand true -fill both
}
proc modules::showConstraintHelp {Root text } {
    if ![winfo exists $Root.help] {
	set bg $modules::helpConstraintBg
	set root $Root.help
	set wmx [winfo rootx $Root]
	set wmy [expr [winfo rooty $Root]+[winfo height $Root]]

	toplevel $root -bg $bg
	wm geometry $root  +$wmx+$wmy

	frame $root.f -bg $bg
	pack $root.f -expand true -fill both
	frame $root.buttons -bg $bg
	button $root.buttons.destroy -text "Done" -command "destroy $root" -bg $bg
	pack  $root.buttons.destroy -side left -expand true -fill both -padx 20 -pady 10
	pack $root.buttons -expand true -fill both

	pack [message $root.f.message -width 500  -bg $bg -text $text]

#	raise $root

    }
}
proc modules::addConstraintHelp {root text } {
    if ![winfo exists $root.process.buttons.help] {
	button $root.process.buttons.help -text "Help" -command "modules::showConstraintHelp $root \"$text\"" -bg $modules::constraintBg
	pack  $root.process.buttons.help -side right -expand true -fill both -padx 20 -pady 10
    }
}
proc modules::setTooltip {widget text} {
        if { $text != "" } {
                bind $widget <Any-Enter> [list after 1000 [list modules::showTooltip %W $text]]
                bind $widget <Any-Leave> {destroy %W.balloon}
        }
 }
 proc modules::showTooltip {w arg} {
       if {[eval winfo containing  [winfo pointerxy .]]!=$w} {return}
    set top $w.balloon
    catch {destroy $top}
    toplevel $top -bd 1 -bg black
    wm overrideredirect $top 1
    if {$::tcl_platform(platform) == "macintosh"} {
     unsupported1 style $top floating sideTitlebar
    }
    pack [message $top.txt -aspect 10000 -bg lightyellow \
            -justify left -fg black -text $arg]
    set wmx [winfo rootx $w]
    set wmy [expr [winfo rooty $w]+[winfo height $w]]
    wm geometry $top \
      [winfo reqwidth $top.txt]x[winfo reqheight $top.txt]+$wmx+$wmy
    raise $top
 }
proc modules::addConstraintEntry {root title cmd {width 10} } {
    set bg $modules::constraintBg
    set n 1
    while {[winfo exists $root.process.f.entry$n]} {
	incr n
    }

    frame $root.process.f.entry$n -bd 2 -bg $bg

    entry $root.process.f.entry$n.entry -relief raised -width $width -bg $bg
    bind $root.process.f.entry$n.entry <Return> "evaluate \"$cmd \[$root.process.f.entry$n.entry get\]\""

    $root.process.f.entry$n.entry delete 0 $width
    $root.process.f.entry$n.entry insert 0 [eval $cmd]
    label $root.process.f.entry$n.label -text $title -bg $bg
    pack $root.process.f.entry$n.entry -side right -padx 20 -pady 5
    pack $root.process.f.entry$n.label -side left -padx 20 -pady 5
    pack  $root.process.f.entry$n -side top -fill x
}
proc modules::addConstraintScale {root title cmd from to resolution {length 200} } {
    set bg $modules::constraintBg
    set n 1
    while {[winfo exists $root.process.f.scale$n]} {
	incr n
    }

#    frame $root.process.f.scale$n -bd 2 -bg $bg

    scale $root.process.f.scale$n -orient horizontal -label $title -from $from -to $to -length $length -resolution $resolution -bg $bg
    $root.process.f.scale$n set [eval $cmd]
    bind $root.process.f.scale$n <ButtonRelease-1> "evaluate \"$cmd \[$root.process.f.scale$n get\]\""
    
    pack $root.process.f.scale$n -side top -padx 20 -pady 10
}
proc modules::addConstraintMenuButton {root title menuRoutine constraint {width 20} {side top} } {
    set bg $modules::constraintBg
    set n 1
    while {[winfo exists $root.process.f.menuButton$n]} {
	incr n
    }
    menubutton $root.process.f.menuButton$n -text $title -width $width -menu $root.process.f.menuButton$n.m -bg $bg -relief raised
    set m [menu $root.process.f.menuButton$n.m -bg $bg ]
#    $menuRoutine $m $constraint $bg
    eval "$menuRoutine $m $constraint $bg"
    pack $root.process.f.menuButton$n -side $side -padx 20 -pady 10
}
proc modules::addConstraintOptionMenu {root title cmd option {width 20} {side top}} {
    global toto
    set bg $modules::constraintBg
    set n 1
    while {[winfo exists $root.process.f.optionMenu$n]} {
	incr n
    }
    menubutton $root.process.f.optionMenu$n -text $title -width $width -menu $root.process.f.optionMenu$n.m -bg $bg  -relief raised
    set m [menu $root.process.f.optionMenu$n.m -bg $bg -postcommand "set toto \[eval $cmd\] "]
    foreach e $option {
	$m add radio  -label "[lindex $e 0]" -variable toto -value "[lindex $e 1]" -command "evaluate \"$cmd [lindex $e 1]  \""
    }
    pack $root.process.f.optionMenu$n -side $side -padx 20 -pady 10
}
proc modules::addConstraintButton {root title cmd {width 20} {side top}} {
    set bg $modules::constraintBg
    set n 1
    while {[winfo exists $root.process.f.button$n]} {
	incr n
    }

    button $root.process.f.button$n -text $title -command "eval $cmd" -bg $bg -width $width
    
    pack $root.process.f.button$n -side $side -padx 20 -pady 10
}

proc modules::deleteConstraint { constraint} {
    global availableConstraintEditProcedure
    set name [$constraint -constraintName]
    foreach d $availableConstraintEditProcedure {
	if [string match $name [lindex $d 0] ] {
	    set kindname [lindex $d 2]
	    foreach w [which $kindname] {
		if [string match [$w -sameConstraint $constraint] "true"] {

		    evaluate "$w -deleteConstraint"
		}
	    }
	}
    }
    $constraint -destroy
}

proc modules::launchConstraintEdition { constraint object root} {
    global availableConstraintEditProcedure
    set name [$constraint -constraintName]
    foreach d $availableConstraintEditProcedure {
	if [string match $name [lindex $d 0] ] {
	    set kindname [lindex $d 2]
	    foreach w [which $kindname] {
		if [string match [$w -sameConstraint $constraint] "true"] {
		    eval "[lindex $d 1] $root $w $object"
		}
	    }
	}
    }
}

proc modules::listConstraint { root cmd object } {
    $root.f.list delete 0 end
    set cList [eval $cmd]
    if ![string match $cList {}] {
	set i 0
	foreach const $cList {
	    $root.f.list insert $i [$const -constraintName]
	    incr i
	}
    }
    bind $root.f.list <Double-1> "set val \[ lindex \[ $root.f.list curselection \] 0 \];if ![string match \$val {}] {set i 0;foreach const {$cList} { if \[expr \$i == \$val\] {modules::launchConstraintEdition \$const $object $root};incr i; }; }"
    return $cList
}

proc modules::editConstraint { cmd title object creationList id} {
    set root .{$id}Constraint$object
    global constraintList
    if {![winfo exists $root]} {
	set bg #A0D0A0
	set bg2  #80C080
	toplevel $root -bg $bg
	wm title $root $title 
	label $root.msg -justify center -text $title -bg $bg
	pack $root.msg  -side top
	
	frame $root.f -borderwidth .5c -bg $bg 
	pack $root.f -side top -expand yes -fill y -fill x
	
	scrollbar $root.f.scroll -command "$root.f.list yview" -bg $bg
	listbox $root.f.list -yscroll "$root.f.scroll set" -setgrid 1 -height 6 -bg $bg -width 30
	pack $root.f.scroll -side right -fill y
	pack $root.f.list -side left -expand true -fill both


	set constraintList [modules::listConstraint $root "$cmd" $object]

	frame $root.new -bg $bg
	pack $root.new -side bottom -pady 2m -expand 1
	button $root.new.add -text {Add...} -command "if ![string match $constraintList {}] {foreach const $constraintList {\$const -destroy}};modules::addConstraintDialog $root \"$creationList\" $object \"$cmd\";  " -bg lightGrey
	button $root.new.edit -text {Edit} -command "set val \[ lindex \[ $root.f.list curselection \] 0 \];if ![string match \$val {}] {set i 0;foreach const \$constraintList { if \[expr \$i == \$val\] {modules::launchConstraintEdition \$const $object $root};incr i; }; }"  -bg lightGrey
	button $root.new.remove -text {Remove} -command "foreach w \[winfo children $root.process\] {destroy \$w };set val \[ lindex \[ $root.f.list curselection \] 0 \];if ![string match \$val {}] {set i 0;foreach const \$constraintList { if \[expr \$i == \$val\] {modules::deleteConstraint \$const} else { \$const -destroy}; incr i}; 	set constraintList \[modules::listConstraint $root \"$cmd\" $object\] }"  -bg lightGrey
	pack  $root.new.add  $root.new.edit $root.new.remove -side left -expand 1 -padx 15

	frame $root.process -bg $bg -relief sunken -highlightthickness 4
	pack $root.process -side bottom -pady 2m -fill both -expand true

    }
}
proc modules::histoTopdfDialog { histo color} {
set root .pdffrom$histo
if {![winfo exists $root]} {
	toplevel $root -bg $color
	wm title $root "Create Probability Density Function from Histogram"
	frame $root.d -bg $color
	frame $root.b -bg $color
	
	labelframe $root.d.parzen -bg $color -text "Choose Parzen window"
	scale $root.d.parzen.sc -orient horizontal -width 10 -from 1 -to 15 -resolution 1 -bg $color
	pack $root.d.parzen.sc 
	pack $root.d.parzen -side left

	labelframe $root.d.quantile -bg $color -text "Choose Quantile to zero distribution"
	scale $root.d.quantile.sc -orient horizontal -width 10 -from 0 -to 1.0 -resolution 0.05 -bg $color
	pack $root.d.quantile.sc 
	pack $root.d.quantile
	
	button $root.b.app -text "Apply" -command "evaluate \" new pdf $histo \[$root.d.parzen.sc get \] \[$root.d.quantile.sc get \] \"; destroy $root" -bg $color
	button $root.b.dismiss -text "Dismiss"  -command "destroy $root" -bg $color
	
	pack  $root.b.app $root.b.dismiss -side left -padx 10 -pady 10

	pack $root.d $root.b -side top
}	

}
proc modules::addConstraintDialog { Root creationList object cmd} {
    global constraintList
    set root ${Root}Creation
    if {![winfo exists $root]} {

	set wmx [winfo rootx $Root]
	set wmy [expr [winfo rooty $Root]+[winfo height $Root]]



	set bg #A0D0A0
	set bg2  #80C080
	toplevel $root -bg $bg
	wm title $root "Constraint Creation"
	wm geometry $root  +$wmx+$wmy

	label $root.msg -justify center -text "Adding Constraints" -bg $bg
	pack $root.msg  -side top
	
	frame $root.f -borderwidth .5c -bg $bg 
	pack $root.f -side top -expand yes -fill y -fill x
	
	scrollbar $root.f.scroll -command "$root.f.list yview" -bg $bg
	listbox $root.f.list -yscroll "$root.f.scroll set" -setgrid 1 -height 6 -bg $bg -width 50
	pack $root.f.scroll -side right -fill y
	pack $root.f.list -side left -expand 1 -fill both


	if ![string match $creationList {}] {
	    set i 0
	    foreach {r} $creationList {
		set label [lindex $r 0]
		$root.f.list insert $i $label
		incr i
	    }
	}
	bind $root.f.list <Double-1>  "set val \[ lindex \[ $root.f.list curselection \] 0 \];if ![string match \$val {}] {set i 0;foreach {r} {$creationList} { if \[expr \$i == \$val\] {eval \"\[lindex \$r 1\] $object $root\";destroy $root;set constraintList  \[modules::listConstraint $Root \"$cmd\" $object\]; }; incr i };}"
	
	frame $root.process -bg $bg -relief sunken -highlightthickness 4
	pack $root.process -side bottom -pady 2m -fill both -expand true
	

	frame $root.button -bg $bg
	pack $root.button -side bottom -pady 2m
	button $root.button.dismiss -text Dismiss -command "destroy $root;set constraintList  \[modules::listConstraint $Root \"$cmd\" $object\]" -bg $bg
	pack  $root.button.dismiss -side bottom -expand 1
    }
    
}
proc modules::displayFCMProfile { fcm text1D clusterIndex} {
    set root .displayFCMProfile$fcm$clusterIndex
    if ![winfo exists $root] {
	set color white
	toplevel $root -bg $color
	wm title $root "Cluster #$clusterIndex center ($fcm)"
	frame $root.f -bg $color
	set x {}
	set y {}
	set counter 0
	foreach {i} [$fcm -clusterCenter $clusterIndex]   {
	    lappend x  $counter 
	    lappend y $i
	    set counter [ expr $counter +1 ]  
	}

	set graph $root.f.g
	#load BLT
	package require BLT

	blt::graph $graph -plotbackground $color 
	
	$graph element create $fcm -xdata $x -ydata $y  -linewidth 1 -scalesymbols yes -pixels 2 -symbol "" -label "Cluster #$clusterIndex center" -color [$text1D -color $clusterIndex]
	$graph axis configure y -hide false

	Blt_ZoomStack $graph
	Blt_ClosestPoint $graph
	pack  $graph -fill both 
	pack $root.f -side top

	frame $root.buttons -bg white

	button $root.buttons.destroy -text "Dismiss" -command "destroy $root" -bg white
	pack $root.buttons.destroy -side left -expand true -fill both -padx 20 -pady 10
	pack $root.buttons -expand true -fill both

    }
    
}

proc modules::displayEMProfile { em text1D modeIndex} {
    set root .displayEMProfile$em$modeIndex
    if ![winfo exists $root] {
	set color white
	toplevel $root -bg $color
	wm title $root "Mode #$modeIndex center+SDV ($em)"
	frame $root.f -bg $color
	set x {}
	set y {}
	set counter 0
	foreach {i} [$em -modeCenter $modeIndex]   {
	    lappend x  $counter 
	    lappend y $i
	    set counter [ expr $counter +1 ]
	}

	set graph $root.f.g
	#load BLT
	package require BLT

	blt::graph $graph -plotbackground $color 
	
	$graph element create $em -xdata $x -ydata $y  -linewidth 1 -scalesymbols yes -pixels 2 -symbol "" -label "Mode #$modeIndex center" -color [$text1D -color $modeIndex]
	$graph axis configure y -hide false

	set yplus {}
	set yminus {}
	set counter3 0

	foreach {i} [$em -modeStdDeviation $modeIndex]   {
		set mean [lindex $y $counter3]
		set vp [expr $mean + $i]
		lappend yplus [expr $mean + $i]
		lappend yminus [expr $mean - $i]
		incr counter3
	}

	$graph element create varPlus$em$modeIndex -xdata $x -ydata $yplus  -linewidth 1 -scalesymbols yes -pixels 2 -dashes {4 4} -symbol "" -label "Mode #$modeIndex SDV+" -color [$text1D -color $modeIndex]

	$graph element create varMinus$em$modeIndex -xdata $x -ydata $yminus  -linewidth 1 -scalesymbols yes -pixels 2 -dashes {4 4} -symbol "" -label "Mode #$modeIndex SDV-" -color [$text1D -color $modeIndex]

	Blt_ZoomStack $graph
	Blt_ClosestPoint $graph
	pack  $graph -fill both 
	pack $root.f -side top

	frame $root.buttons -bg white

	button $root.buttons.destroy -text "Dismiss" -command "destroy $root" -bg white
	pack $root.buttons.destroy -side left -expand true -fill both -padx 20 -pady 10
	pack $root.buttons -expand true -fill both

    }
    
}

proc modules::displayFCMInfo { fcm subobject text1D color } {
    set name .infofcm
    if ![winfo exists $name] {
	toplevel $name -bg $color

	frame ${name}.fv -bg \#A0D0A0
	set counter 0
	set counter2 [$fcm -nbrClusters]
	set counter3 2

	# Membership part
	label ${name}.fv.stat1title -text "Membership" -bg lightgrey -width 22
	grid ${name}.fv.stat1title -row 1 -column 1 -columnspan 2 -sticky ew

	foreach {i} [$fcm -membership $subobject] {
		set counter2 [ expr $counter2-1 ]
		set col [$text1D -color $counter2]
		label ${name}.fv.stat1$counter -text "Cluster #$counter2" -bg $col -width 10
		label ${name}.fv.stat1r$counter -bg lightgreen -width 12
		grid ${name}.fv.stat1$counter -column 1 -row $counter3 -sticky ew
		grid ${name}.fv.stat1r$counter -column 2 -row $counter3 -sticky ew
		set counter [ expr $counter+1 ]
		set counter3 [ expr $counter3+1 ]
	}

	# Data part
	label ${name}.fv.stat2title -text "Data" -bg lightgrey
	grid ${name}.fv.stat2title -row $counter3 -column 1 -columnspan 2 -sticky ew
	set counter 0
	set counter3 [ expr $counter3+1 ]

	foreach {i} [$fcm -data $subobject] {
		label ${name}.fv.stat2$counter -text "Dim #$counter" -bg $color
		label ${name}.fv.stat2r$counter -bg lightgreen
		grid ${name}.fv.stat2$counter -column 1 -row $counter3 -sticky ew
		grid ${name}.fv.stat2r$counter -column 2 -row $counter3 -sticky ew
		set counter [ expr $counter+1 ]
		set counter3 [ expr $counter3+1 ]
	}

	# Buttons part
	frame ${name}.fb -bg $color
	button ${name}.fb.refresh -text "Refresh" -command "modules::displayFCMInfo $fcm $subobject $text1D $color" -bg $color
	button ${name}.fb.dismiss -text "Dismiss" -command "evaluate \"sm1 -infofcm false\"" -bg $color
	pack ${name}.fb.refresh ${name}.fb.dismiss -side left -expand true -pady 5
	grid ${name}.fb -row 2 -column 1
    }

	# Refresh part
	wm title $name "$subobject"
	grid ${name}.fv -row 1 -column 1
	set counter 0

	foreach {i} [$fcm -membership $subobject] {
    		${name}.fv.stat1r$counter configure -text $i
		set counter [ expr $counter+1 ]
	}

	set counter 0

	foreach {i} [$fcm -data $subobject] {
    		${name}.fv.stat2r$counter configure -text $i
		set counter [ expr $counter+1 ]
	}
}

proc modules::displayEMInfo { em subobject text1D color } {
    set name .infoem
    if ![winfo exists $name] {
	toplevel $name -bg $color

	frame ${name}.fv -bg \#A0D0A0
	set counter 0
	set counter2 [$em -nbrModes]
	set counter3 2

	# Membership part
	label ${name}.fv.stat1title -text "Probability" -bg lightgrey -width 22
	grid ${name}.fv.stat1title -row 1 -column 1 -columnspan 2 -sticky ew

	foreach {i} [$em -proba $subobject] {
		set counter2 [ expr $counter2-1 ]
		set col [$text1D -color $counter2]
		label ${name}.fv.stat1$counter -text "Mode #$counter2" -bg $col -width 10
		label ${name}.fv.stat1r$counter -bg lightgreen -width 12
		grid ${name}.fv.stat1$counter -column 1 -row $counter3 -sticky ew
		grid ${name}.fv.stat1r$counter -column 2 -row $counter3 -sticky ew
		set counter [ expr $counter+1 ]
		set counter3 [ expr $counter3+1 ]
	}

	# Data part
	label ${name}.fv.stat2title -text "Data" -bg lightgrey
	grid ${name}.fv.stat2title -row $counter3 -column 1 -columnspan 2 -sticky ew
	set counter 0
	set counter3 [ expr $counter3+1 ]

	foreach {i} [$em -data $subobject] {
		label ${name}.fv.stat2$counter -text "Dim #$counter" -bg $color
		label ${name}.fv.stat2r$counter -bg lightgreen
		grid ${name}.fv.stat2$counter -column 1 -row $counter3 -sticky ew
		grid ${name}.fv.stat2r$counter -column 2 -row $counter3 -sticky ew
		set counter [ expr $counter+1 ]
		set counter3 [ expr $counter3+1 ]
	}

	# Buttons part
	frame ${name}.fb -bg $color
	button ${name}.fb.refresh -text "Refresh" -command "modules::displayEMInfo $em $subobject $text1D $color" -bg $color
	button ${name}.fb.dismiss -text "Dismiss" -command "evaluate \"sm1 -infoem false\"" -bg $color
	pack ${name}.fb.refresh ${name}.fb.dismiss -side left -expand true -pady 5
	grid ${name}.fb -row 2 -column 1
    }

	# Refresh part
	wm title $name "$subobject"
	grid ${name}.fv -row 1 -column 1
	set counter 0

	foreach {i} [$em -proba $subobject] {
    		${name}.fv.stat1r$counter configure -text $i
		set counter [ expr $counter+1 ]
	}

	set counter 0

	foreach {i} [$em -data $subobject] {
    		${name}.fv.stat2r$counter configure -text $i
		set counter [ expr $counter+1 ]
	}
}

proc modules::displayAllFCMProfiles { fcm text1D } {
    set root .displayAllFCMProfile$fcm
    if ![winfo exists $root] {
	set color white
	toplevel $root -bg $color
	wm title $root "Clusters centers ($fcm)"
	frame $root.f -bg $color

	set graph $root.f.g
	#load BLT
	package require BLT

	blt::graph $graph -plotbackground $color 

	set x {}
	set counter 0
	set clusterIndex 0
	foreach {i} [$fcm -clusterCenter $clusterIndex]   {
	    lappend x  $counter 
	    set counter [ expr $counter +1 ]  
	}

	set counter2 [$fcm -nbrClusters]

	for {set clusterIndex 0} {$clusterIndex< [$fcm -nbrClusters]} {incr clusterIndex} {
		set y {}
		set counter2 [ expr $counter2-1 ]
		foreach {i} [$fcm -clusterCenter $counter2]   {
	    		lappend y $i
		}
		$graph element create $fcm$clusterIndex -xdata $x -ydata $y  -linewidth 1 -scalesymbols yes -pixels 2 -symbol "" -label "Cluster #$counter2 center" -color [$text1D -color $counter2]
	}
	$graph axis configure y -hide false

	Blt_ZoomStack $graph
	Blt_ClosestPoint $graph
	pack  $graph -fill both 
	pack $root.f -side top

	frame $root.buttons -bg white

	button $root.buttons.destroy -text "Dismiss" -command "destroy $root" -bg white
	pack $root.buttons.destroy -side left -expand true -fill both -padx 20 -pady 10
	pack $root.buttons -expand true -fill both
    }
}

proc modules::displayAllEMProfiles { em text1D {showVariance 0}} {
    set root .displayAllEMProfile$em$showVariance
    if ![winfo exists $root] {
	set color white
	toplevel $root -bg $color

	if $showVariance {
		wm title $root "Modes centers+SDV ($em)"
	} else {
		wm title $root "Modes centers ($em)"
	}

	frame $root.f -bg $color

	set graph $root.f.g
	#load BLT
	package require BLT

	if $showVariance {
		blt::graph $graph -plotbackground $color -width 10i -height 7i
	} else {
		blt::graph $graph -plotbackground $color
	}

	set x {}
	set counter 0
	set modeIndex 0
	foreach {i} [$em -modeCenter $modeIndex]   {
	    lappend x  $counter 
	    set counter [ expr $counter +1 ]  
	}

	set counter2 [$em -nbrModes]

	for {set modeIndex 0} {$modeIndex< [$em -nbrModes]} {incr modeIndex} {
		set y {}
		set counter2 [ expr $counter2-1 ]
		foreach {i} [$em -modeCenter $counter2]   {
	    		lappend y $i
		}

		$graph element create $em$modeIndex -xdata $x -ydata $y  -linewidth 1 -scalesymbols yes -pixels 2 -symbol "" -label "Mode #$counter2 center" -color [$text1D -color $counter2]

		if $showVariance {
			set yplus {}
			set yminus {}
			set counter3 0

			foreach {i} [$em -modeStdDeviation $counter2]   {
				set mean [lindex $y $counter3]
				set vp [expr $mean + $i]
	    			lappend yplus [expr $mean + $i]
				lappend yminus [expr $mean - $i]
				incr counter3
			}

			$graph element create varPlus$em$modeIndex -xdata $x -ydata $yplus  -linewidth 1 -scalesymbols yes -pixels 2 -dashes {4 4} -symbol "" -label "Mode #$counter2 SDV+" -color [$text1D -color $counter2]

			$graph element create varMinus$em$modeIndex -xdata $x -ydata $yminus  -linewidth 1 -scalesymbols yes -pixels 2 -dashes {4 4} -symbol "" -label "Mode #$counter2 SDV-" -color [$text1D -color $counter2]
		}
	}
	$graph axis configure y -hide false

	Blt_ZoomStack $graph
	Blt_ClosestPoint $graph
	pack  $graph -fill both 
	pack $root.f -side top

	frame $root.buttons -bg white

	button $root.buttons.destroy -text "Dismiss" -command "destroy $root" -bg white
	pack $root.buttons.destroy -side left -expand true -fill both -padx 20 -pady 10
	pack $root.buttons -expand true -fill both
    }
}

proc modules::displayFCMStatistics { fcm } {
    set root .displayFCMStatistics$fcm
    if ![winfo exists $root] {
	toplevel $root -bg lightgreen
	wm title $root "About $fcm"

	# Statistics part
	frame $root.stat1
	label $root.stat1.stat11 -text "Number of clusters :" -anchor w -width 28 -bg lightgrey
	label $root.stat1.stat12 -text "[$fcm -nbrClusters]" -anchor w -width 10 -bg lightgrey
	pack  $root.stat1.stat11 -side left
	pack  $root.stat1.stat12 -side left
	pack  $root.stat1 -expand true -side top 
	frame $root.stat2
	label $root.stat2.stat21 -text "Number of data :" -anchor w -width 28 -bg lightgreen
	label $root.stat2.stat22 -text "[$fcm -nbrData]" -anchor w -width 10 -bg lightgreen
	pack  $root.stat2.stat21 -side left
	pack  $root.stat2.stat22 -side left
	pack  $root.stat2 -expand true -side top
	frame $root.stat3
	label $root.stat3.stat31 -text "Number of dimensions :" -anchor w -width 28 -bg lightgrey
	label $root.stat3.stat32 -text "[$fcm -nbrCoord]" -anchor w -width 10 -bg lightgrey
	pack  $root.stat3.stat31 -side left
	pack  $root.stat3.stat32 -side left
	pack  $root.stat3 -expand true -side top
	frame $root.stat4
	label $root.stat4.stat41 -text "Fuzzy exponent :" -anchor w -width 28 -bg lightgreen
	label $root.stat4.stat42 -text "[$fcm -expo]" -anchor w -width 10 -bg lightgreen
	pack  $root.stat4.stat41 -side left
	pack  $root.stat4.stat42 -side left
	pack  $root.stat4 -expand true -side top
	frame $root.stat5
	label $root.stat5.stat51 -text "Max number of iterations :" -anchor w -width 28 -bg lightgrey
	label $root.stat5.stat52 -text "[$fcm -maxIter]" -anchor w -width 10 -bg lightgrey
	pack  $root.stat5.stat51 -side left
	pack  $root.stat5.stat52 -side left
	pack  $root.stat5 -expand true -side top
	frame $root.stat6
	label $root.stat6.stat61 -text "Used number of iterations :" -anchor w -width 28 -bg lightgreen
	label $root.stat6.stat62 -text "[$fcm -nbrIter]" -anchor w -width 10 -bg lightgreen
	pack  $root.stat6.stat61 -side left
	pack  $root.stat6.stat62 -side left
	pack  $root.stat6 -expand true -side top
	frame $root.stat7
	label $root.stat7.stat71 -text "Min amount of improvement :" -anchor w -width 28 -bg lightgrey
	label $root.stat7.stat72 -text "[$fcm -minImpro]" -anchor w -width 10 -bg lightgrey
	pack  $root.stat7.stat71 -side left
	pack  $root.stat7.stat72 -side left
	pack  $root.stat7 -expand true -side top
	frame $root.stat8
	label $root.stat8.stat81 -text "Partition coefficient (Vpc) :" -anchor w -width 28 -bg lightgreen
	label $root.stat8.stat82 -text "[$fcm -vpc]" -anchor w -width 10 -bg lightgreen
	pack  $root.stat8.stat81 -side left
	pack  $root.stat8.stat82 -side left
	pack  $root.stat8 -expand true -side top
	frame $root.stat9
	label $root.stat9.stat91 -text "Partition entropy (Vpe) :" -anchor w -width 28 -bg lightgrey
	label $root.stat9.stat92 -text "[$fcm -vpe]" -anchor w -width 10 -bg lightgrey
	pack  $root.stat9.stat91 -side left
	pack  $root.stat9.stat92 -side left
	pack  $root.stat9 -expand true -side top
	frame $root.stat10
	label $root.stat10.stat101 -text "Overlap :" -anchor w -width 28 -bg lightgreen
	label $root.stat10.stat102 -text "[$fcm -overlap]" -anchor w -width 10 -bg lightgreen
	pack  $root.stat10.stat101 -side left
	pack  $root.stat10.stat102 -side left
	pack  $root.stat10 -expand true -side top
	frame $root.stat11
	label $root.stat11.stat111 -text "Separation :" -anchor w -width 28 -bg lightgrey
	label $root.stat11.stat112 -text "[$fcm -separation]" -anchor w -width 10 -bg lightgrey
	pack  $root.stat11.stat111 -side left
	pack  $root.stat11.stat112 -side left
	pack  $root.stat11 -expand true -side top
	frame $root.stat12
	label $root.stat12.stat122 -text "[$fcm -metric]" -width 38 -bg #90C090
	pack  $root.stat12.stat122 -side left
	pack  $root.stat12 -expand true -side top

	# Exit button part
	frame $root.fexit
	button $root.fexit.butexit -text "Dismiss" -command "destroy $root" -bg lightgreen
	pack $root.fexit.butexit -side bottom
	pack $root.fexit  -expand true -side bottom
    }
}

proc modules::displayEMStatistics { em } {
    set root .displayEMStatistics$em
    if ![winfo exists $root] {
	toplevel $root -bg lightgreen
	wm title $root "About $em"

	# Statistics part
	frame $root.stat1
	label $root.stat1.stat11 -text "Number of modes :" -anchor w -width 28 -bg lightgrey
	label $root.stat1.stat12 -text "[$em -nbrModes]" -anchor w -width 10 -bg lightgrey
	pack  $root.stat1.stat11 -side left
	pack  $root.stat1.stat12 -side left
	pack  $root.stat1 -expand true -side top 
	frame $root.stat2
	label $root.stat2.stat21 -text "Number of data :" -anchor w -width 28 -bg lightgreen
	label $root.stat2.stat22 -text "[$em -nbrData]" -anchor w -width 10 -bg lightgreen
	pack  $root.stat2.stat21 -side left
	pack  $root.stat2.stat22 -side left
	pack  $root.stat2 -expand true -side top
	frame $root.stat3
	label $root.stat3.stat31 -text "Number of dimensions :" -anchor w -width 28 -bg lightgrey
	label $root.stat3.stat32 -text "[$em -nbrCoord]" -anchor w -width 10 -bg lightgrey
	pack  $root.stat3.stat31 -side left
	pack  $root.stat3.stat32 -side left
	pack  $root.stat3 -expand true -side top
	frame $root.stat5
	label $root.stat5.stat51 -text "Max number of iterations :" -anchor w -width 28 -bg lightgrey
	label $root.stat5.stat52 -text "[$em -maxIter]" -anchor w -width 10 -bg lightgrey
	pack  $root.stat5.stat51 -side left
	pack  $root.stat5.stat52 -side left
	pack  $root.stat5 -expand true -side top
	frame $root.stat6
	label $root.stat6.stat61 -text "Used number of iterations :" -anchor w -width 28 -bg lightgreen
	label $root.stat6.stat62 -text "[$em -nbrIter]" -anchor w -width 10 -bg lightgreen
	pack  $root.stat6.stat61 -side left
	pack  $root.stat6.stat62 -side left
	pack  $root.stat6 -expand true -side top
	frame $root.stat7
	label $root.stat7.stat71 -text "Min amount of improvement :" -anchor w -width 28 -bg lightgrey
	label $root.stat7.stat72 -text "[$em -minImpro]" -anchor w -width 10 -bg lightgrey
	pack  $root.stat7.stat71 -side left
	pack  $root.stat7.stat72 -side left
	pack  $root.stat7 -expand true -side top
	frame $root.stat8
	label $root.stat8.stat81 -text "Partition coefficient (Vpc) :" -anchor w -width 28 -bg lightgreen
	label $root.stat8.stat82 -text "[$em -vpc]" -anchor w -width 10 -bg lightgreen
	pack  $root.stat8.stat81 -side left
	pack  $root.stat8.stat82 -side left
	pack  $root.stat8 -expand true -side top
	frame $root.stat9
	label $root.stat9.stat91 -text "Partition entropy (Vpe) :" -anchor w -width 28 -bg lightgrey
	label $root.stat9.stat92 -text "[$em -vpe]" -anchor w -width 10 -bg lightgrey
	pack  $root.stat9.stat91 -side left
	pack  $root.stat9.stat92 -side left
	pack  $root.stat9 -expand true -side top
	frame $root.stat10
	label $root.stat10.stat101 -text "Overlap :" -anchor w -width 28 -bg lightgreen
	label $root.stat10.stat102 -text "[$em -overlap]" -anchor w -width 10 -bg lightgreen
	pack  $root.stat10.stat101 -side left
	pack  $root.stat10.stat102 -side left
	pack  $root.stat10 -expand true -side top
	frame $root.stat11
	label $root.stat11.stat111 -text "Separation :" -anchor w -width 28 -bg lightgrey
	label $root.stat11.stat112 -text "[$em -separation]" -anchor w -width 10 -bg lightgrey
	pack  $root.stat11.stat111 -side left
	pack  $root.stat11.stat112 -side left
	pack  $root.stat11 -expand true -side top

	# Exit button part
	frame $root.fexit
	button $root.fexit.butexit -text "Dismiss" -command "destroy $root" -bg lightgreen
	pack $root.fexit.butexit -side bottom
	pack $root.fexit  -expand true -side bottom
    }
}

proc modules::addMenuRadioButtons { title menu radioList cmd id} {
    global val$id
    set val$id ""
    $menu add cascade -label $title  -menu   $menu.cascade$id
    menu $menu.cascade$id -tearoff 0 -postcommand "set val$id  \[$cmd\] "
    foreach r $radioList {
	set label [lindex $r 0]
	set value [lindex $r 1]
	$menu.cascade$id add radio -label $label -variable val$id -value $value -command "evaluate \"$cmd $value \""
    }
}

proc modules::scaleListDialog  { root color scaleList}  {
    if ![winfo exists $root] {
	toplevel $root  -bg $color
	frame $root.f -bg $color
	set counter 1
	foreach s $scaleList {
	    set label [lindex $s 0]
	    set from [lindex $s 1]
	    set to [lindex $s 2]
	    set initVal [lindex $s 3]
	    set cmd [lindex $s 4]
	    set resolution [lindex $s 5]
	    scale $root.f.sc$counter -orient horizontal -label $label -from $from -to $to -resolution $resolution -bg $color -length 200 
	    bind $root.f.sc$counter <ButtonRelease-1> "evaluate \"$cmd \[$root.f.sc$counter get\]\""
	    pack $root.f.sc$counter
	    $root.f.sc$counter  set $initVal
	    incr counter
	}
	frame $root.buttons -bg $color
	button $root.buttons.destroy -text "Dismiss" -command "destroy $root" -bg $color
	pack $root.buttons.destroy -side left -expand true -fill both -padx 20 -pady 10
	pack $root.f $root.buttons -expand true -fill both
    }
}

proc modules::scaleDialog  { root color title from to initVal cmd {resolution 1.0} {fg black} } {
    if ![winfo exists $root] {
	toplevel $root  -bg $color
	frame $root.f -bg $color
	label $root.f.l -text $title -bg $color
	scale $root.f.s -from $from -to $to -resolution $resolution -orient horizontal -bg $color -fg $fg
	$root.f.s set $initVal
	pack $root.f.l $root.f.s -side left 
	frame $root.buttons -bg $color
	button $root.buttons.apply -text "Apply" -command " eval \[modules::menusubstitute {$cmd} ...  \[$root.f.s get \] \] " -bg $color
	button $root.buttons.destroy -text "Destroy" -command "destroy $root" -bg $color
	pack $root.buttons.apply $root.buttons.destroy -side left -expand true -fill both -padx 20 -pady 10
	pack $root.f $root.buttons -expand true -fill both
    }
}
proc modules::entryDialog  { root color title  initVal cmd  {fg black} } {
    if ![winfo exists $root] {
	toplevel $root  -bg $color
	frame $root.f -bg $color
	label $root.f.l -text $title -bg $color
	entry $root.f.e  -bg $color -fg $fg -relief sunken
	$root.f.e insert 0  $initVal
	pack $root.f.l $root.f.e -side left 
	frame $root.buttons -bg $color
	button $root.buttons.apply -text Apply -command " eval \[modules::menusubstitute {$cmd} ...  {\[$root.f.e get \]} \] " -bg $color
	button $root.buttons.destroy -text "Destroy" -command "destroy $root" -bg $color
	pack $root.buttons.apply $root.buttons.destroy -side left -expand true -fill both -padx 20 -pady 10
	pack $root.f $root.buttons -expand true -fill both
    }
}

proc modules::displayString { str color } {
    $modules::parent.msg.r.b.list insert [$modules::parent.msg.r.b.list size] [$modules::parent.msg.r.txt cget -text]
    $modules::parent.msg.r.b.list see end
    $modules::parent.msg.r.txt configure -text $str -fg $color
}


proc modules::displayMessage { str } {
    modules::displayString $str black
}


proc modules::evaluate { commandStr } {
    # evaluate command at toplevel
    set res [uplevel \#0 $commandStr]
    
    if [winfo exists $modules::parent] {

    # set result
    modules::displayMessage $res
    $modules::parent.cmd.r.txt delete 0 1000
    
    $modules::parent.cmd.r.b.list insert $modules::yavncmds $commandStr
    $modules::parent.cmd.r.b.list see end
    incr modules::yavncmds
    set modules::yavcrtcmd $modules::yavncmds
    }
    return $res
}



proc modules::displayHelp { msg } {
    set l 1

    while { [string length $msg] > 75 } {
	set i 75
	while { $i > 0 && [string compare [string index $msg $i] "|"] &&
		[string compare [string index $msg $i] " "] } {
	    incr i -1
	}
	if { $i == 0 } {
	    set i 75
	    while { $i < [string length $msg] &&
		    [string compare [string index $msg $i] "|"] &&
		    [string compare [string index $msg $i] " "] } { 
		incr i
	    }		
	}
	if { $l > 1 } {
	    .help.txt.l insert end "   [string range $msg 0 $i]"
	} else {
	    .help.txt.l insert end [string range $msg 0 $i]
	}
	set msg [string range $msg [expr $i + 1] end]
	incr l
    }
    if { $l > 1 } {
	.help.txt.l insert end "   $msg"
    } else {
	.help.txt.l insert end $msg
    }
}

proc modules::help { color } {
    if ![winfo exists .help] {
	toplevel .help -bg $color
	wm title .help "Help me!"

	frame .help.lists -bg $color

	frame .help.lists.modules -bg $color
	label .help.lists.modules.t -text "Modules" -relief raised -bg $color
	frame .help.lists.modules.l -bg $color
	listbox .help.lists.modules.l.l -height 8 -bg white -yscrollcommand [list .help.lists.modules.l.sy set]
	scrollbar .help.lists.modules.l.sy -orient vertical -bg $color -command [list .help.lists.modules.l.l yview]
	pack .help.lists.modules.l.l -fill both -expand true -side left
	pack .help.lists.modules.l.sy -fill y -side right
	pack .help.lists.modules.t -fill x -expand true
	pack .help.lists.modules.l -fill both -expand true
	foreach module [module -list] {
	    .help.lists.modules.l.l insert end $module
	}
	bind .help.lists.modules.l.l <1> ".help.lists.functions.l.l delete 0 end; .help.lists.options.l.l delete 0 end; .help.txt.l delete 0 end; foreach fct \[functions \[.help.lists.modules.l.l get @%x,%y\]\] {.help.lists.functions.l.l insert end \$fct}"

	frame .help.lists.functions -bg $color
	label .help.lists.functions.t -text "Functions" -relief raised -bg $color
	frame .help.lists.functions.l -bg $color
	listbox .help.lists.functions.l.l -height 8 -bg white -yscrollcommand [list .help.lists.functions.l.sy set]
	scrollbar .help.lists.functions.l.sy -orient vertical -bg $color -command [list .help.lists.functions.l.l yview]
	pack .help.lists.functions.l.l -fill both -expand true -side left
	pack .help.lists.functions.l.sy -fill y -side right
	pack .help.lists.functions.t -fill x -expand true
	pack .help.lists.functions.l -fill both -expand true
	bind .help.lists.functions.l.l <1> ".help.lists.options.l.l delete 0 end; .help.txt.l delete 0 end; set modules::helpFct \[.help.lists.functions.l.l get @%x,%y\]; modules::displayHelp \[help \$modules::helpFct synopsis\]; modules::displayHelp \[help \$modules::helpFct description\]; modules::displayHelp \[help \$modules::helpFct creation\]; .help.txt.l yview end; modules::displayHelp {}; foreach opt \[options \$modules::helpFct\] {.help.lists.options.l.l insert end \$opt}"

	frame .help.lists.options -bg $color
	label .help.lists.options.t -text "Options" -relief raised -bg $color
	frame .help.lists.options.l -bg $color
	listbox .help.lists.options.l.l -height 8 -bg white -yscrollcommand [list .help.lists.options.l.sy set]
	scrollbar .help.lists.options.l.sy -orient vertical -bg $color -command [list .help.lists.options.l.l yview]
	pack .help.lists.options.l.l -fill both -expand true -side left
	pack .help.lists.options.l.sy -fill y -side right
	pack .help.lists.options.t -fill x -expand true
	pack .help.lists.options.l -fill both -expand true
	bind .help.lists.options.l.l <1> "modules::displayHelp \[help \$modules::helpFct \[.help.lists.options.l.l get @%x,%y\] synopsis\]; modules::displayHelp \[help \$modules::helpFct \[.help.lists.options.l.l get @%x,%y\] description\]; .help.txt.l yview end; modules::displayHelp {}"

	pack .help.lists.modules .help.lists.functions .help.lists.options -fill both -expand true -padx 10 -pady 10 -side left

	frame .help.txt -bg $color
	listbox .help.txt.l -height 8 -width 80 -bg white -yscrollcommand [list .help.txt.sy set]
	scrollbar .help.txt.sy -orient vertical -bg $color -command [list .help.txt.l yview]
	pack .help.txt.l -fill both -expand true -side left
	pack .help.txt.sy -fill y -side right

	frame .help.bt -bg $color
	button .help.bt.quit -text "Dismiss" -command "destroy .help" -bg $color
	pack .help.bt.quit -pady 10 

	pack .help.lists .help.txt -fill both -padx 10 -pady 10 -expand true
	pack .help.bt -fill x -expand true
    }
}


proc modules::dumpResultsToFile { } {
    set fname [ tk_getSaveFile -filetypes {{Results {.dump}} {All *}} -parent . -title {Select an image} ]
    if { [string compare $fname ""] } {
	set fres [open $fname w]
	for {set i 1} {$i < [$modules::parent.msg.r.b.list size]} {incr i} {
	    puts $fres [$modules::parent.msg.r.b.list get $i]
	}
	puts $fres [$modules::parent.msg.r.txt cget -text]
	close $fres
    }
}

proc modules::dumpCommandsToFile { } {
    set fname [ tk_getSaveFile -filetypes {{Results {.tcl}} {All *}} -parent . -title {Select an image} ]
    if { [string compare $fname ""] } {
	set fres [open $fname w]
	for {set i 0} {$i < [$modules::parent.cmd.r.b.list size]} {incr i} {
	    puts $fres [$modules::parent.cmd.r.b.list get $i]
	}
	puts $fres "mainLoop\n"
	close $fres
    }
}
proc modules::rgbToColor {str } {
    scan $str {%f %f %f} r g b
    return [ format "#%02x%02x%02x" [expr round($r*255)] [expr round($g*255)] [expr round($b*255)] ]
}

proc modules::selectColor {str}   {
    set color [tk_chooseColor -title "Choose a color"  -initialcolor [modules::rgbToColor $str] ]
    if ![string compare color " "] {
	return[ modules::colorToRgb str]
    } else {
	return [modules::colorToRgb $color]
    }
}
proc modules::selectColorStr {str}   {
    set color [tk_chooseColor -title "Choose a color" -initialcolor  $str]
    if ![string compare color " "] {
	return[ modules::colorToRgb str]
    } else {
	return [modules::colorToRgb $color]
    }
}

proc modules::createPDFGraph { pdf graph } {
	scan [$pdf -range] "%f %f %f" minVal maxVal nbBuckets
    set xval $minVal
    set x {}
    set y {}
    set nb $nbBuckets
    set s  [ expr ( $maxVal -$minVal)/($nbBuckets -1) ]
    for { set i 0} { $i < $nb} { incr i} {
	lappend x $xval
	lappend y [ $pdf -histogramValue $xval ]
	set xval [ expr  $xval + $s ]
    }
    
    if {![$graph element exists $pdf]} {
	$graph element create $pdf -xdata $x -ydata $y -linewidth 1 -scalesymbols yes -pixels 2 -color red -symbol ""
    } else {
	$graph element configure $pdf  -xdata $x -ydata $y
    }
}
proc modules::createHistogramGraph { histo graph minVal nbBuckets} {
    set xval $minVal
    set x {}
    set y {}
    set nb $nbBuckets
    set s  [ $histo -bucket size]
    for { set i 0} { $i < $nb} { incr i} {
	lappend x $xval
	lappend y [ $histo -value $xval ]
	set xval [ expr  $xval + $s ] 
    }
    if {![$graph element exists $histo]} {
	$graph element create $histo -xdata $x -ydata $y -linewidth 1 -scalesymbols yes -pixels 2 -symbol ""
    } else {
	$graph element configure $histo  -xdata $x -ydata $y
    }
    # affichage des coordonnees
    if { [winfo exists .$histo.position.p]} {
	bind $graph <Motion> ".$histo.position.p configure -text \"\[format \"%%.2f\" \[$graph axis invtransform x \[lindex \[split \[lindex \[split \[$graph crosshairs cget -position\] ,\] 0\] @\] 1\]\] \] \[format \"%%.2f\" \[$graph axis invtransform y \[lindex \[split \[$graph crosshairs cget -position\] ,\] 1\]\]\]\""
    }
    # rajoute les barres verticales et horizontales du curseur
    Blt_Crosshairs $graph
}
proc modules::drawHistogramInFrame { frame histo  {color "#E8E0D2"}} {
    set graph $frame.g
    #load BLT
    package require BLT
    
    blt::graph $frame.g -plotbackground white

    modules::createHistogramGraph $histo $graph [ $histo -minValue ] [ $histo -bucket number]
    Blt_ZoomStack $graph
    Blt_ClosestPoint $graph
    
    scrollbar $frame.xbar -command " $graph axis view x " -orient horizontal
    scrollbar $frame.ybar -command " $graph axis view y " -orient vertical
    $graph axis configure x -scrollcommand " $frame.xbar set " -logscale no -loose no
    $graph axis configure y -scrollcommand " $frame.ybar set " -logscale no 
    
    pack  $frame.ybar -fill y -side left
    pack  $frame.g -fill both 
    pack  $frame.xbar -fill x

}
proc modules::plotHistogram { histo  {color "#E8E0D2"} {subcolor "#D8D0C2"}} {
    set root ".${histo}"
    if {![winfo exists $root]} {

	toplevel $root -bg $color
	wm title $root "Plot of Histogram $histo"

	frame $root.position -bg $color
	label $root.position.l -text "Position:"  -bg $color
	label $root.position.p -text "" -bg $color

	frame $root.menu -bg $color
	menubutton $root.menu.file -text "File" -menu $root.menu.file.m -anchor w -bg $color
	set mfile [ menu  $root.menu.file.m -bg $color]
	$mfile add command -label "Save as Postcript.." -command "$root.f.g postscript output \[tk_getSaveFile -filetypes {{\"Postcript File\" {*.ps}}  {\"All\" {*} }  } -parent . -title {Select a Postcript File}\] -maxpect yes -decorations no"
	$mfile add command -label "Save as PPM File" -command "set im \[ image create photo\]; blt::winop snap $root.f.g \$im; \$im write \[tk_getSaveFile -filetypes {{\"PPM File\" {*.ppm}}  {\"All\" {*} }  } -parent . -title {Select a PPM File}\] -format ppm "
	$mfile add separator
	$mfile add command -label "Destroy" -command "destroy $root; $histo -destroy"
	menubutton $root.menu.graph -text "Graphics" -menu $root.menu.graph.m -anchor w -bg $color
	set mgraph [ menu  $root.menu.graph.m -bg $color]
	$mgraph add command -label "Set X Range..."
	$mgraph add command -label "Set Y Range..."
	$mgraph add command -label "Set Background Color..." -command "$root.f.g configure -plotbackground \[ tk_chooseColor -title \"Choose a background color\" -initialcolor \[ $root.f.g cget -plotbackground\]  \] "
	menubutton $root.menu.hist -text "Histogram" -menu $root.menu.hist.m -anchor w -bg $color
	set mhist [ menu  $root.menu.hist.m -bg $color]
	$mhist add command -label "Set Number of Buckets..." -command "modules::scaleDialog .{$histo}NbBucket $color \"Set the number of buckets\" 2 256 \[$histo -bucket number \] \"evaluate \{ $histo -bucket number ... \};modules::createHistogramGraph $histo $root.f.g \[$histo -minValue\] \\\[$histo -bucket number \\\] \""
	$mhist add command -label "Set Bucket Size..." -command "modules::entryDialog .{$histo}BucketSize $color \"Set the size of each bucket\" \[$histo -bucket size \] \"evaluate \{ $histo -bucket size ... \};modules::createHistogramGraph $histo $root.f.g \[$histo -minValue\] \\\[$histo -bucket number \\\] \""
	$mhist add command -label "Show Statistics..." -command "tk_messageBox -icon info -message \" Mean Value : \[$histo -stat mean\] \n\n Median Value : \[$histo -stat median\] \n\n Min Value : \[$histo -minValue\]\n\n Max Value : \[$histo -maxValue\]\" " 
	$mhist add command -label "Smooth Histogram..." -command  "modules::scaleDialog .{$histo}Smooth $color \"Set the filter size\" 1 10 2.0 \"evaluate \{ $histo -smooth ... \};modules::createHistogramGraph $histo $root.f.g \[$histo -minValue\] \\\[$histo -bucket number \\\] \""
#	$mhist add command -label "Create Probability Density Function..." -command  "modules::scaleDialog .{$histo}PDF $color \"Set the size of Parzen Window\" 1 30 1.0 \"evaluate \{ new pdf $histo ... \} \""
	$mhist add command -label "Create Probability Density Function..." -command  "modules::histoTopdfDialog  $histo $color"
	$mhist add cascade -label "Plot Probability Density Function." -menu $mhist.pdf
	menu $mhist.pdf -bg $color -postcommand "
	$mhist.pdf delete 0 last; 
	foreach p \[which proba\] {
	if \[ $root.f.g element exists \$p \] {
	set  modules::draw(\$p) true
	} else {
	set  modules::draw(\$p) false
	}
	$mhist.pdf add checkbutton  -onvalue true -offvalue false  -variable modules::draw(\$p) -label \$p -command \"if !\$modules::draw(\$p) \{modules::createPDFGraph \$p  $root.f.g \ \} else \{ $root.f.g element delete \$p \} \" 
	}
	foreach p \[which gauss\] {
	if \[ $root.f.g element exists \$p \] {
	set  modules::draw(\$p) true
	} else {
	set  modules::draw(\$p) false
	}
	$mhist.pdf add checkbutton  -onvalue true -offvalue false  -variable modules::draw(\$p) -label \$p -command \"if !\$modules::draw(\$p) \{modules::createPDFGraph \$p  $root.f.g \ \} else \{ $root.f.g element delete \$p \} \" 
	}
	foreach p \[which pdf\] {
	if \[ $root.f.g element exists \$p \] {
	set  modules::draw(\$p) true
	} else {
	set  modules::draw(\$p) false
	}
	$mhist.pdf add checkbutton  -onvalue true -offvalue false  -variable modules::draw(\$p) -label \$p -command \"if !\$modules::draw(\$p) \{modules::createPDFGraph \$p  $root.f.g \ \} else \{ $root.f.g element delete \$p \} \" 
	}
	"
	$mhist add command -label "Fit Gaussian..."
	
	pack $root.menu.file $root.menu.graph $root.menu.hist -side left
	pack $root.position.l $root.position.p -side left
	pack $root.position $root.menu -fill x
	
	frame $root.f -bg $color
	modules::drawHistogramInFrame $root.f $histo $color
	pack $root.f 
	
    }
    
}
proc modules::colorToRgb { color } {
    scan $color "#%02x%02x%02x" r g b
    set r [expr $r/255.0]
    set g [expr $g/255.0]
    set b [expr $b/255.0]
    return [list $r $g $b]
}
proc modules::mpeg { name {encode ppmtompeg} {play mplayer}  {factor 1} {from 0} {nb 4} } {
    set rname [file rootname $name]
    set ename [file extension $name]
    if [string compare $ename ".ppm"] {
	set conv 1
    } else {
	set conv 0
    }

    set i $from
    while { [ file exists ${rname}[set id [format "%0${nb}i" $i]]${ename}  ] } {
	for { set j 0 } { $j < $factor } { incr j } {
	    set did [format "%0${nb}i" [expr $factor * $i + $j]]
	    if $conv {
#		puts "CONVERSION convert ${rname}${id}${ename} ${rname}${id}.ppm"
		exec convert ${rname}${id}${ename} ${rname}${id}.ppm
	    }
	    file link  tmp.${rname}.$did.ppm ${rname}${id}.ppm 
	}
	incr i
    }
    set to $i
    set bg [format "%0${nb}i" [expr $factor * $from]]
    set nd [format "%0${nb}i" [expr $factor * $to - 1]]

    set f [open tmp.${rname}.param w]
    fconfigure $f -translation lf
    puts $f "PATTERN I"
    puts $f "OUTPUT ${rname}.mpeg"
    puts $f "INPUT_DIR ."
    puts $f "INPUT"
    puts $f "tmp.${rname}.*.ppm \[${bg}-${nd}\]"
    puts $f "END_INPUT"
    puts $f "BASE_FILE_FORMAT PNM"
    puts $f "INPUT_CONVERT *"
    puts $f "GOP_SIZE 1"
    puts $f "SLICES_PER_FRAME 1"
    puts $f "PIXEL FULL"
    puts $f "RANGE 10"
    puts $f "PSEARCH_ALG EXHAUSTIVE"
    puts $f "BSEARCH_ALG EXHAUSTIVE"
    puts $f "IQSCALE 8"
    puts $f "PQSCALE 10"
    puts $f "BQSCALE 25"
    puts $f "REFERENCE_FRAME DECODED"
    close $f
        
    exec ${encode} tmp.${rname}.param  

    file delete tmp.${rname}.param
    
    for { set i $from } { $i < $to } { incr i } {
	set id [format "%0${nb}i" [expr $i]]
	if $conv {
	    file delete ${rname}${id}.ppm
	}
	for { set j 0 } { $j < $factor } { incr j } {
	    set did [format "%0${nb}i" [expr $factor * $i + $j]]
	    file delete tmp.$rname.$did.ppm
	}
    }
    exec ${play}  ${rname}.mpeg &
}
proc modules::sceneGraph { { dim  2D}  } {
    package require BLT	
    set root .sceneGraph$dim
    if  ![winfo exists $root]  {
	toplevel $root
	#scrolll bar
	set vs $root.vs
	set hs $root.hs
	#create hierbox
	set list [blt::hierbox $root.h  \
		      -yscrollcommand "$vs set " \
		      -xscrollcommand "$hs set " ] 
	$list configure -width 300

	scrollbar $vs -orient vertical -command " $list yview "
	scrollbar $hs -orient horizontal -command "$list xview "

	
	blt::table $root 0,0 $list  -fill both 0,1 $vs  -fill y 1,0 $hs -fill x

	blt::table configure $root c1 r1 -resize non
	#	    blt::table configure $root c0 -width $list
	
	$list configure -separator "/" -autocreate yes -activebackground white

	$list entry configure root -label "$dim Scene Graph"  -labelfont { Helvetica 24 }   
	foreach i [which scn$dim] {
	    scan $i "scn$dim%u" val
	    set label "Scene no$val ($i)"
	    $list insert end "$label" -labelfont { Helvetica 16 }  -labelcolor  magenta
	    $list insert end "$label/Graphical Models" -labelcolor darkorange1
	    foreach j [ $i -objects ] {
		set realName [scn$dim -realName [$j -kind] ]
		set name "$realName ($j)"
		$list insert end "$label/Graphical Models/$name" -labelcolor darkorange1
		#parse inrview
		if ![string compare [$j -kind] inrview] {
		    foreach cam [$i -camera] {
			if ![string compare [$cam -kind] slicecam ] {
			    set k [$j -background $cam]
			    set otherRealName [scn$dim -realName [$k -kind] ]
			    set otherName "$otherRealName ($k)"
			    $list insert end "$label/Graphical Models/$name/$otherName"  
			}
		    }
		}

	    }
	    set val [$list insert end "$label/Deformable Models"]
	    $list entry configure $val -labelcolor forestgreen
	    foreach j [ $i -deformable ] {
		set realName [scn$dim -realName [$j -kind] ]
		set name "$realName ($j)"
		set val [$list insert end "$label/Deformable Models/$name"]
		$list entry configure $val -labelcolor forestgreen
	    }
	    set val [$list insert end "$label/Range Data Models"]
	    $list entry configure $val -labelcolor red
	    foreach j [ $i -range ] {
		set realName [scn$dim -realName [$j -kind] ]
		set name "$realName ($j)"
		set val [$list insert end "$label/Range Data Models/$name"]
		$list entry configure $val -labelcolor red
	    }
	    set val [$list insert end "$label/Cameras"]
	    $list entry configure $val -labelcolor blue
	    foreach j [ $i -camera ] {
		set realName [scn$dim -realName [$j -kind] ]
		set name "$realName ($j)"
		set val [$list insert end "$label/Cameras/$name"]
		$list entry configure $val -labelcolor blue
		#parse slicecam
		if ![string compare [$j -kind] slicecam] {
		    foreach k [$j -slicescale] {
			set otherRealName [scn$dim -realName [$k -kind] ]
			set otherName "$otherRealName ($k)"
			$list insert end "$label/Cameras/$name/$otherName"  
		    }
		}
	    }
	    
	}
    }
}

proc modules::defaultUninstallMenu {viewer camera object} {
    # remove frame children
    set list [winfo children $viewer.context]
    foreach w $list { destroy $w}
    pack $viewer.context -fill x
    bind [$camera -widget] <Button-3> ""
}


proc bgerror { str } {
    global errorInfo
    namespace eval onerror { global errCmd }
    
    if { [winfo exists $modules::parent.msg.r.b.list] } {
	if { ![info exists onerror::errCmd] } {
	    set onerror::errCmd $errorInfo
	    modules::displayString "error: $str" red
	} else {
	    if { ![string compare $onerror::errCmd $errorInfo] } {
		modules::displayString "$errorInfo" red
	    } else {
		set onerror::errCmd $errorInfo
		modules::displayString "error: $str" red
	    }
	}
    } else {
	puts "********* TCL error ***********"
	puts "$errorInfo"
    }
    
    if { [winfo exists $modules::parent.cmd.r.b.list] } {
	$modules::parent.cmd.r.b.list insert $modules::yavncmds [$modules::parent.cmd.r.txt get]
	$modules::parent.cmd.r.txt delete 0 1000
	$modules::parent.cmd.r.b.list see end
	incr modules::yavncmds
	set modules::yavcrtcmd $modules::yavncmds
    }
}


# Proc to display a balloon help
# Example:
#  button  .b -text Exit -command exit
#  balloon .b "Push me if you're done with this"
#  pack    .b

proc balloon {w help} {
    bind $w <Any-Enter> "after 5 [list balloon:show %W [list $help]]"
    bind $w <Any-Leave> "destroy %W.balloon"
}
proc balloon:show {w arg} {
    if {[eval winfo containing  [winfo pointerxy .]]!=$w} {return}
    set top $w.balloon
    catch {destroy $top}
    toplevel $top -bd 1 -bg black
    wm overrideredirect $top 1
    if {$::tcl_platform(platform) == "macintosh"} {
	unsupported1 style $top floating sideTitlebar
    }
    pack [message $top.txt -aspect 10000 -bg lightyellow \
	      -font fixed -text $arg]
    set wmx [winfo rootx $w]
    set wmy [expr [winfo rooty $w]+[winfo height $w]]
    wm geometry $top \
	[winfo reqwidth $top.txt]x[winfo reqheight $top.txt]+$wmx+$wmy
    raise $top
}
