/* epi_close.c 2.2 5/14/98 */
/*==========================================================================
 *
 * File : $EPI/src/epi_close.c
 * Module : ?
 * $ Id : ? 
 * Test file : $EPI/test/test.im $EPI/test/test.tr
 * Author :  Jacques Feldmar
 * Creation: June 2 1992
 *===========================================================================
 *   (c) Copyright 1992, Inria, all right reserved
 *===========================================================================
 *
 *------------------------------Description----------------------------------
 *
 * Calcul le point d'une liste le plus proche d'un point donne.
 * Methode de Zhang basee sur des kd_trees
 *
 *-----------------------------Modification----------------------------------
 *
     Le 7 aout 94, Jacques Feldmar : On remplace le trie par arbre binaire, par le tri de 
     la bibliotheque (QuickSort).

     Le 25 aout 94, Jacques Feldmar
     Telle que la construction du kdtree etait programmee, beaucoup de temps etait passe
     a copier les listes gauches et les listes droites : on a  supprimer ca, car c'etait inutile. 

     Le 10 octobre 94, jacques Feldmar
     On essaie d'eviter les sequences allocation/desallocation des vecteurs n-dimensionels. Pour cela,
     ces vecteurs sont locaux statics, et a chaque appel a la fonction, on verifie que la dimension
     allouee correspond bien a la dimension necessaire. Si ce n'est pas le cas on libere et on realloue.
 */

/*********************************************************************************/
/* epi_close.c							                 */
/*		(c) Jacques Feldmar -- INRIA -- 9/3/92	 */
/*										 */
/*********************************************************************************/

#include <epidaure.h>
#include <epidaure.ei>
#include <epidaure.ee>


/*-------Definition des fonctions statiques----------*/
#ifndef NO_PROTO
#ifdef COMPILE_UNUSED
static void trie_liste( t_ListPtr liste, ARBRE_TRI *p_arbre, void_func_type comp, void_func_type func, int dim  );
static int compare_sommet( t_NPointPtr sommet1, t_NPointPtr sommet2 );
#endif /* COMPILE_UNUSED */
static ARBRE_TRI ajoute_noeud( ARBRE_TRI arbre, char *obj , void_func_type comp, void_func_type func, t_NPointPtr point, int dim );
static void parcours_arbre_GRD( ARBRE_TRI arbre, void_func_type action, int profondeur );
static ARBRE_TRI alloue_noeud( char *obj, void_func_type func ) ;
static void arbre_to_liste( ARBRE_TRI arbre, t_ListPtr *nouvelle_liste ) ;
static void divise_liste( t_ListPtr liste, t_ListPtr *p_liste_g, t_ListPtr *p_liste_d, char **p_sommet_milieu ) ;
static void libere_arbre( ARBRE_TRI arbre );
static void construit_kd_tree( t_ListPtr * liste_sommets, ARBRE_TRI *kd_tree, int profondeur, void_func_type func, int dim );
static void search_kdtree( ARBRE_TRI arbre, t_NPointPtr sommet_ori, int profondeur );
static float calc_distance( t_NPointPtr  pt1, t_NPointPtr  pt2);
static void search_liste_kdtree( ARBRE_TRI arbre, t_NPointPtr sommet_ori, int profondeur, float rayon );
static int crest_to_3d( P_SOMMET pt, t_NPointPtr sommet ) ;
static void  CopyListNonRecursiveAndCorrect( t_ListPtr liste_sommets, t_ListPtr * p_nouvelle_liste, void_func_type func, int dim  ) ;
static int compare_sommet_epi( char *  * sommet1, char * * sommet2 );
#else 
#ifdef COMPILE_UNUSED
static void trie_liste();
static int compare_sommet();
#endif /* COMPILE_UNUSED */
static ARBRE_TRI ajoute_noeud();
static void parcours_arbre_GRD();
static ARBRE_TRI alloue_noeud();
static void arbre_to_liste();
static void divise_liste();
static void libere_arbre();
static void construit_kd_tree();
static void search_kdtree();
static float calc_distance();
static void search_liste_kdtree();
static int crest_to_3d();
static void  CopyListNonRecursiveAndCorrect( ) ;
static int compare_sommet_epi( );
#endif
/*-------Fin des proto statiques----------*/


/* Construit le kd_tree associe a une liste d'objet, chaque objet etant converti en un point n dimensionnel par la fonction pointee par func.
 

  Un kd-tree est une structure de donnee (developpee initialement en base de donnees)
  que l'on precalcule sur un ensemble E d'objets pouvant etre convertis en
  en points N-dimensionnels. Cette structure permettra, etant donne un objet O, de 
  trouver dans E l'objet le plus proche de O, au sens de la norme Euclidienne 
  (dimension N).


RETURN  
  fonction void.
  
ERROR 
 Cette fonction ne gere pas les erreurs de type Epidaure.
  */
#if defined(_ANSI_)
void E_BuildKDTree(t_ListPtr liste, /* liste des objets a stoquer dans le kd_tree. Ces objets 
				       doivent etre convertibles par func. Parametre d'entree */
		   void_func_type func,  /* pointeur sur la fonction de conversion 
					    objet -> T_NPointPtr. 
					    Cette fonction doit retourner TRUE si l'objet doit etre 
					    insere dans le kd_tree, FALSE sinon. Parametre d'entree.
					    */
		   t_KDTreePtr arbre,  /* le kd_tree construit a partir la liste liste. Parametre 
					 de sortie */
		   int dim /* dimension des vecteurs T_NPoint convertis par func. Parametre 
					 d'entree */ ) 
#else
void E_BuildKDTree(liste, func, arbre, dim)
t_ListPtr liste; /* liste des objets a stoquer dans le kd_tree. Ces objets 
		    doivent etre convertibles par func. Parametre d'entree */
void_func_type func; /* pointeur sur la fonction de conversion 
			objet -> T_NPointPtr. 
			Cette fonction doit retourner TRUE si l'objet doit etre 
			insere dans le kd_tree, FALSE sinon. Parametre d'entree.
			*/
 t_KDTreePtr arbre;  /* le kd_tree construit a partir la liste liste. Parametre 
			de sortie */
 int dim;             /* dimension des vecteurs T_NPoint convertis par func. Parametre 
			 d'entree */
#endif

{
  t_ListPtr nouvelle_liste = NULL ;
  *arbre = NULL ;
  CopyListNonRecursiveAndCorrect( liste, &nouvelle_liste, func, dim ) ;
  construit_kd_tree(  &nouvelle_liste, arbre , 0, func, dim   ) ;
}




/* =============================================================
   GESTION DES ARBRES de sommet  
   ============================================================= */

#ifdef COMPILE_UNUSED
#if defined(_ANSI_)
static void trie_liste( t_ListPtr liste, ARBRE_TRI *p_arbre, void_func_type comp, void_func_type func, int dim  )
#else
static void trie_liste(liste, p_arbre, comp, func, dim)
 t_ListPtr liste;
 ARBRE_TRI *p_arbre;
void_func_type comp;
 void_func_type func;
 int dim;
#endif
{
  
  char *       sommet;
  t_ListPtr       aList;
  t_charNPPtr_func_type funcp;
  
  t_NPointPtr point = NULL ;
  
  
  T_AllocNPoint(&point ) ;
  T_AllocNPointTb(point, dim ) ; 
  *p_arbre = NULL ;
  aList = liste;
  while (aList != NULL)
    {
      sommet = (char *)aList->item ;
      funcp = (t_charNPPtr_func_type ) func;
      if( funcp ( sommet, point ) )
	{
	  *p_arbre = ajoute_noeud( *p_arbre, sommet, comp, func, point, dim  ) ;
	}
      aList = aList->next;
    }
  T_FreeNPoint( &point ) ;
  
}

#endif /* COMPILE_UNUSED */


/* =============================================================
   GESTION DES ARBRES generale 
   ============================================================= */


#if defined(_ANSI_)
static ARBRE_TRI ajoute_noeud( ARBRE_TRI arbre, char *obj , void_func_type comp, void_func_type func, t_NPointPtr point, int dim )
#else
static ARBRE_TRI ajoute_noeud(arbre, obj, comp, func, point, dim)
 ARBRE_TRI arbre;
 char *obj;
 void_func_type comp;
 void_func_type func;
 t_NPointPtr point;
 int dim;
#endif

{
  static t_NPointPtr point_arbre = NULL ;
  t_charNPPtr_func_type arbfunc;
  t_2NPointPtr_func_type compp;
  
  
  if( arbre == NULL )
    return( alloue_noeud( obj, func ) ) ; 
  T_AllocNPoint(&point_arbre ) ;
  T_AllocNPointTb(point_arbre, dim ) ; 
  
  arbfunc = ( t_charNPPtr_func_type ) arbre->func;
  arbfunc ( arbre->obj, point_arbre ) ;
  
  compp = ( t_2NPointPtr_func_type ) comp;
  if( compp ( point, point_arbre ) )
    {
      /* on ajoute dans le sous arbre gauche */
      T_FreeNPoint( &point_arbre ) ;
      arbre->fils_g = ajoute_noeud( arbre->fils_g, obj , comp, func, point, dim ) ;
    }
  else
    {
      /* on ajoute dans le sous arbre gauche */
      T_FreeNPoint( &point_arbre ) ;
      arbre->fils_d = ajoute_noeud( arbre->fils_d, obj , comp, func, point, dim ) ;
    }
  return( arbre ) ;
}

#if defined(_ANSI_)
static void parcours_arbre_GRD( ARBRE_TRI arbre, void_func_type action, int profondeur )
#else
static void parcours_arbre_GRD(arbre, action, profondeur)
 ARBRE_TRI arbre;
 void_func_type action;
 int profondeur;
#endif

{
  charint_func_type actionp;
  
  if( arbre == NULL )
    return ; 
  parcours_arbre_GRD( arbre->fils_g , action, profondeur + 1 ) ;
  actionp = (charint_func_type) action;
  actionp ( arbre->obj, profondeur ) ;
  parcours_arbre_GRD( arbre->fils_d , action, profondeur + 1  ) ;
  
}


#if defined(_ANSI_)
static ARBRE_TRI alloue_noeud( char *obj, void_func_type func ) 
#else
static ARBRE_TRI alloue_noeud(obj, func)
 char *obj;
 void_func_type func;
#endif
{
  ARBRE_TRI pt ;
  
  pt = (ARBRE_TRI)malloc( sizeof( NOEUD ) ) ;
  if( pt == NULL )
    {
      fprintf( stderr, "erreur allocation noeud \n" ) ;
      return( NULL ) ;
    }
  
  pt->obj = obj ;
  pt->fils_g = NULL ;
  pt->fils_d = NULL ;
  pt->func = func ;
  return( pt ) ;
}


/* on fait un parcours droite, racine, gauche
   car  T_AddObjList fait de l'ajout en tete */

#if defined(_ANSI_)
static void arbre_to_liste( ARBRE_TRI arbre, t_ListPtr *nouvelle_liste ) 
#else
static void arbre_to_liste(arbre, nouvelle_liste )
ARBRE_TRI arbre;
t_ListPtr *nouvelle_liste ;

#endif

{
  if( arbre == NULL )
    return ; 
  arbre_to_liste( arbre->fils_d, nouvelle_liste ) ;
  T_AddObjList( (t_ObjPtr)arbre->obj, nouvelle_liste ) ;
  arbre_to_liste( arbre->fils_g, nouvelle_liste ) ;
}

/* =============================================================
   GESTION DES Listes generale 
   ============================================================= */

#if defined(_ANSI_)
static void divise_liste( t_ListPtr liste, t_ListPtr *p_liste_g, t_ListPtr *p_liste_d, char **p_sommet_milieu ) 
#else
static void divise_liste(liste, p_liste_g, p_liste_d, p_sommet_milieu )
t_ListPtr liste;
t_ListPtr *p_liste_g;
t_ListPtr *p_liste_d;
char **p_sommet_milieu ;

#endif

{
  int taille ;
  int i ;
  
  
  t_ListPtr aList ;
  
  *p_liste_g = NULL ;
  *p_liste_d = NULL ;
  *p_sommet_milieu = NULL ;
  
  if( liste == NULL )
	  return ;
  
  taille = T_CountList( liste ) ;
  
  /* on repere le sommet milieu de la liste */
  aList = liste ;
  for( i = 0 ; i < ( taille - 1 ) / 2 ; i++ )
    aList = aList->next ;
  *p_sommet_milieu = (char *)aList->item ;
  if( aList->back != NULL )
    aList->back->next = NULL ;
  if( aList->next != NULL )
    aList->next->back = NULL ;
  
  *p_liste_d = aList->next ;
  T_Free( (void **) &aList ) ;
  if( taille <= 2 )
    return ;
  *p_liste_g = liste ;
}



#if defined(_ANSI_)
static void libere_arbre( ARBRE_TRI arbre )
#else
static void libere_arbre(arbre )
ARBRE_TRI arbre ;

#endif

{
  if( arbre == NULL )
    return ;
  
  libere_arbre( arbre->fils_g ) ;	
  libere_arbre( arbre->fils_d ) ;
  free( arbre ) ;
}


/* ==================================================*/
/* ==================================================*/
/* Gestion des kd_trees                              */
/* ==================================================*/
/* ==================================================*/

/* TROIS variables globale : je n'ai pas trouve d'autre solution, mais ils y a des gens super
   intelligents chez Epidaure... Et ailleurs.  */
static int profondeur_courante ;
static void_func_type fonction_conversion_courante ;
static int dimension_courante ;


#if defined(_ANSI_)
static void construit_kd_tree( t_ListPtr * liste_sommets, ARBRE_TRI *kd_tree, int profondeur, void_func_type func, int dim )
#else
static void construit_kd_tree(liste_sommets, kd_tree, profondeur, func, dim)
t_ListPtr * liste_sommets ;
ARBRE_TRI *kd_tree ;
int profondeur ;
void_func_type func;
int dim ;
#endif
{
  t_ListPtr liste_gauche = NULL ;
  t_ListPtr liste_droite = NULL ;
  char * sommet = NULL ;



  if( *liste_sommets == NULL )
    {
      *kd_tree = NULL ;
      return ;
    }
  
  profondeur_courante = profondeur ;
  if( ( dim != 1 ) || ( profondeur == 0 )  )
  {
	  fonction_conversion_courante = func ;
	  dimension_courante = dim ;
	  T_SortList( liste_sommets, (void_func_type) compare_sommet_epi,  OBJ_UNKNOWN );
  }
  /* une petite optimisation dans le cas ou 
     la dimension de l'espace est 1 
     et ou la profondeur n'est pas 0 : 
     en effet dans ce cas,
     liste sommet est deja triee : ceci explique le if */

	
  /* attention, la fonction divise_liste a un effet de bord sur 
     liste_sommets. De plus, divise_liste libere le doublet auquel etait attache 
     sommet. Ca explique que a la fin de construit_kd_tree, liste_sommets 
     est liberee. */
  
  divise_liste( *liste_sommets, &liste_gauche, &liste_droite, &sommet ) ;
  
  *kd_tree = alloue_noeud( sommet, func ) ;
  profondeur_courante = profondeur ;
  construit_kd_tree( &liste_gauche, &(*kd_tree)->fils_g, profondeur + 1, func, dim  ) ;
  profondeur_courante = profondeur ;
  construit_kd_tree( &liste_droite, &(*kd_tree)->fils_d, profondeur + 1, func, dim   ) ;
}

#ifdef COMPILE_UNUSED
/* retourne TRUE si sommet1 est plus petit que sommet2 */
#if defined(_ANSI_)
static int compare_sommet( t_NPointPtr sommet1, t_NPointPtr sommet2 )
#else
static int compare_sommet(sommet1, sommet2 )
t_NPointPtr sommet1;
t_NPointPtr sommet2 ;

#endif

{
  int i ;
  
  i = ( profondeur_courante % sommet1->dim ) ;
  return( sommet1->v[i] < sommet2->v[i] )  ;
}
#endif /* COMPILE_UNUSED */



/* retourne TRUE si sommet1 est plus petit que sommet2 */
#if defined(_ANSI_)
static int compare_sommet_epi( char * *sommet1, char * *sommet2 )
#else
static int compare_sommet_epi(sommet1, sommet2 )
char  * * sommet1;
char  * * sommet2 ;

#endif

{
  int i ;
  t_charNPPtr_func_type funcp;
  static t_NPointPtr point1 = NULL,  point2 = NULL ;
  reeltype  difference  ;


  /* dans le cas ou point1 et point2 sont deja alloues, on verifie que leur dimension correspond bien a 
     dimension_courante. Si ce n'est pas le cas, on libere point1 et point2 => ils seront realloue avec la bonne taille
     a l'instruction suivante */
  if( point1 != NULL )
  {
	  if( point1->dim != dimension_courante )
	  {
		  T_FreeNPoint( &point1 ) ;
		  T_FreeNPoint( &point2 ) ;
		  point1 = NULL ; point2 = NULL ;
	  }
  }
  if( point1 == NULL )
  {
	  T_AllocNPoint(&point1 ) ;
	  T_AllocNPointTb(point1, dimension_courante ) ; 
	  T_AllocNPoint(&point2 ) ;
	  T_AllocNPointTb(point2, dimension_courante ) ;
  }

  funcp = (t_charNPPtr_func_type )fonction_conversion_courante ;
  funcp( *sommet1, point1 ) ;
  funcp( *sommet2, point2 ) ;
  i = ( profondeur_courante % dimension_courante ) ;
  difference = (point1)->v[i] - (point2)->v[i]   ;


  if( difference == 0. )
	  return( 0 ) ;
   if( difference < 0 )
	  return( -1 ) ;
  return( 1 ) ;
}


#ifndef MAXFLOAT
#  define _MAXFLOAT
#  define MAXFLOAT    ((float) 3.40282346638528860e+38)
#endif

static t_NPointPtr sommet_arbre = NULL ;
static float meilleure_distance = MAXFLOAT ;
static char * meilleur_sommet = NULL ;



/*  Trouve, parmi les objets stoques dans le kd_tree, l'objet le plus proche de l'objet pointe  par pt, au sens de la norme Euclidienne. 

    La recherche ne se fait que dans une 
    sphere de rayon rayon, centree sur pt. Si cette sphere est vide, le resultat sera NULL.
RETURN
 fonction void.

ERROR
  Cette fonction ne gere pas les erreurs de type Epidaure.
  
*/

#if defined(_ANSI_)
void E_ObjClosest( t_KDTree arbre /* le kd_tree dans lequel sont stoque les objets. Parametre
				     d'entree */, 
		   char * pt /* un pointeur sur l'objet dont on cherche le point le plus proche.
				Parametre d'entree. Bien entendu, cet objet doit etre
				convertible par la function de conversion qui a servie 
				a construire le kd-tree. Parametre d'entree */, 
		   char ** res /* pointe en sortie sur l'objet du kd_tree le plus proche de pt
				  se trouvant dans la sphere de centre pt, et de rayon rayon. 
				  Sera NULL si cette sphere est vide.
				  Parametre de sortie */, 
		   float rayon /* rayon de la sphere dans laquelle on cherche.
				  Parametre d'entree */, 
		   int dim  /* dimension des vecteurs T_NPoint convertis. Parametre d'entree */ )
#else
void E_ObjClosest(arbre,  pt,  res, rayon, dim )
t_KDTree arbre ; /* le kd_tree dans lequel sont stoque les objets. Parametre
		 d'entree */ 
char * pt ; /* un pointeur sur l'objet dont on cherche le point le plus proche.
	     Parametre d'entree. Bien entendu, cet objet doit etre
	     convertible par la function de conversion qui a servie 
	     a construire le kd-tree. Parametre d'entree */
char ** res ; /* pointe en sortie sur l'objet du kd_tree le plus proche de pt
	       se trouvant dans la sphere de centre pt, et de rayon rayon. 
	       Sera NULL si cette sphere est vide.
	       Parametre de sortie */
float rayon ; /* rayon de la sphere dans laquelle on cherche.
	       Parametre d'entree */
int dim ; /* dimension des vecteurs T_NPoint convertis. Parametre d'entree */ 

#endif

{
  t_charNPPtr_func_type arbfunc;
  static t_NPointPtr sommet = NULL ;
  
  if (arbre == NULL )
    return ;
  

  if( sommet != NULL )
  {
	  if( sommet->dim != dim )
	  {
		  T_FreeNPoint( &sommet  );
		  T_FreeNPoint( &sommet_arbre  );
		  sommet = NULL ;
		  sommet_arbre = NULL ;
	  }
  }

  if( sommet == NULL )
  {
	  T_AllocNPoint(&sommet ) ;
	  T_AllocNPointTb(sommet, dim ) ; 
	  T_AllocNPoint(&sommet_arbre ) ;
	  T_AllocNPointTb( sommet_arbre, dim ) ; 
  }


  meilleur_sommet = NULL ;  
  arbfunc = ( t_charNPPtr_func_type ) arbre->func;
  arbfunc ( pt, sommet ) ;
  meilleure_distance = rayon ;
  search_kdtree( arbre , sommet, 0 ) ;
  *res = meilleur_sommet ;
  
  
}



#if defined(_ANSI_)
static void search_kdtree( ARBRE_TRI arbre, t_NPointPtr sommet_ori, int profondeur )
#else
static void search_kdtree(arbre, sommet_ori, profondeur )
ARBRE_TRI arbre;
t_NPointPtr sommet_ori;
int profondeur ;

#endif

{
  reeltype c1, c2 ;
  double abs_c1_moins_c2 ;
  t_charNPPtr_func_type arbfunc;
  int i ;
  
  if( arbre == NULL )
    return ;
  
  arbfunc = ( t_charNPPtr_func_type ) arbre->func;
  arbfunc ( arbre->obj, sommet_arbre ) ;
  
  i = ( profondeur % sommet_ori->dim ) ;
  c1 = sommet_ori->v[i] ;
  c2 = sommet_arbre->v[i] ;
  
  
  abs_c1_moins_c2 = fabs( (double)( c1 - c2 ) ) ;
  
  if( ( abs_c1_moins_c2 <= meilleure_distance ) &&
     ( calc_distance( sommet_ori, sommet_arbre ) <= meilleure_distance )
     )
    {
      meilleur_sommet = (char *)arbre->obj ; 
      meilleure_distance = calc_distance( sommet_ori, sommet_arbre ) ;
    }
  if( c1 - meilleure_distance < c2 )
    search_kdtree( arbre->fils_g, sommet_ori, profondeur + 1 ) ;
  
  if( c2 - meilleure_distance < c1 )
    search_kdtree( arbre->fils_d, sommet_ori, profondeur + 1 ) ;
  
}


#if defined(_ANSI_)
static float calc_distance( t_NPointPtr  pt1, t_NPointPtr  pt2)
#else
static float calc_distance( pt1,  pt2)
t_NPointPtr  pt1;
t_NPointPtr  pt2;

#endif

{
  int i ;
  double val = 0. ;
  
  for( i = 0 ; i < pt1->dim ; i++ )
    val += (double)(pt1->v[i] - pt2->v[i]) *  (double)(pt1->v[i] - pt2->v[i]);
  val = sqrt( val ) ;
  return( (float)val ) ;
}




/* Libere le kd_tree arbre. Ne libere pas les objets qui lui sont attache.
 
RETURN
 fonction void.

ERROR
  Cette fonction ne gere pas les erreurs de type Epidaure.
   
   */
#if defined(_ANSI_)
void E_ClearKDTree( t_KDTreePtr arbre /* le kd_tree a liberer. Parametre d'entree/sortie */) 
#else
void E_ClearKDTree(arbre )
t_KDTreePtr arbre ; /* le kd_tree a liberer. Parametre d'entree/sortie */
#endif
{
  libere_arbre( *arbre  );
  *arbre = NULL  ;
}






static t_ListPtr liste_sphere = NULL ;


/* Trouve, parmi les objets stoques dans le kd_tree, l'ensemble des objets qui se trouve dans la sphere de rayon rayon, centree sur pt. Le resultat est stoque dans la liste pointee par p_liste.

RETURN
 fonction void.

ERROR
  Cette fonction ne gere pas les erreurs de type Epidaure.
  */

#if defined(_ANSI_)
void E_ObjLClosest( t_KDTree arbre /* le kd_tree dans lequel sont stoques les objets.
				      Parametre d'entree */,
		    char * pt /* un pointeur sur l'objet autour duquel on cherche. 
				 Bien entendu, cet objet doit etre
				 convertible par la function de conversion qui a servie
				 a construire le kd-tree. 
			         Parametre d'entree. */, 
		    t_ListPtr * p_liste /* pointe en sortie sur la liste des objets qui se trouvent 
					   dans la sphere de centre pt et de rayon rayon.
					   Parametre de sortie */, 
		    float rayon /* rayon de la sphere dans laquelle on cherche.
				    Parametre d'entree */ , 
		    int dim /*  dimension des vecteurs T_NPoint convertis.
				Parametre d'entree */ )
#else
void E_ObjLClosest(arbre,  pt,  p_liste, rayon, dim )
t_KDTree arbre; /* le kd_tree dans lequel sont stoques les objets.
		  Parametre d'entree */
char * pt; /* un pointeur sur l'objet autour duquel on cherche. 
	    Bien entendu, cet objet doit etre
	    convertible par la function de conversion qui a servie
	    a construire le kd-tree. 
	    Parametre d'entree. */ 
t_ListPtr * p_liste ; /* pointe en sortie sur la liste des objets qui se trouvent 
		       dans la sphere de centre pt et de rayon rayon.
		       Parametre de sortie */
float rayon ; /* rayon de la sphere dans laquelle on cherche.
	       Parametre d'entree */ 
int dim ; /*  dimension des vecteurs T_NPoint convertis.
	    Parametre d'entree */

#endif

{
  static t_NPointPtr sommet = NULL ;
  t_charNPPtr_func_type arbfunc;
  
  if (arbre == NULL )
    return ;
  
  if( sommet != NULL )
  {
	  if( sommet->dim != dim )
	  {
		  T_FreeNPoint( &sommet  );
		  T_FreeNPoint( &sommet_arbre  );
		  sommet = NULL ;
		  sommet_arbre = NULL ;
	  }
  }

  if( sommet == NULL )
  {
	  T_AllocNPoint(&sommet ) ;
	  T_AllocNPointTb(sommet, dim ) ; 
	  T_AllocNPoint(&sommet_arbre ) ;
	  T_AllocNPointTb( sommet_arbre, dim ) ; 
  }
  
  arbfunc = ( t_charNPPtr_func_type ) arbre->func;
  arbfunc ( pt, sommet ) ;
  liste_sphere = NULL ;
  search_liste_kdtree( arbre , sommet, 0, rayon ) ;
  *p_liste = liste_sphere ;
  
  
}



#if defined(_ANSI_)
static void search_liste_kdtree( ARBRE_TRI arbre, t_NPointPtr sommet_ori, int profondeur, float rayon )
#else
static void search_liste_kdtree(arbre, sommet_ori, profondeur, rayon )
ARBRE_TRI arbre;
t_NPointPtr sommet_ori;
int profondeur;
float rayon ;

#endif

{
  t_charNPPtr_func_type arbfunc;
  reeltype c1, c2 ;
  double abs_c1_moins_c2 ;
  int i ;
  
  if( arbre == NULL )
    return ;
  
  arbfunc = ( t_charNPPtr_func_type ) arbre->func;
  arbfunc ( arbre->obj, sommet_arbre ) ;
  
  i = ( profondeur % sommet_ori->dim ) ;
  c1 = sommet_ori->v[i] ;
  c2 = sommet_arbre->v[i] ;
  
  
  abs_c1_moins_c2 = fabs( (double)( c1 - c2 ) ) ;
  
  if( ( abs_c1_moins_c2 <= rayon ) &&
     ( calc_distance( sommet_ori, sommet_arbre ) <= rayon )
     )
    T_AddObjList( (t_ObjPtr)arbre->obj, &liste_sphere ) ; 
  if( c1 - rayon < c2 )
    search_liste_kdtree( arbre->fils_g, sommet_ori, profondeur + 1, rayon ) ;
  
  if( c2 - rayon < c1 )
    search_liste_kdtree( arbre->fils_d, sommet_ori, profondeur + 1, rayon ) ;
  
}






/* ================================================================================= */
/* ================================================================================= */
/* Compatibilite avec l'ancienne version. Ces fonctions ne doivent pas etre utilisees, elles ne sont pas maintenues et ne se trouvent dans la librairie  que pour les quelques personnes d'Epidaure qui les utilisent. Elles devraient disparaitrent a terme. */
#define DIM_KD_TREE 3

/*=======================================================================
 * E_get_point_proche_kdtree(  P_SOMMET sommet_ori, P_SOMMET * p_sommet_proche,
 t_ListPtr liste_sommets_modele, float DMAX  )
 *
 *
 * Action :
 * calcul le point d'une liste le plus proche d'un point donne
 *
 * Inclure dans l'ordre :
 * <epidaure.h>
 *
 * Parametres de sortie :
 *  p_sommet_proche     : contient en sortie un pointeur sur le sommet le plus proche
 *
 * Parametres d'entree :
 *  sommet_ori         : le sommet dont on cherche le point le plus proche. Si ce parametre
 est NULL, la fonction ne fait que construire un nouveau kd_tree.
 *  liste_sommet_modele : liste des sommet dans laquelle chercher le point le plus proche. A chaque fois que 
 ce parametre n'est pas NULL, on construit un nouveau kd_tree.
 Si ce parametre est NULL, on utilse le kd_tree construit a l'appel 
 precedent.
 *  DMAX                : rayon de la sphere centree sur sommet ori dans laquelle on cherche
 le sommet le plus proche
 *
 * Parametres d'entree/sortie :
 * aucun.
 *
 * Retour :
 * 0 si une erreur c'est produite, 1 si tout c'est bien passe.
 *
 * Erreurs :
 * 
 * Cette fonction est A TESTER.
 *
 * Observations :
 *  a chaque fois que l'on appelle la fonction avec une nouvelle liste de sommets ( l'argumenet liste_sommet_modele 
 *  n'est pas a NULL, ou l'argument sommet_ori est a NULL ),
 *  on construit un nouveau kd tree. Chaque appel avec une nouvelle liste est un peut 
 *  couteux. Mais en general, on utilise plusieurs fois de suite la meme liste...
 * 
 * 
 *
 * Exemple :
 * On trouvera un exemple d'utilisation dans le main qui se trouve dans ce meme fichier
 * entoure d'un #if 0 ...  #endif
 *========================================================================
 */

/* UTILISATION du kd_tree */
static ARBRE_TRI arbre_courant = NULL ;


/* Cette fonction est volontairement non commentee.  Cette fonction ne doit pas etre utilisee, elle n'est pas maintenue et ne se trouve dans la librairie  que pour les quelques personnes d'Epidaure qui l'utilise. Elle devrait disparaitrent a terme. */
#if defined(_ANSI_)
int E_get_point_proche_kdtree(  P_SOMMET sommet_ori, P_SOMMET * p_sommet_proche, t_ListPtr liste_sommets_modele, float DMAX  )
#else
int E_get_point_proche_kdtree(sommet_ori,  p_sommet_proche, liste_sommets_modele, DMAX  )
P_SOMMET sommet_ori;
P_SOMMET * p_sommet_proche;
t_ListPtr liste_sommets_modele;
float DMAX  ;

#endif
{
  t_NPointPtr sommet = NULL ;
  t_CrestPtPtr_func_type arbfunc;
  t_ListPtr nouvelle_liste = NULL ;


  if( ( liste_sommets_modele != NULL ) || ( sommet_ori == NULL )  )
    {
      libere_arbre( arbre_courant  ) ;
      arbre_courant  = NULL ;
      CopyListNonRecursiveAndCorrect( liste_sommets_modele, &nouvelle_liste, (void_func_type) crest_to_3d, 3   ) ;
      construit_kd_tree(  &nouvelle_liste, &arbre_courant , 0, (void_func_type) crest_to_3d, 3   ) ;
    }
  else
    if( arbre_courant  == NULL )
      {
	fprintf( stderr, " attention, il n'y a pas de kd_tree deja existant\n" ) ;
	return( 0 ) ;
      }
  
  if( sommet_ori == NULL )
    return( 1 ) ; 
  
  
  T_AllocNPoint(&sommet ) ;
  T_AllocNPointTb(sommet, 3 ) ; 
  T_AllocNPoint(&sommet_arbre ) ;
  T_AllocNPointTb( sommet_arbre, 3 ) ;
  meilleure_distance = DMAX ;
  meilleur_sommet = NULL ;
  arbfunc = ( t_CrestPtPtr_func_type ) arbre_courant->func;
  arbfunc ( sommet_ori, sommet ) ;
  search_kdtree( arbre_courant , sommet, 0 ) ;
  
  *p_sommet_proche = (P_SOMMET)meilleur_sommet ;
  T_FreeNPoint( &sommet  );
  T_FreeNPoint( &sommet_arbre  );
  
  return( 1 ) ;
}


/*=======================================================================
 * E_free_kdtree(  ) ;
 *
 *
 * Action :
 * libere le kd_tree courant
 *
 * Inclure dans l'ordre :
 * <epidaure.h>
 *
 =======================================================================*/


/* Cette fonction est volontairement non commentee.  Cette fonction ne doit pas etre  utilisee, elle n'est pas maintenue et ne se trouve dans la librairie  que pour les quelques personnes d'Epidaure qui l'utilise. Elle devrait  disparaitrent a terme. */
#if defined(_ANSI_)
int E_free_kdtree( void ) 
#else
int E_free_kdtree()

#endif
{
  libere_arbre( arbre_courant  ) ;
  arbre_courant  = NULL ;
  return( 1 ) ;
}




#if defined(_ANSI_)
static int crest_to_3d( P_SOMMET pt, t_NPointPtr sommet ) 
#else
static int crest_to_3d(pt, sommet )
P_SOMMET pt;
t_NPointPtr sommet ;

#endif

{
  sommet->v[0] = pt->pt.x ;
  sommet->v[1] = pt->pt.y ;
  sommet->v[2] = pt->pt.z ;
  
  return( TRUE ) ;
}



#if defined(_ANSI_)
static void CopyListNonRecursiveAndCorrect( t_ListPtr liste, t_ListPtr * p_nouvelle_liste,  void_func_type func, int dim  ) 
#else
static void CopyListNonRecursiveAndCorrect( liste, p_nouvelle_liste, func, dim ) 
t_ListPtr liste ;
t_ListPtr * p_nouvelle_liste ; 
void_func_type func;
int dim;
#endif
{
	t_charNPPtr_func_type funcp;
	t_NPointPtr point = NULL ;
	t_ListPtr aList  = liste ;
  
	T_AllocNPoint(&point ) ;
	T_AllocNPointTb(point, dim ) ; 
	funcp = (t_charNPPtr_func_type ) func;
      
	while( aList != NULL )
        {	
		
		if( funcp ( (char *)aList->item, point ) )
			T_AddObjList( aList->item, p_nouvelle_liste ) ;
		aList = aList->next ;
	}
	T_ReverseList( p_nouvelle_liste ) ;
	T_FreeNPoint( &point ) ;
}



