/* epi_connex.c 2.5 5/14/98 */

#include <epidaure.h>
#include <epidaure.ee>
#include <epidaure.ei>

/*--- calcule les coordonnees (i,j) d'un label l dans une table t ---*/
/* #define _COMPUTE_LABEL(I,J,L,T) { (I) = (int)((L) * (T)->isz); \
                                  (J) = (L) - (I) * (T)->sz  ; } */
/*--- cette premiere solution ne fonctionne pas : il y a parfois des
      erreurs (d'arrondi ?) lors de la multiplication par l'inverse :
      on peut donc essayer la solution "normale" suivante, qui est
      toutefois plus couteuse                                         ---*/
/* #define _COMPUTE_LABEL(I,J,L,T) { (I) = (int)((L) / (T)->sz); \
                                  (J) = (L) - (I) * (T)->sz  ; } */
/*--- cette derniere solution (ci-dessous) semble convenir. Elle semble plus
      rapide que la multiplication par l'inverse, et ne peut pas comporter
      d'erreurs de calculs                                                   ---*/
#define _COMPUTE_LABEL(I,J,L,T) { (I) = 0;   (J) = (L); \
                                  while( (J) >= (T)->sz1000 ) { (I) += 1000;   (J) -= (T)->sz1000; } \
                                  while( (J) >= (T)->sz100 )  { (I) += 100;    (J) -= (T)->sz100;  } \
                                  while( (J) >= (T)->sz10 )   { (I) += 10;     (J) -= (T)->sz10;   } \
                                  while( (J) >= (T)->sz )     { (I) ++ ;       (J) -= (T)->sz;     } }

/*--- valeurs des seuils bas et haut apres seuillage ---*/
static unsigned short int _con_low_thres = 100;
static unsigned short int _con_hig_thres = 200;

/* structure : classe d'equivalence 

   Le champ eq est le numero de la classe d'equivalence.

   Le champ sz est la taille de la classe d'equivalence.

   Si le champ ok est a TRUE, la classe d'equivalence
   est a conserver.
*/
typedef struct _con_ce{
    int eq;
    int sz;
    int ok;
} _con_ce;

/* structure : table d'equivalence 

   Le champ nb est le nombre de labels utilises.

   Le champ sz est la taille utilisee pour les
   allocations.

   Le champ tab est un tableau a 2 dimensions
   contenant les classes d'equivalences.
*/
typedef struct _con_te{
    int nb;
    int sz;
    int sz10;
    int sz100;
    int sz1000;
    double isz;
    _con_ce **tb;
} _con_te;

/*-------Definition des fonctions statiques----------*/
#ifndef NO_PROTO
static int  _CON_Threshold( t_Image *theIm, t_Image *resIm, float t, int v, int flag );
static int  _CON_Convert( t_Image *theIm, t_Image *resIm );
static int  _CON_ComputeCC( t_Image *theIm, e_ParamConnex *par, e_CC_results *res );
static int  _CON_CreateLabel( _con_te *table );
static void _CON_FreeTable( _con_te *table );
static int  _CON_VoisinsPpAlCg( unsigned short int *b, _con_te *t, int *l, int *o, e_ParamConnex *p );
static int  _CON_VoisinsPpAlCe( unsigned short int *b, _con_te *t, int *l, int *o, e_ParamConnex *p );
static int  _CON_VoisinsPpAlCd( unsigned short int *b, _con_te *t, int *l, int *o, e_ParamConnex *p );
static int  _CON_VoisinsApPlCg( unsigned short int *b, _con_te *t, int *l, int *o, e_ParamConnex *p );
static int  _CON_VoisinsApPlCe( unsigned short int *b, _con_te *t, int *l, int *o, e_ParamConnex *p );
static int  _CON_VoisinsApPlCd( unsigned short int *b, _con_te *t, int *l, int *o, e_ParamConnex *p );
static int  _CON_VoisinsApAlCg( unsigned short int *b, _con_te *t, int *l, int *o, e_ParamConnex *p );
static int  _CON_VoisinsApAlCe( unsigned short int *b, _con_te *t, int *l, int *o, e_ParamConnex *p );
static int  _CON_VoisinsApAlCd( unsigned short int *b, _con_te *t, int *l, int *o, e_ParamConnex *p );
static int  _CON_VoisinsApDlCg( unsigned short int *b, _con_te *t, int *l, int *o, e_ParamConnex *p );
static int  _CON_VoisinsApDlCe( unsigned short int *b, _con_te *t, int *l, int *o, e_ParamConnex *p );
static int  _CON_VoisinsApDlCd( unsigned short int *b, _con_te *t, int *l, int *o, e_ParamConnex *p );
static void _CON_Tri_CC( e_CC *tab, int left, int right );
/* static void _CON_Swap_CC( e_CC *tab, int left, int right ); */
#else 
static int  _CON_Threshold();
static int  _CON_Convert();
static int  _CON_ComputeCC();
static int  _CON_CreateLabel();
static void _CON_FreeTable();
static int  _CON_VoisinsPpAlCg();
static int  _CON_VoisinsPpAlCe();
static int  _CON_VoisinsPpAlCd();
static int  _CON_VoisinsApPlCg();
static int  _CON_VoisinsApPlCe();
static int  _CON_VoisinsApPlCd();
static int  _CON_VoisinsApAlCg();
static int  _CON_VoisinsApAlCe();
static int  _CON_VoisinsApAlCd();
static int  _CON_VoisinsApDlCg();
static int  _CON_VoisinsApDlCe();
static int  _CON_VoisinsApDlCd();
static void _CON_Tri_CC();
/* static void _CON_Swap_CC(); */
#endif
/*-------Fin des proto statiques----------*/

/* Procedure d'initialisation des parametres de connexite.

   Le champ lowThres est mis a 0.5.

   Le champ higThres est mis a 0.5.

   Le champ minSize est mis a 1.

   Le champ maxSize est mis a FALSE.

   Le champ maxNbCC est mis a 0, ce qui le desactive.

   Le champ connexite est mis a CONNEXE_26.

   Le champ mode est mis a MODE_3D.

   Le champ verbose est mis a FALSE.
*/

#if defined(_ANSI_)
void E_InitConnexe( e_ParamConnex* par /* parameters structure to be initialized */ )
#else
void E_InitConnexe( par )
e_ParamConnex* par; /* parameters structure to be initialized */
#endif
{
        par->lowThres  = 0.5;
        par->higThres  = 0.5;
        par->minSize   = 1;
        par->maxSize   = FALSE;
	par->maxNbCC   = 0;
        par->connexite = CONNEXE_26;
        par->mode      = MODE_3D;
	par->verbose   = FALSE;
}

/* Procedure de recherche de composantes connexes dans une image.
   
   Le champ lowThres de la structure de parametre permet de considerer 
   l'image d'entree comme binaire : on conserve les points dont la
   valeur est superieure ou egale a ce seuil.

   La connexite utilisee est specifiee par le champ connexite de
   la structure de parametre :

   - CONNEXE_26 : relation 3D, 9 voisins au dessus, 8 dans le plan
                  9 voisins au dessous
   
   - CONNEXE_18 : relation 3D, 5 voisins au dessus, 8 dans le plan
                  5 voisins au dessous
   
   - CONNEXE_10 : relation 3D, 1 voisins au dessus, 8 dans le plan
                  1 voisins au dessous
   
   - CONNEXE_6 : relation 3D, 1 voisins au dessus, 4 dans le plan
                 1 voisins au dessous
   
   - CONNEXE_8 : relation 2D, 8 voisins dans le plan
   
   - CONNEXE_4 : relation 2D, 4 voisins dans le plan

   La taille minimale des composantes connexes a conserver est 
   specifiee par le champ minSize de la structure de parametre.

   Le champ maxNbCC donne le nombre maximal de composantes connexes
   a retenir. Comme elles seront numerotees par ordre de taille
   decroissante, on ne conserve que les maxNbCC plus grandes.

   Le champ mode donne la dimension du traitement. Il peut
   prendre les valeurs MODE_3D ou MODE_2D. S'il est a MODE_2D,
   et que l'image est 3D, chaque plan de l'image est traitee
   comme etant une image independante. Par exemple, si maxNbCC
   est a 1, on conserve une composante connexe par plan.

RETURN
   Retourne FALSE en cas d'erreur. C'est le cas si le type de l'image
   resultat n'est pas suffisant pour coder tous les labels.
*/

#ifdef _ANSI_
int E_Connexe( t_ImagePtr theIm /* image where connected components are to be labeled */, 
	      t_ImagePtr resIm /* result image */, 
	      e_ParamConnex *par /* parameters */ )
#else
int E_Connexe( theIm, resIm, par )
t_ImagePtr theIm;     /* image where connected components are to be labeled */
t_ImagePtr resIm;     /* result image */
e_ParamConnex *par; /* parameters */
#endif
{
    e_CC_results res;

    res.nb = 0;
    res.bool_alloc = FALSE;
    res.tab = (e_CC*)NULL;

    return( E_ConnexePlus( theIm, resIm, par, &res ) );
}

/* Procedure (interne) de recherche de composantes connexes dans une image.
   
   Le champ lowThres de la structure de parametre permet de considerer 
   l'image d'entree comme binaire : on conserve les points dont la
   valeur est superieure ou egale a ce seuil.

   La connexite utilisee est specifiee par le champ connexite de
   la structure de parametre :

   - CONNEXE_26 : relation 3D, 9 voisins au dessus, 8 dans le plan
                  9 voisins au dessous
   
   - CONNEXE_18 : relation 3D, 5 voisins au dessus, 8 dans le plan
                  5 voisins au dessous
   
   - CONNEXE_10 : relation 3D, 1 voisins au dessus, 8 dans le plan
                  1 voisins au dessous
   
   - CONNEXE_6 : relation 3D, 1 voisins au dessus, 4 dans le plan
                 1 voisins au dessous
   
   - CONNEXE_8 : relation 2D, 8 voisins dans le plan
   
   - CONNEXE_4 : relation 2D, 4 voisins dans le plan

   La taille minimale des composantes connexes a conserver est 
   specifiee par le champ minSize de la structure de parametre.

   Le champ maxNbCC donne le nombre maximal de composantes connexes
   a retenir. Comme elles seront numerotees par ordre de taille
   decroissante, on ne conserve que les maxNbCC plus grandes.

   Le champ mode donne la dimension du traitement. Il peut
   prendre les valeurs MODE_3D ou MODE_2D. S'il est a MODE_2D,
   et que l'image est 3D, chaque plan de l'image est traitee
   comme etant une image independante. Par exemple, si maxNbCC
   est a 1, on conserve une composante connexe par plan.

   La structure de resultats supplementaires res contient les champs
   suivants :

   nb : nombre de composantes connexes valides

   bool_alloc : si TRUE, alors le tableau tab est alloue et rempli

   tab : tableau de type e_CC* de (nb+1) elements ;
         tab[i].size donne la taille de la composante connexe numero i

   Cette structure n'est pas remplie si le traitement est 2D
   et que l'image a plus d'un plan.

   Avant appel, il faut mettre le champ bool_alloc a TRUE ou FALSE,
   et etre sur que le tableau tab ne contiennent pas une adresse
   valide (qui risque alors d'etre perdue).

   Apres appel, et si bool_alloc etait a TRUE, il ne faut pas oublier
   de desallouer tab ( T_Free( (void**)&(res.tab) ) ).

RETURN
   Retourne FALSE en cas d'erreur. C'est le cas si le type de l'image
   resultat n'est pas suffisant pour coder tous les labels.
*/

#ifdef _ANSI_
int E_ConnexePlus( t_Image *theIm /* image where connected components are to be labeled */, 
		   t_Image *resIm /* result image */, 
		   e_ParamConnex *par /* parameters */,
		   e_CC_results *res /* additional results */ )
#else
int E_ConnexePlus( theIm, resIm, par, res )
t_Image *theIm;     /* image where connected components are to be labeled */
t_Image *resIm;     /* result image */
e_ParamConnex *par; /* parameters */
e_CC_results *res;  /* additional results */
#endif
{
    t_Image auxIm, auxIm2D;
    register int i;
    register unsigned short int *in;
    int x, y, z;
    e_ParamConnex local_par;
    e_CC_results local_res;
    char message[256];

    E_ERR( (theIm==NULL) || (theIm->buf==NULL), E_ERRNULL, FALSE,
	   "E_ConnexePlus", "input image not allocated");
    E_ERR( (resIm==NULL) || (resIm->buf==NULL), E_ERRNULL, FALSE,
	   "E_ConnexePlus", "output image not allocated");
    E_ERR( (theIm->tx!=resIm->tx) || (theIm->ty!=resIm->ty) || (theIm->tz!=resIm->tz),
	   E_ERRFATAL, FALSE, "E_ConnexePlus", 
	   "input and output have different dimensions");
    E_ERR( (theIm->tx<2) || (theIm->ty<2),
	   E_ERRFATAL, FALSE, "E_ConnexePlus",
	   "input has too little dimensions in X or Y");

    /*--- initialisation ---*/
    local_par = *par;
    if ( par->maxSize == TRUE ) local_par.maxNbCC = 1;
    local_res.nb = res->nb = 0;
    local_res.bool_alloc = ( (par->verbose == TRUE) || (res->bool_alloc == TRUE) ) ? TRUE : FALSE;
    local_res.tab = (e_CC*)NULL;

    /*--- preparation de l'image intermediaire ---*/
    T_Init3Image( &auxIm, theIm->tx, theIm->ty, theIm->tz, (int)TYPEIM_16B, CNULL );
    T_AllocImage( &auxIm );
    if ( E_Error(E_ERRNO,(char*)NULL,(char*)NULL) ) {
	E_SETERR( E_ERRFATAL, "E_ConnexePlus", "can not allocate auxiliary image" );
        return( FALSE );
    }
    if ( _CON_Threshold( theIm, &auxIm, (float)(par->lowThres),
			 (int)_con_hig_thres, (int)TRUE ) != TRUE ) {
	E_SETERR( E_ERRFATAL, "E_ConnexePlus", "pre-computing failed" );
	T_FreeImage( &auxIm );
	return( FALSE );
    }

    /*--- calcul proprement dit ---*/
    if ( (par->mode == MODE_2D) && (theIm->tz > 1) ) {
	for ( z = 0; z < theIm->tz; z++ ) {
	    T_Init3Image( &auxIm2D, auxIm.tx, auxIm.ty, (int)1, auxIm.type, CNULL );	
	    auxIm2D.buf = auxIm.buf + z * auxIm.tx * auxIm.ty * sizeof(unsigned short int);
	    /*--- calcul ---*/
	    if ( _CON_ComputeCC( &auxIm2D, &local_par, &local_res ) != TRUE ) {
		sprintf( message, "unable to extract connected components in slice %d", z );
		E_SETERR( E_ERRFATAL, "E_ConnexePlus", message );
		T_FreeImage( &auxIm );
		if ( local_res.tab != (e_CC*)NULL ) T_Free( (void**)(&local_res.tab) );
		return( FALSE );
	    }
	    /*--- verification ----*/
	    if ( ((resIm->type == TYPEIM_256) && (local_res.nb > 255)) ||
		 ((resIm->type == TYPEIM_S16B) && (local_res.nb > 32767)) ) {
		sprintf( message, "too much connected components for output image in slice %d", z );
		E_SETERR( E_ERRFATAL, "E_ConnexePlus", message );
		T_FreeImage( &auxIm );
		if ( local_res.tab != (e_CC*)NULL ) T_Free( (void**)(&local_res.tab) );
		return( FALSE );
	    }
	    /*--- ecriture ---*/
	    if ( par->verbose == TRUE ) {
		fprintf( stderr, "IMAGE %s slice %d : %5d valid connected components\n", theIm->fi.name, z, local_res.nb );
		if ( local_res.tab != (e_CC*)NULL ) {
		    for ( i = 1; i <= local_res.nb ; i++ ) 
			fprintf( stderr, "      the connected component # %5d contains %10d points\n", i, local_res.tab[i].size );
		}
	    }
	    if ( local_res.tab != (e_CC*)NULL ) T_Free( (void**)(&local_res.tab) );
	    local_res.nb = 0;
	}
    }
    else {
	/*--- calcul ---*/
	if ( _CON_ComputeCC( &auxIm, &local_par, &local_res ) != TRUE ) {
	    E_SETERR( E_ERRFATAL, "E_ConnexePlus", "unable to extract connected components" );
	    T_FreeImage( &auxIm );
	    if ( local_res.tab != (e_CC*)NULL ) T_Free( (void**)(&local_res.tab) );
	    return( FALSE );
	}
	/*--- verification ----*/
	if ( ((resIm->type == TYPEIM_256) && (local_res.nb > 255)) ||
	     ((resIm->type == TYPEIM_S16B) && (local_res.nb > 32767)) ) {
	    E_SETERR( E_ERRFATAL, "E_ConnexePlus", "too much connected components for output image" );
	    T_FreeImage( &auxIm );
	    if ( local_res.tab != (e_CC*)NULL ) T_Free( (void**)(&local_res.tab) );
	    return( FALSE );
	}
	/*--- ecriture ---*/
	if ( par->verbose == TRUE ) {
	    fprintf( stderr, "IMAGE %s : %5d valid connected components\n", theIm->fi.name, local_res.nb );
	    if ( local_res.tab != (e_CC*)NULL ) {
		for ( i = 1; i <= local_res.nb ; i++ ) 
		    fprintf( stderr, "      the connected component # %5d contains %10d points\n", i, local_res.tab[i].size );
	    }
	}
	res->nb = local_res.nb;
	if ( local_res.tab != (e_CC*)NULL ) {
	    if ( res->bool_alloc == TRUE ) res->tab = local_res.tab;
	    else T_Free( (void**)(&local_res.tab) ); 
	}
	local_res.nb = 0;
    }

    /*--- ecriture de l'image resultat ---*/
    z = resIm->tx * resIm->ty * resIm->tz;
    in = (unsigned short int*)(auxIm.buf);
    
    if ( par->maxSize == TRUE ) {
	if ( _CON_Convert( &auxIm, resIm ) != TRUE ) {
	    E_SETERR( E_ERRFATAL, "E_ConnexePlus", "unable to convert auxiliary image" );
	    T_FreeImage( &auxIm );
	    return( FALSE );
	}
    }
    else {
	switch( resIm->type ) {
	case TYPEIM_256 :
            {
	    unsigned char *out;
	    out = (unsigned char*)(resIm->buf);
	    for ( i = 0; i < z; i++ )
		*out++ = (unsigned char)(*in++);
	    }
	    break;
	case TYPEIM_S16B :
            {
	    short int *out;
	    out = (short int*)(resIm->buf);
	    for ( i = 0; i < z; i++ )
		*out++ = (short int)(*in++);
	    }
	    break;
	case TYPEIM_16B :
            {
	    unsigned short int *out;
	    out = (unsigned short int*)(resIm->buf);
	    for ( i = 0; i < z; i++ )
		*out++ = (unsigned short int)(*in++);
	    }
	    break;
	case TYPEIM_FLOAT :
            {
            float *out;
	    out = (float*)(resIm->buf);
	    for ( i = 0; i < z; i++ )
		*out++ = (float)(*in++);
	    }
	    break;
	default :
	    for ( z = 0; z < resIm->tz; z++ )
	    for ( y = 0; y < resIm->ty; y++ )
            for ( x = 0; x < resIm->tx; x++ ) {
		T_Set3Point( x, y, z, resIm, (reeltype)(*in++), MODE_SET );
	    }
	}
    }
	
    T_FreeImage( &auxIm );
    return( TRUE );
}

/* Procedure de recherche de la plus grande composante connexe dans une image.
   
   Le champ lowThres de la structure de parametre permet de considerer 
   l'image d'entree comme binaire : on conserve les points dont la
   valeur est superieure ou egale a ce seuil.

   La connexite utilisee est specifiee par le champ connexite de
   la structure de parametre :

   - CONNEXE_26 : relation 3D, 9 voisins au dessus, 8 dans le plan
                  9 voisins au dessous
   
   - CONNEXE_18 : relation 3D, 5 voisins au dessus, 8 dans le plan
                  5 voisins au dessous
   
   - CONNEXE_10 : relation 3D, 1 voisins au dessus, 8 dans le plan
                  1 voisins au dessous
   
   - CONNEXE_6 : relation 3D, 1 voisins au dessus, 4 dans le plan
                 1 voisins au dessous
   
   - CONNEXE_8 : relation 2D, 8 voisins dans le plan
   
   - CONNEXE_4 : relation 2D, 4 voisins dans le plan

   La taille minimale des composantes connexes a examiner est 
   specifiee par le champ minSize de la structure de parametre.
   La plus grande composante connexe est recherchee parmi celles-ci.

   Le champ mode donne la dimension du traitement. Il peut
   prendre les valeurs MODE_3D ou MODE_2D. S'il est a MODE_2D,
   et que l'image est 3D, chaque plan de l'image est traitee
   comme etant une image independante. 
   On conserve donc une composante connexe par plan.

RETURN
   Retourne FALSE en cas d'erreur.
*/

#ifdef _ANSI_
int E_ConnexeMax( t_ImagePtr theIm /* image where the greatest connected components is to be extracted */, 
	      t_ImagePtr resIm /* result image */, 
	      e_ParamConnex *par /* parameters */ )
#else
int E_ConnexeMax( theIm, resIm, par )
t_ImagePtr theIm;     /* image where the greatest connected components is to be extracted */
t_ImagePtr resIm;     /* result image */
e_ParamConnex *par; /* parameters */
#endif
{
    e_ParamConnex local_par;

    local_par = *par;
    local_par.maxSize = TRUE;
    local_par.maxNbCC = 1;

    return( E_Connexe( theIm, resIm, &local_par ) );
}

/* Procedure dite de reconstruction d'une image par rapport a une autre.
   
   Le champ lowThres de la structure de parametre permet de considerer 
   les images d'entree comme binaire : on conserve les points dont la
   valeur est superieure ou egale a ce seuil.

   La reconstruction est une selection des composantes connexes
   de l'image theIm contenant au moins un point de l'image supIm.

   Les points appartenant a l'image supIm, mais n'appartenant pas
   a l'image theIm, sont comptes egalement (s'ils existent). Cela
   revient donc a selectionner les composantes connexes de l'union
   des deux images (theIm et supIm) contenant au moins un point 
   de l'image supIm.

   La connexite utilisee est specifiee par le champ connexite de
   la structure de parametre :

   - CONNEXE_26 : relation 3D, 9 voisins au dessus, 8 dans le plan
                  9 voisins au dessous
   
   - CONNEXE_18 : relation 3D, 5 voisins au dessus, 8 dans le plan
                  5 voisins au dessous
   
   - CONNEXE_10 : relation 3D, 1 voisins au dessus, 8 dans le plan
                  1 voisins au dessous
   
   - CONNEXE_6 : relation 3D, 1 voisins au dessus, 4 dans le plan
                 1 voisins au dessous
   
   - CONNEXE_8 : relation 2D, 8 voisins dans le plan
   
   - CONNEXE_4 : relation 2D, 4 voisins dans le plan

   La taille minimale des composantes connexes a conserver est 
   specifiee par le champ minSize de la structure de parametre.

   Le champ maxNbCC donne le nombre maximal de composantes connexes
   a retenir. Comme elles seront numerotees par ordre de taille
   decroissante, on ne conserve que les maxNbCC plus grandes.

   Le champ mode donne la dimension du traitement. Il peut
   prendre les valeurs MODE_3D ou MODE_2D. S'il est a MODE_2D,
   et que l'image est 3D, chaque plan de l'image est traitee
   comme etant une image independante. Par exemple, si maxNbCC
   est a 1, on conserve une composante connexe par plan.

RETURN
   Retourne FALSE en cas d'erreur. 
*/

#ifdef _ANSI_
int E_Reconstruct( t_Image *theIm /* image where connected components are to be searched */, 
		   t_Image *supIm /* image containing valid parts */,
		   t_Image *resIm /* result image */, 
		   e_ParamConnex *par /* parameters */ )
#else
int E_Reconstruct( theIm, resIm, supIm, par )
t_Image *theIm;     /* image where connected components are to be searched */
t_Image *supIm;     /* image containing valid parts */
t_Image *resIm;     /* result image */
e_ParamConnex *par; /* parameters */
#endif
{
    t_Image auxIm, auxIm2D;
    register int i;
    int z;
    e_ParamConnex local_par;
    e_CC_results res;
    char message[256];

    E_ERR( (theIm==NULL) || (theIm->buf==NULL), E_ERRNULL, FALSE,
	   "E_Reconstruct", "first input image not allocated");
    E_ERR( (supIm==NULL) || (supIm->buf==NULL), E_ERRNULL, FALSE,
	   "E_Reconstruct", "second input image not allocated");
    E_ERR( (resIm==NULL) || (resIm->buf==NULL), E_ERRNULL, FALSE,
	   "E_Reconstruct", "output image not allocated");
    E_ERR( (theIm->tx!=resIm->tx) || (theIm->ty!=resIm->ty) || (theIm->tz!=resIm->tz),
	   E_ERRFATAL, FALSE, "E_Reconstruct", 
	   "first input and output have different dimensions");
    E_ERR( (supIm->tx!=resIm->tx) || (supIm->ty!=resIm->ty) || (supIm->tz!=resIm->tz),
	   E_ERRFATAL, FALSE, "E_Reconstruct", 
	   "first input and output have different dimensions");
    E_ERR( (theIm->tx<2) || (theIm->ty<2),
	   E_ERRFATAL, FALSE, "E_Reconstruct",
	   "input has too little dimensions in X or Y");

    /*--- initialisation ---*/
    local_par = *par;
    if ( par->maxSize == TRUE ) local_par.maxNbCC = 1;
    res.nb = 0;
    res.bool_alloc = ( par->verbose == TRUE ) ? TRUE : FALSE;
    res.tab = (e_CC*)NULL;

    /*--- preparation de l'image intermediaire ---*/
    T_Init3Image( &auxIm, theIm->tx, theIm->ty, theIm->tz, (int)TYPEIM_16B, CNULL );
    T_AllocImage( &auxIm );
    if ( E_Error(E_ERRNO,(char*)NULL,(char*)NULL) ) {
	E_SETERR( E_ERRFATAL, "E_Reconstruct", "can not allocate auxiliary image" );
        return( FALSE );
    }
    if ( _CON_Threshold( theIm, &auxIm, (float)(par->lowThres),
			 (int)_con_low_thres, (int)TRUE ) != TRUE ) {
	E_SETERR( E_ERRFATAL, "E_Reconstruct", "pre-computing failed (first input image)" );
	T_FreeImage( &auxIm );
	return( FALSE );
    }
    if ( _CON_Threshold( supIm, &auxIm, (float)(par->lowThres),
			 (int)_con_hig_thres, (int)FALSE ) != TRUE ) {
	E_SETERR( E_ERRFATAL, "E_Reconstruct", "pre-computing failed (second input image)" );
	T_FreeImage( &auxIm );
	return( FALSE );
    }

    /*--- calcul proprement dit ---*/
    if ( (par->mode == MODE_2D) && (theIm->tz > 1) ) {
	for ( z = 0; z < theIm->tz; z++ ) {
	    T_Init3Image( &auxIm2D, auxIm.tx, auxIm.ty, (int)1, auxIm.type, CNULL );	
	    auxIm2D.buf = auxIm.buf + z * auxIm.tx * auxIm.ty * sizeof(unsigned short int);
	    /*--- calcul ---*/
	    if ( _CON_ComputeCC( &auxIm2D, &local_par, &res ) != TRUE ) {
		sprintf( message, "unable to extract connected components in slice %d", z );
		E_SETERR( E_ERRFATAL, "E_Reconstruct", message );
		T_FreeImage( &auxIm );
		if ( res.tab != (e_CC*)NULL ) T_Free( (void**)(&res.tab) );
		return( FALSE );
	    }
	    /*--- ecriture ---*/
	    if ( par->verbose == TRUE ) {
		fprintf( stderr, "IMAGE %s slice %d : %5d valid connected components\n", theIm->fi.name, z, res.nb );
		if ( res.tab != (e_CC*)NULL ) {
		    for ( i = 1; i <= res.nb ; i++ ) 
			fprintf( stderr, "      the connected component # %5d contains %10d points\n", i, res.tab[i].size );
		}
	    }
	    if ( res.tab != (e_CC*)NULL ) T_Free( (void**)(&res.tab) );
	    res.nb = 0;
	}
    }
    else {
	/*--- calcul ---*/
	if ( _CON_ComputeCC( &auxIm, &local_par, &res ) != TRUE ) {
	    E_SETERR( E_ERRFATAL, "E_Reconstruct", "unable to extract connected components" );
	    T_FreeImage( &auxIm );
	    if ( res.tab != (e_CC*)NULL ) T_Free( (void**)(&res.tab) );
	    return( FALSE );
	}
	/*--- ecriture ---*/
	if ( par->verbose == TRUE ) {
	    fprintf( stderr, "IMAGE %s : %5d valid connected components\n", theIm->fi.name, res.nb );
	    if ( res.tab != (e_CC*)NULL ) {
		for ( i = 1; i <= res.nb ; i++ ) 
		    fprintf( stderr, "      the connected component # %5d contains %10d points\n", i, res.tab[i].size );
	    }
	}
	if ( res.tab != (e_CC*)NULL ) T_Free( (void**)(&res.tab) );
	res.nb = 0;
    }

    /*--- ecriture de l'image resultat ---*/
    if ( _CON_Convert( &auxIm, resIm ) != TRUE ) {
	E_SETERR( E_ERRFATAL, "E_Reconstruct", "unable to convert auxiliary image" );
	T_FreeImage( &auxIm );
	return( FALSE );
    }
	
    T_FreeImage( &auxIm );
    return( TRUE );
}

/* Fonction de seuillage par hysteresis.

   Le seuillage par hysteresis est l'extraction de composantes connexes
   qui verifient :

   - tous les points ont une valeur superieure a un seuil bas 
     (champ lowThres de la structure de parametre)

   - chaque composante connexe contient au moins un point dont la valeur 
     est superieure a un seuil haut 
     (champ higThres de la structure de parametre)

   La connexite utilisee est specifiee par le champ connexite de
   la structure de parametre :

   - CONNEXE_26 : relation 3D, 9 voisins au dessus, 8 dans le plan
                  9 voisins au dessous
   
   - CONNEXE_18 : relation 3D, 5 voisins au dessus, 8 dans le plan
                  5 voisins au dessous
   
   - CONNEXE_10 : relation 3D, 1 voisins au dessus, 8 dans le plan
                  1 voisins au dessous
   
   - CONNEXE_6 : relation 3D, 1 voisins au dessus, 4 dans le plan
                 1 voisins au dessous
   
   - CONNEXE_8 : relation 2D, 8 voisins dans le plan
   
   - CONNEXE_4 : relation 2D, 4 voisins dans le plan

   La taille minimale des composantes connexes a conserver est 
   specifiee par le champ minSize de la structure de parametre.

   Le champ maxNbCC donne le nombre maximal de composantes connexes
   a retenir. Comme elles seront numerotees par ordre de taille
   decroissante, on ne conserve que les maxNbCC plus grandes.

   Le champ mode donne la dimension du traitement. Il peut
   prendre les valeurs MODE_3D ou MODE_2D. S'il est a MODE_2D,
   et que l'image est 3D, chaque plan de l'image est traitee
   comme etant une image independante. Par exemple, si maxNbCC
   est a 1, on conserve une composante connexe par plan.

RETURN
   Retourne FALSE en cas d'erreur. 
*/

#ifdef _ANSI_
int E_Hysteresis( t_Image *theIm /* image to be thresholded */, 
		   t_Image *resIm /* result image */, 
		   e_ParamConnex *par /* parameters */ )
#else
int E_Hysteresis( theIm, resIm, par )
t_Image *theIm;     /* image to be thresholded */
t_Image *resIm;     /* result image */
e_ParamConnex *par; /* parameters */
#endif
{
    t_Image auxIm, auxIm2D;
    register int i;
    register unsigned short int *out;
    int x, y, z, sb, sh;
    e_ParamConnex local_par;
    e_CC_results res;
    char message[256];

    E_ERR( (theIm==NULL) || (theIm->buf==NULL), E_ERRNULL, FALSE,
	   "E_Hysteresis", "first input image not allocated");
    E_ERR( (resIm==NULL) || (resIm->buf==NULL), E_ERRNULL, FALSE,
	   "E_Hysteresis", "output image not allocated");
    E_ERR( (theIm->tx!=resIm->tx) || (theIm->ty!=resIm->ty) || (theIm->tz!=resIm->tz),
	   E_ERRFATAL, FALSE, "E_Hysteresis", 
	   "input and output have different dimensions");
    E_ERR( (theIm->tx<2) || (theIm->ty<2),
	   E_ERRFATAL, FALSE, "E_Hysteresis",
	   "input has too little dimensions in X or Y");

    /*--- initialisation ---*/
    local_par = *par;
    if ( par->maxSize == TRUE ) local_par.maxNbCC = 1;
    res.nb = 0;
    res.bool_alloc = ( par->verbose == TRUE ) ? TRUE : FALSE;
    res.tab = (e_CC*)NULL;

    /*--- preparation de l'image intermediaire ---*/
    T_Init3Image( &auxIm, theIm->tx, theIm->ty, theIm->tz, (int)TYPEIM_16B, CNULL );
    T_AllocImage( &auxIm );
    if ( E_Error(E_ERRNO,(char*)NULL,(char*)NULL) ) {
	E_SETERR( E_ERRFATAL, "E_Hysteresis", "can not allocate auxiliary image" );
        return( FALSE );
    }
    z = theIm->tx * theIm->ty * theIm->tz;
    out = (unsigned short int*)(auxIm.buf);
    switch ( theIm->type ) {
    case TYPEIM_256 :
        {
	unsigned char *in;
	in = (unsigned char*)(theIm->buf);
	sb = (int)(par->lowThres * 255.0 + 0.5);
	sh = (int)(par->higThres * 255.0 + 0.5);
	for ( i = 0; i < z; i++ ) {
	    if ( *in < sb ) { *out++ = (unsigned short int)0;   in++;   continue; }
	    if ( *in < sh ) { *out++ = _con_low_thres;          in++;   continue; }
	    *out++ = _con_hig_thres;          
	    in++;
	}}
	break;
    case TYPEIM_16B :
        {
	unsigned short int *in;
	in = (unsigned short int*)(theIm->buf);
	sb = (int)(par->lowThres * 65535.0 + 0.5);
	sh = (int)(par->higThres * 65535.0 + 0.5);
	for ( i = 0; i < z; i++ ) {
	    if ( *in < sb ) { *out++ = (unsigned short int)0;   in++;   continue; }
	    if ( *in < sh ) { *out++ = _con_low_thres;          in++;   continue; }
	    *out++ = _con_hig_thres;          
	    in++;
	}}
	break;
    case TYPEIM_S16B :
        {
	short int *in;
	in = (short int*)(theIm->buf);
	sb = (par->lowThres < 0) ? (int)(par->lowThres * 32767.0 - 0.5) : (int)(par->lowThres * 32767.0 + 0.5);
	sh = (par->higThres < 0) ? (int)(par->higThres * 32767.0 - 0.5) : (int)(par->higThres * 32767.0 + 0.5);
	for ( i = 0; i < z; i++ ) {
	    if ( *in < sb ) { *out++ = (unsigned short int)0;   in++;   continue; }
	    if ( *in < sh ) { *out++ = _con_low_thres;          in++;   continue; }
	    *out++ = _con_hig_thres;          
	    in++;
	}}
	break;
    case TYPEIM_FLOAT :
        {
	float *in;
	in = (float*)(theIm->buf);
	for ( i = 0; i < z; i++ ) {
	    if ( *in < par->lowThres ) { *out++ = (unsigned short int)0;   in++;   continue; }
	    if ( *in < par->higThres ) { *out++ = _con_low_thres;          in++;   continue; }
	    *out++ = _con_hig_thres;          
	    in++;
	}}
	break;
    default :
        {
	reeltype val;    
	for ( z = 0; z < resIm->tz; z++ )
        for ( y = 0; y < resIm->ty; y++ )
        for ( x = 0; x < resIm->tx; x++ ) {
	    val = T_Get3Point( x, y, z, theIm );
	    if ( val < par->lowThres ) { *out++ = (unsigned short int)0;   continue; }
	    if ( val < par->higThres ) { *out++ = _con_low_thres;          continue; }
	    *out++ = _con_hig_thres;          
	}}
    }

    /*--- calcul proprement dit ---*/
    if ( (par->mode == MODE_2D) && (theIm->tz > 1) ) {
	for ( z = 0; z < theIm->tz; z++ ) {
	    T_Init3Image( &auxIm2D, auxIm.tx, auxIm.ty, (int)1, auxIm.type, CNULL );	
	    auxIm2D.buf = auxIm.buf + z * auxIm.tx * auxIm.ty * sizeof(unsigned short int);
	    /*--- calcul ---*/
	    if ( _CON_ComputeCC( &auxIm2D, &local_par, &res ) != TRUE ) {
		sprintf( message, "unable to extract connected components in slice %d", z );
		E_SETERR( E_ERRFATAL, "E_Hysteresis", message );
		T_FreeImage( &auxIm );
		if ( res.tab != (e_CC*)NULL ) T_Free( (void**)(&res.tab) );
		return( FALSE );
	    }
	    /*--- ecriture ---*/
	    if ( par->verbose == TRUE ) {
		fprintf( stderr, "IMAGE %s slice %d : %5d valid connected components\n", theIm->fi.name, z, res.nb );
		if ( res.tab != (e_CC*)NULL ) {
		    for ( i = 1; i <= res.nb ; i++ ) 
			fprintf( stderr, "      the connected component # %5d contains %10d points\n", i, res.tab[i].size );
		}
	    }
	    if ( res.tab != (e_CC*)NULL ) T_Free( (void**)(&res.tab) );
	    res.nb = 0;
	}
    }
    else {
	/*--- calcul ---*/
	if ( _CON_ComputeCC( &auxIm, &local_par, &res ) != TRUE ) {
	    E_SETERR( E_ERRFATAL, "E_Hysteresis", "unable to extract connected components" );
	    T_FreeImage( &auxIm );
	    if ( res.tab != (e_CC*)NULL ) T_Free( (void**)(&res.tab) );
	    return( FALSE );
	}
	/*--- ecriture ---*/
	if ( par->verbose == TRUE ) {
	    fprintf( stderr, "IMAGE %s : %5d valid connected components\n", theIm->fi.name, res.nb );
	    if ( res.tab != (e_CC*)NULL ) {
		for ( i = 1; i <= res.nb ; i++ ) 
		    fprintf( stderr, "      the connected component # %5d contains %10d points\n", i, res.tab[i].size );
	    }
	}
	if ( res.tab != (e_CC*)NULL ) T_Free( (void**)(&res.tab) );
	res.nb = 0;
    }

    /*--- ecriture de l'image resultat ---*/
    if ( _CON_Convert( &auxIm, resIm ) != TRUE ) {
	E_SETERR( E_ERRFATAL, "E_Hysteresis", "unable to convert auxiliary image" );
	T_FreeImage( &auxIm );
	return( FALSE );
    }
	
    T_FreeImage( &auxIm );
    return( TRUE );
}

/* Fonction de seuillage adaptee a la recherche de composantes connexes.

   Cette fonction seuille l'image d'entree theIm avec le seuil
   flottant t (qui sera normalise selon les conventions de
   T_Get3Point et T_Set3Point pour les images entieres).

   Les points superieurs ou egaux a ce seuil sont mis a
   la valeur v dans l'image resultat resIm (obligatoirement
   de type TYPEIM_16B).

   Si la valeur flag est a TRUE, les autres points sont mis a 0,
   sinon ils restent inchanges.
*/

#if defined(_ANSI_)
static int _CON_Threshold( t_Image *theIm, t_Image *resIm, float t, int v, int flag )
#else
static int _CON_Threshold( theIm, resIm, t, v, flag )
t_Image *theIm;
t_Image *resIm;
float t;
int v;
int flag;
#endif
{
    register int i, is=1;
    int x, y, z;
    register float fs;
    register unsigned short int *res;
    register unsigned short int iv;
    int dxyz;

    E_ERR( (theIm==NULL) || (theIm->buf==NULL), E_ERRNULL, FALSE,
	   "_CON_Threshold", "input image not allocated");
    E_ERR( (resIm==NULL) || (resIm->buf==NULL), E_ERRNULL, FALSE,
	   "_CON_Threshold", "output image not allocated");
    E_ERR( (theIm->tx!=resIm->tx) || (theIm->ty!=resIm->ty) || (theIm->tz!=resIm->tz),
	   E_ERRFATAL, FALSE, "_CON_Threshold", 
	   "input and output have different dimensions");
    E_ERR( (resIm->type != TYPEIM_16B), E_ERRNULL, FALSE,
	   "_CON_Threshold", "output image has bad type");

    /*--- initialisations ---*/
    fs = t;
    iv = (unsigned short int)v;
    switch ( theIm->type ) {
    case TYPEIM_256 :
	if ( fs <= 0.0 )     is = 0;
	else if ( fs > 1.0 ) is = 255;
	else                 is = (int)(fs * 255.0 + 0.5);
	break;
    case TYPEIM_16B :
	if ( fs <= 0.0 )     is = 0;
	else if ( fs > 1.0 ) is = 65535;
	else                 is = (int)(fs * 65535.0 + 0.5);
	break;
    case TYPEIM_S16B :
	if ( fs > 1.0 ) is = 32767;
	else  is = ( fs < 0 ) ? (int)(fs * 32767.0 - 0.5) : (int)(fs * 32767.0 + 0.5);
    }
    dxyz = theIm->tx * theIm->ty * theIm->tz;
    res = (unsigned short int*)(resIm->buf);

    /*--- seuillage ----*/
    if ( flag == TRUE ) {
	switch ( theIm->type ) {
	case TYPEIM_256 :
            { 
	    unsigned char *buf;
	    buf = (unsigned char*)(theIm->buf);
	    for (i = 0; i < dxyz; i++) {
		if ( *buf++ >= is ) *res++ = iv;
		else                *res++ = 0;
	    }}
	    break;
	case TYPEIM_16B :
            {
	    unsigned short int *buf;
	    buf = (unsigned short int*)(theIm->buf);
	    for (i = 0; i < dxyz; i++) {
		if ( *buf++ >= is ) *res++ = iv;
		else                *res++ = 0;
	    }}
	    break;
	case TYPEIM_S16B :
            {
	    short int *buf;
	    buf = (short int*)(theIm->buf);
	    for (i = 0; i < dxyz; i++) {
		if ( *buf++ >= is ) *res++ = iv;
		else                *res++ = 0;
	    }}
	    break;
	case TYPEIM_FLOAT :
            {
	    float *buf;
	    buf = (float*)(theIm->buf);
	    for (i = 0; i < dxyz; i++) {
		if ( *buf++ >= fs ) *res++ = iv;
		else                *res++ = 0;
	    }}
	    break;
	default :
	    for ( z = 0; z < theIm->tz; z ++ )
            for ( y = 0; y < theIm->ty; y ++ )
            for ( x = 0; x < theIm->tx; x ++ ) {
		if ( T_Get3Point( x, y, z, theIm ) >= fs ) *res++ = iv;
		else                                       *res++ = 0;
	    }
	}
    }
    else {
	switch ( theIm->type ) {
	case TYPEIM_256 :
            { 
	    unsigned char *buf;
	    buf = (unsigned char*)(theIm->buf);
	    for (i = 0; i < dxyz; i++) {
		if ( *buf++ >= is ) *res++ = iv;
		else res++;
	    }}
	    break;
	case TYPEIM_16B :
            {
	    unsigned short int *buf;
	    buf = (unsigned short int*)(theIm->buf);
	    for (i = 0; i < dxyz; i++) {
		if ( *buf++ >= is ) *res++ = iv;
		else res++;
	    }}
	    break;
	case TYPEIM_S16B :
            {
	    short int *buf;
	    buf = (short int*)(theIm->buf);
	    for (i = 0; i < dxyz; i++) {
		if ( *buf++ >= is ) *res++ = iv;
		else res++;
	    }}
	    break;
	case TYPEIM_FLOAT :
            {
	    float *buf;
	    buf = (float*)(theIm->buf);
	    for (i = 0; i < dxyz; i++) {
		if ( *buf++ >= fs ) *res++ = iv;
		else res++;
	    }}
	    break;
	default :
	    for ( z = 0; z < theIm->tz; z ++ )
            for ( y = 0; y < theIm->ty; y ++ )
            for ( x = 0; x < theIm->tx; x ++ ) {
		if ( T_Get3Point( x, y, z, theIm ) >= fs ) *res++ = iv;
		else res++;
	    }
	}
    }
    return( TRUE );
}

/* Fonction de conversion adaptee aux operations de connexites.

   Si dans l'image d'entree (image des composantes connexes
   numerotees), la valeur est differente de 0, alors
   elle est mise a 1.0 dans l'image resultat, sinon elle
   a mise a 0.0.
*/

#if defined(_ANSI_)
static int _CON_Convert( t_Image *theIm, t_Image *resIm )
#else
static int _CON_Convert( theIm, resIm )
t_Image *theIm;
t_Image *resIm;
#endif
{
    register int i;
    int x, y, z;
    register unsigned short int *in;

    E_ERR( (theIm==NULL) || (theIm->buf==NULL), E_ERRNULL, FALSE,
	   "_CON_Convert", "input image not allocated");
    E_ERR( (resIm==NULL) || (resIm->buf==NULL), E_ERRNULL, FALSE,
	   "_CON_Convert", "output image not allocated");
    E_ERR( (theIm->tx!=resIm->tx) || (theIm->ty!=resIm->ty) || (theIm->tz!=resIm->tz),
	   E_ERRFATAL, FALSE, "_CON_Convert", 
	   "input and output have different dimensions");
    E_ERR( (theIm->type != TYPEIM_16B), E_ERRNULL, FALSE,
	   "_CON_Convert", "input image has bad type");

    z = resIm->tx * resIm->ty * resIm->tz;
    in = (unsigned short int*)(theIm->buf);
    switch( resIm->type ) {
    case TYPEIM_256 :
        {
	unsigned char *out;
	out = (unsigned char*)(resIm->buf);
	for ( i = 0; i < z; i++ )
	    *out++ = ( *in++ > 0 ) ? (unsigned char)255 : (unsigned char)0;
	}
	break;
    case TYPEIM_S16B :
        {
        short int *out;
	out = (short int*)(resIm->buf);
	for ( i = 0; i < z; i++ )
	    *out++ = ( *in++ > 0 ) ? (short int)32767 : (short int)0;
	}
	break;
    case TYPEIM_16B :
        {
	unsigned short int *out;
	out = (unsigned short int*)(resIm->buf);
	for ( i = 0; i < z; i++ )
	    *out++ = ( *in++ > 0 ) ? (unsigned short int)65535 : (unsigned short int)0;
	}
	break;
    case TYPEIM_FLOAT :
        {
	float *out;
	out = (float*)(resIm->buf);
	for ( i = 0; i < z; i++ )
	    *out++ = ( *in++ > 0 ) ? (float)1.0 : (float)0.0;
	}
	break;
    default :
	for ( z = 0; z < resIm->tz; z++ )
        for ( y = 0; y < resIm->ty; y++ )
	for ( x = 0; x < resIm->tx; x++ ) {
	    if ( *in++ > 0 ) 
		T_Set3Point( x, y, z, resIm, (reeltype)1.0, MODE_SET );
	    else
		T_Set3Point( x, y, z, resIm, (reeltype)0.0, MODE_SET );
	}
    }
    return( TRUE );
}

/* Procedure de recherche de composantes connexes.

   Cette procedure est a usage interne.
   
   Le principe est le suivant. L'image est deja pre-traitee.
   Les points des composantes connexes ont une valeur superieure
   ou egale a _con_low_thres, les composantes connexes a conserver
   ont au moins un point de valeur superieure ou egale
   a _con_hig_thres.

   Pour chaque point de l'image de valeur superieure ou egale
   a _con_low_thres :

   - on recherche ses voisins deja numerotes

   - on le numerote et on gere eventuellement
     les equivalences
     
   - on teste si sa valeur est superieure ou egale a
     _con_hig_thres
 
*/
#if defined(_ANSI_)
static int _CON_ComputeCC( t_Image *theIm /* image des labels */,
			 e_ParamConnex *par /* parametres */,
			 e_CC_results *res/* resultats */ )
#else
static int _CON_ComputeCC( theIm, par, res )
t_Image *theIm;     /* image des labels */
e_ParamConnex *par; /* parametres */
e_CC_results *res;  /* resultats */ 
#endif
{
    register int x, y, z;
    register int i1, i2, label;
    int j1, j2;
    unsigned short int *buf, val;
    int dimx, dimy, dimz, plan, offset[13], labels[13];
    _con_te table, *add_table;
    e_CC *first_tab, *ordered_tab;

    /*--- image : parametres ---*/
    dimx = theIm->tx;
    dimy = theIm->ty;
    dimz = theIm->tz;
    plan = dimx * dimy;
    buf  = (unsigned short int*)(theIm->buf);

    /*--- pre-calcul des offsets ---*/
    offset[0]  = - plan - dimx - 1;   /* point (-1, -1, -1) */
    offset[1]  = - plan - dimx;       /* point ( 0, -1, -1) */
    offset[2]  = - plan - dimx + 1;   /* point ( 1, -1, -1) */
    offset[3]  = - plan - 1;          /* point (-1,  0, -1) */
    offset[4]  = - plan;              /* point ( 0,  0, -1) */
    offset[5]  = - plan + 1;          /* point ( 1,  0, -1) */
    offset[6]  = - plan + dimx - 1;   /* point (-1,  1, -1) */
    offset[7]  = - plan + dimx;       /* point ( 0,  1, -1) */
    offset[8]  = - plan + dimx + 1;   /* point ( 1,  1, -1) */
    offset[9]  =        - dimx - 1;   /* point (-1, -1,  0) */
    offset[10] =        - dimx;       /* point ( 0, -1,  0) */
    offset[11] =        - dimx + 1;   /* point ( 1, -1,  0) */
    offset[12] =        - 1;          /* point (-1,  0,  0) */

    /*--- table d'equivalence : initialisation ---*/
    table.nb = 0;
    table.sz = (int)( sqrt( (double)(dimz * plan) ) + 1 );
    table.sz10   = 10   * table.sz;
    table.sz100  = 100  * table.sz;
    table.sz1000 = 1000 * table.sz;
    table.isz = ( table.sz > 0 ) ? ( 1.0 / (double)table.sz ) : 0.0;
    table.tb = (_con_ce**)(NULL);
    add_table = &table;

    /*--- traitement de l'image ---*/
    
    /*--- premier plan : coin superieur gauche ---*/
    val = *buf;
    if ( val >= _con_low_thres ) {
	if ( _CON_CreateLabel( &table ) != TRUE ) {
	    _CON_FreeTable( &table );
	    return( FALSE );
	}
	label = table.nb;
	_COMPUTE_LABEL(i1,i2,label,add_table)
	*buf = label;
	if ( val >= _con_hig_thres ) table.tb[i1][i2].ok = TRUE;
    }
    buf++;
    /*--- premier plan : ligne superieure ---*/
    for ( x = 1; x < dimx; x++ ) {
	val = *buf;
	if ( val >= _con_low_thres ) {
	    if ( *(buf-1) > 0 ) {
		label =  *(buf-1);
		_COMPUTE_LABEL(i1,i2,label,add_table)
		table.tb[i1][i2].sz ++;
	    }
	    else {
		if ( _CON_CreateLabel( &table ) != TRUE ) {
		    _CON_FreeTable( &table );
		    return( FALSE );
		}
		label = table.nb;
		_COMPUTE_LABEL(i1,i2,label,add_table)
	    }
	    *buf = label;
	    if ( val >= _con_hig_thres ) table.tb[i1][i2].ok = TRUE;
	}
	buf++;
    }
    /*--- premier plan : autres lignes ---*/
    for ( y = 1; y < dimy; y++ ) {
	/*--- premier plan : autres lignes : point de gauche ---*/
	val = *buf;
	if ( val >= _con_low_thres ) {
	    label = _CON_VoisinsPpAlCg( buf, &table, labels, offset, par );
	    if ( label != 0 ) {
		_COMPUTE_LABEL(i1,i2,label,add_table)
		table.tb[i1][i2].sz ++;
	    }
	    else {
		if ( _CON_CreateLabel( &table ) != TRUE ) {
		    _CON_FreeTable( &table );
		    return( FALSE );
		}
		label = table.nb;
		_COMPUTE_LABEL(i1,i2,label,add_table)
	    }
	    *buf = label;
	    if ( val >= _con_hig_thres ) table.tb[i1][i2].ok = TRUE;
	}
	buf++;
	/*--- premier plan : autres lignes : points du centre ---*/
	for ( x = 1; x < dimx - 1; x++ ) {
	    val = *buf;
	    if ( val >= _con_low_thres ) {
		label = _CON_VoisinsPpAlCe( buf, &table, labels, offset, par );
		if ( label != 0 ) {
		    _COMPUTE_LABEL(i1,i2,label,add_table)
		    table.tb[i1][i2].sz ++;
		}
		else {
		    if ( _CON_CreateLabel( &table ) != TRUE ) {
			_CON_FreeTable( &table );
			return( FALSE );
		    }
		    label = table.nb;
		    _COMPUTE_LABEL(i1,i2,label,add_table)
		}
		*buf = label;
		if ( val >= _con_hig_thres ) table.tb[i1][i2].ok = TRUE;
	    }
	    buf++;
	}
	/*--- premier plan : autres lignes : point de droite ---*/
	val = *buf;
	if ( val >= _con_low_thres ) {
	    label = _CON_VoisinsPpAlCd( buf, &table, labels, offset, par );
	    if ( label != 0 ) {
		_COMPUTE_LABEL(i1,i2,label,add_table)
		table.tb[i1][i2].sz ++;
	    }
	    else {
		if ( _CON_CreateLabel( &table ) != TRUE ) {
		    _CON_FreeTable( &table );
		    return( FALSE );
		}
		label = table.nb;
		_COMPUTE_LABEL(i1,i2,label,add_table)
	    }
	    *buf = label;
	    if ( val >= _con_hig_thres ) table.tb[i1][i2].ok = TRUE;
	}
	buf++;
    }
    /*--- autres plans ---*/
    for ( z = 1; z < dimz; z ++ ) {
	/*--- autres plans : premiere ligne : point de gauche ---*/
	val = *buf;
	if ( val >= _con_low_thres ) {
	    label = _CON_VoisinsApPlCg( buf, &table, labels, offset, par );
	    if ( label != 0 ) {
		_COMPUTE_LABEL(i1,i2,label,add_table)
		table.tb[i1][i2].sz ++;
	    }
	    else {
		if ( _CON_CreateLabel( &table ) != TRUE ) {
		    _CON_FreeTable( &table );
		    return( FALSE );
		}
		label = table.nb;
		_COMPUTE_LABEL(i1,i2,label,add_table)
	    }
	    *buf = label;
	    if ( val >= _con_hig_thres ) table.tb[i1][i2].ok = TRUE;
	}
	buf++;
	/*--- autres plans : premiere ligne : points du centre ---*/
	for ( x = 1; x < dimx - 1; x++ ) {
	    val = *buf;
	    if ( val >= _con_low_thres ) {
		label = _CON_VoisinsApPlCe( buf, &table, labels, offset, par );
		if ( label != 0 ) {
		    _COMPUTE_LABEL(i1,i2,label,add_table)
		    table.tb[i1][i2].sz ++;
		}
		else {
		    if ( _CON_CreateLabel( &table ) != TRUE ) {
			_CON_FreeTable( &table );
			return( FALSE );
		    }
		    label = table.nb;
		    _COMPUTE_LABEL(i1,i2,label,add_table)
		}
		*buf = label;
		if ( val >= _con_hig_thres ) table.tb[i1][i2].ok = TRUE;
	    }
	    buf++;
	}
	/*--- autres plans : premiere ligne : point de droite ---*/
	val = *buf;
	if ( val >= _con_low_thres ) {
	    label = _CON_VoisinsApPlCd( buf, &table, labels, offset, par );
	    if ( label != 0 ) {
		_COMPUTE_LABEL(i1,i2,label,add_table)
		table.tb[i1][i2].sz ++;
	    }
	    else {
		if ( _CON_CreateLabel( &table ) != TRUE ) {
		    _CON_FreeTable( &table );
		    return( FALSE );
		}
		label = table.nb;
		_COMPUTE_LABEL(i1,i2,label,add_table)
	    }
	    *buf = label;
	    if ( val >= _con_hig_thres ) table.tb[i1][i2].ok = TRUE;
	}
	buf++;
	/*--- autres plans : autres lignes ---*/
	for ( y = 1; y < dimy - 1 ; y++ ) {
	    /*--- autres plans : autres lignes : point de gauche ---*/
	    val = *buf;
	    if ( val >= _con_low_thres ) {
		label = _CON_VoisinsApAlCg( buf, &table, labels, offset, par );
		if ( label != 0 ) {
		    _COMPUTE_LABEL(i1,i2,label,add_table)
		    table.tb[i1][i2].sz ++;
		}
		else {
		    if ( _CON_CreateLabel( &table ) != TRUE ) {
			_CON_FreeTable( &table );
			return( FALSE );
		    }
		    label = table.nb;
		    _COMPUTE_LABEL(i1,i2,label,add_table)
		}
		*buf = label;
		if ( val >= _con_hig_thres ) table.tb[i1][i2].ok = TRUE;
	    }
	    buf++;
	    /*--- autres plans : autres lignes : points du centre ---*/
	    for ( x = 1; x < dimx - 1; x++ ) {
		val = *buf;
		if ( val >= _con_low_thres ) {
		    label = _CON_VoisinsApAlCe( buf, &table, labels, offset, par );
		    if ( label != 0 ) {
			_COMPUTE_LABEL(i1,i2,label,add_table)
			table.tb[i1][i2].sz ++;
		    }
		    else {
			if ( _CON_CreateLabel( &table ) != TRUE ) {
			    _CON_FreeTable( &table );
			    return( FALSE );
			}
			label = table.nb;
			_COMPUTE_LABEL(i1,i2,label,add_table)
		    }
		    *buf = label;
		    if ( val >= _con_hig_thres ) table.tb[i1][i2].ok = TRUE;
		}
		buf++;
	    }
	    /*--- autres plans : autres lignes : point de droite ---*/
	    val = *buf;
	    if ( val >= _con_low_thres ) {
		label = _CON_VoisinsApAlCd( buf, &table, labels, offset, par );
		if ( label != 0 ) {
		    _COMPUTE_LABEL(i1,i2,label,add_table)
		    table.tb[i1][i2].sz ++;
		}
		else {
		    if ( _CON_CreateLabel( &table ) != TRUE ) {
			_CON_FreeTable( &table );
			return( FALSE );
		    }
		    label = table.nb;
		    _COMPUTE_LABEL(i1,i2,label,add_table)
		}
		*buf = label;
		if ( val >= _con_hig_thres ) table.tb[i1][i2].ok = TRUE;
	    }
	    buf++;
	}
	/*--- autres plans : derniere ligne : point de gauche ---*/
	val = *buf;
	if ( val >= _con_low_thres ) {
	    label = _CON_VoisinsApDlCg( buf, &table, labels, offset, par );
	    if ( label != 0 ) {
		_COMPUTE_LABEL(i1,i2,label,add_table)
		table.tb[i1][i2].sz ++;
	    }
	    else {
		if ( _CON_CreateLabel( &table ) != TRUE ) {
		    _CON_FreeTable( &table );
		    return( FALSE );
		}
		label = table.nb;
		_COMPUTE_LABEL(i1,i2,label,add_table)
	    }
	    *buf = label;
	    if ( val >= _con_hig_thres ) table.tb[i1][i2].ok = TRUE;
	}
	buf++;
	/*--- autres plans : derniere ligne : points du centre ---*/
	for ( x = 1; x < dimx - 1; x++ ) {
	    val = *buf;
	    if ( val >= _con_low_thres ) {
		label = _CON_VoisinsApDlCe( buf, &table, labels, offset, par );
		if ( label != 0 ) {
		    _COMPUTE_LABEL(i1,i2,label,add_table)
		    table.tb[i1][i2].sz ++;
		}
		else {
		    if ( _CON_CreateLabel( &table ) != TRUE ) {
			_CON_FreeTable( &table );
			return( FALSE );
		    }
		    label = table.nb;
		    _COMPUTE_LABEL(i1,i2,label,add_table)
		}
		*buf = label;
		if ( val >= _con_hig_thres ) table.tb[i1][i2].ok = TRUE;
	    }
	    buf++;
	}
	/*--- autres plans : derniere ligne : point de droite ---*/
	val = *buf;
	if ( val >= _con_low_thres ) {
	    label = _CON_VoisinsApDlCd( buf, &table, labels, offset, par );
	    if ( label != 0 ) {
		_COMPUTE_LABEL(i1,i2,label,add_table)
		table.tb[i1][i2].sz ++;
	    }
	    else {
		if ( _CON_CreateLabel( &table ) != TRUE ) {
		    _CON_FreeTable( &table );
		    return( FALSE );
		}
		label = table.nb;
		_COMPUTE_LABEL(i1,i2,label,add_table)
	    }
	    *buf = label;
	    if ( val >= _con_hig_thres ) table.tb[i1][i2].ok = TRUE;
	}
	buf++;
    }
    
    /*--- le traitement de l'image est maintenant fini : 
          on rearrange la table d'equivalence
	  z va designe le nombre de composantes connexes ---*/
    z = 0;
    for ( x = 1 ; x <= table.nb ; x ++ ) {
	_COMPUTE_LABEL(i1,i2,x,add_table)
	/*--- si c'est le representant d'une classe d'equivalence :
	      on lui attribue un nouveau label                      ---*/
	if ( table.tb[i1][i2].eq == x ) {
	    if ( (table.tb[i1][i2].ok == TRUE) && (table.tb[i1][i2].sz >= par->minSize) ) 
		table.tb[i1][i2].eq = ++z;
	    else
		table.tb[i1][i2].eq = 0;
	}
	/*--- sinon, on lui attribue le label de son representant ---*/
	else {
	    label = table.tb[i1][i2].eq;
	    _COMPUTE_LABEL(j1,j2,label,add_table)
	    table.tb[i1][i2].eq =  table.tb[j1][j2].eq;
	}
    }

    /*--- resultats : construction du tableau des composantes connexes ---*/
    res->nb = z;
    first_tab = ordered_tab = (e_CC*)NULL;
    if ( (res->bool_alloc == TRUE) || (par->maxNbCC > 0) ) {
	/*--- allocation ---*/
	first_tab = (e_CC*)T_Alloc((z+1)*sizeof(e_CC));
	if ( first_tab == (e_CC*)NULL ) {
	    _CON_FreeTable( &table );
	    return( FALSE );
	}
	/*--- determination des labels ---*/
	first_tab[0].label = first_tab[0].size = 0;
	for ( x = 1 ; x <= table.nb ; x ++ ) {
	    _COMPUTE_LABEL(i1,i2,x,add_table)
	    if ( table.tb[i1][i2].sz > 0 ) {
		first_tab[ table.tb[i1][i2].eq ].label = table.tb[i1][i2].eq;
		first_tab[ table.tb[i1][i2].eq ].size  = table.tb[i1][i2].sz;
	    }
	}
	/*--- faut-il faire un tri ? ---*/
	if ( par->maxNbCC > 0 ) {
	    /*--- allocation ---*/
	    ordered_tab = (e_CC*)T_Alloc((z+1)*sizeof(e_CC));
	    if ( ordered_tab == (e_CC*)NULL ) {
		T_Free( (void**)(&first_tab) );
		_CON_FreeTable( &table );
		return( FALSE );
	    }
	    /*--- tri ---*/
	    _CON_Tri_CC( first_tab, (int)1, z );
	    /*--- copie ---*/
	    for ( x = 1; x <= z; x++ ) {
		if ( x > par->maxNbCC ) {
		    ordered_tab[ first_tab[x].label ].label = 0;
		    first_tab[x].label = first_tab[x].size = 0;
		}
		else {
		    ordered_tab[ first_tab[x].label ].label = x;
		}
	    }
	    /*--- repercution sur la table des labels ---*/
	    for ( x = 1 ; x <= table.nb ; x ++ ) {
		_COMPUTE_LABEL(i1,i2,x,add_table)
		label = table.tb[i1][i2].eq;
		table.tb[i1][i2].eq = ordered_tab[label].label;
	    }
	    T_Free( (void**)(&ordered_tab) );
	    /*--- resultat ---*/
	    if ( z > par->maxNbCC ) res->nb = par->maxNbCC;
	}
	if ( res->bool_alloc == TRUE ) res->tab = first_tab;
	else                     T_Free( (void**)(&first_tab) );
    }
    /*--- on renumerote l'image des labels ---*/
    buf = (unsigned short int*)(theIm->buf);
    for ( x = 0; x < dimz * plan ; x++ ) {
	label = *buf;
	if ( label > 0 ) {
	    _COMPUTE_LABEL(i1,i2,label,add_table)
	    *buf = table.tb[i1][i2].eq;
	}
	buf++;
    }

    /*--- fin ---*/
    _CON_FreeTable( &table );
    return( TRUE );
}

#if defined(_ANSI_)
static int _CON_CreateLabel( _con_te *table )
#else
static int _CON_CreateLabel( table )
_con_te *table;
#endif
{
    int i, j;
    
    /*--- premiere allocation ---*/
    if ( table->nb == 0 ) {
	table->tb = (_con_ce**)T_Alloc( table->sz * sizeof(_con_ce*) );
	E_ISERR( FALSE, "CON_CreateLabel", (char*)NULL);
	for ( i = 0; i < table->sz; i++) table->tb[i] = (_con_ce*)NULL;

	table->tb[0] = (_con_ce*)T_Alloc( table->sz * sizeof(_con_ce) );
	E_ISERR( FALSE, "CON_CreateLabel", (char*)NULL);
	table->nb = 1;
	table->tb[0][1].eq = 1;
	table->tb[0][1].sz = 1;
	table->tb[0][1].ok = FALSE;
    }
    /*--- autre allocation ---*/
    else {
	table->nb ++;
	_COMPUTE_LABEL(i,j,table->nb,table)
	if ( j == 0 ) {
	    table->tb[i] = (_con_ce*)T_Alloc( table->sz * sizeof(_con_ce) );
	    E_ISERR( FALSE, "CON_CreateLabel", (char*)NULL);
	}
	table->tb[i][j].eq = table->nb;
	table->tb[i][j].sz = 1;
	table->tb[i][j].ok = FALSE;
    }
    E_ERR( (table->nb > 65535), E_ERRNULL, FALSE, 
	   "CON_CreateLabel", "too much labels for auxiliary image" );
    return( TRUE );
}
    
#if defined(_ANSI_)
static void _CON_FreeTable( _con_te *table )
#else
static void _CON_FreeTable( table )
_con_te *table;
#endif
{
    int i, j;
    
    if ( table->nb == 0 ) return;
    if ( table->tb == (_con_ce**)NULL ) {
	table->nb = 0;
	return;
    }

    _COMPUTE_LABEL(i,j,table->nb,table)
    for ( j = 0; j <= i ; j ++) {
	if ( table->tb[i] != (_con_ce*)NULL )
	    T_Free( (void**)( &(table->tb[i]) ) );
    }
    T_Free( (void**)( &(table->tb) ) );
    table->nb = 0;
}

/* Test des voisins : premier plan
                      autres lignes 
		      colonne de gauche 
*/

#if defined(_ANSI_)
static int _CON_VoisinsPpAlCg( unsigned short int *buf,
			       _con_te *table,
			       int *labels,
			       int *offset,
			       e_ParamConnex *par )
#else
static int _CON_VoisinsPpAlCg( buf, table, labels, offset, par )
unsigned short int *buf;
_con_te *table;
int *labels;
int *offset;
e_ParamConnex *par;
#endif
{
    register int i, j, i1, i2, m1, m2;
    int nb = 0;
    int min, val;
    
    /*--- test des voisins ---*/
    switch ( par->connexite ) {
    case CONNEXE_26 :
    case CONNEXE_18 :
    case CONNEXE_10 :
    case CONNEXE_08 :
     	if ( *(buf + offset[11]) > 0 ) labels[nb++] = *(buf + offset[11]);
    case CONNEXE_06 :
    case CONNEXE_04 :
     	if ( *(buf + offset[10]) > 0 ) labels[nb++] = *(buf + offset[10]);
    }
    /*--- pas de voisins ---*/
    if ( nb == 0 ) return( 0 );
    /*--- recherche des labels equivalents ---*/
    for ( i = 0; i < nb ; i ++ ) {
	_COMPUTE_LABEL(i1,i2,labels[i],table)
	labels[i] = table->tb[i1][i2].eq;
    }
    /*--- recherche du plus petit label ---*/
    min = labels[0];
    for ( i = 1; i < nb ; i ++ ) 
	if ( min > labels[i] ) min = labels[i];
    /*--- un seul label ---*/
    if ( nb == 1 ) return( min );
    /*--- gestion des equivalences ---*/
    _COMPUTE_LABEL(m1,m2,min,table)
    for ( i = 0; i < nb ; i ++ ) {
	_COMPUTE_LABEL(i1,i2,labels[i],table)
	if ( table->tb[i1][i2].eq != min ) {
	    val = table->tb[i1][i2].eq;
	    table->tb[m1][m2].sz += table->tb[i1][i2].sz;
	    if ( table->tb[i1][i2].ok == TRUE ) table->tb[m1][m2].ok = TRUE;
	    table->tb[i1][i2].sz = 0;
	    table->tb[i1][i2].eq = min;
	    for ( j = (labels[i] + 1); j <= table->nb; j ++ ) {
		_COMPUTE_LABEL(i1,i2,j,table)
		if ( table->tb[i1][i2].eq == val ) table->tb[i1][i2].eq = min;
	    }
	}
    }
    return( min );
}

/* Test des voisins : premier plan
                      autres lignes 
		      colonnes du centre
*/

#if defined(_ANSI_)
static int _CON_VoisinsPpAlCe( unsigned short int *buf,
			       _con_te *table,
			       int *labels,
			       int *offset,
			       e_ParamConnex *par )
#else
static int _CON_VoisinsPpAlCe( buf, table, labels, offset, par )
unsigned short int *buf;
_con_te *table;
int *labels;
int *offset;
e_ParamConnex *par;
#endif
{
    register int i, j, i1, i2, m1, m2;
    int nb = 0;
    int min, val;
    
    /*--- test des voisins ---*/
    switch ( par->connexite ) {
    case CONNEXE_26 :
    case CONNEXE_18 :
    case CONNEXE_10 :
    case CONNEXE_08 :
     	if ( *(buf + offset[9]) > 0 ) labels[nb++] = *(buf + offset[9]);
     	if ( *(buf + offset[11]) > 0 ) labels[nb++] = *(buf + offset[11]);
    case CONNEXE_06 :
    case CONNEXE_04 :
     	if ( *(buf + offset[10]) > 0 ) labels[nb++] = *(buf + offset[10]);
     	if ( *(buf + offset[12]) > 0 ) labels[nb++] = *(buf + offset[12]);
    }
    /*--- pas de voisins ---*/
    if ( nb == 0 ) return( 0 );
    /*--- recherche des labels equivalents ---*/
    for ( i = 0; i < nb ; i ++ ) {
	_COMPUTE_LABEL(i1,i2,labels[i],table)
	labels[i] = table->tb[i1][i2].eq;
    }
    /*--- recherche du plus petit label ---*/
    min = labels[0];
    for ( i = 1; i < nb ; i ++ ) 
	if ( min > labels[i] ) min = labels[i];
    /*--- un seul label ---*/
    if ( nb == 1 ) return( min );
    /*--- gestion des equivalences ---*/
    _COMPUTE_LABEL(m1,m2,min,table)
    for ( i = 0; i < nb ; i ++ ) {
	_COMPUTE_LABEL(i1,i2,labels[i],table)
	if ( table->tb[i1][i2].eq != min ) {
	    val = table->tb[i1][i2].eq;
	    table->tb[m1][m2].sz += table->tb[i1][i2].sz;
	    if ( table->tb[i1][i2].ok == TRUE ) table->tb[m1][m2].ok = TRUE;
	    table->tb[i1][i2].sz = 0;
	    table->tb[i1][i2].eq = min;
	    for ( j = (labels[i] + 1); j <= table->nb; j ++ ) {
		_COMPUTE_LABEL(i1,i2,j,table)
		if ( table->tb[i1][i2].eq == val ) table->tb[i1][i2].eq = min;
	    }
	}
    }
    return( min );
}

/* Test des voisins : premier plan
                      autres lignes 
		      colonne de droite
*/

#if defined(_ANSI_)
static int _CON_VoisinsPpAlCd( unsigned short int *buf,
			       _con_te *table,
			       int *labels,
			       int *offset,
			       e_ParamConnex *par )
#else
static int _CON_VoisinsPpAlCd( buf, table, labels, offset, par )
unsigned short int *buf;
_con_te *table;
int *labels;
int *offset;
e_ParamConnex *par;
#endif
{
    register int i, j, i1, i2, m1, m2;
    int nb = 0;
    int min, val;
    
    /*--- test des voisins ---*/
    switch ( par->connexite ) {
    case CONNEXE_26 :
    case CONNEXE_18 :
    case CONNEXE_10 :
    case CONNEXE_08 :
     	if ( *(buf + offset[9]) > 0 ) labels[nb++] = *(buf + offset[9]);
    case CONNEXE_06 :
    case CONNEXE_04 :
     	if ( *(buf + offset[10]) > 0 ) labels[nb++] = *(buf + offset[10]);
     	if ( *(buf + offset[12]) > 0 ) labels[nb++] = *(buf + offset[12]);
    }
    /*--- pas de voisins ---*/
    if ( nb == 0 ) return( 0 );
    /*--- recherche des labels equivalents ---*/
    for ( i = 0; i < nb ; i ++ ) {
	_COMPUTE_LABEL(i1,i2,labels[i],table)
	labels[i] = table->tb[i1][i2].eq;
    }
    /*--- recherche du plus petit label ---*/
    min = labels[0];
    for ( i = 1; i < nb ; i ++ ) 
	if ( min > labels[i] ) min = labels[i];
    /*--- un seul label ---*/
    if ( nb == 1 ) return( min );
    /*--- gestion des equivalences ---*/
    _COMPUTE_LABEL(m1,m2,min,table)
    for ( i = 0; i < nb ; i ++ ) {
	_COMPUTE_LABEL(i1,i2,labels[i],table)
	if ( table->tb[i1][i2].eq != min ) {
	    val = table->tb[i1][i2].eq;
	    table->tb[m1][m2].sz += table->tb[i1][i2].sz;
	    if ( table->tb[i1][i2].ok == TRUE ) table->tb[m1][m2].ok = TRUE;
	    table->tb[i1][i2].sz = 0;
	    table->tb[i1][i2].eq = min;
	    for ( j = (labels[i] + 1); j <= table->nb; j ++ ) {
		_COMPUTE_LABEL(i1,i2,j,table)
		if ( table->tb[i1][i2].eq == val ) table->tb[i1][i2].eq = min;
	    }
	}
    }
    return( min );
}

/* Test des voisins : autres plan
                      premiere ligne 
		      colonne de gauche 
*/

#if defined(_ANSI_)
static int _CON_VoisinsApPlCg( unsigned short int *buf,
			       _con_te *table,
			       int *labels,
			       int *offset,
			       e_ParamConnex *par )
#else
static int _CON_VoisinsApPlCg( buf, table, labels, offset, par )
unsigned short int *buf;
_con_te *table;
int *labels;
int *offset;
e_ParamConnex *par;
#endif
{
    register int i, j, i1, i2, m1, m2;
    int nb = 0;
    int min, val;
    
    /*--- test des voisins ---*/
    switch ( par->connexite ) {
    case CONNEXE_26 :
	if ( *(buf + offset[8]) > 0 ) labels[nb++] = *(buf + offset[8]);
    case CONNEXE_18 :
	if ( *(buf + offset[5]) > 0 ) labels[nb++] = *(buf + offset[5]);
	if ( *(buf + offset[7]) > 0 ) labels[nb++] = *(buf + offset[7]);
    case CONNEXE_10 :
     	if ( *(buf + offset[4]) > 0 ) labels[nb++] = *(buf + offset[4]);
    }
    /*--- pas de voisins ---*/
    if ( nb == 0 ) return( 0 );
    /*--- recherche des labels equivalents ---*/
    for ( i = 0; i < nb ; i ++ ) {
	_COMPUTE_LABEL(i1,i2,labels[i],table)
	labels[i] = table->tb[i1][i2].eq;
    }
    /*--- recherche du plus petit label ---*/
    min = labels[0];
    for ( i = 1; i < nb ; i ++ ) 
	if ( min > labels[i] ) min = labels[i];
    /*--- un seul label ---*/
    if ( nb == 1 ) return( min );
    /*--- gestion des equivalences ---*/
    _COMPUTE_LABEL(m1,m2,min,table)
    for ( i = 0; i < nb ; i ++ ) {
	_COMPUTE_LABEL(i1,i2,labels[i],table)
	if ( table->tb[i1][i2].eq != min ) {
	    val = table->tb[i1][i2].eq;
	    table->tb[m1][m2].sz += table->tb[i1][i2].sz;
	    if ( table->tb[i1][i2].ok == TRUE ) table->tb[m1][m2].ok = TRUE;
	    table->tb[i1][i2].sz = 0;
	    table->tb[i1][i2].eq = min;
	    for ( j = (labels[i] + 1); j <= table->nb; j ++ ) {
		_COMPUTE_LABEL(i1,i2,j,table)
		if ( table->tb[i1][i2].eq == val ) table->tb[i1][i2].eq = min;
	    }
	}
    }
    return( min );
}

/* Test des voisins : autres plan
                      premiere ligne 
		      colonnes du centre
*/

#if defined(_ANSI_)
static int _CON_VoisinsApPlCe( unsigned short int *buf,
			       _con_te *table,
			       int *labels,
			       int *offset,
			       e_ParamConnex *par )
#else
static int _CON_VoisinsApPlCe( buf, table, labels, offset, par )
unsigned short int *buf;
_con_te *table;
int *labels;
int *offset;
e_ParamConnex *par;
#endif
{
    register int i, j, i1, i2, m1, m2;
    int nb = 0;
    int min, val;
    
    /*--- test des voisins ---*/
    switch ( par->connexite ) {
    case CONNEXE_26 :
	if ( *(buf + offset[6]) > 0 ) labels[nb++] = *(buf + offset[6]);
	if ( *(buf + offset[8]) > 0 ) labels[nb++] = *(buf + offset[8]);
    case CONNEXE_18 :
	if ( *(buf + offset[3]) > 0 ) labels[nb++] = *(buf + offset[3]);
	if ( *(buf + offset[5]) > 0 ) labels[nb++] = *(buf + offset[5]);
	if ( *(buf + offset[7]) > 0 ) labels[nb++] = *(buf + offset[7]);
    case CONNEXE_10 :
     	if ( *(buf + offset[4]) > 0 ) labels[nb++] = *(buf + offset[4]);
    case CONNEXE_08 :
    case CONNEXE_04 :
     	if ( *(buf + offset[12]) > 0 ) labels[nb++] = *(buf + offset[12]);
	break;
    case CONNEXE_06 :
     	if ( *(buf + offset[4]) > 0 ) labels[nb++] = *(buf + offset[4]);
     	if ( *(buf + offset[12]) > 0 ) labels[nb++] = *(buf + offset[12]);
    }
    /*--- pas de voisins ---*/
    if ( nb == 0 ) return( 0 );
    /*--- recherche des labels equivalents ---*/
    for ( i = 0; i < nb ; i ++ ) {
	_COMPUTE_LABEL(i1,i2,labels[i],table)
	labels[i] = table->tb[i1][i2].eq;
    }
    /*--- recherche du plus petit label ---*/
    min = labels[0];
    for ( i = 1; i < nb ; i ++ ) 
	if ( min > labels[i] ) min = labels[i];
    /*--- un seul label ---*/
    if ( nb == 1 ) return( min );
    /*--- gestion des equivalences ---*/
    _COMPUTE_LABEL(m1,m2,min,table)
    for ( i = 0; i < nb ; i ++ ) {
	_COMPUTE_LABEL(i1,i2,labels[i],table)
	if ( table->tb[i1][i2].eq != min ) {
	    val = table->tb[i1][i2].eq;
	    table->tb[m1][m2].sz += table->tb[i1][i2].sz;
	    if ( table->tb[i1][i2].ok == TRUE ) table->tb[m1][m2].ok = TRUE;
	    table->tb[i1][i2].sz = 0;
	    table->tb[i1][i2].eq = min;
	    for ( j = (labels[i] + 1); j <= table->nb; j ++ ) {
		_COMPUTE_LABEL(i1,i2,j,table)
		if ( table->tb[i1][i2].eq == val ) table->tb[i1][i2].eq = min;
	    }
	}
    }
    return( min );
}

/* Test des voisins : autres plan
                      premiere ligne 
		      colonne de droite
*/

#if defined(_ANSI_)
static int _CON_VoisinsApPlCd( unsigned short int *buf,
			       _con_te *table,
			       int *labels,
			       int *offset,
			       e_ParamConnex *par )
#else
static int _CON_VoisinsApPlCd( buf, table, labels, offset, par )
unsigned short int *buf;
_con_te *table;
int *labels;
int *offset;
e_ParamConnex *par;
#endif
{
    register int i, j, i1, i2, m1, m2;
    int nb = 0;
    int min, val;
    
    /*--- test des voisins ---*/
    switch ( par->connexite ) {
    case CONNEXE_26 :
	if ( *(buf + offset[6]) > 0 ) labels[nb++] = *(buf + offset[6]);
    case CONNEXE_18 :
	if ( *(buf + offset[3]) > 0 ) labels[nb++] = *(buf + offset[3]);
	if ( *(buf + offset[7]) > 0 ) labels[nb++] = *(buf + offset[7]);
    case CONNEXE_10 :
     	if ( *(buf + offset[4]) > 0 ) labels[nb++] = *(buf + offset[4]);
    case CONNEXE_08 :
    case CONNEXE_04 :
     	if ( *(buf + offset[12]) > 0 ) labels[nb++] = *(buf + offset[12]);
	break;
    case CONNEXE_06 :
     	if ( *(buf + offset[4]) > 0 ) labels[nb++] = *(buf + offset[4]);
     	if ( *(buf + offset[12]) > 0 ) labels[nb++] = *(buf + offset[12]);
    }
    /*--- pas de voisins ---*/
    if ( nb == 0 ) return( 0 );
    /*--- recherche des labels equivalents ---*/
    for ( i = 0; i < nb ; i ++ ) {
	_COMPUTE_LABEL(i1,i2,labels[i],table)
	labels[i] = table->tb[i1][i2].eq;
    }
    /*--- recherche du plus petit label ---*/
    min = labels[0];
    for ( i = 1; i < nb ; i ++ ) 
	if ( min > labels[i] ) min = labels[i];
    /*--- un seul label ---*/
    if ( nb == 1 ) return( min );
    /*--- gestion des equivalences ---*/
    _COMPUTE_LABEL(m1,m2,min,table)
    for ( i = 0; i < nb ; i ++ ) {
	_COMPUTE_LABEL(i1,i2,labels[i],table)
	if ( table->tb[i1][i2].eq != min ) {
	    val = table->tb[i1][i2].eq;
	    table->tb[m1][m2].sz += table->tb[i1][i2].sz;
	    if ( table->tb[i1][i2].ok == TRUE ) table->tb[m1][m2].ok = TRUE;
	    table->tb[i1][i2].sz = 0;
	    table->tb[i1][i2].eq = min;
	    for ( j = (labels[i] + 1); j <= table->nb; j ++ ) {
		_COMPUTE_LABEL(i1,i2,j,table)
		if ( table->tb[i1][i2].eq == val ) table->tb[i1][i2].eq = min;
	    }
	}
    }
    return( min );
}

/* Test des voisins : autres plan
                      autres lignes
		      colonne de gauche 
*/

#if defined(_ANSI_)
static int _CON_VoisinsApAlCg( unsigned short int *buf,
			       _con_te *table,
			       int *labels,
			       int *offset,
			       e_ParamConnex *par )
#else
static int _CON_VoisinsApAlCg( buf, table, labels, offset, par )
unsigned short int *buf;
_con_te *table;
int *labels;
int *offset;
e_ParamConnex *par;
#endif
{
    register int i, j, i1, i2, m1, m2;
    int nb = 0;
    int min, val;
    
    /*--- test des voisins ---*/
    switch ( par->connexite ) {
    case CONNEXE_26 :
	if ( *(buf + offset[2]) > 0 ) labels[nb++] = *(buf + offset[2]);
	if ( *(buf + offset[8]) > 0 ) labels[nb++] = *(buf + offset[8]);
    case CONNEXE_18 :
	if ( *(buf + offset[1]) > 0 ) labels[nb++] = *(buf + offset[1]);
	if ( *(buf + offset[5]) > 0 ) labels[nb++] = *(buf + offset[5]);
	if ( *(buf + offset[7]) > 0 ) labels[nb++] = *(buf + offset[7]);
    case CONNEXE_10 :
     	if ( *(buf + offset[4]) > 0 ) labels[nb++] = *(buf + offset[4]);
    case CONNEXE_08 :
     	if ( *(buf + offset[11]) > 0 ) labels[nb++] = *(buf + offset[11]);
    case CONNEXE_04 :
     	if ( *(buf + offset[10]) > 0 ) labels[nb++] = *(buf + offset[10]);
	break;
    case CONNEXE_06 :
     	if ( *(buf + offset[4]) > 0 ) labels[nb++] = *(buf + offset[4]);
     	if ( *(buf + offset[10]) > 0 ) labels[nb++] = *(buf + offset[10]);
    }
    /*--- pas de voisins ---*/
    if ( nb == 0 ) return( 0 );
    /*--- recherche des labels equivalents ---*/
    for ( i = 0; i < nb ; i ++ ) {
	_COMPUTE_LABEL(i1,i2,labels[i],table)
	labels[i] = table->tb[i1][i2].eq;
    }
    /*--- recherche du plus petit label ---*/
    min = labels[0];
    for ( i = 1; i < nb ; i ++ ) 
	if ( min > labels[i] ) min = labels[i];
    /*--- un seul label ---*/
    if ( nb == 1 ) return( min );
    /*--- gestion des equivalences ---*/
    _COMPUTE_LABEL(m1,m2,min,table)
    for ( i = 0; i < nb ; i ++ ) {
	_COMPUTE_LABEL(i1,i2,labels[i],table)
	if ( table->tb[i1][i2].eq != min ) {
	    val = table->tb[i1][i2].eq;
	    table->tb[m1][m2].sz += table->tb[i1][i2].sz;
	    if ( table->tb[i1][i2].ok == TRUE ) table->tb[m1][m2].ok = TRUE;
	    table->tb[i1][i2].sz = 0;
	    table->tb[i1][i2].eq = min;
	    for ( j = (labels[i] + 1); j <= table->nb; j ++ ) {
		_COMPUTE_LABEL(i1,i2,j,table)
		if ( table->tb[i1][i2].eq == val ) table->tb[i1][i2].eq = min;
	    }
	}
    }
    return( min );
}

/* Test des voisins : autres plan
                      autres lignes 
		      colonnes du centre 
*/

#if defined(_ANSI_)
static int _CON_VoisinsApAlCe( unsigned short int *buf,
			       _con_te *table,
			       int *labels,
			       int *offset,
			       e_ParamConnex *par )
#else
static int _CON_VoisinsApAlCe( buf, table, labels, offset, par )
unsigned short int *buf;
_con_te *table;
int *labels;
int *offset;
e_ParamConnex *par;
#endif
{
    register int i, j, i1, i2, m1, m2;
    int nb = 0;
    int min, val;
    
    /*--- test des voisins ---*/
    switch ( par->connexite ) {
    case CONNEXE_26 :
	if ( *(buf + offset[0]) > 0 ) labels[nb++] = *(buf + offset[0]);
	if ( *(buf + offset[2]) > 0 ) labels[nb++] = *(buf + offset[2]);
	if ( *(buf + offset[6]) > 0 ) labels[nb++] = *(buf + offset[6]);
	if ( *(buf + offset[8]) > 0 ) labels[nb++] = *(buf + offset[8]);
    case CONNEXE_18 :
	if ( *(buf + offset[1]) > 0 ) labels[nb++] = *(buf + offset[1]);
	if ( *(buf + offset[3]) > 0 ) labels[nb++] = *(buf + offset[3]);
	if ( *(buf + offset[5]) > 0 ) labels[nb++] = *(buf + offset[5]);
	if ( *(buf + offset[7]) > 0 ) labels[nb++] = *(buf + offset[7]);
    case CONNEXE_10 :
     	if ( *(buf + offset[4]) > 0 ) labels[nb++] = *(buf + offset[4]);
    case CONNEXE_08 :
     	if ( *(buf + offset[9]) > 0 ) labels[nb++] = *(buf + offset[9]);
     	if ( *(buf + offset[11]) > 0 ) labels[nb++] = *(buf + offset[11]);
    case CONNEXE_04 :
     	if ( *(buf + offset[10]) > 0 ) labels[nb++] = *(buf + offset[10]);
     	if ( *(buf + offset[12]) > 0 ) labels[nb++] = *(buf + offset[12]);
	break;
    case CONNEXE_06 :
     	if ( *(buf + offset[4]) > 0 ) labels[nb++] = *(buf + offset[4]);
     	if ( *(buf + offset[10]) > 0 ) labels[nb++] = *(buf + offset[10]);
     	if ( *(buf + offset[12]) > 0 ) labels[nb++] = *(buf + offset[12]);
    }
    /*--- pas de voisins ---*/
    if ( nb == 0 ) return( 0 );
    /*--- recherche des labels equivalents ---*/
    for ( i = 0; i < nb ; i ++ ) {
	_COMPUTE_LABEL(i1,i2,labels[i],table)
	labels[i] = table->tb[i1][i2].eq;
    }
    /*--- recherche du plus petit label ---*/
    min = labels[0];
    for ( i = 1; i < nb ; i ++ ) 
	if ( min > labels[i] ) min = labels[i];
    /*--- un seul label ---*/
    if ( nb == 1 ) return( min );
    /*--- gestion des equivalences ---*/
    _COMPUTE_LABEL(m1,m2,min,table)
    for ( i = 0; i < nb ; i ++ ) {
	_COMPUTE_LABEL(i1,i2,labels[i],table)
	if ( table->tb[i1][i2].eq != min ) {
	    val = table->tb[i1][i2].eq;
	    table->tb[m1][m2].sz += table->tb[i1][i2].sz;
	    if ( table->tb[i1][i2].ok == TRUE ) table->tb[m1][m2].ok = TRUE;
	    table->tb[i1][i2].sz = 0;
	    table->tb[i1][i2].eq = min;
	    for ( j = (labels[i] + 1); j <= table->nb; j ++ ) {
		_COMPUTE_LABEL(i1,i2,j,table)
		if ( table->tb[i1][i2].eq == val ) table->tb[i1][i2].eq = min;
	    }
	}
    }
    return( min );
}

/* Test des voisins : autres plan
                      autres lignes 
		      colonne de droite
*/

#if defined(_ANSI_)
static int _CON_VoisinsApAlCd( unsigned short int *buf,
			       _con_te *table,
			       int *labels,
			       int *offset,
			       e_ParamConnex *par )
#else
static int _CON_VoisinsApAlCd( buf, table, labels, offset, par )
unsigned short int *buf;
_con_te *table;
int *labels;
int *offset;
e_ParamConnex *par;
#endif
{
    register int i, j, i1, i2, m1, m2;
    int nb = 0;
    int min, val;
    
    /*--- test des voisins ---*/
    switch ( par->connexite ) {
    case CONNEXE_26 :
	if ( *(buf + offset[0]) > 0 ) labels[nb++] = *(buf + offset[0]);
	if ( *(buf + offset[6]) > 0 ) labels[nb++] = *(buf + offset[6]);
    case CONNEXE_18 :
	if ( *(buf + offset[1]) > 0 ) labels[nb++] = *(buf + offset[1]);
	if ( *(buf + offset[3]) > 0 ) labels[nb++] = *(buf + offset[3]);
	if ( *(buf + offset[7]) > 0 ) labels[nb++] = *(buf + offset[7]);
    case CONNEXE_10 :
     	if ( *(buf + offset[4]) > 0 ) labels[nb++] = *(buf + offset[4]);
    case CONNEXE_08 :
     	if ( *(buf + offset[9]) > 0 ) labels[nb++] = *(buf + offset[9]);
    case CONNEXE_04 :
     	if ( *(buf + offset[10]) > 0 ) labels[nb++] = *(buf + offset[10]);
     	if ( *(buf + offset[12]) > 0 ) labels[nb++] = *(buf + offset[12]);
	break;
    case CONNEXE_06 :
     	if ( *(buf + offset[4]) > 0 ) labels[nb++] = *(buf + offset[4]);
     	if ( *(buf + offset[10]) > 0 ) labels[nb++] = *(buf + offset[10]);
     	if ( *(buf + offset[12]) > 0 ) labels[nb++] = *(buf + offset[12]);
    }
    /*--- pas de voisins ---*/
    if ( nb == 0 ) return( 0 );
    /*--- recherche des labels equivalents ---*/
    for ( i = 0; i < nb ; i ++ ) {
	_COMPUTE_LABEL(i1,i2,labels[i],table)
	labels[i] = table->tb[i1][i2].eq;
    }
    /*--- recherche du plus petit label ---*/
    min = labels[0];
    for ( i = 1; i < nb ; i ++ ) 
	if ( min > labels[i] ) min = labels[i];
    /*--- un seul label ---*/
    if ( nb == 1 ) return( min );
    /*--- gestion des equivalences ---*/
    _COMPUTE_LABEL(m1,m2,min,table)
    for ( i = 0; i < nb ; i ++ ) {
	_COMPUTE_LABEL(i1,i2,labels[i],table)
	if ( table->tb[i1][i2].eq != min ) {
	    val = table->tb[i1][i2].eq;
	    table->tb[m1][m2].sz += table->tb[i1][i2].sz;
	    if ( table->tb[i1][i2].ok == TRUE ) table->tb[m1][m2].ok = TRUE;
	    table->tb[i1][i2].sz = 0;
	    table->tb[i1][i2].eq = min;
	    for ( j = (labels[i] + 1); j <= table->nb; j ++ ) {
		_COMPUTE_LABEL(i1,i2,j,table)
		if ( table->tb[i1][i2].eq == val ) table->tb[i1][i2].eq = min;
	    }
	}
    }
    return( min );
}

/* Test des voisins : autres plan
                      derniere ligne
		      colonne de gauche
*/

#if defined(_ANSI_)
static int _CON_VoisinsApDlCg( unsigned short int *buf,
			       _con_te *table,
			       int *labels,
			       int *offset,
			       e_ParamConnex *par )
#else
static int _CON_VoisinsApDlCg( buf, table, labels, offset, par )
unsigned short int *buf;
_con_te *table;
int *labels;
int *offset;
e_ParamConnex *par;
#endif
{
    register int i, j, i1, i2, m1, m2;
    int nb = 0;
    int min, val;
    
    /*--- test des voisins ---*/
    switch ( par->connexite ) {
    case CONNEXE_26 :
	if ( *(buf + offset[2]) > 0 ) labels[nb++] = *(buf + offset[2]);
    case CONNEXE_18 :
	if ( *(buf + offset[1]) > 0 ) labels[nb++] = *(buf + offset[1]);
	if ( *(buf + offset[5]) > 0 ) labels[nb++] = *(buf + offset[5]);
    case CONNEXE_10 :
     	if ( *(buf + offset[4]) > 0 ) labels[nb++] = *(buf + offset[4]);
    case CONNEXE_08 :
     	if ( *(buf + offset[11]) > 0 ) labels[nb++] = *(buf + offset[11]);
    case CONNEXE_04 :
     	if ( *(buf + offset[10]) > 0 ) labels[nb++] = *(buf + offset[10]);
	break;
    case CONNEXE_06 :
     	if ( *(buf + offset[4]) > 0 ) labels[nb++] = *(buf + offset[4]);
     	if ( *(buf + offset[10]) > 0 ) labels[nb++] = *(buf + offset[10]);
    }
    /*--- pas de voisins ---*/
    if ( nb == 0 ) return( 0 );
    /*--- recherche des labels equivalents ---*/
    for ( i = 0; i < nb ; i ++ ) {
	_COMPUTE_LABEL(i1,i2,labels[i],table)
	labels[i] = table->tb[i1][i2].eq;
    }
    /*--- recherche du plus petit label ---*/
    min = labels[0];
    for ( i = 1; i < nb ; i ++ ) 
	if ( min > labels[i] ) min = labels[i];
    /*--- un seul label ---*/
    if ( nb == 1 ) return( min );
    /*--- gestion des equivalences ---*/
    _COMPUTE_LABEL(m1,m2,min,table)
    for ( i = 0; i < nb ; i ++ ) {
	_COMPUTE_LABEL(i1,i2,labels[i],table)
	if ( table->tb[i1][i2].eq != min ) {
	    val = table->tb[i1][i2].eq;
	    table->tb[m1][m2].sz += table->tb[i1][i2].sz;
	    if ( table->tb[i1][i2].ok == TRUE ) table->tb[m1][m2].ok = TRUE;
	    table->tb[i1][i2].sz = 0;
	    table->tb[i1][i2].eq = min;
	    for ( j = (labels[i] + 1); j <= table->nb; j ++ ) {
		_COMPUTE_LABEL(i1,i2,j,table)
		if ( table->tb[i1][i2].eq == val ) table->tb[i1][i2].eq = min;
	    }
	}
    }
    return( min );
}

/* Test des voisins : autres plan
                      derniere ligne
		      colonnes du centre 
*/

#if defined(_ANSI_)
static int _CON_VoisinsApDlCe( unsigned short int *buf,
			       _con_te *table,
			       int *labels,
			       int *offset,
			       e_ParamConnex *par )
#else
static int _CON_VoisinsApDlCe( buf, table, labels, offset, par )
unsigned short int *buf;
_con_te *table;
int *labels;
int *offset;
e_ParamConnex *par;
#endif
{
    register int i, j, i1, i2, m1, m2;
    int nb = 0;
    int min, val;
    
    /*--- test des voisins ---*/
    switch ( par->connexite ) {
    case CONNEXE_26 :
	if ( *(buf + offset[0]) > 0 ) labels[nb++] = *(buf + offset[0]);
	if ( *(buf + offset[2]) > 0 ) labels[nb++] = *(buf + offset[2]);
    case CONNEXE_18 :
	if ( *(buf + offset[1]) > 0 ) labels[nb++] = *(buf + offset[1]);
	if ( *(buf + offset[3]) > 0 ) labels[nb++] = *(buf + offset[3]);
	if ( *(buf + offset[5]) > 0 ) labels[nb++] = *(buf + offset[5]);
    case CONNEXE_10 :
     	if ( *(buf + offset[4]) > 0 ) labels[nb++] = *(buf + offset[4]);
    case CONNEXE_08 :
     	if ( *(buf + offset[9]) > 0 ) labels[nb++] = *(buf + offset[9]);
     	if ( *(buf + offset[11]) > 0 ) labels[nb++] = *(buf + offset[11]);
    case CONNEXE_04 :
     	if ( *(buf + offset[10]) > 0 ) labels[nb++] = *(buf + offset[10]);
     	if ( *(buf + offset[12]) > 0 ) labels[nb++] = *(buf + offset[12]);
	break;
    case CONNEXE_06 :
     	if ( *(buf + offset[4]) > 0 ) labels[nb++] = *(buf + offset[4]);
     	if ( *(buf + offset[10]) > 0 ) labels[nb++] = *(buf + offset[10]);
     	if ( *(buf + offset[12]) > 0 ) labels[nb++] = *(buf + offset[12]);
    }
    /*--- pas de voisins ---*/
    if ( nb == 0 ) return( 0 );
    /*--- recherche des labels equivalents ---*/
    for ( i = 0; i < nb ; i ++ ) {
	_COMPUTE_LABEL(i1,i2,labels[i],table)
	labels[i] = table->tb[i1][i2].eq;
    }
    /*--- recherche du plus petit label ---*/
    min = labels[0];
    for ( i = 1; i < nb ; i ++ ) 
	if ( min > labels[i] ) min = labels[i];
    /*--- un seul label ---*/
    if ( nb == 1 ) return( min );
    /*--- gestion des equivalences ---*/
    _COMPUTE_LABEL(m1,m2,min,table)
    for ( i = 0; i < nb ; i ++ ) {
	_COMPUTE_LABEL(i1,i2,labels[i],table)
	if ( table->tb[i1][i2].eq != min ) {
	    val = table->tb[i1][i2].eq;
	    table->tb[m1][m2].sz += table->tb[i1][i2].sz;
	    if ( table->tb[i1][i2].ok == TRUE ) table->tb[m1][m2].ok = TRUE;
	    table->tb[i1][i2].sz = 0;
	    table->tb[i1][i2].eq = min;
	    for ( j = (labels[i] + 1); j <= table->nb; j ++ ) {
		_COMPUTE_LABEL(i1,i2,j,table)
		if ( table->tb[i1][i2].eq == val ) table->tb[i1][i2].eq = min;
	    }
	}
    }
    return( min );
}

/* Test des voisins : autres plan
                      derniere ligne
		      colonne de droite
*/

#if defined(_ANSI_)
static int _CON_VoisinsApDlCd( unsigned short int *buf,
			       _con_te *table,
			       int *labels,
			       int *offset,
			       e_ParamConnex *par )
#else
static int _CON_VoisinsApDlCd( buf, table, labels, offset, par )
unsigned short int *buf;
_con_te *table;
int *labels;
int *offset;
e_ParamConnex *par;
#endif
{
    register int i, j, i1, i2, m1, m2;
    int nb = 0;
    int min, val;
    
    /*--- test des voisins ---*/
    switch ( par->connexite ) {
    case CONNEXE_26 :
	if ( *(buf + offset[0]) > 0 ) labels[nb++] = *(buf + offset[0]);
    case CONNEXE_18 :
	if ( *(buf + offset[1]) > 0 ) labels[nb++] = *(buf + offset[1]);
	if ( *(buf + offset[3]) > 0 ) labels[nb++] = *(buf + offset[3]);
    case CONNEXE_10 :
     	if ( *(buf + offset[4]) > 0 ) labels[nb++] = *(buf + offset[4]);
    case CONNEXE_08 :
     	if ( *(buf + offset[9]) > 0 ) labels[nb++] = *(buf + offset[9]);
    case CONNEXE_04 :
     	if ( *(buf + offset[10]) > 0 ) labels[nb++] = *(buf + offset[10]);
     	if ( *(buf + offset[12]) > 0 ) labels[nb++] = *(buf + offset[12]);
	break;
    case CONNEXE_06 :
     	if ( *(buf + offset[4]) > 0 ) labels[nb++] = *(buf + offset[4]);
     	if ( *(buf + offset[10]) > 0 ) labels[nb++] = *(buf + offset[10]);
     	if ( *(buf + offset[12]) > 0 ) labels[nb++] = *(buf + offset[12]);
    }
    /*--- pas de voisins ---*/
    if ( nb == 0 ) return( 0 );
    /*--- recherche des labels equivalents ---*/
    for ( i = 0; i < nb ; i ++ ) {
	_COMPUTE_LABEL(i1,i2,labels[i],table)
	labels[i] = table->tb[i1][i2].eq;
    }
    /*--- recherche du plus petit label ---*/
    min = labels[0];
    for ( i = 1; i < nb ; i ++ ) 
	if ( min > labels[i] ) min = labels[i];
    /*--- un seul label ---*/
    if ( nb == 1 ) return( min );
    /*--- gestion des equivalences ---*/
    _COMPUTE_LABEL(m1,m2,min,table)
    for ( i = 0; i < nb ; i ++ ) {
	_COMPUTE_LABEL(i1,i2,labels[i],table)
	if ( table->tb[i1][i2].eq != min ) {
	    val = table->tb[i1][i2].eq;
	    table->tb[m1][m2].sz += table->tb[i1][i2].sz;
	    if ( table->tb[i1][i2].ok == TRUE ) table->tb[m1][m2].ok = TRUE;
	    table->tb[i1][i2].sz = 0;
	    table->tb[i1][i2].eq = min;
	    for ( j = (labels[i] + 1); j <= table->nb; j ++ ) {
		_COMPUTE_LABEL(i1,i2,j,table)
		if ( table->tb[i1][i2].eq == val ) table->tb[i1][i2].eq = min;
	    }
	}
    }
    return( min );
}

/* Realise un tri sur les composantes connexes.

   Le resultat est un tableau dont les composantes 
   connexes sont rangees par taille decroissante.
*/

#if defined(_ANSI_)
static void _CON_Tri_CC( e_CC *tab, int left, int right )
#else
static void _CON_Tri_CC( tab, left, right )
e_CC *tab;
int left;
int right;
#endif
{
    int i, last;
    e_CC tmp;

    if ( left >= right ) return;

    /* _CON_Swap_CC( tab, left, (left+right)/2 ); */
    tmp = tab[left];   tab[left] = tab[(left+right)/2];   tab[(left+right)/2] = tmp;

    last = left;

    for ( i = left+1; i <= right; i++ )       
	if ( tab[i].size > tab[left].size ) {
	    /* _CON_Swap_CC( tab, ++last, i ); */
	    tmp = tab[++last];   tab[last] = tab[i];   tab[i] = tmp;
	}

    /* _CON_Swap_CC( tab, left, last ); */
    tmp = tab[left];   tab[left] = tab[last];   tab[last] = tmp;

    _CON_Tri_CC( tab, left, last-1 );
    _CON_Tri_CC( tab, last+1, right );
}

/*
#if defined(_ANSI_)
static void _CON_Swap_CC( e_CC *tab, int left, int right )
#else
static void _CON_Swap_CC( tab, left, right )
e_CC *tab;
int left;
int right;
#endif
{
    e_CC tmp;
    tmp = tab[left];
    tab[left] = tab[right];
    tab[right] = tmp;
}
*/


