/* epi_connex.h 2.2 5/14/98 */
#ifndef _epi_connex_h_
#define _epi_connex_h_

#define CONNEXE_26  1
#define CONNEXE_6   2
#define CONNEXE_06  2
#define CONNEXE_18  3
#define CONNEXE_10  4
#define CONNEXE_4  11
#define CONNEXE_04 11
#define CONNEXE_8  12
#define CONNEXE_08 12

/* Structure de parametres pour les recherches de composantes connexes.

STRUCTURE
   Cette structure contient les champs suivants :

   float lowThres : seuil bas

   float higThres : seuil haut
   
   int minSize    : taille minimum requise des composantes connexes

   int maxSize    : variable booleenne

   int maxNbCC    : nombre maximal de composantes connexes

   int connexite  : type de connexite

   int mode       : dimension du traitement

   int verbose    : si TRUE, ecriture de quelques informations

FIELDS
   Pour la recherche de composantes connexes et la reconstruction,
   les images sont considerees comme binaire grace au seuil bas
   (le passage aux valeurs entieres se fait avec la normalisation
   de T_Get3Point et T_Set3Point). Ce seuil bas est egalement le
   seuil bas du seuillage par hysteresis.

   Le seuil haut ne sert que pour le seuillage par hysteresis.

   Le champ minSize determine la taille minimale des composantes
   connexes a retenir.

   Si le champ maxSize est a TRUE, on ne conserve que la plus grande
   composante connexe.

   Le champ maxNbCC donne le nombre maximal de composantes connexes
   a retenir. Comme elles seront numerotees par ordre de taille
   decroissante, on ne conserve que les maxNbCC plus grandes.
   Mettre maxSize a TRUE equivaut donc a mettre maxNbCC a 1.
   Si ce champ a une valeur negative ou nulle, il n'y a pas 
   de limitations au nombre de composantes connexes.

   Le champ connexite donne la relation de voisinage (de connexite)
   a employer. Il peut prendre les valeurs :

   - CONNEXE_26 : relation 3D, 9 voisins au dessus, 8 dans le plan
                  9 voisins au dessous
   
   - CONNEXE_18 : relation 3D, 5 voisins au dessus, 8 dans le plan
                  5 voisins au dessous
   
   - CONNEXE_10 : relation 3D, 1 voisins au dessus, 8 dans le plan
                  1 voisins au dessous
   
   - CONNEXE_6 : relation 3D, 1 voisins au dessus, 4 dans le plan
                 1 voisins au dessous
   
   - CONNEXE_8 : relation 2D, 8 voisins dans le plan
   
   - CONNEXE_4 : relation 2D, 4 voisins dans le plan

   Le champ mode donne la dimension du traitement. Il peut
   prendre les valeurs MODE_3D ou MODE_2D. S'il est a MODE_2D,
   et que l'image est 3D, chaque plan de l'image est traitee
   comme etant une image independante. Par exemple, si maxNbCC
   est a 1, on conserve une composante connexe par plan.
*/   
   
typedef struct e_ParamConnex{
	float lowThres;      /* low threshold */
	float higThres;      /* high threshold */
	int minSize;         /* minimum size of connected component */
	int maxSize;         /* TRUE if component of maximum size is required */
        int maxNbCC;         /* maximal number of connected components */
	int connexite;       /* type of the connectivity */
	int mode;            /* mode 2D ou 3D */
        int verbose;
} e_ParamConnex, *e_ParamConnexPtr;

/* structure decrivant une composante connexe.

   label est le numero de ladite composante

   size est sa taille (nombre de points).
*/
typedef struct e_CC{
    int label;
    int size;
} e_CC;

/* structure resultat pour le calcul des composantes connexes.

   nb est le nombre de composantes connexes retenues

   si bool_alloc est a TRUE, le tableau tab sera alloue
   et contiendra la description des nb composantes
   connexes (tab[nb] contient la description
   de la nb-ieme composante, c'est donc un tableau
   de nb+1 elements).
*/
typedef struct e_CC_results{  
    int nb;
    int bool_alloc;
    e_CC *tab;
} e_CC_results;

#endif
