/* epi_distance.c 2.2 5/14/98 */

#include <epidaure.h>
#include <epidaure.ei>
#include <epidaure.ee>

#ifndef NO_PROTO
static int E_IntDistance( t_ImagePtr theIm, t_ImagePtr resIm, e_ParamDistancePtr par );
static int E_IDtraiteDCsg( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par );
static int E_IDtraiteDLs(  unsigned short *vl, int i, int *offset, e_ParamDistancePtr par );
static int E_IDtraiteDCsd( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par );
static int E_IDtraiteDCg(  unsigned short *vl, int i, int *offset, e_ParamDistancePtr par );
#ifdef COMPILE_UNUSED
static int E_IDtraiteDCe(  unsigned short *vl, int i, int *offset, e_ParamDistancePtr par );
static int E_IDtraiteRCe(  unsigned short *vl, int i, int *offset, e_ParamDistancePtr par );
#endif /* COMPILE_UNUSED */
static int E_IDtraiteDCd(  unsigned short *vl, int i, int *offset, e_ParamDistancePtr par );
static int E_IDtraiteDCig( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par );
static int E_IDtraiteDLi(  unsigned short *vl, int i, int *offset, e_ParamDistancePtr par );
static int E_IDtraiteDCid( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par );
static int E_IDtraiteRCid( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par );
static int E_IDtraiteRLi(  unsigned short *vl, int i, int *offset, e_ParamDistancePtr par );
static int E_IDtraiteRCig( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par );
static int E_IDtraiteRCd(  unsigned short *vl, int i, int *offset, e_ParamDistancePtr par );
static int E_IDtraiteRCg(  unsigned short *vl, int i, int *offset, e_ParamDistancePtr par );
static int E_IDtraiteRCsd( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par );
static int E_IDtraiteRLs(  unsigned short *vl, int i, int *offset, e_ParamDistancePtr par );
static int E_IDtraiteRCsg( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par );
static int E_V3DNorme2( short int x, short int y , short int z );
static int E_V3DtraiteAp( short int *vx, short int *vy , short int *vz, int x, int y, int z, int dim[4], int dz );
static int E_V3DtraiteMpHG( short int *vx, short int *vy , short int *vz, int x, int y, int z, int dim[4] );
static int E_V3DtraiteMpBD( short int *vx, short int *vy , short int *vz, int x, int y, int z, int dim[4] );
static int E_V3DtraiteMpG( short int *vx, short int *vy , short int *vz, int x, int y, int z, int dim[4] );
static int E_V3DtraiteMpD( short int *vx, short int *vy , short int *vz, int x, int y, int z, int dim[4] );
static void _VD_Seuil( t_ImagePtr theIm, t_ImagePtr resIm, e_ParamDistancePtr par );
static void _ID_Seuil( t_ImagePtr theIm, t_ImagePtr resIm, int *offset, e_ParamDistancePtr par );
#else 
static int E_IntDistance();
static int E_IDtraiteDCsg();
static int E_IDtraiteDLs();
static int E_IDtraiteDCsd();
static int E_IDtraiteDCg();
static int E_IDtraiteDCe();
static int E_IDtraiteDCd();
static int E_IDtraiteDCig();
static int E_IDtraiteDLi();
static int E_IDtraiteDCid();
static int E_IDtraiteRCid();
static int E_IDtraiteRLi();
static int E_IDtraiteRCig();
static int E_IDtraiteRCd();
static int E_IDtraiteRCe();
static int E_IDtraiteRCg();
static int E_IDtraiteRCsd();
static int E_IDtraiteRLs();
static int E_IDtraiteRCsg();
static int E_V3DNorme2();
static int E_V3DtraiteAp();
static int E_V3DtraiteMpHG();
static int E_V3DtraiteMpBD();
static int E_V3DtraiteMpG();
static int E_V3DtraiteMpD();
static void _VD_Seuil();
static void _ID_Seuil();
#endif

#define UINFINI 65535
#define SINFINI 26753 
/* on definit SINFINI a 26753 au lieu de 32767, car 3*32767*32767 = 3221028867
   depasse les capacites de l'entier 2^31-1 = 2147483647. on choisit donc
   l'infini comme (int)sqrt(2147483647/3)-1
*/

/* Calcul de distance : valeurs par defaut pour la structure de parametres.

   Le champ float thres de la structure est mis a 0.5
   Le champ int choix de la structure de parametre est mis a DIST_CHAMFER
   Le champ int mode de la structure de parametre est mis a MODE_3D
   Les 5 elements du champ int inc[5] sont mis a 1
   Les 3 elements du champ t_3Point hm sont mis a 1.0
   
*/   

#if defined(_ANSI_)
void E_InitDistance( e_ParamDistancePtr par /* parameters */ )
#else
void E_InitDistance( par )
e_ParamDistancePtr par; /* parameters */ 
#endif
{
	/*------------------------------------------------------------*/
	/*                                                            */
	/*   set the default parameters for distance computation      */
	/*                                                            */
	/*------------------------------------------------------------*/
	par->thres = 0.5;
	par->choix = DIST_CHAMFER;
	par->mode = MODE_3D;
	par->inc[0] = par->inc[1] = par->inc[2] = par->inc[3] = par->inc[4] = 1;
	par->hm.x = par->hm.y = par->hm.z = (reeltype)1.0;
}

/* Calcul d'une image de distance.

   Le champ thres de la structure de parametre definit le seuil pour
   binariser l'image. Les points dont la valeur est superieure ou
   egale au seuil definissent l'objet. Attention, cette valeur
   de seuil est sujette aux normalisations de T_Get3Point.

   Le champ choix de la structure de parametre definit le type
   de distance :
   - DIST_6 : 6-distance, les points partageant une face avec le 
     point central en sont a une distance de 1.
   - DIST_18 : 18-distance, les points partageant une face ou une arete
     avec le point central en sont a une distance de 1.
   - DIST_26 : 26-distance, les points partageant une face, une arete
     ou un sommet avec le point central en sont a une distance de 1.
   - DIST_CHAMFER : distance de chamfer. Coefficients (16, 21, 26)
     puis division par 16.
   - DIST_CHAMFER2 : distance de chamfer. Coefficients (16, 21, 26)
     sans division. La distance obtenue est donc 16 fois la "vraie"
     distance. 
   - DIST_CHAMFER3 : distance de chamfer. Coefficients (32, 42, 52)
     pour le calcul. Toutefois les points 6-voisins, 18-voisins et
     26-voisins de l'objet en sont a une distance de 16, 21 et 26.
     Pas de division ensuite. La distance obtenue est donc 32 fois 
     la "vraie" distance. 
   - DIST_EUCLI : distance euclidienne (algorithme de Danielsson).
     On calcule d'abord les 3 images formant les vecteurs designant 
     les points les plus proches, puis on en prend la norme.
     Ce calcul est forcement 3D.
   - DIST_SQR_EUCLI : distance euclidienne au carre (algorithme de 
     Danielsson). On calcule d'abord les 3 images formant les vecteurs 
     designant les points les plus proches, puis on en prend la norme
     au carre. Ce calcul est forcement 3D.
   - DIST_4 : 4-distance = 6-distance dans le plan.
   - DIST_8 : 8-distance = 18- ou 26-distance dans le plan.
   - DIST_CHAMFER_2D : distance de chamfer DIST_CHAMFER dans le plan.
   - DIST_CHAMFER2_2D : distance de chamfer DIST_CHAMFER2 dans le plan.
   - DIST_CHAMFER3_2D : distance de chamfer DIST_CHAMFER3 dans le plan.
   - DIST_USER : increments specifies par l'utilisateur (ce calcul est
     forcement 3D).

   Le champ mode specifie la dimension du traitement. S'il est egal a
   MODE_2D, le calcul se fera dans le plan : le choix (dans le champ
   du meme nom peut etre modifie), sinon on respectera le choix fait.

   Les 5 elements du champ inc[5] specifient les increments pour le calcul
   de la distance (facon chamfer) si le choix est DIST_USER, ils concernent
   - inc[0] : voisin par une face dans le plan
   - inc[1] : voisin par une arete dans le plan
   - inc[2] : voisin par une face dans le plan superieur ou inferieur
   - inc[3] : voisin par une arete dans le plan superieur ou inferieur
   - inc[4] : voisin par un sommet dans le plan superieur ou inferieur
   
RETURN
   Renvoie FALSE en cas d'erreur.
*/

#if defined(_ANSI_)
int E_Distance( t_ImagePtr theIm /* input image */, 
	       t_ImagePtr resIm /* result image */, 
	       e_ParamDistancePtr par /* parameters */ )
#else
int E_Distance( theIm, resIm, par )
t_ImagePtr theIm;       /* input image */
t_ImagePtr resIm;       /* result image */
e_ParamDistancePtr par; /* parameters */ 
#endif
{
  t_Image auxIm, auxX, auxY, auxZ;
  register int i, x, y, z, n;
  unsigned short *vl;
  short int *vx, *vy, *vz;
  
  E_ERR( (theIm==NULL) || (theIm->buf==NULL), E_ERRNULL, FALSE,
	"E_Distance", "input image not allocated");
  E_ERR( (resIm==NULL) || (resIm->buf==NULL), E_ERRNULL, FALSE,
	"E_Distance", "output image not allocated");
  E_ERR( (theIm->tx!=resIm->tx) || (theIm->ty!=resIm->ty) || (theIm->tz!=resIm->tz),
	E_ERRFATAL, FALSE, "E_Distance", 
	"input and output have different dimensions");
  E_ERR( (theIm->tx < 3) || (theIm->ty < 3), E_ERRFATAL, FALSE, "E_Distance", 
	"images have bad dimensions");

  switch ( par->choix) {
  case DIST_6 :
  case DIST_18 :
  case DIST_26 :
  case DIST_CHAMFER :
  case DIST_CHAMFER2 :
  case DIST_CHAMFER3 :
  case DIST_4 :
  case DIST_8 :
  case DIST_CHAMFER_2D :
  case DIST_CHAMFER2_2D :
  case DIST_CHAMFER3_2D :
  case DIST_USER :
	  T_Init3Image( &auxIm, theIm->tx, theIm->ty, theIm->tz, TYPEIM_16B, CNULL );
	  if ( (resIm->type != TYPEIM_16B) && (resIm->type != TYPEIM_S16B) )
		  T_AllocImage( &auxIm );
	  else
		  auxIm.buf = resIm->buf;
	  E_REPERR( goto END_E_Distance );
	  
	  /*--- calcul effectif dans l'image auxilliaire ---*/
	  if ( E_IntDistance( theIm, &auxIm, par ) == FALSE ) {
		  if ( (resIm->type != TYPEIM_16B) && (resIm->type != TYPEIM_S16B) )
			  T_FreeImage( &auxIm );
		  return( FALSE );
	  }
	  /*--- ecriture de l'image de sortie ---*/
	  if ( resIm->type != TYPEIM_16B) {
		  vl = (unsigned short*)auxIm.buf;
		  i = 0;
		  switch (resIm->type) {
		  case TYPEIM_BW :
			  for (z=0; z<resIm->tz; z++)
			  for (y=0; y<resIm->ty; y++)
			  for (x=0; x<resIm->tx; x++) {
				  T_SetIPoint( x, y, z, resIm,
					      (int)(*(vl + i)), MODE_SET );
				  i++;
			  }
			  break;
		  case TYPEIM_256 :
			  {
			  unsigned char *buf;
			  buf = (unsigned char*)(resIm->buf);
			  for (z=0; z<resIm->tz; z++)
			  for (y=0; y<resIm->ty; y++)
			  for (x=0; x<resIm->tx; x++) {
				  if ( *(vl+i) > 255 ) *(buf+i) = (unsigned char)255;
				  else *(buf+i) = (unsigned char)(*(vl+i));
				  i++;
			  }}
			  break;
		  case TYPEIM_S16B :
			  {
			  short int *buf;
			  buf = (short int*)(resIm->buf);
			  for (z=0; z<resIm->tz; z++)
			  for (y=0; y<resIm->ty; y++)
			  for (x=0; x<resIm->tx; x++) {
				  if ( *(vl+i) > 32767 ) *(buf+i) = (short int)32767;
				  else *(buf+i) = (short int)(*(vl+i));
				  i++;
			  }}
			  break;
		  case TYPEIM_FLOAT :
			  {
			  float *buf;
			  buf = (float*)(resIm->buf);
			  for (z=0; z<resIm->tz; z++)
			  for (y=0; y<resIm->ty; y++)
			  for (x=0; x<resIm->tx; x++) {
				  *(buf+i) = (float)(*(vl+i));
			  }}
			  break;
		  default :
			  if ( (resIm->type != TYPEIM_16B) && (resIm->type != TYPEIM_S16B) )
				  T_FreeImage( &auxIm );
			  E_XERR( E_ERRMISMATCH, FALSE, "E_Distance", "output has bad type");
		  }
	  }
	  if ( (resIm->type != TYPEIM_16B) && (resIm->type != TYPEIM_S16B) )
		  T_FreeImage( &auxIm );
	  return( TRUE );
	  break;
  case DIST_EUCLI :
  case DIST_SQR_EUCLI :
	  T_Init3Image( &auxX, theIm->tx, theIm->ty, theIm->tz, TYPEIM_S16B, CNULL );
	  T_Init3Image( &auxY, theIm->tx, theIm->ty, theIm->tz, TYPEIM_S16B, CNULL );
	  T_Init3Image( &auxZ, theIm->tx, theIm->ty, theIm->tz, TYPEIM_S16B, CNULL );
	  T_AllocImage( &auxX );
	  T_AllocImage( &auxY );
	  if ( (resIm->type != TYPEIM_16B) && (resIm->type != TYPEIM_S16B) )
		  T_AllocImage( &auxZ );
	  else
		  auxZ.buf = resIm->buf;
	  E_REPERR( goto END_E_Distance );
	  
	  E_Vec3Distance( theIm, &auxX, &auxY, &auxZ, par );
	  
	  /*   ecriture de l'image de sortie   */
	  /*************************************/
	  vx = (short int*)auxX.buf;
	  vy = (short int*)auxY.buf;
	  vz = (short int*)auxZ.buf;
	  /*************************************************************/
	  /*   le point le plus proche du point (x,y,z) est le point   */
	  /*   coordonnees                                             */
	  /*   ( x + T_GetIPoint( x, y, z, &auxX),                     */
	  /*     y + T_GetIPoint( x, y, z, &auxY),                     */
	  /*     z + T_GetIPoint( x, y, z, &auxZ) )                    */
	  /*************************************************************/
	  i = 0;
	  switch (resIm->type) {
	  case TYPEIM_BW :
		  for (z=0; z<resIm->tz; z++)
		  for (y=0; y<resIm->ty; y++)
		  for (x=0; x<resIm->tx; x++) {
			  n = E_V3DNorme2( *(vx+i), *(vy+i), *(vz+i) );
			  if (par->choix == DIST_EUCLI ) n = (int)(sqrt((double)n) + 0.5);
			  T_SetIPoint( x, y, z, resIm, n, MODE_SET );
			  i++;
		  }
		  break;
	  case TYPEIM_256 :
		  {
		  unsigned char *buf;
		  buf = (unsigned char*)(resIm->buf);
		  for (z=0; z<resIm->tz; z++)
		  for (y=0; y<resIm->ty; y++)
		  for (x=0; x<resIm->tx; x++) {
			  n = E_V3DNorme2( *(vx+i), *(vy+i), *(vz+i) );
			  if (par->choix == DIST_EUCLI ) n = (int)(sqrt((double)n) + 0.5);
			  if ( n > 255 ) *(buf+i) = (unsigned char)255;
			  else *(buf+i) = (unsigned char)(n);
			  i++;
		  }}
		  break;
	  case TYPEIM_16B :
		  {
		  unsigned short int *buf;
		  buf = (unsigned short int*)(resIm->buf);
		  for (z=0; z<resIm->tz; z++)
		  for (y=0; y<resIm->ty; y++)
		  for (x=0; x<resIm->tx; x++) {
			  n = E_V3DNorme2( *(vx+i), *(vy+i), *(vz+i) );
			  if (par->choix == DIST_EUCLI ) n = (int)(sqrt((double)n) + 0.5);
			  if ( n > 65535 ) *(buf+i) = (unsigned short int)65535;
			  else *(buf+i) = (unsigned short int)(n);
			  i++;
		  }}
		  break;
	  case TYPEIM_S16B :
		  {
		  short int *buf;
		  buf = (short int*)(resIm->buf);
		  for (z=0; z<resIm->tz; z++)
		  for (y=0; y<resIm->ty; y++)
		  for (x=0; x<resIm->tx; x++) {
			  n = E_V3DNorme2( *(vx+i), *(vy+i), *(vz+i) );
			  if (par->choix == DIST_EUCLI ) n = (int)(sqrt((double)n) + 0.5);
			  if ( n > 32767 ) *(buf+i) = (short int)32767;
			  else *(buf+i) = (short int)(n);
			  i++;
		  }}
		  break;
	  case TYPEIM_FLOAT :
		  {
		  float *buf;
		  buf = (float*)(resIm->buf);
		  for (z=0; z<resIm->tz; z++)
		  for (y=0; y<resIm->ty; y++)
		  for (x=0; x<resIm->tx; x++) {
			  n = E_V3DNorme2( *(vx+i), *(vy+i), *(vz+i) );
			  if (par->choix == DIST_EUCLI ) 
			      *(buf+i) = (float)sqrt((double)n);
			  else 
			      *(buf+i) = (float)n;
			  i++;
		  }}
		  break;
	  default :
		  T_FreeImage( &auxX );
		  T_FreeImage( &auxY );
		  if ( (resIm->type != TYPEIM_16B) && (resIm->type != TYPEIM_S16B) )
			  T_FreeImage( &auxZ );
		  E_XERR( E_ERRMISMATCH, FALSE, "E_Distance", "output has bad type");
	  }
	  T_FreeImage( &auxX );
	  T_FreeImage( &auxY );
	  if ( (resIm->type != TYPEIM_16B) && (resIm->type != TYPEIM_S16B) )
		  T_FreeImage( &auxZ );
	  return( TRUE );
  default :
	  E_XERR( E_ERRNULL, FALSE, "E_Distance", "error in parameters" );
  }

END_E_Distance:
  switch ( par->choix) {
  case DIST_6 :
  case DIST_18 :
  case DIST_26 :
  case DIST_CHAMFER :
  case DIST_CHAMFER2 :
  case DIST_CHAMFER3 :
  case DIST_4 :
  case DIST_8 :
  case DIST_CHAMFER_2D :
  case DIST_CHAMFER2_2D :
  case DIST_CHAMFER3_2D :
  case DIST_USER :
	  if ( (resIm->type != TYPEIM_16B) && (resIm->type != TYPEIM_S16B) )
		  T_FreeImage( &auxIm );
	  break;
  case DIST_EUCLI :
  case DIST_SQR_EUCLI :
	  T_FreeImage( &auxX );
	  T_FreeImage( &auxY );
	  if ( (resIm->type != TYPEIM_16B) && (resIm->type != TYPEIM_S16B) )
		  T_FreeImage( &auxZ );
  }
  E_ISERR( FALSE, "E_Distance", (char*)NULL );  
  return( TRUE );
}

#if defined(_ANSI_)
static int E_IntDistance( t_ImagePtr theIm, t_ImagePtr resIm, e_ParamDistancePtr par )
#else
static int E_IntDistance( theIm, resIm, par )
t_ImagePtr theIm;
t_ImagePtr resIm;
e_ParamDistancePtr par ;
#endif
{
  int offset[27];
  register int i, x, y, z, test;
  int dimx, dimy, dimz, plan, local_mode;
  unsigned short *vl;
  e_ParamDistance local_par;
  
  E_ERR( (theIm==NULL) || (theIm->buf==NULL), E_ERRNULL, FALSE,
	"E_IntDistance", "input image not allocated");
  E_ERR( (resIm==NULL) || (resIm->buf==NULL), E_ERRNULL, FALSE,
	"E_IntDistance", "output image not allocated");
  E_ERR( (resIm->type != TYPEIM_16B), E_ERRNULL, FALSE,"E_IntDistance",
	"output image has a bad type");
  E_ERR( (theIm->tx < 3) || (theIm->ty < 3), E_ERRFATAL, FALSE, "E_IntDistance", 
	"images have bad dimensions");
  E_ERR( (theIm->tx!=resIm->tx) || (theIm->ty!=resIm->ty) || (theIm->tz!=resIm->tz),
	E_ERRFATAL, FALSE, "E_IntDistance", 
	"input and output have different dimensions");
  
  /*--- mise a jour des parametres locaux : resolution des conflits ---*/
  local_par = *par;
  if ( local_par.mode == MODE_2D ) {
	  switch ( local_par.choix ) {
	  case DIST_6 :
		  local_par.choix = DIST_6_2D;
		  break;
	  case DIST_18 :
		  local_par.choix = DIST_18_2D;
		  break;
	  case DIST_26 :
		  local_par.choix = DIST_26_2D;
		  break;
	  case DIST_CHAMFER :
		  local_par.choix = DIST_CHAMFER_2D;
		  break;
	  case DIST_CHAMFER2 :
		  local_par.choix = DIST_CHAMFER2_2D;
		  break;
	  case DIST_CHAMFER3 :
		  local_par.choix = DIST_CHAMFER3_2D;
	  }
  }
  switch ( local_par.choix ) {
  case DIST_4 :
  case DIST_8 :
  case DIST_CHAMFER_2D :
  case DIST_CHAMFER2_2D :
  case DIST_CHAMFER3_2D :
	  local_par.mode = MODE_2D;
  }
  /*--- specification des increments ---*/
  switch ( local_par.choix ) {
  case DIST_6 :
  case DIST_18 :
  case DIST_26 :	  
  case DIST_4 :
  case DIST_8 :
	  local_par.inc[0] = 1;
	  local_par.inc[1] = 1;
	  local_par.inc[2] = 1;
	  local_par.inc[3] = 1;
	  local_par.inc[4] = 1;
	  break;
  case DIST_CHAMFER :
  case DIST_CHAMFER2 :
  case DIST_CHAMFER_2D :
  case DIST_CHAMFER2_2D :
	  local_par.inc[0] = 16;
	  local_par.inc[1] = 21;
	  local_par.inc[2] = 16;
	  local_par.inc[3] = 21;
	  local_par.inc[4] = 26;
	  break;
  case DIST_CHAMFER3 :
  case DIST_CHAMFER3_2D :
	  local_par.inc[0] = 32;
	  local_par.inc[1] = 42;
	  local_par.inc[2] = 32;
	  local_par.inc[3] = 42;
	  local_par.inc[4] = 52;
  }

  /*--- pre-calcul des offsets ---*/
  dimx = theIm->tx;
  dimy = theIm->ty;
  dimz = theIm->tz;
  plan = dimx*dimy;
  offset[0]  = - plan - dimx - 1;   /* point (-1, -1, -1) */
  offset[1]  = - plan - dimx;       /* point ( 0, -1, -1) */
  offset[2]  = - plan - dimx + 1;   /* point ( 1, -1, -1) */
  offset[3]  = - plan - 1;          /* point (-1,  0, -1) */
  offset[4]  = - plan;              /* point ( 0,  0, -1) */
  offset[5]  = - plan + 1;          /* point ( 1,  0, -1) */
  offset[6]  = - plan + dimx - 1;   /* point (-1,  1, -1) */
  offset[7]  = - plan + dimx;       /* point ( 0,  1, -1) */
  offset[8]  = - plan + dimx + 1;   /* point ( 1,  1, -1) */
  offset[9]  =        - dimx - 1;   /* point (-1, -1,  0) */
  offset[10] =        - dimx;       /* point ( 0, -1,  0) */
  offset[11] =        - dimx + 1;   /* point ( 1, -1,  0) */
  offset[12] =        - 1;          /* point (-1,  0,  0) */
  offset[13] =          0;          /* point ( 0,  0,  0) */
  offset[14] =          1;          /* point ( 1,  0,  0) */
  offset[15] =          dimx - 1;   /* point (-1,  1,  0) */
  offset[16] =          dimx;       /* point ( 0,  1,  0) */
  offset[17] =          dimx + 1;   /* point ( 1,  1,  0) */
  offset[18] =   plan - dimx - 1;   /* point (-1, -1,  1) */
  offset[19] =   plan - dimx;       /* point ( 0, -1,  1) */
  offset[20] =   plan - dimx + 1;   /* point ( 1, -1,  1) */
  offset[21] =   plan - 1;          /* point (-1,  0,  1) */
  offset[22] =   plan;              /* point ( 0,  0,  1) */
  offset[23] =   plan + 1;          /* point ( 1,  0,  1) */
  offset[24] =   plan + dimx - 1;   /* point (-1,  1,  1) */
  offset[25] =   plan + dimx;       /* point ( 0,  1,  1) */
  offset[26] =   plan + dimx + 1;   /* point ( 1,  1,  1) */

  /*--- dernieres affectations ---*/
  local_mode = local_par.mode;
  vl = (unsigned short*)(resIm->buf);
  
  /*--- preparation de l'image d'entree ---*/
  _ID_Seuil( theIm, resIm, offset, &local_par );

  /*   boucle de la descente   */
  /*****************************/
  local_par.mode = MODE_2D;
  i = 0;
  for (z=0; z<dimz; z++) {
	  /*--- coin superieur gauche ---*/
	  *(vl+i) = (unsigned short)E_IDtraiteDCsg(vl, i, offset, &local_par);
	  i++;
	  /*--- ligne superieure ---*/
	  for (x=1; x<dimx-1; x++) {
		  *(vl+i) = (unsigned short)E_IDtraiteDLs(vl, i, offset, &local_par);
		  i++;
	  }
	  /*--- coin superieur droit ---*/
	  *(vl+i) = (unsigned short)E_IDtraiteDCsd(vl, i, offset, &local_par);
	  i++;
	  /*--- autres lignes ---*/
	  for (y=1; y<dimy-1; y++) {
		  /*--- autres lignes : colonne de gauche ---*/
		  *(vl+i) = (unsigned short)E_IDtraiteDCg(vl, i, offset, &local_par);
		  i++;
		  /*--- autres lignes : centre ---*/
		  for (x=1; x<dimx-1; x++) {
			  /*--- *(vl+i) = (unsigned short)E_IDtraiteDCe(vl, i, offset, &local_par); ---*/
			  test = *(vl+i);
			  switch ( local_par.choix ) {
			  case DIST_CHAMFER :
			  case DIST_CHAMFER2 :
			  case DIST_CHAMFER3 :
			  case DIST_CHAMFER_2D :
			  case DIST_CHAMFER2_2D :
			  case DIST_CHAMFER3_2D :
			  case DIST_USER :
			  case DIST_26 :
			  case DIST_18 :
			  case DIST_8 :
				  test = EPI_MIN( test, local_par.inc[1] + *(vl+i+offset[9]) );    /* (-1, -1,  0) */
				  test = EPI_MIN( test, local_par.inc[1] + *(vl+i+offset[11]) );   /* ( 1, -1,  0) */
			  case DIST_6 :
			  case DIST_4 :
				  test = EPI_MIN( test, local_par.inc[0] + *(vl+i+offset[10]) );   /* ( 0, -1,  0) */
				  test = EPI_MIN( test, local_par.inc[0] + *(vl+i+offset[12]) );   /* (-1,  0,  0) */ 
			  }
			  if ( local_par.mode == MODE_2D ) { *(vl+i) = test; i++; continue; }
			  switch ( local_par.choix ) {
			  case DIST_CHAMFER :
			  case DIST_CHAMFER2 :
			  case DIST_CHAMFER3 :
			  case DIST_USER :
			  case DIST_26 :
				  test = EPI_MIN( test, local_par.inc[4] + *(vl+i+offset[0]) );    /* (-1, -1, -1) */
				  test = EPI_MIN( test, local_par.inc[4] + *(vl+i+offset[2]) );    /* ( 1, -1, -1) */
				  test = EPI_MIN( test, local_par.inc[4] + *(vl+i+offset[6]) );    /* (-1,  1, -1) */
				  test = EPI_MIN( test, local_par.inc[4] + *(vl+i+offset[8]) );    /* ( 1,  1, -1) */
			  case DIST_18 :
				  test = EPI_MIN( test, local_par.inc[3] + *(vl+i+offset[1]) );    /* ( 0, -1, -1) */
				  test = EPI_MIN( test, local_par.inc[3] + *(vl+i+offset[3]) );    /* (-1,  0, -1) */
				  test = EPI_MIN( test, local_par.inc[3] + *(vl+i+offset[5]) );    /* ( 1,  0, -1) */
				  test = EPI_MIN( test, local_par.inc[3] + *(vl+i+offset[7]) );    /* ( 0,  1, -1) */
			  case DIST_6 :
				  test = EPI_MIN( test, local_par.inc[2] + *(vl+i+offset[4]) );    /* ( 0,  0, -1) */
			  }
			  *(vl+i) = test; i++;
		  }
		  /*--- autres lignes : colonne de droite ---*/
		  *(vl+i) = (unsigned short)E_IDtraiteDCd(vl, i, offset, &local_par);
		  i++;
	  }
	  /*--- coin inferieur gauche ---*/
	  *(vl+i) = (unsigned short)E_IDtraiteDCig(vl, i, offset, &local_par);
	  i++;
	  
	  /*--- ligne inferieure ---*/
	  for (x=1; x<dimx-1; x++) {
		  *(vl+i) = (unsigned short)E_IDtraiteDLi(vl, i, offset, &local_par);
		  i++;
	  }
	  /*--- coin inferieur droit ---*/
	  *(vl+i) = (unsigned short)E_IDtraiteDCid(vl, i, offset, &local_par);
	  i++;
	  local_par.mode = local_mode;
  }
  
  /*   boucle de la remontee   */
  /*****************************/
  local_par.mode = MODE_2D;
  i = dimz*plan-1;
  for (z=0; z<dimz; z++) {
	  /*--- coin inferieur droit ---*/
	  *(vl+i) = (unsigned short)E_IDtraiteRCid(vl, i, offset, &local_par);
	  i--;
	  /*--- ligne inferieure ---*/
	  for (x=1; x<dimx-1; x++) {
		  *(vl+i) = (unsigned short)E_IDtraiteRLi(vl, i, offset, &local_par);
		  i--;
	  }
	  /*--- coin inferieur gauche ---*/
	  *(vl+i) = (unsigned short)E_IDtraiteRCig(vl, i, offset, &local_par);
	  i--;
	  /*--- autres lignes ---*/
	  for (y=1; y<dimy-1; y++) {
		  /*--- autres lignes : colonne droite ---*/
		  *(vl+i) = (unsigned short)E_IDtraiteRCd(vl, i, offset, &local_par);
		  i--;
		  /*--- autres lignes : centre ---*/
		  for (x=1; x<dimx-1; x++) {
			  /*--- *(vl+i) = (unsigned short)E_IDtraiteRCe(vl, i, offset, &local_par); ---*/
			  test = *(vl+i);
			  switch ( local_par.choix ) {
			  case DIST_CHAMFER :
			  case DIST_CHAMFER2 :
			  case DIST_CHAMFER3 :
			  case DIST_CHAMFER_2D :
			  case DIST_CHAMFER2_2D :
			  case DIST_CHAMFER3_2D :
			  case DIST_USER :
			  case DIST_26 :
			  case DIST_18 :
			  case DIST_8 :
				  test = EPI_MIN( test, local_par.inc[1] + *(vl+i+offset[15]) );   /* (-1,  1,  0) */
				  test = EPI_MIN( test, local_par.inc[1] + *(vl+i+offset[17]) );   /* ( 1,  1,  0) */
			  case DIST_6 :
			  case DIST_4 :
				  test = EPI_MIN( test, local_par.inc[0] + *(vl+i+offset[14]) );   /* ( 1,  0,  0) */ 
				  test = EPI_MIN( test, local_par.inc[0] + *(vl+i+offset[16]) );   /* ( 0,  1,  0) */
			  }
			  if ( local_par.mode == MODE_2D ) { *(vl+i) = test; i--; continue; }
			  switch ( local_par.choix ) {
			  case DIST_CHAMFER :
			  case DIST_CHAMFER2 :
			  case DIST_CHAMFER3 :
			  case DIST_USER :
			  case DIST_26 :
				  test = EPI_MIN( test, local_par.inc[4] + *(vl+i+offset[18]) );   /* (-1, -1,  1) */
				  test = EPI_MIN( test, local_par.inc[4] + *(vl+i+offset[20]) );   /* ( 1, -1,  1) */
				  test = EPI_MIN( test, local_par.inc[4] + *(vl+i+offset[24]) );   /* (-1,  1,  1) */
				  test = EPI_MIN( test, local_par.inc[4] + *(vl+i+offset[26]) );   /* ( 1,  1,  1) */
			  case DIST_18 :
				  test = EPI_MIN( test, local_par.inc[3] + *(vl+i+offset[19]) );   /* ( 0, -1,  1) */
				  test = EPI_MIN( test, local_par.inc[3] + *(vl+i+offset[21]) );   /* (-1,  0,  1) */
				  test = EPI_MIN( test, local_par.inc[3] + *(vl+i+offset[23]) );   /* ( 1,  0,  1) */
				  test = EPI_MIN( test, local_par.inc[3] + *(vl+i+offset[25]) );   /* ( 0,  1,  1) */
			  case DIST_6 :
				  test = EPI_MIN( test, local_par.inc[2] + *(vl+i+offset[22]) );   /* ( 0,  0,  1) */
			  }
			  *(vl+i) = test; i--; 
		  }
		  /*--- autres lignes : colonne de gauche ---*/
		  *(vl+i) = (unsigned short)E_IDtraiteRCg(vl, i, offset, &local_par);
		  i--;
	  }
	  /*--- coin superieur droit ---*/
	  *(vl+i) = (unsigned short)E_IDtraiteRCsd(vl, i, offset, &local_par);
	  i--;
	  /*--- ligne superieure ---*/
	  for (x=1; x<dimx-1; x++) {
		  *(vl+i) = (unsigned short)E_IDtraiteRLs(vl, i, offset, &local_par);
		  i--;
	  }
	  /*--- coin superieur gauche ---*/
	  *(vl+i) = (unsigned short)E_IDtraiteRCsg(vl, i, offset, &local_par);
	  i--;
	  local_par.mode = local_mode;
  }
  
  if ( ( par->choix == DIST_CHAMFER ) || ( par->choix == DIST_CHAMFER_2D ) ) {
	  /*--- division par 16 ---*/
	  for (i=0; i<dimz*plan; i++)
		  *(vl+i) = *(vl+i) >> 4;
  }
  
  return( TRUE );
}

#if defined(_ANSI_)
static int E_IDtraiteDCsg( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par )
#else
static int E_IDtraiteDCsg(vl, i, offset, par )
unsigned short *vl;
int i;
int *offset;
e_ParamDistancePtr par ;
#endif
{
	/*--- boucle de la descente ---*/
	/*--- coin superieur gauche ---*/
	int test;
	test = *(vl+i);
  
	if ( par->mode == MODE_2D ) return( test );
	
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_USER :
	case DIST_26 :
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[8]) );    /* ( 1,  1, -1) */
	case DIST_18 :
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[5]) );    /* ( 1,  0, -1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[7]) );    /* ( 0,  1, -1) */
	case DIST_6 :
		test = EPI_MIN( test, par->inc[2] + *(vl+i+offset[4]) );    /* ( 0,  0, -1) */
	}
	
	return( test );
}

#if defined(_ANSI_)
static int E_IDtraiteDLs( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par )
#else
static int E_IDtraiteDLs(vl, i, offset, par )
unsigned short *vl;
int i;
int *offset;
e_ParamDistancePtr par ;
#endif
{
	/*--- boucle de la descente  ---*/
	/*--- ligne superieure       ---*/
	int test;
	test = *(vl+i);
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_CHAMFER_2D :
	case DIST_CHAMFER2_2D :
	case DIST_CHAMFER3_2D :
	case DIST_USER :
	case DIST_26 :
	case DIST_18 :
	case DIST_8 :
	case DIST_6 :
	case DIST_4 :
		test = EPI_MIN( test, par->inc[0] + *(vl+i+offset[12]) );   /* (-1,  0,  0) */ 
	}
	
	if ( par->mode == MODE_2D ) return( test );
  
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_USER :
	case DIST_26 :
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[6]) );    /* (-1,  1, -1) */
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[8]) );    /* ( 1,  1, -1) */
	case DIST_18 :
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[3]) );    /* (-1,  0, -1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[5]) );    /* ( 1,  0, -1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[7]) );    /* ( 0,  1, -1) */
	case DIST_6 :
		test = EPI_MIN( test, par->inc[2] + *(vl+i+offset[4]) );    /* ( 0,  0, -1) */
	}
  
	return( test );
}

#if defined(_ANSI_)
static int E_IDtraiteDCsd( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par )
#else
static int E_IDtraiteDCsd(vl, i, offset, par )
unsigned short *vl;
int i;
int *offset;
e_ParamDistancePtr par ;
#endif
{
	/*--- boucle de la descente  ---*/
	/*--- cois superieur droit   ---*/
	int test;
	test = *(vl+i);
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_CHAMFER_2D :
	case DIST_CHAMFER2_2D :
	case DIST_CHAMFER3_2D :
	case DIST_USER :
	case DIST_26 :
	case DIST_18 :
	case DIST_8 :
	case DIST_6 :
	case DIST_4 :
		test = EPI_MIN( test, par->inc[0] + *(vl+i+offset[12]) );   /* (-1,  0,  0) */ 
	}
	
	if ( par->mode == MODE_2D ) return( test );
  
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_USER :
	case DIST_26 :
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[6]) );    /* (-1,  1, -1) */
	case DIST_18 :
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[3]) );    /* (-1,  0, -1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[7]) );    /* ( 0,  1, -1) */
	case DIST_6 :
		test = EPI_MIN( test, par->inc[2] + *(vl+i+offset[4]) );    /* ( 0,  0, -1) */
	}
	
	return( test );
}

#if defined(_ANSI_)
static int E_IDtraiteDCg( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par )
#else
static int E_IDtraiteDCg(vl, i, offset, par )
unsigned short *vl;
int i;
int *offset;
e_ParamDistancePtr par ;
#endif
{
	/*--- boucle de la descente  ---*/
	/*--- colonne de gauche      ---*/
	int test;
	test = *(vl+i);
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_CHAMFER_2D :
	case DIST_CHAMFER2_2D :
	case DIST_CHAMFER3_2D :
	case DIST_USER :
	case DIST_26 :
	case DIST_18 :
	case DIST_8 :
		test = EPI_MIN( test, par->inc[1] + *(vl+i+offset[11]) );   /* ( 1, -1,  0) */
	case DIST_6 :
	case DIST_4 :
		test = EPI_MIN( test, par->inc[0] + *(vl+i+offset[10]) );   /* ( 0, -1,  0) */
	}

	if ( par->mode == MODE_2D ) return( test );
  
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_USER :
	case DIST_26 :
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[2]) );    /* ( 1, -1, -1) */
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[8]) );    /* ( 1,  1, -1) */
	case DIST_18 :
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[1]) );    /* ( 0, -1, -1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[5]) );    /* ( 1,  0, -1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[7]) );    /* ( 0,  1, -1) */
	case DIST_6 :
		test = EPI_MIN( test, par->inc[2] + *(vl+i+offset[4]) );    /* ( 0,  0, -1) */
	}

	return( test );
}

#ifdef COMPILE_UNUSED

#if defined(_ANSI_)
static int E_IDtraiteDCe( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par )
#else
static int E_IDtraiteDCe(vl, i, offset, par )
unsigned short *vl;
int i;
int *offset;
e_ParamDistancePtr par ;
#endif
{
	/*--- boucle de la descente  ---*/
	/*--- autres lignes : centre ---*/
	int test;
	test = *(vl+i);
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_CHAMFER_2D :
	case DIST_CHAMFER2_2D :
	case DIST_CHAMFER3_2D :
	case DIST_USER :
	case DIST_26 :
	case DIST_18 :
	case DIST_8 :
		test = EPI_MIN( test, par->inc[1] + *(vl+i+offset[9]) );    /* (-1, -1,  0) */
		test = EPI_MIN( test, par->inc[1] + *(vl+i+offset[11]) );   /* ( 1, -1,  0) */
	case DIST_6 :
	case DIST_4 :
		test = EPI_MIN( test, par->inc[0] + *(vl+i+offset[10]) );   /* ( 0, -1,  0) */
		test = EPI_MIN( test, par->inc[0] + *(vl+i+offset[12]) );   /* (-1,  0,  0) */ 
	}
	
	if ( par->mode == MODE_2D ) return( test );
	
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_USER :
	case DIST_26 :
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[0]) );    /* (-1, -1, -1) */
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[2]) );    /* ( 1, -1, -1) */
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[6]) );    /* (-1,  1, -1) */
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[8]) );    /* ( 1,  1, -1) */
	case DIST_18 :
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[1]) );    /* ( 0, -1, -1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[3]) );    /* (-1,  0, -1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[5]) );    /* ( 1,  0, -1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[7]) );    /* ( 0,  1, -1) */
	case DIST_6 :
		test = EPI_MIN( test, par->inc[2] + *(vl+i+offset[4]) );    /* ( 0,  0, -1) */
	}
	
	return( test );
}

#endif /* COMPILE_UNUSED */


#if defined(_ANSI_)
static int E_IDtraiteDCd( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par )
#else
static int E_IDtraiteDCd(vl, i, offset, par )
unsigned short *vl;
int i;
int *offset;
e_ParamDistancePtr par ;
#endif
{
 	/*--- boucle de la descente  ---*/
	/*--- colonne de droite      ---*/
	int test;
	test = *(vl+i);
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_CHAMFER_2D :
	case DIST_CHAMFER2_2D :
	case DIST_CHAMFER3_2D :
	case DIST_USER :
	case DIST_26 :
	case DIST_18 :
	case DIST_8 :
		test = EPI_MIN( test, par->inc[1] + *(vl+i+offset[9]) );    /* (-1, -1,  0) */
	case DIST_6 :
	case DIST_4 :
		test = EPI_MIN( test, par->inc[0] + *(vl+i+offset[10]) );   /* ( 0, -1,  0) */
		test = EPI_MIN( test, par->inc[0] + *(vl+i+offset[12]) );   /* (-1,  0,  0) */ 
	}
	
	if ( par->mode == MODE_2D ) return( test );
  
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_USER :
	case DIST_26 :
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[0]) );    /* (-1, -1, -1) */
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[6]) );    /* (-1,  1, -1) */
	case DIST_18 :
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[1]) );    /* ( 0, -1, -1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[3]) );    /* (-1,  0, -1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[7]) );    /* ( 0,  1, -1) */
	case DIST_6 :
		test = EPI_MIN( test, par->inc[2] + *(vl+i+offset[4]) );    /* ( 0,  0, -1) */
	}
  
	return( test );
}

#if defined(_ANSI_)
static int E_IDtraiteDCig( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par )
#else
static int E_IDtraiteDCig(vl, i, offset, par )
unsigned short *vl;
int i;
int *offset;
e_ParamDistancePtr par ;
#endif
{
	/*--- boucle de la descente  ---*/
	/*--- coin inferieur gauche  ---*/
	int test;
	test = *(vl+i);
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_CHAMFER_2D :
	case DIST_CHAMFER2_2D :
	case DIST_CHAMFER3_2D :
	case DIST_USER :
	case DIST_26 :
	case DIST_18 :
	case DIST_8 :
		test = EPI_MIN( test, par->inc[1] + *(vl+i+offset[11]) );   /* ( 1, -1,  0) */
	case DIST_6 :
	case DIST_4 :
		test = EPI_MIN( test, par->inc[0] + *(vl+i+offset[10]) );   /* ( 0, -1,  0) */
	}
	
	if ( par->mode == MODE_2D ) return( test );
	
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_USER :
	case DIST_26 :
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[2]) );    /* ( 1, -1, -1) */
	case DIST_18 :
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[1]) );    /* ( 0, -1, -1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[5]) );    /* ( 1,  0, -1) */
	case DIST_6 :
		test = EPI_MIN( test, par->inc[2] + *(vl+i+offset[4]) );    /* ( 0,  0, -1) */
	}

	return( test );
}

#if defined(_ANSI_)
static int E_IDtraiteDLi( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par )
#else
static int E_IDtraiteDLi(vl, i, offset, par )
unsigned short *vl;
int i;
int *offset;
e_ParamDistancePtr par ;
#endif
{
	/*--- boucle de la descente  ---*/
	/*--- ligne inferieure       ---*/
	int test;
	test = *(vl+i);
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_CHAMFER_2D :
	case DIST_CHAMFER2_2D :
	case DIST_CHAMFER3_2D :
	case DIST_USER :
	case DIST_26 :
	case DIST_18 :
	case DIST_8 :
		test = EPI_MIN( test, par->inc[1] + *(vl+i+offset[9]) );    /* (-1, -1,  0) */
		test = EPI_MIN( test, par->inc[1] + *(vl+i+offset[11]) );   /* ( 1, -1,  0) */
	case DIST_6 :
	case DIST_4 :
		test = EPI_MIN( test, par->inc[0] + *(vl+i+offset[10]) );   /* ( 0, -1,  0) */
		test = EPI_MIN( test, par->inc[0] + *(vl+i+offset[12]) );   /* (-1,  0,  0) */ 
	}
	
	if ( par->mode == MODE_2D ) return( test );
  
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_USER :
	case DIST_26 :
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[0]) );    /* (-1, -1, -1) */
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[2]) );    /* ( 1, -1, -1) */
	case DIST_18 :
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[1]) );    /* ( 0, -1, -1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[3]) );    /* (-1,  0, -1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[5]) );    /* ( 1,  0, -1) */
	case DIST_6 :
		test = EPI_MIN( test, par->inc[2] + *(vl+i+offset[4]) );    /* ( 0,  0, -1) */
	}
  
	return( test );
}

#if defined(_ANSI_)
static int E_IDtraiteDCid( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par )
#else
static int E_IDtraiteDCid(vl, i, offset, par )
unsigned short *vl;
int i;
int *offset;
e_ParamDistancePtr par ;
#endif
{
	/*--- boucle de la descente  ---*/
	/*--- coin inferieur droit   ---*/
	int test;
	test = *(vl+i);
	switch ( par->choix ) {
 	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_CHAMFER_2D :
	case DIST_CHAMFER2_2D :
	case DIST_CHAMFER3_2D :
	case DIST_USER :
	case DIST_26 :
	case DIST_18 :
	case DIST_8 :
		test = EPI_MIN( test, par->inc[1] + *(vl+i+offset[9]) );    /* (-1, -1,  0) */
	case DIST_6 :
	case DIST_4 :
		test = EPI_MIN( test, par->inc[0] + *(vl+i+offset[10]) );   /* ( 0, -1,  0) */
		test = EPI_MIN( test, par->inc[0] + *(vl+i+offset[12]) );   /* (-1,  0,  0) */ 
	}
	
	if ( par->mode == MODE_2D ) return( test );
	  
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_USER :
	case DIST_26 :
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[0]) );    /* (-1, -1, -1) */
	case DIST_18 :
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[1]) );    /* ( 0, -1, -1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[3]) );    /* (-1,  0, -1) */
	case DIST_6 :
		test = EPI_MIN( test, par->inc[2] + *(vl+i+offset[4]) );    /* ( 0,  0, -1) */
	}

	return( test );
}

#if defined(_ANSI_)
static int E_IDtraiteRCid( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par )
#else
static int E_IDtraiteRCid(vl, i, offset, par )
unsigned short *vl;
int i;
int *offset;
e_ParamDistancePtr par ;
#endif
{
	/*--- boucle de la remontee  ---*/
	/*--- coin inferieur droit   ---*/
	int test;
	test = *(vl+i);
  
	if ( par->mode == MODE_2D ) return( test );
	
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_USER :
	case DIST_26 :
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[18]) );   /* (-1, -1,  1) */
	case DIST_18 :
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[19]) );   /* ( 0, -1,  1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[21]) );   /* (-1,  0,  1) */
	case DIST_6 :
		test = EPI_MIN( test, par->inc[2] + *(vl+i+offset[22]) );   /* ( 0,  0,  1) */
	}
	
	return( test );
}

#if defined(_ANSI_)
static int E_IDtraiteRLi( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par )
#else
static int E_IDtraiteRLi(vl, i, offset, par )
unsigned short *vl;
int i;
int *offset;
e_ParamDistancePtr par ;
#endif
{
	/*--- boucle de la remontee  ---*/
	/*--- ligne inferieure       ---*/
	int test;
	test = *(vl+i);
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_CHAMFER_2D :
	case DIST_CHAMFER2_2D :
	case DIST_CHAMFER3_2D :
	case DIST_USER :
	case DIST_26 :
	case DIST_18 :
	case DIST_8 :
	case DIST_6 :
	case DIST_4 :
		test = EPI_MIN( test, par->inc[0] + *(vl+i+offset[14]) );   /* ( 1,  0,  0) */ 
	}
	
	if ( par->mode == MODE_2D ) return( test );
	
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_USER :
	case DIST_26 :
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[18]) );   /* (-1, -1,  1) */
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[20]) );   /* ( 1, -1,  1) */
	case DIST_18 :
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[19]) );   /* ( 0, -1,  1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[21]) );   /* (-1,  0,  1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[23]) );   /* ( 1,  0,  1) */
	case DIST_6 :
		test = EPI_MIN( test, par->inc[2] + *(vl+i+offset[22]) );   /* ( 0,  0,  1) */
	}
  
	return( test );
}

#if defined(_ANSI_)
static int E_IDtraiteRCig( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par )
#else
static int E_IDtraiteRCig(vl, i, offset, par )
unsigned short *vl;
int i;
int *offset;
e_ParamDistancePtr par ;
#endif
{
	/*--- boucle de la remontee  ---*/
	/*--- coin inferieur gauche  ---*/
	int test;
	test = *(vl+i);
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_CHAMFER_2D :
	case DIST_CHAMFER2_2D :
	case DIST_CHAMFER3_2D :
	case DIST_USER :
	case DIST_26 :
	case DIST_18 :
	case DIST_8 :
	case DIST_6 :
	case DIST_4 :
		test = EPI_MIN( test, par->inc[0] + *(vl+i+offset[14]) );   /* ( 1,  0,  0) */ 
	}

	if ( par->mode == MODE_2D ) return( test );
  
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_USER :
	case DIST_26 :
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[20]) );   /* ( 1, -1,  1) */
	case DIST_18 :
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[19]) );   /* ( 0, -1,  1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[23]) );   /* ( 1,  0,  1) */
	case DIST_6 :
		test = EPI_MIN( test, par->inc[2] + *(vl+i+offset[22]) );   /* ( 0,  0,  1) */
	}
	
	return( test );
}

#if defined(_ANSI_)
static int E_IDtraiteRCd( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par )
#else
static int E_IDtraiteRCd(vl, i, offset, par )
unsigned short *vl;
int i;
int *offset;
e_ParamDistancePtr par ;
#endif
{
	/*--- boucle de la remontee  ---*/
	/*--- colonne de droite      ---*/
	int test;
	test = *(vl+i);
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_CHAMFER_2D :
	case DIST_CHAMFER2_2D :
	case DIST_CHAMFER3_2D :
	case DIST_USER :
	case DIST_26 :
	case DIST_18 :
	case DIST_8 :
		test = EPI_MIN( test, par->inc[1] + *(vl+i+offset[15]) );   /* (-1,  1,  0) */
	case DIST_6 :
	case DIST_4 :
		test = EPI_MIN( test, par->inc[0] + *(vl+i+offset[16]) );   /* ( 0,  1,  0) */
	}
	
	if ( par->mode == MODE_2D ) return( test );

	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_USER :
	case DIST_26 :
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[18]) );   /* (-1, -1,  1) */
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[24]) );   /* (-1,  1,  1) */
	case DIST_18 :
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[19]) );   /* ( 0, -1,  1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[21]) );   /* (-1,  0,  1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[25]) );   /* ( 0,  1,  1) */
	case DIST_6 :
		test = EPI_MIN( test, par->inc[2] + *(vl+i+offset[22]) );   /* ( 0,  0,  1) */
	}

	return( test );
}


#ifdef COMPILE_UNUSED
#if defined(_ANSI_)
static int E_IDtraiteRCe( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par )
#else
static int E_IDtraiteRCe(vl, i, offset, par )
unsigned short *vl;
int i;
int *offset;
e_ParamDistancePtr par ;
#endif
{
	/*--- boucle de la remontee  ---*/
	/*--- autres lignes : centre ---*/
	int test;
	test = *(vl+i);
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_CHAMFER_2D :
	case DIST_CHAMFER2_2D :
	case DIST_CHAMFER3_2D :
	case DIST_USER :
	case DIST_26 :
	case DIST_18 :
	case DIST_8 :
		test = EPI_MIN( test, par->inc[1] + *(vl+i+offset[15]) );   /* (-1,  1,  0) */
		test = EPI_MIN( test, par->inc[1] + *(vl+i+offset[17]) );   /* ( 1,  1,  0) */
	case DIST_6 :
	case DIST_4 :
		test = EPI_MIN( test, par->inc[0] + *(vl+i+offset[14]) );   /* ( 1,  0,  0) */ 
		test = EPI_MIN( test, par->inc[0] + *(vl+i+offset[16]) );   /* ( 0,  1,  0) */
	}
	
	if ( par->mode == MODE_2D ) return( test );
	
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_USER :
	case DIST_26 :
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[18]) );   /* (-1, -1,  1) */
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[20]) );   /* ( 1, -1,  1) */
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[24]) );   /* (-1,  1,  1) */
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[26]) );   /* ( 1,  1,  1) */
	case DIST_18 :
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[19]) );   /* ( 0, -1,  1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[21]) );   /* (-1,  0,  1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[23]) );   /* ( 1,  0,  1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[25]) );   /* ( 0,  1,  1) */
	case DIST_6 :
		test = EPI_MIN( test, par->inc[2] + *(vl+i+offset[22]) );   /* ( 0,  0,  1) */
	}
	
	return( test );
}
#endif /* COMPILE_UNUSED */


#if defined(_ANSI_)
static int E_IDtraiteRCg( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par )
#else
static int E_IDtraiteRCg(vl, i, offset, par )
unsigned short *vl;
int i;
int *offset;
e_ParamDistancePtr par ;
#endif
{
	/*--- boucle de la remontee  ---*/
	/*--- autres lignes : centre ---*/
	int test;
	test = *(vl+i);
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_CHAMFER_2D :
	case DIST_CHAMFER2_2D :
	case DIST_CHAMFER3_2D :
	case DIST_USER :
	case DIST_26 :
	case DIST_18 :
	case DIST_8 :
		test = EPI_MIN( test, par->inc[1] + *(vl+i+offset[17]) );   /* ( 1,  1,  0) */
	case DIST_6 :
	case DIST_4 :
		test = EPI_MIN( test, par->inc[0] + *(vl+i+offset[14]) );   /* ( 1,  0,  0) */ 
		test = EPI_MIN( test, par->inc[0] + *(vl+i+offset[16]) );   /* ( 0,  1,  0) */
	}
	
	if ( par->mode == MODE_2D ) return( test );

	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_USER :
	case DIST_26 :
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[20]) );   /* ( 1, -1,  1) */
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[26]) );   /* ( 1,  1,  1) */
	case DIST_18 :
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[19]) );   /* ( 0, -1,  1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[23]) );   /* ( 1,  0,  1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[25]) );   /* ( 0,  1,  1) */
	case DIST_6 :
		test = EPI_MIN( test, par->inc[2] + *(vl+i+offset[22]) );   /* ( 0,  0,  1) */
	}
  
	return( test );
}

#if defined(_ANSI_)
static int E_IDtraiteRCsd( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par )
#else
static int E_IDtraiteRCsd(vl, i, offset, par )
unsigned short *vl;
int i;
int *offset;
e_ParamDistancePtr par ;
#endif
{
	/*--- boucle de la remontee  ---*/
	/*--- coin superieur droit   ---*/
	int test;
	test = *(vl+i);
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_CHAMFER_2D :
	case DIST_CHAMFER2_2D :
	case DIST_CHAMFER3_2D :
	case DIST_USER :
	case DIST_26 :
	case DIST_18 :
	case DIST_8 :
		test = EPI_MIN( test, par->inc[1] + *(vl+i+offset[15]) );   /* (-1,  1,  0) */
	case DIST_6 :
	case DIST_4 :
		test = EPI_MIN( test, par->inc[0] + *(vl+i+offset[16]) );   /* ( 0,  1,  0) */
	}
	
	if ( par->mode == MODE_2D ) return( test );
	
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_USER :
	case DIST_26 :
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[24]) );   /* (-1,  1,  1) */
	case DIST_18 :
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[21]) );   /* (-1,  0,  1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[25]) );   /* ( 0,  1,  1) */
	case DIST_6 :
		test = EPI_MIN( test, par->inc[2] + *(vl+i+offset[22]) );   /* ( 0,  0,  1) */
	}
  
	return( test );
}

#if defined(_ANSI_)
static int E_IDtraiteRLs( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par )
#else
static int E_IDtraiteRLs(vl, i, offset, par )
unsigned short *vl;
int i;
int *offset;
e_ParamDistancePtr par ;
#endif
{
	/*--- boucle de la remontee  ---*/
	/*--- ligne superieure       ---*/
	int test;
	test = *(vl+i);
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_CHAMFER_2D :
	case DIST_CHAMFER2_2D :
	case DIST_CHAMFER3_2D :
	case DIST_USER :
	case DIST_26 :
	case DIST_18 :
	case DIST_8 :
		test = EPI_MIN( test, par->inc[1] + *(vl+i+offset[15]) );   /* (-1,  1,  0) */
		test = EPI_MIN( test, par->inc[1] + *(vl+i+offset[17]) );   /* ( 1,  1,  0) */
	case DIST_6 :
	case DIST_4 :
		test = EPI_MIN( test, par->inc[0] + *(vl+i+offset[14]) );   /* ( 1,  0,  0) */ 
		test = EPI_MIN( test, par->inc[0] + *(vl+i+offset[16]) );   /* ( 0,  1,  0) */
	}
	
	if ( par->mode == MODE_2D ) return( test );

	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_USER :
	case DIST_26 :
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[24]) );   /* (-1,  1,  1) */
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[26]) );   /* ( 1,  1,  1) */
	case DIST_18 :
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[21]) );   /* (-1,  0,  1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[23]) );   /* ( 1,  0,  1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[25]) );   /* ( 0,  1,  1) */
	case DIST_6 :
		test = EPI_MIN( test, par->inc[2] + *(vl+i+offset[22]) );   /* ( 0,  0,  1) */
	}
	
	return( test );
}

#if defined(_ANSI_)
static int E_IDtraiteRCsg( unsigned short *vl, int i, int *offset, e_ParamDistancePtr par )
#else
static int E_IDtraiteRCsg(vl, i, offset, par )
unsigned short *vl;
int i;
int *offset;
e_ParamDistancePtr par ;
#endif
{
	/*--- boucle de la remontee  ---*/
	/*--- coin superieur gauche  ---*/
	int test;
	test = *(vl+i);
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_CHAMFER_2D :
	case DIST_CHAMFER2_2D :
	case DIST_CHAMFER3_2D :
	case DIST_USER :
	case DIST_26 :
	case DIST_18 :
	case DIST_8 :
		test = EPI_MIN( test, par->inc[1] + *(vl+i+offset[17]) );   /* ( 1,  1,  0) */
	case DIST_6 :
	case DIST_4 :
		test = EPI_MIN( test, par->inc[0] + *(vl+i+offset[14]) );   /* ( 1,  0,  0) */ 
		test = EPI_MIN( test, par->inc[0] + *(vl+i+offset[16]) );   /* ( 0,  1,  0) */
	}
	
	if ( par->mode == MODE_2D ) return( test );
			
	switch ( par->choix ) {
	case DIST_CHAMFER :
	case DIST_CHAMFER2 :
	case DIST_CHAMFER3 :
	case DIST_USER :
	case DIST_26 :
		test = EPI_MIN( test, par->inc[4] + *(vl+i+offset[26]) );   /* ( 1,  1,  1) */
	case DIST_18 :
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[23]) );   /* ( 1,  0,  1) */
		test = EPI_MIN( test, par->inc[3] + *(vl+i+offset[25]) );   /* ( 0,  1,  1) */
	case DIST_6 :
		test = EPI_MIN( test, par->inc[2] + *(vl+i+offset[22]) );   /* ( 0,  0,  1) */
	}
  
	return( test );
}

/* Calcul des 3 images formant les vecteurs designant les points les plus proches.

   Le champ thres de la structure de parametre definit le seuil pour
   binariser l'image. Les points dont la valeur est superieure ou
   egale au seuil definissent l'objet. Attention, cette valeur
   de seuil est sujette aux normalisations de T_Get3Point.

   Les 3 images resVx, resVy et resVz doivent etre de type
   TYPEIM_S16B (short int).

   Apres calcul, le point le plus proche du point (x,y,z)  de l'image
   theIm est le point de coordonnees ( x + T_GetIPoint( x, y, z, resVx),
   y + T_GetIPoint( x, y, z, resVy), z + T_GetIPoint( x, y, z, resVz) ) 

RETURN
   Retourne FALSE en cas d'erreur.
*/

#if defined(_ANSI_)
int E_Vec3Distance( t_ImagePtr theIm /* input image */, 
		   t_ImagePtr resVx /* image of X coordinates of vectors pointing towards nearest points */, 
		   t_ImagePtr resVy /* image of Y coordinates of vectors pointing towards nearest points */, 
		   t_ImagePtr resVz /* image of Z coordinates of vectors pointing towards nearest points */, 
		   e_ParamDistancePtr par /* parameters */ )
#else
int E_Vec3Distance( theIm, resVx, resVy, resVz, par )
t_ImagePtr theIm;       /* input image */
t_ImagePtr resVx;       /* image of X coordinates of vectors pointing towards nearest points */
t_ImagePtr resVy;       /* image of Y coordinates of vectors pointing towards nearest points */
t_ImagePtr resVz;       /* image of Z coordinates of vectors pointing towards nearest points */
e_ParamDistancePtr par; /* parameters */
#endif
{
  int dim[4];
  register int i, x, y, z;
  int dimx, dimy;
  short int *vlx, *vly, *vlz;
  
  E_ERR( (theIm==NULL) || (theIm->buf==NULL), E_ERRNULL, FALSE,
	"E_Vec3Distance", "input image not allocated");
  E_ERR( (resVx==NULL) || (resVx->buf==NULL), E_ERRNULL, FALSE,
	"E_Vec3Distance", "output image not allocated");
  E_ERR( (resVy==NULL) || (resVy->buf==NULL), E_ERRNULL, FALSE,
	"E_Vec3Distance", "output image not allocated");
  E_ERR( (resVz==NULL) || (resVz->buf==NULL), E_ERRNULL, FALSE,
	"E_Vec3Distance", "output image not allocated");
  E_ERR( (resVx->type != TYPEIM_S16B), E_ERRNULL, FALSE,"E_Vec3Distance",
	"output images have bad type");
  E_ERR( (resVy->type != TYPEIM_S16B), E_ERRNULL, FALSE,"E_Vec3Distance",
	"output images have bad type");
  E_ERR( (resVz->type != TYPEIM_S16B), E_ERRNULL, FALSE,"E_Vec3Distance",
	"output images have bad type");
  E_ERR( (theIm->tx!=resVx->tx) || (theIm->ty!=resVx->ty) || (theIm->tz!=resVx->tz),
	E_ERRFATAL, FALSE, "E_Vec3Distance", 
	"input and output have different dimensions");
  E_ERR( (resVy->tx!=resVx->tx) || (resVy->ty!=resVx->ty) || (resVy->tz!=resVx->tz),
	E_ERRFATAL, FALSE, "E_Vec3Distance", 
	"output images have different dimensions");
  E_ERR( (resVy->tx!=resVz->tx) || (resVy->ty!=resVz->ty) || (resVy->tz!=resVz->tz),
	E_ERRFATAL, FALSE, "E_Vec3Distance", 
	"output images have different dimensions");
  E_ERR( (theIm->tx < 3) || (theIm->ty < 3), E_ERRFATAL, FALSE, "E_Vec3Distance", 
	"images have bad dimensions");
  
  dim[0] = dimx = theIm->tx;
  dim[1] = dimy = theIm->ty;
  dim[2] = theIm->tz;
  dim[3] = dimx*dimy;
  vlx = (short int*)resVx->buf;
  vly = (short int*)resVy->buf;
  vlz = (short int*)resVz->buf;

  /*--- seuillage ---*/
  _VD_Seuil( theIm, resVx, par );

  /*   boucle de la descente   */
  /*****************************/
  for (z=0; z<dim[2]; z++) {
	  for (y=0; y<dim[1]; y++) {
		  i = z*dim[3] + y*dim[0];
		  for (x=0; x<dim[0]; x++) {
			  *(vly + i) = *(vlz + i) = *(vlx + i);
			  if ( *(vlx + i) != 0 ) 
				  E_V3DtraiteAp( vlx, vly, vlz, x, y, z, dim, -1 );
			  i++;
		  }
		  i = z*dim[3] + y*dim[0];
		  for (x=0; x<dim[0]; x++) {
			  if ( *(vlx + i) || *(vly + i) || *(vlz + i) )
				  E_V3DtraiteMpHG( vlx, vly, vlz, x, y, z, dim );
			  i++;
		  }
		  i = z*dim[3] + (y+1)*dim[0] - 2;
		  for (x=dim[0]-2; x>=0; x--) {
			  if ( *(vlx + i) || *(vly + i) || *(vlz + i) )
				  E_V3DtraiteMpD( vlx, vly, vlz, x, y, z, dim );
			  i--;
		  }
	  }
	  i = (z+1)*dim[3] - 1;
	  for (y=dim[1]-1; y>=0; y--) {
		  for (x=dim[0]-1; x>=0; x--) {
			  if ( *(vlx + i) || *(vly + i) || *(vlz + i) )
				  E_V3DtraiteMpBD( vlx, vly, vlz, x, y, z, dim );
			  i--;
		  }
	  }
  }
  /*   boucle de la remontee   */
  /*****************************/
  for (z=dim[2]-1; z>=0; z--) {
	  for (y=dim[1]-1; y>=0; y--) {
		  i = z*dim[3] + (y+1)*dim[0] - 1;
		  for (x=dim[0]-1; x>=0; x--) {
			  if ( *(vlx + i) || *(vly + i) || *(vlz + i) )
				  E_V3DtraiteAp( vlx, vly, vlz, x, y, z, dim, 1 );
			  i--;
		  }
		  i = z*dim[3] + (y+1)*dim[0] - 1;
		  for (x=dim[0]-1; x>=0; x--) {
			  if ( *(vlx + i) || *(vly + i) || *(vlz + i) )
				  E_V3DtraiteMpBD( vlx, vly, vlz, x, y, z, dim );
			  i--;
		  }
		  i = z*dim[3] + y*dim[0] + 1;
		  for (x=1; x<dim[0]; x++) {
			  if ( *(vlx + i) || *(vly + i) || *(vlz + i) )
				  E_V3DtraiteMpG( vlx, vly, vlz, x, y, z, dim );
			  i++;
		  }
	  }
	  i = z*dim[3];
	  for (y=0; y<dim[1]; y++) {
		  for (x=0; x<dim[0]; x++) {
			  if ( *(vlx + i) || *(vly + i) || *(vlz + i) )
				  E_V3DtraiteMpHG( vlx, vly, vlz, x, y, z, dim );
			  i++;
		  }
	  }
  }
  return( TRUE );
}

#if defined(_ANSI_)
static int E_V3DNorme2( short int x, short int y , short int z )
#else
static int E_V3DNorme2(x, y , z )
short int x;
short int y ;
short int z ;
#endif
{
  int ax, ay, az;
  ax = x;
  ay = y;
  az = z;
  return( ax*ax + ay*ay + az*az );
}

#if defined(_ANSI_)
static int E_V3DtraiteAp( short int *vx, short int *vy , short int *vz, int x, int y, int z, int dim[4], int dz )
#else
static int E_V3DtraiteAp(vx, vy , vz, x, y, z, dim, dz )
short int *vx;
short int *vy ;
short int *vz;
int x;
int y;
int z;
int dim[4];
int dz ;
#endif
{
  register int i, dx, dy, ddx, ddy, ddz;
  int n,m;
  
  if ( (z+dz < 0) || (z+dz >= dim[2]) )
    return( TRUE );
  
  ddx = ddy = ddz = 0;
  i = z*dim[3] + y*dim[0] + x;
  n = E_V3DNorme2( *(vx+i), *(vy+i), *(vz+i) );
  i += dz*dim[3];
  
  for (dy= -1; dy<2; dy++)
    for (dx= -1; dx<2; dx++) {
      if ( (x+dx<0) || (x+dx>dim[0]-1) || (y+dy<0) || (y+dy>dim[1]-1) )
	continue;
      m = E_V3DNorme2( (short int)(*(vx+i+dy*dim[0]+dx)+dx), 
		      (short int)(*(vy+i+dy*dim[0]+dx)+dy),
		      (short int)(*(vz+i+dy*dim[0]+dx)+dz) );
      if (m >= n)
	continue;
      ddx = dx;
      ddy = dy;
      ddz = dz;
      n = m;
    }
  if (ddz) {
    *(vx+i-dz*dim[3]) = *(vx+i+ddy*dim[0]+ddx) + ddx;
    *(vy+i-dz*dim[3]) = *(vy+i+ddy*dim[0]+ddx) + ddy;
    *(vz+i-dz*dim[3]) = *(vz+i+ddy*dim[0]+ddx) + ddz;
  }
  return( TRUE );
}

#if defined(_ANSI_)
static int E_V3DtraiteMpHG( short int *vx, short int *vy , short int *vz, int x, int y, int z, int dim[4] )
#else
static int E_V3DtraiteMpHG(vx, vy , vz, x, y, z, dim)
short int *vx;
short int *vy ;
short int *vz;
int x;
int y;
int z;
int dim[4] ;
#endif
{
  register int i, dx, dy, ddx, ddy;
  int n,m;
  
  ddx = ddy = 0;
  i = z*dim[3] + y*dim[0] + x;
  n = E_V3DNorme2( *(vx+i), *(vy+i), *(vz+i) );
  
  if ( y > 0 ) {
    dy = -1;
    for (dx= -1; dx<2; dx++) {
      if ( (x+dx<0) || (x+dx>dim[0]-1) )
	continue;
      m = E_V3DNorme2( (short int)(*(vx+i-dim[0]+dx)+dx), 
		      (short int)(*(vy+i-dim[0]+dx)+dy),
		      *(vz+i-dim[0]+dx) );
      if (m >= n)
	continue;
      ddx = dx;
      ddy = dy;
      n = m;
    }
  }
  if ( x > 0 ) {
    dx = -1;
    dy = 0;
    m = E_V3DNorme2( (short int)(*(vx+i+dx)+dx), 
		    (short int)(*(vy+i+dx)+dy), 
		    *(vz+i+dx) );
    if (m < n) {
      ddx = dx;
      ddy = dy;
    }
  }
  if ( ddx || ddy ) {
    *(vx+i) = *(vx+i+ddy*dim[0]+ddx) + ddx;
    *(vy+i) = *(vy+i+ddy*dim[0]+ddx) + ddy;
    *(vz+i) = *(vz+i+ddy*dim[0]+ddx);
  }
  return( TRUE );
}

#if defined(_ANSI_)
static int E_V3DtraiteMpBD( short int *vx, short int *vy , short int *vz, int x, int y, int z, int dim[4] )
#else
static int E_V3DtraiteMpBD(vx, vy , vz, x, y, z, dim)
short int *vx;
short int *vy ;
short int *vz;
int x;
int y;
int z;
int dim[4] ;
#endif
{
  register int i, dx, dy, ddx, ddy;
  int n,m;
  
  ddx = ddy = 0;
  i = z*dim[3] + y*dim[0] + x;
  n = E_V3DNorme2( *(vx+i), *(vy+i), *(vz+i) );
  
  if ( y < dim[1]-1 ) {
    dy = 1;
    for (dx= -1; dx<2; dx++) {
      if ( (x+dx<0) || (x+dx>dim[0]-1) )
	continue;
      m = E_V3DNorme2( (short int)(*(vx+i+dim[0]+dx)+dx), 
		      (short int)(*(vy+i+dim[0]+dx)+dy),
		      *(vz+i+dim[0]+dx) );
      if (m >= n)
	continue;
      ddx = dx;
      ddy = dy;
      n = m;
    }
  }
  if ( x < dim[0]-1 ) {
    dx = 1;
    dy = 0;
    m = E_V3DNorme2( (short int)(*(vx+i+dx)+dx), 
		    *(vy+i+dx), *(vz+i+dx) );
    if (m < n) {
      ddx = dx;
      ddy = dy;
    }
  }
  if ( ddx || ddy ) {
    *(vx+i) = *(vx+i+ddy*dim[0]+ddx) + ddx;
    *(vy+i) = *(vy+i+ddy*dim[0]+ddx) + ddy;
    *(vz+i) = *(vz+i+ddy*dim[0]+ddx);
  }
  return( TRUE );
}

#if defined(_ANSI_)
static int E_V3DtraiteMpG( short int *vx, short int *vy , short int *vz, int x, int y, int z, int dim[4] )
#else
static int E_V3DtraiteMpG(vx, vy , vz, x, y, z, dim)
short int *vx;
short int *vy ;
short int *vz;
int x;
int y;
int z;
int dim[4] ;
#endif
{
  register int i, dx, ddx;
  int n,m;
  
  ddx = 0;
  i = z*dim[3] + y*dim[0] + x;
  n = E_V3DNorme2( *(vx+i), *(vy+i), *(vz+i) );
  
  dx = -1;
  m = E_V3DNorme2( (short int)(*(vx+i+dx)+dx), 
		  *(vy+i+dx), *(vz+i+dx) );
  if (m < n) {
    ddx = dx;
  }
  if ( ddx ) {
    *(vx+i) = *(vx+i+ddx) + ddx;
    *(vy+i) = *(vy+i+ddx);
    *(vz+i) = *(vz+i+ddx);
  }
  return( TRUE );
}

#if defined(_ANSI_)
static int E_V3DtraiteMpD( short int *vx, short int *vy , short int *vz, int x, int y, int z, int dim[4] )
#else
static int E_V3DtraiteMpD(vx, vy , vz, x, y, z, dim)
short int *vx;
short int *vy ;
short int *vz;
int x;
int y;
int z;
int dim[4] ;
#endif
{
  register int i, dx, ddx;
  int n,m;
  
  ddx = 0;
  i = z*dim[3] + y*dim[0] + x;
  n = E_V3DNorme2( *(vx+i), *(vy+i), *(vz+i) );
  
  dx = 1;
  m = E_V3DNorme2( (short int)(*(vx+i+dx)+dx), 
		  *(vy+i+dx), *(vz+i+dx) );
  if (m < n) {
    ddx = dx;
  }
  if ( ddx ) {
    *(vx+i) = *(vx+i+ddx) + ddx;
    *(vy+i) = *(vy+i+ddx);
    *(vz+i) = *(vz+i+ddx);
  }
  return( TRUE );
}

#if defined(_ANSI_)
static void _VD_Seuil( t_ImagePtr theIm, t_ImagePtr resIm, e_ParamDistancePtr par )
#else
static void _VD_Seuil( theIm, resIm, par )
t_ImagePtr theIm;
t_ImagePtr resIm;
e_ParamDistancePtr par ;
#endif
{
	register int is, i, x, y, z;
	int v;
	register float fs;
	short int *vl;

	fs = par->thres;
	v = theIm->tx * theIm->ty * theIm->tz;
	vl = (short int*)(resIm->buf);

	switch ( theIm->type ) {
	case TYPEIM_256 :
		{
		unsigned char *buf;
		buf = (unsigned char*)(theIm->buf);
		is = (int)(fs * 255.0 + 0.5);
		for (i=0; i<v; i++) 
			*vl++ = ( *buf++ < is ) ? SINFINI : 0;
	        }
		break;
	case TYPEIM_16B :
		{
		unsigned short int *buf;
		buf = (unsigned short int*)(theIm->buf);
		is = (int)(fs * 65535.0 + 0.5);
		for (i=0; i<v; i++) 
			*vl++ = ( *buf++ < is ) ? SINFINI : 0;
	        }
		break;
	case TYPEIM_S16B :
		{
		short int *buf;
		buf = (short int*)(theIm->buf);
		is = ( fs < 0 ) ? (int)(fs * 32767.0 - 0.5) : (int)(fs * 32767.0 + 0.5);
		for (i=0; i<v; i++) 
			*vl++ = ( *buf++ < is ) ? SINFINI : 0;
	        }
		break;
	case TYPEIM_FLOAT :
		{
		float *buf;
		buf = (float*)(theIm->buf);
		for (i=0; i<v; i++) 
			*vl++ = ( *buf++ < fs ) ? SINFINI : 0;
	        }
		break;
	default :
		for (z=0; z<theIm->tz; z++)
		for (y=0; y<theIm->ty; y++)
		for (x=0; x<theIm->tx; x++) 
			*vl++ = ( T_Get3Point(x, y, z, theIm) < fs ) ? SINFINI : 0;
	}
}

#if defined(_ANSI_)
static void _ID_Seuil( t_ImagePtr theIm, t_ImagePtr resIm, int *offset, e_ParamDistancePtr par )
#else
static void _ID_Seuil( theIm, resIm, offset, par )
t_ImagePtr theIm;
t_ImagePtr resIm;
int *offset;
e_ParamDistancePtr par ;
#endif
{
	register int is, i, x, y, z;
	int v, dx1, dy1, dz1;
	register float fs;
	unsigned short int *vl;

	fs = par->thres;
	v = theIm->tx * theIm->ty * theIm->tz;
	vl = (unsigned short int*)(resIm->buf);

	switch ( theIm->type ) {
	case TYPEIM_256 :
		{
		unsigned char *buf;
		buf = (unsigned char*)(theIm->buf);
		is = (int)(fs * 255.0 + 0.5);
		for (i=0; i<v; i++) 
			*vl++ = ( *buf++ < is ) ? UINFINI : 0;
	        }
		break;
	case TYPEIM_16B :
		{
		unsigned short int *buf;
		buf = (unsigned short int*)(theIm->buf);
		is = (int)(fs * 65535.0 + 0.5);
		for (i=0; i<v; i++) 
			*vl++ = ( *buf++ < is ) ? UINFINI : 0;
	        }
		break;
	case TYPEIM_S16B :
		{
		short int *buf;
		buf = (short int*)(theIm->buf);
		is = ( fs < 0 ) ? (int)(fs * 32767.0 - 0.5) : (int)(fs * 32767.0 + 0.5);
		for (i=0; i<v; i++) 
			*vl++ = ( *buf++ < is ) ? UINFINI : 0;
	        }
		break;
	case TYPEIM_FLOAT :
		{
		float *buf;
		buf = (float*)(theIm->buf);
		for (i=0; i<v; i++) 
			*vl++ = ( *buf++ < fs ) ? UINFINI : 0;
	        }
		break;
	default :
		for (z=0; z<theIm->tz; z++)
		for (y=0; y<theIm->ty; y++)
		for (x=0; x<theIm->tx; x++) 
			*vl++ = ( T_Get3Point(x, y, z, theIm) < fs ) ? UINFINI : 0;
	}

	if ( (par->choix != DIST_CHAMFER3_2D) && (par->choix != DIST_CHAMFER3) ) 
		return;

	dx1 = theIm->tx - 1;
	dy1 = theIm->ty - 1;
	dz1 = theIm->tz - 1;
	i = 0;
	for (z=0; z<theIm->tz; z++)
	for (y=0; y<theIm->ty; y++)
	for (x=0; x<theIm->tx; x++) {
		if ( *(vl + i) == 0 ) continue;
		/*--- 6-voisins ---*/
		if ( x > 0 ) {
			if ( *(vl + i + offset[12]) == 0 ) *(vl + i) = 16;
			continue;
		}
		if ( x < dx1 ) {
			if ( *(vl + i + offset[14]) == 0 ) *(vl + i) = 16;
			continue;
		}
		if ( y > 0 ) {
			if ( *(vl + i + offset[10]) == 0 ) *(vl + i) = 16;
			continue;
		}
		if ( y < dy1 ) {
			if ( *(vl + i + offset[16]) == 0 ) *(vl + i) = 16;
			continue;
		}
		if ( par->choix == DIST_CHAMFER3 ) {
			if ( z > 0 ) {
				if ( *(vl + i + offset[4]) == 0 ) *(vl + i) = 16;
				continue;
			}
			if ( z < dz1 ) {
				if ( *(vl + i + offset[22]) == 0 ) *(vl + i) = 16;
				continue;
			}
		}
		/*--- 18-voisins ---*/
		if ( (x > 0) && (y > 0) ) {
			if ( *(vl + i + offset[9]) == 0 ) *(vl + i) = 21;
			continue;
		}
		if ( (x > 0) && (y < dy1) ) {
			if ( *(vl + i + offset[15]) == 0 ) *(vl + i) = 21;
			continue;
		}
		if ( (x < dx1) && (y > 0) ) {
			if ( *(vl + i + offset[11]) == 0 ) *(vl + i) = 21;
			continue;
		}
		if ( (x < dx1) && (y < dy1) ) {
			if ( *(vl + i + offset[17]) == 0 ) *(vl + i) = 21;
			continue;
		}
		if ( par->choix == DIST_CHAMFER3 ) {
			if ( (x > 0) && (z > 0) ) {
				if ( *(vl + i + offset[3]) == 0 ) *(vl + i) = 21;
				continue;
			}
			if ( (x > 0) && (z < dz1) ) {
				if ( *(vl + i + offset[21]) == 0 ) *(vl + i) = 21;
				continue;
			}
			if ( (x < dx1) && (z > 0) ) {
				if ( *(vl + i + offset[5]) == 0 ) *(vl + i) = 21;
				continue;
			}
			if ( (x < dx1) && (z < dz1) ) {
				if ( *(vl + i + offset[23]) == 0 ) *(vl + i) = 21;
				continue;
			}
			if ( (z > 0) && (y > 0) ) {
				if ( *(vl + i + offset[1]) == 0 ) *(vl + i) = 21;
				continue;
			}
			if ( (z > 0) && (y < dy1) ) {
				if ( *(vl + i + offset[7]) == 0 ) *(vl + i) = 21;
				continue;
			}
			if ( (z < dz1) && (y > 0) ) {
				if ( *(vl + i + offset[19]) == 0 ) *(vl + i) = 21;
				continue;
			}
			if ( (z < dz1) && (y < dy1) ) {
				if ( *(vl + i + offset[25]) == 0 ) *(vl + i) = 21;
				continue;
			}
		}
		/*--- 26-voisins ---*/
		if ( par->choix == DIST_CHAMFER3_2D ) continue;
		if ( (x > 0) && (y > 0) && (z > 0) ) {
			if ( *(vl + i + offset[0]) == 0 ) *(vl + i) = 26;
			continue;
		}
		if ( (x < dx1) && (y > 0) && (z > 0) ) {
			if ( *(vl + i + offset[2]) == 0 ) *(vl + i) = 26;
			continue;
		}
		if ( (x > 0) && (y < dy1) && (z > 0) ) {
			if ( *(vl + i + offset[6]) == 0 ) *(vl + i) = 26;
			continue;
		}
		if ( (x < dx1) && (y < dy1) && (z > 0) ) {
			if ( *(vl + i + offset[8]) == 0 ) *(vl + i) = 26;
			continue;
		}
		if ( (x > 0) && (y > 0) && (z < dz1) ) {
			if ( *(vl + i + offset[18]) == 0 ) *(vl + i) = 26;
			continue;
		}
		if ( (x < dx1) && (y > 0) && (z < dz1) ) {
			if ( *(vl + i + offset[20]) == 0 ) *(vl + i) = 26;
			continue;
		}
		if ( (x > 0) && (y < dy1) && (z < dz1) ) {
			if ( *(vl + i + offset[24]) == 0 ) *(vl + i) = 26;
			continue;
		}
		if ( (x < dx1) && (y < dy1) && (z < dz1) ) {
			if ( *(vl + i + offset[26]) == 0 ) *(vl + i) = 26;
			continue;
		}
	}
}
