/* epi_extent.c 2.1 5/14/98 */

/*=============================================================================
 *
 * File           : $EPI/usr/people/haibin/epidaure/include/epi_extent.c
 * Module         : 
 * $Id: epi_extent.c,v 1.1 2004/01/26 16:00:27 xpennec Exp $
 * $Header: /proj/epidaure/home/private/CVS/mips/AuxPkgs/EpidaureLibPkg/epi_extent.c,v 1.1 2004/01/26 16:00:27 xpennec Exp $
 * Test File      : 
 * Author         : Haibin yIN
 * Creation       : Fri Oct 15 16:00:31 1993
 *
 *=============================================================================
 *     (c) Copyright 1992, ALEPH MED, all rights reserved
 *=============================================================================
 *
 *------------------------- Description ---------------------------------------
 *
 *   Functions defined for the volume voxel quick access
 *   
 *   Ces fonctions internes a` Epidaure doivent etre utilisee avec precautions, 
 *   car on suppose que les parametres d'entree sont corrects, donc aucun control
 *   des donnees.
 *------------------------- Modifications -------------------------------------
 *   
 * $Log: epi_extent.c,v $
 * Revision 1.1  2004/01/26 16:00:27  xpennec
 * Added Package EpidaureLibPkg in AuxPkgs
 * source of EpidaureLib taken from Roboscope (~ nov 99)
 *
 * Revision 3.0  1994/05/05  13:08:18  isabelle
 * version fixe'e a 3.0.0 avant demo GE
 *
 * Revision 2.6  94/03/30  11:34:03  haibin
 * ajout fonction EPI_set(get)_current_value
 * 
 * Revision 2.5  94/02/03  13:37:50  haibin
 * Epidaure New version 2.5 
 * 
 * Revision 2.1  94/02/03  13:32:43  haibin
 * new version Epidaure
 * 
 * Revision 2.0  93/11/08  09:48:31  haibin
 * version fixed to 2.0
 * 
 * Revision 1.5  93/10/21  15:13:41  haibin
 * introduire un tableau de i / 255.0 pour eviter la division
 * 
 * Revision 1.4  93/10/20  12:33:40  haibin
 * suppression du control des donnees
 * 
 * Revision 1.3  93/10/20  12:09:09  haibin
 * amelioration d'acces au volume
 * 
 * Revision 1.2  93/10/19  12:24:58  haibin
 * correction bugs et integration dans dilatation et eorsion les fonction d'acces
 * 
 * Revision 1.1  93/10/18  18:22:55  haibin
 * optimisation de l'acces aux donnees
 * 
 *   
 * 
 */


#include <epidaure.h>
#include <epidaure.ee>
#include <epidaure.ei>


/*-------Definition des fonctions statiques----------*/
#ifndef NO_PROTO
static int _EPI_compute_inv_255_array(void);
#else 
static int _EPI_compute_inv_255_array();
#endif
/*-------Fin des proto statiques----------*/


static reeltype INV_FLOAT_65535  = (reeltype) (1.0 / 65535.0);  
static reeltype INV_FLOAT_32767  = (reeltype) (1.0 / 32767.0);  
static reeltype INV_FLOAT_255_ARRAY[256];
static int inv_255_array_computed = 0;


/*---------------------------------------------------------------------------
 *      A c c e s s     F u n c t i o n s 
 *
 *  The access functions are used to change the read-write pointer
 *  the offset is computed from the last access pointer.
 *
 *---------------------------------------------------------------------------
 */


/*---------------------------------------------------------------------------
 * Compute 1.0 / i where i is in [0, 255] and stockes the result in 
 * a floatting array. This array is used in the functions EPI_inq_*
 *---------------------------------------------------------------------------
 */

#if defined(_ANSI_)
static int _EPI_compute_inv_255_array(void)
#else
static int _EPI_compute_inv_255_array()

#endif

{ int i;
  
  for (i = 0; i < 256; i ++)
    INV_FLOAT_255_ARRAY[i] = (reeltype) (i / 255.0);
  
  return((int)1); 
}


/*---------------------------------------------------------------------------
 * int EPI_init_access_point(EPI_Imaccess *access, t_ImagePtr ima)
 * 
 * Action :
 *   Initialize the acces structure for starting a image acces.
 * 
 * Include :
 *  <epi_extent.h> 
 *
 * Parameters in :
 *   t_ImagePtr ima : input image pointer ( <> null).   
 *
 * Parameters out :
 *   
 * Parameters in/out :
 *  EPI_Imaccess *access : acess structure for the input image.
 *
 * Return value : 
 *   int :   1 if OK;  0 if not.
 *
 * Description :
 *   Initialize the info for the access and the current acees adresss to
 *   the image buffer (according to the position (x, y, z) = (0, 0, 0)).
 *
 * Errors :
 *   
 *   
 *---------------------------------------------------------------------------
 */
#if defined(_ANSI_)
int EPI_init_access_point(EPI_Imaccess *access, t_ImagePtr ima)
#else
int EPI_init_access_point(access, ima)
EPI_Imaccess *access;
t_ImagePtr ima;

#endif
{ int result = 0;
  
  if ((ima != NULL) && (access != NULL)) {
    access -> ima = ima;
    access -> buff = ima->buf; 
    access -> type = ima->type; 
    
    switch (ima->type) {
    case TYPEIM_256:
      if (inv_255_array_computed == 0)
	inv_255_array_computed = _EPI_compute_inv_255_array();
      access -> voxel_size = sizeof(unsigned char); 
      break;
    case TYPEIM_16B: 
      access -> voxel_size = sizeof(unsigned short); 
      break;
    case TYPEIM_S16B: 
      access -> voxel_size = sizeof(short); 
      break;
    case TYPEIM_FLOAT: 
      access -> voxel_size = sizeof(reeltype); 
      break;
    } 
    access -> tx = ima -> tx;
    access -> ty = ima -> ty;
    access -> tz = ima -> tz;
    access -> ln_voxel_nb = ima -> tx; 
    access -> pln_voxel_nb = (ima -> tx) * (ima -> ty); 
    access -> ln_size = (access -> ln_voxel_nb) * (access -> voxel_size); 
    access -> pln_size = (access -> pln_voxel_nb) * (access -> voxel_size); 
    access -> curr_adr = access -> buff; 
    access -> curr_x = access -> curr_y = access -> curr_z = 0;        
    result = 1;
  }
  
  return(result);
}


/*---------------------------------------------------------------------------
 * 
 * int EPI_go_to_point(EPI_Imaccess *access, int x, int y, int z)
 * 
 * Action :
 *   set the current access pointer to the position according to (x, y, z).
 * 
 * Include :
 *   <epi_extent.h>
 *
 * Parameters in :
 *   int x, y, z :  position in the image 
 *                  (x, y, z in [0, tx - 1] x [0, ty -1] x [0, tz - 1]).
 *   
 * Parameters out :
 *   
 * Parameters in/out :
 *    EPI_Imaccess *access : acess structure for the input image.
 *
 * Return value : 
 *    int :  1 if OK;  0 if not.
 *
 * Description :
 *   The resulting current access pointer is
 *       buff + z * pln_size + y * ln_size + x * voxel_size.
 *
 *   But this adress is computed from the old current adress and its offset
 *   possible.
 *     i.e. curr_adr = curr_adr + (z - curr_z) * pln_size +
 *                     (y - curr_y) * ln_size + (x - curr_x) * voxel_size.
 *   This algorithm is optimized if 
 *     z - curr_z = 0 or (and) y - curr_y = 0 or (and) x - curr_x = 0.
 *
 * Errors :
 *   
 *---------------------------------------------------------------------------
 */
#if defined(_ANSI_)
int EPI_go_to_point(EPI_Imaccess *access, int x, int y, int z)
#else
int EPI_go_to_point(access, x, y, z)
EPI_Imaccess *access;
int x;
int y;
int z;

#endif
{ int result = 0;
  int offset_x, offset_y, offset_z;
  
  /*----------------------------------------------------------------------
   *  If the image is considered as toric (Inria)
   * 
   *     if (x < 0) 
   *       x += access -> tx;
   *     else
   *       if (x >= access -> tx)
   * 	        x  %= access -> tx;
   *     
   *     if (y < 0) 
   *       y += access -> ty;
   *     else
   *       if (y >= access -> ty)
   * 	       y %= access -> ty;
   *     
   *     if (z < 0) 
   *       z += access -> tz;
   *     else
   *       if (z >= access -> tz)
   *         z %= access -> tz;
   *---------------------------------------------------------------------
   */
  
  if ((x >= 0) && (x < access -> tx) &&
      (y >= 0) && (y < access -> ty) &&
      (z >= 0) && (z < access -> tz)) {
    
    if ((offset_x = x - access -> curr_x) != 0) {
      access -> curr_x = x;
      access -> curr_adr += (offset_x * (access -> voxel_size));
    }
    if ((offset_y = y - access -> curr_y) != 0) {
      access -> curr_y = y;
      access -> curr_adr += (offset_y * (access -> ln_size));
    }
    if ((offset_z = z - access -> curr_z) != 0) {
      access -> curr_z = z;
      access -> curr_adr += (offset_z * (access -> pln_size));
    }      
    result = 1;
  }
  
  return(result);
}


/*---------------------------------------------------------------------------
 * 
 * int EPI_go_to_next_x_point(EPI_Imaccess *access)
 * 
 * Action :
 *  set the current access adress to the next position along x.
 *
 * Include :
 *  <epi_extrent.h>
 *
 * Parameters in :
 *   
 * Parameters out :
 *   
 * Parameters in/out :
 *   EPI_Imaccess *access : Access structure (<> NULL).
 *
 * Return value : 
 *   int :  1 if OK;  0 if not.
 *
 * Description :
 *   curr_adr = curr_adr + voxel_size;   
 *   a) if (old curr_x < ln_voxel_nb - 1)
 *       new curr_x = old curr_x + 1.
 *   b) else 
 *       new curr_x = 0;
 *       if (old curr_y < tz - 1)
 *           new curr_y = old curr_y + 1;
 *       else 
 *           new curr_y = 0;
 *           new curr_z = old curr_z + 1;
 *
 * Errors :
 *   
 *---------------------------------------------------------------------------
 */
#if defined(_ANSI_)
int EPI_go_to_next_x_point(EPI_Imaccess *access)
#else
int EPI_go_to_next_x_point(access)
EPI_Imaccess *access;

#endif
{ int result = 0;
  
  access -> curr_adr += access -> voxel_size;
  
  if (access -> curr_x < access -> tx - 1) 
    /*------ 
      x = x + 1
      ------*/
    (access -> curr_x) ++;
  else {
    if (access -> curr_y < access -> ty - 1) {
      /*------- 
	y = y + 1
	-------*/
      access -> curr_x = 0;
      (access -> curr_y) ++;
    }
    else {
      /*-------
	z = z + 1
	-------*/
      access -> curr_x = 0;
      access -> curr_y = 0;
      (access -> curr_z) ++;
    }
  }
  result = 1;
  
  return(result);
}


/*---------------------------------------------------------------------------
 * 
 * int EPI_go_to_prev_x_point(EPI_Imaccess *access)
 * 
 * Action :
 *  set the current access adress to the previous position along x.
 *
 * Include :
 *  <epi_extrent.h>
 *
 * Parameters in :
 *
 * Parameters out :
 *   
 * Parameters in/out :
 *   EPI_Imaccess *access : Access structure (<> NULL).
 *
 * Return value : 
 *   int :  1 if OK;  0 if not.
 *   
 * Description :
 *   curr_adr = curr_adr - voxel_size;   
 *   but  
 *   a) if (old curr_x > 0)
 *       new curr_x = old curr_x - 1.
 *   b) else 
 *       new curr_x = tx - 1;
 *       if (old curr_y == 0)
 *           new curr_y = ty - 1;
 *           new curr_z = old curr_z - 1
 *       else 
 *           new curr_y = old_curr_y - 1.
 *
 * Errors :
 *   
 *---------------------------------------------------------------------------
 */
#if defined(_ANSI_)
int EPI_go_to_prev_x_point(EPI_Imaccess *access)
#else
int EPI_go_to_prev_x_point(access)
EPI_Imaccess *access;

#endif
{ int result = 0;

  access -> curr_adr -= access -> voxel_size;
  
  if (access -> curr_x > 0)
    /*------ 
      x = x - 1
      ------*/
    (access -> curr_x) --;
  else {
    if (access -> curr_y > 0) {
      /*------- 
	y = y - 1
	-------*/
      access -> curr_x = access -> tx - 1;
      (access -> curr_y) --;
    }
    else {
      /*-------
	z = z - 1
	-------*/
      access -> curr_x = access -> tx - 1;
      access -> curr_y = access -> ty - 1;
      (access -> curr_z) --;
    }
  }
  result = 1;
  return(result);
}


/*---------------------------------------------------------------------------
 * 
 * int EPI_go_to_next_y_point(EPI_Imaccess *access)
 * 
 * Action :
 *  set the current access adress to the next position along y.
 *
 * Include :
 *  <epi_extrent.h>
 *
 * Parameters in :
 *   
 * Parameters out :
 *   
 * Parameters in/out :
 *   EPI_Imaccess *access : Access structure (<> NULL).
 *
 * Return value : 
 *   int :  1 if OK;  0 if not.
 *   
 * Description :
 *   curr_adr = curr_adr + ln_size;
 *   but  
 *       if (old curr_y < ty- 1)
 *           new curr_y = old curr_y + 1;
 *       else 
 *           new curr_y = 0;
 *           new curr_z = old curr_z + 1;
 *
 * Errors :
 *   
 *---------------------------------------------------------------------------
 */
#if defined(_ANSI_)
int EPI_go_to_next_y_point(EPI_Imaccess *access)
#else
int EPI_go_to_next_y_point(access)
EPI_Imaccess *access;

#endif
{ int result = 0;
  
  access -> curr_adr += access -> ln_size;
  
  if (access -> curr_y < access -> ty -1) 
    /*-----
      y = y + 1
      ------*/
    (access -> curr_y) ++;
  else {
    /*-----
      z = z + 1
      -------*/
    access -> curr_y = 0;
    (access -> curr_z) ++;
  }
  
  result = 1;
  return(result);
}

/*---------------------------------------------------------------------------
 * 
 * int EPI_go_to_prev_y_point(EPI_Imaccess *access)
 * 
 * Action :
 *  set the current access adress to the previous position along y.
 *
 * Include :
 *  <epi_extrent.h>
 *
 * Parameters in :
 *   
 * Parameters out :
 *   
 * Parameters in/out :
 *   EPI_Imaccess *access : Access structure (<> NULL).
 *
 * Return value : 
 *   int :  1 if OK;  0 if not.
 *   
 * Description :
 *   curr_adr = curr_adr - ln_size;
 *   but  
 *       if (old curr_y == 0)
 *           new curr_y = ty - 1;
 *           new curr_z = old curr_z - 1
 *       else 
 *           new curr_y = old_curr_y - 1.
 *
 * Errors :
 *   
 *---------------------------------------------------------------------------
 */
#if defined(_ANSI_)
int EPI_go_to_prev_y_point(EPI_Imaccess *access)
#else
int EPI_go_to_prev_y_point(access)
EPI_Imaccess *access;

#endif
{ int result = 0;
  
  access -> curr_adr -= access -> ln_size;
  
  if (access -> curr_y > 0)
    /*-----
      y = y - 1
      ------*/
    (access -> curr_y) --;
  else {
    /*-----
      z = z - 1
      -------*/
    access -> curr_y = access -> ty - 1;
    (access -> curr_z) --;
  }
  
  result = 1;
  return(result);
}


/*---------------------------------------------------------------------------
 * 
 * int EPI_go_to_next_z_point(EPI_Imaccess *access)
 * 
 * Action :
 *  set the current access adress to the next position along z.
 *
 * Include :
 *  <epi_extrent.h>
 *
 * Parameters in :
 *   
 * Parameters out :
 *   
 * Parameters in/out :
 *   EPI_Imaccess *access : Access structure (<> NULL).
 *
 * Return value : 
 *   int :  1 if OK;  0 if not.
 *   
 * Description :
 *   curr_adr = curr_adr + pln_size.
 *
 *   Remark:  The new acces adress is valid only if (old curr_z < tz - 1),
 *            User must verify this condition.
 *
 * Errors :
 *   
 *---------------------------------------------------------------------------
 */

#if defined(_ANSI_)
int EPI_go_to_next_z_point(EPI_Imaccess *access)
#else
int EPI_go_to_next_z_point(access)
EPI_Imaccess *access;

#endif
{ int result = 0;
  
  /*-----
    z = z + 1
    ------*/
  access -> curr_adr += access -> pln_size;
  (access -> curr_z) ++;
  result = 1;
  return(result);
}


/*---------------------------------------------------------------------------
 * 
 * int EPI_go_to_prev_z_point(EPI_Imaccess *access)
 * 
 * Action :
 *  set the current access adress to the previous position along z.
 *
 * Include :
 *  <epi_extrent.h>
 *
 * Parameters in :
 *   
 * Parameters out :
 *   
 * Parameters in/out :
 *   EPI_Imaccess *access : Access structure (<> NULL).
 *
 * Return value : 
 *   int :  1 if OK;  0 if not.
 *   
 * Description :
 *   curr_adr = curr_adr - pln_size.
 *
 *   Remark:  The new acces adress is valid only if (old curr_z > 0).
 *            User must verify this condition.
 *
 * Errors :
 *   
 *---------------------------------------------------------------------------
 */
#if defined(_ANSI_)
int EPI_go_to_prev_z_point(EPI_Imaccess *access)
#else
int EPI_go_to_prev_z_point(access)
EPI_Imaccess *access;

#endif
{ int result = 0;
  
  /*-----
    z = z - 1
    ------*/
  access -> curr_adr -= access -> pln_size;
  (access -> curr_z) --;
  result = 1;
  return(result);
}



/*---------------------------------------------------------------------------
 *   R e a d i  n g    F u n c t i o n  s 
 *---------------------------------------------------------------------------
 */

/*---------------------------------------------------------------------------
 * 
 * int EPI_inq_current_value(EPI_Imaccess *access, reeltype *real_val) 
 * 
 * Action :
 *   Get the value not normalized accoording to the current access pointer.
 * 
 * Include :
 *   <epi_extent.h>
 *
 * Parameters in :
 *    EPI_Imaccess *access : acess structure for the input image.
 *   
 * Parameters out :
 *    reeltype *real_val : current point value (not normalized).
 *
 * Parameters in/out :
 *
 * Return value : 
 *   int :  1 if OK; 0 if not.
 *
 * Description :
 *
 * Errors :
 *   
 *---------------------------------------------------------------------------
 */
#if defined(_ANSI_)
int EPI_inq_current_value(EPI_Imaccess *access, reeltype *real_val)
#else
int EPI_inq_current_value(access, real_val)
EPI_Imaccess *access;
reeltype *real_val;

#endif
{ int result = 0;
  unsigned char *ubyte_ptr;
  unsigned short *ushort_ptr;
  short *short_ptr;
  reeltype *float_ptr;
  
  switch(access -> type) {
  case TYPEIM_256:
    ubyte_ptr = (unsigned char *)(access -> curr_adr);
    *real_val = (reeltype)(*ubyte_ptr);
    break;
  case TYPEIM_16B:
    ushort_ptr = (unsigned short *)(access -> curr_adr);
    *real_val = (reeltype)(*ushort_ptr);
    break;
  case TYPEIM_S16B:
    short_ptr = (short *)(access -> curr_adr);
    *real_val = (reeltype)(*short_ptr);
    break;
  case TYPEIM_FLOAT:
    float_ptr = (reeltype *)(access -> curr_adr);
    *real_val = *float_ptr;
    break;
  default:
    *real_val = (reeltype)T_GetIPoint(access -> curr_x,
				      access -> curr_y,
				      access -> curr_z,
				      access -> ima);
    break;
  }
  
  result = 1;
  return(result);
}



/*---------------------------------------------------------------------------
 * 
 * int EPI_inq_current_point(EPI_Imaccess *access, reeltype *real_val) 
 * 
 * Action :
 *   Get the value accoording to the current access pointer.
 * 
 * Include :
 *   <epi_extent.h>
 *
 * Parameters in :
 *    EPI_Imaccess *access : acess structure for the input image.
 *   
 * Parameters out :
 *    reeltype *real_val : current point value (normalized in [0, ..1.0]
 *                    if the image type is TYPEIM_256 or TYPEIM_16B)
 *
 * Parameters in/out :
 *
 * Return value : 
 *   int :  1 if OK; 0 if not.
 *
 * Description :
 *
 * Errors :
 *   
 *---------------------------------------------------------------------------
 */
#if defined(_ANSI_)
int EPI_inq_current_point(EPI_Imaccess *access, reeltype *real_val)
#else
int EPI_inq_current_point(access, real_val)
EPI_Imaccess *access;
reeltype *real_val;

#endif
{ int result = 0;
  unsigned char *ubyte_ptr;
  unsigned short *ushort_ptr;
  short *short_ptr;
  reeltype *float_ptr;
  
  switch(access -> type) {
  case TYPEIM_256:
    ubyte_ptr = (unsigned char *)(access -> curr_adr);
    *real_val = INV_FLOAT_255_ARRAY[(int)*ubyte_ptr];
    break;
  case TYPEIM_16B:
    ushort_ptr = (unsigned short *)(access -> curr_adr);
    *real_val = (reeltype)(*ushort_ptr) * INV_FLOAT_65535;      
    break;
  case TYPEIM_S16B:
    short_ptr = (short *)(access -> curr_adr);
    *real_val = (reeltype)(*short_ptr) * INV_FLOAT_32767;
    break;
  case TYPEIM_FLOAT:
    float_ptr = (reeltype *)(access -> curr_adr);
    *real_val = *float_ptr;
    break;
  default:
    *real_val = T_Get3Point(access -> curr_x,
			    access -> curr_y,
			    access -> curr_z,
			    access -> ima);
    break;
  }
  
  result = 1;
  
  return(result);
}


/*---------------------------------------------------------------------------
 * 
 * int EPI_inq_next_x_point(EPI_Imaccess *access, reeltype *real_val)
 * 
 * Action :
 *  Get the value accoording to the next position along x of the current access
 *  adress.
 *
 * Include :
 *  <epi_extrent.h>
 *
 * Parameters in :
 *    EPI_Imaccess *access : acess structure for the input image.
 *   
 * Parameters out :
 *    reeltype *real_val : output value (normalized in [0, ..1.0]
 *                    if the image type is TYPEIM_256 or TYPEIM_16B)
 *
 * Parameters in/out :
 *   
 * Return value : 
 *   int :  1 if OK;  0 if not.
 *
 * Description :
 *    real_val = (reetype)*(curr_adr + voxel_size)/ CTE.
 *    Remark :
 *    This value accords to the position (x, y, z), where
 *    if (curr_x < tx - 1),  (x, y, z) = (curr_x + 1, curr_y, curr_z) 
 *    else
 *      if (cuur_y < ty -1), (x, y, z) = (0, curr_y + 1, curr_z); 
 *      else 
 *         (x, y, z) = (0, 0, curr_z + 1).
 *
 * Errors :
 *   
 *---------------------------------------------------------------------------
 */
#if defined(_ANSI_)
int EPI_inq_next_x_point(EPI_Imaccess *access, reeltype *real_val)
#else
int EPI_inq_next_x_point(access, real_val)
EPI_Imaccess *access;
reeltype *real_val;

#endif
{ int result = 0;
  unsigned char *ubyte_ptr;
  unsigned short *ushort_ptr;
  short *short_ptr;
  reeltype *float_ptr;
  int x, y, z;

  switch(access -> type) {
  case TYPEIM_256:
    ubyte_ptr = (unsigned char *)(access -> curr_adr);
    ubyte_ptr ++;
    *real_val = INV_FLOAT_255_ARRAY[(int)*ubyte_ptr];
    break;
  case TYPEIM_16B:
    ushort_ptr = (unsigned short *)(access -> curr_adr);
    ushort_ptr ++;
    *real_val = (reeltype)(*ushort_ptr) * INV_FLOAT_65535;      
    break;
  case TYPEIM_S16B:
    short_ptr = (short *)(access -> curr_adr);
    short_ptr ++;
    *real_val = (reeltype)(*short_ptr) * INV_FLOAT_32767;
    break;
  case TYPEIM_FLOAT:
    float_ptr = (reeltype *)(access -> curr_adr);
    float_ptr ++;
    *real_val = *float_ptr;
    break;
  default:
    if (access -> curr_x < access -> tx - 1) {
      /*------ 
	x = x + 1
	------*/
      x = access -> curr_x + 1;
      y = access -> curr_y;
      z = access -> curr_z;
    }
    else {
      if (access -> curr_y < access -> ty - 1) {
	/*------- 
	  y = y + 1
	  -------*/
	x = 0;
	y = access -> curr_y + 1;
	z = access -> curr_z;
      }
      else {
	/*-------
	  z = z + 1
	  -------*/
	x = 0;
	y = 0;
	z = access -> curr_z + 1;
      }
    }
      
    *real_val = T_Get3Point(x, y, z, access -> ima);
    
    break;
  }
  
  result = 1;
  return(result);
}


/*---------------------------------------------------------------------------
 * 
 * int EPI_inq_prev_x_point(EPI_Imaccess *access, reeltype *real_val)
 * 
 * Action :
 *  Get the value accoording to the previous position along x of the current access
 *  adress.
 *
 * Include :
 *  <epi_extrent.h>
 *
 * Parameters in :
 *   EPI_Imaccess *access : acess structure for the input image.
 *   
 * Parameters out :
 *    reeltype *real_val : output value (normalized in [0, ..1.0]
 *                    if the image type is TYPEIM_256 or TYPEIM_16B)
 *   
 * Parameters in/out :
 *   
 * Return value : 
 *    int :  1 if OK;  0 if not.
 *
 * Description :
 *    real_val = (reetype)*(curr_adr - voxel_size)/ CTE.
 *    Remark :
 *    This value accords to the position (x, y, z), where
 *    if (curr_x > 0),  (x, y, z) = (curr_x - 1, curr_y, curr_z) 
 *    else
 *      if (cuur_y > 0), (x, y, z) = (tx - 1, curr_y - 1, curr_z); 
 *      else 
 *         (x, y, z) = (tx - 1, ty -1, curr_z - 1).
 *
 * Errors :
 *   
 *---------------------------------------------------------------------------
 */
#if defined(_ANSI_)
int EPI_inq_prev_x_point(EPI_Imaccess *access, reeltype *real_val)
#else
int EPI_inq_prev_x_point(access, real_val)
EPI_Imaccess *access;
reeltype *real_val;

#endif
{ int result = 0;
  unsigned char *ubyte_ptr;
  unsigned short *ushort_ptr;
  short *short_ptr;
  reeltype *float_ptr;
  int x, y, z;

  switch(access -> type) {
  case TYPEIM_256:
    ubyte_ptr = (unsigned char *)(access -> curr_adr);
    ubyte_ptr --;
    *real_val = INV_FLOAT_255_ARRAY[(int)*ubyte_ptr];
    break;
  case TYPEIM_16B:
    ushort_ptr = (unsigned short *)(access -> curr_adr);
    ushort_ptr --;
    *real_val = (reeltype)(*ushort_ptr) * INV_FLOAT_65535;      
    break;
  case TYPEIM_S16B:
    short_ptr = (short *)(access -> curr_adr);
    short_ptr --;
    *real_val = (reeltype)(*short_ptr) * INV_FLOAT_32767;
    break;
  case TYPEIM_FLOAT:
    float_ptr = (reeltype *)(access -> curr_adr);
    float_ptr --;
    *real_val = *float_ptr;
    break;
  default:
    if (access -> curr_x > 0) {
      /*------ 
	x = x - 1
	------*/
      x = access -> curr_x - 1;
      y = access -> curr_y;
      z = access -> curr_z;
    }
    else {
      if (access -> curr_y > 0) {
	/*------- 
	  y = y - 1
	  -------*/
	x = access -> tx - 1;
	y = access -> curr_y - 1;
	z = access -> curr_z;
      }
      else {
	/*-------
	  z = z - 1
	  -------*/
	x = access -> tx - 1;
	y = access -> ty - 1;
	z = access -> curr_z - 1;
      }
    }
    
    *real_val = T_Get3Point(x, y, z, access -> ima);
    
    break;
  }
  
  result = 1;
  return(result);
}



/*---------------------------------------------------------------------------
 * 
 * int EPI_inq_next_y_point(EPI_Imaccess *access, reeltype *real_val)
 * 
 * Action :
 *  Get the value accoording to the next position along y of the current access
 *  adress.
 *
 * Include :
 *  <epi_extrent.h>
 *
 * Parameters in :
 *   EPI_Imaccess *access : acess structure for the input image.
 *   
 * Parameters out :
 *    reeltype *real_val : output value (normalized in [0, ..1.0]
 *                    if the image type is TYPEIM_256 or TYPEIM_16B)
 *
 * Parameters in/out :
 *   
 * Return value : 
 *   int :  1 if OK;  0 if not.
 *
 * Description :
 *    real_val = (reetype)*(curr_adr + ln_size)/ CTE.
 *    Remark :
 *    This value accords to the position (x, y, z), where
 *    if (curr_y < ty -1), (x, y, z) = (curr_x, curr_y + 1, curr_z); 
 *    else 
 *         (x, y, z) = (curr_x, 0, curr_z + 1).
 *
 * Errors :
 *   
 *---------------------------------------------------------------------------
 */
#if defined(_ANSI_)
int EPI_inq_next_y_point(EPI_Imaccess *access, reeltype *real_val)
#else
int EPI_inq_next_y_point(access, real_val)
EPI_Imaccess *access;
reeltype *real_val;

#endif
{ int result = 0;
  unsigned char *ubyte_ptr;
  unsigned short *ushort_ptr;
  short *short_ptr;
  reeltype *float_ptr;
  int x, y, z;

  switch(access -> type) {
  case TYPEIM_256:
    ubyte_ptr = (unsigned char *)(access -> curr_adr);
    ubyte_ptr += (access -> ln_voxel_nb);
    *real_val = INV_FLOAT_255_ARRAY[(int)*ubyte_ptr];
    break;
  case TYPEIM_16B:
    ushort_ptr = (unsigned short *)(access -> curr_adr);
    ushort_ptr += (access -> ln_voxel_nb);
    *real_val = (reeltype)(*ushort_ptr) * INV_FLOAT_65535;      
    break;
  case TYPEIM_S16B:
    short_ptr = (short *)(access -> curr_adr);
    short_ptr += (access -> ln_voxel_nb);
    *real_val = (reeltype)(*short_ptr) * INV_FLOAT_32767;
    break;
  case TYPEIM_FLOAT:
    float_ptr = (reeltype *)(access -> curr_adr);
    float_ptr += (access -> ln_voxel_nb);
    *real_val = *float_ptr;
    break;
  default:
    x = access -> curr_x;
    y = access -> curr_y  + 1;
    z = access -> curr_z;
    if (access -> curr_y == access -> ty -1)  {
      /*-----
	z = z + 1
	-------*/
      y = 0;
      z ++;
    }
    
    *real_val = T_Get3Point(x, y, z, access -> ima);
    
    break;
  }
  
  result = 1;
  return(result);
}


/*---------------------------------------------------------------------------
 * 
 * int EPI_inq_prev_y_point(EPI_Imaccess *access, reeltype *real_val)
 * 
 * Action :
 *  Get the value accoording to the previous position along y of the current access
 *  adress.
 *
 * Include :
 *  <epi_extrent.h>
 *
 * Parameters in :
 *   EPI_Imaccess *access : acess structure for the input image.
 *   
 * Parameters out :
 *   reeltype *real_val : output value (normalized in [0, ..1.0]
 *                    if the image type is TYPEIM_256 or TYPEIM_16B)
 *   
 * Parameters in/out :
 *   
 * Return value : 
 *   int :  1 if OK;  0 if not.
 *
 * Description :
 *    real_val = (reetype)*(curr_adr - ln_size)/ CTE.
 *    Remark :
 *    This value accords to the position (x, y, z), where
 *    if (curr_y > 0), (x, y, z) = (curr_x, curr_y - 1, curr_z); 
 *    else 
 *      (x, y, z) = (curr_x, ty -1, curr_z - 1).
 *
 * Errors :
 *   
 *---------------------------------------------------------------------------
 */
#if defined(_ANSI_)
int EPI_inq_prev_y_point(EPI_Imaccess *access, reeltype *real_val)
#else
int EPI_inq_prev_y_point(access, real_val)
EPI_Imaccess *access;
reeltype *real_val;

#endif
{ int result = 0;
  unsigned char *ubyte_ptr;
  unsigned short *ushort_ptr;
  short *short_ptr;
  reeltype *float_ptr;
  int x, y, z;

  switch(access -> type) {
  case TYPEIM_256:
    ubyte_ptr = (unsigned char *)(access -> curr_adr);
    ubyte_ptr -= (access -> ln_voxel_nb);
    *real_val = INV_FLOAT_255_ARRAY[(int)*ubyte_ptr];
    break;
  case TYPEIM_16B:
    ushort_ptr = (unsigned short *)(access -> curr_adr);
    ushort_ptr -= (access -> ln_voxel_nb);
    *real_val = (reeltype)(*ushort_ptr) * INV_FLOAT_65535;      
    break;
  case TYPEIM_S16B:
    short_ptr = (short *)(access -> curr_adr);
    short_ptr -= (access -> ln_voxel_nb);
    *real_val = (reeltype)(*short_ptr) * INV_FLOAT_32767;
    break;
  case TYPEIM_FLOAT:
    float_ptr = (reeltype *)(access -> curr_adr);
    float_ptr -= (access -> ln_voxel_nb);
    *real_val = *float_ptr;
    break;
  default:
    x = access -> curr_x;
    y = access -> curr_y - 1;
    z = access -> curr_z;
    
    if (access -> curr_y == 0) {
      /*-----
	z = z - 1
	-------*/
      y = access -> ty - 1;
      z --;
    }
    
    *real_val = T_Get3Point(x, y, z, access -> ima);
    
    break;
  }
    
  result = 1;
  return(result);
}



/*---------------------------------------------------------------------------
 * 
 * int EPI_inq_next_z_point(EPI_Imaccess *access, reeltype *real_val)
 * 
 * Action :
 *  Get the value accoording to the next position along z of the current access
 *  adress.
 *
 * Include :
 *  <epi_extrent.h>
 *
 * Parameters in :
 *   EPI_Imaccess *access : acess structure for the input image.
 *   
 * Parameters out :
 *    reeltype *real_val : output value (normalized in [0, ..1.0]
 *                    if the image type is TYPEIM_256 or TYPEIM_16B)
 *
 * Parameters in/out :
 *   
 * Return value : 
 *   int :  1 if OK;  0 if not.
 *
 * Description :
 *    real_val = (reetype)*(curr_adr + pln_size)/ CTE.
 *    Remark :
 *    This value accords to the position (x, y, z), where
 *    (x, y, z) = (curr_x, curr_y, curr_z + 1).
 *    User must verify the condition (curr_z < tz - 1).
 * 
 * Errors :
 *   
 *---------------------------------------------------------------------------
 */
#if defined(_ANSI_)
int EPI_inq_next_z_point(EPI_Imaccess *access, reeltype *real_val)
#else
int EPI_inq_next_z_point(access, real_val)
EPI_Imaccess *access;
reeltype *real_val;

#endif
{ int result = 0;
  unsigned char *ubyte_ptr;
  unsigned short *ushort_ptr;
  short *short_ptr;
  reeltype *float_ptr;
  
  switch(access -> type) {
  case TYPEIM_256:
    ubyte_ptr = (unsigned char *)(access -> curr_adr);
    ubyte_ptr += (access -> pln_voxel_nb);
    *real_val = INV_FLOAT_255_ARRAY[(int)*ubyte_ptr];
      break;
  case TYPEIM_16B:
    ushort_ptr = (unsigned short *)(access -> curr_adr);
    ushort_ptr += (access -> pln_voxel_nb);
    *real_val = (reeltype)(*ushort_ptr) * INV_FLOAT_65535;      
    break;
  case TYPEIM_S16B:
    short_ptr = (short *)(access -> curr_adr);
    short_ptr += (access -> pln_voxel_nb);
    *real_val = (reeltype)(*short_ptr) * INV_FLOAT_32767;
    break;
  case TYPEIM_FLOAT:
    float_ptr = (reeltype *)(access -> curr_adr);
    float_ptr += (access -> pln_voxel_nb);
    *real_val = *float_ptr;
    break;
  default:
    *real_val = T_Get3Point(access -> curr_x,
			    access -> curr_y,
			    access -> curr_z + 1,
			    access -> ima);
  }
  
  result = 1;
  return(result);
}


/*---------------------------------------------------------------------------
 * 
 * int EPI_inq_prev_z_point(EPI_Imaccess *access, reeltype *real_val)
 * 
 * Action :
 *  Get the value accoording to the previous position along z of the current access
 *  adress.
 *
 * Include :
 *  <epi_extrent.h>
 *
 * Parameters in :
 *   EPI_Imaccess *access : acess structure for the input image.
 *   
 * Parameters out :
 *    reeltype *real_val : output value (normalized in [0, ..1.0]
 *                    if the image type is TYPEIM_256 or TYPEIM_16B)
 *
 * Parameters in/out :
 *   
 * Return value : 
 *   int :  1 if OK;  0 if not.
 *
 * Description :
 *    real_val = (reetype)*(curr_adr - pln_size)/ CTE.
 *    Remark :
 *    This value accords to the position (x, y, z), where
 *    (x, y, z) = (curr_x, curr_y, curr_z - 1).
 *    User must verify the condition (curr_z > 0).
 * 
 * Errors :
 *   
 *---------------------------------------------------------------------------
 */
#if defined(_ANSI_)
int EPI_inq_prev_z_point(EPI_Imaccess *access, reeltype *real_val)
#else
int EPI_inq_prev_z_point(access, real_val)
EPI_Imaccess *access;
reeltype *real_val;

#endif

{ int result = 0;
  unsigned char *ubyte_ptr;
  unsigned short *ushort_ptr;
  short *short_ptr;
  reeltype *float_ptr;
  
  switch(access -> type) {
  case TYPEIM_256:
    ubyte_ptr = (unsigned char *)(access -> curr_adr);
    ubyte_ptr -= (access -> pln_voxel_nb);
    *real_val = INV_FLOAT_255_ARRAY[(int)*ubyte_ptr];
    break;
  case TYPEIM_16B:
    ushort_ptr = (unsigned short *)(access -> curr_adr);
    ushort_ptr -= (access -> pln_voxel_nb);
    *real_val = (reeltype)(*ushort_ptr) * INV_FLOAT_65535;      
    break;
  case TYPEIM_S16B:
    short_ptr = (short *)(access -> curr_adr);
    short_ptr -= (access -> pln_voxel_nb);
    *real_val = (reeltype)(*short_ptr) * INV_FLOAT_32767;
    break;
  case TYPEIM_FLOAT:
    float_ptr = (reeltype *)(access -> curr_adr);
    float_ptr -= (access -> pln_voxel_nb);
    *real_val = *float_ptr;
    break;
  default:
    *real_val = T_Get3Point(access -> curr_x,
			    access -> curr_y,
			    access -> curr_z - 1,
			    access -> ima);
    break;
  }
  
  result = 1;
  return(result);
}


/*---------------------------------------------------------------------------
 * 
 * int EPI_inq_current_offset_point(EPI_Imaccess *access, 
 *                           int offset_x, 
 *                           int offset_y, 
 *                           int offset_z,
 *                           reeltype *real_val)
 * 
 * Action :
 *  Get the position value according to an offset along x, y and z from the 
 *  current access adress.
 *
 * Include :
 *  <epi_extrent.h>
 *
 * Parameters in :
 *   EPI_Imaccess *access : acess structure for the input image.
 *    int offset_x, offset_y, offset_z : offset along the x, y and z axis
 *                                       to the current position.
 * Parameters out :
 *    reeltype *real_val : output value (normalized in [0, ..1.0]
 *                    if the image type is TYPEIM_256 or TYPEIM_16B)
 *
 * Parameters in/out :
 *   
 * Return value : 
 *   int :  1 if OK;  0 if not.
 *
 * Description :
 *   The output value accords to the position 
 *     (curr_x + offset_x, curr_y + offset_y, curr_z + offset_z).
 *
 *   But this adress is computed from the old current adress and its offset
 *   possible.
 *     i.e. curr_adr = curr_adr + offset_z * pln_size +
 *                     offset_y * (access -> ln_size) + offset_x * voxel_size.
 *   This algorithm is optimized if 
 *     offset_z = 0 or (and) offset_y = 0 or (and) offset_x = 0.
 *
 * Errors :
 *   
 *---------------------------------------------------------------------------
 */
#if defined(_ANSI_)
int EPI_inq_current_offset_point(EPI_Imaccess *access,
				 int offset_x,
				 int offset_y,
				 int offset_z,
				 reeltype *real_val)
#else
int EPI_inq_current_offset_point(access, 
				 offset_x, 
				 offset_y, 
				 offset_z,
				 real_val)
EPI_Imaccess *access;
int offset_x;
int offset_y;
int offset_z;
reeltype *real_val;

#endif
{ 
  int result = 0;
  unsigned char *ubyte_ptr;
  unsigned short *ushort_ptr;
  short *short_ptr;
  reeltype *float_ptr;
  char *tmp_ptr;

  tmp_ptr = (char *)(access -> curr_adr);
  
  if (offset_z != 0) 
    tmp_ptr += offset_z * (access -> pln_size);
  
  if (offset_y != 0) 
    tmp_ptr += offset_y * (access -> ln_size);
  
  if (offset_x != 0) 
    tmp_ptr += offset_x * (access -> voxel_size);
  
  switch(access -> type) {
  case TYPEIM_256:
    ubyte_ptr = (unsigned char *)tmp_ptr;
    *real_val = INV_FLOAT_255_ARRAY[(int)*ubyte_ptr];
    break;
  case TYPEIM_16B:
    ushort_ptr = (unsigned short *)tmp_ptr;
    *real_val = (reeltype)(*ushort_ptr) * INV_FLOAT_65535;      
    break;
  case TYPEIM_S16B:
    short_ptr = (short *)tmp_ptr;
    *real_val = (reeltype)(*short_ptr) * INV_FLOAT_32767;
    break;
  case TYPEIM_FLOAT:
    float_ptr = (reeltype *)tmp_ptr;
    *real_val = *float_ptr;
    break;
  default:
    *real_val = T_Get3Point(access -> curr_x + offset_x,
			    access -> curr_y + offset_y,
			    access -> curr_z + offset_z,
			    access -> ima);
    break;
  }
  
  result = 1;
  return(result);
}


/*---------------------------------------------------------------------------
 *    W r i t i n g    F u n c t i o n s 
 *---------------------------------------------------------------------------
 */


/*---------------------------------------------------------------------------
 * 
 * int EPI_set_current_value(EPI_Imaccess *access, reeltype real_val) 
 * 
 * Action :
 *   set the value not nrmalized accoording to the current access pointer.
 * 
 * Include :
 *   <epi_extent.h>
 *
 * Parameters in :
 *   EPI_Imaccess *access : acess structure for the input image.
 *   reeltype real_val : current point value (not normalized in [0, ..1.0]).
 *   
 * Parameters out :
 *
 * Parameters in/out :
 *
 * Return value : 
 *   int :  1 if OK;  0 if not.
 *
 * Description :
 *
 * Errors :
 *   
 *---------------------------------------------------------------------------
 */
#if defined(_ANSI_)
int EPI_set_current_value(EPI_Imaccess *access, reeltype real_val)
#else
int EPI_set_current_value(access, real_val)
EPI_Imaccess *access;
reeltype real_val;

#endif
{ 
  int result = 0;
  unsigned char *ubyte_ptr;
  unsigned short *ushort_ptr;
  short *short_ptr;
  reeltype *float_ptr;
  
  switch(access -> type) {
  case TYPEIM_256:
    ubyte_ptr = (unsigned char *)(access -> curr_adr);
    *ubyte_ptr = (int)real_val;
    break;
  case TYPEIM_16B:
    ushort_ptr = (unsigned short *)(access -> curr_adr);
    *ushort_ptr = (int)real_val;
    break;
  case TYPEIM_S16B:
    short_ptr = (short *)(access -> curr_adr);
    *short_ptr = (int)real_val;
    break;
  case TYPEIM_FLOAT:
    float_ptr = (reeltype *)(access -> curr_adr);
    *float_ptr = real_val;
    break;
  default:
    T_Set3Point(access -> curr_x,
		access -> curr_y,
		access -> curr_z,
		access -> ima,
		real_val, MODE_SET);
    /* T_SetIPoint changed in T_Set3Point by GM 12/09/94 */
    break;
  }
  
  result = 1;
  return(result);
}


/*---------------------------------------------------------------------------
 * 
 * int EPI_set_current_point(EPI_Imaccess *access, reeltype real_val) 
 * 
 * Action :
 *   set the value accoording to the current access pointer.
 * 
 * Include :
 *   <epi_extent.h>
 *
 * Parameters in :
 *   EPI_Imaccess *access : acess structure for the input image.
 *    reeltype real_val : current point value (normalized in [0, ..1.0]
 *                    if the image type is TYPEIM_256 or TYPEIM_16B)
 *   
 * Parameters out :
 *
 * Parameters in/out :
 *
 * Return value : 
 *   int :  1 if OK;  0 if not.
 *
 * Description :
 *
 * Errors :
 *   
 *---------------------------------------------------------------------------
 */
#if defined(_ANSI_)
int EPI_set_current_point(EPI_Imaccess *access, reeltype real_val)
#else
int EPI_set_current_point(access, real_val)
EPI_Imaccess *access;
reeltype real_val;

#endif
{ 
  int result = 0;
  unsigned char *ubyte_ptr;
  unsigned short *ushort_ptr;
  short *short_ptr;
  reeltype *float_ptr;
  
  switch(access -> type) {
  case TYPEIM_256:
    ubyte_ptr = (unsigned char *)(access -> curr_adr);
    *ubyte_ptr = (int) (real_val * 255.0);
    break;
  case TYPEIM_16B:
    ushort_ptr = (unsigned short *)(access -> curr_adr);
    *ushort_ptr = (int) (real_val * 65535.0);
    break;
  case TYPEIM_S16B:
    short_ptr = (short *)(access -> curr_adr);
    *short_ptr = (int) (real_val * 32767.0);
    break;
  case TYPEIM_FLOAT:
    float_ptr = (reeltype *)(access -> curr_adr);
    *float_ptr = real_val;
    break;
  default:
    T_Set3Point(access -> curr_x,
		access -> curr_y,
		access -> curr_z,
		access -> ima,
		real_val, MODE_SET);
    break;
  }
  
  result = 1;
  return(result);
}



/*---------------------------------------------------------------------------
 * 
 * int EPI_set_next_x_point(EPI_Imaccess *access, reeltype real_val)
 * 
 * Action :
 *  set the value accoording to the next position along x of the current access
 *  adress.
 *
 * Include :
 *  <epi_extrent.h>
 *
 * Parameters in :
 *   EPI_Imaccess *access : acess structure for the input image.
 *    reeltype real_val : output value (normalized in [0, ..1.0]
 *                    if the image type is TYPEIM_256 or TYPEIM_16B)
 *   
 * Parameters out :
 *
 * Parameters in/out :
 *   
 * Return value : 
 *   int :  1 if OK;  0 if not.
 *
 * Description :
 *    *(curr_adr + voxel_size) = real_val * CTE.
 *    Remark :
 *    This value accords to the position (x, y, z), where
 *    if (curr_x < tx - 1),  (x, y, z) = (curr_x + 1, curr_y, curr_z) 
 *    else
 *      if (cuur_y < ty -1), (x, y, z) = (0, curr_y + 1, curr_z); 
 *      else 
 *         (x, y, z) = (0, 0, curr_z + 1).
 *
 * Errors :
 *   
 *---------------------------------------------------------------------------
 */
#if defined(_ANSI_)
int EPI_set_next_x_point(EPI_Imaccess *access, reeltype real_val)
#else
int EPI_set_next_x_point(access, real_val)
EPI_Imaccess *access;
reeltype real_val;

#endif
{ 
  int result = 0;
  unsigned char *ubyte_ptr;
  unsigned short *ushort_ptr;
  short *short_ptr;
  reeltype *float_ptr;
  int x, y, z;

  switch(access -> type) {
  case TYPEIM_256:
    ubyte_ptr = (unsigned char *)(access -> curr_adr);
    ubyte_ptr ++;
    *ubyte_ptr = (int)(real_val * 255.0);
    break;
  case TYPEIM_16B:
    ushort_ptr = (unsigned short *)(access -> curr_adr);
    ushort_ptr ++;
    *ushort_ptr = (int)(real_val * 65535.0);
    break;
  case TYPEIM_S16B:
    short_ptr = (short *)(access -> curr_adr);
    short_ptr ++;
    *short_ptr = (int)(real_val * 32767.0);
    break;
  case TYPEIM_FLOAT:
    float_ptr = (reeltype *)(access -> curr_adr);
    float_ptr ++;
    *float_ptr = real_val;
    break;
  default:
    if (access -> curr_x < access -> tx - 1) {
      /*------ 
	x = x + 1
	------*/
      x = access -> curr_x + 1;
      y = access -> curr_y;
      z = access -> curr_z;
    }
    else {
      if (access -> curr_y < access -> ty - 1) {
	/*------- 
	  y = y + 1
	  -------*/
	x = 0;
	y = access -> curr_y + 1;
	z = access -> curr_z;
      }
      else {
	/*-------
	  z = z + 1
	  -------*/
	x = 0;
	y = 0;
	z = access -> curr_z + 1;
      }
    }
    
    T_Set3Point(x, y, z, access -> ima, real_val, MODE_SET);
    break;
  }
  
  result = 1;
  return(result);
}


/*---------------------------------------------------------------------------
 * 
 * int EPI_set_prev_x_point(EPI_Imaccess *access, reeltype real_val)
 * 
 * Action :
 *  set the value accoording to the previous position along x of the current access
 *  adress.
 *
 * Include :
 *  <epi_extrent.h>
 *
 * Parameters in :
 *   EPI_Imaccess *access : acess structure for the input image.
 *    reeltype real_val : input value (normalized in [0, ..1.0]
 *                    if the image type is TYPEIM_256 or TYPEIM_16B)
 *   
 * Parameters out :
 *   
 * Parameters in/out :
 *   
 * Return value : 
 *   int :  1 if OK;  0 if not.
 *
 * Description :
 *    *(curr_adr - voxel_size) = real_val * CTE.
 *    Remark :
 *    This value accords to the position (x, y, z), where
 *    if (curr_x > 0),  (x, y, z) = (curr_x - 1, curr_y, curr_z) 
 *    else
 *      if (cuur_y > 0), (x, y, z) = (tx - 1, curr_y - 1, curr_z); 
 *      else 
 *         (x, y, z) = (tx - 1, ty -1, curr_z - 1).
 *
 * Errors :
 *   
 *---------------------------------------------------------------------------
 */
#if defined(_ANSI_)
int EPI_set_prev_x_point(EPI_Imaccess *access, reeltype real_val)
#else
int EPI_set_prev_x_point(access, real_val)
EPI_Imaccess *access;
reeltype real_val;

#endif
{ int result = 0;
  unsigned char *ubyte_ptr;
  unsigned short *ushort_ptr;
  short *short_ptr;
  reeltype *float_ptr;
  int x, y, z;
  
  switch(access -> type) {
  case TYPEIM_256:
    ubyte_ptr = (unsigned char *)(access -> curr_adr);
    ubyte_ptr --;
    *ubyte_ptr = (int)(real_val * 255.0);
    break;
  case TYPEIM_16B:
    ushort_ptr = (unsigned short *)(access -> curr_adr);
    ushort_ptr --;
    *ushort_ptr = (int)(real_val * 65535.0);
    break;
  case TYPEIM_S16B:
    short_ptr = (short *)(access -> curr_adr);
    short_ptr --;
    *short_ptr = (int)(real_val * 32767.0);
    break;
  case TYPEIM_FLOAT:
    float_ptr = (reeltype *)(access -> curr_adr);
    float_ptr --;
    *float_ptr = real_val;
    break;
  default:
    if (access -> curr_x > 0) {
      /*------ 
	x = x - 1
	------*/
      x = access -> curr_x - 1;
      y = access -> curr_y;
      z = access -> curr_z;
    }
    else {
      if (access -> curr_y > 0) {
	/*------- 
	  y = y - 1
	  -------*/
	x = access -> tx - 1;
	y = access -> curr_y - 1;
	z = access -> curr_z;
      }
      else {
	/*-------
	  z = z - 1
	  -------*/
	x = access -> tx - 1;
	y = access -> ty - 1;
	z = access -> curr_z - 1;
      }
    }
    
    T_Set3Point(x, y, z, access -> ima, real_val, MODE_SET);
    break;
  }
  
  result = 1;
  return(result);
}


/*---------------------------------------------------------------------------
 * 
 * int EPI_set_next_y_point(EPI_Imaccess *access, reeltype real_val)
 * 
 * Action :
 *  set the value accoording to the next position along y of the current access
 *  adress.
 *
 * Include :
 *  <epi_extrent.h>
 *
 * Parameters in :
 *   EPI_Imaccess *access : acess structure for the input image.
 *    reeltype real_val : input value (normalized in [0, ..1.0]
 *                    if the image type is TYPEIM_256 or TYPEIM_16B)
 *   
 * Parameters out :
 *
 * Parameters in/out :
 *   
 * Return value : 
 *   
 * Description :
 *    *(curr_adr + ln_size) = real_val * CTE.
 *    Remark :
 *    This value accords to the position (x, y, z), where
 *    if (curr_y < ty -1), (x, y, z) = (curr_x, curr_y + 1, curr_z); 
 *    else 
 *         (x, y, z) = (curr_x, 0, curr_z + 1).
 *
 * Errors :
 *   
 *---------------------------------------------------------------------------
 */
#if defined(_ANSI_)
int EPI_set_next_y_point(EPI_Imaccess *access, reeltype real_val)
#else
int EPI_set_next_y_point(access, real_val)
EPI_Imaccess *access;
reeltype real_val;

#endif
{ int result = 0;
  unsigned char *ubyte_ptr;
  unsigned short *ushort_ptr;
  short *short_ptr;
  reeltype *float_ptr;
  int x, y, z;

  switch(access -> type) {
  case TYPEIM_256:
    ubyte_ptr = (unsigned char *)(access -> curr_adr);
    ubyte_ptr += (access -> ln_voxel_nb);
    *ubyte_ptr = (int)(real_val * 255.0);
    break;
  case TYPEIM_16B:
    ushort_ptr = (unsigned short *)(access -> curr_adr);
    ushort_ptr += (access -> ln_voxel_nb);
    *ushort_ptr = (int)(real_val * 65535.0);
    break;
  case TYPEIM_S16B:
    short_ptr = (short *)(access -> curr_adr);
    short_ptr += (access -> ln_voxel_nb);
    *short_ptr = (int)(real_val * 32767.0);
    break;
  case TYPEIM_FLOAT:
    float_ptr = (reeltype *)(access -> curr_adr);
    float_ptr += (access -> ln_voxel_nb);
    *float_ptr = real_val;
    break;
  default:
    x = access -> curr_x;
    y = access -> curr_y  + 1;
    z = access -> curr_z;
    if (access -> curr_y == access -> ty -1)  {
      /*-----
	z = z + 1
	-------*/
      y = 0;
      z ++;
    }
    
    T_Set3Point(x, y, z, access -> ima, real_val, MODE_SET);
    
    break;
  }
  
  result = 1;
  return(result);
}


/*---------------------------------------------------------------------------
 * 
 * int EPI_set_prev_y_point(EPI_Imaccess *access, reeltype real_val)
 * 
 * Action :
 *  set the value accoording to the previous position along y of the current access
 *  adress.
 *
 * Include :
 *  <epi_extrent.h>
 *
 * Parameters in :
 *   EPI_Imaccess *access : acess structure for the input image.
 *    reeltype real_val : input value (normalized in [0, ..1.0]
 *                    if the image type is TYPEIM_256 or TYPEIM_16B)
 *   
 * Parameters out :
 *   
 * Parameters in/out :
 *   
 * Return value : 
 *    int :  1 if OK;  0 if not.
 *
 * Description :
 *    *(curr_adr - ln_size) = real_val * CTE.
 *    Remark :
 *    This value accords to the position (x, y, z), where
 *    if (curr_y > 0), (x, y, z) = (curr_x, curr_y - 1, curr_z); 
 *    else 
 *      (x, y, z) = (curr_x, ty -1, curr_z - 1).
 *
 * Errors :
 *   
 *---------------------------------------------------------------------------
 */
#if defined(_ANSI_)
int EPI_set_prev_y_point(EPI_Imaccess *access, reeltype real_val)
#else
int EPI_set_prev_y_point(access, real_val)
EPI_Imaccess *access;
reeltype real_val;

#endif
{ int result = 0;
  unsigned char *ubyte_ptr;
  unsigned short *ushort_ptr;
  short *short_ptr;
  reeltype *float_ptr;
  int x, y, z;
  
  switch(access -> type) {
  case TYPEIM_256:
    ubyte_ptr = (unsigned char *)(access -> curr_adr);
    ubyte_ptr -= (access -> ln_voxel_nb);
    *ubyte_ptr = (int)(real_val * 255.0);
    break;
  case TYPEIM_16B:
    ushort_ptr = (unsigned short *)(access -> curr_adr);
    ushort_ptr -= (access -> ln_voxel_nb);
    *ushort_ptr = (int)(real_val * 65535.0);
    break;
  case TYPEIM_S16B:
    short_ptr = (short *)(access -> curr_adr);
    short_ptr -= (access -> ln_voxel_nb);
    *short_ptr = (int)(real_val * 32767.0);
    break;
  case TYPEIM_FLOAT:
    float_ptr = (reeltype *)(access -> curr_adr);
    float_ptr -= (access -> ln_voxel_nb);
    *float_ptr = real_val;
    break;
  default:
    x = access -> curr_x;
    y = access -> curr_y - 1;
    z = access -> curr_z;
    
    if (access -> curr_y == 0) {
      /*-----
	z = z - 1
	-------*/
      y = access -> ty - 1;
      z --;
    }
    
    T_Set3Point(x, y, z, access -> ima, real_val, MODE_SET);
    
    break;
  }
  
  result = 1;
  return(result);
}


/*---------------------------------------------------------------------------
 * 
 * int EPI_set_next_z_point(EPI_Imaccess *access, reeltype real_val)
 * 
 * Action :
 *  set the value accoording to the next position along z of the current access
 *  adress.
 *
 * Include :
 *  <epi_extrent.h>
 *
 * Parameters in :
 *   EPI_Imaccess *access : acess structure for the input image.
 *    reeltype real_val : input value (normalized in [0, ..1.0]
 *                    if the image type is TYPEIM_256 or TYPEIM_16B)
 *
 * Parameters out :
 *
 * Parameters in/out :
 *   
 * Return value : 
 *   
 * Description :
 *    *(curr_adr + pln_size) = real_val * CTE.
 *    Remark :
 *    This value accords to the position (x, y, z), where
 *    (x, y, z) = (curr_x, curr_y, curr_z + 1).
 *    User must verify the condition (curr_z < tz - 1).
 * 
 * Errors :
 *   
 *---------------------------------------------------------------------------
 */
#if defined(_ANSI_)
int EPI_set_next_z_point(EPI_Imaccess *access, reeltype real_val)
#else
int EPI_set_next_z_point(access, real_val)
EPI_Imaccess *access;
reeltype real_val;

#endif
{ int result = 0;
  unsigned char *ubyte_ptr;
  unsigned short *ushort_ptr;
  short *short_ptr;
  reeltype *float_ptr;
  
  switch(access -> type) {
  case TYPEIM_256:
    ubyte_ptr = (unsigned char *)(access -> curr_adr);
    ubyte_ptr += (access -> pln_voxel_nb);
    *ubyte_ptr = (int)(real_val * 255.0);
    break;
  case TYPEIM_16B:
    ushort_ptr = (unsigned short *)(access -> curr_adr);
    ushort_ptr += (access -> pln_voxel_nb);
    *ushort_ptr = (int)(real_val * 65535.0);
    break;
  case TYPEIM_S16B:
    short_ptr = (short *)(access -> curr_adr);
    short_ptr += (access -> pln_voxel_nb);
    *short_ptr = (int)(real_val * 32767.0);
    break;
  case TYPEIM_FLOAT:
    float_ptr = (reeltype *)(access -> curr_adr);
    float_ptr += (access -> pln_voxel_nb);
    *float_ptr = real_val;
    break;
  default:
    T_Set3Point(access -> curr_x,
		access -> curr_y,
		access -> curr_z + 1,
		access -> ima,
		real_val, MODE_SET);
    break;
  }
  
  result = 1;
  return(result);
}


/*---------------------------------------------------------------------------
 * 
 * int EPI_set_prev_z_point(EPI_Imaccess *access, reeltype real_val)
 * 
 * Action :
 *  set the value accoording to the previous position along z of the current access
 *  adress.
 *
 * Include :
 *  <epi_extrent.h>
 *
 * Parameters in :
 *   EPI_Imaccess *access : acess structure for the input image.
 *    reeltype real_val : input value (normalized in [0, ..1.0]
 *                    if the image type is TYPEIM_256 or TYPEIM_16B)
 *   
 * Parameters out :
 *
 * Parameters in/out :
 *   
 * Return value : 
 *   int :   1 if OK;  0 if not.
 *
 * Description :
 *    *(curr_adr - pln_size) = real * CTE.
 *    Remark :
 *    This value accords to the position (x, y, z), where
 *    (x, y, z) = (curr_x, curr_y, curr_z - 1).
 *    User must verify the condition (curr_z > 0).
 * 
 * Errors :
 *   
 *---------------------------------------------------------------------------
 */
#if defined(_ANSI_)
int EPI_set_prev_z_point(EPI_Imaccess *access, reeltype real_val)
#else
int EPI_set_prev_z_point(access, real_val)
EPI_Imaccess *access;
reeltype real_val;

#endif
{ int result = 0;
  unsigned char *ubyte_ptr;
  unsigned short *ushort_ptr;
  short *short_ptr;
  reeltype *float_ptr;
  
  switch(access -> type) {
  case TYPEIM_256:
    ubyte_ptr = (unsigned char *)(access -> curr_adr);
    ubyte_ptr -= (access -> pln_voxel_nb);
    *ubyte_ptr = (int)(real_val * 255.0);
    break;
  case TYPEIM_16B:
    ushort_ptr = (unsigned short *)(access -> curr_adr);
    ushort_ptr -= (access -> pln_voxel_nb);
    *ushort_ptr = (int)(real_val * 65535.0);
    break;
  case TYPEIM_S16B:
    short_ptr = (short *)(access -> curr_adr);
    short_ptr -= (access -> pln_voxel_nb);
    *short_ptr = (int)(real_val * 32767.0);
    break;
  case TYPEIM_FLOAT:
    float_ptr = (reeltype *)(access -> curr_adr);
    float_ptr -= (access -> pln_voxel_nb);
    *float_ptr = real_val;
    break;
  default:
    T_Set3Point(access -> curr_x,
		access -> curr_y,
		access -> curr_z - 1,
		access -> ima,
		real_val, MODE_SET);
    break;
  }
  
  result = 1;
  return(result);
}


/*---------------------------------------------------------------------------
 * 
 * int EPI_set_current_offset_point(EPI_Imaccess *access, 
 *                               int offset_x, 
 *                               int offset_y, 
 *                               int offset_z, 
 *                               reeltype real_val)
 * 
 * Action :
 *  set the position value according to an offset along x, y and z from the 
 *  current access adress.
 *
 * Include :
 *  <epi_extrent.h>
 *
 * Parameters in :
 *   EPI_Imaccess *access : acess structure for the input image.
 *    int offset_x, offset_y, offset_z : offset along the x, y and z axis
 *    reeltype real_val : intput value (normalized in [0, ..1.0]
 *                    if the image type is TYPEIM_256 or TYPEIM_16B)
 *
 * Parameters out :
 *
 * Parameters in/out :
 *   
 * Return value : 
 *   int :  1 if OK;  0 if not.
 *   
 * Description :
 *   The input value accords to the position 
 *     (curr_x + offset_x, curr_y + offset_y, curr_z + offset_z).
 *
 *   But this adress is computed from the old current adress and its offset
 *   possible.
 *     i.e. curr_adr = curr_adr + offset_z * pln_size +
 *                     offset_y * ln_size + offset_x * voxel_size.
 *   This algorithm is optimized if 
 *     offset_z = 0 or (and) offset_y = 0 or (and) offset_x = 0.
 *
 * Errors :
 *   
 *---------------------------------------------------------------------------
 */
#if defined(_ANSI_)
int EPI_set_current_offset_point(EPI_Imaccess *access, 
				 int offset_x, 
				 int offset_y, 
				 int offset_z, 
				 reeltype real_val)
#else
int EPI_set_current_offset_point(access, offset_x, offset_y, offset_z, real_val)
EPI_Imaccess *access;
int offset_x;
int offset_y;
int offset_z;
reeltype real_val;

#endif
{ int result = 0;
  unsigned char *ubyte_ptr;
  unsigned short *ushort_ptr;
  short *short_ptr;
  reeltype *float_ptr;
  char *tmp_ptr;
  
  tmp_ptr = (char *)(access -> curr_adr);
  
  if (offset_z != 0) 
    tmp_ptr += offset_z * (access -> pln_size);
  
  if (offset_y != 0) 
    tmp_ptr += offset_y * (access -> ln_size);
  
  if (offset_x != 0) 
    tmp_ptr += offset_x * (access -> voxel_size);
  
  switch(access -> type) {
  case TYPEIM_256:
    ubyte_ptr = (unsigned char *)tmp_ptr;
    *ubyte_ptr = (int)(real_val * 255.0);
    break;
  case TYPEIM_16B:
    ushort_ptr = (unsigned short *)tmp_ptr;
    *ushort_ptr = (int)(real_val * 65535.0);
    break;
  case TYPEIM_S16B:
    short_ptr = (short *)tmp_ptr;
    *short_ptr = (int)(real_val * 32767.0);
    break;
  case TYPEIM_FLOAT:
    float_ptr = (reeltype *)tmp_ptr;
    *float_ptr = real_val;
    break;
  default:
      T_Set3Point(access -> curr_x + offset_x,
		  access -> curr_y + offset_y,
		  access -> curr_z + offset_z,
		  access -> ima, real_val, MODE_SET);
    break;
  }
  
  result = 1;
  return(result);
}


/*---------------------------------------------------------------------------
 *      Fin de fichier 
 *---------------------------------------------------------------------------
 */















