/* epi_extrem.c 2.1 5/14/98 */
/*==========================================================================
*
* File : $EPI/src/epi_extrem.c
* Module :
* $ Id : ? $
* Test file : ?
* Author : J-P Thirion
* Creation: August 4 1992
*===========================================================================
*   (c) Copyright 1992, Inria, all right reserved
*===========================================================================
*
*------------------------------Description----------------------------------
*
* Calcul des coefficients d'extremalite', et des caracte'ristiques
* differentielles des iso-surfaces, a` partir des diffe'rentielles
* d'image.
*
*-----------------------------Modification----------------------------------
*
*/

/*======================================================================*/
/*
  epi_extrem.c

  Compute the extremality coefficients for 3D surfaces
  
  (c) Jean-Philippe THIRION -- INRIA -- 2/19/92
*/
/*======================================================================*/

/* ATTENTION ATTENTION ATTENTION ATTENTION ATTENTION ATTENTION ATTENTIO */
/* Attention cDer est deja un type defini dans epi_spline.h, j'ai donc  */
/* transforme tous les cDer du fichier en cDeri                         */
/* 05 avril 1994 Gregoire                                               */
/* ATTENTION ATTENTION ATTENTION ATTENTION ATTENTION ATTENTION ATTENTIO */

#if defined(INC_MIN) /* inclusions minimales */
 /* rajouter -DINC_MIN dans les options de compilation */
 /* GM 12/09/94                                        */
#include <epi_kernel.ei>
#include <epi_extrem.ei>
#include <epi_filter.ei>
#include <epi_crest.h>
#else /* inclusions maximales */
#include <epidaure.h>
#include <epidaure.ee>
#include <epidaure.ei>
#endif

/*-------Definition des fonctions statiques----------*/
#ifndef NO_PROTO
static void _T_GetSurfCaract( t_3DerivatesPtr der3, t_SurfDerPtr theDer, t_3PointPtr normal);
static void _T_GetAllDifCaract( t_3DerivatesPtr der3, t_SurfDerPtr theDer, t_3PointPtr normal,int flags);

#ifdef COMPILE_UNUSED
static void _T_3DerTest( reeltype x, reeltype y, t_3DerivatesPtr der3 );
static void _T_3DerTest2( reeltype x, reeltype y, t_3DerivatesPtr der3 );
static void _T_3DerTest3( reeltype x, reeltype y, t_3DerivatesPtr der3 );
static void _T_3DerTest4( reeltype x, reeltype y, t_3DerivatesPtr der3 );
static void _T_3DerTest5( reeltype x, reeltype y, t_3DerivatesPtr der3 );
static void _T_3DerTest6( reeltype x, reeltype y, t_3DerivatesPtr der3 );
static void _T_3ExtremTest( t_ImagePtr theIm );
#endif /* COMPILE_UNUSED */
#else 
static void _T_GetSurfCaract();
static void _T_GetAllDifCaract();
#ifdef COMPILE_UNUSED
static void _T_3DerTest();
static void _T_3DerTest2();
static void _T_3DerTest3();
static void _T_3DerTest4();
static void _T_3DerTest5();
static void _T_3DerTest6();
static void _T_3ExtremTest();
#endif /* COMPILE_UNUSED */
#endif
/*-------Fin des proto statiques----------*/

#ifndef E_ISO_SECOND
#define E_ISO_SECOND        0x0200      /* extremal lines of second extr.   */
#endif

#ifndef E_ISO_QUADRIC
#define E_ISO_QUADRIC       0x4000      /* (internally used)                */
#endif


/************************************************************************/
/*
  Extrem3D

  Compute the extremality coefficients for 3D surfaces
  
  (c) Jean-Philippe THIRION -- INRIA -- 2/19/92
*/
/************************************************************************/

/*----------------------------------------------------------------------*/
/*	
	Compute all the differential characteristics of the surface.

	It starts with
	the differencials of the image up to order 3. Doesn't take into
	account any orientation problems.

	Also, it doesn't switch the principal directions according to their
	absolute values.
*/

#if defined(_ANSI_)
static void _T_GetSurfCaract( t_3DerivatesPtr der3, t_SurfDerPtr theDer, t_3PointPtr normal)
#else
static void _T_GetSurfCaract(der3, theDer, normal)
t_3DerivatesPtr der3;
t_SurfDerPtr theDer;
t_3PointPtr normal;

#endif

{
    static double fx,fy,fz,fxx,fxy,fxz,fyy,fyz,fzz;
    static double fxxx,fxxy,fxxz,fxyy,fxyz,fxzz,fyyy,fyyz,fyzz,fzzz;
	
    static double _H,S,K;
    static double A1,A2,A3;
    static double B1,B2,B3;
    static double D1;
    static double E1,E2,E3;
    static double F1;
    static double delta,delta2;
    static double norm1,norm2;

    static double dHx, dHy, dHz, dSx, dSy, dSz, dKx, dKy, dKz;
    static double dA1x, dA1y, dA1z, dA2x, dA2y, dA2z, dA3x, dA3y, dA3z; 
    static double dB1x, dB1y, dB1z, dB2x, dB2y, dB2z, dB3x, dB3y, dB3z; 
    static double dC1x, dC1y, dC1z, dC2x, dC2y, dC2z, dC3x, dC3y, dC3z; 

    static double dD1x, dD1y, dD1z;
    static double dE1x, dE1y, dE1z, dE2x, dE2y, dE2z, dE3x, dE3y, dE3z; 
    static double dF1x, dF1y, dF1z;
    static double dDeltax, dDeltay, dDeltaz, dDelta2x, dDelta2y, dDelta2z;
    static double dk1x, dk1y, dk1z, dk2x, dk2y, dk2z;
    static t_3Point a,b,t1,t2,n;
    
    theDer->g= 0.0;
    theDer->m= 0.0;
    theDer->k1= 0.0;
    theDer->k2= 0.0;
    theDer->norm= 0.0;
    T_3Form(0.0,0.0,0.0,&theDer->grad );
    T_3Form(0.0,0.0,0.0,&theDer->t1 );
    T_3Form(0.0,0.0,0.0,&theDer->t2 );
    T_3Form(0.0,0.0,0.0,&theDer->dk1 );
    T_3Form(0.0,0.0,0.0,&theDer->dk2 );
    theDer->e1= 0.0;
    theDer->e2= 0.0;
    theDer->good= TRUE;
    
    fx= der3->fx;
    fy= der3->fy;
    fz= der3->fz;

    fxy= der3->fxy;
    fxz= der3->fxz;
    fyz= der3->fyz;

    fxx= der3->fxx;
    fyy= der3->fyy;
    fzz= der3->fzz;

    fxxx= der3->fxxx;
    fyyy= der3->fyyy;
    fzzz= der3->fzzz;

    fxxy= der3->fxxy;
    fxxz= der3->fxxz;
    fxyy= der3->fxyy;
    fxyz= der3->fxyz;
    fyyz= der3->fyyz;
    fyzz= der3->fyzz;
    fxzz= der3->fxzz;

    _H= fx*fx+fy*fy+fz*fz;
    n.x= (reeltype) fx;
    n.y= (reeltype) fy;
    n.z= (reeltype) fz;
    T_3Normalize( &n, &n );
    if ( normal ) T_3Copy( &n, normal );

    theDer->norm= (reeltype) sqrt(_H);

    if ( theDer->norm < 1e-20 ){
		theDer->good= FALSE; 
		return; 
		}

	T_3Form((reeltype) fx,(reeltype) fy,(reeltype) fz,&theDer->grad );

	A1= fyy*fzz - fyz*fyz;		B1= fxz*fxy - fxx*fyz;	
	A2= fxx*fzz - fxz*fxz;		B2= fyz*fxy - fyy*fxz;	
	A3= fxx*fyy - fxy*fxy;		B3= fxz*fyz - fzz*fxy;	

	D1= fx*fx*A1 + 2.0*fy*fz*B1 +
	  fy*fy*A2 + 2.0*fx*fz*B2 +
	  fz*fz*A3 + 2.0*fx*fy*B3;

	K= D1 / (_H*_H);

/*.................	computation of Kx	...........*/

	dA1x= fxyy*fzz + fyy*fxzz - 2.0*fyz*fxyz;
	dA2x= fxxx*fzz + fxx*fxzz - 2.0*fxz*fxxz;
	dA3x= fxxx*fyy + fxx*fxyy - 2.0*fxy*fxxy;

	dB1x= fxxz*fxy + fxz*fxxy - fxxx*fyz - fxx*fxyz;
	dB2x= fxyz*fxy + fyz*fxxy - fxyy*fxz - fyy*fxxz;
	dB3x= fxxz*fyz + fxz*fxyz - fxzz*fxy - fzz*fxxy;

	dC1x= fx*fx*dA1x + 2.0*fx*fxx*A1 + 2.0*fy*fz*dB1x + 2.0*(fxy*fz+fy*fxz)*B1;
	dC2x= fy*fy*dA2x + 2.0*fy*fxy*A2 + 2.0*fx*fz*dB2x + 2.0*(fxx*fz+fx*fxz)*B2;
	dC3x= fz*fz*dA3x + 2.0*fz*fxz*A3 + 2.0*fx*fy*dB3x + 2.0*(fxx*fy+fx*fxy)*B3;
	
	dD1x= dC1x + dC2x + dC3x;
	dHx= fx*fxx*2.0 + fy*fxy*2.0 + fz*fxz*2.0;
	dKx= (dD1x*_H - 2.0*dHx*D1) / (_H*_H*_H);

/*.................	computation of Ky	...........*/

	dA1y= fyyy*fzz + fyy*fyzz - 2.0*fyz*fyyz;
	dA2y= fxxy*fzz + fxx*fyzz - 2.0*fxz*fxyz;
	dA3y= fxxy*fyy + fxx*fyyy - 2.0*fxy*fxyy;

	dB1y= fxyz*fxy + fxz*fxyy - fxxy*fyz - fxx*fyyz;
	dB2y= fyyz*fxy + fyz*fxyy - fyyy*fxz - fyy*fxyz;
	dB3y= fxyz*fyz + fxz*fyyz - fyzz*fxy - fzz*fxyy;

	dC1y= fx*fx*dA1y + 2.0*fx*fxy*A1 + 2.0*fy*fz*dB1y + 2.0*(fyy*fz+fy*fyz)*B1;
	dC2y= fy*fy*dA2y + 2.0*fy*fyy*A2 + 2.0*fx*fz*dB2y + 2.0*(fxy*fz+fx*fyz)*B2;
	dC3y= fz*fz*dA3y + 2.0*fz*fyz*A3 + 2.0*fx*fy*dB3y + 2.0*(fxy*fy+fx*fyy)*B3;

	dD1y= dC1y + dC2y + dC3y;
	dHy= 2.0*fx*fxy + 2.0*fy*fyy + 2.0*fz*fyz;
	dKy= (dD1y*_H - 2.0*dHy*D1) / (_H*_H*_H);

/*.................	computation of Kz	...........*/

	dA1z= fyyz*fzz + fyy*fzzz - 2.0*fyz*fyzz;
	dA2z= fxxz*fzz + fxx*fzzz - 2.0*fxz*fxzz;
	dA3z= fxxz*fyy + fxx*fyyz - 2.0*fxy*fxyz;

	dB1z= fxzz*fxy + fxz*fxyz - fxxz*fyz - fxx*fyzz;
	dB2z= fyzz*fxy + fyz*fxyz - fyyz*fxz - fyy*fxzz;
	dB3z= fxzz*fyz + fxz*fyzz - fzzz*fxy - fzz*fxyz;

	dC1z= fx*fx*dA1z + 2.0*fx*fxz*A1 + 2.0*fy*fz*dB1z + 2.0*(fyz*fz+fy*fzz)*B1;
	dC2z= fy*fy*dA2z + 2.0*fy*fyz*A2 + 2.0*fx*fz*dB2z + 2.0*(fxz*fz+fx*fzz)*B2;
	dC3z= fz*fz*dA3z + 2.0*fz*fzz*A3 + 2.0*fx*fy*dB3z + 2.0*(fxz*fy+fx*fyz)*B3;

	dD1z= dC1z + dC2z + dC3z;
	dHz= 2.0*fx*fxz + 2.0*fy*fyz + 2.0*fz*fzz;
	dKz= (dD1z*_H - 2.0*dHz*D1) / (_H*_H*_H);

	/* T_3Copy( &dK, &theDer->dK ); */
	theDer->dK.x = (reeltype) dKx;
	theDer->dK.y = (reeltype) dKy;
	theDer->dK.z = (reeltype) dKz;
	
/*.................	computation of S	...........*/
	
	E1= fx*fx*(fyy+fzz) - 2.0*fy*fz*fyz;	
	E2= fy*fy*(fxx+fzz) - 2.0*fx*fz*fxz;	
	E3= fz*fz*(fxx+fyy) - 2.0*fx*fy*fxy;	
	F1= E1+E2+E3;
	
	S= F1 / (_H*sqrt(_H));

/*.................	computation of Sx	...........*/

	dE1x= 2.0*fx*fxx*(fyy+fzz) + fx*fx*(fxyy+fxzz)
	  -2.0*( fxy*fz*fyz + fy*fxz*fyz + fy*fz*fxyz );

	dE2x= 2.0*fy*fxy*(fxx+fzz) + fy*fy*(fxxx+fxzz)
	  -2.0*( fxx*fz*fxz + fx*fxz*fxz + fx*fz*fxxz );

	dE3x= 2.0*fz*fxz*(fxx+fyy) + fz*fz*(fxxx+fxyy)
	  -2.0*( fxx*fy*fxy + fx*fxy*fxy + fx*fy*fxxy );

	dF1x= dE1x + dE2x + dE3x;
	dSx= (dF1x*_H - 3*F1*dHx/2) / (_H*_H*sqrt(_H));

/*.................	computation of Sy	...........*/

	dE1y= 2.0*fx*fxy*(fyy+fzz) + fx*fx*(fyyy+fyzz)
	  -2.0*( fyy*fz*fyz + fy*fyz*fyz + fy*fz*fyyz );

	dE2y= 2.0*fy*fyy*(fxx+fzz) + fy*fy*(fxxy+fyzz)
	  -2.0*( fxy*fz*fxz + fx*fyz*fxz + fx*fz*fxyz );

	dE3y= 2.0*fz*fyz*(fxx+fyy) + fz*fz*(fxxy+fyyy)
	  -2.0*( fxy*fy*fxy + fx*fxy*fyy + fx*fy*fxyy );

	dF1y= dE1y + dE2y + dE3y;
	dSy= (dF1y*_H - 3*F1*dHy/2) / (_H*_H*sqrt(_H));

/*.................	computation of Sz	...........*/

	dE1z= 2.0*fx*fxz*(fyy+fzz) + fx*fx*(fyyz+fzzz)
	  -2.0*( fyz*fz*fyz + fy*fzz*fyz + fy*fz*fyzz );
	
	dE2z= 2.0*fy*fyz*(fxx+fzz) + fy*fy*(fxxz+fzzz)
	  -2.0*( fxz*fz*fxz + fx*fzz*fxz + fx*fz*fxzz );

	dE3z= 2.0*fz*fzz*(fxx+fyy) + fz*fz*(fxxz+fyyz)
	  -2.0*( fxz*fy*fxy + fx*fyz*fxy + fx*fy*fxyz );

	dF1z= dE1z + dE2z + dE3z;
	dSz= (dF1z*_H - 3*F1*dHz/2) / (_H*_H*sqrt(_H));

	/* T_3Copy( &dS, &theDer->dS ); */
	theDer->dS.x = (reeltype) dSx; 
	theDer->dS.y = (reeltype) dSy; 
	theDer->dS.z = (reeltype) dSz; 
	
/*.................	computation of k1,k2 and derivatives ..........*/

	delta2= S*S - 4*K;
	if ( delta2 > 0 ){
		delta= sqrt(delta2);
		dDelta2x= 2.0*S*dSx - 4.0*dKx;
		dDelta2y= 2.0*S*dSy - 4.0*dKy;
		dDelta2z= 2.0*S*dSz - 4.0*dKz;
		dDeltax= dDelta2x / (2.0*delta);
		dDeltay= dDelta2y / (2.0*delta);
		dDeltaz= dDelta2z / (2.0*delta);
		}
	else{
		delta= 0.0;
		dDeltax= 0.0;
		dDeltay= 0.0;
		dDeltaz= 0.0;
		}


	theDer->g= (reeltype) K;
	theDer->m= (reeltype) (S/2);
	theDer->k1= (reeltype) ((S + delta) / 2.0);
	theDer->k2= (reeltype) ((S - delta) / 2.0);

	dk1x= (dSx + dDeltax) / 2.0;
	dk1y= (dSy + dDeltay) / 2.0;
	dk1z= (dSz + dDeltaz) / 2.0;

	dk2x= (dSx - dDeltax) / 2.0;
	dk2y= (dSy - dDeltay) / 2.0;
	dk2z= (dSz - dDeltaz) / 2.0;

	/* T_3Copy( &dk1, &theDer->dk1 ); */
	/* T_3Copy( &dk2, &theDer->dk2 ); */
	theDer->dk1.x = (reeltype) dk1x;
	theDer->dk1.y = (reeltype) dk1y;
	theDer->dk1.z = (reeltype) dk1z;

	theDer->dk2.x = (reeltype) dk2x;
	theDer->dk2.y = (reeltype) dk2y;
	theDer->dk2.z = (reeltype) dk2z;

/*............... computation of the principal directions ........*/

	T_PrincDir( der3, &a, &b );

/*..................................................................*/


	t1.x= (reeltype) (a.x - delta * b.x / 2.0);
	t1.y= (reeltype) (a.y - delta * b.y / 2.0);
	t1.z= (reeltype) (a.z - delta * b.z / 2.0);
	t2.x= (reeltype) (a.x + delta * b.x / 2.0);
	t2.y= (reeltype) (a.y + delta * b.y / 2.0);
	t2.z= (reeltype) (a.z + delta * b.z / 2.0);

	norm1= T_3Norm( &t1 );
	norm2= T_3Norm( &t2 );
	if ( (norm1 < 1e-20) && (norm2 < 1e-20) ){ 
		theDer->good= FALSE; 
		return; 
		}

	T_3Copy( &t1, &theDer->t1 );
	T_3Copy( &t2, &theDer->t2 );
	}

/*----------------------------------------------------------------------*/
/*	
   Differential characteristics using quadric approximation
	
   This time, these are the characteristics
   of the quadric approximation of the signal, that is, only the 2nd
   order differencials are used.
*/

#if defined(_ANSI_)
void T_GetQuadricCaract( t_3DerivatesPtr der3, t_SurfDerPtr theDer, t_3PointPtr normal)
#else
void T_GetQuadricCaract(der3, theDer, normal)
t_3DerivatesPtr der3;
t_SurfDerPtr theDer;
t_3PointPtr normal;

#endif

{
    static double fx,fy,fz,fxx,fxy,fxz,fyy,fyz,fzz;
	
    static double _H,S,K;
    static double A1,A2,A3;
    static double B1,B2,B3;
    static double D1;
    static double E1,E2,E3;
    static double F1;
    static double delta,delta2;
    static double norm1,norm2;


    static double dHx, dHy, dHz, dSx, dSy, dSz, dKx, dKy, dKz;
    static double dC1x, dC1y, dC1z, dC2x, dC2y, dC2z, dC3x, dC3y, dC3z; 

    static double dD1x, dD1y, dD1z;
    static double dE1x, dE1y, dE1z, dE2x, dE2y, dE2z, dE3x, dE3y, dE3z; 
    static double dF1x, dF1y, dF1z;
    static double dDeltax, dDeltay, dDeltaz, dDelta2x, dDelta2y, dDelta2z;
    static double dk1x, dk1y, dk1z, dk2x, dk2y, dk2z;

    static t_3Point a,b,t1,t2,n;
    
    theDer->g= 0.0;
    theDer->m= 0.0;
    theDer->k1= 0.0;
    theDer->k2= 0.0;
    theDer->norm= 0.0;
    T_3Form(0.0,0.0,0.0,&theDer->grad );
    T_3Form(0.0,0.0,0.0,&theDer->t1 );
    T_3Form(0.0,0.0,0.0,&theDer->t2 );
    T_3Form(0.0,0.0,0.0,&theDer->dk1 );
    T_3Form(0.0,0.0,0.0,&theDer->dk2 );
    theDer->e1= 0.0;
    theDer->e2= 0.0;
    theDer->good= TRUE;
    
    fx= der3->fx;
    fy= der3->fy;
    fz= der3->fz;

    fxy= der3->fxy;
    fxz= der3->fxz;
    fyz= der3->fyz;

    fxx= der3->fxx;
    fyy= der3->fyy;
    fzz= der3->fzz;

    _H= fx*fx+fy*fy+fz*fz;
    n.x= (reeltype) fx;
    n.y= (reeltype) fy;
    n.z= (reeltype) fz;
    T_3Normalize( &n, &n );
    if ( normal ) T_3Copy( &n, normal );

    theDer->norm= (reeltype) sqrt(_H);

    if ( theDer->norm < 1e-20 ){
      theDer->good= FALSE; 
      return; 
    }
    
    T_3Form( (reeltype) fx, (reeltype) fy, (reeltype) fz,&theDer->grad );
    
    A1= fyy*fzz - fyz*fyz;		B1= fxz*fxy - fxx*fyz;	
    A2= fxx*fzz - fxz*fxz;		B2= fyz*fxy - fyy*fxz;	
    A3= fxx*fyy - fxy*fxy;		B3= fxz*fyz - fzz*fxy;	
    
    D1= fx*fx*A1 + 2.0*fy*fz*B1 +
      fy*fy*A2 + 2.0*fx*fz*B2 +
      fz*fz*A3 + 2.0*fx*fy*B3;
    
    K= D1 / (_H*_H);
    
    /*.................	computation of Kx	...........*/
    
    dC1x= 2.0*fx*fxx*A1 + 2.0*(fxy*fz+fy*fxz)*B1;
    dC2x= 2.0*fy*fxy*A2 + 2.0*(fxx*fz+fx*fxz)*B2;
    dC3x= 2.0*fz*fxz*A3 + 2.0*(fxx*fy+fx*fxy)*B3;
    
    dD1x= dC1x + dC2x + dC3x;
    dHx= fx*fxx*2.0 + fy*fxy*2.0 + fz*fxz*2.0;
    dKx= (dD1x*_H - 2.0*dHx*D1) / (_H*_H*_H);
    
    /*.................	computation of Ky	...........*/
    
    dC1y= 2.0*fx*fxy*A1 + 2.0*(fyy*fz+fy*fyz)*B1;
    dC2y= 2.0*fy*fyy*A2 + 2.0*(fxy*fz+fx*fyz)*B2;
    dC3y= 2.0*fz*fyz*A3 + 2.0*(fxy*fy+fx*fyy)*B3;
    
    dD1y= dC1y + dC2y + dC3y;
    dHy= 2.0*fx*fxy + 2.0*fy*fyy + 2.0*fz*fyz;
    dKy= (dD1y*_H - 2.0*dHy*D1) / (_H*_H*_H);
    
    /*.................	computation of Kz	...........*/
    
    dC1z= 2.0*fx*fxz*A1 + 2.0*(fyz*fz+fy*fzz)*B1;
    dC2z= 2.0*fy*fyz*A2 + 2.0*(fxz*fz+fx*fzz)*B2;
    dC3z= 2.0*fz*fzz*A3 + 2.0*(fxz*fy+fx*fyz)*B3;
    
    dD1z= dC1z + dC2z + dC3z;
    dHz= 2.0*fx*fxz + 2.0*fy*fyz + 2.0*fz*fzz;
    dKz= (dD1z*_H - 2.0*dHz*D1) / (_H*_H*_H);
    
    /* T_3Copy( &dK, &theDer->dK ); */
    theDer->dK.x = (reeltype) dKx;
    theDer->dK.y = (reeltype) dKy;
    theDer->dK.z = (reeltype) dKz;
    
    /*.................	computation of S	...........*/
    
    E1= fx*fx*(fyy+fzz) - 2.0*fy*fz*fyz;	
    E2= fy*fy*(fxx+fzz) - 2.0*fx*fz*fxz;	
    E3= fz*fz*(fxx+fyy) - 2.0*fx*fy*fxy;	
    F1= E1+E2+E3;
    
    S= F1 / (_H*sqrt(_H));
    
    /*.................	computation of Sx	...........*/
    
    dE1x= 2.0*fx*fxx*(fyy+fzz) - 2.0*( fxy*fz*fyz + fy*fxz*fyz );
    dE2x= 2.0*fy*fxy*(fxx+fzz) - 2.0*( fxx*fz*fxz + fx*fxz*fxz );
    dE3x= 2.0*fz*fxz*(fxx+fyy) - 2.0*( fxx*fy*fxy + fx*fxy*fxy );
    
    dF1x= dE1x + dE2x + dE3x;
    dSx= (dF1x*_H - 3.0*F1*dHx/2.0) / (_H*_H*sqrt(_H));
    
    /*.................	computation of Sy	...........*/
    
    dE1y= 2.0*fx*fxy*(fyy+fzz) - 2.0*( fyy*fz*fyz + fy*fyz*fyz );
    dE2y= 2.0*fy*fyy*(fxx+fzz) - 2.0*( fxy*fz*fxz + fx*fyz*fxz );
    dE3y= 2.0*fz*fyz*(fxx+fyy) - 2.0*( fxy*fy*fxy + fx*fxy*fyy );
    
    dF1y= dE1y + dE2y + dE3y;
    dSy= (dF1y*_H - 3.0*F1*dHy/2.0) / (_H*_H*sqrt(_H));
    
    /*.................	computation of Sz	...........*/
    
    dE1z= 2.0*fx*fxz*(fyy+fzz) - 2.0*( fyz*fz*fyz + fy*fzz*fyz );
    dE2z= 2.0*fy*fyz*(fxx+fzz) - 2.0*( fxz*fz*fxz + fx*fzz*fxz );
    dE3z= 2.0*fz*fzz*(fxx+fyy) - 2.0*( fxz*fy*fxy + fx*fyz*fxy );
    
    dF1z= dE1z + dE2z + dE3z;
    dSz= (dF1z*_H - 3.0*F1*dHz/2.0) / (_H*_H*sqrt(_H));
    
    /* T_3Copy( &dS, &theDer->dS ); */
    theDer->dS.x =  (reeltype) dSx;
    theDer->dS.y =  (reeltype) dSy;
    theDer->dS.z =  (reeltype) dSz;
    
    /*.................	computation of k1,k2 and derivatives ..........*/
    
    delta2= S*S - 4.0*K;
    if ( delta2 > 0.0 ){
      delta= sqrt(delta2);
      dDelta2x= 2.0*S*dSx - 4.0*dKx;
      dDelta2y= 2.0*S*dSy - 4.0*dKy;
      dDelta2z= 2.0*S*dSz - 4.0*dKz;
      dDeltax= dDelta2x / (2.0*delta);
      dDeltay= dDelta2y / (2.0*delta);
      dDeltaz= dDelta2z / (2.0*delta);
    }
    else{
      delta= 0.0;
      dDeltax= 0.0;
      dDeltay= 0.0;
      dDeltaz= 0.0;
    }
    
    theDer->g=  (reeltype) K;
    theDer->m=  (reeltype) (S/2.0);
    theDer->k1=  (reeltype) ((S + delta) / 2.0);
    theDer->k2=  (reeltype) ((S - delta) / 2.0);
    
    dk1x= (dSx + dDeltax) / 2.0;
    dk1y= (dSy + dDeltay) / 2.0;
    dk1z= (dSz + dDeltaz) / 2.0;
    
    dk2x= (dSx - dDeltax) / 2.0;
    dk2y= (dSy - dDeltay) / 2.0;
    dk2z= (dSz - dDeltaz) / 2.0;
    
    /* T_3Copy( &dk1, &theDer->dk1 );
       T_3Copy( &dk2, &theDer->dk2 ); */
    theDer->dk1.x = (reeltype) dk1x;
    theDer->dk1.y = (reeltype) dk1y;
    theDer->dk1.z = (reeltype) dk1z;
    
    theDer->dk2.x = (reeltype) dk2x;
    theDer->dk2.y = (reeltype) dk2y;
    theDer->dk2.z = (reeltype) dk2z;
  
    /*............... computation of the principal directions ........*/
    
    T_PrincDir( der3, &a, &b );
    
    /*..................................................................*/
    
    
    t1.x= (reeltype) (a.x - delta * b.x / 2.0);
    t1.y= (reeltype) (a.y - delta * b.y / 2.0);
    t1.z= (reeltype) (a.z - delta * b.z / 2.0);
    t2.x= (reeltype) (a.x + delta * b.x / 2.0);
    t2.y= (reeltype) (a.y + delta * b.y / 2.0);
    t2.z= (reeltype) (a.z + delta * b.z / 2.0);
    
    norm1= T_3Norm( &t1 );
    norm2= T_3Norm( &t2 );
    if ( (norm1 < 1e-20) && (norm2 < 1e-20) ){ 
      theDer->good= FALSE; 
      return; 
    }
    
    T_3Copy( &t1, &theDer->t1 );
    T_3Copy( &t2, &theDer->t2 );
}

/*----------------------------------------------------------------------*/
/*
  Used in the computation of surface characteristics

	Compute all the differential characteristics of the surface from
	the differencials of the image up to order 3. Doesn't take into
	acount any orientation problems, but switches the principal values
	according to their absolute values.
*/

#if defined(_ANSI_)
static void _T_GetAllDifCaract( t_3DerivatesPtr der3, t_SurfDerPtr theDer, t_3PointPtr normal, int flags )
#else
static void _T_GetAllDifCaract(der3, theDer, normal,flags)
t_3DerivatesPtr der3;
t_SurfDerPtr theDer;
t_3PointPtr normal;
int flags;

#endif

{
	if ( flags & E_ISO_QUADRIC )	T_GetQuadricCaract(der3,theDer,normal);
	else							_T_GetSurfCaract(der3,theDer,normal);

	/* 1 is principal direct */

	if ( fabs(theDer->k2) > fabs(theDer->k1) ){

		T_Switch( &theDer->k1, &theDer->k2 );
		T_3Switch( &theDer->dk1, &theDer->dk2 );
		T_3Switch( &theDer->t1, &theDer->t2 );
		}
	}

/*----------------------------------------------------------------------*/
/*
  Orientation of the differentials
  
  Orient the principal directions and compute the extremalities,
  according to a given direction for either t1 or t2
*/

#if defined(_ANSI_)
void T_CptExtrem(
t_3DerivatesPtr der3, t_SurfDerPtr theDer, 
t_3PointPtr dir, /* vector to orientate principal directions	*/
int flags )
#else
void T_CptExtrem(der3, theDer, dir, flags )
t_3DerivatesPtr der3;
t_SurfDerPtr theDer;
t_3PointPtr dir;
int flags ;

#endif

{
  t_3Point n,t1,t2;
  reeltype aux,norm1,norm2;
  
  _T_GetAllDifCaract(der3,theDer,&n,flags);
  
  if ( !theDer->good ) return;		/* not able to compute it */
  
  T_3Copy( &theDer->t1, &t1 );
  T_3Copy( &theDer->t2, &t2 );
  
  if ( flags & E_ISO_SECOND ){	   /* align first principal direction */
    
    norm1= T_3Norm( &t1 );
    if (norm1 < EPSY ){
      T_3VMul( &t2, &n, &t1 );
      T_3Normalize( &t1, &t1 );
    }
    else T_3RDiv( &t1,norm1,&t1 );
    aux= T_3SMul( &t1, dir );		/* orient direction t1		*/
    if (aux > 0) T_3RMul( &t1, -1.0, &t1 );
    T_3VMul( &n, &t1, &t2 );	        /* compute direction t2	*/	
    T_3Normalize( &t2, &t2 );		/* normalize direction t2	*/
  }
  
  else{		       	/* align second principal direction */
    
    norm2= T_3Norm( &t2 );
    if (norm2 < EPSY){		  /*	normalize direction t2	*/
      T_3VMul( &n, &t1, &t2 );
      T_3Normalize( &t2, &t2 );
    }
    else T_3RDiv( &t2,norm2,&t2 );
    
    aux= T_3SMul( &t2, dir );	    /* orient direction t2		*/
    if (aux < 0) T_3RMul( &t2, -1.0, &t2 );
    T_3VMul( &t2, &n, &t1 );	    /* compute direction t1	*/	
    T_3Normalize( &t1, &t1 );	    /* normalize direction t1	*/
  }
  
  T_3Copy( &t1, &theDer->t1 );
  T_3Copy( &t2, &theDer->t2 );
  
  theDer->e1= T_3SMul( &theDer->t1, &theDer->dk1 );
  theDer->e2= T_3SMul( &theDer->t2, &theDer->dk2 );
  
  /*.... The sign of the curvature influence the sign of the extremality ..*/
  
  if ( theDer->k1 < 0 ) theDer->e1= -theDer->e1;
  if ( theDer->k2 < 0 ) theDer->e2= -theDer->e2;
}

/*----------------------------------------------------------------------*/
/*
  Compute color codes for extremal values

  Compute a color code corresponding to the 2 extremalities and the
  two cross-extremalities on the surface
*/

#if defined(_ANSI_)
int T_ExtremColorCode( t_3DerivatesPtr der3, int flags )
#else
int T_ExtremColorCode( der3, flags )
t_3DerivatesPtr der3 ;
int flags;

#endif

{
	t_SurfDer theDer;
	reeltype e11,e12,e22,e21;
	int code;
	t_3Point dir;

	T_3Form( 1.0, 1.0, 1.0, &dir );
	T_CptExtrem( der3, &theDer, &dir, flags );

	if ( !theDer.good ) return(-1);		/* not able to compute it */

	e11= T_3SMul( &theDer.t1, &theDer.dk1 );
	e12= T_3SMul( &theDer.t1, &theDer.dk2 );
	e22= T_3SMul( &theDer.t2, &theDer.dk2 );
	e21= T_3SMul( &theDer.t2, &theDer.dk1 );

	code= 0;
	if ( theDer.e1 * theDer.e2 >= 0 ) code |= 1;

	return( code );
	}

	
/*----------------------------------------------------------------------*/
/*
  Get all differential characteristics of the surface

Compute all the values corresponding to the successive derivates
of the 3D surface, including the extremality function.
good = FALSE if there is a degenerated case.
1 is the maximum curvature, and 1,2, and n is direct
*/

#if defined(_ANSI_)
void T_GetAllCurvDer(
t_3DerivatesPtr der3,
t_SurfDerPtr theDer,
t_3PointPtr dir /* vector to orientate principal directions	*/
)
#else
void T_GetAllCurvDer(der3, theDer, dir)
t_3DerivatesPtr der3;
t_SurfDerPtr theDer;
t_3PointPtr dir;

#endif

{
	T_CptExtrem( der3, theDer, dir, 0 );
	}
	
/*----------------------------------------------------------------------*/
/*
  Verify the computation of the curvatures
*/

#if defined(_ANSI_)
int T_CurvVerif( 
t_SurfDerPtr theDer,
t_3PointPtr dir /* vector to orientate principal direction */
)
#else
int T_CurvVerif(theDer, dir )
t_SurfDerPtr theDer;
t_3PointPtr dir ;

#endif

{
	reeltype aux;

	if ( theDer->good == FALSE ){
		T_Warn("unable to compute one princ. dir in T_CurvVerif\n", NULL );
		return( FALSE );
		}

	aux= T_3Deter( &theDer->t1, &theDer->t2, &theDer->grad );
	if ( aux < 0 ){
		T_Warn("indirect orientation in T_CurvVerif\n", NULL );
		return( FALSE );
		}

	aux= T_3SMul( &theDer->t2, dir );
	if ( aux < 0 ){
		T_Warn("wrong orientation in T_CurvVerif\n", NULL );
		return( FALSE );
		}

	if ( fabs( theDer->k1 ) < fabs( theDer->k2 ) ){
		T_Warn("wrong maximal curvature in T_CurvVerif\n", NULL );
		return( FALSE );
		}

	return( TRUE );
	}
	
/*----------------------------------------------------------------------*/
/*		Compute the principal directions								*/

#if defined(_ANSI_)
void T_PrincDir( t_3DerivatesPtr der3, t_3PointPtr a, t_3PointPtr b )
#else
void T_PrincDir(der3, a, b )
t_3DerivatesPtr der3;
t_3PointPtr a;
t_3PointPtr b ;

#endif

{
  static double fx,fy,fz,fxx,fyy,fzz,fxy,fxz,fyz;
  static double fx2,fy2,fz2,fx3,fy3;
  static double _H,delta;
  
  fx= der3->fx;	fx2= fx*fx; fx3= fx*fx2;
  fy= der3->fy;	fy2= fy*fy; fy3= fy*fy2;
  fz= der3->fz;	fz2= fz*fz;
  fxx= der3->fxx;
  fyy= der3->fyy;
  fzz= der3->fzz;
  fxy= der3->fxy;
  fxz= der3->fxz;
  fyz= der3->fyz;
  
  _H= fx*fx+fy*fy+fz*fz;
  delta= _H * sqrt(_H);
  
  if ( delta < 1e-15 ){
    T_3Form( 0.0, 0.0, 0.0, a );
    T_3Form( 0.0, 0.0, 0.0, b );
    return;
  }
  
  a->x=(reeltype) 
    (0.5*(-2.0*fy2*fxy*fx + 2.0*fz*fx2*fyz - fx2*fy*fzz - 2.0*fz2*fxy*fx
	    +  fx2*fy*fyy +   fy3*fxx    + fz2*fxx*fy + 2.0*fz*fy2*fyz 
	    -  fz2*fyy*fy -   fy3*fzz) / delta);
  
  a->y=(reeltype) 
    (0.5*(-  fx*fy2*fxx - 2.0*fz*fx2*fxz + fx*fz2*fxx +   fx3*fzz 
	    -  fx3*fyy 	-   fx*fz2*fyy + fx*fy2*fzz + 2.0*fy*fx2*fxy
	    -2.0*fz*fy2*fxz + 2.0*fy*fz2*fxy) / delta);
  
  a->z=(reeltype)
    (-  ( (fy2+fx2)*(fx*fyz-fy*fxz) + fx*fy*fz*(fxx-fyy) 
	     + (fy2-fx2)*fz*fxy ) / delta);
  
  b->x= (reeltype) -fy;
  b->y= (reeltype) fx;
  b->z= (reeltype) 0.0;
}

/*----------------------------------------------------------------------*/
/*		Compute the principal directions								*/

#if defined(_ANSI_)
void T_PrincDir2( t_3DerivatesPtr der3, t_3PointPtr a, t_3PointPtr b )
#else
void T_PrincDir2(der3, a, b )
t_3DerivatesPtr der3;
t_3PointPtr a;
t_3PointPtr b ;

#endif

{
  static double fx,fy,fz,fxx,fyy,fzz,fxy,fxz,fyz;
  static double fx2,fy2,fz2,fx3,fy3,fz3;
  static double _H,delta;
  
  fx= der3->fx;	fx2= fx*fx;		fx3= fx2*fx;
  fy= der3->fy;	fy2= fy*fy;		fy3= fy2*fy;
  fz= der3->fz;	fz2= fz*fz;		fz3= fz2*fz;
  fxx= der3->fxx;
  fyy= der3->fyy;
  fzz= der3->fzz;
  fxy= der3->fxy;
  fxz= der3->fxz;
  fyz= der3->fyz;
  
  _H= fx*fx+fy*fy+fz*fz;
  delta= _H * sqrt(_H);
  
  if ( delta < EPSY ){
    T_3Form( 0.0, 0.0, 0.0, a );
    T_3Form( 0.0, 0.0, 0.0, b );
    return;
  }
  
  a->x= (reeltype) (
    -0.5*(-2*fz3*fxy       +fy3*fzz        +2*fy3*fxz	 -2*fy2*fz*fxy	
	  +2*fz2*fx*fyz	   +2*fz2*fy*fxz   -2*fy2*fx*fyz -2*fz*fx*fy*fzz
	  +2*fx*fy*fz*fyy  +fy2*fz*fxx     -2*fz2*fx*fxz +fz*fx2*fzz
	  -fx2*fz*fyy      +2*fz2*fy*fyz   -fz*fy2*fzz	 +fz3*fxx
	  -fz3*fyy	   -2*fy2*fx*fxz   +2*fx2*fy*fyz -fy3*fxx
	  +2*fx*fz2*fxy    -fy*fz2*fxx     -2*fz*fy2*fyz +fy*fz2*fyy
	  -2*fz*fx2*fyz    +2*fx*fy2*fxy   +fx2*fy*fzz   -fx2*fy*fyy )/delta);
  
  a->y= (reeltype) (
       0.5*(  -2*fz3*fxy     +2*fz2*fx*fyz   +2*fz2*fy*fxz  -2*fz*fx*fy*fzz
	      -fy2*fz*fxx    +2*fz2*fx*fxz   -fz*fx2*fzz    +fx2*fz*fyy
	      -2*fz2*fy*fyz  +fz*fy2*fzz     -fz3*fxx       +fz3*fyy
	      +2*fy2*fx*fxz  -2*fx2*fy*fyz   +2*fx3*fyz     -2*fx2*fz*fxy
	      -2*fx2*fy*fxz  +2*fx*fy*fz*fxx -fx3*fyy       +2*fy*fz2*fxy
	      +fx*fz2*fxx    +fx3*fzz        -fx*fz2*fyy    -2*fz*fy2*fxz
	      -2*fz*fx2*fxz  -fx*fy2*fxx     +fx*fy2*fzz    +2*fy*fx2*fxy)/
       delta );
  
  a->z= (reeltype) (
       0.5*(  -fy3*fzz         +2*fy3*fxz     -2*fy2*fz*fxy  -2*fy2*fx*fyz
	      +2*fx*fy*fz*fyy  -2*fx3*fyz     +2*fx2*fz*fxy  +2*fx2*fy*fxz
	      -2*fx*fy*fz*fxx  -fx3*fyy       +2*fy*fz2*fxy  +fx*fz2*fxx
	      +fx3*fzz         -fx*fz2*fyy    -2*fz*fy2*fxz  -2*fz*fx2*fxz
	      -fx*fy2*fxx      +fx*fy2*fzz    +2*fy*fx2*fxy  +fy3*fxx
	      -2*fx*fz2*fxy    +fy*fz2*fxx    +2*fz*fy2*fyz  -fy*fz2*fyy
	      +2*fz*fx2*fyz    -2*fx*fy2*fxy  -fx2*fy*fzz    +fx2*fy*fyy)/
       delta);
  
  b->x= (reeltype) (fz - fy);
  b->y= (reeltype) (fx - fz);
  b->z= (reeltype) (fy - fx);
}

/*----------------------------------------------------------------------*/
/*		Compute the principal directions								*/

#if defined(_ANSI_)
void T_PrincDir3( t_3DerivatesPtr der3, t_3PointPtr t1, t_3PointPtr t2 )
#else
void T_PrincDir3(der3, t1, t2 )
t_3DerivatesPtr der3;
t_3PointPtr t1;
t_3PointPtr t2 ;

#endif

{
  static double fx,fy,fz,fxx,fyy,fzz,fxy,fxz,fyz;
  static double fx2,fy2,fz2,fz3;
  static double _H,delta;
  static double A,B,C,_D,u,v,s;
  static double e,f,g,l,m,n;
  static double k;
  static double a1x, a1y, a1z, b1x, b1y, b1z;
  static double a2x, a2y, a2z, b2x, b2y, b2z;
  
  fx= der3->fx;	fx2= fx*fx;		/* fx3= fx2*fx; */
  fy= der3->fy;	fy2= fy*fy;		/* fy3= fy2*fy; */
  fz= der3->fz;	fz2= fz*fz;		fz3= fz2*fz;
  fxx= der3->fxx;
  fyy= der3->fyy;
  fzz= der3->fzz;
  fxy= der3->fxy;
  fxz= der3->fxz;
  fyz= der3->fyz;
  
  if ( fabs(fz) < EPSY ){
    T_3Form( 0.0, 0.0, 0.0, t1 );
    T_3Form( 0.0, 0.0, 0.0, t2 );
    return;
  }
  
  _H= (fx2+fy2+fz2)/ fz2;
  
  e= (fx2 + fz2)/fz2;
  f= fx*fy/fz2;
  g= (fy2 + fz2)/fz2;
  l= (2.0*fx*fz*fxz - fx2*fzz - fz2*fxx) / (sqrt(_H)*fz3);
  m= (fx*fz*fyz + fy*fz*fxz - fx*fy*fzz - fz2*fxy) / (sqrt(_H)*fz3);
  n= (2.0*fy*fz*fyz - fy2*fzz - fz2*fyy) / (sqrt(_H)*fz3);
  
  A= (g*l - f*m)/_H;
  B= (g*m - f*n)/_H;
  C= (e*m - f*l)/_H;
  _D= (e*n - f*m)/_H;
  
  s= (A+_D)/2;
  k= A*_D - B*C;
  
  delta= sqrt(s*s - k);
  
  u= B;
  v= (s - A);	
  
  a1x= u;
  a1y= v;
  a1z= -fx*u/fz -fy*v/fz;
  
  b1x= 0.0;
  b1y= 1.0; 
  b1z= -fy/fz;
  
  u= (s-_D);
  v= C;	
  
  a2x= u;
  a2y= v;
  a2z= -fx*u/fz -fy*v/fz;
  
  b2x= 1.0;
  b2y= 0.0; 
  b2z= -fx/fz;
  
  t1->x= (reeltype) (a1x - delta * b1x);
  t1->y= (reeltype) (a1y - delta * b1y);
  t1->z= (reeltype) (a1z - delta * b1z);
  t2->x= (reeltype) (a2x + delta * b2x);
  t2->y= (reeltype) (a2y + delta * b2y);
  t2->z= (reeltype) (a2z + delta * b2z);
}

/*----------------------------------------------------------------------*/
/*
  Compute the value of the extremality coefficient

  with a priviledged direction for orientation
*/

#if defined(_ANSI_)
reeltype T_GetExtremCurv( int x, int y, int z, t_ImagePtr theIm, t_3rdFiltersPtr filter)
#else
reeltype T_GetExtremCurv(x, y, z, theIm, filter)
int x;
int y;
int z;
t_ImagePtr theIm;
t_3rdFiltersPtr filter;

#endif

{
	t_SurfDer theDer;
	t_3Derivates der3;
	t_3Point dir;

	T_3Form( 1.0, 1.0, 1.0, &dir );
	T_3DerivatesComp(x,y,z,theIm,filter,&der3 );
	T_GetAllCurvDer(&der3,&theDer,&dir);
/*	
	return( sqrt( theDer.m * theDer.m - theDer.g/2 ) );
*/
	
	if (theDer.good == FALSE) return(-1.0);

	if ( fabs(theDer.k1) > fabs(theDer.k2) )	return( theDer.e1 );
	else										return( theDer.e2 );
	}

/*----------------------------------------------------------------------*/
/*
	Compute the 3rd derivates of a point
*/

#if defined(_ANSI_)
void T_3DerivatesComp( int x, int y, int z, t_ImagePtr theIm, t_3rdFiltersPtr filter, t_3DerivatesPtr der3 )
#else
void T_3DerivatesComp(x, y, z, theIm, filter, der3 )
int x;
int y;
int z;
t_ImagePtr theIm;
t_3rdFiltersPtr filter;
t_3DerivatesPtr der3 ;

#endif

{
    T_LoadFilter(x,y,z,theIm,filter);

	der3->f= T_ApplyFilters(filter,0,0,0);
    der3->fx= T_ApplyFilters(filter,1,0,0);
    der3->fy= T_ApplyFilters(filter,0,1,0);
    der3->fz= T_ApplyFilters(filter,0,0,1);

    der3->fxy= T_ApplyFilters(filter,1,1,0);
    der3->fxz= T_ApplyFilters(filter,1,0,1);
    der3->fyz= T_ApplyFilters(filter,0,1,1);

    der3->fxx= T_ApplyFilters(filter,2,0,0);
    der3->fyy= T_ApplyFilters(filter,0,2,0);
    der3->fzz= T_ApplyFilters(filter,0,0,2);

    der3->fxxx= T_ApplyFilters(filter,3,0,0);
    der3->fyyy= T_ApplyFilters(filter,0,3,0);
    der3->fzzz= T_ApplyFilters(filter,0,0,3);

    der3->fxxy= T_ApplyFilters(filter,2,1,0);
    der3->fxxz= T_ApplyFilters(filter,2,0,1);
    der3->fxyy= T_ApplyFilters(filter,1,2,0);
    der3->fxyz= T_ApplyFilters(filter,1,1,1);
    der3->fyyz= T_ApplyFilters(filter,0,2,1);
    der3->fyzz= T_ApplyFilters(filter,0,1,2);
    der3->fxzz= T_ApplyFilters(filter,1,0,2);
	}

/*----------------------------------------------------------------------*/
/*
	Compute only the 2nd derivates of a point
*/

#if defined(_ANSI_)
void T_2DerivatesComp( int x, int y, int z, t_ImagePtr theIm, t_3rdFiltersPtr filter, t_3DerivatesPtr der3 )
#else
void T_2DerivatesComp(x, y, z, theIm, filter, der3 )
int x;
int y;
int z;
t_ImagePtr theIm;
t_3rdFiltersPtr filter;
t_3DerivatesPtr der3 ;

#endif

{
    T_LoadFilter(x,y,z,theIm,filter);

	der3->f= T_ApplyFilters(filter,0,0,0);
    der3->fx= T_ApplyFilters(filter,1,0,0);
    der3->fy= T_ApplyFilters(filter,0,1,0);
    der3->fz= T_ApplyFilters(filter,0,0,1);

    der3->fxy= T_ApplyFilters(filter,1,1,0);
    der3->fxz= T_ApplyFilters(filter,1,0,1);
    der3->fyz= T_ApplyFilters(filter,0,1,1);

    der3->fxx= T_ApplyFilters(filter,2,0,0);
    der3->fyy= T_ApplyFilters(filter,0,2,0);
    der3->fzz= T_ApplyFilters(filter,0,0,2);

    der3->fxxx= 0.0;
    der3->fyyy= 0.0;
    der3->fzzz= 0.0;

    der3->fxxy= 0.0;
    der3->fxxz= 0.0;
    der3->fxyy= 0.0;
    der3->fxyz= 0.0;
    der3->fyyz= 0.0;
    der3->fyzz= 0.0;
    der3->fxzz= 0.0;
	}

/*----------------------------------------------------------------------*/
/*		
	Compute the 3rd derivates of a point,

	using only the 2D filtering functions.
*/

#if defined(_ANSI_)
void T_3Derivates2DComp( int x, int y, int z, t_ImagePtr theIm, t_3rdFiltersPtr filter, t_3DerivatesPtr der3 )
#else
void T_3Derivates2DComp(x, y, z, theIm, filter, der3 )
int x;
int y;
int z;
t_ImagePtr theIm;
t_3rdFiltersPtr filter;
t_3DerivatesPtr der3 ;

#endif

{
    T_LoadFilter(x,y,z,theIm,filter);

	der3->f= T_ApplyFilters(filter,0,0,0);
    der3->fx= T_ApplyFilters(filter,1,0,0);
    der3->fy= T_ApplyFilters(filter,0,1,0);

    der3->fxy= T_ApplyFilters(filter,1,1,0);
    der3->fxx= T_ApplyFilters(filter,2,0,0);
    der3->fyy= T_ApplyFilters(filter,0,2,0);

    der3->fxxx= T_ApplyFilters(filter,3,0,0);
    der3->fyyy= T_ApplyFilters(filter,0,3,0);
    der3->fxxy= T_ApplyFilters(filter,2,1,0);
    der3->fxyy= T_ApplyFilters(filter,1,2,0);
	}

/*----------------------------------------------------------------------*/
/*
  Compute the 3rd derivatives of a test function (obsolete)

<pre>
2 * r6 - 15 * r4 + 24 * r2	with r2 = (X2 + Y2)
X= 1.5 when x=1 and X= -1.5 when x= 0, etc...
</pre>
*/


#ifdef COMPILE_UNUSED

#if defined(_ANSI_)
static void _T_3DerTest( reeltype x, reeltype y, t_3DerivatesPtr der3 )
#else
static void _T_3DerTest(x, y, der3 )
reeltype x;
reeltype y;
t_3DerivatesPtr der3 ;

#endif

{
	reeltype X,Y,R2;

	X= x * 3 - 1.5;
	Y= y * 3 - 1.5;
	R2= X*X+Y*Y;
	
	der3->fx= 12 * R2*R2*X - 60*R2*X + 48*X;
	der3->fy= 12 * R2*R2*Y - 60*R2*Y + 48*Y;
	der3->fz= -10;
	
    der3->fxy= 48*R2*X*Y - 120*X*Y;
    der3->fxz= 0;
    der3->fyz= 0;

    der3->fxx= 48*R2*X*X + 12*R2*R2-180*X*X -60*Y*Y + 48;
    der3->fyy= 48*R2*Y*Y + 12*R2*R2-180*Y*Y -60*X*X + 48;
    der3->fzz= 0;

    der3->fxxx= 96*X*X*X + 144*R2*X - 360*X; 
    der3->fyyy= 96*Y*Y*Y + 144*R2*Y - 360*Y; 
    der3->fzzz= 0;

    der3->fxxy= 96*Y*X*X + 48 * R2*Y - 120*Y;
    der3->fxxz= 0;
    der3->fxyy= 96*X*Y*Y + 48 * R2*X - 120*X;
    der3->fxyz= 0;
    der3->fyyz= 0;
    der3->fyzz= 0;
    der3->fxzz= 0;
	}

/*
  Compute the 3rd derivatives of a test function (obsolete)
  */

#if defined(_ANSI_)
static void _T_3DerTest2( reeltype x, reeltype y, t_3DerivatesPtr der3 )
#else
static void _T_3DerTest2(x, y, der3 )
reeltype x;
reeltype y;
t_3DerivatesPtr der3 ;

#endif

{
	reeltype X,Y,R2;

	X= x * 3 - 1.5;
	Y= y * 3 - 1.5;
	R2= X*X+Y*Y;
	
	der3->fx= 12 * R2*R2*X - 60*R2*X + 48*X;
	der3->fy= -10;
	der3->fz= 12 * R2*R2*Y - 60*R2*Y + 48*Y;
	
    der3->fxy= 0;
    der3->fxz= 48*R2*X*Y - 120*X*Y;
    der3->fyz= 0;

    der3->fxx= 48*R2*X*X + 12*R2*R2-180*X*X -60*Y*Y + 48;
    der3->fyy= 0;
    der3->fzz= 48*R2*Y*Y + 12*R2*R2-180*Y*Y -60*X*X + 48;

    der3->fxxx= 96*X*X*X + 144*R2*X - 360*X; 
    der3->fyyy= 0;
    der3->fzzz= 96*Y*Y*Y + 144*R2*Y - 360*Y; 

    der3->fxxy= 0;
    der3->fxxz= 96*Y*X*X + 48 * R2*Y - 120*Y;
    der3->fxyy= 0;
    der3->fxyz= 0;
    der3->fyyz= 0;
    der3->fyzz= 0;
    der3->fxzz= 96*X*Y*Y + 48 * R2*X - 120*X;
	}

/*
  Compute the 3rd derivatives of a test function (obsolete)
  */

#if defined(_ANSI_)
static void _T_3DerTest3( reeltype x, reeltype y, t_3DerivatesPtr der3 )
#else
static void _T_3DerTest3(x, y, der3 )
reeltype x;
reeltype y;
t_3DerivatesPtr der3 ;

#endif
							/* x4 - 1 */
{
	reeltype Y;

	Y= y * 3 - 1.5;
	
	der3->fx= 0;
	der3->fy= 4*Y*Y*Y;
	der3->fz= -1;
	
    der3->fxy= 0;
    der3->fxz= 0;
    der3->fyz= 0;

    der3->fxx= 0;
    der3->fyy= 12*Y*Y;
    der3->fzz= 0;

    der3->fxxx= 0; 
    der3->fyyy= 24*Y;
    der3->fzzz= 0; 

    der3->fxxy= 0;
    der3->fxxz= 0;
    der3->fxyy= 0;
    der3->fxyz= 0;
    der3->fyyz= 0;
    der3->fyzz= 0;
    der3->fxzz= 0;
	}
/*
  Compute the 3rd derivatives of a test function (obsolete)
  */

#if defined(_ANSI_)
static void _T_3DerTest4( reeltype x, reeltype y, t_3DerivatesPtr der3 )
#else
static void _T_3DerTest4(x, y, der3 )
reeltype x;
reeltype y;
t_3DerivatesPtr der3 ;

#endif
							/* x4 + y4 */
{
	reeltype X,Y;

	X= x * 3 - 1.5;
	Y= y * 3 - 1.5;
	
	der3->fx= 4*X*X*X;
	der3->fy= 4*Y*Y*Y;
	der3->fz= -1;
	
    der3->fxy= 0;
    der3->fxz= 0;
    der3->fyz= 0;

    der3->fxx= 12*X*X;
    der3->fyy= 12*Y*Y;
    der3->fzz= 0;

    der3->fxxx= 24*X; 
    der3->fyyy= 24*Y;
    der3->fzzz= 0; 

    der3->fxxy= 0;
    der3->fxxz= 0;
    der3->fxyy= 0;
    der3->fxyz= 0;
    der3->fyyz= 0;
    der3->fyzz= 0;
    der3->fxzz= 0;
	}

/*
  Compute the 3rd derivatives of a test function (obsolete)
  */

#if defined(_ANSI_)
static void _T_3DerTest5( reeltype x, reeltype y, t_3DerivatesPtr der3 )
#else
static void _T_3DerTest5(x, y, der3 )
reeltype x;
reeltype y;
t_3DerivatesPtr der3 ;

#endif
							/* 2*x2 + y2 */
{
	reeltype X,Y;

	X= x * 3 - 1.5;
	Y= y * 3 - 1.5;
	
	der3->fx= 4*X;
	der3->fy= 2*Y;
	der3->fz= -1;
	
    der3->fxy= 0;
    der3->fxz= 0;
    der3->fyz= 0;

    der3->fxx= 4;
    der3->fyy= 2;
    der3->fzz= 0;

    der3->fxxx= 0; 
    der3->fyyy= 0;
    der3->fzzz= 0; 

    der3->fxxy= 0;
    der3->fxxz= 0;
    der3->fxyy= 0;
    der3->fxyz= 0;
    der3->fyyz= 0;
    der3->fyzz= 0;
    der3->fxzz= 0;
	}

/*
  Compute the 3rd derivatives of a test function (obsolete)
  */

#if defined(_ANSI_)
static void _T_3DerTest6( reeltype x, reeltype y, t_3DerivatesPtr der3 )
#else
static void _T_3DerTest6(x, y, der3 )
reeltype x;
reeltype y;
t_3DerivatesPtr der3 ;

#endif
						/* x4 + y4  rotation of pi/4 */
{
	reeltype X,Y;

	X= x * 3.0 - 1.5;
	Y= y * 3.0 - 1.5;
	
	der3->fx= 2*X*X*X + 6*X*Y*Y;
	der3->fy= 2*Y*Y*Y + 6*X*X*Y;
	der3->fz= -1;
	
    der3->fxy= 12*X*Y;
    der3->fxz= 0;
    der3->fyz= 0;

    der3->fxx= 6*X*X + 6*Y*Y;
    der3->fyy= 6*X*X + 6*Y*Y;
    der3->fzz= 0;

    der3->fxxx= 12*X; 
    der3->fyyy= 12*Y;
    der3->fzzz= 0; 

    der3->fxxy= 12*Y;
    der3->fxxz= 0;
    der3->fxyy= 12*X;
    der3->fxyz= 0;
    der3->fyyz= 0;
    der3->fyzz= 0;
    der3->fxzz= 0;
	}


/*----------------------------------------------------------------------*/
/*		Compute the Image of the extremality of the test function		*/

#if defined(_ANSI_)
static void _T_3ExtremTest( t_ImagePtr theIm )
#else
static void _T_3ExtremTest(theIm )
t_ImagePtr theIm ;

#endif

{
	int i,j;
	t_SurfDer theDer;
	t_3Derivates der3;
	reeltype value,x,y;
	t_3Point dir;

	T_3Form( 0.0, 0.0, 0.0, &dir );

	for( j=0; j< theIm->ty; j++ ){

		printf("Line %d\n", j );
		for( i=0; i< theIm->tx; i++ ){

			x= ((reeltype) i) / theIm->tx;
			y= ((reeltype) j) / theIm->ty;
			_T_3DerTest(x,y,&der3 );
			T_GetAllCurvDer(&der3,&theDer,&dir);

			T_Set3Point( i,j,0,theIm,theDer.t1.x,MODE_SET );
			T_Set3Point( i,j,1,theIm,theDer.t1.y,MODE_SET );
			T_Set3Point( i,j,2,theIm,theDer.t1.z,MODE_SET );
			T_Set3Point( i,j,3,theIm,theDer.t2.x,MODE_SET );
			T_Set3Point( i,j,4,theIm,theDer.t2.y,MODE_SET );
			T_Set3Point( i,j,5,theIm,theDer.t2.z,MODE_SET );
			if (fabs(theDer.k1) > fabs(theDer.k2))	value= fabs(theDer.k1);
			else									value= fabs(theDer.k2);
			T_Set3Point( i,j,6,theIm,value,MODE_SET );

			if (theDer.good == FALSE) value= -1.0;
			else if ( fabs(theDer.k1) > fabs(theDer.k2) )	value= theDer.e1;
			else 											value= theDer.e2;
			T_Set3Point( i,j,7,theIm,value,MODE_SET );

			T_Set3Point( i,j,8,theIm,theDer.k1,MODE_SET );
			T_Set3Point( i,j,9,theIm,theDer.dk1.x,MODE_SET );
			T_Set3Point( i,j,10,theIm,theDer.dk1.y,MODE_SET );
			T_Set3Point( i,j,11,theIm,theDer.dk1.z,MODE_SET );
			T_Set3Point( i,j,12,theIm,theDer.k2,MODE_SET );
			T_Set3Point( i,j,13,theIm,theDer.dk2.x,MODE_SET );
			T_Set3Point( i,j,14,theIm,theDer.dk2.y,MODE_SET );
			T_Set3Point( i,j,15,theIm,theDer.dk2.z,MODE_SET );
			T_Set3Point( i,j,16,theIm,theDer.g,MODE_SET );
			T_Set3Point( i,j,17,theIm,theDer.dK.x,MODE_SET );
			T_Set3Point( i,j,18,theIm,theDer.dK.y,MODE_SET );
			T_Set3Point( i,j,19,theIm,theDer.dK.z,MODE_SET );
			T_Set3Point( i,j,20,theIm,theDer.m,MODE_SET );
			T_Set3Point( i,j,21,theIm,theDer.dS.x,MODE_SET );
			T_Set3Point( i,j,22,theIm,theDer.dS.y,MODE_SET );
			T_Set3Point( i,j,23,theIm,theDer.dS.z,MODE_SET );
			T_Set3Point( i,j,24,theIm,theDer.e1,MODE_SET );
			T_Set3Point( i,j,25,theIm,theDer.e2,MODE_SET );
			}
		}
	}


#endif /* COMPILE_UNUSED */
/*----------------------------------------------------------------------*/
/************************************************************************/
/*
  FiniteExtr
  
  Compute extremality functions with finite differences (obsolete)
  (c) Jean-Philippe THIRION -- INRIA -- 2/19/92
*/
/************************************************************************/

/*----------------------------------------------------------------------*/
/*		Get the extremality function using finite differences			*/

#if defined(_ANSI_)
reeltype T_GetFiniteCrest( int x, int y, int z, t_ImagePtr theIm, t_3rdFiltersPtr filter)
#else
reeltype T_GetFiniteCrest(x, y, z, theIm, filter)
int x;
int y;
int z;
t_ImagePtr theIm;
t_3rdFiltersPtr filter;

#endif

{
    t_SurfDer theDer;
    t_3Point dir;

    T_3Form( 1.0, 1.0, 1.0, &dir );
    T_FiniteCurv(x,y,z,theIm,&theDer,&dir);

    if (theDer.good == FALSE) return(-1.0);
    if ( fabs(theDer.k1) > fabs(theDer.k2) )    return( theDer.e1 );
    else                                        return( theDer.e2 );
    }

/*----------------------------------------------------------------------*/
/*
  Compute all the differential characteristics (obsolete)

  of iso-surfaces using a finite difference method.
*/

#if defined(_ANSI_)
void T_FiniteCurv( int x, int y, int z, t_ImagePtr theIm, t_SurfDerPtr theDer, t_3PointPtr dir )
#else
void T_FiniteCurv(x, y, z, theIm, theDer, dir )
int x;
int y;
int z;
t_ImagePtr theIm;
t_SurfDerPtr theDer;
t_3PointPtr dir ;

#endif

{
	static t_SurfDer pnn,mnn,npn,nmn,nnp,nnm;
	static t_3Derivates aDer;
	
	T_FiniteDer2( x, y, z, theIm, &aDer );
	T_CurvDer2( &aDer, theDer );
	T_PrincDer2( &aDer, theDer, dir );
	if ( theDer->good == FALSE ) return;
	
	T_FiniteDer2( x+1, y, z, theIm, &aDer );
	T_CurvDer2( &aDer, &pnn );
	if ( pnn.good == FALSE ) { theDer->good= FALSE; return; }
	
	T_FiniteDer2( x-1, y, z, theIm, &aDer );
	T_CurvDer2( &aDer, &mnn );
	if ( mnn.good == FALSE ) { theDer->good= FALSE; return; }

	T_FiniteDer2( x, y+1, z, theIm, &aDer );
	T_CurvDer2( &aDer, &npn );
	if ( npn.good == FALSE ) { theDer->good= FALSE; return; }

	T_FiniteDer2( x, y-1, z, theIm, &aDer );
	T_CurvDer2( &aDer, &nmn );
	if ( nmn.good == FALSE ) { theDer->good= FALSE; return; }

	T_FiniteDer2( x, y, z+1, theIm, &aDer );
	T_CurvDer2( &aDer, &nnp );
	if ( nnp.good == FALSE ) { theDer->good= FALSE; return; }

	T_FiniteDer2( x, y, z-1, theIm, &aDer );
	T_CurvDer2( &aDer, &nnm );
	if ( nnm.good == FALSE ) { theDer->good= FALSE; return; }

	theDer->dk1.x= (pnn.k1 - mnn.k1) / 2;
	theDer->dk1.y= (npn.k1 - nmn.k1) / 2;
	theDer->dk1.z= (nnp.k1 - nnm.k1) / 2;

	theDer->dk2.x= (pnn.k2 - mnn.k2) / 2;
	theDer->dk2.y= (npn.k2 - nmn.k2) / 2;
	theDer->dk2.z= (nnp.k2 - nnm.k2) / 2;

	theDer->e1= T_3SMul( &theDer->t1, &theDer->dk1 );
	theDer->e2= T_3SMul( &theDer->t2, &theDer->dk2 );
	}

/*----------------------------------------------------------------------*/
/*
  Compute the derivatives of the function (obsolete)

  using finite differences until order 2.
*/

#if defined(_ANSI_)
void T_FiniteDer2( int x, int y, int z, t_ImagePtr theIm, t_3DerivatesPtr der )
#else
void T_FiniteDer2(x, y, z, theIm, der )
int x;
int y;
int z;
t_ImagePtr theIm;
t_3DerivatesPtr der ;

#endif

{
	reeltype nnn;
	reeltype pnn,mnn,npn,nmn,nnp,nnm;
	reeltype qnn,lnn,nqn,nln,nnq,nnl;
	reeltype ppn,mmn,pmn,mpn;
	reeltype pnp,mnm,pnm,mnp;
	reeltype npp,nmm,npm,nmp;

	nnn= T_Get3Point(x,y,z,theIm);

	pnn= T_Get3Point(x+1,y,z,theIm);
	mnn= T_Get3Point(x-1,y,z,theIm);
	npn= T_Get3Point(x,y+1,z,theIm);
	nmn= T_Get3Point(x,y-1,z,theIm);
	nnp= T_Get3Point(x,y,z+1,theIm);
	nnm= T_Get3Point(x,y,z-1,theIm);

	qnn= T_Get3Point(x+2,y,z,theIm);
	lnn= T_Get3Point(x-2,y,z,theIm);
	nqn= T_Get3Point(x,y+2,z,theIm);
	nln= T_Get3Point(x,y-2,z,theIm);
	nnq= T_Get3Point(x,y,z+2,theIm);
	nnl= T_Get3Point(x,y,z-2,theIm);

	ppn= T_Get3Point(x+1,y+1,z,theIm);
	mmn= T_Get3Point(x-1,y-1,z,theIm);
	pmn= T_Get3Point(x+1,y-1,z,theIm);
	mpn= T_Get3Point(x-1,y+1,z,theIm);

	pnp= T_Get3Point(x+1,y,z+1,theIm);
	mnm= T_Get3Point(x-1,y,z-1,theIm);
	pnm= T_Get3Point(x+1,y,z-1,theIm);
	mnp= T_Get3Point(x-1,y,z+1,theIm);

	npp= T_Get3Point(x,y+1,z+1,theIm);
	nmm= T_Get3Point(x,y-1,z-1,theIm);
	npm= T_Get3Point(x,y+1,z-1,theIm);
	nmp= T_Get3Point(x,y-1,z+1,theIm);

	der->f= (reeltype) nnn;

	der->fx= (reeltype) ((pnn - mnn) / 2.0);
	der->fy= (reeltype) ((npn - nmn) / 2.0);
	der->fz= (reeltype) ((nnp - nnm) / 2.0);

	der->fxx= (reeltype) ((qnn + lnn - 2*nnn) / 4.0); 
	der->fyy= (reeltype) ((nqn + nln - 2*nnn) / 4.0); 
	der->fzz= (reeltype) ((nnq + nnl - 2*nnn) / 4.0); 

	der->fxy= (reeltype) ((ppn + mmn - pmn - mpn) / 4.0);
	der->fyz= (reeltype) ((npp + nmm - npm - nmp) / 4.0);
	der->fxz= (reeltype) ((pnp + mnm - pnm - mnp) / 4.0);
	}

/*----------------------------------------------------------------------*/
/*
  Compute the first and second differentials (obsolete)

  of the iso-surface from the derivatives up to two
*/

#if defined(_ANSI_)
void T_CurvDer2( t_3DerivatesPtr der3, t_SurfDerPtr theDer )
#else
void T_CurvDer2(der3, theDer )
t_3DerivatesPtr der3;
t_SurfDerPtr theDer ;

#endif

{
    static double fx,fy,fz,fxx,fxy,fxz,fyy,fyz,fzz;
	
    static double _H,S,K;
    static double A1,A2,A3;
    static double B1,B2,B3;
	static double D1;
	static double E1,E2,E3;
	static double F1;
	static double delta,delta2;

	theDer->g= 0.0;
	theDer->m= 0.0;
	theDer->k1= 0.0;
	theDer->k2= 0.0;
	theDer->norm= 0.0;
	T_3Form(0.0,0.0,0.0,&theDer->grad );
	theDer->good= TRUE;

    fx= der3->fx;
    fy= der3->fy;
    fz= der3->fz;

    fxy= der3->fxy;
    fxz= der3->fxz;
    fyz= der3->fyz;

    fxx= der3->fxx;
    fyy= der3->fyy;
    fzz= der3->fzz;

    _H= fx*fx+fy*fy+fz*fz;

    if ( _H < EPSY ){ theDer->good= FALSE; return; }

    T_3Form((reeltype) fx,(reeltype) fy,(reeltype) fz,&theDer->grad );
    theDer->norm= (reeltype) sqrt(_H);
    
    A1= fyy*fzz - fyz*fyz;		B1= fxz*fxy - fxx*fyz;	
    A2= fxx*fzz - fxz*fxz;		B2= fyz*fxy - fyy*fxz;	
    A3= fxx*fyy - fxy*fxy;		B3= fxz*fyz - fzz*fxy;	
    
    D1= fx*fx*A1 + 2*fy*fz*B1 +
      fy*fy*A2 + 2*fx*fz*B2 +
      fz*fz*A3 + 2*fx*fy*B3;
    
    K= D1 / (_H*_H);
    
    /*.................	computation of S	...........*/
    
    E1= fx*fx*(fyy+fzz) - 2*fy*fz*fyz;	
    E2= fy*fy*(fxx+fzz) - 2*fx*fz*fxz;	
    E3= fz*fz*(fxx+fyy) - 2*fx*fy*fxy;	
    F1= E1+E2+E3;
    
    S= F1 / (_H*sqrt(_H));
    
    /*.................	computation of k1,k2 and derivatives ..........*/
    
    delta2= S*S - 4*K;
    if ( delta2 > 0 ) delta= sqrt(delta2);
    else delta= 0.0;
    
    theDer->g= (reeltype) K;
    theDer->m= (reeltype) (S/2.0);
    theDer->k1= (reeltype) ((S + delta) / 2.0);
    theDer->k2= (reeltype) ((S - delta) / 2.0);
}

/*----------------------------------------------------------------------*/
/*
  computation of the principal directions (obsolete)
*/

#if defined(_ANSI_)
void T_PrincDer2( t_3DerivatesPtr der3, t_SurfDerPtr theDer, t_3PointPtr dir )
#else
void T_PrincDer2(der3, theDer, dir )
t_3DerivatesPtr der3;
t_SurfDerPtr theDer;
t_3PointPtr dir ;

#endif

{
  static t_3Point a,b,t1,t2,n;
  static reeltype delta,delta2;
  static reeltype norm1,norm2;
  static reeltype aux;
  
  T_3Form(0.0,0.0,0.0,&theDer->t1 );
  T_3Form(0.0,0.0,0.0,&theDer->t2 );
  
  T_PrincDir( der3, &a, &b );
  
  delta2= theDer->m*theDer->m - theDer->g;
  if ( delta2 > 0 ) delta= (reeltype) sqrt(delta2);
  else delta= 0.0;
  
  T_3Normalize( &theDer->grad, &n );
  
  t1.x= a.x - delta * b.x;
  t1.y= a.y - delta * b.y;
  t1.z= a.z - delta * b.z;
  t2.x= a.x + delta * b.x;
  t2.y= a.y + delta * b.y;
  t2.z= a.z + delta * b.z;
  
  norm1= T_3Norm( &t1 );
  norm2= T_3Norm( &t2 );
  
  if ( (norm1 < EPSY) && (norm2 < EPSY) ){ theDer->good= FALSE; return; }
  
  if ( fabs(theDer->k1) > fabs(theDer->k2) ){
    if (norm2 < EPSY){
      T_3VMul( &n, &t1, &t2 );
      T_3Normalize( &t2, &t2 );
    }
    else T_3RDiv( &t2,norm2,&t2 );
    aux= T_3SMul( &t2, dir );
    if (aux < 0) T_3RMul( &t2, -1.0, &t2 );
    T_3VMul( &t2, &n, &t1 );
    T_3Normalize( &t1, &t1 );
  }
  
  else{	
    if (norm1 < EPSY){
      T_3VMul( &t2, &n, &t1 );
      T_3Normalize( &t1, &t1 );
    }
    else T_3RDiv( &t1,norm1,&t1 );
    aux= T_3SMul( &t1, dir );
    if (aux < 0) T_3RMul( &t1, -1.0, &t1 );
    T_3VMul( &n, &t1, &t2 );
    T_3Normalize( &t2, &t2 );
  }
  
  T_3Copy( &t1, &theDer->t1 );
  T_3Copy( &t2, &theDer->t2 );
}

/*----------------------------------------------------------------------*/
/************************************************************************/
/*
  Wedge

	Compute the value of the extremality in the 2D case (for wedges)

	(c) Jean-Philippe THIRION -- INRIA -- November 13, 1992
*/
/************************************************************************/

/*----------------------------------------------------------------------*/
/*
    Compute curvature for 2D
*/

#if defined(_ANSI_)
reeltype T_Get2DCurvVal( int x, int y, int z, t_ImagePtr theIm, t_3rdFiltersPtr filter)
#else
reeltype T_Get2DCurvVal(x, y, z, theIm, filter)
int x;
int y;
int z;
t_ImagePtr theIm;
t_3rdFiltersPtr filter;

#endif

{
	t_CurvDer c_Der;

	T_GetAllWedgeVal( x, y, z, theIm, filter, &c_Der );

	if ( c_Der.good )	return( c_Der.c );
	else				return( 0.0 );
	}

/*----------------------------------------------------------------------*/
/*
	Compute extremality for 2D
*/

#if defined(_ANSI_)
reeltype T_GetWedgeVal( int x, int y, int z, t_ImagePtr theIm, t_3rdFiltersPtr filter)
#else
reeltype T_GetWedgeVal(x, y, z, theIm, filter)
int x;
int y;
int z;
t_ImagePtr theIm;
t_3rdFiltersPtr filter;

#endif

{
	t_CurvDer c_Der;

	T_GetAllWedgeVal( x, y, z, theIm, filter, &c_Der );

	if ( c_Der.good )	return( c_Der.e );
	else				return( 0.0 );
	}

/*----------------------------------------------------------------------*/
/*
	Compute extremality for 2D
*/

#if defined(_ANSI_)
void T_GetAllWedgeVal( int x, int y, int z, t_ImagePtr theIm, t_3rdFiltersPtr filter, t_CurvDerPtr cDeri)
#else
void T_GetAllWedgeVal(x, y, z, theIm, filter, cDeri)
int x;
int y;
int z;
t_ImagePtr theIm;
t_3rdFiltersPtr filter;
t_CurvDerPtr cDeri;

#endif

{
	t_3Derivates der3;

	T_3Derivates2DComp( x,y,z,theIm,filter, &der3 );
    E_ISERRN( "T_GetWedgeVal", NULL );

	T_GetAllWedgeDer( &der3, cDeri );
    E_ISERRN( "T_GetWedgeVal", NULL );
	}

/*----------------------------------------------------------------------*/
/*
	Compute the 2D extremality from all the derivatives
*/

#if defined(_ANSI_)
void T_GetAllWedgeDer( t_3DerivatesPtr der3, t_CurvDerPtr cDeri )
#else
void T_GetAllWedgeDer(der3, cDeri )
t_3DerivatesPtr der3;
t_CurvDerPtr cDeri ;

#endif

{
	double fx,fy;
	double fxx,fyy,fxy;
	double fxxx,fxxy,fxyy,fyyy;
	double A,B,C,_D,E,K2,K;
	double Ax,Bx,Cx,Dx,Ex,K2x;
	double Ay,By,Cy,Dy,Ey,K2y;
	double courbx,courby;


	fx= der3->fx;
	fy= der3->fy;

	fxx= der3->fxx;
	fxy= der3->fxy;
	fyy= der3->fyy;

	fxxx= der3->fxxx;
	fxxy= der3->fxxy;
	fxyy= der3->fxyy;
	fyyy= der3->fyyy;

	/* courb = A / B */

	C= 2*fx*fy*fxy;
	_D= fx*fx*fyy;
	E= fy*fy*fxx;
	A= C - _D - E;

	K2= fx*fx + fy*fy;
	K= sqrt( K2 );
	B= K * K2;

	if ( B < 1e-30 ){
		cDeri->good= FALSE;
		return;
		}

	Cx= 2*fxx*fy*fxy + 2*fx*fxy*fxy + 2*fx*fy*fxxy;
	Dx= 2*fx*fxx*fyy + fx*fx*fxyy;
	Ex= 2*fy*fxy*fxx + fy*fy*fxxx;
	Ax= Cx - Dx - Ex;
	
	Cy= 2*fxy*fy*fxy + 2*fx*fyy*fxy + 2*fx*fy*fxyy;
	Dy= 2*fx*fxy*fyy + fx*fx*fyyy;
	Ey= 2*fy*fyy*fxx + fy*fy*fxxy;
	Ay= Cy - Dy - Ey;

	K2x= 2*fx*fxx + 2*fy*fxy;
	K2y= 2*fx*fxy + 2*fy*fyy;

	Bx= 1.5 * K * K2x;
	By= 1.5 * K * K2y;

	courbx= (Ax*B - A*Bx) / (B*B);
	courby= (Ay*B - A*By) / (B*B);

	cDeri->c= (reeltype) (A / B);
	cDeri->norm= (reeltype) K;
	cDeri->n[0]= (reeltype) (fx / K);
	cDeri->n[1]= (reeltype) (fy / K);
	cDeri->dC[0]= (reeltype) courbx;
	cDeri->dC[1]= (reeltype) courby;
	cDeri->e= (reeltype) ((courbx * fy - courby * fx) / K);
	cDeri->good= TRUE;
	}

/*----------------------------------------------------------------------*/






