/* epi_filter.c 2.2 5/14/98 */
/*==========================================================================
 *
 * File : $EPI/src/epidaure.c
 * Module :
 * $ Id : ? $
 * Test file : ?
 * Author : J-P Thirion
 * Creation: November 1991
 *===========================================================================
 *   (c) Copyright 1990-1992, Inria, all right reserved
 *===========================================================================
 *
 *------------------------------Description----------------------------------
 *
 * fonctions bas niveau ge'ne'rales de la bibliothe`que Epidaure.lib
 *
 *-----------------------------Modification----------------------------------
 *
 * la normalisation de la gaussienne etait modifiee dans:
 *
 * _T_LinearGaussFilDir
 * _T_InitLinearGaussFil();
 * T_FilterGauss();
 * T_FilterGaussFunc();
 * T_GaussDer0();
 * T_GaussDer1();
 * T_GaussDer2();
 * T_GaussDer3();
 *
 * Marta Fidrich, mai 1994
 */

/*======================================================================*/
/*
	epi_filter.c

  Compute and apply filters to 3D images
  
  (c) Jean-Philippe THIRION -- INRIA -- Nov. 1991
  */
/*======================================================================*/

#ifndef _INC_MIN_
#include <epidaure.h>
#include <epidaure.ee>
#include <epidaure.ei>
#else /* _INC_MIN_ */
#include <epi_kernel.ei>
#include <epi_var.ee>
#include <epi_filter.ei>
#endif /* _INC_MIN_ */


/*-------------------- prototypes of static functions ------------------*/
#ifndef NO_PROTO
static void T_FilterGauss(reeltype *filter, 
			  int mx, int my, int mz, 
			  int Mx, int My, int Mz, 
			  int dx, int dy, int dz, 
			  t_3rdFiltersPtr ftr );
static reeltype T_FilterGaussFunc(double x, double y, double z, 
				  int mx, int my, int mz, 
				  int Mx, int My, int Mz,
				  int dx, int dy, int dz,
				  t_3rdFiltersPtr ftr,
				  double gx, double gy, double gz,
				  double sx, double sy, double sz);
static double T_GaussDer0( int mx, int Mx, double sx );
static double T_GaussDer1( double x, int mx, int Mx, double sx, int mode );
static double T_GaussDer2( double x, int mx, int Mx, double sx, int mode );
static double T_GaussDer3( double x, int mx, int Mx, double sx, int mode );
static void _T_InitLinearGaussFil( t_3rdFiltersPtr filter, int size, reeltype sigma, reeltype scale, int direct, int mode );
static void _T_LinearGaussFilDir( t_ImagePtr theIm, e_SepFilterParPtr par, int dir, t_ImagePtr resIm );
static void _T_LinGaussFilDir1( t_ImagePtr theIm, e_SepFilterParPtr par, int dir, t_3rdFiltersPtr filter, t_ImagePtr anIm, t_3IPointPtr vDer );
static void _TOP_LinGaussFilDir1( 
	t_ImagePtr theIm, e_SepFilterParPtr par, int dir, t_3rdFiltersPtr filter, 
	t_ImagePtr resIm, t_3IPointPtr vDer );
static void _TOP_LinGaussFilDir2( 
	t_ImagePtr im, e_SepFilterParPtr par,t_ImagePtr res,t_3rdFiltersPtr filter,
	reeltype *bf, int m, int M,
	int inc, int size, int pos );
#else 
static void T_FilterGauss();
static reeltype T_FilterGaussFunc();
static double T_GaussDer0();
static double T_GaussDer1();
static double T_GaussDer2();
static double T_GaussDer3();
static void _T_InitLinearGaussFil();
static void _T_LinearGaussFilDir();
static void _T_LinGaussFilDir1();
static void _TOP_LinGaussFilDir1();
static void _TOP_LinGaussFilDir2();
#endif




/************************************************************************/
/*
  GaussFilt

  Compute and apply gaussien filters to images
  
  (c) Jean-Philippe THIRION -- INRIA -- Nov. 1991
  */
/************************************************************************/

/*----------------------------------------------------------------------*/
/*
  Compute the filter corresponding to a given derivative
  of the gaussian, up to the 3rd order.
*/
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
static void T_FilterGauss(
reeltype *filter, /* the initialized filter structure */ 
int mx, /* minimal position in x */
int my, /* minimal position in y */
int mz, /* minimal position in z */
int Mx, /* maximal position in x */
int My, /* maximal position in y */
int Mz, /* maximal position in z */
int dx, /* derivative order in x */
int dy, /* derivative order in y */
int dz, /* derivative order in z */
t_3rdFiltersPtr ftr
)
#else
static void T_FilterGauss(filter, mx, my, mz, Mx, My, Mz, dx, dy, dz, ftr )
reeltype *filter;
int mx;
int my;
int mz;
int Mx;
int My;
int Mz;
int dx;
int dy;
int dz;
t_3rdFiltersPtr ftr ;

#endif

{
  reeltype *ptr;
  double x,y,z;
  int i,j,k;
  t_3PointPtr sc;
  double vsc=1.0;
  double gx,gy,gz;
  double sx,sy,sz;

  ptr= filter;
  
  sc= &ftr->scale;
  vsc= 1.0;
  gx= 1.0; gy= 1.0; gz= 1.0;

  sx= ftr->sx / ftr->scale.x;
  sy= ftr->sy / ftr->scale.y;
  sz= ftr->sz / ftr->scale.z;

  /* limitation of sigma in order to avoid degenerated cases */
  
  if ( sx < 0.4 ) sx= 0.4;
  if ( sy < 0.4 ) sy= 0.4;
  if ( sz < 0.4 ) sz= 0.4;

  
  if ( mx != Mx ) switch(dx){

  case 0: gx= T_GaussDer0(mx, Mx, sx);  break;
  case 1: vsc *= sc->x; break;
  case 2: vsc *= sc->x*sc->x; break;
  case 3: vsc *= sc->x*sc->x*sc->x; break;
  default: E_XERRN(E_ERRFATAL,"T_FilterGauss","no such power"); break;
  }
  if ( my != My ) switch(dy){
  case 0: gy = T_GaussDer0 (my, My, sy); break;
  case 1: vsc *= sc->y; break;
  case 2: vsc *= sc->y*sc->y; break;
  case 3: vsc *= sc->y*sc->y*sc->y; break;
  default: E_XERRN(E_ERRFATAL,"T_FilterGauss","no such power"); break;
  }
  if ( mz != Mz ) switch(dz){
  case 0: gz = T_GaussDer0 (mz, Mz, sz); break;
  case 1: vsc *= sc->z; break;
  case 2: vsc *= sc->z*sc->z; break;
  case 3: vsc *= sc->z*sc->z*sc->z; break;
  default: E_XERRN(E_ERRFATAL,"T_FilterGauss","no such power"); break;
  }
  
  E_ERRN( vsc < 1e-15, E_ERRDIV, "T_FilterGauss","scale too small");
  
  vsc= 1.0/vsc;
  
  for( k=mz; k<= Mz; k++ )
	  for( j=my; j<= My; j++ ){
		  for( i=mx; i<= Mx; i++ ){
			  x= i; y= j; z= k;
			  
			  *ptr++= (reeltype) (vsc * T_FilterGaussFunc(
				  x,y,z,mx,my,mz,Mx,My,Mz,
				  dx,dy,dz,ftr,gx,gy,gz,sx,sy,sz) );
			  
			  E_ISERRN("in T_FilterGauss", NULL );
		  }
	  }
  
  E_ISERRN("in T_FilterGauss",NULL);
}


/* Internally used in the initialization of the gaussian filter */

#if defined(_ANSI_)
static reeltype T_FilterGaussFunc(double x, double y, double z, 
				  int mx, int my, int mz, 
				  int Mx, int My, int Mz,
				  int dx, int dy, int dz,
				  t_3rdFiltersPtr ftr,
	double gx, double gy, double gz,
	double sx, double sy, double sz)
#else
static reeltype T_FilterGaussFunc(x, y, z, mx, my, mz, Mx, My, Mz, dx, dy, dz, ftr, gx, gy, gz, sx, sy, sz )
double x;
double y;
double z;
int mx;
int my;
int mz;
int Mx;
int My;
int Mz;
int dx;
int dy;
int dz;
t_3rdFiltersPtr ftr;
double gx;
double gy;
double gz;
double sx;
double sy;
double sz;

#endif

{	
  double vx,vy,vz;
  int mode;
  
  vx= 1.0; vy= 1.0; vz= 1.0;
  mode = ftr->normModes;

  if ( mx != Mx) switch( dx ){
  case 0: 
      vx= gx; 
      break;
  case 1: 
      vx= T_GaussDer1( x, mx, Mx, sx, mode ); 
      if (mode & NORM_DIFF) vx *= sx;
      break;
  case 2: 
      vx= T_GaussDer2( x, mx, Mx, sx, mode ); 
      if (mode & NORM_DIFF) vx *= sx*sx;
      break;
  case 3: 
      vx= T_GaussDer3( x, mx, Mx, sx, mode ); 
      if (mode & NORM_DIFF) vx *= sx*sx*sx;
      break;
  default: 
      E_XERR(E_ERRFATAL,0,"T_FilterGaussFunc","no such power"); 
      break;
  }

  if ( my != My) switch( dy ){
  case 0: 
      vy= gy; 
      break;
  case 1: 
      vy= T_GaussDer1( y, my, My, sy, mode ); 
      if (mode & NORM_DIFF) vy *= sy; 
      break;
  case 2: 
      vy= T_GaussDer2( y, my, My, sy, mode ); 
      if (mode & NORM_DIFF) vy *= sy*sy; 
      break;
  case 3: 
      vy= T_GaussDer3( y, my, My, sy, mode ); 
      if (mode & NORM_DIFF) vy *= sy*sy*sy; 
      break;
  default:
      E_XERR(E_ERRFATAL,0,"T_FilterGaussFunc","no such power"); break;
  }

  if ( mz != Mz) switch( dz ){
  case 0: 
      vz= gz; 
      break;
  case 1: 
      vz= T_GaussDer1( z, mz, Mz, sz, mode ); 
      if (mode & NORM_DIFF) vz *= sz;
      break;
  case 2: 
      vz= T_GaussDer2( z, mz, Mz, sz, mode ); 
      if (mode & NORM_DIFF) vz *= sz*sz;
      break;
  case 3: 
      vz= T_GaussDer3( z, mz, Mz, sz, mode ); 
      if (mode & NORM_DIFF) vz *= sz*sz*sz;
      break;
  default:
    E_XERR(E_ERRFATAL,0,"T_FilterGaussFunc","no such power"); break;
  }
  
  E_ISERR( 0.0, "T_FilterGaussFunc", NULL );
  return( (reeltype) (vx*vy*vz*
		      exp( -(x*x/(sx*sx)+y*y/(sy*sy)+z*z/(sz*sz)) / 2.0 ) ) );
}


/* Normalize the Gaussian filter */

#define SIGMA_MIN 9.0
#define SIGMA_MAX 12.0

#if defined(_ANSI_)
static double T_GaussDer0( int mx, int Mx, double sx )
#else
static double T_GaussDer0( mx, Mx, sx )
int mx;
int Mx;
double sx ;

#endif
{
    double sx2, sum=0.0;
    int i;

    sx2 = sx*sx;
    /* sum = sqrt(2*Pi)*sx theoretically, in practice it is smaller */
    for (i=mx; i<=Mx; i++)
	sum += exp (-i*i / (2*sx2));
    if (fabs(sum)<1e-20){
	T_Warn ("T_GaussDer0: can't compute the gaussian\n", NULL);
	return (0.0);
    }
    return (1/sum);
    /* theoretic value: 1/(sqrt(2*Pi)*sx) */
}

/* Normalize the Gaussian filter for first derivatives */

#if defined(_ANSI_)
static double T_GaussDer1( double x, int mx, int Mx, double sx, int mode )
#else
static double T_GaussDer1(x, mx, Mx, sx, mode )
double x;
int mx;
int Mx;
double sx ;
int mode;

#endif
{
    double den, sx2, sum=0.0, v0=0.0;
    int i;

    den = sqrt (2*PI) * sx;
    sx2 = sx*sx;

    if (mode & NORM_SUM){
	/* normalization by the sum of the coefficients */

	if (sx > 2.0){
	    /* sum = sx2 theoretically, in practice it is smaller */
	    for (i=mx; i<0; i++)
		sum -= i * exp (-i*i / (2*sx2));
	    if (sum < 1e-20){
		T_Warn ("T_GaussDer1: can't compute the gaussian\n", NULL);
		return (0.0);
	    }
	}
	
	if (sx <= 3.0){
	    for (i=mx; i<= Mx; i++)
		v0 += i*i*exp(-i*i/(2*sx2));
	    if (v0 < 1e-20){
		T_Warn ("T_GaussDer1: can't compute the gaussian\n", NULL);
		return (0.0);
	    }
	}
	
	if (sx <= 2.0)
	    return (-x / v0); 
	else if (sx <= 3.0)
	    return ((3.0-sx) * -x/v0 - (sx-2.0) * x/(den*sum));
	else
	    return (-x / (den*sum));
	/* theoretic value: -x / (den*sx2) */
    }

    else{
	/* normalization by polynomials */

	for (i=mx; i<= Mx; i++)
	    v0 += i*i*exp(-i*i/(2*sx2));
	if (v0 < 1e-20){
	    T_Warn ("T_GaussDer1: can't compute the gaussian\n", NULL);
	    return (0.0);
	}
	return (-x / v0); 
    }
}

/* Normalize the Gaussian filter for second derivatives */

#if defined(_ANSI_)
static double T_GaussDer2( double x, int mx, int Mx, double sx, int mode )
#else
static double T_GaussDer2(x, mx, Mx, sx, mode )
double x;
int mx;
int Mx;
double sx ;
int mode;

#endif
{
    double den, sx2, sum=0.0;
    double v0=0.0, v1=0.0, v2=0.0, a=0.0, b=0.0, tmp;

    int i;

    den = sqrt (2*PI) * sx;
    sx2 = sx*sx;

    if (mode & NORM_SUM){
	/* normalization by the sum of the coefficients */

	if (sx > 2.0){
	    /* sum = 0.0 theoretically, in practice it is negative */
	    for (i=mx; i<=0; i++)
		sum += (i*i/sx2 - 1) * exp (-i*i / (2*sx2));
	    sum = fabs(sum) / den;
	}
	
	if (sx <= 3.0){
	    for (i= mx; i<= Mx; i++){
		v0 += exp(-i*i/(2*sx2));
		v1 += i*i * exp(-i*i/(2*sx2));
		v2 += i*i*i*i * exp(-i*i/(2*sx2));
	    }
	    tmp = v0*v2 - v1*v1;
	    if ((fabs(v0) < 1e-20) || (fabs(tmp) < 1e-20)){
		T_Warn ("T_GaussDer2: can't compute the gaussian\n", NULL);
		return (0.0);
	    }
	    b = 2*v0 / tmp;
	    a = - v1*b / v0;
	}
	
	if (sx <= 2.0)
	    return (a + b*x*x);
	
	else if (sx <= 3.0){
	    
	    if (sum < 1e-6){
		tmp = (x*x/sx2 - 1) / (den*sx2);
		return ((3.0-sx) * (a + b*x*x) + (sx-2.0) * tmp);
	    }
	    if (x > sx){
		tmp = ((x+sum)*(x+sum) / sx2 - 1+sum/sx) / (den*sx2);
		return ((3.0-sx) * (a + b*x*x) + (sx-2.0) * tmp);
	    }
	    else if (-x > sx){
		tmp = ((x-sum)*(x-sum) / sx2 - 1+sum/sx) / (den*sx2);
		return ((3.0-sx) * (a + b*x*x) + (sx-2.0) * tmp);
	    }
	    else{
		tmp = (x*x/sx2 - 1) / (den*sx2);
		return ((3.0-sx) * (a + b*x*x) + (sx-2.0) * tmp);
	    }
	}
	else{
	    
	    if (sum < 1e-6)
		return ((x*x/sx2 - 1) / (den*sx2));
	    if (x > sx)
		return (( (x+sum)*(x+sum) / sx2 - 1+sum/sx) / (den*sx2));
	    else if (-x > sx)
		return (( (x-sum)*(x-sum) / sx2 - 1+sum/sx) / (den*sx2));
	    else
		return ((x*x/sx2 - 1) / (den*sx2));
	}
	/* theoretic value: (x*x/sx2 - 1) / (den*sx2) */
    }

    else{
	/* normalization by polynomials */

	for (i= mx; i<= Mx; i++){
	    v0 += exp(-i*i/(2*sx2));
	    v1 += i*i * exp(-i*i/(2*sx2));
	    v2 += i*i*i*i * exp(-i*i/(2*sx2));
	}
	tmp = v0*v2 - v1*v1;
	if ((fabs(v0) < 1e-20) || (fabs(tmp) < 1e-20)){
	    T_Warn ("T_GaussDer2: can't compute the gaussian\n", NULL);
	    return (0.0);
	}
	b = 2*v0 / tmp;
	a = - v1*b / v0;
	return (a + b*x*x);
    }
}

/* Normalize the Gaussian filter for third derivatives */

#if defined(_ANSI_)
static double T_GaussDer3( double x, int mx, int Mx, double sx, int mode )
#else
static double T_GaussDer3(x, mx, Mx, sx, mode )
double x;
int mx;
int Mx;
double sx ;
int mode;

#endif
{
    double den, sx2, sum=0.0;
    double v0=0.0, v1=0.0, v2=0.0, a=0.0, b=0.0, tmp; 

    int i;

    den = sqrt (2*PI) * sx;
    sx2 = sx*sx;

    if (mode & NORM_SUM){
	/* normalization by the sum of the coefficients */

	if (sx > 2.0){
	    /* sum = -1.0 theoretically, in practice it is bigger */
	    for (i=mx; i<=0; i++)
		sum += (3*i/sx2 - i*i*i/(sx2 *sx2)) * exp (-i*i / (2*sx2));
	    sum = fabs (sum);
	    if (sum < 1e-20){
		T_Warn ("T_GaussDer3: can't compute the gaussian\n", NULL);
		return (0.0);
	    }
	}
	
	if (sx <= 3.0){
	    for (i=mx; i<= Mx; i++){
		v0 += i*i * exp(-i*i/(2*sx2));
		v1 += i*i*i*i * exp(-i*i/(2*sx2));
		v2 += i*i*i*i*i*i * exp(-i*i/(2*sx2));
	    }
	    tmp = v0*v2 - v1*v1;
	    if ((fabs(v0) < 1e-20) || (fabs(tmp) < 1e-20)){
		T_Warn ("T_TestGaussDer3: can't compute the gaussian\n", NULL);
		return (0.0);
	    }
	    b = 6*v0 / tmp;
	    a = -v1*b / v0;
	}
	
	if (sx <= 2.0)
	    return (-a*x - b*x*x*x);
	
	else if (sx <= 3.0){
	    tmp = (3*x - x*x*x/sx2) / (den*sx2 *sx2*sum);
	    return ((3.0-sx) * (-a*x - b*x*x*x) + (sx-2.0) * tmp);
	}
	else
	    return ((3*x - x*x*x/sx2) / (den*sx2 *sx2*sum));
	/* theoretic value: (3*x - x*x*x/sx2) / (den*sx2 *sx2) */
    }

    else{
	/* normalization by polynomials */

	for (i=mx; i<= Mx; i++){
	    v0 += i*i * exp(-i*i/(2*sx2));
	    v1 += i*i*i*i * exp(-i*i/(2*sx2));
	    v2 += i*i*i*i*i*i * exp(-i*i/(2*sx2));
	}
	tmp = v0*v2 - v1*v1;
	if ((fabs(v0) < 1e-20) || (fabs(tmp) < 1e-20)){
	    T_Warn ("T_TestGaussDer3: can't compute the gaussian\n", NULL);
	    return (0.0);
	}
	b = 6*v0 / tmp;
	a = -v1*b / v0;
	return (-a*x - b*x*x*x);
    }
}


/*----------------------------------------------------------------------*/
/*
  Initialize the Gaussian filter structure with default values
  */
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
void T_DefaultFilters( t_3rdFiltersPtr filter )
#else
void T_DefaultFilters(filter )
t_3rdFiltersPtr filter ;

#endif

{
  filter->tx= 5;
  filter->ty= 5;
  filter->tz= 5;
  filter->sx= 1.0;
  filter->sy= 1.0;
  filter->sz= 1.0;
  T_3Form( 1.0, 1.0, 1.0, &filter->scale );
  filter->init= FALSE;
  filter->boardModes= 0;
  filter->normModes= 0;
}

/*----------------------------------------------------------------------*/
/*		Allocate and compute the filters of a gaussien up to the 3rd	*/
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
void T_InitFilters( t_3rdFiltersPtr filter )
#else
void T_InitFilters(filter )
t_3rdFiltersPtr filter ;

#endif

{
  int i,j,k;
  int tx,ty,tz,mx,my,mz,Mx,My,Mz;
  
  tx= filter->tx;
  ty= filter->ty;
  tz= filter->tz;
  
  Mx= tx/2;	My= ty/2;	Mz= tz/2;
  if (tx % 2) mx= -tx/2;	else mx= -tx/2 + 1;
  if (ty % 2) my= -ty/2;	else my= -ty/2 + 1;
  if (tz % 2) mz= -tz/2;	else mz= -tz/2 + 1;
  
  filter->mx= mx; filter->my= my; filter->mz= mz;
  filter->Mx= Mx; filter->My= My; filter->Mz= Mz;
  
  if ( fabs( filter->sx ) < 1e-15 ) filter->sx= (reeltype) 1e-15;
  if ( fabs( filter->sy ) < 1e-15 ) filter->sy= (reeltype) 1e-15;
  if ( fabs( filter->sz ) < 1e-15 ) filter->sz= (reeltype) 1e-15;
  
  for( k=0; k<4; k++ )
    for( j=0; j<4; j++ )
      for( i=0; i<4; i++ ){
	
	filter->d[i][j][k]= 
	  (reeltype *) T_Alloc((int)(sizeof(reeltype)*tx*ty*tz));
	
	E_REPERR(T_FreeFilters(filter));
	E_ISERRN("T_InitFilters",NULL);
	
	T_FilterGauss(filter->d[i][j][k],mx,my,mz,Mx,My,Mz,
		      i,j,k, filter);
      }
  
  filter->im= (reeltype *) T_Alloc( (int) (sizeof(reeltype)*tx*ty*tz) );
  filter->init= TRUE;
  
  E_REPERR(T_FreeFilters(filter));
  E_ISERRN("T_InitFilters",NULL);
}

#if defined(_ANSI_)
void T_FreeFilters( t_3rdFiltersPtr filter )
#else
void T_FreeFilters(filter )
t_3rdFiltersPtr filter ;

#endif

{
  int i,j,k;
  
  E_ERRN( filter->init != TRUE, E_ERRNULL, "FreeFilters", 
	 "freed before initialized");
  
  for( k=0; k<4; k++ )
    for( j=0; j<4; j++ )
      for( i=0; i<4; i++ ) free( (char *) filter->d[i][j][k] );
  
  free( (char *) filter->im );
  filter->init= FALSE;
}

/*----------------------------------------------------------------------*/
/*      Load the filter with a sub image corresponding to the data      */
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
void T_LoadFilter( int x, int y, int z, t_ImagePtr theIm, t_3rdFiltersPtr filter )
#else
void T_LoadFilter(x, y, z, theIm, filter )
int x;
int y;
int z;
t_ImagePtr theIm;
t_3rdFiltersPtr filter ;

#endif

{
  int i,j,k;
  reeltype *ptr;
  
  ptr= filter->im;
  
  for( k= filter->mz; k<= filter->Mz; k++ )
    for( j= filter->my; j<= filter->My; j++ )
      for( i= filter->mx; i<= filter->Mx; i++ ){

		if ( filter->boardModes & MODE_FOLD_MASK )
				*ptr++= T_Get3PointMode( x+i, y+j, z+k, theIm, filter->boardModes );
		else	*ptr++= T_Get3Point( x+i, y+j, z+k, theIm );
		}
  
  E_ISERRN("in T_LoadFilter",NULL);
}

/*----------------------------------------------------------------------*/
/*
  Optimized routine for filtering, with a pre-loaded filter.

  Shift the content of a loaded filter in the "axis" direction, 
  with a given sens, and load the point for the new position.
  
  x,y,z, is the former position of the point
  sens is either PLUS or MINUS.
  
  note: this function has not been tested yet !!!
  */
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
void T_ShiftFilter( int x, int y, int z, t_ImagePtr theIm, t_3rdFiltersPtr filter, int axis, int sens )
#else
void T_ShiftFilter(x, y, z, theIm, filter, axis, sens )
int x;
int y;
int z;
t_ImagePtr theIm;
t_3rdFiltersPtr filter;
int axis;
int sens ;

#endif

{
  int tx,ty,tz;
  int i,j,k;
  int off=0;
	int inc= 1;
	int new= 2;

  register reeltype *ptr;
  
  tx= filter->tx;
  ty= filter->ty;
  tz= filter->tz;
  
  switch( sens ){
  case PLUS:
    switch( axis ){
    case DIRECT_X: off= 0; inc= 1; new= filter->Mx+1; break;
    case DIRECT_Y: off= 0; inc= tx; new= filter->My+1; break;
    case DIRECT_Z: off= 0; inc= tx*ty; new= filter->Mz+1; break;
    }
    break;
  case MINUS:
    switch( axis ){
    case DIRECT_X: off=tx-1;         inc= -1;    new=filter->mx-1;break;
    case DIRECT_Y: off=tx*(ty-1);    inc= -tx;   new=filter->my-1;break;
    case DIRECT_Z: off=tx*ty*(tz-1); inc= -tx*ty;new=filter->mz-1;break;
    }
    break;
  }
  
  switch( axis ){
    
  case DIRECT_X:
    
    for( k= filter->mz; k<= filter->Mz; k++ )
      for( j= filter->my; j<= filter->My; j++ ){
	
	ptr= filter->im + j*tx + k*tx*ty + off;
	for( i=0; i<tx-1; i++ ){
	  *ptr= *(ptr+inc);
	  ptr += inc;
	}
	if ( filter->boardModes & MODE_FOLD_MASK )
			*ptr= T_Get3PointMode( x+new, y+j, z+k, theIm, filter->boardModes);
	else	*ptr= T_Get3Point( x+new, y+j, z+k, theIm);
      }
    break;
    
  case DIRECT_Y:
    
    for( k= filter->mz; k<= filter->Mz; k++ )
      for( i= filter->mx; i<= filter->Mx; i++ ){
	
	ptr= filter->im + i + k*tx*ty + off;
	for( j=0; j<ty-1; j++ ){
	  *ptr= *(ptr+inc);
	  ptr += inc;
	}
	if ( filter->boardModes & MODE_FOLD_MASK )
			*ptr= T_Get3PointMode( x+i, y+new, z+k, theIm, filter->boardModes);
	else	*ptr= T_Get3Point( x+i, y+new, z+k, theIm);
      }
    break;
    
  case DIRECT_Z:
    
    for( j= filter->my; j<= filter->My; j++ )
      for( i= filter->mx; i<= filter->Mx; i++ ){
	
	ptr= filter->im + i + j*tx + off;
	for( k=0; k<tz-1; k++ ){
	  *ptr= *(ptr+inc);
	  ptr += inc;
	}
	if ( filter->boardModes & MODE_FOLD_MASK )
			*ptr= T_Get3PointMode( x+i, y+j, z+new, theIm, filter->boardModes);
	else	*ptr= T_Get3Point( x+i, y+j, z+new, theIm);
      }
    break;
  }
  
  E_ISERRN("in T_LoadFilter",NULL);
}

/*----------------------------------------------------------------------*/
/*
  Optimized Gaussian filtering for the 1D case (separate filtering)

  Once again recall that x,y,z, is the former position of the point
  (and not the shifted one).
  */
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
void T_ShiftFilter1D( int x, int y, int z, t_ImagePtr theIm, t_3rdFiltersPtr filter, int axis, int sens )
#else
void T_ShiftFilter1D(x, y, z, theIm, filter, axis, sens )
int x;
int y;
int z;
t_ImagePtr theIm;
t_3rdFiltersPtr filter;
int axis;
int sens ;

#endif

{
  register reeltype *ptr;
  
  switch( axis ){
    
  case DIRECT_X:
    
    if (sens==PLUS){
      for( ptr=filter->im; ptr<filter->im+filter->tx-1; ptr++) 
	*ptr= *(ptr+1); 
      *ptr= T_Get3PointMode( x+filter->Mx+1, y, z, theIm, filter->boardModes);
    }
    else{
      for( ptr=filter->im+filter->tx-2; ptr>filter->im; ptr++) 
	*ptr= *(ptr-1); 
      *ptr= T_Get3PointMode( x+filter->mx-1, y, z, theIm, filter->boardModes);
    }
    break;
    
    
  case DIRECT_Y:
    
    if (sens==PLUS){
      for( ptr=filter->im; ptr<filter->im+filter->ty-1; ptr++) 
	*ptr= *(ptr+1); 
      *ptr= T_Get3PointMode( x, y+filter->My+1, z, theIm, filter->boardModes );
    }
    else{
      for( ptr=filter->im+filter->ty-2; ptr>filter->im; ptr++) 
	*ptr= *(ptr-1); 
      *ptr= T_Get3PointMode( x, y+filter->my-1, z, theIm, filter->boardModes );
    }
    break;
    
  case DIRECT_Z:
    
    if (sens==PLUS){
      for( ptr=filter->im; ptr<filter->im+filter->tz-1; ptr++) 
	*ptr= *(ptr+1); 
      *ptr= T_Get3PointMode( x, y, z+filter->Mz+1, theIm, filter->boardModes );
    }
    else{
      for( ptr=filter->im+filter->tz-2; ptr>filter->im; ptr++) 
	*ptr= *(ptr-1); 
      *ptr= T_Get3PointMode( x, y, z+filter->mz-1, theIm, filter->boardModes );
    }
    break;
    
  default: 
    E_XERRN( E_ERRUNKNOWN, "T_ShiftFilter1D","unknown direction");
    break;
  }
  
  E_ISERRN("T_ShiftFilter1D",NULL);
}

/*----------------------------------------------------------------------*/
/*      Compute the application of one of the filter                    */
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
reeltype T_ApplyFilters( t_3rdFiltersPtr filter, int dx, int dy, int dz )
#else
reeltype T_ApplyFilters(filter, dx, dy, dz )
t_3rdFiltersPtr filter;
int dx;
int dy;
int dz ;

#endif

{
  int j,k;
  register int i;
  register reeltype *ptr,*dptr;
  double value;
  
  ptr= filter->im + filter->tx*filter->ty*filter->tz -1;
  dptr= filter->d[dx][dy][dz];
  value= 0.0;
  
  for( k= filter->mz; k<= filter->Mz; k++ )
    for( j= filter->my; j<= filter->My; j++ )
      for( i= filter->mx; i<= filter->Mx; i++ )
	value += (double) (*(ptr--) * *(dptr++));
  
  return( (reeltype) value );
}

/*----------------------------------------------------------------------*/
/*		Compute a derivative of a given order.							*/
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
reeltype T_GetFiltersPt( int x, int y, int z, t_ImagePtr theIm, t_3rdFiltersPtr filter, int dx, int dy, int dz )
#else
reeltype T_GetFiltersPt(x, y, z, theIm, filter, dx, dy, dz )
int x;
int y;
int z;
t_ImagePtr theIm;
t_3rdFiltersPtr filter;
int dx;
int dy;
int dz ;

#endif

{
  E_ERR( filter->init != TRUE, E_ERRNULL, 0.0,
	"in T_GetFiltersPt", "filter not initialized" );
  
  return( T_GetFiltredPt(x,y,z,theIm,filter->d[dx][dy][dz],
			 filter->tx,filter->ty,filter->tz) );
}

/*----------------------------------------------------------------------*/
/*		
  Get the gaussien filtered value at a given point
  */
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
reeltype T_GetFilteredValue( int x, int y, int z, t_ImagePtr theIm, t_3rdFiltersPtr filter)
#else
reeltype T_GetFilteredValue(x, y, z, theIm, filter)
int x;
int y;
int z;
t_ImagePtr theIm;
t_3rdFiltersPtr filter;

#endif

{
  reeltype f;
  
  T_LoadFilter(x,y,z,theIm,filter);
  E_ISERR( 0.0, "T_GetGradNorm", NULL );
  
  f= T_ApplyFilters(filter,0,0,0);
  
  return(f);
}

/*----------------------------------------------------------------------*/
/*		Get the norm of the gradient at a given point					*/
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
reeltype T_GetGradNorm( int x, int y, int z, t_ImagePtr theIm, t_3rdFiltersPtr filter)
#else
reeltype T_GetGradNorm(x, y, z, theIm, filter)
int x;
int y;
int z;
t_ImagePtr theIm;
t_3rdFiltersPtr filter;

#endif

{
  double fx,fy,fz;
  
  T_LoadFilter(x,y,z,theIm,filter);
  E_ISERR( 0.0, "T_GetGradNorm", NULL );
  
  fx= T_ApplyFilters(filter,1,0,0);
  fy= T_ApplyFilters(filter,0,1,0);
  fz= T_ApplyFilters(filter,0,0,1);
  
  return( (reeltype) sqrt( fx*fx + fy*fy + fz*fz ) );
}

/*----------------------------------------------------------------------*/
/*		Get the norm of the gradient at a given point					*/
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
reeltype T_GetLaplace( int x, int y, int z, t_ImagePtr theIm, t_3rdFiltersPtr filter)
#else
reeltype T_GetLaplace(x, y, z, theIm, filter)
int x;
int y;
int z;
t_ImagePtr theIm;
t_3rdFiltersPtr filter;

#endif

{
  reeltype fxx,fyy,fzz;
  
  T_LoadFilter(x,y,z,theIm,filter);
  
  fxx= T_ApplyFilters(filter,2,0,0);
  fyy= T_ApplyFilters(filter,0,2,0);
  fzz= T_ApplyFilters(filter,0,0,2);
  
  return( fxx + fyy + fzz );
}

/*----------------------------------------------------------------------*/
/*		Get the gaussian curvature at a given point						*/
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
reeltype T_GetGaussCurv( int x, int y, int z, t_ImagePtr theIm, t_3rdFiltersPtr filter)
#else
reeltype T_GetGaussCurv(x, y, z, theIm, filter)
int x;
int y;
int z;
t_ImagePtr theIm;
t_3rdFiltersPtr filter;

#endif

{
  double grad2;
  double fx,fy,fz,fxx,fxy,fxz,fyy,fyz,fzz;
  
  T_LoadFilter(x,y,z,theIm,filter);
  
  fx= T_ApplyFilters(filter,1,0,0);
  fy= T_ApplyFilters(filter,0,1,0);
  fz= T_ApplyFilters(filter,0,0,1);
  
  grad2= fx*fx+fy*fy+fz*fz;
  if ( grad2 < EPSY ) return( 0.0 );
  
  fxy= T_ApplyFilters(filter,1,1,0);
  fxz= T_ApplyFilters(filter,1,0,1);
  fyz= T_ApplyFilters(filter,0,1,1);
  
  fxx= T_ApplyFilters(filter,2,0,0);
  fyy= T_ApplyFilters(filter,0,2,0);
  fzz= T_ApplyFilters(filter,0,0,2);
  
  return(	(reeltype) ((fx*fx*(fyy*fzz-fyz*fyz) + 2*fy*fz*(fxz*fxy-fxx*fyz) +
			     fy*fy*(fxx*fzz-fxz*fxz) + 2*fx*fz*(fyz*fxy-fyy*fxz) +
			     fz*fz*(fxx*fyy-fxy*fxy) + 2*fx*fy*(fxz*fyz-fzz*fxy)) /
			    ( grad2 * grad2 ) ) );
}

/*----------------------------------------------------------------------*/
/*		Get the average curvature at a given point						*/
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
reeltype T_GetAverCurv( int x, int y, int z, t_ImagePtr theIm, t_3rdFiltersPtr filter)
#else
reeltype T_GetAverCurv(x, y, z, theIm, filter)
int x;
int y;
int z;
t_ImagePtr theIm;
t_3rdFiltersPtr filter;

#endif

{
  double grad2;
  double fx,fy,fz,fxx,fxy,fxz,fyy,fyz,fzz;
  
  T_LoadFilter(x,y,z,theIm,filter);
  
  fx= T_ApplyFilters(filter,1,0,0);
  fy= T_ApplyFilters(filter,0,1,0);
  fz= T_ApplyFilters(filter,0,0,1);
  
  grad2= fx*fx+fy*fy+fz*fz;
  if ( grad2 < EPSY ) return( 0.0 );
  
  fxy= T_ApplyFilters(filter,1,1,0);
  fxz= T_ApplyFilters(filter,1,0,1);
  fyz= T_ApplyFilters(filter,0,1,1);
  
  fxx= T_ApplyFilters(filter,2,0,0);
  fyy= T_ApplyFilters(filter,0,2,0);
  fzz= T_ApplyFilters(filter,0,0,2);
  
  return(	(reeltype) ((	fx*fx*(fyy + fzz) - 2*fy*fz*fyz +
			     fy*fy*(fxx + fzz) - 2*fx*fz*fxz +
			     fz*fz*(fxx + fyy) - 2*fx*fy*fxy ) /
			    ( grad2 * sqrt(grad2) ) ) );
}

/*----------------------------------------------------------------------*/
/*		Get the norm sqrt(k1*k1 + k2*k2) curvature at a given point		*/
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
reeltype T_GetNormCurv( int x, int y, int z, t_ImagePtr theIm, t_3rdFiltersPtr filter)
#else
reeltype T_GetNormCurv(x, y, z, theIm, filter)
int x;
int y;
int z;
t_ImagePtr theIm;
t_3rdFiltersPtr filter;

#endif

{
  double grad2;
  double fx,fy,fz,fxx,fxy,fxz,fyy,fyz,fzz;
  double s,k,n;
  
  T_LoadFilter(x,y,z,theIm,filter);
  
  fx= T_ApplyFilters(filter,1,0,0);
  fy= T_ApplyFilters(filter,0,1,0);
  fz= T_ApplyFilters(filter,0,0,1);
  
  grad2= fx*fx+fy*fy+fz*fz;
  if ( grad2 < EPSY ) return( 0.0 );
  
  fxy= T_ApplyFilters(filter,1,1,0);
  fxz= T_ApplyFilters(filter,1,0,1);
  fyz= T_ApplyFilters(filter,0,1,1);
  
  fxx= T_ApplyFilters(filter,2,0,0);
  fyy= T_ApplyFilters(filter,0,2,0);
  fzz= T_ApplyFilters(filter,0,0,2);
  
  s=	(	fx*fx*(fyy + fzz) - 2*fy*fz*fyz +
	 fy*fy*(fxx + fzz) - 2*fx*fz*fxz +
	 fz*fz*(fxx + fyy) - 2*fx*fy*fxy ) /
	   (grad2 * sqrt(grad2));
  
  k=	(	fx*fx*(fyy*fzz-fyz*fyz) + 2*fy*fz*(fxz*fxy-fxx*fyz) +
	 fy*fy*(fxx*fzz-fxz*fxz) + 2*fx*fz*(fyz*fxy-fyy*fxz) +
	 fz*fz*(fxx*fyy-fxy*fxy) + 2*fx*fy*(fxz*fyz-fzz*fxy) ) /
	   (grad2 * grad2);
  
  n= s*s-2*k;
  if ( n > 0 ) n= sqrt(n);
  else		 n= 0.0;
  return( (reeltype) n );
}

/*----------------------------------------------------------------------*/
/*		Get the principal curvatures k1 and k2							*/
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
reeltype T_GetFirstCurv( int x, int y, int z, t_ImagePtr theIm, t_3rdFiltersPtr filter)
#else
reeltype T_GetFirstCurv(x, y, z, theIm, filter)
int x;
int y;
int z;
t_ImagePtr theIm;
t_3rdFiltersPtr filter;

#endif

{
  reeltype k1,k2;
  T_GetPrincCurv(x,y,z,theIm,filter,&k1,&k2);
  return( (reeltype) fabs(k1) );
}

/* Get maximal curvature absolute value when it is negative */

#if defined(_ANSI_)
reeltype T_GetGyriCurv( int x, int y, int z, t_ImagePtr theIm, t_3rdFiltersPtr filter)
#else
reeltype T_GetGyriCurv(x, y, z, theIm, filter)
int x;
int y;
int z;
t_ImagePtr theIm;
t_3rdFiltersPtr filter;

#endif

{
  reeltype k1,k2;
  T_GetPrincCurv(x,y,z,theIm,filter,&k1,&k2);
  if (k1 < 0) return(-k1);
  else		return(0.0);
}

/* Get maximal curvature when it is positive */

#if defined(_ANSI_)
reeltype T_GetSulciCurv( int x, int y, int z, t_ImagePtr theIm, t_3rdFiltersPtr filter)
#else
reeltype T_GetSulciCurv(x, y, z, theIm, filter)
int x;
int y;
int z;
t_ImagePtr theIm;
t_3rdFiltersPtr filter;

#endif

{
  reeltype k1,k2;
  T_GetPrincCurv(x,y,z,theIm,filter,&k1,&k2);
  if (k1 > 0) return(k1);
  else		return(0.0);
}

/* Get the maximal curvature in absolute value */

#if defined(_ANSI_)
void T_GetPrincCurv( int x, int y, int z, t_ImagePtr theIm, t_3rdFiltersPtr filter, reeltype *pk1, reeltype *pk2)
#else
void T_GetPrincCurv(x, y, z, theIm, filter, pk1, pk2)
int x;
int y;
int z;
t_ImagePtr theIm;
t_3rdFiltersPtr filter;
reeltype *pk1;
reeltype *pk2;

#endif

{
  double grad2;
  double fx,fy,fz,fxx,fxy,fxz,fyy,fyz,fzz;
  double s,k,delta;
  
  T_LoadFilter(x,y,z,theIm,filter);
  
  fx= T_ApplyFilters(filter,1,0,0);
  fy= T_ApplyFilters(filter,0,1,0);
  fz= T_ApplyFilters(filter,0,0,1);
  
  grad2= fx*fx+fy*fy+fz*fz;
  if ( grad2 < EPSY ){ *pk1= 0; *pk2= 0; return; }
  
  fxy= T_ApplyFilters(filter,1,1,0);
  fxz= T_ApplyFilters(filter,1,0,1);
  fyz= T_ApplyFilters(filter,0,1,1);
  
  fxx= T_ApplyFilters(filter,2,0,0);
  fyy= T_ApplyFilters(filter,0,2,0);
  fzz= T_ApplyFilters(filter,0,0,2);
  
  s=	(	fx*fx*(fyy + fzz) - 2.0*fy*fz*fyz +
	 fy*fy*(fxx + fzz) - 2.0*fx*fz*fxz +
	 fz*fz*(fxx + fyy) - 2.0*fx*fy*fxy ) /
	   (grad2 * sqrt(grad2));
  
  k=	(	fx*fx*(fyy*fzz-fyz*fyz) + 2.0*fy*fz*(fxz*fxy-fxx*fyz) +
	 fy*fy*(fxx*fzz-fxz*fxz) + 2.0*fx*fz*(fyz*fxy-fyy*fxz) +
	 fz*fz*(fxx*fyy-fxy*fxy) + 2.0*fx*fy*(fxz*fyz-fzz*fxy) ) /
	   (grad2 * grad2);
  
  delta= s*s - 4.0*k;
  if ( delta < 0 ) delta= 0;
  else		   delta= sqrt(delta);
  
  if ( s > 0 ){ 
    *pk1= (reeltype) ((s+delta)/2.0); 
    *pk2= (reeltype) ((s-delta)/2.0);
  }
  else{ 
    *pk1= (reeltype) ((s-delta)/2.0); 
    *pk2= (reeltype) ((s+delta)/2.0);
  }
}

/*----------------------------------------------------------------------*/
/*
  Apply a gaussien filter of a given size to an image (obsolete)

sigma= 1 seems to be a good default value for smoothing
*/
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
void T_ImSizedGauss( t_ImagePtr theIm, t_ImagePtr resIm, int tx, int ty, int tz, reeltype sigma )
#else
void T_ImSizedGauss(theIm, resIm, tx, ty, tz, sigma )
t_ImagePtr theIm;
t_ImagePtr resIm;
int tx;
int ty;
int tz;
reeltype sigma ;

#endif

{
  reeltype *filter,*ptr;
  reeltype x,y,z,sigma2,value;
  int i,j,k;
  
  filter= (reeltype *) T_Alloc((int) (sizeof(reeltype) * tx * ty * tz) );
  E_REPERR( free((char *) filter) );
  E_ISERRN( "in T_ImSizedGauss", NULL );
  
  ptr= filter;
  sigma2= sigma*sigma;
  
  for( k=0; k< tz; k++ )
    for( i=0; i< ty; i++ )
      for( j=0; j< tx; j++ ){
	x= (reeltype) (j - tx / 2.0);
	y= (reeltype) (i - ty / 2.0);
	z= (reeltype) (k - tz / 2.0);
	value= (reeltype) (exp(-(x*x+y*y+z*z)/(2.0*sigma2)));
	*ptr++ = value;
      }
  
  T_NormalizeFilter( filter, tx*ty*tz );
  T_ImApplyFilter( theIm, resIm, filter, tx, ty, tz );
  free( (char *) filter );
}

/*----------------------------------------------------------------------*/
/*
  Apply a gaussian filter of a given size to the image (obsolete)

sigma= 1 seems to be a good default value. This time, three separated filters
are used to compute the result.
*/
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
void T_ImSizedGauss2( t_ImagePtr theIm, t_ImagePtr resIm, int tx, int ty, int tz, reeltype sigma )
#else
void T_ImSizedGauss2(theIm, resIm, tx, ty, tz, sigma )
t_ImagePtr theIm;
t_ImagePtr resIm;
int tx;
int ty;
int tz;
reeltype sigma ;

#endif

{
  t_Image auxIm,auxIm2;
  reeltype *filter_x,*filter_y,*filter_z,*ptr;
  double sigma2,v;
  int i;
  
  filter_x= (reeltype *) T_Alloc( (int) (sizeof(reeltype) * tx) );
  filter_y= (reeltype *) T_Alloc( (int) (sizeof(reeltype) * ty) );
  filter_z= (reeltype *) T_Alloc( (int) (sizeof(reeltype) * tz) );
  
  E_REPERR( goto end_ImSizedGauss2 );
  E_ISERRN( "in T_ImSizedGauss2", NULL );
  
  sigma2= sigma*sigma;
  
  ptr= filter_x;
  for( i=0; i<tx; i++ ){ v= i - tx / 2.0; *ptr++= (reeltype) exp( -(v*v)/(2.0*sigma2) ); }
  T_NormalizeFilter( filter_x, tx );
  
  ptr= filter_y;
  for( i=0; i<ty; i++ ){ v= i - ty / 2.0; *ptr++= (reeltype) exp( -(v*v)/(2.0*sigma2) ); }
  T_NormalizeFilter( filter_y, ty );
  
  ptr= filter_z;
  for( i=0; i<tz; i++ ){ v= i - tz / 2.0; *ptr++= (reeltype) exp( -(v*v)/(2.0*sigma2) ); }
  T_NormalizeFilter( filter_z, tz );
  
  T_CopyHeaderIm( resIm, &auxIm );
  T_CopyHeaderIm( resIm, &auxIm2 );
  
  T_ImDirectFilter( theIm, &auxIm, filter_x, tx, DIRECT_X );
  T_ImDirectFilter( &auxIm, &auxIm2, filter_y, ty, DIRECT_Y );
  T_FreeImage( &auxIm );
  T_ImDirectFilter( &auxIm2, resIm, filter_z, tz, DIRECT_Z );
  T_FreeImage( &auxIm2 );
  
 end_ImSizedGauss2:
  free( (char *) filter_x );
  free( (char *) filter_y );
  free( (char *) filter_z );
}

/*----------------------------------------------------------------------*/
/*
  Normalize a filter buffer (obsolete)

  Such that the integral of positive values	is 1.0, and the integral of
 negative value is -1.0
*/
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
void T_NormalizeFilter( reeltype *filter, int nb )
#else
void T_NormalizeFilter(filter, nb )
reeltype *filter;
int nb ;

#endif

{
  reeltype posit,negat;
  int i;
  
  posit= 0.0; negat= 0.0;
  
  for(i=0; i< nb; i++ ){
    if ( filter[i] > 0 ) 	posit += filter[i];
    else					negat -= filter[i];
  }
  
  if (posit > EPSY)
    for(i=0; i< nb; i++ )
      if ( filter[i] > 0 ) 	filter[i] /= posit;
  
  if (negat > EPSY)
    for(i=0; i< nb; i++ )
      if ( filter[i] < 0 )	filter[i] /= negat;
}

/*----------------------------------------------------------------------*/
/*		Normalize the filter such as int(f,-inf..inf)= 1.0				*/
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
void T_Normalize2Filter( reeltype *filter, int nb )
#else
void T_Normalize2Filter(filter, nb )
reeltype *filter;
int nb ;

#endif

{
  reeltype value;
  int i;
  
  value= 0.0;
  
  for(i=0; i< nb; i++ ) value += filter[i];
  
  if ( fabs(value) > EPSY )
    for(i=0; i< nb; i++ ) filter[i] /= value;
}

/*----------------------------------------------------------------------*/
/************************************************************************/
/*
  LinGaussFil

  Linear seperate Gaussien filtering
  
  (c) Jean-Philippe THIRION -- INRIA -- 19/11/92
  */
/************************************************************************/
/*----------------------------------------------------------------------*/
/*
  Initialize the filter for linear filtering
*/
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
static void _T_InitLinearGaussFil( t_3rdFiltersPtr filter, int size, reeltype sigma, reeltype scale, int direct, int mode )
#else
static void _T_InitLinearGaussFil(filter, size, sigma, scale, direct, mode )
t_3rdFiltersPtr filter;
int size;
reeltype sigma;
reeltype scale;
int direct;
int mode;

#endif

{
  T_DefaultFilters( filter );
  
  filter->tx= 1;
  filter->ty= 1;
  filter->tz= 1;
  filter->sx= 1.0;
  filter->sy= 1.0;
  filter->sz= 1.0;
  
  switch( direct ){
    
  case DIRECT_X:
    
    filter->tx= size;
    filter->sx= sigma;
    filter->scale.x= scale;
    break;
    
  case DIRECT_Y:
    
    filter->ty= size;
    filter->sy= sigma;
    filter->scale.y= scale;
    break;
    
  case DIRECT_Z:
    
    filter->tz= size;
    filter->sz= sigma;
    filter->scale.z= scale;
    break;
    
  default: 
    
    E_XERRN( E_ERRMISMATCH, "_T_InitLinearGaussFil", 
	    "unknown direction" );
    break;
  }
  
  filter->normModes= mode;
  
  T_InitFilters( filter );
  E_ISERRN( "_T_InitLinearGaussFil", NULL );
}

/*----------------------------------------------------------------------*/
/*
  Internally used in directional linear Gaussian filtering
*/
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
static void _T_LinGaussFilDir1( 
	t_ImagePtr theIm, e_SepFilterParPtr par, int dir, t_3rdFiltersPtr filter,
	t_ImagePtr anIm, t_3IPointPtr vDer )
#else
static void _T_LinGaussFilDir1(theIm, par, dir, filter, anIm, vDer)
t_ImagePtr theIm;
e_SepFilterParPtr par;
int dir;
t_3rdFiltersPtr filter;
t_ImagePtr anIm ;
t_3IPointPtr vDer;

#endif

{
  reeltype value;
  int i,j,k;
  
  /* check for optimization routines */
  
  if ( (anIm->type == TYPEIM_FLOAT) &&
       ((par->mode == MODE_SET) ||
	(par->mode == MODE_ADD) ||
	(par->mode == MODE_ADD2)) &&
       ((theIm->type == TYPEIM_FLOAT) ||
	(theIm->type == TYPEIM_256)) &&
       (filter->boardModes == 0)  ){
    
    _TOP_LinGaussFilDir1( theIm, par, dir, filter, anIm, vDer );	
    return;
  }
  
  switch( dir ){
    
  case DIRECT_X:
    for( k=0; k< anIm->tz; k++ ){
      
      if ( par->verb ){
	sprintf( tmpStr1, " plane no %d\r", k );
	T_Mess( tmpStr1, NULL );
      }
      
      for( j=0; j< anIm->ty; j++ ){
	T_LoadFilter(0,j,k,theIm,filter);
	for( i=0; i< anIm->tx; i++ ){
	  value= T_ApplyFilters(filter,vDer->x,vDer->y,vDer->z);
	  T_Set3Point( i,j,k, anIm, value, (char) par->mode );
	  T_ShiftFilter1D(i,j,k,theIm,filter,DIRECT_X,PLUS);
	  E_ISERRN( "_T_LinGaussFilDir1", NULL );
	}
      }
    }
    break;
    
  case DIRECT_Y:
    
    for( k=0; k< anIm->tz; k++ ){
      if ( par->verb ){
	sprintf( tmpStr1, " plane no %d\r", k );
	T_Mess( tmpStr1, NULL );
      }
      for( i=0; i< anIm->tx; i++ ){
	T_LoadFilter(i,0,k,theIm,filter);
	for( j=0; j< anIm->ty; j++ ){
	  value= T_ApplyFilters(filter,vDer->x,vDer->y,vDer->z);
	  T_Set3Point( i,j,k, anIm, value,  (char) par->mode );
	  
	  T_ShiftFilter1D(i,j,k,theIm,filter,DIRECT_Y,PLUS);
	  E_ISERRN( "_T_LinGaussFilDir1", NULL );
	}
      }
    }
    break;
    
  case DIRECT_Z:
    for( j=0; j< anIm->ty; j++ ){
      if ( par->verb ){
	sprintf( tmpStr1, " plane no %d\r", j );
	T_Mess( tmpStr1, NULL );
      }
      for( i=0; i< anIm->tx; i++ ){
	T_LoadFilter(i,j,0,theIm,filter);
	for( k=0; k< anIm->tz; k++ ){
	  value= T_ApplyFilters(filter,vDer->x,vDer->y,vDer->z);
	  T_Set3Point( i,j,k, anIm, value,  (char) par->mode );
	  T_ShiftFilter1D(i,j,k,theIm,filter,DIRECT_Z,PLUS);
	  E_ISERRN( "_T_LinGaussFilDir1", NULL );
	}
      }
    }
    break;
    
  default: E_SETERR(E_ERRUNKNOWN,"_T_LinearGaussFilDir","unknown direction");
    break;
  }
  
  E_ISERRN( "_T_LinGaussFilDir1", NULL );
}
  
/*----------------------------------------------------------------------*/
/*
  Apply a directional linear Gaussian filtering to the image
*/
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
static void _T_LinearGaussFilDir( t_ImagePtr theIm, e_SepFilterParPtr par, int dir, t_ImagePtr resIm )
#else
static void _T_LinearGaussFilDir(theIm, par, dir, resIm )
t_ImagePtr theIm;
e_SepFilterParPtr par;
int dir;
t_ImagePtr resIm ;

#endif

{
  int size= 0;
  reeltype sigma= 1.0;
	reeltype scale= 1.0;
  t_3IPoint vDer;
  t_ImagePtr anIm;
  t_Image auxIm;
  t_3rdFilters filter;
  
  if ( (resIm->buf == NULL) || (resIm->buf == theIm->buf) ){
    T_CopyHeaderIm( resIm, &auxIm );
    T_AllocImage( &auxIm );
    anIm= &auxIm;
  }
  else anIm= resIm;
  
  T_3IForm( 0, 0, 0, &vDer );
  
  switch( dir ){
	case DIRECT_X:
		sigma= par->vSigma.x;
		scale= par->vScale.x;
		size= par->vSize.x;
		vDer.x= par->vDer.x;
		break;
	case DIRECT_Y:
		sigma= par->vSigma.y;
		scale= par->vScale.y;
	 	size= par->vSize.y;
		vDer.y= par->vDer.y;
	 	break;
	case DIRECT_Z:
		sigma= par->vSigma.z;
		scale= par->vScale.z;
		size= par->vSize.z;
		vDer.z= par->vDer.z;
		break;
		}
  
	_T_InitLinearGaussFil( &filter, size, sigma, scale, dir, par->normModes );
	E_ISERRN( "_T_LinearGaussFilDir", NULL );

	filter.boardModes= par->boardModes;

	_T_LinGaussFilDir1( theIm, par, dir, &filter, anIm, &vDer );

	T_FreeFilters( &filter );
	if ( resIm != anIm ){
    	T_FreeImage( resIm );
		resIm->buf= anIm->buf;
		}
	E_REPERR( T_FreeImage( resIm ) );
	E_ISERRN( "_T_LinearGaussFilDir", NULL );
	}

/*----------------------------------------------------------------------*/
/*
  Perform the linear filtering according to the 3 space directions
*/
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
void E_SepFilter( t_ImagePtr theIm, e_SepFilterParPtr par, t_ImagePtr resIm )
#else
void E_SepFilter(theIm, par, resIm )
t_ImagePtr theIm;
e_SepFilterParPtr par;
t_ImagePtr resIm ;

#endif

{
  t_ImagePtr anIm;
  t_Image auxIm;
  
  if ( (resIm->buf == NULL) || (resIm->buf == theIm->buf) ){
    T_CopyHeaderIm( resIm, &auxIm );
    T_AllocImage( &auxIm );
    anIm= &auxIm;
  }
  else anIm= resIm;
  E_REPERR( goto END_LinearGaussFil );
  
  if ( par->verb ) T_Mess("filtering in X direction\n",NULL);
  if ((par->vSize.x > 1) && (theIm->tx > 1))
       _T_LinearGaussFilDir(theIm,par,DIRECT_X,anIm);
  else T_CopyImageBuffer( theIm, anIm );
  
  E_REPERR( goto END_LinearGaussFil );
  
  if ( par->verb ) T_Mess("filtering in Y direction\n",NULL);
  if ((par->vSize.y > 1) && (theIm->ty > 1))
    _T_LinearGaussFilDir(anIm,par,DIRECT_Y,anIm);
  
  E_REPERR( goto END_LinearGaussFil );
  
  if ( par->verb ) T_Mess("filtering in Z direction\n",NULL);
  if ((par->vSize.z > 1) && (theIm->tz > 1))
    _T_LinearGaussFilDir(anIm,par,DIRECT_Z,anIm);
  
  E_REPERR( goto END_LinearGaussFil );
  
 END_LinearGaussFil:
  
  if ( resIm != anIm ){
    T_FreeImage( resIm );
    resIm->buf= anIm->buf;
  }
  E_REPERR( T_FreeImage( resIm ) );
  E_ISERRN( "_T_LinearGaussFil", NULL );
}

/*----------------------------------------------------------------------*/
/*
  Computation of the Laplacien with separated filters
  */
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
void E_SepLaplace( t_ImagePtr theIm, e_SepFilterParPtr par, t_ImagePtr resIm )
#else
void E_SepLaplace(theIm, par, resIm )
t_ImagePtr theIm;
e_SepFilterParPtr par;
t_ImagePtr resIm ;

#endif

{
  t_ImagePtr anIm;
  t_Image auxIm,aux2Im;
  e_SepFilterPar aPar;
  
  if ( (resIm->buf == NULL) || (resIm->buf == theIm->buf) ){
    T_CopyHeaderIm( resIm, &auxIm );
    T_AllocImage( &auxIm );
    anIm= &auxIm;
  }
  else anIm= resIm;
  E_REPERR( goto END_SepLaplace );
  
  T_CopyHeaderIm( resIm, &aux2Im );
  T_AllocImage( &aux2Im );
  
  aPar= *par;
  if (par->verb) aPar.verb= par->verb - 1;
  if ( theIm->tz <= 1 ) aPar.vSize.z= 1;
  
  if (par->verb) T_Mess("compute differentials according to X\n", NULL );
  T_3IForm( 2, 0, 0, &aPar.vDer );
  E_SepFilter( theIm, &aPar, anIm );
  
  if (par->verb) T_Mess("compute differentials according to Y\n", NULL );
  T_3IForm( 0, 2, 0, &aPar.vDer );
  E_SepFilter( theIm, &aPar, &aux2Im );
  T_ImageCompose( anIm, &aux2Im, anIm, MODE_ADD );
  
  if ( theIm->tz > 1 ){
    if (par->verb) T_Mess("compute differentials according to Z\n",NULL);
    T_3IForm( 0, 0, 2, &aPar.vDer );
    E_SepFilter( theIm, &aPar, &aux2Im );
    T_ImageCompose( anIm, &aux2Im, anIm, MODE_ADD );
  }
  
 END_SepLaplace:
  
  T_FreeImage( &aux2Im );
  if ( resIm != anIm ){
    T_FreeImage( resIm );
    resIm->buf= anIm->buf;
  }
  E_REPERR( T_FreeImage( resIm ) );
  E_ISERRN( "E_SepLaplace", NULL );
}

/*----------------------------------------------------------------------*/
/*
  Computation of the gradient norm with separated filters
  */
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
void E_SepGradientNorm( t_ImagePtr theIm, e_SepFilterParPtr par, t_ImagePtr resIm )
#else
void E_SepGradientNorm(theIm, par, resIm )
t_ImagePtr theIm;
e_SepFilterParPtr par;
t_ImagePtr resIm ;

#endif

{
  t_ImagePtr anIm;
  t_Image auxIm,aux2Im;
  e_SepFilterPar aPar;
  
  if ( (resIm->buf == NULL) || (resIm->buf == theIm->buf) ){
    T_CopyHeaderIm( resIm, &auxIm );
    T_AllocImage( &auxIm );
    anIm= &auxIm;
  }
  else anIm= resIm;
  E_REPERR( goto END_SepGradientNorm );
  
  T_CopyHeaderIm( resIm, &aux2Im );
  T_AllocImage( &aux2Im );
  
  aPar= *par;
  if (par->verb) aPar.verb= par->verb - 1;
  if ( theIm->tz <= 1 ) aPar.vSize.z= 1;
  
  if (par->verb) T_Mess("compute differentials according to X\n", NULL );
  T_3IForm( 1, 0, 0, &aPar.vDer );
  E_SepFilter( theIm, &aPar, anIm );
  T_ImageExp2( anIm, anIm );
  
  if (par->verb) T_Mess("compute differentials according to Y\n", NULL );
  T_3IForm( 0, 1, 0, &aPar.vDer );
  E_SepFilter( theIm, &aPar, &aux2Im );
  T_ImageExp2( &aux2Im, &aux2Im );
  T_ImageCompose( anIm, &aux2Im, anIm, MODE_ADD );
  
  if ( theIm->tz > 1 ){
    if (par->verb) T_Mess("compute differentials according to Z\n",NULL);
    T_3IForm( 0, 0, 1, &aPar.vDer );
    E_SepFilter( theIm, &aPar, &aux2Im );
    T_ImageExp2( &aux2Im, &aux2Im );
    T_ImageCompose( anIm, &aux2Im, anIm, MODE_ADD );
  }
  
  T_ImageSqrt( anIm, anIm );
  
 END_SepGradientNorm:
  
  T_FreeImage( &aux2Im );
  if ( resIm != anIm ){
    T_FreeImage( resIm );
    resIm->buf= anIm->buf;
  }
  E_REPERR( T_FreeImage( resIm ) );
  E_ISERRN( "E_SepGradientNorm", NULL );
}

/*----------------------------------------------------------------------*/
/*
  Initialize the separate filtering structure with default values
  */
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
void E_DefaultSepFilter( e_SepFilterParPtr par )
#else
void E_DefaultSepFilter(par )
e_SepFilterParPtr par ;

#endif

{
  T_3Form( 1.0, 1.0, 1.0, &par->vSigma );
  T_3Form( 1.0, 1.0, 1.0, &par->vScale );
  T_3IForm( 11, 11, 11, &par->vSize );
  T_3IForm( 0, 0, 0, &par->vDer );
  par->verb= FALSE;
  par->mode= MODE_SET;
  par->boardModes= 0;
  par->normModes= 0;
}

/*----------------------------------------------------------------------*/
/************************************************************************/
/*
  LocalOptim

  	Local optimizations of the filtering

	Those functions depends on the type of the image
  
  (c) Jean-Philippe THIRION -- INRIA -- 29/4/94
  */
/************************************************************************/
/*----------------------------------------------------------------------*/
/*
  Linear Gaussian filtering of an image (optimized)

  Perform the linear filtering with a gaussien of an image, according
  to one given direction. This is an optimized version
*/
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
static void _TOP_LinGaussFilDir1( 
	t_ImagePtr theIm, e_SepFilterParPtr par, int dir, t_3rdFiltersPtr filter, 
	t_ImagePtr resIm, t_3IPointPtr vDer )
#else
static void _TOP_LinGaussFilDir1(theIm, par, dir, filter, resIm, vDer)
t_ImagePtr theIm;
e_SepFilterParPtr par;
int dir;
t_3rdFiltersPtr filter;
t_ImagePtr resIm ;
t_3IPointPtr vDer;

#endif

{
	int i,j,k;
	int tx,ty,tz;
	reeltype *bf;

	E_ERRN( (par->mode != MODE_SET) &&
		(par->mode != MODE_ADD) &&
		(par->mode != MODE_ADD2), E_ERRFATAL, "_TOP_LinearGaussFilDir1",
		"unsupported image mode");

	E_ERRN( resIm->type != TYPEIM_FLOAT, E_ERRFATAL, "_TOP_LinearGaussFilDir1",
		"wrong type for output image" );

	E_ERRN( (theIm->type != TYPEIM_256) &&
			(theIm->type != TYPEIM_FLOAT), E_ERRFATAL,"_TOP_LinearGaussFilDir1",
		"wrong type for input image" );

	tx= resIm->tx;
	ty= resIm->ty;
	tz= resIm->tz;

	bf= filter->d[vDer->x][vDer->y][vDer->z];

	switch( dir ){
    
		case DIRECT_X:

			for( k=0; k< tz; k++ ){

				if ( par->verb ){
					sprintf( tmpStr1, " plane no %d\r", k );
					T_Mess( tmpStr1, NULL );
					}

				for( j=0; j< ty; j++ ){

					_TOP_LinGaussFilDir2( theIm, par, resIm, filter,
						bf, filter->mx, filter->Mx,
						1, tx, j * tx + k * tx * ty );
					}
    			}
    		break;
    
		case DIRECT_Y:

			for( k=0; k< tz; k++ ){

				if ( par->verb ){
					sprintf( tmpStr1, " plane no %d\r", k );
					T_Mess( tmpStr1, NULL );
					}

				for( i=0; i< tx; i++ ){

					_TOP_LinGaussFilDir2( theIm, par, resIm, filter,
						bf, filter->my, filter->My,
						tx, ty, i + k * tx * ty );
					}
    			}
    		break;
    
		case DIRECT_Z:

			for( j=0; j< ty; j++ ){
			
				if ( par->verb ){
					sprintf( tmpStr1, " plane no %d\r", j );
					T_Mess( tmpStr1, NULL );
      				}

				for( i=0; i< tx; i++ ){

					_TOP_LinGaussFilDir2( theIm, par, resIm, filter,
						bf, filter->mz, filter->Mz,
						tx*ty, tz, i + j * tx );
					}
      			}
			break;
    
		default: 
			E_XERRN(E_ERRUNKNOWN,"_TOP_LinearGaussFilDir1","unknown direction");
			break;
		}
	}
 
/*----------------------------------------------------------------------*/
/*
	First subroutine for optimized filtering
*/
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
static void _TOP_LinGaussFilDir2( 
	t_ImagePtr im, e_SepFilterParPtr par,t_ImagePtr res,t_3rdFiltersPtr filter,
	reeltype *bf, int m, int M,
	int inc, int size, int pos )
#else
static void _TOP_LinGaussFilDir2(im, par, res, filter, bf, m, M, inc, size, pos)
t_ImagePtr im;
e_SepFilterParPtr par;
t_ImagePtr res;
t_3rdFiltersPtr filter;
reeltype *bf;
int m,M,inc,size,pos;

#endif

{
  int i,bSize,p;
  reeltype *resP;
  
  register double value;
  register reeltype v;
  register reeltype *dptr;
  register reeltype *ptr;
  register int j;
  
  /* start to load the values */
  
  bSize= M - m + 1;
  ptr= filter->im;
  
  for( j= m; j <= M; j++ ){
    
    p= j % size;
    if ( p < 0 ) p += size;		/* avoid side effects, p in 0..size-1 */
    
    switch( im->type ){
      
    case TYPEIM_256:
      *ptr++ = (reeltype) 
	( *(((unsigned char *) im->buf) + pos + p * inc) / 255.0 );
      break;
      
    case TYPEIM_FLOAT:
      *ptr++ = *(((reeltype *) im->buf) + pos + p * inc);
      break;
    }
  }
  
  /* compute all values */
  
  resP= ((reeltype *) res->buf) + pos;
  
  for( i=0; i< size; i++ ){
    
    ptr= filter->im;
    dptr= bf;
    value= 0.0;
    
    for( j=0; j < bSize; j++ ){
      v= *ptr; *ptr= *(ptr+1); ptr++;
      value += (double) (*(dptr++) * v);
    }
    
    switch( par->mode ){
      
    case MODE_SET: *resP= (reeltype) value; break;
    case MODE_ADD: *resP += (reeltype) value; break;
    case MODE_ADD2: *resP += (reeltype) (value * value); break;
    }
    
    resP += inc;
    
    p= (i + M + 1) % size;
    if ( p < 0 ) p += size;
    
    switch( im->type ){		/* load next image value */
      
    case TYPEIM_256:
      filter->im[bSize-1]= (reeltype)
	( *(((unsigned char *) im->buf)+pos+p*inc) / 255.0 );
      break;
      
    case TYPEIM_FLOAT:
      filter->im[bSize-1]= *(((reeltype *) im->buf)+pos+p*inc);
      break;
    }
  }
}

/*----------------------------------------------------------------------*/
/*
    Computation of the pseudo-gradient square norm

	It is computed with separated filters.
	pseudo because the filtering is computed independently according
	to x,y, and z, which is acceptable for an already smooth image.
*/
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
void T_PseudoGradientN2( t_ImagePtr theIm, e_SepFilterParPtr par, t_ImagePtr resIm )
#else
void T_PseudoGradientN2(theIm, par, resIm )
t_ImagePtr theIm;
e_SepFilterParPtr par;
t_ImagePtr resIm ;

#endif

{
  t_ImagePtr anIm;
  t_Image auxIm;
  e_SepFilterPar aPar;
  
  if ( (resIm->buf == NULL) || (resIm->buf == theIm->buf) ){
    T_CopyHeaderIm( resIm, &auxIm );
    T_AllocImage( &auxIm );
	T_ClearImage( &auxIm );
    anIm= &auxIm;
  }
  else anIm= resIm;
  E_REPERR( goto END_PseudoGradientN2 );
  
  if ( theIm->tz <= 1 ) aPar.vSize.z= 1;
 
  if (par->verb) T_Mess("compute differentials according to X\n", NULL );
  aPar= *par;
  if (par->verb) aPar.verb= par->verb - 1;
  aPar.vSize.y= 1;
  aPar.vSize.z= 1;
  aPar.mode= MODE_ADD2;
  T_3IForm( 1, 0, 0, &aPar.vDer );
  _T_LinearGaussFilDir(theIm,&aPar,DIRECT_X,anIm);
  
  if (par->verb) T_Mess("compute differentials according to Y\n", NULL );
  aPar= *par;
  if (par->verb) aPar.verb= par->verb - 1;
  aPar.vSize.x= 1;
  aPar.vSize.z= 1;
  aPar.mode= MODE_ADD2;
  T_3IForm( 0, 1, 0, &aPar.vDer );
  _T_LinearGaussFilDir(theIm,&aPar,DIRECT_Y,anIm);
  
  if ( (theIm->tz > 1) && (par->vSize.z > 1) ){
   	if (par->verb) T_Mess("compute differentials according to Z\n",NULL);
  	aPar= *par;
  	if (par->verb) aPar.verb= par->verb - 1;
  	aPar.vSize.x= 1;
  	aPar.vSize.y= 1;
    aPar.mode= MODE_ADD2;
  	T_3IForm( 0, 0, 1, &aPar.vDer );
  	_T_LinearGaussFilDir(theIm,&aPar,DIRECT_Z,anIm);
  }
  
END_PseudoGradientN2:
  
  if ( resIm != anIm ){
    T_FreeImage( resIm );
    resIm->buf= anIm->buf;
  }
  E_REPERR( T_FreeImage( resIm ) );
  E_ISERRN( "T_PseudoGradientN2", NULL );
}

/*----------------------------------------------------------------------*/
/*
  Computation of the pseudo-laplacien with separated filters

  pseudo because the filtering is computed independently according
  to x,y, and z, which is acceptable for an already smooth image.
*/
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
void T_PseudoLaplace(t_ImagePtr theIm, e_SepFilterParPtr par, t_ImagePtr resIm)
#else
void T_PseudoLaplace(theIm, par, resIm )
t_ImagePtr theIm;
e_SepFilterParPtr par;
t_ImagePtr resIm ;

#endif

{
  t_ImagePtr anIm;
  t_Image auxIm;
  e_SepFilterPar aPar;
  
  if ( (resIm->buf == NULL) || (resIm->buf == theIm->buf) ){
    T_CopyHeaderIm( resIm, &auxIm );
    T_AllocImage( &auxIm );
	T_ClearImage( &auxIm );
    anIm= &auxIm;
  }
  else anIm= resIm;
  E_REPERR( goto END_PseudoLaplace );
  
  if (par->verb) T_Mess("compute differentials according to X\n", NULL );
  aPar= *par;
  aPar.vSize.y= 1;
  aPar.vSize.z= 1;
  aPar.mode= MODE_ADD;
  T_3IForm( 2, 0, 0, &aPar.vDer );
  if (par->verb) aPar.verb= par->verb - 1;
  _T_LinearGaussFilDir(theIm,&aPar,DIRECT_X,anIm);
  
  if (par->verb) T_Mess("compute differentials according to Y\n", NULL );
  aPar= *par;
  aPar.vSize.x= 1;
  aPar.vSize.z= 1;
  aPar.mode= MODE_ADD;
  T_3IForm( 0, 2, 0, &aPar.vDer );
  if (par->verb) aPar.verb= par->verb - 1;
  _T_LinearGaussFilDir(theIm,&aPar,DIRECT_Y,anIm);
  
  if ( (theIm->tz > 1) && (par->vSize.z > 1) ){
    if (par->verb) T_Mess("compute differentials according to Z\n",NULL);
  	aPar= *par;
  	aPar.vSize.x= 1;
  	aPar.vSize.y= 1;
    aPar.mode= MODE_ADD;
  	T_3IForm( 0, 0, 2, &aPar.vDer );
  	if (par->verb) aPar.verb= par->verb - 1;
  	_T_LinearGaussFilDir(theIm,&aPar,DIRECT_Z,anIm);
  }
  
 END_PseudoLaplace:
  
  if ( resIm != anIm ){
    T_FreeImage( resIm );
    resIm->buf= anIm->buf;
  }
  E_REPERR( T_FreeImage( resIm ) );
  E_ISERRN( "T_PseudoLaplace", NULL );
}

/*---------------------------------------------------------------------------
 *      End of file
 *---------------------------------------------------------------------------
 */





