/* epi_ima.c 2.3 5/14/98 */
/*
 * File : $EPI/src/epi_ima.c
 * Module :
 * $ Id : ? $ 
 * Test file : ?
 * Author : J-P Thirion
 * Creation: August 4 1992
 *===========================================================================
 *	(c) Copyright 1992, Inria, all right reserved
 *===========================================================================
 *
 *------------------------------Description----------------------------------
 *
 * Fonctions generales d'acces aux voxels et de traitement d'images
 *
 *-----------------------------Modification----------------------------------
 *
 * juillet 1993 : G. Malandain
 *                introduction du type short int : TYPEIM_S16B
 *                fonctions modifiees :
 *                - *T_GetPtr3Point
 *                - T_Get3Point
 *                - T_GetIntPoint
 *                - T_Interpol3
 *                - T_Set3Point
 *                - T_SetIPoint
 *                - T_GetIPoint
 *                - T_ImageSize
 *                - T_WriteInrimLow
 *                - T_ReadInrim
 *                - T_ReadInrimHead
 *                - T_WriteInrimHead
 *                - T_InvertIm
 *
 * Mars 1999 : X. Pennec
 *             add voxel size in Inrimage headers and modified functions:
 *             - T_ReadInrimHead
 *             - T_WriteInrimHead
 *             - T_Init3Image
 *             Add function:
 *             - T_SetVoxelSize
 *  
 */

/*==========================================================================*/
/* 
	epi_ima.c

	Low level image processing functions

	(c) Jean-Philippe THIRION -- INRIA -- 25/10/90
*/
/*==========================================================================*/

#if defined(INC_MIN) /* inclusions minimales */
 /* rajouter -DINC_MIN dans les options de compilation */
 /* GM 12/09/94                                        */

#include <epi_kernel.ei>
#include <epi_filter.ei>
#include <epi_var.ee>
/* #include <sys/types.h> */
/* #include <unistd.h> */
/* #include <sys/stat.h> */
#include <fcntl.h>
#include <stdlib.h>

#else /* INC_MIN */

#include <epidaure.h>
#include <epidaure.ee>
#include <epidaure.ei>

#endif /* INC_MIN */

/*-------Definition des fonctions statiques----------*/
#ifndef NO_PROTO
static int T_Compare256( unsigned char *r1, unsigned char *r2 );
#else 
static int T_Compare256();
#endif
/*-------Fin des proto statiques----------*/


/*-------------------- prototype of extern functions -------------------*/
#ifndef NO_PROTO
extern int abs (int i);
#else 
extern int abs ( );
#endif

/*****************************************************************************/
/* 
	lowVoxels

	Low level routines of access to pixels and voxels values.

It includes the bresenham algorithm of segments drawing

   (c) Jean-Philippe THIRION -- INRIA -- 25/10/90
*/
/*****************************************************************************/

/*---------------------------------------------------------------------------*/

/*
	Old function of access to a point pointer in 2D images

	Used in order to have the compatibility with the old version.
*/


#if defined(_ANSI_)
char *T_GetPtrPoint( int xx, int yy, t_ImagePtr theIm )
#else
char *T_GetPtrPoint(xx, yy, theIm )
int xx;
int yy; 
t_ImagePtr theIm ;

#endif
{ 
  return( T_GetPtr3Point(xx, yy, 0, theIm) ); 
}

/*
	Old function of access to a point value in 2D images

	Used in order to have the compatibility with the old version.
*/

#if defined(_ANSI_)
reeltype T_GetPoint( int xx, int yy, t_ImagePtr theIm )
#else
reeltype T_GetPoint(xx, yy, theIm )
int xx;
int yy;
t_ImagePtr theIm ;

#endif
{ 
  return( T_Get3Point(xx, yy, 0, theIm) ); 
}

/*
	Old function to set the value of a point in 2D images

	Used in order to have the compatibility with the old version.
*/

#if defined(_ANSI_)
void T_SetPoint( int xx, int yy, t_ImagePtr theIm, reeltype color, char mode )
#else
void T_SetPoint(xx, yy, theIm, color, mode )
int xx;
int yy;
t_ImagePtr theIm;
reeltype color;
char mode ;

#endif
{ 
  T_Set3Point( xx, yy, 0, theIm, color, mode ); 
}

/*
	Old function to draw a 2D segment in a 2D image

	Used in order to have the compatibility with the old version.
*/

#if defined(_ANSI_)
void T_Segment( t_2PointPtr p1, t_2PointPtr p2, t_ImagePtr theIm, reeltype color, char mode )
#else
void T_Segment(p1, p2, theIm, color, mode )
t_2PointPtr p1;
t_2PointPtr p2;
t_ImagePtr theIm;
reeltype color;
char mode ;

#endif
{ 
  T_Segment3( p1, p2, 0, theIm, color, mode ); 
}


/*
	Get a pointer toward the point (seldom used).

	You should then convert this pointer to the type described in the image header. Other high level fonctions are frequently prefered to this one.
*/

#if defined(_ANSI_)
char *T_GetPtr3Point( int xx, int yy, int zz, t_ImagePtr theIm )
#else
char *T_GetPtr3Point(xx, yy, zz, theIm )
int xx;
int yy;
int zz;
t_ImagePtr theIm ;

#endif

{
  int plan;
  
  xx= xx % theIm->tx;
  yy= yy % theIm->ty;
  
  if ( xx < 0 ) xx += theIm->tx;
  if ( yy < 0 ) yy += theIm->ty;
  
  if ( zz ){	zz= zz % theIm->tz;
		if ( zz < 0 ) zz += theIm->tz;
		plan= zz * theIm->tx * theIm->ty; }
  else		plan=0;
  
  switch( theIm->type ){
    
  case TYPEIM_BW:
    
    return( theIm->buf + ((xx+yy*theIm->tx+plan) >> 3) );	break;
    
  case TYPEIM_256:
    
    return( theIm->buf + xx + yy * theIm->tx + plan ); break;
    
  case TYPEIM_16B :
  case TYPEIM_S16B :
    return((char *) (((short int*) theIm->buf)+xx+yy*theIm->tx+plan) );
    break;
    
  case TYPEIM_FLOAT:
    
    return((char *) (((reeltype *) theIm->buf)+xx+yy*theIm->tx+plan) );
    break;
    
  default: 
    E_XERR( E_ERRUNKNOWN, NULL, "GetPtrPoint","unknown image mode" );
    break;
  }
  
  return( (char *) NULL );
}

/*----------------------------------------------------------------------*/

/*
	Get the color intensity of a point.

Returns:
	A real value which is generally normalized within 0 and 1, whatever the type of the image. For signed images it can return also signed values within -1 and 1, and for floating points images, the values are not obliged to be normalized.
*/

#if defined(_ANSI_)
reeltype T_Get3Point( int xx, int yy, int zz, t_ImagePtr theIm )
#else
reeltype T_Get3Point(xx, yy, zz, theIm )
int xx,yy,zz; /* position in the image */
t_ImagePtr theIm ; /* the image */

#endif

{
  int plan;
  unsigned char value;
  unsigned short sval;
  short int ssval;
  

  xx= xx % theIm->tx;
  yy= yy % theIm->ty;
  
  if ( xx < 0 ) xx += theIm->tx;
  if ( yy < 0 ) yy += theIm->ty;
  
  if ( zz ){	zz= zz % theIm->tz;
		if ( zz < 0 ) zz += theIm->tz;
		plan= zz * theIm->tx * theIm->ty; }
  else		plan=0;
  
  switch( theIm->type ){
    
  case TYPEIM_BW:
    
    value= 	(unsigned char) theIm->buf[(xx+yy*theIm->tx+plan) >> 3] & 
      (0x80 >> (xx % 8));
    if ( value ) return( 1.0 ); 
    else		 return( 0.0 );
    break;
    
  case TYPEIM_256:
    
    return( (reeltype) (((unsigned char) 
			 theIm->buf[xx+yy*theIm->tx+plan]) / 255.0 ));
    break;
    
  case TYPEIM_16B:
    
    sval= *( ((unsigned short *) theIm->buf) + xx+yy*theIm->tx+plan );
    return( (reeltype) (sval / 65535.0) );
    break;
    
  case TYPEIM_S16B:
    
    ssval= *( ((short int*) theIm->buf) + xx+yy*theIm->tx+plan );
    return( (reeltype) (ssval / 32767.0) );
    break;
    
  case TYPEIM_FLOAT:
    return( *(((reeltype *) theIm->buf) + xx + yy * theIm->tx + plan));
    break;
    
  default: 
    E_XERR( E_ERRUNKNOWN, 0.0, "T_Get3Point","unknown image type" );
    break;
  }
  
  return( 0.0 );
}

/*----------------------------------------------------------------------*/
/*
	Process special side effects modes for voxel access. Return TRUE
	if inside the image (or extended image). Return FALSE otherwise.
*/
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
static int _T_FoldMode( int *x, int tx, int mode )
#else
static int _T_FoldMode( x, tx, mode )
int *x;
int tx;
int mode;

#endif

{
	int even,xx;

	xx= *x;

	switch( mode ){

		case MODE_1_EXTEND:

			if ( xx < 0 ) *x= 0;
			else if ( xx >= tx ) *x= tx - 1;
			break;

		case MODE_1_UNFOLD:

			if ( !tx || (xx / tx) % 2 ) even= TRUE;
			else						even= FALSE;

			xx= xx % tx;
			if ( xx < 0 ) xx += tx;
			if ( even )	*x= xx;
			else		*x= tx - xx - 1;
			break;

		case MODE_1_CLIP:

			if ( xx < 0 )			return( FALSE );
			else if ( xx >= tx )	return( FALSE );
			break;

		default:
	
			xx= xx % tx;
			if ( xx < 0 ) 	*x= xx + tx;
			else			*x= xx;
			break;
		}

	return( TRUE );
	}

/*----------------------------------------------------------------------*/

/*
	Get a 3D image point value, with special access mode.

Returns:
A reelype value (see also T_Get3Point).
*/

#if defined(_ANSI_)
reeltype T_Get3PointMode( int xx, int yy, int zz, t_ImagePtr theIm, int mode )
#else
reeltype T_Get3PointMode( xx, yy, zz, theIm, mode )
int xx,yy,zz; /* position in the image */
t_ImagePtr theIm ; /* the image */
int mode; /* access mode to the image */

#endif

{
	int x,y,z;

	x= xx;
	y= yy;
	z= zz;

	if ( mode & MODE_FOLD_MASK ){	/* special side effects mode */


		if ( !_T_FoldMode(&x,theIm->tx,(mode&MODE_X_MASK)>>MODE_X_SHIFT) )
			return( 0.0 );

		if ( !_T_FoldMode(&y,theIm->ty,(mode&MODE_Y_MASK)>>MODE_Y_SHIFT) )
			return( 0.0 );

		if ( !_T_FoldMode(&z,theIm->tz,(mode&MODE_Z_MASK)>>MODE_Z_SHIFT) )
			return( 0.0 );
		}

	return( T_Get3Point( x, y, z, theIm ) );
	}

/*----------------------------------------------------------------------*/

/*
  Get the color of a point in integer images (old function). 

This function has been replaced by T_GetIPoint(). It is maintained for compatibility with earlier versions.

Returns: the exact integer number coded by this voxel. Works only for integer images.
*/

#if defined(_ANSI_)
int T_GetIntPoint( int xx, int yy, int zz, t_ImagePtr theIm )
#else
int T_GetIntPoint(xx, yy, zz, theIm )
int xx,yy,zz; /* position in the image */
t_ImagePtr theIm ; /* the image */

#endif

{
  int plan;
  unsigned char value;
  unsigned short sval;
  short int ssval;
  
  xx= xx % theIm->tx;
  yy= yy % theIm->ty;
  
  if ( xx < 0 ) xx += theIm->tx;
  if ( yy < 0 ) yy += theIm->ty;
  
  if ( zz ){	zz= zz % theIm->tz;
		if ( zz < 0 ) zz += theIm->tz;
		plan= zz * theIm->tx * theIm->ty; }
  else		plan=0;
  
  switch( theIm->type ){
    
  case TYPEIM_BW:
    
    value= 	(unsigned char) theIm->buf[(xx+yy*theIm->tx+plan) >> 3] & 
      (0x80 >> (xx % 8));
    if ( value ) return( 1 ); 
    else		 return( 0 );
    break;
    
  case TYPEIM_256:
    
    return( (unsigned char) theIm->buf[xx+yy*theIm->tx+plan] );
    break;
    
  case TYPEIM_16B:
    
    sval= *( ((unsigned short *) theIm->buf) + xx+yy*theIm->tx+plan );
    return( sval );
    break;
    
  case TYPEIM_S16B:
    
    ssval= *( ((short int*) theIm->buf) + xx+yy*theIm->tx+plan );
    return( ssval );
    break;
    
  default: 
    E_XERR( E_ERRMISMATCH,0,"T_Get3Point","not supported image type" );
    break;
  }
  
  return( 0 );
}

/*----------------------------------------------------------------------*/

/*
  Get the Nth point value of an image.

 This function can be useful for histogram computation, because it is a faster access the pixels/voxels values.

Returns:

A generally normalized intensity value (see T_Get3Point()).
*/

#if defined(_ANSI_)
reeltype T_GetNthPoint( int num, t_ImagePtr theIm )
#else
reeltype T_GetNthPoint(num, theIm )
int num;	/* position in the image */
t_ImagePtr theIm ; /* the image itself */

#endif

{
  switch( theIm->type ){
    
  case TYPEIM_FLOAT:
    
    return( *(((reeltype *) theIm->buf) + num) );
    break;
    
  case TYPEIM_256:
    
    return( (reeltype) ( *(theIm->buf + num) / 255.0 ) );
    break;
    
  default:
    E_XERR( E_ERRMISMATCH, -1.0, "_T_GetNthPoint", "type of images" );
    return(-1.0);
	break;
  }
	return( -1.0 );
}

/*----------------------------------------------------------------------*/

/*
	(centered) linear interpolation of a point in a 2D image.

	This function is based on the hypothesis that the value of the pixel represents	the value of the signal for the center of a regular rectangular grid, and not the value at the vertices of a grid. Such conventions can make half pixel differencies in the result. The origin is the upper left grid cell center.

Returns:
A generally normalized intensity value (see T_Get3Point()).
*/

#if defined(_ANSI_)
reeltype T_Interpol( reeltype xx, reeltype yy, t_ImagePtr theIm )
#else
reeltype T_Interpol(xx, yy, theIm )
reeltype xx,yy; /* position in the image */
t_ImagePtr theIm ; /* the image */

#endif

{
  return( T_Interpol2D( xx, yy, 0.0, theIm ) );
}

/*----------------------------------------------------------------------*/

/*
	(centered) linear 2D interpolation of a point in a 3D image.

	This function is based on the hypothesis that the value of the pixel represents	the value of the signal for the center of a regular rectangular 2D grid, for each image plane in 3D image (and not the value at the vertices of a grid). Such conventions can make half pixel differencies in the result. The origin is the first plane, upper left grid cell center.

Returns:
A generally normalized intensity value (see T_Get3Point()).
*/

#if defined(_ANSI_)
reeltype T_Interpol2D( reeltype xx, reeltype yy, reeltype zz, t_ImagePtr theIm )
#else
reeltype T_Interpol2D(xx, yy, zz, theIm )
reeltype xx,yy;	/* position in the slice */
reeltype zz; /* number of the slice in the 3D image (converted to integer position) */
t_ImagePtr theIm ; /* the image */

#endif

{
  int x,y,z;
  reeltype val1,val2,val3,val4,val5,val6;
  
  xx -= 0.5;		/* go to the center of the pixel */
  yy -= 0.5;
  
  x= (int) xx;
  y= (int) yy;
  z= (int) zz;
  
  val1= T_Get3Point( x, y, z, theIm );
  val2= T_Get3Point( x + 1, y, z, theIm );
  val3= T_Get3Point( x, y + 1, z, theIm );
  val4= T_Get3Point( x + 1, y + 1, z, theIm );
  
  val5= (1 - (xx - x)) * val1 + (xx - x) * val2;
  val6= (1 - (xx - x)) * val3 + (xx - x) * val4;
  
  E_ISERR( 0.0, "T_Interpol2D", NULL );	
  return( (1 - (yy - y)) * val5 + (yy - y) * val6 ); 
}

/*
	(centered) linear 3D interpolation of a point in a 3D image.

	This function is based on the hypothesis that the value of the voxel represents	the value of the signal for the center of a regular rectangular 3D grid (and not the value at the vertices of the grid, see T_Interp3D()). Such conventions can make half pixel differencies in the result. The origin is the first plane, upper left grid cell center.

Returns:
A generally normalized intensity value (see T_Get3Point()).
*/

#if defined(_ANSI_)
reeltype T_Interpol3( reeltype xx, reeltype yy, reeltype zz, t_ImagePtr theIm )
#else
reeltype T_Interpol3(xx, yy, zz, theIm )
reeltype xx,yy,zz; /* floating point position in the image */
t_ImagePtr theIm ; /* the image */

#endif

{
  int x,y,z,plane,pos;
  double val1,val2,val3,val4,val5,val6,val7,val8,val9;
  double val10,val11,val12,val13,val14;
  double a,b,c;
  
  xx -= 0.5;		/* go to the center of the grid */
  yy -= 0.5;
  zz -= 0.5;
  
  x= (int) floor(xx);
  y= (int) floor(yy);
  z= (int) floor(zz);
  
  a= (xx - x);
  b= (yy - y);
  c= (zz - z);
  
  /* Haibin, don't forget to avoid side effects... */
  
  if ( (x < 0) || (y < 0) || (z < 0) ||
      (x >= theIm->tx-1) ||
      (y >= theIm->ty-1) ||
      (z >= theIm->tz-1) ||
      (	(theIm->type != TYPEIM_256) &&
       (theIm->type != TYPEIM_FLOAT) &&
       (theIm->type != TYPEIM_16B) &&
       (theIm->type != TYPEIM_S16B)) ){
    
    val1= (double) T_Get3Point( x  , y  , z, theIm );
    val2= (double) T_Get3Point( x+1, y  , z, theIm );
    val3= (double) T_Get3Point( x  , y+1, z, theIm );
    val4= (double) T_Get3Point( x+1, y+1, z, theIm );
    
    val5= (double) T_Get3Point( x  , y  , z+1, theIm );
    val6= (double) T_Get3Point( x+1, y  , z+1, theIm );
    val7= (double) T_Get3Point( x  , y+1, z+1, theIm );
    val8= (double) T_Get3Point( x+1, y+1, z+1, theIm );
  }
  
  else{ 
    plane= theIm->tx*theIm->ty;
    pos= x + y*theIm->tx + z*plane;
    
    switch( theIm->type ){
      
    case TYPEIM_256: 
      
      { unsigned char *ubyte_pt;
	
	ubyte_pt= ((unsigned char *)(theIm -> buf)) + pos;
	val1= *ubyte_pt / 255.0;
	val2= *(ubyte_pt+1) / 255.0;
	val3= *(ubyte_pt+theIm->tx) / 255.0;
	val4= *(ubyte_pt+theIm->tx+1) / 255.0;
	ubyte_pt += plane;
	val5= *ubyte_pt / 255.0;
	val6= *(ubyte_pt+1) / 255.0;
	val7= *(ubyte_pt+theIm->tx) / 255.0;
	val8= *(ubyte_pt+theIm->tx+1) / 255.0; }
      break;
      
    case TYPEIM_16B:
      
      { unsigned short *ushort_pt;				
	
	ushort_pt = ((unsigned short *)(theIm -> buf)) + pos; 
	val1= *ushort_pt / 65535.0;
	val2= *(ushort_pt+1) / 65535.0;
	val3= *(ushort_pt+theIm->tx) / 65535.0;
	val4= *(ushort_pt+theIm->tx+1) / 65535.0;
	ushort_pt += plane;
	val5= *ushort_pt / 65535.0;
	val6= *(ushort_pt+1) / 65535.0;
	val7= *(ushort_pt+theIm->tx) / 65535.0;
	val8= *(ushort_pt+theIm->tx+1) / 65535.0; }
      break;
      
    case TYPEIM_S16B:
      
      {short int *sshort_pt;				
       
       sshort_pt = ((short int*)(theIm -> buf)) + pos; 
       val1= *sshort_pt / 32767.0;
       val2= *(sshort_pt+1) / 32767.0;
       val3= *(sshort_pt+theIm->tx) / 32767.0;
       val4= *(sshort_pt+theIm->tx+1) / 32767.0;
       sshort_pt += plane;
       val5= *sshort_pt / 32767.0;
       val6= *(sshort_pt+1) / 32767.0;
       val7= *(sshort_pt+theIm->tx) / 32767.0;
       val8= *(sshort_pt+theIm->tx+1) / 32767.0; }
      break;
      
    case TYPEIM_FLOAT:
      
      { float *float_pt;
      
      float_pt = ((float *)(theIm -> buf)) + pos; 
      val1= (double) *float_pt;
      val2= (double) *(float_pt+1);
      val3= (double) *(float_pt+theIm->tx);
      val4= (double) *(float_pt+theIm->tx+1);
      float_pt += plane;
      val5= (double) *float_pt;
      val6= (double) *(float_pt+1);
      val7= (double) *(float_pt+theIm->tx);
      val8=(double)  *(float_pt+theIm->tx+1); }
      break;
      
    default:
      val1= 0.0; val2= 0.0; val3= 0.0; val4= 0.0;
      val5= 0.0; val6= 0.0; val7= 0.0; val8= 0.0;
      break;
    }
  }
  
  val9 = (1-a) * val1 + a * val2;
  val10= (1-a) * val3 + a * val4;
  val11= (1-a) * val5 + a * val6;
  val12= (1-a) * val7 + a * val8;
  
  val13= (1 - b) * val9  + b * val10; 
  val14= (1 - b) * val11 + b * val12; 
  
  E_ISERR( 0.0, "T_Interpol3", NULL );	
  return( (reeltype) ((1 - c) * val13 + c * val14) );
}


/*----------------------------------------------------------------------*/

/*
	Linear 3D interpolation of a point in a 3D image.

	This function is based on the hypothesis that the value of the voxel represents	the value of the signal for the vertices of a regular rectangular 3D grid (and not the value at the center of the grid cell, see T_Interpol(),T_Interpol3D()). Such conventions can make half voxel differencies in the result. The origin is the first plane, upper left grid cell vertex.

Returns:
A generally normalized intensity value (see T_Get3Point()).
*/

#if defined(_ANSI_)
reeltype T_Interp3D( reeltype xx, reeltype yy, reeltype zz, t_ImagePtr theIm )
#else
reeltype T_Interp3D(xx, yy, zz, theIm )
reeltype xx,yy,zz; /* floating point position in the image */
t_ImagePtr theIm ; /* the image */

#endif

{
  return( T_Interpol3( (reeltype) (xx+0.5), (reeltype) (yy+0.5), (reeltype) (zz+0.5), theIm ) );
}

/*----------------------------------------------------------------------*/

/*
	Set the intensity value of a 3D image.
*/

#if defined(_ANSI_)
void T_Set3Point( int xx, int yy, int zz, t_ImagePtr theIm, reeltype color, char mode )
#else
void T_Set3Point(xx, yy, zz, theIm, color, mode )
int xx,yy,zz; /* position in the image */
t_ImagePtr theIm; /* the image */
reeltype color; 
/* 
the intensity in the image. You should generally normalize this value between 0 and 1. For integer images, 0 represents the minimal value (generally 0), and 1 represents the maximal value (for example 255 for a 1 byte/voxel image). It can be signed values for signed images, or any value for floating point images.
*/
char mode ;

#endif

{
  int pos,plan;
  reeltype *realPt;
  unsigned char *ptr;
  unsigned short *sbuf;
  short int *ssbuf;
  unsigned int value;
  int svalue;
  unsigned char bitmask;
  
  if ( mode & MODE_VER ){			/* verify if point in the image		*/
    if ( (xx<0) || (yy<0) || (zz<0) ||
	(xx>=theIm->tx) || (yy>=theIm->ty) || (zz>=theIm->tz) ) return;
  }

	else if ( mode & MODE_FOLD_MASK ){

		if ( !_T_FoldMode(&xx,theIm->tx,(mode&MODE_X_MASK)>>MODE_X_SHIFT) )
			return;

		if ( !_T_FoldMode(&yy,theIm->ty,(mode&MODE_Y_MASK)>>MODE_Y_SHIFT) )
			return;

		if ( !_T_FoldMode(&zz,theIm->tz,(mode&MODE_Z_MASK)>>MODE_Z_SHIFT) )
			return;
		}

	else{
		xx= xx % theIm->tx;
		yy= yy % theIm->ty;
		if ( xx < 0 ) xx += theIm->tx;
		if ( yy < 0 ) yy += theIm->ty;
		}
  
  if ( zz ){
		zz= zz % theIm->tz;
		if ( zz < 0 ) zz += theIm->tz;
		plan= zz * theIm->tx * theIm->ty;
		}
  else plan=0;
  
  mode &= MODE_MASK;
  
  switch( theIm->type ){
    
  case TYPEIM_BW:
    
    bitmask= 0x80 >> (xx % 8);
    switch( mode & 255 ){
      
    case MODE_SET:
      if ( color >= 0.5 )
	theIm->buf[(xx+yy*theIm->tx+plan) >> 3] |= bitmask;
      else	theIm->buf[(xx+yy*theIm->tx+plan) >> 3] &= ~bitmask;
      break;
      
    default: 
      E_XERRN(E_ERRUNKNOWN,"T_Set3Point","unknown color mode"); 
      break;
    }
    break;
    
  case TYPEIM_256:
    
    value= (int) (color * 255);
    
    switch( mode ){
      
    case MODE_SET:
      theIm->buf[ xx + yy * theIm->tx + plan ] = value; break;
      
    case MODE_ADD:
      ptr= (unsigned char *) &theIm->buf[xx+yy*theIm->tx+plan];
      value += (unsigned int) *ptr; 
      if (value < 256) { *ptr= (unsigned char) value; }
      else			 { *ptr= 255; }
      break;
      
    case MODE_ADD2:
      ptr= (unsigned char *) &theIm->buf[xx+yy*theIm->tx+plan];
      value += ((unsigned int) *ptr) * ((unsigned int) *ptr); 
      if (value < 256) { *ptr= (unsigned char) value; }
      else			 { *ptr= 255; }
      break;
      
    case MODE_BUF:
      pos= xx + yy * theIm->tx + plan;
      if ( theIm->buf[ pos ] == 0 )
	theIm->buf[ pos ]= value + MODE_BUFLEVEL;
      else	theIm->buf[ pos ] += value;
      break;
      
    default: 
      E_XERRN(E_ERRUNKNOWN,"T_Set3Point","unknown color mode"); 
      break;
    }
    break;
    
  case TYPEIM_16B:
    
    value= (int) (color * 65535);
    sbuf= ((unsigned short *) theIm->buf) + xx+yy*theIm->tx+plan;
    
    switch( mode ){
      
    case MODE_SET: *sbuf= value; break;
    case MODE_ADD: value += *sbuf; 
      if (value < 65535) { *sbuf= (unsigned short) value; }
      else			   { *sbuf= 65535; }
      break;
    case MODE_ADD2: value += *sbuf * *sbuf; 
      if (value < 65535) { *sbuf= (unsigned short) value; }
      else			   { *sbuf= 65535; }
      break;
    default: 
      E_XERRN(E_ERRUNKNOWN,"T_Set3Point","unknown color mode"); 
      break;
    }
    break;
    
  case TYPEIM_S16B:
    
    svalue= (int) (color * 32767);
    ssbuf= ((short int*) theIm->buf) + xx+yy*theIm->tx+plan;
    
    switch( mode ){
      
    case MODE_SET: *ssbuf= svalue; break;
    case MODE_ADD: svalue += *ssbuf; 
      if (svalue > 32768) {
	*ssbuf= 32767;
      }
      else if (svalue < -32768) {
	*ssbuf= -32767;
      }
      else {
	*ssbuf= (short int)svalue;
      }
      break;
    case MODE_ADD2: svalue += *ssbuf * *ssbuf; 
      if (svalue > 32768) {
	*ssbuf= 32767;
      }
      else if (svalue < -32768) {
	*ssbuf= -32767;
      }
      else {
	*ssbuf= (short int)svalue;
      }
      break;
    default: 
      E_XERRN(E_ERRUNKNOWN,"T_Set3Point","unknown color mode"); 
      break;
    }
    break;
    
  case TYPEIM_FLOAT:
    
    realPt= ((reeltype *) theIm->buf) + xx + yy * theIm->tx + plan;
    switch( mode ){
      
    case MODE_SET:
      *realPt = color;
      break;
      
    case MODE_ADD:
      *realPt += color;
      break;
      
    case MODE_ADD2:
      *realPt += color * color;
      break;
      
    default: 
      E_XERRN(E_ERRUNKNOWN,"T_Set3Point","unknown color mode"); 
      break;
    }
    break;
    
  default: 
    E_XERRN(E_ERRUNKNOWN,"T_Set3Point","unknown image type"); 
    break;
  }
}


/*----------------------------------------------------------------------*/

/*
	Draw a 2D segment in image slices, without clipping.

Starts from real coordinates, to set the error term of the bresenham algorithm. Every pixel of the screen that is intersected by the segment is modified, exept the last pixel, in order not to plot twice the same point. No verification on the position of the ending points of the segment, according to the screen position, is performed.
*/

#if defined(_ANSI_)
void T_Segment3( t_2PointPtr p1, t_2PointPtr p2, int tz, t_ImagePtr theIm, reeltype color, char mode )
#else
void T_Segment3(p1, p2, tz, theIm, color, mode )
t_2PointPtr p1;	/* starting point of the segment */
t_2PointPtr p2; /* ending point of the segment */
int tz;			/* image plane number (for 3D images) */
t_ImagePtr theIm; /* the image */
reeltype color;	/* intensity value */
char mode;		/*  image access mode */

#endif

{	
  int delta_x,delta_y,xx,yy,x1,y1,x2,y2;
  int error;
  
  delta_x= (int) ((p2->x - p1->x) * 100);	
  delta_y= (int) ((p2->y - p1->y) * 100);
  
  xx= x1= (int) p1->x;
  yy= y1= (int) p1->y;
  x2= (int) p2->x;
  y2= (int) p2->y;
  
  /*..................... if the slope is less or equal to 1 ............*/
  
  if ( abs(delta_x) >= abs(delta_y) ) {
    
    if ( delta_x >= 0 ){
      
      if ( delta_y >= 0 ){		/* cas 0 <= delta_y <= delta_x	*/
	
	if (( delta_x == 0 ) &&
	    ( delta_y == 0 ) ){
	  T_Set3Point( xx, yy, tz, theIm, color, mode );
	  return;
	}
	
	error=	delta_x * ( (int)(p1->y) - y1) - 
	  delta_y * ( (int) (p1->x) - x1) - delta_x;
	
	do{
	  T_Set3Point( xx, yy, tz, theIm, color, mode );
	  if ( error < 0 )	{ xx++; error += delta_y; }
	  else				{ yy++; error -= delta_x; }
	} while( xx < x2 );
	/*				if ( yy < y2 ) T_Set3Point( xx, yy, tz, theIm, color, mode ); */
      }
      
      else{						/* cas delta_y <= 0 <= delta_x */
	
	error=	delta_x * ((int)(p1->y) - y1 ) -
	  delta_y * ((int)(p1->x) - x1 );
	
	do{
	  T_Set3Point( xx, yy, tz, theIm, color, mode );
	  if ( error > 0 )	{ xx++; error += delta_y; }
	  else				{ yy--; error += delta_x; }
	} while( xx < x2 );
	/*				if ( yy > y2 ) T_Set3Point( xx, yy, tz, theIm, color, mode ); */
      }
    }
    
    else{
      
      if ( delta_y >= 0 ){		/* cas delta_x <= 0 <= delta_y */
	
	error= -delta_x * ((int)(p1->y) - yy) +
	  delta_y * ((int)(p1->x) - xx) + delta_x;
	
	do{
	  T_Set3Point( xx, yy, tz, theIm, color, mode );
	  if ( error < 0 )	{ xx--; error += delta_y; }
	  else				{ yy++; error += delta_x; }
	} while( xx > x2 );
	/*				if ( yy < y2 ) T_Set3Point( xx, yy, tz, theIm, color, mode ); */
      }
      
      else{						/* cas delta_x <= delta_y <= 0	*/
	
	error= -delta_x * ((int)(p1->y) - yy) +
	  delta_y * ((int)(p1->x) - xx);
	
	do{
	  T_Set3Point( xx, yy, tz, theIm, color, mode );
	  if ( error > 0 )	{ xx--; error += delta_y; }
	  else				{ yy--; error -= delta_x; }
	} while ( xx > x2 );
	/*				if ( yy > y2 ) T_Set3Point( xx, yy, tz, theIm, color, mode );*/
      }
    }
  }
  
  /*......................................................................*/
  /*		The following case comes mecanically when y is replaced by x	*/
  /*	and x replaced by y													*/
  /*......................................................................*/
  
  else{
    
    if ( delta_y >= 0 ){
      
      if ( delta_x >= 0 ){		/* cas 0 <= delta_x <= delta_y	*/
	
	error=	delta_y * ( (int)(p1->x) - x1) - 
	  delta_x * ( (int)(p1->y) - y1) - delta_y;
	
	do{
	  T_Set3Point( xx, yy, tz, theIm, color, mode );
	  if ( error < 0 )	{ yy++; error += delta_x; }
	  else				{ xx++; error -= delta_y; }
	} while( yy < y2 );
	/*				if ( xx < x2 ) T_Set3Point( xx, yy, tz, theIm, color, mode ); */
      }
      
      else{						/* cas delta_x <= 0 <= delta_y */
	
	error=	delta_y * ( (int)(p1->x) - x1 ) -
	  delta_x * ( (int)(p1->y) - y1 );
	
	do{
	  T_Set3Point( xx, yy, tz, theIm, color, mode );
	  if ( error > 0 )	{ yy++; error += delta_x; }
	  else				{ xx--; error += delta_y; }
	} while( yy < y2 );
	/*				if ( xx > x2 ) T_Set3Point( xx, yy, tz, theIm, color, mode ); */
      }
    }
    
    else{
      
      if ( delta_x >= 0 ){		/* cas delta_y <= 0 <= delta_x */
	
	error= -delta_y * ( (int)(p1->x) - xx) +
	  delta_x * ( (int)(p1->y) - yy) + delta_y;
	
	do{
	  T_Set3Point( xx, yy, tz, theIm, color, mode );
	  if ( error < 0 )	{ yy--; error += delta_x; }
	  else				{ xx++; error += delta_y; }
	} while( yy > y2 );
	/*				if ( xx < x2 ) T_Set3Point( xx, yy, tz, theIm, color, mode ); */
      }
      
      else{						/* cas delta_y <= delta_x <= 0	*/
	
	error= -delta_y * ( (int)(p1->x) - xx) +
	  delta_x * ( (int)(p1->y) - yy);
	
	do{
	  T_Set3Point( xx, yy, tz, theIm, color, mode );
	  if ( error > 0 )	{ yy--; error += delta_x; }
	  else				{ xx--; error -= delta_y; }
	} while ( yy > y2 );
	/*				if ( xx > x2 ) T_Set3Point( xx, yy, tz, theIm, color, mode ); */
      }
    }
  }
}

/*----------------------------------------------------------------------*/

/*
	1D sampling along a segment in a 2D image.

It extract a 1D buffer corresponding to the values of the image around that point, and according to that direction.
*/

#if defined(_ANSI_)
void T_ReSampleSeg( t_ImagePtr theIm, reeltype *buf, int size, t_2PointPtr pos, t_2PointPtr dir, reeltype length )
#else
void T_ReSampleSeg(theIm, buf, size, pos, dir, length )
t_ImagePtr theIm; /* the 2D image */
reeltype *buf; /* the 1D output buffer */
int size; /* size of the 1D output buffer */
t_2PointPtr pos; /* 2D position in the image */
t_2PointPtr dir; /* direction of the sampling */
reeltype length ; /* length of the sampling segment */

#endif

{
  int i;
  t_2Point incr,depPt;
  
  if (size < 2) return;
  
  depPt.x= (reeltype) (pos->x - length / 2.0 * dir->x);
  depPt.y= (reeltype) (pos->y - length / 2.0 * dir->y);
  incr.x= dir->x * length / (size - 1);
  incr.y= dir->y * length / (size - 1);
  
  for( i=0; i<size; i++ ){
    buf[i]= T_Interpol( depPt.x, depPt.y, theIm );
    depPt.x += incr.x;
    depPt.y += incr.y;
  }
}

/*----------------------------------------------------------------------*/

/*
	Get simultaneously the 8 values in a 8-cell of a 3D image.
*/

#if defined(_ANSI_)
void T_GetCube( int i, int j, int k, t_ImagePtr theIm, reeltype tb[8] )
#else
void T_GetCube(i, j, k, theIm, tb)
int i,j,k;			/* position of the first vertex of the 8-cell */
t_ImagePtr theIm;	/* the image */
reeltype tb[8] ;	/* the 8-cell values */

#endif

{
  tb[0]= T_Get3Point( i, j, k, theIm );
  tb[1]= T_Get3Point( i+1, j, k, theIm );
  tb[2]= T_Get3Point( i, j+1, k, theIm );
  tb[3]= T_Get3Point( i+1, j+1, k, theIm );
  
  tb[4]= T_Get3Point( i, j, k+1, theIm );
  tb[5]= T_Get3Point( i+1, j, k+1, theIm );
  tb[6]= T_Get3Point( i, j+1, k+1, theIm );
  tb[7]= T_Get3Point( i+1, j+1, k+1, theIm );
}

/*----------------------------------------------------------------------*/

/*
	Set the intensity of an integer 3D image.
*/

#if defined(_ANSI_)
void T_SetIPoint( int xx, int yy, int zz, t_ImagePtr theIm, int color, char mode )
#else
void T_SetIPoint(xx, yy, zz, theIm, color, mode )
int xx,yy,zz;		/* position in the image */
t_ImagePtr theIm;	/* the image */
int color;			/* intensity value */
char mode ;			/* image access mode */

#endif

{
  unsigned char bitmask;
  int plan;
  unsigned char *buf1;
  unsigned short *buf2;
  short int *buf3;
  int addr;
  
  if ( mode & MODE_VER ){			/* verify if point in the image		*/
    if ( (xx<0) || (yy<0) || (zz<0) ||
	(xx>=theIm->tx) || (yy>=theIm->ty) || (zz>=theIm->tz) ) return;
  }
  else{
    xx= xx % theIm->tx;
    yy= yy % theIm->ty;
    zz= zz % theIm->tz;
    if ( xx < 0 ) xx += theIm->tx;
    if ( yy < 0 ) yy += theIm->ty;
  }
  
  if ( zz ){	zz= zz % theIm->tz;
		if ( zz < 0 ) zz += theIm->tz;
		plan= zz * theIm->tx * theIm->ty; }
  else		plan=0;
  
  addr= xx + yy * theIm->tx + plan;
  mode &= MODE_MASK;
  
  switch( theIm->type ){
    
  case TYPEIM_BW:
    
    bitmask= 0x80 >> (xx % 8);
    buf1= (unsigned char *) (theIm->buf + (addr >> 3));
    
    switch( mode & 255 ){
      
    case MODE_SET:	if ( color )	*buf1 |= bitmask;
    else			*buf1 &= ~bitmask; break;
      
    default: 
      E_XERRN(E_ERRUNKNOWN,"T_SetIPoint","unknown color mode"); 
      break;
    }
    break;
    
    
  case TYPEIM_256:
    
    buf1= (unsigned char *) (theIm->buf + addr);
    switch( mode ){
    case MODE_SET: *buf1= color; break;
    case MODE_ADD:
      color += *buf1;
      if (color < 256) *buf1= (unsigned char) color;
      else			 *buf1= 255;
      break;
    case MODE_ADD2:
      color += *buf1 * *buf1;
      if (color < 256) *buf1= (unsigned char) color;
      else			 *buf1= 255;
      break;
    default:
      E_XERRN(E_ERRUNKNOWN,"T_SetIPoint","unknown color mode"); 
      break;
    }
    break;
    
    
  case TYPEIM_16B:
    
    buf2= ((unsigned short *) theIm->buf) + addr;
    switch( mode ){
    case MODE_SET: *buf2= color; break;
    case MODE_ADD:
      color += *buf2;
      if (color < 0x10000) *buf2= (unsigned short) color;
      else			 	 *buf2= 0xFFFF;
      break;
    case MODE_ADD2:
      color += *buf2 * *buf2;
      if (color < 0x10000) *buf2= (unsigned short) color;
      else			 	 *buf2= 0xFFFF;
      break;
    default:
      E_XERRN(E_ERRUNKNOWN,"T_SetIPoint","unknown color mode"); 
      break;
    }
    break;
    
  case TYPEIM_S16B:
    
    buf3= ((short int*) theIm->buf) + addr;
    switch( mode ){
    case MODE_SET: *buf3= color; break;
    case MODE_ADD:
      color += *buf3;
      if (color > 32768) {
	*buf3= 32767;
      }
      else if (color < -32768) {
	*buf3= -32767;
      }
      else {
	*buf3= (short int)color;
      }
      break;
    case MODE_ADD2:
      color += *buf3 * *buf3;
      if (color > 32768) {
	*buf3= 32767;
      }
      else if (color < -32768) {
	*buf3= -32767;
      }
      else {
	*buf3= (short int)color;
      }
      break;
    default:
      E_XERRN(E_ERRUNKNOWN,"T_SetIPoint","unknown color mode"); 
      break;
    }
    break;
    
  default: 
    E_XERRN(E_ERRUNKNOWN,"T_SetIPoint","unknown image type"); 
    break;
  }
}

/*----------------------------------------------------------------------*/
/*	Get a point with an integer value									*/
/*----------------------------------------------------------------------*/

/*
  Get the color of a point for integer images. 

This function has been replaced by T_GetIPoint(). It is maintained for compatibility with earlier versions.

Returns: the exact integer number coded by this voxel. Works only for integer images.
*/

#if defined(_ANSI_)
int T_GetIPoint( int xx, int yy, int zz, t_ImagePtr theIm )
#else
int T_GetIPoint(xx, yy, zz, theIm )
int xx,yy,zz;		/* position in the image */
t_ImagePtr theIm ;	/* the image */

#endif

{
  unsigned char bitmask;
  int plan;
  unsigned char *buf1;
  unsigned short *buf2;
  short int *buf3;
  int addr;
  
  xx= xx % theIm->tx;
  yy= yy % theIm->ty;
  zz= zz % theIm->tz;
  if ( xx < 0 ) xx += theIm->tx;
  if ( yy < 0 ) yy += theIm->ty;
  if ( zz ){	zz= zz % theIm->tz;
		if ( zz < 0 ) zz += theIm->tz;
		plan= zz * theIm->tx * theIm->ty; }
  else		plan=0;
  addr= xx + yy * theIm->tx + plan;
  
  switch( theIm->type ){
    
  case TYPEIM_BW:
    
    bitmask= 0x80 >> (xx % 8);
    buf1= (unsigned char *) (theIm->buf + (addr >> 3));
    if ( *buf1 & bitmask )	return(1);
    else				  	return(0);
    break;
    
    
  case TYPEIM_256:
    
    buf1= (unsigned char *) (theIm->buf + addr);
    return( *buf1 );
    break;
    
    
  case TYPEIM_16B:
    
    buf2= ((unsigned short *) theIm->buf) + addr;
    return( *buf2 );
    break;
    
  case TYPEIM_S16B:
    
    buf3= ((short int*) theIm->buf) + addr;
    return( *buf3 );
    break;
    
  default: 
    E_XERR( E_ERRUNKNOWN, 0, "T_GetIPoint", "unknown image mode" ); 
    break;
  }
  
  return(0);
}


/*----------------------------------------------------------------------*/

/*
	Apply a kernel filter to a point in a 3D image

Returns:

The value of filtered point.
*/

#if defined(_ANSI_)
reeltype T_GetFiltredPt( int x, int y, int z, t_ImagePtr theIm, reeltype *filter, int dx, int dy, int dz )
#else
reeltype T_GetFiltredPt(x, y, z, theIm, filter, dx, dy, dz )
int x,y,z;			/* position in the image */
t_ImagePtr theIm;	/* the image */
reeltype *filter;	/* buffer containing the 3D kernel values */
int dx,dy,dz ;		/* size of the 3D kernel */

#endif

{
  int i,j,k,mx,my,mz,Mx,My,Mz;
  reeltype value;
  
  Mx= dx/2; My= dy/2; Mz= dz/2;
  if (dx % 2) mx= -dx/2;	else mx= -dx/2+1;
  if (dy % 2) my= -dy/2;	else my= -dy/2+1;
  if (dx % 2) mz= -dz/2;	else mz= -dz/2+1;
  
  value= 0.0;
  for( k= mz; k<= Mz; k++ )
    for( j= my; j<= My; j++ )
      for( i= mx; i<= Mx; i++ ){
	value += T_Get3Point( x+i, y+j, z+k, theIm ) * (*filter++);
	E_ISERR( 0.0, "T_GetFiltredPt", NULL );
      }
  
  return( value );
}

/*----------------------------------------------------------------------*/
/************************************************************************/
/*
	lowImages
	
	Low level operations on image processing

		(c) Jean-Philippe THIRION -- INRIA --
*/
/************************************************************************/

#ifdef INRIMAGE
#include <inrimage/image.h>
#endif

/*----------------------------------------------------------------------*/

/*
	Gives the size in bytes of a 3D image
*/

#if defined(_ANSI_)
int T_ImageSize( t_ImagePtr theIm )
#else
int T_ImageSize(theIm )
t_ImagePtr theIm ;	/* the image */

#endif

{
  switch( theIm->type ){
    
  case TYPEIM_BW:
    
    return( ((theIm->tx * theIm->ty) >> 3) * theIm->tz  );
    break;
    
  case TYPEIM_256:
    
    return( theIm->tx * theIm->ty * theIm->tz );
    break;
    
  case TYPEIM_16B:
  case TYPEIM_S16B:
    
    return( theIm->tx * theIm->ty * theIm->tz * 2 );
    break;
    
  case TYPEIM_FLOAT:
    
    return( theIm->tx * theIm->ty * sizeof(reeltype) * theIm->tz );
    break;
    
  default:
    E_XERR( E_ERRUNKNOWN, 0, "T_ImageSize", "Unknown image type" );
    break;
  }
  
  return( 0 );
}

/*----------------------------------------------------------------------*/

/*
	Allocate the buffer for the image

	If the image buffer has not been allocated yet, it performs the memory allocation for the image buffer. This is a convenient way to keep the information about images (in the image header), but to allocate or free the image buffer, which can be very huge, only when it is needed. See T_FreeImage() for the free operation, and T_Init3Image() for the image header initialization.
*/

#if defined(_ANSI_)
void T_AllocImage( t_ImagePtr theIm )
#else
void T_AllocImage(theIm )
t_ImagePtr theIm ;	/* the image */

#endif

{
  int size;
  
  if ( theIm->buf != NULL ) return;
  size= T_ImageSize( theIm );
  theIm->buf= T_Alloc((unsigned) size );
  E_ISERRN("T_AllocImage", NULL);
  
  (void) memset( theIm->buf, 0, size );
}

/*----------------------------------------------------------------------*/

/*
	Set the image voxels to zero

This function does not free the image buffer, but only set its values to zero. See T_FreeImage() for the image buffer destruction.
*/

#if defined(_ANSI_)
void T_ClearImage( t_ImagePtr theIm )
#else
void T_ClearImage(theIm )
t_ImagePtr theIm ;	/* the image to clear */

#endif

{
  int size;
  
  if ( theIm->buf == NULL ) return;
  
  size= T_ImageSize( theIm );
  (void) memset( theIm->buf, 0, size );
}

/*----------------------------------------------------------------------*/

/*
	Set the image voxels to a given value
*/

#if defined(_ANSI_)
void T_SetImage( t_ImagePtr theIm, reeltype value )
#else
void T_SetImage(theIm, value )
t_ImagePtr theIm;	/* the image */
reeltype value ;	/* the intensity value, generally normalized between 0 and 1 (see T_Set3Point()). */

#endif

{
  int i,j,k;
  
  if ( theIm->buf == NULL ){
    T_Warn("image not allocated in T_SetImage", NULL );
    return;
  }
  
  for( k=0; k< theIm->tz; k++ )
    for( j=0; j< theIm->ty; j++ )
      for( i=0; i< theIm->tx; i++ ){
	T_Set3Point( i, j, k, theIm, value, MODE_SET );
	E_ISERRN("T_SetImage", NULL);
      }
}

/*----------------------------------------------------------------------*/

/*
  Compute the minimal and maximal values of an image
*/

#if defined(_ANSI_)
void T_ImFloatBound( t_ImagePtr theIm, reeltype *v_min, reeltype *v_max )
#else
void T_ImFloatBound(theIm, v_min, v_max )
t_ImagePtr theIm;	/* the image */
reeltype *v_min;	/* a pointer toward a reeltype which receive the minimal value */
reeltype *v_max;	/* a pointer toward a reeltype which receive the maximal value */

#endif

{
  int j,k;
  register reeltype value,mini,maxi,*buf;
  register int i;
  
  /* mini= (reeltype) INFINITY;
     maxi= (reeltype) -INFINITY; */
  mini= maxi = T_Get3Point( 0,0,0, theIm );
  buf = (reeltype *) theIm->buf;
  
  if ( theIm->type ==  TYPEIM_FLOAT ){
    
    buf = (reeltype *) theIm->buf;
    
    for( i= theIm->tx * theIm->ty * theIm->tz; i>0; i-- ){
      
      value= *buf++;
      if ( value < mini ) mini= value;
      if ( value > maxi ) maxi= value;
    }
  }
  
  else{
    for( k=0; k< theIm->tz; k++ )
      for( j=0; j< theIm->ty; j++ )
	for( i=0; i< theIm->tx; i++ ){
	  value= T_Get3Point( i,j,k, theIm );
	  if ( value < mini ) mini= value;
	  if ( value > maxi ) maxi= value;
	}
  }
  
  *v_min= mini;
  *v_max= maxi;
}

/*----------------------------------------------------------------------*/

/*
	Convert an image into an 8 bits/pixel image
*/

#if defined(_ANSI_)
void T_ConvIm256( t_ImagePtr theIm )
#else
void T_ConvIm256(theIm )
t_ImagePtr theIm ;	/* the image to convert */

#endif

{
  t_Image		auxIm;
  reeltype 	v_min,v_max;
  
  register 	reeltype min,delta,*buf_float;
  register 	unsigned char *buf_byte;
  register	int i;
  
  switch( theIm->type ){
    
  case TYPEIM_256: break;
  case TYPEIM_FLOAT:
    
    T_ImFloatBound( theIm, &v_min, &v_max );
    E_ERRN( v_max - v_min < EPSY, E_ERRDIV,
	   "T_ConvIm256", "flat histogram");
    
    delta= 255 / ( v_max - v_min );
    min= v_min;
    
    T_CopyHeaderIm( theIm, &auxIm );
    auxIm.type= TYPEIM_256;
    
    T_AllocImage( &auxIm );
    E_ISERRN("T_ConvIm256", NULL);
    
    buf_float	= (reeltype *)		theIm->buf;
    buf_byte	= (unsigned char *) auxIm.buf;
    
    for( i= theIm->tx * theIm->ty * theIm->tz; i>0; i-- )
      *buf_byte++ = (unsigned char) (((*buf_float++)-min)*delta);
    
    free( theIm->buf );
    theIm->buf= auxIm.buf;
    theIm->type= TYPEIM_256;
    auxIm.buf= NULL;
    break;
    
  default:
    E_XERRN( E_ERRUNKNOWN, "T_ConvIm256", "Unknown image type" );
    break;
  }
}

/*----------------------------------------------------------------------*/

/*
Free the buffer of an image

The image buffer and image header can be processed independently. See T_AllocImage() for image buffer allocation and T_Init3Image() for the image header initialization.
*/

#if defined(_ANSI_)
void T_FreeImage( t_ImagePtr theIm )
#else
void T_FreeImage(theIm )
t_ImagePtr theIm ;	/* the image */

#endif

{
  if ( theIm->buf != NULL ){
    (void) free( theIm->buf );
    theIm->buf= NULL;
  }
}

/*----------------------------------------------------------------------*/

/*
	Initialize a 2D image header (old function).

	This function is now replaced with T_Init3Image(), which can be used to
initialize also 3D images.
*/

#if defined(_ANSI_)
void T_InitImage( t_ImagePtr theIm, int tx, int ty, char type, char *name )
#else
void T_InitImage(theIm, tx, ty, type, name )
t_ImagePtr theIm;
int tx;			/* number of vertical columns in the image */
int ty;			/* number of horizontal lines in the image */
char type;		/* type of image */
char *name ;	/* file name of the image on disk (or NULL) */

#endif

{
  T_Init3Image( theIm, tx, ty, 1, type, name );
}

/*----------------------------------------------------------------------*/

/*
	Initialize a 3D image header.

You should absolutly initialize image headers before using them. In particular, the image buffer pointer of the image is set to NULL by the initialization procedure, which means by convention that the image buffer is not allocated (see T_AllocImage()). The normal course is to initialize the image header, to allocate the image buffer with T_AllocImage()) or to read it from disk with T_ReadInrim(), to process the image, eventually to write it on disk with T_WriteInrim(), and atlast to destroy it with T_FreeImage().
*/

#if defined(_ANSI_)
void T_Init3Image( t_ImagePtr theIm, int tx, int ty, int tz, char type, char *name )
#else
void T_Init3Image(theIm, tx, ty, tz, type, name )
t_ImagePtr theIm;
int tx;
int ty;
int tz;
char type;
char *name ;

#endif

{
  if ( name == NULL )	theIm->fi.name[0]= 0;
  else 
    if ( theIm->fi.name != name)
      strncpy( theIm->fi.name, name, B_FNAME_LENGTH );
  
  theIm->fi.type= TYPEFI_IMAGE;
  
  if ( tx < 1 ) 	theIm->tx= 1;
  else			theIm->tx= tx;
  if ( ty < 1 ) 	theIm->ty= 1;
  else			theIm->ty= ty;
  if ( tz < 1 ) 	theIm->tz= 1;
  else			theIm->tz= tz;
  
  theIm->type		= type;
  theIm->buf		= NULL;
  theIm->sx = 1.0;
  theIm->sy = 1.0;
  theIm->sz = 1.0;
  
}

/*----------------------------------------------------------------------*/

/*
	Set the disk file name associated to the image.

The image header must be initialized with T_Init3Image() before any operations on images, and the name of a disk file, associated to this image, can be defined then. The present function is used to change this file name. It is better to to this that working directly with the image structure, which can be subject to changes. One reason to use the present function is to read an image with T_ReadInrim(), to change its value with an image function, to change the file name and to write the result into this new file using T_WriteInrim(). See also T_Write2Inrim().
*/

#if defined(_ANSI_)
void T_SetImName( t_ImagePtr theIm, char *name )
#else
void T_SetImName(theIm, name )
t_ImagePtr theIm;	/* the image header */
char *name ;		/* the new disk file name */

#endif

{
  strncpy( theIm->fi.name, name, B_FNAME_LENGTH );
}

/*----------------------------------------------------------------------*/

/*
	Get the disk file name associated to the image.

The image header must be initialized with T_Init3Image() before any operations on images, and the name of a disk file, associated to this image, can be defined then. See also T_SetImName().
*/

#if defined(_ANSI_)
void T_GetImName( t_ImagePtr theIm, char *name )
#else
void T_GetImName(theIm, name )
t_ImagePtr theIm;	/* the image header */
char *name ;		/* to get the disk file name (output) */

#endif

{
	if ( name == NULL ) return;

	strncpy( name, theIm->fi.name, B_FNAME_LENGTH );
	}

/*----------------------------------------------------------------------*/

/*
	Set the voxel size of the image.

The image header must be initialized with T_Init3Image() before any operations on images, where the voxel size is set to 1.0. The present function is used to change these sizes. 
*/

#if defined(_ANSI_)
void T_SetVoxelSize( t_ImagePtr theIm, reeltype sx, reeltype sy, reeltype sz)
#else
void T_SetVoxelSize(theIm, sx, sy, sz )
t_ImagePtr theIm;	/* the image header */
reeltype sx,sy,sz;      /* the new voxel sizes */

#endif

{
  theIm->sx = sx;
  theIm->sy = sy;
  theIm->sz = sz;
}

/*----------------------------------------------------------------------*/

/*
	Get the voxel sizes of the image.

The image header must be initialized with T_Init3Image() before any operations on images.
*/

#if defined(_ANSI_)
void T_GetVoxelSize( t_ImagePtr theIm, reeltype *sx, reeltype *sy, reeltype *sz)
#else
void T_GetVoxelSize(theIm, sx, sy, sz )
t_ImagePtr theIm;	/* the image header */
reeltype *sx,*sy,*sz;   /* the voxel sizes */

#endif

{
  E_ERRN((sx == NULL) || (sy == NULL) || (sz == NULL) , E_ERRNULL,
	   "T_GetVoxelSize", "Null pointer");
  *sx = theIm->sx;
  *sy = theIm->sy;
  *sz = theIm->sz;

}

/*----------------------------------------------------------------------*/
#ifdef INRIMAGE
/*----------------------------------------------------------------------*/

/*
	Write an image on disk, with inrimage format.

	The name of the file has been given by the initialization function T_Init3Image(), or has been changed to a new disk file name by using T_SetImName(). See also T_Write2Inrim() to write an image with a given name.
*/

#ifndef NO_PROTO
static void T_WriteInrimLow( t_ImagePtr theIm, char *l1, char *l2 );
#else 
static void T_WriteInrimLow();
#endif	/* NO_PROTO */

#if defined(_ANSI_)
void T_WriteInrim( t_ImagePtr theIm )
#else
void T_WriteInrim(theIm )
t_ImagePtr theIm ;	/* the image */

#endif

{ 
  T_WriteInrimLow( theIm, "cx", "" ); 
}

#if defined(_ANSI_)
void T_WriteInrimLoop( t_ImagePtr theIm )
#else
void T_WriteInrimLoop(theIm )
t_ImagePtr theIm ;

#endif

{ 
  T_WriteInrimLow( theIm, "sx", "dx" ); 
}

/*----------------------------------------------------------------------*/
/*	write a t image on the disk, using the inrimage format				*/
/*	low level internal routine.											*/
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
static void T_WriteInrimLow( t_ImagePtr theIm, char *l1, char *l2 )
#else
static void T_WriteInrimLow(theIm, l1, l2 )
t_ImagePtr theIm;
char *l1;
char *l2 ;

#endif

{
  extern struct image *image_();
  
  struct nf_fmt theFmt;
  struct image *inrIm;
  
  if ( theIm->buf == NULL ){
    sprintf( tmpStr1, "void image %s", theIm->fi.name );
    E_XERRN( E_ERRNULL, "T_WriteInrimLow", tmpStr1 );
  }
  
  theFmt.lfmt[0]= theIm->tx;
  theFmt.lfmt[1]= theIm->ty * theIm->tz;
  theFmt.lfmt[4]= theIm->tx;	/* number of vectors by line			*/
  theFmt.lfmt[5]= theIm->ty;	/* number of lines per plans	*/
  theFmt.lfmt[6]= 1;			/* number of colors of the pixels		*/
  theFmt.lfmt[7]= theIm->tz;	/* number of plans of the image		*/
  theFmt.lfmt[8]= 0;			/* scale factor of the image			*/
  
  theFmt.offsets[0]= 0;		/* offset in x,y and z					*/
  theFmt.offsets[1]= 0;
  theFmt.offsets[2]= 0;
  
  theFmt.maille= 0;			/* square network						*/
  theFmt.bias= 0.0;
  theFmt.scale= 0.0;
  
  switch( theIm->type ){
    
  case TYPEIM_BW:
    theFmt.lfmt[2]= -1;	/* number of bits for point encoding	        */
    theFmt.lfmt[3]= -1;	/* packed bit by bit				*/
    break;
    
  case TYPEIM_256:
    theFmt.lfmt[2]= 1;
    theFmt.lfmt[3]= 0;	/* packed bit by bit			        */
    theFmt.lfmt[8]= 200;    /* scale factor of the image			*/
    break;
    
  case TYPEIM_16B:
    theFmt.lfmt[2]= 2;
    theFmt.lfmt[3]= 0;	/* packed bit by bit				*/
    theFmt.lfmt[8]= 200;    /* scale factor of the image			*/
    break;
    
  case TYPEIM_S16B:
    theFmt.lfmt[2]= 2;
    theFmt.lfmt[3]= 0;	/* packed bit by bit				*/
    theFmt.lfmt[8]= -200;    /* scale factor of the image			*/
    break;
    
  case TYPEIM_FLOAT:
    /*		T_ConvIm256( theIm );	*/
    theFmt.lfmt[2]= 4;
    theFmt.lfmt[3]= 1;	/* packed bit by bit				 */
    break;
    
  default:
    sprintf( tmpStr1, "Unsupported image type (file %s)", 
	    theIm->fi.name); 
    E_XERRN( E_ERRUNKNOWN, "T_WriteInrimLow", tmpStr1 ); 
    break;
  }
  
  inrIm= image_( theIm->fi.name, l1, l2, &theFmt ); 
  c_ecr( inrIm, theIm->ty * theIm->tz, theIm->buf );
  fermnf_( &inrIm );
}

/*----------------------------------------------------------------------*/

/*
	Write an inrimage image on disk, with a given image name.

The image header must be initialized with T_Init3Image() before any operations on images, and the name of a disk file, associated to this image, can be defined then. It can be more convenient to simply use this default image file name, with the function T_WriteInrim().
*/

#if defined(_ANSI_)
void T_Write2Inrim( t_ImagePtr theIm, char *name )
#else
void T_Write2Inrim(theIm, name )
t_ImagePtr theIm;	/* the image */
char *name ;		/* disk file name */

#endif

{
	char name[B_FNAME_LENGTH];

	T_GetImName( theIm, name );
	T_WriteInrim( theIm );
	T_SetImName( theIm, name );
  	E_ISERRN("T_Write2Inrim", NULL);
	}

/*----------------------------------------------------------------------*/

/*
	Read an inrimage image from disk.

The image header must be initialized with T_Init3Image() before any operations on images, and the name of a disk file, associated to this image, can be defined then. See also T_WriteInrim() to write inrimage images on disk.
*/

#if defined(_ANSI_)
void T_ReadInrim( t_ImagePtr theIm )
#else
void T_ReadInrim(theIm )
t_ImagePtr theIm ;	/* the image */

#endif

{
  extern struct image *image_();
  
  struct nf_fmt theFmt;
  struct image *inrIm;
  
  inrIm= image_( theIm->fi.name, "ex", "", &theFmt ); 
  
  theIm->type= TYPEIM_UNKNOWN;
  
  if (	( (theFmt.lfmt[2] == -8) ||
	 (theFmt.lfmt[2] == 1) ) &&
      (theFmt.lfmt[6] == 1 ) ) theIm->type= TYPEIM_256;
  
  if (	( (theFmt.lfmt[2] == -16) || (theFmt.lfmt[2] == 2) ) &&
      (theFmt.lfmt[6] == 1 ) && (theFmt.lfmt[8] == 200 ) ) theIm->type= TYPEIM_16B;
  
  if (	( (theFmt.lfmt[2] == -16) || (theFmt.lfmt[2] == 2) ) &&
      (theFmt.lfmt[6] == 1 ) && (theFmt.lfmt[8] == -200 ) ) theIm->type= TYPEIM_S16B;
  
  if (	(theFmt.lfmt[2] == -1) &&
      (theFmt.lfmt[3] == -1) &&
      (theFmt.lfmt[6] == 1 ) ) theIm->type= TYPEIM_BW;
  
  if (	(theFmt.lfmt[2] == 4) &&
      (theFmt.lfmt[3] == 1) &&
      (theFmt.lfmt[6] == 1 ) ) theIm->type= TYPEIM_FLOAT;
  
  if( theIm->type == TYPEIM_UNKNOWN ){
    sprintf( tmpStr1, "unknown image type in file %s", theIm->fi.name );
    E_XERRN( E_ERRUNKNOWN, "T_ReadInrim", tmpStr1 );
  }
  
  /*
    theIm->tx= theFmt.lfmt[0];
    theIm->ty= theFmt.lfmt[1] / theFmt.lfmt[7];
    theIm->tz= theFmt.lfmt[7];
    */
  theIm->tx= theFmt.lfmt[4];
  theIm->ty= theFmt.lfmt[5];
  theIm->tz= theFmt.lfmt[7];
  
  T_AllocImage( theIm );
  E_ISERRN("T_ReadInrim", NULL );
  
  c_lect( inrIm, theIm->ty * theIm->tz, theIm->buf );
  fermnf_( &inrIm );
}

/*----------------------------------------------------------------------*/
#endif /* INRIMAGE */
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
static void E_Swap4octets( char *tab1 /* buffer to be swapped */, 
		   char *tab2 /* result buffer */, 
		   int taille /* size of the buffer, when considered as char* */ )
#else 
static void E_Swap4octets( tab1, tab2, taille )
char *tab1; /* buffer to be swapped */
char *tab2; /* result buffer */
int taille; /* size of the buffer, when considered as char* */
#endif
{
    register unsigned int *theBuf = (unsigned int *)tab1;
    register unsigned int i;
    register unsigned int *resBuf = (unsigned int *)tab2;
    register int nb = taille/4;
    
    while ( nb-- > 0 ) {
      i = *theBuf++;
      *resBuf++ = (i << 24) | ((i & 0xff00) << 8) | ((i >> 8) & 0xff00) | ((i >> 24) & 0xff);
    }

}


/*
	Read an inrimage image from disk.

The image header must be initialized with T_Init3Image() before any operations on images, and the name of a disk file, associated to this image, can be defined then. See also T_WriteInrim() to write inrimage images on disk.
*/

#if defined(_ANSI_)
void T_ReadInrim( t_ImagePtr theIm )
#else
void T_ReadInrim(theIm )
t_ImagePtr theIm ;	/* the image */

#endif

{
  /* int fd */
  FILE* fd;
  int size;
  char aux[4];
  
  T_ReadInrimHead( theIm );
  /*   fd= open( theIm->fi.name, O_RDONLY, 0644 ); */
  /* if ( fd < 0 ) E_XERRN( E_ERRO, "T_MyReadInrim", theIm->fi.name ); */
  fd = fopen( theIm->fi.name, "rb");
  if ( fd == NULL ) 
    E_XERRN( E_ERRO, "T_MyReadInrim", theIm->fi.name );
 
  
/*   status= lseek( fd, 252L, SEEK_SET ); */
/*   if ( status != 252L )  */
/*     E_SETERR( E_ERRUNKNOWN, "Seeking in file", theIm->fi.name); */
  if ( fseek( fd, 252L, SEEK_SET) == EOF ) 
    E_SETERR( E_ERRUNKNOWN, "Seeking in file", theIm->fi.name);
  

/*   status= read( fd, aux, 4 ); */
/*   if ( status < 0 ) E_SETERR(E_ERRR, "reading in file", theIm->fi.name); */
       

  fread( (void *) aux, sizeof(char), 4, fd );
  if ( ferror(fd) )
    E_SETERR(E_ERRR, "reading in file", theIm->fi.name);
 
  if ( strncmp( aux, "##}", 3 ) != 0 ){

/* 	Status= Lseek( Fd, 1020l, Seek_Set ); */
/*   	If ( Status != 1020l ) E_Seterr( E_Errunknown, "Seeking In File", */
/*                  Theim->Fi.Name); */

/*   	Status= read( Fd, Aux, 4 ); */
/*   	If ( Status < 0 ) E_Seterr(E_Errr, "Reading In File", Theim->Fi.Name); */

  	if ( fseek( fd, 1020L, SEEK_SET ) == EOF ) 
	  E_SETERR( E_ERRUNKNOWN, "Seeking in file", theIm->fi.name);

	fread( (void *) aux, sizeof(char), 4, fd );
  	if ( ferror(fd)   ) 
	  E_SETERR(E_ERRR, "reading in file", theIm->fi.name);
 
 
  	E_ERRN( strncmp( aux, "##}", 3 ) != 0, E_ERRFATAL, 
		"Unknown Inrimage header size", theIm->fi.name );

	}

  size= T_ImageSize( theIm );
  T_AllocImage( theIm );
  
/*   status= read( fd, theIm->buf, size ); */
/*   if ( status < 0 )  */
/*     E_SETERR(E_ERRR, "reading in file", theIm->fi.name); */
    
  fread( (void *) theIm->buf, sizeof(char), size, fd );
  if ( ferror(fd) ) 
    E_SETERR(E_ERRR, "reading in file", theIm->fi.name);

  E_REPERR( fclose(fd) );
  E_ISERRN("T_MyReadInrim", NULL );
  fclose( fd );

  switch ( theIm->type ) {
  case TYPEIM_16B :
  case TYPEIM_S16B :
    switch ( B_MACHINE ) {
    case MACHINE_DEC :
      switch( theIm->fi.machine ) {
      case MACHINE_SUN :
	E_Swap2octets( theIm->buf, theIm->buf, size );
      }
      break;
    case MACHINE_SUN :
      switch( theIm->fi.machine ) {
      case MACHINE_DEC :
	E_Swap2octets( theIm->buf, theIm->buf, size );
      }
    }
    break;
  case TYPEIM_FLOAT :
    switch ( B_MACHINE ) {
    case MACHINE_DEC :
      switch( theIm->fi.machine ) {
      case MACHINE_SUN :
	E_Swap4octets( theIm->buf, theIm->buf, size );
      }
      break;
    case MACHINE_SUN :
      switch( theIm->fi.machine ) {
      case MACHINE_DEC :
	E_Swap4octets( theIm->buf, theIm->buf, size );
      }
    }
    break;
    /*--- reste a faire ---*/
    break;
  }
}

/*----------------------------------------------------------------------*/

/*
  Read the header of an inrimage image from disk file

Normally, you should't have to use this function, because T_ReadInrim() does all what is needed to read an inrimage from disk. This function can be used if only the image header information is needed, for example to get the image type or size, without loading the entire image. The image header must be initialized with T_Init3Image() before any operations on images, and the name of a disk file, associated to this image, can be defined then. See also T_ReadInrim() and T_WriteInrim(), T_WriteInrimHead() to write inrimage images on disk.
*/

#if defined(_ANSI_)
void T_ReadInrimHead( t_ImagePtr theIm )
#else
void T_ReadInrimHead(theIm )
t_ImagePtr theIm ;	/* the image */

#endif

{	
  FILE *fd;
  int status,auxI;
  char str[B_FNAME_LENGTH];
  int isUInt= FALSE;
  int isSInt= FALSE;
  float tmp_fl;
  
  fd= fopen(theIm->fi.name,"r");
  if ( fd == NULL ){
    E_XERRN( E_ERRO, "T_ReadInrimHead", theIm->fi.name );
  }
  
  status= fscanf( fd, "%s", str );
  if ( status == EOF ) goto End_ReadInrimHead;
  if ( strncmp(str,"#INRIMAGE-4#{",13) != 0 ) goto End_ReadInrimHead;
  
  status= fscanf( fd, "%s", str );
  if ( status == EOF ) goto End_ReadInrimHead;
  sscanf( str, "XDIM=%d", &theIm->tx );
  
  status= fscanf( fd, "%s", str );
  if ( status == EOF ) goto End_ReadInrimHead;
  sscanf( str, "YDIM=%d", &theIm->ty );
  
  status= fscanf( fd, "%s", str );
  if ( status == EOF ) goto End_ReadInrimHead;
  sscanf( str, "ZDIM=%d", &theIm->tz );
  
  status= fscanf( fd, "%s", str );
  if ( status == EOF ) goto End_ReadInrimHead;
  sscanf( str, "VDIM=%d", &auxI );
  
  status= fscanf( fd, "%s", str );
  if ( status == EOF ) goto End_ReadInrimHead;
  sscanf( str, "TYPE=%s", str );
  
  if ( strcmp(str,"float") == 0 ) 		theIm->type= TYPEIM_FLOAT;
  else if ( strcmp(str,"unsigned") == 0 ) isUInt= TRUE;
  else if ( strcmp(str,"signed") == 0 ) isSInt= TRUE;
  else goto End_ReadInrimHead;
  
  if ( isUInt || isSInt ){
    status= fscanf( fd, "%s", str );
    if ( status == EOF ) goto End_ReadInrimHead;
    if ( strcmp( str, "fixed" ) != 0 ) goto End_ReadInrimHead;
  }
  
  status= fscanf( fd, "%s", str );
  if ( status == EOF ) goto End_ReadInrimHead;
  
  status= sscanf( str, "PIXSIZE=%d", &auxI );
  if ( status == EOF ) goto End_ReadInrimHead;
  
  if ( isUInt ) switch (auxI){
  case 8: theIm->type= TYPEIM_256; break;
  case 16: theIm->type= TYPEIM_16B; break;
  default: goto End_ReadInrimHead; break;
  }
  else if ( isSInt ) switch (auxI){
  case 16: theIm->type= TYPEIM_S16B; break;
  default: goto End_ReadInrimHead; break;
  }
  else switch (auxI){
  case 32: break;
  default: goto End_ReadInrimHead; break;
  }
  
  status= fscanf( fd, "%s", str );
  if ( status == EOF ) goto End_ReadInrimHead;
  if ( strcmp( str, "bits" ) != 0 ) goto End_ReadInrimHead;
  
  if ( (isUInt==TRUE) || (isSInt==TRUE) ){
	/* read SCALE=.... */
	/* ajout du test sur isSInt, sinon impossibilite de lecture */  
	/* des images short int,                    GM : 12/09/1994 */  
    status= fscanf( fd, "%s", str );
    if ( status == EOF ) goto End_ReadInrimHead;
  }
  
  status= fscanf( fd, "%s", str );
  if ( status == EOF ) goto End_ReadInrimHead;
  status= sscanf( str, "CPU=%s", str );
  
  if ( strcmp(str, "decm") == 0 ) 		theIm->fi.machine= MACHINE_DEC;
  else if ( strcmp(str, "sun") == 0 ) 	theIm->fi.machine= MACHINE_SUN;
  else goto End_ReadInrimHead;

  if ( (theIm->fi.machine != B_MACHINE) &&
       (theIm->type == TYPEIM_FLOAT) )
    T_Warn( "Image type and machine type not coherent", NULL );
  

  /* Read Voxel sizes */
  status= fscanf( fd, "%s", str );
  if ( status != EOF ) {
    status= sscanf( str, "VX=%f", &tmp_fl );
    if ( status == 1 ) theIm->sx = (reeltype) tmp_fl;
  }
  status= fscanf( fd, "%s", str );
  if ( status != EOF ) {
    status= sscanf( str, "VY=%f", &tmp_fl );
    if ( status == 1 ) theIm->sy = (reeltype) tmp_fl;
  }
  status= fscanf( fd, "%s", str );
  if ( status != EOF ) {
    status= sscanf( str, "VZ=%f", &tmp_fl );
    if ( status == 1 ) theIm->sz = (reeltype) tmp_fl;
  }

  fclose( fd );
  return;
  
 End_ReadInrimHead:
  
  fclose(fd);
  sprintf( tmpStr1, "unsuported inrimage type if file %s", theIm->fi.name );
  E_XERRN( E_ERRUNKNOWN, "T_ReadInrimHead", tmpStr1 );
}

/*----------------------------------------------------------------------*/

/*	
	Write an image on disk, with inrimage format.

	The name of the file has been given by the initialization function T_Init3Image(), or has been changed to a new disk file name by using T_SetImName(). See also T_Write2Inrim() to write an image with a given name.
*/

#if defined(_ANSI_)
void T_WriteInrim( t_ImagePtr theIm )
#else
void T_WriteInrim(theIm )
t_ImagePtr theIm ;	/* the image */

#endif

{	
  /* int fd; */
  FILE *fd;
  int size;
  
  T_WriteInrimHead( theIm );

  
/*   fd= open( theIm->fi.name, O_WRONLY, 0644 );   */
/*   if ( fd < 0 ) E_XERRN( E_ERRO, "T_MyWriteInrim", theIm->fi.name ); */
  fd = fopen(theIm->fi.name, "r+b");
  if ( fd == NULL ) 
    E_XERRN( E_ERRO, "T_MyWriteInrim", theIm->fi.name );

  
/*   status= lseek( fd, 256L, SEEK_SET ); */
/*   if ( status != 256L ){ */
/*     close(fd); */
  
  if ( fseek( fd, 256L, SEEK_SET ) == EOF ){
    fclose(fd);
    E_XERRN( E_ERRUNKNOWN,"T_MyWriteInrim", theIm->fi.name );
  }


  size= T_ImageSize( theIm );

/*   status= write( fd, theIm->buf, size ); */
/*   if ( status < 0 ) { */
/*     close(fd);  */


  fwrite( (void *) theIm->buf, sizeof(char), size, fd );
  if ( ferror(fd) ){
    fclose(fd);
  E_XERRN( E_ERRW, "T_MyWriteInrim", theIm->fi.name );
  }
  fclose (fd );
}

/*----------------------------------------------------------------------*/

/*
  Write the header of an inrimage image from disk file

Normally, you should't have to use this function, because T_WriteInrim() does all what is needed to read an inrimage from disk. This function can be used if only the image header information is needed if you have to write the image by yourself, for example on a line by line, or slice by slice basis. The image header must be initialized with T_Init3Image() before any operations on images, and the name of a disk file, associated to this image, can be defined then. See also T_ReadInrim(), T_ReadInrimHead() to write inrimage images on disk.
*/

#if defined(_ANSI_)
void T_WriteInrimHead( t_ImagePtr theIm )
#else
void T_WriteInrimHead(theIm )
t_ImagePtr theIm ;	/* the image */

#endif

{
  FILE *fd;
  int status,i;
  
  switch( theIm->type ){
    
  case TYPEIM_256: break;
  case TYPEIM_16B: break;
  case TYPEIM_S16B: break;
  case TYPEIM_FLOAT: break;
  default: 
    E_XERRN( E_ERRUNKNOWN, "T_WriteInrimHead (unknown image type)", 
	    theIm->fi.name); 
    break;
  }
  
  fd= fopen( theIm->fi.name, "w" );
  E_ERRN( fd == NULL, E_ERRC, "T_WriteInrimHead", theIm->fi.name );
  
  fprintf( fd, "#INRIMAGE-4#{\n" );
  fprintf( fd, "XDIM=%d \nYDIM=%d \nZDIM=%d \nVDIM=1 \n", 
	  theIm->tx, theIm->ty, theIm->tz );
  
  switch( theIm->type ){
    
  case TYPEIM_256: 
    fprintf(fd,"TYPE=unsigned fixed\nPIXSIZE=8 bits\nSCALE=2**0\n"); 
    break;
  case TYPEIM_16B:
    /* fprintf(fd,"TYPE=unsigned\nfixed\nPIXSIZE=16 bits\nSCALE=2**0\n"); */
    /* modifie par Gregoire 09/09/1994 : incompatibilite avec Inrimage    */
    fprintf(fd,"TYPE=unsigned fixed\nPIXSIZE=16 bits\nSCALE=2**0\n"); 			break;
  case TYPEIM_S16B:
    /* fprintf(fd,"TYPE=signed\nfixed\nPIXSIZE=16 bits\nSCALE=2**0\n");   */
    /* modifie par Gregoire 09/09/1994 : incompatibilite avec Inrimage    */	
    fprintf(fd,"TYPE=signed fixed\nPIXSIZE=16 bits\nSCALE=2**0\n"); 			break;
  case TYPEIM_FLOAT:
    fprintf(fd,"TYPE=float\nPIXSIZE=32 bits\n"); 
    break;
  }

  switch( B_MACHINE ){

	case MACHINE_SUN:	fprintf( fd, "CPU=sun\n" ); break;
	case MACHINE_DEC:	
	default:
		fprintf( fd, "CPU=decm\n" ); break;
	}
	
  /* write voxel sizes */
  fprintf(fd,"VX=%f\nVY=%f\nVZ=%f\n",theIm->sx, theIm->sy, theIm->sz);

  fprintf( fd, "#*[H]*<1> \n\n\n\n");
  status= ftell( fd );
  if ( status < 0 ){
    fclose(fd);
    E_XERRN( E_ERRW, "T_WriteInrimHead", theIm->fi.name );
  }
  
  for( i= status; i< 252; i++ ) fprintf(fd,"\n");
  fprintf( fd, "##}\n" );
  
  fclose( fd );
}

/************************************************************************/
/*
	highImages

	High level general routines for image processing
  
  (c) Jean-Philippe THIRION -- INRIA -- 1990
*/
/************************************************************************/

/*----------------------------------------------------------------------*/

/*
	Invert the sign of an image intensity

The effect of this function is different on signed or unsigned images. On signedor float images, the sign of the image intensity is simply reversed. On unsigned integer images the value is the complement to 1. For example, if the intensity is 1.045 in a float image, the result is -1.045. For an unsigned 1 byte / pixel image, an intensity 31 is transformed into 255 - 31 = 224. See T_ImInvertColor() for a completion to 1 in all cases.
*/
	

#if defined(_ANSI_)
void T_InvertIm( t_ImagePtr theIm )
#else
void T_InvertIm(theIm )
t_ImagePtr theIm ;	/* the image to invert */

#endif

{
  unsigned char *buf;
  reeltype *rbuf;
  unsigned short *usbuf;
  short int *ssbuf;
  int i;
  
  if ( theIm->buf == NULL ) return;
  
  switch( theIm->type ){
    
  case TYPEIM_BW:
    
    buf= (unsigned char *) theIm->buf;
    for( i=0; i< theIm->tx * theIm->ty * theIm->tz / 8; i++ ){
      *(buf) = !(*buf);
      buf++;
    }
    break;
    
  case TYPEIM_256:
    
    buf= (unsigned char *) theIm->buf;
    for( i=0; i< theIm->tx * theIm->ty * theIm->tz; i++ ) {
      *(buf) = 255 - *buf;
      buf++;
    }
    break;
    
  case TYPEIM_16B:
    
    usbuf= (unsigned short *) theIm->buf;
    for( i=0; i< theIm->tx * theIm->ty * theIm->tz; i++ ){ 
      *(usbuf) = 65535 - *usbuf;
      usbuf++;
    }
    break;
    
  case TYPEIM_S16B:
    
    ssbuf= (short int*) theIm->buf;
    for( i=0; i< theIm->tx * theIm->ty * theIm->tz; i++ ) {
      if ( *ssbuf == -32768 )
	*ssbuf = -32767;
      *(ssbuf) = - *ssbuf;
      ssbuf++;
    }
    break;
    
  case TYPEIM_FLOAT:
    
    rbuf= (reeltype *) theIm->buf;
    for( i=0; i< theIm->tx * theIm->ty * theIm->tz; i++ ){
      *(rbuf) = (- *rbuf);
      rbuf++;
    }
    break;
    
  default:
    E_SETERR( E_ERRUNKNOWN, "T_InvertIm", "unknown type of image");
    break;
  }
  E_ISERRN( "T_InvertIm", NULL );
}

/*----------------------------------------------------------------------*/

/*
  Reverse the value of an image by complement to 1

It is assumed that the values in the images are normalized (that is, between 0 and 1 for floating point images). In that case, the result for each voxel is the complement to 1. For example, for a floating point image, if the intensity is 0.92, the result is 0.08. For a 1 byte/voxel image, if the intensity is 31, the result is 255 - 31 = 224. See T_InvertIm() for sign inversion of signed or floating point images. Note that you can use the same image for input and output. If the output image buffer is not allocated, this function allocates one. At last, the output image should have been initialized, for example with B_Init3Image(), orwith T_CopyHeaderIm().
*/

#if defined(_ANSI_)
void T_ImInvertColor( t_ImagePtr theIm, t_ImagePtr resIm )
#else
void T_ImInvertColor(theIm, resIm )
t_ImagePtr theIm;	/* input image */
t_ImagePtr resIm;	/* output image */

#endif

{
  int i,j,k;
  reeltype value;
  t_Image aux;
  t_ImagePtr res;
  
  if ( resIm->buf == NULL ) T_AllocImage( resIm );
  
  if ( resIm->buf == theIm->buf ){
    T_CopyHeaderIm( resIm, &aux );
    T_AllocImage( &aux );
    res= &aux;
  }
  else res= resIm;
  
  for( k=0; k< resIm->tz; k++ )
    for( j=0; j< resIm->ty; j++ )
      for( i=0; i< resIm->tx; i++ ){
	
	value= T_Get3Point( i,j,k, theIm );
	T_Set3Point( i,j,k, res, 1 - value, MODE_SET );
      }
  
  if ( resIm->buf == theIm->buf ){
    T_FreeImage( resIm );
    resIm->buf= aux.buf;
  }
}

/*----------------------------------------------------------------------*/

/*	
	Vertical mirror of an image

	Reverse the order of the lines in the image
*/

#if defined(_ANSI_)
void T_ImageVMirror( t_ImagePtr theIm, t_ImagePtr resIm )
#else
void T_ImageVMirror(theIm, resIm )
t_ImagePtr theIm; /* input image */
t_ImagePtr resIm ; /* output reversed image */

#endif

{
  reeltype value;
  t_Image auxIm;
  int i,j,k;
  
  T_CopyHeaderIm( resIm, &auxIm );
  T_AllocImage( &auxIm );
  E_ISERRN("T_ImageVMirror",NULL);
  
  for( k=0; k< theIm->tz; k++ )	
    for( j=0; j< theIm->ty; j++ )
      for( i=0; i< theIm->tx; i++ ){
	value= T_Get3Point( i,j,k, theIm );
	T_Set3Point( theIm->tx-i-1,j,k, &auxIm, value, MODE_SET );
      }
  
  T_FreeImage( resIm );
  resIm->buf= auxIm.buf;
}

/*----------------------------------------------------------------------*/

/*
	Horizontal mirror of an image

	Reverse the order of the vertical columns of an image
*/

#if defined(_ANSI_)
void T_ImageHMirror( t_ImagePtr theIm, t_ImagePtr resIm )
#else
void T_ImageHMirror(theIm, resIm )
t_ImagePtr theIm;	/* input image */
t_ImagePtr resIm ;	/* output mirrored image */

#endif

{
  reeltype value;
  t_Image auxIm;
  int i,j,k;
  
  T_CopyHeaderIm( resIm, &auxIm );
  T_AllocImage( &auxIm );
  E_ISERRN("T_ImageHMirror", NULL);
  
  for( k=0; k<theIm->tz;k++)	
    for( j=0; j < theIm->ty; j++ )
      for( i=0; i< theIm->tx; i++ ){
	value= T_Get3Point( i,j,k, theIm );
	T_Set3Point(i,theIm->ty-j-1,k,&auxIm,value,MODE_SET );
      }
  T_FreeImage( resIm );
  resIm->buf= auxIm.buf;
}

/*----------------------------------------------------------------------*/

/*
	Subsample an image by taking only one line every n lines
*/

#if defined(_ANSI_)
void T_ImExtLines( t_ImagePtr theIm, t_ImagePtr resIm, int n )
#else
void T_ImExtLines(theIm, resIm, n )
t_ImagePtr theIm;	/* input image */
t_ImagePtr resIm;	/* subsample output image */
int n ;				/* one line over n is kept */

#endif

{
  reeltype value;
  t_Image auxIm;
  int i,j,k,ty,line;
  
  E_ERRN( n <= 0, E_ERRNULL, "T_ImExtLines","null number of lines");
  
  ty= theIm->ty / n;
  T_CopyHeaderIm( resIm, &auxIm );
  auxIm.ty= ty;
  T_AllocImage( &auxIm );
  E_ISERRN("T_ImExtLines",NULL);
  
  for( k=0; k<theIm->tz;k++)	
    for( j=0; j < ty; j++ ){
      line= j * n;
      for( i=0; i< theIm->tx; i++ ){
	value= T_Get3Point( i,line,k, theIm );
	T_Set3Point(i,j,k,&auxIm,value,MODE_SET);
      }
    }
  
  T_FreeImage( resIm );
  resIm->ty= ty;
  resIm->buf= auxIm.buf;
}

/*----------------------------------------------------------------------*/

/*
	Rotate an image of an angle multiple of PI/2
*/

#if defined(_ANSI_)
void T_ImageRotate( t_ImagePtr theIm, t_ImagePtr resIm, int angle )
#else
void T_ImageRotate(theIm, resIm, angle )
t_ImagePtr theIm;	/* input image */
t_ImagePtr resIm;	/* rotated outpu image */
int angle ;			/* rotation angle in degrees (example: 90) */

#endif

{
  int line,col,c,l,off_x,off_y;
  reeltype value;
  t_Image auxIm;
  
  T_CopyHeaderIm( resIm, &auxIm );
  
  switch( angle ){ 
  case 90:
    off_x= theIm->tx / 2 - theIm->ty / 2;
    off_y= theIm->tx / 2 + theIm->ty / 2;
    auxIm.tx= theIm->ty;
    auxIm.ty= theIm->tx;
    break;
  case -90:
    off_x= theIm->tx / 2 + theIm->ty / 2;
    off_y= - theIm->tx / 2 + theIm->ty / 2;
    auxIm.tx= theIm->ty;
    auxIm.ty= theIm->tx;
    break;
  default:
    T_Warn("unknown angle %d in T_ImageRotate\n", CNULL);
    return;
    break;
  }
  
  T_AllocImage( &auxIm );
  E_ISERRN("T_ImageRotate",NULL);
  
  for( line=0; line < theIm->ty; line++ )
    for( col=0; col < theIm->tx; col++ ){
      
      value= T_GetPoint( col, line, theIm );
      switch( angle ){
      case 90:
	c= line + off_x;
	l= -col	+ off_y;
	break;
      case -90:
	c= -line + off_x;
	l= col	+ off_y;
	break;
		default:
			c= col;
			l= line;
			break;
      }
      
      T_SetPoint( c, l, &auxIm, value, MODE_SET );
    }
  
  T_FreeImage( resIm );
  resIm->buf= auxIm.buf;
}

/*----------------------------------------------------------------------*/

/*
	Vertical shift of an image

	The image is considered to be toric according to y, that is, lines disappearing at the bottom of the image will reappear at the top.

Limitations:
	works only for 2D images.
*/

#if defined(_ANSI_)
void T_ImageVShift( t_ImagePtr theIm, t_ImagePtr resIm, int shift )
#else
void T_ImageVShift(theIm, resIm, shift )
t_ImagePtr theIm;	/* input image */
t_ImagePtr resIm;	/* shifted image */
int shift ;			/* number of lines of the shift */

#endif

{
  int line,col;
  reeltype value;
  t_Image auxIm;
  
  T_CopyHeaderIm( resIm, &auxIm );
  T_AllocImage( &auxIm );
  E_ISERRN("T_ImageVShift",NULL);
  
  for( line=0; line < theIm->ty; line++ )
    for( col=0; col < theIm->tx; col++ ){
      value= T_GetPoint( col, line, theIm );
      T_SetPoint( col, (line + shift) % theIm->ty, 
		 &auxIm, value, MODE_SET );
    }
  
  T_FreeImage( resIm );
  resIm->buf= auxIm.buf;
}

/*----------------------------------------------------------------------*/

/*
	Transform each voxel value of an image into its square
*/

#if defined(_ANSI_)
void T_ImageExp2( t_ImagePtr theIm, t_ImagePtr resIm )
#else
void T_ImageExp2(theIm, resIm )
t_ImagePtr theIm;	/* the image */
t_ImagePtr resIm ;	/* the square valued image */

#endif

{
  reeltype value;
  int j,k;
	register int i;
	register reeltype *rPt1,*rPt2;
  
  T_CopyImage( theIm, resIm );
  E_ISERRN("T_ImageExp2",NULL);
 
	if ( theIm->type == TYPEIM_FLOAT ){

		rPt1= (reeltype *) theIm->buf;
		rPt2= (reeltype *) resIm->buf;

		for( i=0; i< resIm->tx * resIm->ty * resIm->tz; i++ ){

			*rPt2++ = *rPt1 * *rPt1;
			rPt1++;
			}
		}

	else{
	  	for( k=0; k< resIm->tz; k++ )
    		for( j=0; j< resIm->ty; j++ )
      			for( i=0; i< resIm->tx; i++ ){
	
					value= T_Get3Point( i,j,k, resIm );
					T_Set3Point( i,j,k, resIm, value*value, MODE_SET );
					}
      }
}

/*----------------------------------------------------------------------*/

/*
	Transform each voxel of an image into its square root
*/

#if defined(_ANSI_)
void T_ImageSqrt( t_ImagePtr theIm, t_ImagePtr resIm )
#else
void T_ImageSqrt(theIm, resIm )
t_ImagePtr theIm;	/* the input image */
t_ImagePtr resIm ;	/* the output square root valued image */

#endif

{
  reeltype value;
  int i,j,k;
  
  T_CopyImage( theIm, resIm );
  E_ISERRN("T_ImageSqrt",NULL);
  
  for( k=0; k< resIm->tz; k++ )
    for( j=0; j< resIm->ty; j++ )
      for( i=0; i< resIm->tx; i++ ){
	
	value= T_Get3Point( i,j,k, resIm );
	T_Set3Point( i,j,k, resIm, (reeltype) sqrt(fabs(value)), 
		    MODE_SET );
      }
}

/*----------------------------------------------------------------------*/

/*
	Compose two images with a given composition mode.
*/

#if defined(_ANSI_)
void T_ImageCompose( t_ImagePtr im1, t_ImagePtr im2, t_ImagePtr resIm, int mode )
#else
void T_ImageCompose(im1, im2, resIm, mode )
t_ImagePtr im1;		/* first image */
t_ImagePtr im2;		/* second image */
t_ImagePtr resIm;	/* result image */
int mode ;			/* image composition mode */

#endif

{
  t_Image auxIm;
  t_ImagePtr anIm;
  int l,z;
  reeltype value1,value2, tmp;
  register int c;
  register reeltype *rPt1,*rPt2,*rPt3;
  
  E_ERRN( 	(im1->tx > im2->tx) ||
		(im1->ty > im2->ty) ||
		(im1->tz > im2->tz), E_ERRARRAY,
		"T_ImageCompose", "Incompatible size of image");
  
  if ( resIm->buf == NULL){
    T_CopyHeaderIm( resIm, &auxIm );
    T_AllocImage( &auxIm );
    anIm= &auxIm;
  }
  else anIm= resIm;
  
  E_ISERRN("T_ImageCompose",NULL);
  
  if ( (im1->type == TYPEIM_FLOAT) &&
       (im2->type == TYPEIM_FLOAT) &&
       (anIm->type == TYPEIM_FLOAT) &&
       ((mode == MODE_ADD) ||
	(mode == MODE_ADD2)) ) {
    
    rPt1= (reeltype *) im1->buf;
    rPt2= (reeltype *) im2->buf;
    rPt3= (reeltype *) anIm->buf;
    
    switch( mode ){
      
    case MODE_ADD:
      for( c=0; c < im1->tx * im1->ty * im1->tz; c++ )
	*(rPt3++) = *(rPt1++) + *(rPt2++);
      break;
      
    case MODE_ADD2:
      for( c=0; c < im1->tx * im1->ty * im1->tz; c++ ){
	tmp = *(rPt2++);
	*(rPt3++) = *(rPt1++) + tmp*tmp;
      }
      break;
    }
  }
  
  else{
    for( z=0; z< im1->tz; z++ )
    		for( l=0; l< im1->ty; l++ )
		  for( c=0; c < im1->tx; c++ ){
		    value1= T_Get3Point( c, l, z, im1 );
		    value2= T_Get3Point( c, l, z, im2 );
		    T_Set3Point( c, l, z, anIm, value1, MODE_SET );
		    T_Set3Point( c, l, z, anIm, value2, (char) mode );
					}
  }
  
  if ( resIm != anIm ){
    T_FreeImage( resIm );
    resIm->buf= anIm->buf;
  }
  
  E_REPERR( T_FreeImage( resIm ) );
  E_ISERRN( "T_ImageCompose", NULL );
}

/*----------------------------------------------------------------------*/

/*
	Complete copy of an image into another one

	All is copied, the mage header information and the image buffer information. See also T_CopyHeaderIm() and T_CopyImageBuffer().
*/

#if defined(_ANSI_)
void T_CopyImage( t_ImagePtr theIm, t_ImagePtr resIm )
#else
void T_CopyImage(theIm, resIm )
t_ImagePtr theIm;	/* input image */
t_ImagePtr resIm ;	/* the image copy */ 

#endif

{
  if ( theIm == resIm ) return;
  T_CopyHeaderIm( theIm, resIm );
  
  if ( theIm->buf == NULL ){
    T_Warn("Nothing in the image buffer in T_CopyImage\n",CNULL );
    return;
  }
  
  T_AllocImage( resIm );
  E_ISERRN("T_CopyImage",NULL);
  memcpy( (void *) resIm->buf, (void *) theIm->buf, T_ImageSize(theIm) );
}

/*----------------------------------------------------------------------*/

/*
	Copy the buffer of the image into the other one

	This operation does not copy the image header information. One can use T_CopyImage() for a complete copy and T_CopyHeaderIm() for a copy of the image header.
*/

#if defined(_ANSI_)
void T_CopyImageBuffer( t_ImagePtr theIm, t_ImagePtr resIm )
#else
void T_CopyImageBuffer(theIm, resIm )
t_ImagePtr theIm;	/* input image */
t_ImagePtr resIm ;	/* image where the buffer is copied */

#endif

{
  int c,l,z;
  reeltype value;
  
  if ( theIm == resIm ) return;
  if ( (theIm->buf == NULL) || (resIm->buf == NULL) ){
    T_Warn("Nothing in an image buffer in T_CopyImageBuffer\n",CNULL );
    return;
  }
  
  if ( 	(theIm->tx != resIm->tx) ||
      (theIm->ty != resIm->ty) ||
      (theIm->tz != resIm->tz) ||
      (theIm->type != resIm->type) ){
    
    for ( z=0; z< resIm->tz; z++ )
      for ( l=0; l< resIm->ty; l++ )
	for( c=0; c< resIm->tx; c++ ){
	  value= T_Get3Point( c,l,z,theIm );
	  T_Set3Point(c,l,z,resIm,value,MODE_SET);
	}
    return;
  }
  memcpy( (void *) resIm->buf, (void *) theIm->buf, T_ImageSize(theIm) );
}

/*----------------------------------------------------------------------*/

/*
	Copy only the header of an image into another EMPTY image

It copies the image header information but not the image buffer.  Use T_CopyImage() for a complete copy of the image. Also, the output image must not have an allocated image buffer, because the image buffer is superseted by this operation (and lots of memory can be lost). This operation can be used also simply to initialize the values of the output image header, and can advantageously replace T_InitImage() in some particular cases. After this operation, the image buffer of resIm is NULL, you should then use T_AllocImage() if you want to allocate the image buffer.
*/

#if defined(_ANSI_)
void T_CopyHeaderIm( t_ImagePtr theIm, t_ImagePtr resIm )
#else
void T_CopyHeaderIm(theIm, resIm )
t_ImagePtr theIm;	/* the input image header */
t_ImagePtr resIm ;	/* the output image header */

#endif

{
  if ( theIm == resIm ) return;
  memcpy( (void *) resIm, (void *) theIm, sizeof(t_Image) );
  resIm->buf= NULL;
}

/*----------------------------------------------------------------------*/

/*
	Copy sizes and type information between two image headers

If you want to make complete image copies you should use T_CopyImage(). You should use T_CopyHeaderIm() if you want to copy all header information. The present operation doesn't change the image buffer pointer. In particular, it doens't initialize this pointer. You should have a look to T_CopyHeaderIm() or T_Init3Image() for those operations.
*/

#if defined(_ANSI_)
void T_CopyCaractIm( t_ImagePtr theIm, t_ImagePtr resIm )
#else
void T_CopyCaractIm(theIm, resIm )
t_ImagePtr theIm;	/* input header */
t_ImagePtr resIm ;	/* output header */

#endif

{
  if ( theIm == resIm ) return;
  resIm->tx= theIm->tx;
  resIm->ty= theIm->ty;
  resIm->tz= theIm->tz;
  resIm->type= theIm->type;
}

/*----------------------------------------------------------------------*/

/*
	Transform 2D image gradient vectors into polar forms

If the input are 3D images, this function process each image slices independently.
*/

#if defined(_ANSI_)
void T_ImGradToMod( t_ImagePtr gradXIm, t_ImagePtr gradYIm, t_ImagePtr modIm, t_ImagePtr dirIm )
#else
void T_ImGradToMod(gradXIm, gradYIm, modIm, dirIm )
t_ImagePtr gradXIm;		/* image of the x coordinates of the gradient vector */
t_ImagePtr gradYIm;		/* image of the y coordinates of the gradient vector */
t_ImagePtr modIm;		/* image of the module of the gradient */
t_ImagePtr dirIm ;		/* image of the phase of the gradiet */

#endif

{
  int l,c,z;
  reeltype xx,yy,mod,dir;
  t_Image auxMod,auxDir;
  
  E_ERRN(
	 (gradXIm->buf == NULL) ||
	 (gradXIm->buf == NULL) ||
	 (gradXIm->tx != gradYIm->tx) ||
	 (gradXIm->ty != gradYIm->ty) ||
	 (gradXIm->tz != gradYIm->tz) ||
	 (gradXIm->tx != modIm->tx) ||
	 (gradXIm->ty != modIm->ty) ||
	 (gradXIm->tz != modIm->tz) ||
	 (gradXIm->tx != dirIm->tx) ||
	 (gradXIm->ty != dirIm->ty) ||
	 (gradXIm->tz != dirIm->tz), E_ERRARRAY,
	 "T_ImGradToMod", "Incompatible images" );
  
  if ( gradXIm->type != TYPEIM_FLOAT )
    T_Warn("Not a float image in T_ImGradToMod\n", NULL );
  
  T_CopyHeaderIm( modIm, &auxMod );
  T_CopyHeaderIm( dirIm, &auxDir );
  T_AllocImage( &auxMod );
  T_AllocImage( &auxDir );
  
  E_REPERR({ T_FreeImage(modIm); T_FreeImage(dirIm); });
  E_ISERRN("T_ImGradToMod", NULL);
  
  for( z=0; z < gradXIm->tz; z++ )
    for( l=0; l < gradXIm->ty; l++ )
      for( c=0; c < gradXIm->tx; c++ ){
	
	xx= T_Get3Point( c, l, z, gradXIm );
	yy= T_Get3Point( c, l, z, gradYIm );
	mod= (reeltype) sqrt( xx * xx + yy * yy );
	if ( mod > EPSY )	dir= (reeltype) atan2( yy, xx );
	else				dir= 0.0;
	T_Set3Point( c, l, z, &auxMod, mod, MODE_SET );
	T_Set3Point( c, l, z, &auxDir, dir, MODE_SET );
      }
  
  T_FreeImage( modIm );
  T_FreeImage( dirIm );
  modIm->buf= auxMod.buf;
  dirIm->buf= auxDir.buf;
}

/*----------------------------------------------------------------------*/

/*
    Apply a filter kernel to a 3D image, normalized output

The 3D image is considered to be toric according to the 3 directions. The
output image values are limited to 0..1 
*/

#if defined(_ANSI_)
void T_ImApplyFilterNorm( t_ImagePtr theIm, t_ImagePtr resIm, reeltype *filter, 
int tx, int ty, int tz )
#else
void T_ImApplyFilterNorm(theIm, resIm, filter, tx, ty, tz )
t_ImagePtr theIm;   /* input image */
t_ImagePtr resIm;   /* output filtered image */
reeltype *filter;   /* 3d filter kernel buffer */
int tx,ty,tz;       /* 3d filter kernel size */

#endif

{
  t_Image auxIm;
  int i,j,k,l,m,n;
  reeltype *ptr;
  reeltype value;
  
  T_CopyHeaderIm( resIm, &auxIm );
  T_AllocImage( &auxIm );
  E_ISERRN("T_ImApplyFilter",NULL);
  
  for( k=0; k< theIm->tz; k++ )
    for( i=0; i< theIm->ty; i++ )
      for( j=0; j< theIm->tx; j++ ){
	
	ptr= filter;
	value= 0.0;
	
	for( n=0; n< tz; n++ )
	  for( m=0; m< ty; m++ )
	    for( l=0; l< tx; l++ )
	      value += T_Get3Point(j+l,i+m,k+n,theIm) * *(ptr++);

		if ( value <= 0.0 ) value= 0.0;
   		else if ( value >= 1.0 ) value= 1.0;
		T_Set3Point( j+tx/2, i+ty/2, k+tz/2, &auxIm, value, MODE_SET );
      }
  
  T_FreeImage( resIm );
  resIm->buf= auxIm.buf;
}

/*----------------------------------------------------------------------*/

/*
	Apply a filter kernel to a 3D image

The 3D image is considered to be toric according to the 3 directions
*/

#if defined(_ANSI_)
void T_ImApplyFilter( t_ImagePtr theIm, t_ImagePtr resIm, reeltype *filter, int tx, int ty, int tz )
#else
void T_ImApplyFilter(theIm, resIm, filter, tx, ty, tz )
t_ImagePtr theIm;	/* input image */
t_ImagePtr resIm;	/* output filtered image */
reeltype *filter;	/* 3d filter kernel buffer */
int tx,ty,tz;		/* 3d filter kernel size */

#endif

{
  t_Image auxIm;
  int i,j,k,l,m,n;
  reeltype *ptr;
  reeltype value;
  
  T_CopyHeaderIm( resIm, &auxIm );
  T_AllocImage( &auxIm );
  E_ISERRN("T_ImApplyFilter",NULL);
  
  for( k=0; k< theIm->tz; k++ )
    for( i=0; i< theIm->ty; i++ )
      for( j=0; j< theIm->tx; j++ ){
	
	ptr= filter;
	value= 0.0;
	
	for( n=0; n< tz; n++ )
	  for( m=0; m< ty; m++ )
	    for( l=0; l< tx; l++ )
	      value += T_Get3Point(j+l,i+m,k+n,theIm) * *(ptr++);
	
	T_Set3Point( j+tx/2, i+ty/2, k+tz/2, &auxIm, value, MODE_SET );
      }
  
  T_FreeImage( resIm );
  resIm->buf= auxIm.buf;
}

/*----------------------------------------------------------------------*/

/*
	Filter the image in one direction with external kernel

This function applies a filter kernel to an image in one of the principal direction (x,y,z). When the filter is a Gaussian filter, it is much faster to use a separate filtering (see E_SepFilter()).


Restrictions:
	The image is considered to be toric.
*/

#if defined(_ANSI_)
void T_ImDirectFilter( t_ImagePtr theIm, t_ImagePtr resIm, reeltype *filter, int t, int direct )
#else
void T_ImDirectFilter(theIm, resIm, filter, t, direct )
t_ImagePtr theIm;	/* the input image */
t_ImagePtr resIm;	/* the output filtered image */
reeltype *filter;	/* the one dimentional filter kernel */
int t;				/* number of samples (size) of the image kernel */
int direct ;		/* direction of filtering */

#endif

{
  t_Image auxIm;
  int i,j,k,n;
  reeltype *ptr;
  reeltype value;
  
  T_CopyHeaderIm( resIm, &auxIm );
  T_AllocImage( &auxIm );
  E_ISERRN("T_ImDirectFilter",NULL);
  
  switch( direct ){
    
  case DIRECT_X:
    
    for( k=0; k< theIm->tz; k++ )
      for( j=0; j< theIm->ty; j++ )
	for( i=0; i< theIm->tx; i++ ){
	  
	  ptr= filter;
	  value= 0.0;
	  for( n=0; n< t; n++ )
	    value += T_Get3Point(i+n,j,k,theIm) * *(ptr++);
	  T_Set3Point( i+t/2, j, k, &auxIm, value, MODE_SET );
	}
    break;
    
  case DIRECT_Y:
    
    for( k=0; k< theIm->tz; k++ )
      for( j=0; j< theIm->ty; j++ )
	for( i=0; i< theIm->tx; i++ ){
	  
	  ptr= filter;
	  value= 0.0;
	  for( n=0; n< t; n++ )
	    value += T_Get3Point(i,j+n,k,theIm) * *(ptr++);
	  T_Set3Point( i, j+t/2, k, &auxIm, value, MODE_SET );
	}
    break;
    
  case DIRECT_Z:
    
    for( k=0; k< theIm->tz; k++ )
      for( j=0; j< theIm->ty; j++ )
	for( i=0; i< theIm->tx; i++ ){
	  
	  ptr= filter;
	  value= 0.0;
	  for( n=0; n< t; n++ )
	    value += T_Get3Point(i,j,k+n,theIm) * *(ptr++);
	  T_Set3Point( i, j, k+t/2, &auxIm, value, MODE_SET );
	}
    break;
    
  default: 
    E_SETERR( E_ERRUNKNOWN, "T_ImDirectFilter","unknown direction");
    break;
  }
  
  E_REPERR( T_FreeImage( resIm ) );
  E_ISERRN("T_ImDirectFilter",NULL);
  
  T_FreeImage( resIm );
  resIm->buf= auxIm.buf;
}

/*----------------------------------------------------------------------*/

/*
	Sort the image buffer with decreasing voxel values

This function is using quicksort to transform the buffer ofthe image in such a way that the voxel values are ordered from the largest voxel value to the smallest. This strange operation is very useful to compute statistics about images, or histograms.

Restrictions:
	Works only with float or 1 byte per voxel images.
*/

#if defined(_ANSI_)
void T_QuickImage( t_ImagePtr theIm )
#else
void T_QuickImage(theIm )
t_ImagePtr theIm ;	/* the image to sort */

#endif

{
  if ( theIm->buf == NULL ) return;
  
  switch( theIm->type ){
    
  case TYPEIM_FLOAT:
    qsort( (char *) theIm->buf, theIm->tx * theIm->ty * theIm->tz,
	  sizeof(reeltype), (func_type_2_arg)T_CompareFloat );
    break;
    
  case TYPEIM_256:
    qsort( (char *) theIm->buf, theIm->tx * theIm->ty * theIm->tz,
	  1, (func_type_2_arg)T_Compare256 );
    break;
    
  default:
    E_XERRN( E_ERRMISMATCH, "T_QuickImage", "type of images" );
    break;
  }
}

/*----------------------------------------------------------------------*/

/*
	Median filtering of a 2D image

Apply a median filter to a 2D image, using a given window size in x and y.
*/


#if defined(_ANSI_)
void T_ImMedianFilter( t_ImagePtr theIm, t_ImagePtr resIm, int tx, int ty )
#else
void T_ImMedianFilter(theIm, resIm, tx, ty )
t_ImagePtr theIm;	/* input image */
t_ImagePtr resIm;	/* output filtered image */
int tx;				/* window size in x */
int ty ;			/* window size in y */

#endif

{
  t_Image auxIm;
  reeltype *buffer,*ptr;
  int i,j,l,c,size,size_2;
  
  T_CopyHeaderIm( resIm, &auxIm );
  T_AllocImage( &auxIm );
  
  size= tx * ty;
  size_2= size / 2;
  
  buffer= (reeltype *) T_Alloc( (int) (sizeof(reeltype) * size) );
  
  E_REPERR({ T_FreeImage(&auxIm); free((char *) buffer ); });
  E_ISERRN("T_ImMedianFilter",NULL);
  
  for( l=0; l< auxIm.ty; l++ )
    for( c=0; c< auxIm.tx; c++ ){
      
      ptr= buffer;
      for( i=0; i< ty; i++ )
	for( j=0; j< tx; j++ )
	  (*ptr++)= T_GetPoint( c+j, l+i, theIm ); 
      
      qsort( (char *) buffer, size, sizeof(reeltype), (func_type_2_arg)T_CompareFloat );
      T_SetPoint( c+tx/2, l+ty/2, &auxIm, buffer[size_2], MODE_SET );
    }
  
  T_FreeImage( resIm );
  free( (char *) buffer );
  resIm->buf= auxIm.buf;
  auxIm.buf= NULL;
}

/*----------------------------------------------------------------------*/

/*
	Keeps only the positive values of an image

The negative values in the image are replaced with zero.
*/

#if defined(_ANSI_)
void T_TakePositIm( t_ImagePtr theIm, t_ImagePtr resIm )
#else
void T_TakePositIm(theIm, resIm )
t_ImagePtr theIm;	/* input image */
t_ImagePtr resIm ;	/* positive output image */

#endif

{
	int i,j,k;
	reeltype value;

	if ( !resIm->buf ){
		resIm->tx= theIm->tx;
		resIm->ty= theIm->ty;
		resIm->tz= theIm->tz;
		T_AllocImage( resIm );
		}

	for ( k=0; k< theIm->tz; k++ )
		for( j=0; j< theIm->ty; j++ )
			for( i=0; i< theIm->tx; i++ ){
				value= T_Get3Point( i, j, k, theIm );
      			if ( value > 0.0 )
						T_Set3Point( i, j, k, resIm, value, MODE_SET);
				else	T_Set3Point( i, j, k, resIm, 0.0, MODE_SET);
				}
    }
  
/*----------------------------------------------------------------------*/

/*
	Change the principal directions of a 3D image

This functions allows you to exchange the principal directions of a 3D image. It reorder the voxels such that x,y,z directions become pX,pY,pZ.
*/

#if defined(_ANSI_)
void T_ExchangeIm( t_ImagePtr theIm, t_ImagePtr resIm, int pX, int pY, int pZ )
#else
void T_ExchangeIm(theIm, resIm, pX, pY, pZ )
t_ImagePtr theIm;	/* the input image */
t_ImagePtr resIm;	/* the output reordered image */
int pX;		/* x axis direction of output image */
int pY;		/* y axis direction of output image */
int pZ ;	/* z axis direction of output image */

#endif

{
  t_Image auxIm;
  int i,j,k,trans[3];
  reeltype value;
  
  T_CopyHeaderIm( resIm, &auxIm );
  if ( auxIm.type == TYPEIM_UNKNOWN ) auxIm.type= theIm->type;
  
  switch( pX ){
  case DIRECT_X: auxIm.tx= theIm->tx; break;
  case DIRECT_Y: auxIm.tx= theIm->ty; break;
  case DIRECT_Z: auxIm.tx= theIm->tz; break;
  }
  
  switch( pY ){
  case DIRECT_X: auxIm.ty= theIm->tx; break;
  case DIRECT_Y: auxIm.ty= theIm->ty; break;
  case DIRECT_Z: auxIm.ty= theIm->tz; break;
  }
  
  switch( pZ ){
  case DIRECT_X: auxIm.tz= theIm->tx; break;
  case DIRECT_Y: auxIm.tz= theIm->ty; break;
  case DIRECT_Z: auxIm.tz= theIm->tz; break;
  }
  
  T_AllocImage( &auxIm );
  E_ISERRN("T_ExchangeIm",NULL);
  
  for( k=0; k< theIm->tz; k++ )
    for( j=0; j< theIm->ty; j++ )
      for( i=0; i< theIm->tx; i++ ){
	value= T_Get3Point( i,j,k,theIm );
	trans[ DIRECT_X % 3 ]= i;
	trans[ DIRECT_Y % 3 ]= j;
	trans[ DIRECT_Z % 3 ]= k;
	T_Set3Point( trans[pX % 3], trans[pY % 3],
		    trans[pZ % 3], &auxIm, value, MODE_SET );
      }
  
  T_FreeImage( resIm );
  T_CopyHeaderIm( &auxIm, resIm );
  resIm->buf= auxIm.buf;
}

/*----------------------------------------------------------------------*/

/*
  Extract a smaller rectangular image from a given image
*/

#if defined(_ANSI_)
void T_ExtractIm( t_ImagePtr theIm, t_ImagePtr resIm, int ox, int oy, int oz  )
#else
void T_ExtractIm(theIm, resIm, ox, oy, oz  )
t_ImagePtr theIm;	/* input image */
t_ImagePtr resIm;	/* extracted output image */
int ox,oy,oz;		/* origin of the extracted image in the original image */

#endif

{
  t_Image auxIm;
  int i,j,k;
  reeltype value;
  
  T_CopyHeaderIm( resIm, &auxIm );
  T_AllocImage( &auxIm );
  E_ISERRN("T_ExtractIm",NULL);
  
  for( k=0; k< auxIm.tz; k++ )
    for( j=0; j< auxIm.ty; j++ )
      for( i=0; i< auxIm.tx; i++ ){
	value= T_Get3Point( i+ox,j+oy,k+oz,theIm );
	T_Set3Point( i,j,k, &auxIm, value, MODE_SET );
      }
  
  T_FreeImage( resIm );
  resIm->buf= auxIm.buf;
}

/*----------------------------------------------------------------------*/

/*
  Compute the average value of an image

The image is generally normalized between 0 and 1, thus the result of this function is also generally between 0 and 1. See T_Get3Point() and T_Set3Point() for discussions about image normalisation.
*/

#if defined(_ANSI_)
int T_ImageAverage( t_ImagePtr theIm, reeltype *pAver )
#else
int T_ImageAverage(theIm, pAver )
t_ImagePtr theIm;	/* input image */
reeltype *pAver ;	/* average value of the voxels */

#endif

{
  int x,y,z;
  double aver= 0.0;
  long size;
  
  for( z=0; z< theIm->tz; z++ )
    for( y=0; y< theIm->ty; y++ )
      for( x=0; x< theIm->tx; x++ )
		aver += T_Get3Point( x,y,z,theIm );
  
	size= theIm->tx * theIm->ty * theIm->tz;
	if ( !size ) { *pAver= 0.0; return( FALSE ); }
	else { *pAver= (reeltype) (aver / size); return( TRUE ); }
	}

/*----------------------------------------------------------------------*/

/*
	used to compare two voxel values in 1 byte/voxel images
*/

#if defined(_ANSI_)
static int T_Compare256( unsigned char *r1, unsigned char *r2 )
#else
static int T_Compare256(r1, r2 )
unsigned char *r1;
unsigned char *r2 ;

#endif

{
  if ( *r1 > *r2 ) return(1);
  else    if ( *r1 < *r2 ) return(-1);
  else    return(0);
}

/*----------------------------------------------------------------------*/
/************************************************************************/
/*
	oldImages

	Old image processing functions kept for compatibility

Those functions have been kept only for compatibility reasons. Don't use
them because they could deseapear in a near future.
  
  (c) Jean-Philippe THIRION -- INRIA -- 1990
*/
/************************************************************************/
/*----------------------------------------------------------------------*/

/* 
	Transform a single image line into the 2D intensity image

The intensity in a 2D image can be considered as an altitude function i=f(x,y). In that case, this operation is similar to constructing the image of a cross-section of this surface f.

Limitations:
	Use only 2D 1 byte/pixel images.
*/

#if defined(_ANSI_)
void T_Slice( t_ImagePtr theIm, int line, t_ImagePtr resIm )
#else
void T_Slice(theIm, line, resIm )
t_ImagePtr theIm;	/* original 2D image */
int line;			/* number of the line to process in this image */
t_ImagePtr resIm ;	/* result image */

#endif

{
  int value;
  register int c,l;
  
  E_ERRN( theIm->type != TYPEIM_256, E_ERRUNKNOWN, 
	 "T_Slice", "wrong image type" );
  
  resIm->tx= theIm->tx;
  resIm->ty= 256;
  resIm->type= TYPEIM_256;
  
  T_AllocImage( resIm );
  E_ISERRN("T_Slice", NULL);
  
  E_ERRN( (line < 0 ) || (line >= theIm->ty),
	 E_ERRARRAY, "T_Slice", "number of line doesn't fit" );
  
  for( c=0; c<theIm->tx; c++ ){
    
    value= (unsigned char) theIm->buf[theIm->tx * line + c];
    for( l=255; l >= 255 - value; l-- )
      resIm->buf[ resIm->tx * l + c ]= (unsigned char) 255;
  }
}

/*----------------------------------------------------------------------*/

/*
	Unfold a moebius strip image

A moebius strip image is such that its last line can be reversed in x, and is then continuous with the first line. It is transformed into a double sized image which is toric in y, that is, the last line is continuous with the first line. Sinograms for 1/2 turn are moebius strip images.
*/

/*
*/

#if defined(_ANSI_)
void T_UnfoldMoebius( t_ImagePtr theIm )
#else
void T_UnfoldMoebius(theIm )
t_ImagePtr theIm ; /* the image to unfold */

#endif

{
  t_Image auxIm;
  reeltype color;
  int line,col,plan;
  
  if ( theIm->buf == NULL ) return;
  
  T_CopyHeaderIm( theIm, &auxIm );
  auxIm.ty= theIm->ty * 2;
  auxIm.buf= NULL;
  
  T_AllocImage( &auxIm );
  E_ISERRN("T_UnfoldMoebius", NULL );
  
  for( plan= 0; plan < theIm->tz; plan++ )
    for( line= 0; line < theIm->ty; line++ )
      for( col= 0; col < theIm->tx; col++ ){
	
	color= T_GetPoint( col, line, theIm );
	T_SetPoint( col, line, &auxIm, color, MODE_SET );
	T_SetPoint( auxIm.tx - col - 1, line + theIm->ty, &auxIm, 
		   color, MODE_SET );
      }
  
  T_FreeImage( theIm );
  T_CopyHeaderIm( &auxIm, theIm );
  theIm->buf= auxIm.buf;
}

/*----------------------------------------------------------------------*/

/*
	2D median filter of an image (obsolete)

The support of the filter is 3x3. This function has been replaced by T_ImMedianFilter().
*/

#if defined(_ANSI_)
void T_ImageMFilter( t_ImagePtr theIm, t_ImagePtr resIm )
#else
void T_ImageMFilter(theIm, resIm )
t_ImagePtr theIm;	/* input image */
t_ImagePtr resIm ;	/* filtered output image */

#endif

{
  int line,col,k,l;
  int tx,ty;
  reeltype value;
  t_Image auxIm;
  
  T_CopyImage( theIm, &auxIm );
  E_ISERRN("T_ImageMFilter",NULL);
  
  tx= 3;
  ty= 3;
  
  for( k=0; k < ty / 2; k++ )
    for( col=0; col < theIm->tx; col++ ){
      value = T_GetPoint( col, k, theIm );
      T_SetPoint( col, k, &auxIm, value, MODE_SET );
    }
  
  for( line=0; line < theIm->ty - ty  + 1; line++ ){
    
    for( l=0; l< tx / 2; l++ ){
      value = T_GetPoint( l, line, theIm );
      T_SetPoint( l, line, &auxIm, value, MODE_SET );
    }
    
    for( col=0; col < theIm->tx - tx + 1; col++ ){
      
      value= 0;
      for( k=0; k< ty; k++ )
	for( l=0; l< tx; l++ )
	  value += T_GetPoint( col + l, line + k, theIm );
      
      value /= tx * ty;
      T_SetPoint( col + tx / 2, line + ty / 2,
		 &auxIm, value, MODE_SET );
    }
    
    for( l= theIm->tx - tx / 2; l< theIm->tx; l++ ){
      value = T_GetPoint( l, line, theIm );
      T_SetPoint( l, line, &auxIm, value, MODE_SET );
    }
  }
  
  for( k= theIm->ty - ty / 2; k < theIm->ty; k++ )
    for( col= 0; col < theIm->tx; col++ ){
      value = T_GetPoint( col, k, theIm );
      T_SetPoint( col, k, &auxIm, value, MODE_SET );
    }
  
  T_FreeImage( resIm );
  T_CopyImage( &auxIm, resIm );
  T_FreeImage( &auxIm );
  
  E_ISERRN("T_ImageMFilter",NULL);
}

/*----------------------------------------------------------------------*/

/*
	Keep only the first half horizontal part of an image
*/

#if defined(_ANSI_)
void T_ImDivide( t_ImagePtr theIm )
#else
void T_ImDivide(theIm )
t_ImagePtr theIm ;	/* the image to divide */

#endif

{	
  int size;
  char *newBuf;
  
  size= T_ImageSize( theIm ) / 2;
  newBuf= T_Alloc((unsigned) size );
  E_ISERRN( "T_ImDivide",NULL);
  
  (void) memcpy( (void *) newBuf, (void *) theIm->buf, (unsigned) size );
  free( theIm->buf );
  theIm->buf= newBuf;
  theIm->ty= theIm->ty / 2;
}

/*----------------------------------------------------------------------*/

/*
	Compute the gradient in one direction (obsolete)

This function can be advantageously replaced by separate gaussian filter (for example E_SepGradient()). It is only maintained for compatibility reasons.
*/


#if defined(_ANSI_)
void T_ImSizedGrad( t_ImagePtr theIm, t_ImagePtr resIm, int tx, int ty, int tz, reeltype sigma, int dir )
#else
void T_ImSizedGrad(theIm, resIm, tx, ty, tz, sigma, dir )
t_ImagePtr theIm;	/* input image */
t_ImagePtr resIm;	/* filtered output image */
int tx,ty,tz;		/* size of the filter kernel */
reeltype sigma;		/* sigma of the gaussian filter */
int dir ;			/* direction for filtering */

#endif

{
  reeltype *filter,*ptr;
  double x,y,z;
  int i,j,k;
  double sigma2;
  
  filter= (reeltype *) T_Alloc( (unsigned) (sizeof(reeltype) * tx * ty * tz) );
  E_ISERRN("T_ImSizedGrad",NULL);
  
  ptr= filter;
  sigma2= sigma * sigma;
  
  
  for( k=0; k< tz; k++ )
    for( i=0; i< ty; i++ )
      for( j=0; j< tx; j++ ){
	
	x= j - tx / 2.0;
	y= i - ty / 2.0;
	z= k - tz / 2.0;
	
	switch( dir ){
	  
	case DIRECT_X:	
	  *ptr++ = (reeltype) (- x / sigma2 * exp(-(x*x+y*y+z*z)/(2.0*sigma2)));
	  break;
	case DIRECT_Y:	
	  *ptr++ = (reeltype) (- y / sigma2 * exp(-(x*x+y*y+z*z)/(2.0*sigma2)));
	  break;
	case DIRECT_Z:	
	  *ptr++ = (reeltype) (- z / sigma2 * exp(-(x*x+y*y+z*z)/(2.0*sigma2)));
	  break;
	  
	default: 
	  free( (char *) filter );
	  E_XERRN(E_ERRUNKNOWN,"T_ImSizedGrad",
		  "unknown direction"); break; 
	}
      }
  
  T_NormalizeFilter( filter, tx*ty*tz );
  T_ImApplyFilter( theIm, resIm, filter, tx, ty, tz );
  free( (char *) filter );
}

/*----------------------------------------------------------------------*/

/*
	Compute the laplacian of an image (obsolete)

This function can be advantageously replaced by separated gaussian filtering (for example E_SepLaplace()), which is much faster. It is only maintained for compatibility reasons.
*/

#if defined(_ANSI_)
void T_ImSizedLap( t_ImagePtr theIm, t_ImagePtr resIm, int tx, int ty, int tz, reeltype sigma )
#else
void T_ImSizedLap(theIm, resIm, tx, ty, tz, sigma )
t_ImagePtr theIm;	/* input image */
t_ImagePtr resIm;	/* filtered output image */
int tx,ty,tz;		/* size of the filter kernel */
reeltype sigma;		/* sigma of the gaussian filter */

#endif

{
  reeltype *filter,*ptr;
  double x,y,z,sigma2,value;
  int i,j,k;
  
  filter= (reeltype *) T_Alloc( (int) (sizeof(reeltype) * tx * ty * tz) );
  E_ISERRN("T_ImSizedLap", NULL );
  
  ptr= filter;
  sigma2= sigma*sigma;
  
  for( k=0; k< tz; k++ )
    for( i=0; i< ty; i++ ){
      for( j=0; j< tx; j++ ){
	x= j - tx / 2.0;
	y= i - ty / 2.0;
	z= k - tz / 2.0;
	value= (x*x+y*y+z*z-3*sigma2)/(sigma2*sigma2) * 
	  exp(-(x*x+y*y+z*z)/(2*sigma2));
	*ptr++ = (reeltype) value;
      }
    }
  
  T_NormalizeFilter( filter, tx*ty*tz );
  T_ImApplyFilter( theIm, resIm, filter, tx, ty, tz );
  free( (char *) filter );
}

/*----------------------------------------------------------------------*/
