/* epi_kernel.h 2.3 5/14/98 */
/*==========================================================================*/
/*
	epi_kernel.h

	General includes and types for EpidaureLib files

	(c) Jean-Philippe THIRION -- INRIA -- October 25 1992

 *-----------------------------Modification----------------------------------
 *
 * Mars 1999 : X. Pennec
 *                Add voxel size in image headers
 *  
*/
/*==========================================================================*/

#ifndef _epi_kernel_h_
#define _epi_kernel_h_

#ifndef _epi_unix_ 
#include <epi_unix.h>		/* unix principal included files	*/
#endif /* _epi_unix_ */

#ifndef _epi_err_		/* error messages			*/
#include <epi_err.h>
#endif

/*......................................................................*/
/*
	Different types of machines constants
*/
/*......................................................................*/


#define MACHINE_UNKNOWN 0
#define MACHINE_SUN 1
#define MACHINE_DEC 2
#define MACHINE_SGI 3

#ifdef sun 
#define B_MACHINE MACHINE_SUN
#else
#define B_MACHINE MACHINE_DEC
#endif

/*......................................................................*/
/*
	Swap of the integers and shorts on 68000 machines
*/
/*......................................................................*/

#if (sgi || sun || ridge || convex || hpux)
#define SWAP(x) ( ((((unsigned int) (x)) & 0x000000ff ) << 24 ) |\
                  ((((unsigned int) (x)) & 0x0000ff00 ) << 8 ) |\
                  ((((unsigned int) (x)) & 0x00ff0000 ) >> 8 ) |\
                  ((((unsigned int) (x)) & 0xff000000 ) >> 24 ) )
#else
#define SWAP(x) (x)
#endif
#define SWP(x) ( ((((unsigned int) (x)) & 0x000000ff ) << 24 ) |\
                 ((((unsigned int) (x)) & 0x0000ff00 ) << 8 ) |\
                 ((((unsigned int) (x)) & 0x00ff0000 ) >> 8 ) |\
                 ((((unsigned int) (x)) & 0xff000000 ) >> 24 ) )

/*......................................................................*/
/*
	Principal pre-defined constants and flags
*/
/*......................................................................*/

#define TRUE		1
#define FALSE		0
#define PLUS		1
#define MINUS		-1
#define CLOSED		1
#define OPENED		0
#ifndef PI
#define PI		3.14159265358979323846
#endif
#define DEUXPI	 	6.28318530717958647692	
#define EPSY		0.00001
/* #define INFINITY	1e+20 */
#define RNULL		(reeltype *) NULL
#define CNULL		(char *) NULL

#define DIRECT_X		0		/* directions				*/
#define DIRECT_Y		1
#define DIRECT_Z		2
#define UNKNOWN			0
#define WEST			1
#define EAST			2
#define NORTH			4
#define SOUTH			8
#define DOWN			16	
#define UP				32	
#define DIRECT			1
#define REVERSE			-1

#define E_LINEAR		1
#define E_SPLINE		2	

/* typedef float reeltype; */

typedef float reeltype;

#ifndef NO_PROTO
typedef int (*void_func_type)( void );
typedef int (*func_type_2_arg)(const void *, const void *);
#else
typedef int (*void_func_type)( );
typedef int (*func_type_2_arg)( );
#endif


/*
extern char program[];
extern unsigned char verbose;
*/

/*......................................................................*/
/*
	Principal low level data structures
*/
/*......................................................................*/


typedef struct t_2IPoint{		/* 2D integer point */

	int x,y;
	} t_2IPoint, *t_2IPointPtr;


typedef struct t_2Point{		/* a 2d point	*/

	reeltype	x,y;
	} t_2Point, *t_2PointPtr;


typedef struct t_3Point{		/* a 3d point	*/

	reeltype	x,y,z;
	} t_3Point, *t_3PointPtr;


typedef struct t_3IPoint{		/* a 3d integer point	*/

	int	x,y,z;
	} t_3IPoint, *t_3IPointPtr;


typedef struct t_3IBox{		/* a 3D integer box	*/

	t_3IPoint min,max;
	} t_3IBox, *t_3IBoxPtr;

/*......................................................................*/
/*
	structure definitions for epidaure.lib objects
*/
/*......................................................................*/

#define OBJ_BEGINARRAY -5 
#define OBJ_ENDARRAY -4 
#define OBJ_ENDLOOP -3 
#define OBJ_ENDLIST -2 
#define OBJ_EOF -1 

#define OBJ_UNKNOWN 0
#define OBJ_LIST 1
#define OBJ_3MAP 2
#define OBJ_3PDUAL 3
#define OBJ_CREST 4
#define OBJ_INT 5
#define OBJ_POINT 6
#define OBJ_ARRAY 7
#define OBJ_MAT4 8
#define OBJ_OBJDEF 9
#define OBJ_FLOAT 10 
#define OBJ_STRING 11 
#define OBJ_MAT 12 				/* N-dimensional matrix		*/
#define OBJ_NPOINT 13			/* no more cross compatible with array */
#define OBJ_NB 14

#define OBJ_USER 0x80000000		/* user's fixed length types */

#define OBJF_UNKNOWN 0
#define OBJF_ALLOC 1
#define OBJF_FREE 2
#define OBJF_FREAD 3
#define OBJF_FWRITE 4
#define OBJF_COPY 5
#define OBJF_DEFAULT 6
#define OBJF_NB 7


typedef struct t_Obj{	/* EpidaureLib generic object */

	long type;			/* object type */

	} t_Obj,*t_ObjPtr,**t_ObjHand;

#ifndef NO_PROTO
typedef int (*t_ObjPtr_func_type )( t_ObjPtr );
typedef int (*t_ObjHand_func_type )( t_ObjHand );
typedef int (*t_PtrHand_func_type )( t_ObjPtr, t_ObjHand );
typedef int (*t_fdcharObjP_func_type )( FILE *, char *, t_ObjPtr );
typedef int (*t_fdcharObjH_func_type )( FILE *, char *, t_ObjHand );
#else
typedef int (*t_ObjPtr_func_type )( );
typedef int (*t_ObjHand_func_type )( );
typedef int (*t_PtrHand_func_type )( );
typedef int (*t_fdcharObjP_func_type )( );
typedef int (*t_fdcharObjH_func_type )( );
#endif

typedef struct t_3Map{				/* element of a map (obsolete)  	*/

	t_Obj obj;						/* a 3Map is an object				*/
	struct t_3Map *n[4];			/* neigbours						*/
	t_3Point pt;					/* position of the node				*/
	reeltype weight;				/* confidance in that point			*/
	int num;						/* number of the map				*/
	int state;						/* index for global processing		*/
	
	} t_3Map,*t_3MapPtr,**t_3MapHand;


typedef struct t_List{				/* generic structure of lists		*/

	t_Obj obj;						/* a list is an object 				*/
	struct t_List *next;			/* next element of the list			*/
	struct t_List *back;			/* list in the reverse order		*/
	t_ObjPtr item;					/* an element of the list			*/
	} t_List,*t_ListPtr,**t_ListHand;

#ifndef NO_PROTO
typedef void (*theFunc_type)( t_ObjPtr, int, char *[] );
#else
typedef void (*theFunc_type)( );
#endif

#define LNULL (t_ListPtr) NULL

typedef struct t_3PDual{			/* (obsolete) */
	
	t_Obj obj;						/* a 3PDual is an object			*/
	t_3Point pt;					/* position of the point			*/
	t_3Point nm;					/* direction of the normal			*/
	reeltype weight;				/* confidence in that point			*/

	} t_3PDual,*t_3PDualPtr,**t_3PDualHand;


typedef struct t_Crest{				/* crest line element				*/
	
	t_Obj obj;						/* a Crest elem. is an object		*/
	t_3Point pt;					/* position of the point			*/
	t_3Point t1;					/* first princ. direction			*/
	t_3Point t2;					/* second princ. direction			*/
	reeltype k1;					/* first princ. curvature			*/
	reeltype k2;					/* second princ. curvature			*/
	reeltype conf;					/* confidence in that point			*/
									/* principal curvature				*/
	} t_Crest,*t_CrestPtr,**t_CrestHand;


typedef struct t_Point{				/* 3D point as an object			*/
	
	t_Obj obj;						/* a Crest elem. is an object		*/
	t_3Point pt;					/* position of the point			*/
	} t_Point,*t_PointPtr,**t_PointHand;


typedef struct t_Int{				/* an integer object			*/

	t_Obj obj;						
	int index;
	} t_Int,*t_IntPtr,**t_IntHand;


typedef struct t_Float{				/* a floating point object		*/

	t_Obj obj;						
	float v;
	} t_Float,*t_FloatPtr,**t_FloatHand;


typedef struct t_String{			/* a character string object	*/

	t_Obj obj;						
	char *str;
	} t_String,*t_StringPtr,**t_StringHand;
	

typedef struct t_Array{     /* array of fixed size objects          */

    t_Obj obj;              /* this is an object (OBJ_ARRAY)        */
    int nbAlloc;            /* present number of allocated objects  */
    char *tb;               /* theTber toward the objects buffer    */
							/* more internal information			*/
    t_Obj elem;             /* type of the objects in the array     */
    int nb;                 /* number of objects in the array       */

    } t_Array, *t_ArrayPtr, **t_ArrayHand;


typedef struct t_NPoint{    /* N-Dimensional point (no more compatible with arrays) */
    t_Obj obj;              /* this is an object (OBJ_NPOINT)       */
    int dim;           		/* dimension of the vector				*/
    reeltype *v;            /* array representing the vector		*/

							/* no more compatible with arrays	    */
    t_Obj reserved1;        /* reserved DON'T USE THIS FIELD		*/
    int reserved2;          /* reserved DON'T USE THIS FIELD		*/

    } t_NPoint, *t_NPointPtr, **t_NPointHand;

							/* give the element Pi of the point		*/
							/* EPP= pointer, EPV= value				*/

#define EPP(V,i) ((V)->v + (i))
#define EPV(V,i) (*EPP(V,i))

typedef struct t_Mat{		/* N-Dimensional matrix					*/

	t_Obj obj;				/* the matrix is an object				*/
	int nL;					/* number of lines						*/
	int nC;					/* number of column						*/
    reeltype *v;            /* array representing the vector		*/

	} t_Mat, *t_MatPtr, **t_MatHand;

							/* give the element Aij of the matrix	*/
							/* EMP= pointer, EMV= value				*/

#define EMP(A,i,j) ((A)->v + (j) * (A)->nL + (i))
#define EMV(A,i,j) (*EMP(A,i,j))


typedef struct t_Mat4{		/* 4x4 matrix for homogeneous transforms	*/
	t_Obj obj;						
	reeltype mt[4][4];
	} t_Mat4,*t_Mat4Ptr,**t_Mat4Hand;


typedef struct t_ObjDef{		/* new object user type definition	*/
	t_Obj obj;						

	int init;					/* TRUE if initialized, else FALSE	*/
	int sizeUtil;
	void_func_type f[ OBJF_NB ];

	} t_ObjDef, *t_ObjDefPtr, **t_ObjDefHand;


/*......................................................................*/
/*
	Epidaure.lib file system
*/
/*......................................................................*/


#define MAGIC_NAME "(c) INRIA - Epidaure files (Header 512 bytes)"

#define TYPEFI_UNKNOWN 0        /* types of thomy files                 */
#define TYPEFI_IMAGE 1
#define TYPEFI_SEGMENTS 2
#define TYPEFI_CUBESURF 3
#define TYPEFI_OBJLIST 4

#define MODE_BIN 0
#define MODE_ASCII 1

#define B_FNAME_LENGTH 512

typedef struct t_File{		/* file description */

    int     fd;             		/* unix file descriptor     */
    FILE    *fdt;           		/* ascii file descriptor    */
    int     type;           		/* type of thomy file       */
    char    name[B_FNAME_LENGTH];	/* name of the file         */
    int     mode;     				/* MODE_BIN or MODE_ASCII   */
    int     machine;        		/* machine who produce the file */

    } t_File,*t_FilePtr;


/*......................................................................*/
/*
	Epidaure.lib images 
*/
/*......................................................................*/


#define TYPEIM_UNKNOWN 	0   /* not a known type of thomy images		*/
#define TYPEIM_BW 	1   /* black and white picture		     	*/
#define TYPEIM_256	2   /* grey level image one byte encoded	*/
#define TYPEIM_FLOAT 	3   /* a floating point image	      		*/
#define TYPEIM_16B 	4   /* a 2 byte points image	       		*/
#define TYPEIM_S16B 	5   /* a signed 2 byte points image    		*/

/*................ access modes to images .............................*/

#define MODE_SET 	1		/* set the value of the point			*/
#define MODE_ADD	2		/* add the value to the point			*/
#define MODE_ADD2 	3		/* add the square value to the point	*/
#define MODE_MASK 15		/* mask for imformation about coposit.	*/

							/* ways to process side effects			*/

#define MODE_1_EXTEND	1		/* repeat the last value of image   */
#define MODE_1_CLIP		2		/* 0 out of the bounds			    */
#define MODE_1_UNFOLD	3		/* repeat reversed image            */

#define MODE_X_SHIFT	16		/* (0x00030000)						*/
#define MODE_Y_SHIFT	18		/* (0x000C0000)						*/
#define MODE_Z_SHIFT	20		/* (0x00300000)						*/

#define MODE_X_MASK		(3 << MODE_X_SHIFT)
#define MODE_Y_MASK		(3 << MODE_Y_SHIFT)
#define MODE_Z_MASK		(3 << MODE_Z_SHIFT)

#define MODE_X_EXTEND	(MODE_1_EXTEND	<< MODE_X_SHIFT)
#define MODE_Y_EXTEND	(MODE_1_EXTEND	<< MODE_Y_SHIFT)
#define MODE_Z_EXTEND	(MODE_1_EXTEND	<< MODE_Z_SHIFT)

#define MODE_X_CLIP		(MODE_1_CLIP	<< MODE_X_SHIFT)
#define MODE_Y_CLIP		(MODE_1_CLIP	<< MODE_Y_SHIFT)
#define MODE_Z_CLIP		(MODE_1_CLIP	<< MODE_Z_SHIFT)

#define MODE_X_UNFOLD	(MODE_1_UNFOLD	<< MODE_X_SHIFT)
#define MODE_Y_UNFOLD	(MODE_1_UNFOLD	<< MODE_Y_SHIFT)
#define MODE_Z_UNFOLD	(MODE_1_UNFOLD	<< MODE_Z_SHIFT)

#define MODE_FOLD_MASK  (MODE_X_MASK	| MODE_Y_MASK	| MODE_Z_MASK)
#define MODE_EXTEND		(MODE_X_EXTEND	| MODE_Y_EXTEND	| MODE_Z_EXTEND)
#define MODE_CLIP		(MODE_X_CLIP	| MODE_Y_CLIP	| MODE_Z_CLIP)
#define MODE_UNFOLD		(MODE_X_UNFOLD	| MODE_Y_UNFOLD	| MODE_Z_UNFOLD)


								/* old modes kept for consistency		*/
#define MODE_BUF 15				/* bufferize information about segments	*/
#define MODE_BUFLEVEL 128		/* level to bufferize information		*/
#define MODE_CONFI 0x02000000	/* display confidence value (of dual pt)*/
#define MODE_VEC 0x04000000		/* display vectors						*/
#define MODE_PTS 0x08000000		/* display only points for segments		*/
#define MODE_VER 0x10000000		/* verifie if the points in the images	*/

/*........... image structure .......................................*/

typedef struct t_Image{	/* 3D image header */
  
  t_Obj	obj;		 /* unused now */
  int	tx,ty,tz;        /* size of the image	*/
  char	type;	         /* type of the image   */
  char	*buf;		 /* the buffer with the image	*/
  t_File  fi;            /* thomy file descr. */
  reeltype sx,sy,sz;     /* voxel size */
} t_Image, *t_ImagePtr;

#ifndef NO_PROTO
typedef reeltype (*ReechParPtr_func_type)( reeltype, reeltype, reeltype, t_ImagePtr, void * );
#else
typedef reeltype (*ReechParPtr_func_type)( );
#endif

/*----------------------------------------------------------------------*/
/*		3D transform using 4D homogeneous operations					*/
/*----------------------------------------------------------------------*/

#define TR_DIR 0			/*	direct transform		*/
#define TR_INV 1			/*	inverse transform		*/
#define TR_NORM 2			/*	transforms for normals	*/

typedef struct t_4Tr{		/* direct and inverse 3D transform */

	t_Mat4 dir;						/* direct transform					*/
	t_Mat4 inv;						/* inverse transform				*/
	} t_4Tr, *t_4TrPtr, **t_4TrHand;

/*----------------------------------------------------------------------*/
/*		Structures for hashing											*/
/*----------------------------------------------------------------------*/

typedef struct t_Entry{	/* entry of a hash table */
	struct t_Entry *next;	/* next entry to process collisions */
	int code;				/* complete code associated to the entry */
	char *item;				/* item attached to that entry */
	} t_Entry,*t_EntryPtr,**t_EntryHand;

#ifndef NO_PROTO
typedef int (*fCode_type)( char * );
typedef int (*filter_type)(char *, char * );
#else
typedef int (*fCode_type)( );
typedef int (*filter_type)( );
#endif

typedef struct t_HTable{	/* Hash table structure */

	int size;				/* number of bits for code truncation */
	int nb;					/* number of separated buckets (2^size) */
	int mask;				/* pre-computed mask for code trucation */
	t_EntryHand tb;			/* array of buckets */
	t_EntryPtr search;		/* last explored bucket for search */
	int code;				/* code memorized for next search */
	char *data;				/* reference item memorized for search */
	fCode_type fCode;		/* for data encoding	*/
	filter_type filter;		/* filtering function	*/

	} t_HTable,*t_HTablePtr,**t_HTableHand;


/*----------------------------------------------------------------------*/
/*		Data structure for the computation of any derivatives filters	*/
/*	up to 3rd order.													*/
/*----------------------------------------------------------------------*/

#define NORM_SUM  1
#define NORM_DIFF 2
#define NORM_INV  4

typedef struct t_3rdFilters{	/* Structure for 3d image filtering */

    reeltype sx,sy,sz;	        /* sigma according to x,y and z */
    int tx,ty,tz;		        /* size of the filter			*/
    int mx,my,mz;				/* lower bound of filter		*/
    int Mx,My,Mz;				/* uper bound of filter			*/
    t_3Point scale;				/* scale of the voxel			*/
    reeltype *d[4][4][4];       /* arrays of filters			*/
    reeltype *im;				/* values of the pixels			*/
    int init;					/* true if initialized			*/
    int boardModes;				/* treat board effects          */
    int normModes;              /* normalize the filter         */

	} t_3rdFilters, *t_3rdFiltersPtr;

typedef struct t_3Derivates{ /* function and function derivatives up to order 3 */

	reeltype f;
	reeltype fx,fy,fz;
	reeltype fxx,fxy,fxz,fyy,fyz,fzz;
	reeltype fxxx,fxxy,fxxz,fxyy,fxyz,fxzz,fyyy,fyyz,fyzz,fzzz;

	} t_3Derivates, *t_3DerivatesPtr;


typedef struct t_SurfDer{	/* surface differential characteristics */

    reeltype g,m;           /* gaussian and mean curvature              */
    reeltype k1,k2;         /* principal curvature                      */
    reeltype norm;          /* of the gradient                          */
    t_3Point grad;          /* gradient of the surface                  */
    t_3Point t1,t2;         /* principal directions                     */
    t_3Point dk1,dk2;       /* derivatives of princ. curv.              */
    t_3Point dK,dS;         /* derivatives of gaussian and mean curv.   */
    reeltype e1,e2;         /* extremality coefficient                  */
    int good;               /* false if the extremality is degenerated  */

    } t_SurfDer, *t_SurfDerPtr;


typedef struct t_CurvDer{	/* differentials for iso-contours in 2D		*/

	t_Obj obj;				/* is an object...							*/
	
	t_3Point pt;			/* position of the point in 3D				*/
    reeltype c;           	/* implicit curvature  			            */
    reeltype norm;          /* of the gradient                          */
    reeltype n[2];          /* normal to the curve						*/
    reeltype dC[2];         /* gradient of the curvature				*/
    reeltype e;         	/* extremality coefficient                  */
    int good;               /* false if the extremality is degenerated  */

    } t_CurvDer, *t_CurvDerPtr;

typedef struct e_ReechPar{		/* parameters for resampling 3D images	*/

	t_Mat4 trsf;				/* transform in homogeneous coordinates	*/
	int type;					/* type of the resampling				*/
	int verbose;				/* display messages if TRUE				*/

	} e_ReechPar, *e_ReechParPtr;

/*----------------------------------------------------------------------*/

#endif
