/* epi_math.c 2.1 5/14/98 */
/*==========================================================================
 *
 * File : $EPI/src/epi_math.c
 * Module : ?
 * $ Id : ? 
 * Test file :
 * Author : S Benayoun & J-P Thirion
 * Creation: June 28, 1993
 *===========================================================================
 *   (c) Copyright 1992-93, Inria, all right reserved
 *===========================================================================
 *
 *------------------------------Description----------------------------------
 *
 * Operations mathematiques et matricielles
 *
 *-----------------------------Modification----------------------------------
 *
 */

/*=========================================================================*/
/*
	epi_math.c

  General mathematical operations on points and matrices

	S. Benayoun and J-P Thirion
*/
/*=========================================================================*/

#if defined(INC_MIN) /* inclusions minimales */
 /* rajouter -DINC_MIN dans les options de compilation */
 /* GM 12/09/94                                        */
#include <epi_kernel.ei>
#include <epi_math.ei>
#else /* inclusions maximales */
#include <epidaure.h>
#include <epidaure.ee>
#include <epidaure.ei>
#endif

/************************************************************************/
/*
  OpersMat

  General operations with N-Dimensional Matrix
  */
/************************************************************************/

/*
  Test function on matrices
  */

#if defined(_ANSI_)
void T_TestMat( void )
#else
void T_TestMat()

#endif

{
  t_MatPtr aMat= NULL;
  
  T_FReadOne( (FILE *) NULL, "test.mat", (t_ObjHand) &aMat );
  E_ERRN( aMat->obj.type != OBJ_MAT, E_ERRMISMATCH, "T_MatTest",
	 "not a matrix" );
  
  T_MatTrans( aMat, aMat );	
  T_FWriteOne( (FILE *) NULL, "tmp.resmat", (t_ObjPtr) aMat );
  T_FreeMat( &aMat );
  
  E_ISERRN( "T_MatTest", (char*)NULL );
}

/*
  Transpose a given matrix
  */

#if defined(_ANSI_)
void T_MatTrans( t_MatPtr mat, t_MatPtr resMat )
#else
void T_MatTrans(mat, resMat )
t_MatPtr mat;
t_MatPtr resMat ;

#endif

{
  int l,c;
  t_MatPtr aux=NULL;
  
  T_AllocMat( &aux );
  T_AllocMatTb( aux, mat->nC, mat->nL );
  E_ISERRN( "T_MatTrans", (char*)NULL );
  
  for( l=0; l< mat->nL; l++ )
    for( c=0; c< mat->nC; c++ )
      EMV( aux, c, l )= EMV( mat, l, c );
  
  T_Free( (void **) &resMat->v );
  resMat->v= aux->v;
  resMat->nC= aux->nC;
  resMat->nL= aux->nL;
}


/*
  Add two matrices
  */

#if defined(_ANSI_)
void T_MatAdd( t_MatPtr mat1, t_MatPtr mat2, t_MatPtr resMat )
#else
void T_MatAdd(mat1, mat2, resMat )
t_MatPtr mat1;
t_MatPtr mat2;
t_MatPtr resMat ;

#endif

{
  int l,c;
  
  E_ERRN( (mat1->nL != mat2->nL) ||
	 (mat1->nC != mat2->nC), E_ERRFATAL, "T_MatAdd",
	 "matrix sizes differ" );
  
  T_AllocMatTb( resMat, mat1->nL, mat1->nC );
  E_ISERRN( "T_MatAdd", (char*)NULL );
  
  for( l=0; l< mat1->nL; l++ )
    for( c=0; c< mat1->nC; c++ )
      EMV( resMat, l, c )= EMV(mat1,l,c) + EMV(mat2,l,c);
}	

/*
  Fill a matrix with zero values
  */

#if defined(_ANSI_)
void T_ZeroMat( t_MatPtr mat )
#else
void T_ZeroMat(mat )
t_MatPtr mat ;

#endif

{
  int l,c;
  
  if ( mat->v == NULL )
    return;
  
  for( c=0; c<  mat->nC; c++ )
    for( l=0; l< mat->nL; l++ )
      EMV( mat, l, c )= 0.0;
}

/*
  Multiply two matrices : resMat= mat1 x mat2= mat1(mat2())
  */

#if defined(_ANSI_)
void T_MatMul( t_MatPtr mat1, t_MatPtr mat2, t_MatPtr resMat )
#else
void T_MatMul(mat1, mat2, resMat )
t_MatPtr mat1;
t_MatPtr mat2;
t_MatPtr resMat ;

#endif

{
  t_MatPtr aux= NULL;
  int i,l,c;
  
  E_ERRN( mat1->nC != mat2->nL, E_ERRFATAL, "T_MatMul",
	 "matrix sizes incompatible" );
  
  T_AllocMat( &aux );
  T_AllocMatTb( aux, mat1->nL, mat2->nC );
  E_ISERRN( "T_MatMul", (char*)NULL );
  
  for( l=0; l< aux->nL; l++ )
    for( c=0; c< aux->nC; c++ ){
      EMV(aux,l,c)= 0.0;
      for( i=0; i< mat1->nC; i++ )
	EMV(aux,l,c) += EMV(mat1,l,i) * EMV(mat2,i,c);
    }
  
  T_CopyMat( aux, &resMat );
  T_FreeMat( &aux );
}	

/* 
  Compute the scalar product between two N-points
  */

#if defined(_ANSI_)
reeltype T_Scalar( t_NPointPtr p1, t_NPointPtr p2 )
#else
reeltype T_Scalar(p1, p2 )
t_NPointPtr p1;
t_NPointPtr p2 ;

#endif

{
  int i;
  reeltype aux;
  reeltype res= 0.0;
  
  E_ERR( (p1->dim != p2->dim) , E_ERRFATAL, 0.0, "T_Scalar",
	 "vectors dimension differ" );
  
  for( i=0; i< p1->dim; i++ ){
    aux= EPV(p1,i) * EPV(p2,i);
    res += aux;
  }
  return( (reeltype) res );
}


/*
  Apply a vector to a matrix
  */

#if defined(_ANSI_)
void T_ApplyVecToMat( t_MatPtr mat, t_NPointPtr vec, t_NPointPtr resvec )
#else
void T_ApplyVecToMat(mat, vec, resvec )
t_MatPtr mat;
t_NPointPtr vec;
t_NPointPtr resvec ;

#endif

{
  int l,m;
  t_NPointPtr aux=NULL;
  
  E_ERRN( (mat->nC != vec->dim), E_ERRFATAL, "T_ApplyVecToMat",
	 "vector and matrix dimensions not compatible" );
  
  T_ReAllocNPointTb( resvec, mat->nL );
  E_ISERRN( "T_ApplyVecToMat", (char*)NULL );
  T_NClean( resvec );              /* met les valeurs a zero */
  
  if( vec == resvec){
    T_CopyNPoint(vec,&aux);
    vec = aux;
  }
  
  for( l=0; l< mat->nL; l++ )
    for( m=0; m< mat->nC; m++ )
      EPV(resvec,l) += EMV(mat,l,m) * EPV(vec,m);
  
  if(aux)
    T_FreeNPoint(&aux);
}


/*
  Compute the trace of a square matrix
  */

#if defined(_ANSI_)
void T_MatTrace( t_MatPtr mat, reeltype *t )
#else
void T_MatTrace(mat, t )
t_MatPtr mat;
reeltype *t ;

#endif

{
  int l;
  reeltype tr=0.0;
  
  E_ERRN( (mat->nL != mat->nC) , E_ERRFATAL, "T_MatTrace",
	 "not a square matrix" );
  
  for( l=0; l< mat->nL; l++ )
    tr += EMV(mat,l,l);
  
  *t = tr;
}



/*
  Multiply a matrix with a scalar : resMat= mat1 x s 
  */

#if defined(_ANSI_)
void T_MatMulScal( t_MatPtr mat1, reeltype s, t_MatPtr resMat )
#else
void T_MatMulScal(mat1, s, resMat )
t_MatPtr mat1;
reeltype s;
t_MatPtr resMat ;

#endif

{
  int l,c;
  
  T_AllocMatTb( resMat, mat1->nL, mat1->nC );
  E_ISERRN( "T_MatMulScal", (char*)NULL );
  
  for( l=0; l< mat1->nL; l++ )
    for( c=0; c< mat1->nC; c++ )
      EMV(resMat,l,c)= EMV(mat1,l,c) * s;
  
}	

/*----------------------------------------------------------------------*/


/************************************************************************/
/*
  GradConj

  Conjugate gradient method to solve linear systems
  */
/************************************************************************/

/* +++++++++++++++++++++++++++++++++++++++++++++

GGGG   
   ALGORITHME DU GRADIENT CONJUGUE  
   METHODE DE FLETCHER-REEVES
   et  METHODE DE POLAK RIBIERE
   
   REMARQUE IMPORTANTE :
   Cette procedure garantit la convergence
   vers le minimum global seulement si la
   matrice est SYMETRIQUE DEFINIE POSITIVE.
   Sinon, elle converge vers un minimum
   local dependant de la solution initiale.
   
   ++++++++++++++++++++++++++++++++++++++++++++ */

/*  Conjugate gradient method to solve linear systems */
  
#if defined(_ANSI_)
void E_GradConj( t_MatPtr A, t_NPointPtr b, t_NPointPtr x )
#else
void E_GradConj(A, b, x )
t_MatPtr A;
t_NPointPtr b;
t_NPointPtr x ;

#endif
{
  t_GradConjPar par;
  
  T_DefaultGradConj( &par  );
  E_GradConjGeneral( A, b, x, &par );
}


/*
  Initialize parameter values for the conjugate gradient method
  */


#if defined(_ANSI_)
void T_DefaultGradConj( t_GradConjParPtr par  )
#else
void T_DefaultGradConj(par  )
t_GradConjParPtr par  ;

#endif
{
  par->nbIterMax= 25;
  par->eps= (reeltype) 1e-10;
  par->func= T_ApplyVecToMat;
  par->nbIter= 0;
  par->vtest=100.0;
  par->mode= E_GCONJ_POLAK;
  par->affiche= NON_AFFICHAGE;
}


/*
  General procedure for the conjugate gradient method
  */


#if defined(_ANSI_)
void E_GradConjGeneral( t_MatPtr A, t_NPointPtr c, t_NPointPtr y, t_GradConjParPtr par )
#else
void E_GradConjGeneral(A, c, y, par )
t_MatPtr A;
t_NPointPtr c;
t_NPointPtr y;
t_GradConjParPtr par ;

#endif
{
  
  reeltype lambda,ro,sc1,sc2,sc3,sc4,sc5,Ei,Ef=0.0,ng;
  int i,iter=1;
  reeltype *v1,*v2,*v1dep,*v2dep;
  reeltype *g,*w,*gf,*gdep,*wdep,*gfdep;
  reeltype *x,*b,*xdep,*bdep;
  t_NPointPtr tbV[NB_TAB];
  
  /* allocattions dynamiques */
  
  for(i=0;i<NB_TAB;i++){
    
    T_AllocNPoint( &tbV[i] );
    T_AllocNPointTb( tbV[i], c->dim);
    T_NClean( tbV[i] );
  }
  
  E_REPERR( goto END_GRAD_CONJ);
  
  v1 = tbV[0]->v;
  v2 = tbV[1]->v;
  g = tbV[2]->v;
  w = tbV[3]->v;
  gf =tbV[4]->v;
  x = y->v;
  b = c->v;
  
  v1dep=v1;v2dep=v2;gdep=g;
  gfdep=gf;wdep=w;xdep=x;bdep=b;
  
  par->func ( A, y, tbV[0]);
  
  /* calcul de v1 = Ax0 - b */
  
  for(i=0;i<c->dim;i++)
    
    *v1++ -= *b++;
  
  v1=v1dep;b=bdep;
  
  /* erreur initiale */
  
  Ei = T_NNorm( tbV[0] );
  Ei = Ei/(reeltype)c->dim;
  v1 = v1dep;
  
  /* calcul de g = A ( Ax0 - b ) */
  
  par->func ( A, tbV[0], tbV[2]);
  
  /* assignation   w = g */
  memcpy((void*)w,(void*)g,sizeof(reeltype)*c->dim);
  w=wdep;g=gdep;
  
  ng = T_NNorm( tbV[2] );
  g=gdep;
  ng = ng/(reeltype)c->dim;
  
  
  while(ng > par->eps && iter < par->nbIterMax){

        if  (par->affiche == AFFICHAGE)
            printf("Nombre d'iterations = %3d  Erreur Finale = %5.2e \n",
                   iter,ng);
    
    iter++;
    
    par->nbIter ++;

    /* calcul de v2 = A w */
    
    par->func ( A, tbV[3], tbV[1]);
    
    /* calcul de ro */
    
    sc1 = T_Scalar( tbV[1], tbV[0] );
    
    v2=v2dep;v1=v1dep;
    
    sc2 = T_NNorm( tbV[1] );
    sc2 *= sc2;
    
    v2=v2dep;
    
    if (sc2 != 0.0)
      ro = sc1/sc2;
	else ro= 0.0;
    
    /* calcul de Xn+1 = Xn - ro . w */
    
    for(i=0;i<c->dim;i++)
      
      *x++ -= ro * (*w++);
    
    x=xdep;w=wdep;
    
    /* calcul de v1 = A Xn */
    
    par->func ( A, y, tbV[0]);
    
    /* calcul de v1 = A Xn - b */
    
    for(i=0;i<c->dim;i++)
      
      *v1++ -= *b++;
    
    v1=v1dep;b=bdep;
    
    /* erreur finale */
    
    Ef = T_NNorm( tbV[0] );
    v1=v1dep;
    
    /* calcul de Gn+1 = A v1 */
    
    par->func ( A, tbV[0], tbV[4]);
    
    /* calcul de lambda */
    
    sc3 = T_NNorm( tbV[4] );
    sc3 *= sc3;
    gf=gfdep;
    sc4 = T_NNorm( tbV[2] );
    sc4 *= sc4;
    g=gdep;
    
    if (par->mode == E_GCONJ_FLETCHER)
      
      lambda = sc3/sc4;
    
    else{
      
      sc5 = T_Scalar( tbV[2], tbV[4] );
      g=gdep;gf=gfdep;
      lambda = (sc3-sc5)/sc4;
    }
    
    /* calcul de Wn+1 = Gn+1 + lambda Wn */
    
    for(i=0;i<c->dim;i++)
      
      *w++ *= lambda;
    
    w=wdep;
    
    for(i=0;i<c->dim;i++)
      
      *w++ += *gf++;
    
    w=wdep;gf=gfdep;
    
    /* assignation g = gf */
    
    memcpy((void*)g,(void*)gf,sizeof(reeltype)*c->dim);
    g=gdep;gf=gfdep;
    
    ng = T_NNorm( tbV[2] );
    g=gdep;
    ng = ng/(reeltype)c->dim;
  }

  par->vtest = ng;
  
  /* ecriture du resultat */
  
  for(i=0;i<c->dim;i++)
    
    EPV( y, i ) = *x++;
  
  x=xdep;
  
  Ef = (reeltype) sqrt(Ef/((reeltype)c->dim)); 
  
  if  (par->affiche == AFFICHAGE)
        printf("Nombre d'iterations = %3d  Erreur Finale = %5.2e (%5.2e) \n",
               iter,ng,Ef);

  x=xdep;
  
  /* liberations de memoire */
  
 END_GRAD_CONJ:
  
  for(i=0;i<NB_TAB;i++)
    
    T_FreeNPoint( &tbV[i] );
  
  E_ISERRN("E_GradConjGeneral",(char*)NULL);
  
}

/*----------------------------------------------------------------------*/

/* +++++++++++++++++++++++++++++++++++++++++++++
   
   ALGORITHME DU GRADIENT CONJUGUE  
   VERSION OPTIMISEE
   
   REMARQUE IMPORTANTE :
   Cette procedure garantit la convergence
   vers le minimum global seulement si la
   matrice est SYMETRIQUE DEFINIE POSITIVE.
   Sinon, elle converge vers un minimum
   local dependant de la solution initiale.
   
   ++++++++++++++++++++++++++++++++++++++++++++ */

/*
  Optimized version of the conjugate gradient method
  */

#if defined(_ANSI_)
void E_FastGradConj( t_MatPtr A, t_NPointPtr b, t_NPointPtr x )
#else
void E_FastGradConj(A, b, x )
t_MatPtr A;
t_NPointPtr b;
t_NPointPtr x ;

#endif
{
  t_FastGradConjPar par;
  
  T_DefaultFastGradConj( &par  );
  E_FastGradConjGeneral( A, b, x, &par );
}

/*
  Initialization for the optimized conjugate gradient method
  */

#if defined(_ANSI_)
void T_DefaultFastGradConj( t_FastGradConjParPtr par  )
#else
void T_DefaultFastGradConj(par  )
t_FastGradConjParPtr par  ;

#endif
{
  par->nbIterMax= 25;
  par->eps= (reeltype) 1e-10;
  par->func= T_ApplyVecToMat;
  par->nbIter= 0;
  par->vtest=100.0;
  par->affiche= NON_AFFICHAGE;
}

/*
  High level routine for the optimized conjugate gradient method
  */

#if defined(_ANSI_)
void E_FastGradConjGeneral( t_MatPtr A, t_NPointPtr c, t_NPointPtr y, t_FastGradConjParPtr par )
#else
void E_FastGradConjGeneral(A, c, y, par )
t_MatPtr A;
t_NPointPtr c;
t_NPointPtr y;
t_FastGradConjParPtr par ;

#endif
{
  
  reeltype sc1,sc2,sc3,Ei,Ef,ng,alpha,beta;
  int i;
  reeltype *p0,*p1,*r0,*r1,*Ap;
  reeltype *x,*b;
  reeltype *p0dep,*p1dep,*r0dep,*r1dep,*Apdep;
  reeltype *xdep,*bdep;
  reeltype nb;
  t_NPointPtr tbV[NB_TAB];
  int iterG=0;
  
  /* allocattions dynamiques */
  
  for(i=0;i<NB_TAB;i++){
    
    T_AllocNPoint( &tbV[i] );
    T_AllocNPointTb( tbV[i], c->dim);
    T_NClean( tbV[i] );
  }
  
  p0 = tbV[0]->v;
  r0 = tbV[1]->v;
  Ap = tbV[2]->v;
  r1 = tbV[3]->v;
  p1 =tbV[4]->v;
  x = y->v;
  b = c->v;
  
  p0dep=p0;p1dep=p1;r0dep=r0;r1dep=r1;Apdep=Ap;
  xdep=x;bdep=b;

  E_REPERR( goto END_FAST_GRAD_CONJ);
  
  /* calcul de p(0) = b - A x(0) */

  nb = T_NNorm(c);
			   
  par->func ( A, y, tbV[0]);
  
  for(i=0;i<c->dim; i++)
	  p0[i] = b[i] - p0[i];
  
  /* assignation r(0) = p(0) */
  for(i=0;i<c->dim;i++)
	  r0[i] = p0[i];
  
  /* calcul de norme de r(0) */
  ng = T_NNorm(tbV[1]);
  r0=r0dep;
  ng = (reeltype) sqrt(ng/nb);
  
  /* erreur initiale */
  Ei = ng;
  
  while(ng>par->eps){
	  
	  iterG++;
	  /* calcul de Ap = A p(0) */
	  par->func(A,tbV[0],tbV[2]);
	  
	  /* calcul de alpha */
	  sc1 = T_NNorm(tbV[1]);
	  r0=r0dep;
	  sc2 = T_Scalar(tbV[2],tbV[0]);
	  Ap=Apdep;p0=p0dep;
	  if(sc2 != 0.0)
		 alpha = sc1/sc2;
	  else 
		 alpha = 0.0;
	  
	  /* calcul de Xn+1 = Xn + alpha . p(k) */
	  for(i=0;i<c->dim;i++)
		  *x++ += alpha * (*p0++);
	  x=xdep;p0=p0dep;
	  
	  /* calcul de r(k+1) = r(k) - alpha A p(k) */
	  for(i=0;i<c->dim;i++)
		  *r1++ = *r0++ - alpha * *Ap++;
	  r1=r1dep;r0=r0dep;Ap=Apdep;
	  
	  /* nouvelle erreur */
	  ng = T_NNorm(tbV[3]);
	  r1=r1dep;
	  ng = (reeltype) sqrt(ng/nb);
	  
	  /* calcul de beta: rapport entre erreurs  */
	  sc3 = T_NNorm(tbV[3]);
	  r1=r1dep;
	  if(sc1 != 0.0)
		 beta = sc3/sc1;
	  else 
		 beta = 0.0;
	  
	  /* calcul de p(k+1) = r(k+1) + beta * p(k) */
	  for(i=0;i<c->dim;i++)
		  *p1++ = *r1++ + beta * *p0++;
	  p1=p1dep;p0=p0dep;r1=r1dep;
	  
	  /* assignation r(k) = r(k+1) */
	  for(i=0;i<c->dim;i++)
		  *r0++ = *r1++;
	  r0=r0dep;r1=r1dep;
	  
	  /* assignation p(k) = p(k+1) */
	  for(i=0;i<c->dim;i++)
		  *p0++ = *p1++;
	  
	  p0=p0dep;p1=p1dep;
	  
	  if  (par->affiche == AFFICHAGE)
		  printf("Erreur %e \n",ng);
  } 
  
  /* erreur finale */
  Ef = ng;
  
  if  (par->affiche == AFFICHAGE)
	  printf("\n G.C.-->nb d'iterations = %d Erreur initiale = %4.2e et finale = %4.2e ",
			 iterG,Ei,Ef);
  
 END_FAST_GRAD_CONJ:
  
  for(i=0;i<NB_TAB;i++)
	  
	  T_FreeNPoint( &tbV[i] );
  
  E_ISERRN("E_FastGradConjGeneral",NULL);
  
}

/*----------------------------------------------------------------------*/
/************************************************************************/
/*
  OpersNPoint

  Operations on N-Dimensional vectors.
  */
/************************************************************************/

/*----------------------------------------------------------------------*/
/* 
  Fill the vector with zero values
  */
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
void T_NClean( t_NPointPtr pt )
#else
void T_NClean(pt )
t_NPointPtr pt ;

#endif

{
  int i;
  
  for( i=0; i< pt->dim; i++ ) EPV(pt,i)= 0.0;
}

/* 
  Compute the euclidean norm of a N-dimensional vector
  */

#if defined(_ANSI_)
reeltype T_NNorm( t_NPointPtr pt )
#else
reeltype T_NNorm(pt )
t_NPointPtr pt ;

#endif

{
  int i;
  reeltype res= 0.0;
  
  for( i=0; i< pt->dim; i++ ) res += EPV(pt,i) * EPV(pt,i);
  return( (reeltype) sqrt( (double) res ) );
}

/* 
  Compute the euclidean distance between two N-points
  */

#if defined(_ANSI_)
reeltype T_NDist( t_NPointPtr p1, t_NPointPtr p2 )
#else
reeltype T_NDist(p1, p2 )
t_NPointPtr p1;
t_NPointPtr p2 ;

#endif

{
  int i;
  reeltype aux;
  reeltype res= 0.0;
  
  for( i=0; i< p1->dim; i++ ){
    aux= EPV(p1,i) - EPV(p2,i);
    res += aux * aux;
  }
  return( (reeltype) sqrt( (double) res ) );
}

/*
  Multiply a N-dimensional vector by a real value.
  */

#if defined(_ANSI_)
void T_NRMul( t_NPointPtr p1, reeltype val, t_NPointPtr p2 )
#else
void T_NRMul(p1, val, p2 )
t_NPointPtr p1;
reeltype val;
t_NPointPtr p2 ;

#endif

{
  int i;
  
  T_ReAllocNPointTb( p2, p1->dim );	/* verify size of output */
  
  for( i=0; i< p1->dim; i++ ) EPV( p2, i )= EPV( p1, i ) * val;
}

/*
  Multiply each coordinates of two N-dimensional vectors two by two
  */

#if defined(_ANSI_)
void T_NMult( t_NPointPtr p1, t_NPointPtr p2, t_NPointPtr res )
#else
void T_NMult(p1, p2, res )
t_NPointPtr p1;
t_NPointPtr p2;
t_NPointPtr res ;

#endif

{
  int i;
  
  E_ERRN( p1->dim != p2->dim, E_ERRFATAL, "T_NMult", 
	 "inclompatible vector sizes" );
  
  T_ReAllocNPointTb( res, p1->dim );   /* verify size of output */
  
  for( i=0; i< p1->dim; i++ ) EPV(res,i)= EPV(p1,i) * EPV(p2,i);
}

/*
  Transform each coordinate of this vector into its square root
  */

#if defined(_ANSI_)
void T_NSqrt( t_NPointPtr p1, t_NPointPtr res )
#else
void T_NSqrt(p1, res )
t_NPointPtr p1;
t_NPointPtr res ;

#endif

{
  int i;
  reeltype val;
  
  T_ReAllocNPointTb( res, p1->dim );   /* verify size of output */
  
  for( i=0; i< p1->dim; i++ ){
    val= EPV(p1,i);
    if ( val > 0.0 )    EPV(res,i)= (reeltype) sqrt( EPV(p1,i) );
    else		EPV(res,i)= 0.0;
  }
}

/*
  Divide each coefficient by each coefficient p1/p2 -> res
  
  (valid between 1e10 and 1e-10)
  */

#if defined(_ANSI_)
void T_NDiv( t_NPointPtr p1, t_NPointPtr p2, t_NPointPtr res )
#else
void T_NDiv(p1, p2, res )
t_NPointPtr p1;
t_NPointPtr p2;
t_NPointPtr res ;

#endif

{
  int i;
  reeltype aux;
  
  E_ERRN( p1->dim != p2->dim, E_ERRFATAL, "T_NMult", 
	 "inclompatible vector sizes" );
  
  T_ReAllocNPointTb( res, p1->dim );   /* verify size of output */
  
  for( i=0; i< p1->dim; i++ ){
    aux= EPV(p2,i);
    if ( fabs(aux) > 1e-20 )	EPV(res,i)= EPV(p1,i) / aux;
    else			EPV(res,i)= (reeltype) 1e+20;
  }
}

/*
  Add two NDimensional vectors into a third one
  */

#if defined(_ANSI_)
void T_NAdd( t_NPointPtr p1, t_NPointPtr p2, t_NPointPtr res )
#else
void T_NAdd(p1, p2, res )
t_NPointPtr p1;
t_NPointPtr p2;
t_NPointPtr res ;

#endif

{
  int i;
  
  E_ERRN( p1->dim != p2->dim, E_ERRFATAL, "T_NMult", 
	 "inclompatible vector sizes" );
  
  T_ReAllocNPointTb( res, p1->dim );   /* verify size of output */
  
  for( i=0; i< p1->dim; i++ )
    EPV(res,i)= EPV(p1,i) + EPV(p2,i);
}

/*
  substratct two NDimensional vectors into a third one (res= p1 - p2)
  */

#if defined(_ANSI_)
void T_NSub( t_NPointPtr p1, t_NPointPtr p2, t_NPointPtr res )
#else
void T_NSub(p1, p2, res )
t_NPointPtr p1;
t_NPointPtr p2;
t_NPointPtr res ;

#endif

{
  int i;
  
  E_ERRN( p1->dim != p2->dim, E_ERRFATAL, "T_NMult", 
	 "inclompatible vector sizes" );
  
  T_ReAllocNPointTb( res, p1->dim );   /* verify size of output */
  
  for( i=0; i< p1->dim; i++ )
    EPV(res,i)= EPV(p1,i) - EPV(p2,i);
}

/*
  Compute the average N-dimensional value

  It takes of a list of N-dimensional points.
  */

#if defined(_ANSI_)
void T_NAverage( t_ListPtr points, t_NPointPtr res )
#else
void T_NAverage(points, res )
t_ListPtr points;
t_NPointPtr res ;

#endif

{
  t_NPointPtr p;
  int nb= 0;
  
  if ( !points ) { T_NClean( res ); return; }
  
  T_ListItemErr( points, (t_ObjHand) &p, OBJ_NPOINT );
  E_ISERRN( "T_NAverage", (char*)NULL );
  
  T_ReAllocNPointTb( res, p->dim );
  E_ISERRN( "T_NAverage", (char*)NULL );
  
  T_NClean( res );
  
  while( points ){
    
    T_ListItemErr( points, (t_ObjHand) &p, OBJ_NPOINT );
    E_ISERRN( "T_NAverage", (char*)NULL );
    
    T_NAdd( p, res, res );
    nb += 1;
    points= points->next;
  }
  
  T_NRMul( res, 1 / (reeltype) nb, res );
}

/*
  Compute the variance vector of ND points

  This variance vector is a ND point
  */

#if defined(_ANSI_)
void T_NVariance( t_ListPtr points, t_NPointPtr res )
#else
void T_NVariance(points, res )
t_ListPtr points;
t_NPointPtr res ;

#endif

{
  t_NPointPtr aver= NULL;
  t_NPointPtr aux= NULL;
  t_NPointPtr p;
  int nb= 0;
  
  if ( !points ) { T_NClean( res ); return; }
  
  T_ListItemErr( points, (t_ObjHand) &p, OBJ_NPOINT );
  E_ISERRN( "T_NVariance", (char*)NULL );
  
  T_ReAllocNPointTb( res, p->dim );
  E_ISERRN( "T_NVariance", (char*)NULL );
  
  T_NClean( res );
  
  T_AllocNPoint( &aver );
  T_NAverage( points, aver );
  E_ISERRN( "T_NVariance", (char*)NULL );
  
  T_AllocNPoint( &aux );
  T_AllocNPointTb( aux, p->dim );
  E_ISERRN( "T_NVariance", (char*)NULL );
  
  while( points ){
    
    T_ListItemErr( points, (t_ObjHand) &p, OBJ_NPOINT );
    E_ISERRN( "T_NAverage", (char*)NULL );
    
    T_NSub( p, aver, aux );
    T_NMult( aux, aux, aux );
    T_NAdd( aux, res, res );
    
    nb += 1;
    points= points->next;
  }
  
  T_NRMul( res, 1/ (reeltype) nb, res );
  
  T_FreeNPoint( &aver );
  T_FreeNPoint( &aux );
}

/*----------------------------------------------------------------------*/












