/* epi_math2.c 2.2 5/14/98 */

#include <epidaure.h>
#include <epidaure.ee>

#ifndef NO_PROTO
static void   E_DMAddRKLtoLL( double *m, double r, int k, int l, int nL, int nC );
static void   E_DMEchLKandLL( double *m, int k, int l, int dim );
static void   E_DMEchCKandCL( double *m, int k, int l, int dim );
static void   E_DMSwapI( int *pts, int l, int r );
static void   E_DMSwapD( double *pts, int l, int r );
static void   E_DMTriPointsDbl( double *nbPts, int *perm, int left, int right );
static int    E_DMRangMat( double *m, int nL, int nC );
static double E_DMDetMat( double *m, int dim );
static int    E_DMSystLinMat( double *a, double *b, double *x, int dim );
static int    E_DMVVpropMatSym( double *m, double *val, double *vec, int dim );
static int    E_DMTriDiagonal( double *tr, double *td, int dim );
static int    E_DMVVpropMatTriDiag( double *td, double *tr, int dim );
#else 
static void   E_DMAddRKLtoLL();
static void   E_DMEchLKandLL();
static void   E_DMEchCKandCL();
static void   E_DMSwapI();
static void   E_DMSwapD();
static void   E_DMTriPointsDbl();
static int    E_DMRangMat();
static double E_DMDetMat();
static int    E_DMSystLinMat();
static int    E_DMVVpropMatSym();
static int    E_DMTriDiagonal();
static int    E_DMVVpropMatTriDiag();
#endif

/* Procedure de calcul du rang d'une matrice.

   La matrice est un objet d'EpidaureLib.

RETURN
   Retourne le rang d'une matrice, ou -1 en cas d'erreur ou de 
   calcul impossible.
*/

#if defined(_ANSI_)
int T_RangMat( t_MatPtr mat /* matrix whose rank is to be computed */ )
#else
int T_RangMat( mat )
t_MatPtr mat; /* matrix whose rank is to be computed */
#endif
{
	double *aux=NULL;
	register int i, j;
	
	if ( mat->v == NULL )
		E_XERR( E_ERRFATAL, (int)(-1), "T_RangMat", "Error in parameters" );
	
	aux = (double*)T_Alloc( (unsigned)((mat->nL)*(mat->nC)*sizeof(double)) );
	E_ISERR( (int)(-1), "T_RangMat", (char*)NULL );
	
	for (i=0; i < mat->nL; i++)
		for (j=0; j < mat->nC; j++)
			*(aux + i*mat->nC + j) = (double)EMV( mat, i, j );
	
	i = E_DMRangMat( aux, mat->nL, mat->nC );
	T_Free((void **) &aux );
	E_ISERR( (int)(-1), "T_RangMat", (char*)NULL );
	return( i );
}

/* Calcul du determinant d'une matrice carree.

   La matrice est un objet d'EpidaureLib.

RETURN
   Retourne le determinant d'une matrice, ou 0.0 en cas d'erreur ou de 
   calcul impossible.
*/

#if defined(_ANSI_)
reeltype T_DetMat( t_MatPtr mat /* matrix whose determinant is to be computed */ )
#else
reeltype T_DetMat(mat )
t_MatPtr mat; /* matrix whose determinant is to be computed */
#endif
{
	double *aux=NULL;
	register int i, j;
	reeltype d;
	
	if ( (mat->nL != mat->nC) || (mat->v == NULL) )
		E_XERR( E_ERRFATAL, (reeltype)0.0, "T_DetMat", "Error in parameters" );
	
	aux = (double*)T_Alloc( (unsigned)((mat->nL)*(mat->nL)*sizeof(double)) );
	E_ISERR( (reeltype)0.0, "T_DetMat", (char*)NULL );
	
	for (i=0; i < mat->nL; i++)
		for (j=0; j < mat->nC; j++)
			*(aux + i*mat->nC + j) = (double)EMV( mat, i, j );
	
	d = (reeltype)E_DMDetMat( aux, mat->nL );
	T_Free((void **) &aux );
	E_ISERR( (reeltype)0.0, "T_DetMat", (char*)NULL );
	return( d );
}

/* Procedure de resolution d'un systeme lineaire : A X = B.

   A doit etre une matrice carree
   La matrice et les vecteurs sont des objets d'EpidaureLib.

RETURN
   Retourne FALSE en cas d'erreur ou de resolution impossible,
   TRUE sinon.
*/

#if defined(_ANSI_)
int T_SystLinMat( t_MatPtr matA /* (square) matrix of the linear system */, 
		 t_NPointPtr vecB /* result vector of the linear system */, 
		 t_NPointPtr vecX /* unknown vector of the linear system */ )
#else
int T_SystLinMat( matA, vecB, vecX )
t_MatPtr matA;    /* (square) matrix of the linear system */
t_NPointPtr vecB; /* result vector of the linear system */
t_NPointPtr vecX; /* unknown vector of the linear system */
#endif
{
	double *auxA=NULL, *auxB=NULL, *auxX=NULL;
	register int i, j, r = FALSE;
	
	if ( (matA->nL != matA->nC) || (matA->v == NULL) )
		E_XERR( E_ERRFATAL, FALSE, "T_SystLinMat", "Error in parameters" );
	if ( (vecB->dim != matA->nC) || (vecB->v == NULL) )
		E_XERR( E_ERRFATAL, FALSE, "T_SystLinMat", "Error in parameters" );
	if ( (vecX->dim != matA->nC) || (vecX->v == NULL) )
		E_XERR( E_ERRFATAL, FALSE, "T_SystLinMat", "Error in parameters" );
	
	auxA = (double*)T_Alloc( (unsigned)((matA->nL)*(matA->nL)*sizeof(double)) );
	auxB = (double*)T_Alloc( (unsigned)((matA->nL)*sizeof(double)) );
	auxX = (double*)T_Alloc( (unsigned)((matA->nL)*sizeof(double)) );
	E_REPERR( goto END_T_SystLinMat );
	
	for (i=0; i < matA->nL; i++)
		for (j=0; j < matA->nC; j++)
			*(auxA + i*matA->nC + j) = (double)EMV( matA, i, j );
	for (i=0; i<(vecB->dim); i++)
		*(auxB + i) = (double)EPV( vecB, i );
	
	r = E_DMSystLinMat( auxA, auxB, auxX, matA->nL );
	if ( r ==TRUE )
		for (i=0; i<(vecX->dim); i++)
			EPV( vecX, i ) = (reeltype)*(auxX + i);
	
END_T_SystLinMat:
	T_Free((void **) &auxA );
	T_Free((void **) &auxB );
	T_Free((void **) &auxX );
	E_ISERR(FALSE,"T_SystLinMat",(char*)NULL);
	return( r );
}

/* Procedure d'inversion d'une matrice carree.

   Les matrices sont des objets d'EpidaureLib.

RETURN
   Retourne FALSE en cas d'erreur ou d'inversion impossible,
   TRUE sinon.
*/

#if defined(_ANSI_)
int T_InvMat( t_MatPtr mat /* (square) matrix to be inversed */, 
	     t_MatPtr Imat /* result matrix */ )
#else
int T_InvMat(mat, Imat )
t_MatPtr mat;  /* (square) matrix to be inversed */
t_MatPtr Imat; /* result matrix */
#endif
{
	double *aux=NULL, *Iaux=NULL;
	register int i, j, r=FALSE;
	
	if ( (mat->nL != mat->nC) || (mat->v == NULL) )
		E_XERR( E_ERRFATAL, FALSE, "T_InvMat", "Error in parameters" );
	if ( (Imat->nL != Imat->nC) || (Imat->nL != mat->nL) || (Imat->v == NULL) )
		E_XERR( E_ERRFATAL, FALSE, "T_InvMat", "Error in parameters" );
	
	aux = (double*)T_Alloc( (unsigned)((mat->nL)*(mat->nL)*sizeof(double)) );
	Iaux = (double*)T_Alloc( (unsigned)((mat->nL)*(mat->nL)*sizeof(double)) );
	E_REPERR( goto END_T_InvMat );
	
	for (i=0; i < mat->nL; i++)
		for (j=0; j < mat->nC; j++)
			*(aux + i*mat->nC + j) = (double)EMV( mat, i, j );
	
	r = E_DMInvMat( aux, Iaux, mat->nL ) ;
	if ( r == TRUE )
		for (i=0; i < mat->nL; i++)
			for (j=0; j < mat->nC; j++)
				EMV( Imat, i, j ) = (reeltype)*(Iaux + i*mat->nC + j);
	
END_T_InvMat:
	T_Free((void **) &aux );
	T_Free((void **) &Iaux );
	E_ISERR(FALSE,"T_InvMat",(char*)NULL);
	return( r );
}

/* Calcul des valeurs et vecteurs propres d'une matrice carree symetrique.

   Matrices et vecteurs sont des objets d'EpidaureLib.
   A cause de leur codage particulier, il est conseille d'acceder
   aux valeurs par les macros EMV et EPV.
   Apres calcul (s'il est possible), les valeurs propres sont 
   ordonnees par ordre croissant et rangees dans le vecteur
   val. 
   Leurs vecteurs propres correspondants sont ranges dans 
   le meme ordre dans la matrice vec (chaque colonne
   correspond a un vecteur propre) et ont ete normalises.

RETURN
   Retourne FALSE en cas d'erreur ou de calcul impossible.
*/

#if defined(_ANSI_)
int T_VVpropresMatSym( t_MatPtr mat /* (square) symmetric matrix whose eigenvectors and values are to be computed */, 
		      t_NPointPtr val /* computed eigenvalues (it is a vector) */, 
		      t_MatPtr vec /* computed eigenvectors (it is a matrix whose columns are the eigenvectors */ )
#else
int T_VVpropresMatSym( mat, val, vec )
t_MatPtr mat;    /* (square) symmetric matrix whose eigenvectors and values are to be computed */
t_NPointPtr val; /* computed eigenvalues (it is a vector) */
t_MatPtr vec;    /* computed eigenvectors (it is a matrix whose columns are the eigenvectors */
#endif
{
	double *auxM=NULL, *auxA=NULL, *auxE=NULL;
	register int i, j, r=FALSE;
	
  if ( (mat->nL != mat->nC) || (mat->v == NULL) )
	  E_XERR( E_ERRFATAL, FALSE, "T_VVpropresMatSym", "Error in parameters" );
	if ( (vec->nL != mat->nL) || (vec->nL != vec->nC) || (vec->v == NULL) )
		E_XERR( E_ERRFATAL, FALSE, "T_VVpropresMatSym", "Error in parameters" );
	if ( (val->dim != mat->nL) || (val->v == NULL) )
		E_XERR( E_ERRFATAL, FALSE, "T_VVpropresMatSym", "Error in parameters" );
	
	auxM = (double*)T_Alloc( (unsigned)((mat->nL)*(mat->nL)*sizeof(double)) );
	auxE = (double*)T_Alloc( (unsigned)((mat->nL)*(mat->nL)*sizeof(double)) );
	auxA = (double*)T_Alloc( (unsigned)((mat->nL)*sizeof(double)) );
	E_REPERR( goto END_T_InvMat );
	
	for (i=0; i < mat->nL; i++)
		for (j=0; j < mat->nC; j++)
			*(auxM + i*mat->nC + j) = (double)EMV( mat, i, j );
	
	r = E_DMVVpropresMatSym( auxM, auxA, auxE, mat->nL );
	if ( r == TRUE ) {
		for (i=0; i < mat->nL; i++) {
			EPV( val, i ) = (reeltype)*(auxA + i);
			for (j=0; j < mat->nC; j++)
				EMV( vec, i, j ) = (reeltype)*(auxE + i*mat->nC + j);
		}
	}
	
END_T_InvMat:
	T_Free((void **) &auxM );
	T_Free((void **) &auxE );
	T_Free((void **) &auxA );
	E_ISERR(FALSE,"T_VVpropresMatSym",(char*)NULL);
	return( r );
}

#if defined(_ANSI_)
static void E_DMAddRKLtoLL( double *m, double r, int k, int l, int nL, int nC )
#else
static void E_DMAddRKLtoLL(m, r, k, l, nL, nC )
double *m;
double r;
int k;
int l;
int nL;
int nC ;
#endif
{
	/*--------------------------------------------------
	  matrice double
	  ajoute r fois la ligne k a la ligne l
	  --------------------------------------------------*/
	register int i;
	for (i=0; i<nL; i++)
		*(m + l*nC + i ) += r * *(m + k*nC + i );
}

#if defined(_ANSI_)
static void E_DMEchLKandLL( double *m, int k, int l, int dim )
#else
static void E_DMEchLKandLL(m, k, l, dim )
double *m;
int k;
int l;
int dim ;
#endif
{
	/*--------------------------------------------------
	  matrice double
	  echange la ligne k avec la ligne l
	  --------------------------------------------------*/
	register int i;
	register double t;
	for (i=0; i<dim; i++) {
		t = *(m + i + k*dim);
		*(m + i + k*dim) = *(m + i + l*dim);
		*(m + i + l*dim) = t;
	}
}

#if defined(_ANSI_)
static void E_DMEchCKandCL( double *m, int k, int l, int dim )
#else
static void E_DMEchCKandCL(m, k, l, dim )
double *m;
int k;
int l;
int dim ;
#endif
{
	/*--------------------------------------------------
	  matrice double
	  echange la colonne k avec la colonne l
	  --------------------------------------------------*/
	register int i;
	register double t;
	for (i=0; i<dim; i++) {
		t = *(m + k + i*dim);
		*(m + k + i*dim) = *(m + l + i*dim);
		*(m + l + i*dim) = t;
	}
}

/* Multiplication d'une matrice carree par un vecteur.

   La matrice et les vecteurs sont codes en double.
   Il ne faut que le vecteur resultat soit le meme
   que le vecteur a multiplier, sinon le resultat sera
   nul (et surement faux).
*/

#if defined(_ANSI_)
void E_DMMatMulVect( double *a /* (square) matrix */, 
		   double *b /* vector to be multiplied to the matrix */, 
		   double *c /* result vector */, 
		   int dim /* dimension of the vectors and the matrix */ )
#else
void E_DMMatMulVect( a, b, c, dim )
double *a; /* (square) matrix */
double *b; /* vector to be multiplied to the matrix */
double *c; /* result vector */
int dim;   /* dimension of the vectors and the matrix */
#endif
{
	register int i,k;
	for (i=0;i<dim;i++) {
		c[i] = 0.0;
		for (k=0;k<dim;k++)
			c[i] += a[i*dim+k]*b[k];
	}
}

/* Transposition d'une matrice carree.

   Les matrices sont codees en double. La
   matrice resultat doit etre differente de
   la matrice a transposer.
*/

#if defined(_ANSI_)
void E_DMMatTrans( double *a /* (square) matrix to be transposed */, 
		  double *aT /* result matrix */, 
		  int dim /* dimension of the square matrix */ )
#else
void E_DMMatTrans(a, aT, dim )
double *a;  /* (square) matrix to be transposed */
double *aT; /* result matrix */
int dim;    /* dimension of the square matrix */
#endif
{
	register int i,j;
	for (i=0;i<dim;i++)
		for (j=0;j<dim;j++)
			aT[i*dim+j] = a[j*dim+i];
}

/* Multiplication de deux matrices carrees.

   Les matrices sont codees en doubles.
   La matrice resultat doit etre differente des
   deux matrices a multiplier.
*/

#if defined(_ANSI_)
void E_DMMatMul( double *a /* first (square) matrix to be multiplied */, 
		double *b /* second (square) matrix to be multiplied */, 
		double *ab /* result matrix = a * b */, 
		int dim /* dimension of the matrix */ )
#else
void E_DMMatMul(a, b, ab, dim )
double *a;  /* first (square) matrix to be multiplied */
double *b;  /* second (square) matrix to be multiplied */
double *ab; /* result matrix = a * b */
int dim;    /* dimension of the matrix */
#endif
{
	register int i,j,k,l;
	for (i=0;i<dim;i++)
		for (j=0;j<dim;j++) {
			l = i*dim+j;
			ab[l] = 0.0;
			for (k=0;k<dim;k++)
				ab[l] += a[i*dim+k]*b[k*dim+j];
		}
}

#if defined(_ANSI_)
static void E_DMSwapI( int *pts, int l, int r )
#else
static void E_DMSwapI( pts, l, r )
int *pts;
int l;
int r;
#endif
{
	/*--------------------------------------------------
	  procedures de tri
	  --------------------------------------------------*/
	register int t;
	t = pts[l];
	pts[l] = pts[r];
	pts[r] = t;
}

#if defined(_ANSI_)
static void E_DMSwapD( double *pts, int l, int r )
#else
static void E_DMSwapD( pts, l, r )
double *pts;
int l;
int r;
#endif
{
	/*--------------------------------------------------
	  procedures de tri
	  --------------------------------------------------*/
	register double t;
	t = pts[l];
	pts[l] = pts[r];
	pts[r] = t;
}

#if defined(_ANSI_)
static void E_DMTriPointsDbl( double *nbPts, int *perm, int left, int right )
#else
static void E_DMTriPointsDbl( nbPts, perm, left, right )
double *nbPts;
int *perm;
int left;
int right;
#endif
{
	/****************************************************************/
	/*   procedure de tri qui est une reecriture de qsort           */
	/*   avec un terme conservant la permutation realisant le tri   */
	/*   le resultat du tri est dans l'ordre croissant              */
	/****************************************************************/
	int i,last;
	
	if (left>=right)                             /* do nothing if array contains */
		return;                              /* fewer than two elements      */
	E_DMSwapD(nbPts,left,(left+right)/2);        /* move partition elem to v[0]  */
	E_DMSwapI(perm,left,(left+right)/2);
	last = left;
	for (i=left+1;i<=right;i++)                  /* partition                    */
		if ( nbPts[i] < nbPts[left] ) {
			E_DMSwapD(nbPts,++last,i);
			E_DMSwapI(perm,last,i);
		}
	E_DMSwapD(nbPts,left,last);                  /* restore partition elem       */
	E_DMSwapI(perm,left,last);
	E_DMTriPointsDbl(nbPts,perm,left,last-1);
	E_DMTriPointsDbl(nbPts,perm,last+1,right);
	return;
}

#if defined(_ANSI_)
static int E_DMRangMat( double *m, int nL, int nC )
#else
static int E_DMRangMat(m, nL, nC )
double *m;
int nL;
int nC;
#endif
{
	/*--------------------------------------------------
	  rang d'une matrice double
	  --------------------------------------------------*/
	double *n=NULL, max;
	register int i, j, k, ra, jmax;
	
	n = (double*)T_Alloc( (unsigned)(nL*nC*sizeof(double)) );
	E_ISERR( 0, "E_DMRangMat", (char*)NULL );
	
	for (i=0; i<nL*nC; i++)
		*(n + i) = *(m + i);
	ra = EPI_MIN( nL, nC );
	
	/*   on suppose la matrice de rang min( nL, nC )   */
	/***************************************************/
	for (i=0; i<nL; i++) {
		jmax = 0;
		max = 0.0;
		/* recherche du plus grand element non nul sur la ligne i */
		for (j=0; j<nC; j++)
			if ( fabs(n[i*nC+j]) > max ) {
				jmax = j;
				max = fabs(n[i*nC+j]);
			}
		if ( max < DBLZERO )
			ra--;
		else
			for (k=i+1; k<nL; k++)
				E_DMAddRKLtoLL( n, -n[k*nC+jmax]/n[i*nC+jmax], i, k, nL, nC );
	}
	T_Free((void **) &n );
	return( ra );
}

#if defined(_ANSI_)
static double E_DMDetMat( double *m, int dim )
#else
static double E_DMDetMat(m, dim )
double *m;
int dim;
#endif
{
	/*--------------------------------------------------
	  determinant d'une matrice double
	  --------------------------------------------------*/
	double *n=NULL, det, max;
	register int i, j, k, jmax;
	
	n = (double*)T_Alloc( (unsigned)(dim*dim*sizeof(double)) );
	E_ISERR( (double)0.0, "E_DMRangMat", (char*)NULL );
	
	for (i=0; i<dim*dim; i++)
		*(n + i) = *(m + i);
	det = 1.0;
	
	/*   on va triangulariser la matrice   */
	/***************************************/
	for (i=0; i<dim-1; i++) {
		/* recherche du plus grand element non nul sur la colonne i */
		max = 0.0;
		jmax = 0;
		for (j=i; j<dim; j++)
			if ( fabs(n[j*dim+i]) > max ) {
				jmax = j;
				max = fabs(n[j*dim+i]);
			}
		if (max) {
			/* il y a un terme non nul */
			if (jmax!=i) {
				E_DMEchLKandLL( n, jmax, i, dim );
				det *= -1.0;
			}
			for (k=i+1;k<dim;k++)
				E_DMAddRKLtoLL( n, -n[k*dim+i]/n[i*dim+i], i, k, dim, dim);
		}
	}
	
	/*   calcul du determinant   */
	/*****************************/
	for (i=0; i<dim; i++)
		det *= n[i*dim+i];
	T_Free((void **) &n );
	return( det );
}

#if defined(_ANSI_)
static int E_DMSystLinMat( double *a, double *b, double *x, int dim )
#else
static int E_DMSystLinMat( a, b, x, dim )
double *a;
double *b;
double *x;
int dim;
#endif
{
	/*--------------------------------------------------
	  matrice double
	  resolution d'un systeme lineaire
	  --------------------------------------------------*/
	double *n=NULL, max;
	register int i, j, k, jmax, ra;
	
	n = (double*)T_Alloc( (unsigned)(dim*dim*sizeof(double)) );
	E_ISERR( FALSE, "E_DMSystLinMat", (char*)NULL );
  
	for (i=0; i<dim*dim; i++)
		*(n + i) = *(a + i);
	for (i=0; i<dim; i++)
		*(x+i) = 0.0;
	ra = dim;
	
	for (i=0; (i<dim && ra==dim); i++) {
		jmax = 0;
		max = 0.0;
		/* recherche du plus grand element non nul sur la ligne i */
		for (j=0; j<dim; j++)
			if ( fabs(n[i*dim+j]) > max ) {
				max = fabs(n[i*dim+j]);
				jmax = j;
			}
		if ( max == 0.0) {
			/* la ligne est nulle */
			ra--;
			continue;
		}
		/* sinon, on annule la colonne du terme non nul */
		for (k=0; k<dim; k++)
			if (k!=i) {
				b[k] += b[i]*(-n[k*dim+jmax]/n[i*dim+jmax]);
				E_DMAddRKLtoLL( n, -n[k*dim+jmax]/n[i*dim+jmax], i, k, dim, dim );
			}
	}
	
	if ( ra<dim ) {
		/* le systeme n'est pas de rang dim */
		T_Free((void **) &n );
		return( FALSE );
	}
	/* sinon */
	for (i=0; i<dim; i++) {
		jmax = 0;
		max = 0.0;
		/* recherche du terme non nul sur la ligne */
		for (j=0; j<dim; j++)
			if ( fabs(n[i*dim+j]) > fabs(max) ) {
				max = n[i*dim+j];
				jmax = j;
			}
		x[jmax] = b[i]/max;
	}
	T_Free((void **) &n );
	return( TRUE );
}

/* Procedure d'inversion d'une matrice carree.

   Les matrices sont codees en double. 
   La matrice resultat doit etre differente de
   la matrice a inverser.

RETURN
   Retourne FALSE en cas d'erreur ou d'inversion impossible.
*/

#if defined(_ANSI_)
int E_DMInvMat( double *a /* (square) matrix to be inversed */, 
	       double *b /* result matrix */, 
	       int dim /* dimension of the matrix */ )
#else
int E_DMInvMat(a, b, dim )
double *a; /* (square) matrix to be inversed */
double *b; /* result matrix */
int dim;   /* dimension of the matrix */
#endif
{
	double *v=NULL, *x=NULL;
	register int i, j, r=FALSE;
	
	if ( E_DMRangMat( a, dim, dim ) < dim )
		/* la matrice n'est pas inversible */
		return( FALSE );
	
	x = (double*)T_Alloc( (unsigned)(dim*sizeof(double)) );
	v = (double*)T_Alloc( (unsigned)(dim*sizeof(double)) );
	E_REPERR( goto END_E_DMInvMat );
	
	r = TRUE;
	for (i=0; (i<dim && r); i++) {
		for (j=0; j<dim; j++)
			v[j] = 0.0;
		v[i] = 1.0;
		r = E_DMSystLinMat( a, v, x, dim );
		for (j=0; j<dim; j++)
			b[j*dim+i] = x[j];
	}
END_E_DMInvMat:
	T_Free((void **) &v );
	T_Free((void **) &x );
	E_ISERR(FALSE,"E_DMInvMat",(char*)NULL);
	return( r );
}

/* Calcul des valeurs et vecteurs propres d'une matrice carree symetrique.

   Matrices et vecteurs sont codes en double.
   Apres calcul (s'il est possible), les valeurs propres sont 
   ordonnees par ordre croissant et rangees dans le vecteur
   val. 
   Leurs vecteurs propres correspondants sont ranges dans 
   le meme ordre dans la matrice vec (chaque colonne
   correspond a un vecteur propre) et ont ete normalises.

RETURN
   Retourne FALSE en cas d'erreur ou de calcul impossible.
*/

#if defined(_ANSI_)
int E_DMVVpropresMatSym( double *m /* (square) symmetric matrix whose eigenvectors and values are to be computed */, 
			double *val /* computed eigenvalues (it is a vector) */, 
			double *vec /* computed eigenvectors (it is a matrix whose columns are the eigenvectors */, 
			int dim /* dimension of vecors and matrix */ )
#else
int E_DMVVpropresMatSym( m, val, vec, dim )
double *m;   /* (square) symmetric matrix whose eigenvectors and values are to be computed */
double *val; /* computed eigenvalues (it is a vector) */
double *vec; /* computed eigenvectors (it is a matrix whose columns are the eigenvectors */
int dim;     /* dimension of vecors and matrix */
#endif
{
	register int i, j;
	int ret, *perm=NULL, *perm2=NULL;
	double norme;
	
	/*   algorithme de calcul proprement dit   */
	/*******************************************/
	if ( !(ret = E_DMVVpropMatSym( m, val, vec, dim )) )
		return( FALSE );
	E_ISERR( FALSE, "E_DMVVpropresMatSym", (char*)NULL );
	
	/*   on ordonne les valeurs propres   */
	/**************************************/
	perm = (int*)T_Alloc( (unsigned)(dim*sizeof(int)) );
	perm2 = (int*)T_Alloc( (unsigned)(dim*sizeof(int)) );
	E_REPERR( goto END_E_DMVVpropresMatSym );
	
	for (i=0; i<dim; i++)
		perm[i] = perm2[i] = i;
	
	/*--- tri des valeurs propres ---*/
	E_DMTriPointsDbl( val, perm, 0, dim-1 );
	
	/*--- rearrangements des vecteurs propres ---*/
	for (i=0; i<dim; i++) {
		for (j=i; ( (j<dim) && (perm2[j]!=perm[i]) ); j++) ;
		if ( j!=i ) {
			E_DMSwapI(perm2,i,j);
			E_DMEchCKandCL(vec,i,j,dim);
		}
	}
	
	/*--- normalisation des vecteurs propres ---*/
	for (i=0; i<dim; i++) {
		norme = 0.0;
		for (j=0; j<dim; j++)
			norme += vec[j*dim+i]*vec[j*dim+i];
		if ( norme < DBLZERO )
			continue;
		norme = sqrt(norme);
		for (j=0; j<dim; j++)
			vec[j*dim+i] /= norme;
	}
	
END_E_DMVVpropresMatSym:
	T_Free((void **) &perm );
	T_Free((void **) &perm2 );
	E_ISERR(FALSE,"T_VVpropresMatSym",(char*)NULL);
	return( ret );
}

#if defined(_ANSI_)
static int E_DMVVpropMatSym( double *m, double *val, double *vec, int dim )
#else
static int E_DMVVpropMatSym( m, val, vec, dim )
double *m;
double *val;
double *vec;
int dim;
#endif
{
	/*--------------------------------------------------
	  matrice double
	  calcul des valeurs et vecteurs propres d'une
	  matrice symetrique
	  --------------------------------------------------*/
	register int i,j;
	double *td=NULL;
	
	/*   allocations memoires   */
	/****************************/
	td = (double*)T_Alloc( (unsigned)(2*dim*sizeof(double)) );       
	/* matrice tri-diagonale */
	E_REPERR( goto END_E_DMVVpropMatSym );
	
	/*   transformation de la matrice symetrique en tri-diagonale   */
	/*   et calcul des valeurs et vecteurs propres                  */
	/****************************************************************/
	for (i=0; i<dim; i++)
		for (j=0; j<dim; j++)
			vec[i*dim+j] = m[i*dim+j];
	E_DMTriDiagonal( vec, td, dim );
	E_DMVVpropMatTriDiag( td, vec, dim);
	E_REPERR( goto END_E_DMVVpropMatSym );
	
	for (i=0; i<dim; i++)
		val[i] = td[dim+i];
	
END_E_DMVVpropMatSym:
	T_Free((void **)&td);
	E_ISERR(FALSE,"E_DMVVpropMatSym",(char*)NULL);
	return( TRUE );
}

#if defined(_ANSI_)
static int E_DMTriDiagonal( double *tr, double *td, int dim )
#else
static int E_DMTriDiagonal(tr, td, dim )
double *tr;
double *td;
int dim;
#endif
{  
	/*   tr : matrice originale symetrique                                  */
	/*        au retour, elle contiendra la matrice de passage              */
	/*   td : matrice tridiagonale : td[0-(dim-2)] elements non diagonaux   */
	/*                               td[n-(2dim-1)] elements diagonaux      */
	/*   dim : dimension de la matrice                                      */
	/************************************************************************/
	
	register int i, j, k, l;
	double scale, hh, h, g, f;
	
	for (i=dim-1; i>=1 ; i--) {
		l = i-1;
		h = scale = 0.0;
		if (l > 0) {
			for (k=0; k<=l; k++)
				scale += fabs(tr[i*dim+k]);
			if (scale == 0.0)
				td[i] = tr[i*dim+l];
			else {
				for (k=0; k<=l; k++) {
					tr[i*dim+k] /= scale;
					h += tr[i*dim+k]*tr[i*dim+k];
				}
				f = tr[i*dim+l];
				g = f>0 ? -sqrt(h) : sqrt(h);
				td[i] = scale*g;
				h -= f*g;
				tr[i*dim+l] = f-g;
				f = 0.0;
				for (j=0; j<=l; j++) {
					/* Next statement can be omitted if eigenvectors not wanted */
					tr[j*dim+i] = tr[i*dim+j]/h;
					g = 0.0;
					for (k=0 ;k<=j; k++)
						g += tr[j*dim+k]*tr[i*dim+k];
					for (k=j+1; k<=l; k++)
						g += tr[k*dim+j]*tr[i*dim+k];
					td[j] = g/h;
					f += td[j]*tr[i*dim+j];
				}
				hh = f/(h+h);
				for (j=0; j<=l; j++) {
					f = tr[i*dim+j];
					td[j] = g = td[j]-hh*f;
					for (k=0; k<=j; k++)
						tr[j*dim+k] -= (f*td[k]+g*tr[i*dim+k]);
				}
			}
		} else
			td[i] = tr[i*dim+l];
		td[dim+i] = h;
	}
	/* Next statement can be omitted if eigenvectors not wanted */
	td[dim] = 0.0;
	td[0] = 0.0;
	/* Contents of this loop can be omitted if eigenvectors not
	   wanted except for statement d[i]=tr[i*dim+i]; */
	for (i=0;i<dim;i++) {
		l = i-1;
		if (td[dim+i]) {
			for (j=0; j<=l; j++) {
				g = 0.0;
				for (k=0; k<=l; k++)
					g += tr[i*dim+k]*tr[k*dim+j];
				for (k=0; k<=l; k++)
					tr[k*dim+j] -= g*tr[k*dim+i];
			}
		}
		td[dim+i] = tr[i*dim+i];
		tr[i*dim+i] = 1.0;
		for (j=0; j<=l; j++)
			tr[j*dim+i] = tr[i*dim+j]=0.0;
	}
	for (i=1;i<dim;i++)
		td[i-1] = td[i];
	td[dim-1] = 0.0;
	
	return( TRUE );
}

#if defined(_ANSI_)
static int E_DMVVpropMatTriDiag( double *td, double *tr, int dim )
#else
static int E_DMVVpropMatTriDiag(td, tr, dim )
double *td;
double *tr;
int dim;
#endif
{
	register int i, l, k, iter, m;
	double b, c, dd, f, g, p, r, s;
	
	for (l=0; l<dim; l++) {
		iter = 0;
		do {
			for (m=l; m<dim-1; m++) {
				dd = fabs( td[dim+m] ) + fabs( td[dim+m+1] );
				if ((double)(fabs(td[m]) + dd) == dd)
					break;
			}
			if (m != l) {
				if (iter++ == 30)
					E_ISERR(FALSE,"E_DMVVpropMatTriDiag","Too many iterations");
				g = (td[dim+l+1] - td[dim+l]) / (2.0*td[l]);
				r = sqrt((g*g)+1.0);
				g = td[dim+m] - td[dim+l] + td[l]/(g+EPI_MSIGN(r,g));
				s = c =1.0;
				p = 0.0;
				for (i=m-1; i>=l; i--) {
					f = s*td[i];
					b = c*td[i];
					if (fabs(f) >= fabs(g)) {
						c = g/f;
						r = sqrt((c*c)+1.0);
						td[i+1] = f*r;
						c *= (s=1.0/r);
					}
					else {
						s = f/g;
						r = sqrt((s*s)+1.0);
						td[i+1] = g*r;
						s *= (c=1.0/r);
					}
					g = td[dim+i+1] - p;
					r = (td[dim+i]-g)*s + 2.0*c*b;
					p = s*r;
					td[dim+i+1] =g+p;
					g = c*r-b;
					/* Next loop can be omitted if eigenvectors not wanted */
					for (k=0; k<dim; k++) {
						f = tr[k*dim+i+1];
						tr[k*dim+i+1] = s*tr[k*dim+i]+c*f;
						tr[k*dim+i] = c*tr[k*dim+i]-s*f;
					}
				}
				td[dim+l] = td[dim+l]-p;
				td[l] = g;
				td[m] = 0.0;
			}
		} while (m != l);
	}
	return( TRUE );
}
