/* epi_obj.c 2.4 5/14/98 */
/*==========================================================================
 *
 * File : $EPI/src/epi_obj.c
 * Module :
 * $ Id : ? $ 
 * Test file : ?
 * Author : J-P Thirion
 * Creation: August 4 1992
 *===========================================================================
 *	(c) Copyright 1992, Inria, all right reserved
 *===========================================================================
 *
 *------------------------------Description----------------------------------
 *
 * Fonctions ge'ne'rales de de'finition et de gestion des objets
 *
 *-----------------------------Modification----------------------------------
 *
 */

/*======================================================================*/
/*
	epi_obj.c
		
	General operations for EpidaureLib objects

	(c) Jean-Philippe THIRION -- INRIA
*/
/*======================================================================*/

#if defined(INC_MIN) /* inclusions minimales */
 /* rajouter -DINC_MIN dans les options de compilation */
 /* GM 12/09/94                                        */
#include <epi_kernel.ei>
#include <epi_var.ee>
/* #include <testecle.ei> */
#else /* inclusions maximales */
#include <epidaure.h>
#include <epidaure.ee>
#include <epidaure.ei>
#endif

#define OBJ_USERUTIL 0x4FFFFFFF

typedef struct t_ObjDefUtil{
  int init;
  int sizeUtil;
  void_func_type f[ OBJF_NB ];
} t_ObjDefUtil, *t_ObjDefUtilPtr;

typedef struct t_MyOwnType{		/* my own object type definition		*/
  t_Obj obj;
  int index;
} t_MyOwnType, *t_MyOwnTypePtr;

/*------ Local vars ------*/
static t_ArrayPtr gUserObj= NULL;


/*--------------------- Prototypes of static functions -----------------*/
/*-------Definition des fonctions statiques----------*/
#ifndef NO_PROTO
static void T_ObjInit( void );
static void T_ObjInitAll( void );
static t_ObjDefUtilPtr _T_GetObjDef( int type );
static void_func_type _T_GetFunc( int objType, int opType );
static int _T_GetSizeUtil( int objType );
static int T_GetNextChar( FILE *fd, char *name );
static void T_RecSetNum3Map( t_3MapPtr theMap, int state, int num );
#else 
static void T_ObjInit();
static void T_ObjInitAll();
static t_ObjDefUtilPtr _T_GetObjDef();
static void_func_type _T_GetFunc();
static int _T_GetSizeUtil();
static int T_GetNextChar();
static void T_RecSetNum3Map();
#endif


/************************************************************************/
/*
	EpiInits
		
	Initialisation operations in EpidaureLib

	(c) Jean-Philippe THIRION -- INRIA --
*/
/************************************************************************/

/*
  Start all initialisations in EpidaureLib

	This function has to be called once, at the very beginning of
	your application.	
*/

#if defined(_ANSI_)
void T_InitStandard( int argc, char **argv )
#else
void T_InitStandard(argc, argv )
int argc;
char **argv ;

#endif
{
  /* Tests the key file */
  /* TestKey(argc, argv); */
  if (argc >= 1)
    strncpy( program, argv[0], 80 );
  else
    strncpy( program, "Undefined_program_name", 80 );
  verbose= TRUE;
  E_FatalInit();	      /* initialize error message array		*/
  
#ifdef INRIMAGE
  initopts( argc, argv );     /* initialize inrimage                  */
#endif
  
  T_ObjInitAll();             /* initialize all the object functions  */
}

/*
  Perform the desallocation of global structures

	This function has to be called once, at the very end of
	your application.	
*/

#if defined(_ANSI_)
void T_FreeStandard( void )
#else
void T_FreeStandard()

#endif
{
  T_FreeObj( (t_ObjHand) &gUserObj );
}

/*----------------------------------------------------------------------*/
/************************************************************************/
/*
	OpersObj

	Generic operations with objects

		(c) Jean-Philippe THIRION -- INRIA
*/
/************************************************************************/

void_func_type gObjTb[OBJ_NB][OBJF_NB];	/* could be static */
static int gObjStart= FALSE;
static int gObjSizeUtil[OBJ_NB];

/*		Void function. This function do nothing							*/

#if defined(_ANSI_)
int fNULL( void )
#else
int fNULL()

#endif

{ 
  return(TRUE); 
}

/*	Initialize the object table.										*/

#if defined(_ANSI_)
static void T_ObjInit( void )
#else
static void T_ObjInit()

#endif

{
  int obj;
  int fc;
  
  if ( gObjStart == TRUE ) return;
  
  for( obj=0; obj< OBJ_NB; obj++ )
    for( fc=0; fc< OBJF_NB; fc++ )
      gObjTb[ obj ][ fc ]= NULL;
  
  gUserObj= NULL;
}

/*	Initialize all the functions 										*/

#if defined(_ANSI_)
static void T_ObjInitAll( void )
#else
static void T_ObjInitAll()

#endif

{
  int size;
  
  T_ObjInit();
  
  T_ObjListInit();
  T_Obj3PDualInit();
  T_ObjCrestInit();
  T_ObjIntInit();
  T_ObjPointInit();
  T_ObjArrayInit();
  T_ObjMat4Init();
  T_ObjObjDefInit();
  T_ObjFloatInit();
  T_ObjStringInit();
  T_ObjMatInit();
  T_ObjNPointInit();
  
  size= sizeof(t_Obj);
  
  gObjSizeUtil[OBJ_UNKNOWN]= 0;
  gObjSizeUtil[OBJ_LIST]		= sizeof(t_List) - size;
  gObjSizeUtil[OBJ_3MAP]		= sizeof(t_3Map) - size;
  gObjSizeUtil[OBJ_3PDUAL]	= sizeof(t_3PDual) - size;
  gObjSizeUtil[OBJ_CREST]		= sizeof(t_Crest) - size;
  gObjSizeUtil[OBJ_INT]		= sizeof(t_Int) - size;
  gObjSizeUtil[OBJ_POINT]		= sizeof(t_Point) - size;
  gObjSizeUtil[OBJ_ARRAY]		= sizeof(t_Array) - size;
  gObjSizeUtil[OBJ_MAT4]		= sizeof(t_Mat4) - size;
  gObjSizeUtil[OBJ_OBJDEF]	= sizeof(t_ObjDef) - size;
  gObjSizeUtil[OBJ_FLOAT]		= sizeof(t_Float) - size;
  gObjSizeUtil[OBJ_STRING]	= sizeof(t_String) - size;
  gObjSizeUtil[OBJ_MAT]		= sizeof(t_Mat) - size;
  gObjSizeUtil[OBJ_NPOINT]	= sizeof(t_NPoint) - size;
}

/*	
  Verify that the type of the object is known and initialized
  */

#if defined(_ANSI_)
int T_ObjType( int type )
#else
int T_ObjType(type )
int type ;

#endif
{
  t_ObjDefUtilPtr objDef;
  
  if ( type & OBJ_USER ){
    
    objDef= _T_GetObjDef( type );
    if ( objDef == NULL )	return( FALSE );
    else 					return( TRUE );
  }
  
  switch( type ){
    
  case OBJ_UNKNOWN:
  case OBJ_LIST:
  case OBJ_3MAP:
  case OBJ_3PDUAL:
  case OBJ_CREST:
  case OBJ_INT:
  case OBJ_POINT:
  case OBJ_ARRAY:
  case OBJ_MAT4:
  case OBJ_OBJDEF:
  case OBJ_FLOAT:
  case OBJ_STRING:
  case OBJ_MAT:
  case OBJ_NPOINT:
    return(TRUE); break;
    
  default: return(FALSE); break;
  }
  
  return( FALSE );
}

/*
  Initialize an object with default values
  */

#if defined(_ANSI_)
int T_DefaultObj( int type, t_ObjPtr obj )
#else
int T_DefaultObj(type, obj )
int type;
t_ObjPtr obj ;

#endif

{
  t_ObjPtr_func_type func;
  
  E_ERR(  !T_ObjType(type), E_ERRUNKNOWN, FALSE, "T_DefaultObj", "epidaure object initialization");
  
  func = ( t_ObjPtr_func_type ) _T_GetFunc( type, OBJF_DEFAULT );
  if ( func ) func( obj );
/*  else obj->type= (signed long ) type; */
  else obj->type= (long) type;   /* Gregoire 05 avril 1994 */
  
  E_ISERR( FALSE, "T_DefaultObj", NULL );
  return( TRUE );
}

/*		Allocate an object of a given type								*/

#if defined(_ANSI_)
int T_AllocObj( int type, t_ObjHand objHand )
#else
int T_AllocObj(type, objHand )
int type;
t_ObjHand objHand ;

#endif

{
  t_ObjHand_func_type func;
  t_ObjPtr_func_type func2;
  unsigned size;
  
  E_ERR(	!T_ObjType(type), E_ERRUNKNOWN, FALSE, "T_AllocObj", 
	"epidaure object allocation");
  
  func= ( t_ObjHand_func_type ) _T_GetFunc( type, OBJF_ALLOC );
  
  if ( func == NULL ){
    size= _T_GetSizeUtil( type ) + sizeof( t_Obj );
    *objHand= (t_ObjPtr) T_Alloc( size );
    (*objHand)->type= (long)type;
  }
  else func( objHand );
  
  /* perform also default initialization */
  
  func2= ( t_ObjPtr_func_type ) _T_GetFunc( type, OBJF_DEFAULT );
  if ( func2 ) func2( *objHand );
  
  E_ISERR( FALSE, "T_AllocObj", "epidaure object allocation" );
  return( TRUE );
}

/*
  Free an object of a given type
  */

#if defined(_ANSI_)
int T_FreeObj( t_ObjHand objHand )
#else
int T_FreeObj(objHand )
t_ObjHand objHand ;

#endif

{
  t_ObjHand_func_type func;
  int type;
  
  if (*objHand == NULL ) return(TRUE);
  
  type= (*objHand)->type;
 
	if ( !T_ObjType(type) ){
		sprintf( tmpStr1, "epidaure object free (%d)", type ); 
  		E_XERR( E_ERRUNKNOWN, FALSE, "T_FreeObj", tmpStr1 );
		}
  
  func= ( t_ObjHand_func_type ) _T_GetFunc( type, OBJF_FREE );
  
  if ( func == NULL ) T_Free( (void **) objHand );
  else func( objHand );
  
  /*	E_ISERR( FALSE, "T_FreeObj", "epidaure object desallocation" ); */
  
  return( TRUE );
}

/*
  Copy an object of a given type
  */

#if defined(_ANSI_)
int T_CopyObj( t_ObjPtr theObj, t_ObjHand objHand )
#else
int T_CopyObj(theObj, objHand )
t_ObjPtr theObj;
t_ObjHand objHand ;

#endif

{
  t_PtrHand_func_type func;
  int size;
  
  if ( theObj == NULL ){ *objHand= NULL; return(FALSE); }
  
  if ( !T_ObjType( (int) theObj->type) ){
    
    sprintf( tmpStr1, "copy object %ld", theObj->type );
    E_XERR( E_ERRUNKNOWN, FALSE, "T_CopyObj", tmpStr1 );
  }
  
  func= ( t_PtrHand_func_type ) _T_GetFunc( (int) theObj->type, OBJF_COPY );
  
  if ( func == NULL ){
    if ( *objHand == NULL ) T_AllocObj( (int) theObj->type, objHand );
    size= _T_GetSizeUtil( (int) theObj->type ) + sizeof( t_Obj ); 
    memcpy( (void *) *objHand, (void *) theObj, (size_t) size );
  }
  
  else func( theObj, objHand );
  
  E_ISERR( FALSE, "T_CopyObj", "epidaure object copy" );
  return( TRUE );
}

/*
  Write an object of a given type
  */

#if defined(_ANSI_)
int T_FWriteObj( FILE *fd, char *name, t_ObjPtr objPtr )
#else
int T_FWriteObj(fd, name, objPtr )
FILE *fd;
char *name;
t_ObjPtr objPtr ;

#endif

{
  t_fdcharObjP_func_type func;
  
  E_ERR(	!T_ObjType( (int) objPtr->type), E_ERRUNKNOWN, FALSE, "T_FWriteObj", 
	"epidaure object write");
  
  func= ( t_fdcharObjP_func_type ) _T_GetFunc( (int) objPtr->type, OBJF_FWRITE );
  E_ERR( func == NULL, E_ERRNULL, FALSE, "T_FWriteObj", "no write function");
  
  return( func( fd, name, objPtr ) );
}

/*
  Read an object of a given type
  */

#if defined(_ANSI_)
int T_FReadObj( int type, FILE *fd, char *name, t_ObjHand objHand )
#else
int T_FReadObj(type, fd, name, objHand )
int type;
FILE *fd;
char *name;
t_ObjHand objHand ;

#endif

{
  t_fdcharObjH_func_type func;
  
  E_ERR(	!T_ObjType(type), E_ERRUNKNOWN, FALSE, "T_FReadObj", 
	"epidaure object read");
  
  if ( *objHand == NULL ) T_AllocObj( type, objHand );
  
  func= ( t_fdcharObjH_func_type ) _T_GetFunc( type, OBJF_FREAD );
  E_ERR( func == NULL, E_ERRNULL, FALSE, "T_FReadObj", "no read function");
  
  return( func( fd, name, objHand ) );
}

/*
  Write a single basic object
  */

#if defined(_ANSI_)
int T_FWriteOne( FILE *fd, char *name, t_ObjPtr theObj )
#else
int T_FWriteOne(fd, name, theObj )
FILE *fd;
char *name;
t_ObjPtr theObj ;

#endif

{
  int create;
  int status;
  
  if ( fd == NULL ){
    create= TRUE;
    fd= fopen( name, "w" );
    E_ERR( fd==NULL, E_ERRC, FALSE, "T_FWriteOne", name );
  }
  else create= FALSE;
  
  status= T_WriteObjectType( fd, name, (int) theObj->type );
  E_ERR( status == EOF, E_ERRW, FALSE, "T_FWriteOne", name );
  
  status= T_FWriteObj( fd, name, theObj );
  
  switch( (int) theObj->type ){
  case OBJ_FLOAT:
  case OBJ_INT:
    fprintf( fd, "\n" );
    break;
  }
  
  if ( create ) fclose( fd );
  E_ISERR( FALSE, "T_FWriteOne", NULL );
  return( status );
}

/*
  Read a single basic object
  */

#if defined(_ANSI_)
int T_FReadOne( FILE *fd, char *name, t_ObjHand handObj )
#else
int T_FReadOne(fd, name, handObj )
FILE *fd;
char *name;
t_ObjHand handObj ;

#endif

{
  int opening;
  int status;
  int type;
  
  if ( fd == NULL ){
    opening= TRUE;
    fd= fopen( name, "r" );
    E_ERR( fd==NULL, E_ERRO, FALSE, "T_FReadOne", name );
  }
  else opening= FALSE;
  
  type= T_ReadObjectType( fd, name );
  E_ISERR( FALSE, "T_FReadOne", NULL );
  
  status= T_FReadObj( type, fd, name, handObj );
  
  if ( opening ) fclose( fd );
  E_ISERR( FALSE, "T_FReadOne", NULL );
  return( status );
}

/*
  Return the definition of a user object definition.

Returns: NULL if out of bounds
*/

#if defined(_ANSI_)
static t_ObjDefUtilPtr _T_GetObjDef( int type )
#else
static t_ObjDefUtilPtr _T_GetObjDef(type )
int type ;

#endif

{
  t_ObjDefUtilPtr objDef;
  
  type= type & OBJ_USERUTIL;
  
  if ( (type < 0) || (type >= gUserObj->nbAlloc) ) return( NULL );
  
  objDef= (t_ObjDefUtilPtr) &gUserObj->tb[type*sizeof(t_ObjDefUtil)];
  
  if ( objDef->init ) return( objDef );
  else				return( NULL );
}

/*
  Returns the function corresponding to the object type
 */

#if defined(_ANSI_)
static void_func_type _T_GetFunc( 
int objType, 	/* object type */
int opType		/* function type */
)
#else
static void_func_type _T_GetFunc(objType, opType )
int objType;
int opType ;

#endif

{
  t_ObjDefUtilPtr objDef;
  
  if ( objType & OBJ_USER ){
    
    objDef= _T_GetObjDef( objType );
    E_ERR( !objDef, E_ERRFATAL, NULL, "T_GetFunc",
	  "invalid user type number");
    return( objDef->f[opType] );
  }
  else{
    E_ERR( (objType >= OBJ_NB) ||
	  (opType >= OBJF_NB), E_ERRARRAY, NULL, "T_GetFunc",
	  "invalid object or function number");
    return( gObjTb[ objType ][ opType ] );
  }
}

/*
  Gives length in bytes of the "util" part of the object definition. 

	You must add sizeof(t_Obj) to have the entire length of the structure.
  This function is dirty because it don't use the array operation. But
  as it is basic operation, it is optimized to run fast.
  */

#if defined(_ANSI_)
static int _T_GetSizeUtil( int objType )
#else
static int _T_GetSizeUtil(objType )
int objType ;

#endif

{
  t_ObjDefUtilPtr objDef;
  
  if ( objType & OBJ_USER ){
    
    objDef= _T_GetObjDef( objType );
    E_ERR( !objDef, E_ERRFATAL, 0, "_T_GetSizeUtil",
	  "invalid user object type");
    return( objDef->sizeUtil );
  }
  else{
    E_ERR( objType >= OBJ_NB, E_ERRARRAY, 0, "_T_GetSizeUtil",
	  "invalid object type");
    return( gObjSizeUtil[objType] );
  }
}

/*
  Dynamic addition of a new object type.

	This function also adds the basic functions to its definition.
  Return the new object type number. If it can't allocate a new object
  definition, then it returns OBJ_UNKNOWN.
  */

#if defined(_ANSI_)
int T_AddObjDef( t_ObjDefPtr objDef )
#else
int T_AddObjDef(objDef )
t_ObjDefPtr objDef ;

#endif

{
  int pos,num,nbAlloc;
  t_ObjDefUtilPtr freeDef;
  
  if ( gUserObj == NULL ){
    T_AllocArray( &gUserObj );
    T_AllocArrayTb( gUserObj, OBJ_OBJDEF, 16 );
    
    freeDef= (t_ObjDefUtilPtr) gUserObj->tb;
    for( num=0; num< gUserObj->nbAlloc; num++ ){
      freeDef->init= FALSE;
      freeDef += 1;
    }
  }
  
  pos= T_ArrayAdd( gUserObj, (t_ObjPtr) objDef );
  
  if ( pos < 0 ){			/* try to find a free object definition		*/
    
    num=0;
    freeDef= (t_ObjDefUtilPtr) gUserObj->tb;
    
    while( (num < gUserObj->nbAlloc) && (freeDef->init) ){
      num += 1;
      freeDef += 1;
    }
    if ( num == gUserObj->nbAlloc ) pos= -1;
    else{
      pos= num;
      T_ArraySet( gUserObj, (t_ObjPtr) objDef, num );
    }
  }
  
  if ( pos < 0 ){			/* double the size of object def		*/
    
    nbAlloc= gUserObj->nbAlloc;
    T_ReAllocArrayTb( gUserObj, nbAlloc * 2 );
    
    freeDef= (t_ObjDefUtilPtr) &gUserObj->tb;
    freeDef += nbAlloc;
    
    for( num=0; num< nbAlloc; num++ ){
      freeDef->init= FALSE;
      freeDef += 1;
    }
    
    pos= T_ArrayAdd( gUserObj, (t_ObjPtr) objDef );
  }
  
  E_ISERR( OBJ_UNKNOWN, "T_AddObjDef", NULL );
  if ( pos == -1 ) return( OBJ_UNKNOWN );
  return( pos | OBJ_USER ); 
}

/*
  Remove the object definition from the array. 

Returns: TRUE if removed, FALSE if an error occured.
*/

#if defined(_ANSI_)
int T_RemoveObjDef( int *pType )
#else
int T_RemoveObjDef(pType )
int *pType ;

#endif

{
  t_ObjDefUtilPtr objDef;
  
  if( !(*pType & OBJ_USER) ) return( FALSE );
  
  objDef= _T_GetObjDef( *pType );
  if ( !objDef ) return( FALSE );
  
  objDef->init= FALSE;
  *pType= OBJ_UNKNOWN;
  return( TRUE );
}

/*----------------------------------------------------------------------*/
/************************************************************************/
/*
	OpersList

		General operations on lists of objects

	(c) Jean-Philippe THIRION -- INRIA -- 8/7/91
*/
/************************************************************************/

/*		Perform the initialization of all the functions on lists		*/

#if defined(_ANSI_)
void T_ObjListInit( void )
#else
void T_ObjListInit()

#endif

{
  gObjTb[ OBJ_LIST ][ OBJF_ALLOC ]= ( void_func_type ) T_AllocList;
  gObjTb[ OBJ_LIST ][ OBJF_FREE ]= ( void_func_type ) T_FreeList;
  gObjTb[ OBJ_LIST ][ OBJF_FREAD ]= ( void_func_type ) T_FReadList;
  gObjTb[ OBJ_LIST ][ OBJF_FWRITE ]= ( void_func_type ) T_FWriteList;
  gObjTb[ OBJ_LIST ][ OBJF_COPY ]= ( void_func_type ) T_CopyList;
  gObjTb[ OBJ_LIST ][ OBJF_DEFAULT ]= ( void_func_type ) T_DefaultList;
}

/*
  Perform the initialization of list structure
*/

#if defined(_ANSI_)
void T_DefaultList( t_ListPtr theList )
#else
void T_DefaultList(theList )
t_ListPtr theList ;

#endif

{
  theList->obj.type= OBJ_LIST;
  theList->next= NULL;
  theList->back= NULL;
  theList->item= NULL;
}

/*		Allocate a new element list 									*/

#if defined(_ANSI_)
int T_AllocList( t_ListHand handList )
#else
int T_AllocList(handList )
t_ListHand handList ;

#endif

{
  *handList= (t_ListPtr) T_Alloc(sizeof(t_List));
  E_ISERR( FALSE, "T_AllocList", "list of epidaure objects");
  
  T_DefaultList( *handList );
  return(TRUE);
}

/*
	Allocate a new element of a list.

	It allocates also the corresponding item. Return a pointer toward both
 the element and the item. Are you sure that you want to use this procedure ?
... just have a look on the procedure T_AddObjList!
*/

#if defined(_ANSI_)
int T_AllocListItem( t_ListHand elemHand, t_ObjHand objHand, int type )
#else
int T_AllocListItem(elemHand, objHand, type )
t_ListHand elemHand;
t_ObjHand objHand;
int type ;

#endif

{
  t_ListPtr theElem;
  
  T_AllocList( &theElem );
  E_ISERR( FALSE, "T_AllocListItem", "prefix of a list");
  
  T_AllocObj( type, &(theElem->item) );
  E_ISERR( FALSE, "T_AllocListItem", NULL );
  
  *elemHand= theElem;
  *objHand= theElem->item;
  return( TRUE );
}

/*		Free a list and all the elements in this list					*/

#if defined(_ANSI_)
int T_FreeList( t_ListHand handList )
#else
int T_FreeList(handList )
t_ListHand handList ;

#endif

{
  t_ListPtr theList,start,nextList;
  
  if ( *handList == NULL ) return(TRUE);
  theList= *handList;
  
  start= theList;
  
  do{
    nextList= theList->next;
    if (theList->item != NULL){
      T_FreeObj( &theList->item );
      E_ISERR( FALSE, "T_FreeList", "list of epidaure objects"); 
    }
    free( (void *) theList );
    theList= nextList;
    
  } while( (theList != NULL) && (theList != start) );
  
  *handList= NULL;
  return(TRUE);
}

/*
  Delete a list, but not the elements which are attached to it.
  */

#if defined(_ANSI_)
int T_DeleteList( t_ListHand handList )
#else
int T_DeleteList(handList )
t_ListHand handList ;

#endif

{
  t_ListPtr theList,start,nextList;
  
  if ( *handList == NULL ) return(TRUE);
  theList= *handList;
  
  start= theList;
  
  do{
    nextList= theList->next;
    free( (void *) theList );
    theList= nextList;
    
  } while( (theList != NULL) && (theList != start) );
  
  *handList= NULL;
  return(TRUE);
}

/*		Reverse a list.													*/

#if defined(_ANSI_)
void T_ReverseList( t_ListHand handList )
#else
void T_ReverseList(handList )
t_ListHand handList ;

#endif

{
  t_ListPtr theList,start,next;
  t_ListPtr newList= NULL;
  
  theList= *handList;
  if ( theList == NULL ) return;
  
  start= theList;
  
  do{
    next= theList->next;
    
    T_ListLink( theList, newList );
    newList= theList;
    theList= next;
  } while( (theList != NULL) && (theList != start) );
  
  if ( theList != NULL )	T_ListLink( theList, newList );
  else					T_ListLink( (t_ListPtr) NULL, newList );
  *handList= newList;
}

/*		Indicate if a list is closed or not								*/

#if defined(_ANSI_)
int T_IsListClosed( t_ListPtr theList )
#else
int T_IsListClosed(theList )
t_ListPtr theList ;

#endif

{
  t_ListPtr start;
  
  start= theList;
  if ( theList == NULL ) return( FALSE );
  
  do{
    theList= theList->next;
  } while( (theList != NULL) && (theList != start) );
  
  if ( theList == NULL )	return( FALSE );
  else					return( TRUE );
}

/*		Get the last element of a list									*/

#if defined(_ANSI_)
void T_ListLast( t_ListPtr theList, t_ListHand handLast )
#else
void T_ListLast(theList, handLast )
t_ListPtr theList;
t_ListHand handLast ;

#endif

{
  t_ListPtr start,prev;
  
  if ( theList == NULL ){
    *handLast= NULL; 
    return;
  }
  
  if ( theList->back != NULL ){
    *handLast= theList->back;
    return;
  }
  
  start= theList;
  
  do{
    prev= theList;
    theList= theList->next;
  } while( (theList != NULL) && (theList != start) );
  
  *handLast= prev;
}

/*		Close a linked list of points									*/

#if defined(_ANSI_)
void T_ListClose( t_ListPtr theList )
#else
void T_ListClose(theList )
t_ListPtr theList ;

#endif

{
  t_ListPtr last;
  
  T_ListLast( theList, &last );
  T_ListLink( last, theList );
}

/*		Open a linked list of points									*/

#if defined(_ANSI_)
void T_ListOpen( t_ListPtr theList )
#else
void T_ListOpen(theList )
t_ListPtr theList ;

#endif

{
  if ( theList ) T_LinkCut( theList->back, theList );
}

/*		Return the item of the list, but verify that the type match		*/

#if defined(_ANSI_)
void T_ListItemErr( t_ListPtr theList, t_ObjHand handObj, int type )
#else
void T_ListItemErr(theList, handObj, type )
t_ListPtr theList;
t_ObjHand handObj;
int type ;

#endif

{
  if ( type == OBJ_UNKNOWN )
    { *handObj= theList->item; return; }
  
  E_ERRN( theList->item == NULL, E_ERRNULL,"T_ListItemErr","void list node");
  E_ERRN( type != theList->item->type, E_ERRMISMATCH, "T_ListItemErr", 
	 "not the expected object in list" );
  
  *handObj= theList->item;
}

/*		Write the content of a list in an ascii file on disk			*/

#if defined(_ANSI_)
int T_FWriteList( FILE *fd, char *name, t_ListPtr list )
#else
int T_FWriteList(fd, name, list )
FILE *fd;
char *name;
t_ListPtr list ;

#endif

{
  int create;
  int type,status;
  t_ListPtr start;
  
  if ( fd == NULL ) {
    create= TRUE;
    fd= fopen( name, "w" );
    E_ERR( fd == NULL, E_ERRC, FALSE, "T_FWriteList", name );
  }
  else create= FALSE;
  status= T_WriteObjectType( fd, name, OBJ_LIST );
  if ( status == EOF ) goto END_T_FWriteList;
  
  if ( list == NULL ){
    status= T_WriteObjectType( fd, name, OBJ_ENDLIST );
    if ( create ) fclose( fd );
    return( TRUE );
  }
  
  start= list;
  
  do{
    if ( list->item != NULL ){
      type= list->item->type;
      if (type != OBJ_LIST){
	status= T_WriteObjectType( fd, name, type );
	if( status == EOF ) goto END_T_FWriteList;
      }
      status= T_FWriteObj( fd, name, list->item );
      if( status == FALSE ) goto END_T_FWriteList;
    }
    list= list->next;
  } while( (list != NULL) && (list != start) );
  
  if ( list != NULL ){			/* closed loop	*/
    status= T_WriteObjectType( fd, name, OBJ_ENDLOOP );
    if( status == EOF ) goto END_T_FWriteList;
  }
  else{
    status= T_WriteObjectType( fd, name, OBJ_ENDLIST );
    if( status == EOF ) goto END_T_FWriteList;
  }
  
  if ( create ) fclose( fd );
  return( TRUE );
  
 END_T_FWriteList:
  if ( create ) fclose( fd );
  E_XERR( E_ERRW, FALSE, "T_FWriteList", name );
  return( FALSE );
}

/*		Read the content of a list in an ascii file on disk				*/

#if defined(_ANSI_)
int T_FReadList( FILE *fd, char *name, t_ListHand listHand )
#else
int T_FReadList(fd, name, listHand )
FILE *fd;
char *name;
t_ListHand listHand ;

#endif

{
  int closed= FALSE;
  int newOpen,stillRead;
  int type,status;
  t_ListPtr theList= NULL;
  t_ObjPtr newItem= NULL;
  
  if ( fd == NULL ) {
    newOpen= TRUE;
    fd= fopen( name, "r" );
    if ( fd == NULL ){
      E_XERR( E_ERRR, FALSE, "T_FReadList", name);
      return(FALSE);
    }
    
    type= T_ReadObjectType( fd, name );
    if ( type == OBJ_EOF ) T_Warn("reading void object file %s\n", name);
    else if (type != OBJ_LIST){
      E_XERR( E_ERRMISMATCH, FALSE, "not a list in file ", name );
      return(FALSE);
    }
  }
  else newOpen= FALSE;
  
  stillRead= TRUE;
  
  do{
    type= T_ReadObjectType( fd, name );
    switch( type ){
    case OBJ_EOF: stillRead= FALSE; closed= FALSE; break;
    case OBJ_ENDLIST: stillRead= FALSE; closed= FALSE; break;
    case OBJ_ENDLOOP: stillRead= FALSE; closed= TRUE; break;
    case OBJ_LIST:
      newItem = NULL;  /* Differencie l'objet liste des autres, sinon il alloue
			* un objet qui ne sera plus trace dans la liste et donc
			* cet objet ne sera pas libere par T_FreeList */
      status= T_FReadList( fd, name, (t_ListHand)& newItem );
      if ( status == 0 )
	{
	  T_Warn("error reading in file %s\n", name );
	  return( FALSE );
	}
      T_AddObjList( newItem, & theList );
      break;
    default:
      T_AllocObj( type, &newItem );
      status= T_FReadObj( type, fd, name, &newItem );
      if ( status == 0 ){
	T_Warn("error reading in file %s\n", name );
	return( FALSE );
      }
      T_AddObjList( newItem, &theList );
      break;
    }
    
  } while( stillRead );
  
  if (closed) T_ListClose( theList );
  T_ReverseList( &theList );
  *listHand= theList;	
  if ( newOpen ) fclose( fd );
  return( TRUE );
}

/*		Return the value of the next token								*/

#if defined(_ANSI_)
int T_ReadObjectType( FILE *fd, char *name )
#else
int T_ReadObjectType(fd, name )
FILE *fd;
char *name ;

#endif

{	
  int theChar;
  int type,status;
  
  theChar= T_GetNextChar( fd, name );
  
  E_ERR( theChar==OBJ_EOF, E_ERRR, OBJ_EOF, "T_ReadObjType", name );
  
  switch( theChar ){
  case EOF: return( OBJ_EOF ); break;
  case '(': return( OBJ_LIST ); break;
  case ')': return( OBJ_ENDLIST ); break;
  case '}': return( OBJ_ENDLOOP ); break;
  case '[': return( OBJ_BEGINARRAY ); break;
  case ']': return( OBJ_ENDARRAY ); break;
  case 'O': 
    status= fscanf( fd, "%d", &type );
    E_ERR( status==EOF, E_ERRR, OBJ_EOF, "T_ReadObjType", name );
    return( type );
    break;
    
  default:
    E_XERR( E_ERRR, OBJ_EOF, "T_ReadObjType", name );
    break;
  }
  
  return( OBJ_EOF );
}

/*		
	Get the next character which is not a separator
*/
#if defined(_ANSI_)
static int T_GetNextChar( FILE *fd, char *name )
#else
static int T_GetNextChar(fd, name )
FILE *fd;
char *name ;

#endif

{
  int theChar;
  
  do{
    theChar= getc( fd );
  } while((theChar == '\n') || (theChar == '\t') || (theChar == ' '));
  
  return( theChar );
}

/*		Write the type of an object on disk								*/

#if defined(_ANSI_)
int T_WriteObjectType( FILE *fd, char *name, int type )
#else
int T_WriteObjectType(fd, name, type )
FILE *fd;
char *name;
int type ;

#endif

{
  int status;
  
  switch( type ){
  case OBJ_LIST: status= fprintf( fd, "(\n" ); break;
  case OBJ_ENDLIST: status= fprintf( fd, ")\n" ); break;
  case OBJ_ENDLOOP: status= fprintf( fd, "}\n" ); break;
  case OBJ_BEGINARRAY: status= fprintf( fd, "[ " ); break;
  case OBJ_ENDARRAY: status= fprintf( fd, "]\n" ); break;
  default: 
    status= fprintf( fd, "O%d", type ); 
    E_ERR( status==EOF, E_ERRW, EOF, "T_WriteObjectType", name );
    break;
  }
  E_ERR( status==EOF, E_ERRW, status, "T_WriteObjectType", name );
  return( status );
}

/*		Count the number of elements of a 3D list						*/

#if defined(_ANSI_)
int T_CountList( t_ListPtr theList )
#else
int T_CountList(theList )
t_ListPtr theList ;

#endif

{
  t_ListPtr start;
  int i=0;
  
  if ( theList == NULL ) return(0);
  start= theList;
  do{
    i += 1;
    theList= theList->next;
  } while( (theList != NULL) && (theList != start) );
  return(i);
}

/*
  Copy a list into another one.

This is a recursive procedure which copies recursively the items into the list.
*/

#if defined(_ANSI_)
int T_CopyList( t_ListPtr theList, t_ListHand handList )
#else
int T_CopyList(theList, handList )
t_ListPtr theList;
t_ListHand handList ;

#endif

{
  t_ListPtr newList= NULL;
  t_ListPtr start;
  t_ObjPtr newItem;
  
  if ( theList == NULL ){
    T_FreeList( handList );
    return(TRUE);
  }
  
  start= theList;
  do{
    if ( theList->item != NULL ){
      newItem= NULL;
      T_CopyObj( theList->item, &newItem );
      T_AddObjList( newItem, &newList );
    }
    theList= theList->next;
    
  } while ( (theList != NULL) && (theList != start) );
  
  T_ReverseList( &newList );
  T_FreeList( handList );
  *handList= newList;
  return( TRUE );
}

/*		Apply a function to a list, with a list of arguments			*/

#if defined(_ANSI_)
void T_ApplyList( t_ListPtr theList, theFunc_type theFunc, int argc, char *argv[] )
#else
void T_ApplyList(theList, theFunc, argc, argv)
t_ListPtr theList;
theFunc_type theFunc;
int argc;
char *argv[] ;

#endif

{
  t_ListPtr start;
  
  if ( theList == NULL ) return;
  
  start= theList;
  do{
    theFunc( theList->item, argc, argv );
    theList= theList->next;
  } while( (theList != NULL) && (theList != start) );
}

/*
	Link two elements of a list, first -> second
*/

#if defined(_ANSI_)
void T_ListLink( t_ListPtr elem1, t_ListPtr elem2 )
#else
void T_ListLink(elem1, elem2 )
t_ListPtr elem1;
t_ListPtr elem2 ;

#endif

{
  if ( elem1 != NULL ) elem1->next= elem2;
  if ( elem2 != NULL ) elem2->back= elem1;
}

/*
	Cuts the bi-directional link between two elements
*/


#if defined(_ANSI_)
void T_LinkCut( t_ListPtr elem1, t_ListPtr elem2 )
#else
void T_LinkCut(elem1, elem2 )
t_ListPtr elem1;
t_ListPtr elem2 ;

#endif

{
  if ( elem1 != NULL ) elem1->next= NULL;
  if ( elem2 != NULL ) elem2->back= NULL;
}

/*		Add an object to the begining of a list							*/

#if defined(_ANSI_)
void T_AddObjList( t_ObjPtr theObj, t_ListHand listHand )
#else
void T_AddObjList(theObj, listHand )
t_ObjPtr theObj;
t_ListHand listHand ;

#endif

{
  t_ListPtr newElem;
  t_ListPtr list,back;
  
  if ( theObj == NULL ) return;
  
  T_AllocList( &newElem );
  E_ISERRN("T_AddObjList", "adding an object to a list");
  
  newElem->item= theObj;
  list= *listHand;
  if ( list && list->back ) back= list->back;
  else					  back= NULL;
  T_ListLink( newElem, list );
  if ( back ) T_ListLink( back, newElem );
  *listHand= newElem;
}

/*
  Remove a list element from the list it belongs to

Returns: the corresponding item.
*/

#if defined(_ANSI_)
t_ObjPtr T_SubObjList( t_ListPtr theElem, t_ListHand listHand )
#else
t_ObjPtr T_SubObjList(theElem, listHand )
t_ListPtr theElem;
t_ListHand listHand ;

#endif

{
  t_ListPtr prev,next;
  t_ObjPtr theObj;
  
  if ( theElem == NULL ) return(NULL);
  if ( theElem == *listHand ){
    *listHand= theElem->next;
    if ( *listHand == theElem ) *listHand= NULL;
  }	
  
  prev= theElem->back;
  next= theElem->next;
  
  T_ListLink( prev, next );
  
  theObj= theElem->item;
  T_Free( (void **) &theElem );
  return( theObj );
}

/*
  Remove the element from a list, and destroy it.

Returns: TRUE on success else return FALSE.
*/

#if defined(_ANSI_)
int T_DeleteObjList( t_ListPtr theElem, t_ListHand handList )
#else
int T_DeleteObjList(theElem, handList )
t_ListPtr theElem;
t_ListHand handList ;

#endif

{
  t_ObjPtr theObj;
  
  theObj= T_SubObjList( theElem, handList );
  if ( theObj == NULL ) return( FALSE );
  return( T_FreeObj( &theObj ) );
}

/*
  Append the content of a list to another list
  
  list1,list2 => list1->list2. second list must be an open list.
*/

#if defined(_ANSI_)
void T_AppendList( t_ListPtr list1, t_ListHand handList2 )
#else
void T_AppendList(list1, handList2 )
t_ListPtr list1;
t_ListHand handList2 ;

#endif

{
  t_ListPtr last;
  
  if ( list1 == NULL ) return;
  
  E_ERRN( T_IsListClosed( *handList2 ), E_ERRFATAL, "T_AppendList",
	 "second list is closed" );
  
  T_ListLast( list1, &last );
  T_ListLink( last, *handList2 );
  *handList2= list1;
}

/*
  Sort the list with the quick sort procedure. 
  
  func is a function which takes two HANDLES! toward the objects o1 and o2,
  and return:
  
  1 if o1 > o2
  0 if o1 == o2
  -1 if o1 < o2
  
  It takes as input a handle toward the list, to remember you that
  the order of the elements may have changed ! type is used for type
  checking, but can be ignored if set to OBJ_UNKNOWN.
  
*/


#if defined(_ANSI_)
void T_SortList( t_ListHand handList, void_func_type func, int type )
#else
void T_SortList(handList, func, type)
t_ListHand handList;
void_func_type func;
int type;
#endif

{
  int count,i;
  t_ListPtr theList;
  t_ObjHand tb= NULL;
  t_ObjHand tbPtr;
  
  count= T_CountList( *handList );
  if ( count <= 1 ) return;
  
  tb= (t_ObjHand) T_Alloc( count * sizeof(t_ObjPtr) );
  
  theList= *handList;
  tbPtr= tb;
  
  for( i=0; i< count; i++ ){
    
    T_ListItemErr( theList, tbPtr, type );
    tbPtr += 1;
    theList= theList->next;
  }
  
  qsort( (void *) tb, (size_t) count, sizeof(t_ObjPtr), (func_type_2_arg)func );
  
  theList= *handList;
  tbPtr= tb;
  
  for( i=0; i< count; i++ ){
    
    theList->item= *tbPtr;
    tbPtr += 1;
    theList= theList->next;
  }
  
  T_Free( (void **) &tb );	
}

/*
  Transform a bucket of a hash table into a list of objects.

  type is used for type checking, if it is not OBJ_UNKNOWN.
  BEWARE ! the objects are not created, only the pointers are copied.
  If you wand to detatch this list from all previous operation, you
  have to copy it with T_CopyList, and to delete the previous one
  with T_DeleteList, such as:
 
<pre> 
  T_HashBucket2List( theHTable, code, &auxList, OBJ_UNKNOWN );
  T_CopyList( auxList, &theList );
  T_DeleteList( &auxList );
  ...
</pre>
*/

#if defined(_ANSI_)
void T_HashBucket2List( t_HTablePtr theHTable, int code, t_ListHand handList, int type )
#else
void T_HashBucket2List(theHTable, code, handList, type )
t_HTablePtr theHTable;
int code;
t_ListHand handList;
int type ;

#endif

{   
  t_ObjPtr obj;
  t_EntryPtr theEntry;
  
  code= code & theHTable->mask;
  theEntry= theHTable->tb[ code ];
  
  while( theEntry ){ 
    
    obj= (t_ObjPtr) theEntry->item;
    E_ERRN( (type != OBJ_UNKNOWN) && (obj->type != type),
	   E_ERRMISMATCH, "T_HashBucket2List", "wrong item type" );
    
    T_AddObjList( obj, handList );
    theEntry= theEntry->next;
  }
}

/*
  Expand a list of list of ... list of list of objects.

 (which are not lists) into a single list of object. Place NULL in *handRes to create a new list, else the elements are appened to this list.
  
This procedure do not copy the elementary object, but only use their
pointers.
*/

#if defined(_ANSI_)
void T_ExpandList( t_ListPtr theList, t_ListHand handRes )
#else
void T_ExpandList(theList, handRes )
t_ListPtr theList;
t_ListHand handRes ;

#endif

{
  t_ListPtr start;
  t_ObjPtr aObj;
  
  if ( theList == NULL ) return;
  start= theList;
  do{
    T_ListItemErr( theList, &aObj, OBJ_UNKNOWN );	/* no type checking */
    if ( aObj ){
      if ( aObj->type == OBJ_LIST ) 			/* recursive call */
	T_ExpandList( (t_ListPtr) aObj, handRes );
      else 	T_AddObjList( aObj, handRes );
    }
    theList= theList->next;
  } while( theList && (theList != start) );
}

/*----------------------------------------------------------------------*/
/************************************************************************/
/*
	OpersMap

	General operations on maps (obsolete)

	(c) Jean-Philippe THIRION -- INRIA -- 8/7/91
*/
/************************************************************************/

/*		Allocate a new map element										*/

#if defined(_ANSI_)
int T_Alloc3Map( t_3MapHand handMap )
#else
int T_Alloc3Map(handMap )
t_3MapHand handMap ;

#endif

{
  t_3MapPtr theMap;
  int i;
  
  theMap= (t_3MapPtr) T_Alloc( sizeof(t_3Map) );
  E_ISERR( FALSE, "T_Alloc3Map", NULL );
  
  theMap->obj.type= OBJ_3MAP;
  for( i=0; i< 4; i++ ) theMap->n[i]= NULL;
  theMap->weight= 0;
  theMap->num= -1;
  theMap->state= 0;
  
  return( TRUE );
}

/*		Set all the flags of the map to a given value. 					*/

#if defined(_ANSI_)
void T_Set3Map( t_3MapPtr theMap, int value )
#else
void T_Set3Map(theMap, value )
t_3MapPtr theMap;
int value ;

#endif

{
  int i;
  
  if ( theMap == NULL ) return;
  if ( theMap->state == value ) return;
  
  theMap->state= value;
  for(i=0;i<4;i++) T_Set3Map( theMap->n[i], value );
}

/*		Set the number of a map											*/

#if defined(_ANSI_)
void T_SetNum3Map( t_3MapPtr theMap, int num )
#else
void T_SetNum3Map(theMap, num )
t_3MapPtr theMap;
int num ;

#endif

{
  if ( theMap->num == num ) return;
  T_RecSetNum3Map( theMap, theMap->state, num );
}

#if defined(_ANSI_)
static void T_RecSetNum3Map( t_3MapPtr theMap, int state, int num )
#else
static void T_RecSetNum3Map(theMap, state, num )
t_3MapPtr theMap;
int state;
int num ;

#endif

{
  int i;
  
  if ( theMap == NULL ) return;
  if ( theMap->state != state ) return;
  
  theMap->state += 1;
  theMap->num= num;
  for( i=0; i< 4; i++ ){
    T_RecSetNum3Map( theMap->n[i], state, num );
  }
}

/*
	Connect a new map to a given map.

The second map number begomes the first map number, and the states are uniformized.
*/

#if defined(_ANSI_)
void T_Connect3Map( t_3MapPtr theMap, int v1, t_3MapPtr newMap, int v2 )
#else
void T_Connect3Map(theMap, v1, newMap, v2 )
t_3MapPtr theMap;
int v1;
t_3MapPtr newMap;
int v2 ;

#endif

{
  E_ERRN( (theMap == NULL) || (newMap == NULL), E_ERRNULL,
	 "T_Connect3Map" , "empty map(s)" );
  
  E_ERRN( (theMap->n[v1] != NULL) ||
	 (newMap->n[v2] != NULL), E_ERRNULL,
	 "T_Connect3Map", "not empty vertex" );
  
  T_SetNum3Map( newMap, theMap->num );
  T_Set3Map( newMap, theMap->state );
  theMap->n[v1]= newMap;
  newMap->n[v2]= theMap;
}

/*		Free a map 														*/

#if defined(_ANSI_)
void T_Free3Map( t_3MapHand handMap )
#else
void T_Free3Map(handMap )
t_3MapHand handMap ;
#endif

{
}

/*----------------------------------------------------------------------*/
/************************************************************************/
/*
	Opers3PDual

	General operations on 3PDual points (obsolete)

	(c) Jean-Philippe THIRION -- INRIA
*/
/************************************************************************/

/*		Perform the initialization of all the functions on 3PDual		*/

#if defined(_ANSI_)
void T_Obj3PDualInit( void )
#else
void T_Obj3PDualInit()

#endif

{
  
  gObjTb[ OBJ_3PDUAL ][ OBJF_ALLOC ]= ( void_func_type ) T_Alloc3PDual;
  gObjTb[ OBJ_3PDUAL ][ OBJF_FREE ]= ( void_func_type ) T_Free3PDual;
  gObjTb[ OBJ_3PDUAL ][ OBJF_FREAD ]= ( void_func_type ) T_FRead3PDual;
  gObjTb[ OBJ_3PDUAL ][ OBJF_FWRITE ]= ( void_func_type ) T_FWrite3PDual;
  gObjTb[ OBJ_3PDUAL ][ OBJF_COPY ]= ( void_func_type ) T_Copy3PDual;
  gObjTb[ OBJ_3PDUAL ][ OBJF_DEFAULT ]= NULL;
}

/*		Free a 3PDual point												*/

#if defined(_ANSI_)
void T_Free3PDual( t_3PDualHand dualHand )
#else
void T_Free3PDual(dualHand )
t_3PDualHand dualHand ;

#endif

{
  free( (void *) *dualHand );
  *dualHand= NULL;
}

/*		Allocate a 3PDual point											*/

#if defined(_ANSI_)
int T_Alloc3PDual( t_3PDualHand dualHand )
#else
int T_Alloc3PDual(dualHand )
t_3PDualHand dualHand ;

#endif

{
  *dualHand= (t_3PDualPtr) T_Alloc( sizeof(t_3PDual) );
  E_ISERR( FALSE, "T_Alloc3PDual", NULL );
  
  (*dualHand)->obj.type= OBJ_3PDUAL;
  return( TRUE );
}

/*		Write a 3PDual point in an ascii file							*/

#if defined(_ANSI_)
int T_FWrite3PDual( FILE *fd, char *name, t_3PDualPtr theDual )
#else
int T_FWrite3PDual(fd, name, theDual )
FILE *fd;
char *name;
t_3PDualPtr theDual ;

#endif

{
  int status;
  
  status= fprintf( fd, " %f %f\n", theDual->pt.x, theDual->pt.y );
  E_ERR( status == EOF, E_ERRW, FALSE, "T_FWrite3PDual", name);
  return( TRUE );
}

/*		Read a 3PDual point in an ascii file							*/

#if defined(_ANSI_)
int T_FRead3PDual( FILE *fd, char *name, t_3PDualHand handDual )
#else
int T_FRead3PDual(fd, name, handDual )
FILE *fd;
char *name;
t_3PDualHand handDual ;

#endif

{
  int status;
  t_3PDualPtr theDual;
  float x,y;
  
  if ( *handDual == NULL ) T_Alloc3PDual( handDual );
  E_ISERR( FALSE, "T_FRead3Dual", name );
  
  theDual= *handDual;
  
  status= fscanf( fd, "%f %f\n", &x, &y);
  if ( status == EOF ) {
    T_Warn("error reading in file %s in T_FReadDual\n", name);
    return( FALSE );
  }
  theDual->pt.x = (reeltype) x;
  theDual->pt.y = (reeltype) y;

  return( TRUE );
}

/*		Copy a dual point into another one								*/

#if defined(_ANSI_)
int T_Copy3PDual( t_3PDualPtr theDual, t_3PDualHand handDual )
#else
int T_Copy3PDual(theDual, handDual )
t_3PDualPtr theDual;
t_3PDualHand handDual ;

#endif

{
  if ( theDual == NULL ) { *handDual= NULL; return( TRUE ); }
  if ( *handDual == NULL ){
    *handDual= (t_3PDualPtr) T_Alloc( sizeof(t_3PDual) );
  }
  E_ISERR( FALSE, "T_Copy3PDual", NULL);
  
  memcpy( (void *) *handDual, (void *) theDual, (size_t) sizeof(t_3PDual) );
  return( TRUE );
}

/*----------------------------------------------------------------------*/
/************************************************************************/
/*
	OpersCrest

	General operations on Crest points

	(c) Jean-Philippe THIRION -- INRIA -- 8 janvier 1992
*/
/************************************************************************/

/*		Perform the initialization of all the functions on Crest		*/

#if defined(_ANSI_)
void T_ObjCrestInit( void )
#else
void T_ObjCrestInit()

#endif

{
  
  gObjTb[ OBJ_CREST ][ OBJF_ALLOC ]= ( void_func_type ) T_AllocCrest;
  gObjTb[ OBJ_CREST ][ OBJF_FREE ]= ( void_func_type ) T_FreeCrest;
  gObjTb[ OBJ_CREST ][ OBJF_FREAD ]= ( void_func_type ) T_FReadCrest;
  gObjTb[ OBJ_CREST ][ OBJF_FWRITE ]= ( void_func_type ) T_FWriteCrest;
  gObjTb[ OBJ_CREST ][ OBJF_COPY ]= ( void_func_type ) T_CopyCrest;
  gObjTb[ OBJ_CREST ][ OBJF_DEFAULT ]= ( void_func_type ) T_DefaultCrest;
}

/*		Free a Crest point												*/

#if defined(_ANSI_)
void T_FreeCrest( t_CrestHand crestHand )
#else
void T_FreeCrest(crestHand )
t_CrestHand crestHand ;

#endif

{
  free( (void *) *crestHand );
  *crestHand= NULL;
}

/*
  Initialize crest structure
  */

#if defined(_ANSI_)
void T_DefaultCrest( t_CrestPtr crest )
#else
void T_DefaultCrest(crest )
t_CrestPtr crest ;

#endif

{
  (void *) memset( (void *) crest, 0, (size_t) sizeof(t_Crest) );
  crest->obj.type= OBJ_CREST;
}

/*		Allocate a Crest point											*/

#if defined(_ANSI_)
int T_AllocCrest( t_CrestHand crestHand )
#else
int T_AllocCrest(crestHand )
t_CrestHand crestHand ;

#endif

{
  *crestHand= (t_CrestPtr) T_Alloc( sizeof(t_Crest) );
  E_ISERR( FALSE, "T_AllocCrest", NULL );
  
  T_DefaultCrest( *crestHand );
  return( TRUE );
}

/*		Write a Crest point in an ascii file							*/

#if defined(_ANSI_)
int T_FWriteCrest( FILE *fd, char *name, t_CrestPtr crest )
#else
int T_FWriteCrest(fd, name, crest )
FILE *fd;
char *name;
t_CrestPtr crest ;

#endif

{
  int status;
  
  status= fprintf( fd, "\n%f %f %f\n", crest->pt.x,crest->pt.y,crest->pt.z );
  E_ERR( status == EOF, E_ERRW, FALSE, "T_FWriteCrest", name);
  
  status= fprintf( fd, "%f %f %f\n", crest->t1.x,crest->t1.y,crest->t1.z );
  E_ERR( status == EOF, E_ERRW, FALSE, "T_FWriteCrest", name);
  
  status= fprintf( fd, "%f %f %f\n", crest->t2.x,crest->t2.y,crest->t2.z );
  E_ERR( status == EOF, E_ERRW, FALSE, "T_FWriteCrest", name);
  
  status= fprintf( fd, "%f %f %f\n", crest->k1, crest->k2, crest->conf );
  E_ERR( status == EOF, E_ERRW, FALSE, "T_FWriteCrest", name);
  return( TRUE );
}

/*		Read a Crest point in an ascii file								*/

#if defined(_ANSI_)
int T_FReadCrest( FILE *fd, char *name, t_CrestHand handCrest )
#else
int T_FReadCrest(fd, name, handCrest )
FILE *fd;
char *name;
t_CrestHand handCrest ;

#endif

{
  int status;
  t_CrestPtr crest;
  float x,y,z;
  
  if ( *handCrest == NULL ){
    T_AllocCrest( handCrest );
    E_ISERR( FALSE, "T_FReadCrest", name );
  }
  
  crest= *handCrest;
  
  status= fscanf( fd, "%f %f %f\n", &x, &y, &z);
  E_ERR( status == EOF, E_ERRW, FALSE, "T_FReadCrest", name);
  crest->pt.x = (reeltype) x;
  crest->pt.y = (reeltype) y;
  crest->pt.z = (reeltype) z;
  
  status= fscanf( fd, "%f %f %f\n",  &x, &y, &z);
  E_ERR( status == EOF, E_ERRW, FALSE, "T_FReadCrest", name);
  crest->t1.x = (reeltype) x;
  crest->t1.y = (reeltype) y;
  crest->t1.z = (reeltype) z;
  
  status= fscanf( fd, "%f %f %f\n",  &x, &y, &z);
  E_ERR( status == EOF, E_ERRW, FALSE, "T_FReadCrest", name);
  crest->t2.x = (reeltype) x;
  crest->t2.y = (reeltype) y;
  crest->t2.z = (reeltype) z;

  status= fscanf( fd, "%f %f %f\n", &x, &y, &z);
  E_ERR( status == EOF, E_ERRW, FALSE, "T_FReadCrest", name);
  crest->k1 = (reeltype) x; 
  crest->k2 = (reeltype) y;
  crest->conf = (reeltype) z;


  return( TRUE );
}

/*		Copy a crest point into another one								*/

#if defined(_ANSI_)
int T_CopyCrest( t_CrestPtr theCrest, t_CrestHand handCrest )
#else
int T_CopyCrest(theCrest, handCrest )
t_CrestPtr theCrest;
t_CrestHand handCrest ;

#endif

{
  if ( theCrest == NULL ) { *handCrest= NULL; return( TRUE ); }
  E_ERR( (theCrest->obj.type) != OBJ_CREST,
	E_ERRMISMATCH, FALSE, "T_CopyCrest", "not a good type" );
  
  if ( *handCrest == NULL ){
    T_AllocCrest( handCrest );
    E_ISERR( FALSE, "T_CopyCrest", NULL );
  }
  
  memcpy( (void *) *handCrest, (void *) theCrest, (size_t) sizeof(t_Crest) );
  return( TRUE );
}

/*----------------------------------------------------------------------*/
/************************************************************************/
/*
	OpersInt

	General operations on Int points

	(c) Jean-Philippe THIRION -- INRIA -- 8 janvier 1992
*/
/************************************************************************/

/*		Perform the initialization of all the functions on Int		*/

#if defined(_ANSI_)
void T_ObjIntInit( void )
#else
void T_ObjIntInit()

#endif

{
  
  gObjTb[ OBJ_INT ][ OBJF_ALLOC ]= ( void_func_type ) T_AllocInt;
  gObjTb[ OBJ_INT ][ OBJF_FREE ]= ( void_func_type ) T_FreeInt;
  gObjTb[ OBJ_INT ][ OBJF_FREAD ]= ( void_func_type ) T_FReadInt;
  gObjTb[ OBJ_INT ][ OBJF_FWRITE ]= ( void_func_type ) T_FWriteInt;
  gObjTb[ OBJ_INT ][ OBJF_COPY ]= ( void_func_type ) T_CopyInt;
  gObjTb[ OBJ_INT ][ OBJF_DEFAULT ]= ( void_func_type ) T_DefaultInt;
}

/*		Free a Int point												*/

#if defined(_ANSI_)
void T_FreeInt( t_IntHand IntHand )
#else
void T_FreeInt(IntHand )
t_IntHand IntHand ;

#endif

{
  free( (void *) *IntHand );
  *IntHand= NULL;
}

/*
  default value
  */

#if defined(_ANSI_)
void T_DefaultInt( t_IntPtr theInt )
#else
void T_DefaultInt(theInt )
t_IntPtr theInt ;

#endif

{
  (void *) memset( (void *) theInt, 0, (size_t) sizeof(t_Int) );
  theInt->obj.type= OBJ_INT;
}

/*		Allocate a Int point											*/

#if defined(_ANSI_)
int T_AllocInt( t_IntHand IntHand )
#else
int T_AllocInt(IntHand )
t_IntHand IntHand ;

#endif

{
  *IntHand= (t_IntPtr) T_Alloc( sizeof(t_Int) );
  E_ISERR( FALSE, "T_AllocInt", NULL );
  
  T_DefaultInt( *IntHand );
  return( TRUE );
}

/*		Write a Int point in an ascii file							*/

#if defined(_ANSI_)
int T_FWriteInt( FILE *fd, char *name, t_IntPtr Int )
#else
int T_FWriteInt(fd, name, Int )
FILE *fd;
char *name;
t_IntPtr Int ;

#endif

{
  int status;
  
  status= fprintf( fd, " %d", Int->index );
  E_ERR( status == EOF, E_ERRW, FALSE, "T_FWriteInt", name );
  
  return( TRUE );
}

/*		Read a Int point in an ascii file								*/

#if defined(_ANSI_)
int T_FReadInt( FILE *fd, char *name, t_IntHand handInt )
#else
int T_FReadInt(fd, name, handInt )
FILE *fd;
char *name;
t_IntHand handInt ;

#endif

{
  int status;
  t_IntPtr Int;
  
  if ( *handInt == NULL ){
    T_AllocInt( handInt );
    E_ISERR( FALSE, "T_FReadInt", name );
  }
  
  Int= *handInt;
  
  status= fscanf( fd, "%d ", &Int->index);
  E_ERR( status == EOF, E_ERRR, FALSE, "T_FReadInt", name );
  
  return( TRUE );
}

/*		Copy a Int point into another one								*/

#if defined(_ANSI_)
int T_CopyInt( t_IntPtr theInt, t_IntHand handInt )
#else
int T_CopyInt(theInt, handInt )
t_IntPtr theInt;
t_IntHand handInt ;

#endif

{
  if ( theInt == NULL ) { *handInt= NULL; return( TRUE ); }
  E_ERR( (theInt->obj.type) != OBJ_INT,E_ERRMISMATCH,FALSE,"T_CopyInt",NULL);
  
  if ( *handInt == NULL ){
    T_AllocInt( handInt );
    E_ISERR( FALSE, "T_CopyInt", NULL);
  }
  
  memcpy( (void *) *handInt, (void *) theInt, (size_t) sizeof(t_Int) );
  return( TRUE );
}

/*----------------------------------------------------------------------*/
/************************************************************************/
/*
	OpersFloat

	General operations on floats objects

	(c) Jean-Philippe THIRION -- INRIA -- 15 juin 1993
*/
/************************************************************************/

/*		Perform the initialization of all the functions on floats		*/

#if defined(_ANSI_)
void T_ObjFloatInit( void )
#else
void T_ObjFloatInit()

#endif

{
  
  gObjTb[ OBJ_FLOAT ][ OBJF_ALLOC ]= NULL;
  gObjTb[ OBJ_FLOAT ][ OBJF_FREE ]= NULL;
  gObjTb[ OBJ_FLOAT ][ OBJF_FREAD ]= ( void_func_type ) T_FReadFloat;
  gObjTb[ OBJ_FLOAT ][ OBJF_FWRITE ]= ( void_func_type ) T_FWriteFloat;
  gObjTb[ OBJ_FLOAT ][ OBJF_COPY ]= NULL;
  gObjTb[ OBJ_FLOAT ][ OBJF_DEFAULT ]= NULL;
}

/*
  Write a float in an ascii file
  */

#if defined(_ANSI_)
int T_FWriteFloat( FILE *fd, char *name, t_FloatPtr aFloat )
#else
int T_FWriteFloat(fd, name, aFloat )
FILE *fd;
char *name;
t_FloatPtr aFloat ;

#endif

{
  int status;
  
  status= fprintf( fd, " %f ", aFloat->v );
  E_ERR( status == EOF, E_ERRW, FALSE, "T_FWriteFloat", name );
  
  return( TRUE );
}

/*
  Read a float in an ascii file
  */

#if defined(_ANSI_)
int T_FReadFloat( FILE *fd, char *name, t_FloatHand handFloat )
#else
int T_FReadFloat(fd, name, handFloat )
FILE *fd;
char *name;
t_FloatHand handFloat ;

#endif

{
  int status;
  t_FloatPtr aFloat;
  
  if ( *handFloat == NULL ){
    T_AllocObj( OBJ_FLOAT, (t_ObjHand) handFloat );
    E_ISERR( FALSE, "T_FReadFloat", name );
  }
  
  aFloat= *handFloat;
  
  status= fscanf( fd, "%f ", &aFloat->v );
  E_ERR( status == EOF, E_ERRR, FALSE, "T_FReadFloat", name );
  
  return( TRUE );
}

/*----------------------------------------------------------------------*/
/************************************************************************/
/*
	OpersPoint

  General operations on 3D vectors of floating point
  
  (c) Jean-Philippe THIRION -- INRIA --	july 8 1992
  */
/************************************************************************/

/*		Perform the initialization of all the functions on Point		*/

#if defined(_ANSI_)
void T_ObjPointInit( void )
#else
void T_ObjPointInit()

#endif

{
  
  gObjTb[ OBJ_POINT ][ OBJF_ALLOC ]= ( void_func_type ) T_AllocPoint;
  gObjTb[ OBJ_POINT ][ OBJF_FREE ]= ( void_func_type ) T_FreePoint;
  gObjTb[ OBJ_POINT ][ OBJF_FREAD ]= ( void_func_type ) T_FReadPoint;
  gObjTb[ OBJ_POINT ][ OBJF_FWRITE ]= ( void_func_type ) T_FWritePoint;
  gObjTb[ OBJ_POINT ][ OBJF_COPY ]= ( void_func_type ) T_CopyPoint;
  gObjTb[ OBJ_POINT ][ OBJF_DEFAULT ]= ( void_func_type ) T_DefaultPoint;
}

/*		Free a Point point												*/

#if defined(_ANSI_)
void T_FreePoint( t_PointHand handPoint )
#else
void T_FreePoint(handPoint )
t_PointHand handPoint ;

#endif

{
  free( (void *) *handPoint );
  *handPoint= NULL;
}

/*
  Initialize point structure
  */

#if defined(_ANSI_)
void T_DefaultPoint( t_PointPtr point )
#else
void T_DefaultPoint(point )
t_PointPtr point ;

#endif

{
  (void *) memset( (void *) point, 0, (size_t) sizeof(t_Point) );
  point->obj.type= OBJ_POINT;
}

/*		Allocate a Point point											*/

#if defined(_ANSI_)
int T_AllocPoint( t_PointHand handPoint )
#else
int T_AllocPoint(handPoint )
t_PointHand handPoint ;

#endif

{
  *handPoint= (t_PointPtr) T_Alloc( sizeof(t_Point) );
  E_ISERR( FALSE, "T_AllocPoint", NULL );
  
  T_DefaultPoint( *handPoint );
  return( TRUE );
}

/*		Write a point in an ascii file									*/

#if defined(_ANSI_)
int T_FWritePoint( FILE *fd, char *name, t_PointPtr point )
#else
int T_FWritePoint(fd, name, point )
FILE *fd;
char *name;
t_PointPtr point ;

#endif

{
  int status;
  
  status= fprintf(fd," %f %f %f\n", point->pt.x, point->pt.y, point->pt.z );
  E_ERR( status == EOF, E_ERRW, FALSE, "T_FWritePoint", name);
  
  return( TRUE );
}

/*		Read a Point poPoint in an ascii file								*/

#if defined(_ANSI_)
int T_FReadPoint( FILE *fd, char *name, t_PointHand handPoint )
#else
int T_FReadPoint(fd, name, handPoint )
FILE *fd;
char *name;
t_PointHand handPoint ;

#endif

{
  int status;
  t_PointPtr point;
  float x,y,z;
  
  if ( *handPoint == NULL ){
    T_AllocPoint( handPoint );
    E_ISERR( FALSE, "T_FReadPoint", name );
  }
  
  point= *handPoint;
  
  status= fscanf( fd, "%f %f %f", &x, &y, &z);
  E_ERR( status == EOF, E_ERRR, FALSE, "T_FReadPoint", name);
  point->pt.x = (reeltype) x;
  point->pt.y = (reeltype) y;
  point->pt.z = (reeltype) z;
  
  return( TRUE );
}

/*		Copy a Point point into another one								*/

#if defined(_ANSI_)
int T_CopyPoint( t_PointPtr thePoint, t_PointHand handPoint )
#else
int T_CopyPoint(thePoint, handPoint )
t_PointPtr thePoint;
t_PointHand handPoint ;

#endif

{
  if ( thePoint == NULL ) { *handPoint= NULL; return( TRUE ); }
  E_ERR( (thePoint->obj.type) != OBJ_POINT,
	E_ERRMISMATCH, FALSE, "T_CopyPoint", "not the expected type" );
  
  if ( *handPoint == NULL ){
    T_AllocPoint( handPoint );
    E_ISERR( FALSE, "T_FReadPoint", NULL );
  }
  
  memcpy( (void *) *handPoint, (void *) thePoint, (size_t) sizeof(t_Point) );
  return( TRUE );
}

/*----------------------------------------------------------------------*/
/************************************************************************/
/*
	OpersArray

  General operations on Object Arrays. This structure is very useful
  to process and store relatively large numbers of fixed sized objects
  of the same type. It is a very compact representation of the 
  information and can be fully dynamic, due to the use of the
  realloc function.
  
  (c) Jean-Philippe THIRION -- INRIA --	August 13 1992
  */
/************************************************************************/

/*
  Perform the initialization of all the functions on Array
  */

#if defined(_ANSI_)
void T_ObjArrayInit( void )
#else
void T_ObjArrayInit()

#endif

{
  
  gObjTb[ OBJ_ARRAY ][ OBJF_ALLOC ]= ( void_func_type ) T_AllocArray;
  gObjTb[ OBJ_ARRAY ][ OBJF_FREE ]= ( void_func_type ) T_FreeArray;
  gObjTb[ OBJ_ARRAY ][ OBJF_FREAD ]= ( void_func_type ) T_FReadArray;
  gObjTb[ OBJ_ARRAY ][ OBJF_FWRITE ]= ( void_func_type ) T_FWriteArray;
  gObjTb[ OBJ_ARRAY ][ OBJF_COPY ]= ( void_func_type ) T_CopyArray;
  gObjTb[ OBJ_ARRAY ][ OBJF_DEFAULT ]= ( void_func_type ) T_DefaultArray;
}

/*
  Free a Array structure
  */

#if defined(_ANSI_)
void T_FreeArray( t_ArrayHand handArray )
#else
void T_FreeArray(handArray )
t_ArrayHand handArray ;

#endif

{
  t_ArrayPtr theArray;
  
  if ( *handArray == NULL ) return;
  theArray= *handArray;
  
  T_Free( (void **) &theArray->tb );
  theArray->tb= NULL;
  theArray->nbAlloc= 0;
  theArray->nb= 0;
  
  T_Free( (void **) handArray );
  *handArray= NULL;
}

/*
  Initialize a Array with default values
*/

#if defined(_ANSI_)
int T_DefaultArray( t_ArrayPtr array )
#else
int T_DefaultArray(array )
t_ArrayPtr array ;

#endif

{
  (void *) memset( (void *) array, 0, (size_t) sizeof(t_Array) );
  array->obj.type= OBJ_ARRAY;
  return( TRUE );
}

/*
  Allocate and initialize a Array
*/

#if defined(_ANSI_)
int T_AllocArray( t_ArrayHand handArray )
#else
int T_AllocArray(handArray )
t_ArrayHand handArray ;

#endif

{
  *handArray= (t_ArrayPtr) T_Alloc( sizeof(t_Array) );
  E_ISERR( FALSE, "T_AllocArray", NULL );
  T_DefaultArray( *handArray );
  return( TRUE );
}

/*
  Re-allocate a Array, with the same elements
*/

#if defined(_ANSI_)
void T_ReAllocArrayTb( t_ArrayPtr theArray, int nb )
#else
void T_ReAllocArrayTb(theArray, nb )
t_ArrayPtr theArray;
int nb ;

#endif

{
  int size;
  
  if ( !nb ){
    T_Free( (void **) &theArray->tb );
    theArray->nbAlloc= 0;
    theArray->nb= 0;
    return;
  }
  
  size= _T_GetSizeUtil( (int) theArray->elem.type );
  theArray->tb = T_ReAlloc( &(theArray->tb), size * nb );
  E_REPERR( {theArray->nbAlloc= 0; theArray->tb= NULL;} );
  E_ISERRN( "T_ReAllocArrayTb", NULL );
  theArray->nbAlloc= nb;
}

/*		Write an array in an ascii file									*/

#if defined(_ANSI_)
int T_FWriteArray( FILE *fd, char *name, t_ArrayPtr theArray )
#else
int T_FWriteArray(fd, name, theArray )
FILE *fd;
char *name;
t_ArrayPtr theArray ;

#endif

{
  int status,size,create,i,pretty;
  char *auxPtr,*objPtr;
  t_ObjPtr auxObj=NULL;
  
  E_ERR( theArray->obj.type!=OBJ_ARRAY,E_ERRMISMATCH, FALSE,"T_FWriteArray",
	"not the expected object type" );
  
  E_ERR( !T_ObjType((int)theArray->elem.type),E_ERRUNKNOWN, FALSE,"T_FWriteArray",
	"unknown type of elements");
  
  if( fd == NULL ){
    create= TRUE;
    fd= fopen( name, "w" );
    E_ERR( fd==NULL, E_ERRC, FALSE, "T_FWriteArray", name );
    status= T_WriteObjectType( fd, name,(int) theArray->obj.type );
    E_REPERR( goto END_FWriteArray );
  }
  else create= FALSE;
  
  T_AllocObj((int) theArray->elem.type, &auxObj );
  E_REPERR( goto END_FWriteArray );
  
  if (( theArray->elem.type != OBJ_INT) || (theArray->nb>16)) pretty= FALSE; 
  else pretty= TRUE;
  
  status= fprintf(fd," " );
  if (status == EOF){ E_SETERR(E_ERRW,NULL,NULL); goto END_FWriteArray; }
  
  status= T_WriteObjectType( fd, name, (int) theArray->elem.type );
  E_REPERR( goto END_FWriteArray );
  
  status= fprintf(fd," %d ", theArray->nb );
  if (status == EOF){ E_SETERR(E_ERRW,NULL,NULL); goto END_FWriteArray; }
  
  if ( !pretty ) fprintf(fd,"\n");
  
  status= T_WriteObjectType( fd, name, OBJ_BEGINARRAY );
  E_REPERR( goto END_FWriteArray; );
  
  if ( !pretty ) fprintf(fd,"\n");
  
  size= _T_GetSizeUtil( (int) theArray->elem.type );	
  E_REPERR( goto END_FWriteArray );
  
  auxPtr= ((char *) auxObj) + sizeof(t_Obj);
  objPtr= theArray->tb;
  
  for( i=0; i< theArray->nb; i++ ){
    
    memcpy( (void *) auxPtr, (void *) objPtr, (size_t) size ); 
    status= T_FWriteObj( fd,name,auxObj);
    E_REPERR( goto END_FWriteArray );
    
    if( !pretty ) fprintf(fd,"\n");
    objPtr += size;
  }
  
 END_FWriteArray:
  
  T_WriteObjectType( fd, name, OBJ_ENDARRAY );
  
  T_Free((void **) &auxObj );	/* free object, but not linked structures	*/
  if ( create ) fclose( fd );
  
  E_ISERR( FALSE, "T_FWriteArray", NULL ); 
  return( TRUE );
}

/*		Read a Array array in an ascii file								*/

#if defined(_ANSI_)
int T_FReadArray( FILE *fd, char *name, t_ArrayHand handArray )
#else
int T_FReadArray(fd, name, handArray )
FILE *fd;
char *name;
t_ArrayHand handArray ;

#endif

{
  int status,opening,size,type,nb;
  char *auxPtr,*objPtr;
  t_ObjPtr auxObj;
  t_ArrayPtr theArray;
  
  /* test, allocation, opening */
  
  if ( *handArray == NULL ) T_AllocArray( handArray );
  theArray= *handArray;
  E_ISERR( FALSE, "T_FReadArray", name );
  
  if ( fd == NULL ){
    opening= TRUE;
    fd= fopen( name, "r" );
    if ( fd == NULL ){
      T_FreeArray( handArray );
      E_XERR( E_ERRO, FALSE, "T_FReadArray", name );
    }
    
    type= T_ReadObjectType( fd, name );
    if ( type != OBJ_ARRAY ){
      T_FreeArray( handArray );
      fclose(fd);
      E_XERR(E_ERRMISMATCH,FALSE,"T_FReadArray","not the expected type");
    }
  }
  else opening= FALSE;
  
  /* processing... */
  
  type= T_ReadObjectType( fd, name );
  if ( !T_ObjType(type) ){
    E_SETERR(E_ERRUNKNOWN,"T_FReadArray","unknown sub-objects type");
    goto END_FReadArray;
  }
  
  theArray->elem.type= type;
  T_AllocObj( (int) theArray->elem.type, &auxObj );
  E_REPERR( goto END_FReadArray; );
  
  status= fscanf(fd,"%d", &nb );
  if (status == EOF){ 
    E_SETERR(E_ERRR,"reading","total size"); 
    goto END_FReadArray;
  }
  
  type= T_ReadObjectType( fd, name );
  if (type != OBJ_BEGINARRAY) 
    E_SETERR(E_ERRMISMATCH,"T_FReadArray","array delimiter expected"); 
  
  E_REPERR( goto END_FReadArray; );
  
  T_AllocArrayTb( theArray, (int) theArray->elem.type, nb );
  
  size= _T_GetSizeUtil( (int) auxObj->type );	
  E_REPERR( goto END_FReadArray );
  
  auxPtr= ((char *) auxObj) + sizeof(t_Obj);
  objPtr= theArray->tb;
  
  for( theArray->nb=0; theArray->nb< nb; theArray->nb++ ){
    
    status= T_FReadObj( (int) theArray->elem.type, fd, name, &auxObj );
    if (status == FALSE) goto END_FReadArray;
    memcpy( (void *) objPtr, (void *) auxPtr, (size_t) size ); 
    objPtr += size;
  }
  
  type= T_ReadObjectType( fd, name );
  if (type != OBJ_ENDARRAY) 
    E_SETERR(E_ERRMISMATCH,"T_FReadArray","end array delimiter expected"); 
  
 END_FReadArray:
  
  if ( opening ) fclose( fd );
  T_Free( (void **) &auxObj );
  E_REPERR( T_FreeArray(handArray) );
  E_ISERR( FALSE, "T_FReadArray", NULL );
  return( TRUE );
}

/*
  Copy a Array into another one
  */

#if defined(_ANSI_)
int T_CopyArray( t_ArrayPtr theArray, t_ArrayHand handArray )
#else
int T_CopyArray(theArray, handArray )
t_ArrayPtr theArray;
t_ArrayHand handArray ;

#endif

{
  t_ArrayPtr resArray;
  int size;
  
  if ( theArray == NULL ) { *handArray= NULL; return( TRUE ); }
  E_ERR( (theArray->obj.type) != OBJ_ARRAY,
	E_ERRMISMATCH, FALSE, "T_CopyArray", "not the expected type" );
  
  if ( *handArray == NULL ){
    T_AllocArray( handArray );
    E_ISERR( FALSE, "T_CopyArray", NULL );
  }
  
  resArray= *handArray;
  T_Free( (void **) &resArray->tb );
  memcpy( (void *) resArray, (void *) theArray, (size_t) sizeof(t_Array) );
  resArray->nbAlloc= 0;
  resArray->tb= NULL;
  
  T_AllocArrayTb( resArray, (int) theArray->elem.type, theArray->nbAlloc );
  size= _T_GetSizeUtil( (int) resArray->elem.type );
  
  E_REPERR( { T_Free( (void **) &resArray->tb); T_Free((void **) handArray); } );
  E_ISERR( FALSE, "T_CopyArray", NULL);
  
  memcpy( (void *) resArray->tb, (void *) theArray->tb, (size_t) (size * theArray->nb) );
  resArray->nb= theArray->nb;
  return( TRUE );
}

/*
  Allocate the array of a Array with a given number of element
  */

#if defined(_ANSI_)
void T_AllocArrayTb( t_ArrayPtr theArray, int type, int nb )
#else
void T_AllocArrayTb(theArray, type, nb )
t_ArrayPtr theArray;
int type;
int nb ;

#endif

{
  int size;
  
  if (theArray->nbAlloc != 0){T_Free( (void **) &theArray->tb);theArray->nbAlloc= 0;}
  
  E_ERRN(!T_ObjType(type),E_ERRUNKNOWN,"T_AllocArrayArray",
	 "element type unknown" );
  
  theArray->elem.type= type;
  theArray->nb= 0;
  size= _T_GetSizeUtil( (int) theArray->elem.type );
  theArray->tb= T_Alloc(( unsigned ) ( size * nb ));
  E_REPERR( {theArray->nbAlloc= 0; theArray->tb= NULL;} );
  E_ISERRN( "T_AllocArrayTb", NULL );
  theArray->nbAlloc= nb;
}

/*
  Set the ith element of the array with given object values
  */

#if defined(_ANSI_)
void T_ArraySet( t_ArrayPtr theArray, t_ObjPtr theObj, int num )
#else
void T_ArraySet(theArray, theObj, num )
t_ArrayPtr theArray;
t_ObjPtr theObj;
int num ;

#endif

{
  int size;
  char *auxPtr,*objPtr;
  
  E_ERRN((num < 0) || (num >= theArray->nbAlloc), E_ERRARRAY, "T_ObjArraySet",
	 "out of the bounds of the Array array" );
  
  E_ERRN( theObj->type != theArray->elem.type, E_ERRMISMATCH, "T_ObjArraySet",
	 "object and Array elements has not the same type" );
  
  size= _T_GetSizeUtil( (int) theObj->type );
  
  auxPtr= theArray->tb + size * num;
  objPtr= ((char *) theObj) + sizeof(t_Obj);
  memcpy( (void *) auxPtr, (void *) objPtr, (size_t) size );
  if ( num >= theArray->nb ) theArray->nb= num+1;
}

/*
  Get the ith element of an array, and put the result into an
  object record
  */

#if defined(_ANSI_)
void T_ArrayGet( t_ArrayPtr theArray, t_ObjPtr theObj, int num )
#else
void T_ArrayGet(theArray, theObj, num )
t_ArrayPtr theArray;
t_ObjPtr theObj;
int num ;

#endif

{
  int size;
  char *auxPtr,*objPtr;
  
  E_ERRN((num < 0) || (num > theArray->nbAlloc), E_ERRARRAY, "T_ObjArrayGet",
	 "out of the bounds of the Array array" );
  
  E_ERRN( theObj->type != theArray->elem.type, E_ERRMISMATCH, "T_ObjArrayGet",
	 "object and Array elements has not the same type" );
  
  size= _T_GetSizeUtil((int) theObj->type);
  
  auxPtr= theArray->tb + size * num;
  objPtr= ((char *) theObj) + sizeof(t_Obj);
  memcpy( (void *) objPtr, (void *) auxPtr, (size_t) size );
}

/*
  Add a new element to an array, without reallocating memory. Return
  the position in the array on success, return -1 if out of bounds.
  */

#if defined(_ANSI_)
int T_ArrayAdd( t_ArrayPtr theArray, t_ObjPtr theObj )
#else
int T_ArrayAdd(theArray, theObj )
t_ArrayPtr theArray;
t_ObjPtr theObj ;

#endif

{
  int size,pos;
  char *auxPtr,*objPtr;
  
  E_ERR( theObj->type != theArray->elem.type, E_ERRMISMATCH, -1, "T_ArrayAdd",
	"object and Array elements has not the same type" );
  
  pos= theArray->nb;
  if ( pos >= theArray->nbAlloc ) return( -1 );
  
  size= _T_GetSizeUtil( (int) theObj->type );
  
  auxPtr= theArray->tb + size * pos;
  objPtr= ((char *) theObj) + sizeof(t_Obj);
  memcpy( (void *) auxPtr, (void *) objPtr, (size_t) size );
  
  theArray->nb += 1;
  return( pos );
}

/*
  Add a new element to an array (same as before). This time, it reallocate
  the memory, when the array is full, with 1/2 more than already allocated. 
  Return the position in the array on success, return -1 on failure.
*/

#if defined(_ANSI_)
int T_ArrayAddNew( t_ArrayPtr theArray, t_ObjPtr theObj )
#else
int T_ArrayAddNew(theArray, theObj )
t_ArrayPtr theArray;
t_ObjPtr theObj ;

#endif

{
  int size,pos;
  char *auxPtr,*objPtr;
  
  E_ERR( theObj->type != theArray->elem.type, E_ERRMISMATCH, -1, 
	"T_ArrayAddNew",
	"object and Array elements has not the same type" );
  
  pos= theArray->nb;
  if ( pos >= theArray->nbAlloc ){
	if ( theArray->nbAlloc <= 1 )	/* start with 16 objects */	
 			T_ReAllocArrayTb( theArray, 16 );
	else	T_ReAllocArrayTb( theArray, (int) (1.5 * (float) theArray->nbAlloc) );
	E_ISERR( -1, "T_ArrayAddNew", NULL );
	}
  
  size= _T_GetSizeUtil( (int) theObj->type );
  
  auxPtr= theArray->tb + size * pos;
  objPtr= ((char *) theObj) + sizeof(t_Obj);
  memcpy( (void *) auxPtr, (void *) objPtr, (size_t) size );
  
  theArray->nb += 1;
  return( pos );
}

/*
  Test program for the arrays
  */

#if defined(_ANSI_)
void T_TestArray( void )
#else
void T_TestArray()

#endif

{
  int i,j;
  t_IntPtr theI=NULL;
  t_ArrayPtr array=NULL;
  t_ListPtr theList= NULL;
  t_ListPtr aList;
  
  T_AllocInt( &theI );
  
  for( j=0; j< 10; j++ ){
    
    T_AllocArray( &array );
    T_AllocArrayTb( array, OBJ_INT, 10 );
    E_REPERR( goto END_ArrayTest; );
    
    for( i=0; i< j; i++ ){
      theI->index= i;
      T_ArrayAdd( array, (t_ObjPtr) theI );
      E_REPERR( goto END_ArrayTest; );
    }
    
    T_AddObjList( (t_ObjPtr) array, &theList );
    array= NULL;
  }
  
  T_ReverseList( &theList );
  T_FWriteList( (FILE *) NULL, "tmp.array", theList );
  E_REPERR( goto END_ArrayTest; );
  
  T_FreeList( &theList );
  
  T_FReadList( (FILE *) NULL, "tmp.array", &theList );
  E_REPERR( goto END_ArrayTest; );
  
  aList= theList;
  while( aList != NULL ){	
    
    T_ListItemErr( aList, (t_ObjHand) &array, OBJ_ARRAY );
    E_REPERR( goto END_ArrayTest; );
    
    for( i=0; i< array->nb; i++ ){
      T_ArrayGet( array, (t_ObjPtr) theI, i );
      E_REPERR( {array= NULL; goto END_ArrayTest;} );
      printf("%d ", theI->index );
    }
    printf("\n");
    aList= aList->next;
  }
  
 END_ArrayTest:
  T_FreeInt( &theI );
  T_FreeArray( &array );
  T_FreeList( &theList );
  E_ISERRN( "T_ArrayTest", NULL);
}

/*----------------------------------------------------------------------*/
/************************************************************************/
/*
	OpersNPoint

  Functions on arrays, particular to N-Dimensional points.
  Use ->dim for the dimension of the vector and ->v for the
  vector values (in [0..dim-1])
  
  (c) Jean-Philippe THIRION -- INRIA --	August 25, 1992
  */
/************************************************************************/

/*
	Perform the initialization of all the functions on NPoints
*/

#if defined(_ANSI_)
void T_ObjNPointInit( void )
#else
void T_ObjNPointInit()

#endif

{
	gObjTb[ OBJ_NPOINT ][ OBJF_ALLOC ]= ( void_func_type ) T_AllocNPoint;
	gObjTb[ OBJ_NPOINT ][ OBJF_FREE ]= ( void_func_type ) T_FreeNPoint;
	gObjTb[ OBJ_NPOINT ][ OBJF_FREAD ]= ( void_func_type ) T_FReadNPoint;
	gObjTb[ OBJ_NPOINT ][ OBJF_FWRITE ]= ( void_func_type ) T_FWriteNPoint;
	gObjTb[ OBJ_NPOINT ][ OBJF_COPY ]= ( void_func_type ) T_CopyNPoint;
	gObjTb[ OBJ_NPOINT ][ OBJF_DEFAULT ]= ( void_func_type ) T_DefaultNPoint;
	}

/*
  Free a N-Dimensional point 
  */

#if defined(_ANSI_)
void T_FreeNPoint( t_NPointHand handPoint )
#else
void T_FreeNPoint(handPoint )
t_NPointHand handPoint ;

#endif

{ 
	T_Free( (void **) &(*handPoint)->v );
	T_Free( (void **) handPoint );
	}

/*
  Initialize an N-Point structure
  */

#if defined(_ANSI_)
void T_DefaultNPoint( t_NPointPtr point )
#else
void T_DefaultNPoint(point )
t_NPointPtr point ;

#endif

{ 
  point->obj.type= OBJ_NPOINT;
  point->dim= 0;
  point->v= NULL;
  point->reserved1.type= OBJ_FLOAT;
  point->reserved2= 0;
}

/*
  Allocate a N-Dimensional point
  */

#if defined(_ANSI_)
void T_AllocNPoint( t_NPointHand handPoint )
#else
void T_AllocNPoint(handPoint )
t_NPointHand handPoint ;

#endif

{ 
	t_NPointPtr aPt= NULL;

	aPt= (t_NPointPtr) T_Alloc( sizeof( t_NPoint ) );
	E_ISERRN( "T_AllocNPoint", NULL );

	T_DefaultNPoint( aPt );
	*handPoint= aPt;
	}

/*
  Initialize a N-dimensionnal point to a given dimension
  */

#if defined(_ANSI_)
int T_InitNPoint( t_NPointPtr point, int dim )
#else
int T_InitNPoint(point, dim )
t_NPointPtr point;
int dim ;

#endif

{
  T_DefaultNPoint( point );
  T_AllocNPointTb( point, dim );
  E_ISERR( FALSE,"T_InitNPoint", NULL );
  return( TRUE );
}

/*
  Allocate a vector of dimension N of floating points
  */

#if defined(_ANSI_)
void T_AllocNPointTb( t_NPointPtr point, int dim )
#else
void T_AllocNPointTb(point, dim )
t_NPointPtr point;
int dim ;

#endif

{
	
	point->v= (reeltype *) T_Alloc( sizeof(reeltype) * dim );
	E_ISERRN( "T_AllocNPointTb", NULL );

	point->dim= dim;
	point->reserved2= point->dim;
	}

/*
  Change the dimension of a N-Dimensional point
  */

#if defined(_ANSI_)
void T_ReAllocNPointTb( t_NPointPtr point, int dim )
#else
void T_ReAllocNPointTb(point, dim )
t_NPointPtr point;
int dim ;

#endif

{	
	if ( !dim )	T_Free( (void **) &point->v );
	else		T_ReAlloc( (char **) &point->v, sizeof( reeltype ) * dim );
	point->dim= dim;
  	point->reserved2= point->dim;
	}

/*
  Copy a NPoint into another one
  */

#if defined(_ANSI_)
void T_CopyNPoint( t_NPointPtr point, t_NPointHand hRes )
#else
void T_CopyNPoint(point, hRes )
t_NPointPtr point;
t_NPointHand hRes ;

#endif

{
	t_NPointPtr newP;

	if ( point == NULL ) { *hRes= NULL; return; }

	if ( *hRes == NULL ){
		T_AllocNPoint( hRes );
		E_ISERRN( "T_CopyNPoint", NULL );
		}

	newP= *hRes;
	T_Free( (void **) &newP->v );
	memcpy( (void *) newP, (void *) point, sizeof(t_NPoint) );
	newP->v= NULL;
	newP->dim= 0;
	newP->reserved2= 0;

	T_AllocNPointTb( newP, point->dim );
	
	E_REPERR( { T_Free( (void **) hRes ); } );
	E_ISERRN( "T_CopyNPoint", NULL );

	memcpy( (void *) newP->v, (void *) point->v, sizeof(reeltype)*point->dim );
	}

/*
  Test program for the N-Dimensional vectors 
  */

#if defined(_ANSI_)
void T_TestNPoint( void )
#else
void T_TestNPoint()

#endif

{
  int i,j;
  t_NPointPtr point=NULL;
  t_ListPtr theList= NULL;
  t_ListPtr aList;
  
  for( j=0; j< 10; j++ ){
    
    T_AllocNPoint( &point );		/* allocate NPoint structure		*/
    T_AllocNPointTb( point, j );	/* allocate a j dimension vector	*/
    E_REPERR( goto END_TestNPoint; );
    
    for( i=0; i< j; i++ ) point->v[i]= (float) i;	
    
    T_AddObjList( (t_ObjPtr) point, &theList );
    point= NULL;
  }
  
  T_ReverseList( &theList );
  T_FWriteList( (FILE *) NULL, "tmp.npoints", theList );
  E_REPERR( goto END_TestNPoint; );
  
  T_FreeList( &theList );
  
  T_FReadList( (FILE *) NULL, "tmp.npoints", &theList );
  E_REPERR( goto END_TestNPoint; );
  
  aList= theList;
  while( aList != NULL ){	
    
    T_ListItemErr( aList, (t_ObjHand) &point, OBJ_NPOINT );
    E_REPERR( goto END_TestNPoint; );
    
    for( i=0; i< point->dim; i++ ) printf("%f ", point->v[i] );
    
    printf("\n");
    aList= aList->next;
    point= NULL;
  }
  
 END_TestNPoint:
  T_FreeNPoint( &point );
  T_FreeList( &theList );
  E_ISERRN( "T_TestNPoint", NULL);
}

/*
  Convert a t_3Point into an N-Point. Allocate it if null pointer.
  */

#if defined(_ANSI_)
void T_3Pt2NPt( t_3PointPtr pt, t_NPointHand handNPt )
#else
void T_3Pt2NPt(pt, handNPt )
t_3PointPtr pt;
t_NPointHand handNPt ;

#endif

{
  t_NPointPtr nPt;
  
  if ( *handNPt == NULL ) T_AllocObj( OBJ_NPOINT, (t_ObjHand) handNPt );
  
  nPt= *handNPt;
  if(nPt->dim < 3) T_ReAllocNPointTb( nPt, 3 );
  E_ISERRN( "T_3Pt2NPt", NULL );
  
  EPV( nPt, 0 )= pt->x;
  EPV( nPt, 1 )= pt->y;
  EPV( nPt, 2 )= pt->z;
}

/*
  Convert a t_3Point into an N-Point
  */

#if defined(_ANSI_)
void T_NPt23Pt( t_NPointPtr nPt, t_3PointPtr pt )
#else
void T_NPt23Pt(nPt, pt )
t_NPointPtr nPt;
t_3PointPtr pt ;

#endif

{
  E_ERRN( nPt->dim < 3, E_ERRFATAL, "T_NPt23Pt",
	 "incompatible sizes" );
  
  pt->x= EPV( nPt, 0 );
  pt->y= EPV( nPt, 1 );
  pt->z= EPV( nPt, 2 );
}

/*
  Write a N-dimensional vector in an ascii file
*/

#if defined(_ANSI_)
int T_FWriteNPoint( FILE *fd, char *name, t_NPointPtr pt )
#else
int T_FWriteNPoint(fd, name, pt )
FILE *fd;
char *name;
t_NPointPtr pt;

#endif

{
	int create= FALSE;
	int status;
	int i;
	reeltype *aux;

	if ( fd == NULL ){
		create= TRUE;
		fd= fopen( name, "w" );
		E_ERR( fd == NULL, E_ERRC, FALSE, "T_FWriteNPoint", name );
		status= T_WriteObjectType( fd, name, (int) pt->obj.type );
		E_REPERR( goto END_FWriteNPoint );
		}
	else create= FALSE;

	status= fprintf( fd, " " );
	if ( status == EOF ){ 
		E_SETERR( E_ERRW, NULL, NULL );
		goto END_FWriteNPoint;
		}

	status= T_WriteObjectType( fd, name, OBJ_FLOAT );
	E_REPERR( goto END_FWriteNPoint );

	status= fprintf( fd, " %d\n", pt->dim );
	if ( status == EOF ){
		E_SETERR( E_ERRW, NULL, NULL );
		goto END_FWriteNPoint;
		}

	status= T_WriteObjectType( fd, name, OBJ_BEGINARRAY );
	E_REPERR( goto END_FWriteNPoint; );

	fprintf( fd, "\n" );

	aux= pt->v;
	for( i=0; i< pt->dim; i++ ){
		status= fprintf( fd, " %f\n", *aux++ );
		if ( status == EOF ){
			E_SETERR( E_ERRW, NULL, NULL );
			goto END_FWriteNPoint;
			}
		}
		
END_FWriteNPoint:

	T_WriteObjectType( fd, name, OBJ_ENDARRAY );
	
	if ( create ) fclose( fd );

	E_ISERR( FALSE, "T_FWriteNPoint", NULL );
	return( TRUE );
	}

/*		
  Read a N-Dimensional vector from an ascii file
*/

#if defined(_ANSI_)
int T_FReadNPoint( FILE *fd, char *name, t_NPointHand hPt )
#else
int T_FReadNPoint(fd, name, hPt )
FILE *fd;
char *name;
t_NPointHand hPt;

#endif

{
	int status,opening,type,nb,i;
	reeltype *aux;
	t_NPointPtr pt;
	float tmp;
  
  /* test, allocation, opening */
  
	if ( *hPt == NULL ) T_AllocNPoint( hPt );
	pt= *hPt;
	E_ISERR( FALSE, "T_FReadNPoint", name );
  
	if ( fd == NULL ){
		opening= TRUE;
		fd= fopen( name, "r" );
		if ( fd == NULL ){
			T_FreeNPoint( hPt );
			E_XERR( E_ERRO, FALSE, "T_FReadNPoint", name );
   		 	}
    
    	type= T_ReadObjectType( fd, name );
    	if ( type != OBJ_NPOINT ){
   		  	T_FreeNPoint( hPt );
			fclose(fd);
			E_XERR(E_ERRMISMATCH,FALSE,"T_FReadNPoint","not the expected type");
			}
		}
	else opening= FALSE;
  
	/* processing... */
  
	type= T_ReadObjectType( fd, name );
	if ( !T_ObjType(type) ){
		E_SETERR(E_ERRUNKNOWN,"T_FReadNPoint","unknown sub-objects type");
		goto END_FReadNPoint;
	}
  
	pt->reserved1.type= type;
  
	status= fscanf(fd,"%d", &nb );
	if (status == EOF){ 
		E_SETERR(E_ERRR,"reading","total size"); 
		goto END_FReadNPoint;
		}
  
	type= T_ReadObjectType( fd, name );
	if (type != OBJ_BEGINARRAY) 
		E_SETERR(E_ERRMISMATCH,"T_FReadNPoint","array delimiter expected"); 
  
	E_REPERR( goto END_FReadNPoint; );
  
	T_AllocNPointTb( pt, nb );
 
	aux= pt->v; 
  
	for( i=0; i< nb; i++ ){

		status= fscanf( fd, "%f", &tmp );
		if ( status == EOF ){
		  E_SETERR(E_ERRR,"reading",NULL); 
		  goto END_FReadNPoint;
		}
		*aux = (reeltype) tmp;
		aux++;
	}
  
	type= T_ReadObjectType( fd, name );
	if (type != OBJ_ENDARRAY) 
		E_SETERR(E_ERRMISMATCH,"T_FReadNPoint","end array delimiter expected"); 
  
END_FReadNPoint:
  
	if ( opening ) fclose( fd );
	E_REPERR( T_FreeNPoint(hPt) );
	E_ISERR( FALSE, "T_FReadNPoint", NULL );
	return( TRUE );
	}

/*----------------------------------------------------------------------*/
/************************************************************************/
/*
	OpersMat4

  General operations on 4x4 matrix for homogeneous coordinates
  transforms.
  
  (c) Jean-Philippe THIRION -- INRIA --	August 25, 1992
  */
/************************************************************************/

/*		Perform the initialization of all the functions on 4x4 Matrix	*/

#if defined(_ANSI_)
void T_ObjMat4Init( void )
#else
void T_ObjMat4Init()

#endif

{
  
  gObjTb[ OBJ_MAT4 ][ OBJF_ALLOC ]= ( void_func_type ) T_AllocMat4;
  gObjTb[ OBJ_MAT4 ][ OBJF_FREE ]= ( void_func_type ) T_FreeMat4;
  gObjTb[ OBJ_MAT4 ][ OBJF_FREAD ]= ( void_func_type ) T_FReadMat4;
  gObjTb[ OBJ_MAT4 ][ OBJF_FWRITE ]= ( void_func_type ) T_FWriteMat4;
  gObjTb[ OBJ_MAT4 ][ OBJF_COPY ]= ( void_func_type ) T_CopyMat4;
  gObjTb[ OBJ_MAT4 ][ OBJF_DEFAULT ]= ( void_func_type ) T_DefaultMat4;
}

/*		Free a Mat4form												*/

#if defined(_ANSI_)
void T_FreeMat4( t_Mat4Hand handMat4 )
#else
void T_FreeMat4(handMat4 )
t_Mat4Hand handMat4 ;

#endif

{
  T_Free( (void **) handMat4 );
}

/*
  Initialize a Mat4 structure with default values
  */

#if defined(_ANSI_)
int T_DefaultMat4( t_Mat4Ptr mat4 )
#else
int T_DefaultMat4(mat4 )
t_Mat4Ptr mat4 ;

#endif

{
  (void *) memset( (void *) mat4, 0, (size_t) sizeof(t_Mat4) );
  mat4->obj.type= OBJ_MAT4;
  return( TRUE );
}

/*		Allocate a Mat4form											*/

#if defined(_ANSI_)
int T_AllocMat4( t_Mat4Hand handMat4 )
#else
int T_AllocMat4(handMat4 )
t_Mat4Hand handMat4 ;

#endif

{
  *handMat4= (t_Mat4Ptr) T_Alloc( sizeof(t_Mat4) );
  E_ISERR( FALSE, "T_AllocMat4", NULL );
  T_DefaultMat4( *handMat4 );
  return( TRUE );
}

/*		Write a Mat4form in an ascii file								*/

#if defined(_ANSI_)
int T_FWriteMat4( FILE *fd, char *name, t_Mat4Ptr trans )
#else
int T_FWriteMat4(fd, name, trans )
FILE *fd;
char *name;
t_Mat4Ptr trans ;

#endif

{
  int status,i;
  
  if ( fd == NULL ) return( T_FWriteOne( fd, name, (t_ObjPtr) trans) );
  
  for( i=0; i<4; i++ ){
    status= fprintf(fd," %f %f %f %f\n", trans->mt[0][i], 
		    trans->mt[1][i], trans->mt[2][i], trans->mt[3][i] );
    E_ERR( status == EOF, E_ERRW, FALSE, "T_FWriteMat4", name);
  }
  
  return( TRUE );
}

/*		Read a Mat4 poMat4 in an ascii file								*/

#if defined(_ANSI_)
int T_FReadMat4( FILE *fd, char *name, t_Mat4Hand handMat4 )
#else
int T_FReadMat4(fd, name, handMat4 )
FILE *fd;
char *name;
t_Mat4Hand handMat4 ;

#endif

{
  int status,i;
  t_Mat4Ptr trans;
  float tmp0, tmp1, tmp2, tmp3;
  
  if ( *handMat4 == NULL ){
    T_AllocMat4( handMat4 );
    E_ISERR( FALSE, "T_FReadMat4", name );
  }
  
  if ( fd == NULL ) return( T_FReadOne( fd, name, (t_ObjHand) handMat4) );
  
  trans= *handMat4;
  
  for( i=0; i<4; i++ ){
    status= fscanf(fd," %f %f %f %f", &tmp0, &tmp1, &tmp2, &tmp3); 
    E_ERR( status == EOF, E_ERRR, FALSE, "T_FReadMat4", name);
    trans->mt[0][i] = (reeltype) tmp0;
    trans->mt[1][i] = (reeltype) tmp1;
    trans->mt[2][i] = (reeltype) tmp2;
    trans->mt[3][i] = (reeltype) tmp3;
  }
  
  return( TRUE );
}

/*		Copy a Mat4 trans into another one								*/

#if defined(_ANSI_)
int T_CopyMat4( t_Mat4Ptr theMat4, t_Mat4Hand handMat4 )
#else
int T_CopyMat4(theMat4, handMat4 )
t_Mat4Ptr theMat4;
t_Mat4Hand handMat4 ;

#endif

{
  if ( theMat4 == NULL ) { *handMat4= NULL; return( TRUE ); }
  E_ERR( (theMat4->obj.type) != OBJ_MAT4,
	E_ERRMISMATCH, FALSE, "T_CopyMat4", "not the expected type" );
  
  if ( *handMat4 == NULL ){
    T_AllocMat4( handMat4 );
    E_ISERR( FALSE, "T_FReadMat4", NULL );
  }
  
  memcpy( (void *) *handMat4, (void *) theMat4, (size_t) sizeof(t_Mat4) );
  return( TRUE );
}

/*----------------------------------------------------------------------*/
/************************************************************************/
/*
	OpersObjDef

  General operations on object definition, which are used to
create dynamically new object types.
  
  (c) Jean-Philippe THIRION -- INRIA --	October 12, 1992
*/
/************************************************************************/

/*
  Perform the initialization of all the functions
  */

#if defined(_ANSI_)
void T_ObjObjDefInit( void )
#else
void T_ObjObjDefInit()

#endif

{
  
  gObjTb[ OBJ_OBJDEF ][ OBJF_ALLOC ]= NULL;
  gObjTb[ OBJ_OBJDEF ][ OBJF_FREE ]= NULL;
  gObjTb[ OBJ_OBJDEF ][ OBJF_FREAD ]= NULL;
  gObjTb[ OBJ_OBJDEF ][ OBJF_FWRITE ]= NULL;
  gObjTb[ OBJ_OBJDEF ][ OBJF_COPY ]= NULL;
  gObjTb[ OBJ_OBJDEF ][ OBJF_DEFAULT ]= ( void_func_type ) T_DefaultObjDef;
}

/*
  default values for the object
  */

#if defined(_ANSI_)
void T_DefaultObjDef( t_ObjDefPtr objDef )
#else
void T_DefaultObjDef(objDef )
t_ObjDefPtr objDef ;

#endif

{
  int i;
  
  objDef->obj.type= OBJ_OBJDEF;
  objDef->init= TRUE;
  objDef->sizeUtil= 0;
  for( i=0; i< OBJF_NB; i++ ) objDef->f[i]= NULL;
}

/*
  Initialise the Object definition with default values.

Always to be used before any definition of a new object.
*/

#if defined(_ANSI_)
void T_InitObjDef( t_ObjDefPtr objDef, int sizeUtil )
#else
void T_InitObjDef(objDef, sizeUtil )
t_ObjDefPtr objDef;
int sizeUtil ;

#endif

{
  T_DefaultObjDef( objDef );
  objDef->sizeUtil= sizeUtil;
}

/*
  Test function for the dynamic object definition.

Redefinition of the integer type...
*/

#if defined(_ANSI_)
void T_TestObjDef( void )
#else
void T_TestObjDef()

#endif

{
  int i,j;
  t_MyOwnTypePtr theI=NULL;
  t_ArrayPtr array=NULL;
  t_ListPtr theList= NULL;
  t_ListPtr aList;
  t_ObjDef myObjDef;
  int myOwnType= OBJ_UNKNOWN;
  
  /*........ defines a new object type which is mine ....................*/
  
  T_InitObjDef( &myObjDef, sizeof(t_MyOwnType) - sizeof(t_Obj) );
  
  myObjDef.f[ OBJF_FWRITE ]= (void_func_type) T_FWriteInt;
  myObjDef.f[ OBJF_FREAD ]= (void_func_type) T_FReadInt;
  
  myOwnType= T_AddObjDef( &myObjDef );
  
  /*.....................................................................*/
  
  T_AllocObj( myOwnType, (t_ObjHand) &theI );
  
  for( j=0; j< 10; j++ ){
    
    T_AllocArray( &array );
    T_AllocArrayTb( array, myOwnType, 10 );
    E_REPERR( goto END_TestObjDef; );
    
    for( i=0; i< j; i++ ){
      theI->index= i;
      T_ArrayAdd( array, (t_ObjPtr) theI );
      E_REPERR( goto END_TestObjDef; );
    }
    
    T_AddObjList( (t_ObjPtr) array, &theList );
    array= NULL;
  }
  
  T_ReverseList( &theList );
  T_FWriteList( (FILE *) NULL, "tmp.objDef", theList );
  E_REPERR( goto END_TestObjDef; );
  
  T_FreeList( &theList );
  
  T_FReadList( (FILE *) NULL, "tmp.objDef", &theList );
  E_REPERR( goto END_TestObjDef; );
  
  aList= theList;
  while( aList != NULL ){	
    
    T_ListItemErr( aList, (t_ObjHand) &array, OBJ_ARRAY );
    E_REPERR( goto END_TestObjDef; );
    
    for( i=0; i< array->nb; i++ ){
      T_ArrayGet( array, (t_ObjPtr) theI, i );
      E_REPERR( {array= NULL; goto END_TestObjDef;} );
      printf("%d ", theI->index );
    }
    printf("\n");
    aList= aList->next;
  }
  
 END_TestObjDef:
  T_FreeObj( (t_ObjHand) &theI );
  T_RemoveObjDef( &myOwnType );	/* remove the user oriented type	*/
  T_FreeArray( &array );
  T_FreeList( &theList );
  E_ISERRN( "T_ArrayTest", NULL);
}

/*----------------------------------------------------------------------*/
/************************************************************************/
/*
	OpersString

  Operations on character strings
  
  (c) Jean-Philippe THIRION -- INRIA --	June 23, 1993
  */
/************************************************************************/

/*		Perform the initialization of all the functions on strings		*/

#if defined(_ANSI_)
void T_ObjStringInit( void )
#else
void T_ObjStringInit()

#endif

{
  
  gObjTb[ OBJ_STRING ][ OBJF_ALLOC ]= ( void_func_type ) T_AllocString;
  gObjTb[ OBJ_STRING ][ OBJF_FREE ]= ( void_func_type ) T_FreeString;
  gObjTb[ OBJ_STRING ][ OBJF_FREAD ]= ( void_func_type ) T_FReadString;
  gObjTb[ OBJ_STRING ][ OBJF_FWRITE ]= ( void_func_type ) T_FWriteString;
  gObjTb[ OBJ_STRING ][ OBJF_COPY ]= ( void_func_type ) T_CopyString;
  gObjTb[ OBJ_STRING ][ OBJF_DEFAULT ]= ( void_func_type ) T_DefaultString;
}

/*
  Free a string
  */

#if defined(_ANSI_)
void T_FreeString( t_StringHand handString )
#else
void T_FreeString(handString )
t_StringHand handString ;

#endif

{
  T_Free( (void **) &((*handString)->str) );
  T_Free( (void **) handString );
}

/*
  Initialize s string structure
  */

#if defined(_ANSI_)
int T_DefaultString( t_StringPtr theString )
#else
int T_DefaultString(theString )
t_StringPtr theString ;

#endif

{
  theString->str= NULL;
  theString->obj.type= OBJ_STRING;	
  return( TRUE );
}

/*
  Allocate a new string structure
  */

#if defined(_ANSI_)
int T_AllocString( t_StringHand handString )
#else
int T_AllocString(handString )
t_StringHand handString ;

#endif

{
  *handString= (t_StringPtr) T_Alloc( sizeof(t_String) );
  E_ISERR( FALSE, "T_AllocString", NULL );
  
  T_DefaultString( *handString );	
  return( TRUE );
}

/*
  Write a string in an ascii file
  */

#if defined(_ANSI_)
int T_FWriteString( FILE *fd, char *name, t_StringPtr string )
#else
int T_FWriteString(fd, name, string )
FILE *fd;
char *name;
t_StringPtr string ;

#endif

{
  int status;
  
  if ( string->str )	status= fprintf(fd," %s\n", string->str );
  else				status= fprintf(fd,"(void string)\n");
  E_ERR( status == EOF, E_ERRW, FALSE, "T_FWriteString", name);
  
  return( TRUE );
}

static char gAuxStr[4096];

/*		
  Read a string in an ascii file
*/

#if defined(_ANSI_)
int T_FReadString( FILE *fd, char *name, t_StringHand handString )
#else
int T_FReadString(fd, name, handString )
FILE *fd;
char *name;
t_StringHand handString ;

#endif

{
  int status;
  
  status= fscanf( fd, "%s", gAuxStr );
  E_ERR( status == EOF, E_ERRR, FALSE, "T_FReadString", name);
  
  if ( strcmp( gAuxStr, "(void string)" ) == 0 ){
    if ( *handString && (*handString)->str ){
      free( (void *) (*handString)->str );
      (*handString)->str= NULL;
    }
    return( TRUE );
  }
  
  T_BuildString( gAuxStr, handString );
  E_ISERR( FALSE, "T_FReadString", name );
  return( TRUE );
}

/*	
  Copy a string into another one
  */

#if defined(_ANSI_)
int T_CopyString( t_StringPtr string, t_StringHand handString )
#else
int T_CopyString(string, handString )
t_StringPtr string;
t_StringHand handString ;

#endif

{
  if ( string == NULL ) { *handString= NULL; return( TRUE ); }
  
  E_ERR( (string->obj.type) != OBJ_STRING,
	E_ERRMISMATCH, FALSE, "T_CopyString", "not the expected type" );
  
  T_BuildString( string->str, handString );
  E_ISERR( FALSE, "T_CopyString", string->str );
  return( TRUE );
}

/*
  Build an EpidaureLib object string from a C string
  */

#if defined(_ANSI_)
void T_BuildString( char *aStr, t_StringHand handString )
#else
void T_BuildString(aStr, handString )
char *aStr;
t_StringHand handString ;

#endif

{
  t_StringPtr string;
  int length;
  
  if ( *handString == NULL ){
    T_AllocString( handString );
    E_ISERRN( "T_BuildString", aStr );
  }
  else ( T_Free( (void **) &(*handString)->str ) );
  
  if ( aStr == NULL ) return;
  
  string= *handString;
  length= (int) strlen(aStr);
  string->str= (char *) T_Alloc(( unsigned ) ( length + 1 ) );
  E_ISERRN( "T_BuildString", aStr );
  
  strncpy( string->str, aStr, length );
  string->str[length]= 0;
}

/*----------------------------------------------------------------------*/
/************************************************************************/
/*
	OpersMat

  Operations on Matrices in any dimensions
  
  (c) Jean-Philippe THIRION -- INRIA --	June 25, 1993
  */
/************************************************************************/

/*		Perform the initialization of all the functions on strings		*/

#if defined(_ANSI_)
void T_ObjMatInit( void )
#else
void T_ObjMatInit()

#endif

{
  
  gObjTb[ OBJ_MAT ][ OBJF_ALLOC ]= ( void_func_type ) T_AllocMat;
  gObjTb[ OBJ_MAT ][ OBJF_FREE ]= ( void_func_type ) T_FreeMat;
  gObjTb[ OBJ_MAT ][ OBJF_FREAD ]= ( void_func_type ) T_FReadMat;
  gObjTb[ OBJ_MAT ][ OBJF_FWRITE ]= ( void_func_type ) T_FWriteMat;
  gObjTb[ OBJ_MAT ][ OBJF_COPY ]= ( void_func_type ) T_CopyMat;
  gObjTb[ OBJ_MAT ][ OBJF_DEFAULT ]= ( void_func_type ) T_DefaultMat;
}

/*
  Free an N-Dimensional Matrix 
  */

#if defined(_ANSI_)
void T_FreeMat( t_MatHand handMat )
#else
void T_FreeMat(handMat )
t_MatHand handMat ;

#endif

{
  if ( !*handMat ) return;
  T_Free( (void **) &((*handMat)->v) );
  T_Free( (void **) handMat );
}

/*
  Initialize the fields of a matrix structure
  */

#if defined(_ANSI_)
void T_DefaultMat( t_MatPtr mat )
#else
void T_DefaultMat(mat )
t_MatPtr mat ;

#endif

{
  mat->nC= 0;
  mat->nL= 0;
  mat->v= NULL;
  mat->obj.type= OBJ_MAT;
}

/*
  Allocate a new matrix structure
  */

#if defined(_ANSI_)
int T_AllocMat( t_MatHand handMat )
#else
int T_AllocMat(handMat )
t_MatHand handMat ;

#endif

{
  *handMat= (t_MatPtr) T_Alloc( sizeof(t_Mat) );
  E_ISERR( FALSE, "T_AllocMat", NULL );
  
  T_DefaultMat( *handMat );	
  return( TRUE );
}

/*
  Initialize a matrix to a given size
  */

#if defined(_ANSI_)
int T_InitMat( t_MatPtr mat, int nLine, int nCol )
#else
int T_InitMat(mat, nLine, nCol )
t_MatPtr mat;
int nLine;
int nCol ;

#endif

{
  T_DefaultMat( mat );
  T_AllocMatTb( mat, nLine, nCol );
  E_ISERR( FALSE, "T_InitMat", NULL );
  return( TRUE );
}

/*
  Allocate the Matrix array for a given number of lines and columns
  */

#if defined(_ANSI_)
void T_AllocMatTb( t_MatPtr theMat, int nLine, int nCol )
#else
void T_AllocMatTb(theMat, nLine, nCol )
t_MatPtr theMat;
int nLine;
int nCol ;

#endif

{
  if ( (nCol == theMat->nC) &&
      (nLine == theMat->nL) ) return;	/* as already the good size */
  
  T_Free( (void **) &theMat->v );			/* allocate a new buffer	*/
  
  if ( !nLine || !nCol ){					/* desallocate all and exit */
    theMat->nC= nCol;
    theMat->nL= nLine;
    return;
  }
  
  theMat->v= (reeltype *) T_Alloc( nCol * nLine * sizeof(reeltype) );
  E_ISERRN( "T_AllocMatTb", NULL );
  
  theMat->nC= nCol;
  theMat->nL= nLine;
}

/*
  Write a N-dimensional matrix in an ascii file
  */

#if defined(_ANSI_)
int T_FWriteMat( FILE *fd, char *name, t_MatPtr mat )
#else
int T_FWriteMat(fd, name, mat )
FILE *fd;
char *name;
t_MatPtr mat ;

#endif

{
  int status;
  int l,c;
  
  status= fprintf( fd, " %d %d\n", mat->nL, mat->nC );
  E_ERR( status == EOF, E_ERRW, FALSE, "T_FWriteMat", name);
  
  for( l=0; l< mat->nL; l++ ){
    for( c=0; c< mat->nC; c++ ){
      status= fprintf(fd, " %f", EMV(mat,l,c) );
      E_ERR( status == EOF, E_ERRW, FALSE, "T_FWriteMat", name);
      if ( mat->nC > 8 ) fprintf( fd, "\n");
    }
    fprintf( fd, "\n");
  }
  
  return( TRUE );
}

/*		
  Read a N-Dimensional matrix in an ascii file
  */

#if defined(_ANSI_)
int T_FReadMat( FILE *fd, char *name, t_MatHand handMat )
#else
int T_FReadMat(fd, name, handMat )
FILE *fd;
char *name;
t_MatHand handMat ;

#endif

{
  int status;
  t_MatPtr mat;
  int nLine,nCol;
  int l,c;
  float tmp;
  
  if ( *handMat == NULL ){
    T_AllocPoint( (t_PointHand) handMat );
    E_ISERR( FALSE, "T_FReadMat", name );
  }
  mat= *handMat;
  
  status= fscanf( fd, "%d %d", &nLine, &nCol );
  E_ERR( status == EOF, E_ERRR, FALSE, "T_FReadMat", name);
  
  T_AllocMatTb( mat, nLine, nCol );
  
  for( l=0; l< mat->nL; l++ )
    for( c=0; c< mat->nC; c++ ){
      status= fscanf( fd, "%f", &tmp );
      E_ERR( status == EOF, E_ERRR, FALSE, "T_FReadMat", name);
      EMV(mat,l,c) = tmp;
    }
  
  E_ISERR( FALSE, "T_FReadMat", name );
  return( TRUE );
}

/*	
  Copy a matrix into another one
  */

#if defined(_ANSI_)
int T_CopyMat( t_MatPtr mat, t_MatHand handMat )
#else
int T_CopyMat(mat, handMat )
t_MatPtr mat;
t_MatHand handMat ;

#endif

{
  t_MatPtr aMat;
  
  if ( mat == NULL ) { *handMat= NULL; return( TRUE ); }
  
  E_ERR( (mat->obj.type) != OBJ_MAT,
	E_ERRMISMATCH, FALSE, "T_CopyMat", "not the expected type" );
  
  if ( *handMat == NULL ) T_AllocMat( handMat );
  E_ISERR( FALSE, "T_CopyMat", NULL );
  
  aMat= *handMat;
  T_AllocMatTb( aMat, mat->nL, mat->nC );
  E_ISERR( FALSE, "T_CopyMat", NULL );
  
  memcpy( (void *) aMat->v, (void *) mat->v, 
	 (size_t) (mat->nL * mat->nC * sizeof(float)));
  
  E_ISERR( FALSE, "T_CopyMat", NULL );
  return( TRUE );
}

/*
  Transforms a ND point into a l x c matrix, line by line
  
  truncate or complete with 0.0 when the sizes of the matrix and
  the point are differents
  */

#if defined(_ANSI_)
void T_NPoint2Mat( t_NPointPtr point, t_MatHand hmat, int l, int c )
#else
void T_NPoint2Mat(point, hmat, l, c )
t_NPointPtr point;
t_MatHand hmat;
int l;
int c ;

#endif

{
  t_MatPtr mat;
  int dMax;
  int i=0;
  
  if ( *hmat == NULL ) T_AllocObj( OBJ_MAT, (t_ObjHand) hmat );
  mat= *hmat;
  
  T_AllocMatTb( mat, l, c );
  E_ISERRN( "T_NPoint2Mat", NULL );
  
  dMax= point->dim;
  
  for( l=0; l< mat->nL; l++ )
    for( c=0; c< mat->nC; c++ )
      if ( i < dMax ) EMV(mat,l,c)= EPV(point,i++);
      else			EMV(mat,l,c)= 0.0;
} 

/*
  Transforms a l x c matrix into a l x c vector, line by line 
  */

#if defined(_ANSI_)
void T_Mat2NPoint( t_MatPtr mat, t_NPointHand hpoint )
#else
void T_Mat2NPoint(mat, hpoint )
t_MatPtr mat;
t_NPointHand hpoint ;

#endif

{
  t_NPointPtr point;
  int i,l,c;
  
  if ( *hpoint == NULL ) T_AllocObj( OBJ_NPOINT, (t_ObjHand) hpoint );
  point= *hpoint;
  
  T_ReAllocNPointTb( point, mat->nC * mat->nL );
  E_ISERRN( "T_Mat2NPoint", NULL );
  
  
  i= 0;
  
  for( l=0; l< mat->nL; l++ )
    for( c=0; c< mat->nC; c++ )
      EPV(point,i++)= EMV(mat,l,c);
}

/*
  Convert a general 4x4 matrix into a transform matrix (t_Mat4)
  */

#if defined(_ANSI_)
void T_Mat2Mat4( t_MatPtr mat, t_Mat4Hand hmat4 )
#else
void T_Mat2Mat4(mat, hmat4 )
t_MatPtr mat;
t_Mat4Hand hmat4 ;

#endif

{
  t_Mat4Ptr mat4;
  int l,c;
  
  E_ERRN( (mat->nL != 4) || (mat->nC != 4),
	 E_ERRFATAL, "T_Mat2Mat4", "incompatible sizes" );
  
  if ( *hmat4 == NULL ) T_AllocObj( OBJ_MAT4, (t_ObjHand) hmat4 );
  mat4= *hmat4;
  
  for( l=0; l< mat->nL; l++ )
    for( c=0; c< mat->nC; c++ )
      mat4->mt[c][l]= EMV(mat,l,c);
}

/*
  convert a transform matrix (t_Mat4) into a general 4x4 matrix
  */

#if defined(_ANSI_)
void T_Mat42Mat( t_Mat4Ptr mat4, t_MatHand hMat )
#else
void T_Mat42Mat(mat4, hMat )
t_Mat4Ptr mat4;
t_MatHand hMat ;

#endif

{
  t_MatPtr mat;
  int l,c;
  
  if ( *hMat == NULL ) T_AllocObj( OBJ_MAT, (t_ObjHand) hMat );
  mat= *hMat;
  
  T_AllocMatTb( mat, 4, 4 );
  E_ISERRN( "T_Mat42Mat", NULL );
  
  for( l=0; l< mat->nL; l++ )
    for( c=0; c< mat->nC; c++ )
      EMV(mat,l,c)= mat4->mt[c][l];
}
	
/*----------------------------------------------------------------------*/
