/* epi_quater.c 2.3 5/14/98 */

#include <epidaure.h>
#include <epidaure.ee>

#ifndef NO_PROTO
static int E_EvalScalRotTrans(t_ListPtr list1,
                              t_ListPtr list2,
                              int nb,
                              double rot[3],
                              double b1[3],
                              double b2[3],
                              double *scale);
static int E_EvalRotTrans( t_ListPtr list1, t_ListPtr list2, int nb, double rot[3], double b1[3], double b2[3]);
#else 
static int E_EvalScalRotTrans();
static int E_EvalRotTrans();
#endif

/* Calcul d'une transformation rigide (aux moindres carres) entre deux listes de points.

   Etant donnee une liste de nb points apparies (donnee sous la forme de deux
   listes),la fonction evalue la matrice de transformation en coordonnees
   homogenes (4x4 : rotation et translation) pour passer de la liste list1 a
   la liste list2 en utilisant les quaternions.

RETURN 
   FALSE si il est impossible de calculer ladite transformation.
*/

#if defined(_ANSI_)
int E_EvalTransform( t_ListPtr list1 /* first list of points */, 
		    t_ListPtr list2  /* second list of points */, 
		    int nb           /* number of points */, 
		    t_Mat4Ptr mat4   /* result transformation matrix */ )
#else
int E_EvalTransform(list1, list2, nb, mat4 )
t_ListPtr list1; /* first list of points */
t_ListPtr list2; /* second list of points */
int nb;          /* number of points */
t_Mat4Ptr mat4;  /* result transformation matrix */
#endif
{
  double rot[4], b1[3], b2[3];
  double rog[3], mr[9];
  
  /*   on evalue la rotation (ramenee aux barycentres) entre les deux listes de points   */
  /*   ainsi que les barycentres                                                         */
  /***************************************************************************************/
  if ( !E_EvalRotTrans( list1, list2, nb, rot, b1, b2 ) ) {
    mat4->mt[0][0] = mat4->mt[1][0] = 0.0;
    mat4->mt[2][0] = mat4->mt[3][0] = 0.0;
    mat4->mt[0][1] = mat4->mt[1][1] = 0.0;
    mat4->mt[2][1] = mat4->mt[3][1] = 0.0;
    mat4->mt[0][2] = mat4->mt[1][2] = 0.0;
    mat4->mt[2][2] = mat4->mt[3][2] = 0.0;
    mat4->mt[0][3] = mat4->mt[1][3] = 0.0;
    mat4->mt[2][3] = mat4->mt[3][3] = 0.0;
    return( FALSE );
  }
  E_ISERR(FALSE,"E_EvalTransform",(char*)NULL);
  
  /*   calcul de la matrice de transformation   */
  /**********************************************/
  E_QMatRotFromQuat( rot, mr );
  E_DMMatMulVect(mr,b1,rog,3);
  
  mat4->mt[0][0] = (reeltype)( mr[0] );
  mat4->mt[1][0] = (reeltype)( mr[1] );
  mat4->mt[2][0] = (reeltype)( mr[2] );
  mat4->mt[3][0] = (reeltype)( b2[0] - rog[0] );
  mat4->mt[0][1] = (reeltype)( mr[3] );
  mat4->mt[1][1] = (reeltype)( mr[4] );
  mat4->mt[2][1] = (reeltype)( mr[5] );
  mat4->mt[3][1] = (reeltype)( b2[1] - rog[1] );
  mat4->mt[0][2] = (reeltype)( mr[6] );
  mat4->mt[1][2] = (reeltype)( mr[7] );
  mat4->mt[2][2] = (reeltype)( mr[8] );
  mat4->mt[3][2] = (reeltype)( b2[2] - rog[2] );
  mat4->mt[0][3] = (reeltype)( 0.0 );
  mat4->mt[1][3] = (reeltype)( 0.0 );
  mat4->mt[2][3] = (reeltype)( 0.0 );
  mat4->mt[3][3] = (reeltype)( 1.0 );
  
  return( TRUE );
}

#if defined(_ANSI_)
static int E_EvalRotTrans( t_ListPtr list1, t_ListPtr list2, int nb, double rot[3], double b1[3], double b2[3])
#else
static int E_EvalRotTrans(list1, list2, nb, rot, b1, b2)
t_ListPtr list1;
t_ListPtr list2;
int nb;
double rot[3];
double b1[3];
double b2[3];
#endif
{
  register int i, j;
  double a[16], aT[16], aTa[16], som[16], r[4];
  register double x1, y1, z1, x2, y2, z2;
  t_ListPtr aux1, aux2;
  t_PointPtr thePt1, thePt2;
  
  /*   evaluation des barycentres                   */
  /*   et de la translation entre les barycentres   */
  /**************************************************/
  b1[0] = b1[1] = b1[2] = 0.0;
  b2[0] = b2[1] = b2[2] = 0.0;
  aux1 = list1;
  aux2 = list2;
  for (i=0;i<nb;i++) {
    T_ListItemErr( aux1, (t_ObjHand) &thePt1, OBJ_UNKNOWN );
    T_ListItemErr( aux2, (t_ObjHand) &thePt2, OBJ_UNKNOWN );
    b1[0] += thePt1->pt.x;
    b1[1] += thePt1->pt.y;
    b1[2] += thePt1->pt.z;
    b2[0] += thePt2->pt.x;
    b2[1] += thePt2->pt.y;
    b2[2] += thePt2->pt.z;
    aux1 = aux1->next;
    aux2 = aux2->next;
  }
  for (i=0;i<3;i++) {
    b1[i] /= nb;
    b2[i] /= nb;
  }
  /*   evaluation de la rotation   */
  /*********************************/
  for (i=0;i<16;i++)
    som[i] = 0.0;
  aux1 = list1;
  aux2 = list2;
  for (i=0;i<nb;i++) {
    T_ListItemErr( aux1, (t_ObjHand) &thePt1, OBJ_UNKNOWN );
    T_ListItemErr( aux2, (t_ObjHand) &thePt2, OBJ_UNKNOWN );
    x1 = thePt1->pt.x - b1[0];
    y1 = thePt1->pt.y - b1[1];
    z1 = thePt1->pt.z - b1[2];
    x2 = thePt2->pt.x - b2[0];
    y2 = thePt2->pt.y - b2[1];
    z2 = thePt2->pt.z - b2[2];
    aux1 = aux1->next;
    aux2 = aux2->next;
    a[0] = a[5] = a[10] = a[15] = 0.0;
    a[1] = x1 - x2; 
    a[2] = y1 - y2;
    a[3] = z1 - z2;
    a[6] = -(z1 + z2);
    a[7] = y1 + y2;
    a[11] = -(x1 + x2);
    a[4] = -a[1];
    a[8] = -a[2];
    a[9] = -a[6];
    a[12] = -a[3];
    a[13] = -a[7];
    a[14] = -a[11];
    E_DMMatTrans(a,aT,4);
    E_DMMatMul(aT,a,aTa,4);
    for (j=0;j<16;j++)
      som[j] += aTa[j];
  }
  /*   calcul du vecteur propre associe a la plus petite valeur propre   */
  /***********************************************************************/
  if ( !E_DMVVpropresMatSym( som, r ,a ,4) )
    return( FALSE );
  E_ISERR(FALSE,"E_EvalRotTrans","Evaluation des valeurs et vecteurs propres");
  
  /*   l'erreur sur la rotation est dans r[0] (valeur propre)   */
  /**************************************************************/
  for (i=0;i<4;i++)
    rot[i] = a[i*4];
  
  return( TRUE );
}


/* Calcul d'une transformation rigide + scaling uniforme
   aux moindres carres entre deux listes de points.

   Etant donnee une liste de nb points apparies (donnee sous la forme
   de deux listes), la fonction evalue la matrice de transformation en
   coordonnees homogenes (4x4 : rotation*scaling et translation) pour
   passer de la liste list1 a la liste list2 en utilisant les
   quaternions.

RETURN 
   FALSE s'il est impossible de calculer ladite transformation.
*/

#if defined(_ANSI_)

int E_EvalScalTransform( t_ListPtr list1  /* first list of points */, 
                         t_ListPtr list2  /* second list of points */, 
                         int nb           /* number of points */, 
                         t_Mat4Ptr mat4   /* result transformation matrix */ )

#else


int E_EvalScalTransform(list1, list2, nb, mat4 )

t_ListPtr list1; /* first list of points */
t_ListPtr list2; /* second list of points */
int nb;          /* number of points */
t_Mat4Ptr mat4;  /* result transformation matrix */

#endif
{
  double rot[4], b1[3], b2[3];
  double rog[3], mr[9],scale;
  int i;
  
  /* on evalue la rotation (ramenee aux barycentres) entre les deux  */
  /* listes de points ainsi que les barycentres                      */
  /*******************************************************************/
  if ( !E_EvalScalRotTrans( list1, list2, nb, rot, b1, b2 , &scale) ) {
    mat4->mt[0][0] = mat4->mt[1][0] = 0.0;
    mat4->mt[2][0] = mat4->mt[3][0] = 0.0;
    mat4->mt[0][1] = mat4->mt[1][1] = 0.0;
    mat4->mt[2][1] = mat4->mt[3][1] = 0.0;
    mat4->mt[0][2] = mat4->mt[1][2] = 0.0;
    mat4->mt[2][2] = mat4->mt[3][2] = 0.0;
    mat4->mt[0][3] = mat4->mt[1][3] = 0.0;
    mat4->mt[2][3] = mat4->mt[3][3] = 0.0;
    return( FALSE );
  }
  E_ISERR(FALSE,"E_EvalScalTransform",(char*)NULL);
  
  /*   calcul de la matrice de transformation   */
  /**********************************************/
  E_QMatRotFromQuat( rot, mr );
  for(i = 0 ; i<9;i++)
      mr[i] *= scale;
  E_DMMatMulVect(mr,b1,rog,3);
  
  mat4->mt[0][0] = (reeltype)( mr[0] );
  mat4->mt[1][0] = (reeltype)( mr[1] );
  mat4->mt[2][0] = (reeltype)( mr[2] );
  mat4->mt[3][0] = (reeltype)( b2[0] - rog[0] );
  mat4->mt[0][1] = (reeltype)( mr[3] );
  mat4->mt[1][1] = (reeltype)( mr[4] );
  mat4->mt[2][1] = (reeltype)( mr[5] );
  mat4->mt[3][1] = (reeltype)( b2[1] - rog[1] );
  mat4->mt[0][2] = (reeltype)( mr[6] );
  mat4->mt[1][2] = (reeltype)( mr[7] );
  mat4->mt[2][2] = (reeltype)( mr[8] );
  mat4->mt[3][2] = (reeltype)( b2[2] - rog[2] );
  mat4->mt[0][3] = (reeltype)( 0.0 );
  mat4->mt[1][3] = (reeltype)( 0.0 );
  mat4->mt[2][3] = (reeltype)( 0.0 );
  mat4->mt[3][3] = (reeltype)( 1.0 );
  
  return( TRUE );
}



#if defined(_ANSI_)

static int E_EvalScalRotTrans( t_ListPtr list1,
                               t_ListPtr list2,
                               int nb,
                               double rot[3],
                               double b1[3],
                               double b2[3],
                               double *scale)

#else
static int E_EvalScalRotTrans(list1, list2, nb, rot, b1, b2, scale)

t_ListPtr list1;
t_ListPtr list2;
int nb;
double rot[3];
double b1[3];
double b2[3];
double *scale;

#endif
{
  register int i, j;
  double a[16], aT[16], aTa[16], som[16], r[4];
  register double x1, y1, z1, x2, y2, z2, sig_1, sig_2;
  t_ListPtr aux1, aux2;
  t_PointPtr thePt1, thePt2;
  
  /*   evaluation des barycentres                   */
  /*   et de la translation entre les barycentres   */
  /**************************************************/
  b1[0] = b1[1] = b1[2] = 0.0;
  b2[0] = b2[1] = b2[2] = 0.0;
  aux1 = list1;
  aux2 = list2;
  for (i=0;i<nb;i++) {
    T_ListItemErr( aux1, (t_ObjHand) &thePt1, OBJ_UNKNOWN );
    T_ListItemErr( aux2, (t_ObjHand) &thePt2, OBJ_UNKNOWN );
    b1[0] += thePt1->pt.x;
    b1[1] += thePt1->pt.y;
    b1[2] += thePt1->pt.z;
    b2[0] += thePt2->pt.x;
    b2[1] += thePt2->pt.y;
    b2[2] += thePt2->pt.z;
    aux1 = aux1->next;
    aux2 = aux2->next;
  }
  for (i=0;i<3;i++) {
    b1[i] /= nb;
    b2[i] /= nb;
  }
  /*   evaluation de la rotation   */
  /*********************************/
  for (i=0;i<16;i++)
    som[i] = 0.0;
  aux1 = list1;
  aux2 = list2;
  sig_1 = sig_2 = 0;
  for (i=0;i<nb;i++) {
    T_ListItemErr( aux1, (t_ObjHand) &thePt1, OBJ_UNKNOWN );
    T_ListItemErr( aux2, (t_ObjHand) &thePt2, OBJ_UNKNOWN );
    x1 = thePt1->pt.x - b1[0];
    y1 = thePt1->pt.y - b1[1];
    z1 = thePt1->pt.z - b1[2];
    sig_1 += x1*x1 + y1*y1 + z1*z1;
    x2 = thePt2->pt.x - b2[0];
    y2 = thePt2->pt.y - b2[1];
    z2 = thePt2->pt.z - b2[2];
    sig_2 += x2*x2 + y2*y2 + z2*z2;
    aux1 = aux1->next;
    aux2 = aux2->next;
    a[0] = a[5] = a[10] = a[15] = 0.0;
    a[1] = x1 - x2; 
    a[2] = y1 - y2;
    a[3] = z1 - z2;
    a[6] = -(z1 + z2);
    a[7] = y1 + y2;
    a[11] = -(x1 + x2);
    a[4] = -a[1];
    a[8] = -a[2];
    a[9] = -a[6];
    a[12] = -a[3];
    a[13] = -a[7];
    a[14] = -a[11];
    E_DMMatTrans(a,aT,4);
    E_DMMatMul(aT,a,aTa,4);
    for (j=0;j<16;j++)
      som[j] += aTa[j];
  }
  /*   calcul du vecteur propre associe a la plus petite valeur propre   */
  /***********************************************************************/
  if ( !E_DMVVpropresMatSym( som, r ,a ,4) )
    return( FALSE );
  E_ISERR(FALSE,"E_EvalScalRotTrans","Evaluation des valeurs et vecteurs propres");
  
  /*   l'erreur sur la rotation est dans r[0] (valeur propre)   */
  /**************************************************************/
  for (i=0;i<4;i++)
    rot[i] = a[i*4];

  /*  calcul du scaling */
  *scale = ( sig_1 + sig_2 - r[0])/(2.0 * sig_1);
  
  return( TRUE );
}



/* Calcul d'une matrice de rotation a partir d'un quaternion.

   Le titre dit tout.

BUGS
   Aucune verification sur la validite du quaternion.
*/

#if defined(_ANSI_)
void E_QMatRotFromQuat( double qr[4] /* input quaternion */, 
		       double rot[9] /* result rotation matrix */ )
#else
void E_QMatRotFromQuat(qr, rot)
double qr[4]; /* input quaternion */
double rot[9]; /* result rotation matrix */
#endif
{
  /*   calcul de la matrice de rotation a partir du quaternion   */
  /***************************************************************/
  double norme;
  int i;
  
  norme = qr[3]*qr[3];
  norme += qr[2]*qr[2];
  norme += qr[1]*qr[1];
  norme += qr[0]*qr[0];
  
  rot[0] = qr[0]*qr[0] + qr[1]*qr[1] - qr[2]*qr[2] - qr[3]*qr[3];
  rot[4] = qr[0]*qr[0] - qr[1]*qr[1] + qr[2]*qr[2] - qr[3]*qr[3];
  rot[8] = qr[0]*qr[0] - qr[1]*qr[1] - qr[2]*qr[2] + qr[3]*qr[3];
  rot[1] = (qr[1]*qr[2] - qr[0]*qr[3]) * 2.0;
  rot[2] = (qr[1]*qr[3] + qr[0]*qr[2]) * 2.0;
  rot[3] = (qr[1]*qr[2] + qr[0]*qr[3]) * 2.0;
  rot[5] = (qr[2]*qr[3] - qr[0]*qr[1]) * 2.0;
  rot[6] = (qr[1]*qr[3] - qr[0]*qr[2]) * 2.0;
  rot[7] = (qr[2]*qr[3] + qr[0]*qr[1]) * 2.0;
  
  for (i=0;i<9;i++)
    rot[i] /= norme;
}


