/* epi_recfilt.c 2.3 5/14/98 */

#include <epidaure.h>
#include <epidaure.ee>
#include <epidaure.ei>

#ifndef NO_PROTO
static int E_RecFilalongX( t_ImagePtr theIm, t_ImagePtr resIm, e_ParamRecFilPtr par );
static int E_RecFilalongY( t_ImagePtr theIm, t_ImagePtr resIm, e_ParamRecFilPtr par );
static int E_RecFilalongZ( t_ImagePtr theIm, t_ImagePtr resIm, e_ParamRecFilPtr par );
static void E_RecFil0Deriche( float *i, float *o, float *w1, float *w2, e_CoeffRecFilPtr par, int dim );
static void E_RecFil1Deriche( float *i, float *o, float *w1, float *w2, e_CoeffRecFilPtr par, int dim );
static void E_RecFil2Deriche( float *i, float *o, float *w1, float *w2, e_CoeffRecFilPtr par, int dim );
static void E_RecFil3Deriche( float *i, float *o, float *w1, float *w2, e_CoeffRecFilPtr par, int dim );
static void E_RecFilGenerique4( float *i, float *o, float *w1, float *w2, e_CoeffRecFilPtr par, int dim );
static void E_Coeff0Deriche( e_CoeffRecFilPtr par, float alpha );
static void E_Coeff1Deriche( e_CoeffRecFilPtr par, float alpha );
static void E_Coeff2Deriche( e_CoeffRecFilPtr par, float alpha );
static void E_Coeff3Deriche( e_CoeffRecFilPtr par, float alpha );
static void E_RecGaussCoeff( int der, e_CoeffRecFilPtr coeff, float sigma, int mode );
#else 
static int E_RecFilalongX();
static int E_RecFilalongY();
static int E_RecFilalongZ();
static void E_RecFil0Deriche();
static void E_RecFil1Deriche();
static void E_RecFil2Deriche();
static void E_RecFil3Deriche();
static void E_RecFilGenerique4();
static void E_Coeff0Deriche();
static void E_Coeff1Deriche();
static void E_Coeff2Deriche();
static void E_Coeff3Deriche();
static void E_RecGaussCoeff();
#endif

/* Filtrage recursif : valeurs par defaut pour la structure de parametres.

   Par defaut, les coefficients de filtrage sont mis hors valeurs
   permises et aucun filtrage n'est fait.
*/

#if defined(_ANSI_)
void E_InitRecFilters( e_ParamRecFilPtr par /* parameters */ )
#else
void E_InitRecFilters( par )
e_ParamRecFilPtr par; /* parameters */ 
#endif
{
	/*------------------------------------------------------------*/
	/*                                                            */
	/*   set the default parameters for recursive filtering       */
	/*                                                            */
	/*------------------------------------------------------------*/
	par->deriveX = -1;     /* pas d'operation dans la direction X */
	par->deriveY = -1;
	par->deriveZ = -1;
	par->alphaX = -1.0;    /* le coefficient alpha est hors des limites */
	par->alphaY = -1.0;
	par->alphaZ = -1.0;
	par->bord_typeX = 0;
	par->bord_longX = 0;
	par->bord_typeY = 0;
	par->bord_longY = 0;
	par->bord_typeZ = 0;
	par->bord_longZ = 0;
	par->extr_mode = MODE_3D;
	par->esti_mode = ESTI_EUCLI;
	par->norm_mode = 0;   /* default */
	par->verb = FALSE;    /* pas de message */
}

/* Filtrage recursif d'une image.

   Cette procedure realise les filtrages d'ordre 0 (lissage) a 3 (derivee
   d'ordre 3) selon les trois directions X, Y ou Z. 

   Pour chacune de ces directions, le type du filtrage doit etre specifie
   dans le champ derive(X,Y,Z) de la structure de parametre :
   -1 : pas de filtrage
   0 : lissage
   1 : derivee premiere
   2 : derivee seconde
   3 : derivee troisieme

   Le filtrage est separable. 

   Pour attenuer l'effet de "cut-off", des valeurs peuvent etre ajoutees a 
   gauche et a droite de la ligne a traiter. Le nombre de ces valeurs 
   ajoutees est contenu dans le champ bord_long(X,Y,Z) de la structure 
   de parametre. Le mode de calcul de ces valeurs est specifie par le
   champ bord_type(X,Y,Z) de la structure de parametre :
   BORD_TORUS : la fin de la ligne est poursuivie par le debut de la 
                ligne, on considere celle-ci comme un tore
   BORD_MIRROR : on realise une symetrie des valeurs par rapport au dernier
                 point
   BORD_CONT : on conserve la valeur du dernier point comme valeur des 
               points ajoutes.
   
   Le filtrage realise par defaut est de type "Deriche" auquel cas la valeur
   du coefficient de filtrage doit etre specifie dans le champ 
   alpha(X,Y,Z) de la structure de parametre et doit appartenir a 
   l'intervalle [0.1,1.9].

   Si le test (norm_mode & NORM_INV), ou norm_mode est un champ de la 
   structure de parametre, est verifie alors le filtrage utilise
   est l'approximation du filtrage gaussien par un filtre recursif
   d'ordre 4. En ce cas, le champ alpha(X,Y,Z) de la structure de 
   parametre correspond au sigma de la gaussienne et doit etre 
   dans l'intervalle [0.4,theIm->t(x,y,z)/8].

   Les calculs sont faits en flottants, de ce fait pour garder toute la
   precision il est souhaitable que l'image de sortie soit en flottants.
   En outre, afin d'eviter les problemes de normalisation dus a l'utilisation
   de T_Get3Point() et T_Set3Point(), il vaut egalement mieux que l'image
   d'entree soit codee en flottants. Ainsi, les valeurs seront comparables
   et pourront etre utilisees telles quelles ensemble dans des calculs
   futurs.

   L'adresse du buffer de l'image re'sultat peut changer (en ce cas,
   l'ancien buffer a ete libere)

RETURN 
   TRUE en cas de succes, FALSE en cas d'erreur. Les erreurs peuvent provenir
   de coefficients hors de la plage per
*/   

#if defined(_ANSI_)
int E_RecFilters( t_ImagePtr theIm /* input image */, 
		 t_ImagePtr resIm /* result image */, 
		 e_ParamRecFilPtr par /* parameters */ )
#else
int E_RecFilters( theIm, resIm, par )
t_ImagePtr theIm; /* input image */ 
t_ImagePtr resIm; /* result image */
e_ParamRecFilPtr par; /* parameters */
#endif
{
	int actionX, actionY, actionZ;
	t_Image AuxIm;
	
	actionX = actionY = actionZ = FALSE;

	E_ERR( (theIm==NULL) || (theIm->buf==NULL), E_ERRNULL, FALSE,
	     "E_RecFilters", "input image not allocated");
	E_ERR( (resIm==NULL) || (resIm->buf==NULL), E_ERRNULL, FALSE,
	     "E_RecFilters", "output image not allocated");
	E_ERR( (theIm->tx!=resIm->tx) || (theIm->ty!=resIm->ty) || (theIm->tz!=resIm->tz),
	       E_ERRFATAL, FALSE, "E_RecFilters", 
	       "input and output have different dimensions");

	if (par->norm_mode & NORM_INV){
	    E_ERR( (par->deriveX>=0) && (par->deriveX<=3) && 
		   ((par->alphaX<0.4) || (par->alphaX>theIm->tx/8)||(theIm->tx<4)),
		   E_ERRFATAL, FALSE, "E_RecFilters", "error in parameters");
	    E_ERR( (par->deriveY>=0) && (par->deriveY<=3) && 
		   ((par->alphaY<0.4) || (par->alphaY>theIm->ty/8)||(theIm->ty<4)),
		   E_ERRFATAL, FALSE, "E_RecFilters", "error in parameters");
	    E_ERR( (par->deriveZ>=0) && (par->deriveZ<=3) && 
		   ((par->alphaZ<0.4) || (par->alphaZ>theIm->tz/8)||(theIm->tz<4)),
		   E_ERRFATAL, FALSE, "E_RecFilters", "error in parameters");
	}
	else{  /* default */
	    E_ERR( (par->deriveX>=0) && (par->deriveX<=3) &&
		   ((par->alphaX<0.1) || (par->alphaX>1.9)||(theIm->tx<3)),
		   E_ERRFATAL, FALSE, "E_RecFilters", "error in parameters");
	    E_ERR( (par->deriveY>=0) && (par->deriveY<=3) && 
		   ((par->alphaY<0.1) || (par->alphaY>1.9)||(theIm->ty<3)),
		   E_ERRFATAL, FALSE, "E_RecFilters", "error in parameters");
	    E_ERR( (par->deriveZ>=0) && (par->deriveZ<=3) && 
		   ((par->alphaZ<0.1) || (par->alphaZ>1.9)||(theIm->tz<3)),
		   E_ERRFATAL, FALSE, "E_RecFilters", "error in parameters");
	}

	if ( (par->deriveX>=0) && (par->deriveX<=3) )
	    actionX = TRUE;
	if ( (par->deriveY>=0) && (par->deriveY<=3) )
	    actionY = TRUE;
	if ( (par->deriveZ>=0) && (par->deriveZ<=3) )
	    actionZ = TRUE;
	
	/*   images auxiliaires   */
	/**************************/
	/*--- il y a une image intermediaire en flottant :
	      o si l'image resultat est en flottant, on ne l'alloue pas
	        mais on lui attribue le buffer de l'image resultat
              o si l'image d'entree est differente de l'image resultat
                on recopie l'image d'entree dans l'image intermediaire ---*/
		
	T_Init3Image( &AuxIm, theIm->tx, theIm->ty, theIm->tz, TYPEIM_FLOAT, CNULL );

	if ( resIm->type != TYPEIM_FLOAT ) 
	    T_AllocImage( &AuxIm );
	else
	    AuxIm.buf = resIm->buf;

	E_REPERR( goto END_E_RecFilters );

	/*--- copie eventuelle de l'image d'entree ---*/
	if ( (theIm->type != TYPEIM_FLOAT) || ( (theIm->type == TYPEIM_FLOAT) && (theIm->buf != resIm->buf) ) ) {
	    if ( E_CopyImage( theIm, &AuxIm ) != TRUE ) {
		E_REPERR( goto END_E_RecFilters );
	    }
	}

	/*   filtrage   */
	/****************/
	if ( actionX == TRUE ) {
	    E_RecFilalongX( &AuxIm, &AuxIm, par );
	    E_REPERR( goto END_E_RecFilters );
	}
	if ( actionY == TRUE ) {
	    E_RecFilalongY( &AuxIm, &AuxIm, par );
	    E_REPERR( goto END_E_RecFilters );
	}
	if ( actionZ == TRUE ) {
	    E_RecFilalongZ( &AuxIm, &AuxIm, par );
	    E_REPERR( goto END_E_RecFilters );
	}

	/*   images finales   */
	/**********************/
	if ( resIm->type != TYPEIM_FLOAT) {
	    if ( E_CopyImage( &AuxIm, resIm ) != TRUE ) {
		E_REPERR( goto END_E_RecFilters );
	    }
	}

END_E_RecFilters:

	if ( resIm->type != TYPEIM_FLOAT)
	    T_FreeImage( &AuxIm );
	E_ISERR(FALSE,"E_RecFilters",(char*)NULL);

	return( TRUE );
}

#if defined(_ANSI_)
static int E_RecFilalongX( t_ImagePtr theIm, t_ImagePtr resIm, e_ParamRecFilPtr par )
#else
static int E_RecFilalongX( theIm, resIm, par )
t_ImagePtr theIm;
t_ImagePtr resIm;
e_ParamRecFilPtr par ;
#endif
{
	/*------------------------------------------------------------------------*/
	/*   recursive filtering along the X axis                                 */
	/*------------------------------------------------------------------------*/

	register int i, j, k, l, t, y, z;
	float *in=NULL, *work1=NULL, *out=NULL, *vol1, *vol2;
	int dim, ndim, edim;
	e_CoeffRecFil coeff;

	if (par->norm_mode & NORM_INV)
	    E_RecGaussCoeff (par->deriveX, &coeff, par->alphaX, par->norm_mode);
	else /* default */
	    switch (par->deriveX) {
	    case 0 :
		E_Coeff0Deriche( &coeff, par->alphaX );
		break;
	    case 1 :
		E_Coeff1Deriche( &coeff, par->alphaX );
		break;
	    case 2 :
		E_Coeff2Deriche( &coeff, par->alphaX );
		break;
	    case 3 :
		E_Coeff3Deriche( &coeff, par->alphaX );
		break;
	    default :
		E_ISERR(FALSE,"E_RecFilalongX", (char*)NULL);
	    }

	dim = theIm->tx;
	edim = EPI_MIN( dim-1, EPI_MAX( par->bord_longX, 0 ) );
	ndim = dim+2*edim;
        vol1 = (float*)(theIm->buf);
        vol2 = (float*)(resIm->buf);

	in = (float*)T_Alloc( (unsigned)(ndim*sizeof(float)) );
	work1 = (float*)T_Alloc( (unsigned)(ndim*sizeof(float)) );
        out = (float*)T_Alloc( (unsigned)(ndim*sizeof(float)) );
	E_REPERR( goto END_E_RecFilalongX );
	
	for (z=0; z<theIm->tz; z++)
		for (y=0; y<theIm->ty; y++) {
			t = (z * theIm->ty + y) * theIm->tx;
			/*   saisie de la ligne   */
			/**************************/
			for (i=0,j=edim,k=t; i<dim; i++,j++,k++)
				*(in+j) = *(vol1+k);
			/*   calcul de la bordure   */
			/****************************/
			if (edim) {
				switch (par->bord_typeX) {
				case BORD_TORUS :
					for (i=0,j=dim,k=dim+edim,l=edim; i<edim; i++,j++,k++,l++) {
						*(in+i) = *(in+j);
						*(in+k) = *(in+l);
					}
					break;
				case BORD_MIRROR :
					for (i=0,j=2*edim,k=dim+edim,l=dim+edim-2; i<edim; i++,j--,k++,l--) {
						*(in+i) = 2 * *(in+edim) - *(in+j);
						*(in+k) = 2 * *(in+dim+edim-1) - *(in+l);
					}
					break;
				case BORD_CONT :
					for (i=0,j=edim,k=dim+edim,l=dim+edim-1; i<edim; i++,k++) {
						*(in+i) = *(in+j);
						*(in+k) = *(in+l);
					}
					break;
				}
			}
			/*   calcul   */
			/**************/
			if (par->norm_mode & NORM_INV)
			    E_RecFilGenerique4( in, out, work1, out, &coeff, ndim );
			else /* default */
			    switch (par->deriveX) {
			    case 0 :
				E_RecFil0Deriche( in, out, work1, out, &coeff, ndim );
				break;
			    case 1 :
				E_RecFil1Deriche( in, out, work1, out, &coeff, ndim );
				break;
			    case 2 :
				E_RecFil2Deriche( in, out, work1, out, &coeff, ndim );
				break;
			    case 3 :
				E_RecFil3Deriche( in, out, work1, out, &coeff, ndim );
				break;
			    default :
				E_REPERR( goto END_E_RecFilalongX );
			    }
			/*   resultat   */
			/****************/
			for (i=0,j=edim,k=t; i<dim; i++,j++,k++)
				*(vol2+k) = *(out+j);
		}
	
END_E_RecFilalongX:

        T_Free((void **)&in);
	T_Free((void **)&work1);
        T_Free((void **)&out);
        E_ISERR(FALSE,"E_RecFilalongX", (char*)NULL);

        return( TRUE );
}
	
#if defined(_ANSI_)
static int E_RecFilalongY( t_ImagePtr theIm, t_ImagePtr resIm, e_ParamRecFilPtr par )
#else
static int E_RecFilalongY(theIm, resIm, par )
t_ImagePtr theIm;
t_ImagePtr resIm;
e_ParamRecFilPtr par ;
#endif
{
	/*------------------------------------------------------------------------*/
	/*   recursive filtering along the Y axis                                 */
	/*------------------------------------------------------------------------*/

	register int i, j, k, l, t, x, z, dimx;
	float *in=NULL, *work1=NULL, *out=NULL, *vol1, *vol2;
	int dim, ndim, edim;
	e_CoeffRecFil coeff;

	if (par->norm_mode & NORM_INV)
	    E_RecGaussCoeff (par->deriveY, &coeff, par->alphaY, par->norm_mode);
	else /* default */
	    switch (par->deriveY) {
	    case 0 :
		E_Coeff0Deriche( &coeff, par->alphaY );
		break;
	    case 1 :
		E_Coeff1Deriche( &coeff, par->alphaY );
		break;
	    case 2 :
		E_Coeff2Deriche( &coeff, par->alphaY );
		break;
	    case 3 :
		E_Coeff3Deriche( &coeff, par->alphaY );
		break;
	    default :
		E_ISERR(FALSE,"E_RecFilalongY", (char*)NULL);
	    }

	dim = theIm->ty;
	dimx = theIm->tx;
	edim = EPI_MIN( dim-1, EPI_MAX( par->bord_longY, 0 ) );
	ndim = dim+2*edim;
        vol1 = (float*)(theIm->buf);
        vol2 = (float*)(resIm->buf);

	in = (float*)T_Alloc( (unsigned)(ndim*sizeof(float)) );
	work1 = (float*)T_Alloc( (unsigned)(ndim*sizeof(float)) );
        out = (float*)T_Alloc( (unsigned)(ndim*sizeof(float)) );
	E_REPERR( goto END_E_RecFilalongY );
	
	for (z=0; z<theIm->tz; z++)
		for (x=0; x<theIm->tx; x++) {
			t = z * theIm->ty * theIm->tx + x;
			/*   saisie de la ligne   */
			/**************************/
			for (i=0,j=edim,k=t; i<dim; i++,j++,k+=dimx)
				*(in+j) = *(vol1+k);
			/*   calcul de la bordure   */
			/****************************/
			if (edim) {
				switch (par->bord_typeY) {
				case BORD_TORUS :
					for (i=0,j=dim,k=dim+edim,l=edim; i<edim; i++,j++,k++,l++) {
						*(in+i) = *(in+j);
						*(in+k) = *(in+l);
					}
					break;
				case BORD_MIRROR :
					for (i=0,j=2*edim,k=dim+edim,l=dim+edim-2; i<edim; i++,j--,k++,l--) {
						*(in+i) = 2 * *(in+edim) - *(in+j);
						*(in+k) = 2 * *(in+dim+edim-1) - *(in+l);
					}
					break;
				case BORD_CONT :
					for (i=0,j=edim,k=dim+edim,l=dim+edim-1; i<edim; i++,k++) {
						*(in+i) = *(in+j);
						*(in+k) = *(in+l);
					}
					break;
				}
			}
			/*   calcul   */
			/**************/
			if (par->norm_mode & NORM_INV)
			    E_RecFilGenerique4( in, out, work1, out, &coeff, ndim );
			else /* default */
			    switch (par->deriveY) {
			    case 0 :
				E_RecFil0Deriche( in, out, work1, out, &coeff, ndim );
				break;
			    case 1 :
				E_RecFil1Deriche( in, out, work1, out, &coeff, ndim );
				break;
			    case 2 :
				E_RecFil2Deriche( in, out, work1, out, &coeff, ndim );
				break;
			    case 3 :
				E_RecFil3Deriche( in, out, work1, out, &coeff, ndim );
				break;
			    default :
				E_REPERR( goto END_E_RecFilalongY );
			}
			/*   resultat   */
			/****************/
			for (i=0,j=edim,k=t; i<dim; i++,j++,k+=dimx)
				*(vol2+k) = *(out+j);
		}
	
END_E_RecFilalongY:

        T_Free((void **)&in);
	T_Free((void **)&work1);
        T_Free((void **)&out);
        E_ISERR(FALSE,"E_RecFilalongY", (char*)NULL);

        return( TRUE );
}
	
#if defined(_ANSI_)
static int E_RecFilalongZ( t_ImagePtr theIm, t_ImagePtr resIm, e_ParamRecFilPtr par )
#else
static int E_RecFilalongZ(theIm, resIm, par )
t_ImagePtr theIm;
t_ImagePtr resIm;
e_ParamRecFilPtr par ;
#endif
{
	/*------------------------------------------------------------------------*/
	/*   recursive filtering along the Z axis                                 */
	/*------------------------------------------------------------------------*/

	register int i, j, k, l, t, y, x, plan;
	float *in=NULL, *work1=NULL, *out=NULL, *vol1, *vol2;
	int dim, ndim, edim;
	e_CoeffRecFil coeff;

	if (par->norm_mode & NORM_INV)
	    E_RecGaussCoeff (par->deriveZ, &coeff, par->alphaZ, par->norm_mode);
	else /* default */
	    switch (par->deriveZ) {
	    case 0 :
		E_Coeff0Deriche( &coeff, par->alphaZ );
		break;
	    case 1 :
		E_Coeff1Deriche( &coeff, par->alphaZ );
		break;
	    case 2 :
		E_Coeff2Deriche( &coeff, par->alphaZ );
		break;
	    case 3 :
		E_Coeff3Deriche( &coeff, par->alphaZ );
		break;
	    default :
		E_ISERR(FALSE,"E_RecFilalongZ", (char*)NULL);
	    }

	dim = theIm->tz;
	plan = theIm->tx * theIm->ty;
	edim = EPI_MIN( dim-1, EPI_MAX( par->bord_longZ, 0 ) );
	ndim = dim+2*edim;
        vol1 = (float*)(theIm->buf);
        vol2 = (float*)(resIm->buf);

	in = (float*)T_Alloc( (unsigned)(ndim*sizeof(float)) );
	work1 = (float*)T_Alloc( (unsigned)(ndim*sizeof(float)) );
        out = (float*)T_Alloc( (unsigned)(ndim*sizeof(float)) );
	E_REPERR( goto END_E_RecFilalongZ );
	
	for (x=0; x<theIm->tx; x++)
		for (y=0; y<theIm->ty; y++) {
			t = x + y * theIm->tx;
			/*   saisie de la ligne   */
			/**************************/
			for (i=0,j=edim,k=t; i<dim; i++,j++,k+=plan)
				*(in+j) = *(vol1+k);
			/*   calcul de la bordure   */
			/****************************/
			if (edim) {
				switch (par->bord_typeZ) {
				case BORD_TORUS :
					for (i=0,j=dim,k=dim+edim,l=edim; i<edim; i++,j++,k++,l++) {
						*(in+i) = *(in+j);
						*(in+k) = *(in+l);
					}
					break;
				case BORD_MIRROR :
					for (i=0,j=2*edim,k=dim+edim,l=dim+edim-2; i<edim; i++,j--,k++,l--) {
						*(in+i) = 2 * *(in+edim) - *(in+j);
						*(in+k) = 2 * *(in+dim+edim-1) - *(in+l);
					}
					break;
				case BORD_CONT :
					for (i=0,j=edim,k=dim+edim,l=dim+edim-1; i<edim; i++,k++) {
						*(in+i) = *(in+j);
						*(in+k) = *(in+l);
					}
					break;
				}
			}
			/*   calcul   */
			/**************/
			if (par->norm_mode & NORM_INV)
			    E_RecFilGenerique4( in, out, work1, out, &coeff, ndim );
			else /* default */
			    switch (par->deriveZ) {
			    case 0 :
				E_RecFil0Deriche( in, out, work1, out, &coeff, ndim );
				break;
			    case 1 :
				E_RecFil1Deriche( in, out, work1, out, &coeff, ndim );
				break;
			    case 2 :
				E_RecFil2Deriche( in, out, work1, out, &coeff, ndim );
				break;
			    case 3 :
				E_RecFil3Deriche( in, out, work1, out, &coeff, ndim );
				break;
			    default :
				E_REPERR( goto END_E_RecFilalongZ );
			    }
			/*   resultat   */
			/****************/
			for (i=0,j=edim,k=t; i<dim; i++,j++,k+=plan)
				*(vol2+k) = *(out+j);
		}
	
END_E_RecFilalongZ:

        T_Free((void **)&in);
	T_Free((void **)&work1);
        T_Free((void **)&out);
        E_ISERR(FALSE,"E_RecFilalongZ", (char*)NULL);

        return( TRUE );
}

#if defined(_ANSI_)
static void E_RecFil0Deriche( float *in, float *out, float *work1, float *work2, e_CoeffRecFilPtr par, int dim )
#else
static void E_RecFil0Deriche(in, out, work1, work2, par, dim )
float *in;
float *out;
float *work1;
float *work2;
e_CoeffRecFilPtr par;
int dim ;
#endif
{
	/*------------------------------------------------------------------------*/
	/*   modified smoothing Deriche filter                                    */
	/*------------------------------------------------------------------------*/

	register float a0, a1, a2, a3, b1, b2;
	register int i,j,k;

	a0 = par->p0;
	a1 = par->p1;
	b1 = par->d1;
	b2 = par->d2;
	a2 = par->n1;
	a3 = par->n2;

        /*   calcul de y+   */
        /********************/
	*(work1)   = a0 * *(in);
	*(work1+1) = a0 * *(in+1) + a1 * *(in) - b1 * *(work1);     
        for (i=2,j=1,k=0;  i<dim; i++,j++,k++)
		*(work1+i) = a0 * *(in+i) + a1 * *(in+j)
			- b1 * *(work1+j) - b2 * *(work1+k);

        /*   calcul de y-   */
        /********************/
        *(work2+dim-1) = 0.0;
	*(work2+dim-2) = a2 * *(in+dim-1);
	for (i=dim-3,j=dim-2,k=dim-1; i>=0; i--,j--,k--)
		*(work2+i) = a2 * *(in+j) + a3 * *(in+k)
			- b1 * *(work2+j) - b2 * *(work2+k);
	/*   calcul final   */
	/********************/
	for (i=0 ; i<dim ; i++)
		*(out+i) = *(work1+i) + *(work2+i);
}

#if defined(_ANSI_)
static void E_RecFil1Deriche( float *in, float *out, float *work1, float *work2, e_CoeffRecFilPtr par, int dim )
#else
static void E_RecFil1Deriche(in, out, work1, work2, par, dim )
float *in;
float *out;
float *work1;
float *work2;
e_CoeffRecFilPtr par;
int dim ;
#endif
{
	/*------------------------------------------------------------------------*/
	/*   modified 1st derivative Deriche filter                               */
	/*------------------------------------------------------------------------*/

	register float a1, a2, b1, b2;
	register int i,j,k;

	a1 = par->p1;
	a2 = par->n1;
	b1 = par->d1;
	b2 = par->d2;

	/*   calcul de y+   */
        /********************/
	*(work1)   = 0.0;
	*(work1+1) = a1 * *(in);     
        for (i=2,j=1,k=0;  i<dim; i++,j++,k++)
		*(work1+i) = a1 * *(in+j)
			- b1 * *(work1+j) - b2 * *(work1+k);

        /*   calcul de y-   */
        /********************/
        *(work2+dim-1) = 0.0;
	*(work2+dim-2) = a2 * *(in+dim-1);
	for (i=dim-3,j=dim-2,k=dim-1; i>=0; i--,j--,k--)
		*(work2+i) = a2 * *(in+j)
			- b1 * *(work2+j) - b2 * *(work2+k);
	/*   calcul final   */
	/********************/
	for (i=0 ; i<dim ; i++)
		*(out+i) = *(work1+i) + *(work2+i);
}

#if defined(_ANSI_)
static void E_RecFil2Deriche( float *in, float *out, float *work1, float *work2, e_CoeffRecFilPtr par, int dim )
#else
static void E_RecFil2Deriche( in, out, work1, work2, par, dim )
float *in;
float *out;
float *work1;
float *work2;
e_CoeffRecFilPtr par;
int dim ;
#endif
{
	/*------------------------------------------------------------------------*/
	/*   modified 2nd derivative Deriche filter                               */
	/*------------------------------------------------------------------------*/

	register float a0, a1, a2, a3, b1, b2;
	register int i,j,k;

	a0 = par->p0;
	a1 = par->p1;
	b1 = par->d1;
	b2 = par->d2;
	a2 = par->n1;
	a3 = par->n2;

        /*   calcul de y+   */
        /********************/
	*(work1)   = a0 * *(in);
	*(work1+1) = a0 * *(in+1) + a1 * *(in) - b1 * *(work1);     
        for (i=2,j=1,k=0;  i<dim; i++,j++,k++)
		*(work1+i) = a0 * *(in+i) + a1 * *(in+j)
			- b1 * *(work1+j) - b2 * *(work1+k);

        /*   calcul de y-   */
        /********************/
        *(work2+dim-1) = 0.0;
	*(work2+dim-2) = a2 * *(in+dim-1);
	for (i=dim-3,j=dim-2,k=dim-1; i>=0; i--,j--,k--)
		*(work2+i) = a2 * *(in+j) + a3 * *(in+k)
			- b1 * *(work2+j) - b2 * *(work2+k);
	/*   calcul final   */
	/********************/
	for (i=0 ; i<dim ; i++)
		*(out+i) = *(work1+i) + *(work2+i);
}

#if defined(_ANSI_)
static void E_RecFil3Deriche( float *in, float *out, float *work1, float *work2, e_CoeffRecFilPtr par, int dim )
#else
static void E_RecFil3Deriche(in, out, work1, work2, par, dim )
float *in;
float *out;
float *work1;
float *work2;
e_CoeffRecFilPtr par;
int dim ;
#endif
{
	/*------------------------------------------------------------------------*/
	/*   modified 3rd derivative Deriche filter                               */
	/*------------------------------------------------------------------------*/

	register float a0, a1, a2, a3, b1, b2;
	register int i,j,k;

	a0 = par->p0;
	a1 = par->p1;
	b1 = par->d1;
	b2 = par->d2;
	a2 = par->n0;
	a3 = par->n1;

        /*   calcul de y+   */
        /********************/
	*(work1)   = a0 * *(in);
	*(work1+1) = a0 * *(in+1) + a1 * *(in) - b1 * *(work1);     
        for (i=2,j=1,k=0;  i<dim; i++,j++,k++)
		*(work1+i) = a0 * *(in+i) + a1 * *(in+j)
			- b1 * *(work1+j) - b2 * *(work1+k);

        /*   calcul de y-   */
        /********************/
        *(work2+dim-1) = a2 * *(in+dim-1);
	*(work2+dim-2) = a2 * *(in+dim-2) + a3 * *(in+dim-1) - b1 * *(work2+dim-1);
	for (i=dim-3,j=dim-2,k=dim-1; i>=0; i--,j--,k--)
		*(work2+i) = a2 * *(in+i) + a3 * *(in+j)
			- b1 * *(work2+j) - b2 * *(work2+k);
	/*   calcul final   */
	/********************/
	for (i=0 ; i<dim ; i++)
		*(out+i) = *(work1+i) + *(work2+i);
}

#if defined(_ANSI_)
static void E_RecFilGenerique4( float *in, float *out, float *work1, float *work2, e_CoeffRecFilPtr par, int dim )
#else
static void E_RecFilGenerique4(in, out, work1, work2, par, dim )
float *in;
float *out;
float *work1;
float *work2;
e_CoeffRecFilPtr par;
int dim ;
#endif
{
	/*------------------------------------------------------------------------*/
	/*   generic recursive filter of 4th order                                */
	/*------------------------------------------------------------------------*/
	/*
	 * une autre type de normalization et de filtrage est rajoute:
	 * l'option MODE_INV utilise le filtrage generique d'ordre 4, qui permet 
	 * de calculer les derivees du gaussien pour obtenir les invariants geometriques
	 * les fonctions modifiees: E_RecFilGenerique4
	 * M. Fidrich
	 */

	register int i;
	register float pp0, pp1, pp2, pp3;
	register float pd1, pd2, pd3 ,pd4;
	register float pn1, pn2, pn3, pn4;
	
	pp0 = par->p0;
	pp1 = par->p1;
	pp2 = par->p2;
	pp3 = par->p3;
	pd1 = par->d1;
	pd2 = par->d2;
	pd3 = par->d3;
	pd4 = par->d4;
	pn1 = par->n1;
	pn2 = par->n2;
	pn3 = par->n3;
	pn4 = par->n4;

        /*   calcul de y+   */
        /********************/
	*(work1)   = pp0 * *(in);
	*(work1+1) = pp0 * *(in+1) + pp1 * *(in) -
		     pd1 * *(work1);   
	*(work1+2) = pp0 * *(in+2) + pp1 * *(in+1) + pp2 * *(in) -
	             pd1 * *(work1+1) - pd2 * *(work1);
	*(work1+3) = pp0 * *(in+3) + pp1 * *(in+2) + pp2 * *(in+1) + pp3 * *(in) -
		     pd1 * *(work1+2) - pd2 * *(work1+1) - pd3 * *(work1);
	for (i=4; i<dim; i++)
            *(work1+i) = pp0 * *(in+i) + pp1 * *(in+i-1) +
			 pp2 * *(in+i-2) + pp3 * *(in+i-3) -
			 pd1 * *(work1+i-1) - pd2 * *(work1+i-2) -
			 pd3 * *(work1+i-3) - pd4 * *(work1+i-4);
        /*   calcul de y-   */
        /********************/
        *(work2+dim-1) = 0;
	*(work2+dim-2) = pn1 * *(in+dim-1);
	*(work2+dim-3) = pn1 * *(in+dim-2) + pn2 * *(in+dim-1) - pd1 * *(work2+dim-2);
	*(work2+dim-4) = pn1 * *(in+dim-3) + pn2 * *(in+dim-2) + pn3 * *(in+dim-1) -
		         pd1 * *(work2+dim-3) - pd2 * *(work2+dim-2);
	for (i=dim-5; i>=0; i--)
	    *(work2+i) = pn1 * *(in+i+1) + pn2 * *(in+i+2) + pn3 * *(in+i+3) +
                         pn4 * *(in+i+4) - pd1 * *(work2+i+1) - pd2 * *(work2+i+2) -
			 pd3 * *(work2+i+3) - pd4 * *(work2+i+4);
	/*   calcul final   */
	/********************/
	for (i=0 ; i<dim ; i++)
		*(out+i) = *(work1+i) + *(work2+i);
}

#if defined(_ANSI_)
static void E_Coeff0Deriche( e_CoeffRecFilPtr par, float alpha )
#else
static void E_Coeff0Deriche(par, alpha )
e_CoeffRecFilPtr par;
float alpha ;
#endif
{
	/*------------------------------------------------------------------------*/
	/*   coefficients of the modified smoothing Deriche filter                */
	/*------------------------------------------------------------------------*/
	double ea;
	
	ea = exp((double)(-alpha));
	par->p0 = (float) ((1.0 - ea)*(1.0 - ea) / (1.0 + 2.0*alpha*ea - ea*ea));
	par->p1 = (float) (par->p0 * (alpha - 1.0) * ea);
	par->p2 = 0.0;
	par->p3 = 0.0;
	par->n0 = 0.0;
	par->n1 = (float) (par->p0 * (alpha + 1.0) * ea);
	par->n2 = (float) ( - par->p0 * ea * ea);
	par->n3 = 0.0;
	par->n4 = 0.0;
	par->d1 = (float) (-2.0 * ea);
	par->d2 = (float) (ea * ea);
	par->d3 = 0.0;
	par->d4 = 0.0;
}

#if defined(_ANSI_)
static void E_Coeff1Deriche( e_CoeffRecFilPtr par, float alpha )
#else
static void E_Coeff1Deriche(par, alpha )
e_CoeffRecFilPtr par;
float alpha ;
#endif
{
	/*------------------------------------------------------------------------*/
	/*   coefficients of the modified 1st derivative Deriche filter           */
	/*------------------------------------------------------------------------*/
	double ea;
	
	ea = exp((double)(-alpha));
	par->p0 = 0.0;
	par->p1 = (float) (- (1.0 - ea)*(1.0 - ea)*(1.0 - ea) / (2.0*(1.0 + ea)));
	par->p2 = 0.0;
	par->p3 = 0.0;
	par->n0 = 0.0;
	par->n1 = - par->p1;
	par->n2 = 0.0;
	par->n3 = 0.0;
	par->n4 = 0.0;
	par->d1 = (float) (-2.0 * ea);
	par->d2 = (float) (ea * ea);
	par->d3 = 0.0;
	par->d4 = 0.0;
}

#if defined(_ANSI_)
static void E_Coeff2Deriche( e_CoeffRecFilPtr par, float alpha )
#else
static void E_Coeff2Deriche(par, alpha )
e_CoeffRecFilPtr par;
float alpha ;
#endif
{
	/*------------------------------------------------------------------------*/
	/*   coefficients of the modified 2nd derivative Deriche filter           */
	/*------------------------------------------------------------------------*/
	double ea, k1, k2;
	
	ea = exp((double)(-alpha));
	k1 = -2.0 * (1.0 - ea) * (1.0 - ea) * (1.0 - ea);
	k1 /= (1.0 + ea) * (1.0 + ea) * (1.0 + ea);
	k2 = (1.0 - ea*ea)/(2.0*ea);
	par->p0 = (float) k1;
	par->p1 = (float) (- k1 * ( 1.0 + k2 )*ea);
	par->p2 = 0.0;
	par->p3 = 0.0;
	par->n0 = 0.0;
	par->n1 = (float) (k1 * ( 1.0 - k2 )*ea);
	par->n2 = (float) (-k1*ea*ea);
	par->n3 = 0.0;
	par->n4 = 0.0;
	par->d1 = (float) (-2.0 * ea);
	par->d2 = (float) (ea * ea);
	par->d3 = 0.0;
	par->d4 = 0.0;
}

#if defined(_ANSI_)
static void E_Coeff3Deriche( e_CoeffRecFilPtr par, float alpha )
#else
static void E_Coeff3Deriche(par, alpha )
e_CoeffRecFilPtr par;
float alpha ;
#endif
{
	/*------------------------------------------------------------------------*/
	/*   coefficients of the modified 3rd derivative Deriche filter           */
	/*------------------------------------------------------------------------*/
	double ea, eb, k1, k2;
	
	ea = exp((double)(-alpha));
	eb = 1.0 - ea;
	k1 = (1.0 + alpha)*ea + (alpha - 1.0);
	k2 = eb / k1;
	k1 *= eb * eb * eb * eb;
	k1 /= 2.0 * alpha * alpha * ea * ea;
	k1 /= ea + 1.0;
	par->p0 = (float) (k1 * alpha * (k2 + 1.0));
	par->p1 = (float) (- k1 * alpha * ( 1.0 + k2 + k2*alpha )*ea);
	par->p2 = 0.0;
	par->p3 = 0.0;
	par->n0 = - par->p0;
	par->n1 = - par->p1;
	par->n2 = 0.0;
	par->n3 = 0.0;
	par->n4 = 0.0;
	par->d1 = (float) (-2.0 * ea);
	par->d2 = (float) (ea * ea);
	par->d3 = 0.0;
	par->d4 = 0.0;
}

#if defined(_ANSI_)
static void E_RecGaussCoeff (int der, e_CoeffRecFilPtr coeff, float sigma, int mode)
#else
static void E_RecGaussCoeff (der, coeff, sigma, mode)
int der;
e_CoeffRecFilPtr coeff;
float sigma;
int mode;
#endif
{
  /*------------------------------------------------------------------------*/
  /*   coefficients of the recursive 4th order Deriche filter               */
  /*   to get invariant gaussian derivatives                                */
  /*------------------------------------------------------------------------*/
  /*
   * une autre type de normalization et de filtrage est rajoute:
   * l'option MODE_INV utilise le filtrage generique d'ordre 4, qui permet 
   * de calculer les derivees du gaussien pour obtenir les invariants geometriques
   * les fonctions ajoutees:  E_RecGaussCoeff
   * M. Fidrich
   */
  
  double a0, a1, c0, c1, omega0, omega1, b0, b1;
  
  a0 = a1 = 0.0;
  b0 = b1 = 0.0;
  c0 = c1 = 0.0;
  omega0 = omega1 = 0.0;
  
  /*--- Entree des coefficients du filtre d'approximation---*/
  
  if (der == 0) {
    a0 = 0.6570033214 / sigma;
    a1 = 1.978946687 / sigma;
    c0 = -0.2580640608 / sigma;
    c1 = -0.2391206463 / sigma;
    omega0 =0.6512453378;
    omega1 = 2.05339943;
    b0 = 1.906154352;
    b1 = 1.881305409;
  }
  else if (der == 1) {
    if (mode & NORM_DIFF){
      a0 = -0.1726729496 / sigma;
      a1 = -2.003565572 / sigma;
      c0 = 0.1726730777 / sigma;
      c1 = 0.4440126835 / sigma;
    }
    else{
      a0 = -0.1726729496 / (sigma*sigma);
      a1 = -2.003565572 / (sigma*sigma);
      c0 = 0.1726730777 / (sigma*sigma);
      c1 = 0.4440126835 / (sigma*sigma);
    }
    b0 = 1.560644213;
    b1 = 1.594202256;
    omega0 = 0.6995461735;
    omega1 = 2.144671764;
  }
  else if (der == 2) {
    if (mode & NORM_DIFF){
      a0 = -0.7241334169 / sigma;
      a1 = 1.688628765 / sigma;
      c0 = 0.3251949838 / sigma;
      c1 = -0.7211796018 / sigma;
    }
    else{
      a0 = -0.7241334169 / (sigma*sigma*sigma);
      a1 = 1.688628765 / (sigma*sigma*sigma);
      c0 = 0.3251949838 / (sigma*sigma*sigma);
      c1 = -0.7211796018 / (sigma*sigma*sigma);
    }
    b0 = 1.294951143;
    b1 = 1.427007123;
    omega0 = 0.7789803775;
    omega1 = 2.233566862;
  }
  else if (der == 3) {
    if (mode & NORM_DIFF){
      a0 = 1.285774106 / sigma;
      a1 = -0.2896378408 / sigma;
      c0 = -1.28577129 / sigma;
      c1 = 0.26249833 / sigma;
    }
    else{
      a0 = 1.285774106 / pow(sigma,4.0);
      a1 = -0.2896378408 / pow(sigma,4.0);
      c0 = -1.28577129 / pow(sigma,4.0);
      c1 = 0.26249833 / pow(sigma,4.0);
    }
    b0 = 1.01162886;
    b1 = 1.273344739;
    omega0 = 0.9474270928;
    omega1 = 2.337607006;
  }
  else
    E_XERRN (E_ERRFATAL, "E_RecGaussCoeff", "Derivative is out of [0  3]");
  
	/*---Calcul des coefficients de la reponse impulsionnelle du filtre----*/
  
  coeff->p0 = (float) (a0 + c0);
  
  coeff->p1 = (float) ( exp(-b1/sigma) * 
    (c1 * sin(omega1/sigma) - (c0 + 2*a0) * cos(omega1/sigma)) +
    exp(-b0/sigma) * 
    (a1 * sin(omega0/sigma) - (2*c0 + a0) * cos(omega0/sigma)));
  
  coeff->p2 = (float) (2 * exp(-b0/sigma - b1/sigma) * 
    ((a0 + c0) * cos(omega1/sigma) * cos(omega0/sigma) -
     cos(omega1/sigma) * a1 * sin(omega0/sigma) - cos(omega0/sigma) * 
		 c1 * sin(omega1/sigma)) +
    c0 * exp(-2 * b0/sigma) + a0 * exp(-2 * b1/sigma));
	
  coeff->p3 = (float) (exp(-b1/sigma - 2*b0/sigma) * 
    (c1 * sin(omega1/sigma) - c0 * cos(omega1/sigma)) +	
    exp(-b0/sigma - 2 * b1/sigma) * (a1 * sin(omega0/sigma) - a0 * cos(omega0/sigma)));
  
  coeff->d1 = (float) (-2 * exp(-b1/sigma) * cos(omega1/sigma) - 
    2 * exp(-b0/sigma) * cos(omega0/sigma));
  
  coeff->d2 = (float) (4 * cos(omega1/sigma) * cos(omega0/sigma) * exp(-b0/sigma - b1/sigma) +
    exp(-2 * b1/sigma) + exp(-2 * b0/sigma));
  
  coeff->d3 = (float) (-2 * cos(omega0/sigma) * exp(-b0/sigma - 2 * b1/sigma) -
    2 * cos(omega1/sigma) * exp(-b1/sigma - 2 * b0/sigma));
  
  coeff->d4 = (float) (exp(-2 * b0/sigma - 2 * b1/sigma));
  
  if (der == 0 || der == 2) {
    coeff->n1 = coeff->p1 - coeff->d1 * coeff->p0;
    coeff->n2 = coeff->p2 - coeff->d2 * coeff->p0;
    coeff->n3 = coeff->p3 - coeff->d3 * coeff->p0;
    coeff->n4 = -coeff->d4 * coeff->p0;
  } 
  else {
    coeff->n1 = - coeff->p1 + coeff->d1 * coeff->p0;
    coeff->n2 = - coeff->p2 + coeff->d2 * coeff->p0;
    coeff->n3 = - coeff->p3 + coeff->d3 * coeff->p0;
    coeff->n4 = coeff->d4 * coeff->p0;
  }
}
