/* epi_seuil.c 2.1 5/14/98 */

#include <epidaure.h>
#include <epidaure.ee>
#include <epidaure.ei>


#ifndef NO_PROTO
#else 
#endif

/* Ancienne fonction de seuillage par hysteresis.

   Fait appel a E_Hysteresis desormais.
*/

#ifdef _ANSI_
int E_Hysteresys( t_Image *theIm /* image to be thresholded */, 
		 t_Image *resIm /* result image */, 
		 e_ParamConnex *par /* parameters */ )
#else
int E_Hysteresys( theIm, resIm, par )
t_Image *theIm; /* image to be thresholded */
t_Image *resIm; /* result image */
e_ParamConnex *par; /* parameters */
#endif
{
	return(  E_Hysteresis( theIm, resIm, par ) );
}

/* Fonction de seuillage simple.

   On ne conserve que les points dont la valeur est superieure
   ou egale a un seuil bas (champ lowThres de la structure de 
   parametre) si le champ maxSize de la structure de parametre
   est a TRUE, ou dont la valeur est superieure ou egale a un 
   seuil bas (meme champ que precedemment) et inferieure ou
   egale a un seuil haut (champ higThres de la structure de 
   parametre).
   
   Dans l'image resultat, les points verifiant la condition sont 
   mis a (255, 32767, 65535, 1.0) selon le codage de ladite
   image.
*/
   
#ifdef _ANSI_
int E_Seuillage( t_ImagePtr theIm /* image to be thresholded */, 
		 t_ImagePtr resIm /* result image */, 
		 e_ParamConnexPtr par /* parameters */ )
#else
int E_Seuillage( theIm, resIm, par )
t_ImagePtr theIm; /* image to be thresholded */
t_ImagePtr resIm; /* result image */
e_ParamConnexPtr par; /* parameters */
#endif
{
  register int i, j, k, t;
  register float rval;
  float fsl, fsh;
  reeltype val;
  int isl, ish, dim;
  EPI_Imaccess theAccess, resAccess;
  
  E_ERR( (theIm==NULL) || (theIm->buf==NULL), E_ERRNULL, FALSE,
	"E_Seuillage", "input image not allocated");
  E_ERR( (resIm==NULL) || (resIm->buf==NULL), E_ERRNULL, FALSE,
	"E_Seuillage", "output image not allocated");
  E_ERR( (theIm->tx!=resIm->tx) || (theIm->ty!=resIm->ty) || (theIm->tz!=resIm->tz),
	E_ERRFATAL, FALSE, "E_Seuillage", 
	"input and output have different dimensions");
  E_ERR( ((par->maxSize != TRUE) && (par->lowThres > par->higThres)), E_ERRFATAL,
	FALSE, "E_Seuillage", "error in parameters");
  

  switch( resIm->type ) {
  case TYPEIM_256 :
          {
          unsigned char *buf;
	  buf = (unsigned char*)(resIm->buf);
	  dim = theIm->tx * theIm->ty * theIm->tz;
	  fsl = par->lowThres; fsh = par->higThres; 
	  switch ( theIm->type ) {
	  case TYPEIM_256 :
		  {
	          unsigned char *tmp;
		  tmp = (unsigned char*)(theIm->buf);
		  isl = (int)(fsl * 255.0 + 0.5);   ish = (int)(fsh * 255.0 + 0.5);
		  for (i=0; i<dim; i++) {
			  t = *(tmp++);
			  if ( (t >= isl) && ((par->maxSize == TRUE) || (t <= ish)) )
				  *(buf++) |= 0xff;
			  else    *(buf++) &= ~0xff;
		  }}
		  break;
	  case TYPEIM_16B :
		  {
	          unsigned short int *tmp;
		  tmp = (unsigned short int*)(theIm->buf);
		  isl = (int)(fsl * 65535.0 + 0.5);   ish = (int)(fsh * 65535.0 + 0.5);
		  for (i=0; i<dim; i++) {
			  t = *(tmp++);
			  if ( (t >= isl) && ((par->maxSize == TRUE) || (t <= ish)) )
				  *(buf++) |= 0xff;
			  else    *(buf++) &= ~0xff;
		  }}
		  break;
	  case TYPEIM_S16B :
		  {
	          short int *tmp;
		  tmp = (short int*)(theIm->buf);
		  isl = (int)(fsl * 32767.0 + 0.5);   ish = (int)(fsh * 32767.0 + 0.5);
		  for (i=0; i<dim; i++) {
			  t = *(tmp++);
			  if ( (t >= isl) && ((par->maxSize == TRUE) || (t <= ish)) )
				  *(buf++) |= 0xff;
			  else    *(buf++) &= ~0xff;
		  }}
		  break;
	  case TYPEIM_FLOAT :
		  {
		  float *tmp;
		  tmp = (float*)(theIm->buf);
		  for (i=0; i<dim; i++) {
			  rval = *(tmp++);
			  if ( (rval >= fsl) && ((par->maxSize == TRUE) || (rval <= fsh)) )
				  *(buf++) |= 0xff;
			  else    *(buf++) &= ~0xff;
		  }}
		  break;
	  default :
		  for (k=0; k<theIm->tz; k++)
		  for (j=0; j<theIm->ty; j++)
		  for (i=0; i<theIm->tx; i++) {
			  val = T_Get3Point(i, j, k, theIm);
			  if ( (val >= par->lowThres) && ((par->maxSize) || (val <= par->higThres)) )
				  *(buf++) |= 0xff;
			  else
				  *(buf++) &= ~0xff;
		  }
	  }}
	  return( TRUE );
  case TYPEIM_16B :
          {
          unsigned short int *buf;
	  buf = (unsigned short int*)(resIm->buf);
	  dim = theIm->tx * theIm->ty * theIm->tz;
	  fsl = par->lowThres; fsh = par->higThres; 
	  switch ( theIm->type ) {
	  case TYPEIM_256 :
		  {
	          unsigned char *tmp;
		  tmp = (unsigned char*)(theIm->buf);
		  isl = (int)(fsl * 255.0 + 0.5);   ish = (int)(fsh * 255.0 + 0.5);
		  for (i=0; i<dim; i++) {
			  t = *(tmp++);
			  if ( (t >= isl) && ((par->maxSize == TRUE) || (t <= ish)) )
				  *(buf++) = (unsigned short int)65535;
			  else    *(buf++) = (unsigned short int)0;
		  }}
		  break;
	  case TYPEIM_16B :
		  {
	          unsigned short int *tmp;
		  tmp = (unsigned short int*)(theIm->buf);
		  isl = (int)(fsl * 65535.0 + 0.5);   ish = (int)(fsh * 65535.0 + 0.5);
		  for (i=0; i<dim; i++) {
			  t = *(tmp++);
			  if ( (t >= isl) && ((par->maxSize == TRUE) || (t <= ish)) )
				  *(buf++) = (unsigned short int)65535;
			  else    *(buf++) = (unsigned short int)0;
		  }}
		  break;
	  case TYPEIM_S16B :
		  {
	          short int *tmp;
		  tmp = (short int*)(theIm->buf);
		  isl = (int)(fsl * 32767.0 + 0.5);   ish = (int)(fsh * 32767.0 + 0.5);
		  for (i=0; i<dim; i++) {
			  t = *(tmp++);
			  if ( (t >= isl) && ((par->maxSize == TRUE) || (t <= ish)) )
				  *(buf++) = (unsigned short int)65535;
			  else    *(buf++) = (unsigned short int)0;
		  }}
		  break;
	  case TYPEIM_FLOAT :
		  {
		  float *tmp;
		  tmp = (float*)(theIm->buf);
		  for (i=0; i<dim; i++) {
			  rval = *(tmp++);
			  if ( (rval >= fsl) && ((par->maxSize == TRUE) || (rval <= fsh)) )
				  *(buf++) = (unsigned short int)65535;
			  else    *(buf++) = (unsigned short int)0;
		  }}
		  break;
	  default :
		  for (k=0; k<theIm->tz; k++)
		  for (j=0; j<theIm->ty; j++)
		  for (i=0; i<theIm->tx; i++) {
			  val = T_Get3Point(i, j, k, theIm);
			  if ( (val >= par->lowThres) && ((par->maxSize) || (val <= par->higThres)) )
				  *(buf++) = (unsigned short int)65535;
			  else
				  *(buf++) = (unsigned short int)0;
		  }
	  }}
	  return( TRUE );
  case TYPEIM_S16B :
          {
          short int *buf;
	  buf = (short int*)(resIm->buf);
	  dim = theIm->tx * theIm->ty * theIm->tz;
	  fsl = par->lowThres; fsh = par->higThres; 
	  switch ( theIm->type ) {
	  case TYPEIM_256 :
		  {
	          unsigned char *tmp;
		  tmp = (unsigned char*)(theIm->buf);
		  isl = (int)(fsl * 255.0 + 0.5);   ish = (int)(fsh * 255.0 + 0.5);
		  for (i=0; i<dim; i++) {
			  t = *(tmp++);
			  if ( (t >= isl) && ((par->maxSize == TRUE) || (t <= ish)) )
				  *(buf++) = (short int)32767;
			  else    *(buf++) = (short int)0;
		  }}
		  break;
	  case TYPEIM_16B :
		  {
	          unsigned short int *tmp;
		  tmp = (unsigned short int*)(theIm->buf);
		  isl = (int)(fsl * 65535.0 + 0.5);   ish = (int)(fsh * 65535.0 + 0.5);
		  for (i=0; i<dim; i++) {
			  t = *(tmp++);
			  if ( (t >= isl) && ((par->maxSize == TRUE) || (t <= ish)) )
				  *(buf++) = (short int)32767;
			  else    *(buf++) = (short int)0;
		  }}
		  break;
	  case TYPEIM_S16B :
		  {
	          short int *tmp;
		  tmp = (short int*)(theIm->buf);
		  isl = (int)(fsl * 32767.0 + 0.5);   ish = (int)(fsh * 32767.0 + 0.5);
		  for (i=0; i<dim; i++) {
			  t = *(tmp++);
			  if ( (t >= isl) && ((par->maxSize == TRUE) || (t <= ish)) )
				  *(buf++) = (short int)32767;
			  else    *(buf++) = (short int)0;
		  }}
		  break;
	  case TYPEIM_FLOAT :
		  {
		  float *tmp;
		  tmp = (float*)(theIm->buf);
		  for (i=0; i<dim; i++) {
			  rval = *(tmp++);
			  if ( (rval >= fsl) && ((par->maxSize == TRUE) || (rval <= fsh)) )
				  *(buf++) = (short int)32767;
			  else    *(buf++) = (short int)0;
		  }}
		  break;
	  default :
		  for (k=0; k<theIm->tz; k++)
		  for (j=0; j<theIm->ty; j++)
		  for (i=0; i<theIm->tx; i++) {
			  val = T_Get3Point(i, j, k, theIm);
			  if ( (val >= par->lowThres) && ((par->maxSize) || (val <= par->higThres)) )
				  *(buf++) = (short int)32767;
			  else
				  *(buf++) = (short int)0;
		  }
	  }}
	  return( TRUE );
  default :	  
          /*   ecriture de l'image de sortie   */
          /*************************************/
          EPI_init_access_point(&theAccess, theIm);
          EPI_init_access_point(&resAccess, resIm);
          for (k=0; k<theIm->tz; k++)
          for (j=0; j<theIm->ty; j++)
          for (i=0; i<theIm->tx; i++) {
	    EPI_inq_current_point(&theAccess, &val);
	    if ( (val >= par->lowThres) && ((par->maxSize) || (val <= par->higThres)) )
	      EPI_set_current_point(&resAccess, 1.0);
	    else
	      EPI_set_current_point(&resAccess, 0.0);
	    EPI_go_to_next_x_point(&theAccess);
	    EPI_go_to_next_x_point(&resAccess);
	  }
  }

  E_ISERR(FALSE,"E_Seuillage",(char*)NULL);
  
  return( TRUE );
}
