/* epi_spline.c 2.2 5/14/98 */
/*==========================================================================
* File : $EPI/src/epi_spline.c
* Module : epi_spline
* $ Id : ? $
* Test file : epi_test.c
* Author : A. Gueziec
* Creation: September 15 1992
* Modification October 30 1992
*
*===========================================================================
*       (c) Copyright 1992, Inria, all rights reserved
*===========================================================================
*
*------------------------------Description----------------------------------
*
* Calcul d'une image de coefficients B-splines en 3-D
*
*-----------------------------Modification----------------------------------
*
*/

/************************************************************************/
/*
        Determination of spline coefficients for images in 3-D

        (c) Andre' Gueziec -- INRIA -- Sep. 1992
*/
/************************************************************************/

#if defined(INC_MIN) /* inclusions minimales */
 /* rajouter -DINC_MIN dans les options de compilation */
 /* GM 12/09/94                                        */
#include <epi_kernel.ei>
#include <epi_spline.ee>
#else /* inclusions maximales */
#include <epidaure.h>
#include <epidaure.ee>
#include <epidaure.ei>
#endif

/*------- Definition of static functions ----------*/
#ifndef NO_PROTO
static void S_InitLS(sPtr SP, int dim1x, int dim1y, int dim1z, 
		     int dimcx, int dimcy, int dimcz, 
		     int derx, int dery, int derz, 
		     float taux, float tauy, float tauz);
static void S_ComputeUUt( sPtr SP, int *Inf, float *Bsp, register float *UUt,
			 int dim1, int dimc, int der, float tau);
static void S_InitDegreeCoef( sPtr SP);
static void S_InitUBsp( sPtr SP, int dim1x, int dim1y, int dim1z, int dimcx, int dimcy, int dimcz);
static void S_UVW( float *U, int dim);
static void S_ParametrizeU( sPtr SP, register float *UNIF, register int dim1x, register int dimcx);
static void S_ComputeU( sPtr SP, register float *U, register float *UNIF, 
		       register float *Bsp, int *Inf, int dim1, int dimc);
static void S_EvalB_Spline( register int inf, register int deg, 
			   float *U, register float *BSP, 
			   register float val);
static void init_B( float *Bsp, int desQ);
static void S_AllocUBsp( sPtr SP, int dim1x, int dim1y, int dim1z, int dimcx, int dimcy, int dimcz);
static  void S_AllocLS( sPtr SP, int dim1x, int dim1y, int dim1z, int dimcx, int dimcy, int dimcz);
static void S_FreeLS( sPtr SP);
static void S_WriteSystem( sPtr SP, t_ImagePtr imIn, t_ImagePtr imOut);
static void S_SolveSystem( sPtr SP, t_ImagePtr imOut);
static int S_Chobandet( int *n, int *m, 
		       float *a, int *ia, float *rl, 
		       int *il, int *mm, double *d1,
		       int *id, int *lfail);
static int S_Chobandsol( int *n, int *m, int *ir, 
			float *rl, int *il, int *m1, 
			float *b, int *ib, float *x, int *ix);
static void ES_Coefs2Spline( t_ImagePtr imCtrl, t_ImagePtr imSpl);
static void S_PolyInit( float *x, float *U, 
		       float thres, int nP, int *nC, int *I, 
		       int siz, int deg, int comp_fact);
static float S_Dist( float *Pt1, float *Pt2);
static void S_PolyTree( int ib, int ie, float *x, float *U, float *l, int cdep,
		  int *I, int *nVert, int nVertM, int siz, int depth, float thres);
static void S_PtSegDist( float *Pt, float *A, float *B, float *AB, 
			float *dv, float *dh, float lpoly);
static void S_InitSplCurv( sPtr SP, int dimu, int dimcu, short close_u);
static void S_AllocSplCurv( sPtr SP, int dimu, int dimcu);
static void S_FreeSplCurv( sPtr SP);
static void S_ParamCurve( sPtr SP, register int dimcu, short close);
static int S_InitLSCurv( sPtr SP, int dimu, int dimcu, int dimcircu, 
			int siz, int deru);
static void S_AllocLSCurv( sPtr SP, int dimcu, int siz);
static void S_FreeLSCurv( sPtr SP);
static void S_WSystCurv( sPtr SP, float *x, int nP, int nCirc, int siz, int deriv);
static void S_ComputeCircUUt( sPtr SP, int *Inf, float *Bsp, register float *UUt, 
			     int dim1, int dimc, int der, float tau_u);
static void S_CircBandUUt( sPtr SP, int dim);
static void S_SolvCurv( sPtr SP, int dimcu, int dimcircu, int siz);
static void S_DiffVectFieldNe1e2( sPtr SP, cSPtr CP, int nP, int nCirc, float *e1x);
static void S_VectProd( register float *u, register float *v, register float *w);
static void S_ScalProd( register float *u, register float *v, 
		       register float *res, register int n);
static void S_Norm( register float *u, register float *res, register int n);
static void S_Normalise( register float *u, register int n);
static void S_Multiply( register float *u, 
		       register float lambda,
		       register int n);
static void S_LinSamp( sPtr SP, int dimu, int dimcu);
static  void S_EvalN( sPtr SP, cSPtr CP, int nP, int nCirc);
static void S_EvalXu( sPtr SP, cSPtr CP, int nP, int nCirc);
static void S_UpdaNormConv( sPtr SP, cSPtr CP, 
			   int nP, int nCirc, 
			   float *cos_psi_max, float *cos2_psi_moy, 
			   float lambda, short close);
static void S_AllCurvSurfDer( sPtr SP, cSPtr CP, int nP, int nCirc);
static void S_VectDiffMult( register float *u, register float *v,
			   register float lambda, register int n) /* u=u-lambda*v */;
static void S_CurvSurfSelPar( sPtr SP, cSPtr CP, int nP);
static void S_Multiply_in( register float *u, register float *v, 
			  register float lambda, register int n);
static void S_Angle( register float *u, register float *v,
		    register float *res, int n) /* n should be 3, otherwise ? */;
static void S_3Det( register float *u, register float *v, 
		   register float *w, register float *res);
#else 
static void S_InitLS();
static void S_ComputeUUt();
static void S_InitDegreeCoef();
static void S_InitUBsp();
static void S_UVW();
static void S_ParametrizeU();
static void S_ComputeU();
static void S_EvalB_Spline();
static void init_B();
static void S_AllocUBsp();
static  void S_AllocLS();
static void S_FreeLS();
static void S_WriteSystem();
static void S_SolveSystem();
static int S_Chobandet();
static int S_Chobandsol();
static void ES_Coefs2Spline();
static void S_PolyInit();
static float S_Dist();
static void S_PolyTree();
static void S_PtSegDist();
static void S_InitSplCurv();
static void S_AllocSplCurv();
static void S_FreeSplCurv();
static void S_ParamCurve();
static int S_InitLSCurv();
static void S_AllocLSCurv();
static void S_FreeLSCurv();
static void S_WSystCurv();
static void S_ComputeCircUUt();
static void S_CircBandUUt();
static void S_SolvCurv();
static void S_DiffVectFieldNe1e2();
static void S_VectProd();
static void S_ScalProd();
static void S_Norm();
static void S_Normalise();
static void S_Multiply();
static void S_LinSamp();
static void S_EvalN();
static void S_EvalXu();
static void S_UpdaNormConv();
static void S_AllCurvSurfDer();
static void S_VectDiffMult();
static void S_CurvSurfSelPar();
static void S_Multiply_in();
static void S_Angle();
static void S_3Det();
#endif




/*=======================================================================
* int E_CoefSpline(        t_ImagePtr imIn,
*                                          t_ImagePtr imOut,
*                                          float sx,float sy,float sz)
*
* Action :
* Calcul des coefficients B_splines relatifs a` une
* image tri-dimensionnelle pour la re'e'chantillonner
* pour interpoler par des fonctions cubiques entre les voxels
* pour calculer des derivees ou des integrales
*
* Inclure dans l'ordre :
* <epidaure.h>
* <spline.h>
*
* Parametres de sortie :
* imOut : buffer re'sultat des coefficients
*
* Parametres d'entree :
* imIn : Image a approximer ou a interpoler par des B_splines volumiques
* imOut->tx   : nombre de coefficient en x
* imOut->ty   : nombre de coefficient en y
* imOut->tz   : nombre de coefficient en z
* sx   : coefficient de lissage suivant x
* sy   : coefficient de lissage suivant y
* sz   : coefficient de lissage suivant z
*
* Parametres d'entree/sortie :
* aucun.
*
* Retour :
* 0 si une erreur c'est produite, 1 si tout c'est bien passe.
*
* Erreurs :
* pas assez de memoire 
* dans le cas de l'interpolation, coefficients de lissage nuls
* 
* Observations :
* Les 2 images imOut et imIn doivent etre differentes et doivent
* etre allouees en entree
* 
* imOut->tx=dimcx,imOut->ty=dimcy,imOut->tz=dimcz sont les dimension
* de l'image de coefficients en sortie
* si ces dimensions sont les memes que l'image de depart sx,sy et sz
* doivent necessairement >0. pour la stabilite numerique
* des valeurs faibles 1e-2, 1e-3 .. suffiront
* 
* Exemple :
* t_Image imIn,imOut;
*
* T_Init3Image( &imIn, 0,0,0,TYPEIM_UNKNOWN, "image.file" );
* T_ReadInrim( &imIn );
* T_Init3Image( &imOut, dimcx,dimcy,dimcz,TYPEIM_FLOAT, "coef_image.file" );
* T_AllocImage( &imOut);
* E_CoefSpline( &imIn,&imOut,sx,sy,sz);
* T_WriteInrim( &imOut );
* T_FreeImage( &imIn );
* T_FreeImage( &imOut );
*
*========================================================================
*/

static int c__1 = 1;

/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
int E_CoefSpline(t_ImagePtr imIn,t_ImagePtr imOut,float sx,float sy,float sz)
#else
int E_CoefSpline(imIn, imOut, sx, sy, sz)
t_ImagePtr imIn;
t_ImagePtr imOut;
float sx;
float sy;
float sz;

#endif

{
	sPar_ param_S;
	
	/* cubic B_spline functions */
	param_S.De=4;

	ES_InitSpline(&param_S,imIn->tx,imIn->ty,imIn->tz,
		     imOut->tx,imOut->ty,imOut->tz);

	S_InitLS(&param_S,imIn->tx,imIn->ty,imIn->tz,
		 imOut->tx,imOut->ty,imOut->tz,
		 (int)0,(int)0,(int)0,sx,sy,sz);

	S_WriteSystem(&param_S,imIn,imOut);
       
	S_SolveSystem(&param_S,imOut);
	

	ES_FreeSpline(&param_S);
	S_FreeLS(&param_S);
	
	return (1);
}
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
void ES_InitSpline( sPtr SP, int dim1x, int dim1y, int dim1z, int dimcx, int dimcy, int dimcz)
#else
void ES_InitSpline(SP, dim1x, dim1y, dim1z, dimcx, dimcy, dimcz)
sPtr SP;
int dim1x;
int dim1y;
int dim1z;
int dimcx;
int dimcy;
int dimcz;

#endif

{
	S_InitDegreeCoef(SP);
	S_AllocUBsp(SP,dim1x,dim1y,dim1z,dimcx,dimcy,dimcz);
	E_ISERRN( "ES_InitSpline", NULL );

	S_InitUBsp(SP,dim1x,dim1y,dim1z,dimcx,dimcy,dimcz);
	return;
}
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
static void S_InitLS(sPtr SP, int dim1x, int dim1y, int dim1z, 
		int dimcx, int dimcy, int dimcz, 
		int derx, int dery, int derz, 
		float taux, float tauy, float tauz)
#else
static void S_InitLS(SP, dim1x, dim1y, dim1z, dimcx, dimcy, dimcz, derx, dery, derz, taux, tauy, tauz)
sPtr SP;
int dim1x;
int dim1y;
int dim1z;
int dimcx;
int dimcy;
int dimcz;
int derx;
int dery;
int derz;
float taux;
float tauy;
float tauz;

#endif

{
	S_AllocLS(SP,dim1x,dim1y,dim1z,dimcx,dimcy,dimcz);
	E_ISERRN( "S_InitLS", NULL );

	
	S_ComputeUUt(SP,SP->Inf,SP->BspU,SP->UUt,dim1x,dimcx,derx,taux);
	
	if(SP->UNIF!=SP->VNIF)
	      S_ComputeUUt(SP,SP->Jnf,SP->BspV,SP->VVt,dim1y,dimcy,dery,tauy);

	if(SP->UNIF!=SP->WNIF && SP->VNIF!=SP->WNIF)
	      S_ComputeUUt(SP,SP->Knf,SP->BspW,SP->WWt,dim1z,dimcz,derz,tauz);
}
/*___________________________________________________________________________*/

#if defined(_ANSI_)
static void S_ComputeUUt( sPtr SP, int *Inf, float *Bsp, register float *UUt,
			 int dim1, int dimc, int der, float tau)
#else
static void S_ComputeUUt(SP, Inf, Bsp, UUt, dim1, dimc, der, tau)
sPtr SP;
int *Inf;
float *Bsp;
register float *UUt;
int dim1;
int dimc;
int der;
float tau;

#endif

{
	register int f;int e,i,ind_u,iu,dim=dimc-1;
	int mc=SP->m*dimc,mc1=SP->De*dimc,mc2=mc1+mc1;
	register float *Bsp0,*BspD0;float bu,bud,bd,*BspD,magder,tmpmag;

	tmpmag=(float)(dimc+1-SP->De);
	tmpmag=(tmpmag<=1.0)?(float)1.: (float)(dim1-1)/tmpmag;
	for(magder=1.,e=SP->m;e>der;e--){magder*=tmpmag;}
	/* order of magnitude for the <<der .th>> derivative*/
	
	for(Bsp0=UUt,i=0;i<mc2;i++) *Bsp0++ =0.;

    /*----------         special way to fill the array     ------------------*/
    
    /* original symmetric matrix a(oe,of)                                    */
    /*   a00 a01 a02 */ /*  stored :   */ /*   x   x   a00 */
    /*   a10 a11 a12 */ /*   \         */ /*   x   a10 a11 */
    /*   a20 a21 a22 */ /*     \       */ /*   a20 a21 a22 */

    /* new Indexing   A(ce,cf) c stands for Choleski                         */
    /*   x   x   A02 */ /*  correspondance : */
    /*   x   A11 A12 */ /*   SAME ROW        */  /* ce    =    oe */
    /*   A20 A21 A22 */ /*   COLUMN          */  /* cf    =   of+m-oe  */

    /*----------                    classiquely            ------------------*/
    /* we step down the diagonal of the matrix by taking the lowest Bj!=0    */
    /* and then we update all cells jk with bjbk!=0                          */
    /*         \                                                             */
    /*          \                                                            */
    /*           \                                                           */
    /*            \________________                                          */
    /*            |BjBj                                                      */
    /*            |Bj+1Bj  Bj+1Bj+1                                          */
    /*            |Bj+2Bj  Bj+2Bj+1                                          */
    /*            |                                                          */

   
    /*----------                    now                    ------------------*/
    /* we went down the m+1 column and expand diagonally                     */

    /*       x  x /|                                                         */
    /*       x   / |                                                         */
    /*        _/_ _|    size of triangle : m+1                               */
    /*            /|                                                         */
    /*           / |                                                         */
    /*        _/_ _|                                                         */
    /*             |                                                         */
	
    /*----------                       UUt                 ------------------*/

      for(BspD=Bsp+der*SP->De,Bsp+=SP->des0,
	  i=0;i<dim1;i++,Bsp+=SP->desQ,BspD+=SP->desQ,Inf++){

	    for(e=0,ind_u= *Inf +mc; e<SP->De; e++, ind_u++) 

		 {bu= *(Bsp+e);

		  bud= *(BspD+e);

		 for(f=e,iu=ind_u,Bsp0=Bsp+e,BspD0=BspD+e;
		     f<SP->De;f++,iu-=dim)  
		   
			 {bd= bud* *(BspD0++) *magder;

			  *(UUt+iu+mc1) += bd;
       
			  *(UUt+iu) += bu * *(Bsp0++)
				    +  bd * tau;}}/*end for f,e */

        }/* end for i */
    /*-----------     highest derivative : last columns of UUt    -----------*/
    /*                            2 column                                 */

    /*       x  x /| x  x /|                                               */
    /*       x   / | x   / |                                               */
    /*        _/_ _|  _/_ _|                                               */
  return;
}
/*___________________________________________________________________________*/

#if defined(_ANSI_)
static void S_InitDegreeCoef( sPtr SP)
#else
static void S_InitDegreeCoef(SP)
sPtr SP;

#endif

{
	SP->m=SP->De-1;
	
	SP->desQ=SP->De*SP->De;
	SP->des0=SP->desQ-SP->De;
	SP->des1=SP->des0-SP->De;
	SP->des2=SP->des1-SP->De;
	SP->des3=SP->des2-SP->De;
	return;
}
/*___________________________________________________________________________*/

#if defined(_ANSI_)
static void S_InitUBsp( sPtr SP, int dim1x, int dim1y, int dim1z, int dimcx, int dimcy, int dimcz)
#else
static void S_InitUBsp(SP, dim1x, dim1y, dim1z, dimcx, dimcy, dimcz)
sPtr SP;
int dim1x;
int dim1y;
int dim1z;
int dimcx;
int dimcy;
int dimcz;

#endif

{
	
	S_UVW(SP->U,dim1x);S_UVW(SP->V,dim1y);S_UVW(SP->W,dim1z);

	S_ParametrizeU(SP,SP->UNIF,dim1x,dimcx);

	if(SP->UNIF!=SP->VNIF)
		S_ParametrizeU(SP,SP->VNIF,dim1y,dimcy);

	if(SP->UNIF!=SP->WNIF && SP->VNIF!=SP->WNIF)
		S_ParametrizeU(SP,SP->WNIF,dim1z,dimcz);

	
	S_ComputeU(SP,SP->U,SP->UNIF,SP->BspU,SP->Inf,dim1x,dimcx);

	if(SP->UNIF!=SP->VNIF)
		S_ComputeU(SP,SP->V,SP->VNIF,SP->BspV,SP->Jnf,dim1y,dimcy);

	if(SP->UNIF!=SP->WNIF && SP->VNIF!=SP->WNIF)
		S_ComputeU(SP,SP->W,SP->WNIF,SP->BspW,SP->Knf,dim1z,dimcz);
}
/*___________________________________________________________________________*/

#if defined(_ANSI_)
static void S_UVW( float *U, int dim)
#else
static void S_UVW(U, dim)
float *U;
int dim;

#endif

{
	register int i;
	for(i=0;i<dim;i++)
		*U++ =(float)i;
	return;
}
/*___________________________________________________________________________*/

#if defined(_ANSI_)
static void S_ParametrizeU( sPtr SP, register float *UNIF, register int dim1x, register int dimcx)
#else
static void S_ParametrizeU(SP, UNIF, dim1x, dimcx)
sPtr SP;
register float *UNIF;
register int dim1x;
register int dimcx;

#endif

{
	register int i;
	register float d1x= (float)(dim1x-1);
	register float incu= (float)(dimcx+1-SP->De),valu=0.;

	incu= (incu == 0.0) ? (float)0.0 : d1x/ incu;

	for(i=0;i<SP->De;i++) *UNIF++ =0.;
	
	for(i=SP->De;i<dimcx;i++){valu+=incu;
				  *UNIF++ =valu;}
	
	for(i=dimcx;i<dimcx+SP->De;i++) *UNIF++ =d1x;
	/* rather than valu to prevent roundoff errors */

}
/*___________________________________________________________________________*/

#if defined(_ANSI_)
static void S_ComputeU( sPtr SP, register float *U, register float *UNIF, register float *Bsp, int *Inf, int dim1, int dimc)
#else
static void S_ComputeU(SP, U, UNIF, Bsp, Inf, dim1, dimc)
sPtr SP;
register float *U;
register float *UNIF;
register float *Bsp;
int *Inf;
int dim1;
int dimc;

#endif

{
	int i;
	register int inf=SP->m,dimC=dimc-1;
	for(i=0;i<dim1;i++,Bsp+=SP->desQ,U++,Inf++){
		while (*U >= *(UNIF+1+inf)  && inf<dimC){inf++;}
		*Inf =inf-SP->m;
		S_EvalB_Spline(inf,SP->De,UNIF,Bsp,*U);}

}
/*___________________________________________________________________________*/

/*   SI  : Compute B-spline functions  
                    
       no need to initialise BSP to zero          
*/
/*___________________________________________________________________________*/

#if defined(_ANSI_)
static void S_EvalB_Spline( register int inf, register int deg, 
			   float *U, register float *BSP, 
			   register float val)
#else
static void S_EvalB_Spline(inf, deg, U, BSP, val)
register int inf;
register int deg;
float *U;
register float *BSP;
register float val;

#endif

{
  /* content of matrix BSP[lig][col]~= BSP[lig+deg*col]                      */
  /*                                           *//* slowest varying index || */
  /*___________________________________________*//*                       \/ */

  
  /*                           ..       2nd deriv    1st Deriv     BSPLINE  */

  /* stored columnwise */


  /* d stands for deg */
  /*lig/col 0           1               2            3           ..   d-1    */
 
  /*           ||               ||                                           */
  /*         storage           order                                         */
  /*           \/               \/                               function    */
  /*   |----------------  derivatives  -------------------|         \/       */
  /*0Betad-1inf-d+1,1 Beta3inf-d+1,2 Beta2inf-d+1,3 Betainf-d+1,4..Binf-d+1,d*/

  /* 1  Betad-1 inf-3,1 Beta3 inf-3,2 Beta2 inf-3,3  Beta inf-3,4..Binf-d+2,d*/

  /* 2  Betad-1 inf-2,1 Beta3 inf-2,2 Beta2 inf-2,3  Beta inf-2,4..Binf-d+3,d*/

  /* 3  Betad-1 inf-1,1 Beta3 inf-1,2 Beta2 inf-1,3  Beta inf-1,4 ..Binf-1,d*/

  /*      ..             ..        ..        ..       ..                     */

  /* d-1  Betad-1 inf,1  Beta3 inf,2    Beta2 inf,3    Beta inf,4   ..Binf,d */


  register int cdd,d_deg,lig,col;
  int dcol,dc,ind_beg;  
  float omega;
  register float tmp,tmpB;

  if ( val < *(U+inf) || val > *(U+inf+1)){init_B(BSP,deg*deg);return;}

  *(BSP+deg-1)=1.0;/*  BSPinf,1 */

  for (dc=deg,dcol=deg-1,col=1;col<deg;col++,dc+=deg){
    dcol--;
    for(d_deg=cdd=0;cdd<=col;cdd++,d_deg+=deg)  *(BSP+dcol+d_deg)=0.0;
    ind_beg=inf-col; /* = 2 */
    for(lig=dcol;lig<deg-1;lig++){ 
      ind_beg++;
      omega=  *(U+ind_beg+col) - *(U+ind_beg);
      tmp=(omega == 0.0) ? (float)0.0 : (float)col/ omega;
      omega=(omega == 0.0) ? (float)0.0 : (val- *(U+ind_beg))/omega;
      
      *(BSP+lig+dc)  += (float)((1.0-omega) * *(BSP+lig+1+dc-deg));
      *(BSP+lig+1+dc) = omega *  *(BSP+lig+1+dc-deg);
      
      
      for(d_deg=cdd=0;cdd<col;cdd++,d_deg+=deg)
	{/*DERIVATIVES*/
	  tmpB=  *(BSP+lig+1+d_deg)*tmp;
	  *(BSP+lig+d_deg)-= tmpB ; 
	  *(BSP+lig+1+d_deg)= tmpB ;
	}
    }
  }
}
/*___________________________________________________________________________*/

/*         initialise B-spline construction array                            */
/*___________________________________________________________________________*/

#if defined(_ANSI_)
static void init_B( float *Bsp, int desQ)
#else
static void init_B(Bsp, desQ)
float *Bsp;
int desQ;

#endif

{
	register int k;
	for (k=0;k<desQ;k++) *Bsp++ = 0.0;
	return;
}
/*___________________________________________________________________________*/

/*        SI  : Allocations                                                  */

/*___________________________________________________________________________*/

#if defined(_ANSI_)
static void S_AllocUBsp( sPtr SP, int dim1x, int dim1y, int dim1z, int dimcx, int dimcy, int dimcz)
#else
static void S_AllocUBsp(SP, dim1x, dim1y, dim1z, dimcx, dimcy, dimcz)
sPtr SP;
int dim1x;
int dim1y;
int dim1z;
int dimcx;
int dimcy;
int dimcz;

#endif

{
	/*     parametrization                                             */
	
	SP->U= NULL;
	SP->V= NULL;
	SP->W= NULL;

	SP->UNIF= NULL;
	SP->VNIF= NULL;
	SP->WNIF= NULL;

	SP->Inf= NULL;
	SP->Jnf= NULL;
	SP->Knf= NULL;

	SP->BspU= NULL;
	SP->BspV= NULL;
	SP->BspW= NULL;

	SP->CspU= NULL;
	SP->CspV= NULL;
	SP->CspW= NULL;

	SP->U=(float *) calloc(dim1x,sizeof(float));
	SP->V=(float *) calloc(dim1y,sizeof(float));
	SP->W=(float *) calloc(dim1z,sizeof(float));

	if(SP->U==NULL || SP->V==NULL || SP->W==NULL){

		E_SETERR( E_ERRMEM, "calloc", "no more memory" );
		E_REPERR( goto END_S_AllocUBsp );
	}

	SP->UNIF=(float *) calloc(dimcx+SP->De,sizeof(float));

	SP->VNIF= (dim1x==dim1y && dimcx==dimcy) ? SP->UNIF :
		(float *) calloc(dimcy+SP->De,sizeof(float));

	SP->WNIF= (dim1x==dim1z && dimcx==dimcz) ? SP->UNIF :
		  (dim1y==dim1z && dimcy==dimcz) ? SP->VNIF :
		(float *) calloc(dimcz+SP->De,sizeof(float));

	if(SP->UNIF==NULL || SP->VNIF==NULL || SP->WNIF==NULL)
		{
			E_SETERR( E_ERRMEM, "calloc", 
				 "UNIF-VNIF-WNIF bad alloc" );
			E_REPERR( goto END_S_AllocUBsp );}

	/*     Index                                                       */
	

	SP->Inf=(int *) calloc(dim1x,sizeof(int));

	SP->Jnf=  (SP->UNIF==SP->VNIF) ? SP->Inf  :
		(int *) calloc(dim1y,sizeof(int));

	SP->Knf=  (SP->UNIF==SP->WNIF) ? SP->Inf  :
		  (SP->VNIF==SP->WNIF) ? SP->Jnf  :
		(int *) calloc(dim1z,sizeof(int));

	if(SP->Inf==NULL || SP->Jnf==NULL || SP->Knf==NULL)
		{
			E_SETERR( E_ERRMEM, "calloc", 
				 "Inf-Jnf-Knf bad alloc" );
			E_REPERR( goto END_S_AllocUBsp );}


	/*    Bspline functions                                            */

	SP->BspU=(float *) calloc(dim1x*SP->desQ,sizeof(float));
	
	SP->BspV=  (SP->UNIF==SP->VNIF) ? SP->BspU :
		(float *) calloc(dim1y*SP->desQ,sizeof(float));

	SP->BspW=  (SP->UNIF==SP->WNIF) ? SP->BspU :
		   (SP->VNIF==SP->WNIF) ? SP->BspV :
		 (float *) calloc(dim1z*SP->desQ,sizeof(float));

	if(SP->BspU==NULL || SP->BspV==NULL || SP->BspW==NULL)
		{
			E_SETERR( E_ERRMEM, "calloc", 
				 "Bsp bad alloc" );
			E_REPERR( goto END_S_AllocUBsp );}

	SP->CspU=(float *) calloc(SP->desQ,sizeof(float));
	
	SP->CspV= (float *) calloc(SP->desQ,sizeof(float));

	SP->CspW=  (float *) calloc(SP->desQ,sizeof(float));

	if(SP->CspU==NULL || SP->CspV==NULL || SP->CspW==NULL)
		{
			E_SETERR( E_ERRMEM, "calloc", 
				 "Csp bad alloc" );
			E_REPERR( goto END_S_AllocUBsp );}
	return;

END_S_AllocUBsp:

	E_REPERR(  ES_FreeSpline(SP) );
	E_ISERRN( "S_AllocUBsp", NULL );
	return;
}

/*___________________________________________________________________________*/

#if defined(_ANSI_)
static  void S_AllocLS( sPtr SP, int dim1x, int dim1y, int dim1z, int dimcx, int dimcy, int dimcz)
#else
static  void S_AllocLS(SP, dim1x, dim1y, dim1z, dimcx, dimcy, dimcz)
sPtr SP;
int dim1x;
int dim1y;
int dim1z;
int dimcx;
int dimcy;
int dimcz;

#endif

{
	int m2=SP->De*2;
	int dimc=(dimcx>dimcy)?dimcx:dimcy;

	if (dimcz>dimc) dimc=dimcz;/* highest dimension*/
	
	SP->UUt= NULL;
	SP->VVt= NULL;
	SP->WWt= NULL;

	SP->LLt= NULL;
	SP->VUt= NULL;

	SP->UUt= (float *) calloc(dimcx*m2,sizeof(float));
	
	SP->VVt= (SP->UNIF==SP->VNIF) ? SP->UUt :
		(float *) calloc(dimcy*m2,sizeof(float));

	SP->WWt= (SP->UNIF==SP->WNIF) ? SP->UUt :
		(SP->VNIF==SP->WNIF) ? SP->VVt :
		(float *) calloc(dimcz*m2,sizeof(float));

	
	/* result of  choleski decomposition                   */
	SP->LLt= (float *) calloc( dimc*SP->De,sizeof(float));

	SP->VUt= (float *) calloc( dimcx*dimcy*dimcz,sizeof(float));

	if(SP->UUt==NULL || SP->VVt==NULL || SP->WWt==NULL || 
	   SP->LLt==NULL)
		{
			E_SETERR( E_ERRMEM, "calloc", 
				 "UUt-VVt-WWt-LLt-VUt bad alloc" );
			E_REPERR( goto END_S_AllocLS );}

	
/*
E_ERRN( SP->VUt==NULL, E_ERRMEM, "S_AllocLS", NULL );
*/

	/* after choleski decomposition                        */
	/*SP->RU is stored in SP->UUt                          */
	/*SP->RV is stored in SP->VVt                          */
	
	/* we need to solve 2 following systems                */
	/* VVt Gv = VUt                                        */
	/* UUt Guvt =Gvt                                       */
	/* AX = B is solved so that X is stored in B           */
	/* We need an other buffer than VUt, say Gvt
	   to make transpositions but
	   the program is called with a buffer in imOut
	   so we do not need to declare Gvt */
	return;

END_S_AllocLS:
	E_REPERR(  S_FreeLS(SP)  );
	E_ISERRN( "S_AllocLS", NULL );
	return;


}
/*___________________________________________________________________________*/
/* 
       SI  : free buffers
*/
/*___________________________________________________________________________*/

#if defined(_ANSI_)
void ES_FreeSpline( sPtr SP)
#else
void ES_FreeSpline(SP)
sPtr SP;

#endif

{
	T_Free((void **) &SP->U);
	T_Free((void **) &SP->V);
	T_Free((void **) &SP->W);
	
	if((SP->WNIF!=SP->VNIF) && (SP->WNIF != SP->UNIF))
		{T_Free((void **) &SP->WNIF);
		 T_Free((void **) &SP->Knf);
		 T_Free((void **) &SP->BspW);}

	if((SP->VNIF != SP->UNIF))
		{T_Free((void **) &SP->VNIF);
		 T_Free((void **) &SP->Jnf);
		 T_Free((void **) &SP->BspV);}

	T_Free((void **) &SP->UNIF);
	T_Free((void **) &SP->Inf);
	T_Free((void **) &SP->BspU);
	
	T_Free((void **) &SP->CspU);
	T_Free((void **) &SP->CspV);
	T_Free((void **) &SP->CspW);
	
	return;
}
/*___________________________________________________________________________*/

#if defined(_ANSI_)
static void S_FreeLS( sPtr SP)
#else
static void S_FreeLS(SP)
sPtr SP;

#endif

{
	if((SP->WNIF!=SP->VNIF) && (SP->WNIF != SP->UNIF))
		{T_Free((void **) &SP->WWt);}

	if((SP->VNIF != SP->UNIF)){T_Free((void **) &SP->VVt);}

	T_Free((void **) &SP->UUt);
	
	T_Free((void **) &SP->LLt);
	T_Free((void **) &SP->VUt);

	return;
}
/*___________________________________________________________________________*/
/* 
       SI  : write spline image system
*/
/*___________________________________________________________________________*/

#if defined(_ANSI_)
static void S_WriteSystem( sPtr SP, t_ImagePtr imIn, t_ImagePtr imOut)
#else
static void S_WriteSystem(SP, imIn, imOut)
sPtr SP;
t_ImagePtr imIn;
t_ImagePtr imOut;

#endif

{

    int e,g,i,j,k,dimcxz=imOut->tx*imOut->tz,dimcxyz=dimcxz*imOut->ty;

    int dimCx=imOut->tx-1,dimCy=imOut->ty-1,dimCz=imOut->tz-1;

    int inf_i,inf_j,inf_k,DEG_h;

    int offset_v,offset_u;

    int ind_uv,ind_v;

    register int h;

    register float buv, *Bsw;

    float ifl,jfl,kfl,bv, *BSU,*BSV,*BSW,imval;

    register float *VUt;

    
    /*---------------------        initialisation        --------------------*/
    DEG_h=(imIn->tz==1) ? 1 : SP->De;
    for(VUt=SP->VUt,e=0;e<dimcxyz;e++)  *VUt++ =0.;
    

    /*----------                       VUt                 ------------------*/
    for(BSW=SP->BspW+SP->des0,inf_k=SP->m,k=0;k<imIn->tz;k++,BSW+=SP->desQ){
         kfl= (float)k;
         while(kfl>= *(SP->WNIF+1+inf_k) && inf_k<dimCz ){inf_k++;}

         for(BSV=SP->BspV+SP->des0,inf_j=SP->m,offset_v=j=0;
	     j<imIn->ty;j++,BSV+=SP->desQ){
	     jfl= (float)j;
	     while(jfl>= *(SP->VNIF+1+inf_j)&&inf_j<dimCy)
		     {inf_j++;offset_v+=dimcxz;}

	     for(BSU=SP->BspU+SP->des0,inf_i=SP->m,offset_u=0,
		 i=0;i<imIn->tx;i++,BSU+=SP->desQ){
		 ifl= (float)i;
		 while(ifl>= *(SP->UNIF+1+inf_i) && inf_i<dimCx)
				  {inf_i++;offset_u+=imOut->tz;}
		 imval= (float)T_Get3Point( i, j, k, imIn );

		  for(g=0,ind_v=offset_v; g<SP->De; g++, ind_v+=dimcxz) 

			  {bv= *(BSV+g);

			   for(e=0,ind_uv=ind_v+offset_u;
			       e<SP->De; e++, ind_uv+=imOut->tz)

				   {buv=bv* *(BSU+e);
					     
				    for(h=0,VUt=SP->VUt+ind_uv+inf_k-SP->m,
					Bsw=BSW;h<DEG_h;h++)
					   {/*        REGISTER	     */      
					    *(VUt++) += buv * *(Bsw++) * 
						    imval;}}}
			  
		   }/* end for  i */

	 }/* end for j */
    
     }/* end for k */
   return;
}
/*___________________________________________________________________________*/
/* 
   SI : solve linear system 3 times 
*/
/*___________________________________________________________________________*/
#if defined(_ANSI_)
static void S_SolveSystem( sPtr SP, t_ImagePtr imOut)
#else
static void S_SolveSystem(SP, imOut)
sPtr SP;
t_ImagePtr imOut;

#endif

{
	/*                       3 D - S P A C E                       */
	/* ____________                                                */ 
	/* |x x x x x x                                                */
	/* |z y                                                        */
	/* |z   y                                                      */
	/* |z     y                                                    */
	/* |z       y                ...                               */

	/* ____________          write as                              */ 
	/* |z z z z z z                                                */
	/* |y x                                                        */
	/* |y   x                                                      */
	/* |y     x                                                    */
	/* |y       x                ...                               */

	/* ____________          finally                               */ 
	/* |y y y y y y                                                */
	/* |x z                                                        */
	/* |x   z                                                      */
	/* |x     z                                                    */
	/* |x       z                                                  */

	
	int dimcxy=imOut->tx*imOut->ty,dimcxz=imOut->tx*imOut->tz;
	int dimcyz=imOut->ty*imOut->tz;
	int i,j,k,ii,jj;
	int iFail=0;
	int idetUUt,idetVVt,idetWWt;
	double detUUt,detVVt,detWWt;
	register float *TheGv;
	float *TheImbuf=(float *)imOut->buf;

	/*______________      solve WWt Gv = VUt         ______________*/

	/*...............      Choleski decomposition     .............*/

	S_Chobandet(&imOut->tz,&SP->m,SP->WWt,&imOut->tz,SP->LLt,&imOut->tz,
		    &SP->De,&detWWt,&idetWWt,&iFail);

	if (iFail!=0){fprintf(stderr,"s.solve.c f03agf choleski ::W  FAILS");}
	 
	S_Chobandsol(&imOut->tz,&SP->m,&dimcxy,SP->LLt,&imOut->tz,&SP->De,
		    SP->VUt,&imOut->tz,TheImbuf,&imOut->tz);

	/*...............        transpose matrix          .............*/
	
	for(TheGv=SP->VUt,ii=i=0;i<imOut->tx;i++,ii+=imOut->tz)
		for(k=0;k<imOut->tz;k++)
			for(jj=k+ii,j=0;j<imOut->ty;j++,jj+=dimcxz)

				*(TheGv++) = *(TheImbuf+jj);

	/*______________      solve VVt Guv = Gvt         ______________*/
	
	if(SP->VNIF!=SP->WNIF)
		S_Chobandet(&imOut->ty,&SP->m,SP->VVt,&imOut->ty,SP->LLt,
			    &imOut->ty,&SP->De,&detVVt,&idetVVt,&iFail);

	
	if (iFail!=0){fprintf(stderr,"s.solve.c f03agf choleski ::V  FAILS");}

	S_Chobandsol(&imOut->ty,&SP->m,&dimcxz,SP->LLt,&imOut->ty,&SP->De,
		    SP->VUt,&imOut->ty,TheImbuf,&imOut->ty);

	
	for(TheGv=SP->VUt,jj=k=0;k<imOut->tz;k++,jj+=imOut->ty)
		for(j=0;j<imOut->ty;j++)
			for(ii=j+jj,i=0;i<imOut->tx;i++,ii+=dimcyz)
		
				*(TheGv++) =  *(TheImbuf+ii);

	/*______________      solve UUt Guvt =Gvt       ______________*/
	
	if(SP->UNIF!=SP->VNIF)
		S_Chobandet(&imOut->tx,&SP->m,SP->UUt,&imOut->tx,SP->LLt,
			    &imOut->tx,&SP->De,&detUUt,&idetUUt,&iFail);

	
	if (iFail!=0){fprintf(stderr,"s.solve.c f03agf choleski ::U  FAILS");}
	
	S_Chobandsol(&imOut->tx,&SP->m,&dimcyz,SP->LLt,&imOut->tx,&SP->De,
		    SP->VUt,&imOut->tx,TheImbuf,&imOut->tx);

	
	return;
}
/*___________________________________________________________________________*/
/* 
   SI : choleski decomposition 
*/
/*___________________________________________________________________________*/
#if defined(_ANSI_)
static int S_Chobandet( int *n, int *m, 
		       float *a, int *ia, float *rl, 
		       int *il, int *mm, double *d1,
		       int *id, int *lfail)
#else
static int S_Chobandet(n, m, a, ia, rl, il, mm, d1, id, lfail)
int *n;
int *m;
float *a;
int *ia;
float *rl;
int *il;
int *mm;
double *d1;
int *id;
int *lfail;

#endif

{
    /* System generated locals */
    int a_dim1, a_offset, rl_dim1, rl_offset, i_1, i_2, i_3;

    /* Local variables */
    static int i, j, k;
    static float y;
    static int m1, ip, ir, is, iq;

    /* Parameter adjustments */
    a_dim1 = *ia;
    a_offset = a_dim1 + 1;
    a -= a_offset;
    rl_dim1 = *il;
    rl_offset = rl_dim1 + 1;
    rl -= rl_offset;

    /* Function Body */


    *lfail = 0;
    *d1 = 1.;
    *id = 0;
    m1 = *m + 1;
    i_1 = *n;
    for (i = 1; i <= i_1; ++i) {

/* Computing MAX */
	i_2 = 1, i_3 = *m - i + 2;
	ip = (i_3 > i_2) ? i_3 : i_2;
	ir = i - *m + ip - 1;
	i_2 = m1;
	for (j = ip; j <= i_2; ++j) {
	    is = j - 1;
	    iq = *m + 1 - j + ip;
	    y = a[i + j * a_dim1];
	    if (is < ip) {
		goto L40;
	    }
	    i_3 = is;
	    for (k = ip; k <= i_3; ++k) {
		y -= rl[i + k * rl_dim1] * rl[ir + iq * rl_dim1];
		++iq;
/* L20: */
	    }
L40:
	    if (j == m1) {
		goto L60;
	    }
	    rl[i + j * rl_dim1] = y * rl[ir + m1 * rl_dim1];
	    goto L180;
L60:
	    *d1 *= (double)y;
	    if (y != 0.) {
		goto L80;
	    }
	    *id = 0;
	    goto L140;
L80:
	    if (fabs(*d1) < 1.) {
		goto L100;
	    }
	    *d1 *= .0625;
	    *id += 4;
	    goto L80;
L100:
	    if (fabs(*d1) >= .0625) {
		goto L120;
	    }
	    *d1 *= 16.;
	    *id -= 4;
	    goto L100;
L120:
	    if (y >= 0.) {
		goto L160;
	    }
L140:
	    *lfail = c__1;
	    fprintf(stderr,"the matrix is not positive definite\n");
	    return 0;
L160:
	    rl[i + j * rl_dim1] = (float)(1. / sqrt(y));
L180:
	    ++ir;
/* L200: */
	}
/* L220: */
    }
    return 0;
} 
/*___________________________________________________________________________*/
/* 
   SI : choleski solution
*/
/*___________________________________________________________________________*/
#if defined(_ANSI_)
static int S_Chobandsol( int *n, int *m, int *ir, 
			float *rl, int *il, int *m1, 
			float *b, int *ib, float *x, int *ix)
#else
static int S_Chobandsol(n, m, ir, rl, il, m1, b, ib, x, ix)
int *n;
int *m;
int *ir;
float *rl;
int *il;
int *m1;
float *b;
int *ib;
float *x;
int *ix;

#endif

{
    /* System generated locals */
    int rl_dim1, rl_offset, b_dim1, b_offset, x_dim1, x_offset, i_1, i_2, 
	    i_3;

    /* Local variables */
    static int i, j, k;
    static float y;
    static int ii, kk, ip, iq, is;

    /* Parameter adjustments */
    rl_dim1 = *il;
    rl_offset = rl_dim1 + 1;
    rl -= rl_offset;
    b_dim1 = *ib;
    b_offset = b_dim1 + 1;
    b -= b_offset;
    x_dim1 = *ix;
    x_offset = x_dim1 + 1;
    x -= x_offset;

 
    is = *m;
    i_1 = *ir;
    for (j = 1; j <= i_1; ++j) {
/*     SOLUTION OF LY = B */
	i_2 = *n;
	for (i = 1; i <= i_2; ++i) {
	    ip = 1;
	    if (i <= *m + 1) {
		ip = *m - i + 2;
	    }
	    iq = i;
	    y = b[i + j * b_dim1];
	    k = is + 1;
	    if (is < ip) {
		goto L40;
	    }
	    i_3 = is;
	    for (kk = ip; kk <= i_3; ++kk) {
		--k;
		--iq;
		y -= rl[i + k * rl_dim1] * x[iq + j * x_dim1];
/* L20: */
	    }
L40:
	    x[i + j * x_dim1] = y * rl[i + (*m + 1) * rl_dim1];
/* L60: */
	}
/*     SOLUTION OF UX = Y */
	i = *n + 1;
	i_2 = *n;
	for (ii = 1; ii <= i_2; ++ii) {
	    --i;
	    ip = 1;
	    if (*n - i <= *m) {
		ip = *m - *n + i + 1;
	    }
	    y = x[i + j * x_dim1];
	    iq = i;
	    k = is + 1;
	    if (is < ip) {
		goto L100;
	    }
	    i_3 = is;
	    for (kk = ip; kk <= i_3; ++kk) {
		--k;
		++iq;
		y -= rl[iq + k * rl_dim1] * x[iq + j * x_dim1];
/* L80: */
	    }
L100:
	    x[i + j * x_dim1] = y * rl[i + (*m + 1) * rl_dim1];
/* L120: */
	}
/* L140: */
    }
    return 0;
} 
/*---------------------------------------------------------------------------*/

/*        
                  SI  : evaluation from spline coefficients
*/

/*t_ImagePtr imCtrl; 	coefficients  image 								*/
/*t_ImagePtr imSpl;   	spline image 										*/
/*___________________________________________________________________________*/

#if defined(_ANSI_)
static void ES_Coefs2Spline( t_ImagePtr imCtrl, t_ImagePtr imSpl)
#else
static void ES_Coefs2Spline(imCtrl, imSpl)
t_ImagePtr imCtrl;
t_ImagePtr imSpl;

#endif

{
	int i,j;
	register int k;
	t_3Derivates Der3; /* derivative values */
	sPar_ param_S; /* spline parameters */

	/* cubic B_spline functions */
        param_S.De=4;

        ES_InitSpline(&param_S,imSpl->tx,imSpl->ty,imSpl->tz,
                     imCtrl->tx,imCtrl->ty,imCtrl->tz);

	for(i=0;i<imSpl->tx;i++)
	      for(j=0;j<imSpl->ty;j++)
		    for(k=0;k<imSpl->tz;k++)
			     {ES_ComputeDerivates(&param_S,&Der3,imCtrl,i,j,k);
			      if (Der3.f<0.) Der3.f=0.;
			      else if (Der3.f>1.) Der3.f=1.;
			      T_Set3Point(i,j,k,imSpl,Der3.f, MODE_SET);}
	ES_FreeSpline(&param_S);
	return;
}
/*---------------------------------------------------------------------------*/
/* evaluation inside the mesh                                                */
/* t_ImagePtr imCtrl;   spline coefficients image 			     */
/*---------------------------------------------------------------------------*/

#if defined(_ANSI_)
void ES_ContinuousF( sPtr SP, t_ImagePtr imCtrl, reeltype u, reeltype v, reeltype w, reeltype *f)
#else
void ES_ContinuousF(SP, imCtrl, u, v, w, f)
sPtr SP;
t_ImagePtr imCtrl;
reeltype u;
reeltype v;
reeltype w;
reeltype *f;

#endif

{
	int g,h,DEG_h=(imCtrl->tz==1)?1:SP->De;
	int dimcxy=imCtrl->tx*imCtrl->ty;
	register int e;
	int inf=SP->m,jnf=SP->m,knf=SP->m;

	float uf=(float)u,vf=(float)v,wf=(float)w;

	register float *Ctge;
        float *Unif=SP->UNIF+1,*Vnif=SP->VNIF+1,*Wnif=SP->WNIF+1;
        float *Ctg,*Ct;
	
	float *BSU,*Bsu=SP->CspU;
        float *BSV,*Bsv=SP->CspV;
        float *BSW=SP->CspW;

        while (uf >= *(Unif+inf)  && inf<imCtrl->tx){inf++;}
        while (vf >= *(Vnif+jnf)  && jnf<imCtrl->ty){jnf++;}
        while (wf >= *(Wnif+knf)  && knf<imCtrl->tz){knf++;}

        Ct= (float *)imCtrl->buf+
		(knf-SP->m)*dimcxy+(jnf-SP->m)*imCtrl->tx+(inf-SP->m);

        S_EvalB_Spline(inf,SP->De,SP->UNIF,Bsu,uf);
        S_EvalB_Spline(jnf,SP->De,SP->VNIF,Bsv,vf);
        S_EvalB_Spline(knf,SP->De,SP->WNIF,BSW,wf);

	 *f=0;

	  for(h=0;h<DEG_h;h++,Ct+=dimcxy,BSW++)

           for(g=0,Ctg=Ct,BSV=Bsv; g<SP->De; g++, Ctg+=imCtrl->tx,BSV++)

               for(e=0,Ctge=Ctg,BSU=Bsu; e<SP->De; e++,BSU++)

                {*f += (reeltype)
			 (*Ctge++  * *(BSU+SP->des0) * *(BSV+SP->des0) *
                         *(BSW+SP->des0));}


	return;
}
/*---------------------------------------------------------------------------*/
/*   evaluation at mesh points */
/* t_3DerivatesPtr Der3P;  derivatives */
/* t_ImagePtr imCtrl;   spline coefficients image 			 */
/*---------------------------------------------------------------------------*/

#if defined(_ANSI_)
void ES_ComputeDerivates( sPtr SP, t_3DerivatesPtr Der3P, t_ImagePtr imCtrl, int i, int j, int k)
#else
void ES_ComputeDerivates(SP, Der3P, imCtrl, i, j, k)
sPtr SP;
t_3DerivatesPtr Der3P;
t_ImagePtr imCtrl;
int i;
int j;
int k;

#endif

{

	
	int g,h,DEG_h=(imCtrl->tz==1)?1:SP->De;

	int dimcxy=imCtrl->tx*imCtrl->ty;

	register int e;

	
	float *Ct=(float *)imCtrl->buf+
		        SP->Knf[k]*dimcxy+
			SP->Jnf[j]*imCtrl->tx+SP->Inf[i];
	register float *Ctge;float *Ctg;

	float *BSU,*Bsu=SP->BspU+i*SP->desQ;
	float *BSV,*Bsv=SP->BspV+j*SP->desQ;
	float *BSW=SP->BspW+k*SP->desQ;
	
	 
	Der3P->f=Der3P->fx=Der3P->fy=Der3P->fz=0;
	Der3P->fxx=Der3P->fxy=Der3P->fxz=Der3P->fyy=Der3P->fyz=Der3P->fzz=0;
	Der3P->fxxx=Der3P->fxxy=Der3P->fxxz=Der3P->fxyy=Der3P->fxyz=
		Der3P->fxzz=Der3P->fyyy=Der3P->fyyz=Der3P->fyzz=Der3P->fzzz=0;

	for(h=0;h<DEG_h;h++,Ct+=dimcxy,BSW++)
			 
	   for(g=0,Ctg=Ct,BSV=Bsv; g<SP->De; g++, Ctg+=imCtrl->tx,BSV++) 

	       for(e=0,Ctge=Ctg,BSU=Bsu; e<SP->De; e++,BSU++)

	        {Der3P->fxxx+= (reeltype) *Ctge * *(BSU+SP->des3) * 
			 *(BSV+SP->des0) * *(BSW+SP->des0);
		 Der3P->fxxy+= (reeltype) *Ctge * *(BSU+SP->des2) * 
			 *(BSV+SP->des1) * *(BSW+SP->des0);
		 Der3P->fxxz+= (reeltype) *Ctge * *(BSU+SP->des2) * 
			 *(BSV+SP->des0) * *(BSW+SP->des1);
		 Der3P->fxyy+= (reeltype) *Ctge * *(BSU+SP->des1) * 
			 *(BSV+SP->des2) * *(BSW+SP->des0);
		 Der3P->fxyz+= (reeltype) *Ctge * *(BSU+SP->des1) * 
			 *(BSV+SP->des1) * *(BSW+SP->des1);
		 Der3P->fxzz+= (reeltype) *Ctge * *(BSU+SP->des1) * 
			 *(BSV+SP->des0) * *(BSW+SP->des2);
		 Der3P->fyyy+= (reeltype) *Ctge * *(BSU+SP->des0) * 
			 *(BSV+SP->des3) * *(BSW+SP->des0);
		 Der3P->fyyz+= (reeltype) *Ctge * *(BSU+SP->des0) * 
			 *(BSV+SP->des2) * *(BSW+SP->des1);
		 Der3P->fyzz+= (reeltype) *Ctge * *(BSU+SP->des0) * 
			 *(BSV+SP->des1) * *(BSW+SP->des2);
		 Der3P->fzzz+= (reeltype) *Ctge * *(BSU+SP->des0) * 
			 *(BSV+SP->des0) * *(BSW+SP->des3);

		 
		 Der3P->fxx+= (reeltype) *Ctge * *(BSU+SP->des2) * 
			 *(BSV+SP->des0) * *(BSW+SP->des0);
		 Der3P->fxy+= (reeltype) *Ctge * *(BSU+SP->des1) * 
			 *(BSV+SP->des1) * *(BSW+SP->des0);
		 Der3P->fxz+= (reeltype) *Ctge * *(BSU+SP->des1) * 
			 *(BSV+SP->des0) * *(BSW+SP->des1);
		 Der3P->fyy+= (reeltype) *Ctge * *(BSU+SP->des0) * 
			 *(BSV+SP->des2) * *(BSW+SP->des0);
		 Der3P->fyz+= (reeltype) *Ctge * *(BSU+SP->des0) * 
			 *(BSV+SP->des1) * *(BSW+SP->des1);
		 Der3P->fzz+= (reeltype) *Ctge * *(BSU+SP->des0) * 
			 *(BSV+SP->des0) * *(BSW+SP->des2);

		 Der3P->fx+= (reeltype) *Ctge * *(BSU+SP->des1) * 
			 *(BSV+SP->des0) * *(BSW+SP->des0);
		 Der3P->fy+= (reeltype) *Ctge * *(BSU+SP->des0) * 
			 *(BSV+SP->des1) * *(BSW+SP->des0);
		 Der3P->fz+= (reeltype) *Ctge * *(BSU+SP->des0) * 
			 *(BSV+SP->des0) * *(BSW+SP->des1);

		 Der3P->f+= (reeltype) *Ctge++  * *(BSU+SP->des0) * 
			 *(BSV+SP->des0) * *(BSW+SP->des0);
	  }
	return;
}
/*___________________________________________________________________________*/
/* 
  fit splines to curves on surfaces
*/
/*---------------------------------------------------------------------------*/

#if defined(_ANSI_)
int ES_CurveSurfPar( cSPtr CP, float *x, float *U, int *I, 
		    float thres_pt, int nP, int siz, 
		    int comp_fact, float sf, float sfn, 
		    int iter_nor, int de, 
		    short closed, short verbose)
#else
int ES_CurveSurfPar(CP, x, U, I, thres_pt, nP, siz, comp_fact, sf, sfn, iter_nor, de, closed, verbose)
cSPtr CP;
float *x;
float *U;
int *I;
float thres_pt;
int nP;
int siz;
int comp_fact;
float sf;
float sfn;
int iter_nor;
int de;
short closed;
short verbose;

#endif

{
	sPar_ par_S;
	int nCirc,nC=0; /* number of control vertices */
	int deru=0;
	int itern=0;
	float cos_psi_max,cos2_psi_moy;
	int status= TRUE;
	
	if(nP-de+1<2) return( FALSE );  /* JPT 28-6-95 j'ai rajoute' false */
	par_S.De=de;
	par_S.taux=sf; 
	par_S.U=U;
	par_S.Inf=I;

	S_PolyInit(x,U,thres_pt,nP,&nC,I,siz,de,comp_fact);
	
	if(verbose) printf("nP=%d,nC=%d ",nP,nC);
	S_InitSplCurv(&par_S,nP,nC,closed);
	E_REPERR( goto END_ES_CurveSurfPar );
	nCirc=(closed)?nC-par_S.m:nC;


/* jpt 15/12/92 recuperation en cas d'erreur */
	
	status= S_InitLSCurv(&par_S,nP,nC,nCirc,siz,deru);
	if ( !status ) goto END_ES_CurveSurfPar;
	E_REPERR( goto END_ES_CurveSurfPar );

	S_WSystCurv(&par_S,x,nP,nCirc,siz,par_S.m);
	S_SolvCurv(&par_S,nC,nCirc,siz);
	
	/* transforms e1 and e2 into differentiable vector fields */

	S_DiffVectFieldNe1e2(&par_S,CP,nP,nCirc,x);
	
	/* rewrite and solve the system*/
	
	S_WSystCurv(&par_S,x,nP,nCirc,siz,par_S.m);
	S_SolvCurv(&par_S,nC,nCirc,siz);
	
	/* now for the reconstruction part, we want to have a uniform
	   sampling
	   
	   the sampling provided by the crest line extraction routine
	   is extremely irregular/sporadic
	   we use a linear sampling based on parameter values 
	   */
	
	  S_LinSamp(&par_S,nP,nC);
	 
	/* eval normal at new samples */
	S_EvalN(&par_S,CP,nP,nCirc);

	if(sfn>0.)do{
		/* eval tangent vector at new samples 
		   with new control vertices */
		S_EvalXu(&par_S,CP,nP,nCirc);
		S_UpdaNormConv(&par_S,CP,nP,nCirc,&cos_psi_max,&cos2_psi_moy,
			       sfn,closed);
		if(verbose)printf("s_u*.. cos_psi_max=%f cos2_psi_moy=%f\n",
				  cos_psi_max,cos2_psi_moy);}
	while(++itern < iter_nor);

 
	S_AllCurvSurfDer(&par_S,CP,nP,nCirc);

	S_CurvSurfSelPar(&par_S,CP,nP);
	
END_ES_CurveSurfPar:
	S_FreeSplCurv(&par_S);
	S_FreeLSCurv(&par_S);
	E_ISERR( status, "ES_CurveSurfPar", NULL );
	return( status );
}

/*___________________________________________________________________________*/
/* 

  polygonal approximation to initialise knot values
*/
/*---------------------------------------------------------------------------*/
#if defined(_ANSI_)
static void S_PolyInit( float *x, float *U, 
		       float thres, int nP, int *nC, int *I, 
		       int siz, int deg, int comp_fact)
#else
static void S_PolyInit(x, U, thres, nP, nC, I, siz, deg, comp_fact)
float *x;
float *U;
float thres;
int nP;
int *nC;
int *I;
int siz;
int deg;
int comp_fact;

#endif

{
	int depth,nVertM=nP-deg+1;
	float l=0.0;

	depth= (int)floor(log((double)nP/(double)comp_fact-
			      (double)deg+1.)/log(2.)+.5);
	if (depth==0) depth = 1;
	S_PolyTree(0,nP,x,U,&l,0,I,nC,nVertM,siz,depth,thres);
	I[*nC]=nP-1;
	*nC += deg -1;
	return;
}
/*___________________________________________________________________________*/
/* 
  distance between two points
*/
/*---------------------------------------------------------------------------*/
#if defined(_ANSI_)
static float S_Dist( float *Pt1, float *Pt2)
#else
static float S_Dist(Pt1, Pt2)
float *Pt1;
float *Pt2;

#endif

{
	
	register short i;
	float sum,tmp;
	for(sum=0.0,i=0;i<3;i++){
		tmp=Pt1[i]-Pt2[i];
	        tmp*=tmp;
		sum+=tmp;
	}
	return (float) sqrt((double)sum);
}
/*___________________________________________________________________________*/
/* 

  recursive call of the polygonal approximation
*/
/*---------------------------------------------------------------------------*/
#if defined(_ANSI_)
static void S_PolyTree( int ib, int ie, float *x, float *U, float *l, int cdep,
		  int *I, int *nVert, int nVertM, int siz, int depth, float thres)
#else
static void S_PolyTree(ib, ie, x, U, l, cdep, I, nVert, nVertM, siz, depth, thres)
int ib;
int ie;
float *x;
float *U;
float *l;
int cdep;
int *I;
int *nVert;
int nVertM;
int siz;
int depth;
float thres;

#endif

{
	int iPt=0;
	static int i,iE;
	float dv,dmax=0.0;
	static float LR[3];
	short m;
	
	iE=ie-1;
	if (*nVert >= nVertM) return;
	
	for(m=0;m<3;m++)LR[m]=x[m+iE*siz]-x[m+ib*siz];
	
	for(i=ib;i<ie;i++)
		{S_PtSegDist(x+i*siz,x+ib*siz,x+iE*siz,LR,&dv,U+i,*l);
		 if(dv>dmax){dmax=dv;iPt=i;}}

	if((cdep==depth) || (dmax<thres) || (iE-ib<=1)){
		I[*nVert]=ib;
		*nVert += 1;
		/*  *nVert++ increments the pointer and not the value */

		*l += S_Dist(x+ib*siz,x+iE*siz);
	}
	
	else {/* separate in left and right lists 
		 iPt will be the separation point
		 */

	      S_PolyTree(ib,iPt+1,x,U,l,cdep+1,I,nVert,nVertM,siz,depth,thres);
	      S_PolyTree(iPt,ie,x,U,l,cdep+1,I,nVert,nVertM,siz,depth,thres);
	}
	return;
}
/*___________________________________________________________________________*/
/* 
  distance from a point to a segment
*/
/*---------------------------------------------------------------------------*/
#if defined(_ANSI_)
static void S_PtSegDist( float *Pt, float *A, float *B, float *AB, 
			float *dv, float *dh, float lpoly)
#else
static void S_PtSegDist(Pt, A, B, AB, dv, dh, lpoly)
float *Pt;
float *A;
float *B;
float *AB;
float *dv;
float *dh;
float lpoly;

#endif

{
	static float AM[3],BM[3], p,n;
	short j;
	for(j=0;j<3;j++){AM[j]=Pt[j]-A[j];BM[j]=Pt[j]-B[j];}

	p=AM[0]*AB[0]+AM[1]*AB[1]+AM[2]*AB[2]; /*(AM,AB)*/
	       
	n= AB[0]*AB[0]+AB[1]*AB[1]+AB[2]*AB[2]; /*p-n= (BM,AB) */

	if (p <= 0.0) {*dv=S_Dist(Pt,A);*dh=lpoly;}
	
	else if (p-n >= 0.0){*dv=S_Dist(Pt,B);
			*dh= (float)sqrt((double)n)+lpoly;}
	else {  
		S_VectProd(AM,AB,BM);
		n=(float)sqrt((double)n);
		*dv=(float)sqrt((double)BM[0]*BM[0]+BM[1]*BM[1]+BM[2]*BM[2])
			/n;
		*dh=p/n+lpoly;}
	return;
}
/*___________________________________________________________________________*/
/* 
  initialise spline curve
*/
/*---------------------------------------------------------------------------*/
#if defined(_ANSI_)
static void S_InitSplCurv( sPtr SP, int dimu, int dimcu, short close_u)
#else
static void S_InitSplCurv(SP, dimu, dimcu, close_u)
sPtr SP;
int dimu;
int dimcu;
short close_u;

#endif

{
	S_InitDegreeCoef(SP);
	
	S_AllocSplCurv(SP,dimu,dimcu);
	E_ISERRN( "InitSplCurv", NULL );
	
	S_ParamCurve(SP,dimcu,close_u);
	S_ComputeU(SP,SP->U,SP->UNIF,SP->BspU,SP->Inf,dimu,dimcu);
	return;
}
/*___________________________________________________________________________*/
/* 
  alloc spline curve arrays
*/
/*---------------------------------------------------------------------------*/
#if defined(_ANSI_)
static void S_AllocSplCurv( sPtr SP, int dimu, int dimcu)
#else
static void S_AllocSplCurv(SP, dimu, dimcu)
sPtr SP;
int dimu;
int dimcu;

#endif

{
	SP->UNIF=NULL;
	SP->BspU=NULL;
	SP->CspU=NULL;

	SP->UNIF=(float*) calloc(dimcu+SP->De,sizeof(int));
	SP->BspU=(float *) calloc(dimu*SP->desQ,sizeof(float));
	SP->CspU=(float *) calloc(SP->desQ,sizeof(float));
	if(   SP->UNIF==NULL || SP->BspU==NULL || 
	      SP->CspU==NULL   ){
		E_SETERR( E_ERRMEM, "calloc", 
			 "UNIF-BspU-CspU bad alloc" );
		E_REPERR( goto END_S_AllocSplCurv );
	}
END_S_AllocSplCurv:
	E_REPERR( S_FreeSplCurv(SP));
	E_ISERRN( "S_AllocSplCurv", NULL );
	return;
}
/*___________________________________________________________________________*/
/* 
  free spline curve arrays
*/
/*---------------------------------------------------------------------------*/
#if defined(_ANSI_)
static void S_FreeSplCurv( sPtr SP)
#else
static void S_FreeSplCurv(SP)
sPtr SP;

#endif

{
	T_Free((void **) &SP->UNIF);
	T_Free((void **) &SP->BspU);
	T_Free((void **) &SP->CspU);
	return;
}
/*___________________________________________________________________________*/
/* 
  parametrize spline curve
*/
/*---------------------------------------------------------------------------*/
#if defined(_ANSI_)
static void S_ParamCurve( sPtr SP, register int dimcu, short close)
#else
static void S_ParamCurve(SP, dimcu, close)
sPtr SP;
register int dimcu;
short close;

#endif

{
	
	register int i;
	
	for (i=0;i<dimcu+2-SP->De;i++)
		 SP->UNIF[i+SP->m]=SP->U[SP->Inf[i]];

	for (i=SP->De-2;i>=0;i--) /*first*/
		{SP->UNIF[i]=SP->U[0];
		 if (close)
			 SP->UNIF[i]+=SP->UNIF[dimcu+i-SP->m]-SP->UNIF[dimcu];}

	for (i=0;i<SP->De-1;i++)/*last */
                {SP->UNIF[dimcu+1+i]=SP->UNIF[dimcu];
		  if (close)
			  SP->UNIF[dimcu+1+i]+= SP->UNIF[i+SP->De]-SP->U[0];}
	
	return;
}
/*___________________________________________________________________________*/
/* 
  initialise spline curve for least squares
*/
/*---------------------------------------------------------------------------*/
#if defined(_ANSI_)
static int S_InitLSCurv( sPtr SP, int dimu, int dimcu, int dimcircu, int siz, int deru)
#else
static int S_InitLSCurv(SP, dimu, dimcu, dimcircu, siz, deru)
sPtr SP;
int dimu;
int dimcu;
int dimcircu;
int siz;
int deru;

#endif

{
	int idet;
	int iFail=0;
	int dimcircu1=dimcircu-1;
	int dim2circu=dimcircu*dimcircu;
	double det;
	S_AllocLSCurv(SP,dimcircu,siz);
	E_ISERR( FALSE, "InitLSCurv", NULL );	/* JPT j'ai rajoute' false */
	
	if(dimcircu < dimcu)
		{S_ComputeCircUUt(SP,SP->Inf,SP->BspU,
			     SP->LLt+dim2circu,dimu,dimcircu,deru,SP->taux);
		 S_CircBandUUt(SP,dimcircu);
		 S_Chobandet(&dimcircu,&dimcircu1,SP->LLt,&dimcircu,SP->LLt,
			     &dimcircu,&dimcircu,&det,&idet,&iFail);}
	

	else
		{S_ComputeUUt(SP,SP->Inf,SP->BspU,SP->UUt,
			     dimu,dimcu,deru,SP->taux);
		 S_Chobandet(&dimcu,&SP->m,SP->UUt,&dimcu,SP->LLt,&dimcu,
		    &SP->De,&det,&idet,&iFail);}
	
/* jpt 15/12/92 recuperation en cas d'erreur 
	if (iFail!=0){E_SETERR( E_ERRDIV, "InitLSCurv", "choleski failure" );
		      E_ISERRN( "InitLSCurv", NULL );}
*/
	if (iFail!=0) return( FALSE );
	return( TRUE );
}
/*___________________________________________________________________________*/
/* 
  alloc spline curve arrays for least squares
*/
/*---------------------------------------------------------------------------*/
#if defined(_ANSI_)
static void S_AllocLSCurv( sPtr SP, int dimcu, int siz)
#else
static void S_AllocLSCurv(SP, dimcu, siz)
sPtr SP;
int dimcu;
int siz;

#endif

{
	
	SP->UUt=NULL;
	SP->LLt=NULL;
	SP->VUt=NULL;
	SP->Gvt=NULL;

	SP->UUt=(float*) calloc(dimcu*SP->De*2,sizeof(float));

	SP->LLt=(float*) calloc(dimcu*dimcu*2,sizeof(float));

	SP->VUt=(float*) calloc(dimcu*siz,sizeof(float));

	SP->Gvt=(float*) calloc(dimcu*siz,sizeof(float));

	if(   SP->UUt==NULL || SP->LLt==NULL || 
	      SP->VUt==NULL || SP->Gvt==NULL   ){
		E_SETERR( E_ERRMEM, "calloc", "UUt-LLt-VUt-Gvt bad alloc" );
		E_REPERR( goto END_S_AllocLSCurv );
	}

END_S_AllocLSCurv:
	E_REPERR( S_FreeLSCurv(SP));
	E_ISERRN( "S_AllocLSCurv", NULL );
	return;
}
/*___________________________________________________________________________*/
/* 
  free spline curve arrays for least square
*/
/*---------------------------------------------------------------------------*/
#if defined(_ANSI_)
static void S_FreeLSCurv( sPtr SP)
#else
static void S_FreeLSCurv(SP)
sPtr SP;

#endif

{
	T_Free((void **) &SP->UUt);
	T_Free((void **) &SP->LLt);
	T_Free((void **) &SP->VUt);
	T_Free((void **) &SP->Gvt);
	return;
}
/*___________________________________________________________________________*/
/* 
  compute matrix Bx
*/
/*---------------------------------------------------------------------------*/
#if defined(_ANSI_)
static void S_WSystCurv( sPtr SP, float *x, int nP, int nCirc, int siz, int deriv)
#else
static void S_WSystCurv(SP, x, nP, nCirc, siz, deriv)
sPtr SP;
float *x;
int nP;
int nCirc;
int siz;
int deriv;

#endif

{
	int des=deriv*SP->De;
	int i,j,*I;
	register int e,ind;
	register float *BSU,*VUt,*x1;

	for(VUt=SP->VUt,i=0;i<siz*nCirc;i++) *VUt++ =0.;

	for(VUt=SP->VUt,j=0;j<siz;j++,x++,VUt+=nCirc)

		for(x1=x,BSU=SP->BspU+des,I=SP->Inf,
		    i=0;i<nP;i++,BSU+=SP->desQ,I++,x1+=siz)
			
				for(ind= *I,e=0;e<SP->De; e++,ind++)
					*(VUt+ ind % nCirc)
						+= *x1 *  *(BSU+e);
	return;
}
/*___________________________________________________________________________*/
/* 
  circular parametrization:
  write least squares matrix
*/
/*___________________________________________________________________________*/
/* run with dimc=dimc-m */
#if defined(_ANSI_)
static void S_ComputeCircUUt( sPtr SP, int *Inf, float *Bsp, register float *UUt, 
			     int dim1, int dimc, int der, float tau_u)
#else
static void S_ComputeCircUUt(SP, Inf, Bsp, UUt, dim1, dimc, der, tau_u)
sPtr SP;
int *Inf;
float *Bsp;
register float *UUt;
int dim1;
int dimc;
int der;
float tau_u;

#endif

{
	register int f;
	int e,i,ind_u,iu,iv,dim=dimc+1;
	int dc1=dimc*dimc,dc=dc1+dimc;
	register float *Bsp0,*BspD0;
	float  bu,bud,bd,*BspD;
	float magder,tmpmag;

	tmpmag=(float)(dimc+1-SP->De);
	tmpmag=(tmpmag<=1.0)?(float)1.: (float)(dim1-1)/tmpmag;
	for(magder=1.,e=SP->m;e>der;e--){magder*=tmpmag;}
	
	for(Bsp0=UUt,f=0;f<dc1;f++) *Bsp0++ = 0.0;

	for(BspD=Bsp+der*SP->De,Bsp+=SP->desQ-SP->De,
	  i=0;i<dim1;i++,Bsp+=SP->desQ,BspD+=SP->desQ,Inf++){

	    for(e=0,ind_u= *Inf *dim; e<SP->De; e++, ind_u+=dim) 

		 {bu= *(Bsp+e);

		  bud= *(BspD+e);

		 for(f=e,iu=iv=ind_u%dc,Bsp0=Bsp+e,BspD0=BspD+e;
		     f<SP->De;f++,iu+=dimc,iv++)  

			 {bd=  (bud* *(BspD0++)) *magder;
			 if (iu<dc1)
			 {*(UUt+iu) +=  (bu* *(Bsp0++))+bd*tau_u;}

			 else
			{*(UUt+iv-dimc) += (bu* *(Bsp0++))+bd*tau_u;}
	} }/*for f,e */
    }/* end for i */	

}

/*___________________________________________________________________________*/
/* 
  circular parametrization:
  convert symetrix matrix LLt+dim^2 in circular band matrix LLt
*/
/*---------------------------------------------------------------------------*/
#if defined(_ANSI_)
static void S_CircBandUUt( sPtr SP, int dim)
#else
static void S_CircBandUUt(SP, dim)
sPtr SP;
int dim;

#endif

{
	int dim2=dim*dim;
	int dim0=dim2-dim;
	int dim1=dim+1;
	int di=dim-1;

	register int i,j;
	register float *L, *U;
	float *L1, *U1;

	
	for(i=0,L=SP->LLt;i<dim2;i++) *L++ =0.0;

	/* diagonal by diagonal */

	for(L1=SP->LLt+dim0,
	    U1=SP->LLt+dim2,
	    i=0;i<dim;i++,
	    U1+=dim,L1-=di)
		
		for(L=L1,U=U1,j=i;
		    j<dim;j++,
		    U+=dim1,L++)
			*L= *U;
	
	/* 
	  how to print a matrix computed for input of a fortran routine:
	  the matrix is written colum by column
	  since we necessarily write it row by row
	  we make each time a jump of the dimention of one column

	  
	 for(i=0;i<dim;i++)       ..for each row
		{for(j=0;j<dim;j++)  ..print one row     \/ column jump
			fprintf(stderr,"%f ", SP->LLt[i+dim*j]);
		 fprintf(stderr,"\n");}

	*/
	return;
}
/*___________________________________________________________________________*/
/* 
  choleski method, solving part
*/
/*---------------------------------------------------------------------------*/
#if defined(_ANSI_)
static void S_SolvCurv( sPtr SP, int dimcu, int dimcircu, int siz)
#else
static void S_SolvCurv(SP, dimcu, dimcircu, siz)
sPtr SP;
int dimcu;
int dimcircu;
int siz;

#endif

{
	int dimcircu1=dimcircu-1;
	
	if(dimcircu < dimcu)
		S_Chobandsol(&dimcircu,&dimcircu1,&siz,SP->LLt,&dimcircu,
			     &dimcircu,SP->VUt,&dimcircu,SP->Gvt,&dimcircu);
	else
		S_Chobandsol(&dimcu,&SP->m,&siz,SP->LLt,&dimcu,&SP->De,SP->VUt,
			     &dimcu,SP->Gvt,&dimcu);
	return;
}
/*___________________________________________________________________________*/
/* 
  transforms N e1 and e2 into differentiable vector fields
*/
/*---------------------------------------------------------------------------*/
#if defined(_ANSI_)
static void S_DiffVectFieldNe1e2( sPtr SP, cSPtr CP, int nP, int nCirc, float *e1x)
#else
static void S_DiffVectFieldNe1e2(SP, CP, nP, nCirc, e1x)
sPtr SP;
cSPtr CP;
int nP;
int nCirc;
float *e1x;

#endif

{
	int i,*I,ind,ind0,N=5*nCirc; 

	short j;
	register short e;

	register cSPtr CP1=CP;

	float *VUt,*BSU,pscal;
	register float *BU;
	
	/*******************     Initialisation      ******************/

	for(CP1=CP,i=0;i<nP;i++,CP1++) 
	   for(j=0;j<3;j++)CP1->N[j]=CP1->xu[j]=0.;

	/*****************       frame and position      **************/
	
	for(VUt=SP->Gvt,j=0;j<3;j++,VUt+=nCirc)
		for(CP1=CP,BSU=SP->BspU,I=SP->Inf,
		    i=0;i<nP;i++,BSU+=SP->desQ,I++,CP1++)
			
		      for(ind= *I ,
			  BU=BSU,e=0;e<SP->De; e++,ind++)
			      { ind0=ind % nCirc;
				CP1->N[j]   += *(VUt+ind0+N) 
					*  *(BU   +SP->des0);
				CP1->xu[j]  += *(VUt+ind0) 
					*  *(BU++ +SP->des1);}
	
	
	/*e1 must be re-read from the file 
	  because of orientation problems we cannot smooth
	  e1 directly
	  we use the fact that the direction of the crest lines
	  is near the e2 direction
	  and also the fact that there MUST be a differentiable
	  field of e1s at least along the line 

	  on the other hand N is smoothed directly
	  because we smooth the normal */

	for(CP1=CP,e1x+=8,i=0;i<nP;i++,CP1++,e1x+=8)
		{S_Normalise(CP1->N,3);
		 for(j=0;j<3;j++,e1x++)
			CP1->e1[j]= *e1x;
		 S_VectProd(CP1->N,CP1->e1,CP1->e2);
		 S_ScalProd(CP1->xu,CP1->e2,&pscal,3);
		 if (pscal<0.) 
			{for(e1x -=3,j=0;j<3;j++,e1x++)
			   *e1x = - *e1x;}
	 }

	/* in order to provide a differentiable field of principal directions
	   we want e2 to be as close to t, curve tangent as possible
	   
	   so if the scalar product S_ScalProd(CP1->xu,CP1->e2)
	   is negative we inverse CP1->e1 and thus CP1->e2
	*/

	return;

}
/*___________________________________________________________________________*/
 
#if defined(_ANSI_)
static void S_VectProd( register float *u, register float *v, register float *w)
#else
static void S_VectProd(u, v, w)
register float *u;
register float *v;
register float *w;

#endif

{
        w[0]=u[1]*v[2]-u[2]*v[1];
        w[1]=u[2]*v[0]-u[0]*v[2];
        w[2]=u[0]*v[1]-u[1]*v[0];
	return;
}
/*---------------------------------------------------------------------------*/

#if defined(_ANSI_)
static void S_ScalProd( register float *u, register float *v, register float *res, register int n)
#else
static void S_ScalProd(u, v, res, n)
register float *u;
register float *v;
register float *res;
register int n;

#endif

{
	register int i;
	for (*res=0.,i=0;i<n;i++) *res  += *u++ * *v++ ;

	return;
}
/*___________________________________________________________________________*/

#if defined(_ANSI_)
static void S_Norm( register float *u, register float *res, register int n)
#else
static void S_Norm(u, res, n)
register float *u;
register float *res;
register int n;

#endif

{
	S_ScalProd(u,u,res,n);
	*res=(float)sqrt((double) *res);
}
/*___________________________________________________________________________*/

#if defined(_ANSI_)
static void S_Normalise( register float *u, register int n)
#else
static void S_Normalise(u, n)
register float *u;
register int n;

#endif

{
	float norm_u; register float inv_norm_u;
	S_Norm(u, &norm_u, n);
	inv_norm_u= (norm_u==0.0) ? (float)0.0 : 1/norm_u;
	S_Multiply(u,inv_norm_u,n);
	return;
}
/*___________________________________________________________________________*/

#if defined(_ANSI_)
static void S_Multiply( register float *u, register float lambda, register int n)
#else
static void S_Multiply(u, lambda, n)
register float *u;
register float lambda;
register int n;

#endif

{
	register int i;
	for (i=0;i<n;i++) *u++ *= lambda;

	return;
}
/*___________________________________________________________________________*/
/* 
  Linear sampling for the spline curve parametrize spline curve fo
*/
/*---------------------------------------------------------------------------*/
#if defined(_ANSI_)
static void S_LinSamp( sPtr SP, int dimu, int dimcu)
#else
static void S_LinSamp(SP, dimu, dimcu)
sPtr SP;
int dimu;
int dimcu;

#endif

{
	register int i;
	int dim1u=dimu-1;
	float utot= *(SP->U+dim1u);
	register float incu= utot/ (float)dim1u,valu=0.0;
	register float *U=SP->U;

	for (i=0;i<dim1u;i++,
	     valu+=incu) 
		*U++ =valu;

	S_ComputeU(SP,SP->U,SP->UNIF,SP->BspU,SP->Inf,dimu,dimcu);
	return;
}
/*___________________________________________________________________________*/
/* 
  eval surface normal vectors
*/
/*---------------------------------------------------------------------------*/
#if defined(_ANSI_)
static  void S_EvalN( sPtr SP, cSPtr CP, int nP, int nCirc)
#else
static  void S_EvalN(SP, CP, nP, nCirc)
sPtr SP;
cSPtr CP;
int nP;
int nCirc;

#endif

{
	register short e;
	short j;
	register cSPtr CP1=CP;

	int i, ind,*I, N=5*nCirc;
	float *BSU,*VUt;
	register float *BU;
	
	for(CP1=CP,i=0;i<nP;i++,CP1++) 
	   for(j=0;j<3;j++)CP1->N[j]=0.;

	for(VUt=SP->Gvt,j=0;j<3;j++,VUt+=nCirc)
		for(CP1=CP,BSU=SP->BspU,I=SP->Inf,
		    i=0;i<nP;i++,BSU+=SP->desQ,I++,CP1++)
			
		      for(ind= *I,BU=BSU,e=0;e<SP->De; e++,ind++)
			      {CP1->N[j]  += *(VUt+N+ind%nCirc) *  
				       *(BU++ +SP->des0);}

	for(CP1=CP,i=0;i<nP;i++,CP1++) S_Normalise(CP1->N,3);
	return;
}
/*___________________________________________________________________________*/
/* 
  eval curve tangent vectors
*/
/*---------------------------------------------------------------------------*/
#if defined(_ANSI_)
static void S_EvalXu( sPtr SP, cSPtr CP, int nP, int nCirc)
#else
static void S_EvalXu(SP, CP, nP, nCirc)
sPtr SP;
cSPtr CP;
int nP;
int nCirc;

#endif

{
	register short e;
	short j;
	register cSPtr CP1=CP;
	int i,ind, *I;
	float *VUt,*BSU;
	register float *BU;
	
	for(CP1=CP,i=0;i<nP;i++,CP1++) 
	   for(j=0;j<3;j++)CP1->xu[j]=0.;

	for(VUt=SP->Gvt,j=0;j<3;j++,VUt+=nCirc)
		for(CP1=CP,BSU=SP->BspU,I=SP->Inf,
		    i=0;i<nP;i++,BSU+=SP->desQ,I++,CP1++)
			
		      for(ind= *I,BU=BSU,e=0;e<SP->De; e++,ind++)
			      {CP1->xu[j]  += *(VUt+ ind%nCirc)
				      *  *(BU++ +SP->des1);}
	return;
}
/*___________________________________________________________________________*/
/*
   substract projection on tangent vector onto normal vector 
   (t.N).N             
*/                                                  
/*___________________________________________________________________________*/

#if defined(_ANSI_)
static void S_UpdaNormConv( sPtr SP, cSPtr CP, 
			   int nP, int nCirc, 
			   float *cos_psi_max, float *cos2_psi_moy, 
			   float lambda, short close)
#else
static void S_UpdaNormConv(SP, CP, nP, nCirc, cos_psi_max, cos2_psi_moy, lambda, close)
sPtr SP;
cSPtr CP;
int nP;
int nCirc;
float *cos_psi_max;
float *cos2_psi_moy;
float lambda;
short close;

#endif

{
	cSPtr CP1;
	int i,j,*I,ind,ind0;
	register int e;
	register float *BSU;
	float *VUt;
	
	/* BtB cx + BddtBd cx + lambda Nb'tb'Nt cx = Bx                */
	/* cxi -cxi_1 = lambda b'(b'cxN)N                              */

	
	for(*cos_psi_max= *cos2_psi_moy=0.,CP1=CP,i=0;i<nP;i++,CP1++){

		 S_ScalProd(CP1->N,CP1->xu,&CP1->cos_psi,3);
		 
		 if (CP1->cos_psi> *cos_psi_max) *cos_psi_max=CP1->cos_psi;
		 if (- CP1->cos_psi> *cos_psi_max) *cos_psi_max= -CP1->cos_psi;
		 *cos2_psi_moy+= CP1->cos_psi*CP1->cos_psi;
	 }
	*cos2_psi_moy/= nP;
	
     for(VUt=SP->Gvt,j=0;j<3;j++,VUt+=nCirc)

	 for(CP1=CP,BSU=SP->BspU,I=SP->Inf,
	     i=0;i<nP;i++,BSU+=SP->desQ,I++,CP1++)

	     for(ind= (*I),e=0;e<SP->De; e++,ind++)

		     {ind0= ind % nCirc;

		      if( (close) || 
			  ( (ind0 >0) && (ind0 < nCirc-1) ) )
		            /*     ^except end-points^ */

			     *(VUt+ind0)  -=   *(BSU+SP->des1+e) 
				             /* b_s */

				     * CP1->cos_psi   * lambda  
					     /* b' . cxN N normalised */
					     * CP1->N[j];}
	return;
}
/*___________________________________________________________________________*/
/*
   eval all curve derivatives using splines          
*/                                                  
/*___________________________________________________________________________*/

#if defined(_ANSI_)
static void S_AllCurvSurfDer( sPtr SP, cSPtr CP, int nP, int nCirc)
#else
static void S_AllCurvSurfDer(SP, CP, nP, nCirc)
sPtr SP;
cSPtr CP;
int nP;
int nCirc;

#endif

{
	int i,*I,N=5*nCirc,K1=3*nCirc,K2=K1+nCirc,E1=N+K1; 

	register int ind,ind0;
	short j;

	register short e;

	register cSPtr CP1=CP;

	float *VUt,*BSU,pscal;
	register float *BU;

	/*..................     Initialisation      .................*/

	for(CP1=CP,i=0;i<nP;i++,CP1++) 
	   {CP1->k1=CP1->k2=0.;
	    for(j=0;j<3;j++)
	      CP1->x[j]=CP1->xu[j]=CP1->xuu[j]=CP1->xuuu[j]=CP1->N[j]=
		       CP1->Nu[j]=CP1->e1[j]=0.;}

	/*................   computation of curvatures   .............*/
	 
	for(CP1=CP,BSU=SP->BspU,I=SP->Inf,
	    i=0;i<nP;i++,BSU+=SP->desQ,I++,CP1++)
			
		      for(ind= *I,BU=BSU,
			  e=0;e<SP->De; e++,ind++)
		/* surface curvatures */		 
				{ind0=ind%nCirc;
				 CP1->k1  += *(SP->Gvt+ ind0 +K1)
					 *  *(BU+SP->des0);

				 CP1->k2  += *(SP->Gvt+ ind0 +K2)
					 *  *(BU++ +SP->des0);}


	/*................       frame and position      ..............*/

	for(VUt=SP->Gvt,j=0;j<3;j++,VUt+=nCirc)
		for(CP1=CP,BSU=SP->BspU,I=SP->Inf,
		    i=0;i<nP;i++,BSU+=SP->desQ,I++,CP1++)
			
		      for(ind= *I,
			  BU=BSU,e=0;e<SP->De; e++,ind++)
		/* normal and normal derivative*/	
			      {ind0=ind%nCirc;
			       CP1->N[j]   += *(VUt+ind0+N) * *(BU+SP->des0);
			       CP1->Nu[j]  += *(VUt+ind0+N) * *(BU+SP->des1);
		/* principal frame as a rotation vector */		 
			       CP1->e1[j] += *(VUt+ind0+E1) * *(BU+SP->des0);
		
		/* position */CP1->x[j] += *(VUt+ind0) * *(BU+SP->des0);
			      CP1->xu[j] += *(VUt+ind0) * *(BU+SP->des1);
			      CP1->xuu[j] += *(VUt+ind0) * *(BU+SP->des2);
			      CP1->xuuu[j] += *(VUt+ind0) * *(BU++ +SP->des3);}

	/*................        normalise frame       ..............*/
	
	/*in case no jumps appear on N we can normalise the smoothed N*/
	/* e1 must then be orthogonal to N and also normalized */

	for(CP1=CP,i=0;i<nP;i++,CP1++)
		{S_Normalise(CP1->N,3);
		 S_ScalProd(CP1->N,CP1->e1,&pscal,3);
		 S_VectDiffMult(CP1->e1,CP1->N,pscal,3);
		 S_Normalise(CP1->e1,3);}

	return;
}
/*___________________________________________________________________________*/

#if defined(_ANSI_)
static void S_VectDiffMult( register float *u, register float *v,
			   register float lambda, register int n) /* u=u-lambda*v */
#else
static void S_VectDiffMult(u, v, lambda, n)
register float *u;
register float *v;
register float lambda;
register int n;

#endif

{
	register int i;
	for (i=0;i<n;i++) *u++ -= lambda* *v++ ;

	return;
}
/*___________________________________________________________________________*/
/*

   selected curve surface parameters are evaluated        
*/                                                  
/*___________________________________________________________________________*/
#if defined(_ANSI_)
static void S_CurvSurfSelPar( sPtr SP, cSPtr CP, int nP)
#else
static void S_CurvSurfSelPar(SP, CP, nP)
sPtr SP;
cSPtr CP;
int nP;

#endif

{
	register cSPtr CP1=CP;register int i;

    for(i=0;i<nP;i++,CP1++){
	/* normalise normal and principal direction */
	
	
	S_VectProd(CP1->N,CP1->e1,CP1->e2); 

	/*...........arclengh derivative ...........*/
	
	S_Norm(CP1->xu, &CP1->s_u, 3);

	/*........... vector product xu xuu ...........*/

	S_VectProd(CP1->xu, CP1->xuu, CP1->b);

	S_Norm(CP1->b,&CP1->ks_u3,3);

	/*........... curvature ...........*/

	CP1->k=CP1->ks_u3 / CP1->s_u / CP1->s_u / CP1->s_u;

	/*........... torsion ...........*/

	S_3Det(CP1->xu,CP1->xuu,CP1->xuuu, &CP1->tau);
	
	CP1->tau /= CP1->ks_u3 * CP1->ks_u3;

	CP1->tau = -CP1->tau;

	/*............ binormal vector and frenet frame ............*/
	S_Multiply(CP1->b,1/CP1->ks_u3,3);

	S_Multiply_in(CP1->xu, CP1->t,1  / CP1->s_u,3); /* tangent vector */
	
	S_VectProd(CP1->b,CP1->t,CP1->n); /* normal vector */

	/*...........         SURFACE PARAMETERS          ............*/

	/*...........           geodesic torsion          ............*/
	
	S_3Det(CP1->Nu,CP1->N,CP1->t, &CP1->taug_N);
	
	CP1->taug_N /= CP1->s_u;

	/*...........           geodesic curvature        ............*/

	S_3Det(CP1->xuu,CP1->N,CP1->t, &CP1->kg_t);
	
	CP1->kg_t /= CP1->s_u;CP1->kg_t /= CP1->s_u;

	S_3Det(CP1->xuu,CP1->N,CP1->t, &CP1->kg_t);
	
	CP1->kg_t /= CP1->s_u;CP1->kg_t /= CP1->s_u;

	/*...........           normal  curvature         ............*/

	S_ScalProd(CP1->Nu,CP1->t,&CP1->kn_N,3);
	
	CP1->kn_N /= CP1->s_u;
	
	/*...........                ANGLES               ............*/


	S_Angle(CP1->N,CP1->n,&CP1->theta,3);
	
	S_Angle(CP1->e1,CP1->t,&CP1->phi,3);
	
	S_Angle(CP1->e2,CP1->b,&CP1->rho,3);

	S_ScalProd(CP1->N,CP1->t,&CP1->cos_psi,3);
		 
	
   }/* end for i */

	return;
}
/*___________________________________________________________________________*/

#if defined(_ANSI_)
static void S_Multiply_in( register float *u, register float *v, register float lambda, register int n)
#else
static void S_Multiply_in(u, v, lambda, n)
register float *u;
register float *v;
register float lambda;
register int n;

#endif

{
	register int i;
	for (i=0;i<n;i++) *v++ = *u++ * lambda;

	return;
}

/*___________________________________________________________________________*/

#if defined(_ANSI_)
static void S_Angle( register float *u, register float *v,
		    register float *res, int n) /* n should be 3, otherwise ? */
#else
static void S_Angle(u, v, res, n)
register float *u;
register float *v;
register float *res;
int n;

#endif

{
	float p,q;
	S_ScalProd(v,v, &q,n);
	S_ScalProd(u,v, &p,n);
	*res= (float) atan2(q-p*p,p);
}
/*---------------------------------------------------------------------------*/

#if defined(_ANSI_)
static void S_3Det( register float *u, register float *v, register float *w, register float *res)
#else
static void S_3Det(u, v, w, res)
register float *u;
register float *v;
register float *w;
register float *res;

#endif

{
        *res=u[0]*v[1]*w[2]+v[0]*w[1]*u[2]+w[0]*u[1]*v[2];
        *res-=w[0]*v[1]*u[2]+u[0]*w[1]*v[2]+v[0]*u[1]*w[2];
	return;
}

/*___________________________________________________________________________*/
/* 
  test function
*/
/*---------------------------------------------------------------------------*/
#if defined(_ANSI_)
void ES_TestSpline( void )
#else
void ES_TestSpline()

#endif
{
	t_Image imIn,imCtrl,imSpl;	/*  epidaure.lib images	*/
	int dimcx=24,dimcy=24,dimcz=24;
	float sx=(float) .01,sy=(float) .01,sz=(float) .01;

	T_Init3Image( &imIn, 0,0,0,TYPEIM_UNKNOWN, "test.im" );

	T_ReadInrim( &imIn );
	
	T_Init3Image( &imCtrl, dimcx,dimcy,dimcz,TYPEIM_FLOAT, 
		     "tmp.ctrl" );
	T_AllocImage( &imCtrl);
	E_CoefSpline( &imIn,&imCtrl,sx,sy,sz);
	T_WriteInrim( &imCtrl );
	/* write the Ctrl image on disk		*/

	/* back from coefficients to spline */
	T_Init3Image( &imSpl, imIn.tx,imIn.ty,imIn.tz,TYPEIM_256, 
		     "tmp.spline" );
	T_AllocImage( &imSpl );
	E_ISERRN("ES_TestSpline",NULL);

	ES_Coefs2Spline(&imCtrl,&imSpl);
	T_WriteInrim( &imSpl );
	/* write the spline image on disk		*/

	T_FreeImage( &imIn );
	T_FreeImage( &imCtrl );
	T_FreeImage( &imSpl );
	/* free the image buffers		*/
	E_ISERRN("ES_TestSpline",NULL);
	
	return;
}



/*---------------------------------------------------------------------------
 *      Fin de fichier 
 *---------------------------------------------------------------------------
 */



