/* epi_vec.c 2.1 5/14/98 */
/*==========================================================================
 *
 * File : $EPI/src/epi_vec.c
 * Module :
 * $ Id : ? $ 
 * Test file : ?
 * Author : J-P Thirion
 * Creation: August 4 1992
 *===========================================================================
 *	(c) Copyright 1992, Inria, all right reserved
 *===========================================================================
 *
 *------------------------------Description----------------------------------
 *
 * Fonctions ge'ne'rales d'ope'rations vectorielles et matricielles
 *
 *-----------------------------Modification----------------------------------
 *
 */

/*======================================================================*/
/*
	epi_vec.c
  
	Basic operations on vectors

	(c) Jean-Philippe Thirion -- INRIA
*/
/*======================================================================*/

#if defined(INC_MIN) /* inclusions minimales */
 /* rajouter -DINC_MIN dans les options de compilation */
 /* GM 12/09/94                                        */
#include <epi_kernel.ei>
#else /* inclusions maximales */
#include <epidaure.h>
#include <epidaure.ee>
#include <epidaure.ei>
#endif

/************************************************************************/
/*
  BasicFloats
  
  Basic operations on floating points
*/
/************************************************************************/


/*-------Definition des fonctions statiques----------*/
#ifndef NO_PROTO
static void T_4MInvertAux( reeltype v[4], reeltype v1[4], reeltype v2[4], reeltype v3[4], reeltype det );
#else 
static void T_4MInvertAux();
#endif
/*-------Fin des proto statiques----------*/

/*
  Compare two integers

Returs: 1, if the first argument is greater, 0 if equal, -1 otherwise
*/

#if defined(_ANSI_)
int T_CompareInt( int *i1, int *i2 )
#else
int T_CompareInt(i1, i2 )
int *i1;
int *i2 ;

#endif

{
	if ( *i1 > *i2 ) return(1);
	else    if ( *i1 < *i2 ) return(-1);
	else    return(0);
}

/*
  Compare two floats

Returs: 1, if the first argument is greater, 0 if equal, -1 otherwise
*/

#if defined(_ANSI_)
int T_CompareFloat( reeltype *r1, reeltype *r2 )
#else
int T_CompareFloat(r1, r2 )
reeltype *r1;
reeltype *r2 ;

#endif

{
	if ( *r1 > *r2 ) return(1);
	else    if ( *r1 < *r2 ) return(-1);
	else    return(0);
}

/*
  Sort two floating point values

  After this operation, the first real number is the smallest.
*/

#if defined(_ANSI_)
void T_OrderFloat( reeltype *r1, reeltype *r2 ) /* r1 minimum and r2 maximum */
#else
void T_OrderFloat(r1, r2 )
reeltype *r1;
reeltype *r2 ;

#endif

{
	reeltype aux;
	if ( *r1 > *r2 ) { aux= *r1; *r1= *r2; *r2= aux; }
}

/************************************************************************/
/*
  Basic2Point

  Basic operations on 2D points
  */
/************************************************************************/

/*      Copy a point into another point                                 */

#if defined(_ANSI_)
void T_2Copy( t_2PointPtr pt1, t_2PointPtr pt2 )
#else
void T_2Copy(pt1, pt2 )
t_2PointPtr pt1;
t_2PointPtr pt2 ;

#endif

{
	pt2->x= pt1->x;
	pt2->y= pt1->y;
}

/*  Compute the norm of a point                                         */

#if defined(_ANSI_)
reeltype T_2Norm( t_2PointPtr thePt )
#else
reeltype T_2Norm(thePt )
t_2PointPtr thePt ;

#endif

{
	return( (reeltype) sqrt((double) thePt->x*thePt->x + thePt->y*thePt->y) );
}

/*      Normalize a vector                                              */

#if defined(_ANSI_)
void T_2Normalize( t_2PointPtr thePt, t_2PointPtr resPt )
#else
void T_2Normalize(thePt, resPt )
t_2PointPtr thePt;
t_2PointPtr resPt ;

#endif

{
	reeltype norm;
  
	norm= (reeltype) sqrt( (double) thePt->x*thePt->x + thePt->y*thePt->y );
	resPt->x= thePt->x / norm;
	resPt->y= thePt->y / norm;
}

/*      Compute the distance between two points                         */

#if defined(_ANSI_)
reeltype T_2Dist( t_2PointPtr pt1, t_2PointPtr pt2 )
#else
reeltype T_2Dist(pt1, pt2 )
t_2PointPtr pt1;
t_2PointPtr pt2 ;

#endif

{
	double xx,yy;
  
	xx= pt2->x - pt1->x;
	yy= pt2->y - pt1->y;
	return( (reeltype) sqrt( xx * xx + yy * yy ) );
}

/*      Multiply a vector with a real value                             */

#if defined(_ANSI_)
void T_2RMul( t_2PointPtr pt1, reeltype val, t_2PointPtr pt2 )
#else
void T_2RMul(pt1, val, pt2 )
t_2PointPtr pt1;
reeltype val;
t_2PointPtr pt2 ;

#endif

{
  pt2->x= pt1->x * val;
  pt2->y= pt1->y * val;
}

/*      Divide a vector with a real value                               */

#if defined(_ANSI_)
void T_2RDiv( t_2PointPtr pt1, reeltype val, t_2PointPtr pt2 )
#else
void T_2RDiv(pt1, val, pt2 )
t_2PointPtr pt1;
reeltype val;
t_2PointPtr pt2 ;

#endif

{
	if ( fabs( val ) < EPSY ){
		T_Warn("divide by zero in T_2RDiv\n", CNULL );
		pt2->x= 0.0;
		pt2->y= 0.0;
	}
  
	pt2->x= pt1->x / val;
	pt2->y= pt1->y / val;
}
/*      Compute the vector perpendicular to another one.                */

#if defined(_ANSI_)
void T_2VMul( t_2PointPtr pt1, t_2PointPtr pt2 )
#else
void T_2VMul(pt1, pt2 )
t_2PointPtr pt1;
t_2PointPtr pt2 ;

#endif

{
  pt2->x= -pt1->y;
  pt2->y= pt1->x;
}

/*
      Rescale the value of an angle around the value of another

	  The operation is performed +- PI
 */

#if defined(_ANSI_)
reeltype T_ModeAngle( reeltype theA, reeltype refA )
#else
reeltype T_ModeAngle(theA, refA )
reeltype theA;
reeltype refA ;

#endif
{
  double aux;
  aux= theA - refA;
  aux -= DEUXPI*floor(aux / DEUXPI);
  if ( aux > PI ) aux -= DEUXPI;
  return( (reeltype) ( aux + refA) );
}

/* idem with cartesian coord    */

#if defined(_ANSI_)
reeltype T_Mode2Angle( reeltype yy, reeltype xx, reeltype refA )
#else
reeltype T_Mode2Angle(yy, xx, refA )
reeltype yy;
reeltype xx;
reeltype refA ;

#endif
{
  double theA,aux;
  theA= atan2(yy,xx);
  aux= theA - refA;
  aux= aux - DEUXPI*floor(aux / DEUXPI);
  if ( aux > PI ) aux -= DEUXPI;
  return( (reeltype)( aux + refA) );
}

/*----------------------------------------------------------------------*/
/************************************************************************/
/*
  Basic3Point

  Basic operation on 3D points
  */
/************************************************************************/

/*		Find the maximum value of two value								*/

#if defined(_ANSI_)
reeltype T_MaxVal( reeltype v1, reeltype v2 )
#else
reeltype T_MaxVal(v1, v2 )
reeltype v1;
reeltype v2 ;

#endif

{
  if ( v1 > v2 ) 	return( v1 );
  else			return( v2 );
}

/*		Copy a point into another point									*/

#if defined(_ANSI_)
void T_3Copy( t_3PointPtr pt1, t_3PointPtr pt2 )
#else
void T_3Copy(pt1, pt2 )
t_3PointPtr pt1;
t_3PointPtr pt2 ;

#endif

{
  (void) memcpy( (void *) pt2, (void *) pt1, (size_t) sizeof(t_3Point) );
}

/*		form a 3D vector with 3 reeltypes								*/	

#if defined(_ANSI_)
void T_3Form( reeltype x, reeltype y, reeltype z, t_3PointPtr pt)
#else
void T_3Form(x, y, z, pt)
reeltype x;
reeltype y;
reeltype z;
t_3PointPtr pt;

#endif

{
  pt->x= x; pt->y= y; pt->z= z;
}

/*	Compute the norm of a point											*/

#if defined(_ANSI_)
reeltype T_3Norm( t_3PointPtr thePt )
#else
reeltype T_3Norm(thePt )
t_3PointPtr thePt ;

#endif

{
  return( (reeltype) sqrt((double) thePt->x*thePt->x + 
			  thePt->y*thePt->y +
			  thePt->z*thePt->z) );
}

/*		Normalize a vector												*/

#if defined(_ANSI_)
void T_3Normalize( t_3PointPtr thePt, t_3PointPtr resPt )
#else
void T_3Normalize(thePt, resPt )
t_3PointPtr thePt;
t_3PointPtr resPt ;

#endif

{
  reeltype norm;
  
  norm= (reeltype) sqrt( (double) thePt->x*thePt->x + thePt->y*thePt->y 
	     + thePt->z*thePt->z );
  if ( norm > 1e-20 ){
    resPt->x= thePt->x / norm;
    resPt->y= thePt->y / norm;
    resPt->z= thePt->z / norm;
  }
  else{ resPt->x= 0.0; resPt->y= 0.0; resPt->z= 0.0; }
}

/*		Compute the distance between two points							*/

#if defined(_ANSI_)
reeltype T_3Dist( t_3PointPtr pt1, t_3PointPtr pt2 )
#else
reeltype T_3Dist(pt1, pt2 )
t_3PointPtr pt1;
t_3PointPtr pt2 ;

#endif

{
  double xx,yy,zz;
  
  xx= pt2->x - pt1->x;
  yy= pt2->y - pt1->y;
  zz= pt2->z - pt1->z;
  return( (reeltype) sqrt( xx * xx + yy * yy + zz * zz ) );
}

/*		Multiply a vector with a real value								*/

#if defined(_ANSI_)
void T_3RMul( t_3PointPtr pt1, reeltype val, t_3PointPtr pt2 )
#else
void T_3RMul(pt1, val, pt2 )
t_3PointPtr pt1;
reeltype val;
t_3PointPtr pt2 ;

#endif

{
  pt2->x= pt1->x * val;
  pt2->y= pt1->y * val;
  pt2->z= pt1->z * val;
}

/*		Divide a vector with a real value								*/

#if defined(_ANSI_)
void T_3RDiv( t_3PointPtr pt1, reeltype val, t_3PointPtr pt2 )
#else
void T_3RDiv(pt1, val, pt2 )
t_3PointPtr pt1;
reeltype val;
t_3PointPtr pt2 ;

#endif

{
  if ( fabs( val ) < EPSY ){
    T_Warn("divide by zero in T_3RDiv\n", CNULL );
    pt2->x= 0.0;
    pt2->y= 0.0;	
    pt2->z= 0.0;	
  }
  
  pt2->x= pt1->x / val;
  pt2->y= pt1->y / val;
  pt2->z= pt1->z / val;
}

/*		Compute the scalar product										*/

#if defined(_ANSI_)
reeltype T_3SMul( t_3PointPtr pt1, t_3PointPtr pt2 )
#else
reeltype T_3SMul(pt1, pt2 )
t_3PointPtr pt1;
t_3PointPtr pt2 ;

#endif

{
  return( pt1->x*pt2->x + pt1->y*pt2->y + pt1->z*pt2->z );
}

/*		Compute the vectorial product									*/

#if defined(_ANSI_)
void T_3VMul( t_3PointPtr pt1, t_3PointPtr pt2, t_3PointPtr ptRes )
#else
void T_3VMul(pt1, pt2, ptRes )
t_3PointPtr pt1;
t_3PointPtr pt2;
t_3PointPtr ptRes ;

#endif

{
  reeltype xx,yy,zz;
  
  xx= pt1->y * pt2->z - pt1->z * pt2->y;
  yy= pt1->z * pt2->x - pt1->x * pt2->z;
  zz= pt1->x * pt2->y - pt1->y * pt2->x;
  ptRes->x= xx;
  ptRes->y= yy;
  ptRes->z= zz;
}

/*		Compute the determinant of three vectors						*/

#if defined(_ANSI_)
reeltype T_3Deter( t_3PointPtr p1, t_3PointPtr p2, t_3PointPtr p3 )
#else
reeltype T_3Deter(p1, p2, p3 )
t_3PointPtr p1;
t_3PointPtr p2;
t_3PointPtr p3 ;

#endif

{
  reeltype value;
  
  value=	p1->x*p2->y*p3->z + p2->x*p3->y*p1->z + p3->x*p1->y*p2->z -
    p3->x*p2->y*p1->z - p2->x*p1->y*p3->z - p1->x*p3->y*p2->z;
  return( value );
}

/*		Multiply each coefficient by each coefficient					*/

#if defined(_ANSI_)
void T_3Mult( t_3PointPtr pt1, t_3PointPtr pt2, t_3PointPtr ptRes )
#else
void T_3Mult(pt1, pt2, ptRes )
t_3PointPtr pt1;
t_3PointPtr pt2;
t_3PointPtr ptRes ;

#endif

{
  ptRes->x= pt1->x * pt2->x;
  ptRes->y= pt1->y * pt2->y;
  ptRes->z= pt1->z * pt2->z;
}

/*		Divide each coefficient by each coefficient pt1/pt2 -> ptRes	*/

#if defined(_ANSI_)
void T_3Div( t_3PointPtr pt1, t_3PointPtr pt2, t_3PointPtr ptRes )
#else
void T_3Div(pt1, pt2, ptRes )
t_3PointPtr pt1;
t_3PointPtr pt2;
t_3PointPtr ptRes ;

#endif

{
  if ( fabs(pt2->x) > EPSY ) ptRes->x= pt1->x / pt2->x;
  if ( fabs(pt2->y) > EPSY ) ptRes->y= pt1->y / pt2->y;
  if ( fabs(pt2->z) > EPSY ) ptRes->z= pt1->z / pt2->z;
}

/*		Add a vector to another one										*/

#if defined(_ANSI_)
void T_3VAdd( t_3PointPtr pt1, t_3PointPtr pt2, t_3PointPtr ptRes )
#else
void T_3VAdd(pt1, pt2, ptRes )
t_3PointPtr pt1;
t_3PointPtr pt2;
t_3PointPtr ptRes ;

#endif

{
  ptRes->x= pt1->x + pt2->x;
  ptRes->y= pt1->y + pt2->y;
  ptRes->z= pt1->z + pt2->z;
}

/*		Substract a vector to another one		pt1 - pt2 -> pt3	   	*/

#if defined(_ANSI_)
void T_3VSub( t_3PointPtr pt1, t_3PointPtr pt2, t_3PointPtr ptRes )
#else
void T_3VSub(pt1, pt2, ptRes )
t_3PointPtr pt1;
t_3PointPtr pt2;
t_3PointPtr ptRes ;

#endif

{
  ptRes->x= pt1->x - pt2->x;
  ptRes->y= pt1->y - pt2->y;
  ptRes->z= pt1->z - pt2->z;
}

/*		Verify that the vector is in the corresponding bounding box		*/

#if defined(_ANSI_)
int T_3Bound( t_3PointPtr pt, t_3PointPtr ptMin, t_3PointPtr ptMax )
#else
int T_3Bound(pt, ptMin, ptMax )
t_3PointPtr pt;
t_3PointPtr ptMin;
t_3PointPtr ptMax ;

#endif

{
  if ( (pt->x < ptMin->x) ||
      (pt->y < ptMin->y) ||
      (pt->z < ptMin->z) ||
      (pt->x > ptMax->x) ||
      (pt->y > ptMax->y) ||
      (pt->z > ptMax->z) ) return( FALSE );
  
  return( TRUE );
}

/*		Compute the two euler angles corresponding to a given vector	*/

#if defined(_ANSI_)
void T_3VEuler( t_3PointPtr thePt, reeltype *thetaPtr, reeltype *phiPtr )
#else
void T_3VEuler(thePt, thetaPtr, phiPtr )
t_3PointPtr thePt;
reeltype *thetaPtr;
reeltype *phiPtr ;

#endif

{
  double xx;
  
  xx= sqrt(thePt->x * thePt->x + thePt->y * thePt->y);
  if ( ( fabs(xx) < EPSY ) && (fabs(thePt->z) < EPSY) ){
    *thetaPtr= 0.0;
    *phiPtr= 0.0;
    return;	
  }
  
  *phiPtr= (reeltype) atan2( thePt->z, xx );
  if ( fabs(xx) < EPSY ){
    *thetaPtr= 0.0;
    return;
  }
  *thetaPtr= (reeltype) atan2( thePt->y, thePt->x );
}

/*		Compute the vector from the 2 polar angles						*/

#if defined(_ANSI_)
void T_3VEulerInv( reeltype theta, reeltype phi, t_3PointPtr thePt )
#else
void T_3VEulerInv(theta, phi, thePt )
reeltype theta;
reeltype phi;
t_3PointPtr thePt ;

#endif

{
  thePt->x=  (reeltype) (cos(theta)*cos(phi));
  thePt->y=  (reeltype) (sin(theta)*cos(phi));
  thePt->z=  (reeltype) sin(phi);
}

/*		Write a vector in an ascii file									*/

#if defined(_ANSI_)
int T_FWriteVect( FILE *fd, t_3PointPtr v1 )
#else
int T_FWriteVect(fd, v1 )
FILE *fd;
t_3PointPtr v1 ;

#endif

{
  return( fprintf( fd, "%f %f %f ", v1->x, v1->y, v1->z ) );
}

#if defined(_ANSI_)
int T_FWriteVectR( FILE *fd, t_3PointPtr v1 )	/* same with carriage return */
#else
int T_FWriteVectR(fd, v1 )
FILE *fd;
t_3PointPtr v1 ;

#endif

{
  return( fprintf( fd, "%f %f %f\n", v1->x, v1->y, v1->z ) );
}

#if defined(_ANSI_)
int T_FReadVect( FILE *fd, t_3PointPtr v1 )
#else
int T_FReadVect(fd, v1 )
FILE *fd;
t_3PointPtr v1 ;

#endif

{
  float x,y,z;
  int status;

  status = fscanf( fd, "%f %f %f", &x, &y, &z );
  v1->x =x; v1->y =y; v1->z =z; 
  return( status );
}

/*	Switch the content of two vectors									*/

#if defined(_ANSI_)
void T_3Switch( t_3PointPtr pt1, t_3PointPtr pt2 )
#else
void T_3Switch(pt1, pt2 )
t_3PointPtr pt1;
t_3PointPtr pt2 ;

#endif

{
  reeltype aux;
  
  aux= pt2->x; pt2->x= pt1->x; pt1->x= aux;
  aux= pt2->y; pt2->y= pt1->y; pt1->y= aux;
  aux= pt2->z; pt2->z= pt1->z; pt1->z= aux;
}

#if defined(_ANSI_)
void T_Switch( reeltype *pv1, reeltype *pv2 )
#else
void T_Switch(pv1, pv2 )
reeltype *pv1;
reeltype *pv2 ;

#endif

{
  reeltype aux;
  aux= *pv1; *pv1= *pv2; *pv2= aux;
}

/*	
  Tells if two 3D vectors are strictly equal (TRUE) else return FALSE
  */

#if defined(_ANSI_)
int T_3VEqual( t_3PointPtr pt1, t_3PointPtr pt2 )
#else
int T_3VEqual(pt1, pt2 )
t_3PointPtr pt1;
t_3PointPtr pt2 ;

#endif

{
  if ( (pt1->x != pt2->x) ||
      (pt1->y != pt2->y) ||
      (pt1->z != pt2->z) )	return( FALSE );
  else						return( TRUE );
}

/*
  Invert the direction of a given vector
  */

#if defined(_ANSI_)
void T_3VInvert( t_3PointPtr pt1 )
#else
void T_3VInvert(pt1 )
t_3PointPtr pt1 ;

#endif

{
  pt1->x= -pt1->x;
  pt1->y= -pt1->y;
  pt1->z= -pt1->z;
}

/*----------------------------------------------------------------------*/
/************************************************************************/
/*
  InterpPoints

  Fast interpolation of 3D and ND vector values

  The following routines can be used to perform a fast interpolation
  of a NDimensional vector of floating point values.
  
  the typical sequence is:

<pre>  
  T_NVClean( n, ptRes );
  T_NVIncrem( n, pt, 1.0, ptRes );
  ...
  T_NVIncrem( n, pt, 1.0, ptRes );
  T_NVRMul( n, ptRes, 1.0 / nb, ptRes );
</pre>
  */
/************************************************************************/

/*		interpolate a vector from two other vectors						*/

#if defined(_ANSI_)
void T_3VInter(t_3PointPtr pt1, t_3PointPtr pt2, reeltype a, t_3PointPtr ptRes )
#else
void T_3VInter(pt1, pt2, a, ptRes )
t_3PointPtr pt1;
t_3PointPtr pt2;
reeltype a;
t_3PointPtr ptRes ;

#endif

{
  reeltype b;
  
  b= 1 - a;
  ptRes->x= a*pt1->x + b*pt2->x;
  ptRes->y= a*pt1->y + b*pt2->y;
  ptRes->z= a*pt1->z + b*pt2->z;
}

/*
  Interpolate an n-dimensionnal vector from two other vectors
  */

#if defined(_ANSI_)
void T_NVInter(register int n,reeltype *pt1,reeltype *pt2,reeltype a,reeltype *ptRes)
#else
void T_NVInter(n, pt1, pt2, a, ptRes)
register int n;
reeltype *pt1;
reeltype *pt2;
reeltype a;
reeltype *ptRes;

#endif

{
  reeltype b;
  
  b= 1 - a;
  for( ;n>0;n--) (*ptRes++) = a * (*pt1++) + b * (*pt2++);
}


/*	Clean a n-dimensionnal vector */

#if defined(_ANSI_)
void T_NVClean( register int n, reeltype *ptRes )
#else
void T_NVClean(n, ptRes )
register int n;
reeltype *ptRes ;

#endif

{
  for( ;n>0;n--) (*ptRes++)= 0.0;
}

#if defined(_ANSI_)
void T_3Clean( t_3PointPtr ptRes )
#else
void T_3Clean(ptRes )
t_3PointPtr ptRes ;

#endif

{
  ptRes->x= 0.0;
  ptRes->y= 0.0;
  ptRes->z= 0.0;
}

/*	Increment a ND vector with a proportion of another vector */

#if defined(_ANSI_)
void T_NVIncrem( register int n, reeltype *pt1, reeltype a, reeltype *ptRes )
#else
void T_NVIncrem(n, pt1, a, ptRes )
register int n;
reeltype *pt1;
reeltype a;
reeltype *ptRes ;

#endif

{
  if ( a == 1.0 )
    for( ;n>0;n--) (*ptRes++) += (*pt1++);
  else	for( ;n>0;n--) (*ptRes++) += a * (*pt1++);
}

/*
  increment a 3D vector with a proportion of another vector
  */

#if defined(_ANSI_)
void T_3Increm( t_3PointPtr pt, reeltype a, t_3PointPtr res )
#else
void T_3Increm(pt, a, res )
t_3PointPtr pt;
reeltype a;
t_3PointPtr res ;

#endif

{
  T_NVIncrem( 3, (reeltype *) pt, a, (reeltype *) res );
}

/*	Multiply a ND vector with a real value */

#if defined(_ANSI_)
void T_NVRMul( register int n, reeltype *pt1, reeltype a, reeltype *ptRes )
#else
void T_NVRMul(n, pt1, a, ptRes )
register int n;
reeltype *pt1;
reeltype a;
reeltype *ptRes ;

#endif

{
  for( ;n>0;n--) (*ptRes++)= a * (*pt1++);
}

/*
  Bi-linear interpolation of 4 n-dimensional vectors
  
<pre>
  a1	  a2	
  pt0---pt1	b1
  |		|
  pt2---pt3	b2
</pre>
  */

#if defined(_ANSI_)
void T_NVBiInter( register int n, reeltype *pt0, reeltype *pt1, reeltype *pt2, reeltype *pt3, reeltype a1, reeltype b1, reeltype *ptRes )
#else
void T_NVBiInter(n, pt0, pt1, pt2, pt3, a1, b1, ptRes )
register int n;
reeltype *pt0;
reeltype *pt1;
reeltype *pt2;
reeltype *pt3;
reeltype a1;
reeltype b1;
reeltype *ptRes ;

#endif

{
  reeltype a2,b2;
  
  a2= 1 - a1;
  b2= 1 - b1;
  
  for( ;n>0;n--)
    (*ptRes++)= b1 * ( a1 * (*pt0++) + a2 * (*pt1++) ) +
      b2 * ( a1 * (*pt2++) + a2 * (*pt3++) );
}

/************************************************************************/
/*
  Basic3IPoint

  Basic operation on 3D integer points

  (c) Jean-Philippe THIRION -- INRIA -- 7/7/92
  */
/************************************************************************/

/*
  copy a 3D integer vector
*/

#if defined(_ANSI_)
void T_3ICopy( t_3IPointPtr pt1, t_3IPointPtr pt2 )
#else
void T_3ICopy(pt1, pt2 )
t_3IPointPtr pt1;
t_3IPointPtr pt2 ;

#endif
{
  (void) memcpy( (void *) pt2, (void *) pt1, (size_t) sizeof(t_3IPoint) );
}

/* transform a 3D point into an integer 3D point */
  
#if defined(_ANSI_)
void T_3to3I( t_3PointPtr pt, t_3IPointPtr ptI )
#else
void T_3to3I(pt, ptI )
t_3PointPtr pt;
t_3IPointPtr ptI ;

#endif
{
  ptI->x= (int) pt->x;
  ptI->y= (int) pt->y;
  ptI->z= (int) pt->z;
}

/* transform a 3D integer point into a 3D point */

#if defined(_ANSI_)
void T_3Ito3( t_3IPointPtr ptI, t_3PointPtr pt )
#else
void T_3Ito3(ptI, pt )
t_3IPointPtr ptI;
t_3PointPtr pt ;

#endif
{
  pt->x= (reeltype) ptI->x;
  pt->y= (reeltype) ptI->y;
  pt->z= (reeltype) ptI->z;
}

/*
  form a 3D integer point with 3 integers
*/	

#if defined(_ANSI_)
void T_3IForm( int x, int y, int z, t_3IPointPtr pt )
#else
void T_3IForm(x, y, z, pt )
int x;
int y;
int z;
t_3IPointPtr pt ;

#endif
{
  pt->x= x; pt->y= y; pt->z= z;
}

/*
  Compute the maximum value of 3 integers
*/

#if defined(_ANSI_)
int T_3IMax( int x, int y, int z )
#else
int T_3IMax(x, y, z )
int x;
int y;
int z ;

#endif
{
  if ( y > x ) x= y;
  if ( z > x ) x= z;
  return( x );
}

/*
  tell if a 3D integer point belongs to a bounding box

  The maximum values according to x,y,z of the box are not considered
  to be inside the box.

  Returns:
  TRUE if inside, else return FALSE
*/

#if defined(_ANSI_)
int T_3IBound( t_3IPointPtr pt, t_3IPointPtr ptMin, t_3IPointPtr ptMax )
#else
int T_3IBound(pt, ptMin, ptMax )
t_3IPointPtr pt;
t_3IPointPtr ptMin;
t_3IPointPtr ptMax ;

#endif
{
  if ( (pt->x < ptMin->x) ||
      (pt->y < ptMin->y) ||
      (pt->z < ptMin->z) ||
      (pt->x >= ptMax->x) ||
      (pt->y >= ptMax->y) ||
      (pt->z >= ptMax->z) ) return( FALSE );
  
  return( TRUE );
}

/*----------------------------------------------------------------------*/
/************************************************************************/
/*
  Basic4Mat

  Basic operations on 3D transformations

  Those 3D transforms are defined with homogeneous coordinates.

		(c) Jean-Philippe THIRION -- INRIA -- 13/8/91
*/
/************************************************************************/
/*		Apply a transform on a vector									*/

#if defined(_ANSI_)
void T_4TrApply( t_3PointPtr thePt, t_3PointPtr resPt, t_4TrPtr tr, int mode )
#else
void T_4TrApply(thePt, resPt, tr, mode )
t_3PointPtr thePt;
t_3PointPtr resPt;
t_4TrPtr tr;
int mode ;

#endif
{ 
  if ( mode & TR_NORM ){
    if ( mode & TR_INV )	T_4MTransApply( thePt, resPt, &tr->dir ); 
    else					T_4MTransApply( thePt, resPt, &tr->inv ); 
  }
  else{
    if ( mode & TR_INV )	T_4MApply( thePt, resPt, &tr->inv ); 
    else					T_4MApply( thePt, resPt, &tr->dir ); 
  }
}

/*		Apply the affine transform to a 3D vector						*/

#if defined(_ANSI_)
void T_4MApply( t_3PointPtr thePt, t_3PointPtr resPt, t_Mat4Ptr tr )
#else
void T_4MApply(thePt, resPt, tr )
t_3PointPtr thePt;
t_3PointPtr resPt;
t_Mat4Ptr tr ;

#endif
{
  reeltype x,y,z;
  
  x= thePt->x;
  y= thePt->y;
  z= thePt->z;
  
  resPt->x= tr->mt[0][0]*x + tr->mt[1][0]*y + tr->mt[2][0]*z + tr->mt[3][0];
  resPt->y= tr->mt[0][1]*x + tr->mt[1][1]*y + tr->mt[2][1]*z + tr->mt[3][1];
  resPt->z= tr->mt[0][2]*x + tr->mt[1][2]*y + tr->mt[2][2]*z + tr->mt[3][2];
}

/*		Apply the vectorial transform to a 3D vector					*/

#if defined(_ANSI_)
void T_4MApplyVect( t_3PointPtr thePt, t_3PointPtr resPt, t_Mat4Ptr tr )
#else
void T_4MApplyVect(thePt, resPt, tr )
t_3PointPtr thePt;
t_3PointPtr resPt;
t_Mat4Ptr tr ;

#endif
{
  reeltype x,y,z;
  
  x= thePt->x;
  y= thePt->y;
  z= thePt->z;
  
  resPt->x= tr->mt[0][0]*x + tr->mt[1][0]*y + tr->mt[2][0]*z;
  resPt->y= tr->mt[0][1]*x + tr->mt[1][1]*y + tr->mt[2][1]*z;
  resPt->z= tr->mt[0][2]*x + tr->mt[1][2]*y + tr->mt[2][2]*z;
}

/*		apply the transposed matrix the the 3D vector					*/

#if defined(_ANSI_)
void T_4MTransApply( t_3PointPtr thePt, t_3PointPtr resPt, t_Mat4Ptr tr )
#else
void T_4MTransApply(thePt, resPt, tr )
t_3PointPtr thePt;
t_3PointPtr resPt;
t_Mat4Ptr tr ;

#endif
{
  reeltype x,y,z;
  
  x= thePt->x;
  y= thePt->y;
  z= thePt->z;
  
  resPt->x= tr->mt[0][0]*x + tr->mt[0][1]*y + tr->mt[0][2]*z + tr->mt[0][3];
  resPt->y= tr->mt[1][0]*x + tr->mt[1][1]*y + tr->mt[1][2]*z + tr->mt[1][3];
  resPt->z= tr->mt[2][0]*x + tr->mt[2][1]*y + tr->mt[2][2]*z + tr->mt[2][3];
}

/*		Invert a 4D matrix												*/

#if defined(_ANSI_)
void T_4MInvert( t_Mat4Ptr m1, t_Mat4Ptr m2 )
#else
void T_4MInvert(m1, m2 )
t_Mat4Ptr m1;
t_Mat4Ptr m2 ;

#endif
{
  reeltype det;
  t_Mat4 m;
  
  det= T_4MDeterm( m1 );
  E_ERRN( fabs(det) < 1e-20, E_ERRDIV, "T_4MInvert", "null determinant" );
  
  T_4MInvertAux( m.mt[0], m1->mt[1], m1->mt[2], m1->mt[3], det );
  T_4MInvertAux( m.mt[1], m1->mt[2], m1->mt[3], m1->mt[0], -det );
  T_4MInvertAux( m.mt[2], m1->mt[3], m1->mt[0], m1->mt[1], det );
  T_4MInvertAux( m.mt[3], m1->mt[0], m1->mt[1], m1->mt[2], -det );
  T_4MTranspose( &m, m2 );
}

/* internally used in T_4MInver */

#if defined(_ANSI_)
static void T_4MInvertAux( reeltype v[4], reeltype v1[4], reeltype v2[4], reeltype v3[4], reeltype det )
#else
static void T_4MInvertAux(v, v1, v2, v3, det )
reeltype v[4];
reeltype v1[4];
reeltype v2[4];
reeltype v3[4];
reeltype det ;

#endif
{
  register int i0,i1,i2,i3;
  	double signe;
	double a1,a2,a3,a4,a5,a6;	/* to avoid compiler optimisations */
  
  for( i0=0; i0<4; i0++ ){
    i1= (i0+1)%4; i2= (i0+2)%4; i3= (i0+3)%4;
    if ( i0 % 2 ) 	signe= -1;
    else			signe= 1;
	a1= v1[i1] * v2[i2] * v3[i3];
	a2= v1[i2] * v2[i3] * v3[i1];
	a3= v1[i3] * v2[i1] * v3[i2];
	a4= v1[i1] * v2[i3] * v3[i2];
	a5= v1[i2] * v2[i1] * v3[i3];
	a6= v1[i3] * v2[i2] * v3[i1];
    v[i0]=  (reeltype) ( ( a1 + a2 + a3 - a4 - a5 - a6 ) * signe / det);
  }
}

/*		Compute the determinant of a 4x4 matrix							*/

#if defined(_ANSI_)
reeltype T_4MDeterm( t_Mat4Ptr m )
#else
reeltype T_4MDeterm(m )
t_Mat4Ptr m ;

#endif
{
  reeltype det;
  reeltype signe;
  register int i0,i1,i2,i3;
  
  det= 0;
  
  for( i0=0; i0<4; i0++ ){
    i1= (i0+1)%4; i2= (i0+2)%4; i3= (i0+3)%4;
    if ( i0 % 2 ) 	signe= -1;
    else			signe= 1;
    det += ( m->mt[1][i1] * m->mt[2][i2] * m->mt[3][i3] +
	    m->mt[1][i2] * m->mt[2][i3] * m->mt[3][i1] +
	    m->mt[1][i3] * m->mt[2][i1] * m->mt[3][i2] -
	    m->mt[1][i1] * m->mt[2][i3] * m->mt[3][i2] -
	    m->mt[1][i2] * m->mt[2][i1] * m->mt[3][i3] -
	    m->mt[1][i3] * m->mt[2][i2] * m->mt[3][i1]) * 
	      m->mt[0][i0] * signe;
  }
  return( det );
}
/*		Transpose a 4D matrix											*/

#if defined(_ANSI_)
void T_4MTranspose( t_Mat4Ptr m1, t_Mat4Ptr m2 )
#else
void T_4MTranspose(m1, m2 )
t_Mat4Ptr m1;
t_Mat4Ptr m2 ;

#endif

{
  reeltype aux;
  register int i,j;
  
  if ( m1 == m2 ){
    for( i=0; i<4; i++ )
      for( j=i+1; j<4; j++ ){
	aux= m1->mt[i][j];
	m1->mt[i][j]= m2->mt[j][i];
	m2->mt[j][i]= aux;
      }
  }
  else{
    for( i=0; i<4; i++ )
      for( j=0; j<4; j++ ) m2->mt[i][j]= m1->mt[j][i];
  }
}

/*		Compute thge identity matrix									*/

#if defined(_ANSI_)
void T_4MIdentity( t_Mat4Ptr m )
#else
void T_4MIdentity(m )
t_Mat4Ptr m ;

#endif
{
  int i,j;
  
  m->obj.type= OBJ_MAT4;
  
  for( i=0; i<4; i++ )
    for( j=0; j<4; j++ ){
      if ( i == j )	m->mt[i][j]= 1.0;
      else			m->mt[i][j]= 0.0;
    }
}

/*
  Set all the matrix values to 0
  */

#if defined(_ANSI_)
void T_4MClear( t_Mat4Ptr m )
#else
void T_4MClear(m )
t_Mat4Ptr m ;

#endif
{
  int i,j;
  
  m->obj.type= OBJ_MAT4;
  
  for( i=0; i<4; i++ )
    for( j=0; j<4; j++ ) m->mt[i][j]= 0.0;
}

/*	 Multiply two matrix: mRes= m1 * m2,	mRes(x)= m1( m2( x ) )		*/

#if defined(_ANSI_)
void T_4MMult( t_Mat4Ptr m1, t_Mat4Ptr m2, t_Mat4Ptr resM )
#else
void T_4MMult(m1, m2, resM )
t_Mat4Ptr m1;
t_Mat4Ptr m2;
t_Mat4Ptr resM ;

#endif
{
  t_Mat4 m;
  register int i,j,k;
  
  for( i=0; i<4; i++ )
    for( j=0; j<4; j++ ){
      
      m.mt[i][j]= 0;
      for( k=0; k<4; k++ )
	m.mt[i][j] += m1->mt[k][j] * m2->mt[i][k];	
    }
  
  T_4MCopy( &m, resM );	
}

/*	 Add two matrix: mRes= m1 + m2										*/

#if defined(_ANSI_)
void T_4MAdd( t_Mat4Ptr m1, t_Mat4Ptr m2, t_Mat4Ptr resM )
#else
void T_4MAdd(m1, m2, resM )
t_Mat4Ptr m1;
t_Mat4Ptr m2;
t_Mat4Ptr resM ;

#endif
{
  register int i,j;
  
  for( i=0; i<4; i++ )
    for( j=0; j<4; j++ )
      resM->mt[i][j]= m1->mt[i][j] + m2->mt[i][j];	
}

/*
  multiply the matrix coefficient with a real number
  */

#if defined(_ANSI_)
void T_4MRMul( t_Mat4Ptr m1, reeltype val, t_Mat4Ptr resM )
#else
void T_4MRMul(m1, val, resM )
t_Mat4Ptr m1;
reeltype val;
t_Mat4Ptr resM ;

#endif
{
  register int i,j;
  
  for( i=0; i<4; i++ )
    for( j=0; j<4; j++ )
      resM->mt[i][j]= m1->mt[i][j] * val;	
}

/*		Copy a 4d matrix into another one								*/

#if defined(_ANSI_)
void T_4MCopy( t_Mat4Ptr theM, t_Mat4Ptr resM )
#else
void T_4MCopy(theM, resM )
t_Mat4Ptr theM;
t_Mat4Ptr resM ;

#endif
{
  memcpy( (void *) resM, (void *) theM, (size_t) sizeof(t_Mat4) );
  resM->obj.type= OBJ_MAT4;
}

/*		Apply a translation to the matrix								*/

#if defined(_ANSI_)
void T_4MTranslate( t_Mat4Ptr theM, t_3PointPtr v )
#else
void T_4MTranslate(theM, v )
t_Mat4Ptr theM;
t_3PointPtr v ;

#endif
{
  theM->mt[3][0] += v->x;
  theM->mt[3][1] += v->y;
  theM->mt[3][2] += v->z;
}

/*		Apply a rotation to a matrix									*/

#if defined(_ANSI_)
void T_4MRotate( t_Mat4Ptr theM, reeltype angle, int dir )
#else
void T_4MRotate(theM, angle, dir )
t_Mat4Ptr theM;
reeltype angle;
int dir ;

#endif
{
  t_Mat4 m;
  register int i1,i2;
  
  T_4MIdentity( &m );
  i1= (dir+1) % 3;
  i2= (dir+2) % 3;
  m.mt[i1][i1]=  (reeltype) cos( angle );
  m.mt[i2][i1]=  (reeltype) -sin( angle );
  m.mt[i1][i2]=  (reeltype) sin( angle );
  m.mt[i2][i2]=  (reeltype) cos( angle );
  T_4MMult( &m, theM, theM );
}
/*
  Compute rotation matrix from its rotation vector

  Compute the rotation matrix corresponding to the rotation vector
the axis of the rotation is the rotation vector and the angle is
the norm of the rotation vector
*/

#if defined(_ANSI_)
void T_ComputeRotation(t_3PointPtr v ,t_Mat4Ptr m)
#else
void T_ComputeRotation(v,m)
t_3PointPtr v;
t_Mat4Ptr m;
#endif
{
	reeltype norm,cosinus,f,g,xx,yy,zz,xy,yz,zx;
	
	xx = v->x * v->x ;
	yy = v->y * v->y ;
	zz = v->z * v->z ;
	norm = (reeltype) sqrt((double)(xx + yy + zz));

	if(norm <EPSY)T_4MIdentity(m);
	else{
		cosinus = (reeltype)cos((double)norm);
		f = (reeltype)sin((double)norm)/norm ;
		g = (1 - cosinus)/(norm * norm);
		
		xy = v->x * v->y * g;
		yz = v->y * v->z * g;
		zx = v->z * v->x * g;

		xx *= g ;
		yy *= g;
		zz *= g;

		m->mt[0][0] = cosinus + xx  ; m->mt[1][0] = xy - v->z * f ; m->mt[2][0] = zx + v->y * f ; 
		m->mt[0][1] = xy + v->z * f ; m->mt[1][1] = cosinus + yy  ; m->mt[2][1] = yz - v->x * f ; 
		m->mt[0][2] = zx - v->y * f ; m->mt[1][2] = yz + v->x * f ; m->mt[2][2] = cosinus + zz  ; 
	}
}

/*
  Get the rotation vector from the rotation matrix (not tested)
*/

#if defined(_ANSI_)
void T_ComputeRotationVector(t_Mat4Ptr m,t_3PointPtr axis)
#else
void T_ComputeRotationVector(m,axis)
t_Mat4Ptr m;
t_3PointPtr axis;
#endif
{
  double xx,yy,zz,xy,yz,zx,g,cosinus,sinus,angle,A;
  
  cosinus =  0.5   * (m->mt[0][0] + m->mt[1][1] +  m->mt[2][2] -1.0);
  
  if(cosinus>1.0) cosinus = 1.0;
  else if (cosinus<-1.0)  cosinus = -1.0;
  
  angle = acos( cosinus);
  
  if (angle==0.0){
    axis->x = 0.0 ;
    axis->y = 0.0 ;	
    axis->z = 0.0 ;
  }
  else{ 
    sinus = (reeltype) sin((double)angle);
    
    if(cosinus > -0.5){
      if((angle<EPSY)&&(angle>-EPSY))
	A = 0.5 * ( 1.0 + angle * angle /6.0);
      else A = 0.5 * angle / sinus;
      axis->x = (reeltype) (A * (m->mt[1][2] - m->mt[2][1]) );
      axis->y = (reeltype) (A * (m->mt[2][0] - m->mt[0][2]) );	
      axis->z = (reeltype) (A * (m->mt[0][1] - m->mt[1][0]) );
    }
    else {
      g = (1.0 - cosinus ) / (angle * angle) ;
      
      xx = (m->mt[0][0] - cosinus) / g;
      yy = (m->mt[1][1] - cosinus) / g;
      zz = (m->mt[2][2] - cosinus) / g;
      
      if(xx<0.0) xx *= -1.0;
      if(yy<0.0) yy *= -1.0;
      if(zz<0.0) zz *= -1.0;
      
      yz = (m->mt[2][1] + m->mt[1][2]) / g ;
      zx = (m->mt[0][2] + m->mt[2][0]) / g ;
      xy = (m->mt[1][0] + m->mt[0][1]) / g ;
      
      if(cosinus == -1.0){
	axis->x = (reeltype) sqrt((double)xx);
	if(xy > 0.0) axis->y = (reeltype) sqrt((double)yy);
	else axis->y = (reeltype)( - sqrt((double)yy));
	if(zx > 0.0)axis->z = (reeltype)sqrt((double)zz);
	else axis->z = (reeltype) ( - sqrt((double)zz));
      }
      else{
	if(m->mt[1][2] > m->mt[2][1])axis->x = (reeltype)sqrt((double)xx);
	else axis->x = - (reeltype) sqrt((double)xx);
	if(m->mt[2][0] > m->mt[0][2])axis->y = (reeltype)sqrt((double)yy);
	else axis->y = - (reeltype)sqrt((double)yy);
	if(m->mt[0][1] > m->mt[1][0])axis->z = (reeltype)sqrt((double)zz);
	else axis->z = - (reeltype)sqrt((double)zz);
      }
    }
  }
}
/*		Get the translation vector of the matrix						*/

#if defined(_ANSI_)
void T_4MTranslation( t_Mat4Ptr theM, t_3PointPtr v )
#else
void T_4MTranslation(theM, v )
t_Mat4Ptr theM;
t_3PointPtr v ;

#endif
{
  
	v->x = theM->mt[3][0] ;
	v->y = theM->mt[3][1] ;
	v->z = theM->mt[3][2] ;
}

/*		Multiply the matrix by 3 given values							*/

#if defined(_ANSI_)
void T_4MHomo( t_Mat4Ptr theM, t_3PointPtr v )
#else
void T_4MHomo(theM, v )
t_Mat4Ptr theM;
t_3PointPtr v ;

#endif
{
  register int i;
  
  for( i=0; i< 4; i++ ){
    theM->mt[i][0] *= v->x;
    theM->mt[i][1] *= v->y;
    theM->mt[i][2] *= v->z;
  }
}

/*		Verify the matrix operations									*/

#if defined(_ANSI_)
int T_4MVerify( void )
#else
int T_4MVerify()

#endif

{
  t_Mat4 m1,m2,m3;
  reeltype det1,det2;
  int j,good= TRUE;
  
  for( j=0; j<4; j++ ) m1.mt[0][j]= (reeltype) 1;
  for( j=0; j<4; j++ ) m1.mt[1][j]= (reeltype) j;
  for( j=0; j<4; j++ ) m1.mt[2][j]= (reeltype) j*j;
  for( j=0; j<4; j++ ) m1.mt[3][j]= (reeltype) j*j*j;
  
  det1= T_4MDeterm( &m1 );
  T_4MInvert( &m1, &m2 );
  T_4MMult( &m1, &m2, &m3 );
  det2= T_4MDeterm( &m3 );
  if ( fabs(det2 - 1) > EPSY ) good= FALSE;
  
  T_4MTranspose( &m1, &m2 );
  T_4MInvert( &m2, &m3 );
  T_4MTranspose( &m3, &m2 );
  T_4MInvert( &m2, &m3 );
  det2= T_4MDeterm( &m3 );
  if ( fabs(det1 - det2) > 1e-20 ) good= FALSE;
  
  return( good );
}

/*----------------------------------------------------------------------*/
