/* epidaure.c 2.3 5/14/98 */
/*==========================================================================
*
* File : $EPI/src/epidaure.c
* Module :
* $ Id : ? $ 
* Test file : ?
* Author : J-P Thirion
* Creation: August 4 1992
*===========================================================================
*	(c) Copyright 1990-1992, Inria, all right reserved
*===========================================================================
*
*------------------------------Description----------------------------------
*
* fonctions bas niveau ge'ne'rales de la bibliothe`que Epidaure.lib
*
*-----------------------------Modification----------------------------------
*
*/

/*======================================================================*/
/*
	epidaure.c

  	Low level functions in EpidaureLib

	It includes error processing, memory allocation, hash table
	processing...

  (c) Jean-Philippe THIRION -- INRIA -- october 25 1990
*/
/*======================================================================*/

#if defined(INC_MIN) /* inclusions minimales */
 /* rajouter -DINC_MIN dans les options de compilation */
 /* GM 12/09/94                                        */
#include <epi_kernel.ei>
#include <epi_var.ee>
#else /* inclusions maximales */
#include <epidaure.h>
#include <epidaure.ee>
#include <epidaure.ei>
#endif

/*-------Definition des fonctions statiques----------*/
#ifndef NO_PROTO
#else 
#endif
/*-------Fin des proto statiques----------*/


/************************************************************************/
/*
	Errors

  	Error processing

  (c) Jean-Philippe THIRION -- INRIA -- october 25 1990
*/
/************************************************************************/

static int e_gError= E_ERRNO;		/* fatal error reference number			*/
char *e_gTbErr[E_ERRNB];		/* list of standard fatal messages		*/

/*
	Error management system. 

This routine handle the fatal errors in a more appropriate way than T_Err
*/

#if defined(_ANSI_)
int E_Error(
int no, 	/* Reference number of the error */
char *proc, /* procedure where error occured */
char *mess 	/* complementary message */
)
#else
int E_Error(no, proc, mess )
int no;
char *proc;
char *mess ;

#endif

{
    if ( no && !e_gError ){
        if ( (no >= E_ERRNB) || (no < 0) ) e_gError= E_ERRFATAL;
        else e_gError= no;
        if ( verbose )
            fprintf( stderr, "%s error: %s\n", program, e_gTbErr[e_gError] );
    }

    if ( e_gError ){
        if ( ((mess != NULL)||(proc !=NULL)) && verbose ){
			if (mess != NULL) 	fprintf( stderr, "  in %s (%s)\n", proc, mess );
			else 				fprintf( stderr, "  in %s\n", proc );
			}
        return(TRUE);
    }

    return( FALSE );
}

/*
	Reset the error flag
*/

#if defined(_ANSI_)
void E_ErrorReset( void )
#else
void E_ErrorReset()

#endif

{
    e_gError= E_ERRNO;
	}

/*
	Initialize the standard fatal error message array
*/

#if defined(_ANSI_)
void E_FatalInit( void )
#else
void E_FatalInit()

#endif

{
	e_gTbErr[E_ERRNO]		= EM_ERRNO;
	e_gTbErr[E_ERRUNKNOWN]	= EM_ERRUNKNOWN;
	e_gTbErr[E_ERRC]		= EM_ERRC;
	e_gTbErr[E_ERRO]		= EM_ERRO;
	e_gTbErr[E_ERRW]		= EM_ERRW;
	e_gTbErr[E_ERRR]		= EM_ERRR;
	e_gTbErr[E_ERRMEM]		= EM_ERRMEM;
	e_gTbErr[E_ERRDIV]		= EM_ERRDIV;
	e_gTbErr[E_ERRFATAL]	= EM_ERRFATAL;
	e_gTbErr[E_ERRARRAY]	= EM_ERRARRAY;
	e_gTbErr[E_ERRNULL]		= EM_ERRNULL;
	e_gTbErr[E_ERRMISMATCH]	= EM_ERRMISMATCH;
	}

/*	Set an extend postfixed name to a given name						*/

#if defined(_ANSI_)
void T_SetExtend( char *name1, char *name2, char *extend )
#else
void T_SetExtend(name1, name2, extend )
char *name1;
char *name2;
char *extend ;

#endif

{
	char *pos;
	int l;

	l= strlen(name1) + strlen( extend );

	E_ERRN( l > B_FNAME_LENGTH, E_ERRFATAL, "T_SetExtend", "name too long" );

	strncpy( name2, name1, B_FNAME_LENGTH );
	pos= strchr( name2, '.' );
	if ( pos != NULL ) *pos= 0;
	strcat( name2, extend );
	}
	
/*			Print an error message on the screen						*/

#if defined(_ANSI_)
void T_Err( char *str1, char *str2 )
#else
void T_Err(str1, str2 )
char *str1;
char *str2 ;

#endif

{
	if( verbose ){
		fprintf( stderr, "%s erreur :", program );
		fprintf( stderr, str1, str2 );
		}
	(void) exit(0);
	}

/*			Print an warning message on the screen						*/

#if defined(_ANSI_)
void T_Warn( char *str1, char *str2 )
#else
void T_Warn(str1, str2 )
char *str1;
char *str2 ;

#endif

{
	if ( verbose ){
		fprintf( stderr, "%s warning :", program );
		fprintf( stderr, str1, str2 );
		}
	}

/*			Print a message on the screen								*/

#if defined(_ANSI_)
void T_Mess( char *str1, char *str2 )
#else
void T_Mess(str1, str2 )
char *str1;
char *str2 ;

#endif

{
	if ( verbose ){
		fprintf( stderr, "%s message :", program );
		fprintf( stderr, str1, str2 );
		}
	}

/*	Dynamic allocation of memory										*/

#if defined(_ANSI_)
char *T_Alloc(unsigned size )
#else
char *T_Alloc(size )
unsigned size ;

#endif

{
	char *buf;

	buf= malloc( (unsigned) size );
	E_ERR( buf==NULL, E_ERRMEM, NULL, "T_Alloc", NULL );
	return( buf );
}

/*
	Dynamic re-allocation of memory
*/

#if defined(_ANSI_)
char *T_ReAlloc( char **handle,int size )
#else
char *T_ReAlloc(handle, size )
char **handle;
int size ;

#endif

{
	char *buf ;

	buf= realloc( *handle, (unsigned) size );
	E_ERR( buf==NULL, E_ERRMEM, NULL, "T_ReAlloc", NULL );
	return( buf );
}
/*
	Free a handle, with the convention that a null pointer means not
	allocated handle
*/

#if defined(_ANSI_)
void T_Free( void **handle )
#else
void T_Free( handle )
void **handle;
#endif

{
	if (*handle == NULL) return;
	free( *handle );
	*handle= NULL;
	}

/*----------------------------------------------------------------------*/
/************************************************************************/
/*	
	stacks

			stack manager (obsolete)
		(c) Jean-Philippe THIRION -- INRIA --
*/
/************************************************************************/

#define NB_STACK_MAX 128		/* maximum number of stacks				*/
#define B_SNAME_LENGTH 80

typedef struct stackRec{

	char name[B_SNAME_LENGTH];		/* name of the items				*/
	int	size;			/* of an item									*/
	int nb_max;			/* maximum number of items in the stack			*/
	char *first;		/* pointer toward the first item of the stack	*/
	char *next;			/* pointer toward the next item of the stack	*/
	char *maxi;			/* first not valid address in the stack			*/

	} stackRec, *stackRecPtr;

static stackRecPtr stackTb[128];
static char stackStarted= FALSE;

/*
	Create a new stack record, and return the number of the stack

	Return -1 if not able to do so.
*/

#if defined(_ANSI_)
int StackCreate(
char *name, /*  name of the items */
int size, 	/* size in bytes of an item */
int nb_max 	/* maximum number of items */
)
#else
int StackCreate(name, size, nb_max )
char *name;
int size;
int nb_max ;

#endif

{
	int i;

	E_ERR( nb_max < 1,E_ERRFATAL,-1,
		   "StackCreate", "Wrong max number of items");

	if ( stackStarted == FALSE ){
		for( i=0; i< NB_STACK_MAX; i++ ) stackTb[i]= NULL;
		stackStarted= TRUE;
		}

	i=0;
	while( (i < NB_STACK_MAX ) && (stackTb[i] != NULL) ) i++;

	E_ERR(i >= NB_STACK_MAX,E_ERRFATAL,-1,"StackCreate", 
			"No free stack available");

	stackTb[i]= (stackRecPtr) T_Alloc(sizeof(stackRec));
	E_ISERR( -1,"StackCreate", NULL);

	strncpy( stackTb[i]->name, name, B_SNAME_LENGTH );
	stackTb[i]->size	= size;
	stackTb[i]->nb_max	= nb_max;
	stackTb[i]->first	= T_Alloc( size * nb_max );
	stackTb[i]->next	= stackTb[i]->first;
	stackTb[i]->maxi	= stackTb[i]->first + size * nb_max;

	E_REPERR({ free((char *) stackTb[i]); stackTb[i]= NULL; });
	E_ISERR( -1,"StackCreate",NULL);
	return(i);
	}

/*
	Allocate a new item of a given stack
*/

#if defined(_ANSI_)
char *StackPush(
int stackNum 	/* number of the stack */
)
#else
char *StackPush(stackNum )
int stackNum ;

#endif

{
	stackRecPtr theStack;
	char		*theItem;

	E_ERR( ( stackNum < 0 ) || ( stackNum >= NB_STACK_MAX ),
		E_ERRARRAY, NULL, "StackPush", "Out of range stack number" );

	theStack= stackTb[ stackNum ];
	E_ERR( theStack == NULL, E_ERRUNKNOWN, NULL, "StackPush",
		   "Not a created stack" );

	theItem= theStack->next;
	if ( theItem >= theStack->maxi ){
		sprintf( tmpStr1, "No more items available in stack %s", 
				 theStack->name );
		E_SETERR( E_ERRARRAY, "StackPush", tmpStr1 );
		}

	E_ISERR( NULL, "StackPush", NULL );

	theStack->next += theStack->size;
	return( theItem );
	}

/*
	Release a given number of items into a stack
*/

#if defined(_ANSI_)
char *StackPop(
int stackNum,	/* number of the stack */
int nb			/* number of items to release */
)
#else
char *StackPop(stackNum, nb )
int stackNum;
int nb ;

#endif

{
	stackRecPtr theStack;

	E_ERR( ( stackNum < 0 ) || ( stackNum >= NB_STACK_MAX ), 
		E_ERRARRAY, NULL, "StackPop", "Out of range stack number" );

	theStack= stackTb[ stackNum ];
	E_ERR( theStack == NULL, E_ERRUNKNOWN, NULL,
		   "StackPop", "Not a created stack" );

	theStack->next -= theStack->size * nb;
	if ( theStack->next < theStack->first ){
		sprintf( tmpStr1, "Not so many allocated items %s", 
				 theStack->name );
		E_XERR( E_ERRARRAY, NULL, "StackPop", tmpStr1 );
		}

	if( theStack->next == theStack->first ) 
			return( NULL );								/* empty stack	*/
	else	return( theStack->next - theStack->size );
	}

/*
	Delete a given stack
*/

#if defined(_ANSI_)
void StackDelete( 
int stackNum 	/* number of the stack */
)
#else
void StackDelete(stackNum )
int stackNum ;

#endif

{
	E_ERRN( ( stackNum < 0 ) || ( stackNum >= NB_STACK_MAX ), 
		E_ERRARRAY, "StackDelete", "Out of range stack number" );

	E_ERRN( (stackTb[stackNum] == NULL), E_ERRUNKNOWN, 
		 	  "StackDelete", "Not a created stack" );

	free( (char *) stackTb[ stackNum ]->first );
	free( (char *) stackTb[ stackNum ] );
	stackTb[stackNum]= NULL;
	}

/*----------------------------------------------------------------------*/
/************************************************************************/
/*
	HashTables

		Building and using hashing tables

		(c) Jean-Philippe THIRION -- INRIA -- 24/7/91
*/
/************************************************************************/

/*
	Initialize a H table
*/

#if defined(_ANSI_)
void T_InitHTable( 
t_HTablePtr theHTable, /* the hash table */
int size, /* number of significant digits in the HTable (12 ?) */
fCode_type fCode	/* code function, returns an int charact. the item */
)
#else
void T_InitHTable(theHTable, size, fCode )
t_HTablePtr theHTable;
int size;
fCode_type fCode ;

#endif
{
	int i;

	E_ERRN( size > 31, E_ERRARRAY,"T_AllocHTable", "Too large htable" );

	theHTable->size= size;
	theHTable->nb= 1 << size;
	theHTable->tb= 
		(t_EntryHand) T_Alloc((int) (theHTable->nb * sizeof(t_EntryPtr)));

	E_ISERRN("T_InitHTable",NULL);

	for( i=0; i< theHTable->nb; i++ ) theHTable->tb[i]= NULL;
	theHTable->search= NULL;
	theHTable->code= 0;
	theHTable->fCode= fCode;
	theHTable->filter= NULL;
	theHTable->mask= (1 << size) - 1;
	theHTable->data= NULL;
	}

/*
	Clear the hashing table, but do not remove the items

	which are attached to them.
*/

#if defined(_ANSI_)
void T_ClearHTable( t_HTablePtr theHTable )
#else
void T_ClearHTable(theHTable )
t_HTablePtr theHTable ;

#endif
{
	int i;
	for( i=0; i< theHTable->nb; i++ ) T_FreeLEntry( &(theHTable->tb[i]) );
	}

/*
	Free the hash table, and remove the buffer

	Do not destroy the items which are attached to it.
*/

#if defined(_ANSI_)
void T_DeleteHTable( t_HTablePtr theHTable )
#else
void T_DeleteHTable(theHTable )
t_HTablePtr theHTable ;

#endif
{
	T_ClearHTable( theHTable );
	T_Free( (void **) &theHTable->tb );
	theHTable->nb= 0;
	} 

/*
	Free a linked list of entries.

	But not the items which are attached to it.
*/

#if defined(_ANSI_)
void T_FreeLEntry( t_EntryHand handEntry )
#else
void T_FreeLEntry(handEntry )
t_EntryHand handEntry ;

#endif
{
	t_EntryPtr theEntry,nextEntry;

	theEntry= *handEntry;

	while( theEntry != NULL ){

		nextEntry= theEntry->next;
		free( (char *) theEntry );
		theEntry= nextEntry;
		}

	*handEntry= NULL;
	}

/*
	Free all the entries with most than nbMax entries. 

	This can be used to remove efficiently non-discriminent points.
*/

#if defined(_ANSI_)
void T_HashDiscriminate( t_HTablePtr theHTable, int nbMax )
#else
void T_HashDiscriminate(theHTable, nbMax )
t_HTablePtr theHTable;
int nbMax ;

#endif
{
	int count,i;

	for( i=0; i< theHTable->nb; i++ ){

		count= T_HashNbEntries( theHTable, i );
		if ( count > nbMax ) T_FreeLEntry( &theHTable->tb[i] );
		}
	}

/*		Compute statistiques about the hash table						*/

#if defined(_ANSI_)
void T_HashStat( t_HTablePtr theHTable )
#else
void T_HashStat(theHTable )
t_HTablePtr theHTable ;

#endif
{
	int nbTotal,nbNull,lMax,i,length;
	t_EntryPtr theEntry;

	nbTotal= 0;
	nbNull= 0;
	lMax= 0;

	for( i=0; i<theHTable->nb; i++ ){

		theEntry= theHTable->tb[i];
		length= 0;
		
		if ( theEntry == NULL ) nbNull += 1;
		else{
			length= T_HashNbEntries( theHTable, i );
			nbTotal += length;
			if ( length > lMax ) lMax= length;
			}
		}

	printf("Number of entry lists:              %d\n", theHTable->nb );
	printf("Total number of entries:            %d\n", nbTotal );
	printf("Number of void entry lists:         %d\n", nbNull );
	printf("Maximum number of entry in a list:  %d\n", lMax );
	printf("\n");
	}

/*
	Return the number of entries of a given entry of the hash table.
*/

#if defined(_ANSI_)
int T_HashNbEntries( t_HTablePtr theHTable, int i )
#else
int T_HashNbEntries(theHTable, i )
t_HTablePtr theHTable;
int i ;

#endif
{
	int count;
	t_EntryPtr theEntry;
	
	E_ERR( (i < 0) || (i >= theHTable->nb), E_ERRARRAY, 0, 
			"T_HashNbEntries", "not that much buckets int the Hash table" );
	
	theEntry= theHTable->tb[i];
	count= 0;
	while( theEntry ){ count +=1; theEntry= theEntry->next; }
	return( count );
	}

/*
	Return the number of the longest entry list of the HTable
*/

#if defined(_ANSI_)
int T_HashLongest( t_HTablePtr theHTable )
#else
int T_HashLongest(theHTable )
t_HTablePtr theHTable ;

#endif
{
	int i;
	int count;
	int countMax= 0;
	int numMax= 0;

	for( i=0; i< theHTable->nb; i++ ){

		count= T_HashNbEntries( theHTable, i );
		if ( count > countMax ){
			countMax= count;
			numMax= i;
			}
		}

	return( numMax );
	}

/*
	Add a new entry into the table
*/

#if defined(_ANSI_)
void T_AddEntry( 
char *theItem, 	/* item to be added the the hashing table */
char *data, 	/* data used for the hashing code computation */
t_HTablePtr theHTable	/* the Hash table itself */
)
#else
void T_AddEntry(theItem, data, theHTable )
char *theItem;
char *data;
t_HTablePtr theHTable ;

#endif
{
	t_EntryPtr theEntry;
	int pos;

	theEntry= (t_EntryPtr) T_Alloc( sizeof(t_Entry) );
	E_ISERRN("T_AddEntry", NULL);

	if ( data == NULL ) data= theItem;
	theEntry->code= theHTable->fCode( data );
	theEntry->item= theItem;

	pos= theEntry->code & theHTable->mask;
	theEntry->next= theHTable->tb[pos];
	theHTable->tb[pos]= theEntry;
	}

/*
	Find the first occurence of an item with a given code

	The filter takes two arguments as imput: a pointer toward the
	item to check, and a pointer toward the data which has been used
	to compute the code. It returns zero if the item is "a good one"
*/

#if defined(_ANSI_)
char *T_FindEntry( 
t_HTablePtr theHTable, 	/* data to be compared with */
char *data, 			/* filter function, can be NULL */
filter_type filter 		/* return 0 if ok */
)
#else
char *T_FindEntry(theHTable, data, filter )
t_HTablePtr theHTable;
char *data;
filter_type filter ;

#endif

{
	int pos,code;

	code= theHTable->fCode( data ); 
	pos= code & theHTable->mask;
	theHTable->code= code;
	theHTable->data= data;
	theHTable->search= theHTable->tb[pos];
	theHTable->filter= filter;

	return( T_NextEntry( theHTable ) );
	}

/*
	Find the next occurence of an entry with the same code. 

This function must be called only after T_FindEntry is called.
*/

#if defined(_ANSI_)
char *T_NextEntry( t_HTablePtr theHTable )
#else
char *T_NextEntry(theHTable )
t_HTablePtr theHTable ;

#endif
{ 
	t_EntryPtr theEntry; 
	int code,good;

	theEntry= theHTable->search;
	code= theHTable->code;
	good= FALSE;

	if ( theHTable->filter != NULL ){
		while( 	(theEntry != NULL) && (good != TRUE) ){ 
			if ((theEntry->code == code) && 
				(theHTable->filter(theEntry->item,theHTable->data) == 0) )
				good= TRUE;
			else theEntry= theEntry->next;
			}
		}
	else{
		while( 	(theEntry != NULL) && (good != TRUE) ){ 
			if (theEntry->code == code) good= TRUE;
			else theEntry= theEntry->next;
			}
		}

	if (good == FALSE) return( CNULL );
	else{
		theHTable->search= theEntry->next;
		return( theEntry->item );
		}
	}

/*	
	Compute the code for 2D hashing. 

	It assumes that the size of the	hashing table >= 8 (256 elements)
*/

#if defined(_ANSI_)
int T_2DCode( t_2PointPtr pt )
#else
int T_2DCode(pt )
t_2PointPtr pt ;

#endif
{
	int code;
	unsigned int x,y;

	x= (int) pt->x;
	y= (int) pt->y;

	code= x ^ y ^ (y << 4);
	return( code );
	}

/*
	Compute the code for 3D hashing with floating point coordinates.

	It assumes that the size of the	hashing table >= 12 (4096 elements)
*/

#if defined(_ANSI_)
int T_3DCode( t_3PointPtr pt )
#else
int T_3DCode(pt )
t_3PointPtr pt ;

#endif
{
	int code;
	unsigned int x,y,z;

	x= (int) pt->x;
	y= (int) pt->y;
	z= (int) pt->z;

	code= x^y^z^(z << 8)^(y<< 4);
	return( code );
	}

/*
	Compute the code for 3D hashing with integer coordinates.

	It assumes that the size of the	hashing table >= 12 (4096 elements)
*/

#if defined(_ANSI_)
int T_3DICode( t_3IPointPtr pt )
#else
int T_3DICode(pt )
t_3IPointPtr pt ;

#endif
{
  int code;
  
  code= pt->x^pt->y^pt->z^(pt->z << 8)^(pt->y<< 4);
  return( code );
}

/*
	Compute the code for an EpidaureLib 3D point

	It assumes that the size of the	hashing table >= 12 (4096 elements)
*/

#if defined(_ANSI_)
int T_PointCode( t_PointPtr aPt )
#else
int T_PointCode(aPt )
t_PointPtr aPt ;

#endif
{
  return( T_3DCode(&aPt->pt) );
}

/*		Example of filtering function on 3D points						*/

#if defined(_ANSI_)
int T_3DIntFilter( t_3PointPtr pt1, t_3PointPtr pt2 )
#else
int T_3DIntFilter(pt1, pt2 )
t_3PointPtr pt1;
t_3PointPtr pt2 ;

#endif
{
	if ( (((int) pt1->x) == ((int) pt2->x)) &&
		 (((int) pt1->y) == ((int) pt2->y)) &&
		 (((int) pt1->z) == ((int) pt2->z)) )	return(0);
	else										return(1);
	}

/*		Another example of filtering function on 3D points				*/

#if defined(_ANSI_)
int T_3DFilter( t_3PointPtr pt1, t_3PointPtr pt2 )
#else
int T_3DFilter(pt1, pt2 )
t_3PointPtr pt1;
t_3PointPtr pt2 ;

#endif
{
  if ( (pt1->x == pt2->x) &&
      (pt1->y == pt2->y) &&
      (pt1->z == pt2->z) )	return(0);
  else						return(1);
}

/*
	Filtering function for 3D integer points
*/

#if defined(_ANSI_)
int T_3DIFilter( t_3IPointPtr pt1, t_3IPointPtr pt2 )
#else
int T_3DIFilter(pt1, pt2 )
t_3IPointPtr pt1;
t_3IPointPtr pt2 ;

#endif
{
  if ( (pt1->x == pt2->x) &&
      (pt1->y == pt2->y) &&
      (pt1->z == pt2->z) )	return(0);
  else	
					return(1);
}

/*
	Filtering function for EpidaureLib 3D points
*/

#if defined(_ANSI_)
int T_PointFilter( t_PointPtr pt1, t_PointPtr pt2 )
#else
int T_PointFilter(pt1, pt2 )
t_PointPtr pt1;
t_PointPtr pt2 ;

#endif
{
  if ( (pt1->pt.x == pt2->pt.x) &&
      (pt1->pt.y == pt2->pt.y) &&
      (pt1->pt.z == pt2->pt.z) )	return(0);
  else							return(1);
}

/*
	Code function for a pointer value
*/

#if defined(_ANSI_)
int T_PointerCode( char *thePtr )
#else
int T_PointerCode(thePtr )
char *thePtr ;
#endif

{ 
	unsigned long i;

	i= (unsigned long) thePtr;
	return( i );
	}

/*
	Filter function for a pointer value
*/

#if defined(_ANSI_)
int T_PointerFilter( char *ptr1, char *ptr2 )
#else
int T_PointerFilter(ptr1, ptr2 )
char *ptr1;
char *ptr2 ;

#endif
{
	if ( ptr1 == ptr2 ) return(0);
	else				return(1);
	}

/*		Add a new item into all the boxes covered with a given box		*/

#if defined(_ANSI_)
void T_AddEntryBox( char *theItem, t_2PointPtr ptMin, t_2PointPtr ptMax, t_HTablePtr theHTable )
#else
void T_AddEntryBox(theItem, ptMin, ptMax, theHTable )
char *theItem;
t_2PointPtr ptMin;
t_2PointPtr ptMax;
t_HTablePtr theHTable ;

#endif
{
  t_2Point pt;
  int x,y;
  int xmin,xmax,ymin,ymax;
  
  xmin= (int) ptMin->x; xmax= (int) ptMax->x; ymin= (int) ptMin->y; ymax= (int) ptMax->y;
  
  for( x= xmin; x<= xmax; x++ )
    for( y= ymin; y<= ymax; y++ ){
      pt.x= (reeltype) x;
      pt.y= (reeltype) y;
      T_AddEntry( theItem, (char *) &pt, theHTable );
    }
}

/*	Add a 3D box entry													*/

#if defined(_ANSI_)
void T_AddEntry3Box( char *theItem, t_3PointPtr pMin, t_3PointPtr pMax, t_HTablePtr theHTable )
#else
  void T_AddEntry3Box(theItem, pMin, pMax, theHTable )
  char *theItem;
t_3PointPtr pMin;
t_3PointPtr pMax;
t_HTablePtr theHTable ;
#endif
{
  t_3Point pt;
  int xmin,ymin,zmin,xmax,ymax,zmax;
  int x,y,z;
  
  xmin= (int) pMin->x; ymin= (int) pMin->y; zmin= (int) pMin->z;
  xmax= (int) pMax->x; ymax= (int) pMax->y; zmax= (int) pMax->z;
  
  for( x= xmin; x<= xmax; x++ )
    for( y= ymin; y<= ymax; y++ )
      for( z= zmin; z<= zmax; z++ ){
	
	pt.x= (reeltype) x;
	pt.y= (reeltype) y;
	pt.z= (reeltype) z;
	T_AddEntry( theItem, (char *) &pt, theHTable );
      }
}

/*----------------------------------------------------------------------*/

/* srand48 and drand48 are not implemented on windows */
/* Equivalent using standard functions */
/* srand48 -> srand */
/* drand48 -> rand()/(RAND_MAX+1.0) */


#include <stdlib.h>     /* for rand() and srand() declaration */
#include <time.h>       /* for time() declaration */

/* Initialize random generator  ~ srand48 */
#if defined(_ANSI_)
void RndInit(void) 
#else
void RndInit()
#endif
{srand((long int) time( NULL) );}

/* returns a random value between 0 and 1 ~ drand48 */
#if defined(_ANSI_)
double UniformRnd(void)
#else
double UniformRnd()
#endif
{return rand()/(RAND_MAX+1.0);}

/* Returns a random Gaussian value */
#if defined(_ANSI_)
double GaussianRnd(void)
#else
 double GaussianRnd()
#endif
{return ( sqrt(-2.0*log(UniformRnd()))*cos(2*PI*UniformRnd()));}

/*----------------------------------------------------------------------*/
