#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#include <imagedata/Image3DViewer.h>

namespace yav {

Image3DViewer::Image3DViewer(Scene3D *scene, Image3DBackgroundManager *_manager)  : Image3DViewerInterface(_manager),
Data3D(scene)
{
	init();
}

void Image3DViewer::init() 
{

  gridDisplay = FRAME;
  
  setBoxColor(1.0f, 1.0f, 1.0f);
  boxWidth = 3.0;
  setGridColor(1.0f, 0.0f, 0.0f);
  gridWidth = 1.0;

 // updateCenterAndRadius();
}

void Image3DViewer::setGridDisplay(GRID_DISPLAY gd) {
  if(gd != gridDisplay) {
    gridDisplay = gd;
    redrawScene();
  }
}



void Image3DViewer::updateCenter(Vec3<double>& center, unsigned int& n) {
  if(getImage()) {
	  Inrimage *inr=getImage();
    if(inr->getGeometry() == Inrimage::CARTESIAN) {
      center[0] = inr->getX() * inr->getVX() / 2.0;
      center[1] = inr->getY() * inr->getVY() / 2.0;
      center[2] = inr->getZ() * inr->getVZ() / 2.0;
    }
    else {
      center[0] = center[1] = inr->getX() * inr->getVX() / 2.0;
      center[2] = inr->getY() * inr->getVY() / 2.0;
    }
    center=inr->getTransformationMatrix()*center;
    n = 1;
  }
  else {
    center[0] = center[1] = center[2] = 0.0;
    n = 0;
  }
}

double Image3DViewer::updateRadius(const Vec3<double>& center) 
{
	if(getImage()) {
		Inrimage *inr=getImage();
		double r1=center.norm2();
		double r2=(Vec3<double>(inr->getX() * inr->getVX(),
			inr->getY() * inr->getVY(),
			inr->getZ() * inr->getVZ()) -center).norm2();
		if (r1>r2)
			return sqrt(r1);
		else
			return sqrt(r2);
	}
	else
		return 0.0;
}

double Image3DViewer::computeClosestDistance(DataTypes::MouseClick& mc) const {
  return maxTypeValue<double>(0);
}


void Image3DViewer::getBoxColor(float &r, float &g, float &b) const {
  r = boxColor[0];
  g = boxColor[1];
  b = boxColor[2];
}


void Image3DViewer::getGridColor(float &r, float &g, float &b) const {
  r = gridColor[0];
  g = gridColor[1];
  b = gridColor[2];
}


void Image3DViewer::setBoxColor(float r, float g, float b) {
  boxColor[0] = r < 0.0f ? 0.0f : (r > 1.0f ? 1.0f : r);
  boxColor[1] = g < 0.0f ? 0.0f : (g > 1.0f ? 1.0f : g);
  boxColor[2] = b < 0.0f ? 0.0f : (b > 1.0f ? 1.0f : b);
  redrawScene();
}


void Image3DViewer::setGridColor(float r, float g, float b) {
  gridColor[0] = r < 0.0f ? 0.0f : (r > 1.0f ? 1.0f : r);
  gridColor[1] = g < 0.0f ? 0.0f : (g > 1.0f ? 1.0f : g);
  gridColor[2] = b < 0.0f ? 0.0f : (b > 1.0f ? 1.0f : b);
  redrawScene();
}


void Image3DViewer::setBoxWidth(double width) {
  if(width >= 0.0) {
    boxWidth = width;
    redrawScene();
  }
}

void Image3DViewer::setGridWidth(double width) {
  if(width > 0.0) {
    gridWidth = width;
    redrawScene();
  }
}


void Image3DViewer::setShowBlackPixels(const bool b) {
  if (b!=showBlackPixels)
    {
      showBlackPixels=b;
      redrawScene();
    }
}
void Image3DViewer::redraw(Camera3D *) 
{

	Inrimage *image=getImage();
	if (image->useTransformationMatrix()) {
		Matrix4x4<double> mat=Matrix4x4<double>(image->getRotationMatrix(),
			image->getTranslation());
		glPushMatrix();
		mat.transpose();
		glMultMatrixd((const double *)mat);  
	}
	double x=0.0, y=0.0, z=0.0;

	// display bounding box
	glColor3f(boxColor[0], boxColor[1], boxColor[2]);
	glLineWidth((GLfloat)boxWidth);
	if(boxWidth > 0.0) {
		if(image->getGeometry() == Inrimage::CARTESIAN) {
			x = image->getX() * image->getVX();
			y = image->getY() * image->getVY();
			z = image->getZ() * image->getVZ();
		}
		else {
			x =y=image->getX() * image->getVX();
			z = image->getY() * image->getVY();
		}
		glDisable(GL_LIGHTING);
		glBegin(GL_LINE_LOOP);
		glVertex3f(0.0, 0.0, 0.0);
		glVertex3f((GLfloat)x, 0.0, 0.0);
		glVertex3f((GLfloat)x, (GLfloat)y, 0.0);
		glVertex3f(0.0, (GLfloat)y, 0.0);  
		glEnd();
		glBegin(GL_LINE_LOOP);
		glVertex3f(0.0, 0.0, (GLfloat)z);
		glVertex3f((GLfloat)x, 0.0, (GLfloat)z);
		glVertex3f((GLfloat)x, (GLfloat)y, (GLfloat)z);
		glVertex3f(0.0, (GLfloat)y, (GLfloat)z);  
		glEnd();
		glBegin(GL_LINES);
		glVertex3f(0.0, 0.0, 0.0);
		glVertex3f(0.0, 0.0, (GLfloat)z);
		glVertex3f((GLfloat)x, 0.0, 0.0);
		glVertex3f((GLfloat)x, 0.0, (GLfloat)z);
		glVertex3f(0.0, (GLfloat)y, 0.0);
		glVertex3f(0.0, (GLfloat)y,(GLfloat) z);
		glVertex3f((GLfloat)x,(GLfloat) y, 0.0);
		glVertex3f((GLfloat)x, (GLfloat)y, (GLfloat)z);
		glEnd();
		
	}
	glDisable(GL_LIGHTING);
	glColor3f(1.0f,1.0f,1.0f);
	/// returns an iterator on the first background
	ObjectsList<Image3DCameraBackground>::iterator itb=manager->firstBackground();
	for (;itb!=manager->lastBackground();itb++)
	{
		Image3DTextureInformation *mainTexture=(*itb)->getMainTexture();
		mainTexture->enableTexture(getImage(),
			manager->getRedColorMap(),
			manager->getGreenColorMap(),
			manager->getBlueColorMap(),
			manager->getMinWindow(),
			manager->getMaxWindow(),1.0);

		glColor4f(1,1,1,0.5);

		GLfloat glInterpolation;
		Image3DBackgroundManager::IMAGE_INTERPOLATION interpolation=manager->getInterpolationMethod();
		if (interpolation==Image3DBackgroundManager::NEAREST)
			glInterpolation=GL_NEAREST;
		else 
			glInterpolation=GL_LINEAR;

		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, glInterpolation);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, glInterpolation);
		mainTexture->drawTexture3D();	
		if ( ((*itb)->getSecondTexture()) && ((*itb)->getSecondTexture()->isInsideBackground())){
			Image3DTextureInformation *secondTexture=(*itb)->getSecondTexture();

			glBindTexture(GL_TEXTURE_2D, 0);
			glDisable(GL_LIGHTING);
			glEnable(GL_BLEND);
			glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);


			secondTexture->enableTexture(manager->getSecondImage(),
				manager->getRedColorMap2(),
				manager->getGreenColorMap2(),
				manager->getBlueColorMap2(),
				manager->getMinWindow2(),
				manager->getMaxWindow2(),1.0);

			glColor4f(1,1,1,(GLfloat)manager->getOpacity());
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, glInterpolation);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, glInterpolation);

			secondTexture->drawTexture3D();	
			glDisable(GL_BLEND);	

		}	
	}
	(*itb)->disableTexture();
	glEnable(GL_LIGHTING);

	if (image->useTransformationMatrix()) {
		glPopMatrix();
	}

}
void Image3DViewer::translate(const Vec3<double>& t) 
{
  getImage()->addTranslation(t);
}


void Image3DViewer::rotate(const Rotation3D &rotation) 
{
  getImage()->composeRotation(rotation);
}
void Image3DViewer::scale(const double scale) 
{
}
} // namespace yav
