#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "imagedata/RangeInrimage.h"
#include <imagedata/RangeInrimage.h>
#include <inrimage/ImageGeometry.h>
#include <basics/FileInfo.h>

namespace yav {

/**********************************************************************/
/**********************************************************************/
/**************************   RangeInrimage   *************************/
/**********************************************************************/
/**********************************************************************/

const TclString GradientBalloonInrimageInformation::informationName="GradientBalloon";
const TclString GradientGradientInrimageInformation::informationName="GradientGradient";
const TclString IsosurfaceInrimageInformation::informationName="Isosurface";
const TclString GradientInrimageInformation::informationName="Gradient";
const TclString RegionInrimageInformation::informationName="Region";
const TclString RegionProbabilityMapInformation::informationName="ProbabilityMap";
const TclString ClassInrimageInformation::informationName="Classify";
const TclString IntensityProfileInformation::informationName="Intensity Profile";
const TclString BlockMatchingInformation::informationName="Block Matching";

static std::map<TclString , Inrimage *> bmRefImages;
  

#define GRADIENT_SECOND_DERIVATIVE_RATIO 1.5

RangeInrimage::InvalidSize::InvalidSize( const std::string &from) :
  Exception(from,"image with wrong size ") { }

/**********************************************************************/
/**********************************************************************/
/*********************   RangeInrimage::Force   ***********************/
/**********************************************************************/
/**********************************************************************/


RangeInrimage::Force::InvalidRangeInrimage::InvalidRangeInrimage( const std::string &from)
  : Exception(from,"force with a null RangeInrimage pointer") { }

RangeInrimage::Force::InvalidImages::InvalidImages( const std::string &from)
  : Exception(from,"unvalid intensity, gradient images") { }

RangeInrimage::Force::Force(RangeInrimage *_range,TclString _name)
  throw (InvalidRangeInrimage) : name(_name) ,range(_range) {
  if (!range)
    throw(InvalidRangeInrimage("RangeInrimage::Force::Force"));
}
void RangeInrimage::initFormat() {  
  // Sets the necessary module to read and write a file
  GeneralFormat::addModuleVersion("Main", true, 1.0,
				  (GeneralFormatModuleVersion::INPUTREADFUNCTION) &RangeInrimage::readMain,
				  (GeneralFormatModuleVersion::INPUTSKIPFUNCTION) &RangeInrimage::skipMain,
				  (GeneralFormatModuleVersion::OUTPUTWRITEFUNCTION) &RangeInrimage::writeMain);
}
void RangeInrimage::skipMain(GFInput &input) {
  unsigned int i,n;
  input >> n;
  TclString name,filename;
  for(i=0;i<n;++i) {
    input >> name >> filename;
  }
}
void RangeInrimage::writeMain(GFOutput &output) {
  unsigned int n;
  n=0;
  if (intensity) n++;
  if (gradient) n++;
  if (gradientGradient) n++;
  if (gradientDirection) n++;
  if (classify) n++;
  output << n << ENDL;
  if (intensity) {
    output << "intensity" <<  ENDL << intensity->getName() << ENDL;
  }
  if (gradient) {
    output << "gradient" << ENDL << gradient->getName() << ENDL;
  }
  if (gradientDirection) {
    output << "gradientDirection" << ENDL << gradientDirection->getName() << ENDL;
  }
  if (gradientGradient) {
    output << "gradientGradient" << ENDL << gradientGradient->getName() << ENDL;
  }
}
void RangeInrimage::readMain(GFInput &input) {
  unsigned int i,n;
  input >> n;
  TclString name,filename;
  for(i=0;i<n;++i) {
    input >> name >> filename;
    if (name=="intensity") {
      intensity=new Inrimage(filename);
      setSize(intensity);
    }
    else if (name=="gradient") {
      setGradient(new Inrimage(filename));
    }
    else if (name=="gradientDirection") {
      setGradientDirection(new Inrimage(filename));
    }
    else if (name=="gradientGradient") {
      setGradientGradient(new Inrimage(filename));
    }
    else if (name=="classify") {
      setClassify(new Inrimage(filename));
    }
  }
}

RangeInrimage::RangeInrimage(const TclString& fileName,
			     Inrimage *_intensity)
  throw(Inrimage::ImageNotFound, Inrimage::UnknownImageType,
	Inrimage::InvalidHeader, Inrimage::InvalidData,
	Inrimage::UnknownType)
  :  Owner<Inrimage>(0), GeneralFormat("RangeImage File","RANGEIMAGE"),intensity(this, _intensity), gradient(this, 0), 
     gradientDirection(this, 0), edge(this, 0),
     gradientGradient(this, 0), classify(this, 0),
     sizeX(0),sizeY(0),sizeZ(0),
     forceGeometry(FORCE_3D)
{
  TclString gradientFileName, gradientDirectionFileName,
    gradientGradientFileName, classifyFileName;

  initFormat();
  

  if (imageType((char *)fileName)==0) {
    GeneralFormat::read((char *)fileName,0);
  }  else {
    if(!intensity) {
      intensity = new Inrimage(fileName);
    }
    else setIntensity(_intensity);
    
    setSize(intensity);

    forceList += new RangeInrimage::IntensityProfileForce(this);
    
    gradientFile(fileName, gradientFileName);
    if(gradientFileName != EMPTY_STRING) {
      try {
	setGradient(new Inrimage(gradientFileName));
      }
      catch(InvalidSize&) {
	CERR << "Invalid gradient file size \'" << gradientFileName
	     << "\'. No gradient information available." << ENDL;
      }
    }
    gradientDirectionFile(fileName, gradientDirectionFileName);
    if(gradientDirectionFileName != EMPTY_STRING) {
      try {
	setGradientDirection(new Inrimage(gradientDirectionFileName));
      }
      catch(InvalidSize&) {
	CERR << "Invalid gradient direction file size \'" << gradientFileName
	     << "\'. No gradient direction information available." << ENDL;
      }
    }
    gradientGradientFile(fileName, gradientGradientFileName);
    if(gradientGradientFileName != EMPTY_STRING) {
      try {
	setGradientGradient(new Inrimage(gradientGradientFileName));
      }
      catch(InvalidSize&) {
	CERR << "Invalid gradient of the gradient file size \'" << gradientFileName
	     << "\'. No gradient of the gradient information available." << ENDL;
      }
    }
    classifyFile(fileName, classifyFileName);
    if(classifyFileName != EMPTY_STRING) {
      try {
	setClassify(new Inrimage(classifyFileName));
      }
      catch(InvalidSize&) {
	CERR << "Invalid classify file size \'" << classifyFileName
	     << "\'. No classify information available." << ENDL;
      }
    }
  }
  initForces();
  sampledDirection=Inrimage::createDirectionArray();
}

RangeInrimage::RangeInrimage(Inrimage *_intensity,
			     Inrimage *_gradient,
			     Inrimage *_gradientDirection,
			     Inrimage *_edge,
			     Inrimage *_gradientGradient,
			     Inrimage *_classify,
			     RangeInrimage::ForceGeometry _forceGeometry) :
  Owner<Inrimage>(0),GeneralFormat("RangeImage File","RANGEIMAGE"),intensity(this,_intensity), gradient(this,_gradient), 
  gradientDirection(this,_gradientDirection), edge(this,_edge),
  gradientGradient(this,_gradientGradient), classify(this, _classify),
  forceGeometry(_forceGeometry)
{
  /// sets the size of all images
  if (intensity()) setSize(intensity);
  else if (gradient()) setSize(gradient);
  else if (gradientDirection()) setSize(gradientDirection);
  else if (edge()) setSize(edge);
  else if (gradientGradient()) setSize(gradientGradient);
  else if (classify()) setSize(classify);
  else { sizeX = sizeY = sizeZ = 0;}

  if (intensity()) testSize(intensity);
  if (gradient()) testSize(gradient);
  if (gradientDirection()) testSize(gradientDirection);
  if (edge()) testSize(edge);
  if (gradientGradient()) testSize(gradientGradient);
  if (classify()) testSize(classify);
  
  initForces();
  sampledDirection=Inrimage::createDirectionArray();
}

RangeInrimage::~RangeInrimage() {
  delete [] sampledDirection;
}

void RangeInrimage::initForces() {
  intensityProfileForce= new RangeInrimage::IntensityProfileForce(this);
  blockMatchingForce = new RangeInrimage::BlockMatchingForce(this);
  regionForce= new RangeInrimage::RegionForce(this); 
  isosurfaceForce= new RangeInrimage::IsosurfaceForce(this); 

 // probabilityMapForce= new RangeInrimage::RegionProbabilityMapForce(this);
  if(gradient) {
    gradientForce= new RangeInrimage::GradientForce(this);
    gradientBalloonForce= new RangeInrimage::GradientBalloonForce(this); 
  }
  else {
    gradientForce=0;
    gradientBalloonForce=0;
  }
  if(gradientGradient)
    gradientGradientForce= new RangeInrimage::GradientGradientForce(this);
  else
    gradientGradientForce=0;
  if(classify) {
    classForce= new RangeInrimage::ClassForce(this);
  }
  else
    classForce=0;
}

void RangeInrimage::setSize(Inrimage *im) {
  if (im) {
    sizeX=im->getX();
    sizeY=im->getY();
    sizeZ=im->getZ();
  }
}

void RangeInrimage::testSize(Inrimage *im) const throw (RangeInrimage::InvalidSize) {
  if ((im) && (sizeX) && (sizeY) && (sizeZ)) {
    if ((im->getX()!=sizeX) || (im->getY()!=sizeY) || (im->getZ()!=sizeZ))
      throw(RangeInrimage::InvalidSize("RangeInrimage::testSize()"));
  }
}

void RangeInrimage::gradientFile(const TclString& fileName, 
				 TclString& gradientName) const {
  FileInfo file(fileName);
  file = FileInfo(file.noextension(".gz"));
  file = FileInfo(file.noextension(".inr"));
  file = FileInfo(file.getName() + ".gradient");
  if(file.exists()) gradientName = file.getName();
  else {
    file = FileInfo(file.getName() + ".gz");
    if(file.exists()) gradientName = file.getName();
    else gradientName = EMPTY_STRING;
  }
}


void RangeInrimage::gradientDirectionFile(const TclString& fileName,
					  TclString& gradientDirectionName) const {
  FileInfo file(fileName);
  file = FileInfo(file.noextension(".gz"));
  file = FileInfo(file.noextension(".inr"));
  file = FileInfo(file.getName() + ".gradient_direction");
  if(file.exists()) gradientDirectionName = file.getName();
  else {
    file = FileInfo(file.getName() + ".gz");
    if(file.exists()) gradientDirectionName = file.getName();
    else gradientDirectionName = EMPTY_STRING;
  }
}
void RangeInrimage::gradientGradientFile(const TclString& fileName,
					 TclString& gradientGradientName) const {
  FileInfo file(fileName);
  file = FileInfo(file.noextension(".gz"));
  file = FileInfo(file.noextension(".inr"));
  file = FileInfo(file.getName() + ".gradientGradient");
  if(file.exists()) gradientGradientName = file.getName();
  else {
    file = FileInfo(file.getName() + ".gz");
    if(file.exists()) gradientGradientName = file.getName();
    else gradientGradientName = EMPTY_STRING;
  }
}

void RangeInrimage::classifyFile(const TclString& fileName, 
				 TclString& classifyName) const {
  FileInfo file(fileName);
  file = FileInfo(file.noextension(".gz"));
  file = FileInfo(file.noextension(".inr"));
  file = FileInfo(file.getName() + ".classify");
  if(file.exists()) classifyName = file.getName();
  else {
    file = FileInfo(file.getName() + ".gz");
    if(file.exists()) classifyName = file.getName();
    else classifyName = EMPTY_STRING;
  }
}

void RangeInrimage::updateForceImages() {
  for(ObjectsList<Force>::Iterator it(forceList); !it.isAtEnd(); it++)
    it()->updateImage();
}


void RangeInrimage::setGradient(Inrimage *inr) throw(InvalidSize) {
  if ((!sizeX) && (!sizeY) && (!sizeZ))
    setSize(inr);
  else
    testSize(inr);
  if(!gradient) 
    {
      forceList += new RangeInrimage::GradientForce(this);
      forceList += new RangeInrimage::GradientBalloonForce(this);
    }
  gradient = inr;
  updateForceImages();
}

void RangeInrimage::setGradientDirection(Inrimage *inr) throw(InvalidSize) {
  if ((!sizeX) && (!sizeY) && (!sizeZ))
    setSize(inr);
  else
    testSize(inr);
  gradientDirection = inr;
  updateForceImages();
}

void RangeInrimage::setIntensity(Inrimage *inr) throw(InvalidSize) {
  if ((!sizeX) && (!sizeY) && (!sizeZ))
    setSize(inr);
  else
    testSize(inr);
  intensity = inr;
  updateForceImages();
}

void RangeInrimage::setEdge(Inrimage *inr) throw(InvalidSize) {
  if ((!sizeX) && (!sizeY) && (!sizeZ))
    setSize(inr);
  else
    testSize(inr);
  edge = inr;
  updateForceImages();
}

void RangeInrimage::setGradientGradient(Inrimage *inr) throw(InvalidSize) {
  if ((!sizeX) && (!sizeY) && (!sizeZ))
    setSize(inr);
  else
    testSize(inr);
  if(!gradientGradient) {
	  gradientGradientForce= new RangeInrimage::GradientGradientForce(this);
    forceList += gradientGradientForce;
  }
  gradientGradient = inr;
  updateForceImages();
}

void RangeInrimage::setClassify(Inrimage *inr) throw(InvalidSize) {
  if ((!sizeX) && (!sizeY) && (!sizeZ))
    setSize(inr);
  else
    testSize(inr);
  if(!classify) {
    classForce= new RangeInrimage::ClassForce(this);
    forceList += classForce;
  }
  classify = inr;
  updateForceImages();
}


void RangeInrimage::computeSobelGradient(unsigned int sigma,bool twoD,
					 bool maximum,double lt)
{
  Inrimage *gradientNorm,*gradientDir;
  if (intensity())
    {
      yav::sobel(intensity,gradientNorm,gradientDir,sigma,yav::GD_SAMPLE,
		 maximum,twoD,lt);
      setGradient(gradientNorm);
      setGradientDirection(gradientDir);
    }
}
void RangeInrimage::computeRecursiveGradient(yav::RECURSIVE_FILTER_TYPE filter,double sigma,bool twoD,bool maximum,double lt)
{
  Inrimage *gradientNorm,*gradientDir;
  if (intensity)
    {
      yav::gradient(intensity,gradientNorm,gradientDir,sigma,filter,
		    yav::GD_SAMPLE,maximum,twoD,lt);
      setGradient(gradientNorm);
      setGradientDirection(gradientDir);
    }
}
void RangeInrimage::computeSobelGradientGradient(unsigned int sigma,bool twoD)
{
  Inrimage *gradientGradientNorm,*gradientGradientDir;
  if (gradient)
    {
      yav::sobel(gradient,gradientGradientNorm,gradientGradientDir,sigma,yav::GD_VECTORS,false,0.0);
      delete gradientGradientNorm;
      setGradientGradient(gradientGradientDir);
    }
}
void RangeInrimage::computeRecursiveGradientGradient(yav::RECURSIVE_FILTER_TYPE filter,double sigma,bool twoD)
{
  Inrimage *gradientGradientNorm,*gradientGradientDir;
  if (gradient)
    {
      yav::gradient(gradient,gradientGradientNorm,gradientGradientDir,sigma,filter,
		    yav::GD_SAMPLE,false,twoD,0.0);
      delete gradientGradientNorm;
      setGradientGradient(gradientGradientDir);
    }
}
void RangeInrimage::dying(const ObjectPointer<Inrimage>& image) 
{
  if (image()==intensity())  
    intensity=0;
  else if (image()==gradient())  
    gradient=0;
  else if (image()==gradientGradient())  
    gradientGradient=0;
  else if (image()==gradientDirection())  
    gradientDirection=0;
  else if (image()==edge())  
    edge=0;

}




/**********************************************************************/
/**********************************************************************/
/*****************   RangeInrimage::Gradient Force   ******************/
/**********************************************************************/
/**********************************************************************/


RangeInrimage::GradientForce::GradientForce(RangeInrimage *_range) 
  throw(RangeInrimage::Force::InvalidRangeInrimage)
  :  VoxelMax(0.0, (bool (VoxelMax::*)(ImageVoxel *, Voxel&))
	     &RangeInrimage::GradientForce::stopFromGradient), 
     Force(_range,"gradient")
{
  inr=range->getGradient();
  maxVoxel=0;
  if (inr)
    converter=inr->newConverter();
  else
    converter=0;
  if (inr)
	  vsize=(inr->getVX()+inr->getVY()+inr->getVZ())/3.0;
}

void RangeInrimage::GradientForce::start(Voxel& voxel) {
  if(testVoxel(startvoxel)) maxVoxel = startvoxel;
  else maxVoxel = 0;
}


RangeInrimage::GradientForce::~GradientForce() {
  if (converter)
    delete converter;
}


void RangeInrimage::GradientForce::updateImage(void) {
  inr = range->getGradient();
  if (inr)
	  vsize=(inr->getVX()+inr->getVY()+inr->getVZ())/3.0;
  if (converter)
    delete converter;
  if (inr)
    converter=inr->newConverter();
  else
    converter=0;
}

													 

bool RangeInrimage::GradientForce::testVoxel(ImageVoxel *voxel) {
	bool test = false; 
	if(voxel->getValue() > gi->getThreshold()) {

		if(gi->getOrientationConstraint() == GradientInrimageInformation::NONE) 
			test = true;
		else if (range->getGradientDirection()){
			Vec3<double> gdir; 
			double val;
			unsigned int dir;
			Voxel *gdvoxel =
				Voxel::voxelByType(range->getGradientDirection()->getType(),range->getGradientDirection()->getVdim());
			range->getGradientDirection()->getVoxel(voxel->getX(), voxel->getY(), 
				voxel->getZ(), *gdvoxel);

			if (range->getGradientDirection()->getVdim()==1)
			{
				dir = (unsigned int) gdvoxel->getValueAsLong();
				delete gdvoxel;
				if(dir >= range->getDirectionArraySize()) test = false;
				else 
					gdir = range->getSampledDirection()[dir];
			}
			else if (gdvoxel->getType()==Inrimage::WT_FLOAT_VECTOR) {
				VoxelFV *vf=(VoxelFV *) gdvoxel;
				gdir= Vec3<double> ((double)(vf->getValue()[0U]),
					(double)(vf->getValue()[1U]),
					(double)(vf->getValue()[2U]));
				delete gdvoxel;
			}
			val = dot<double>(gdir, direction);
			if((val <= 0.5 && gi->getOrientationConstraint() == GradientInrimageInformation::ALONG_GRADIENT) || 
				(val >= -0.5 && gi->getOrientationConstraint() == GradientInrimageInformation::ALONG_OPPOSITE_GRADIENT))
				test = false;
			else
				test = true;
		}

		if(gi->getIntensityConstraint() && test) {
			double val;
			Voxel *ivoxel =
				Voxel::voxelByType(range->getIntensity()->getType());
			range->getIntensity()->getVoxel(voxel->getX(), voxel->getY(), 
				voxel->getZ(), *ivoxel);
			val = ivoxel->getValueAsDouble();
			delete ivoxel;	
			if((val>gi->getMaximumIntensity()) || (val<gi->getMinimumIntensity())) test = false;
		}
	}
	return test;
}


bool RangeInrimage::GradientForce::stopFromGradient(ImageVoxel *voxel,
						    Voxel&)  {

  // already found a closest value along the other direction
  if(gi->getStopCondition() && maxVoxel &&
     (voxel->getDistance() > maxVoxel->getDistance())) return true;

  // is this voxel better than max ?
  if(testVoxel(voxel) &&
     (!maxVoxel || (voxel->getValue() > maxVoxel->getValue()) || ((voxel->getValue() == maxVoxel->getValue()) && (voxel->getDistance()<  maxVoxel->getDistance() )))) {
    maxVoxel = voxel;
    if(gi->getStopCondition()) return true;
  }

  // have we reached the scan-line end ?
  if(voxel->getDistance() > maxDistance) return true;

  // we passed all tests : continue scanning
  return false;
}

bool RangeInrimage::GradientForce::isPossible() const {
  if(range->getGradient() && range->getGradientDirection())
    return true;
  return false;
}

bool RangeInrimage::GradientForce::getPoint(const Vec3<double> &position, 
					    const Vec3<double> &direction,
					    double r,
					    Vec3<double> &point,double &confidence,
					    GradientInrimageInformation *rii) 
{
  if (rii==0)
    return false;
  gi=rii;
  maxDistance=rii->getMaxDistance();
  maxVoxel = 0;
  confidence=0;
  compute(r, position, direction);
  if(!startvoxel || !maxVoxel) return false;
  Vec3<unsigned int> v;
  v = maxVoxel->getVoxelPosition();
  converter->setImage((double)(v[0]+.5),(double)(v[1]+.5),
		      (double)(v[2]+.5));
  double p[3];
  converter->toCartesian(p);

  point=Vec3<double>(p[0],p[1],p[2]);

  if (rii->getStopCondition()) {
	  /// if first voxel above threshold then set the confidence as amount above threshold
	  confidence=maxVoxel->getValue()-rii->getThreshold();
  } else {
	  /// if max gradient then set the confidence as the second derivative of gradient at that point 
	  confidence=maxVoxel->getValue()-rii->getThreshold();
	  Vec3<double> nextPoint=point+GRADIENT_SECOND_DERIVATIVE_RATIO*vsize*direction;
	  Voxel *gvoxel=inr->getRealVoxel(nextPoint);
	  double nextGradient=gvoxel->getValueAsDouble();

	  Vec3<double> previousPoint=point-GRADIENT_SECOND_DERIVATIVE_RATIO*vsize*direction;
	  inr->getRealVoxel(previousPoint.getX(), previousPoint.getY(), 
		  previousPoint.getZ(), *gvoxel);
	  double previousGradient=gvoxel->getValueAsDouble();
	  confidence= - (previousGradient+nextGradient - 2*maxVoxel->getValue());
	  confidence /= vsize*vsize*GRADIENT_SECOND_DERIVATIVE_RATIO*GRADIENT_SECOND_DERIVATIVE_RATIO;
	  if (confidence<0.0)
		  confidence=0;
  }
  return true;
}



/**********************************************************************/
/**********************************************************************/
/*****************   RangeInrimage::Isosurface Force   ******************/
/**********************************************************************/
/**********************************************************************/


RangeInrimage::IsosurfaceForce::IsosurfaceForce(RangeInrimage *_range) 
  throw(RangeInrimage::Force::InvalidRangeInrimage)
  :  VoxelMax(0.0, (bool (VoxelMax::*)(ImageVoxel *, Voxel&))
	     &RangeInrimage::IsosurfaceForce::stopFromDistance), 
     Force(_range,"Isosurface")
{
  inr=range->getIntensity();
  sigma=2*(inr->getVX()+inr->getVY()+inr->getVZ())/3;
  maxVoxel=0;
}

void RangeInrimage::IsosurfaceForce::start(Voxel& voxel) {
  maxVoxel = 0;
}


RangeInrimage::IsosurfaceForce::~IsosurfaceForce() {
}


void RangeInrimage::IsosurfaceForce::updateImage(void) {
  inr = range->getIntensity();
}


bool RangeInrimage::IsosurfaceForce::findClosestIsosurfaceVoxel(double &distance) {
  std::vector<ImageVoxel *>::iterator it;
  double previousVal,currentVal,threshold,l,previousDistance=0.0;
  bool found=false,foundp,foundm;

  threshold=isi->getIsovalue();

  distance=maxDistance;
  it=positive.begin();
  previousVal=(*it)->getValue();
  if (previousVal==threshold) {
    foundp=true;
    distance=(*it)->getDistance();
  } else 
    foundp=false;

  for(it++;(foundp==false) && (it!=positive.end());++it) {
    currentVal=(*it)->getValue();

    if (currentVal==threshold) {
      distance=(*it)->getDistance();
      foundp=true;
    } else if  (((currentVal-threshold)*(previousVal-threshold))<=0.0) {
      l=(threshold-previousVal)/(currentVal-previousVal);
      distance=previousDistance+l*((*it)->getDistance()-previousDistance);
      foundp=true;
    } else {
      previousVal=currentVal;
      previousDistance=(*it)->getDistance();
    }
  }

  if (foundp)
    found=true;

  it=positive.begin();
  previousVal=(*it)->getValue();
  previousDistance=(*it)->getDistance();
  foundm=false;


  for(it=negative.begin();(foundm==false) && (it!=negative.end()) && (fabs((*it)->getDistance())<distance);++it) {
    currentVal=(*it)->getValue();

    if (currentVal==threshold) {
      distance= -(*it)->getDistance();
      foundm=true;
    } else if  (((currentVal-threshold)*(previousVal-threshold))<=0.0) {
      l=(threshold-previousVal)/(currentVal-previousVal);
      distance=previousDistance+l*(-(*it)->getDistance()-previousDistance);
      foundm=true;
    } else {
      previousVal=currentVal;
      previousDistance= -(*it)->getDistance();
    }
  }

  if (foundm)
    found=true;
  return found;
}


bool RangeInrimage::IsosurfaceForce::stopFromDistance(ImageVoxel *voxel,
						      Voxel&)  {
  return (voxel->getDistance() > maxDistance) ? true : false;
}

bool RangeInrimage::IsosurfaceForce::isPossible() const {
  if(range->getIntensity())
    return true;
  return false;
}

bool RangeInrimage::IsosurfaceForce::findRegionIsosurfaceVoxel(double &distance,IsosurfaceInrimageInformation::IsosurfaceConstraint constraint) {
  bool insideRegion=false;
  bool finish=false;
  double voxelValue,voxelDistance,l;
  double previousValue,previousDistance;
  int i;
  bool inside=true;
  if ((constraint==IsosurfaceInrimageInformation::OUTSIDE_OVER_ISOVALUE) ||
	  (constraint==IsosurfaceInrimageInformation::OUTSIDE_BELOW_ISOVALUE))
	  inside=false;
  double transitionSign=1;
  if ((constraint==IsosurfaceInrimageInformation::INSIDE_BELOW_ISOVALUE) ||
	  (constraint==IsosurfaceInrimageInformation::OUTSIDE_BELOW_ISOVALUE))
	  transitionSign= -1;


 if ((positive.size()+negative.size())>4) {
	 int minIndex= -(int)negative.size()-1;
	 int maxIndex= positive.size();
	 if (maxIndex== 0)
		 maxIndex= -1;

	 if (inside) {
		 i=minIndex+1;
	 }
	 else {
		 i=maxIndex-1;
	 }
	 double isoValue=isi->getIsovalue();
	 bool test;

	 do {
		 voxelValue=getVoxel(i)->getValue();
		 voxelDistance=getDistance(i);

		 if(insideRegion) {
			 /// test if we are on the other side of the isoValue
			 if((voxelValue-isoValue)*transitionSign<0.0) {
				 // we found the isosurface intersection
				 l=(isoValue-previousValue)/(voxelValue-previousValue);
				 distance=previousDistance+l*(voxelDistance-previousDistance);
				 finish=true;
			 }
			 else {
				 /// still on the same side of the isosurface
				 previousValue=voxelValue;
				 previousDistance=voxelDistance;
			 }
		 }
		 else {
			 /// test if we are in the right region (depending on the transition sign)
			 if ((voxelValue-isoValue)*transitionSign>=0) {
				 insideRegion=true;	
				 previousValue=voxelValue;
				 previousDistance=voxelDistance;
			 }
		 }
		 

		 if (inside) {
			 ++i;
			 test=(i<maxIndex);
		 } else {
			 --i;
			 test=(i>minIndex);
		 }
	 } while ((finish==false) && test);

	 return finish;

 } else
	 return false;
}
	

bool RangeInrimage::IsosurfaceForce::getPoint(const Vec3<double> &position, 
					      const Vec3<double> &direction,
					      double r,
					      Vec3<double> &point,double &confidence,
					      IsosurfaceInrimageInformation *rii) 
{
  if (rii==0)
    return false;
  isi=rii;
  maxDistance=rii->getMaxDistance();
  confidence=0;
  maxVoxel = 0;
  compute(r, position, direction);

  double d;
  if (rii->getIsosurfaceConstraint()==IsosurfaceInrimageInformation::CLOSEST) {
	  if (findClosestIsosurfaceVoxel(d)) {
		  point=position+d*direction;
		  /// compute the confidence as the square gradient
		  confidence=computeGradientAtDistance(d,sigma);
		  confidence*=confidence;
		  return true;
	  } else 
		  return false;
  } else {
	  if (findRegionIsosurfaceVoxel(d,rii->getIsosurfaceConstraint())) {
		  point=position+d*direction;
  		  /// compute the confidence as the square gradient
		  confidence=computeGradientAtDistance(d,sigma);
		  confidence*=confidence;
		  return true;
	  } else 
		  return false;
  }
}


/**********************************************************************/
/**********************************************************************/
/**************   RangeInrimage::GradientGradientForce   **************/
/**********************************************************************/
/**********************************************************************/

RangeInrimage::GradientGradientForce::GradientGradientForce(RangeInrimage *_range)
  throw(RangeInrimage::Force::InvalidRangeInrimage)  : Force(_range, "gradient2")
{
}

bool RangeInrimage::GradientGradientForce::isPossible() const {
  if (range->getGradientGradient())
    return(true);
  else
    return(false);
}
bool RangeInrimage::GradientGradientForce::getPoint(const Vec3<double> &position,
						    const Vec3<double> &direction,
						    double _range,Vec3<double> &point,double &confidence,
						    GradientGradientInrimageInformation *rii)
{
  if (rii==0)
    return false;
  int pos[3];
  Inrimage *inr = range->getGradientGradient();

  ImageGeometry *converter = inr->newConverter();
  converter->set(position.getX(), position.getY(), position.getZ());
  if(!converter->in()) {
    delete converter;
    return false;
  }
  converter->toImage(pos);
  delete converter;

  Voxel *voxel = inr->getVoxel(pos[0], pos[1], pos[2]);
  switch(inr->getType()) {

  case Inrimage::WT_UNSIGNED_CHAR:
  case Inrimage::WT_UNSIGNED_SHORT:
    {
      Vec3<double> gradientVector = 
	range->getSampledDirection()[(unsigned int) voxel->getValueAsDouble()];
      point = position+dot(direction,gradientVector)*direction;
      return(true);
    }
    break;

  case Inrimage::WT_FLOAT_VECTOR:
    {
      VoxelFV *gradientVoxel = (VoxelFV *) (voxel);
      if (gradientVoxel->getVdim() == 3) {
	Vec3<double> gradientVector ((*gradientVoxel)[0], (*gradientVoxel)[1],
				     (*gradientVoxel)[2]);
	point = position + dot<double>(direction,gradientVector)*direction;
	return(true);
      }
      else if (gradientVoxel->getVdim() == 2) {
	Vec3<double> gradientVector ((*gradientVoxel)[0], (*gradientVoxel)[1],
				     0.0);
	point = position + dot<double>(direction,gradientVector)*direction;
	return(true);
      }
      else 
	return(false);
      delete voxel;
      break;
    }

  default:
    return false;
  }
  return false;
}


/**********************************************************************/
/**********************************************************************/
/***************   RangeInrimage::GradientBalloonForce   **************/
/**********************************************************************/
/**********************************************************************/

RangeInrimage::GradientBalloonForce::GradientBalloonForce(RangeInrimage *_range)
  throw(RangeInrimage::Force::InvalidRangeInrimage)
  : Force(_range, "gradientBalloon")
{
}

bool RangeInrimage::GradientBalloonForce::isPossible() const {
  if (range->getGradient()) 
    return(true);
  else
    return(false);
}

bool RangeInrimage::GradientBalloonForce::getPoint(const Vec3<double> &position,
						   const Vec3<double> &direction,
						   double _range,Vec3<double> &point,double &confidence,
						   GradientBalloonInrimageInformation *rii)
{
  if (rii==0)
    return false;
  Inrimage *inr = range->getGradient();
  int pos[3];

  ImageGeometry *converter = inr->newConverter();
  converter->set(position.getX(), position.getY(), position.getZ());
  if(!converter->in()) {
    delete converter;
    return false;
  }
  converter->toImage(pos);
  delete converter;

  Voxel *voxel = inr->getVoxel(pos[0],pos[1],pos[2]);
  double gradientValue = voxel->getValueAsDouble();
  delete voxel;
  if(gradientValue>=rii->getThreshold()) point = position;
  else {
	  switch(rii->getOrientationConstraint()) {
	case GradientBalloonInrimageInformation::NONE:
	case GradientBalloonInrimageInformation::ALONG_GRADIENT:
		point = position + direction * (rii->getBalloonForce() *
			(1.0 - gradientValue/rii->getThreshold()));
		break;
	case GradientBalloonInrimageInformation::ALONG_OPPOSITE_GRADIENT:
		point = position - direction * (rii->getBalloonForce() *
			(1.0 - gradientValue/rii->getThreshold()));
		break;
	  }
  }
  return true;
}


void RangeInrimage::GradientBalloonForce::getSpeed(
                                          const Vec3<double> &pos,
					  double &mult, double &add,GradientBalloonInrimageInformation *rii) {
  if (rii==0)
    {
      mult=1.0;
    }
  else
    {
      add = 0.0;
      Inrimage *inr = range->getGradient();
      if(pos.getX() > 0 && pos.getY() > 0 && pos.getZ() > 0 &&
	 pos.getX() < inr->getX() && pos.getY() < inr->getY() &&
	 pos.getZ() < inr->getZ()) {
	Voxel *voxel = inr->getVoxel((unsigned int)pos.getX(),(unsigned int)pos.getY(),(unsigned int)pos.getZ());
	double gradientValue = voxel->getValueAsDouble();
	delete voxel;
	if(gradientValue > rii->getThreshold()) mult = 0.0;
	else mult = 1.0 - (gradientValue / rii->getThreshold());
      }
      else mult = 1.0;
    }
}

/**********************************************************************/
/**********************************************************************/
/******************   RangeInrimage::RegionForce   ********************/
/**********************************************************************/
/**********************************************************************/

RangeInrimage::RegionForce::RegionForce(RangeInrimage *_range)
  throw(RangeInrimage::Force::InvalidRangeInrimage)
  :  VoxelMax(0.0, (bool (VoxelMax::*)(ImageVoxel *, Voxel&))
	    &RangeInrimage::RegionForce::stopFromDistance),
     Force(_range,"region")
{
  inr=range->getIntensity();
  maxVoxel=0;
  if (inr)
	  vsize=(inr->getVX()+inr->getVX()+inr->getVX())/3.0;


}

void RangeInrimage::RegionForce::start(Voxel& voxel) {
 maxVoxel = 0;
}


RangeInrimage::RegionForce::~RegionForce() {}

void RangeInrimage::RegionForce::updateImage(void) {
  inr = range->getIntensity();
  vsize=(inr->getVX()+inr->getVX()+inr->getVX())/3.0;

}

void RangeInrimage::RegionForce::smoothProfile() 
{
  std::vector<double> array(negative.size()+positive.size()+1);

  int minIndex= 0-(negative.size());
  int maxIndex= positive.size();
  int i;

  double variance=1/(ri->getMinimumLength()*ri->getMinimumLength());

  for(i=minIndex+1;i<maxIndex;++i)
    {
      int j=i;
      double value=getVoxel(i)->getValue();
      double totalWeight=1.0;
      double weight;
      double distance;
	  double referenceDistance=getVoxel(i)->getDistance();
	  do {
		  j--;
		  if (j>minIndex)
		  {
			  distance=fabs(getVoxel(j)->getDistance()-referenceDistance);
			  weight= exp(-(distance*distance)*variance);
			  value+=weight*getVoxel(j)->getValue();
			  totalWeight+=weight;
		  }
		  else
			  distance=2*ri->getMinimumLength();
	  } while (distance<2*ri->getMinimumLength());

	  j=i;
	  do {
		  j++;
		  if (j<maxIndex)
		  {
			  distance=fabs(getVoxel(j)->getDistance()-referenceDistance);
			  weight= exp(-(distance*distance)*variance);
			  value+=weight*getVoxel(j)->getValue();
			  totalWeight+=weight;
		  }
		  else
			  distance=2*ri->getMinimumLength();
	  } while (distance<2*ri->getMinimumLength());
	  array[i-minIndex+1]=value/totalWeight;
	}
	for(i=minIndex+1;i<maxIndex;++i)
	{
		getVoxel(i)->setValue(array[i-minIndex+1]);
	}
	array.erase(array.begin(),array.end());
}

double RangeInrimage::RegionForce::findRegion() {
  bool insideRegion=false;
  bool searchGradient=false;
  bool searchRegion=false;
  bool finish=false;
  double voxelValue;
  double lastRegionDistance=0;
  double firstRegionDistance=0;
  double valGradient = 0.0f ,maxGradient=0.0f;
 int i;

 if ((positive.size()+negative.size())>4) {
	 int minIndex= -(int)negative.size()-1;
	 int maxIndex= positive.size();
	 if (maxIndex== 0)
		 maxIndex= -1;
	 bool inside=ri->getOrientationConstraint()==RegionInrimageInformation::INSIDE;

	 computeGradient(ri->getMinimumLength());   
	 if (inside) {
		 i=minIndex+1;
	 }
	 else {
		 i=maxIndex-1;
	 }
	 double minValue=ri->getMinimumIntensity();
	 double maxValue=ri->getMaximumIntensity();
	 double gradientThreshold=ri->getThreshold();
	 bool test;

	 do {
		 voxelValue=getVoxel(i)->getValue();

		 if(insideRegion) {
			 /// we are still a priori inside the region
			 if((voxelValue<minValue)||(voxelValue>maxValue)) {
				 // but we are leaving it
				 valGradient=getGradient(i);
				 if(fabs(valGradient)>gradientThreshold) {
					 // we found the edge so stop now
					 maxVoxel=getVoxel(i);	 
					 finish=true;
			  }
				 else {
					 // start to look for a gradient we have a short distance to find one
					 lastRegionDistance=getDistance(i);
					 searchGradient=true;
			  }
				 // we have left the region
				 insideRegion=false;
			 }
		 }
		 else if(searchGradient) {
			 // we are looking for a high gradient Voxel
			 valGradient=getGradient(i);
			 if(fabs(valGradient)>gradientThreshold) {
				 // gotcha
				 maxVoxel=getVoxel(i);
				 finish=true;
			 }
			 else if(fabs(getDistance(i)-lastRegionDistance)>ri->getMinimumLength()) {
				 // too late we could not find any high gradient voxel
				 searchGradient=false;
			 }
		 }
		 else if(searchRegion) {
			 // we have already been in the region and we try to get the minimum length
			 if((voxelValue>=minValue) && (voxelValue<=maxValue)) {
				 // found one voxel
				 if(fabs(getDistance(i)-firstRegionDistance)>ri->getMinimumLength()) {
					 // and we have found enough previous voxels to asses that we are inside the region
					 insideRegion=true;
			  }
			 }
			 else
				 searchRegion=false;
		 }
		 else {
			 if((voxelValue>=minValue) && (voxelValue<=maxValue)) {
				 searchRegion=true;
				 firstRegionDistance=getDistance(i);
			 }
		 }
		 if (inside) {
			 ++i;
			 test=(i<maxIndex);
		 } else {
			 --i;
			 test=(i>minIndex);
		 }
	 } while ((finish==false) && test);

	 if (finish==true) {
		 maxGradient=fabs(valGradient);
		 if (inside) {
		   i-=3;
		   test=(i<maxIndex)&& (fabs(getDistance(i)-lastRegionDistance)<3);
		 } else {
		   i+=3;
		   test=(i>minIndex)&& (fabs(getDistance(i)-lastRegionDistance)<3);
		 }
		 while (test) {
			 valGradient=fabs(getGradient(i));
			 if (valGradient>maxGradient) {
				 maxGradient=valGradient;
				 maxVoxel=getVoxel(i);
			 }
			 if (inside) {
				 ++i;
				 test=(i<maxIndex)&& (fabs(getDistance(i)-lastRegionDistance)<3);
			 } else {
				 --i;
				 test=(i>minIndex)&& (fabs(getDistance(i)-lastRegionDistance)<3);
			 }
		 } 
	 }
#ifdef old
  Voxel *gvoxel = Voxel::voxelByType(range->getGradient()->getType());

  if (ri->getSmoothLine())
    smoothProfile();
 
  
  for(;(itNegative!=itNegativeEnd) && (finish==false);++itNegative) {
    voxelValue=(*itNegative)->getValue();
    if(insideRegion) {
      if((voxelValue<ri->getMinimumIntensity())||(voxelValue>ri->getMaximumIntensity())) {
	range->getGradient()->getVoxel((*itNegative)->getX(),
				       (*itNegative)->getY(), 
				       (*itNegative)->getZ(), 
				       *gvoxel);
	if(gvoxel->getValueAsDouble()>ri->getThreshold()) {
	  if((*itNegative)->getX()!=pvoxel->getX()) {
	    if((*itNegative)->getX()<=pvoxel->getX())
	      maxVoxel=pvoxel;
	    else
	      maxVoxel=(*itNegative);
	  }
	  else if((*itNegative)->getY()!=pvoxel->getY()) {
	    if((*itNegative)->getY()<=pvoxel->getY())
	      maxVoxel=pvoxel;
	    else
	      maxVoxel=(*itNegative);
	  }
	  else {
	    if((*itNegative)->getZ()<=pvoxel->getZ())
	      maxVoxel=pvoxel;
	    else
	      maxVoxel=(*itNegative);
	  }
	  //		  maxVoxel=(*itNegative);
	  finish=true;
	}
	else {
	  lastRegionDistance=(*itNegative)->getDistance();
	  searchGradient=true;
	  pvoxel=(*itNegative);
	}
	pvoxel=(*itNegative);
	insideRegion=false;
      }
      else
	pvoxel=(*itNegative);
    }
    else if(searchGradient) {
      range->getGradient()->getVoxel((*itNegative)->getX(),
				     (*itNegative)->getY(), 
				     (*itNegative)->getZ(), 
				     *gvoxel);
      if(gvoxel->getValueAsDouble()>ri->getThreshold()) {
	maxVoxel=(*itNegative);
	finish=true;
      }
      else if(fabs((*itNegative)->getDistance()-lastRegionDistance)>ri->getMinimumLength()) {
	searchGradient=false;
      }
    }
    else if(searchRegion) {
      if((voxelValue>=ri->getMinimumIntensity()) && (voxelValue<=ri->getMaximumIntensity())) {
	if(fabs((*itNegative)->getDistance()-firstRegionDistance)>ri->getMinimumLength()) {
	  insideRegion=true;
	  pvoxel=(*itNegative);
	}
      }
      else
	searchRegion=false;
    }
    else {
      if((voxelValue>=ri->getMinimumIntensity()) && (voxelValue<=ri->getMaximumIntensity())) {
	searchRegion=true;
	firstRegionDistance=(*itNegative)->getDistance();
      }
    }
  }
  std::vector<ImageVoxel *>::iterator itPositive,itPositiveEnd;
  
  if(ri->getOrientationConstraint()==RegionInrimageInformation::INSIDE) {
    itPositive=positive.begin();
    itPositiveEnd=positive.end();
  }
  else {
    itPositive=negative.begin();
    itPositiveEnd=negative.end();
  }
  
  for(;(itPositive!=itPositiveEnd) && (finish==false);++itPositive) {
    voxelValue=(*itPositive)->getValue();
    if(insideRegion) {
      if((voxelValue<ri->getMinimumIntensity())||(voxelValue>ri->getMaximumIntensity())) {
	range->getGradient()->getVoxel((*itPositive)->getX(),
				       (*itPositive)->getY(), 
				       (*itPositive)->getZ(), 
				       *gvoxel);
	if(gvoxel->getValueAsDouble()>ri->getThreshold()) {
	  if((*itPositive)->getX()!=pvoxel->getX()) {
	    if((*itPositive)->getX()<=pvoxel->getX())
	      maxVoxel=pvoxel;
	    else
	      maxVoxel=(*itPositive);
	  }
	  else if((*itPositive)->getY()!=pvoxel->getY()) {
	    if((*itPositive)->getY()<=pvoxel->getY())
	      maxVoxel=pvoxel;
	    else
	      maxVoxel=(*itPositive);
	  }
	  else {
	    if((*itPositive)->getZ()<=pvoxel->getZ())
	      maxVoxel=pvoxel;
	    else
	      maxVoxel=(*itPositive);
	  }
	  //		  maxVoxel=(*itPositive);
	  finish=true;
	}
	else {
	  lastRegionDistance=(*itPositive)->getDistance();
	  searchGradient=true;
	  pvoxel=(*itPositive);
	}
	insideRegion=false;
      }
      else
	pvoxel=(*itPositive);
    }
    else if(searchGradient) {
      range->getGradient()->getVoxel((*itPositive)->getX(),
				     (*itPositive)->getY(), 
				     (*itPositive)->getZ(), 
				     *gvoxel);
      if(gvoxel->getValueAsDouble()>ri->getThreshold()) {
	maxVoxel=(*itPositive);
	finish=true;
      }
      else if(fabs((*itPositive)->getDistance()-lastRegionDistance)>ri->getMinimumLength()) {
	searchGradient=false;
      }
    }
    else if(searchRegion) {
      if((voxelValue>=ri->getMinimumIntensity()) && (voxelValue<=ri->getMaximumIntensity())) {
	if(fabs((*itPositive)->getDistance()-firstRegionDistance)>ri->getMinimumLength()) {
	  insideRegion=true;
	  pvoxel=(*itPositive);
	}
      }
      else
	searchRegion=false;
    }
    else {
      if((voxelValue>=ri->getMinimumIntensity()) && (voxelValue<=ri->getMaximumIntensity())) {
	searchRegion=true;
	firstRegionDistance=(*itPositive)->getDistance();
      }
    }
  }
  delete gvoxel;
#endif
  }
  return maxGradient;
}


bool RangeInrimage::RegionForce::stopFromRegion(ImageVoxel *voxel,
						Voxel&)  {
  return (voxel->getDistance() > maxDistance) ? true : false;
}

bool RangeInrimage::RegionForce::isPossible() const {
	if(range->getIntensity())
		return(true);
	else
		return(false);
}

bool RangeInrimage::RegionForce::getPoint(const Vec3<double> &position, 
					  const Vec3<double> &direction,
					  double range,
					  Vec3<double> &point,double &confidence,
					  RegionInrimageInformation *rii) 
{
  if (rii==0)
    return false;
  ri=rii;
  maxDistance=rii->getMaxDistance();
  maxVoxel = 0;
  confidence=0;
  compute(range, position, direction);
  //if (startvoxel) 
  double maxGradient=findRegion();
  if(!maxVoxel) return false;


  point = maxVoxel->getPosition();
  /// confidence computed as amount above gradientThreshold
  confidence=fabs(maxGradient)-rii->getThreshold();

  /** alternative computes confidence from the 2nd derivative of the gradient profile 
  /// compute confidence value
  int minIndex= -(int)negative.size()-1;
  int maxIndex= positive.size();
  double d=maxVoxel->getDistance();
  double d1,d2,g1,g2;
  if ((d-vsize)<getDistance(-minIndex)) {
	  d1=getDistance(-minIndex);
  } else
	  d1=d-2.0;
  if ((d+2.0)>getDistance(maxIndex-1)) {
	  d2=getDistance(maxIndex-1);
  } else
	  d2=d+2.0;
  g1=fabs(getGradient(d1));
  g2=fabs(getGradient(d2));
  confidence= -(getGradient(d) -(d1*g2-d2*g1)/(d1-d2))/(d1*d2);
*/
  return true;
}


/**********************************************************************/
/**********************************************************************/
/******************   RangeInrimage::RegionProbabilityMapForce   ******/
/**********************************************************************/
/**********************************************************************/

RangeInrimage::RegionProbabilityMapForce::RegionProbabilityMapForce(RangeInrimage *_range)
  throw(RangeInrimage::Force::InvalidRangeInrimage)
  :  VoxelMax(0.0, (bool (VoxelMax::*)(ImageVoxel *, Voxel&))
	    &RangeInrimage::RegionProbabilityMapForce::stopFromDistance),
     Force(_range,"probabilityMap")
{
  inr=range->getIntensity();
  maxVoxel=0;
}

void RangeInrimage::RegionProbabilityMapForce::start(Voxel& voxel) {
 maxVoxel = 0;
}

RangeInrimage::RegionProbabilityMapForce::~RegionProbabilityMapForce() {}

void RangeInrimage::RegionProbabilityMapForce::updateImage(void) {
  inr = range->getIntensity();
}

void RangeInrimage::RegionProbabilityMapForce::findRegion() {
 /* bool insideRegion=false;
  bool searchRegion=false;
  bool finish=false;
  double voxelValue;
  Voxel *gvoxel = Voxel::voxelByType(range->getGradient()->getType());
  ImageVoxel *pvoxel=0;
  double lastRegionDistance=0;
  double firstRegionDistance=0;*/
  std::vector<ImageVoxel *>::reverse_iterator itNegative,itNegativeEnd;


  if (ri->getOrientationConstraint()==RegionProbabilityMapInformation::INSIDE) {
    itNegative=negative.rbegin();
    itNegativeEnd=negative.rend();
  }
  else {
    itNegative=positive.rbegin();
    itNegativeEnd=positive.rend();
  }
  
  std::vector<ImageVoxel *>::iterator itPositive,itPositiveEnd;
  
  if(ri->getOrientationConstraint()==RegionInrimageInformation::INSIDE) {
    itPositive=positive.begin();
    itPositiveEnd=positive.end();
  }
  else {
    itPositive=negative.begin();
    itPositiveEnd=negative.end();
  }
  
}


bool RangeInrimage::RegionProbabilityMapForce::stopFromRegion(ImageVoxel *voxel,
						Voxel&)  {
  return (voxel->getDistance() > maxDistance) ? true : false;
}

bool RangeInrimage::RegionProbabilityMapForce::isPossible() const {
  return ( range->getIntensity()!=0);
}

bool RangeInrimage::RegionProbabilityMapForce::getPoint(const Vec3<double> &position, 
					  const Vec3<double> &direction,
					  double range,
					  Vec3<double> &point,double &confidence,
					  RegionProbabilityMapInformation *rii) 
{
  if (rii==0)
    return false;
  ri=rii;
  maxDistance=rii->getMaxDistance();
  maxVoxel = 0;
  compute(range, position, direction);
  findRegion();
  if(!maxVoxel) return false;
  point = maxVoxel->getPosition();
  return true;
}



/**********************************************************************/
/**********************************************************************/
/******************   RangeInrimage::LocalForce   *********************/
/**********************************************************************/
/**********************************************************************/

RangeInrimage::IntensityProfileForce::IntensityProfileForce(RangeInrimage *range)
  throw(Force::InvalidRangeInrimage)
  : IntensityProfile(10.0), Force(range, "local") {
  inr = range->getIntensity();
}


bool RangeInrimage::IntensityProfileForce::getPoint(const Vec3<double> &position, 
					 const Vec3<double> &direction,
					 double range,
					 Vec3<double> &point,double &confidence,
					 IntensityProfileInformation *rii) {
  if(!rii) return false;
  compute(rii->getMaxDistance(), position, direction);
  if(!startvoxel) return false;
  else return rii->getPoint(*this, point);
}
RegionInrimageInformation::RegionInrimageInformation(double max,
						     double _gradientThreshold,
						     OrientationConstraint _regionConstraint,
						     double _minimumRegionLength, 
						     double _minimumIntensity,
						     double _maximumIntensity,
						     bool _smoothLine) :
  RangeInformation(RegionInrimageInformation::informationName),range(max),minimumRegionLength(_minimumRegionLength),gradientThreshold(_gradientThreshold),regionConstraint(_regionConstraint),minimumIntensity(_minimumIntensity),maximumIntensity(_maximumIntensity),smoothLine(_smoothLine) {
}
  RegionProbabilityMapInformation::RegionProbabilityMapInformation(double max,
		 OrientationConstraint _regionConstraint,
		 double _minimumRegionLength,
		 ProbabilityDensityFunction *_pdfInside,
		 ProbabilityDensityFunction *_pdfOutside) :   RangeInformation(RegionProbabilityMapInformation::informationName),
		 range(max),minimumRegionLength(_minimumRegionLength),regionConstraint(_regionConstraint),
		 insidePDF(_pdfInside),outsidePDF(_pdfOutside) {
		 }
RangeInformation *RegionInrimageInformation::readRegionInformation(GFInput &input) {
  double _range,length,threshold;
  unsigned int val;
  double minI,maxI;
  bool smooth;
  input >> _range >> length >> threshold >> val >> minI >> maxI >> smooth ;
  OrientationConstraint constraint=(OrientationConstraint) val;
  return((RangeInformation *)new RegionInrimageInformation(_range,threshold,constraint,length,minI,maxI,smooth));
}

void RegionInrimageInformation::writeInFile(GFOutput &output) {
  output << range << minimumRegionLength << gradientThreshold << (unsigned int )regionConstraint << minimumIntensity <<maximumIntensity << smoothLine <<ENDL;
}

GradientInrimageInformation::GradientInrimageInformation(double _max,
							 double _gradientThreshold,
							 OrientationConstraint _gradientConstraint,
							 bool _stopAtFirst ,
							 bool _useIntensity,
							 double _minIntensity,
							 double _maxIntensity):RangeInformation("Gradient"),range(_max),gradientThreshold(_gradientThreshold),gradientConstraint(_gradientConstraint),stopAtFirst(_stopAtFirst),useIntensity(_useIntensity),minimumIntensity(_minIntensity),maximumIntensity(_maxIntensity) {
}

RangeInformation *GradientInrimageInformation::readGradientInformation(GFInput &input) {
  double _range,threshold;
  unsigned int val;
  double minI,maxI;
  bool _useIntensity,_stopAtFirst;
  input >> _range >> threshold >> val >> _stopAtFirst >> _useIntensity >> minI >> maxI ;
  OrientationConstraint constraint=(OrientationConstraint) val;
  GradientInrimageInformation *gii=new GradientInrimageInformation(_range,threshold,constraint,_stopAtFirst,_useIntensity,minI,maxI);
  RangeInformation *ri=(RangeInformation *) gii;
  gii=(GradientInrimageInformation *)ri;
  return(ri);
}


void GradientInrimageInformation::writeInFile(GFOutput &output) {
  output << range << gradientThreshold << (unsigned int )gradientConstraint << stopAtFirst << useIntensity << minimumIntensity <<maximumIntensity << ENDL;
}

ClassInrimageInformation::ClassInrimageInformation(double _max,
						   unsigned int _nin,
						   unsigned char *_in,
						   unsigned int _nout,
						   unsigned char *_out) : 
  RangeInformation(ClassInrimageInformation::informationName),range(_max),nin(_nin),nout(_nout), inlength(1), outlength(1) 
{
  if (nin)
    {
      in= new unsigned char [nin];
    }
  else 
    in=0;
  if (nout)
    {
      out= new unsigned char [nout];
    }
  else
    out=0;
  unsigned int i;
  for(i = 0; i < nin; i++) in[i] = _in[i];
  for(i = 0; i < nout; i++) out[i] = _out[i];
}
void ClassInrimageInformation::setInlength(double d) {
  if(d < 0.0) d = 1.0;
  inlength = d;
}
void ClassInrimageInformation::setOutlength(double d) {
  if(d < 0.0) d = 1.0;
  outlength = d;
}
RangeInformation *ClassInrimageInformation::readClassInformation(GFInput &input) {
  double _range;
  unsigned int _nin,_nout,i;
  unsigned char *_in,*_out;

  input >> _range >> _nin >> _nout;
  _in= new unsigned char[_nin];
  _out= new unsigned char[_nout];

  for(i=0;i<_nin;++i)
    input >> _in[i];
  for(i=0;i<_nout;++i)
    input >> _out[i];

  return((RangeInformation *)new ClassInrimageInformation(_range,_nin,_in,
							  _nout,_out));
}

void ClassInrimageInformation::writeInFile(GFOutput &output) {
  output << range << nin << nout << ENDL;
  unsigned int i;
  for(i=0;i<nin;++i)
    output << (const char &)in[i];
  for(i=0;i<nout;++i)
    output << (const char &)out[i];
}

IsosurfaceInrimageInformation::IsosurfaceInrimageInformation(double _isovalue,
							     double r, IsosurfaceConstraint def) : RangeInformation(informationName),isovalue(_isovalue),range(r), isosurfaceConstraint(def){
}

RangeInformation *IsosurfaceInrimageInformation::readIsosurfaceInformation(GFInput &input) {
  double threshold,r;
  unsigned int c;
  input >> threshold >> r >> c;
  return((RangeInformation *)new IsosurfaceInrimageInformation(threshold,r,(IsosurfaceConstraint)c));
}

void IsosurfaceInrimageInformation::writeInFile(GFOutput &output) {
  output << isovalue << range << isosurfaceConstraint <<ENDL;
}


GradientGradientInrimageInformation::GradientGradientInrimageInformation(double _gradientThreshold,
  OrientationConstraint _gradientConstraint) : RangeInformation(informationName),gradientThreshold(_gradientThreshold),gradientConstraint(_gradientConstraint) {
}
RangeInformation *GradientGradientInrimageInformation::readGradientGradientInformation(GFInput &input) {
  double threshold;
  unsigned int val;
  input >> threshold >> val ;
  OrientationConstraint constraint=(OrientationConstraint) val;
  return((RangeInformation *)new GradientGradientInrimageInformation(threshold,constraint));
}

void GradientGradientInrimageInformation::writeInFile(GFOutput &output) {
  output << gradientThreshold << (unsigned int )gradientConstraint << ENDL;
}
GradientBalloonInrimageInformation::GradientBalloonInrimageInformation(double _gradientThreshold, OrientationConstraint _gradientConstraint, double _balloonForce) : RangeInformation(GradientBalloonInrimageInformation::informationName),gradientThreshold(_gradientThreshold),gradientConstraint(_gradientConstraint),balloonForce(_balloonForce) {
}
RangeInformation *GradientBalloonInrimageInformation::readGradientBalloonInformation(GFInput &input) {
  double threshold,force;
  unsigned int val;
  input >> threshold >> val >> force;
  OrientationConstraint constraint=(OrientationConstraint) val;
  return((RangeInformation *)new GradientBalloonInrimageInformation(threshold,constraint,force));
}

void GradientBalloonInrimageInformation::writeInFile(GFOutput &output) {
  output << gradientThreshold << (unsigned int )gradientConstraint << balloonForce << ENDL;
}

/**********************************************************************/
/**********************************************************************/
/******************   RangeInrimage::ClassForce   *********************/
/**********************************************************************/
/**********************************************************************/


RangeInrimage::ClassForce::ClassForce(RangeInrimage *range, double forceRange)
  throw(Force::InvalidRangeInrimage)
  : IntensityProfile(forceRange), Force(range, "classify") {
  inr = range->getClassify();
}

bool RangeInrimage::ClassForce::isPossible() const {
  return range->getClassify() != 0;
}



void RangeInrimage::ClassForce::updateImage(void) {
  inr = range->getClassify();
}





bool RangeInrimage::ClassForce::getPoint(const Vec3<double> &pos, 
					 const Vec3<double> &dir,
					 double max,
					 Vec3<double> &point,double &confidence,
					 ClassInrimageInformation *rii) {
if (rii)
  {
    compute(max, pos, dir);
    
    const unsigned char *in,*out;
    unsigned int nin,nout;
    in=rii->getInArray(nin);
    out=rii->getOutArray(nout);

    unsigned int p = positiveArraySize(), n = negativeArraySize();
    if(p + n < 3) return false;
    
    bool *incriterion = new bool[p + n - 1];
    bool *outcriterion = new bool[p + n - 1]; 
    double startd = MIN_DOUBLE;
    int i;
    
    for(i = -(int) n+1; i < (int) p; i++) {
      incriterion[i+n-1] = false;
      
      ImageVoxel *v = (*this)[i];
      bool match = false;
      for(unsigned int j = 0; j < nin; j++)
	if((unsigned int) v->getValue() == in[j]) match = true;
      
      if(match) {
	if(startd > MIN_DOUBLE) {
	  double d = (i >= 0) ? v->getDistance() : -v->getDistance();
	  if((d - startd) >= rii->getInlength()) incriterion[i+n-1] = true;
	}
	else startd = (i >= 0) ? v->getDistance() : -v->getDistance();
      }
      else startd = MIN_DOUBLE;
    }
    
    for(i = (int) p-1; i > - (int) n; i--) {
      outcriterion[i+n-1] = false;
      
      ImageVoxel *v = (*this)[i];
      bool match = false;
      for(unsigned int j = 0; j < nout; j++)
	if((unsigned int) v->getValue() == out[j]) match = true;
      
      if(match) {
	if(startd > MIN_DOUBLE) {
	  double d = (i >= 0) ? v->getDistance() : -v->getDistance();
	  if(startd - d >= rii->getOutlength()) outcriterion[i+n-1] = true;
	}
	else startd = (i >= 0) ? v->getDistance() : -v->getDistance();
      }
      else startd = MIN_DOUBLE;
    }
    
    int besti=0;
    double bestd = -1.0;
    for(i = 1; i < (int) (p+n-1); i++) {
      if(incriterion[i-1] && outcriterion[i]) {
	double d = ((*this)[i-n]->getDistance() +
		    (*this)[i-n+1]->getDistance()) / 2.0;
	if(bestd < 0.0 || d < bestd) {
	  bestd = d;
	  besti = i-n;
	}
      }
    }
    delete incriterion;
    delete outcriterion;
    
    if(bestd < 0.0) return false;
    
    Vec3<double> position = (*this)[besti]->getPosition();
    position += (*this)[besti+1]->getPosition();
    position /= 2.0;
#ifndef OLD_STUFF
    point = Vec3<double>
      (position.getX(), position.getY(), position.getZ());
    
#else
    point = Vec3<double>
      (position.getX() * range->getIntensity()->getVX(),
       position.getY() * range->getIntensity()->getVY(),
       position.getZ() * range->getIntensity()->getVZ());
#endif
    return true;
  }
else
   return false;
}


/**********************************************************************/
/**********************************************************************/
/******************   RangeInrimage::BlockMatchingForce   *************/
/**********************************************************************/
/**********************************************************************/


RangeInrimage::BlockMatchingForce::BlockMatchingForce(RangeInrimage *range)
  throw(Force::InvalidRangeInrimage)
  : Force(range, "blockMatching") {
}

bool RangeInrimage::BlockMatchingForce::isPossible() const {
  return range->getIntensity() != 0;
}

bool RangeInrimage::BlockMatchingForce::getPoint(const Vec3<double> &pos, 
					 const Vec3<double> &dir,
					 double max,
					 Vec3<double> &point,double &confidence,
					 BlockMatchingInformation *rii) {
  if(rii->isInsideRef()) {
    Inrimage * tarImage = range->getIntensity();
    Inrimage * refImage = rii->getRefImage();
    Voxel * tarVox = Voxel::voxelByType(tarImage->getType());
    Voxel * refVox = Voxel::voxelByType(refImage->getType());
    
    Vec3<int> bs = (Vec3<int>) rii->getBlockSize();
    Vec3<int> sr = (Vec3<int>) rii->getSearchRange();
    Vec3<int> st = (Vec3<int>) rii->getSearchStep();
    Vec3<unsigned int> refCenter = rii->getBlockCenter();
    BlockMatchingInformation::SimilarityMeasure sm = rii->getSimilarityMeasure();
    double vx = tarImage->getVX(), vy = tarImage->getVY(), vz = tarImage->getVZ();
    Vec3<int> tarCenter = Vec3<int>((int)floor(pos[0]/vx),(int)floor(pos[1]/vy),(int)floor(pos[2]/vz));
    
    Vec3<unsigned int> res;
    
    int xTarMin = tarCenter[0]-sr[0];
    int xTarMax = tarCenter[0]+sr[0];
    int yTarMin = tarCenter[1]-sr[1];
    int yTarMax = tarCenter[1]+sr[1];
    int zTarMin = tarCenter[2]-sr[2];
    int zTarMax = tarCenter[2]+sr[2];    
    int i, j, k, u, v, w;    
    
    if(sm==BlockMatchingInformation::CORRELATION_COEFFICIENT) {
      double nbVox = (double) (2.0*bs[0]+1.0)*(2.0*bs[1]+1.0)*(2.0*bs[2]+1.0);
      double val;
      double simMax=0.0, cov=0.0, tarMean=0.0, tarVariance=0.0, sim;
      double refMean = rii->getRefMean(), refVariance=rii->getRefVariance();
      
      for(i=xTarMin;i<=xTarMax;i+=st[0]) {
	for(j=yTarMin;j<=yTarMax;j+=st[1]) {
	  for(k=zTarMin;k<=zTarMax;k+=st[2]) {
	    if(tarImage->inside(i-bs[0],j-bs[1],k-bs[2]) && tarImage->inside(i+bs[0],j+bs[1],k+bs[2])) {
	      cov=0.0;
	      tarMean=0.0;
	      tarVariance=0.0;	       
	      for(u=-bs[0];u<=bs[0];++u) {
		for(v=-bs[1];v<=bs[1];++v) {
		  for(w=-bs[2];w<=bs[2];++w) {
		    val = tarImage->getVoxel(i+u,j+v,k+w,*tarVox).getValueAsDouble();
		    tarMean += val;
		    tarVariance += val * val;
		    cov +=val*refImage->getVoxel(refCenter[0]+u,refCenter[1]+v,refCenter[2]+w,*refVox).getValueAsDouble();
		  }
		}
	      }
	      tarMean /= nbVox;
	      tarVariance /= nbVox-1.0;
	      tarVariance -= tarMean*tarMean;
	      cov /= nbVox-1.0;
	      cov -= tarMean*refMean;
	      if(refVariance==0.0 || tarVariance==0.0)
		sim = 0.0;
	      else
		sim = cov*cov/(refVariance*tarVariance);
	      if(sim>simMax) {
		simMax=sim;
		res.setValue(i,j,k);
	      }
	      else if(sim==simMax) {
		if(((i-tarCenter[0])*(i-tarCenter[0])*vx*vx+(j-tarCenter[1])*(j-tarCenter[1])*vy*vy+(k-tarCenter[2])*(k-tarCenter[2])*vz*vz)<((res[0]-tarCenter[0])*(res[0]-tarCenter[0])*vx*vx+(res[1]-tarCenter[1])*(res[1]-tarCenter[1])*vy*vy+(res[2]-tarCenter[2])*(res[2]-tarCenter[2])*vz*vz)) {
		  simMax=sim;
		  res.setValue(i,j,k);
		}
	      }
	    }
	  }
	}
      }
      // we are computing the square correlation coefficient and 0.5 ~ 0.7^2 
      if(simMax<0.5) 
	return false;
    }
    
    else if(sm==BlockMatchingInformation::SSD) {
      double valt, valr;
      double simMax=MAX_DOUBLE, sim;
      double nbVox = (double) (2.0*bs[0]+1.0)*(2.0*bs[1]+1.0)*(2.0*bs[2]+1.0);
      for(i=xTarMin;i<=xTarMax;i+=st[0]) {
	for(j=yTarMin;j<=yTarMax;j+=st[1]) {
	  for(k=zTarMin;k<=zTarMax;k+=st[2]) {
	    if(tarImage->inside(i-bs[0],j-bs[1],k-bs[2]) && tarImage->inside(i+bs[0],j+bs[1],k+bs[2])) {
	      sim=0.0;
	      for(u=-bs[0];u<=bs[0];++u) {
		for(v=-bs[1];v<=bs[1];++v) {
		  for(w=-bs[2];w<=bs[2];++w) {
		    valt = tarImage->getVoxel(i+u,j+v,k+w,*tarVox).getValueAsDouble();
		    valr = refImage->getVoxel(refCenter[0]+u,refCenter[1]+v,refCenter[2]+w,*refVox).getValueAsDouble();
		    sim += (valt-valr)*(valt-valr);
		  }
		}
	      }
	      if(sim<simMax) {
		simMax=sim;
		res.setValue(i,j,k);
	      }
	      else if(sim==simMax) {
		if(((i-tarCenter[0])*(i-tarCenter[0])*vx*vx+(j-tarCenter[1])*(j-tarCenter[1])*vy*vy+(k-tarCenter[2])*(k-tarCenter[2])*vz*vz)<((res[0]-tarCenter[0])*(res[0]-tarCenter[0])*vx*vx+(res[1]-tarCenter[1])*(res[1]-tarCenter[1])*vy*vy+(res[2]-tarCenter[2])*(res[2]-tarCenter[2])*vz*vz)) {
		  simMax=sim;
		  res.setValue(i,j,k);
		}
	      }
	    }
	  }
	}
      }
      // we are computing the ssd. Good threshold ?
      if((simMax/nbVox)>(2.0*rii->getRefVariance()))
	return false;
    }
    
    else if(sm==BlockMatchingInformation::PAIREDTTEST_STUDENT) {
      double valt, valr;
      double simMax=MAX_DOUBLE, sim;
      double refMean = rii->getRefMean(), tarMean, dif;
      double nbVox = (double) (2.0*bs[0]+1.0)*(2.0*bs[1]+1.0)*(2.0*bs[2]+1.0);
      for(i=xTarMin;i<=xTarMax;i+=st[0]) {
	for(j=yTarMin;j<=yTarMax;j+=st[1]) {
	  for(k=zTarMin;k<=zTarMax;k+=st[2]) {
	    if(tarImage->inside(i-bs[0],j-bs[1],k-bs[2]) && tarImage->inside(i+bs[0],j+bs[1],k+bs[2])) {
	      sim=0.0;
	      tarMean = 0.0;
	      dif = 0.0;
	      for(u=-bs[0];u<=bs[0];++u) {
		for(v=-bs[1];v<=bs[1];++v) {
		  for(w=-bs[2];w<=bs[2];++w) {
		    valt = tarImage->getVoxel(i+u,j+v,k+w,*tarVox).getValueAsDouble();
		    valr = refImage->getVoxel(refCenter[0]+u,refCenter[1]+v,refCenter[2]+w,*refVox).getValueAsDouble();
		    tarMean += valt;
		    dif += (valt-valr)*(valt-valr);
		  }
		}
	      }
	      tarMean /= nbVox;
	      sim = dif - nbVox*(refMean-tarMean)*(refMean-tarMean);
	      if(sim!=0.0)
		sim = (refMean-tarMean)*(refMean-tarMean)*(nbVox-1.0)*nbVox/sim;
	      if(sim<simMax) {
		simMax=sim;
		res.setValue(i,j,k);
	      }
	      else if(sim==simMax) {
		if(((i-tarCenter[0])*(i-tarCenter[0])*vx*vx+(j-tarCenter[1])*(j-tarCenter[1])*vy*vy+(k-tarCenter[2])*(k-tarCenter[2])*vz*vz)<((res[0]-tarCenter[0])*(res[0]-tarCenter[0])*vx*vx+(res[1]-tarCenter[1])*(res[1]-tarCenter[1])*vy*vy+(res[2]-tarCenter[2])*(res[2]-tarCenter[2])*vz*vz)) {
		  simMax=sim;
		  res.setValue(i,j,k);
		}
	      }
	    }
	  }
	}
      }
      // we are computing the Paired t-Test. The threshold is from the Student's distribution.
      // (two tailed distribution -> interval*2 : 95% -> 97.5% confidence interval, nbVox > 30)
      // 97.5 % : threshold = 2.0, we are using t^2 ->4.0 ... big threshold ...
      if(simMax>4)
	return false;
    }
    
    delete tarVox;
    delete refVox;
    if (tarImage->getGeometry()==Inrimage::CARTESIAN) {
      point.setValue(((double)res[0]+0.5)*vx,((double)res[1]+0.5)*vy,((double)res[2]+0.5)*vz);
    }
    else if (tarImage->getGeometry()==Inrimage::CYLINDRICAL) {
      ImageGeometry * converter=tarImage->newConverter();
      converter->setImage((double)(res[0]+0.5),(double)(res[1]+0.5),(double)(res[2]+0.5));
      double p[3];
      converter->toCartesian(p);
      delete converter;
      point.setValue(p[0],p[1],p[2]);
    }
    if(rii->isAlongNormal())
      point = pos + dot(dir,point-pos)*dir;

    point -= rii->getOffset();

    return true;
  }  
  else
    return false;
}

BlockMatchingInformation::BlockMatchingInformation(Inrimage * refim, Vec3<unsigned int> center, Vec3<unsigned int> bs, Vec3<unsigned int> sr, Vec3<unsigned int> st, SimilarityMeasure sm, Vec3<double> os) : RangeInformation(BlockMatchingInformation::informationName), refImage(refim), blockCenter(center), blockSize(bs), searchRange(sr), searchStep(st), similarityMeasure(sm), offset(os) {
  init();
}


void BlockMatchingInformation::init(void) {
  alongNormal = false;
  if(refImage->inside(blockCenter[0]-blockSize[0],blockCenter[1]-blockSize[1],blockCenter[2]-blockSize[2]) && refImage->inside(blockCenter[0]+blockSize[0],blockCenter[1]+blockSize[1],blockCenter[2]+blockSize[2]))
    insideRef=true;
  else
    insideRef=false;
  if(insideRef && (similarityMeasure != BlockMatchingInformation::SSD)) {
    computeRefBlockStats();
  }
}
  
RangeInformation *BlockMatchingInformation::readBlockMatchingInformation(GFInput &input) {
  TclString refname, sim;
  Vec3<unsigned int> center, bs, sr, st;
  SimilarityMeasure sm=SSD;
  input >> refname >> center >> bs >> sr >> st >> sim;
  Inrimage * ref; 
  if(bmRefImages.find(refname)==bmRefImages.end()) {
    ref = new Inrimage(refname);
    bmRefImages[refname]=ref;
  }
  else
    ref = bmRefImages[refname];
  if(sim=="CORRELATION_COEFFICIENT") sm=CORRELATION_COEFFICIENT;
  else if(sim=="SSD") sm=SSD;
  else if(sim=="PAIREDTTEST_STUDENT") sm = PAIREDTTEST_STUDENT;
  else CERR<<"No similarity measure, assuming SSD"<<ENDL;
  return((RangeInformation *)new BlockMatchingInformation(ref,center,bs,sr,st,sm));
}

void BlockMatchingInformation::writeInFile(GFOutput &output) {
  output << refImage->getName() << blockCenter << blockSize << searchRange << searchStep << similarityMeasure << ENDL;
}

void BlockMatchingInformation::computeRefBlockStats(void) {
  Voxel * refVox = Voxel::voxelByType(refImage->getType());
  refMean=0.0;
  refVariance=0.0;
  unsigned int xmin = blockCenter[0] - blockSize[0];
  unsigned int xmax = blockCenter[0] + blockSize[0];
  unsigned int ymin = blockCenter[1] - blockSize[1];
  unsigned int ymax = blockCenter[1] + blockSize[1];
  unsigned int zmin = blockCenter[2] - blockSize[2];
  unsigned int zmax = blockCenter[2] + blockSize[2];

  double val;
  unsigned int i,j,k;
  for(i=xmin;i<=xmax;++i) {
    for(j=ymin;j<=ymax;++j) {
      for(k=zmin;k<=zmax;++k) {
	val = refImage->getVoxel(i,j,k,*refVox).getValueAsDouble();
	refMean += val;
	refVariance += val * val;
      }
    }
  }
  double nbVox = (double) ((2*blockSize[0]+1)*(2*blockSize[1]+1)*(2*blockSize[2]+1));
  refMean /= nbVox;
  refVariance /= nbVox-1.0;
  refVariance -= refMean*refMean;
}

IntensityProfileInformation::IntensityProfileInformation(const IntensityProfile& line, double o)
  : RangeInformation(IntensityProfileInformation::informationName),origin(o), criterion(DIFFERENCES), threshold(50.0) {
  unsigned int j = 0;
  double d = line.negativeArraySize() > 0
    ? line[1-(line.negativeArraySize())]->getDistance()
    : 0.0;
  for(int i = -((int) line.negativeArraySize() - 1);
      i < (int) line.positiveArraySize(); i++, j++) {
    ImageVoxel *v = line[i];
    profile.push_back(element(i < 0 ? d - v->getDistance()
			            : d + v->getDistance(), v->getValue()));
  }
}

IntensityProfileInformation::IntensityProfileInformation(unsigned int size, double o,
				   double *distance, double *intensity,
				   Criterion _criterion,double _threshold)
  : RangeInformation(IntensityProfileInformation::informationName),origin(o), criterion(_criterion), threshold(_threshold) {

  for(unsigned int i = 0; i < size; i++)
    profile.push_back(element(distance[i], intensity[i]));
  range=10.0;
}

IntensityProfileInformation::~IntensityProfileInformation() {
}


double IntensityProfileInformation::getIntensity(double d) const {
  if(d < 0.0 || d > profile[profile.size() - 1].distance) return 0.0;
  int i = 0;
  while(profile[i + 1].distance < d) i++;
  double a = d - profile[i].distance;
  double b = a + profile[i + 1].distance - d;
  if(b != 0.0) a /= b;
  //a /= a + profile[i + 1].distance - d;
  return a * profile[i + 1].value + (1.0 - a) * profile[i].value;
}



bool IntensityProfileInformation::getPoint(const IntensityProfile& line, Vec3<double>& point) {
  if(profile.size() < 1) return false;

  int index = 1-(line.negativeArraySize());
  double mind = line[index]->getDistance();
  double llength = line[index]->getDistance() +
    line[line.positiveArraySize() - 1]->getDistance();
  double plength = profile.back().distance;
  double based = 0.0;
  int bindex = index;
  double bgamma;

  switch(criterion) {
  case DIFFERENCES:
    {
      bgamma = 1E37;
      while(based + plength < llength) {
	double d = 0.0;
	int ii = index, n = 0;
	double gamma = 0.0;
      
	while(d < plength) {
	  gamma += fabs(line[ii]->getValue() - getIntensity(d));
	  n++;
	  ii++;
	  d = ((ii < 0) ? mind - line[ii]->getDistance()
	       : mind + line[ii]->getDistance()) - based;
	}
	if(n > 0) {
	  gamma /= n;
	  
	  if(gamma < bgamma) {
	    bgamma = gamma;
	    bindex = index;
	  }
	}
	index++;
	based = (index < 0) ? mind - line[index]->getDistance()
	  : mind + line[index]->getDistance();
      }
      
      if(bgamma < threshold) {
	double d = (bindex < 0) ? -line[bindex]->getDistance()
	  :  line[bindex]->getDistance();
	point = line.getOrigin() + (d + origin) * line.getDirection();
	return true;
      }
    }
    break;

  case LINEAR:
    {
      bgamma = 0.0;
      while(based + plength < llength) {
	double d = 0.0, gamma = 0.0;
	int ii = index;
	
 	double nx = 0.0, ny = 0.0;
	while(d < plength) {
	  double x = line[ii]->getValue();
	  double y = getIntensity(d);
	  nx += x * x;
	  ny += y * y;
	  gamma += x * y;
	  ii++;
	  d = ((ii < 0) ? mind - line[ii]->getDistance()
	       : mind + line[ii]->getDistance()) - based;
 	}
	
	gamma *= gamma;
	if(nx > 0.0 && ny > 0.0) {
	  gamma /= (nx * ny);
	  if(gamma > bgamma) {
	    bgamma = gamma;
	    bindex = index;
	  }
	}
 	index++;
	based = (index < 0) ? mind - line[index]->getDistance()
	  : mind + line[index]->getDistance();
     }

      if(bgamma > threshold) {
	double d = (bindex < 0) ? -line[bindex]->getDistance()
	  :  line[bindex]->getDistance();
	point = line.getOrigin() + (d + origin) * line.getDirection();
	return true;
      }
    }
    break;

  case CORRELATION:
    {
      bgamma = 0.0;
      while(based + plength < llength) {
	double d = 0.0, ex = 0.0, ey = 0.0;
	int ii = index, n = 0;
	
	while(d < plength) {
	  ex += line[ii]->getValue();
	  ey += getIntensity(d);
	  n++;
	  ii++;
	  d = ((ii < 0) ? mind - line[ii]->getDistance()
	       : mind + line[ii]->getDistance()) - based;
	}
 	if(n > 0) {
	  ex /= n;
	  ey /= n;
	
	  double vx = 0.0, vy = 0.0;
	  d = 0.0;
	  ii = index;
	  while(d < plength) {
	    double t = line[ii]->getValue() - ex;
	    vx += t * t;
	    t = getIntensity(d) - ey;
	    vy += t * t;
	    ii++;
	    d = ((ii < 0) ? mind - line[ii]->getDistance()
		 : mind + line[ii]->getDistance()) - based;
	  }
	
	  double gamma = 0.0;
	  if(vx > 0.0 && vy > 0.0) {
	    d = 0.0;
	    ii = index;
	    while(d < plength) {
	      gamma += (line[ii]->getValue() - ex) * (getIntensity(d) - ey);
	      ii++;
	      d = ((ii < 0) ? mind - line[ii]->getDistance()
		   : mind + line[ii]->getDistance()) - based;
	    }
	    gamma *= gamma;
	    gamma /= (vx * vy);
	  }

	  if(gamma > bgamma) {
	    bgamma = gamma;
	    bindex = index;
	  }
	}
 	index++;
	based = (index < 0) ? mind - line[index]->getDistance()
	  : mind + line[index]->getDistance();
      }

      if(bgamma > threshold) {
	double d = (bindex < 0) ? -line[bindex]->getDistance()
	  :  line[bindex]->getDistance();
	point = line.getOrigin() + (d + origin) * line.getDirection();
	return true;
      }
    }
    break;

  case PATTERN_INTENSITY:
    {
      bgamma = maxTypeValue<float>(0);
      double *diff = new double[line.negativeArraySize() +
			       line.positiveArraySize()];
      while(based + plength < llength) {
	const double sigma = 10.0 * 10.0;
	const int r = 3;

	double d = 0.0;
	int ii = index, n = 0;
	while(d < plength) {
	  diff[n] = line[ii]->getValue() - getIntensity(d);
	  n++;
	  ii++;
	  d = ((ii < 0) ? mind - line[ii]->getDistance()
	       : mind + line[ii]->getDistance()) - based;
	}
 	if(n > 0) {
	  double gamma = 0.0;
	  for(int i = 0; i < n; i++) {
	    for(int j = ((i - r) < 0) ? 0 : i - r;
		j < ((i + r > n) ? n : i + r); j++) {
	      double t = diff[i] - diff[j];
	      gamma += 1.0 / (sigma + t * t);
	    }
	  }
	  gamma *= sigma;
	  
	  if(gamma < bgamma) {
	    bgamma = gamma;
	    bindex = index;
	  }
	}
	index++;
	based = (index < 0) ? mind - line[index]->getDistance()
	  : mind + line[index]->getDistance();
      }
      delete[] diff;

      if(bgamma < threshold) {
	double d = (bindex < 0) ? -line[bindex]->getDistance()
	  :  line[bindex]->getDistance();
	point = line.getOrigin() + (d + origin) * line.getDirection();
	return true;
      }
    }
    break;
  }

  return false;
}


void IntensityProfileInformation::setCriterion(Criterion c, double t) {
  switch(c) {
  case DIFFERENCES:
    criterion = DIFFERENCES;
    threshold = (t < 0.0) ? 50.0 : t;
    break;
  case LINEAR:
    criterion = LINEAR;
    threshold = (t < 0.0 || t > 1.0) ? 0.5 : t;
    break;
  case CORRELATION:
    criterion = CORRELATION;
    threshold = (t < 0.0 || t > 1.0) ? 0.5 : t;
    break;
  case PATTERN_INTENSITY:
    criterion = PATTERN_INTENSITY;
    threshold = (t < 0.0) ? 30 : t;
    break;
  }
}
RangeInformation *IntensityProfileInformation::readIntensityProfile(GFInput &input)
{
  unsigned int i,nb;
  double o,t;
  input >> nb >> o >> t;
  double *intensity= new double[nb];
  double *distance= new double[nb];
  for(i=0;i<nb;++i)
    {
      input >> distance[i] >> intensity[i];
    }
  IntensityProfileInformation *ip= new IntensityProfileInformation(nb,o,distance,intensity);
  delete [] distance;
  delete [] intensity;

  unsigned int val;
  input >> val;
  IntensityProfileInformation::Criterion c= (IntensityProfileInformation::Criterion) val;
  ip->setCriterion(c,t);
  return ((RangeInformation *)ip);
}
void IntensityProfileInformation::writeInFile(GFOutput &output) {
  output << (int)profile.size() << origin << threshold << ENDL;
  std::vector<element>::iterator it;
  for(it=profile.begin();it!=profile.end();it++)
    {
      output << (*it).distance << (*it).value;
    }
  output << ENDL;

  output << (unsigned int) criterion << ENDL;
}
RangeInformation *IntensityProfileInformation::duplicate() {
  IntensityProfileInformation *ip=new IntensityProfileInformation(*this);
  return (ip);
}

} // namespace yav
