#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "imagedata/RangeInrimage2D.h"
#include <imagedata/RangeInrimage2D.h>

namespace yav {
RangeInrimage2D::RangeInrimage2D(Inrimage *intensity,
				 Inrimage *gradient, Inrimage *gradientdir,
				 Inrimage *edge, Inrimage *gradient2,
				 Inrimage *classify)
  throw(Inrimage::ImageNotFound, Inrimage::UnknownImageType,
	Inrimage::InvalidHeader, Inrimage::InvalidData,
	Inrimage::UnknownType)
  :    RangeInrimage(intensity, gradient, gradientdir, edge, gradient2, classify)
 {
   initFormat();

}


RangeInrimage2D::RangeInrimage2D(const TclString& fileName)
  throw(Inrimage::ImageNotFound, Inrimage::UnknownImageType,
	Inrimage::InvalidHeader, Inrimage::InvalidData,
	Inrimage::UnknownType)
  : 
RangeInrimage(fileName)
 {
 initFormat();
}


RangeInrimage2D::~RangeInrimage2D() {
}

bool RangeInrimage2D::getClosestPoint(const Vec2<double> &position,
				      const Vec2<double> &normal,
				      const double range,
				      Vec2<double> &closest,double &confidence,
				      SLICE_ORIENTATION geometry,
				      double sliceNumber,
				      const RangeInformation *clientData) const
{
  Vec3<double> pos,norm;
  switch(geometry)
    {
    case SLICE_X:
      pos.setValue(sliceNumber+.5,position.getX(),
		   position.getY());
      norm.setValue(0.0,normal.getX(),normal.getY());
      break;
    case SLICE_Y:
      pos.setValue(position.getY(),sliceNumber+.5,
		   position.getX());
      norm.setValue(normal.getY(),0.0,normal.getX());
      break;
    default:
      pos.setValue(position.getX(),intensity->getY()-position.getY()-1,
		   sliceNumber+.5);
      norm.setValue(normal.getX(),-normal.getY(),0.0);
      break;
    }
  Vec3<double> closest3D; 
  bool found;
  if (clientData) {
  if ((clientData->getName()==GradientInrimageInformation::informationName) && gradientForce)
    found=gradientForce->getPoint(pos, norm, range, closest3D,confidence,
				  (GradientInrimageInformation *) clientData);
  else if ((clientData->getName()==GradientGradientInrimageInformation::informationName) && gradientGradientForce)
    found=gradientGradientForce->getPoint(pos, norm, range, closest3D,confidence,
					  (GradientGradientInrimageInformation *) clientData);
  else if ((clientData->getName()==GradientBalloonInrimageInformation::informationName) && gradientBalloonForce)
    found=gradientBalloonForce->getPoint(pos, norm, range, closest3D,confidence,
					 (GradientBalloonInrimageInformation *) clientData);
  else if ((clientData->getName()==RegionInrimageInformation::informationName) && regionForce)
    found=regionForce->getPoint(pos, norm, range, closest3D,confidence,
				(RegionInrimageInformation *) clientData);
    else if ((clientData->getName()==IsosurfaceInrimageInformation::informationName) && isosurfaceForce)
    found=isosurfaceForce->getPoint(pos, norm, range, closest3D,confidence,
				(IsosurfaceInrimageInformation *) clientData);
  else if ((clientData->getName()==ClassInrimageInformation::informationName) && classForce)
    found=classForce->getPoint(pos, norm, range, closest3D,confidence,
			       (ClassInrimageInformation *) clientData);
  else if ((clientData->getName()==IntensityProfileInformation::informationName) && intensityProfileForce)
    return(intensityProfileForce->getPoint(pos, norm, range, closest3D,confidence,
					   (IntensityProfileInformation *) clientData));
  else
    found=false;
  if (found)
    {
      switch(geometry)
	{
	case SLICE_X:
	  closest.setValue(closest3D.getY(),closest3D.getZ());
	  break;
	case SLICE_Y:
	  closest.setValue(closest3D.getZ(),closest3D.getX());
	  break;
	default:
	  closest.setValue(closest3D.getX(),intensity->getY()-closest3D.getY()-1);
	  break;
	}
      return true;
    }
  }
  return false;
}



void RangeInrimage2D::getSpeed(const Vec2<double> &position, double &mul,
			       double &add,
			       const RangeInformation *clientData) const {
  
  if ((clientData->getName()==GradientBalloonInrimageInformation::informationName) && gradientBalloonForce)
    gradientBalloonForce->getSpeed(Vec3<double>(position.getX(),
						position.getY(), 0.5), mul, add,
				   (GradientBalloonInrimageInformation *)clientData);
  else {
    mul = 1.0;
    add = 0.0;
  }
}




} // namespace yav
