#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#include <imagedata/TclImage3DViewer.h>
#include <inrimage/TclImage3DBackgroundManager.h>


namespace yav {


TclImage3DViewer::TclImage3DViewer(Image3DBackgroundManager *manager,TclScene3D *scene)
  : TclScene3DObject(typeid(TclImage3DViewer), new Image3DViewer(scene->getClient(),manager), scene->getModule()) 
{
	done(typeid(TclImage3DViewer),"imagedata::installViewer3DMenu", 
       "imagedata::uninstallViewer3DMenu");
}






TclImage3DViewer::~TclImage3DViewer() {
  deleted(typeid(TclImage3DViewer));
}


int TclImage3DViewer::destroyCmd(int argc, char **argv) {
  if(argc > 0) return COMMAND_SYNOPSIS;
  //unsigned int ref = getRef(); // commented by fpoyer. reason : unused.
  TclModule *m = getModule();
  TclString n=name();
  delete this;
  m->setResult("%s does not exist any longer",(char *) n);
  return TCL_OK;  
}

int TclImage3DViewer::boxCmd(int argc, char **argv) {
  int i;
  float r, g, b;
  TclString res;

  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "color")) {
      if(i+3 >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[++i], "%f", &r) != 1) return COMMAND_SYNOPSIS;
      if(sscanf(argv[++i], "%f", &g) != 1) return COMMAND_SYNOPSIS;
      if(sscanf(argv[++i], "%f", &b) != 1) return COMMAND_SYNOPSIS;
      getClient()->setBoxColor(r, g, b);
    }
    else if(!strcmp(argv[i], "width")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%f", &r) != 1) return COMMAND_SYNOPSIS;
      getClient()->setBoxWidth(r);
    }
    else return COMMAND_SYNOPSIS;
  }

  getClient()->getBoxColor(r,g,b);
  res.sprintf("%f %f %f %f", r,g,b, getClient()->getBoxWidth());
  setResult(res);
  return TCL_OK;
}


int TclImage3DViewer::gridCmd(int argc, char **argv) {
  int i;
  float r, g, b;
  TclString res, p;

  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[0], "frame"))
      getClient()->setGridDisplay(Image3DViewer::FRAME);
    else if(!strcmp(argv[0], "grid"))
      getClient()->setGridDisplay(Image3DViewer::GRID);
    else if(!strcmp(argv[0], "hide"))
      getClient()->setGridDisplay(Image3DViewer::HIDE);
    else if(!strcmp(argv[0], "voxel"))
      getClient()->setGridDisplay(Image3DViewer::VOXELS);
    else if(!strcmp(argv[i], "color")) {
      if(i+3 >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[++i], "%f", &r) != 1) return COMMAND_SYNOPSIS;
      if(sscanf(argv[++i], "%f", &g) != 1) return COMMAND_SYNOPSIS;
      if(sscanf(argv[++i], "%f", &b) != 1) return COMMAND_SYNOPSIS;
      getClient()->setGridColor(r, g, b);
    }
    else if(!strcmp(argv[i], "width")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%f", &r) != 1) return COMMAND_SYNOPSIS;
      getClient()->setGridWidth(r);
    }
    else return COMMAND_SYNOPSIS;
  }

  switch(getClient()->getGridDisplay()) {
  case Image3DViewer::FRAME:
    res = "frame";
    break;
  case Image3DViewer::HIDE:
    res = "hide";
    break;
  case Image3DViewer::GRID:
    res = "grid";
    break;
  case Image3DViewer::VOXELS:
    res = "voxel";
    break;
  }
  getClient()->getGridColor(r,g,b);
  p.sprintf(" %f %f %f %f", r,g,b, getClient()->getGridWidth());
  setResult(res + p);
  return TCL_OK;
}






int TclImage3DViewer::create(TclModule *m, int argc, char **argv) {
  TclImage3DViewer *viewer;
  TclImage3DBackgroundManager *manager;
  TclScene3D *scene = 0;

  if(argc != 2 && argc != 4) return COMMAND_SYNOPSIS;
  if(argc == 4) {
    if(strcmp(argv[2], "-scene")) return COMMAND_SYNOPSIS;
    scene = (TclScene3D *) TclObjectKind::getObject(typeid(TclScene3D), argc, argv, 3);
  } else {
    scene = TclScene3D::getDefaultScene(m);
  }
  if(!scene) return TCL_ERROR;

 manager =(TclImage3DBackgroundManager *) TclObjectKind::getObject(typeid(TclImage3DBackgroundManager), argc, argv, 1);

  if(manager) {
     viewer = new TclImage3DViewer((Image3DBackgroundManager *)manager, scene);
     m->setResult(viewer->name());
	return TCL_OK;
  } 
  else 
  {
	  m->setResult("%s is not a background manager of 3D images",argv[2]);
	  return TCL_ERROR;
  }
   
 
}


/// creates the Tcl commands when the module is loaded
void  TclImage3DViewer::createCommands(TclModule *m)
{
	TclObjectKind *kind;	
  kind = new Scene3DItem(m, typeid(TclImage3DViewer), "viewImg3D","3D Display of an Image Slicer",
			 &TclImage3DViewer::create, &TclObject::list,
			 TclScene3D::SCRIPT, 
			 "multiview%u [-scene scn3D%u]");
  m->newCommand("viewImg3D%u", kind->getOptions(),
		"3D image viewer management command");

 ((Scene3DItem *)kind)->setCreationScript("imagedata::createViewer3D");
  kind->newOption("-destroy", (TclObject::OPTION) &TclImage3DViewer::destroyCmd, 0,
		  "destroy this viewer");
  kind->newOption("-box", (TclObject::OPTION) &TclImage3DViewer::boxCmd,
		  "[width %f] [color %f %f %f]",
		  "get/set bounding box line widht and color");
  kind->newOption("-grid", (TclObject::OPTION) &TclImage3DViewer::gridCmd,
		  "[frame|grid|voxel] [width %f] [color %f %f %f]",
		  "get/set grid display mode, width and color");
 TclScene3DObject::createOptionCommands(kind); 
}



} // namespace yav

