#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "imagedata/TclImageData.h"
#include <graphics/SceneItem.h>
#include <graphics3D/TclScene3D.h>
#include <graphics2D/SliceScale.h>
#include <imagedata/TclImageData.h>
#include <imagedata/ImageLine.h>


namespace yav {


TclImageData::TclImageData(TclScene3D *scene, const std::type_info& id,
			   ImageData *dimg)
  : TclScene3DObject(id, dimg, scene->getModule()) {
}


TclImageData::TclImageData(TclInrimage *inr, TclScene3D *scene,
			   const std::type_info& id, ImageData *dimg)
  : TclScene3DObject(id, dimg ? dimg : new ImageData(inr->getClient(),
						     scene->getClient()),
		     scene->getModule()) {
  done(typeid(TclImageData),"imagedata::installMenu", 
       "imagedata::uninstallMenu");
  if (getClient()->sliceMap.size()==0) {
    getModule()->tickle("view %s",(char *)inr->name());
  }
  getClient()->display(true);
}




TclImageData::~TclImageData() {
  deleted(typeid(TclImageData));
}


int TclImageData::destroyCmd(int argc, char **argv) {
  if(argc > 0) return COMMAND_SYNOPSIS;
  unsigned int ref = getRef();
  TclModule *m = getModule();

  delete this;
  m->setResult("%s%u does not exist any longer",
	       TclImageData::getTCLObjectName(), ref);
  return TCL_OK;  
}
int TclImageData::showBlackPixelsCmd(int argc, char **argv) {
  if (argc==0)
    {
      if (getClient()->AreBlackPixelsShown())
	setResult("true");
      else
	setResult("false");
      return TCL_OK;
    }
  else if (argc==1)
    {
      bool show;
      if(!TclModule::boolValue(argv[0], show)) return COMMAND_SYNOPSIS;
      getClient()->setShowBlackPixels(show);
      if (show)
	setResult("true");
      else
	setResult("false");
      return TCL_OK;
    }
  else
    return COMMAND_SYNOPSIS;
}

int TclImageData::boxCmd(int argc, char **argv) {
  int i;
  float r, g, b;
  TclString res;

  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "color")) {
      if(i+3 >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[++i], "%f", &r) != 1) return COMMAND_SYNOPSIS;
      if(sscanf(argv[++i], "%f", &g) != 1) return COMMAND_SYNOPSIS;
      if(sscanf(argv[++i], "%f", &b) != 1) return COMMAND_SYNOPSIS;
      getClient()->setBoxColor(r, g, b);
    }
    else if(!strcmp(argv[i], "width")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%f", &r) != 1) return COMMAND_SYNOPSIS;
      getClient()->setBoxWidth(r);
    }
    else return COMMAND_SYNOPSIS;
  }

  res.sprintf("%f %f %f %f", getClient()->boxColor[0], getClient()->boxColor[1],
	      getClient()->boxColor[2], getClient()->boxWidth);
  setResult(res);
  return TCL_OK;
}


int TclImageData::gridCmd(int argc, char **argv) {
  int i;
  float r, g, b;
  TclString res, p;

  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[0], "frame"))
      getClient()->setGridDisplay(ImageData::FRAME);
    else if(!strcmp(argv[0], "grid"))
      getClient()->setGridDisplay(ImageData::GRID);
    else if(!strcmp(argv[0], "hide"))
      getClient()->setGridDisplay(ImageData::HIDE);
    else if(!strcmp(argv[0], "voxel"))
      getClient()->setGridDisplay(ImageData::VOXELS);
    else if(!strcmp(argv[i], "color")) {
      if(i+3 >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[++i], "%f", &r) != 1) return COMMAND_SYNOPSIS;
      if(sscanf(argv[++i], "%f", &g) != 1) return COMMAND_SYNOPSIS;
      if(sscanf(argv[++i], "%f", &b) != 1) return COMMAND_SYNOPSIS;
      getClient()->setGridColor(r, g, b);
    }
    else if(!strcmp(argv[i], "width")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%f", &r) != 1) return COMMAND_SYNOPSIS;
      getClient()->setGridWidth(r);
    }
    else return COMMAND_SYNOPSIS;
  }

  switch(getClient()->gridDisplay) {
  case ImageData::FRAME:
    res = "frame";
    break;
  case ImageData::HIDE:
    res = "hide";
    break;
  case ImageData::GRID:
    res = "grid";
    break;
  case ImageData::VOXELS:
    res = "voxel";
    break;
  }
  p.sprintf(" %f %f %f %f", getClient()->gridColor[0], getClient()->gridColor[1],
	    getClient()->gridColor[2], getClient()->gridWidth);
  setResult(res + p);
  return TCL_OK;
}


int TclImageData::renderCmd(int argc, char **argv) {
  if(argc > 1) return COMMAND_SYNOPSIS;
  if(argc == 1) {
    if(!strcmp(argv[0], "none")) {
      if(getClient()->slicesDisplay == ImageData::VOLUME) tickle("destroy .volren");
      getClient()->setSlicesDisplay(ImageData::NOSLICES);
    }
    else if(!strcmp(argv[0], "solid")) {
      if(getClient()->slicesDisplay == ImageData::VOLUME) tickle("destroy .volren");
      getClient()->setSlicesDisplay(ImageData::SOLID);
    }
    else if(!strcmp(argv[0], "texture")) {
      if(getClient()->slicesDisplay == ImageData::VOLUME) tickle("destroy .volren");
      getClient()->setSlicesDisplay(ImageData::TEXTURE);
    }
    else if(!strcmp(argv[0], "volume")) {
      getClient()->textbg[3] = 0.1f;
      tickle("imagedata::volumeRendering %s%u",
	     TclObjectKind::getObjectName(typeid(*this)), getRef());
      getClient()->setSlicesDisplay(ImageData::VOLUME);
    }
    else return COMMAND_SYNOPSIS;
  }
  switch(getClient()->slicesDisplay) {
  case ImageData::NOSLICES:
    setResult("none");
    break;
  case ImageData::SOLID:
    setResult("solid");
    break;
  case ImageData::TEXTURE:
    setResult("texture");
    break;
  case ImageData::VOLUME:
    setResult("volume");
    break;
  }
  return TCL_OK;
}


int TclImageData::displayCmd(int argc, char **argv) {
  if(argc > 1) return COMMAND_SYNOPSIS;
  if(argc == 1) {
    if(!strcmp(argv[0], "true"))
      getClient()->display(true);
    else if(!strcmp(argv[0], "false"))
      getClient()->display(false);
    else return COMMAND_SYNOPSIS;
  }

  switch(getClient()->showImage) {
  case true:
    setResult("true");
    break;
  case false:
    setResult("false");
    break;
  }
  return TCL_OK;
}


int TclImageData::textureCmd(int argc, char **argv) {
  TclString res, bg;
  int i;

  if(argc > 0) {
    for(i = 0; i < argc; i++) {
      if(!strcmp(argv[i], "mapping")) {
	if(++i >= argc) return COMMAND_SYNOPSIS;
	if(!strcmp(argv[i], "decal")) getClient()->textmap = GL_DECAL;
	else if(!strcmp(argv[i], "modulate")) getClient()->textmap = GL_MODULATE;
	else if(!strcmp(argv[i], "blend")) getClient()->textmap = GL_BLEND;
	else return COMMAND_SYNOPSIS;
      }
      else if(!strcmp(argv[i], "bg")) {
	if(i + 4 >= argc) return COMMAND_SYNOPSIS;
	if(sscanf(argv[++i], "%f", &getClient()->textbg[0]) != 1)
	  return COMMAND_SYNOPSIS;
	if(sscanf(argv[++i], "%f", &getClient()->textbg[1]) != 1)
	  return COMMAND_SYNOPSIS;
	if(sscanf(argv[++i], "%f", &getClient()->textbg[2]) != 1)
	  return COMMAND_SYNOPSIS;
	if(sscanf(argv[++i], "%f", &getClient()->textbg[3]) != 1)
	  return COMMAND_SYNOPSIS;
      }
      else return COMMAND_SYNOPSIS;
    }
    if(getClient()->slicesDisplay == ImageData::TEXTURE ||
       getClient()->slicesDisplay == ImageData::VOLUME) getClient()->clearSliceLists();
  }

  if(getClient()->textmap == GL_DECAL) res = "decal";
  else if(getClient()->textmap == GL_BLEND) res = "blend";
  else if(getClient()->textmap == GL_MODULATE) res = "modulate";
  bg.sprintf(" %f %f %f %f", getClient()->textbg[0], getClient()->textbg[1],
	     getClient()->textbg[2], getClient()->textbg[3]);
  setResult(res + bg);
  return TCL_OK;
}


int TclImageData::resolutionCmd(int argc, char **argv) {
  unsigned int res;

  if(argc > 1) return COMMAND_SYNOPSIS;
  if(argc == 1) {
    if(sscanf(argv[0], "%u", &res) != 1) return COMMAND_SYNOPSIS;
    getClient()->setResolution(res);
  }
  setResult("%u", getClient()->resolution);
  return TCL_OK;
}


int TclImageData::transparencyCmd(int argc, char **argv) {
  double a;

  if(argc > 2) return COMMAND_SYNOPSIS;
  if(argc > 0) {
    bool blend;
    if(!TclModule::boolValue(argv[0], blend)) return COMMAND_SYNOPSIS;
    if(blend) {
      if(argc > 1) {
	if(sscanf(argv[1], "%lf", &a) != 1) return COMMAND_SYNOPSIS;
      }
      else a = getClient()->alpha;
    }
    else if(argc > 1) return COMMAND_SYNOPSIS;
    getClient()->setBlending(blend, a);
  }

  if(getClient()->blending) setResult("true %f", getClient()->alpha);
  else setResult("false", getClient()->alpha);
  return TCL_OK;
}


int TclImageData::create(TclModule *m, int argc, char **argv) {
  TclImageData *img;
  TclInrimage *inr;
  TclScene3D *scene = 0;

  if(argc != 2 && argc != 4) return COMMAND_SYNOPSIS;
  if(argc == 4) {
    if(strcmp(argv[2], "-scene")) return COMMAND_SYNOPSIS;
    scene = (TclScene3D *) TclObjectKind::getObject(typeid(TclScene3D), argc, argv, 3);
  } else {
    scene = TclScene3D::getDefaultScene(m);
  }
  if(!scene) return TCL_ERROR;

  inr = TclInrimage::getObject(argc, argv, 1);
  if(inr) {
    img = new TclImageData(inr, scene);
  }
  else {
    try {
      img = new TclImageData(new TclInrimage(m, argv[1]), scene);
    }
    catch(Inrimage::ImageNotFound&) {
      m->setResult("file not found");
      return TCL_ERROR;
    }
    catch(Inrimage::UnknownImageType&) {
      m->setResult("unknown image type");
      return TCL_ERROR;
    }
    catch(Inrimage::InvalidHeader&) {
      m->setResult("invalid inrimage header");
      return TCL_ERROR;
    }
    catch(Inrimage::InvalidData&) {
      m->setResult("invalid inrimage data");
      return TCL_ERROR;
    }
    catch(Inrimage::UnknownType&) {
      m->setResult("non supported inrimage type");
      return TCL_ERROR;
    }
  }
   
  m->setResult("%s%u", TclImageData::getTCLObjectName(), img->getRef());
  return TCL_OK;
}


int TclImageData::sceneCmd(int argc, char **argv) 
{
  if (argc==0)
    {
      setResult("%s%u", TclObjectKind::getObjectName(typeid(TclScene3D)),
		getTclScene()->getRef());
      return TCL_OK;
    }
    else
    return COMMAND_SYNOPSIS;
}


void TclImageData::load(const TclString &filename, Scene3D *scn) {
  TclInrimage *inr = new TclInrimage(((TclScene3D *)scn->getInterface())->getModule(),
				     filename);
  (void) new TclImageData(inr, (TclScene3D *)scn->getInterface());
}

int TclImageData::inrimageCmd(int argc, char **argv) {
  if (argc!=0) 
    return COMMAND_SYNOPSIS;
  if (getClient()->getImage())
    {
      if (TclInrimage::getInterface(getClient()->getImage())==0)
	new TclInrimage(getModule(), typeid(TclInrimage), getClient()->getImage() );
      
      setResult("%s%u",TclInrimage::getTCLObjectName(),
		TclInrimage::getInterface(getClient()->getImage())->getRef());
      return TCL_OK;
    }
  else {
    setResult("none");
    return TCL_OK;
  }
}

int TclImageData::volumeCmd(int argc, char **argv) {
  unsigned int ns;
  ImageData::BLENDING vb;
  TclString res;
  int i, j, k;
  double *tr;
  float a;
#if TK_MAJOR_VERSION==8 && TK_MINOR_VERSION==0
  Tcl_Obj *rarray, *garray, *barray, *aarray, *nb;
#else
  char nb[5];
#endif

  for(i = 0; i < argc; i++) {

    if(!strcmp(argv[i], "nslices")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%u", &ns) != 1) return COMMAND_SYNOPSIS;
      getClient()->setNSlices(ns);
    }
    else if(!strcmp(argv[i], "2D")) {
      getClient()->setVolumeRenderingType(ImageData::VR_2D);
      getClient()->redrawScene();
    }
    else if(!strcmp(argv[i], "3D")) {
      getClient()->setVolumeRenderingType(ImageData::VR_3D);
      getClient()->redrawScene();
    }
    else if(!strcmp(argv[i], "minGreyValue")) {
      double ming;
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%lf", &ming) != 1) return COMMAND_SYNOPSIS;
      getClient()->setMinGreyValue(ming);
    }
    else if(!strcmp(argv[i], "maxGreyValue")) {
      double maxg;
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%lf", &maxg) != 1) return COMMAND_SYNOPSIS;
      getClient()->setMaxGreyValue(maxg);
    }
    else if(!strcmp(argv[i], "blending")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(!strcmp(argv[i], "over")) vb = ImageData::OVER;
      else if(!strcmp(argv[i], "attenuate")) vb = ImageData::ATTENUATE;
      else if(!strcmp(argv[i], "MIP")) vb = ImageData::MIP;
      else if(!strcmp(argv[i], "mip")) vb = ImageData::MIP;
      else if(!strcmp(argv[i], "under")) vb = ImageData::UNDER;
      else return COMMAND_SYNOPSIS;
      getClient()->setVolumeBlending(vb);
    }
    else if(!strcmp(argv[i], "lookupTable")) {
      if (i + 2 >= argc) return COMMAND_SYNOPSIS;
      if(!strcmp(argv[i+1], "read")) {
	getClient()->readLookupTable(TclString(argv[i+2]));
	char *str="{#%02x%02x%02x}";
	const float *table=getClient()->getLookupTable();
	for(int i=0;i<256;++i) {
	  tickle("set imagedata::cline(%u) [.volren.c create line %u 300 %u 330 -fill [ format %s %u %u %u ] ]",i,60+i,60+i,str,(unsigned int)(255*table[4*i]),
		 (unsigned int)(255*table[4*i+1]),
		 (unsigned int)(255*table[4*i+2]));
	  tickle("set imagedata::red(%u) %f; set imagedata::green(%u) %f;set imagedata::blue(%u) %f; set imagedata::alpha(%u) %f  ",i,table[4*i],i,table[4*i+1],i,table[4*i+2],table[4*i+3]);
	  tickle(".volren.c coords $imagedata::vline(%u)  %u 270 %u %u ",i,60+i,60+i,(unsigned int)(270-255*table[4*i+3]));
	}
	
      }
      else if(!strcmp(argv[i+1], "write")) {
	getClient()->writeLookupTable(TclString(argv[i+2]));
      }
      else {
	return COMMAND_SYNOPSIS;
      }
      i+=3;
    }
    else if(!strcmp(argv[i], "transparency")) {
      if(i + 5 >= argc) return COMMAND_SYNOPSIS;
      Tcl_Obj *obj;
      tr = new double[4 * 256];
      for(j = 0, k = 0; j < 256; j++) {

	sprintf(nb, "%i", j);
	obj = Tcl_GetVar2Ex(getModule()->getInterp(), argv[i + 1], nb,
			     TCL_NAMESPACE_ONLY);
	if(!obj) goto freeandexit;
	if(Tcl_GetDoubleFromObj(getModule()->getInterp(), obj, &tr[k++])
	   == TCL_ERROR) goto freeandexit;
	obj = Tcl_GetVar2Ex(getModule()->getInterp(), argv[i + 2], nb,
			     TCL_NAMESPACE_ONLY);
	if(!obj) goto freeandexit;
	if(Tcl_GetDoubleFromObj(getModule()->getInterp(), obj, &tr[k++])
	   == TCL_ERROR) goto freeandexit;
	obj = Tcl_GetVar2Ex(getModule()->getInterp(), argv[i + 3], nb,
			     TCL_NAMESPACE_ONLY);
	if(!obj) goto freeandexit;
	if(Tcl_GetDoubleFromObj(getModule()->getInterp(), obj, &tr[k++])
	   == TCL_ERROR) goto freeandexit;
	obj = Tcl_GetVar2Ex(getModule()->getInterp(), argv[i + 4], nb,
			     TCL_NAMESPACE_ONLY);
	if(!obj) goto freeandexit;
	if(Tcl_GetDoubleFromObj(getModule()->getInterp(), obj, &tr[k++])
	   == TCL_ERROR) goto freeandexit;

      }
      if(sscanf(argv[i + 5], "%f", &a) != 1) goto freeandexit;
      getClient()->setVolumeTransparency(tr, a);
      delete[] tr;
      i += 5;
    }

    else return COMMAND_SYNOPSIS;
  }

  res.sprintf("%u ", getClient()->nslices);
  switch(getClient()->volumeBlending) {
  case ImageData::OVER:
    res += "over";
    break;
  case ImageData::ATTENUATE:
    res += "attenuate";
    break;
  case ImageData::MIP:
    res += "MIP";
    break;
  case ImageData::UNDER:
    res += "under";
    break;
  }
  setResult(res);
  return TCL_OK;

 freeandexit:
#if TK_MAJOR_VERSION==8 && TK_MINOR_VERSION==0
  Tcl_DecrRefCount(rarray);
  Tcl_DecrRefCount(garray);
  Tcl_DecrRefCount(barray);
  Tcl_DecrRefCount(aarray);
  Tcl_DecrRefCount(nb);
#endif
  delete[] tr;
  return COMMAND_SYNOPSIS;
}


} // namespace yav

