#ifdef _MSC_VER
#pragma warning ( disable : 4068 )
#pragma warning ( disable : 4786 )
#pragma warning ( disable : 4081 )
#endif

#pragma implementation "imagedata/TclIntensityProfile.h"

#include <inrimage/TclInrimage.h>
#include <imagedata/TclIntensityProfile.h>
#include <modules/TclFuzzyCMeans.h>
#include <modules/TclExpectationMaximization.h>

namespace yav {

int TclIntensityProfile::create(TclModule * m, int argc, char ** argv)
{
	if ((argc!=2) && (argc!=9))
		return COMMAND_SYNOPSIS;

	TclIntensityProfile * til, * til2;
	TclInrimage * tinr;
	Vec3<double> p, d;
	double dist;

	if (argc==2)
	{
		til = (TclIntensityProfile *) TclObjectKind::getObject(typeid(TclIntensityProfile), argc, argv, 1);

		if (til==0)
			return COMMAND_SYNOPSIS;
		else
		{
			til2 = new TclIntensityProfile(m,til->il);
			m->setResult("%s",(char *)til2->name());
			return TCL_OK;
		}
	}
	else 
	{
		tinr = (TclInrimage *) TclObjectKind::getObject(typeid(TclInrimage), argc,argv,1);

		if (tinr==0)
			return COMMAND_SYNOPSIS;
		else
		{
			p=Vec3<double>((double) atof(argv[2]), (double) atof(argv[3]), (double) atof(argv[4]));
			d=Vec3<double>((double) atof(argv[5]), (double) atof(argv[6]), (double) atof(argv[7]));
			dist = (double) atof(argv[8]);
			TclIntensityProfile * til = new TclIntensityProfile(m,tinr->getClient(),p,d,dist);
			m->setResult("%s",(char *)til->name());

			return TCL_OK;
		}
	}
}

int TclIntensityProfile::destroyCmd(int argc, char ** argv)
{
	if(argc!=0)
		return COMMAND_SYNOPSIS;

	TclModule * m = getModule();
	TclString n = name();
	delete this;
	m->setResult("%s", (char *) n);

	return TCL_OK;
}

int TclIntensityProfile::dumpCmd(int argc, char ** argv)
{
	if(argc != 0)
		return COMMAND_SYNOPSIS;
	
	int i;
	TclString val, str;
	str="{ ";

	for(i= -(int)il.negativeArraySize()+1; i<(int)il.positiveArraySize(); ++i)
	{
		if (i<0) 
			val.sprintf(" { %.5E %.5E } ",il.getDistance(i),il[i]->getValue());
		else
			val.sprintf(" { %.5E %.5E } ",il.getDistance(i),il[i]->getValue());

		str += val;
	}

	str += " } ";
	getModule()->setResult("%s",(char *) str);

	return TCL_OK;
}

int TclIntensityProfile::dumpGradientCmd(int argc, char ** argv)
{
	if(argc != 1)
		return COMMAND_SYNOPSIS;

	int i;
	double sigma;
	TclString str, val;

	sigma = atof(argv[0]);
	il.computeGradient(sigma);
	str="{ ";
	
	for(i= -(int)il.negativeArraySize()+1; i<(int)il.positiveArraySize(); ++i)
	{
		if (i<0) 
			val.sprintf(" { %.2E %.2E } ",il.getDistance(i),il.getGradient(i));
		else
			val.sprintf(" { %.2E %.2E } ",il.getDistance(i),il.getGradient(i));

		str += val;
	}

	str += " } ";
	getModule()->setResult("%s",(char *) str);

	return TCL_OK;
}

int TclIntensityProfile::drawCmd(int argc, char ** argv)
{
	TclModule * module = getModule();

	if (argc==1)
		module->tickle("imagedata::displayIntensityProfile %s %s", (char *)name(), argv[0]);
	else if (argc==0)
		module->tickle("imagedata::displayIntensityProfile %s none", (char *)name());
	else
		return COMMAND_SYNOPSIS;

	return TCL_OK;
}
// Intensity profiles using voxels
TclIntensityProfileSet::TclIntensityProfileSet (TclModule * m, Inrimage * inr, const std::vector< Vec3<double> > &pos, const std::vector< Vec3<double> > &dir,const double dist, const double step) : TclObject(typeid(TclIntensityProfileSet),m), ils(inr,pos,dir,dist,step)
{ 
	done(typeid(TclIntensityProfileSet));
}

TclIntensityProfileSet::TclIntensityProfileSet(TclModule * m) : TclObject(typeid(TclIntensityProfileSet),m) {
	done(typeid(TclIntensityProfileSet));
}

TclIntensityProfileSet::~TclIntensityProfileSet()
{}

int TclIntensityProfileSet::create(TclModule *m, int argc, char ** argv)
{
	if ((argc!=2) && (argc!=0))
		return COMMAND_SYNOPSIS;

	if (argc==2)
	{ 
		TclIntensityProfileSet * til = (TclIntensityProfileSet *) TclObjectKind::getObject(typeid(TclIntensityProfile), argc, argv, 1);

		if (til==0)
			return COMMAND_SYNOPSIS;
		else
		{
			TclIntensityProfileSet *til2 = new TclIntensityProfileSet(m,til->ils);
			m->setResult("%s",(char *)til2->name());
			return TCL_OK;
		}
    	}
	else 
	{
		TclInrimage * tinr = (TclInrimage *) TclObjectKind::getObject(typeid(TclInrimage), argc, argv, 1);

		if (tinr==0)
			return COMMAND_SYNOPSIS;

		return TCL_OK;
	}
}

int TclIntensityProfileSet::addIntensityProfileCmd(int argc, char ** argv)
{
	if(argc != 1)
		return COMMAND_SYNOPSIS;

	TclIntensityProfile * til = (TclIntensityProfile *) TclObjectKind::getObject(typeid(TclIntensityProfile), argc, argv, 1);

	if (til==0)
	{
		setResult("%s is not an image line", argv[0]);
		return TCL_ERROR;
	}

	IntensityProfile * i = new IntensityProfile(*til->getIntensityProfile());
	addIntensityProfile(i);
	setResult("%s added in the list", (char *) argv[0]);

	return TCL_OK;
}


int TclIntensityProfileSet::destroyCmd(int argc, char ** argv)
{
	TclModule * m;

	if(argc != 0)
		return COMMAND_SYNOPSIS;

	m = getModule();
	TclString n = name();
	delete this;
	m->setResult("%s", (char *) n);

	return TCL_OK;
}

int TclIntensityProfileSet::drawCmd(int argc, char ** argv)
{
	if (argc!=1)
		return COMMAND_SYNOPSIS;

	TclString inside = "inside";
	TclString outside = "outside";
	TclString bothsides = "bothsides";
	IntensityProfileSet::IMAGETYPE imagetype;

	if (inside.compare((TclString) argv[0])==0)
		imagetype = IntensityProfileSet::NEGATIVE_PROFILES;
	else if (outside.compare((TclString) argv[0])==0)
		imagetype = IntensityProfileSet::POSITIVE_PROFILES;
	else if (bothsides.compare((TclString) argv[0])==0)
		imagetype = IntensityProfileSet::ALL_PROFILES;
	else
		return COMMAND_SYNOPSIS;

	TclInrimage *tinr=new TclInrimage(getModule(),typeid(TclInrimage),ils.createIntensityProfileImage(imagetype));
	setResult("%s",(char *)tinr->name());

	return TCL_OK;
}

int TclIntensityProfileSet::fuzzyCMeansCmd(int argc, char **argv)
{
	if (argc!=2)
		return COMMAND_SYNOPSIS;

	unsigned int nbClusters = (unsigned int) atoi(argv[0]);
	FuzzyCMeans *fcm;
	TclString inside = "inside";
	TclString outside = "outside";
	TclString bothsides = "bothsides";
	IntensityProfileSet::IMAGETYPE imagetype;

	if (inside.compare((TclString) argv[1])==0)
		imagetype = IntensityProfileSet::NEGATIVE_PROFILES;
	else if (outside.compare((TclString) argv[1])==0)
		imagetype = IntensityProfileSet::POSITIVE_PROFILES;
	else if (bothsides.compare((TclString) argv[1])==0)
		imagetype = IntensityProfileSet::ALL_PROFILES;
	else
		return COMMAND_SYNOPSIS;

	fcm = ils.classifyByFuzzyCMeans(nbClusters,imagetype);
	TclInrimage *tinr = new TclInrimage(getModule(),typeid(TclInrimage), ils.sortFCMClassification(fcm));
	TclFuzzyCMeans *tfcm = new TclFuzzyCMeans(getModule(),fcm);
	setResult("Output image %s and Fuzzy C-Means %s" ,(char *)tinr->name(),(char *)tfcm->name() );

	return TCL_OK;
}

}
