#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "imagedata/TclRangeInrimage.h"
#include <modules/TclObject.h>
#include <imagedata/TclRangeInrimage.h>
#include <imagedata/TclIntensityProfile.h>
#include <inrimage/TclInrimage.h>
#include <modules/TclProbabilityDistribution.h>

namespace yav {

unsigned int TclRangeInrimage::nbIntensityProfileWindow = 0;

int TclRangeInrimage::forceCmd(int argc, char **argv,
			       RangeInrimage *client, TclString& result) {
  if(argc != 1) return COMMAND_SYNOPSIS;
  ObjectsList<RangeInrimage::Force>::Iterator it(client->forceList);
  if (!strcmp(argv[0], "list")) {
    result = EMPTY_STRING;
    for(; !it.isAtEnd();) {
      result += it()->getName();
      it++;
      if(!it.isAtEnd()) result += SPACE_STRING;
    }
    return TCL_OK;
  }
  else
    return COMMAND_SYNOPSIS;
}



int TclGradientBalloonInformation::create(TclModule *m, int argc, char **argv) {
  if ((argc!=4) && (argc!=2))
    return COMMAND_SYNOPSIS; 
  if (argc==2)
    {
      TclRangeInformation *ri= (TclRangeInformation *) TclObjectKind::getObject(typeid(TclRangeInformation), argc,argv,1);
      if (ri)
	{
	  GradientBalloonInrimageInformation *gi=dynamic_cast<GradientBalloonInrimageInformation *>(ri->getRangeInformation());
	  if (gi)
	    {
	      TclGradientBalloonInformation *tgi= new TclGradientBalloonInformation(m,gi);
	      m->setResult("%s",(char *)tgi->name());
	      return TCL_OK;
	    }
	  else
	    {
	      m->setResult("%s is not a gradient Balloon information",argv[1]);
	      return TCL_ERROR;
	    }
	}
      else return COMMAND_SYNOPSIS; 
    }
  else
    {
      double threshold=(double) atof(argv[1]);
      GradientBalloonInrimageInformation::OrientationConstraint constraint;
      if (!strcmp(argv[2],"direct"))
	constraint= GradientBalloonInrimageInformation::ALONG_GRADIENT;
      else   if (!strcmp(argv[2],"opposite"))
	constraint= GradientBalloonInrimageInformation::ALONG_OPPOSITE_GRADIENT;
      else   if (!strcmp(argv[2],"none"))
	constraint= GradientBalloonInrimageInformation::NONE;
      else return COMMAND_SYNOPSIS; 
      double force=(double) atof(argv[3]);
      GradientBalloonInrimageInformation *gi=new GradientBalloonInrimageInformation(threshold,constraint,force);
      
      TclGradientBalloonInformation *tgi= new TclGradientBalloonInformation(m,gi);
      m->setResult("%s",(char *)tgi->name());
      return TCL_OK;
    }
}
int TclGradientBalloonInformation::editCmd(int argc, char **argv) {

  char *directions[] = { "none", "direct", "opposite" };
  int i;

  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "threshold")) {
      if(++i >= argc) {
	  setResult("%f", getGradientBalloonInformation()->getThreshold());
	  return TCL_OK;
	}
	double r;
	if(sscanf(argv[i], "%lf", &r) != 1) return COMMAND_SYNOPSIS;
	getGradientBalloonInformation()->setThreshold(r);
    }
    else if(!strcmp(argv[i], "force")) {
      if(++i >= argc) {
	  setResult("%f", getGradientBalloonInformation()->getBalloonForce());
	  return TCL_OK;
	}
	double r;
	if(sscanf(argv[i], "%lf", &r) != 1) return COMMAND_SYNOPSIS;
	getGradientBalloonInformation()->setBalloonForce(r);
    }
    else if(!strcmp(argv[i], "orientation")) {
	if(++i >= argc) {
	  setResult(directions[getGradientBalloonInformation()->getOrientationConstraint()]);
	  return TCL_OK;
	}
	if(!strcmp(argv[i], directions[GradientBalloonInrimageInformation::NONE]))
	  getGradientBalloonInformation()->setOrientationConstraint(GradientBalloonInrimageInformation::NONE);
	else if(!strcmp(argv[i], directions[GradientBalloonInrimageInformation::ALONG_GRADIENT]))
	  getGradientBalloonInformation()->setOrientationConstraint(GradientBalloonInrimageInformation::ALONG_GRADIENT);
	else if(!strcmp(argv[i], directions[GradientBalloonInrimageInformation::ALONG_OPPOSITE_GRADIENT]))
	  getGradientBalloonInformation()->setOrientationConstraint(GradientBalloonInrimageInformation::ALONG_OPPOSITE_GRADIENT);
	else return COMMAND_SYNOPSIS;
    }
    else return COMMAND_SYNOPSIS;
  }
  
  setResult("orientation %s threshold %f force %f",
	    directions[getGradientBalloonInformation()->getOrientationConstraint()],
	    getGradientBalloonInformation()->getThreshold(),
	    getGradientBalloonInformation()->getBalloonForce());
  return TCL_OK;
}
int TclGradientGradientInformation::create(TclModule *m, int argc, char **argv) {
  if ((argc!=3) && (argc!=2))
    return COMMAND_SYNOPSIS; 
  if (argc==2)
    {
      TclRangeInformation *ri= (TclRangeInformation *) TclObjectKind::getObject(typeid(TclRangeInformation), argc,argv,1);
      if (ri)
	{
	  GradientGradientInrimageInformation *gi=dynamic_cast<GradientGradientInrimageInformation *>(ri->getRangeInformation());
	  if (gi)
	    {
	      TclGradientGradientInformation *tgi= new TclGradientGradientInformation(m,gi);
	      m->setResult("%s",(char *)tgi->name());
	      return TCL_OK;
	    }
	  else
	    {
	      m->setResult("%s is not a gradient of the gradient information",argv[1]);
	      return TCL_ERROR;
	    }
	}
      else return COMMAND_SYNOPSIS; 
    }
  else
    {
      double threshold=(double) atof(argv[1]);
      GradientGradientInrimageInformation::OrientationConstraint constraint;
      if (!strcmp(argv[2],"direct"))
	constraint= GradientGradientInrimageInformation::ALONG_GRADIENT;
      else   if (!strcmp(argv[2],"opposite"))
	constraint= GradientGradientInrimageInformation::ALONG_OPPOSITE_GRADIENT;
      else   if (!strcmp(argv[2],"none"))
	constraint= GradientGradientInrimageInformation::NONE;
      else return COMMAND_SYNOPSIS; 
      GradientGradientInrimageInformation *gi=new GradientGradientInrimageInformation(threshold,constraint);
      
      TclGradientGradientInformation *tgi= new TclGradientGradientInformation(m,gi);
      m->setResult("%s",(char *)tgi->name());
      return TCL_OK;
    }
}
int TclGradientGradientInformation::editCmd(int argc, char **argv) {

  char *directions[] = { "none", "direct", "opposite" };
  int i;

  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "threshold")) {
      if(++i >= argc) {
	  setResult("%f", getGradientGradientInformation()->getThreshold());
	  return TCL_OK;
	}
	double r;
	if(sscanf(argv[i], "%lf", &r) != 1) return COMMAND_SYNOPSIS;
	getGradientGradientInformation()->setThreshold(r);
    }
    else if(!strcmp(argv[i], "orientation")) {
	if(++i >= argc) {
	  setResult(directions[getGradientGradientInformation()->getOrientationConstraint()]);
	  return TCL_OK;
	}
	if(!strcmp(argv[i], directions[GradientGradientInrimageInformation::NONE]))
	  getGradientGradientInformation()->setOrientationConstraint(GradientGradientInrimageInformation::NONE);
	else if(!strcmp(argv[i], directions[GradientGradientInrimageInformation::ALONG_GRADIENT]))
	  getGradientGradientInformation()->setOrientationConstraint(GradientGradientInrimageInformation::ALONG_GRADIENT);
	else if(!strcmp(argv[i], directions[GradientGradientInrimageInformation::ALONG_OPPOSITE_GRADIENT]))
	  getGradientGradientInformation()->setOrientationConstraint(GradientGradientInrimageInformation::ALONG_OPPOSITE_GRADIENT);
	else return COMMAND_SYNOPSIS;
    }
    else return COMMAND_SYNOPSIS;
  }
  
  setResult("orientation %s threshold %f",
	    directions[getGradientGradientInformation()->getOrientationConstraint()],
	    getGradientGradientInformation()->getThreshold());
  return TCL_OK;
}
int TclRegionInformation::create(TclModule *m, int argc, char **argv) {
  if ((argc!=8) && (argc!=2))
    return COMMAND_SYNOPSIS; 
  if (argc==2)
    {
      TclRangeInformation *ri= (TclRangeInformation *) TclObjectKind::getObject(typeid(TclRangeInformation), argc,argv,1);
      if (ri)
	{
	  RegionInrimageInformation *gi=dynamic_cast<RegionInrimageInformation *>(ri->getRangeInformation());
	  if (gi)
	    {
	      TclRegionInformation *tgi= new TclRegionInformation(m,gi);
	      m->setResult("%s",(char *)tgi->name());
	      return TCL_OK;
	    }
	  else
	    {
	      m->setResult("%s is not a region information",argv[1]);
	      return TCL_ERROR;
	    }
	}
      else return COMMAND_SYNOPSIS; 
    }
  else
    {
      double range=(double) atof(argv[1]);
      double length=(double) atof(argv[2]);
      double threshold=(double) atof(argv[3]);
      RegionInrimageInformation::OrientationConstraint constraint;
      if (!strcmp(argv[4],"inside"))
	constraint= RegionInrimageInformation::INSIDE;
      else   if (!strcmp(argv[4],"outside"))
	constraint= RegionInrimageInformation::OUTSIDE;
      else return COMMAND_SYNOPSIS; 
      double min=(double) atof(argv[5]);
      double max=(double) atof(argv[6]);
      bool smooth;
      if(!TclModule::boolValue(argv[7],smooth)) return COMMAND_SYNOPSIS;
      
      RegionInrimageInformation *gi=new RegionInrimageInformation(range,threshold,constraint,length,min,max,smooth);
      
      TclRegionInformation *tgi= new TclRegionInformation(m,gi);
      m->setResult("%s",(char *)tgi->name());
      return TCL_OK;
    }
}
int TclRegionInformation::editCmd(int argc, char **argv) {
  char *directions[] = { "inside", "outside" };
  int i;

  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "range")) {
      if(++i >= argc) {
	setResult("%f", getRegionInformation()->getMaxDistance());
	return TCL_OK;
      }
      double r;
      if(sscanf(argv[i], "%lf", &r) != 1) return COMMAND_SYNOPSIS;
      getRegionInformation()->setMaxDistance(r);
    }
    else if(!strcmp(argv[i], "threshold")) {
      if(++i >= argc) {
	setResult("%f", getRegionInformation()->getThreshold());
	return TCL_OK;
      }
      double r;
      if(sscanf(argv[i], "%lf", &r) != 1) return COMMAND_SYNOPSIS;
      getRegionInformation()->setThreshold(r);
    }
    else if(!strcmp(argv[i], "length")) {
      if(++i >= argc) {
	setResult("%f", getRegionInformation()->getMinimumLength());
	return TCL_OK;
      }
      double r;
      if(sscanf(argv[i], "%lf", &r) != 1) return COMMAND_SYNOPSIS;
      getRegionInformation()->setMinimumLength(r);
    }
    else if(!strcmp(argv[i], "min")) {
      if(++i >= argc) {
	setResult("%f", getRegionInformation()->getMinimumIntensity());
	return TCL_OK;
      }
      double r;
      if(sscanf(argv[i], "%lf", &r) != 1) return COMMAND_SYNOPSIS;
      getRegionInformation()->setMinimumIntensity(r);
    }
    else if(!strcmp(argv[i], "max")) {
      if(++i >= argc) {
	setResult("%f", getRegionInformation()->getMaximumIntensity());
	return TCL_OK;
      }
      double r;
      if(sscanf(argv[i], "%lf", &r) != 1) return COMMAND_SYNOPSIS;
      getRegionInformation()->setMaximumIntensity(r);
    }
    else if(!strcmp(argv[i], "orientation")) {
      if(++i >= argc) {
	setResult(directions[getRegionInformation()->getOrientationConstraint()]);
	return TCL_OK;
      }
      if(!strcmp(argv[i], directions[RegionInrimageInformation::INSIDE]))
	getRegionInformation()->setOrientationConstraint(RegionInrimageInformation::INSIDE);
      else if(!strcmp(argv[i],
		      directions[RegionInrimageInformation::OUTSIDE]))
	  getRegionInformation()->setOrientationConstraint(RegionInrimageInformation::OUTSIDE);
      else return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "smooth")) {
      if(++i >= argc) {
	setResult(TclModule::boolString(getRegionInformation()->getSmoothLine()));
	return TCL_OK;
      }
      bool b;
      if(!TclModule::boolValue(argv[i], b)) return COMMAND_SYNOPSIS;
      getRegionInformation()->setSmoothLine(b);
    }
    else return COMMAND_SYNOPSIS;
  }

  setResult("range %f orientation %s threshold %f length %f min %f max %f smooth %s",
	      getRegionInformation()->getMaxDistance(),
	      directions[getRegionInformation()->getOrientationConstraint()],
	      getRegionInformation()->getThreshold(),
	      getRegionInformation()->getMinimumLength(),
	      getRegionInformation()->getMinimumIntensity(),
	      getRegionInformation()->getMaximumIntensity(),
	      TclModule::boolString(getRegionInformation()->getSmoothLine()));
    return TCL_OK;
}
int TclIsosurfaceInformation::create(TclModule *m, int argc, char **argv) {
  if ((argc!=4) && (argc!=2))
    return COMMAND_SYNOPSIS; 
  if (argc==2)
    {
      TclRangeInformation *ri= (TclRangeInformation *) TclObjectKind::getObject(typeid(TclRangeInformation), argc,argv,1);
      if (ri)
	{
	  IsosurfaceInrimageInformation *gi=dynamic_cast<IsosurfaceInrimageInformation *>(ri->getRangeInformation());
	  if (gi)
	    {
	      TclIsosurfaceInformation *tgi= new TclIsosurfaceInformation(m,gi);
	      m->setResult("%s",(char *)tgi->name());
	      return TCL_OK;
	    }
	  else
	    {
	      m->setResult("%s is not an isosurface information",argv[1]);
	      return TCL_ERROR;
	    }
	}
      else return COMMAND_SYNOPSIS; 
    }
  else
    {
		char *constraints[] = { "closest", "insideOver","insideBelow","outsideOver","outsideBelow" };
		double range=(double) atof(argv[1]);
		double isovalue=(double) atof(argv[2]);
		unsigned int i;
		for (i=0; (i<5) && (strcmp(argv[3],constraints[i]));++i);
		if (i==5) {
			m->setResult("%s is not an isosurface constraint",argv[3]);
			return TCL_ERROR;
		}

		IsosurfaceInrimageInformation *gi=new IsosurfaceInrimageInformation(isovalue,range,(IsosurfaceInrimageInformation::IsosurfaceConstraint) i);

		TclIsosurfaceInformation *tgi= new TclIsosurfaceInformation(m,gi);
		m->setResult("%s",(char *)tgi->name());
		return TCL_OK;
    }
}
int TclIsosurfaceInformation::editCmd(int argc, char **argv) {
	char *constraints[] = { "closest", "insideOver","insideBelow","outsideOver","outsideBelow" };
  int i;

  for(i = 0; i < argc; i++) {
	  if(!strcmp(argv[i], "range")) {
		  if(++i >= argc) {
			  setResult("%f", getIsosurfaceInformation()->getMaxDistance());
			  return TCL_OK;
		  }
		  double r;
		  if(sscanf(argv[i], "%lf", &r) != 1) return COMMAND_SYNOPSIS;
		  getIsosurfaceInformation()->setMaxDistance(r);
	  }
	  else if(!strcmp(argv[i], "isovalue")) {
		  if(++i >= argc) {
			  setResult("%f", getIsosurfaceInformation()->getIsovalue());
			  return TCL_OK;
		  }
		  double r;
		  if(sscanf(argv[i], "%lf", &r) != 1) return COMMAND_SYNOPSIS;
		  getIsosurfaceInformation()->setIsovalue(r);
	  }
	   else if(!strcmp(argv[i], "constraint")) {
		  if(++i >= argc) {
			  setResult("%s",constraints[ getIsosurfaceInformation()->getIsosurfaceConstraint()]);
			  return TCL_OK;
		  }
		  unsigned int j;
		  for (j=0; (j<5) && (strcmp(argv[i],constraints[j]));++j);
		  if (j==5) return COMMAND_SYNOPSIS;

		  getIsosurfaceInformation()->setIsosurfaceConstraint((IsosurfaceInrimageInformation::IsosurfaceConstraint)j);
	  }
    else return COMMAND_SYNOPSIS;
  }

  setResult("range %f isovalue %f constraint $s",
	      getIsosurfaceInformation()->getMaxDistance(),
	      getIsosurfaceInformation()->getIsovalue(),
		  constraints[ getIsosurfaceInformation()->getIsosurfaceConstraint()]);
    return TCL_OK;
}


int TclRegionProbabilityMapInformation::create(TclModule *m, int argc, char **argv) {
  if ((argc!=8) && (argc!=2))
    return COMMAND_SYNOPSIS; 
  if (argc==2)
    {
      TclRangeInformation *ri= (TclRangeInformation *) TclObjectKind::getObject(typeid(TclRangeInformation), argc,argv,1);
      if (ri)
	{
	  RegionProbabilityMapInformation *gi=dynamic_cast<RegionProbabilityMapInformation *>(ri->getRangeInformation());
	  if (gi)
	    {
	      TclRegionProbabilityMapInformation *tgi= new TclRegionProbabilityMapInformation(m,gi);
	      m->setResult("%s",(char *)tgi->name());
	      return TCL_OK;
	    }
	  else
	    {
	      m->setResult("%s is not a region probability map information",argv[1]);
	      return TCL_ERROR;
	    }
	}
      else return COMMAND_SYNOPSIS; 
    }
  else
    {
      double range=(double) atof(argv[1]);
      double length=(double) atof(argv[2]);
      RegionProbabilityMapInformation::OrientationConstraint constraint;
      if (!strcmp(argv[3],"inside"))
	constraint= RegionProbabilityMapInformation::INSIDE;
      else   if (!strcmp(argv[3],"outside"))
	constraint= RegionProbabilityMapInformation::OUTSIDE;
      else return COMMAND_SYNOPSIS;
	  TclProbabilityDistribution *pdfIn= (TclProbabilityDistribution *) TclObjectKind::getObject(typeid(TclProbabilityDistribution), argc,argv,4);
	  if (pdfIn==0) {
		  m->setResult("%s is not a probability density function",argv[4]);
		  return TCL_ERROR;
	  }
	  TclProbabilityDistribution *pdfOut= (TclProbabilityDistribution *) TclObjectKind::getObject(typeid(TclProbabilityDistribution), argc,argv,5);
	  if (pdfOut==0) {
		  m->setResult("%s is not a probability density function",argv[5]);
		  return TCL_ERROR;
	  }

      
      RegionProbabilityMapInformation *gi=new RegionProbabilityMapInformation(range,constraint,length,pdfIn->getProbabilityFunction(),
		  pdfOut->getProbabilityFunction());
      
      TclRegionProbabilityMapInformation *tgi= new TclRegionProbabilityMapInformation(m,gi);
      m->setResult("%s",(char *)tgi->name());
      return TCL_OK;
    }
}
int TclRegionProbabilityMapInformation::editCmd(int argc, char **argv) {
  char *directions[] = { "inside", "outside" };
  int i;

  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "range")) {
      if(++i >= argc) {
	setResult("%f", getRegionInformation()->getMaxDistance());
	return TCL_OK;
      }
      double r;
      if(sscanf(argv[i], "%lf", &r) != 1) return COMMAND_SYNOPSIS;
      getRegionInformation()->setMaxDistance(r);
    }
    else if(!strcmp(argv[i], "length")) {
      if(++i >= argc) {
	setResult("%f", getRegionInformation()->getMinimumLength());
	return TCL_OK;
      }
      double r;
      if(sscanf(argv[i], "%lf", &r) != 1) return COMMAND_SYNOPSIS;
      getRegionInformation()->setMinimumLength(r);
    }
    else if(!strcmp(argv[i], "pdfInside")) {
		if(++i >= argc) {
			if (getRegionInformation()->getInsidePDF()==0) {
				setResult("none");
				return TCL_OK;
			} else {
				TclProbabilityDistribution *pdfIn= new TclProbabilityDistribution(typeid(TclProbabilityDistribution),
					getModule(),getRegionInformation()->getInsidePDF());
				setResult(pdfIn->name());
				return TCL_OK;
			}
		}
	  TclProbabilityDistribution *pdfIn= (TclProbabilityDistribution *) TclObjectKind::getObject(typeid(TclProbabilityDistribution), argc,argv,i);
	  if (pdfIn==0) {
		  setResult("%s is not a probability density function",argv[i]);
		  return TCL_ERROR;
	  }

      getRegionInformation()->setInsidePDF(pdfIn->getProbabilityFunction());
    }
    else if(!strcmp(argv[i], "pdfOutside")) {
		if(++i >= argc) {
			if (getRegionInformation()->getOutsidePDF()==0) {
				setResult("none");
				return TCL_OK;
			} else {
				TclProbabilityDistribution *pdfOut= new TclProbabilityDistribution(typeid(TclProbabilityDistribution),
					getModule(),getRegionInformation()->getOutsidePDF());
				setResult(pdfOut->name());
				return TCL_OK;
			}
		}
   	  TclProbabilityDistribution *pdfOut= (TclProbabilityDistribution *) TclObjectKind::getObject(typeid(TclProbabilityDistribution), argc,argv,i);
	  if (pdfOut==0) {
		  setResult("%s is not a probability density function",argv[i]);
		  return TCL_ERROR;
	  }

      getRegionInformation()->setOutsidePDF(pdfOut->getProbabilityFunction());

    }
    else if(!strcmp(argv[i], "orientation")) {
      if(++i >= argc) {
	setResult(directions[getRegionInformation()->getOrientationConstraint()]);
	return TCL_OK;
      }
      if(!strcmp(argv[i], directions[RegionProbabilityMapInformation::INSIDE]))
	getRegionInformation()->setOrientationConstraint(RegionProbabilityMapInformation::INSIDE);
      else if(!strcmp(argv[i],
		      directions[RegionProbabilityMapInformation::OUTSIDE]))
	  getRegionInformation()->setOrientationConstraint(RegionProbabilityMapInformation::OUTSIDE);
      else return COMMAND_SYNOPSIS;
    }
   
    else return COMMAND_SYNOPSIS;
  }
  TclString inString;
  if (getRegionInformation()->getInsidePDF()==0) {
	  inString="none";
  } else {
	  TclProbabilityDistribution *pdfIn= new TclProbabilityDistribution(typeid(TclProbabilityDistribution),
		  getModule(),getRegionInformation()->getInsidePDF());
	  inString=pdfIn->name();
  }
  TclString outString;
  if (getRegionInformation()->getOutsidePDF()==0) {
	  outString="none";
  } else {
	  TclProbabilityDistribution *pdfOut= new TclProbabilityDistribution(typeid(TclProbabilityDistribution),
		  getModule(),getRegionInformation()->getOutsidePDF());
	  outString=pdfOut->name();
  }
  setResult("range %f orientation %s length %f pdfInside %f pdfOutside %f",
	      getRegionInformation()->getMaxDistance(),
	      directions[getRegionInformation()->getOrientationConstraint()],
	      getRegionInformation()->getMinimumLength(),(char *) inString,
		  (char *) outString);
    return TCL_OK;
}

int TclGradientInformation::create(TclModule *m, int argc, char **argv) {
  if ((argc!=8) && (argc!=2))
    return COMMAND_SYNOPSIS; 
  if (argc==2)
    {
      TclRangeInformation *ri= (TclRangeInformation *) TclObjectKind::getObject(typeid(TclRangeInformation), argc,argv,1);
      if (ri)
	{
	  GradientInrimageInformation *gi=dynamic_cast<GradientInrimageInformation *>(ri->getRangeInformation());
	  if (gi)
	    {
	      TclGradientInformation *tgi= new TclGradientInformation(m,gi);
	      m->setResult("%s",(char *)tgi->name());
	      return TCL_OK;
	    }
	  else
	    {
	      m->setResult("%s is not a gradient information",argv[1]);
	      return TCL_ERROR;
	    }
	}
      else return COMMAND_SYNOPSIS; 
    }
  else
    {
      double range=(double) atof(argv[1]);
      double threshold=(double) atof(argv[2]);
      GradientInrimageInformation::OrientationConstraint constraint;
      if (!strcmp(argv[3],"none"))
	constraint= GradientInrimageInformation::NONE;
      else   if (!strcmp(argv[3],"direct"))
	constraint= GradientInrimageInformation::ALONG_GRADIENT;
      else   if (!strcmp(argv[3],"opposite"))
	constraint= GradientInrimageInformation::ALONG_OPPOSITE_GRADIENT;
      else return COMMAND_SYNOPSIS; 
      bool _stopAtFirst,_useIntensity;
      if(!TclModule::boolValue(argv[4],_stopAtFirst)) return COMMAND_SYNOPSIS;
      if(!TclModule::boolValue(argv[5],_useIntensity)) return COMMAND_SYNOPSIS;
      double min=(double) atof(argv[6]);
      double max=(double) atof(argv[7]);
      
      GradientInrimageInformation *gi=new GradientInrimageInformation(range,threshold,constraint,_stopAtFirst,_useIntensity,min,max);
      
      TclGradientInformation *tgi= new TclGradientInformation(m,gi);
      m->setResult("%s",(char *)tgi->name());
      return TCL_OK;
    }
}
 
int TclGradientInformation::editCmd(int argc, char **argv) {

  char *directions[] = { "none", "direct", "opposite" };
  int i;

  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "range")) {
      if(++i >= argc) {
	setResult("%f", getGradientInformation()->getMaxDistance());
	return TCL_OK;
      }
      double r;
      if(sscanf(argv[i], "%lf", &r) != 1) return COMMAND_SYNOPSIS;
      getGradientInformation()->setMaxDistance(r);
    }
    else if(!strcmp(argv[i], "threshold")) {
      if(++i >= argc) {
	  setResult("%f", getGradientInformation()->getThreshold());
	  return TCL_OK;
	}
	double r;
	if(sscanf(argv[i], "%lf", &r) != 1) return COMMAND_SYNOPSIS;
	getGradientInformation()->setThreshold(r);
    }
    else if(!strcmp(argv[i], "min")) {
	if(++i >= argc) {
	  setResult("%f", getGradientInformation()->getMinimumIntensity());
	  return TCL_OK;
	}
	double r;
	if(sscanf(argv[i], "%lf", &r) != 1) return COMMAND_SYNOPSIS;
	getGradientInformation()->setMinimumIntensity(r);
      }
    else if(!strcmp(argv[i], "max")) {
	if(++i >= argc) {
	  setResult("%f", getGradientInformation()->getMaximumIntensity());
	  return TCL_OK;
	}
	double r;
	if(sscanf(argv[i], "%lf", &r) != 1) return COMMAND_SYNOPSIS;
	getGradientInformation()->setMaximumIntensity(r);
      }
    else if(!strcmp(argv[i], "intensity")) {
	if(++i >= argc) {
	  setResult(TclModule::boolString(getGradientInformation()->getIntensityConstraint()));
	  return TCL_OK;
	}
	bool b;
	if(!TclModule::boolValue(argv[i], b)) return COMMAND_SYNOPSIS;
	getGradientInformation()->setIntensityConstraint(b);
      }
    else if(!strcmp(argv[i], "orientation")) {
	if(++i >= argc) {
	  setResult(directions[getGradientInformation()->getOrientationConstraint()]);
	  return TCL_OK;
	}
	if(!strcmp(argv[i], directions[GradientInrimageInformation::NONE]))
	  getGradientInformation()->setOrientationConstraint(GradientInrimageInformation::NONE);
	else if(!strcmp(argv[i], directions[GradientInrimageInformation::ALONG_GRADIENT]))
	  getGradientInformation()->setOrientationConstraint(GradientInrimageInformation::ALONG_GRADIENT);
	else if(!strcmp(argv[i], directions[GradientInrimageInformation::ALONG_OPPOSITE_GRADIENT]))
	  getGradientInformation()->setOrientationConstraint(GradientInrimageInformation::ALONG_OPPOSITE_GRADIENT);
	else return COMMAND_SYNOPSIS;
      }
    else if(!strcmp(argv[i], "stop")) {
	if(++i >= argc) {
	  setResult(getGradientInformation()->getStopCondition() ? "first" : "highest");
	  return TCL_OK;
	}
	if(!strcmp(argv[i], "first"))
	  getGradientInformation()->setStopCondition(true);
	else if(!strcmp(argv[i], "highest"))
	  getGradientInformation()->setStopCondition(false);
	else return COMMAND_SYNOPSIS;
      }
    else return COMMAND_SYNOPSIS;
  }
  
  setResult("range %f orientation %s threshold %f stop %s intensity %s min %f max %f",
		 getGradientInformation()->getMaxDistance(),
		 directions[getGradientInformation()->getOrientationConstraint()],
		 getGradientInformation()->getThreshold(),
		 getGradientInformation()->getStopCondition() ? "first" : "highest",
		 TclModule::boolString(getGradientInformation()->getIntensityConstraint()),
		 getGradientInformation()->getMinimumIntensity(),
		 getGradientInformation()->getMaximumIntensity());
  return TCL_OK;
}


int TclClassInformation::create(TclModule *m, int argc, char **argv) {
  if (argc<2) 
    return COMMAND_SYNOPSIS; 
  if (argc==2)
    {
      TclRangeInformation *ri= (TclRangeInformation *) TclObjectKind::getObject(typeid(TclRangeInformation), argc,argv,1);
      if (ri)
	{
	  ClassInrimageInformation *gi=dynamic_cast<ClassInrimageInformation *>(ri->getRangeInformation());
	  if (gi)
	    {
	      TclClassInformation *tgi= new TclClassInformation(m,gi);
	      m->setResult("%s",(char *)tgi->name());
	      return TCL_OK;
	    }
	  else
	    {
	      m->setResult("%s is not a class information",argv[1]);
	      return TCL_ERROR;
	    }
	}
      else return COMMAND_SYNOPSIS; 
    }
  else
    {
      unsigned int nin,nout,i;
      double range=(double) atof(argv[1]);
      nin=(unsigned int) atof(argv[2]);
      unsigned char *in;
      if (nin)
	in=new unsigned char [nin];
      else
	in=0;
      if(argc< (int) (nin +3)) return COMMAND_SYNOPSIS;
      for(i=0;i<nin;++i)
	in[i]=atoi(argv[i+3]);
      nout=(unsigned int) atof(argv[nin+3]);
      unsigned char *out;
      if (nout)
	out=new unsigned char [nout];
      else
	out=0;
      if(argc< (int)(nout+ nin +4)) return COMMAND_SYNOPSIS;
      for(i=0;i<nout;++i)
	out[i]=atoi(argv[i+4+nin]);
      ClassInrimageInformation *ci=new ClassInrimageInformation(range,nin,in,nout,out);
      
      TclClassInformation *tci= new TclClassInformation(m,ci);
      m->setResult("%s",(char *)tci->name());
      return TCL_OK;
    }
}


int TclClassInformation::editCmd(int argc, char **argv) {
  int i, j;
  
  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "inclasses")) {
      unsigned int nin;
      if(++i >= argc) {
	char tmp[6];
	const unsigned char *in = getClassInformation()->getInArray(nin);
	TclString result;

	result.sprintf("%i", nin);
	for(j = 0; j < (int) nin; j++) {
	  sprintf(tmp, " %i", in[j]);
	  result += tmp;
	}
	setResult(result);
	return TCL_OK;
      }
      if(sscanf(argv[i], "%i", &nin) != 1) return COMMAND_SYNOPSIS;
      if(i + (int) nin >= argc) return COMMAND_SYNOPSIS;
      unsigned char *in = new unsigned char[nin];
      for(j = 0, i++; j < (int) nin; j++, i++) {
	int k;
	if(sscanf(argv[i], "%i", &k) != 1 || k < 0 || k > 255) {
	  delete[] in;
	  return COMMAND_SYNOPSIS;
	}
	else in[j] = (unsigned char) k;
      }
      i--;
      getClassInformation()->setInArray(nin, in);
      delete[] in;
    }
    
    else if(!strcmp(argv[i], "outclasses")) {
      unsigned int nout;
      if(++i >= argc) {
	char tmp[6];
	const unsigned char *out = getClassInformation()->getOutArray(nout);
	TclString result;
	
	result.sprintf("%i", nout);
	for(j = 0; j < (int) nout; j++) {
	  sprintf(tmp, " %i", out[j]);
	  result += tmp;
	}
	setResult(result);
	return TCL_OK;
      }
      if(sscanf(argv[i], "%i", &nout) != 1) return COMMAND_SYNOPSIS;
      if(i + (int) nout >= argc) return COMMAND_SYNOPSIS;
      unsigned char *out = new unsigned char[nout];
      for(j = 0, i++; j < (int) nout; j++, i++) {
	int k;
	if(sscanf(argv[i], "%i", &k) != 1 || k < 0 || k > 255) {
	  delete[] out;
	  return COMMAND_SYNOPSIS;
	}
	else out[j] = (unsigned char) k;
      }
      i--;
      getClassInformation()->setOutArray(nout, out);
      delete[] out;
    }
    
    else if(!strcmp(argv[i], "range")) {
      if(++i >= argc) {
	setResult("%f", getClassInformation()->getMaxDistance());
	return TCL_OK;
      }
      double dist;
      if(sscanf(argv[i], "%lf", &dist) != 1) return COMMAND_SYNOPSIS;
      getClassInformation()->setMaxDistance(dist);
    }
    else if(!strcmp(argv[i], "inlength")) {
      if(++i >= argc) {
	setResult("%f", getClassInformation()->getInlength());
	return TCL_OK;
      }
      double dist;
      if(sscanf(argv[i], "%lf", &dist) != 1) return COMMAND_SYNOPSIS;
      getClassInformation()->setInlength(dist);
    }
    else if(!strcmp(argv[i], "outlength")) {
      if(++i >= argc) {
	setResult("%f", getClassInformation()->getOutlength());
	return TCL_OK;
      }
      double dist;
      if(sscanf(argv[i], "%lf", &dist) != 1) return COMMAND_SYNOPSIS;
      getClassInformation()->setOutlength(dist);
    }
    else return COMMAND_SYNOPSIS;
  }
  
  
  unsigned int nin, nout;
  const unsigned char *in, *out;
  char tmp[20];
  TclString result;

  in = getClassInformation()->getInArray(nin);
  out = getClassInformation()->getOutArray(nout);
  
  result.sprintf("range %f", getClassInformation()->getMaxDistance());
  sprintf(tmp, " inclasses %i", nin);
  result += tmp;
  for(j = 0; j < (int) nin; j++) {
    sprintf(tmp, " %i", (int) in[j]);
    result += tmp;
  }
  sprintf(tmp, " outclasses %i", nout);
  result += tmp;
  for(j = 0; j < (int) nout; j++) {
    sprintf(tmp, " %i", (int) out[j]);
    result += tmp;
  }
  sprintf(tmp, " inlength %f", getClassInformation()->getInlength());
  result += tmp;
  sprintf(tmp, " outlength %f", getClassInformation()->getOutlength());
  result += tmp;

  setResult(result);
  return TCL_OK;
}


int TclRangeInrimage::imagesCmd(int argc, char **argv,
				RangeInrimage *rimg, TclString& result) {
  TclString res, one;
  TclInrimage *inr = 0, *grad = 0, *dir = 0, *edge = 0, *grad2 = 0, *clf = 0;
  int i;

  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "intensity")) {
      if(++i >= argc) 
	{
	  if (rimg->intensity)
	    result.sprintf("%s%u", TclInrimage::getTCLObjectName(),
			   getIntensity(rimg)->getRef());
	  else
	    result.sprintf("none");
	  return TCL_OK;
	}
      inr = TclInrimage::getObject(argc, argv, i);
      if(!inr) return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "gradient")) {
      if(++i >= argc) 
	{
	  if (rimg->gradient)
	    result.sprintf("%s%u", TclInrimage::getTCLObjectName(),
			   getGradient(rimg)->getRef());
	  else
	    result.sprintf("none");
	  return TCL_OK;
	}
      grad = TclInrimage::getObject(argc, argv, i);
      if(!grad) return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "gradientdir")) {
      if(++i >= argc)
	{
	  if (rimg->gradientDirection)
	    result.sprintf("%s%u", TclInrimage::getTCLObjectName(),
			   getGradientDirection(rimg)->getRef());
	  else
	    result.sprintf("none");
	  return TCL_OK;
	}
      dir = TclInrimage::getObject(argc, argv, i);
      if(!dir) return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "edge")) {
      if(++i >= argc) 
	{
	  if (rimg->edge)
	    result.sprintf("%s%u", TclInrimage::getTCLObjectName(),
			   getEdge(rimg)->getRef());
	  else
	    result.sprintf("none");
	  return TCL_OK;
	}
      edge = TclInrimage::getObject(argc, argv, i);
      if(!edge) return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "gradient2")) {
      if(++i >= argc) 
	{
	  if (rimg->gradientGradient)
	    result.sprintf("%s%u", TclInrimage::getTCLObjectName(),
			   getGradientGradient(rimg)->getRef());
	  else
	    result.sprintf("none");
	  return TCL_OK;
	}
      grad2 = TclInrimage::getObject(argc, argv, i);
      if(!grad2) return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "classify")) {
      if(++i >= argc) 
	{
	  if (rimg->classify)
	    result.sprintf("%s%u", TclInrimage::getTCLObjectName(),
			   getClassify(rimg)->getRef());
	  else
	    result.sprintf("none");
	  return TCL_OK;
	}
      clf = TclInrimage::getObject(argc, argv, i);
      if(!clf) return COMMAND_SYNOPSIS;
    }
    else return COMMAND_SYNOPSIS;
  }
  
  if(inr) {
    try {
      rimg->setIntensity(inr->getClient());
    }
    catch(RangeInrimage::InvalidSize&) {
      result.sprintf("invalid intensity image size");
      return TCL_ERROR;
    }
   
  }
  if(grad) {
    try {
      rimg->setGradient(grad->getClient());
    }
    catch(RangeInrimage::InvalidSize&) {
      result.sprintf("invalid gradient image size");
      return TCL_ERROR;
    }
  }
  if(dir) {
    try {
      rimg->setGradientDirection(dir->getClient());
    }
    catch(RangeInrimage::InvalidSize&) {
      result.sprintf("invalid gradient direction image size");
      return TCL_ERROR;
    }
  }
  if(edge) {
    try {
      rimg->setEdge(edge->getClient());
    }
    catch(RangeInrimage::InvalidSize&) {
      result.sprintf("invalid edge image size");
      return TCL_ERROR;
    }
    
  }
  if(grad2) {
    try {
      rimg->setGradientGradient(grad2->getClient());
    }
    catch(RangeInrimage::InvalidSize&) {
      result.sprintf("invalid gradient derivative image size");
      return TCL_ERROR;
    }
   
  }
  if(clf) {
    try {
      rimg->setClassify(clf->getClient());
    }
    catch(RangeInrimage::InvalidSize&) {
      result.sprintf("invalid classify image size");
      return TCL_ERROR;
    }
    
  }

  
  res.sprintf("intensity %s%u", TclInrimage::getTCLObjectName(),
	     getIntensity(rimg)->getRef());
  
  if(rimg->gradient) {
    one.sprintf(" gradient %s%u", TclInrimage::getTCLObjectName(),
		getGradient(rimg)->getRef());
    res += one;
  }
  if(rimg->gradientDirection) {
    one.sprintf(" gradientdir %s%u", TclInrimage::getTCLObjectName(),
		getGradientDirection(rimg)->getRef());
    res += one;
  }
  if(rimg->edge) {
    one.sprintf(" edge %s%u", TclInrimage::getTCLObjectName(),
		getEdge(rimg)->getRef());
    res += one;
  }
  if(rimg->gradientGradient) {
    one.sprintf(" gradient2 %s%u", TclInrimage::getTCLObjectName(),
		getGradientGradient(rimg)->getRef());
    res += one;
  }
  if(rimg->classify) {
    one.sprintf(" classify %s%u", TclInrimage::getTCLObjectName(),
		getClassify(rimg)->getRef());
    res += one;
  }
  result.sprintf(res);
  return TCL_OK;
}


int TclRangeInrimage::saveCmd(int argc, char **argv,
			      RangeInrimage *rimg, TclString& result) {
  if (argc!=1)
    return COMMAND_SYNOPSIS;
  else {
    try {
      rimg->write(argv[0]);
      result.sprintf("Range Inrimage saved as %s",argv[0]);
      return TCL_OK;
    }
    catch (...) {
      return TCL_ERROR;
    }
  }
}


void TclRangeInrimage::displayGraph(char *frameName, char *title,
				    IntensityProfile &line,
				    double width, double height,
				    TclModule *module) {
  double val;
  /// computes the min and max value of the distance and image value
  double minValue=line[0]->getValue();
  double maxValue=line[0]->getValue();
  double minDistance= -line[-(int)line.negativeArraySize()+1]->getDistance();
  double maxDistance= line[line.positiveArraySize()-1]->getDistance();
  int maximumIndex=line.positiveArraySize();
  int i;

  for(i= -(int)line.negativeArraySize()+1;i<maximumIndex;++i) {
      val=line[i]->getValue();
      if (val<minValue)
	minValue=val;
      if (val>maxValue)
	maxValue=val;
  }
 
  module->tickle("\
frame %s;\
label %s.l -font {Times 18} -justify center -text \"%s\";\
label %s.v -font {Times 16} -justify center -text \"Distance : 0.0 Value : 0.0\";\
pack  %s.l %s.v -side top -expand true -fill x",
		      frameName, frameName, title, frameName, frameName,
		      frameName);
  module->tickle("\
canvas %s.c -relief raise -width %f -height %f;\
pack %s.c -side top -fill x;\
pack %s",
		      frameName, width, height, frameName, frameName);
  double distanceStep = (width-3)/(maxDistance - minDistance);
  
  double valueStep;
  if (maxValue==minValue) valueStep=height;
  else valueStep=(height-3)/(maxValue - minValue);
  TclString lineCmd;
  lineCmd.sprintf("%s.c create line ",frameName);
  char pixel[100];
  for(i= -(int)line.negativeArraySize()+1; i<maximumIndex; ++i) {
    sprintf(pixel,"%f %f ", (line[i]->getDistance()*signd(i)-minDistance)*
	    distanceStep, height-((line[i]->getValue()-minValue)*valueStep));
    lineCmd+=pixel;
  }
  lineCmd+=" -width 2 -fill {#000000}";
  module->tickle(lineCmd);
  lineCmd.sprintf("%s.c create line %f %f %f %f -width 2 -fill {#FF0000} -arrow last -arrowshape {8 10 8}",
		  frameName, (double)0.0,
		  height-(line[0]->getValue()-minValue)*valueStep,
		  (maxDistance-minDistance)*distanceStep,
		  height-(line[0]->getValue()-minValue)*valueStep);
  module->tickle(lineCmd);
  lineCmd.sprintf("%s.c create line %f %f %f %f -width 2 -fill {#FF0000} -arrow last -arrowshape {8 10 8}",
		  frameName, -minDistance*distanceStep,
		  (double)height, -minDistance*distanceStep, 0.0);
  module->tickle(lineCmd);
  lineCmd.sprintf("bind %s.c <Motion> \"%s.v configure -text \\\"Distance : \\[ expr \\[ expr round ( \\[ expr \\[ expr \\[expr %%x / %f \\] + %f \\] * 10.0 \\])  \\] / 10.0 \\] Value : \\[ expr \\[ expr round ( \\[ expr \\[ expr \\[expr %%y / %f \\] + %f \\]  * 10.0 \\])  \\] / 10.0 \\] \\\" \"",
		  frameName, frameName, distanceStep, minDistance,
		  -valueStep, height/valueStep + minValue);
  module->tickle(lineCmd);
}

int TclRangeInrimage::displayLineCmd(int argc, char **argv,
				     RangeInrimage *client,
				     TclString& result, TclModule *module) {
  result = "";
  if (argc==7) {
    module->tickle("imagedata::profileCreation %u",
			nbIntensityProfileWindow);
    
    Vec3<double> position((double) atof(argv[0]),(double) atof(argv[1]),
			  (double) atof(argv[2]));
    Vec3<double> normal((double) atof(argv[3]),(double) atof(argv[4]),
			(double) atof(argv[5]));
    double range= (double) atof(argv[6]);
    
    IntensityProfile line= IntensityProfile(range==0 ? 10 : range);
    char frameName[200];
    unsigned int nbImages=0;
    double width,height;
      
    if (client->intensity()) nbImages++;
    if (client->gradient()) nbImages++;
    if (nbImages==1) {
      width=450;
      height=300;
    }
    else if (nbImages==2) {
      width=350;
      height=200;
    }
    else width = height = 100;

    if (client->intensity()) {
      line.setIntensity(client->intensity());
      line.compute(range,position,normal);
      sprintf(frameName,".lineDisplay%u.intensity",nbIntensityProfileWindow);
      displayGraph(frameName, "Intensity Image", line, width, height, module);
    }
    if (client->gradient()) {
      line.setIntensity(client->gradient());
      line.compute(range,position,normal);
      sprintf(frameName,".lineDisplay%u.gradient",nbIntensityProfileWindow);
      displayGraph(frameName, "Gradient Image", line, width, height, module);
    }
    module->tickle("imagedata::profileEnd %u", nbIntensityProfileWindow);
		      
    nbIntensityProfileWindow++;
    return TCL_OK;
  }
  else
    return COMMAND_SYNOPSIS;
}

int TclIntensityProfileInformation::create(TclModule *m, int argc, char **argv)
{
  if (argc<2) 
    return COMMAND_SYNOPSIS; 
  if (argc==2)
    {
      TclRangeInformation *ri= (TclRangeInformation *) TclObjectKind::getObject(typeid(TclRangeInformation), argc,argv,1);
      if (ri)
	{
	  IntensityProfileInformation *ip=dynamic_cast<IntensityProfileInformation *>(ri->getRangeInformation());
	  if (ip)
	    {
	      TclIntensityProfileInformation *tip= new TclIntensityProfileInformation(m,ip);
	      m->setResult("%s",(char *)tip->name());
	      return TCL_OK;
	    }
	  else
	    {
	      m->setResult("%s is not an intensity profile information",argv[1]);
	      return TCL_ERROR;
	    }
	}
      else return COMMAND_SYNOPSIS; 
    }
  else
    {
      TclIntensityProfile *il= (TclIntensityProfile *) TclObjectKind::getObject(typeid(TclIntensityProfile), argc,argv,1);
      if (il)
	{
	  double threshold=(double) atof(argv[2]);
	  double origin=(double) atof(argv[3]);
	  double range=(double) atof(argv[4]);
	  IntensityProfileInformation::Criterion criterion;
	  if (!strcmp(argv[5],"differences"))
	    criterion=IntensityProfileInformation::DIFFERENCES;
	  else if (!strcmp(argv[5],"linear"))
	    criterion=IntensityProfileInformation::LINEAR;
	  else if (!strcmp(argv[5],"correlation"))
	    criterion=IntensityProfileInformation::CORRELATION;
	  else if (!strcmp(argv[5],"pattern"))
	    criterion=IntensityProfileInformation::PATTERN_INTENSITY;
	  else
	    return COMMAND_SYNOPSIS;
	  IntensityProfileInformation *ip= 
	    new IntensityProfileInformation(*(il->getIntensityProfile()),origin);
	  ip->setCriterion(criterion);
	  ip->setThreshold(threshold);
	  ip->setMaxDistance(range);
	  TclIntensityProfileInformation *tip= new TclIntensityProfileInformation(m,ip);
	  m->setResult("%s",(char *)tip->name());
	  return TCL_OK;

    }
      else
	{

	  unsigned int i,size;
	  double threshold=(double) atof(argv[1]);
	  double origin=(double) atof(argv[2]);
	  IntensityProfileInformation::Criterion criterion;
	  if (!strcmp(argv[3],"differences"))
	    criterion=IntensityProfileInformation::DIFFERENCES;
	  else if (!strcmp(argv[3],"linear"))
	    criterion=IntensityProfileInformation::LINEAR;
	  else if (!strcmp(argv[3],"correlation"))
	    criterion=IntensityProfileInformation::CORRELATION;
	  else if (!strcmp(argv[3],"pattern"))
	    criterion=IntensityProfileInformation::PATTERN_INTENSITY;
	  else
	    return COMMAND_SYNOPSIS;
	  double range=(double) atof(argv[4]);
	  
	  size=(unsigned int) atoi(argv[5]);
	  
	  if(argc< (int) (size +6)) return COMMAND_SYNOPSIS;
	  double *dist=new double[size];
	  double *val=new double[size];
	  for(i=0;i<size;i+=2)
	    {
	      dist[i]=(double) atof(argv[i+6]);
	      val[i]=(double) atof(argv[i+7]);
	    }
	  IntensityProfileInformation *ip= 
	    new IntensityProfileInformation(size,origin,dist, val,criterion,
					    threshold);
	  ip->setMaxDistance(range);
	  TclIntensityProfileInformation *tip= new TclIntensityProfileInformation(m,ip);
	  delete dist;
	  delete val;
	  m->setResult("%s",(char *)tip->name());
	  return TCL_OK;
	}
    }
}

int TclIntensityProfileInformation::drawCmd(int argc, char **argv) {
  TclModule *module=getModule();
  module->tickle("imagedata::profileCreation %uip%u",++displayIndex,
		 TclObject::getRef());
  double width=450;
  double height=300;
  double val,dist;
  /// computes the min and max value of the distance and image value
  double minValue=getIntensityProfile()->getIntensity((unsigned int)0);
  double maxValue=getIntensityProfile()->getIntensity((unsigned int)0);
  double minDistance= getIntensityProfile()->getDistance((unsigned int)0);
  double maxDistance= minDistance;
  unsigned int i;
  char frameName[200];
  sprintf(frameName,".lineDisplay%uip%u.intensity",displayIndex,
	  TclObject::getRef());

  for(i=0;i<getIntensityProfile()->getSize();++i) {
      val=getIntensityProfile()->getIntensity(i);
      if (val<minValue)
	minValue=val;
      if (val>maxValue)
	maxValue=val;
      dist=getIntensityProfile()->getDistance(i);
      if (val<minDistance)
	minDistance=dist;
      if (val>maxDistance)
	maxDistance=dist;
  }
 
  double val0=getIntensityProfile()->getIntensity((double) 0.0);
  module->tickle("\
frame %s;\
label %s.l -font {Times 18} -justify center -text \"Image Intensity Profile\";\
label %s.v -font {Times 16} -justify center -text \"Distance : 0.0 Value : 0.0\";\
pack  %s.l %s.v -side top -expand true -fill x",
		      frameName, frameName, frameName, frameName,
		      frameName);
  module->tickle("\
canvas %s.c -relief raise -width %f -height %f;\
pack %s.c -side top -fill x;\
pack %s",
		      frameName, width, height, frameName, frameName);
  double distanceStep = (width-3)/(maxDistance - minDistance);
  
  double valueStep;
  if (maxValue==minValue) valueStep=height;
  else valueStep=(height-3)/(maxValue - minValue);
  TclString lineCmd;
  lineCmd.sprintf("%s.c create line ",frameName);
  char pixel[100];
  for(i=0;i<getIntensityProfile()->getSize();++i) {
    sprintf(pixel,"%f %f ", (getIntensityProfile()->getDistance(i)-minDistance)*
	    distanceStep, height-((getIntensityProfile()->getIntensity(i)-minValue)*valueStep));
    lineCmd+=pixel;
  }
  lineCmd+=" -width 2 -fill {#000000}";
  module->tickle(lineCmd);
  lineCmd.sprintf("%s.c create line %f %f %f %f -width 2 -fill {#FF0000} -arrow last -arrowshape {8 10 8}",
		  frameName, (double)0.0,
		  height-(val0-minValue)*valueStep,
		  (maxDistance-minDistance)*distanceStep,
		  height-(val0-minValue)*valueStep);
  module->tickle(lineCmd);
  lineCmd.sprintf("%s.c create line %f %f %f %f -width 2 -fill {#FF0000} -arrow last -arrowshape {8 10 8}",
		  frameName, -minDistance*distanceStep,
		  (double)height, -minDistance*distanceStep, 0.0);
  module->tickle(lineCmd);
  lineCmd.sprintf("bind %s.c <Motion> \"%s.v configure -text \\\"Distance : \\[ expr \\[ expr round ( \\[ expr \\[ expr \\[expr %%x / %f \\] + %f \\] * 10.0 \\])  \\] / 10.0 \\] Value : \\[ expr \\[ expr round ( \\[ expr \\[ expr \\[expr %%y / %f \\] + %f \\]  * 10.0 \\])  \\] / 10.0 \\] \\\" \"",
		  frameName, frameName, distanceStep, minDistance,
		  -valueStep, height/valueStep + minValue);
  module->tickle(lineCmd);
  module->tickle("imagedata::profileEnd %s %uip%u",(char *) name(),
		 displayIndex,
		 TclObject::getRef());
  return TCL_OK;
}


int TclIntensityProfileInformation::editCmd(int argc, char **argv)
{
  char *criter[] = { "differences", "linear", "correlation","pattern" };
  int i,j;
   for(i = 0; (int)i < argc; i++) {
     if(!strcmp(argv[i], "profile")) {
       unsigned int nval;
       if(++i >= argc) {
	 char tmp[200];
	 nval = getIntensityProfile()->getSize();

	 TclString result;

	result.sprintf("%i", nval);
	for(j = 0; j < (int) nval; j++) {
	  sprintf(tmp, " %f %f",getIntensityProfile()->getDistance((unsigned int)j),
		  getIntensityProfile()->getIntensity((unsigned int)j));
	  result += tmp;
	}
	setResult(result);
	return TCL_OK;
      }
      if(sscanf(argv[i], "%i", &nval) != 1) return COMMAND_SYNOPSIS;
      if((int)(i +  2*nval) >= argc) return COMMAND_SYNOPSIS;
      double *dist= new double [nval];
      double *val= new double [nval];
      for(j = 0, i++; j < (int) nval; j++, i+=2) {
	double v;
	if(sscanf(argv[i], "%lf", &v) != 1 ) {
	  delete[] dist;
	  delete[] val;
	  return COMMAND_SYNOPSIS;
	}
	else dist[j] = v;
	if(sscanf(argv[i+1], "%lf", &v) != 1 ) {
	  delete[] dist;
	  delete[] val;
	  return COMMAND_SYNOPSIS;
	}
	else val[j] = v;
      }
     }
     else if (!strcmp(argv[i], "threshold")) {
       if(++i < argc) {
	 getIntensityProfile()->setThreshold((double)atof(argv[i]));
       }
       setResult("%f", getIntensityProfile()->getThreshold());
       return TCL_OK;

     }
     else if (!strcmp(argv[i], "range")) {
       if(++i < argc) {
	 getIntensityProfile()->setMaxDistance((double)atof(argv[i]));
       }
       setResult("%f", getIntensityProfile()->getMaxDistance());
       return TCL_OK;
     }
     else if (!strcmp(argv[i], "origin")) {
       if(++i < argc) {
	 getIntensityProfile()->setOrigin((double)atof(argv[i]));
       }
       setResult("%f", getIntensityProfile()->getOrigin());
       return TCL_OK;
     }
     else if (!strcmp(argv[i], "criterion")) {
       if(++i < argc) {
	 IntensityProfileInformation::Criterion criterion;
	 if (!strcmp(argv[i],"differences"))
	   criterion=IntensityProfileInformation::DIFFERENCES;
	 else if (!strcmp(argv[i],"linear"))
	   criterion=IntensityProfileInformation::LINEAR;
	 else if (!strcmp(argv[i],"correlation"))
	   criterion=IntensityProfileInformation::CORRELATION;
	 else if (!strcmp(argv[i],"pattern"))
	   criterion=IntensityProfileInformation::PATTERN_INTENSITY;
	 else
	   return COMMAND_SYNOPSIS;
	 getIntensityProfile()->setCriterion(criterion);
       }
       setResult("%s",criter[ (unsigned int)getIntensityProfile()->getCriterion()]);
       
       return TCL_OK;
     }
     else
       return COMMAND_SYNOPSIS;
     
   }
   TclString result,res;
   result.sprintf("origin %f threshold %f criterion %s profile",
		  getIntensityProfile()->getOrigin(),
		  getIntensityProfile()->getThreshold(),
		  criter[ (unsigned int)getIntensityProfile()->getCriterion()]);
   unsigned int nval = getIntensityProfile()->getSize();

   res.sprintf( "%i ", nval);
   result+=res;
   for(i = 0; i < (int) nval; i++) {
     res.sprintf(" %f %f",getIntensityProfile()->getDistance((unsigned int)i),
	     getIntensityProfile()->getIntensity((unsigned int)i));
     result += res;
   }
   setResult(result);
   return TCL_OK;
   
}
int TclIntensityProfileInformationSet::create(TclModule *m, int argc, char **argv)
{
  if ((argc!=2) && (argc!=1))
    return COMMAND_SYNOPSIS; 
  if (argc==2)
    {
      TclRangeInformationSet *ris= (TclRangeInformationSet *) TclObjectKind::getObject(typeid(TclRangeInformationSet), argc,argv,1);
      if (ris)
	{
	  std::list<RangeInformation *> ril=ris->getRangeInformationList();
	  std::list<RangeInformation *>::iterator it;
	  bool b=true;
	  for(it=ril.begin();it!=ril.end();it++)
	    {
	      IntensityProfileInformation *ip=dynamic_cast<IntensityProfileInformation *>(*it);
	      if (ip==0)
		b=false;
	    }
	  if (b)
	    {
	      TclIntensityProfileInformationSet *tip= new TclIntensityProfileInformationSet(m,ril);
	      m->setResult("%s",(char *)tip->name());
	      return TCL_OK;
	    }
	  else
	    {
	      m->setResult("%s is not a set of intensity profile information",argv[1]);
	      return TCL_ERROR;
	    }
	}
      else return COMMAND_SYNOPSIS; 
    }
  else
    {
      TclIntensityProfileInformationSet *tip= new TclIntensityProfileInformationSet(m);
      m->setResult("%s",(char *)tip->name());
      return TCL_OK;
    }
}
int TclIntensityProfileInformationSet::editCmd(int argc, char **argv)
{
  char *criter[] = { "differences", "linear", "correlation","pattern" };
  int i;

  if (rangeInformationList.size()==0)
    {
      setResult("The set of range information is empty");
      return TCL_ERROR;
    }
  else
    {
      for(i = 0; (int)i < argc; i++) {
	if (!strcmp(argv[i], "threshold")) {
	  std::list<RangeInformation *>::iterator it;
	  if(++i < argc) {
	    double val=(double) atof(argv[i]);
	    for(it=rangeInformationList.begin();it!=rangeInformationList.end();
		it++)
	      {
		IntensityProfileInformation *ip=dynamic_cast<IntensityProfileInformation *>(*it);
		if (ip)
		  ip->setThreshold(val);
	      }
	  }
	  it=rangeInformationList.begin();
	  IntensityProfileInformation *ip=dynamic_cast<IntensityProfileInformation *>(*it);
	  if (ip)
	    setResult("%f", ip->getThreshold());
	  return TCL_OK;
	}
	else if (!strcmp(argv[i], "origin")) {
	  std::list<RangeInformation *>::iterator it;
	  if(++i < argc) {
	    double val=(double) atof(argv[i]);
	    for(it=rangeInformationList.begin();it!=rangeInformationList.end();
		it++)
	      {
		IntensityProfileInformation *ip=dynamic_cast<IntensityProfileInformation *>(*it);
		if (ip)
		  ip->setOrigin(val);
	      }
	  }
	  it=rangeInformationList.begin();
	  IntensityProfileInformation *ip=dynamic_cast<IntensityProfileInformation *>(*it);
	  if (ip)
	    setResult("%f", ip->getOrigin());
	  return TCL_OK;
	}
	else if (!strcmp(argv[i], "range")) {
	  std::list<RangeInformation *>::iterator it;
	  if(++i < argc) {
	    double val=(double) atof(argv[i]);
	    for(it=rangeInformationList.begin();it!=rangeInformationList.end();
		it++)
	      {
		IntensityProfileInformation *ip=dynamic_cast<IntensityProfileInformation *>(*it);
		if (ip)
		  ip->setMaxDistance(val);
	      }
	  }
	  it=rangeInformationList.begin();
	  IntensityProfileInformation *ip=dynamic_cast<IntensityProfileInformation *>(*it);
	  if (ip)
	    setResult("%f", ip->getMaxDistance());
	  return TCL_OK;
	}
	else if (!strcmp(argv[i], "criterion")) {
	  std::list<RangeInformation *>::iterator it;
	  if(++i < argc) {
	    IntensityProfileInformation::Criterion criterion;
	    if (!strcmp(argv[i],"differences"))
	      criterion=IntensityProfileInformation::DIFFERENCES;
	    else if (!strcmp(argv[i],"linear"))
	      criterion=IntensityProfileInformation::LINEAR;
	    else if (!strcmp(argv[i],"correlation"))
	      criterion=IntensityProfileInformation::CORRELATION;
	    else if (!strcmp(argv[i],"pattern"))
	      criterion=IntensityProfileInformation::PATTERN_INTENSITY;
	    else
	      return COMMAND_SYNOPSIS;
	    for(it=rangeInformationList.begin();it!=rangeInformationList.end();
		it++)
	      {
		IntensityProfileInformation *ip=dynamic_cast<IntensityProfileInformation *>(*it);
		if (ip)
		  ip->setCriterion(criterion);
	      }
	  }
	  it=rangeInformationList.begin();
	  IntensityProfileInformation *ip=dynamic_cast<IntensityProfileInformation *>(*it);
	  if (ip)
	    setResult("%s",criter[ (unsigned int)(ip->getCriterion())]);
	  
	  return TCL_OK;
	}

	else
	  return COMMAND_SYNOPSIS;
      }
    }
  TclString result;
  std::list<RangeInformation *>::iterator it=rangeInformationList.begin();
  IntensityProfileInformation *ip=dynamic_cast<IntensityProfileInformation *>(*it);
  if (ip)
    result.sprintf("origin %f threshold %f range %f criterion %s profile",
		   ip->getOrigin(),
		   ip->getThreshold(),
		   ip->getMaxDistance(),
		   criter[ (unsigned int)(ip->getCriterion())]);
  setResult(result);
  return TCL_OK;
   
}

int TclBlockMatchingInformation::create(TclModule *m, int argc, char **argv) {
  if(argc<2) 
    return COMMAND_SYNOPSIS; 
  if(argc==2) {
    TclRangeInformation *ri= (TclRangeInformation *) TclObjectKind::getObject(typeid(TclRangeInformation), argc,argv,1);
    if (ri) {
      BlockMatchingInformation * bi = dynamic_cast<BlockMatchingInformation *>(ri->getRangeInformation());
      if(bi) {
	TclBlockMatchingInformation * tbi = new TclBlockMatchingInformation(m,bi);
	m->setResult("%s",(char *)tbi->name());
	return TCL_OK;
      }
      else {
	m->setResult("%s is not a block matching information",argv[1]);
	return TCL_ERROR;
      }
    }
    else return COMMAND_SYNOPSIS; 
  }
  else if (argc==14) {
    Inrimage * inr = new Inrimage((TclString)argv[1]);
    Vec3<unsigned int> bc = Vec3<unsigned int>(atoi(argv[2]),atoi(argv[3]),atoi(argv[4]));
    Vec3<unsigned int> bs = Vec3<unsigned int>(atoi(argv[5]),atoi(argv[6]),atoi(argv[7]));
    Vec3<unsigned int> sr = Vec3<unsigned int>(atoi(argv[8]),atoi(argv[9]),atoi(argv[10]));
    Vec3<unsigned int> st = Vec3<unsigned int>(atoi(argv[11]),atoi(argv[12]),atoi(argv[13]));
    BlockMatchingInformation::SimilarityMeasure sm=BlockMatchingInformation::CORRELATION_COEFFICIENT;
    if (!strcmp(argv[14],"CORRELATION_COEFFICIENT"))
      sm = BlockMatchingInformation::CORRELATION_COEFFICIENT;
    else if (!strcmp(argv[14],"SSD"))
      sm = BlockMatchingInformation::SSD;
    BlockMatchingInformation * bi = new BlockMatchingInformation(inr,bc,bs,sr,st,sm);
    TclBlockMatchingInformation * tbi = new TclBlockMatchingInformation(m,bi);
    m->setResult("%s",(char *)tbi->name());
    return TCL_OK;
  }
  else
    return COMMAND_SYNOPSIS;
}

int TclBlockMatchingInformation::editCmd(int argc, char **argv) {
  return TCL_OK;
}

int TclBlockMatchingInformationSet::create(TclModule *m, int argc, char **argv) {
  if((argc!=2) && (argc!=1))
    return COMMAND_SYNOPSIS; 
  if(argc==2) {
    TclRangeInformationSet *ris= (TclRangeInformationSet *) TclObjectKind::getObject(typeid(TclRangeInformationSet), argc,argv,1);
    if(ris) {
      std::list<RangeInformation *> ril=ris->getRangeInformationList();
      std::list<RangeInformation *> ril2;
      std::list<RangeInformation *>::iterator it;
      for(it=ril.begin();it!=ril.end();it++) {
	BlockMatchingInformation *ip=dynamic_cast<BlockMatchingInformation *>(*it);
	if (ip)
	  ril2.push_back(*it);	
      }
      if(ril2.size()>0) {
	TclBlockMatchingInformationSet *tip= new TclBlockMatchingInformationSet(m,ril2);
	m->setResult("%s",(char *)tip->name());
	return TCL_OK;
      }
      else {
	m->setResult("%s do not have any block matching information",argv[1]);
	return TCL_ERROR;
      }
    }
    else return COMMAND_SYNOPSIS; 
  }
  else {
    TclBlockMatchingInformationSet *tip= new TclBlockMatchingInformationSet(m);
    m->setResult("%s",(char *)tip->name());
    return TCL_OK;
  }
}
int TclBlockMatchingInformationSet::initCmd(int argc, char **argv) {
  if(rangeInformationList.size()==0) {
    setResult("The set of range information is empty");
    return TCL_ERROR;
  }
  else {
    std::list<RangeInformation *>::iterator it=rangeInformationList.begin();
    for(;it!=rangeInformationList.end();++it) {
      BlockMatchingInformation *ip=dynamic_cast<BlockMatchingInformation *>(*it);
      if(ip)
	ip->computeRefBlockStats();
    }
    return TCL_OK;
  }
}

int TclBlockMatchingInformationSet::editCmd(int argc, char **argv) {
  int i;
  if(rangeInformationList.size()==0) {
    setResult("The set of range information is empty");
    return TCL_ERROR;
  }
  else {
    for(i=0;i<argc;++i) {
      if(!strcmp(argv[i], "blockSize")) {
	std::list<RangeInformation *>::iterator it;
	if(++i<argc) {
	  unsigned int sx = atoi(argv[i]);
	  unsigned int sy = atoi(argv[i+1]);
	  unsigned int sz = atoi(argv[i+2]);
	  Vec3<unsigned int> bs = Vec3<unsigned int>(sx,sy,sz);
	  for(it=rangeInformationList.begin();it!=rangeInformationList.end();++it) {
	    BlockMatchingInformation *ip=dynamic_cast<BlockMatchingInformation *>(*it);
	    if(ip) {
	      ip->setBlockSize(bs);
	      ip->init();
	    }
	  }
	}
	it=rangeInformationList.begin();
	BlockMatchingInformation *ip=dynamic_cast<BlockMatchingInformation *>(*it);
	if(ip)
	  setResult("%i %i %i",ip->getBlockSize().getX(),ip->getBlockSize().getY(),ip->getBlockSize().getZ());
	return TCL_OK;
      }
      if(!strcmp(argv[i], "searchRange")) {
	std::list<RangeInformation *>::iterator it;
	if(++i<argc) {
	  unsigned int sx = atoi(argv[i]);
	  unsigned int sy = atoi(argv[i+1]);
	  unsigned int sz = atoi(argv[i+2]);
	  Vec3<unsigned int> bs = Vec3<unsigned int>(sx,sy,sz);
	  for(it=rangeInformationList.begin();it!=rangeInformationList.end();++it) {
	    BlockMatchingInformation *ip=dynamic_cast<BlockMatchingInformation *>(*it);
	    if(ip)
	      ip->setSearchRange(bs);
	  }
	}
	it=rangeInformationList.begin();
	BlockMatchingInformation *ip=dynamic_cast<BlockMatchingInformation *>(*it);
	if(ip)
	  setResult("%i %i %i",ip->getSearchRange().getX(),ip->getSearchRange().getY(),ip->getSearchRange().getZ());
	return TCL_OK;
      }
      if(!strcmp(argv[i], "searchStep")) {
	std::list<RangeInformation *>::iterator it;
	if(++i<argc) {
	  unsigned int sx = atoi(argv[i]);
	  unsigned int sy = atoi(argv[i+1]);
	  unsigned int sz = atoi(argv[i+2]);
	  Vec3<unsigned int> bs = Vec3<unsigned int>(sx,sy,sz);
	  for(it=rangeInformationList.begin();it!=rangeInformationList.end();++it) {
	    BlockMatchingInformation *ip=dynamic_cast<BlockMatchingInformation *>(*it);
	    if(ip)
	      ip->setSearchStep(bs);
	  }
	}
	it=rangeInformationList.begin();
	BlockMatchingInformation *ip=dynamic_cast<BlockMatchingInformation *>(*it);
	if(ip)
	  setResult("%i %i %i",ip->getSearchStep().getX(),ip->getSearchStep().getY(),ip->getSearchStep().getZ());
	return TCL_OK;
      }
      if(!strcmp(argv[i], "similarityMeasure")) {
	std::list<RangeInformation *>::iterator it;
	if(++i<argc)
	  TclString sm = (TclString) argv[i];
	BlockMatchingInformation::SimilarityMeasure sm=BlockMatchingInformation::CORRELATION_COEFFICIENT;
	if (!strcmp(argv[14],"CORRELATION_COEFFICIENT"))
	  sm = BlockMatchingInformation::CORRELATION_COEFFICIENT;
	else if (!strcmp(argv[14],"SSD"))
	  sm = BlockMatchingInformation::SSD;
	else if (!strcmp(argv[14],"PAIREDTTEST_STUDENT"))
	  sm = BlockMatchingInformation::PAIREDTTEST_STUDENT;
	else
	  sm = BlockMatchingInformation::SSD;
	for(it=rangeInformationList.begin();it!=rangeInformationList.end();++it) {
	  BlockMatchingInformation *ip=dynamic_cast<BlockMatchingInformation *>(*it);
	  if(ip)
	    ip->setSimilarityMeasure(sm);
	}
	it=rangeInformationList.begin();
	BlockMatchingInformation *ip=dynamic_cast<BlockMatchingInformation *>(*it);
	if(ip)
	  setResult("%s",ip->getSimilarityMeasure());
	return TCL_OK;
      }
      else
	return COMMAND_SYNOPSIS;
    }
  }
  return TCL_OK;
}
void TclRangeInrimage::createCommands(TclModule *m) {
	TclObjectKind *kind;
	kind = new TclObjectKind(m, typeid(TclGradientInformation), "ginfo","Gradient Information", &TclGradientInformation::create, &TclGradientInformation::list,"[range %f]  [threshold %f] [orientation <none | direct | opposite>] [stop first %b] [intensity %b] [min %f] [max %f] ");
	kind->addOptions(typeid(TclRangeInformation));  
	m->newCommand("ginfo%u", kind->getOptions(),
		"Gradient Information management command");
	kind->newOption("-edit", (TclObject::OPTION) &TclGradientInformation::editCmd, 
		"[range %f] [orientation <none | direct | opposite>] [threshold %f] [stop <first | highest>] [intensity %b] [min %f] [max %f]",
		"get/set gradient information options with its orientation constraint, gradient voxel threshsold, intensity constraint, minimum and maximum intensity values, and first/maximum gradient intensity voxel lock");

	kind = new TclObjectKind(m, typeid(TclRegionInformation), "reginfo","Region Information", &TclRegionInformation::create, &TclRegionInformation::list,"[range %f]  [length %f] [threshold %f] [orientation <inside | outside>]  [min %f] [max %f] [smooth %b] ");
	kind->addOptions(typeid(TclRangeInformation));  
	m->newCommand("reginfo%u", kind->getOptions(),
		"Region Information management command");
	kind->newOption("-edit", (TclObject::OPTION) &TclRegionInformation::editCmd, 
		"[range %f] [orientation <inside | outside>] [threshold %f] [length %f] [min %f] [max %f] [smooth %b]",
		"get/set region force options  with its orientation constraint, gradient voxel threshsold, region length, minimum and maximum intensity values");

	kind = new TclObjectKind(m, typeid(TclIsosurfaceInformation), "isosurfinfo","Isosurface Information", &TclIsosurfaceInformation::create, &TclIsosurfaceInformation::list,"[range %f]  [isovalue %f] [method closest| insideOver | insideBelow | outsideOver | outsideBelow ]");
	kind->addOptions(typeid(TclRangeInformation));  
	m->newCommand("isosurfinfo%u", kind->getOptions(),
		"Isosurface Information management command");
	kind->newOption("-edit", (TclObject::OPTION) &TclIsosurfaceInformation::editCmd, 
		"[range %f] [isovalue %f] [constraint <closest| insideOver | insideBelow | outsideOver | outsideBelow> ]",
		"get/set isosurface force options  with its maximum range search value, isovalue and the constraint for defining the closest isosurface point");


	kind = new TclObjectKind(m, typeid(TclGradientGradientInformation), "gginfo","Gradient of the Gradient Information", &TclGradientGradientInformation::create, &TclGradientGradientInformation::list,"[threshold %f] [orientation <direct | opposite | none>] ");
	kind->addOptions(typeid(TclRangeInformation));  
	m->newCommand("gginfo%u", kind->getOptions(),
		"Gradient of the Gradient Information management command");
	kind->newOption("-edit", (TclObject::OPTION) &TclGradientGradientInformation::editCmd, 
		"[orientation <direct | opposite | none>] [threshold %f]",
		"get/set gradient of the gradient force options  with its orientation constraint, gradient voxel threshold");

	kind = new TclObjectKind(m, typeid(TclGradientBalloonInformation), "gbinfo","Gradient Balloon Information", &TclGradientBalloonInformation::create, &TclGradientBalloonInformation::list,"[threshold %f] [orientation <direct | opposite | none>] [force %f]");
	kind->addOptions(typeid(TclRangeInformation));  
	m->newCommand("gbinfo%u", kind->getOptions(),
		"Gradient Balloon Information management command");
	kind->newOption("-edit", (TclObject::OPTION) &TclGradientBalloonInformation::editCmd, 
		"[orientation <direct | opposite | none>] [threshold %f] [force %f]",
		"get/set gradient balloon force options  with its orientation constraint, gradient voxel threshold and maximum normal displacement");


	kind = new TclObjectKind(m, typeid(TclClassInformation), "clinfo","Classification Information", &TclClassInformation::create, &TclClassInformation::list,
		"[range %f] [inclasses %i+] [outclasses %i+] [inlength %f] [outlength %f]");
	kind->addOptions(typeid(TclRangeInformation));  
	m->newCommand("clinfo%u", kind->getOptions(),
		"Classification Information management command");
	kind->newOption("-edit", (TclObject::OPTION) &TclClassInformation::editCmd, 
		"[range %f] [inclasses %i+] [outclasses %i+] [inlength %f] [outlength %f]",
		"modify a class information given 1) its range value 2) the number of inner classes followed by the value of each inner class 3) the number of outer classes followed by the value of outer classes 4) the mininmum length for the inside region 5) the minimim length for the outside region");

	kind = new TclObjectKind(m, typeid(TclBlockMatchingInformation), "bminfo","Block Matching Information", &TclBlockMatchingInformation::create, &TclBlockMatchingInformation::list,
		"[refImage %s] [blockSize %i %i %i] [searchRange %i %i %i] [searchStep %i %i %i] [similarityMeasure %s]");
	kind->addOptions(typeid(TclRangeInformation));  
	m->newCommand("bminfo%u", kind->getOptions(),
		"Block Matching Information management command");
	kind->newOption("-edit", (TclObject::OPTION) &TclBlockMatchingInformation::editCmd, 
		"[refImage %s] [blockSize %i %i %i] [searchRange %i %i %i] [searchStep %i %i %i] [similarityMeasure %s]",
		"modify or get value of a block matching information");

	kind = new TclObjectKind(m, typeid(TclBlockMatchingInformationSet), "bminfos","Block Matching Information Set", &TclBlockMatchingInformationSet::create, &TclBlockMatchingInformationSet::list,
		"[reference rimg %s] [half block size %i %i %i] [half search block size %i %i %i] [step size in the search block %i %i %i] [similarity measure %s]");
	kind->addOptions(typeid(TclRangeInformationSet));  
	m->newCommand("bminfos%u", kind->getOptions(),
		"Block Matching Information Set management command");
	kind->newOption("-edit", (TclObject::OPTION) &TclBlockMatchingInformationSet::editCmd, 
		"[reference rimg %s] [half block size %i %i %i] [half search block size %i %i %i] [step size in the search block %i %i %i] [similarity measure %s]",
		"modify or get value of a Block Matching information Set");
	kind->newOption("-init", (TclObject::OPTION) &TclBlockMatchingInformationSet::initCmd, 
		"",
		"initialize the block matching stats");

	kind = new TclObjectKind(m, typeid(TclIntensityProfileInformation), "ipinfo","Intensity Profile Information", &TclIntensityProfileInformation::create, &TclIntensityProfileInformation::list,
		"[threshold %f] [origin %f] [ differences | linear | correlation | pattern ] [profile %i+]");
	kind->addOptions(typeid(TclRangeInformation));  
	m->newCommand("ipinfo%u", kind->getOptions(),
		"Intensity Profile Information management command");
	kind->newOption("-edit", (TclObject::OPTION) &TclIntensityProfileInformation::editCmd, 
		"[profile %i+] [threshold %f] [range %f] [origin %f] [ criterion <differences | linear | correlation | pattern> ]",
		"modify or get value of an intensity profile information given 1) its profile (ie the number of items in the profile followed by the list ofdistance and intensity values) 2) the threshold used to find the best point 3) the origin of the intensity profile 4) the criterion used to find the best point");
	kind->newOption("-draw", (TclObject::OPTION) &TclIntensityProfileInformation::drawCmd, 0,
		"draw the intensity profile stored in the intensity profile");

	kind = new TclObjectKind(m, typeid(TclIntensityProfileInformationSet), "ipinfos","Intensity Profile Information Set", &TclIntensityProfileInformationSet::create, &TclIntensityProfileInformationSet::list,
		"[rinfos%u]");
	kind->addOptions(typeid(TclRangeInformationSet));  
	m->newCommand("ipinfos%u", kind->getOptions(),
		"Intensity Profile Information Set management command");
	kind->newOption("-edit", (TclObject::OPTION) &TclIntensityProfileInformationSet::editCmd, 
		"[threshold %f] [origin %f] [range %f] [ criterion <differences | linear | correlation | pattern> ]",
		"modify or get value of an intensity profile information given 1) the threshold used to find the best point 2) the origin of the intensity profile 3) the criterion used to find the best point");


	kind = new TclObjectKind(m, typeid(TclIntensityProfile), "il","Image Line", &TclIntensityProfile::create, &TclIntensityProfile::list,
		"[il%u | inr%u   pos=<%f %f %f> dir=< %f %f %f> range=%f]");
	m->newCommand("il%u", kind->getOptions(),
		"Image Line management command");
	kind->newOption("-destroy", (TclObject::OPTION) &TclIntensityProfile::destroyCmd,0, 
		"destroy the image line object");
	kind->newOption("-draw", (TclObject::OPTION) &TclIntensityProfile::drawCmd,0, 
		"draw the intensity profile stored in the image line");
	kind->newOption("-dump", (TclObject::OPTION) &TclIntensityProfile::dumpCmd,0, 
		"returns a TCL list corresponding to the set of pair value {distance, intensity value } ");
	kind->newOption("-dumpGradient", (TclObject::OPTION) &TclIntensityProfile::dumpGradientCmd,0, 
		"returns a TCL list corresponding to the set of pair value {distance, gradient value } ");


	kind = new TclObjectKind(m, typeid(TclIntensityProfileSet), "ils","Image Line Set", &TclIntensityProfileSet::create, &TclIntensityProfileSet::list,
		"[ils%u | ]");
	m->newCommand("ils%u", kind->getOptions(),
		"Image Line set management command");
	kind->newOption("-destroy", (TclObject::OPTION) &TclIntensityProfileSet::destroyCmd,0, 
		"destroy the image line set object");
	kind->newOption("-draw", (TclObject::OPTION) &TclIntensityProfileSet::drawCmd,"%s", 
		"draw the intensity profiles into a 2D image. The parameter specifies if the positive side (outside), the negative side (inside) or both of them (bothsides) are taken into account");
	kind->newOption("-fuzzyCMeans", (TclObject::OPTION) &TclIntensityProfileSet::fuzzyCMeansCmd,"%u %s", 
		"classify intensity profiles using the Fuzzy C-Means clustering algorithm and then draw a 2D image. The first parameter is the number of clusters used by the Fuzzy C-Means clustering algorithm. The second parameter specifies if the positive side (outside), the negative side (inside) or both of them (bothsides) are taken into account");

}
} // namespace yav
