#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "imagedata/TclRangeInrimage3D.h"
#include <imagedata/TclRangeInrimage3D.h>

namespace yav {

TclRangeInrimage3D::TclRangeInrimage3D(TclInrimage *intensity,
				       TclScene3D *scene,
				       TclInrimage *gradient,
				       TclInrimage *gradientdir,
				       TclInrimage *edge,
				       TclInrimage *gradient2,
				       TclInrimage *classify,
				       const std::type_info& id,
				       RangeInrimage3DGraphics *data)
  : TclScene3DObject( id,
       data ? data
	    : new RangeInrimage3DGraphics(intensity->getClient(), scene->getClient(),
				  gradient ? gradient->getClient() : 0,
				  gradientdir ? gradientdir->getClient() : 0,
				  edge ? edge->getClient() : 0,
				  gradient2 ? gradient2->getClient() : 0,
				  classify ? classify->getClient() : 0)) 
{ 
  done(typeid(TclRangeInrimage3D),
       "imagedata::range3Dinstall", "imagedata::range3Duninstall");
}

TclRangeInrimage3D::TclRangeInrimage3D(const TclString& filename,
				       TclScene3D *scene,
				       const std::type_info& id,
				       RangeInrimage3DGraphics *data)
  : TclScene3DObject(id,
		 data ? data
		      : new RangeInrimage3DGraphics(filename, scene->getClient())) {
  //TclInrimage *image=0;  // commented by fpoyer. reason : unused.

 
  if (getClient()->getIntensity())  {
    (void) new TclInrimage(getModule(), typeid(TclInrimage),
			   (getClient()->getIntensity()));
   // getClient()->InrimageView::init();
  }
  if (getClient()->getGradient())
    (void) new TclInrimage(getModule(), typeid(TclInrimage),
			   (getClient()->getGradient()));
  if (getClient()->getGradientDirection())
    (void) new TclInrimage(getModule(), typeid(TclInrimage),
			   (getClient()->getGradientDirection()));

  if(getClient()->getGradientGradient())
    (void) new TclInrimage(getModule(), typeid(TclInrimage),
			   (getClient()->getGradientGradient()));
  if(getClient()->getEdge())
    (void) new TclInrimage(getModule(), typeid(TclInrimage),
			   (getClient()->getEdge()));
  if (getClient()->getClassify())
    (void) new TclInrimage(getModule(), typeid(TclInrimage),
			   (getClient()->getClassify()));

 

  done(typeid(TclRangeInrimage3D),
       "imagedata::range3Dinstall", "imagedata::range3Duninstall");
}

TclRangeInrimage3D::TclRangeInrimage3D(TclScene3D *scene,
				       const std::type_info& id,
				       RangeInrimage3DGraphics *data)
  : TclScene3DObject( id, data) {
  
  done(typeid(TclRangeInrimage3D),
       "imagedata::range3Dinstall", "imagedata::range3Duninstall");
}

TclRangeInrimage3D::~TclRangeInrimage3D() {
  deleted(typeid(TclRangeInrimage3D));
}

int TclRangeInrimage3D::create(TclModule *m, int argc, char **argv) {
  TclRangeInrimage3D *img;
  TclInrimage *inr = 0, *gradient = 0, *gradientdir = 0, *edge = 0,
    *gradient2 = 0, *classify = 0;
  TclScene3D *scene = 0;
  int i, filename = -1;
  
  for(i = 1; i < argc; i++) {
    if(argv[i][0] != '-') {
      if(inr || filename >= 0) return COMMAND_SYNOPSIS;
      inr = TclInrimage::getObject(argc, argv, i);
      if(!inr) filename = i;
    }
    else if(!strcmp(argv[i], "-scene")) {
      scene = (TclScene3D *)
	TclObjectKind::getObject(typeid(TclScene3D), argc, argv, ++i);
      if(!scene) return TCL_ERROR;
    }
    else if(!strcmp(argv[i], "-gradient")) {
      if(gradient) return COMMAND_SYNOPSIS;
      gradient = TclInrimage::getObject(argc, argv, ++i);
      if(!gradient) return TCL_ERROR;
    }
    else if(!strcmp(argv[i], "-gradientdir")) {
      if(gradientdir) return COMMAND_SYNOPSIS;
      gradientdir = TclInrimage::getObject(argc, argv, ++i);
      if(!gradientdir) return TCL_ERROR;
    }
    else if(!strcmp(argv[i], "-edge")) {
      if(edge) return COMMAND_SYNOPSIS;
      edge = TclInrimage::getObject(argc, argv, ++i);
      if(!edge) return TCL_ERROR;
    }
    else if(!strcmp(argv[i], "-gradient2")) {
      if(gradient2) return COMMAND_SYNOPSIS;
      gradient2 = TclInrimage::getObject(argc, argv, ++i);
      if(!gradient2) return TCL_ERROR;
    }
    else if(!strcmp(argv[i], "-classify")) {
      if(classify) return COMMAND_SYNOPSIS;
      classify = TclInrimage::getObject(argc, argv, ++i);
      if(!classify) return TCL_ERROR;
    }
    else return COMMAND_SYNOPSIS;
  }

  if(!scene) scene = TclScene3D::getDefaultScene(m);
  if(filename < 0) {
    if(!inr) return COMMAND_SYNOPSIS;
    img = new TclRangeInrimage3D(inr, scene, gradient, gradientdir, edge,
				 gradient2, classify);
  }
  else {
    try {
      img = new TclRangeInrimage3D(TclString(argv[filename]), scene);
    }
    catch(Inrimage::ImageNotFound&) {
      m->setResult("file not found");
      return TCL_ERROR;
    }
    catch(Inrimage::UnknownImageType&) {
      m->setResult("unknown image type");
      return TCL_ERROR;
    }
    catch(Inrimage::InvalidHeader&) {
      m->setResult("invalid inrimage header");
      return TCL_ERROR;
    }
    catch(Inrimage::InvalidData&) {
      m->setResult("invalid inrimage data");
      return TCL_ERROR;
    }
    catch(Inrimage::UnknownType&) {
      m->setResult("non supported inrimage type");
      return TCL_ERROR;
    }
  }
   
  m->setResult("%s%u", img->getObjectName(), img->getRef());
  return TCL_OK;
}

int TclRangeInrimage3D::imagesCmd(int argc, char **argv) {
  TclString res;
  int error = TclRangeInrimage::imagesCmd(argc, argv, getClient(), res);
  setResult(res);
  return error;
}

int TclRangeInrimage3D::saveCmd(int argc, char **argv) {
  TclString res;
  int error = TclRangeInrimage::saveCmd(argc, argv, getClient(), res);
  setResult(res);
  return error;
}


int TclRangeInrimage3D::displayLineCmd(int argc, char **argv) {
  TclString res;
  int error = TclRangeInrimage::displayLineCmd(argc, argv, getClient(), res,
					       getModule());
  setResult(res);
  return error;
}

int TclRangeInrimage3D::destroyCmd(int argc, char **argv) 
{
  if (argc>0)
    return COMMAND_SYNOPSIS;
  else
    {
      TclModule *m = getModule();
     TclString n=name();
      delete this;
      m->setResult("%s does not exist any longer",(char *)n);;
      return TCL_OK;
    }
}
int TclRangeInrimage3D::forceCmd(int argc, char **argv) {
  TclString res;
  int error = TclRangeInrimage::forceCmd(argc, argv, getClient(), res);
  setResult(res);
  return error;
}

void TclRangeInrimage3D::load(const TclString &filename, Scene3D *scn) {
  TclInrimage *inr = new TclInrimage(((TclScene3D *)scn->getInterface())->getModule(),
				     filename);
  (void) new TclRangeInrimage3D(inr, (TclScene3D *)scn->getInterface());
}

TclRangeInrimage3D *TclRangeInrimage3D::getInterface(RangeInrimage3D *range) 
{
	TclRangeInrimage3D *r,*res;
	res=0;
	TclObjectKind *ok=TclObjectKind::findByKind(typeid(RangeInrimage3D));
	if ( !ok )
		return res;
	  ObjectsList<TclObject>::iterator it=ok->firstObject();
	  while (it!=ok->lastObject()) {
		  r=(TclRangeInrimage3D *)(*it);
		  if (r->getClient()==range)
			  res=r;
		  it++;
	  }
	  return res;
}


/// creates the Tcl commands when the module is loaded
void  TclRangeInrimage3D::createCommands(TclModule *m)
{
	TclObjectKind *kind;	
 kind = new Scene3DItem(m, typeid(TclRangeInrimage3D), "rimg3D","3D Image Deformation",
			 &TclRangeInrimage3D::create,
			 &TclObject::list,
			 TclScene3D::FILE | TclScene3D::OTHER,
			 "", ".rimg3D,.inr,.inr.gz", &TclRangeInrimage3D::load, 0,
			 "<filename=%s|inr%u [-gradient inr%u] [-gradientdir inr%u] [-edge inr%u] [-gradient2 inr%u] [-classify inr%u]> [-scene scn3D%u]");
  m->newCommand("rimg3D%u", kind->getOptions(), "range image management command");
  TclScene3DObject::createOptionCommands(kind);
  kind->newOption("-images", (TclObject::OPTION) &TclRangeInrimage3D::imagesCmd,
		  "[[intensity inr%u] [gradient inr%u] [gradientdir inr%u] [edge inr%u] [gradient2 inr%u] [classify inr%u] ]",
		  "get/set inrimages composing data image");
  kind->newOption("-destroy", (TclObject::OPTION) &TclRangeInrimage3D::destroyCmd,
	  0,"destroy a 3D range image");

  kind->newOption("-force",
		  (TclObject::OPTION) &TclRangeInrimage3D::forceCmd,
		  "[ list | none | gradient | ... ]",
		  "get/set external force type or return available forces list");
  kind->newOption("-save",
		  (TclObject::OPTION) &TclRangeInrimage3D::saveCmd,
		  "%s",
		  "save the range inrimage 3D in a General Format file");
  kind->newOption("-displayLine",
		  (TclObject::OPTION) &TclRangeInrimage3D::displayLineCmd,
		  "position=(%f %f %f) normal=(%f %f %f) range=%f color=%s title=%s",
		  "display in a window the image intensity, gradient and gradientDirection profile along a 3D segment. The first 3 parameters are the coordinates of the middle profile. The second 3 parameters are the coordinates of the direction of the line segment. The last parameter is the half length of the segment");
}

} // namespace yav
