#ifndef RANGEINRIMAGE_H
#define RANGEINRIMAGE_H


#ifdef _MSC_VER
#pragma warning ( disable : 4068 )
#pragma warning ( disable : 4786 )
#pragma warning ( disable : 4081 )
#endif

#include <framework/RangeInformation.h>
#include <inrimage/Inrimage.h>
#include <imagedata/IntensityProfile.h>
#include <imagedata/IntensityProfile.h>
#include <inrimage/recursiveFilters.h>
#include <basics/Probability.h>
namespace yav {


/** \class GradientBalloonInrimageInformation
    \ingroup ImageData
    a class that store all information necessary for computing an external
    force based on the gradient of the gradient norm  and a ballon force*/
class GradientBalloonInrimageInformation : public RangeInformation {
 public:
  /// name stored in the Range Information
  static const  TclString informationName;
 /// test gradient orientation constraint
  enum OrientationConstraint {
    /// accept all voxels
    NONE = 0,
    /// accept only voxels whose gradient orientation match line orientation
    ALONG_GRADIENT = 1,
    /// accept only voxels whose gradient orientation match opposite line orientation
    ALONG_OPPOSITE_GRADIENT = 2,
  };
 protected:
  /// the gradient norm value above which  a voxel is considered for test
  double gradientThreshold;
  /// if there is a constraint on the voxel gradient orientation
  OrientationConstraint gradientConstraint;
  /// the maximum displacement allowed in the normal direction
  double balloonForce;
 public:
  /** @name constructor */
  //@{
  /**  builds a data structure for a force computed as the gradient of gradient norm : requires a gradient threshold and its orientation constraint  
       @param _gradientThreshold the threshold on the gradient norm value
       @param _gradientConstraint the side on which to look for the region
       @param _balloonForce the maximum displacement allowed in the normal direction */
  GradientBalloonInrimageInformation(double _gradientThreshold = 0.0,
				     OrientationConstraint _gradientConstraint = NONE,
				     double _balloonForce=0.0);
  //@}

    /** @name destructor */
    //@{
    /// destroys the object
    virtual ~GradientBalloonInrimageInformation(){}
    //@}
       /// get the gradient threshold
    double getThreshold() const {
      return (gradientThreshold);
    }
    /// set the gradient threshold
    void setThreshold(const double threshold) {
      gradientThreshold = threshold;
    }
    /// get the gradient constraint
    OrientationConstraint getOrientationConstraint() const {
      return (gradientConstraint);
    }
    /// set the gradient constraint
    void setOrientationConstraint(const OrientationConstraint constraint) {
      gradientConstraint=constraint;}
    /** returns the normal displacement used to compute the
	gradient balloon Force */
    double getBalloonForce() const {
      return(balloonForce);
    }
    /** sets  the normal displacement used to compute the
	gradient balloon Force */
    void setBalloonForce(const double force) {
      balloonForce=force;
    }
    /// create a gradient of the gradient information Object from a file
    static RangeInformation *readGradientBalloonInformation(GFInput &input);
    /** create a copy of this region information object. This is used to apply 
	the same algorithm to all vertices of a mesh */
    virtual RangeInformation* duplicate() {
      return new GradientBalloonInrimageInformation(*this); }
    /** write in a file the data stored in a gradient of the gradient information object
	@param output a pointer on a file */
    virtual void writeInFile(GFOutput &output);
};

/** \class GradientGradientInrimageInformation
    \ingroup ImageData
    a class that store all information necessary for computing an external
    force based on the gradient of the gradient norm  */
class GradientGradientInrimageInformation : public RangeInformation {
 public:
  /// name stored in the Range Information
  static const  TclString informationName;
 /// test gradient orientation constraint
  enum OrientationConstraint {
    /// accept all voxels
    NONE = 0,
    /// accept only voxels whose gradient orientation match line orientation
    ALONG_GRADIENT = 1,
    /// accept only voxels whose gradient orientation match opposite line orientation
    ALONG_OPPOSITE_GRADIENT = 2,
  };
 protected:
  /// the gradient norm value above which  a voxel is considered for test
  double gradientThreshold;
  /// if there is a constraint on the voxel gradient orientation
  OrientationConstraint gradientConstraint;
 public:
  /** @name constructor */
  //@{
  /**  builds a data structure for a force computed as the gradient of gradient norm : requires a gradient threshold and its orientation constraint  
       @param _gradientThreshold the threshold on the gradient norm value
       @param _gradientConstraint the side on which to look for the region*/
  GradientGradientInrimageInformation(double _gradientThreshold = 0.0,
				      OrientationConstraint _gradientConstraint = NONE);
  //@}

    /** @name destructor */
    //@{
    /// destroys the object
    virtual ~GradientGradientInrimageInformation(){}
    //@}
       /// get the gradient threshold
    double getThreshold() const {
      return (gradientThreshold);
    }
    /// set the gradient threshold
    void setThreshold(const double threshold) {
      gradientThreshold = threshold;
    }
    /// get the gradient constraint
    OrientationConstraint getOrientationConstraint() const {
      return (gradientConstraint);
    }
    /// set the gradient constraint
    void setOrientationConstraint(const OrientationConstraint constraint) {
      gradientConstraint=constraint;}
    /// create a gradient of the gradient information Object from a file
    static RangeInformation *readGradientGradientInformation(GFInput &input);
    /** create a copy of this region information object. This is used to apply 
	the same algorithm to all vertices of a mesh */
    virtual RangeInformation* duplicate() {
      return new GradientGradientInrimageInformation(*this); }
    /** write in a file the data stored in a gradient of the gradient information object
	@param output a pointer on a file */
    virtual void writeInFile(GFOutput &output);
};

/** \class IsosurfaceInrimageInformation
    \ingroup ImageData
    a class that store all information necessary for computing an external
    force based on the iso-intensity value  */
class IsosurfaceInrimageInformation : public RangeInformation {
 public:
  /// name stored in the Range Information
  static const  TclString informationName;
   /// contraint isosurface definition
  enum IsosurfaceConstraint {
    /// finds the closest isosurface
    CLOSEST = 0,
	/// defines isosurface from a transition + -> - starting from the inside voxels 
    INSIDE_OVER_ISOVALUE = 1,
	/// defines isosurface from a transition - -> + starting from the inside voxels 
    INSIDE_BELOW_ISOVALUE = 2,
	/// defines isosurface from a transition + -> - starting from the outside voxels 
    OUTSIDE_OVER_ISOVALUE = 3,
	/// defines isosurface from a transition - -> + starting from the outside voxels 
    OUTSIDE_BELOW_ISOVALUE = 4,
  };
 protected:
  /// the gradient norm value above which  a voxel is considered for test
  double isovalue;
  /// the half-size of intensity profile
  double range;
  /// how the isosurface is defined
  IsosurfaceConstraint isosurfaceConstraint;
 
 public:
  /** @name constructor */
  //@{
  /**  builds a data structure for a force that find the closest intersection 
       with an isosurface : requires am isovalue 
       @param _isovalue the intensity level defining the isosurface
       @param _r the maximum half-distance of search from the origin */
  IsosurfaceInrimageInformation(double _isovalue= 0.0,double _r=10.0,IsosurfaceConstraint def=INSIDE_OVER_ISOVALUE);
  //@}
  
  /** @name destructor */
  //@{
  /// destroys the object
  virtual ~IsosurfaceInrimageInformation(){}
  //@}
  /// get the gradient threshold
  double getIsovalue() const {
    return (isovalue);
  }
  /// set the gradient threshold
  void setIsovalue(const double iso) {
    isovalue = iso;
  }
  /// get the maximum range distance 
  double getMaxDistance(void) const {
    return (range);
  }
  /// set the maximum range distance 
  void setMaxDistance(double r) {
    range = r;
  }
  /// get the isosurface constraint
  IsosurfaceConstraint getIsosurfaceConstraint() const {
	  return (isosurfaceConstraint);
  }
  /// set the isosurface constraint
  void setIsosurfaceConstraint(const IsosurfaceConstraint constraint) {
	  isosurfaceConstraint=constraint;}
  /// create a gradient of the gradient information Object from a file
  static RangeInformation *readIsosurfaceInformation(GFInput &input);
  /** create a copy of this region information object. This is used to apply 
      the same algorithm to all vertices of a mesh */
  virtual RangeInformation* duplicate() {
    return new IsosurfaceInrimageInformation(*this); }
  /** write in a file the data stored in a gradient of the gradient information object
      @param output a pointer on a file */
  virtual void writeInFile(GFOutput &output);
};


/** \class RegionInrimageInformation
    \ingroup ImageData
    a class that store all information necessary for computing an external
    force based on a region definition */
class RegionInrimageInformation : public RangeInformation {
 public:
  /// name stored in the Range Information
  static const  TclString informationName;
  /// test gradient orientation constraint
  enum OrientationConstraint {
    /// look for an inside region
    INSIDE = 0,
    /// look for an outside region 
    OUTSIDE = 1
  };
  
 private:
  /// the half-size of intensity profile
  double range;
  /// the minimum length for identifying a region
  double minimumRegionLength;
  /// the gradient norm value above which a voxel is considered for test
  double gradientThreshold;
  /// whether the region is defined in the inside or outside
  OrientationConstraint regionConstraint;
  /// the minimum Intensity value when checking if a voxel should be retained
  double minimumIntensity;
  /// the maximum Intensity value when checking if a voxel should be retained
  double maximumIntensity;
  /// whether the line should be smoothed
  bool smoothLine;
        
 public:
  /** @name constructor */
  //@{
  /**  builds a data structure for a region force given  the gradient threshold and its orientation constraint and its intensity range 
       @param max the associated range inrimage
       @param regionThreshold the threshold on the gradient norm value
       @param regionConstraint the side on which to look for the region
       @param _minimumRegionLength the minimum length for detecting a region
       @param _minimumIntensity the lower threshold defining a region
       @param _maximumIntensity the upper threshold defining a region
       @param _smoothLine if the image profile should be smoothed*/
  RegionInrimageInformation(double max = 10.0,
			    double regionThreshold = 0.0,
			    OrientationConstraint regionConstraint = INSIDE,
			    double _minimumRegionLength=4,
			    double _minimumIntensity=0,
			    double _maximumIntensity=255,
			    bool _smoothLine=false);
  //@}

    /** @name destructor */
    //@{
    /// destroys any line voxel
    virtual ~RegionInrimageInformation(){}
    //@}

    /// get the maximum range distance 
    double getMaxDistance(void) const {
      return (range);
    }
    /// set the maximum range distance 
    void setMaxDistance(double r) {
      range = r;
    }

    /// Returns the minimum length for identifying a region
    double getMinimumLength() const {
      return(minimumRegionLength);
    }
    /// Sets the minimum length for identifying a region
    void setMinimumLength(const double _minimumRegionLength) {
      minimumRegionLength=_minimumRegionLength;
    }
    /// get the gradient threshold
    double getThreshold() const {
      return (gradientThreshold);
    }
    /// set the gradient threshold
    void setThreshold(const double threshold) {
      gradientThreshold = threshold;
    }
    /// get the gradient constraint
    OrientationConstraint getOrientationConstraint() const {
      return (regionConstraint);
    }
    /// set the gradient constraint
    void setOrientationConstraint(const OrientationConstraint constraint) {
      regionConstraint=constraint;}
    /// returns the minimum intensity value for accepting a voxel as edge voxel
    double getMinimumIntensity() const {
      return(minimumIntensity);
    }
    /// returns the maximum intensity value for accepting a voxel as edge voxel
    double getMaximumIntensity() const {
      return(maximumIntensity);
    }
    /// sets the minimum intensity value for accepting a voxel as edge voxel
    void setMinimumIntensity(const double _minimumIntensity) {
      minimumIntensity=_minimumIntensity;
    }
    /// sets the maximum intensity value for accepting a voxel as edge voxel
    void setMaximumIntensity(const double _maximumIntensity) {
      maximumIntensity=_maximumIntensity;
    }
    /// return the boolean indicating if the line should be smoothed
    bool getSmoothLine() const {
      return(smoothLine);
    }
    /// sets the boolean indicating if the line should be smoothed
    void setSmoothLine(const bool _smoothLine)  {
      smoothLine= _smoothLine;
    }
    /// create an region information Object from a file
    static RangeInformation *readRegionInformation(GFInput &input);
    /** create a copy of this region information object. This is used to apply 
	the same algorithm to all vertices of a mesh */
    virtual RangeInformation* duplicate() {
      return new RegionInrimageInformation(*this); }
    /** write in a file the data stored in a region information object
	@param output a pointer on a file */
    virtual void writeInFile(GFOutput &output);
};

/** \class RegionInrimageInformation
    \ingroup ImageData
    a class that store all information necessary for computing an external
    force based on a region definition */
class RegionProbabilityMapInformation : public RangeInformation {
 public:
  /// name stored in the Range Information
  static const  TclString informationName;
  /// test gradient orientation constraint
  enum OrientationConstraint {
    /// look for an inside region
    INSIDE = 0,
    /// look for an outside region 
    OUTSIDE = 1
  };
  
 private:
  /// the half-size of intensity profile
  double range;
  /// the minimum length for identifying a region
  double minimumRegionLength;
  /// whether the region is defined in the inside or outside
  OrientationConstraint regionConstraint;
  /// the pdf function that gives the probability that a voxel is inside the seeked region
  ProbabilityDensityFunction *insidePDF;
  /// the pdf function that gives the probability that a voxel is outside the seeked region
  ProbabilityDensityFunction *outsidePDF;
        
 public:
  /** @name constructor */
  //@{
  /**  builds a data structure for a region force given  the gradient threshold and its orientation constraint and its intensity range 
       @param max the associated range inrimage
       @param regionConstraint the side on which to look for the region
       @param _minimumRegionLength the minimum length for detecting a region
       @param _pdfInside the pdf that indicates if a voxel intensity is likely to be inside the region
       @param _pdfOutside the pdf that indicates if a voxel intensity is likely to be outside the region */
	 RegionProbabilityMapInformation(double max = 10.0,
		 OrientationConstraint regionConstraint = INSIDE,
		 double _minimumRegionLength=4,
		 ProbabilityDensityFunction *_pdfInside=0,
		 ProbabilityDensityFunction *_pdfOutside=0);
	 //@}

    /** @name destructor */
    //@{
    /// destroys any line voxel
    virtual ~RegionProbabilityMapInformation(){}
    //@}

    /// get the maximum range distance 
    double getMaxDistance(void) const {
      return (range);
    }
    /// set the maximum range distance 
    void setMaxDistance(double r) {
      range = r;
    }

    /// Returns the minimum length for identifying a region
    double getMinimumLength() const {
      return(minimumRegionLength);
    }
    /// Sets the minimum length for identifying a region
    void setMinimumLength(const double _minimumRegionLength) {
      minimumRegionLength=_minimumRegionLength;
    }
   
    /// get the gradient constraint
    OrientationConstraint getOrientationConstraint() const {
      return (regionConstraint);
    }
    /// set the gradient constraint
    void setOrientationConstraint(const OrientationConstraint constraint) {
      regionConstraint=constraint;}

  /// returns the pdf function that gives the probability that a voxel is inside the seeked region
    ProbabilityDensityFunction *getInsidePDF() const {
      return(insidePDF);
    }
  /// returns the pdf function that gives the probability that a voxel is outside the seeked region
    ProbabilityDensityFunction *getOutsidePDF() const {
      return(outsidePDF);
    }
    /// sets the pdf function that gives the probability that a voxel is inside the seeked region
    void setInsidePDF( ProbabilityDensityFunction *pdf) {
      insidePDF=pdf;
    }
    /// sets the pdf function that gives the probability that a voxel is outside the seeked region
    void setOutsidePDF( ProbabilityDensityFunction *pdf) {
      outsidePDF=pdf;
    }
 
    virtual RangeInformation* duplicate() {
      return new RegionProbabilityMapInformation(*this); }
};
/** \class GradientInrimageInformation
    \ingroup ImageData
    a class that store all information necessary for computing an external
    force based on the gradient */
class GradientInrimageInformation : public RangeInformation {
 public:
  /// name stored in the Range Information
  static const  TclString informationName;
  /// test gradient orientation constraint
  enum OrientationConstraint {
    /// accept all voxels
    NONE = 0,
    /// accept only voxels whose gradient orientation match line orientation
    ALONG_GRADIENT = 1,
    /// accept only voxels whose gradient orientation match opposite line orientation
    ALONG_OPPOSITE_GRADIENT = 2,
  };
  
 private:
  /// the half-distance along which the highest gradient voxel is searched
  double range;
  /// the gradient norm value above which  a voxel is considered for test
  double gradientThreshold;
  /// if there is a constraint on the voxel gradient orientation
  OrientationConstraint gradientConstraint;
  /// should the scan-line stop on highest gradient vertex or first one higher than threshold?
  bool stopAtFirst;
  /// should the intensity be used  for looking for the optimal voxel
  bool useIntensity;
  /// the minimum Intensity value when checking if a voxel should be retained
  double minimumIntensity;
  /// the maximum Intensity value when checking if a voxel should be retained
  double maximumIntensity;
  
 public:
  /** @name constructor */
  //@{
  /**  builds a data structure storing the gradient threshold, its orientation constraint and its intensity range 
       @param max the half-distance used for the search algorithm
       @param gradientThreshold the threshold on the gradient norm value
       @param gradientConstraint the constraint on the gradient direction,
       @param stopAtFirst should the scan-line stop on highest gradient vertex or first one higher than threshold?
       @param useIntensity if intensity should be considered when finding the best voxel
       @param minIntensity the minimum intensity of the intensity range
       @param maxIntensity the maximum intensity of the intensity range */
  GradientInrimageInformation(double max = 10.0,
			      double gradientThreshold = 0.0,
			      OrientationConstraint gradientConstraint = NONE,
			      bool stopAtFirst = false,
			      bool useIntensity = false,
			      double minIntensity = 0.0,
			      double maxIntensity = 255.0); 
  //@}
  
  /** @name destructor */
  //@{
  /// destroys the objwct
  virtual ~GradientInrimageInformation(){}
  //@}

  /// get the maximum range distance 
  double getMaxDistance(void) const {
    return (range);
  }
  /// set the maximum range distance 
  void setMaxDistance(double r) {
    range = r;
  }
  /// get the gradient threshold
  double getThreshold(void) const {
    return (gradientThreshold);
  }
  /// set the gradient threshold
  void setThreshold(double threshold) {
    gradientThreshold = (threshold < 0.0) ? 0.0 : threshold;
  }
  /// get the gradient constraint
  OrientationConstraint getOrientationConstraint(void) const {
    return (gradientConstraint);
  }
  /// set the gradient constraint
  void setOrientationConstraint(OrientationConstraint constraint) {
    gradientConstraint=constraint; }
  /// returns a boolean indicating if voxel intensity should be considered
  bool getIntensityConstraint(void) const {
    return(useIntensity);
  }
  /// set the intensity constraint
  void setIntensityConstraint(const bool _useIntensity) {
    useIntensity=_useIntensity;
  }
  /// returns the minimum intensity value for accepting a voxel as edge voxel
  double getMinimumIntensity(void) const {
    return(minimumIntensity);
  }
  /// returns the maximum intensity value for accepting a voxel as edge voxel
  double getMaximumIntensity(void) const {
    return(maximumIntensity);
  }
  /// sets the minimum intensity value for accepting a voxel as edge voxel
  void setMinimumIntensity(const double _minimumIntensity) {
    minimumIntensity=_minimumIntensity;
  }
  /// sets the maximum intensity value for accepting a voxel as edge voxel
  void setMaximumIntensity(const double _maximumIntensity) {
    maximumIntensity=_maximumIntensity;
  }
  /// returns stop condition (true means stop at first gradient voxel)
  double getStopCondition(void) const {
    return stopAtFirst;
  }
  /// set stop condition (true means stop at first gradient voxel)
  void setStopCondition(bool stop) {
    stopAtFirst = stop;
  }
  /// create a gradient information Object from a file
  static RangeInformation *readGradientInformation(GFInput &input);
  /** create a copy of this gradient information object. This is used to apply 
      the same algorithm to all vertices of a mesh */
  virtual RangeInformation* duplicate() {
    return new GradientInrimageInformation(*this); }
  /** write in a file the data stored in a gradient information object
      @param output a pointer on a file */
  virtual void writeInFile(GFOutput &output);

  friend class GradientForce;
};

/** \class ClassInrimageInformation
    \ingroup ImageData
    a class that store all information necessary for computing an external
    force based on the image classification  */
 class ClassInrimageInformation : public RangeInformation {
 public:
   /// name stored in the Range Information
   static const  TclString informationName;
 protected:
   /// the half-distance along which the highest gradient voxel is searched
   double range;
  /// number of elements in <code>in</code> array
   unsigned int nin;
   /// number of elements in <code>out</code> array
   unsigned int nout;
   /// class values that should be inside the model
   unsigned char *in;
   /// class values that should be outside the model
   unsigned char *out;
   /// minimum length of inside voxels
   double inlength;
   /// minimum length of outside voxels
   double outlength;

 public:
   /** @name constructor */
   //@{
   /**  builds a data structure storing the gradient threshold, its orientation constraint and its intensity range 
	@param max the half-distance used for the search algorithm
	@param _nint the number of elements in the array describing the inside class labels
	@param _in the array describing the inside class labels
	@param _nout the number of elements in the array describing the outside class labels
	@param _out the array describing the outside class labels*/
   ClassInrimageInformation(double max = 10.0, 
			    unsigned int _nint = 0,
			    unsigned char *_in=0,
			    unsigned int _nout = 0,
			    unsigned char *_out=0);
   //@}
   
   /** @name destructor */
   //@{
   /// destroys the object
   virtual ~ClassInrimageInformation(){}
   //@}
   /// create a class information Object from a file
   static RangeInformation *readClassInformation(GFInput &input);
   /** create a copy of this class information object. This is used to 
       apply the same algorithm to all vertices of a mesh */
   virtual RangeInformation* duplicate() {
     return new ClassInrimageInformation(*this); }
   /** write in a file the data stored in a gradient information object
       @param output a pointer on a file */
   virtual void writeInFile(GFOutput &output);
   /// get the maximum range distance 
   double getMaxDistance(void) const {
     return (range);
   }
   /// set the maximum range distance 
   void setMaxDistance(double r) {
     range = r;
   }
   /// returns minimum length of inside voxels
   double getInlength(void) const { return inlength; }
   /// returns minimum length of outside voxels
   double getOutlength(void) const { return outlength; }
   /// set minimum length of inside voxels
   void setInlength(double d);
   /// set minimum length of outside voxels
   void setOutlength(double d);
   /** return the maximum number of inner classes */
   unsigned int getMaxInClasses() const {
     return nin;}
   /** return the maximum number of outer classes */
   unsigned int getMaxOutClasses() const {
     return nout;}
   /** return class value that should lie inside the model
       @param i index of the  inner class values
       @return inner class value of the ith class */
   unsigned char getIn(unsigned int i) {
     if (i<nin)
       return in[i];
     else 
       return 0;}
   /** return class value that should lie outside the model
       @param i index of the  outer class values
       @return outer class value of the ith class */
   unsigned char getOut(unsigned int i) {
     if (i<nout)
       return out[i];
     else 
       return 0;}
   /** return class values that should lie inside the model
       @param _nin number of inner class values
       @return array of inner class values */
   const unsigned char *getInArray(unsigned int& _nin) const {
     _nin=nin;
     return in;}
   /** return class values that should lie inside the model
       @param _nout number of outer class values
       @return array of outer class values */
   const unsigned char *getOutArray(unsigned int& _nout) const {
     _nout=nout;
     return out;}
   /** set class values that should lie inside the model
       @param _nin number of inner class values
       @param _in of inner class values */
   void setInArray(unsigned int _nin, const unsigned char *_in)
     {
       nin = _nin;
       if(in) delete[] in;
       in = new unsigned char[nin];
       for(unsigned int i = 0; i < nin; i++) in[i] = _in[i];
     }
   /** set class values that should lie outside the model
       @param _nout number of outter class values
       @param _out of outter class values */
   void setOutArray(unsigned int _nout, const unsigned char *_out)
     {
       nout = _nout;
       if(out) delete[] out;
       out = new unsigned char[nout];
       for(unsigned int i = 0; i < nout; i++) out[i] = _out[i];
     }

   friend class ClassForce;
 };

/** \class BlockMatchingInformation
    \ingroup ImageData
    a class that store all information necessary for computing an external
    force based on the block matching algorithm */
class BlockMatchingInformation : public RangeInformation {
 public:
  /// name stored in the Range Information
  static const  TclString informationName;
  /// supported similarity measures
  enum SimilarityMeasure {
    /// correlation coefficient
    CORRELATION_COEFFICIENT,
    /// correlation ratio
    CORRELATION_RATIO,
    /// sum of squared difference
    SSD,
    /// paired t-test measure + Student distribution
    PAIREDTTEST_STUDENT
  };
  
 protected:
  /// reference image to compute the block matching
  Inrimage * refImage;
  /// central voxel of the block in the reference image
  Vec3<unsigned int> blockCenter;
  /// half of reference image block size
  Vec3<unsigned int> blockSize;
  /// half search block range size
  Vec3<unsigned int> searchRange;
  /// search step resolution in the search block
  Vec3<unsigned int> searchStep;
  /// similarity measure used
  SimilarityMeasure similarityMeasure;
  /// mean of the reference block
  double refMean;
  /// variance of the reference block
  double refVariance;
  /// is the reference block inside the reference image
  bool insideRef;
  /// apply a force along the normal
  bool alongNormal;
  /// offset between the vertex position and the voxel center
  Vec3<double> offset;

 public:
  /** @name constructor */
  //@{
  /**  builds a data structure for a force computed with block matching  
   */
  BlockMatchingInformation(Inrimage * refim, Vec3<unsigned int> center, Vec3<unsigned int> bs, Vec3<unsigned int> sr, Vec3<unsigned int> st, SimilarityMeasure sm, Vec3<double> os = Vec3<double>());
  //@}
  
  /** @name destructor */
  //@{
  /// destroys the object
  virtual ~BlockMatchingInformation(){}
  //@}
  inline void setBlockSize(const Vec3<unsigned int> bs) {
    blockSize=bs;
  }
  inline Vec3<unsigned int> getBlockSize(void) const {
    return(blockSize);
  }
  inline bool isAlongNormal(void) const {
    return(alongNormal);
  }
  inline bool isInsideRef(void) const {
    return(insideRef);
  }
  inline void setBlockCenter(const Vec3<unsigned int> bc) {
    blockCenter=bc;
  }
  inline Vec3<unsigned int> getBlockCenter(void) const {
    return(blockCenter);
  }
  inline void setSearchStep(const Vec3<unsigned int> s) {
    searchStep=s;
  }
  inline Vec3<unsigned int> getSearchStep(void) const {
    return(searchStep);
  }
  inline void setSearchRange(const Vec3<unsigned int> r) {
    searchRange=r;
  }
  inline Vec3<unsigned int> getSearchRange(void) const {
    return(searchRange);
  }
  inline Vec3<double> getOffset(void) const {
    return(offset);
  }
  inline void setSimilarityMeasure(const BlockMatchingInformation::SimilarityMeasure s) {
    similarityMeasure=s;
  }
  inline BlockMatchingInformation::SimilarityMeasure getSimilarityMeasure(void) const {
    return(similarityMeasure);
  }
  
  /// get the reference image
  inline Inrimage * getRefImage(void) const {
    return(refImage);
  }
  /// set the reference image
  void setRefImage(Inrimage* im) {
    refImage = im;
  }
   /// get the reference mean
  inline double getRefMean(void) const {
    return(refMean);
  }
  /// set the reference mean
  void setRefMean(const double m) {
    refMean = m;
  }
     /// get the reference variance
  inline double getRefVariance(void) const {
    return(refVariance);
  }
  /// set the reference variance
  void setRefVariance(const double m) {
    refVariance = m;
  }
  void computeRefBlockStats(void);

  void init(void);
  
  /// create a block matching information Object from a file
  static RangeInformation *readBlockMatchingInformation(GFInput &input);
  /** create a copy of this region information object. This is used to apply 
      the same algorithm to all vertices of a mesh */
  virtual RangeInformation* duplicate() {
    return new BlockMatchingInformation(*this); }
  /** write in a file the data stored in a block matching information object
      @param output a pointer on a file */
  virtual void writeInFile(GFOutput &output);
};
  
/** \class IntensityProfileInformation
    \ingroup ImageData
    \brief an intensity profile stored at a vertex */
class IntensityProfileInformation : public RangeInformation {
 public:
  /// name stored in the Range Information
  static const  TclString informationName;

  /// registration evaluation criterion
  enum Criterion {
    /// minimize sum |x - y|
    DIFFERENCES=0,
    /// maximise [sum xy]^2 / [sum x^2 sum y^2]
    LINEAR=1,
    /// maximise [sum (x - E(x))(y - E(y))]^2 / [n^2 Var(x) Var(y)]
    CORRELATION=2,
    /// maximise sum s^2 / (s^2 + (Idiff(i, j) - Idiff(v, w))^2)
    PATTERN_INTENSITY=3
  };
  
 private:
  /// an intensity element
  struct element {
    /// constructor
    element(double d, double v) : distance(d), value(v) {}
    /// distance from profile origin
    double distance;
    /// intensity value at that point
    double value;
  };
  /// intensity values
  std::vector<element> profile;
  /// profile origin
  double origin;
  /// registration evaluation criterion
  Criterion criterion;
  /// criterion match threshold
  double threshold;
  /// the half length of the image line to be matched with the stored profile
  double range;
 public:
  /** @name constructors */
  //@{
  /** constructs a profile representing given image line
      @param line profile line in image 
      @param origin profile origin */
  IntensityProfileInformation(const IntensityProfile& line, double origin);
  /** constructs a profile from values array
      @param size profile number of points
      @param origin profile origin
      @param distance size elements array of points distance
      @param intensity  size elements array of points intensity 
      @param criterion the criterion used to find the matching point
      @param threshold */
  IntensityProfileInformation(unsigned int size, double origin, double *distance,
		   double *intensity, Criterion criterion=DIFFERENCES,
		   double threshold=50.0);
  //@}
  /** @name destructor */
  //@{
  /// destroys profile intensity values
  ~IntensityProfileInformation();
  //@}

  /// get the maximum range distance 
  double getMaxDistance(void) const {
    return (range);
  }
  /// set the maximum range distance 
  void setMaxDistance(double r) {
    range = r;
  }
  /// returns number of sampled vector values
  unsigned int getSize(void) const { return profile.size(); }
  /// returns profile origin
  double getOrigin(void) const { return origin; }

  /** returns a sampled intensity value
      @param i sampled value index
      @return intensity value or 0 if index falls out of profile */
  double getIntensity(unsigned int i) const {
    if(i < profile.size()) return profile[i].value;
    return 0.0;
  }

  /** returns a sampled point distance
      @param i sampled value index
      @return ith point distance from profile origin or 0 if index falls out of profile */
  double getDistance(unsigned int i) const {
    if(i < profile.size()) return profile[i].distance;
    return 0.0;
  }

  /** returns a linear interpolation of intensity value at given distance
      @param d euclidian point distance from profile origin
      @return intensity value at distance d or 0.0 if d falls out of profile */
  double getIntensity(double d) const;

  /// returns registration evaluation criterion
  Criterion getCriterion(void) const { return criterion; }
  /** set registration evaluation criterion
      @param criterion criterion to use
      @param threshold criterion threshold */
  void setCriterion(Criterion criterion, double threshold = -1.0);
  /// return the threshold related to the image criterion
  double getThreshold() const {
    return threshold;}
  /// set the threshold related to the image criterion
  void setThreshold(const double t)  {
    threshold=t;}
  /// set the profile origin
  void setOrigin(const double o)  {
    origin=o;}

  /** computes attractive point in given image line using local
      information
      @param line image line
      @param point returned attractive point
      @return true if a point was found in image line, false otherwise */
  virtual bool getPoint(const IntensityProfile& line, Vec3<double>& point);
  /// create an IntensityProfile Object from a file
  static RangeInformation *readIntensityProfile(GFInput &input);
  /** create a copy of this IntensityProfile object. This is used to apply 
      the same algorithm to all vertices of a mesh */
  virtual RangeInformation* duplicate();
  /** write in a file the data stored in an Intensity Profile object
      @param output a pointer on a file */
  virtual void writeInFile(GFOutput &output);
 
};

/** \class RangeInrimage
    \ingroup ImageData
    \brief base class for 2D and 3D image used to deform models */
class RangeInrimage : public Owner<Inrimage>, public GeneralFormat {
 public:
  /** @name exception */
  //@{
  /** \class InvalidSize
      \ingroup ImageData
  */
  class InvalidSize : public Exception {
  public:
    InvalidSize(const std::string& from);
  };
  //@}

  /** \class Force
      \ingroup ImageData
  */
  class Force : public ListElement<Force>{
  public:
    /** @name exception thrown when the associated RangeInrimage is not valid*/
    //@{
    /** \class InvalidRangeInrimage
        \ingroup ImageData
    */
    class InvalidRangeInrimage : public Exception {
    public:
      InvalidRangeInrimage(const std::string& from);
    };
    //@}

    /** @name exception */
    //@{
    /** \class InvalidImages
        \ingroup ImageData
    */
    class InvalidImages : public Exception {
    public:
      InvalidImages(const std::string& from);
    };
    //@}

  private:
    /// the force name
    TclString name;
  protected:
    /// the associated Range image
    RangeInrimage *range;

  public:

    /** @name constructor */
    //@{
    /** builds a new force applying on a volumetric image
	@param range the associated range inrimage
	@param _name the force name
	@exception InvalidRangeInrimage range pointer should not be 0
    */
    Force(RangeInrimage *range, TclString _name) throw(InvalidRangeInrimage);
    //@}

    /** @name destructor */
    //@{
    /// empty
    virtual ~Force() {}
    //@}

    /// whether the force can be computed
    virtual bool isPossible() const = 0;


    /** routines that computes a level set speed at a given point in image
	@param position the position of a 3D point in the image
	@param mult returned multiplicative speed term
	@param add returned additive speed term */
    virtual void getSpeed(const Vec3<double> &position,
			  double &mult, double &add) {
      mult = 1.0;
      add = 0.0;
    }

    /// returns the name of the force
    TclString getName() { return(name); }
    /// function called whenever an image of the pointed RangeImage is changed 
    virtual void updateImage() {}
  };




/** \class GradientForce
    \ingroup ImageData
    \brief a force that detects voxel of maximum gradient norm */
class GradientForce :  public VoxelMax, public Force {
 protected:
  /// use this geometry converter to properly handle cylindrical images
  ImageGeometry *converter;
  /// gradient Inrimage Information
  GradientInrimageInformation *gi;
  /// average voxel size
  double vsize;

  /** returns true if voxel verify orientation and intensity constraints, false otherwise
      @param voxel tested voxel */
  bool testVoxel(ImageVoxel *voxel);
  
 public:
  /** @name constructor */
  //@{
  /**  builds a gradient force given the gradient threshold, its orientation constraint and its intensity range 
       @param range the associated range inrimage
       @exception InvalidRangeInrimage range pointer should not be 0 */
  GradientForce(RangeInrimage *range) 
    throw(Force::InvalidRangeInrimage);
  //@}
  
  /** @name destructor */
  //@{
  /// destroys any line voxel
  virtual ~GradientForce();
  //@}
  
  /** initializes the closest voxel
      @param voxel an image type voxel */
  void start(Voxel& voxel);
  
  /// whether this force can be used
  bool isPossible() const ;
  
  /** routines that computes a point in a volumetric image lying on a 3D line defined by (position, direction) and that that best matches a given criterion
      @param position the position of a 3D point in the image
      @param direction a unit vector representing a direction of a 3D line in the image
      @param range the maximum distance from the 3D point (position) where to search for that point
      @param point the matched point in the 3D image that optimized a given criterion
	  @param confidence a confidence value (>=0) that is proportional to the confidence in finding the closest point
      @param rii local range information at given point
      @return a boolean that is set to true is a point is found that matches a given criterion
  */
  bool getPoint(const Vec3<double> &position,const Vec3<double> &direction,
		double range,Vec3<double> &point,double &confidence,
		GradientInrimageInformation *rii = 0);
  
  /// set the image pointer to the updated gradient image
  void updateImage(void);
  /// is it possible to use the force
  bool isPossible();
  /** gradient force stop condition
      @param v current image line voxel
      @param iv an image type voxel */
  bool stopFromGradient(ImageVoxel *v, Voxel& iv);
};

/** \class IsosurfaceForce
    \ingroup ImageData
    \brief a force that detects a voxel whose intensity corresponds to a given isovalue */
class IsosurfaceForce :  public VoxelMax, public Force {
 protected:
  /// gradient Inrimage Information
  IsosurfaceInrimageInformation *isi;
  /// the sigma to compute the gradient information (for oomputing the confidence) */
  double sigma;

  /** returns true if the intensity profile intersects the isosurface , false otherwise
      @param d the distance from the origin where the closest point on the isosurface lies */
  bool findClosestIsosurfaceVoxel(double &d);
  /** returns true if the intensity profile intersects the isosurface , false otherwise.
  The intersected isosurface voxel is constrained to correspond to a transition + -> - or - -> + starting from the inside or outside voxels 
  @param d the distance from the origin where the closest point on the isosurface lies 
  @param constraint the type of isosurface constraint */
  bool findRegionIsosurfaceVoxel(double &d,IsosurfaceInrimageInformation::IsosurfaceConstraint constraint);
  
 public:
  /** @name constructor */
  //@{
  /**  builds a gradient force given the gradient threshold, its orientation constraint and its intensity range 
       @param range the associated range inrimage
       @exception InvalidRangeInrimage range pointer should not be 0 */
  IsosurfaceForce(RangeInrimage *range) 
    throw(Force::InvalidRangeInrimage);
  //@}
  
  /** @name destructor */
  //@{
  /// destroys any line voxel
  virtual ~IsosurfaceForce();
  //@}
  
  /** initializes the closest voxel
      @param voxel an image type voxel */
  void start(Voxel& voxel);
  
  /// whether this force can be used
  bool isPossible() const ;
  
  /** routines that computes a point in a volumetric image lying on a 3D line defined by (position, direction) and that that best matches a given criterion
      @param position the position of a 3D point in the image
      @param direction a unit vector representing a direction of a 3D line in the image
      @param range the maximum distance from the 3D point (position) where to search for that point
      @param point the matched point in the 3D image that optimized a given criterion
	  @param confidence a confidence value (>=0) that is proportional to the confidence in finding the closest point
      @param rii local range information at given point
      @return a boolean that is set to true is a point is found that matches a given criterion
  */
  bool getPoint(const Vec3<double> &position,const Vec3<double> &direction,
		double range,Vec3<double> &point,double &confidence,
		IsosurfaceInrimageInformation *rii = 0);
  
  /// set the image pointer to the updated gradient image
  void updateImage(void);
  /// is it possible to use the force
  bool isPossible();
  /** isosurface force stop condition
      @param v current image line voxel
      @param iv an image type voxel */
  bool stopFromDistance(ImageVoxel *v, Voxel& iv);
};


/** \class RegionForce
    \ingroup ImageData
    a force that detects the closest voxel whose gradient value is above a
    given threshold and eventually whose gradient orientation follows the
    gradient orientation constraint and finally whose intensity belongs to a
    given intensity range */
class RegionForce :  public VoxelMax, public Force  {
  private:
  /// necessary information for computing matching point
  RegionInrimageInformation *ri;
  protected:
    /** Finds the region boundary and store the boundary voxel in maxVoxel
	@return the gradient value at the best point*/
    double  findRegion();
    /** Smooth the image profile */
    void smoothProfile();
	/// average voxel size
	double vsize;
    
  public:
    /** @name constructor */
    //@{
    /**  builds an edge force given  the gradient threshold and its orientation constraint and its intensity range 
	 @param range the associated range inrimage
	 @exception InvalidRangeInrimage range pointer should not be 0 */
    RegionForce(RangeInrimage *range)
      throw(Force::InvalidRangeInrimage);
    //@}

    /** @name destructor */
    //@{
    /// destroys any line voxel
    virtual ~RegionForce();
    //@}

    /** initializes the closest voxel
	@param voxel an image type voxel */
    void start(Voxel& voxel);

    /// whether this force can be used
    bool isPossible() const ;

    /** routines that computes a point in a volumetric image lying on a 3D line defined by (position, direction) and that that best matches a given criterion
      @param position the position of a 3D point in the image
      @param direction a unit vector representing a direction of a 3D line in the image
      @param range the maximum distance from the 3D point (position) where to search for that point
      @param point the matched point in the 3D image that optimized a given criterion
	  @param confidence a confidence value (>=0) that is proportional to the confidence in finding the closest point
      @param rii local range information at given point
      @return a boolean that is set to true is a point is found that matches a given criterion
  */
	bool getPoint(const Vec3<double> &position,const Vec3<double> &direction,
		double range,Vec3<double> &point,double &confidence,
		RegionInrimageInformation *rii = 0);

    /// set the image pointer to the updated gradient image
    void updateImage(void);
    /** edge force stop condition
	@param v current image line voxel
	@param iv an image type voxel */
    bool stopFromRegion(ImageVoxel *v, Voxel& iv);
  };



/** \class RegionProbabilityMapForce
    \ingroup ImageData
    a force that detects the closest voxel whose gradient value is above a
    given threshold and eventually whose gradient orientation follows the
    gradient orientation constraint and finally whose intensity belongs to a
    given intensity range */
class RegionProbabilityMapForce :  public VoxelMax, public Force  {
  private:
  /// necessary information for computing matching point
  RegionProbabilityMapInformation *ri;
  protected:
    /** Finds the region boundary and store the boundary voxel in maxVoxel */
    void findRegion();
    
  public:
    /** @name constructor */
    //@{
    /**  builds an edge force given  the gradient threshold and its orientation constraint and its intensity range 
	 @param range the associated range inrimage
	 @exception InvalidRangeInrimage range pointer should not be 0 */
    RegionProbabilityMapForce(RangeInrimage *range)
      throw(Force::InvalidRangeInrimage);
    //@}

    /** @name destructor */
    //@{
    /// destroys any line voxel
    virtual ~RegionProbabilityMapForce();
    //@}

    /** initializes the closest voxel
	@param voxel an image type voxel */
    void start(Voxel& voxel);

    /// whether this force can be used
    bool isPossible() const ;

    /** routines that computes a point in a volumetric image lying on a 3D line defined by (position, direction) and that that best matches a given criterion
      @param position the position of a 3D point in the image
      @param direction a unit vector representing a direction of a 3D line in the image
      @param range the maximum distance from the 3D point (position) where to search for that point
      @param point the matched point in the 3D image that optimized a given criterion
	  @param confidence a confidence value (>=0) that is proportional to the confidence in finding the closest point
      @param rii local range information at given point
      @return a boolean that is set to true is a point is found that matches a given criterion
  */
	bool getPoint(const Vec3<double> &position,const Vec3<double> &direction,
		double range,Vec3<double> &point,double &confidence,
		  RegionProbabilityMapInformation *rii = 0);

    /// set the image pointer to the updated gradient image
    void updateImage(void);
    /** edge force stop condition
	@param v current image line voxel
	@param iv an image type voxel */
    bool stopFromRegion(ImageVoxel *v, Voxel& iv);
  };


/** \class GradientGradientForce
    \ingroup ImageData
    the force that apply a force in the direction of the gradient vector
    of the gradient   */
  class GradientGradientForce : public Force {
  public:
    /** @name constructor */
    //@{
    /**  builds a GradientGradient force
	 @param _range the associated range inrimage 
	 @exception InvalidRangeInrimage range pointer should not be 0 */
    GradientGradientForce(RangeInrimage *_range)
      throw(Force::InvalidRangeInrimage);
    //@}

    /** @name destructor */
    //@{
    /// destroys any line voxel
    virtual ~GradientGradientForce() {}
    //@}
    

    /// whether this force can  be computed
    bool isPossible() const;
    /** routines that computes a point in a volumetric image lying on a 3D line defined by (position, direction) and that that best matches a given criterion
      @param position the position of a 3D point in the image
      @param direction a unit vector representing a direction of a 3D line in the image
      @param range the maximum distance from the 3D point (position) where to search for that point
      @param point the matched point in the 3D image that optimized a given criterion
	  @param confidence a confidence value (>=0) that is proportional to the confidence in finding the closest point
      @param rii local range information at given point
      @return a boolean that is set to true is a point is found that matches a given criterion
  */
	bool getPoint(const Vec3<double> &position,const Vec3<double> &direction,
		double range,Vec3<double> &point,double &confidence,
		  GradientGradientInrimageInformation *rii = 0);
  };


/** \class GradientBalloonForce
    \ingroup ImageData
    the force that moves the deformable model in the normal direction only of
    the gradient value is less than a given threhold  */
  class GradientBalloonForce : public Force {
  public:
    /** @name constructor */
    //@{
    /**  builds a GradientGradient force
	 @param _range the associated range inrimage 
	 @exception InvalidRangeInrimage range pointer should not be 0 */
    GradientBalloonForce(RangeInrimage *_range)
      throw(Force::InvalidRangeInrimage);
    //@}
    /// whether this force can  be computed
    bool isPossible() const;
    /** routines that computes a point in a volumetric image lying on a 3D line defined by (position, direction) and that that best matches a given criterion
      @param position the position of a 3D point in the image
      @param direction a unit vector representing a direction of a 3D line in the image
      @param range the maximum distance from the 3D point (position) where to search for that point
      @param point the matched point in the 3D image that optimized a given criterion
	  @param confidence a confidence value (>=0) that is proportional to the confidence in finding the closest point
      @param rii local range information at given point
      @return a boolean that is set to true is a point is found that matches a given criterion
  */
	bool getPoint(const Vec3<double> &position,const Vec3<double> &direction,
		double range,Vec3<double> &point,double &confidence,
		  GradientBalloonInrimageInformation *rii = 0);

    /** compute level set speed as 1 /|\ nabla I| truncated function
	@param position the position of a 3D point in the image
	@param mult returned multiplicative speed term
	@param add returned additive speed term 
	@param rii local information necessary to compute the speed */
    virtual void getSpeed(const Vec3<double> &position,
			  double &mult, double &add,
			  GradientBalloonInrimageInformation *rii);

    /** @name destructor */
    //@{
    /// destroys any line voxel
    virtual ~GradientBalloonForce(){}
    //@}
    
  };


  /** \class IntensityProfileForce
      \ingroup ImageData
      this force deforms the model so that pre-stored intensity profile
      at each mesh vertex match the image intensity profile */
  class IntensityProfileForce : public IntensityProfile,  public Force {
  public:
    /** @name constructor */
    //@{
    /** builds a profile force
	@param range range inrimage the force is computed in
	@exception InvalidRangeInrimage range pointer should not be 0 */
    IntensityProfileForce(RangeInrimage *range)
      throw(Force::InvalidRangeInrimage);
    //@}

    /** @name destructor */
    //@{
    /// destroys any line voxel
    virtual ~IntensityProfileForce() {}
    //@}
    
    /// whether the force can be computed
    virtual bool isPossible() const { return true; }

    /** routines that computes a point in a volumetric image lying on a 3D line defined by (position, direction) and that that best matches a given criterion
      @param position the position of a 3D point in the image
      @param direction a unit vector representing a direction of a 3D line in the image
      @param range the maximum distance from the 3D point (position) where to search for that point
      @param point the matched point in the 3D image that optimized a given criterion
	  @param confidence a confidence value (>=0) that is proportional to the confidence in finding the closest point
      @param rii local range information at given point
      @return a boolean that is set to true is a point is found that matches a given criterion
  */
	bool getPoint(const Vec3<double> &position,const Vec3<double> &direction,
		double range,Vec3<double> &point,double &confidence,
		  IntensityProfileInformation *rii);
  };


  /** \class ClassForce
      \ingroup ImageData
      this force deforms the model in a classified image at the interface
      of two classes */
  class ClassForce :  public IntensityProfile, public Force {
  private:
    /// necessary information for computing matching point
    ClassInrimageInformation *ri;
  public:
    /** @name constructor */
    //@{
    /** builds a class force
	@param range range inrimage the force is computed in
	@param forceRange range to scan voxels for an appropriate candidate
	@exception InvalidRangeInrimage range pointer should not be 0 */
    ClassForce(RangeInrimage *range, double forceRange = 10.0)
      throw(Force::InvalidRangeInrimage);
    //@}

    /** @name destructor */
    //@{
    /// destroys any line voxel
    virtual ~ClassForce() {}
    //@}

    /// whether the force can be computed
    virtual bool isPossible() const;

    /** routines that computes a point in a volumetric image lying on a 3D line defined by (position, direction) and that that best matches a given criterion
      @param position the position of a 3D point in the image
      @param direction a unit vector representing a direction of a 3D line in the image
      @param range the maximum distance from the 3D point (position) where to search for that point
      @param point the matched point in the 3D image that optimized a given criterion
	  @param confidence a confidence value (>=0) that is proportional to the confidence in finding the closest point
      @param rii local range information at given point
      @return a boolean that is set to true is a point is found that matches a given criterion
  */
	bool getPoint(const Vec3<double> &position,const Vec3<double> &direction,
		double range,Vec3<double> &point,double &confidence,
		  ClassInrimageInformation *rii);

    /** set the image pointer on the updated classify image */
    void updateImage(void);
  };
  
  /** \class BlockMatchingForce
      \ingroup ImageData
      the force that moves the deformable model in the direction given by the
      block matching algorithm */
  class BlockMatchingForce : public Force {
  public:
    /** @name constructor */
    //@{
    /**  builds a BlockMatching force
	 @param _range the associated range inrimage 
	 @exception InvalidRangeInrimage range pointer should not be 0 */
    BlockMatchingForce(RangeInrimage *_range)
      throw(Force::InvalidRangeInrimage);
    //@}
    /// whether this force can  be computed
    bool isPossible() const;
    /** routines that computes a point in a volumetric image that best matches a given criterion
      @param position the position of a 3D point in the image
      @param direction a unit vector representing a direction of a 3D line in the image
      @param range the maximum distance from the 3D point (position) where to search for that point
      @param point the matched point in the 3D image that optimized a given criterion
	  @param confidence a confidence value (>=0) that is proportional to the confidence in finding the closest point
      @param rii local range information at given point
      @return a boolean that is set to true is a point is found that matches a given criterion
  */
	bool getPoint(const Vec3<double> &position,const Vec3<double> &direction,
		double range,Vec3<double> &point,double &confidence,
		  BlockMatchingInformation *rii);

    /** @name destructor */
    //@{
    /// destroys any line voxel
    virtual ~BlockMatchingForce(){}
    //@}
    
  };
  
 public:
  /// the geometry of attraction forces
  enum ForceGeometry { FORCE_2D, FORCE_3D};

 protected:
  /// intensity image
  ObjectPointer<Inrimage> intensity; 
  /// gradient image
  ObjectPointer<Inrimage> gradient; 
  /// gradient direction image
  ObjectPointer<Inrimage> gradientDirection;
  /// gradient direction image
  ObjectPointer<Inrimage> edge;
  /// gradient direction image
  ObjectPointer<Inrimage> gradientGradient;
  /// classification image
  ObjectPointer<Inrimage> classify;
  /// array of vertex corresponding to coding of the gradientDirection image
  Vec3<double> *sampledDirection;

  /// the size in the X direction of all images
  unsigned int sizeX;
  /// the size in the Y direction of all images
  unsigned int sizeY;
  /// the size in the Z direction of all images
  unsigned int sizeZ;

  /// list of available forces
  ObjectsList<Force> forceList;
  /// the geometry of attraction forces
  ForceGeometry forceGeometry;
  /// gradient force
  GradientForce *gradientForce;
  /// region force
  RegionForce *regionForce;
  /// isosurface force
  IsosurfaceForce *isosurfaceForce;
  /// gradientGradient force
  GradientGradientForce *gradientGradientForce;
  /// gradient balloon force
  GradientBalloonForce *gradientBalloonForce;
  /// intensityProfile force
  IntensityProfileForce *intensityProfileForce;
  /// classify force
  ClassForce *classForce;
  /// block matching force
  BlockMatchingForce *blockMatchingForce;



 protected:
  /** an image owned by this object is being destroyed
      @param obj element before destruction */
  virtual void dying(const ObjectPointer<Inrimage>& obj);
  /// set sizeX sizeY sizeZ as the size of a given image
  void setSize(Inrimage *im);
  /// test if  im has the size corresponding to sizeX sizeY sizeZ 
  void testSize(Inrimage *im) const throw (InvalidSize);
  /** looks for gradient file from intensity file name
      @param fileName intensity file name
      @param gradientName returned gradient file name, EMPTY_STRING if no gradient file was found */
  void gradientFile(const TclString& fileName, TclString& gradientName) const;
  /** looks for gradient direction file from intensity file name
      @param fileName intensity file name
      @param gradientDirectionName returned gradient direction file name, EMPTY_STRING if no gradient file was found */
  void gradientDirectionFile(const TclString& fileName, 
			     TclString& gradientDirectionName) const;
  /** looks for gradient of the gradient file from intensity file name
      @param fileName intensity file name
      @param gradientGradientName returned gradient of the gradient norme file name, EMPTY_STRING if no gradient file was found */
  void gradientGradientFile(const TclString& fileName, 
			    TclString& gradientGradientName) const;
  /** looks for classify file from intensity file name
      @param fileName intensity file name
      @param classifyName returned classify file name, EMPTY_STRING if no classify file was found */
  void classifyFile(const TclString& fileName, TclString& classifyName) const;
  /// call the updateImage method of all existing force in its lists
  void updateForceImages();

  /// write the main description of a RangeInrimage
  void writeMain(GFOutput &output);
  /// read the main description of a RangeInrimage
  void readMain(GFInput &input);
  /// skip the main description of a RangeInrimage
  void skipMain(GFInput &input);
 protected:
  /// init general format function
  void initFormat();
 public:
  /** @name constructor */
  //@{
  /** Builds a RangeInrimage given the intensity image and its name.
      If name.inr.gz is given, name.gradient and name.gradient.gz are
      used for gradient image candidates, and name.gradient_direction
      and name.gradient_direction.gz are used for gradient direction
      image candidate.
      @param filename intensity image file name
      @param intensity intensity image (loaded if null)
      @exception Inrimage::ImageNotFound file not found 
      @exception Inrimage::UnknownImageType unknown image file format
      @exception Inrimage::InvalidHeader invalid inrimage header
      @exception Inrimage::InvalidData invalid inrimage data
      @exception Inrimage::UnknownType unhandled inrimage type */
  RangeInrimage(const TclString& filename,
		Inrimage *intensity = 0)
    throw(Inrimage::ImageNotFound, Inrimage::UnknownImageType,
	  Inrimage::InvalidHeader, Inrimage::InvalidData,
	  Inrimage::UnknownType);
  /** build a RangeInrimage given the list of images
      @param intensity the intensity imag
      @param gradient the image of the norm of the gradient
      @param gradientDirection the image of gradient direction coded in a unsigned char
      @param edge the edge image coded as a binary unsigned char image
      @param gradientGradient the image corresponding to the gradient of the gradient norm
      @param classify the image corresponding to the classification of intensity image
      @param _forceGeometry if it should be a 2D or 3D force */
  RangeInrimage(Inrimage *intensity = 0,
		Inrimage *gradient = 0,
		Inrimage *gradientDirection = 0,
		Inrimage *edge = 0,
		Inrimage *gradientGradient = 0,
		Inrimage *classify = 0,
		ForceGeometry _forceGeometry = FORCE_3D);
  //@}

  /** @name destructor */
  //@{
  /// destroys any line voxel
  virtual ~RangeInrimage();
  //@}

  /// creates the set of forces and adds then in the list
  void initForces();
  /** unit vectors of the spherical dome number when the unit sphere is
      tesselated with frequency DEFAULT_SPHERE_FREQUENCY */
  unsigned int getDirectionArraySize() {
    return(20*DEFAULT_SPHERE_FREQUENCY*DEFAULT_SPHERE_FREQUENCY);}

  /** set a new intensity image
      @param inr intensity image
      @exception InvalidSize inr size does not match data image size */
  void setIntensity(Inrimage *inr) throw(RangeInrimage::InvalidSize);
  /** set a new gradient image
      @param inr gradient image
      @exception InvalidSize inr size does not match data image size */
  void setGradient(Inrimage *inr) throw(RangeInrimage::InvalidSize);
  /** set a new gradient direction image
      @param inr gradient direction image
      @exception InvalidSize inr size does not match data image size */
  void setGradientDirection(Inrimage *inr) throw(RangeInrimage::InvalidSize);
  /** set a new edge image
      @param inr edge image
      @exception InvalidSize inr size does not match data image size */
  void setEdge(Inrimage *inr) throw(RangeInrimage::InvalidSize);
  /** set a new gradient of the gradient norm image
      @param inr gradient of the gradient norm image
      edge image
      @exception InvalidSize inr size does not match data image size */
  void setGradientGradient(Inrimage *inr) throw(RangeInrimage::InvalidSize);
  /** set a new classify image
      @param inr classify image
      @exception InvalidSize inr size does not match data image size */
  void setClassify(Inrimage *inr) throw(RangeInrimage::InvalidSize);
  /** get the intensity image
      @return the intensity image */
  Inrimage *getIntensity(void) const {
    return(intensity);}
  /** get the gradient image
      @return the gradient image */
  Inrimage *getGradient(void) const {
    return(gradient);  }
  /** get the gradientDirection image
      @return the gradientDirection image */
  Inrimage *getGradientDirection(void) const {
    return(gradientDirection);  }
  /** get the edge image
      @return the edge image */
  Inrimage *getEdge(void) const {
    return(edge);  }
  /** get the gradientGradient image
      @return the gradientGradient image */
  Inrimage *getGradientGradient(void) const {
    return(gradientGradient);  }
  /** get the classify image
      @return the gradientGradient image */
  Inrimage *getClassify(void) const {
    return(classify);  }
  /** get the gradientGradient image
      @return the gradientGradient image */
  const Vec3<double> *getSampledDirection(void) const {
    return(sampledDirection);  }
  /** compute the gradient norm and direction using sobel filters : it requires that the intensity image already exists
      @param sigma the width of the line recursive filter
      @param twoD whether the computation should be 2D or 3D
      @param maximum whether the maximum of the gradient in the gradient direction should be computed
      @param lt the low threshold on the gradient norm */
  void computeSobelGradient(unsigned int sigma=3,bool twoD=false,
			    bool maximum=false,double lt=0.0);
  /** compute the gradient norm and direction using recursive filters  : it requires that the intensity image already exists
      @param filter the type of recursive filter to apply
      @param sigma the width of the filter
      @param twoD whether the computation should be 2D or 3D
      @param maximum whether the maximum of the gradient in the gradient direction should be computed
      @param lt the low threshold on the gradient norm */
  void computeRecursiveGradient(yav::RECURSIVE_FILTER_TYPE filter=yav::DERICHE_STEP_EDGE,
				double sigma=3.0,bool twoD=false,
				bool maximum=false,double lt=0.0);
  /** compute the gradient of the gradient norm using sobel filters : it requires that the gradient image already exists
      @param sigma the width of the line recursive filter
      @param twoD whether the computation should be 2D or 3D */
  void computeSobelGradientGradient(unsigned int sigma=3,bool twoD=false);
  /** compute the gradient of the gradient norm using recursive filters : it requires that the gradient image already exists
      @param filter the type of recursive filter to apply
      @param sigma the width of the filter
      @param twoD whether the computation should be 2D or 3D
  */
  void computeRecursiveGradientGradient(
    yav::RECURSIVE_FILTER_TYPE filter=yav::DERICHE_STEP_EDGE,
    double sigma=3.0,
    bool twoD=false);


  friend class TclRangeInrimage;
};

} // namespace yav

#endif
