#ifndef RANGEINRIMAGE2D_H
#define RANGEINRIMAGE2D_H


#ifdef _MSC_VER
#pragma warning ( disable : 4068 )
#pragma warning ( disable : 4786 )
#pragma warning ( disable : 4081 )
#endif

#pragma interface "imagedata/RangeInrimage2D.h"

#include <imagedata/RangeInrimage.h>
#include <framework/SliceOrientation.h>

namespace yav {


/** \class RangeInrimage2D
    \ingroup ImageData
    \brief 2D range inrimage */
class RangeInrimage2D :  public RangeInrimage,
			 Pointed<RangeInrimage2D>
{
 public:
  
  /** @name constructors */
  //@{
  /** builds a range inrimage from intensity image
      @param inr intensity inrimage
      @param scene scene to create range image in
      @param gradient gradient image
      @param gradientdir gradient direction image
      @param edge edge image
      @param gradient2 gradient derivative image
      @param classified classified image */
  RangeInrimage2D(Inrimage *inr, 
		  Inrimage *gradient = 0, Inrimage *gradientdir = 0,
		  Inrimage *edge = 0, Inrimage *gradient2 = 0,
		  Inrimage *classified = 0)
    throw(Inrimage::ImageNotFound, Inrimage::UnknownImageType,
	  Inrimage::InvalidHeader, Inrimage::InvalidData,
	  Inrimage::UnknownType);

  /** Builds a range inrimage from intensity image file name.
      Given file name is used as intensity image.
      If name.inr.gz is given, name.gradient and name.gradient.gz are
      used for gradient image candidates, and name.gradient_direction
      and name.gradient_direction.gz are used for gradient direction
      image candidate.
      @param fileName intensity image file name
      @param scene scene to create range image in */
  RangeInrimage2D(const TclString& fileName)
    throw(Inrimage::ImageNotFound, Inrimage::UnknownImageType,
	  Inrimage::InvalidHeader, Inrimage::InvalidData,
	  Inrimage::UnknownType);
  //@}
  
  /** @name destructor */
  //@{
  /// empty
  virtual ~RangeInrimage2D();
  //@}
  


  /** RangeData2D closest point computation callback
      @param position vertex position
      @param normal vertex normal
      @param range search range
      @param closest returned closest data point found
	  @param confidence value (>=0) that provides confidence in the search for the closest point
      @param clientData vertex local information
      @return true if some data point was found, false otherwise */
  bool getClosestPoint (const Vec2<double> &position,
			const Vec2<double> &normal,
			const double range,
			Vec2<double> &closest,double &confidence,
			const RangeInformation *clientData) const {
    return false; }
  
  /** routine that computes the closest point on a 3D image given the slice number and slice geometry
      @param position the position of the point from which the closest point should be computed 
      @param normalDirection the normal direction of the point from which the closest point should be computed 
      @param rangeSearchDistance the maximum distance (= cut-off distance) along the normal direction at which we should search for the closest point
      @param closestPoint the closestPoint of the dataset from a given position and normal direction. This output value is only set if a closest point were found. Otherwise it is undetermined
	  @param confidence value (>=0) that provides confidence in the search for the closest point
      @param geometry slice geometry used to define the slice where the computation is performed
      @param sliceNumber slice number used to define the slice where the computation is performed
      @param clientData a data structure that can be used to constrain the search of the closest point. This pointer may be 0
      @return a boolean equal to true if a closest Point were found and false otherwise */
  virtual bool getClosestPoint(const Vec2<double> &position,
			       const Vec2<double> &normalDirection,
			       const double rangeSearchDistance,
			       Vec2<double> &closestPoint,double &confidence,
			       SLICE_ORIENTATION geometry=SLICE_Z,
			       double sliceNumber=0,
			       const RangeInformation *clientData=0) const;

  /** compute level set speed as 1 /|\ nabla I| truncated function
      @param position the position of a 2D point in the image
      @param mul returned multiplicative speed term
      @param add returned additive speed term 
      @param clientData range information associated for finding the boundary point*/
  void getSpeed(const Vec2<double> &position, double &mul, double &add, 
		const RangeInformation *clientData) const;


};


} // namespace yav

#endif
