#ifndef RANGEINRIMAGE3D_H
#define RANGEINRIMAGE3D_H

#include <inrimage/Inrimage.h>
#include <imagedata/RangeInrimage.h>

namespace yav {



/** \class RangeInrimage3D
    \ingroup ImageData
    \brief 3D range inrimage */
class RangeInrimage3D :	public RangeInrimage,
			 public Pointed<RangeInrimage3D>
 {

 public:

  /** @name constructors */
  //@{
  /** builds a range inrimage from intensity image
      @param inr intensity inrimage
      @param scene scene to create range image in
      @param gradient gradient image
      @param gradientdir gradient direction image
      @param edge edge image
      @param gradient2 gradient derivative image
      @param classify classified image
      @param allCameras ImageData allCameras flag */
  RangeInrimage3D(Inrimage *inr, 
		  Inrimage *gradient = 0, Inrimage *gradientdir = 0,
		  Inrimage *edge = 0, Inrimage *gradient2 = 0,
		  Inrimage *classify = 0,
		  bool allCameras = false)
    throw(Inrimage::ImageNotFound, Inrimage::UnknownImageType,
	  Inrimage::InvalidHeader, Inrimage::InvalidData,
	  Inrimage::UnknownType);

  /** Builds a range inrimage from intensity image file name.
      Given file name is used as intensity image.
      If name.inr.gz is given, name.gradient and name.gradient.gz are
      used for gradient image candidates, and name.gradient_direction
      and name.gradient_direction.gz are used for gradient direction
      image candidate.
      @param fileName intensity image file name
      @param scene scene to create range image in
      @param allCameras ImageData allCameras flag */
  RangeInrimage3D(const TclString& fileName, 
		  bool allCameras = false)
    throw(Inrimage::ImageNotFound, Inrimage::UnknownImageType,
	  Inrimage::InvalidHeader, Inrimage::InvalidData,
	  Inrimage::UnknownType);
  //@}
  
  /** @name destructor */
  //@{
  /// empty
  virtual ~RangeInrimage3D();
  //@}


  /** RangeData3D closest point computation callback
      @param position vertex position
      @param normal vertex normal
      @param range search range
      @param closest returned closest data point found
      @param clientData vertex local information
      @return true if some data point was found, false otherwise */
  bool getClosestPoint (const Vec3<double> &position,
			const Vec3<double> &normal,
			const double range,
			Vec3<double> &closest,
			double &confidence,
			const RangeInformation *clientData) const;

  /** compute level set speed as 1 /|\ nabla I| truncated function
      @param position the position of a 3D point in the image
      @param mul returned multiplicative speed term
      @param add returned additive speed term   
      @param clientData vertex local information*/
  void getSpeed(const Vec3<double> &position, double &mul, double &add,
		const RangeInformation *clientData)
    const {
    mul = 1.0;
    add = 0.0;
  }


};

} // namespace yav

#endif // ifndef RANGEINRIMAGE3D_H


