#ifndef RANGEINRIMAGE3DGRAPHICS_H
#define RANGEINRIMAGE3DGRAPHICS_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 )
#pragma warning ( disable : 4786 )
#pragma warning ( disable : 4081 )
#endif

#include <imagedata/RangeInrimage3D.h>
#include <graphics3D/Data3D.h>
#include <framework/Client.h>

namespace yav {


/** \class RangeInrimage3DGraphics
    \ingroup ImageData
    \brief Graphics Interface over a 3D range inrimage */
class RangeInrimage3DGraphics :  public RangeInrimage3D, public Data3D
{
 public:
  
  /** @name constructors */
  //@{
  /** builds a range inrimage from intensity image
      @param inr intensity inrimage
      @param scene scene to create range image in
      @param gradient gradient image
      @param gradientdir gradient direction image
      @param edge edge image
      @param gradient2 gradient derivative image
      @param classified classified image */
  RangeInrimage3DGraphics(Inrimage *inr, Scene3D *scene,
		  Inrimage *gradient = 0, Inrimage *gradientdir = 0,
		  Inrimage *edge = 0, Inrimage *gradient2 = 0,
		  Inrimage *classified = 0)
    throw(Inrimage::ImageNotFound, Inrimage::UnknownImageType,
	  Inrimage::InvalidHeader, Inrimage::InvalidData,
	  Inrimage::UnknownType);

  /** Builds a range inrimage from intensity image file name.
      Given file name is used as intensity image.
      If name.inr.gz is given, name.gradient and name.gradient.gz are
      used for gradient image candidates, and name.gradient_direction
      and name.gradient_direction.gz are used for gradient direction
      image candidate.
      @param fileName intensity image file name
      @param scene scene to create range image in */
  RangeInrimage3DGraphics(const TclString& fileName, Scene3D *scene)
    throw(Inrimage::ImageNotFound, Inrimage::UnknownImageType,
	  Inrimage::InvalidHeader, Inrimage::InvalidData,
	  Inrimage::UnknownType);
  //@}
  
  /** @name destructor */
  //@{
  /// empty
  virtual ~RangeInrimage3DGraphics();
  //@}
  
 protected:
  /** an image owned by this object is being destroyed
      @param obj element before destruction */
  virtual void dying(const ObjectPointer<Inrimage>& obj);
 public:
  /** update image display (when the displayed image change)
      @param inr new displayed image */
 virtual void updateImageDisplay(Inrimage *inr){
 }
  /// Returns the graphic object center and weight.
  virtual void updateCenter(Vec3<double>& center, unsigned int& n)  {
    center = Vec3<double>(0.0, 0.0,0.0);
    n = 0;
  }
   /// Returns the graphic object radius given its center
  virtual double updateRadius(const Vec3<double>& center)  { return 0.0; }
  /** Return MAXDOUBLE, an image is considered as background
      and may not be selected */
  virtual double computeClosestDistance(DataTypes::MouseClick& mc) const {
    return -1.0;
  }
  /** default redraw function for sliceable objects
      @param cam camera to redraw */
  virtual void redraw(Camera3D *cam){}

};


} // namespace yav

#endif
