#ifndef TCLIMAGEDATA_H
#define TCLIMAGEDATA_H

#pragma interface "imagedata/TclImageData.h"
#include <map>
#include <graphics3D/TclScene3DObject.h>
#include <graphics3D/TclScene3D.h>
#include <inrimage/TclInrimage.h>
#include <imagedata/ImageData.h>


namespace yav {


/** \class TclImageData
    \ingroup ImageData
    \brief ImageData interface */
class TclImageData : public TclScene3DObject {

 public:
  /** @name constructors */
  //@{
  /** builds an empty TclImageData
      @param scene host scene for created object
      @param id object dynamic type
      @param dimg data to interface
*/
  TclImageData(TclScene3D *scene,
	       const std::type_info& id,
	       ImageData *dimg);

  /** builds a TclImageData interface for an existing inrimage.
      @param inr inrimage to display
      @param scene host scene for created object
      @param id object dynamic type
      @param dimg data to interface
 */
  TclImageData(TclInrimage *inr, TclScene3D *scene,
	       const std::type_info& id = typeid(TclImageData),
	       ImageData *dimg = 0);
  //@}

  /** @name destructor */
  //@{
  /// destroys image data (but not concerned inrimages)
  virtual ~TclImageData();
  //@}


  /// returns client
  ImageData *getClient(void) { return (ImageData *) client; }
  

  /// returns TclImageData objects list head
  static TclImageData* getHead(void) {
    return (TclImageData *) TclObject::getHead(typeid(TclImageData));
  }
  /// returns following element in TclImageData objects list
  TclImageData* following(void) const {
    return (TclImageData *) TclObject::following();
  }
  // return object id by command line arguments
  static const char *getTCLObjectName(void) {
    return TclObjectKind::getObjectName(typeid(TclImageData));
  }

  /// -destroy TCL option
  int destroyCmd(int argc, char **argv);
  /// -scene TCL option
  int sceneCmd(int argc, char **argv);
  /// -box TCL option
  int boxCmd(int argc, char **argv);
  /// -grid TCL option
  int gridCmd(int argc, char **argv);
  /// -render TCL option
  int renderCmd(int argc, char **argv);
  /// -inrimage TCL option
  int inrimageCmd(int argc, char **argv);
  /// -display TCL option
  int displayCmd(int argc, char **argv);
  /// -resolution TCL option
  int resolutionCmd(int argc, char **argv);
  /// -transparency TCL option
  int transparencyCmd(int argc, char **argv);
  /// -texture TCL option
  int textureCmd(int argc, char **argv);
  /// -volume TCL option
  int volumeCmd(int argc, char **argv);
  /// -showBlackPixels TCL option
  int showBlackPixelsCmd(int argc, char **argv);

  /// raises ambigous list() function call
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }

  /// new operator call back function
  static int create(TclModule *m, int argc, char **argv);

  /// load callback function
  static void load(const TclString &filename, Scene3D *scn);

  /// Scalable::objectName() callback
  TclString objectName(void) const {
    return name();
  }

  /// client is being redrawn
  void redraw(void);
};


} // namespace yav

#endif  // ifndef TCLIMAGEDATA_H
