#ifndef TCLRANGEINRIMAGE_H
#define TCLRANGEINRIMAGE_H
#pragma interface "imagedata/TclRangeInrimage.h"
#include <imagedata/RangeInrimage.h>
#include <imagedata/IntensityProfile.h>
#include <graphics/Data.h>
#include <inrimage/TclInrimage.h>
#include <modules/TclRangeInformation.h>

namespace yav {

/** \class TclRangeInrimage
    \ingroup ImageData
    \brief common code for 2D and 3D range inrimage interfaces */
class TclRangeInrimage {
 public:
  /** -force TCL option
      @param argc number of option arguments
      @param argv arguments array
      @param client RangeInrimage client to operate on
      @param result resulting string
      @return TCL_OK, TCL_ERROR or COMMAND_SYNOPSIS */
  int forceCmd(int argc, char **argv,
	       RangeInrimage *client, TclString& result);

  /** -images TCL option
      @param argc number of option arguments
      @param argv arguments array
      @param client RangeInrimage client to operate on
      @param result resulting string
      @return TCL_OK, TCL_ERROR or COMMAND_SYNOPSIS */
  int imagesCmd(int argc, char **argv,
		RangeInrimage *client, TclString& result);


  /** -displayLine TCL option
      @param argc number of option arguments
      @param argv arguments array
      @param client RangeInrimage client to operate on
      @param result resulting string
      @param module
      @return TCL_OK, TCL_ERROR or COMMAND_SYNOPSIS */
  int displayLineCmd(int argc, char **argv,
		     RangeInrimage *client, TclString& result,
		     TclModule *module);

  /** -save TCL option
      @param argc number of option arguments
      @param argv arguments array
      @param client RangeInrimage client to operate on
      @param result resulting string
      @return TCL_OK, TCL_ERROR or COMMAND_SYNOPSIS */
  int saveCmd(int argc, char **argv,
	      RangeInrimage *client, TclString& result);

  static void createCommands(TclModule *m);
 protected:
  /// get Tcl interface of intensity image
  TclInrimage *getIntensity(RangeInrimage *rimg) const {
    return (TclInrimage::getInterface(rimg->intensity()));
  }
  /// get Tcl interface of gradient image
  TclInrimage *getGradient(RangeInrimage *rimg) const {
    return (TclInrimage::getInterface(rimg->gradient()));
  }
  /// get Tcl interface of gradient image
  TclInrimage *getGradientDirection(RangeInrimage *rimg) const {
    return (TclInrimage::getInterface(rimg->gradientDirection()));
  }
  /// get Tcl interface of edge image
  TclInrimage *getEdge(RangeInrimage *rimg) const {
    return (TclInrimage::getInterface(rimg->edge()));
  }
 /// get Tcl interface of gradient of the gradient  image
  TclInrimage *getGradientGradient(RangeInrimage *rimg) const {
    return (TclInrimage::getInterface(rimg->gradientGradient()));
  }
  /// get Tcl interface of classify  image
  TclInrimage *getClassify(RangeInrimage *rimg) const {
    return (TclInrimage::getInterface(rimg->classify()));
  }
 private:
  /** the number of image line windows that have been displayed. It is
      used to avoid to create a unique Tcl window*/
  static unsigned int nbIntensityProfileWindow;
  /** draws a canvas with an image profile */
  void displayGraph(char *frameName, char *title, IntensityProfile &line,
		    double width, double height, TclModule *module);
  
};

/** \class TclGradientBalloonInformation
    \ingroup ImageData
    \brief a class that include information about a gradient of the gradient force */
class TclGradientBalloonInformation : public TclRangeInformation {
 public:
  /** @name constructor */
  //@{
  /** creates a Tcl Interface of a GradientInformation object
      @param m module to create TCL object
      @param ri a pointer to a RangeInformation object */
  TclGradientBalloonInformation(TclModule *m,
				 GradientBalloonInrimageInformation *ri) : TclRangeInformation(typeid(TclGradientBalloonInformation),m,ri) {
    done(typeid(TclGradientBalloonInformation));
  }
  //@}
  
  /** @name destructor */
  //@{
  /// destroys this interface
  virtual ~TclGradientBalloonInformation(){}
  //@}
  /// returns the gradient information object
  GradientBalloonInrimageInformation *getGradientBalloonInformation() const{
    return ((GradientBalloonInrimageInformation *) rangeInformation);}

  /// new TCL command callback
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }
  /// new TCL command callback
  static int create(TclModule *m, int argc, char **argv);
  /// -edit option
  int editCmd(int argc, char **argv);
};

/** \class TclGradientGradientInformation
    \ingroup ImageData
    \brief a class that include information about a gradient of the gradient force */
class TclGradientGradientInformation : public TclRangeInformation {
 public:
  /** @name constructor */
  //@{
  /** creates a Tcl Interface of a GradientInformation object
      @param m module to create TCL object
      @param ri a pointer to a RangeInformation object */
  TclGradientGradientInformation(TclModule *m,
				 GradientGradientInrimageInformation *ri) : TclRangeInformation(typeid(TclGradientGradientInformation),m,ri) {
    done(typeid(TclGradientGradientInformation));
  }
  //@}
  
  /** @name destructor */
  //@{
  /// destroys this interface
  virtual ~TclGradientGradientInformation(){}
  //@}
  /// returns the gradient information object
  GradientGradientInrimageInformation *getGradientGradientInformation() const{
    return ((GradientGradientInrimageInformation *) rangeInformation);}

  /// new TCL command callback
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }
  /// new TCL command callback
  static int create(TclModule *m, int argc, char **argv);
  /// -edit option
  int editCmd(int argc, char **argv);
};

/** \class TclGradientInformation
    \ingroup ImageData
    \brief a class that include information about a gradient force */
class TclGradientInformation : public TclRangeInformation {
 public:
  /** @name constructor */
  //@{
  /** creates a Tcl Interface of a GradientInformation object
      @param m module to create TCL object
      @param ri a pointer to a RangeInformation object */
  TclGradientInformation(TclModule *m,
			 GradientInrimageInformation *ri) : TclRangeInformation(typeid(TclGradientInformation),m,ri) {
    done(typeid(TclGradientInformation));
  }
  //@}
  
  /** @name destructor */
  //@{
  /// destroys this interface
  virtual ~TclGradientInformation(){}
  //@}
  /// returns the gradient information object
  GradientInrimageInformation *getGradientInformation() const{
    return ((GradientInrimageInformation *) rangeInformation);}

  /// new TCL command callback
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }
  /// new TCL command callback
  static int create(TclModule *m, int argc, char **argv);
  /// -edit option
  int editCmd(int argc, char **argv);
};

/** \class TclRegionInformation
    \ingroup ImageData
    \brief a class that include information about a region force */
class TclRegionInformation : public TclRangeInformation {
 public:
  /** @name constructor */
  //@{
  /** creates a Tcl Interface of a GradientInformation object
      @param m module to create TCL object
      @param ri a pointer to a RangeInformation object */
  TclRegionInformation(TclModule *m,
		       RegionInrimageInformation *ri) : TclRangeInformation(typeid(TclRegionInformation),m,ri) {
    done(typeid(TclRegionInformation));
  }
  //@}
  
  /** @name destructor */
  //@{
  /// destroys this interface
  virtual ~TclRegionInformation(){}
  //@}
  /// returns the region information object
  RegionInrimageInformation *getRegionInformation() const{
    return ((RegionInrimageInformation *) rangeInformation);}

  /// new TCL command callback
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }
  /// new TCL command callback
  static int create(TclModule *m, int argc, char **argv);
  /// -edit option
  int editCmd(int argc, char **argv);
};


/** \class TclRegionProbabilityMapInformation
    \ingroup ImageData
    \brief a class that include information about a region force */
class TclRegionProbabilityMapInformation : public TclRangeInformation {
 public:
  /** @name constructor */
  //@{
  /** creates a Tcl Interface of a GradientInformation object
      @param m module to create TCL object
      @param ri a pointer to a RangeInformation object */
  TclRegionProbabilityMapInformation(TclModule *m,
		       RegionProbabilityMapInformation *ri) : TclRangeInformation(typeid(TclRegionProbabilityMapInformation),m,ri) {
    done(typeid(TclRegionProbabilityMapInformation));
  }
  //@}
  
  /** @name destructor */
  //@{
  /// destroys this interface
  virtual ~TclRegionProbabilityMapInformation(){}
  //@}
  /// returns the region information object
  RegionProbabilityMapInformation *getRegionInformation() const{
    return ((RegionProbabilityMapInformation *) rangeInformation);}

  /// new TCL command callback
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }
  /// new TCL command callback
  static int create(TclModule *m, int argc, char **argv);
  /// -edit option
  int editCmd(int argc, char **argv);
};
/** \class TclIsosurfaceInformation
    \ingroup ImageData
    \brief a class that include information about an isosurface force */
class TclIsosurfaceInformation : public TclRangeInformation {
 public:
  /** @name constructor */
  //@{
  /** creates a Tcl Interface of a GradientInformation object
      @param m module to create TCL object
      @param ri a pointer to a RangeInformation object */
  TclIsosurfaceInformation(TclModule *m,
		      IsosurfaceInrimageInformation *ri) : TclRangeInformation(typeid(TclIsosurfaceInformation),m,ri) {
    done(typeid(TclIsosurfaceInformation));
  }
  //@}
  
  /** @name destructor */
  //@{
  /// destroys this interface
  virtual ~TclIsosurfaceInformation(){}
  //@}
  /// returns the region information object
  IsosurfaceInrimageInformation *getIsosurfaceInformation() const{
    return ((IsosurfaceInrimageInformation *) rangeInformation);}

  /// new TCL command callback
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }
  /// new TCL command callback
  static int create(TclModule *m, int argc, char **argv);
  /// -edit option
  int editCmd(int argc, char **argv);
};
/** \class TclClassInformation
    \ingroup ImageData
    \brief a class that include information about a class force */
class TclClassInformation : public TclRangeInformation {
 public:
  /** @name constructor */
  //@{
  /** creates a Tcl Interface of a GradientInformation object
      @param m module to create TCL object
      @param ri a pointer to a RangeInformation object */
  TclClassInformation(TclModule *m,
		      ClassInrimageInformation *ri) : TclRangeInformation(typeid(TclClassInformation),m,ri) {
    done(typeid(TclClassInformation));
  }
  //@}
  
  /** @name destructor */
  //@{
  /// destroys this interface
  virtual ~TclClassInformation(){}
  //@}
  /// returns the region information object
  ClassInrimageInformation *getClassInformation() const{
    return ((ClassInrimageInformation *) rangeInformation);}

  /// new TCL command callback
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }
  /// new TCL command callback
  static int create(TclModule *m, int argc, char **argv);
  /// -edit option
  int editCmd(int argc, char **argv);
};

/** \class TclIntensityProfileInformation
    \ingroup ImageData
    \brief a class that include information about an intensity profile */
class TclIntensityProfileInformation : public TclRangeInformation {
 private:
  /// the index of the display
  unsigned int displayIndex;
 public:
  /** @name constructor */
  //@{
  /** creates a Tcl Interface of a Intensity Profile object
      @param m module to create TCL object
      @param ri a pointer to a RangeInformation object */
  TclIntensityProfileInformation(TclModule *m,
				 IntensityProfileInformation *ri) : TclRangeInformation(typeid(TclIntensityProfileInformation),m,ri) {
    displayIndex=0;
    done(typeid(TclIntensityProfileInformation));
  }
  //@}
  
  /** @name destructor */
  //@{
  /// destroys this interface
  virtual ~TclIntensityProfileInformation(){}
  //@}
  /// returns the region information object
  IntensityProfileInformation *getIntensityProfile() const{
    return ((IntensityProfileInformation *) rangeInformation);}

  /// new TCL command callback
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }
  /// new TCL command callback
  static int create(TclModule *m, int argc, char **argv);
  /// -edit option
  int editCmd(int argc, char **argv);
  /// -draw option
  int drawCmd(int argc, char **argv);
};

/** \class TclIntensityProfileInformationSet
    \ingroup ImageData
    \brief a class that include information about an intensity profile */
class TclIntensityProfileInformationSet : public TclRangeInformationSet {
 public:
  /** @name constructor */
  //@{
  /** creates a Tcl Interface of a Intensity Profile object
      @param m module to create TCL object
      @param ri a pointer to a RangeInformation object */
  TclIntensityProfileInformationSet(TclModule *m,
				 std::list<RangeInformation *>ri=std::list<RangeInformation *>()) : TclRangeInformationSet(typeid(TclIntensityProfileInformationSet),m,ri) {
    done(typeid(TclIntensityProfileInformationSet));
  }
  //@}
  
  /** @name destructor */
  //@{
  /// destroys this interface
  virtual ~TclIntensityProfileInformationSet(){}
  //@}
  /// returns the first intensity profile in list
  std::list<RangeInformation *>::iterator getFirstIntensityProfile()  {
    return (rangeInformationList.begin());}
  /// returns the last intensity profile in list
  std::list<RangeInformation *>::iterator getLastIntensityProfile()  {
    return (rangeInformationList.end());}
  /// add an IntensityProfileInformation object in list 
  void addInList( IntensityProfileInformation *ip) {
    rangeInformationList.push_back((RangeInformation *)ip); }
  /// remove all IntensityProfileInformation object in list 
  void resetList() {
    rangeInformationList.erase(rangeInformationList.begin(),
				rangeInformationList.end()); }

  /// new TCL command callback
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }
  /// new TCL command callback
  static int create(TclModule *m, int argc, char **argv);
  /// -edit option
  int editCmd(int argc, char **argv);
};

/** \class TclBlockMatchingInformation
    \ingroup ImageData
    \brief a class that include information about a block matching */
class TclBlockMatchingInformation : public TclRangeInformation {
 private:
  /// the index of the display
  unsigned int displayIndex;
 public:
  /** @name constructor */
  //@{
  /** creates a Tcl Interface of a Block Matching object
      @param m module to create TCL object
      @param ri a pointer to a RangeInformation object */
  TclBlockMatchingInformation(TclModule *m,
				 BlockMatchingInformation *ri) : TclRangeInformation(typeid(TclBlockMatchingInformation),m,ri) {
    displayIndex=0;
    done(typeid(TclBlockMatchingInformation));
  }
  //@}
  
  /** @name destructor */
  //@{
  /// destroys this interface
  virtual ~TclBlockMatchingInformation(){}
  //@}
  /// returns the region information object
  BlockMatchingInformation *getBlockMatching() const {
    return ((BlockMatchingInformation *) rangeInformation);}

  /// new TCL command callback
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }
  /// new TCL command callback
  static int create(TclModule *m, int argc, char **argv);
  /// -edit option
  int editCmd(int argc, char **argv);
  /// -draw option
  int drawCmd(int argc, char **argv);
};

/** \class TclBlockMatchingInformationSet
    \ingroup ImageData
    \brief a class that include a set of block matching information */
class TclBlockMatchingInformationSet : public TclRangeInformationSet {
 public:
  /** @name constructor */
  //@{
  /** creates a Tcl Interface of a Block Matching object
      @param m module to create TCL object
      @param ri a pointer to a RangeInformation object */
  TclBlockMatchingInformationSet(TclModule *m,
				 std::list<RangeInformation *>ri=std::list<RangeInformation *>()) : TclRangeInformationSet(typeid(TclBlockMatchingInformationSet),m,ri) {
    done(typeid(TclBlockMatchingInformationSet));
  }
  //@}
  
  /** @name destructor */
  //@{
  /// destroys this interface
  virtual ~TclBlockMatchingInformationSet(){}
  //@}
  /// returns the first intensity profile in list
  std::list<RangeInformation *>::iterator getFirstBlockMatching()  {
    return (rangeInformationList.begin());}
  /// returns the last intensity profile in list
  std::list<RangeInformation *>::iterator getLastBlockMatching()  {
    return (rangeInformationList.end());}
  /// add a BlockMatchingInformation object in list 
  void addInList(BlockMatchingInformation *ip) {
    rangeInformationList.push_back((RangeInformation *)ip); }
  /// remove all BlockMatchingInformation object in list 
  void resetList() {
    rangeInformationList.erase(rangeInformationList.begin(),
			       rangeInformationList.end()); }
  
  /// new TCL command callback
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }
  /// new TCL command callback
  static int create(TclModule *m, int argc, char **argv);

  /// -edit option
  int editCmd(int argc, char **argv);
  /// -init option
  int initCmd(int argc, char **argv);
};
  
} // namespace yav

#endif // #ifndef TCLRANGEINRIMAGE_H
