#ifndef TCLRANGEINRIMAGE2D_H
#define TCLRANGEINRIMAGE2D_H


#include <graphics2D/TclScene2DObject.h>
#include <imagedata/RangeInrimage2DGraphics.h>
#include <inrimage/TclInrimage.h>
#include <imagedata/TclRangeInrimage.h>

namespace yav {

/** \class TclRangeInrimage2D
    \ingroup ImageData
    \brief RangeImage2D class interface */
class TclRangeInrimage2D : public TclScene2DObject, public TclRangeInrimage {
 public:
  /** @name constructors */
  //@{
  /** Builds a new range inrimage from an images set.
      @param intensity intensity image (mandatory)
      @param scene 2D scene to set object in
      @param gradient gradient image
      @param gradientDirection gradient direction image
      @param edge edge image
      @param gradientGradient gradient derivative image
      @param classify classified image
      @param id dynamic object id
      @param data client */
  TclRangeInrimage2D(TclInrimage *intensity,
		     TclScene2D *scene,
		     TclInrimage *gradient = 0,
		     TclInrimage *gradientDirection = 0,
		     TclInrimage *edge = 0,
		     TclInrimage *gradientGradient = 0,
		     TclInrimage *classify = 0,
		     const std::type_info& id = typeid(TclRangeInrimage2D),
		     RangeInrimage2DGraphics *data = 0);
		     
  /** Builds a new range inrimage from an inrimage file name
      @param filename inrimage file name
      @param scene 2D scene to set object in
      @param id dynamic object id
      @param data client */
  TclRangeInrimage2D(const TclString& filename,
		     TclScene2D *scene,
		     const std::type_info& id = typeid(TclRangeInrimage2D),
		     RangeInrimage2DGraphics *data = 0);
   /** Builds a Tcl interface from an already existing range data 
      @param scene 2D scene to set object in
      @param id dynamic object id
      @param data client */
  TclRangeInrimage2D(TclScene2D *scene,
		     RangeInrimage2DGraphics *data );
  //@}

  /** @name destructor */
  //@{
  /// empty
  virtual ~TclRangeInrimage2D();
  //@}
  
  /// -save TCL option
  int saveCmd(int argc, char **argv);
  /// -images TCL option
  int imagesCmd(int argc, char **argv);

  /// -force TCL option
  int forceCmd(int argc, char **argv);
  /// -displayLine TCL option
  int displayLineCmd(int argc, char **argv);

  /// returns client real type
  RangeInrimage2DGraphics *getClient(void) { return (RangeInrimage2DGraphics *) client; }
  /// returns the Tcl interface associated to a given RangeInrimage2D
  static TclRangeInrimage2D *getInterface(RangeInrimage2D *range); 
  /// new operator call back function
  static int create(TclModule *m, int argc, char **argv);

  /// load callback function
  static void load(const TclString &filename, Scene2D *scn);
 
  /// creates the Tcl commands when the module is loaded
  static void  createCommands(TclModule *m);

};

} // namespace yav

#endif  // TCLRANGEINRIMAGE2D_H
