#ifndef TCLRANGEINRIMAGE3D_H
#define TCLRANGEINRIMAGE3D_H

#include <imagedata/TclRangeInrimage.h>
#include <imagedata/RangeInrimage3DGraphics.h>
#include <graphics3D/TclScene3DObject.h>

namespace yav {

/** \class TclRangeInrimage3D
    \ingroup ImageData
    \brief RangeImage3D class interface */
class TclRangeInrimage3D : public TclScene3DObject, public TclRangeInrimage {
 public:
  /** @name constructors */
  //@{
  /** Builds a new range inrimage from an images set.
      @param intensity intensity image (mandatory)
      @param scene 3D scene to set object in
      @param gradient gradient image
      @param gradientDirection gradient direction image
      @param edge edge image
      @param gradientGradient gradient derivative image
      @param classify classified image
      @param id dynamic object id
      @param data client */
  TclRangeInrimage3D(TclInrimage *intensity,
		     TclScene3D *scene,
		     TclInrimage *gradient = 0,
		     TclInrimage *gradientDirection = 0,
		     TclInrimage *edge = 0,
		     TclInrimage *gradientGradient = 0,
		     TclInrimage *classify = 0,
		     const std::type_info& id = typeid(TclRangeInrimage3D),
		     RangeInrimage3DGraphics *data = 0);
		     
  /** Builds a new range inrimage from an inrimage file name
      @param filename inrimage file name
      @param scene 3D scene to set object in
      @param id dynamic object id
      @param data client */
  TclRangeInrimage3D(const TclString& filename,
		     TclScene3D *scene,
		     const std::type_info& id = typeid(TclRangeInrimage3D),
		     RangeInrimage3DGraphics *data = 0);
		     
  /** Builds a new range inrimage from an existing client
      @param scene 3D scene to set object in
      @param id dynamic object id
      @param data client */
  TclRangeInrimage3D(TclScene3D *scene, const std::type_info& id,
		     RangeInrimage3DGraphics *data);
  //@}

  /** @name destructor */
  //@{
  /// empty
  virtual ~TclRangeInrimage3D();
  //@}
  
  

  /// -save TCL option
  int saveCmd(int argc, char **argv);
  /// -images TCL option
  int imagesCmd(int argc, char **argv);
   /// -destroy TCL option
  int destroyCmd(int argc, char **argv);

  /// -force TCL option
  int forceCmd(int argc, char **argv);
  /// -displayLine TCL option
  int displayLineCmd(int argc, char **argv);

   /// returns client real type
  RangeInrimage3DGraphics *getClient(void) { return (RangeInrimage3DGraphics *) client; }
  /// returns the Tcl interface associated to a given RangeInrimage3D
  static TclRangeInrimage3D *getInterface(RangeInrimage3D *range); 

  /// new operator call back function
  static int create(TclModule *m, int argc, char **argv);

  /// load callback function
  static void load(const TclString &filename, Scene3D *scn);
  
  /// creates the Tcl commands when the module is loaded
  static void  createCommands(TclModule *m);

};

} // namespace yav

#endif  // TCLRANGEINRIMAGE3D_H
