#include "mipsInrimageDemoGui.h"

#include "mipsDataSetManager.h"
#include "mipsViewSpace.h"
#include "mipsViewer.h"
#include "mipsInrimageToVtkExporter.h"
#include "MIPS_GUIPKG_config.h"

#include "vtkActor.h"
#include "vtkPolyDataMapper.h"
#include "vtkProperty.h"
#include "vtkFloatArray.h"
#include "vtkPolyData.h"
#include "vtkStructuredPointsReader.h"
#include "vtkStructuredPoints.h"

#include <inrimage/Inrimage.h>

namespace mips
{


InrimageDemoGui::InrimageDemoGui() :
  m_gladeGui("mipsInrimageDemo.glade")
{
  // Get the top widget from the plugin glade file
  // and attach it on the main plugin notebook
  GtkWidget *main_window = m_gladeGui.GetWidget("main_window");
  gtk_widget_hide(main_window);  
  m_topWidget = m_gladeGui.GetWidget("main_vbox");
  g_object_ref(this->m_topWidget);
  gtk_container_remove(GTK_CONTAINER(main_window), m_topWidget);
  GtkWidget *label = gtk_label_new("Inrimage Iterate Demo");  
  gtk_notebook_append_page_menu(GTK_NOTEBOOK(Viewer::GetWorkSpaceNotebook()),
                                m_topWidget,
                                label,
                                NULL);
  
  // Connect signal for the test button. Muahahahaha !!!
  m_gladeGui.SignalConnect("test_button", "clicked",
                           G_CALLBACK (InrimageDemoGui::OnTestButtonClickedCallback),
                           this);

  //
  // Initialize the application for Graphical run
  //
  m_zSliceScale = m_gladeGui.GetWidget("ZSliceScale");
  gtk_range_set_range(GTK_RANGE(m_zSliceScale),0, 255);
  g_signal_connect (G_OBJECT (m_zSliceScale), "value-changed", 
                    G_CALLBACK (InrimageDemoGui::OnZScaleValueChangedCallback),this );
  m_x = 0;

  m_ySliceScale = m_gladeGui.GetWidget("YSliceScale");
  gtk_range_set_range(GTK_RANGE(m_ySliceScale),0, 255);
  g_signal_connect (G_OBJECT (m_ySliceScale), "value-changed", 
                    G_CALLBACK (InrimageDemoGui::OnYScaleValueChangedCallback),this );
  m_y = 0;
  
  m_xSliceScale = m_gladeGui.GetWidget("XSliceScale");
  gtk_range_set_range(GTK_RANGE(m_xSliceScale),0, 255);
  g_signal_connect (G_OBJECT (m_xSliceScale), "value-changed", 
                    G_CALLBACK (InrimageDemoGui::OnXScaleValueChangedCallback),this );
  m_z = 0;

  m_Min1Scale = m_gladeGui.GetWidget("Min1Scale");
  g_signal_connect (G_OBJECT (m_Min1Scale), "value-changed", 
                    G_CALLBACK (InrimageDemoGui::OnMin1ScaleValueChangedCallback),this );


  m_Max1Scale  = m_gladeGui.GetWidget("Max1Scale");
  gtk_range_set_range(GTK_RANGE(m_Max1Scale),0, 20000);
  g_signal_connect (G_OBJECT (m_Max1Scale), "value-changed", 
                    G_CALLBACK (InrimageDemoGui::OnMax1ScaleValueChangedCallback),this );
  
  m_Min2Scale = m_gladeGui.GetWidget("Min2Scale");
  g_signal_connect (G_OBJECT (m_Min2Scale), "value-changed", 
                    G_CALLBACK (InrimageDemoGui::OnMin2ScaleValueChangedCallback),this );
  

  m_Max2Scale  = m_gladeGui.GetWidget("Max2Scale");
  gtk_range_set_range(GTK_RANGE(m_Max2Scale),0, 20000);
  g_signal_connect (G_OBJECT (m_Max2Scale), "value-changed", 
                    G_CALLBACK (InrimageDemoGui::OnMax2ScaleValueChangedCallback),this );
  
  // Creating the 3 views in the viewspace for the vtk image
  m_viewXY = new VtkViewImage2D();
  mips::Viewer::GetViewSpace()->AddView(m_viewXY);
  m_viewZY = new VtkViewImage2D();
  mips::Viewer::GetViewSpace()->AddView(m_viewZY);
  m_viewXZ = new VtkViewImage2D();
  mips::Viewer::GetViewSpace()->AddView(m_viewXZ);

  // Creating the 3 views in the viewspace for the inrimage
  m_viewInrXY = new VtkViewImage2D();
  mips::Viewer::GetViewSpace()->AddView(m_viewInrXY);
  m_viewInrZY = new VtkViewImage2D();
  mips::Viewer::GetViewSpace()->AddView(m_viewInrZY);
  m_viewInrXZ = new VtkViewImage2D();
  mips::Viewer::GetViewSpace()->AddView(m_viewInrXZ);
    
  // Open the image
  std::string filename = std::string(TESTDATA_DIR) +
    std::string("/Images/MRI16bitsFull.vtk");
  vtkStructuredPointsReader* reader = vtkStructuredPointsReader::New();
  reader->SetFileName(filename.c_str());
  vtkStructuredPoints* image = (reader->GetOutput());
  image->Update();
  m_viewXY->SetImage(image);
  m_viewXY->SetOrientation(0);
  m_viewZY->SetImage(image);
  m_viewZY->SetOrientation(1);
  m_viewXZ->SetImage(image);
  m_viewXZ->SetOrientation(2);
  image->Delete();  // vtk smart Delete
  

  // Open an inrimage and test the exporter
  std::string filename1 = std::string(TESTDATA_DIR) +
    std::string("/Images/MRI16bitsFull.inr.gz");

  cout << "Image inr: " << filename1 << endl;
  
  yav::Inrimage* image1 = new yav::Inrimage(filename1.c_str());
  m_Exporter = new InrimageToVtkExporter;
  m_Exporter->SetInput(image1);
  vtkImageData* image2 = m_Exporter->GetOutput();
  image2->Update();
  m_viewInrXY->SetImage(image2);
  m_viewInrXY->SetOrientation(0);
  m_viewInrZY->SetImage(image2);
  m_viewInrZY->SetOrientation(1);
  m_viewInrXZ->SetImage(image2);
  m_viewInrXZ->SetOrientation(2);
  //image2->Delete();  // vtk smart Delete

  // Extract first view image range, and set color window and level
  float range[2];
  m_viewXY->GetInput()->GetScalarRange(range);
  cout << "Range: " << range[0] << ", " << range[1] << endl;

  gtk_range_set_range(GTK_RANGE(m_Min1Scale), range[0], range[1]);
  gtk_range_set_range(GTK_RANGE(m_Max1Scale),  range[0], range[1]);
  gtk_range_set_value(GTK_RANGE(m_Min1Scale), range[0]);
  gtk_range_set_value(GTK_RANGE(m_Max1Scale),  range[1]);
  
  m_viewInrXY->GetInput()->GetScalarRange(range);
  cout << "Range: " << range[0] << ", " << range[1] << endl;

  gtk_range_set_range(GTK_RANGE(m_Min2Scale), range[0], range[1]);
  gtk_range_set_range(GTK_RANGE(m_Max2Scale),  range[0], range[1]);
  gtk_range_set_value(GTK_RANGE(m_Min2Scale), range[0]);
  gtk_range_set_value(GTK_RANGE(m_Max2Scale),  range[1]);

  // Update x, y, z slider corresponding to image size
  int* dims = m_viewXY->GetInput()->GetDimensions();
  gtk_range_set_range(GTK_RANGE(m_xSliceScale), 0, dims[0]);
  gtk_range_set_range(GTK_RANGE(m_ySliceScale), 0, dims[1]);
  gtk_range_set_range(GTK_RANGE(m_zSliceScale), 0, dims[2]);
  gtk_range_set_value(GTK_RANGE(m_xSliceScale), dims[0] / 2.);
  gtk_range_set_value(GTK_RANGE(m_ySliceScale), dims[1] / 2.);
  gtk_range_set_value(GTK_RANGE(m_zSliceScale), dims[2] / 2.);

}

InrimageDemoGui::~InrimageDemoGui()
{
  g_object_unref(this->GetTopWidget());
}


//   void
//   InrimageDemoGui::AddInrimageDemo()
//   {
//   } 

  
GtkWidget*
InrimageDemoGui::GetTopWidget()
{
  return m_topWidget;  
}


void 
InrimageDemoGui::OnTestButtonClicked(GtkButton *btn)
{
  gtk_button_set_label(btn, "Muahahahahaha !!!!!!!!");
  gtk_range_set_value(GTK_RANGE(m_Min1Scale), 0);
  gtk_range_set_value(GTK_RANGE(m_Max1Scale), 10000);
}
void 
InrimageDemoGui::OnTestButtonClickedCallback(GtkButton *btn, gpointer data)
{
  static_cast<InrimageDemoGui*>(data)->OnTestButtonClicked(btn);  
}

void 
InrimageDemoGui::OnZScaleValueChanged(GtkRange *range)
{
  m_z = (int)(gtk_range_get_value(range));

  m_viewXY->UpdatePosition(m_x, m_y, m_z);
  m_viewInrXY->UpdatePosition(m_x, m_y, m_z);
  m_viewXY->Render();
  m_viewInrXY->Render();
}

void 
InrimageDemoGui::OnZScaleValueChangedCallback(GtkRange *range, gpointer data)
{
  static_cast<InrimageDemoGui*>(data)->OnZScaleValueChanged(range);  
}

void 
InrimageDemoGui::OnYScaleValueChanged(GtkRange *range)
{
  m_y = (int)(gtk_range_get_value(range));

  m_viewXZ->UpdatePosition(m_x, m_y, m_z);
  m_viewInrXZ->UpdatePosition(m_x, m_y, m_z);
  m_viewXZ->Render();
  m_viewInrXZ->Render();
}

void 
InrimageDemoGui::OnYScaleValueChangedCallback(GtkRange *range, gpointer data)
{
  static_cast<InrimageDemoGui*>(data)->OnYScaleValueChanged(range);  
}

void 
InrimageDemoGui::OnXScaleValueChanged(GtkRange *range)
{
  m_x = (int)(gtk_range_get_value(range));
  
  m_viewZY->UpdatePosition(m_x, m_y, m_z);
  m_viewInrZY->UpdatePosition(m_x, m_y, m_z);
  m_viewZY->Render();
  m_viewInrZY->Render();
}

void 
InrimageDemoGui::OnXScaleValueChangedCallback(GtkRange *range, gpointer data)
{
  static_cast<InrimageDemoGui*>(data)->OnXScaleValueChanged(range);  
}

void 
InrimageDemoGui::OnMin1ScaleValueChangedCallback(
  GtkRange *range, gpointer data)
{
  //static_cast<InrimageDemoGui*>(data)->OnMin1ScaleValueChanged(range);
  static_cast<InrimageDemoGui*>(data)->UpdateColorWindow1();
}

void 
InrimageDemoGui::OnMax1ScaleValueChangedCallback(
  GtkRange *range, gpointer data)
{
  //static_cast<InrimageDemoGui*>(data)->OnMax1ScaleValueChanged(range);  
  static_cast<InrimageDemoGui*>(data)->UpdateColorWindow1();
}

void InrimageDemoGui::UpdateColorWindow1(float minValue, float maxValue)
{
  gtk_range_set_value(GTK_RANGE(m_Min1Scale), minValue);
  gtk_range_set_value(GTK_RANGE(m_Max1Scale), maxValue);
}

void InrimageDemoGui::UpdateColorWindow1()
{
  float minValue = (float)(gtk_range_get_value(GTK_RANGE(m_Min1Scale)));
  float maxValue = (float)(gtk_range_get_value(GTK_RANGE(m_Max1Scale)));

  float window = maxValue - minValue;
  float level = (maxValue + minValue) / 2.;

  m_viewXY->SetColorLevel(level);
  m_viewZY->SetColorLevel(level);
  m_viewXZ->SetColorLevel(level);
  m_viewXY->SetColorWindow(window);
  m_viewZY->SetColorWindow(window);
  m_viewXZ->SetColorWindow(window);
  
  m_viewXY->Render();
  m_viewZY->Render();
  m_viewXZ->Render();
}

void 
InrimageDemoGui::OnMin2ScaleValueChangedCallback(
  GtkRange *range, gpointer data)
{
  //static_cast<InrimageDemoGui*>(data)->OnMin2ScaleValueChanged(range);
  static_cast<InrimageDemoGui*>(data)->UpdateColorWindow2();
}

void 
InrimageDemoGui::OnMax2ScaleValueChangedCallback(
  GtkRange *range, gpointer data)
{
  //static_cast<InrimageDemoGui*>(data)->OnMax2ScaleValueChanged(range);  
  static_cast<InrimageDemoGui*>(data)->UpdateColorWindow2();
}

void InrimageDemoGui::UpdateColorWindow2()
{
  float minValue = (float)(gtk_range_get_value(GTK_RANGE(m_Min2Scale)));
  float maxValue = (float)(gtk_range_get_value(GTK_RANGE(m_Max2Scale)));

  float window = maxValue - minValue;
  float level = (maxValue + minValue) / 2.;

  m_viewInrXY->SetColorLevel(level);
  m_viewInrZY->SetColorLevel(level);
  m_viewInrXZ->SetColorLevel(level);
  m_viewInrXY->SetColorWindow(window);
  m_viewInrZY->SetColorWindow(window);
  m_viewInrXZ->SetColorWindow(window);
  
  m_viewInrXY->Render();
  m_viewInrZY->Render();
  m_viewInrXZ->Render();
}

} // namespace mips 
