#ifndef MIPS_INRIMAGE_DEMO_GUI_H
#define MIPS_INRIMAGE_DEMO_GUI_H

#include "mipsApplicationGUI.h"
#include "mipsInrimageDemoFacade.h"
#include "mipsInrimageToVtkExporter.h"
#include "mipsVtkViewImage2D.h"
#include "mipsGladeGui.h"

#include "gtk/gtk.h"

namespace mips
{

/** \class InrimageDemoGui mipsInrimageDemoGui.h "mipsInrimageDemoGui.h"
    \ingroup ImageGUI
    \brief Temporary and hacky first version of an inrimage viewer within Mips

    The plugin loads a test vtk image, and a test inrimage. It displays an
    axial, coronal and sagital view of each image. The 3 sliders allow to
    control the position of each slice.

    Currently, the UI shortcuts the controller, which means the views are only
    synchronized via the UI.
*/
class InrimageDemoGui : public ApplicationGUI
{
  // Attributes
private:
  GtkWidget*              m_topWidget;
  InrimageDemoFacade      m_application;
  GladeGui                m_gladeGui;
  InrimageToVtkExporter   m_inrimageExporter;
  VtkViewImage2D*         m_viewXY;
  VtkViewImage2D*         m_viewZY;
  VtkViewImage2D*         m_viewXZ;
  VtkViewImage2D*         m_viewInrXY;
  VtkViewImage2D*         m_viewInrZY;
  VtkViewImage2D*         m_viewInrXZ;
  InrimageToVtkExporter*  m_Exporter;
  GtkWidget*              m_zSliceScale;
  GtkWidget*              m_ySliceScale;
  GtkWidget*              m_xSliceScale;
  GtkWidget*              m_Min1Scale;
  GtkWidget*              m_Max1Scale;
  GtkWidget*              m_Min2Scale;
  GtkWidget*              m_Max2Scale;

  int                     m_x;
  int                     m_y;
  int                     m_z;
  // Public Methods
public:
  /// Default constructor: loads the UI, registers it within the plugin
  /// notebook and connects the signals to the views
  InrimageDemoGui();
  /// Default destructor: detach the UI from the plugin notebook
  virtual ~InrimageDemoGui();
  /// Extract the top widget, which is usually a box to be inserted in the
  /// notebook
  virtual GtkWidget* GetTopWidget();


  /// Test button callback
  void OnTestButtonClicked(GtkButton *btn);
  /// Test button callback: corresponding gtk callback
  static void OnTestButtonClickedCallback(GtkButton *btn, gpointer data);

  /// Z slider callback
  void OnZScaleValueChanged(GtkRange *range);
  /// Z slider callback: corresponding gtk callback
  static void OnZScaleValueChangedCallback(GtkRange *range, gpointer user_data);
  
  /// Y slider callback
  void OnYScaleValueChanged(GtkRange *range);
  /// Y slider callback: corresponding gtk callback
  static void OnYScaleValueChangedCallback(GtkRange *range, gpointer user_data);
  
  /// X slider callback
  void OnXScaleValueChanged(GtkRange *range);
  /// X slider callback: corresponding gtk callback
  static void OnXScaleValueChangedCallback(GtkRange *range, gpointer user_data);


  /// Color Window Slider callback
  //void OnMin1ScaleValueChanged(GtkRange *range);
  /// Color Window Slider callback: corresponding gtk callback
  static void OnMin1ScaleValueChangedCallback(
    GtkRange *range, gpointer user_data);
  
  /// Color Level Slider callback
  //void OnMax1ScaleValueChanged(GtkRange *range);
  /// Color Level Slider callback: corresponding gtk callback
  static void OnMax1ScaleValueChangedCallback(
    GtkRange *range, gpointer user_data);

  
  void UpdateColorWindow1(float minValue, float maxValue);
  void UpdateColorWindow1();
  

  /// Color Window Slider callback
  //void OnMin2ScaleValueChanged(GtkRange *range);
  /// Color Window Slider callback: corresponding gtk callback
  static void OnMin2ScaleValueChangedCallback(
    GtkRange *range, gpointer user_data);
  
  /// Color Level Slider callback
  //void OnMax2ScaleValueChanged(GtkRange *range);
  /// Color Level Slider callback: corresponding gtk callback
  static void OnMax2ScaleValueChangedCallback(
    GtkRange *range, gpointer user_data);
  
  void UpdateColorWindow2();
  
  //Private Methods
private:
};

} //namespace mips

#endif /* MIPS_MESH_ITERATION_DEMO_GUI_H */
