#ifndef MIPS_IMAGE_DATASET_H
#define MIPS_IMAGE_DATASET_H

#include "mipsDataSet.h"

#include <ostream>

class vtkImageData;
class vtkMatrix4x4;
class vtkTransform;
class vtkImageReslice;
class vtkIndent;

namespace mips {

/** \class ImageDataSet mipsImageDataSet.h 
    \brief Base image dataset API abstract class */
class ImageDataSet : public DataSet
{
 public:
  ImageDataSet();            /**< Default constructor */
  virtual ~ImageDataSet();   /**< Default destructor */
  
  // Return the transform to apply to fit to the the mips visualization referential.
  vtkMatrix4x4         *GetOriginalTransform();
  void                  SetOriginalTransform(vtkMatrix4x4*);

  // Return the transformed vtkImageData.
  vtkImageData*         GetVtkImageData();
  
  virtual void          UpdateVtkImageData()=0;
  virtual void          CreatePageInViewSpace(ViewSpace* p_viewSpace);

  void                  PrintSelf(std::ostream& os, vtkIndent indent);
 
 protected:
  // The vtkImageData have to be set by the child. The child keep the responsability  
  // of the memory de-allocation (vtk Delete()).
  void                  SetVtkImageData(vtkImageData* p_image);  
  vtkImageData         *m_vtkImageData;  

 private:
  vtkImageReslice*      GetReslicer();  

 private:
  vtkImageReslice      *m_reslicer;  
  vtkTransform         *m_transform;  
};
 
} // namespace mips

#endif //MIPS_IMAGE_DATASET_H
