#include <iostream>
#include "mipsInrimageToVTKImageData.h"
#include "mipsVTKImageDataToInrimage.h"
#include <vtkStructuredPointsReader.h>
#include <vtkStructuredPointsWriter.h>
#include <vtkStructuredPoints.h>


void PrintHelp (const char* exec)
{
  std::cout << "Usage: " << std::endl;
  std::cout << exec << " [vtkImageData_in] [Inrimage_out] [vtkImageData_out]" << std::endl;
  exit (0);
}



int main (int narg, char* arg[])
{

  if( narg!=4 )
    PrintHelp (arg[0]);
  
  
  typedef mips::VTKImageDataToInrimage VTKToInrimage;
  typedef mips::InrimageToVTKImageData InrimageToVTK;
  typedef VTKToInrimage::Inrimage      Inrimage;


  char* filename = arg[1];
  
  
  vtkStructuredPointsReader* reader = vtkStructuredPointsReader::New();
  reader->SetFileName (filename);

  vtkStructuredPoints* vtkImage = reader->GetOutput();
  vtkImage->Update();
  vtkImage->SetReferenceCount (2); // tricky: to register the image

  // convert VTK -> Inrimage
  std::cout << "Converting vtkImageData to Inrimage...";
  std::cout << std::flush;
  VTKToInrimage* myConverter = new VTKToInrimage;
  myConverter->SetInput (vtkImage);
  myConverter->Update();

  Inrimage* myInrimage = myConverter->GetOutput();
  myInrimage->write (arg[2]);
  std::cout << "Done." << std::endl;

  
  // convert Inrimage -> VTK
  std::cout << "Converting Inrimage to vtkImageData ...";
  std::cout << std::flush;
  InrimageToVTK* myConverter2 = new InrimageToVTK;
  myConverter2->SetInput ( myInrimage );
  myConverter2->Update();
  vtkImageData* vtkOutput = myConverter2->GetOutput();
  vtkOutput->SetReferenceCount (2);  
  std::cout << "Done." << std::endl;
  
  vtkStructuredPointsWriter* writer = vtkStructuredPointsWriter::New();
  writer->SetFileName (arg[3]);
  writer->SetInput (vtkOutput);
  writer->Write();

  reader->Delete();
  delete myConverter;
  delete myConverter2;
  delete myInrimage;
  writer->Delete();
  
  return 0;
}
