#include "mipsInrimageVTKImageExport.h"

using namespace yav;

namespace mips
{

  InrimageVTKImageExport::InrimageVTKImageExport()
  {
    m_Input = 0;
    m_ScalarTypeName = "double";
  }

  InrimageVTKImageExport::~InrimageVTKImageExport()
  {}

  void
  InrimageVTKImageExport::SetInput(const Inrimage* im)
  {
    m_Input = im;
  }

  const Inrimage*
  InrimageVTKImageExport::GetInput()
  {
    return m_Input;
  }


  int*
  InrimageVTKImageExport::WholeExtentCallback()
  {
    // std::cout << "InrimageVTKImageExport::WholeExtentCallback()" << std::endl;
    m_WholeExtent[0] = 0;
    m_WholeExtent[1] = m_Input->getX()-1;
    m_WholeExtent[2] = 0;
    m_WholeExtent[3] = m_Input->getY()-1;
    m_WholeExtent[4] = 0;
    m_WholeExtent[5] = m_Input->getZ()-1;

    return m_WholeExtent;
  }

  
  /**
   * Implements the SpacingCallback.  This returns a pointer to an array
   * of three floating point values describing the spacing of the image.
   */
  double*
  InrimageVTKImageExport::SpacingCallback()
  {

    // std::cout << "InrimageVTKImageExport::SpacingCallback()" << std::endl;
    m_DataSpacing[0] = (double)(m_Input->getVX());
    m_DataSpacing[1] = (double)(m_Input->getVY());
    m_DataSpacing[2] = (double)(m_Input->getVZ());
   
    return m_DataSpacing;
  }

  /**
   * Implements the SpacingCallback.  This returns a pointer to an array
   * of three floating point values describing the spacing of the image.
   */
  float* 
  InrimageVTKImageExport::FloatSpacingCallback()
  {
    m_FloatDataSpacing[0] = m_Input->getVX();
    m_FloatDataSpacing[1] = m_Input->getVY();
    m_FloatDataSpacing[2] = m_Input->getVZ();
   
    return m_FloatDataSpacing;
  }

  /**
 * Implements the OriginCallback.  This returns a pointer to an array
 * of three floating point values describing the origin of the image.
 */
  double*
  InrimageVTKImageExport::OriginCallback()
  {
    m_DataOrigin[0] = (double)(m_Input->getTX());
    m_DataOrigin[1] = (double)(m_Input->getTY());
    m_DataOrigin[2] = (double)(m_Input->getTZ());
    return m_DataOrigin;
  }

  /**
 * Implements the OriginCallback.  This returns a pointer to an array
 * of three floating point values describing the origin of the image.
 */
  float*
  InrimageVTKImageExport::FloatOriginCallback()
  {
    m_FloatDataOrigin[0] = m_Input->getTX();
    m_FloatDataOrigin[1] = m_Input->getTY();
    m_FloatDataOrigin[2] = m_Input->getTZ();
    return m_FloatDataOrigin;
  }

  /**
 * Implements the ScalarTypeCallback.  This returns the name of the
 * scalar value type of the image.
 */
  const char*
  InrimageVTKImageExport::ScalarTypeCallback()
  {

    Inrimage::WORD_TYPE imType = m_Input->getType();

    switch(imType)
      {
      case Inrimage::WT_UNSIGNED_CHAR:
	m_ScalarTypeName = "unsigned char";
	break;
      case Inrimage::WT_SIGNED_CHAR:
	m_ScalarTypeName = "char";
	break;
      case Inrimage::WT_UNSIGNED_SHORT:
	m_ScalarTypeName = "unsigned short";
	break;
      case Inrimage::WT_SIGNED_SHORT:
	m_ScalarTypeName = "short";
	break;
      case Inrimage::WT_UNSIGNED_LONG:
	m_ScalarTypeName = "unsigned long";
	break;
      case Inrimage::WT_SIGNED_LONG:
	m_ScalarTypeName = "long";
	break;
      case Inrimage::WT_FLOAT:
	m_ScalarTypeName = "float";
	break;
      case Inrimage::WT_DOUBLE:
	m_ScalarTypeName = "double";
	break;
      case Inrimage::WT_RGB:
	m_ScalarTypeName = "unsigned char";
	break;
      case Inrimage::WT_RGBA:
	m_ScalarTypeName = "unsigned char";
	break;	
      case Inrimage::WT_FLOAT_VECTOR:
	m_ScalarTypeName = "float";
	break;
      case Inrimage::WT_DOUBLE_VECTOR:
	m_ScalarTypeName = "double";
	break;
      case Inrimage::WT_UNSIGNED_INT:
	m_ScalarTypeName = "unsigned int";
	break;
      case Inrimage::WT_SIGNED_INT:
	m_ScalarTypeName = "int";
	break;
      default:
	std::cerr << "Scalar format not supported." << std::endl;
	// EXCEPTION SOMEWHERE HERE!!
	break;
    }
    
    return m_ScalarTypeName.c_str();
  }

  /**
 * Implements the NumberOfComponentsCallback.  This returns the number of
 * components per pixel for the image.
 *
 */
  int
  InrimageVTKImageExport::NumberOfComponentsCallback()
  {

    int numberOfComponents;
    Inrimage::WORD_TYPE imType = m_Input->getType();
    
    switch(imType)
      {
      
      case Inrimage::WT_RGB:
	numberOfComponents = 3;
	break;
      case Inrimage::WT_RGBA:
	numberOfComponents = 4;
	break;	
      case Inrimage::WT_FLOAT_VECTOR:
	numberOfComponents = m_Input->getVdim();
	break;
      case Inrimage::WT_DOUBLE_VECTOR:
	numberOfComponents = m_Input->getVdim();
	break;
      default:
	numberOfComponents = 1;
	break;
      }
    
    return numberOfComponents;
  }

  
  /**
   * Implements the BufferPointerCallback.  This returns a pointer to
   * the image's memory buffer.
   */
  void*
  InrimageVTKImageExport::BufferPointerCallback()
  {
    return m_Input->getData();
  }

  int*
  InrimageVTKImageExport::DataExtentCallback()
  {
    return this->WholeExtentCallback();
  }


}// end of namespace mips
