#include "mipsInrimageVTKImageImport.h"

using namespace yav;

namespace mips
{

  InrimageVTKImageImport::InrimageVTKImageImport()
  {
    m_Image = 0;
    
  m_SpacingCallback = 0;
  m_FloatSpacingCallback = 0;
  m_OriginCallback = 0;
  m_FloatOriginCallback = 0;
  m_ScalarTypeCallback = 0;
  m_NumberOfComponentsCallback = 0;
  m_DataExtentCallback = 0;
  m_BufferPointerCallback = 0;
  m_FloatSpacingCallback = 0;
  m_FloatOriginCallback = 0;

  for( int i=0; i<3; i++)
  {
    m_DataExtent[i]=0;
    m_Spacing[i]=1.0;
    m_Origin[i]=0.0;
  }
  m_NumberOfComponents = 1;
  m_Buffer = 0;
  m_WordType = Inrimage::WT_DOUBLE;
}


void InrimageVTKImageImport::Update()
{

  this->ExecuteCallbacks();

  // do some checks here
  if( m_NumberOfComponents == 3 && (m_WordType==Inrimage::WT_UNSIGNED_CHAR) )
    m_WordType = Inrimage::WT_RGB;
  else if( m_NumberOfComponents == 4 && (m_WordType==Inrimage::WT_UNSIGNED_CHAR) )
    m_WordType = Inrimage::WT_RGBA;
  else if ( m_NumberOfComponents>1 )
  {
    
    WORD_TYPE newType = m_WordType;
    
    switch( m_WordType )
    {
        case Inrimage::WT_DOUBLE:
          newType = Inrimage::WT_DOUBLE_VECTOR;
          break;

        case Inrimage::WT_FLOAT:
          newType = Inrimage::WT_FLOAT_VECTOR;
          break;

        default:
          break;          
    }

    m_WordType = newType;
  }

  // create the output image
  m_Image = new Inrimage ( m_DataExtent[0], m_DataExtent[1], m_DataExtent[2],
                           m_WordType, m_NumberOfComponents, VM_INTERLACED,
                           (float)m_Spacing[0], (float)m_Spacing[1], (float)m_Spacing[2] );
  
  
  m_Image->setData (m_Buffer);
  // origin not yet implemented
  
}


void InrimageVTKImageImport::ExecuteCallbacks()
{
  if( !m_CallbackUserData )
    return;
  
  if( m_DataExtentCallback )
    this->SetDataExtent ( (m_DataExtentCallback)(m_CallbackUserData) );

  if( m_SpacingCallback )
    this->SetSpacing ( (m_SpacingCallback)(m_CallbackUserData) );
  
  if( m_FloatSpacingCallback )
    this->SetFloatSpacing ( (m_FloatSpacingCallback)(m_CallbackUserData) );

  if( m_OriginCallback )
    this->SetOrigin ( (m_OriginCallback)(m_CallbackUserData) );

  if( m_FloatOriginCallback )
    this->SetFloatOrigin ( (m_FloatOriginCallback)(m_CallbackUserData) );

  if( m_ScalarTypeCallback )
    this->SetScalarType ( (m_ScalarTypeCallback)(m_CallbackUserData) );

  if( m_NumberOfComponentsCallback )
    this->SetNumberOfComponents( (m_NumberOfComponentsCallback)(m_CallbackUserData) );

  if( m_BufferPointerCallback )
    this->SetBufferPointer ( (m_BufferPointerCallback)(m_CallbackUserData) );
}


void InrimageVTKImageImport::SetDataExtent (int* extent)
{
  for( int i=0; i<3; i++)
    m_DataExtent[i]=extent[2*i+1]-extent[2*i]+1;
}

void InrimageVTKImageImport::SetSpacing (double* spacing)
{
  for( int i=0; i<3; i++)
    m_Spacing[i]=spacing[i];
}

void InrimageVTKImageImport::SetOrigin (double* origin)
{
  for( int i=0; i<3; i++)
    m_Origin[i]=origin[i];
}

void InrimageVTKImageImport::SetFloatSpacing (float* origin)
{
  for( int i=0; i<3; i++)
    m_Origin[i]=(double)(origin[i]);
}

void InrimageVTKImageImport::SetFloatOrigin (float* spacing)
{
  for( int i=0; i<3; i++)
    m_Spacing[i]=(double)(spacing[i]);
}

void InrimageVTKImageImport::SetScalarType (const char* scalarType)
{
  if( strcmp (scalarType, "double")==0 )
    m_WordType = Inrimage::WT_DOUBLE;
  else if( strcmp (scalarType, "float")==0 )
    m_WordType = Inrimage::WT_FLOAT;
  else if( strcmp (scalarType, "long")==0 )
    m_WordType = Inrimage::WT_SIGNED_LONG;
  else if( strcmp (scalarType, "unsigned long")==0 )
    m_WordType = Inrimage::WT_UNSIGNED_LONG;
  else if( strcmp (scalarType, "int")==0 )
    m_WordType = Inrimage::WT_SIGNED_INT;
  else if( strcmp (scalarType, "unsigned int")==0 )
    m_WordType = Inrimage::WT_UNSIGNED_INT;
  else if( strcmp (scalarType, "short")==0 )
    m_WordType = Inrimage::WT_SIGNED_SHORT;
  else if( strcmp (scalarType, "unsigned short")==0 )
    m_WordType = Inrimage::WT_UNSIGNED_SHORT;
  else if( strcmp (scalarType, "char")==0 )
    m_WordType = Inrimage::WT_SIGNED_CHAR;
  else if( strcmp (scalarType, "unsigned char")==0 )
    m_WordType = Inrimage::WT_UNSIGNED_CHAR;    
}

void InrimageVTKImageImport::SetNumberOfComponents ( int n )
{
  m_NumberOfComponents = n;
}

void InrimageVTKImageImport::SetBufferPointer (void* buffer)
{
  m_Buffer = buffer;
}


} // end of namespace 
