#ifndef mips_InrimageVTKImageImport_h_
#define mips_InrimageVTKImageImport_h_

#include <inrimage/Inrimage.h>

namespace yav {
  class Inrimage;
};

//using namespace yav;

namespace mips
{
  
 class InrimageVTKImageImport
 {

 public:
   InrimageVTKImageImport();
   ~InrimageVTKImageImport(){};

   /** These are function pointer types for the pipeline connection
    * callbacks. */
   typedef double* (*SpacingCallbackType)(void*);
   typedef double* (*OriginCallbackType)(void*);
   typedef const char* (*ScalarTypeCallbackType)(void*); 
   typedef int (*NumberOfComponentsCallbackType)(void*);
   typedef int* (*DataExtentCallbackType)(void*);
   typedef void* (*BufferPointerCallbackType)(void*);

   typedef yav::Inrimage::WORD_TYPE WORD_TYPE;
   
   /** Compatibility for VTK older than 4.4.  */
   typedef float* (*FloatSpacingCallbackType)(void*);
   typedef float* (*FloatOriginCallbackType)(void*);

   void SetSpacingCallback            (SpacingCallbackType cb)
   { m_SpacingCallback=cb; }
   void SetOriginCallback             (OriginCallbackType cb)
   { m_OriginCallback=cb; }
   void SetScalarTypeCallback         (ScalarTypeCallbackType cb)
   { m_ScalarTypeCallback=cb; }
   void SetNumberOfComponentsCallback (NumberOfComponentsCallbackType cb)
   { m_NumberOfComponentsCallback=cb; }
   void SetDataExtentCallback         (DataExtentCallbackType cb)
   { m_DataExtentCallback=cb; }
   void SetBufferPointerCallback      (BufferPointerCallbackType cb)
   { m_BufferPointerCallback=cb; }
   void SetFloatSpacingCallback       (FloatSpacingCallbackType cb)
   { m_FloatSpacingCallback=cb;}
   void SetFloatOriginCallback        (FloatOriginCallbackType cb)
   { m_FloatOriginCallback=cb;}

   void SetCallbackUserData (void* data)
   { m_CallbackUserData = data; }
   
   void Update (void);
   
   yav::Inrimage* GetOutput (void) const
   { return m_Image; }
   
   
 protected:

   void ExecuteCallbacks (void);

   void SetDataExtent ( int* );
   void SetSpacing (double* );
   void SetFloatSpacing ( float* );
   void SetOrigin ( double* );
   void SetFloatOrigin ( float* );
   void SetScalarType ( const char* );
   void SetNumberOfComponents ( int );
   void SetBufferPointer ( void* );
   
 private:
   
   SpacingCallbackType               m_SpacingCallback;
   FloatSpacingCallbackType          m_FloatSpacingCallback;
   OriginCallbackType                m_OriginCallback;
   FloatOriginCallbackType           m_FloatOriginCallback;
   ScalarTypeCallbackType            m_ScalarTypeCallback;
   NumberOfComponentsCallbackType    m_NumberOfComponentsCallback;  
   DataExtentCallbackType            m_DataExtentCallback;
   BufferPointerCallbackType         m_BufferPointerCallback;
   void*                             m_CallbackUserData;
   
   WORD_TYPE m_WordType;
   double m_Spacing[3];
   double m_Origin[3];
   int    m_NumberOfComponents;
   int    m_DataExtent[3];
   void*  m_Buffer;
   

   yav::Inrimage * m_Image;
      
 };
  
} // end of namespace mips


#endif
