#include <stdlib.h>
#include <string.h>

#include <inrimage/Inrimage.h>

// Convert 3D PGM image to Inrimage
// So far, only 3D PGM images whose synthaxis is described here 
// (http://laurentnajman.org/heart/index.html) can be read successfully.
//
// Format:
// P7
// sizeX sizeY sizeZ
// max
// Content as unsigned char.

int main(int argc, char** argv)
{
	if (argc <= 2)
	{
		CERR << "Usage: " << argv[0] 
				 << " <input.pgm> <output.inr.gz> [<spacing x>] [<spacing y>] [<spacing z>]\n";
		return 1;
	}

	// Initialization
	float vx, vy, vz;
	vx = vy = vz = 1;

	if ( argc > 5 )
		vz = atof(argv[5]);
	if ( argc > 4 )
		vy = atof(argv[4]);
	if ( argc > 3 )
		vx = atof(argv[3]);

	// Open input file
	FILE* fp;
	fp = fopen(argv[1], "r");
	if ( !fp )
	{
		CERR << "Error: impossible to read " << argv[1] << ENDL;
		return 1;
	}

	// Read the header
	char buffer[256];
	unsigned int sizeX, sizeY, sizeZ;
	unsigned int max;

	fscanf(fp, "%s", buffer);
 	if ( (buffer[0] != 'P') || (buffer[1] != '7') )
	{
 		CERR << "Error: Input file is not a valid 3D PGM\n";
 		return 1;
 	}
  // some images have 3 lines with #xdim 1, #ydim 1, #zdim 1
  fscanf(fp, "%s", buffer);
  while (buffer[0] == '#') {
    fscanf(fp, "%s", buffer);
    fscanf(fp, "%s", buffer);
  }
  sizeX = atoi(buffer);
  
  //fscanf(fp, "%d", &sizeX);
	fscanf(fp, "%d", &sizeY);
	fscanf(fp, "%d", &sizeZ);
	fscanf(fp, "%d", &max);

	// Create the inrimage
	COUT << "Converting file: " << std::flush;

	yav::Inrimage* img = new yav::Inrimage(sizeX, sizeY, sizeZ,
																				 yav::Inrimage::WT_UNSIGNED_CHAR,
																				 1, VM_INTERLACED,
																				 vx, vy, vz);
	char*** imgArray = (char***) img->getArray();
	int c;
	for (unsigned int k = 0; k < sizeZ; k++)
		for (unsigned int j = 0; j < sizeY; j++)
			for (unsigned int i = 0; i < sizeX; i++)
			{
				c = fgetc(fp);
				if ( c == EOF )
				{
					CERR << "Error: End of file has been reached prematurely. Conversion aborted\n";
					fclose(fp);
					delete img;
					return 1;
				}
				imgArray[k][j][i] = (unsigned char) c;
			}

	// Save the result
	img->write(argv[2]);

	COUT << "Ok\n";

	// Close input file and delete img
	fclose(fp);
	delete img;

	return 0;
}
