#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/BaseInrimage.h"
#include <inrimage/BaseInrimage.h>

#ifdef WIN32
#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>
#endif

namespace yav {

static void *CPLUSPLUSAlloc(size_t size) {
  return (void *) new unsigned char[size];
}


static void CPLUSPLUSDelete(void *pointer) {
  delete[] (unsigned char *) pointer;
}
/** dummy class whose constructor updates the list of supported file format */
class BaseInrimageInit {
public:
  /** actually register TclInrimage_Init in the hash table */
  BaseInrimageInit() {
    // XP 15-04-05 Set correct allocation schemes here otherwise
    // createInrimageFormat uses  malloc instead of new as allocation  
    setImageIOAllocationRoutines(CPLUSPLUSAlloc, CPLUSPLUSDelete);
    initSupportedFileFormat();
  }
  ~BaseInrimageInit() {
    //removeSupportedFileFormat();
  }
};
/// create a static object for automatic initialization
static BaseInrimageInit baseInrimageInit;

// exceptions
#ifndef LIBINRIMAGE
BaseInrimage::ImageNotFound::ImageNotFound(const std::string& from)
  : Exception(from, "unable to open inrimage file") {
}


BaseInrimage::UnknownImageType::UnknownImageType(const std::string& from)
  : Exception(from,
	      "unknown image type (known types are INRIMAGE4, GIF, IRIS ans ANALYZE)") {
}


BaseInrimage::InvalidHeader::InvalidHeader(const std::string& from)
  : Exception(from, "invalid inrimage header") {
}


BaseInrimage::InvalidData::InvalidData(const std::string& from)
  : Exception(from, "truncated image data") {
}


BaseInrimage::InvalidSize::InvalidSize(const std::string& from)
  : Exception(from,
	      "an image size cannot be null in one dimension") {
}


BaseInrimage::NotEnoughMemory::NotEnoughMemory(const std::string& from)
  : Exception(from, "unable to allocate memory") {
}


BaseInrimage::InvalidVectorialMode::InvalidVectorialMode(const std::string& from)
  : Exception(from,
	      "attempt to set a scalar mode to a vectorial image or a vectorial mode to a scalar image") {
}

BaseInrimage::InvalidVoxelSize::InvalidVoxelSize(const std::string& from)
  : Exception(from, "voxel size cannot be null or negative") {
}


BaseInrimage::UnableToCreateFile::UnableToCreateFile(const std::string& from)
  : Exception(from, "unable to open image file for writting") {
}


BaseInrimage::UnableToWriteHeader::UnableToWriteHeader(const std::string& from)
  : Exception(from,
	      "unable to write image header (no space left on device?)") {
}


BaseInrimage::UnableToWriteData::UnableToWriteData(const std::string& from)
  : Exception(from,
	      "unable to write image data (no space left on device?)") {
}
#endif






// Reads inrimage from file whose path is given as argument.
// File can be an INRIMAGE(4) gzipped or not, a GIF or an RGB
// irix file.
//
// if (vdim > 1 && filevm == VM_NON_INTERLACED)
//    if( vm == VM_NON_INTERLACED) 
//        _readImageData(img)
//    else 
//        _readNonInterlacedFileData(img)
// else 
//    if(vdim > 1 && vm == VM_NON_INTERLACED)
//        _readNonInterlacedImageData(img)
//    else
//        _readImageData(img)
//
//  

BaseInrimage::BaseInrimage(const char *name,
			   bool headerOnly,
			   VECTORIAL_MODE vm,
			   VECTORIAL_MODE filevm)
  throw(BaseInrimage::ImageNotFound, BaseInrimage::UnknownImageType,
	BaseInrimage::InvalidHeader, BaseInrimage::InvalidData)
  : useTransformation(false),xdim(1), ydim(1), zdim(1), vdim(1), vx(1.0), vy(1.0), vz(1.0),
    data(0), wdim(1), wordKind(WK_UNKNOWN), sign(SGN_UNKNOWN),
    vectMode(VM_SCALAR), user(0), nuser(0), imageFormat(0),
    info(0) {

  _image *img;
#ifndef LIBINRIMAGE
  char from[] = "BaseInrimage::BaseInrimage(const char *, bool, VECTORIAL_MODE)";
#endif
  
  // be sure to use correct memory allocation scheme
  setImageIOAllocationRoutines(CPLUSPLUSAlloc, CPLUSPLUSDelete);

  // read header
  img = readHeader(name);

  // initialize BaseInrimage
  xdim = img->xdim;
  ydim = img->ydim;
  zdim = img->zdim;
  vdim = img->vdim;
  wdim = img->wdim;
  vx = img->vx;
  vy = img->vy;
  vz = img->vz;

  translation=Vec3<double>(img->tx,img->ty,img->tz);
  RotVec3D rv=RotVec3D(img->rx,img->ry,img->rz);
  RotVec3DToRotMat3D(rv,rotation);


  if ((img->tx!=0) || (img->ty!=0) || (img->tz!=0) || (img->rx) || (img->ry) || (img->rz))
    useTransformation=true;
  updateTransformationMatrix();

  wordKind = img->wordKind;
  vectMode = img->vectMode;
  sign =img->sign;
  nuser = img->nuser;
  
  //  user = img->user;
  // copy header comments. Must reallocate memory because it has been allocated
  // with ImageIO_Alloc() in img data structure. (XP + HD)
   if( nuser > 0 )
   {
     unsigned int i;
     user = new char*[nuser];
     
     if( user == 0 )
#ifndef LIBINRIMAGE
       throw NotEnoughMemory(from);
#else
     throw NotEnoughMemory();
#endif

     for(i = 0 ; i < nuser ; ++i )
       {
         user[i] = new char[strlen(img->user[i]) + 1];
	 if( user[i] == 0 ) 
#ifndef LIBINRIMAGE
	   throw NotEnoughMemory(from) ;
#else
	 throw NotEnoughMemory();
#endif
         strcpy( user[i], img->user[i] ) ;
       }
   }   


  imageFormat = img->imageFormat;

  if(!headerOnly) {
    
    // read body
    if(img->openMode != OM_CLOSE) {


      // different vectorial mode #1
      // file is non interlaced
      if ( vdim > 1 && filevm == VM_NON_INTERLACED && vm != VM_NON_INTERLACED ) {
	if( _readNonInterlacedFileData(img) < 0 ) {
	  _freeImage(img);
#ifndef LIBINRIMAGE
	  throw InvalidData(from);
#else
	  throw InvalidData();
#endif
	}
      }

      // different vectorial mode #2
      // file is interlaced
      // read non interlaced vector image
      else if (vdim > 1 && filevm != VM_NON_INTERLACED && vm == VM_NON_INTERLACED ) {
	vectMode=VM_NON_INTERLACED;
	if( _readNonInterlacedImageData(img) < 0 ) {
	  _freeImage(img);
#ifndef LIBINRIMAGE
	  throw InvalidData(from);
#else
	  throw InvalidData();
#endif
	}
      }
      
      // scalar image
      // or same vectorial mode
      else {
	if( _readImageData(img) < 0 ) {
	  _freeImage(img);
#ifndef LIBINRIMAGE
	  throw InvalidData(from);
#else
	  throw InvalidData();
#endif
	}
      }
      ImageIO_close(img);
    }
  }

  else ImageIO_close(img);

  data = img->data;
  /// set buffer to zero before freeing memory
  img->data=0;
  _freeImage(img);
}









bool BaseInrimage::userField(const char *name, char *val, unsigned int n) const {
  unsigned int i, j, o, lu, ln;

  for(i = 0; i < nuser; i++) {
    o = 0;
    lu = strlen(user[i]);
    ln = strlen(name);
    while(o < lu && (user[i][o] == ' ' || user[i][o] == '\t')) o++;
    if(!strncmp(user[i] + o, name, ln) &&
       (user[i][o + ln] == '\0' || user[i][o + ln] == '='
	|| user[i][o + ln] == ' ' || user[i][o + ln] == '\t')) {
      if(val) {
	val[0] = '\0';
	o += ln;
	if(user[i][o] != '\0') {
	  while(o < lu && (user[i][o] == ' ' || user[i][o] == '\t')) o++;
	  if(user[i][o] == '=') {
	    o++;
	    while(o < lu && (user[i][o] == ' ' || user[i][o] == '\t')) o++;
	    if(user[i][o] != '\0') {
	      for(j = 0; (j < n) && (o < lu); j++, o++)
		val[j] = user[i][o];
	      val[j] = '\0';
	    }
	  }
	}
      }
      return true;
    }
  }

  return false;
}


void BaseInrimage::addUserField(const char *name, const char *val) {
   unsigned int ln;
   
   if( updateUserField( name, val ) == false ) {
      char** cpp = new char*[nuser+1] ;
      if( user != 0 ) {
         memcpy( cpp, user, nuser*sizeof(char*) ) ;
         delete[] user ;
      }
      user = cpp ;
      ln = strlen(name);
      if( val != 0 ) {
         user[nuser] = new char[ln+1+strlen(val)+1] ;
         strcpy( user[nuser], name ) ;
         user[nuser][ln]='=';
         strcpy( user[nuser]+ln+1, val ) ;
      } else {
         user[nuser] = new char[ln+1];
         strcpy( user[nuser], name ) ;
      }
      nuser++ ;
   }
}

bool BaseInrimage::removeUserField(const char *name) {
   size_t ln = strlen(name);
   for(size_t i = 0; i < nuser; i++) {
      size_t o = 0;
      size_t lu = strlen(user[i]);
      while(o < lu && (user[i][o] == ' ' || user[i][o] == '\t')) o++;
      if(!strncmp(user[i] + o, name, ln) &&
         (user[i][o + ln] == '\0' || user[i][o + ln] == '=' || user[i][o + ln] == ' ' 
          || user[i][o + ln] == '\t')) {
	 delete[] user[i] ;
	 for(size_t j=i;j<nuser-1;++j) user[j]=user[j+1];
	 user[--nuser]=0;
	 return true;
      }
   }
   return false;
}

bool BaseInrimage::updateUserField(const char *name, const char *val) const {
   unsigned int i, o, lu, ln;
   
   for(i = 0; i < nuser; i++) {
      ln = strlen(name);
      o = 0;
      lu = strlen(user[i]);
      while(o < lu && (user[i][o] == ' ' || user[i][o] == '\t')) o++;
      if(!strncmp(user[i] + o, name, ln) &&
         (user[i][o + ln] == '\0' || user[i][o + ln] == '=' || user[i][o + ln] == ' ' 
          || user[i][o + ln] == '\t')) {
         if(val) {
            delete[] user[i] ;
            user[i] = new char[ln+1+strlen(val)+1];
            strcpy( user[i], name ) ;
            user[i][ln]='=';
            strcpy( user[i]+ln+1, val ) ;
         }
         return true;
      }
   }
   return false;
}






















// Writes inrimage to file whose path is given as argument.
// If file name ends with '.gz' suffix, file is gzipped.
void BaseInrimage::write(const char *name /* inrimage file name */) const 
  throw(BaseInrimage::UnableToCreateFile,
	BaseInrimage::UnableToWriteHeader, BaseInrimage::UnableToWriteData) {

  _image *img;
  int error;

#ifndef LIBINRIMAGE
  char from[] = "BaseInrimage::BaseInrimage(const char *)";
#endif

  img = BaseInrimage2_image(*this);


  // special case
  // MINC image
#ifdef MINC_FILES
  if ( name && !strncmp(name + i - 4, ".mnc", 4) ) {
    img->imageFormat = IF_MINC;
    img->openMode = OM_FILE;
    {
      MincInformation *minfo = 0;
      double startx = 0.0, starty = 0.0, startz = 0.0, stepx, stepy, stepz;
      double def[3] = { 0.0, 0.0, 0.0 }, range[2];
      const char *sfile = 0;
      
      if(info && (minfo = dynamic_cast<MincInformation *>(info))) {
	startx = minfo->getStartX();
	starty = minfo->getStartY();
	startz = minfo->getStartZ();
	sfile = (const char *) minfo->getSourceFilename();
      }
      const double *Xcosine = minfo ? minfo->getXcosines() : def;
      const double *Ycosine = minfo ? minfo->getYcosines() : def;
      const double *Zcosine = minfo ? minfo->getZcosines() : def;
      
      stepx = minfo ? (minfo->getStepX() > 0.0 ? 1.0 : -1.0) * vx : vx;
      stepy = minfo ? (minfo->getStepY() > 0.0 ? 1.0 : -1.0) * vy : vy;
      stepz = minfo ? (minfo->getStepZ() > 0.0 ? 1.0 : -1.0) * vz : vz;
      
      getBounds(range[0], range[1]);
      if(writeMincFile(img, name, sfile, startx, starty, startz,
		       stepx, stepy, stepz,
		       Xcosine, Ycosine, Zcosine, range) < 0) {
	  // img allocated using ImageIO_alloc through  BaseInrimage2_image -> _initImage()
  // delete img;
  // should be deleted using ImageIO_free 
  ImageIO_free(img);
#ifndef LIBINRIMAGE
	throw UnableToWriteData(from);
#else
	throw UnableToWriteData();
#endif
      }
      img->openMode = OM_CLOSE; 
    }
  }
  else 
#endif
    
    {
      error = _writeImage( img, name );
#ifndef LIBINRIMAGE
      switch( error ) {
      case ImageIO_WRITING_HEADER :
	throw UnableToWriteHeader(from );
	break;
      case ImageIO_WRITING_IMAGE :
      case ImageIO_WRITING_DATA :
	throw UnableToWriteData(from );
	break;
      case ImageIO_UNKNOWN_TYPE :
      case ImageIO_OPENING :
	throw UnableToCreateFile(from );
	break;
      }
#else
      switch( error ) {
      case ImageIO_WRITING_HEADER :
	throw UnableToWriteHeader( );
	break;
      case ImageIO_WRITING_IMAGE :
      case ImageIO_WRITING_DATA :
	throw UnableToWriteData( );
	break;
      case ImageIO_UNKNOWN_TYPE :
      case ImageIO_OPENING :
	throw UnableToCreateFile( );
    break;
      }
#endif
    }

  // img allocated using ImageIO_alloc through  BaseInrimage2_image -> _initImage()
  // delete img;
  // should be deleted using ImageIO_free 
  ImageIO_free(img);

}






#ifdef OLDVERSION

// Writes inrimage to file whose path is given as argument.
// If file name ends with '.gz' suffix, file is gzipped.
void BaseInrimage::write(const char *name /* inrimage file name */) const 
  throw(BaseInrimage::UnableToCreateFile,
	BaseInrimage::UnableToWriteHeader, BaseInrimage::UnableToWriteData) {

  _image *img;
  int i;
#ifndef LIBINRIMAGE
  char from[] = "BaseInrimage::BaseInrimage(const char *)";
#endif

  img = BaseInrimage2_image(*this);

  if(!name) {
    img->openMode = OM_STD;
#if (defined _LINUX_) || (defined _SOLARIS_) || (defined _SGI_)
    img->fd = gzdopen(1, "wb");
#else
    img->fd = gzdopen(fileno(stdout), "wb");
#endif
  }
  
  else {
    i = strlen(name);
    if(!strncmp(name + i - 3, ".gz", 3)) {
      img->openMode = OM_GZ;
#ifdef WIN32
    int ffd=_open(name,_O_RDWR | _O_CREAT| _O_TRUNC | _O_BINARY, _S_IREAD|_S_IWRITE);
	img->fd=gzdopen(ffd,"wb");
#else
      img->fd = gzopen(name, "wb");
#endif
      img->imageFormat = IF_INR;
      if(!img->fd) {
	// img allocated using ImageIO_alloc through  BaseInrimage2_image -> _initImage()
	// delete img;
	// should be deleted using ImageIO_free 
	ImageIO_free(img);
#ifndef LIBINRIMAGE
	throw UnableToCreateFile(from);
#else
	throw UnableToCreateFile();
#endif
      }
    }
    
#ifdef MINC_FILES
    else if(!strncmp(name + i - 4, ".mnc", 4)) {
      img->imageFormat = IF_MINC;
      img->openMode = OM_FILE;
    }
#endif

    else {
      if(!strncmp(name + i - 4, ".hdr", 4)) img->imageFormat = IF_ANALYZE;
      else img->imageFormat = IF_INR;

      img->openMode = OM_FILE;
      img->fd = (gzFile) fopen(name, "wb");
      if(!img->fd) {
	// img allocated using ImageIO_alloc through  BaseInrimage2_image -> _initImage()
	// delete img;
	// should be deleted using ImageIO_free 
	ImageIO_free(img);
#ifndef LIBINRIMAGE
	throw UnableToCreateFile(from);
#else
	throw UnableToCreateFile();
#endif
      }
    }
  }


  switch(img->imageFormat) {

  case IF_ANALYZE:
    /* write header */
    if(writeAnalyzeHeader(img) < 0) {
      // img allocated using ImageIO_alloc through  BaseInrimage2_image -> _initImage()
      // delete img;
      // should be deleted using ImageIO_free 
      ImageIO_free(img);
#ifndef LIBINRIMAGE
      throw UnableToWriteHeader(from);
#else
      throw UnableToWriteHeader();
#endif
    }

    if(name) {
      int length = strlen(name);
      char* data_filename = new char[length + 5];
      
      strcpy(data_filename, name);
      if(!strcmp(name + length - 4, ".hdr"))
	strcpy(data_filename + length - 3, "img");
      else
	strcpy(data_filename + length, ".img");
      
      _closeImage(img);
      _openWriteImage(img, data_filename);
      delete[] data_filename;

      if(!img->fd) {
	// img allocated using ImageIO_alloc through  BaseInrimage2_image -> _initImage()
	// delete img;
	// should be deleted using ImageIO_free 
	ImageIO_free(img);
#ifndef LIBINRIMAGE
	throw UnableToCreateFile(from);
#else
	throw UnableToCreateFile();
#endif
      }
    }
	   
    /* write body */
    if(writeAnalyzeData(img) < 0) {
      // img allocated using ImageIO_alloc through  BaseInrimage2_image -> _initImage()
      // delete img;
      // should be deleted using ImageIO_free 
      ImageIO_free(img);
#ifndef LIBINRIMAGE
      throw UnableToWriteData(from);
#else
      throw UnableToWriteData();
#endif
    }
    break;

#ifdef MINC_FILES
  case IF_MINC:
    {
      MincInformation *minfo = 0;
      double startx = 0.0, starty = 0.0, startz = 0.0, stepx, stepy, stepz;
      double def[3] = { 0.0, 0.0, 0.0 }, range[2];
      const char *sfile = 0;

      if(info && (minfo = dynamic_cast<MincInformation *>(info))) {
	startx = minfo->getStartX();
	starty = minfo->getStartY();
	startz = minfo->getStartZ();
	sfile = (const char *) minfo->getSourceFilename();
      }
      const double *Xcosine = minfo ? minfo->getXcosines() : def;
      const double *Ycosine = minfo ? minfo->getYcosines() : def;
      const double *Zcosine = minfo ? minfo->getZcosines() : def;
      
      stepx = minfo ? (minfo->getStepX() > 0.0 ? 1.0 : -1.0) * vx : vx;
      stepy = minfo ? (minfo->getStepY() > 0.0 ? 1.0 : -1.0) * vy : vy;
      stepz = minfo ? (minfo->getStepZ() > 0.0 ? 1.0 : -1.0) * vz : vz;

      getBounds(range[0], range[1]);
      if(writeMincFile(img, name, sfile, startx, starty, startz,
		       stepx, stepy, stepz,
		       Xcosine, Ycosine, Zcosine, range) < 0) {
	// img allocated using ImageIO_alloc through  BaseInrimage2_image -> _initImage()
	// delete img;
	// should be deleted using ImageIO_free 
  ImageIO_free(img);
#ifndef LIBINRIMAGE
	throw UnableToWriteData(from);
#else
	throw UnableToWriteData();
#endif
      }
      img->openMode = OM_CLOSE;
    }
    break;
#endif

  case IF_INR:
  case IF_GIF:
  case IF_IRIS:
  default:
    /* write header */
    if(_writeInrimageHeader(img, END_UNKNOWN) < 0) {
      // img allocated using ImageIO_alloc through  BaseInrimage2_image -> _initImage()
      // delete img;
      // should be deleted using ImageIO_free 
  ImageIO_free(img);
#ifndef LIBINRIMAGE
      throw UnableToWriteHeader(from);
#else
      throw UnableToWriteHeader();
#endif
    }
      
    /* write body */
    else if(_writeInrimageData(img) < 0) {
      // img allocated using ImageIO_alloc through  BaseInrimage2_image -> _initImage()
      // delete img;
      // should be deleted using ImageIO_free 
      ImageIO_free(img);
#ifndef LIBINRIMAGE
      throw UnableToWriteData(from);
#else
      throw UnableToWriteData();
#endif
    }
    break;
  }

  /* close file descriptor */
  switch(img->openMode) {
  case OM_STD:
  case OM_CLOSE:
    break;
  case OM_GZ:
    gzclose(img->fd);
    break;
  case OM_FILE:
    fclose((FILE *) img->fd);
    break;
  }
  
  // img allocated using ImageIO_alloc through  BaseInrimage2_image -> _initImage()
  // delete img;
  // should be deleted using ImageIO_free 
  ImageIO_free(img);
}

#endif

















// Creates an inrimage with given size and word type.
// The inrimage data buffer is allocated
BaseInrimage::BaseInrimage(unsigned int _xdim,
			   unsigned int _ydim,
			   unsigned int _zdim,
			   unsigned int _vdim,
			   unsigned int _wdim,
			   WORD_KIND _wordKind,
			   SIGN _sign,
			   float _vx,
			   float _vy,
			   float _vz)
  throw(InvalidSize, NotEnoughMemory, InvalidVoxelSize)
  : useTransformation(false),xdim(_xdim), ydim(_ydim), zdim(_zdim), vdim(_vdim),
  vx(_vx), vy(_vy), vz(_vz), data(0), wdim(_wdim),
  wordKind(_wordKind), sign(_sign), user(0), nuser(0),
  imageFormat(0), info(0) {
  
   if(vx <= 0.0 || vy <= 0.0 || vz <= 0.0)
#ifndef LIBINRIMAGE
      throw InvalidVoxelSize("BaseInrimage::BaseInrimage(unsigned int, unsigned int, unsigned int, unsigned int, unsigned int, WORD_KIND, SIGN");
#else
      throw InvalidVoxelSize();
#endif
      
   if(xdim <= 0 || ydim <= 0 || zdim <= 0 || vdim <= 0 || wdim <= 0)
#ifndef LIBINRIMAGE
      throw InvalidSize("BaseInrimage::BaseInrimage(unsigned int, unsigned int, unsigned int, unsigned int, unsigned int, WORD_KIND, SIGN");
#else
      throw InvalidSize();
#endif
  updateTransformationMatrix();

  if(vdim == 1) vectMode = VM_SCALAR;
  else vectMode = VM_INTERLACED;
  allocate();
}


// Creates an inrimage with given size and word type.
// If data is 0, the image is not allocated.
// If data is non 0, image data points to it.
BaseInrimage::BaseInrimage(unsigned int _xdim,
			   unsigned int _ydim,
			   unsigned int _zdim,
			   void *_data,
			   unsigned int _vdim,
			   unsigned int _wdim,
			   WORD_KIND _wordKind,
			   SIGN _sign,
			   float _vx,
			   float _vy,
			   float _vz) throw(BaseInrimage::InvalidSize, InvalidVoxelSize) :
  useTransformation(false),xdim(_xdim), ydim(_ydim), zdim(_zdim), vdim(_vdim), 
  vx(_vx), vy(_vy), vz(_vz), data(_data), wdim(_wdim),
  wordKind(_wordKind), sign(_sign), user(0), nuser(0),
  imageFormat(0), info(0) {
 
   if(vx <= 0.0 || vy <= 0.0 || vz <= 0.0)
#ifndef LIBINRIMAGE
      throw InvalidVoxelSize("BaseInrimage::BaseInrimage(unsigned int, unsigned int, unsigned int, unsigned int, unsigned int, WORD_KIND, SIGN");
#else
      throw InvalidVoxelSize();
#endif
      
   if(xdim <= 0 || ydim <= 0 || zdim <= 0 || vdim <= 0 || wdim <= 0)
  if(xdim <= 0 || ydim <= 0 || zdim <= 0 || vdim <= 0 || wdim <= 0)
#ifndef LIBINRIMAGE
      throw InvalidSize("BaseInrimage::BaseInrimage(unsigned int, unsigned int, unsigned int, void *, unsigned int, unsigned int, WORD_KIND, SIGN");
#else
      throw InvalidSize();
#endif
  updateTransformationMatrix();
  if(vdim == 1) vectMode = VM_SCALAR;
  else vectMode = VM_INTERLACED;
}


/// Creates an replica of img.
BaseInrimage::BaseInrimage(const BaseInrimage& img)
   throw(InvalidSize,InvalidVectorialMode,InvalidVoxelSize,NotEnoughMemory)
  :  rotation(img.rotation),
     translation(img.translation),
     useTransformation(img.useTransformation),
     xdim(img.getX()),
     ydim(img.getY()),
     zdim(img.getZ()),
     vdim(img.getVdim()),
     vx(img.getVX()),
     vy(img.getVY()),
     vz(img.getVZ()),
     data(0),
     wdim(img.getWdim()),
     wordKind(img.getWordKind()),
     sign(img.getWordSign()),
     vectMode(img.getVectorialMode()),
     user(0),
     nuser(img.nuser),
     imageFormat(img.getFileFormat()),
     info(0)
{
  updateTransformationMatrix();
#ifndef LIBINRIMAGE
    char from[] = "BaseInrimage::BaseInrimage(const BaseInrimage&)";
#endif

   unsigned long int size =
      img.getX()*img.getY()*img.getZ()*img.getVdim()*img.getWordSize() ;
#ifdef __GNUC__
   try{
#endif
      data = new unsigned char[size];
#ifdef __GNUC__
   } catch(std::bad_alloc&){
#ifndef LIBINRIMAGE
      throw NotEnoughMemory(from);
#else
      throw NotEnoughMemory();
#endif
   }
#endif
   memcpy( getData(), img.getData(), size ) ;
   
   // copy header comments
   if( img.nuser > 0 )
   {
      user = new char*[img.nuser];

      if( user == 0 )
#ifndef LIBINRIMAGE
	  throw NotEnoughMemory(from);
#else
	  throw NotEnoughMemory();
#endif
      
      for( unsigned int i = 0 ; i < nuser ; ++i )
      {
         user[i] = new char[strlen(img.user[i]) + 1];
	 if( user[i] == 0 ) 
#ifndef LIBINRIMAGE
	     throw NotEnoughMemory(from) ;
#else
	     throw NotEnoughMemory();
#endif
         strcpy( user[i], img.user[i] ) ;
      }
   }   
}

/// Make *this an replica of img.
BaseInrimage& BaseInrimage::operator=(const BaseInrimage& img)
   throw(InvalidSize,InvalidVectorialMode,InvalidVoxelSize,NotEnoughMemory)
{
   if( &img != this )
   {
#ifndef LIBINRIMAGE
       char from[] = "BaseInrimage::operator=(const BaseInrimage&)";
#endif

      // resize image
      resize( img.getX(), img.getY(), img.getZ(), img.getVdim(),
	      img.getWordSize(), img.getWordKind(), img.getWordSign() ) ;
      
      // copy data 
      unsigned long int size =
	 img.getX()*img.getY()*img.getZ()*img.getVdim()*img.getWordSize() ;
      if( data == 0 )
      {
#ifdef __GNUC__
	 try{
#endif
	    data = new unsigned char[size];
#ifdef __GNUC__
	 }catch(std::bad_alloc&){
#ifndef LIBINRIMAGE
	    throw NotEnoughMemory(from) ;
#else
	    throw NotEnoughMemory();
#endif
	 }
#endif
      }
      // should check for overlap?
      memcpy( getData(), img.getData(), size ) ;
      
      // copy header comments
      nuser = img.nuser ;
      if( img.nuser > 0 )
      {
	 user = new char*[img.nuser];
	 
	 if( user == 0 )
#ifndef LIBINRIMAGE
	     throw NotEnoughMemory(from) ;
#else
	     throw NotEnoughMemory();
#endif
	 
	 for( unsigned int i = 0 ; i < nuser ; ++i )
	 {
	    user[i] = new char[strlen(img.user[i])];
	    if( user[i] == 0 )
#ifndef LIBINRIMAGE
	     throw NotEnoughMemory(from) ;
#else
	     throw NotEnoughMemory();
#endif
	    strcpy( user[i], img.user[i] ) ;
	 }
      }
      
      // update format fields
      setVectorialMode( img.getVectorialMode() ) ;
      setFileFormat( img.getFileFormat() ) ;
      setVoxelSize( img.getVX(), img.getVY(), img.getVZ() ) ;
   }
   return *this ;
}   

// Properly frees image object
BaseInrimage::~BaseInrimage() {
  if(data) delete[] (unsigned char *) data;
  data = 0 ;
  if(nuser > 0) {
    for(unsigned int i = 0; i < nuser; i++) {
       delete[] user[i];
       user[i] = 0;
    }
    delete[] user;
    user = 0;
  }
  if(info) delete info;
}


// Allocates and fills with zero image data buffer. If data
// is not null, buffer is first fried.
void BaseInrimage::allocate(void) throw(NotEnoughMemory) {
  unsigned LONGINT size;

  if(data) delete[] (unsigned char *) data;
  size = xdim * ydim * zdim * vdim * wdim;
#ifdef __GNUC__
  try{
#endif
     data = new unsigned char[size];
#ifdef __GNUC__
  }catch(std::bad_alloc&){
#ifndef LIBINRIMAGE
     throw NotEnoughMemory("BaseInrimage::allocate(void)") ;
#else
     throw NotEnoughMemory();
#endif
  }
#endif
  memset((char *) data, 0, size);
}


// Frees image data buffer
void BaseInrimage::free(void) {
  if(data) delete[] (unsigned char *) data;
  data = 0;
}

// Set a new image voxel size
void BaseInrimage::setVoxelSize(float _vx, float _vy, float _vz)
  throw(BaseInrimage::InvalidVoxelSize) {

  if(_vx <= 0.0 || _vy <= 0.0 || _vz <= 0.0)
#ifndef LIBINRIMAGE
      throw InvalidVoxelSize("BaseInrimage::setVoxelSize");
#else
      throw InvalidVoxelSize();
#endif
  vx = _vx;
  vy = _vy;
  vz = _vz;
}

// Set a new data buffer. Returns old image data buffer.
void *BaseInrimage::setData(void *_data ) {
  void *tmp = data;
  data = _data;
  return tmp;
}


// Set user defined parameters strings to write into image file
// header.
void BaseInrimage::setUserParams(unsigned int n,
				 char **params) {
  if(nuser > 0) {
    for(unsigned int i = 0; i < nuser; i++) delete[] user[i];
    delete[] user;
  }
  nuser = n;
  user = params;
}

// Changes image size. If image data buffer existed it is fried
// and reallocated.
void BaseInrimage::resize(unsigned int _xdim,
			  unsigned int _ydim,
			  unsigned int _zdim,
			  unsigned int _vdim,
			  unsigned int _wdim,
			  WORD_KIND _wordKind,
			  SIGN _sign)
  throw(BaseInrimage::InvalidSize, BaseInrimage::NotEnoughMemory) {
#ifndef LIBINRIMAGE
    char from[] = "BaseInrimage::resize";
#endif

  if(_xdim <= 0 || _ydim <= 0 || _zdim <= 0 || _vdim <= 0 || _wdim <= 0)
#ifndef LIBINRIMAGE
    throw InvalidSize(from);
#else
    throw InvalidSize();
#endif

  wordKind = _wordKind;
  sign = _sign;
  if(_xdim != xdim || _ydim != ydim || _zdim != zdim || _vdim != vdim || _wdim != wdim)
  {
     xdim = _xdim;
     ydim = _ydim;
     zdim = _zdim;
     vdim = _vdim;
     wdim = _wdim;
     if(data) allocate();
  }
}
  

// Changes image word type. If image data buffer existed it is fried
// and reallocated.
void BaseInrimage::setWords(WORD_KIND _wordKind, SIGN _sign)
  throw(BaseInrimage::NotEnoughMemory) {

  wordKind = _wordKind;
  sign = _sign;
  if(data) allocate();
}



// Changes image vectorial mode. A vectorial image buffer can be
// stored either interlaced (default) or non interlaced.
void BaseInrimage::setVectorialMode(VECTORIAL_MODE vm)
  throw(BaseInrimage::InvalidVectorialMode) {

#ifndef LIBINRIMAGE
    char from[] = "BaseInrimage::setVectorialMode";
#endif

  if(vdim == 1) {
    if(vm != VM_SCALAR)
#ifndef LIBINRIMAGE
	throw InvalidVectorialMode(from);
#else
	throw InvalidVectorialMode();
#endif
    vectMode = VM_SCALAR;
  }

  else {
    if(vm == VM_SCALAR)
#ifndef LIBINRIMAGE
	throw InvalidVectorialMode(from);
#else
	throw InvalidVectorialMode();
#endif
    vectMode = vm;    
  }
}



// Converts an _image structure to a BaseInrimage
BaseInrimage *BaseInrimage::_image2BaseInrimage(const _image& img)
  throw(BaseInrimage::InvalidSize, BaseInrimage::InvalidVoxelSize,
	BaseInrimage::InvalidVectorialMode) {

  BaseInrimage *inr = 0;

  inr = new BaseInrimage(img.xdim, img.ydim, img.zdim, 0,
			 img.vdim, img.zdim, img.wordKind, img.sign);

  try {
    inr->setVectorialMode(img.vectMode);
    inr->setVoxelSize(img.vx, img.vy, img.vz);
  }
#ifndef LIBINRIMAGE
  catch(Exception& e) {
    delete inr;
    throw e;
  }
#else
  catch(ImageException& e) {
    delete inr;
    throw e;
  }
#endif

  inr->setUserParams(img.nuser, img.user);

  inr->data = img.data;
  inr->imageFormat = img.imageFormat;

  return inr;
}



// Converts a BaseImage structure to an _image
_image *BaseInrimage::BaseInrimage2_image(const BaseInrimage& inr) {
  _image *img;

  img = _initImage();
  
  img->xdim = inr.xdim;
  img->ydim = inr.ydim;
  img->zdim = inr.zdim;
  img->vdim = inr.vdim;
  img->wdim = inr.wdim;
  img->wordKind = inr.wordKind;
  img->sign = inr.sign;
  img->data = inr.data;
  img->vectMode = inr.vectMode;
  img->nuser = inr.nuser;
  img->user = inr.user;
  img->vx = inr.vx;
  img->vy = inr.vy;
  img->vz = inr.vz;
  img->tx=inr.translation.getX();
  img->ty=inr.translation.getY();
  img->tz=inr.translation.getZ();
  
  RotVec3D rv;
  RotMat3DToRotVec3D(inr.rotation,rv);
  img->rx=rv(0);
  img->ry=rv(1);
  img->rz=rv(2);

  img->imageFormat = inr.imageFormat;

  return img;
}





// exceptions:
//   ImageNotFound:    erreur lors de l'ouverture
//   InvalidHeader:    erreur lors de la lecture de l'entete
//                     ou du fichier
//   UnknownImageType: type d'image inconnu

_image *BaseInrimage::readHeader(const char *name)
  throw(BaseInrimage::ImageNotFound, BaseInrimage::InvalidHeader,
	BaseInrimage::UnknownImageType) {
  _image *img;
  int error = ImageIO_NO_ERROR;
#ifndef LIBINRIMAGE
  char from[] = "BaseInrimage::readHeader";
#endif

  img = _readImageHeaderAndGetError( name, &error );
  
#ifndef LIBINRIMAGE
  switch( error ) {
  case ImageIO_UNKNOWN_TYPE :
    throw UnknownImageType(from);
    break;
  case ImageIO_READING_HEADER :
  case ImageIO_READING_IMAGE :
    throw InvalidHeader(from);
    break;
  case ImageIO_OPENING :
    throw ImageNotFound(from);
    break;
  }
#else
  switch( error ) {
  case ImageIO_UNKNOWN_TYPE :
    throw UnknownImageType( );
    break;
  case ImageIO_READING_HEADER :
  case ImageIO_READING_IMAGE :
    throw InvalidHeader( );
    break;
  case ImageIO_OPENING :
    throw ImageNotFound( );
    break;
  }
#endif

  return img;
}


#ifdef OLDVERSION

_image *BaseInrimage::oldReadHeader(const char *name)
  throw(BaseInrimage::ImageNotFound, BaseInrimage::InvalidHeader,
	BaseInrimage::UnknownImageType) {

  _image *img;
  int res = 0;
  char magic[5];
#ifndef LIBINRIMAGE
  char from[] = "BaseInrimage::readHeader";
#endif

  img = _initImage();
  _openReadImage(img, name);
  if(!img->fd)
#ifndef LIBINRIMAGE
      throw ImageNotFound(from);
#else
      throw ImageNotFound();
#endif

  /* read magic number */
  if(img->openMode != OM_STD) {
    gzread(img->fd, (void *) magic, 4);
    magic[4] = '\0';
    gzseek(img->fd, 0L, SEEK_SET);
  }

  /* openned image is an inrimage */
  if(img->openMode == OM_STD || !strcmp(magic, INR_MAGIC)) {
    if(readInrimageHeader(img) < 0) {
      _freeImage(img);
#ifndef LIBINRIMAGE
     throw InvalidHeader(from);
#else
      throw InvalidHeader();
#endif
    }
    img->imageFormat = IF_INR;
  }

  /* openned image is an ANALYZE */
  else if(!memcmp(magic, ANALYZE_LE_MAGIC, 4) ||
	  !memcmp(magic, ANALYZE_BE_MAGIC, 4)) {
    if( !readAnalyzeHeader(img, name) ) {
      _freeImage(img);
#ifndef LIBINRIMAGE
     throw InvalidHeader(from);
#else
      throw InvalidHeader();
#endif
    }
    img->imageFormat = IF_ANALYZE;
  }

#ifdef MINC_FILES
  else if(!memcmp(magic, MINC_MAGIC, 3)) {
    double startx, starty, startz, stepx, stepy, stepz,
      Xcosine[3], Ycosine[3], Zcosine[3];

    if( !readMincHeader(img, name, &startx, &starty, &startz,
			&stepx, &stepy, &stepz,
			Xcosine, Ycosine, Zcosine) ) {
      _freeImage(img);
#ifndef LIBINRIMAGE
     throw InvalidHeader(from);
#else
     throw InvalidHeader();
#endif
    }
    img->imageFormat = IF_MINC;
    info = new MincInformation(name, startx, starty, startz,
			       stepx, stepy, stepz,
			       Xcosine, Ycosine, Zcosine);
  }
#endif

  else {
    /* opened image is a GIF: full read */
    if(!strcmp(magic, GIF_MAGIC)) {
      res = readGifImage(img,name);
      img->imageFormat = IF_GIF;
    }  
    
    /* opened image is an RGB: full read */
    else if((((unsigned char *)magic)[0]<<8) + ((unsigned char *)magic)[1]
	    == IRIS_MAGIC) {
      res = readIrisImage(img, name);
      img->imageFormat = IF_IRIS;
    }
    
    /* unknown image type */
    else {
      _freeImage(img);
#ifndef LIBINRIMAGE
     throw UnknownImageType(from);
#else
      throw UnknownImageType();
#endif
    }

    /* close image file and return */
    if(res > 0) {
      _closeImage(img);
      img->fd = 0;
      img->openMode = OM_CLOSE;
    }
    else {
      _freeImage(img);
#ifndef LIBINRIMAGE
     throw InvalidHeader(from);
#else
      throw InvalidHeader();
#endif
    }
  }

  return img;
}

#endif // OLDVERSION




// toggle buffer to produce an interlaced vector image (for
// vectorial images only)
void BaseInrimage::setInterlacedData(void) {
  unsigned char *src, *dest;
  unsigned long int size, i;
  unsigned int j, k;

  if(vdim > 1 && vectMode == VM_NON_INTERLACED) {
    size = xdim * ydim * zdim;
    dest = new unsigned char[size * vdim * wdim];
    for(i = 0; i < size; i++) {
      src = (unsigned char *) data + i * wdim;
      for(j = 0; j < vdim; j++, src += size * wdim) {
	for(k = 0; k < wdim; k++) *dest++ = *(src + k);
      }
    }
    dest -= size * vdim * wdim;
    delete[] (unsigned char *) setData((void *) dest);
    setVectorialMode(VM_INTERLACED);
  }
}


// toggle buffer to produce an interlaced vector image (for
// vectorial images only)
void BaseInrimage::setNonInterlacedData(void) {
  unsigned char *src, *dest, *ptr;
  unsigned long int size, i;
  unsigned int j, k;

  if(vdim > 1 && vectMode == VM_INTERLACED) {
    size = xdim * ydim * zdim;
    dest = new unsigned char[size * vdim * wdim];
    src = (unsigned char *) data;
    for(i = 0; i < size; i++) {
      ptr = (unsigned char *) dest + i * wdim;
      for(j = 0; j < vdim; j++, ptr += size * wdim) {
	for(k = 0; k < wdim; k++) *(ptr+k) = *src++;
      }
    }
    delete[] (unsigned char *) setData((void *) dest);
    setVectorialMode(VM_NON_INTERLACED);
  }
}

#define INTERNALGETBOUNDS(TYPE,A,B) internalGetBounds((TYPE) 0,A,B)

void BaseInrimage::getBounds(double &min, double &max) const {
  switch(wordKind) {
  case WK_FIXED:

    switch(wdim) {
    case 1:
        INTERNALGETBOUNDS(unsigned char,min,max);
      return; 
    case 2:
      if(sign == SGN_SIGNED)
        INTERNALGETBOUNDS(short int,min,max);
      else
         INTERNALGETBOUNDS(unsigned short int,min,max);
      return;
    case 4:
      if(sign == SGN_SIGNED)
	INTERNALGETBOUNDS(int,min,max);
      else
	INTERNALGETBOUNDS(unsigned int,min,max);
      return;
    case 8:
      if(sign == SGN_SIGNED)
	INTERNALGETBOUNDS(LONGINT,min,max);
      else
	INTERNALGETBOUNDS(unsigned LONGINT,min,max);
      return;
    }
    break;

  case WK_FLOAT:

    switch(wdim) {
    case 4:
     	INTERNALGETBOUNDS(float,min,max);
      return;
    case 8:
          	INTERNALGETBOUNDS(double,min,max);
      return;
    }
    break;
    
  default:
    break;
  }

  min = 0.0;
  max = 1.0;
}

void BaseInrimage::updateTransformationMatrix() {
  transformationMatrix=Matrix4x4<double>(rotation,translation);
  transformationMatrix[0][0]*=vx;
  transformationMatrix[1][0]*=vx;
  transformationMatrix[2][0]*=vx;  
  transformationMatrix[0][1]*=vy;
  transformationMatrix[1][1]*=vy;
  transformationMatrix[2][1]*=vy;
  transformationMatrix[0][2]*=vz;
  transformationMatrix[1][2]*=vz;
  transformationMatrix[2][2]*=vz;
  if (translation.norm2()>0.0 || rotation.Norm()>0.0)
    useTransformation=true;      
}


#ifdef MINC_FILES

#include <unistd.h>

MincInformation::MincInformation(const char *sn,
				 double sx, double sy, double sz,
				 double px, double py, double pz,
				 double *xc, double *yc, double *zc)
  : sourceName((const char *) 0) {
  if(sn) {
    if(sn[0] != '/') {
      char path[4096];
      getcwd(path, 4096);
      sourceName = path;
      sourceName += "/";
      sourceName += sn;
    }
    else sourceName = sn;
  }
  start[0] = sx;
  start[1] = sy;
  start[2] = sz;
  step[0] = px;
  step[1] = py;
  step[2] = pz;

  for(int i = 0; i < 3; i++) {
    Xcosine[i] = xc[i];
    Ycosine[i] = yc[i];
    Zcosine[i] = zc[i];
  }
}
#endif


} // namespace yav

