/******************************************************************************/
/* ChamferMask.cpp                                                            */
/******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifdef _YAV_
#include "ChamferMask.h"
#include "preDefinedMasks.h"
using namespace yav;
#else
#include "distancemap/ChamferMask.h"
#include "distancemap/preDefinedMasks.h"
#endif


/****************************************************************************/
/*                                                                          */
/*                               CONSTRUCTORS                               */
/*                                                                          */
/****************************************************************************/
ChamferMask::ChamferMask(int size, int dim, int flag)
{
  preDefinedMasks preDefMask;
  
  switch(dim)
    {
    case 2 :
      switch(size)
	{
	default :
	  fprintf(stderr, "No such mask implemented yet\n Using 3x3 chamfer mask by default\n");
	case 3 :
	  if (flag == 1) // initialization mask
	    initMask2D(3, preDefMask.chamfer2D3x3Init, preDefMask.epsilon2D3x3);
	  else           // computation mask
	    initMask2D(3, preDefMask.chamfer2D3x3, preDefMask.epsilon2D3x3);
	  this->maxRelError = preDefMask.maxRelErr2D3x3;
	  break;
	case 5 :
	  if (flag == 1) // initialization mask
	    initMask2D(3, preDefMask.chamfer2D5x5Init,preDefMask.epsilon2D5x5);
	  else           // computation mask
	    initMask2D(5, preDefMask.chamfer2D5x5, preDefMask.epsilon2D5x5);
	  this->maxRelError = preDefMask.maxRelErr2D5x5;
	  break;
	case 7 :
	  if (flag == 1) // initialization mask
	    initMask2D(3, preDefMask.chamfer2D7x7Init, preDefMask.epsilon2D7x7);
	  else           // computation mask
	    initMask2D(7, preDefMask.chamfer2D7x7, preDefMask.epsilon2D7x7);
	  this->maxRelError = preDefMask.maxRelErr2D7x7;
	  break;
	}
      break;
      
    default :
      fprintf(stderr, "No such mask implemented yet\n Using 3D chamfer mask by default\n");
    case 3 :
      switch(size)
	{
	default :
	  fprintf(stderr, "No such mask implemented yet\n Using 3x3x3 chamfer mask by default\n");
	case 3 :
	  if (flag == 1) // initialization mask
	    initMask3D(3, preDefMask.chamfer3D3x3x3Init, preDefMask.epsilon3D3x3x3);
	  else           // computation mask
	    initMask3D(3, preDefMask.chamfer3D3x3x3, preDefMask.epsilon3D3x3x3);
	  this->maxRelError = preDefMask.maxRelErr3D3x3x3;
	  break;
	case 5 :
	  if (flag == 1) // initialization mask
	    initMask3D(3, preDefMask.chamfer3D5x5x5Init, preDefMask.epsilon3D5x5x5);
	  else           // computation mask
	    initMask3D(5, preDefMask.chamfer3D5x5x5, preDefMask.epsilon3D5x5x5);
	  break;
	  this->maxRelError = preDefMask.maxRelErr3D5x5x5;
	case 7 :
	  if (flag == 1) // initialization mask
	    initMask3D(3, preDefMask.chamfer3D7x7x7Init, preDefMask.epsilon3D7x7x7);
	  else           // computation mask
	    initMask3D(7, preDefMask.chamfer3D7x7x7, preDefMask.epsilon3D7x7x7);
	  this->maxRelError = preDefMask.maxRelErr3D7x7x7;
	  break;
	}
      break;
    }
}

/************** 2D constructor **********************************************/
ChamferMask::ChamferMask(int size, unsigned char ** coeffs, float epsilon, float maxRelError)
{
  if (epsilon == -1)
    initMask2D(size, coeffs, (float) coeffs[1][0]);
  else
    initMask2D(size, coeffs, epsilon);

  this->maxRelError = maxRelError;
}

/************** 3D constructor *********************************************/
ChamferMask::ChamferMask(int size, unsigned char *** coeffs, float epsilon, float maxRelError)
{
  if (epsilon == -1)
    initMask3D(size, coeffs, (float) coeffs[1][0][0]);
  else
    initMask3D(size, coeffs, epsilon);
  
  this->maxRelError = maxRelError;
}


/************** Initialization functions ************************************/
void
ChamferMask::initMask2D(int size, unsigned char **  coeffs, float epsilon)
{
  int i, j;
  int x, y, z, m;
  int xBis, yBis, zBis;
  int dimT;
  bool cond;
  chamferElmt * maskElmt;
  
  chamferSize = size;
  iBegin = chamferSize / 2;
  jBegin = chamferSize / 2;
  kBegin = 0;  // every slice can be treated

  isSetOffsets = false;

  this->epsilon = epsilon;

  // we considere only the forward neighborhood, or only the backward one, so
  // k = 0 and { j >= 0, and for j = 0, {i > 0}}
  //  privateMask = new (chamferElement)[chamferSize][chamferSize][1];
  privateMask = (chamferElmt ***) malloc(chamferSize * sizeof(chamferElmt **));
  for (i = 0; i < chamferSize; i++)
    {
      privateMask[i] = (chamferElmt **) malloc(chamferSize * sizeof(chamferElmt *));
      for (j = 0; j < chamferSize; j++)
	privateMask[i][j] = (chamferElmt *) malloc(1 * sizeof(chamferElmt));
    }

  dimT = 1 * chamferSize * chamferSize;
  chamferMask =  (unsigned char *) malloc(dimT * sizeof(unsigned char));
  neighborhood = (int *)  malloc(dimT * sizeof(int));

  maskSize = 0;
  m = 0;
  
  z = 0; 
  zBis = z;
  for (y = - jBegin; y <= jBegin; y++)
    {
      yBis = y + jBegin;
      for (x = - iBegin; x <= iBegin; x++)
	{
	  xBis = x + iBegin;
	  maskElmt = &(privateMask[xBis][yBis][zBis]);
	  
	  // we consider only the forward mask.
	  cond = (y < 0) || ((y == 0) && (x <= 0));
	  
	  if ((coeffs[abs(x)][abs(y)] == 0) || cond)
	    {
	      maskElmt->exist = false;
	      maskElmt->value = 0;
	    }
	  else 
	    {
	      maskElmt->exist = true;
	      maskElmt->value = (coeffs[abs(x)][abs(y)]);
	      
	      chamferMask[m]  = maskElmt->value;
	      m++;
	      maskSize++;
	    }
	}
    }
}

void
ChamferMask::initMask3D(int size, unsigned char *** coeffs, float epsilon)
{
  int i, j;
  int x, y, z, m;
  int xBis, yBis, zBis;
  int dim_pos, dimT;
  bool cond;
  chamferElmt * maskElmt;
  
  chamferSize = size;
  iBegin = chamferSize / 2;
  jBegin = chamferSize / 2;
  kBegin = chamferSize / 2;

  isSetOffsets = false;

  this->epsilon = epsilon;

  // we considere only the forward neighborhood, or only the backward one, so
  // k >= 0 and for k = 0 { j >= 0, and for j = 0, {i > 0}}
  dim_pos = chamferSize / 2 + 1;  
  //  privateMask = new (chamferElement)[chamferSize][chamferSize][dim_pos];
  privateMask = (chamferElmt ***) malloc(chamferSize * sizeof(chamferElmt **));
  for (i = 0; i < chamferSize; i++)
    {
      privateMask[i] = (chamferElmt **) malloc(chamferSize * sizeof(chamferElmt *));
      for (j = 0; j < chamferSize; j++)
	privateMask[i][j] = (chamferElmt *) malloc(dim_pos * sizeof(chamferElmt));
    }

  dimT = dim_pos * chamferSize * chamferSize;
  chamferMask =  (unsigned char *) malloc(dimT * sizeof(unsigned char));
  neighborhood = (int *)  malloc(dimT * sizeof(int));

  maskSize = 0;
  m = 0;
  for (z = 0; z <= kBegin; z++)  // k >= 0
    {
      zBis = z;
      for (y = - jBegin; y <= jBegin; y++)
	{
	  yBis = y + jBegin;
	  for (x = - iBegin; x <= iBegin; x++)
	    {
	      xBis = x + iBegin;
	      maskElmt = &(privateMask[xBis][yBis][zBis]);

	      // we consider only the forward mask.
	      cond = (z == 0) && ((y < 0) || ((y == 0) && (x <= 0)));

	      if ((coeffs[abs(x)][abs(y)][abs(z)] == 0) || cond)
		{
		  maskElmt->exist = false;
		  maskElmt->value = 0;
		}
	      else 
		{
		  maskElmt->exist = true;
		  maskElmt->value = (coeffs[abs(x)][abs(y)][abs(z)]);
		  
		  chamferMask[m]  = maskElmt->value;
		  m++;
		  maskSize++;
		}
	    }
	}
    }
}
/****************************************************************************/
/*                                                                          */
/*                             DESTRUCTORS                                  */
/*                                                                          */
/****************************************************************************/
ChamferMask::~ChamferMask()
{
  int i, j;
  
  for (i = 0; i < chamferSize; i++)
    {
      for (j = 0; j < chamferSize; j++)
	free (privateMask[i][j]);
      free (privateMask[i]);
    }
  
  free (privateMask) ;

  free (chamferMask) ;
  free (neighborhood) ;
}


/****************************************************************************/
/*                                                                          */
/*                             ACCESSORS                                    */
/*                                                                          */
/****************************************************************************/
int 
ChamferMask::getMaskSize()
{
  return maskSize;
}

unsigned char * 
ChamferMask::getChamferMask()
{
  return chamferMask;
}

int * 
ChamferMask::getNeighborhood()
{
  return neighborhood;
}

int 
ChamferMask::getChamferSize()
{
  return chamferSize;
}


int
ChamferMask::getIBegin()
{
  return iBegin;
}

int
ChamferMask::getJBegin()
{
  return jBegin;
}

int 
ChamferMask::getKBegin()
{
  return kBegin;
}

bool 
ChamferMask::getIsSetOffsets()
{
  return isSetOffsets;
}

float
ChamferMask::getEpsilon()
{
  return epsilon;
}

float
ChamferMask::getMaxRelativeError()
{
  return maxRelError;
}

void
ChamferMask::getPrivatMask(int x, int y, int z, chamferElmt * elmt)
{
  bool cond;
  int xBis, yBis, zBis;
  cond = ( (z < 0)        || (z > kBegin) || 
	   (y < - jBegin) || (y > jBegin) ||
	   (x < - iBegin) || (x > iBegin) );
  
  if (cond) {
    printf("privateMask[%i][%i][%i] does not exists, sorry\n", x, y, z);
    exit(-1);
  }
  
  zBis = z;
  yBis = y + jBegin;
  xBis = x + iBegin;

  elmt->exist  = (privateMask[xBis][yBis][zBis]).exist;
  elmt->offset = (privateMask[xBis][yBis][zBis]).offset;
  elmt->value  = (privateMask[xBis][yBis][zBis]).value;
}


/****************************************************************************/
/*                                                                          */
/*                             PUBLIC METHODS                               */
/*                                                                          */
/****************************************************************************/

/****************************************************************************/
/*                             Initialization                               */
/****************************************************************************/
void
ChamferMask::initOffsets(int dimX, int dimY, int dimZ)
{
  int m;
  int offset;
  int x, y, z, xBis, yBis, zBis;
  chamferElmt * maskElmt;
  
  m = 0;
  for (z = 0; z <= kBegin; z++)
    {
      zBis = z;
      for (y = -jBegin; y <= jBegin; y++)
	{
	  yBis = y + jBegin;

	  for (x = -iBegin; x <= iBegin; x++)
	    {
	      xBis = x + iBegin;
	      maskElmt = &(privateMask[xBis][yBis][zBis]);

	      offset = x + (dimX * y) + (dimX * dimY * z);
	      maskElmt->offset = offset;
	      if (maskElmt->exist)
		{
		  neighborhood[m] = maskElmt->offset;
		  m++;
		}
	    }
	}
    }
  isSetOffsets = true;
}


