#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/ConnectingElement.h"

#include <inrimage/Voxel.h>
#include <inrimage/threshold.h>
#include <inrimage/ConnectingElement.h>


namespace yav {



  /*-------------------------------------------------------------------------*/
  CONNEXE_PARAMETER *connectedParameterAlloc() {
    CONNEXE_PARAMETER *par;

    par = new CONNEXE_PARAMETER;
    par->min_size = 1;
    par->max_nbcc = -1;
    par->output = COMPONENT_LABEL;
    return(par);
  }














  /* * structure describing a connected component
   */

  typedef struct {
    unsigned int label;
    int pointsAboveLowThreshold;
    int pointsAboveHighThreshold;
    char completelyProcessed;
    char representative;
  } typeConnectedComponent;


  /* * structure used for connected components processing
   */ 
  typedef struct {
    /* what is used for managing the equivalences
     * we have an array of cc_array_size components
     * indexed from 0 to cc_array_size-1
     */
    typeConnectedComponent *cc;
    int cc_array_size;

    /* what is used to select the valid connected components
     */
    int minNumberOfPtsAboveLow;
    int minNumberOfPtsAboveHigh;

    /* COMPONENT_BINARY or COMPONENT_LABEL
     */
    COMPONENT_OUTPUT output;
    int max_number_of_components;

    /* the color of a connected component when it is not valide,
       a fortiori, it should be zero
    */
    int invalid_label; 

  } typeConnectedComponentsManagement;





#include <_connexe.C>




  /*
    static int _GetMinMaxVoxels( Voxel *bg, Voxel *fg, Inrimage::WORD_TYPE type )
    {
    (bg)->setMin();
    (fg)->setMax();
    return( 1 );
  
    switch( type ) {   
    case Inrimage::WT_UNSIGNED_CHAR:
    dynamic_cast<Voxel8u *>(bg)->setMin();
    dynamic_cast<Voxel8u *>(fg)->setMax();
    break;
    case Inrimage::WT_SIGNED_CHAR:
    dynamic_cast<Voxel8s *>(bg)->setMin();
    dynamic_cast<Voxel8s *>(fg)->setMax();
    break;
    case Inrimage::WT_UNSIGNED_SHORT:
    dynamic_cast<Voxel16u *>(bg)->setMin();
    dynamic_cast<Voxel16u *>(fg)->setMax();
    break;
    case Inrimage::WT_SIGNED_SHORT:
    dynamic_cast<Voxel16s *>(bg)->setMin();
    dynamic_cast<Voxel16s *>(fg)->setMax();
    break;
    case Inrimage::WT_UNSIGNED_INT:
    dynamic_cast<Voxel32u *>(bg)->setMin();
    dynamic_cast<Voxel32u *>(fg)->setMax();
    break;
    case Inrimage::WT_SIGNED_INT:
    dynamic_cast<Voxel32s *>(bg)->setMin();
    dynamic_cast<Voxel32s *>(fg)->setMax();
    break;
    case Inrimage::WT_UNSIGNED_LONG:
    dynamic_cast<Voxel64u *>(bg)->setMin();
    dynamic_cast<Voxel64u *>(fg)->setMax();
    break;
    case Inrimage::WT_SIGNED_LONG:
    dynamic_cast<Voxel64s *>(bg)->setMin();
    dynamic_cast<Voxel64s *>(fg)->setMax();
    break;
    case Inrimage::WT_FLOAT:
    dynamic_cast<VoxelFloat *>(bg)->setMin();
    dynamic_cast<VoxelFloat *>(fg)->setMax();
    break;
    case Inrimage::WT_DOUBLE:
    dynamic_cast<VoxelDouble *>(bg)->setMin();
    dynamic_cast<VoxelDouble *>(fg)->setMax();
    break;
    default :
    return -1;;
    }
    return 1;
    }
  */




  /** where the computation take place ...
   */

  static int _ObjectConnectedComponentsExtraction( Inrimage *inr,
						   ConnectingElement *CE,
						   int background_label,
						   int v1, 
						   int v2,
						   int binary_label ) 
{
  int connexite;
  typeConnectedComponentsManagement env;
  int r;
  
  switch( CE->getConnectivity() ) {
  case FOUR_CONNECTIVITY :       connexite =  4; break;
  case EIGHT_CONNECTIVITY :      connexite =  8; break;
  case SIX_CONNECTIVITY :        connexite =  6; break;
  case TEN_CONNECTIVITY :        connexite = 10; break;
  case EIGHTEEN_CONNECTIVITY :   connexite = 18; break;
  default :
  case TWENTY_SIX_CONNECTIVITY : connexite = 26; break;      
  }


  env.cc_array_size = CE->getCCArraySize();
  env.cc = (typeConnectedComponent *)malloc( env.cc_array_size * sizeof( typeConnectedComponent ) );
  if ( env.cc == NULL ) return -1;

  /* filling the parameters
     structure 
  */
  env.minNumberOfPtsAboveLow   = CE->getMinimalSize();
  env.minNumberOfPtsAboveHigh  = CE->getMinimalSizeHighThreshold();
  env.output                   = CE->getOutputType();
  env.max_number_of_components = CE->getMaximalNumberOfComponents();
  env.invalid_label            = background_label;


  /* extracting the connected components
     of the foreground
  */
  switch( inr->getType() ) {
  default :
    free( env.cc );
    return -1;
  case Inrimage::WT_UNSIGNED_INT :
    r =  _ForegroundConnectedComponentsExtraction( (unsigned int*)inr->getData(),
						   inr->getX(),
						   inr->getY(),
						   inr->getZ(),
						   (unsigned int)v1,
						   (unsigned int)v2,
						   connexite, &env );
    break;
  case Inrimage::WT_UNSIGNED_SHORT :
    r =  _ForegroundConnectedComponentsExtraction( (unsigned short int*)inr->getData(),
						   inr->getX(),
						   inr->getY(),
						   inr->getZ(),
						   (unsigned short int)v1,
						   (unsigned short int)v2,
						   connexite, &env );
    break;
  case Inrimage::WT_UNSIGNED_CHAR :
    r =  _ForegroundConnectedComponentsExtraction( (unsigned char*)inr->getData(),
						   inr->getX(),
						   inr->getY(),
						   inr->getZ(),
						   (unsigned char)v1,
						   (unsigned char)v2,
						   connexite, &env );
    break;
  }
  

  /* easy things: error or zero components
   */
  switch( r ) {
  case -1 :
    free( env.cc );
    return -1;
  case 0 :
    inr->zeroRegion( 0, 0, 0, inr->getX()-1, inr->getY()-1, inr->getZ()-1 );
    free( env.cc );
    return 0;
  }
  
  /* how many valid connected components
     they will be numbered from 1 to the number of components
     in case of binary output, the output image is binary (either 'background_label' or 1)
  */
  r = _ConnectedComponentsSelection( &env, 1, r, binary_label );

  switch( r ) {
  case -1 :
    free( env.cc );
    return -1;
  case 0 :
    inr->zeroRegion( 0, 0, 0, inr->getX()-1, inr->getY()-1, inr->getZ()-1 );
    free( env.cc );
    return 0;
  }
 
  /* relabeling the connected components
   */
  switch( inr->getType() ) {
  default :
    free( env.cc );
    return -1;
  case Inrimage::WT_UNSIGNED_INT :
    _ConnectedComponentsRelabeling( (unsigned int*)inr->getData(),
				    inr->getX(),
				    inr->getY(),
				    inr->getZ(),
				    &env );
    break;
  case Inrimage::WT_UNSIGNED_SHORT :
    _ConnectedComponentsRelabeling( (unsigned short int*)inr->getData(),
				    inr->getX(),
				    inr->getY(),
				    inr->getZ(),
				    &env );
    break;
  case Inrimage::WT_UNSIGNED_CHAR :
    _ConnectedComponentsRelabeling( (unsigned char*)inr->getData(),
				    inr->getX(),
				    inr->getY(),
				    inr->getZ(),
				    &env );
    break;
  }

  return( r );
}











  /* numerote les composantes connexes du fond et de l'objet

     les composantes du fond seront numerotees de 1 a back_labels,
     celles de l'objet de back_labels+1 a fore_labels
   
     Il n'y a pas de selection des composantes
  */

  static int _AllConnectedComponentsExtraction( Inrimage *inr,
						ConnectingElement *CE,
						int background_label,
						int *back_labels,
						int *fore_labels,
						unsigned int z_first,
						unsigned int z_last ) 
{
  int fg_connexite, bg_connexite;
  typeConnectedComponentsManagement env;
  int bg_first = -1;
  int bg_last = -1;
  int fg_first = -1;
  int fg_last = -1;
  int r;

  *back_labels = *fore_labels = 0;

  switch( CE->getConnectivity() ) {
  case FOUR_CONNECTIVITY :       fg_connexite =  4; bg_connexite =  8; break;
  case EIGHT_CONNECTIVITY :      fg_connexite =  8; bg_connexite =  4; break;
  case SIX_CONNECTIVITY :        fg_connexite =  6; bg_connexite = 26; break;
  case TEN_CONNECTIVITY :        fg_connexite = 10; bg_connexite =  6; break;
  case EIGHTEEN_CONNECTIVITY :   fg_connexite = 18; bg_connexite =  6; break;
  default :
  case TWENTY_SIX_CONNECTIVITY : fg_connexite = 26; bg_connexite =  6; break;      
  }
  

  env.cc_array_size = CE->getCCArraySize();
  env.cc = (typeConnectedComponent *)malloc( env.cc_array_size * sizeof( typeConnectedComponent ) );
  if ( env.cc == NULL ) return -1;

  /* filling the parameters
     structure 
  */
  env.minNumberOfPtsAboveLow   = CE->getMinimalSize();
  env.minNumberOfPtsAboveHigh  = CE->getMinimalSizeHighThreshold();
  env.output                   = CE->getOutputType();
  env.max_number_of_components = CE->getMaximalNumberOfComponents();
  env.invalid_label            = background_label;


  /* extracting the connected components
     of the foreground
  */
  switch( inr->getType() ) {
  default :
    free( env.cc );
    return -1;
  case Inrimage::WT_UNSIGNED_INT :
    r =  _ForeBackConnectedComponentsExtraction( ((unsigned int*)inr->getData())+z_first*inr->getX()*inr->getY(),
						 inr->getX(),
						 inr->getY(),
						 z_last - z_first + 1,
						 (unsigned int)background_label,
						 bg_connexite, fg_connexite, &env,
						 &bg_first, &bg_last, &fg_first, &fg_last );
    break;
  case Inrimage::WT_UNSIGNED_SHORT :
    r =  _ForeBackConnectedComponentsExtraction( ((unsigned short int*)inr->getData())+z_first*inr->getX()*inr->getY(),
						 inr->getX(),
						 inr->getY(),
						 z_last - z_first + 1,
						 (unsigned short int)background_label,
						 bg_connexite, fg_connexite, &env,
						 &bg_first, &bg_last, &fg_first, &fg_last );
    break;
  case Inrimage::WT_UNSIGNED_CHAR :
    r =  _ForeBackConnectedComponentsExtraction( ((unsigned char*)inr->getData())+z_first*inr->getX()*inr->getY(),
						 inr->getX(),
						 inr->getY(),
						 z_last - z_first + 1,
						 (unsigned char)background_label,
						 bg_connexite, fg_connexite, &env,
						 &bg_first, &bg_last, &fg_first, &fg_last );
    break;
  }

  /* easy things: error or zero components
   */
  switch( r ) {
  case -1 :
    free( env.cc );
    return -1;
  }

  /* 'bg_first > bg_last' means that the processed image was completely foreground
     'fg_first > fg_last' means that the processed image was completely background
     whatever, no error should be returned (empty slices do occur in the 2D processing
     of a 3D image)
  */

  /* how many valid connected components
     they will be numbered from 1 to the number of components
     in case of binary output, the output image is binary (either 'background_label' or 1)
  */
  r = _ConnectedComponentsSelection( &env, &bg_first, &bg_last, &fg_first, &fg_last );

  /* relabeling the connected components
   */
  switch( inr->getType() ) {
  default :
    free( env.cc );
    return -1;
  case Inrimage::WT_UNSIGNED_INT :
    _ConnectedComponentsRelabeling( ((unsigned int*)inr->getData())+z_first*inr->getX()*inr->getY(),
				    inr->getX(),
				    inr->getY(),
				    z_last - z_first + 1,
				    &env );
    break;
  case Inrimage::WT_UNSIGNED_SHORT :
    _ConnectedComponentsRelabeling( ((unsigned short int*)inr->getData())+z_first*inr->getX()*inr->getY(),
				    inr->getX(),
				    inr->getY(),
				    z_last - z_first + 1,
				    &env );
    break;
  case Inrimage::WT_UNSIGNED_CHAR :
    _ConnectedComponentsRelabeling( ((unsigned char*)inr->getData())+z_first*inr->getX()*inr->getY(),
				    inr->getX(),
				    inr->getY(),
				    z_last - z_first + 1,
				    &env );
    break;
  }

  *back_labels = bg_last - bg_first + 1;
  *fore_labels = fg_last - fg_first + 1;

  return( r );
}













  /*-------------------------------------------------------------------------*/
  /*-------------------------------------------------------------------------*/
  /*-------------------------------------------------------------------------*/

  /** Hysteresis thresholding
   */

void hysteresis( Inrimage *inr, 
		 double min,  double max,
		 ConnectingElement *param_CE,
		 Voxel *param_bg, Voxel *param_fg )
  throw( Exception )
{
  hysteresis( inr, inr->getType(), min, max, param_CE,
	      param_bg, param_fg );
}

void hysteresis( Inrimage *inr, 
		 Inrimage::WORD_TYPE newtype,
		 double min,  double max,
		 ConnectingElement *param_CE,
		 Voxel *param_bg, Voxel *param_fg )
  throw( Exception )
{
  Voxel *vmin = Voxel::voxelByType( inr->getType() );
  Voxel *vmax = Voxel::voxelByType( inr->getType() );
  vmin->setValueFromDouble( min );
  vmax->setValueFromDouble( max );
  hysteresis( inr, newtype, vmin, vmax, param_CE,
	      param_bg, param_fg );
  delete vmin;
  delete vmax;
}

void hysteresis( Inrimage *inr, 
		 Voxel *min,  Voxel *max,
		 ConnectingElement *param_CE,
		 Voxel *param_bg, Voxel *param_fg )
  throw( Exception )
{
  hysteresis( inr, inr->getType(), min, max, param_CE,
	      param_bg, param_fg );
}

void hysteresis( Inrimage *inr, 
		 Inrimage::WORD_TYPE newtype,
		 Voxel *min,  Voxel *max,
		 ConnectingElement *param_CE,
		 Voxel *param_bg, Voxel *param_fg )
  throw( Exception )
{
  char *from = "hysteresis(Inrimage *inr, Inrimage::WORD_TYPE, Voxel *, Voxel *,\n ConnectingElement *, Voxel *, Voxel *)";
  Voxel *vox_binary;
  Voxel *vox_back;
  Voxel *vox_low;
  Voxel *vox_high;
  int background_label = 0;
  int lowthres_label = 100;
  int highthres_label = 200;
  ConnectingElement *CE;
  Voxel *bg;
  Voxel *fg;
  

  if ( param_bg == NULL ) {
    bg = Voxel::voxelByType( newtype );
  }
  else {
    bg = param_bg;
  }
  if ( param_fg == NULL ) {
    fg = Voxel::voxelByType( newtype );
    fg->setMax();
  }
  else {
    fg = param_bg;
  }
  if ( param_CE == NULL ) {
    CE = new ConnectingElement();
  }
  else {
    CE = param_CE;
  }

  /* what is needed for preprocessing
   */
  switch( CE->getWorkType() ) {
  default :
  case Inrimage::WT_UNSIGNED_SHORT :
    if ( 1.0 <= fg->getValueAsDouble() && fg->getValueAsDouble() <= 65535.0 )
      vox_binary = new Voxel16u( (unsigned short int)fg->getValueAsDouble() );
    else {
      vox_binary = new Voxel16u( MAX_UNSIGNED_SHORT );
    }
    vox_back = new Voxel16u( (unsigned short int)background_label );
    vox_low  = new Voxel16u( (unsigned short int)lowthres_label );
    vox_high = new Voxel16u( (unsigned short int)highthres_label );
    break;
  case Inrimage::WT_UNSIGNED_INT :
    if ( 1.0 <= fg->getValueAsDouble() && fg->getValueAsDouble() <= (double)MAX_SIGNED_INT )
      vox_binary = new Voxel32u( (unsigned int)fg->getValueAsDouble() );
    else
      vox_binary = new Voxel32u( MAX_SIGNED_INT );
    vox_back = new Voxel32u( (unsigned int)background_label );
    vox_low  = new Voxel32u( (unsigned int)lowthres_label );
    vox_high = new Voxel32u( (unsigned int)highthres_label );
    break;
  case Inrimage::WT_UNSIGNED_CHAR :
    if ( 1.0 <= fg->getValueAsDouble() && fg->getValueAsDouble() <= 255.0 )
      vox_binary = new Voxel8u( (unsigned char)fg->getValueAsDouble() );
    else
      vox_binary = new Voxel8u( MAX_UNSIGNED_CHAR );
    vox_back = new Voxel8u( (unsigned char)background_label );
    vox_low  = new Voxel8u( (unsigned char)lowthres_label );
    vox_high = new Voxel8u( (unsigned char)highthres_label );
    break;
  }
  
  
  try {
    /* pre-processed the input image
     */
    if ( CE->getNumberSeedPoints() >= 1 ) {
      threshold( inr, CE->getWorkType(), min, NULL, vox_back, vox_low );
      CE->markSeedPoints( inr, vox_high );
    }
    else {
      ternarize(inr, CE->getWorkType(), min, max, vox_back, vox_low, vox_high );
    }
    if ( CE->getMinimalSizeHighThreshold() <= 0 )
      CE->setMinimalSizeHighThreshold( 1 );
    if ( _ObjectConnectedComponentsExtraction( inr, CE,
					       background_label,
					       lowthres_label,
					       highthres_label,
					       (int)vox_binary->getValueAsLong() ) < 0 ) {
      throw Exception( from, "Error while extracting connected components" );
    }

    if ( CE->getOutputType() == COMPONENT_LABEL ) {
      inr->convertCast( newtype );
      return;
    }
    
    if ( CE->getWorkType() == newtype 
	 && CE->getOutputType() == COMPONENT_BINARY ) {
      if ( bg->getValueAsLong() == background_label 
	   && fg->getValueAsLong() == vox_binary->getValueAsLong() )
	return;
    }
    binarize( inr, newtype, vox_back, bg, fg );
  }
  catch( Exception e)  {
    e.push( from );
    throw e;
  }
  
}





void getHysteresisThresholding(Inrimage *inr, 
			       Voxel *min,
			       Voxel *max,
			       Voxel *bg,
			       Voxel *fg,
			       CONNECTIVITY_TYPE connectivity,
			       CONNEXE_PARAMETER *par )
{
  ConnectingElement *CE = new ConnectingElement(connectivity);
  if ( par != NULL ) {
    CE->setMinimalSize( (int)par->min_size );
    CE->setMaximalNumberOfComponents( par->max_nbcc );
    CE->setOutputType( par->output );
  }
  fprintf( stderr, "WARNING: getHysteresisThresholding() is deprecated\n" );
  fprintf( stderr, "\t use hysteresis() instead\n" );
  hysteresis( inr, inr->getType(), min, max, CE, bg, fg );
}














/*-------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------*/

/** Connected components extraction and/or labeling
 */

void connectedComponents( Inrimage *inr, 
			  double min,
			  ConnectingElement *param_CE,
			  Voxel *param_bg, Voxel *param_fg )
  throw( Exception )
{
  connectedComponents( inr, inr->getType(), min, param_CE,
		       param_bg, param_fg );
}

void connectedComponents( Inrimage *inr, 
			  Inrimage::WORD_TYPE newtype,
			  double min,
			  ConnectingElement *param_CE,
			  Voxel *param_bg, Voxel *param_fg )
  throw( Exception )
{
  Voxel *vmin = Voxel::voxelByType( inr->getType() );
  vmin->setValueFromDouble( min );
  connectedComponents( inr, newtype, vmin, NULL, param_CE,
		       param_bg, param_fg );
  delete vmin;
}

void connectedComponents( Inrimage *inr, 
			  double min,  double max,
			  ConnectingElement *param_CE,
			  Voxel *param_bg, Voxel *param_fg )
  throw( Exception )
{
  connectedComponents( inr, inr->getType(), min, max, param_CE,
		       param_bg, param_fg );
}

void connectedComponents( Inrimage *inr, 
			  Inrimage::WORD_TYPE newtype,
			  double min,  double max,
			  ConnectingElement *param_CE,
			  Voxel *param_bg, Voxel *param_fg )
  throw( Exception )
{
  Voxel *vmin = Voxel::voxelByType( inr->getType() );
  Voxel *vmax = Voxel::voxelByType( inr->getType() );
  vmin->setValueFromDouble( min );
  vmax->setValueFromDouble( max );
  connectedComponents( inr, newtype, vmin, vmax, param_CE,
		       param_bg, param_fg );
  delete vmin;
  delete vmax;
}

void connectedComponents( Inrimage *inr, 
			  Voxel *low_threshold,  Voxel *high_threshold,
			  ConnectingElement *param_CE,
			  Voxel *param_bg, Voxel *param_fg )
  throw( Exception )
{
  connectedComponents( inr, inr->getType(), low_threshold, high_threshold, param_CE,
		       param_bg, param_fg );
}

void connectedComponents( Inrimage *inr, 
			  Inrimage::WORD_TYPE newtype,
			  Voxel *low_threshold,  Voxel *high_threshold,
			  ConnectingElement *param_CE,
			  Voxel *param_bg, Voxel *param_fg )
  throw( Exception )
{
  char *from = "connectedComponents(Inrimage *inr, Inrimage::WORD_TYPE, Voxel *, Voxel *,\n ConnectingElement *, Voxel *, Voxel *)";
  Voxel *vox_binary;
  Voxel *vox_back;
  Voxel *vox_low;
  Voxel *vox_high;
  int background_label = 0;
  int lowthres_label = 100;
  int highthres_label = 200;
  ConnectingElement *CE;
  Voxel *bg;
  Voxel *fg;
  

  if ( param_bg == NULL ) {
    bg = Voxel::voxelByType( newtype );
  }
  else {
    bg = param_bg;
  }
  if ( param_fg == NULL ) {
    fg = Voxel::voxelByType( newtype );
    fg->setMax();
  }
  else {
    fg = param_bg;
  }
  if ( param_CE == NULL ) {
    CE = new ConnectingElement();
  }
  else {
    CE = param_CE;
  }

  /* what is needed for preprocessing
   */
  switch( CE->getWorkType() ) {
  default :
  case Inrimage::WT_UNSIGNED_SHORT :
    if ( 1.0 <= fg->getValueAsDouble() && fg->getValueAsDouble() <= 65535.0 )
      vox_binary = new Voxel16u( (unsigned short int)fg->getValueAsDouble() );
    else {
      vox_binary = new Voxel16u( MAX_UNSIGNED_SHORT );
    }
    vox_back = new Voxel16u( (unsigned short int)background_label );
    vox_low  = new Voxel16u( (unsigned short int)lowthres_label );
    vox_high = new Voxel16u( (unsigned short int)highthres_label );
    break;
  case Inrimage::WT_UNSIGNED_INT :
    if ( 1.0 <= fg->getValueAsDouble() && fg->getValueAsDouble() <= (double)MAX_SIGNED_INT )
      vox_binary = new Voxel32u( (unsigned int)fg->getValueAsDouble() );
    else
      vox_binary = new Voxel32u( MAX_SIGNED_INT );
    vox_back = new Voxel32u( (unsigned int)background_label );
    vox_low  = new Voxel32u( (unsigned int)lowthres_label );
    vox_high = new Voxel32u( (unsigned int)highthres_label );
    break;
  case Inrimage::WT_UNSIGNED_CHAR :
    if ( 1.0 <= fg->getValueAsDouble() && fg->getValueAsDouble() <= 255.0 )
      vox_binary = new Voxel8u( (unsigned char)fg->getValueAsDouble() );
    else
      vox_binary = new Voxel8u( MAX_UNSIGNED_CHAR );
    vox_back = new Voxel8u( (unsigned char)background_label );
    vox_low  = new Voxel8u( (unsigned char)lowthres_label );
    vox_high = new Voxel8u( (unsigned char)highthres_label );
    break;
  }
  
  
  try {
    /* pre-processed the input image
     */
    if ( CE->getNumberSeedPoints() >= 1 ) {
      threshold( inr, CE->getWorkType(), low_threshold, high_threshold, vox_back, vox_low );
      CE->markSeedPoints( inr, vox_high );
      CE->setMinimalSizeHighThreshold( 1 );
    }
    else {
      threshold( inr, CE->getWorkType(), low_threshold, high_threshold, vox_back, vox_high );
    }

    if ( _ObjectConnectedComponentsExtraction( inr, CE,
					       background_label,
					       lowthres_label,
					       highthres_label,
					       (int)vox_binary->getValueAsLong() ) < 0 ) {
      throw Exception( from, "Error while extracting connected components" );
    }

    if ( CE->getOutputType() == COMPONENT_LABEL ) {
      inr->convertCast( newtype );
      return;
    }
    if ( CE->getWorkType() == newtype 
	 && CE->getOutputType() == COMPONENT_BINARY ) {
      if ( bg->getValueAsLong() == background_label 
	   && fg->getValueAsLong() == vox_binary->getValueAsLong() )
	return;
    }
    binarize( inr, newtype, vox_back, bg, fg );
  }
  catch( Exception e)  {
    e.push( from );
    throw e;
  }

}


void getConnectedComponentBinary(Inrimage *inr, Voxel *bg,
				 Voxel *fg,
				 CONNECTIVITY_TYPE connectivity,
				 CONNEXE_PARAMETER *par) 
{
  ConnectingElement *CE = new ConnectingElement(connectivity);

  Voxel *v1 = Voxel::voxelByType( inr->getType() );
  Voxel *v2 = Voxel::voxelByType( inr->getType() );
  v1->setValueFromDouble( 0.0 );
  v2->setValueFromDouble( 1.0 );
  if ( par != NULL ) {
    CE->setMinimalSize( (int)par->min_size );
    CE->setMaximalNumberOfComponents( par->max_nbcc );
    CE->setOutputType( par->output );
  }
  fprintf( stderr, "WARNING: getConnectedComponentBinary() is deprecated\n" );
  fprintf( stderr, "\t use connectedComponents() instead\n" );
  binarize( inr, bg, v1, v2 );
  connectedComponents(inr, inr->getType(), v2, NULL, CE, bg, fg);

  delete v1;
  delete v2;
}


void getConnectedComponentThreshold(Inrimage *inr, 
				    Voxel *min,
				    Voxel *max,
				    Voxel *bg,
				    Voxel *fg,
				    CONNECTIVITY_TYPE connectivity,
				    CONNEXE_PARAMETER *par) 
{
  ConnectingElement *CE = new ConnectingElement(connectivity);
  if ( par != NULL ) {
    CE->setMinimalSize( (int)par->min_size );
    CE->setMaximalNumberOfComponents( par->max_nbcc );
    CE->setOutputType( par->output );
  }
  fprintf( stderr, "WARNING: getConnectedComponentThreshold() is deprecated\n" );
  fprintf( stderr, "\t use connectedComponents() instead\n" );
  connectedComponents(inr, inr->getType(), min, max, CE, bg, fg);
}

















/*-------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------*/



void fillContours( Inrimage *inr, 
		   ConnectingElement *CE, 
		   Voxel *bg, Voxel *fg, 
		   bool threeD )
  throw( Exception )
{
  fillContours( inr, inr->getType(), CE, bg, fg, threeD );
}

void fillContours( Inrimage *inr, Inrimage::WORD_TYPE newtype,
		   ConnectingElement *param_CE, 
		   Voxel *param_bg, Voxel *param_fg, 
		   bool threeD )
  throw( Exception )
{
  char *from = "fillContours( Inrimage *, Inrimage::WORD_TYPE, Voxel *, Voxel *, ConnectingElement *, bool )";
  Voxel *vox_input_fg;
  Voxel *vox_input_bg;
  Voxel *vox_work_fg;
  Voxel *vox_work_bg;
  Voxel *vox_output_fg;
  Voxel *vox_output_bg;

  //   Voxel *vox_binary;
  //   Voxel *vox_back;
  unsigned int background_label = 0;
  //   Voxel *bg;
  //   Voxel *fg;
  ConnectingElement *CE;
  int nb_bg_labels=0, nb_fg_labels=0;
  unsigned int z;

  /* param_bg : voxel designant le fond  dans l'image de depart,
     de type inr->getType()
     vox_input_bg <- on le recupere s'il existe 
     si l'image d'arrivee est de meme  type
     vox_output_bg <- on recupere vox_input_bg
  */
  if ( param_bg == NULL ) {
    vox_input_bg  = Voxel::voxelByType( inr->getType() );
  }
  else {
    vox_input_bg = param_bg;
  }
  if ( inr->getType() == newtype ) {
    vox_output_bg = vox_input_bg;
  } else {
    vox_output_bg = Voxel::voxelByType( newtype );
  }

  if ( param_fg == NULL ) {
    vox_input_fg = Voxel::voxelByType( inr->getType() );
    vox_input_fg->setMax();
  } else {
    vox_input_fg = param_fg;
  }
  if ( inr->getType() == newtype ) {
    vox_output_fg = vox_input_fg;
  } else {
    vox_output_fg = Voxel::voxelByType( newtype );
    vox_output_fg->setMax();
  }

 
  /* connecting element
   */
  if ( param_CE == NULL ) {
    CE = new ConnectingElement();
  }
  else {
    CE = param_CE;
  }

  
  
  /* what is needed for preprocessing
   */
  switch( CE->getWorkType() ) {
  default :
  case Inrimage::WT_UNSIGNED_SHORT :
    if ( 1.0 <= vox_input_fg->getValueAsDouble() 
	 && vox_input_fg->getValueAsDouble() <= 65535.0 )
      vox_work_fg = new Voxel16u( (unsigned short int)vox_input_fg->getValueAsDouble() );
    else {
      vox_work_fg = new Voxel16u( MAX_UNSIGNED_SHORT );
    }
    vox_work_bg = new Voxel16u( (unsigned short int)background_label );
    break;

  case Inrimage::WT_UNSIGNED_INT :
    if ( 1.0 <= vox_input_fg->getValueAsDouble() 
	 && vox_input_fg->getValueAsDouble() <= (double)MAX_SIGNED_INT )
      vox_work_fg = new Voxel32u( (unsigned int)vox_input_fg->getValueAsDouble() );
    else
      vox_work_fg = new Voxel32u( MAX_SIGNED_INT );
    vox_work_bg = new Voxel32u( (unsigned int)background_label );
    break;

  case Inrimage::WT_UNSIGNED_CHAR :
    if ( 1.0 <= vox_input_fg->getValueAsDouble() 
	 && vox_input_fg->getValueAsDouble() <= 255.0 )
      vox_work_fg = new Voxel8u( (unsigned char)vox_input_fg->getValueAsDouble() );
    else
      vox_work_fg = new Voxel8u( MAX_UNSIGNED_CHAR );
    vox_work_bg = new Voxel8u( (unsigned char)background_label );
    break;
  }

  CE->setOutputType( COMPONENT_LABEL );

  if ( threeD == false ) {
    switch( CE->getConnectivity() ) {
    default :
      break;
    case SIX_CONNECTIVITY :
      CE->setConnectivity( FOUR_CONNECTIVITY );
      break;
    case TEN_CONNECTIVITY :      
    case EIGHTEEN_CONNECTIVITY : 
    case TWENTY_SIX_CONNECTIVITY :
      CE->setConnectivity( EIGHT_CONNECTIVITY );
      break;
    }
  }

  binarize( inr, CE->getWorkType(), vox_input_bg, vox_work_bg, vox_work_fg );


  if ( threeD == false ) {

    for ( z = 0; z < inr->getZ(); z ++ ) {
      if ( _AllConnectedComponentsExtraction( inr, CE,
					      background_label,
					      &nb_bg_labels,  &nb_fg_labels,
					      z, z ) < 0 ) {
	throw Exception( from, "Error while extracting connected components" );
      }
  
      /* here we have background connected components from 1 to nb_bg_labels
	 and foreground connected components from nb_bg_labels+1 to nb_bg_labels+nb_fg_labels
      */

      switch( CE->getWorkType() ) {
      default :
      case Inrimage::WT_UNSIGNED_SHORT :
	_reallyFillContours( ((unsigned short int*)inr->getData())+z*inr->getX()*inr->getY(), 
			     inr->getX(), inr->getY(), 1,
			     (unsigned int)1, (unsigned int)nb_bg_labels, 
			     (unsigned int)(nb_bg_labels+1), 
			     (unsigned int)(nb_bg_labels+nb_fg_labels), 
			     background_label, ((Voxel16u*)vox_work_fg)->getValue(), threeD ); 
	break;
      case Inrimage::WT_UNSIGNED_INT :
	_reallyFillContours( ((unsigned int*)inr->getData())+z*inr->getX()*inr->getY(), 
			     inr->getX(), inr->getY(), 1,
			     (unsigned int)1, (unsigned int)nb_bg_labels, 
			     (unsigned int)(nb_bg_labels+1), 
			     (unsigned int)(nb_bg_labels+nb_fg_labels), 
			     background_label, ((Voxel32u*)vox_work_fg)->getValue(), threeD ); 
	break;
      case Inrimage::WT_UNSIGNED_CHAR :
	_reallyFillContours( ((unsigned char*)inr->getData())+z*inr->getX()*inr->getY(), 
			     inr->getX(), inr->getY(), 1,
			     (unsigned int)1, (unsigned int)nb_bg_labels, 
			     (unsigned int)(nb_bg_labels+1), 
			     (unsigned int)(nb_bg_labels+nb_fg_labels), 
			     background_label, ((Voxel8u*)vox_work_fg)->getValue(), threeD ); 
	break;
      }
    }

  }
  else {

    if ( _AllConnectedComponentsExtraction( inr, CE,
					    background_label,
					    &nb_bg_labels,  &nb_fg_labels,
					    0, inr->getZ()-1 ) < 0 ) {
      throw Exception( from, "Error while extracting connected components" );
    }
    switch( CE->getWorkType() ) {
    default :
    case Inrimage::WT_UNSIGNED_SHORT :
      _reallyFillContours( (unsigned short int*)inr->getData(), 
			   inr->getX(), inr->getY(), inr->getZ(),
			   (unsigned int)1, (unsigned int)nb_bg_labels, 
			   (unsigned int)(nb_bg_labels+1), 
			   (unsigned int)(nb_bg_labels+nb_fg_labels), 
 			   background_label, ((Voxel16u*)vox_work_fg)->getValue(), threeD ); 
      break;
    case Inrimage::WT_UNSIGNED_INT :
      _reallyFillContours( (unsigned int*)inr->getData(), 
			   inr->getX(), inr->getY(), inr->getZ(),
			   (unsigned int)1, (unsigned int)nb_bg_labels, 
			   (unsigned int)(nb_bg_labels+1), 
			   (unsigned int)(nb_bg_labels+nb_fg_labels), 
			   background_label, ((Voxel32u*)vox_work_fg)->getValue(), threeD ); 
      break;
    case Inrimage::WT_UNSIGNED_CHAR :
      _reallyFillContours( (unsigned char*)inr->getData(), 
			   inr->getX(), inr->getY(), inr->getZ(),
			   (unsigned int)1, (unsigned int)nb_bg_labels, 
			   (unsigned int)(nb_bg_labels+1), 
			   (unsigned int)(nb_bg_labels+nb_fg_labels), 
			   background_label, ((Voxel8u*)vox_work_fg)->getValue(), threeD ); 
      break;
    }
    
  }

  if ( newtype == CE->getWorkType() ) return;
  binarize( inr, newtype, vox_work_bg, vox_output_bg, vox_output_fg );
}












/*-------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------*/


void reconstruct( Inrimage *inr, 
		  ConnectingElement *CE, 
		  Inrimage *seeds,
		  Voxel *bg, Voxel *fg )
  throw( Exception )
{
  reconstruct( inr, inr->getType(), CE, seeds, bg, fg );
}

void reconstruct( Inrimage *inr, Inrimage::WORD_TYPE newtype,
		  ConnectingElement *param_CE, 
		  Inrimage *seeds,
		  Voxel *param_bg, Voxel *param_fg )
  throw( Exception )
{
  char *from = "reconstruct(Inrimage *inr, Inrimage::WORD_TYPE, ConnectingElement *, Inrimage *, Voxel *, Voxel *)";
  Voxel *vox_binary;
  Voxel *vox_back;
  Voxel *vox_low;
  Voxel *vox_high;
  int background_label = 0;
  int lowthres_label = 100;
  int highthres_label = 200;
  Voxel *bg;
  Voxel *fg;
  ConnectingElement *CE;


  /* no seeds
   */
  if ( (param_CE == NULL || param_CE->getNumberSeedPoints() <= 0 )
       && seeds == NULL ) {
    inr->zeroRegion( 0, 0, 0, inr->getX()-1, inr->getY()-1, inr->getZ()-1 );
  }
  if ( seeds != NULL ) {
    if (inr->sameSize( *seeds ) != true ||
	inr->getVdim() != seeds->getVdim() ||
	// inr->getType() != seeds->getType() ||
	inr->getVectorialMode() != seeds->getVectorialMode() ||
	inr->getVdim() > 1 ||
	inr->getVectorialMode() != VM_SCALAR ) {
      throw Exception( from, "Input images header mismatch" );
    }  
  }

  /* check parameters
   */
  if ( param_bg == NULL ) {
    bg = Voxel::voxelByType( inr->getType() );
  }
  else {
    bg = param_bg;
  }
  if ( param_fg == NULL ) {
    fg = Voxel::voxelByType( inr->getType() );
    fg->setMax();
  }
  else {
    fg = param_bg;
  }
  if ( param_CE == NULL ) {
    CE = new ConnectingElement();
  }
  else {
    CE = param_CE;
  }

  

  /* what is needed for preprocessing
   */
  switch( CE->getWorkType() ) {
  default :
  case Inrimage::WT_UNSIGNED_SHORT :
    if ( 1.0 <= fg->getValueAsDouble() && fg->getValueAsDouble() <= 65535.0 )
      vox_binary = new Voxel16u( (unsigned short int)fg->getValueAsDouble() );
    else {
      vox_binary = new Voxel16u( MAX_UNSIGNED_SHORT );
    }
    vox_back = new Voxel16u( (unsigned short int)background_label );
    vox_low  = new Voxel16u( (unsigned short int)lowthres_label );
    vox_high = new Voxel16u( (unsigned short int)highthres_label );
    break;
  case Inrimage::WT_UNSIGNED_INT :
    if ( 1.0 <= fg->getValueAsDouble() && fg->getValueAsDouble() <= (double)MAX_SIGNED_INT )
      vox_binary = new Voxel32u( (unsigned int)fg->getValueAsDouble() );
    else
      vox_binary = new Voxel32u( MAX_SIGNED_INT );
    vox_back = new Voxel32u( (unsigned int)background_label );
    vox_low  = new Voxel32u( (unsigned int)lowthres_label );
    vox_high = new Voxel32u( (unsigned int)highthres_label );
    break;
  case Inrimage::WT_UNSIGNED_CHAR :
    if ( 1.0 <= fg->getValueAsDouble() && fg->getValueAsDouble() <= 255.0 )
      vox_binary = new Voxel8u( (unsigned char)fg->getValueAsDouble() );
    else
      vox_binary = new Voxel8u( MAX_UNSIGNED_CHAR );
    vox_back = new Voxel8u( (unsigned char)background_label );
    vox_low  = new Voxel8u( (unsigned char)lowthres_label );
    vox_high = new Voxel8u( (unsigned char)highthres_label );
    break;
  }



  try {

    binarize( inr, CE->getWorkType(), bg, vox_back, vox_low );
    if ( CE->getNumberSeedPoints() >= 1 ) {
      CE->markSeedPoints( inr, vox_high );
    }

    if ( seeds != NULL ) {
      
      Voxel *bg_seeds = Voxel::voxelByType( seeds->getType() );
      
      switch( CE->getWorkType() ) {
      default :
      case Inrimage::WT_UNSIGNED_SHORT :
	seedImageToBuffer( (unsigned short int *)inr->getData(),
			   (unsigned int)background_label, (unsigned int)highthres_label, 
			   seeds, bg_seeds );
	break;
      case Inrimage::WT_UNSIGNED_INT :
	seedImageToBuffer( (unsigned int *)inr->getData(),
			   (unsigned int)background_label, (unsigned int)highthres_label, 
			   seeds, bg_seeds );
	break;
      case Inrimage::WT_UNSIGNED_CHAR :
	seedImageToBuffer( (unsigned char *)inr->getData(),
			   (unsigned int)background_label, (unsigned int)highthres_label, 
			   seeds, bg_seeds );
	break;
      }
    }
    
    CE->setMinimalSizeHighThreshold( 1 );
   
    if ( _ObjectConnectedComponentsExtraction( inr, CE,
					       background_label,
					       lowthres_label,
					       highthres_label,
					       (int)vox_binary->getValueAsLong() ) < 0 ) {
      throw Exception( from, "Error while extracting connected components" );
    }
    
    if ( CE->getOutputType() == COMPONENT_LABEL ) {
      inr->convertCast( newtype );
      return;
    }

    if ( CE->getWorkType() == newtype 
	 && CE->getOutputType() == COMPONENT_BINARY ) {
      if ( bg->getValueAsLong() == background_label 
	   && fg->getValueAsLong() == vox_binary->getValueAsLong() )
	return;
    }
    
    binarize( inr, newtype, vox_back, bg, fg );
  }
  catch( Exception e)  {
    e.push( from );
    throw e;
  }
  
 
}








/*-------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------*/

//
// classe ConnectingElement
//



//
// initialisation des champs
//
void ConnectingElement::init( ) 
{
  connectivity = TWENTY_SIX_CONNECTIVITY;
  minimal_size = 1;
  minimal_size_high_threshold = 0;
  maximal_number_of_components = -1;
  output = COMPONENT_BINARY;
  work_type = Inrimage::WT_UNSIGNED_SHORT;
  cc_array_size = 1 + (int)MAX_UNSIGNED_SHORT;

  number_seed_points = 0;
  seeds = NULL;
}

void ConnectingElement::setConnectivity( int c ) 
{ 
  CONNECTIVITY_TYPE CNT = connectivity;
  switch ( c ) {
  case 4 :  CNT = FOUR_CONNECTIVITY; break;
  case 8 :  CNT = EIGHT_CONNECTIVITY; break;
  case 6 :  CNT = SIX_CONNECTIVITY; break;
  case 10 : CNT = TEN_CONNECTIVITY; break;
  case 18 : CNT = EIGHTEEN_CONNECTIVITY; break;
  case 26 : CNT = TWENTY_SIX_CONNECTIVITY; break;
  default :
    break;
  }
  if ( connectivity != CNT ) connectivity = CNT;
}

void ConnectingElement::setWorkType( Inrimage::WORD_TYPE T ) 
{ 
  switch( T ) {
  default : break;
  case Inrimage::WT_UNSIGNED_CHAR :
    work_type = T;
    cc_array_size = 1 + (int)MAX_UNSIGNED_CHAR;
    break;
  case Inrimage::WT_UNSIGNED_SHORT :
    work_type = T;
    cc_array_size = 1 + (int)MAX_UNSIGNED_SHORT;
    break;
  case Inrimage::WT_UNSIGNED_INT :
    work_type = T;
    cc_array_size = 2 * (int)MAX_UNSIGNED_SHORT;
    break;
  }
}

void ConnectingElement::addSeedPoint( int x, int y, int z )
{
  typeSeedPoint *newseeds;
  newseeds = new typeSeedPoint[ number_seed_points + 1 ];
  if ( number_seed_points > 0 ) {
    memcpy( newseeds, seeds, number_seed_points*sizeof(typeSeedPoint) );
    delete[] seeds;
  }
  newseeds[number_seed_points][0] = x;
  newseeds[number_seed_points][1] = y;
  newseeds[number_seed_points][2] = z;
  seeds = newseeds;
  number_seed_points ++;
}





template <class TYPE> void markSeedPointsInBuffer( TYPE *theBuf,
						   unsigned int dimx,
						   unsigned int dimy,
						   unsigned int dimz,
						   const TYPE val,
						   typeSeedPoint *seeds,
						   unsigned int nb_seeds )
{
  unsigned int dimxy = dimx*dimy;
  unsigned int o, z, n;
  if ( nb_seeds == 0 ) return;
  for ( n=0; n<nb_seeds; n++ ) {
    if ( seeds[n][2] >= (int)dimz ) continue;
    if ( seeds[n][0] < 0 || seeds[n][0] >= (int)dimx ) continue;
    if ( seeds[n][1] < 0 || seeds[n][1] >= (int)dimy ) continue;
    o = seeds[n][0] + seeds[n][1]*dimx;
    if ( seeds[n][2] >= 0 ) {
      if ( theBuf[ o + seeds[n][2]*dimxy ] > 0 )
	theBuf[ o + seeds[n][2]*dimxy ] = val;
      continue;
    }
    for ( z=0; z<dimz; z++ ) {
      if ( theBuf[ o + z*dimxy ] > 0 )
	theBuf[ o + z*dimxy ] = val;
    }
  }
}


void ConnectingElement::markSeedPoints( Inrimage *inr, Voxel *v )
  throw( ErrTypeMismatch )
{
  char *from = "ConnectingElement::markSeedPoints( Inrimage *, Voxel * )";
  switch( inr->getType() ) {
  default:
    throw ErrTypeMismatch( from );
    return;
  case Inrimage::WT_UNSIGNED_CHAR:
    markSeedPointsInBuffer( (unsigned char *)inr->getData(),
			    inr->getX(), inr->getY(), inr->getZ(), 
			    ((Voxel8u *)v)->getValue(), seeds, number_seed_points );
    break;
  case Inrimage::WT_UNSIGNED_SHORT:
    markSeedPointsInBuffer( (unsigned short int *)inr->getData(),
			    inr->getX(), inr->getY(), inr->getZ(), 
			    ((Voxel16u *)v)->getValue(), seeds, number_seed_points );
    break;
  case Inrimage::WT_UNSIGNED_INT:
    markSeedPointsInBuffer( (unsigned int *)inr->getData(),
			    inr->getX(), inr->getY(), inr->getZ(), 
			    ((Voxel32u *)v)->getValue(), seeds, number_seed_points );
    break;
  }
}

// Constructeur: avec une connexite
ConnectingElement::ConnectingElement( const CONNECTIVITY_TYPE C,
				      const unsigned int min_size_low_threshold ,
				      const unsigned int min_size_high_threshold,
				      const int max_number_of_components,
				      const COMPONENT_OUTPUT type_output,
				      const Inrimage::WORD_TYPE label_type)
{
  this->init();
  setConnectivity( C );
  setMinimalSize( min_size_low_threshold );
  setMinimalSizeHighThreshold( min_size_high_threshold );
  setMaximalNumberOfComponents( max_number_of_components );
  setOutputType( type_output );
}

ConnectingElement::ConnectingElement( const int c,
				      const unsigned int min_size_low_threshold,
				      const unsigned int min_size_high_threshold,
				      const int max_number_of_components,
				      const COMPONENT_OUTPUT type_output,
				      const Inrimage::WORD_TYPE label_type)
{
  this->init();
  setConnectivity( c );
  setMinimalSize( min_size_low_threshold );
  setMinimalSizeHighThreshold( min_size_high_threshold );
  setMaximalNumberOfComponents( max_number_of_components );
  setOutputType( type_output );
}

// Destructeur

ConnectingElement::~ConnectingElement() {
}


} // namespace yav
