#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/DisplacementField3D.h"

#include <inrimage/DisplacementField3D.h>
#include <inrimage/arithmetic.h>

namespace yav {

DisplacementField3D::~DisplacementField3D() {

  delete inrx;
  delete inry;
  delete inrz;

} 


DisplacementField3D::DisplacementField3D (FIELD_MODE m, Inrimage *imx, Inrimage *imy, Inrimage *imz, float _vx, float _vy, float _vz) 
  throw(DisplacementField3D::DisplacementField3DMismatchType, DisplacementField3D::DisplacementField3DMismatchSize, DisplacementField3D::DisplacementField3DMismatchVoxelSize) : GeneralFormat("3D DisplacementField File", "3D FIELD"), Owner<Inrimage>(0) {

  char from[] = "DisplacementField3D::DisplacementField3D (FIELD_MODE m, Inrimage *imx, Inrimage *imy, Inrimage *imz, float vx, float vy, float vz ) ";
  
  mode = m;
  
  if ( (imx->getX() != imy->getX()) || (imx->getX() != imz->getX()) || 
       (imz->getX() != imz->getX()) ||
       (imx->getY() != imy->getY()) || (imx->getY() != imz->getY()) || 
       (imz->getY() != imz->getY()) ||
       (imx->getZ() != imy->getZ()) || (imx->getZ() != imz->getZ()) || 
       (imz->getZ() != imz->getZ()) )
    throw (DisplacementField3DMismatchSize("DisplacementField3D::DisplacementField3D (FIELD_MODE m, Inrimage *imx, Inrimage *imy, Inrimage *imz, float vx, float vy, float vz ) "));
       

  if ( (imx->getType() == imy->getType()) && (imy->getType() == imz->getType()) &&
       (imx->getType() == imz->getType()) )
    type = imx->getType();
  else
    throw (DisplacementField3DMismatchType("DisplacementField3D::DisplacementField3D (FIELD_MODE m, Inrimage *imx, Inrimage *imy, Inrimage *imz, float vx, float vy, float vz ) "));
  
  if ( mode==VOXEL )
    if (!( (imx->getVX() == imy->getVX()) && (imx->getVX() == imz->getVX()) && (imy->getVX() == imz->getVX()) &&
	 (imx->getVY() == imy->getVY()) && (imx->getVY() == imz->getVY()) && (imy->getVY() == imz->getVY()) &&
	 (imx->getVZ() == imy->getVZ()) && (imx->getVZ() == imz->getVZ()) && (imy->getVZ() == imz->getVZ()) ))
      throw DisplacementField3DMismatchVoxelSize(from);

  vx = _vx;
  vy = _vy;
  vz = _vz;
    
  
  if ( (imx==(Inrimage *)NULL) || (imy==(Inrimage *)NULL) || (imz==(Inrimage *)NULL) )
    throw (InvalidDisplacementField3DImages("DisplacementField3D::DisplacementField3D (FIELD_MODE m, Inrimage *imx, Inrimage *imy, Inrimage *imz, float vx, float vy, float vz ) "));
  else {
    inrx = new ObjectPointer<Inrimage>(this, imx);
    inry = new ObjectPointer<Inrimage>(this, imy);
    inrz = new ObjectPointer<Inrimage>(this, imz);
  }

  initGeneralFormat();

}


DisplacementField3D::DisplacementField3D( const char *file)
  throw(GeneralFormat::FileNotFound, Inrimage::ImageNotFound, InvalidDisplacementField3DImagesNb,
	DisplacementField3DMismatchType, DisplacementField3DMismatchSize)
  : GeneralFormat("3D DisplacementField File", "3D FIELD"), Owner<Inrimage>(0) {

  char from[] = "DisplacementField3D::DisplacementField3D( const char *filename) throw(GeneralFormat::FileNotFound, Inrimage::ImageNotFound, InvalidDisplacementField3DImagesNb, DisplacementField3DMismatchType, DisplacementField3DMismatchSize)";


  initGeneralFormat();
  filename = file;

  try {
    read((char *)filename, 0);
  }
  /* catch(Inrimage4D::ReadError &e) {
     e.push(from);
     throw e;
    } */
  catch(Inrimage::ImageNotFound &e) {
    e.push(from);
    throw e;
  }
  catch(GeneralFormat::FileNotFound &e) {
    e.push(from);
    throw e;
  }
  /* catch(Exception &e) {
     throw ReadError(from, filename);
    }*/

}


float DisplacementField3D::getVX() const throw(DisplacementField3D::InvalidDisplacementField3DMode) {

  char from[] = " DisplacementField3D::getVX() ";
 
  if ( mode == VOXEL )
    return vx;
  else 
    throw InvalidDisplacementField3DMode(from);
  
}


float DisplacementField3D::getVY() const throw (DisplacementField3D::InvalidDisplacementField3DMode) {
  
  char from[] = " DisplacementField3D::getVY() ";
 
  if ( mode == VOXEL )
    return vy;
  else 
    throw InvalidDisplacementField3DMode(from);
  
}

float DisplacementField3D::getVZ() const throw (DisplacementField3D::InvalidDisplacementField3DMode) {
  
  char from[] = " DisplacementField3D::getVZ() ";
 
  if ( mode == VOXEL )
    return vz;
  else 
    throw InvalidDisplacementField3DMode(from);
  
}


void DisplacementField3D::setVX(float _vx) throw(DisplacementField3D::InvalidDisplacementField3DMode) {

  char from[] = " DisplacementField3D::setVX(float) ";
 
  if ( mode == VOXEL )
    vx = _vx;
  else 
    throw InvalidDisplacementField3DMode(from);
  
}


void DisplacementField3D::setVY(float _vy) throw (DisplacementField3D::InvalidDisplacementField3DMode) {
  
  char from[] = " DisplacementField3D::setVY() ";
 
  if ( mode == VOXEL )
    vy = _vy;
  else 
    throw InvalidDisplacementField3DMode(from);
  
}

void DisplacementField3D::setVZ(float _vz) throw (DisplacementField3D::InvalidDisplacementField3DMode) {
  
  char from[] = " DisplacementField3D::setVZ() ";
 
  if ( mode == VOXEL )
    vz = _vz;
  else 
    throw InvalidDisplacementField3DMode(from);
  
}

void DisplacementField3D::initGeneralFormat() {

  filename = "NoName.fld3D";
  addModuleVersion("Inrimage names", true, 1.0,
		   (GeneralFormatModuleVersion::INPUTREADFUNCTION)
		   &DisplacementField3D::readInrNames,
		   (GeneralFormatModuleVersion::INPUTSKIPFUNCTION)
		   &DisplacementField3D::skipInrNames,
		   (GeneralFormatModuleVersion::OUTPUTWRITEFUNCTION)
		   &DisplacementField3D::writeInrNames);
		   
}

void DisplacementField3D::readInrNames(GFInput &input) throw(InvalidDisplacementField3DImagesNb) {
  
  TclString name;
  Inrimage *inr,*im=0;
  char from[] = "DisplacementField3D::readInrNames(GFInput &)";
  unsigned int nb = 3;
  
  TclString txt;

  input >> txt;

  if (!strncmp(txt, "MODE=",  5))
    if (!strncmp((char*)txt+5, "real", 4)) mode = REAL;
    else if (!strncmp((char*)txt+5, "voxel", 5)) mode = VOXEL;
    else CERR << "MODE NON RECONNU DANS " << from << ENDL;
  
  input >> txt;
  if (!strncmp(txt, "VX=", 3))
    if(sscanf((char*)txt+3, "%f", &vx) != 1) CERR << "Impossible de lire VX dans " <<  from << ENDL;

  input >> txt;
  if (!strncmp(txt, "VY=", 3))
    if(sscanf((char*)txt+3, "%f", &vy) != 1) CERR << "Impossible de lire VY dans " <<  from << ENDL;

  input >> txt;
  if (!strncmp(txt, "VZ=", 3))
    if(sscanf((char*)txt+3, "%f", &vz) != 1) CERR << "Impossible de lire VZ dans " <<  from << ENDL;

  
  for(unsigned int i = 0; i < nb; i++) {
    input >> name;
    if(name[0] != '/') {
      int j = filename.getLength() - 1;
      while(j >= 0 && filename[j] != '/') j--;
      if(j > 0) name = TclString(filename, 0, j) + name;
    }
    try {
      inr = new Inrimage(name);
    }
    catch(Inrimage::ImageNotFound &e) {
      e.push(from);
      throw e;
    }
    /* catch(Exception &) {
       throw Inrimage::ReadError(from, (char *) name);
      } */
    if (i>0)
      {
	if (inr->getType()!=im->getType())
	  throw (DisplacementField3DMismatchType("DisplacementField3D::DisplacementField3D (FIELD_MODE m, Inrimage *imx, Inrimage *imy, Inrimage *imz, float vx, float vy, float vz ) "));
	if ((inr->getX()!=im->getX())|| (inr->getY()!=im->getY()) ||
	    (inr->getZ()!=im->getZ())  || (inr->getVdim()!=im->getVdim()) )
	  throw (DisplacementField3DMismatchSize("DisplacementField3D::DisplacementField3D (FIELD_MODE m, Inrimage *imx, Inrimage *imy, Inrimage *imz, float vx, float vy, float vz ) "));
      }
    else 
      im=inr;
    
    switch (i) {
    case 0:
      inrx = new ObjectPointer<Inrimage>(this, inr);
      break;
    case 1:
      inry = new ObjectPointer<Inrimage>(this, inr);
      break;
    case 2:
      inrz = new ObjectPointer<Inrimage>(this, inr);
      break;
    default:
      break;
    }
  }
 
  type = inr->getType();
 
}





void DisplacementField3D::skipInrNames(GFInput &input) {
  unsigned int nb;
  input >> nb;
  TclString name;
  for(unsigned int i = 0; i < nb; i++) input >> name;
}

void DisplacementField3D::writeInrNames(GFOutput &output) {

  char from[] = "void DisplacementField3D::writeInrNames(GFOutput &output) ";
  TclString txt;

  if ( mode == REAL )
     txt.sprintf("MODE=real\nVX=%f\nVY=%f\nVZ=%f\n", vx, vy, vz);
  else if ( mode == VOXEL )
    txt.sprintf("MODE=voxel\nVX=%f\nVY=%f\nVZ=%f\n", vx, vy, vz);
  else
    CERR << "Mode non reconnu dans "  << from << ENDL;
     
  /* get base name of the 4D image */
  TclString name;
  TclString base = output.getFileName();

  if((base.getLength() > 6) &&
     (TclString(base, base.getLength()-6, base.getLength()-1) == ".inr4D"))
    base = TclString(base, 0, base.getLength()-7);  
  if((base.getLength() > 2) &&
     (TclString(base, base.getLength()-2, base.getLength()-1) == "4D"))
    base = TclString(base, 0, base.getLength()-3);      


  if ( ((*inrx)->getName()=="noname.inr.gz") || ((*inrx)->getName()==EMPTY_STRING)) {
    name.sprintf("%s%03u.inr.gz\n", (char *) base, 0);
    //output << name << endl;
    txt.append(name);
    (*inrx)->write(name);
  }
  else {
    //output << (*inrx)->getName() << endl;
    txt.append((*inrx)->getName());
    txt.append("\n");
  }

  if ( ((*inry)->getName()=="noname.inr.gz") || ((*inry)->getName()==EMPTY_STRING)) {
    name.sprintf("%s%03u.inr.gz\n", (char *) base, 1);
    //output << name << endl;
    txt.append(name);
    (*inry)->write(name);
  }
  else {
    //output << (*inry)->getName() << endl;
    txt.append((*inry)->getName());
    txt.append("\n");
  }

  if ( ((*inrz)->getName()=="noname.inr.gz") || ((*inrz)->getName()==EMPTY_STRING)) {
    name.sprintf("%s%03u.inr.gz\n", (char *) base, 1);
    //output << name << endl;
    txt.append(name);
    (*inrz)->write(name);
  }
  else {
    //output << (*inrz)->getName() << endl;
    txt.append((*inrz)->getName());
    txt.append("\n");
  }



  output << txt;

}

void DisplacementField3D::VoxelToReal(void) throw (DisplacementField3D::InvalidDisplacementField3DMode) {

  char from[] = "void DisplacementField3D::VoxelToReal(void) throw (DisplacementField3D::InvalidDisplacementField3DMode)";

  if ( mode != VOXEL )
    throw DisplacementField3D::InvalidDisplacementField3DMode(from);
  
  yav::mul((*inrx)(),vx);
  yav::mul((*inry)(),vy);
  yav::mul((*inrz)(),vz);

}



void DisplacementField3D::RealToVoxel(float _vx, float _vy, float _vz) throw (DisplacementField3D::InvalidDisplacementField3DMode, Inrimage::InvalidVoxelSize) {

  char from[] = "void DisplacementField3D::RealToVoxel(float vx, float vy, float vz) throw (DisplacementField3D::InvalidDisplacementField3DMode)";

  if ( mode != REAL )
    throw DisplacementField3D::InvalidDisplacementField3DMode(from);
  
  if ( (_vx <=0) || (_vy <=0) || (_vz <=0) )
    throw Inrimage::InvalidVoxelSize(from);
  
  vx = _vx;
  vy = _vy;
  vz = _vz;

  yav::div((*inrx)(),vx);
  yav::div((*inry)(),vy);
  yav::div((*inrz)(),vz);

}

} // namespace yav
