#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/GLScale.h"
#include <inrimage/GLScale.h>


namespace yav {



GLScale::GLScale(TclModule *module, const TclString& scaleName,Scalable *elt)
  : TclScale(typeid(GLScale), scaleName,module,elt,"inrimage::glscalemenu") {
  init();
}
 

GLScale::GLScale(TclModule *module, const TclString& scaleName,
		 std::list<Scalable *>& list)
  : TclScale(typeid(GLScale), scaleName, module,list,"inrimage::glscalemenu") {
  init();
}


GLScale::~GLScale() {
}


void GLScale::init(void) {
  firstupdate = false;
  tickle("bind %s <ButtonRelease-1> {%s%u -command [%s get]};\
bind %s <Enter> \"focus %s\";\
bind %s <KeyRelease-Left> {%s%u -command [%s get] };\
bind %s <KeyRelease-Right> {%s%u -command [%s get] };\
bind %s <KeyRelease-Down> {%s%u -command [%s get]};\
bind %s <KeyRelease-Up> {%s%u -command [%s get]}",
		 (char *) TCLwindow, getObjectName(),
		 getRef(), (char *) TCLwindow,
		 (char *) TCLwindow, (char *) TCLwindow,
		 (char *) TCLwindow, getObjectName(),
		 getRef(), (char *) TCLwindow,
		 (char *) TCLwindow, getObjectName(),
		 getRef(), (char *) TCLwindow,
		 (char *) TCLwindow, getObjectName(),
		 getRef(), (char *) TCLwindow,
		 (char *) TCLwindow, getObjectName(),
		 getRef(), (char *) TCLwindow);
	 
  tickle("bind %s <ButtonRelease-2> [bind %s <ButtonRelease-1>]",
		 (char *) TCLwindow, (char *) TCLwindow);
}


void GLScale::reset(Scalable *sa, const TclString& lab) {
  min = sa->getMin(this);
  max = sa->getMax(this);

  
  if(min == (int) min && max == (int) max)
    tickle("%s configure -from %i -to %i -resolution %f",
	   (char *) TCLwindow, (int) min, (int) max, sa->getStep(this));
  else
    tickle("%s configure -from %f -to %f -resolution %f",
	   (char *) TCLwindow, min, max, sa->getStep(this));

  if(lab != EMPTY_STRING) setLabel(lab);
  command(get());
}


void GLScale::set(double val) {
  to(val);
  command(val);
}


void GLScale::activation(void) {
  TclScale::activation();
  command(get());
}


MinGLScale::MinGLScale(TclModule *module, const TclString& scaleName,Scalable *elt)
  : GLScale(module, scaleName,elt) {
}


MinGLScale::MinGLScale(TclModule *module, const TclString& scaleName,
		       std::list<Scalable *> list)
  : GLScale(module, scaleName, list) {
}


MinGLScale::~MinGLScale() {
}


void MinGLScale::reset(Scalable *sa, const TclString& lab) {
  GLScale::reset(sa, lab); 
  value = min;
  sa->scaleValue(this,value);
  tickle("%s set %f", (char *) TCLwindow, value);
}


MaxGLScale::~MaxGLScale() {
}


MaxGLScale::MaxGLScale(TclModule *module, const TclString& scaleName,Scalable *elt)
  : GLScale(module, scaleName,elt) {
}


MaxGLScale::MaxGLScale(TclModule *module, const TclString& scaleName,
		       std::list<Scalable *> list)
  : GLScale(module, scaleName, list) {
}


void MaxGLScale::reset(Scalable *sa, const TclString& lab) {
  GLScale::reset(sa, lab);
  value = max;
  sa->scaleValue(this,value);
  tickle("%s set %f", (char *) TCLwindow, value);
}

 

/* "-in %w [-add slicecam%u+]",
   "Create a new grey level scale object associated to given TCL scale (-in argument) and add any given slicecam to scale. Only slicecam with the same type can be added to a glscale."*/
int GLScale::create(TclModule *m, int argc, char **argv) {
  GLScale *gsc;
  std::list<Scalable *> inrvl;
  char *name = 0;
  int i;
  int minmax = 0;

  for(i = 1; i < argc; i++) {
    // -in %w
    if(!strcmp(argv[i], "-in")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(!m->windowExists(argv[i])) return COMMAND_SYNOPSIS;
      name = argv[i];
    }

    // -min
    else if(!strcmp(argv[i], "-min")) minmax = 1;

    // -max
    else if(!strcmp(argv[i], "-max")) minmax = 2;


   
    else return COMMAND_SYNOPSIS;
  }

  if(!name || !minmax) return COMMAND_SYNOPSIS;
  if(minmax == 1)
    gsc = new MinGLScale(m, name);
  else
    gsc = new MaxGLScale(m, name);
  m->setResult("%s%u", (char *) gsc->getObjectName(), gsc->getRef());
  return TCL_OK;
}



} // namespace yav

