#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/Image3DCameraBackground.h"
#include <inrimage/Image3DCameraBackground.h>
#include <graphics2D/Camera2DActivation.h>
#include <inrimage/Image3DBackgroundManager.h>

namespace yav {

AUTOLINK_STATIC_VARIABLES(Image3DCameraBackground);

   /// function which computes and store the vector norm and which fills the buffer with the right texture
  template <class T> void computeAndDrawVectorNormSliceZ(T *type,const  Inrimage *inr,double ** &normVector, unsigned int *buffer,
	  const double minVal, const double maxVal,
	  const unsigned int nrows, const unsigned int ncols, const unsigned int nbColumnsTexture,
	  const unsigned int sliceNumber,
		const unsigned char *redColorMap,
		const unsigned char *greenColorMap,
		const unsigned char *blueColorMap,
		const double opacity){
	  T *in;
	  unsigned int i,j,k,q;
	  double a= 1.0/(maxVal-minVal);

	  unsigned int vdim=inr->getVdim();
	  double *n;
	  unsigned int *b;
	  unsigned char *p;
	  unsigned char v;
	  unsigned char vo=(unsigned char)(opacity*255);
	  double val;
	  if (normVector==0) {
		  unsigned int componentXYZOffset=inr->getX()*inr->getY()*inr->getZ();
  		  unsigned int componentXYOffset=inr->getX()*inr->getY();
		  normVector= new (double *[nrows]);
		  /// store the norm of each vector in an array as to avoid unnecessary computation
		  if (inr->getVectorialMode()==VM_INTERLACED) {
			  in=&(((T ***)inr->getArray())[0][0][0]);
			  in+=componentXYOffset*vdim*sliceNumber;
			  for (j=0;j<nrows;j++) {
				  n=normVector[j]=new double [ncols];
				  for (i=0;i<ncols;i++,n++) {
					  *n=0;
					  for (q=0;q<vdim;++q,++in)
						  (*n)+=(*in)*(*in);
					  (*n)=sqrt(*n);
				  }
			  }
		  } else {
			  in=&(((T ***)inr->getArray())[0][0][0]);
			  T *inp;

			  for (j=0;j<nrows;j++) {
				  n=normVector[j]=new double [ncols];
				  for (i=0;i<ncols;i++,n++,in++) {
					  *n=0;
					  for (inp=in,q=0;q<vdim;++q,inp+=componentXYZOffset)
						  (*n)+=(*inp)*(*inp);
					  (*n)=sqrt(*n);
				  }
			  }
		  }
	  }
	  k=sliceNumber;
	  for (j=0;j<nrows;j++) {
		  b=&buffer[j*nbColumnsTexture];
		  n=normVector[j];
		  for (i=0;i<ncols;i++,b++,n++) {

			  val=(*n);
			  p=(unsigned char *) b;
			  p[3]=vo;
			  if (val<minVal) {
				  p[0]=redColorMap[256];
				  p[1]=greenColorMap[256];
				  p[2]=blueColorMap[256];
			  }  
			  else if  (val>maxVal){
				  p[0]=redColorMap[257];
				  p[1]=greenColorMap[257];
				  p[2]=blueColorMap[257];
			  } 
			  else {
				  v=(unsigned char )(255*(val-minVal)*a);
				  p[0]=redColorMap[v];
				  p[1]=greenColorMap[v];
				  p[2]=blueColorMap[v];
			  }
		  }
	  }
  }
/// function which computes and store the vector norm and which fills the buffer with the right texture
  template <class T> void computeAndDrawVectorNormSliceX(T *type,const  Inrimage *inr,double ** &normVector, unsigned int *buffer,
	  const double minVal, const double maxVal,
	  const unsigned int nrows, const unsigned int ncols, const unsigned int nbColumnsTexture,
	  const unsigned int sliceNumber,
		const unsigned char *redColorMap,
		const unsigned char *greenColorMap,
		const unsigned char *blueColorMap,
		const double opacity){
	  T *in;
	  unsigned int i,j,q;
	  double a= 1.0/(maxVal-minVal);

	  unsigned int vdim=inr->getVdim();
	  double *n;
	  unsigned int *b;
	  unsigned char *p;
	  unsigned char v;
	  unsigned char vo=(unsigned char)(opacity*255);
	  double val;
	  if (normVector==0) {
  		  unsigned int componentXYOffset=inr->getX()*inr->getY();
		  normVector= new (double *[nrows]);
		  /// store the norm of each vector in an array as to avoid unnecessary computation
		  if (inr->getVectorialMode()==VM_INTERLACED) {
				componentXYOffset *=vdim;
			  for (j=0;j<nrows;j++) {
				  in=&(((T ***)inr->getArray())[0][0][0]);

				  in += (sliceNumber*vdim+j*vdim*inr->getX());
				  
				  n=normVector[j]=new double [ncols];
				  for (i=0;i<ncols;i++,n++,in+=componentXYOffset-vdim) {
					  *n=0;
					  for (q=0;q<vdim;++q,++in)
						  (*n)+=(*in)*(*in);
					  (*n)=sqrt(*n);
				  }
			  }
		  } else {
			  T *inp;

			  for (j=0;j<nrows;j++) {
  				  in=&(((T ***)inr->getArray())[0][0][0]);
				  in+=sliceNumber+j*inr->getX();
				  n=normVector[j]=new double [ncols];
				  for (i=0;i<ncols;i++,n++,in+=componentXYOffset*vdim) {
					  *n=0;
					  for (inp=in,q=0;q<vdim;++q,inp+=componentXYOffset)
						  (*n)+=(*inp)*(*inp);
					  (*n)=sqrt(*n);
				  }
			  }
		  }
	  }
	  for (j=0;j<nrows;j++) {
		  b=&buffer[j*nbColumnsTexture];
		  n=normVector[j];
		  for (i=0;i<ncols;i++,b++,n++) {

			  val=(*n);
			  p=(unsigned char *) b;
			  p[3]=vo;
			  if (val<minVal) {
				  p[0]=redColorMap[256];
				  p[1]=greenColorMap[256];
				  p[2]=blueColorMap[256];
			  }  
			  else if  (val>maxVal){
				  p[0]=redColorMap[257];
				  p[1]=greenColorMap[257];
				  p[2]=blueColorMap[257];
			  } 
			  else {
				  v=(unsigned char )(255*(val-minVal)*a);
				  p[0]=redColorMap[v];
				  p[1]=greenColorMap[v];
				  p[2]=blueColorMap[v];
			  }
		  }
	  }
  }
/// function which computes and store the vector norm and which fills the buffer with the right texture
  template <class T> void computeAndDrawVectorNormSliceY(T *type,const  Inrimage *inr,double ** &normVector, unsigned int *buffer,
	  const double minVal, const double maxVal,
	  const unsigned int nrows, const unsigned int ncols, const unsigned int nbColumnsTexture,
	  const unsigned int sliceNumber,
	  const unsigned char *redColorMap,
	  const unsigned char *greenColorMap,
	  const unsigned char *blueColorMap,
	  const double opacity){
		  T *in;
		  unsigned int i,q;
		  int j;
		  double a= 1.0/(maxVal-minVal);

		  unsigned int vdim=inr->getVdim();
		  double *n;
		  unsigned int *b;
		  unsigned char *p;
		  unsigned char v;
	  	  unsigned char vo=(unsigned char)(opacity*255);

		  double val;
		  if (normVector==0) {
			  unsigned int componentXYOffset=inr->getX()*inr->getY();
			  normVector= new (double *[nrows]);
			  /// store the norm of each vector in an array as to avoid unnecessary computation
			  if (inr->getVectorialMode()==VM_INTERLACED) {
				  componentXYOffset *=vdim;
				  for (j=nrows-1;j>=0;j--) {
				  in=&(((T ***)inr->getArray())[0][0][0]);
				  in+=j*componentXYOffset+(sliceNumber)*vdim*inr->getX();
				  n=normVector[j]=new double [ncols];
				  for (i=0;i<ncols;i++,n++) {
					  *n=0;
					  for (q=0;q<vdim;++q,++in)
						  (*n)+=(*in)*(*in);
					  (*n)=sqrt(*n);
				  }
			  }
		  } else {
			  T *inp;

			  for (j=0;j<(int)nrows;j++) {
				  in=&(((T ***)inr->getArray())[0][0][0]);
				  in+=j*vdim*componentXYOffset+(sliceNumber)*inr->getX();
		
				  n=normVector[j]=new double [ncols];
				  for (i=0;i<ncols;i++,n++,in++) {
					  *n=0;
					  for (inp=in,q=0;q<vdim;++q,inp+=componentXYOffset)
						  (*n)+=(*inp)*(*inp);
					  (*n)=sqrt(*n);
				  }
			  }
		  }
	  }
	  for (j=0;j<(int)nrows;j++) {
		  b=&buffer[(nrows-1-j)*nbColumnsTexture];
		  n=normVector[j];
		  for (i=0;i<ncols;i++,b++,n++) {

			  val=(*n);
			  p=(unsigned char *) b;
			  p[3]=vo;
			  if (val<minVal) {
				  p[0]=redColorMap[256];
				  p[1]=greenColorMap[256];
				  p[2]=blueColorMap[256];
			  }  
			  else if  (val>maxVal){
				  p[0]=redColorMap[257];
				  p[1]=greenColorMap[257];
				  p[2]=blueColorMap[257];
			  } 
			  else {
				  v=(unsigned char )(255*(val-minVal)*a);
				  p[0]=redColorMap[v];
				  p[1]=greenColorMap[v];
				  p[2]=blueColorMap[v];
			  }
		  }
	  }
  }
 template <class T> void drawVectorSliceZ(T *type,const Image3DTextureInformation *t)
{
		Image3DCameraBackground *b=t->getBackground();
        unsigned int i,j,k;
		Inrimage *inr=b->getInrimage();
		unsigned int vdim=inr->getVdim();
		T *in;
		unsigned int nrows=t->getNbRealRows();
		unsigned int ncols=t->getNbRealColumns();
		k=t->getSliceNumber();
		unsigned int step=b->getVectorXYStep();
		// the origin of the vector
		double px,py;
		double vectorScale=b->getVectorScale();
		// the vector to be drawn
		double vx,vy;
		double vr=t->getVr();
		double vc=t->getVc();
		double arrowWidth=b->getArrowWidth();
		double arrowHeight=1.0-b->getArrowHeight();

		bool showArrows=b->getShowArrows();
		/// define the color
		glColor3f(b->getVectorColor(0),
			b->getVectorColor(1),
			b->getVectorColor(2));
		glBegin(GL_LINES);

		/// store the norm of each vector in an array as to avoid unnecessary computation
		if (inr->getVectorialMode()==VM_INTERLACED) {
			for (j=0;j<nrows;j+=step) {
				in=&(((T ***)inr->getArray())[k][j][0]);
				for (i=0;i<ncols;i+=step,in+=step*vdim) {
					// get the vector length
					vx=(double)(*in);
					vy=(double) -(*(in+1));
					// get the vector origin
					px=i*vc;
					py=(nrows-1-j)*vr;
					glVertex2d(px,py);
					glVertex2d(px+vx*vectorScale,
						py+vy*vectorScale);
					if (showArrows) {
						glVertex2d(px+vectorScale*(arrowHeight*vx-arrowWidth*vy),
							py+vectorScale*(arrowHeight*vy+arrowWidth*vx));
						glVertex2d(px+vx*vectorScale,
							py+vy*vectorScale);
						glVertex2d(px+vectorScale*(arrowHeight*vx+arrowWidth*vy),
							py+vectorScale*(arrowHeight*vy-arrowWidth*vx));
						glVertex2d(px+vx*vectorScale,
							py+vy*vectorScale);
					}
				}
			}

		} else  {
			for (j=0;j<nrows;j+=step) {
				unsigned int componentOffset=inr->getX()*inr->getY();

				in=&(((T ***)inr->getArray())[k*vdim][j][0]);
				for (i=0;i<ncols;i+=step,in+=step) {
					// get the vector length
					vx=(double)(*in);
					vy=(double) (*(in+componentOffset));
					// get the vector origin
					px=i*vc;
					py=(nrows-1-j)*vr;
					glVertex2d(px,py);
					glVertex2d(px+vx*vectorScale,
						py+vy*vectorScale);
					if (showArrows) {
						glVertex2d(px+vectorScale*(arrowHeight*vx-arrowWidth*vy),
							py+vectorScale*(arrowHeight*vy+arrowWidth*vx));
						glVertex2d(px+vx*vectorScale,
							py+vy*vectorScale);
						glVertex2d(px+vectorScale*(arrowHeight*vx+arrowWidth*vy),
							py+vectorScale*(arrowHeight*vy-arrowWidth*vx));
						glVertex2d(px+vx*vectorScale,
							py+vy*vectorScale);
					}
				}
			}
		}
		glEnd();
	  }
template <class T> void drawVectorSliceX(T *type,const Image3DTextureInformation *t)
{
		Image3DCameraBackground *b=t->getBackground();
        unsigned int i,j,k;
		Inrimage *inr=b->getInrimage();
		unsigned int vdim=inr->getVdim();
		T *in;
		unsigned int nrows=t->getNbRealRows();
		unsigned int ncols=t->getNbRealColumns();
		k=t->getSliceNumber();
		unsigned int step=b->getVectorXYStep();
		unsigned int zstep=b->getVectorZStep();
		// the origin of the vector
		double px,py;
		double vectorScale=b->getVectorScale();
		// the vector to be drawn
		double vx,vy;
		double vr=t->getVr();
		double vc=t->getVc();
		double arrowWidth=b->getArrowWidth();
		double arrowHeight=1.0-b->getArrowHeight();

		bool showArrows=b->getShowArrows();
		/// define the color
		glColor3f(b->getVectorColor(0),
			b->getVectorColor(1),
			b->getVectorColor(2));
		glBegin(GL_LINES);
   	   unsigned int componentXYOffset=inr->getX()*inr->getY();

		/// store the norm of each vector in an array as to avoid unnecessary computation
		if (inr->getVectorialMode()==VM_INTERLACED) {
				componentXYOffset *=vdim*zstep;
			for (j=0;j<nrows;j+=step) {
				  in=&(((T ***)inr->getArray())[0][0][0]);
				  in += vdim*(k+j*inr->getX());

				for (i=0;i<ncols;i+=zstep,in+=componentXYOffset) {
					// get the vector length
					vx=(double)(*(in+2));
					vy=(double) -(*(in+1));
					// get the vector origin
					px=i*vc;
					py=(nrows-1-j)*vr;
					glVertex2d(px,py);
					glVertex2d(px+vx*vectorScale,
						py+vy*vectorScale);
					if (showArrows) {
						glVertex2d(px+vectorScale*(arrowHeight*vx-arrowWidth*vy),
							py+vectorScale*(arrowHeight*vy+arrowWidth*vx));
						glVertex2d(px+vx*vectorScale,
							py+vy*vectorScale);
						glVertex2d(px+vectorScale*(arrowHeight*vx+arrowWidth*vy),
							py+vectorScale*(arrowHeight*vy-arrowWidth*vx));
						glVertex2d(px+vx*vectorScale,
							py+vy*vectorScale);
					}
				}
			}
		} else  {
			for (j=0;j<nrows;j+=step) {
				  in=&(((T ***)inr->getArray())[0][0][0]);
				  in += (k*vdim+j*vdim*inr->getX());
				for (i=0;i<ncols;i+=zstep,in+=zstep*vdim*componentXYOffset) {
					// get the vector length
					vx=(double)(*(in+2*componentXYOffset));
					vy=(double) -(*(in+componentXYOffset));
					// get the vector origin
					px=i*vc;
					py=(nrows-1-j)*vr;
					glVertex2d(px,py);
					glVertex2d(px+vx*vectorScale,
						py+vy*vectorScale);
					if (showArrows) {
						glVertex2d(px+vectorScale*(arrowHeight*vx-arrowWidth*vy),
							py+vectorScale*(arrowHeight*vy+arrowWidth*vx));
						glVertex2d(px+vx*vectorScale,
							py+vy*vectorScale);
						glVertex2d(px+vectorScale*(arrowHeight*vx+arrowWidth*vy),
							py+vectorScale*(arrowHeight*vy-arrowWidth*vx));
						glVertex2d(px+vx*vectorScale,
							py+vy*vectorScale);
					}
				}
			}
		}
		glEnd();
	  }
template <class T> void drawVectorSliceY(T *type,const Image3DTextureInformation *t)
{
		Image3DCameraBackground *b=t->getBackground();
        unsigned int i,k,j;
		Inrimage *inr=b->getInrimage();
		unsigned int vdim=inr->getVdim();
		T *in;
		unsigned int nrows=t->getNbRealRows();
		unsigned int ncols=t->getNbRealColumns();
		k=b->getSliceNumber();
		unsigned int step=b->getVectorXYStep();
		unsigned int zstep=b->getVectorZStep();
		// the origin of the vector
		double px,py;
		double vectorScale=b->getVectorScale();
		// the vector to be drawn
		double vx,vy;
		double vr=t->getVr();
		double vc=t->getVc();
		double arrowWidth=b->getArrowWidth();
		double arrowHeight=1.0-b->getArrowHeight();

		bool showArrows=b->getShowArrows();
		/// define the color
		glColor3f(b->getVectorColor(0),
			b->getVectorColor(1),
			b->getVectorColor(2));
		glBegin(GL_LINES);
   	   unsigned int componentXYOffset=inr->getX()*inr->getY();
		/// store the norm of each vector in an array as to avoid unnecessary computation
		if (inr->getVectorialMode()==VM_INTERLACED) {
				componentXYOffset *=vdim;
			for (j=0;j<nrows;j+=zstep) {
				  in=&(((T ***)inr->getArray())[0][0][0]);
  				  in+=j*componentXYOffset+k*vdim*inr->getX();

				for (i=0;i<ncols;i+=step,in+=vdim*step) {
					// get the vector length
					vx=(double)(*(in));
					vy=(double) (*(in+2));
					// get the vector origin
					px=i*vc;
					py=j*vr;
					glVertex2d(px,py);
					glVertex2d(px+vx*vectorScale,
						py+vy*vectorScale);
					if (showArrows) {
						glVertex2d(px+vectorScale*(arrowHeight*vx-arrowWidth*vy),
							py+vectorScale*(arrowHeight*vy+arrowWidth*vx));
						glVertex2d(px+vx*vectorScale,
							py+vy*vectorScale);
						glVertex2d(px+vectorScale*(arrowHeight*vx+arrowWidth*vy),
							py+vectorScale*(arrowHeight*vy-arrowWidth*vx));
						glVertex2d(px+vx*vectorScale,
							py+vy*vectorScale);
					}
				}
			}
		} else  {
			for (j=0;j<(unsigned int)nrows;j+=zstep) {
				  in=&(((T ***)inr->getArray())[0][0][0]);
				  in+=j*vdim*componentXYOffset+(k)*inr->getX();
				for (i=0;i<ncols;i+=step,in+=step) {
					// get the vector length
					vx=(double)(*(in));
					vy=(double) (*(in+2*componentXYOffset));
					// get the vector origin
					px=i*vc;
					py=j*vr;
					glVertex2d(px,py);
					glVertex2d(px+vx*vectorScale,
						py+vy*vectorScale);
					if (showArrows) {
						glVertex2d(px+vectorScale*(arrowHeight*vx-arrowWidth*vy),
							py+vectorScale*(arrowHeight*vy+arrowWidth*vx));
						glVertex2d(px+vx*vectorScale,
							py+vy*vectorScale);
						glVertex2d(px+vectorScale*(arrowHeight*vx+arrowWidth*vy),
							py+vectorScale*(arrowHeight*vy-arrowWidth*vx));
						glVertex2d(px+vx*vectorScale,
							py+vy*vectorScale);
					}
				}
			}
		}
		glEnd();
	  }

/// function which fills the buffer with the right texture
  template <class T> void drawSliceZ(T *type,const  Inrimage *inr,unsigned int *buffer,
	  const double minWindow, const double maxWindow,
	  const unsigned int nbRealRows, const unsigned int nbRealColumns, const unsigned int ncols,
	  const unsigned int sliceNumber,
	  const unsigned char *redColorMap,
	  const unsigned char *greenColorMap,
	  const unsigned char *blueColorMap,
	  double opacity)
  {
	  T *in;
	  double a= 1.0/(maxWindow-minWindow);
	  double val;
	  unsigned int i,j,k,*b;
	  unsigned char *p,v;
	  unsigned char vo=(unsigned char)(opacity*255);

	  k=sliceNumber;
	  in=&(((T ***)inr->getArray())[k][0][0]);
	  for (j=0;j<nbRealRows;j++) {
		  b=&buffer[j*ncols];
		  for (i=0;i<nbRealColumns;i++,b++,in++) {
			  val= *in;
			  p=(unsigned char *) b;
			  p[3]=vo;
			  if (val<minWindow) {
				  p[0]=redColorMap[256];
				  p[1]=greenColorMap[256];
				  p[2]=blueColorMap[256];
			  }  
			  else if  (val>maxWindow){
				  p[0]=redColorMap[257];
				  p[1]=greenColorMap[257];
				  p[2]=blueColorMap[257];
			  } 
			  else {
				  v=(unsigned char )(255*(val-minWindow)*a);
				  p[0]=redColorMap[v];
				  p[1]=greenColorMap[v];
				  p[2]=blueColorMap[v];
			  }
		  }
	  }
  }
  /// function which fills the buffer with the right texture
  template <class T> void drawSliceY(T *type,const  Inrimage *inr,unsigned int *buffer,
	  const double minWindow, const double maxWindow,
	  const unsigned int nbRealRows, const unsigned int nbRealColumns, const unsigned int ncols,
	  const unsigned int sliceNumber,
	  const unsigned char *redColorMap,
	  const unsigned char *greenColorMap,
	  const unsigned char *blueColorMap,
	  const double opacity)
  {
	  T *in;
	  double a= 1.0/(maxWindow-minWindow);
	  double val;
	  unsigned int i,j,k,*b;
	  unsigned char *p,v;
	  unsigned char vo=(unsigned char)(opacity*255);

	  j=sliceNumber;
	  for (k=0;k<nbRealRows;k++) {
		  b=&buffer[k*ncols];
		  in=&(((T ***)inr->getArray())[nbRealRows-1-k][j][0]);

		  for (i=0;i<nbRealColumns;i++,b++,in++) {

			  val= *in;
			  p=(unsigned char *) b;
			  p[3]=vo;
			  if (val<minWindow) {
				  p[0]=redColorMap[256];
				  p[1]=greenColorMap[256];
				  p[2]=blueColorMap[256];
			  }  
			  else if  (val>maxWindow){
				  p[0]=redColorMap[257];
				  p[1]=greenColorMap[257];
				  p[2]=blueColorMap[257];
			  } 
			  else {
				  v=(unsigned char )(255*(val-minWindow)*a);
				  p[0]=redColorMap[v];
				  p[1]=greenColorMap[v];
				  p[2]=blueColorMap[v];
			  }
		  }
	  }
  }
    /// function which fills the buffer with the right texture
  template <class T> void drawSliceX(T *type,const  Inrimage *inr,unsigned int *buffer,
	  const double minWindow, const double maxWindow,
	  const unsigned int nbRealRows, const unsigned int nbRealColumns, const unsigned int ncols,
	  const unsigned int sliceNumber,
	  const unsigned char *redColorMap,
	  const unsigned char *greenColorMap,
	  const unsigned char *blueColorMap,
	  const double opacity)
  {
	  T *in;
	  double a= 1.0/(maxWindow-minWindow);
	  double val;
	  unsigned int i,j,k,*b,offset;
	  unsigned char *p,v;
	  unsigned char vo=(unsigned char)(opacity*255);

	  i=sliceNumber;
	  offset=inr->getX()*inr->getY();
	  for (j=0;j<nbRealRows;j++) {
		  b=&buffer[j*ncols];
		  in=&(((T ***)inr->getArray())[0][j][i]);

		  for (k=0;k<nbRealColumns;k++,b++,in+=offset) {

			  val= *in;
			  p=(unsigned char *) b;
			  p[3]=vo;
			  if (val<minWindow) {
				  p[0]=redColorMap[256];
				  p[1]=greenColorMap[256];
				  p[2]=blueColorMap[256];
			  }  
			  else if  (val>maxWindow){
				  p[0]=redColorMap[257];
				  p[1]=greenColorMap[257];
				  p[2]=blueColorMap[257];
			  } 
			  else {
				  v=(unsigned char )(255*(val-minWindow)*a);
				  p[0]=redColorMap[v];
				  p[1]=greenColorMap[v];
				  p[2]=blueColorMap[v];
			  }
		  }
	  }
  }


void Image3DCameraBackground::mouseButtonPressed(CameraTypes::WHICH_BUTTON button,
													  CameraTypes::KEY_MODIFIER modifier,
													  Camera2D *cam,int x,int y)
{
	if (manager)
		manager->mouseButtonPressed(button,modifier,this,x,y);
}
void Image3DCameraBackground::mouseButtonReleased(CameraTypes::WHICH_BUTTON button,
													   CameraTypes::KEY_MODIFIER modifier,Camera2D *cam,int x,int y)
{
	if (manager)
		manager->mouseButtonReleased(button,modifier,this,x,y);
}
void Image3DCameraBackground::mouseButtonMoved(Camera2D *cam,int x,int y	)
{
	if (manager)
		manager->mouseButtonMoved(this,x,y);
}

Image3DCameraBackground::Image3DCameraBackground(Camera2D *cam,
															   Image3DBackgroundManager *_manager,
															   SLICE_ORIENTATION geom) : Camera2DBackground(cam),
geometry(geom) {
	
  showAxis=false;
  manager=_manager;
	manager->addBackground(this);
		switch (geometry) {
			case SLICE_Y:
				sliceName="sliceY";
				break;
			case SLICE_X:
				sliceName="sliceX";
				break;
			case SLICE_Z:
				sliceName="sliceZ";
				break;
			default:
				break;
		}
	mainTexture = new Image3DTextureInformation(manager->getInrimage(),this);
	secondTexture=0;
	sliceNumber=mainTexture->sliceNumber;
	
  showVectors=true;
  showArrows=false;
  vectorColor[0]=1.0;
  vectorColor[1]=0.0;
  vectorColor[2]=0.0;
  arrowWidth=0.1;
  arrowHeight=0.3;

  vectorXYStep=9;
  vectorZStep=6;
  vectorId=0;
  vectorScale=5*vectorXYStep/manager->getMaxWindow();
  recomputeVectorDisplayList=true;

  if (manager->getSecondImage()) {
	  setSecondImage(manager->getSecondImage());
  }
}
Image3DTextureInformation::Image3DTextureInformation(Inrimage *inr,
													 Image3DCameraBackground *b)
{
	buffer=0;
	id=0;
	background=b;
   init(inr);
}
Image3DTextureInformation::~Image3DTextureInformation(){
	if (buffer)
	delete [] buffer;
	emptyNormVector();
}
Image3DCameraBackground::~Image3DCameraBackground() {
	if (mainTexture)
		delete mainTexture;
}
Inrimage *Image3DCameraBackground::getInrimage() const{
if (manager)
return(manager->getInrimage());
else 
return ((Inrimage *)0);
}

void Image3DTextureInformation::init(Inrimage *inr) {
	/// set the array as zero
	normVector=0;
    /// texture will be binded at the first redraw
	binded=false;
	offsetx=offsety=0;
	insideBackground=true;
	/// initialize the sliceNumber and sliceName
	switch (background->getGeometry()) {
		case SLICE_X:
			sliceNumber=inr->getX()/2;
			vr=inr->getVY();
			vc=inr->getVZ();
			nbRealRows=inr->getY();
			nbRealColumns=inr->getZ();
			break;
		case SLICE_Y:
			sliceNumber=inr->getY()/2;
			vr=inr->getVZ();
			vc=inr->getVX();
			nbRealRows=inr->getZ();
			nbRealColumns=inr->getX();
			break;
		case SLICE_Z:
			sliceNumber=inr->getZ()/2;
			vr=inr->getVY();
			vc=inr->getVX();
			nbRealRows=inr->getY();
			nbRealColumns=inr->getX();
			break;
		default:
			break;
	}
 /// computes the size of the texture buffer (must be a multiple of 2)
	unsigned int i;
  for(i=1;i<nbRealRows;i*=2);
  nrows=i;
  for(i=1;i<nbRealColumns;i*=2);
  ncols=i;
  if (buffer) 
	  delete [] buffer;
  /// alloc memory for buffer
  buffer=new unsigned int[nrows*ncols];
  /// put 0 in the image
  memset((void *) buffer,0,nrows*ncols*sizeof(unsigned int));

}
void Image3DTextureInformation::createTexture() {
  glGenTextures(1,&id);
  binded=true;
  glBindTexture(GL_TEXTURE_2D, id);
  updateTexture();
  
}

void Image3DTextureInformation::updateTexture() const {

	glPixelStorei(GL_UNPACK_ALIGNMENT, 4);


	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, ncols,nrows, 0,
		GL_RGBA, GL_UNSIGNED_BYTE, (void *)buffer);

	glBindTexture(GL_TEXTURE_2D, 0);
}


void Image3DCameraBackground::redraw(){


	glDisable(GL_LIGHTING);

	mainTexture->enableTexture(getInrimage(),
		manager->getRedColorMap(),
		manager->getGreenColorMap(),
		manager->getBlueColorMap(),
		manager->getMinWindow(),
		manager->getMaxWindow(),1.0);

	glColor4f(1,1,1,0.5);

	GLfloat glInterpolation;
	Image3DBackgroundManager::IMAGE_INTERPOLATION interpolation=manager->getInterpolationMethod();
	if (interpolation==Image3DBackgroundManager::NEAREST)
        glInterpolation=GL_NEAREST;
	else 
        glInterpolation=GL_LINEAR;

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, glInterpolation);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, glInterpolation);

	mainTexture->drawTexture();	

	if ( (secondTexture) && (secondTexture->insideBackground)){

		glBindTexture(GL_TEXTURE_2D, 0);
		glDisable(GL_LIGHTING);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
//		      glDepthMask(GL_FALSE);

		secondTexture->enableTexture(manager->getSecondImage(),
			manager->getRedColorMap2(),
			manager->getGreenColorMap2(),
			manager->getBlueColorMap2(),
			manager->getMinWindow2(),
			manager->getMaxWindow2(),1.0);

		glColor4f(1,1,1,(GLfloat)manager->getOpacity());
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, glInterpolation);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, glInterpolation);

		secondTexture->drawTexture();	
		glDisable(GL_BLEND);	

	}

	disableTexture();
	if (showAxis) {

		if ((currentMousePosition.getX()<mainTexture->vc*(mainTexture->getNbRealColumns()-.5)) &&
			(currentMousePosition.getY()<mainTexture->vr*(mainTexture->getNbRealRows()-.5))) { 
				glLineWidth(1.0);
				glBegin(GL_LINES);
				glColor3f(1.0,0.0,0.0);
				glVertex2d(-0.5*mainTexture->vc,currentMousePosition.getY());
				glVertex2d(mainTexture->vc*(mainTexture->getNbRealColumns()+.5),currentMousePosition.getY());
				glVertex2d(currentMousePosition.getX(),-0.5*mainTexture->vr);
				glVertex2d(currentMousePosition.getX(),mainTexture->vr*(mainTexture->getNbRealRows()+.5));
				glEnd();
			}
	}
//	showAxis=false;
	if ((showVectors) && (getInrimage()->getVdim()>1)) {
		if (recomputeVectorDisplayList)
			destroyVectorDisplayList();
		callVectorDisplayList();
	}
	
}

void Image3DTextureInformation::drawTexture() const
{
	double s=getNbRealRows()/(double)getNbRows();
	double t=getNbRealColumns()/(double) getNbColumns();

	glBegin(GL_QUADS);
	glTexCoord2d(0,0);
	glVertex2d(offsetx-0.5*vc, offsety+vr*(getNbRealRows()-.5)); 
	glTexCoord2d(t,0);
	glVertex2d(offsetx+vc*(getNbRealColumns()-.5),offsety+vr*(getNbRealRows()-.5));
	glTexCoord2d(t,s);
	glVertex2d(offsetx+vc*(getNbRealColumns()-.5),offsety-0.5*vr);
	glTexCoord2d(0,s);
	glVertex2d(offsetx-0.5*vc,offsety-0.5*vr); 
	glEnd();

}

void Image3DTextureInformation::drawTexture3D() const
{
	double s=getNbRealRows()/(double)getNbRows();
	double t=getNbRealColumns()/(double) getNbColumns();

	Vec3<double> corner3D;
	
	Vec3<double> org,normal;
	background->getImagePlane(org,normal);
	glNormal3dv((const double *)normal);
	glBegin(GL_QUADS);
	glTexCoord2d(0,0);
	Vec2<double> corner2D=Vec2<double>(offsetx-0.5*vc, offsety+vr*(getNbRealRows()-.5));
    background->map2DWorldToRaw3DWorldCoordinates(corner2D,corner3D);		
	glVertex3dv((const double *) corner3D); 
	glTexCoord2d(t,0);
	corner2D=Vec2<double>(offsetx+vc*(getNbRealColumns()-.5),offsety+vr*(getNbRealRows()-.5));
	 background->map2DWorldToRaw3DWorldCoordinates(corner2D,corner3D);		
	glVertex3dv((const double *) corner3D); 
	glTexCoord2d(t,s);
	corner2D=Vec2<double>(offsetx+vc*(getNbRealColumns()-.5),offsety-0.5*vr);
	 background->map2DWorldToRaw3DWorldCoordinates(corner2D,corner3D);		
	glVertex3dv((const double *) corner3D); 
	glTexCoord2d(0,s);
	corner2D=Vec2<double>(offsetx-0.5*vc,offsety-0.5*vr);
	 background->map2DWorldToRaw3DWorldCoordinates(corner2D,corner3D);		
	glVertex3dv((const double *) corner3D); 
	glEnd();

}

void Image3DCameraBackground::setSecondSliceNumber()
{
	Inrimage *secondImage=manager->getSecondImage();
	Vec3<double> offset=secondImage->getTranslation();
	double  sliceOffset;
	switch(geometry) {
		case SLICE_Z:
			sliceOffset=offset.getZ()/getInrimage()->getVZ();
			secondTexture->offsetx=offset.getX();
			secondTexture->offsety=offset.getY();
			break;
		case SLICE_Y:
			sliceOffset=offset.getY()/getInrimage()->getVY();
			secondTexture->offsetx=offset.getX();
			secondTexture->offsety=offset.getZ();
			break;
		case SLICE_X:
			sliceOffset=offset.getX()/getInrimage()->getVX();
			secondTexture->offsetx=offset.getZ();
			secondTexture->offsety=offset.getY();
			break;
		default:
			break;
	}
	int sn=(int)((sliceNumber/secondTexture->sliceRatio) +sliceOffset +.5);
	if ((sn<0)|| (sn>(int)secondTexture->maxSlice))
			secondTexture->insideBackground=false;
	else {
	secondTexture->insideBackground=true;
	secondTexture->sliceNumber= (unsigned int)sn;
	}
}

void Image3DCameraBackground::setSliceNumber(unsigned int s){
unsigned int smax;

switch (getGeometry()) {
case SLICE_X:
	smax=getInrimage()->getX();		
	break;
case SLICE_Y:
	smax=getInrimage()->getY();
	break;
case SLICE_Z:
	smax=getInrimage()->getZ();
	break;
default:
	break;
}
if (s<smax) {
    mainTexture->emptyNormVector();
	recomputeVectorDisplayList=true;
	sliceNumber=mainTexture->sliceNumber=s;
	if (secondTexture) {
		setSecondSliceNumber();
	}
	mainTexture->binded=false;
	if (secondTexture)
		secondTexture->binded=false;
	if (camera->getCameraActivation()) {
		camera->getCameraActivation()->redraw();
	}
	manager->redrawViewers();
	}
}

void Image3DCameraBackground::setColormap()
{
	mainTexture->binded=false;
if (secondTexture)
		secondTexture->binded=false;
	if (camera->getCameraActivation()) {
		camera->getCameraActivation()->redraw();
	}

}
const char *Image3DCameraBackground::getLabel(const Vec2<double> &pt){
	Vec4<double> voxel;
	TclString sval;
	map2DWorldToImageCoordinates(pt, voxel);
	if (getInrimage()->inside((int)voxel.getX(),(int)voxel.getY(),(int)voxel.getZ())) {
	Voxel *v = getInrimage()->getVoxel((int)(voxel.getX()),(int)(voxel.getY()),
		(int)(voxel.getZ()));
	v->print(sval);
	delete v;
	} else {
		sval="out";
	}
	voxelLabel.sprintf("Row %3u Col %3u Plane %3u Value %s", (unsigned int)(voxel.getY()),
			(unsigned int)(voxel.getX()),(unsigned int)(voxel.getZ()),(char *) sval);
	return((const char *)voxelLabel);
}
bool Image3DCameraBackground::mapImageTo2DWorldCoordinates(const Vec4<double> &pixel,Vec2<double> &pt ){
	switch(geometry) {
		case SLICE_Z:
			if (pixel.getZ()==sliceNumber) {
				pt.setValue(getInrimage()->getVX()*pixel.getX(),
					getInrimage()->getVY()*(getInrimage()->getY()-pixel.getY()-1));
				return true;
			} else {
				return false;
			}
			break;
		case SLICE_X:
			if (pixel.getX()==sliceNumber) {
				pt.setValue(getInrimage()->getVZ()*pixel.getZ(),
					getInrimage()->getVY()*(getInrimage()->getY()-pixel.getY()-1));
				return true;
			} else {
				return false;
			}
			break;
		case SLICE_Y:
			if (pixel.getY()==sliceNumber) {
				pt.setValue(getInrimage()->getVX()*pixel.getX(),
					getInrimage()->getVZ()*(pixel.getZ()));
				return true;
			} else {
				return false;
			}
			break;
		default:
			return false;
			break;
	}
}

void Image3DCameraBackground::map2DWorldToImageCoordinates(const Vec2<double> &pt, Vec4<double> &pixel)
{
	switch(geometry) {
		case SLICE_Z:
			pixel=Vec4<double>(pt.getX()/getInrimage()->getVX(),
				(getInrimage()->getY()-1-(pt.getY()/getInrimage()->getVY())),(double) sliceNumber,0.0);
			break;
		case SLICE_X:
			pixel=Vec4<double>((double) sliceNumber,
				(getInrimage()->getY()-1-(pt.getY()/getInrimage()->getVY())),
				pt.getX()/getInrimage()->getVZ(),0.0);
			break;
		case SLICE_Y:
			pixel=Vec4<double>(pt.getX()/getInrimage()->getVX(),(double) sliceNumber,
				(pt.getY()/getInrimage()->getVZ()),0.0);
			break;
		default:
			break;
	}
}
/** return the width of the main texture in world coordinates */ 
double Image3DCameraBackground::getRequiredWidth() const {
	return (mainTexture->vc)*(mainTexture->nbRealColumns);
}
/** return the width of the main texture in world coordinates */ 
double Image3DCameraBackground::getRequiredHeight() const {
	return (mainTexture->vr)*(mainTexture->nbRealRows);
}
double Image3DCameraBackground::getPixelSize() const {
	return (mainTexture->vr+mainTexture->vc)/2.0;
}
void Image3DCameraBackground::moveToPoint(Vec3<double> &pos) {
int sn;
	switch(geometry) {
		case SLICE_Z:
			sn=(int)((pos.getZ()/getInrimage()->getVZ())+.5);
			if (sn<0) {
				sn=0; 
			}
			else {
              if (sn>(int)(getInrimage()->getZ()-1))
				  sn=(int)(getInrimage()->getZ()-1);
			}
			setSliceNumber((unsigned int)sn);
			break;
		case SLICE_X:
			sn=( int)((pos.getX()/getInrimage()->getVX())+.5);
			if (sn<0) {
				sn=0; 
			}
			else {
              if (sn>(int)(getInrimage()->getX()-1))
				  sn=(int)(getInrimage()->getX()-1);
			}
			setSliceNumber((unsigned int)sn);
			break;
		case SLICE_Y:
			sn=( int)((pos.getY()/getInrimage()->getVY())+.5);
//			sn=(int)((getInrimage()->getY()-1-(pos.getY()/getInrimage()->getVY()))+.5);
			if (sn<0) {
				sn=0; 
			}
			else {
              if (sn>(int)(getInrimage()->getY()-1))
				  sn=(int)(getInrimage()->getY()-1);
			}
			setSliceNumber((unsigned int)sn);
			break;
		default:
			break;
	}

}
void Image3DCameraBackground::getViewportSize(Vec2<double> &vmin,Vec2<double> &vmax){
vmin=Vec2<double>(-mainTexture->vc*0.5,-mainTexture->vr*0.5);
vmax=Vec2<double>(mainTexture->vc*(mainTexture->getNbRealColumns()-0.5),mainTexture->vr*(mainTexture->getNbRealRows()-.5));
}
void Image3DCameraBackground::getImagePlane(Vec3<double> &org, 
												   Vec3<double> &normal)
{
  switch(geometry) {
		case SLICE_Z:
			org=Vec3<double>(0,0,sliceNumber*getInrimage()->getVZ());
			normal=Vec3<double>(0,0,1.0);
			break;
		case SLICE_Y:
			org=Vec3<double>(0,(sliceNumber)*getInrimage()->getVY(),0);
			normal=Vec3<double>(0,1.0,0.0);
			break;
		case SLICE_X:
			org=Vec3<double>(sliceNumber*getInrimage()->getVX(),0,0);
			normal=Vec3<double>(1,0,0.0);
			break;
		default:
			break;
	}								   
  if (getInrimage()->useTransformationMatrix()) {
	  Rotation3D rot=getInrimage()->getRotationMatrix();
	  normal=rot*normal;
	  org+=getInrimage()->getTranslation();
  }
}
bool Image3DCameraBackground::isPointInSlice(const Vec3<double> &p) 
{
	switch(geometry) {
		case SLICE_Z:
			if (fabs(p.getZ()-sliceNumber*getInrimage()->getVZ())<(getInrimage()->getVZ()/2.0)) {
				return true;
			} else 
				return false;
			break;
		case SLICE_Y:
			if (fabs(p.getY()-(sliceNumber)*getInrimage()->getVY())<(getInrimage()->getVY()/2.0)) {
				return true;
			} else 
				return false;
			break;
		case SLICE_X:
		default:
			if (fabs(p.getX()-sliceNumber*getInrimage()->getVX())<(getInrimage()->getVX()/2.0)) {
				return true;
			} else 
				return false;
			break;
	}
}
bool Image3DCameraBackground::map3DWorldTo2DWorldCoordinates(const Vec3<double> &pp, Vec2<double> &pt){
	Vec3<double> p=pp;

	if (getInrimage()->useTransformationMatrix()) {
		p=p-getInrimage()->getTranslation();
	}
	switch(geometry) {
		case SLICE_Z:
			if (fabs(p.getZ()-sliceNumber*getInrimage()->getVZ())<=(getInrimage()->getVZ()/2.0)) {
				pt=Vec2<double>(p.getX(),(getInrimage()->getY()-1)*getInrimage()->getVY()-p.getY());
				return true;
			} else 
				return false;
			break;
			case SLICE_Y:
			if (fabs(p.getY()-(sliceNumber)*getInrimage()->getVY())<=(getInrimage()->getVY()/2.0)) {

				//if (fabs(p.getY()-(getInrimage()->getY()-1-sliceNumber)*getInrimage()->getVY())<(getInrimage()->getVY()/2.0)) {
				pt=Vec2<double>(p.getX(),p.getZ());
				return true;
			} else 
				return false;
			break;
			case SLICE_X:
			default:
			if (fabs(p.getX()-sliceNumber*getInrimage()->getVX())<=(getInrimage()->getVX()/2.0)) {
				pt=Vec2<double>(p.getZ(),(getInrimage()->getY()-1)*getInrimage()->getVY()-p.getY());
				return true;
			} else 
				return false;
			break;
	}
}
void Image3DCameraBackground::map2DWorldTo3DWorldCoordinates(const Vec2<double> &pt, Vec3<double> &p){
	switch(geometry) {
		case SLICE_Z:
			p=Vec3<double>(pt.getX(),(getInrimage()->getY()-1)*getInrimage()->getVY()-pt.getY(),sliceNumber*getInrimage()->getVZ());
			break;
		case SLICE_Y:
			p=Vec3<double>(pt.getX(),(sliceNumber)*getInrimage()->getVY(),pt.getY());
			break;
		case SLICE_X:
			p=Vec3<double>(sliceNumber*getInrimage()->getVX(),(getInrimage()->getY()-1)*getInrimage()->getVY()-pt.getY(),pt.getX());
			break;
		default:
			break;
	}

	if (getInrimage()->useTransformationMatrix()) {
		p=p+getInrimage()->getTranslation();
	}
	
}
void Image3DCameraBackground::map2DWorldToRaw3DWorldCoordinates(const Vec2<double> &pt, Vec3<double> &p){
	switch(geometry) {
		case SLICE_Z:
			p=Vec3<double>(pt.getX(),(getInrimage()->getY()-1)*getInrimage()->getVY()-pt.getY(),sliceNumber*getInrimage()->getVZ());
			break;
		case SLICE_Y:
			p=Vec3<double>(pt.getX(),(sliceNumber)*getInrimage()->getVY(),pt.getY());
			break;
		case SLICE_X:
			p=Vec3<double>(sliceNumber*getInrimage()->getVX(),(getInrimage()->getY()-1)*getInrimage()->getVY()-pt.getY(),pt.getX());
			break;
		default:
			break;
	}
}
void Image3DTextureInformation::emptyNormVector() {
	if (normVector) {
		unsigned int i;
		for (i=0;i<nbRealRows;++i) {
			delete [] normVector[i];
		}
		delete [] normVector;
	}
	normVector=0;
}



void Image3DTextureInformation::drawVectors(void) {
	Inrimage *inr=background->getInrimage();
	switch(background->getGeometry()) {
  case SLICE_Z:
	  if (inr->getType()==Inrimage::WT_DOUBLE_VECTOR) {
		  drawVectorSliceZ<double>((double *)0,this);
	  } else   if (inr->getType()==Inrimage::WT_FLOAT_VECTOR) {
		  drawVectorSliceZ<float>((float *)0,this);
	  }
	  break;
  case SLICE_X:
	  if (inr->getVdim()>2) {
		  if (inr->getType()==Inrimage::WT_DOUBLE_VECTOR) {
			  drawVectorSliceX<double>((double *)0,this);
		  } else   if (inr->getType()==Inrimage::WT_FLOAT_VECTOR) {
			  drawVectorSliceX<float>((float *)0,this);
		  }
	  }
	  break;
 case SLICE_Y:
	  if (inr->getVdim()>2) {
		  if (inr->getType()==Inrimage::WT_DOUBLE_VECTOR) {
			  drawVectorSliceY<double>((double *)0,this);
		  } else   if (inr->getType()==Inrimage::WT_FLOAT_VECTOR) {
			  drawVectorSliceY<float>((float *)0,this);
		  }
	  }
	  break;
  default:
	  break;
	}
}
void Image3DCameraBackground::destroyVectorDisplayList(void) {
  if(vectorId > 0) {
    glDeleteLists(vectorId, 1);
    vectorId=0;
  }
}
void Image3DCameraBackground::callVectorDisplayList(void) {

  if(vectorId == 0) {
    vectorId = glGenLists(1);
    glNewList(vectorId, GL_COMPILE_AND_EXECUTE);
    mainTexture->drawVectors();
    glEndList();
	recomputeVectorDisplayList=false;
  }
  else glCallList(vectorId);
}
void Image3DTextureInformation::fillSliceZTexture(const Inrimage *inr,
												  const unsigned char *redColorMap,
												  const unsigned char *greenColorMap,
												  const unsigned char *blueColorMap,
												  const double minWindow,
												  const double maxWindow,
												  const double opacity) 
{
  switch(inr->getType()) {
  case Inrimage::WT_UNSIGNED_CHAR:
	  drawSliceZ<unsigned char>((unsigned char *)0,inr,buffer,minWindow,
		  maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity);   
	  break;
  case Inrimage::WT_SIGNED_CHAR:
	  drawSliceZ<char>((char *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity);    
	  break;
  case Inrimage::WT_UNSIGNED_SHORT:
	  drawSliceZ<unsigned short>((unsigned short *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity);    
	  break;
  case Inrimage::WT_SIGNED_SHORT:
	  drawSliceZ<short int>((short int *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity);   
	  break;
  case Inrimage::WT_UNSIGNED_INT:
	  drawSliceZ<unsigned int>((unsigned int *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity); 
	  break;
  case Inrimage::WT_SIGNED_INT:
	  drawSliceZ<int>((int *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity);
	  break;
  case Inrimage::WT_UNSIGNED_LONG:
	  drawSliceZ<unsigned LONGINT>((unsigned LONGINT *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity); 
	  break;
  case Inrimage::WT_SIGNED_LONG:
	  drawSliceZ<LONGINT>((LONGINT *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity); 
	  break;
  case Inrimage::WT_FLOAT:
	  drawSliceZ<float>((float *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity);    
	  break;
  case Inrimage::WT_DOUBLE:
	  drawSliceZ<double>((double *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity);  
	  break;
  case Inrimage::WT_FLOAT_VECTOR:
	  computeAndDrawVectorNormSliceZ<float>((float *)0, inr,normVector, 
		  buffer,minWindow,maxWindow,nbRealRows,nbRealColumns,ncols,sliceNumber,
		  redColorMap,greenColorMap,blueColorMap,opacity);
	  break;
  case Inrimage::WT_DOUBLE_VECTOR:
	  computeAndDrawVectorNormSliceZ<double>((double *)0, inr,normVector, 
		  buffer,minWindow,maxWindow,nbRealRows,nbRealColumns,ncols,sliceNumber,
		  redColorMap,greenColorMap,blueColorMap,opacity);
	  break;
  default:
	  {
		  double a= 1.0/(maxWindow-minWindow);
		  double val;
		  unsigned int i,j,k;
		  Voxel *vox=Voxel::voxelByType(inr->getType(),inr->getVdim());
		  unsigned int *b=buffer;
		  unsigned char *p,v;

		  k=sliceNumber;
		  for (j=0;j<nbRealRows;j++) {
			  b=&buffer[j*ncols];
			  for (i=0;i<nbRealColumns;i++,b++) {
				  inr->getVoxel(i,j,k,*vox);
				  val=vox->getValueAsDouble();
				  p=(unsigned char *) b;
				  if (val<minWindow) {
					  p[0]=redColorMap[256];
					  p[1]=greenColorMap[256];
					  p[2]=blueColorMap[256];
				  }  
				  else if  (val>maxWindow){
					  p[0]=redColorMap[257];
					  p[1]=greenColorMap[257];
					  p[2]=blueColorMap[257];
				  } 
				  else {
					  v=(unsigned char )(255*(val-minWindow)*a);
					  p[0]=redColorMap[v];
					  p[1]=greenColorMap[v];
					  p[2]=blueColorMap[v];
				  }
			  }
		  }
		  delete vox;
	  }

  }
}
void Image3DTextureInformation::fillSliceYTexture(const Inrimage *inr,
												  const unsigned char *redColorMap,
												  const unsigned char *greenColorMap,
												  const unsigned char *blueColorMap,
												  const double minWindow,
												  const double maxWindow,
												  const double opacity) 
{

  switch(inr->getType()) {
  case Inrimage::WT_UNSIGNED_CHAR:
	  drawSliceY<unsigned char>((unsigned char *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity);   
	  break;
  case Inrimage::WT_SIGNED_CHAR:
	  drawSliceY<char>((char *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity);    
	  break;
  case Inrimage::WT_UNSIGNED_SHORT:
	  drawSliceY<unsigned short>((unsigned short *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity);    
	  break;
  case Inrimage::WT_SIGNED_SHORT:
	  drawSliceY<short int>((short int *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity);   
	  break;
  case Inrimage::WT_UNSIGNED_INT:
	  drawSliceY<unsigned int>((unsigned int *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity); 
	  break;
  case Inrimage::WT_SIGNED_INT:
	  drawSliceY<int>((int *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity);
	  break;
  case Inrimage::WT_UNSIGNED_LONG:
	  drawSliceY<unsigned LONGINT>((unsigned LONGINT *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity); 
	  break;
  case Inrimage::WT_SIGNED_LONG:
	  drawSliceY<LONGINT>((LONGINT *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity); 
	  break;
  case Inrimage::WT_FLOAT:
	  drawSliceY<float>((float *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity);    
	  break;
  case Inrimage::WT_DOUBLE:
	  drawSliceY<double>((double *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity);  
	  break;
 case Inrimage::WT_FLOAT_VECTOR:
	 computeAndDrawVectorNormSliceY<float>((float *)0, inr,normVector, 
		  buffer,minWindow,maxWindow,nbRealRows,nbRealColumns,ncols,sliceNumber,
		  redColorMap,greenColorMap,blueColorMap,opacity);
	  break;
  case Inrimage::WT_DOUBLE_VECTOR:
	 computeAndDrawVectorNormSliceY<double>((double *)0, inr,normVector, 
		  buffer,minWindow,maxWindow,nbRealRows,nbRealColumns,ncols,sliceNumber,
		  redColorMap,greenColorMap,blueColorMap,opacity);
	  break;
  default:
	  {
		  double a= 1.0/(maxWindow-minWindow);
		  double val;
		  unsigned int i,j,k;
		  Voxel *vox=Voxel::voxelByType(inr->getType(),inr->getVdim());
		  unsigned int *b=buffer;
		  unsigned char *p,v;

		  j=sliceNumber;
		  for (k=0,b=buffer;k<nbRealRows;k++) {
			  b=&buffer[k*ncols];
			  for (i=0;i<nbRealColumns;i++,b++) {
				 inr->getVoxel(i,j,nbRealRows-1-k,*vox);
				  val=vox->getValueAsDouble();
				  p=(unsigned char *) b;
				  if (val<minWindow) {
					  p[0]=redColorMap[256];
					  p[1]=greenColorMap[256];
					  p[2]=blueColorMap[256];
				  }  
				  else if  (val>maxWindow){
					  p[0]=redColorMap[257];
					  p[1]=greenColorMap[257];
					  p[2]=blueColorMap[257];
				  } 
				  else {
					  v=(unsigned char )(255*(val-minWindow)*a);
					  p[0]=redColorMap[v];
					  p[1]=greenColorMap[v];
					  p[2]=blueColorMap[v];
				  }
			  }
		  }
	    delete vox;

	  }
  }
}
void Image3DTextureInformation::fillSliceXTexture(const Inrimage *inr,
												  const unsigned char *redColorMap,
												  const unsigned char *greenColorMap,
												  const unsigned char *blueColorMap,
												  const double minWindow,
												  const double maxWindow,
												  const double opacity)
{
  switch(inr->getType()) {
  case Inrimage::WT_UNSIGNED_CHAR:
	  drawSliceX<unsigned char>((unsigned char *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity);   
	  break;
  case Inrimage::WT_SIGNED_CHAR:
	  drawSliceX<char>((char *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity);    
	  break;
  case Inrimage::WT_UNSIGNED_SHORT:
	  drawSliceX<unsigned short>((unsigned short *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity);    
	  break;
  case Inrimage::WT_SIGNED_SHORT:
	  drawSliceX<short int>((short int *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity);   
	  break;
  case Inrimage::WT_UNSIGNED_INT:
	  drawSliceX<unsigned int>((unsigned int *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity); 
	  break;
  case Inrimage::WT_SIGNED_INT:
	  drawSliceX<int>((int *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity);
	  break;
  case Inrimage::WT_UNSIGNED_LONG:
	  drawSliceX<unsigned LONGINT>((unsigned LONGINT *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity); 
	  break;
  case Inrimage::WT_SIGNED_LONG:
	  drawSliceX<LONGINT>((LONGINT *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity); 
	  break;
  case Inrimage::WT_FLOAT:
	  drawSliceX<float>((float *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity);    
	  break;
  case Inrimage::WT_DOUBLE:
	  drawSliceX<double>((double *)0,inr,buffer,minWindow,maxWindow,nbRealRows,
		  nbRealColumns,ncols,sliceNumber,redColorMap,greenColorMap,blueColorMap,opacity);  
	  break;
 case Inrimage::WT_FLOAT_VECTOR:
	  computeAndDrawVectorNormSliceX<float>((float *)0, inr,normVector, 
		  buffer,minWindow,maxWindow,nbRealRows,nbRealColumns,ncols,sliceNumber,
		  redColorMap,greenColorMap,blueColorMap,opacity);
	  break;
  case Inrimage::WT_DOUBLE_VECTOR:
	  computeAndDrawVectorNormSliceX<double>((double *)0, inr,normVector, 
		  buffer,minWindow,maxWindow,nbRealRows,nbRealColumns,ncols,sliceNumber,
		  redColorMap,greenColorMap,blueColorMap,opacity);
	  break;
  default:
	  {
		  double a= 1.0/(maxWindow-minWindow);
		  double val;
		  unsigned int i,j,k;
		  Voxel *vox=Voxel::voxelByType(inr->getType(),inr->getVdim());
		  unsigned int *b=buffer;
		  unsigned char *p,v;

		  i=sliceNumber;
		  for (j=0;j<nbRealRows;j++) {
			  b=&buffer[j*ncols];
			  for (k=0;k<nbRealColumns;k++,b++) {
				  inr->getVoxel(i,j,k,*vox);
				  val=vox->getValueAsDouble();
				  p=(unsigned char *) b;
				  if (val<minWindow) {
					  p[0]=redColorMap[256];
					  p[1]=greenColorMap[256];
					  p[2]=blueColorMap[256];
				  }  
				  else if  (val>maxWindow){
					  p[0]=redColorMap[257];
					  p[1]=greenColorMap[257];
					  p[2]=blueColorMap[257];
				  } 
				  else {
					  v=(unsigned char )(255*(val-minWindow)*a);
					  p[0]=redColorMap[v];
					  p[1]=greenColorMap[v];
					  p[2]=blueColorMap[v];
				  }
			  }
		  }
		  delete vox;
	  }

  }
}
void Image3DTextureInformation::enableTexture(const Inrimage *inr,
											  const unsigned char *redColorMap,
											  const unsigned char *greenColorMap,
											  const unsigned char *blueColorMap,
											  const double minWindow,
											  const double maxWindow,
											  const double opacity) 
{
	glEnable(GL_TEXTURE_2D);

	if (binded==false) {
		if (id>0) 
			glDeleteTextures(1,&id);
		writeTexture(inr,redColorMap,greenColorMap,blueColorMap,minWindow,maxWindow,opacity);
		createTexture();
	}
	glBindTexture(GL_TEXTURE_2D, id);

}
void Image3DTextureInformation::writeTexture(const Inrimage *inr,
											 const unsigned char *redColorMap,
											 const unsigned char *greenColorMap,
											 const unsigned char *blueColorMap,
											 const double minWindow,
											 const double maxWindow,
											 const double opacity) 
{
	// must differentiate depending on slice orientation
	switch(background->getGeometry()) {
  case SLICE_Z:
	  fillSliceZTexture(inr,redColorMap,greenColorMap,blueColorMap,
		  minWindow,maxWindow,opacity);
	  break;
  case SLICE_X:
	  fillSliceXTexture(inr,redColorMap,greenColorMap,blueColorMap,
		  minWindow,maxWindow,opacity);	 
	  break;

  case SLICE_Y:
	  fillSliceYTexture(inr,redColorMap,greenColorMap,blueColorMap,
		  minWindow,maxWindow,opacity);
	  break;

  default:
	  break;
	}
}

void Image3DCameraBackground::imageHasChangedSize(Inrimage *)
{
	recomputeVectorDisplayList=true;
	mainTexture->emptyNormVector();
	mainTexture->init(getInrimage());
	mainTexture->binded=true;
	showAxis=false;
}
void Image3DCameraBackground::setSecondImage(Inrimage *inr){
	if (secondTexture)
		delete secondTexture;
	if (inr) {
		secondTexture= new Image3DTextureInformation(inr,this);
		switch (geometry) {
		case SLICE_Z:
			secondTexture->sliceRatio=((double)(inr->getVZ()))/((double)(getInrimage()->getVZ()));
			secondTexture->maxSlice=inr->getZ()-1;
			break;
		case SLICE_X:
			secondTexture->sliceRatio=((double)(inr->getVX()))/((double)(getInrimage()->getVX()));
			secondTexture->maxSlice=inr->getX()-1;
			break;
		case SLICE_Y:
			secondTexture->sliceRatio=((double)(inr->getVY()))/((double)(getInrimage()->getVY()));
			secondTexture->maxSlice=inr->getY()-1;
			break;
		default:
			break;
		}
		setSecondSliceNumber();
	} else {
		secondTexture=0;
	}
}

void Image3DCameraBackground::setOpacity()
{
	if (camera->getCameraActivation()) {
		camera->getCameraActivation()->redraw();
	}
}

/** warn the background that the image has changed it content (but not its size) */
void Image3DCameraBackground::imageHasChangedTexture(Inrimage *inr)
{
	mainTexture->emptyNormVector();
	recomputeVectorDisplayList=true;
	mainTexture->binded=false;

	if (camera->getCameraActivation()) {
		camera->getCameraActivation()->redraw();
	}
}


} // namespace yav
