#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/ImageGeometry.h"
#include <basics/Math.h>
#include <linearAlgebra/Vec2.h>
#include <inrimage/ImageGeometry.h>

namespace yav {

const double ImageGeometry::epsilon = 1e-6;

/**********************************************************************/
/**********************************************************************/
/***********************   CartesianGeometry   ************************/
/**********************************************************************/
/**********************************************************************/

bool CartesianGeometry::in(void) const {
  double _x,_y,_z;
  _x=x;
  _y=y;
  _z=z;
  if (inr.useTransformationMatrix()) {
    Vec3<double> translation=inr.getTranslationFromConstImage();
    Vec3<double> p=Vec3<double>(x,y,z);
    p-= translation;
    /// transform from world coordinates into image coordinates
    Rotation3D r=inr.getRotationMatrixFromConstImage();
    r.transpose();
    
    p=r*p;
    _x= p.getX();
    _y= p.getY();
    _z= p.getZ();
    
  }
  
  double _vx=_x/inr.getVX();
  double _vy=_y/inr.getVY();
  double _vz=_z/inr.getVZ();
  
  return (_vx >= (-0.5f)) && (_vx < ( inr.getX()-0.5f)) &&
    (_vy >= (-0.5f)) && (_vy < (inr.getY()-0.5f)) &&
    (_vz >= (-0.5f)) && (_vz < (inr.getZ()-0.5f));
}


void CartesianGeometry::set(double _x, double _y, double _z) {
  x = _x;
  y = _y;
  z = _z;

  if (inr.useTransformationMatrix()) {
     /// transform from world coordinates into image coordinates
    Vec3<double> p=Vec3<double>(x,y,z)-inr.getTranslationFromConstImage();
    Rotation3D r=inr.getRotationMatrixFromConstImage();
    r.transpose();
    p=r*p;
    _x=p.getX(); _y=p.getY(); _z=p.getZ();
  }

  ic[0] = signedFloor(_x / inr.getVX());
  ic[1] = signedFloor(_y / inr.getVY());
  ic[2] = signedFloor(_z / inr.getVZ());
}
void CartesianGeometry::setImage(double _x, double _y, double _z) {
  x = _x*inr.getVX();
  y = _y*inr.getVY();
  z = _z*inr.getVZ();

 if (inr.useTransformationMatrix()) {
   Vec3<double> p=inr.getTransformationMatrixFromConstImage()*Vec3<double>(x,y,z);
   x=p.getX(); y=p.getY(); z=p.getZ();
 }

  ic[0] = signedFloor(_x +0.5);
  ic[1] = signedFloor(_y +0.5);
  ic[2] = signedFloor(_z +0.5);
}

void CartesianGeometry::toImage(int *pos) const {
  pos[0] = ic[0];
  pos[1] = ic[1];
  pos[2] = ic[2];
}


double CartesianGeometry::interpolate(void) const {
  inr.getRealVoxel(x, y, z, *voxel);
  return voxel->getValueAsDouble();
}


double CartesianGeometry::nextVoxel(const Vec3<double>& pp) {
  double d, dlx,dly,dlz;
  Vec3<double> dp=pp;
  double _x,_y,_z;
  _x=x;
  _y=y;
  _z=z;
  
  if (inr.useTransformationMatrix()) {
	  Vec3<double> translation;
     /// transform from world coordinates into image coordinates
    Rotation3D r=inr.getRotationMatrixFromConstImage();
    r.transpose();
    dp=r*dp;
	translation=inr.getTranslationFromConstImage();
	_x-= translation.getX();
	_y-= translation.getY();
	_z-= translation.getZ();

  }


  if(dp.getX() >= 0.0) {
    dlx = ((ic[0] + 1) * inr.getVX() - _x) / dp.getX();
    if(dlx <= 0) dlx = ((ic[0] + 2.0) * inr.getVX() - _x) / dp.getX();
  }
  else {
    dlx = ((ic[0] ) * inr.getVX() - _x) / dp.getX();
    if(dlx <= 0) dlx = ((ic[0] - 1.0) * inr.getVX() - _x) / dp.getX();
  }

  if(dp.getY() >= 0.0) {
    dly = ((ic[1] + 1) * inr.getVY() - _y) / dp.getY();
    if(dly <= 0) dly = ((ic[1] + 2.0) * inr.getVY() - _y) / dp.getY();
  }
  else {
    dly = ((ic[1] ) * inr.getVY() - _y) / dp.getY();
    if(dly <= 0) dly = ((ic[1] - 1.0) * inr.getVY() - _y) / dp.getY();
  }


  if(dp.getZ() >= 0.0) {
    dlz = ((ic[2] + 1) * inr.getVZ() - _z) / dp.getZ();
    if(dlz <= 0) dlz = ((ic[2] + 2.0) * inr.getVZ() - _z) / dp.getZ();
  }
  else {
    dlz = ((ic[2] )* inr.getVZ() - _z) / dp.getZ();
    if(dlz <= 0) dlz = ((ic[2] - 1.0) * inr.getVZ() - _z) / dp.getZ();
  }
  if ((dlx<dly) && (dlx<dlz))
    {
      d=dlx;
      ic[0]+=signi(dp.getX());
    }
  else if ((dly<dlz) && (dly<dlx))
    {
      d=dly;
      ic[1]+=signi(dp.getY());
    }
  else
    {
      d=dlz;
      ic[2]+=signi(dp.getZ());
    }
  x+=pp.getX()*d;
  y+=pp.getY()*d;
  z+=pp.getZ()*d;



  return d;
}

double CartesianGeometry::nextBorderVoxel(const Vec3<double>& pp) {
  double dl,dly,dlz;
  Vec3<double> dp=pp;
  double _x,_y,_z;
  _x=x;
  _y=y;
  _z=z;
  
  if (inr.useTransformationMatrix()) {
	  Vec3<double> translation;
     /// transform from world coordinates into image coordinates
    Rotation3D r=inr.getRotationMatrixFromConstImage();
    r.transpose();
    dp=r*dp;
	translation=inr.getTranslationFromConstImage();
	_x-= translation.getX();
	_y-= translation.getY();
	_z-= translation.getZ();

  }

  dly=dlz=dl=0;
  if (_x< 0.5f*inr.getVX()) {
	  dl= -x/dp.getX();
  } 
  else if (_x> ((inr.getX()-1)*inr.getVX())) {
	  dl= ((inr.getX()-1)*inr.getVX()-x)/dp.getX();
  }
  if (_y< 0.5f*inr.getVY()) {
	  dly= -y/dp.getY();
  } 
  else if (_y> ((inr.getY()-1)*inr.getVY())) {
	  dly= ((inr.getY()-1)*inr.getVY()-y)/dp.getY();
  }
  if (_z< 0.5f*inr.getVZ()) {
	  dlz= -z/dp.getZ();
  } 
  else if (_z> ((inr.getZ()-1)*inr.getVZ())) {
	  dlz= ((inr.getZ()-1)*inr.getVZ()-z)/dp.getZ();
  }
  if (fabs(dly)>fabs(dl))
	  dl=dly;
  if (fabs(dlz)>fabs(dl))
	  dl=dlz;

  x+=pp.getX()*dl;
  y+=pp.getY()*dl;
  z+=pp.getZ()*dl;

  _x+=dp.getX()*dl;
  _y+=dp.getY()*dl;
  _z+=dp.getZ()*dl;

  ic[0] = signedFloor(_x / inr.getVX());
  ic[1] = signedFloor(_y / inr.getVY());
  ic[2] = signedFloor(_z / inr.getVZ());

  return dl;
}


/**********************************************************************/
/**********************************************************************/
/**********************   CylindricalGeometry   ***********************/
/**********************************************************************/
/**********************************************************************/

bool CylindricalGeometry::in(void) const {
  return (getRadius() < getX0() &&
	  getZ() >= 0.0 && getZ() < inr.getY() * inr.getVY());
}

void CylindricalGeometry::set(double xx, double yy, double zz) {
  double x=0.0,y=0.0,_z=0.0;
 if (inr.useTransformationMatrix()) {
     /// transform from world coordinates into image coordinates
    Vec3<double> p=Vec3<double>(xx,yy,zz)-inr.getTranslationFromConstImage();
    Rotation3D r=inr.getRotationMatrixFromConstImage();
    r.transpose();
    p=r*p;
    x=p.getX(); y=p.getY(); _z=p.getZ();
  }

  double a = x - x0;
  double b = y - x0;
  r = sqrt(a * a + b * b);
  theta = (x <= x0) ? acos((y - x0) / r) : (2*M_PI - acos((y - x0) / r));
  if(theta >= 2.0 * M_PI) theta -= 2.0 * M_PI;
  z = _z;
}

void CylindricalGeometry::setImage(double x, double y, double _z) {

  //  double dt = M_PI / (2.0 * inr.getZ());
  z=y*inr.getVY();
  theta= _z*M_PI/inr.getZ();
  //  if (theta < M_PI - dt || theta >= 2.0 * M_PI - dt )
    r= x- inr.getX() / 2.0 + 1.5;
    /*  else
	r= -(x- inr.getX() / 2.0);*/
  r*=inr.getVX();


}

void CylindricalGeometry::toCartesian(double *pos) const {
  pos[0] = x0 - r * sin(theta);
  pos[1] = x0 + r * cos(theta);
  pos[2] = z;

  if (inr.useTransformationMatrix()) {
    Vec3<double> p=inr.getTransformationMatrixFromConstImage()*Vec3<double>(pos[0],pos[1],pos[2]);
    pos[0]=p.getX();pos[1]=p.getY();pos[2]=p.getZ();
  }
  
}

void CylindricalGeometry::toImage(int *pos) const {
  double dt = M_PI / (2.0 * inr.getZ());
  pos[0] = (theta < M_PI - dt || theta >= 2.0 * M_PI - dt )
    ? (int) ceil(r / inr.getVX() + inr.getX() / 2.0 - 1.5)
    : (int) floor(inr.getX() / 2.0 - r / inr.getVX());
  pos[1] = (int) floor(z / inr.getVY());
  pos[2] = ((int) floor(theta * inr.getZ() / M_PI + 0.5)) % inr.getZ();
}

double CylindricalGeometry::interpolate(void) const {
  double V[8], n, alpha, r0, r1, theta0, theta1, z0, z1, res;
  int i;
  int X0, Y0, Z0, X1, Y1, Z1;

  if(r <= 0.5) {
    Y0 = (int) floor(z / inr.getVY() + 0.5);
    if(Y0 < 0) Y0 = 0;
    if(Y0 >= (int) inr.getY()) Y0 = inr.getY() - 1;
    inr.getVoxel((int) (x0 / inr.getVX() + 0.5), Y0, 0, *voxel);
    return voxel->getValueAsDouble();
  }

  alpha = M_PI / inr.getZ();
  
  // compute tri-linear weights
  r0 = (floor(r / inr.getVX() - 0.5) + 0.5) * inr.getVX();
  r1 = r0 + inr.getVX();
  theta0 = (floor(theta / alpha - 0.5) + 0.5) * alpha;
  theta1 = theta0 + alpha;
  z0 = (floor(z / inr.getVY() - 0.5) + 0.5) * inr.getVY();
  z1 = z0 + inr.getVY();

  V[0] = volume(r1, r, theta1 - theta, z1 - z);
  V[1] = volume(r, r0, theta1 - theta, z1 - z);
  V[2] = volume(r, r0, theta - theta0, z1 - z);
  V[3] = volume(r1, r, theta - theta0, z1 - z);
  V[4] = volume(r1, r, theta1 - theta, z - z0);
  V[5] = volume(r, r0, theta1 - theta, z - z0);
  V[6] = volume(r, r0, theta - theta0, z - z0);
  V[7] = volume(r1, r, theta - theta0, z - z0);

  for(i = 0, n = 0.0; i < 8; i++) n += V[i];

  // compute weighted greylevel value
  double dt = M_PI / (2.0 * inr.getZ());
  X0 = (theta0 < M_PI - dt || theta0 >= 2.0 * M_PI - dt)
    ? (int) ceil(r0 / inr.getVX() + inr.getX() / 2.0 - 1.5)
    : (int) floor(inr.getX() / 2.0 - r0 / inr.getVX());
  Y0 = (int) floor(z0 / inr.getVY());
  Z0 = ((int) floor(theta0 * inr.getZ() / M_PI + 0.5)) % inr.getZ();
  if(X0 < 0) X0 = 0;
  if(X0 >= (int) inr.getX()) X0 = inr.getX() - 1;
  if(Y0 < 0) Y0 = 0;
  if(Y0 >= (int) inr.getY()) Y0 = inr.getY() - 1;
  if(Z0 < 0) Z0 = 0;
  if(Z0 >= (int) inr.getZ()) Z0 = inr.getZ() - 1;
  X1 = (theta1 < M_PI - dt || theta1 >= 2.0 * M_PI - dt)
    ? (int) ceil(r1 / inr.getVX() + inr.getX() / 2.0 - 1.5)
    : (int) floor(inr.getX() / 2.0 - r1 / inr.getVX());
  Y1 = (int) floor(z1 / inr.getVY());
  Z1 = ((int) floor(theta1 * inr.getZ() / M_PI + 0.5)) % inr.getZ();
  if(X1 < 0) X1 = 0;
  if(X1 >= (int) inr.getX()) X1 = inr.getX() - 1;
  if(Y1 < 0) Y1 = 0;
  if(Y1 >= (int) inr.getY()) Y1 = inr.getY() - 1;
  if(Z1 < 0) Z1 = 0;
  if(Z1 >= (int) inr.getZ()) Z1 = inr.getZ() - 1;

  inr.getVoxel(X0, Y0, Z0, *voxel);
  res = V[0] * voxel->getValueAsDouble();
  inr.getVoxel(X1, Y0, Z0, *voxel);
  res += V[1] * voxel->getValueAsDouble();
  inr.getVoxel(X1, Y0, Z1, *voxel);
  res += V[2] * voxel->getValueAsDouble();
  inr.getVoxel(X0, Y0, Z1, *voxel);
  res += V[3] * voxel->getValueAsDouble();
  inr.getVoxel(X0, Y1, Z0, *voxel);
  res += V[4] * voxel->getValueAsDouble();
  inr.getVoxel(X1, Y1, Z0, *voxel);
  res += V[5] * voxel->getValueAsDouble();
  inr.getVoxel(X1, Y1, Z1, *voxel);
  res += V[6] * voxel->getValueAsDouble();
  inr.getVoxel(X0, Y1, Z1, *voxel);
  res += V[7] * voxel->getValueAsDouble();
  
  return res / n;
}

static double computeAngle(const Vec2<double>& v) {
  if(v.getY() > 0.0)
    return (v.getX() <= 0.0)
      ? atan(-v.getX() / v.getY())
      : 2.0 * M_PI + atan(-v.getX() / v.getY());
  else if(v.getY() < 0.0)
    return atan(-v.getX() / v.getY()) + M_PI;
  else {
    if(v.getX() >= 0.0) return 3.0 * M_PI / 2.0;
    else return M_PI / 2.0;
  }
}

/** check whether angle is in between a and b
    @param angle considered angle in [0, 2 PI[
    @param a lower bound in [0, 2 PI[
    @param b upper bound in [0, 2 PI[
    @return true if a <= angle < b, false otherwise */
static bool isAngleBetween(double angle, double a, double b) {
  return (b >= a)
    ? (angle >= a && angle < b)
    : ((angle >= a && angle < 2 * M_PI) || (angle < b && angle >= 0.0));
}

double CylindricalGeometry::nextVoxel(const Vec3<double>& direction) {
  double a, az, art, theta[4], r0, r1, alpha, theta0, theta1, dtheta, z0, z1;
  Vec2<double> dp(direction.getX(), direction.getY());
  double dz = direction.getZ();

  alpha = M_PI / inr.getZ();
  dtheta = computeAngle(dp);
  r0 = floor(getRadius() / inr.getVX()) * inr.getVX();
  if(r0 == getRadius()) r0 -= epsilon;
  r1 = r0 + inr.getVX();
  theta0 = floor(getAngle() / alpha) * alpha;
  if(theta0 == getAngle()) theta0 -= epsilon;
  theta1 = theta0 + alpha;
  z0 = floor(getZ() / inr.getVY()) * inr.getVY();
  if(z0 == getZ()) z0 -= epsilon;
  z1 = z0 + inr.getVY();
  Vec2<double> u(-sin(getAngle()), cos(getAngle()));
  Vec2<double> u0(-sin(theta0), cos(theta0));
  Vec2<double> u1(-sin(theta1), cos(theta1));
  
  if(dz >= 0.0) az = (z1 - getZ()) / dz;
  else az = (z0 - getZ()) / dz;
  
  Vec2<double> v = (r1 * u1) - (getRadius() * u);
  theta[0] = computeAngle(v);
  
  v = (r0 * u1) - (getRadius() * u);
  theta[1] = computeAngle(v);
  
  if(isAngleBetween(dtheta, theta[0], theta[1])) {
    art = getRadius()
      * (u.getY() * u1.getX() - u.getX() * u1.getY())
      / (dp.getX() * u1.getY() - dp.getY() * u1.getX());
  }
  else {
    v = (r0 * u0) - (getRadius() * u);
    theta[2] = computeAngle(v);
    
    if(isAngleBetween(dtheta, theta[1], theta[2])) {
      double dt = dot<double>(dp, u), n2 = dp.norm2(),
	rc = sqrt(getRadius() * getRadius() * dt * dt -
		  n2 * (getRadius() * getRadius()
			- r0 * r0)), s1, s2;
      s1 = (-getRadius() * dt + rc) / n2;
      s2 = (-getRadius() * dt - rc) / n2;
      if(s1 >= 0.0) {
	if(s2 >= 0.0) art = (s2 > s1) ? s1 : s2;
	else art = s1;
      }
      else art = s2;
    }
    
    else {
      v = (r1 * u0) - (getRadius() * u);
      theta[3] = computeAngle(v);
      
      if(isAngleBetween(dtheta, theta[2], theta[3])) {
	art = getRadius() * (u.getY() * u0.getX()
				  - u.getX() * u0.getY())
	  / (dp.getX() * u0.getY() - dp.getY() * u0.getX());
      }
      
      else {
	double dt = dot<double>(dp, u), n2 = dp.norm2(),
	  rc = sqrt(getRadius() * getRadius() * dt * dt -
		    n2 * (getRadius() * getRadius()
			  - r1 * r1)), s1, s2;
	s1 = (-getRadius() * dt + rc) / n2;
	s2 = (-getRadius() * dt - rc) / n2;
	if(s1 >= 0.0) {
	  if(s2 >= 0.0) art = (s2 > s1) ? s1 : s2;
	  else art = s1;
	}
	else art = s2;
      }
    }
  }
  
  if(az < art) a = az;
  else a = art;
  double p[3];
  toCartesian(p);
  Vec3<double> q(p);
  q+= (a+epsilon)*direction;
  set(q.getX(),q.getY(),q.getZ());
  return a + epsilon;
}

double CylindricalGeometry::volume(double ra, double rb,
				   double dtheta, double dz) const {
  return (ra * ra - rb * rb) * dtheta * dz / 2.0;
}

double CylindricalGeometry::nextBorderVoxel(const Vec3<double>& pp) {
	CERR<< "nextBorderVoxel(const Vec3<double>& pp) : this function is not supported in Cylindrical Geometry"<<ENDL;
	return 0.0;
}

/**********************************************************************/
/**********************************************************************/
/************************  SphericalGeometry   ************************/
/**********************************************************************/
/**********************************************************************/

bool SphericalGeometry::in(void) const {
  return /*(getRadius() < inr.getX() * inr.getVX() &&
	   0.0 <= getAlpha() <= inr.getAlpha0 )*/ 0.0;
}

void SphericalGeometry::set(double , double , double ) {
  /*
  double a = x - x0;
  double b = y - x0;
  double c = r0 - z;
  r = sqrt(a*a + b*b + c*c);
  alpha = atan2(sqrt(a*a + b*b),c);
  theta = (x <= x0) ? atan2(-a,b) : atan2(a,b)+M_PI;
  */
}
void SphericalGeometry::setImage(double , double , double ) {
  /*
  double a = x - x0;
  double b = y - x0;
  double c = r0 - z;
  r = sqrt(a*a + b*b + c*c);
  alpha = atan2(sqrt(a*a + b*b),c);
  theta = (x <= x0) ? atan2(-a,b) : atan2(a,b)+M_PI;
  */
}

void SphericalGeometry::toCartesian(double *) const {
  /*
  pos[0] = x0 - r * sin(alpha0/2 - alpha) * sin(theta);
  pos[1] = x0 + r * sin(alpha0/2 - alpha) * cos(theta);
  pos[2] = r0 - r * cos(alpha0/2 - alpha);
  */
}

void SphericalGeometry::toImage(int *) const {
  /*
  double da = inr.getAlpha0 / (inr.getY()-1);
  double dt = M_PI / inr.getZ();
  pos[0] = floor(r/inr.getVX()+0.5);
  pos[1] = floor(alpha/da+0.5);
  pos[2] = floor(theta/dt+0.5);
  */
}

double SphericalGeometry::interpolate(void) const {
  return 0.0;
}

double SphericalGeometry::nextVoxel(const Vec3<double>& ) {
  return 0.0;
}
double SphericalGeometry::nextBorderVoxel(const Vec3<double>& pp) {
	CERR<< "nextBorderVoxel(const Vec3<double>& pp) : this function is not supported in Cylindrical Geometry"<<ENDL;
		return 0.0;
}

} // namespace yav
