/***************************************************************************
 ***************************************************************************/

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#include "inrimage/InrChamferMap.h"
#include "ChamferMap.h"
#include <limits.h>
#include <float.h>
#include <stdlib.h>
#include <math.h>

using namespace yav;

InrChamferMap::InrChamferMap(Inrimage * src,
			     int location, int maskSize, int maskDim)
  : Inrimage(*src)
{
  int i;
  int dimT = xdim * ydim * zdim;
  int dims[3];
  dims[0] = xdim;
  dims[1] = ydim;
  dims[2] = zdim;

  float voxelSize = this->vx;
  float dy = this->vy / voxelSize;
  float dz = this->vz / voxelSize;

  isExact = false;

  //  union {
    ChamferMap<unsigned short int> * ushortChamferMap = 0;
    ChamferMap<short int>          *  shortChamferMap = 0;
    ChamferMap<unsigned int>       *   uintChamferMap = 0;
    ChamferMap<int>                *    intChamferMap = 0;
    ChamferMap<unsigned LONGINT>   *  ulongChamferMap = 0;
    ChamferMap<LONGINT>            *   longChamferMap = 0;
    ChamferMap<float>              *  floatChamferMap = 0;
    ChamferMap<double>             * doubleChamferMap = 0;
    //  };

  /*-- step I : needed conversions --*/
  switch(location)
    {
    default :
      printf("Bad chamfer map location, default location : foreground\n");
    case 1 : // foreground
      switch(type)
	{
	case WT_UNSIGNED_CHAR :
	  {
	    setType(WT_UNSIGNED_SHORT);
	    unsigned short int * pixel  = (unsigned short int *) data;
	    unsigned char      * srcPix = (unsigned char *)      src->getData();
	    for (i = 0; i < dimT; i++)
	      pixel[i] = (unsigned short int) srcPix[i];
	  }
	  break;
	case WT_SIGNED_CHAR :
	  {
	    setType(WT_SIGNED_SHORT);
	    short int * pixel  = (short int *) data;
	    char      * srcPix = (char *)      src->getData();
	    for (i = 0; i < dimT; i++)
	      pixel[i] = (short int) srcPix[i];
	  }
	  break;
	default :
	  break;
	} // ens switch(type)
      break;
    case 0 : // background
    case 2 : // both foreground and background
      switch(type)
	{
	case WT_UNSIGNED_CHAR :
	  {
	    setType(WT_SIGNED_SHORT);
	    short int      * pixel  = (short int *)     data;
	    unsigned char  * srcPix = (unsigned char *) src->getData();
	    for (i = 0; i < dimT; i++)
	      pixel[i] = (short int) srcPix[i];
	  }
	  break;
	case WT_SIGNED_CHAR :
	  {
	    setType(WT_SIGNED_SHORT);
	    short int * pixel  = (short int *) data;
	    char      * srcPix = (char *)      src->getData();
	    for (i = 0; i < dimT; i++)
	      pixel[i] = (short int) srcPix[i];
	  }
	  break;
	case WT_UNSIGNED_SHORT :
	  {
	    setType(WT_SIGNED_INT);
	    int * pixel = (int *) data;
	    unsigned short int * srcPix = (unsigned short int *) src->getData();
	    for (i = 0; i < dimT; i++)
	      pixel[i] = (int) srcPix[i];
	  }
	  break;
	case WT_UNSIGNED_INT :
	  {
	  setType(WT_SIGNED_LONG);
	  LONGINT * pixel = (LONGINT *) data;
	  unsigned int * srcPix = (unsigned int *) src->getData();
	  for (i = 0; i < dimT; i++)
	    pixel[i] = (LONGINT) srcPix[i];
	  }
	  break;
	  
	default :
	  break;

	} // end switch(type)
      break;
    } // end switch(location)


  /*-- Step II ChamferMap creation and calculation --*/
  
      switch(type)
	{
	case WT_UNSIGNED_SHORT :
//	  printf("image type : unsigned short int\n");
	  chamferType = WT_UNSIGNED_SHORT;
	  if ((fabs(dy - 1.0) < 0.000001) && (fabs(dz - 1.0) < 0.000001)) {
//	    printf("Masque isotrope, maskSize : %i, maskDim : %i, voxelSize : %f\n", maskSize, maskDim, voxelSize);
	    ushortChamferMap = new ChamferMap<unsigned short int>((unsigned short int) USHRT_MAX, (int) maskSize, (int) maskDim, voxelSize);}
	  else {
//	    printf("Masque non isotrope, dy : %f, dz : %f, voxelSize : %f\n", dy, dz, voxelSize);
	    ushortChamferMap = new ChamferMap<unsigned short int>((unsigned short int) USHRT_MAX, (float) dy, (float) dz, voxelSize);}
	  
	  switch(location) 
	    {
	    case 0 : // background
	    case 2 : // both foreground and background
	    default :
	      printf("Bad chamfer map location, default location : foreground\n");
	    case 1 : // foreground
	      ushortChamferMap->computeChamferMap((unsigned short int *) data, dims);
	      break;
	    }
	      
	  break;
	
	case WT_SIGNED_SHORT :
	  chamferType = WT_SIGNED_SHORT;
	  if ((fabs(dy - 1.0) < 0.000001) && (fabs(dz - 1.0) < 0.000001)) 
	    shortChamferMap = new ChamferMap<short int>((short int) SHRT_MAX, (int) maskSize, (int) maskDim, voxelSize);
	  else 
	    shortChamferMap = new ChamferMap<short int>((short int) SHRT_MAX, (float) dy, (float) dz, voxelSize);

	  switch(location)
	    {
	    case 0 : // background
	      shortChamferMap->computeChamferMapOut((short int *) data, dims);
	      break;
	    case 2 : // both foreground and background
	      shortChamferMap->computeChamferMapInOut((short int *) data, dims);
	      break;
	    default :
	      printf("Bad chamfer map location, default location : foreground\n");
	    case 1 : // foreground
	      shortChamferMap->computeChamferMap((short int *) data, dims);
	      break;
	    }
	  
	  break;

	case WT_UNSIGNED_INT :
	  chamferType = WT_UNSIGNED_INT;
	  if ((fabs(dy - 1.0) < 0.000001) && (fabs(dz - 1.0) < 0.000001)) 
	    uintChamferMap = new ChamferMap<unsigned int>((unsigned int) UINT_MAX, (int) maskSize, (int) maskDim, voxelSize);
	  else
	    uintChamferMap = new ChamferMap<unsigned int>((unsigned int) UINT_MAX, (float) dy, (float) dz, voxelSize);

	  switch(location)
	    {
	    case 0 : // background
	    case 2 : // both foreground and background
	    default :
	      printf("Bad chamfer map location, default location : foreground\n");
	    case 1 :// foreground
		uintChamferMap->computeChamferMap((unsigned int *) data, dims);
	      break;
	    }

	  break;
	  
	case WT_SIGNED_INT :
	  chamferType = WT_SIGNED_INT;
	  if ((fabs(dy - 1.0) < 0.000001) && (fabs(dz - 1.0) < 0.000001)) 
	    intChamferMap = new ChamferMap<int>((int) INT_MAX, (int) maskSize, (int) maskDim, voxelSize);
	  else
	    intChamferMap = new ChamferMap<int>((int) INT_MAX, (float) dy, (float) dz, voxelSize);

	  switch(location)
	    {
	    case 0 : // background
	      intChamferMap->computeChamferMapOut((int *) data, dims);
	      break;
	    case 2 : // both foreground and background
	      intChamferMap->computeChamferMapInOut((int *) data, dims);
	      break;
	    default :
	      printf("Bad chamfer map location, default location : foreground\n");
	    case 1 : // foreground
	      intChamferMap->computeChamferMap((int *) data, dims);
	      break;
	    }
	  break;

	case WT_UNSIGNED_LONG :
	  chamferType = WT_UNSIGNED_LONG;
	  if ((fabs(dy - 1.0) < 0.000001) && (fabs(dz - 1.0) < 0.000001)) 
	    ulongChamferMap = new ChamferMap<unsigned LONGINT>((unsigned LONGINT) ULONG_MAX, (int) maskSize, (int) maskDim, voxelSize);
	  else
	    ulongChamferMap = new ChamferMap<unsigned LONGINT>((unsigned LONGINT) ULONG_MAX, (float) dy, (float) dz, voxelSize);

	  switch(location)
	    {
	    case 0 : // background
	    case 2 : // both foreground and background
	    default :
	      printf("Bad chamfer map location, default location : foreground\n");
	    case 1 : // foreground
	      ulongChamferMap->computeChamferMap((unsigned LONGINT *) data, dims);
	      break;
	    }
	  break;
	  
	case WT_SIGNED_LONG :
	  chamferType = WT_SIGNED_LONG;
	  if ((fabs(dy - 1.0) < 0.000001) && (fabs(dz - 1.0) < 0.000001)) 
	    longChamferMap = new ChamferMap<LONGINT>((LONGINT) LONG_MAX, (int) maskSize, (int) maskDim, voxelSize);
	  else
	    longChamferMap = new ChamferMap<LONGINT>((LONGINT) LONG_MAX, (float) dy, (float) dz, voxelSize);

	  switch(location)
	    {
	    case 0 : // background
	      longChamferMap->computeChamferMapOut((LONGINT *) data, dims);
	      break;
	    case 2 : // both foreground and background
	      longChamferMap->computeChamferMapInOut((LONGINT *) data, dims);
	      break;
	    default :
	      printf("Bad chamfer map location, default location : foreground\n");
	    case 1 : // foreground
	      longChamferMap->computeChamferMap((LONGINT *) data, dims);
	      break;
	    }
	  break;
	  
	case WT_FLOAT :
	  chamferType = WT_FLOAT;
	  if ((fabs(dy - 1.0) < 0.000001) && (fabs(dz - 1.0) < 0.000001)) 
	    floatChamferMap = new ChamferMap<float>((float) FLT_MAX, (int) maskSize, (int) maskDim, voxelSize);
	  else
	    floatChamferMap = new ChamferMap<float>((float) FLT_MAX, (float) dy, (float) dz, voxelSize);
	  
	  switch(location)
	    {
	    case 0 : // background
	      floatChamferMap->computeChamferMapOut((float *) data, dims);
	      break;
	    case 2 : // both foreground and background
	      floatChamferMap->computeChamferMapInOut((float *) data, dims);
	      break;
	    default :
	      printf("Bad chamfer map location, default location : foreground\n");
	    case 1 : // foreground
	      floatChamferMap->computeChamferMap((float *) data, dims);
	      break;
	    }
	  break;

	case WT_DOUBLE :
	  chamferType = WT_DOUBLE;
	  if ((fabs(dy - 1.0) < 0.000001) && (fabs(dz - 1.0) < 0.000001)) 
	    doubleChamferMap = new ChamferMap<double>((double) DBL_MAX, (int) maskSize, (int) maskDim, voxelSize);
	  else
	    doubleChamferMap = new ChamferMap<double>((double) DBL_MAX, (float) dy, (float) dz, voxelSize);

	  switch(location)
	    {
	    case 0 : // background
	      doubleChamferMap->computeChamferMapOut((double *) data, dims);
	      break;
	    case 2 : // both foreground and background
	      doubleChamferMap->computeChamferMapInOut((double *) data, dims);
	      break;
	    default :
	      printf("Bad chamfer map location, default location : foreground\n");
	    case 1 : // foreground
	      doubleChamferMap->computeChamferMap((double *) data, dims);
	      break;
	    }

	  break;

	default :
	  printf("Cannot handle this type of data in InrChamferMap...\n");
	  break;
	  
	} // end switch(type)


      switch(chamferType)
	{
	  
	case WT_UNSIGNED_SHORT :
	  myChamferMap =  (ChamferMap<unsigned short int> *) ushortChamferMap;
	  break;
	case WT_SIGNED_SHORT :
	  myChamferMap = (ChamferMap<short int> *) shortChamferMap;
	  break;
	case WT_UNSIGNED_INT :
	  myChamferMap = (ChamferMap<unsigned int> *) uintChamferMap;
	  break;
	case WT_SIGNED_INT :
	  myChamferMap = (ChamferMap<int> *) intChamferMap;
	  break;
	case WT_UNSIGNED_LONG :
	  myChamferMap = (ChamferMap<unsigned LONGINT> *) ulongChamferMap;
	  break;
	case WT_SIGNED_LONG :
	  myChamferMap = (ChamferMap<LONGINT> *)  longChamferMap;
	  break;
	case WT_FLOAT :
	  myChamferMap = (ChamferMap<float> *) floatChamferMap;
	  break;
	case WT_DOUBLE :
	  myChamferMap = (ChamferMap<double> *) doubleChamferMap;
	  break;
	default :
	  myChamferMap = 0;
	  break;
	}
}


InrChamferMap::InrChamferMap(Inrimage * src,
			     Inrimage * output,
			     int flag,
			     int location, int maskSize, int maskDim)
  : Inrimage(*src)
{
  int i;
  int dimT = xdim * ydim * zdim;
  int dims[3];
  dims[0] = xdim;
  dims[1] = ydim;
  dims[2] = zdim;

  float voxelSize = this->vx;
  float dy = this->vy / voxelSize;
  float dz = this->vz / voxelSize;

  isExact = false;

  //  union {
    ChamferMap<unsigned short int> * ushortChamferMap = 0;
    ChamferMap<short int>          *  shortChamferMap = 0;
    ChamferMap<unsigned int>       *   uintChamferMap = 0;
    ChamferMap<int>                *    intChamferMap = 0;
    ChamferMap<unsigned LONGINT>   *  ulongChamferMap = 0;
    ChamferMap<LONGINT>            *   longChamferMap = 0;
    ChamferMap<float>              *  floatChamferMap = 0;
    ChamferMap<double>             * doubleChamferMap = 0;
    //  };

  switch(flag)
    {
    case POSITION_PROPAGATION :
      {
	/*-- step I : needed conversions and output initialisaion --*/
	setType(WT_SIGNED_INT);
	int * pixel = (int *) data;
	switch(src->getType())
	{
	case WT_SIGNED_CHAR :
	  {
	    char * srcPix = (char *) src->getData();
	    for (i = 0; i < dimT; i++)
	      pixel[i] = (int) srcPix[i];
	  }
	  break;
	case WT_UNSIGNED_CHAR :
	  {
	    unsigned char * srcPix = (unsigned char *) src->getData();
	    for (i = 0; i < dimT; i++)
	      pixel[i] = (int) srcPix[i];
	  }
	  break;
	case WT_SIGNED_SHORT :
	  {
	    short int * srcPix = (short int *) src->getData();
	    for (i = 0; i < dimT; i++)
	      pixel[i] = (int) srcPix[i];
	  }
	  break;
	case WT_UNSIGNED_SHORT :
	  {
	    unsigned short int * srcPix = (unsigned short int *) src->getData();
	    for (i = 0 ; i < dimT; i++)
	      pixel[i] = srcPix[i];
	  }
	  break;
	case WT_SIGNED_INT :
	  break;
	default :
	  printf("Cannot handle this type of data in InrChamferMap...\n");
	  break;
	} // end switch(type)

      /*-- Step II ChamferMap creation and calculation --*/
      switch(type)
	{
	case WT_SIGNED_INT :
	  chamferType = WT_SIGNED_INT;
	  output->setType(WT_SIGNED_INT);
	  if ((fabs(dy - 1.0) < 0.000001) && (fabs(dz - 1.0) < 0.000001)) {
	    printf("Masque Opt isotrope, maskSize : %i, maskDim : %i, voxelSize : %f\n", maskSize, maskDim, voxelSize);
	    intChamferMap = (ChamferMap<int> *) new ChamferMap<int>((int) INT_MAX, flag, (int) maskSize, (int) maskDim, voxelSize);
	  }
	  else {
	    printf("Masque Opt non isotrope, dy : %f, dz : %f, voxelSize : %f\n", dy, dz, voxelSize);
	    intChamferMap = (ChamferMap<int> *) new ChamferMap<int>((int) INT_MAX, flag, (float) dy, (float) dz, voxelSize);
	  }
	  switch(location)
	    {
	    case 0 : // background
	      intChamferMap->computeChamferMapOut((int *) data, (int *) output->getData(), dims);
	      break;
	    case 2 : // both foreground and background
	      intChamferMap->computeChamferMapInOut((int *) data, (int *) output->getData(), dims);
	      break;
	    default :
	      printf("Bad chamfer map location, default location : foreground\n");
	    case 1 : // foreground
	      intChamferMap->computeChamferMap((int *) data, (int *) output->getData(), dims);
	      break;
	    }
	  break;
	default :
	  printf("Cannot handle this type of data in InrChamferMap...\n");
	  break;
	} // end switch(type)
      } // end case POSITION_PROPAGATION
      break;

    case COLOR_PROPAGATION :
      {
	switch(location)
	  {
	  default :
	    printf("Bad chamfer map location, default location : foreground\n");
	  case 1 : // foreground
	    switch(type)
	      {
	      case WT_UNSIGNED_CHAR :
		{
		  setType(WT_UNSIGNED_SHORT);
		  unsigned short int * pixel  = (unsigned short int *) data;
		  unsigned char      * srcPix = (unsigned char *)      src->getData();
		  for (i = 0; i < dimT; i++)
		    pixel[i] = (unsigned short int) srcPix[i];
		}
		break;
	      case WT_SIGNED_CHAR :
		{
		  setType(WT_SIGNED_SHORT);
		  short int * pixel  = (short int *) data;
		  char      * srcPix = (char *)      src->getData();
		  for (i = 0; i < dimT; i++)
		    pixel[i] = (short int) srcPix[i];
		}
		break;
	      default :
		break;
	      } // ens switch(type)
	    break;
	  case 0 : // background
	  case 2 : // both foreground and background
	    switch(type)
	      {
	      case WT_UNSIGNED_CHAR :
		{
		  setType(WT_SIGNED_SHORT);
		  short int      * pixel  = (short int *)     data;
		  unsigned char  * srcPix = (unsigned char *) src->getData();
		  for (i = 0; i < dimT; i++)
		    pixel[i] = (short int) srcPix[i];
		}
		break;
	      case WT_SIGNED_CHAR :
		{
		  setType(WT_SIGNED_SHORT);
		  short int * pixel  = (short int *) data;
		  char      * srcPix = (char *)      src->getData();
		  for (i = 0; i < dimT; i++)
		    pixel[i] = (short int) srcPix[i];
		}
		break;
	      case WT_UNSIGNED_SHORT :
		{
		  setType(WT_SIGNED_INT);
		  int * pixel = (int *) data;
		  unsigned short int * srcPix = (unsigned short int *) src->getData();
		  for (i = 0; i < dimT; i++)
		    pixel[i] = (int) srcPix[i];
		}
		break;
	      case WT_UNSIGNED_INT :
		{
		  setType(WT_SIGNED_LONG);
		  LONGINT * pixel = (LONGINT *) data;
		  unsigned int * srcPix = (unsigned int *) src->getData();
		  for (i = 0; i < dimT; i++)
	    pixel[i] = (LONGINT) srcPix[i];
		}
		break;
		
	      default :
		break;
		
	      } // end switch(type)
	    break;
	  } // end switch(location)
	
	
	/*-- Step II ChamferMap creation and calculation --*/
	
	switch(type)
	  {
	  case WT_UNSIGNED_SHORT :
	    printf("image type : unsigned short int\n");
	    chamferType = WT_UNSIGNED_SHORT;
	    output->setType(WT_UNSIGNED_SHORT);
	    if ((fabs(dy - 1.0) < 0.000001) && (fabs(dz - 1.0) < 0.000001)) {
	      printf("Masque isotrope, maskSize : %i, maskDim : %i, voxelSize : %f\n", maskSize, maskDim, voxelSize);
	      ushortChamferMap = new ChamferMap<unsigned short int>((unsigned short int) USHRT_MAX, flag, (int) maskSize, (int) maskDim, voxelSize);}
	    else {
	      printf("Masque non isotrope, dy : %f, dz : %f, voxelSize : %f\n", dy, dz, voxelSize);
	      ushortChamferMap = new ChamferMap<unsigned short int>((unsigned short int) USHRT_MAX, flag, (float) dy, (float) dz, voxelSize);}
	    
	    switch(location) 
	      {
	      case 0 : // background
	      case 2 : // both foreground and background
	      default :
		printf("Bad chamfer map location, default location : foreground\n");
	      case 1 : // foreground
		ushortChamferMap->computeChamferMap((unsigned short int *) data, 
						    (unsigned short int *) output->getData(), dims);
		break;
	      }
	    
	    break;
	    
	  case WT_SIGNED_SHORT :
	    chamferType = WT_SIGNED_SHORT;
	    output->setType(WT_SIGNED_SHORT);
	  if ((fabs(dy - 1.0) < 0.000001) && (fabs(dz - 1.0) < 0.000001)) 
	    shortChamferMap = new ChamferMap<short int>((short int) SHRT_MAX, flag, (int) maskSize, (int) maskDim, voxelSize);
	  else 
	    shortChamferMap = new ChamferMap<short int>((short int) SHRT_MAX, flag, (float) dy, (float) dz, voxelSize);
	  
	  switch(location)
	    {
	    case 0 : // background
	      shortChamferMap->computeChamferMapOut((short int *) data, 
						    (short int *) output->getData(), dims);
	      break;
	    case 2 : // both foreground and background
	      shortChamferMap->computeChamferMapInOut((short int *) data, 
						      (short int *) output->getData(), dims);
	      break;
	    default :
	      printf("Bad chamfer map location, default location : foreground\n");
	    case 1 : // foreground
	      shortChamferMap->computeChamferMap((short int *) data, 
						 (short int *) output->getData(), dims);
	      break;
	    }
	  
	  break;

	case WT_UNSIGNED_INT :
	  chamferType = WT_UNSIGNED_INT;
	  output->setType(WT_UNSIGNED_INT);
	  if ((fabs(dy - 1.0) < 0.000001) && (fabs(dz - 1.0) < 0.000001)) 
	    uintChamferMap = new ChamferMap<unsigned int>((unsigned int) UINT_MAX, flag, (int) maskSize, (int) maskDim, voxelSize);
	  else
	    uintChamferMap = new ChamferMap<unsigned int>((unsigned int) UINT_MAX, flag, (float) dy, (float) dz, voxelSize);

	  switch(location)
	    {
	    case 0 : // background
	    case 2 : // both foreground and background
	    default :
	      printf("Bad chamfer map location, default location : foreground\n");
	    case 1 :// foreground
		uintChamferMap->computeChamferMap((unsigned int *) data, 
						  (unsigned int *) output->getData(), dims);
	      break;
	    }

	  break;
	  
	case WT_SIGNED_INT :
	  chamferType = WT_SIGNED_INT;
	  output->setType(WT_SIGNED_INT);
	  if ((fabs(dy - 1.0) < 0.000001) && (fabs(dz - 1.0) < 0.000001)) 
	    intChamferMap = new ChamferMap<int>((int) INT_MAX, flag, (int) maskSize, (int) maskDim, voxelSize);
	  else
	    intChamferMap = new ChamferMap<int>((int) INT_MAX, flag, (float) dy, (float) dz, voxelSize);

	  switch(location)
	    {
	    case 0 : // background
	      intChamferMap->computeChamferMapOut((int *) data, 
						  (int *) output->getData(), dims);
	      break;
	    case 2 : // both foreground and background
	      intChamferMap->computeChamferMapInOut((int *) data, 
						    (int *) output->getData(), dims);
	      break;
	    default :
	      printf("Bad chamfer map location, default location : foreground\n");
	    case 1 : // foreground
	      intChamferMap->computeChamferMap((int *) data, 
					       (int *) output->getData(), dims);
	      break;
	    }
	  break;

	case WT_UNSIGNED_LONG :
	  chamferType = WT_UNSIGNED_LONG;
	  output->setType(WT_UNSIGNED_LONG);
	  if ((fabs(dy - 1.0) < 0.000001) && (fabs(dz - 1.0) < 0.000001)) 
	    ulongChamferMap = new ChamferMap<unsigned LONGINT>((unsigned LONGINT) ULONG_MAX, flag, (int) maskSize, (int) maskDim, voxelSize);
	  else
	    ulongChamferMap = new ChamferMap<unsigned LONGINT>((unsigned LONGINT) ULONG_MAX, flag, (float) dy, (float) dz, voxelSize);

	  switch(location)
	    {
	    case 0 : // background
	    case 2 : // both foreground and background
	    default :
	      printf("Bad chamfer map location, default location : foreground\n");
	    case 1 : // foreground
	      ulongChamferMap->computeChamferMap((unsigned LONGINT *) data,
						 (unsigned LONGINT *) output->getData(), dims);
	      break;
	    }
	  break;
	  
	case WT_SIGNED_LONG :
	  chamferType = WT_SIGNED_LONG;
	  output->setType(WT_SIGNED_LONG);
	  if ((fabs(dy - 1.0) < 0.000001) && (fabs(dz - 1.0) < 0.000001)) 
	    longChamferMap = new ChamferMap<LONGINT>((LONGINT) LONG_MAX, flag, (int) maskSize, (int) maskDim, voxelSize);
	  else
	    longChamferMap = new ChamferMap<LONGINT>((LONGINT) LONG_MAX, flag, (float) dy, (float) dz, voxelSize);

	  switch(location)
	    {
	    case 0 : // background
	      longChamferMap->computeChamferMapOut((LONGINT *) data, 
						   (LONGINT *) output->getData(), dims);
	      break;
	    case 2 : // both foreground and background
	      longChamferMap->computeChamferMapInOut((LONGINT *) data, 
						     (LONGINT *) output->getData(), dims);
	      break;
	    default :
	      printf("Bad chamfer map location, default location : foreground\n");
	    case 1 : // foreground
	      longChamferMap->computeChamferMap((LONGINT *) data, 
						(LONGINT *) output->getData(), dims);
	      break;
	    }
	  break;
	  
	case WT_FLOAT :
	  chamferType = WT_FLOAT;
	  output->setType(WT_FLOAT);
	  if ((fabs(dy - 1.0) < 0.000001) && (fabs(dz - 1.0) < 0.000001)) 
	    floatChamferMap = new ChamferMap<float>((float) FLT_MAX, flag, (int) maskSize, (int) maskDim, voxelSize);
	  else
	    floatChamferMap = new ChamferMap<float>((float) FLT_MAX, flag, (float) dy, (float) dz, voxelSize);
	  
	  switch(location)
	    {
	    case 0 : // background
	      floatChamferMap->computeChamferMapOut((float *) data, 
						    (float *) output->getData(), dims);
	      break;
	    case 2 : // both foreground and background
	      floatChamferMap->computeChamferMapInOut((float *) data, 
						      (float *) output->getData(), dims);
	      break;
	    default :
	      printf("Bad chamfer map location, default location : foreground\n");
	    case 1 : // foreground
	      floatChamferMap->computeChamferMap((float *) data, 
						 (float *) output->getData(), dims);
	      break;
	    }
	  break;

	case WT_DOUBLE :
	  chamferType = WT_DOUBLE;
	  output->setType(WT_DOUBLE);
	  if ((fabs(dy - 1.0) < 0.000001) && (fabs(dz - 1.0) < 0.000001)) 
	    doubleChamferMap = new ChamferMap<double>((double) DBL_MAX, flag, (int) maskSize, (int) maskDim, voxelSize);
	  else
	    doubleChamferMap = new ChamferMap<double>((double) DBL_MAX, flag, (float) dy, (float) dz, voxelSize);

	  switch(location)
	    {
	    case 0 : // background
	      doubleChamferMap->computeChamferMapOut((double *) data, 
						     (double *) output->getData(), dims);
	      break;
	    case 2 : // both foreground and background
	      doubleChamferMap->computeChamferMapInOut((double *) data, 
						       (double *) output->getData(), dims);
	      break;
	    default :
	      printf("Bad chamfer map location, default location : foreground\n");
	    case 1 : // foreground
	      doubleChamferMap->computeChamferMap((double *) data, 
						  (double *) output->getData(), dims);
	      break;
	    }

	  break;

	default :
	  printf("Cannot handle this type of data in InrChamferMap...\n");
	  break;
	  
	} // end switch(type)
      } // end case COLOR_PROPAGATION
      break;
      
    default :
      break;
    } // end switch(flag) 


  switch(chamferType)
    {
    case WT_UNSIGNED_SHORT :
      myChamferMap =  (ChamferMap<unsigned short int> *) ushortChamferMap;
      break;
    case WT_SIGNED_SHORT :
      myChamferMap = (ChamferMap<short int> *) shortChamferMap;
      break;
    case WT_UNSIGNED_INT :
      myChamferMap = (ChamferMap<unsigned int> *) uintChamferMap;
      break;
    case WT_SIGNED_INT :
      myChamferMap = (ChamferMap<int> *) intChamferMap;
      break;
    case WT_UNSIGNED_LONG :
      myChamferMap = (ChamferMap<unsigned LONGINT> *) ulongChamferMap;
      break;
    case WT_SIGNED_LONG :
      myChamferMap = (ChamferMap<LONGINT> *)  longChamferMap;
      break;
    case WT_FLOAT :
      myChamferMap = (ChamferMap<float> *) floatChamferMap;
      break;
    case WT_DOUBLE :
      myChamferMap = (ChamferMap<double> *) doubleChamferMap;
      break;
    default :
      myChamferMap = 0;
      break;
    }

}






InrChamferMap::~InrChamferMap()
{
  union {
    ChamferMap<unsigned short int> * ushortChamferMap;
    ChamferMap<short int>          *  shortChamferMap;
    ChamferMap<unsigned int>       *   uintChamferMap;
    ChamferMap<int>                *    intChamferMap;
    ChamferMap<unsigned LONGINT>   *  ulongChamferMap;
    ChamferMap<LONGINT>            *   longChamferMap;
    ChamferMap<float>              *  floatChamferMap;
    ChamferMap<double>             * doubleChamferMap;
  };

  switch(chamferType)
    {
    case WT_UNSIGNED_SHORT :
      ushortChamferMap = (ChamferMap<unsigned short int> *) myChamferMap;
      delete ushortChamferMap;
      break;
    case WT_SIGNED_SHORT :
      shortChamferMap = (ChamferMap<short int> *) myChamferMap;
      delete shortChamferMap;
      break;
    case WT_UNSIGNED_INT :
      uintChamferMap = (ChamferMap<unsigned int> *) myChamferMap;
      delete uintChamferMap;
      break;
    case WT_SIGNED_INT :
      intChamferMap = (ChamferMap<int> *) myChamferMap;
      delete intChamferMap;
      break;
    case WT_UNSIGNED_LONG :
      ulongChamferMap = (ChamferMap<unsigned LONGINT> *) myChamferMap;
      delete ulongChamferMap;
      break;
    case WT_SIGNED_LONG :
      longChamferMap = (ChamferMap<LONGINT> *) myChamferMap;
      delete longChamferMap;
      break;
    case WT_FLOAT :
      floatChamferMap = (ChamferMap<float> *) myChamferMap;
      delete floatChamferMap;
      break;
    case WT_DOUBLE :
      doubleChamferMap = (ChamferMap<double> *) myChamferMap;
      delete doubleChamferMap;
      break;
    default :
      break;
    }

  //  delete myChamferMap;

//  printf("Arghh ! I am dead...\n");
}



void
InrChamferMap::exactMap()
{
  int dims[3];
  dims[0] = (int) xdim;
  dims[1] = (int) ydim;
  dims[2] = (int) zdim;
  int dimT= xdim*ydim*zdim;
  
  float * tmpBuff;

  if (isExact)
    return;

  union {
    ChamferMap<unsigned short int> * ushortChamferMap;
    ChamferMap<short int>          *  shortChamferMap;
    ChamferMap<unsigned int>       *   uintChamferMap;
    ChamferMap<int>                *    intChamferMap;
    ChamferMap<unsigned LONGINT>   *  ulongChamferMap;
    ChamferMap<LONGINT>            *   longChamferMap;
    ChamferMap<float>              *  floatChamferMap;
    ChamferMap<double>             * doubleChamferMap;
  };

  switch(chamferType)
    {
    case WT_UNSIGNED_SHORT :
      ushortChamferMap = (ChamferMap<unsigned short int> *) myChamferMap;
      tmpBuff = ushortChamferMap->getExactMap((unsigned short int *) data, dims);
      break;
    case WT_SIGNED_SHORT :
      shortChamferMap = (ChamferMap<short int> *) myChamferMap;
      tmpBuff = shortChamferMap->getExactMap((short int *) data, dims);
      break;
    case WT_UNSIGNED_INT :
      uintChamferMap = (ChamferMap<unsigned int> *) myChamferMap;
      tmpBuff = uintChamferMap->getExactMap((unsigned int *) data, dims);
      break;
    case WT_SIGNED_INT :
      intChamferMap = (ChamferMap<int> *) myChamferMap;
      tmpBuff = intChamferMap->getExactMap((int *) data, dims);
      break;
    case WT_UNSIGNED_LONG :
      ulongChamferMap = (ChamferMap<unsigned LONGINT> *) myChamferMap;
      tmpBuff = ulongChamferMap->getExactMap((unsigned LONGINT *) data, dims);
      break;
    case WT_SIGNED_LONG :
      longChamferMap = (ChamferMap<LONGINT> *) myChamferMap;
      tmpBuff = longChamferMap->getExactMap((LONGINT *) data, dims);
      break;
    case WT_FLOAT :
      floatChamferMap = (ChamferMap<float> *) myChamferMap;
      tmpBuff = floatChamferMap->getExactMap((float *) data, dims);
      break;
    case WT_DOUBLE :
      doubleChamferMap = (ChamferMap<double> *) myChamferMap;
      tmpBuff = doubleChamferMap->getExactMap((double *) data, dims);
      break;
    default :
      tmpBuff = 0;
      break;
    } // end switch(chamferType)
  
  if (tmpBuff)
    {
      setType(WT_FLOAT);
      float * pixel = (float *) data;
      for (int i = 0; i < dimT; i++)
	pixel[i] = (float) tmpBuff[i];
      
      std::free (tmpBuff);
    }
}



float
InrChamferMap::getExactValue(int pos)
{
  float result = 0.0;

  union {
    ChamferMap<unsigned short int> * ushortChamferMap;
    ChamferMap<short int>          *  shortChamferMap;
    ChamferMap<unsigned int>       *   uintChamferMap;
    ChamferMap<int>                *    intChamferMap;
    ChamferMap<unsigned LONGINT>   *  ulongChamferMap;
    ChamferMap<LONGINT>            *   longChamferMap;
    ChamferMap<float>              *  floatChamferMap;
    ChamferMap<double>             * doubleChamferMap;
  };
  
  if(isExact)
    {
      if (this->type != WT_FLOAT) {
	printf("Warning, InrChamferMap type should be float\n");
	return 0.0;
      }
      else {
	float * ptr = (float *) data;
	return ptr[pos];
      }
    }

  else
    {
      switch(chamferType)
	{
	case WT_UNSIGNED_SHORT :
	  ushortChamferMap = (ChamferMap<unsigned short int> *) myChamferMap;
	  result = ushortChamferMap->getExactValue((unsigned short int *) data, pos);
	  break;
	case WT_SIGNED_SHORT :
	  shortChamferMap = (ChamferMap<short int> *) myChamferMap;
	  result = shortChamferMap->getExactValue((short int *) data, pos);
	  break;
	case WT_UNSIGNED_INT :
	  uintChamferMap = (ChamferMap<unsigned int> *) myChamferMap;
	  result = uintChamferMap->getExactValue((unsigned int *) data, pos);
	  break;
	case WT_SIGNED_INT :
	  intChamferMap = (ChamferMap<int> *) myChamferMap;
	  result = intChamferMap->getExactValue((int *) data, pos);
	  break;
	case WT_UNSIGNED_LONG :
	  ulongChamferMap = (ChamferMap<unsigned LONGINT> *) myChamferMap;
	  result = ulongChamferMap->getExactValue((unsigned LONGINT *) data, pos);
	  break;
	case WT_SIGNED_LONG :
	  longChamferMap = (ChamferMap<LONGINT> *) myChamferMap;
	  result = longChamferMap->getExactValue((LONGINT *) data, pos);
	  break;
	case WT_FLOAT :
	  floatChamferMap = (ChamferMap<float> *) myChamferMap;
	  result = floatChamferMap->getExactValue((float *) data, pos);
	  break;
	case WT_DOUBLE :
	  doubleChamferMap = (ChamferMap<double> *) myChamferMap;
	  result = doubleChamferMap->getExactValue((double *) data, pos);
	  break;
	default :
	  break;
	}
      
      return result;
    }
}


float
InrChamferMap::getExactValue(double x, double y, double z)
{
  // x -> x position of the point in Space
  // y -> y position of the point in Space
  // z -> z position of the point in Space

  float newVal;
  // table to store the position of the different exact values.
  int pos[8]; 
  // table to store the different exact values.
  float val[8];
  
  //  int dimX = this->lattice.dims()[0];
  //  int dimY = this->lattice.dims()[1];
  //  int dimZ = this->lattice.dims()[2];
  int dimXY = xdim*ydim;

  //  float vx = this->getVoxelSize().x;
  //  float vy = this->getVoxelSize().y;
  //  float vz = this->getVoxelSize().z;

  float xFloat = (float)(x / vx); // x position of the point in the image
  float yFloat = (float)(y / vy); // y position of the point in the image
  float zFloat = (float)(z / vz); // z position of the point in the image
  
  int xInt = (int) xFloat;
  int yInt = (int) yFloat;
  int zInt = (int) zFloat;

  float a = xFloat - (float) xInt; // x - E(x)
  float b = yFloat - (float) yInt; // y - E(y)
  float c = zFloat - (float) zInt; // z - E(z)

  float a1 = 1.0f - a;
  float b1 = 1.0f - b;
  float c1 = 1.0f - c;

  pos[0] = xInt + xdim*yInt + xdim*ydim*zInt; // 0 0 0
  pos[1] = pos[0] + 1;               // 1 0 0
  pos[2] = pos[0] + xdim;            // 0 1 0
  pos[3] = pos[1] + xdim;            // 1 1 0
  pos[4] = pos[0] + dimXY;           // 0 0 1
  pos[5] = pos[2] + dimXY;           // 0 1 1 
  pos[6] = pos[3] + dimXY;           // 1 1 1
  pos[7] = pos[1] + dimXY;           // 1 0 1

  for (int p = 0; p < 8; p++)
    val[p] = this->getExactValue(pos[p]);
  
  newVal = a1*b1*c1*val[0] + a*b1*c1*val[1] + a1*b*c1*val[2] + a*b*c1*val[3]
    + a1*b1*c*val[4] + a1*b*c*val[5] + a*b*c*val[6] + a*b1*c*val[7];
  
  return newVal;
}


float
InrChamferMap::getMaxRelativeError()
{
  float maxRelError = -1.0;

  union {
    ChamferMap<unsigned short int> * ushortChamferMap;
    ChamferMap<short int>          *  shortChamferMap;
    ChamferMap<unsigned int>       *   uintChamferMap;
    ChamferMap<int>                *    intChamferMap;
    ChamferMap<unsigned LONGINT>   *  ulongChamferMap;
    ChamferMap<LONGINT>            *   longChamferMap;
    ChamferMap<float>              *  floatChamferMap;
    ChamferMap<double>             * doubleChamferMap;
  };
  
  switch(chamferType)
    {
    case WT_UNSIGNED_SHORT :
      ushortChamferMap = (ChamferMap<unsigned short int> *) myChamferMap;
      maxRelError = ushortChamferMap->getMaxRelativeError();
      break;
    case WT_SIGNED_SHORT :
      shortChamferMap = (ChamferMap<short int> *) myChamferMap;
      maxRelError = shortChamferMap->getMaxRelativeError();
      break;
    case WT_UNSIGNED_INT :
      uintChamferMap = (ChamferMap<unsigned int> *) myChamferMap;
      maxRelError = uintChamferMap->getMaxRelativeError();
      break;
    case WT_SIGNED_INT :
      intChamferMap = (ChamferMap<int> *) myChamferMap;
      maxRelError = intChamferMap->getMaxRelativeError();
      break;
    case WT_UNSIGNED_LONG :
      ulongChamferMap = (ChamferMap<unsigned LONGINT> *) myChamferMap;
      maxRelError = ulongChamferMap->getMaxRelativeError();
      break;
    case WT_SIGNED_LONG :
      longChamferMap = (ChamferMap<LONGINT> *) myChamferMap;
      maxRelError = longChamferMap->getMaxRelativeError();
      break;
    case WT_FLOAT :
      floatChamferMap = (ChamferMap<float> *) myChamferMap;
      maxRelError = floatChamferMap->getMaxRelativeError();
      break;
    case WT_DOUBLE :
      doubleChamferMap = (ChamferMap<double> *) myChamferMap;
      maxRelError = doubleChamferMap->getMaxRelativeError();
      break;
    default :
      break;
    }
  
  return maxRelError;
}
