#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/Inrimage4D.h"
#include <inrimage/Inrimage4D.h>

namespace yav {

/** returns the default format for reading a 4D Image 
     (based on General Format) */
static  IMAGE4D_FORMAT *getInrimage4DGeneralFormat();


/// static list of 4D formats
static IMAGE4D_FORMAT *first4DFormat=0;
/** dummy class whose constructor updates the list of supported file format */
class Inrimage4DInitFormat {
public:
  /** actually register TclInrimage_Init in the hash table */
  Inrimage4DInitFormat() {
    IMAGE4D_FORMAT *f= new IMAGE4D_FORMAT();
    f->readImage=&Inrimage4D::readGeneralFormat;
    f->writeImage=&Inrimage4D::writeGeneralFormat;
    f->testImageFormat=&Inrimage4D::testGeneralFormat;
    strcpy(f->fileExtension,".inr4D");
    strcpy(f->realName,"Inr4D");
    first4DFormat=f;
    f->next=0;
    }
  virtual ~Inrimage4DInitFormat() {}
};
  
static Inrimage4DInitFormat inrimage4DInitFormat;



void add4DImageFormat(IMAGE4D_FORMAT *f) {
  IMAGE4D_FORMAT *format=first4DFormat;
  if (f) {
    if (format==0)
      first4DFormat=f;
    else {
      while (format->next!=0)
	format=format->next;
      format->next=f;
    }
    f->next=0;
  }
}
IMAGE4D_FORMAT *getInrimage4DGeneralFormat() {
  IMAGE4D_FORMAT *format=first4DFormat;  
  while((format!=0) &&(strcmp(format->realName,"Inr4D"))) 
    format=format->next;
  return format;
}
IMAGE4D_FORMAT *firstImage4DFormat() {
  return(first4DFormat);
}

int Inrimage4D::readGeneralFormat(Inrimage4D *inr4D,
				  const char *fileName) {
  try {
    inr4D->read((char *) fileName, 0);
  }
  catch(Inrimage4D::ReadError &e) {
    e.push("int Inrimage4D::readGeneralFormat(Inrimage4D *,const char *)");
    throw e;
  }
  catch(Inrimage::ImageNotFound &e) {
    e.push("int Inrimage4D::readGeneralFormat(Inrimage4D *,const char *)");
    throw e;
  }
  catch(GeneralFormat::FileNotFound &e) {
    e.push("int Inrimage4D::readGeneralFormat(Inrimage4D *,const char *)");
    throw e;
  }
  return 1;
}
int Inrimage4D::writeGeneralFormat(Inrimage4D *inr4D,
				    const char *fileName) {
  inr4D->GeneralFormat::write(fileName);
  return 1;
}
int Inrimage4D::testGeneralFormat(char *magic,const char *fileName) {
  if (!strcmp(magic,"# 4D"))
    return 1;
  else return -1;
}


Inrimage4D::ReadError::ReadError(const std::string& from,
				 const std::string& filename)
  : Exception(from, "invalid inrimage file \'" + filename + "\'") {
}

Inrimage4D::Inrimage4D(void) : Owner<Inrimage>(0),GeneralFormat("4D Inrimage File", "4D INRIMAGE"), name(filename),array(0),vt(1.0),format(0) {}

Inrimage4D::Inrimage4D(const char *filename)
  throw(GeneralFormat::FileNotFound, Inrimage::ImageNotFound,
	Inrimage4D::ReadError,Inrimage4D::EmptyInrimageList, Inrimage4D::DifferentImageType, Inrimage4D::DifferentImageSize)
    :  Owner<Inrimage>(0),GeneralFormat("4D Inrimage File", "4D INRIMAGE"), name(filename),array(0),vt(1.0),format(0) {

  char from[] = "Inrimage4D::Inrimage4D(const char *, Scene4D *)";

  init();
  try {
    IMAGE4D_FORMAT *selectedFormat=0,*format;
    format=first4DFormat;  
    bool found=false;
    /** get magic */
    char magic[5];
    _image *im = _initImage();
    _openReadImage(im, name);
    ImageIO_read(im, magic, 4);
    magic[4] = '\0';
    ImageIO_close(im);
    
    while((found==false) && (format!=0)) {
      if ((*format->testImageFormat)(magic,filename)>=0)
	{
	  selectedFormat=format;
	  found=true;
	}
      format=format->next;
    }
    if (found==false)
      throw(Inrimage4D::UnknownImageFormat("Inrimage4D::Inrimage4D(const char *)"));
    if ((*selectedFormat->readImage)(this,filename)<0) {
      throw(Inrimage4D::ReadError("Inrimage4D::Inrimage4D(const char *)",filename));
    }
    format=selectedFormat;
  }
  catch(Inrimage4D::ReadError &e) {
    e.push(from);
    throw e;
  }
  catch(Inrimage::ImageNotFound &e) {
    e.push(from);
    throw e;
  }
  catch(GeneralFormat::FileNotFound &e) {
    e.push(from);
    throw e;
  }
  catch(Exception e) {
    throw ReadError(from, filename);
  }
  buildArray();
}

Inrimage4D::Inrimage4D(std::list<Inrimage *> sinrs) throw(Inrimage4D::EmptyInrimageList, Inrimage4D::DifferentImageType, Inrimage4D::DifferentImageSize)
  :  Owner<Inrimage>(0),GeneralFormat("4D Inrimage File", "4D INRIMAGE"), name("noname.inr4D"),array(0),vt(1.0),format(0) {

  std::list<Inrimage *>::iterator it;
  it=sinrs.begin();
  Inrimage *im;
  if (sinrs.size()==0)
    {
      throw (Inrimage4D::EmptyInrimageList("Inrimage4D::Inrimage4D(std::list<Inrimage *>&"));
    }
  im = (*it);
  for(;it!=sinrs.end();it++)
    {
      if ((*it)->getType()!=im->getType())
	throw (Inrimage4D::DifferentImageType("Inrimage4D::Inrimage4D(std::list<Inrimage *>&"));
      if (((*it)->getX()!=im->getX())|| ((*it)->getY()!=im->getY()) ||
	  ((*it)->getZ()!=im->getZ())  || ((*it)->getVdim()!=im->getVdim()) )
	throw (Inrimage4D::DifferentImageSize("Inrimage4D::Inrimage4D(std::list<Inrimage *>&"));
      inrs.push_back(*it);
    }
  
  init();

  buildArray();
}
void Inrimage4D::addInrimage(Inrimage *inr, bool bArray) throw (Inrimage4D::DifferentImageType,Inrimage4D::DifferentImageSize){

   std::vector<Inrimage *>::iterator it=inrs.begin();
   if (it==inrs.end())
     inrs.push_back(inr);
   else {
     Inrimage *im=(*it);
     if (inr->getType()!=im->getType())
       throw (Inrimage4D::DifferentImageType("Inrimage4D::Inrimage4D(std::list<Inrimage *>&"));
     if ((inr->getX()!=im->getX())|| (inr->getY()!=im->getY()) ||
	  (inr->getZ()!=im->getZ())  || (inr->getVdim()!=im->getVdim()) )
       throw (Inrimage4D::DifferentImageSize("Inrimage4D::Inrimage4D(std::list<Inrimage *>&"));
     inrs.push_back(inr);
   }
   if (bArray)
     buildArray();
     
}
void Inrimage4D::setVoxelSize(float _vx, float _vy, float _vz, float _vt)  
  throw(BaseInrimage::InvalidVoxelSize) {
  if (_vt<=0.0)
    throw BaseInrimage::InvalidVoxelSize("Inrimage4D::setVoxelSize");
  vt=_vt;

  std::vector<Inrimage *>::iterator it;
  for( it = inrs.begin(); it != inrs.end(); it++)
    {
      (*it)->setVoxelSize(_vx,_vy,_vz);
    }
  
}
void Inrimage4D::init(void) {
  addModuleVersion("Inrimage names", true, 1.0,
		   (GeneralFormatModuleVersion::INPUTREADFUNCTION)
		   &Inrimage4D::readInrNames,
		   (GeneralFormatModuleVersion::INPUTSKIPFUNCTION)
		   &Inrimage4D::skipInrNames,
		   (GeneralFormatModuleVersion::OUTPUTWRITEFUNCTION)
		   &Inrimage4D::writeInrNames);
}

void Inrimage4D::getBounds(Voxel *min, Voxel *max, bool vectorNorm) const {
  getFirstInrimage()->getBounds(min,max,vectorNorm);
}


void Inrimage4D::buildArray(void) {
  if (array) {
    delete [] array;
  }

  array = new void***[inrs.size()];
  std::vector<Inrimage *>::iterator it;
  unsigned int i;
  for(i =0, it = inrs.begin(); it != inrs.end(); it++, i++)
    {
      array[i] = (*it)->getArray();
      // create array of Object Pointers
      inrPointers.push_back( new ObjectPointer<Inrimage>(this,*it) );
    }
}

Inrimage4D::~Inrimage4D() { 
  if (array) {
    delete [] array;
  }
  std::vector<ObjectPointer<Inrimage> *>::iterator ito;
  for(ito = inrPointers.begin(); ito != inrPointers.end(); ito++) delete *ito;
  std::vector<Inrimage *>::iterator it;
  for(it = inrs.begin(); it != inrs.end(); it++) delete *it;
  inrs.erase(inrs.begin(), inrs.end());
}


void Inrimage4D::readInrNames(GFInput &input) {
  TclString name;
  Inrimage *inr,*im=0;
  char from[] = "Inrimage4D::readFile(GFInput &)";
  unsigned int nb;

  input >> nb;
  if (nb==0)
    throw (Inrimage4D::EmptyInrimageList("Inrimage4D::readInrNames(GFInput &)"));
  for(unsigned int i = 0; i < nb; i++) {
    input >> name;
    if(name[0] != '/') {
      int j = filename.getLength() - 1;
      while(j >= 0 && filename[j] != '/') j--;
      if(j > 0) name = TclString(filename, 0, j) + name;
    }
    try {
      inr = new Inrimage(name);
    }
    catch(Inrimage::ImageNotFound &e) {
      e.push(from);
      throw e;
    }
    catch(Exception &) {
      throw ReadError(from, (char *) name);
    }
    if (i>0)
      {
	if (inr->getType()!=im->getType())
	  throw (Inrimage4D::DifferentImageType("Inrimage4D::readInrNames(GFInput &)"));
	if ((inr->getX()!=im->getX())|| (inr->getY()!=im->getY()) ||
	    (inr->getZ()!=im->getZ())  || (inr->getVdim()!=im->getVdim()) )
	  throw (Inrimage4D::DifferentImageSize("Inrimage4D::readInrNames(GFInput &)"));
      }
    else 
      im=inr;
    inrs.push_back(inr);
  }
  
}


void Inrimage4D::skipInrNames(GFInput &input) {
  unsigned int nb;
  input >> nb;
  TclString name;
  for(unsigned int i = 0; i < nb; i++) input >> name;
}


void Inrimage4D::writeInrNames(GFOutput &output) {
  output << (int)inrs.size() << ENDL;
  unsigned int i;
  std::vector<Inrimage *>::iterator itd;

  /* get base name of the 4D image */
  TclString name;
  TclString base = output.getFileName();
  if((base.getLength() > 6) &&
     (TclString(base, base.getLength()-6, base.getLength()-1) == TclString(".inr4D")))
    base = TclString(base, 0, base.getLength()-7);  
  if((base.getLength() > 2) &&
     (TclString(base, base.getLength()-2, base.getLength()-1) == TclString("4D")))
    base = TclString(base, 0, base.getLength()-3);      


  for(i = 0, itd = inrs.begin(); itd != inrs.end(); itd++, i++) {
    /** tests oif that 3D image have been saved on disk previously */
    if (((*itd)->getName()==TclString("noname.inr.gz")) || ((*itd)->getName()==EMPTY_STRING))
      {
	/* if not rename the 3D image based on the name of the 4D image and
	   save it on disk */
	name.sprintf("%s%03u.inr.gz", (char *) base, i);
	output << name << ENDL;
	(*itd)->write(name);
      }
    else
      output << (*itd)->getName() << ENDL;
  }
  
}

void Inrimage4D::write(const char *filename) {
 IMAGE4D_FORMAT *selectedFormat=0;
 IMAGE4D_FORMAT *format=first4DFormat;
 unsigned int i,extLength,length;
 char *ptr;
 char ext[300];

 length=strlen(filename);
 while((selectedFormat==0) && (format!=0)) {
   /* scan all extensions for that format */
   ptr=&format->fileExtension[0];
   
   do {
     /* get next file extension */
     i=0;
     for(i=0;((*ptr)!=',' && (*ptr)!='\0');i++,ptr++) {
       ext[i]=(*ptr);
     }
     if ((*ptr)==',') {
       ext[i]='\0';
       ptr++;
     }
     else {
       ext[i]='\0';
     }
     extLength=strlen(ext);
     /* test if the tail of name matches the extension */
     if ( !strcmp( filename + length - extLength, ext) ) {
       selectedFormat=format;
     }
     
   } while (((*ptr)!='\0') && (selectedFormat==0));
   format=format->next;
 }
 if ((selectedFormat==0) && (format!=0))
   selectedFormat=format;
 else if (selectedFormat==0){
   selectedFormat=getInrimage4DGeneralFormat();
 }
 (*selectedFormat).writeImage(this,filename);
}

Inrimage *Inrimage4D::getInrimage(unsigned int i) {
  unsigned int max = inrs.size() - 1;
  if(i > max) i = max;
  return inrs[i];
}



Voxel *Inrimage4D::getVoxel(unsigned int x, unsigned int y, unsigned int z,
			    unsigned int t)
  const {

  switch(getType()) {
  case Inrimage::WT_SIGNED_CHAR:
    return new Voxel8s(((char ****) array)[t][z][y][x]);
  case Inrimage::WT_UNSIGNED_CHAR:
    return new Voxel8u(((unsigned char ****) array)[t][z][y][x]);
  case Inrimage::WT_UNSIGNED_SHORT:
    return new Voxel16u(((unsigned short int ****) array)[t][z][y][x]);
  case Inrimage::WT_SIGNED_SHORT:
    return new Voxel16s(((short int ****) array)[t][z][y][x]);
  case Inrimage::WT_UNSIGNED_INT:
    return new Voxel32u(((unsigned int ****) array)[t][z][y][x]);
  case Inrimage::WT_SIGNED_INT:
    return new Voxel32s(((int ****) array)[t][z][y][x]);
  case Inrimage::WT_UNSIGNED_LONG:
    return new Voxel64u(((unsigned LONGINT ****) array)[t][z][y][x]);
  case Inrimage::WT_SIGNED_LONG:
    return new Voxel64s(((LONGINT ****) array)[t][z][y][x]);
  case Inrimage::WT_FLOAT:
    return new VoxelFloat(((float ****) array)[t][z][y][x]);
  case Inrimage::WT_DOUBLE:
    return new VoxelDouble(((double ****) array)[t][z][y][x]);
  case Inrimage::WT_RGB:
    return new VoxelRGB(((unsigned char ****) array)[t][z][y][3*x],
			((unsigned char ****) array)[t][z][y][3*x+1],
			((unsigned char ****) array)[t][z][y][3*x+2]);
  case Inrimage::WT_RGBA:
#if (defined (_ALPHA_) || defined (_LINUX_))
    return new VoxelRGBA(((unsigned char ****) array)[t][z][y][4*x+3],
			 ((unsigned char ****) array)[t][z][y][4*x+2],
			 ((unsigned char ****) array)[t][z][y][4*x+1],
			 ((unsigned char ****) array)[t][z][y][4*x]);
#else
    return new VoxelRGBA(((unsigned char ****) array)[t][z][y][4*x],
			 ((unsigned char ****) array)[t][z][y][4*x+1],
			 ((unsigned char ****) array)[t][z][y][4*x+2],
			 ((unsigned char ****) array)[t][z][y][4*x+3]);
#endif
  case Inrimage::WT_FLOAT_VECTOR:
    if(getVectorialMode() == VM_INTERLACED) {
      return new VoxelFV(((float ****) array)[t][z][y] + getVdim()*x, getVdim());
    }
    else {
      float *val = new float[getVdim()];
      for(unsigned int i = 0; i < getVdim(); i++)
	val[i] = ((float ****) inrs[t]->getVarray())[i][z][y][x];
      Voxel *v = new VoxelFV(val, getVdim());
      delete val;
      return v;
    }
    break;
  case Inrimage::WT_DOUBLE_VECTOR:
     if(getVectorialMode() == VM_INTERLACED) {
      return new VoxelDV(((double ****) array)[t][z][y] + getVdim()*x, getVdim());
    }
    else {
      double *val = new double[getVdim()];
      for(unsigned int i = 0; i < getVdim(); i++)
	val[i] = ((double ****) inrs[t]->getVarray())[i][z][y][x];
      Voxel *v = new VoxelDV(val, getVdim());
      delete val;
      return v;
    }
   break;
  }

  return 0;
}


Voxel& Inrimage4D::getVoxel(unsigned int x, unsigned int y, unsigned int z,
			    unsigned int t,Voxel &v) const {

  switch(getType()) {
  case Inrimage::WT_SIGNED_CHAR:
    (static_cast<Voxel8s *>(&v))->setValue(((char ****) array)[t][z][y][x]);
    break;
  case Inrimage::WT_UNSIGNED_CHAR:
    (static_cast<Voxel8u *>(&v))->setValue(((unsigned char ****) array)[t][z][y][x]);
    break;
  case Inrimage::WT_UNSIGNED_SHORT:
    (static_cast<Voxel16u *>(&v))->setValue(((unsigned short int ****) array)[t][z][y][x]);
    break;
  case Inrimage::WT_SIGNED_SHORT:
    (static_cast<Voxel16s *>(&v))->setValue(((short int ****) array)[t][z][y][x]);
    break;
  case Inrimage::WT_UNSIGNED_INT:
    (static_cast<Voxel32u *>(&v))->setValue(((unsigned int ****) array)[t][z][y][x]);
    break;
  case Inrimage::WT_SIGNED_INT:
    (static_cast<Voxel32s *>(&v))->setValue(((int ****) array)[t][z][y][x]);
    break;
  case Inrimage::WT_UNSIGNED_LONG:
    (static_cast<Voxel64u *>(&v))->setValue(((unsigned LONGINT ****) array)[t][z][y][x]);
    break;
  case Inrimage::WT_SIGNED_LONG:
    (static_cast<Voxel64s *>(&v))->setValue(((LONGINT ****) array)[t][z][y][x]);
    break;
  case Inrimage::WT_FLOAT:
    (static_cast<VoxelFloat *>(&v))->setValue(((float ****) array)[t][z][y][x]);
    break;
  case Inrimage::WT_DOUBLE:
    (static_cast<VoxelDouble *>(&v))->setValue(((double ****) array)[t][z][y][x]);
    break;
  case Inrimage::WT_RGB:
    (static_cast<VoxelRGB *>(&v))->setValue(
                       ((unsigned char ****) array)[t][z][y][3 * x],
		       ((unsigned char ****) array)[t][z][y][3 * x + 1],
		       ((unsigned char ****) array)[t][z][y][3 * x + 2]);
    break;
  case Inrimage::WT_RGBA:
#if (defined (_ALPHA_) || defined (_LINUX_))
    (static_cast<VoxelRGBA *>(&v))->setValue(
                       ((unsigned char ****) array)[t][z][y][3 * x + 3],
		       ((unsigned char ****) array)[t][z][y][3 * x + 2],
		       ((unsigned char ****) array)[t][z][y][3 * x + 1],
		       ((unsigned char ****) array)[t][z][y][3 * x]);
    break;
#else
    (static_cast<VoxelRGBA *>(&v))->setValue(
                       ((unsigned char ****) array)[t][z][y][3 * x],
		       ((unsigned char ****) array)[t][z][y][3 * x + 1],
		       ((unsigned char ****) array)[t][z][y][3 * x + 2],
		       ((unsigned char ****) array)[t][z][y][3 * x + 3]);
    break;
#endif
  case Inrimage::WT_FLOAT_VECTOR:
    if(getVectorialMode() == VM_INTERLACED) {
      (static_cast<VoxelFV *>(&v))->setValue(((float ****) array)[t][z][y] + x*getVdim(), getVdim());
    }
    else {
      VoxelFV *fv = static_cast<VoxelFV *>(&v);
      float *f = ((float ****) array)[t][z][y] + x;
      unsigned long size = getZ()* getY()* getX();
      for(unsigned int i = 0; i < getVdim(); i++, f += size) fv->operator[](i) = *f;
    }
    break;
  case Inrimage::WT_DOUBLE_VECTOR:
    if(getVectorialMode() == VM_INTERLACED) {
      (static_cast<VoxelDV *>(&v))->setValue(((double ****) array)[t][z][y] + x*getVdim(), getVdim());
    }
    else {
      VoxelDV *fv = static_cast<VoxelDV *>(&v);
      double *f = ((double ****) array)[t][z][y] + x;
      unsigned long size = getZ()* getY()* getX();
      for(unsigned int i = 0; i < getVdim(); i++, f += size) fv->operator[](i) = *f;
    }
    break;
  }
  return v;
}


void Inrimage4D::setVoxel(unsigned int x, unsigned int y, unsigned int z,
			  unsigned int t,const Voxel& v) {
  unsigned int i;
  const VoxelRGB *vrgb;
  const VoxelRGBA *vrgba;
  const VoxelFV *vfv;
  const VoxelDV *vdv;

  switch(getType()) {
  case Inrimage::WT_UNSIGNED_CHAR:
    ((unsigned char ****) array)[t][z][y][x] =
      static_cast<const Voxel8u *>(&v)->getValue();
    break;
  case Inrimage::WT_UNSIGNED_SHORT:
    ((unsigned short int ****) array)[t][z][y][x] =
      static_cast<const Voxel16u *>(&v)->getValue();
    break;
  case Inrimage::WT_SIGNED_SHORT:
    ((short int ****) array)[t][z][y][x] =
      static_cast<const Voxel16s *>(&v)->getValue();
    break;
  case Inrimage::WT_UNSIGNED_INT:
    ((unsigned int ****) array)[t][z][y][x] =
      static_cast<const Voxel32u *>(&v)->getValue();
    break;
  case Inrimage::WT_SIGNED_INT:
    ((int ****) array)[t][z][y][x] =
      static_cast<const Voxel32s *>(&v)->getValue();
    break;
  case Inrimage::WT_UNSIGNED_LONG:
    ((unsigned LONGINT ****) array)[t][z][y][x] =
      static_cast<const Voxel64u *>(&v)->getValue();
    break;
  case Inrimage::WT_SIGNED_LONG:
    ((LONGINT ****) array)[t][z][y][x] =
      static_cast<const Voxel64s *>(&v)->getValue();
    break;
  case Inrimage::WT_FLOAT:
    ((float ****) array)[t][z][y][x] =
      static_cast<const VoxelFloat *>(&v)->getValue();
    break;
  case Inrimage::WT_DOUBLE:
    ((double ****) array)[t][z][y][x] =
      static_cast<const VoxelDouble *>(&v)->getValue();
    break;
  case Inrimage::WT_RGB:
    vrgb = static_cast<const VoxelRGB *>(&v);
    ((unsigned char ****) array)[t][z][y][3 * x] = vrgb->getValue()[0];
    ((unsigned char ****) array)[t][z][y][3 * x + 1] = vrgb->getValue()[1];
    ((unsigned char ****) array)[t][z][y][3 * x + 2] = vrgb->getValue()[2];
    break;
  case Inrimage::WT_RGBA:
    vrgba = static_cast<const VoxelRGBA *>(&v);
#if (defined (_ALPHA_) || defined (_LINUX_))
    ((unsigned char ****) array)[t][z][y][3 * x + 3] = vrgba->getValue()[0];
    ((unsigned char ****) array)[t][z][y][3 * x + 2] = vrgba->getValue()[1];
    ((unsigned char ****) array)[t][z][y][3 * x + 1] = vrgba->getValue()[2];
    ((unsigned char ****) array)[t][z][y][3 * x] = vrgba->getValue()[3];
#else
    ((unsigned char ****) array)[t][z][y][3 * x] = vrgba->getValue()[0];
    ((unsigned char ****) array)[t][z][y][3 * x + 1] = vrgba->getValue()[1];
    ((unsigned char ****) array)[t][z][y][3 * x + 2] = vrgba->getValue()[2];
    ((unsigned char ****) array)[t][z][y][3 * x + 3] = vrgba->getValue()[3];
#endif
    break;
  case Inrimage::WT_FLOAT_VECTOR:
    vfv = static_cast<const VoxelFV *>(&v);
    if(getVectorialMode() == VM_INTERLACED) {
      for(i = 0; i < getVdim(); i++)
	((float ****) array)[t][z][y][getVdim() * x + i] = vfv->getValue()[i];
    }
    else {
      for(i = 0; i < getVdim(); i++)
	((float ****) inrs[t]->getVarray())[i][z][y][x] = vfv->getValue()[i];
    }
    break;
  case Inrimage::WT_DOUBLE_VECTOR:
    vdv = static_cast<const VoxelDV *>(&v);
    if(getVectorialMode() == VM_INTERLACED) {
      for(i = 0; i < getVdim(); i++)
	((double ****) array)[t][z][y][getVdim() * x + i] = vdv->getValue()[i];
    }
    else {
      for(i = 0; i < getVdim(); i++)
	((double ****) inrs[t]->getVarray())[i][z][y][x] = vdv->getValue()[i];
    }
    break;
  default:
    break;
  }
}

void Inrimage4D::crop(unsigned int x1, unsigned int y1, unsigned int z1,
		    unsigned int x2, unsigned int y2, unsigned int z2)
  throw(Inrimage::InvalidArguments) {

  std::vector<Inrimage *>::iterator it;
  for(it=inrs.begin();it!=inrs.end();++it) {
    (*it)->crop(x1,y1,z1,x2,y2,z2);
  }
}

std::list<Inrimage *> Inrimage4D::getInrimageList(void) {
  std::list<Inrimage *> inrList;
  std::vector<Inrimage *>::iterator it;
  for(it=inrs.begin();it!=inrs.end();++it) {
    inrList.push_back((*it));
  }
  return inrList;
  
}

} // namespace yav
