#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/Inrimage4DView.h"
#ifdef GRAPHICS
#include <inrimage/Inrimage4DView.h>
#ifdef TCL_INTERFACE
#include <graphics2D/TimeSliceCamera.h>
#include <inrimage/TclInrimage4DBackground.h>
#include <inrimage/TclInrimage4DView.h>
#endif

namespace yav {

Inrimage4DView::Inrimage4DView(Inrimage4D *inr, Scene2D *scene,bool _createGLScale)
  : TimeSliceable(scene, true), Owner<Inrimage4D>(inr),
    createGLScale(_createGLScale) {
  init();
}

Inrimage4DView::Inrimage4DView(Inrimage4D *inr, Inrimage4D *i,Scene2D *scene,
			   bool _createGLScale )
  : TimeSliceable(scene, true), Owner<Inrimage4D>(inr),
    createGLScale(_createGLScale) {
}
Inrimage4DView::Inrimage4DView(const TclString &filename, Scene2D *scene)
  : TimeSliceable(scene, true), Owner<Inrimage4D>(0),
    createGLScale(true) {
  pointer=new Inrimage4D(filename);
}

Inrimage4DView::~Inrimage4DView() {
  while(!backgroundList.empty()) delete backgroundList.front();
}


void Inrimage4DView::init(void) {
  TimeSliceCamera *slicecam;
  ObjectsList<Camera2D>::iterator it = scene->firstCamera();
  for(; it!=scene->lastCamera(); it++)
    {
      slicecam= dynamic_cast<TimeSliceCamera *>(*it);
      if (slicecam)
	createdCamera(*slicecam);
    }
}


void Inrimage4DView::dying(const ObjectPointer<Inrimage4D>&) {
  std::list<Inrimage4DBackground *>::iterator it;
  for(it=backgroundList.begin();it!=backgroundList.end();it=backgroundList.begin())
    delete (*it);
  delete this;
}


void Inrimage4DView::createdCamera(SliceCamera& cam) {
  if(cam.getScene() == getScene() && !cam.getBackground())
    doCreateCamera(cam);
}


void Inrimage4DView::doCreateCamera(SliceCamera& cam) {
  cameras[&cam] = new Sliceable::CamInfo();
  double m;

  if(pointer->getVX() < pointer->getVY()) m = pointer->getVX();
  else m = pointer->getVY();
  if(pointer->getVZ() < m) m = pointer->getVZ();
  //  cam.scaleCamera(1.0 / (m * cam.getScale()));
  cam.scaleCamera(1.0 / (m));
  switch(((TimeSliceCamera &)cam).getGeometry()) {
  case TimeSliceCamera::SLICE_X:
  case TimeSliceCamera::SLICE_XT:
    cam.setThickness(pointer->getVX());
    break;
  case TimeSliceCamera::SLICE_Y:
  case TimeSliceCamera::SLICE_YT:
    cam.setThickness(pointer->getVY());
    break;
  case TimeSliceCamera::SLICE_Z:
  case TimeSliceCamera::SLICE_ZT:
  case TimeSliceCamera::SLICE_ALPHA:
    cam.setThickness(pointer->getVZ());
    break;
  }
#ifdef TCL_INTERFACE
  TclInrimage4DBackground *b;
  if (createGLScale)
    b = new TclInrimage4DBackground(getInterface()->getModule(),this,
				    (TimeSliceCamera *)&cam);
  else
    b = new TclInrimage4DBackground(getInterface()->getModule(),this,
				    (TimeSliceCamera *)&cam,false,0,false,0);
  backgroundList.push_back(b);  
  
#endif
  cam.focusOnBackground();
}

Inrimage4DBackground *Inrimage4DView::getBackgroundFromCamera(TimeSliceCamera *cam)
{
  std::list<Inrimage4DBackground *>::iterator it;
  for(it=backgroundList.begin();(it!=backgroundList.end()) && ((*it)->getCamera()!=cam);it++);
  if (it==backgroundList.end())
    return (0);
  else
    return (*it);
}

void Inrimage4DView::deletedCamera(SliceCamera& cam) {
  std::list<Inrimage4DBackground *>::iterator it;
  for(it=backgroundList.begin();(it!=backgroundList.end()) && ((*it)->getCamera()!=&cam);it++);
  if (it!=backgroundList.end())
    {
      cam.setBackground(0);
      Inrimage4DBackground *b=(*it);
      // avoid calling background dying
      b->slicecam=0;
      delete b;
    }
}


void Inrimage4DView::changeInrimage(Inrimage4D *inr) {
  if (inr!=pointer()) {
    pointer=inr;
    std::list<Inrimage4DBackground *>::iterator it;
    for(it = backgroundList.begin(); it != backgroundList.end(); it++) {
      (*it)->inrimageChanged();
    }
  }
}


void Inrimage4DView::refreshCameras(void) {
  std::list<Inrimage4DBackground *>::iterator it;
  for(it = backgroundList.begin(); it != backgroundList.end(); it++) {
    (*it)->refreshCamera();
  }
}

void Inrimage4DView::redraw(Camera2D *cam) {
  std::map<SliceCamera *, CamInfo *>::iterator it =
    cameras.find((SliceCamera *) cam);
  if(it != cameras.end() && (it->second->needRedraw ||
			     redrawFlag == Data2D::UPDATE ||
			     redrawFlag == Data2D::CREATION)) {
    TimeSliceCamera *tcam= dynamic_cast<TimeSliceCamera *>(it->first);
    if (tcam)
      {
	Inrimage4DBackground *img = getBackgroundFromCamera(tcam);
	if(img) {
	  img->redraw();
	}
      }
    it->second->needRedraw = false;
  }
}



Inrimage4D *Inrimage4DView::getInrimage(void) const {
  return( (Inrimage4D *)pointer()); }

#ifdef TCL_INTERFACE
TclInrimage4DView *Inrimage4DView::getInterface(void) const {
  return (TclInrimage4DView *) Client::getInterface();
}
#endif



} // namespace yav

#endif
