#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/InrimageROI.h"

#include <inrimage/InrimageROI.h>
#include <inrimage/Image3DCameraBackground.h>
#include <inrimage/ConnectingElement.h>
#include <inrimage/arithmetic.h>
#include <inrimage/StructuringElement.h>

#include <vector>
#if ((defined(__GNUC__)) && (__GNUC__ <3 ))
#include <algo.h>
#else
#include <algorithm>
#endif

namespace yav {


void InrimageROI::addVoxel(std::list< Vec3<unsigned int>  > &plist) {
  if (getPointer()) { 
    Inrimage *inr=getPointer();
    Voxel *greyVoxel= Voxel::voxelByType(inr->getType());
    unsigned int dimx,nplanes,index;
    std::list< Vec3<unsigned int>  >::iterator it;

    std::list< std::pair<unsigned int, ROIVoxel *> > newEntries;

    dimx=getPointer()->getX();
    nplanes=getPointer()->getX()*getPointer()->getY();
    
    for(it=plist.begin();it!=plist.end();it++) {
      ROIVoxel *voxel= new ROIVoxel();
      inr->getVoxel((*it)[0],(*it)[1],(*it)[2],*greyVoxel);
      if (greyVoxel) {
	voxel->voxelValue=greyVoxel->getValueAsDouble();
	voxel->position=(*it);
    
	index=(*it)[0] + (*it)[1] *dimx + (*it)[2]*nplanes;
	voxelMap[index]=voxel;
	newEntries.push_back(std::pair<unsigned int,ROIVoxel *>( index,voxel) );
      } else 
      delete voxel;
    }
	redrawScene();
  }
}
void InrimageROI::addVoxel(std::list<ROIVoxel *> vlist) {
  std::list<ROIVoxel *>::iterator itv;
  
  unsigned int dimx,nplanes,index;

  dimx=getPointer()->getX();
  nplanes=getPointer()->getX()*getPointer()->getY();
  std::list< std::pair<unsigned int, ROIVoxel *> > newEntries;

  for(itv=vlist.begin();itv!=vlist.end();itv++) {
    if (*itv) {
      index=(*itv)->position[0] + (*itv)->position[1] *dimx +
	(*itv)->position[2]*nplanes;
      voxelMap[index]=(*itv);
      newEntries.push_back(std::pair<unsigned int,ROIVoxel *>( index,(*itv) ));
    }
    
  }
redrawScene();
//  addDisplay(newEntries);
}


bool InrimageROI::removeVoxel(std::list< Vec3<unsigned int>  > &plist) {
  unsigned int dimx,nplanes,index;

  dimx=getPointer()->getX();
  nplanes=getPointer()->getX()*getPointer()->getY();
  std::list< Vec3<unsigned int>  >::iterator it;
  std::map<unsigned int,ROIVoxel *>::iterator itm;

  std::list<unsigned int> oldEntries;

  bool out=true;
  for(it=plist.begin();it!=plist.end();it++) {
    index=(*it)[0] + (*it)[1] *dimx + (*it)[2]*nplanes;
    itm=voxelMap.find(index);
    if (itm!=voxelMap.end()) {
      oldEntries.push_back( index);
      voxelMap.erase(itm);
    }
    else
      out=false;
  }
   redrawScene();
  return out;
}

void InrimageROI::redraw(Camera2D *cam) 
{
	if ((cam->getBackground()) && (visibility)) {
		Image3DCameraBackground *b=static_cast<Image3DCameraBackground *> (cam->getBackground());
		Vec2<double> pos2D;
		std::map<unsigned int,ROIVoxel *>::iterator itv;
		//unsigned int sliceNumber=b->getSliceNumber(); // commented by fpoyer. reason : unused variable
		glColor3fv((const float *)roiColor);
		double px,py;
		if (!strcmp(b->getSliceName(),"sliceZ")) { 
			px=voxelSize[0]*0.5;
			py=voxelSize[1]*0.5;
		} else if (!strcmp(b->getSliceName(),"sliceY")) {
			px=voxelSize[0]*0.5;
			py=voxelSize[2]*0.5;
		} else if (!strcmp(b->getSliceName(),"sliceX")) {
			px=voxelSize[2]*0.5;
			py=voxelSize[1]*0.5;
		}
    	Vec4<double> voxel;
	int i;
	for(i=0,itv=voxelMap.begin();itv!=voxelMap.end();itv++,i++)
		{
			voxel.setValue((*itv).second->position.getX(),
				(*itv).second->position.getY(),
				(*itv).second->position.getZ(),0.0);

			if (b->mapImageTo2DWorldCoordinates(voxel,pos2D)) {
					glBegin(GL_POLYGON);
					glVertex2d(pos2D.getX()+px,pos2D.getY()+py);
					glVertex2d(pos2D.getX()+px,pos2D.getY()-py);
					glVertex2d(pos2D.getX()-px,pos2D.getY()-py);
					glVertex2d(pos2D.getX()-px,pos2D.getY()+py);
					glEnd();
			}
		}

	}	

}

void InrimageROI::init() {

  roiColor[0]=0.0;
  roiColor[1]=0.0;
  roiColor[2]=1.0;

  opacityLevel=FULL_OPACITY;
  useGreyLevel=false;

  visibility=true;

  leftButtonOn=false;
  selectionMode=IDLE_MODE;
  brushWidth=1;
  eraserWidth=3;

}
InrimageROI::InrimageROI(Scene2D *scene,  char *filename) : ROIVoxelList(filename), Data2D(scene), surface(0)
{
  init();
}

InrimageROI::InrimageROI(Scene2D *scene, const std::list<ROIVoxel *> &vl,
			 Vec3<double> vsize):  ROIVoxelList(vl,vsize),
 Data2D(scene),surface(0)
{
  init();
}
InrimageROI::InrimageROI(Scene2D *scene, 
			 Inrimage *inr, 
			 const Inrimage *bin) 
  throw(InrimageROI::ImageMissing,InrimageROI::BinaryImageWrongSize) :  ROIVoxelList(inr,bin),Data2D(scene), surface(0)
{
  init();
}

InrimageROI::InrimageROI(Scene2D *scene, 
			 Inrimage *inr) 
  throw(InrimageROI::ImageMissing) :  ROIVoxelList(inr),Data2D(scene), surface(0)
{
  init();
}

InrimageROI::InrimageROI(Scene2D *scene, 
			 Inrimage *inr, 
			 Vec3<double> worldPosition,
			 double radius) : ROIVoxelList(),Data2D(scene),surface(0)
{
	voxelSize= Vec3<double> (inr->getVX(),inr->getVY(), inr->getVZ());
	imageSize=Vec3<unsigned int> (inr->getX(),inr->getY(),inr->getZ());
	Vec3<double> voxelPosition,pos;
	inr->map3DWorldToVoxelCoordinates(worldPosition,voxelPosition);
	Vec3<double> dp;
	double r2=radius*radius;
	VoxelIterator it=inr->getVoxelIterator();
	unsigned int x,y,z,index;
	unsigned int dimx=inr->getX();
	unsigned int nplanes=inr->getX()*inr->getY();
	for(;!it.isAtEnd();it++) {
		it.getPosition(x,y,z);
		inr->map3DVoxelToWorldCoordinates(Vec3<double>(x,y,z),pos);
		dp=pos-worldPosition;
		if (dp.norm2()<r2) {
			ROIVoxel *voxel=new ROIVoxel();
			voxel->voxelValue=it();
			voxel->position=Vec3<unsigned int>(x,y,z);
			index=x + y *dimx +z*nplanes;
			voxelMap[index]=voxel;
		}
	}
	init();
}


InrimageROI::InrimageROI(Scene2D *scene, 
			 Inrimage *inr, 
			 SurfaceSliceable *s,
			 InrimageROI::SurfaceRoiMethod method,
			 double param,int selectedZone) : ROIVoxelList(),Data2D(scene),surface(s)
{
  voxelSize= Vec3<double> (inr->getVX(),inr->getVY(), inr->getVZ());
  imageSize=Vec3<unsigned int> (inr->getX(),inr->getY(),inr->getZ());


  Voxel *fg= Voxel::voxelByType(Inrimage::WT_UNSIGNED_CHAR);
  fg->setMax();
  Voxel *bg= Voxel::voxelByType(Inrimage::WT_UNSIGNED_CHAR);
  bg->setValueFromDouble(0);
  Voxel *v= Voxel::voxelByType(inr->getType());
  // set the grey-level image
  Owner<Inrimage>::pointer=inr;
  // slice in the z direction
  std::list<SliceContour *> contours;
  unsigned int i,z;
  int x, y;
  Vec3<double> p1,p2;
  unsigned int index,nplanes,dimx;
  unsigned char ***array;
  dimx=inr->getX();
  nplanes=inr->getX()*inr->getY();
  
  /// perform slice by slice in the z direction
  slice= new Inrimage(inr->getX(), inr->getY(),1,Inrimage::WT_UNSIGNED_CHAR);
  array=(unsigned char ***) slice->getArray();
  memset(&(array[0][0][0]),0,inr->getX()*inr->getY());
  
  Vec3<double> translation=inr->getTranslation();


  for(z = 0; z < inr->getZ(); z++) {

    switch(inr->getGeometry()) {
    case Inrimage::CARTESIAN:
      contours = s->intersectWithPlane(Vec3<double>(0.0, 0.0, 1.0),Vec3<double>(0.0, 0.0,z * inr->getVZ())+translation);

      break;
    default:
      break;
    }
	if (contours.size()>0)
	{
		std::list<SliceContour *>::iterator it;
		Vec3<double> worldPos,imagePos;
		for(it = contours.begin(); it != contours.end(); it++) {
			worldPos=Vec3<double>((*it)->X[0],(*it)->Y[0],(*it)->Z[0]);
			inr->map3DWorldToVoxelCoordinates(worldPos,imagePos);
			p1=imagePos;
			p1[2]=0.0;
			for (i = 1; (int) i < (*it)->nbvertices ; i++) {
				worldPos=Vec3<double>((*it)->X[i],(*it)->Y[i],(*it)->Z[i]);
				inr->map3DWorldToVoxelCoordinates(worldPos,imagePos);
				p2=imagePos;
				p2[2]=0.0;
				if ((selectedZone== -1) || ((selectedZone>=0)&&((*it)->zone[i]==selectedZone))) {
					slice->line3D(p1,p2,*fg,true);
				}
				p1=p2;
			}
			worldPos=Vec3<double>((*it)->X[0],(*it)->Y[0],(*it)->Z[0]);
			inr->map3DWorldToVoxelCoordinates(worldPos,imagePos);
			p2=imagePos;
			p2[2]=0.0;
			if ((selectedZone== -1) || ((selectedZone>=0)&&((*it)->zone[0]==selectedZone))) {
				slice->line3D(p1,p2,*fg,true);
			}
			delete *it;
		}

	switch(method) {
	case SURFACE_REGION:
		//yav::fill(slice,bg,fg,false);
	  fillContours(slice, slice->getType(), NULL, bg,fg,false);
	  break;
	case SURFACE_REGION_OUTSIDE:
	  fillContours(slice, slice->getType(), NULL, bg,fg,false);
	  logicalNot(slice);
	  break;
	case SURFACE_NEAR_BOUNDARY:
	dilation(slice, EIGHT_CONNECTIVITY, (int)param,false);
	break;
	default:
	  break;
	}

	array=(unsigned char ***) slice->getArray();
	for(y=0;y<(int)slice->getY();y++) 
		for(x=0;x<(int)slice->getX();x++) 
		{
			if (array[0][y][x]!=0)
			{
				ROIVoxel *voxel= new ROIVoxel();
				inr->getVoxel(x,y,z,*v);
				voxel->voxelValue=v->getValueAsDouble();
				voxel->position=Vec3<unsigned int>(x,y,z);
				index=x + y *dimx +z*nplanes;
				voxelMap[index]=voxel;
			}
		}
		memset(&(array[0][0][0]),0,inr->getX()*inr->getY());

	}
  }

  delete slice; 
  if (method!=SURFACE_BOUNDARY_Z) {
  /// perform slice by slice in the x direction
  slice= new Inrimage(inr->getY(), inr->getZ(),1,Inrimage::WT_UNSIGNED_CHAR);
  array=(unsigned char ***) slice->getArray();
  memset(&(array[0][0][0]),0,inr->getZ()*inr->getY());


  for(x = 0; x < (int)inr->getX(); x++) {
    switch(inr->getGeometry()) {
    case Inrimage::CARTESIAN:
      contours = s->intersectWithPlane(Vec3<double>(1.0, 0.0, 0.0),
				       Vec3<double>(x * inr->getVX(), 0.0,0.0 )+translation);
      break;
    default:
      break;
    }
	if (contours.size()>0)
	{
		std::list<SliceContour *>::iterator it;
		double vy=inr->getVY();
		double vz=inr->getVZ();
		for(it = contours.begin(); it != contours.end(); it++) {
		  (*it)->X[0]-= (float)translation.getX();
		  (*it)->Y[0]-= (float)translation.getY();
		  (*it)->Z[0]-= (float)translation.getZ();
			p1[2]=p2[2]=0;
			p1[0] = ((*it)->Y[0]/vy );
			p1[1] = ((*it)->Z[0]/vz );
			for (i = 1; (int) i < (*it)->nbvertices ; i++) {
		  (*it)->X[i]-= (float)translation.getX();
		  (*it)->Y[i]-= (float)translation.getY();
		  (*it)->Z[i]-= (float)translation.getZ();

				p2[0] = ((*it)->Y[i]/vy);
				p2[1] = ((*it)->Z[i]/vz);
				if ((selectedZone== -1) || ((selectedZone>=0)&&((*it)->zone[i]==selectedZone))) {
					slice->line3D(p1,p2,*fg,true);
				}
				p1=p2;
			}
			p2[0] = ((*it)->Y[0]/vy);
			p2[1] = ((*it)->Z[0]/vz);
			if ((selectedZone== -1) || ((selectedZone>=0)&&((*it)->zone[0]==selectedZone))) {
				slice->line3D(p1,p2,*fg,true);
			}
			delete *it;
		}
		switch(method) {
	case SURFACE_REGION:
		//yav::fill(slice,bg,fg,false);
		fillContours(slice, slice->getType(), NULL, bg,fg,false);
		break;
	case SURFACE_REGION_OUTSIDE:
		fillContours(slice, slice->getType(), NULL, bg,fg,false);
		logicalNot(slice);
		break;
	case SURFACE_NEAR_BOUNDARY:
		dilation(slice, EIGHT_CONNECTIVITY, (int)param,false);
		break;
	default:
		break;
		}

		array=(unsigned char ***) slice->getArray();
		for(z=0;z<slice->getY();z++) 
			for(y=0;y<(int)slice->getX();y++) 
			{
				if ((array[0][z][y]!=0) && (getVoxel(x,y,z)==0))
				{
					ROIVoxel *voxel= new ROIVoxel();
					inr->getVoxel(x,y,z,*v);
					voxel->voxelValue=v->getValueAsDouble();
					voxel->position=Vec3<unsigned int>(x,y,z);
					index=x + y *dimx +z*nplanes;
					voxelMap[index]=voxel;
				}
			}
			memset(&(array[0][0][0]),0,inr->getZ()*inr->getY());

	}
  }
  }

  init();
}


void InrimageROI::mouseButtonPressed(CameraTypes::WHICH_BUTTON button,
				     CameraTypes::KEY_MODIFIER modifier,
				     Camera2D *cam,int mx,int my)
{
	leftButtonOn=true;

}
void InrimageROI::mouseButtonMoved(Camera2D *cam, 
				   int  mx, int my) 
{
	if (cam->getBackground()) {

		Image3DCameraBackground *b=static_cast<Image3DCameraBackground *> (cam->getBackground());
		if ( (selectionMode==BRUSH_MODE) || (selectionMode==ERASER_MODE)){
			int i,j;
			int bmin=0,bmax=0;
			Vec2<double> pixel(mx,my);
			Vec2<double> pos2D;
			Vec4<double> posVoxel;
			cam->mapScreenTo2DWorldCoordinates(pixel,pos2D);
			b->map2DWorldToImageCoordinates(pos2D,posVoxel);
			unsigned int vx=(unsigned int)posVoxel.getX();
			unsigned int vy=(unsigned int)posVoxel.getY();
			unsigned int vz=(unsigned int)posVoxel.getZ();
			Vec3<unsigned int> p=Vec3<unsigned int>(vx,vy,vz);
			std::list<Vec3<unsigned int> > plist;
			if  (selectionMode==BRUSH_MODE){
				if (brushWidth==1) {
					plist.push_back(p);
				} else {
					bmin=(-(int)brushWidth)+1;
					bmax=(int)brushWidth-1;
				}
			}
			else {
				if (eraserWidth==1) {
					plist.push_back(p);
				} else {
					bmin=(-(int)eraserWidth)+1;
					bmax=(int)eraserWidth-1;
				}
			}
			if (plist.size()==0) {
				if (!strcmp(b->getSliceName(),"sliceZ"))
				{
					for(i=bmin ; i<=bmax;++i) {
						if ((vx+i)>=0) {
							for(j= bmin; j<=bmax;++j) {
								if ((vy+j)>=0) {
									plist.push_back(Vec3<unsigned int> (vx+i,vy+j,vz));
								}
							}
						}
					}
				} else 	if (!strcmp(b->getSliceName(),"sliceY"))
				{
					for(i= bmin; i<=bmax;++i) {
						if ((vx+i)>=0) {
							for(j= bmin; j<=bmax;++j) {
								if ((vz+j)>=0) {
									plist.push_back(Vec3<unsigned int> (vx+i,vy,vz+j));
								}
							}
						}

					}
				}
			 else 	if (!strcmp(b->getSliceName(),"sliceX"))
			{
				for(i= bmin; i<=bmax;++i) {
					if ((vy+i)>=0) {
						for(j= bmin; j<=bmax;++j) {
							if ((vz+j)>=0) {
								plist.push_back(Vec3<unsigned int> (vx,vy+i,vz+j));
							}
						}
					}

				}
			}
		}

		if (selectionMode==BRUSH_MODE) 
			addVoxel(plist);
		else 
			removeVoxel(plist);
	}
	}
}  

void InrimageROI::mouseButtonReleased(CameraTypes::WHICH_BUTTON button,
				      CameraTypes::KEY_MODIFIER modifier,
				      Camera2D *cam,int mx,int my) {
  
}


}
