#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/InrimageView.h"
#ifdef GRAPHICS
#include <inrimage/InrimageView.h>
#ifdef TCL_INTERFACE
#include <graphics2D/TclSliceCamera.h>
#include <inrimage/TclInrimage.h>
#include <inrimage/TclInrimageView.h>
#include <inrimage/TclInrimageBackground.h>
#endif

namespace yav {

InrimageView::InrimageView(Inrimage *inr, Scene2D *scene,bool _createGLScale)
  : Sliceable(scene, true), Owner<Inrimage>(inr),
    createGLScale(_createGLScale) {
  InrimageSliceInterface *any;
  for(any = AutoLink<InrimageSliceInterface>::getHead(); 
      any && (any->getInrimage()!=inr); any = any->following());
  if (any)
    sliceInterface=any;
  else 
    sliceInterface= new InrimageSliceInterface(inr);
  init();
}

InrimageView::InrimageView(Inrimage *inr, Inrimage *i,Scene2D *scene,
			   bool _createGLScale )
  : Sliceable(scene, true), Owner<Inrimage>(inr),
    createGLScale(_createGLScale) {
  InrimageSliceInterface *any;
  for(any = AutoLink<InrimageSliceInterface>::getHead(); 
      any && (any->getInrimage()!=inr); any = any->following());
  if (any)
    sliceInterface=any;
  else 
    sliceInterface= new InrimageSliceInterface(inr);
}
InrimageView::InrimageView(const TclString &filename, Scene2D *scene)
  : Sliceable(scene, true), Owner<Inrimage>(0),
    createGLScale(true) {
  pointer=new Inrimage(filename);
  InrimageSliceInterface *any;
  for(any = AutoLink<InrimageSliceInterface>::getHead(); 
      any && (any->getInrimage()!=pointer()); any = any->following());
  if (any)
    sliceInterface=any;
  else 
    sliceInterface= new InrimageSliceInterface(pointer());

}

InrimageView::~InrimageView() {
  while(!backgroundList.empty()) delete backgroundList.front();
}


void InrimageView::init(void) {
  SliceCamera *slicecam;
  ObjectsList<Camera2D>::iterator it = scene->firstCamera();
  for(; it!=scene->lastCamera(); it++)
    {
      slicecam= dynamic_cast<SliceCamera *>(*it);
      if (slicecam)
	createdCamera(*slicecam);
    }
}

void InrimageView::redrawBackgrounds() {
  std::list<InrimageBackground *>::iterator it = backgroundList.begin();
  for (; it!=backgroundList.end(); ++it){
    (*it)->redraw();
  }
}
void InrimageView::dying(const ObjectPointer<Inrimage>&) {
  std::list<SliceCamera *> tmplist;
  std::list<InrimageBackground *>::iterator it;
  for(it=backgroundList.begin();it!=backgroundList.end();it=backgroundList.begin())
    delete (*it);
  delete this;
}


void InrimageView::createdCamera(SliceCamera& cam) {
  if(cam.getScene() == getScene() && !cam.getBackground())
    doCreateCamera(cam);
}


void InrimageView::doCreateCamera(SliceCamera& cam) {
  cameras[&cam] = new Sliceable::CamInfo();
  double m;

  if(pointer->getVX() < pointer->getVY()) m = pointer->getVX();
  else m = pointer->getVY();
  if(pointer->getVZ() < m) m = pointer->getVZ();
  //  cam.scaleCamera(1.0 / (m * cam.getScale()));
  // cam.scaleCamera(1.0 / (m));
  switch(cam.getGeometry()) {
  case SliceCamera::SLICE_X:
    cam.setThickness(pointer->getVX());
    break;
  case SliceCamera::SLICE_Y:
    cam.setThickness(pointer->getVY());
    break;
  case SliceCamera::SLICE_Z:
  case SliceCamera::SLICE_ALPHA:
    cam.setThickness(pointer->getVZ());
    break;
  }
#ifdef TCL_INTERFACE
  TclInrimageBackground *b;
  if (createGLScale)
    b = new TclInrimageBackground(cam.getInterface()->getModule(),this,&cam);
  else
    b = new TclInrimageBackground(getInterface()->getModule(),this,&cam,false,0,false,0);
  backgroundList.push_back(b);  
#endif
  cam.focusOnBackground();
}

InrimageBackground *InrimageView::getBackgroundFromCamera(SliceCamera *cam)
{
  std::list<InrimageBackground *>::iterator it;
  for(it=backgroundList.begin();(it!=backgroundList.end()) && ((*it)->getCamera()!=cam);it++);
  if (it==backgroundList.end())
    return (0);
  else
    return (*it);
}

void InrimageView::deletedCamera(SliceCamera& cam) {
  std::list<InrimageBackground *>::iterator it;
  for(it=backgroundList.begin();(it!=backgroundList.end()) && ((*it)->getCamera()!=&cam);it++);
  if (it!=backgroundList.end())
    {
      cam.setBackground(0);
      InrimageBackground *b=(*it);
      // avoid calling background dying
      b->slicecam=0;
      delete b;
    }
}


void InrimageView::changeInrimage(Inrimage *inr) {
  if (inr!=pointer()) {
    pointer=inr;
    std::list<InrimageBackground *>::iterator it;
    for(it = backgroundList.begin(); it != backgroundList.end(); it++) {
      (*it)->inrimageChanged();
    }
  }
  InrimageSliceInterface *any;
  for(any = AutoLink<InrimageSliceInterface>::getHead(); 
      any && (any->getInrimage()!=(Inrimage *)inr); any = any->following());
  if (any)
    sliceInterface=any;
  else 
    sliceInterface= new InrimageSliceInterface(inr);
}


void InrimageView::refreshCameras(void) {
  std::list<InrimageBackground *>::iterator it;
  for(it = backgroundList.begin(); it != backgroundList.end(); it++) {
    (*it)->refreshCamera();
  }
}

void InrimageView::redraw(Camera2D *cam) {
  std::map<SliceCamera *, CamInfo *>::iterator it =
    cameras.find((SliceCamera *) cam);
  if(it != cameras.end() && (it->second->needRedraw ||
			     redrawFlag == Data2D::UPDATE ||
			     redrawFlag == Data2D::CREATION)) {
    InrimageBackground *img = getBackgroundFromCamera(it->first);
    if(img) {
      img->redraw();
    }
    it->second->needRedraw = false;
  }
}




Inrimage *InrimageView::getInrimage(void) const {
  return((Inrimage *)pointer()); }

#ifdef TCL_INTERFACE
TclInrimageView *InrimageView::getInterface(void) const {
	return (TclInrimageView *) Client::getInterface();
}
#endif


} // namespace yav


#endif
