#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/Interpolator.h"
#include <inrimage/Interpolator.h>
#include <inrimage/Inrimage.h>
#include <inrimage/Voxel.h>
#include <inrimage/ImageGeometry.h>

namespace yav {


/**********************************************************************/
/**********************************************************************/
/*************************   Interpolator  ****************************/
/**********************************************************************/
/**********************************************************************/


Interpolator *
Interpolator::trilinearInterpolatorByType(Inrimage::GEOMETRY geom,
					  Inrimage::WORD_TYPE t, int vdim){
  switch(geom) {
    
  case Inrimage::CARTESIAN:
    switch(vdim) {
    case 1:
      switch(t) {
      case Inrimage::WT_UNSIGNED_CHAR:
	return new TypedTrilinInterpolator<Voxel8u, unsigned char>();
      case Inrimage::WT_UNSIGNED_SHORT:
	return new TypedTrilinInterpolator<Voxel16u, unsigned short>();
      case Inrimage::WT_SIGNED_SHORT:
	return new TypedTrilinInterpolator<Voxel16s, short>();
      case Inrimage::WT_UNSIGNED_INT:
	return new TypedTrilinInterpolator<Voxel32u, unsigned int>();
      case Inrimage::WT_SIGNED_INT:
	return new TypedTrilinInterpolator<Voxel32s, int>();
      case Inrimage::WT_UNSIGNED_LONG:
	return new TypedTrilinInterpolator<Voxel64u, unsigned LONGINT>();
      case Inrimage::WT_SIGNED_LONG:
	return new TypedTrilinInterpolator<Voxel64s, LONGINT>();
      default:
	return new TrilinInterpolator();
      }
    case 2:
      switch(t) {
      case Inrimage::WT_FLOAT_VECTOR:
//	return new TypedTrilinVectInterpolator<VoxelFV, float, 2>();
	return new TrilinVectInterpolator();
      case Inrimage::WT_DOUBLE_VECTOR:
	return new TrilinVectInterpolator();
//	return new TypedTrilinVectInterpolator<VoxelDV, double, 2>();
      default:
	return new TrilinVectInterpolator();
      }
    case 3:
      switch(t) {
      case Inrimage::WT_FLOAT_VECTOR:
//	return new TypedTrilinVectInterpolator<VoxelFV, float, 3>();
	return new TrilinVectInterpolator();
      case Inrimage::WT_DOUBLE_VECTOR:
//	return new TypedTrilinVectInterpolator<VoxelDV, double, 3>();
	return new TrilinVectInterpolator();
      default:
	return new TrilinVectInterpolator();
      }
     default:
      return new TrilinVectInterpolator();
    }

  case Inrimage::CYLINDRICAL:
    return new CylTrilinInterpolator();

  case Inrimage::SPHERICAL:
    return new SphTrilinInterpolator();
  }

  return new TrilinInterpolator();
}



/**********************************************************************/
/**********************************************************************/
/*********************   NearestInterpolator  *************************/
/**********************************************************************/
/**********************************************************************/


Voxel& NearestInterpolator::operator()(Inrimage const& img, double x,
				       double y, double z, Voxel& v) const{
//   return img.getVoxel((unsigned int)(x / img.getVX() + 0.5),
// 		      (unsigned int)(y / img.getVY() + 0.5),
// 		      (unsigned int)(z / img.getVZ() + 0.5), v);
  return img.getVoxel((unsigned int)(x / img.getVX()),
		      (unsigned int)(y / img.getVY()),
		      (unsigned int)(z / img.getVZ()), v);
}

Interpolator* NearestInterpolator::clone() const{
  return new NearestInterpolator(*this);
}



/**********************************************************************/
/**********************************************************************/
/**********************   ToZeroExtrapolator **************************/
/**********************************************************************/
/**********************************************************************/


Voxel& ToZeroExtrapolator::operator()(Inrimage const& ,
				      double , double , double ,
				      Voxel& v) const{
  v.setValueFromDouble(0.0);
  return v;
}

Extrapolator* ToZeroExtrapolator::clone() const{
  return new ToZeroExtrapolator(*this);
}

/**********************************************************************/
/**********************************************************************/
/*********************   NearestExtrapolator **************************/
/**********************************************************************/
/**********************************************************************/


Voxel& NearestExtrapolator::operator()(Inrimage const& img,
				       double x, double y, double z,
				       Voxel& v) const{
  if(x<0) x = 0;
  else if(x >= img.getX()) x = img.getX()-1;
  if(y<0) y = 0;
  else if(y >= img.getY()) y = img.getY()-1;
  if(z<0) z = 0;
  else if(z >= img.getZ()) z = img.getZ()-1;
  return img.getVoxel((unsigned int)(x), (unsigned int)(y),
		      (unsigned int)(z), v);
}


Extrapolator* NearestExtrapolator::clone() const{
   return new NearestExtrapolator(*this);
}

/**********************************************************************/
/**********************************************************************/
/***********************   TrilinInterpolator *************************/
/**********************************************************************/
/**********************************************************************/

Voxel& TrilinInterpolator::operator()(Inrimage const& img,double x,double y,double z,Voxel& v) const{

  // here, the point lies on the borders or completely inside the image
  x /= img.getVX();
  y /= img.getVY();
  z /= img.getVZ();

  int  const dimx1 = img.getX()-1;
  int  const dimy1 = img.getY()-1;
  int  const dimz1 = img.getZ()-1;

  double dx,dy,dz;

  int  ix =  (int) (x);
  int  iy =  (int) (y);
  int  iz =  (int) (z);
  int ix1 = ix+1;
  int iy1 = iy+1;
  int iz1 = iz+1;
  // are we on the border or not ?
  if ( (x >= 0) && (ix < dimx1) && (y >= 0) && (iy < dimy1) && (z >= 0) && (iz < dimz1) )  
  {  
	  // the corresponding point is in the box defined by (ix[+1],iy[+1],iz[+1])
	  dx = x - ix;
	  dy = y - iy;
	  dz = z - iz;
  } else {
	  if (x<0)    
	  { 
		  ix=ix1=0;
		  dx=0;
	  }
	  else if (ix>=dimx1) {
		  ix1=dimx1;
		  dx=0;
	  }
	  else 
		  dx = x - ix;      

	  if (y<0)    
	  { 
		  iy=iy1=0;
		  dy=0;
	  }
	  else if (iy>=dimy1) {
		  iy1=dimx1;
		  dy=0;
	  }
	  else 
		  dy = y - iy;      

	  if (z<0)    
	  { 
		  iz=iz1=0;
		  dz=0;
	  }
	  else if (iz>=dimz1) {
		  iz1=dimz1;
		  dz=0;
	  }
	  else 
		  dz = z - iz;      
  }
  double const dxdy = dx*dy;
  double const dxdz = dx*dz;
  double const dydz = dy*dz;
  double const dxdydz = dxdy*dz;

  // we have
  // v[7]=dxdydz;            coefficient of (ix+1,iy+1,iz+1)
  // v[6]=dxdz-dxdydz;       coefficient of (ix+1,iy,  iz+1)
  // v[5]=dxdy-dxdydz;       coefficient of (ix+1,iy+1,iz  )
  // v[4]=dx-dxdy-v[6];      coefficient of (ix+1,iy  ,iz  )
  // v[3]=dydz-dxdydz;       coefficient of (ix  ,iy+1,iz+1)
  // v[2]=dz-dydz-v[6];      coefficient of (ix  ,iy  ,iz+1)
  // v[1]=dy-dydz-v[5];      coefficient of (ix  ,iy+1,iz  )
  // v[0]=1-dy-dz+dydz-v[4]; coefficient of (ix  ,iy  ,iz  )

  double const v6 = dxdz-dxdydz;
  double const v5 = dxdy-dxdydz;
  double const v4 = dx-dxdy-v6;


  img.getVoxel(ix1,iy1,iz1,v);
  double res = dxdydz * v.getValueAsDouble();  // (ix+1,iy+1,iz+1)
  img.getVoxel(ix,iy1,iz1,v);
  res += (dydz-dxdydz) *v.getValueAsDouble();  // (ix  ,iy+1,iz+1) 
  img.getVoxel(ix1,iy,iz1,v);
  res += v6 *v.getValueAsDouble();		   // (ix+1,iy,  iz+1) 
  img.getVoxel(ix,iy,iz1,v);
  res += (dz-dydz-v6) *v.getValueAsDouble();   // (ix  ,iy  ,iz+1) 
  img.getVoxel(ix1,iy1,iz,v);
  res += v5 *v.getValueAsDouble();		   // (ix+1,iy+1,iz  ) 
  img.getVoxel(ix,iy1,iz,v);
  res += (dy-dydz-v5) * v.getValueAsDouble();  // (ix  ,iy+1,iz  ) 
  img.getVoxel(ix1,iy,iz,v);
  res += v4 * v.getValueAsDouble();		   // (ix+1,iy  ,iz  ) 
  img.getVoxel(ix,iy,iz,v);
  v.setValueFromDouble(res+(1-dy-dz+dydz-v4)*v.getValueAsDouble()); // (ix  ,iy  ,iz  ) 



#ifdef old
  double  const dimx1 = img.getX()-0.5;
  double  const dimy1 = img.getY()-0.5;
  double  const dimz1 = img.getZ()-0.5;

  int  ix =  (int) (x-0.5);
  int  iy =  (int) (y-0.5);
  int  iz =  (int) (z-0.5);
  int ix1 = ix+1;
  int iy1 = iy+1;
  int iz1 = iz+1;
  // are we on the border or not ?
  if ( (x > 0.5) && (ix < dimx1) && (y > 0.5) && (iy < dimy1) && (z > 0.5) && (iz < dimz1) )  
    {  
      // the corresponding point is in the box defined by (ix[+1],iy[+1],iz[+1])
      double const dx = x - ix-0.5;
      double const dy = y - iy-0.5;
      double const dz = z - iz-0.5;
      double const dxdy = dx*dy;
      double const dxdz = dx*dz;
      double const dydz = dy*dz;
      double const dxdydz = dxdy*dz;
      
      // we have
      // v[7]=dxdydz;            coefficient of (ix+1,iy+1,iz+1)
      // v[6]=dxdz-dxdydz;       coefficient of (ix+1,iy,  iz+1)
      // v[5]=dxdy-dxdydz;       coefficient of (ix+1,iy+1,iz  )
      // v[4]=dx-dxdy-v[6];      coefficient of (ix+1,iy  ,iz  )
      // v[3]=dydz-dxdydz;       coefficient of (ix  ,iy+1,iz+1)
      // v[2]=dz-dydz-v[6];      coefficient of (ix  ,iy  ,iz+1)
      // v[1]=dy-dydz-v[5];      coefficient of (ix  ,iy+1,iz  )
      // v[0]=1-dy-dz+dydz-v[4]; coefficient of (ix  ,iy  ,iz  )
      
      double const v6 = dxdz-dxdydz;
      double const v5 = dxdy-dxdydz;
      double const v4 = dx-dxdy-v6;

      
      img.getVoxel(ix1,iy1,iz1,v);
      double res = dxdydz * v.getValueAsDouble();  // (ix+1,iy+1,iz+1)
      img.getVoxel(ix,iy1,iz1,v);
      res += (dydz-dxdydz) *v.getValueAsDouble();  // (ix  ,iy+1,iz+1) 
      img.getVoxel(ix1,iy,iz1,v);
      res += v6 *v.getValueAsDouble();		   // (ix+1,iy,  iz+1) 
      img.getVoxel(ix,iy,iz1,v);
      res += (dz-dydz-v6) *v.getValueAsDouble();   // (ix  ,iy  ,iz+1) 
      img.getVoxel(ix1,iy1,iz,v);
      res += v5 *v.getValueAsDouble();		   // (ix+1,iy+1,iz  ) 
      img.getVoxel(ix,iy1,iz,v);
      res += (dy-dydz-v5) * v.getValueAsDouble();  // (ix  ,iy+1,iz  ) 
      img.getVoxel(ix1,iy,iz,v);
      res += v4 * v.getValueAsDouble();		   // (ix+1,iy  ,iz  ) 
      img.getVoxel(ix,iy,iz,v);
      v.setValueFromDouble(res+(1-dy-dz+dydz-v4)*v.getValueAsDouble()); // (ix  ,iy  ,iz  ) 
    } 
  else {
    int const maxx = img.getX()-1;
    int const maxy = img.getY()-1;
    int const maxz = img.getZ()-1;
    double dx,dy,dz;

    if (ix<0)    
      { 
	ix=ix1=0;
	dx=0;
      }
    else if (ix1>maxx) {
      ix1=maxx;
      dx=0;
    }
    else 
      dx = x - ix-0.5;      

    if (iy<0)    
      { 
	iy=iy1=0;
	dy=0;
      }
    else if (iy1>maxy) {
      iy1=maxy;
      dy=0;
    }
    else 
      dy = y - iy-0.5;      

    if (iz<0)    
      { 
	iz=iz1=0;
	dz=0;
      }
    else if (iz1>maxz) {
      iz1=maxz;
      dz=0;
    }
    else 
      dz = z - iz-0.5;      

    double const dxdy = dx*dy;
    double const dxdz = dx*dz;
    double const dydz = dy*dz;
    double const dxdydz = dxdy*dz;
    double const v6 = dxdz-dxdydz;
    double const v5 = dxdy-dxdydz;
    double const v4 = dx-dxdy-v6;
    
    img.getVoxel(ix1,iy1,iz1,v);
    double res = dxdydz * v.getValueAsDouble();                       // (ix+1,iy+1,iz+1)
    img.getVoxel(ix,iy1,iz1,v);
    res += (dydz-dxdydz) *v.getValueAsDouble();                       // (ix  ,iy+1,iz+1) 
    img.getVoxel(ix1,iy,iz1,v);
    res += v6 *v.getValueAsDouble();		                        // (ix+1,iy,  iz+1) 
    img.getVoxel(ix,iy,iz1,v);
    res += (dz-dydz-v6) *v.getValueAsDouble();                        // (ix  ,iy  ,iz+1) 
    img.getVoxel(ix1,iy1,iz,v);
    res += v5 *v.getValueAsDouble();		                        // (ix+1,iy+1,iz  ) 
    img.getVoxel(ix,iy1,iz,v);
    res += (dy-dydz-v5) * v.getValueAsDouble();                       // (ix  ,iy+1,iz  ) 
    img.getVoxel(ix1,iy,iz,v);
    res += v4 * v.getValueAsDouble();		                        // (ix+1,iy  ,iz  ) 
    img.getVoxel(ix,iy,iz,v);
    v.setValueFromDouble(res+(1-dy-dz+dydz-v4)*v.getValueAsDouble()); // (ix  ,iy  ,iz  ) 
    
    // OLD CODE
    //     double res = dxdydz * tpt[iz1][iy1][ix1];	               // (ix+1,iy+1,iz+1)
    //     res += (dydz-dxdydz) * tpt[iz1][iy1][ix];	               // (ix  ,iy+1,iz+1) 
    //     res += v6 * tpt[iz1][iy][ix1];			       // (ix+1,iy,  iz+1) 
    //     res += (dz-dydz-v6) * tpt[iz1][iy][ix];		       // (ix  ,iy  ,iz+1) 
    //     res += v5 * tpt[iz][iy1][ix1];			       // (ix+1,iy+1,iz  ) 
    //     res += (dy-dydz-v5) * tpt[iz][iy1][ix];		       // (ix  ,iy+1,iz  ) 
    //     res += v4 * tpt[iz][iy][ix1];			       // (ix+1,iy  ,iz  ) 
    //    v.setValueFromDouble(res+(1-dy-dz+dydz-v4)*tpt[iz][iy][ix]); // (ix  ,iy  ,iz  ) 
  }
#endif
  return v;
}

Interpolator* TrilinInterpolator::clone() const{
   return new TrilinInterpolator(*this);
}

TrilinInterpolator* TrilinInterpolator::interpolatorByType(Inrimage::WORD_TYPE t){
  switch(t) {
  case Inrimage::WT_UNSIGNED_CHAR:
    return new TypedTrilinInterpolator<Voxel8u,unsigned char>();
  case Inrimage::WT_UNSIGNED_SHORT:
    return new TypedTrilinInterpolator<Voxel16u,unsigned short>();
  case Inrimage::WT_SIGNED_SHORT:
    return new TypedTrilinInterpolator<Voxel16s,short>();
  case Inrimage::WT_UNSIGNED_INT:
    return new TypedTrilinInterpolator<Voxel32u,unsigned int>();
  case Inrimage::WT_SIGNED_INT:
    return new TypedTrilinInterpolator<Voxel32s,int>();
  case Inrimage::WT_UNSIGNED_LONG:
    return new TypedTrilinInterpolator<Voxel64u,unsigned LONGINT>();
  case Inrimage::WT_SIGNED_LONG:
    return new TypedTrilinInterpolator<Voxel64s,LONGINT>();
  default: return new TrilinInterpolator();
  }
}

/**********************************************************************/
/**********************************************************************/
/********************   TypedTrilinInterpolator ***********************/
/**********************************************************************/
/**********************************************************************/

template<class V,typename T>
T& TypedTrilinInterpolator<V,T>::operator()(Inrimage const& img,
					    double x,double y,double z,
					    T& v) const{
  // here, the point lies on the borders or completely inside the image
  x /= img.getVX();
  y /= img.getVY();
  z /= img.getVZ();

  int  const dimx1 = img.getX()-1;
  int  const dimy1 = img.getY()-1;
  int  const dimz1 = img.getZ()-1;

  T const* const* const* tpt =
    reinterpret_cast<T const* const* const*>(img.getArray());
  double dx,dy,dz;

  int  ix =  (int) (x);
  int  iy =  (int) (y);
  int  iz =  (int) (z);
  int ix1 = ix+1;
  int iy1 = iy+1;
  int iz1 = iz+1;
  // are we on the border or not ?
  if ( (x >= 0) && (ix < dimx1) && (y >= 0) && (iy < dimy1) && (z >= 0) && (iz < dimz1) )  
  {  
	  // the corresponding point is in the box defined by (ix[+1],iy[+1],iz[+1])
	  dx = x - ix;
	  dy = y - iy;
	  dz = z - iz;
  } else {
	  if (x<0)    
	  { 
		  ix=ix1=0;
		  dx=0;
	  }
	  else if (ix>=dimx1) {
		  ix1=dimx1;
		  dx=0;
	  }
	  else 
		  dx = x - ix;      

	  if (y<0)    
	  { 
		  iy=iy1=0;
		  dy=0;
	  }
	  else if (iy>=dimy1) {
		  iy1=dimx1;
		  dy=0;
	  }
	  else 
		  dy = y - iy;      

	  if (z<0)    
	  { 
		  iz=iz1=0;
		  dz=0;
	  }
	  else if (iz>=dimz1) {
		  iz1=dimz1;
		  dz=0;
	  }
	  else 
		  dz = z - iz;      
  }

							

  double const dxdy = dx*dy;
  double const dxdz = dx*dz;
  double const dydz = dy*dz;
  double const dxdydz = dxdy*dz;

  // we have
  // v[7]=dxdydz;            coefficient of (ix+1,iy+1,iz+1)
  // v[6]=dxdz-dxdydz;       coefficient of (ix+1,iy,  iz+1)
  // v[5]=dxdy-dxdydz;       coefficient of (ix+1,iy+1,iz  )
  // v[4]=dx-dxdy-v[6];      coefficient of (ix+1,iy  ,iz  )
  // v[3]=dydz-dxdydz;       coefficient of (ix  ,iy+1,iz+1)
  // v[2]=dz-dydz-v[6];      coefficient of (ix  ,iy  ,iz+1)
  // v[1]=dy-dydz-v[5];      coefficient of (ix  ,iy+1,iz  )
  // v[0]=1-dy-dz+dydz-v[4]; coefficient of (ix  ,iy  ,iz  )

  double const v6 = dxdz-dxdydz;
  double const v5 = dxdy-dxdydz;
  double const v4 = dx-dxdy-v6;

  double res = dxdydz * tpt[iz1][iy1][ix1];		       // (ix+1,iy+1,iz+1)
  res += (dydz-dxdydz) * tpt[iz1][iy1][ix];		       // (ix  ,iy+1,iz+1) 
  res += v6 * tpt[iz1][iy][ix1];			       // (ix+1,iy,  iz+1) 
  res += (dz-dydz-v6) * tpt[iz1][iy][ix];		       // (ix  ,iy  ,iz+1) 
  res += v5 * tpt[iz][iy1][ix1];			       // (ix+1,iy+1,iz  ) 
  res += (dy-dydz-v5) * tpt[iz][iy1][ix];		       // (ix  ,iy+1,iz  ) 
  res += v4 * tpt[iz][iy][ix1];		                 // (ix+1,iy  ,iz  ) 
  res+= (1-dy-dz+dydz-v4)*tpt[iz][iy][ix];
  v = castFromDouble<T>(res); // (ix  ,iy  ,iz  ) 
  return v;
}
  
template<class V,typename T>
Interpolator* TypedTrilinInterpolator<V,T>::clone() const{
  return new TypedTrilinInterpolator<V,T>(*this);
}


/**********************************************************************/
/**********************************************************************/
/*********************   TrilinVectInterpolator ***********************/
/**********************************************************************/
/**********************************************************************/

Voxel& TrilinVectInterpolator::operator()(Inrimage const& img,
					  double x, double y, double z,
					  Voxel& v) const{
 // here, the point lies on the borders or completely inside the image
  x /= img.getVX();
  y /= img.getVY();
  z /= img.getVZ();

  int  const dimx1 = img.getX()-1;
  int  const dimy1 = img.getY()-1;
  int  const dimz1 = img.getZ()-1;


  Voxel* val=Voxel::voxelByType(v.getType(),img.getVdim());
  v.setValueFromDouble(0.0);

  double dx,dy,dz;

  int  ix =  (int) (x);
  int  iy =  (int) (y);
  int  iz =  (int) (z);
  int ix1 = ix+1;
  int iy1 = iy+1;
  int iz1 = iz+1;
  // are we on the border or not ?
  if ( (x >= 0) && (ix < dimx1) && (y >= 0) && (iy < dimy1) && (z >= 0) && (iz < dimz1) )  
  {  
	  // the corresponding point is in the box defined by (ix[+1],iy[+1],iz[+1])
	  dx = x - ix;
	  dy = y - iy;
	  dz = z - iz;
  } else {
	  if (x<0)    
	  { 
		  ix=ix1=0;
		  dx=0;
	  }
	  else if (ix>=dimx1) {
		  ix1=dimx1;
		  dx=0;
	  }
	  else 
		  dx = x - ix;      

	  if (y<0)    
	  { 
		  iy=iy1=0;
		  dy=0;
	  }
	  else if (iy>=dimy1) {
		  iy1=dimx1;
		  dy=0;
	  }
	  else 
		  dy = y - iy;      

	  if (z<0)    
	  { 
		  iz=iz1=0;
		  dz=0;
	  }
	  else if (iz>=dimz1) {
		  iz1=dimz1;
		  dz=0;
	  }
	  else 
		  dz = z - iz;      
  } 

  double const dxdy = dx*dy;
  double const dxdz = dx*dz;
  double const dydz = dy*dz;
  double const dxdydz = dxdy*dz;

  // we have
  // v[7]=dxdydz;            coefficient of (ix+1,iy+1,iz+1)
  // v[6]=dxdz-dxdydz;       coefficient of (ix+1,iy,  iz+1)
  // v[5]=dxdy-dxdydz;       coefficient of (ix+1,iy+1,iz  )
  // v[4]=dx-dxdy-v[6];      coefficient of (ix+1,iy  ,iz  )
  // v[3]=dydz-dxdydz;       coefficient of (ix  ,iy+1,iz+1)
  // v[2]=dz-dydz-v[6];      coefficient of (ix  ,iy  ,iz+1)
  // v[1]=dy-dydz-v[5];      coefficient of (ix  ,iy+1,iz  )
  // v[0]=1-dy-dz+dydz-v[4]; coefficient of (ix  ,iy  ,iz  )

  double const v6 = dxdz-dxdydz;
  double const v5 = dxdy-dxdydz;
  double const v4 = dx-dxdy-v6;

  // (ix+1,iy+1,iz+1)
  img.getVoxel(ix1,iy1,iz1,*val);
  addMulVoxDbl(v,*val,dxdydz);
  // (ix  ,iy+1,iz+1) 
  img.getVoxel(ix,iy1,iz1,*val);
  addMulVoxDbl(v,*val,dydz-dxdydz);
  // (ix+1  ,iy,  iz+1) 
  img.getVoxel(ix1,iy,iz1,*val);
  addMulVoxDbl(v,*val,v6);
  // (ix  ,iy  ,iz+1) 
  img.getVoxel(ix,iy,iz1,*val);
  addMulVoxDbl(v,*val,dz-dydz-v6);
  // (ix+1,iy+1,iz  ) 
  img.getVoxel(ix1,iy1,iz,*val);
  addMulVoxDbl(v,*val,v5);
  // (ix  ,iy+1,iz  ) 
  img.getVoxel(ix,iy1,iz,*val);
  addMulVoxDbl(v,*val,dy-dydz-v5);
  // (ix+1,iy  ,iz  ) 
  img.getVoxel(ix1,iy,iz,*val);
  addMulVoxDbl(v,*val,v4);
  // (ix  ,iy  ,iz  ) 
  img.getVoxel(ix,iy,iz,*val);
  v = addMulVoxDbl(v,*val,1-dy-dz+dydz-v4);
  delete val;
  return v;

}

Interpolator* TrilinVectInterpolator::clone() const{
  return new TrilinVectInterpolator(*this);
}

TrilinVectInterpolator *
TrilinVectInterpolator::interpolatorByType(Inrimage::WORD_TYPE t,size_t v){
  switch(t) {
  case Inrimage::WT_FLOAT_VECTOR:
    switch(v){
    case 2: return new TypedTrilinVectInterpolator<VoxelFV,float,2>();
    case 3: return new TypedTrilinVectInterpolator<VoxelFV,float,3>();
    default: return new TrilinVectInterpolator();
    }
  default: return new TrilinVectInterpolator();
  }
}

Voxel& TrilinVectInterpolator::addMulVoxDbl(Voxel& r,Voxel const& v,double d){
  switch(r.getType()){
  case Inrimage::WT_FLOAT_VECTOR:{
    VoxelFV& rfv=static_cast<VoxelFV&>(r);
    VoxelFV const& vfv=static_cast<VoxelFV const&>(v);
    for(size_t i=0;i<rfv.getVdim();++i) rfv[i]+=vfv[i]*(float)d;
    break;
  }
  case Inrimage::WT_DOUBLE_VECTOR:{
    VoxelDV& rdv=static_cast<VoxelDV&>(r);
    VoxelDV const& vdv=static_cast<VoxelDV const&>(v);
    for(size_t i=0;i<rdv.getVdim();++i) rdv[i]+=vdv[i]*d;
    break;
  }
  case Inrimage::WT_RGB:{
    unsigned char *rv=static_cast<VoxelRGB&>(r).getValue();
    unsigned char const*vv=static_cast<VoxelRGB const&>(v).getValue();
    for(size_t i=0;i<3;++i) rv[i]+=(unsigned char)(vv[i]*d+0.5);
    break;
  }
  case Inrimage::WT_RGBA:{
    unsigned char *rv=static_cast<VoxelRGBA&>(r).getValue();
    unsigned char const*vv=static_cast<VoxelRGBA const&>(v).getValue();
    for(size_t i=0;i<4;++i) rv[i]+=(unsigned char)(vv[i]*d+0.5);
    break;
  }
  default: throw VoxelNotAVector("Voxel& TrilinVectInterpolator::addMulVoxDbl(Voxel& r,Voxel const& v,double d)");
  }
  return r;
}
   
template<class V,typename T,size_t D>
T* TypedTrilinVectInterpolator<V,T,D>::operator()(Inrimage const& img,
						  double x, double y,double z,
						  T v[D]) const{
//  T const* const* const* const* tpt=reinterpret_cast<T const* const* const*
//  const*>(img.getVarray());
  T ****tpt=(T ****) img.getVarray();
  size_t i;

 // here, the point lies on the borders or completely inside the image
  x /= img.getVX();
  y /= img.getVY();
  z /= img.getVZ();

  int  const dimx1 = img.getX()-1;
  int  const dimy1 = img.getY()-1;
  int  const dimz1 = img.getZ()-1;
  double dx,dy,dz;

  int  ix =  (int) (x);
  int  iy =  (int) (y);
  int  iz =  (int) (z);
  int ix1 = ix+1;
  int iy1 = iy+1;
  int iz1 = iz+1;
  // are we on the border or not ?
  if ( (x >= 0) && (ix < dimx1) && (y >= 0) && (iy < dimy1) && (z >= 0) && (iz < dimz1) )  
  {  
	  // the corresponding point is in the box defined by (ix[+1],iy[+1],iz[+1])
	  dx = x - ix;
	  dy = y - iy;
	  dz = z - iz;
  } else {
	  if (x<0)    
	  { 
		  ix=ix1=0;
		  dx=0;
	  }
	  else if (ix>=dimx1) {
		  ix1=dimx1;
		  dx=0;
	  }
	  else 
		  dx = x - ix;      

	  if (y<0)    
	  { 
		  iy=iy1=0;
		  dy=0;
	  }
	  else if (iy>=dimy1) {
		  iy1=dimx1;
		  dy=0;
	  }
	  else 
		  dy = y - iy;      

	  if (z<0)    
	  { 
		  iz=iz1=0;
		  dz=0;
	  }
	  else if (iz>=dimz1) {
		  iz1=dimz1;
		  dz=0;
	  }
	  else 
		  dz = z - iz;      
  } 

  double const dxdy = dx*dy;
  double const dxdz = dx*dz;
  double const dydz = dy*dz;
  double const dxdydz = dxdy*dz;

  // we have
  // v[7]=dxdydz;            coefficient of (ix+1,iy+1,iz+1)
  // v[6]=dxdz-dxdydz;       coefficient of (ix+1,iy,  iz+1)
  // v[5]=dxdy-dxdydz;       coefficient of (ix+1,iy+1,iz  )
  // v[4]=dx-dxdy-v[6];      coefficient of (ix+1,iy  ,iz  )
  // v[3]=dydz-dxdydz;       coefficient of (ix  ,iy+1,iz+1)
  // v[2]=dz-dydz-v[6];      coefficient of (ix  ,iy  ,iz+1)
  // v[1]=dy-dydz-v[5];      coefficient of (ix  ,iy+1,iz  )
  // v[0]=1-dy-dz+dydz-v[4]; coefficient of (ix  ,iy  ,iz  )

  double const v6 = dxdz-dxdydz;
  double const v5 = dxdy-dxdydz;
  double const v4 = dx-dxdy-v6;

  for(i=0;i<D;++i) v[i]=(T)(dxdydz * tpt[i][iz1][iy1][ix1]);	   // (ix+1,iy+1,iz+1)
  for(i=0;i<D;++i) v[i]+=(T)((dydz-dxdydz) * tpt[i][iz1][iy1][ix]); // (ix  ,iy+1,iz+1) 
  for(i=0;i<D;++i) v[i]+=(T)(v6 * tpt[i][iz1][iy][ix1]);		   // (ix+1,iy,  iz+1) 
  for(i=0;i<D;++i) v[i]+=(T)((dz-dydz-v6) * tpt[i][iz1][iy][ix]);   // (ix  ,iy  ,iz+1) 
  for(i=0;i<D;++i) v[i]+=(T)(v5 * tpt[i][iz][iy1][ix1]);		   // (ix+1,iy+1,iz  ) 
  for(i=0;i<D;++i) v[i]+=(T)((dy-dydz-v5) * tpt[i][iz][iy1][ix]);   // (ix  ,iy+1,iz  ) 
  for(i=0;i<D;++i) v[i]+=(T)(v4 * tpt[i][iz][iy][ix1]);		   // (ix+1,iy  ,iz  ) 
  for(i=0;i<D;++i) v[i]+=(T)((1-dy-dz+dydz-v4)*tpt[i][iz][iy][ix]); // (ix  ,iy,iz  )
  return v;
 
}
  
template<class V,typename T,size_t D>
Interpolator* TypedTrilinVectInterpolator<V,T,D>::clone() const{
   return new TypedTrilinVectInterpolator<V,T,D>(*this);
}

/**********************************************************************/
/**********************************************************************/
/*********************   CylTrilinInterpolator ************************/
/**********************************************************************/
/**********************************************************************/

Voxel& CylTrilinInterpolator::operator()(Inrimage const& img,
					 double x, double y, double z,
					 Voxel& v) const{
  CylindricalGeometry converter(img);
  converter.set(x, y, z);
  v.setValueFromDouble(converter.interpolate());
  return v;
}

Interpolator* CylTrilinInterpolator::clone() const{
  return new CylTrilinInterpolator(*this);
}


/**********************************************************************/
/**********************************************************************/
/*********************   SphTrilinInterpolator ************************/
/**********************************************************************/
/**********************************************************************/

Voxel& SphTrilinInterpolator::operator()(Inrimage const& img,
					 double x, double y, double z,
					 Voxel& v) const{
  SphericalGeometry converter(img);
  converter.set(x, y, z);
  v.setValueFromDouble(converter.interpolate());
  return v;
}

Interpolator* SphTrilinInterpolator::clone() const{
  return new SphTrilinInterpolator(*this);
}

} // namespace yav
