#ifdef _MSC_VER
#pragma warning ( disable : 4068 )
#pragma warning ( disable : 4786 )
#pragma warning ( disable : 4081 )
#endif


#include <inrimage/ImageLandmarkSet.h>


namespace yav {


LandmarkSet::LandmarkSet() 
{
}
LandmarkSet::LandmarkSet(const char *filename) 
{
	IFSTREAM file(filename);

	unsigned int nbLandmarks,i;
	ImageLandmark *il;
	file >> nbLandmarks;
	unsigned int vx,vy,vz;
	double px,py,pz;
	char lname[240];
	for(i=0;i<nbLandmarks;++i) {
		file >> vx >> vy >> vz >> px >> py >> pz >> lname;
		il= new ImageLandmark(Vec3<unsigned int>(vx,vy,vz),Vec3<double>(px,py,pz)
			,TclString(lname));
		landmarkList.append(il);
	}

}

LandmarkSet::LandmarkSet(std::list<ImageLandmark *> list) 
{
  std::list<ImageLandmark *>::iterator it;
  for(it=list.begin();it!=list.end();it++) {
    landmarkList.append(*it);
  }
}
LandmarkSet::~LandmarkSet() {}

void LandmarkSet::addLandmark(ImageLandmark *imagelandmark)
{
  if (imagelandmark)
    {
      landmarkList.append(imagelandmark);
    }
}


void LandmarkSet::removeLandmark(ImageLandmark *imagelandmark)
{ 
  if (imagelandmark)    {
    delete imagelandmark;
  }

}
void LandmarkSet::write(char *name)
{
  OFSTREAM file(name);

  file<<landmarkList.getCount() <<ENDL;
  ObjectsList<ImageLandmark>::iterator it;

 for(it=firstLandmark();it!=lastLandmark();it++) {
    file << (*it)->voxelPosition.getX() << ' ' << (*it)->voxelPosition.getY() << ' ' << (*it)->voxelPosition.getZ() << ' ' << (*it)->realPosition.getX() << ' ' << (*it)->realPosition.getY() << ' '  << (*it)->realPosition.getZ() << ' ' << (char *)((*it)->name) << ENDL;
  }

}

void LandmarkSet::detachLandmark(ImageLandmark *imagelandmark)
{ 
  if (imagelandmark)
    {
      landmarkList.remove(imagelandmark);
    }

}



std::list<double> LandmarkSet::computeError(LandmarkSet *lset){

  std::list<double> error;
  ObjectsList<ImageLandmark>::iterator it1 = firstLandmark(), it2 = lset->firstLandmark();
  double tmp=0.0;
  double x1, y1, z1, x2, y2, z2;

  for (int i=0; it1 != lastLandmark(); ++it1, ++it2, ++i){
    x1 = (*it1)->getRealPosition()[0]; y1 = (*it1)->getRealPosition()[1]; z1 = (*it1)->getRealPosition()[2]; 
    x2 = (*it2)->getRealPosition()[0]; y2 = (*it2)->getRealPosition()[1]; z2 = (*it2)->getRealPosition()[2]; 
    tmp = sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2)+(z1-z2)*(z1-z2));
    error.push_back(tmp);
  }

  return error;
}


Transformation *LandmarkSet::computeTransfo(LandmarkSet *lset, TclString type){
  char from[]="Transformation *LandmarkSet::computeTransfo(LandmarkSet *, TclString)";
  Transformation *theTransfo = (Transformation *)NULL;

  PointSetIterator *it1 = new ImageLandmarkIterator(this), *it2 = new ImageLandmarkIterator(lset);
  it1->first();
  it2->first();

  if (!(strcmp((char *)type,"rigid")))  
    theTransfo = new RigidTransformation(it1, it2);
  else if (!(strcmp((char *)type,"similarity")))  
    theTransfo = new SimilarityTransformation(it1, it2);
  else if (!(strcmp((char *)type,"affine")))  
    theTransfo = new AffineTransformation(it1, it2);
  else
    CERR << " Invalid Transfo type in " << from << ENDL;

  theTransfo->compute();

  return theTransfo;

}

ImageLandmarkIterator::ImageLandmarkIterator(LandmarkSet *set) :lset(set) {
}

ImageLandmarkIterator::~ImageLandmarkIterator(void){
}

void ImageLandmarkIterator::first(void){
  it = lset->firstLandmark();
}

void ImageLandmarkIterator::operator++(void){
  ++it;
}

void ImageLandmarkIterator::operator++(int i){
  it++;
}
bool ImageLandmarkIterator::isAtEnd(void) const {
  return ( it==lset->lastLandmark());
}
const Vec3<double> ImageLandmarkIterator::get(void) const {
  return ((*it)->getRealPosition());
}
double ImageLandmarkIterator::getX(void) const {
  return (((*it)->getRealPosition()).getX());
}

double ImageLandmarkIterator::getY(void) const {
  return (((*it)->getRealPosition()).getY());
}

double ImageLandmarkIterator::getZ(void) const {
  return (((*it)->getRealPosition()).getZ());
}

void ImageLandmarkIterator::set(const Vec3<double>& Coordinates) {
  (*it)->setRealPosition(Coordinates);
}

double ImageLandmarkIterator::getWeight() const {
  return 1.0;
}
ImageLandmark::ImageLandmark(Vec3<unsigned int> voxelposition, 
			     Vec3<double> _realPosition, 
			     const TclString n) { 
  voxelPosition = voxelposition;
  realPosition=_realPosition;
  name = n;
}

} 



























