/******************************************************************************/
/* MaskCalcAnisotropYZ.cpp                                                    */
/******************************************************************************/
#ifdef _YAV_
#include "Tri3x3x3AnisotropYZ.h"
#include "MaskCalcAnisotropYZ.h"
#include <stdio.h>
using namespace yav;
#if ((defined(WIN32)) || ((defined(__GNUC__)) && (__GNUC__>=3)))
template <class T> T min(T a, T b) { if (a < b) return a; else return b;}

template <class T> T max(T a, T b) { if (a > b) return a; else return b;}
#include <math.h>
#endif
#else
template <class T> T min(T a, T b) { if (a < b) return a; else return b;}

template <class T> T max(T a, T b) { if (a > b) return a; else return b;}

#include "distancemap/Tri3x3x3AnisotropYZ.h"
#include "distancemap/MaskCalcAnisotropYZ.h"

#endif // _YAV_

#include <cmath>
#include <cstdlib>





MaskCalcAnisotropYZ::MaskCalcAnisotropYZ(MaskTri * aTriangulation, float dy, float dz)
  : MaskCalculation()
{
  myTriangulation = aTriangulation;

  this->myTris = &(myTriangulation->myTris);
  this->myElmts = &(myTriangulation->myElmts);
  this->nbPts = myTriangulation->nbPts; 
  
  int * tabWeight = (int*) malloc(nbPts*sizeof(int));
  for (int p = 0; p < nbPts; p++)
    tabWeight[p] = 0;
  wheightsTable.push_back(tabWeight);

  RelErrorElmt maxErr;
  maxErr.tauMin = -1;
  maxErr.tauMax = 1;
  maxErr.tau    = 1;
  maxErr.tauOpt = 1;
  errorsTable.push_back(maxErr);
  etIt = 0;

  a = 1;

  this->dy = dy;
  this->dz = dz;
}



MaskCalcAnisotropYZ::MaskCalcAnisotropYZ(float dy, float dz)
  : MaskCalculation()
{
  myTriangulation = new Tri3x3x3AnisotropYZ();

  this->myTris = &(myTriangulation->myTris);
  this->myElmts = &(myTriangulation->myElmts);
  this->nbPts = myTriangulation->nbPts; 
  
  int * tabWeight = (int*) malloc(nbPts*sizeof(int));
  for (int p = 0; p < nbPts; p++)
    tabWeight[p] = 0;
  wheightsTable.push_back(tabWeight);

  RelErrorElmt maxErr;
  maxErr.tauMin = -1;
  maxErr.tauMax = 1;
  maxErr.tau    = 1;
  maxErr.tauOpt = 1;
  errorsTable.push_back(maxErr);
  etIt = 0;

  a = 1;

  this->dy = dy;
  this->dz = dz;
}

MaskCalcAnisotropYZ::~MaskCalcAnisotropYZ()
{

}


float
MaskCalcAnisotropYZ::normeInf(chamferTriElmt elmt)
{
	return max((float) max((float) elmt.x, dy*elmt.y), dz*elmt.z);
}

float
MaskCalcAnisotropYZ::norme1(chamferTriElmt elmt)
{
  return (elmt.x + dy*elmt.y + dz*elmt.z);
}


void
MaskCalcAnisotropYZ::relativeErrorCalc(chamferTriangle tri, RelErrorElmt * errElmt)
{

  float epsilon;
  float fMax;
  float fABMax, fBCMax, fACMax;
  float fAMax, fBMax, fCMax;
  float xA1, yA1, zA1, xB1, yB1, zB1, xC1, yC1, zC1;
  float cond1, cond2, cond3;
  bool isInTriFMax;
  bool isInTriAB;
  bool isInTriBC;
  bool isInTriAC;


  float P, Q;
  float xMax, zMax, yMax;
  int alpha, beta, gamma;
  
  int wA = (*myElmts)[tri.A].w;
  int wB = (*myElmts)[tri.B].w;
  int wC = (*myElmts)[tri.C].w;
  
  alpha = wA*tri.nC[0][0] + wB*tri.nC[0][1] + wC*tri.nC[0][2];
  beta  = wA*tri.nC[1][0] + wB*tri.nC[1][1] + wC*tri.nC[1][2];
  gamma = wA*tri.nC[2][0] + wB*tri.nC[2][1] + wC*tri.nC[2][2];


	float beta1 = beta;

	float gamma1 = gamma;

  epsilon = (float) a;
  
  xA1 = tri.xA1;
  yA1 = tri.yA1;
  zA1 = tri.zA1;
  xB1 = tri.xB1;
  yB1 = tri.yB1;
  zB1 = tri.zB1;
  xC1 = tri.xC1;
  yC1 = tri.yC1;
  zC1 = tri.zC1;
  
  switch(tri.whichPlan)
    {
    case 0 : // projection sur le plan x = 1
      /*** Maximum global ***/
      /* Obtained for :
	 xPx = 1, 
	 yPxMax = beta  / (alpha*dy^2), 
	 zPxMax = gamma / (alpha*dz^2)
	 
	 xP = alpha, yP = beta/(dy^2), zP = gamma/(dz^2)
	 This maximum is obtained if P1(1, yPxMax, zPxMax) is in (A1, B1, C1), i.e.
	 Delta(A1B1P1) > 0 && Delta(A1P1C1) > 0 && Delta(P1B1C1) > 0 i.e.
	 Delta(A1B1P)  > 0 && Delta(A1PC1)  > 0 && Delta(PB1C1)  > 0  
	 indeed : Delta(A1B1P1) =  1/alpha Delta(A1B1P) and alpha > 0
      */
      beta1  = (float) beta  / (dy*dy);
      gamma1 = (float) gamma / (dz*dz);
      fMax = (sqrt(alpha*alpha + beta*beta1 + gamma*gamma1)/epsilon) - 1;
      cond1 = (yB1*gamma1 + beta1*zA1 + alpha*yA1*zB1 - alpha*yB1*zA1 - yA1*gamma1 - beta1*zB1);
      cond2 = (beta1*zC1 + alpha*yC1*zA1 + yA1*gamma1 - beta1*zA1 - alpha*yA1*zC1 - yC1*gamma1);
      cond3 = (alpha*yB1*zC1 + yC1*gamma1 + beta1*zB1 - yB1*gamma1 - beta1*zC1 - alpha*yC1*zB1);
      isInTriFMax = (cond1 >= 0) && (cond2 >= 0) && (cond3 >= 0);

      /*** Maxima sur les axes ***/
      /** Axe AB **/
      if (tri.denoAB != 0.0) // On a une droite du type y = Pz + Q
	{
	  P = tri.PAB;
	  Q = tri.QAB;
	  zMax = (-alpha*P*Q*dy*dy + beta*P + gamma*(1+dy*dy*Q*Q)) / 
	    (alpha*(P*P*dy*dy+dz*dz) + beta*Q*dz*dz - gamma*dy*dy*P*Q);
	  float yTmp = P*zMax + Q; 
	  fABMax = ((alpha + beta*yTmp + gamma*zMax) /
		    (epsilon*sqrt(1 + dy*dy*yTmp*yTmp + dz*dz*zMax*zMax))) - 1;
	  isInTriAB = (((zMax > zA1) && (zMax < zB1)) || 
		       ((zMax < zA1) && (zMax > zB1))); 
	}// fin if (deno != 0)
      else // On a une droite du type z = Q
	{
	  Q = tri.QAB;
	  yMax = (beta*(1 + Q*Q*dz*dz)) / ((alpha + gamma*Q)*dy*dy);
	  fABMax = (alpha + beta*yMax + gamma*Q) / (epsilon*(sqrt(1 + dy*dy*yMax*yMax + Q*Q*dz*dz))) - 1;
	  isInTriAB = (((yMax > yA1) && (yMax < yB1)) || 
		       ((yMax < yA1) && (yMax > yB1)));
	}// fin else (deno != 0)
      
      /** Axe BC **/
      if (tri.denoBC != 0.0) // On a une droite du type y = Pz + Q
	{
	  P = tri.PBC;
	  Q = tri.QBC;
	  zMax = (-alpha*P*Q*dy*dy + beta*P + gamma*(1+dy*dy*Q*Q)) / 
	    (alpha*(P*P*dy*dy+dz*dz) + beta*Q*dz*dz - gamma*dy*dy*P*Q);
	  float yTmp = P*zMax + Q; 
	  fBCMax = ((alpha + beta*yTmp + gamma*zMax) /
		    (epsilon*sqrt(1 + dy*dy*yTmp + dz*dz*zMax*zMax))) - 1;
	  isInTriBC = (((zMax > zB1) && (zMax < zC1)) || 
		       ((zMax < zB1) && (zMax > zC1))); 
	}// fin if (deno != 0)
      else // On a une droite du type z = Q
	{
	  Q = tri.QBC;
	  yMax = (beta*(1 + Q*Q*dz*dz)) / ((alpha + gamma*Q)*dy*dy);
	  fBCMax = (alpha + beta*yMax + gamma*Q) / (epsilon*(sqrt(1 + dy*dy*yMax*yMax + Q*Q*dz*dz))) - 1;
	  isInTriBC = (((yMax > yB1) && (yMax < yC1)) || 
		       ((yMax < yB1) && (yMax > yC1)));
	}// fin else (deno != 0)
      
      /** Axe AC **/
      if (tri.denoAC != 0.0) // On a une droite du type y = Pz + Q
	{
	  P = tri.PAC;
	  Q = tri.QAC;
	  zMax = (-alpha*P*Q*dy*dy + beta*P + gamma*(1+dy*dy*Q*Q)) / 
	    (alpha*(P*P*dy*dy+dz*dz) + beta*Q*dz*dz - gamma*dy*dy*P*Q);
	  float yTmp = P*zMax + Q;
	  fACMax = ((alpha + beta*yTmp + gamma*zMax) /
		    (epsilon*sqrt(1 + dy*dy*yTmp + dz*dz*zMax*zMax))) - 1;
	  isInTriAC = (((zMax > zA1) && (zMax < zC1)) || 
		       ((zMax < zA1) && (zMax > zC1))); 
	}// fin if (deno != 0)
      else // On a une droite du type z = Q
	{
	  Q = tri.QAC;
	  yMax = (beta*(1 + Q*Q*dz*dz)) / ((alpha + gamma*Q)*dy*dy);
	  fACMax = (alpha + beta*yMax + gamma*Q) / (epsilon*(sqrt(1 + dy*dy*yMax*yMax + Q*Q*dz*dz))) - 1;
	  isInTriAC = (((yMax > yA1) && (yMax < yC1)) || 
		       ((yMax < yA1) && (yMax > yC1)));
	}// fin else (deno != 0)

      /*** Maxima sur les sommets ***/
      /** Sommet A **/
      fAMax = ((alpha + beta*yA1 + gamma*zA1) / (epsilon*sqrt(1 + dy*dy*yA1*yA1 + dz*dz*zA1*zA1))) - 1;
      /** Sommet B **/
      fBMax = ((alpha + beta*yB1 + gamma*zB1) / (epsilon*sqrt(1 + dy*dy*yB1*yB1 + dz*dz*zB1*zB1))) - 1;
      /** Sommet C **/
      fCMax = ((alpha + beta*yC1 + gamma*zC1) / (epsilon*sqrt(1 + dy*dy*yC1*yC1 + dz*dz*zC1*zC1))) - 1;

      break;

    case 1 : // projection sur le plan y = 1
      /*** Maximum global ***/
      /* Obtained for :
	 xPyMax = (alpha*dy^2) / beta, 
	 yPyMax = 1, 
	 zPyMax = (gamma*dy^2) / (beta*dz^2)
	 
	 xP = alpha, yP = beta/(dy^2), zP = gamma/(dz^2) 
	 xP = (beta/dy^2) PyMax 
	 This maximum is obtained if P1(1, yPxMax, zPxMax) is in (A1, B1, C1), i.e.
	 Delta(A1B1P1) > 0 && Delta(A1P1C1) > 0 && Delta(P1B1C1) > 0 i.e.
	 Delta(A1B1P)  > 0 && Delta(A1PC1)  > 0 && Delta(PB1C1)  > 0  
	 indeed : Delta(A1B1P1) =  (beta/dy^2) Delta(A1B1P) and (beta  / dy^2) > 0
      */
      beta1  = (float) beta  / (dy*dy);
      gamma1 = (float) gamma / (dz*dz);
      fMax = (sqrt(alpha*alpha + beta*beta1 + gamma*gamma1)/epsilon) - 1;
      cond1 = (xA1*gamma1 + xB1*beta1*zA1 + alpha*zB1 - alpha*zA1 - xB1*gamma1 - xA1*beta1*zB1);
      cond2 = (xA1*beta1*zC1 + alpha*zA1 + xC1*gamma1 - xC1*beta1*zA1 - alpha*zC1 - xA1*gamma1);
      cond3 = (alpha*zC1 + xB1*gamma1 + xC1*beta1*zB1 - xC1*gamma1 - xB1*beta1*zC1 - alpha*zB1);
      isInTriFMax = (cond1 >= 0) && (cond2 >= 0) && (cond3 >= 0);

      /*** Maxima sur les axes ***/
      /** Axe AB **/
      if (tri.denoAB != 0.0) // On a une droite du type x = Pz + Q
	{
	  P = tri.PAB;
	  Q = tri.QAB;
	  zMax = (alpha*P*dy*dy - beta*P*Q + gamma*(dy*dy + Q*Q)) / 
	    (alpha*Q*dz*dz + beta*(P*P + dz*dz) - gamma*P*Q);
	  float xTmp = P*zMax + Q; 
	  fABMax = ((alpha*xTmp + beta + gamma*zMax) /
		    (epsilon*sqrt(xTmp*xTmp + dy*dy + dz*dz*zMax*zMax))) - 1;
	  isInTriAB = (((zMax > zA1) && (zMax < zB1)) || 
		       ((zMax < zA1) && (zMax > zB1))); 
	}// fin if (deno != 0)
      else // On a une droite du type z = Q
	{
	  Q = tri.QAB;
	  xMax = (alpha*(dy*dy + Q*Q*dz*dz)) / (beta + gamma*Q);
	  fABMax = (alpha*xMax + beta + gamma*Q) / (epsilon*(sqrt(xMax*xMax + dy*dy + Q*Q*dz*dz))) - 1;
	  isInTriAB = (((xMax > xA1) && (xMax < xB1)) || 
		       ((xMax < xA1) && (xMax > xB1)));
	}// fin else (deno != 0)
      
      /** Axe BC **/
      if (tri.denoBC != 0.0) // On a une droite du type x = Pz + Q
	{
	  P = tri.PBC;
	  Q = tri.QBC;
	  zMax = (alpha*P*dy*dy - beta*P*Q + gamma*(dy*dy + Q*Q)) / 
	    (alpha*Q*dz*dz + beta*(P*P + dz*dz) - gamma*P*Q);
	  float xTmp = P*zMax + Q; 
	  fBCMax = ((alpha*xTmp + beta + gamma*zMax) /
		    (epsilon*sqrt(xTmp*xTmp + dy*dy + dz*dz*zMax*zMax))) - 1;
	  isInTriBC = (((zMax > zB1) && (zMax < zC1)) || 
		       ((zMax < zB1) && (zMax > zC1))); 
	}// fin if (deno != 0)
      else // On a une droite du type z = Q
	{
	  Q = tri.QBC;
	  xMax = (alpha*(dy*dy + Q*Q*dz*dz)) / (beta + gamma*Q);
	  fBCMax = (alpha*xMax + beta + gamma*Q) / (epsilon*(sqrt(xMax*xMax + dy*dy + Q*Q*dz*dz))) - 1;
	  isInTriBC = (((xMax > xB1) && (xMax < xC1)) || 
		       ((xMax < xB1) && (xMax > xC1)));
	}// fin else (deno != 0)
      
      /** Axe AC **/
      if (tri.denoAC != 0.0) // On a une droite du type y = Pz + Q
	{
	  P = tri.PAC;
	  Q = tri.QAC;
	  zMax = (alpha*P*dy*dy - beta*P*Q + gamma*(dy*dy + Q*Q)) / 
	    (alpha*Q*dz*dz + beta*(P*P + dz*dz) - gamma*P*Q);
	  float xTmp = P*zMax + Q; 
	  fACMax = ((alpha*xTmp + beta + gamma*zMax) /
		    (epsilon*sqrt(xTmp*xTmp + dy*dy + dz*dz*zMax*zMax))) - 1;
	  isInTriAC = (((zMax > zA1) && (zMax < zC1)) || 
		       ((zMax < zA1) && (zMax > zC1))); 
	}// fin if (deno != 0)
      else // On a une droite du type z = Q
	{
	  Q = tri.QAC;
	  xMax = (alpha*(dy*dy + Q*Q*dz*dz)) / (beta + gamma*Q);
	  fACMax = (alpha*xMax + beta + gamma*Q) / (epsilon*(sqrt(xMax*xMax + dy*dy + Q*Q*dz*dz))) - 1;
	  isInTriAC = (((xMax > xA1) && (xMax < xC1)) || 
		       ((xMax < xA1) && (xMax > xC1)));
	}// fin else (deno != 0)

      /*** Maxima sur les sommets ***/
      /** Sommet A **/
      fAMax = ((alpha*xA1 + beta + gamma*zA1) / (epsilon*sqrt(xA1*xA1 + dy*dy + dz*dz*zA1*zA1))) - 1;
      /** Sommet B **/
      fBMax = ((alpha*xB1 + beta + gamma*zB1) / (epsilon*sqrt(xB1*xB1 + dy*dy + dz*dz*zB1*zB1))) - 1;
      /** Sommet C **/
      fCMax = ((alpha*xC1 + beta + gamma*zC1) / (epsilon*sqrt(xC1*xC1 + dy*dy + dz*dz*zC1*zC1))) - 1;

      break;

    case 2 : // projection sur le plan z = 1
      /*** Maximum global ***/
      /* Obtained for :
	 xPyMax = (alpha*dz^2) / gamma 
	 yPyMax = (beta*dz^2)  / (gamma*dy^2)
	 zPyMax = 1
	 
	 xP = alpha, yP = beta/(dy^2), zP = gamma/(dz^2) 
	 xP = (gamma/dz^2) PyMax 
	 This maximum is obtained if P1(1, yPxMax, zPxMax) is in (A1, B1, C1), i.e.
	 Delta(A1B1P1) > 0 && Delta(A1P1C1) > 0 && Delta(P1B1C1) > 0 i.e.
	 Delta(A1B1P)  > 0 && Delta(A1PC1)  > 0 && Delta(PB1C1)  > 0  
	 indeed : Delta(A1B1P1) =  (gamma/dz^2) Delta(A1B1P) and (gamma / dz^2) > 0
      */
      beta1  = (float) beta  / (dy*dy);
      gamma1 = (float) gamma / (dz*dz);
      fMax = (sqrt(alpha*alpha + beta*beta1 + gamma*gamma1)/epsilon) - 1;
      cond1 = (xA1*yB1*gamma1 + xB1*beta1 + alpha*yA1 - alpha*yB1 - xB1*yA1*gamma1 - xA1*beta1);
      cond2 = (xA1*beta1 + alpha*yC1 + xC1*yA1*gamma1 - xC1*beta1 - alpha*yA1 - xA1*yC1*gamma1);
      cond3 = (alpha*yB1 + xB1*yC1*gamma1 + xC1*beta1 - xC1*yB1*gamma1 - xB1*beta1 - alpha*yC1);
      isInTriFMax = (cond1 >= 0) && (cond2 >= 0) && (cond3 >= 0);

      /*** Maxima sur les axes ***/
      /** Axe AB **/
      if (tri.denoAB != 0.0) // On a une droite du type x = Py + Q
	{
	  P = tri.PAB;
	  Q = tri.QAB;
	  yMax = (alpha*P*dz*dz + beta*(Q*Q + dz*dz) - gamma*P*Q) / 
	    (alpha*Q*dy*dy - beta*P*Q + gamma*(P*P + dy*dy));
	  float xTmp = P*yMax + Q; 
	  fABMax = ((alpha*xTmp + beta*yMax + gamma) /
		    (epsilon*sqrt(xTmp*xTmp + dy*dy*yMax*yMax + dz*dz))) - 1;
	  isInTriAB = (((yMax > yA1) && (yMax < yB1)) || 
		       ((yMax < yA1) && (yMax > yB1))); 
	}// fin if (deno != 0)
      else // On a une droite du type y = Q
	{
	  Q = tri.QAB;
	  xMax = (alpha*(dy*dy*Q*Q + dz*dz)) / (beta*Q + gamma);
	  fABMax = ((alpha*xMax + beta*Q + gamma) / 
		    (epsilon*(sqrt(xMax*xMax + dy*dy*Q*Q + dz*dz)))) - 1;
	  isInTriAB = (((xMax > xA1) && (xMax < xB1)) || 
		       ((xMax < xA1) && (xMax > xB1)));
	}// fin else (deno != 0)
      
      /** Axe BC **/
      if (tri.denoBC != 0.0) // On a une droite du type x = Py + Q
	{
	  P = tri.PBC;
	  Q = tri.QBC;
	  yMax = (alpha*P*dz*dz + beta*(Q*Q + dz*dz) - gamma*P*Q) / 
	    (alpha*Q*dy*dy - beta*P*Q + gamma*(P*P + dy*dy));
	  float xTmp = P*yMax + Q; 
	  fBCMax = ((alpha*xTmp + beta*yMax + gamma) /
		    (epsilon*sqrt(xTmp*xTmp + dy*dy*yMax*yMax + dz*dz))) - 1;
	  isInTriBC = (((yMax > yB1) && (yMax < yC1)) || 
		       ((yMax < yB1) && (yMax > yC1))); 
	}// fin if (deno != 0)
      else // On a une droite du type z = Q
	{
	  Q = tri.QBC;
	  xMax = (alpha*(dy*dy*Q*Q + dz*dz)) / (beta*Q + gamma);
	  fBCMax = ((alpha*xMax + beta*Q + gamma) / 
		    (epsilon*(sqrt(xMax*xMax + dy*dy*Q*Q + dz*dz)))) - 1;
	  isInTriBC = (((xMax > xB1) && (xMax < xC1)) || 
		       ((xMax < xB1) && (xMax > xC1)));
	}// fin else (deno != 0)
      
      /** Axe AC **/
      if (tri.denoAC != 0.0) // On a une droite du type y = Pz + Q
	{
	  P = tri.PAC;
	  Q = tri.QAC;
	  yMax = (alpha*P*dz*dz + beta*(Q*Q + dz*dz) - gamma*P*Q) / 
	    (alpha*Q*dy*dy - beta*P*Q + gamma*(P*P + dy*dy));
	  float xTmp = P*yMax + Q; 
	  fACMax = ((alpha*xTmp + beta*yMax + gamma) /
		    (epsilon*sqrt(xTmp*xTmp + dy*dy*yMax*yMax + dz*dz))) - 1;
	  isInTriAC = (((yMax > yA1) && (yMax < yC1)) || 
		       ((yMax < yA1) && (yMax > yC1))); 
	}// fin if (deno != 0)
      else // On a une droite du type z = Q
	{
	  Q = tri.QAC;
	  xMax = (alpha*(dy*dy*Q*Q + dz*dz)) / (beta*Q + gamma);
	  fACMax = ((alpha*xMax + beta*Q + gamma) / 
		    (epsilon*(sqrt(xMax*xMax + dy*dy*Q*Q + dz*dz)))) - 1;
	  isInTriAC = (((xMax > xA1) && (xMax < xC1)) || 
		       ((xMax < xA1) && (xMax > xC1)));
	}// fin else (deno != 0)
    
      /*** Maxima sur les sommets ***/
      /** Sommet A **/
      fAMax = ((alpha*xA1 + beta*yA1 + gamma) / (epsilon*sqrt(xA1*xA1 + dy*dy*yA1*yA1 + dz*dz))) - 1;
      /** Sommet B **/
      fBMax = ((alpha*xB1 + beta*yB1 + gamma) / (epsilon*sqrt(xB1*xB1 + dy*dy*yB1*yB1 + dz*dz))) - 1;
      /** Sommet C **/
      fCMax = ((alpha*xC1 + beta*yC1 + gamma) / (epsilon*sqrt(xC1*xC1 + dy*dy*yC1*yC1 + dz*dz))) - 1;
      break;
      
    default :
      printf("Warning, this case should not occur\n");
      exit(-1);
      break;

    } // end switch(tri.whichPlan)
    

      
  errElmt->tauMin = fCMax;
  errElmt->tauMax = fCMax;
  errElmt->tauMin = min(fBMax, errElmt->tauMin);
  errElmt->tauMax = max(fBMax, errElmt->tauMax);
  errElmt->tauMin = min(fAMax, errElmt->tauMin);
  errElmt->tauMax = max(fAMax, errElmt->tauMax);
  if (isInTriAC) {
    errElmt->tauMin = min(fACMax, errElmt->tauMin);
    errElmt->tauMax = max(fACMax, errElmt->tauMax);
  }
  if (isInTriAB) {
    errElmt->tauMin = min(fABMax, errElmt->tauMin);
    errElmt->tauMax = max(fABMax, errElmt->tauMax);
  }
  if (isInTriBC) {
    errElmt->tauMin = min(fBCMax, errElmt->tauMin);
    errElmt->tauMax = max(fBCMax, errElmt->tauMax);
  }
  if (isInTriFMax) {
    errElmt->tauMin = min(fMax, errElmt->tauMin);
    errElmt->tauMax = max(fMax, errElmt->tauMax);
  }
    
}
