/******************************************************************************/
/* MaskCalculation.cpp                                                        */
/******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef _YAV_
#include "MaskCalculation.h"
using namespace yav;
#if ((defined(WIN32)) || ((defined(__GNUC__)) && (__GNUC__>=3)))
template <class T> T min(T a, T b) { if (a < b) return a; else return b;}
template <class T> T max(T a, T b) { if (a > b) return a; else return b;}
#endif
#else
template <class T> T min(T a, T b) { if (a < b) return a; else return b;}
template <class T> T max(T a, T b) { if (a > b) return a; else return b;}
#include "distancemap/MaskCalculation.h"
#endif // _YAV_

#include <cmath>

MaskCalculation::MaskCalculation()
{

}


MaskCalculation::~MaskCalculation()
{

}

void
MaskCalculation::printMask(char * filename)
{
  FILE * fp = fopen(filename, "w");

  fprintf(fp, "Wheights...\t\t TauOpt \t Tau \t\t TauMin \t TauMax \t epsOpt\n");
  for(unsigned int p = 1; p <= etIt; p++) {
    for (int q = 0; q < nbPts; q++)
      fprintf(fp, "%i, ", (wheightsTable[p])[q]);
    fprintf(fp, "\t %f \t %f \t %f \t %f \t %f\n", 
	    errorsTable[p].tauOpt, errorsTable[p].tau, 
	    errorsTable[p].tauMin, errorsTable[p].tauMax, errorsTable[p].epsOpt);
  }
  fclose(fp);
}



void
MaskCalculation::initLCC()
{

  int i, j;
  int a, b, c, aSym, bSym, cSym;
  int nbTris = myTris->size();
  int elmtIt = myElmts->size() - 1;

  chamferTriangle * tri;
  chamferTriElmt newPoint;
  LCCElmt newLCC;

  int newIndex;

  chamferTriangle a1bc;
  chamferTriangle aa1c;
  chamferTriangle aba1;

  chamferTriangle b1bc;
  chamferTriangle ab1c;
  chamferTriangle abb1;
  
  chamferTriangle c1bc;
  chamferTriangle ac1c;
  chamferTriangle abc1;

  for (i = 0; i < nbTris; i++)
    {
      tri = &((*myTris)[i]);
      a = tri->A;
      b = tri->B;
      c = tri->C;

      /*-- Segment BC du triangle --*/
      if (tri->Abc < 0) // sinon, ca veut dire qu'on l'a deja determine
	{
	  switch (tri->Abc)
	    {
	    case -2 : // Symetrique par rapport au plan P : (x = y)
	      newPoint.x = (*myElmts)[a].y;
	      newPoint.y = (*myElmts)[a].x;
	      newPoint.z = (*myElmts)[a].z;
	      newPoint.w = (*myElmts)[a].w;
	      
	      newIndex = myTriangulation->isInSet(newPoint);
	      if (newIndex == -1) {
		myElmts->push_back(newPoint);
		elmtIt++;
		tri->Abc = elmtIt;
	      }
	      else
		tri->Abc = newIndex;

	      newLCC.w0 = tri->A;
	      break;
	      
	    case -3 : // Symetrique par rapport au plan P : (y = z)
	      newPoint.x = (*myElmts)[a].x;
	      newPoint.y = (*myElmts)[a].z;
	      newPoint.z = (*myElmts)[a].y;
	      newPoint.w = (*myElmts)[a].w;
	      
	      newIndex = myTriangulation->isInSet(newPoint);
	      if (newIndex == -1) {
		myElmts->push_back(newPoint);
		elmtIt++;
		tri->Abc = elmtIt;
	      }
	      else
		tri->Abc = newIndex;

	      newLCC.w0 = tri->A;
	      break;
	      
	    case -4 : // Symtrie par rapport au plan P : (z = 0)
	      newPoint.x = (*myElmts)[a].x;
	      newPoint.y = (*myElmts)[a].y;
	      newPoint.z = - (*myElmts)[a].z;
	      newPoint.w = (*myElmts)[a].w;
	      
	      newIndex = myTriangulation->isInSet(newPoint);
	      if (newIndex == -1) {
		myElmts->push_back(newPoint);
		elmtIt++;
		tri->Abc = elmtIt;
	      }
	      else
		tri->Abc = newIndex;
	      
	      newLCC.w0 = tri->A;
	      break;
	      
	    case -5 : // Symtrie par rapport au plan P : (y = 0)
	      newPoint.x = (*myElmts)[a].x;
	      newPoint.y = - (*myElmts)[a].y;
	      newPoint.z = (*myElmts)[a].z;
	      newPoint.w = (*myElmts)[a].w;
	      
	      newIndex = myTriangulation->isInSet(newPoint);
	      if (newIndex == -1) {
		myElmts->push_back(newPoint);
		elmtIt++;
		tri->Abc = elmtIt;
	      }
	      else
		tri->Abc = newIndex;
	      
	      newLCC.w0 = tri->A;
	      break;
	      
	    case -6 : // Symtrie par rapport au plan P : (x = 0)
	      newPoint.x = - (*myElmts)[a].x;
	      newPoint.y = (*myElmts)[a].y;
	      newPoint.z = (*myElmts)[a].z;
	      newPoint.w = (*myElmts)[a].w;
	      
	      newIndex = myTriangulation->isInSet(newPoint);
	      if (newIndex == -1) {
		myElmts->push_back(newPoint);
		elmtIt++;
		tri->Abc = elmtIt;
	      }
	      else
		tri->Abc = newIndex;
	      
	      newLCC.w0 = tri->A;
	      break;
	      
	    case -1 : // Symetrique dans la triangulation
	      for (j = 0; j < nbTris; j++)
		{
		  if (j != i) {
		    aSym = (*myTris)[j].A;
		    bSym = (*myTris)[j].B;
		    cSym = (*myTris)[j].C;
		    // On a trois possibilits d'artes commune (triangles directs)
		    if ((b == aSym) && (c ==  cSym)) {
		      tri->Abc = bSym;
		      (*myTris)[j].Bac = a;
		    }
		    if ((b == bSym) && (c == aSym)) {
		      tri->Abc = cSym;
		      (*myTris)[j].Cab = a;
		    }
		    if ((b == cSym) && (c == bSym)) {
		      tri->Abc = aSym;
		      (*myTris)[j].Abc = a;
		    }
		  }
		}

	      newLCC.w0 = tri->Abc;
	      
	      break;
	      
	    default :
	      printf("warning, this case should not occur\n");
	      exit(-1);
	      break;   
	    }// end switch(tri->Abc)
	  
	
	  a1bc.A = tri->Abc; a1bc.B = tri->B;   a1bc.C = tri->C;
	  aa1c.A = tri->A;   aa1c.B = tri->Abc; aa1c.C = tri->C;
	  aba1.A = tri->A;   aba1.B = tri->B;   aba1.C = tri->Abc;

	  newLCC.w1 = tri->A;
	  newLCC.w2 = tri->B;
	  newLCC.w3 = tri->C;
	  newLCC.D1 = delta(a1bc);
	  newLCC.D2 = delta(aa1c);
	  newLCC.D3 = delta(aba1);
	  newLCC.flag = 2;

	  LCCTable.push_back(newLCC);

	} // end if (tri->Abc < 0)




      /*-- Segment AC du triangle --*/
      if (tri->Bac < 0) // sinon, ca veut dire qu'on l'a deja determine
	{
	  switch (tri->Bac)
	    {
	    case -2 : // Symetrique par rapport au plan P : (x = y)
	      newPoint.x = (*myElmts)[b].y;
	      newPoint.y = (*myElmts)[b].x;
	      newPoint.z = (*myElmts)[b].z;
	      newPoint.w = (*myElmts)[b].w;

	      newIndex = myTriangulation->isInSet(newPoint);
	      if (newIndex == -1) {
		myElmts->push_back(newPoint);
		elmtIt++;
		tri->Bac = elmtIt;
	      }
	      else
		tri->Bac = newIndex;

	      newLCC.w0 = tri->B;
	      break;

	    case -3 : // Symetrique par rapport au plan P : (y = z)
	      newPoint.x = (*myElmts)[b].x;
	      newPoint.y = (*myElmts)[b].z;
	      newPoint.z = (*myElmts)[b].y;
	      newPoint.w = (*myElmts)[b].w;
	      
	      newIndex = myTriangulation->isInSet(newPoint);
	      if (newIndex == -1) {
		myElmts->push_back(newPoint);
		elmtIt++;
		tri->Bac = elmtIt;
	      }
	      else
		tri->Bac = newIndex;

	      newLCC.w0 = tri->B;
	      break;
	      
	    case -4 : // Symtrie par rapport au plan P : (z = 0)
	      newPoint.x = (*myElmts)[b].x;
	      newPoint.y = (*myElmts)[b].y;
	      newPoint.z = - (*myElmts)[b].z;
	      newPoint.w = (*myElmts)[b].w;
	      
	      newIndex = myTriangulation->isInSet(newPoint);
	      if (newIndex == -1) {
		myElmts->push_back(newPoint);
		elmtIt++;
		tri->Bac = elmtIt;
	      }
	      else
		tri->Bac = newIndex;

	      newLCC.w0 = tri->B;
	      break;

	    case -5 : // Symtrie par rapport au plan P : (y = 0)
	      newPoint.x = (*myElmts)[b].x;
	      newPoint.y = - (*myElmts)[b].y;
	      newPoint.z = (*myElmts)[b].z;
	      newPoint.w = (*myElmts)[b].w;
	      
	      newIndex = myTriangulation->isInSet(newPoint);
	      if (newIndex == -1) {
		myElmts->push_back(newPoint);
		elmtIt++;
		tri->Bac = elmtIt;
	      }
	      else
		tri->Bac = newIndex;

	      newLCC.w0 = tri->B;
	      break;

	    case -6 : // Symtrie par rapport au plan P : (x = 0)
	      newPoint.x = - (*myElmts)[b].x;
	      newPoint.y = (*myElmts)[b].y;
	      newPoint.z = (*myElmts)[b].z;
	      newPoint.w = (*myElmts)[b].w;
	      
	      newIndex = myTriangulation->isInSet(newPoint);
	      if (newIndex == -1) {
		myElmts->push_back(newPoint);
		elmtIt++;
		tri->Bac = elmtIt;
	      }
	      else
		tri->Bac = newIndex;

	      newLCC.w0 = tri->B;
	      break;

	    case -1 : // Symetrique dans la triangulation
	      for (j = 0; j < nbTris; j++)
		{	      
		  if (j != i) // sauf celui sur lequel on travaille...
		    {
		      aSym = (*myTris)[j].A;
		      bSym = (*myTris)[j].B;
		      cSym = (*myTris)[j].C;
		      
		      if ((a == aSym) && (c == bSym)) {
			tri->Bac = cSym;
			(*myTris)[j].Cab = b;
		      }
		      if ((a == bSym) && (c == cSym)) {
			tri->Bac = aSym;
			(*myTris)[j].Abc = b;
		      }
		      if ((a == cSym) && (c == aSym)) {
			tri->Bac = bSym;
			(*myTris)[j].Bac = b;
		      }
		    }
		}

	      newLCC.w0 = tri->Bac;
	      break;

	    default :
	      printf("warning, this case should not occur\n");
	      exit(-1);
	      break;   
	      
	    }// end switch (tri->Bac)

	  b1bc.A = tri->Bac; b1bc.B = tri->B;   b1bc.C = tri->C;
	  ab1c.A = tri->A;   ab1c.B = tri->Bac; ab1c.C = tri->C;
	  abb1.A = tri->A;   abb1.B = tri->B;   abb1.C = tri->Bac;
	  
	  newLCC.w1 = tri->A;
	  newLCC.w2 = tri->B;
	  newLCC.w3 = tri->C;
	  newLCC.D1 = delta(b1bc);
	  newLCC.D2 = delta(ab1c);
	  newLCC.D3 = delta(abb1);
	  newLCC.flag = 2;

	  LCCTable.push_back(newLCC);

	} //end if (tri->Bac < 0)


      /*-- Segment AB du triangle --*/
      if (tri->Cab < 0) // sinon, ca veut dire qu'on l'a deja determine
	{
	  switch (tri->Cab)
	    {
	    case -2 : // Symetrique par rapport au plan P : (x = y)
	      newPoint.x = (*myElmts)[c].y;
	      newPoint.y = (*myElmts)[c].x;
	      newPoint.z = (*myElmts)[c].z;
	      newPoint.w = (*myElmts)[c].w;

	      newIndex = myTriangulation->isInSet(newPoint);
	      if (newIndex == -1) {
		myElmts->push_back(newPoint);
		elmtIt++;
		tri->Cab = elmtIt;
	      }
	      else
		tri->Cab = newIndex;
	      
	      newLCC.w0 = tri->C;
	      break;
	      
	    case -3 :  // Symetrique par rapport au plan P : (y = z)
	      newPoint.x = (*myElmts)[c].x;
	      newPoint.y = (*myElmts)[c].z;
	      newPoint.z = (*myElmts)[c].y;
	      newPoint.w = (*myElmts)[c].w;
	      
	      newIndex = myTriangulation->isInSet(newPoint);
	      if (newIndex == -1) {
		myElmts->push_back(newPoint);
		elmtIt++;
		tri->Cab = elmtIt;
	      }
	      else
		tri->Cab = newIndex;

	      newLCC.w0 = tri->C;
	      break;
	  
	    case -4 : // Symtrie par rapport au plan P : (z = 0)
	      newPoint.x = (*myElmts)[c].x;
	      newPoint.y = (*myElmts)[c].y;
	      newPoint.z = - (*myElmts)[c].z;
	      newPoint.w = (*myElmts)[c].w;
	      
	      newIndex = myTriangulation->isInSet(newPoint);
	      if (newIndex == -1) {
		myElmts->push_back(newPoint);
		elmtIt++;
		tri->Cab = elmtIt;
	      }
	      else
		tri->Cab = newIndex;

	      newLCC.w0 = tri->C;
	      break;
	      
	    case -5 : // Symtrie par rapport au plan P : (y = 0)
	      newPoint.x = (*myElmts)[c].x;
	      newPoint.y = - (*myElmts)[c].y;
	      newPoint.z = (*myElmts)[c].z;
	      newPoint.w = (*myElmts)[c].w;
	      
	      newIndex = myTriangulation->isInSet(newPoint);
	      if (newIndex == -1) {
		myElmts->push_back(newPoint);
		elmtIt++;
		tri->Cab = elmtIt;
	      }
	      else
		tri->Cab = newIndex;

	      newLCC.w0 = tri->C;
	      break;
	      
	    case -6 : // Symtrie par rapport au plan P : (x = 0)
	      newPoint.x = - (*myElmts)[c].x;
	      newPoint.y = (*myElmts)[c].y;
	      newPoint.z = (*myElmts)[c].z;
	      newPoint.w = (*myElmts)[c].w;
	      
	      newIndex = myTriangulation->isInSet(newPoint);
	      if (newIndex == -1) {
		myElmts->push_back(newPoint);
		elmtIt++;
		tri->Cab = elmtIt;
	      }
	      else
		tri->Cab = newIndex;

	      newLCC.w0 = tri->C;
	      break;
	      
	    case -1 : // Symetrique dans la triangulation
	      for (j = 0; j < nbTris; j++)
		{	      
		  if (j != i) // sauf celui sur lequel on travaille...
		    {
		      aSym = (*myTris)[j].A;
		      bSym = (*myTris)[j].B;
		      cSym = (*myTris)[j].C;
		      
		      if ((a == aSym) && (b == cSym)) {
			tri->Cab = bSym;
			(*myTris)[j].Bac = c;
		      }
		      if ((a == bSym) && (b == aSym)) {
			tri->Cab = cSym;
			(*myTris)[j].Cab = c;
		      }
		      if ((a == cSym) && (b == bSym)) {
			tri->Cab = aSym;
			(*myTris)[j].Abc = c;
		      }
		    }
		}
	      
	      newLCC.w0 = tri->Cab;
	      break;

	    default :
	      printf("warning, this case should not occur\n");
	      exit(-1);
	      break;   
	      
	    }// end switch (tri->Bac)

	  c1bc.A = tri->Cab; c1bc.B = tri->B;   c1bc.C = tri->C;
	  ac1c.A = tri->A;   ac1c.B = tri->Cab; ac1c.C = tri->C;
	  abc1.A = tri->A;   abc1.B = tri->B;   abc1.C = tri->Cab;
	  
	  newLCC.w1 = tri->A;
	  newLCC.w2 = tri->B;
	  newLCC.w3 = tri->C;
	  newLCC.D1 = delta(c1bc);
	  newLCC.D2 = delta(ac1c);
	  newLCC.D3 = delta(abc1);
	  newLCC.flag = 2;

	  LCCTable.push_back(newLCC);
	}// end if (tri->Cab < 0)
	  

    }// fin for (i = 0; i <nbTris; i++)

}

void
MaskCalculation::printInitialisation(char * filename)
{
  unsigned int i;
  FILE * fp = fopen(filename, "w");

  fprintf(fp, "\n Points : \n");
  for (i = 0; i < myElmts->size(); i++)
    fprintf(fp, "%i.x : %i, %i.y : %i, %i.z : %i\n",
	    i, (*myElmts)[i].x,
	    i, (*myElmts)[i].y,
	    i, (*myElmts)[i].z);
  
  fprintf(fp, "\n Triangles\n");
  for (i = 0; i < myTris->size(); i++)
    fprintf(fp, "%i.A : %i, %i.B : %i, %i.C : %i,   %i.Abc : %i, %i.Bac : %i, %i.Cab : %i\n",
	    i, (*myTris)[i].A,
	    i, (*myTris)[i].B,
	    i, (*myTris)[i].C,
	    i, (*myTris)[i].Abc,
	    i, (*myTris)[i].Bac,
	    i, (*myTris)[i].Cab);
  
  fprintf(fp, "\n Conditions de convexite locale \n");
  for (i = 0; i < LCCTable.size(); i++)
    fprintf(fp, "w%i - %i.w%i - %i.w%i - %i.w%i >= 0\n",
	    (LCCTable[i]).w0, 
	    (LCCTable[i]).D1, (LCCTable[i]).w1,
	    (LCCTable[i]).D2, (LCCTable[i]).w2,
	    (LCCTable[i]).D3, (LCCTable[i]).w3);
  
  fprintf(fp, "\n Ordre des points\n");
  for(int k = 0; k < nbPts; k++)
    fprintf(fp, "%i %i %i,   ", (*myElmts)[k].x, (*myElmts)[k].y, (*myElmts)[k].z);

  fprintf(fp, "\n\n");

  fclose(fp);

}



void
MaskCalculation::preCalcError()
{
  chamferTriangle * tri;
  int xA, yA, zA, xB, yB, zB, xC, yC, zC;
  float xA1, yA1, zA1, xB1, yB1, zB1, xC1, yC1, zC1;
  int nbTris = myTris->size();

  for (int i = 0; i < nbTris; i++)
    {
      tri = &((*myTris)[i]);


      xA = (*myElmts)[tri->A].x;
      yA = (*myElmts)[tri->A].y;
      zA = (*myElmts)[tri->A].z;
      xB = (*myElmts)[tri->B].x;
      yB = (*myElmts)[tri->B].y;
      zB = (*myElmts)[tri->B].z;
      xC = (*myElmts)[tri->C].x;
      yC = (*myElmts)[tri->C].y;
      zC = (*myElmts)[tri->C].z;

      
      switch(tri->whichPlan)
	{
	case 0 : // projection sur le plan x = 1
	  tri->xA1 = (float) 1.0;
	  tri->xB1 = (float) 1.0;
	  tri->xC1 = (float) 1.0;
	  tri->yA1 = (float) yA / xA;
	  tri->zA1 = (float) zA / xA;
	  tri->yB1 = (float) yB / xB;
	  tri->zB1 = (float) zB / xB;
	  tri->yC1 = (float) yC / xC;
	  tri->zC1 = (float) zC / xC;
	  
	  xA1 = tri->xA1; yA1 = tri->yA1; zA1 = tri->zA1;
	  xB1 = tri->xB1; yB1 = tri->yB1; zB1 = tri->zB1;
	  xC1 = tri->xC1; yC1 = tri->yC1; zC1 = tri->zC1;

	  tri->denoAB = zB1 - zA1;
	  if (tri->denoAB != 0.0) {
	    tri->PAB = (yB1 - yA1) / tri->denoAB;
	    tri->QAB = (yA1*zB1 - yB1*zA1) / tri->denoAB;
	  }
	  else {
	    tri->PAB = 0.0;
	    tri->QAB = (float) tri->zA1;
	  }
      
	  tri->denoBC = zC1 - zB;
	  if (tri->denoBC != 0.0) {
	    tri->PBC = (yC1 - yB1) / tri->denoBC;
	    tri->QBC = (yB1*zC1 - yC1*zB1) / tri->denoBC;
	  }
	  else {
	    tri->PBC = 0.0;
	    tri->QBC = (float) tri->zB1;
	  }

	  tri->denoAC = zC1 - zA1;
	  if (tri->denoAC != 0.0) {
	    tri->PAC = (yC1 - yA1) / tri->denoAC;
	    tri->QAC = (yA1*zC1 - yC1*zA1) / tri->denoAC;
	  }
	  else {
	    tri->PAC = 0.0;
	    tri->QAC = (float) tri->zA1;
	  }
	  break;

	case 1 : // projection sur le plan y = 1
	  tri->xA1 = (float) xA / yA;
	  tri->xB1 = (float) xB / yB;
	  tri->xC1 = (float) xC / yC;
	  tri->yA1 = (float) 1.0;
	  tri->yB1 = (float) 1.0;
	  tri->yC1 = (float) 1.0;
	  tri->zA1 = (float) zA / yA;
	  tri->zB1 = (float) zB / yB;
	  tri->zC1 = (float) zC / yC;

	  xA1 = tri->xA1; yA1 = tri->yA1; zA1 = tri->zA1;
	  xB1 = tri->xB1; yB1 = tri->yB1; zB1 = tri->zB1;
	  xC1 = tri->xC1; yC1 = tri->yC1; zC1 = tri->zC1;

	  tri->denoAB = zB1 - zA1;
	  if (tri->denoAB != 0.0) {
	    tri->PAB = (xB1 - xA1) / tri->denoAB;
	    tri->QAB = (xA1*zB1 - xB1*zA1) / tri->denoAB;
	  }
	  else {
	    tri->PAB = 0.0;
	    tri->QAB = (float) tri->zA1;
	  }
      
	  tri->denoBC = zC1 - zB;
	  if (tri->denoBC != 0.0) {
	    tri->PBC = (xC1 - zB1) / tri->denoBC;
	    tri->QBC = (xB1*zC1 - xC1*zB1) / tri->denoBC;
	  }
	  else {
	    tri->PBC = 0.0;
	    tri->QBC = (float) tri->zB1;
	  }

	  tri->denoAC = zC1 - zA1;
	  if (tri->denoAC != 0.0) {
	    tri->PAC = (xC1 - xA1) / tri->denoAC;
	    tri->QAC = (xA1*zC1 - xC1*zA1) / tri->denoAC;
	  }
	  else {
	    tri->PAC = 0.0;
	    tri->QAC = (float) tri->zA1;
	  }
	  break;

	case 2 : // projection sur le plan z = 1
	  tri->xA1 = (float) xA / zA;
	  tri->xB1 = (float) xB / zB;
	  tri->xC1 = (float) xC / zC;
	  tri->yA1 = (float) yA / zA;
	  tri->yB1 = (float) yB / zB;
	  tri->yC1 = (float) yC / zC;
	  tri->zA1 = (float) 1.0;
	  tri->zB1 = (float) 1.0;
	  tri->zC1 = (float) 1.0;

	  xA1 = tri->xA1; yA1 = tri->yA1; zA1 = tri->zA1;
	  xB1 = tri->xB1; yB1 = tri->yB1; zB1 = tri->zB1;
	  xC1 = tri->xC1; yC1 = tri->yC1; zC1 = tri->zC1;

	  tri->denoAB = yB1 - yA1;
	  if (tri->denoAB != 0.0) {
	    tri->PAB = (xB1 - xA1) / tri->denoAB;
	    tri->QAB = (xA1*yB1 - xB1*yA1) / tri->denoAB;
	  }
	  else {
	    tri->PAB = 0.0;
	    tri->QAB = (float) tri->yA1;
	  }
      
	  tri->denoBC = yC1 - yB;
	  if (tri->denoBC != 0.0) {
	    tri->PBC = (xC1 - yB1) / tri->denoBC;
	    tri->QBC = (xB1*yC1 - xC1*yB1) / tri->denoBC;
	  }
	  else {
	    tri->PBC = 0.0;
	    tri->QBC = (float) tri->yB1;
	  }

	  tri->denoAC = yC1 - yA1;
	  if (tri->denoAC != 0.0) {
	    tri->PAC = (xC1 - xA1) / tri->denoAC;
	    tri->QAC = (xA1*yC1 - xC1*yA1) / tri->denoAC;
	  }
	  else {
	    tri->PAC = 0.0;
	    tri->QAC = (float) tri->yA1;
	  }
	  break;
	  
	default :
	  break;
	  
	} // end switch



      tri->nC[0][0] = yB*zC - yC*zB; 
      tri->nC[0][1] = yC*zA - yA*zC;
      tri->nC[0][2] = yA*zB - yB*zA;
      
      tri->nC[1][0] = xC*zB - xB*zC;
      tri->nC[1][1] = xA*zC - xC*zA;
      tri->nC[1][2] = xB*zA - xA*zB;

      tri->nC[2][0] = xB*yC - xC*yB;
      tri->nC[2][1] = xC*yA - xA*yC;
      tri->nC[2][2] = xA*yB - xB*yA;
    }
  
}

void
MaskCalculation::wheightsCalc(char * filename, int maxWheight, int minWheight)
{

  //  char newFileName[255];
  initLCC();
  preCalcError();

  //  sprintf(newFileName, "%s_init.txt", filename);
  //  printInitialisation(newFileName);

  for (int wheight = minWheight; wheight <= maxWheight; wheight++)
    {
      a = wheight;
      (*myElmts)[0].w = wheight;
      wheightsCalcRec(1);
      
    }
      

  for(int i = 0; i < nbPts; i++)
    {
      (*myElmts)[i].w = (wheightsTable[etIt])[i];
    }
}


void
MaskCalculation::wheightsCalcRec(int k)
{
  RelErrorElmt maxError;
  RelErrorElmt tmpError;

  int minW, maxW;

  //  minW = a * max(max((*myElmts)[k].x, (*myElmts)[k].y), (*myElmts)[k].z);
  //  maxW = a * ((*myElmts)[k].x + (*myElmts)[k].y + (*myElmts)[k].z);
  //  maxW = min(maxW, 255);
  
  minW = (int) (a * normeInf((*myElmts)[k]));
  maxW = (int) (a * norme1((*myElmts)[k]));
  maxW = min(maxW, 255);

  // Cas de terminaison
  if (k == nbPts - 1)
    {
      for (int wheight = minW; wheight <= maxW; wheight++)
	{
	  (*myElmts)[k].w = wheight;
	  // Le poids de tous les points sont affects
	  
	  if (localConvexityCond(k)) // Si le masque rationnel quivalent est convexe
	    {
	      // On calcule l'erreur relative maximale du masque
	      relativeErrorCalc((*myTris)[0], &maxError);

	      for (unsigned int triangle = 1; triangle < (*myTris).size(); triangle++)
		{
		  relativeErrorCalc((*myTris)[triangle], &tmpError);
		  if (tmpError.tauMin < maxError.tauMin)
		    maxError.tauMin = tmpError.tauMin;
		  if (tmpError.tauMax > maxError.tauMax)
		    maxError.tauMax = tmpError.tauMax;
		}

	      maxError.tau = max(fabs(maxError.tauMin), fabs(maxError.tauMax));
	      maxError.epsOpt = a*((maxError.tauMin + maxError.tauMax)/2 + 1);
	      maxError.tauOpt = (a/(maxError.epsOpt))*(maxError.tauMax + 1) - 1;
	      // On stoke les poids et l'erreur de ce masque 
	      // (uniquement si l'erreur globale est infrieure  celles trouves prcdemment)

	      //	      if ((maxError.tau < errorsTable[etIt].tau) ||
	      //		  (maxError.tauOpt < errorsTable[etIt].tauOpt))
	      //	      if (maxError.tau < errorsTable[etIt].tau)
	      if (maxError.tauOpt < errorsTable[etIt].tauOpt)
		{
		  int * tabWeight = (int*) malloc(nbPts*sizeof(int));
		  if(tabWeight == NULL) {
		    printf("Impossible d'allourer tabWeight\n");
		    return;
		  }
		  for (int p = 0; p < nbPts; p++)
		    tabWeight[p] = (*myElmts)[p].w;
		  wheightsTable.push_back(tabWeight);
		  errorsTable.push_back(maxError);

// for (int q = 0; q < nbPts; q++)
//   fprintf(stderr, "%i, ", tabWeight[q]);
// fprintf(stderr, "\t %f \t %f \t %f \t %f \t %f\n", 
//        maxError.tauOpt, maxError.tau, maxError.tauMin, maxError.tauMax, maxError.epsOpt);
		  
		  etIt++;
		}
	      
	    }
	} //fin for (int wheight = 1; wheight <= 10; wheight++)
    } //fin if (k == nbPts - 1)
  
  else
    {
      for (int wheight = minW; wheight <= maxW; wheight++)
	{
	  (*myElmts)[k].w = wheight;
	  if (localConvexityCond(k))
	    wheightsCalcRec(k + 1);
	}
    }
}



int
MaskCalculation::delta(chamferTriangle tri)
{
  int res;
  int xA = (*myElmts)[tri.A].x;
  int yA = (*myElmts)[tri.A].y;
  int zA = (*myElmts)[tri.A].z;
  
  int xB = (*myElmts)[tri.B].x;
  int yB = (*myElmts)[tri.B].y;
  int zB = (*myElmts)[tri.B].z;
  
  int xC = (*myElmts)[tri.C].x;
  int yC = (*myElmts)[tri.C].y;
  int zC = (*myElmts)[tri.C].z;
  
  // On applique la regle de Sarrus.
  res = xA*yB*zC + xB*yC*zA + xC*yA*zB - xC*yB*zA - xB*yA*zC - xA*yC*zB;
  
  return res;      
}


bool
MaskCalculation::localConvexityCond(int k)
{
  int expr;
  int nbConds = LCCTable.size();
  int w0, w1, w2, w3;


  for (int i = 0; i < nbConds; i++)
    {
      // if one coefficient is greater than k,
      // the condition connot be tested
      if ((LCCTable[i].w0 > k) ||
	  (LCCTable[i].w1 > k) ||
	  (LCCTable[i].w2 > k) ||
	  (LCCTable[i].w3 > k))
	LCCTable[i].flag = 2;

      // all coefficient are smaller than k
      // if one is equal to k, the kst coefficient has changed
      // the condition must be re-tested
      else if ((LCCTable[i].w0 == k) ||
	       (LCCTable[i].w1 == k) ||
	       (LCCTable[i].w2 == k) ||
	       (LCCTable[i].w3 == k))
	{
	  w0 = (*myElmts)[LCCTable[i].w0].w;
	  w1 = (*myElmts)[LCCTable[i].w1].w;
	  w2 = (*myElmts)[LCCTable[i].w2].w;
	  w3 = (*myElmts)[LCCTable[i].w3].w;
	  
	  expr = w0 - LCCTable[i].D1 * w1 - LCCTable[i].D2 * w2 - LCCTable[i].D3 * w3;
	  if (expr >= 0)
	    LCCTable[i].flag = 1;
	  else
	    {
	      LCCTable[i].flag = 0;
	      return false;
	    }
	}
      else // all coefficient are strictly smaller than k 
	{  // this condition must have been already tested and must be true
	  //	  if ((LCCTable[i] == 2) || (LCCTable[i] == 0)) {
	  if (LCCTable[i].flag != 1) {
	    printf("Warning, this case should not occur\n");
	    return false;
	  }
	}
    }
  
  return true;
}


unsigned char ***
MaskCalculation::getFinalMask()
{
  return myTriangulation->editFinalMask();
}

float 
MaskCalculation::getEpsilon()
{
  return (errorsTable[etIt]).epsOpt;
}

float
MaskCalculation::getMaxRelativeError()
{
  return (errorsTable[etIt]).tauOpt;
}
