/******************************************************************************/
/* MaskCalculation.h                                                          */
/******************************************************************************/
#ifndef MASKCALCULATION_H
#define MASKCALCULATION_H

#ifdef _YAV_ // namespace for INRIA
#include "chamferTriangleStruct.h"
#include "MaskTri.h"
namespace yav {
#else
#include "distancemap/chamferTriangleStruct.h"
#include "distancemap/MaskTri.h"
#endif //_YAV_


class MaskCalculation
{
public :
  // Constructor
  MaskCalculation();

  // Destructor
  virtual ~MaskCalculation();

  void printMask(char * filename);

  void wheightsCalc(char * filename, int maxWheight, int minWheight = 1);
  void initLCC();

  unsigned char *** getFinalMask();
  float getEpsilon();
  float getMaxRelativeError();

  void printInitialisation(char * filename);

protected :

  bool localConvexityCond(int k);
  void preCalcError();
  int delta(chamferTriangle tri);
  

  virtual float normeInf(chamferTriElmt elmt) = 0;
  virtual float norme1(chamferTriElmt elmt) = 0;


  void wheightsCalcRec(int k);
  virtual void relativeErrorCalc(chamferTriangle tri, RelErrorElmt * errElmt)=0;

  /// Fields 
  MaskTri * myTriangulation;
  
  unsigned int etIt;
  std::vector < int * > wheightsTable;
  std::vector < RelErrorElmt > errorsTable;
 

  std::vector < LCCElmt > LCCTable;
  /// vector where triangulation triangles are stored
  std::vector < chamferTriangle > * myTris;
  /// vector where points of the triangulation are stored
  std::vector < chamferTriElmt > *  myElmts;

  int nbPts;

  int a;


};

#ifdef _YAV_
} // end namespace yav
#endif

#endif // MASKCALCULATION_H
