/******************************************************************************/
/* MaskTri.h                                                                  */
/******************************************************************************/

#ifndef MASKTRI_H
#define MASKTRI_H

#include <vector>
#ifdef _YAV_
#include "chamferTriangleStruct.h"
namespace yav {
#else
#include "distancemap/chamferTriangleStruct.h"
#endif // _YAV_
  


  /** MaskTri class : virtual class to compute chamfer triangulations 
      used for chamfer mask coefficients determination */ 
class MaskTri
{
  friend class MaskCalculation;
  friend class MaskCalc;
  friend class MaskCalcAnisotropZ;
  friend class MaskCalcAnisotropYZ;

public :

  /** @name Constructors */
  //@{
  ///
  MaskTri(int _order);
  ///@}

  /** @name Destructor */
  //@{ 
  /// 
  virtual ~MaskTri();
  //@} 

  /** @name accessors */
  ///
  int getNbPts();
  ///
  int getNbTri();
  ///
  void deleteTri(int index);

  /** @name Other methods */

  /// 
  virtual unsigned char *** editFinalMask();

  ///
  virtual void printTriangulation() = 0;


protected :
  /** returns the indice of elmt in fareyElmts,
      returns -1 if elmt is not in fareyElmts.
  **/
  int isInSet(chamferTriElmt elmt);
 
  //  void initNC();
 
  /// vector where triangulation triangles are stored
  std::vector < chamferTriangle > myTris;
  /// vector where points of the triangulation are stored
  std::vector < chamferTriElmt >  myElmts;

  /** The nbPts first points of myElmts are the points of the chamfer mask generator. 
      After nbPts, myElmts' points are possible symetric points of the mask generator 
      used to compute local convexity conditions 
  **/
  int nbPts;
  /// Iterator of myElmts
  int elmtIt;
  /// fareyTris iterator
  int triIt;

  /// Order of the triangulation (i.e. width of the mask)
  int order;
  /// Size of the final mask
  int maskSize;
  /// Final mask
  unsigned char *** finalMask;
};

#ifdef _YAV_
} //namespace yav
#endif

#endif // FAREYTRI_H
