#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/MultiView4D.h"
#include <inrimage/MultiView4D.h>
#include <inrimage/TclInrimage4DBackground.h>

namespace yav {


MultiView4D::MultiView4D(TclModule *m,
			 const std::list<TclTimeSliceCamera *>& _views,
			 const TclString& window)
  : TclSon(m, window), TclObject(typeid(MultiView4D), m), DPOwner<TclTimeSliceCamera>(_views),
    axis(false), zoom(1.0), ignoreVoxelSize(false),
    axisColor("red"), axisColor2("cyan"), dieWhenEmpty(true)  {
}



MultiView4D::MultiView4D(TclModule *m, const TclString& window)
  : TclSon(m, window), TclObject(typeid(MultiView4D), m), DPOwner<TclTimeSliceCamera>(),
    axis(false), zoom(1.0), ignoreVoxelSize(false),
    axisColor("red"), axisColor2("cyan"), dieWhenEmpty(true) {
}



MultiView4D::~MultiView4D() {
  DPList<TclTimeSliceCamera>::iterator it;
  for(it = elements.begin(); it != elements.end(); it=elements.begin()) {
    {
      TclTimeSliceCamera *slicecam=(*it);
      remove(slicecam);
      delete slicecam;
    }
  }
}


void MultiView4D::dying(const TclTimeSliceCamera *elt) {
   if(elements.size() == 0 && dieWhenEmpty) delete this;
}


void MultiView4D::append(TclTimeSliceCamera *inrv)
  throw(DPList<TclTimeSliceCamera>::NullPointer,
	DPList<TclTimeSliceCamera>::AlreadyInList) {
  double slice = -1.0;

  if(elements.empty()) {
    zoom = inrv->getClient()->getScale();
    TclInrimage4DBackground *b=dynamic_cast<TclInrimage4DBackground *> (inrv->getClient()->getBackground());
    if (b)
      {
	ignoreVoxelSize = b->getIVS();
      }
  }
  else {
    inrv->getClient()->scaleCamera(zoom / inrv->getClient()->getScale());
    TclInrimage4DBackground *b=dynamic_cast<TclInrimage4DBackground *> (inrv->getClient()->getBackground());
    if (b)
      {
	b->setIVS(ignoreVoxelSize);
      }
    DPList<TclTimeSliceCamera>::iterator it;
    for(it = elements.begin(); it != elements.end() && slice < 0.0; it++) {
      if((*it)->getClient()->getGeometry() == inrv->getClient()->getGeometry())
	slice = (*it)->getClient()->getSlice();
    }
    if(slice >= 0.0) inrv->getClient()->setSlice(slice);
  }

  getModule()->tickle("if [winfo exists %s] {bind %s <Motion> {%s%u -mousevoxel %s%u %%x %%y}}",
		      (char *) inrv->getCanvas(),(char *) inrv->getCanvas(), getObjectName(), getRef(),
		      inrv->getObjectName(), inrv->getRef());
  getModule()->tickle("bind %s <ButtonPress-2> {%s%u -changeview %s%u %%x %%y; bind %s <Motion> {%s%u -changeview %s%u %%%%x %%%%y}}",
		      (char *) inrv->getCanvas(), getObjectName(), getRef(),
		      inrv->getObjectName(), inrv->getRef(),
		      (char *) inrv->getCanvas(), getObjectName(), getRef(),
		      inrv->getObjectName(), inrv->getRef());
  getModule()->tickle("bind %s <ButtonRelease-2> {bind %s <Motion> {%s%u -mousevoxel %s%u %%%%x %%%%y}}",
		      (char *) inrv->getCanvas(), (char *) inrv->getCanvas(),
		      getObjectName(), getRef(),
		      inrv->getObjectName(), inrv->getRef());
  DPOwner<TclTimeSliceCamera>::append(inrv);
}


void MultiView4D::remove(TclTimeSliceCamera *inrv)
  throw(DPList<TclTimeSliceCamera>::NotFound) {

  try {
    DPOwner<TclTimeSliceCamera>::remove(inrv);
    getModule()->tickle("if [winfo exists %s] {bind %s <ButtonPress-2> {} }",
			(char *) inrv->getCanvas(),(char *) inrv->getCanvas());
    getModule()->tickle("if [winfo exists %s] {bind %s <ButtonRelease-2> {} }",
			(char *) inrv->getCanvas(),(char *) inrv->getCanvas());
    getModule()->tickle("if [winfo exists %s] {bind %s <Motion> {} }",
			(char *) inrv->getCanvas(),(char *) inrv->getCanvas());
    getModule()->tickle("if [winfo exists %s] {%s delete axis}",
			(char *) inrv->getCanvas(),(char *) inrv->getCanvas());
    inrv->bindMouseMotion(true);
  }
  catch(DPList<TclTimeSliceCamera>::NotFound& e) {
    e.push("MultiView4D::operator-=(TclTimeSliceCamera *)");
    throw e;    
  }
}


void MultiView4D::setAxis(bool _axis) {
  axis = _axis;
  if(!axis) {
    DPList<TclTimeSliceCamera>::iterator it;
    for(it = elements.begin(); it != elements.end(); it++) {
      getModule()->tickle("%s delete axis",
			  (char *) (*it)->getCanvas());
    }
  }
}


void MultiView4D::setScale(double val) {
  if(val > 0.0) {  
    getModule()->tickle("set graphics2D::busy 1");
    zoom *= val; 
    double w;
    DPList<TclTimeSliceCamera>::iterator it;
    for(it = elements.begin(); it != elements.end(); it++) {
      w=(*it)->getClient()->getWidth();
      (*it)->getClient()->resize((unsigned int)((*it)->getClient()->getWidth() * val),
				 (unsigned int)((*it)->getClient()->getHeight() * val));
      (*it)->getClient()->scaleCamera((*it)->getClient()->getWidth()/w);
    } 
    getModule()->tickle("set graphics2D::busy 0");
  }
}


void MultiView4D::setIVS(bool val) {
  ignoreVoxelSize = val;
  DPList<TclTimeSliceCamera>::iterator it;
  for(it = elements.begin(); it != elements.end(); it++)
    {
      TclInrimage4DBackground *b=dynamic_cast<TclInrimage4DBackground *> ((*it)->getClient()->getBackground());
      if (b)
	{
	  b->setIVS(val);
	}
    }
}

void MultiView4D::activation(void) { }

void MultiView4D::on(void) {
  DPList<TclTimeSliceCamera>::iterator it;
  for(it = elements.begin(); it != elements.end(); it++)
    if ((*it)->getClient()->getScaleCamera())
      (*it)->getClient()->getScaleCamera()->on();
  Activable::on();
}

void MultiView4D::off(void) {
  DPList<TclTimeSliceCamera>::iterator it;
  for(it = elements.begin(); it != elements.end(); it++)
    if ((*it)->getClient()->getScaleCamera())
      (*it)->getClient()->getScaleCamera()->off();
  Activable::off();
}


int MultiView4D::destroyCmd(int argc, char **argv) {
  TclModule *m;
  unsigned int ref;

  if(argc != 0) return COMMAND_SYNOPSIS;
  m = getModule();
  ref = getRef();
  delete this;
  m->setResult("%s%u does not exist any longer",
	       TclObjectKind::getObjectName(typeid(MultiView4D)), ref);
  return TCL_OK;
}

int MultiView4D::addCmd(int argc, char **argv) {
  if(argc < 1) return COMMAND_SYNOPSIS;

  int i = 0;
  TclString res;
  while(i < argc) {
    TclTimeSliceCamera *inrv;
    inrv = TclTimeSliceCamera::getObject(argc, argv, i);
    if(!inrv) return TCL_ERROR;
    try {
      *this += inrv;
      res += TclString(argv[i]) + " ";
    }
    catch(DPList<TclTimeSliceCamera>::AlreadyInList&) {
    }
    i++;
  }
  res += "added";
  setResult(res);
  return TCL_OK;
}


int MultiView4D::removeCmd(int argc, char **argv) {
  if(argc < 1) return COMMAND_SYNOPSIS;

  int i = 0;
  TclString res;
  while(i < argc) {
    TclTimeSliceCamera *inrv;
    inrv = TclTimeSliceCamera::getObject(argc, argv, i);
    if(!inrv) return TCL_ERROR;
    try {
      *this -= inrv;
    }
    catch(DPList<TclTimeSliceCamera>::NotFound e) {
      setResult("%s%u not handled by %s%u", inrv->getObjectName(), inrv->getRef(),
		getObjectName(), getRef());
      return TCL_ERROR;
    }
    res += TclString(argv[i]) + SPACE_STRING;
    i++;
  }
  res += "supressed";
  setResult(res);
  return TCL_OK;
}


int MultiView4D::viewsCmd(int argc, char **argv) {
  TclString res, one;
  DPList<TclTimeSliceCamera>::iterator it;

  if(argc > 1) return COMMAND_SYNOPSIS;
  if(argc == 0) {
    for(it = elements.begin(); it != elements.end(); it++) {
      one.sprintf("%s%u ", (*it)->getObjectName(), (*it)->getRef());
      res += one;
    }
  }

  else {
    // -views X
    if(!strcmp(argv[0], "X")) {
      for(it = elements.begin(); it != elements.end(); it++) {
	if((*it)->getClient()->getGeometry() == TimeSliceCamera::SLICE_X) {
	  one.sprintf("%s%u ", (*it)->getObjectName(), (*it)->getRef());
	  res += one;
	}
      }
    }
    // -views Y
    else if(!strcmp(argv[0], "Y")) {
      for(it = elements.begin(); it != elements.end(); it++) {
	if((*it)->getClient()->getGeometry() == TimeSliceCamera::SLICE_Y) {
	  one.sprintf("%s%u ", (*it)->getObjectName(), (*it)->getRef());
	  res += one;
	}
      }
    }
    // -views Z
    else if(!strcmp(argv[0], "Z")) {
      for(it = elements.begin(); it != elements.end(); it++) {
	if((*it)->getClient()->getGeometry() == TimeSliceCamera::SLICE_Z) {
	  one.sprintf("%s%u ", (*it)->getObjectName(), (*it)->getRef());
	  res += one;
	}
      }
    }
    // -views alpha
    else if(!strcmp(argv[0], "alpha")) {
      for(it = elements.begin(); it != elements.end(); it++) {
	if((*it)->getClient()->getGeometry() == TimeSliceCamera::SLICE_ALPHA) {
	  one.sprintf("%s%u ", (*it)->getObjectName(), (*it)->getRef());
	  res += one;
	}
      }
    }
    else if(!strcmp(argv[0], "XT")) {
      for(it = elements.begin(); it != elements.end(); it++) {
	if((*it)->getClient()->getGeometry() == TimeSliceCamera::SLICE_XT) {
	  one.sprintf("%s%u ", (*it)->getObjectName(), (*it)->getRef());
	  res += one;
	}
      }
    }
    else if(!strcmp(argv[0], "YT")) {
      for(it = elements.begin(); it != elements.end(); it++) {
	if((*it)->getClient()->getGeometry() == TimeSliceCamera::SLICE_YT) {
	  one.sprintf("%s%u ", (*it)->getObjectName(), (*it)->getRef());
	  res += one;
	}
      }
    }
    else if(!strcmp(argv[0], "ZT")) {
      for(it = elements.begin(); it != elements.end(); it++) {
	if((*it)->getClient()->getGeometry() == TimeSliceCamera::SLICE_ZT) {
	  one.sprintf("%s%u ", (*it)->getObjectName(), (*it)->getRef());
	  res += one;
	}
      }
    }
    else return COMMAND_SYNOPSIS;
  }

  setResult(res);
  return TCL_OK;
}


int MultiView4D::mousevoxelCmd(int argc, char **argv) {
  if(argc != 3) return COMMAND_SYNOPSIS;
  TclTimeSliceCamera *inrv=0;
  DPList<TclTimeSliceCamera>::iterator it; 
  for(it = elements.begin(); it != elements.end(); it++) {
    if ((*it)->name()==(const char *)argv[0])
      inrv=(*it);
  }
  if (!inrv)
    return TCL_ERROR;
    
  int mx, my;
  if(sscanf(argv[1], "%i", &mx) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[2], "%i", &my) != 1) return COMMAND_SYNOPSIS;

  
  TimeSliceCamera::TIME_GEOMETRY geom = inrv->getClient()->getGeometry();
  TclInrimage4DBackground *bg=(TclInrimage4DBackground *)inrv->getClient()->getBackground();
  VoxelCoordinate4D *vc=0;
  if (bg)
    vc=bg->coordinates(mx, my);  
  for(it = elements.begin(); it != elements.end(); it++) {
    if((axis)&& (vc)) (*it)->drawAxis(geom,mx,my,vc->x,vc->y,vc->z,vc->t, axisColor, axisColor2);
    if((vc)&&((*it)->getClient()->getGeometry() == geom)) (*it)->displayValue(((TclInrimage4DBackground *)(*it)->getClient()->getBackground())->getLabel(mx,my));
  }
  
  if (vc)
    {
      setResult("%u %u %u", vc->x, vc->y, vc->z);
      delete vc;
    }
  return TCL_OK;

}
int MultiView4D::sceneCmd(int argc, char **argv) {
  if (argc!=0)  return COMMAND_SYNOPSIS;
  DPList<TclTimeSliceCamera>::iterator it=elements.begin();
  if (it != elements.end())
    {
      setResult("%s",(char *)(*it)->getScene()->name());
      return TCL_OK;
    }
  else
    {
      setResult("none");
      return TCL_OK;
    }
}
int MultiView4D::changeViewCmd(int argc, char **argv) {
  if(argc != 3) return COMMAND_SYNOPSIS;
  TclTimeSliceCamera *inrv=0;
  DPList<TclTimeSliceCamera>::iterator it; 
  for(it = elements.begin(); it != elements.end(); it++) {
    if ((*it)->name()==(const char *)argv[0])
      inrv=(*it);
  }
  if (!inrv)
    return TCL_ERROR;

  int mx, my;
  if(sscanf(argv[1], "%i", &mx) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[2], "%i", &my) != 1) return COMMAND_SYNOPSIS;

  TimeSliceCamera::TIME_GEOMETRY geom = inrv->getClient()->getGeometry();
  TclInrimage4DBackground *bg=(TclInrimage4DBackground *)inrv->getClient()->getBackground();
  VoxelCoordinate4D *vc=0;
  if (bg)
    vc=bg->coordinates(mx, my);  
  for(it = elements.begin(); it != elements.end(); it++) {
    if (vc) (*it)->changeView(vc->x,vc->y,vc->z,vc->t);
    if((axis)&& (vc)) (*it)->drawAxis(geom,mx,my,vc->x,vc->y,vc->z,vc->t, axisColor, axisColor2);
  }
  delete vc;
  setResult("%u %u %u %u", vc->x, vc->y, vc->z,vc->t);

  return TCL_OK;
}


int MultiView4D::axisCmd(int argc, char **argv) {
  bool bval;

  if(argc > 1) return COMMAND_SYNOPSIS;
  if(argc == 1) {
    if(TclModule::boolValue(argv[0], bval)) {
      setAxis(bval);
    }
    else return COMMAND_SYNOPSIS;
  }
  setResult(TclModule::boolString(getAxis()));
  return TCL_OK;
}


int MultiView4D::axisColorCmd(int argc, char **argv) {
  if(argc != 0 && argc != 2) return COMMAND_SYNOPSIS;
  if(argc == 2) {
    setAxisColors(argv[0], argv[1]);
  }
  setResult("%s %s", (char *) getAxisColor1(), (char *) getAxisColor2());
  return TCL_OK;
}


int MultiView4D::activateCmd(int argc, char **argv) {
  bool bval;

  if(argc > 1) return COMMAND_SYNOPSIS;
  if(argc == 1) {
    if(!TclModule::boolValue(argv[0], bval)) return COMMAND_SYNOPSIS;
    if(bval) on();
    else off();
  }
  setResult(TclModule::boolString(active()));
  return TCL_OK;
}


int MultiView4D::ignoreVoxelSizeCmd(int argc, char **argv) {
  bool bval;

  if(argc > 1) return COMMAND_SYNOPSIS;
  if(argc == 1) {
    if(!TclModule::boolValue(argv[0], bval)) return COMMAND_SYNOPSIS;
    setIVS(bval);
  }
  setResult(TclModule::boolString(getIVS()));
  return TCL_OK;
}


int MultiView4D::scaleCmd(int argc, char **argv) {
  double val;

  if(argc > 2) return COMMAND_SYNOPSIS;
  if(argc > 0) {
    if(argv[0][0] == '*' || argv[0][0] == '/') {
      if(strlen(argv[0]) > 1) {
	if(sscanf(argv[0] + 1, "%lf", &val) != 1) return COMMAND_SYNOPSIS;
	if(argv[0][0] == '/') val = 1.0 / val;
	setScale(val);
      }
      else {
	if(argc > 1) {
	  if(sscanf(argv[1], "%lf", &val) != 1) return COMMAND_SYNOPSIS;
	  if(argv[0][0] == '/') val = 1.0 / val;
	  setScale(val);
	}
	else return COMMAND_SYNOPSIS;
      }
    }
    else {
      if(sscanf(argv[0], "%lf", &val) != 1) return COMMAND_SYNOPSIS;
      setScale(val / getScale());
    }
  }
  setResult("%f", getScale());
  return TCL_OK;
}


int MultiView4D::redrawCmd(int argc, char **argv) {
  if(argc > 0) return COMMAND_SYNOPSIS;

  DPList<TclTimeSliceCamera>::iterator it;
  for(it = elements.begin(); it != elements.end(); it++) {
    tickle("%s%u -redraw", (*it)->getObjectName(), (*it)->getRef());
  }
  return TCL_OK;
}


int MultiView4D::minCmd(int argc, char **argv) {
  if(argc != 1) return COMMAND_SYNOPSIS;

  DPList<TclTimeSliceCamera>::iterator it;
  for(it = elements.begin(); it != elements.end(); it++) {
    tickle("%s%u -min %s", (*it)->getObjectName(),
	   argv[0]);
  }
  return TCL_OK;
}


int MultiView4D::maxCmd(int argc, char **argv) {
  if(argc != 1) return COMMAND_SYNOPSIS;

  DPList<TclTimeSliceCamera>::iterator it;
  for(it = elements.begin(); it != elements.end(); it++) {
    tickle("%s%u -max %s", (*it)->getObjectName(), (*it)->getRef(),
	   argv[0]);
  }
  return TCL_OK;
}


int MultiView4D::inrimageCmd(int argc, char **argv) {
  if(argc != 1) return COMMAND_SYNOPSIS;

  DPList<TclTimeSliceCamera>::iterator it;
  for(it = elements.begin(); it != elements.end(); it++) {
    tickle("%s%u -inrimage %s", (*it)->getObjectName(), (*it)->getRef(),
	   argv[0]);
  }
  return TCL_OK;
}
int MultiView4D::parentWidgetCmd(int argc, char **argv) {
  if(argc != 1) return COMMAND_SYNOPSIS;

  setTCLwindow(argv[0]);
  return TCL_OK;
}


int MultiView4D::visibleCmd(int argc, char **argv) {
  if(argc != 1 && argc != 8) return COMMAND_SYNOPSIS;

  DPList<TclTimeSliceCamera>::iterator it;
  if(argc == 1) {
    if(strcmp(argv[0], "all")) return COMMAND_SYNOPSIS;
    for(it = elements.begin(); it != elements.end(); it++)
      (*it)->getClient()->getBackground()->setVisibleArea();
  }    
  else {
    unsigned int v[8];
    if(sscanf(argv[0], "%u", &v[0]) != 1) return COMMAND_SYNOPSIS;
    if(sscanf(argv[1], "%u", &v[1]) != 1) return COMMAND_SYNOPSIS;
    if(sscanf(argv[2], "%u", &v[2]) != 1) return COMMAND_SYNOPSIS;
    if(sscanf(argv[3], "%u", &v[3]) != 1) return COMMAND_SYNOPSIS;
    if(sscanf(argv[4], "%u", &v[4]) != 1) return COMMAND_SYNOPSIS;
    if(sscanf(argv[5], "%u", &v[5]) != 1) return COMMAND_SYNOPSIS;
    if(sscanf(argv[6], "%u", &v[6]) != 1) return COMMAND_SYNOPSIS;
    if(sscanf(argv[7], "%u", &v[7]) != 1) return COMMAND_SYNOPSIS;
    for(it = elements.begin(); it != elements.end(); it++)
      (*it)->getClient()->getBackground()->setVisibleArea(v[0], v[1],
							  v[2], v[3],
							  v[4], v[5],
							  v[6], v[7]);
  }
  return TCL_OK;
}


int MultiView4D::dieCmd(int argc, char **argv) {
  if(argc>1) return COMMAND_SYNOPSIS;
  if(argc==1){
    if(!TclModule::boolValue(argv[0],dieWhenEmpty)) return  COMMAND_SYNOPSIS;
  }
  setResult(TclModule::boolString(dieWhenEmpty));
  return TCL_OK;
}


/* "Create a new multiviewer object. -in %w: the given TCL window is associated to the multi-viewer and will be destroyed at the same time. -add slicecam%u+: add all given viewers to the multiviewer" */
int MultiView4D::create(TclModule *m, int argc, char **argv) {
  MultiView4D *mv;
  std::list<TclTimeSliceCamera *> inrvl;
  char *name = 0;
  int i;

  for(i = 1; i < argc; i++) {
    // -in %w
    if(!strcmp(argv[i], "-in")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(!m->windowExists(argv[i])) return COMMAND_SYNOPSIS;
      name = argv[i];
    }

    // -add inrview%u...
    else if(!strcmp(argv[i], "-add")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      TclTimeSliceCamera *inrv;
      while(i < argc && argv[i][0] != '-') {
	inrv = TclTimeSliceCamera::getObject(argc, argv, i);
	if(!inrv) return TCL_ERROR;
	inrvl.push_back(inrv);
	i++;
      }
    }

    else return COMMAND_SYNOPSIS;
  }

  mv = new MultiView4D(m, inrvl, name);
  m->setResult("%s%u", (char *) mv->getObjectName(), mv->getRef());
  return TCL_OK;
}


} // namespace yav


