#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/OpacityScale.h"
#include <inrimage/OpacityScale.h>
#include <graphics2D/TclCamera2D.h>

namespace yav {

OpacityScale::OpacityScale(TclModule *module, const TclString& scaleName,Scalable *elt)
  : TclScale(typeid(OpacityScale), scaleName,module,elt,"inrimage::opacityscalemenu") {
    init();
}
 

OpacityScale::OpacityScale(TclModule *module, const TclString& scaleName,
		 std::list<Scalable *>& list)
  : TclScale(typeid(OpacityScale), scaleName, module,list,"inrimage::opacityscalemenu") {
  init();
}


OpacityScale::~OpacityScale() {
}


void OpacityScale::init(void) {
  max=1.0;
  min=0.0;
  firstupdate = false;
  tickle("bind %s <ButtonRelease-1> {%s%u -command [%s get]};\
bind %s <Enter> \"focus %s\";\
bind %s <Left> {%s%u -command [expr [%s get] - 1]};\
bind %s <Right> {%s%u -command [expr [%s get] + 1]};\
bind %s <Down> {%s%u -command [expr [%s get] - 1]};\
bind %s <Up> {%s%u -command [expr [%s get] + 1]}",
		 (char *) TCLwindow, getObjectName(),
		 getRef(), (char *) TCLwindow,
		 (char *) TCLwindow, (char *) TCLwindow,
		 (char *) TCLwindow, getObjectName(),
		 getRef(), (char *) TCLwindow,
		 (char *) TCLwindow, getObjectName(),
		 getRef(), (char *) TCLwindow,
		 (char *) TCLwindow, getObjectName(),
		 getRef(), (char *) TCLwindow,
		 (char *) TCLwindow, getObjectName(),
		 getRef(), (char *) TCLwindow);
  tickle("bind %s <ButtonRelease-2> [bind %s <ButtonRelease-1>]",
		 (char *) TCLwindow, (char *) TCLwindow);
}


void OpacityScale::reset(Scalable *sa, const TclString& lab) {

  
  tickle("%s configure -from 0 -to 1 -resolution 0.01",
	 (char *) TCLwindow);

  if(lab != EMPTY_STRING) setLabel(lab);
  command(get());
  value = 0.5;
  sa->scaleValue(this,value);
  tickle("%s set %f", (char *) TCLwindow, value);
}


void OpacityScale::set(double val) {
  to(val);
  command(val);
}


void OpacityScale::activation(void) {
  TclScale::activation();
  command(value);
  /* artificially ask to redraw camera in order to create movies */
  DPList<Scalable>::const_iterator it;
  const DPList<Scalable>& lst = getObjects();
  for(it = lst.begin(); it != lst.end(); it++) {
#ifdef OLD_CODE
    TclOverlayBackground *b=dynamic_cast<TclOverlayBackground *> (*it);
    if (b)
      {
	if ((TclCamera2D *)(b->getCamera()->getInterface())->isRecording()) {
	  b->getCamera()->updateDisplay();
	}
      }
#endif
  }
  
}

/* "-in %w [-add slicecam%u+]",
   "Create a new opacity level scale object associated to given TCL scale (-in argument) and add any given slicecam to scale. Only slicecam with the same type can be added to a opacityscale."*/
int OpacityScale::create(TclModule *m, int argc, char **argv) {
  OpacityScale *gsc;
  std::list<Scalable *> inrvl;
  char *name = 0;
  int i;

  for(i = 1; i < argc; i++) {
    // -in %w
    if(!strcmp(argv[i], "-in")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(!m->windowExists(argv[i])) return COMMAND_SYNOPSIS;
      name = argv[i];
    }

    // -add inrview%u
    else if(!strcmp(argv[i], "-add")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      while(i < argc && argv[i][0] != '-') {
	Scalable *scalable=0;
	Scalable *sc=AutoLink<Scalable>::getHead();
	for(;sc;sc=sc->AutoLink<Scalable>::following())
	  if (sc->objectName()==(const char *)argv[i])
	    scalable=sc;
	if (scalable==0)
	  return TCL_ERROR;
	i++;
      }
    }

    else return COMMAND_SYNOPSIS;
  }

  if(!name ) return COMMAND_SYNOPSIS;
  gsc = new OpacityScale(m, name, inrvl);
  m->setResult("%s%u", (char *) gsc->getObjectName(), gsc->getRef());
  return TCL_OK;
}



} // namespace yav

