#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/OverlayInrimage4DView.h"

#ifdef GRAPHICS
#include <inrimage/OverlayInrimage4DView.h>
#ifdef TCL_INTERFACE
#include <graphics2D/TclTimeSliceCamera.h>
#include <inrimage/TclInrimage4D.h>
#include <inrimage/TclOverlayInrimage4DView.h>
#include <inrimage/TclOverlay4DBackground.h>
#endif

namespace yav {

OverlayInrimage4DView::NoImage::NoImage(const std::string& from)
  : Exception(from,"provided an empty image while attempting to fuse two images") {
}
OverlayInrimage4DView::MismatchedImageType::MismatchedImageType(const std::string& from)
  : Exception(from,"attempt to fuse two images having different types") {
}
OverlayInrimage4DView::MismatchedImageSize::MismatchedImageSize(const std::string& from)
  : Exception(from,"attempt to fuse two images having different sizes") {
}
OverlayInrimage4DView::MismatchedImageTime::MismatchedImageTime(const std::string& from)
  : Exception(from,"attempt to fuse two images having different time dimensions") {
}
OverlayInrimage4DView::OverlayInrimage4DView(Inrimage4D *inr1, Inrimage4D *inr2,
					 Scene2D *scene,bool _createGLScale) 
  throw(OverlayInrimage4DView::MismatchedImageType,OverlayInrimage4DView::MismatchedImageSize,OverlayInrimage4DView::MismatchedImageTime,OverlayInrimage4DView::NoImage)
  : Inrimage4DView(inr1,inr2,scene,_createGLScale),secondImage(this,inr2) {
  if (!inr1)
    throw(OverlayInrimage4DView::NoImage("OverlayInrimageView::OverlayInrimageView(Inrimage *,Inrimage *,Scene2D *,bool)"));
  if (!inr2)
    throw(OverlayInrimage4DView::NoImage("OverlayInrimageView::OverlayInrimageView(Inrimage *,Inrimage *,Scene2D *,bool)"));
  if (inr1->getType()!=inr2->getType())
    throw(OverlayInrimage4DView::MismatchedImageType("OverlayInrimageView::OverlayInrimageView(Inrimage *,Inrimage *,Scene2D *,bool)"));
  if ((inr1->getX()!=inr2->getX()) || (inr1->getY()!=inr2->getY()) || (inr1->getZ()!=inr2->getZ()) || (inr1->getVdim()!=inr2->getVdim()))
    throw(OverlayInrimage4DView::MismatchedImageSize("OverlayInrimageView::OverlayInrimageView(Inrimage *,Inrimage *,Scene2D *,bool)"));
  if (inr1->getTimeDimension() != inr2->getTimeDimension())
    throw(OverlayInrimage4DView::MismatchedImageTime("OverlayInrimageView::OverlayInrimageView(Inrimage *,Inrimage *,Scene2D *,bool)"));
  init();
}
  

  
void OverlayInrimage4DView::changeInrimage(Inrimage4D *inr) {
  if (inr!=pointer()) {
    if ((inr->getX()==secondImage()->getX()) && (inr->getY()==secondImage()->getY()) &&
	(inr->getZ()==secondImage()->getZ()) && (inr->getType()==secondImage()->getType())) {
      pointer=inr;
      std::list<Inrimage4DBackground *>::iterator it;
      for(it = backgroundList.begin(); it != backgroundList.end(); it++) {
	(*it)->inrimageChanged();
      }
    }
  }
}
void OverlayInrimage4DView::changeSecondInrimage(Inrimage4D *inr) {
  if (inr!=secondImage()) {
    if ((inr->getX()==pointer()->getX()) && (inr->getY()==pointer()->getY()) &&
	(inr->getZ()==pointer()->getZ()) && (inr->getType()==pointer()->getType())) {
      secondImage=inr;
      std::list<Inrimage4DBackground *>::iterator it;
      for(it = backgroundList.begin(); it != backgroundList.end(); it++) {
	(*it)->inrimageChanged();
      }
    }
  }
}


void OverlayInrimage4DView::doCreateCamera(SliceCamera& cam) {
  cameras[&cam] = new Sliceable::CamInfo();
  double m;

  if(pointer->getVX() < pointer->getVY()) m = pointer->getVX();
  else m = pointer->getVY();
  if(pointer->getVZ() < m) m = pointer->getVZ();
  //  cam.scaleCamera(1.0 / (m * cam.getScale()));
  cam.scaleCamera(1.0 / (m));
  switch(((TimeSliceCamera &)cam).getGeometry()) {
  case TimeSliceCamera::SLICE_X:
  case TimeSliceCamera::SLICE_XT:
    cam.setThickness(pointer->getVX());
    break;
  case TimeSliceCamera::SLICE_Y:
  case TimeSliceCamera::SLICE_YT:
    cam.setThickness(pointer->getVY());
    break;
  case TimeSliceCamera::SLICE_Z:
  case TimeSliceCamera::SLICE_ZT:
  case TimeSliceCamera::SLICE_ALPHA:
    cam.setThickness(pointer->getVZ());
    break;
  }
#ifdef TCL_INTERFACE
  TclInrimage4DBackground *b;
  if (createGLScale)
    b = new TclOverlay4DBackground(cam.getInterface()->getModule(),this,(TimeSliceCamera *)&cam);
  else
    b = new TclOverlay4DBackground(getInterface()->getModule(),this,(TimeSliceCamera *)&cam,false,0,false,0,false,0,false,0,false,0);
  backgroundList.push_back(b);  
#endif
  cam.focusOnBackground();
}


#ifdef TCL_INTERFACE
TclOverlayInrimage4DView *OverlayInrimage4DView::getInterface(void) const {
	return (TclOverlayInrimage4DView *) Client::getInterface();
}
#endif



} // namespace yav

#endif
