#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/OverlayInrimageView.h"

#ifdef GRAPHICS
#include <inrimage/OverlayInrimageView.h>
#ifdef TCL_INTERFACE
#include <graphics2D/TclSliceCamera.h>
#include <inrimage/TclInrimage.h>
#include <inrimage/TclOverlayInrimageView.h>
#include <inrimage/TclOverlayBackground.h>
#endif

namespace yav {


OverlayInrimageView::NoImage::NoImage(const std::string& from)
  : Exception(from,"provided an empty image while attempting to fuse two images") {
}
OverlayInrimageView::MismatchedImageType::MismatchedImageType(const std::string& from)
  : Exception(from,"attempt to fuse two images having different types") {
}
OverlayInrimageView::MismatchedImageSize::MismatchedImageSize(const std::string& from)
  : Exception(from,"attempt to fuse two images having different sizes") {
}
OverlayInrimageView::OverlayInrimageView(Inrimage *inr1, Inrimage *inr2,
					 Scene2D *scene,bool _createGLScale) 
  throw(OverlayInrimageView::MismatchedImageType,OverlayInrimageView::MismatchedImageSize,OverlayInrimageView::NoImage)
  : InrimageView(inr1,inr2,scene,_createGLScale), 
    secondImage(this,inr2) {
  if (!inr1)
    throw(OverlayInrimageView::NoImage("OverlayInrimageView::OverlayInrimageView(Inrimage *,Inrimage *,Scene2D *,bool)"));
  if (!inr2)
    throw(OverlayInrimageView::NoImage("OverlayInrimageView::OverlayInrimageView(Inrimage *,Inrimage *,Scene2D *,bool)"));
  if (inr1->getType()!=inr2->getType())
    throw(OverlayInrimageView::MismatchedImageType("OverlayInrimageView::OverlayInrimageView(Inrimage *,Inrimage *,Scene2D *,bool)"));
  if ((inr1->getX()!=inr2->getX()) || (inr1->getY()!=inr2->getY()) || (inr1->getZ()!=inr2->getZ()) || (inr1->getVdim()!=inr2->getVdim()))
    throw(OverlayInrimageView::MismatchedImageSize("OverlayInrimageView::OverlayInrimageView(Inrimage *,Inrimage *,Scene2D *,bool)"));

  InrimageSliceInterface *any;
  for(any = AutoLink<InrimageSliceInterface>::getHead(); 
      any && (any->getInrimage()!=inr2); any = any->following());
  if (any)
    secondSliceInterface=any;
  else 
    secondSliceInterface= new InrimageSliceInterface(inr2);

  init();
}
  

  
void OverlayInrimageView::changeInrimage(Inrimage *inr) {
  if (inr!=pointer()) {
    if ((inr->getX()==secondImage()->getX()) && (inr->getY()==secondImage()->getY()) &&
	(inr->getZ()==secondImage()->getZ()) && (inr->getType()==secondImage()->getType())) {
      InrimageView::changeInrimage(inr);
      
    }
  }
}
void OverlayInrimageView::changeSecondInrimage(Inrimage *inr) {
  if (inr!=secondImage()) {
    if ((inr->getX()==pointer()->getX()) && (inr->getY()==pointer()->getY()) &&
	(inr->getZ()==pointer()->getZ()) && (inr->getType()==pointer()->getType())) {
      secondImage=inr;
      std::list<InrimageBackground *>::iterator it;
      for(it = backgroundList.begin(); it != backgroundList.end(); it++) {
	(*it)->inrimageChanged();
      }
      InrimageSliceInterface *any;
      for(any = AutoLink<InrimageSliceInterface>::getHead(); 
	  any && (any->getInrimage()!=(Inrimage *)inr); any = any->following());
      if (any)
	secondSliceInterface=any;
      else 
	secondSliceInterface= new InrimageSliceInterface(inr);
    }
  }
}


void OverlayInrimageView::doCreateCamera(SliceCamera& cam) {
  cameras[&cam] = new Sliceable::CamInfo();
  double m;

  if(pointer->getVX() < pointer->getVY()) m = pointer->getVX();
  else m = pointer->getVY();
  if(pointer->getVZ() < m) m = pointer->getVZ();
  //  cam.scaleCamera(1.0 / (m * cam.getScale()));
  //  cam.scaleCamera(1.0 / (m));
  switch(cam.getGeometry()) {
  case SliceCamera::SLICE_X:
    cam.setThickness(pointer->getVX());
    break;
  case SliceCamera::SLICE_Y:
    cam.setThickness(pointer->getVY());
    break;
  case SliceCamera::SLICE_Z:
  case SliceCamera::SLICE_ALPHA:
    cam.setThickness(pointer->getVZ());
    break;
  }
#ifdef TCL_INTERFACE
  TclInrimageBackground *b;
  if (createGLScale)
    b = new TclOverlayBackground(cam.getInterface()->getModule(),this,&cam);
  else
    b = new TclOverlayBackground(getInterface()->getModule(),this,&cam,false,0,false,0,false,0,false,0,false,0);
  backgroundList.push_back(b);  
#endif
  cam.focusOnBackground();
}


#ifdef TCL_INTERFACE
TclOverlayInrimageView *OverlayInrimageView::getInterface(void) const {
	return (TclOverlayInrimageView *) Client::getInterface();
}
#endif



} // namespace yav

#endif
