#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#include <inrimage/ROIVoxelList.h>
#include <inrimage/ImageGeometry.h>
//#include <inrimage/ConnectingElement.h>
//#include <inrimage/Image3DCameraBackground.h>

#include <vector>
#if ((defined(__GNUC__)) && (__GNUC__ <3 ))
#include <algo.h>
#else
#include <algorithm>
#endif

namespace yav {



void ROIVoxelList::initFormat() {  
  // Sets the necessary module to read and write a file
  GeneralFormat::addModuleVersion("Main", true, 1.0,
				  (GeneralFormatModuleVersion::INPUTREADFUNCTION) &ROIVoxelList::readMain,
				  (GeneralFormatModuleVersion::INPUTSKIPFUNCTION) &ROIVoxelList::skipMain,
				  (GeneralFormatModuleVersion::OUTPUTWRITEFUNCTION) &ROIVoxelList::writeMain);
  realName="Region of Interest";
}

ROIVoxelList::ROIVoxelList(char *filename) :   GeneralFormat("ImageROI File","IMAGEROI"),Owner<Inrimage>(0),slice(0) {
  initFormat();
  GeneralFormat::read(filename,0);
}

ROIVoxelList::ROIVoxelList(const std::list<ROIVoxel *> &vl,
			 Vec3<double> vsize, Vec3<unsigned int> _imageSize,
			   Inrimage *im): GeneralFormat("ImageROI File","IMAGEROI"),Owner<Inrimage>(im), voxelSize(vsize), slice(0),imageSize(_imageSize) {
  initFormat(); 
  if (im) {
    voxelSize= Vec3<double> (im->getVX(),im->getVY(), im->getVZ());
    imageSize=Vec3<unsigned int> (im->getX(),im->getY(),im->getZ());
  }
  buildVoxelMap(vl);
}
ROIVoxelList::ROIVoxelList(Inrimage *im) throw (ROIVoxelList::ImageMissing): GeneralFormat("ImageROI File","IMAGEROI"),Owner<Inrimage>(im), slice(0) {
  initFormat();
  if (!im)
    throw(ROIVoxelList::ImageMissing("ROIVoxelList::ROIVoxelList(const Inrimage *inr)"));
  voxelSize= Vec3<double> (im->getVX(),im->getVY(), im->getVZ());
  imageSize=Vec3<unsigned int> (im->getX(),im->getY(),im->getZ());

}
ROIVoxelList::ROIVoxelList(Vec3<double> vsize,Vec3<unsigned int> _imageSize): GeneralFormat("ImageROI File","IMAGEROI"), Owner<Inrimage>(0),voxelSize(vsize), slice(0),imageSize(_imageSize) {
  initFormat();
}
ROIVoxelList::ROIVoxelList(Inrimage *inr, 
			   const Inrimage *bin) 
  throw(ROIVoxelList::ImageMissing,ROIVoxelList::BinaryImageWrongSize) : GeneralFormat("ImageROI File","IMAGEROI"),Owner<Inrimage>(inr),slice(0)
{
  if ((!inr) || (!bin))
    throw(ROIVoxelList::ImageMissing("ROIVoxelList::ROIVoxelList(const Inrimage *inr,const Inrimage *bin)"));
  if ((bin->getX()!=inr->getX()) || (bin->getY()!=inr->getY()) || (bin->getZ()!=inr->getZ())) {
    throw(ROIVoxelList::BinaryImageWrongSize("ROIVoxelList::ROIVoxelList(const Inrimage *inr,const Inrimage *bin)"));    
  }
  voxelSize= Vec3<double> (inr->getVX(),inr->getVY(), inr->getVZ());
  imageSize=Vec3<unsigned int> (inr->getX(),inr->getY(),inr->getZ());
  
  Voxel *greyVoxel= Voxel::voxelByType(inr->getType());
  Voxel *binaryVoxel= Voxel::voxelByType(bin->getType());

  unsigned int index,nplanes,dimx;

  dimx=getPointer()->getX();
  nplanes=getPointer()->getX()*getPointer()->getY();

  for(unsigned int z = 0; z < inr->getZ(); z++) {
    for(unsigned int y = 0; y < inr->getY(); y++) {
      for(unsigned int x = 0; x < inr->getX(); x++) {
	bin->getVoxel(x,y,z,*binaryVoxel);
	if (binaryVoxel->getValueAsDouble()>0.0)
	  {
	    inr->getVoxel(x,y,z,*greyVoxel);
	    ROIVoxel *voxel= new ROIVoxel();
	    voxel->voxelValue=greyVoxel->getValueAsDouble();
	    voxel->position=Vec3<unsigned int>(x,y,z);
	    index=x + y *dimx +z*nplanes;
	    voxelMap[index]=voxel;
	  }
      }
    }
  }
  delete greyVoxel;
  delete binaryVoxel;

  initFormat();
}
void ROIVoxelList::dying(const ObjectPointer<Inrimage>& obj){
  pointer=0;
}

ROIVoxelList::~ROIVoxelList() {
  if (slice)
    delete slice;
  std::map<unsigned int,ROIVoxel *>::iterator itv;	  
  for(itv=voxelMap.begin();itv!=voxelMap.end();itv++)
    {
      delete (*itv).second;
    }
}
void ROIVoxelList::readMain(GFInput &input) 
{
  std::list<ROIVoxel *> voxelList;
  input >> realName;
  input >> voxelSize;
  input >> imageSize;
  unsigned int i,nvoxels;
  input >> nvoxels;
  double val;
  Vec3<unsigned int> pos;
  ROIVoxel *voxel;
  for(i=0;i<nvoxels;++i)
    {
      input >> val >> pos;
      voxel= new ROIVoxel();
      voxel->voxelValue=val;
      voxel->position=pos;
      voxelList.push_back(voxel);
    }
  buildVoxelMap(voxelList);
}
void ROIVoxelList::skipMain(GFInput &input) 
{
  Vec3<double> v;
  input >> v;
  unsigned int i,nvoxels;
  input >> nvoxels;
  double val;
  Vec3<unsigned int> pos;
  for(i=0;i<nvoxels;++i)
    {
      input >> val >> pos;
    }
  
}
void ROIVoxelList::writeMain(GFOutput &output) 
{
  output << realName << ENDL;
  output << voxelSize << ENDL;
  output << imageSize << ENDL;
  output << (int)voxelMap.size() << ENDL;
  std::map<unsigned int,ROIVoxel *>::iterator itv;  
  for(itv=voxelMap.begin();itv!=voxelMap.end();itv++) {
    output << (*itv).second->voxelValue << (*itv).second->position << ENDL;
  }
  
}
GenericHistogram *ROIVoxelList::createHistogram() const {
  std::map<unsigned int,ROIVoxel *>::const_iterator it;
  double key;
  std::map< double, unsigned long int> histo;
  std::map< double, unsigned long int>::iterator ith;
  for(it=getFirstVoxel();it!=getLastVoxel();it++) {
    key=(*it).second->voxelValue;
    ith=histo.find(key);
    if (ith==histo.end())
      {
	histo[key]=1;
      }
    else
      {
	histo[key]=((*ith).second)+1;
      }
  }
  GenericHistogram *h= new GenericHistogram(histo);
  histo.erase(histo.begin(),histo.end());
  return h;
}
Inrimage *ROIVoxelList::createBinaryImage() const {
  if (!getPointer())
    return 0;
  else {
    Inrimage *bin=new Inrimage(getPointer()->getX(), getPointer()->getY(),
			       getPointer()->getZ(),
			       Inrimage::WT_UNSIGNED_CHAR);
    memset(bin->getData(),0,bin->getX()*bin->getY()*bin->getZ());
	bin->setVoxelSize(getPointer()->getVX(),getPointer()->getVY(),getPointer()->getVZ());
    unsigned char ***array=(unsigned char ***)bin->getArray();
    std::map<unsigned int,ROIVoxel *>::const_iterator itv=voxelMap.begin();
    for(;itv!=voxelMap.end();itv++) {
      array[(*itv).second->position[2]][(*itv).second->position[1]][(*itv).second->position[0]]=255;
    }
    return bin;
  }
}
void ROIVoxelList::dilate(CONNECTIVITY_TYPE connectivity, unsigned int niter) {
  dilateOrErode(connectivity,niter,true);
}
void ROIVoxelList::erode(CONNECTIVITY_TYPE connectivity, unsigned int niter) {
  dilateOrErode(connectivity,niter,false);
}
void ROIVoxelList::buildVoxelMap(const std::list<ROIVoxel *> &voxelList) { 
  std::list<ROIVoxel *>::const_iterator itv=voxelList.begin();
  unsigned int index,nplanes,dimx;

  dimx=imageSize[0];
  nplanes=imageSize[0]*imageSize[1];

  voxelMap.erase(voxelMap.begin(),voxelMap.end());
  for(;itv!=voxelList.end();itv++) {
    index=(*itv)->position[0] + (*itv)->position[1] *dimx +
      (*itv)->position[2]*nplanes;
    voxelMap[index]=(*itv);
  }
}
void ROIVoxelList::getCenterAndRadius(Vec3<double> &center, double &radius) const 
{
	std::map<unsigned int,ROIVoxel *>::const_iterator it;
	center= Vec3<double>();
	Vec3<unsigned int> voxelPos;
	Vec3<double> pos3D;
	double p[3];
	unsigned int i;
	double dist;

	ImageGeometry *converter = getPointer()->newConverter();

	for(i=0,it=voxelMap.begin();it!=voxelMap.end();it++,i++) {
		voxelPos=(*it).second->position;
		converter->setImage((double)voxelPos.getX(),(double)voxelPos.getY(),
			(double)voxelPos.getZ());
		converter->toCartesian(p);
		pos3D=Vec3<double>(p);
		center+=pos3D;
	}
	center/=(double) i;
	radius= 0;
	for(it=voxelMap.begin();it!=voxelMap.end();it++) {
		voxelPos=(*it).second->position;
		converter->setImage((double)voxelPos.getX(),(double)voxelPos.getY(),
			(double)voxelPos.getZ());
		converter->toCartesian((double *)p);
		pos3D=Vec3<double>(p);
		dist = (pos3D-center).norm2();
		if (dist>radius)
			radius=dist;
	}
	radius=sqrt(radius);
}

void ROIVoxelList::dilateOrErode(CONNECTIVITY_TYPE connectivity, 
				 unsigned int niter, bool dilate) {
  if (getPointer()) {
    std::map<unsigned int,ROIVoxel *> voxelDilatedMap;
    unsigned int index,nplanes,dimx,i,j;

    dimx=getPointer()->getX();
    nplanes=getPointer()->getX()*getPointer()->getY();
  
    std::vector< Vec3<int> > offsetVector(26);
    unsigned int nbOffsets=0;
    switch(connectivity) {
    case FOUR_CONNECTIVITY:
      nbOffsets=4;
      offsetVector[0]=Vec3<int>(1,0,0);
      offsetVector[1]=Vec3<int>(0,1,0);
      offsetVector[2]=Vec3<int>(-1,0,0);
      offsetVector[3]=Vec3<int>(0,-1,0);
      break;
    case EIGHT_CONNECTIVITY:
      nbOffsets=8;
      offsetVector[0]=Vec3<int>(1,0,0);
      offsetVector[1]=Vec3<int>(0,1,0);
      offsetVector[2]=Vec3<int>(-1,0,0);
      offsetVector[3]=Vec3<int>(0,-1,0);
      offsetVector[4]=Vec3<int>(1,1,0);
      offsetVector[5]=Vec3<int>(-1,-1,0);
      offsetVector[6]=Vec3<int>(-1,1,0);
      offsetVector[7]=Vec3<int>(1,-1,0);
      break;
    case TEN_CONNECTIVITY:
    case SIX_CONNECTIVITY:
      nbOffsets=6;
      offsetVector[0]=Vec3<int>(1,0,0);
      offsetVector[1]=Vec3<int>(0,1,0);
      offsetVector[2]=Vec3<int>(-1,0,0);
      offsetVector[3]=Vec3<int>(0,-1,0);
      offsetVector[4]=Vec3<int>(0,0,1);
      offsetVector[5]=Vec3<int>(0,0,-1);
      break;
    case EIGHTEEN_CONNECTIVITY:
      nbOffsets=18;
      offsetVector[0]=Vec3<int>(1,0,0);
      offsetVector[1]=Vec3<int>(0,1,0);
      offsetVector[2]=Vec3<int>(-1,0,0);
      offsetVector[3]=Vec3<int>(0,-1,0);
      offsetVector[4]=Vec3<int>(0,0,1);
      offsetVector[5]=Vec3<int>(0,0,-1);

      offsetVector[6]=Vec3<int>(1,1,0);
      offsetVector[7]=Vec3<int>(-1,-1,0);
      offsetVector[8]=Vec3<int>(-1,1,0);
      offsetVector[9]=Vec3<int>(1,-1,0);

      offsetVector[10]=Vec3<int>(1,0,1);
      offsetVector[11]=Vec3<int>(-1,0,1);
      offsetVector[12]=Vec3<int>(-1,0,-1);
      offsetVector[13]=Vec3<int>(1,0,-1);

      offsetVector[14]=Vec3<int>(0,1,1);
      offsetVector[15]=Vec3<int>(0,-1,1);
      offsetVector[16]=Vec3<int>(0,-1,-1);
      offsetVector[17]=Vec3<int>(0,1,-1);
    case TWENTY_SIX_CONNECTIVITY:
      nbOffsets=26;
      offsetVector[0]=Vec3<int>(1,0,0);
      offsetVector[1]=Vec3<int>(0,1,0);
      offsetVector[2]=Vec3<int>(-1,0,0);
      offsetVector[3]=Vec3<int>(0,-1,0);
      offsetVector[4]=Vec3<int>(0,0,1);
      offsetVector[5]=Vec3<int>(0,0,-1);

      offsetVector[6]=Vec3<int>(1,1,0);
      offsetVector[7]=Vec3<int>(-1,-1,0);
      offsetVector[8]=Vec3<int>(-1,1,0);
      offsetVector[9]=Vec3<int>(1,-1,0);

      offsetVector[10]=Vec3<int>(1,0,1);
      offsetVector[11]=Vec3<int>(-1,0,1);
      offsetVector[12]=Vec3<int>(-1,0,-1);
      offsetVector[13]=Vec3<int>(1,0,-1);

      offsetVector[14]=Vec3<int>(0,1,1);
      offsetVector[15]=Vec3<int>(0,-1,1);
      offsetVector[16]=Vec3<int>(0,-1,-1);
      offsetVector[17]=Vec3<int>(0,1,-1);

      offsetVector[18]=Vec3<int>(1,1,1);
      offsetVector[19]=Vec3<int>(1,-1,1);
      offsetVector[20]=Vec3<int>(1,-1,-1);
      offsetVector[21]=Vec3<int>(1,1,-1);

      offsetVector[22]=Vec3<int>(-1,1,1);
      offsetVector[23]=Vec3<int>(-1,-1,1);
      offsetVector[24]=Vec3<int>(-1,-1,-1);
      offsetVector[25]=Vec3<int>(-1,1,-1);

      break;

    }

    int x,y,z;
    Voxel *v= Voxel::voxelByType(getPointer()->getType());
    ROIVoxel *rv;
    std::map<unsigned int,ROIVoxel *>::iterator it,testVoxel;
    
    if (dilate) {
      for(i=0;i<niter;++i) {
	for(it=voxelMap.begin();it!=voxelMap.end();it++) {
	  for(j=0;j<nbOffsets;++j) {
	    x=(*it).second->position[0]+offsetVector[j][0];
	    y=(*it).second->position[1]+offsetVector[j][1];
	    z=(*it).second->position[2]+offsetVector[j][2];
	    
	    if (getPointer()->inside(x,y,z)) {
	      index=x+y*dimx+z*nplanes;
	      testVoxel=voxelMap.find(index);
	      
	      if (testVoxel==voxelMap.end()) {
		rv= new ROIVoxel();
		rv->position=Vec3<unsigned int>(x,y,z);
		getPointer()->getVoxel(x,y,z,*v);
		rv->voxelValue=v->getValueAsDouble();
		voxelDilatedMap[index]=rv;
	      }
	    }
	  }
	}
	for(it=voxelDilatedMap.begin();it!=voxelDilatedMap.end();it++) {
	  voxelMap[(*it).first]=(*it).second;
	}
	voxelDilatedMap.erase(voxelDilatedMap.begin(),voxelDilatedMap.end());
      }
    }
    else {
      //      std::list < std::map<unsigned int, ROIVoxel *>::iterator > removeList;
      for(i=0;i<niter;++i) {
	for(it=voxelMap.begin();it!=voxelMap.end();it++) {
	  for(j=0;j<nbOffsets;++j) {
	    x=(*it).second->position[0]+offsetVector[j][0];
	    y=(*it).second->position[1]+offsetVector[j][1];
	    z=(*it).second->position[2]+offsetVector[j][2];
	    
	    if (getPointer()->inside(x,y,z)) {
	      index=x+y*dimx+z*nplanes;
	      testVoxel=voxelMap.find(index);
	      
	      if (testVoxel==voxelMap.end()) {
		//		removeList.push_back(it);
		voxelDilatedMap[(*it).first]=(*it).second;
	      }
	    }
	  }
	}
	for(it=voxelDilatedMap.begin();it!=voxelDilatedMap.end();it++) {
	  testVoxel=voxelMap.find((*it).first);
	  if (testVoxel!=voxelMap.end())
	    voxelMap.erase(testVoxel);
	}
	voxelDilatedMap.erase(voxelDilatedMap.begin(),voxelDilatedMap.end());
      }
    }
  }
}
void ROIVoxelList::computeMeanVariance(double &mean, double &variance, 
				      double rejectRate) const{
  std::map<unsigned int,ROIVoxel *>::const_iterator it;
  double val=0;
  double vv;
  for(it=getFirstVoxel();it!=getLastVoxel();it++) {
    val+=(*it).second->voxelValue;
  }
  val/=(double)voxelMap.size();
  if (rejectRate>0)
    {
      std::vector< std::pair< double, double> > residual;
      std::vector< std::pair< double, double> >::iterator itc;

      unsigned int n=(unsigned int)(voxelMap.size()*(1.0-rejectRate));
      unsigned int i,j;
      for(j=0;j<3;++j) {
	/// create the vector of residual from the mean
	residual.erase(residual.begin(),residual.end());
	vv=(*it).second->voxelValue;
	for(i=0,it=getFirstVoxel();it!=getLastVoxel();it++,i++) {
	  if (i==n)
	    itc=residual.insert(residual.end(),std::pair<double,double>(fabs(vv-val),vv));
	  else
	    residual.push_back(std::pair<double,double>(fabs(vv-val),vv));
	}
	// sort vector array
	std::nth_element(residual.begin(),itc,residual.end());
	// compute the mean from the smallest residuals
	val=0;
	for(i=0;i<n;++i)
	  {
	    val+=residual[i].second;
	  }
	val/=(double)n;

	//	cerr << i << " " << n << " " <<  val << endl;

      }
      // set mean value
      mean=val;
      // 
      residual.erase(residual.begin(),residual.end());
      for(i=0,it=getFirstVoxel();it!=getLastVoxel();it++,i++) {
	if (i==n)
	    itc=residual.insert(residual.end(),std::pair<double,double>(fabs((*it).second->voxelValue-val),(*it).second->voxelValue ));
	else
	  residual.push_back(std::pair<double,double>(fabs((*it).second->voxelValue-val),(*it).second->voxelValue ));
      }
      // sort vector array
      std::nth_element(residual.begin(),itc,residual.end());
      // compute the variance from the smallest residuals
      variance=0;
      for(itc=residual.begin(),i=0;i<n;++i,itc++)
	{
	  variance+=(*itc).first*(*itc).first;
	}
      variance/=(double)n;
    }
  else
    {
      mean=val;
      variance=0;
      unsigned int i;
      for( i=0,it=getFirstVoxel();it!=getLastVoxel();it++,i++) {
	val=(*it).second->voxelValue-mean;
	variance+=val*val;
      }
      variance/=(double)voxelMap.size();
    }
  
  

}
void ROIVoxelList::write(TclString fname)
{
  if (fname!=EMPTY_STRING)
    {
      filename=fname;
    }
  GeneralFormat::write(( const char *)filename);
}


void ROIVoxelList::addVoxel(std::list< Vec3<unsigned int>  > &plist) {
  if (getPointer()) { 
    Inrimage *inr=getPointer();
    Voxel *greyVoxel= Voxel::voxelByType(inr->getType());
    unsigned int dimx,nplanes,index;
    std::list< Vec3<unsigned int>  >::iterator it;

    dimx=getPointer()->getX();
    nplanes=getPointer()->getX()*getPointer()->getY();
    
    for(it=plist.begin();it!=plist.end();it++) {
      ROIVoxel *voxel= new ROIVoxel();
      inr->getVoxel((*it)[0],(*it)[1],(*it)[2],*greyVoxel);
      if (greyVoxel) {
	voxel->voxelValue=greyVoxel->getValueAsDouble();
	voxel->position=(*it);
    
	index=(*it)[0] + (*it)[1] *dimx + (*it)[2]*nplanes;
	voxelMap[index]=voxel;
      } else 
      delete voxel;
    }
  }
}
void ROIVoxelList::addVoxel(std::list<ROIVoxel *> vlist) {
  std::list<ROIVoxel *>::iterator itv;
  
  unsigned int dimx,nplanes,index;

  dimx=getPointer()->getX();
  nplanes=getPointer()->getX()*getPointer()->getY();
  
  for(itv=vlist.begin();itv!=vlist.end();itv++) {
    if (*itv) {
      index=(*itv)->position[0] + (*itv)->position[1] *dimx +
	(*itv)->position[2]*nplanes;
      voxelMap[index]=(*itv);
      
    }
  }
}
bool ROIVoxelList::removeVoxel(std::list< Vec3<unsigned int>  > &plist) {
  unsigned int dimx,nplanes,index;

  dimx=getPointer()->getX();
  nplanes=getPointer()->getX()*getPointer()->getY();
  std::list< Vec3<unsigned int>  >::iterator it;
  std::map<unsigned int,ROIVoxel *>::iterator itm;
  
  bool out=true;
  for(it=plist.begin();it!=plist.end();it++) {
    index=(*it)[0] + (*it)[1] *dimx + (*it)[2]*nplanes;
    itm=voxelMap.find(index);
    if (itm!=voxelMap.end()) {
      voxelMap.erase(itm);
    }
    else
      out=false;
  }
  return out;
}

ROIVoxel *ROIVoxelList::getVoxel(const unsigned int x,const unsigned int y,const unsigned int z) const {
  unsigned int index;
  index=x+getPointer()->getX()*y+getPointer()->getX()*getPointer()->getY()*z;
  std::map<unsigned int,ROIVoxel *>::const_iterator itm;   
  itm=voxelMap.find(index);
  if (itm!=voxelMap.end()) 
    return (*itm).second;
  else
    return 0;
}

}
