#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/StructuringElement.h"

#include <inrimage/StructuringElement.h>
#include <inrimage/arithmetic.h>

namespace yav {


#define _morpho_ErrorInParameters_ -1
#define _morpho_ErrMemFull_ -2




#include "_predef_morpho.C"
#include "_usrdef_morpho.C"
#include "_median.C"
#include "_watershed.C"



// Number of iterations to perform to simulate a dilation with an euclidian element
static int _iterations_04connectivity( int R )
{
  switch( R ) {
  default : return( -1 );
  case 1 : return( 0 );
  case 2 : return( 1 );
  case 3 :
  case 4 : return( 2 );
  case 5 : return( 3 );
  case 6 :
  case 7 : return( 4 );
  case 8 :
  case 9 : return( 5 );
  case 10 : return( 6 );
  case 11 : return( 7 );
  case 12 :
  case 13 :
  case 14 : return( 8 );
  case 15 : return( 9 );
  case 16 :
  case 17 : return( 10 );
  case 18 : return( 11 );
  case 19 :
  case 20 : return( 12 );
  case 21 : return( 13 );
  case 22 : return( 11 );
  case 23 :
  case 24 : return( 12 );
  case 25 :
  case 26 : return( 16 );
  case 27 : return( 14 );
  case 28 : return( 15 );
  case 29 :
  case 30 : return( 16 );
  case 31 :
  case 32 : return( 17 );
  case 33 : return( 18 );
  case 34 : return( 19 );
  case 35 :
  case 36 :
  case 37 : return( 20 );
  case 38 : return( 21 );
  case 39 :
  case 40 : return( 22 );
  case 41 : return( 23 );
  case 42 :
  case 43 : return( 24 );
  case 44 :
  case 45 : return( 25 );
  case 46 : return( 26 );
  case 47 : return( 27 );
  case 48 :
  case 49 :
  case 50 : return( 28 );
    break;
  }
  return( -1 );
}


static int _iterations_08connectivity( int R )
{
  switch( R ) {
  default : return( -1 );
  case 1 :
  case 2 :
  case 3 : return( 1 );
  case 4 :
  case 5 :
  case 6 : return( 2 );
  case 7 :
  case 8 : return( 3 );
  case 9 :
  case 10 :
  case 11 :
  case 12 : return( 4 );
  case 13 : return( 5 );
  case 14 :
  case 15 :
  case 16 : return( 6 );
  case 17 :
  case 18 :
  case 19 : return( 7 );
  case 20 :
  case 21 : return( 8 );
  case 22 :
  case 23 : return( 10 );
  case 24 : return( 11 );
  case 25 : return( 9 );
  case 26 : return( 10 );
  case 27 :
  case 28 :
  case 29 : return( 12 );
  case 30 :
  case 31 : return( 13 );
  case 32 :
  case 33 :
  case 34 :
  case 35 : return( 14 );
  case 36 : return( 15 );
  case 37 :
  case 38 :
  case 39 : return( 16 );
  case 40 :
  case 41 :
  case 42 : return( 17 );
  case 43 :
  case 44 : return( 18 );
  case 45 :
  case 46 :
  case 47 :
  case 48 : return( 19 );
  case 49 : return( 20 );
  case 50 : return( 21 );
    break;
  }
  return( -1 );
}

static int _iterations_06connectivity( int R )
{
  switch( R ) {
  default : return( -1 );
  case 1 : return( 0 );
  case 2 : return( 1 );
  case 3 :
  case 4 : return( 2 );
  case 5 :
  case 6 : return( 3 );
  case 7 : return( 2 );
  case 8 : return( 5 );
  case 9 : return( 3 );
  case 10 : return( 4 );
  case 11 : return( 5 );
  case 12 :
  case 13 : return( 6 );
  case 14 :
  case 15 : return( 7 );
  case 16 : return( 8 );
  case 17 :
  case 18 : return( 9 );
  case 19 : return( 10 );
  case 20 :
  case 21 : return( 11 );
  case 22 : return( 12 );
  case 23 :
  case 24 : return( 13 );
  case 25 : return( 14 );
  case 26 : return( 12 );
  case 27 : return( 13 );
  case 28 : return( 16 );
  case 29 : return( 14 );
  case 30 :
  case 31 : return( 15 );
  case 32 : return( 16 );
  case 33 :
  case 34 : return( 17 );
  case 35 : return( 18 );
  case 36 :
  case 37 : return( 19 );
  case 38 : return( 20 );
  case 39 :
  case 40 : return( 21 );
  case 41 : return( 22 );
  case 42 :
  case 43 : return( 23 );
  case 44 : return( 24 );
  case 45 : return( 25 );
  case 46 : return( 23 );
  case 47 :
  case 48 : return( 24 );
  case 49 : return( 25 );
  case 50 : return( 26 );
    break;
  }
  return( -1 );
}

static int _iterations_18connectivity( int R )
{
  switch( R ) {
  default : return( -1 );
  case 1 : return( 1 );
  case 2 :
  case 3 : return( 0 );
  case 4 : return( 2 );
  case 5 : return( 0 );
  case 6 :
  case 7 : return( 2 );
  case 8 : return( 0 );
  case 9 : return( 3 );
  case 10 : return( 2 );
  case 11 : return( 1 );
  case 12 : return( 0 );
  case 13 : return( 2 );
  case 14 : return( 1 );
  case 15 : return( 3 );
  case 16 : return( 2 );
  case 17 : return( 1 );
  case 18 : return( 3 );
  case 19 : return( 2 );
  case 20 : return( 1 );
  case 21 : return( 3 );
  case 22 : return( 2 );
  case 23 : return( 1 );
  case 24 : return( 3 );
  case 25 : return( 2 );
  case 26 :
  case 27 : return( 4 );
  case 28 : return( 2 );
  case 29 : return( 4 );
  case 30 : return( 3 );
  case 31 : return( 5 );
  case 32 :
  case 33 : return( 4 );
  case 34 : return( 5 );
  case 35 : return( 4 );
  case 36 : return( 3 );
  case 37 : return( 5 );
  case 38 : return( 4 );
  case 39 : return( 3 );
  case 40 : return( 5 );
  case 41 : return( 4 );
  case 42 : return( 3 );
  case 43 : return( 5 );
  case 44 : return( 4 );
  case 45 : return( 3 );
  case 46 : return( 6 );
  case 47 : return( 5 );
  case 48 : return( 7 );
  case 49 : return( 6 );
  case 50 : return( 5 );
    break;
  }
  return( -1 );
}

static int _iterations_26connectivity( int R )
{
  switch( R ) {
  default : return( -1 );
  case 1 : return( 0 );
  case 2 :
  case 3 : return( 1 );
  case 4 : return( 0 );
  case 5 : return( 2 );
  case 6 : return( 1 );
  case 7 : return( 2 );
  case 8 : return( 3 );
  case 9 : return( 2 );
  case 10 : return( 3 );
  case 11 : return( 4 );
  case 12 : return( 5 );
  case 13 : return( 4 );
  case 14 : return( 5 );
  case 15 : return( 4 );
  case 16 : return( 5 );
  case 17 : return( 6 );
  case 18 : return( 5 );
  case 19 : return( 6 );
  case 20 : return( 7 );
  case 21 : return( 6 );
  case 22 : return( 7 );
  case 23 : return( 8 );
  case 24 : return( 7 );
  case 25 :
  case 26 :
  case 27 : return( 8 );
  case 28 :
  case 29 : return( 9 );
  case 30 : return( 10 );
  case 31 : return( 9 );
  case 32 :
  case 33 :
  case 34 : return( 10 );
  case 35 : return( 11 );
  case 36 : return( 12 );
  case 37 : return( 11 );
  case 38 : return( 12 );
  case 39 : return( 13 );
  case 40 : return( 12 );
  case 41 : return( 13 );
  case 42 : return( 14 );
  case 43 : return( 13 );
  case 44 : return( 14 );
  case 45 : return( 15 );
  case 46 : return( 14 );
  case 47 : return( 15 );
  case 48 : return( 14 );
  case 49 : return( 15 );
  case 50 : return( 16 );
    break;
  }
  return( -1 );
}




void erosion( Inrimage *image, 
	      StructuringElement *se,
	      int iterations,
	      bool binary )
  throw( Exception )
{
  erosion( image, image, se, iterations, binary );
}

void erosion( Inrimage *image, 
	      CONNECTIVITY_TYPE C,
	      int iterations,
	      bool binary )
  throw( Exception )
{
  erosion( image, image, C, iterations, binary );
}

void erosion( Inrimage *imageIn, 
	      Inrimage *imageOut, 
	      CONNECTIVITY_TYPE C,
	      int iterations,
	      bool binary )
  throw( Exception )
{
  StructuringElement *se = new StructuringElement( C );
  erosion( imageIn, imageOut, se, iterations, binary );
  delete se;
}

//
// Erosion with euclidian kernel
//

void erosion(Inrimage *image, int radius, int iterations, 
						 int dimension, bool binary)
	throw( Exception )
{
	const char* from = "erosion(Inrimage*, int, bool)";

	if ( dimension != 2 && dimension != 3)
		throw Exception( from, "Dimension must be 2 or 3" );
	if ( radius <= 0 )
		throw Exception( from, "Negative or null radius" );
	if ( iterations <= 0 )
		throw Exception( from, "Negative or null number of iterations" );
  if ( image == (Inrimage *)NULL )
    throw Exception( from, "Input image is NULL. You must supply a valid input image" );
  if ( image->getData() == NULL)
    throw Exception( from, "Input image buffer is NULL. You must supply a valid input image" );

  try 
	{
		int iter;
		for (int i = 0; i < iterations; i++)
		{
			if ( dimension == 2 )
			{
				COUT << "2D mode\n";
				// 8-connectivity
				iter = _iterations_08connectivity(radius);
				COUT << "8-Connectivity, nb iterations: " << iter << ENDL;
				if ( iter < 0 )
					throw Exception( from, "Radius is too large\n");
				if ( iter != 0 )
					erosion(image, EIGHT_CONNECTIVITY, iter, binary);

				// 4-connectivity first
				iter = _iterations_04connectivity(radius);
				COUT << "4-Connectivity, nb iterations: " << iter << ENDL;
				if ( iter < 0 )
					throw Exception( from, "Radius is too large\n");
				if ( iter != 0 )
					erosion(image, FOUR_CONNECTIVITY, iter, binary);
			}
			else
			{
				COUT << "3D mode\n";
				// 26-connectivity
				iter = _iterations_26connectivity(radius);
				COUT << "26-Connectivity, nb iterations: " << iter << ENDL;
				if ( iter < 0 )
					throw Exception( from, "Radius is too large\n");
				if ( iter != 0 )
					erosion(image, TWENTY_SIX_CONNECTIVITY, iter, binary);

				// 18-connectivity
				iter = _iterations_18connectivity(radius);
				COUT << "18-Connectivity, nb iterations: " << iter << ENDL;
				if ( iter < 0 )
					throw Exception( from, "Radius is too large\n");
				if ( iter != 0 )
					erosion(image, EIGHTEEN_CONNECTIVITY, iter, binary);

				// 6-connectivity
				iter = _iterations_06connectivity(radius);
				COUT << "6-Connectivity, nb iterations: " << iter << ENDL;
				if ( iter < 0 )
					throw Exception( from, "Radius is too large\n");
				if ( iter != 0 )
					erosion(image, SIX_CONNECTIVITY, iter, binary);
			}
		}
  }
  catch(Exception e)
	{
		COUT << "An error has occured in erosion\n";
    e.push(from);

    throw e;
  }
}




void dilation( Inrimage *image, 
	       StructuringElement *se,
	       int iterations,
	       bool binary )
  throw( Exception )
{
  dilation( image, image, se, iterations, binary );
}

void dilation( Inrimage *image, 
	       CONNECTIVITY_TYPE C,
	       int iterations,
	       bool binary )
  throw( Exception )
{
  dilation( image, image, C, iterations, binary );
}

void dilation( Inrimage *imageIn, 
	       Inrimage *imageOut, 
	       CONNECTIVITY_TYPE C,
	       int iterations,
	       bool binary )
  throw( Exception )
{
  StructuringElement *se = new StructuringElement( C );
  dilation( imageIn, imageOut, se, iterations, binary );
  delete se;
}



//
// Dilation with euclidian kernel
//

void dilation(Inrimage *image, int radius, int iterations, 
							int dimension, bool binary)
	throw( Exception )
{
	const char* from = "dilation(Inrimage*, int, bool)";

	if ( dimension != 2 && dimension != 3)
		throw Exception( from, "Dimension must be 2 or 3" );
	if ( radius <= 0 )
		throw Exception( from, "Negative or null radius" );
	if ( iterations <= 0 )
		throw Exception( from, "Negative or null number of iterations" );
  if ( image == (Inrimage *)NULL )
    throw Exception( from, "Input image is NULL. You must supply a valid input image" );
  if ( image->getData() == NULL)
    throw Exception( from, "Input image buffer is NULL. You must supply a valid input image" );

  try 
	{
		int iter;
		for (int i = 0; i < iterations; i++)
		{
			if ( dimension == 2 )
			{
				COUT << "2D mode\n";
				// 8-connectivity
				iter = _iterations_08connectivity(radius);
				COUT << "8-Connectivity, nb iterations: " << iter << ENDL;
				if ( iter < 0 )
					throw Exception( from, "Radius is too large\n");
				if ( iter != 0 )
					dilation(image, EIGHT_CONNECTIVITY, iter, binary);

				// 4-connectivity first
				iter = _iterations_04connectivity(radius);
				COUT << "4-Connectivity, nb iterations: " << iter << ENDL;
				if ( iter < 0 )
					throw Exception( from, "Radius is too large\n");
				if ( iter != 0 )
					dilation(image, FOUR_CONNECTIVITY, iter, binary);
			}
			else
			{
				COUT << "3D mode\n";
				// 26-connectivity
				iter = _iterations_26connectivity(radius);
				COUT << "26-Connectivity, nb iterations: " << iter << ENDL;
				if ( iter < 0 )
					throw Exception( from, "Radius is too large\n");
				if ( iter != 0 )
					dilation(image, TWENTY_SIX_CONNECTIVITY, iter, binary);

				// 18-connectivity
				iter = _iterations_18connectivity(radius);
				COUT << "18-Connectivity, nb iterations: " << iter << ENDL;
				if ( iter < 0 )
					throw Exception( from, "Radius is too large\n");
				if ( iter != 0 )
					dilation(image, EIGHTEEN_CONNECTIVITY, iter, binary);

				// 6-connectivity
				iter = _iterations_06connectivity(radius);
				COUT << "6-Connectivity, nb iterations: " << iter << ENDL;
				if ( iter < 0 )
					throw Exception( from, "Radius is too large\n");
				if ( iter != 0 )
					dilation(image, SIX_CONNECTIVITY, iter, binary);
			}
		}
  }
  catch(Exception e)
	{
		COUT << "An error has occured in dilation\n";
    e.push(from);
    throw e;
  }
}



void opening( Inrimage *image, 
	      StructuringElement *se,
	      int iterations,
	      bool binary )
  throw( Exception )
{
  opening( image, image, se, iterations, binary );
}

void opening( Inrimage *image, 
	      CONNECTIVITY_TYPE C,
	      int iterations,
	      bool binary )
  throw( Exception )
{
  opening( image, image, C, iterations, binary );
}

void opening( Inrimage *imageIn, 
	      Inrimage *imageOut, 
	      CONNECTIVITY_TYPE C,
	      int iterations,
	      bool binary )
  throw( Exception )
{
  StructuringElement *se = new StructuringElement( C );
  opening( imageIn, imageOut, se, iterations, binary );
  delete se;
}




void closing( Inrimage *image, 
	      StructuringElement *se,
	      int iterations,
	      bool binary )
  throw( Exception )
{
  closing( image, image, se, iterations, binary );
}

void closing( Inrimage *image, 
	      CONNECTIVITY_TYPE C,
	      int iterations,
	      bool binary )
  throw( Exception )
{
  closing( image, image, C, iterations, binary );
}

void closing( Inrimage *imageIn, 
	      Inrimage *imageOut, 
	      CONNECTIVITY_TYPE C,
	      int iterations,
	      bool binary )
  throw( Exception )
{
  StructuringElement *se = new StructuringElement( C );
  closing( imageIn, imageOut, se, iterations, binary );
  delete se;
}




void topHat( Inrimage *image, 
	     StructuringElement *se,
	     int iterations,
	     bool binary )
  throw( Exception )
{
  topHat( image, image, se, iterations, binary );
}

void topHat( Inrimage *image, 
	     CONNECTIVITY_TYPE C,
	     int iterations,
	     bool binary )
  throw( Exception )
{
  topHat( image, image, C, iterations, binary );
}

void topHat( Inrimage *imageIn, 
	     Inrimage *imageOut, 
	     CONNECTIVITY_TYPE C,
	     int iterations,
	     bool binary )
  throw( Exception )
{
  StructuringElement *se = new StructuringElement( C );
  topHat( imageIn, imageOut, se, iterations, binary );
  delete se;
}




void invTopHat( Inrimage *image, 
		StructuringElement *se,
		int iterations,
		bool binary )
  throw( Exception )
{
  invTopHat( image, image, se, iterations, binary );
}

void invTopHat( Inrimage *image, 
		CONNECTIVITY_TYPE C,
		int iterations,
		bool binary )
  throw( Exception )
{
  invTopHat( image, image, C, iterations, binary );
}

void invTopHat( Inrimage *imageIn, 
		Inrimage *imageOut, 
		CONNECTIVITY_TYPE C,
		int iterations,
		bool binary )
  throw( Exception )
{
  StructuringElement *se = new StructuringElement( C );
  invTopHat( imageIn, imageOut, se, iterations, binary );
  delete se;
}




void morphologicalGradient( Inrimage *image, 
			    StructuringElement *se,
			    int iterations,
			    bool binary )
  throw( Exception )
{
  morphologicalGradient( image, image, se, iterations, binary );
}

void morphologicalGradient( Inrimage *image, 
			    CONNECTIVITY_TYPE C,
			    int iterations,
			    bool binary )
  throw( Exception )
{
  morphologicalGradient( image, image, C, iterations, binary );
}

void morphologicalGradient( Inrimage *imageIn, 
			    Inrimage *imageOut, 
			    CONNECTIVITY_TYPE C,
			    int iterations,
			    bool binary )
  throw( Exception )
{
  StructuringElement *se = new StructuringElement( C );
  morphologicalGradient( imageIn, imageOut, se, iterations, binary );
  delete se;
}




void morphologicalLowerGradient( Inrimage *image, 
				 StructuringElement *se,
				 int iterations,
				 bool binary )
  throw( Exception )
{
  morphologicalLowerGradient( image, image, se, iterations, binary );
}

void morphologicalLowerGradient( Inrimage *image, 
				 CONNECTIVITY_TYPE C,
				 int iterations,
				 bool binary )
  throw( Exception )
{
  morphologicalLowerGradient( image, image, C, iterations, binary );
}

void morphologicalLowerGradient( Inrimage *imageIn, 
				 Inrimage *imageOut, 
				 CONNECTIVITY_TYPE C,
				 int iterations,
				 bool binary )
  throw( Exception )
{
  StructuringElement *se = new StructuringElement( C );
  morphologicalLowerGradient( imageIn, imageOut, se, iterations, binary );
  delete se;
}




void morphologicalUpperGradient( Inrimage *image, 
				 StructuringElement *se,
				 int iterations,
				 bool binary )
  throw( Exception )
{
  morphologicalUpperGradient( image, image, se, iterations, binary );
}

void morphologicalUpperGradient( Inrimage *image, 
				 CONNECTIVITY_TYPE C,
				 int iterations,
				 bool binary )
  throw( Exception )
{
  morphologicalUpperGradient( image, image, C, iterations, binary );
}

void morphologicalUpperGradient( Inrimage *imageIn, 
				 Inrimage *imageOut, 
				 CONNECTIVITY_TYPE C,
				 int iterations,
				 bool binary )
  throw( Exception )
{
  StructuringElement *se = new StructuringElement( C );
  morphologicalUpperGradient( imageIn, imageOut, se, iterations, binary );
  delete se;
}




void medianFiltering( Inrimage *image, 
		      StructuringElement *se )
  throw( Exception )
{
  medianFiltering( image, image, se );
}

void medianFiltering( Inrimage *image, 
		      CONNECTIVITY_TYPE C )
  throw( Exception )
{
  medianFiltering( image, image, C );
}

void medianFiltering( Inrimage *imageIn, 
		      Inrimage *imageOut, 
		      CONNECTIVITY_TYPE C )
  throw( Exception )
{
  StructuringElement *se = new StructuringElement( C );
  medianFiltering( imageIn, imageOut, se );
  delete se;
}




//
// Opening
//

void opening( Inrimage *inputImage, Inrimage *outputImage, 
	      StructuringElement *se,
	      int iterations,
	      bool binary )
  throw( Exception, StructuringElement::InvalidElement )
{
  char *from = "opening( Inrimage *, Inrimage *, StructuringElement *s, int, bool )";
  
  if (se == (StructuringElement *)NULL)
    throw StructuringElement::InvalidElement( from, "Structuring element is NULL. You must supply a valid Structuring Element");

  try {
    StructuringElement newse = *se;
    erosion( inputImage, outputImage, se, iterations, binary );
    newse.inverse();
    dilation( outputImage, outputImage, &newse, iterations, binary );
  }
  catch( Exception e)  {
    e.push( from );
    throw e;
  }
}



//
// closing
//

void closing( Inrimage *inputImage, Inrimage *outputImage, 
	      StructuringElement *se,
	      int iterations,
	      bool binary )
  throw( Exception, StructuringElement::InvalidElement )
{
  char *from = "closing( Inrimage *, Inrimage *, StructuringElement *s, int, bool )";

  if (se == (StructuringElement *)NULL)
    throw StructuringElement::InvalidElement( from, "Structuring element is NULL. You must supply a valid Structuring Element");

  try {
    //
    // nouvel objet, appel explicite au constructeur par recopie
    //    StructuringElement::StructuringElement( const StructuringElement &elem )
    // StructuringElement *newse = new StructuringElement(*se);
    // ...
    // delete newse;
    //
    // c'est la meme chose, appel implicite au constructeur par recopie
    // StructuringElement newse(*se);
    //g
    // nouvel objet, construction par defaut, 
    // puis utilisation de l'operateur egal
    //   StructuringElement& StructuringElement::operator=(const StructuringElement& s)
    // StructuringElement newse;
    // newse = *se;
    //
    // mouvel objet,  appel implicite au constructeur par recopie
    // StructuringElement newse = *se;
    // 
    StructuringElement newse = *se;

    dilation( inputImage, outputImage, se, iterations, binary );
    newse.inverse();
    erosion( outputImage, outputImage, &newse, iterations, binary );
  }
  
  catch( Exception e)  {
    e.push( from );
    throw e;
  }
}





//
// topHat
// Image - Ouvert
//

void topHat( Inrimage *inputImage, Inrimage *outputImage, 
	     StructuringElement *se,
	     int iterations,
	     bool binary )
  throw( Exception )
{
  char *from = "topHat( Inrimage *, Inrimage *, StructuringElement *s, int, bool )";

  if (inputImage == (Inrimage *)NULL)
    throw Exception( from, "Input image is NULL. You must supply a valid input image");
  if (inputImage->getData() == NULL)
    throw Exception( from, "Input image buffer is NULL. You must supply a valid input image");
  if (outputImage == (Inrimage *)NULL)
    throw Exception( from, "Output image is NULL. You must supply a valid output image");
  if (outputImage->getData() == NULL )
    throw Exception( from, "Output image buffer is NULL. You must supply a valid output image");

  try {
    Inrimage *ouvert = new Inrimage( *inputImage );
    opening( inputImage, ouvert, se, iterations, binary );
    if ( inputImage != outputImage ) {
      outputImage = inputImage;
    }
    sub( outputImage, ouvert );
    delete ouvert;
  }
  catch( Exception e)  {
    e.push( from );
    throw e;
  }
}




//
// InvTopHat
// Ferme - Image
//

void invTopHat( Inrimage *inputImage, Inrimage *outputImage, 
		StructuringElement *se,
		int iterations,
		bool binary )
  throw( Exception, StructuringElement::InvalidElement )
{
  char *from = "invTopHat( Inrimage *, Inrimage *, StructuringElement *s, int, bool )";

  if (inputImage == (Inrimage *)NULL)
    throw Exception( from, "Input image is NULL. You must supply a valid input image");
  if (inputImage->getData() == NULL)
    throw Exception( from, "Input image buffer is NULL. You must supply a valid input image");
  if (outputImage->getData() == NULL )
    throw Exception( from, "Output image buffer is NULL. You must supply a valid output image");
  if (se == (StructuringElement *)NULL)
    throw StructuringElement::InvalidElement( from, "Structuring element is NULL. You must supply a valid Structuring Element");

  try {
    if ( inputImage != outputImage ) {
      closing( inputImage, outputImage, se, iterations, binary );
      sub( outputImage, inputImage );
    }
    else {
      Inrimage *tmp = new Inrimage( *inputImage );
      closing( inputImage, outputImage, se, iterations, binary );
      sub( outputImage, tmp );
      delete tmp;
    }
  }
  catch( Exception e)  {
    e.push( from );
    throw e;
  }
}




//
// Gradient Morphologique
//

void morphologicalGradient( Inrimage *inputImage, Inrimage *outputImage, 
			    StructuringElement *se,
			    int iterations,
			    bool binary )
  throw( Exception )
{
  char *from = "morphologicalGradient( Inrimage *, Inrimage *, StructuringElement *s, int, bool )";
  
  if (inputImage == (Inrimage *)NULL)
    throw Exception( from, "Input image is NULL. You must supply a valid input image");
  if (inputImage->getData() == NULL)
    throw Exception( from, "Input image buffer is NULL. You must supply a valid input image");

  try {
    Inrimage *tmp = new Inrimage( *inputImage );
    dilation( inputImage, outputImage, se, iterations, binary );
    erosion( tmp, tmp, se, iterations, binary );
    sub( outputImage, tmp );
    // div( outputImage, 2.0 );
  }
  catch( Exception e)  {
    e.push( from );
    throw e;
  }
}



//
// Lower Gradient Morphologique
//

void morphologicalLowerGradient( Inrimage *inputImage, Inrimage *outputImage, 
				 StructuringElement *se,
				 int iterations,
				 bool binary )
  throw( Exception )
{
  char *from = "morphologicalLowerGradient( Inrimage *, Inrimage *, StructuringElement *s, int, bool )";

  if (inputImage == (Inrimage *)NULL)
    throw Exception( from, "Input image is NULL. You must supply a valid input image");
  if (inputImage->getData() == NULL)
    throw Exception( from, "Input image buffer is NULL. You must supply a valid input image");

  try {
    Inrimage *tmp = new Inrimage( *inputImage );
    erosion( inputImage, tmp, se, iterations, binary );
    outputImage = inputImage;
    sub( outputImage, tmp );
    // div( outputImage, 2.0 );
  }
  catch( Exception e)  {
    e.push( from );
    throw e;
  }
}



//
// Upper Gradient Morphologique
//

void morphologicalUpperGradient( Inrimage *inputImage, Inrimage *outputImage, 
				 StructuringElement *se,
				 int iterations,
				 bool binary )
  throw( Exception )
{
  char *from = "morphologicalUpperGradient( Inrimage *, Inrimage *, StructuringElement *s, int, bool )";

  if (inputImage == (Inrimage *)NULL)
    throw Exception( from, "Input image is NULL. You must supply a valid input image");
  if (inputImage->getData() == NULL)
    throw Exception( from, "Input image buffer is NULL. You must supply a valid input image");

  try {
    Inrimage *tmp = new Inrimage( *inputImage );
    dilation( inputImage, outputImage, se, iterations, binary );
    sub( outputImage, tmp );
    // div( outputImage, 2.0 );
  }
  catch( Exception e)  {
    e.push( from );
    throw e;
  }
}



//
// Erosion
//

void erosion( Inrimage *inputImage, Inrimage *outputImage, 
	      StructuringElement *se,
	      int iterations,
	      bool binary )
  throw( Exception, ErrMemFull, ErrUnknown, StructuringElement::InvalidElement)
{
  char *from = "erosion( Inrimage *, Inrimage *, StructuringElement *, int, bool )";
  int ret = 0;

  
  if (inputImage == (Inrimage *)NULL)
    throw Exception( from, "Input image is NULL. You must supply a valid input image");
  if (inputImage->getData() == NULL)
    throw Exception( from, "Input image buffer is NULL. You must supply a valid input image");
  if (outputImage == (Inrimage *)NULL)
    throw Exception( from, "Output image is NULL. You must supply a valid output image");
  if (outputImage->getData() == NULL )
    throw Exception( from, "Output image buffer is NULL. You must supply a valid output image");
  if (se == (StructuringElement *)NULL)
    throw StructuringElement::InvalidElement( from, "Structuring element is NULL. You must supply a valid Structuring Element");

  if( inputImage->sameSize( *outputImage ) != true ||
      inputImage->getVdim() != outputImage->getVdim() ||
      inputImage->getType() != outputImage->getType() ||
      inputImage->getVectorialMode() != outputImage->getVectorialMode() ||
      inputImage->getVdim() > 1 ||
      inputImage->getVectorialMode() != VM_SCALAR ) {
    throw Exception( from, "Input images header mismatch" );
  }
  
  
  if ( iterations <= 0 )
    throw Exception( from, "Negative or null number of iterations" );

    
  //
  // User defined structuring element?
  //
  if ( se != NULL &&
       se->getArrayOfPoints() != NULL &&
       se->getNumberOfPoints() > 0 ) {

    switch ( inputImage->getType() ) {
    case Inrimage::WT_UNSIGNED_CHAR :
		ret = _usrdef_valued_Operation<unsigned char, std::less<unsigned char> >( (unsigned char *)0, 
									   (std::less<unsigned char> *)0,
									   (unsigned char*)inputImage->getData(),
									   (unsigned char*)outputImage->getData(),
									   inputImage->getX(),
									   inputImage->getY(),
									   inputImage->getZ(),
									   se->getArrayOfPoints(),
									   se->getNumberOfPoints(),
									   iterations );
      break;
    case Inrimage::WT_SIGNED_CHAR :
      ret = _usrdef_valued_Operation<char, std::less<char> > ( (char *)0, 
							  (std::less<char> *)0,
							  (char*)inputImage->getData(),
							  (char*)outputImage->getData(),
							  inputImage->getX(),
							  inputImage->getY(),
							  inputImage->getZ(),
							  se->getArrayOfPoints(),
							  se->getNumberOfPoints(),
							  iterations );
      break;
    case Inrimage::WT_UNSIGNED_SHORT :
      ret = _usrdef_valued_Operation<unsigned short int, std::less<unsigned short int> >( (unsigned short int *)0, 
										     (std::less<unsigned short int> *)0,
										     (unsigned short int*)inputImage->getData(),
										     (unsigned short int*)outputImage->getData(),
										     inputImage->getX(),
										     inputImage->getY(),
										     inputImage->getZ(),
										     se->getArrayOfPoints(),
										     se->getNumberOfPoints(),
										     iterations );
      break;
    case Inrimage::WT_SIGNED_SHORT :
      ret = _usrdef_valued_Operation<short int, std::less<short int> >( (short int *)0, 
								   (std::less<short int> *)0,
								   (short int*)inputImage->getData(),
								   (short int*)outputImage->getData(),
								   inputImage->getX(),
								   inputImage->getY(),
								   inputImage->getZ(),
								   se->getArrayOfPoints(),
								   se->getNumberOfPoints(),
								   iterations );
      break;
    case Inrimage::WT_UNSIGNED_INT :
      ret = _usrdef_valued_Operation<unsigned int, std::less<unsigned int> >( (unsigned int *)0, 
								   (std::less<unsigned int> *)0,
								   (unsigned  int*)inputImage->getData(),
								   (unsigned int*)outputImage->getData(),
								   inputImage->getX(),
								   inputImage->getY(),
								   inputImage->getZ(),
								   se->getArrayOfPoints(),
								   se->getNumberOfPoints(),
								   iterations );
      break;
    case Inrimage::WT_SIGNED_INT :
      ret = _usrdef_valued_Operation<int, std::less<int> >( (int *)0, 
						       (std::less<int> *)0,
						       (int*)inputImage->getData(),
						       (int*)outputImage->getData(),
						       inputImage->getX(),
						       inputImage->getY(),
						       inputImage->getZ(),
						       se->getArrayOfPoints(),
						       se->getNumberOfPoints(),
						       iterations );
      break;
    case Inrimage::WT_UNSIGNED_LONG :
      ret = _usrdef_valued_Operation<unsigned long, std::less<unsigned long> >( (unsigned long *)0, 
						       (std::less<unsigned long> *)0,
						       (unsigned long*)inputImage->getData(),
						       (unsigned long*)outputImage->getData(),
						       inputImage->getX(),
						       inputImage->getY(),
						       inputImage->getZ(),
						       se->getArrayOfPoints(),
						       se->getNumberOfPoints(),
						       iterations );
      break;
    case Inrimage::WT_SIGNED_LONG :
      ret = _usrdef_valued_Operation<long, std::less<long> >( (long *)0, 
						       (std::less<long> *)0,
						       (long*)inputImage->getData(),
						       (long*)outputImage->getData(),
						       inputImage->getX(),
						       inputImage->getY(),
						       inputImage->getZ(),
						       se->getArrayOfPoints(),
						       se->getNumberOfPoints(),
						       iterations );
      break;
    case Inrimage::WT_FLOAT :
      ret = _usrdef_valued_Operation<float, std::less<float> >( (float* )0,
							   (std::less<float> *)0,
							   (float*)inputImage->getData(),
							   (float*)outputImage->getData(),
							   inputImage->getX(),
							   inputImage->getY(),
							   inputImage->getZ(),
							   se->getArrayOfPoints(),
							   se->getNumberOfPoints(),
							   iterations );
      break;
    case Inrimage::WT_DOUBLE :
      ret = _usrdef_valued_Operation<double, std::less<double> >( (double* )0,
							   (std::less<double> *)0,
							   (double*)inputImage->getData(),
							   (double*)outputImage->getData(),
							   inputImage->getX(),
							   inputImage->getY(),
							   inputImage->getZ(),
							   se->getArrayOfPoints(),
							   se->getNumberOfPoints(),
							   iterations );
      break;
    case Inrimage::WT_RGB :
    case Inrimage::WT_RGBA :
    case Inrimage::WT_FLOAT_VECTOR :
    case Inrimage::WT_DOUBLE_VECTOR :
    default :
      throw Exception( from, "Image type not handled yet" );
      break;
    }
    
    switch( ret ) {
    case 1 : 
      return;
    case _morpho_ErrorInParameters_ :
      throw Exception( from, "Error in parameters during computation" );
    case _morpho_ErrMemFull_ :
      throw ErrMemFull( from );
    default :
      throw ErrUnknown( from );
    }
  }
  
  
  //
  // Pre-defined connectivity
  //
  
  int c = 26;
  if ( se != NULL ) {
    switch( se->getConnectivity() ) {
    case FOUR_CONNECTIVITY :       c =  4; break;
    case EIGHT_CONNECTIVITY :      c =  8; break;
    case SIX_CONNECTIVITY :        c =  6; break;
    case TEN_CONNECTIVITY :        c = 10; break;
    case EIGHTEEN_CONNECTIVITY :   c = 18; break;
    case TWENTY_SIX_CONNECTIVITY : c = 26; break;      
    }
  }


  switch ( inputImage->getType() ) {
  case Inrimage::WT_UNSIGNED_CHAR :
    if ( binary == true ) {
      ret = _predef_binary_Operation<unsigned char, std::less<unsigned char>, std::logical_and<unsigned char> >( (unsigned char *)0, 
												       (std::less<unsigned char> *)0,
												       (std::logical_and<unsigned char> *)0,
												       (unsigned char*)inputImage->getData(),
												       (unsigned char*)outputImage->getData(),
												       inputImage->getX(),
												       inputImage->getY(),
												       inputImage->getZ(),
												       c, iterations );
    }
    else {
      ret = _predef_valued_Operation<unsigned char, std::less<unsigned char> >( (unsigned char *)0, 
									   (std::less<unsigned char> *)0,
									   (unsigned char*)inputImage->getData(),
									   (unsigned char*)outputImage->getData(),
									   inputImage->getX(),
									   inputImage->getY(),
									   inputImage->getZ(),
									   c, iterations );
    }
    break;
  case Inrimage::WT_SIGNED_CHAR :
    if ( binary == true ) {
      ret = _predef_binary_Operation<char, std::less<char>, std::logical_and<char> >( (char *)0, 
									    (std::less<char> *)0,
									    (std::logical_and<char> *)0,		    
									    (char*)inputImage->getData(),
									    (char*)outputImage->getData(),
									    inputImage->getX(),
									    inputImage->getY(),
									    inputImage->getZ(),
									    c, iterations );
    }
    else {
      ret = _predef_valued_Operation<char, std::less<char> >( (char *)0, 
							 (std::less<char> *)0,
							 (char*)inputImage->getData(),
							 (char*)outputImage->getData(),
							 inputImage->getX(),
							 inputImage->getY(),
							 inputImage->getZ(),
							 c, iterations );
    }
    break;
  case Inrimage::WT_UNSIGNED_SHORT :
    if ( binary == true ) {
      ret = _predef_binary_Operation<unsigned short int, std::less<unsigned short int>, std::logical_and<unsigned short int> >( (unsigned short int *)0, 
														      (std::less<unsigned short int> *)0,
														      (std::logical_and<unsigned short int> *)0,			      
														      (unsigned short int*)inputImage->getData(),
														      (unsigned short int*)outputImage->getData(),
														      inputImage->getX(),
														      inputImage->getY(),
														      inputImage->getZ(),
														      c, iterations );
    }
    else {
      ret = _predef_valued_Operation<unsigned short int, std::less<unsigned short int> >( (unsigned short int *)0, 
										     (std::less<unsigned short int> *)0,
										     (unsigned short int*)inputImage->getData(),
										     (unsigned short int*)outputImage->getData(),
										     inputImage->getX(),
										     inputImage->getY(),
										     inputImage->getZ(),
										     c, iterations );
    }
    break;
  case Inrimage::WT_SIGNED_SHORT :
    if ( binary == true ) {
      ret = _predef_binary_Operation<short int, std::less<short int>, std::logical_and<short int> >( (short int *)0, 
											   (std::less<short int> *)0,
											   (std::logical_and<short int> *)0,
											   (short int*)inputImage->getData(),
											   (short int*)outputImage->getData(),
											   inputImage->getX(),
											   inputImage->getY(),
											   inputImage->getZ(),
											   c, iterations );
    }
    else {
      ret = _predef_valued_Operation<short int, std::less<short int> >( (short int *)0, 
								   (std::less<short int> *)0,
								   (short int*)inputImage->getData(),
								   (short int*)outputImage->getData(),
								   inputImage->getX(),
								   inputImage->getY(),
								   inputImage->getZ(),
								   c, iterations );
    }
    break;
  case Inrimage::WT_UNSIGNED_INT :
    if ( binary == true ) {
      ret = _predef_binary_Operation<unsigned int, std::less<unsigned int>, std::logical_and<unsigned int> >( (unsigned int *)0, 
											   (std::less<unsigned int> *)0,
											   (std::logical_and<unsigned int> *)0,
											   (unsigned int*)inputImage->getData(),
											   (unsigned int*)outputImage->getData(),
											   inputImage->getX(),
											   inputImage->getY(),
											   inputImage->getZ(),
											   c, iterations );
    }
    else {
      ret = _predef_valued_Operation<unsigned int, std::less<unsigned int> >( (unsigned int *)0, 
								   (std::less<unsigned int> *)0,
								   (unsigned int*)inputImage->getData(),
								   (unsigned int*)outputImage->getData(),
								   inputImage->getX(),
								   inputImage->getY(),
								   inputImage->getZ(),
								   c, iterations );
    }
    break;
  case Inrimage::WT_SIGNED_INT :
    if ( binary == true ) {
      ret = _predef_binary_Operation<int, std::less<int>, std::logical_and<int> >( (int *)0, 
									 (std::less<int> *)0,
									 (std::logical_and<int> *)0,
									 (int*)inputImage->getData(),
									 (int*)outputImage->getData(),
									 inputImage->getX(),
									 inputImage->getY(),
									 inputImage->getZ(),
									 c, iterations );
    }
    else {
      ret = _predef_valued_Operation<int, std::less<int> >( (int *)0, 
						       (std::less<int> *)0,
						       (int*)inputImage->getData(),
						       (int*)outputImage->getData(),
						       inputImage->getX(),
						       inputImage->getY(),
						       inputImage->getZ(),
						       c, iterations );
    }
    break;
  case Inrimage::WT_UNSIGNED_LONG :
    if ( binary == true ) {
      ret = _predef_binary_Operation<unsigned long, std::less<unsigned long>, std::logical_and<unsigned long> >( (unsigned long *)0, 
												       (std::less<unsigned long> *)0,
												       (std::logical_and<unsigned long> *)0,
												       (unsigned long*)inputImage->getData(),
												       (unsigned long*)outputImage->getData(),
												       inputImage->getX(),
												       inputImage->getY(),
												       inputImage->getZ(),
												       c, iterations );
    }
    else {
      ret = _predef_valued_Operation<unsigned long, std::less<unsigned long> >( (unsigned long *)0, 
									   (std::less<unsigned long> *)0,
									   (unsigned long*)inputImage->getData(),
									   (unsigned long*)outputImage->getData(),
									   inputImage->getX(),
									   inputImage->getY(),
									   inputImage->getZ(),
									   c, iterations );
    }
    break;
  case Inrimage::WT_SIGNED_LONG :
    if ( binary == true ) {
      ret = _predef_binary_Operation<long, std::less<long>, std::logical_and<long> >( (long *)0, 
									    (std::less<long> *)0,
									    (std::logical_and<long> *)0,
									    (long*)inputImage->getData(),
									    (long*)outputImage->getData(),
									    inputImage->getX(),
									    inputImage->getY(),
									    inputImage->getZ(),
									    c, iterations );
    }
    else {
      ret = _predef_valued_Operation<long, std::less<long> >( (long *)0, 
							 (std::less<long> *)0,
							 (long*)inputImage->getData(),
							 (long*)outputImage->getData(),
							 inputImage->getX(),
							 inputImage->getY(),
							 inputImage->getZ(),
							 c, iterations );
    }
    break;
  case Inrimage::WT_FLOAT :
    ret = _predef_valued_Operation<float, std::less<float> >( (float* )0,
							 (std::less<float> *)0,
							 (float*)inputImage->getData(),
							 (float*)outputImage->getData(),
							 inputImage->getX(),
							 inputImage->getY(),
							 inputImage->getZ(),
							 c, iterations );
    break;
  case Inrimage::WT_DOUBLE :
    ret = _predef_valued_Operation<double, std::less<double> >( (double* )0,
							 (std::less<double> *)0,
							 (double*)inputImage->getData(),
							 (double*)outputImage->getData(),
							 inputImage->getX(),
							 inputImage->getY(),
							 inputImage->getZ(),
							 c, iterations );
    break;
  case Inrimage::WT_RGB :
  case Inrimage::WT_RGBA :
  case Inrimage::WT_FLOAT_VECTOR :
  case Inrimage::WT_DOUBLE_VECTOR :
  default :
    throw Exception( from, "Image type not handled yet" );
    break;
  }

  switch( ret ) {
  case 1 : 
    return;
  case _morpho_ErrorInParameters_ :
    throw Exception( from, "Error in parameters during computation" );
  case _morpho_ErrMemFull_ :
    throw ErrMemFull( from );
  default :
    throw ErrUnknown( from );
  }
  
  return;

}

//
// Dilation
//

void dilation( Inrimage *inputImage, Inrimage *outputImage, 
	       StructuringElement *se,
	       int iterations,
	       bool binary )
  throw( Exception, ErrMemFull, ErrUnknown, StructuringElement::InvalidElement )
{
  char *from = "Dilation( Inrimage *, Inrimage *, StructuringElement *s, int, bool )";
  int ret = 0;
  
  if (inputImage == (Inrimage *)NULL)
    throw Exception( from, "Input image is NULL. You must supply a valid input image");
  if (inputImage->getData() == NULL)
    throw Exception( from, "Input image buffer is NULL. You must supply a valid input image");
  if (outputImage == (Inrimage *)NULL)
    throw Exception( from, "Output image is NULL. You must supply a valid output image");
  if (outputImage->getData() == NULL )
    throw Exception( from, "Output image buffer is NULL. You must supply a valid output image");
  if (se == (StructuringElement *)NULL)
    throw StructuringElement::InvalidElement( from, "Structuring element is NULL. You must supply a valid Structuring Element");
  
  if ( inputImage->sameSize( *outputImage ) != true ||
      inputImage->getVdim() != outputImage->getVdim() ||
      inputImage->getType() != outputImage->getType() ||
      inputImage->getVectorialMode() != outputImage->getVectorialMode() ||
      inputImage->getVdim() > 1 ||
      inputImage->getVectorialMode() != VM_SCALAR ) {
    throw Exception( from, "Input images header mismatch" );
  }
  
  
  if ( iterations <= 0 )
    throw Exception( from, "Negative or null number of iterations" );

    
  //
  // User defined structuring element?
  //
  if ( se != NULL &&
       se->getArrayOfPoints() != NULL &&
       se->getNumberOfPoints() > 0 ) {

    switch ( inputImage->getType() ) {
    case Inrimage::WT_UNSIGNED_CHAR :
      ret = _usrdef_valued_Operation<unsigned char, std::greater<unsigned char> > ( (unsigned char *)0, 
									       (std::greater<unsigned char> *)0,
									       (unsigned char*)inputImage->getData(),
									       (unsigned char*)outputImage->getData(),
									       inputImage->getX(),
									       inputImage->getY(),
									       inputImage->getZ(),
									       se->getArrayOfPoints(),
									       se->getNumberOfPoints(),
									       iterations );
      break;
    case Inrimage::WT_SIGNED_CHAR :
      ret = _usrdef_valued_Operation<char, std::greater<char> >( (char *)0, 
							    (std::greater<char> *)0,
							    (char*)inputImage->getData(),
							    (char*)outputImage->getData(),
							    inputImage->getX(),
							    inputImage->getY(),
							    inputImage->getZ(),
							    se->getArrayOfPoints(),
							    se->getNumberOfPoints(),
							    iterations );
      break;
    case Inrimage::WT_UNSIGNED_SHORT :
      ret = _usrdef_valued_Operation<unsigned short int, std::greater<unsigned short int> >( (unsigned short int *)0, 
											(std::greater<unsigned short int> *)0,
											(unsigned short int*)inputImage->getData(),
											(unsigned short int*)outputImage->getData(),
											inputImage->getX(),
											inputImage->getY(),
											inputImage->getZ(),
											se->getArrayOfPoints(),
											se->getNumberOfPoints(),
											iterations );
      break;
    case Inrimage::WT_SIGNED_SHORT :
      ret = _usrdef_valued_Operation<short int, std::greater<short int> >( (short int *)0, 
								      (std::greater<short int> *)0,
								      (short int*)inputImage->getData(),
								      (short int*)outputImage->getData(),
								      inputImage->getX(),
								      inputImage->getY(),
								      inputImage->getZ(),
								      se->getArrayOfPoints(),
								      se->getNumberOfPoints(),
								      iterations );
      break;
    case Inrimage::WT_UNSIGNED_INT :
      ret = _usrdef_valued_Operation<unsigned int, std::greater<unsigned int> >( (unsigned int *)0, 
								      (std::greater<unsigned int> *)0,
								      (unsigned int*)inputImage->getData(),
								      (unsigned int*)outputImage->getData(),
								      inputImage->getX(),
								      inputImage->getY(),
								      inputImage->getZ(),
								      se->getArrayOfPoints(),
								      se->getNumberOfPoints(),
								      iterations );
      break;
    case Inrimage::WT_SIGNED_INT :
      ret = _usrdef_valued_Operation<int, std::greater<int> >( (int *)0, 
							  (std::greater<int> *)0,
							  (int*)inputImage->getData(),
							  (int*)outputImage->getData(),
							  inputImage->getX(),
							  inputImage->getY(),
							  inputImage->getZ(),
							  se->getArrayOfPoints(),
							  se->getNumberOfPoints(),
							  iterations );
      break;
    case Inrimage::WT_UNSIGNED_LONG :
      ret = _usrdef_valued_Operation<unsigned long, std::greater<unsigned long> >( (unsigned long *)0, 
									      (std::greater<unsigned long> *)0,
									      (unsigned long*)inputImage->getData(),
									      (unsigned long*)outputImage->getData(),
									      inputImage->getX(),
									      inputImage->getY(),
									      inputImage->getZ(),
									      se->getArrayOfPoints(),
									      se->getNumberOfPoints(),
									      iterations );
      break;
    case Inrimage::WT_SIGNED_LONG :
      ret = _usrdef_valued_Operation<long, std::greater<long> >( (long *)0, 
							    (std::greater<long> *)0,
							    (long*)inputImage->getData(),
							    (long*)outputImage->getData(),
							    inputImage->getX(),
							    inputImage->getY(),
							    inputImage->getZ(),
							    se->getArrayOfPoints(),
							    se->getNumberOfPoints(),
							    iterations );
      break;
    case Inrimage::WT_FLOAT :
      ret = _usrdef_valued_Operation<float, std::greater<float> >( (float* )0,
							      (std::greater<float> *)0,
							      (float*)inputImage->getData(),
							      (float*)outputImage->getData(),
							      inputImage->getX(),
							      inputImage->getY(),
							      inputImage->getZ(),
							      se->getArrayOfPoints(),
							      se->getNumberOfPoints(),
							      iterations );
      break;
    case Inrimage::WT_DOUBLE :
      ret = _usrdef_valued_Operation<double, std::greater<double> >( (double* )0,
								(std::greater<double> *)0,
								(double*)inputImage->getData(),
								(double*)outputImage->getData(),
								inputImage->getX(),
								inputImage->getY(),
								inputImage->getZ(),
								se->getArrayOfPoints(),
								se->getNumberOfPoints(),
								iterations );
      break;
    case Inrimage::WT_RGB :
    case Inrimage::WT_RGBA :
    case Inrimage::WT_FLOAT_VECTOR :
    case Inrimage::WT_DOUBLE_VECTOR :
    default :
      throw Exception( from, "Image type not handled yet" );
      break;
    }
    
    switch( ret ) {
    case 1 : 
      return;
    case _morpho_ErrorInParameters_ :
      throw Exception( from, "Error in parameters during computation" );
    case _morpho_ErrMemFull_ :
      throw ErrMemFull( from );
    default :
      throw ErrUnknown( from );
    }
  }
  

  //
  // Pre-defined connectivity
  //
  
  int c = 26;
  if ( se != NULL ) c = se->getConnectivity();
  


  switch ( inputImage->getType() ) {
  case Inrimage::WT_UNSIGNED_CHAR :
    if ( binary == true ) {
      ret = _predef_binary_Operation<unsigned char, std::greater<unsigned char>, std::logical_or<unsigned char> >( (unsigned char *)0, 
													 (std::greater<unsigned char> *)0,
													 (std::logical_or<unsigned char> *)0,
													 (unsigned char*)inputImage->getData(),
													 (unsigned char*)outputImage->getData(),
													 inputImage->getX(),
													 inputImage->getY(),
													 inputImage->getZ(),
													 c, iterations );
    }
    else {
      ret = _predef_valued_Operation<unsigned char, std::greater<unsigned char> >( (unsigned char *)0, 
									      (std::greater<unsigned char> *)0,
									      (unsigned char*)inputImage->getData(),
									      (unsigned char*)outputImage->getData(),
									      inputImage->getX(),
									      inputImage->getY(),
									      inputImage->getZ(),
									      c, iterations );
    }
    break;
  case Inrimage::WT_SIGNED_CHAR :
    if ( binary == true ) {
      ret = _predef_binary_Operation<char, std::greater<char>, std::logical_or<char> >( (char *)0, 
									      (std::greater<char> *)0,
									      (std::logical_or<char> *)0,
									      (char*)inputImage->getData(),
									      (char*)outputImage->getData(),
									      inputImage->getX(),
									      inputImage->getY(),
									      inputImage->getZ(),
									      c, iterations );
    }
    else {
      ret = _predef_valued_Operation<char, std::greater<char> >( (char *)0, 
							    (std::greater<char> *)0,
							    (char*)inputImage->getData(),
							    (char*)outputImage->getData(),
							    inputImage->getX(),
							    inputImage->getY(),
							    inputImage->getZ(),
							    c, iterations );
    }
    break;
  case Inrimage::WT_UNSIGNED_SHORT :
    if ( binary == true ) {
      ret = _predef_binary_Operation<unsigned short int, std::greater<unsigned short int>, std::logical_or<unsigned short int> >( (unsigned short int *)0, 
															(std::greater<unsigned short int> *)0,
															(std::logical_or<unsigned short int> *)0,
															(unsigned short int*)inputImage->getData(),
															(unsigned short int*)outputImage->getData(),
															inputImage->getX(),
															inputImage->getY(),
															inputImage->getZ(),
															c, iterations );
    }
    else {
      ret = _predef_valued_Operation<unsigned short int, std::greater<unsigned short int> >( (unsigned short int *)0, 
											(std::greater<unsigned short int> *)0,
											(unsigned short int*)inputImage->getData(),
											(unsigned short int*)outputImage->getData(),
											inputImage->getX(),
											inputImage->getY(),
											inputImage->getZ(),
											c, iterations );
    }
    break;
  case Inrimage::WT_SIGNED_SHORT :
    if ( binary == true ) {
      ret = _predef_binary_Operation<short int, std::greater<short int>, std::logical_or<short int> >( (short int *)0, 
											     (std::greater<short int> *)0,
											     (std::logical_or<short int> *)0,
											     (short int*)inputImage->getData(),
											     (short int*)outputImage->getData(),
											     inputImage->getX(),
											     inputImage->getY(),
											     inputImage->getZ(),
											     c, iterations );
    }
    else {
      ret = _predef_valued_Operation<short int, std::greater<short int> >( (short int *)0, 
								      (std::greater<short int> *)0,
								      (short int*)inputImage->getData(),
								      (short int*)outputImage->getData(),
								      inputImage->getX(),
								      inputImage->getY(),
								      inputImage->getZ(),
								      c, iterations );
    }
    break;
  case Inrimage::WT_UNSIGNED_INT :
    if ( binary == true ) {
      ret = _predef_binary_Operation<unsigned int, std::greater<unsigned int>, std::logical_or<unsigned int> >( (unsigned int *)0, 
												      (std::greater<unsigned int> *)0,
												      (std::logical_or<unsigned int> *)0,
												      (unsigned int*)inputImage->getData(),
												      (unsigned int*)outputImage->getData(),
												      inputImage->getX(),
												      inputImage->getY(),
												      inputImage->getZ(),
												      c, iterations );
    }
    else {
      ret = _predef_valued_Operation<unsigned int, std::greater<unsigned int> >( (unsigned int *)0, 
									    (std::greater<unsigned int> *)0,
									    (unsigned int*)inputImage->getData(),
									    (unsigned int*)outputImage->getData(),
									    inputImage->getX(),
									    inputImage->getY(),
									    inputImage->getZ(),
									    c, iterations );
    }
    break;
  case Inrimage::WT_SIGNED_INT :
    if ( binary == true ) {
      ret = _predef_binary_Operation<int, std::greater<int>, std::logical_or<int> >( (int *)0, 
									   (std::greater<int> *)0,
									   (std::logical_or<int> *)0,
									   (int*)inputImage->getData(),
									   (int*)outputImage->getData(),
									   inputImage->getX(),
									   inputImage->getY(),
									   inputImage->getZ(),
									   c, iterations );
    }
    else {
      ret = _predef_valued_Operation<int, std::greater<int> >( (int *)0, 
							  (std::greater<int> *)0,
							  (int*)inputImage->getData(),
							  (int*)outputImage->getData(),
							  inputImage->getX(),
							  inputImage->getY(),
							  inputImage->getZ(),
							  c, iterations );
    }
    break;
  case Inrimage::WT_UNSIGNED_LONG :
    if ( binary == true ) {
      ret = _predef_binary_Operation<unsigned long, std::greater<unsigned long>, std::logical_or<unsigned long> >( (unsigned long *)0, 
													 (std::greater<unsigned long> *)0,
													 (std::logical_or<unsigned long> *)0,
													 (unsigned long*)inputImage->getData(),
													 (unsigned long*)outputImage->getData(),
													 inputImage->getX(),
													 inputImage->getY(),
													 inputImage->getZ(),
													 c, iterations );
    }
    else {
      ret = _predef_valued_Operation<unsigned long, std::greater<unsigned long> >( (unsigned long *)0, 
									      (std::greater<unsigned long> *)0,
									      (unsigned long*)inputImage->getData(),
									      (unsigned long*)outputImage->getData(),
									      inputImage->getX(),
									      inputImage->getY(),
									      inputImage->getZ(),
									      c, iterations );
    }
    break;
  case Inrimage::WT_SIGNED_LONG :
    if ( binary == true ) {
      ret = _predef_binary_Operation<long, std::greater<long>, std::logical_or<long> >( (long *)0, 
									      (std::greater<long> *)0,
									      (std::logical_or<long> *)0,
									      (long*)inputImage->getData(),
									      (long*)outputImage->getData(),
									      inputImage->getX(),
									      inputImage->getY(),
									      inputImage->getZ(),
									      c, iterations );
    }
    else {
      ret = _predef_valued_Operation<long, std::greater<long> >( (long *)0, 
							    (std::greater<long> *)0,
							    (long*)inputImage->getData(),
							    (long*)outputImage->getData(),
							    inputImage->getX(),
							    inputImage->getY(),
							    inputImage->getZ(),
							    c, iterations );
    }
    break;
  case Inrimage::WT_FLOAT :
    ret = _predef_valued_Operation<float, std::greater<float> >( (float* )0,
							    (std::greater<float> *)0,	    
							    (float*)inputImage->getData(),
							    (float*)outputImage->getData(),
							    inputImage->getX(),
							    inputImage->getY(),
							    inputImage->getZ(),
							    c, iterations );
    break;
  case Inrimage::WT_DOUBLE :
    ret = _predef_valued_Operation<double, std::greater<double> >( (double* )0,
							      (std::greater<double> *)0,	    
							      (double*)inputImage->getData(),
							      (double*)outputImage->getData(),
							      inputImage->getX(),
							      inputImage->getY(),
							      inputImage->getZ(),
							      c, iterations );
    break;
  case Inrimage::WT_RGB :
  case Inrimage::WT_RGBA :
  case Inrimage::WT_FLOAT_VECTOR :
  case Inrimage::WT_DOUBLE_VECTOR :
  default :
    throw Exception( from, "Image type not handled yet" );
    break;
  }

  switch( ret ) {
  case 1 : 
    return;
  case _morpho_ErrorInParameters_ :
    throw Exception( from, "Error in parameters during computation" );
  case _morpho_ErrMemFull_ :
    throw ErrMemFull( from );
  default :
    throw ErrUnknown( from );
  }
  
  return;

}







//
//
//
void medianFiltering( Inrimage *inputImage, Inrimage *outputImage, 
		      StructuringElement *se )
  throw( Exception, ErrMemFull, ErrUnknown, StructuringElement::InvalidElement)
{
  char *from = "medianFiltering( Inrimage *, Inrimage *, StructuringElement * )";
  int ret;

  if ( inputImage == (Inrimage *)NULL)
    throw Exception( from, "Input image is NULL. You must supply a valid input image");
  if ( outputImage == (Inrimage *)NULL)
    throw Exception( from, "Output image is NULL. You must supply a valid output image");
  if ( se == (StructuringElement *)NULL)
    throw StructuringElement::InvalidElement(from, "Structuring Element is NULL. You must supply a valid Structuring Element");
  
  if( inputImage->sameSize( *outputImage ) != true ||
      inputImage->getVdim() != outputImage->getVdim() ||
      inputImage->getType() != outputImage->getType() ||
      inputImage->getVectorialMode() != outputImage->getVectorialMode() ||
      inputImage->getVdim() > 1 ||
      inputImage->getVectorialMode() != VM_SCALAR ) {
    throw Exception( from, "Input images header mismatch" );
  }

  StructuringElement newse = *se;

  newse.convertToList();

  switch ( inputImage->getType() ) {
  case Inrimage::WT_UNSIGNED_CHAR :
    ret = _medianFiltering( (unsigned char *)0, 
			    (unsigned char*)inputImage->getData(),
			    (unsigned char*)outputImage->getData(),
			    inputImage->getX(),
			    inputImage->getY(),
			    inputImage->getZ(),
			    newse.getArrayOfPoints(),
			    newse.getNumberOfPoints() );
    break;
  case Inrimage::WT_SIGNED_CHAR :
    ret = _medianFiltering( (char *)0, 
			    (char*)inputImage->getData(),
			    (char*)outputImage->getData(),
			    inputImage->getX(),
			    inputImage->getY(),
			    inputImage->getZ(),
			    newse.getArrayOfPoints(),
			    newse.getNumberOfPoints() );
    break;
  case Inrimage::WT_UNSIGNED_SHORT :
    ret = _medianFiltering( (unsigned short int *)0, 
			    (unsigned short int*)inputImage->getData(),
			    (unsigned short int*)outputImage->getData(),
			    inputImage->getX(),
			    inputImage->getY(),
			    inputImage->getZ(),
			    newse.getArrayOfPoints(),
			    newse.getNumberOfPoints() );
    break;
  case Inrimage::WT_SIGNED_SHORT :
    ret = _medianFiltering( (short int *)0, 
			    (short int*)inputImage->getData(),
			    (short int*)outputImage->getData(),
			    inputImage->getX(),
			    inputImage->getY(),
			    inputImage->getZ(),
			    newse.getArrayOfPoints(),
			    newse.getNumberOfPoints() );
    break;
  case Inrimage::WT_UNSIGNED_INT :
    ret = _medianFiltering( (unsigned int *)0, 
			    (unsigned int*)inputImage->getData(),
			    (unsigned int*)outputImage->getData(),
			    inputImage->getX(),
			    inputImage->getY(),
			    inputImage->getZ(),
			    newse.getArrayOfPoints(),
			    newse.getNumberOfPoints() );
    break;
  case Inrimage::WT_SIGNED_INT :
    ret = _medianFiltering( (int *)0, 
			    (int*)inputImage->getData(),
			    (int*)outputImage->getData(),
			    inputImage->getX(),
			    inputImage->getY(),
			    inputImage->getZ(),
			    newse.getArrayOfPoints(),
			    newse.getNumberOfPoints() );
    break;
  case Inrimage::WT_UNSIGNED_LONG :
    ret = _medianFiltering( (unsigned long *)0, 
			    (unsigned long*)inputImage->getData(),
			    (unsigned long*)outputImage->getData(),
			    inputImage->getX(),
			    inputImage->getY(),
			    inputImage->getZ(),
			    newse.getArrayOfPoints(),
			    newse.getNumberOfPoints() );
    break;
  case Inrimage::WT_SIGNED_LONG :
    ret = _medianFiltering( ( long *)0, 
			    ( long *)inputImage->getData(),
			    ( long *)outputImage->getData(),
			    inputImage->getX(),
			    inputImage->getY(),
			    inputImage->getZ(),
			    newse.getArrayOfPoints(),
			    newse.getNumberOfPoints() );
    break;
  case Inrimage::WT_FLOAT :
    ret = _medianFiltering( (float* )0,
			    (float*)inputImage->getData(),
			    (float*)outputImage->getData(),
			    inputImage->getX(),
			    inputImage->getY(),
			    inputImage->getZ(),
			    newse.getArrayOfPoints(),
			    newse.getNumberOfPoints() );
    break;
  case Inrimage::WT_DOUBLE :
    ret = _medianFiltering( (double* )0,
			    (double*)inputImage->getData(),
			    (double*)outputImage->getData(),
			    inputImage->getX(),
			    inputImage->getY(),
			    inputImage->getZ(),
			    newse.getArrayOfPoints(),
			    newse.getNumberOfPoints() );
    break;
  case Inrimage::WT_RGB :
  case Inrimage::WT_RGBA :
  case Inrimage::WT_FLOAT_VECTOR :
  case Inrimage::WT_DOUBLE_VECTOR :
  default :
    throw Exception( from, "Image type not handled yet" );
    break;
  }
  
  
  switch( ret ) {
  case 1 : 
    return;
  case _morpho_ErrorInParameters_ :
    throw Exception( from, "Error in parameters during computation");
  case _morpho_ErrMemFull_ :
    throw ErrMemFull( from );
  default :
    throw ErrUnknown( from );
  }
}




//
//
//
void watershed( Inrimage *gradientImage,
		Inrimage *inputLabelsImage,
		Inrimage *outputLabelsImage )
  throw( Exception, ErrMemFull, ErrUnknown )
{
  char *from = "watershed( Inrimage *, Inrimage *, Inrimage * )";
  int ret;


  if ( inputLabelsImage->getType() != outputLabelsImage->getType() ) {
    throw Exception( from, "Labels image have different types" );
  }
  
  //
  // C'est une implementation par tableaux
  // le type de l'image 'norme de gradient' doit etre
  // entier non signe
  //
  // id pour les images de labels
  //
  switch ( gradientImage->getType() ) {
  default :
    throw Exception( from, "Gradient image type not handled yet" );
    break;
  case Inrimage::WT_UNSIGNED_CHAR :

    switch ( inputLabelsImage->getType() ) {
    default :
      throw Exception( from, "Labels image type not handled yet" );
      break;
    case Inrimage::WT_UNSIGNED_CHAR :
      ret = _watershed( (unsigned char *)0,
			(unsigned char *)0,
			(unsigned char *)gradientImage->getData(),
			(unsigned char *)inputLabelsImage->getData(),
			(unsigned char *)outputLabelsImage->getData(),
			gradientImage->getX(),
			gradientImage->getY(),
			gradientImage->getZ(),
			gradientImage->getWdim() );
      break;
    case Inrimage::WT_UNSIGNED_SHORT :
      ret = _watershed( (unsigned char *)0,
			(unsigned short int *)0,
			(unsigned char *)gradientImage->getData(),
			(unsigned short int *)inputLabelsImage->getData(),
			(unsigned short int *)outputLabelsImage->getData(),
			gradientImage->getX(),
			gradientImage->getY(),
			gradientImage->getZ(),
			gradientImage->getWdim() );
      break;
    case Inrimage::WT_UNSIGNED_INT :
      ret = _watershed( (unsigned char *)0,
			(unsigned int *)0,
			(unsigned char *)gradientImage->getData(),
			(unsigned int *)inputLabelsImage->getData(),
			(unsigned int *)outputLabelsImage->getData(),
			gradientImage->getX(),
			gradientImage->getY(),
			gradientImage->getZ(),
			gradientImage->getWdim() );
      break;
    }

    break;

  case Inrimage::WT_UNSIGNED_SHORT :

    switch ( inputLabelsImage->getType() ) {
    default :
      throw Exception( from, "Labels image type not handled yet" );
      break;
    case Inrimage::WT_UNSIGNED_CHAR :
      ret = _watershed( (unsigned short int *)0,
			(unsigned char *)0,
			(unsigned short int *)gradientImage->getData(),
			(unsigned char *)inputLabelsImage->getData(),
			(unsigned char *)outputLabelsImage->getData(),
			gradientImage->getX(),
			gradientImage->getY(),
			gradientImage->getZ(),
			gradientImage->getWdim() );
      break;
    case Inrimage::WT_UNSIGNED_SHORT :
      ret = _watershed( (unsigned short int *)0,
			(unsigned short int *)0,
			(unsigned short int *)gradientImage->getData(),
			(unsigned short int *)inputLabelsImage->getData(),
			(unsigned short int *)outputLabelsImage->getData(),
			gradientImage->getX(),
			gradientImage->getY(),
			gradientImage->getZ(),
			gradientImage->getWdim() );
      break;
    case Inrimage::WT_UNSIGNED_INT :
      ret = _watershed( (unsigned short int *)0,
			(unsigned int *)0,
			(unsigned short int *)gradientImage->getData(),
			(unsigned int *)inputLabelsImage->getData(),
			(unsigned int *)outputLabelsImage->getData(),
			gradientImage->getX(),
			gradientImage->getY(),
			gradientImage->getZ(),
			gradientImage->getWdim() );
      break;
    }

    break;

  case Inrimage::WT_UNSIGNED_INT :

    switch ( inputLabelsImage->getType() ) {
    default :
      throw Exception( from, "Labels image type not handled yet" );
      break;
    case Inrimage::WT_UNSIGNED_CHAR :
      ret = _watershed( (unsigned int *)0,
			(unsigned char *)0,
			(unsigned int *)gradientImage->getData(),
			(unsigned char *)inputLabelsImage->getData(),
			(unsigned char *)outputLabelsImage->getData(),
			gradientImage->getX(),
			gradientImage->getY(),
			gradientImage->getZ(),
			gradientImage->getWdim() );
      break;
    case Inrimage::WT_UNSIGNED_SHORT :
      ret = _watershed( (unsigned int *)0,
			(unsigned short int *)0,
			(unsigned int *)gradientImage->getData(),
			(unsigned short int *)inputLabelsImage->getData(),
			(unsigned short int *)outputLabelsImage->getData(),
			gradientImage->getX(),
			gradientImage->getY(),
			gradientImage->getZ(),
			gradientImage->getWdim() );
      break;
    case Inrimage::WT_UNSIGNED_INT :
      ret = _watershed( (unsigned int *)0,
			(unsigned int *)0,
			(unsigned int *)gradientImage->getData(),
			(unsigned int *)inputLabelsImage->getData(),
			(unsigned int *)outputLabelsImage->getData(),
			gradientImage->getX(),
			gradientImage->getY(),
			gradientImage->getZ(),
			gradientImage->getWdim() );
      break;
    }

    break;
  }
}









//
// templates pour la construction des elements structurants
//


//
// recherche du centre et du nombre de points
// cas des images de type signe
//
template <class TYPE> static int get_nb_and_center_from_signed( TYPE* T, 
								int *c,
								TYPE ***buf,
								unsigned int dimx, 
								unsigned int dimy, 
								unsigned int dimz,
								int CENTRE_VALUE )
{
  int n = 0;
  unsigned int x, y, z;
  c[0] = dimx/2;
  c[1] = dimy/2;
  c[2] = dimz/2;
  for ( z=0; z<dimz; z++ )
    for ( y=0; y<dimy; y++ )
      for ( x=0; x<dimx; x++ ) {
	if ( buf[z][y][x] <= - CENTRE_VALUE ) {
	  c[0] = x;
	  c[1] = y;
	  c[2] = z;
	}
	if ( buf[z][y][x] <= 0 ) continue;
	if ( buf[z][y][x] >= CENTRE_VALUE ) {
	  c[0] = x;
	  c[1] = y;
	  c[2] = z;
	}
	n++;
      }
  return( n );
}

//
// recherche du centre et du nombre de points
// cas des images de type non signe
//
template <class TYPE> static int get_nb_and_center_from_unsigned( TYPE* T, 
								  int *c,
								  TYPE ***buf,
								  unsigned int dimx, 
								  unsigned int dimy, 
								  unsigned int dimz,
								  int CENTRE_VALUE )
{
  int n = 0;
  unsigned int x, y, z;
  c[0] = dimx/2;
  c[1] = dimy/2;
  c[2] = dimz/2;
  for ( z=0; z<dimz; z++ )
    for ( y=0; y<dimy; y++ )
      for ( x=0; x<dimx; x++ ) {
	if ( buf[z][y][x] <= 0 ) continue;
	if ( buf[z][y][x] >= (TYPE)CENTRE_VALUE ) {
	  c[0] = x;
	  c[1] = y;
	  c[2] = z;
	}
	n++;
      }
  return( n );
}

  

//
// construction de l'element structurant
//

template <class TYPE > static int buildFromBuffer(  TYPE* T, 
						    StructuringElement *s,
						    int n, int *c,
						    TYPE ***buf,
						    unsigned int dimx, 
						    unsigned int dimy, 
						    unsigned int dimz )
{
  unsigned int x, y, z;

  if ( n <= 0 ) return( _morpho_ErrorInParameters_ );
  typeStructuringElementPoint *myList = NULL;
  myList = (typeStructuringElementPoint *) new typeStructuringElementPoint [n];
  if ( myList == NULL ) return( _morpho_ErrMemFull_ );
  
  for ( n=0, z=0; z<dimz; z++ )
    for ( y=0; y<dimy; y++ )
      for ( x=0; x<dimx; x++ ) {
	if ( buf[z][y][x] <= 0 ) continue;
	myList[n].x = x - c[0];
 	myList[n].y = y - c[1];
	myList[n].z = z - c[2];
	myList[n].o = 0;
	n ++;
      }
  
  s->setNumberOfPoints( n );
  s->setArrayOfPoints( myList );
  return( 1 );
}
					     




//
// exceptions
//
#ifndef LIBINRIMAGE

StructuringElement::InvalidElement::InvalidElement(const std::string& from)
  : Exception(from, "structuring element is not valid") {
}
StructuringElement::InvalidElement::InvalidElement(const std::string& from, const std::string& error)
  : Exception(from, error) {
}

#endif




//
// initialisation des champs
//
void StructuringElement::init( ) 
{
  connectivity = TWENTY_SIX_CONNECTIVITY;
  list = NULL;
  nb = 0;  
  centre_value = 100;
}


//
//
//
void StructuringElement::print( )
{
  COUT<<"*** Structuring Element ***"<<ENDL;
  if ( nb > 0 && list != NULL ) {
    COUT<<" number of points = "<<nb<<ENDL;
    for (int i=0; i<nb; i++ ) {
      printf( " #%4d: %3d %3d %3d\n",
	      i, list[i].x, list[i].y, list[i].z );
    }
  }
  else {
    COUT<<" connectivity = "<<connectivity<<ENDL;
  }
  COUT<<ENDL;
}


//
// Construction d'un element structurant a partir d'une image
//
int StructuringElement::buildFromImage( const Inrimage *image )
{
  int type;
  int ret;
  int n=0, c[3];
  
  if ( image == NULL ) return( _morpho_ErrorInParameters_ );
  if ( image->getVdim() > 1 ) return( _morpho_ErrorInParameters_ );

  type = image->getType();

  switch( type ) {
  default :
    return( _morpho_ErrorInParameters_ );
  case Inrimage::WT_UNSIGNED_CHAR :
    n = get_nb_and_center_from_unsigned( (unsigned char *)0, c, (unsigned char ***)image->getArray(),
					 image->getX(), image->getY(), image->getZ(), centre_value );
    break;
  case Inrimage::WT_SIGNED_CHAR :
    n = get_nb_and_center_from_signed( (char *)0, c, (char ***)image->getArray(),
				       image->getX(), image->getY(), image->getZ(), centre_value );
    break;
  case Inrimage::WT_UNSIGNED_SHORT :
    n = get_nb_and_center_from_unsigned( (unsigned short int *)0, c, (unsigned short int ***)image->getArray(),
					 image->getX(), image->getY(), image->getZ(), centre_value );
    break;
  case Inrimage::WT_SIGNED_SHORT :
    n = get_nb_and_center_from_signed( (short int *)0, c, (short int ***)image->getArray(),
				       image->getX(), image->getY(), image->getZ(), centre_value );
    break;
  case Inrimage::WT_UNSIGNED_INT :
    n = get_nb_and_center_from_unsigned( (unsigned int *)0, c, (unsigned int ***)image->getArray(),
					 image->getX(), image->getY(), image->getZ(), centre_value );
    break;
  case Inrimage::WT_SIGNED_INT :
    n = get_nb_and_center_from_signed( (int *)0, c, (int ***)image->getArray(),
				       image->getX(), image->getY(), image->getZ(), centre_value );
    break;
  case Inrimage::WT_UNSIGNED_LONG :
    n = get_nb_and_center_from_unsigned( (unsigned long *)0, c, (unsigned long ***)image->getArray(),
					 image->getX(), image->getY(), image->getZ(), centre_value );
    break;
  case Inrimage::WT_SIGNED_LONG :
    n = get_nb_and_center_from_signed( (long *)0, c, (long ***)image->getArray(),
				       image->getX(), image->getY(), image->getZ(), centre_value );
    break;
  case Inrimage::WT_FLOAT :
    n = get_nb_and_center_from_signed( (float *)0, c, (float ***)image->getArray(),
				       image->getX(), image->getY(), image->getZ(), centre_value );
    break;
  case Inrimage::WT_DOUBLE :
    n = get_nb_and_center_from_signed( (double *)0, c, (double ***)image->getArray(),
				       image->getX(), image->getY(), image->getZ(), centre_value );
    break;
  }

  if ( n <= 0 ) return( _morpho_ErrorInParameters_ );

  switch( type ) {
  default :
    return( _morpho_ErrorInParameters_ );
  case Inrimage::WT_UNSIGNED_CHAR :
    ret = buildFromBuffer( (unsigned char *)0, this, n, c, (unsigned char ***)image->getArray(),
			   image->getX(), image->getY(), image->getZ() );
    break;
  case Inrimage::WT_SIGNED_CHAR :
    ret = buildFromBuffer( (char *)0, this, n, c, (char ***)image->getArray(),
			   image->getX(), image->getY(), image->getZ() );
    break;
  case Inrimage::WT_UNSIGNED_SHORT :
    ret = buildFromBuffer( (unsigned short int *)0, this, n, c, (unsigned short int ***)image->getArray(),
			   image->getX(), image->getY(), image->getZ() );
    break;
  case Inrimage::WT_SIGNED_SHORT :
    ret = buildFromBuffer( (short int *)0, this, n, c, (short int ***)image->getArray(),
			   image->getX(), image->getY(), image->getZ() );
    break;
  case Inrimage::WT_UNSIGNED_INT :
    ret = buildFromBuffer( (unsigned int *)0, this, n, c, (unsigned int ***)image->getArray(),
			   image->getX(), image->getY(), image->getZ() );
    break;
  case Inrimage::WT_SIGNED_INT :
    ret = buildFromBuffer( (int *)0, this, n, c, (int ***)image->getArray(),
			   image->getX(), image->getY(), image->getZ() );
    break;
  case Inrimage::WT_UNSIGNED_LONG :
    ret = buildFromBuffer( (unsigned long int *)0, this, n, c, (unsigned long int ***)image->getArray(),
			   image->getX(), image->getY(), image->getZ() );
    break;
  case Inrimage::WT_SIGNED_LONG :
    ret = buildFromBuffer( (long *)0, this, n, c, (long ***)image->getArray(),
			   image->getX(), image->getY(), image->getZ() );
    break;
  case Inrimage::WT_FLOAT :
    ret = buildFromBuffer( (float *)0, this, n, c, (float ***)image->getArray(),
			   image->getX(), image->getY(), image->getZ() );
    break;
  case Inrimage::WT_DOUBLE :
    ret = buildFromBuffer( (double *)0, this, n, c, (double ***)image->getArray(),
			   image->getX(), image->getY(), image->getZ() );
    break;
  }
  
  if ( ret != 1 ) return( ret );

  return( 1 );
}



// Constructeur: avec une connexite
StructuringElement::StructuringElement( const CONNECTIVITY_TYPE C )
  throw( InvalidElement )
{
#ifndef LIBINRIMAGE
  char *from = "StructuringElement::StructuringElement( int )";
#endif

  this->init();

  switch( C ) {
  default :
#ifndef LIBINRIMAGE
    throw InvalidElement( from );
#else
    throw InvalidElement();
#endif
    break;
  case FOUR_CONNECTIVITY :
  case EIGHT_CONNECTIVITY :
  case SIX_CONNECTIVITY :
  case TEN_CONNECTIVITY :
  case EIGHTEEN_CONNECTIVITY :
  case TWENTY_SIX_CONNECTIVITY :
    connectivity = C;
  }
}


// Constructeur: avec un voisinage rectangulaire
StructuringElement::StructuringElement( const int x, const int y, const int z )
  throw( InvalidElement )
{
  char *from = "StructuringElement::StructuringElement( const int, const int, const int )";
  int cx, cy, cz;
  int i,j,k;
  int n;
  typeStructuringElementPoint *myList = NULL;

  this->init();

  if ( x <= 0 || y <= 0 || z <= 0 ) {
#ifndef LIBINRIMAGE
    throw InvalidElement( from );
#else
    throw InvalidElement();
#endif
  }

  cx = x/2;
  cy = y/2;
  cz = z/2;
  n = x * y * z;

  myList = (typeStructuringElementPoint *) new typeStructuringElementPoint [n];
  if ( myList == NULL ) 
    throw ErrMemFull( from );
  
  for ( n=0, k=0; k<z; k++ )
    for ( j=0; j<y; j++ )
      for ( i=0; i<x; i++, n++ ) {
	myList[n].x = i - cx;
	myList[n].y = j - cy;
	myList[n].z = k - cz;
	myList[n].o = 0;
      }
  
  this->setNumberOfPoints( n );
  this->setArrayOfPoints( myList );
}



// Constructeur: avec un nom d'image
StructuringElement::StructuringElement( const char *name )
{
  char *from = "StructuringElement::StructuringElement( const char * )";
  Inrimage *image = NULL;

  this->init();

  try {
    image = new Inrimage( name );
  }
  catch( Exception e)  {
    e.push( from );
    throw e;
  }

  int r = this->buildFromImage( image );


  switch ( r ) {
  case 1 :
    break;
  case _morpho_ErrMemFull_ :
    throw ErrMemFull( from );
  case _morpho_ErrorInParameters_ :
  default :
#ifndef LIBINRIMAGE
    throw InvalidElement( from );
#else
    throw InvalidElement();
#endif
  }

  delete image;
}



// Constructeur: avec une image
StructuringElement::StructuringElement( const Inrimage *image )
  throw( InvalidElement )
{
  char *from = "SStructuringElement::StructuringElement( const Inrimage * )";
  
  this->init();
  
  int r = this->buildFromImage( image );

  switch ( r ) {
  case 1 :
    break;
  case _morpho_ErrMemFull_ :
    throw ErrMemFull( from );
  case _morpho_ErrorInParameters_ :
  default :
#ifndef LIBINRIMAGE
    throw InvalidElement( from );
#else
    throw InvalidElement();
#endif
  }

}
  
  
//
// redefinition de =
//
StructuringElement& StructuringElement::operator=(const StructuringElement& s)
{
  char *from = "StructuringElement& StructuringElement::operator=( const StructuringElement& )";
  if ( this != &s  ) {
    this->connectivity = s.connectivity;
    if ( s.nb > 0 && s.list != NULL) {
      if ( this->nb < s.nb ) {
	if ( this->list != NULL ) free( this->list );
	this->list = NULL;
	this->list = (typeStructuringElementPoint *) new typeStructuringElementPoint [s.nb];
	if ( this->list == NULL ) 
	  throw ErrMemFull( from );
      }
      memcpy( this->list, s.list, s.nb * sizeof( typeStructuringElementPoint ) );
      this->nb = s.nb;
    } 
    else {
      this->nb = 0;
      this->list = NULL;
    }
  }
  return *this;
}
  
//
// Create an replica of elem
// 
StructuringElement::StructuringElement( const StructuringElement &elem ) {
  
  connectivity = elem.connectivity;
  nb = elem.nb;
  if (nb >0 )
    list = new typeStructuringElementPoint [nb];
  else
    list = (typeStructuringElementPoint *) NULL;
  typeStructuringElementPoint *tmp = list;
  typeStructuringElementPoint *tmp2 = elem.list;
  
  for (int i=0; i<nb; ++i, ++tmp, ++tmp2) {
    tmp->x = tmp2->x;
    tmp->y = tmp2->y;
    tmp->z = tmp2->z;
    tmp->o = tmp2->o;
  }
  
}


//
// 
//
void StructuringElement::inverse( )
{
#ifndef LIBINRIMAGE
  char *from = "StructuringElement::inverse( )";
#endif
  int n;
  if ( this->nb == 0 && this->list == NULL ) return;
  if ( this->nb > 0 && this->list != NULL ) {
    for (n=0; n<this->nb; n++ ) {
      this->list[n].x = -this->list[n].x;
      this->list[n].y = -this->list[n].y;
      this->list[n].z = -this->list[n].z;
    }
    return;
  }
#ifndef LIBINRIMAGE
  throw InvalidElement( from );
#else
  throw InvalidElement();
#endif
}

//
// 
//
void StructuringElement::convertToList( )
{
  char *from = "StructuringElement::convertToList()";
  int n;
  int i, j, k;

  if ( this->nb > 0 && this->list != NULL ) return;
  if ( this->nb == 0 && this->list == NULL ) {
    switch( this->connectivity ) {
    default :
#ifndef LIBINRIMAGE
      throw InvalidElement( from );
#else
      throw InvalidElement();
#endif
    case FOUR_CONNECTIVITY :       n =  5; break;
    case EIGHT_CONNECTIVITY :      n =  9; break;
    case SIX_CONNECTIVITY :        n =  7; break;
    case TEN_CONNECTIVITY :        n = 11; break;
    case EIGHTEEN_CONNECTIVITY :   n = 19; break;
    case TWENTY_SIX_CONNECTIVITY : n = 27; break;
    }
    this->list = (typeStructuringElementPoint *) new typeStructuringElementPoint [n];
    if ( this->list == NULL ) 
      throw ErrMemFull( from );
    this->nb = n;
    this->list[0].x = 0;   this->list[0].y = 0;   this->list[0].z = 0;   
    switch( this->connectivity ) {
    default :
#ifndef LIBINRIMAGE
      throw InvalidElement( from );
#else
      throw InvalidElement();
#endif
    case  FOUR_CONNECTIVITY :
      this->list[1].x =  0;   this->list[1].y = -1;   this->list[1].z =  0;   
      this->list[2].x = -1;   this->list[2].y =  0;   this->list[2].z =  0;   
      this->list[3].x =  1;   this->list[3].y =  0;   this->list[3].z =  0;   
      this->list[4].x =  0;   this->list[4].y =  1;   this->list[4].z =  0;   
      break;
    case  EIGHT_CONNECTIVITY :
      for ( n=1, j= -1; j<=1; j++ )
	for ( i= -1; i<=1; i++ )
	  if ( i != 0 || j != 0 ) {
	    this->list[n].x =  i;   this->list[n].y =  j;   this->list[n].z =  0;   
	    n++;
	  }
      break;
    case  SIX_CONNECTIVITY :
      this->list[1].x =  0;   this->list[1].y =  0;   this->list[1].z = -1;   
      this->list[2].x =  0;   this->list[2].y = -1;   this->list[2].z =  0;   
      this->list[3].x = -1;   this->list[3].y =  0;   this->list[3].z =  0;   
      this->list[4].x =  1;   this->list[4].y =  0;   this->list[4].z =  0;   
      this->list[5].x =  0;   this->list[5].y =  1;   this->list[5].z =  0;   
      this->list[6].x =  0;   this->list[6].y =  0;   this->list[6].z =  1;   
      break;
    case TEN_CONNECTIVITY :
      this->list[1].x =  0;   this->list[1].y =  0;   this->list[1].z = -1;
      for ( n=2, j= -1; j<=1; j++ )
	for ( i= -1; i<=1; i++ )
	  if ( i != 0 || j != 0 ) {
	    this->list[n].x =  i;   this->list[n].y =  j;   this->list[n].z =  0;   
	    n++;
	  }
      this->list[10].x =  0;   this->list[10].y =  0;   this->list[10].z =  1;   
      break;
    case EIGHTEEN_CONNECTIVITY :
      for ( n=1, j= -1; j<=1; j++ )
	for ( i= -1; i<=1; i++ )
	  if ( j == 0 || i == 0 ) {
	    this->list[n].x =  i;   this->list[n].y =  j;   this->list[n].z = -1; 
	    n++;
	  }
      for ( j= -1; j<=1; j++ )
	for ( i= -1; i<=1; i++ )
	  if ( i != 0 || j != 0 ) {
	    this->list[n].x =  i;   this->list[n].y =  j;   this->list[n].z =  0;   
	    n++;
	  }
      for ( j= -1; j<=1; j++ )
	for ( i= -1; i<=1; i++ )
	  if ( j == 0 || i == 0 ) {
	    this->list[n].x =  i;   this->list[n].y =  j;   this->list[n].z =  1; 
	    n++;
	  }
      
      break;
    case TWENTY_SIX_CONNECTIVITY :
      for ( n=1, k= -1; k<=1; k++ )
	for ( j= -1; j<=1; j++ )
	  for ( i= -1; i<=1; i++ )
	    if ( i != 0 || j != 0 || k != 0 ) {
	      this->list[n].x =  i;   this->list[n].y =  j;   this->list[n].z =  k; 
	      n++;
	    }
      break;
    }
    for ( n=0; n<this->nb; n++ ) 
      this->list[n].o = 0;
    return;
  }
#ifndef LIBINRIMAGE
  throw InvalidElement( from );
#else
  throw InvalidElement();
#endif
}




// Destructeur
// StructuringElement::~StructuringElement() {
//   if ( list != NULL ) free( list );
// }

StructuringElement::~StructuringElement() {
  if ( list != NULL )
    delete list;
}

} // namespace yav
