#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/TclColorMap.h"

#include <inrimage/TclColorMap.h>
#include <inrimage/TclInrimage.h>

namespace yav {


TclColorMap::TclColorMap(TclModule *m, const unsigned char *r,
			 const  unsigned char *g,const  unsigned char *b) : 
  TclObject(typeid(TclColorMap),m) {
  done(typeid(TclColorMap));
  unsigned int i;
  for (i=0;i<258;++i)
    {
      red[i]=r[i];
      green[i]=g[i];
      blue[i]=b[i];
    }
}

TclColorMap::TclColorMap(TclModule *m, Inrimage *inr) throw(TclColorMap::ImageWrongSize) : TclObject(typeid(TclColorMap),m)
{
  done(typeid(TclColorMap));
  if (((inr->getX()!=256) && (inr->getX()!=257) && (inr->getX()!=258)) || 
      (inr->getY()!=3) || (inr->getZ()!=1) || (inr->getType()!=Inrimage::WT_UNSIGNED_CHAR))
    {
      throw (TclColorMap::ImageWrongSize("TclColorMap::TclColorMap(TclModule *, Inrimage *)"));
    }
  else
    {
      unsigned int sizeX=inr->getX();
      unsigned int i;
      unsigned char ***c=(unsigned char ***) inr->getArray();

      for(i=0;i<sizeX;++i) 
	red[i]=c[0][0][i];
      for(i=0;i<sizeX;++i) 
	green[i]=c[0][1][i];
      for(i=0;i<sizeX;++i) 
	blue[i]=c[0][2][i];
	  if (sizeX==256)
	  {
		  red[257]=red[255];
		  green[257]=green[255];
		  blue[257]=blue[255];
		  red[256]=red[0];
		  green[256]=green[0];
		  blue[256]=blue[0];
	  } else if (sizeX==257)
	  {
		  red[257]=red[256];
		  green[257]=green[256];
		  blue[257]=blue[256];
		  red[256]=red[0];
		  green[256]=green[0];
		  blue[256]=blue[0];
	  } 
    }
}
TclColorMap::TclColorMap(TclModule *m, char *fname) throw(TclColorMap::FileNotFound) : TclObject(typeid(TclColorMap),m)
{
  done(typeid(TclColorMap));
  unsigned int i,v;
  IFSTREAM input(fname);
  if (!input)
    throw FileNotFound("TclColorMap::TclColorMap(TclModule *m, char *fname)",(char *)fname);
  for (i=0;i<257;++i)
    {
      input >> v;
      red[i]=v;
    }
   for (i=0;i<257;++i)
    {
      input >> v;
      green[i]=v;
    }
    for (i=0;i<257;++i)
      {
      input >> v;
      blue[i]=v;
    }
	  red[257]=red[256];
	  green[257]=green[256];
	  blue[257]=blue[256];
	  red[256]=red[0];
	  green[256]=green[0];
	  blue[256]=blue[0];
}
int TclColorMap::create(TclModule *m, int argc, char **argv) {
  TclColorMap *tc=0;

  if (argc!=2)
    return COMMAND_SYNOPSIS;
  try {
    TclInrimage *inr = new TclInrimage(m,argv[1]);
    tc=new TclColorMap(m,inr->getClient());
  }
  catch (Exception e) {
    CERR << " in exception" << ENDL;
    try { 
      tc=new TclColorMap(m,argv[1]);
    }
    catch (Exception e)  {
      m->setResult("%s is not an image or a colormap file",argv[1]);
      return TCL_ERROR;
    }
  }
  m->setResult("%s%u", tc->getObjectName(), tc->TclObject::getRef());
  return TCL_OK;
}
int TclColorMap::destroyCmd(int argc, char **argv) {
  TclModule *m;
  unsigned int ref;

  if(argc != 0) return COMMAND_SYNOPSIS;
  m = getModule();
  ref = TclObject::getRef();
  delete this;
  m->setResult("%s%u has been destroyed", TclColorMap::getObjectName(), ref);
  return TCL_OK;
}
int TclColorMap::getColorCmd(int argc, char **argv) {
  if (argc==0)
    {
      TclString res="{ ";
      unsigned int i;
      TclString tmp;
      for(i=0;i<258;++i)
	{
	  if (i==257)
	    tmp.sprintf(" { %u %u %u } }",red[i],green[i],blue[i]);
	  else
	    tmp.sprintf(" { %u %u %u }, ",red[i],green[i],blue[i]);
	  res+=tmp;
	}
      setResult(res);
      return TCL_OK;
    }
  else if (argc==1)
    {
      unsigned int v=(unsigned int ) atoi(argv[0]);
      if (v<258)
	{
	  setResult("%u %u %u",red[v],green[v],blue[v]);
	  return TCL_OK;
	}
      else 
	{
	  setResult("%u must be between 0 and 257",v);
	  return TCL_ERROR;
	}
      
    }
  else return COMMAND_SYNOPSIS;
}
int TclColorMap::setColorCmd(int argc, char **argv) {
  if (argc==4)
    {
      unsigned int v=(unsigned int ) atoi(argv[0]);
      unsigned char r=(unsigned char ) atoi(argv[1]);
      unsigned char g=(unsigned char ) atoi(argv[2]);
      unsigned char b=(unsigned char ) atoi(argv[3]);
      if (v<258)
	{
	  red[v]=r;
	  green[v]=g;
	  blue[v]=b;
	  setResult("%u %u %u",red[v],green[v],blue[v]);
	  return TCL_OK;
	}
      else 
	{
	  setResult("%u must be between 0 and 257",v);
	  return TCL_ERROR;
	}
      
    }
  else return COMMAND_SYNOPSIS;
}


} // namespace yav



