#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/TclImage3DCameraBackground.h"

#include <inrimage/TclImage3DCameraBackground.h>
#include <graphics2D/Camera2DActivation.h>
#include <inrimage/TclImage3DBackgroundManager.h>
#include <inrimage/TclInrimage.h>
#include <graphics2D/TclCamera2D.h>
#include <graphics2D/SliceScale.h>
#include <basics/Constants.h>

namespace yav {

	TclImage3DCameraBackground::TclImage3DCameraBackground(TclCamera2D *cam,TclImage3DBackgroundManager *m,
		SLICE_ORIENTATION geom) : Image3DCameraBackground(cam->getClient(),(Image3DBackgroundManager *)m,
		geom), TclObject(typeid(TclImage3DCameraBackground),cam->getModule()) 
	{

		cam->getClient()->toggleCopyright();
		cam->bindMouseMotion(true);
		cam->getClient()->setBackgroundColor((float)(232.0/255.0),(float)(224/255.0),
			(float)(210/255.0));
		/// install menu
		tickle("inrimage::backgroundMenu [%s -widget].popup %s",(char *)cam->name(),
			(char *) cam->name());
		done(typeid(TclImage3DCameraBackground));
		}

	TclString TclImage3DCameraBackground::objectName(void) const{
		return name();
	}

	void TclImage3DCameraBackground::imageHasChangedSize(Inrimage *inr){
		Image3DCameraBackground::imageHasChangedSize(inr);
		if (getOwners().size()>0) {
			 DPOwner<Scalable> *spc=  (*(getOwners().begin()))->getOwner();
			 SliceScale *slScale=dynamic_cast<SliceScale *>(spc);
			if (slScale) {
				slScale->reset(this);
			} 
		}
	}
	void TclImage3DCameraBackground::scaleValue(const TclScale *scale, double val){
		Image3DCameraBackground::setSliceNumber((unsigned int)val);
		if (getManager()) 
			getManager()->moveOtherManagers();
	}

  
	void TclImage3DCameraBackground::scaleTo(const TclScale *scale, double val){
	}

  
	double TclImage3DCameraBackground::getMin(const TclScale *scale) const{
		return 0.0;
	}

 
	double TclImage3DCameraBackground::getMax(const TclScale *scale) const{
		switch(geometry) {
			case SLICE_Z:
				return ((double)getInrimage()->getZ()-1);
				break;
			case SLICE_X:
				return ((double)getInrimage()->getX()-1);
				break;
			case SLICE_Y:
				return ((double)getInrimage()->getY()-1);
				break;
			default:
				break;
		}
		return 0.0;
	}

	double TclImage3DCameraBackground::getStep(const TclScale *scale) const{
		return 1.0;
	}

	double TclImage3DCameraBackground::getSlice(const TclScale *scale) const{
		return (double) sliceNumber;
	}

	void TclImage3DCameraBackground::setSliceNumber(unsigned int s) {
		Image3DCameraBackground::setSliceNumber(s);
		if (getOwners().size()>0) {
			DPOwner<Scalable> *spc=  (*(getOwners().begin()))->getOwner();
			 SliceScale *slScale=dynamic_cast< SliceScale *>(spc);
			if (slScale) {
				slScale->set((double)s);

			} 
		}

	}

TclImage3DCameraBackground::~TclImage3DCameraBackground(){
	deleted(typeid(TclImage3DCameraBackground));
}
 /*		if(useSliceScale) {
			if(!getClient()->getScaleCamera()) {
				TclString sname = getWindow() + ".s";
				getModule()->tickle("scale %s -orient horizontal -width 10 -font fixed -label %s -bg %s", (char *) sname, getClient()->sliceName(getClient()->getGeometry()), color);
				getClient()->setScaleCamera( new SliceScale(getModule(), sname,getClient()));
				getModule()->tickle("pack %s -side bottom  -fill x -before %s",
					(char *) sname, (char *) getCanvas() );
			}
*/
int TclImage3DCameraBackground::sliceScaleCmd(int argc, char **argv){
	if (argc==0) {
		if (getOwners().size()>0) {
			const DPOwner<Scalable> *spc=  (*(getOwners().begin()))->getOwner();
			const SliceScale *slScale=dynamic_cast<const SliceScale *>(spc);
			if (slScale) {
				setResult(slScale->name());
				return TCL_OK;
			} 
		}
		setResult("none");
		return TCL_OK;
	} 
	else if (argc==1) {
		if (getOwners().size()>0) {
			 DPOwner<Scalable> *spc=  (*(getOwners().begin()))->getOwner();
			 SliceScale *slScale=dynamic_cast< SliceScale *>(spc);
			if (slScale) {
				delete slScale; 
			} 
		}
		SliceScale *sliceScale=new SliceScale(getModule(), argv[0],this);
		sliceScale->reset(this);
		setResult(sliceScale->name());
		return TCL_OK;
	} else return COMMAND_SYNOPSIS;
	
}
 int TclImage3DCameraBackground::sliceNumberCmd(int argc, char **argv){
	 if (argc==0) {
		 setResult("%d",sliceNumber);
		 return TCL_OK;
	 } else if (argc==1) {
		 unsigned int s=(unsigned int) atoi(argv[0]);
		 setSliceNumber(s);
		 setResult("%d",sliceNumber);
		 return TCL_OK;
	 } else return COMMAND_SYNOPSIS;
 }
 int TclImage3DCameraBackground::cameraCmd(int argc, char **argv){
	 if (argc!=0)
		 return COMMAND_SYNOPSIS;
	 setResult(((TclCamera2D *) (getCamera()->getInterface()))->name());
	 return TCL_OK;
 }
 int TclImage3DCameraBackground::sliceNameCmd(int argc, char **argv){
	 if (argc!=0)
		 return COMMAND_SYNOPSIS;
	 setResult(sliceName);
	 return TCL_OK;
 }
  int TclImage3DCameraBackground::requiredSizeCmd(int argc, char **argv){
	 if (argc!=0)
		 return COMMAND_SYNOPSIS;
	 setResult("%f %f",getRequiredWidth(),getRequiredHeight());
	 return TCL_OK;
 }
  /// -destroy option
int TclImage3DCameraBackground::destroyCmd(int argc, char **argv){
	 if(argc != 0) return COMMAND_SYNOPSIS;
  TclModule *m = getModule();
  unsigned int u = TclObject::getRef();
  m->setResult("%s%u does not exist any longer",
	       TclImage3DCameraBackground::getObjectName(), u);
  delete this;
  return TCL_OK;
}

  /// new TCL command callback
int TclImage3DCameraBackground::create(TclModule *m, int argc, char **argv){
	TclImage3DCameraBackground *b=0;
	TclImage3DBackgroundManager *manager=0;
	SLICE_ORIENTATION geom;
	TclCamera2D *cam;
	if(argc == 4) {
		cam= (TclCamera2D *) TclObjectKind::getObject(typeid(TclCamera2D), argc, argv, 1);
		if (!cam) {
			m->setResult("%s must be a 2D camera",argv[1]);
			return TCL_ERROR;
		}	
		manager= (TclImage3DBackgroundManager *) TclObjectKind::getObject(typeid(TclImage3DBackgroundManager), argc, argv, 2);
		if (!manager) {
			m->setResult("%s must be a background manager",argv[2]);
			return TCL_ERROR;
		}
		if (!strcmp(argv[3],"X")) 
			geom=SLICE_X;
		else 	  if (!strcmp(argv[3],"Y")) 
			geom=SLICE_Y;
		else if (!strcmp(argv[3],"Z")) 
			geom=SLICE_Z;
		else {
			m->setResult("%s must be a slice geometry : either X,Y or Z",argv[3]);
			return TCL_ERROR;
		}
		try{
			b= new TclImage3DCameraBackground(cam,manager,geom);
		}
		catch (Exception e) {
			CERR << e << ENDL;
			m->setResult("Could not create Image3DCameraBackground");
		}

	}
	else
		return COMMAND_SYNOPSIS;
	m->setResult(b->name());
	return TCL_OK;
}
void TclImage3DCameraBackground::createCommands(TclModule *m) {
   TclObjectKind *kind;

	kind = new TclObjectKind(m, typeid(TclImage3DCameraBackground),
		"imageBgnd",
		"Background associated with a 3D image in 2D camera","cam%u multiview%u X|Y|Z",
			   &TclImage3DCameraBackground::create, 
			   &TclImage3DCameraBackground::list);

  kind->newOption("-destroy", (TclObject::OPTION) &TclImage3DCameraBackground::destroyCmd, 0,
		  "destroy a 2D camera background");
    kind->newOption("-sliceNumber", (TclObject::OPTION) &TclImage3DCameraBackground::sliceNumberCmd, "[sliceNumber=%u]",
		  "set or get the current slice Number");
    kind->newOption("-sliceScale", (TclObject::OPTION) &TclImage3DCameraBackground::sliceScaleCmd, "[scaleWidget=%s]",
		  "get the sliceScale object associated with that background or create a sliceScale object (and connect it to the background) given the the scale widget name");
    kind->newOption("-sliceName", (TclObject::OPTION) &TclImage3DCameraBackground::sliceNameCmd, 0,
		  "get the slice name associated with that background");
    kind->newOption("-camera", (TclObject::OPTION) &TclImage3DCameraBackground::cameraCmd, 0,
		  "returns the camera associated with this background");
    kind->newOption("-requiredSize", (TclObject::OPTION) &TclImage3DCameraBackground::requiredSizeCmd, 0,
		  "returns the width and height required to draw the texture");

}

}
