#ifdef _MSC_VER
#pragma warning ( disable : 4068 )
#pragma warning ( disable : 4786 )
#pragma warning ( disable : 4081 )
#endif

#pragma implementation "inrimage/TclImageLandmarkSet.h"

#include <inrimage/TclImageLandmarkSet.h>
#include <inrimage/Image3DCameraBackground.h>
#include <graphics2D/TclCamera2D.h>
#include <modules/TclMatrix4D.h>

namespace yav {

TclImageLandmarkSet::TclImageLandmarkSet(TclModule *m, const std::type_info& info, ImageLandmarkSet *data, TclScene2D *scene)
  : TclScene2DObject(info, data ? data : new ImageLandmarkSet(scene->getClient()), m)
{
  init();
}

TclImageLandmarkSet::TclImageLandmarkSet(TclModule *m, ImageLandmarkSet *data, TclScene2D *scene)
  : TclScene2DObject(typeid(TclImageLandmarkSet), data ? data : new ImageLandmarkSet(scene->getClient()), m)
{
  init();
}
TclImageLandmarkSet::TclImageLandmarkSet(TclModule *m, char *filename, TclScene2D *scene)
  : TclScene2DObject(typeid(TclImageLandmarkSet), new ImageLandmarkSet(scene->getClient(),filename), m)
{
  init();
}
  
TclImageLandmarkSet::~TclImageLandmarkSet(){
  deleted(typeid(TclImageLandmarkSet));
}
void TclImageLandmarkSet::buttonMotionCallback(Camera2D *slicecam, 
					    int  mx, int my) {

}
void TclImageLandmarkSet::buttonReleasedCallback(Camera2D *slicecam, 
						 int  mx, int my) {

//  if ((selectionMode==MOVE_LANDMARK) && (movingLandmark)) 
 //   movingLandmark=0;
}
void TclImageLandmarkSet::buttonPressedCallback(Camera2D *slicecam, 
					    int  mx, int my) {

}
void TclImageLandmarkSet::init() {
  initSelectionModes();
  setInstallPopupMenuCommandName("inrimage::installLandmarkSetPopupMenu");
  done(typeid(TclImageLandmarkSet),"inrimage::installLandmarkSetMenu","inrimage::uninstallLandmarkSetMenu");
}
int TclImageLandmarkSet::create(TclModule *m, int argc, char **argv)
{

  if ( argc < 1) return COMMAND_SYNOPSIS;
  
  TclScene2D *scn2D=0; 
  char *filename=0;
  TclImageLandmarkSet *lset;

  for(int i = 1; i < argc; i++) {
    if (!strcmp(argv[i], "-scene")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      else
	scn2D = (TclScene2D *) TclObjectKind::getObject(typeid(TclScene2D), argc, argv, i);
    }
    else
      filename=argv[i];
    }
    if(!scn2D) scn2D =  TclScene2D::getDefaultScene(m);
    
    if (filename) {
      try {
	lset = new TclImageLandmarkSet(m, filename, scn2D);
      } 
      catch(...) {
	m->setResult("Illegal landmark set filename %s",filename);
	return TCL_ERROR;
      }
    }
    else
      lset = new TclImageLandmarkSet(m, (ImageLandmarkSet *)NULL, scn2D);
    m->setResult("%s%u",lset->getObjectName(), lset->getRef());
    return TCL_OK;
}
int TclImageLandmarkSet::destroyCmd(int argc, char **argv) {
  TclModule *m;

  if(argc != 0) return COMMAND_SYNOPSIS;
  m = getModule();
  TclString n=name();
  delete this;
  m->setResult("%s destroyed",(char *)n);
  return TCL_OK;
}


void TclImageLandmarkSet::initSelectionModes() {
 SelectionMode *mode= new SelectionMode();
 mode->name="select";
 mode->label="Select Landmark";
 mode->TclUnInstallProc="inrimage::uninstallLandmarkMenu";
 mode->installProc=&SelectLandmarkInstallProc;
 selectionModeMap[(char *)mode->name]=mode;

 mode= new SelectionMode();
 mode->name="create";
 mode->label="Add Landmark";
 mode->TclInstallProc="inrimage::installLandmarkMenu";
 mode->installProc=&CreateLandmarkInstallProc;
 mode->TclUnInstallProc="inrimage::uninstallLandmarkMenu";
 selectionModeMap[(char *)mode->name]=mode;
 selectedMode=0;

 mode= new SelectionMode();
 mode->name="move";
 mode->label="Move Landmark";
 mode->installProc=&MoveLandmarkInstallProc;
 selectionModeMap[(char *)mode->name]=mode;


 mode= new SelectionMode();
 mode->name="delete";
 mode->label="Delete Landmark";
 mode->installProc=&DeleteLandmarkInstallProc;
 selectionModeMap[(char *)mode->name]=mode;

}

int TclImageLandmarkSet::nbLandmarksCmd(int argc, char **argv) 
{
  if (argc!=0) return COMMAND_SYNOPSIS;
  setResult("%u", getClient()->getNumberOfLandmarks());
  return TCL_OK;
}

int TclImageLandmarkSet::addLandmarkCmd(int argc, char **argv)
{

  if ((argc!=6)&& (argc!=7)) return COMMAND_SYNOPSIS;

  double x = atof(argv[0]);
  double y = atof(argv[1]);
  double z = atof(argv[2]);

  double a = atof(argv[3]);
  double b = atof(argv[4]);
  double c = atof(argv[5]);

  TclString identifiant;
  unsigned int   rank = getClient()->getNumberOfLandmarks()+1;

  if (argc==6) {
    identifiant.sprintf("landmark_no_%u",rank);
  } else {
    identifiant=argv[6];
  }
 
  ImageLandmark *imagelandmark = new ImageLandmark(Vec3<double>(x,y,z), Vec3<double>(a,b,c), identifiant);
  getClient()->addLandmark(imagelandmark);
  getClient()->redrawScene();

  setResult("landmark %u has been added", rank );
 
  return TCL_OK;
}



int TclImageLandmarkSet::removeLandmarkCmd(int argc, char **argv)
{
  if (argc!=1) return COMMAND_SYNOPSIS;

  ObjectsList<ImageLandmark>::iterator it;
  it=getClient()->firstLandmark();
  
  for(;it!=getClient()->lastLandmark() &&  ( (*it)->getName()!=(const char *)argv[0]);it++);

  if ( it == getClient()->lastLandmark() )
    {
      setResult("Unkwown landmark : could not remove it");
    }
  else
    {
      getClient()->removeLandmark( (*it) );
      getClient()->redrawScene();
      setResult("landmark removed");
    }

  return TCL_OK;
}




int TclImageLandmarkSet::sceneCmd(int argc, char **argv) 
{
  if(argc > 1) return COMMAND_SYNOPSIS;
  else if (argc==0) {
    setResult("%s%u",
	      TclObjectKind::getObjectName(typeid(TclScene2D)),
	      ((TclScene2D *)getClient()->getScene()->getInterface())->getRef());
    return TCL_OK;
  } else {
    TclScene2D *scn=(TclScene2D *) TclObjectKind::getObject(typeid(TclScene2D), argc, argv, 0);	
    if (!scn) {
      setResult("%s must be a 2D scene",argv[0]);
      return TCL_ERROR;
    }
    getClient()->Data2D::setScene(scn->getClient());
    setResult("%s%u",
	      TclObjectKind::getObjectName(typeid(TclScene2D)),
	      getTclScene()->getRef());
    return TCL_OK;
  }
}
int TclImageLandmarkSet::displayNameCmd(int argc, char **argv) {
  if (argc>1)
    return COMMAND_SYNOPSIS;
  if (argc==1) {
    bool selected;
    if (TclModule::boolValue(argv[0],selected)) {
      getClient()->setDisplayName(selected);
    }
    else {
      setResult("%s must be true or false",argv[0]);
      return TCL_ERROR;
    }
  }
  setResult("%s",TclModule::boolString(getClient()->getDisplayName()));
  return TCL_OK;
}
int TclImageLandmarkSet::crossSizeCmd(int argc, char **argv) {
  if (argc>1)
    return COMMAND_SYNOPSIS;
  if (argc==1) {
    unsigned int s=(unsigned int) atoi(argv[0]);
    if (s>0) {
      getClient()->setCrossSize(s);
      getClient()->redrawScene();
    }
    else {
      setResult("%s must be greater than 0",argv[0]);
      return TCL_ERROR;
    }
  }
  setResult("%u",getClient()->getCrossSize());
  return TCL_OK;
}
int TclImageLandmarkSet::colorCmd(int argc, char **argv) {

  if (argc==0)
    {  
      float r,g,b;
      getClient()->getColor(r,g,b);
      setResult("%f %f %f",r,g,b);
      return TCL_OK;
    }
  else if (argc==3)
    {
      float r,g,b;
      getClient()->setColor((float)atof(argv[0]),
			    (float)atof(argv[1]),(float)atof(argv[2]));
      getClient()->redrawScene();
      getClient()->getColor(r,g,b);
      setResult("%f %f %f",r,g,b);
      return TCL_OK;
    }
  else 
    return COMMAND_SYNOPSIS;
}
int TclImageLandmarkSet::selectedColorCmd(int argc, char **argv) {

  if (argc==0)
    {  
      float r,g,b;
      getClient()->getSelectedColor(r,g,b);
      setResult("%f %f %f",r,g,b);
      return TCL_OK;
    }
  else if (argc==3)
    {
      float r,g,b;
      getClient()->setSelectedColor((float)atof(argv[0]),
				    (float)atof(argv[1]),(float)atof(argv[2]));
      getClient()->redrawScene();
      getClient()->getSelectedColor(r,g,b);
      setResult("%f %f %f",r,g,b);
      return TCL_OK;
    }
  else 
    return COMMAND_SYNOPSIS;
}

int TclImageLandmarkSet::displayModeCmd(int argc, char **argv) {
  if (argc>1) return COMMAND_SYNOPSIS;
  if (argc==1) {
    if (!strcmp(argv[0],"voxel")) {
      getClient()->setDisplayMode(ImageLandmarkSet::VOXEL);
      getClient()->redrawScene();
    }
    else if (!strcmp(argv[0],"cross")) {
      getClient()->setDisplayMode(ImageLandmarkSet::CROSS);
      getClient()->redrawScene();
    }
    else if (!strcmp(argv[0],"hide")) {
      getClient()->setDisplayMode(ImageLandmarkSet::HIDE);
      getClient()->redrawScene();
    }
    else {
      setResult("%s must be 'voxel' or 'cross'",argv[0]);
      return TCL_ERROR;
    }
  }
  if (getClient()->getDisplayMode()==ImageLandmarkSet::VOXEL) 
    setResult("voxel");
  else if (getClient()->getDisplayMode()==ImageLandmarkSet::HIDE) 
    setResult("hide");
  else 
    setResult("cross");
  return TCL_OK;
}

int TclImageLandmarkSet::saveAsCmd(int argc, char **argv) {
  if (argc!=1) 
    return COMMAND_SYNOPSIS;    
  getClient()->write(argv[0]);
  setResult("landmark set saved in %s", argv[0]);
  return TCL_OK;
}
int TclImageLandmarkSet::selectedLandmarkCmd(int argc, char **argv) 
{
//	char from[] = "int TclImageLandmarkSet::selectedLandmarkCmd(int argc, char **argv)";
	if (argc>1) 
		return COMMAND_SYNOPSIS;    
	if (argc==1) {
		if (!strcmp(argv[0],"info")) {
			ImageLandmark *land= getClient()->getSelectedLandmark();
			if (land) {
				Vec3<unsigned int> voxelPosition=land->getVoxelPosition();
				Vec3<double> realPosition=land->getRealPosition();

				setResult("%u %u %u %f %f %f \"%s\" ",voxelPosition.getX(),voxelPosition.getY(),voxelPosition.getZ(),
					realPosition.getX(),realPosition.getY(),realPosition.getZ(),(char *)land->getName());
			}
			return TCL_OK;
		}
		else if (!strcmp(argv[0],"none"))
			getClient()->setSelectedLandmark((ImageLandmark *)NULL);
		else {
			getClient()->setSelectedLandmark((ImageLandmark *)NULL);
			ObjectsList<ImageLandmark>::iterator it = getClient()->firstLandmark();
			for(;it!= getClient()->lastLandmark();it++)
			{
				if ((*it)->getName()==(const char *)argv[0]){
					getClient()->setSelectedLandmark(*it);
					break;
				}
			}
			if (getClient()->getSelectedLandmark()){
				Scene2D *scn2D = getClient()->getScene();
				ObjectsList< Camera2D >::iterator it =  scn2D->firstCamera();
				for (; it != scn2D->lastCamera(); ++it){
					Image3DCameraBackground *b=dynamic_cast<Image3DCameraBackground *> ((*it)->getBackground());
					if (b) {
					  Vec3<double> p=getClient()->getSelectedLandmark()->getRealPosition();
						b->moveToPoint(p);
					}
				}
			}
			getClient()->redrawScene();
		}
	}

	if (getClient()->getSelectedLandmark())
		setResult(getClient()->getSelectedLandmark()->getName());
	else
		setResult("none");
	return TCL_OK;
}

	


int TclImageLandmarkSet::landmarksCmd(int argc, char **argv) 
{
  if (argc>0)
    return COMMAND_SYNOPSIS;    
  TclString out;
  TclString l;
  Vec3<unsigned int> voxelPosition;
  Vec3<double> realPosition;

  ObjectsList<ImageLandmark>::iterator it = getClient()->firstLandmark();
  for(;it!= getClient()->lastLandmark();it++) {
    voxelPosition=(*it)->getVoxelPosition();
    realPosition=(*it)->getRealPosition();

    l.sprintf("{ %u %u %u %f %f %f {%s} } ",voxelPosition.getX(),voxelPosition.getY(),voxelPosition.getZ(),
	      realPosition.getX(),realPosition.getY(),realPosition.getZ(),
	      (char *) (*it)->getName());
    out+=l;
  };

  setResult(out);
  return TCL_OK;
}




int TclImageLandmarkSet::computeErrorCmd(int argc, char *argv[]){
  if (argc!=1) return COMMAND_SYNOPSIS;
  TclImageLandmarkSet *lset2 = TclImageLandmarkSet::getObject(argc, argv, 0);
  if (!lset2)
    return TCL_ERROR;
  
  std::list<double> l = getClient()->computeError(lset2->getClient());

  std::list<double>::iterator it = l .begin();
  TclString res, out;
  
  for (; it != l.end(); ++it){
    res.sprintf("%f ", (*it));
    out += res;
  }

  setResult(out);
  return TCL_OK;
}


int TclImageLandmarkSet::registerCmd(int argc, char *argv[]){
  bool resample=false;
  TclScene2D *scn2D = (TclScene2D *)NULL;

  if (argc>6) return COMMAND_SYNOPSIS;

  TclString transfotype = "rigid";
  TclImageLandmarkSet *lset2 = TclImageLandmarkSet::getObject(argc, argv, 0);
  if (!lset2)
    return TCL_ERROR;
   
  if (argc>1) {
    transfotype=argv[1];
    if (strcmp(transfotype, "rigid"))
      if (strcmp(transfotype, "similarity"))
	if (strcmp(transfotype, "affine"))
	  return COMMAND_SYNOPSIS;
  }
    
  int lset2resample;

 if (argc>=3) {
   if (!(strcmp(argv[2], "-resample"))) resample=true;
   else return COMMAND_SYNOPSIS;
   if (argc>3) {
     if (argc!=6) return COMMAND_SYNOPSIS;
     if (sscanf(argv[3], "%u", &lset2resample) != 1) return COMMAND_SYNOPSIS;
     if ( (lset2resample!=1)&&(lset2resample!=2)) return COMMAND_SYNOPSIS;
     if (!(strcmp(argv[4], "-scene"))){
       scn2D = (TclScene2D *) TclObjectKind::getObject(typeid(TclScene2D), argc, argv, 5);
       if (!scn2D) return TCL_ERROR;
     }
     else return COMMAND_SYNOPSIS;
   }
   else 
     scn2D = getTclScene();
 }

 Transformation *transfo = getClient()->computeTransfo(lset2->getClient(), transfotype);
 // si on doit resampler lset au lieu de this

 MatrixTransformation *transfo_inverse = (MatrixTransformation *)NULL;
 MatrixTransformation *matrixtransfo = dynamic_cast<MatrixTransformation *> (transfo);
 if (!matrixtransfo)
   return TCL_ERROR;
 
 // on recupere la matrice de transformation qui transforme le getClient() en lset2->getClient()
  Matrix4x4<double> mat = matrixtransfo->getMatrix();
  TclMatrix4D *theMat = new TclMatrix4D(getModule(), mat);  


  // on applique la transfo au lset a resampler
  ImageLandmarkSet *lset_resample = (ImageLandmarkSet *)NULL;
  TclImageLandmarkSet *tclset = (TclImageLandmarkSet *)NULL;
  PointSetIterator *it = (ImageLandmarkIterator *)NULL;
  ImageLandmark *new_lmark = (ImageLandmark *)NULL;

  if (resample) {
    lset_resample = new ImageLandmarkSet(scn2D->getClient());

    if (lset2resample == 1) {
      ObjectsList<ImageLandmark>::iterator tmp1 = getClient()->firstLandmark();
      for (; tmp1!=getClient()->lastLandmark(); ++tmp1){
	new_lmark = new ImageLandmark((*tmp1)->getVoxelPosition(), (*tmp1)->getRealPosition(), (*tmp1)->getName());
	lset_resample->addLandmark(new_lmark);
      }
      
      it = new ImageLandmarkIterator(lset_resample);
      it->first();
      transfo->apply(it);
      tclset = new TclImageLandmarkSet(getModule(), lset_resample, scn2D);
    } else {
      Matrix4x4<double> mat_inverse(mat);
      
      mat_inverse.inv(mat_inverse);
      transfo_inverse = dynamic_cast<MatrixTransformation *> (transfo);
      
      transfo_inverse->setMatrix(mat_inverse);      
      
      ObjectsList<ImageLandmark>::iterator tmp1 = lset2->getClient()->firstLandmark();
      for (; tmp1!=lset2->getClient()->lastLandmark(); ++tmp1){
	new_lmark = new ImageLandmark((*tmp1)->getVoxelPosition(), (*tmp1)->getRealPosition(), (*tmp1)->getName());
	lset_resample->addLandmark(new_lmark);
      }
      it = new ImageLandmarkIterator(lset_resample);
      it->first();
      transfo_inverse->apply(it);
      tclset = new TclImageLandmarkSet(getModule(), lset_resample, scn2D);
    }    
    setResult("%s%u", tclset->getObjectName(), tclset->getRef());
  }
  
  delete transfo;
  
  TclString res;
  res.sprintf("%s%u", theMat->getObjectName(), theMat->getRef());

  if (resample) {
    TclString tmp;
    tmp.sprintf(" %s%u", tclset->getObjectName(), tclset->getRef());
    res+= tmp;
  }
  
  setResult(res);
  
  return TCL_OK;
  
}

int TclImageLandmarkSet::initCmd(int argc, char **argv) {

  if (argc!=1) return COMMAND_SYNOPSIS;

  for (unsigned int i=0; i<getClient()->getNumberOfLandmarks(); ++i)
    getClient()->removeLandmark((*(getClient()->firstLandmark())));
  
  TclImageLandmarkSet *lset2 = TclImageLandmarkSet::getObject(argc, argv, 0);
  if (!lset2) return TCL_ERROR;

  ObjectsList<ImageLandmark>::iterator it = lset2->getClient()->firstLandmark();
  for (; it != lset2->getClient()->lastLandmark(); ++it) {
    getClient()->addLandmark(new ImageLandmark((*it)->getVoxelPosition(), (*it)->getRealPosition(), (*it)->getName()));
  }

  setResult("%s%u initialized with %s%u", getObjectName(), getRef(), lset2->getObjectName(), lset2->getRef());

  return TCL_OK;

}
void TclImageLandmarkSet::createCommands(TclModule *m)
{
	TclObjectKind *kind;

	kind = new TclObjectKind(m, typeid(TclImageLandmarkSet), "lset",
		"Set of Image Landmarks", &TclImageLandmarkSet::create,
		&TclImageLandmarkSet::list, "[%s] [-scene scn2D%u]");
	m->newCommand("lset%u", kind->getOptions(),
		"Set of Image Landmarks management command");
	TclScene2DObject::createOptionCommands(kind);
	kind->newOption("-init", (TclObject::OPTION) &TclImageLandmarkSet::initCmd, "lset%u", "init a landmarkset from another landmarkset affecting all the second landmarkset landmarks positions to the first landmarkset landmarks. All the landmarks of the first lset are removed");
	kind->newOption("-register", (TclObject::OPTION) &TclImageLandmarkSet::registerCmd, "lset%u [rigid|similarity|affine] [-resample %u [-scene scn2D%u]]", "compute the transfo between two lsets. The transfo may be rigid similarity or affine. If -resample, create a new resampled lset for the first or second lset (-resample 1 or -resample 2). It return the resampling real matrix transforming the first lset into the second and the new lset resampled");
	kind->newOption("-computeError", (TclObject::OPTION) &TclImageLandmarkSet::computeErrorCmd, "lset%u", "compute the errors between each landmark of each set");
	kind->newOption("-nbLandmarks", (TclObject::OPTION) &TclImageLandmarkSet::nbLandmarksCmd, 0, "get the nb of landmark points" );
	kind->newOption("-addLandmark", (TclObject::OPTION) &TclImageLandmarkSet::addLandmarkCmd, "%u %u %u %f %f %f %s", "create a new landmark given its 3 voxel coordinates, its 3 world coordinates and its name" );
	kind->newOption("-removeLandmark", (TclObject::OPTION) &TclImageLandmarkSet::removeLandmarkCmd, "%s", "destroy a landmark given its name" );
	kind->newOption("-scene", (TclObject::OPTION) &TclImageLandmarkSet::sceneCmd, "[scn2D%u]", "set/get the 2D scene where the set is stored" );
	kind->newOption("-displayMode", (TclObject::OPTION) &TclImageLandmarkSet::displayModeCmd, "[voxel|cross]", "set/get the manner to display landmarks in slicecameras : 'voxel' for drawing a full voxel and 'cross' for drawing a cross " );
	kind->newOption("-displayName", (TclObject::OPTION) &TclImageLandmarkSet::displayNameCmd, "[%b]", "set/get the boolean that indicates if the landmark should be written during display" );
	kind->newOption("-crossSize", (TclObject::OPTION) &TclImageLandmarkSet::crossSizeCmd, "[%u]", "set/get the size of crosses for the display of landmarks");
	kind->newOption("-color", (TclObject::OPTION) &TclImageLandmarkSet::colorCmd, "[%f %f %f]", "set/get the color of each landmark during display");
	kind->newOption("-selectedColor", (TclObject::OPTION) &TclImageLandmarkSet::selectedColorCmd, "[%f %f %f]", "set/get the color of the selected landmark during display");
	kind->newOption("-landmarks", (TclObject::OPTION) &TclImageLandmarkSet::landmarksCmd, 0, "get the list of landmarks : for each landmarks returns its voxel position, real position and name");
	kind->newOption("-destroy", (TclObject::OPTION) &TclImageLandmarkSet::destroyCmd, 0, "delete landmark set");
	kind->newOption("-selectedLandmark", (TclObject::OPTION) &TclImageLandmarkSet::selectedLandmarkCmd, "[%s]", "get/set the name of the selected landmark : 'none' is no landmark selected");
	kind->newOption("-saveAs", (TclObject::OPTION) &TclImageLandmarkSet::saveAsCmd, "%s", "save the current landmark set in a file");

}
void SelectLandmarkInstallProc(TclScene2DObject *obj,TclCamera2D *cam)
{
	TclImageLandmarkSet *lset=(TclImageLandmarkSet *) obj;
	lset->getClient()->setSelectionMode(ImageLandmarkSet::SELECT_LANDMARK);
}
void CreateLandmarkInstallProc(TclScene2DObject *obj,TclCamera2D *cam)
{
	TclImageLandmarkSet *lset=(TclImageLandmarkSet *) obj;
	lset->getClient()->setSelectionMode(ImageLandmarkSet::CREATE_LANDMARK);
}
void MoveLandmarkInstallProc(TclScene2DObject *obj,TclCamera2D *cam)
{
	TclImageLandmarkSet *lset=(TclImageLandmarkSet *) obj;
	lset->getClient()->setSelectionMode(ImageLandmarkSet::MOVE_LANDMARK);
}
void DeleteLandmarkInstallProc(TclScene2DObject *obj,TclCamera2D *cam)
{
	TclImageLandmarkSet *lset=(TclImageLandmarkSet *) obj;
	lset->getClient()->setSelectionMode(ImageLandmarkSet::DELETE_LANDMARK);
}
}
