#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/TclInrimage.h"
#include <modules/TclMatrix4D.h>
#include <inrimage/TclInrimage.h>
#include <modules/TclGenericHistogram.h>
#include <modules/TclProbabilityDistribution.h>
#include <inrimage/Interpolator.h>
#include <inrimage/Image3DBackgroundManager.h>
#include <inrimage/fparser.h>
#include <inrimage/threshold.h>
#include <inrimage/arithmetic.h>
#include <inrimage/recursiveFilters.h>
#include <inrimage/anisotropicDiffusion.h>
#include <inrimage/InrChamferMap.h>
#include <inrimage/resample.h>
#include <inrimage/blockProcessing.h>
#include <inrimage/ConnectingElement.h>
#include <inrimage/sliceImage.h>
#include <inrimage/TclStructuringElement.h>
#include "tkwrapper.h"

/* added defined to remove warnings on VC++ 8.0 */
//#define sscanf sscanf_s
//#define strcpy strcpy_s

#ifdef WIN32
/*
#define _CRT_SECURE_CPP_OVERLOAD_STANDARD_NAMES 1
#define _CRT_NONSTDC_NO_DEPRECATE 1
#define _CRT_SECURE_NO_DEPRECATE 
*/
#endif
namespace yav {


char *TclInrimage::types[] = {
  "8bits",
  "s8bits",
  "16bits",
  "s16bits",
  "32bits",
  "s32bits",
  "64bits",
  "s64bits",
  "float",
  "double",
  "rgb",
  "rgba",
  "floatVector",
  "doubleVector"
};


TclInrimage::TclInrimage(TclModule *m, const std::type_info& info, Inrimage *data) 
  : TclObject(info, m), Owner<Inrimage>(data)
{
  done(typeid(TclInrimage));
}


TclInrimage::TclInrimage(const std::type_info& info,
			 TclModule *m, const TclString& _name,
			 bool headerOnly,
			 VECTORIAL_MODE vm, VECTORIAL_MODE fvm)
    throw(Inrimage::ImageNotFound, Inrimage::UnknownImageType,
	  Inrimage::InvalidHeader, Inrimage::InvalidData, 
	  Inrimage::UnknownType)
  : TclObject(info,m), Owner<Inrimage>( new Inrimage(_name, headerOnly, vm, fvm))
{
  done(typeid(TclInrimage));
}


TclInrimage::TclInrimage(const std::type_info& info,
			 TclModule *m,
			 unsigned int xdim,
			 unsigned int ydim,
			 unsigned int zdim,
			 unsigned int vdim,
			 unsigned int wdim,
			 WORD_KIND wordKind,
			 SIGN sign,
			 float vx,
			 float vy,
			 float vz)
  throw(Inrimage::InvalidSize, Inrimage::UnknownType, Inrimage::InvalidVoxelSize)
  : TclObject(info, m), Owner<Inrimage>(new Inrimage(xdim, ydim, zdim, vdim, wdim, wordKind, sign, vx, vy, vz))
{
  done(typeid(TclInrimage));
}

  
TclInrimage::TclInrimage(const std::type_info& info,
			 TclModule *m,
			 unsigned int xdim,
			 unsigned int ydim,
			 unsigned int zdim,
			 Inrimage::WORD_TYPE wt,
			 unsigned int vdim,
			 VECTORIAL_MODE vm,
			 float vx,
			 float vy,
			 float vz)
  throw(Inrimage::InvalidSize, Inrimage::UnknownType, Inrimage::InvalidVoxelSize)
  : TclObject(info,m),
    Owner<Inrimage>(new Inrimage(xdim, ydim, zdim, wt, vdim, vm, vx, vy, vz))
{
  done(typeid(TclInrimage));
}





TclInrimage::TclInrimage(TclModule *m,
			 const TclString& _name,
			 bool headerOnly, 
			 VECTORIAL_MODE vm, VECTORIAL_MODE fvm)
    throw(Inrimage::ImageNotFound, Inrimage::UnknownImageType,
	  Inrimage::InvalidHeader, Inrimage::InvalidData, 
	  Inrimage::UnknownType)
  : TclObject(typeid(TclInrimage),m),
    Owner<Inrimage>(new Inrimage(_name, headerOnly, vm, fvm))
{
  done(typeid(TclInrimage));
}


TclInrimage::TclInrimage(TclModule *m,
			 unsigned int xdim,
			 unsigned int ydim,
			 unsigned int zdim,
			 unsigned int vdim,
			 unsigned int wdim,
			 WORD_KIND wordKind ,
			 SIGN sign,
			 float vx,
			 float vy,
			 float vz)
  throw(Inrimage::InvalidSize, Inrimage::UnknownType, Inrimage::InvalidVoxelSize)
  : TclObject(typeid(TclInrimage), m),
    Owner<Inrimage>(new Inrimage(xdim, ydim, zdim, vdim, wdim, wordKind, sign, vx, vy, vz))
{
  done(typeid(TclInrimage));
}

  
TclInrimage::TclInrimage(TclModule *m,
			 unsigned int xdim,
			 unsigned int ydim,
			 unsigned int zdim,
			 void *data,
			 unsigned int vdim, 
			 unsigned int wdim,
			 WORD_KIND wordKind,
			 SIGN sign,
			 float vx,
			 float vy,
			 float vz)
  throw(Inrimage::InvalidSize, Inrimage::UnknownType, Inrimage::InvalidVoxelSize)
  : TclObject(typeid(TclInrimage),m),
    Owner<Inrimage>( new Inrimage(xdim, ydim, zdim, data, vdim, wdim,
				  wordKind, sign, vx, vy, vz))

{
  done(typeid(TclInrimage));
}


TclInrimage::TclInrimage(TclModule *m,
			 unsigned int xdim,
			 unsigned int ydim,
			 unsigned int zdim,
			 Inrimage::WORD_TYPE wt,
			 unsigned int vdim,
			 VECTORIAL_MODE vm,
			 float vx,
			 float vy,
			 float vz)
  throw(Inrimage::InvalidSize, Inrimage::UnknownType, Inrimage::InvalidVoxelSize)
  : TclObject(typeid(TclInrimage),m),
    Owner<Inrimage>(new Inrimage(xdim, ydim, zdim, wt, vdim, vm, vx, vy, vz))
{
  done(typeid(TclInrimage));
}


/// Creates an replica of img.
TclInrimage::TclInrimage(const TclInrimage& img)
  throw(Inrimage::InvalidSize, Inrimage::InvalidVectorialMode,
	Inrimage::InvalidVoxelSize, Inrimage::NotEnoughMemory)
  : TclObject(typeid(TclInrimage),img.getModule()),
    Owner<Inrimage>(new Inrimage((*img.getClient())))

{
  done(typeid(TclInrimage));
}


// destructor
TclInrimage::~TclInrimage() {
  deleted(typeid(TclInrimage)); 
  if (getClient()) {
    Inrimage *i=getClient();
    pointer=0;
    delete i;
  }
}

void TclInrimage::save(const TclString& newname) {
  if(newname != EMPTY_STRING) {
    getClient()->write((char*)newname);
  }
  else {
    if(getClient()->getName() != EMPTY_STRING)
      getClient()->write((char*)getClient()->getName());
  }
}

void TclInrimage::redraw(void) const {

	std::list<ObjectPointer<Inrimage> *>::iterator itp;
	// first ask all backgrounds to change their internal buffer
	for(itp=getClient()->pointers.begin();itp!=getClient()->pointers.end();itp++)
	{
		Image3DBackgroundManager *b=dynamic_cast<Image3DBackgroundManager *>((*itp)->getOwner());
		if (b)
		{
			b->imageHasChangedTexture(getClient());

		}
	}


}


void TclInrimage::resizeAndRedraw(void) const {
	
	std::list<ObjectPointer<Inrimage> *>::iterator itp;
	// first ask all backgrounds to change their internal buffer
	for(itp=getClient()->pointers.begin();itp!=getClient()->pointers.end();itp++)
	{
		Image3DBackgroundManager *b=dynamic_cast<Image3DBackgroundManager *>((*itp)->getOwner());
		if (b)
		{
			b->imageHasChangedSize(getClient());

		}
			
	}
	/// then redraw
	for(itp=getClient()->pointers.begin();itp!=getClient()->pointers.end();itp++)
	{
		Image3DBackgroundManager *b=dynamic_cast<Image3DBackgroundManager *>((*itp)->getOwner());
		if (b)
		{
			b->imageHasChangedTexture(getClient());

		}
	}
#ifdef OLD_CODE
  getModule()->tickle("set graphics2D::busy 1");
  std::list<ObjectPointer<Inrimage> *>::iterator itp;
  for(itp=getClient()->pointers.begin();itp!=getClient()->pointers.end();itp++)
    {
      InrimageView *inrv=dynamic_cast<InrimageView *>((*itp)->getOwner());
      if (inrv)
	{
	  std::list<InrimageBackground *>::iterator itb=inrv->firstBackground();
	  for(;itb!=inrv->lastBackground();itb++)
	    {
	      (*itb)->setVisibleArea();
	    }
	}
    }
  getModule()->tickle("set graphics2D::busy 0");
#endif
}


void TclInrimage::resizeVoxelsAndRedraw(void) const {
#ifdef OLD_CODE
  std::list<ObjectPointer<Inrimage> *>::iterator itp;
  for(itp=getClient()->pointers.begin();itp!=getClient()->pointers.end();itp++)
    {
      InrimageView *inrv=dynamic_cast<InrimageView *>((*itp)->getOwner());
      if (inrv)
	{
	  std::list<InrimageBackground *>::iterator itb=inrv->firstBackground();
	  for(;itb!=inrv->lastBackground();itb++)
	    {
	      ((TclInrimageBackground *)(*itb))->imageVoxelSizeChanged();
	    }
	}
    }
#endif
}






void TclInrimage::changeAndRedraw(void) 
{
	std::list<ObjectPointer<Inrimage> *>::iterator itp;
	for(itp=getClient()->pointers.begin();itp!=getClient()->pointers.end();itp++)
	{
		Image3DBackgroundManager *b=dynamic_cast<Image3DBackgroundManager *>((*itp)->getOwner());
		if (b)
		{
			b->imageHasChangedTexture(getClient());

		}
		
	}
}



Inrimage::WORD_TYPE TclInrimage::typeFromString(char *stype) {
  if(!strcmp(stype, types[0])) return Inrimage::WT_UNSIGNED_CHAR;
  else if(!strcmp(stype, types[1])) return Inrimage::WT_SIGNED_CHAR;
  else if(!strcmp(stype, types[2])) return Inrimage::WT_UNSIGNED_SHORT;
  else if(!strcmp(stype, types[3])) return Inrimage::WT_SIGNED_SHORT;
  else if(!strcmp(stype, types[4])) return Inrimage::WT_UNSIGNED_INT;
  else if(!strcmp(stype, types[5])) return Inrimage::WT_SIGNED_INT;
  else if(!strcmp(stype, types[6])) return Inrimage::WT_UNSIGNED_LONG;
  else if(!strcmp(stype, types[7])) return Inrimage::WT_SIGNED_LONG;
  else if(!strcmp(stype, types[8])) return Inrimage::WT_FLOAT;
  else if(!strcmp(stype, types[9])) return Inrimage::WT_DOUBLE;
  else if(!strcmp(stype, types[10])) return Inrimage::WT_RGB;
  else if(!strcmp(stype, types[11])) return Inrimage::WT_RGBA;
  else if(!strcmp(stype, types[12])) return Inrimage::WT_FLOAT_VECTOR;
  else if(!strcmp(stype, types[13])) return Inrimage::WT_DOUBLE_VECTOR;
  else return (Inrimage::WORD_TYPE) -1;
}


char *TclInrimage::typeToString(Inrimage::WORD_TYPE wt) {
  switch(wt) {
  case Inrimage::WT_UNSIGNED_CHAR:
    return types[0];
  case Inrimage::WT_SIGNED_CHAR:
    return types[1];
  case Inrimage::WT_UNSIGNED_SHORT:
    return types[2];
  case Inrimage::WT_SIGNED_SHORT:
    return types[3];
  case Inrimage::WT_UNSIGNED_INT:
    return types[4];
  case Inrimage::WT_SIGNED_INT:
    return types[5];
  case Inrimage::WT_UNSIGNED_LONG:
    return types[6];
  case Inrimage::WT_SIGNED_LONG:
    return types[7];
  case Inrimage::WT_FLOAT:
    return types[8];
  case Inrimage::WT_DOUBLE:
    return types[9];
  case Inrimage::WT_RGB:
    return types[10];
  case Inrimage::WT_RGBA:
    return types[11];
  case Inrimage::WT_FLOAT_VECTOR:
    return types[12];
  case Inrimage::WT_DOUBLE_VECTOR:
    return types[13];
  default:
    return 0;
  }
}


CONNECTIVITY_TYPE TclInrimage::connectivityFromString(char *str) {
  if(!strcmp(str, "4")) return FOUR_CONNECTIVITY;
  else if(!strcmp(str, "6")) return SIX_CONNECTIVITY;
  else if(!strcmp(str, "8")) return EIGHT_CONNECTIVITY;
  else if(!strcmp(str, "10")) return TEN_CONNECTIVITY;
  else if(!strcmp(str, "18")) return EIGHTEEN_CONNECTIVITY;
  else if(!strcmp(str, "26")) return TWENTY_SIX_CONNECTIVITY;
  else return (CONNECTIVITY_TYPE) -1;
}



TclInrimage *TclInrimage::load(const TclString& name, TclModule *m,
			       bool interlace, bool fileinterlace) {

  TclInrimage *inr;

  try {
    VECTORIAL_MODE vm = interlace ? VM_INTERLACED : VM_NON_INTERLACED;
    VECTORIAL_MODE fvm = fileinterlace ? VM_INTERLACED : VM_NON_INTERLACED;
    inr = new TclInrimage(m, name, false, vm, fvm);
  }
  catch(Inrimage::ImageNotFound) {
    m->setResult("file not found");
    return 0;
  }
  catch(Inrimage::UnknownImageType) {
    m->setResult("unknown image type (should be INRIMAGE, GIF or IRIS image file format)");
    return 0;
  }
  catch(Inrimage::InvalidHeader) {
    m->setResult("invalid inrimage header");
    return 0;
  }
  catch(Inrimage::InvalidData) {
    m->setResult("invalid inrimage data");
    return 0;
  }
  catch(Inrimage::UnknownType) {
    m->setResult("unhandled inrimage type");
    return 0;
  }

  return inr;
}


int TclInrimage::load(TclModule *m, int argc, char **argv) {
  TclInrimage *inr;
  int i;

  bool interlace = true;
  bool fileinterlace = true;
  for(i = 2; i < argc; i++) {
    if(!strcmp(argv[i], "-interlace")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(!TclModule::boolValue(argv[i], interlace)) return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "-fileInterlace")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(!TclModule::boolValue(argv[i], fileinterlace)) return COMMAND_SYNOPSIS;
    }
    else return COMMAND_SYNOPSIS;
  }

  inr = TclInrimage::load(argv[1], m, interlace, fileinterlace);


  if(!inr) {
    m->setResult("no file %s", argv[1]);
    return TCL_ERROR;
  }
  
  m->setResult("%s%u", TclInrimage::getTCLObjectName(), inr->getRef());
  return TCL_OK;
}


/*  "Create a new inrimage with given size and type. Vectorial images
 dimension is given after floatVector or doubleVector.");*/
int TclInrimage::create(TclModule *m, int argc, char **argv) {

  unsigned int x = 0, y = 0, z = 0, vdim = 1;
  Inrimage::WORD_TYPE wt = Inrimage::WT_UNSIGNED_CHAR;
  TclInrimage *inr;
  int i;

  if(argc < 2) return COMMAND_SYNOPSIS;

  if(strcmp(argv[1], "x") && strcmp(argv[1], "y") && strcmp(argv[1], "z")
     && strcmp(argv[1], "type")) {
    return TclInrimage::load(m, argc, argv);
  }

  for(i = 1; i < argc; i++) {
    if(!strcmp(argv[i], "x")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%u", &x) != 1) return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "y")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%u", &y) != 1) return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "z")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%u", &z) != 1) return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "type")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      wt = TclInrimage::typeFromString(argv[i]);
      if(wt < 0) return COMMAND_SYNOPSIS;
      if(wt == Inrimage::WT_RGB) vdim = 3;
      if(wt == Inrimage::WT_RGBA) vdim = 4;
      if(wt == Inrimage::WT_FLOAT_VECTOR || wt == Inrimage::WT_DOUBLE_VECTOR) {
	if(++i >= argc) return COMMAND_SYNOPSIS;
	if(sscanf(argv[i], "%u", &vdim) != 1) return COMMAND_SYNOPSIS;
      }
    }
    else return COMMAND_SYNOPSIS;
  }

  if(x == 0 || y == 0 || z == 0 || vdim == 0) return COMMAND_SYNOPSIS;

  inr = new TclInrimage(m, x, y, z, wt, vdim, VM_INTERLACED);

  m->setResult("%s%u", inr->getObjectName(), inr->getRef());
  return TCL_OK;
}


int TclInrimage::view(TclModule *m, int argc, char **argv) {
  if ((argc!=2) && (argc!=3))
	  return  COMMAND_SYNOPSIS;
  else {
	  TclInrimage *inr = TclInrimage::getObject(argc, argv, 1);
	  TclScene2D *scn=0;
	  if (!inr)
	  {
		  m->setResult("%s is not an image",argv[1]);
		  return TCL_ERROR;
	  }
	  if (argc==3) {
		  scn = (TclScene2D *) TclObjectKind::getObject(typeid(TclScene2D), argc, argv, 2);   
		  if (!scn) {
			  TclInrimage *inr2 = TclInrimage::getObject(argc, argv, 2);
			  if (inr2) {
				  m->tickle("inrimage::viewImage %s %s; ",argv[1],argv[2]);
				  return TCL_OK;
			  } else {
				  m->setResult("%s must be a 2D scene or an image",argv[2]);
				  return TCL_ERROR;
			  }
		  }
	  }
	  if (inr->getClient()->getGeometry()==Inrimage::CARTESIAN)
	  {
		  if (scn)
			  m->tickle("inrimage::viewImage %s none %s; ",argv[1],argv[2]);
		  else
			  m->tickle("inrimage::viewImage %s; ",argv[1]);
		  return TCL_OK;
	  }
	  else
	  {
		  if (scn)
			  m->tickle("inrimage::singleViewImage %s alpha %s",argv[1],argv[2]);
		  else
			  m->tickle("inrimage::singleViewImage %s alpha",argv[1]);
		  return TCL_OK;
	  }

  }
}



/************************************************************************/
/************************************************************************/
/*********************        TCL COMMANDS      *************************/
/************************************************************************/
/************************************************************************/

/* parse options:
      <org %u %u %u | center %u %u %u> <end %u %u %u| size %u %u %u>
   in command line and returns box bounds */
int TclInrimage::parseBoxOptions(TclModule *m, TclInrimage *inr,
			   int argc, char **argv,
			   unsigned int& x1, unsigned int& y1,
			   unsigned int& z1,
			   unsigned int& x2, unsigned int& y2,
			   unsigned int& z2) {
  int i;
  int center = 0, size = 0;

  /* parse arguments */
  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "org") || !strcmp(argv[i], "center")) {
      if(i+3 < argc) {
	if(!strcmp(argv[i], "center")) center = 1;
 	if(sscanf(argv[++i], "%u", &x1) != 1) return 0;
	if(sscanf(argv[++i], "%u", &y1) != 1) return 0;
	if(sscanf(argv[++i], "%u", &z1) != 1) return 0;
      }
      else goto error;
    }
    else if(!strcmp(argv[i], "end") || !strcmp(argv[i], "size")) {
      if(i+3 < argc) {
	if(!strcmp(argv[i], "size")) size = 1;
	if(sscanf(argv[++i], "%u", &x2) != 1) return 0;
	if(sscanf(argv[++i], "%u", &y2) != 1) return 0;
	if(sscanf(argv[++i], "%u", &z2) != 1) return 0;
      }
      else goto error;
    }
  }

  if(size) {
    x2 /= 2;
    y2 /= 2;
    z2 /= 2;
    if(!x2 || !y2 || !z2) goto error;
  }

  if(center) {
    if(!size) {
      x2 = std::abs((int)(x1 - x2)) / 2;
      y2 = std::abs((int)(y1 - y2)) / 2;
      z2 = std::abs((int)(z1 - z2)) / 2;
    }

    if(x1 > x2) x1 -= x2;
    else x1 = 0;
    if(y1 > y2) y1 -= y2;
    else y1 = 0;
    if(z1 > z2) z1 -= z2;
    else z1 = 0;
    x2 = x1 + 2 * x2;
    y2 = y1 + 2 * y2;
    z2 = z1 + 2 * z2;
  }

  else {
    if(size) {
      x2 += x1 - 1;
      y2 += y1 - 1;
      z2 += z1 - 1;
    }
  }

  if(x2 < x1) { i = x1; x1 = x2; x2 = i; }
  if(y2 < y1) { i = y1; y1 = y2; y2 = i; }
  if(z2 < z1) { i = z1; z1 = z2; z2 = i; }

  if(x1 >= inr->getX() || y1 >= inr->getY() || z1 >= inr->getZ())
    goto error;

  if(x2 >= inr->getX()) x2 = inr->getX() - 1;
  if(y2 >= inr->getY()) y2 = inr->getY() - 1;
  if(z2 >= inr->getZ()) z2 = inr->getZ() - 1;

  return 1;

 error:
  m->setResult("invalid region");
  return 0;
}



int TclInrimage::destroyCmd(int argc, char **argv) {
  unsigned int ref;
  TclModule *m;

  if(argc > 0) return COMMAND_SYNOPSIS;
  ref = getRef();
  m = getModule();
  delete this;
  m->setResult("%s%u does not exist any longer",
	       TclInrimage::getTCLObjectName(), ref);
  return TCL_OK;
}


int TclInrimage::nameCmd(int argc, char **argv) {
  if(argc > 1) return COMMAND_SYNOPSIS;
  if(argc == 1) getClient()->setName(argv[0]);
  setResult("%s", (char *) getClient()->getName());
  return TCL_OK;
}


int TclInrimage::sizeCmd(int argc, char **argv) {
  if(argc > 0) return COMMAND_SYNOPSIS;
  if(Inrimage::vectorial(getType()))
    setResult("%u %u %u %u", getX(), getY(),
	      getZ(), getVdim());
  else
    setResult("%u %u %u", getX(), getY(),
	      getZ());
  return TCL_OK;
}


int TclInrimage::saveCmd(int argc, char **argv) {
  if(argc > 1) return COMMAND_SYNOPSIS;
  if(argc == 0) save();
  else save((TclString) argv[0]);
  setResult("%s%u saved as %s", getObjectName(), getRef(),
	    (char *) getClient()->getName());
  return TCL_OK;
}


int TclInrimage::copyCmd(int argc, char **argv) {
  TclInrimage *res;
  if(argc > 0) return COMMAND_SYNOPSIS;
  res = new TclInrimage(*this);
  setResult("%s%u", getObjectName(), res->getRef());
  return TCL_OK;
}


int TclInrimage::redrawCmd(int argc, char **argv) {
  if(argc > 0) return COMMAND_SYNOPSIS;
  redraw();
  setResult("%s%u redrawn", getObjectName(),
	    getRef());
  return TCL_OK;
}




int TclInrimage::vectorialCmd(int argc, char **argv) {
  if(argc > 0) return COMMAND_SYNOPSIS;
  setResult(TclModule::boolString(Inrimage::vectorial(getType())));
  return TCL_OK;
}


int TclInrimage::floatingCmd(int argc, char **argv) {
  if(argc > 0) return COMMAND_SYNOPSIS;
  setResult(TclModule::boolString(Inrimage::floating(getType())));
  return TCL_OK;
}



int TclInrimage::glboundsCmd(int argc, char **argv) {
  Voxel *min = 0, *max = 0;
  TclString smin, smax;
  int i;

  if(argc != 4 && argc != 2 && argc != 0) goto error;

  min = Voxel::voxelByType(getType());
  max = Voxel::voxelByType(getType());
  getClient()->getBounds(min, max);

  if(argc != 0) {
    for(i = 0; i < argc; i++) {
      if(!strcmp(argv[i], "min")) {
	if(++i >= argc) goto error;
	try {
	  min->scan(argv[i]);
	}
	catch(Voxel::InvalidValue) {
	  goto error;
	}
      }
      else if(!strcmp(argv[i], "max")) {
	if(++i >= argc) goto error;
	try {
	  max->scan(argv[i]);
	}
	catch(Voxel::InvalidValue) {
	  goto error;
	}
      }
      else return TCL_ERROR;
    }
    getClient()->setBounds(min, max);
    changeAndRedraw();
  }

  min->print(smin);
  max->print(smax);
  setResult("%s %s", (char *) smin, (char *) smax);
  return TCL_OK;

 error:
  if(min) delete min;
  if(max) delete max;
  return COMMAND_SYNOPSIS;
}
int TclInrimage::probabilityMapCmd(int argc, char **argv) {
	if (argc!=1)  return COMMAND_SYNOPSIS;
	TclProbabilityDistribution *pdf=(TclProbabilityDistribution *)
		TclObjectKind::getObject(typeid(TclProbabilityDistribution),
		argc, argv, 0);
	if (pdf==0) {
		setResult("%s is not a probability density function",argv[0]);
		return TCL_ERROR;
	}
	else {
		Inrimage *output=createProbabilityImage(getClient(),pdf->getProbabilityFunction());
		TclInrimage *out=new TclInrimage(getModule(),typeid(TclInrimage),output);
		setResult(out->name());
		return TCL_OK;
	}
}
int TclInrimage::createSlicePhotoCmd(int argc, char **argv) {
  if ((argc==8) || (argc==10)) {
    SLICE_ORIENTATION orientation;
    if (!strcmp(argv[0],"X")) 
      orientation=SLICE_X;
    else if (!strcmp(argv[0],"Y")) 
      orientation=SLICE_Y;
    else if (!strcmp(argv[0],"Z")) 
      orientation=SLICE_Z;
    else if (!strcmp(argv[0],"YFY")) 
      orientation=SLICE_Y_FLIPY;
    else if (!strcmp(argv[0],"YFX")) 
      orientation=SLICE_Y_FLIPX;
    else if (!strcmp(argv[0],"XR")) 
      orientation=SLICE_X_ROTATED;
    else {
      setResult("unknown orientation %s",argv[0]);
      return TCL_ERROR;
    }
    unsigned char **colormap=0;

    unsigned int sliceNumber=(unsigned int)atoi(argv[1]); 
    double scaleFactor=(double) atof(argv[2]);
    bool useVoxelSize;
    if(!TclModule::boolValue(argv[3], useVoxelSize)) return COMMAND_SYNOPSIS;
    double  minWindow=(double) atof(argv[4]);
    double  maxWindow=(double) atof(argv[5]);
    bool interpolate;
    if(!TclModule::boolValue(argv[6], interpolate)) return COMMAND_SYNOPSIS;

    if (!strcmp(argv[7],"spect")) {
      colormap= new (unsigned char *[3]);
      colormap[0]=Image3DBackgroundManager::SPECTColorMapRed;
      colormap[1]=Image3DBackgroundManager::SPECTColorMapGreen;
      colormap[2]=Image3DBackgroundManager::SPECTColorMapBlue;
    }
    else if (!strcmp(argv[7],"affine")) {
    }
    else if (!strcmp(argv[7],"pet")) {
      colormap= new (unsigned char *[3]);
      colormap[0]=Image3DBackgroundManager::PETColorMapRed;
      colormap[1]=Image3DBackgroundManager::PETColorMapGreen;
      colormap[2]=Image3DBackgroundManager::PETColorMapBlue;
    }
    else if (!strcmp(argv[7],"threshold")) {
      colormap= new (unsigned char *[3]);
      colormap[0]=Image3DBackgroundManager::ThresholdColorMap;
      colormap[1]=Image3DBackgroundManager::ThresholdColorMap;
      colormap[2]=Image3DBackgroundManager::ThresholdColorMap;
    }
    else if (!strcmp(argv[7],"hotiron")) {
      colormap= new (unsigned char *[3]);
      colormap[0]=Image3DBackgroundManager::HotIronColorMapRed;
      colormap[1]=Image3DBackgroundManager::HotIronColorMapGreen;
      colormap[2]=Image3DBackgroundManager::HotIronColorMapBlue;
    }
    else if (!strcmp(argv[7],"jet")) {
      colormap= new (unsigned char *[3]);
      colormap[0]=Image3DBackgroundManager::JetColorMapRed;
      colormap[1]=Image3DBackgroundManager::JetColorMapGreen;
      colormap[2]=Image3DBackgroundManager::JetColorMapBlue;
    }
    else {
      setResult("%s must be either jet,spect, pet, hotiron or threshold",argv[7]);
      return TCL_ERROR;
    }

    Inrimage *slice=createSliceImage(getClient(),orientation,
				     sliceNumber,scaleFactor,useVoxelSize,
				     minWindow,maxWindow,interpolate,colormap);
    
    if (colormap)
      delete [] colormap;
    
    TclString  TKimageID;
    if (argc==10) {
      if (!strcmp(argv[8],"-image")){
	TKimageID=argv[9];
      }
      else 
	return COMMAND_SYNOPSIS;
    } else {
      getModule()->tickle("image create photo -width %u -height %u",
			  slice->getX(),slice->getY());
      TKimageID = getModule()->getInterp()->result;
    }


    Tk_PhotoHandle handle=Tk_FindPhoto(getModule()->getInterp(),(char *)TKimageID);
    if (argc==10) {
      
#if TK_MAJOR_VERSION==8 && TK_MINOR_VERSION<=4
      Tk_PhotoSetSize(handle, slice->getX(), slice->getY());
#else
      Tk_PhotoSetSize(getModule()->getInterp(), handle, slice->getX(), slice->getY());
#endif
    }
    if (handle==0)
    {
      setResult("Could not find any image named %s",(char *)TKimageID);
      return TCL_ERROR;
    }
    Tk_PhotoImageBlock *block = new Tk_PhotoImageBlock;
    block->pixelSize = 3;
    block->offset[0] = 0;
    block->offset[1] = 1;
    block->offset[2] = 2;
    block->pitch = 3 * slice->getX();
    block->pixelPtr = 0;
    block->width = slice->getX();
    block->height = slice->getY();
    block->pixelPtr = new unsigned char[slice->getX() * slice->getY() * 3];
    unsigned long size=slice->getX() * slice->getY();
    unsigned char *ptr=(unsigned char *)slice->getData();
    unsigned char *out;
    unsigned int i;
    if (!colormap) {
      for (out=block->pixelPtr,i=0;i<size;i++,ptr++,out+=3) {
	out[0]=*ptr;
	out[1]=*ptr;
	out[2]=*ptr;
      }
    } else {
      for (out=block->pixelPtr,i=0;i<size;i++,ptr+=3,out+=3) {
	out[0]=ptr[0];
	out[1]=ptr[1];
	out[2]=ptr[2];
      }
    }
#if TK_MAJOR_VERSION==8 && TK_MINOR_VERSION<=4     
    TkWrapper_PhotoPutBlock(handle, block, 0, 0, block->width, block->height);
#else
    TkWrapper_PhotoPutBlock(getModule()->getInterp(), handle, block, 0, 0, block->width, block->height);
#endif
    delete block->pixelPtr;
    delete block;
    delete slice;
    setResult(TKimageID);
    return TCL_OK;
  }
  else
    return COMMAND_SYNOPSIS;
}
  
int TclInrimage::isolateRegionCmd(int argc, char **argv) {
  unsigned int x1, y1, z1, x2, y2, z2;
  int i;

  if(argc != 8) return COMMAND_SYNOPSIS;

  if(!parseBoxOptions(getModule(), this, argc, argv,
		      x1, y1, z1, x2, y2, z2))
    return TCL_ERROR;

  // parse arguments
  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "org") || !strcmp(argv[i], "center")
       || !strcmp(argv[i], "end") || !strcmp(argv[i], "size")) i = i + 3;
    else return COMMAND_SYNOPSIS;
  }

  getClient()->isolateRegion(x1, y1, z1, x2, y2, z2);
  redraw();
  setResult("(%u, %u, %u) - (%u %u %u)", x1, y1, z1, x2, y2, z2);
  return TCL_OK;
}

int TclInrimage::zeroRegionCmd(int argc, char **argv) {
  unsigned int x1, y1, z1, x2, y2, z2;
  int i;

  if(argc != 8) return COMMAND_SYNOPSIS;

  if(!parseBoxOptions(getModule(), this, argc, argv,
		      x1, y1, z1, x2, y2, z2))
    return TCL_ERROR;

  // parse arguments
  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "org") || !strcmp(argv[i], "center")
       || !strcmp(argv[i], "end") || !strcmp(argv[i], "size")) i = i + 3;
    else return COMMAND_SYNOPSIS;
  }

  getClient()->zeroRegion(x1, y1, z1, x2, y2, z2);
  redraw();
  setResult("(%u, %u, %u) - (%u %u %u)", x1, y1, z1, x2, y2, z2);
  return TCL_OK;
}

int TclInrimage::globalTransformationToVectorCmd(int argc, char **argv) {
  if (argc==16) {
    // apply a rotation and translation to vectorial data through a 4x4 matrix
    Vec3<double> trans=Vec3<double>(atof(argv[3]),atof(argv[7]),atof(argv[11]));
    Matrix3x3<double> r=Matrix3x3<double>(atof(argv[0]),atof(argv[1]),atof(argv[2]),atof(argv[4]),atof(argv[5]),atof(argv[6]),atof(argv[8]),atof(argv[9]),atof(argv[10]));
    mul(getClient(),r);
    add(getClient(),trans);
    changeAndRedraw();
    return TCL_OK;
  }
  else 
    return COMMAND_SYNOPSIS;
}

int TclInrimage::globalTransformationCmd(int argc, char **argv) {
  if (argc==0) {
    Matrix4x4<double> mat=getClient()->getTransformationMatrix();
    setResult("%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f",
	      mat[0][0], mat[0][1], mat[0][2], mat[0][3],
	      mat[1][0], mat[1][1], mat[1][2], mat[1][3],
	      mat[2][0], mat[2][1], mat[2][2], mat[2][3],
	      mat[3][0], mat[3][1], mat[3][2], mat[3][3]);
    
    return TCL_OK;
  }
  else if (argc==1) {
    if (!strcmp(argv[0],"translation")) {
      Vec3<double> p=getClient()->getTranslation();
      setResult("%f %f %f",p[0],p[1],p[2]);
      return TCL_OK;
    }
    else if (!strcmp(argv[0],"rotation")) {
      Rotation3D r=getClient()->getRotationMatrix();
      setResult("%f %f %f %f %f %f %f %f %f",
		r[0][0],r[0][1],r[0][2],
		r[1][0],r[1][1],r[1][2],
		r[2][0],r[2][1],r[2][2]);	
      return TCL_OK;
    }
    else
      return COMMAND_SYNOPSIS;
  }
  else if (argc==2) {
    IFSTREAM input(argv[1]);
    input.setf(IOS::scientific,IOS::floatfield);
    TclString line;
    Matrix4x4<double> tab;
    unsigned int i=0;
    for(;i<4;++i) {
      input >> tab(i,0) >> tab(i,1) >> tab(i,2) >> tab(i,3);
    }
    Vec3<double> t = Vec3<double>(tab(0,3),tab(1,3),tab(2,3));
    Rotation3D r=Rotation3D(Vec3<double>(tab(0,0),tab(1,0),tab(2,0)),Vec3<double>(tab(0,1),tab(1,1),tab(2,1)),Vec3<double>(tab(0,2),tab(1,2),tab(2,2)));
    if (!strcmp(argv[0],"set")) {
      getClient()->setTranslation(t);
      getClient()->setRotation(r);
    }
    else if (!strcmp(argv[0],"compose")) {
      getClient()->addTranslation(t);
      getClient()->composeRotation(r);
    }
    else
      return COMMAND_SYNOPSIS;
    Matrix4x4<double> mat=getClient()->getTransformationMatrix();
    setResult("%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f",
	      mat[0][0], mat[0][1], mat[0][2], mat[0][3],
	      mat[1][0], mat[1][1], mat[1][2], mat[1][3],
	      mat[2][0], mat[2][1], mat[2][2], mat[2][3],
	      mat[3][0], mat[3][1], mat[3][2], mat[3][3]);
    return TCL_OK;
  }
  else if (argc==4) {
    if (!strcmp(argv[0],"addTranslation")) {
      Vec3<double> p=Vec3<double>(atof(argv[1]),atof(argv[2]),atof(argv[3]));
      getClient()->addTranslation(p);
      Matrix4x4<double> mat=getClient()->getTransformationMatrix();
      setResult("%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f",
		mat[0][0], mat[0][1], mat[0][2], mat[0][3],
		mat[1][0], mat[1][1], mat[1][2], mat[1][3],
		mat[2][0], mat[2][1], mat[2][2], mat[2][3],
		mat[3][0], mat[3][1], mat[3][2], mat[3][3]);
      return TCL_OK;
    }
    else if (!strcmp(argv[0],"translation")) {
      Vec3<double> p=Vec3<double>(atof(argv[1]),atof(argv[2]),atof(argv[3]));
      getClient()->setTranslation(p);
      Matrix4x4<double> mat=getClient()->getTransformationMatrix();
      setResult("%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f",
		mat[0][0], mat[0][1], mat[0][2], mat[0][3],
		mat[1][0], mat[1][1], mat[1][2], mat[1][3],
		mat[2][0], mat[2][1], mat[2][2], mat[2][3],
		mat[3][0], mat[3][1], mat[3][2], mat[3][3]);
      return TCL_OK;
    }
    else if (!strcmp(argv[0],"composeRotation")) {
      Vec3<double> p=Vec3<double>(atof(argv[1]),atof(argv[2]),atof(argv[3]));
      Rotation3D r=Rotation3D(p);
      getClient()->composeRotation(r);
      Matrix4x4<double> mat=getClient()->getTransformationMatrix();
      setResult("%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f",
		mat[0][0], mat[0][1], mat[0][2], mat[0][3],
		mat[1][0], mat[1][1], mat[1][2], mat[1][3],
		mat[2][0], mat[2][1], mat[2][2], mat[2][3],
		mat[3][0], mat[3][1], mat[3][2], mat[3][3]);
      return TCL_OK;
    }
    else if (!strcmp(argv[0],"rotation")) {
	Vec3<double> p=Vec3<double>(atof(argv[1]),atof(argv[2]),atof(argv[3]));
	Rotation3D r=Rotation3D(p);
	getClient()->setRotation(r);
	Matrix4x4<double> mat=getClient()->getTransformationMatrix();
	setResult("%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f",
		  mat[0][0], mat[0][1], mat[0][2], mat[0][3],
		  mat[1][0], mat[1][1], mat[1][2], mat[1][3],
		  mat[2][0], mat[2][1], mat[2][2], mat[2][3],
		  mat[3][0], mat[3][1], mat[3][2], mat[3][3]);
	return TCL_OK;
    }
    else 
      return COMMAND_SYNOPSIS;
  }
  else if (argc==17) {
    if (!strcmp(argv[0],"set")) {
      // set the rotation and the translation through a 4x4 matrix
      Vec3<double> trans=Vec3<double>(atof(argv[4]),atof(argv[8]),atof(argv[12]));
      Rotation3D r=Rotation3D(Vec3<double>(atof(argv[1]),atof(argv[5]),atof(argv[9])),Vec3<double>(atof(argv[2]),atof(argv[6]),atof(argv[10])),Vec3<double>(atof(argv[3]),atof(argv[7]),atof(argv[11])));
      getClient()->setTranslation(trans);
      getClient()->setRotation(r);
      Matrix4x4<double> mat=getClient()->getTransformationMatrix();
      setResult("%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f",
		mat[0][0], mat[0][1], mat[0][2], mat[0][3],
		mat[1][0], mat[1][1], mat[1][2], mat[1][3],
		mat[2][0], mat[2][1], mat[2][2], mat[2][3],
		mat[3][0], mat[3][1], mat[3][2], mat[3][3]);
      return TCL_OK;
    }
    else if (!strcmp(argv[0],"compose")) {
      // compose the rotation and the translation through a 4x4 matrix
      Vec3<double> trans=Vec3<double>(atof(argv[4]),atof(argv[8]),atof(argv[12]));
      Rotation3D r=Rotation3D(Vec3<double>(atof(argv[1]),atof(argv[5]),atof(argv[9])),Vec3<double>(atof(argv[2]),atof(argv[6]),atof(argv[10])),Vec3<double>(atof(argv[3]),atof(argv[7]),atof(argv[11])));
      getClient()->addTranslation(trans);
      getClient()->composeRotation(r);
      Matrix4x4<double> mat=getClient()->getTransformationMatrix();
      setResult("%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f",
		mat[0][0], mat[0][1], mat[0][2], mat[0][3],
		mat[1][0], mat[1][1], mat[1][2], mat[1][3],
		mat[2][0], mat[2][1], mat[2][2], mat[2][3],
		mat[3][0], mat[3][1], mat[3][2], mat[3][3]);
      return TCL_OK;
    }
    else 
      return COMMAND_SYNOPSIS;
  }
  else 
    return COMMAND_SYNOPSIS;
}

int TclInrimage::cropCmd(int argc, char **argv) {
  unsigned int x1, y1, z1, x2, y2, z2;
  int i;

  if(argc != 8) return COMMAND_SYNOPSIS;

  if(!parseBoxOptions(getModule(), this, argc, argv,
		      x1, y1, z1, x2, y2, z2))
    return TCL_ERROR;

  // parse arguments
  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "org") || !strcmp(argv[i], "center")
       || !strcmp(argv[i], "end") || !strcmp(argv[i], "size")) i = i + 3;
    else return COMMAND_SYNOPSIS;
  }

  getClient()->crop(x1, y1, z1, x2, y2, z2);
  resizeAndRedraw();
  setResult("(%u, %u, %u) - (%u %u %u)", x1, y1, z1, x2, y2, z2);
  return TCL_OK;
}

int TclInrimage::voxelInterpolatedValueCmd(int argc, char **argv) {
  if(argc != 3) return COMMAND_SYNOPSIS;
  double x,y,z;
  if(sscanf(argv[0], "%lf", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%lf", &y) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[2], "%lf", &z) != 1) return COMMAND_SYNOPSIS;

  TclString output;
  Voxel *v = Voxel::voxelByType(getClient()->getType());
  getClient()->getRealVoxel(x,y,z,*v);
  v->print(output);
  delete v;
  setResult(output);
  return TCL_OK;
 
}

int TclInrimage::voxelValueCmd(int argc, char **argv) {
  if(argc != 3 && argc != 4) return COMMAND_SYNOPSIS;
  unsigned int x,y,z;
  if(sscanf(argv[0], "%u", &x) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%u", &y) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[2], "%u", &z) != 1) return COMMAND_SYNOPSIS;
  if(argc == 3) {
    TclString output;
	Voxel *v = Voxel::voxelByType(getClient()->getType());
    getClient()->getVoxel(x,y,z,*v);
	v->print(output);
	delete v;
    setResult(output);
    return TCL_OK;
  }
  else if(argc == 4) {
    double d;
    TclString output;
    if(sscanf(argv[3], "%lf", &d) != 1) return COMMAND_SYNOPSIS;
    Voxel *v = Voxel::voxelByType(getClient()->getType());
    v->setValueFromDouble(d);
    getClient()->setVoxel(x,y,z,*v);
	getClient()->getVoxel(x,y,z,*v);
	v->print(output);
	delete v;
    setResult(output);
    return TCL_OK;
  }
  return 0;
}

int TclInrimage::voxelCmd(int argc, char **argv) {
  float vx, vy, vz;
  
  if(argc > 3) return COMMAND_SYNOPSIS;

  if(argc > 0) {
    if(argc != 3) return COMMAND_SYNOPSIS;
    if(sscanf(argv[0], "%f", &vx) != 1) return COMMAND_SYNOPSIS;
    if(sscanf(argv[1], "%f", &vy) != 1) return COMMAND_SYNOPSIS;
    if(sscanf(argv[2], "%f", &vz) != 1) return COMMAND_SYNOPSIS;

    try {
      getClient()->setVoxelSize(vx, vy, vz);
    }
    catch(BaseInrimage::InvalidVoxelSize&) {
      setResult("voxel size should be > 0");
      return TCL_ERROR;      
    }
    resizeVoxelsAndRedraw();
  }

  setResult("%f %f %f", getVX(), getVY(), getVZ());
  return TCL_OK;
}



int TclInrimage::flipCmd(int argc, char **argv) {
  SLICE_ORIENTATION flipgeom = SLICE_Z;

  if(argc > 1) return COMMAND_SYNOPSIS;
  if(argc == 1) {
    if(!strcmp(argv[0], "X")) flipgeom = SLICE_X;
    else if(!strcmp(argv[0], "Y")) flipgeom = SLICE_Y;
    else if(!strcmp(argv[0], "Z")) flipgeom = SLICE_Z;
    else if(!strcmp(argv[0], "alpha")) flipgeom = SLICE_Z;
    else return COMMAND_SYNOPSIS;
  }

  getClient()->flip(flipgeom);
  redraw();
  switch(flipgeom) {
  case SLICE_X:
    setResult("X");
    break;
  case SLICE_Y:
    setResult("Y");
    break;
  case SLICE_Z:
  case SLICE_ALPHA:
    setResult("Z");
    break;
  default:
    break;
  }
  return TCL_OK;
}


int TclInrimage::switchCmd(int argc, char **argv) {

  if(argc != 1) return COMMAND_SYNOPSIS;
  if(strlen(argv[0]) != 2) return COMMAND_SYNOPSIS;
  SLICE_ORIENTATION geom1, geom2;
  switch(argv[0][0]) {
  case 'X':
  case 'x':
    geom1 = SLICE_X;
    break;
  case 'Y':
  case 'y':
    geom1 = SLICE_Y;
    break;
  case 'Z':
  case 'z':
    geom1 = SLICE_Z;
    break;
  default:
    return COMMAND_SYNOPSIS;
  }
  switch(argv[0][1]) {
  case 'X':
  case 'x':
    geom2 = SLICE_X;
    break;
  case 'Y':
  case 'y':
    geom2 = SLICE_Y;
    break;
  case 'Z':
  case 'z':
    geom2 = SLICE_Z;
    break;
  default:
    return COMMAND_SYNOPSIS;
  }

  getClient()->switchGeometry(geom1, geom2);
  resizeAndRedraw();
  setResult("%c <-> %c", argv[0][0], argv[0][1]);
  return TCL_OK;
}


int TclInrimage::melCmd(int argc, char **argv) {
  unsigned int ixo = 0, iyo = 0, izo = 0, idx, idy, idz;
  int ixi = 0, iyi = 0, izi = 0, i;
  TclInrimage *src;

  if(argc < 1) return COMMAND_SYNOPSIS;
  src = TclInrimage::getObject(argc, argv, 0);
  if(!src) return TCL_ERROR;

  if(src->getType() != getType()) {
    setResult("%s%i and %s%i size differ", 
		 getObjectName(), getRef(),
		 getObjectName(), src->getRef());
    return TCL_ERROR;
  }

  idx = src->getX();
  idy = src->getY();
  idz = src->getZ();

  for(i = 1; i < argc; i++) {
    if(!strcmp(argv[i], "ixo")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%u", &ixo) != 1) return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "iyo")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%u", &iyo) != 1) return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "izo")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%u", &izo) != 1) return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "ixi")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%i", &ixi) != 1) return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "iyi")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%i", &iyi) != 1) return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "izi")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%i", &izi) != 1) return COMMAND_SYNOPSIS;
	}
    else if(!strcmp(argv[i], "idx")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%u", &idx) != 1) return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "idy")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%u", &idy) != 1) return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "idz")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%u", &idz) != 1) return COMMAND_SYNOPSIS;
    }
    else return COMMAND_SYNOPSIS;
  }

  getClient()->mel(src->getClient(), ixo, iyo, izo, ixi, iyi, izi,
		   idx, idy, idz);
  changeAndRedraw();
  setResult("%s%u", getObjectName(), getRef());
  return TCL_OK;
}

int TclInrimage::typeCmd(int argc, char **argv) {
  bool scale = true;
  Inrimage::WORD_TYPE type = (Inrimage::WORD_TYPE) -1;
  int i;
  
  if(argc > 2) return COMMAND_SYNOPSIS;

  if(argc > 0) {
    for(i = 0; i < argc; i++) {
      if(!strcmp(argv[i], "cast")) scale = false;
      else if(!strcmp(argv[i], "scale")) scale = true;
      else {
	if((type = TclInrimage::typeFromString(argv[i])) < 0)
	  return COMMAND_SYNOPSIS;
      }
    }

    if(type < 0) return COMMAND_SYNOPSIS;

    if(scale) getClient()->convertScale(type);
    else getClient()->convertCast(type);
    changeAndRedraw();
  }

  setResult(TclInrimage::typeToString(getType()));
  return TCL_OK;
}

int TclInrimage::line3DCmd(int argc, char **argv) {
  
	if(argc != 7) return COMMAND_SYNOPSIS;
	Vec3<double> p1=Vec3<double>((double) atof(argv[0]),(double) atof(argv[1]),
		(double) atof(argv[2]));
	Vec3<double> p2=Vec3<double>((double) atof(argv[3]),(double) atof(argv[4]),
		(double) atof(argv[5]));
	Voxel *fg = Voxel::voxelByType(getType());
	fg->setValueFromDouble((double) atof(argv[6]));
	getClient()->line3D(p1,p2,*fg);
   redraw();
  return TCL_OK;
}

int TclInrimage::arithCmd(int argc, char **argv) {
  enum {
    NONE,
    ADD,
    ADDCST,
    SUB,
    SUBCST,
    MUL,
    MULCST,
    DIV,
    DIVCST,
    INV,
    ABS,
    SETCST
  } operation = NONE;
  TclInrimage *op=0;
  double val;
  int n = 0;

  if(argc < 1) return COMMAND_SYNOPSIS;

  if(!strcmp(argv[0], "add")) {
    operation = ADD;
    n = 2;
  }
  else if(!strcmp(argv[0], "set")) {
    operation = SETCST;
    n = 2;
  }
  else if(!strcmp(argv[0], "sub")) {
    operation = SUB;
    n = 2;
  }
  else if(!strcmp(argv[0], "mul")) {
    operation = MUL;
    n = 2;
  }
  else if(!strcmp(argv[0], "div")) {
    operation = DIV;
    n = 2;
  }
  else if(!strcmp(argv[0], "inv")) {
    operation = INV;
    n = 1;
  }
  else if(!strcmp(argv[0], "abs")) {
    operation = ABS;
    n = 1;
  }
  else return COMMAND_SYNOPSIS;

  if(n == 1) {
    if(argc != 1) return COMMAND_SYNOPSIS; 
    switch(operation) {
    case INV:
      inv(getClient());
      changeAndRedraw();
      setResult("%s%i = -%s%i", getObjectName(), getRef(),
		getObjectName(), getRef());
      break;
    case ABS:
      abs(getClient());
      changeAndRedraw();
      setResult("%s%i = |%s%i|", getObjectName(), getRef(),
		getObjectName(), getRef());
      break;
    default:
      break;
    }
    return TCL_OK;
  }

  if(argc != 2) return COMMAND_SYNOPSIS; 
  // operate with constant
  if(sscanf(argv[1], "%lf", &val) == 1) {
    switch(operation) {
    case SETCST:
      break;
    case ADD:
      operation = ADDCST;
      break;
    case SUB:
      operation = SUBCST;
      break;
    case MUL:
      operation = MULCST;
      break;
    case DIV:
      operation = DIVCST;
      break;
    default:
      return COMMAND_SYNOPSIS;
    }
  }

  // operate with another inrimage
  else {
    op = TclInrimage::getObject(argc, argv, 1);
    if(!op) return TCL_ERROR;

    if(op->getX() != getX() || op->getY() != getY() ||
       op->getZ() != getZ()) {
      setResult("%s%i and %s%i size differ", 
		   getObjectName(), getRef(),
		   getObjectName(), op->getRef());
      return TCL_ERROR;
    } 
    if(op->getType() != getType()) {
      setResult("%s%i and %s%i type differ", 
		   getObjectName(), getRef(),
		   getObjectName(), op->getRef());
      return TCL_ERROR;
    }
  }

  switch(operation) {
  case ADD:
    add(getClient(), op->getClient());
    changeAndRedraw();
    setResult("%s%i += %s%i", getObjectName(), getRef(),
	      getObjectName(), op->getRef());
    break;
  case SUB:
    sub(getClient(), op->getClient());
    changeAndRedraw();
    setResult("%s%i -= %s%i", getObjectName(), getRef(),
	      getObjectName(), op->getRef());
    break;
  case MUL:
    mul(getClient(), op->getClient());
    changeAndRedraw();
    setResult("%s%i *= %s%i", getObjectName(), getRef(),
	      getObjectName(), op->getRef());
    break;
  case DIV:
    div(getClient(), op->getClient());
    changeAndRedraw();
    setResult("%s%i /= %s%i", getObjectName(), getRef(),
	      getObjectName(), op->getRef());
    break;
  case SETCST:
    setValue(getClient(), val);
    changeAndRedraw();
    setResult("%s%i = %f", getObjectName(), getRef(),
		 val);
    break;
  case ADDCST:
    add(getClient(), val);
    changeAndRedraw();
    setResult("%s%i += %f", getObjectName(), getRef(),
		 val);
    break;
  case SUBCST:
    sub(getClient(), val);
    changeAndRedraw();
    setResult("%s%i -= %f", getObjectName(), getRef(),
	      val);
    break;
  case MULCST:
    mul(getClient(), val);
    changeAndRedraw();
    setResult("%s%i *= %f", getObjectName(), getRef(),
	      val);
    break;
  case DIVCST:
    if(val != 0.0) {
      div(getClient(), val);
      changeAndRedraw();
      setResult("%s%i /= %f", getObjectName(), getRef(),
		val);
    }
    else {
      setResult("0 divide error");
      return TCL_ERROR;
    }
    break;
  default:
    break;
  }
  return TCL_OK;
}

int TclInrimage::blockMeanCmd(int argc, char **argv) {
	if ((argc < 1) && (argc>2)) return COMMAND_SYNOPSIS;
	unsigned int sxy,sz;
	sxy=(unsigned int) atoi(argv[0]);
	if (argc==2) {
		sz=(unsigned int) atoi(argv[1]);
	} else
		sz=sxy;

	Inrimage *out=computeBlockMean(getClient(),sxy,sz);
    TclInrimage *tclOut = new TclInrimage(getModule(), typeid(TclInrimage), out);
	setResult(tclOut->name());
	return TCL_OK;
}
int TclInrimage::blockStandardDeviationCmd(int argc, char **argv) {
	if ((argc < 1) && (argc>2)) return COMMAND_SYNOPSIS;
	unsigned int sxy,sz;
	sxy=(unsigned int) atoi(argv[0]);
	if (argc==2) {
		sz=(unsigned int) atoi(argv[1]);
	} else
		sz=sxy;

	Inrimage *out=computeBlockStandardDeviation(getClient(),sxy,sz);
    TclInrimage *tclOut = new TclInrimage(getModule(), typeid(TclInrimage), out);
	setResult(tclOut->name());
	return TCL_OK;
}
int TclInrimage::blockMedianCmd(int argc, char **argv) {
	if ((argc < 1) && (argc>2)) return COMMAND_SYNOPSIS;
	unsigned int sxy,sz;
	sxy=(unsigned int) atoi(argv[0]);
	if (argc==2) {
		sz=(unsigned int) atoi(argv[1]);
	} else
		sz=sxy;

	Inrimage *out=computeBlockMedian(getClient(),sxy,sz);
    TclInrimage *tclOut = new TclInrimage(getModule(), typeid(TclInrimage), out);
	setResult(tclOut->name());
	return TCL_OK;
}
int TclInrimage::maskCorrelationCmd(int argc, char **argv) {
	if (argc != 2)  return COMMAND_SYNOPSIS;

	TclInrimage *mask = TclInrimage::getObject(argc, argv, 0);
	if(!mask) return TCL_ERROR;

	if(mask->getType() != getType()) {
		setResult("%s%i and %s%i type differ", 
			getObjectName(), getRef(),
			getObjectName(), mask->getRef());
		return TCL_ERROR;
	}
	enum {
    NONE,
    SSD,
    SAD,
    CR
  } correlationMethod;

	if(!strcmp(argv[1], "cr")) correlationMethod = CR;
	else if(!strcmp(argv[1], "ssd")) correlationMethod = SSD;
	else if(!strcmp(argv[1], "sad")) correlationMethod = SAD;
	else return COMMAND_SYNOPSIS;

	Inrimage *out;
	switch(correlationMethod) {
		case SSD:
			out=computeSSDWithMask(getClient(),mask->getClient());
			break;
		case SAD:
			out=computeSADWithMask(getClient(),mask->getClient());
			break;
		case CR:
			out=computeCorrelationRatioWithMask(getClient(),mask->getClient());
			break;
	}
	if (out==0) {
		setResult("cannot compute correlation with mask : size or voxel size not correct ");
		return TCL_ERROR;
	}
    TclInrimage *tclOut = new TclInrimage(getModule(), typeid(TclInrimage), out);
	setResult(tclOut->name());
	return TCL_OK;
}
int TclInrimage::logicCmd(int argc, char **argv) {
  enum {
    NONE,
    NOT,
    OR,
    XOR,
    AND
  } operation = NONE;
  TclInrimage *op=0;

  if(argc < 1) return COMMAND_SYNOPSIS;
  if(!strcmp(argv[0], "not")) operation = NOT;
  else if(!strcmp(argv[0], "and")) operation = AND;
  else if(!strcmp(argv[0], "or")) operation = OR;
  else if(!strcmp(argv[0], "xor")) operation = XOR;
  else return COMMAND_SYNOPSIS;

  switch(operation) {
  case NOT:
  case NONE:
    break;
  default:
    op = TclInrimage::getObject(argc, argv, 1);
    if(!op) return TCL_ERROR;
    break;
  }

  switch(operation) {
  case NOT:
    logicalNot(getClient());
    redraw();
    setResult("!%s%u", getObjectName(), getRef());
    break;
  case AND:
    logicalAnd(getClient(), op->getClient());
    redraw();
    setResult("%s%u &= %s%u", getObjectName(), getRef(),
	      getObjectName(), op->getRef());
    break;
  case OR:
    logicalOr(getClient(), op->getClient());
    redraw();
    setResult("%s%u |= %s%u", getObjectName(), getRef(),
	      getObjectName(), op->getRef());
    break;
  case XOR:
    logicalXor(getClient(), op->getClient());
    redraw();
    setResult("%s%u ^= %s%u", getObjectName(), getRef(),
	      getObjectName(), op->getRef());
    break;
  default:
    break;
  }
  return TCL_OK;
}


int TclInrimage::chamferCmd(int argc, char **argv) {

  TclString name;
  TclInrimage *chamfermapimg;
  TclInrimage * chamfermapimg2;
  InrChamferMap * output;
  Inrimage * output2;

   int maskDim  = 3;
   int maskSize = 3;
   int mapLocation = 1;
   int exactMap = 0;
   //   float dy, dz, voxelSize;
   int propFlag = 0;
   int flag;

   for (int i = 0; i < argc; i++)
     {
       if (!strcmp(argv[i], "2D"))
	 maskDim = 2;
       else if (!strcmp(argv[i], "3D"))
	 maskDim = 3;
       else if (!strcmp(argv[i], "size"))
	 if (++i < argc) sscanf(argv[i], "%i", &maskSize);
	 else return COMMAND_SYNOPSIS;
       else if (!strcmp(argv[i], "location"))
	 if (++i < argc) sscanf(argv[i], "%i", &mapLocation);
	 else return COMMAND_SYNOPSIS;
       else if (!strcmp(argv[i], "exactmap"))
	 if (++i < argc) sscanf(argv[i], "%i", &exactMap);
	 else return COMMAND_SYNOPSIS;
       else if (!strcmp(argv[i], "propagate"))
	 if (++i < argc) sscanf(argv[i], "%i", &propFlag);
	 else return COMMAND_SYNOPSIS;

       else return COMMAND_SYNOPSIS;
     }

   if (propFlag == 1)
     flag = COLOR_PROPAGATION;
   else if (propFlag == 2)
     flag = POSITION_PROPAGATION;
   else 
     flag = NO_PROPAGATION;

   name = getClient()->getName();

   if (flag == NO_PROPAGATION)
     {
       output = new InrChamferMap(getClient(), mapLocation, maskSize, maskDim);
       if (exactMap)
	 output->exactMap();
       
       chamfermapimg = new TclInrimage(getModule(), typeid(TclInrimage), output);
       chamfermapimg->getClient()->setName(name + ".chamfermap.gz");
       getModule()->setResult("%s%u", TclInrimage::getTCLObjectName(), chamfermapimg->getRef());
     }
   else
     {
       output2 = new Inrimage(*getClient());
       output = new InrChamferMap(getClient(), output2, flag, mapLocation, maskSize, maskDim);
       if (exactMap)
	 output->exactMap();

       chamfermapimg  = new TclInrimage(getModule(), typeid(TclInrimage), output);
       chamfermapimg->getClient()->setName(name + ".chamfermap.gz");
       
       chamfermapimg2 = new TclInrimage(getModule(), typeid(TclInrimage), output2);
       chamfermapimg2->getClient()->setName(name + ".voronoi.gz");
   
       getModule()->setResult("%s%u %s%u", 
			      TclInrimage::getTCLObjectName(), chamfermapimg->getRef(), 
			      TclInrimage::getTCLObjectName(), chamfermapimg2->getRef());
       
     }

   printf("Maximum relative error with the euclidian distance : %f\n", output->getMaxRelativeError());
   return TCL_OK;
   
}


int TclInrimage::thresholdCmd(int argc, char **argv) {
  int i;
	for (i=0; i<argc; ++i)
    CERR << argv[i] << ENDL;
  CERR << ENDL;
  Voxel *min, *max, *fg, *bg;

  bg = Voxel::voxelByType(getType());
  fg = Voxel::voxelByType(getType());
  min = Voxel::voxelByType(getType());
  max = Voxel::voxelByType(getType());
  bg->setMin();
  min->setMin();
  fg->setMax();
  max->setMax();

  // parse arguments
  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "min")) {
      if(++i < argc) {
	try {
	  min->scan(argv[i]);
	}
	catch(Voxel::InvalidValue&) {
	  goto error;
	}
      }
      else goto error;
    }
    else if(!strcmp(argv[i], "max")) {
      if(++i < argc) {
	try {
	  max->scan(argv[i]);
	}
	catch(Voxel::InvalidValue&) {
	  goto error;
	}
      }
      else goto error;
    }
    else if(!strcmp(argv[i], "fg")) {
      if(++i < argc) {
	try {
	  fg->scan(argv[i]);
	}
	catch(Voxel::InvalidValue&) {
	  goto error;
	}
      }
      else goto error;
    }
    else if(!strcmp(argv[i], "bg")) {
      if(++i < argc) {
	try {
	  bg->scan(argv[i]);
	}
	catch(Voxel::InvalidValue&) {
	  goto error;
	}
      }
      else goto error;
    }
    else goto error;
  }

  // threshold
  threshold(getClient(), min, max, bg, fg);
  changeAndRedraw();
  setResult("%s%u", getObjectName(), getRef());

  delete fg;
  delete bg;
  delete min;
  delete max;
  return TCL_OK;

 error:
  delete fg;
  delete bg;
  delete min;
  delete max;
  return COMMAND_SYNOPSIS;
}


int TclInrimage::windowingCmd(int argc, char **argv) {
  int i;
  Voxel *min, *max;

  min = Voxel::voxelByType(getType());
  max = Voxel::voxelByType(getType());
  min->setMin();
  max->setMax();

  // parse arguments
  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "min")) {
      if(++i < argc) {
	try {
	  min->scan(argv[i]);
	}
	catch(Voxel::InvalidValue&) {
	  goto error;
	}
      }
      else goto error;
    }
    else if(!strcmp(argv[i], "max")) {
      if(++i < argc) {
	try {
	  max->scan(argv[i]);
	}
	catch(Voxel::InvalidValue&) {
	  goto error;
	}
      }
      else goto error;
    }
    else goto error;
  }
  { 
    // threshold
    Inrimage *out=windowing(getClient(), min, max);
    TclInrimage *windowimg = new TclInrimage(getModule(), typeid(TclInrimage),
					     out);
    TclString name = getClient()->getName();
    if(name.getLength() > 3 &&
       TclString(name, name.getLength() - 3, name.getLength() - 1) == ".gz")
      name = TclString(name, 0, name.getLength() - 4);
    if(name.getLength() > 4 &&
       TclString(name, name.getLength() - 4, name.getLength() - 1) == ".inr")
      name = TclString(name, 0, name.getLength() - 5);
    windowimg->getClient()->setName(name + ".window.inr.gz");
    setResult("%s%u", getObjectName(), windowimg->getRef());
  }

  delete min;
  delete max;
  return TCL_OK;

 error:
  delete min;
  delete max;
  return COMMAND_SYNOPSIS;
}

int TclInrimage::cropHistogrammCmd(int argc, char **argv) {
  int i;
  Voxel *min, *max, *bg;

  bg = Voxel::voxelByType(getType());
  min = Voxel::voxelByType(getType());
  max = Voxel::voxelByType(getType());
  bg->setMin();
  min->setMin();
  max->setMax();

  // parse arguments
  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "min")) {
      if(++i < argc) {
	try {
	  min->scan(argv[i]);
	}
	catch(Voxel::InvalidValue&) {
	  goto error;
	}
      }
      else goto error;
    }
    else if(!strcmp(argv[i], "max")) {
      if(++i < argc) {
	try {
	  max->scan(argv[i]);
	}
	catch(Voxel::InvalidValue&) {
	  goto error;
	}
      }
      else goto error;
    }
    else if(!strcmp(argv[i], "bg")) {
      if(++i < argc) {
	try {
	  bg->scan(argv[i]);
	}
	catch(Voxel::InvalidValue&) {
	  goto error;
	}
      }
      else goto error;
    }
    else goto error;
  }

  // threshold
  cropHistogramm(getClient(), min, max, bg);
  changeAndRedraw();
  setResult("%s%u", getObjectName(), getRef());

  delete bg;
  delete min;
  delete max;
  return TCL_OK;

 error:
  delete bg;
  delete min;
  delete max;
  return COMMAND_SYNOPSIS;
}


int TclInrimage::noiseCmd(int argc, char **argv) {
  int i;
  unsigned int x1, y1, z1, x2, y2, z2;
  double mean = 255.0, dev = 0.0;
  Inrimage::Noise kind = Inrimage::GAUSSIAN_NOISE;

  x1 = y1 = z1 = 0;
  x2 = getClient()->getX() - 1;
  y2 = getClient()->getY() - 1;
  z2 = getClient()->getZ() - 1;
  if(!parseBoxOptions(getModule(), this, argc, argv,
		      x1, y1, z1, x2, y2, z2)) return COMMAND_SYNOPSIS;

  // parse arguments
  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "org") || !strcmp(argv[i], "center")
       || !strcmp(argv[i], "end") || !strcmp(argv[i], "size")) i = i + 3;
    else if(!strcmp(argv[i], "white")) {
      if(++i < argc) {
	kind = Inrimage::WHITE_NOISE;
	if(sscanf(argv[i], "%lf", &mean) != 1) return COMMAND_SYNOPSIS;
      }
      else return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "gaussian")) kind = Inrimage::GAUSSIAN_NOISE;
    else if(!strcmp(argv[i], "us")) kind = Inrimage::US_NOISE;
    else if(!strcmp(argv[i], "dev")) {
      if(++i < argc) {
	if(sscanf(argv[i], "%lf", &dev) != 1) return COMMAND_SYNOPSIS;
      }
      else return COMMAND_SYNOPSIS;
    }
    else return COMMAND_SYNOPSIS;
  }

  getClient()->noise(x1, y1, z1, x2, y2, z2, kind, mean, dev);
  redraw();
  setResult("%s%u", getObjectName(), getRef());
  return TCL_OK;
}


int TclInrimage::fillCmd(int argc, char **argv) {
  int i, dim = 2;
  Voxel *bg, *fg;

  bg = Voxel::voxelByType(getType());
  fg = Voxel::voxelByType(getType());

  if(argc < 4) goto error;

  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "fg")) {
      if(++i < argc) {
	try {
	  fg->scan(argv[i]);
	}
	catch(Voxel::InvalidValue&) {
	  goto error;
	}
      }
      else goto error;
    }
    else if(!strcmp(argv[i], "bg")) {
      if(++i < argc) {
	try {
	  bg->scan(argv[i]);
	}
	catch(Voxel::InvalidValue&) {
	  goto error;
	}
      }
      else goto error;
    }
    else if(!strcmp(argv[i], "2D")) dim = 2;
    else if(!strcmp(argv[i], "3D")) dim = 3;
    else goto error;
  }

  if(dim == 2) fillContours( getClient(), getClient()->getType(), NULL, bg, fg, false );
  else fillContours( getClient(), getClient()->getType(), NULL, bg, fg, true );
  redraw();
  setResult("%s%u", getObjectName(), getRef());

  delete fg;
  delete bg;
  return TCL_OK;

 error:
  delete fg;
  delete bg;
  return COMMAND_SYNOPSIS;
}


int TclInrimage::connexeCmd(int argc, char **argv) {
  
  ConnectingElement CE;
  int tcc_ht;
  CONNEXE_PARAMETER par = {1, -1, COMPONENT_BINARY};
  int i;
  bool bsb = false;
  Voxel *sb, *sh, *bg, *fg;
  CONNECTIVITY_TYPE con = TWENTY_SIX_CONNECTIVITY;
  bool thereIsAHighThreshold = false;

  enum {
    BINARY,
    THRESHOLD,
    HYSTERESIS
  } operation = BINARY;

  typedef enum {
    _COMPONENTS_,
    _HYSTERESIS_,
    _FILL_CONTOURS_,
    _RECONSTRUCT_
  } enumComputation;
  
  enumComputation typeComputation = _COMPONENTS_;


  if(argc < 2) return COMMAND_SYNOPSIS;

  sb = Voxel::voxelByType(getType());
  sh = Voxel::voxelByType(getType());
  bg = Voxel::voxelByType(getType());
  fg = Voxel::voxelByType(getType());

  sb->setValueFromDouble(1.0);
  sh->setValueFromDouble(1.0);
  
  getClient()->getBounds(bg, fg);


  // parse parameters
  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "con")) {
      if(++i < argc) {
	con = TclInrimage::connectivityFromString(argv[i]);
	if(con < 0) return COMMAND_SYNOPSIS;
	CE.setConnectivity(con);
      }
      else return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "tcc")) {
      if(++i < argc) {
	if(sscanf(argv[i], "%i", &par.min_size) != 1) return COMMAND_SYNOPSIS;
	CE.setMinimalSize(par.min_size);
      }
      else return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "ncc")) {
      if(++i < argc) {
	if(sscanf(argv[i], "%i", &par.max_nbcc) != 1) return COMMAND_SYNOPSIS;
	CE.setMaximalNumberOfComponents(par.max_nbcc);
      }
      else return COMMAND_SYNOPSIS;
    }
   else if(!strcmp(argv[i], "sb") || !strcmp(argv[i], "lt")) {
      if(++i < argc) {
	bsb = true;
	try {
	  sb->scan(argv[i]);
	  sh->setValueFromDouble(sb->getValueAsDouble());
	}
	catch(Voxel::InvalidValue&) {
	  return COMMAND_SYNOPSIS;
	}
	//operation = THRESHOLD;
      }
      else return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "sh") || !strcmp(argv[i], "ht")) {
      if(++i < argc) {
	try {
	  sh->scan(argv[i]);
	  typeComputation = _HYSTERESIS_;
	  CE.setOutputType( COMPONENT_BINARY );
	  thereIsAHighThreshold = true;
	}
	catch(Voxel::InvalidValue&) {
	  return COMMAND_SYNOPSIS;
	}
	//operation = HYSTERESIS;
      }
      else return COMMAND_SYNOPSIS;
    }
    
    else if(!strcmp(argv[i], "max")) { par.max_nbcc = 1; CE.setMaximalNumberOfComponents(par.max_nbcc);}
    else if(!strcmp(argv[i], "output")) {
      if(++i < argc) {
	if(!strcmp(argv[i], "binary"))
	  {par.output = COMPONENT_BINARY; CE.setOutputType(par.output);}
	else if(!strcmp(argv[i], "label"))
	  {par.output = COMPONENT_LABEL; CE.setOutputType(par.output);}
	else return COMMAND_SYNOPSIS;
      }
      else return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "hyster")) {
      typeComputation = _HYSTERESIS_;
      // on recupere le mode hyster
      if(++i < argc) {
	if(!strcmp(argv[i], "tcc_ht"))
	  if(++i < argc) {
	    if(sscanf(argv[i], "%i", &tcc_ht) != 1) return COMMAND_SYNOPSIS;
	    CE.setMinimalSizeHighThreshold(tcc_ht);
	  }	  
	  else return COMMAND_SYNOPSIS;
	else return COMMAND_SYNOPSIS;
      }
    }
    
    else return COMMAND_SYNOPSIS;
  }
  
  if(operation == HYSTERESIS && !bsb) return COMMAND_SYNOPSIS;

  // allocate result image
  switch (typeComputation) {
  case _COMPONENTS_ :
      if ( thereIsAHighThreshold == true )
	connectedComponents( getClient(), sb, sh, &CE );
      else
	connectedComponents( getClient(), sb->getValueAsDouble(), &CE );
    break;
  case _HYSTERESIS_ :
    hysteresis(getClient(), sb, sh, &CE);
    break;
  default:
    CERR << ENDL << " Error in connexeCmd: invalid operation or operation not yet implemented" << ENDL;
    break;
  }
     
  redraw();
  if(par.output == COMPONENT_BINARY)
    setResult("%s%u connexe components extracted",
	      getObjectName(), getRef());
  else
    setResult("%s%u labeled",
	      getObjectName(), getRef());
  return TCL_OK;  
}


int TclInrimage::histogramCmd(int argc, char **argv) {
  if (argc>0)
    return COMMAND_SYNOPSIS;
  else {
    GenericHistogram *h=getClient()->createHistogram();
    TclGenericHistogram *th= new TclGenericHistogram(getModule(),h);
    setResult("%s",(char *) th->name());
    return TCL_OK;
  }
}

int TclInrimage::maskedHistogramCmd(int argc, char **argv) {
  if (argc!=1)
    return COMMAND_SYNOPSIS;
  else {
    Inrimage* mask = new Inrimage( argv[0] );
    GenericHistogram *h=getClient()->createMaskedHistogram( mask );
    TclGenericHistogram *th= new TclGenericHistogram(getModule(),h);
    setResult("%s",(char *) th->name());
    return TCL_OK;
  }
}

int TclInrimage::morphoCmd(int argc, char **argv) {

  int i;
  unsigned int it = 1;
  bool with_struct = false, with_con = false, mode_binary = false;
  TclStructuringElement *Struct = (TclStructuringElement *) NULL;
	unsigned int R = 1;
	bool with_radius = false;

  enum {
    DILATION,
    EROSION,
    CLOSURE,
    OPENING,
    TOPHAT,
    INVTOPHAT,
    MORPHOGRAD,
    MORPHO_LOWER_GRAD,
    MORPHO_UPPER_GRAD,
    MEDIAN_FILTER
  } operation;

  CONNECTIVITY_TYPE con = TWENTY_SIX_CONNECTIVITY;
  Voxel *fg, *bg;

  fg = Voxel::voxelByType(getType());
  bg = Voxel::voxelByType(getType());
  getClient()->getBounds(bg, fg);
  operation = DILATION;

  // parse arguments
  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "dil")) operation = DILATION;
    else if(!strcmp(argv[i], "ero")) operation = EROSION;
    else if(!strcmp(argv[i], "close")) operation = CLOSURE;
    else if(!strcmp(argv[i], "open")) operation = OPENING;
    else if(!strcmp(argv[i], "tophat")) operation = TOPHAT;
    else if(!strcmp(argv[i], "invtophat")) operation = INVTOPHAT;
    else if(!strcmp(argv[i], "grad")) operation = MORPHOGRAD;
    else if(!strcmp(argv[i], "lowgrad")) operation = MORPHO_LOWER_GRAD;
    else if(!strcmp(argv[i], "uppgrad")) operation = MORPHO_UPPER_GRAD;
    else if(!strcmp(argv[i], "median")) operation = MEDIAN_FILTER;
    else if(!strcmp(argv[i], "binary")) mode_binary = true;
    else if(!strcmp(argv[i], "greylevel")) mode_binary = false;
    else if(!strcmp(argv[i], "R")) {
      if(++i < argc) {
				if(sscanf(argv[i], "%u", &R) != 1) goto error;
				with_radius = true;
      }
      else goto error;
    }
    else if(!strcmp(argv[i], "i")) {
      if(++i < argc) {
				if(sscanf(argv[i], "%u", &it) != 1) goto error;
      }
      else goto error;
    }
    else if(!strcmp(argv[i], "con")) {
      if ( with_struct ) 
	goto error;
      else
	with_con = true;
      if(++i < argc) {
	con = TclInrimage::connectivityFromString(argv[i]);
	if(con < 0) goto error;
      }
      else goto error;
    }
    else if(!strcmp(argv[i], "struct")) {
      if ( with_con ) 
	goto error;
      else {
	with_struct = true;
	if(++i>=argc) return COMMAND_SYNOPSIS;
	else
	  Struct = TclStructuringElement::getObject(argc,argv,i);
      }
    }
    else 
      goto error;
  }

  // proceed morphomathematical operators
  switch(operation) {
  case DILATION:
    if (with_con) 
			dilation(getClient(), con, it, mode_binary);
    else 
			if (with_radius)
				dilation(getClient(), R, it, 3, mode_binary);
			else
				dilation(getClient(), Struct, it, mode_binary);
    setResult("%s%u dilated", getObjectName(), getRef());
    break;
  case EROSION:
    if (with_con) 
			erosion(getClient(), con, it, mode_binary);
    else 
			if (with_radius)
				erosion(getClient(), R, it, 3, mode_binary);
			else
				erosion(getClient(), Struct, it, mode_binary);
    setResult("%s%u eroded", getObjectName(), getRef());
    break;
  case CLOSURE:
    if (with_con)
	closing(getClient(), con, it, mode_binary);
    else 
	closing(getClient(), Struct, it, mode_binary);
    setResult("%s%u closure", getObjectName(), getRef());
    break;
  case OPENING:
    if (with_con) 
	opening(getClient(), con, it, mode_binary);
    else 
	opening(getClient(), Struct, it, mode_binary);
    setResult("%s%u opening", getObjectName(), getRef());
    break;
  case TOPHAT:
    if (with_con) 
	topHat(getClient(), con, it, mode_binary);
    else 
	topHat(getClient(), Struct, it, mode_binary);
    setResult("%s%u Top Hat", getObjectName(), getRef());
    break;
  case INVTOPHAT:
    if (with_con) 
	invTopHat(getClient(), con, it, mode_binary);
    else 
	invTopHat(getClient(), Struct, it, mode_binary);
    setResult("%s%u Inv Top Hat", getObjectName(), getRef());
    break;
  case MORPHOGRAD:
    if (with_con) 
	morphologicalGradient(getClient(), con, it, mode_binary);
    else 
        morphologicalGradient(getClient(), Struct, it, mode_binary);
    setResult("%s%u Morphological Gradient", getObjectName(), getRef());
    break;
  case MORPHO_LOWER_GRAD:
    if (with_con) 
      morphologicalLowerGradient(getClient(), con, it, mode_binary);
    else 
      morphologicalLowerGradient(getClient(), Struct, it, mode_binary);
    setResult("%s%u Morphological Lower Gradient", getObjectName(), getRef());
    break;
  case MORPHO_UPPER_GRAD:
    if (with_con) 
      morphologicalUpperGradient(getClient(), con, it, mode_binary);
    else 
      morphologicalUpperGradient(getClient(), Struct, it, mode_binary);
    setResult("%s%u Morphological Upper Gradient", getObjectName(), getRef());
    break;
  case MEDIAN_FILTER:
    if (with_con) 
      medianFiltering(getClient(), con);
    else 
      medianFiltering(getClient(), Struct);
    setResult("%s%u Median Filtering", getObjectName(), getRef());
    break;
  }

  redraw();
 
  delete fg;
  delete bg;
  return TCL_OK;

 error:
  delete fg;
  delete bg;
  return COMMAND_SYNOPSIS;
}


int TclInrimage::gradientCmd(int argc, char **argv) {
  int i;
  bool max = false, compute2D = false, useSobel = false;
  float lt = 0.0;
  RECURSIVE_FILTER_TYPE filter = DERICHE_STEP_EDGE;
  Inrimage *output = 0, *outdir = 0;
  TclInrimage *gradientimg, *gradientdir;
  double sigma = 3.0;
  TclString name;
  GRADIENT_DIRECTION dir = GD_NONE;

  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "sigma")) {
      if(++i < argc) {
	if(sscanf(argv[i], "%lf", &sigma) != 1) return COMMAND_SYNOPSIS;
      }
      else return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "2D")) compute2D = true;
    else if(!strcmp(argv[i], "3D")) compute2D = false;
    else if(!strcmp(argv[i], "dir")) {
      if(++i < argc) {
	if(!strcmp(argv[i], "vectors")) dir = GD_VECTORS;
	else if(!strcmp(argv[i], "sample")) dir = GD_SAMPLE;
	else return COMMAND_SYNOPSIS;
      }
      else return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "max")) max = true;
    else if(!strcmp(argv[i], "sobel")) useSobel = true;
    else if(!strcmp(argv[i], "rec")) useSobel = false;
    else if(!strcmp(argv[i], "lt")) {
      if(++i < argc) sscanf(argv[i], "%f", &lt);
      else return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "gaussian")) filter = DERICHE_GAUSSIAN;
    else if(!strcmp(argv[i], "marta")) filter = MARTA_GAUSSIAN;
    else if(!strcmp(argv[i], "deriche")) filter = DERICHE_STEP_EDGE;
    else return COMMAND_SYNOPSIS;
  }

  if(useSobel)
    sobel(getClient(), output, outdir, (int) sigma, dir, max, compute2D, lt);
  else
    gradient(getClient(), output, outdir, sigma, filter, dir, max,
			  compute2D, lt);

  gradientimg = new TclInrimage(getModule(), typeid(TclInrimage),
				output);
  
  name = getClient()->getName();
  if(name.getLength() > 3 &&
     TclString(name, name.getLength() - 3, name.getLength() - 1) == ".gz")
    name = TclString(name, 0, name.getLength() - 4);
  if(name.getLength() > 4 &&
     TclString(name, name.getLength() - 4, name.getLength() - 1) == ".inr")
    name = TclString(name, 0, name.getLength() - 5);
  gradientimg->getClient()->setName(name + ".gradient.gz");

  if(dir) {
    gradientdir = new TclInrimage(getModule(), typeid(TclInrimage), outdir);
    gradientdir->getClient()->setName(name + ".gradient_direction.gz");
    setResult("%s%u %s%u", getObjectName(), gradientimg->getRef(),
	      getObjectName(), gradientdir->getRef());
  }
  else
    setResult("%s%u", getObjectName(), gradientimg->getRef());
  return TCL_OK;
}


int TclInrimage::convoluteCmd(int argc, char **argv) {
  FILTER_TYPE filter = (FILTER_TYPE) -1;
  bool twoD = false;
  unsigned int filterSize = 3;
  int i;

  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "2D")) twoD = true;
    else if(!strcmp(argv[i], "3D")) twoD = false;
    else if(!strcmp(argv[i], "gauss")) filter = GAUSSIAN;
    else if(!strcmp(argv[i], "smoothX")) filter = SMOOTH_X;
    else if(!strcmp(argv[i], "smoothY")) filter = SMOOTH_Y;
    else if(!strcmp(argv[i], "smoothZ")) filter = SMOOTH_Z;
    else if(!strcmp(argv[i], "derX")) filter = DERIVATIVE_X;
    else if(!strcmp(argv[i], "derY")) filter = DERIVATIVE_Y;
    else if(!strcmp(argv[i], "derZ")) filter = DERIVATIVE_Z;
    else if(!strcmp(argv[i], "der2X")) filter = SECOND_DERIVATIVE_X;
    else if(!strcmp(argv[i], "der2Y")) filter = SECOND_DERIVATIVE_Y;
    else if(!strcmp(argv[i], "der2Z")) filter = SECOND_DERIVATIVE_Z;
    else if(!strcmp(argv[i], "filterSize")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%u", &filterSize) != 1) return COMMAND_SYNOPSIS;
    }
    else return COMMAND_SYNOPSIS;
  }

  if(filter < 0) return COMMAND_SYNOPSIS;
  convolute(getClient(), filter, filterSize, twoD);
  changeAndRedraw();
  setResult("%s%u convoluted", getObjectName(), getRef());
  return TCL_OK;
}


int TclInrimage::curvatureCmd(int argc, char **argv) {
  if(argc > 0) return COMMAND_SYNOPSIS;
  curvature(getClient());
  changeAndRedraw();
  setResult("%s%u curvature", getObjectName(), getRef());
  return TCL_OK;
}


int TclInrimage::recfilterCmd(int argc, char **argv) {
  RECURSIVE_FILTER_TYPE rfilter = DERICHE_STEP_EDGE;
  FILTER_TYPE filter = (FILTER_TYPE) -1;
  bool twoD = false;
  double sigma = 3.0;
  int i;

  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "2D")) twoD = true;
    else if(!strcmp(argv[i], "3D")) twoD = false;
    else if(!strcmp(argv[i], "gaussian")) rfilter = DERICHE_GAUSSIAN;
    else if(!strcmp(argv[i], "marta")) rfilter = MARTA_GAUSSIAN;
    else if(!strcmp(argv[i], "deriche")) rfilter = DERICHE_STEP_EDGE;
    else if(!strcmp(argv[i], "gauss")) filter = GAUSSIAN;
    else if(!strcmp(argv[i], "smoothX")) filter = SMOOTH_X;
    else if(!strcmp(argv[i], "smoothY")) filter = SMOOTH_Y;
    else if(!strcmp(argv[i], "smoothZ")) filter = SMOOTH_Z;
    else if(!strcmp(argv[i], "derX")) filter = DERIVATIVE_X;
    else if(!strcmp(argv[i], "derY")) filter = DERIVATIVE_Y;
    else if(!strcmp(argv[i], "derZ")) filter = DERIVATIVE_Z;
    else if(!strcmp(argv[i], "der2X")) filter = SECOND_DERIVATIVE_X;
    else if(!strcmp(argv[i], "der2Y")) filter = SECOND_DERIVATIVE_Y;
    else if(!strcmp(argv[i], "der2Z")) filter = SECOND_DERIVATIVE_Z;
    else if(!strcmp(argv[i], "sigma")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%lf", &sigma) != 1) return COMMAND_SYNOPSIS;
    }
    else return COMMAND_SYNOPSIS;
  }

  if(filter < 0) return COMMAND_SYNOPSIS;
  recursiveFilter(getClient(), rfilter, filter, sigma, twoD);
  changeAndRedraw();
  setResult("%s%u filtered", getObjectName(), getRef());
 
  return TCL_OK;
}


int TclInrimage::laplacianCmd(int argc, char **argv) {
  RECURSIVE_FILTER_TYPE rfilter = DERICHE_STEP_EDGE;
  bool twoD = false;
  double sigma = 3.0;
  int i;

  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "2D")) twoD = true;
    else if(!strcmp(argv[i], "3D")) twoD = false;
    else if(!strcmp(argv[i], "gaussian")) rfilter = DERICHE_GAUSSIAN;
    else if(!strcmp(argv[i], "marta")) rfilter = MARTA_GAUSSIAN;
    else if(!strcmp(argv[i], "deriche")) rfilter = DERICHE_STEP_EDGE;
    else if(!strcmp(argv[i], "sigma")) {
      if(++i >= argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%lf", &sigma) != 1) return COMMAND_SYNOPSIS;
    }
    else return COMMAND_SYNOPSIS;
  }

  laplacian(getClient(), rfilter, sigma, twoD);
  changeAndRedraw();
  setResult("%s%u laplacian", getObjectName(), getRef());
  return TCL_OK;
}


int TclInrimage::splitCmd(int argc, char **argv) {
  TclString res, one;
  unsigned int i;
  TclInrimage *cur;
  Inrimage **inrs;

  if(argc != 0) return COMMAND_SYNOPSIS;
  inrs = getClient()->splitVectors();
  cur = new TclInrimage(getModule(), typeid(TclInrimage),inrs[0]);
  res.sprintf("%s%u", getObjectName(), cur->getRef());
  for(i = 1; i < getVdim(); i++) {
    cur = new TclInrimage(getModule(), typeid(TclInrimage),inrs[i]);
    one.sprintf(" %s%u", getObjectName(), cur->getRef());
    res += one;
  }
  delete[] inrs;

  setResult(res);
  return TCL_OK;
}


int TclInrimage::componentsCmd(int argc, char **argv) {
  int i;
  Inrimage **inrs;

  if(argc == 0) return COMMAND_SYNOPSIS;
  if(argc != (int) getVdim()) {
    setResult("number of inrimage components should be %u", getVdim());
    return TCL_ERROR;
  }

  inrs = new Inrimage*[argc];
  for(i = 0; i < argc; i++) {
    inrs[i] = TclInrimage::getObject(argc, argv, i)->getClient();
    if(!inrs[i]) {
      delete[] inrs;
      return COMMAND_SYNOPSIS;
    }
  }
  try {
    getClient()->setComponentsFrom(inrs);
  }
  catch(Inrimage::InvalidSize& ) {
    delete[] inrs;
    setResult("all inrimage should have the same size");
    return TCL_ERROR;
  }
  catch(Inrimage::InvalidArguments& ) {
    delete[] inrs;
    setResult("all inrimage should be scalar");
    return TCL_ERROR;
  }

  delete[] inrs;
  setResult("%s%u", getObjectName(), getRef());
  changeAndRedraw();
  return TCL_OK;
}


int TclInrimage::dftCmd(int argc, char **argv) {
  bool twoD = false;
  int i;

  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "2D")) twoD = true;
    else if(!strcmp(argv[i], "3D")) twoD = false;
    else return COMMAND_SYNOPSIS;
  }

  fft(getClient(), twoD);
  changeAndRedraw();
  if(getVdim() == 1)
    setResult("%s%u inverse DFT",
	      getObjectName(), getRef());
  else
    setResult("%s%u DFT",
	      getObjectName(), getRef());
  return TCL_OK;
}
TclInrimage *TclInrimage::getInterface(Inrimage *inr) {
  TclInrimage *tinr=0;

  std::list<ObjectPointer<Inrimage> *>::iterator itp;
  for(itp=inr->getFirstPointer();(tinr==0) &&(itp!=inr->getLastPointer());itp++) {
    TclInrimage *t = dynamic_cast<TclInrimage *>((*itp)->getOwner());
    if (t)
      tinr=t;
  }
  return tinr;
}

int TclInrimage::undersampleCmd(int argc, char **argv) {
  unsigned int sx = 1, sy = 1, sz = 1;

  if(argc < 1) return COMMAND_SYNOPSIS;
  if(argc == 1) {
    if(sscanf(argv[0], "%u", &sx) != 1) return COMMAND_SYNOPSIS;
    sy = sz = sx;
  }
  else {
    for(int i = 0; i < argc; i++) {
      if(!strcmp(argv[i], "X")) {
	if(++i >= argc) return COMMAND_SYNOPSIS;
	if(sscanf(argv[i], "%u", &sx) != 1) return COMMAND_SYNOPSIS;
      }
      else if(!strcmp(argv[i], "Y")) {
	if(++i >= argc) return COMMAND_SYNOPSIS;
	if(sscanf(argv[i], "%u", &sy) != 1) return COMMAND_SYNOPSIS;
      }
      else if(!strcmp(argv[i], "Z")) {
	if(++i >= argc) return COMMAND_SYNOPSIS;
	if(sscanf(argv[i], "%u", &sz) != 1) return COMMAND_SYNOPSIS;
      }
      else return COMMAND_SYNOPSIS;
    }
  }

  getClient()->underSample(sx, sy, sz);
  resizeAndRedraw();
  setResult("undersampled");
  return TCL_OK;
}

int TclInrimage::oversampleCmd(int argc, char **argv) {
  unsigned int sx = 0, sy = 0, sz = 0;

  if(argc != 3) return COMMAND_SYNOPSIS;
  if(sscanf(argv[0], "%u", &sx) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%u", &sy) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[2], "%u", &sz) != 1) return COMMAND_SYNOPSIS;

  try {
    getClient()->overSample(sx, sy, sz);
  }
  catch(Inrimage::InvalidSize &) {
    setResult("Invalid new size (should be greater than image size)");
    return TCL_ERROR;
  }
  resizeAndRedraw();
  setResult("oversampled");
  return TCL_OK;
}

int TclInrimage::resampleCmd(int argc, char **argv) {
  unsigned int sx = 0, sy = 0, sz = 0;

  if(argc != 3) return COMMAND_SYNOPSIS;
  if(sscanf(argv[0], "%u", &sx) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[1], "%u", &sy) != 1) return COMMAND_SYNOPSIS;
  if(sscanf(argv[2], "%u", &sz) != 1) return COMMAND_SYNOPSIS;

  getClient()->resample(sx, sy, sz);
  resizeAndRedraw();
  setResult("resampled");
  return TCL_OK;
}

int TclInrimage::imageFormatCmd(TclModule *m,int argc, char **argv) {
  if (argc!=2) 
    return COMMAND_SYNOPSIS;
  else if (!strcmp(argv[1],"input")) {
    
    TclString res,out;
    PTRIMAGE_FORMAT f;
    for(f=firstImageFormat();(f!=NULL);f=f->next) {
      out.sprintf("{%s %s}",f->realName,f->fileExtension);
      res+=out;
    }
    m->setResult(res);
    return TCL_OK;
  }
  else if (!strcmp(argv[1],"output")) { 
    TclString res,out;
    PTRIMAGE_FORMAT f;
    for(f=firstImageFormat();(f!=NULL);f=f->next) {
      if (f->writeImage) {
	out.sprintf("{%s %s}",f->realName,f->fileExtension);
	res+=out;
      }
    }
    m->setResult(res);
    return TCL_OK;

  }
  else {
    m->setResult("%s must be either input or output",argv[1]);
    return TCL_ERROR;
  }
}
int TclInrimage::transformCmd(int argc, char **argv) {

  RESAMPLE_TYPE interp = NEAREST;
  bool twoD = false, with_def = false, inv = false, newDim = false, worldCoordinates=false;
  TclMatrix4D *mat=0;
  Matrix4x4<double> mat_voxel;
  TclInrimage *defx=0, *defy=0, *defz=0;
  unsigned int xdim=0, ydim=0, zdim=0;
  double ovx=getClient()->getVX(), ovy=getClient()->getVY(), ovz=getClient()->getVZ();

  yav::RESAMPLE_FILTER_DERIVATIVE derivative[3] = { SAMPLE_NO_DERIVATIVE, SAMPLE_NO_DERIVATIVE, SAMPLE_NO_DERIVATIVE } ;
  float gain = 1, bias = 0;

  if (argc<1) return COMMAND_SYNOPSIS;

  for(int i=0; i<argc; ++i) {
    if(!strcmp(argv[i], "nearest"))
      interp=NEAREST;
    if(!strcmp(argv[i], "trilinear"))
      interp=TRILINEAR;
    if(!strcmp(argv[i], "cspline"))
      interp=CSPLINE;
    if(!strcmp(argv[i], "2D"))
      twoD=true;
    if(!strcmp(argv[i], "-inv"))
      inv=true;
     if(!strcmp(argv[i], "-world"))
      worldCoordinates=true;
    if(!strcmp(argv[i], "-dim")){
      newDim=true;
      if(++i>=argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%u", &xdim) != 1) return COMMAND_SYNOPSIS;
      if(++i>=argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%u", &ydim) != 1) return COMMAND_SYNOPSIS;
      if(++i>=argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%u", &zdim) != 1) return COMMAND_SYNOPSIS;
    }
    if(!strcmp(argv[i], "-vsize")){
      if(++i>=argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%lf", &ovx) != 1) return COMMAND_SYNOPSIS;
      if(++i>=argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%lf", &ovy) != 1) return COMMAND_SYNOPSIS;
      if(++i>=argc) return COMMAND_SYNOPSIS;
      if(sscanf(argv[i], "%lf", &ovz) != 1) return COMMAND_SYNOPSIS;
    }
   
    if(!strcmp(argv[i], "-mat")) 
      if(++i>=argc) return COMMAND_SYNOPSIS;
      else mat = TclMatrix4D::getObject(argc, argv, i);
    
    if(!strcmp(argv[i], "-def")) {
      with_def=true;

      if(++i>=argc) return COMMAND_SYNOPSIS;
      else
	defx = TclInrimage::getObject(argc, argv, i);
      if(++i>=argc) return COMMAND_SYNOPSIS;
      else
	defy = TclInrimage::getObject(argc, argv, i);
      if(++i>=argc) return COMMAND_SYNOPSIS;
      else
	defz = TclInrimage::getObject(argc, argv, i);
      
    }

  }

  Inrimage *reech_inr = (Inrimage *)NULL;
  
  if (!mat) {
    mat = new TclMatrix4D(typeid(TclMatrix4D), getModule());
  } 
  mat_voxel = (*mat);

  if (worldCoordinates) {
   
   double ivx=getClient()->getVX(); double ivy=getClient()->getVY(); double ivz= getClient()->getVZ();
 //  Vec3<double> ip=getClient()->getTranslation();
   mat_voxel(0,0)*=ovx/ivx;  mat_voxel(0,1)*=ovy/ivx;  mat_voxel(0,2)*=ovz/ivx;  mat_voxel(0,3)/=ivx;
   mat_voxel(1,0)*=ovx/ivy;  mat_voxel(1,1)*=ovy/ivy;  mat_voxel(1,2)*=ovz/ivy;  mat_voxel(1,3)/=ivy;
   mat_voxel(2,0)*=ovx/ivz;  mat_voxel(2,1)*=ovy/ivz;  mat_voxel(2,2)*=ovz/ivz;  mat_voxel(2,3)/=ivz;
  }

  




    
  if (inv) mat_voxel=mat_voxel.Inv(mat_voxel);
  
  if (with_def)
      switch (interp) {
     case TRILINEAR:
	if (newDim)
	  reech_inr = resampleDefTriLinear ( getClient(), defx->getClient(), defy->getClient(), defz->getClient(), mat_voxel, getClient()->getType(),
	  xdim,ydim,zdim, gain, bias );
	else
	  reech_inr = resampleDefTriLinear ( getClient(), defx->getClient(), defy->getClient(), defz->getClient(), mat_voxel, getClient()->getType(),ovx,ovy,ovz, gain, bias );
	break;
      case CSPLINE:
	if (newDim)
	  reech_inr = resampleDefCSpline ( getClient(), defx->getClient(), defy->getClient(), defz->getClient(), mat_voxel, getClient()->getType(), derivative,
	  xdim,ydim,zdim);
	else
	  reech_inr = resampleDefCSpline ( getClient(), defx->getClient(), defy->getClient(), defz->getClient(), mat_voxel, getClient()->getType(), derivative,
	  ovx,ovy,ovz);
	  break;
      case NEAREST:
	  default:
	if (newDim)
	  reech_inr = resampleDefNearest ( getClient(), defx->getClient(), defy->getClient(), defz->getClient(), mat_voxel, getClient()->getType(),
	  xdim,ydim,zdim);
	else
	  reech_inr = resampleDefNearest ( getClient(), defx->getClient(), defy->getClient(), defz->getClient(), mat_voxel, getClient()->getType(),
	  ovx,ovy,ovz);
	break;
      }

    else
      switch (interp) {
      case NEAREST:
	if (newDim)
	  reech_inr = resample4x4Nearest ( getClient(), mat_voxel, getClient()->getType(), xdim, ydim, zdim,  ovx,ovy,ovz,twoD);
	else
	  reech_inr = resample4x4Nearest ( getClient(), mat_voxel, getClient()->getType(), getClient()->getX(), getClient()->getY(), getClient()->getZ(), ovx,ovy,ovz, twoD);
	break;
      case TRILINEAR:
	if (newDim)
	  reech_inr = resample4x4TriLinear ( getClient(), mat_voxel, getClient()->getType(), xdim, ydim, zdim,  ovx,ovy,ovz, twoD, gain, bias);
	else
	  reech_inr = resample4x4TriLinear ( getClient(), mat_voxel, getClient()->getType(), getClient()->getX(), getClient()->getY(), getClient()->getZ(), ovx,ovy,ovz, twoD, gain, bias);
	break;
      case CSPLINE:
	if (newDim)
	  reech_inr = resample4x4CSpline ( getClient(), mat_voxel, getClient()->getType(), xdim, ydim, zdim, derivative);
	else
	  reech_inr = resample4x4CSpline ( getClient(), mat_voxel, getClient()->getType(), getClient()->getX(), getClient()->getY(), getClient()->getZ(), derivative);
	break;
      default:
	if (newDim)
	  reech_inr = resample4x4Nearest ( getClient(), mat_voxel, getClient()->getType(), xdim, ydim, zdim, ovx,ovy,ovz, twoD);
	else
	  reech_inr = resample4x4Nearest ( getClient(), mat_voxel, getClient()->getType(), getClient()->getX(), getClient()->getY(), getClient()->getZ(), ovx,ovy,ovz, twoD);
	break;
      }

  TclInrimage *res = new TclInrimage(getModule(), typeid(TclInrimage), reech_inr);
  
  setResult("%s%u", res->getObjectName(), res->getRef()); 

  return TCL_OK;
}

int TclInrimage::anisotropicDiffusionCmd(int argc, char **argv) {
   Inrimage* img=getClient();
   if(argc==12){
     float dt= (float)atof(argv[0]);
     if (dt<0.0)
       {
	 setResult("Time step %f must be positive",dt);
	 return TCL_ERROR;
       }
     float tmax = (float)atof(argv[1]);
     if (tmax<dt)
       {
	 setResult("Diffusion time %f must be greater than time step %f",tmax,dt);
	 return TCL_ERROR;
       }
     float quantx = (float)atof(argv[2]);
     if ((quantx<0.0)||(quantx>100))
       {
	 setResult("Gradient quantile in the x direction %f must be between 0 and 100",quantx);
	 return TCL_ERROR;
       }   
     float quanty = (float)atof(argv[3]);
     if ((quanty<0.0)||(quanty>100))
       {
	 setResult("Gradient quantile in the y direction %f must be between 0 and 100",quanty);
	 return TCL_ERROR;
       }   
     float quantz = (float)atof(argv[4]);
     if ((quantz<0.0)||(quantz>100))
       {
	 setResult("Gradient quantile in the z direction %f must be between 0 and 100",quantz);
	 return TCL_ERROR;
       } 
     float k = (float)atof(argv[5]);
    
     bool lisse;
     if(!TclModule::boolValue(argv[6],lisse)) 
       {
	 setResult("%s must be true or false",argv[6]);
	 return TCL_ERROR;
       }
     float sigma = (float)atof(argv[7]);
     if (sigma<0.0)
       {
	 setResult("Standard Deviation %f must be greater than 0",sigma);
	 return TCL_ERROR;
       }
     DIFFUSION_FUNCTION df;
     if (!strcmp(argv[8],"perona"))
       df=PERONA_MALIK;
     else if (!strcmp(argv[8],"perona_2"))
       df=PERONA_MALIK_2;
     else if (!strcmp(argv[8],"perona_3"))
       df=PERONA_MALIK_3;
     else if (!strcmp(argv[8],"aubert"))
       df=AUBERT;
     else if (!strcmp(argv[8],"weickert"))
        df=WEICKERT; 
     else if (!strcmp(argv[8],"green"))
       df=GREEN;
     else
       {
	 setResult("%s must be perona, perona_2,  perona_3, weickert, green or aubert",argv[8]);
	 return TCL_ERROR;
       }
     bool disk;
     if(!TclModule::boolValue(argv[9],disk)) 
       {
	 setResult("%s must be true or false",argv[9]);
	 return TCL_ERROR;
       }
     DIFFUSION_TYPE td;
     if (!strcmp(argv[10],"scalar"))
       td=SCALAR;
     else if (!strcmp(argv[10],"matrix"))
       td=MATRIX;
     else if (!strcmp(argv[10],"krissian"))
       td=KRISSIAN; 
     else if (!strcmp(argv[10],"matkrissian"))
       td=MATRIX_KRISSIAN;
     else
       {
	 setResult("%s must be scalar, matrix, krissian, or matkrissian",argv[10]);
	 return TCL_ERROR;
       }

     NUMERICAL_SCHEME sc;
     if (!strcmp(argv[11],"explicite"))
       sc=EXPLICIT;
     else if (!strcmp(argv[11],"aos"))
       sc=AOS_scheme;
     else
       {
	 setResult("%s must be explicite or aos",argv[11]);
	 return TCL_ERROR;
       }
    
     TclInrimage * tclOutput = new TclInrimage(*this);
     Inrimage * output = tclOutput->getClient();
#ifdef USE_MTL
     anisotropicDiffusion(img, output,dt,tmax,quantx, quanty,quantz,k,lisse,sigma,df,disk,td,sc);
#endif
     setResult("%s%u", getObjectName(), tclOutput->getRef());
     return TCL_OK;
   }
   else return COMMAND_SYNOPSIS;
}

int TclInrimage::interpCmd(int argc, char **argv) {
   if(argc>1) return COMMAND_SYNOPSIS;
   Inrimage* img=getClient();
   if(argc==1){
	   if(!strcmp(argv[0],"nearest")){
		   img->setInterpolator(new NearestInterpolator());
	   }
	   else if(!strcmp(argv[0],"trilin")) {
		   if(img->vdim==1)  img->setInterpolator(TrilinInterpolator::interpolatorByType(img->type));
		   else img->setInterpolator(TrilinVectInterpolator::interpolatorByType(img->type,img->vdim));
	   }
	   else if(!strcmp(argv[0],"trilinGeneric")) {
			img->setInterpolator(new TrilinInterpolator());
	   }
	   else return COMMAND_SYNOPSIS;
   }
   if(dynamic_cast<NearestInterpolator const*>(img->getInterpolator())) setResult("nearest");
   else if(dynamic_cast<TrilinInterpolator const*>(img->getInterpolator())) setResult("trilin");
   else setResult("unknown");
   return TCL_OK;
}


int TclInrimage::extrapCmd(int argc, char **argv) {
   if(argc>1) return COMMAND_SYNOPSIS;
   Inrimage* img=getClient();
   if(argc==1){
      if(!strcmp(argv[0],"nearest")) img->setExtrapolator(new NearestExtrapolator());
      else if(!strcmp(argv[0],"tozero")) img->setExtrapolator(new ToZeroExtrapolator());
      else return COMMAND_SYNOPSIS;
   }
   if(dynamic_cast<NearestExtrapolator const*>(img->getExtrapolator())) setResult("nearest");
   else if(dynamic_cast<ToZeroExtrapolator const*>(img->getExtrapolator())) setResult("tozero");
   else setResult("unknown");
   return TCL_OK;
}


int TclInrimage::regionCmd(int argc, char **argv) {
   if(argc!=0 && argc!=6) return COMMAND_SYNOPSIS;
   Inrimage* img=getClient();
   if(argc==6) img->setRegion(atoi(argv[0]),atoi(argv[1]),atoi(argv[2]),atoi(argv[3]),atoi(argv[4]),atoi(argv[5]));
   unsigned int xmin,ymin,zmin,xmax,ymax,zmax;
   img->getWorkRegion(xmin,ymin,zmin,xmax,ymax,zmax);
   setResult("%u %u %u %u %u %u",xmin,ymin,zmin,xmax,ymax,zmax);
   return TCL_OK;
}



void TclInrimage::read(const TclString &filename, Scene2D *scn) {
  (void) new TclInrimage(typeid(TclInrimage),
			 ((TclScene2D *) scn->getInterface())->getModule(), filename);
}


 
int TclInrimage::worldToVoxelCmd(int argc, char **argv) {
	if (argc!=3) 
		return COMMAND_SYNOPSIS;
	double px,py,pz;
	px=(double) atof(argv[0]);
	py=(double) atof(argv[1]);
	pz=(double) atof(argv[2]);
    Vec3<double> pos(px,py,pz);
	Vec3<double> res;
	getClient()->map3DWorldToVoxelCoordinates(pos,res);
	setResult("%f %f %f",res.getX(),res.getY(),res.getZ());
	return TCL_OK;
}
int TclInrimage::voxelToWorldCmd(int argc, char **argv) {
	if (argc!=3) 
		return COMMAND_SYNOPSIS;
	double px,py,pz;
	px=(double) atof(argv[0]);
	py=(double) atof(argv[1]);
	pz=(double) atof(argv[2]);
    Vec3<double> pos(px,py,pz);
	Vec3<double> res;
	getClient()->map3DVoxelToWorldCoordinates(pos,res);
	setResult("%f %f %f",res.getX(),res.getY(),res.getZ());
	return TCL_OK;
}
int TclInrimage::evalCmd(int argc, char **argv) {
  TclInrimage *op;
  Var_list *op_list=NULL,*w_op_list;
  
  int c,i;
  int nmbr=0;   // number of images
  int ref_nmbr; // inr reference number
  char str_func[512],var_name[512],h_name[512],*str_p;
 
  /* check if function string exists */
  if (argc!=1){
    return COMMAND_SYNOPSIS;
  }
  
  /* find image names in string */
  strcpy(str_func,argv[0]);
  str_p=str_func;
  nmbr=0;
  /* check for inr */
  while((str_p=strstr(str_p,"inr"))!=NULL){
    c=3;
    /*check for number of image */
    while(isdigit(str_p[c])){c++;}
    strcpy(h_name,str_p+3);
    h_name[c]='\0';
    ref_nmbr=atoi(h_name);
    strcpy(var_name,str_p);
    var_name[c]='\0';
    /* try to get pointer to inrimage */
    op = TclInrimage::getObject(ref_nmbr);
    if(!op) {
      setResult("Image (%s) does not exist",var_name);
      return TCL_ERROR;
    }
    /* check sizes of images */
    if((op)->getX() != getX() || (op)->getY() != getY() ||
       (op)->getZ() != getZ()) {
      setResult("%s%i and %s%i size differ", 
		   getObjectName(), getRef(),
		   getObjectName(), (op)->getRef());
      return TCL_ERROR;
    }
    /* fill linked list of Var_list object (fparser.h) */
    w_op_list= new (Var_list);
    w_op_list->next=op_list;
    op_list=w_op_list;
    w_op_list->im=(void*)((op)->getClient());
    strcpy(w_op_list->name,var_name);
    nmbr++;
    str_p++;
  }
  
  /* evaluate function string */
  getClient()->eval(argv[0],nmbr,op_list);
  changeAndRedraw();
  setResult("%s%u evaluated %s", getObjectName(), getRef(),
	    (char *) argv[0]);
  /* TODO delete linked list w_op_list */
  w_op_list=op_list;
  for(i=0;i<nmbr;i++){
    op_list=w_op_list->next;
    delete(w_op_list);
    w_op_list=op_list;
  }
  return TCL_OK;
}


  int TclInrimage::reformatCmd(int argc, char *argv[]) {
/*    
    if (argc != 6) return COMMAND_SYNOPSIS;
    
    char direction;
    int orix=atoi(argv[1]), oriy=atoi(argv[2]), endx=atoi(argv[3]), endy=atoi(argv[4]);

    Matrix4x4<double> *mat = (Matrix4x4<double> *)NULL;

    TclSliceCamera *cam = TclSliceCamera::getObject(argc, argv, 0);
    if (!cam) return TCL_ERROR;

    if(sscanf(argv[5], "%c", &direction) != 1) return COMMAND_SYNOPSIS;

    // les voxels
    unsigned int  vorix, voriy, voriz, vendx, vendy, vendz;

    cam->getClient()->getBackground()->getImageCoordinates(orix, oriy,
							   vorix, voriy,voriz);
    cam->getClient()->getBackground()->getImageCoordinates(endx, endy,
							   vendx, vendy,vendz);
    
    TclInrimage *res;
    if (direction == 'h')
      switch (cam->getClient()->getGeometry()) {
      case SliceCamera::SLICE_X:
	res = new TclInrimage(getModule(), typeid(TclInrimage), reformat(getClient(), vorix, voriy, voriz, vendx, vendy, vendz, Inrimage::SLICE_X, mat, false));
	break;
      case SliceCamera::SLICE_Y:
	res = new TclInrimage(getModule(), typeid(TclInrimage), reformat(getClient(), vorix, voriy, voriz, vendx, vendy, vendz, Inrimage::SLICE_Y, mat, false));
	break;
      case SliceCamera::SLICE_Z:
	res = new TclInrimage(getModule(), typeid(TclInrimage), reformat(getClient(), vorix, voriy, voriz, vendx, vendy, vendz, Inrimage::SLICE_Z, mat, false));
	break;
      default:
	CERR << " Invalid slicecam orientation in reformatCmd " ;
	return TCL_ERROR;
      }
    else if (direction == 'v')
      switch (cam->getClient()->getGeometry()) {
      case SliceCamera::SLICE_X:
	res = new TclInrimage(getModule(), typeid(TclInrimage), reformat(getClient(), vorix, voriy, voriz, vendx, vendy, vendz, Inrimage::SLICE_X, mat, true));
	break;
      case SliceCamera::SLICE_Y:
	res = new TclInrimage(getModule(), typeid(TclInrimage), reformat(getClient(),  vorix, voriy, voriz, vendx, vendy, vendz, Inrimage::SLICE_Y, mat, true));
	break;
      case SliceCamera::SLICE_Z:
	res = new TclInrimage(getModule(), typeid(TclInrimage), reformat(getClient(),  vorix, voriy, voriz, vendx, vendy, vendz, Inrimage::SLICE_Z, mat, true));
	break;
      default:
	CERR << " Invalid slicecam orientation in reformatCmd " ;
	return TCL_ERROR;
      }
    else {
      CERR << " Invalid direction horizontal (h) or vertical (v)  in reformatCmd " ;
      return TCL_ERROR;
    }
    
    TclMatrix4D *tmat = new TclMatrix4D(getModule(), *mat);

    setResult("%s%u", tmat->getObjectName(), tmat->getRef() );
    setResult("%s%u", res->getObjectName(), res->getRef() );
	*/
    return TCL_OK;
  }

void TclInrimage::createCommands(TclModule *m){
  TclObjectKind *kind;

 kind = new Scene2DItem(m, typeid(TclInrimage), "inr","3D Image",
			 &TclInrimage::create, &TclInrimage::list,
			 TclScene2D::FILE | TclScene2D::OTHER,
			 "#INRIMAGE-4", ".inr,.inr.gz .hdr .dim", &TclInrimage::read, 0,
			 "<filename=%s [-interlace %b] [-fileInterlace %b]| x %u y %u z [type <8bits | 16bits | s16bits | 32bits | s32bits | 64bits | s64bits | float | double | rgb | rgba | floatVector %u | doubleVector %u>] >");
  kind->setModule(m);
  m->newCommand("inr%u", kind->getOptions(), "inrimage management command");
  kind->newOption("-destroy", (TclObject::OPTION) &TclInrimage::destroyCmd, 0,
		  "destroy inrimage");
  kind->newOption("-name", (TclObject::OPTION) &TclInrimage::nameCmd, "%s",
		  "get/set image name");
  kind->newOption("-size", (TclObject::OPTION) &TclInrimage::sizeCmd, 0,
		  "get image size");
  kind->newOption("-save", (TclObject::OPTION) &TclInrimage::saveCmd, "%s",
		  "save image (using image name if no new name is given)");
  kind->newOption("-copy", (TclObject::OPTION) &TclInrimage::copyCmd, 0,
		  "duplicate image and return new image id");
  kind->newOption("-redraw", (TclObject::OPTION) &TclInrimage::redrawCmd, 0,
		  "force inrimage to be redisplayed in any viewer it appears");
  kind->newOption("-vectorial", (TclObject::OPTION) &TclInrimage::vectorialCmd, 0,
		  "return true if inrimage type is vectorial, false otherwise");
  kind->newOption("-floating", (TclObject::OPTION) &TclInrimage::floatingCmd, 0,
		  "return true if inrimage type is floating point, false otherwise");
  kind->newOption("-glbounds", (TclObject::OPTION) &TclInrimage::glboundsCmd,
		  "[min %%v] [max %%v]",
		  "get/set inrimage min and max grey level values");
  kind->newOption("-crop", (TclObject::OPTION) &TclInrimage::cropCmd,
		  "<org %u %u %u | center %u %u %u> <end %u %u %u | size %u %u %u>",
		  "crop inrimage with given box (from org to end or centered on center and with given size)");
  kind->newOption("-isolateRegion", (TclObject::OPTION) &TclInrimage::isolateRegionCmd,
		  "<org %u %u %u | center %u %u %u> <end %u %u %u | size %u %u %u>",
		  "zero voxel outside region  (from org to end or centered on center and with given size)"); 
  kind->newOption("-zeroRegion", (TclObject::OPTION) &TclInrimage::zeroRegionCmd,
		  "<org %u %u %u | center %u %u %u> <end %u %u %u | size %u %u %u>",
		  "zero voxel inside region  (from org to end or centered on center and with given size)");
  kind->newOption("-voxel", (TclObject::OPTION) &TclInrimage::voxelCmd, "[%f %f %f]",
		  "get/set inrimage voxels size");
  kind->newOption("-flip", (TclObject::OPTION) &TclInrimage::flipCmd, "[X|Y|Z]",
		  "flip along given direction");
  kind->newOption("-histogram", (TclObject::OPTION) &TclInrimage::histogramCmd, 0,
		  "returns the image histogram ");
  kind->newOption("-maskedHistogram", (TclObject::OPTION) &TclInrimage::maskedHistogramCmd,
		  "<%s>",
		  "returns the image histogram within a given mask");
  kind->newOption("-switch", (TclObject::OPTION) &TclInrimage::switchCmd,
		  "<X|Y|Z><X|Y|Z>",
		  "switch two directions");
  kind->newOption("-mel", (TclObject::OPTION) &TclInrimage::melCmd,
		  "inr%u [ixo %u] [iyo %u] [izo %u] [ixi %i] [iyi %i] [izi %i] [idx %u] [idy %u] [idz %u]",
		  "mel two inrimages (ixi, iyi, izi is input image block corner, and idx, idy, idz its size; ixo, iyo, izo is output image block corner).");
  /*  kind->newOption("-xcrop", (TclObject::OPTION) &TclInrimage::xcropCmd,
      "<init | drop | done>", "allow interactive crop");*/
  kind->newOption("-type", (TclObject::OPTION) &TclInrimage::typeCmd,
		  "[[cast|scale] <8bits|16bits|s16bits|32bits|s32bits|64bits|s64bits|float|double|rgb|rgba|floatVector|doubleVector>]",
		  "get/set inrimage type");
  kind->newOption("-arith", (TclObject::OPTION) &TclInrimage::arithCmd,
		  "<inv | abs | <set | add | sub | mul | div> <%f | inr%u> >",
		  "Perform given arithmetic operation. Either a constant is use with each voxel or another same size inrimage is used. Zero division returns 0. Overflows are no handled.");
  kind->newOption("-logic", (TclObject::OPTION) &TclInrimage::logicCmd,
		  "<<and | or | xor> inr%u | not>",
		  "apply given logical operator");
  
  kind->newOption("-eval", (TclObject::OPTION) &TclInrimage::evalCmd,
		  "function string (do not use spaces)",
		  "Evaluates expression in function string.Supports the following arithmetic operators in the following order: ^(power),/(division) or *(multiplication) and +(addition) or -(subtraction). Operators with the same priority are processed from left to right. All other operators have lowest priority and therefor processed from left to right as well. Special operators: |<(minimum), |>(maximum). Logical Operators(returns 1=true,0=false):  <=(Smaller or equal), >=(Larger or equal), ==(equal), >(larger), <(smaller), &&(and), ||(or). Standard functions (angles are in radians): abs(), exp(), ceil()(rounds upwards), floor()(rounds downwards), log(), sqrt(), sin(), cos(), tan(), asin(), acos(), atan(), sinh(), cosh(), tanh(). Global functions (returns a scalar and calculates value based on the expression given as input to the function)): mean, min(minimum value), max(maximum value), med(median), std(estimated standard deviation), var(estimated variance). Variables: x,y,z (pixel positions), inrX(X can be any number of a valid image). Examples: inr3 -eval inr1|<inr2  (each pixel of inr3 contains the maximum of the corresponding pixels in inr1 and inr2), inr3 -eval inr1 -eval (x-64)^2+(y-64)^2+(z-64)^2<200(creates a sphere), inr3 -eval inr1-mean(inr1)(set mean to zero), inr3 -eval (inr1-min(inr1))/(max(inr1)-min(inr1))*100(normalizes image between 0 and 100), inr3 -eval med(sqrt(inr1^2/inr2)/(x+y+z)^3)*inr2(I don't know).");

  kind->newOption("-windowing", (TclObject::OPTION) &TclInrimage::windowingCmd,
		  "[min %v] [max %v]",
		  "Creates an unsigned char image after Windowing the image between min and max values");

//   kind->newOption("-chamfer", (TclObject::OPTION) &TclInrimage::chamferCmd,
// 		  "[2D|3D] [thres %f] [malandain3x3x3|chamfer5x5x5] ",
// 		  "Compute a chamfer map with a threshold an a type of distance. ");

 kind->newOption("-chamfer", (TclObject::OPTION) &TclInrimage::chamferCmd,
		 "[2D|3D] [size %i] [location %i] [exactmap %i] [propagate %i]",
		 "Compute a chamfer map. ");

 kind->newOption("-blockMean", (TclObject::OPTION) &TclInrimage::blockMeanCmd,
		  "[sizeXY %u] [sizeZ %u]",
		  "output image with average block intensity");
 kind->newOption("-blockStandardDeviation", (TclObject::OPTION) &TclInrimage::blockStandardDeviationCmd,
		  "[sizeXY %u] [sizeZ %u]",
		  "output image with intensity block standard deviation");
 kind->newOption("-blockMedian", (TclObject::OPTION) &TclInrimage::blockMedianCmd,
		  "[sizeXY %u] [sizeZ %u]",
		  "output image with intensity block median value");
  kind->newOption("-maskCorrelation", (TclObject::OPTION) &TclInrimage::maskCorrelationCmd,
		  "[mask inr%u] [method ssd|sad|cr]",
		  "output image corresponding to the correlation between a block around each voxel and a mask. The mask image must have the same type and voxel size as the input image. Correlation methods can be  SSD (Sum of Square Differences), SAD (Sum of Absolute Differences) and CR (Correlation Ratio)");
 kind->newOption("-threshold", (TclObject::OPTION) &TclInrimage::thresholdCmd,
		  "[min %v] [max %v] [bg %v] [fg %v]",
		  "Threshold inrimage between min and max values. Voxel between min an max included are set to fg value. Other voxels are set to bg value.");
  kind->newOption("-cropHistogramm", (TclObject::OPTION) &TclInrimage::cropHistogrammCmd,
		  "[min %v] [max %v] [bg %v]",
		  "Threshold inrimage between min and max values. Voxel between min an max included are not modified. Other voxels are set to bg value.");
  kind->newOption("-noise", (TclObject::OPTION) &TclInrimage::noiseCmd,
		  "[white mean=%f|gaussian|us] [dev %f] [org %u %u %u | center %u %u %u] [end %u %u %u | size %u %u %u]",
		  "Creates a noisy box inside an inrimage. The output is either a white gaussian noise with given mean (white), a gaussian noise added to image (gaussian, default) or an ultrasound like multiplicative noise (us)");
  kind->newOption("-connexe", (TclObject::OPTION) &TclInrimage::connexeCmd,
		  "[con <4|6|8|18|26>] [tcc %i] [ncc %u|max] [lt %v ht %v] [hyster [tcc_ht %i] ] [output <binary|label>] ",
		  "Extract connexe components from an inrimage. Binarize input image using low threshold (lt) and high threshold (ht) if any (hysteresis thresholding). Extract connexe components with given (con) connectivity. tcc is the minimal size of cc kept, ncc is the number of cc kept (sorted with respect to their size), max selects only largest cc. The output type indicates whether to return a binary image or a labelled cc image.");
  kind->newOption("-probabilityMap", (TclObject::OPTION) &TclInrimage::probabilityMapCmd,
		  "proba%u",
		  "Returns an image corresponding to the probability of each voxel value");
   kind->newOption("-line3D", (TclObject::OPTION) &TclInrimage::line3DCmd,
		  "%f %f %f %f %f %f %lf",
		  "Draws a 3D line in an image. The first 3 float is the position of the first vertex. The next 3 floats are the position of the second vertex. The last float is the color of the foreground.");
  kind->newOption("-fill", (TclObject::OPTION) &TclInrimage::fillCmd,
		  "bg %v fg %v [2D|3D]",
		  "Fill-in 2D or 3D regions of inrimage. Contours with fg color on a bg background are filled with fg value.");
  kind->newOption("-morpho", (TclObject::OPTION) &TclInrimage::morphoCmd,
		  "[dil|ero|close|open|tophat|invtophat|grad|lowgrad|uppgrad|median] [ [R %u] | [con <4|6|8|10|18|26>] | [struct %s] ] [i %u] [binary | greylevel]" ,
		  "Binarize inrimage and apply given morphological operator. ");
  kind->newOption("-gradient", (TclObject::OPTION) &TclInrimage::gradientCmd,
		  "[sigma %f] [2D|3D] [rec [marta|gaussian|deriche] | sobel] [dir <vectors|sample>] [max] [lt %f]",
		  "Compute inrimage gradient either using recursive filters (rec with Marta, Deriche gaussian or Deriche step-edge approximation) or sobel masks (sobel). sigma represents filter standard deviation (rec) or filter size (sobel). A low thresholding is applied on resulting gradient image if lt option is specified. Only gradient extrema along gradient direction are kept with max. The dir option allow to get gradient vectors (dir vectors) or to generate a gradient directions map (as a 8-bits image whose voxel value index a direction vector among 180 vectors sampling all directions, dir sample).");
  kind->newOption("-convolute", (TclObject::OPTION) &TclInrimage::convoluteCmd,
		  "[filterSize %u] <gauss [2D|3D]| smoothX | smoothY | smoothZ |derX | derY | derZ | der2X | der2Y | der2Z>",
		  "Convolute inrimage with a filter of given size. Available filters are gaussian (X or Y or Z or 2D or 3D), first derivative X, Y or Z-wise and second derivative X, Y or Z-wise.");
  kind->newOption("-curvature", (TclObject::OPTION) &TclInrimage::curvatureCmd,
		  0, "compute 2D inrimage curvature: (Ixx*Iy^2 - 2*Ix*Iy*Ixy + Iyy*Ix^2) / (Ix^2 + Iy^2)^(3/2)");
  kind->newOption("-recfilter", (TclObject::OPTION) &TclInrimage::recfilterCmd,
		  "[sigma %f] [marta|gaussian|deriche] <gauss [2D|3D] | smoothX | smoothY | smoothZ | derX | derY | derZ | der2X | der2Y | der2Z>",
		  "Apply recursive filter on inrimage (with Marta, Deriche gaussian or Deriche step-edge approximation). Available filters are gaussian (X or Y or Z or 2D or 3D), first derivative X, Y or Z-wise and second derivative X, Y or Z-wise.");
  kind->newOption("-laplacian", (TclObject::OPTION) &TclInrimage::laplacianCmd,
		  "[sigma %f] [marta|gaussian|deriche]",
		  "Compute inrimage laplacian using recursive filters.");
  kind->newOption("-split", (TclObject::OPTION) &TclInrimage::splitCmd, 0,
		  "split a vectorial inrimage in as many scalar inrimages as vector dimensions and return list of created inrimages");
  kind->newOption("-createSlicePhoto", (TclObject::OPTION) &TclInrimage::createSlicePhotoCmd,
		  "[X|Y|Z]  sliceNumber=%u scaleFactor=%f useVoxelSize=%b minWindow=%u  maxWindow=%u interpolate=%b [spect|pet|hotiron|jet|threshold|affine] [-create imageName=%s]",
		  "creates a photo (that can be drawn in a canvas widget) as a slice of a tridimensional image. Either rovides the image name with the \"-image\" option or creates a new image ");
  kind->newOption("-components", (TclObject::OPTION) &TclInrimage::componentsCmd,
		  "inr%u+",
		  "Set vectorial components of a vectorial inrimage with given scalar inrimages. There should be as many components as inrimage vectorial dimension.");
  kind->newOption("-dft", (TclObject::OPTION) &TclInrimage::dftCmd,
		  "[2D|3D]",
		  "Compute DFT (Discrete Fourier Transform) of image using the FFT algorithm. If image is scalar, the direct transform is computed. If image is vectorial and vdim is 2, the inverse transform is computed.");
  kind->newOption("-undersample", (TclObject::OPTION) &TclInrimage::undersampleCmd,
		  "<step=%u | X stepX=%u Y stepY=%u Z stepZ=%u>",
		  "undersample image with given step in X, Y and Z direction, or with stepX step in X direction, stepY step in Y direction, stepZ step in Z direction.");
  kind->newOption("-oversample", (TclObject::OPTION) &TclInrimage::oversampleCmd,
		  "<xdim=%u ydim=%u zdim=%u>",
		  "Oversample an image so that its new size is xdim*ydim*zdim. New size should be greater thant oler one.");
  kind->newOption("-voxelValue", (TclObject::OPTION) &TclInrimage::voxelValueCmd,
		  "<x=%u y=%u z=%u (val=%f)>",
		  "Get value of voxel (x,y,z) or set value of voxel (x,y,z) to value val");
  kind->newOption("-voxelInterpolatedValue", (TclObject::OPTION) &TclInrimage::voxelInterpolatedValueCmd,
		  "<x=%u y=%u z=%u>",
		  "Get Interpolated value at position (x,y,z) based on the chosen interpolator/extrapolator");

  kind->newOption("-transform", (TclObject::OPTION) &TclInrimage::transformCmd,
		  "[nearest|trilinear|cspline] [2D|3D] <-mat %s> [-def inr%u inr%u inr%u] [-inv] [-dim %u %u %u] [-vsize %lf %lf %lf] [-world]",
		  "Resample an image with a mat4 and a deformation field. Resampling may be nearest, trilinear, cspline and in 2D or 3D. The matrix argument transforms a point in voxel coordinates of the result image into a point in voxel coordinates of the input image. The -inv option inverse the matrix. The -world option modifies the matrix if it is provided in world coordinates. If you have a resampling matrix issued from a registration, you well need this option. You can use deformation field given as inr%u. The field transform apoint of the result image into a point of the input image. You can use a matrix and a field at the same time. In this case, given a point in the result image, we apply the field, then the matrix to get the corresponding point intensity in the input mage. You can also specify the dimension of the result image and its voxel size (by default it is using the same voxel size as the input image). ");

  kind->newOption("-resample", (TclObject::OPTION) &TclInrimage::resampleCmd,
		  "<xdim=%u ydim=%u zdim=%u>",
		  "Resample an image so that its new size is xdim*ydim*zdim. Under sampled images are approximated by picking a trilinearly interpolated value in the originale image (i.e. there is no voxel averaging computation).");
  kind->newOption("-interpolation", (TclObject::OPTION) &TclInrimage::interpCmd,
		  "[nearest|trilin|trilinGeneric]",
		  "get or set interpolation method");
  kind->newOption("-extrapolation", (TclObject::OPTION) &TclInrimage::extrapCmd,
		  "[nearest|tozero]",
		  "get or set extrapolation method");
  kind->newOption("-region", (TclObject::OPTION) &TclInrimage::regionCmd,
		  "[%d %d %d %d %d %d]",
		  "get or set region in which operations should be performed method. not used by all operations. negative values mean lower or higher bound, as appropriate.");
  kind->newOption("-anisotropicDiffusion", (TclObject::OPTION) &TclInrimage::anisotropicDiffusionCmd,
		  "[%f %f %f %f %f %f %b %f %s %b %s %s ]",
		  "compute the anisotropic diffusion of the image. The first parameter is the time step dt (dt>0). The second parameter is the diffusion time tmax with tmax>=dt. Indeed, (tmax/dt) is the number of iterations. The third parameter is the quantile quant that set the gradient threshold (0<quant<100) in the x direction. The fourth parameter is the quantile quant that set the gradient threshold (0<quant<100) in the y direction. The fifth parameter is the quantile quant that set the gradient threshold (0<quant<100) in the z direction. The sixth parameter allows to have a constant threshold, if k < 0 computation of the quantile else use k instead of the quantile.The seventh paramater is a boolean that indicates if we smooth the gradient information in the diffusity matrix. The eighth parameter is the standard deviation of smoothing kernel. The ninth parameter indicates the diffusion function df used. Currently, df is either \"aubert\" or \"perona\"or \"perona_2\" or \"weickert\" or \"green\". The tenth parameter indicates wether the temporary buffers are saved to disk to save memory (for big images) or not. The eleventh parameter is the diffusion type. Currently, td is either \"scalar\", \"matrix\", \"krissian\" or \"matkrissian\". The last paramater is the numerical scheme : \"explicite\" or \"aos\".");
  kind->newOption("-reformat", (TclObject::OPTION) &TclInrimage::reformatCmd,
		  "slicecam%u %d %d %d %d %c",
		  "reformat an image using clipping plane defined by a camera and two points of the canvas. We choose the vertical (v) or horizontal (h) axis to align the image and the geomtry in which this axis (v or h) has been defined. ");
  kind->newOption("-globalTransformation", (TclObject::OPTION) &TclInrimage::globalTransformationCmd,
		  "[ translation | rotation | translation %f %f %f | addTranslation %f %f %f | rotation %f %f %f  | addRotation %f %f %f | [ set | compose ] %s | [set | compose ] %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f (value of the 4x4 translation + rotation matrix to compose the image with)]",
		  "get or set the transformation matrix that transforms voxel based coordinates into world coordinates. It is possible to get or set the rotation or translation part of the transformation. For the rotation, one has to specify a rotation vector.");
  kind->newOption("-globalTransformationToVector", (TclObject::OPTION) &TclInrimage::globalTransformationToVectorCmd,"%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f (value of the 4x4 translation + rotation matrix to compose the image vectors with)]","apply a transformation to the vectors of the vectorial image (dimension=3)");
  kind->newOption("-worldToVoxel", (TclObject::OPTION) &TclInrimage::worldToVoxelCmd,"%f %f %f (value of a 3D vector)",
	  "returns the 3 coordinates (float) of the input in voxel coordinates");
  kind->newOption("-voxelToWorld", (TclObject::OPTION) &TclInrimage::voxelToWorldCmd,"%f %f %f (value of a 3D vector)",
	  "returns the 3 coordinates (float) of the input in world coordinates");

  m->newCommand("imageFormat", TclInrimage::imageFormatCmd,
		"input | output",
		"returns the available file format for reading or writing an image. For each format gives the \"real name\" and the file extensions separated with a comma");
  m->newCommand("view", TclInrimage::view,
		"inr%u [scn2D%u | inr%u]", 
		"Create viewers for all given images. view inr%u creates 3 viewers in X, Y and Z in a predefined window. view inr%u <X|Y|Z|alpha> ... creates a viewer in given direction. Option are: -in <view%u|%w>: view image in an existing view or an existing TCL frame; -slice <no|create|scale%u|%w>: do not use/create/use existing scale/use TCL scale object for slice scale; -glmin <no|create|scale%u|%w>: do not use/create/use existing scale/use TCL scale object for min image grey level threshold; -glmax <no|create|scale%u|%w>: do not use/create/use existing scale/use TCL scale object for max image grey level threshold; -motion <no|create|%w>: do not use/create/use existing Tcl label for mouse motion tracking. If a single viewer is created, slicecam%u is returned. If more than one viewer is created, a multiview%u managing all those viewers is returned."); 

}
} // namespace yav



