#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#pragma implementation "inrimage/TclInrimage4D.h"
#include <inrimage/TclInrimage4D.h>
//#include <graphics2D/TclTimeSliceCamera.h>
//#include <inrimage/Inrimage4DView.h>
//#include <inrimage/TclInrimage4DBackground.h>
#include <inrimage/TclInrimage.h>
#include <inrimage/recursiveFilters.h>
#include <inrimage/anisotropicDiffusion.h>


namespace yav {


TclInrimage4D::TclInrimage4D(TclModule *m, const char *name,
			     const std::type_info& info, Inrimage4D *data)
  throw(GeneralFormat::FileNotFound,
	Inrimage::ImageNotFound, Inrimage4D::ReadError)
  : TclObject(info, m), 
    Owner<Inrimage4D> (data ? data : new Inrimage4D(name)) {

  init(m);

  done(typeid(TclInrimage4D));
}


TclInrimage4D::TclInrimage4D(TclModule *m, std::list<Inrimage *> inrs)
  : TclObject(typeid(TclInrimage4D),m),
    Owner<Inrimage4D>(new Inrimage4D(inrs)){

  init(m);

  std::vector<Inrimage *>::iterator it;
  for(it = getClient()->inrs.begin(); it != getClient()->inrs.end(); it++) {
    if (!TclInrimage::getInterface(*it))
      new TclInrimage(m, typeid(TclInrimage), (Inrimage *) *it);
  }


  done(typeid(TclInrimage4D));
}



TclInrimage4D::TclInrimage4D(TclModule *m, 
			     const std::type_info& info, Inrimage4D *data)
  : TclObject(info, m),
    Owner<Inrimage4D>(data){

  init(m);
  done(typeid(TclInrimage4D));
}


void TclInrimage4D::init(TclModule *m) {
  std::vector<Inrimage *>::iterator iti;
  for(iti = getClient()->inrs.begin(); iti != getClient()->inrs.end(); iti++) {
    if (! TclInrimage::getInterface(*iti))
      new TclInrimage(m, typeid(TclInrimage),  *iti);
  }
}


TclInrimage4D::~TclInrimage4D() {
  deleted(typeid(TclInrimage4D));
   if (getClient()) {
     Inrimage4D *i=getClient();
     pointer=0;
    delete i;
  }
}



int TclInrimage4D::create(TclModule *m, int argc, char **argv) {
  std::list<Inrimage *> sinrs;
  int j= -1;
  TclInrimage4D *inr4;

  if(argc < 2) return COMMAND_SYNOPSIS;
  if (argc==2)
    j=1;
  else
    for(int i = 1; i < argc; i++) {
      TclInrimage *inr= TclInrimage::getObject(argc, argv, i);
      if (!inr)
	{
	  m->setResult("%s is not an image  name",argv[i]);
	  return TCL_ERROR;
	}
      else
	{
	  sinrs.push_back(inr->getClient());
	}
    }

  
  try {
    if (sinrs.size()==0)
      {
	inr4 = new TclInrimage4D(m, argv[1]);
      }
    else
      {
	inr4 = new TclInrimage4D(m, sinrs);
      }
  }
  catch(GeneralFormat::FileNotFound &) {
    m->setResult("file not found");
    return TCL_ERROR;
  }
  catch(Inrimage::ImageNotFound &) {
    m->setResult("3D inrimage file not found");
    return TCL_ERROR;    
  }
  catch(Inrimage4D::ReadError &) {
    m->setResult("3D inrimage file corrupted");
    return TCL_ERROR;
  }
  catch(Inrimage4D::EmptyInrimageList &) {
    m->setResult("3D inrimage list is empty");
    return TCL_ERROR;
  }
  catch(Inrimage4D::DifferentImageType &) {
    m->setResult("images are of different image type");
    return TCL_ERROR;
  }
  catch(Inrimage4D::DifferentImageSize &) {
    m->setResult("images are of different image size");
    return TCL_ERROR;
  }

  m->setResult("%s%u", TclInrimage4D::getTCLObjectName(), inr4->getRef());
  return TCL_OK;
}




TclInrimage4D *TclInrimage4D::getInterface(Inrimage4D *inr) {
  TclInrimage4D *tinr=0;

  std::list<ObjectPointer<Inrimage4D> *>::iterator itp;
  for(itp=inr->getFirstPointer();(tinr==0) &&(itp!=inr->getLastPointer());itp++) {
    TclInrimage4D *t = dynamic_cast<TclInrimage4D *>((*itp)->getOwner());
    if (t)
      tinr=t;
  }
  return tinr;
}


void TclInrimage4D::load(const TclString &filename, Scene2D *scene) {
  (void) new TclInrimage4D(((TclScene2D *)scene->getInterface())->getModule(), filename);
}


int TclInrimage4D::destroyCmd(int argc, char **argv) {
  unsigned int u = getRef();
  TclModule *m = getModule();
  if(argc > 0) return COMMAND_SYNOPSIS;

  delete this;
  m->setResult("%s%u does not exist any longer",
	       TclInrimage4D::getTCLObjectName(), u);
  return TCL_OK;
}


void TclInrimage4D::redraw(void) const {
#ifdef OLD_CODE
  std::list<ObjectPointer<Inrimage4D> *>::iterator itp;
  for(itp=getClient()->getFirstPointer();itp!=getClient()->getLastPointer();itp++)
    {
      Inrimage4DView *inrv=dynamic_cast<Inrimage4DView *>((*itp)->getOwner());
      if (inrv)
	{
	  std::list<Inrimage4DBackground *>::iterator itb=inrv->firstBackground();
	  for(;itb!=inrv->lastBackground();itb++)
	    {
	      (*itb)->redraw();
	    }
	}
    }
#endif
}

int TclInrimage4D::applyCmd(int (TclInrimage::*cmd)(int, char **),
			    int argc, char **argv) {
  std::vector<Inrimage *>::iterator itd = getClient()->inrs.begin();

    
  if(argc > 0) {
    int res = (TclInrimage::getInterface(*itd)->*cmd)(argc, argv);
    if(res != TCL_OK) return res;
    for(itd++; itd != getClient()->inrs.end(); itd++)
      (TclInrimage::getInterface(*itd)->*cmd)(argc, argv);
    redraw();
    itd = getClient()->inrs.begin();
    (TclInrimage::getInterface(*itd)->*cmd)(0, 0);
    return TCL_OK;
  }
  else return (TclInrimage::getInterface(*itd)->*cmd)(argc, argv);
}

int TclInrimage4D::typeCmd(int argc, char **argv) {
  return applyCmd(&TclInrimage::typeCmd, argc, argv);
}

int TclInrimage4D::glboundsCmd(int argc, char **argv) {
  return applyCmd(&TclInrimage::glboundsCmd, argc, argv);
}

int TclInrimage4D::imageFormatCmd(TclModule *m,int argc, char **argv) {
  if (argc!=2) 
    return COMMAND_SYNOPSIS;
  else if (!strcmp(argv[1],"input")) {
    
    TclString res,out;
    PTRIMAGE4D_FORMAT f;
    for(f=firstImage4DFormat();(f!=NULL);f=f->next) {
      out.sprintf("{%s %s}",f->realName,f->fileExtension);
      res+=out;
    }
    m->setResult(res);
    return TCL_OK;
  }
  else if (!strcmp(argv[1],"output")) { 
    TclString res,out;
    PTRIMAGE4D_FORMAT f;
    for(f=firstImage4DFormat();(f!=NULL);f=f->next) {
      if (f->writeImage) {
	out.sprintf("{%s %s}",f->realName,f->fileExtension);
	res+=out;
      }
    }
    m->setResult(res);
    return TCL_OK;

  }
  else {
    m->setResult("%s must be either input or output",argv[1]);
    return TCL_ERROR;
  }
}
int TclInrimage4D::voxelCmd(int argc, char **argv) {
   float vx, vy, vz,vt;
  
  if(argc > 4) return COMMAND_SYNOPSIS;

  if(argc > 0) {
    if(argc != 4) return COMMAND_SYNOPSIS;
    if(sscanf(argv[0], "%f", &vx) != 1) return COMMAND_SYNOPSIS;
    if(sscanf(argv[1], "%f", &vy) != 1) return COMMAND_SYNOPSIS;
    if(sscanf(argv[2], "%f", &vz) != 1) return COMMAND_SYNOPSIS;
    if(sscanf(argv[3], "%f", &vt) != 1) return COMMAND_SYNOPSIS;

    try {
      getClient()->setVoxelSize(vx, vy, vz,vt);
    }
    catch(BaseInrimage::InvalidVoxelSize&) {
      setResult("voxel size should be > 0");
      return TCL_ERROR;      
    }
    resizeVoxelsAndRedraw();
  }
  setResult("%f %f %f %f", getClient()->getVX(), getClient()->getVY(), 
	    getClient()->getVZ(),getClient()->getVT());
  return TCL_OK;
}

void TclInrimage4D::resizeAndRedraw(void) const {
#ifdef OLD_CODE
	std::list<ObjectPointer<Inrimage4D> *>::iterator itp;
  for(itp=getClient()->pointers.begin();itp!=getClient()->pointers.end();itp++)
    {
      Inrimage4DView *inrv=dynamic_cast<Inrimage4DView *>((*itp)->getOwner());
      if (inrv)
	{
	  std::list<Inrimage4DBackground *>::iterator itb=inrv->firstBackground();
	  for(;itb!=inrv->lastBackground();itb++)
	    {
	      (*itb)->setVisibleArea();
	    }
	}
    }
#endif
}


void TclInrimage4D::resizeVoxelsAndRedraw(void) const {
	#ifdef OLD_CODE
  std::list<ObjectPointer<Inrimage4D> *>::iterator itp;
  for(itp=getClient()->pointers.begin();itp!=getClient()->pointers.end();itp++)
    {
      Inrimage4DView *inrv=dynamic_cast<Inrimage4DView *>((*itp)->getOwner());
      if (inrv)
	{
	  std::list<Inrimage4DBackground *>::iterator itb=inrv->firstBackground();
	  for(;itb!=inrv->lastBackground();itb++)
	    {
	      ((TclInrimage4DBackground *)(*itb))->imageVoxelSizeChanged();
	    }
	}
    }
#endif
}




void TclInrimage4D::rescaleGLAndRedraw(void) const {
	#ifdef OLD_CODE

  std::list<ObjectPointer<Inrimage4D> *>::iterator itp;
  for(itp=getClient()->pointers.begin();itp!=getClient()->pointers.end();itp++)
    {
      Inrimage4DView *inrv=dynamic_cast<Inrimage4DView *>((*itp)->getOwner());
      if (inrv)
	{
	  std::list<Inrimage4DBackground *>::iterator itb=inrv->firstBackground();
	  for(;itb!=inrv->lastBackground();itb++)
	    {
	      ((TclInrimage4DBackground *)(*itb))->imageGLChanged();
	    }
	}
    }
#endif
}


void TclInrimage4D::changeAndRedraw(void) {
	#ifdef OLD_CODE

  std::list<ObjectPointer<Inrimage4D> *>::iterator itp;
  for(itp=getClient()->pointers.begin();itp!=getClient()->pointers.end();itp++)
    {
      Inrimage4DView *inrv=dynamic_cast<Inrimage4DView *>((*itp)->getOwner());
      if (inrv)
	{
	  std::list<Inrimage4DBackground *>::iterator itb=inrv->firstBackground();
	  for(;itb!=inrv->lastBackground();itb++)
	    {
	      ((TclInrimage4DBackground *)(*itb))->imageTypeChanged();
	    }
	}
    }
#endif
}

int TclInrimage4D::redrawCmd(int argc, char **argv) {
  if(argc > 0) return COMMAND_SYNOPSIS;
  redraw();
  setResult("%s%u redrawn", getObjectName(),
	    getRef());
  return TCL_OK;
}

int TclInrimage4D::slicecamsCmd(int argc, char **argv) {
	#ifdef OLD_CODE

  if(argc > 0) return COMMAND_SYNOPSIS;
  TclString s="";
  bool first=true;
  std::list<ObjectPointer<Inrimage4D> *>::iterator itp;
  TclTimeSliceCamera *slicecam;
  for(itp=getClient()->pointers.begin();itp!=getClient()->pointers.end();itp++)
    {
      Inrimage4DView *inrv=dynamic_cast<Inrimage4DView *>((*itp)->getOwner());
      if (inrv)
	{
	  std::list<Inrimage4DBackground *>::iterator itb=inrv->firstBackground();
	  for(;itb!=inrv->lastBackground();itb++)
	    {
	      slicecam= ((TclInrimage4DBackground *)(*itb))->getCamera()->getInterface();
	      if(!first) s+=" ";
	      else first=false;
	      s+=slicecam->name();
	    }
	}
    }
  
  setResult("%s",(char*)s);
#endif
  return TCL_OK;
}


int TclInrimage4D::gradientCmd(int argc, char **argv) {
  int i;
  bool max = false, compute2D = false, useSobel = false;
  float lt = 0.0;
  RECURSIVE_FILTER_TYPE filter = DERICHE_STEP_EDGE;
  Inrimage *output, *outdir;
  TclInrimage *gradientimg, *gradientdir;
  double sigma = 3.0;
  TclString name, name4D;
  GRADIENT_DIRECTION dir = GD_NONE;
  std::list<Inrimage *> grads;
  std::list<Inrimage *> dirs;

  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "sigma")) {
      if(++i < argc) {
	if(sscanf(argv[i], "%lf", &sigma) != 1) return COMMAND_SYNOPSIS;
      }
      else return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "2D")) compute2D = true;
    else if(!strcmp(argv[i], "3D")) compute2D = false;
    else if(!strcmp(argv[i], "dir")) {
      if(++i < argc) {
	if(!strcmp(argv[i], "vectors")) dir = GD_VECTORS;
	else if(!strcmp(argv[i], "sample")) dir = GD_SAMPLE;
	else return COMMAND_SYNOPSIS;
      }
      else return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "max")) max = true;
    else if(!strcmp(argv[i], "sobel")) useSobel = true;
    else if(!strcmp(argv[i], "rec")) useSobel = false;
    else if(!strcmp(argv[i], "lt")) {
      if(++i < argc) sscanf(argv[i], "%f", &lt);
      else return COMMAND_SYNOPSIS;
    }
    else if(!strcmp(argv[i], "gaussian")) filter = DERICHE_GAUSSIAN;
    else if(!strcmp(argv[i], "marta")) filter = MARTA_GAUSSIAN;
    else if(!strcmp(argv[i], "deriche")) filter = DERICHE_STEP_EDGE;
    else return COMMAND_SYNOPSIS;
  }

  std::vector<Inrimage *>::iterator it;
  for(it = getClient()->inrs.begin(); it != getClient()->inrs.end(); it++) {
    output = 0;
    outdir = 0;
    if(useSobel)
      sobel(*it, output, outdir, (int) sigma, dir, max, compute2D, lt);
    else
      gradient(*it, output, outdir, sigma, filter, dir, max,
		      compute2D, lt);

    gradientimg = new TclInrimage(getModule(), typeid(TclInrimage),
				  output);
    grads.push_back(output);

    name = getClient()->getName();
    if(name.getLength() > 3 &&
       TclString(name, name.getLength() - 3, name.getLength() - 1) == ".gz")
      name = TclString(name, 0, name.getLength() - 4);
    if(name.getLength() > 4 &&
       TclString(name, name.getLength() - 4, name.getLength() - 1) == ".inr")
      name = TclString(name, 0, name.getLength() - 5);
    gradientimg->getClient()->setName(name + ".gradient.gz");

    if(dir) {
      gradientdir = new TclInrimage(getModule(), typeid(TclInrimage), outdir);
      dirs.push_back(outdir);
      gradientdir->getClient()->setName(name + ".gradient_direction.gz");
    }
  }

  TclInrimage4D *grad4D = new TclInrimage4D(getModule(), grads);
  name4D = getClient()->getName();
  if(name4D.getLength() > 6 &&
     TclString(name4D, name4D.getLength() - 6, name4D.getLength() - 1)
     == ".inr4D")
    name4D = TclString(name4D, 0, name4D.getLength() - 7);
  grad4D->getClient()->setName(name4D + ".grad4D");
  if(dir) {
    TclInrimage4D *dir4D = new TclInrimage4D(getModule(), grads);
    dir4D->getClient()->setName(name4D + ".gradir4D");
    setResult("%s%u %s%u", getObjectName(), grad4D->getRef(),
	      getObjectName(), dir4D->getRef());
  }
  else
    setResult("%s%u", getObjectName(), grad4D->getRef());
  return TCL_OK;
}



int TclInrimage4D::nameCmd(int argc, char **argv) {
  if(argc > 1) return COMMAND_SYNOPSIS;
  if(argc == 1) getClient()->setName(argv[0]);
  setResult("%s", (char *) getClient()->getName());
  return TCL_OK;
}

int TclInrimage4D::saveCmd(int argc, char **argv) {
  if(argc > 1) return COMMAND_SYNOPSIS;
  if(argc > 0) {
    if(argv[0][0] != '\0') getClient()->setName(argv[0]);
    else {
      setResult("a non-empty name is expected");
      return TCL_ERROR;
    }    
  }
  getClient()->write(getClient()->getName());
  setResult("%s%u saved as %s", getObjectName(), getRef(),
	    (char *) getClient()->getName());
  return TCL_OK;
}

int TclInrimage4D::anisotropicDiffusionCmd(int argc, char **argv) {
   Inrimage4D * img=getClient();
   if(argc==7){
     float dt= (float)atof(argv[0]);
     if (dt<0.0)
       {
	 setResult("Time step %f must be positive",dt);
	 return TCL_ERROR;
       }
     float tmax = (float)atof(argv[1]);
     if (tmax<dt)
       {
	 setResult("Diffusion time %f must be greater than time step %f",tmax,dt);
	 return TCL_ERROR;
       }
     float quant = (float)atof(argv[2]);
     if ((quant<0.0)||(quant>100))
       {
	 setResult("Gradient quantile %f must be between 0 and 100",quant);
	 return TCL_ERROR;
       }
     bool lisse;
     if(!TclModule::boolValue(argv[3],lisse)) 
       {
	 setResult("%s must be true or false",argv[3]);
	 return TCL_ERROR;
       }
     float sigma = (float)atof(argv[4]);
     if (sigma<0.0)
       {
	 setResult("Standard Deviation %f must be greater than 0",sigma);
	 return TCL_ERROR;
       }
     DIFFUSION_FUNCTION df;
     if (!strcmp(argv[5],"perona"))
       df=PERONA_MALIK;
     else if (!strcmp(argv[5],"perona_2"))
       df=PERONA_MALIK_2;
     else if (!strcmp(argv[5],"perona_3"))
       df=PERONA_MALIK_3;
     else if (!strcmp(argv[5],"aubert"))
       df=AUBERT;
     else if (!strcmp(argv[5],"weickert"))
       df=WEICKERT; 
     else if (!strcmp(argv[5],"green"))
       df=GREEN; 
     else
       {
	 setResult("%s must be perona, aubert, weickert or green",argv[5]);
	 return TCL_ERROR;
       }
     bool disk;
     if(!TclModule::boolValue(argv[6],disk)) 
       {
	 setResult("%s must be true or false",argv[6]);
	 return TCL_ERROR;
       }
     
     std::list<Inrimage *> sinrs;
     unsigned int n = img->getTimeDimension(), i;
     
     for(i=0;i<n;i++) {
       sinrs.push_back(new Inrimage(*img->getInrimage(i)));
     }
     
     Inrimage4D * output = new Inrimage4D(sinrs);

#ifdef USE_MTL
     anisotropicDiffusion(img,output,dt,tmax,quant,lisse,sigma,df,disk);
#endif

     TclModule *m = getModule();
     TclInrimage4D * tclOutput = new TclInrimage4D(m,typeid(TclInrimage4D),output);
     setResult("%s%u", getObjectName(), tclOutput->getRef());
     
     return TCL_OK;
   }
   else return COMMAND_SYNOPSIS;
}

int TclInrimage4D::cropCmd(int argc, char **argv) {
  unsigned int x1, y1, z1, x2, y2, z2;
  int i;

  if(argc != 8) return COMMAND_SYNOPSIS;

  TclInrimage *  img = new TclInrimage(getModule(), typeid(TclInrimage), (getClient()->getFirstInrimage()));
  if(!TclInrimage::parseBoxOptions(getModule(), img, argc, argv,
		      x1, y1, z1, x2, y2, z2))
    return TCL_ERROR;

  // parse arguments
  for(i = 0; i < argc; i++) {
    if(!strcmp(argv[i], "org") || !strcmp(argv[i], "center")
       || !strcmp(argv[i], "end") || !strcmp(argv[i], "size")) i = i + 3;
    else return COMMAND_SYNOPSIS;
  }

  getClient()->crop(x1, y1, z1, x2, y2, z2);
  resizeAndRedraw();
  setResult("(%u, %u, %u) - (%u %u %u)", x1, y1, z1, x2, y2, z2);
  return TCL_OK;
}

}
