#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/TclInrimage4DBackground.h"
#pragma implementation "inrimage/background4DRedraw.h"
#include <inrimage/TclInrimage4DBackground.h>
#include <inrimage/TclColorMap.h>
#include <graphics2D/SliceScale.h>
#include <graphics2D/TclTimeSliceCamera.h>
#include <inrimage/TclInrimage4D.h>
#include "backgroundRedraw.h"
#include "background4DRedraw.h"
#include "tkwrapper.h"

namespace yav {



TclInrimage4DBackground::TclInrimage4DBackground(TclModule *m,
					     const std::type_info& type,
						 Inrimage4DView *_inrv, 
						 TimeSliceCamera *_cam,
						 MinGLScale *minScale,
						 MaxGLScale *maxScale) :  Inrimage4DBackground(_cam),Scalable(),TclObject(type,m),Owner<MinGLScale>(minScale), Owner<MaxGLScale>(maxScale),handle(0),block(0),glmin(0), glmax(0),  inrv(_inrv), blockGeom((TimeSliceCamera::TIME_GEOMETRY) -1) {

  Inrimage4D *inr=inrv->getInrimage();

  visible[0] = 0;
  visible[1] = 0;
  visible[2] = 0;
  visible[3] = inr->getX() - 1;
  visible[4] = inr->getY() - 1;
  visible[5] = inr->getZ() - 1;
  visible[6] = 0;
  visible[7] = inr->getTimeDimension()-1;
  
  voxelSize[0]=inr->getVX();
  voxelSize[1]=inr->getVY();
  voxelSize[2]=inr->getVZ();
  voxelSize[3]=inr->getVT();

  glmin = Voxel::voxelByType(inr->getType());
  glmax = Voxel::voxelByType(inr->getType());

  redColorMap=AffineColorMap;
  greenColorMap=AffineColorMap;
  blueColorMap=AffineColorMap;
  
}


TclInrimage4DBackground::TclInrimage4DBackground(TclModule *m,
						 Inrimage4DView *_inrv, 
						 TimeSliceCamera *_cam,
						 bool useMinGLScale,
						 MinGLScale *minScale,
						 bool useMaxGLScale,
						 MaxGLScale *maxScale, 
						 const char *color)
  :  Inrimage4DBackground(_cam),Scalable(),TclObject(typeid(TclInrimage4DBackground),m),Owner<MinGLScale>(minScale), Owner<MaxGLScale>(maxScale), handle(0),block(0),glmin(0), glmax(0),inrv(_inrv),blockGeom((TimeSliceCamera::TIME_GEOMETRY) -1) {
  done(typeid(TclInrimage4DBackground));
  
  Inrimage4D *inr=inrv->getInrimage();

  visible[0] = 0;
  visible[1] = 0;
  visible[2] = 0;
  visible[3] = inr->getX() - 1;
  visible[4] = inr->getY() - 1;
  visible[5] = inr->getZ() - 1;
  visible[6] = 0;
  visible[7] = inr->getTimeDimension()-1;
  
  voxelSize[0]=inr->getVX();
  voxelSize[1]=inr->getVY();
  voxelSize[2]=inr->getVZ();
  voxelSize[3]=inr->getVT();

  redColorMap=AffineColorMap;
  greenColorMap=AffineColorMap;
  blueColorMap=AffineColorMap;

  glmin = Voxel::voxelByType(inr->getType());
  glmax = Voxel::voxelByType(inr->getType());
  glBounds();
  
  slicecam->setBackground(this);
  init(useMinGLScale,minScale,useMaxGLScale,maxScale,color);
  if (slicecam->getScaleCamera())
    {
      ((SliceScale *)slicecam->getScaleCamera())->reset(slicecam,slicecam->sliceName(slicecam->getGeometry()));
    }
  if (getCamera()->getSecondScaleCamera())
    {
      ((SliceScale *)getCamera()->getSecondScaleCamera())->reset(slicecam,getCamera()->secondSliceName(getCamera()->getGeometry()));
    }
  if (getMinScale())
    getMinScale()->reset((Scalable *)this);
  if (getMaxScale())
    getMaxScale()->reset((Scalable *)this);
  //  inr->addSlice(this);
}


TclInrimage4DBackground::~TclInrimage4DBackground(void) {
  //  inrv->getInrimage()->deleteSlice(this);
  inrv->backgroundList.remove(this);
  getModule()->tickle("image delete %s",
		      (char *) TKimageID);
  if (slicecam)
    slicecam->backgroundDying();
  
}
MinGLScale *TclInrimage4DBackground::getMinScale(void) const {
  MinGLScale *sc=this->Owner<MinGLScale>::pointer();
  return(sc);
}
MaxGLScale *TclInrimage4DBackground::getMaxScale(void) const {
  MaxGLScale *sc=this->Owner<MaxGLScale>::pointer();
  return(sc);
}

void TclInrimage4DBackground::addMinGLScale(MinGLScale *min)  {
  Owner<MinGLScale>::pointer = min;
  *Owner<MinGLScale>::pointer += this;
}

void TclInrimage4DBackground::addMaxGLScale(MaxGLScale *max)  {
  Owner<MaxGLScale>::pointer = max;
  *Owner<MaxGLScale>::pointer += this;
}

int TclInrimage4DBackground::ignoreVoxelSizeCmd(int argc, char **argv) {
  bool IVS;

  if(argc > 1) return COMMAND_SYNOPSIS;
  if(argc == 1) {
    if(!TclModule::boolValue(argv[0], IVS)) return COMMAND_SYNOPSIS;
    setIVS(IVS);
  }
  setResult(TclModule::boolString(getIVS()));
  return TCL_OK;
}

void TclInrimage4DBackground::setIVS(const bool val){
  double _scale,m;
  if(ignoreVoxelSize != val) {
    ignoreVoxelSize = val;
    if(ignoreVoxelSize) _scale = 1.0;
    else {
      if(getVX() < getVY()) m = getVX();
      else m = getVY();
      if(getVZ() < m) m = getVZ();
      _scale = 1.0 / m;
    }
    slicecam->scaleCamera(_scale);
    slicecam->focusOnBackground();
    slicecam->updateSliceables();
  }
}
void TclInrimage4DBackground::init(bool useMinScale,
				   MinGLScale *minScale,
				   bool useMaxScale,
				   MaxGLScale *maxScale, 
				   const char *color)
{ 
  Inrimage4D *inr=inrv->getInrimage();

  // get inrimage photo handle
  tickle("image create photo -data \"inrimage4D %s%u %s%u\" -format inrimage4D",
	 TclObjectKind::getObjectName(typeid(TclInrimage4D)),
	 TclInrimage4D::getInterface(inr)->getRef(),
	 getCamera()->getInterface()->getObjectName(),
	 getCamera()->getInterface()->getRef());
  
  TKimageID = getModule()->getInterp()->result;
  
  // set photo image in canvas view
  tickle("[%s%u -widget] create image 0 0 -anchor nw -image %s -tag inrimage4D",
	 getCamera()->getInterface()->getObjectName(),
	 getCamera()->getInterface()->getRef(),
	 (char *) TKimageID);
  
  if(useMaxScale) {
    if(!Owner<MaxGLScale>::pointer) {
      TclString sname = getCamera()->getInterface()->getWindow() + ".max";
      getModule()->tickle("scale %s -orient horizontal -label {max grey level} -width 10 -font fixed -bg %s", (char *) sname, color);
      Owner<MaxGLScale>::pointer = new MaxGLScale(getModule(), sname,this);
      getModule()->tickle("pack %s -side bottom -expand true -fill x",
			  (char *) sname);
    } 
    else
      *Owner<MaxGLScale>::pointer += this;
  }
  
  if(useMinScale) {
    if(!Owner<MinGLScale>::pointer) {
      TclString sname =  getCamera()->getInterface()->getWindow() + ".min";
      getModule()->tickle("scale %s -orient horizontal -label {min grey level} -width 10 -font fixed -bg %s", (char *) sname, color);
      Owner<MinGLScale>::pointer = new MinGLScale(getModule(), sname,this);
      getModule()->tickle("pack %s -side bottom -expand true -fill x",
			  (char *) sname);
    } 
    else
      *Owner<MinGLScale>::pointer += this;
  }

}


Inrimage4D *TclInrimage4DBackground::getInrimage(void) const
{
  return(inrv->getInrimage());
}


void TclInrimage4DBackground::glBounds(void) {
  getInrimage()->getBounds(glmin, glmax);
}



void TclInrimage4DBackground::realloc(void) {
  if(!block) {
    // creates photo image data structure
    block = new Tk_PhotoImageBlock;
    block->pixelSize = 3;
    block->offset[0] = 0;
    block->offset[1] = 1;
    block->offset[2] = 2;
    block->pixelPtr = 0;
  }

  if(block->pixelPtr) delete[] block->pixelPtr;

  blockGeom = getCamera()->getGeometry();
  switch(blockGeom) {
  // create YZ block
  case TimeSliceCamera::SLICE_X:
    block->pixelPtr = new unsigned char[sy * sz * 3];
    block->width = sz;
    block->height = sy;
    block->pitch = 3 * sz;
    break;
  // create XZ block
  case TimeSliceCamera::SLICE_Y:
    block->pixelPtr = new unsigned char[sx * sz * 3];
    block->width = sx;
    block->height = sz;
    block->pitch = 3 * sx;
    break;
  // create XY block
  case TimeSliceCamera::SLICE_Z:
  case TimeSliceCamera::SLICE_ALPHA:
    block->pixelPtr = new unsigned char[sx * sy * 3];
    block->width = sx;
    block->height = sy;
    block->pitch = 3 * sx;
    break;
  // create XT block
  case TimeSliceCamera::SLICE_XT:
    block->pixelPtr = new unsigned char[sx * st * 3];
    block->width = sx;
    block->height = st;
    block->pitch = 3 * sx;
    break;
  // create YT block
  case TimeSliceCamera::SLICE_YT:
    block->pixelPtr = new unsigned char[sy * st * 3];
    block->width = sy;
    block->height = st;
    block->pitch = 3 * sy;
    break;
  // create ZT block
  case TimeSliceCamera::SLICE_ZT:
    block->pixelPtr = new unsigned char[sz * st * 3];
    block->width = sz;
    block->height = st;
    block->pitch = 3 * sz;
    break;
  default:
    block->pixelPtr = 0;
    block->width = block->height = block->pitch = 0;
    break;
  }
}

void TclInrimage4DBackground::voxelRatio(void) {
  double _rx=0.0, _ry=0.0, _rz=0.0,_rt=0.0;
  unsigned int _sx, _sy, _sz,_st;

  if(getIVS())
    _rx = _ry = _rz = _rt=slicecam->getScale();
  else if (getInrimage()){
    Inrimage4D *inr=getInrimage();
    _rx = inr->getVX() * slicecam->getScale();
    _ry = inr->getVY() * slicecam->getScale();
    _rz = inr->getVZ() * slicecam->getScale();
    _rt = inr->getVT() * slicecam->getScale();
  }

  _sx = (int) ((double) (visible[3] - visible[0] + 1) * _rx);
  _sy = (int) ((double) (visible[4] - visible[1] + 1) * _ry);
  _sz = (int) ((double) (visible[5] - visible[2] + 1) * _rz);
  _st = (int) ((double) (visible[7] - visible[6] + 1) * _rt);

  if(rx != _rx || ry != _ry || rz != _rz || rt != _rt ||
     sx != _sx || sy != _sy || sz != _sz || st != _st ||
     blockGeom != getCamera()->getGeometry()) {
    rx = _rx;
    ry = _ry;
    rz = _rz;
    rt = _rt;
    sx = _sx;
    sy = _sy;
    sz = _sz;
    st = _st;

    realloc();
    switch(getCamera()->getGeometry()) {
    case TimeSliceCamera::SLICE_X:
      Tk_PhotoSetSize(handle, sz, sy);
      break;
    case TimeSliceCamera::SLICE_Y:
      Tk_PhotoSetSize(handle, sx, sz);
      break;
    case TimeSliceCamera::SLICE_Z:
    case TimeSliceCamera::SLICE_ALPHA:
      Tk_PhotoSetSize(handle, sx, sy);
      break;
    case TimeSliceCamera::SLICE_XT:
      Tk_PhotoSetSize(handle, sx, st);
      break;
    case TimeSliceCamera::SLICE_YT:
      Tk_PhotoSetSize(handle, sy, st);
      break;
    case TimeSliceCamera::SLICE_ZT:
      Tk_PhotoSetSize(handle, sz, st);
      break;
    }
  }
}


#define DRAWXY(TYPE,INR) drawXY((TYPE) 0,INR,slicecam,visible, \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                block, \
				redColorMap,greenColorMap, \
				blueColorMap,sx,sy,rx,ry,interpolation)


#define DRAWXYVECTOR(TYPE,INR) drawXYVector((TYPE) 0,INR,slicecam,visible, \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                block, \
				redColorMap,greenColorMap, \
				blueColorMap,sx,sy,rx,ry)


#define DRAWXZ(TYPE,INR) drawXZ((TYPE) 0,INR,slicecam,visible, \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                block, \
				redColorMap,greenColorMap, \
				blueColorMap,sx,sz,rx,rz,interpolation)


#define DRAWXZVECTOR(TYPE,INR) drawXZVector((TYPE) 0,INR,slicecam,visible, \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                block, \
				redColorMap,greenColorMap, \
				blueColorMap,sx,sz,rx,rz)


#define DRAWYZ(TYPE,INR) drawYZ((TYPE) 0,INR,slicecam,visible, \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                block, \
				redColorMap,greenColorMap, \
				blueColorMap,sy,sz,ry,rz,interpolation)


#define DRAWYZVECTOR(TYPE,INR) drawYZVector((TYPE) 0,INR,slicecam,visible, \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                block, \
				redColorMap,greenColorMap, \
				blueColorMap,sy,sz,ry,rz)

#define DRAWXT(TYPE,INR) drawXT((TYPE) 0,INR, slicecam,visible, \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
				block, \
                                redColorMap,greenColorMap, \
				blueColorMap,sx,rx,st,rt)

#define DRAWXTVECTOR(TYPE,INR) drawXTVector((TYPE) 0,INR, slicecam,visible, \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                block, \
                                redColorMap,greenColorMap, \
				blueColorMap,sx,rx,st,rt)


#define DRAWYT(TYPE,INR) drawYT((TYPE) 0,INR, slicecam,visible, \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                block, \
                                redColorMap,greenColorMap, \
				blueColorMap,sy,ry,st,rt)

#define DRAWYTVECTOR(TYPE,INR) drawYTVector((TYPE) 0,INR, slicecam,visible, \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                block, \
                                redColorMap,greenColorMap, \
				blueColorMap,sy,ry,st,rt)


#define DRAWZT(TYPE,INR) drawZT((TYPE) 0,INR, slicecam,visible, \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                block, \
                                redColorMap,greenColorMap, \
				blueColorMap,sz,rz,st,rt)

#define DRAWZTVECTOR(TYPE,INR) drawZTVector((TYPE) 0,INR, slicecam,visible, \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                block, \
                                redColorMap,greenColorMap, \
				blueColorMap,sz,rz,st,rt)


void TclInrimage4DBackground::redraw(void) {
  unsigned int plane, i, j=0, x, y, z, *offset, o,t,p;
  register int r;
  
  voxelRatio();
  o = (unsigned int) slicecam->getSlice();
  p = (unsigned int) getCamera()->getSecondSlice();
  Inrimage *inr=getInrimage()->getInrimage(p);
  plane = inr->getX() * inr->getY();

  switch(getCamera()->getGeometry()) {

  /* slice perpendicular to X */
  case TimeSliceCamera::SLICE_X:
    if(o < visible[0] || o > visible[3]) {
      memset((void *) block->pixelPtr, 0, 3 * sy * sz);
      return;
    }

    switch(inr->getType()) {

    case Inrimage::WT_UNSIGNED_CHAR:
      DRAWYZ(unsigned char,inr);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      DRAWYZ(signed char,inr);
      break;
 
    case Inrimage::WT_UNSIGNED_SHORT:
      DRAWYZ(unsigned short int,inr);
      break;

    case Inrimage::WT_SIGNED_SHORT:
      DRAWYZ(short int,inr);
      break;

    case Inrimage::WT_UNSIGNED_INT:
      DRAWYZ(unsigned int,inr);
      break;

    case Inrimage::WT_SIGNED_INT:
      DRAWYZ(int,inr);
      break;

    case Inrimage::WT_UNSIGNED_LONG:
      DRAWYZ(unsigned LONGINT,inr);
      break;

    case Inrimage::WT_SIGNED_LONG:
      DRAWYZ(LONGINT,inr);
      break;

    case Inrimage::WT_FLOAT:
      DRAWYZ(float,inr);
      break;

    case Inrimage::WT_DOUBLE:
      DRAWYZ(double,inr);
      break;

    case Inrimage::WT_RGB:
      offset = new unsigned int[sz];
      for(i = 0; i < sz; i++) offset[i] =
				((int) ((double) i / rz) + visible[2]) * plane;
      
      for(y = 0, j = 0; y < sy; y++) {
	i = ((int)((double) y / ry) + visible[1]) * inr->getX() + o;
	for(z = 0; z < sz; z++) {
	  r = 3 * (offset[z] + i);
	  block->pixelPtr[j++] =
	    ((unsigned char *) inr->getData())[r];
	  block->pixelPtr[j++] =
	    ((unsigned char *) inr->getData())[r+1];
	  block->pixelPtr[j++] =
	    ((unsigned char *) inr->getData())[r+2];
	}
      }
      
      delete[] offset;
      break;

    case Inrimage::WT_RGBA:
      offset = new unsigned int[sz];
      for(i = 0; i < sz; i++) offset[i] =
				((int) ((double) i / rz) + visible[2]) * plane;
      
      for(y = 0, j = 0; y < sy; y++) {
	i = ((int)((double) y / ry) + visible[1]) * inr->getX() + o;
	for(z = 0; z < sz; z++) {
	  r = 4 * (offset[z] + i);
#if (defined (_ALPHA_) || defined (_LINUX_) || defined(WIN32))
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+3];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+2];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+1];
#else
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+1];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+2];
#endif
	}
      }
     
      delete[] offset;
      break;

    case Inrimage::WT_FLOAT_VECTOR:
      DRAWYZVECTOR(float,inr);
      break;

    case Inrimage::WT_DOUBLE_VECTOR:
      DRAWYZVECTOR(double,inr);
      break;

    default:
      break;
    }

    break;

    /* slice perpendicular to Y */
  case TimeSliceCamera::SLICE_Y:
    if(o < visible[1] || o > visible[4]) {
      memset((void *) block->pixelPtr, 0, 3 * sx * sz);
      return;
    }

    switch(inr->getType()) {

    case Inrimage::WT_UNSIGNED_CHAR:
      DRAWXZ(unsigned char,inr);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      DRAWXZ(signed char,inr);
      break;

    case Inrimage::WT_UNSIGNED_SHORT:
      DRAWXZ(unsigned short int,inr);
      break;

    case Inrimage::WT_SIGNED_SHORT:
      DRAWXZ(short int,inr);
      break;

    case Inrimage::WT_UNSIGNED_INT:
      DRAWXZ(unsigned int,inr);
      break;

    case Inrimage::WT_SIGNED_INT:
      DRAWXZ(int,inr);
      break;

    case Inrimage::WT_UNSIGNED_LONG:
      DRAWXZ(unsigned LONGINT,inr);
      break;

    case Inrimage::WT_SIGNED_LONG:
      DRAWXZ(LONGINT,inr);
      break;

    case Inrimage::WT_FLOAT:
      DRAWXZ(float,inr);
      break;

    case Inrimage::WT_DOUBLE:
      DRAWXZ(double,inr);
      break;

    case Inrimage::WT_RGB:
      offset = new unsigned int[sx];
      for(i = 0; i < sx; i++) offset[i] = (int) ((double) i / rx) + visible[0];
      o = (unsigned int) slicecam->getSlice() * inr->getX();

      for(z = 0, j = 0; z < sz; z++) {
	i = ((int)((double) z / rz) + visible[2]) * plane + o;
	for(x = 0; x < sx; x++) {
	  r = 3 * (offset[x] + i);
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+1];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+2];
	}
      }
      
      delete[] offset;
      break;

    case Inrimage::WT_RGBA:
      offset = new unsigned int[sx];
      for(i = 0; i < sx; i++) offset[i] = (int) ((double) i / rx) + visible[0];
      o = (unsigned int) slicecam->getSlice() * inr->getX();

      for(z = 0, j = 0; z < sz; z++) {
	i = ((int)((double) z / rz) + visible[2]) * plane + o;
	for(x = 0; x < sx; x++) {
	  r = 4 * (offset[x] + i);
#if (defined (_ALPHA_) || defined (_LINUX_) || defined(WIN32))
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+3];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+2];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+1];
#else
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+1];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+2];
#endif
	}
      }
       
      delete[] offset;
      break;

    case Inrimage::WT_FLOAT_VECTOR:
      DRAWXZVECTOR(float,inr);
      break;

    case Inrimage::WT_DOUBLE_VECTOR:
      DRAWXZVECTOR(double,inr);
      break;

    default:
      break;
    }

    break;

    /* slice perpendicular to Z */
  case TimeSliceCamera::SLICE_ALPHA:
  case TimeSliceCamera::SLICE_Z:
    if(getCamera()->getGeometry() == TimeSliceCamera::SLICE_Z &&
       (o < visible[2] || o > visible[5])) {
      memset((void *) block->pixelPtr, 0, 3 * sx * sy);
      return;
    }

    switch(inr->getType()) {

    case Inrimage::WT_UNSIGNED_CHAR:
      DRAWXY(unsigned char,inr);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      DRAWXY(signed char,inr);
      break;

    case Inrimage::WT_UNSIGNED_SHORT:
      DRAWXY(unsigned short int,inr);
      break;

    case Inrimage::WT_SIGNED_SHORT:
      DRAWXY(short int,inr);
      break;

    case Inrimage::WT_UNSIGNED_INT:
      DRAWXY(unsigned int,inr);
      break;

    case Inrimage::WT_SIGNED_INT:
      DRAWXY(int,inr);
      break;

    case Inrimage::WT_UNSIGNED_LONG:
      DRAWXY(unsigned LONGINT,inr);
      break;

    case Inrimage::WT_SIGNED_LONG:
      DRAWXY(LONGINT,inr);
      break;

    case Inrimage::WT_FLOAT:
      DRAWXY(float,inr);
      break;

    case Inrimage::WT_DOUBLE:
      DRAWXY(double,inr);
      break;

    case Inrimage::WT_RGB:
      offset = new unsigned int[sx];
      for(i = 0; i < sx; i++) offset[i] = (int) ((double) i / rx) + visible[0];
      o = (unsigned int) slicecam->getSlice() * plane;
    
      for(y = 0, j = 0; y < sy; y++) {
	i = o + ((int) ((double) y / ry) + visible[1]) * inr->getX();
	for(x = 0; x < sx; x++) {
	  r = 3 * (offset[x] + i);
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+1];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+2];
	}
      }

      delete[] offset;
      break;

    case Inrimage::WT_RGBA:
      offset = new unsigned int[sx];
      for(i = 0; i < sx; i++) offset[i] = (int) ((double) i / rx) + visible[0];
      o = (unsigned int) slicecam->getSlice() * plane;
    
      for(y = 0, j = 0; y < sy; y++) {
	i = o + ((int) ((double) y / ry) + visible[1]) * inr->getX();
	for(x = 0; x < sx; x++) {
	  r = 4 * (offset[x] + i);
#if (defined (_ALPHA_) || defined (_LINUX_) || defined(WIN32))
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+3];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+2];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+1];
#else
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+1];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+2];
#endif
	}
      }

      delete[] offset;
      break;

    case Inrimage::WT_FLOAT_VECTOR:
      DRAWXYVECTOR(float,inr);
      break;

    case Inrimage::WT_DOUBLE_VECTOR:
      DRAWXYVECTOR(double,inr);
      break;

    default:
      break;
    }

    break;
  case TimeSliceCamera::SLICE_XT:
    if(o < visible[1] || o > visible[4] || p < visible [2] || p> visible[5]) {
      memset((void *) block->pixelPtr, 0, 3 * sx * st);
      return;
    }

    switch(inr->getType()) {

    case Inrimage::WT_UNSIGNED_CHAR:
      DRAWXT(unsigned char,getInrimage());
      break;
 
    case Inrimage::WT_UNSIGNED_SHORT:
      DRAWXT(unsigned short int,getInrimage());
      break;

    case Inrimage::WT_SIGNED_SHORT:
      DRAWXT(short int,getInrimage());
      break;

    case Inrimage::WT_UNSIGNED_INT:
      DRAWXT(unsigned int,getInrimage());
      break;

    case Inrimage::WT_SIGNED_INT:
      DRAWXT(int,getInrimage());
      break;

    case Inrimage::WT_UNSIGNED_LONG:
      DRAWXT(unsigned LONGINT,getInrimage());
      break;

    case Inrimage::WT_SIGNED_LONG:
      DRAWXT(LONGINT,getInrimage());
      break;

    case Inrimage::WT_FLOAT:
      DRAWXT(float,getInrimage());
      break;

    case Inrimage::WT_DOUBLE:
      DRAWXT(double,getInrimage());
      break;

    case Inrimage::WT_RGB:
      offset = new unsigned int[sx];
      o = (unsigned int) getCamera()->getSecondSlice() * plane +
	(unsigned int)slicecam->getSlice()*getX();
      for(i = 0; i < sx; i++) offset[i] = o+
				((int) ((double) i / rx) + visible[0]) * plane;
      for(t = 0; t < st; t++) {
	i = ((int)((double) t / rt) + visible[6]) * getInrimage()->getTimeDimension();
	inr=getInrimage()->getInrimage(i);

	for(x = 0; x < sx; x++) {
	  r = 3 * (offset[x]);
	  block->pixelPtr[j++] =
	    ((unsigned char *) inr->getData())[r];
	  block->pixelPtr[j++] =
	    ((unsigned char *) inr->getData())[r+1];
	  block->pixelPtr[j++] =
	    ((unsigned char *) inr->getData())[r+2];
	}
      }
      
      delete[] offset;
      break;

    case Inrimage::WT_RGBA:
      offset = new unsigned int[sx];
      o = (unsigned int) getCamera()->getSecondSlice() * plane +
	(unsigned int)slicecam->getSlice()*getX();
      for(i = 0; i < sx; i++) offset[i] = o+
				((int) ((double) i / rx) + visible[0]) * plane;

      for(t = 0; t < st; t++) {
	i = ((int)((double) t / rt) + visible[6]) * getInrimage()->getTimeDimension();
	inr=getInrimage()->getInrimage(i);

	for(x = 0; x < sx; x++) {
	  r = 4 * (offset[x]);
#if (defined (_ALPHA_) || defined (_LINUX_) || defined(WIN32))
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+3];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+2];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+1];
#else
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+1];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+2];
#endif
	}
      }
           
      delete[] offset;
      break;

    case Inrimage::WT_FLOAT_VECTOR:
      DRAWXTVECTOR(float,getInrimage());
      break;

    case Inrimage::WT_DOUBLE_VECTOR:
      DRAWXTVECTOR(double,getInrimage());
      break;

    default:
      break;
    }
    break;
  case TimeSliceCamera::SLICE_YT:
    if(o < visible[0] || o > visible[3] || p < visible [2] || p> visible[5]) {
      memset((void *) block->pixelPtr, 0, 3 * sy * st);
      return;
    }

    switch(inr->getType()) {

    case Inrimage::WT_UNSIGNED_CHAR:
      DRAWYT(unsigned char,getInrimage());
      break;
 
    case Inrimage::WT_UNSIGNED_SHORT:
      DRAWYT(unsigned short int,getInrimage());
      break;

    case Inrimage::WT_SIGNED_SHORT:
      DRAWYT(short int,getInrimage());
      break;

    case Inrimage::WT_UNSIGNED_INT:
      DRAWYT(unsigned int,getInrimage());
      break;

    case Inrimage::WT_SIGNED_INT:
      DRAWYT(int,getInrimage());
      break;

    case Inrimage::WT_UNSIGNED_LONG:
      DRAWYT(unsigned LONGINT,getInrimage());
      break;

    case Inrimage::WT_SIGNED_LONG:
      DRAWYT(LONGINT,getInrimage());
      break;

    case Inrimage::WT_FLOAT:
      DRAWYT(float,getInrimage());
      break;

    case Inrimage::WT_DOUBLE:
      DRAWYT(double,getInrimage());
      break;

    case Inrimage::WT_RGB:
      offset = new unsigned int[sy];
      o = (unsigned int) getCamera()->getSecondSlice() * plane +
	(unsigned int)slicecam->getSlice();
      for(i = 0; i < sy; i++) offset[i] = o+
				((int) ((double) i / ry) + visible[1]) * inr->getX();
      for(t = 0; t < st; t++) {
	i = ((int)((double) t / rt) + visible[6]) * getInrimage()->getTimeDimension();
	inr=getInrimage()->getInrimage(i);

	for(y = 0; y < sy; y++) {
	  r = 3 * (offset[y]);
	  block->pixelPtr[j++] =
	    ((unsigned char *) inr->getData())[r];
	  block->pixelPtr[j++] =
	    ((unsigned char *) inr->getData())[r+1];
	  block->pixelPtr[j++] =
	    ((unsigned char *) inr->getData())[r+2];
	}
      }
      
      delete[] offset;
      break;

    case Inrimage::WT_RGBA:
      offset = new unsigned int[sy];
      o = (unsigned int) getCamera()->getSecondSlice() * plane +
	(unsigned int)slicecam->getSlice();
      for(i = 0; i < sy; i++) offset[i] = o+
				((int) ((double) i / ry) + visible[1]) * inr->getX();
      for(t = 0; t < st; t++) {
	i = ((int)((double) t / rt) + visible[6]) * getInrimage()->getTimeDimension();
	inr=getInrimage()->getInrimage(i);

	for(y = 0; y < sy; y++) {
	  r = 4 * (offset[y]);
#if (defined (_ALPHA_) || defined (_LINUX_) || defined(WIN32))
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+3];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+2];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+1];
#else
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+1];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+2];
#endif
	}
      }
           
      delete[] offset;
      break;

    case Inrimage::WT_FLOAT_VECTOR:
      DRAWYTVECTOR(float,getInrimage());
      break;

    case Inrimage::WT_DOUBLE_VECTOR:
      DRAWYTVECTOR(double,getInrimage());
      break;

    default:
      break;
    }
    break;
  case TimeSliceCamera::SLICE_ZT:
    if(o < visible[0] || o > visible[3] || p < visible [1] || p> visible[4]) {
      memset((void *) block->pixelPtr, 0, 3 * sz * st);
      return;
  }

    switch(inr->getType()) {

    case Inrimage::WT_UNSIGNED_CHAR:
      DRAWZT(unsigned char,getInrimage());
      break;
 
    case Inrimage::WT_UNSIGNED_SHORT:
      DRAWZT(unsigned short int,getInrimage());
      break;

    case Inrimage::WT_SIGNED_SHORT:
      DRAWZT(short int,getInrimage());
      break;

    case Inrimage::WT_UNSIGNED_INT:
      DRAWZT(unsigned int,getInrimage());
      break;

    case Inrimage::WT_SIGNED_INT:
      DRAWZT(int,getInrimage());
      break;

    case Inrimage::WT_UNSIGNED_LONG:
      DRAWZT(unsigned LONGINT,getInrimage());
      break;

    case Inrimage::WT_SIGNED_LONG:
      DRAWZT(LONGINT,getInrimage());
      break;

    case Inrimage::WT_FLOAT:
      DRAWZT(float,getInrimage());
      break;

    case Inrimage::WT_DOUBLE:
      DRAWZT(double,getInrimage());
      break;

    case Inrimage::WT_RGB:
      offset = new unsigned int[sz];
      o = (unsigned int) getCamera()->getSecondSlice() * inr->getX() +
	(unsigned int)slicecam->getSlice();
      for(i = 0; i < sz; i++) offset[i] = o+
				((int) ((double) i / rz) + visible[2]) * plane;
      for(t = 0; t < st; t++) {
	i = ((int)((double) t / rt) + visible[6]) * getInrimage()->getTimeDimension();
	inr=getInrimage()->getInrimage(i);

	for(z = 0; z < sz; z++) {
	  r = 3 * (offset[z]);
	  block->pixelPtr[j++] =
	    ((unsigned char *) inr->getData())[r];
	  block->pixelPtr[j++] =
	    ((unsigned char *) inr->getData())[r+1];
	  block->pixelPtr[j++] =
	    ((unsigned char *) inr->getData())[r+2];
	}
      }
      
      delete[] offset;
      break;

    case Inrimage::WT_RGBA:
      offset = new unsigned int[sz];
      o = (unsigned int) getCamera()->getSecondSlice() * getX() +
	(unsigned int)slicecam->getSlice();
      for(i = 0; i < sz; i++) offset[i] = o+
				((int) ((double) i / rz) + visible[2]) * plane;
      for(t = 0; t < st; t++) {
	i = ((int)((double) t / rt) + visible[6]) * getInrimage()->getTimeDimension();
	inr=getInrimage()->getInrimage(i);

	for(z = 0; z < sz; z++) {
	  r = 4 * (offset[z]);
#if (defined (_ALPHA_) || defined (_LINUX_) || defined(WIN32))
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+3];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+2];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+1];
#else
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+1];
	  block->pixelPtr[j++] =
	    ((unsigned char *)inr->getData())[r+2];
#endif
	}
      }
           
      delete[] offset;
      break;

    case Inrimage::WT_FLOAT_VECTOR:
      DRAWZTVECTOR(float,getInrimage());
      break;

    case Inrimage::WT_DOUBLE_VECTOR:
      DRAWZTVECTOR(double,getInrimage());
      break;

    default:
      break;
    }

    break;

  }

  Vec2<double> tr(slicecam->getWidth() / 2.0 -
		  (slicecam->getCenter().getX() * slicecam->getScale()),
		  slicecam->getHeight() / 2.0 -
		  (slicecam->getCenter().getY() * slicecam->getScale()));

  TkWrapper_PhotoPutBlock(handle, block, 0, 0, block->width, block->height);
  getModule()->tickle("%s coords inrimage %i %i",
		      (char *) getCamera()->getInterface()->getCanvas(),
		      (int) (tr[0] + 0),
		      (int) (tr[1] + 0));
  //  inrv->getInrimage()->updateSlice(this);
}


void TclInrimage4DBackground::imageTypeChanged(void) {
  delete glmin;
  delete glmax;
  Inrimage4D *inr=getInrimage();
  glmin = Voxel::voxelByType(inr->getType());
  glmax = Voxel::voxelByType(inr->getType());
  slicecam->focusOnBackground();
  imageGLChanged();
}
void TclInrimage4DBackground::scaleValue(const TclScale *scale, double val) {
 if (((TclScale *)this->Owner<MinGLScale>::pointer())==scale)
   setMinThreshold(val);
 else
   {
     if (((TclScale *)this->Owner<MaxGLScale>::pointer())==scale)
       setMaxThreshold(val);
   }
}

void TclInrimage4DBackground::setGLTransfo(TclInrimage4DBackground::GLTRANSFO t,const unsigned char *red,
					   const  unsigned char *green,const  unsigned char *blue) {
  if ((t!=transfo)|| (t==USER))
    {
      if ((t!=USER) || (red && green && blue))
	{
	  transfo = t;
	  switch(t) {
	  case AFFINE:
	    redColorMap=AffineColorMap;
	    greenColorMap=AffineColorMap;
	    blueColorMap=AffineColorMap;
	    break;
	  case THRESHOLD:
	    redColorMap=ThresholdColorMap;
	    greenColorMap=ThresholdColorMap;
	    blueColorMap=ThresholdColorMap;
	    break;
	  case PET_COLORS:
	    redColorMap=PETColorMapRed;
	    greenColorMap=PETColorMapGreen;
	    blueColorMap=PETColorMapBlue;
	    break;
	  case JET_COLORS:
	    redColorMap=JetColorMapRed;
	    greenColorMap=JetColorMapGreen;
	    blueColorMap=JetColorMapBlue;
	    break; 
	  case SPECT_COLORS:
	    redColorMap=SPECTColorMapRed;
	    greenColorMap=SPECTColorMapGreen;
	    blueColorMap=SPECTColorMapBlue;
	    break;
	  case HOTIRON_COLORS:
	    redColorMap=HotIronColorMapRed;
	    greenColorMap=HotIronColorMapGreen;
	    blueColorMap=HotIronColorMapBlue;
	    break;
	  case BLUE:
	    redColorMap=ZeroColorMap;
	    greenColorMap=ZeroColorMap;
	    blueColorMap=AffineColorMap;
	    break;
	  case USER:
	    redColorMap=red;
	    greenColorMap=green;
	    blueColorMap=blue;
	    break;
	  }
	}
    }

  redraw();
}


void TclInrimage4DBackground::imageVoxelSizeChanged(void) {
  double m;
  Inrimage4D *inr=getInrimage();
  if(inr->getVX() < inr->getVY()) m = inr->getVX();
  else m = inr->getVY();
  if(inr->getVZ() < m) m = inr->getVZ();
  slicecam->scaleCamera(1.0 / (m * slicecam->getScale()));
  slicecam->focusOnBackground();
  redraw();
}


void TclInrimage4DBackground::imageGLChanged(void) {
  glBounds();
  if(getMinScale())
    getMinScale()->reset(this);
  if(getMaxScale())
    getMaxScale()->reset(this);
  redraw();
}


void TclInrimage4DBackground::setMinThreshold(double val)
  throw(Voxel::InvalidValue) {
  if(glmin->getValueAsDouble() != val) {
    glmin->setValueFromDouble(val);
    redraw();
  }
}


void TclInrimage4DBackground::setMaxThreshold(double val)
  throw(Voxel::InvalidValue) {
  if(glmax->getValueAsDouble() != val) {
    glmax->setValueFromDouble(val);
    redraw();
  }
}


void TclInrimage4DBackground::setGlThresholds(double min, double max)
  throw(Voxel::InvalidValue) {
  if(glmin->getValueAsDouble() != min && glmax->getValueAsDouble() != max) {
    glmin->setValueFromDouble(min);
    glmax->setValueFromDouble(max);
    redraw();
  }
}


void TclInrimage4DBackground::setGlThresholds(const TclString& min, const TclString& max)
  throw(Voxel::InvalidValue) {

  if(min != EMPTY_STRING) glmin->scan(min);
  if(max != EMPTY_STRING) glmax->scan(max);

  redraw();
}


/* compute screen coordinate (mx, my) of voxel (x, y, z) */
void TclInrimage4DBackground::screen(VoxelCoordinate4D *vc,
				     int& mx, int& my) const {
  if(!getIVS()) {
    Inrimage4D *inr=getInrimage();
    Vec2<double> r;

    switch(getCamera()->getGeometry()) {
    case TimeSliceCamera::SLICE_X:
      r = Vec2<double>((vc->z  ) * inr->getVZ(),
		       (vc->y  ) * inr->getVY());
      break;
    case TimeSliceCamera::SLICE_Y:
      r = Vec2<double>((vc->x  ) * inr->getVX(),
		       (vc->z  ) * inr->getVZ());
      break;
    case TimeSliceCamera::SLICE_Z:
    case TimeSliceCamera::SLICE_ALPHA:
      r = Vec2<double>((vc->x ) * inr->getVX(),
		       (vc->y ) * inr->getVY());
      break;
    case TimeSliceCamera::SLICE_XT:
      r = Vec2<double>((vc->x ) * inr->getVX(),
		       (vc->t ) * inr->getVT());
      break;
    case TimeSliceCamera::SLICE_YT:
      r = Vec2<double>((vc->y ) * inr->getVY(),
		       (vc->t ) * inr->getVT());
      break;
    case TimeSliceCamera::SLICE_ZT:
      r = Vec2<double>((vc->z ) * inr->getVZ(),
		       (vc->t ) * inr->getVT());
      break;
    default:
      mx = 0;
      my = 0;
      return;
    }
    
    r = (r - slicecam->getCenter()) * slicecam->getScale() + slicecam->getCameraCenter();
    mx = (int) r[0] ;
    my = (int) r[1] ;
  }
  
  else {
    switch(getCamera()->getGeometry()) {
    case TimeSliceCamera::SLICE_X:
      mx = (int) ((vc->z - visible[2] ) * rz) ;
      my = (int) ((vc->y - visible[1] ) * ry) ;
      break;
    case TimeSliceCamera::SLICE_Y:
      mx = (int) ((vc->x - visible[0] ) * rx) ;
      my = (int) ((vc->z - visible[2] ) * rz) ;
      break;
    case TimeSliceCamera::SLICE_Z:
    case TimeSliceCamera::SLICE_ALPHA:
      mx = (int) ((vc->x - visible[0] ) * rx) ;
      my = (int) ((vc->y - visible[1] ) * ry) ;
      break;
    case TimeSliceCamera::SLICE_XT:
      mx = (int) ((vc->x - visible[0]) * rx) ;
      my = (int) ((vc->t - visible[6]) * rt) ;
      break;
    case TimeSliceCamera::SLICE_YT:
      mx = (int) ((vc->y - visible[1] ) * ry) ;
      my = (int) ((vc->t - visible[6] ) * rt) ;
      break;
    case TimeSliceCamera::SLICE_ZT:
      mx = (int) ((vc->z - visible[2] ) * rz) ;
      my = (int) ((vc->t - visible[6] ) * rt) ;
      break;
    default:
      mx = 0;
      my = 0;
      break;
    }
    
  }
}
TclString TclInrimage4DBackground::getLabel(int &x,int &y){
  TclString s;
  VoxelCoordinate4D *vc = coordinates(x, y);
  Voxel *v = getInrimage()->getVoxel(vc->x, vc->y, vc->z,vc->t);
  TclString sval;
  v->print(sval);
  s.sprintf("%3u %3u %3u %3u: %s", vc->x, vc->y, vc->z,vc->t,(char *) sval);
  delete vc;
  delete v;
  return(s);
}

// Converts a canvas coordinate (i, j) in the corresponding image voxel
// (x, y, z)
VoxelCoordinate4D *TclInrimage4DBackground::coordinates(unsigned int i,
							unsigned int j) const {
  VoxelCoordinate4D *vc = new VoxelCoordinate4D();
  int si, sj;
  Inrimage4D *inr=getInrimage();
  Vec2<double> pix=Vec2<double>(i, j);
  Vec2<double> t;
  slicecam->mapToWorld(pix,t);
  
  
  switch(getCamera()->getGeometry()) {
  case TimeSliceCamera::SLICE_X:
    vc->x = (unsigned int) slicecam->getSlice();
    vc->t = (unsigned int) getCamera()->getSecondSlice();
    if(!getIVS()) {
      si = (unsigned int) (t[1] / inr->getVY());
      sj = (unsigned int) (t[0] / inr->getVZ());
    }
    else {
      si=(unsigned int) t[1];
      sj=(unsigned int) t[0];
    }
    if(si < 0) vc->y = 0;
    else if(si >= (int) inr->getY()) vc->y = inr->getY() - 1;
    else vc->y = si;
    if(sj < 0) vc->z = 0;
    else if(sj >= (int) inr->getZ()) vc->z = inr->getZ() - 1;
    else vc->z = sj;
    break;
  case TimeSliceCamera::SLICE_Y:
    if(!getIVS()) {
      si = (unsigned int) (t[0] / inr->getVX());
      sj = (unsigned int) (t[1] / inr->getVZ());
    } else {
      si= (unsigned int) t[0];
      sj= (unsigned int) t[1];
    }
    if(si < 0) vc->x = 0;
    else if(si >= (int) inr->getX()) vc->x = inr->getX() - 1;
    else vc->x = si;
    vc->y = (unsigned int) slicecam->getSlice();
    vc->t = (unsigned int) getCamera()->getSecondSlice();
    if(sj < 0) vc->z = 0;
    else if(sj >= (int) inr->getZ()) vc->z = inr->getZ() - 1;
    else vc->z = sj;
    break;
    case TimeSliceCamera::SLICE_Z:
     if(!getIVS()) {
      si = (unsigned int) (t[0] / inr->getVX());
      sj = (unsigned int) (t[1] / inr->getVZ());
    } else {
      si= (unsigned int) t[0];
      sj= (unsigned int) t[1];
    }
     if(si < 0) vc->x = 0;
     else if(si >= (int) inr->getX()) vc->x = inr->getX() - 1;
     else vc->x = si;
     if(sj < 0) vc->y = 0;
     else if(sj >= (int) inr->getY()) vc->y = inr->getY() - 1;
     else vc->y = sj;
     vc->z = (unsigned int) slicecam->getSlice();
     vc->t = (unsigned int) getCamera()->getSecondSlice();
     break;
    case TimeSliceCamera::SLICE_ALPHA:
      if(!getIVS()) {
	si = (unsigned int) (t[0] / inr->getVX());
	sj = (unsigned int) (t[1] / inr->getVY());
      } else {
	si= (unsigned int) t[0];
	sj= (unsigned int) t[1];
      }
      if(si < 0) vc->x = 0;
      else if(si >= (int) inr->getX()) vc->x = inr->getX() - 1;
      else vc->x = si;
      if(sj < 0) vc->y = 0;
      else if(sj >= (int) inr->getY()) vc->y = inr->getY() - 1;
      else vc->y = sj;
      vc->z = ((unsigned int) slicecam->getSlice()) % inr->getZ();
      vc->t = (unsigned int) getCamera()->getSecondSlice();
      if(vc->z < 0) vc->z = 0;
      else if(vc->z >= inr->getZ()) vc->z = inr->getZ() - 1;
      break;
    case TimeSliceCamera::SLICE_XT:
      vc->y = (unsigned int) slicecam->getSlice();
      vc->z = (unsigned int) getCamera()->getSecondSlice();
     if(!getIVS()) {
       si = (unsigned int) (t[0] / inr->getVX());
       sj = (unsigned int) (t[1] / inr->getVT());
     } else {
       si = (unsigned int) t[0];
       sj = (unsigned int) t[1];       
     }
     if(si < 0) vc->x = 0;
     else if(si >= (int) inr->getX()) vc->x = inr->getX() - 1;
     else vc->x = si;
     if(sj < 0) vc->t = 0;
     else if(sj >= (int) inr->getTimeDimension()) vc->t = inr->getTimeDimension() - 1;
     else vc->t = sj;
     break;
    case TimeSliceCamera::SLICE_YT:
      vc->x = (unsigned int) slicecam->getSlice();
      vc->z = (unsigned int) getCamera()->getSecondSlice();
      if(!getIVS()) {
	si = (unsigned int) (t[0] / inr->getVY());
	sj = (unsigned int) (t[1] / inr->getVT());
      } else {
	si = (unsigned int) t[0];
	sj = (unsigned int) t[1];       
      }
      if(si < 0) vc->y = 0;
      else if(si >= (int) inr->getY()) vc->y = inr->getY() - 1;
      else vc->y = si;
      vc->y += visible[1];
      if(sj < 0) vc->t = 0;
      else if(sj >= (int) inr->getTimeDimension()) vc->t = inr->getTimeDimension() - 1;
      else vc->t = sj;
      vc->t += visible[6];
      break;
    case TimeSliceCamera::SLICE_ZT:
      vc->x = (unsigned int) slicecam->getSlice();
      vc->y = (unsigned int) getCamera()->getSecondSlice();
      if(!getIVS()) {
	si = (unsigned int) (t[0] / inr->getVZ());
	sj = (unsigned int) (t[1] / inr->getVT());
      } else {
	si = (unsigned int) t[0];
	sj = (unsigned int) t[1];       
      }
      if(si < 0) vc->z = 0;
      else if(si >= (int) inr->getZ()) vc->z = inr->getZ() - 1;
      else vc->z = si;
      if(sj < 0) vc->t = 0;
      else if(sj >= (int) inr->getTimeDimension()) 
	vc->t = inr->getTimeDimension() - 1;
      else vc->t = sj;
      break;
    default:
      break;
    }

  return vc;
}

Rotation3D TclInrimage4DBackground::getWorldFrame() const{
  return getInrimage()->getInrimage(0)->getRotationMatrixFromConstImage();
}
Vec3<double>  TclInrimage4DBackground::getImageOffset() const{
  return getInrimage()->getInrimage(0)->getTranslationFromConstImage();
}

void TclInrimage4DBackground::screen(CartesianCoordinate4D *vc,
				    int& mx, int& my) const {  
  if(!getIVS()) {
    Vec2<double> r;
    Inrimage4D *inr=getInrimage();
    switch(getCamera()->getGeometry()) {
    case TimeSliceCamera::SLICE_X:
      r=Vec2<double>(vc->z,vc->y);
      break;
    case TimeSliceCamera::SLICE_Y:
      r=Vec2<double>(vc->x,vc->z);
      break;
    case TimeSliceCamera::SLICE_Z:
      r=Vec2<double>(vc->x,vc->y);
      break;
    case TimeSliceCamera::SLICE_ALPHA:
      {
	double x0 = inr->getX() * inr->getVX() / 2.0;
	/*double alpha = slicecam->getSlice() / 180.0 * M_PI;*/
	double alpha = (double) slicecam->getSlice() / (double) (inr->getZ()-1)
	  * M_PI;
	if(alpha >= M_PI) alpha -= M_PI;
	double ax = x0 * (1 + sin(alpha)) - vc->x;
	double ay = x0 * (1 - cos(alpha)) - vc->y;
	r = Vec2<double>(sqrt(ax * ax + ay * ay), vc->z);
      }
      break;
    case TimeSliceCamera::SLICE_XT:
      r=Vec2<double>(vc->x,vc->t);
      break;
    case TimeSliceCamera::SLICE_YT:
      r=Vec2<double>(vc->y,vc->t);
      break;
    case TimeSliceCamera::SLICE_ZT:
      r=Vec2<double>(vc->z,vc->t);
      break;
    default:
      mx = 0;
      my = 0;
      return;
    }
    
    r = (r - slicecam->getCenter()) * slicecam->getScale() + slicecam->getCameraCenter();
    mx = (int) r[0] ;
    my = (int) r[1] ;
  }

  else {
    Inrimage4D *inr=getInrimage();
    switch(getCamera()->getGeometry()) {
    case TimeSliceCamera::SLICE_X:
      mx = (int) ((vc->z / inr->getVZ() - visible[2]) * rz) ;
      my = (int) ((vc->y / inr->getVY() - visible[1]) * ry) ;
      break;
    case TimeSliceCamera::SLICE_Y:
      mx = (int) ((vc->x / inr->getVX() - visible[0]) * rx) ;
      my = (int) ((vc->z / inr->getVZ() - visible[2]) * rz) ;
      break;
    case TimeSliceCamera::SLICE_Z:
      mx = (int) ((vc->x / inr->getVY() - visible[0]) * rx) ;
      my = (int) ((vc->y / inr->getVX() - visible[1]) * ry) ;
      break;
    case TimeSliceCamera::SLICE_ALPHA:
      mx = (vc->y <= 180)
	? (int) ((inr->getX() / 2.0 + vc->x + 0.5) * rx)
	: (int) ((inr->getX() / 2.0 - vc->x + 0.5) * rx);
      my = (int) ((vc->z + 0.5) * ry);
      break;
    case TimeSliceCamera::SLICE_XT:
      mx = (int) ((vc->x / inr->getVX() - visible[0]) * rx) ;
      my = (int) ((vc->t / inr->getVT() - visible[6]) * rt) ;
      break;
    case TimeSliceCamera::SLICE_YT:
      mx = (int) ((vc->y / inr->getVY() - visible[1]) * ry) ;
      my = (int) ((vc->t / inr->getVT() - visible[6]) * rt) ;
      break;
    case TimeSliceCamera::SLICE_ZT:
      mx = (int) ((vc->z / inr->getVZ() - visible[2]) * rz) ;
      my = (int) ((vc->t / inr->getVT() - visible[6]) * rt) ;
      break;
    default:
      mx = 0;
      my = 0;
      break;
    }
    
  }
}


CartesianCoordinate4D *
TclInrimage4DBackground::cartesianCoordinates(unsigned int i,
					      unsigned int j) const {
  CartesianCoordinate4D *vc = new CartesianCoordinate4D();
  Inrimage4D *inr=getInrimage();
  if(!getIVS()) {
    Vec2<double> t = (Vec2<double>(i, j) - slicecam->getCameraCenter())
      / slicecam->getScale() + slicecam->getCenter();
  
    switch(getCamera()->getGeometry()) {
    case TimeSliceCamera::SLICE_X:
      vc->x = (slicecam->getSlice() + 0.5) * inr->getVX();
      vc->y = t[1];
      vc->z = t[0];
      break;
    case TimeSliceCamera::SLICE_Y:
      vc->x = t[0];
      vc->y = (slicecam->getSlice() + 0.5) * inr->getVY();
      vc->z = t[1];
      break;
    case TimeSliceCamera::SLICE_Z:
      vc->x = t[0];
      vc->y = t[1];
      vc->z = (slicecam->getSlice() + 0.5) * inr->getVZ();
      break;
    case TimeSliceCamera::SLICE_ALPHA:
      {
	double x0 = (double) inr->getX() * inr->getVX() / 2.0;
	/*double alpha = slicecam->getSlice() / 180.0 * M_PI;*/
	double alpha = (double) slicecam->getSlice() / (double) (inr->getZ()-1)
	  * M_PI;
	if(alpha >= M_PI) alpha -= M_PI;
	if(t[0] >= x0) {
	  vc->x = x0 - (t[0] - x0) * sin(alpha); 
	  vc->y = x0 + (t[0] - x0) * cos(alpha);
	}
	else {
	  vc->x = x0 - (x0 - t[0]) * sin(alpha + M_PI); 
	  vc->y = x0 + (x0 - t[0]) * cos(alpha + M_PI);
	}
	vc->z = t[1];
      }
      break;
    case TimeSliceCamera::SLICE_XT:
      vc->y = (slicecam->getSlice() + 0.5) * inr->getVY();
      vc->z = (getCamera()->getSecondSlice() + 0.5) * inr->getVZ();
      vc->x = t[0];
      vc->t = t[1];
      break;
    case TimeSliceCamera::SLICE_YT:
      vc->x = (slicecam->getSlice() + 0.5) * inr->getVX();
      vc->z = (getCamera()->getSecondSlice() + 0.5) * inr->getVZ();
      vc->y = t[0];
      vc->t = t[1];
      break;
    case TimeSliceCamera::SLICE_ZT:
      vc->x = (slicecam->getSlice() + 0.5) * inr->getVX();
      vc->y = (getCamera()->getSecondSlice() + 0.5) * inr->getVY();
      vc->z = t[0];
      vc->t = t[1];
      break;
    default:
      break;
    }
  }

  else {
    switch(getCamera()->getGeometry()) {
    case TimeSliceCamera::SLICE_X:
      vc->x = (slicecam->getSlice() + 0.5) * inr->getVX();
      vc->y = (j / ry + visible[1])  * inr->getVY();
      vc->z = (i / rz + visible[2])  * inr->getVZ();
      break;
    case TimeSliceCamera::SLICE_Y:
      vc->x = (i / rx + visible[0])  * inr->getVX();
      vc->y = (slicecam->getSlice() + 0.5) * inr->getVY();
      vc->z = (j / rz + visible[2])  * inr->getVZ();
      break;
    case TimeSliceCamera::SLICE_Z:
      vc->x = (i / rx + visible[0])  * inr->getVX();
      vc->y = (j / ry + visible[1])  * inr->getVY();
      vc->z = (slicecam->getSlice() + 0.5) * inr->getVZ();
      break;
    case TimeSliceCamera::SLICE_ALPHA:
      {
	double x0 = (double) inr->getX() / 2.0;
	/*double alpha = slicecam->getSlice() / 180.0 * M_PI;*/
	double alpha = (double) slicecam->getSlice() / (double) (inr->getZ()-1)
	  * M_PI;
	if(alpha >= M_PI) alpha -= M_PI;
	if(i / rx >= x0) {
	  vc->x = (x0 - (i / rx - x0) * sin(alpha)) * inr->getVX(); 
	  vc->y = (x0 + (i / rx - x0) * cos(alpha)) * inr->getVX();
	}
	else {
	  vc->x = (x0 - (x0 - i / rx) * sin(alpha + M_PI)) * inr->getVX(); 
	  vc->y = (x0 + (x0 - i / rx) * cos(alpha + M_PI)) * inr->getVX();
	}
	vc->z = ((inr->getY() - 1) - j / ry) * inr->getVY();
      }
      break;
    case TimeSliceCamera::SLICE_XT:
      vc->x = (i / rx + visible[0])  * inr->getVX();
      vc->t = (j / rt + visible[6])  * inr->getVT();
      vc->y = (slicecam->getSlice() + 0.5) * inr->getVY();
      vc->z = (getCamera()->getSecondSlice() + 0.5) * inr->getVZ();
      break;
    case TimeSliceCamera::SLICE_YT:
      vc->y = (i / ry + visible[1])  * inr->getVY();
      vc->t = (j / rt + visible[6])  * inr->getVT();
      vc->x = (slicecam->getSlice() + 0.5) * inr->getVX();
      vc->z = (getCamera()->getSecondSlice() + 0.5) * inr->getVZ();
      break;
    case TimeSliceCamera::SLICE_ZT:
      vc->z = (i / rz + visible[2])  * inr->getVZ();
      vc->t = (j / rt + visible[6])  * inr->getVT();
      vc->x = (slicecam->getSlice() + 0.5) * inr->getVX();
      vc->y = (getCamera()->getSecondSlice() + 0.5) * inr->getVY();
      break;
    default:
      break;
    }
  }

  return vc;  
}
  


void TclInrimage4DBackground::setVisibleArea(void) {
  setVisibleArea(0, 0, 0,0, getInrimage()->getX() - 1, getInrimage()->getY() - 1, getInrimage()->getZ() - 1,getInrimage()->getTimeDimension() - 1);
}


void TclInrimage4DBackground::setVisibleArea(unsigned int x1, unsigned int y1,
					     unsigned int z1, unsigned int t1,
					     unsigned int x2, unsigned int y2, 
					     unsigned int z2, unsigned int t2) {
  Inrimage4D *inr=getInrimage();
  if(x2 >= inr->getX()) x2 = inr->getX() - 1;
  if(y2 >= inr->getY()) y2 = inr->getY() - 1;
  if(z2 >= inr->getZ()) z2 = inr->getZ() - 1;
  if(t2 >= inr->getTimeDimension()) t2 = inr->getTimeDimension() - 1;

  if(x1 <= x2 && y1 <= y2 && z1 <= z2 && t1<=t2) {
    visible[0] = x1;
    visible[1] = y1;
    visible[2] = z1;
    visible[3] = x2;
    visible[4] = y2;
    visible[5] = z2;
    visible[6] = t1;
    visible[7] = t2;

    visibleAreaChanged();
  }
}
void TclInrimage4DBackground::setVisibleArea(unsigned int x1, unsigned int y1,
					     unsigned int z1,
					     unsigned int x2, unsigned int y2, 
					     unsigned int z2) {
  setVisibleArea(x1,y1,z1,0,x2,y2,z2,getInrimage()->getTimeDimension()-1);
}

void TclInrimage4DBackground::getWorldCoordinates(const int x,const int y, 
						  double &px, double &py, 
						  double &pz ) const {
  CartesianCoordinate4D *cc=cartesianCoordinates((unsigned int) x,
						 (unsigned int) y);
  px=cc->x; py=cc->y; pz=cc->z;
  delete cc;
}
void TclInrimage4DBackground::getPixelPositionFromImageCoordinates(const unsigned int vx,
								 const unsigned int vy,
								 const unsigned int vz,
								 int &x, int&y) const {
  VoxelCoordinate4D *vc=new VoxelCoordinate4D();
  vc->x=vx;  vc->y=vy;  vc->z=vz;
  screen(vc,x,y);
  delete vc;
}
void TclInrimage4DBackground::getPixelPositionFromImageCoordinates(const unsigned int vx,
								 const unsigned int vy,
								 const unsigned int vz,
								 const unsigned int vt,
								 int &x, int&y) const {
  VoxelCoordinate4D *vc=new VoxelCoordinate4D();
  vc->x=vx;  vc->y=vy;  vc->z=vz;vc->t=vt;
  screen(vc,x,y);
  delete vc;
}
void TclInrimage4DBackground::getPixelPositionFromWorldCoordinates(const double px,
								 const double py,
								 const double pz,
								 int &x, int&y) const {
  CartesianCoordinate4D *cc=new CartesianCoordinate4D();
  cc->x=px;  cc->y=py;  cc->z=pz;cc->t=0;
  screen(cc,x,y);
  delete cc;
}

void TclInrimage4DBackground::getImageCoordinates(const int x,const int y, 
						  unsigned int &px, 
						  unsigned int &py, 
						  unsigned int &pz,
						  unsigned int &pt) const {
  VoxelCoordinate4D *vc=coordinates((unsigned int) x,
				    (unsigned int) y);
  px=vc->x; py=vc->y; pz=vc->z;;pt=vc->t;
  delete vc;
}
void TclInrimage4DBackground::getImageCoordinates(const int x,const int y, 
						  unsigned int &px, 
						  unsigned int &py, 
						  unsigned int &pz ) const {
  VoxelCoordinate4D *vc=coordinates((unsigned int) x,
				    (unsigned int) y);
  px=vc->x; py=vc->y; pz=vc->z;
  delete vc;
}

void TclInrimage4DBackground::visibleAreaChanged(void) {

  slicecam->focusOnBackground();
  if(slicecam->getScaleCamera())
    slicecam->resetScaleCamera();
  redraw();
}


void TclInrimage4DBackground::refreshCamera(void) {
  slicecam->setBackground(this);
}


void TclInrimage4DBackground::inrimageChanged(void) {
  
  Inrimage4D *inr=getInrimage();
  visible[0] = 0;
  visible[1] = 0;
  visible[2] = 0;
  visible[3] = inr->getX() - 1;
  visible[4] = inr->getY() - 1;
  visible[5] = inr->getZ() - 1;
  visible[6] = 0;
  visible[7] = inr->getTimeDimension() - 1;
  
  if(glmin) delete glmin;
  if(glmax) delete glmax;
  glmin = Voxel::voxelByType(inr->getType());
  glmax = Voxel::voxelByType(inr->getType());
  glBounds();

  slicecam->focusOnBackground();
  slicecam->updateDisplay();

  if (getMinScale())
    getMinScale()->reset(this);
  if (getMaxScale())
    getMaxScale()->reset(this);
  if (slicecam->getScaleCamera())
    slicecam->getScaleCamera()->reset(slicecam,
				      slicecam->sliceName(slicecam->getGeometry()));
  if (getCamera()->getSecondScaleCamera())
    {
      getCamera()->getSecondScaleCamera()->reset(slicecam,getCamera()->secondSliceName(getCamera()->getGeometry()));
    }
  redraw();
}
int TclInrimage4DBackground::destroyCmd(int argc, char **argv) {
  TclModule *m;
  unsigned int ref;

  if(argc != 0) return COMMAND_SYNOPSIS;
  m = getModule();
  ref = TclObject::getRef();
  delete this;
  m->setResult("%s%u", TclInrimage4DBackground::getObjectName(), ref);
  return TCL_OK;
}
int TclInrimage4DBackground::scalemodeCmd(int argc, char **argv) {
  if(argc > 0) {
    GLTRANSFO t;
    if(!strcmp(argv[0], "affine")) t = AFFINE;
    else if(!strcmp(argv[0], "threshold")) t = THRESHOLD;
    else if(!strcmp(argv[0], "spect")) t = SPECT_COLORS;
    else if(!strcmp(argv[0], "hotiron")) t = HOTIRON_COLORS;
    else if(!strcmp(argv[0], "blue")) t = BLUE;
    else if(!strcmp(argv[0], "pet")) t = PET_COLORS;
    else if(!strcmp(argv[0], "jet")) t = JET_COLORS;
    else if(!strcmp(argv[0], "user")) 
      {
	t = USER;
	if (argc>1)
	  {
	    TclColorMap *tc=(TclColorMap *)
	      TclObjectKind::getObject(typeid(TclColorMap), argc, argv, 1);
	    if (tc)
	      setGLTransfo(t,tc->getRed(),tc->getGreen(),tc->getBlue());
	  }
      }
    else return COMMAND_SYNOPSIS;
    setGLTransfo(t);
  }
  switch(getGLTransfo()) {
  case AFFINE:
    setResult("affine");
    break;
  case THRESHOLD:
    setResult("threshold");
    break;
  case PET_COLORS:
    setResult("pet");
    break;
  case JET_COLORS:
    setResult("jet");
    break;  
  case SPECT_COLORS:
    setResult("spect");
    break;
  case HOTIRON_COLORS:
    setResult("hotiron");
    break;
  case BLUE:
    setResult("blue");
    break;
  case USER:
    setResult("user");
    break;
  }
  return TCL_OK;
}
int TclInrimage4DBackground::interpolationCmd(int argc, char **argv) {
  if (argc>1)
    return COMMAND_SYNOPSIS;
  if (argc==1)
    {
      if (!strcmp(argv[0],"nearest"))
	{
	  interpolation=NEAREST;
	  redraw();
	}
      else if (!strcmp(argv[0],"bilinear"))
	{
	  interpolation=BILINEAR;
	  redraw();
	}
      else {
	setResult("%s must be one of the following : nearest, bilinear",argv[0]);
	return TCL_ERROR;
      }
    }
  switch (interpolation) {
  case NEAREST:
    setResult("nearest");
    break;
  case BILINEAR:
    setResult("bilinear");
    break;
  }
  return TCL_OK;
}

int TclInrimage4DBackground::greyLevelCmd(int argc, char **argv) {
  if (argc==1)
    {
      if (!strcmp(argv[0],"min"))
	{
	  if (Owner<MinGLScale>::pointer)
	    setResult("%s",(char *)Owner<MinGLScale>::pointer->name());
	  return TCL_OK;
	}
      else if (!strcmp(argv[0],"max"))
	{
	  if (Owner<MaxGLScale>::pointer)
	    setResult("%s",(char *)Owner<MaxGLScale>::pointer->name());
	  return TCL_OK;
	}
      else
	return COMMAND_SYNOPSIS;	
    }
  else if (argc==2)
    { 
      if (!strcmp(argv[0],"min"))
	{
	  MinGLScale *gl= dynamic_cast<MinGLScale *>(	
	    TclObjectKind::getObject(typeid(GLScale), argc, argv, 1));
	  if (gl)
	    {
	      addMinGLScale(gl);
	      setResult("%s",(char *)getMinScale()->name());
	      return TCL_OK;
	    }
	  else
	    {
	      setResult("%s is not a minGLScale",argv[1]);
	      return TCL_ERROR;
	    }
	}
      else if (!strcmp(argv[0],"max"))
	{
	  MaxGLScale *gl= dynamic_cast<MaxGLScale *>(	
	    TclObjectKind::getObject(typeid(GLScale), argc, argv, 1));
	  if (gl)
	    {
	      addMaxGLScale(gl);
	      setResult("%s",(char *)getMaxScale()->name());
	      return TCL_OK;
	    }
	  else
	    {
	      setResult("%s is not a maxGLScale",argv[1]);
	      return TCL_ERROR;
	    }
	}
      else
	return COMMAND_SYNOPSIS;

    }
  else
    return COMMAND_SYNOPSIS;
}


int TclInrimage4DBackground::create(TclModule *m, int argc, char **argv) {
  TclScene2D *scene = TclScene2D::getDefaultScene(m);
  int i;
  TclInrimage4DBackground *b;
  TclInrimage4D *inr;
  MinGLScale *minScale=0;
  MaxGLScale *maxScale=0;
  bool useMinScale=false;
  bool useMaxScale=false;
  if (argc<3)
    return COMMAND_SYNOPSIS;
  inr = TclInrimage4D::getObject(argc, argv, 1);
  if (!inr)
    {
      m->setResult("%s is not an image",argv[1]);
      return TCL_ERROR;
    }
  TclTimeSliceCamera *cam=TclTimeSliceCamera::getObject(argc, argv, 2);
  if (!cam)
    {
      m->setResult("%s is not an slice camera",argv[2]);
      return TCL_ERROR;
    }

  for(i=3; i < argc; i++) {
    if(!strcmp(argv[i], "-scene")) {
      if( ++i >= argc) return COMMAND_SYNOPSIS;
      scene = (TclScene2D *)
	TclObjectKind::getObject(typeid(TclScene2D), argc, argv, i);
      if(!scene) {
	m->setResult("Illegal scene name : %s",argv[i]);
	return TCL_ERROR;
      }
    }
    if(!strcmp(argv[i], "-min")) {
      if( ++i >= argc) return COMMAND_SYNOPSIS;
      minScale= (MinGLScale *) 	TclObjectKind::getObject(typeid(MinGLScale), argc, argv, i);
      if (!minScale) {
	m->setResult("Illegal MinGLScale name : %s",argv[i]);
	return TCL_ERROR;
      }
      else
	useMinScale=true;
    }
    if(!strcmp(argv[i], "-max")) {
      if( ++i >= argc) return COMMAND_SYNOPSIS;
      maxScale= (MaxGLScale *) 	TclObjectKind::getObject(typeid(MaxGLScale), argc, argv, i);
      if (!maxScale) {
	m->setResult("Illegal MaxGLScale name : %s",argv[i]);
	return TCL_ERROR;
      }
      else
	useMaxScale=true;
    }
    else return COMMAND_SYNOPSIS;
  }

  b = new TclInrimage4DBackground(m,new Inrimage4DView(inr->getClient(),scene->getClient()),cam->getClient(),
				useMinScale,minScale,
				useMaxScale,maxScale);
  m->setResult("%s%u", b->getObjectName(), b->TclObject::getRef());
  return TCL_OK;
}

}
