#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/TclInrimage4DView.h"
#include <inrimage/TclInrimage4DView.h>
#include <inrimage/TclInrimage4D.h>
#include <graphics2D/TclTimeSliceCamera.h>
#include <inrimage/TclInrimage4DBackground.h>


namespace yav {


TclInrimage4DView::TclInrimage4DView(TclModule *m, TclInrimage4D *inr,
				 TclScene2D *scene, bool _createGLScale)
  :TclScene2DObject(typeid(TclInrimage4DView),
		  new Inrimage4DView(inr->getClient(), scene->getClient(),_createGLScale), m)
{}

TclInrimage4DView::TclInrimage4DView(TclScene2D *scene, const std::type_info& id,
				 Inrimage4DView *inrview) :
  TclScene2DObject(id,inrview,scene->getModule())
{
}

TclInrimage4DView::~TclInrimage4DView() {
  deleted(typeid(TclInrimage4DView));
}

void TclInrimage4DView::init(const char *install, const char *uninstall) {
}

int TclInrimage4DView::destroyCmd(int argc, char **argv) {
  TclModule *m;
  unsigned int ref;

  if(argc != 0) return COMMAND_SYNOPSIS;
  m = getModule();
  ref = getRef();
  delete this;
  m->setResult("%s%u", TclInrimage4DView::getObjectName(), ref);
  return TCL_OK;
}


int TclInrimage4DView::sceneCmd(int argc, char **argv) {
  if(argc != 0) return COMMAND_SYNOPSIS;
  setResult("%s%u",
	    TclObjectKind::getObjectName(typeid(TclScene2D)),
	    ((TclScene2D *)getClient()->getScene()->getInterface())->getRef());
  return TCL_OK;
}

int TclInrimage4DView::GLScaleCmd(int argc, char **argv) {
  if(argc != 2) return COMMAND_SYNOPSIS;
  TclTimeSliceCamera *slicecam= (TclTimeSliceCamera *) TclObjectKind::getObject(typeid(TclTimeSliceCamera),argc, argv, 0);
  if (!slicecam)
    {
      setResult("%s is not a slice camera",argv[0]);
      return TCL_ERROR;
    }
  TclInrimage4DBackground *b=(TclInrimage4DBackground *)
    getClient()->getBackgroundFromCamera(slicecam->getClient());
  if (!strcmp(argv[1],"-min"))
    {
      if (b->getMinScale())
	{
	  setResult(b->getMinScale()->name());
	}
      return TCL_OK;
    }
  else if (!strcmp(argv[1],"-max"))
    {
      if (b->getMaxScale())
	{
	  setResult(b->getMaxScale()->name());
	}
      return TCL_OK;
    }
  else return COMMAND_SYNOPSIS;
}

int TclInrimage4DView::inrimageCmd(int argc, char **argv) {
  TclInrimage4D *inr;

  if(argc > 1) return COMMAND_SYNOPSIS;
  if(argc == 1) {
    inr = TclInrimage4D::getObject(argc, argv, 0);
    if(!inr) return COMMAND_SYNOPSIS;
    getClient()->changeInrimage(inr->getClient());
  }
  inr = TclInrimage4D::getInterface(getClient()->getInrimage());
  if(inr)
    setResult("%s%u", 
	      TclObjectKind::getObjectName(typeid(TclInrimage4D)),
	      inr->getRef());
  else
    setResult("no inrimage");
  return TCL_OK;
}
int TclInrimage4DView::backgroundCmd(int argc, char **argv) {
  TclTimeSliceCamera *slicecam;

  if(argc != 1) return COMMAND_SYNOPSIS;
  if (!strcmp(argv[0],"all")) {
    std::list<Inrimage4DBackground *>::iterator itb;
    TclString out;
    for (itb=getClient()->firstBackground();itb!=getClient()->lastBackground();
	 itb++) {
      out+=((TclInrimage4DBackground *)(*itb))->name();
      out+=" ";
    }
    if (out==EMPTY_STRING) 
      setResult("no background");
    else
      setResult(out);
  } else {
    slicecam = TclTimeSliceCamera::getObject(argc, argv, 0);
    if(!slicecam) return COMMAND_SYNOPSIS;
    
    TclInrimage4DBackground *b= (TclInrimage4DBackground *)
      getClient()->getBackgroundFromCamera(slicecam->getClient());
    
    if(b)
      setResult("%s",(char *) b->name()); 
    else
      setResult("no background");
  }
  return TCL_OK;
}




/* "inr%u [scn2D%u]",
   "create an inrview oject to view given inrimage in given scene (a new scene is created if none is provided)" */
int TclInrimage4DView::create(TclModule *m, int argc, char **argv) {
  TclInrimage4D *inr = 0;
  TclScene2D *scn = 0;
  TclInrimage4DView *inrv;
  int i;
  bool addGLScale=true;

  if(argc > 4) return COMMAND_SYNOPSIS;
  for(i = 1; i < argc; i++) {
    if(!inr) {
      inr = TclInrimage4D::getObject(argc, argv, i);
      if(!inr) {
	if(!scn) scn = (TclScene2D *) TclObjectKind::getObject(typeid(TclScene2D), argc, argv, i);
	else return COMMAND_SYNOPSIS;
	if(!scn) return COMMAND_SYNOPSIS;
      }
    }
    else if (!strcmp(argv[i],"-noGLScale"))
      {
	addGLScale=false;
      }
    else if(!scn) {
      scn = (TclScene2D *) TclObjectKind::getObject(typeid(TclScene2D), argc, argv, i);
      if(!scn) return COMMAND_SYNOPSIS;
    }
    else return COMMAND_SYNOPSIS;
  }
  if(!inr) return COMMAND_SYNOPSIS;
  if(!scn) scn = new TclScene2D(m);

  inrv = new TclInrimage4DView(m, inr, scn,addGLScale);
  m->setResult("%s%u", TclObjectKind::getObjectName(typeid(TclInrimage4DView)), inrv->getRef());
  return TCL_OK;
}


} // namespace yav

