#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#pragma implementation "inrimage/TclInrimageROI.h"
#include <inrimage/TclInrimage.h>
#include <inrimage/TclInrimageROI.h>
#include <modules/TclGenericHistogram.h>

namespace yav {


TclInrimageROI::TclInrimageROI(TclScene2D *scene, Inrimage *image, 
			       SurfaceSliceable *s,
			       InrimageROI::SurfaceRoiMethod m, 
			       double param,int selectedZone) : TclScene2DObject(typeid(TclInrimageROI), new InrimageROI(scene->getClient(),image,s,m,param,selectedZone),scene->getModule()) {
  init();
}
TclInrimageROI::TclInrimageROI(TclScene2D *scene,     
			       char *filename) :  TclScene2DObject(typeid(TclInrimageROI), new InrimageROI(scene->getClient(),filename),scene->getModule()) {
  init();
}

TclInrimageROI::TclInrimageROI(TclScene2D *scene,     
			       Inrimage *image, 
			       const Inrimage *binary) throw(InrimageROI::ImageMissing,
							     InrimageROI::BinaryImageWrongSize) : TclScene2DObject(typeid(TclInrimageROI), new InrimageROI(scene->getClient(),image,binary),scene->getModule()) {
  init();
}

TclInrimageROI::TclInrimageROI(TclScene2D *scene,     
			       Inrimage *image)  
  throw(InrimageROI::ImageMissing) : TclScene2DObject(typeid(TclInrimageROI), new InrimageROI(scene->getClient(),image),scene->getModule()) {
  init();
}

TclInrimageROI::TclInrimageROI(TclScene2D *scene,const std::type_info& info,
			       InrimageROI *ir) : TclScene2DObject(info,ir,scene->getModule()) {
  TclString n;
}
TclInrimageROI::~TclInrimageROI() {
  deleted(typeid(TclInrimageROI));
}
void TclInrimageROI::init() {
  done(typeid(TclInrimageROI),"inrimage::installROIMenu"  ,"inrimage::uninstallROIMenu");
  getClient()->focusScene();

  TclString n;
  n.sprintf("region no %u",getRef());
  getClient()->setRealName(n);
}
int TclInrimageROI::destroyCmd(int argc, char **argv) {
  TclModule *m;

  if(argc != 0) return COMMAND_SYNOPSIS;
  m = getModule();
  TclString n=name();
  delete this;
  m->setResult(n);
  return TCL_OK;
}


int TclInrimageROI::sceneCmd(int argc, char **argv) {
  if(argc > 1) return COMMAND_SYNOPSIS;
  else if (argc==0) {
    setResult("%s%u",
	      TclObjectKind::getObjectName(typeid(TclScene2D)),
	      ((TclScene2D *)getClient()->getScene()->getInterface())->getRef());
    return TCL_OK;
  } else {
    TclScene2D *scn=(TclScene2D *) TclObjectKind::getObject(typeid(TclScene2D), argc, argv, 0);	
    if (!scn) {
      setResult("%s must be a 2D scene",argv[0]);
      return TCL_ERROR;
    }
    getClient()->Data2D::setScene(scn->getClient());
    setResult("%s%u",
	      TclObjectKind::getObjectName(typeid(TclScene2D)),
	      ((TclScene2D *)getClient()->getScene()->getInterface())->getRef());
    return TCL_OK;
  }
}
int TclInrimageROI::meshCmd(int argc, char **argv) {
  if(argc != 0) return COMMAND_SYNOPSIS;
  if (getClient()->getSurface())
    setResult("%s",(char *) getClient()->getSurface()->getSurfaceSliceableName());
  else
    setResult("none");
  return TCL_OK;
}
int TclInrimageROI::meanVarianceCmd(int argc, char **argv) {
  double rejectRate=0.0;
  if(argc == 1) {
    rejectRate= (double) atof(argv[0]);
  }
  else if (argc>1)
    return COMMAND_SYNOPSIS;
  double mean, variance;
  getClient()->computeMeanVariance(mean,variance,rejectRate);
  setResult("%f %f",mean, variance);
  return TCL_OK;
}
int TclInrimageROI::opacityCmd(int argc, char **argv) {
  if(argc == 0) {
    switch(getClient()->getOpacity())
      {
      case InrimageROI::FULL_OPACITY:
	setResult("full");
	break;
      case InrimageROI::OPACITY_75:
	setResult("opacity75");
	break;
      case InrimageROI::OPACITY_50:
	setResult("opacity50");
	break;
      case InrimageROI::OPACITY_25:
	setResult("opacity25");
	break;
      case InrimageROI::OPACITY_12:
	setResult("opacity12");
	break;
      }
    return TCL_OK;
  }
  else if (argc==1) {
    InrimageROI::OpacityLevel opacityLevel;
    if (!strcmp(argv[0],"full")) {
      opacityLevel=InrimageROI::FULL_OPACITY;
    }
    else if (!strcmp(argv[0],"opacity75")) {
      opacityLevel=InrimageROI::OPACITY_75;
    }
    else if (!strcmp(argv[0],"opacity50")) {
      opacityLevel=InrimageROI::OPACITY_50;
    }
    else if (!strcmp(argv[0],"opacity25")) {
      opacityLevel=InrimageROI::OPACITY_25;
    }
    else if (!strcmp(argv[0],"opacity12")) {
      opacityLevel=InrimageROI::OPACITY_12;
    }
    else {
      setResult("%s must be either full or opacity75, opacity50, opacity25 or opacity12");
      return TCL_ERROR;
    }
    getClient()->setOpacity(opacityLevel);
    getClient()->redrawScene();
    setResult("%s",argv[0]);
    return TCL_OK;
  }
  else
    return COMMAND_SYNOPSIS;
}


int TclInrimageROI::create(TclModule *m, int argc, char **argv) {
  TclInrimage *inr = 0;
  TclInrimage *bin = 0;
  TclScene2D *scn = 0;
  TclInrimageROI *roi;
  InrimageROI::SurfaceRoiMethod method=  InrimageROI::SURFACE_BOUNDARY;
  SurfaceSliceable *s=0;
  int i;
  double p=1.0;
  int selectedZone= -1;

  if (argc==1)
    return COMMAND_SYNOPSIS;
  inr = TclInrimage::getObject(argc, argv, 1);  
  if (!inr)
    {
		if (argc>3){
			m->setResult("cannot read image %s",argv[1]);
			return TCL_ERROR;
		}
		if (argc==3)
		{
			if(!scn) scn = (TclScene2D *) TclObjectKind::getObject(typeid(TclScene2D), argc, argv, 2);	
			if (!scn) 
			{
				m->setResult("%s : invalid scene name",argv[2]);
				return TCL_ERROR;
			}
		}
		if(!scn) scn =  TclScene2D::getDefaultScene(m);
		roi= new TclInrimageROI(scn,argv[1]);
		m->setResult("%s%u", TclObjectKind::getObjectName(typeid(TclInrimageROI)), roi->getRef());
		return TCL_OK;
	}
  bin=TclInrimage::getObject(argc, argv, 2);  
  if (bin) {
    if ((bin->getX()!=inr->getX()) || (bin->getY()!=inr->getY()) || (bin->getZ()!=inr->getZ())) {
      m->setResult("the two images do not have the same size");
      return TCL_ERROR;
    }
    for(i = 3; i < argc; i++) {
      if(!scn) scn = (TclScene2D *) TclObjectKind::getObject(typeid(TclScene2D), argc, argv, i);	
      if (!scn) 
	{
	  m->setResult("%s : invalid scene name ",argv[i]);
	  return TCL_ERROR;
	}
    }

  }
  else {
    s=AutoLink<SurfaceSliceable>::getHead();
    for(;s && (s->getSurfaceSliceableName()!=(const char *)argv[2]);s=s->AutoLink<SurfaceSliceable>::following());
    if (!s)
      {
		  if (argc>=6) {
			  double vx,vy,vz,r;
			  vx=atof(argv[2]);
			  vy=atof(argv[3]);
			  vz=atof(argv[4]);
			  r=atof(argv[5]);
			  for(i = 6; i < argc; i++) {
				  if(!scn) scn = (TclScene2D *) TclObjectKind::getObject(typeid(TclScene2D), argc, argv, i);	
				  if (!scn) 
				  {
					  m->setResult("%s : invalid scene name ",argv[i]);
					  return TCL_ERROR;
				  }
			  }
			  if(!scn) scn =  TclScene2D::getDefaultScene(m);
			  roi = new TclInrimageROI(scn,typeid(TclInrimageROI),new InrimageROI(scn->getClient(),inr->getClient(),
				  Vec3<double>(vx,vy,vz),r));
			  m->setResult("%s", (char *) roi->name());
			  return TCL_OK;

		  } else {
		  if(!scn) scn =  TclScene2D::getDefaultScene(m);
		  roi= new TclInrimageROI(scn,inr->getClient());
		  m->setResult("%s%u", TclObjectKind::getObjectName(typeid(TclInrimageROI)), roi->getRef());
		  return TCL_OK;
		  }
	  }
    
    for(i = 3; i < argc; i++) {
      if (!strcmp(argv[i],"boundary"))
	method=InrimageROI::SURFACE_BOUNDARY;
	  else if (!strcmp(argv[i],"boundaryZ"))
	method=InrimageROI::SURFACE_BOUNDARY_Z;
      else if (!strcmp(argv[i],"region"))
	method=InrimageROI::SURFACE_REGION;
      else if (!strcmp(argv[i],"band"))
	method=InrimageROI::SURFACE_NEAR_BOUNDARY;
      else if (!strcmp(argv[i],"outside"))
	method=InrimageROI::SURFACE_REGION_OUTSIDE;

      else if ((!strcmp(argv[i],"-param")) && ((i+1)<argc))
	{
	  p=(double) atof(argv[++i]);
	}
      else if ((!strcmp(argv[i],"-zone")) && ((i+1)<argc))
	{
	  selectedZone=(int) atoi(argv[++i]);
	}

      else
	{
	  if(!scn) scn = (TclScene2D *) TclObjectKind::getObject(typeid(TclScene2D), argc, argv, i);	
	  if (!scn) 
	    {
	      m->setResult("%s : invalid parameter",argv[i]);
	      return TCL_ERROR;
	    }
	}
    }
  }
  if(!scn) scn=TclScene2D::getDefaultScene(m);
  if (!bin)
    {
      roi  = new TclInrimageROI(scn,inr->getClient(),s,method,p,selectedZone);
    }
  else
    {
      roi  = new TclInrimageROI(scn,inr->getClient(),bin->getClient());
    }
  m->setResult("%s%u", TclObjectKind::getObjectName(typeid(TclInrimageROI)), roi->getRef());
  return TCL_OK;

}
int TclInrimageROI::intensityImageCmd(int argc, char **argv) {
  if (argc>1)
    return COMMAND_SYNOPSIS;
  else if (argc==1)
  {

	  TclInrimage *inr = TclInrimage::getObject(argc, argv, 0);
	  if (!inr) {
		  setResult("%s must be an image",argv[0]);
		  return TCL_ERROR;
	  }
	  else {
		  getClient()->setImage(inr->getClient());
		  setResult("%s",argv[0]);
		  return TCL_OK;
	  }
  } else {
	  if (getClient()->getImage()) {
		  setResult(TclInrimage::getInterface(getClient()->getImage())->name());
		  return TCL_OK;
	  } else {
		  setResult("none");
		  return TCL_OK;
	  }
  }
	
}
int TclInrimageROI::binaryImageCmd(int argc, char **argv) {
  if (argc>0)
    return COMMAND_SYNOPSIS;
  else {
    Inrimage *im=getClient()->createBinaryImage();
    if (!im)  {
      setResult("none");
      return TCL_OK;
    }
    else {
      TclInrimage *tinr=new TclInrimage(getModule(), typeid(TclInrimage), (Inrimage *)im);
      setResult("%s",(char *)tinr->name());
      return TCL_OK;
    }
  }

}
int TclInrimageROI::morphoCmd(int argc, char **argv) {
  if ((argc==0))
    return COMMAND_SYNOPSIS;
  int i;
  if (!strcmp(argv[0],"dilation")) {
    unsigned int niter=1;
    CONNECTIVITY_TYPE con=SIX_CONNECTIVITY;
    for (i=1;i<argc;++i) {
      if ((!strcmp(argv[i],"iter")) && ((i+1)<argc)) {
	i++;
	niter=(unsigned int)atoi(argv[i]);
      }
      if ((!strcmp(argv[i],"con")) && ((i+1)<argc)) {
	i++;
	con=TclInrimage::connectivityFromString(argv[i]);
      }
    }
    getClient()->dilate(con,niter);
    getClient()->redrawScene();
    setResult("ROI dilated %d times",niter);
    return TCL_OK;
  }
  else if (!strcmp(argv[0],"erosion")) {
    unsigned int niter=1;
    CONNECTIVITY_TYPE con=SIX_CONNECTIVITY;
    for (i=1;i<argc;++i) {
      if ((!strcmp(argv[i],"iter")) && ((i+1)<argc)) {
	i++;
	niter=(unsigned int)atoi(argv[i]);
      }
      if ((!strcmp(argv[i],"con")) && ((i+1)<argc)) {
	i++;
	con=TclInrimage::connectivityFromString(argv[i]);
      }
    }
    getClient()->erode(con,niter);
    getClient()->redrawScene();
    setResult("ROI eroded %d times",niter);
    return TCL_OK;
  }
  else {
    setResult("%s must be either dilation or erosion",argv[0]);
    return TCL_ERROR;
  }
}
int TclInrimageROI::colorCmd(int argc, char **argv) {
 
  if (argc==0)
    {  
      float r,g,b;
      getClient()->getColor(r,g,b);
      setResult("%f %f %f",r,g,b);
      return TCL_OK;
    }
  else if (argc==3)
    {
      float r,g,b;
      getClient()->setColor((float)atof(argv[0]),
			    (float)atof(argv[1]),(float)atof(argv[2]));
      getClient()->redrawScene();
      getClient()->getColor(r,g,b);
      setResult("%f %f %f",r,g,b);
      return TCL_OK;
    }
  else 
    return COMMAND_SYNOPSIS;
}
int TclInrimageROI::histogramCmd( int argc, char **argv) {
  if (argc!=0)
    return COMMAND_SYNOPSIS;
  else {
    TclGenericHistogram *h= new TclGenericHistogram(getModule(),getClient()->createHistogram());
    setResult("%s",(char *) h->name());
    return TCL_OK;
  }
}
int TclInrimageROI::visibilityCmd( int argc, char **argv) {
  if (argc>1) 
    return COMMAND_SYNOPSIS;
  if (argc==1) {
    bool vis;
    if (TclModule::boolValue(argv[0],vis)) {
      getClient()->setVisibility(vis);
      getClient()->redrawScene();
    } else {
      setResult("%s must be either true or false",argv[0]);
      return TCL_ERROR;
    }
  }
  setResult("%s",TclModule::boolString(getClient()->getVisibility()));
  return TCL_OK;
}
int TclInrimageROI::realNameCmd( int argc, char **argv) 
{
  if (argc>1) 
    return COMMAND_SYNOPSIS;
  if (argc==1) {
    TclString n(argv[0]);
    getClient()->setRealName(n);
  }
  // saves the file

  setResult("%s",(char *)getClient()->getRealName());
  return TCL_OK;
}
int TclInrimageROI::saveCmd( int argc, char **argv) 
{
  if (argc>1) 
    return COMMAND_SYNOPSIS;

  if (argc==0)
    getClient()->write(getClient()->getFilename());
  else 
    getClient()->write(argv[0]);
  // saves the file

  setResult("%s has been saved as %s",(char *) name(),
	    (char *)(getClient()->getFilename()));
  return TCL_OK;
}
int TclInrimageROI::saveOptionCmd( int argc, char **argv) 
{
  if (argc==0) 
    {
      TclString res;
      ObjectsList<GeneralFormatModule>::iterator it=
	getClient()->getFirstModule();
      for(;it!=getClient()->getLastModule();it++)
	res+=(*it)->getName()+SPACE_STRING;
      setResult(res);
      return TCL_OK;
    }
  else if (argc==1)
    {
      ObjectsList<GeneralFormatModule>::iterator it=
	getClient()->getFirstModule();
      GeneralFormatModule *module=0;
      for(;it!=getClient()->getLastModule();it++)
	{
	  if (!strcmp(argv[0],(char *)(*it)->getName()))
	    {
	      module=(*it);
	    }
	}
      if (module)
	{
	  setResult("%s",TclModule::boolString(module->isSelectedWriting()));
	  return TCL_OK;
	}
      else
	{
	  setResult("Unknown module name : %s",argv[0]);
	  return TCL_ERROR;
	}
    }
  else if (argc==2)
    {
      ObjectsList<GeneralFormatModule>::iterator it=
	getClient()->getFirstModule();
      GeneralFormatModule *module=0;
      for(;it!=getClient()->getLastModule();it++)
	{
	  if (!strcmp(argv[0],(char *)(*it)->getName()))
	    {
	      module=(*it);
	    }
	}
      if (module)
	{
	  bool selected;
	  if (TclModule::boolValue(argv[1],selected))
	    {
	      if (selected)
		module->selectWriting();
	      else
		module->unselectWriting();
	      setResult("selection mode of module %s is set to %s",
			argv[0],argv[1]);
	      return TCL_OK;
	    }
	  else
	    {
	      setResult("%s must be true or false",argv[1]);
	      return TCL_ERROR;
	    }
	}
      else
	{
	  setResult("Unknown module name : %s",argv[0]);
	  return TCL_ERROR;
	}

    }
  else
    return COMMAND_SYNOPSIS;
}
int TclInrimageROI::centerAndRadius3DCmd( int argc, char **argv) {
	if (argc!=0)
		 return COMMAND_SYNOPSIS;
	Vec3<double> center;
	double radius;
	getClient()->getCenterAndRadius(center,radius);
	setResult("%f %f %f %f",center.getX(),center.getY(),center.getZ(),
	  radius);
  return TCL_OK;
}
int TclInrimageROI::loadOptionCmd( int argc, char **argv) 
{
  if (argc==0) 
    {
      TclString res;
      ObjectsList<GeneralFormatModule>::iterator it=
	getClient()->getFirstModule();
      for(;it!=getClient()->getLastModule();it++)
	res+=(*it)->getName()+SPACE_STRING;
      setResult(res);
      return TCL_OK;
    }
  else if (argc==1)
    {
      ObjectsList<GeneralFormatModule>::iterator it=
	getClient()->getFirstModule();
      GeneralFormatModule *module=0;
      for(;it!=getClient()->getLastModule();it++)
	{
	  if (!strcmp(argv[0],(char *)(*it)->getName()))
	    {
	      module=(*it);
	    }
	}
      if (module)
	{
	  setResult("%s",TclModule::boolString(module->isSelectedReading()));
	  return TCL_OK;
	}
      else
	{
	  setResult("Unknown module name : %s",argv[0]);
	  return TCL_ERROR;
	}
    }
  else if (argc==2)
    {
      ObjectsList<GeneralFormatModule>::iterator it=
	getClient()->getFirstModule();
      GeneralFormatModule *module=0;
      for(;it!=getClient()->getLastModule();it++)
	{
	  if (!strcmp(argv[0],(char *)(*it)->getName()))
	    {
	      module=(*it);
	    }
	}
      if (module)
	{
	  bool selected;
	  if (TclModule::boolValue(argv[1],selected))
	    {
	      if (selected)
		module->selectReading();
	      else
		module->unselectReading();
	      setResult("selection mode of module %s is set to %s",
			argv[0],argv[1]);
	      return TCL_OK;
	    }
	  else
	    {
	      setResult("%s must be true or false",argv[1]);
	      return TCL_ERROR;
	    }
	}
      else
	{
	  setResult("Unknown module name : %s",argv[0]);
	  return TCL_ERROR;
	}

    }
  else
    return COMMAND_SYNOPSIS;
}

void TclInrimageROI::load(const TclString &filename,Scene2D *scn)
{
  new TclInrimageROI((TclScene2D *)scn->getInterface(),(char *)filename);
}


void TclInrimageROI::createCommands(TclModule *m)
{
	TclObjectKind *kind;

  kind = new Scene2DItem(m, typeid(TclInrimageROI), "roi","Region of Interest",
			 &TclInrimageROI::create, &TclInrimageROI::list,
			 TclScene2D::FILE | TclScene2D::OTHER,"ImageROI", ".roi",&TclInrimageROI::load,0,"%s.roi [scn2D%u] | inr%u vx vy vz radius [scn2D%u] |  inr%u [ [sm%u | tr3D%u | tet3D%u] [boundary | region | band | scn2D%u | -param %f -zone %u] | inr%u [scn2D%u] ]");
  m->newCommand("roi%u", kind->getOptions(), "inrimage region of interest management command");
 
  TclInrimageROI::createOptionCommands(kind);
  kind->newOption("-destroy", (TclObject::OPTION) &TclInrimageROI::destroyCmd, 0,
		  "destroy region of interest");
  kind->newOption("-scene", (TclObject::OPTION) &TclInrimageROI::sceneCmd, 0,
		  "return region of interest's scene");
  kind->newOption("-histogram", (TclObject::OPTION) &TclInrimageROI::histogramCmd,0,
		  "returns the histogram associated with the region of interest"); 
  kind->newOption("-save", (TclObject::OPTION) &TclInrimageROI::saveCmd, "[%s]",
		  "save a region of interest"); 
  kind->newOption("-realName", (TclObject::OPTION) &TclInrimageROI::realNameCmd, "[%s]",
		  "get/set the name of the ROI"); 
  kind->newOption("-binary", (TclObject::OPTION) &TclInrimageROI::binaryImageCmd, 0,"creates and returns a binary image having the same size as the grey level image and that represent the same ROI");

  kind->newOption("-morpho", (TclObject::OPTION) &TclInrimageROI::morphoCmd, "[erosion | dilation] [ iter %u | con <4|6|8|10|18|26> ]",
		  "apply dilation or erosion to the region of interest"); 
  kind->newOption("-saveOption",(TclObject::OPTION)  &TclInrimageROI::saveOptionCmd,"[moduleName=%s] [true|false]",
		  "when no parameters is provided, returns the list of modules. If one module name is provided then returns a boolean indicating if the module should be saved or not. If 2 parameters are provided then set the selection mode of a module for writing");
  kind->newOption("-loadOption",(TclObject::OPTION)  &TclInrimageROI::loadOptionCmd,"[moduleName=%s] [true|false]",
		  "when no parameters is provided, returns the list of modules. If one module name is provided then returns a boolean indicating if the module should be read. If 2 parameters are provided then set the selection mode of a module for reading");
  kind->newOption("-intensityImage", (TclObject::OPTION) &TclInrimageROI::intensityImageCmd, "[inr%u]",
		  "set or get the intensity image attached to the ROI");
  kind->newOption("-color", (TclObject::OPTION) &TclInrimageROI::colorCmd, "[%f %f %f]",
		  "get/set the color of the region of interest");
  kind->newOption("-meanVariance", (TclObject::OPTION) &TclInrimageROI::meanVarianceCmd, "[%f]",
		  "returns the robust estimate of the mean and variance given the rejectRate (default to 0.0)");
  kind->newOption("-visibility", (TclObject::OPTION) &TclInrimageROI::visibilityCmd, "[%b]",
		  "get/set the visibility of a Region of Interest");
  kind->newOption("-opacity", (TclObject::OPTION) &TclInrimageROI::opacityCmd, "[full|opacity75|opacity50|opacity25|opacity12]",
		  "get/set the opacity of the region for display");
  kind->newOption("-mesh", (TclObject::OPTION) &TclInrimageROI::meshCmd, 0,
		  "get/set the name of the mesh that has generated the roi. returns \"none\" if no surface is available");
  kind->newOption("-centerAndRadius3D",(TclObject::OPTION) &TclInrimageROI::centerAndRadius3DCmd,0,
		  "returns the center and radius of the object in world coordinates");

}
} // namespace yav
