#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/TclInrimageView.h"
#include <inrimage/TclInrimageView.h>
#include <inrimage/TclInrimage.h>
#include <graphics2D/TclSliceCamera.h>
#include <inrimage/TclInrimageBackground.h>


namespace yav {
  
  TclInrimageView::InvalidDisplayMode::InvalidDisplayMode(const std::string& from, 
							  const std::string& error) : Exception(from, error) {
  };





TclInrimageView::NORM_MODE TclInrimageView::normModeByName (const char *name) const {

  if(!strcmp(name, "LINEAR")) return TclInrimageView::LINEAR;
  else if(!strcmp(name, "EXP")) return TclInrimageView::EXPONENTIAL;
  else return (TclInrimageView::NORM_MODE) -1;

}

  const char *TclInrimageView::normModeName(TclInrimageView::NORM_MODE mode) const {
    char proc[] = "const char *TclInrimageView::normModeName(TclInrimageView::NORM_MODE mode)";
    switch(mode) {
    default:
      CERR << "Wrong normMode in " << proc << ENDL;
      return "wrong normMode";
      break;
    case LINEAR:
      return "LINEAR";
    case EXPONENTIAL:
      return "EXPONENTIAL";
    }
  }


  const char *TclInrimageView::displayModeName(TclInrimageView::DISPLAY_MODE mode) const {
    char proc[] = "const char *TclInrimageView::displayModeName(TclInrimageView::DISPLAY_MODE mode)";
    switch (getClient()->getInrimage()->getVectorialMode()) {
    default:
      CERR << "Wrong displayMode in " << proc << ENDL;
      return "wrong displayMode";
      break;
    case VM_SCALAR:
      return "Information non available - Image should be Vectorial ";
      break;
    case VM_INTERLACED:
    case VM_NON_INTERLACED:
      switch(mode) {
      default:
	CERR << "Wrong displayMode in " << proc << ENDL;
	return "wrong displayMode";
	break;
      case TclInrimageView::WITHOUT_VECTOR:
	return "WITHOUT_VECTOR";
      case TclInrimageView::WITH_VECTOR:
	return "WITH_VECTOR";
      case TclInrimageView::VECTOR_ONLY:
	return "VECTOR_ONLY";
      }
    }
  }
      
TclInrimageView::DISPLAY_MODE TclInrimageView::displayModeByName (const char *name) const  throw (InvalidDisplayMode){
  char from[] = "TclInrimageView::DISPLAY_MODE TclInrimageView::displayModeByName (const char *name) const  throw (InvalidDisplayMode)";
  if(!strcmp(name, "WITHOUT_VECTOR")) return TclInrimageView::WITHOUT_VECTOR;
  else if(!strcmp(name, "WITH_VECTOR")) return TclInrimageView::WITH_VECTOR;
  else if(!strcmp(name, "VECTOR_ONLY")) return TclInrimageView::VECTOR_ONLY;
  else throw InvalidDisplayMode(from, "Display mode must be WITHOUT_VECTOR, WITH_VECTOR or VECTOR_ONLY");

}



TclInrimageView::TclInrimageView(TclModule *m, TclInrimage *inr,
				 TclScene2D *scene, bool _createGLScale)
  :TclScene2DObject(typeid(TclInrimageView),
		  new InrimageView(inr->getClient(), scene->getClient(),_createGLScale), m) 
{
  switch (inr->getClient()->getVectorialMode()) {
  default:
    displayMode = (TclInrimageView::DISPLAY_MODE) -1;
    break;
  case VM_SCALAR:
    displayMode = (TclInrimageView::DISPLAY_MODE) -1;
    break;
  case VM_INTERLACED:
  case VM_NON_INTERLACED:
    switch(inr->getType()){
    default:
      CERR << "Wrong type in template <class T> void computeNorms(T *type);" << ENDL;
      break;
    case Inrimage::WT_FLOAT_VECTOR:
      computeNorms((float *)0);
      break;
    case Inrimage::WT_DOUBLE_VECTOR:
      computeNorms((double *)0);
      break;
    }

    displayMode = TclInrimageView::WITHOUT_VECTOR;
    break;
  }
  normMode = TclInrimageView::LINEAR;
  under_sample_x=9;
  under_sample_y=9;
  under_sample_z=9;
  if (inr->getClient()->getZ()<9)
	  under_sample_z=1;

  vector_width =1.0;
  vector_scale = 1.0;
  withArrow=true;

  done(typeid(TclInrimageView),"inrimage::installInrviewMenu","inrimage::uninstallInrviewMenu");
  getClient()->focusScene();


}

TclInrimageView::TclInrimageView(TclScene2D *scene, const std::type_info& id,
				 InrimageView *inrview) :
  TclScene2DObject(id,inrview,scene->getModule()), displayMode(TclInrimageView::WITHOUT_VECTOR)
{
  

  switch (inrview->getInrimage()->getVectorialMode()) {
  default:
    displayMode = (TclInrimageView::DISPLAY_MODE) -1;
    break;
  case VM_SCALAR:
    displayMode = (TclInrimageView::DISPLAY_MODE) -1;
    break;
  case VM_INTERLACED:
  case VM_NON_INTERLACED:
    switch(inrview->getInrimage()->getType()){
    default:
      CERR << "Wrong type in template <class T> void computeNorms(T *type);" << ENDL;
      break;
    case Inrimage::WT_FLOAT_VECTOR:
      computeNorms((float *)0);
      break;
    case Inrimage::WT_DOUBLE_VECTOR:
      computeNorms((double *)0);
      break;
    }

    displayMode = TclInrimageView::WITHOUT_VECTOR;
    break;
  }
  under_sample_x=9;
  under_sample_y=9;
  under_sample_z=9;
  if (inrview->getInrimage()->getZ()<9)
	  under_sample_z=1;

  vector_width =1.0;
  vector_scale =1.0;
  normMode = TclInrimageView::LINEAR;
  vector_scale = 1;
  withArrow=true;

}

TclInrimageView::~TclInrimageView() {
  deleted(typeid(TclInrimageView));
}

void TclInrimageView::init(const char *install, const char *uninstall) {

}

int TclInrimageView::destroyCmd(int argc, char **argv) {
  TclModule *m;
  unsigned int ref;

  if(argc != 0) return COMMAND_SYNOPSIS;
  m = getModule();
  ref = getRef();
  delete this;
  m->setResult("%s%u", TclInrimageView::getObjectName(), ref);
  return TCL_OK;
}


int TclInrimageView::sceneCmd(int argc, char **argv) {
  if(argc != 0) return COMMAND_SYNOPSIS;
  setResult("%s%u",
	    TclObjectKind::getObjectName(typeid(TclScene2D)),
	    ((TclScene2D *)getClient()->getScene()->getInterface())->getRef());
  return TCL_OK;
}

int TclInrimageView::GLScaleCmd(int argc, char **argv) {
  if(argc != 2) return COMMAND_SYNOPSIS;
  TclSliceCamera *slicecam= (TclSliceCamera *) TclObjectKind::getObject(typeid(TclSliceCamera),argc, argv, 0);
  if (!slicecam)
    {
      setResult("%s is not a slice camera",argv[0]);
      return TCL_ERROR;
    }
  TclInrimageBackground *b=(TclInrimageBackground *)
    getClient()->getBackgroundFromCamera(slicecam->getClient());
  if (!strcmp(argv[1],"-min"))
    {
      if (b->getMinScale())
	{
	  setResult(b->getMinScale()->name());
	}
      return TCL_OK;
    }
  else if (!strcmp(argv[1],"-max"))
    {
      if (b->getMaxScale())
	{
	  setResult(b->getMaxScale()->name());
	}
      return TCL_OK;
    }
  else return COMMAND_SYNOPSIS;
}

int TclInrimageView::inrimageCmd(int argc, char **argv) {
  TclInrimage *tinr;

  
  if(argc > 1) return COMMAND_SYNOPSIS;
  if(argc == 1) {
    tinr = TclInrimage::getObject(argc, argv, 0);
    if(!tinr) return COMMAND_SYNOPSIS;
    getClient()->changeInrimage(tinr->getClient());
  }
  Inrimage *inr=getClient()->getInrimage();
  tinr=TclInrimage::getInterface(inr);

  if(tinr)
    setResult("%s%u", 
	      TclObjectKind::getObjectName(typeid(TclInrimage)),
	      tinr->getRef());
  else
    setResult("no inrimage");
  return TCL_OK;
}

int TclInrimageView::vectorScaleCmd(int argc, char *argv[]) {
  if (argc>1) return COMMAND_SYNOPSIS;

  if (argc == 1) {
    if(sscanf(argv[0], "%f", &vector_scale) != 1) return COMMAND_SYNOPSIS;
    std::list<InrimageBackground *>::iterator it = getClient()->backgroundList.begin();
    for (; it!=getClient()->backgroundList.end(); ++it){
      (*it)->redraw();
      (*it)->refreshCamera();
    }
  }
  setResult("vector rescaled %f", vector_scale);
  return TCL_OK;
}

int TclInrimageView::vectorWidthCmd(int argc, char *argv[]) {
  if (argc>1) return COMMAND_SYNOPSIS;

  if (argc == 1) {
    if(sscanf(argv[0], "%f", &vector_width) != 1) return COMMAND_SYNOPSIS;
    std::list<InrimageBackground *>::iterator it = getClient()->backgroundList.begin();
    for (; it!=getClient()->backgroundList.end(); ++it){
      (*it)->redraw();
      (*it)->refreshCamera();
    }
  }
  setResult("vector rescaled %f", vector_scale);
  return TCL_OK;
}

int TclInrimageView::displayModeCmd(int argc, char *argv[]) {
  
  char from[] = "int TclInrimageView::displayModeCmd(int argc, char *argv[])";
  if (argc >= 1) {
      if ((argc!=1) && ( argc!=6) && ( argc!=10) && ( argc!=7)) return COMMAND_SYNOPSIS;
      
      try {
	displayMode = displayModeByName(argv[0]);
      }
      catch(InvalidDisplayMode& e) {
	e.push(from);
	return COMMAND_SYNOPSIS;
      }

      if ((argc>1) &&(displayMode == TclInrimageView::WITH_VECTOR || displayMode == TclInrimageView::VECTOR_ONLY)) {
	if ((argc!=6) && (argc!=10) && (argc!=7)) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[1], "%u", &(under_sample_x)) != 1) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[2], "%u", &(under_sample_y)) != 1) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[3], "%u", &(under_sample_z)) != 1) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[4], "%f", &(vector_width)) != 1) return COMMAND_SYNOPSIS;
	  if(sscanf(argv[5], "%f", &(vector_scale)) != 1) return COMMAND_SYNOPSIS;
	  if (argc==6)
	    normMode = LINEAR;
	  if (argc==7)
	    if(!strcmp(argv[6], "normalized")) normMode = NORMALIZED;
	    else return COMMAND_SYNOPSIS;
	  if (argc==10) {
	    if(!strcmp(argv[6], "alpha"))
	      if(sscanf(argv[7], "%f", &(alpha)) != 1) return COMMAND_SYNOPSIS;
	    if(!strcmp(argv[8], "beta"))
	      if(sscanf(argv[9], "%f", &(beta)) != 1) return COMMAND_SYNOPSIS;
	    normMode = EXPONENTIAL;
	  }
      }
      
      std::list<InrimageBackground *>::iterator it = getClient()->backgroundList.begin();
      for (; it!=getClient()->backgroundList.end(); ++it){
	(*it)->redraw();
	//(*it)->refreshCamera();
      }
  };
  if (normMode==LINEAR)
  setResult("%s %u %u %u %f %f", displayModeName(displayMode),under_sample_x,under_sample_y,under_sample_z,vector_width,vector_scale); 
  else if (normMode==NORMALIZED)
  setResult("%s %u %u %u %f %f normalized", displayModeName(displayMode),under_sample_x,under_sample_y,under_sample_z,vector_width,vector_scale); 
  else 
  setResult("%s %u %u %u %f %f alpha %f beta %f", displayModeName(displayMode),under_sample_x,under_sample_y,
  under_sample_z,vector_width,vector_scale,alpha,beta); 

  return TCL_OK;
}

int TclInrimageView::arrowCmd(int argc, char **argv) {

  if ((argc!=0) && (argc!=1))  return COMMAND_SYNOPSIS;
  if (argc==0)
    {
      getModule()->setResult(TclModule::boolString(withArrow));
      return TCL_OK;
    }
  else
    {
      if (!strcmp(argv[0],"true"))
	withArrow=true;
      else if (!strcmp(argv[0],"false"))
	withArrow=false;
      else  return COMMAND_SYNOPSIS;
      getModule()->setResult(TclModule::boolString(withArrow));
      std::list<InrimageBackground *>::iterator it = getClient()->backgroundList.begin();
      for (; it!=getClient()->backgroundList.end(); ++it){
	(*it)->redraw();
	(*it)->refreshCamera();
      }
      return TCL_OK;
    }
}

int TclInrimageView::backgroundCmd(int argc, char **argv) {
  TclSliceCamera *slicecam;

  if(argc != 1) return COMMAND_SYNOPSIS;
  if (!strcmp(argv[0],"all")) {
    std::list<InrimageBackground *>::iterator itb;
    TclString out;
    for (itb=getClient()->firstBackground();itb!=getClient()->lastBackground();
	 itb++) {
      out+=((TclInrimageBackground *)(*itb))->name();
      out+=" ";
    }
    if (out==EMPTY_STRING) 
      setResult("no background");
    else
    setResult(out);
  }
  else {
    slicecam = TclSliceCamera::getObject(argc, argv, 0);
    if(!slicecam) return COMMAND_SYNOPSIS;

    TclInrimageBackground *b= (TclInrimageBackground *)
      getClient()->getBackgroundFromCamera(slicecam->getClient());

    if(b)
      setResult("%s",(char *) b->name()); 
    else
      setResult("no background");
  }
  return TCL_OK;
}




/* "inr%u [scn2D%u]",
   "create an inrview oject to view given inrimage in given scene (a new scene is created if none is provided)" */
int TclInrimageView::create(TclModule *m, int argc, char **argv) {
  TclInrimage *inr = 0;
  TclScene2D *scn = 0;
  TclInrimageView *inrv;
  int i;
  bool addGLScale=true;

  if(argc > 4) return COMMAND_SYNOPSIS;
  for(i = 1; i < argc; i++) {
    if(!inr) {
      inr = TclInrimage::getObject(argc, argv, i);
      if(!inr) {
	if(!scn) scn = (TclScene2D *) TclObjectKind::getObject(typeid(TclScene2D), argc, argv, i);
	else return COMMAND_SYNOPSIS;
	if(!scn) return COMMAND_SYNOPSIS;
      }
    }
    else if (!strcmp(argv[i],"-noGLScale"))
      {
	addGLScale=false;
      }
    else if(!scn) {
      scn = (TclScene2D *) TclObjectKind::getObject(typeid(TclScene2D), argc, argv, i);
      if(!scn) return COMMAND_SYNOPSIS;
    }
    else return COMMAND_SYNOPSIS;
  }
  if(!inr) return COMMAND_SYNOPSIS;
  if(!scn) scn = new TclScene2D(m);

  inrv = new TclInrimageView(m, inr, scn,addGLScale);
  m->setResult("%s%u", TclObjectKind::getObjectName(typeid(TclInrimageView)), inrv->getRef());
  return TCL_OK;
}


} // namespace yav


