#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/TclOverlayBackground.h"
#pragma implementation "inrimage/backgroundOverlayRedraw.h"

#include <inrimage/TclOverlayBackground.h>
#include <graphics2D/TclSliceCamera.h>
#include <inrimage/TclInrimage.h>
#include <inrimage/TclColorMap.h>
#include "backgroundRedraw.h"
#include "backgroundOverlayRedraw.h"
#include "tkwrapper.h"

namespace yav {


TclOverlayBackground::TclOverlayBackground(TclModule *m,
					   OverlayInrimageView *inrv, SliceCamera *cam,
					   bool useMinGLScale ,
					   MinGLScale *minScale,
					   bool useMaxGLScale,
					   MaxGLScale *maxScale, 
					   bool useSecondMinScale,
					   MinGLScale *_secondMinScale,
					   bool useSecondMaxScale,
					   MaxGLScale *_secondMaxScale,
					   bool useOpacityScale,
					   OpacityScale *opacityScale,  
					   const char *color) : 
  TclInrimageBackground(m,typeid(TclOverlayBackground),inrv,cam,minScale,maxScale),Owner<OpacityScale>(opacityScale),secondMinScale(this,_secondMinScale),secondMaxScale(this,_secondMaxScale) {

  done(typeid(TclOverlayBackground));
  secondGlmin = Voxel::voxelByType(getSecondInrimage()->getType());
  secondGlmax = Voxel::voxelByType(getSecondInrimage()->getType());
  glBounds();

  
  redColorMap2=AffineColorMap;
  greenColorMap2=AffineColorMap;
  blueColorMap2=AffineColorMap;
  transfo2=AFFINE;

  slicecam->setBackground(this);
  init(useMinGLScale,minScale,useMaxGLScale,maxScale,color);
  if(useSecondMaxScale) {
    if(!secondMaxScale) {
      TclString sname = slicecam->getInterface()->getWindow() + ".max2";
      getModule()->tickle("scale %s -orient horizontal -label {max grey level (2)} -width 10 -font fixed -bg %s", (char *) sname, color);
      secondMaxScale = new MaxGLScale(getModule(), sname,this);
      getModule()->tickle("pack %s -side bottom -expand true -fill x",
			  (char *) sname);
    } 
    else
      *secondMaxScale += this;
  }

   if(useSecondMinScale) {
    if(!secondMinScale) {
      TclString sname =  slicecam->getInterface()->getWindow() + ".min2";
      getModule()->tickle("scale %s -orient horizontal -label {min grey level (2)} -width 10 -font fixed -bg %s", (char *) sname, color);
      secondMinScale = new MinGLScale(getModule(), sname,this);
      getModule()->tickle("pack %s -side bottom -expand true -fill x",
			  (char *) sname);
    } 
    else
      *secondMinScale += this;
  }
   if(useOpacityScale) {
    if(!Owner<OpacityScale>::pointer) {
      TclString sname =  slicecam->getInterface()->getWindow() + ".opacity";
      getModule()->tickle("scale %s -orient horizontal -label {opacity} -width 10 -font fixed  -from 0 -to 1 -resolution 0.01 -bg %s", (char *) sname, color);
      Owner<OpacityScale>::pointer = new OpacityScale(getModule(), sname,this);
      getModule()->tickle("pack %s -side bottom -expand true -fill x",
			  (char *) sname);
    } 
    else
      *Owner<OpacityScale>::pointer += this;
  }

  if (slicecam->getScaleCamera())
    {
      ((SliceScale *)slicecam->getScaleCamera())->reset(slicecam,slicecam->sliceName(slicecam->getGeometry()));
    }

  if (getMinScale())
    getMinScale()->reset((Scalable *)this);
  if (getMaxScale())
    getMaxScale()->reset((Scalable *)this);

  if (getSecondMinScale())
    getSecondMinScale()->reset((Scalable *)this);
  if (getSecondMaxScale())
    getSecondMaxScale()->reset((Scalable *)this);

  getOverlayView()->getSliceInterface()->addSlice(this);
  getOverlayView()->getSecondSliceInterface()->addSlice(this);
}

TclOverlayBackground::~TclOverlayBackground() {
  getOverlayView()->getSecondSliceInterface()->deleteSlice(this);
}
void TclOverlayBackground::scaleValue(const TclScale *scale, double val) {
  if (((TclScale *)secondMinScale())==scale)
    setSecondMinThreshold(val);
  else if (((TclScale *)secondMaxScale())==scale)
    setSecondMaxThreshold(val);
  else if (((TclScale *)Owner<OpacityScale>::pointer)==scale)
    setOpacity(val);
  else
    TclInrimageBackground::scaleValue(scale,val);
}
void TclOverlayBackground::addSecondMinGLScale(MinGLScale *min)  {
  secondMinScale = min;
  *secondMinScale += this;
}
void TclOverlayBackground::addSecondMaxGLScale(MaxGLScale *max)  {
  secondMaxScale = max;
  *secondMaxScale += this;
}
double TclOverlayBackground::getMax(const TclScale *scale) const {
  if ((((TclScale *)this->Owner<MinGLScale>::pointer())==scale) ||
      (((TclScale *)this->Owner<MaxGLScale>::pointer())==scale))
    return glmax->getValueAsDouble();
  else if ((((TclScale *)getSecondMinScale())==scale) ||
	   (((TclScale *)getSecondMaxScale())==scale))
    return secondGlmax->getValueAsDouble();
  else 
    return 1.0;

}
double TclOverlayBackground::getMin(const TclScale *scale) const {
  if ((((TclScale *)this->Owner<MinGLScale>::pointer())==scale) ||
      (((TclScale *)this->Owner<MaxGLScale>::pointer())==scale))
    return glmin->getValueAsDouble();
  else if ((((TclScale *)getSecondMinScale())==scale) ||
	   (((TclScale *)getSecondMaxScale())==scale))
    return secondGlmin->getValueAsDouble();
  else 
    return 0.0;
}


void TclOverlayBackground::imageGLChanged(void) {
  glBounds();
  if(getMinScale())
    getMinScale()->reset(this);
  if(getMaxScale())
    getMaxScale()->reset(this);
  if(getSecondMinScale())
    getSecondMinScale()->reset(this);
  if(getSecondMaxScale())
    getSecondMaxScale()->reset(this);
  redraw();
}

void TclOverlayBackground::setGLTransfo2(TclInrimageBackground::GLTRANSFO t,const unsigned char *red,
					 const  unsigned char *green,const  unsigned char *blue) {
  if ((t!=transfo2)|| (t==USER))
    {
      if ((t!=USER) || (red && green && blue))
	{
	  transfo2 = t;
	  switch(t) {
	  case AFFINE:
	    redColorMap2=AffineColorMap;
	    greenColorMap2=AffineColorMap;
	    blueColorMap2=AffineColorMap;
	    break;
	  case THRESHOLD:
	    redColorMap2=ThresholdColorMap;
	    greenColorMap2=ThresholdColorMap;
	    blueColorMap2=ThresholdColorMap;
	    break;
	  case PET_COLORS:
	    redColorMap2=PETColorMapRed;
	    greenColorMap2=PETColorMapGreen;
	    blueColorMap2=PETColorMapBlue;
	    break;
	  case SPECT_COLORS:
	    redColorMap2=SPECTColorMapRed;
	    greenColorMap2=SPECTColorMapGreen;
	    blueColorMap2=SPECTColorMapBlue;
	    break;
	  case JET_COLORS:
	    redColorMap2=JetColorMapRed;
	    greenColorMap2=JetColorMapGreen;
	    blueColorMap2=JetColorMapBlue;
	    break;
	  case HOTIRON_COLORS:
	    redColorMap2=HotIronColorMapRed;
	    greenColorMap2=HotIronColorMapGreen;
	    blueColorMap2=HotIronColorMapBlue;
	    break;
	  case BLUE:
	    redColorMap2=ZeroColorMap;
	    greenColorMap2=ZeroColorMap;
	    blueColorMap2=AffineColorMap;
	    break; 
	  case USER:
	    redColorMap2=(unsigned char *)red;
	    greenColorMap2=(unsigned char *)green;
	    blueColorMap2=(unsigned char *)blue;
	    break;
	  }
	}
    }

  redraw();
}


void TclOverlayBackground::inrimageChanged(void) {
  if (secondGlmin)
    delete secondGlmin;
  if (secondGlmax)
    delete secondGlmax;
  secondGlmin = Voxel::voxelByType(getInrimage()->getType());
  secondGlmax = Voxel::voxelByType(getInrimage()->getType());
  TclInrimageBackground::inrimageChanged();
  if(getSecondMinScale())
    getSecondMinScale()->reset(this);
  if(getSecondMaxScale())
    getSecondMaxScale()->reset(this);
  
}
Inrimage *TclOverlayBackground::getSecondInrimage(void) const {
  return (getOverlayView()->getSecondInrimage());
}
  
void TclOverlayBackground::setSecondMinThreshold(double val)
  throw(Voxel::InvalidValue) {
  if(secondGlmin->getValueAsDouble() != val) {
    secondGlmin->setValueFromDouble(val);
    redraw();
  }
}


void TclOverlayBackground::setSecondMaxThreshold(double val)
  throw(Voxel::InvalidValue) {
  if(secondGlmax->getValueAsDouble() != val) {
    secondGlmax->setValueFromDouble(val);
    redraw();
  }
}

void TclOverlayBackground::setOpacity(double val)  {
  if ((val>=0) && (val<=1.0))
    getOpacityScale()->set(val);
    redraw();
}



void TclOverlayBackground::setSecondGlThresholds(double min, double max)
  throw(Voxel::InvalidValue) {
  if(secondGlmin->getValueAsDouble() != min && secondGlmax->getValueAsDouble() != max) {
    secondGlmin->setValueFromDouble(min);
    secondGlmax->setValueFromDouble(max);
    redraw();
  }
}


void TclOverlayBackground::setSecondGlThresholds(const TclString& min, const TclString& max)
  throw(Voxel::InvalidValue) {

  if(min != EMPTY_STRING) secondGlmin->scan(min);
  if(max != EMPTY_STRING) secondGlmax->scan(max);

  redraw();
}
TclString TclOverlayBackground::getLabel(int &x,int &y){
  TclString s;
  VoxelCoordinate *vc = coordinates(x, y);
  Voxel *v = getInrimage()->getVoxel(vc->x, vc->y, vc->z);
  TclString sval;
  v->print(sval);
  Voxel *v2 = getSecondInrimage()->getVoxel(vc->x, vc->y, vc->z);
  TclString sval2;
  v2->print(sval2);
  s.sprintf("%3u %3u %3u : (%s %s)", vc->x, vc->y, vc->z,(char *) sval,
	    (char *) sval2);
  delete vc;
  delete v;
  delete v2;
  return(s);
}

void TclOverlayBackground::glBounds(void) {
  getInrimage()->getBounds(glmin, glmax);
  getSecondInrimage()->getBounds(secondGlmin, secondGlmax);
}

int TclOverlayBackground::create(TclModule *m, int argc, char **argv) {
  TclScene2D *scene = TclScene2D::getDefaultScene(m);
  int i;
  TclInrimageBackground *b;
  TclInrimage *inr,*inr2;
  MinGLScale *minScale=0;
  MaxGLScale *maxScale=0;
  MinGLScale *minScale2=0;
  MaxGLScale *maxScale2=0;
  OpacityScale *opac=0;
  bool useMinScale=false;
  bool useMaxScale=false;
  bool useOpacity=false;
  bool useSecondMinScale=false;
  bool useSecondMaxScale=false;
  if (argc<3)
    return COMMAND_SYNOPSIS;
  inr = TclInrimage::getObject(argc, argv, 1);
  if (!inr)
    {
      m->setResult("%s is not an image",argv[1]);
      return TCL_ERROR;
    }
  inr2 = TclInrimage::getObject(argc, argv, 2);
  if (!inr2)
    {
      m->setResult("%s is not an image",argv[2]);
      return TCL_ERROR;
    }
  TclSliceCamera *cam=TclSliceCamera::getObject(argc, argv, 3);
  if (!cam)
    {
      m->setResult("%s is not an slice camera",argv[3]);
      return TCL_ERROR;
    }

  for(i=4; i < argc; i++) {
    if(!strcmp(argv[i], "-scene")) {
      if( ++i >= argc) return COMMAND_SYNOPSIS;
      scene = (TclScene2D *)
	TclObjectKind::getObject(typeid(TclScene2D), argc, argv, i);
      if(!scene) {
	m->setResult("Illegal scene name : %s",argv[i]);
	return TCL_ERROR;
      }
    }
    if(!strcmp(argv[i], "-min")) {
      if( ++i >= argc) return COMMAND_SYNOPSIS;
      minScale= (MinGLScale *) 	TclObjectKind::getObject(typeid(MinGLScale), argc, argv, i);
      if (!minScale) {
	m->setResult("Illegal MinGLScale name : %s",argv[i]);
	return TCL_ERROR;
      }
      else
	useMinScale=true;
    }
    if(!strcmp(argv[i], "-min2")) {
      if( ++i >= argc) return COMMAND_SYNOPSIS;
      minScale2= (MinGLScale *)	TclObjectKind::getObject(typeid(MinGLScale), argc, argv, i);
      if (!minScale2) {
	m->setResult("Illegal MinGLScale name : %s",argv[i]);
	return TCL_ERROR;
      }
      else
	useSecondMinScale=true;
    }
    if(!strcmp(argv[i], "-max")) {
      if( ++i >= argc) return COMMAND_SYNOPSIS;
      maxScale= (MaxGLScale *) 	TclObjectKind::getObject(typeid(MaxGLScale), argc, argv, i);
      if (!maxScale) {
	m->setResult("Illegal MaxGLScale name : %s",argv[i]);
	return TCL_ERROR;
      }
      else
	useMaxScale=true;
    }
    if(!strcmp(argv[i], "-max2")) {
      if( ++i >= argc) return COMMAND_SYNOPSIS;
      maxScale2= (MaxGLScale *) 	TclObjectKind::getObject(typeid(MaxGLScale), argc, argv, i);
      if (!maxScale2) {
	m->setResult("Illegal MaxGLScale name : %s",argv[i]);
	return TCL_ERROR;
      }
      else
	useSecondMaxScale=true;
    }
    if(!strcmp(argv[i], "-opac")) {
      if( ++i >= argc) return COMMAND_SYNOPSIS;
      opac= (OpacityScale *) 	TclObjectKind::getObject(typeid(OpacityScale), argc, argv, i);
      if (!opac) {
	m->setResult("Illegal OpacityScale name : %s",argv[i]);
	return TCL_ERROR;
      }
      else
	useOpacity=true;
    }
    else return COMMAND_SYNOPSIS;
  }

  b = new TclOverlayBackground(m,new OverlayInrimageView(inr->getClient(),inr2->getClient(),scene->getClient()),cam->getClient(),
			       useMinScale,minScale,
			       useMaxScale,maxScale,
			       useSecondMinScale,minScale2,
			       useSecondMaxScale,maxScale2,useOpacity,opac);
  m->setResult("%s%u", b->getObjectName(), b->TclObject::getRef());
  return TCL_OK;
}

int TclOverlayBackground::scalemode2Cmd(int argc, char **argv) {
  if(argc > 0) {
    GLTRANSFO t;
    if(!strcmp(argv[0], "affine")) t = AFFINE;
    else if(!strcmp(argv[0], "threshold")) t = THRESHOLD;
    else if(!strcmp(argv[0], "spect")) t = SPECT_COLORS;
    else if(!strcmp(argv[0], "hotiron")) t = HOTIRON_COLORS;
    else if(!strcmp(argv[0], "jet")) t = JET_COLORS;
    else if(!strcmp(argv[0], "blue")) t = BLUE;
    else if(!strcmp(argv[0], "pet")) t = PET_COLORS;
    else if(!strcmp(argv[0], "user")) 
      {
	t = USER;
	if (argc>1)
	  {
	    TclColorMap *tc=(TclColorMap *)
	      TclObjectKind::getObject(typeid(TclColorMap), argc, argv, 1);
	    if (tc)
	      setGLTransfo2(t,tc->getRed(),tc->getGreen(),tc->getBlue());
	  }
      }
    else return COMMAND_SYNOPSIS;
    setGLTransfo2(t);
  }
  switch(getGLTransfo2()) {
  case AFFINE:
    setResult("affine");
    break;
  case THRESHOLD:
    setResult("threshold");
    break;
  case PET_COLORS:
    setResult("pet");
    break;
  case JET_COLORS:
    setResult("jet");
    break;
  case SPECT_COLORS:
    setResult("spect");
    break;
  case HOTIRON_COLORS:
    setResult("hotiron");
    break;
  case BLUE:
    setResult("blue");
    break; 
  case USER:
    setResult("blue");
    break;
  }
  return TCL_OK;
}
int TclOverlayBackground::opacityCmd(int argc, char **argv) {
  if (argc==0)
    {
      if (Owner<OpacityScale>::pointer)
	setResult("%s",(char *)Owner<OpacityScale>::pointer->name());
      return TCL_OK;
    }
  else if (argc==1)
    { 
      OpacityScale *opac= (OpacityScale *) 	
	TclObjectKind::getObject(typeid(OpacityScale), argc, argv, 0);
      if (opac)
	{
	  Owner<OpacityScale>::pointer=opac;
	  *Owner<OpacityScale>::pointer+=this;
	  setResult("%s",(char *)Owner<OpacityScale>::pointer->name());
	  return TCL_OK;
	}
      else
	{
	  setResult("%s is not an opacity scale",argv[0]);
	  return TCL_ERROR;
	}
    }
  else
    return COMMAND_SYNOPSIS;
}
int TclOverlayBackground::greyLevelCmd(int argc, char **argv) {
  if (argc==1)
    {
      if (!strcmp(argv[0],"min"))
	{
	  if (Owner<MinGLScale>::pointer)
	    setResult("%s",(char *)Owner<MinGLScale>::pointer->name());
	  return TCL_OK;
	}
      else if (!strcmp(argv[0],"max"))
	{
	  if (Owner<MaxGLScale>::pointer)
	    setResult("%s",(char *)Owner<MaxGLScale>::pointer->name());
	  return TCL_OK;
	}
      else if (!strcmp(argv[0],"min2"))
	{
	  if (secondMinScale())
	    setResult("%s",(char *)secondMinScale()->name());
	  return TCL_OK;
	}
      else if (!strcmp(argv[0],"max2"))
	{
	  if (secondMaxScale())
	    setResult("%s",(char *)secondMaxScale()->name());
	  return TCL_OK;
	}
      else
	return COMMAND_SYNOPSIS;	
    }
  else if (argc==2)
    { 
      if (!strcmp(argv[0],"min"))
	{
	  MinGLScale *gl= dynamic_cast<MinGLScale *>(	
	    TclObjectKind::getObject(typeid(GLScale), argc, argv, 1));
	  if (gl)
	    {
	      addMinGLScale(gl);
	      setResult("%s",(char *)getMinScale()->name());
	      return TCL_OK;
	    }
	  else
	    {
	      setResult("%s is not a minGLScale",argv[1]);
	      return TCL_ERROR;
	    }
	}
      else if (!strcmp(argv[0],"max"))
	{
	  MaxGLScale *gl= dynamic_cast<MaxGLScale *>(	
	    TclObjectKind::getObject(typeid(GLScale), argc, argv, 1));
	  if (gl)
	    {
	      addMaxGLScale(gl);
	      setResult("%s",(char *)getMaxScale()->name());
	      return TCL_OK;
	    }
	  else
	    {
	      setResult("%s is not a maxGLScale",argv[1]);
	      return TCL_ERROR;
	    }
	}
      else if (!strcmp(argv[0],"min2"))
	{
	  MinGLScale *gl= dynamic_cast<MinGLScale *>(	
	    TclObjectKind::getObject(typeid(GLScale), argc, argv, 1));
	  if (gl)
	    {
	      addSecondMinGLScale(gl);
	      setResult("%s",(char *)getSecondMinScale()->name());
	      return TCL_OK;
	    }
	  else
	    {
	      setResult("%s is not a minGLScale",argv[1]);
	      return TCL_ERROR;
	    }
	}
      else if (!strcmp(argv[0],"max2"))
	{
	  MaxGLScale *gl= dynamic_cast<MaxGLScale *>(	
	    TclObjectKind::getObject(typeid(GLScale), argc, argv, 1));
	  if (gl)
	    {
	      addSecondMaxGLScale(gl);
	      setResult("%s",(char *)getSecondMaxScale()->name());
	      return TCL_OK;
	    }
	  else
	    {
	      setResult("%s is not a maxGLScale",argv[1]);
	      return TCL_ERROR;
	    }
	}
      else
	return COMMAND_SYNOPSIS;

    }
  else
    return COMMAND_SYNOPSIS;
}

#define DRAWXY(TYPE,INR,QQ) drawOverlayXY((TYPE) 0,INR,QQ, slicecam, visible, \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                secondGlmin->getValueAsDouble(), \
			        secondGlmax->getValueAsDouble(),\
                                block, \
				redColorMap,greenColorMap, \
				blueColorMap, \
				redColorMap2,greenColorMap2, \
				blueColorMap2,sx,sy,rx,ry, getOpacity())

#define DRAWOVERLAYXYVECTOR(TYPE,INR,QQ) drawOverlayXYVector((TYPE) 0,INR,QQ, slicecam, visible, \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                secondGlmin->getValueAsDouble(), \
			        secondGlmax->getValueAsDouble(),\
                                block, \
				redColorMap,greenColorMap, \
				blueColorMap, \
				redColorMap2,greenColorMap2, \
				blueColorMap2,sx,sy,rx,ry, getOpacity())

#define DRAWXZ(TYPE,INR,QQ) drawOverlayXZ((TYPE) 0,INR,QQ, slicecam, visible, \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                secondGlmin->getValueAsDouble(), \
			        secondGlmax->getValueAsDouble(),\
                                block, \
				redColorMap,greenColorMap, \
				blueColorMap, \
				redColorMap2,greenColorMap2, \
				blueColorMap2,sx,sz,rx,rz, getOpacity())

#define DRAWOVERLAYXZVECTOR(TYPE,INR,QQ) drawOverlayXZVector((TYPE) 0,INR,QQ, slicecam, visible, \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                secondGlmin->getValueAsDouble(), \
			        secondGlmax->getValueAsDouble(),\
                                block, \
				redColorMap,greenColorMap, \
				blueColorMap, \
				redColorMap2,greenColorMap2, \
				blueColorMap2,sx,sz,rx,rz, getOpacity())

#define DRAWYZ(TYPE,INR,QQ) drawOverlayYZ((TYPE) 0,INR,QQ, slicecam, visible, \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                secondGlmin->getValueAsDouble(), \
			        secondGlmax->getValueAsDouble(),\
                                block, \
				redColorMap,greenColorMap, \
				blueColorMap, \
				redColorMap2,greenColorMap2, \
				blueColorMap2,sy,sz,ry,rz, getOpacity())

#define DRAWOVERLAYYZVECTOR(TYPE,INR,QQ) drawOverlayYZVector((TYPE) 0,INR,QQ, slicecam, visible, \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                secondGlmin->getValueAsDouble(), \
			        secondGlmax->getValueAsDouble(),\
                                block, \
				redColorMap,greenColorMap, \
				blueColorMap, \
				redColorMap2,greenColorMap2, \
				blueColorMap2,sy,sz,ry,rz, getOpacity())

#define DRAWXYVECTOR(TYPE,INR) drawXYVector((TYPE) 0,INR,slicecam,visible, \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                block, \
				redColorMap,greenColorMap, \
				blueColorMap,sx,sy,rx,ry)

#define DRAWXYVECTOR_VECTOR(TYPE,INR,SAMPLE_X,SAMPLE_Y,SAMPLE_Z,V_WIDTH,V_SCALE,NORMMODE,A,B,WA,NORM2D,NORM3D,N3D_MAX,N3D_MIN) drawXYVector_Vector((TYPE) 0,INR,SAMPLE_X,SAMPLE_Y,SAMPLE_Z, \
								slicecam,visible, \
								glmin->getValueAsDouble(), \
								glmax->getValueAsDouble(),\
								block, \
								redColorMap,greenColorMap, \
								blueColorMap,sx,sy,rx,ry, \
                                                                V_WIDTH,V_SCALE,NORMMODE,A,B,WA,NORM2D,NORM3D,N3D_MAX,N3D_MIN)

#define DRAWXZVECTOR(TYPE,INR) drawXZVector((TYPE) 0,INR,slicecam,visible, \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                block, \
				redColorMap,greenColorMap, \
				blueColorMap,sx,sz,rx,rz)
  
#define DRAWXZVECTOR_VECTOR(TYPE,INR,SAMPLE_X,SAMPLE_Y,SAMPLE_Z,V_WIDTH,V_SCALE,NORMMODE,A,B,WA,NORM2D,NORM3D,N3D_MAX,N3D_MIN) drawXZVector_Vector((TYPE) 0,INR,SAMPLE_X,SAMPLE_Y,SAMPLE_Z, \
								slicecam,visible, \
								glmin->getValueAsDouble(), \
								glmax->getValueAsDouble(),\
								block, \
								redColorMap,greenColorMap, \
				                                blueColorMap,sx,sz,rx,rz, \
                                                                V_WIDTH,V_SCALE,NORMMODE,A,B,WA,NORM2D,NORM3D,N3D_MAX,N3D_MIN)
#define DRAWYZVECTOR(TYPE,INR) drawYZVector((TYPE) 0,INR,slicecam,visible, \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                block, \
				redColorMap,greenColorMap, \
				blueColorMap,sy,sz,ry,rz)

#define DRAWYZVECTOR_VECTOR(TYPE,INR,SAMPLE_X,SAMPLE_Y,SAMPLE_Z,V_WIDTH,V_SCALE,NORMMODE,A,B,WA,NORM2D,NORM3D,N3D_MAX,N3D_MIN) drawYZVector_Vector((TYPE) 0,INR,SAMPLE_X,SAMPLE_Y,SAMPLE_Z, \
								slicecam,visible, \
				                                glmin->getValueAsDouble(), \
				                                glmax->getValueAsDouble(),\
                                                                block, \
				                                redColorMap,greenColorMap, \
				                                blueColorMap,sy,sz,ry,rz, \
                                                                V_WIDTH,V_SCALE,NORMMODE,A,B,WA,NORM2D,NORM3D,N3D_MAX,N3D_MIN)

void TclOverlayBackground::redraw(void) {
  unsigned int plane, i, j, x, y, z, *offset, o;
  register int r;
  
  voxelRatio();
 
  Inrimage *inr=getInrimage();
  Inrimage *inr2=getSecondInrimage();
  plane = inr->getX() * inr->getY();
  o = (unsigned int) slicecam->getSlice();
  double a=getOpacity();
  double b= 1-a;
  switch(slicecam->getGeometry()) {

  /* slice perpendicular to X */
  case SliceCamera::SLICE_X:
    if(o < visible[0] || o > visible[3]) {
      memset((void *) block->pixelPtr, 0, 3 * sy * sz);
      return;
    }

    switch(inr->getType()) {

    case Inrimage::WT_UNSIGNED_CHAR:
      DRAWYZ(unsigned char,inr,inr2);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      DRAWYZ(signed char,inr,inr2);
      break;
 
    case Inrimage::WT_UNSIGNED_SHORT:
      DRAWYZ(unsigned short int,inr,inr2);
      break;

    case Inrimage::WT_SIGNED_SHORT:
      DRAWYZ(short int,inr,inr2);
      break;

    case Inrimage::WT_UNSIGNED_INT:
      DRAWYZ(unsigned int,inr,inr2);
      break;

    case Inrimage::WT_SIGNED_INT:
      DRAWYZ(int,inr,inr2);
      break;

    case Inrimage::WT_UNSIGNED_LONG:
      DRAWYZ(unsigned LONGINT,inr,inr2);
      break;

    case Inrimage::WT_SIGNED_LONG:
      DRAWYZ(LONGINT,inr,inr2);
      break;

    case Inrimage::WT_FLOAT:
      DRAWYZ(float,inr,inr2);
      break;

    case Inrimage::WT_DOUBLE:
      DRAWYZ(double,inr,inr2);
      break;

    case Inrimage::WT_RGB:
      
      offset = new unsigned int[sz];
      for(i = 0; i < sz; i++) offset[i] =
				((int) ((double) i / rz) + visible[2]) * plane;
      
      for(y = 0, j = 0; y < sy; y++) {
	i = ((int)((double) y / ry) + visible[1]) * inr->getX() + o;
	for(z = 0; z < sz; z++) {
	  r = 3 * (offset[z] + i);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r] +
	    b*((unsigned char *) inr2->getData())[r]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+1] +
	    b*((unsigned char *) inr2->getData())[r+1]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+2] +
	    b*((unsigned char *) inr2->getData())[r+2]);
	}
      }
      
      delete[] offset;
      break;

    case Inrimage::WT_RGBA:
      offset = new unsigned int[sz];
      for(i = 0; i < sz; i++) offset[i] =
				((int) ((double) i / rz) + visible[2]) * plane;
      
      for(y = 0, j = 0; y < sy; y++) {
	i = ((int)((double) y / ry) + visible[1]) * inr->getX() + o;
	for(z = 0; z < sz; z++) {
	  r = 4 * (offset[z] + i);
#if (defined (_ALPHA_) || defined (_LINUX_) || defined(WIN32))
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+3] +
	    b*((unsigned char *) inr2->getData())[r+3]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+2] +
	    b*((unsigned char *) inr2->getData())[r+2]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+1] +
	    b*((unsigned char *) inr2->getData())[r+1]);
#else
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r] +
	    b*((unsigned char *) inr2->getData())[r]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+1] +
	    b*((unsigned char *) inr2->getData())[r+1]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+2] +
	    b*((unsigned char *) inr2->getData())[r+2]);
#endif
	}
      }
     
      delete[] offset;
      break;

    case Inrimage::WT_FLOAT_VECTOR:
      //      DRAWOVERLAYYZVECTOR(float,inr,inr2);
      DRAWYZVECTOR(float,inr);
      DRAWYZVECTOR_VECTOR(float,inr2,inrv->getInterface()->under_sample_x,inrv->getInterface()->under_sample_y,inrv->getInterface()->under_sample_z,inrv->getInterface()->vector_width, inrv->getInterface()->vector_scale, inrv->getInterface()->normMode, inrv->getInterface()->alpha, inrv->getInterface()->beta, inrv->getInterface()->withArrow,inrv->getInterface()->norme3D,inrv->getInterface()->norme2D_X,inrv->getInterface()->norme3D_max,inrv->getInterface()->norme3D_min);
      break;

    case Inrimage::WT_DOUBLE_VECTOR:
      DRAWOVERLAYYZVECTOR(double,inr,inr2);
      break;

    default:
      break;
    }

    break;

    /* slice perpendicular to Y */
  case SliceCamera::SLICE_Y:
    if(o < visible[1] || o > visible[4]) {
      memset((void *) block->pixelPtr, 0, 3 * sx * sz);
      return;
    }

    switch(inr->getType()) {

    case Inrimage::WT_UNSIGNED_CHAR:
      DRAWXZ(unsigned char,inr,inr2);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      DRAWXZ(signed char,inr,inr2);
      break;

    case Inrimage::WT_UNSIGNED_SHORT:
      DRAWXZ(unsigned short int,inr,inr2);
      break;

    case Inrimage::WT_SIGNED_SHORT:
      DRAWXZ(short int,inr,inr2);
      break;

    case Inrimage::WT_UNSIGNED_INT:
      DRAWXZ(unsigned int,inr,inr2);
      break;

    case Inrimage::WT_SIGNED_INT:
      DRAWXZ(int,inr,inr2);
      break;

    case Inrimage::WT_UNSIGNED_LONG:
      DRAWXZ(unsigned LONGINT,inr,inr2);
      break;

    case Inrimage::WT_SIGNED_LONG:
      DRAWXZ(LONGINT,inr,inr2);
      break;

    case Inrimage::WT_FLOAT:
      DRAWXZ(float,inr,inr2);
      break;

    case Inrimage::WT_DOUBLE:
      DRAWXZ(double,inr,inr2);
      break;

    case Inrimage::WT_RGB:
      offset = new unsigned int[sx];
      for(i = 0; i < sx; i++) offset[i] = (int) ((double) i / rx) + visible[0];
      o = (unsigned int) slicecam->getSlice() * inr->getX();

      for(z = 0, j = 0; z < sz; z++) {
	i = ((int)((double) z / rz) + visible[2]) * plane + o;
	for(x = 0; x < sx; x++) {
	  r = 3 * (offset[x] + i);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r] +
	    b*((unsigned char *) inr2->getData())[r]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+1] +
	    b*((unsigned char *) inr2->getData())[r+1]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+2] +
	    b*((unsigned char *) inr2->getData())[r+2]);
	}
      }
      
      delete[] offset;
      break;

    case Inrimage::WT_RGBA:
      offset = new unsigned int[sx];
      for(i = 0; i < sx; i++) offset[i] = (int) ((double) i / rx) + visible[0];
      o = (unsigned int) slicecam->getSlice() * inr->getX();

      for(z = 0, j = 0; z < sz; z++) {
	i = ((int)((double) z / rz) + visible[2]) * plane + o;
	for(x = 0; x < sx; x++) {
	  r = 4 * (offset[x] + i);
#if (defined (_ALPHA_) || defined (_LINUX_) || defined(WIN32))
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+3] +
	    b*((unsigned char *) inr2->getData())[r+3]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+2] +
	    b*((unsigned char *) inr2->getData())[r+2]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+1] +
	    b*((unsigned char *) inr2->getData())[r+1]);
#else
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r] +
	    b*((unsigned char *) inr2->getData())[r]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+1] +
	    b*((unsigned char *) inr2->getData())[r+1]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+2] +
	    b*((unsigned char *) inr2->getData())[r+2]);
#endif
	}
      }
       
      delete[] offset;
      break;

    case Inrimage::WT_FLOAT_VECTOR:
      DRAWXZVECTOR(float,inr);
      DRAWXZVECTOR_VECTOR(float,inr2,inrv->getInterface()->under_sample_x,inrv->getInterface()->under_sample_y,inrv->getInterface()->under_sample_z,inrv->getInterface()->vector_width, inrv->getInterface()->vector_scale, inrv->getInterface()->normMode, inrv->getInterface()->alpha, inrv->getInterface()->beta, inrv->getInterface()->withArrow,inrv->getInterface()->norme3D,inrv->getInterface()->norme2D_Y,inrv->getInterface()->norme3D_max,inrv->getInterface()->norme3D_min );
      break;

    case Inrimage::WT_DOUBLE_VECTOR:
      DRAWOVERLAYXZVECTOR(double,inr,inr2);
      break;

    default:
      break;
    }

    break;

    /* slice perpendicular to Z */
  case SliceCamera::SLICE_ALPHA:
  case SliceCamera::SLICE_Z:
    if(slicecam->getGeometry() == SliceCamera::SLICE_Z &&
       (o < visible[2] || o > visible[5])) {
      memset((void *) block->pixelPtr, 0, 3 * sx * sy);
      return;
    }

    switch(inr->getType()) {

    case Inrimage::WT_UNSIGNED_CHAR:
      DRAWXY(unsigned char,inr,inr2);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      DRAWXY(signed char,inr,inr2);
      break;

    case Inrimage::WT_UNSIGNED_SHORT:
      DRAWXY(unsigned short int,inr,inr2);
      break;

    case Inrimage::WT_SIGNED_SHORT:
      DRAWXY(short int,inr,inr2);
      break;

    case Inrimage::WT_UNSIGNED_INT:
      DRAWXY(unsigned int,inr,inr2);
      break;

    case Inrimage::WT_SIGNED_INT:
      DRAWXY(int,inr,inr2);
      break;

    case Inrimage::WT_UNSIGNED_LONG:
      DRAWXY(unsigned LONGINT,inr,inr2);
      break;

    case Inrimage::WT_SIGNED_LONG:
      DRAWXY(LONGINT,inr,inr2);
      break;

    case Inrimage::WT_FLOAT:
      DRAWXY(float,inr,inr2);
      break;

    case Inrimage::WT_DOUBLE:
      DRAWXY(double,inr,inr2);
      break;

    case Inrimage::WT_RGB:
      offset = new unsigned int[sx];
      for(i = 0; i < sx; i++) offset[i] = (int) ((double) i / rx) + visible[0];
      o = (unsigned int) slicecam->getSlice() * plane;
    
      for(y = 0, j = 0; y < sy; y++) {
	i = o + ((int) ((double) y / ry) + visible[1]) * inr->getX();
	for(x = 0; x < sx; x++) {
	  r = 3 * (offset[x] + i);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r] +
	    b*((unsigned char *) inr2->getData())[r]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+1] +
	    b*((unsigned char *) inr2->getData())[r+1]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+2] +
	    b*((unsigned char *) inr2->getData())[r+2]);
	}
      }

      delete[] offset;
      break;

    case Inrimage::WT_RGBA:
      offset = new unsigned int[sx];
      for(i = 0; i < sx; i++) offset[i] = (int) ((double) i / rx) + visible[0];
      o = (unsigned int) slicecam->getSlice() * plane;
    
      for(y = 0, j = 0; y < sy; y++) {
	i = o + ((int) ((double) y / ry) + visible[1]) * inr->getX();
	for(x = 0; x < sx; x++) {
	  r = 4 * (offset[x] + i);
#if (defined (_ALPHA_) || defined (_LINUX_) || defined(WIN32))
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+3] +
	    b*((unsigned char *) inr2->getData())[r+3]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+2] +
	    b*((unsigned char *) inr2->getData())[r+2]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+1] +
	    b*((unsigned char *) inr2->getData())[r+1]);
#else
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r] +
	    b*((unsigned char *) inr2->getData())[r]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+1] +
	    b*((unsigned char *) inr2->getData())[r+1]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+2] +
	    b*((unsigned char *) inr2->getData())[r+2]);
#endif
	}
      }

      delete[] offset;
      break;

    case Inrimage::WT_FLOAT_VECTOR:
      DRAWXYVECTOR(float,inr);
      /*  DRAWOVERLAYXYVECTOR(float,inr,inr2);*/
      DRAWXYVECTOR_VECTOR(float,inr2,inrv->getInterface()->under_sample_x,inrv->getInterface()->under_sample_y,inrv->getInterface()->under_sample_z,inrv->getInterface()->vector_width, inrv->getInterface()->vector_scale, inrv->getInterface()->normMode, inrv->getInterface()->alpha, inrv->getInterface()->beta, inrv->getInterface()->withArrow,inrv->getInterface()->norme3D,inrv->getInterface()->norme2D_Z,inrv->getInterface()->norme3D_max,inrv->getInterface()->norme3D_min);
      break;

    case Inrimage::WT_DOUBLE_VECTOR:
      DRAWOVERLAYXYVECTOR(double,inr,inr2);
      break;

    default:
      break;
    }

    break;
  }

  Vec2<double> tr(slicecam->getWidth() / 2.0 -
		  (slicecam->getCenter().getX() * slicecam->getScale()),
		  slicecam->getHeight() / 2.0 -
		  (slicecam->getCenter().getY() * slicecam->getScale()));

  TkWrapper_PhotoPutBlock(handle, block, 0, 0, block->width, block->height);
  slicecam->getInterface()->getModule()->tickle("%s coords inrimage %i %i",
						(char *) slicecam->getInterface()->getCanvas(),
						(int) (tr[0] + 0.5),
						(int) (tr[1] + 0.5));
  inrv->getSliceInterface()->updateSlice(this);
}


} // namespace yav
