#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#pragma implementation "inrimage/TclROIEdit.h"
#include <inrimage/TclInrimage.h>
#include <inrimage/TclROIEdit.h>
#include <modules/TclGenericHistogram.h>

namespace yav {

static float roiColor[10][3]={ {1.,0.,0.},
			       {0.,1.,0.},
			       {0.,0.,1.},
			       {1.,1.,0.},
			       {1.,0.,1.},
			       {0.,1.,1.},
			       {1.,0.5,0.5},
			       {0.5,1.0,0.5},
			       {0.5,0.5,1.0},
			       {0.5,0.,0.5}};

TclROIEdit::TclROIEdit(TclScene2D *scene, Inrimage *image, 
		       SurfaceSliceable *s,
		       InrimageROI::SurfaceRoiMethod m, 
		       double param) : TclInrimageROI(scene,typeid(TclROIEdit), new InrimageROI(scene->getClient(),image,s,m,param)) {
  init();
}
TclROIEdit::TclROIEdit(TclScene2D *scene,     
		       char *filename) :  TclInrimageROI(scene,typeid(TclROIEdit), new InrimageROI(scene->getClient(),filename)) {
  init();
}

TclROIEdit::TclROIEdit(TclScene2D *scene,     
		       Inrimage *image, 
		       const Inrimage *binary) throw(InrimageROI::ImageMissing,
						     InrimageROI::BinaryImageWrongSize) : TclInrimageROI(scene,typeid(TclROIEdit),new InrimageROI(scene->getClient(),image,binary)) {

  init();
}
TclROIEdit::TclROIEdit(TclScene2D *scene,     
		       Inrimage *image) throw(InrimageROI::ImageMissing) : TclInrimageROI(scene,typeid(TclROIEdit),new InrimageROI(scene->getClient(),image)) {
  init();

}
TclROIEdit::~TclROIEdit() {
  deleted(typeid(TclROIEdit));
}

void TclROIEdit::init() {


getClient()->setSelectionMode(InrimageROI::BRUSH_MODE);
 setInstallPopupMenuCommandName("inrimage::installROIEditPopupMenu");
  done(typeid(TclROIEdit),"inrimage::installROIEditMenu","inrimage::uninstallROIEditMenu");
 getClient()->focusScene(); 
 
 
}

int TclROIEdit::create(TclModule *m, int argc, char **argv) {
  TclInrimage *inr = 0;
  TclInrimage *bin = 0;
  TclScene2D *scn = 0;
  TclROIEdit *roi;
  InrimageROI::SurfaceRoiMethod method=  InrimageROI::SURFACE_BOUNDARY;
  SurfaceSliceable *s=0;
  int i;
  double p=1.0;

  if (argc==1)
    return COMMAND_SYNOPSIS;
  inr = TclInrimage::getObject(argc, argv, 1);  
  if (!inr)
    {
      if (argc>3){
	m->setResult("cannot read image %s",argv[1]);
	return TCL_ERROR;
      }
      if (argc==3)
	{
	  if(!scn) scn = (TclScene2D *) TclObjectKind::getObject(typeid(TclScene2D), argc, argv, 2);	
	  if (!scn) 
	    {
	      m->setResult("%s : invalid scene name",argv[2]);
	      return TCL_ERROR;
	    }
	}
      if(!scn) scn =  TclScene2D::getDefaultScene(m);
      roi= new TclROIEdit(scn,argv[1]);
      m->setResult("%s%u", TclObjectKind::getObjectName(typeid(TclROIEdit)), roi->getRef());
      return TCL_OK;
    }
  bin=TclInrimage::getObject(argc, argv, 2);  
  if (bin) {
    if ((bin->getX()!=inr->getX()) || (bin->getY()!=inr->getY()) || (bin->getZ()!=inr->getZ())) {
      m->setResult("the two images do not have the same size");
      return TCL_ERROR;
    }
    for(i = 3; i < argc; i++) {
      if(!scn) scn = (TclScene2D *) TclObjectKind::getObject(typeid(TclScene2D), argc, argv, i);	
      if (!scn) 
	{
	  m->setResult("%s : invalid scene name ",argv[i]);
	  return TCL_ERROR;
	}
    }
  }
  else {
    s=AutoLink<SurfaceSliceable>::getHead();
    for(;s && (s->getSurfaceSliceableName()!=(const char *)argv[2]);s=s->AutoLink<SurfaceSliceable>::following());
    if (!s)
      { 
	if(!scn) scn = (TclScene2D *) TclObjectKind::getObject(typeid(TclScene2D), argc, argv, 2);	
	if (!scn) 
	  {
	    m->setResult("%s : invalid scene name ",argv[2]);
	    return TCL_ERROR;
	  }
	roi= new TclROIEdit(scn,inr->getClient());
	m->setResult("%s%u", TclObjectKind::getObjectName(typeid(TclROIEdit)), roi->getRef());
	return TCL_OK;
      }
    
    for(i = 3; i < argc; i++) {
      if (!strcmp(argv[i],"boundary"))
	method=InrimageROI::SURFACE_BOUNDARY;
      else if (!strcmp(argv[i],"region"))
	method=InrimageROI::SURFACE_REGION;
      else if (!strcmp(argv[i],"band"))
	method=InrimageROI::SURFACE_NEAR_BOUNDARY;
      else if ((!strcmp(argv[i],"-param ")) && ((i+1)<argc))
	{
	  p=(double) atof(argv[++i]);
	}
      else
	{
	  if(!scn) scn = (TclScene2D *) TclObjectKind::getObject(typeid(TclScene2D), argc, argv, i);	
	  if (!scn) 
	    {
	      m->setResult("%s : invalid parameter",argv[i]);
	      return TCL_ERROR;
	    }
	}
    }
  }
  if(!scn) scn = new TclScene2D(m);
  if (!bin)
    {
      roi  = new TclROIEdit(scn,inr->getClient(),s,method,p);
    }
  else
    {
      roi  = new TclROIEdit(scn,inr->getClient(),bin->getClient());
    }
  m->setResult("%s%u", TclObjectKind::getObjectName(typeid(TclROIEdit)), 
	       roi->getRef());
  return TCL_OK;

}
void TclROIEdit::load(const TclString &filename,Scene2D *scn)
{
  new TclROIEdit((TclScene2D *)scn->getInterface(),(char *)filename);
}

int TclROIEdit::selectionModeCmd(int argc, char **argv) {
  if (argc>1) return COMMAND_SYNOPSIS;
  if (argc==1) {
    if (!strcmp(argv[0],"brush")) {
		getClient()->setSelectionMode(InrimageROI::BRUSH_MODE);
    }
    else if (!strcmp(argv[0],"eraser")) {
      getClient()->setSelectionMode(InrimageROI::ERASER_MODE);
    }
    else if (!strcmp(argv[0],"line")) {
      getClient()->setSelectionMode(InrimageROI::LINE_MODE);
    }
     else if (!strcmp(argv[0],"idle")) {
      getClient()->setSelectionMode(InrimageROI::IDLE_MODE);
    }
    else {
      setResult("%s must either 'brush' or 'eraser' or 'line' or 'idle'",argv[1]);
      return TCL_ERROR;
    }
  }
  if (getClient()->getSelectionMode()==InrimageROI::BRUSH_MODE) 
    setResult("brush");
  else if (getClient()->getSelectionMode()==InrimageROI::ERASER_MODE) 
    setResult("eraser");
  else if (getClient()->getSelectionMode()==InrimageROI::LINE_MODE) 
    setResult("line");
  else if (getClient()->getSelectionMode()==InrimageROI::IDLE_MODE) 
    setResult("idle");
  return TCL_OK;
}
int TclROIEdit::brushWidthCmd(int argc, char **argv) {
  if (argc==0) {
    setResult("%u",getClient()->getBrushWidth());
    return TCL_OK;
  }
  else if (argc==1) {
    unsigned int w=(unsigned int) atoi(argv[0]);
    getClient()->setBrushWidth(w);
    setResult("%u",getClient()->getBrushWidth());
    return TCL_OK;
  }
  else
    return COMMAND_SYNOPSIS;
}
int TclROIEdit::eraserWidthCmd(int argc, char **argv) {
  if (argc==0) {
    setResult("%u",getClient()->getEraserWidth());
    return TCL_OK;
  }
  else if (argc==1) {
    unsigned int w=(unsigned int) atoi(argv[0]);
    getClient()->setEraserWidth(w);
    setResult("%u",getClient()->getEraserWidth());
    return TCL_OK;
  }
  else
    return COMMAND_SYNOPSIS;
}
TclROIList::TclROIList(TclInrimage *inr, TclScene2D *_scene)  throw(ImageMissing): TclObject(typeid(TclROIList),_scene->getModule()),Owner<Inrimage>(inr->getClient()) {
  scene=_scene;
  filename="defaultName.rl";
  tickle("inrimage::createROIList %s %s %s",
	 (char *) name(),
	 (char *)inr->name(),
	 (char *)scene->name());
  token=Tcl_CreateTimerHandler(60000,TclROIList::timerHandler,(ClientData)this);

  done(typeid(TclROIList));
}
TclROIList::TclROIList(TclScene2D *_scene,TclString name) throw(ImageMissing) : TclObject(typeid(TclROIList),_scene->getModule()),Owner<Inrimage>(0) {
  scene=_scene;
  filename=name;
  IFSTREAM input((char *)name);
  TclString inrName;
  TclInrimage *inr;
  input >> inrName;
  try {
    inr=new TclInrimage(getModule(),inrName);
    if (!inr) throw(ImageMissing("TclROIList::TclROIList(TclScene2D *scene,TclString name)"));
  pointer=inr->getClient();
  unsigned int nb,i;
 // TclString roiName;
  char rname[300];
  TclROIEdit *r;
  input >> nb;
  
  for(i=0;i<nb;++i) {
    input >> rname;
    r=new TclROIEdit(scene,(char *)rname);
    r->getClient()->setImage(inr->getClient());
    if (roiList.getCount()<10)
      r->getClient()->setColor(roiColor[roiList.getCount()][0],
			       roiColor[roiList.getCount()][1],
			       roiColor[roiList.getCount()][2]);
    else
      r->getClient()->setColor((float)drand(),(float)drand(),(float)drand());

    roiList.append(r);
  }
  tickle("inrimage::createROIList %s %s %s",
	 (char *) this->name(),
	 (char *)inr->name(),
	 (char *)scene->name());
 
  token=Tcl_CreateTimerHandler(300000,TclROIList::timerHandler,(ClientData)this);
  }
  catch(...) {
    throw(ImageMissing("TclROIList::TclROIList(TclScene2D *scene,TclString name)"));
  }
  done(typeid(TclROIList));
}
int TclROIList::writeCmd(int argc, char **argv) {
  if (argc==0) {
    write(filename);
  }
  else if (argc==1) {
    if (!strcmp(argv[0],"name")) {
      setResult(filename);
      return TCL_OK;
    }
    else 
      write(argv[0]);
  }  else return COMMAND_SYNOPSIS;
  setResult("%s has been saved",(char *)filename);
  return TCL_OK;
}
int TclROIList::roisCmd(int argc, char **argv) {
  if (argc>0) 
    return COMMAND_SYNOPSIS;
  else {
    ObjectsList<TclROIEdit>::iterator itr;
    TclString output;
    for(itr=firstROI();itr!=lastROI();itr++) {
      output+=(*itr)->name();
      output+=" ";
    }
    setResult(output);
    return TCL_OK;
  }
}
int TclROIList::sceneCmd(int argc, char **argv) {
  if (argc>0) 
    return COMMAND_SYNOPSIS;
  else {
    setResult("%s",(char *)scene->name());
    return TCL_OK;
  }
}
int TclROIList::destroyCmd(int argc, char **argv) {
  TclModule *m;
  unsigned int ref;

  if(argc != 0) return COMMAND_SYNOPSIS;
  m = getModule();
  ref = getRef();
  TclString n=name();
  delete this;
  m->setResult(n);
  return TCL_OK;
}

int TclROIList::newROICmd(int argc, char **argv) {
  if (argc>0) 
    return COMMAND_SYNOPSIS;
  else {
    TclROIEdit *roi=new TclROIEdit(scene,getPointer());
    if (roiList.getCount()<10)
      roi->getClient()->setColor(roiColor[roiList.getCount()][0],
				 roiColor[roiList.getCount()][1],
				 roiColor[roiList.getCount()][2]);
    else
      roi->getClient()->setColor((float)drand(),(float)drand(),(float)drand());
    roiList.append(roi);
    setResult("%s",(char *)roi->name());
    return TCL_OK;
  }
}
void TclROIList::write(TclString fname) {
  if (fname!=EMPTY_STRING)
    filename=fname;
  
  TclString base =filename;
  
  if((base.getLength() > 3) &&
     (TclString(base, base.getLength()-3, base.getLength()-1) == ".rl"))
    base = TclString(base, 0, base.getLength()-4);  
  

  OFSTREAM output(filename);
  output << getPointer()->getName() << ENDL;
  ObjectsList<TclROIEdit>::iterator itr;
  output << roiList.getCount() <<ENDL;

  TclString roiName;
  unsigned int i;
  for(i=1,itr=firstROI();itr!=lastROI();itr++,i++) {
    roiName.sprintf("%s.%d.roi",(char *)base,i);
    (*itr)->getClient()->write(roiName);
    output <<(*itr)->getClient()->getFilename()<<ENDL;
  }
  
}
int TclROIList::create(TclModule *m, int argc, char **argv) {
  TclInrimage *inr = 0;
  TclScene2D *scn = 0;
  TclROIList *roi;
  if ((argc==1) || (argc>3))
    return COMMAND_SYNOPSIS;
  inr = TclInrimage::getObject(argc, argv, 1);  
  if (!inr) {
    if (argc==3)
      {
	if(!scn) scn = (TclScene2D *) TclObjectKind::getObject(typeid(TclScene2D), argc, argv, 2);	
	if (!scn) 
	  {
	    m->setResult("%s : invalid scene name",argv[2]);
	    return TCL_ERROR;
	  }
      }
    if(!scn) scn =  TclScene2D::getDefaultScene(m);
    roi= new TclROIList(scn,argv[1]);
    return TCL_OK;
  }
  else {
    if (argc==3)
      {
	if(!scn) scn = (TclScene2D *) TclObjectKind::getObject(typeid(TclScene2D), argc, argv, 2);	
	if (!scn) 
	  {
	    m->setResult("%s : invalid scene name",argv[2]);
	    return TCL_ERROR;
	  }
      }
    if(!scn) scn =  TclScene2D::getDefaultScene(m);
    roi= new TclROIList(inr,scn);
    return TCL_OK;
  }
}
void TclROIList::dying(const ObjectPointer<Inrimage>& obj){
  pointer=0;
  delete this;
}
TclROIList::~TclROIList() {
  Tcl_DeleteTimerHandler(token);
}
void TclROIList::timerHandler(ClientData client) {
  TclROIList *rl=(TclROIList *)client;
  TclString fname=rl->getFilename();
  TclString saveName=fname;
  fname+=".bak";
  rl->write(fname);
  rl->filename=saveName;

  rl->token=Tcl_CreateTimerHandler(300000,TclROIList::timerHandler,(ClientData)client);
}

void TclROIEdit::createCommands(TclModule *m)
{
	TclObjectKind *kind;
	kind = new Scene2DItem(m, typeid(TclROIEdit), "roie","Editable Region of Interest",
		&TclROIEdit::create, &TclROIEdit::list,
		TclScene2D::FILE | TclScene2D::OTHER,"ROI Edition", ".roi",&TclROIEdit::load,0,"%s.roi [scn2D%u] | inr%u [ [sm%u | tr3D%u | tet3D%u] [boundary | region | band | scn2D%u | -param %f] | inr%u [scn2D%u] ]");
	m->newCommand("roie%u", kind->getOptions(), "inrimage region of interest management command");
	kind->addOptions(typeid(TclInrimageROI));
	kind->newOption("-selectionMode", (TclObject::OPTION) &TclROIEdit::selectionModeCmd, "[brush | eraser | line ]", "get/set the current selection mode : either brush, eraser or line in order to select a block of pixels or to erase a block of pixels or to select a line of pixels");
	kind->newOption("-brushWidth", (TclObject::OPTION) &TclROIEdit::brushWidthCmd, "[%u ]", "get/set the current with of the brush");
	kind->newOption("-eraserWidth", (TclObject::OPTION) &TclROIEdit::eraserWidthCmd, "[%u ]", "get/set the current with of the eraser");
}

void TclROIList::createCommands(TclModule *m)
{
	TclObjectKind *kind;
	kind = new TclObjectKind(m, typeid(TclROIList), "rlist",
		"List of ROI", &TclROIList::create,
		&TclROIList::list, "<%s>");
	m->newCommand("rlist%u", kind->getOptions(), "list of  region of interest management command");
	kind->newOption("-scene", (TclObject::OPTION) &TclROIList::sceneCmd, 0,
		"returns the 2D scene where the ROI are located");
	kind->newOption("-rois", (TclObject::OPTION) &TclROIList::roisCmd, 0,
		"returns the list of ROIs");
	kind->newOption("-write", (TclObject::OPTION) &TclROIList::writeCmd, "[%s | name]",
		"save the list in a file. Allows to get the current filename");
	kind->newOption("-newROI", (TclObject::OPTION) &TclROIList::newROICmd, 0,
		"create a new ROI");
	kind->newOption("-destroy", (TclObject::OPTION) &TclROIList::destroyCmd, 0,
		"deletes  ROI list");
}
} // namespace yav
