#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/TclTransformedInrimageBackground.h"
#pragma implementation "inrimage/backgroundTransformedInrimageRedraw.h"

#include <inrimage/TclTransformedInrimageBackground.h>
#include <graphics2D/TclSliceCamera.h>
#include <inrimage/TclInrimage.h>
#include <inrimage/TclColorMap.h>
#include <inrimage/resample.h>
#include "backgroundRedraw.h"
#include "backgroundTransformedInrimageRedraw.h"
#include "tkwrapper.h"

namespace yav {


TclTransformedInrimageBackground::TclTransformedInrimageBackground(TclModule *m,
					   TransformedInrimageView *inrv, SliceCamera *cam,
					   bool useMinGLScale ,
					   MinGLScale *minScale,
					   bool useMaxGLScale,
					   MaxGLScale *maxScale, 
					   const char *color) : 
  TclInrimageBackground(m,typeid(TclTransformedInrimageBackground),inrv,cam,minScale,maxScale) {

  done(typeid(TclTransformedInrimageBackground));

  slicecam->setBackground(this);
  init(useMinGLScale,minScale,useMaxGLScale,maxScale,color);

  if (slicecam->getScaleCamera())
    {
      ((SliceScale *)slicecam->getScaleCamera())->reset(slicecam,slicecam->sliceName(slicecam->getGeometry()));
    }

  if (getMinScale())
    getMinScale()->reset((Scalable *)this);
  if (getMaxScale())
    getMaxScale()->reset((Scalable *)this);

  resampledSlice=0;
  resampleImage=true;
  getTransformedInrimageView()->getSliceInterface()->addSlice(this);
}

TclTransformedInrimageBackground::~TclTransformedInrimageBackground() {
}


void TclTransformedInrimageBackground::imageGLChanged(void) {
  glBounds();
  if(getMinScale())
    getMinScale()->reset(this);
  if(getMaxScale())
    getMaxScale()->reset(this);
  redraw();
}


void TclTransformedInrimageBackground::inrimageChanged(void) {
  TclInrimageBackground::inrimageChanged();
  
}


TclString TclTransformedInrimageBackground::getLabel(int &x,int &y){
  TclString s;
  VoxelCoordinate *vc = coordinates(x, y);
  Voxel *v = getInrimage()->getVoxel(vc->x, vc->y, vc->z);
  TclString sval;
  v->print(sval);
  s.sprintf("%3u %3u %3u : (%s)", vc->x, vc->y, vc->z,(char *) sval);
  delete vc;
  delete v;
  return(s);
}



void TclTransformedInrimageBackground::createResampledSlice() {
  if (getTransformedInrimageView()) {
    if (resampledSlice==0) {
      Inrimage *inr=getInrimage();
     switch (slicecam->getGeometry()) {
     case SliceCamera::SLICE_X:
       resampledSlice= new Inrimage(1,visible[4] - visible[1] + 1,
				    visible[5] - visible[2] + 1, inr->getType(),1,VM_SCALAR,inr->getVX(),
				    inr->getVY(),inr->getVZ());
       break;
      case SliceCamera::SLICE_Y:
	resampledSlice= new Inrimage(visible[3] - visible[0] + 1,1,
				     visible[5] - visible[2] + 1, inr->getType(),1,VM_SCALAR,inr->getVX(),
				     inr->getVY(),inr->getVZ());
	break;
      case SliceCamera::SLICE_Z:
      case SliceCamera::SLICE_ALPHA:
	resampledSlice= new Inrimage(visible[3] - visible[0] + 1,
				     visible[4] - visible[1] + 1, 1,inr->getType(),1,VM_SCALAR,inr->getVX(),
				     inr->getVY(),inr->getVZ());
	break;
     }
    }
    Matrix4x4<double> t;
    t.resetToIdentity();
    switch (slicecam->getGeometry()) {
    case SliceCamera::SLICE_X:
      t(0,3)=(double) slicecam->getSlice();
      t(1,3)=visible[2];
      t(2,3)=visible[1];
      break;
    case SliceCamera::SLICE_Y:
      t(0,3)=visible[0];
      t(1,3)=(double) slicecam->getSlice();
      t(2,3)=visible[2];
      break;
    case SliceCamera::SLICE_Z:
    case SliceCamera::SLICE_ALPHA:
      t(0,3)=visible[0];
      t(1,3)=visible[1];
      t(2,3)=(double) slicecam->getSlice();
      break;
    }
    t = getTransformedInrimageView()->getTransformation() *t;
    resample4x4TriLinear(getInrimage(),resampledSlice,t,false,1.0,0.0);
    if (slicecam->getGeometry()==SliceCamera::SLICE_X) 
      resampledSlice->write("toto.inr");
  }
}
int TclTransformedInrimageBackground::create(TclModule *m, int argc, char **argv) {
  TclScene2D *scene = TclScene2D::getDefaultScene(m);
  int i;
  TclInrimageBackground *b;
  TclInrimage *inr;
  MinGLScale *minScale=0;
  MaxGLScale *maxScale=0;
  bool useMinScale=false;
  bool useMaxScale=false;
  if (argc<3)
    return COMMAND_SYNOPSIS;
  inr = TclInrimage::getObject(argc, argv, 1);
  if (!inr)
    {
      m->setResult("%s is not an image",argv[1]);
      return TCL_ERROR;
    }
  TclSliceCamera *cam=TclSliceCamera::getObject(argc, argv, 3);
  if (!cam)
    {
      m->setResult("%s is not an slice camera",argv[3]);
      return TCL_ERROR;
    }

  for(i=4; i < argc; i++) {
    if(!strcmp(argv[i], "-scene")) {
      if( ++i >= argc) return COMMAND_SYNOPSIS;
      scene = (TclScene2D *)
	TclObjectKind::getObject(typeid(TclScene2D), argc, argv, i);
      if(!scene) {
	m->setResult("Illegal scene name : %s",argv[i]);
	return TCL_ERROR;
      }
    }
    if(!strcmp(argv[i], "-min")) {
      if( ++i >= argc) return COMMAND_SYNOPSIS;
      minScale= (MinGLScale *) 	TclObjectKind::getObject(typeid(MinGLScale), argc, argv, i);
      if (!minScale) {
	m->setResult("Illegal MinGLScale name : %s",argv[i]);
	return TCL_ERROR;
      }
      else
	useMinScale=true;
    }
    if(!strcmp(argv[i], "-max")) {
      if( ++i >= argc) return COMMAND_SYNOPSIS;
      maxScale= (MaxGLScale *) 	TclObjectKind::getObject(typeid(MaxGLScale), argc, argv, i);
      if (!maxScale) {
	m->setResult("Illegal MaxGLScale name : %s",argv[i]);
	return TCL_ERROR;
      }
      else
	useMaxScale=true;
    }
    else return COMMAND_SYNOPSIS;
  }

  b = new TclTransformedInrimageBackground(m,new TransformedInrimageView(inr->getClient(),scene->getClient()),cam->getClient(),
					   useMinScale,minScale,
					   useMaxScale,maxScale);
  m->setResult("%s%u", b->getObjectName(), b->TclObject::getRef());
  return TCL_OK;
}


#define DRAWXY(TYPE,INR) drawTransformedInrimageXY((TYPE) 0,INR,slicecam,  \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                block, \
				redColorMap,greenColorMap, \
				blueColorMap, \
				sx,sy,rx,ry,interpolation)

#define DRAWXYVECTOR(TYPE,INR) drawTransformedInrimageXYVector((TYPE) 0,INR,slicecam,  \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                block, \
				redColorMap,greenColorMap, \
				blueColorMap, \
				sx,sy,rx,ry)

#define DRAWXZ(TYPE,INR) drawTransformedInrimageXZ((TYPE) 0,INR,slicecam, \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                block, \
				redColorMap,greenColorMap, \
				blueColorMap,sx,sz,rx,rz,interpolation)


#define DRAWXZVECTOR(TYPE,INR) drawTransformedInrimageXZVector((TYPE) 0,INR,slicecam, \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                block, \
				redColorMap,greenColorMap, \
				blueColorMap,sx,sz,rx,rz)


#define DRAWYZ(TYPE,INR) drawTransformedInrimageYZ((TYPE) 0,INR,slicecam, \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                block, \
				redColorMap,greenColorMap, \
				blueColorMap,sy,sz,ry,rz,interpolation)


#define DRAWYZVECTOR(TYPE,INR) drawTransformedInrimageYZVector((TYPE) 0,INR,slicecam, \
				glmin->getValueAsDouble(), \
				glmax->getValueAsDouble(),\
                                block, \
				redColorMap,greenColorMap, \
				blueColorMap,sy,sz,ry,rz)




void TclTransformedInrimageBackground::redraw(void) {
  unsigned int plane, /*i, j, x, y, z, *offset,*/ o;
  //register int r;
  
  voxelRatio();
 
  Inrimage *inr=getInrimage();
  plane = inr->getX() * inr->getY();
  o = (unsigned int) slicecam->getSlice();
  // do the image sampling if necessary
  if (resampleImage) 
    createResampledSlice();
  resampleImage=true;

  switch(slicecam->getGeometry()) {

  /* slice perpendicular to X */
  case SliceCamera::SLICE_X:
    if(o < visible[0] || o > visible[3]) {
      memset((void *) block->pixelPtr, 0, 3 * sy * sz);
      return;
    }

    switch(inr->getType()) {

    case Inrimage::WT_UNSIGNED_CHAR:
      DRAWYZ(unsigned char,resampledSlice);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      DRAWYZ(signed char,resampledSlice);
      break;
 
    case Inrimage::WT_UNSIGNED_SHORT:
      DRAWYZ(unsigned short int,resampledSlice);
      break;

    case Inrimage::WT_SIGNED_SHORT:
      DRAWYZ(short int,resampledSlice);
      break;

    case Inrimage::WT_UNSIGNED_INT:
      DRAWYZ(unsigned int,resampledSlice);
      break;

    case Inrimage::WT_SIGNED_INT:
      DRAWYZ(int,resampledSlice);
      break;

    case Inrimage::WT_UNSIGNED_LONG:
      DRAWYZ(unsigned LONGINT,resampledSlice);
      break;

    case Inrimage::WT_SIGNED_LONG:
      DRAWYZ(LONGINT,resampledSlice);
      break;

    case Inrimage::WT_FLOAT:
      DRAWYZ(float,resampledSlice);
      break;

    case Inrimage::WT_DOUBLE:
      DRAWYZ(double,resampledSlice);
      break;

    case Inrimage::WT_RGB:
      /*
      offset = new unsigned int[sz];
      for(i = 0; i < sz; i++) offset[i] =
				((int) ((double) i / rz)) * plane;
      
      for(y = 0, j = 0; y < sy; y++) {
	i = ((int)((double) y / ry) + visible[1]) * inr->getX() + o;
	for(z = 0; z < sz; z++) {
	  r = 3 * (offset[z] + i);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r] +
	    b*((unsigned char *) inr2->getData())[r]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+1] +
	    b*((unsigned char *) inr2->getData())[r+1]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+2] +
	    b*((unsigned char *) inr2->getData())[r+2]);
	}
      }
      
      delete[] offset;
      */
      break;

    case Inrimage::WT_RGBA:

      /*      offset = new unsigned int[sz];
      for(i = 0; i < sz; i++) offset[i] =
				((int) ((double) i / rz) + visible[2]) * plane;
      
      for(y = 0, j = 0; y < sy; y++) {
	i = ((int)((double) y / ry) + visible[1]) * inr->getX() + o;
	for(z = 0; z < sz; z++) {
	  r = 4 * (offset[z] + i);
#if (defined (_ALPHA_) || defined (_LINUX_) || defined(WIN32))
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+3] +
	    b*((unsigned char *) inr2->getData())[r+3]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+2] +
	    b*((unsigned char *) inr2->getData())[r+2]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+1] +
	    b*((unsigned char *) inr2->getData())[r+1]);
#else
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r] +
	    b*((unsigned char *) inr2->getData())[r]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+1] +
	    b*((unsigned char *) inr2->getData())[r+1]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+2] +
	    b*((unsigned char *) inr2->getData())[r+2]);
#endif
	}
      }
     
      delete[] offset;
      */
      break;

    case Inrimage::WT_FLOAT_VECTOR:
      DRAWYZVECTOR(float,resampledSlice);
      break;

    case Inrimage::WT_DOUBLE_VECTOR:
      DRAWYZVECTOR(double,resampledSlice);
      break;

    default:
      break;
    }

    break;

    /* slice perpendicular to Y */
  case SliceCamera::SLICE_Y:
    if(o < visible[1] || o > visible[4]) {
      memset((void *) block->pixelPtr, 0, 3 * sx * sz);
      return;
    }

    switch(inr->getType()) {

    case Inrimage::WT_UNSIGNED_CHAR:
      DRAWXZ(unsigned char,resampledSlice);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      DRAWXZ(signed char,resampledSlice);
      break;

    case Inrimage::WT_UNSIGNED_SHORT:
      DRAWXZ(unsigned short int,resampledSlice);
      break;

    case Inrimage::WT_SIGNED_SHORT:
      DRAWXZ(short int,resampledSlice);
      break;

    case Inrimage::WT_UNSIGNED_INT:
      DRAWXZ(unsigned int,resampledSlice);
      break;

    case Inrimage::WT_SIGNED_INT:
      DRAWXZ(int,resampledSlice);
      break;

    case Inrimage::WT_UNSIGNED_LONG:
      DRAWXZ(unsigned LONGINT,resampledSlice);
      break;

    case Inrimage::WT_SIGNED_LONG:
      DRAWXZ(LONGINT,resampledSlice);
      break;

    case Inrimage::WT_FLOAT:
      DRAWXZ(float,resampledSlice);
      break;

    case Inrimage::WT_DOUBLE:
      DRAWXZ(double,resampledSlice);
      break;

    case Inrimage::WT_RGB:
      /*      offset = new unsigned int[sx];
      for(i = 0; i < sx; i++) offset[i] = (int) ((double) i / rx) + visible[0];
      o = (unsigned int) slicecam->getSlice() * inr->getX();

      for(z = 0, j = 0; z < sz; z++) {
	i = ((int)((double) z / rz) + visible[2]) * plane + o;
	for(x = 0; x < sx; x++) {
	  r = 3 * (offset[x] + i);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r] +
	    b*((unsigned char *) inr2->getData())[r]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+1] +
	    b*((unsigned char *) inr2->getData())[r+1]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+2] +
	    b*((unsigned char *) inr2->getData())[r+2]);
	}
      }
      
      delete[] offset;
      */
      break;

    case Inrimage::WT_RGBA:
      /*
      offset = new unsigned int[sx];
      for(i = 0; i < sx; i++) offset[i] = (int) ((double) i / rx) + visible[0];
      o = (unsigned int) slicecam->getSlice() * inr->getX();

      for(z = 0, j = 0; z < sz; z++) {
	i = ((int)((double) z / rz) + visible[2]) * plane + o;
	for(x = 0; x < sx; x++) {
	  r = 4 * (offset[x] + i);
#if (defined (_ALPHA_) || defined (_LINUX_) || defined(WIN32))
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+3] +
	    b*((unsigned char *) inr2->getData())[r+3]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+2] +
	    b*((unsigned char *) inr2->getData())[r+2]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+1] +
	    b*((unsigned char *) inr2->getData())[r+1]);
#else
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r] +
	    b*((unsigned char *) inr2->getData())[r]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+1] +
	    b*((unsigned char *) inr2->getData())[r+1]);
	  block->pixelPtr[j++] =(unsigned char)(
	    a*((unsigned char *) inr->getData())[r+2] +
	    b*((unsigned char *) inr2->getData())[r+2]);
#endif
	}
      }
       
      delete[] offset;
      */
      break;

    case Inrimage::WT_FLOAT_VECTOR:
      DRAWXZVECTOR(float,resampledSlice);
      break;

    case Inrimage::WT_DOUBLE_VECTOR:
      DRAWXZVECTOR(double,resampledSlice);
      break;

    default:
      break;
    }

    break;

    /* slice perpendicular to Z */
  case SliceCamera::SLICE_ALPHA:
  case SliceCamera::SLICE_Z:
    if(slicecam->getGeometry() == SliceCamera::SLICE_Z &&
       (o < visible[2] || o > visible[5])) {
      memset((void *) block->pixelPtr, 0, 3 * sx * sy);
      return;
    }

    switch(inr->getType()) {

    case Inrimage::WT_UNSIGNED_CHAR:
      DRAWXY(unsigned char,resampledSlice);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      DRAWXY(signed char,resampledSlice);
      break;

    case Inrimage::WT_UNSIGNED_SHORT:
      DRAWXY(unsigned short int,resampledSlice);
      break;

    case Inrimage::WT_SIGNED_SHORT:
      DRAWXY(short int,resampledSlice);
      break;

    case Inrimage::WT_UNSIGNED_INT:
      DRAWXY(unsigned int,resampledSlice);
      break;

    case Inrimage::WT_SIGNED_INT:
      DRAWXY(int,resampledSlice);
      break;

    case Inrimage::WT_UNSIGNED_LONG:
      DRAWXY(unsigned LONGINT,resampledSlice);
      break;

    case Inrimage::WT_SIGNED_LONG:
      DRAWXY(LONGINT,resampledSlice);
      break;

    case Inrimage::WT_FLOAT:
      DRAWXY(float,resampledSlice);
      break;

    case Inrimage::WT_DOUBLE:
      DRAWXY(double,resampledSlice);
      break;

  
      break;

    case Inrimage::WT_RGBA:
     
      break;

    case Inrimage::WT_FLOAT_VECTOR:
      DRAWXYVECTOR(float,resampledSlice);
      break;

    case Inrimage::WT_DOUBLE_VECTOR:
      DRAWXYVECTOR(double,resampledSlice);
      break;

    default:
      break;
    }

    break;
  }

  Vec2<double> tr(slicecam->getWidth() / 2.0 -
		  (slicecam->getCenter().getX() * slicecam->getScale()),
		  slicecam->getHeight() / 2.0 -
		  (slicecam->getCenter().getY() * slicecam->getScale()));
#ifdef OLD_CODE
  TkWrapper_PhotoPutBlock(handle, block, 0, 0, block->width, block->height);
  ((TclSliceCamera *)slicecam->getInterface())->getModule()->tickle("%s coords inrimage %i %i",
						(char *) ((TclSliceCamera *)slicecam->getInterface())->getCanvas(),
						(int) (tr[0] + 0.5),
						(int) (tr[1] + 0.5));
#endif
  inrv->getSliceInterface()->updateSlice(this);
}


} // namespace yav

