#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/TransformedInrimageView.h"

#ifdef GRAPHICS
#include <inrimage/TransformedInrimageView.h>
#ifdef TCL_INTERFACE
#include <graphics2D/TclSliceCamera.h>
#include <inrimage/TclInrimage.h>
#include <inrimage/TclTransformedInrimageView.h>
#include <inrimage/TclTransformedInrimageBackground.h>
#endif

namespace yav {


TransformedInrimageView::NoImage::NoImage(const std::string& from)
  : Exception(from,"provided an empty image while attempting to fuse two images") {
}
TransformedInrimageView::TransformedInrimageView(Inrimage *inr1, 
						 Scene2D *scene,bool _createGLScale) 
  throw(TransformedInrimageView::NoImage)
  : InrimageView(inr1,(Inrimage *)0,scene,_createGLScale) {
  if (!inr1)
    throw(TransformedInrimageView::NoImage("TransformedInrimageView::TransformedInrimageView(Inrimage *,Inrimage *,Scene2D *,bool)"));
  transformation.resetToIdentity();
  init();
  currentSlicecam=0;
  rotationCurrentSlicecam=0;
  scaleCurrentSlicecam=0;

}
  

  
void TransformedInrimageView::changeInrimage(Inrimage *inr) {
  if (inr!=pointer()) {
    InrimageView::changeInrimage(inr);
      
  }
}
void  TransformedInrimageView::shiftButtonPressedCallback(SliceCamera *slicecam, 
							  int  mx, int my){
  previousPixelPosition=slicecam->getTrackballVector(mx,my);
  rotationCurrentSlicecam=slicecam;
}
void TransformedInrimageView::shiftButtonReleasedCallback(SliceCamera *slicecam, 
							  int  mx, int my){
  if (rotationCurrentSlicecam!=0) {
    redrawBackgrounds();
    rotationCurrentSlicecam=0;
  }
  
}
void TransformedInrimageView::shiftButtonMotionCallback(SliceCamera *slicecam, 
							int  mx, int my){
  if (slicecam==rotationCurrentSlicecam) {
    Vec2<double> p=slicecam->getTrackballVector(mx,my);
    Vec2<double> dp1, dp2;
  
    dp1=previousPixelPosition-slicecam->getCameraCenter();
    dp2=p-slicecam->getCameraCenter();
    double norm1=dp1.norm();
    double norm2=dp1.norm();
    const unsigned int *visible=slicecam->getBackground()->getVisibleArea();
    if ((norm1!=0) || (norm2!=0))
      {
	dp1/=norm1;
	dp2/=norm2;
	double angle=atan2(det(dp1,dp2),dot(dp1,dp2));

	Matrix4x4<double> m;
	double ca=cos(angle);
	double sa=sin(angle);
	
	switch(slicecam->getGeometry()) {
	case SliceCamera::SLICE_X:
	  m(1,1)=ca;
	  m(1,2)=-sa;
	  m(2,1)= sa;
	  m(2,2)= ca;
	  m(0,0)=1.0;
	  m(3,3)=1.0;
	  {
	  double cx=0.5*(visible[1]+visible[4]+1);
	  double cy=0.5*(visible[2]+visible[5]+1);
	  m(1,3)=cx*(1-ca)+cy*sa;
	  m(2,3)=-cx*sa+(1-ca)*cy;
	  }
	  break;
	case SliceCamera::SLICE_Y:
	  m(0,0)=ca;
	  m(0,2)=sa;
	  m(2,0)= -sa;
	  m(2,2)= ca;
	  m(1,1)=1.0;
	  m(3,3)=1.0;
	  {
	  double cx=0.5*(visible[0]+visible[3]+1);
	  double cy=0.5*(visible[2]+visible[5]+1);
	  m(0,3)=cx*(1-ca)-cy*sa;
	  m(2,3)=cx*sa+(1-ca)*cy;
	  }

	  break;
	case SliceCamera::SLICE_Z:
	case SliceCamera::SLICE_ALPHA:
	  m(0,0)=ca;
	  m(0,1)=sa;
	  m(1,0)= -sa;
	  m(1,1)= ca;
	  m(2,2)=1.0;
	  m(3,3)=1.0;
	  {
	  double cx=0.5*(visible[0]+visible[3]+1);
	  double cy=0.5*(visible[1]+visible[4]+1);
	  m(0,3)=cx*(1-ca)-cy*sa;
	  m(1,3)=cx*sa+(1-ca)*cy;
	  }
	  break;
	}
	transformation*= m; 
	SliceCameraBackground *b= slicecam->getBackground();
	std::list<InrimageBackground *>::iterator it = backgroundList.begin();
	for (; it!=backgroundList.end(); ++it){
	  if ((SliceCameraBackground *)(*it)==b)
	    (*it)->redraw();
	}
	
	
      }
    
      
    previousPixelPosition=p;
  }
  
  //  redrawScene();
}


void TransformedInrimageView::mouseButtonPressed(CameraTypes::WHICH_BUTTON button,
						 CameraTypes::KEY_MODIFIER modifier,
						 Camera2D *cam,int x,int y) {

  if ((button==CameraTypes::LEFT_BUTTON) && (modifier==CameraTypes::SHIFT_KEY)){ 
    shiftButtonPressedCallback((SliceCamera *)cam,x,y);
  }
  else if ((button==CameraTypes::LEFT_BUTTON) && (modifier==CameraTypes::NO_MODIFIER)){ 
    buttonPressedCallback((SliceCamera *)cam,x,y);
  }
  else if ((button==CameraTypes::LEFT_BUTTON) && (modifier==CameraTypes::CTRL_KEY)){ 
    SliceCamera *slicecam= (SliceCamera *)cam;
    previousPixelPosition=slicecam->getTrackballVector(x,y);
    scaleCurrentSlicecam=slicecam;
  }

}

void TransformedInrimageView::mouseButtonReleased(CameraTypes::WHICH_BUTTON button,
						  CameraTypes::KEY_MODIFIER modifier,
						  Camera2D *cam,int x,int y) {
  if ((button==CameraTypes::LEFT_BUTTON) && (modifier==CameraTypes::SHIFT_KEY)){ 
    shiftButtonReleasedCallback((SliceCamera *)cam,x,y);
  }
  else if ((button==CameraTypes::LEFT_BUTTON) && (modifier==CameraTypes::NO_MODIFIER)){ 
    buttonReleasedCallback((SliceCamera *)cam,x,y);
  }
  else if ((button==CameraTypes::LEFT_BUTTON) && (modifier==CameraTypes::CTRL_KEY)){ 
    if (scaleCurrentSlicecam!=0) {
      redrawBackgrounds();
      scaleCurrentSlicecam=0;
    }
  }
}

void TransformedInrimageView::mouseButtonMoved(Camera2D *cam,int x,int y) {
  if (currentSlicecam)
    buttonMotionCallback((SliceCamera *)cam,x,y);
  else if (rotationCurrentSlicecam)
    shiftButtonMotionCallback((SliceCamera *)cam,x,y);
  else if ((scaleCurrentSlicecam) && (scaleCurrentSlicecam==(SliceCamera *)cam)) {
     SliceCamera *slicecam= (SliceCamera *)cam;
    Vec2<double> p=slicecam->getTrackballVector(x,y);
    double sc=slicecam->getTrackballScale(p,previousPixelPosition);
    previousPixelPosition=p;
    Matrix4x4<double> m;
    m.resetToIdentity();
    m(0,0)=sc;
    m(1,1)=sc;
    m(2,2)=sc;
    const unsigned int *visible=slicecam->getBackground()->getVisibleArea();
    switch(slicecam->getGeometry()) {
    case SliceCamera::SLICE_X:
      {
      double cx=0.5*(visible[1]+visible[4]+1);
      double cy=0.5*(visible[2]+visible[5]+1);
      m(0,3)= -(sc-1.0)*slicecam->getSlice();
      m(1,3)= -(sc-1.0)*cx;
      m(2,3)= -(sc-1.0)*cy;
      }
      break;
    case SliceCamera::SLICE_Y:
      {
      double cx=0.5*(visible[0]+visible[3]+1);
      double cy=0.5*(visible[2]+visible[5]+1);
      m(0,3)= -(sc-1.0)*cx;
      m(1,3)= -(sc-1.0)*slicecam->getSlice();;
      m(2,3)= -(sc-1.0)*cy;
      }
      break;
    case SliceCamera::SLICE_Z:
    case SliceCamera::SLICE_ALPHA:
      {
      double cx=0.5*(visible[0]+visible[3]+1);
      double cy=0.5*(visible[1]+visible[4]+1);
      m(0,3)= -(sc-1.0)*cx;
      m(1,3)= -(sc-1.0)*cy;
      m(2,3)= -(sc-1.0)*slicecam->getSlice();
      }
      break;
    }
    transformation*= m; 
    SliceCameraBackground *b= slicecam->getBackground();
    std::list<InrimageBackground *>::iterator it = backgroundList.begin();
    for (; it!=backgroundList.end(); ++it){
      if ((SliceCameraBackground *)(*it)==b)
	(*it)->redraw();
    }
  }
    

}


void  TransformedInrimageView::buttonPressedCallback(SliceCamera *slicecam, 
			   int  mx, int my){
  unsigned int vx,vy,vz;
  slicecam->getBackground()->getImageCoordinates(mx,my,vx,vy,vz);
  previousVoxelPosition=Vec3<unsigned int>(vx,vy,vz);
  currentSlicecam=slicecam;

}
void TransformedInrimageView::buttonReleasedCallback(SliceCamera *slicecam, 
						     int  mx, int my){
  if (currentSlicecam!=0) {
    redrawBackgrounds();
    currentSlicecam=0;
  }
  
}
void TransformedInrimageView::buttonMotionCallback(SliceCamera *slicecam, 
						   int  mx, int my){
  if (slicecam==currentSlicecam) {
    unsigned int vx,vy,vz;
    slicecam->getBackground()->getImageCoordinates(mx,my,vx,vy,vz);
    Vec3<double> dp=Vec3<double>((double)vx-previousVoxelPosition[0],
				 (double)vy-previousVoxelPosition[1],
				 (double)vz-previousVoxelPosition[2]);
    
    if ((dp[0]!=0) || (dp[1]!=0) || (dp[2]!=0)) {
      Matrix4x4<double> m;
      m.resetToIdentity();
      m(0,3)= -dp[0];
      m(1,3)= -dp[1];
      m(2,3)= -dp[2];
      
      previousVoxelPosition=Vec3<unsigned int>(vx,vy,vz);
      transformation*=m; 
      SliceCameraBackground *b= slicecam->getBackground();
      std::list<InrimageBackground *>::iterator it = backgroundList.begin();
      for (; it!=backgroundList.end(); ++it){
	if ((SliceCameraBackground *)(*it)==b)
	  (*it)->redraw();
      }
    }
  
  //  redrawScene();
  }
}
void TransformedInrimageView::doCreateCamera(SliceCamera& cam) {
  cameras[&cam] = new Sliceable::CamInfo();
  double m;

  if(pointer->getVX() < pointer->getVY()) m = pointer->getVX();
  else m = pointer->getVY();
  if(pointer->getVZ() < m) m = pointer->getVZ();
  //  cam.scaleCamera(1.0 / (m * cam.getScale()));
  //  cam.scaleCamera(1.0 / (m));
  switch(cam.getGeometry()) {
  case SliceCamera::SLICE_X:
    cam.setThickness(pointer->getVX());
    break;
  case SliceCamera::SLICE_Y:
    cam.setThickness(pointer->getVY());
    break;
  case SliceCamera::SLICE_Z:
  case SliceCamera::SLICE_ALPHA:
    cam.setThickness(pointer->getVZ());
    break;
  }
#ifdef TCL_INTERFACE
  TclTransformedInrimageBackground *b;
  if (createGLScale)
    b = new TclTransformedInrimageBackground(cam.getInterface()->getModule(),this,&cam);
  else
    b = new TclTransformedInrimageBackground(getInterface()->getModule(),this,&cam,false,0,false,0);
  backgroundList.push_back(b);  
#endif
  cam.focusOnBackground();
}



} // namespace yav

#endif
