#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/Voxel.h"
#include <inrimage/Voxel.h>
#include <basics/Constants.h>


namespace yav {


Voxel::InvalidValue::InvalidValue(const std::string& from)
  : Exception(from,
	      "invalid voxel value transmitted to Voxel::scan") {
}


Voxel& Voxel8u::operator=(const Voxel& v) {
  val = (unsigned char) (v.getValueAsDouble()+0.5);
  return *this;
}

double Voxel8u::getValueAsDouble(void) const {
  return (double) val;
}

void Voxel8u::setValueFromDouble(double _val) {
  val = (unsigned char) (_val+0.5);
}

LONGINT Voxel8u::getValueAsLong(void) const {
  return (LONGINT) val;
}

void Voxel8u::print(TclString& value) const {
  value.sprintf("%3i", val);
}

void Voxel8u::scan(const TclString& value) throw(InvalidValue) {
  int i;
  if(sscanf((char *) value, "%i", &i) != 1) throw InvalidValue("Voxel8::scan");
  val = i;
}


Voxel& Voxel8s::operator=(const Voxel& v) {
  val = (char) (v.getValueAsDouble()+0.5);
  return *this;
}

double Voxel8s::getValueAsDouble(void) const {
  return (double) val;
}

void Voxel8s::setValueFromDouble(double _val) {
  if ( _val > 0 )
    val = (char) (_val+0.5);
  else
    val = (char) (_val-0.5);
}

LONGINT Voxel8s::getValueAsLong(void) const {
  return (LONGINT) val;
}

void Voxel8s::print(TclString& value) const {
  value.sprintf("%4i", val);
}

void Voxel8s::scan(const TclString& value) throw(InvalidValue) {
  int i;
  if(sscanf((char *) value, "%i", &i) != 1) throw InvalidValue("Voxel8::scan");
  val = i;
}


Voxel& Voxel16s::operator=(const Voxel& v) {
  val = (short int) (v.getValueAsDouble()+0.5);
  return *this;
}

double Voxel16s::getValueAsDouble(void) const {
  return (double) val;
}

void Voxel16s::setValueFromDouble(double _val) {
  if ( _val > 0 )
    val = (short int) (_val+0.5);
  else
    val = (short int) (_val-0.5);
}

LONGINT Voxel16s::getValueAsLong(void) const {
  return (LONGINT) val;
}

void Voxel16s::print(TclString& value) const {
  value.sprintf("%6i", val);
}

void Voxel16s::scan(const TclString& value) throw(InvalidValue) {
  int i;
  if(sscanf((char *) value, "%i", &i) != 1) throw InvalidValue("Voxel16s::scan");
  val = i;
}


Voxel& Voxel16u::operator=(const Voxel& v) {
  val = (unsigned short int) (v.getValueAsDouble()+0.5);
  return *this;
}

double Voxel16u::getValueAsDouble(void) const {
  return (double) val;
}

LONGINT Voxel16u::getValueAsLong(void) const {
  return (LONGINT) val;
}

void Voxel16u::setValueFromDouble(double _val) {
  val = (unsigned short int) (_val+0.5);
}

void Voxel16u::print(TclString& value) const {
  value.sprintf("%5i", val);
}

void Voxel16u::scan(const TclString& value) throw(InvalidValue) {
  int i;
  if(sscanf((char *) value, "%i", &i) != 1) throw InvalidValue("Voxel16u::scan");
  val = i;
}


Voxel& Voxel32s::operator=(const Voxel& v) {
  val = (int) (v.getValueAsDouble()+0.5);
  return *this;
}

double Voxel32s::getValueAsDouble(void) const {
  return (double) val;
}

void Voxel32s::setValueFromDouble(double _val) {
  if ( _val > 0 )
    val = (int) (_val+0.5);
  else
    val = (int) (_val-0.5);
}

LONGINT Voxel32s::getValueAsLong(void) const {
  return (LONGINT) val;
}

void Voxel32s::print(TclString& value) const {
  value.sprintf("%11i", val);
}

void Voxel32s::scan(const TclString& value) throw(InvalidValue) {
  if(sscanf((char *) value, "%i", &val) != 1) throw InvalidValue("Voxel32s::scan");
}


Voxel& Voxel32u::operator=(const Voxel& v) {
  val = (unsigned int) (v.getValueAsDouble()+0.5);
  return *this;
}

double Voxel32u::getValueAsDouble(void) const {
  return (double) val;
}

void Voxel32u::setValueFromDouble(double _val) {
  val = (unsigned int) (_val+0.5);
}

LONGINT Voxel32u::getValueAsLong(void) const {
  return (LONGINT) val;
}

void Voxel32u::print(TclString& value) const {
  value.sprintf("%10u", val);
}

void Voxel32u::scan(const TclString& value) throw(InvalidValue) {
  if(sscanf((char *) value, "%u", &val) != 1) throw InvalidValue("Voxel32u::scan");
}


Voxel& Voxel64s::operator=(const Voxel& v) {
  val = (LONGINT) (v.getValueAsDouble()+0.5);
  return *this;
}

double Voxel64s::getValueAsDouble(void) const {
  return (double) val;
}

void Voxel64s::setValueFromDouble(double _val) {
  val = (LONGINT) (_val+0.5);
}

LONGINT Voxel64s::getValueAsLong(void) const {
  return (LONGINT)val;
}

void Voxel64s::print(TclString& value) const {
  value.sprintf("%20li", (long int)val);
}

void Voxel64s::scan(const TclString& value) throw(InvalidValue) {
  if(sscanf((char *) value, "%li", (long int*)&val) != 1) throw InvalidValue("Voxel64s::scan");
}


Voxel& Voxel64u::operator=(const Voxel& v) {
  val = (unsigned LONGINT) (v.getValueAsDouble()+0.5);
  return *this;
}

double Voxel64u::getValueAsDouble(void) const {
  return (double) val;
}

void Voxel64u::setValueFromDouble(double _val) {
  val = (unsigned LONGINT) (_val+0.5);
}

LONGINT Voxel64u::getValueAsLong(void) const {
  return (LONGINT) val;
}

void Voxel64u::print(TclString& value) const {
  value.sprintf("%19lu", (long unsigned int)val);
}

void Voxel64u::scan(const TclString& value) throw(InvalidValue) {
  if(sscanf((char *) value, "%lu", (long unsigned int*)&val) != 1) throw InvalidValue("Voxel64u::scan");
}


Voxel& VoxelFloat::operator=(const Voxel& v) {
  val = (float) v.getValueAsDouble();
  return *this;
}

double VoxelFloat::getValueAsDouble(void) const {
  return (double) val;
}

void VoxelFloat::setValueFromDouble(double _val) {
  val = (float) _val;
}

LONGINT VoxelFloat::getValueAsLong(void) const {
  return (LONGINT) val;
}

void VoxelFloat::print(TclString& value) const {
  value.sprintf("%+11E", val);
}

void VoxelFloat::scan(const TclString& value) throw(InvalidValue) {
  if(sscanf((char *) value, "%f", &val) != 1) throw InvalidValue("VoxelFloat::scan");
}


Voxel& VoxelDouble::operator=(const Voxel& v) {
  val = v.getValueAsDouble();
  return *this;
}

double VoxelDouble::getValueAsDouble(void) const {
  return val;
}

void VoxelDouble::setValueFromDouble(double _val) {
  val = _val;
}

LONGINT VoxelDouble::getValueAsLong(void) const {
  return (LONGINT) val;
}

void VoxelDouble::print(TclString& value) const {
  value.sprintf("%+11E", val);
}

void VoxelDouble::scan(const TclString& value) throw(InvalidValue) {
  if(sscanf((char *) value, "%lf", &val) != 1) throw InvalidValue("VoxelDouble::scan");
}


Voxel& VoxelRGB::operator=(const Voxel& v) {
  const VoxelRGB *vrgb;
  const VoxelRGBA *vrgba;
  const VoxelFV *vfv;
  const VoxelDV *vdv;

  if((vrgb = dynamic_cast<const VoxelRGB *>(&v)) != 0) {
    val[0] = vrgb->getValue()[0];
    val[1] = vrgb->getValue()[1];
    val[2] = vrgb->getValue()[2];
  }
  else if((vrgba = dynamic_cast<const VoxelRGBA *>(&v)) != 0) {
    val[0] = vrgba->getValue()[0];
    val[1] = vrgba->getValue()[1];
    val[2] = vrgba->getValue()[2];
  }
  else if((vfv = dynamic_cast<const VoxelFV *>(&v)) != 0) {
    if(vfv->getVdim() >= 3) {
      val[0] = (unsigned char) (vfv->getValue()[0]+0.5);
      val[1] = (unsigned char) (vfv->getValue()[1]+0.5);
      val[2] = (unsigned char) (vfv->getValue()[2]+0.5);
    }
    else
      val[0] = val[1] = val[2] = (unsigned char) (vfv->getValueAsDouble()+0.5);
  }
  else if((vdv = dynamic_cast<const VoxelDV *>(&v)) != 0) {
    if(vdv->getVdim() >= 3) {
      val[0] = (unsigned char) (vdv->getValue()[0]+0.5);
      val[1] = (unsigned char) (vdv->getValue()[1]+0.5);
      val[2] = (unsigned char) (vdv->getValue()[2]+0.5);
    }
    else
      val[0] = val[1] = val[2] = (unsigned char) (vdv->getValueAsDouble()+0.5);
  }
  else {
    val[0] = val[1] = val[2] = (unsigned char) (v.getValueAsDouble()+0.5);
  }

  return *this;
}

double VoxelRGB::getValueAsDouble(void) const {
  return (double) (val[0] + val[1] + val[2]) / 3.0;
}

void VoxelRGB::setValueFromDouble(double _val) {
  val[0] = val[1] = val[2] = (unsigned char) (_val+0.5);
}

LONGINT VoxelRGB::getValueAsLong(void) const {
  return (LONGINT) ((double) (val[0] + val[1] + val[2]) / 3.0 + 0.5);
}

void VoxelRGB::print(TclString& value) const {
    value.sprintf("#%02X%02X%02X", val[0], val[1], val[2]);
}

void VoxelRGB::scan(const TclString& value) throw(InvalidValue) {
  int r, g, b;
  char from[] = "VoxelRGB::scan";

  if(value.getLength() != 7) throw InvalidValue(from);
  if(sscanf((char *) value, "#%2x%2x%2x", &r, &g, &b) != 3)
    throw InvalidValue(from);
  val[0] = r;
  val[1] = g;
  val[2] = b;
}


Voxel& VoxelRGBA::operator=(const Voxel& v) {
  const VoxelRGB *vrgb;
  const VoxelRGBA *vrgba;
  const VoxelFV *vfv;
  const VoxelDV *vdv;

  if((vrgba = dynamic_cast<const VoxelRGBA *>(&v)) != 0) {
    val[0] = vrgba->getValue()[0];
    val[1] = vrgba->getValue()[1];
    val[2] = vrgba->getValue()[2];
    val[3] = vrgba->getValue()[3];
  }
  else if((vrgb = dynamic_cast<const VoxelRGB *>(&v)) != 0) {
    val[0] = vrgb->getValue()[0];
    val[1] = vrgb->getValue()[1];
    val[2] = vrgb->getValue()[2];
    val[3] = 255;
  }
  else if((vfv = dynamic_cast<const VoxelFV *>(&v)) != 0) {
    if(vfv->getVdim() >= 4) {
      val[0] = (unsigned char) (vfv->getValue()[0]+0.5);
      val[1] = (unsigned char) (vfv->getValue()[1]+0.5);
      val[2] = (unsigned char) (vfv->getValue()[2]+0.5);
      val[3] = (unsigned char) (vfv->getValue()[3]+0.5);
    }
    else
      val[0] = val[1] = val[2] = val[3] =
	(unsigned char) (vfv->getValueAsDouble()+0.5);
  }
  else if((vdv = dynamic_cast<const VoxelDV *>(&v)) != 0) {
    if(vdv->getVdim() >= 4) {
      val[0] = (unsigned char) (vdv->getValue()[0]+0.5);
      val[1] = (unsigned char) (vdv->getValue()[1]+0.5);
      val[2] = (unsigned char) (vdv->getValue()[2]+0.5);
      val[3] = (unsigned char) (vdv->getValue()[3]+0.5);
    }
    else
      val[0] = val[1] = val[2] = val[3] =
	(unsigned char) (vdv->getValueAsDouble()+0.5);
  }
  else {
    val[0] = val[1] = val[2] = val[3] = (unsigned char) (v.getValueAsDouble()+0.5);
  }
  
  return *this;
}

double VoxelRGBA::getValueAsDouble(void) const {
  return (double) (val[0] + val[1] + val[2]) / 3.0;
}

void VoxelRGBA::setValueFromDouble(double _val) {
  val[0] = val[1] = val[2] = (unsigned char) (_val+0.5);
  val[4] = 255;
}

LONGINT VoxelRGBA::getValueAsLong(void) const {
  return (LONGINT) ((double) (val[0] + val[1] + val[2]) / 3.0 + 0.5);
}

void VoxelRGBA::print(TclString& value) const {
   value.sprintf("#%02X%02X%02X%02X", val[0], val[1], val[2], val[3]);
}

void VoxelRGBA::scan(const TclString& value) throw(InvalidValue) {
  int r, g, b, a;
  char from[] = "VoxelRGBA::scan";

  if(value.getLength() != 9) throw InvalidValue(from);
  if(sscanf((char *) value, "#%2x%2x%2x%2x", &r, &g, &b, &a) != 4)
    throw InvalidValue(from);
  val[0] = r;
  val[1] = g;
  val[2] = b;
  val[3] = a;
}



VoxelFV::VoxelFV(float *_val, unsigned int _vdim) : val(0), vdim(_vdim) {
  val = new float[vdim];
  if(_val)
    for(unsigned int i = 0; i < vdim; i++) val[i] = _val[i];
  else
    for(unsigned int i = 0; i < vdim; i++) val[i] = 0.0;
}

VoxelFV::VoxelFV(float  _val, unsigned int _vdim) : val(0), vdim(_vdim) {
  val = new float[vdim];
  for(unsigned int i = 0; i < vdim; i++) val[i] = _val;
}

VoxelFV::VoxelFV(const VoxelFV& v) : Voxel(){
  vdim = v.vdim;
  val = new float[vdim];
  for(unsigned int i = 0; i < vdim; i++) val[i] = v.val[i];
}

VoxelFV::~VoxelFV() {
  if(val) delete[] val;
}

VoxelFV& VoxelFV::operator=(const VoxelFV& v) {
  if(vdim != v.vdim) {
    delete[] val;
    vdim = v.vdim;
    val = new float[vdim];
  }
  for(unsigned int i = 0; i < vdim; i++) val[i] = v.val[i];  
  return *this;
}

Voxel& VoxelFV::operator=(const Voxel& v) {
  const VoxelRGB *vrgb;
  const VoxelRGBA *vrgba;
  const VoxelFV *vfv;
  const VoxelDV *vdv;

  if((vfv = dynamic_cast<const VoxelFV *>(&v)) != 0) {
    *this = *vfv;
  }
  else if((vdv = dynamic_cast<const VoxelDV *>(&v)) != 0) {
    if(vdim != vdv->getVdim()) {
      delete[] val;
      vdim = vdv->getVdim();
      val = new float[vdim];
    }
    for(unsigned int i = 0; i < vdim; i++)
      val[i] = (float) vdv->getValue()[i];  
  }
  else if((vrgb = dynamic_cast<const VoxelRGB *>(&v)) != 0) {
    if(vdim != 3) {
      delete[] val;
      vdim = 3;
      val = new float[3];
    }
    for(unsigned int i = 0; i < vdim; i++)
      val[i] = (float) vrgb->getValue()[i];
  }
  else if((vrgba = dynamic_cast<const VoxelRGBA *>(&v)) != 0) {
    if(vdim != 4) {
      delete[] val;
      vdim = 4;
      val = new float[4];
    }
    for(unsigned int i = 0; i < vdim; i++)
      val[i] = (float) vrgba->getValue()[i];
  }
  else {
    if(vdim != 1) {
      delete[] val;
      vdim = 1;
      val = new float[1];
    }
    val[0] = (float) v.getValueAsDouble();  
  }

  return *this;
}

void VoxelFV::setValue(float *_val, unsigned int _vdim) {
  if(val) delete[] val;
  vdim = _vdim;
  val = new float[vdim];
  for(unsigned int i = 0; i < vdim; i++) val[i] = _val[i];
}

double VoxelFV::getValueAsDouble(void) const {
  double n = 0.0;
  for(unsigned int i = 0; i < vdim; i++)
    n += val[i] * val[i];
  return sqrt(n);
}

void VoxelFV::setValueFromDouble(double _val) {
  for(unsigned int i = 0; i < vdim; i++)
    val[i] = (float) _val;
}

LONGINT VoxelFV::getValueAsLong(void) const {
  return (LONGINT) getValueAsDouble();
}

void VoxelFV::print(TclString& value) const {
  TclString one;
  unsigned int i;

  if(vdim <= 3) {
    value = "(";
    for(i = 0; i < vdim;++i) {
      if(i < vdim - 1)
	one.sprintf("%+.1E ", val[i]);
      else
	one.sprintf("%+.1E)", val[i]);
      value += one;
    }
  }
  else {
    double norm = 0.0;
    for(i = 0; i < vdim; i++) norm += val[i] * val[i];
    norm = sqrt(norm);
    value.sprintf("|%+.1E|", norm);
  }
}

void VoxelFV::scan(const TclString& value) throw(InvalidValue) {
  unsigned int n = 0, r;

  for(unsigned int i = 0; i < vdim; i++) {
    if(sscanf((char *) value + n, "%f%n", &val[i], &r) != 1)
      throw InvalidValue("VoxelFV::scan");
    n += r;
  }
}

void VoxelFV::setMin(void) {
  for(unsigned int i = 0; i < vdim; i++) val[i] = MIN_FLOAT;
}

void VoxelFV::setMax(void) {
  for(unsigned int i = 0; i < vdim; i++) val[i] = MAX_FLOAT;
}



VoxelDV::VoxelDV(double *_val, unsigned int _vdim) : val(0), vdim(_vdim) {
  val = new double[vdim];
  if(_val)
    for(unsigned int i = 0; i < vdim; i++) val[i] = _val[i];
  else
    for(unsigned int i = 0; i < vdim; i++) val[i] = 0.0;
}

VoxelDV::VoxelDV(double  _val, unsigned int _vdim) : val(0), vdim(_vdim) {
  val = new double[vdim];
  for(unsigned int i = 0; i < vdim; i++) val[i] = _val;
}

VoxelDV::VoxelDV(const VoxelDV& v) : Voxel(){
  vdim = v.vdim;
  val = new double[vdim];
  for(unsigned int i = 0; i < vdim; i++) val[i] = v.val[i];
}

VoxelDV::~VoxelDV() {
  if(val) delete[] val;
}

VoxelDV& VoxelDV::operator=(const VoxelDV& v) {
  if(vdim != v.vdim) {
    delete[] val;
    vdim = v.vdim;
    val = new double[vdim];
  }
  for(unsigned int i = 0; i < vdim; i++) val[i] = v.val[i];  
  return *this;
}

Voxel& VoxelDV::operator=(const Voxel& v) {
  const VoxelRGB *vrgb;
  const VoxelRGBA *vrgba;
  const VoxelFV *vfv;
  const VoxelDV *vdv;

  if((vdv = dynamic_cast<const VoxelDV *>(&v)) != 0) {
    *this = *vdv;
  }
  else if((vfv = dynamic_cast<const VoxelFV *>(&v)) != 0) {
    if(vdim != vfv->getVdim()) {
      delete[] val;
      vdim = vfv->getVdim();
      val = new double[vdim];
    }
    for(unsigned int i = 0; i < vdim; i++) val[i] = 
					     (double) vfv->getValue()[i];  
  }
  else if((vrgb = dynamic_cast<const VoxelRGB *>(&v)) != 0) {
    if(vdim != 3) {
      delete[] val;
      vdim = 3;
      val = new double[3];
    }
    for(unsigned int i = 0; i < vdim; i++) val[i] =
					     (double) vrgb->getValue()[i];  
  }
  else if((vrgba = dynamic_cast<const VoxelRGBA *>(&v)) != 0) {
    if(vdim != 4) {
      delete[] val;
      vdim = 4;
      val = new double[4];
    }
    for(unsigned int i = 0; i < vdim; i++)
      val[i] = (double) vrgba->getValue()[i];  
  }
  else {
    if(vdim != 1) {
      delete[] val;
      vdim = 1;
      val = new double[1];
    }
    val[0] = v.getValueAsDouble();  
  }

  return *this;
}

void VoxelDV::setValue(double *_val, unsigned int _vdim) {
  if(val) delete[] val;
  vdim = _vdim;
  val = new double[vdim];
  for(unsigned int i = 0; i < vdim; i++) val[i] = _val[i];
}

double VoxelDV::getValueAsDouble(void) const {
  double n = 0.0;
  for(unsigned int i = 0; i < vdim; i++)
    n += val[i] * val[i];
  return sqrt(n);
}

void VoxelDV::setValueFromDouble(double _val) {
  for(unsigned int i = 0; i < vdim; i++)
    val[i] = _val;
}

LONGINT VoxelDV::getValueAsLong(void) const {
  return (LONGINT) getValueAsDouble();
}

void VoxelDV::print(TclString& value) const {
  TclString one;
  unsigned int i;
  if(vdim <= 3) {
    value = "(";
    for(i = 0; i < vdim; i++) {
      if(i < vdim - 1)
	one.sprintf("%+.1E ", val[i]);
      else
	one.sprintf("%+.1E)", val[i]);
      value += one;
    }
  }
  else {
    double norm = 0.0;
    for(i = 0; i < vdim; i++) norm += val[i] * val[i];
    norm = sqrt(norm);
    value.sprintf("|%+3.1E|", norm);
  }
}

void VoxelDV::scan(const TclString& value) throw(InvalidValue) {
  unsigned int n = 0, r;

  for(unsigned int i = 0; i < vdim; i++) {
    if(sscanf((char *) value + n, "%lf%n", &val[i], &r) != 1)
      throw InvalidValue("VoxelDV::scan");\
    n += r;
  }
}

void VoxelDV::setMin(void) {
  for(unsigned int i = 0; i < vdim; i++) val[i] = MIN_DOUBLE;
}

void VoxelDV::setMax(void) {
  for(unsigned int i = 0; i < vdim; i++) val[i] = MAX_DOUBLE;
}



Voxel *Voxel::voxelByType(Inrimage::WORD_TYPE wt, unsigned int vdim) {
  switch(wt) {
  case Inrimage::WT_UNSIGNED_CHAR:
    return (Voxel *) new Voxel8u(0);
  case Inrimage::WT_SIGNED_CHAR:
    return (Voxel *) new Voxel8s(0);
  case Inrimage::WT_UNSIGNED_SHORT:
    return (Voxel *) new Voxel16u(0);
  case Inrimage::WT_SIGNED_SHORT:
    return (Voxel *) new Voxel16s(0);
  case Inrimage::WT_UNSIGNED_INT:
    return (Voxel *) new Voxel32u(0);
  case Inrimage::WT_SIGNED_INT:
    return (Voxel *) new Voxel32s(0);
  case Inrimage::WT_UNSIGNED_LONG:
    return (Voxel *) new Voxel64u(0);
  case Inrimage::WT_SIGNED_LONG:
    return (Voxel *) new Voxel64s(0);
  case Inrimage::WT_FLOAT_VECTOR:
    if ( vdim == 0 ) return 0;
    if ( vdim >  1 ) return (Voxel *) new VoxelFV((float)0.0, vdim);
  case Inrimage::WT_FLOAT:
    return (Voxel *) new VoxelFloat(0);
  case Inrimage::WT_DOUBLE_VECTOR:
    if ( vdim == 0 ) return 0;
    if ( vdim >  1 ) return (Voxel *) new VoxelDV((double)0.0, vdim);
  case Inrimage::WT_DOUBLE:
    return (Voxel *) new VoxelDouble(0);
  case Inrimage::WT_RGB:
    return (Voxel *) new VoxelRGB(0, 0, 0);
  case Inrimage::WT_RGBA:
    return (Voxel *) new VoxelRGBA(0, 0, 0, 0);
  }
  return 0;
}


} // namespace yav

