#include <stdio.h>
#include <stdlib.h>




static int _VERBOSE_ = 0;


/* 
 * on verifie, et eventuellement on change (en 2D) la connexite
 */

static int CheckAndEvaluateConnectivity( int connectivity,
					 int dimz )
{
  int c = 26;

  /* seules les connexites 26, 18, 10, 6 (en 3D) et 8 et 4 (en 2D)
     sont permises
  */
  switch ( connectivity ) {
  default : break;
  case 26 :
  case 18 :
  case 10 :
  case 6 :
  case 8 :
  case 4 : c = connectivity;
  }

  /* cas 2D
   */
  if ( dimz <= 1 ) {
    switch ( c ) {
    default :
    case 26 :
    case 18 :
    case 10 : c = 8; break;
    case 6 : 
    case 4 : c = 4;
    }
  }
  return( c );
}





/* 
 * calcule les offsets pour connaitre les voisins du point
 * courant selon la connexite.
 * Retourne le nombre de voisins.
 */

static int InitNeighborsOffsets( int array_offset[3][3][2], 
				 int offsets[13],
				 unsigned int dimx,
				 unsigned int dimy,
				 int connectivity )
{
  int nb=0;
  int i,j,k;
  unsigned int dimxy = dimx*dimy; 

  for ( i = 0; i < 3; i ++ )
  for ( j = 0; j < 3; j ++ )
  for ( k = 0; k < 2; k ++ )
    array_offset[i][j][k] = 0;
  for ( i = 0; i < 13; i ++ )
    offsets[i] = 0;

  switch ( connectivity ) {
  case 8 :
    array_offset[0][0][1] = offsets[ nb++ ] =         - (int)dimx - 1; /* point (-1, -1,  0) */
    array_offset[2][0][1] = offsets[ nb++ ] =         - (int)dimx + 1; /* point ( 1, -1,  0) */
  case 4 :
    array_offset[1][0][1] = offsets[ nb++ ] =         - (int)dimx;     /* point ( 0, -1,  0) */
    array_offset[0][1][1] = offsets[ nb++ ] =                - 1; /* point (-1,  0,  0) */
    break;
  case 26 :
    array_offset[0][0][0] = offsets[ nb++ ] = - (int) dimxy - (int)dimx - 1; /* point (-1, -1, -1) */
    array_offset[2][0][0] = offsets[ nb++ ] = - (int)dimxy - (int)dimx + 1; /* point ( 1, -1, -1) */
    array_offset[0][2][0] = offsets[ nb++ ] = - (int)dimxy + dimx - 1; /* point (-1,  1, -1) */
    array_offset[2][2][0] = offsets[ nb++ ] = - (int)dimxy + dimx + 1; /* point ( 1,  1, -1) */
  case 18 :
    array_offset[1][0][0] = offsets[ nb++ ] = - (int)dimxy - (int)dimx;     /* point ( 0, -1, -1) */
    array_offset[0][1][0] = offsets[ nb++ ] = - (int)dimxy        - 1; /* point (-1,  0, -1) */
    array_offset[2][1][0] = offsets[ nb++ ] = - (int)dimxy        + 1; /* point ( 1,  0, -1) */
    array_offset[1][2][0] = offsets[ nb++ ] = - (int)dimxy + dimx;     /* point ( 0, -1, -1) */
  case 10 :
    array_offset[0][0][1] = offsets[ nb++ ] =         - (int)dimx - 1; /* point (-1, -1,  0) */
    array_offset[2][0][1] = offsets[ nb++ ] =         - (int)dimx + 1; /* point ( 1, -1,  0) */
  case 6 :
    array_offset[1][1][0] = offsets[ nb++ ] = - (int)dimxy;            /* point ( 0,  0, -1) */
    array_offset[1][0][1] = offsets[ nb++ ] =         - (int)dimx;     /* point ( 0, -1,  0) */
    array_offset[0][1][1] = offsets[ nb++ ] =                - 1; /* point (-1,  0,  0) */
  }
  return( nb );
}















int compareConnectedComponents( const void* c1, const void* c2 ) 
{
  typeConnectedComponent *cc1 = (typeConnectedComponent *)c1;
  typeConnectedComponent *cc2 = (typeConnectedComponent *)c2;
  
  if ( cc1->pointsAboveLowThreshold > cc2->pointsAboveLowThreshold ) return( -1 );
  if ( cc1->pointsAboveLowThreshold < cc2->pointsAboveLowThreshold ) return(  1 );
  return( 0 );
}















/*
 * 
 */ 
typedef enum {
  _FROM_LARGEST_,
  _FROM_SMALLEST_ 
} enumRelabelOrder;




/* Gestion du depassement du label maximal utilise.
   
   Pour les composantes deja entierement traitees, on
   determine si elles sont valides ou non.
   => on recupere les labels des composantes
   invalidees.
   
   Pour les composantes valides, et les composantes
   non entierement traitees, on leur attribue a toutes
   un seul label, et on recupere ainsi tous 
   les labels intermediaires.
*/

template <class TYPE > static int _LabelsOverflowManagement( TYPE *inputBuf,
							     /* input buffer, it is already
								labeled until the point (x,y,z)
								not included 
							     */
							     unsigned int dimx, unsigned int dimy,
							     /* dimension of the input buffer */
							     unsigned int x, unsigned int y, unsigned int z,
							     /* next point to be processed */
							     int connectivity,
							     /* connectivity */
							     
							     unsigned int *first_label,
							     unsigned int *last_label,
							     /* first and last labels to be considered */
							     
							     typeConnectedComponentsManagement *env,
							     enumRelabelOrder relabelOrder )
{
  char *proc="_LabelsOverflowManagement";
  int local_connectivity = CheckAndEvaluateConnectivity( connectivity, z+1 );
  int xlast = x;
  int ylast = y;
  int zlast = z;
  int xnum, ynum, znum;

  TYPE *theBuf = inputBuf;
  int new_label = 0;
  int nb_valid_cc = 0;
  int binary_label = -1;

  unsigned int i, j;
  unsigned int minv, maxv;
  unsigned int fv, lv;
  int dv;
  typeConnectedComponent *components = (typeConnectedComponent *)NULL;

  minv = ( (*first_label) <  (*last_label) ) ? (*first_label) : (*last_label);
  maxv = ( (*first_label) >= (*last_label) ) ? (*first_label) : (*last_label);
  

  switch ( relabelOrder ) {
  default :
  case _FROM_SMALLEST_ :
    fv = minv;
    lv = maxv;
    dv = 1;
    break;
  case _FROM_LARGEST_ :
    fv = maxv;
    lv = minv;
    dv = -1;
    break;
  }


  /* determination du dernier point traite
     -> (xlast, ylast, zlast)

     (x,y,z) represente le point courant pour lequel
     l'attribution d'un nouveau label a echoue

     => on cherche le dernier point traite avec succes
     (ce peut etre un point du fond)
   */
  if ( xlast < 0 || ylast < 0 || zlast < 0 ) {
    if ( _VERBOSE_ )
      fprintf( stderr, "%s: bad (negative) last point coordinate(s)\n", proc );
    return( 0 );
  }

  
  if ( xlast > 0 ) {
    xlast --;
  }
  else if ( ylast > 0 ) {
    xlast = dimx - 1;
    ylast --;
  } 
  else if ( zlast > 0 ) {
    xlast = dimx - 1;
    ylast = dimy - 1;
    zlast --;
  } else {
    if ( _VERBOSE_ )
      fprintf( stderr, "%s: all last point coordinates are zeros\n", proc );
    return( 0 );
  }


  /* pour certaines composantes (classes d'equivalence)
     on sait deja si il faut les garder ou non,
     ce sont celles qui n'ont plus de voisins
     dans les points qui restent a parcourir.
   *
     Determination du dernier point 
     -> (xnum, ynum, znum)
     pouvant appartenir a l'une de ces composantes
   *
   * On ne regarde que les cas non-degeneres,
   * c-a-d ou le dernier point traite n'appartient pas a un bord.
   */
  znum = ynum = xnum = -1;


  switch ( local_connectivity ) {
  case 4 :
    if ( ylast > 0 ) {
      xnum = xlast;
      ynum = ylast - 1;
      znum = zlast;
    } 
    else if ( zlast > 0 ) {
      xnum = dimx - 1;
      ynum = dimy - 1;
      znum = zlast - 1;
    }
    break;
  case 8 :
    if ( xlast > 0 ) {
      if ( ylast > 0 ) {
	xnum = xlast - 1;
	ynum = ylast - 1;
	znum = zlast;
      }
      else if ( zlast > 0 ) {
	xnum = dimx - 1;
	ynum = dimy - 1;
	znum = zlast - 1;
      }
    }
    else {
      if ( ylast > 1 ) {
	xnum = dimx - 1;
	ynum = ylast - 2;
	znum = zlast;
      }
      else if ( zlast > 0 ) {
	xnum = dimx - 1;
	ynum = dimy - 1;
	znum = zlast - 1;
      }
    }
    break;
  case 6 :
  case 10 :
    if ( zlast > 0 ) {
      xnum = xlast;
      ynum = ylast;
      znum = zlast - 1;
    }
    break;
  case 18 :
    if ( zlast > 0 && ylast > 0 && xlast > 0 ) {
      xnum = xlast - 1;
      ynum = ylast - 1;
      znum = zlast - 1;
    }
    break;
  case 26 :
  default :
    if ( zlast > 0 && ylast > 0 && xlast > 0 ) {
      xnum = xlast - 1;
      ynum = ylast - 1;
      znum = zlast - 1;
    }
  }




  /* on regarde les composantes pour une elimination eventuelle
     
     1. Si (xnum, ynum, znum) est valide,
        alors les composantes dont tous les points se trouvent
	entre (0,0,0) et (xnum, ynum, znum) inclus sont deja
	entierement traitees et on peut verifier 
	leur validite
	Les autres (dont au moins un point se trouve entre
	(xnum, ynum, znum)+1 et (xlast, ylast, zlast) inclus
	sont conservees telles quelles.
     2. sinon toutes les composantes sont conservees telles quelles
        sans test de validation.
  */
  if ( xnum >= 0 && ynum >= 0 && znum >= 0 ) {
    for ( i = (*first_label); i <= (*last_label); i++ )
      env->cc[i].completelyProcessed = 1;
    i = znum  * dimy * dimx + ynum  * dimx + xnum;
    j = zlast * dimy * dimx + ylast * dimx + xlast;
    for ( i +=1; i <= j; i++ ) {
      if ( theBuf[i] > 0 ) {
	env->cc[ env->cc[ (int)theBuf[i] ].label ].completelyProcessed = 0;
      }
    }
  } else {
    for ( i = (*first_label); i <= (*last_label); i++ )
      env->cc[i].completelyProcessed = 0;
  }
  
  /* ajout pour traiter le cas ou les composantes sont numerotees
     a partir du haut (_FROM_LARGEST_)
  */
  for ( i = (*first_label); i <= (*last_label); i++ )
    env->cc[i].representative = 0;


  /* TO BE DONE
     si un nombre maximum de composantes est specifie,
     on pourrait garder les plus grandes composantes
     parmi celles qui sont specifiees
     
     il faudrait faire un tableau d'entier,
     pour garder les tailles, les trier
     (pour avoir la taille de la n-ieme),
     et marquer (avec invalid_label), celles
     qui sont plus petites que cette taille
  */


  /* allocation d'un nouveau tableau d'equivalence
   */
  components = (typeConnectedComponent *)malloc( env->cc_array_size * sizeof(typeConnectedComponent) );
  if ( components == (typeConnectedComponent *)NULL ) {
    if ( _VERBOSE_ )
      fprintf( stderr, "%s: unable to allocate equivalence array\n", proc );
    return( 0 );
  }
  memcpy( components, env->cc, env->cc_array_size * sizeof(typeConnectedComponent) );





  /* on regarde les labels et on cherche
     les representants des classes d'equivalences (labels[i] = i)
     1. si la composante peut encore grandir 
        (completelyProcessed = 0) => on la garde
     2. sinon, on teste sa validite
  *
     si l'output est binaire, on garde la meme valeur
     pour toutes les composantes valides
     (on fait comme si c'etait une seule composante)
  */

  new_label = fv;
  nb_valid_cc = 0;


  switch ( relabelOrder ) {
  default :
  case _FROM_SMALLEST_ :
    
    for ( i = fv; i <= lv; i ++ ) {

      /* c'est le label representant la classe d'equivalence
       */
      if ( env->cc[ i ].label == i ) {

	/* 
	 * la composante peut encore grandir => on la garde telle quelle
	 * To be done: here we can also check if the component is already valid
	 * and labeled it consequently in case of binary output
	 */
	if ( env->cc[ i ].completelyProcessed == 0 ) {
	  env->cc[ i ].label = new_label; 
	  components[ new_label ] = env->cc[i];
	  new_label += dv;
	}
	
	/* on a deja tous les points;
	   la composante ne grandira plus
	*/
	else {
	  
	  /* on verifie que la composante est valide
	   */
	  if ( (env->cc[ i ].pointsAboveHighThreshold >= env->minNumberOfPtsAboveHigh) &&
	       (env->cc[ i ].pointsAboveLowThreshold  >= env->minNumberOfPtsAboveLow) ) {
	    nb_valid_cc ++;
	    
	    /* mode binaire, on recupere le numero de la premiere composante
	     * complete, et on affectera ce numero a TOUTES les composantes
	     * valides
	     *
	     * 
	     */
	    if ( env->output == COMPONENT_BINARY ) {
	      if ( nb_valid_cc == 1 ) {
		env->cc[ i ].label = binary_label = new_label;
		components[ new_label ] = env->cc[i];
		new_label += dv;
	      } else {
		env->cc[ i ].label = binary_label;
		components[ binary_label ].pointsAboveLowThreshold  += env->cc[ i ].pointsAboveLowThreshold;
		components[ binary_label ].pointsAboveHighThreshold += env->cc[ i ].pointsAboveHighThreshold;
	      }
	    } 
	    
	    /* mode etiquettes:
	       on garde la composante telle quelle
	       et on incremente les labels attribues
	    */
	    else {
	      env->cc[ i ].label = new_label; 
	      components[ new_label ] = env->cc[i];
	      new_label += dv;
	    }
	  } 
	  
	  /* on peut ne plus considerer la composante
	     elle n'est pas valide
	  */
	  else {
	    env->cc[ i ].label = env->invalid_label;
	  }
	}
      } 
      
      /* ce n'est pas le label representant la classe d'equivalence
	 on recupere le nouveau label de celle-ci
      */
      else {
	j = env->cc[ i ].label; 
	env->cc[ i ].label = env->cc[ j ].label;
      }
    }

    break;

  case _FROM_LARGEST_ :

    /* ajout pour traiter le cas ou les composantes sont numerotees
       a partir du haut (_FROM_LARGEST_)
    */
    for ( i = (*first_label); i <= (*last_label); i++ )
      env->cc[i].representative = 0;
    
    for ( i = fv; i >= lv; i -- ) {

      /* c'est le label representant la classe d'equivalence
       */
      if ( env->cc[ i ].label == i ) {
	env->cc[i].representative = 1;

	/* 
	 * la composante peut encore grandir => on la garde telle quelle
	 * To be done: here we can also check if the component is already valid
	 * and labeled it consequently in case of binary output
	 */
	if ( env->cc[ i ].completelyProcessed == 0 ) {
	  env->cc[ i ].label = new_label; 
	  components[ new_label ] = env->cc[i];
	  new_label += dv;
	}
	
	/* on a deja tous les points;
	   la composante ne grandira plus
	*/
	else {
	  
	  /* on verifie que la composante est valide
	   */
	  if ( (env->cc[ i ].pointsAboveHighThreshold >= env->minNumberOfPtsAboveHigh) &&
	       (env->cc[ i ].pointsAboveLowThreshold  >= env->minNumberOfPtsAboveLow) ) {
	    nb_valid_cc ++;
	    
	    /* mode binaire, on recupere le numero de la premiere composante
	     * complete, et on affectera ce numero a TOUTES les composantes
	     * valides
	     *
	     * 
	     */
	    if ( env->output == COMPONENT_BINARY ) {
	      if ( nb_valid_cc == 1 ) {
		env->cc[ i ].label = binary_label = new_label;
		components[ new_label ] = env->cc[i];
		new_label += dv;
	      } else {
		env->cc[ i ].label = binary_label;
		components[ binary_label ].pointsAboveLowThreshold  += env->cc[ i ].pointsAboveLowThreshold;
		components[ binary_label ].pointsAboveHighThreshold += env->cc[ i ].pointsAboveHighThreshold;
	      }
	    } 
	    
	    /* mode etiquettes:
	       on garde la composante telle quelle
	       et on incremente les labels attribues
	    */
	    else {
	      env->cc[ i ].label = new_label; 
	      components[ new_label ] = env->cc[i];
	      new_label += dv;
	    }
	  } 
	  
	  /* on peut ne plus considerer la composante
	     elle n'est pas valide
	  */
	  else {
	    env->cc[ i ].label = env->invalid_label;
	  }
	}
      } 
    }

    
    /* on repercute les labels des representatives
     */
    for ( i = (*first_label); i <= (*last_label); i++ )
      if ( env->cc[i].representative == 0 ) {
	j = env->cc[ i ].label; 
	env->cc[ i ].label = env->cc[ j ].label;
      }
    
  }

  /* renumerote jusqu'au dernier point calcule 
   */
  j = zlast * dimy * dimx + ylast * dimx + xlast;
  for ( i = 0; i <= j; i ++, theBuf++ ) {
    if ( minv <= *theBuf && *theBuf <= maxv )
      *theBuf = (TYPE)( env->cc[ (int)(*theBuf) ].label );
  }

  /* on remonte le nombre de labels utilises
     et la nouvelle table d'equivalence
   */
  switch ( relabelOrder ) {
  default :
  case _FROM_SMALLEST_ :
    *first_label = fv;
    *last_label  = new_label - dv;
    break;
  case _FROM_LARGEST_ :
    *first_label = new_label - dv;
    *last_label  = fv;
    break;
  }

  free( env->cc );
  env->cc = components;

  if ( _VERBOSE_ ) {
    fprintf( stderr, "%s: previously used labels = [%d %d]\n", proc, fv, lv );
    fprintf( stderr, "\t newly used labels = [%d %d]\n", *first_label, *last_label );
    fprintf( stderr, "\t number of already valid components: %d\n", nb_valid_cc );
    if ( _VERBOSE_ > 1 ) {
      fprintf( stderr, "%s: last point for valid component: %3d %3d %3d\n\
\t last processed point:           %3d %3d %3d\n\
\t current point:                  %3d %3d %3d", 
	       proc, xnum, ynum, znum, xlast, ylast, zlast, x, y, z );
    }
    if ( env->output == COMPONENT_BINARY ) {
      fprintf( stderr, "%s: binary mode, label for valid components is %d", 
	       proc, binary_label );
    }
  }
  

  return( 1 );
}




































/*
 * connected components are defined by points which value is larger or equal
 * to _low_value_. Points with value lower than _low_value_ are background.
 * 
 *
 * RETURNED VALUES:
 *  number of used labels : success, 
 *      labels are from 1 to this number (included)
 *      0 means that there was no found connected components
 * -1 : erreur
 */
template <class TYPE > static int _ForegroundConnectedComponentsExtraction( TYPE *inputBuf,
									/* input buffer,
									   should contains at most 3 values
									   0, _low_value_, _hig_value_
									*/
									unsigned int dimx, 
									unsigned int dimy,
									unsigned int dimz,
									/* dimension of the input buffer */
									TYPE low_value,
									TYPE high_value,
									/* necessary values */

									int connectivity,
									/* connectivity:
									   in 2D: 4,8;
									   in 3D: 6, 10, 18, 26
									*/

									typeConnectedComponentsManagement *env
									/* equivalence array
									 */
									)
{
  char *proc = "_ForegroundConnectedComponentsExtraction";
  /* sizes 
   */
  unsigned int dimx1 = dimx-1, dimy1 = dimy-1;
  /* offsets 
   */
  int array_offset[3][3][2], offsets[13], nb_offsets;


  int i, j, k, _INSIDEY_;
  unsigned int x, y, z;
  TYPE *theBuf = (TYPE *)(inputBuf);
  TYPE current_value;

  unsigned int label_neighbors[13];
  unsigned int current_label;
  
  unsigned int first_label = 1;
  unsigned int last_label;
  unsigned int new_label = 0;

  int local_connexite;
  int nb_neighbors;
  /* int valid_labels = 0, total_labels = 0; */

  typeConnectedComponent *cc = env->cc;
  int r;



  /* initialisation de la table des composantes 
     - background value will be 0
     - first label for components will be 1
  */
  cc[0].label = 0;
  first_label = 1;
  new_label = first_label - 1;
  



  /*
   * Computation of a point's label : current_label
   * If this point has some neighbors (i.e. nb_neighbors > 0) then
   *   1. Its label is the minimum value of all the neighbors labels
   *   2. We report the equivalences between all the labels
   * If not (i.e. nb_neighbors == 0) then
   *   we use a new label (potential case of error : label overflow)
   */
#define _LABEL_COMPUTATION {                                     \
      if ( nb_neighbors > 0 ) {                                  \
	/*                                                       \
	 * If there are some neighbors,                          \
	 * the equivalent label for all of them is the minimum   \
	 * of all labels.                                        \
	 */                                                      \
	current_label = label_neighbors[0];                      \
	for ( i = 1; i < nb_neighbors; i ++ )                    \
	  if ( label_neighbors[i] < current_label )              \
            current_label = label_neighbors[i];                  \
	/*                                                       \
	 * we have to report the equivalences inside the equivalence table \
	 */                                                      \
	for ( i = 0; i < nb_neighbors; i ++ ) {                  \
	  if ( label_neighbors[i] != current_label ) {           \
            k = label_neighbors[i];                              \
            /*                                                   \
             * si la classe n'a pas deja ete traitee             \
             */                                                  \
            if ( cc[ k ].label != current_label ) {              \
	      cc[ current_label ].pointsAboveLowThreshold += cc[ k ].pointsAboveLowThreshold; \
              cc[ k ].pointsAboveLowThreshold = 0;               \
	      cc[ current_label ].pointsAboveHighThreshold += cc[ k ].pointsAboveHighThreshold; \
              cc[ k ].pointsAboveHighThreshold = 0;              \
              cc[ k ].label = current_label;                     \
	      for ( j = k+1; j <= (int)new_label; j ++ )            \
	        if ( cc[ j ].label == (unsigned int)k ) cc[ j ].label = current_label; \
            }                                                    \
	  }                                                      \
	}                                                        \
      } else {                                                   \
	/*                                                       \
	 * No neighbors :                                        \
	 * we use a new label for this point.                    \
	 * This is the only case of error in the whole procedure. \
	 * This could be a little bit improved as follows:       \
	 * 1. we re-label the potential equivalence classes (without conditions of \
	 *    validity, i.e. without testing cc[ i ].pointsAboveHighThreshold and  \
         *    cc[ i ].pointsAboveLowThreshold) to get            \
	 *    the effective number of equivalence classes at this time. \
	 * 2. using this re-labeling, we change the image labels up to the current \
	 *    point.                                             \
	 * 3. we re-build the equivalence table to have only one \
         *    entry per equivalence classe.                      \
	 */                                                      \
	current_label = ++new_label;                           \
	if ( new_label > (unsigned int)(env->cc_array_size - 1) ) {    \
          /* we try to relabel the part of the image which has already been \
             processed. This way, we use only the minimal number of labels  \
             necessary to represent this part. We hope to recover a         \
             sufficient number of labels to end the computation             \
	   */ \
          last_label = new_label - 1; \
          r = _LabelsOverflowManagement( inputBuf, dimx, dimy, x, y, z, local_connexite, \
                                            &first_label, &last_label, env, _FROM_SMALLEST_ ); \
          /* does it succeed ?                            \
	   */                                             \
          if ( r == 0 ||  last_label >= new_label - 1 ) { \
	    /* if not, I give up                          \
	     */                                           \
	    if ( _VERBOSE_ )                              \
	      fprintf( stderr, "%s: too much used labels for connected components computation\n", \
		       proc );                            \
	    return( -1 );                                 \
	  } else {                                        \
	    /* if yes,                                    \
	       we go on with the new table of equivalences \
	       and the new number of used labels          \
	    */                                            \
	    cc = env->cc;                          \
	    new_label = last_label;                              \
	    current_label = ++new_label;                \
	  }                                               \
        }                                                 \
	cc[ current_label ].label = current_label;        \
	cc[ current_label ].pointsAboveLowThreshold = 0;  \
	cc[ current_label ].pointsAboveHighThreshold = 0; \
      }                                                   \
}

  /*
   * Generic tests of all the neighbors
   * we make sure that they belong to the image.
   * We check if the neighbor has a label (value > 0)
   * We are looking for the equivalence classe (i.e. cc[ value ].label) 
   * of the neighbors
   */
#define _GENERIC_TEST {                   \
    for ( i = 0; i < nb_offsets; i ++ )   \
      if ( *(theBuf + offsets[i]) > 0 )   \
	label_neighbors[ nb_neighbors++ ] = cc[ (int)( *(theBuf + offsets[i]) ) ].label; \
}

  /*
   * Specific tests for the upper row of the neighborhood of a point
   * We check if the neighbor is connected to the current point
   *          if the neighbor belongs to the image
   *          if the neighbor has a label
   */
#define _UPPER_ROW_TESTS {                \
    if ( y > 0 ) {                        \
      for ( i = -1; i <= 1; i ++ ) {      \
	if ( array_offset[1+i][0][1] == 0 ) continue;  \
	if ( (x+i < 0) || (x+i >= dimx) )   continue;  \
	if ( *(theBuf + array_offset[1+i][0][1]) > 0 ) \
	  label_neighbors[ nb_neighbors++ ] = cc[ (int)( *(theBuf + array_offset[1+i][0][1]) ) ].label; \
      }                                   \
    }                                     \
}

  /*
   * Specific test for the left neighbor of a point
   * We should test if (array_offset[0][1][1] != 0) but we
   * do not because the point (x-1,y,z) always belongs to the
   * neighborhoods we use.
   */
#define _LEFT_NEIGHBOR_TEST {                                                   \
    if ( x > 0 )                                                                \
      if ( *(theBuf - 1) > 0 )                                                  \
	label_neighbors[ nb_neighbors++ ] = cc[ (int)( *(theBuf - 1) ) ].label; \
}

  /*
   * additional informations :
   * - the value of the current label is given to the image buffer
   * - the connect component size is incremented
   * - the validity of the connect component is checked (with respect 
   *   to the current value)
   */
#define _ADDITIONAL_INFORMATIONS {                     \
    *theBuf = (TYPE)current_label;       \
    cc[ current_label ].pointsAboveLowThreshold ++;    \
    if ( current_value >= high_value )                \
      cc[ current_label ].pointsAboveHighThreshold ++; \
}


  /*
   * Computation of all points.
   *
   * We will attribute to each point a value (a label). According to this
   * value and to the equivalence table, we know the equivalence classe 
   * of the point (i.e. cc[ value ].label).
   * 
   * The computation is divided into two parts : the first slice
   * and the others slices. This allows to avoid some tests and
   * to spare some time.
   *
   * The principle of the computation is the following :
   * For each point, we
   * - check if t may belong to a connected component
   *   => ( current_value >= _low_value_)
   * - check if tests are needed for the neighbors, i.e.
   *   if the point is on the image border. This point
   *   is different for the first slice and the other ones.
   *   first slice  -> if ( y == 0 ) _INSIDEY_ = 0;
   *   other slices -> if ( (y == 0) || (y == dimy1) ) _INSIDEY_ = 0;
   *   => ( _INSIDEY_ == 1 ) && ( x > 0 ) && ( x < dimx1 )
   * - depending on the above checking, 
   *   we compute the number of neighbors (nb_neighbors) and
   *   extract all the neighbors' values (label_neighbors[]),
   *   either without additionnal tests (_GENERIC_TEST),
   *   or with additionnal tests for the neighbors in the upper slice,
   *   in the upper row (_UPPER_ROW_TESTS) or the left neighbor
   *   (_LEFT_NEIGHBOR_TEST).
   *   The two offsets structures (offsets[] and array_offset[][][])
   *   are necessary because of this choice.
   * - Once we have the number of neighbors and their values, we can
   *   compute the point's label (_LABEL_COMPUTATION)
   *   - if the point have neighbors, its label will be the minimum
   *     of all labels, and we report the equivalence between this
   *     minimum label and the other labels in the neighborhood inside
   *     the equivalence table.
   *   - if the point does not have any neighbor, we use a new label.
   *     we increment the number of used label (++new_label) and
   *     we initialize a new label. 
   *     WARNING: an error may occur if such initialization is not possible
   *     (it depends on the type used for the image).
   * - Once the point's label is computed, we report this value in the
   *   image, increment the connected component size and test if 
   *   connected component will be valid (current_value >= _hig_value_)
   */


  /*
   * first slice
   */
  local_connexite = CheckAndEvaluateConnectivity( connectivity, (int)1 );
  nb_offsets = InitNeighborsOffsets( array_offset, offsets, dimx, dimy, local_connexite );

  z = 0;
  for ( y = 0; y < dimy; y++ ) {
    _INSIDEY_ = 1;
    if ( y == 0 ) _INSIDEY_ = 0;
    for ( x = 0; x < dimx; x++, theBuf++ ) {
      current_value = *theBuf;
      if ( current_value < low_value ) { 
	*theBuf = 0 ; 
	continue; 
      }
      nb_neighbors = 0;
      if ( ( _INSIDEY_ == 1 ) && ( x > 0 ) && ( x < dimx1 ) ) {
	_GENERIC_TEST
      } else {
	_UPPER_ROW_TESTS
	_LEFT_NEIGHBOR_TEST
      }
      _LABEL_COMPUTATION
      _ADDITIONAL_INFORMATIONS
    }
  }

  /*
   * other slices
   */
  local_connexite = CheckAndEvaluateConnectivity( connectivity, dimz );
  nb_offsets = InitNeighborsOffsets( array_offset, offsets, dimx, dimy, local_connexite );
  for ( z = 1; z < dimz; z ++ ) {
    for ( y = 0; y < dimy; y++ ) {
      _INSIDEY_ = 1;
      if ( (y == 0) || (y == dimy1) ) _INSIDEY_ = 0;
      for ( x = 0; x < dimx; x++, theBuf++ ) {
	current_value = *theBuf;
	if ( current_value < low_value ) {
	  *theBuf = 0 ; 
	  continue;
	}
	nb_neighbors = 0;
	if ( ( _INSIDEY_ == 1 ) && ( x > 0 ) && ( x < dimx1 ) ) {
	  _GENERIC_TEST
	} else {
	  /*
	   * upper slice tests
	   */
	  for ( j = -1; j <= 1; j ++ ) {
	    if ( (y+j < 0) || (y+j >= dimy) )       continue;
	    for ( i = -1; i <= 1; i ++ ) {
	      if ( (x+i < 0) || (x+i >= dimx) )     continue;
	      if ( array_offset[1+i][1+j][0] == 0 ) continue;
	      if ( *(theBuf + array_offset[1+i][1+j][0]) > 0 ) 
		label_neighbors[ nb_neighbors++ ] = cc[ (int)( *(theBuf + array_offset[1+i][1+j][0]) ) ].label;
	    }
	  }
	  _UPPER_ROW_TESTS
	  _LEFT_NEIGHBOR_TEST
	}
	_LABEL_COMPUTATION
	_ADDITIONAL_INFORMATIONS
      }
    }
  }



  /*
   * At this point, all the image's points have been processed.
   */

  if ( _VERBOSE_ ) {
    fprintf( stderr, "%s: number of used labels: %5d\n", proc, new_label );
  }




  return( new_label );
}







/*
 * connected components are defined by points which value is larger or equal
 * to _low_value_. Points with value lower than _low_value_ are background.
 * 
 *
 * RETURNED VALUES:
 *  number of used labels : success, 
 *      labels are from 1 to this number (included)
 *      0 means that there was no found connected components
 * -1 : erreur
 */
template <class TYPE > static int _ForeBackConnectedComponentsExtraction( TYPE *inputBuf,
									    /* input buffer,
									       should contains at most 3 values
									       0, _low_value_, _hig_value_
									    */
									    unsigned int dimx, 
									    unsigned int dimy,
									    unsigned int dimz,
									    /* dimension of the input buffer */
									    TYPE bg_value,
									  /* necessary values */
									    
									    int bg_connectivity,
									    int fg_connectivity,
									    /* connectivity:
									       in 2D: 4,8;
									       in 3D: 6, 10, 18, 26
									    */
									    
									    typeConnectedComponentsManagement *env,
									    /* equivalence array
									     */
									    int *backg_first_label,
									    int *backg_last_label,
									    int *foreg_first_label,
									    int *foreg_last_label
									)
{
  char *proc = "_ForeBackConnectedComponentsExtraction";
  /* sizes 
   */
  unsigned int dimx1 = dimx-1, dimy1 = dimy-1;
  /* offsets 
   */
  int bg_array_offset[3][3][2], bg_offsets[13], bg_nb_offsets;
  int fg_array_offset[3][3][2], fg_offsets[13], fg_nb_offsets;


  int i, j, k, _INSIDEY_;
  unsigned int x, y, z;
  TYPE *theBuf = (TYPE *)(inputBuf);
  TYPE current_value;

  unsigned int label_neighbors[13];
  unsigned int current_label;
  
  unsigned int bg_first_label;
  unsigned int bg_new_label;
  unsigned int fg_first_label;
  unsigned int fg_new_label;
  unsigned int last_label;

  int bg_local_connexite;
  int fg_local_connexite;
  int nb_neighbors;
  /* int valid_labels = 0, total_labels = 0; */

  typeConnectedComponent *cc = env->cc;
  int r;


  /* initialisation de la table des composantes 
     - first label for components will be env->cc_array_size - 1
       these labels are decreasing
     - first label for background is 1
       these labels are increasing
  */
  bg_first_label = 1;
  bg_new_label = bg_first_label - 1;

  fg_first_label = env->cc_array_size - 1;
  fg_new_label = fg_first_label + 1;
 
  
  *backg_first_label = 1;
  *backg_last_label  = 0;
  *foreg_first_label = env->cc_array_size;
  *foreg_last_label  = env->cc_array_size - 1;

  /*
   * Computation of a point's label : current_label
   * If this point has some neighbors (i.e. nb_neighbors > 0) then
   *   1. Its label is the minimum value of all the neighbors labels
   *   2. We report the equivalences between all the labels
   * If not (i.e. nb_neighbors == 0) then
   *   we use a new label (potential case of error : label overflow)
   */
#define _BI_LABEL_COMPUTATION {                                  \
      if ( nb_neighbors > 0 ) {                                  \
	/*                                                       \
	 * If there are some neighbors,                          \
	 * the equivalent label for all of them is the minimum   \
	 * of all labels.                                        \
	 */                                                      \
	current_label = label_neighbors[0];                      \
	for ( i = 1; i < nb_neighbors; i ++ )                    \
	  if ( label_neighbors[i] < current_label )              \
            current_label = label_neighbors[i];                  \
	/*                                                       \
	 * we have to report the equivalences inside the equivalence table \
	 */                                                      \
	for ( i = 0; i < nb_neighbors; i ++ ) {                  \
	  if ( label_neighbors[i] != current_label ) {           \
            k = label_neighbors[i];                              \
            /*                                                   \
             * si la classe n'a pas deja ete traitee             \
             */                                                  \
            if ( cc[ k ].label != current_label ) {              \
	      cc[ current_label ].pointsAboveLowThreshold += cc[ k ].pointsAboveLowThreshold; \
              cc[ k ].pointsAboveLowThreshold = 0;               \
	      cc[ current_label ].pointsAboveHighThreshold += cc[ k ].pointsAboveHighThreshold; \
              cc[ k ].pointsAboveHighThreshold = 0;              \
              cc[ k ].label = current_label;                     \
              if ( current_value == bg_value ) {                 \
	        for ( j = k+1; j <= (int)bg_new_label; j ++ )         \
	          if ( cc[ j ].label == (unsigned int)k ) cc[ j ].label = current_label; \
              } else {                                           \
  	        for ( j = k+1; j <= (int)fg_first_label; j ++ )         \
	          if ( cc[ j ].label == (unsigned int)k ) cc[ j ].label = current_label; \
              }                                                  \
            }                                                    \
	  }                                                      \
	}                                                        \
      } else {                                                   \
	/*                                                       \
	 * No neighbors :                                        \
	 * we use a new label for this point.                    \
	 * This is the only case of error in the whole procedure. \
	 * This could be a little bit improved as follows:       \
	 * 1. we re-label the potential equivalence classes (without conditions of \
	 *    validity, i.e. without testing cc[ i ].pointsAboveHighThreshold and  \
         *    cc[ i ].pointsAboveLowThreshold) to get            \
	 *    the effective number of equivalence classes at this time. \
	 * 2. using this re-labeling, we change the image labels up to the current \
	 *    point.                                             \
	 * 3. we re-build the equivalence table to have only one \
         *    entry per equivalence classe.                      \
	 */                                                      \
        if ( current_value == bg_value ) current_label = ++bg_new_label; \
        else                             current_label = --fg_new_label; \
	if ( bg_new_label >= fg_new_label ) {    \
          /* we try to relabel the part of the image which has already been \
             processed. This way, we use only the minimal number of labels  \
             necessary to represent this part. We hope to recover a         \
             sufficient number of labels to end the computation             \
	   */ \
          if ( current_value == bg_value ) {              \
            last_label = bg_new_label - 1;                \
            r = _LabelsOverflowManagement( inputBuf, dimx, dimy, x, y, z, bg_local_connexite, \
                                            &bg_first_label, &last_label, env, _FROM_SMALLEST_ ); \
            if ( r == 0 ) {                               \
              if ( _VERBOSE_ )                            \
	        fprintf( stderr, "%s: too much used labels for connected components computation\n",  proc ); \
	      return( -1 );                               \
            }                                             \
            bg_new_label  = last_label;                   \
            current_label = ++bg_new_label;               \
            r = _LabelsOverflowManagement( inputBuf, dimx, dimy, x, y, z, fg_local_connexite, \
                                            &fg_new_label, &fg_first_label, env, _FROM_LARGEST_ ); \
          }                                               \
          else {                                          \
            last_label = fg_new_label + 1;                \
            r = _LabelsOverflowManagement( inputBuf, dimx, dimy, x, y, z, fg_local_connexite, \
                                            &last_label, &fg_first_label, env, _FROM_LARGEST_ ); \
            if ( r == 0 ) {                               \
              if ( _VERBOSE_ )                            \
	        fprintf( stderr, "%s: too much used labels for connected components computation\n",  proc ); \
	      return( -1 );                               \
            }                                             \
            fg_new_label = last_label;                    \
            current_label = --fg_new_label;               \
            r = _LabelsOverflowManagement( inputBuf, dimx, dimy, x, y, z, bg_local_connexite, \
                                            &bg_first_label, &bg_new_label, env, _FROM_SMALLEST_ ); \
          }                                               \
          /* does it succeed ?                            \
	   */                                             \
          if ( bg_new_label >= fg_new_label ) {           \
	    /* if not, I give up                          \
	     */                                           \
	    if ( _VERBOSE_ )                              \
	      fprintf( stderr, "%s: too much used labels for connected components computation\n", \
		       proc );                            \
	    return( -1 );                                 \
	  } else {                                        \
	    /* if yes,                                    \
	       we go on with the new table of equivalences \
	       and the new number of used labels          \
	    */                                            \
	    cc = env->cc;                          \
	  }                                               \
        }                                                 \
	cc[ current_label ].label = current_label;        \
	cc[ current_label ].pointsAboveLowThreshold = 0;  \
	cc[ current_label ].pointsAboveHighThreshold = 0; \
      }                                                   \
}

  /*
   * Generic tests of all the neighbors
   * we make sure that they belong to the image.
   * We check if the neighbor has a label (value > 0)
   * We are looking for the equivalence classe (i.e. cc[ value ].label) 
   * of the neighbors
   */
#define _BI_GENERIC_TEST {                \
    if ( current_value == bg_value ) {    \
      for ( i = 0; i < bg_nb_offsets; i ++ ) \
        if ( *(theBuf + bg_offsets[i]) <= bg_new_label ) \
	  label_neighbors[ nb_neighbors++ ] = cc[ (int)( *(theBuf + bg_offsets[i]) ) ].label; \
    }                                     \
    else {                                \
      for ( i = 0; i < fg_nb_offsets; i ++ ) \
        if ( *(theBuf + fg_offsets[i]) >= fg_new_label ) \
	  label_neighbors[ nb_neighbors++ ] = cc[ (int)( *(theBuf + fg_offsets[i]) ) ].label; \
    }                                     \
}

  /*
   * Specific tests for the upper row of the neighborhood of a point
   * We check if the neighbor is connected to the current point
   *          if the neighbor belongs to the image
   *          if the neighbor has a label
   */
#define _BI_UPPER_ROW_TESTS {             \
    if ( y > 0 ) {                        \
      if ( current_value == bg_value ) {  \
        for ( i = -1; i <= 1; i ++ ) {    \
	  if ( bg_array_offset[1+i][0][1] == 0 ) continue;  \
	  if ( (x+i < 0) || (x+i >= dimx) )   continue;  \
	  if ( *(theBuf + bg_array_offset[1+i][0][1]) <= bg_new_label ) \
	    label_neighbors[ nb_neighbors++ ] = cc[ (int)( *(theBuf + bg_array_offset[1+i][0][1]) ) ].label; \
        }                                 \
      }                                   \
      else {                              \
        for ( i = -1; i <= 1; i ++ ) {    \
	  if ( fg_array_offset[1+i][0][1] == 0 ) continue;  \
	  if ( (x+i < 0) || (x+i >= dimx) )   continue;  \
	  if ( *(theBuf + fg_array_offset[1+i][0][1]) >= fg_new_label ) \
	    label_neighbors[ nb_neighbors++ ] = cc[ (int)( *(theBuf + fg_array_offset[1+i][0][1]) ) ].label; \
        }                                 \
      }                                   \
    }                                     \
}

  /*
   * Specific test for the left neighbor of a point
   * We should test if (array_offset[0][1][1] != 0) but we
   * do not because the point (x-1,y,z) always belongs to the
   * neighborhoods we use.
   */
#define _BI_LEFT_NEIGHBOR_TEST {                 \
    if ( x > 0 ) {                               \
      if ( current_value == bg_value ) {         \
        if ( *(theBuf - 1)  <= bg_new_label )    \
	  label_neighbors[ nb_neighbors++ ] = cc[ (int)( *(theBuf - 1) ) ].label; \
      }                                          \
      else {                                     \
        if ( *(theBuf - 1)  >= fg_new_label )    \
	  label_neighbors[ nb_neighbors++ ] = cc[ (int)( *(theBuf - 1) ) ].label; \
      }                                          \
    }                                            \
}

  /*
   * additional informations :
   * - the value of the current label is given to the image buffer
   * - the connect component size is incremented
   * - the validity of the connect component is checked (with respect 
   *   to the current value)
   */
#define _BI_ADDITIONAL_INFORMATIONS {                \
    *theBuf = (TYPE)current_label;     \
    cc[ current_label ].pointsAboveLowThreshold ++;  \
    cc[ current_label ].pointsAboveHighThreshold ++; \
}


  /*
   * Computation of all points.
   *
   * We will attribute to each point a value (a label). According to this
   * value and to the equivalence table, we know the equivalence classe 
   * of the point (i.e. cc[ value ].label).
   * 
   * The computation is divided into two parts : the first slice
   * and the others slices. This allows to avoid some tests and
   * to spare some time.
   *
   * The principle of the computation is the following :
   * For each point, we
   * - check if t may belong to a connected component
   *   => ( current_value >= _low_value_)
   * - check if tests are needed for the neighbors, i.e.
   *   if the point is on the image border. This point
   *   is different for the first slice and the other ones.
   *   first slice  -> if ( y == 0 ) _INSIDEY_ = 0;
   *   other slices -> if ( (y == 0) || (y == dimy1) ) _INSIDEY_ = 0;
   *   => ( _INSIDEY_ == 1 ) && ( x > 0 ) && ( x < dimx1 )
   * - depending on the above checking, 
   *   we compute the number of neighbors (nb_neighbors) and
   *   extract all the neighbors' values (label_neighbors[]),
   *   either without additionnal tests (_GENERIC_TEST),
   *   or with additionnal tests for the neighbors in the upper slice,
   *   in the upper row (_UPPER_ROW_TESTS) or the left neighbor
   *   (_LEFT_NEIGHBOR_TEST).
   *   The two offsets structures (offsets[] and array_offset[][][])
   *   are necessary because of this choice.
   * - Once we have the number of neighbors and their values, we can
   *   compute the point's label (_LABEL_COMPUTATION)
   *   - if the point have neighbors, its label will be the minimum
   *     of all labels, and we report the equivalence between this
   *     minimum label and the other labels in the neighborhood inside
   *     the equivalence table.
   *   - if the point does not have any neighbor, we use a new label.
   *     we increment the number of used label (++new_label) and
   *     we initialize a new label. 
   *     WARNING: an error may occur if such initialization is not possible
   *     (it depends on the type used for the image).
   * - Once the point's label is computed, we report this value in the
   *   image, increment the connected component size and test if 
   *   connected component will be valid (current_value >= _hig_value_)
   */


  /*
   * first slice
   */
  bg_local_connexite = CheckAndEvaluateConnectivity( bg_connectivity, (int)1 );
  fg_local_connexite = CheckAndEvaluateConnectivity( fg_connectivity, (int)1 );

  bg_nb_offsets = InitNeighborsOffsets( bg_array_offset, bg_offsets, dimx, dimy, bg_local_connexite );
  fg_nb_offsets = InitNeighborsOffsets( fg_array_offset, fg_offsets, dimx, dimy, fg_local_connexite );

  z = 0;
  for ( y = 0; y < dimy; y++ ) {
    _INSIDEY_ = 1;
    if ( y == 0 ) _INSIDEY_ = 0;
    for ( x = 0; x < dimx; x++, theBuf++ ) {
      current_value = *theBuf;
      nb_neighbors = 0;
      if ( ( _INSIDEY_ == 1 ) && ( x > 0 ) && ( x < dimx1 ) ) {
	_BI_GENERIC_TEST
      } else {
	_BI_UPPER_ROW_TESTS
	_BI_LEFT_NEIGHBOR_TEST
      }
      _BI_LABEL_COMPUTATION
      _BI_ADDITIONAL_INFORMATIONS
    }
  }

  /*
   * other slices
   */
  bg_local_connexite = CheckAndEvaluateConnectivity( bg_connectivity, dimz );
  fg_local_connexite = CheckAndEvaluateConnectivity( fg_connectivity, dimz );

  bg_nb_offsets = InitNeighborsOffsets( bg_array_offset, bg_offsets, dimx, dimy, bg_local_connexite );
  fg_nb_offsets = InitNeighborsOffsets( fg_array_offset, fg_offsets, dimx, dimy, fg_local_connexite );

  for ( z = 1; z < dimz; z ++ ) {
    for ( y = 0; y < dimy; y++ ) {
      _INSIDEY_ = 1;
      if ( (y == 0) || (y == dimy1) ) _INSIDEY_ = 0;
      for ( x = 0; x < dimx; x++, theBuf++ ) {
	current_value = *theBuf;
	nb_neighbors = 0;
	if ( ( _INSIDEY_ == 1 ) && ( x > 0 ) && ( x < dimx1 ) ) {
	  _BI_GENERIC_TEST
	} else {
	  /*
	   * upper slice tests
	   */
	  if ( current_value == bg_value ) {
	    for ( j = -1; j <= 1; j ++ ) {
	      if ( (y+j < 0) || (y+j >= dimy) )       continue;
	      for ( i = -1; i <= 1; i ++ ) {
		if ( (x+i < 0) || (x+i >= dimx) )     continue;
		if ( bg_array_offset[1+i][1+j][0] == 0 ) continue;
		if ( *(theBuf + bg_array_offset[1+i][1+j][0]) <= bg_new_label ) 
		  label_neighbors[ nb_neighbors++ ] = cc[ (int)( *(theBuf + bg_array_offset[1+i][1+j][0]) ) ].label;
	      }
	    }
	  }
	  else {
	    for ( j = -1; j <= 1; j ++ ) {
	      if ( (y+j < 0) || (y+j >= dimy) )       continue;
	      for ( i = -1; i <= 1; i ++ ) {
		if ( (x+i < 0) || (x+i >= dimx) )     continue;
		if ( fg_array_offset[1+i][1+j][0] == 0 ) continue;
		if ( *(theBuf + fg_array_offset[1+i][1+j][0]) >= fg_new_label ) 
		  label_neighbors[ nb_neighbors++ ] = cc[ (int)( *(theBuf + fg_array_offset[1+i][1+j][0]) ) ].label;
	      }
	    }
	  }
	  _BI_UPPER_ROW_TESTS
	  _BI_LEFT_NEIGHBOR_TEST
	}
	_BI_LABEL_COMPUTATION
	_BI_ADDITIONAL_INFORMATIONS
      }
    }
  }



  /*
   * At this point, all the image's points have been processed.
   */

  if ( _VERBOSE_ ) {
    fprintf( stderr, "%s: used labels: [%5d %5d] - [%5d %5d]\n", proc, 
	     bg_first_label, bg_new_label, fg_new_label, fg_first_label );
  }


  *backg_first_label = (int)bg_first_label;
  *backg_last_label  = (int)bg_new_label;
  *foreg_first_label = (int)fg_new_label;
  *foreg_last_label  = (int)fg_first_label;


  return( 1 );
}








































/*
 * RETURNED VALUES:
 * -1 : some error occurred
 *  0 : no components were found
 * else the number of valid connected components
 */
static int _ConnectedComponentsSelection( typeConnectedComponentsManagement *env,
					  unsigned int first_label,
					  unsigned int last_label,
					  int binary_label )
{
  char *proc = "_ConnectedComponentsSelection";
  unsigned int i, j;
  int total_labels, max_label;
  unsigned int valid_labels;
  typeConnectedComponent *valid_components = (typeConnectedComponent *)NULL;
  /* no components
   */
  if ( last_label < first_label ) return( 0 );



  /* specific case 
     only the largest connected component is required
     
     we look for it.
     if there is any, its label will be 'first_label'
  */
  if ( env->max_number_of_components == 1 ) {

    total_labels = 0;
    max_label = -1;
    
    for ( i = first_label; i <= last_label; i++ ) {
      if ( env->cc[ i ].label == i ) {
	total_labels ++;
	if ( (env->cc[ i ].pointsAboveHighThreshold >= env->minNumberOfPtsAboveHigh)
	     && (env->cc[ i ].pointsAboveLowThreshold >= env->minNumberOfPtsAboveLow) ) {
	  if ( max_label == -1 ) max_label = i;
	  else if ( env->cc[ max_label ].pointsAboveLowThreshold < env->cc[ i ].pointsAboveLowThreshold )
	    max_label = i;
	}
      }
    }
    
    if ( _VERBOSE_ ) {
      fprintf( stderr, "%s: number of valid connected components: ", proc  );
      if ( max_label > 0 ) fprintf( stderr, "1" );
      else fprintf( stderr, "0" );
      fprintf( stderr, " (out of %d)\n", total_labels );
    }

    if ( max_label > 0 ) {
      for ( i = first_label; i <= last_label; i++ ) {
	if ( env->cc[ i ].label == (unsigned int) max_label ) {
	  env->cc[ i ].label = ( env->output == COMPONENT_BINARY ) ? binary_label : first_label;
	}
	else env->cc[ i ].label = env->invalid_label;
      }
      return( 1 );
    }
    return( 0 );

  }
  







  /*
   * counting and re-labeling all the valid connected components
   * this is not optimal, but more elegant because it may be used
   * in several further cases.
   *
   * The representative label of a equivalence's classe
   * (i.e. a connected component) is characterized by
   * ( cc[ i ].label == i ).
   * We have to check if this connected component is valid i.e.
   * - if it contains some points with (value >= _hig_value_),
   *   => (cc[ i ].pointsAboveHighThreshold == minNumberOfPtsAboveHigh )
   * - if its size is large enough
   *   => (cc[ i ].pointsAboveLowThreshold >= minNumberOfPtsAboveLow) )
   * if yes, we increment the number of valid connected components
   * (++valid_labels) and give this value as a new label for the
   * connected component. 
   * if no, we give 0 as a new label and as a new size.
   *
   * If a label is not representative of its equivalence's classe
   * (i.e. cc[ i ].label < i), we give to it the new value
   * of its representative label (which may be 0 if the connected
   * component is not valid). Recall that the representative label
   * of an equivalence's classe is always parsed before the other
   * labels of the same equivalence's classe. Id. for the size.
   *
   * To keep recognizing the valid classes, the representative labels
   * have the numbers of points, while the others not.
   */
  
  total_labels = 0;
  valid_labels = first_label-1;

  for ( i = first_label; i <= last_label; i++ ) {
    if ( env->cc[ i ].label == i ) {
      total_labels ++;
      if ( (env->cc[ i ].pointsAboveHighThreshold >= env->minNumberOfPtsAboveHigh)
	   && (env->cc[ i ].pointsAboveLowThreshold >= env->minNumberOfPtsAboveLow) )
	env->cc[ i ].label = ++valid_labels; 
      else {
	env->cc[ i ].label = env->invalid_label;
	env->cc[ i ].pointsAboveLowThreshold = 0; 
	env->cc[ i ].pointsAboveHighThreshold = 0; 
      }
    }
    else {
      j = env->cc[ i ].label; 
      env->cc[ i ].label = env->cc[ j ].label;
    }
  }






  /*
   * At this point, all valid components have been renumbered,
   * and have the right size.
   */

  if ( _VERBOSE_ ) {
    fprintf( stderr, "%s: number of valid connected components: %5d (out of %d)\n", 
	     proc, valid_labels, total_labels );
  }



  




  /*
   * output: all the valid connected components if specified
   */
  if ( (env->max_number_of_components <= 0) || 
       (env->max_number_of_components >= env->cc_array_size-1) ) {
    if ( env->output == COMPONENT_BINARY ) {
      for ( i = first_label; i <= last_label; i++ ) {
	if ( env->cc[ i ].label != (unsigned int)env->invalid_label )
	  env->cc[ i ].label = binary_label;
      }
    }
    return( valid_labels-first_label+1 );
  }






  /* number of valid components
   */
  valid_labels = valid_labels-first_label+1;

  /*
   * output: the n largest connected component
   *         env->max_number_of_components > 1 
   * we will sort the valid connected components (# = valid_labels)
   * with respect to their size
   */
  if ( env->max_number_of_components > 1 ) {
    valid_components = (typeConnectedComponent *)malloc((valid_labels + 1) * sizeof(typeConnectedComponent));
    if ( valid_components == (typeConnectedComponent*)NULL ) {
      if ( _VERBOSE_ ) 
	fprintf( stderr, "%s: allocation failed for auxiliary array (to sort connected components)\n", proc );
      return( -1 );
    }

    /* store all the valid connected components in an array 
       A valid connected component is recognized by
       (cc[ i ].pointsAboveHighThreshold >= minNumberOfPtsAboveHigh)
       or (cc[ i ].pointsAboveLowThreshold >= minNumberOfPtsAboveLow)

       All attributed labels are in [1...valid_labels]
     */
    env->cc[ 0 ].label = env->cc[ 0 ].pointsAboveLowThreshold = env->cc[ 0 ].pointsAboveHighThreshold = 0;
    j = 0;
    for ( i = first_label; i <= last_label; i++ ) {
      if ( (env->cc[ i ].pointsAboveHighThreshold >= env->minNumberOfPtsAboveHigh)
	   && (env->cc[ i ].pointsAboveLowThreshold >= env->minNumberOfPtsAboveLow) )
	valid_components[ ++j ] = env->cc[i];
    }
    
    if ( j != valid_labels ) {
      if ( _VERBOSE_ ) 
	fprintf( stderr, "%s: something goes wrong\n", proc );
      free( valid_components );
    }

    /* sort them */
    qsort( &(valid_components[1]), valid_labels, sizeof(typeConnectedComponent), &compareConnectedComponents );


    /* At this point, we want to keep the connected
       components represented by valid_components[i].label
       for i=1...min(maxNumberOfConnectedComponent,valid_labels).
     
       and to reject the connected components represented by 
       valid_components[i].label
       for i=min(maxNumberOfConnectedComponent,valid_labels)+1 ... valid_labels
       
       the labels (valid_components[i].label) are in the range 1...valid_labels

       If (valid_labels <= maxNumberOfConnectedComponent) all
       connected components are kept, but they are sorted
       with respect to size.
    */
    
    /* we mark by 
       cc[ i ].pointsAboveHighThreshold = 'index after sort'
       the connected components to be kept
       and by
       cc[ j ].pointsAboveHighThreshold = 0
       the connected components to be removed

    */
    if ( env->output == COMPONENT_BINARY ) {
      for ( i = 1; ((int)i <= env->max_number_of_components) && (i <= valid_labels); i ++ )
	valid_components[ valid_components[i].label ].pointsAboveHighThreshold = first_label;
    }
    else {
      for ( i = 1; ((int)i <= env->max_number_of_components) && (i <= valid_labels); i ++ )
	valid_components[ valid_components[i].label ].pointsAboveHighThreshold = i;
    }

    /* if maxNumberOfConnectedComponent >= valid_labels, 
       the following does not matter 
    */
    for ( i = env->max_number_of_components + 1; i <= valid_labels; i++ )
      valid_components[ valid_components[i].label ].pointsAboveHighThreshold = env->invalid_label;
    

    /* here, for i=1...valid_labels
       if ( valid_components[ i ].pointsAboveHighThreshold > 0 )
       the connected component is to be kept 
       and its index should be 
       valid_components[ i ].pointsAboveHighThreshold (its index after sort)
       instead of valid_components[ i ].label
    */
    for ( i = 1; i <= valid_labels; i++ )
      valid_components[ i ].label = valid_components[ i ].pointsAboveHighThreshold;

    
    /* report the change of value in the complete array */
    for ( i = first_label; i <= last_label; i++ ) {
      if ( env->cc[i].label == (unsigned int)env->invalid_label ) continue;
      j =  env->cc[i].label;
      env->cc[i].label = valid_components[ j ].label;
    }
    
    free( valid_components );
    
    /* nb de composantes valides 
     */
    j = (valid_labels > (unsigned int)env->max_number_of_components) ? env->max_number_of_components : valid_labels ;
    return( j );
  }
  

  /* this case should never occur
   */
  return( -1 );
}






static int _ConnectedComponentsSelection( typeConnectedComponentsManagement *env,
					  int *bg_first_label,
					  int *bg_last_label,
					  int *fg_first_label,
					  int *fg_last_label )
{
  char *proc = "_ConnectedComponentsSelection";
  int i, j;
  int bg_total_labels;
  int fg_total_labels;
  int label = *bg_first_label - 1;

  bg_total_labels = 0;
  for ( i = *bg_first_label; i <= *bg_last_label; i++ ) {
    if ( (int)env->cc[ i ].label == i ) {
      bg_total_labels ++;
      env->cc[ i ].label = ++label;
    }
    else {
     j = env->cc[ i ].label; 
     env->cc[ i ].label = env->cc[ j ].label; 
    }
  }

  if ( _VERBOSE_ ) {
    fprintf( stderr, "%s: background: %d valid connected components out of %d\n",
	     proc, bg_total_labels, *bg_last_label-*bg_first_label+1 );
  }

  *bg_last_label = label;

  fg_total_labels = 0;
  for ( i = *fg_first_label; i <= *fg_last_label; i++ ) {
    if ( (int)env->cc[ i ].label == i ) {
      fg_total_labels ++;
      env->cc[ i ].label = ++label;
    }
    else {
     j = env->cc[ i ].label; 
     env->cc[ i ].label = env->cc[ j ].label; 
    }
  }

  if ( _VERBOSE_ ) {
    fprintf( stderr, "                              " );
    fprintf( stderr, " foreground: %d valid connected components out of %d\n",
	     fg_total_labels, *fg_last_label-*fg_first_label+1 );
  }
  *fg_first_label = *bg_last_label + 1;
  *fg_last_label  = label;


  return( bg_total_labels + fg_total_labels );
}






template <class TYPE > static void _ConnectedComponentsRelabeling( TYPE *inputBuf,
								  unsigned int dimx, 
								  unsigned int dimy,
								  unsigned int dimz,
								  typeConnectedComponentsManagement *env )
{
  TYPE *theBuf = inputBuf;
  unsigned int i, v = dimx*dimy*dimz;

  for (i=0; i<v; i++) 
    theBuf[i] = (TYPE)env->cc[ (int)theBuf[i] ].label;
  
}









template <class TYPE > static void _reallyFillContours( TYPE *theBuf,
							unsigned int dimx, 
							unsigned int dimy,
							unsigned int dimz,
							unsigned int bg_first,
							unsigned int bg_last,
							unsigned int fg_first,
							unsigned int fg_last,
							unsigned int background_label,
							unsigned int foreground_label,
							bool threeD )
{
  int *mat = new int[(bg_last-bg_first+1) * (fg_last-fg_first+1)];
  int *labels = new int[ fg_last+1 ];
  int *bg_frontiers = new int[ (bg_last-bg_first+1) ];
  int *fg_frontiers = new int[ (fg_last-fg_first+1) ];

  unsigned int dimbg = (bg_last-bg_first+1);
  unsigned int dimfg = (fg_last-fg_first+1);
  unsigned int b, f, l;
  unsigned int u, x, y, z;
  unsigned int dimxy = dimx*dimy;

  int nb_bg, nb_fg;
  int mark_bg_as_fg;

  if ( _VERBOSE_ ) {
    fprintf( stderr, "_reallyFillContours: background labels are [%5d - %5d]\n",
	     bg_first, bg_last );
    fprintf( stderr, "                     foreground labels are [%5d - %5d]\n",
	     fg_first, fg_last );
  }
  

  for ( l=bg_first; l<=bg_last; l++ ) labels[l] = background_label;
  for ( l=fg_first; l<=fg_last; l++ ) labels[l] = foreground_label;

  /* if there are background AND foreground components
     we must identify the adjacency relationship
     and relabel the components
  */
  if ( bg_first <= bg_last && fg_first <= fg_last ) {

    for ( b = 0; b < dimbg; b ++ )
    for ( f = 0; f < dimfg; f ++ )
      mat[b*dimfg + f] = 0;

    if ( threeD == true ) {

      for ( u=0, z=0; z<dimz-1; z++ )
      for ( y=0; y<dimy; y++ )
      for ( x=0; x<dimx; x++, u++ ) {
	if ( x < dimx-1 ) {
	  if ( bg_first <= theBuf[ u   ] && theBuf[ u   ] <= bg_last &&
	       fg_first <= theBuf[ u+1 ] && theBuf[ u+1 ] <= fg_last ) {
	    mat[(theBuf[ u ]-bg_first)*dimfg + theBuf[ u+1 ]-fg_first ] = 1;
	  }
	  if ( fg_first <= theBuf[ u   ] && theBuf[ u   ] <= fg_last &&
	       bg_first <= theBuf[ u+1 ] && theBuf[ u+1 ] <= bg_last ) {
	    mat[(theBuf[ u+1 ]-bg_first)*dimfg + theBuf[ u ]-fg_first ] = 1;
	  }
	}
	if ( y < dimy-1 ) {
	  if ( bg_first <= theBuf[ u   ] && theBuf[ u   ] <= bg_last &&
	       fg_first <= theBuf[ u+dimx ] && theBuf[ u+dimx ] <= fg_last ) {
	    mat[(theBuf[ u ]-bg_first)*dimfg + theBuf[ u+dimx ]-fg_first ] = 1;
	  }
	  if ( fg_first <= theBuf[ u   ] && theBuf[ u   ] <= fg_last &&
	       bg_first <= theBuf[ u+dimx ] && theBuf[ u+dimx ] <= bg_last ) {
	    mat[(theBuf[ u+dimx ]-bg_first)*dimfg + theBuf[ u ]-fg_first ] = 1;
	  }
	}
	if ( bg_first <= theBuf[ u   ] && theBuf[ u   ] <= bg_last &&
	     fg_first <= theBuf[ u+dimxy ] && theBuf[ u+dimxy ] <= fg_last ) {
	  mat[(theBuf[ u ]-bg_first)*dimfg + theBuf[ u+dimxy ]-fg_first ] = 1;
	}
	if ( fg_first <= theBuf[ u   ] && theBuf[ u   ] <= fg_last &&
	     bg_first <= theBuf[ u+dimxy ] && theBuf[ u+dimxy ] <= bg_last ) {
	  mat[(theBuf[ u+dimxy ]-bg_first)*dimfg + theBuf[ u ]-fg_first ] = 1;
	}
      }
      /* last slice
       */
      for ( y=0; y<dimy; y++ )
      for ( x=0; x<dimx; x++, u++ ) {
	if ( x < dimx-1 ) {
	  if ( bg_first <= theBuf[ u   ] && theBuf[ u   ] <= bg_last &&
	       fg_first <= theBuf[ u+1 ] && theBuf[ u+1 ] <= fg_last ) {
	    mat[(theBuf[ u ]-bg_first)*dimfg + theBuf[ u+1 ]-fg_first ] = 1;
	  }
	  if ( fg_first <= theBuf[ u   ] && theBuf[ u   ] <= fg_last &&
	       bg_first <= theBuf[ u+1 ] && theBuf[ u+1 ] <= bg_last ) {
	    mat[(theBuf[ u+1 ]-bg_first)*dimfg + theBuf[ u ]-fg_first ] = 1;
	  }
	}
	if ( y < dimy-1 ) {
	  if ( bg_first <= theBuf[ u   ] && theBuf[ u   ] <= bg_last &&
	       fg_first <= theBuf[ u+dimx ] && theBuf[ u+dimx ] <= fg_last ) {
	    mat[(theBuf[ u ]-bg_first)*dimfg + theBuf[ u+dimx ]-fg_first ] = 1;
	  }
	  if ( fg_first <= theBuf[ u   ] && theBuf[ u   ] <= fg_last &&
	       bg_first <= theBuf[ u+dimx ] && theBuf[ u+dimx ] <= bg_last ) {
	    mat[(theBuf[ u+dimx ]-bg_first)*dimfg + theBuf[ u ]-fg_first ] = 1;
	  }
	}
      }
      
    }
    else {
      
      for ( u=0, z=0; z<dimz; z++ )
      for ( y=0; y<dimy; y++ )
      for ( x=0; x<dimx; x++, u++ ) {
	if ( x < dimx-1 ) {
	  if ( bg_first <= theBuf[ u   ] && theBuf[ u   ] <= bg_last &&
	       fg_first <= theBuf[ u+1 ] && theBuf[ u+1 ] <= fg_last ) {
	    mat[(theBuf[ u ]-bg_first)*dimfg + theBuf[ u+1 ]-fg_first ] = 1;
	  }
	  if ( fg_first <= theBuf[ u   ] && theBuf[ u   ] <= fg_last &&
	       bg_first <= theBuf[ u+1 ] && theBuf[ u+1 ] <= bg_last ) {
	    mat[(theBuf[ u+1 ]-bg_first)*dimfg + theBuf[ u ]-fg_first ] = 1;
	  }
	}
	if ( y < dimy-1 ) {
	  if ( bg_first <= theBuf[ u   ] && theBuf[ u   ] <= bg_last &&
	       fg_first <= theBuf[ u+dimx ] && theBuf[ u+dimx ] <= fg_last ) {
	    mat[(theBuf[ u ]-bg_first)*dimfg + theBuf[ u+dimx ]-fg_first ] = 1;
	  }
	  if ( fg_first <= theBuf[ u   ] && theBuf[ u   ] <= fg_last &&
	       bg_first <= theBuf[ u+dimx ] && theBuf[ u+dimx ] <= bg_last ) {
	    mat[(theBuf[ u+dimx ]-bg_first)*dimfg + theBuf[ u ]-fg_first ] = 1;
	  }
	}
      }
      
    }


    /* find all border components
     */
    for ( b = 0; b < dimbg; b ++ ) bg_frontiers[b] = 0;

    for ( z=0; z<dimz; z++ ) {
      for ( x=0; x<dimx; x++ ) {
	u = z*dimxy + x;
	if ( bg_first <= theBuf[ u ] && theBuf[ u ] <= bg_last ) bg_frontiers[ theBuf[ u ]-bg_first ] = 1;
	u = z*dimxy + (dimy-1)*dimx + x;
	if ( bg_first <= theBuf[ u ] && theBuf[ u ] <= bg_last ) bg_frontiers[ theBuf[ u ]-bg_first ] = 1;
      }
      for ( y=1; y<dimy-1; y++ ) {
	u = z*dimxy + y*dimx;
	if ( bg_first <= theBuf[ u ] && theBuf[ u ] <= bg_last ) bg_frontiers[ theBuf[ u ]-bg_first ] = 1;
	u = z*dimxy + y*dimx + dimx-1;
	if ( bg_first <= theBuf[ u ] && theBuf[ u ] <= bg_last ) bg_frontiers[ theBuf[ u ]-bg_first ] = 1;
      }
    }
    if ( threeD == true && dimz > 1 ) {
      for ( u=0, y=0; y<dimy; y++ )
      for ( x=0; x<dimx; x++, u++ )
	if ( bg_first <= theBuf[ u ] && theBuf[ u ] <= bg_last ) bg_frontiers[ theBuf[ u ]-bg_first ] = 1;
      u = (dimz-1)*dimxy;
      for ( y=0; y<dimy; y++ )
      for ( x=0; x<dimx; x++, u++ )
	if ( bg_first <= theBuf[ u ] && theBuf[ u ] <= bg_last ) bg_frontiers[ theBuf[ u ]-bg_first ] = 1;
    }
  
    nb_fg = 0;
    for ( nb_bg = 0, b = 0; b < dimbg; b ++ ) if ( bg_frontiers[b] > 0 ) nb_bg ++;
    mark_bg_as_fg = 0;

    while( nb_bg > 0 || nb_fg > 0 ) {
      
      /* process bg components
       */
      for ( f = 0; f < dimfg; f ++ )  fg_frontiers[f] = 0;
      
      for ( b = 0; b < dimbg; b ++ ) {
	if ( bg_frontiers[b] == 0 ) continue;
	for ( f = 0; f < dimfg; f ++ ) {
	  if ( mat[b*dimfg + f] == 0 ) continue;
	  fg_frontiers[f] = 1;
	  mat[b*dimfg + f] = 0;
	}
	if ( mark_bg_as_fg > 0 ) labels[ b+bg_first ] = foreground_label;
      }
      
      for ( nb_bg = 0, b = 0; b < dimbg; b ++ ) bg_frontiers[b] = 0;
      for ( nb_fg = 0, f = 0; f < dimfg; f ++ ) if ( fg_frontiers[f] > 0 ) nb_fg ++;
      
      /* process fg components
       */
      for ( f = 0; f < dimfg; f ++ ) {
	if ( fg_frontiers[f] == 0 ) continue;
	for ( b = 0; b < dimbg; b ++ ) {
	  if ( mat[b*dimfg + f] == 0 ) continue;
	  bg_frontiers[b] = 1;
	  mat[b*dimfg + f] = 0;
	}
      }
      
      for ( nb_fg = 0, f = 0; f < dimfg; f ++ ) fg_frontiers[f] = 0;
      for ( nb_bg = 0, b = 0; b < dimbg; b ++ ) if ( bg_frontiers[b] > 0 ) nb_bg ++;
      
      mark_bg_as_fg = ( mark_bg_as_fg == 0 ) ? 1 : 0;
    }
    
  }

  for ( u=0, z=0; z<dimz; z++ )
  for ( y=0; y<dimy; y++ )
  for ( x=0; x<dimx; x++, u++ ) {
    theBuf[ u ] = labels[ theBuf[ u ] ];
  }

  delete[] mat;
  delete[] labels;
  delete[] bg_frontiers;
  delete[] fg_frontiers;
}














template <class TYPE1, class TYPE2, class V1> void seedBufferToBuffer( TYPE1 *theBuf, 
								       TYPE2 *resBuf, 
								       const unsigned int size,
								       const unsigned int bgval,
								       const unsigned int seed,
								       V1 *vbg )
{
  unsigned int i;
  TYPE1 *data = theBuf;
  TYPE2 *res = resBuf;
  TYPE1 bg = vbg->getValue();

  for ( i=0; i<size; i++, data++, res++ ) {
    if ( *data == bg ) continue;
    if ( *res == (TYPE2)bgval ) continue;
    *res = seed;
  }
}





template <class TYPE2> void seedImageToBuffer( TYPE2 *resBuf,
					       unsigned int bgValue,
					       unsigned int seedValue,
					       Inrimage *inr,
					       Voxel *bg )
{
  char *from = "seedImageToBuffer(TYPE2 *, unsigned int, Inrimage *, Voxel *, Voxel * )";
  unsigned int size = inr->getX() * inr->getY() * inr->getZ() * inr->getVdim();
  switch( inr->getType() ) {
  default :
    throw ErrTypeMismatch( from );
    return;
  case Inrimage::WT_UNSIGNED_CHAR:
    seedBufferToBuffer( (unsigned char *)inr->getData(), resBuf, size, bgValue, seedValue,
			dynamic_cast<Voxel8u *>(bg) );
    break;
  case Inrimage::WT_SIGNED_CHAR:
    seedBufferToBuffer( (char *)inr->getData(), resBuf, size, bgValue, seedValue, 
			    dynamic_cast<Voxel8s *>(bg) );
    break;
  case Inrimage::WT_UNSIGNED_SHORT:
    seedBufferToBuffer( (unsigned short int *)inr->getData(), resBuf, size, bgValue, seedValue, 
			    dynamic_cast<Voxel16u *>(bg) );
    break;
  case Inrimage::WT_SIGNED_SHORT:
    seedBufferToBuffer( (short int *)inr->getData(), resBuf, size, bgValue, seedValue, 
			    dynamic_cast<Voxel16s *>(bg) );
    break;
  case Inrimage::WT_UNSIGNED_INT:
    seedBufferToBuffer( (unsigned int *)inr->getData(), resBuf, size, bgValue, seedValue, 
			    dynamic_cast<Voxel32u *>(bg) );
    break;
  case Inrimage::WT_SIGNED_INT:
    seedBufferToBuffer( (int *)inr->getData(), resBuf, size, bgValue, seedValue, 
			    dynamic_cast<Voxel32s *>(bg) );
    break;
  case Inrimage::WT_UNSIGNED_LONG:
    seedBufferToBuffer( (unsigned LONGINT *)inr->getData(), resBuf, size, bgValue, seedValue, 
			    dynamic_cast<Voxel64u *>(bg) );
    break;
  case Inrimage::WT_SIGNED_LONG:
    seedBufferToBuffer( (LONGINT *)inr->getData(), resBuf, size, bgValue, seedValue, 
			    dynamic_cast<Voxel64s *>(bg) );
    break;
  case Inrimage::WT_FLOAT:
     seedBufferToBuffer( (float *)inr->getData(), resBuf, size, bgValue, seedValue, 
			    dynamic_cast<VoxelFloat *>(bg) );
    break;
  case Inrimage::WT_DOUBLE:
    seedBufferToBuffer( (double *)inr->getData(), resBuf, size, bgValue, seedValue, 
			    dynamic_cast<VoxelDouble *>(bg) );
    break;
  }

}
