/*************************************************************************
 * convert.c - conversion between types
 *
 * $Id: _convert.C,v 1.5 2001/10/02 15:02:47 jdlemare Exp $
 *
 * CopyrightINRIA 1999
 *
 * AUTHOR:
 * Gregoire Malandain (greg@sophia.inria.fr)
 * 
 * CREATION DATE: 
 * June, 9 1998
 *
 * ADDITIONS, CHANGES
 *
 * - Tue Feb 22 11:25:39 MET 2000 (G. Malandain)
 *   add in ConvertBuffer():
 *       USHORT to UCHAR
 *       USHORT to SHORT
 *
 */
#include <climits>

namespace yav {

#define _CONVERTI_(R) ( (R) >= 0.0 ? ((int)((R)+0.5)) : ((int)((R)-0.5)) )


template <class TO , class FROM> void ConvertBuffer ( TO *t ,
						      FROM *f   ,
						      void *bufferIn,
						      void *bufferOut, 
						      int bufferLength )
  
{

  char *proc = "ConvertBuffer";
  register int i;
  register TO *toBuf = (TO *)bufferOut;
  register FROM *fromBuf = (FROM *)bufferIn;

  if ( (typeid(*t) == typeid(*f)) && (bufferOut == bufferIn) )
    return;

  if ( bufferLength <= 0 ) {
    CERR << " Fatal error in "<< proc << ": buffer length is negative or zero." << ENDL ; 
    return;
  }

  if ( (bufferIn == (void*)NULL) || (bufferOut == (void*)NULL) ) {
    CERR << " Fatal error in " << proc << ": NULL buffer(s)." << ENDL ; 
    return;
  }
  
  if ( typeid(*t) == typeid(*f) )
    (void)memcpy( bufferOut, bufferIn, bufferLength * sizeof(typeid(*t)) );
  
  else {

    if ( typeid(*t) == typeid(unsigned char) )
      for (i=bufferLength; i>0; i--, fromBuf++, toBuf++ ) {
	if ( *fromBuf < 0.0 ) *toBuf = (TO) 0.0;
	else if ( *fromBuf < UCHAR_MAX ) *toBuf = (TO)  _CONVERTI_ ( *fromBuf );
	else *toBuf = (TO) UCHAR_MAX;
      }

    else if ( typeid(*t) == typeid(unsigned short int) )
      for (i=bufferLength; i>0; i--, fromBuf++, toBuf++ ) {
	if ( *fromBuf < 0.0 ) *toBuf = (TO) 0.0;
	else if ( *fromBuf < USHRT_MAX ) *toBuf = (TO) _CONVERTI_ ( *fromBuf );
	else *toBuf = (TO) USHRT_MAX;
      }
    
    else if ( typeid(*t) == typeid( short int) )
      for (i=bufferLength; i>0; i--, fromBuf++, toBuf++ ) {
	if ( *fromBuf < (FROM) SHRT_MIN ) *toBuf = (TO) SHRT_MIN;
	else if ( *fromBuf < SHRT_MAX ) *toBuf = (TO) _CONVERTI_ ( *fromBuf );
	else *toBuf = (TO) SHRT_MAX;
      }
    
    else if ( typeid(*t) == typeid(unsigned int) )
      for (i=bufferLength; i>0; i--, fromBuf++, toBuf++ ) {
	if ( *fromBuf < 0.0 ) *toBuf = (TO) 0.0;
	else if ( *fromBuf < (FROM) UINT_MAX )  *toBuf = (TO) _CONVERTI_ ( *fromBuf );
	else *toBuf = (TO) UINT_MAX;
      }
    
    else if ( typeid(*t) == typeid( int) )
      for (i=bufferLength; i>0; i--, fromBuf++, toBuf++ ) {
	if ( (float) *fromBuf < (float) INT_MIN ) *toBuf = (TO) INT_MIN;
	else if ( *fromBuf < INT_MAX ) *toBuf = (TO) _CONVERTI_ ( *fromBuf );
	else *toBuf = (TO) INT_MAX; 
      }
    
    else if ( typeid(*t) == typeid(unsigned LONGINT) )
      for (i=bufferLength; i>0; i--, fromBuf++, toBuf++ ) {
	if ( *fromBuf < 0.0 ) *toBuf = (TO) 0.0;
	else if ( *fromBuf < (FROM) ULONG_MAX )  *toBuf = (TO) _CONVERTI_ ( *fromBuf );
	else *toBuf = (TO) ULONG_MAX;
      }
    
    else if ( typeid(*t) == typeid( LONGINT) )
      for (i=bufferLength; i>0; i--, fromBuf++, toBuf++ ) {
	if ( (double) *fromBuf < (double) LONG_MIN ) *toBuf = (TO) LONG_MIN;
	else if ( *fromBuf < (FROM) LONG_MAX ) *toBuf = (TO) _CONVERTI_ ( *fromBuf );
	else *toBuf = (TO) LONG_MAX; 
      }
    
    else if ( (typeid(*t) == typeid(float)) || (typeid(*t) == typeid(double)) )
      for (i=bufferLength; i>0; i--, fromBuf++, toBuf++ )
	*toBuf = (TO) _CONVERTI_ ( *fromBuf );
    
    else 
      CERR << "Le type de l'image de l'image de sortie n'est pas supporte dans " << proc  << ENDL;

  }
  
}

} // namespace yav
