/*************************************************************************
 * cspline.h - Cubic splines
 *
 * $Id: _cspline.h,v 1.4 2001/09/25 18:29:31 jdlemare Exp $
 *
 * CopyrightINRIA 2000
 *
 * AUTHOR:
 * Alexis Roche
 * Gregoire Malandain (greg@sophia.inria.fr)
 * 
 * CREATION DATE: 
 * Thu Oct 12 12:08:12 MET DST 2000
 *
 * ADDITIONS, CHANGES
 *	
 *	
 *	
 *
 */

#ifndef _CSPLINE_H_
#define _CSPLINE_H_

namespace yav {

typedef struct {

  /* dimension du buffer
     theDim[0] -> dimension selon X
     theDim[1] -> dimension selon Y
     theDim[2] -> dimension selon Z
  */
  int theDim[3];
  /* buffer contenant les coefficients
   */
  float *theCoeff;

} typeCSplineCoefficients;



template <class T> typeCSplineCoefficients *ComputeCSplineCoefficients( T *theBuf,
									 int *theDim );


template <class FROM, class TO> int CSplineFilterOnBuffer( FROM *bufferIn,  
							   TO *bufferOut,
							   int *bufferDims,
							   int *derivatives );


template <class TO> int CSplineFilterOnBufferWithCoefficients( typeCSplineCoefficients *theCoeff,
							       TO *bufferOut,
							       int *derivatives );


/* met a zero les elements de la structure
   typeCSplineCoefficients
*/
extern void InitTypeCSplineCoefficients( typeCSplineCoefficients *t );

extern void FreeTypeCSplineCoefficients( typeCSplineCoefficients **t );

extern int Reech3DCSpline4x4WithCoefficients( typeCSplineCoefficients *theCoeff,
					      float *resBuf,  /* result buffer */
					      int *resDim,  /* dimensions of this buffer */
					      double *mat,
					      int slice, int *derivative );

extern int Reech2DCSpline4x4WithCoefficients( typeCSplineCoefficients *theCoeff,
					      float* resBuf,  /* result buffer */
					      int *resDim,  /* dimensions of this buffer */
					      double *mat,
					      int *derivative );

extern int Reech3DCSplineDefWithCoefficients( typeCSplineCoefficients *theCoeff,
					      float *resBuf,  /* result buffer */
					      int *resDim,  /* dimensions of this buffer */
					      float **theDef,
					      double *mat,
					      int slice, int *derivative );

extern int Reech2DCSplineDefWithCoefficients( typeCSplineCoefficients *theCoeff,
					      float* resBuf,  /* result buffer */
					      int *resDim,  /* dimensions of this buffer */
					      float **theDef,
					      double *mat,
					      int *derivative );

template <class FROM, class TO > int ReechCSpline4x4( FROM* theBuf, 
						      int *theDim,
						      TO* resBuf, 
						      int *resDim,
						      double *mat,
						      int *derivative);

template <class FROM, class TO > int ReechCSplineDef( FROM* theBuf, 
						      int *theDim,
						      TO* resBuf, 
						      int *resDim,
						      float** theDef,
						      double *mat,
						      int *derivative );
} // namespace yav

#include <_cspline.C>

#endif
