template <class TYPE > static int _medianFiltering( TYPE* T,
						    TYPE* inputBuf, /* buffer to be resampled */
						    TYPE* resultBuf, /* result buffer */
						    unsigned int dx,
						    unsigned int dy,
						    unsigned int dz,
						    typeStructuringElementPoint *list,
						    int nb )
{
  int resMustBeAllocated = 0;
  TYPE *theBuf = inputBuf;
  TYPE *resBuf = NULL;
  TYPE TMP, *TAB   = NULL;
  int bufferSize;

  int centerBelongsToSE = 0;
  int n;

  int mdx = 0, pdx = 0;
  int mdy = 0, pdy = 0;
  int mdz = 0, pdz = 0;
  int born1x, born2x;
  int born1y, born2y;
  int born1z, born2z;

  int i, x, y, z;
  int dimx = dx;
  int dimy = dy;
  int dimz = dz;

  int _INSIDE_Z_;
  int _INSIDE_Y_;

  int l, med;
  int left, right, last, j, k;
  const int right_inside = nb - 1;

  if ( nb <= 0 || list == NULL ) {
    return( _morpho_ErrorInParameters_ );
  }


  //
  // check whether (0,0,0) belongs to the structuring
  // element. If yes, put it at the beginning
  //
  for ( n=0; n<nb; n++ ) {
    if ( list[n].x == 0 && list[n].y == 0 && list[n].z == 0 ) {
      centerBelongsToSE = 1;
      if ( n != 0 ) {
	list[n].x = list[0].x;
	list[n].y = list[0].y;
	list[n].z = list[0].z;
	list[0].x = list[0].y = list[0].z = 0;
      }
    }
  }
  if ( centerBelongsToSE == 0 ) {
    return( _morpho_ErrorInParameters_ );
  }
  
  //
  // spatial extend of the structuring element
  // offset's computation
  //
  // dx is in the interval [mdx pdx]
  //
  mdx = pdx = list[0].x;
  mdy = pdy = list[0].y;
  mdz = pdz = list[0].z;
  list[0].o = list[0].x + list[0].y * dimx + list[0].z * dimx*dimy;
  for ( n=1; n<nb; n++ ) {
    if ( mdx > list[n].x ) mdx = list[n].x;
    if ( pdx < list[n].x ) pdx = list[n].x;
    if ( mdy > list[n].y ) mdy = list[n].y;
    if ( pdy < list[n].y ) pdy = list[n].y;
    if ( mdz > list[n].z ) mdz = list[n].z;
    if ( pdz < list[n].z ) pdz = list[n].z;
    list[n].o = list[n].x + list[n].y * dimx + list[n].z * dimx*dimy;
  }



  //
  // x >= born1x => x + mdx >= 0 
  // x <  born2x => x + pdx <  dimx
  // both => x coordinates of the SE belong to the image
  // 
  // x >= born3x => x + mdx >= dimx
  // x <  born4x => x + pdx <  0
  // One of these => x coordinates of the SE do not belong to the image
  born1x = (-mdx);   born2x = dimx-pdx;
  born1y = (-mdy);   born2y = dimy-pdy;
  born1z = (-mdz);   born2z = dimz-pdz;
  //
  //
  // 
  if ( born1x < 0    ) born1x = 0;
  if ( born2x > dimx ) born2x = dimx;




  //
  // allocation
  //
  if ( inputBuf == resultBuf ) {
    resMustBeAllocated = 1;
  }

  bufferSize = dimx*dimy*dimz*sizeof( TYPE );
  
  if ( resMustBeAllocated ) {
    TAB = (TYPE*)malloc( bufferSize + nb*sizeof( TYPE ) );
  }
  else {
    TAB = (TYPE*)malloc( nb*sizeof( TYPE ) );
  }
  if ( TAB == NULL ) return( _morpho_ErrMemFull_ );
  if ( resMustBeAllocated ) {
    resBuf  = TAB;
    resBuf += nb;
  }
  else {
    resBuf = resultBuf;
  }

  

  for (i=0, z=0; z<dimz; z++ ) {
    
    _INSIDE_Z_ = 0;
    if ( z >= born1z && z < born2z ) _INSIDE_Z_= 1;

    for ( y=0; y<dimy; y++ ) {
      _INSIDE_Y_ = 0;
      if ( _INSIDE_Z_== 1 && y >= born1y && y < born2y ) _INSIDE_Y_ = 1;



      //
      // beginning of line
      //
      for ( x=0; x<born1x; x++, i++ ) {
	TAB[0] = theBuf[i];
	l = 1;
	if ( _INSIDE_Y_ == 1 ) {
	  for ( n=1; n<nb; n++ ) {
	    if ( x+list[n].x < 0 ) continue;
	    TAB[l++] = theBuf[i+list[n].o];
	  }
	}
	else {
	  for ( n=1; n<nb; n++ ) {
	    if ( x+list[n].x < 0 ) continue;
	    if ( y+list[n].y < 0 ) continue;
	    if ( y+list[n].y >= dimy ) continue;
	    if ( z+list[n].z < 0 ) continue;
	    if ( z+list[n].z >= dimz ) continue;
	    TAB[l++] = theBuf[i+list[n].o];
	  }
	}
	med = l/2;   left = 0;   right = l-1;
	do {
	  // swap left et (left+right)/2 
	  j = (left+right)/2;
	  TMP = TAB[left];   TAB[left] = TAB[j];   TAB[j] = TMP;
	  // cut TAB into two */
	  last = left;
	  for ( k = left+1; k <= right; k++ ) {
	    if ( TAB[k] < TAB[left] ) {
	      last ++;
	      TMP = TAB[k];   TAB[k] = TAB[last];   TAB[last] = TMP;
	    }
	  }
	  // from left+1 to last, all elts are < left
	  // by swapping left and last
	  // we've got [left last-1] < last <= [last+1 right]
	  TMP = TAB[left];   TAB[left] = TAB[last]; TAB[last] = TMP;
	  if ( last >  med ) right = last - 1;
	  else if ( last <  med ) left  = last + 1;
	} while ( last != med );
	resBuf[i] = TAB[med];
      } 
      


      //
      // middle of line
      //
      if ( _INSIDE_Y_ == 1 ) {

	med = nb / 2;

	for ( x=born1x; x<born2x; x++, i++ ) {
	  TAB[0] = theBuf[i];
	  for ( n=1; n<nb; n++ )
	    TAB[n] = theBuf[i+list[n].o];
	 
	  left = 0;   right = right_inside;
	  do {
	    // swap left et (left+right)/2 
	    j = (left+right)/2;
	    TMP = TAB[left];   TAB[left] = TAB[j];   TAB[j] = TMP;
	    // cut TAB into two */
	    last = left;
	    for ( k = left+1; k <= right; k++ ) {
	      if ( TAB[k] < TAB[left] ) {
		last ++;
		TMP = TAB[k];   TAB[k] = TAB[last];   TAB[last] = TMP;
	      }
	    }
	    // from left+1 to last, all elts are < left
	    // by swapping left and last
	    // we've got [left last-1] < last <= [last+1 right]
	    TMP = TAB[left];   TAB[left] = TAB[last]; TAB[last] = TMP;
	    if ( last >  med ) right = last - 1;
	    else if ( last <  med ) left  = last + 1;
	  } while ( last != med );
	  resBuf[i] = TAB[med];	
	}

      }
      else {

	for ( x=born1x; x<born2x; x++, i++ ) {
	  TAB[0] = theBuf[i];
	  l = 1;
	  for ( n=1; n<nb; n++ ) {
	    if ( y+list[n].y < 0 ) continue;
	    if ( y+list[n].y >= dimy ) continue;
	    if ( z+list[n].z < 0 ) continue;
	    if ( z+list[n].z >= dimz ) continue;
	    TAB[l++] = theBuf[i+list[n].o];
	  }

	  med = l/2;   left = 0;   right = l-1;
	  do {
	    // swap left et (left+right)/2 
	    j = (left+right)/2;
	    TMP = TAB[left];   TAB[left] = TAB[j];   TAB[j] = TMP;
	    // cut TAB into two */
	    last = left;
	    for ( k = left+1; k <= right; k++ ) {
	      if ( TAB[k] < TAB[left] ) {
		last ++;
		TMP = TAB[k];   TAB[k] = TAB[last];   TAB[last] = TMP;
	      }
	    }
	    // from left+1 to last, all elts are < left
	    // by swapping left and last
	    // we've got [left last-1] < last <= [last+1 right]
	    TMP = TAB[left];   TAB[left] = TAB[last]; TAB[last] = TMP;
	    if ( last >  med ) right = last - 1;
	    else if ( last <  med ) left  = last + 1;
	  } while ( last != med );
	  resBuf[i] = TAB[med];
	}
	
      }



      //
      // end of line
      //
      for ( x=born2x; x<dimx; x++, i++ ) {
	TAB[0] = theBuf[i];
	l = 1;
	if ( _INSIDE_Y_ == 1 ) {
	  for ( n=1; n<nb; n++ ) {
	    if ( x+list[n].x >= dimx ) continue;
	    TAB[l++] = theBuf[i+list[n].o];
	  }
	}
	else {
	  for ( n=1; n<nb; n++ ) {
	    if ( x+list[n].x >= dimx ) continue;
	    if ( y+list[n].y < 0 ) continue;
	    if ( y+list[n].y >= dimy ) continue;
	    if ( z+list[n].z < 0 ) continue;
	    if ( z+list[n].z >= dimz ) continue;
	    TAB[l++] = theBuf[i+list[n].o];
	  }
	}
	med = l/2;   left = 0;   right = l-1;
	do {
	  // swap left et (left+right)/2 
	  j = (left+right)/2;
	  TMP = TAB[left];   TAB[left] = TAB[j];   TAB[j] = TMP;
	  // cut TAB into two */
	  last = left;
	  for ( k = left+1; k <= right; k++ ) {
	    if ( TAB[k] < TAB[left] ) {
	      last ++;
	      TMP = TAB[k];   TAB[k] = TAB[last];   TAB[last] = TMP;
	    }
	  }
	  // from left+1 to last, all elts are < left
	  // by swapping left and last
	  // we've got [left last-1] < last <= [last+1 right]
	  TMP = TAB[left];   TAB[left] = TAB[last]; TAB[last] = TMP;
	  if ( last >  med ) right = last - 1;
	  else if ( last <  med ) left  = last + 1;
	} while ( last != med );
	resBuf[i] = TAB[med];
      }
    }
  }


  
  if ( resMustBeAllocated ) {
    memcpy( resultBuf, resBuf, bufferSize );
  }
  free( TAB );
  return( 1 );
}
