/* for binary operations only. With respect to grey-level operations,
   it seems that these operations are a little bit faster (a few percents
   on the CPU time), more experiments have to be conducted.
#define _TWO_OPS_( RES, A, B ) RES = ( ((int)(A)) | ((int)(B)) ) 
#define _ONE_OP_( RES, A )     RES = ( ((int)(RES)) | ((int)(A)) ) 
*/
/* for grey-level operations (works for binary operations too) */
/*
#define _TWO_OPS_( RES, A, B ) RES = ( (A) _TEST_ (B) ? (A) : (B) )
#define _ONE_OP_( RES, A )     if ( (A) _TEST_ (RES) ) RES = (A)
*/

/* Morphological operation :



*/

template <class TYPE, class PREDICAT > static int _predef_valued_Operation( TYPE *T, 
									    PREDICAT *P,
									    TYPE* inputBuf, /* buffer to be resampled */
									    TYPE* resultBuf, /* result buffer */
									    unsigned int dimx,
									    unsigned int dimy,
									    unsigned int dimz,
									    int connectivity, /* connectivity to be used */ 
									    int iterations  /* number of iterations */ )
{
  PREDICAT *f = new PREDICAT();
  int conn = 0;                 /* "local" connectivity */
  TYPE *localBuf = NULL; /* buffer for auxiliary computation */
  char *localTab[5];
  void *input;

  int dx, dy, dz, dx1, dy1, dz1;
  int dxy;
  int sliceSize, auxIntSize;
  register int x, y, z;
  int iter;

  if ( iterations <= 0 ) return( _morpho_ErrorInParameters_ );


  /* test on connectivity */
  conn = connectivity;
  switch ( conn ) {
  case 4 :
  case 8 :
  case 6 :
  case 10 :
  case 18 :
  case 26 :
    break ;
  default :
    conn = 26;
  }
  
  /* initialization */
  dx = dimx;   dx1 = dx - 1;
  dy = dimy;   dy1 = dy - 1;
  dz = dimz;   dz1 = dz - 1;
  dxy = dx * dy;

  if ( dz == 1 ) {
    switch ( conn ) {
    case 6 :
      conn = 4; break;
    case 10 :
    case 18 :
    case 26 :
      conn = 8; break;
    }
  }

  /* allocation of the auxiliary buffer 
     be sure that the size of each auxiliary slice is 
     a multiple of sizeof(int) for binary computation
  */
  sliceSize = dx * dy * sizeof(TYPE);
  auxIntSize = 1 + sliceSize / sizeof( int );
  localBuf = (TYPE*)malloc( (unsigned int)(5 * auxIntSize * sizeof(int)) );
  if ( localBuf == NULL ) return( _morpho_ErrMemFull_ );

  {
    char *t;
    for ( x = 0; x < 5; x ++ ) {
      t = (char*)localBuf;
      t += x * auxIntSize * sizeof( int );
      localTab[x] = (char*)t;
    }
  }

  input =  inputBuf;
  /* iterative processing */
  for ( iter = 0; iter < iterations; iter ++ ) {
    switch ( conn ) {

      /* 4-connectivity :
	 there is one pointer for each neighbor.
	 the result is computed by comparison between
	 the current point and its [2|3|4] neighbors.
	 There is no "binary acceleration" at the corners.
       
      */
    case 4 :
      {
	register TYPE *cur = (TYPE*)(input);
	TYPE *resBuf = (TYPE*)(resultBuf);
	void *auxCurr = (void*)localTab[0];
	register TYPE *px = cur, *nx = cur, *py = cur, *ny = cur;
	register TYPE *res;
	py -= dx;   px -= 1; 
	ny += dx;   nx += 1; 

	for ( z = 0; z < dz; z ++ ) {
	  res = (TYPE*)auxCurr;
	  /* first row */
	  *res = ( (*f)( (*cur), (*ny)) ? (*cur) : (*ny) ); /* _TWO_OPS_( *res, *cur, *ny ); */
	  if ( (*f)( (*nx), (*res)) ) *res = (*nx); /* _ONE_OP_( *res, *nx ); */
	  cur++; px++; nx++; ny++; res++;

	  for ( x = 1; x < dx1; x++, cur++, px++, nx++, ny++, res++ ) {
	    *res = ( (*f)( (*cur), (*ny)) ? (*cur) : (*ny) ); /* _TWO_OPS_( *res, *cur, *ny ); */
	    if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	    if ( (*f)( (*nx), (*res)) ) *res = (*nx); /* _ONE_OP_( *res, *nx ); */
	  }

	  *res = ( (*f)( (*cur), (*ny)) ? (*cur) : (*ny) ); /* _TWO_OPS_( *res, *cur, *ny ); */
	  if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	  cur++; px++; nx++; ny++; res++;
	  py += dx;

	  /* middle rows */
	  for ( y = 1; y < dy1; y ++ ) {
	    *res = ( (*f)( (*cur), (*py)) ? (*cur) : (*py) ); /* _TWO_OPS_( *res, *cur, *py ); */
	    if ( (*f)( (*ny), (*res)) ) *res = (*ny); /* _ONE_OP_( *res, *ny ); */
	    if ( (*f)( (*nx), (*res)) ) *res = (*nx); /* _ONE_OP_( *res, *nx ); */
	    cur++; px++; nx++; py++; ny++; res++;
	    for ( x = 1; x < dx1; x++, cur++, px++, nx++, py++, ny++, res++ ) {
	      *res = ( (*f)( (*cur), (*py)) ? (*cur) : (*py) ); /* _TWO_OPS_( *res, *cur, *py ); */
	      if ( (*f)( (*ny), (*res)) ) *res = (*ny); /* _ONE_OP_( *res, *ny ); */
	      if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	      if ( (*f)( (*nx), (*res)) ) *res = (*nx); /* _ONE_OP_( *res, *nx ); */
	    }
	    *res = ( (*f)( (*cur), (*py)) ? (*cur) : (*py) ); /* _TWO_OPS_( *res, *cur, *py ); */
	    if ( (*f)( (*ny), (*res)) ) *res = (*ny); /* _ONE_OP_( *res, *ny ); */
	    if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */

	    cur++; px++; nx++; py++; ny++; res++;
	  }

	  /* last row */
	  *res = ( (*f)( (*cur), (*py)) ? (*cur) : (*py) ); /* _TWO_OPS_( *res, *cur, *py ); */
	  if ( (*f)( (*nx), (*res)) ) *res = (*nx); /* _ONE_OP_( *res, *nx ); */
	  cur++; px++; nx++; py++; res++;

	  for ( x = 1; x < dx1; x++, cur++, px++, nx++, py++, res++ ) {

	    *res = ( (*f)( (*cur), (*py)) ? (*cur) : (*py) ); /* _TWO_OPS_( *res, *cur, *py ); */
	    if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	    if ( (*f)( (*nx), (*res)) ) *res = (*nx); /* _ONE_OP_( *res, *nx ); */
	  }

	  *res = ( (*f)( (*cur), (*py)) ? (*cur) : (*py) ); /* _TWO_OPS_( *res, *cur, *py ); */
	  if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	  cur++; px++; nx++; py++; res++;
	  ny += dx;
	  /* copy */

	  memcpy( (void*)resBuf, auxCurr, sliceSize );
	  resBuf += dxy;
	}
      }
      break;
      
      /* 8-connectivity :
	 we use separability. 
	 Ther is no "binary acceleration" at the end of the 
	 X- or Y-lines
      */
    case 8 :
      {
	register TYPE *cur = (TYPE*)(input);
	TYPE *resBuf = (TYPE*)(resultBuf);
	void *auxTmp = (void*)localTab[0];
	register TYPE *px = cur;
	register TYPE *cy, *py, *ny;
	register TYPE *res;

	px -= 1;
	
	for ( z = 0; z < dz; z ++ ) {
	  /* computing along X */
	  res = (TYPE*)auxTmp;
	  for ( y = 0; y < dy; y ++ ) {
	    *res = ( (*f)( (*cur), (px[2])) ? (*cur) : (px[2]) ); /* _TWO_OPS_( *res, *cur, px[2] ); */
	    cur++; px++; res++;
	    for ( x = 1; x < dx1; x++, cur++, px++, res++ ) {

	      *res = ( (*f)( (*cur), (px[2])) ? (*cur) : (px[2]) ); /* _TWO_OPS_( *res, *cur, px[2] ); */
	      if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	    }
	    *res = ( (*f)( (*cur), (*px)) ? (*cur) : (*px) ); /* _TWO_OPS_( *res, *cur, *px ); */
	    cur++; px++; res++;
	  }
	  /* computing along Y */
	  py = cy = ny = (TYPE*)auxTmp;
	  py -= dx; ny += dx;
	  for ( x = 0; x < dx; x++, resBuf++, cy++, ny++ )
	    *resBuf = ( (*f)( (*cy), (*ny)) ? (*cy) : (*ny) ); /* _TWO_OPS_( *resBuf, *cy, *ny ); */
	  py += dx;
	  for ( y = 1; y < dy1; y ++ ) 
	    for ( x = 0; x < dx; x++, cy++, ny++, py++, resBuf++ ) {

	      *resBuf = ( (*f)( (*cy), (*ny)) ? (*cy) : (*ny) ); /* _TWO_OPS_( *resBuf, *cy, *ny ); */
	      if ( (*f)( (*py), (*resBuf)) ) *resBuf = (*py); /* _ONE_OP_( *resBuf, *py ); */
	    }
	  for ( x = 0; x < dx; x++, resBuf++, cy++, py++ ) 
	    *resBuf = ( (*f)( (*cy), (*py)) ? (*cy) : (*py) ); /* _TWO_OPS_( *resBuf, *cy, *py ); */
	}
      }
      break;
      

      /* 6-connectivity :
	 there is one pointer for each neighbor.
	 the result is computed by comparison between
	 the current point and its [2|3|4] neighbors.
	 There is no "binary acceleration" at the corners.
       
      */
    case 6 :
      {
	register TYPE *cur = (TYPE*)(input);
	TYPE *resBuf = (TYPE*)(resultBuf);
	void *auxCurr = (void*)localTab[0];
	void *auxPrev = (void*)localTab[1];
	void *tmp;
	register TYPE *px = cur, *nx = cur, *py = cur, *ny = cur;
	register TYPE *pz = cur, *nz = cur; 
	register TYPE *res;

	pz -= dxy;   py -= dx;   px -= 1; 
	nz += dxy;   ny += dx;   nx += 1; 

	/* first slice */
	res = (TYPE*)auxCurr;
	/* first slice, first row */
	*res = ( (*f)( (*cur), (*ny)) ? (*cur) : (*ny) ); /* _TWO_OPS_( *res, *cur, *ny ); */
	if ( (*f)( (*nx), (*res)) ) *res = (*nx); /* _ONE_OP_( *res, *nx ); */
	if ( (*f)( (*nz), (*res)) ) *res = (*nz); /* _ONE_OP_( *res, *nz ); */
	cur++; px++; nx++; ny++; nz++; res++;

	for ( x = 1; x < dx1; x++, cur++, px++, nx++, ny++, nz++, res++ ) {

	  *res = ( (*f)( (*cur), (*ny)) ? (*cur) : (*ny) ); /* _TWO_OPS_( *res, *cur, *ny ); */
	  if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	  if ( (*f)( (*nx), (*res)) ) *res = (*nx); /* _ONE_OP_( *res, *nx ); */
	  if ( (*f)( (*nz), (*res)) ) *res = (*nz); /* _ONE_OP_( *res, *nz ); */
	}

	*res = ( (*f)( (*cur), (*ny)) ? (*cur) : (*ny) ); /* _TWO_OPS_( *res, *cur, *ny ); */
	if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	if ( (*f)( (*nz), (*res)) ) *res = (*nz); /* _ONE_OP_( *res, *nz ); */
	cur++; px++; nx++; ny++; nz++; res++;
	py += dx;

	/* first slice, middle rows */
	for ( y = 1; y < dy1; y ++ ) {

	  *res = ( (*f)( (*cur), (*py)) ? (*cur) : (*py) ); /* _TWO_OPS_( *res, *cur, *py ); */
	  if ( (*f)( (*ny), (*res)) ) *res = (*ny); /* _ONE_OP_( *res, *ny ); */
	  if ( (*f)( (*nx), (*res)) ) *res = (*nx); /* _ONE_OP_( *res, *nx ); */
	  if ( (*f)( (*nz), (*res)) ) *res = (*nz); /* _ONE_OP_( *res, *nz ); */

	  cur++; px++; nx++; py++; ny++; nz++; res++;
	  for ( x = 1; x < dx1; x++, cur++, px++, nx++, py++, ny++, nz++, res++ ) {

	    *res = ( (*f)( (*cur), (*py)) ? (*cur) : (*py) ); /* _TWO_OPS_( *res, *cur, *py ); */
	    if ( (*f)( (*ny), (*res)) ) *res = (*ny); /* _ONE_OP_( *res, *ny ); */
	    if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	    if ( (*f)( (*nx), (*res)) ) *res = (*nx); /* _ONE_OP_( *res, *nx ); */
	    if ( (*f)( (*nz), (*res)) ) *res = (*nz); /* _ONE_OP_( *res, *nz ); */
	  }
	  *res = ( (*f)( (*cur), (*py)) ? (*cur) : (*py) ); /* _TWO_OPS_( *res, *cur, *py ); */
	  if ( (*f)( (*ny), (*res)) ) *res = (*ny); /* _ONE_OP_( *res, *ny ); */
	  if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	  if ( (*f)( (*nz), (*res)) ) *res = (*nz); /* _ONE_OP_( *res, *nz ); */

	  cur++; px++; nx++; py++; ny++; nz++; res++;
	}

	/* first slice, last row */
	*res = ( (*f)( (*cur), (*py)) ? (*cur) : (*py) ); /* _TWO_OPS_( *res, *cur, *py ); */
	if ( (*f)( (*nx), (*res)) ) *res = (*nx); /* _ONE_OP_( *res, *nx ); */
	if ( (*f)( (*nz), (*res)) ) *res = (*nz); /* _ONE_OP_( *res, *nz ); */
	cur++; px++; nx++; py++; nz++; res++;

	for ( x = 1; x < dx1; x++, cur++, px++, nx++, py++, nz++, res++ ) {

	  *res = ( (*f)( (*cur), (*py)) ? (*cur) : (*py) ); /* _TWO_OPS_( *res, *cur, *py ); */
	  if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	  if ( (*f)( (*nx), (*res)) ) *res = (*nx); /* _ONE_OP_( *res, *nx ); */
	  if ( (*f)( (*nz), (*res)) ) *res = (*nz); /* _ONE_OP_( *res, *nz ); */
	}

	*res = ( (*f)( (*cur), (*py)) ? (*cur) : (*py) ); /* _TWO_OPS_( *res, *cur, *py ); */
	if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	if ( (*f)( (*nz), (*res)) ) *res = (*nz); /* _ONE_OP_( *res, *nz ); */
	cur++; px++; nx++; py++; nz++; res++;
	ny += dx; pz += dxy;
	tmp = auxPrev;   auxPrev = auxCurr;   auxCurr = tmp;
	
	/* middle slice */
	for ( z = 1; z < dz1; z ++ ) {
	  res = (TYPE*)auxCurr;
	  /* middle slice, first row */

	  *res = ( (*f)( (*cur), (*ny)) ? (*cur) : (*ny) ); /* _TWO_OPS_( *res, *cur, *ny ); */
	  if ( (*f)( (*nx), (*res)) ) *res = (*nx); /* _ONE_OP_( *res, *nx ); */
	  if ( (*f)( (*nz), (*res)) ) *res = (*nz); /* _ONE_OP_( *res, *nz ); */
	  if ( (*f)( (*pz), (*res)) ) *res = (*pz); /* _ONE_OP_( *res, *pz ); */

	  cur++; px++; nx++; ny++; nz++; pz++; res++;
	  for ( x = 1; x < dx1; x++, cur++, px++, nx++, ny++, nz++, pz++, res++ ) {

	    *res = ( (*f)( (*cur), (*ny)) ? (*cur) : (*ny) ); /* _TWO_OPS_( *res, *cur, *ny ); */
	    if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	    if ( (*f)( (*nx), (*res)) ) *res = (*nx); /* _ONE_OP_( *res, *nx ); */
	    if ( (*f)( (*nz), (*res)) ) *res = (*nz); /* _ONE_OP_( *res, *nz ); */
	    if ( (*f)( (*pz), (*res)) ) *res = (*pz); /* _ONE_OP_( *res, *pz ); */
	  }

	  *res = ( (*f)( (*cur), (*ny)) ? (*cur) : (*ny) ); /* _TWO_OPS_( *res, *cur, *ny ); */
	  if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	  if ( (*f)( (*nz), (*res)) ) *res = (*nz); /* _ONE_OP_( *res, *nz ); */
	  if ( (*f)( (*pz), (*res)) ) *res = (*pz); /* _ONE_OP_( *res, *pz ); */

	  cur++; px++; nx++; ny++; nz++; pz++; res++;
	  py += dx;
	  /* middle slice, middle rows */
	  for ( y = 1; y < dy1; y ++ ) {

	    *res = ( (*f)( (*cur), (*py)) ? (*cur) : (*py) ); /* _TWO_OPS_( *res, *cur, *py ); */
	    if ( (*f)( (*ny), (*res)) ) *res = (*ny); /* _ONE_OP_( *res, *ny ); */
	    if ( (*f)( (*nx), (*res)) ) *res = (*nx); /* _ONE_OP_( *res, *nx ); */
	    if ( (*f)( (*nz), (*res)) ) *res = (*nz); /* _ONE_OP_( *res, *nz ); */
	    if ( (*f)( (*pz), (*res)) ) *res = (*pz); /* _ONE_OP_( *res, *pz ); */

	    cur++; px++; nx++; py++; ny++; nz++; pz++; res++;
	    for ( x = 1; x < dx1; x++, cur++, px++, nx++, py++, ny++, nz++, pz++, res++ ) {

	      *res = ( (*f)( (*cur), (*py)) ? (*cur) : (*py) ); /* _TWO_OPS_( *res, *cur, *py ); */
	      if ( (*f)( (*ny), (*res)) ) *res = (*ny); /* _ONE_OP_( *res, *ny ); */
	      if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	      if ( (*f)( (*nx), (*res)) ) *res = (*nx); /* _ONE_OP_( *res, *nx ); */
	      if ( (*f)( (*nz), (*res)) ) *res = (*nz); /* _ONE_OP_( *res, *nz ); */
	      if ( (*f)( (*pz), (*res)) ) *res = (*pz); /* _ONE_OP_( *res, *pz ); */
	    }

	    *res = ( (*f)( (*cur), (*py)) ? (*cur) : (*py) ); /* _TWO_OPS_( *res, *cur, *py ); */
	    if ( (*f)( (*ny), (*res)) ) *res = (*ny); /* _ONE_OP_( *res, *ny ); */
	    if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	    if ( (*f)( (*nz), (*res)) ) *res = (*nz); /* _ONE_OP_( *res, *nz ); */
	    if ( (*f)( (*pz), (*res)) ) *res = (*pz); /* _ONE_OP_( *res, *pz ); */

	    cur++; px++; nx++; py++; ny++; nz++; pz++; res++;
	  }
	  /* middle slice, last row */

	  *res = ( (*f)( (*cur), (*py)) ? (*cur) : (*py) ); /* _TWO_OPS_( *res, *cur, *py ); */
	  if ( (*f)( (*nx), (*res)) ) *res = (*nx); /* _ONE_OP_( *res, *nx ); */
	  if ( (*f)( (*nz), (*res)) ) *res = (*nz); /* _ONE_OP_( *res, *nz ); */
	  if ( (*f)( (*pz), (*res)) ) *res = (*pz); /* _ONE_OP_( *res, *pz ); */

	  cur++; px++; nx++; py++; nz++; pz++; res++;
	  for ( x = 1; x < dx1; x++, cur++, px++, nx++, py++, nz++, pz++, res++ ) {

	    *res = ( (*f)( (*cur), (*py)) ? (*cur) : (*py) ); /* _TWO_OPS_( *res, *cur, *py ); */
	    if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	    if ( (*f)( (*nx), (*res)) ) *res = (*nx); /* _ONE_OP_( *res, *nx ); */
	    if ( (*f)( (*nz), (*res)) ) *res = (*nz); /* _ONE_OP_( *res, *nz ); */
	    if ( (*f)( (*pz), (*res)) ) *res = (*pz); /* _ONE_OP_( *res, *pz ); */
	  }

	  *res = ( (*f)( (*cur), (*py)) ? (*cur) : (*py) ); /* _TWO_OPS_( *res, *cur, *py ); */
	  if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	  if ( (*f)( (*nz), (*res)) ) *res = (*nz); /* _ONE_OP_( *res, *nz ); */
	  if ( (*f)( (*pz), (*res)) ) *res = (*pz); /* _ONE_OP_( *res, *pz ); */

	  cur++; px++; nx++; py++; nz++; pz++; res++;
	  ny += dx;
	  /* copy of previous slice */
	  memcpy( (void*)resBuf, auxPrev, sliceSize );
	  resBuf += dxy;
	  tmp = auxPrev;   auxPrev = auxCurr;   auxCurr = tmp;
	}

	/* last slice */
	res = (TYPE*)auxCurr;
	/* last slice, first row */
	*res = ( (*f)( (*cur), (*ny)) ? (*cur) : (*ny) ); /* _TWO_OPS_( *res, *cur, *ny ); */
	if ( (*f)( (*nx), (*res)) ) *res = (*nx); /* _ONE_OP_( *res, *nx ); */
	if ( (*f)( (*pz), (*res)) ) *res = (*pz); /* _ONE_OP_( *res, *pz ); */
	cur++; px++; nx++; ny++; pz++; res++;

	for ( x = 1; x < dx1; x++, cur++, px++, nx++, ny++, pz++, res++ ) {

	  *res = ( (*f)( (*cur), (*ny)) ? (*cur) : (*ny) ); /* _TWO_OPS_( *res, *cur, *ny ); */
	  if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	  if ( (*f)( (*nx), (*res)) ) *res = (*nx); /* _ONE_OP_( *res, *nx ); */
	  if ( (*f)( (*pz), (*res)) ) *res = (*pz); /* _ONE_OP_( *res, *pz ); */
	}

	*res = ( (*f)( (*cur), (*ny)) ? (*cur) : (*ny) ); /* _TWO_OPS_( *res, *cur, *ny ); */
	if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	if ( (*f)( (*pz), (*res)) ) *res = (*pz); /* _ONE_OP_( *res, *pz ); */
	cur++; px++; nx++; ny++; pz++; res++;
	py += dx;

	/* last slice, middle rows */
	for ( y = 1; y < dy1; y ++ ) {

	  *res = ( (*f)( (*cur), (*py)) ? (*cur) : (*py) ); /* _TWO_OPS_( *res, *cur, *py ); */
	  if ( (*f)( (*ny), (*res)) ) *res = (*ny); /* _ONE_OP_( *res, *ny ); */
	  if ( (*f)( (*nx), (*res)) ) *res = (*nx); /* _ONE_OP_( *res, *nx ); */
	  if ( (*f)( (*pz), (*res)) ) *res = (*pz); /* _ONE_OP_( *res, *pz ); */

	  cur++; px++; nx++; py++; ny++; pz++; res++;
	  for ( x = 1; x < dx1; x++, cur++, px++, nx++, py++, ny++, pz++, res++ ) {

	    *res = ( (*f)( (*cur), (*py)) ? (*cur) : (*py) ); /* _TWO_OPS_( *res, *cur, *py ); */
	    if ( (*f)( (*ny), (*res)) ) *res = (*ny); /* _ONE_OP_( *res, *ny ); */
	    if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	    if ( (*f)( (*nx), (*res)) ) *res = (*nx); /* _ONE_OP_( *res, *nx ); */
	    if ( (*f)( (*pz), (*res)) ) *res = (*pz); /* _ONE_OP_( *res, *pz ); */
	  }

	  *res = ( (*f)( (*cur), (*py)) ? (*cur) : (*py) ); /* _TWO_OPS_( *res, *cur, *py ); */
	  if ( (*f)( (*ny), (*res)) ) *res = (*ny); /* _ONE_OP_( *res, *ny ); */
	  if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	  if ( (*f)( (*pz), (*res)) ) *res = (*pz); /* _ONE_OP_( *res, *pz ); */

	  cur++; px++; nx++; py++; ny++; pz++; res++;
	}

	/* last slice, last row */
	*res = ( (*f)( (*cur), (*py)) ? (*cur) : (*py) ); /* _TWO_OPS_( *res, *cur, *py ); */
	if ( (*f)( (*nx), (*res)) ) *res = (*nx); /* _ONE_OP_( *res, *nx ); */
	if ( (*f)( (*pz), (*res)) ) *res = (*pz); /* _ONE_OP_( *res, *pz ); */
	cur++; px++; nx++; py++; pz++; res++;

	for ( x = 1; x < dx1; x++, cur++, px++, nx++, py++, pz++, res++ ) {

	  *res = ( (*f)( (*cur), (*py)) ? (*cur) : (*py) ); /* _TWO_OPS_( *res, *cur, *py ); */
	  if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	  if ( (*f)( (*nx), (*res)) ) *res = (*nx); /* _ONE_OP_( *res, *nx ); */
	  if ( (*f)( (*pz), (*res)) ) *res = (*pz); /* _ONE_OP_( *res, *pz ); */
	}

	*res = ( (*f)( (*cur), (*py)) ? (*cur) : (*py) ); /* _TWO_OPS_( *res, *cur, *py ); */
	if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	if ( (*f)( (*pz), (*res)) ) *res = (*pz); /* _ONE_OP_( *res, *pz ); */
	cur++; px++; nx++; py++; pz++; res++;

	/* copy of previous slice */
	memcpy( (void*)resBuf, auxPrev, sliceSize );
	resBuf += dxy;
	/* copy of current slice */
	memcpy( (void*)resBuf, auxCurr, sliceSize );
      }
      break;
      
      /* 26-connectivity :
	 we use separability. 
	 Ther is no "binary acceleration" at the end of the 
	 X-, Y-lines
      */
    case 26 :
      /* les performances semblent moindres que celles de l'ancienne version
	 (en niveaux de gris), de l'ordre de 2 %
      */
      {
	register TYPE *cur = (TYPE*)(input);
	TYPE *resBuf = (TYPE*)(resultBuf);
	void *auxTmp = (void*)localTab[0];
	void *auxCurr = (void*)localTab[1];
	void *auxNext = (void*)localTab[2];
	void *auxPrev = (void*)localTab[3];
	void *tmp;
	register TYPE *px = cur;
	register TYPE *cy, *py, *ny;
	register TYPE *cz, *pz, *nz;
	register TYPE *res;

	px -= 1;

	/* first slice: 8-connectivity */
	/* computing along X */
	res = (TYPE*)auxTmp;
	for ( y = 0; y < dy; y ++ ) {
	  *res = ( (*f)( (*cur), (px[2])) ? (*cur) : (px[2]) ); /* _TWO_OPS_( *res, *cur, px[2] ); */
	  cur++; px++; res++;
	  for ( x = 1; x < dx1; x++, cur++, px++, res++ ) {

	    *res = ( (*f)( (*cur), (px[2])) ? (*cur) : (px[2]) ); /* _TWO_OPS_( *res, *cur, px[2] ); */
	    if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	  }
	  *res = ( (*f)( (*cur), (*px)) ? (*cur) : (*px) ); /* _TWO_OPS_( *res, *cur, *px ); */
	  cur++; px++; res++;
	}
	/* computing along Y */
	res = (TYPE*)auxCurr;
	py = cy = ny = (TYPE*)auxTmp;
	py -= dx; ny += dx;
	for ( x = 0; x < dx; x++, res++, cy++, ny++ )
	  *res = ( (*f)( (*cy), (*ny)) ? (*cy) : (*ny) ); /* _TWO_OPS_( *res, *cy, *ny ); */
	py += dx;
	for ( y = 1; y < dy1; y ++ ) 
	  for ( x = 0; x < dx; x++, cy++, ny++, py++, res++ ) {

	    *res = ( (*f)( (*cy), (*ny)) ? (*cy) : (*ny) ); /* _TWO_OPS_( *res, *cy, *ny ); */
	    if ( (*f)( (*py), (*res)) ) *res = (*py); /* _ONE_OP_( *res, *py ); */
	  }
	for ( x = 0; x < dx; x++, res++, cy++, py++ ) 
	  *res = ( (*f)( (*cy), (*py)) ? (*cy) : (*py) ); /* _TWO_OPS_( *res, *cy, *py ); */

	/* second slice: 8-connectivity */
	/* computing along X */
	res = (TYPE*)auxTmp;
	for ( y = 0; y < dy; y ++ ) {
	  *res = ( (*f)( (*cur), (px[2])) ? (*cur) : (px[2]) ); /* _TWO_OPS_( *res, *cur, px[2] ); */
	  cur++; px++; res++;
	  for ( x = 1; x < dx1; x++, cur++, px++, res++ ) {

	    *res = ( (*f)( (*cur), (px[2])) ? (*cur) : (px[2]) ); /* _TWO_OPS_( *res, *cur, px[2] ); */
	    if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	  }
	  *res = ( (*f)( (*cur), (*px)) ? (*cur) : (*px) ); /* _TWO_OPS_( *res, *cur, *px ); */
	  cur++; px++; res++;
	}
	/* computing along Y */
	res = (TYPE*)auxNext;
	py = cy = ny = (TYPE*)auxTmp;
	py -= dx; ny += dx;
	for ( x = 0; x < dx; x++, res++, cy++, ny++ )
	  *res = ( (*f)( (*cy), (*ny)) ? (*cy) : (*ny) ); /* _TWO_OPS_( *res, *cy, *ny ); */
	py += dx;
	for ( y = 1; y < dy1; y ++ ) 
	  for ( x = 0; x < dx; x ++, cy++, ny++, py++, res++ ) {

	    *res = ( (*f)( (*cy), (*ny)) ? (*cy) : (*ny) ); /* _TWO_OPS_( *res, *cy, *ny ); */
	    if ( (*f)( (*py), (*res)) ) *res = (*py); /* _ONE_OP_( *res, *py ); */
	  }
	for ( x = 0; x < dx; x++, res++, cy++, py++ ) 
	  *res = ( (*f)( (*cy), (*py)) ? (*cy) : (*py) ); /* _TWO_OPS_( *res, *cy, *py ); */
	
	/* first slice: result 
	   if we cast slices of auxiliary buffer to be divided
	   by sizeof(int) we may use logical operations
	*/

	cz = (TYPE*)auxCurr;   nz = (TYPE*)auxNext;
	for ( x = 0; x < dxy ; x++, resBuf++, cz++, nz++ )
	  *resBuf = ( (*f)( (*cz), (*nz)) ? (*cz) : (*nz) ); /* _TWO_OPS_( *resBuf, *cz, *nz ); */


	tmp = auxPrev; auxPrev = auxCurr; auxCurr = auxNext; auxNext = tmp;
	
	/* middle slices */
	for ( z = 1; z < dz1; z ++ ) {
	  /* next slice: 8-connectivity */
	  /* computing along X */
	  res = (TYPE*)auxTmp;
	  for ( y = 0; y < dy; y ++ ) {
	    *res = ( (*f)( (*cur), (px[2])) ? (*cur) : (px[2]) ); /* _TWO_OPS_( *res, *cur, px[2] ); */
	    cur++; px++; res++;
	    for ( x = 1; x < dx1; x ++, cur++, px++, res++ ) {

	      *res = ( (*f)( (*cur), (px[2])) ? (*cur) : (px[2]) ); /* _TWO_OPS_( *res, *cur, px[2] ); */
	      if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	    }
	    *res = ( (*f)( (*cur), (*px)) ? (*cur) : (*px) ); /* _TWO_OPS_( *res, *cur, *px ); */
	    cur++; px++; res++;
	  }
	  /* computing along Y */
	  res = (TYPE*)auxNext;
	  py = cy = ny = (TYPE*)auxTmp;
	  py -= dx; ny += dx;
	  for ( x = 0; x < dx; x++, res++, cy++, ny++ )
	    *res = ( (*f)( (*cy), (*ny)) ? (*cy) : (*ny) ); /* _TWO_OPS_( *res, *cy, *ny ); */
	  py += dx;
	  for ( y = 1; y < dy1; y ++ ) 
	    for ( x = 0; x < dx; x++, cy++, ny++, py++, res++ ) {

	      *res = ( (*f)( (*cy), (*ny)) ? (*cy) : (*ny) ); /* _TWO_OPS_( *res, *cy, *ny ); */
	      if ( (*f)( (*py), (*res)) ) *res = (*py); /* _ONE_OP_( *res, *py ); */
	    }
	  for ( x = 0; x < dx; x++, res++, cy++, py++ ) 
	    *res = ( (*f)( (*cy), (*py)) ? (*cy) : (*py) ); /* _TWO_OPS_( *res, *cy, *py ); */
	  
	  pz = (TYPE*)auxPrev;   cz = (TYPE*)auxCurr;
	  nz = (TYPE*)auxNext;
	  for ( x = 0; x < dxy ; x++, resBuf++, pz++, cz++, nz++ ) {
	    *resBuf = ( (*f)( (*cz), (*pz)) ? (*cz) : (*pz) ); /* _TWO_OPS_( *resBuf, *cz, *pz ); */
	    if ( (*f)( (*nz), (*resBuf)) ) *resBuf = (*nz); /* _ONE_OP_( *resBuf, *nz ); */
	  }

	  tmp = auxPrev; auxPrev = auxCurr; auxCurr = auxNext; auxNext = tmp;
	}

	/* last slice */
	pz = (TYPE*)auxPrev;   cz = (TYPE*)auxCurr;
	for ( x = 0; x < dxy ; x++, resBuf++, pz++, cz++ )
	  *resBuf = ( (*f)( (*cz), (*pz)) ? (*cz) : (*pz) ); /* _TWO_OPS_( *resBuf, *cz, *pz ); */

      }
      break;

    case 10 :
      {
	register TYPE *cur = (TYPE*)(input);
	TYPE *resBuf = (TYPE*)(resultBuf);
	void *auxTmp = (void*)localTab[0];
	void *auxCurr = (void*)localTab[1];
	void *auxPrev = (void*)localTab[2];
	void *tmp;
	register TYPE *px = cur;
	register TYPE *cy, *py, *ny;
	register TYPE *pz = cur, *nz = cur;
	register TYPE *res;

	px -= 1;
	pz -= dxy;   nz += dxy;

	/* first slice: 8-connectivity */
	/* computing along X */
	res = (TYPE*)auxTmp;
	for ( y = 0; y < dy; y ++ ) {
	  *res = ( (*f)( (*cur), (px[2])) ? (*cur) : (px[2]) ); /* _TWO_OPS_( *res, *cur, px[2] ); */
	  cur++; px++; res++;
	  for ( x = 1; x < dx1; x ++, cur++, px++, res++ ) {

	    *res = ( (*f)( (*cur), (px[2])) ? (*cur) : (px[2]) ); /* _TWO_OPS_( *res, *cur, px[2] ); */
	    if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	  }
	  *res = ( (*f)( (*cur), (*px)) ? (*cur) : (*px) ); /* _TWO_OPS_( *res, *cur, *px ); */
	  cur++; px++; res++;
	}
	/* computing along Y */
	res = (TYPE*)auxCurr;
	py = cy = ny = (TYPE*)auxTmp;
	py -= dx; ny += dx;
	for ( x = 0; x < dx; x++, res++, cy++, ny++ )
	  *res = ( (*f)( (*cy), (*ny)) ? (*cy) : (*ny) ); /* _TWO_OPS_( *res, *cy, *ny ); */
	py += dx;
	for ( y = 1; y < dy1; y ++ ) 
	  for ( x = 0; x < dx; x++, cy++, ny++, py++, res++ ) {

	    *res = ( (*f)( (*cy), (*ny)) ? (*cy) : (*ny) ); /* _TWO_OPS_( *res, *cy, *ny ); */
	    if ( (*f)( (*py), (*res)) ) *res = (*py); /* _ONE_OP_( *res, *py ); */
	  }
	for ( x = 0; x < dx; x++, res++, cy++, py++ ) 
	  *res = ( (*f)( (*cy), (*py)) ? (*cy) : (*py) ); /* _TWO_OPS_( *res, *cy, *py ); */
	/* computing along Z */
	res = (TYPE*)auxCurr;
	for ( x = 0; x < dxy; x ++, res++, nz++ )
	  if ( (*f)( (*nz), (*res)) ) *res = (*nz); /* _ONE_OP_( *res, *nz ); */
	pz += dxy;
	tmp = auxPrev; auxPrev = auxCurr; auxCurr = tmp;

	/* middle slices */
	for ( z = 1; z < dz1; z ++ ) {
	  /* current slice: 8-connectivity */
	  /* computing along X */
	  res = (TYPE*)auxTmp;
	  for ( y = 0; y < dy; y ++ ) {
	    *res = ( (*f)( (*cur), (px[2])) ? (*cur) : (px[2]) ); /* _TWO_OPS_( *res, *cur, px[2] ); */
	    cur++; px++; res++;
	    for ( x = 1; x < dx1; x++, cur++, px++, res++ ) {

	      *res = ( (*f)( (*cur), (px[2])) ? (*cur) : (px[2]) ); /* _TWO_OPS_( *res, *cur, px[2] ); */
	      if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	    }
	    *res = ( (*f)( (*cur), (*px)) ? (*cur) : (*px) ); /* _TWO_OPS_( *res, *cur, *px ); */
	    cur++; px++; res++;
	  }
	  /* computing along Y */
	  res = (TYPE*)auxCurr;
	  py = cy = ny = (TYPE*)auxTmp;
	  py -= dx; ny += dx;
	  for ( x = 0; x < dx; x++, res++, cy++, ny++ )
	    *res = ( (*f)( (*cy), (*ny)) ? (*cy) : (*ny) ); /* _TWO_OPS_( *res, *cy, *ny ); */
	  py += dx;
	  for ( y = 1; y < dy1; y ++ ) 
	    for ( x = 0; x < dx; x++, cy++, ny++, py++, res++ ) {

	      *res = ( (*f)( (*cy), (*ny)) ? (*cy) : (*ny) ); /* _TWO_OPS_( *res, *cy, *ny ); */
	      if ( (*f)( (*py), (*res)) ) *res = (*py); /* _ONE_OP_( *res, *py ); */
	    }
	  for ( x = 0; x < dx; x++, res++, cy++, py++ ) 
	    *res = ( (*f)( (*cy), (*py)) ? (*cy) : (*py) ); /* _TWO_OPS_( *res, *cy, *py ); */
	  /* computing along Z */
	  res = (TYPE*)auxCurr;
	  for ( x = 0; x < dxy; x ++, res++, pz++, nz++ ) {

	    if ( (*f)( (*pz), (*res)) ) *res = (*pz); /* _ONE_OP_( *res, *pz ); */
	    if ( (*f)( (*nz), (*res)) ) *res = (*nz); /* _ONE_OP_( *res, *nz ); */
	  }
	  /* copy of previous slice */
	  memcpy( (void*)resBuf, auxPrev, sliceSize );
	  resBuf += dxy;
	  tmp = auxPrev; auxPrev = auxCurr; auxCurr = tmp;
	}

	/* last slice: 8-connectivity */
	/* computing along X */
	res = (TYPE*)auxTmp;
	for ( y = 0; y < dy; y ++ ) {
	  *res = ( (*f)( (*cur), (px[2])) ? (*cur) : (px[2]) ); /* _TWO_OPS_( *res, *cur, px[2] ); */
	  cur++; px++; res++;
	  for ( x = 1; x < dx1; x++, cur++, px++, res++ ) {

	    *res = ( (*f)( (*cur), (px[2])) ? (*cur) : (px[2]) ); /* _TWO_OPS_( *res, *cur, px[2] ); */
	    if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	  }
	  *res = ( (*f)( (*cur), (*px)) ? (*cur) : (*px) ); /* _TWO_OPS_( *res, *cur, *px ); */
	  cur++; px++; res++;
	}
	/* computing along Y */
	res = (TYPE*)auxCurr;
	py = cy = ny = (TYPE*)auxTmp;
	py -= dx; ny += dx;
	for ( x = 0; x < dx; x++, res++, cy++, ny++ )
	  *res = ( (*f)( (*cy), (*ny)) ? (*cy) : (*ny) ); /* _TWO_OPS_( *res, *cy, *ny ); */
	py += dx;
	for ( y = 1; y < dy1; y ++ ) 
	  for ( x = 0; x < dx; x++, cy++, ny++, py++, res++ ) {

	    *res = ( (*f)( (*cy), (*ny)) ? (*cy) : (*ny) ); /* _TWO_OPS_( *res, *cy, *ny ); */
	    if ( (*f)( (*py), (*res)) ) *res = (*py); /* _ONE_OP_( *res, *py ); */
	  }
	for ( x = 0; x < dx; x++, res++, cy++, py++ ) 
	  *res = ( (*f)( (*cy), (*py)) ? (*cy) : (*py) ); /* _TWO_OPS_( *res, *cy, *py ); */
	/* computing along Z */
	res = (TYPE*)auxCurr;
	for ( x = 0; x < dxy; x ++, res++, pz++ )
	  if ( (*f)( (*pz), (*res)) ) *res = (*pz); /* _ONE_OP_( *res, *pz ); */
	/* copy of previous slice */
	memcpy( (void*)resBuf, auxPrev, sliceSize );
	resBuf += dxy;
	/* copy of current slice */
	memcpy( (void*)resBuf, auxCurr, sliceSize );
      }
      break;

    case 18 :
      {
	register TYPE *cur = (TYPE*)(input);
	TYPE *resBuf = (TYPE*)(resultBuf);
	void *aux4Prev = (void*)localTab[0];
	void *aux4Curr = (void*)localTab[1];
	void *aux4Next = (void*)localTab[2];
	void *auxCurr = (void*)localTab[3];
	void *auxNext = (void*)localTab[4];
	void *tmp;
	register TYPE *px = cur;
	register TYPE *ppy = cur, *nny = cur;
	register TYPE *cy, *py, *ny;

	register TYPE *cz, *pz, *nz;

	register TYPE *res;
	
	px -= 1;
	ppy -= dx;   nny += dx;

	/* first slice: computing along X */
	res = (TYPE*)aux4Curr;
	for ( y = 0; y < dy; y ++ ) {
	  *res = ( (*f)( (*cur), (px[2])) ? (*cur) : (px[2]) ); /* _TWO_OPS_( *res, *cur, px[2] ); */
	  cur++; px++; res++;
	  for ( x = 1; x < dx1; x++, cur++, px++, res++ ) {

	    *res = ( (*f)( (*cur), (px[2])) ? (*cur) : (px[2]) ); /* _TWO_OPS_( *res, *cur, px[2] ); */
	    if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	  }
	  *res = ( (*f)( (*cur), (*px)) ? (*cur) : (*px) ); /* _TWO_OPS_( *res, *cur, *px ); */
	  cur++; px++; res++;
	}
	/* first slice: 8-connectivity: computing along Y */
	res = (TYPE*)auxCurr;
	py = cy = ny = (TYPE*)aux4Curr;
	py -= dx; ny += dx;
	for ( x = 0; x < dx; x++, res++, cy++, ny++ )
	  *res = ( (*f)( (*cy), (*ny)) ? (*cy) : (*ny) ); /* _TWO_OPS_( *res, *cy, *ny ); */
	py += dx;
	for ( y = 1; y < dy1; y ++ ) 
	  for ( x = 0; x < dx; x++, cy++, ny++, py++, res++ ) {

	    *res = ( (*f)( (*cy), (*ny)) ? (*cy) : (*ny) ); /* _TWO_OPS_( *res, *cy, *ny ); */
	    if ( (*f)( (*py), (*res)) ) *res = (*py); /* _ONE_OP_( *res, *py ); */
	  }
	for ( x = 0; x < dx; x++, res++, cy++, py++ ) 
	  *res = ( (*f)( (*cy), (*py)) ? (*cy) : (*py) ); /* _TWO_OPS_( *res, *cy, *py ); */
	/* first slice: 4-connectivity: computing along Y */
	res = (TYPE*)aux4Curr;
	for ( x = 0; x < dx; x++, res++, nny++ )
	  if ( (*f)( (*nny), (*res)) ) *res = (*nny); /* _ONE_OP_( *res, *nny ); */
	ppy += dx;
	for ( y = 1; y < dy1; y ++ ) 
	  for ( x = 0; x < dx; x++, nny++, ppy++, res++ ) {

	    if ( (*f)( (*nny), (*res)) ) *res = (*nny); /* _ONE_OP_( *res, *nny ); */
	    if ( (*f)( (*ppy), (*res)) ) *res = (*ppy); /* _ONE_OP_( *res, *ppy ); */
	  }
	for ( x = 0; x < dx; x++, res++, ppy++ ) 
	  if ( (*f)( (*ppy), (*res)) ) *res = (*ppy); /* _ONE_OP_( *res, *ppy ); */
	nny += dx;

	/* second slice: computing along X */
	res = (TYPE*)aux4Next;
	for ( y = 0; y < dy; y ++ ) {
	  *res = ( (*f)( (*cur), (px[2])) ? (*cur) : (px[2]) ); /* _TWO_OPS_( *res, *cur, px[2] ); */
	  cur++; px++; res++;
	  for ( x = 1; x < dx1; x++, cur++, px++, res++ ) {

	    *res = ( (*f)( (*cur), (px[2])) ? (*cur) : (px[2]) ); /* _TWO_OPS_( *res, *cur, px[2] ); */
	    if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	  }
	  *res = ( (*f)( (*cur), (*px)) ? (*cur) : (*px) ); /* _TWO_OPS_( *res, *cur, *px ); */
	  cur++; px++; res++;
	}
	/* second slice: 8-connectivity: computing along Y */
	res = (TYPE*)auxNext;
	py = cy = ny = (TYPE*)aux4Next;
	py -= dx; ny += dx;
	for ( x = 0; x < dx; x++, res++, cy++, ny++ )
	  *res = ( (*f)( (*cy), (*ny)) ? (*cy) : (*ny) ); /* _TWO_OPS_( *res, *cy, *ny ); */
	py += dx;
	for ( y = 1; y < dy1; y ++ ) 
	  for ( x = 0; x < dx; x++, cy++, ny++, py++, res++ ) {

	    *res = ( (*f)( (*cy), (*ny)) ? (*cy) : (*ny) ); /* _TWO_OPS_( *res, *cy, *ny ); */
	    if ( (*f)( (*py), (*res)) ) *res = (*py); /* _ONE_OP_( *res, *py ); */
	  }
	for ( x = 0; x < dx; x++, res++, cy++, py++ ) 
	  *res = ( (*f)( (*cy), (*py)) ? (*cy) : (*py) ); /* _TWO_OPS_( *res, *cy, *py ); */
	/* second slice: 4-connectivity: computing along Y */
	res = (TYPE*)aux4Next;
	for ( x = 0; x < dx; x++, res++, nny++ )
	  if ( (*f)( (*nny), (*res)) ) *res = (*nny); /* _ONE_OP_( *res, *nny ); */
	ppy += dx;
	for ( y = 1; y < dy1; y ++ ) 
	  for ( x = 0; x < dx; x++, nny++, ppy++, res++ ) {

	    if ( (*f)( (*nny), (*res)) ) *res = (*nny); /* _ONE_OP_( *res, *nny ); */
	    if ( (*f)( (*ppy), (*res)) ) *res = (*ppy); /* _ONE_OP_( *res, *ppy ); */
	  }
	for ( x = 0; x < dx; x++, res++, ppy++ ) 
	  if ( (*f)( (*ppy), (*res)) ) *res = (*ppy); /* _ONE_OP_( *res, *ppy ); */
	nny += dx;
	
	/* first slice: result */

	cz = (TYPE*)auxCurr;   nz = (TYPE*)aux4Next;
	for ( x = 0; x < dxy ; x++, resBuf++, cz++, nz++ )
	  *resBuf = ( (*f)( (*cz), (*nz)) ? (*cz) : (*nz) ); /* _TWO_OPS_( *resBuf, *cz, *nz ); */

	tmp = aux4Prev; aux4Prev = aux4Curr; aux4Curr = aux4Next; aux4Next = tmp;
	tmp = auxCurr; auxCurr = auxNext; auxNext = tmp;

	/* middle slice */
	for ( z = 1; z < dz1; z ++ ) {
	  /* next slice: computing along X */
	  res = (TYPE*)aux4Next;
	  for ( y = 0; y < dy; y ++ ) {
	    *res = ( (*f)( (*cur), (px[2])) ? (*cur) : (px[2]) ); /* _TWO_OPS_( *res, *cur, px[2] ); */
	    cur++; px++; res++;
	    for ( x = 1; x < dx1; x++, cur++, px++, res++ ) {

	      *res = ( (*f)( (*cur), (px[2])) ? (*cur) : (px[2]) ); /* _TWO_OPS_( *res, *cur, px[2] ); */
	      if ( (*f)( (*px), (*res)) ) *res = (*px); /* _ONE_OP_( *res, *px ); */
	    }
	    *res = ( (*f)( (*cur), (*px)) ? (*cur) : (*px) ); /* _TWO_OPS_( *res, *cur, *px ); */
	    cur++; px++; res++;
	  }
	  /* next slice: 8-connectivity: computing along Y */
	  res = (TYPE*)auxNext;
	  py = cy = ny = (TYPE*)aux4Next;
	  py -= dx; ny += dx;
	  for ( x = 0; x < dx; x++, res++, cy++, ny++ )
	    *res = ( (*f)( (*cy), (*ny)) ? (*cy) : (*ny) ); /* _TWO_OPS_( *res, *cy, *ny ); */
	  py += dx;
	  for ( y = 1; y < dy1; y ++ ) 
	    for ( x = 0; x < dx; x++, cy++, ny++, py++, res++ ) {

	      *res = ( (*f)( (*cy), (*ny)) ? (*cy) : (*ny) ); /* _TWO_OPS_( *res, *cy, *ny ); */
	      if ( (*f)( (*py), (*res)) ) *res = (*py); /* _ONE_OP_( *res, *py ); */
	    }
	  for ( x = 0; x < dx; x++, res++, cy++, py++ ) 
	    *res = ( (*f)( (*cy), (*py)) ? (*cy) : (*py) ); /* _TWO_OPS_( *res, *cy, *py ); */
	  /* next slice: 4-connectivity: computing along Y */
	  res = (TYPE*)aux4Next;
	  for ( x = 0; x < dx; x++, res++, nny++ )
	    if ( (*f)( (*nny), (*res)) ) *res = (*nny); /* _ONE_OP_( *res, *nny ); */
	  ppy += dx;
	  for ( y = 1; y < dy1; y ++ ) 
	    for ( x = 0; x < dx; x++, nny++, ppy++, res++ ) {

	      if ( (*f)( (*nny), (*res)) ) *res = (*nny); /* _ONE_OP_( *res, *nny ); */
	      if ( (*f)( (*ppy), (*res)) ) *res = (*ppy); /* _ONE_OP_( *res, *ppy ); */
	    }
	  for ( x = 0; x < dx; x++, res++, ppy++ ) 
	    if ( (*f)( (*ppy), (*res)) ) *res = (*ppy); /* _ONE_OP_( *res, *ppy ); */
	  nny += dx;
	  
	  /* current slice: result */

	  cz = (TYPE*)auxCurr;   nz = (TYPE*)aux4Next;
	  pz = (TYPE*)aux4Prev;
	  for ( x = 0; x < dxy ; x++, resBuf++, cz++, nz++, pz++ ) {
	    *resBuf = ( (*f)( (*cz), (*pz)) ? (*cz) : (*pz) ); /* _TWO_OPS_( *resBuf, *cz, *pz ); */
	    if ( (*f)( (*nz), (*resBuf)) ) *resBuf = (*nz); /* _ONE_OP_( *resBuf, *nz ); */
	  }

	  tmp = aux4Prev; aux4Prev = aux4Curr; aux4Curr = aux4Next; aux4Next = tmp;
	  tmp = auxCurr; auxCurr = auxNext; auxNext = tmp;
	}
	
	/* last slice: result */

	cz = (TYPE*)auxCurr;   pz = (TYPE*)aux4Prev;
	for ( x = 0; x < dxy ; x++, resBuf++, cz++, pz++ )
	  *resBuf = ( (*f)( (*cz), (*pz)) ? (*cz) : (*pz) ); /* _TWO_OPS_( *resBuf, *cz, *pz ); */

      }
      break;

    }
    input = resultBuf;
  }
  /* end */
  free( (void*)localBuf );

  return( 1 );	
}





//
// operation with a centered SE a length 3 along X
//

template <class TYPE, class PREDICAT, class BINOP > static int _predef_binary_X_Operation( TYPE *T, 
											   PREDICAT *P,
											   BINOP *B,
											   TYPE* inputSlice,
											   TYPE* outputSlice,
											   unsigned int dimx,
											   unsigned int dimy )
{
  PREDICAT *f = new PREDICAT();
  BINOP *b = new BINOP();
  unsigned int lineLength = dimx * sizeof( TYPE );
  unsigned int ilength, i, iy, x, y;
  
  int *l1, *l2;
  int lx;
  
  if ( inputSlice == outputSlice ) return( _morpho_ErrorInParameters_ );

  // we have to compare input:line[y][0] with input:line[y][1]
  // for a length of (dimx-1)
  // then output:line[y][1] with input:line[y][0]
  // for the same length
  
  ilength = ( dimx-1 ) * sizeof( TYPE ) / sizeof( int );
  lx      = ilength * sizeof( int ) / sizeof( TYPE );

  memcpy( outputSlice, inputSlice, dimy * lineLength );

  for ( y = 0; y < dimy; y ++ ) {

    iy = y*dimx;

    l2 = (int*)( &outputSlice[iy] );
    l1 = (int*)( &inputSlice [iy+1] );
    for ( i=0; i<ilength; i++ ) *l2++ = (*b)(*l2,*l1++);
    for ( x=lx; x<dimx-1; x++ ) {
      if ( (*f)(inputSlice[iy+x+1], outputSlice[iy+x]) )
	outputSlice[iy+x] = inputSlice[iy+x+1];
    }
    
    l2 = (int*)( &outputSlice[iy+1] );
    l1 = (int*)( &inputSlice [iy] );
    for ( i=0; i<ilength; i++ ) *l2++ = (*b)(*l2,*l1++);
    for ( x=lx; x<dimx-1; x++ ) {
      if ( (*f)(inputSlice[iy+x], outputSlice[iy+x+1]) )
	outputSlice[iy+x+1] = inputSlice[iy+x];
    }
  }
  
  return( 1 );
}






//
// operation with a centered SE a length 3 along Y
//

template <class TYPE, class PREDICAT, class BINOP > static int _predef_binary_Y_Operation( TYPE *T, 
											   PREDICAT *P,
											   BINOP *B,
											   TYPE* inputSlice,
											   TYPE* centerSlice,
											   TYPE* outputSlice,
											   unsigned int dimx,
											   unsigned int dimy )
{
  PREDICAT *f = new PREDICAT();
  BINOP *b = new BINOP();
  unsigned int sliceLength = dimx * dimy * sizeof( TYPE );
  unsigned int ilength, i, x;
  
  int *l1, *l2;
  int lx;
  
  if ( inputSlice == outputSlice ) return( _morpho_ErrorInParameters_ );

  // we have to compare input:line[0][0] with input:line[1][0]
  // for a length of (dimy-1)*dimx
  // then output:line[1][0] with input:line[0][0]
  // for the same length
  
  ilength = ( dimy-1 ) * dimx * sizeof( TYPE ) / sizeof( int );
  lx      = ilength * sizeof( int ) / sizeof( TYPE );
  
  if ( outputSlice != centerSlice )
    memcpy( outputSlice, centerSlice, sliceLength );

  l2 = (int*)(  outputSlice );
  l1 = (int*)( &inputSlice [dimx] );
  for ( i=0; i<ilength; i++ ) *l2++ = (*b)(*l2,*l1++);

  for ( x=lx; x<(dimy-1)*dimx; x++ ) {
    if ( (*f)(inputSlice[dimx+x], outputSlice[x]) )
      outputSlice[x] = inputSlice[dimx+x];
  }
  
  l2 = (int*)( &outputSlice[dimx] );
  l1 = (int*)(  inputSlice  );
  for ( i=0; i<ilength; i++ ) *l2++ = (*b)(*l2,*l1++);

  for ( x=lx; x<(dimy-1)*dimx; x++ ) {
    if ( (*f)(inputSlice[x], outputSlice[dimx+x]) )
      outputSlice[dimx+x] = inputSlice[x];
  }
  
  return( 1 );
}







//
// operation with a centered SE a length 2 along Z
//

template <class TYPE, class PREDICAT, class BINOP > static int _predef_binary_Z_Operation( TYPE *T, 
											   PREDICAT *P,
											   BINOP *B,
											   TYPE* input1Slice,
											   TYPE* input2Slice,
											   TYPE* outputSlice,
											   unsigned int dimx,
											   unsigned int dimy )
{
  PREDICAT *f = new PREDICAT();
  BINOP *b = new BINOP();
  unsigned int sliceLength = dimx * dimy * sizeof( TYPE );
  unsigned int ilength, i, x;
  
  int *l1, *l2;
  int lx;
  

  // we have to compare input:line[0][0] with input:line[1][0]
  // for a length of (dimy-1)*dimx
  // then output:line[1][0] with input:line[0][0]
  // for the same length
  
  ilength = dimy* dimx * sizeof( TYPE ) / sizeof( int );
  lx      = ilength * sizeof( int ) / sizeof( TYPE );
  
  if ( input1Slice == outputSlice ) {

    l2 = (int*)(  outputSlice );
    l1 = (int*)(  input2Slice );
    for ( i=0; i<ilength; i++ ) *l2++ = (*b)(*l2,*l1++);
    
    for ( x=lx; x<dimy*dimx; x++ ) {
      if ( (*f)(input2Slice[x], outputSlice[x]) )
	outputSlice[x] = input2Slice[dimx+x];
    }

  }
  else if ( input2Slice == outputSlice ) {

    l2 = (int*)(  outputSlice );
    l1 = (int*)(  input1Slice );
    for ( i=0; i<ilength; i++ ) *l2++ = (*b)(*l2,*l1++);
    
    for ( x=lx; x<dimy*dimx; x++ ) {
      if ( (*f)(input1Slice[x], outputSlice[x]) )
	outputSlice[x] = input1Slice[dimx+x];
    }

  }
  else {

    memcpy( outputSlice, input2Slice, sliceLength );

    l2 = (int*)(  outputSlice );
    l1 = (int*)(  input1Slice );
    for ( i=0; i<ilength; i++ ) *l2++ = (*b)(*l2,*l1++);
    
    for ( x=lx; x<dimy*dimx; x++ ) {
      if ( (*f)(input1Slice[x], outputSlice[x]) )
	outputSlice[x] = input1Slice[dimx+x];
    }

  }

  return( 1 );
}







template <class TYPE, class PREDICAT, class BINOP > static int _predef_binary_Operation( TYPE *T, 
											 PREDICAT *P,
											 BINOP *B,
											 TYPE* inputBuf, /* buffer to be resampled */
											 TYPE* resultBuf, /* result buffer */
											 unsigned int dimx,
											 unsigned int dimy,
											 unsigned int dimz,
											 int connectivity, /* connectivity to be used */ 
											 int iterations  /* number of iterations */ )
{
  int conn = 0; 
  unsigned int z;
  int sliceSize = dimx * dimy * sizeof( TYPE );

  TYPE *theBuf;
  TYPE *resBuf;
  TYPE *input = inputBuf;
  TYPE *localBuf = NULL;
  TYPE *localTab[5];

  TYPE *prev, *next, *curr, *tmp;
  TYPE *curr8, *next8;

  int iter;

  if ( iterations <= 0 ) return( _morpho_ErrorInParameters_ );


  /* test on connectivity */
  conn = connectivity;
  switch ( conn ) {
  case 4 :
  case 8 :
  case 6 :
  case 10 :
  case 18 :
  case 26 :
    break ;
  default :
    conn = 26;
  }

  if ( dimz == 1 ) {
    switch ( conn ) {
    case 6 :
      conn = 4; break;
    case 10 :
    case 18 :
    case 26 :
      conn = 8; break;
    }
  }

  
  
  localBuf = (TYPE*)malloc( 5 * sliceSize );
  if ( localBuf == NULL ) return( _morpho_ErrMemFull_ );
  for ( z = 0; z < 5; z ++ ) {
    localTab[z] = &localBuf[z*dimx*dimy];
  }



  for ( iter = 0; iter < iterations; iter ++ ) {
    theBuf = input;
    resBuf = resultBuf;
    switch ( conn ) {
    case 4 :
      //
      // pour chaque plan 
      // - operation selon X -> tmp1
      // - operation selon Y -> tmp2
      // - copie dans resultat
      //
      for ( z=0; z<dimz; z++ ) {
	_predef_binary_X_Operation( T, P, B, &theBuf[z*dimx*dimy], localTab[0], dimx, dimy );
	_predef_binary_Y_Operation( T, P, B, &theBuf[z*dimx*dimy], localTab[0], localTab[1], 
				    dimx, dimy );
	memcpy( &resBuf[z*dimx*dimy], localTab[1], sliceSize );
      }
      break;
    case 8 :
      for ( z=0; z<dimz; z++ ) {
	_predef_binary_X_Operation( T, P, B, &theBuf[z*dimx*dimy], localTab[0], dimx, dimy );
	_predef_binary_Y_Operation( T, P, B, localTab[0], localTab[0], localTab[1], 
				    dimx, dimy );
	memcpy( &resBuf[z*dimx*dimy], localTab[1], sliceSize );
      }
      break;
    case 6 :
      prev = localTab[1];
      curr = localTab[2];
      //
      // first slice
      //
      _predef_binary_X_Operation( T, P, B, theBuf, localTab[0], dimx, dimy );
      _predef_binary_Y_Operation( T, P, B, theBuf, localTab[0], prev, dimx, dimy );
      _predef_binary_Z_Operation( T, P, B, &theBuf[dimx*dimy], prev, prev, dimx, dimy );
      //
      // middle slices
      //
      for ( z=1; z<dimz; z++ ) {
	_predef_binary_X_Operation( T, P, B, &theBuf[z*dimx*dimy], localTab[0], dimx, dimy );
	_predef_binary_Y_Operation( T, P, B, &theBuf[z*dimx*dimy], localTab[0], curr, dimx, dimy );
	_predef_binary_Z_Operation( T, P, B, &theBuf[(z-1)*dimx*dimy], curr, curr, dimx, dimy );
	if ( z < dimz-1 ) 
	  _predef_binary_Z_Operation( T, P, B, &theBuf[(z+1)*dimx*dimy], curr, curr, dimx, dimy );
	memcpy( &resBuf[(z-1)*dimx*dimy], prev, sliceSize );
	tmp = prev;   prev = curr;   curr = tmp;
      }
      //
      // last slice
      // 
      memcpy( &resBuf[(dimz-1)*dimx*dimy], prev, sliceSize );
      break;
    case 10 :
      prev = localTab[1];
      curr = localTab[2];
      //
      // first slice
      //
      _predef_binary_X_Operation( T, P, B, theBuf, localTab[0], dimx, dimy );
      _predef_binary_Y_Operation( T, P, B, localTab[0], localTab[0], prev, dimx, dimy );
      _predef_binary_Z_Operation( T, P, B, &theBuf[dimx*dimy], prev, prev, dimx, dimy );
      //
      // middle slices
      //
      for ( z=1; z<dimz; z++ ) {
	_predef_binary_X_Operation( T, P, B, &theBuf[z*dimx*dimy], localTab[0], dimx, dimy );
	_predef_binary_Y_Operation( T, P, B, localTab[0], localTab[0], curr, dimx, dimy );
	_predef_binary_Z_Operation( T, P, B, &theBuf[(z-1)*dimx*dimy], curr, curr, dimx, dimy );
	if ( z < dimz-1 ) 
	  _predef_binary_Z_Operation( T, P, B, &theBuf[(z+1)*dimx*dimy], curr, curr, dimx, dimy );
	memcpy( &resBuf[(z-1)*dimx*dimy], prev, sliceSize );
	tmp = prev;   prev = curr;   curr = tmp;
      }
      //
      // last slice
      // 
      memcpy( &resBuf[(dimz-1)*dimx*dimy], prev, sliceSize );
      break;
    case 18 :
      prev = localTab[0];
      curr = localTab[1];
      next = localTab[2];
      curr8 = localTab[3];
      next8 = localTab[4];
      // 
      // first slice
      //
      _predef_binary_X_Operation( T, P, B, theBuf, curr, dimx, dimy );
      _predef_binary_Y_Operation( T, P, B, curr, curr, curr8, dimx, dimy );
      _predef_binary_Y_Operation( T, P, B, theBuf, curr, curr, dimx, dimy );
      //
      // second slice
      //
      _predef_binary_X_Operation( T, P, B, &theBuf[dimx*dimy], next, dimx, dimy );
      _predef_binary_Y_Operation( T, P, B, next, next, next8, dimx, dimy );
      _predef_binary_Y_Operation( T, P, B, &theBuf[dimx*dimy], next, next, dimx, dimy );
      //
      // result of first slice
      //
      _predef_binary_Z_Operation( T, P, B, curr8, next, curr8, dimx, dimy );
      memcpy( resBuf, curr8, sliceSize );
      //
      // middle slice
      //
      for ( z=1; z<dimz-1; z++ ) {
	tmp = prev;    prev  = curr;    curr  = next;   next = tmp;
	tmp = curr8;   curr8 = next8;   next8 = tmp;
	_predef_binary_X_Operation( T, P, B, &theBuf[(z+1)*dimx*dimy], next, dimx, dimy );
	_predef_binary_Y_Operation( T, P, B, next, next, next8, dimx, dimy );
	_predef_binary_Y_Operation( T, P, B, &theBuf[(z+1)*dimx*dimy], next, next, dimx, dimy );
	_predef_binary_Z_Operation( T, P, B, curr8, prev, curr8, dimx, dimy );
	_predef_binary_Z_Operation( T, P, B, curr8, next, curr8, dimx, dimy );
	memcpy( &resBuf[z*dimx*dimy], curr8, sliceSize );
      }
      //
      // last slice
      //
      _predef_binary_Z_Operation( T, P, B, next8, curr, next8, dimx, dimy );
      memcpy( &resBuf[(dimz-1)*dimx*dimy], next8, sliceSize );
      break;
    case 26 :
      prev = localTab[1];
      curr = localTab[2];
      next = localTab[3];
      //
      // first and second slices
      //
      _predef_binary_X_Operation( T, P, B, theBuf, localTab[0], dimx, dimy );
      _predef_binary_Y_Operation( T, P, B, localTab[0], localTab[0], prev, dimx, dimy );
      
      _predef_binary_X_Operation( T, P, B, &theBuf[dimx*dimy], localTab[0], dimx, dimy );
      _predef_binary_Y_Operation( T, P, B, localTab[0], localTab[0], curr, dimx, dimy );

      _predef_binary_Z_Operation( T, P, B, prev, curr, localTab[0], dimx, dimy );
      memcpy( resBuf, localTab[0], sliceSize );
      //
      // middle slices
      //
      for ( z=1; z<dimz-1; z++ ) {
	_predef_binary_X_Operation( T, P, B, &theBuf[(z+1)*dimx*dimy], localTab[0], 
				    dimx, dimy );
	_predef_binary_Y_Operation( T, P, B, localTab[0], localTab[0], next, dimx, dimy );
	_predef_binary_Z_Operation( T, P, B, prev, curr, localTab[0], dimx, dimy );
	_predef_binary_Z_Operation( T, P, B, next, localTab[0], localTab[0], dimx, dimy );
	memcpy( &resBuf[z*dimx*dimy], localTab[0], sliceSize );
	tmp = prev;   prev = curr;   curr = next;   next = tmp;
      }
      // 
      // last slice
      //
      _predef_binary_Z_Operation( T, P, B, prev, curr, localTab[0], dimx, dimy );
      memcpy( &resBuf[(dimz-1)*dimx*dimy], localTab[0], sliceSize );
      break;
    }
    input = resultBuf;
  }

  return( 1 );
}








