typedef struct {
  int x;
  int y;
  int z;
  int i;
  int l;
} typePoint;

typedef enum {
  _DEFAULT_,
  _USHORT_,
  _UINT_
} enumLabelType;  

/*
 * In this implementation, buffers' types must be integer
 */ 

template <class TYPE_GRD, class TYPE_LBL> static int _watershed( TYPE_GRD *G,
								 TYPE_LBL *L,
								 TYPE_GRD *theGradient,
								 TYPE_LBL *theLabelsInput,
								 TYPE_LBL *theLabelsOutput,
								 unsigned int dimx,
								 unsigned int dimy,
								 unsigned int dimz,
								 unsigned int dimw ) 
{
  unsigned int i, v = dimx*dimy*dimz;
  unsigned int maxPossibleGradient = 0;

  unsigned int *nbPtsByGradient = NULL;
  unsigned int _ITMP_;
  unsigned int maxNbPtsByGradient = 0;

  unsigned int nbPoints;
  typePoint **thePts = (typePoint **)NULL;
  typePoint *tmpPts;

  unsigned int maxLabel = 0;
  enumLabelType labelType;
  unsigned short int  *u16Labels = NULL;
  unsigned int        *u32Labels = NULL;

  unsigned int _ISINQUEUE_;
  unsigned int _CONFLICT_;
  unsigned int _ISINQUEUE_AND_PROCESSED_;
  unsigned int _CONFLICT_AND_PROCESSED_;

  unsigned int x, y, z;
  unsigned int dimxy = dimx*dimy;
  unsigned int dimx1 = dimx-1;
  unsigned int dimy1 = dimy-1;
  unsigned int dimz1 = dimz-1;

  unsigned int j;

  int labels[6];
  int nlabels;

  /*
   * Maximum value of gradient
   */
  maxPossibleGradient = theGradient[0];
  for (i=1;i<v;i++)
    if ( maxPossibleGradient < theGradient[i] ) maxPossibleGradient = theGradient[i];
  if ( maxPossibleGradient == 0 ) {
    return( _morpho_ErrorInParameters_ );
  }

  /*
   * Array of gradient points
   */
  nbPtsByGradient = (unsigned int*)malloc( (maxPossibleGradient+2)*sizeof(unsigned int) );
  if ( nbPtsByGradient == NULL ) {
    return( _morpho_ErrMemFull_ );
  }
  for ( i=0; i<maxPossibleGradient+2; i++ ) nbPtsByGradient[i] = 0;
  _ITMP_ = maxPossibleGradient+1;

  /*
   * Number of points for each gradient value
   */
  for (i=0;i<v;i++) {
    nbPtsByGradient[ theGradient[i] ] ++;
  }
  maxNbPtsByGradient = 0;
  for ( i=0; i<=maxPossibleGradient; i++ ) 
    if ( maxNbPtsByGradient < nbPtsByGradient[i] ) maxNbPtsByGradient = nbPtsByGradient[i];

  /*
   * list allocation
   */
  for ( nbPoints=0,i=0; i<=maxPossibleGradient; i++ ) 
    nbPoints += nbPtsByGradient[i];

  thePts = (typePoint **)malloc( (maxPossibleGradient+2)*sizeof(typePoint *) +
				 (nbPoints+maxNbPtsByGradient)*sizeof(typePoint) );
  if ( thePts == (typePoint **) NULL ) {
    free( nbPtsByGradient );
    return( _morpho_ErrMemFull_ );
  }
  tmpPts = (typePoint *)(thePts + (maxPossibleGradient+2) );
  for (i=0; i<=maxPossibleGradient; i++ ) {
    thePts[i] = tmpPts;
    tmpPts += nbPtsByGradient[i];
  }
  thePts[i] = tmpPts;



  /* 
   * labels preparation
   */
  maxLabel = theLabelsInput[0];
  for (i=1;i<v;i++)
    if ( maxLabel < theLabelsInput[i] ) maxLabel = theLabelsInput[i];

  labelType = _DEFAULT_;
  if ( maxLabel > 251 && maxLabel <= 255 && dimw <= 1 ) {
    labelType = _USHORT_;
  }
  if ( maxLabel > 65531 && maxLabel <= 65535 && dimw <= 2 ) {
    labelType = _UINT_;
  }

  _ISINQUEUE_ = maxLabel+1;
  _CONFLICT_  = maxLabel+2;
  _ISINQUEUE_AND_PROCESSED_ = maxLabel+3;
  _CONFLICT_AND_PROCESSED_  = maxLabel+4;



  /* 
   * allocation is needed
   */
  switch ( labelType ) {
  case _DEFAULT_ :
  default :
    break;
  case _USHORT_ :
    u16Labels = (unsigned short int *)malloc( v * sizeof(unsigned short int) );
    if ( u16Labels == NULL ) {
      free( thePts );
      free( nbPtsByGradient );
      return( _morpho_ErrMemFull_ );
    }
    break;
  case _UINT_ :
    u32Labels = (unsigned int *)malloc( v * sizeof(unsigned int) );
    if ( u32Labels == NULL ) {
      free( thePts );
      free( nbPtsByGradient );
      return( _morpho_ErrMemFull_ );
    }
    break;
  }
    
  /*
   * labels initialisation
   */
  switch ( labelType ) {
  case _DEFAULT_ :
  default :
    if ( theLabelsInput != theLabelsOutput ) 
      memcpy( theLabelsOutput, theLabelsInput, v*dimw );
    break;
  case _USHORT_ :
    for ( i=0; i<v; i++ ) 
      u16Labels[i] = theLabelsInput[i];
    break;
  case _UINT_ :
    for ( i=0; i<v; i++ ) 
      u32Labels[i] = theLabelsInput[i];
    break;
  }
    

  /* 
   * queue initialisation
   */
  for ( i=0; i<maxPossibleGradient+2; i++ ) nbPtsByGradient[i] = 0;

  /*
   * 2D CASE: queue initialisation
   */
#define _QUEUE_INIT_2D( THEBUF ) {                                              \
    for ( i=0, y=0; y<dimy; y++ )						\
    for ( x=0; x<dimx; x++, i++ ) {						\
      if ( THEBUF[i] > 0 && THEBUF[i] <= maxLabel ) continue;			\
      										\
      if ( x < dimx1 && THEBUF[i+1] > 0 && THEBUF[i+1] <= maxLabel ) {		\
	if ( THEBUF[i] == 0 ) {							\
	  tmpPts = &(thePts[ theGradient[i] ][ nbPtsByGradient[ theGradient[i] ] ]); \
	  tmpPts->x = x;   tmpPts->y = y;   tmpPts->z = 0;			\
	  tmpPts->i = i;   tmpPts->l = THEBUF[i+1];				\
	  nbPtsByGradient[ theGradient[i] ] ++;					\
	  THEBUF[i] = _ISINQUEUE_;						\
	} else {								\
	  THEBUF[i] = _CONFLICT_;						\
	}									\
      }										\
										\
      if ( x > 0 && THEBUF[i-1] > 0 && THEBUF[i-1] <= maxLabel ) {		\
	if ( THEBUF[i] == 0 ) {							\
	  tmpPts = &(thePts[ theGradient[i] ][ nbPtsByGradient[ theGradient[i] ] ]); \
	  tmpPts->x = x;   tmpPts->y = y;   tmpPts->z = 0;			\
	  tmpPts->i = i;   tmpPts->l = THEBUF[i-1];				\
	  nbPtsByGradient[ theGradient[i] ] ++;					\
	  THEBUF[i] = _ISINQUEUE_;						\
	} else {								\
	  THEBUF[i] = _CONFLICT_;						\
	}									\
      }										\
      										\
      if ( y < dimy1 && THEBUF[i+dimx] > 0 && THEBUF[i+dimx] <= maxLabel ) {	\
	if ( THEBUF[i] == 0 ) {							\
	  tmpPts = &(thePts[ theGradient[i] ][ nbPtsByGradient[ theGradient[i] ] ]); \
	  tmpPts->x = x;   tmpPts->y = y;   tmpPts->z = 0;			\
	  tmpPts->i = i;   tmpPts->l = THEBUF[i+dimx];				\
	  nbPtsByGradient[ theGradient[i] ] ++;					\
	  THEBUF[i] = _ISINQUEUE_;						\
	} else {								\
	  THEBUF[i] = _CONFLICT_;						\
	}									\
      }										\
      										\
      if ( y > 0 && THEBUF[i-dimx] > 0 && THEBUF[i-dimx] <= maxLabel ) {	\
	if ( THEBUF[i] == 0 ) {							\
	  tmpPts = &(thePts[ theGradient[i] ][ nbPtsByGradient[ theGradient[i] ] ]); \
	  tmpPts->x = x;   tmpPts->y = y;   tmpPts->z = 0;			\
	  tmpPts->i = i;   tmpPts->l = THEBUF[i-dimx];				\
	  nbPtsByGradient[ theGradient[i] ] ++;					\
	  THEBUF[i] = _ISINQUEUE_;						\
	} else {								\
	  THEBUF[i] = _CONFLICT_;						\
	}									\
      }										\
										\
    }										\
  }


#define _QUEUE_INIT_3D( THEBUF ) { 						\
    for ( i=0, z=0; z<dimz; z++ )						\
    for ( y=0; y<dimy; y++ )							\
    for ( x=0; x<dimx; x++, i++ ) {						\
      if ( THEBUF[i] > 0 && THEBUF[i] <= maxLabel ) continue;			\
      										\
      if ( x < dimx1 && THEBUF[i+1] > 0 && THEBUF[i+1] <= maxLabel ) {		\
	if ( THEBUF[i] == 0 ) {							\
	  tmpPts = &(thePts[ theGradient[i] ][ nbPtsByGradient[ theGradient[i] ] ]); \
	  tmpPts->x = x;   tmpPts->y = y;   tmpPts->z = z;			\
	  tmpPts->i = i;   tmpPts->l = THEBUF[i+1];				\
	  nbPtsByGradient[ theGradient[i] ] ++;					\
	  THEBUF[i] = _ISINQUEUE_;						\
	} else {								\
	  THEBUF[i] = _CONFLICT_;						\
	}									\
      }										\
										\
      if ( x > 0 && THEBUF[i-1] > 0 && THEBUF[i-1] <= maxLabel ) {		\
	if ( THEBUF[i] == 0 ) {							\
	  tmpPts = &(thePts[ theGradient[i] ][ nbPtsByGradient[ theGradient[i] ] ]); \
	  tmpPts->x = x;   tmpPts->y = y;   tmpPts->z = z;			\
	  tmpPts->i = i;   tmpPts->l = THEBUF[i-1];				\
	  nbPtsByGradient[ theGradient[i] ] ++;					\
	  THEBUF[i] = _ISINQUEUE_;						\
	} else {								\
	  THEBUF[i] = _CONFLICT_;						\
	}									\
      }										\
      										\
      if ( y < dimy1 && THEBUF[i+dimx] > 0 && THEBUF[i+dimx] <= maxLabel ) {	\
	if ( THEBUF[i] == 0 ) {							\
	  tmpPts = &(thePts[ theGradient[i] ][ nbPtsByGradient[ theGradient[i] ] ]); \
	  tmpPts->x = x;   tmpPts->y = y;   tmpPts->z = z;			\
	  tmpPts->i = i;   tmpPts->l = THEBUF[i+dimx];				\
	  nbPtsByGradient[ theGradient[i] ] ++;					\
	  THEBUF[i] = _ISINQUEUE_;						\
	} else {								\
	  THEBUF[i] = _CONFLICT_;						\
	}									\
      }										\
      										\
      if ( y > 0 && THEBUF[i-dimx] > 0 && THEBUF[i-dimx] <= maxLabel ) {	\
	if ( THEBUF[i] == 0 ) {							\
	  tmpPts = &(thePts[ theGradient[i] ][ nbPtsByGradient[ theGradient[i] ] ]); \
	  tmpPts->x = x;   tmpPts->y = y;   tmpPts->z = z;			\
	  tmpPts->i = i;   tmpPts->l = THEBUF[i-dimx];				\
	  nbPtsByGradient[ theGradient[i] ] ++;					\
	  THEBUF[i] = _ISINQUEUE_;						\
	} else {								\
	  THEBUF[i] = _CONFLICT_;						\
	}									\
      }										\
										\
      if ( z < dimz1 && THEBUF[i+dimxy] > 0 && THEBUF[i+dimxy] <= maxLabel ) {	\
	if ( THEBUF[i] == 0 ) {							\
	  tmpPts = &(thePts[ theGradient[i] ][ nbPtsByGradient[ theGradient[i] ] ]); \
	  tmpPts->x = x;   tmpPts->y = y;   tmpPts->z = z;			\
	  tmpPts->i = i;   tmpPts->l = THEBUF[i+dimxy];				\
	  nbPtsByGradient[ theGradient[i] ] ++;					\
	  THEBUF[i] = _ISINQUEUE_;						\
	} else {								\
	  THEBUF[i] = _CONFLICT_;						\
	}									\
      }										\
      										\
      if ( z > 0 && THEBUF[i-dimxy] > 0 && THEBUF[i-dimxy] <= maxLabel ) {	\
	if ( THEBUF[i] == 0 ) {							\
	  tmpPts = &(thePts[ theGradient[i] ][ nbPtsByGradient[ theGradient[i] ] ]); \
	  tmpPts->x = x;   tmpPts->y = y;   tmpPts->z = z;			\
	  tmpPts->i = i;   tmpPts->l = THEBUF[i-dimxy];				\
	  nbPtsByGradient[ theGradient[i] ] ++;					\
	  THEBUF[i] = _ISINQUEUE_;						\
	} else {								\
	  THEBUF[i] = _CONFLICT_;						\
	}									\
      }										\
										\
    }										\
  }





  

  if ( dimz == 1 ) {

    switch ( labelType ) {
    case _DEFAULT_ :
    default :
      _QUEUE_INIT_2D( theLabelsOutput );
      break;
    case _USHORT_ :
      _QUEUE_INIT_2D( u16Labels );
      break;
    case _UINT_ :
      _QUEUE_INIT_2D( u32Labels );
      break;
    }

  }
  /*
   * 3D CASE: queue initialisation
   */
  else {

    switch ( labelType ) {
    case _DEFAULT_ :
    default :
      _QUEUE_INIT_3D( theLabelsOutput );
      break;
    case _USHORT_ :
      _QUEUE_INIT_3D( u16Labels );
      break;
    case _UINT_ :
      _QUEUE_INIT_3D( u32Labels );
      break;
    }

  }


  /*
   * number of points in queue
   */
  for ( nbPoints=0,i=0; i<=maxPossibleGradient; i++ ) 
    nbPoints += nbPtsByGradient[i];



#define _LABELS_UPDATE( THEBUF ) { 							\
    for ( j = 0; j < nbPtsByGradient[_ITMP_]; j++ ) {					\
      if ( THEBUF[ thePts[_ITMP_][j].i ] == _ISINQUEUE_ ) {				\
	THEBUF[ thePts[_ITMP_][j].i ] = _ISINQUEUE_AND_PROCESSED_;			\
      }											\
      else if ( THEBUF[ thePts[_ITMP_][j].i ] == _CONFLICT_  ) {			\
	THEBUF[ thePts[_ITMP_][j].i ] = _CONFLICT_AND_PROCESSED_;			\
      }											\
      else {										\
	fprintf( stderr, "(1) strange label for (%d,%d,%d) = %d (max=%d)\n",		\
		 thePts[_ITMP_][j].x, thePts[_ITMP_][j].y, thePts[_ITMP_][j].z, 	\
		 THEBUF[ thePts[_ITMP_][j].i ], maxLabel );				\
      }											\
    }						                                  	\
  }



#define _QUEUE_UPDATE_2D( THEBUF ) {                                                    \
      for ( j = 0; j < nbPtsByGradient[_ITMP_]; j++ ) {					\
											\
	x = thePts[_ITMP_][j].x;							\
	y = thePts[_ITMP_][j].y;							\
	z = thePts[_ITMP_][j].z;							\
	i = thePts[_ITMP_][j].i;							\
      											\
	if ( THEBUF[ thePts[_ITMP_][j].i ] == _ISINQUEUE_AND_PROCESSED_ ) {		\
	  THEBUF[ thePts[_ITMP_][j].i ] = thePts[_ITMP_][j].l;				\
	}										\
	else if ( THEBUF[ thePts[_ITMP_][j].i ] == _CONFLICT_AND_PROCESSED_ ) {		\
	  nlabels = 0;									\
	  if ( x < dimx-1 && THEBUF[i+1] > 0 && THEBUF[i+1] <= maxLabel )		\
	    labels[ nlabels ++ ] = THEBUF[i+1];						\
	  if ( x > 0      && THEBUF[i-1] > 0 && THEBUF[i-1] <= maxLabel )		\
	    labels[ nlabels ++ ] = THEBUF[i-1];						\
	  if ( y < dimy-1 && THEBUF[i+dimx] > 0 && THEBUF[i+dimx] <= maxLabel ) 	\
	    labels[ nlabels ++ ] = THEBUF[i+dimx];					\
	  if ( y > 0      && THEBUF[i-dimx] > 0 && THEBUF[i-dimx] <= maxLabel )		\
	    labels[ nlabels ++ ] = THEBUF[i-dimx];					\
	  										\
	  /* il faudrait mieux traiter le cas 'plusieurs labels en conflit' */		\
	  /* mauvais choix #1: THEBUF[ thePts[_ITMP_][j].i ] = labels[ 0 ]; */		\
	  THEBUF[ thePts[_ITMP_][j].i ] = thePts[_ITMP_][j].l;				\
	}										\
	else {										\
	  fprintf( stderr, "(2) strange label for (%d,%d,%d) = %d (max=%d)\n",		\
		   thePts[_ITMP_][j].x, thePts[_ITMP_][j].y, thePts[_ITMP_][j].z, 	\
		   THEBUF[ thePts[_ITMP_][j].i ], maxLabel );				\
	}										\
											\
	/* add neighbors */								\
	if ( x < dimx1 ) {								\
	  if ( THEBUF[i+1] == 0  ) {							\
	    tmpPts = &(thePts[ theGradient[i+1] ][ nbPtsByGradient[ theGradient[i+1] ] ]); \
	    tmpPts->x = x+1;   tmpPts->y = y;     tmpPts->z = z;			\
	    tmpPts->i = i+1;   tmpPts->l = THEBUF[i];					\
	    nbPtsByGradient[ theGradient[i+1] ] ++;					\
	    THEBUF[i+1] = _ISINQUEUE_;							\
	  } 										\
	  else if ( THEBUF[i+1] == _ISINQUEUE_ )					\
	    THEBUF[i+1] = _CONFLICT_;							\
	}										\
	if ( x > 0 ) {									\
	  if ( THEBUF[i-1] == 0  ) {							\
	    tmpPts = &(thePts[ theGradient[i-1] ][ nbPtsByGradient[ theGradient[i-1] ] ]); \
	    tmpPts->x = x-1;   tmpPts->y = y;     tmpPts->z = z;			\
	    tmpPts->i = i-1;   tmpPts->l = THEBUF[i];					\
	    nbPtsByGradient[ theGradient[i-1] ] ++;					\
	    THEBUF[i-1] = _ISINQUEUE_;							\
	  } 										\
	  else if ( THEBUF[i-1] == _ISINQUEUE_ )					\
	    THEBUF[i-1] = _CONFLICT_;							\
	}										\
	if ( y < dimy1 ) {								\
	  if ( THEBUF[i+dimx] == 0  ) {							\
	    tmpPts = &(thePts[ theGradient[i+dimx] ][ nbPtsByGradient[ theGradient[i+dimx] ] ]); \
	    tmpPts->x = x;   tmpPts->y = y+1;     tmpPts->z = z;			\
	    tmpPts->i = i+dimx;   tmpPts->l = THEBUF[i];				\
	    nbPtsByGradient[ theGradient[i+dimx] ] ++;					\
	    THEBUF[i+dimx] = _ISINQUEUE_;						\
	  } 										\
	  else if ( THEBUF[i+dimx] == _ISINQUEUE_ )					\
	    THEBUF[i+dimx] = _CONFLICT_;						\
	}										\
	if ( y > 0 ) {									\
	  if ( THEBUF[i-dimx] == 0  ) {							\
	    tmpPts = &(thePts[ theGradient[i-dimx] ][ nbPtsByGradient[ theGradient[i-dimx] ] ]); \
	    tmpPts->x = x;   tmpPts->y = y-1;     tmpPts->z = z;			\
	    tmpPts->i = i-dimx;   tmpPts->l = THEBUF[i];				\
	    nbPtsByGradient[ theGradient[i-dimx] ] ++;					\
	    THEBUF[i-dimx] = _ISINQUEUE_;						\
	  } 										\
	  else if ( THEBUF[i-dimx] == _ISINQUEUE_ )					\
	    THEBUF[i-dimx] = _CONFLICT_;						\
	}										\
											\
      }                                                                                 \
    }

#define _QUEUE_UPDATE_3D( THEBUF ) {                                                    \
      for ( j = 0; j < nbPtsByGradient[_ITMP_]; j++ ) {					\
											\
	x = thePts[_ITMP_][j].x;							\
	y = thePts[_ITMP_][j].y;							\
	z = thePts[_ITMP_][j].z;							\
	i = thePts[_ITMP_][j].i;							\
      											\
	if ( THEBUF[ thePts[_ITMP_][j].i ] == _ISINQUEUE_AND_PROCESSED_ ) {		\
	  THEBUF[ thePts[_ITMP_][j].i ] = thePts[_ITMP_][j].l;				\
	}										\
	else if ( THEBUF[ thePts[_ITMP_][j].i ] == _CONFLICT_AND_PROCESSED_ ) {		\
	  nlabels = 0;									\
	  if ( x < dimx-1 && THEBUF[i+1] > 0 && THEBUF[i+1] <= maxLabel )		\
	    labels[ nlabels ++ ] = THEBUF[i+1];						\
	  if ( x > 0      && THEBUF[i-1] > 0 && THEBUF[i-1] <= maxLabel )		\
	    labels[ nlabels ++ ] = THEBUF[i-1];						\
	  if ( y < dimy-1 && THEBUF[i+dimx] > 0 && THEBUF[i+dimx] <= maxLabel ) 	\
	    labels[ nlabels ++ ] = THEBUF[i+dimx];					\
	  if ( y > 0      && THEBUF[i-dimx] > 0 && THEBUF[i-dimx] <= maxLabel )		\
	    labels[ nlabels ++ ] = THEBUF[i-dimx];					\
	  if ( z < dimz-1 && THEBUF[i+dimxy] > 0 && THEBUF[i+dimxy] <= maxLabel )	\
	    labels[ nlabels ++ ] = THEBUF[i+dimxy];					\
	  if ( z > 0      && THEBUF[i-dimxy] > 0 && THEBUF[i-dimxy] <= maxLabel )	\
	    labels[ nlabels ++ ] = THEBUF[i-dimxy];					\
	  										\
	  /* il faudrait mieux traiter le cas 'plusieurs labels en conflit' */		\
	  /* mauvais choix #1: THEBUF[ thePts[_ITMP_][j].i ] = labels[ 0 ]; */		\
	  THEBUF[ thePts[_ITMP_][j].i ] = thePts[_ITMP_][j].l;				\
	}										\
	else {										\
	  fprintf( stderr, "(2) strange label for (%d,%d,%d) = %d (max=%d)\n",		\
		   thePts[_ITMP_][j].x, thePts[_ITMP_][j].y, thePts[_ITMP_][j].z, 	\
		   THEBUF[ thePts[_ITMP_][j].i ], maxLabel );				\
	}										\
											\
	/* add neighbors */								\
	if ( x < dimx1 ) {								\
	  if ( THEBUF[i+1] == 0  ) {							\
	    tmpPts = &(thePts[ theGradient[i+1] ][ nbPtsByGradient[ theGradient[i+1] ] ]); \
	    tmpPts->x = x+1;   tmpPts->y = y;     tmpPts->z = z;			\
	    tmpPts->i = i+1;   tmpPts->l = THEBUF[i];					\
	    nbPtsByGradient[ theGradient[i+1] ] ++;					\
	    THEBUF[i+1] = _ISINQUEUE_;							\
	  } 										\
	  else if ( THEBUF[i+1] == _ISINQUEUE_ )					\
	    THEBUF[i+1] = _CONFLICT_;							\
	}										\
	if ( x > 0 ) {									\
	  if ( THEBUF[i-1] == 0  ) {							\
	    tmpPts = &(thePts[ theGradient[i-1] ][ nbPtsByGradient[ theGradient[i-1] ] ]); \
	    tmpPts->x = x-1;   tmpPts->y = y;     tmpPts->z = z;			\
	    tmpPts->i = i-1;   tmpPts->l = THEBUF[i];					\
	    nbPtsByGradient[ theGradient[i-1] ] ++;					\
	    THEBUF[i-1] = _ISINQUEUE_;							\
	  } 										\
	  else if ( THEBUF[i-1] == _ISINQUEUE_ )					\
	    THEBUF[i-1] = _CONFLICT_;							\
	}										\
	if ( y < dimy1 ) {								\
	  if ( THEBUF[i+dimx] == 0  ) {							\
	    tmpPts = &(thePts[ theGradient[i+dimx] ][ nbPtsByGradient[ theGradient[i+dimx] ] ]); \
	    tmpPts->x = x;   tmpPts->y = y+1;     tmpPts->z = z;			\
	    tmpPts->i = i+dimx;   tmpPts->l = THEBUF[i];				\
	    nbPtsByGradient[ theGradient[i+dimx] ] ++;					\
	    THEBUF[i+dimx] = _ISINQUEUE_;						\
	  } 										\
	  else if ( THEBUF[i+dimx] == _ISINQUEUE_ )					\
	    THEBUF[i+dimx] = _CONFLICT_;						\
	}										\
	if ( y > 0 ) {									\
	  if ( THEBUF[i-dimx] == 0  ) {							\
	    tmpPts = &(thePts[ theGradient[i-dimx] ][ nbPtsByGradient[ theGradient[i-dimx] ] ]); \
	    tmpPts->x = x;   tmpPts->y = y-1;     tmpPts->z = z;			\
	    tmpPts->i = i-dimx;   tmpPts->l = THEBUF[i];				\
	    nbPtsByGradient[ theGradient[i-dimx] ] ++;					\
	    THEBUF[i-dimx] = _ISINQUEUE_;						\
	  } 										\
	  else if ( THEBUF[i-dimx] == _ISINQUEUE_ )					\
	    THEBUF[i-dimx] = _CONFLICT_;						\
	}										\
											\
	if ( z < dimz1 ) {								\
	  if ( THEBUF[i+dimxy] == 0  ) {						\
	    tmpPts = &(thePts[ theGradient[i+dimxy] ][ nbPtsByGradient[ theGradient[i+dimxy] ] ]); \
	    tmpPts->x = x;   tmpPts->y = y;     tmpPts->z = z+1;			\
	    tmpPts->i = i+dimxy;   tmpPts->l = THEBUF[i];				\
	    nbPtsByGradient[ theGradient[i+dimxy] ] ++;					\
	    THEBUF[i+dimxy] = _ISINQUEUE_;						\
	  } 										\
	  else if ( THEBUF[i+dimxy] == _ISINQUEUE_ )					\
	    THEBUF[i+dimxy] = _CONFLICT_;						\
	}										\
	if ( z > 0 ) {									\
	  if ( THEBUF[i-dimxy] == 0  ) {						\
	    tmpPts = &(thePts[ theGradient[i-dimxy] ][ nbPtsByGradient[ theGradient[i-dimxy] ] ]); \
	    tmpPts->x = x;   tmpPts->y = y;     tmpPts->z = z-1;			\
	    tmpPts->i = i-dimxy;   tmpPts->l = THEBUF[i];				\
	    nbPtsByGradient[ theGradient[ i-dimxy] ] ++;				\
	    THEBUF[i-dimxy] = _ISINQUEUE_;						\
	  } 										\
	  else if ( THEBUF[i-dimxy] == _ISINQUEUE_ )					\
	    THEBUF[i-dimxy] = _CONFLICT_;						\
	}										\
      }                                                                                 \
    }


  /*
   * queue processing
   */
  while ( nbPoints > 0 ) {

    /* 
     * first non-empty sub-queue
     */
    for (i=0; i<=maxPossibleGradient && nbPtsByGradient[i] == 0; i++ ) 
      ;

    /*
     * processing sub-queue: copy it and give labels to points
     */
    (void)memcpy( thePts[_ITMP_], thePts[i], nbPtsByGradient[i]*sizeof(typePoint) );
    nbPtsByGradient[_ITMP_] =  nbPtsByGradient[i];
    nbPtsByGradient[i] = 0;

    switch ( labelType ) {
    case _DEFAULT_ :
    default :
      _LABELS_UPDATE( theLabelsOutput );
      break;
    case _USHORT_ :
      _LABELS_UPDATE( u16Labels );
      break;
    case _UINT_ :
      _LABELS_UPDATE( u32Labels );
      break;
    }

    /*
     * processing sub-queue: add neighbors to queues
     */
    if ( dimz == 1 ) {
      
      switch ( labelType ) {
      case _DEFAULT_ :
      default :
	_QUEUE_UPDATE_2D( theLabelsOutput );
	break;
      case _USHORT_ :
	_QUEUE_UPDATE_2D( u16Labels );
	break;
      case _UINT_ :
	_QUEUE_UPDATE_2D( u32Labels );
	break;
      }
      
    }
    /*
     * 3D CASE: queue initialisation
     */
    else {
      
      switch ( labelType ) {
      case _DEFAULT_ :
      default :
	_QUEUE_UPDATE_3D( theLabelsOutput );
	break;
      case _USHORT_ :
	_QUEUE_UPDATE_3D( u16Labels );
	break;
      case _UINT_ :
	_QUEUE_UPDATE_3D( u32Labels );
	break;
      }

  }

    /*
     * update number of points in queue
     */
    for ( nbPoints=0,i=0; i<=maxPossibleGradient; i++ ) 
      nbPoints += nbPtsByGradient[i];

  } // while ( nbPoints > 0 )


  /*
   *
   */
  
  free( thePts );
  free( nbPtsByGradient );

  /*
   * output
   */
  switch ( labelType ) {
  case _DEFAULT_ :
  default :
    break;
  case _USHORT_ :
    for ( i=0; i<v; i++ ) 
      theLabelsOutput[i] = static_cast<TYPE_LBL>(u16Labels[i]);
    free( u16Labels );
    break;
  case _UINT_ :
    for ( i=0; i<v; i++ ) 
      theLabelsOutput[i] = u32Labels[i];
    free( u32Labels );
    break;
  }

  return( 1 );

}

