#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/anisotropicDiffusion.h"

#include <inrimage/anisotropicDiffusion.h>
#include <inrimage/recursiveFilters.h>

//using namespace std;

namespace yav {

#ifdef USE_MTL

// Anisotropic diffusion for 2D and 3D images
void anisotropicDiffusion(const Inrimage *inr, Inrimage *& output, const float & dt , const float & tmax , const float & quantx, const float & quanty, const float & quantz,  float  k, const bool smooth, const float & sigma, const DIFFUSION_FUNCTION & p, const bool disk, const DIFFUSION_TYPE & td, const NUMERICAL_SCHEME & ns) {

  Inrimage::WORD_TYPE wt = inr->getType();
  if (wt != Inrimage::WT_FLOAT)
    output->convertCast(Inrimage::WT_FLOAT);
  unsigned int nbiter = (unsigned int) floor(tmax/dt+0.5), x, y, z, t;
  unsigned int xdim = inr->getX(), ydim = inr->getY(), zdim = inr->getZ();
  double vx = 1.0/inr->getX(), vy = 1.0/inr->getVY(), vz = 1.0/inr->getVZ();
  
  Inrimage * star = new Inrimage(*inr);
  if (wt != Inrimage::WT_FLOAT)
    star->convertCast(Inrimage::WT_FLOAT);
  float *** start = (float ***) star -> getArray();
  float *** buf = (float ***) output->getArray();
  for(z=0;z<zdim;++z)
    for(y=0;y<ydim;++y)
      for(x=0;x<xdim;++x)
  	buf[z][y][x]=start[z][y][x];
  unsigned int ms = xdim * ydim * zdim;
  float d = 2.0, thresh;
  MTLVecIt i;
  MTLVector u1(ms), us(ms);
  
  //Scalar Case
  if(td==SCALAR) {
    MTLVector u1(ms), du(ms);
    if(zdim>1) d = 3.0;
    
    if(!disk) {
      MTLVector u2(ms), u3(ms);
      for(t=0;t<nbiter;t++) {
	CERR << " " << (unsigned int) ((float) t / (tmax/dt) * 100.0) << " % " << "\r" ;
	
	i=u1.begin();
	for(z=0;z<zdim;++z) {
	  for(y=0;y<ydim;++y) {
	    for(x=0;x<xdim;++x) {
	      *i = buf[z][y][x];
	      ++i;
	    }
	  }
	}
	if(smooth) {
	  Inrimage * gradient  = new Inrimage(*output);
	  recursiveFilter(gradient,DERICHE_GAUSSIAN,DERIVATIVE_X,sigma,false,false);
	  recursiveFilter(gradient,DERICHE_GAUSSIAN,SMOOTH_Y,sigma,false,false);
	  recursiveFilter(gradient,DERICHE_GAUSSIAN,SMOOTH_Z,sigma,false,false);
	  float *** g = (float ***) gradient->getArray();
	  i=du.begin();
	  for(z=0;z<zdim;z++) {
	    for(y=0;y<ydim;y++) {
	      for(x=0;x<xdim;x++) {
		*i = fabs(g[z][y][x]);
		++i;
	      }
	    }
	  }
	  delete(gradient);
	}
	else derive(du,u1,vx);
	if(k<0.0)
	  thresh = quantile(du,quantx);
	else
	  thresh = k;
	diffusivity(du,thresh,p);
	if(ns==AOS_scheme)
	  AOS(u1,du,dt,d,xdim);
	else
	  explicite(u1,du,dt,xdim);

	i=u2.begin();
	for(z=0;z<zdim;z++) {
	  for(x=0;x<xdim;x++) {
	    for(y=0;y<ydim;y++) {  
	      *i = buf[z][y][x];
	      ++i;
	    }
	  }
	}
	if(smooth) {
	  Inrimage * gradient  = new Inrimage(*output);
	  recursiveFilter(gradient,DERICHE_GAUSSIAN,SMOOTH_X,sigma,false,false);
	  recursiveFilter(gradient,DERICHE_GAUSSIAN,DERIVATIVE_Y,sigma,false,false);
	  recursiveFilter(gradient,DERICHE_GAUSSIAN,SMOOTH_Z,sigma,false,false);
	  float *** g = (float ***) gradient->getArray();
	  i=du.begin();
	  for(z=0;z<zdim;z++) {
	    for(x=0;x<xdim;x++) {
	      for(y=0;y<ydim;y++) {
		*i = fabs(g[z][y][x]);
		++i;
	      }
	    }
	  }
	  delete(gradient);
	}
	else derive(du,u2,vy); 
	if (k<0.0)
	  thresh = quantile(du,quanty);
	else
	  thresh = k;
	diffusivity(du,thresh,p); 
	if(ns==AOS_scheme)
	  AOS(u2,du,dt,d,ydim);
	else
	  explicite(u2,du,dt,ydim);
	
	if(d==3.0) {
	  i=u3.begin();
	  for(x=0;x<xdim;x++) {
	    for(y=0;y<ydim;y++) {
	      for(z=0;z<zdim;z++) {	
		*i = buf[z][y][x];
		++i;
	      }
	    } 
	  }
	  if(smooth) { 
	    Inrimage * gradient  = new Inrimage(*output);
	    recursiveFilter(gradient,DERICHE_GAUSSIAN,SMOOTH_X,sigma,false,false);
	    recursiveFilter(gradient,DERICHE_GAUSSIAN,SMOOTH_Y,sigma,false,false);
	    recursiveFilter(gradient,DERICHE_GAUSSIAN,DERIVATIVE_Z,sigma,false,false);
	    float *** g = (float ***) gradient->getArray();
	    i=du.begin();
	    for(x=0;x<xdim;x++) {
	      for(y=0;y<ydim;y++) {
		for(z=0;z<zdim;z++) {
		  *i = fabs(g[z][y][x]);
		  ++i;
		}
	      }
	    }
	    delete(gradient);
	  }
	  else derive(du,u3,vz); 
	  if(k<0.0)
	    thresh = quantile(du,quantz);
	  else
	    thresh = k;
	  diffusivity(du,thresh,p);
	  if(ns==AOS_scheme)
	    AOS(u3,du,dt,d,zdim);
	  else
	    explicite(u3,du,dt,ydim);
	}

	if(ns==AOS_scheme) {
	  i=u1.begin();
	  for(z=0;z<zdim;z++) {
	    for(y=0;y<ydim;y++) {
	      for(x=0;x<xdim;x++) {
		buf[z][y][x] = *i / d;
		++i;
	      }    
	    }
	  }
	  i=u2.begin();
	  for(z=0;z<zdim;z++) {
	    for(x=0;x<xdim;x++) {
	      for(y=0;y<ydim;y++) {  
		buf[z][y][x] += *i / d;
		++i;
	      }
	    }
	  }
	  if(d==3.0) {
	    i=u3.begin();
	    for(x=0;x<xdim;x++) {
	      for(y=0;y<ydim;y++) {
		for(z=0;z<zdim;z++) {	
		  buf[z][y][x] += *i / d;
		  ++i;
		}
	      }
	    }
	  }
	}
	else {
	  i=u1.begin();
	  for(z=0;z<zdim;z++) {
	    for(y=0;y<ydim;y++) {
	      for(x=0;x<xdim;x++) {
		buf[z][y][x] += *i;
		++i;
	      }    
	    }
	  }
	  i=u2.begin();
	  for(z=0;z<zdim;z++) {
	    for(x=0;x<xdim;x++) {
	      for(y=0;y<ydim;y++) {  
		buf[z][y][x] += *i;
		++i;
	      }
	    }
	  }
	  if(d==3.0) {
	    i=u3.begin();
	    for(x=0;x<xdim;x++) {
	      for(y=0;y<ydim;y++) {
		for(z=0;z<zdim;z++) {	
		  buf[z][y][x] += *i;	 
		  ++i;
		}
	      }
	    }
	  }
	}
      }
    }
    
    else {
      for(t=0;t<nbiter;t++) {
	CERR << " " << (unsigned int) ((float) t / (tmax/dt) * 100.0) << " %" << "\r";
	
	i=u1.begin();
	for(z=0;z<zdim;z++) {
	  for(y=0;y<ydim;y++) {
	    for(x=0;x<xdim;x++) {
	      *i = buf[z][y][x];
	      ++i;
	    }
	  }
	}
	if(smooth) {
	  Inrimage * gradient  = new Inrimage(*output);
	  recursiveFilter(gradient,DERICHE_GAUSSIAN,DERIVATIVE_X,sigma,false,false);
	  recursiveFilter(gradient,DERICHE_GAUSSIAN,SMOOTH_Y,sigma,false,false);
	  recursiveFilter(gradient,DERICHE_GAUSSIAN,SMOOTH_Z,sigma,false,false);
	  float *** g = (float ***) gradient->getArray();
	  i=du.begin();
	  for(z=0;z<zdim;z++) {
	    for(y=0;y<ydim;y++) {
	      for(x=0;x<xdim;x++) {
		*i = fabs(g[z][y][x]);
		++i;
	      }
	    }
	  }
	  delete(gradient);
	}
	else derive(du,u1,vx);
	if(k<0.0)
	  thresh = quantile(du,quantx);
	else
	  thresh = k;
	diffusivity(du,thresh,p);
	if(ns==AOS_scheme)
	  AOS(u1,du,dt,d,xdim);
	else
	  explicite(u1,du,dt,xdim);
	writeMTLVector("anisodiff.tmp1~",u1);
	
	i=u1.begin();
	for(z=0;z<zdim;z++) {
	  for(x=0;x<xdim;x++) {
	    for(y=0;y<ydim;y++) {  
	      *i = buf[z][y][x];
	      ++i;
	    }
	  }
	}
	if(smooth) {
	  Inrimage * gradient  = new Inrimage(*output);
	  recursiveFilter(gradient,DERICHE_GAUSSIAN,SMOOTH_X,sigma,false,false);
	  recursiveFilter(gradient,DERICHE_GAUSSIAN,DERIVATIVE_Y,sigma,false,false);
	  recursiveFilter(gradient,DERICHE_GAUSSIAN,SMOOTH_Z,sigma,false,false);
	  float *** g = (float ***) gradient->getArray();
	  i=du.begin();
	  for(z=0;z<zdim;z++) {
	    for(x=0;x<xdim;x++) {
	      for(y=0;y<ydim;y++) {
		*i = fabs(g[z][y][x]);
		++i;
	      }
	    }
	  }
	  delete(gradient);
	}
	else derive(du,u1,vy);
	if(k<0.0)
	  thresh = quantile(du,quanty);
	else
	  thresh = k;
	diffusivity(du,thresh,p);
	if(ns==AOS_scheme)
	  AOS(u1,du,dt,d,ydim);
	else
	  explicite(u1,du,dt,ydim);
	writeMTLVector("anisodiff.tmp2~",u1);
	
	if(d==3.0) { 
	  i=u1.begin();
	  for(x=0;x<xdim;x++) {
	    for(y=0;y<ydim;y++) {
	      for(z=0;z<zdim;z++) {	
		*i = buf[z][y][x];
		++i;
	      }
	    } 
	  }
	  if(smooth) {
	    Inrimage * gradient  = new Inrimage(*output);
	    recursiveFilter(gradient,DERICHE_GAUSSIAN,SMOOTH_X,sigma,false,false);
	    recursiveFilter(gradient,DERICHE_GAUSSIAN,SMOOTH_Y,sigma,false,false);
	    recursiveFilter(gradient,DERICHE_GAUSSIAN,DERIVATIVE_Z,sigma,false,false);
	    float *** g = (float ***) gradient->getArray();
	    i=du.begin();
	    for(x=0;x<xdim;x++) {
	      for(y=0;y<ydim;y++) {
		for(z=0;z<zdim;z++) {
		  *i = fabs(g[z][y][x]);
		  ++i;
		}
	      }
	    }
	    delete(gradient);
	  }
	  else derive(du,u1,vz);  
	  if(k<0.0)
	    thresh = quantile(du,quantz);
	  else
	    thresh = k;
	  diffusivity(du,thresh,p);
	  if(ns==AOS_scheme) {
	    AOS(u1,du,dt,d,zdim);
	    i=u1.begin();
	    for(x=0;x<xdim;x++) {
	      for(y=0;y<ydim;y++) {
		for(z=0;z<zdim;z++) {	
		  buf[z][y][x] = *i / d;
		  ++i;
		}
	      }
	    }
	  }
	  else { 
	    explicite(u1,du,dt,zdim);
	    i=u1.begin();
	    for(x=0;x<xdim;x++) {
	      for(y=0;y<ydim;y++) {
		for(z=0;z<zdim;z++) {	
		  buf[z][y][x] += *i;
		  ++i;
		}
	      }
	    }
	  }
	  readMTLVector("anisodiff.tmp1~",u1);
	  remove("anisodiff.tmp1~");
	  i=u1.begin();
	  if(ns==AOS_scheme) {
	    for(z=0;z<zdim;z++) {
	      for(y=0;y<ydim;y++) {
		for(x=0;x<xdim;x++) {
		  buf[z][y][x] += *i / d;
		  ++i;
		}
	      }
	    }
	  }
	  else {
	    for(z=0;z<zdim;z++) {
	      for(y=0;y<ydim;y++) {
		for(x=0;x<xdim;x++) {
		  buf[z][y][x] += *i;
		  ++i;
		}
	      }
	    }
	  }
	  
	  readMTLVector("anisodiff.tmp2~",u1);
	  remove("anisodiff.tmp2~");
	  i=u1.begin();
	  if(ns==AOS_scheme) {
	    for(z=0;z<zdim;z++) {
	      for(x=0;x<xdim;x++) {
		for(y=0;y<ydim;y++) {
		  buf[z][y][x] += *i / d;
		  ++i;
		}
	      }
	    }
	  }
	  else {
	    for(z=0;z<zdim;z++) {
	      for(x=0;x<xdim;x++) {
		for(y=0;y<ydim;y++) {
		  buf[z][y][x] += *i;
		  ++i;
		}
	      }
	    }
	  }
	}
	else {
	  readMTLVector("anisodiff.tmp1~",u1);
	  remove("anisodiff.tmp1~");
	  i=u1.begin();
	  if(ns==AOS_scheme) {
	    for(z=0;z<zdim;z++) {
	      for(y=0;y<ydim;y++) {
		for(x=0;x<xdim;x++) {
		  buf[z][y][x] = *i / d;
		  ++i;
		}
	      }
	    }
	  }
	  else {
	    for(z=0;z<zdim;z++) {
	      for(y=0;y<ydim;y++) {
		for(x=0;x<xdim;x++) {
		  buf[z][y][x] += *i;
		  ++i;
		}
	      }
	    }
	  }
	  
	  readMTLVector("anisodiff.tmp2~",u1);
	  remove("anisodiff.tmp2~");
	  i=u1.begin();
	  if(ns==AOS_scheme) {
	    for(z=0;z<zdim;z++) {
	      for(x=0;x<xdim;x++) {
		for(y=0;y<ydim;y++) { 
		  buf[z][y][x] += *i / d;
		  ++i;
		}    
	      }
	    }
	  }
	  else {
	    for(z=0;z<zdim;z++) {
	      for(x=0;x<xdim;x++) {
		for(y=0;y<ydim;y++) { 
		  buf[z][y][x] += *i;
		  ++i;
		}    
	      }
	    }
	  }
	}
      }
    }
  }
  
  
  //matricial diffusion
  
  else if(td==MATRIX) {
    MTLVector du1s(ms),du2s(ms);
    Inrimage *gradi = new Inrimage(*output);
    Inrimage *dir = new Inrimage(*output);
    for(t=0;t<nbiter;t++) {
      CERR << " " << (unsigned int) ((float) t / (tmax/dt) * 100.0) << " %"  << "\r" ;;
      i=u1.begin();
      for(z=0;z<zdim;z++){
	for(y=0;y<ydim;y++) {
	  for(x=0;x<xdim;x++) {
	      *i = buf[z][y][x];
	      ++i;
	  }
	}
      }
      gradient(output,gradi,dir,(double) sigma,DERICHE_GAUSSIAN,GD_NONE,false,true); 
      Inrimage::WORD_TYPE wt = gradi->getType();
      if (wt != Inrimage::WT_FLOAT) {
	gradi->convertCast(Inrimage::WT_FLOAT);
      }
      float *** gradie = (float ***) gradi->getArray();
      i=us.begin(); 
      for(z=0;z<zdim;z++) {
	for(y=0;y<ydim;y++) {
	  for(x=0;x<xdim;x++) {
	    *i = gradie[z][y][x];
	    ++i;
	  }
	}
      }
      
      if(k<0.0)
	thresh = quantile(us,quantx);
      else
	thresh = k;
     
      Inrimage * gradientx  = new Inrimage(*output);
      recursiveFilter(gradientx,DERICHE_GAUSSIAN,DERIVATIVE_X,sigma,false,false);
      recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_Y,sigma,false,false);
      recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_Z,sigma,false,false);
      
      float *** gx = (float ***) gradientx->getArray();
      i=du1s.begin();
      for(z=0;z<zdim;z++) {
	for(y=0;y<ydim;y++) {
	  for(x=0;x<xdim;x++) {
	    *i = gx[z][y][x];
	    ++i;
	  }
	}
      }
     
      Inrimage * gradienty  = new Inrimage(*output);
      recursiveFilter(gradienty,DERICHE_GAUSSIAN,DERIVATIVE_Y,sigma,false,false);
      recursiveFilter(gradienty,DERICHE_GAUSSIAN,SMOOTH_X,sigma,false,false);
      recursiveFilter(gradienty,DERICHE_GAUSSIAN,SMOOTH_Z,sigma,false,false);
      
      float *** gy = (float ***) gradienty->getArray();
      i=du2s.begin();
      for(z=0;z<zdim;z++) {
	for(y=0;y<ydim;y++) {
	  for(x=0;x<xdim;x++) {
	    *i = gy[z][y][x];
	    ++i;
	  }
	}
      }
     
      MTLVector du3s(ms);
      if(d==3.0) {
	
	Inrimage * gradientz  = new Inrimage(*output);
	recursiveFilter(gradientz,DERICHE_GAUSSIAN,DERIVATIVE_Z,sigma,false,false);
	recursiveFilter(gradientz,DERICHE_GAUSSIAN,SMOOTH_X,sigma,false,false);
	recursiveFilter(gradientz,DERICHE_GAUSSIAN,SMOOTH_Y,sigma,false,false);
      
	float *** gz = (float ***) gradienty->getArray();
	i=du3s.begin();
	for(z=0;z<zdim;z++) {
	  for(y=0;y<ydim;y++) {
	    for(x=0;x<xdim;x++) {
	      *i = gz[z][y][x];
	      ++i;
	      }
	  }
	}
	delete(gradientz);
      }
      delete(gradientx);
      delete(gradienty);
      if(d==2.0) {
	explicitematrice(u1,us,du1s,du2s,dt,xdim,ydim,k,p); 
      }
      else if(d==3.0) {
	explicitematrice3D(u1,us,du1s,du2s,du3s,dt,xdim,ydim,k,p);
      }
      
      i=us.begin();
      for(z=0;z<zdim;z++) {
	for(y=0;y<ydim;y++) {
	  for(x=0;x<xdim;x++) {
	    buf[z][y][x] = *i;
	    ++i;
	  }    
	}
      }
      
      }
    }

  // Krissian flow diffusion
  else if(td==KRISSIAN)
    { 
      MTLVector  du1s(ms), du2s(ms), du3s(ms), du11s(ms), du12s(ms), du22s(ms), du13s(ms), du23s(ms), du33s(ms);
  MTLVecIt i;
 
  if(zdim>1)
    d=3.0;
 
  
 
   
    
    for(t=0;t<nbiter;t++) {
      
      CERR << " " << (unsigned int) ((float) t / (tmax/dt) * 100.0) << " %"  << "\r" ;

      i=u1.begin();
       
      for(z=0;z<zdim;z++){
	for(y=0;y<ydim;y++) {
	  for(x=0;x<xdim;x++) {
	    *i = buf[z][y][x];
	    ++i;
	  }
	}
      }
      
      
      
      Inrimage *gradi = new Inrimage(*output);
      Inrimage *dir = new Inrimage(*output);
      
      gradient(output,gradi,dir,(double) sigma,DERICHE_GAUSSIAN,GD_NONE,false,true); 
      Inrimage::WORD_TYPE wt = gradi->getType();
      if (wt != Inrimage::WT_FLOAT) {
	gradi->convertCast(Inrimage::WT_FLOAT);
      }
      float *** grad = (float ***) gradi->getArray();
      
      i=us.begin();
       
      for(z=0;z<zdim;z++){
	for(y=0;y<ydim;y++) {
	  for(x=0;x<xdim;x++) {
	    *i = grad[z][y][x];
	    ++i;
	}
	}
      }
      
      
     
      if(k<0.0)
	thresh=quantile(us,quantx);
      else
	thresh=k;
      
     
      delete(gradi);

      //calculus of the first derivatives
      Inrimage *gradientx=new Inrimage(*output);
      recursiveFilter(gradientx,DERICHE_GAUSSIAN,DERIVATIVE_X,sigma,false,false);
      recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_Y,sigma,false,false);
      recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_Z,sigma,false,false);
      float *** gx = (float ***) gradientx->getArray();
      i=du1s.begin();
      for(z=0;z<zdim;z++) {
	for(y=0;y<ydim;y++) {
	  for(x=0;x<xdim;x++) {
	    *i = gx[z][y][x];
	    ++i;
	  }
	}
      }
      delete(gradientx);
      gradientx=new Inrimage(*output);
      recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_X,sigma,false,false);
      recursiveFilter(gradientx,DERICHE_GAUSSIAN,DERIVATIVE_Y,sigma,false,false);
      recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_Z,sigma,false,false);
      gx = (float ***) gradientx->getArray();
      i=du2s.begin();
      for(z=0;z<zdim;z++) {
	for(y=0;y<ydim;y++) {
	  for(x=0;x<xdim;x++) {
	    *i = gx[z][y][x];
	    ++i;
	  }
	}
      }
      delete(gradientx);
      
      if(d==3.0)
	{
	  gradientx=new Inrimage(*output);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_X,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_Y,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,DERIVATIVE_Z,sigma,false,false);
	  gx = (float ***) gradientx->getArray();
	  i=du3s.begin();
	  for(z=0;z<zdim;z++) {
	    for(y=0;y<ydim;y++) {
	      for(x=0;x<xdim;x++) {
		*i = gx[z][y][x];
		++i;
	      }
	    }
	  }
	  delete(gradientx);
	  
	}

      if(d==3.0)
	{
	  //calculus of the second derivatives
	  gradientx=new Inrimage(*output);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SECOND_DERIVATIVE_X,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_Y,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_Z,sigma,false,false);
	  gx = (float ***) gradientx->getArray();
	  i=du11s.begin();
	  for(z=0;z<zdim;z++) {
	    for(y=0;y<ydim;y++) {
	      for(x=0;x<xdim;x++) {
		*i = gx[z][y][x];
		++i;
	      }
	    }
	  }
	  delete(gradientx);
	  gradientx=new Inrimage(*output);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,DERIVATIVE_X,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,DERIVATIVE_Y,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_Z,sigma,false,false);
	  gx = (float ***) gradientx->getArray();
	  i=du12s.begin();
	  for(z=0;z<zdim;z++) {
	    for(y=0;y<ydim;y++) {
	      for(x=0;x<xdim;x++) {
		*i = gx[z][y][x];
		++i;
	      }
	    }
	  }
	  delete(gradientx);
	  gradientx=new Inrimage(*output);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,DERIVATIVE_X,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_Y,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,DERIVATIVE_Z,sigma,false,false);
	  gx = (float ***) gradientx->getArray();
	  i=du13s.begin();
	  for(z=0;z<zdim;z++) {
	    for(y=0;y<ydim;y++) {
	      for(x=0;x<xdim;x++) {
		*i = gx[z][y][x];
	    ++i;
	      }
	    }
	  }
	  delete(gradientx);
	  gradientx=new Inrimage(*output);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_X,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SECOND_DERIVATIVE_Y,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_Z,sigma,false,false);
	  gx = (float ***) gradientx->getArray();
	  i=du22s.begin();
	  for(z=0;z<zdim;z++) {
	    for(y=0;y<ydim;y++) {
	      for(x=0;x<xdim;x++) {
		*i = gx[z][y][x];
	    ++i;
	      }
	    }
	  }
	  delete(gradientx);
	  gradientx=new Inrimage(*output);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_X,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,DERIVATIVE_Y,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,DERIVATIVE_Z,sigma,false,false);
	  gx = (float ***) gradientx->getArray();
	  i=du23s.begin();
	  for(z=0;z<zdim;z++) {
	    for(y=0;y<ydim;y++) {
	      for(x=0;x<xdim;x++) {
		*i = gx[z][y][x];
	    ++i;
	      }
	    }
	  }
	  delete(gradientx);
	  gradientx=new Inrimage(*output);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_X,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_Y,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SECOND_DERIVATIVE_Z,sigma,false,false);
	  gx = (float ***) gradientx->getArray();
	  i=du33s.begin();
	  for(z=0;z<zdim;z++) {
	    for(y=0;y<ydim;y++) {
	      for(x=0;x<xdim;x++) {
		*i = gx[z][y][x];
		++i;
	      }
	    }
	  }
	  delete(gradientx);


	}
      
      

      
       
       expliciteKrissian(u1,us,du1s,du2s,du3s,du11s,du12s,du13s,du22s,du23s,du33s,dt,k,p,xdim,ydim,zdim);
       
       
       /* On remplie buf */

       i=us.begin();
       for(z=0;z<zdim;z++){
	 for(y=0;y<ydim;y++){
	   for(x=0;x<xdim;x++){
	     buf[z][y][x]=*i;
	     ++i;
	   }
	 }
       }
    
      
     
    }

   
	

    

    i=us.begin();
    for(z=0;z<zdim;z++) {
      for(y=0;y<ydim;y++) {
	for(x=0;x<xdim;x++) {
	  buf[z][y][x] = *i;
	  ++i;
	}    
      }
    }
    
    
    }
else if(td==MATRIX_KRISSIAN)
  {
 
  MTLVector  du1s(ms), du2s(ms), du3s(ms),du11s(ms), du12s(ms), du13s(ms),du22s(ms), du23s(ms), du33s(ms) ;
 
  if(zdim>1)
    d=3.0;
 
  
 
   
    
    for(t=0;t<nbiter;t++) {
      
      CERR << " " << (unsigned int) ((float) t / (tmax/dt) * 100.0) << " %"  << ENDL;
      
      i=u1.begin();
       
      for(z=0;z<zdim;z++){
	for(y=0;y<ydim;y++) {
	  for(x=0;x<xdim;x++) {
	    *i = buf[z][y][x];
	    ++i;
	  }
	}
      }
      
      
      
      
      Inrimage *gradi = new Inrimage(*output);
      Inrimage *dir = new Inrimage(*output);
      gradient(output,gradi,dir,(double) sigma,DERICHE_GAUSSIAN,GD_NONE,false,true); 
      Inrimage::WORD_TYPE wt = gradi->getType();
      if (wt != Inrimage::WT_FLOAT) {
	gradi->convertCast(Inrimage::WT_FLOAT);
      }
      float *** gradie = (float ***) gradi->getArray();
      i=us.begin(); 
      for(z=0;z<zdim;z++) {
	for(y=0;y<ydim;y++) {
	  for(x=0;x<xdim;x++) {
	    *i = gradie[z][y][x];
	    ++i;
	  }
	}
      }
      
      if(k<0.0)
	thresh=quantile(us,quantx);
      else
	thresh=k;
     
      delete(gradi);

      //calculus of the first derivatives
      Inrimage *gradientx=new Inrimage(*output);
      recursiveFilter(gradientx,DERICHE_GAUSSIAN,DERIVATIVE_X,sigma,false,false);
      recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_Y,sigma,false,false);
      recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_Z,sigma,false,false);
      float *** gx = (float ***) gradientx->getArray();
      i=du1s.begin();
      for(z=0;z<zdim;z++) {
	for(y=0;y<ydim;y++) {
	  for(x=0;x<xdim;x++) {
	    *i = gx[z][y][x];
	    ++i;
	  }
	}
      }
      delete(gradientx);
      gradientx=new Inrimage(*output);
      recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_X,sigma,false,false);
      recursiveFilter(gradientx,DERICHE_GAUSSIAN,DERIVATIVE_Y,sigma,false,false);
      recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_Z,sigma,false,false);
      gx = (float ***) gradientx->getArray();
      i=du2s.begin();
      for(z=0;z<zdim;z++) {
	for(y=0;y<ydim;y++) {
	  for(x=0;x<xdim;x++) {
	    *i = gx[z][y][x];
	    ++i;
	  }
	}
      }
      delete(gradientx);
      
      if(d==3.0)
	{
	  gradientx=new Inrimage(*output);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_X,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_Y,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,DERIVATIVE_Z,sigma,false,false);
	  gx = (float ***) gradientx->getArray();
	  i=du3s.begin();
	  for(z=0;z<zdim;z++) {
	    for(y=0;y<ydim;y++) {
	      for(x=0;x<xdim;x++) {
		*i = gx[z][y][x];
		++i;
	      }
	    }
	  }
	  delete(gradientx);
	  
	}

      if(d==3.0)
	{
	  //calculus of the second derivatives
	  gradientx=new Inrimage(*output);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SECOND_DERIVATIVE_X,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_Y,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_Z,sigma,false,false);
	  gx = (float ***) gradientx->getArray();
	  i=du11s.begin();
	  for(z=0;z<zdim;z++) {
	    for(y=0;y<ydim;y++) {
	      for(x=0;x<xdim;x++) {
		*i = gx[z][y][x];
		++i;
	      }
	    }
	  }
	  delete(gradientx);
	  gradientx=new Inrimage(*output);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,DERIVATIVE_X,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,DERIVATIVE_Y,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_Z,sigma,false,false);
	  gx = (float ***) gradientx->getArray();
	  i=du12s.begin();
	  for(z=0;z<zdim;z++) {
	    for(y=0;y<ydim;y++) {
	      for(x=0;x<xdim;x++) {
		*i = gx[z][y][x];
		++i;
	      }
	    }
	  }
	  delete(gradientx);
	  gradientx=new Inrimage(*output);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,DERIVATIVE_X,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_Y,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,DERIVATIVE_Z,sigma,false,false);
	  gx = (float ***) gradientx->getArray();
	  i=du13s.begin();
	  for(z=0;z<zdim;z++) {
	    for(y=0;y<ydim;y++) {
	      for(x=0;x<xdim;x++) {
		*i = gx[z][y][x];
	    ++i;
	      }
	    }
	  }
	  delete(gradientx);
	  gradientx=new Inrimage(*output);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_X,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SECOND_DERIVATIVE_Y,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_Z,sigma,false,false);
	  gx = (float ***) gradientx->getArray();
	  i=du22s.begin();
	  for(z=0;z<zdim;z++) {
	    for(y=0;y<ydim;y++) {
	      for(x=0;x<xdim;x++) {
		*i = gx[z][y][x];
	    ++i;
	      }
	    }
	  }
	  delete(gradientx);
	  gradientx=new Inrimage(*output);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_X,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,DERIVATIVE_Y,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,DERIVATIVE_Z,sigma,false,false);
	  gx = (float ***) gradientx->getArray();
	  i=du23s.begin();
	  for(z=0;z<zdim;z++) {
	    for(y=0;y<ydim;y++) {
	      for(x=0;x<xdim;x++) {
		*i = gx[z][y][x];
	    ++i;
	      }
	    }
	  }
	  delete(gradientx);
	  gradientx=new Inrimage(*output);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_X,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SMOOTH_Y,sigma,false,false);
	  recursiveFilter(gradientx,DERICHE_GAUSSIAN,SECOND_DERIVATIVE_Z,sigma,false,false);
	  gx = (float ***) gradientx->getArray();
	  i=du33s.begin();
	  for(z=0;z<zdim;z++) {
	    for(y=0;y<ydim;y++) {
	      for(x=0;x<xdim;x++) {
		*i = gx[z][y][x];
		++i;
	      }
	    }
	  }
	  delete(gradientx);


	}
      
      if(d==2.0)
	{
	  explicitematKrissian(u1,us,du1s,du2s,dt,xdim,ydim,k,p); 
	}
      else if(d==3.0)
	{
	  explicite3DmatKrissian(u1,us,du1s,du2s,du3s,du11s,du12s,du13s,du22s,du23s,du33s,dt,xdim,ydim,k,p);
	}
	

      i=us.begin();
      for(z=0;z<zdim;z++) {
	for(y=0;y<ydim;y++) {
	  for(x=0;x<xdim;x++) {
	    buf[z][y][x] = *i;
	    ++i;
	  }    
	}
      }
    
    }

  

  }

  if(wt != Inrimage::WT_FLOAT) {
    output->convertCast(wt);
  }
    CERR << "Fini." << ENDL;     
}

// Computation of a quantile of the cumulated gradient histogram
float quantile(const MTLVector & u, const float & quant) {
  unsigned int j, ms = u.size();
  float p = 0.0;
  MTLVector o(256);
  MTLConstVecIt i;
  
  for(i=u.begin();i!=u.end();++i) {
    j = (unsigned int) floor(*i + 0.5);
    if(j>255) o[255]++;
    else o[j]++;
  }
  j=0;
  i=o.begin();
  double m = (double) ms;
  while((p*100/m)<quant) {
    p += *i;
    ++i;
    j++;
  }
  return((float) j);
}

// Diffusivity functions
void diffusivity(MTLVector & u, const float & seuil, const DIFFUSION_FUNCTION & p) {
  MTLVecIt i=u.begin();
  
  if(p==PERONA_MALIK) {
    for(;i!=u.end();++i) {
      if(*i!=0.0)
	*i = 1.0 - exp(-1.0/((*i/seuil)*(*i/seuil)));
      else *i = 1.0;
    }
  }
  
  else if(p==WEICKERT) {
    for(;i!=u.end();++i) {
      if(*i!=0.0)
	*i = 1.0 - exp(-3.315/((*i/seuil)*(*i/seuil)*(*i/seuil)*(*i/seuil)));
      else *i = 1.0;
    }
  }
  
  else if(p==AUBERT) {
    for(;i!=u.end();++i) {
      if(*i!=0.0)
	*i = 1.0 / (sqrt(1.0 + (*i/seuil) * (*i/seuil)));
      else *i = 1.0;
    }
  }
  
  else if(p==GREEN) {
    for(;i!=u.end();++i) {
      if(*i!=0.0)
	*i = tanh(*i/seuil)/(*i/seuil);
      else *i = 1.0;
    }
  }

  else if(p==PERONA_MALIK_2){
     for(;i!=u.end();++i)
       *i = exp(-(*i/seuil)*(*i/seuil));
  }
  
  else if(p==PERONA_MALIK_3)
   for(;i!=u.end();++i)
       *i = 1.0 / (1.0 + (*i/seuil)*(*i/seuil));
  
}

// derivation by finite differences, absolute value of the result
void derive(MTLVector & du, const MTLVector & u, const double& h) {
  unsigned int ms = u.size();
  du[0] = fabs(u[1]-u[0])*h;
  du[ms-1] = fabs(u[ms-1]-u[ms-2])*h;

  MTLConstVecIt i = u.begin() + 1;
  MTLVecIt j = du.begin() + 1;
  double s = h*0.5;
  for(;i!=u.end()-1;++i,++j)
    *j = fabs((*(i+1)-*(i-1)))*s;
}

// AOS Numerical scheme for fast semi-implicit computation
void AOS(MTLVector & r, MTLVector & g, const float & dt, const float & d, const unsigned int dim) {
  unsigned int i, ms = r.size();
  MTLBanded Ax(ms,ms, 1, 1);
  MTLVector L(ms-1), R(ms);
  MTLVecIt itg;
  unsigned int k = (unsigned int) (((float) ms)/((float) dim));
  
  // The Thomas Algorithm for triband matrix inversion
  
  //make Ax
  Ax(0,0) = - (g[0] + g[1]) / 2.0;
  Ax(0,1) =   (g[0] + g[1]) / 2.0;
  Ax(ms-1,ms-1) = -(g[ms-1] + g[ms-2]) / 2.0;
  Ax(ms-1,ms-2) =  (g[ms-1] + g[ms-2]) / 2.0;

  for(i=1,itg=g.begin()+1;itg!=g.end()-1;++itg,++i) {
    Ax(i,i) = - *itg - (*(itg+1) + *(itg-1)) / 2.0;
    Ax(i,i+1) = (*itg + *(itg+1)) / 2.0;
    Ax(i,i-1) = (*itg + *(itg-1)) / 2.0;
  }

  for(i=1;i<k;++i) {
    Ax(i*dim-1,i*dim-1) = - (g[i*dim-1] + g[i*dim-2]) / 2.0;
    Ax(i*dim-1,i*dim) = 0.0;
    Ax(i*dim,i*dim) = - (g[i*dim] + g[i*dim+1]) / 2.0;
    Ax(i*dim,i*dim-1) = 0.0;
  }
  
  //make Dx
  mtl::scale(Ax,-d*dt);
  for(i=0;i<ms;++i)
    Ax(i,i) += 1.0;
  
  //make L and R
  R[0]=Ax(0,0);

  for(i=1;i<ms-1;++i) {
    L[i-1] = Ax(i,i-1) / R[i-1];
    R[i] = Ax(i,i) - L[i-1] * Ax(i-1,i); 
  }
  R[ms-1] = Ax(ms-1,ms-1) - L[ms-3]*Ax(ms-3,ms-2);
  L[ms-2] = Ax(ms-1,ms-2)/R[ms-1];

  //make intermediary result
  g[0]=r[0];
  for(i=1,itg=g.begin()+1;itg!=g.end();++itg,++i)
    *itg = r[i] - L[i-1] * (*(itg-1));
  
  // Result :
  //make r
  r[ms-1] = g[ms-1] / R[ms-1];
  for(i=ms-1;i>0;i--) {
    r[i-1] = (g[i-1] - Ax(i-1,i) * r[i]) / R[i-1];
  }
}

 // Explicit Scheme
void explicite(MTLVector & r, MTLVector & g, const float & dt, const unsigned int dim) {
  unsigned int i, ms = r.size();
  MTLBanded Ax(ms,ms, 1, 1);
  MTLVecIt itg;
  unsigned int k = (unsigned int) (((float) ms)/((float) dim));
  
  
  
  //make Ax
    Ax(0,0) = - (g[0] + g[1]) / 2.0;
    Ax(0,1) =   (g[0] + g[1]) / 2.0;
    Ax(ms-1,ms-1) = -(g[ms-1] + g[ms-2]) / 2.0;
    Ax(ms-1,ms-2) =  (g[ms-1] + g[ms-2]) / 2.0;
    
    i=1;
    for(itg=g.begin()+1;itg!=g.end()-1;++itg) {
      Ax(i,i) = - *itg - (*(itg+1) + *(itg-1)) / 2.0;
      Ax(i,i+1) = (*itg + *(itg+1)) / 2.0;
      Ax(i,i-1) = (*itg + *(itg-1)) / 2.0;
      i++;
    }
    
    
    for(i=1;i<k;i++) {
      Ax(i*dim-1,i*dim-1) = - (g[i*dim-1] + g[i*dim-2]) / 2.0;
      Ax(i*dim-1,i*dim) = 0.0;
      Ax(i*dim,i*dim) = - (g[i*dim] + g[i*dim+1]) / 2.0;
      Ax(i*dim,i*dim-1) = 0.0;
    }
    
    
  //Result
    float temp1;
    float temp=r[0];
    r[0]=dt*(Ax(0,0)*r[0]+Ax(0,1)*r[1]);
    for(i=1;i<ms-1;i++)
      {
	temp1=temp;
	temp=r[i];
      r[i]=dt*(Ax(i,i-1)*temp1+Ax(i,i)*r[i]+Ax(i,i+1)*r[i+1]);
      }
    r[ms-1]=dt*(Ax(ms-1,ms-2)*temp+Ax(ms-1,ms-1)*r[ms-1]);
  }



  //diffusivity function at one point
float diffusion(const float & u, const float & seuil, const DIFFUSION_FUNCTION & p){
  if(p==PERONA_MALIK) {
    if(u==0.0)
      return(1.0);
    else
      return(1.0 - exp(-1.0/((u/seuil)*(u/seuil))));
  }
  
  else if(p==WEICKERT) {
    if(u==0.0)
      return(1.0);
    else
      return(1.0 - exp(-3.315/((u/seuil)*(u/seuil)*(u/seuil)*(u/seuil))));
  }
  
  else if(p==AUBERT) 
    return(1.0 / (sqrt(1.0+(u/seuil)*(u/seuil))));
  
  else if(p==GREEN) {
    if(u==0.0)
      return(1.0);
    else
      return(tanh(u/seuil)/(u/seuil));
  }
  
  else if(p==PERONA_MALIK_2)
    return(exp(-(u/seuil)*(u/seuil)));
  
  else if(p==PERONA_MALIK_3)
    return(1.0 / (1.0 + (u/seuil)*(u/seuil)));
  
  else return(0.0); 
}
  

// calcul de la matrice D de diffusion en chaque point (2D)
void calculD(const MTLVector & du1s, const MTLVector & du2s, MTLVector & d11,  MTLVector & d12, MTLVector &d21, MTLVector &d22,const float & quant, const  DIFFUSION_FUNCTION & p){
    unsigned int i,ms=du1s.size();
    float norme, lambda;
    
 for(i=0;i<ms;i++)
   {
     norme=(du1s[i]*du1s[i]+du2s[i]*du2s[i]);
     if (norme==0.0)
      {
	d11[i]=1.0;
	d12[i]=0.0;
	d21[i]=0.0;
	d22[i]=1.0;
      }
    else
      {
	lambda=diffusion(sqrt(norme),quant,p);
	d11[i]=lambda*du1s[i]*du1s[i]/norme+du2s[i]*du2s[i]/norme;
	d12[i]=(lambda-1)*du1s[i]*du2s[i]/norme;
	d21[i]=d12[i];
	d22[i]=lambda*du2s[i]*du2s[i]/norme+du1s[i]*du1s[i]/norme;
      }
   }
}

// calcul de la matrice D de diffusion en chaque point (3D)
void calculD3D(const MTLVector & du1s, const MTLVector & du2s, const MTLVector & du3s, MTLVector & d11,  MTLVector & d12, MTLVector & d13,  MTLVector &d21, MTLVector &d22, MTLVector &d23, MTLVector &d31, MTLVector &d32,MTLVector &d33,const float & quant, const  DIFFUSION_FUNCTION & p){
  unsigned int i,ms=du1s.size();
  float norme, lambda;
  float a,b,c;
  Matrix<3,float> P;
  Matrix<3,float> D;
  Matrix<3,float> T;
  Matrix<3,float> temp;
  
  for(i=0;i<ms;i++)
   {
     norme=(du1s[i]*du1s[i]+du2s[i]*du2s[i]+du3s[i]*du3s[i]);
     if (norme==0.0)
      {
	d11[i]=1.0;
	d12[i]=0.0;
	d13[i]=0.0;
	d21[i]=0.0;
	d22[i]=1.0;
	d23[i]=0.0;
	d31[i]=0.0;
	d32[i]=0.0;
	d33[i]=1.0;
      }
    else
      {
	lambda=diffusion(sqrt(norme),quant,p);
	a=du1s[i];
	b=du2s[i];
	c=du3s[i];
	if((a==0)&(b==0))
	  {
	    d11[i]=lambda;
	    d12[i]=0;
	    d13[i]=0;
	    d21[i]=0;
	    d22[i]=1;
	    d23[i]=0;
	    d31[i]=0;
	    d32[i]=0;
	    d33[i]=1;
	  }
	else
	  {
	    P(0,0)=a/sqrt(norme);
	    P(1,0)=b/sqrt(norme);
	    P(2,0)=c/sqrt(norme);
	    P(0,1)=-b/sqrt(a*a+b*b);
	    P(1,1)=a/sqrt(a*a+b*b);
	    P(2,1)=0;
	    P(0,2)=a*c/sqrt(a*c*a*c+b*c*b*c+(b*b+a*a)*(b*b+a*a));
	    P(1,2)=b*c/sqrt(a*c*a*c+b*c*b*c+(b*b+a*a)*(b*b+a*a));
	    P(2,2)=-(a*a+b*b)/sqrt(a*c*a*c+b*c*b*c+(b*b+a*a)*(b*b+a*a));
	    D(0,0)=lambda;
	    D(1,0)=0;
	    D(2,0)=0;
	    D(0,1)=0;
	    D(1,1)=1;
	    D(2,1)=0;
	    D(0,2)=0;
	    D(1,2)=0;
	    D(2,2)=1;
	    temp.Mul(D,P);
	    T.Trp(P);
	    temp.Mul(T,temp);
	    d11[i]=temp(0,0);
	    d12[i]=temp(0,1);
	    d13[i]=temp(0,2);
	    d21[i]=temp(1,0);
	    d22[i]=temp(1,1);
	    d23[i]=temp(1,2);
	    d31[i]=temp(2,0);
	    d32[i]=temp(2,1);
	    d33[i]=temp(2,2);
	  }
      }
   }
}




// schema explicite en 2 dimensions expression  matricielle

void explicitematrice(const MTLVector & u, MTLVector & r , const MTLVector & du1s, const MTLVector & du2s,const float & dt,const unsigned int xdim,const unsigned int ydim, const float quant, const DIFFUSION_FUNCTION & p) {
  unsigned int i, ms = r.size();
  MTLVector d11(ms),d12(ms),d21(ms),d22(ms);
  
  calculD(du1s,du2s,d11,d12,d21,d22,quant,p);

  for(i=0;i<ms;i++)
    r[i]=u[i];
  
  
   
  r[0]+=dt*(d11[0]*0.5*(u[1]-u[0])+d12[0]*(0.25*(u[0]-u[xdim])+0.5*0.25*(u[1+xdim]+u[xdim]-u[1]-u[0]))+d22[0]*0.5*(u[xdim]-u[0])+d21[0]*(0.25*(u[0]-u[1])+0.5*0.25*(u[1+xdim]+u[1]-u[xdim]-u[0]))+d11[1]*0.5*(u[1]-u[0])+d12[1]*0.25*0.5*(u[1+xdim]+u[xdim]-u[0]-u[1])+d22[xdim]*0.5*(u[xdim]-u[0])+d21[xdim]*0.25*(u[1+xdim]+u[1]-u[0]-u[xdim]));
  

  r[ms-1]+=dt*(-d11[ms-1]*0.5*(u[ms-1]-u[ms-2])+d12[ms-1]*0.25*(u[ms-1]-u[ms-1-xdim])-0.5*0.25*d12[ms-1]*(u[ms-2]+u[ms-1]-u[ms-2-xdim]-u[ms-1-xdim])-d22[ms-1]*0.5*(u[ms-1]-u[ms-xdim-1])+d21[ms-1]*0.25*(u[ms-1]-u[ms-2])-0.5*0.25*d21[ms-1]*(u[ms-1-xdim]+u[ms-1]-u[ms-2-xdim]-u[ms-2])-d11[ms-2]*0.5*(u[ms-1]-u[ms-2])-d12[ms-2]*0.5*0.25*(u[ms-2]+u[ms-1]-u[ms-2-xdim]-u[ms-1-xdim])-d22[ms-1-xdim]*0.5*(u[ms-1]-u[ms-1-xdim])-d21[ms-xdim-1]*0.5*0.25*(u[ms-1-xdim]+u[ms-1]-u[ms-2-xdim]-u[ms-2]));
  

  for(i=1;i<xdim;i++)
  {
    r[i]+=dt*( 0.5 * d11[i]*(u[i+1]-2*u[i]+u[i-1])+d12[i]*0.25*0.5*(u[i+1+xdim]+u[i-1]-u[i+1]-u[i-1+xdim])+0.5*d11[i+1]*(u[i+1]-u[i])+d12[i+1]*0.25*0.5*(u[i+xdim+1]+u[i+xdim]-u[i+1]-u[i])-d11[i-1]*(u[i]-u[i-1])-d12[i-1]*0.25*0.5*(u[i-1+xdim]+u[i+xdim]-u[i-1]-u[i])+ 0.5*d22[i+xdim]*(u[i+xdim]-u[i])+d21[i+xdim]*0.25*0.5*(u[i+1+xdim]+u[i+1]-u[i-1]-u[i-1+xdim])+0.5*d22[i]*(u[i+xdim]-u[i])-d21[i]*0.25*(u[i+1]-u[i-1])+d21[i]*0.5*0.25*(u[i+1+xdim]+u[i+1]-u[i-1+xdim]-u[i-1]));
     
  }
 
    for(i=ms-xdim;i<ms-1;i++)
  {
    r[i]+=dt*(0.5*d11[i]*(u[i+1]-2*u[i]+u[i-1])+d12[i]*0.5*0.25*(u[i+1]+u[i-1-xdim]-u[i+1-xdim]-u[i-1])+0.5*d11[i+1]*(u[i+1]-u[i])+d12[i+1]*0.5*0.25*(u[i+1]+u[i]-u[i+1-xdim]-u[i-xdim])-0.5*d11[i-1]*(u[i]-u[i-1])-d12[i-1]*0.5*0.25*(u[i-1]+u[i]-u[i-1-xdim]-u[i-xdim])-0.5*d22[i]*(u[i]-u[i-xdim])+d21[i]*0.25*(u[i+1]-u[i-1])-0.5*0.25*d21[i]*(u[i+1-xdim]+u[i+1]-u[i-1]-u[i-1-xdim])-0.5*d22[i-xdim]*(u[i]-u[i-xdim])-d21[i-xdim]*0.5*0.25*(u[i-xdim+1]+u[i+1]-u[i-xdim-1]-u[i-1]));
 
  }
    
  for(i=xdim;i<ms-xdim;i++){
    r[i]+= dt * 0.5 * (d11[i]*(u[i+1]-2*u[i]+u[i-1])+d12[i]*0.25*(u[i+1+xdim]+u[i-1-xdim]-u[i+1-xdim]-u[i-1+xdim])+d22[i]*(u[i+xdim]-2*u[i]+u[i-xdim])+d21[i]*0.25*(u[i+1+xdim]+u[i-1-xdim]-u[i+1-xdim]-u[i-1+xdim]) + d11[i+1]*(u[i+1]-u[i])+d12[i+1]*0.25*(u[i+1+xdim]+u[i+xdim]-u[i+1-xdim]-u[i-xdim]) + d11[i-1]*(u[i-1]-u[i])+d12[i-1]*0.25*(-u[i+xdim]-u[i-1+xdim]+u[i-1-xdim]+u[i-xdim]) + d22[i+xdim]*(u[i+xdim]-u[i])+d21[i+xdim]*0.25*(u[i+1+xdim]+u[i+1]-u[i+xdim-1]-u[i-1]) + d22[i-xdim]*(u[i-xdim]-u[i])+d21[i-xdim]*0.25*(-u[i+1-xdim]-u[i+1]+u[i-xdim-1]+u[i-1]));
    
  }
 
}


// schema explicite en 3 dimensions expression  matricielle

void explicitematrice3D(const MTLVector & u, MTLVector & r , const MTLVector & du1s, const MTLVector & du2s, const MTLVector & du3s , const float & dt,const unsigned int xdim,const unsigned int ydim, const float quant, const DIFFUSION_FUNCTION & p) {
  unsigned int i, ms = r.size();
  MTLVector d11(ms),d12(ms),d13(ms),d21(ms),d22(ms),d23(ms),d31(ms),d32(ms),d33(ms);
  


  calculD3D(du1s,du2s,du3s,d11,d12,d13,d21,d22,d23,d31,d32,d33,quant,p);
  for(i=0;i<ms;i++)
    r[i]=u[i];

r[0]+=dt*(d11[0]*0.5*(u[1]-u[0])+d12[0]*(0.25*(u[0]-u[xdim])+0.5*0.25*(u[1+xdim]+u[xdim]-u[1]-u[0]))+d13[0]*(0.25*(u[0]-u[xdim*ydim])+0.5*0.25*(u[1+xdim*ydim]+u[xdim*ydim]-u[1]-u[0]))+d22[0]*0.5*(u[xdim]-u[0])+d21[0]*(0.25*(u[0]-u[1])+0.5*0.25*(u[1+xdim]+u[1]-u[xdim]-u[0]))+d23[0]*(0.25*(u[0]-u[xdim*ydim])+0.5*0.25*(u[xdim*ydim+xdim]+u[xdim*ydim]-u[xdim]-u[0]))+d11[1]*0.5*(u[1]-u[0])+d12[1]*0.25*0.5*(u[1+xdim]+u[xdim]-u[0]-u[1])+d13[1]*0.25*0.5*(u[1+xdim*ydim]+u[xdim*ydim]-u[0]-u[1])+d22[xdim]*0.5*(u[xdim]-u[0])+d21[xdim]*0.25*(u[1+xdim]+u[1]-u[0]-u[xdim])+d23[xdim]*0.25*(u[xdim*ydim+xdim]+u[xdim*ydim]-u[0]-u[xdim])+d33[0]*0.5*(u[xdim*ydim]-u[0])+d31[0]*(0.25*(u[0]-u[1])+0.5*0.25*(u[1+xdim*ydim]+u[1]-u[xdim*ydim]-u[0]))+d32[0]*0.25*(u[0]-u[xdim])+d33[xdim*ydim]*0.5*(u[xdim*ydim]-u[0])+d31[xdim*ydim]*0.25*(u[1+xdim*ydim]+u[1]-u[0]-u[xdim*ydim])+d32[xdim*ydim]*0.25*(u[xdim*ydim+xdim]+u[xdim]-u[0]-u[xdim*ydim]));

r[ms-1]+=dt*(-d11[ms-1]*0.5*(u[ms-1]-u[ms-2])+d12[ms-1]*0.25*(u[ms-1]-u[ms-1-xdim])-0.5*0.25*d12[ms-1]*(u[ms-2]+u[ms-1]-u[ms-2-xdim]-u[ms-1-xdim])+d13[ms-1]*0.25*(u[ms-1]-u[ms-1-xdim*ydim])-0.5*0.25*d13[ms-1]*(u[ms-2]+u[ms-1]-u[ms-2-xdim*ydim]-u[ms-1-xdim*ydim])-d22[ms-1]*0.5*(u[ms-1]-u[ms-xdim-1])+d21[ms-1]*0.25*(u[ms-1]-u[ms-2])-0.5*0.25*d21[ms-1]*(u[ms-1-xdim]+u[ms-1]-u[ms-2-xdim]-u[ms-2])+d23[ms-1]*0.25*(u[ms-1]-u[ms-1-xdim*ydim])-0.5*0.25*d23[ms-1]*(u[ms-1-xdim]+u[ms-1]-u[ms-1-xdim-xdim*ydim]-u[ms-1-xdim*ydim])-d11[ms-2]*0.5*(u[ms-1]-u[ms-2])-d12[ms-2]*0.5*0.25*(u[ms-2]+u[ms-1]-u[ms-2-xdim]-u[ms-1-xdim])-d13[ms-2]*0.5*0.25*(u[ms-2]+u[ms-1]-u[ms-2-xdim*ydim]-u[ms-1-xdim*ydim])-d22[ms-1-xdim]*0.5*(u[ms-1]-u[ms-1-xdim])-d21[ms-xdim-1]*0.5*0.25*(u[ms-1-xdim]+u[ms-1]-u[ms-2-xdim]-u[ms-2])-d23[ms-xdim-1]*0.5*0.25*(u[ms-1-xdim]+u[ms-1]-u[ms-1-xdim-xdim*ydim]-u[ms-1-xdim*ydim])-d33[ms-1]*0.5*(u[ms-1]-u[ms-xdim*ydim-1])+d31[ms-1]*0.25*(u[ms-1]-u[ms-2])-0.5*0.25*d31[ms-1]*(u[ms-1-xdim*ydim]+u[ms-1]-u[ms-2-xdim*ydim]-u[ms-2])+d32[ms-1]*0.25*(u[ms-1]-u[ms-1-xdim])-0.5*0.25*d32[ms-1]*(u[ms-1-xdim*ydim]+u[ms-1]-u[ms-1-xdim-xdim*ydim]-u[ms-1-xdim])-d33[ms-1-xdim*ydim]*0.5*(u[ms-1]-u[ms-1-xdim*ydim])-d31[ms-xdim*ydim-1]*0.5*0.25*(u[ms-1-xdim*ydim]+u[ms-1]-u[ms-2-xdim*ydim]-u[ms-2])-d32[ms-xdim*ydim-1]*0.5*0.25*(u[ms-1-xdim*ydim]+u[ms-1]-u[ms-1-xdim-xdim*ydim]-u[ms-1-xdim]));
  ;

for(i=1;i<xdim+1;i++)
  {
  r[i]+=dt*( 0.5 * d11[i]*(u[i+1]-2*u[i]+u[i-1])+d12[i]*0.25*0.5*(u[i+1+xdim]+u[i-1]-u[i+1]-u[i-1+xdim])+0.5*d13[i]*0.25*(u[i+1+xdim*ydim]+u[i-1]-u[i+1]-u[i-1+xdim*ydim])+0.5*d11[i+1]*(u[i+1]-u[i])+d12[i+1]*0.25*0.5*(u[i+xdim+1]+u[i+xdim]-u[i+1]-u[i])+0.5*d13[i+1]*0.25*(u[i+1+xdim*ydim]+u[i+xdim*ydim]-u[i]-u[i+1])-d11[i-1]*(u[i]-u[i-1])-d12[i-1]*0.25*0.5*(u[i-1+xdim]+u[i+xdim]-u[i-1]-u[i])-0.5*d13[i-1]*0.25*(u[i-1+xdim*ydim]+u[i+xdim*ydim]-u[i]-u[i-1])+ 0.5*d22[i+xdim]*(u[i+xdim]-u[i])+d23[i]*0.5*0.25*(u[i+xdim+xdim*ydim]+u[i+xdim*ydim]-u[i+xdim]-u[i]-2*(u[i+xdim*ydim]-u[i]))+d21[i+xdim]*0.25*0.5*(u[i+1+xdim]+u[i+1]-u[i-1]-u[i-1+xdim])+0.5*0.25*d23[i+xdim]*(u[i+xdim+xdim*ydim]+u[i+xdim*ydim]-u[i+xdim]-u[i])+0.5*d22[i]*(u[i+xdim]-u[i])-d21[i]*0.25*(u[i+1]-u[i-1])+d21[i]*0.5*0.25*(u[i+1+xdim]+u[i+1]-u[i-1+xdim]-u[i-1])+0.5*0.25*d31[i]*(u[i+xdim*ydim+1]+u[i+1]-u[i+xdim*ydim-1]-u[i-1]-2*(u[i+1]-u[i-1]))+0.5*0.25*d32[i]*(u[i+xdim+xdim*ydim]+u[i+xdim]-u[i]-u[i+xdim*ydim]-2*(u[i+xdim]-u[i]))+0.5*d33[i]*(u[i+xdim*ydim]-u[i])+0.5*0.25*d31[i+xdim*ydim]*(u[i+1+xdim*ydim]+u[i+1]-u[i+xdim*ydim-1]-u[i-1])+0.5*0.25*d32[i+xdim*ydim]*(u[i+xdim+xdim*ydim]+u[i+xdim]-u[i+xdim*ydim]-u[i])+0.5*d33[i+xdim*ydim]*(u[i+xdim*ydim]-u[i]));
       
  }

 

for(i=ms-xdim-1;i<ms-1;i++)
{
r[i]+=dt*(0.5*d11[i]*(u[i+1]+u[i-1]-2*u[i])+0.5*d11[i+1]*(u[i+1]-u[i])-0.5*d11[i-1]*(u[i]-u[i-1]) + 0.5* d12[i]*0.25*(u[i+1]+u[i-1-xdim]-u[i+1-xdim]-u[i-1])+0.5*d12[i+1]*0.25*(u[i+1]+u[i]-u[i+1-xdim]-u[i-xdim])-0.5*d12[i-1]*0.25*(u[i]+u[i-1]-u[i-1-xdim]-u[i-xdim])+  0.5*d13[i]*0.25*(u[i+1]+u[i-1-xdim*ydim]-u[i+1-xdim*ydim]-u[i-1])+0.5*d13[i+1]*0.25*(u[i+1]+u[i]-u[i+1-xdim*ydim]-u[i-xdim*ydim])-0.5*d13[i-1]*0.25*(u[i]+u[i-1]-u[i-1-xdim*ydim]-u[i-xdim*ydim])+ 0.5*d21[i]*0.25*(-u[i-xdim+1]+u[i+1]+u[i-xdim-1]-u[i-1])-0.5*d21[i-xdim]*0.25*(u[i-xdim+1]+u[i+1]-u[i-1]-u[i-xdim-1])-0.5*(d22[i]+d22[i-xdim])*(u[i]-u[i-xdim])+0.5*d23[i]*0.25*(u[i-xdim*ydim-xdim]+u[i]-u[i-xdim]-u[i-xdim*ydim])-0.5*d23[i-xdim]*0.25*(u[i]+u[i-xdim]-u[i-xdim-xdim*ydim]-u[i-xdim*ydim])+0.5*d31[i]*0.25*(u[i+1]+u[i-xdim*ydim-1]-u[i-1]-u[i-xdim*ydim+1])-0.5*d31[i-xdim*ydim]*0.25*(u[i-xdim*ydim+1]+u[i+1]-u[i-xdim*ydim-1]-u[i-1])+0.5*d32[i]*0.25*(u[i-xdim-xdim*ydim])+u[i]-u[i-xdim]-u[i-xdim*ydim]-0.5*d32[i-xdim*ydim]*0.25*(u[i-xdim*ydim]+u[i]-u[i-xdim*ydim-xdim]-u[i-xdim])-0.5*(d33[i]+d33[i-xdim*ydim])*(u[i]-u[i-xdim*ydim])) ;    
}

for(i=xdim+1;i<xdim*ydim+xdim;i++)
  {
   r[i]+=dt*(0.5*d11[i]*(u[i+1]+u[i-1]-2*u[i])+0.5*d11[i+1]*(u[i+1]-u[i])-0.5*d11[i-1]*(u[i]-u[i-1]) + 0.5* d12[i]*0.25*(u[i+1]+u[i-1-xdim]-u[i+1-xdim]-u[i-1])+0.5*d12[i+1]*0.25*(u[i+1]+u[i]-u[i+1-xdim]-u[i-xdim])-0.5*d12[i-1]*0.25*(u[i]+u[i-1]-u[i-1-xdim]-u[i-xdim])+0.5*d13[i]*0.5*(u[i+1+xdim*ydim]-u[i-1+xdim*ydim]-u[i+1]+u[i-1])+0.5*d13[i+1]*0.25*(u[1+1+xdim*ydim]+u[i+xdim*ydim]-u[i]-u[i+1])-0.5*d13[i-1]*0.25*(u[i-1+xdim*ydim]+u[i+xdim*ydim]-u[i]-u[i-1])+0.5*d21[i]*0.25*(u[i+1+xdim]+u[i-1-xdim]-u[i+1-xdim]-u[i-1+xdim])+0.5*d21[i+xdim]*0.25*(u[i+1+xdim]+u[i+1]-u[i-1+xdim]-u[i-1])-0.5*d21[i-xdim]*0.25*(u[i-xdim+1]+u[i+1]-u[i-xdim-1]-u[i-1]) +0.5*d22[i]*(u[i+xdim]+u[i-xdim]-2*u[i])+0.5*d22[i+xdim]*(u[i+xdim]-u[i])-0.5*d22[i-xdim]*(u[i]-u[i-xdim])+ 0.5* d23[i]*0.25*(u[i+xdim+xdim*ydim]+u[i-xdim]-u[i-xdim+xdim*ydim]-u[i+xdim])+0.5*d23[i+xdim]*0.25*(u[i+xdim+xdim*ydim]+u[i+xdim*ydim]-u[i+xdim]-u[i])-0.5*d23[i-1]*0.25*(u[i-xdim+xdim*ydim]+u[i+xdim*ydim]-u[i]-u[i-xdim])+0.5*d31[i]*0.25*(u[i+1]+u[i+xdim*ydim-1]-u[i+xdim*ydim+1]-u[i-1])+0.5*d31[i+xdim*ydim]*0.25*(u[i+xdim*ydim+1]+u[i+1]-u[i+xdim*ydim-1]-u[i-1])+0.5*d32[i]*0.25*(u[i+xdim]+u[i+xdim*ydim+xdim]-u[i-xdim]-u[i+xdim*ydim+xdim])+0.5*d32[i+xdim*ydim]*0.25*(u[i+xdim*ydim+xdim]+u[i+xdim]-u[i+xdim*ydim+xdim]-u[i-xdim])+0.5*d33[i+xdim*ydim]*(u[i+xdim*ydim]-u[i]));
  }

for(i=ms-xdim*ydim-xdim;i<ms-xdim-1;i++)
  {
    
     r[i]+=dt*(0.5*d11[i]*(u[i+1]+u[i-1]-2*u[i])+0.5*d11[i+1]*(u[i+1]-u[i])-0.5*d11[i-1]*(u[i]-u[i-1])+0.5*d12[i]*0.25*(u[i+1+xdim]+u[i-1-xdim]-u[i-1+xdim]-u[i+1-xdim])+0.5*d12[i+1]*0.25*(u[i+1+xdim]+u[i+xdim]-u[i+1-xdim]-u[i-xdim])-0.5*d12[i-1]*0.25*(u[i-1+xdim]+u[i+xdim]-u[i-1-xdim]-u[i-xdim]) + 0.5*d13[i]*0.25*(u[i+1]+u[i-1-xdim*ydim]-u[i-1]-u[i+1-xdim*ydim])+0.5*d13[i+1]*0.25*(u[i]+u[i+1]-u[i+1-xdim*ydim]-u[i-xdim*ydim])-0.5*d13[i-1]*0.25*(u[i]+u[i-1]-u[i-1-xdim*ydim]-u[i-xdim*ydim])+0.5*d21[i]*0.25*(u[i+1+xdim]+u[i-1-xdim]-u[i+1-xdim]-u[i-1+xdim])+0.5*d21[i+xdim]*0.25*(u[i+1+xdim]+u[i+1]-u[i-1+xdim]-u[i-1])-0.5*d21[i-xdim]*0.25*(u[i-xdim+1]+u[i+1]-u[i-xdim-1]-u[i-1]) +0.5*d22[i]*(u[i+xdim]+u[i-xdim]-2*u[i])+0.5*d22[i+xdim]*(u[i+xdim]-u[i])-0.5*d22[i-xdim]*(u[i]-u[i-xdim])+0.5*d23[i]*0.25*(u[i+xdim]+u[i-xdim-xdim*ydim]-u[i+xdim-xdim*ydim]-u[i-xdim])+0.5*d23[i+xdim]*0.25*(u[i+xdim]+u[i]-u[i+xdim-xdim*ydim]-u[i-xdim*ydim])-0.5*d23[i-xdim]*0.25*(u[i]+u[i-xdim]-u[i-xdim-xdim*ydim]-u[i-xdim*ydim])  + 0.5*d31[i]*0.25*(u[i+1]+u[i-xdim*ydim-1]-u[i-1]-u[i-xdim*ydim+1])-0.5*d31[i-xdim*ydim]*0.25*(u[i-xdim*ydim+1]+u[i+1]-u[i-xdim*ydim-1]-u[i-1]) +0.5*d32[i]*0.25*(u[i+xdim]+u[i-xdim*ydim-xdim]-u[i-xdim]-u[i-xdim*ydim+xdim])-0.5*d32[i-xdim*ydim]*0.25*(u[i-xdim*ydim+xdim]+u[i+xdim]-u[i-xdim*ydim-xdim]-u[i-xdim])-0.5*(d33[i]+d33[i-xdim*ydim])*(u[i]-u[i-xdim*ydim]));
  }

for(i=xdim*ydim+xdim;i<ms-xdim*ydim-xdim;i++)
{
  r[i]+=dt * 0.5 * (d11[i]*(u[i+1]-2*u[i]+u[i-1])+d12[i]*0.25*(u[i+1+xdim]+u[i-1-xdim]-u[i+1-xdim]-u[i-1+xdim])+d13[i]*0.25*(u[i+1+xdim*ydim]+u[i-1-xdim*ydim]-u[i+1-xdim*ydim]-u[i-1+xdim*ydim])+d23[i]*0.25*(u[i+xdim+xdim*ydim]+u[i-xdim-xdim*ydim]-u[i-xdim+xdim*ydim]-u[i+xdim-xdim*ydim])+d22[i]*(u[i+xdim]-2*u[i]+u[i-xdim])+d21[i]*0.25*(u[i+1+xdim]+u[i-1-xdim]-u[i+1-xdim]-u[i-1+xdim]) + d31[i]*0.25*(u[i+1+xdim*ydim]+u[i-1-xdim*ydim]-u[i+1-xdim*ydim]-u[i-1+xdim*ydim])+d32[i]*0.25*(u[i+xdim+xdim*ydim]+u[i-xdim-xdim*ydim]-u[i+xdim-xdim*ydim]-u[i-xdim+xdim*ydim])+d33[i]*(u[i+xdim*ydim]+u[i-xdim*ydim]-2*u[i])+d11[i+1]*(u[i+1]-u[i])+d12[i+1]*0.25*(u[i+1+xdim]+u[i+xdim]-u[i+1-xdim]-u[i-xdim]) +d13[i+1]*0.25*(u[i+1+xdim*ydim]+u[i+xdim*ydim]-u[i+1-xdim*ydim]-u[i-xdim*ydim])+ d11[i-1]*(u[i-1]-u[i])+d12[i-1]*0.25*(-u[i+xdim]-u[i-1+xdim]+u[i-1-xdim]+u[i-xdim])+d13[i-1]*0.25*(-u[i-1+xdim*ydim]-u[i+xdim*ydim]+u[i-1-xdim*ydim]+u[i-xdim*ydim]) + d22[i+xdim]*(u[i+xdim]-u[i])+d21[i+xdim]*0.25*(u[i+1+xdim]+u[i+1]-u[i+xdim-1]-u[i-1])+d23[i+xdim]*0.25*(u[i+xdim+xdim*ydim]+u[i+xdim*ydim]-u[i+xdim-xdim*ydim]-u[i-xdim*ydim]) + d22[i-xdim]*(u[i-xdim]-u[i])+d21[i-xdim]*0.25*(-u[i+1-xdim]-u[i+1]+u[i-xdim-1]+u[i-1])+d23[i-xdim]*0.25*(u[i-xdim-xdim*ydim]+u[i-xdim*ydim]-u[i-xdim+xdim*ydim]-u[i+xdim*ydim])+d31[i+xdim*ydim]*0.25*(u[i+1+xdim*ydim]+u[i+1]-u[i-1+xdim*ydim]-u[i-1])+d32[i+xdim*ydim]*0.25*(u[i+xdim+xdim*ydim]+u[i+xdim]-u[i-xdim+xdim*ydim]-u[i-xdim])+d33[i+xdim*ydim]*(u[i+xdim*ydim]-u[i])+d31[i-xdim*ydim]*0.25*(u[i-1-xdim*ydim]+u[i-1]-u[i+1]-u[i+1-xdim*ydim])+d32[i-xdim*ydim]*0.25*(u[i-xdim-xdim*ydim]+u[i-xdim]-u[i+xdim]-u[i+xdim-xdim*ydim])+d33[i-xdim*ydim]*(u[i-xdim*ydim]-u[i]));
  
  
}
 
}
  

// curvature calculus
void courbure(float & k1, float & k2, const float & du1,const float & du2, const float & du3,const float & du11, const float & du12,const float & du13, const float & du22, const float & du23, const float & du33, float & K, float & H){
  float h, delta, temp;
  h=du1*du1+du2*du2+du3*du3;
  if (h<=0.5)
    {
      k1=0.0;
      k2=0.0;
    }
  else
    {
      K=1/(h*h)*(du1*du1*(du22*du33-du23*du23)+2*du2*du3*(du13*du12-du11*du23)+du2*du2*(du11*du33-du13*du13)+2*du1*du3*(du23*du12-du22*du13)+du3*du3*(du11*du22-du12*du12)+2*du1*du2*(du13*du23-du33*du12));
      H=-1/(2*sqrt(h*h*h))*(du1*du1*(du22+du33)-2*du2*du3*du23+du2*du2*(du11+du33)-2*du1*du3*du13+du3*du3*(du11+du22)-2*du1*du2*du12);
      delta=H*H-K;
      if (delta <0)
	{
	  k1=H;
	  k2=H;
	}
      else
	{
    k1=H+sqrt(delta);
    k2=H-sqrt(delta);
    if(fabs(k1)<fabs(k2))
      {
	temp=k1;
	k1=k2;
	k2=temp;
      }
	}
    }
  
}






// explicit scheme for krissian flow diffusion

void expliciteKrissian(const MTLVector & u, MTLVector & r,  const MTLVector & du1s, const MTLVector & du2s, const MTLVector & du3s, const MTLVector & du11s, const MTLVector & du12s, const MTLVector & du13s, const MTLVector & du22s,  const MTLVector & du23s, const MTLVector & du33s, const float & dt, const float k, const DIFFUSION_FUNCTION & p, unsigned int xdim, unsigned int ydim, unsigned int zdim) {
  unsigned int i;
  float K,H;
  unsigned int ms = r.size();
  float Cksi,Ce1,Ce2;
  float k1,k2;
  float dus,du1,du2,du3,du11,du12,du13,du22,du23,du33; 
  
  for(i=0;i<ms;i++)
   r[i]=u[i];

  // si le gradient est nul on n a plus de base, on  fait une diffusion isotrope du/dt = div(g(nabla usigma)nabla u) ...
 
 
  for(i=0;i<ms;i++)
    {
      
      //courbure(k1s,k2s,du1s[i],du2s[i],du3s[i],du11s[i],du12s[i],du13s[i],du22s[i],du23s[i],du33s[i],K,H);
     
     
      dus=sqrt(du1s[i]*du1s[i]+du2s[i]*du2s[i]+du3s[i]*du3s[i]);
      if(dus<k)
	{
	  Cksi=1/((cosh(dus/k))*(cosh(dus/k)))*1/(k*k);
	  if(dus==0)
	    {
	      Ce1=1;
	      Ce2=1;
	    }
	  else
	    {
	      Ce1=tanh(dus/k)/(dus*k);
	      Ce2=tanh(dus/k)/(dus*k);
	    }
	}
      else
	{
	  Cksi=1/((cosh(dus/k))*(cosh(dus/k)))*1/(k*k);
	  Ce1=0;
	  Ce2=tanh(dus/k)/(dus*k);
	}
     
      // calcul des derivees
      
      if(i==0)
	{
	  du1=0.5*(u[1]-u[0]);
	  du11=0.25*(u[2]+u[0]-2*u[1]);
	}
      else if(i==ms-1)
	{
	  du1=0.5*(u[ms-1]-u[ms-2]);
	  du11=0.25*(u[ms-1]+u[ms-3]-2*u[ms-2]);
	}
      else
	{
	  du1=0.5*(u[i+1]-u[i-1]);
	  du11=u[i+1]+u[i-1]-2*u[i];
	}

      if(i<xdim)
	{
	  du2=0.5*(u[i+xdim]-u[i]);
	  du22=0.25*(u[i+2*xdim]+u[i]-2*u[i+xdim]);
	}
      else if(i>=ms-xdim)
	{
	  du2=0.5*(u[i]-u[i-xdim]);
	  du22=0.25*(u[i]+u[i-2*xdim]-2*u[i-xdim]);
	}      
      else
	{
	  du2=0.5*(u[i+xdim]-u[i-xdim]);	
	  du22=u[i+xdim]+u[i-xdim]-2*u[i];	
	}

      if(i<ydim*xdim)
	{
	  du3=0.5*(u[i+xdim*ydim]-u[i]);
	  du33=0.25*(u[i+2*xdim*ydim]+u[i]-2*u[i+xdim*ydim]);
	}
      else if(i>=ms-xdim*ydim)
	{
	  du3=0.5*(u[i]-u[i-xdim*ydim]);
	  du33=0.25*(u[i]+u[i-2*xdim*ydim]-2*u[i-xdim*ydim]);
	}
       else
	 {
	   du3=0.5*(u[i+xdim*ydim]-u[i-xdim*ydim]);
	   du33=u[i+xdim*ydim]+u[i-xdim*ydim]-2*u[i];
	 }
     
     
  
      if(i<xdim+1)
	du12=0.25*(u[i+1+xdim]+u[i]-u[i+1]-u[i+xdim]);
      else if(i>=ms-xdim-1)
	du12=0.25*(u[i]+u[i-1-xdim]-u[i-1]-u[i-xdim]);
      else
	du12=0.25*(u[i+1+xdim]+u[i-1-xdim]-u[i+1-xdim]-u[i-1+xdim]);

      if(i<xdim*ydim+1)
	du13=0.25*(u[i+xdim*ydim+1]+u[i]-u[i+1]-u[i+xdim*ydim]);
      else if(i>=ms-xdim*ydim-1)
	du13=0.25*(u[i]+u[i-xdim*ydim-1]-u[i-xdim*ydim]-u[i-1]);
      else
       du13=0.25*(u[i+xdim*ydim+1]+u[i-xdim*ydim-1]-u[i+xdim*ydim-1]-u[i-xdim*ydim+1]);

      if(i<xdim*ydim+xdim)
	du23=0.25*(u[i+xdim+xdim*ydim]+u[i]-u[i+xdim]-u[i+xdim*ydim]);
      else if(i>=ms-xdim*ydim-xdim)
	du23=0.25*(u[i]+u[i-xdim*ydim-xdim]-u[i-xdim]-u[i-xdim*ydim]);
      else
	du23=0.25*(u[i+xdim*ydim+xdim]+u[i-xdim*ydim-xdim]-u[i+xdim*ydim-xdim]-u[i-xdim*ydim+xdim]);
	 
     
      courbure(k1,k2,du1,du2,du3,du11,du12,du13,du22,du23,du33,K,H);	 
     
     
      if(du1*du1+du2*du2+du3*du3<0.01)
	r[i]=r[i]+dt*(du11+du22+du33);
      else
	r[i]=r[i]+dt*(Cksi*(du1*(du11*du1+du12*du2+du13*du3)+du2*(du12*du1+du22*du2+du3*du23)+du3*(du13*du1+du2*du23+du3*du33))/(du1*du1+du2*du2+du3*du3)-Ce1*(k1*sqrt(du1*du1+du2*du2+du3*du3))-Ce2*(k2*sqrt(du1*du1+du2*du2+du3*du3)));
      
      
    }


	
}


// diffusion matrix for K. Krissian diffusion in 2d
void calculDKarl(const MTLVector & du1s, const MTLVector & du2s, const MTLVector & u, MTLVector & d11,  MTLVector & d12, MTLVector &d21, MTLVector &d22,const float & quant, const  DIFFUSION_FUNCTION & p, const unsigned int & xdim){
  unsigned int i,ms=du1s.size();
  float norme, lambda, e0, e1, ugradient;

 for(i=0;i<ms;i++)
   {
     norme=(du1s[i]*du1s[i]+du2s[i]*du2s[i]);
     if (norme==0.0)
      {
	d11[i]=1.0;
	d12[i]=0.0;
	d21[i]=0.0;
	d22[i]=1.0;
      }
    else
      {	
	e0=du1s[i]/sqrt(norme);
	e1=du2s[i]/sqrt(norme);
	ugradient=(float) fabs((float) ((1.0-e0/2.0)*e1/2.0*u[i+xdim]+e0/2.0*(1.0-e1/2.0)*u[i+1]+e0/2.0*e1/2.0*u[i+1+xdim]-(1.0-e0/2.0)*e1/2*u[i-xdim]+e0/2.0*(1.0-e1/2.0)*u[i-1]+e0/2.0*e1/2.0*u[i-1-xdim]));
	lambda=diffusion((float) ugradient ,quant,p);
	d11[i]=lambda*du1s[i]*du1s[i]/norme+du2s[i]*du2s[i]/norme;
	d12[i]=(lambda-1)*du1s[i]*du2s[i]/norme;
	d21[i]=d12[i];
	d22[i]=lambda*du2s[i]*du2s[i]/norme+du1s[i]*du1s[i]/norme;
      }
   }
}

// diffusion matrix for K. Krissian diffusion in 3d
void calculDKarl3D(const MTLVector & du1s, const MTLVector & du2s, const MTLVector & du3s, const MTLVector & du11s, const MTLVector & du12s, const MTLVector & du13s,const MTLVector & du22s, const MTLVector & du23s, const MTLVector & du33s,const MTLVector & u, MTLVector & d11,  MTLVector & d12,MTLVector & d13, MTLVector &d21,MTLVector &d22,MTLVector & d23, MTLVector &d31,MTLVector &d32,MTLVector & d33, const float & quant, const  DIFFUSION_FUNCTION & p, const unsigned int & xdim){
  unsigned int i,ms=du1s.size();
  float norme, lambda, e0=0.0, e1=0.0,  ugradient, K,H,h,delta, k1, k2,e01, e11,e21, E,F, G, L, M, N;
  float a,b,c,no;
  Matrix<3,float> P;
  Matrix<3,float> D;
  Matrix<3,float> T;
  Matrix<3,float> temp;
#ifdef WIN32
  e0=0.0;
  e1=0.0;
#endif  

 for(i=0;i<ms;i++)
   {
     norme=(du1s[i]*du1s[i]+du2s[i]*du2s[i]+du3s[i]*du3s[i]);
     if (norme==0.0)
      {
	d11[i]=1.0;
	d12[i]=0.0;
	d13[i]=0.0;
	d21[i]=0.0;
	d22[i]=1.0;
	d23[i]=0.0;
	d31[i]=0.0;
	d32[i]=0.0;
	d33[i]=1.0;
      }	
     else
       {
	P(0,0)=du1s[i]/sqrt(norme);
	P(1,0)=du2s[i]/sqrt(norme);
	P(2,0)=du3s[i]/sqrt(norme);
	ugradient=(float) fabs((float) ((1.0-e0/2.0)*e1/2.0*u[i+xdim]+e0/2.0*(1.0-e1/2.0)*u[i+1]+e0/2.0*e1/2.0*u[i+1+xdim]-(1.0-e0/2.0)*e1/2*u[i-xdim]+e0/2.0*(1.0-e1/2.0)*u[i-1]+e0/2.0*e1/2.0*u[i-1-xdim]));
	lambda=diffusion((float) ugradient ,quant,p);
	
	//calcul des courbures 
	h=norme;
	K=1/(h*h)*(du1s[i]*du1s[i]*(du22s[i]*du33s[i]-du23s[i]*du23s[i])+2*du2s[i]*du3s[i]*(du13s[i]*du12s[i]-du11s[i]*du23s[i])+du2s[i]*du2s[i]*(du11s[i]*du33s[i]-du13s[i]*du13s[i])+2*du1s[i]*du3s[i]*(du23s[i]*du12s[i]-du22s[i]*du13s[i])+du3s[i]*du3s[i]*(du11s[i]*du22s[i]-du12s[i]*du12s[i])+2*du1s[i]*du2s[i]*(du13s[i]*du23s[i]-du33s[i]*du12s[i]));
	H=-1/(2*sqrt(h*h*h))*(du1s[i]*du1s[i]*(du22s[i]+du33s[i])-2*du2s[i]*du3s[i]*du23s[i]+du2s[i]*du2s[i]*(du11s[i]+du33s[i])-2*du1s[i]*du3s[i]*du13s[i]+du3s[i]*du3s[i]*(du11s[i]+du22s[i])-2*du1s[i]*du2s[i]*du12s[i]);
	delta=H*H-K;
	if (delta<=0)
	  {
	    k1=H;
	    k2=H;
	  }
	else
	  {
	    k1=H+sqrt(delta);
	    k2=H-sqrt(delta);
	  }
	if (delta<=0.1)
	  {
	    // dans ce cas on une valeur propre double, lambda id  on prend deux vecteurs quelconques cf diffusionMatrice//
	    a=du1s[i];
	    b=du2s[i];
	    c=du3s[i];
	    P(0,1)=-b/sqrt(a*a+b*b);
	    P(1,1)=a/sqrt(a*a+b*b);
	    P(2,1)=0;
	    P(0,2)=a*c/sqrt(a*c*a*c+b*c*b*c+(b*b+a*a)*(b*b+a*a));
	    P(1,2)=b*c/sqrt(a*c*a*c+b*c*b*c+(b*b+a*a)*(b*b+a*a));
	    P(2,2)=-(a*a+b*b)/sqrt(a*c*a*c+b*c*b*c+(b*b+a*a)*(b*b+a*a));
	    D(0,0)=lambda;
	    D(1,0)=0;
	    D(2,0)=0;
	    D(0,1)=0;
	    D(1,1)=0;
	    D(2,1)=0;
	    D(0,2)=0;
	    D(1,2)=0;
	    D(2,2)=1;
	    temp.Mul(D,P);
	    T.Trp(P);
	    temp.Mul(T,temp);
	    d11[i]=temp(0,0);
	    d12[i]=temp(0,1);
	    d13[i]=temp(0,2);
	    d21[i]=temp(1,0);
	    d22[i]=temp(1,1);
	    d23[i]=temp(1,2);
	    d31[i]=temp(2,0);
	    d32[i]=temp(2,1);
	    d33[i]=temp(2,2);
	   
	  }
	
	else if((du1s[i]==du2s[i])&(du2s[i]==du3s[i])){

	    // les vecteurs generaux sont nuls, methode manuelle 
	    E=(du1s[i]*du1s[i]+du3s[i]*du3s[i])/(du3s[i]*du3s[i]);
	    F=du1s[i]*du2s[i]/(du3s[i]*du3s[i]);
	    G=(du2s[i]*du2s[i]+du3s[i]*du3s[i])/(du3s[i]*du3s[i]);
	    H=(du1s[i]*du1s[i]+du2s[i]*du2s[i]+du3s[i]*du3s[i])/(du3s[i]*du3s[i]);
	    L=(2*du1s[i]*du3s[i]*du13s[i]-du1s[i]*du1s[i]*du33s[i]-du3s[i]*du3s[i]*du11s[i])/(sqrt(H)*du3s[i]*du3s[i]*du3s[i]);
	    M=(du1s[i]*du3s[i]*du23s[i]+du2s[i]*du3s[i]*du13s[i]-du1s[i]*du2s[i]*du33s[i]-du3s[i]*du3s[i]*du12s[i])/(sqrt(H)*du3s[i]*du3s[i]*du3s[i]);
	    N=(2*du2s[i]*du3s[i]*du23s[i]-du2s[i]*du2s[i]*du33s[i]-du3s[i]*du3s[i]*du22s[i])/(sqrt(H)*du3s[i]*du3s[i]*du3s[i]);
	    if(fabs(k1)>=fabs(k2))
	      {
		e01=du3s[i]*(G*M-F*N)/H;
		e11=du3s[i]*((E*N-G*L)/(2*H)+sqrt(delta));
		e21=-du1s[i]*(G*M-F*N)/H-du2s[i]*((E*N-G*L)/(2*H)+sqrt(delta));
		if((fabs(e01)<0.01)&(fabs(e11)<0.01)&(fabs(e21)<0.01))
		  {
		   
		    e01=du3s[i]*((G*L-E*N)/(2*H)+sqrt(delta));
		    e11=du3s[i]*(E*M-F*L)/H;
		    e21=-du1s[i]*((G*L-E*N)/(2*H)+sqrt(delta))-du2s[i]*(E*M-F*L)/H;
		  }
		no=sqrt(e01*e01+e11*e11+e21*e21);
		P(0,1)=e01/no;
		P(1,1)=e11/no;
		P(2,1)=e21/no;
	
		P(0,2)=P(1,0)*P(2,1)-P(1,1)*P(2,0);
		P(1,2)=P(0,1)*P(2,0)-P(0,0)*P(2,1);
		P(2,2)=P(0,0)*P(1,1)-P(1,0)*P(0,1);
	      }
	    else
	      {
		
		e01=du3s[i]*(G*M-F*N)/H;
		e11=du3s[i]*((E*N-G*L)/(2*H)-sqrt(delta));
		e21=-du1s[i]*(G*M-F*N)/H-du2s[i]*((E*N-G*L)/(2*H)-sqrt(delta));
		if((fabs(e01)<0.01)&(fabs(e11)<0.01)&(fabs(e21)<0.01))
		  {
		    e01=du3s[i]*((G*L-E*N)/(2*H)-sqrt(delta));
		    e11=du3s[i]*(E*M-F*L)/H;
		    e21=-du1s[i]*((G*L-E*N)/(2*H)-sqrt(delta))-du2s[i]*(E*M-F*L)/H;
		  }
		no=sqrt(e01*e01+e11*e11+e21*e21);
		P(0,1)=e01/no;
		P(1,1)=e11/no;
		P(2,1)=e21/no;
		
		P(0,2)=P(1,0)*P(2,1)-P(1,1)*P(2,0);
		P(1,2)=P(0,1)*P(2,0)-P(0,0)*P(2,1);
		P(2,2)=P(0,0)*P(1,1)-P(1,0)*P(0,1);
	      }
	    D(0,0)=lambda;
	    D(1,0)=0;
	    D(2,0)=0;
	    D(0,1)=0;
	    D(1,1)=0;
	    D(2,1)=0;
	    D(0,2)=0;
	    D(1,2)=0;
	    D(2,2)=1;
	    temp.Mul(D,P);
	    T.Trp(P);
	    temp.Mul(T,temp);
	    d11[i]=temp(0,0);
	    d12[i]=temp(0,1);
	    d13[i]=temp(0,2);
	    d21[i]=temp(1,0);
	    d22[i]=temp(1,1);
	    d23[i]=temp(1,2);
	    d31[i]=temp(2,0);
	    d32[i]=temp(2,1);
	    d33[i]=temp(2,2);
	   
	  }
	
	else 
	  {
	    if(fabs(k1)>=fabs(k2))
	      {
		e01=-1/(2*sqrt(h*h*h))*(-2*du3s[i]*du3s[i]*du3s[i]*du12s[i]+du2s[i]*du2s[i]*du2s[i]*du33s[i]+2*du2s[i]*du2s[i]*du2s[i]*du13s[i]-2*du2s[i]*du2s[i]*du3s[i]*du12s[i]+2*du3s[i]*du3s[i]*du1s[i]*du23s[i]+2*du3s[i]*du3s[i]*du2s[i]*du13s[i]-2*du2s[i]*du2s[i]*du1s[i]*du23s[i]-2*du3s[i]*du2s[i]*du1s[i]*du33s[i]+2*du1s[i]*du2s[i]*du3s[i]*du22s[i]+du2s[i]*du2s[i]*du3s[i]*du11s[i]-2*du3s[i]*du3s[i]*du1s[i]*du13s[i]+du3s[i]*du1s[i]*du1s[i]*du33s[i]-du1s[i]*du1s[i]*du3s[i]*du22s[i]+2*du3s[i]*du3s[i]*du2s[i]*du23s[i]-du3s[i]*du2s[i]*du2s[i]*du33s[i]+du3s[i]*du3s[i]*du3s[i]*du11s[i]-du3s[i]*du3s[i]*du3s[i]*du22s[i]-2*du2s[i]*du2s[i]*du1s[i]*du13s[i]+2*du1s[i]*du1s[i]*du2s[i]*du23s[i]-du2s[i]*du2s[i]*du2s[i]*du11s[i]+2*du1s[i]*du3s[i]*du3s[i]*du12s[i]-du2s[i]*du3s[i]*du3s[i]*du11s[i]-2*du3s[i]*du2s[i]*du2s[i]*du23s[i]+du2s[i]*du3s[i]*du3s[i]*du22s[i]-2*du3s[i]*du1s[i]*du1s[i]*du23s[i]+2*du1s[i]*du2s[i]*du2s[i]*du12s[i]+du1s[i]*du1s[i]*du2s[i]*du33s[i]-du1s[i]*du1s[i]*du2s[i]*du22s[i])+sqrt(delta)*(du3s[i]-du2s[i]);
		e11=-1/(2*sqrt(h*h*h))*(-2*du1s[i]*du1s[i]*du1s[i]*du23s[i]+du3s[i]*du3s[i]*du3s[i]*du11s[i]+2*du3s[i]*du3s[i]*du3s[i]*du12s[i]-2*du3s[i]*du3s[i]*du1s[i]*du23s[i]+2*du1s[i]*du1s[i]*du2s[i]*du13s[i]+2*du1s[i]*du1s[i]*du3s[i]*du12s[i]-2*du3s[i]*du3s[i]*du2s[i]*du13s[i]-2*du1s[i]*du2s[i]*du3s[i]*du11s[i]+2*du2s[i]*du3s[i]*du1s[i]*du33s[i]+du3s[i]*du3s[i]*du1s[i]*du22s[i]-2*du1s[i]*du1s[i]*du2s[i]*du12s[i]+du1s[i]*du2s[i]*du2s[i]*du11s[i]-du2s[i]*du2s[i]*du1s[i]*du33s[i]+2*du1s[i]*du1s[i]*du3s[i]*du13s[i]-du1s[i]*du3s[i]*du3s[i]*du11s[i]+du1s[i]*du1s[i]*du1s[i]*du22s[i]-du1s[i]*du1s[i]*du1s[i]*du33s[i]-2*du3s[i]*du3s[i]*du2s[i]*du12s[i]+2*du2s[i]*du2s[i]*du3s[i]*du13s[i]-du3s[i]*du3s[i]*du3s[i]*du22s[i]+2*du2s[i]*du1s[i]*du1s[i]*du23s[i]-du3s[i]*du1s[i]*du1s[i]*du22s[i]-2*du1s[i]*du3s[i]*du3s[i]*du13s[i]+du3s[i]*du1s[i]*du1s[i]*du33s[i]-2*du1s[i]*du2s[i]*du2s[i]*du13s[i]+2*du2s[i]*du3s[i]*du3s[i]*du23s[i]+du2s[i]*du2s[i]*du3s[i]*du11s[i]-du2s[i]*du2s[i]*du3s[i]*du33s[i])+sqrt(delta)*(du1s[i]-du3s[i]);
		e21=-1/(2*sqrt(h*h*h))*(-2*du2s[i]*du2s[i]*du2s[i]*du13s[i]+du1s[i]*du1s[i]*du1s[i]*du22s[i]+2*du1s[i]*du1s[i]*du1s[i]*du23s[i]-2*du1s[i]*du1s[i]*du2s[i]*du13s[i]+2*du2s[i]*du2s[i]*du3s[i]*du12s[i]+2*du2s[i]*du2s[i]*du1s[i]*du23s[i]-2*du1s[i]*du1s[i]*du3s[i]*du12s[i]-2*du1s[i]*du2s[i]*du3s[i]*du22s[i]+2*du1s[i]*du2s[i]*du3s[i]*du11s[i]+du1s[i]*du1s[i]*du2s[i]*du33s[i]-2*du2s[i]*du2s[i]*du3s[i]*du23s[i]+du3s[i]*du3s[i]*du2s[i]*du22s[i]-du3s[i]*du3s[i]*du2s[i]*du11s[i]+2*du2s[i]*du2s[i]*du1s[i]*du12s[i]-du2s[i]*du1s[i]*du1s[i]*du22s[i]+du2s[i]*du2s[i]*du2s[i]*du33s[i]-du2s[i]*du2s[i]*du2s[i]*du11s[i]-2*du1s[i]*du1s[i]*du3s[i]*du23s[i]+2*du3s[i]*du3s[i]*du1s[i]*du12s[i]-du1s[i]*du1s[i]*du1s[i]*du33s[i]+2*du3s[i]*du2s[i]*du2s[i]*du13s[i]-du1s[i]*du2s[i]*du2s[i]*du33s[i]-2*du2s[i]*du1s[i]*du1s[i]*du12s[i]+du1s[i]*du2s[i]*du2s[i]*du11s[i]-2*du2s[i]*du3s[i]*du3s[i]*du12s[i]+2*du3s[i]*du1s[i]*du1s[i]*du13s[i]+du3s[i]*du3s[i]*du1s[i]*du22s[i]-du3s[i]*du3s[i]*du1s[i]*du11s[i])+sqrt(delta)*(du2s[i]-du1s[i]);
		no=sqrt(e01*e01+e11*e11+e21*e21);
		P(0,1)=e01/no;
		P(1,1)=e11/no;
		P(2,1)=e21/no;
		
		P(0,2)=P(1,0)*P(2,1)-P(1,1)*P(2,0);
		P(1,2)=P(0,1)*P(2,0)-P(0,0)*P(2,1);
		P(2,2)=P(0,0)*P(1,1)-P(1,0)*P(0,1);
		}
	    else
	      {
		e01=-1/(2*sqrt(h*h*h))*(-2*du3s[i]*du3s[i]*du3s[i]*du12s[i]+du2s[i]*du2s[i]*du2s[i]*du33s[i]+2*du2s[i]*du2s[i]*du2s[i]*du13s[i]-2*du2s[i]*du2s[i]*du3s[i]*du12s[i]+2*du3s[i]*du3s[i]*du1s[i]*du23s[i]+2*du3s[i]*du3s[i]*du2s[i]*du13s[i]-2*du2s[i]*du2s[i]*du1s[i]*du23s[i]-2*du3s[i]*du2s[i]*du1s[i]*du33s[i]+2*du1s[i]*du2s[i]*du3s[i]*du22s[i]+du2s[i]*du2s[i]*du3s[i]*du11s[i]-2*du3s[i]*du3s[i]*du1s[i]*du13s[i]+du3s[i]*du1s[i]*du1s[i]*du33s[i]-du1s[i]*du1s[i]*du3s[i]*du22s[i]+2*du3s[i]*du3s[i]*du2s[i]*du23s[i]-du3s[i]*du2s[i]*du2s[i]*du33s[i]+du3s[i]*du3s[i]*du3s[i]*du11s[i]-du3s[i]*du3s[i]*du3s[i]*du22s[i]-2*du2s[i]*du2s[i]*du1s[i]*du13s[i]+2*du1s[i]*du1s[i]*du2s[i]*du23s[i]-du2s[i]*du2s[i]*du2s[i]*du11s[i]+2*du1s[i]*du3s[i]*du3s[i]*du12s[i]-du2s[i]*du3s[i]*du3s[i]*du11s[i]-2*du3s[i]*du2s[i]*du2s[i]*du23s[i]+du2s[i]*du3s[i]*du3s[i]*du22s[i]-2*du3s[i]*du1s[i]*du1s[i]*du23s[i]+2*du1s[i]*du2s[i]*du2s[i]*du12s[i]+du1s[i]*du1s[i]*du2s[i]*du33s[i]-du1s[i]*du1s[i]*du2s[i]*du22s[i])-sqrt(delta)*(du3s[i]-du2s[i]);
		e11=-1/(2*sqrt(h*h*h))*(-2*du1s[i]*du1s[i]*du1s[i]*du23s[i]+du3s[i]*du3s[i]*du3s[i]*du11s[i]+2*du3s[i]*du3s[i]*du3s[i]*du12s[i]-2*du3s[i]*du3s[i]*du1s[i]*du23s[i]+2*du1s[i]*du1s[i]*du2s[i]*du13s[i]+2*du1s[i]*du1s[i]*du3s[i]*du12s[i]-2*du3s[i]*du3s[i]*du2s[i]*du13s[i]-2*du1s[i]*du2s[i]*du3s[i]*du11s[i]+2*du2s[i]*du3s[i]*du1s[i]*du33s[i]+du3s[i]*du3s[i]*du1s[i]*du22s[i]-2*du1s[i]*du1s[i]*du2s[i]*du12s[i]+du1s[i]*du2s[i]*du2s[i]*du11s[i]-du2s[i]*du2s[i]*du1s[i]*du33s[i]+2*du1s[i]*du1s[i]*du3s[i]*du13s[i]-du1s[i]*du3s[i]*du3s[i]*du11s[i]+du1s[i]*du1s[i]*du1s[i]*du22s[i]-du1s[i]*du1s[i]*du1s[i]*du33s[i]-2*du3s[i]*du3s[i]*du2s[i]*du12s[i]+2*du2s[i]*du2s[i]*du3s[i]*du13s[i]-du3s[i]*du3s[i]*du3s[i]*du22s[i]+2*du2s[i]*du1s[i]*du1s[i]*du23s[i]-du3s[i]*du1s[i]*du1s[i]*du22s[i]-2*du1s[i]*du3s[i]*du3s[i]*du13s[i]+du3s[i]*du1s[i]*du1s[i]*du33s[i]-2*du1s[i]*du2s[i]*du2s[i]*du13s[i]+2*du2s[i]*du3s[i]*du3s[i]*du23s[i]+du2s[i]*du2s[i]*du3s[i]*du11s[i]-du2s[i]*du2s[i]*du3s[i]*du33s[i])-sqrt(delta)*(du1s[i]-du3s[i]);
		e21=-1/(2*sqrt(h*h*h))*(-2*du2s[i]*du2s[i]*du2s[i]*du13s[i]+du1s[i]*du1s[i]*du1s[i]*du22s[i]+2*du1s[i]*du1s[i]*du1s[i]*du23s[i]-2*du1s[i]*du1s[i]*du2s[i]*du13s[i]+2*du2s[i]*du2s[i]*du3s[i]*du12s[i]+2*du2s[i]*du2s[i]*du1s[i]*du23s[i]-2*du1s[i]*du1s[i]*du3s[i]*du12s[i]-2*du1s[i]*du2s[i]*du3s[i]*du22s[i]+2*du1s[i]*du2s[i]*du3s[i]*du11s[i]+du1s[i]*du1s[i]*du2s[i]*du33s[i]-2*du2s[i]*du2s[i]*du3s[i]*du23s[i]+du3s[i]*du3s[i]*du2s[i]*du22s[i]-du3s[i]*du3s[i]*du2s[i]*du11s[i]+2*du2s[i]*du2s[i]*du1s[i]*du12s[i]-du2s[i]*du1s[i]*du1s[i]*du22s[i]+du2s[i]*du2s[i]*du2s[i]*du33s[i]-du2s[i]*du2s[i]*du2s[i]*du11s[i]-2*du1s[i]*du1s[i]*du3s[i]*du23s[i]+2*du3s[i]*du3s[i]*du1s[i]*du12s[i]-du1s[i]*du1s[i]*du1s[i]*du33s[i]+2*du3s[i]*du2s[i]*du2s[i]*du13s[i]-du1s[i]*du2s[i]*du2s[i]*du33s[i]-2*du2s[i]*du1s[i]*du1s[i]*du12s[i]+du1s[i]*du2s[i]*du2s[i]*du11s[i]-2*du2s[i]*du3s[i]*du3s[i]*du12s[i]+2*du3s[i]*du1s[i]*du1s[i]*du13s[i]+du3s[i]*du3s[i]*du1s[i]*du22s[i]-du3s[i]*du3s[i]*du1s[i]*du11s[i])-sqrt(delta)*(du2s[i]-du1s[i]);
		no=sqrt(e01*e01+e11*e11+e21*e21);
		P(0,1)=e01/no;
		P(1,1)=e11/no;
		P(2,1)=e21/no;
	
		P(0,2)=P(1,0)*P(2,1)-P(1,1)*P(2,0);
		P(1,2)=P(0,1)*P(2,0)-P(0,0)*P(2,1);
		P(2,2)=P(0,0)*P(1,1)-P(1,0)*P(0,1);
	      }
	    D(0,0)=lambda;
	    D(1,0)=0;
	    D(2,0)=0;
	    D(0,1)=0;
	    D(1,1)=0;
	    D(2,1)=0;
	    D(0,2)=0;
	    D(1,2)=0;
	    D(2,2)=1;
	    temp.Mul(D,P);
	    T.Trp(P);
	    temp.Mul(T,temp);
	    d11[i]=temp(0,0);
	    d12[i]=temp(0,1);
	    d13[i]=temp(0,2);
	    d21[i]=temp(1,0);
	    d22[i]=temp(1,1);
	    d23[i]=temp(1,2);
	    d31[i]=temp(2,0);
	    d32[i]=temp(2,1);
	    d33[i]=temp(2,2);
	    
	  }

	 
       }
   }
}

      

// explicit scheme for 2d matrix Krissian algorithm

void explicitematKrissian(const MTLVector & u, MTLVector & r , const MTLVector & du1s, const MTLVector & du2s,const float & dt,const unsigned int xdim,const unsigned int ydim, const float quant, const DIFFUSION_FUNCTION & p) {
  unsigned int i;
  unsigned int ms = r.size();
  MTLVector d11(ms),d12(ms),d21(ms),d22(ms);

  calculDKarl(du1s,du2s,u,d11,d12,d21,d22,quant,p,xdim);
  
  for(i=0;i<ms;i++)
   r[i]=u[i];

  
   r[0]+=dt*(d11[0]*0.5*(u[1]-u[0])+d12[0]*(0.25*(u[0]-u[xdim])+0.5*0.25*(u[1+xdim]+u[xdim]-u[1]-u[0]))+d22[0]*0.5*(u[xdim]-u[0])+d21[0]*(0.25*(u[0]-u[1])+0.5*0.25*(u[1+xdim]+u[1]-u[xdim]-u[0]))+d11[1]*0.5*(u[1]-u[0])+d12[1]*0.25*0.5*(u[1+xdim]+u[xdim]-u[0]-u[1])+d22[xdim]*0.5*(u[xdim]-u[0])+d21[xdim]*0.25*(u[1+xdim]+u[1]-u[0]-u[xdim]));
 
 r[ms-1]+=dt*(-d11[ms-1]*0.5*(u[ms-1]-u[ms-2])+d12[ms-1]*0.25*(u[ms-1]-u[ms-1-xdim])-0.5*0.25*d12[ms-1]*(u[ms-2]+u[ms-1]-u[ms-2-xdim]-u[ms-1-xdim])-d22[ms-1]*0.5*(u[ms-1]-u[ms-xdim-1])+d21[ms-1]*0.25*(u[ms-1]-u[ms-2])-0.5*0.25*d21[ms-1]*(u[ms-1-xdim]+u[ms-1]-u[ms-2-xdim]-u[ms-2])-d11[ms-2]*0.5*(u[ms-1]-u[ms-2])-d12[ms-2]*0.5*0.25*(u[ms-2]+u[ms-1]-u[ms-2-xdim]-u[ms-1-xdim])-d22[ms-1-xdim]*0.5*(u[ms-1]-u[ms-1-xdim])-d21[ms-xdim-1]*0.5*0.25*(u[ms-1-xdim]+u[ms-1]-u[ms-2-xdim]-u[ms-2]));
  

  for(i=1;i<xdim;i++)
  {
    r[i]+=dt*( 0.5 * d11[i]*(u[i+1]-2*u[i]+u[i-1])+d12[i]*0.25*0.5*(u[i+1+xdim]+u[i-1]-u[i+1]-u[i-1+xdim])+0.5*d11[i+1]*(u[i+1]-u[i])+d12[i+1]*0.25*0.5*(u[i+xdim+1]+u[i+xdim]-u[i+1]-u[i])-d11[i-1]*(u[i]-u[i-1])-d12[i-1]*0.25*0.5*(u[i-1+xdim]+u[i+xdim]-u[i-1]-u[i])+ 0.5*d22[i+xdim]*(u[i+2*xdim]-u[i+xdim])+d21[i+xdim]*0.25*0.5*(u[i+1+xdim]+u[i+1]-u[i-1]-u[i-1+xdim])+0.5*d22[i]*(u[i+xdim]-u[i])-d21[i]*0.25*(u[i+1]-u[i-1])-d21[i]*0.5*0.25*(u[i+1+xdim]+u[i+1]-u[i-1+xdim]-u[i-1]));
     
  }
 
    for(i=ms-xdim;i<ms-1;i++)
  {
    r[i]+=dt*(0.5*d11[i]*(u[i+1]-2*u[i]+u[i-1])+d12[i]*0.5*0.25*(u[i+1]+u[i-1-xdim]-u[i+1-xdim]-u[i-1])+0.5*d11[i+1]*(u[i+1]-u[i])+d12[i+1]*0.5*0.25*(u[i+1]+u[i]-u[i+1-xdim]-u[i-xdim])-0.5*d11[i-1]*(u[i]-u[i-1])-d12[i-1]*0.5*0.25*(u[i-1]+u[i]-u[i-1-xdim]-u[i-xdim])-0.5*d22[i]*(u[i]-u[i-xdim])+d21[i]*0.25*(u[i+1]-u[i-1])-0.5*0.25*d21[i]*(u[i+1-xdim]+u[i+1]-u[i-1]-u[i-1-xdim])-0.5*d22[i-xdim]*(u[i]-u[i-xdim])-d21[i-xdim]*0.5*0.25*(u[i-xdim+1]+u[i+1]-u[i-xdim-1]-u[i-1]));
 
  }
    
 
  for(i=xdim;i<ms-xdim;i++){
    r[i]+= dt * 0.5 * (d11[i]*(u[i+1]-2*u[i]+u[i-1])+d12[i]*0.25*(u[i+1+xdim]+u[i-1-xdim]-u[i+1-xdim]-u[i-1+xdim])+d22[i]*(u[i+xdim]-2*u[i]+u[i-xdim])+d21[i]*0.25*(u[i+1+xdim]+u[i-1-xdim]-u[i+1-xdim]-u[i-1+xdim]) + d11[i+1]*(u[i+1]-u[i])+d12[i+1]*0.25*(u[i+1+xdim]+u[i+xdim]-u[i+1-xdim]-u[i-xdim]) + d11[i-1]*(u[i-1]-u[i])+d12[i-1]*0.25*(-u[i+xdim]-u[i-1+xdim]+u[i-1-xdim]+u[i-xdim]) + d22[i+xdim]*(u[i+xdim]-u[i])+d21[i+xdim]*0.25*(u[i+1+xdim]+u[i+1]-u[i+xdim-1]-u[i-1]) + d22[i-xdim]*(u[i-xdim]-u[i])+d21[i-xdim]*0.25*(-u[i+1-xdim]-u[i+1]+u[i-xdim-1]+u[i-1]));
    
  }
  
}


// explicit scheme for 3d matrix Krissian algorithm

void explicite3DmatKrissian(const MTLVector & u, MTLVector & r , const MTLVector & du1s, const MTLVector & du2s, const MTLVector & du3s ,const MTLVector & du11s, const MTLVector & du12s, const MTLVector & du13s , const MTLVector & du22s, const MTLVector & du23s ,const MTLVector & du33s , const float & dt,const unsigned int xdim,const unsigned int ydim, const float quant, const DIFFUSION_FUNCTION & p) {
  unsigned int i, ms = r.size();
  MTLVector d11(ms),d12(ms),d13(ms),d21(ms),d22(ms),d23(ms),d31(ms),d32(ms),d33(ms);
  
 

  calculDKarl3D(du1s,du2s,du3s,du11s,du12s,du13s,du22s,du23s,du33s,u,d11,d12,d13,d21,d22,d23,d31,d32,d33,quant,p,xdim);

  
  for(i=0;i<ms;i++)
    r[i]=u[i];
  r[0]+=dt*(d11[0]*0.5*(u[1]-u[0])+d12[0]*(0.25*(u[0]-u[xdim])+0.5*0.25*(u[1+xdim]+u[xdim]-u[1]-u[0]))+d13[0]*(0.25*(u[0]-u[xdim*ydim])+0.5*0.25*(u[1+xdim*ydim]+u[xdim*ydim]-u[1]-u[0]))+d22[0]*0.5*(u[xdim]-u[0])+d21[0]*(0.25*(u[0]-u[1])+0.5*0.25*(u[1+xdim]+u[1]-u[xdim]-u[0]))+d23[0]*(0.25*(u[0]-u[xdim*ydim])+0.5*0.25*(u[xdim*ydim+xdim]+u[xdim*ydim]-u[xdim]-u[0]))+d11[1]*0.5*(u[1]-u[0])+d12[1]*0.25*0.5*(u[1+xdim]+u[xdim]-u[0]-u[1])+d13[1]*0.25*0.5*(u[1+xdim*ydim]+u[xdim*ydim]-u[0]-u[1])+d22[xdim]*0.5*(u[xdim]-u[0])+d21[xdim]*0.25*(u[1+xdim]+u[1]-u[0]-u[xdim])+d23[xdim]*0.25*(u[xdim*ydim+xdim]+u[xdim*ydim]-u[0]-u[xdim])+d33[0]*0.5*(u[xdim*ydim]-u[0])+d31[0]*(0.25*(u[0]-u[1])+0.5*0.25*(u[1+xdim*ydim]+u[1]-u[xdim*ydim]-u[0]))+d32[0]*0.25*(u[0]-u[xdim])+d33[xdim*ydim]*0.5*(u[xdim*ydim]-u[0])+d31[xdim*ydim]*0.25*(u[1+xdim*ydim]+u[1]-u[0]-u[xdim*ydim])+d32[xdim*ydim]*0.25*(u[xdim*ydim+xdim]+u[xdim]-u[0]-u[xdim*ydim]));
  
  r[ms-1]+=dt*(-d11[ms-1]*0.5*(u[ms-1]-u[ms-2])+d12[ms-1]*0.25*(u[ms-1]-u[ms-1-xdim])-0.5*0.25*d12[ms-1]*(u[ms-2]+u[ms-1]-u[ms-2-xdim]-u[ms-1-xdim])+d13[ms-1]*0.25*(u[ms-1]-u[ms-1-xdim*ydim])-0.5*0.25*d13[ms-1]*(u[ms-2]+u[ms-1]-u[ms-2-xdim*ydim]-u[ms-1-xdim*ydim])-d22[ms-1]*0.5*(u[ms-1]-u[ms-xdim-1])+d21[ms-1]*0.25*(u[ms-1]-u[ms-2])-0.5*0.25*d21[ms-1]*(u[ms-1-xdim]+u[ms-1]-u[ms-2-xdim]-u[ms-2])+d23[ms-1]*0.25*(u[ms-1]-u[ms-1-xdim*ydim])-0.5*0.25*d23[ms-1]*(u[ms-1-xdim]+u[ms-1]-u[ms-1-xdim-xdim*ydim]-u[ms-1-xdim*ydim])-d11[ms-2]*0.5*(u[ms-1]-u[ms-2])-d12[ms-2]*0.5*0.25*(u[ms-2]+u[ms-1]-u[ms-2-xdim]-u[ms-1-xdim])-d13[ms-2]*0.5*0.25*(u[ms-2]+u[ms-1]-u[ms-2-xdim*ydim]-u[ms-1-xdim*ydim])-d22[ms-1-xdim]*0.5*(u[ms-1]-u[ms-1-xdim])-d21[ms-xdim-1]*0.5*0.25*(u[ms-1-xdim]+u[ms-1]-u[ms-2-xdim]-u[ms-2])-d23[ms-xdim-1]*0.5*0.25*(u[ms-1-xdim]+u[ms-1]-u[ms-1-xdim-xdim*ydim]-u[ms-1-xdim*ydim])-d33[ms-1]*0.5*(u[ms-1]-u[ms-xdim*ydim-1])+d31[ms-1]*0.25*(u[ms-1]-u[ms-2])-0.5*0.25*d31[ms-1]*(u[ms-1-xdim*ydim]+u[ms-1]-u[ms-2-xdim*ydim]-u[ms-2])+d32[ms-1]*0.25*(u[ms-1]-u[ms-1-xdim])-0.5*0.25*d32[ms-1]*(u[ms-1-xdim*ydim]+u[ms-1]-u[ms-1-xdim-xdim*ydim]-u[ms-1-xdim])-d33[ms-1-xdim*ydim]*0.5*(u[ms-1]-u[ms-1-xdim*ydim])-d31[ms-xdim*ydim-1]*0.5*0.25*(u[ms-1-xdim*ydim]+u[ms-1]-u[ms-2-xdim*ydim]-u[ms-2])-d32[ms-xdim*ydim-1]*0.5*0.25*(u[ms-1-xdim*ydim]+u[ms-1]-u[ms-1-xdim-xdim*ydim]-u[ms-1-xdim]));
  
  
  for(i=1;i<xdim+1;i++)
    {
      r[i]+=dt*( 0.5 * d11[i]*(u[i+1]-2*u[i]+u[i-1])+d12[i]*0.25*0.5*(u[i+1+xdim]+u[i-1]-u[i+1]-u[i-1+xdim])+0.5*d13[i]*0.25*(u[i+1+xdim*ydim]+u[i-1]-u[i+1]-u[i-1+xdim*ydim])+0.5*d11[i+1]*(u[i+1]-u[i])+d12[i+1]*0.25*0.5*(u[i+xdim+1]+u[i+xdim]-u[i+1]-u[i])+0.5*d13[i+1]*0.25*(u[i+1+xdim*ydim]+u[i+xdim*ydim]-u[i]-u[i+1])-d11[i-1]*(u[i]-u[i-1])-d12[i-1]*0.25*0.5*(u[i-1+xdim]+u[i+xdim]-u[i-1]-u[i])-0.5*d13[i-1]*0.25*(u[i-1+xdim*ydim]+u[i+xdim*ydim]-u[i]-u[i-1])+ 0.5*d22[i+xdim]*(u[i+xdim]-u[i])+d23[i]*0.5*0.25*(u[i+xdim+xdim*ydim]+u[i+xdim*ydim]-u[i+xdim]-u[i]-2*(u[i+xdim*ydim]-u[i]))+d21[i+xdim]*0.25*0.5*(u[i+1+xdim]+u[i+1]-u[i-1]-u[i-1+xdim])+0.5*0.25*d23[i+xdim]*(u[i+xdim+xdim*ydim]+u[i+xdim*ydim]-u[i+xdim]-u[i])+0.5*d22[i]*(u[i+xdim]-u[i])-d21[i]*0.25*(u[i+1]-u[i-1])+d21[i]*0.5*0.25*(u[i+1+xdim]+u[i+1]-u[i-1+xdim]-u[i-1])+0.5*0.25*d31[i]*(u[i+xdim*ydim+1]+u[i+1]-u[i+xdim*ydim-1]-u[i-1]-2*(u[i+1]-u[i-1]))+0.5*0.25*d32[i]*(u[i+xdim+xdim*ydim]+u[i+xdim]-u[i]-u[i+xdim*ydim]-2*(u[i+xdim]-u[i]))+0.5*d33[i]*(u[i+xdim*ydim]-u[i])+0.5*0.25*d31[i+xdim*ydim]*(u[i+1+xdim*ydim]+u[i+1]-u[i+xdim*ydim-1]-u[i-1])+0.5*0.25*d32[i+xdim*ydim]*(u[i+xdim+xdim*ydim]+u[i+xdim]-u[i+xdim*ydim]-u[i])+0.5*d33[i+xdim*ydim]*(u[i+xdim*ydim]-u[i]));
      
    }
  
  

  for(i=ms-xdim-1;i<ms-1;i++)
    {
      r[i]+=dt*(0.5*d11[i]*(u[i+1]+u[i-1]-2*u[i])+0.5*d11[i+1]*(u[i+1]-u[i])-0.5*d11[i-1]*(u[i]-u[i-1]) + 0.5* d12[i]*0.25*(u[i+1]+u[i-1-xdim]-u[i+1-xdim]-u[i-1])+0.5*d12[i+1]*0.25*(u[i+1]+u[i]-u[i+1-xdim]-u[i-xdim])-0.5*d12[i-1]*0.25*(u[i]+u[i-1]-u[i-1-xdim]-u[i-xdim])+  0.5*d13[i]*0.25*(u[i+1]+u[i-1-xdim*ydim]-u[i+1-xdim*ydim]-u[i-1])+0.5*d13[i+1]*0.25*(u[i+1]+u[i]-u[i+1-xdim*ydim]-u[i-xdim*ydim])-0.5*d13[i-1]*0.25*(u[i]+u[i-1]-u[i-1-xdim*ydim]-u[i-xdim*ydim])+ 0.5*d21[i]*0.25*(-u[i-xdim+1]+u[i+1]+u[i-xdim-1]-u[i-1])-0.5*d21[i-xdim]*0.25*(u[i-xdim+1]+u[i+1]-u[i-1]-u[i-xdim-1])-0.5*(d22[i]+d22[i-xdim])*(u[i]-u[i-xdim])+0.5*d23[i]*0.25*(u[i-xdim*ydim-xdim]+u[i]-u[i-xdim]-u[i-xdim*ydim])-0.5*d23[i-xdim]*0.25*(u[i]+u[i-xdim]-u[i-xdim-xdim*ydim]-u[i-xdim*ydim])+0.5*d31[i]*0.25*(u[i+1]+u[i-xdim*ydim-1]-u[i-1]-u[i-xdim*ydim+1])-0.5*d31[i-xdim*ydim]*0.25*(u[i-xdim*ydim+1]+u[i+1]-u[i-xdim*ydim-1]-u[i-1])+0.5*d32[i]*0.25*(u[i-xdim-xdim*ydim])+u[i]-u[i-xdim]-u[i-xdim*ydim]-0.5*d32[i-xdim*ydim]*0.25*(u[i-xdim*ydim]+u[i]-u[i-xdim*ydim-xdim]-u[i-xdim])-0.5*(d33[i]+d33[i-xdim*ydim])*(u[i]-u[i-xdim*ydim])) ;    
    }
  
  for(i=xdim+1;i<xdim*ydim+xdim;i++)
    {
      r[i]+=dt*(0.5*d11[i]*(u[i+1]+u[i-1]-2*u[i])+0.5*d11[i+1]*(u[i+1]-u[i])-0.5*d11[i-1]*(u[i]-u[i-1]) + 0.5* d12[i]*0.25*(u[i+1]+u[i-1-xdim]-u[i+1-xdim]-u[i-1])+0.5*d12[i+1]*0.25*(u[i+1]+u[i]-u[i+1-xdim]-u[i-xdim])-0.5*d12[i-1]*0.25*(u[i]+u[i-1]-u[i-1-xdim]-u[i-xdim])+0.5*d13[i]*0.5*(u[i+1+xdim*ydim]-u[i-1+xdim*ydim]-u[i+1]+u[i-1])+0.5*d13[i+1]*0.25*(u[1+1+xdim*ydim]+u[i+xdim*ydim]-u[i]-u[i+1])-0.5*d13[i-1]*0.25*(u[i-1+xdim*ydim]+u[i+xdim*ydim]-u[i]-u[i-1])+0.5*d21[i]*0.25*(u[i+1+xdim]+u[i-1-xdim]-u[i+1-xdim]-u[i-1+xdim])+0.5*d21[i+xdim]*0.25*(u[i+1+xdim]+u[i+1]-u[i-1+xdim]-u[i-1])-0.5*d21[i-xdim]*0.25*(u[i-xdim+1]+u[i+1]-u[i-xdim-1]-u[i-1]) +0.5*d22[i]*(u[i+xdim]+u[i-xdim]-2*u[i])+0.5*d22[i+xdim]*(u[i+xdim]-u[i])-0.5*d22[i-xdim]*(u[i]-u[i-xdim])+ 0.5* d23[i]*0.25*(u[i+xdim+xdim*ydim]+u[i-xdim]-u[i-xdim+xdim*ydim]-u[i+xdim])+0.5*d23[i+xdim]*0.25*(u[i+xdim+xdim*ydim]+u[i+xdim*ydim]-u[i+xdim]-u[i])-0.5*d23[i-1]*0.25*(u[i-xdim+xdim*ydim]+u[i+xdim*ydim]-u[i]-u[i-xdim])+0.5*d31[i]*0.25*(u[i+1]+u[i+xdim*ydim-1]-u[i+xdim*ydim+1]-u[i-1])+0.5*d31[i+xdim*ydim]*0.25*(u[i+xdim*ydim+1]+u[i+1]-u[i+xdim*ydim-1]-u[i-1])+0.5*d32[i]*0.25*(u[i+xdim]+u[i+xdim*ydim+xdim]-u[i-xdim]-u[i+xdim*ydim+xdim])+0.5*d32[i+xdim*ydim]*0.25*(u[i+xdim*ydim+xdim]+u[i+xdim]-u[i+xdim*ydim+xdim]-u[i-xdim])+0.5*d33[i+xdim*ydim]*(u[i+xdim*ydim]-u[i]));
  }
  
  for(i=ms-xdim*ydim-xdim;i<ms-xdim-1;i++)
    {
      
      r[i]+=dt*(0.5*d11[i]*(u[i+1]+u[i-1]-2*u[i])+0.5*d11[i+1]*(u[i+1]-u[i])-0.5*d11[i-1]*(u[i]-u[i-1])+0.5*d12[i]*0.25*(u[i+1+xdim]+u[i-1-xdim]-u[i-1+xdim]-u[i+1-xdim])+0.5*d12[i+1]*0.25*(u[i+1+xdim]+u[i+xdim]-u[i+1-xdim]-u[i-xdim])-0.5*d12[i-1]*0.25*(u[i-1+xdim]+u[i+xdim]-u[i-1-xdim]-u[i-xdim]) + 0.5*d13[i]*0.25*(u[i+1]+u[i-1-xdim*ydim]-u[i-1]-u[i+1-xdim*ydim])+0.5*d13[i+1]*0.25*(u[i]+u[i+1]-u[i+1-xdim*ydim]-u[i-xdim*ydim])-0.5*d13[i-1]*0.25*(u[i]+u[i-1]-u[i-1-xdim*ydim]-u[i-xdim*ydim])+0.5*d21[i]*0.25*(u[i+1+xdim]+u[i-1-xdim]-u[i+1-xdim]-u[i-1+xdim])+0.5*d21[i+xdim]*0.25*(u[i+1+xdim]+u[i+1]-u[i-1+xdim]-u[i-1])-0.5*d21[i-xdim]*0.25*(u[i-xdim+1]+u[i+1]-u[i-xdim-1]-u[i-1]) +0.5*d22[i]*(u[i+xdim]+u[i-xdim]-2*u[i])+0.5*d22[i+xdim]*(u[i+xdim]-u[i])-0.5*d22[i-xdim]*(u[i]-u[i-xdim])+0.5*d23[i]*0.25*(u[i+xdim]+u[i-xdim-xdim*ydim]-u[i+xdim-xdim*ydim]-u[i-xdim])+0.5*d23[i+xdim]*0.25*(u[i+xdim]+u[i]-u[i+xdim-xdim*ydim]-u[i-xdim*ydim])-0.5*d23[i-xdim]*0.25*(u[i]+u[i-xdim]-u[i-xdim-xdim*ydim]-u[i-xdim*ydim])  + 0.5*d31[i]*0.25*(u[i+1]+u[i-xdim*ydim-1]-u[i-1]-u[i-xdim*ydim+1])-0.5*d31[i-xdim*ydim]*0.25*(u[i-xdim*ydim+1]+u[i+1]-u[i-xdim*ydim-1]-u[i-1]) +0.5*d32[i]*0.25*(u[i+xdim]+u[i-xdim*ydim-xdim]-u[i-xdim]-u[i-xdim*ydim+xdim])-0.5*d32[i-xdim*ydim]*0.25*(u[i-xdim*ydim+xdim]+u[i+xdim]-u[i-xdim*ydim-xdim]-u[i-xdim])-0.5*(d33[i]+d33[i-xdim*ydim])*(u[i]-u[i-xdim*ydim]));
    }

  for(i=xdim*ydim+xdim;i<ms-xdim*ydim-xdim;i++)
    {
      r[i]+=dt * 0.5 * (d11[i]*(u[i+1]-2*u[i]+u[i-1])+d12[i]*0.25*(u[i+1+xdim]+u[i-1-xdim]-u[i+1-xdim]-u[i-1+xdim])+d13[i]*0.25*(u[i+1+xdim*ydim]+u[i-1-xdim*ydim]-u[i+1-xdim*ydim]-u[i-1+xdim*ydim])+d23[i]*0.25*(u[i+xdim+xdim*ydim]+u[i-xdim-xdim*ydim]-u[i-xdim+xdim*ydim]-u[i+xdim-xdim*ydim])+d22[i]*(u[i+xdim]-2*u[i]+u[i-xdim])+d21[i]*0.25*(u[i+1+xdim]+u[i-1-xdim]-u[i+1-xdim]-u[i-1+xdim]) + d31[i]*0.25*(u[i+1+xdim*ydim]+u[i-1-xdim*ydim]-u[i+1-xdim*ydim]-u[i-1+xdim*ydim])+d32[i]*0.25*(u[i+xdim+xdim*ydim]+u[i-xdim-xdim*ydim]-u[i+xdim-xdim*ydim]-u[i-xdim+xdim*ydim])+d33[i]*(u[i+xdim*ydim]+u[i-xdim*ydim]-2*u[i])+d11[i+1]*(u[i+1]-u[i])+d12[i+1]*0.25*(u[i+1+xdim]+u[i+xdim]-u[i+1-xdim]-u[i-xdim]) +d13[i+1]*0.25*(u[i+1+xdim*ydim]+u[i+xdim*ydim]-u[i+1-xdim*ydim]-u[i-xdim*ydim])+ d11[i-1]*(u[i-1]-u[i])+d12[i-1]*0.25*(-u[i+xdim]-u[i-1+xdim]+u[i-1-xdim]+u[i-xdim])+d13[i-1]*0.25*(-u[i-1+xdim*ydim]-u[i+xdim*ydim]+u[i-1-xdim*ydim]+u[i-xdim*ydim]) + d22[i+xdim]*(u[i+xdim]-u[i])+d21[i+xdim]*0.25*(u[i+1+xdim]+u[i+1]-u[i+xdim-1]-u[i-1])+d23[i+xdim]*0.25*(u[i+xdim+xdim*ydim]+u[i+xdim*ydim]-u[i+xdim-xdim*ydim]-u[i-xdim*ydim]) + d22[i-xdim]*(u[i-xdim]-u[i])+d21[i-xdim]*0.25*(-u[i+1-xdim]-u[i+1]+u[i-xdim-1]+u[i-1])+d23[i-xdim]*0.25*(u[i-xdim-xdim*ydim]+u[i-xdim*ydim]-u[i-xdim+xdim*ydim]-u[i+xdim*ydim])+d31[i+xdim*ydim]*0.25*(u[i+1+xdim*ydim]+u[i+1]-u[i-1+xdim*ydim]-u[i-1])+d32[i+xdim*ydim]*0.25*(u[i+xdim+xdim*ydim]+u[i+xdim]-u[i-xdim+xdim*ydim]-u[i-xdim])+d33[i+xdim*ydim]*(u[i+xdim*ydim]-u[i])+d31[i-xdim*ydim]*0.25*(u[i-1-xdim*ydim]+u[i-1]-u[i+1]-u[i+1-xdim*ydim])+d32[i-xdim*ydim]*0.25*(u[i-xdim-xdim*ydim]+u[i-xdim]-u[i+xdim]-u[i+xdim-xdim*ydim])+d33[i-xdim*ydim]*(u[i-xdim*ydim]-u[i]));
      
    }  
}

void writeMTLVector(const TclString name, const MTLVector & m) {
  MTLConstVecIt i=m.begin();
  OFSTREAM output((const char *) name,IOS::out | IOS::binary);
  unsigned int s = sizeof(float);
  for(;i!=m.end();++i)
	{
	  output.write((const char*)&*i,s);
	}
  output.close();
}

void readMTLVector(const TclString name, MTLVector & m) {
  unsigned int s = sizeof(float);
  float value=0;
  IFSTREAM input((const char *) name,IOS::in | IOS::binary);
  MTLVecIt i=m.begin();
  for(;i!=m.end();++i)
	{
	  input.read((char*)&value,s);
	  *i=value;
	}
  input.close();
}

// Gaussian smoothing to ensure the stability of the process
void AOSGauss(MTLVector & r, const MTLVector & u, const float & dt, const float & d) {
  int i, ms = u.size();
  MTLBanded Dx(ms,ms, 1, 1);
  MTLVector y(ms), L(ms-1), R(ms);
  MTLVecIt ity;
  
  // The Thomas Algorithm in the linear case : diffusivity = 1 => Gaussian smoothing
  
  //make Dx
  Dx(0,0) = 1.0 + d*dt/2.0;
  Dx(ms-1,ms-1) = 1.0 + d*dt/2.0;
  Dx(ms-1,ms-2) = -d*dt/2.0;
  Dx(ms-2,ms-1) = -d*dt/2.0;

  for(i=1;i<ms-1;i++) {
    Dx(i,i) = 1.0 + d * dt;
    Dx(i-1,i) = -d*dt/2.0;
    Dx(i,i-1) = -d*dt/2.0;
  } 
  
  //make L and R
  R[0]=Dx(0,0);
 
  for(i=1;i<ms-1;i++) {
    L[i-1]=Dx(i,i-1)/R[i-1];
    R[i]=Dx(i,i)-L[i-1]*Dx(i-1,i); 
  }
  R[ms-1] = Dx(ms-1,ms-1) - L[ms-3]*Dx(ms-3,ms-2);
  L[ms-2] = Dx(ms-1,ms-2)/R[ms-1];

  //make y
  y[0]=u[0];
  i=1;
  for(ity=y.begin()+1;ity!=y.end();++ity) {
    *ity = u[i] - L[i-1] * (*(ity-1));
    i++;
  }

  // Result :

  //make r
  r[ms-1] = y[ms-1] / R[ms-1];
  for(i=ms-1;i>0;i--) {
    r[i-1] = (y[i-1]-Dx(i-1,i)*r[i])/R[i-1];
  } 
}
#endif
}


