#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/arithmetic.h"

#include <inrimage/arithmetic.h>

namespace yav {



  template <class TYPE>
    void invBuffer(TYPE *T, Inrimage *inr) {
    
    TYPE *target, *end;
    void *data = inr->getData();
    for(target = (TYPE *) data, end = target + inr->getX() * inr->getY() * inr->getZ() * inr->getVdim(); target < end; target++) 
      *target = - *target; 
    
  }
  
  template <class TYPE>
    void absBuffer(TYPE *T, Inrimage *inr) {
    
    void *data = inr->getData();
    TYPE *target, *end;
    for(target = (TYPE *) data, end = target + inr->getX() * inr->getY() * inr->getZ() * inr->getVdim(); target < end; target++) 
      *target = (TYPE) fabs((double)*target);
    
  }
  
  template <class TYPE>
    void setBufferVal(TYPE *T, Inrimage *inr, double val) {
    
    void *data = inr->getData();
    TYPE *target, *end;
    TYPE tval = (TYPE) val;
    for(target = (TYPE *) data, end = target + inr->getX() * inr->getY() * inr->getZ() * inr->getVdim(); target < end; target++)
      *target = tval;
    
  }

  template <class TYPE>
    void addBufferVal(TYPE *T, Inrimage *inr, double val) {
    
    void *data = inr->getData();
    TYPE *target, *end;
    TYPE tval = (TYPE) val;
    for(target = (TYPE *) data, end = target + inr->getX() * inr->getY() * inr->getZ() * inr->getVdim(); target < end; target++)
      *target += tval;
    
  }

  template <class TYPE>
    void subBufferVal(TYPE *T, Inrimage *inr, double val) {
    
    void *data = inr->getData();
    TYPE *target, *end;
    TYPE tval = (TYPE) val;
    for(target = (TYPE *) data, end = target + inr->getX() * inr->getY() * inr->getZ() * inr->getVdim(); target < end; target++)
      *target -= tval;
    
  }

  template <class TYPE>
    void mulBufferVal(TYPE *T, Inrimage *inr, double val) {
    
    void *data = inr->getData();
    TYPE *target, *end;
    for(target = (TYPE *) data, end = target + inr->getX() * inr->getY() * inr->getZ() * inr->getVdim(); target < end; target++)
      *target = (TYPE) (*target * val);
    
  }

  template <class TYPE>
    void divBufferVal(TYPE *T, Inrimage *inr, double val) {
    
    void *data = inr->getData();
    TYPE *target, *end;
    for(target = (TYPE *) data, end = target + inr->getX() * inr->getY() * inr->getZ() * inr->getVdim(); target < end; target++)
      *target = (TYPE) (*target / val);
    
  }

  template <class TO, class FROM>
    void addBuffer(TO *T, FROM *F, Inrimage *inr, const Inrimage *from) {

    TO *target, *end;
    FROM *src;
 
    for(target = (TO *) (inr->getData()), src = (FROM *) (from->getData()), end = target + inr->getX() * inr->getY() * inr->getZ(); target < end; src++, target++)
	  *target += (TO) *src;

  }
    
  template <class TO, class FROM>
    void subBuffer(TO *T, FROM *F, Inrimage *inr, const Inrimage *from) {
    
    TO *target, *end;
    FROM *src;
    
    for(target = (TO *) (inr->getData()), src = (FROM *) (from->getData()), end = target + inr->getX() * inr->getY() * inr->getZ(); target < end; src++, target++)
      *target -= (TO) *src;
    
  }
  
  template <class TO, class FROM>
    void mulBuffer(TO *T, FROM *F, Inrimage *inr, const Inrimage *from) {
    
    TO *target, *end;
    FROM *src;
    
    for(target = (TO *) (inr->getData()), src = (FROM *) (from->getData()), end = target + inr->getX() * inr->getY() * inr->getZ(); target < end; src++, target++)
      *target *= (TO) *src;
    
  }
  
  template <class TO, class FROM>
    void divBuffer(TO *T, FROM *F, Inrimage *inr, const Inrimage *from) {
    
    TO *target, *end;
    FROM *src;
    
    for(target = (TO *) (inr->getData()), src = (FROM *) (from->getData()), end = target + inr->getX() * inr->getY() * inr->getZ(); target < end; src++, target++)
      if(*src != (TO) 0) 
	*target /= (TO) *src;
      else 
	*target = (TO) 0;
  }

  template <class TO, class FROM>
    void addScalarBuffertoVectorialBuffer(TO *T,FROM *F, Inrimage *inr, const Inrimage *from)  { 
    TO *target, *end; 
    FROM *src; 
    unsigned int i; 
    void *data = inr->getData();
    void *from_data = from->getData();
    unsigned int vdim = inr->getVdim();
    
    if(inr->getVectorialMode() == VM_INTERLACED) {
      for(target = (TO *) data, src = (FROM *) from_data, end = target + inr->getX() * inr->getY() * inr->getZ() + vdim; target < end; src++) {
	for(i = 0; i < vdim; i++) *target++ += (TO) *src;
      }
    }
    else { 
      TO *tv; 
      unsigned long size = inr->getX() * inr->getY() * inr->getZ();
      
      for(target = (TO *) data, src = (FROM *) from_data, end = target + size; 
	  target < end; src++, target++) { 
	for(i = 0, tv = target; i < vdim; i++, tv += size) *tv += (TO) *src; 
      } 
    } 
  }

  template <class TO, class FROM, Index D, class T>
  void addVectortoVectorialBuffer(TO *To,FROM *F, Inrimage *inr, const Vect<D,T>& v)  { 
    unsigned int i; 
    void *data = inr->getData();
    unsigned int vdim = inr->getVdim();
    TO *target = (TO *) data;
    
    if(D==vdim) {
      if(inr->getVectorialMode() == VM_INTERLACED) {
	TO *end = target + inr->getX()*inr->getY()*inr->getZ()*vdim;
	for(; target < end;) {
	  for(i = 0; i < vdim; ++i,++target)
	    *target += (TO) v[i];
	}
      }
      else { 
	TO *tv; 
	unsigned long size = inr->getX()*inr->getY()*inr->getZ();
	TO *end = target + size;
	for(;target < end;target++) { 
	  for(i=0,tv=target; i<vdim; ++i, tv += size)
	    *tv += (TO) v[i]; 
	} 
      } 
    }
    else
      CERR << "Vector is not the same dimension as the image vectorial dimension" << ENDL;
  }

  template <class TO, class FROM, Index L, Index C, class T>
  void mulMatrixtoVectorialBuffer(TO *To,FROM *F, Inrimage *inr, const _Mat<L,C,T>& m)  { 
    unsigned int i, j;
    void *data = inr->getData();
    unsigned int vdim = inr->getVdim();
    TO *target = (TO *) data;
    TO *it;
    if(C==vdim && L==vdim) {
      Vect<C,T> temp;
      if(inr->getVectorialMode() == VM_INTERLACED) {
	TO *end = target + inr->getX()*inr->getY()*inr->getZ()*vdim;
	for(; target < end;) {
	  temp.Null();
	  for(i=0;i<vdim;++i) {
	    it=target;
	    for(j=0;j<vdim;++j,++it) 
	      temp.setVal(i,temp.getVal(i) +(*it)* m.getValue(i,j));
	  }
	  for(i=0;i<vdim;++i,++target) {
	    *target = (TO) temp.getVal(i);
	  }
	}
      }
      else { 
	TO *tv; 
	unsigned long size = inr->getX()*inr->getY()*inr->getZ();
	TO *end = target + size;
	for(;target < end;++target) {
	  temp.Null();
	  for(i=0,tv=target; i<vdim;++i, tv += size) {
	    it=target;
	    for(j=0;j<vdim;++j,++it) 
	       temp.setVal(i,temp.getVal(i) +(*it)* m.getValue(i,j));
	  }
	  for(i=0;i<vdim;++i,tv += size)
	    *tv = (TO) temp.getVal(i);
	} 
      } 
    }
    else
      CERR << "Matrix is not square or not the same dimension as the image vectorial dimension" << ENDL;
  }
  
  template <class TO, class FROM>
    void subScalarBuffertoVectorialBuffer(TO *T,FROM *F, Inrimage *inr, const Inrimage *from)  { 
    
    TO *target, *end; 
    FROM *src; 
    void *data = inr->getData();
    void *from_data = from->getData();
    unsigned int xdim, ydim, zdim, vdim;

    xdim = inr->getX(); ydim = inr->getY(); 
    zdim = inr->getZ(); vdim = inr->getVdim();

    unsigned int i; 

    if(inr->getVectorialMode() == VM_INTERLACED) { 
      for(target = (TO *) data, src = (FROM *) from_data, 
	    end = target + xdim * ydim * zdim * vdim; target < end; src++) { 
	for(i = 0; i < vdim; i++) *target++ -= (TO) *src; \
      } 
    } 
    else { 
      TO *tv; 
      unsigned long size = xdim * ydim * zdim; 
      
      for(target = (TO *) data, src = (FROM *) from_data, end = target + size; 
	  target < end; src++, target++) { 
	for(i = 0, tv = target; i < vdim; i++, tv += size) *tv -= (TO) *src; 
      }
    } 
    
  }
  
  template <class TO, class FROM>
    void mulScalarBuffertoVectorialBuffer(TO *T,FROM *F, Inrimage *inr, const Inrimage *from)  { 
    
    TO *target, *end; 
    FROM *src; 
    void *data = inr->getData();
    void *from_data = from->getData();
    unsigned int xdim, ydim, zdim, vdim;

    xdim = inr->getX(); ydim = inr->getY(); 
    zdim = inr->getZ(); vdim = inr->getVdim();

    unsigned int i; 

    if(inr->getVectorialMode() == VM_INTERLACED) { 
      for(target = (TO *) data, src = (FROM *) from_data, 
	    end = target + xdim * ydim * zdim * vdim; target < end; src++) { 
	for(i = 0; i < vdim; i++) *target++ *= (TO) *src; 
      } 
    } 
    else { 
      TO *tv;
      unsigned long size = xdim * ydim * zdim; 
      
      for(target = (TO *) data, src = (FROM *) from_data, end = target + size; 
	  target < end; src++, target++) { 
	for(i = 0, tv = target; i < vdim; i++, tv += size) *tv *= (TO) *src; 
      } 
    } 
  }

  template <class TO, class FROM>
    void divScalarBuffertoVectorialBuffer(TO *T,FROM *F, Inrimage *inr, const Inrimage *from)  { 
    
    TO *target, *end; 
    FROM *src; 
    void *data = inr->getData();
    void *from_data = from->getData();
    unsigned int xdim, ydim, zdim, vdim;

    xdim = inr->getX(); ydim = inr->getY(); 
    zdim = inr->getZ(); vdim = inr->getVdim();
    
    unsigned int i; 
    
    if(inr->getVectorialMode() == VM_INTERLACED) { 
      for(target = (TO *) data, src = (FROM *) from_data, 
	    end = target + xdim * ydim * zdim * vdim; target < end; src++) { 
	if(*src != (FROM) 0) 
	  for(i = 0; i < vdim; i++) *target++ /= (TO) *src; 
	else       
	  for(i = 0; i < vdim; i++) *target++ = (TO) 0; 
      } 
    } 
    else { 
      TO *tv; 
      unsigned long size = xdim * ydim * zdim; 
      
      for(target = (TO *) data, src = (FROM *) from_data, end = target + size; 
	  target < end; src++, target++) { 
	if(*src != (FROM) 0) 
	  for(i = 0, tv = target; i < vdim; i++, tv += size) *tv /= (TO) *src; 
	else 
	  for(i = 0, tv = target; i < vdim; i++, tv += size) *tv = (TO) 0; 
      } 
    } 
  }
  
  template <class TO, class FROM>
    void addVectorialBuffertoVectorialBuffer(TO *T,FROM *F, Inrimage *inr, const Inrimage *from)  { 

    TO *target, *end; 
    FROM *src; 
 
    void *data = inr->getData();
    void *from_data = from->getData();
    unsigned int xdim, ydim, zdim, vdim;

    xdim = inr->getX(); ydim = inr->getY(); 
    zdim = inr->getZ(); vdim = inr->getVdim();
    
    for(target = (TO *) data, src = (FROM *) from_data, 
	  end = target + xdim * ydim * zdim * vdim; target < end; 
	target++, src++) { 
      *target += (TO) *src; 
    } 
  }
  
  template <class TO, class FROM>
    void subVectorialBuffertoVectorialBuffer(TO *T,FROM *F, Inrimage *inr, const Inrimage *from)  { 

    TO *target, *end; 
    FROM *src; 
    void *data = inr->getData();
    void *from_data = from->getData();
    unsigned int xdim, ydim, zdim, vdim;

    xdim = inr->getX(); ydim = inr->getY(); 
    zdim = inr->getZ(); vdim = inr->getVdim();
    
    
    for(target = (TO *) data, src = (FROM *) from_data, 
	  end = target + xdim * ydim * zdim * vdim; target < end; 
	target++, src++) { 
      *target -= (TO)(*src); 
    } 
  }

  
  
  template <class TO, class FROM>
    void mulVectorialBuffertoVectorialBuffer(TO *T,FROM *F, Inrimage *inr, const Inrimage *from)  { 
    
    TO *target, *end; 
    FROM *src; 
    void *data = inr->getData();
    void *from_data = from->getData();
    unsigned int xdim, ydim, zdim, vdim;
    
    xdim = inr->getX(); ydim = inr->getY(); 
    zdim = inr->getZ(); vdim = inr->getVdim();
    
    for(target = (TO *) data, src = (FROM *) from_data, 
	  end = target + xdim * ydim * zdim * vdim; target < end; 
	target++, src++) { 
      *target *= (TO)(*src);
    } 
  }

  template <class TO, class FROM>
    void divVectorialBuffertoVectorialBuffer(TO *T,FROM *F, Inrimage *inr, const Inrimage *from)  { 
    
    TO *target, *end; 
    FROM *src; 
    void *data = inr->getData();
    void *from_data = from->getData();
    unsigned int xdim, ydim, zdim, vdim;
    
    xdim = inr->getX(); ydim = inr->getY(); 
    zdim = inr->getZ(); vdim = inr->getVdim();
    
    for(target = (TO *) data, src = (FROM *) from_data, 
	  end = target + xdim * ydim * zdim * vdim; target < end;
	target++, src++) { 
      if(*src != (FROM) 0) 
	*target /= (TO)(*src); 
      else  
	*target = (TO) 0; 
    } 
  }

  template <class TYPE>
    void notBuffer(TYPE *T, Inrimage *inr) {

    TYPE *target, *end; 
    void *data = inr->getData();
    unsigned int xdim, ydim, zdim;
    
    xdim = inr->getX(); ydim = inr->getY(); 
    zdim = inr->getZ(); 
    
    for(target = (TYPE*) data, end = target + xdim * ydim * zdim; target < end;  target++)
#ifdef __GNUC__
      *target ^= (TYPE) 0xFFFFFFFFFFFFFFFFLL;
#else
    *target ^= (TYPE) 0xFFFFFFFFFFFFFFFF;
#endif
    
  }

  template <class TYPE>
    void orBuffer(TYPE *T, Inrimage *inr, const Inrimage *from) {
    
    TYPE *target, *src, *end; 
    void *data = inr->getData();
    void *from_data = from->getData();
    unsigned int xdim, ydim, zdim;
    
    xdim = inr->getX(); ydim = inr->getY(); 
    zdim = inr->getZ(); 
    
    for(target = (TYPE*) data, src = (TYPE*) from_data, 
	end = target + xdim * ydim * zdim; target < end; src++, target++) 
      *target |= *src; 

  }

  template <class TYPE>
    void andBuffer(TYPE *T, Inrimage *inr, const Inrimage *from) {
    
    TYPE *target, *src, *end; 
    void *data = inr->getData();
    void *from_data = from->getData();
    unsigned int xdim, ydim, zdim;
    
    xdim = inr->getX(); ydim = inr->getY(); 
    zdim = inr->getZ(); 
    
    for(target = (TYPE*) data, src = (TYPE*) from_data, 
	end = target + xdim * ydim * zdim; target < end; src++, target++) 
      *target &= *src; 

  }

  template <class TYPE>
    void xorBuffer(TYPE *T, Inrimage *inr, const Inrimage *from) {
    
    TYPE *target, *src, *end; 
    void *data = inr->getData();
    void *from_data = from->getData();
    unsigned int xdim, ydim, zdim;
    
    xdim = inr->getX(); ydim = inr->getY(); 
    zdim = inr->getZ(); 
    
    for(target = (TYPE*) data, src = (TYPE*) from_data, 
	end = target + xdim * ydim * zdim; target < end; src++, target++) 
      *target |= *src; 
  
  }







void setValue(Inrimage *inr, double val) {
  switch(inr->getType()) {
  case Inrimage::WT_UNSIGNED_CHAR:
    setBufferVal((unsigned char *) 0, inr, val);
    break;
  case Inrimage::WT_SIGNED_CHAR:
    setBufferVal((char *) 0, inr, val);
    break;
  case Inrimage::WT_UNSIGNED_SHORT:
    setBufferVal((unsigned short *) 0, inr, val);
    break;
  case Inrimage::WT_SIGNED_SHORT:
    setBufferVal((short int *) 0, inr, val);
    break;
  case Inrimage::WT_UNSIGNED_INT:
    setBufferVal((unsigned int *) 0, inr, val);
    break;
  case Inrimage::WT_SIGNED_INT:
    setBufferVal((int *) 0, inr, val);
    break;
  case Inrimage::WT_UNSIGNED_LONG:
    setBufferVal((unsigned LONGINT *) 0, inr, val);
    break;
  case Inrimage::WT_SIGNED_LONG:
    setBufferVal((LONGINT *) 0, inr, val);
    break;
  case Inrimage::WT_FLOAT:
    setBufferVal((float *) 0, inr, val);
    break;
  case Inrimage::WT_DOUBLE:
    setBufferVal((double *) 0, inr, val);
    break;
  case Inrimage::WT_RGB:
    setBufferVal((unsigned char *) 0, inr, val);
    break;
  case Inrimage::WT_RGBA:
    setBufferVal((unsigned char *) 0, inr, val);
    break;
  case Inrimage::WT_FLOAT_VECTOR:
    setBufferVal((float *) 0, inr, val);
    break;
  case Inrimage::WT_DOUBLE_VECTOR:
    setBufferVal((double *) 0, inr, val);
    break;
  default:
    break;
  }
}



void inv(Inrimage *inr) {

  switch(inr->getType()) {
  case Inrimage::WT_SIGNED_CHAR:
    invBuffer((char*) 0, inr);
    break;
  case Inrimage::WT_SIGNED_SHORT:
    invBuffer((short int*) 0, inr);
    break;
  case Inrimage::WT_SIGNED_INT:
    invBuffer((int *) 0, inr);
    break;
  case Inrimage::WT_SIGNED_LONG:
    invBuffer((LONGINT *) 0, inr);
    break;
  case Inrimage::WT_FLOAT:
  case Inrimage::WT_FLOAT_VECTOR:
    invBuffer((float *) 0, inr);
    break;
  case Inrimage::WT_DOUBLE:
  case Inrimage::WT_DOUBLE_VECTOR:
    invBuffer((double *) 0, inr);
    break;
  case Inrimage::WT_UNSIGNED_CHAR:
  case Inrimage::WT_UNSIGNED_SHORT:
  case Inrimage::WT_UNSIGNED_INT:
  case Inrimage::WT_UNSIGNED_LONG:
  case Inrimage::WT_RGB:
  case Inrimage::WT_RGBA:
  default:
    break;
  }
}
  
void abs(Inrimage *inr) {
  switch(inr->getType()) {
  case Inrimage::WT_SIGNED_CHAR:
    absBuffer((char *) 0, inr);
    break;
  case Inrimage::WT_SIGNED_SHORT:
    absBuffer((short int *) 0, inr);
    break;
  case Inrimage::WT_SIGNED_INT:
    absBuffer((int *) 0, inr);
    break;
  case Inrimage::WT_SIGNED_LONG:
    absBuffer((LONGINT *) 0, inr);
    break;
  case Inrimage::WT_FLOAT:
  case Inrimage::WT_FLOAT_VECTOR:
    absBuffer((float *) 0, inr);
    break;
  case Inrimage::WT_DOUBLE:
  case Inrimage::WT_DOUBLE_VECTOR:
    absBuffer((double *) 0, inr);
    break;
  case Inrimage::WT_UNSIGNED_CHAR:
  case Inrimage::WT_UNSIGNED_SHORT:
  case Inrimage::WT_UNSIGNED_INT:
  case Inrimage::WT_UNSIGNED_LONG:
  case Inrimage::WT_RGB:
  case Inrimage::WT_RGBA:
  default:
    break;
  }
}
 

void add(Inrimage *inr, double val) {
  switch(inr->getType()) {
  case Inrimage::WT_UNSIGNED_CHAR:
    addBufferVal((unsigned char *) 0, inr, val);
    break;
  case Inrimage::WT_SIGNED_CHAR:
    addBufferVal((char *) 0, inr, val);
    break;
  case Inrimage::WT_UNSIGNED_SHORT:
    addBufferVal((unsigned short *) 0, inr, val);
    break;
  case Inrimage::WT_SIGNED_SHORT:
    addBufferVal((short int *) 0, inr, val);
    break;
  case Inrimage::WT_UNSIGNED_INT:
    addBufferVal((unsigned int *) 0, inr, val);
    break;
  case Inrimage::WT_SIGNED_INT:
    addBufferVal((int *) 0, inr, val);
    break;
  case Inrimage::WT_UNSIGNED_LONG:
    addBufferVal((unsigned LONGINT *) 0, inr, val);
    break;
  case Inrimage::WT_SIGNED_LONG:
    addBufferVal((LONGINT *) 0, inr, val);
    break;
  case Inrimage::WT_FLOAT:
    addBufferVal((float *) 0, inr, val);
    break;
  case Inrimage::WT_DOUBLE:
    addBufferVal((double *) 0, inr, val);
    break;
  case Inrimage::WT_RGB:
    addBufferVal((unsigned char *) 0, inr, val);
    break;
  case Inrimage::WT_RGBA:
    addBufferVal((unsigned char *) 0, inr, val);
    break;
  case Inrimage::WT_FLOAT_VECTOR:
    addBufferVal((float *) 0, inr, val);
    break;
  case Inrimage::WT_DOUBLE_VECTOR:
    addBufferVal((double *) 0, inr, val);
    break;
  default:
    break;
  }
}

 
void add(Inrimage *inr, const Vec3<double>& v) throw(Inrimage::InvalidArguments) {
  switch(inr->getType()) {
  case Inrimage::WT_FLOAT_VECTOR:
    addVectortoVectorialBuffer((float *) 0, (float *) 0, inr, 
			       (const Vect<3,double> &) v);
    break;
  case Inrimage::WT_DOUBLE_VECTOR:
    addVectortoVectorialBuffer((double *) 0, (double *) 0, inr, 
			       (const Vect<3,double> &) v);
    break;
  default:
    CERR << "Not a vectorial Image" << ENDL;
    break;
  }
}
 
void mul(Inrimage *inr, const Matrix3x3<double>& m) throw(Inrimage::InvalidArguments) {
  switch(inr->getType()) {
  case Inrimage::WT_FLOAT_VECTOR:
    mulMatrixtoVectorialBuffer((float *) 0, (float *) 0, inr,  
			       (const _Mat<3,3,double>&) m);
    break;
  case Inrimage::WT_DOUBLE_VECTOR:
    mulMatrixtoVectorialBuffer((double *) 0, (double *) 0, inr, 
			       (const _Mat<3,3,double>&) m);
    break;
  default:
    CERR << "Not a vectorial Image" << ENDL;
    break;
  }
}
  
void sub(Inrimage *inr, double val) {
  switch(inr->getType()) {
  case Inrimage::WT_UNSIGNED_CHAR:
    subBufferVal((unsigned char *) 0, inr, val);
    break;
  case Inrimage::WT_SIGNED_CHAR:
    subBufferVal((char *) 0, inr, val);
    break;
  case Inrimage::WT_UNSIGNED_SHORT:
    subBufferVal((unsigned short *) 0, inr, val);
    break;
  case Inrimage::WT_SIGNED_SHORT:
    subBufferVal((short int *) 0, inr, val);
    break;
  case Inrimage::WT_UNSIGNED_INT:
    subBufferVal((unsigned int *) 0, inr, val);
    break;
  case Inrimage::WT_SIGNED_INT:
    subBufferVal((int *) 0, inr, val);
    break;
  case Inrimage::WT_UNSIGNED_LONG:
    subBufferVal((unsigned LONGINT *) 0, inr, val);
    break;
  case Inrimage::WT_SIGNED_LONG:
    subBufferVal((LONGINT *) 0, inr, val);
    break;
  case Inrimage::WT_FLOAT:
    subBufferVal((float *) 0, inr, val);
    break;
  case Inrimage::WT_DOUBLE:
    subBufferVal((double *) 0, inr, val);
    break;
  case Inrimage::WT_RGB:
    subBufferVal((unsigned char *) 0, inr, val);
    break;
  case Inrimage::WT_RGBA:
    subBufferVal((unsigned char *) 0, inr, val);
    break;
  case Inrimage::WT_FLOAT_VECTOR:
    subBufferVal((float *) 0, inr, val);
    break;
  case Inrimage::WT_DOUBLE_VECTOR:
    subBufferVal((double *) 0, inr, val);
    break;
  default:
    break;
  }
}

void mul(Inrimage *inr, double val) {
  switch(inr->getType()) {
  case Inrimage::WT_UNSIGNED_CHAR:
    mulBufferVal((unsigned char *) 0, inr, val);
    break;
  case Inrimage::WT_SIGNED_CHAR:
    mulBufferVal((char *) 0, inr, val);
    break;
  case Inrimage::WT_UNSIGNED_SHORT:
    mulBufferVal((unsigned short *) 0, inr, val);
    break;
  case Inrimage::WT_SIGNED_SHORT:
    mulBufferVal((short *) 0,inr, val);
    break;
  case Inrimage::WT_UNSIGNED_INT:
    mulBufferVal((unsigned int *) 0, inr, val);
    break;
  case Inrimage::WT_SIGNED_INT:
    mulBufferVal((int *) 0, inr, val);
    break;
  case Inrimage::WT_UNSIGNED_LONG:
    mulBufferVal((unsigned LONGINT *) 0, inr, val);
    break;
  case Inrimage::WT_SIGNED_LONG:
    mulBufferVal((LONGINT *) 0, inr, val);
    break;
  case Inrimage::WT_FLOAT:
    mulBufferVal((float *) 0, inr, val);
    break;
  case Inrimage::WT_DOUBLE:
    mulBufferVal((double *) 0, inr, val);
    break;
  case Inrimage::WT_RGB:
    mulBufferVal((unsigned char *) 0, inr, val);
    break;
  case Inrimage::WT_RGBA:
    mulBufferVal((unsigned char *) 0, inr, val);
    break;
  case Inrimage::WT_FLOAT_VECTOR:
    mulBufferVal((float *) 0, inr, val);
    break;
  case Inrimage::WT_DOUBLE_VECTOR:
    mulBufferVal((double *) 0, inr, val);
    break;
  default:
    break;
  }
}

// div a value to image voxels
void div(Inrimage *inr, double val) {
  switch(inr->getType()) {
  case Inrimage::WT_UNSIGNED_CHAR:
    divBufferVal((unsigned char *) 0, inr, val);
    break;
  case Inrimage::WT_SIGNED_CHAR:
    divBufferVal((char *) 0, inr, val);
    break;
  case Inrimage::WT_UNSIGNED_SHORT:
    divBufferVal((unsigned short *) 0, inr, val);
    break;
  case Inrimage::WT_SIGNED_SHORT:
    divBufferVal((short *) 0, inr, val);
    break;
  case Inrimage::WT_UNSIGNED_INT:
    divBufferVal((unsigned int *) 0, inr, val);
    break;
  case Inrimage::WT_SIGNED_INT:
    divBufferVal((int *) 0, inr, val);
    break;
  case Inrimage::WT_UNSIGNED_LONG:
    divBufferVal((unsigned LONGINT *) 0, inr, val);
    break;
  case Inrimage::WT_SIGNED_LONG:
    divBufferVal((LONGINT *) 0, inr, val);
    break;
  case Inrimage::WT_FLOAT:
    divBufferVal((float *) 0, inr, val);
    break;
  case Inrimage::WT_DOUBLE:
    divBufferVal((double *) 0, inr, val);
    break;
  case Inrimage::WT_RGB:
    divBufferVal((unsigned char *) 0, inr, val);
    break;
  case Inrimage::WT_RGBA:
    divBufferVal((unsigned char *) 0, inr, val);
    break;
  case Inrimage::WT_FLOAT_VECTOR:
    divBufferVal((float *) 0, inr, val);
    break;
  case Inrimage::WT_DOUBLE_VECTOR:
    divBufferVal((double *) 0,inr, val);
    break;
  default:
    break;
  }
}


void add (Inrimage *inr, const Inrimage *from) throw(Inrimage::InvalidArguments) {
  
    char _from[] = "yav::add(const Inrimage *)";

  if(inr->getX() != from->getX() || inr->getY() != from->getY() || inr->getZ() != from->getZ())
    throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");

  switch(inr->getType()) {
  
  case Inrimage::WT_UNSIGNED_CHAR:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      addBuffer((unsigned char *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      addBuffer((unsigned char *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      addBuffer((unsigned char *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      addBuffer((unsigned char *) 0, (short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      addBuffer((unsigned char *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      addBuffer((unsigned char *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      addBuffer((unsigned char *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      addBuffer((unsigned char *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      addBuffer((unsigned char *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      addBuffer((unsigned char *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_SIGNED_CHAR:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      addBuffer((char *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      addBuffer((char *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      addBuffer((char *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      addBuffer((char *) 0, (short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      addBuffer((char *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      addBuffer((char *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      addBuffer((char *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      addBuffer((char *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      addBuffer((char *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      addBuffer((char *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_UNSIGNED_SHORT:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      addBuffer((unsigned short int *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      addBuffer((unsigned short int *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      addBuffer((unsigned short int *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      addBuffer((unsigned short int *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      addBuffer((unsigned short int *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      addBuffer((unsigned short int *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      addBuffer((unsigned short int *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      addBuffer((unsigned short int *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      addBuffer((unsigned short int *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_SIGNED_SHORT:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      addBuffer((short int *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      addBuffer((short int *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      addBuffer((short int *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      addBuffer((short int *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      addBuffer((short int *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      addBuffer((short int *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      addBuffer((short int *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      addBuffer((short int *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      addBuffer((short int *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_UNSIGNED_INT:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      addBuffer((unsigned int *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      addBuffer((unsigned int *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      addBuffer((unsigned int *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      addBuffer((unsigned int *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      addBuffer((unsigned int *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      addBuffer((unsigned int *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      addBuffer((unsigned int *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      addBuffer((unsigned int *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      addBuffer((unsigned int *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_SIGNED_INT:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      addBuffer((int *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      addBuffer((int *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      addBuffer((int *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      addBuffer((int *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      addBuffer((int *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      addBuffer((int *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      addBuffer((int *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      addBuffer((int *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      addBuffer((int *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_UNSIGNED_LONG:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      addBuffer((unsigned LONGINT *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      addBuffer((unsigned LONGINT *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      addBuffer((unsigned LONGINT *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      addBuffer((unsigned LONGINT *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      addBuffer((unsigned LONGINT *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      addBuffer((unsigned LONGINT *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      addBuffer((unsigned LONGINT *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      addBuffer((unsigned LONGINT *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      addBuffer((unsigned LONGINT *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_SIGNED_LONG:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      addBuffer((LONGINT *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      addBuffer((LONGINT *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      addBuffer((LONGINT *) 0, (short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      addBuffer((LONGINT *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      addBuffer((LONGINT *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      addBuffer((LONGINT *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      addBuffer((LONGINT *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      addBuffer((LONGINT *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      addBuffer((LONGINT *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_FLOAT:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      addBuffer((float *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      addBuffer((float *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      addBuffer((float *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      addBuffer((float *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      addBuffer((float *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      addBuffer((float *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      addBuffer((float *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      addBuffer((float *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      addBuffer((float *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_DOUBLE:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      addBuffer((double *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      addBuffer((double *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      addBuffer((double *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      addBuffer((double *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      addBuffer((double *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      addBuffer((double *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      addBuffer((double *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      addBuffer((double *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      addBuffer((double *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_RGB:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      addScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      addScalarBuffertoVectorialBuffer((unsigned char *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      addScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      addScalarBuffertoVectorialBuffer((unsigned char *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      addScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      addScalarBuffertoVectorialBuffer((unsigned char *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      addScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      addScalarBuffertoVectorialBuffer((unsigned char *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      addScalarBuffertoVectorialBuffer((unsigned char *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      addScalarBuffertoVectorialBuffer((unsigned char *) 0, (double *) 0, inr, from);
      break;
    case Inrimage::WT_RGB:
      addVectorialBuffertoVectorialBuffer((unsigned char *) 0, (unsigned char *) 0, inr, from);
    default:
      break;
    }
    break;

  case Inrimage::WT_RGBA:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      addScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      addScalarBuffertoVectorialBuffer((unsigned char *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      addScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      addScalarBuffertoVectorialBuffer((unsigned char *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      addScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      addScalarBuffertoVectorialBuffer((unsigned char *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      addScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      addScalarBuffertoVectorialBuffer((unsigned char *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      addScalarBuffertoVectorialBuffer((unsigned char *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      addScalarBuffertoVectorialBuffer((unsigned char *) 0, (double *) 0, inr, from);
      break;
    case Inrimage::WT_RGBA:
      addVectorialBuffertoVectorialBuffer((unsigned char *) 0, (unsigned char *) 0, inr, from);
    default:
      break;
    }
    break;

  case Inrimage::WT_FLOAT_VECTOR:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      addScalarBuffertoVectorialBuffer((float *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      addScalarBuffertoVectorialBuffer((float *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      addScalarBuffertoVectorialBuffer((float *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      addScalarBuffertoVectorialBuffer((float *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      addScalarBuffertoVectorialBuffer((float *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      addScalarBuffertoVectorialBuffer((float *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      addScalarBuffertoVectorialBuffer((float *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      addScalarBuffertoVectorialBuffer((float *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      addScalarBuffertoVectorialBuffer((float *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      addScalarBuffertoVectorialBuffer((float *) 0, (double *) 0, inr, from);
      break;
    case Inrimage::WT_RGB:
      if(inr->getVdim() == 3)
	addVectorialBuffertoVectorialBuffer((float *) 0, (unsigned char *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    case Inrimage::WT_RGBA:
      if(inr->getVdim() == 4)
	addVectorialBuffertoVectorialBuffer((float *) 0, (unsigned char *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    case Inrimage::WT_FLOAT_VECTOR:
      if(from->getVdim() == inr->getVdim())
	addVectorialBuffertoVectorialBuffer((float *) 0, (float *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    case Inrimage::WT_DOUBLE_VECTOR:
      if(from->getVdim() == inr->getVdim())
	addVectorialBuffertoVectorialBuffer((float *) 0, (double *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_DOUBLE_VECTOR:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      addScalarBuffertoVectorialBuffer((double *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      addScalarBuffertoVectorialBuffer((double *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      addScalarBuffertoVectorialBuffer((double *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      addScalarBuffertoVectorialBuffer((double *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      addScalarBuffertoVectorialBuffer((double *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      addScalarBuffertoVectorialBuffer((double *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      addScalarBuffertoVectorialBuffer((double *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      addScalarBuffertoVectorialBuffer((double *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      addScalarBuffertoVectorialBuffer((double *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      addScalarBuffertoVectorialBuffer((double *) 0, (double *) 0, inr, from);
      break;
    case Inrimage::WT_RGB:
      if(inr->getVdim() == 3)
	addVectorialBuffertoVectorialBuffer((double *) 0, (unsigned char *) 0,inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    case Inrimage::WT_RGBA:
      if(inr->getVdim() == 4)
	addVectorialBuffertoVectorialBuffer((double *) 0, (unsigned char *) 0,inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    case Inrimage::WT_FLOAT_VECTOR:
      if(from->getVdim() == inr->getVdim())
	addVectorialBuffertoVectorialBuffer((double *) 0, (float *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    case Inrimage::WT_DOUBLE_VECTOR:
      if(from->getVdim() == inr->getVdim())
	addVectorialBuffertoVectorialBuffer((double *) 0, (double *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    default:
      break;
    }
    break;

  default:
    break;
  }
}


void sub(Inrimage *inr, const Inrimage *from) throw(Inrimage::InvalidArguments) {

  char _from[] = "yav::sub(const Inrimage *)";

  if(inr->getX() != from->getX() || inr->getY() != from->getY() || inr->getZ() != from->getZ())
    throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");

  switch(inr->getType()) {
  
  case Inrimage::WT_UNSIGNED_CHAR:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      subBuffer((unsigned char *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      subBuffer((unsigned char *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      subBuffer((unsigned char *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      subBuffer((unsigned char *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      subBuffer((unsigned char *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      subBuffer((unsigned char *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      subBuffer((unsigned char *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      subBuffer((unsigned char *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      subBuffer((unsigned char *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      subBuffer((unsigned char *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

   case Inrimage::WT_SIGNED_CHAR:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      subBuffer((char *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      subBuffer((char *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      subBuffer((char *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      subBuffer((char *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      subBuffer((char *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      subBuffer((char *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      subBuffer((char *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      subBuffer((char *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      subBuffer((char *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      subBuffer((char *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

 case Inrimage::WT_UNSIGNED_SHORT:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      subBuffer((unsigned short int *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      subBuffer((unsigned short int *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      subBuffer((unsigned short int *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      subBuffer((unsigned short int *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      subBuffer((unsigned short int *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      subBuffer((unsigned short int *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      subBuffer((unsigned short int *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      subBuffer((unsigned short int *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      subBuffer((unsigned short int *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      subBuffer((unsigned short int *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_SIGNED_SHORT:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      subBuffer((short int *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      subBuffer((short int *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      subBuffer((short int *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      subBuffer((short int *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      subBuffer((short int *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      subBuffer((short int *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      subBuffer((short int *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      subBuffer((short int *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      subBuffer((short int *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      subBuffer((short int *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_UNSIGNED_INT:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      subBuffer((unsigned int *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      subBuffer((unsigned int *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      subBuffer((unsigned int *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      subBuffer((unsigned int *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      subBuffer((unsigned int *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      subBuffer((unsigned int *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      subBuffer((unsigned int *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      subBuffer((unsigned int *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      subBuffer((unsigned int *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      subBuffer((unsigned int *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_SIGNED_INT:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      subBuffer((int *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      subBuffer((int *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      subBuffer((int *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      subBuffer((int *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      subBuffer((int *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      subBuffer((int *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      subBuffer((int *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      subBuffer((int *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      subBuffer((int *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      subBuffer((int *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_UNSIGNED_LONG:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      subBuffer((unsigned LONGINT *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      subBuffer((unsigned LONGINT *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      subBuffer((unsigned LONGINT *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      subBuffer((unsigned LONGINT *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      subBuffer((unsigned LONGINT *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      subBuffer((unsigned LONGINT *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      subBuffer((unsigned LONGINT *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      subBuffer((unsigned LONGINT *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      subBuffer((unsigned LONGINT *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      subBuffer((unsigned LONGINT *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_SIGNED_LONG:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      subBuffer((LONGINT *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      subBuffer((LONGINT *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      subBuffer((LONGINT *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      subBuffer((LONGINT *) 0, (short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      subBuffer((LONGINT *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      subBuffer((LONGINT *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      subBuffer((LONGINT *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      subBuffer((LONGINT *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      subBuffer((LONGINT *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      subBuffer((LONGINT *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_FLOAT:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      subBuffer((float *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      subBuffer((float *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      subBuffer((float *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      subBuffer((float *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      subBuffer((float *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      subBuffer((float *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      subBuffer((float *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      subBuffer((float *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      subBuffer((float *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      subBuffer((float *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_DOUBLE:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      subBuffer((double *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      subBuffer((double *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      subBuffer((double *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      subBuffer((double *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      subBuffer((double *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      subBuffer((double *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      subBuffer((double *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      subBuffer((double *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      subBuffer((double *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      subBuffer((double *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_RGB:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      subScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      subScalarBuffertoVectorialBuffer((unsigned char *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      subScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      subScalarBuffertoVectorialBuffer((unsigned char *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      subScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      subScalarBuffertoVectorialBuffer((unsigned char *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      subScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      subScalarBuffertoVectorialBuffer((unsigned char *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      subScalarBuffertoVectorialBuffer((unsigned char *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      subScalarBuffertoVectorialBuffer((unsigned char *) 0, (double *) 0, inr, from);
      break;
    case Inrimage::WT_RGB:
      subVectorialBuffertoVectorialBuffer((unsigned char *) 0, (unsigned char *) 0, inr, from);
    default:
      break;
    }
    break;

  case Inrimage::WT_RGBA:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      subScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      subScalarBuffertoVectorialBuffer((unsigned char *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      subScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      subScalarBuffertoVectorialBuffer((unsigned char *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      subScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      subScalarBuffertoVectorialBuffer((unsigned char *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      subScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      subScalarBuffertoVectorialBuffer((unsigned char *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      subScalarBuffertoVectorialBuffer((unsigned char *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      subScalarBuffertoVectorialBuffer((unsigned char *) 0, (double *) 0, inr, from);
      break;
    case Inrimage::WT_RGBA:
      subVectorialBuffertoVectorialBuffer((unsigned char *) 0, (unsigned char *) 0, inr, from);
    default:
      break;
    }
    break;

  case Inrimage::WT_FLOAT_VECTOR:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      subScalarBuffertoVectorialBuffer((float *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      subScalarBuffertoVectorialBuffer((float *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      subScalarBuffertoVectorialBuffer((float *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      subScalarBuffertoVectorialBuffer((float *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      subScalarBuffertoVectorialBuffer((float *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      subScalarBuffertoVectorialBuffer((float *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      subScalarBuffertoVectorialBuffer((float *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      subScalarBuffertoVectorialBuffer((float *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      subScalarBuffertoVectorialBuffer((float *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      subScalarBuffertoVectorialBuffer((float *) 0, (double *) 0, inr, from);
      break;
    case Inrimage::WT_RGB:
      if(inr->getVdim() == 3)
	subVectorialBuffertoVectorialBuffer((float *) 0, (unsigned char *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    case Inrimage::WT_RGBA:
      if(inr->getVdim() == 4)
	subVectorialBuffertoVectorialBuffer((float *) 0, (unsigned char *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    case Inrimage::WT_FLOAT_VECTOR:
      if(inr->getVdim() == from->getVdim())
	subVectorialBuffertoVectorialBuffer((float *) 0, (float *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    case Inrimage::WT_DOUBLE_VECTOR:
      if(inr->getVdim() == from->getVdim())
	subVectorialBuffertoVectorialBuffer((float *) 0, (double *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_DOUBLE_VECTOR:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      subScalarBuffertoVectorialBuffer((double *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      subScalarBuffertoVectorialBuffer((double *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      subScalarBuffertoVectorialBuffer((double *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      subScalarBuffertoVectorialBuffer((double *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      subScalarBuffertoVectorialBuffer((double *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      subScalarBuffertoVectorialBuffer((double *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      subScalarBuffertoVectorialBuffer((double *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      subScalarBuffertoVectorialBuffer((double *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      subScalarBuffertoVectorialBuffer((double *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      subScalarBuffertoVectorialBuffer((double *) 0, (double *) 0, inr, from);
      break;
    case Inrimage::WT_RGB:
      if(inr->getVdim() == 3)
	subVectorialBuffertoVectorialBuffer((double *) 0, (unsigned char *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    case Inrimage::WT_RGBA:
      if(inr->getVdim() == 4)
	subVectorialBuffertoVectorialBuffer((double *) 0, (unsigned char *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    case Inrimage::WT_FLOAT_VECTOR:
      if(inr->getVdim() == from->getVdim())
	subVectorialBuffertoVectorialBuffer((double *) 0, (float *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    case Inrimage::WT_DOUBLE_VECTOR:
      if(inr->getVdim() == from->getVdim())
	subVectorialBuffertoVectorialBuffer((double *) 0, (double *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    default:
      break;
    }
    break;

  default:
    break;
  }
}



// multiplicate an inrimage to this
void mul(Inrimage *inr, const Inrimage *from) throw(Inrimage::InvalidArguments) {

  char _from[] = "yav::mul(const Inrimage *)";

  if(inr->getX() != from->getX() || inr->getY() != from->getY() || inr->getZ() != from->getZ())
    throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");

  switch(inr->getType()) {
  
  case Inrimage::WT_UNSIGNED_CHAR:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      mulBuffer((unsigned char *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      mulBuffer((unsigned char *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      mulBuffer((unsigned char *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      mulBuffer((unsigned char *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      mulBuffer((unsigned char *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      mulBuffer((unsigned char *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      mulBuffer((unsigned char *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      mulBuffer((unsigned char *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      mulBuffer((unsigned char *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      mulBuffer((unsigned char *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_SIGNED_CHAR:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      mulBuffer((char *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      mulBuffer((char *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      mulBuffer((char *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      mulBuffer((char *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      mulBuffer((char *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      mulBuffer((char *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      mulBuffer((char *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      mulBuffer((char *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      mulBuffer((char *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      mulBuffer((char *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_UNSIGNED_SHORT:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      mulBuffer((unsigned short int *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      mulBuffer((unsigned short int *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      mulBuffer((unsigned short int *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      mulBuffer((unsigned short int *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      mulBuffer((unsigned short int *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      mulBuffer((unsigned short int *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      mulBuffer((unsigned short int *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      mulBuffer((unsigned short int *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      mulBuffer((unsigned short int *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      mulBuffer((unsigned short int *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_SIGNED_SHORT:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      mulBuffer((short int *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      mulBuffer((short int *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      mulBuffer((short int *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      mulBuffer((short int *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      mulBuffer((short int *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      mulBuffer((short int *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      mulBuffer((short int *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      mulBuffer((short int *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      mulBuffer((short int *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      mulBuffer((short int *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_UNSIGNED_INT:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      mulBuffer((unsigned int *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      mulBuffer((unsigned int *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      mulBuffer((unsigned int *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      mulBuffer((unsigned int *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      mulBuffer((unsigned int *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      mulBuffer((unsigned int *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      mulBuffer((unsigned int *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      mulBuffer((unsigned int *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      mulBuffer((unsigned int *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      mulBuffer((unsigned int *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_SIGNED_INT:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      mulBuffer((int *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      mulBuffer((int *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      mulBuffer((int *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      mulBuffer((int *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      mulBuffer((int *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      mulBuffer((int *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      mulBuffer((int *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      mulBuffer((int *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      mulBuffer((int *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      mulBuffer((int *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_UNSIGNED_LONG:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      mulBuffer((unsigned LONGINT *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      mulBuffer((unsigned LONGINT *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      mulBuffer((unsigned LONGINT *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      mulBuffer((unsigned LONGINT *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      mulBuffer((unsigned LONGINT *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      mulBuffer((unsigned LONGINT *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      mulBuffer((unsigned LONGINT *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      mulBuffer((unsigned LONGINT *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      mulBuffer((unsigned LONGINT *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      mulBuffer((unsigned LONGINT *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_SIGNED_LONG:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      mulBuffer((LONGINT *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      mulBuffer((LONGINT *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      mulBuffer((LONGINT *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      mulBuffer((LONGINT *) 0, (short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      mulBuffer((LONGINT *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      mulBuffer((LONGINT *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      mulBuffer((LONGINT *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      mulBuffer((LONGINT *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      mulBuffer((LONGINT *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      mulBuffer((LONGINT *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_FLOAT:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      mulBuffer((float *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      mulBuffer((float *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      mulBuffer((float *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      mulBuffer((float *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      mulBuffer((float *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      mulBuffer((float *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      mulBuffer((float *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      mulBuffer((float *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      mulBuffer((float *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      mulBuffer((float *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_DOUBLE:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      mulBuffer((double *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      mulBuffer((double *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      mulBuffer((double *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      mulBuffer((double *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      mulBuffer((double *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      mulBuffer((double *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      mulBuffer((double *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      mulBuffer((double *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      mulBuffer((double *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      mulBuffer((double *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_RGB:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      mulScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      mulScalarBuffertoVectorialBuffer((unsigned char *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      mulScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      mulScalarBuffertoVectorialBuffer((unsigned char *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      mulScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      mulScalarBuffertoVectorialBuffer((unsigned char *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      mulScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      mulScalarBuffertoVectorialBuffer((unsigned char *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      mulScalarBuffertoVectorialBuffer((unsigned char *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      mulScalarBuffertoVectorialBuffer((unsigned char *) 0, (double *) 0, inr, from);
      break;
    case Inrimage::WT_RGB:
      mulVectorialBuffertoVectorialBuffer((unsigned char *) 0, (unsigned char *) 0, inr, from);
    default:
      break;
    }
    break;

  case Inrimage::WT_RGBA:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      mulScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      mulScalarBuffertoVectorialBuffer((unsigned char *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      mulScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      mulScalarBuffertoVectorialBuffer((unsigned char *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      mulScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      mulScalarBuffertoVectorialBuffer((unsigned char *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      mulScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      mulScalarBuffertoVectorialBuffer((unsigned char *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      mulScalarBuffertoVectorialBuffer((unsigned char *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      mulScalarBuffertoVectorialBuffer((unsigned char *) 0, (double *) 0, inr, from);
      break;
    case Inrimage::WT_RGBA:
      mulVectorialBuffertoVectorialBuffer((unsigned char *) 0, (unsigned char *) 0, inr, from);
    default:
      break;
    }
    break;

  case Inrimage::WT_FLOAT_VECTOR:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      mulScalarBuffertoVectorialBuffer((float *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      mulScalarBuffertoVectorialBuffer((float *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      mulScalarBuffertoVectorialBuffer((float *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      mulScalarBuffertoVectorialBuffer((float *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      mulScalarBuffertoVectorialBuffer((float *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      mulScalarBuffertoVectorialBuffer((float *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      mulScalarBuffertoVectorialBuffer((float *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      mulScalarBuffertoVectorialBuffer((float *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      mulScalarBuffertoVectorialBuffer((float *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      mulScalarBuffertoVectorialBuffer((float *) 0, (double *) 0, inr, from);
      break;
    case Inrimage::WT_RGB:
      if(inr->getVdim() == 3)
	mulVectorialBuffertoVectorialBuffer((float *) 0, (unsigned char *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    case Inrimage::WT_RGBA:
      if(inr->getVdim() == 4)
	mulVectorialBuffertoVectorialBuffer((float *) 0, (unsigned char *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    case Inrimage::WT_FLOAT_VECTOR:
      if(inr->getVdim() == from->getVdim())
	mulVectorialBuffertoVectorialBuffer((float *) 0, (float *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    case Inrimage::WT_DOUBLE_VECTOR:
      if(inr->getVdim() == from->getVdim())
	mulVectorialBuffertoVectorialBuffer((float *) 0, (double *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_DOUBLE_VECTOR:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      mulScalarBuffertoVectorialBuffer((double *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      mulScalarBuffertoVectorialBuffer((double *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      mulScalarBuffertoVectorialBuffer((double *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      mulScalarBuffertoVectorialBuffer((double *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      mulScalarBuffertoVectorialBuffer((double *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      mulScalarBuffertoVectorialBuffer((double *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      mulScalarBuffertoVectorialBuffer((double *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      mulScalarBuffertoVectorialBuffer((double *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      mulScalarBuffertoVectorialBuffer((double *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      mulScalarBuffertoVectorialBuffer((double *) 0, (double *) 0, inr, from);
      break;
    case Inrimage::WT_RGB:
      if(inr->getVdim() == 3)
	mulVectorialBuffertoVectorialBuffer((double *) 0, (unsigned char *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    case Inrimage::WT_RGBA:
      if(inr->getVdim() == 4)
	mulVectorialBuffertoVectorialBuffer((double *) 0, (unsigned char *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    case Inrimage::WT_FLOAT_VECTOR:
      if(inr->getVdim() == from->getVdim())
	mulVectorialBuffertoVectorialBuffer((double *) 0, (float *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    case Inrimage::WT_DOUBLE_VECTOR:
      if(inr->getVdim() == from->getVdim())
	mulVectorialBuffertoVectorialBuffer((double *) 0, (double *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    default:
      break;
    }
    break;

  default:
    break;
  }
}





// divide an inrimage to this
void div(Inrimage *inr, const Inrimage *from) throw(Inrimage::InvalidArguments) {

  char _from[] = "yav::div(const Inrimage *)";

  if(inr->getX() != from->getX() || inr->getY() != from->getY() || inr->getZ() != from->getZ())
    throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");

  switch(inr->getType()) {
  
  case Inrimage::WT_UNSIGNED_CHAR:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      divBuffer((unsigned char *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      divBuffer((unsigned char *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      divBuffer((unsigned char *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      divBuffer((unsigned char *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      divBuffer((unsigned char *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      divBuffer((unsigned char *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      divBuffer((unsigned char *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      divBuffer((unsigned char *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      divBuffer((unsigned char *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      divBuffer((unsigned char *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_SIGNED_CHAR:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      divBuffer((char *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      divBuffer((char *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      divBuffer((char *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      divBuffer((char *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      divBuffer((char *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      divBuffer((char *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      divBuffer((char *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      divBuffer((char *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      divBuffer((char *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      divBuffer((char *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_UNSIGNED_SHORT:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      divBuffer((unsigned short int *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      divBuffer((unsigned short int *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      divBuffer((unsigned short int *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      divBuffer((unsigned short int *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      divBuffer((unsigned short int *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      divBuffer((unsigned short int *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      divBuffer((unsigned short int *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      divBuffer((unsigned short int *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      divBuffer((unsigned short int *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      divBuffer((unsigned short int *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_SIGNED_SHORT:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      divBuffer((short int *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      divBuffer((short int *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      divBuffer((short int *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      divBuffer((short int *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      divBuffer((short int *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      divBuffer((short int *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      divBuffer((short int *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      divBuffer((short int *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      divBuffer((short int *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      divBuffer((short int *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_UNSIGNED_INT:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      divBuffer((unsigned int *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      divBuffer((unsigned int *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      divBuffer((unsigned int *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      divBuffer((unsigned int *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      divBuffer((unsigned int *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      divBuffer((unsigned int *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      divBuffer((unsigned int *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      divBuffer((unsigned int *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      divBuffer((unsigned int *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      divBuffer((unsigned int *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_SIGNED_INT:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      divBuffer((int *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      divBuffer((int *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      divBuffer((int *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      divBuffer((int *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      divBuffer((int *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      divBuffer((int *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      divBuffer((int *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      divBuffer((int *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      divBuffer((int *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      divBuffer((int *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_UNSIGNED_LONG:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      divBuffer((unsigned LONGINT *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      divBuffer((unsigned LONGINT *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      divBuffer((unsigned LONGINT *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      divBuffer((unsigned LONGINT *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      divBuffer((unsigned LONGINT *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      divBuffer((unsigned LONGINT *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      divBuffer((unsigned LONGINT *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      divBuffer((unsigned LONGINT *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      divBuffer((unsigned LONGINT *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      divBuffer((unsigned LONGINT *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_SIGNED_LONG:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      divBuffer((LONGINT *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      divBuffer((LONGINT *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      divBuffer((LONGINT *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      divBuffer((LONGINT *) 0, (short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      divBuffer((LONGINT *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      divBuffer((LONGINT *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      divBuffer((LONGINT *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      divBuffer((LONGINT *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      divBuffer((LONGINT *) 0, (float* ) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      divBuffer((LONGINT *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_FLOAT:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      divBuffer((float *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      divBuffer((float *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      divBuffer((float *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      divBuffer((float *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      divBuffer((float *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      divBuffer((float *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      divBuffer((float *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      divBuffer((float *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      divBuffer((float *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      divBuffer((float *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_DOUBLE:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      divBuffer((double *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      divBuffer((double *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      divBuffer((double *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      divBuffer((double *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      divBuffer((double *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      divBuffer((double *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      divBuffer((double *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      divBuffer((double *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      divBuffer((double *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      divBuffer((double *) 0, (double *) 0, inr, from);
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_RGB:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      divScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      divScalarBuffertoVectorialBuffer((unsigned char *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      divScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      divScalarBuffertoVectorialBuffer((unsigned char *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      divScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      divScalarBuffertoVectorialBuffer((unsigned char *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      divScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      divScalarBuffertoVectorialBuffer((unsigned char *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      divScalarBuffertoVectorialBuffer((unsigned char *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      divScalarBuffertoVectorialBuffer((unsigned char *) 0, (double *) 0, inr, from);
      break;
    case Inrimage::WT_RGB:
      divVectorialBuffertoVectorialBuffer((unsigned char *) 0, (unsigned char *) 0, inr, from);
    default:
      break;
    }
    break;

  case Inrimage::WT_RGBA:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      divScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      divScalarBuffertoVectorialBuffer((unsigned char *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      divScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      divScalarBuffertoVectorialBuffer((unsigned char *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      divScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      divScalarBuffertoVectorialBuffer((unsigned char *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      divScalarBuffertoVectorialBuffer((unsigned char *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      divScalarBuffertoVectorialBuffer((unsigned char *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      divScalarBuffertoVectorialBuffer((unsigned char *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      divScalarBuffertoVectorialBuffer((unsigned char *) 0, (double *) 0, inr, from);
      break;
    case Inrimage::WT_RGBA:
      divVectorialBuffertoVectorialBuffer((unsigned char *) 0, (unsigned char *) 0, inr, from);
    default:
      break;
    }
    break;

  case Inrimage::WT_FLOAT_VECTOR:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      divScalarBuffertoVectorialBuffer((float *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      divScalarBuffertoVectorialBuffer((float *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      divScalarBuffertoVectorialBuffer((float *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      divScalarBuffertoVectorialBuffer((float *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      divScalarBuffertoVectorialBuffer((float *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      divScalarBuffertoVectorialBuffer((float *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      divScalarBuffertoVectorialBuffer((float *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      divScalarBuffertoVectorialBuffer((float *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      divScalarBuffertoVectorialBuffer((float *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      divScalarBuffertoVectorialBuffer((float *) 0, (double *) 0, inr, from);
      break;
    case Inrimage::WT_RGB:
      if(inr->getVdim() == 3)
	divVectorialBuffertoVectorialBuffer((float *) 0, (unsigned char *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    case Inrimage::WT_RGBA:
      if(inr->getVdim() == 4)
	divVectorialBuffertoVectorialBuffer((float *) 0, (unsigned char *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    case Inrimage::WT_FLOAT_VECTOR:
      if(inr->getVdim() == from->getVdim())
	divVectorialBuffertoVectorialBuffer((float *) 0, (float *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    case Inrimage::WT_DOUBLE_VECTOR:
      if(inr->getVdim() == from->getVdim())
	divVectorialBuffertoVectorialBuffer((float *) 0, (double *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    default:
      break;
    }
    break;

  case Inrimage::WT_DOUBLE_VECTOR:
    switch(from->getType()) {
    case Inrimage::WT_UNSIGNED_CHAR:
      divScalarBuffertoVectorialBuffer((double *) 0, (unsigned char *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_CHAR:
      divScalarBuffertoVectorialBuffer((double *) 0, (char *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      divScalarBuffertoVectorialBuffer((double *) 0, (unsigned short int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      divScalarBuffertoVectorialBuffer((double *) 0,(short int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      divScalarBuffertoVectorialBuffer((double *) 0, (unsigned int *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_INT:
      divScalarBuffertoVectorialBuffer((double *) 0, (int *) 0, inr, from);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      divScalarBuffertoVectorialBuffer((double *) 0, (unsigned LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_SIGNED_LONG:
      divScalarBuffertoVectorialBuffer((double *) 0, (LONGINT *) 0, inr, from);
      break;
    case Inrimage::WT_FLOAT:
      divScalarBuffertoVectorialBuffer((double *) 0, (float *) 0, inr, from);
      break;
    case Inrimage::WT_DOUBLE:
      divScalarBuffertoVectorialBuffer((double *) 0, (double *) 0, inr, from);
      break;
    case Inrimage::WT_RGB:
      if(inr->getVdim() == 3)
	divVectorialBuffertoVectorialBuffer((double *) 0, (unsigned char *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    case Inrimage::WT_RGBA:
      if(inr->getVdim() == 4)
	divVectorialBuffertoVectorialBuffer((double *) 0, (unsigned char *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    case Inrimage::WT_FLOAT_VECTOR:
      if(inr->getVdim() == from->getVdim())
	divVectorialBuffertoVectorialBuffer((double *) 0, (float *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    case Inrimage::WT_DOUBLE_VECTOR:
      if(inr->getVdim() == from->getVdim())
	divVectorialBuffertoVectorialBuffer((double *) 0, (double *) 0, inr, from);
      else
	throw Inrimage::InvalidArguments(_from, "another image type was expected in this method");
      break;
    default:
      break;
    }
    break;

  default:
    break;
  }
}


void logicalNot(Inrimage *inr){
  switch(inr->getType()) {
  case Inrimage::WT_UNSIGNED_CHAR:
    notBuffer((unsigned char *) 0, inr);
    break;
  case Inrimage::WT_SIGNED_CHAR:
    notBuffer((char *) 0, inr);
    break;
  case Inrimage::WT_UNSIGNED_SHORT:
    notBuffer((unsigned short int *) 0, inr);
    break;
  case Inrimage::WT_SIGNED_SHORT:
    notBuffer((short *) 0, inr);
    break;
  case Inrimage::WT_UNSIGNED_INT:
    notBuffer((unsigned int *) 0, inr);
    break;
  case Inrimage::WT_SIGNED_INT:
    notBuffer((int *) 0, inr);
    break;
  case Inrimage::WT_UNSIGNED_LONG:
    notBuffer((unsigned LONGINT *) 0, inr);
    break;
  case Inrimage::WT_SIGNED_LONG:
    notBuffer((LONGINT *) 0, inr);
    break;
  default:
    break;
  }  

}



void logicalOr(Inrimage *inr, const Inrimage *from) throw(Inrimage::InvalidArguments) {
  if (from->getType() != inr->getType())
    throw Inrimage::InvalidArguments("inrimage::logicalAnd(Inrimage *inr, const Inrimage *)",
				     "another image type was expected in this method");
  if (inr->getX() != from->getX() || inr->getY() != from->getY() || inr->getZ() != from->getZ())
    throw Inrimage::InvalidArguments("inrimage::logicalAnd(Inrimage *inr, const Inrimage *)",
				     "argument image size does not match this image size");
  
  switch (inr->getType()) {
  case Inrimage::WT_UNSIGNED_CHAR:
    orBuffer((unsigned char *) 0, inr, from);
    break;
  case Inrimage::WT_SIGNED_CHAR:
    orBuffer((char *) 0, inr, from);
    break;
  case Inrimage::WT_UNSIGNED_SHORT:
    orBuffer((unsigned short int *) 0, inr, from);
    break;
  case Inrimage::WT_SIGNED_SHORT:
    orBuffer((short *) 0, inr, from);
    break;
  case Inrimage::WT_UNSIGNED_INT:
    orBuffer((unsigned int *) 0, inr, from);
    break;
  case Inrimage::WT_SIGNED_INT:
    orBuffer((int *) 0, inr, from);
    break;
  case Inrimage::WT_UNSIGNED_LONG:
    orBuffer((unsigned LONGINT *) 0, inr, from);
    break;
  case Inrimage::WT_SIGNED_LONG:
    orBuffer((LONGINT *) 0, inr, from);
    break;
  default:
    break;
  }  

}

void logicalAnd(Inrimage *inr, const Inrimage *from) throw(Inrimage::InvalidArguments) {

  if(inr->getType() != from->getType())
    throw Inrimage::InvalidArguments("inrimage::logicalAnd(Inrimage *inr, const Inrimage *)",
			   "another image type was expected in this method");
  if(inr->getX() != from->getX() || inr->getY() != from->getY() || inr->getZ() != from->getZ())
    throw Inrimage::InvalidArguments("inrimage::logicalAnd(Inrimage *inr, const Inrimage *)",
			   "argument image size does not match this image size");

  switch(inr->getType()) {
  case Inrimage::WT_UNSIGNED_CHAR:
    andBuffer((unsigned char *) 0, inr, from);
    break;
  case Inrimage::WT_SIGNED_CHAR:
    andBuffer((char *) 0, inr, from);
    break;
  case Inrimage::WT_UNSIGNED_SHORT:
    andBuffer((unsigned short int *) 0, inr, from);
    break;
  case Inrimage::WT_SIGNED_SHORT:
    andBuffer((short *) 0, inr, from);
    break;
  case Inrimage::WT_UNSIGNED_INT:
    andBuffer((unsigned int *) 0, inr, from);
    break;
  case Inrimage::WT_SIGNED_INT:
    andBuffer((int *) 0, inr, from);
    break;
  case Inrimage::WT_UNSIGNED_LONG:
    andBuffer((unsigned LONGINT *) 0, inr, from);
    break;
  case Inrimage::WT_SIGNED_LONG:
    andBuffer((LONGINT *) 0, inr, from);
    break;
  default:
    break;
  }

}

void logicalXor(Inrimage *inr, const Inrimage *from) throw(Inrimage::InvalidArguments){

  if(inr->getType() != from->getType())
    throw Inrimage::InvalidArguments("inrimage::logivalXor(Inrimage *inr, const Inrimage *)",
				     "another image type was expected in this method");
  if(inr->getX() != from->getX() || inr->getY() != from->getY() || inr->getZ() != from->getZ())
    throw Inrimage::InvalidArguments("inrimag::logicalXor(Inrimage *inr, const Inrimage *)",
				     "argument image size does not match this image size");
  
  switch(inr->getType()) {
  case Inrimage::WT_UNSIGNED_CHAR:
    xorBuffer((unsigned char *) 0, inr, from);
    break;
  case Inrimage::WT_SIGNED_CHAR:
    xorBuffer((char *) 0, inr, from);
    break;
  case Inrimage::WT_UNSIGNED_SHORT:
    xorBuffer((unsigned short int *) 0, inr, from);
    break;
  case Inrimage::WT_SIGNED_SHORT:
    xorBuffer((short *) 0, inr, from);
    break;
  case Inrimage::WT_UNSIGNED_INT:
    xorBuffer((unsigned int *) 0, inr, from);
    break;
  case Inrimage::WT_SIGNED_INT:
    xorBuffer((int *) 0, inr, from);
    break;
  case Inrimage::WT_UNSIGNED_LONG:
    xorBuffer((unsigned LONGINT *) 0, inr, from);
    break;
  case Inrimage::WT_SIGNED_LONG:
    xorBuffer((LONGINT *) 0, inr, from);
    break;
  default:
    break;
  }  

}

}
