#ifndef BACKGROUNDOVERLAYREDRAW_H
#define BACKGROUNDOVERLAYREDRAW_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/backgroundOverlayRedraw.h"
namespace yav {


  /// template function for drawing an xy slice
template <class B> void drawOverlayXY(B,Inrimage *inr,
				      Inrimage *inr2, 
				      SliceCamera *slicecam,
				      unsigned int visible[6], 
				      double min, double max,
				      double min2, double max2,
				      Tk_PhotoImageBlock *block,
				      const unsigned char *redColorMap, 
				      const unsigned char *greenColorMap, 
				      const unsigned char *blueColorMap,
				      const unsigned char *redColorMap2, 
				      const unsigned char *greenColorMap2, 
				      const unsigned char *blueColorMap2,
				      unsigned int sx,unsigned int sy, 
				      double rx,double ry, double opacity){

    unsigned int plane, i, j, x, y, *offset, o;
    register int r1,r2;
    double a1,a2, dr;
    float p = (float) opacity;
    float q = 1-p;  

    plane = inr->getX() * inr->getY();
    if(slicecam->getGeometry() == SliceCamera::SLICE_ALPHA) {
      /*o = (((unsigned int) (slicecam->getSlice() * inr->getZ() / 180.0 + 0.5))
	% inr->getZ()) * plane;*/
      o = (((unsigned int) slicecam->getSlice()) % inr->getZ()) * plane;
    }
    else
      o = (unsigned int) slicecam->getSlice() * plane;
    
    offset = new unsigned int[sx];
    for(i = 0; i < sx; i++) offset[i] = (int) ((double) i / rx) + visible[0];
    
    a1 = (double) max - (double) min;
    if(a1 == 0.0) a1 = 255.0;
    else a1 = 255.0 / a1;
    
    a2 = max2 -min2;
    if(a2 == 0.0) a2 = 255.0;
    else a2 = 255.0 / a2;
    
    for(y = 0, j = 0; y < sy; y++) {
      i = o + ((int) ((double) y / ry) + visible[1]) * inr->getX();
      for(x = 0; x < sx; x++) {
	dr = (double) ((B *) inr->getData())[offset[x] + i];
	if(dr < (double) min) r1 = 0;
	else if(dr > (double) max) r1 = 256;
	else r1 = (int) (a1 * (dr - min) + 0.5);
	
	dr = (double) ((B *) inr2->getData())[offset[x] + i];
	if(dr < (double) min2) r2 = 0;
	else if(dr > (double) max2) r2 = 256;
	else r2 = (int) (a2 * (dr - min2) + 0.5);
	
	block->pixelPtr[j++] = (unsigned char)(p*redColorMap[r1] +q*redColorMap2[r2]);
	block->pixelPtr[j++] = (unsigned char)(p*greenColorMap[r1] +q*greenColorMap2[r2]);
	block->pixelPtr[j++] = (unsigned char)(p*blueColorMap[r1] +q*blueColorMap2[r2]);
      }
    }

delete[] offset;
}

template <class B> void drawOverlayXZ(B,Inrimage *inr,
				      Inrimage *inr2, 
				      SliceCamera *slicecam,
				      unsigned int visible[6], 
				      double min, double max,
				      double min2, double max2,
				      Tk_PhotoImageBlock *block,
				      const unsigned char *redColorMap, 
				      const unsigned char *greenColorMap, 
				      const unsigned char *blueColorMap,
				      const unsigned char *redColorMap2, 
				      const unsigned char *greenColorMap2, 
				      const unsigned char *blueColorMap2,
				      unsigned int sx,unsigned int sz, 
				      double rx,double rz, double opacity) {
    unsigned int plane, i, j, x, z, *offset, o;
    register int r1,r2;
    double a1,a2, dr;
    float p= (float) opacity;
    float q= 1-p;  
    
    a1 = (double) max - (double) min;
    if(a1 == 0.0) a1 = 255.0;
    else a1 = 255.0 / a1;
    
    a2 = max2 -min2;
    if(a2 == 0.0) a2 = 255.0;
    else a2 = 255.0 / a2;
    
    plane = inr->getX() * inr->getY();
    offset = new unsigned int[sx];
    for(i = 0; i < sx; i++) offset[i] = (int) ((double) i / rx) + visible[0];
    o = (unsigned int) slicecam->getSlice() * inr->getX();

    for(z = 0, j = 0; z < sz; z++) {
      i = ((int)((double) z / rz) + visible[2]) * plane + o;
      for(x = 0; x < sx; x++) {
	dr = (double) ((B *) inr->getData())[offset[x] + i];
	if(dr < min) r1 = 0;
	else if(dr > max) r1 = 256;
	else r1 = (int) (a1 * (dr - min) + 0.5);

	dr = (double) ((B *) inr2->getData())[offset[x] + i];
	if(dr < (double) min2) r2 = 0;
	else if(dr > (double) max2) r2 = 256;
	else r2 = (int) (a2 * (dr - min2) + 0.5);

	block->pixelPtr[j++] = (unsigned char)(p*redColorMap[r1] +q*redColorMap2[r2]);
	block->pixelPtr[j++] = (unsigned char)(p*greenColorMap[r1] +q*greenColorMap2[r2]);
	block->pixelPtr[j++] = (unsigned char)(p*blueColorMap[r1] +q*blueColorMap2[r2]);
      }
    }

  delete[] offset;
}

  template <class B> void drawOverlayYZ(B,Inrimage *inr,
					Inrimage *inr2, 
				      SliceCamera *slicecam,
				      unsigned int visible[6], 
				      double min, double max,
				      double min2, double max2,
				      Tk_PhotoImageBlock *block,
				      const unsigned char *redColorMap, 
				      const unsigned char *greenColorMap, 
				      const unsigned char *blueColorMap,
				      const unsigned char *redColorMap2, 
				      const unsigned char *greenColorMap2, 
				      const unsigned char *blueColorMap2,
				      unsigned int sy,unsigned int sz, 
				      double ry,double rz, double opacity){
    unsigned int plane, i, j, y, z, *offset, o;
    register int r1,r2;
    double  dr;
    float p= (float)opacity;
    float q= 1-p;  
    
    plane = inr->getX() * inr->getY();
    o = (unsigned int) slicecam->getSlice();
    
    offset = new unsigned int[sz];
    for(i = 0; i < sz; i++) offset[i] = ((int) ((double) i / rz) + visible[2]) * plane;
    
    double a1 = (double) max - (double) min;
    if(a1 == 0.0) a1 = 255.0;
    else a1 = 255.0 / a1;
    
    double a2 = max2 -min2;
    if(a2 == 0.0) a2 = 255.0;
    else a2 = 255.0 / a2;
    
    for(y = 0, j = 0; y < sy; y++) {
      i = ((int)((double) y / ry) + visible[1]) * inr->getX() + o;
      for(z = 0; z < sz; z++) {
	dr = (double) ((B *) inr->getData())[offset[z] + i];
	if(dr < min) r1 = 0;
	else if(dr > max) r1 = 256;
	else r1 = (int) (a1 * (dr - min) + 0.5);

	dr = (double) ((B *) inr2->getData())[offset[z] + i];
	if(dr < (double) min2) r2 = 0;
	else if(dr > (double) max2) r2 = 256;
	else r2 = (int) (a2 * (dr - min2) + 0.5);

	block->pixelPtr[j++] = (unsigned char)(p*redColorMap[r1] +q*redColorMap2[r2]);
	block->pixelPtr[j++] = (unsigned char)(p*greenColorMap[r1] +q*greenColorMap2[r2]);
	block->pixelPtr[j++] = (unsigned char)(p*blueColorMap[r1] +q*blueColorMap2[r2]);
      }
    }

  delete[] offset;
}

  template <class B> void drawOverlayXYVector(B,Inrimage *inr,
					      Inrimage *inr2, 
				      SliceCamera *slicecam,
				      unsigned int visible[6], 
				      double min, double max,
				      double min2, double max2,
				      Tk_PhotoImageBlock *block,
				      const unsigned char *redColorMap, 
				      const unsigned char *greenColorMap, 
				      const unsigned char *blueColorMap,
				      const unsigned char *redColorMap2, 
				      const unsigned char *greenColorMap2, 
				      const unsigned char *blueColorMap2,
				      unsigned int sx,unsigned int sy, 
				      double rx,double ry, double opacity){
    unsigned int i, j, x, y, *offset, o;
    register int r1,r2;
    double a1,a2, dr, *norm,*norm2;

    float p= (float) opacity;
    float q= 1-p;  
    offset = new unsigned int[sx];
    for(i = 0; i < sx; i++) offset[i] = (int) ((double) i / rx);
    if(slicecam->getGeometry() == SliceCamera::SLICE_ALPHA) {
      /*o = (((unsigned int) (slicecam->getSlice() * inr->getZ() / 180.0 + 0.5))
	% inr->getZ());*/
      o = ((unsigned int) slicecam->getSlice()) % inr->getZ();
    }
    else {
      o = (unsigned int) slicecam->getSlice();
    }
    
    j = (visible[3] - visible[0] + 1) * (visible[4] - visible[1] + 1);
    norm = new double[j];
    norm2 = new double[j];
    for(y = visible[1]; y <= visible[4]; y++) {
      for(x = visible[0]; x <= visible[3]; x++, norm++) {
	*norm = 0.0;
	*norm2 = 0.0;
	for(i = 0; i < inr->getVdim(); i++)
	  {
	    *norm += ((B ***) inr->getArray())[o][y][x * inr->getVdim() + i]
	      * ((B ***) inr->getArray())[o][y][x * inr->getVdim() + i];
	    *norm2 += ((B ***) inr2->getArray())[o][y][x * inr2->getVdim() + i]
	    * ((B ***) inr2->getArray())[o][y][x * inr2->getVdim() + i];
	  }
	*norm = sqrt(*norm);
	*norm2 = sqrt(*norm2);
      }
    }

    a1 = (double) max - (double) min;
    if(a1 == 0.0) a1 = 255.0;
    else a1 = 255.0 / a1;
    
    a2 = max2 -min2;
    if(a2 == 0.0) a2 = 255.0;
    else a2 = 255.0 / a2;

    norm -= j;
    

    for(y = 0, j = 0; y < sy; y++) {
      i = ((int) ((double) y / ry)) * (visible[3] - visible[0] + 1);
      for(x = 0; x < sx; x++) {
	dr = norm[offset[x] + i];
	if(dr < min) r1 = 0;
	else if(dr > max) r1 = 256;
	else r1 = (int) (a1 * (dr - min) + 0.5);

	dr = norm2[offset[x] + i];
	if(dr < (double) min2) r2 = 0;
	else if(dr > (double) max2) r2 = 256;
	else r2 = (int) (a2 * (dr - min2) + 0.5);

	block->pixelPtr[j++] = (unsigned char)(p*redColorMap[r1] +q*redColorMap2[r2]);
	block->pixelPtr[j++] = (unsigned char)(p*greenColorMap[r1] +q*greenColorMap2[r2]);
	block->pixelPtr[j++] = (unsigned char)(p*blueColorMap[r1] +q*blueColorMap2[r2]);
      }
    }
    
    delete[] offset;
    delete[] norm;
    delete[] norm2;
  }

  template <class B> void drawOverlayXZVector(B,Inrimage *inr,
					      Inrimage *inr2, 
				      SliceCamera *slicecam,
				      unsigned int visible[6], 
				      double min, double max,
				      double min2, double max2,
				      Tk_PhotoImageBlock *block,
				      const unsigned char *redColorMap, 
				      const unsigned char *greenColorMap, 
				      const unsigned char *blueColorMap,
				      const unsigned char *redColorMap2, 
				      const unsigned char *greenColorMap2, 
				      const unsigned char *blueColorMap2,
				      unsigned int sx,unsigned int sz, 
				      double rx,double rz, double opacity){
    unsigned int i, j, x, z, *offset, o;
    register int r1,r2;
    double a1,a2, dr, *norm,*norm2;
    float p= (float) opacity;
    float q= 1-p;  

    offset = new unsigned int[sx];
    for(i = 0; i < sx; i++) offset[i] = (int) ((double) i / rx);
    o = (unsigned int) slicecam->getSlice();
    
    j = (visible[5] - visible[2] + 1) * (visible[3] - visible[0] + 1);
    norm = new double[j];
    norm2 = new double[j];
    for(z = visible[2]; z <= visible[5]; z++) {
      for(x = visible[0]; x <= visible[3]; x++, norm++) {
	*norm = 0.0;
	*norm2 = 0.0;
	for(i = 0; i < inr->getVdim(); i++)
	  {
	    *norm += ((B ***) inr->getArray())[z][o][inr->getVdim() * x + i]
	      * ((B ***) inr->getArray())[z][o][inr->getVdim() * x + i];
	    *norm2 += ((B ***) inr2->getArray())[z][o][inr2->getVdim() * x + i]
	      * ((B ***) inr2->getArray())[z][o][inr2->getVdim() * x + i];
	  }
	*norm = sqrt(*norm);
	*norm2 = sqrt(*norm2);
      }
    }
    
    
    a1 = (double) max - (double) min;
    if(a1 == 0.0) a1 = 255.0;
    else a1 = 255.0 / a1;
    
    a2 = max2 -min2;
    if(a2 == 0.0) a2 = 255.0;
    else a2 = 255.0 / a2;

    norm -= j;
	
    for(z = 0, j = 0; z < sz; z++) {
      i = ((int)((double) z / rz)) * (visible[3] - visible[0] + 1);
      for(x = 0; x < sx; x++) {
	dr = norm[offset[x] + i];
	if(dr < min) r1 = 0;
	else if(dr > max) r1 = 256;
	else r1 = (int) (a1 * (dr - min) + 0.5);

	dr = norm2[offset[x] + i];
	if(dr < (double) min2) r2 = 0;
	else if(dr > (double) max2) r2 = 256;
	else r2 = (int) (a2 * (dr - min2) + 0.5);

	block->pixelPtr[j++] = (unsigned char)(p*redColorMap[r1] +q*redColorMap2[r2]);
	block->pixelPtr[j++] = (unsigned char)(p*greenColorMap[r1] +q*greenColorMap2[r2]);
	block->pixelPtr[j++] = (unsigned char)(p*blueColorMap[r1] +q*blueColorMap2[r2]);
      }
    }

  delete[] norm;
  delete[] norm2;
  delete[] offset;
}


  template <class B> void drawOverlayYZVector(B,Inrimage *inr,
					      Inrimage *inr2, 
				      SliceCamera *slicecam,
				      unsigned int visible[6], 
				      double min, double max,
				      double min2, double max2,
				      Tk_PhotoImageBlock *block,
				      const unsigned char *redColorMap, 
				      const unsigned char *greenColorMap, 
				      const unsigned char *blueColorMap,
				      const unsigned char *redColorMap2, 
				      const unsigned char *greenColorMap2, 
				      const unsigned char *blueColorMap2,
				      unsigned int sy,unsigned int sz, 
				      double ry,double rz, double opacity) {
    unsigned int i, j, y, z, *offset, o;
    register int r1,r2;
    double a1,a2, dr, *norm,*norm2;
    float p= (float)opacity;
    float q= 1-p;  

    o = (unsigned int) slicecam->getSlice();
    offset = new unsigned int[sz];
    for(i = 0; i < sz; i++) offset[i] = (int) ((double) i / rz);
    
  j = (visible[5] - visible[2] + 1) * (visible[4] - visible[1] + 1);
  norm = new double[j];
  norm2 = new double[j];
  for(y = visible[1]; y <= visible[4]; y++) {
    for(z = visible[2]; z <= visible[5]; z++, norm++) {
      *norm = 0.0;
      *norm2 = 0.0;
      for(i = 0; i < inr->getVdim(); i++)
	{
	  *norm += ((B ***) inr->getArray())[z][y][inr->getVdim() * o + i]
	    * ((B ***) inr->getArray())[z][y][inr->getVdim() * o + i];
	  *norm2 += ((B ***) inr2->getArray())[z][y][inr2->getVdim() * o + i]
	    * ((B ***) inr2->getArray())[z][y][inr2->getVdim() * o + i];
	}
      *norm = sqrt(*norm);
      *norm2 = sqrt(*norm2);
    }
  }


    norm -= j;
           
    a1 = (double) max - (double) min;
    if(a1 == 0.0) a1 = 255.0;
    else a1 = 255.0 / a1;
    
    a2 = max2 -min2;
    if(a2 == 0.0) a2 = 255.0;
    else a2 = 255.0 / a2;


    for(y = 0, j = 0; y < sy; y++) {
      i = ((int)((double) y / ry)) * (visible[5] - visible[2] + 1);
      for(z = 0; z < sz; z++) {
	dr = norm[offset[z] + i];
	if(dr < min) r1 = 0;
	else if(dr > max) r1 = 256;
	else r1 = (int) (a1 * (dr - min) + 0.5);

	dr = norm2[offset[z] + i];
	if(dr < (double) min2) r2 = 0;
	else if(dr > (double) max2) r2 = 256;
	else r2 = (int) (a2 * (dr - min2) + 0.5);

	block->pixelPtr[j++] = (unsigned char)(p*redColorMap[r1] +q*redColorMap2[r2]);
	block->pixelPtr[j++] = (unsigned char)(p*greenColorMap[r1] +q*greenColorMap2[r2]);
	block->pixelPtr[j++] = (unsigned char)(p*blueColorMap[r1] +q*blueColorMap2[r2]);
      }
    }

  delete[] norm;
  delete[] norm2;
  delete[] offset;
}


}
#endif
