#ifndef BACKGROUNDREDRAW_H
#define BACKGROUNDREDRAW_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/backgroundRedraw.h"

#include <inrimage/TclInrimageView.h>

namespace yav {

  template <class B> void drawXY(B,Inrimage *inr,SliceCamera *slicecam,
				 unsigned int visible[6], 
				 double min, double max,
				 Tk_PhotoImageBlock *block,
				 const unsigned char *redColorMap, 
				 const unsigned char *greenColorMap, 
				 const unsigned char *blueColorMap,
				 unsigned int sx,unsigned int sy, 
				 double rx,double ry,
				 BasicInrimageBackground::IMAGE_INTERPOLATION inter){

    unsigned int plane, i, j, x, y, *offset, o;
    register int r;
    double a, dr;
   
    plane = inr->getX() * inr->getY();
    o = (unsigned int) slicecam->getSlice() * plane;
   
    if (inter==BasicInrimageBackground::NEAREST) {
      offset = new unsigned int[sx];
      for(i = 0; i < sx; i++) offset[i] = (int) ((double) i / rx) + visible[0];
     
     
      a = (double) max - (double) min;
      if(a == 0.0) a = 255.0;
      else a = 255.0 / a;
      for(y = 0, j = 0; y < sy; y++) {
	i = o + ((int) ((double) y / ry) + visible[1]) * inr->getX();
	for(x = 0; x < sx; x++) {
	  dr = (double) ((B *) inr->getData())[offset[x] + i];
	  if(dr < (double) min) r = 0;
	  else if(dr > (double) max) r = 256;
	  else r = (int) (a * (dr - min) + 0.5);
	  block->pixelPtr[j++] = redColorMap[r];
	  block->pixelPtr[j++] = greenColorMap[r];
	  block->pixelPtr[j++] = blueColorMap[r];
	}
      }
      delete[] offset;
    } 
    else if (inter==BasicInrimageBackground::BILINEAR) {
      offset = new unsigned int[sx];
      double *weightx= new double[sx];
      unsigned int posy;
      unsigned int dimx=inr->getX();
      unsigned int dimy=inr->getY()-1;
     
      unsigned int incrx,incry;
      double weighty,val00,val01,val10,val11,p;
      for(i = 0; i < sx; i++) {
	p=(((double) i / rx)-0.5);
	offset[i] = (int) p;
 	if (p<0) {
	  offset[i]=visible[0];
	  weightx[i]=0;
	}
	else {
	  offset[i]+=visible[0];
	  weightx[i]=p-((int)p);
	}
      }
      a = (double) max - (double) min;
      if(a == 0.0) a = 255.0;
      else a = 255.0 / a;
      for(y = 0, j = 0; y < sy; y++) {
	p= (((double) y / ry)-0.5);
	posy=(int) p;
	if (p<0) {
	  posy= visible[1];
	  weighty=0;
	}
	else {
	  posy+=visible[1];
	  weighty=p-((int)p);
	}
	i = o + posy * dimx;

	if (posy<dimy)
	  incry=dimx;
	else
	  incry=0;

	for(x = 0; x < sx; x++) {
	  if (offset[x]<(dimx-1))
	    incrx=1;
	  else
	    incrx=0;
	  val00=(double) ((B *) inr->getData())[offset[x] + i];
	  val10=(double) ((B *) inr->getData())[offset[x] + incrx + i];
	  val01=(double) ((B *) inr->getData())[offset[x] + i + incry];
	  val11=(double) ((B *) inr->getData())[offset[x] + incrx + i + incry];
	  dr=val00*(1-weightx[x])*(1-weighty)+val10*weightx[x]*(1-weighty)+
	    val01*weighty*(1-weightx[x])+val11*weighty*weightx[x];
	  if(dr < (double) min) r = 0;
	  else if(dr > (double) max) r = 256;
	  else r = (int) (a * (dr - min) + 0.5);
	  block->pixelPtr[j++] = redColorMap[r];
	  block->pixelPtr[j++] = greenColorMap[r];
	  block->pixelPtr[j++] = blueColorMap[r];
	}
      }
      delete[] offset;
      delete[] weightx;
    }
  }
  template <class B> void drawXZ(B,Inrimage *inr,SliceCamera *slicecam,
				 unsigned int visible[6], 
				 double min, double max,
				 Tk_PhotoImageBlock *block,
				 const unsigned char *redColorMap, 
				 const unsigned char *greenColorMap, 
				 const unsigned char *blueColorMap,
				 unsigned int sx,unsigned int sz, 
				 double rx,double rz,BasicInrimageBackground::IMAGE_INTERPOLATION inter){
    
    unsigned int plane, i, j, x, z, *offset, o;
    register int r;
    double a, dr;
    
    plane = inr->getX() * inr->getY();
    if (inter==BasicInrimageBackground::NEAREST) {    
      offset = new unsigned int[sx];
      for(i = 0; i < sx; i++) offset[i] = (int) ((double) i / rx) + visible[0];
      o = (unsigned int) slicecam->getSlice() * inr->getX();
      a = max - min;
      if(a == 0.0) a = 255.0;
      else a = 255.0 / a;
      for(z = 0, j = 0; z < sz; z++) {
	i = ((int)((double) z / rz) + visible[2]) * plane + o;
	for(x = 0; x < sx; x++) {
	  dr = (double) ((B *) inr->getData())[offset[x] + i];
	  if(dr < min) r = 0;
	  else if(dr > max) r = 256;
	  else r = (int) (a * (dr - min) + 0.5);
	  block->pixelPtr[j++] = redColorMap[r];
	  block->pixelPtr[j++] = greenColorMap[r];
	  block->pixelPtr[j++] = blueColorMap[r];
	}
      }
      
      delete[] offset;
    }
    else if (inter==BasicInrimageBackground::BILINEAR) {
      offset = new unsigned int[sx];
      unsigned int posz;
 
      unsigned int dimx=inr->getX()-1;
      unsigned int dimz=inr->getZ()-1;

      double *weightx= new double[sx];
      double weighty,val00,val01,val10,val11,p;
      unsigned int incrx,incrz;
      for(i = 0; i < sx; i++) {	
	p=(((double) i / rx)-0.5);
	offset[i] = (int) p;
 	if (p<0)
	  {
	    offset[i]=visible[0];
	    weightx[i]=0.0;
	  }
	else {
	  offset[i]+=visible[0];
	  weightx[i]=p-((int)p);
	}

	
      }
      o = (unsigned int) slicecam->getSlice() * inr->getX();
      a = (double) max - (double) min;
      if(a == 0.0) a = 255.0;
      else a = 255.0 / a;
      for(z = 0, j = 0; z < sz; z++) {
	p= (((double) z / rz)-0.5);
	posz=(int)p;
	if (p<0)
	  {
	    posz= visible[2];
	    weighty=0;
	  }
	else {
	  posz+=visible[2];
	  weighty=p-((int)p);
	}
	i = posz * plane + o;

	if (posz<dimz)
	  incrz=plane;
	else
	  incrz=0;
	for(x = 0; x < sx; x++) {
	  if (offset[x]<dimx)
	    incrx=1;
	  else
	    incrx=0;
	  val00=(double) ((B *) inr->getData())[offset[x] + i];
	  val10=(double) ((B *) inr->getData())[offset[x] + incrx + i];
	  val01=(double) ((B *) inr->getData())[offset[x] + i + incrz];
	  val11=(double) ((B *) inr->getData())[offset[x] + incrx + i + incrz];
	  dr=val00*(1-weightx[x])*(1-weighty)+val10*weightx[x]*(1-weighty)+
	    val01*weighty*(1-weightx[x])+val11*weighty*weightx[x];
	  if(dr < (double) min) r = 0;
	  else if(dr > (double) max) r = 256;
	  else r = (int) (a * (dr - min) + 0.5);
	  block->pixelPtr[j++] = redColorMap[r];
	  block->pixelPtr[j++] = greenColorMap[r];
	  block->pixelPtr[j++] = blueColorMap[r];
	}
      }
      delete[] offset;
      delete[] weightx;
      
    }
  }
  template <class B> void drawYZ(B,Inrimage *inr,SliceCamera *slicecam,
				 unsigned int visible[6], 
				 double min, double max,
				 Tk_PhotoImageBlock *block,
				 const unsigned char *redColorMap, 
				 const unsigned char *greenColorMap, 
				 const unsigned char *blueColorMap,
				 unsigned int sy,unsigned int sz, 
				 double ry,double rz,BasicInrimageBackground::IMAGE_INTERPOLATION inter){
    
    unsigned int plane, i, j, y, z, *offset, o;
    register int r;
    double a, dr;
    
    plane = inr->getX() * inr->getY();
    o = (unsigned int) slicecam->getSlice();
    if (inter==BasicInrimageBackground::NEAREST) {        
      offset = new unsigned int[sz];
      for(i = 0; i < sz; i++) offset[i] = ((int) ((double) i / rz) + visible[2]) * plane;
      
      a = max - min;
      if(a == 0.0) a = 255.0;
      else a = 255.0 / a;
      for(y = 0, j = 0; y < sy; y++) {
	i = ((int)((double) y / ry) + visible[1]) * inr->getX() + o;
	for(z = 0; z < sz; z++) {
	  dr = (double) ((B *) inr->getData())[offset[z] + i];
	  if(dr < min) r = 0;
	  else if(dr > max) r = 256;
	  else r = (int) (a * (dr - min) + 0.5);
	  block->pixelPtr[j++] = redColorMap[r];
	  block->pixelPtr[j++] = greenColorMap[r];
	  block->pixelPtr[j++] = blueColorMap[r];
	}
      }
      
      delete[] offset;
    }
    else if (inter==BasicInrimageBackground::BILINEAR) {
      offset = new unsigned int[sz];
      unsigned int *posz = new unsigned int[sz];

      unsigned int dimx=inr->getX();
      unsigned int dimy=inr->getY()-1;
      unsigned int dimz=inr->getZ()-1;

      unsigned int incry,incrz,posy,base;
      double *weightz= new double[sz];
      double weighty,val00,val01,val10,val11,p;

      for(i = 0; i < sz; i++) {
	p=((double) i / rz)-0.5;
	posz[i]=(int) p;
	if (p<0) {
	  posz[i]= visible[2];
	  weightz[i]=0;
	}
	else {
	  posz[i]+=visible[2];
	  weightz[i]=p -((int) p);
	}

	offset[i] = posz[i] * plane;

      }

      a = (double) max - (double) min;
      if(a == 0.0) a = 255.0;
      else a = 255.0 / a;

      for(y = 0, j = 0; y < sy; y++) {
	p=(((double) y / ry)-0.5);
	posy=(int) p;
	if (p<0){
	  posy= visible[1];
	  weighty=0;
	}
	else {
	  posy+=visible[1];
	  weighty= p-((int)p);
	}
	i = posy * inr->getX() + o;



	if (posy<dimy)
	  incry=dimx;
	else
	  incry=0;
	for(z = 0; z < sz; z++) {
	 
	  base=offset[z] + i;
	  if (posz[z]<dimz)
	    incrz=plane;
	  else
	    incrz=0;

	  val00=(double) ((B *) inr->getData())[base];
	  val10=(double) ((B *) inr->getData())[base + incrz];
	  val01=(double) ((B *) inr->getData())[base  + incry];
	  val11=(double) ((B *) inr->getData())[base + incry + incrz];
	  dr=val00*(1-weightz[z])*(1-weighty)+val10*weightz[z]*(1-weighty)+
	    val01*weighty*(1-weightz[z])+val11*weighty*weightz[z];
	  if(dr < (double) min) r = 0;
	  else if(dr > (double) max) r = 256;
	  else r = (int) (a * (dr - min) + 0.5);
	  block->pixelPtr[j++] = redColorMap[r];
	  block->pixelPtr[j++] = greenColorMap[r];
	  block->pixelPtr[j++] = blueColorMap[r];
	}
      }
      delete[] offset;
      delete[] weightz;      
      delete[] posz;      
    }
  }


  template <class B> void drawXYVector(B,Inrimage *inr,SliceCamera *slicecam,
				       unsigned int visible[6],
				       double min, double max,
				       Tk_PhotoImageBlock *block,
				       const unsigned char *redColorMap, 
				       const unsigned char *greenColorMap, 
				       const unsigned char *blueColorMap,
				       unsigned int sx,unsigned int sy, 
				       double rx,double ry){

    unsigned int i, j, x, y, *offset, o;
    register int r;
    double a, dr, *norm;
    
    offset = new unsigned int[sx];
    for(i = 0; i < sx; i++) offset[i] = (int) ((double) i / rx);
    o = (unsigned int) slicecam->getSlice();

    
    j = (visible[3] - visible[0] + 1) * (visible[4] - visible[1] + 1);
    norm = new double[j];
    for(y = visible[1]; y <= visible[4]; y++) {
      for(x = visible[0]; x <= visible[3]; x++, norm++) {
	*norm = 0.0;
	for(i = 0; i < inr->getVdim(); i++)
	  *norm += ((B ***) inr->getArray())[o][y][x * inr->getVdim() + i]
	    * ((B ***) inr->getArray())[o][y][x * inr->getVdim() + i];
	*norm = sqrt(*norm);
      }
    }
    
    a = max - min;
    norm -= j;
    
    if(a == 0.0) a = 255.0;
    else a = 255.0 / a;
    for(y = 0, j = 0; y < sy; y++) {
      i = ((int) ((double) y / ry)) * (visible[3] - visible[0] + 1);
      for(x = 0; x < sx; x++) {
	dr = norm[offset[x] + i];
	if(dr < min) r = 0;
	else if(dr > max) r = 256;
	else r = (int) (a * (dr - min) + 0.5);
	block->pixelPtr[j++] = redColorMap[r];
	block->pixelPtr[j++] = greenColorMap[r];
	block->pixelPtr[j++] = blueColorMap[r];
      }
    }
    
    delete[] offset;
    delete[] norm;
  }



  template <class B> void drawXZVector(B,Inrimage *inr,SliceCamera *slicecam,
				       unsigned int visible[6], 
				       double min, double max,
				       Tk_PhotoImageBlock *block,
				       const unsigned char *redColorMap, 
				       const unsigned char *greenColorMap, 
				       const unsigned char *blueColorMap,
				       unsigned int sx,unsigned int sz, 
				       double rx,double rz){

    unsigned int i, j, x, z, *offset, o;
    register int r;
    double a, dr, *norm;
    
    offset = new unsigned int[sx];
    for(i = 0; i < sx; i++) offset[i] = (int) ((double) i / rx);
    o = (unsigned int) slicecam->getSlice();
    
    j = (visible[5] - visible[2] + 1) * (visible[3] - visible[0] + 1);
    norm = new double[j];
    for(z = visible[2]; z <= visible[5]; z++) {
      for(x = visible[0]; x <= visible[3]; x++, norm++) {
	*norm = 0.0;
	for(i = 0; i < inr->getVdim(); i++)
	  *norm += ((B ***) inr->getArray())[z][o][inr->getVdim() * x + i]
	    * ((B ***) inr->getArray())[z][o][inr->getVdim() * x + i];
	*norm = sqrt(*norm);
      }
    }
    
    a = max - min;
    norm -= j;
	
    if(a == 0.0) a = 255.0;
    else a = 255.0 / a;
    for(z = 0, j = 0; z < sz; z++) {
      i = ((int)((double) z / rz)) * (visible[3] - visible[0] + 1);
      for(x = 0; x < sx; x++) {
	dr = norm[offset[x] + i];
	if(dr < min) r = 0;
	else if(dr > max) r = 256;
	else r = (int) (a * (dr - min) + 0.5);
	block->pixelPtr[j++] = redColorMap[r];
	block->pixelPtr[j++] = greenColorMap[r];
	block->pixelPtr[j++] = blueColorMap[r];
      }
    }

    delete[] norm;
    delete[] offset;
  }


  template <class B> void drawYZVector(B,Inrimage *inr,SliceCamera *slicecam,
				       unsigned int visible[6], 
				       double min, double max,
				       Tk_PhotoImageBlock *block,
				       const unsigned char *redColorMap, 
				       const unsigned char *greenColorMap, 
				       const unsigned char *blueColorMap,
				       unsigned int sy,unsigned int sz, 
				       double ry,double rz) {
    unsigned int i, j, y, z, *offset, o;
    register int r;
    double a, dr, *norm;

    o = (unsigned int) slicecam->getSlice();
    offset = new unsigned int[sz];
    for(i = 0; i < sz; i++) offset[i] = (int) ((double) i / rz);
    
    j = (visible[5] - visible[2] + 1) * (visible[4] - visible[1] + 1);
    norm = new double[j];
    for(y = visible[1]; y <= visible[4]; y++) {
      for(z = visible[2]; z <= visible[5]; z++, norm++) {
	*norm = 0.0;
	for(i = 0; i < inr->getVdim(); i++)
	  *norm += ((B ***) inr->getArray())[z][y][inr->getVdim() * o + i]
	    * ((B ***) inr->getArray())[z][y][inr->getVdim() * o + i];
	*norm = sqrt(*norm);
      }
    }

    a = max - min;
    norm -= j;
	
    if(a == 0.0) a = 255.0;
    else a = 255.0 / a;
    for(y = 0, j = 0; y < sy; y++) {
      i = ((int)((double) y / ry)) * (visible[5] - visible[2] + 1);
      for(z = 0; z < sz; z++) {
	dr = norm[offset[z] + i];
	if(dr < min) r = 0;
	else if(dr > max) r = 256;
	else r = (int) (a * (dr - min) + 0.5);
	block->pixelPtr[j++] = redColorMap[r];
	block->pixelPtr[j++] = greenColorMap[r];
	block->pixelPtr[j++] = blueColorMap[r];
      }
    }

    delete[] norm;
    delete[] offset;
  }

  /* pour dessiner les champs de vecteurs */
  

  static float vec_color[3];

  static void HSVtoRGB( float *r, float *g, float *b, float h, float s, float v )
    {
      int i;
      float f, p, q, t;
      if( s == 0 ) {
	// achromatic (grey)
	*r = *g = *b = v;
	return;
      }
      h /= 60;                        // sector 0 to 5
      i = (int)floor( h );
      f = h - i;                      // factorial part of h
      p = v * ( 1 - s );
      q = v * ( 1 - s * f );
      t = v * ( 1 - s * ( 1 - f ) );
      switch( i ) {
      case 0:
	*r = v;	*g = t;	*b = p;
	break;
      case 1:
	*r = q;	*g = v;	*b = p;
	break;
      case 2:
	*r = p;	*g = v;	*b = t;
	break;
      case 3:
	*r = p;	*g = q;	*b = v;
	break;
      case 4:
	*r = t;	*g = p;	*b = v;
	break;
      default:                // case 5:
	*r = v;	*g = p;	*b = q;
	break;
      }
    }
  
  
  template <class B> void drawYZVector_Vector(B,Inrimage *inr, unsigned int under_x, 
					      unsigned int under_y, unsigned int under_z, 
					      SliceCamera *slicecam,
					      unsigned int visible[6], 
					      double min, double max,
					      Tk_PhotoImageBlock *block,
					      const unsigned char *redColorMap, 
					      const unsigned char *greenColorMap, 
					      const unsigned char *blueColorMap,
					      unsigned int sy,unsigned int sz, 
					      double ry,double rz,
					      float vec_width,
					      float vec_scale,
					      TclInrimageView::NORM_MODE norm_Mode,
					      float alpha,
					      float beta,
					      bool withArrow,
					      double ***norme3D,
					      double ***norme2D,
					      double norme3D_max,
					      double norme3D_min)
    {
      float h = 0.0;
      float v = 1.0;
 

      B *buffer = (B *) inr->getData();   

      unsigned int sl = (unsigned int) (slicecam->getSlice());
     	unsigned int vdim=inr->getVdim();
		/// initialize the color to red for the 2D case
		vec_color[0]=1.0;
		vec_color[1]=0;
		vec_color[2]=0;
 
      // on efface toutes les anciens vecteurs
      slicecam->getInterface()->getModule()->tickle("[%s%u -widget] delete arrow", slicecam->getInterface()->getTCLObjectName(), slicecam->getInterface()->getRef());
      
      // calcul du pas d'echantillonnage => A VERIFIER !!!
      unsigned int offset_x = (inr->getZ()-1) - (inr->getZ()-1)%under_z;
      unsigned int offset_y = inr->getY() -1 - (inr->getY()-1)%under_y;

	  unsigned int width=slicecam->getWidth();
		unsigned int height=slicecam->getHeight();

      // extremite du vecteur a dessiner
      double orig_x=0, orig_y=0;
      double end_x=0, end_y=0;

  		double vx=inr->getVZ();
		double vy=inr->getVY();

      double x=0.0, y=0.0; 
      double norm2D=0.0, norm3D=0.0;

		double cx,cy,ccx,ccy,sc;

		cx=slicecam->getCenter().getX();
		cy=slicecam->getCenter().getY();
		ccx=slicecam->getCameraCenter().getX();
		ccy=slicecam->getCameraCenter().getY();
		sc=slicecam->getScale();

		unsigned int firstComponentOffset=inr->getX()*inr->getY()*inr->getZ();

      for (unsigned int j = 0; j<=offset_y; j+=under_y) 
	for (unsigned int i = 0; i<=offset_x; i+=under_z) {
		if (inr->getVectorialMode() == VM_INTERLACED) { 

			x = buffer[i*inr->getX()*inr->getY()*vdim + j*inr->getX()*vdim + sl*vdim+2];
			y = buffer[i*inr->getX()*inr->getY()*vdim + j*inr->getX()*vdim + sl*vdim+1];
		} else {
			x = buffer[firstComponentOffset*2+i*inr->getX()*inr->getY()+ j*inr->getX() + sl];
			y = buffer[firstComponentOffset+i*inr->getX()*inr->getY()+ j*inr->getX() + sl];

		}
	  norm2D = norme2D[i][j][sl];
	  norm3D = norme3D[i][j][sl];

	  /* Gestion de la couleur */ 
	  /* Formules : H = 240 / ( norme3D_max - norme3D_min) * (norme3D_max - norme3D) */
	  /*            S = 1                               */
	  /*            V = norme2D/norme3D                 */

	  if (norm2D > 0.0) {
		  if (vdim==3) {

			  h = (float)(240.0 / ( norme3D_max - norme3D_min) * (norme3D_max - norm3D));
			  // s = 1
			  v = (float)(norm2D/norm3D);

			  HSVtoRGB(&vec_color[0], &vec_color[1], &vec_color[2], h, 1, v);
		  }
		  if ( (i >= visible[2]) && (i <= visible[5]) && (j >= visible[1]) && (j <= visible[4]) ) {
			  orig_x =  vx*(i-visible[2]);
			  orig_y =  vy*(j-visible[1]);

			  orig_x= sc*(orig_x - cx)  + ccx;
			  orig_y= sc*(orig_y - cy)  + ccy;
			  if ((orig_x>=0) && (orig_y>=0) && (orig_x<width) && (orig_y<height)) {

				  switch (norm_Mode) {
		  default:
			  break;
		  case TclInrimageView::LINEAR:
			  break;
		  case TclInrimageView::EXPONENTIAL:
			  x *= beta * (1-exp(-alpha*norm2D));
			  y *= beta * (1-exp(-alpha*norm2D));
			  break;
		  case TclInrimageView::NORMALIZED:
			  x /= norm2D * inr->getVY();
			  y /= norm2D * inr->getVZ();
			  break;
				  }
				  x *= vec_scale;
				  y *= vec_scale;

				  end_x = orig_x + sc * x;
				  end_y = orig_y + sc * y;
				  if ((end_x>=0) && (end_y>=0) && (end_x<width) && (end_y<height)) {

					  if (withArrow)
						  slicecam->getInterface()->getModule()->tickle("[%s%u -widget] create line %u %u %u %u -arrow last -fill [modules::rgbToColor \"%f %f %f\"] -width %f -tags arrow", slicecam->getInterface()->getTCLObjectName(), slicecam->getInterface()->getRef(), (unsigned int)orig_x, (unsigned int)orig_y, (unsigned int)end_x, (unsigned int)end_y, vec_color[0], vec_color[1], vec_color[2], vec_width);
					  else
						  slicecam->getInterface()->getModule()->tickle("[%s%u -widget] create line %u %u %u %u -arrow none -fill [modules::rgbToColor \"%f %f %f\"] -width %f -tags arrow", slicecam->getInterface()->getTCLObjectName(), slicecam->getInterface()->getRef(), (unsigned int)orig_x, (unsigned int)orig_y, (unsigned int)end_x, (unsigned int)end_y, vec_color[0], vec_color[1], vec_color[2], vec_width);
				  }
			  }
		  }
	  }
	}
    }
	template <class B> void drawXYVector_Vector(B,Inrimage *inr, unsigned int under_x, 
		unsigned int under_y, unsigned int under_z, 
		SliceCamera *slicecam,
		unsigned int visible[6], 
		double min, double max,
		Tk_PhotoImageBlock *block,
		const unsigned char *redColorMap, 
		const unsigned char *greenColorMap, 
		const unsigned char *blueColorMap,
		unsigned int sx,unsigned int sy, 
		double rx,double ry,
		float vec_width,
		float vec_scale,
		TclInrimageView::NORM_MODE norm_Mode,
		float alpha,
		float beta,
		bool withArrow,
		double ***norme3D,
		double ***norme2D,
		double norme3D_max,
		double norme3D_min)
	{
		float h = 0.0;
		float v = 1.0;

		B *buffer = (B *) inr->getData();   

		unsigned int sl = (unsigned int) (slicecam->getSlice());
		unsigned int vdim=inr->getVdim();
		/// initialize the color to red for the 2D case
		vec_color[0]=1.0;
		vec_color[1]=0;
		vec_color[2]=0;


		// on efface toutes les anciens vecteurs
		slicecam->getInterface()->getModule()->tickle("[%s%u -widget] delete arrow", slicecam->getInterface()->getTCLObjectName(), slicecam->getInterface()->getRef());

		// calcul du pas d'echantillonnage => A VERIFIER !!!
		unsigned int offset_x = inr->getX()-1 - (inr->getX()-1)%under_x;
		unsigned int offset_y = inr->getY()-1 - (inr->getY()-1)%under_y;
		
		unsigned int width=slicecam->getWidth();
		unsigned int height=slicecam->getHeight();


		// extremite du vecteur a dessiner
		double orig_x=0, orig_y=0;
		double end_x=0, end_y=0;

		double vx=inr->getVX();
		double vy=inr->getVY();

		double cx,cy,ccx,ccy,sc;

		cx=slicecam->getCenter().getX();
		cy=slicecam->getCenter().getY();
		ccx=slicecam->getCameraCenter().getX();
		ccy=slicecam->getCameraCenter().getY();
		sc=slicecam->getScale();

		double x=0.0, y=0.0 ;
		double norm2D=0.0, norm3D=0.0;

		unsigned int firstComponentOffset=inr->getX()*inr->getY()*inr->getZ();
		unsigned int offset;


		for (unsigned int j = 0; j<=offset_y; j+=under_y) {
				if (inr->getVectorialMode() == VM_INTERLACED) 
			offset=sl*inr->getX()*inr->getY()*vdim + j*inr->getX()*vdim;
				else
				offset=sl*inr->getX()*inr->getY() + j*inr->getX() ;
			for (unsigned int i = 0; i<=offset_x; i+=under_x) {
				if (inr->getVectorialMode() == VM_INTERLACED) { 
					x = buffer[offset + i*vdim];
					y = buffer[offset+ i*vdim+1];
				} else {
					x = buffer[offset + i];
					y = buffer[firstComponentOffset+offset+ i];
				}
				norm2D = norme2D[sl][j][i];
				norm3D = norme3D[sl][j][i];

				/* Gestion de la couleur */ 
				/* Formules : H = 240 / ( norme3D_max - norme3D_min) * (norme3D_max - norme3D) */
				/*            S = 1                               */
				/*            V = norme2D/norme3D                 */

				if (norm2D > 0.0) {
					if (vdim==3) {

						h = (float)(240.0 / ( norme3D_max - norme3D_min) * (norme3D_max - norm3D));
						v = (float)(norm2D/norm3D);

						HSVtoRGB(&vec_color[0], &vec_color[1], &vec_color[2], h, 1, v);
					}
					if ( (i >= visible[0]) && (i <= visible[3]) && (j >= visible[1]) && (j <= visible[4]) ) {
						orig_x =  vx*(i-visible[0]);
						orig_y = vy*(j-visible[1]);

						orig_x= sc*(orig_x - cx)  + ccx;
						orig_y= sc*(orig_y - cy)  + ccy;
						if ((orig_x>=0) && (orig_y>=0) && (orig_x<width) && (orig_y<height)) {

							switch (norm_Mode) {
		  default:
			  break;
		  case TclInrimageView::LINEAR:
			  break;
		  case TclInrimageView::EXPONENTIAL:
			  x *= beta * (1-exp(-alpha*norm2D));
			  y *= beta * (1-exp(-alpha*norm2D));
			  break;
		  case TclInrimageView::NORMALIZED:
			  x /= norm2D * inr->getVX();
			  y /= norm2D * inr->getVY();
			  break;
							}
							x *= vec_scale;
							y *= vec_scale;

							end_x = orig_x + sc* x;
							end_y = orig_y + sc* y;

							if ((end_x>=0) && (end_y>=0) && (end_x<width) && (end_y<height)) {
								if (withArrow)
									slicecam->getInterface()->getModule()->tickle("[%s%u -widget] create line %u %u %u %u -arrow last -fill [modules::rgbToColor \"%f %f %f\"] -width %f -tags arrow", slicecam->getInterface()->getTCLObjectName(), slicecam->getInterface()->getRef(), (unsigned int)orig_x, (unsigned int)orig_y, (unsigned int)end_x, (unsigned int)end_y, vec_color[0], vec_color[1], vec_color[2], vec_width);
								else
									slicecam->getInterface()->getModule()->tickle("[%s%u -widget] create line %u %u %u %u -arrow none -fill [modules::rgbToColor \"%f %f %f\"] -width %f -tags arrow", slicecam->getInterface()->getTCLObjectName(), slicecam->getInterface()->getRef(), (unsigned int)orig_x, (unsigned int)orig_y, (unsigned int)end_x, (unsigned int)end_y, vec_color[0], vec_color[1], vec_color[2], vec_width);
							}
						}
					}
				}
			}
			}
	}


  template <class B> void drawXZVector_Vector(B,Inrimage *inr, unsigned int under_x, 
					      unsigned int under_y, unsigned int under_z, 
					      SliceCamera *slicecam,
					      unsigned int visible[6], 
					      double min, double max,
					      Tk_PhotoImageBlock *block,
					      const unsigned char *redColorMap, 
					      const unsigned char *greenColorMap, 
					      const unsigned char *blueColorMap,
					      unsigned int sx,unsigned int sz, 
					      double rx,double rz,
					      float vec_width,
					      float vec_scale,
					      TclInrimageView::NORM_MODE norm_Mode,
					      float alpha,
					      float beta,
					      bool withArrow,
					      double ***norme3D,
					      double ***norme2D,
					      double norme3D_max,
					      double norme3D_min)
    {
      float h = 0.0;
      float v = 1.0;
     
      B *buffer = (B *) inr->getData();   
      
      unsigned int sl = (unsigned int) (slicecam->getSlice());
       unsigned int vdim=inr->getVdim();
		/// initialize the color to red for the 2D case
		vec_color[0]=1.0;
		vec_color[1]=0;
		vec_color[2]=0;

      
      // on efface toutes les anciens vecteurs
      slicecam->getInterface()->getModule()->tickle("[%s%u -widget] delete arrow", slicecam->getInterface()->getTCLObjectName(), slicecam->getInterface()->getRef());
      
      // calcul du pas d'echantillonnage => A VERIFIER !!!
      unsigned int offset_x = (inr->getX()-1) - (inr->getX()-1)%under_x;
      unsigned int offset_y = (inr->getZ()-1) - (inr->getZ()-1)%under_z;

	  unsigned int height=slicecam->getHeight();
	  unsigned int width=slicecam->getWidth();

      // extremite du vecteur a dessiner
      double orig_x=0, orig_y=0;
      double end_x=0, end_y=0;

	  	double vx=inr->getVX();
		double vy=inr->getVZ();

      double x=0.0, y=0.0; 
      double norm2D=0.0, norm3D=0.0;

		double cx,cy,ccx,ccy,sc;

		cx=slicecam->getCenter().getX();
		cy=slicecam->getCenter().getY();
		ccx=slicecam->getCameraCenter().getX();
		ccy=slicecam->getCameraCenter().getY();
		sc=slicecam->getScale();

		unsigned int firstComponentOffset=inr->getX()*inr->getY()*inr->getZ();

      for (unsigned int j = 0; j<=offset_y; j+=under_z) 
	for (unsigned int i = 0; i<=offset_x; i+=under_x) {

	if (inr->getVectorialMode() == VM_INTERLACED) { 
	  x = buffer[j*inr->getX()*inr->getY()*vdim + sl*inr->getX()*vdim + i*vdim];
	  y = buffer[j*inr->getX()*inr->getY()*vdim + sl*inr->getX()*vdim + i*vdim+2];
	} else {
	  x = buffer[j*inr->getX()*inr->getY() + sl*inr->getX() + i];
	  y = buffer[2*firstComponentOffset+j*inr->getX()*inr->getY() + sl*inr->getX() + i];
	}
	  norm2D = norme2D[j][sl][i];
	  norm3D = norme3D[j][sl][i];

	  /* Gestion de la couleur */ 
	  /* Formules : H = 240 / ( norme3D_max - norme3D_min) * (norme3D_max - norme3D) */
	  /*            S = 1                               */
	  /*            V = norme2D/norme3D                 */

	  if (norm2D > 0.0) {
		  if (vdim==3) {
			  h = (float)(240.0 / ( norme3D_max - norme3D_min) * (norme3D_max - norm3D));
			  // s = 1
			  v = (float)(norm2D/norm3D);

			  HSVtoRGB(&vec_color[0], &vec_color[1], &vec_color[2], h, 1, v);
		  }

	    if ( (i >= visible[0]) && (i <= visible[3]) && (j >= visible[2]) && (j <= visible[5]) ) {

		  orig_x = vx*(i-visible[0]);
	      orig_y = vy*(j-visible[2]);

		  orig_x= sc*(orig_x - cx)  + ccx;
  		  orig_y= sc*(orig_y - cy)  + ccy;
		
		  if ((orig_x>=0) && (orig_y>=0) && (orig_x<width) && (orig_y<height)) {


			  switch (norm_Mode) {
		  default:
			  break;
		  case TclInrimageView::LINEAR:
			  break;
		  case TclInrimageView::EXPONENTIAL:
			  x *= beta * (1-exp(-alpha*norm2D));
			  y *= beta * (1-exp(-alpha*norm2D));
			  break;
		  case TclInrimageView::NORMALIZED:
			  x /= norm2D * inr->getVX();
			  y /= norm2D * inr->getVZ();
			  break;
			  }

			  x *= vec_scale;
			  y *= vec_scale;
			  end_x = orig_x +  sc*x;
			  end_y = orig_y +  sc*y;
			  if ((end_x>=0) && (end_y>=0) && (end_x<width) && (end_y<height)) {

				  if (withArrow)
					  slicecam->getInterface()->getModule()->tickle("[%s%u -widget] create line %u %u %u %u -arrow last -fill [modules::rgbToColor \"%f %f %f\"] -width %f -tags arrow", slicecam->getInterface()->getTCLObjectName(), slicecam->getInterface()->getRef(), (unsigned int)orig_x, (unsigned int)orig_y, (unsigned int)end_x, (unsigned int)end_y, vec_color[0], vec_color[1], vec_color[2], vec_width );
				  else
					  slicecam->getInterface()->getModule()->tickle("[%s%u -widget] create line %u %u %u %u -arrow none -fill [modules::rgbToColor \"%f %f %f\"] -width %f -tags arrow", slicecam->getInterface()->getTCLObjectName(), slicecam->getInterface()->getRef(), (unsigned int)orig_x, (unsigned int)orig_y, (unsigned int)end_x, (unsigned int)end_y, vec_color[0], vec_color[1], vec_color[2], vec_width);
			  }
		  }
		}
	  }
	}
      
    }
  





}
#endif
