#ifndef BACKGROUNDTRANSFORMEDINRIMAGEREDRAW_H
#define BACKGROUNDTRANSFORMEDINRIMAGEREDRAW_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/backgroundTransformedInrimageRedraw.h"
namespace yav {


  /// template function for drawing an xy slice
template <class B> void drawTransformedInrimageXY(B,Inrimage *inr,
						  SliceCamera *slicecam,
						  double min, double max,
						  Tk_PhotoImageBlock *block,
						  const unsigned char *redColorMap, 
						  const unsigned char *greenColorMap, 
						  const unsigned char *blueColorMap,
						  unsigned int sx,unsigned int sy, 
						  double rx,double ry,
						  BasicInrimageBackground::IMAGE_INTERPOLATION inter){

    unsigned int i, j, x, y, *offset;
    register int r1;
    double a1, dr;

    offset = new unsigned int[sx];
    if (inter==BasicInrimageBackground::NEAREST) {

      for(i = 0; i < sx; i++) offset[i] = (int) ((double) i / rx) ;
      
      a1 = (double) max - (double) min;
      if(a1 == 0.0) a1 = 255.0;
      else a1 = 255.0 / a1;
      
      
      for(y = 0, j = 0; y < sy; y++) {
	i =  ((int) ((double) y / ry)) * inr->getX();
	for(x = 0; x < sx; x++) {
	  dr = (double) ((B *) inr->getData())[offset[x] + i];
	  if(dr < (double) min) r1 = 0;
	  else if(dr > (double) max) r1 = 256;
	  else r1 = (int) (a1 * (dr - min) + 0.5);
	  
	  
	  block->pixelPtr[j++] = (unsigned char)(redColorMap[r1]);
	  block->pixelPtr[j++] = (unsigned char)(greenColorMap[r1]);
	  block->pixelPtr[j++] = (unsigned char)(blueColorMap[r1]);
	}
      }
    } 
    else if (inter==BasicInrimageBackground::BILINEAR) {
      double *weightx= new double[sx];
      unsigned int posy;
      unsigned int dimx=inr->getX();
      unsigned int dimy=inr->getY()-1;
      int r;
      unsigned int incrx,incry;
      double weighty,val00,val01,val10,val11,p,a;
      for(i = 0; i < sx; i++) {
	p=(((double) i / rx)-0.5);
	offset[i] = (int) p;
 	if (p<0) {
	  offset[i]=0;
	  weightx[i]=0;
	}
	else {
	  weightx[i]=p-((int)p);
	}
      }
      a = (double) max - (double) min;
      if(a == 0.0) a = 255.0;
      else a = 255.0 / a;
      for(y = 0, j = 0; y < sy; y++) {
	p= (((double) y / ry)-0.5);
	posy=(int) p;
	if (p<0) {
	  posy= 0;
	  weighty=0;
	}
	else {
	  weighty=p-((int)p);
	}
	i = posy * dimx;

	if (posy<dimy)
	  incry=dimx;
	else
	  incry=0;

	for(x = 0; x < sx; x++) {
	  if (offset[x]<(dimx-1))
	    incrx=1;
	  else
	    incrx=0;
	  val00=(double) ((B *) inr->getData())[offset[x] + i];
	  val10=(double) ((B *) inr->getData())[offset[x] + incrx + i];
	  val01=(double) ((B *) inr->getData())[offset[x] + i + incry];
	  val11=(double) ((B *) inr->getData())[offset[x] + incrx + i + incry];
	  dr=val00*(1-weightx[x])*(1-weighty)+val10*weightx[x]*(1-weighty)+
	    val01*weighty*(1-weightx[x])+val11*weighty*weightx[x];
	  if(dr < (double) min) r = 0;
	  else if(dr > (double) max) r = 256;
	  else r = (int) (a * (dr - min) + 0.5);
	  block->pixelPtr[j++] = redColorMap[r];
	  block->pixelPtr[j++] = greenColorMap[r];
	  block->pixelPtr[j++] = blueColorMap[r];
	}
      }
      delete[] weightx;
    }
    delete[] offset;
}

template <class B> void drawTransformedInrimageXZ(B,Inrimage *inr,SliceCamera *slicecam,
				 double min, double max,
				 Tk_PhotoImageBlock *block,
				 const unsigned char *redColorMap, 
				 const unsigned char *greenColorMap, 
				 const unsigned char *blueColorMap,
				 unsigned int sx,unsigned int sz, 
				 double rx,double rz,BasicInrimageBackground::IMAGE_INTERPOLATION inter){

 
  unsigned int plane, i, j, x, z, *offset;
  register int r;
  double a, dr;

  plane = inr->getX() * inr->getY();
  if (inter==BasicInrimageBackground::NEAREST) {    
    offset = new unsigned int[sx];
    for(i = 0; i < sx; i++) offset[i] = (int) ((double) i / rx);
      a = max - min;
      if(a == 0.0) a = 255.0;
      else a = 255.0 / a;
      for(z = 0, j = 0; z < sz; z++) {
	i = ((int)((double) z / rz)) * plane ;
	for(x = 0; x < sx; x++) {
	  dr = (double) ((B *) inr->getData())[offset[x] + i];
	  if(dr < min) r = 0;
	  else if(dr > max) r = 256;
	  else r = (int) (a * (dr - min) + 0.5);
	  block->pixelPtr[j++] = redColorMap[r];
	  block->pixelPtr[j++] = greenColorMap[r];
	  block->pixelPtr[j++] = blueColorMap[r];
	}
      }
      
      delete[] offset;
    }
    else if (inter==BasicInrimageBackground::BILINEAR) {
      offset = new unsigned int[sx];
      unsigned int posz;
 
      unsigned int dimx=inr->getX()-1;
      unsigned int dimz=inr->getZ()-1;

      double *weightx= new double[sx];
      double weighty,val00,val01,val10,val11,p;
      unsigned int incrx,incrz;
      for(i = 0; i < sx; i++) {	
	p=(((double) i / rx)-0.5);
	offset[i] = (int) p;
 	if (p<0)
	  {
	    offset[i]=0;
	    weightx[i]=0.0;
	  }
	else {
	  weightx[i]=p-((int)p);
	}

	
      }
      a = (double) max - (double) min;
      if(a == 0.0) a = 255.0;
      else a = 255.0 / a;
      for(z = 0, j = 0; z < sz; z++) {
	p= (((double) z / rz)-0.5);
	posz=(int)p;
	if (p<0)
	  {
	    posz= 0;
	    weighty=0;
	  }
	else {
	  weighty=p-((int)p);
	}
	i = posz * plane ;

	if (posz<dimz)
	  incrz=plane;
	else
	  incrz=0;
	for(x = 0; x < sx; x++) {
	  if (offset[x]<dimx)
	    incrx=1;
	  else
	    incrx=0;
	  val00=(double) ((B *) inr->getData())[offset[x] + i];
	  val10=(double) ((B *) inr->getData())[offset[x] + incrx + i];
	  val01=(double) ((B *) inr->getData())[offset[x] + i + incrz];
	  val11=(double) ((B *) inr->getData())[offset[x] + incrx + i + incrz];
	  dr=val00*(1-weightx[x])*(1-weighty)+val10*weightx[x]*(1-weighty)+
	    val01*weighty*(1-weightx[x])+val11*weighty*weightx[x];
	  if(dr < (double) min) r = 0;
	  else if(dr > (double) max) r = 256;
	  else r = (int) (a * (dr - min) + 0.5);
	  block->pixelPtr[j++] = redColorMap[r];
	  block->pixelPtr[j++] = greenColorMap[r];
	  block->pixelPtr[j++] = blueColorMap[r];
	}
      }
      delete[] offset;
      delete[] weightx;
      
    }
}

  template <class B> void drawTransformedInrimageYZ(B,Inrimage *inr,SliceCamera *slicecam,
				 double min, double max,
				 Tk_PhotoImageBlock *block,
				 const unsigned char *redColorMap, 
				 const unsigned char *greenColorMap, 
				 const unsigned char *blueColorMap,
				 unsigned int sy,unsigned int sz, 
				 double ry,double rz,BasicInrimageBackground::IMAGE_INTERPOLATION inter){
    
    unsigned int plane, i, j, y, z, *offset;
    register int r;
    double a, dr;
    
    plane = inr->getX() * inr->getY();
    if (inter==BasicInrimageBackground::NEAREST) {        
      offset = new unsigned int[sz];
      for(i = 0; i < sz; i++) offset[i] = ((int) ((double) i / rz) ) * plane;
      
      a = max - min;
      if(a == 0.0) a = 255.0;
      else a = 255.0 / a;
      for(y = 0, j = 0; y < sy; y++) {
	i = ((int)((double) y / ry) ) * inr->getX();
	for(z = 0; z < sz; z++) {
	  dr = (double) ((B *) inr->getData())[offset[z] + i];
	  if(dr < min) r = 0;
	  else if(dr > max) r = 256;
	  else r = (int) (a * (dr - min) + 0.5);
	  block->pixelPtr[j++] = redColorMap[r];
	  block->pixelPtr[j++] = greenColorMap[r];
	  block->pixelPtr[j++] = blueColorMap[r];
	}
      }
      
      delete[] offset;
    }
    else if (inter==BasicInrimageBackground::BILINEAR) {
      offset = new unsigned int[sz];
      unsigned int *posz = new unsigned int[sz];

      unsigned int dimx=inr->getX();
      unsigned int dimy=inr->getY()-1;
      unsigned int dimz=inr->getZ()-1;

      unsigned int incry,incrz,posy,base;
      double *weightz= new double[sz];
      double weighty,val00,val01,val10,val11,p;

      for(i = 0; i < sz; i++) {
	p=((double) i / rz)-0.5;
	posz[i]=(int) p;
	if (p<0) {
	  posz[i]= 0;
	  weightz[i]=0;
	}
	else {
	  weightz[i]=p -((int) p);
	}

	offset[i] = posz[i] * plane;

      }

      a = (double) max - (double) min;
      if(a == 0.0) a = 255.0;
      else a = 255.0 / a;

      for(y = 0, j = 0; y < sy; y++) {
	p=(((double) y / ry)-0.5);
	posy=(int) p;
	if (p<0){
	  posy= 0;
	  weighty=0;
	}
	else {
	  weighty= p-((int)p);
	}
	i = posy * inr->getX() ;



	if (posy<dimy)
	  incry=dimx;
	else
	  incry=0;
	for(z = 0; z < sz; z++) {
	 
	  base=offset[z] + i;
	  if (posz[z]<dimz)
	    incrz=plane;
	  else
	    incrz=0;

	  val00=(double) ((B *) inr->getData())[base];
	  val10=(double) ((B *) inr->getData())[base + incrz];
	  val01=(double) ((B *) inr->getData())[base  + incry];
	  val11=(double) ((B *) inr->getData())[base + incry + incrz];
	  dr=val00*(1-weightz[z])*(1-weighty)+val10*weightz[z]*(1-weighty)+
	    val01*weighty*(1-weightz[z])+val11*weighty*weightz[z];
	  if(dr < (double) min) r = 0;
	  else if(dr > (double) max) r = 256;
	  else r = (int) (a * (dr - min) + 0.5);
	  block->pixelPtr[j++] = redColorMap[r];
	  block->pixelPtr[j++] = greenColorMap[r];
	  block->pixelPtr[j++] = blueColorMap[r];
	}
      }
      delete[] offset;
      delete[] weightz;      
      delete[] posz;      
    }
  }

  template <class B> void drawTransformedInrimageXYVector(B,Inrimage *inr,SliceCamera *slicecam,
							  double min, double max,
							  Tk_PhotoImageBlock *block,
							  const unsigned char *redColorMap, 
							  const unsigned char *greenColorMap, 
							  const unsigned char *blueColorMap,
							  unsigned int sx,unsigned int sy, 
							  double rx,double ry){
    
    unsigned int i, j, x, y, *offset;
    register int r;
    double a, dr, *norm;
    
    offset = new unsigned int[sx];
    for(i = 0; i < sx; i++) offset[i] = (int) ((double) i / rx);

    
    j = inr->getX()*inr->getY();
    norm = new double[j];
    for(y = 0; y < inr->getY(); y++) {
      for(x = 0; x < inr->getX() ; x++, norm++) {
	*norm = 0.0;
	for(i = 0; i < inr->getVdim(); i++)
	  *norm += ((B ***) inr->getArray())[0][y][x * inr->getVdim() + i]
	    * ((B ***) inr->getArray())[0][y][x * inr->getVdim() + i];
	*norm = sqrt(*norm);
      }
    }
    
    a = max - min;
    norm -= j;
    
    if(a == 0.0) a = 255.0;
    else a = 255.0 / a;
    for(y = 0, j = 0; y < sy; y++) {
      i = ((int) ((double) y / ry)) * inr->getX();
      for(x = 0; x < sx; x++) {
	dr = norm[offset[x] + i];
	if(dr < min) r = 0;
	else if(dr > max) r = 256;
	else r = (int) (a * (dr - min) + 0.5);
	block->pixelPtr[j++] = redColorMap[r];
	block->pixelPtr[j++] = greenColorMap[r];
	block->pixelPtr[j++] = blueColorMap[r];
      }
    }
    
    delete[] offset;
    delete[] norm;
  }
  template <class B> void drawTransformedInrimageXZVector(B,Inrimage *inr,SliceCamera *slicecam,
				       double min, double max,
				       Tk_PhotoImageBlock *block,
				       const unsigned char *redColorMap, 
				       const unsigned char *greenColorMap, 
				       const unsigned char *blueColorMap,
				       unsigned int sx,unsigned int sz, 
				       double rx,double rz){
    unsigned int i, j, x, z, *offset;
    register int r;
    double a, dr, *norm;
    
    offset = new unsigned int[sx];
    for(i = 0; i < sx; i++) offset[i] = (int) ((double) i / rx);
    
    j = inr->getZ()*inr->getX();
    norm = new double[j];
    for(z = 0; z < inr->getZ(); z++) {
      for(x = 0; x < inr->getX(); x++, norm++) {
	*norm = 0.0;
	for(i = 0; i < inr->getVdim(); i++)
	  *norm += ((B ***) inr->getArray())[z][0][inr->getVdim() * x + i]
	    * ((B ***) inr->getArray())[z][0][inr->getVdim() * x + i];
	*norm = sqrt(*norm);
      }
    }
    
    a = max - min;
    norm -= j;
	
    if(a == 0.0) a = 255.0;
    else a = 255.0 / a;
    for(z = 0, j = 0; z < sz; z++) {
      i = ((int)((double) z / rz)) * inr->getX();
      for(x = 0; x < sx; x++) {
	dr = norm[offset[x] + i];
	if(dr < min) r = 0;
	else if(dr > max) r = 256;
	else r = (int) (a * (dr - min) + 0.5);
	block->pixelPtr[j++] = redColorMap[r];
	block->pixelPtr[j++] = greenColorMap[r];
	block->pixelPtr[j++] = blueColorMap[r];
      }
    }

    delete[] norm;
    delete[] offset;
}


  template <class B> void drawTransformedInrimageYZVector(B,Inrimage *inr,SliceCamera *slicecam,
							  double min, double max,
							  Tk_PhotoImageBlock *block,
							  const unsigned char *redColorMap, 
							  const unsigned char *greenColorMap, 
							  const unsigned char *blueColorMap,
							  unsigned int sy,unsigned int sz, 
							  double ry,double rz) {
    unsigned int i, j, y, z, *offset;
    register int r;
    double a, dr, *norm;


    offset = new unsigned int[sz];
    for(i = 0; i < sz; i++) offset[i] = (int) ((double) i / rz);
    
    j = inr->getY()*inr->getZ();
    norm = new double[j];
    for(y = 0; y < inr->getY(); y++) {
      for(z = 0; z < inr->getX(); z++, norm++) {
	*norm = 0.0;
	for(i = 0; i < inr->getVdim(); i++)
	  *norm += ((B ***) inr->getArray())[z][y][ i]
	    * ((B ***) inr->getArray())[z][y][i];
	*norm = sqrt(*norm);
      }
    }

    a = max - min;
    norm -= j;
	
    if(a == 0.0) a = 255.0;
    else a = 255.0 / a;
    for(y = 0, j = 0; y < sy; y++) {
      i = ((int)((double) y / ry)) * inr->getZ();
      for(z = 0; z < sz; z++) {
	dr = norm[offset[z] + i];
	if(dr < min) r = 0;
	else if(dr > max) r = 256;
	else r = (int) (a * (dr - min) + 0.5);
	block->pixelPtr[j++] = redColorMap[r];
	block->pixelPtr[j++] = greenColorMap[r];
	block->pixelPtr[j++] = blueColorMap[r];
      }
    }

    delete[] norm;
    delete[] offset;
  }


}
#endif
