#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#include <inrimage/Inrimage.h>
#include <inrimage/blockProcessing.h>
#include <algorithm>
#include <assert.h>

#pragma implementation "inrimage/blockProcessing.h"

namespace yav {

	template <class T>
	  void storeBlockMeanScalar(T ,Inrimage *inputImage,		
		Inrimage *outputImage,unsigned int sxy, unsigned int sz) 
	  {
		  T *inputData= (T *)inputImage->getData();
		  T *outputData= (T *)outputImage->getData();
		  T *pz,*py,*px,*out;
		  unsigned int minx, miny,minz,maxx,maxy,maxz;
		  unsigned int i,j,k,u,v,w;
		  unsigned int nx=inputImage->getX();
		  unsigned int ny=inputImage->getY();
		  unsigned int nz=inputImage->getZ();
		  double val;


		  unsigned int offsetY=inputImage->getX();
		  unsigned int offsetZ=inputImage->getX()*inputImage->getY();

		  out=outputData;
		  for (k=0;k<nz;++k) {
			  if (k<sz)
				  minz=0;
			  else minz=k-sz;
			  if (k<(nz-sz)) {
				  maxz=k+sz;
			  } else 
				  maxz=nz-1;
			  for (j=0;j<ny;++j) {
				  if (j<sxy)
					  miny=0;
				  else miny=j-sxy;
				  if (j<(ny-sxy)) {
					  maxy=j+sxy;
				  } else 
					  maxy=ny-1;

				  for (i=0;i<nx;++i,++out) {
					  if (i<sxy)
						  minx=0;
					  else minx=i-sxy;
					  if (i<(nx-sxy)) {
						  maxx=i+sxy;
					  } else 
						  maxx=nx-1;

					  pz=inputData+offsetZ*minz+offsetY*miny+minx;
					  val=(double)0;

					  for (w=minz;w<=maxz;++w,pz+=offsetZ) {
						  py=pz;
						  for (v=miny;v<=maxy;++v,py+=offsetY) {
							  px=py;
							  for (u=minx;u<=maxx;++u,++px) {
								  val+=(*px);
							  }
						  }
					  }
					  val/=(maxx-minx+1)*(maxy-miny+1)*(maxz-minz+1);
					  *out=(T) val;

				  }
			  }
		  }



	  }
#define BLOCKMEANSCALAR(TYPE,IN,OUT,SXY,SZ) storeBlockMeanScalar((TYPE)0,IN,OUT,SXY,SZ)

	  void storeBlockMean(Inrimage *inputImage,Inrimage *outputImage,
		  unsigned int sxy,unsigned int sz){
		  switch(inputImage->getType()) {
  case Inrimage::WT_UNSIGNED_CHAR:
	  BLOCKMEANSCALAR(unsigned char,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_SIGNED_CHAR:
	  BLOCKMEANSCALAR(char,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_UNSIGNED_SHORT:
	  BLOCKMEANSCALAR(unsigned short,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_SIGNED_SHORT:
	  BLOCKMEANSCALAR(short,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_UNSIGNED_INT:
	  BLOCKMEANSCALAR(unsigned int,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_SIGNED_INT:
	  BLOCKMEANSCALAR(int,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_UNSIGNED_LONG:
	  BLOCKMEANSCALAR(unsigned LONGINT,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_SIGNED_LONG:
	  BLOCKMEANSCALAR(LONGINT,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_FLOAT:
	  BLOCKMEANSCALAR(float,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_DOUBLE:
	  BLOCKMEANSCALAR(double,inputImage,outputImage,sxy,sz);
	  break;
 /* case Inrimage::WT_RGB:
	  BLOCKMEANSCALAR(unsigned char,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_RGBA:
	  BLOCKMEANSCALAR(unsigned char,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_FLOAT_VECTOR:
	  BLOCKMEANSCALAR(float,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_DOUBLE_VECTOR:
	  BLOCKMEANSCALAR(double,inputImage,outputImage,sxy,sz);
	  break;*/
  default:
	  break;
		  }
	  }

	Inrimage *computeBlockMean(Inrimage *inputImage,unsigned int blockSizeXY,unsigned int blockSizeZ){
		Inrimage *outputImage=new Inrimage(*inputImage);
		outputImage->setName(inputImage->getName().baseName()+"-mean.inr.gz");

		unsigned int sxy=blockSizeXY/2;
		unsigned int sz=blockSizeZ/2;

		storeBlockMean(inputImage,outputImage,sxy,sz);

		return outputImage;
	}


	template <class T>
	  void storeBlockStandardDeviationScalar(T ,Inrimage *inputImage,		
		Inrimage *outputImage,unsigned int sxy, unsigned int sz) 
	  {
		  T *inputData= (T *)inputImage->getData();
		  T *outputData= (T *)outputImage->getData();
		  T *pz,*py,*px,*out;
		  unsigned int minx, miny,minz,maxx,maxy,maxz;
		  unsigned int i,j,k,u,v,w;
		  unsigned int nx=inputImage->getX();
		  unsigned int ny=inputImage->getY();
		  unsigned int nz=inputImage->getZ();
		  double val,val2;
		  unsigned int size;


		  unsigned int offsetY=inputImage->getX();
		  unsigned int offsetZ=inputImage->getX()*inputImage->getY();

		  out=outputData;
		  for (k=0;k<nz;++k) {
			  if (k<sz)
				  minz=0;
			  else minz=k-sz;
			  if (k<(nz-sz)) {
				  maxz=k+sz;
			  } else 
				  maxz=nz-1;
			  for (j=0;j<ny;++j) {
				  if (j<sxy)
					  miny=0;
				  else miny=j-sxy;
				  if (j<(ny-sxy)) {
					  maxy=j+sxy;
				  } else 
					  maxy=ny-1;

				  for (i=0;i<nx;++i,++out) {
					  if (i<sxy)
						  minx=0;
					  else minx=i-sxy;
					  if (i<(nx-sxy)) {
						  maxx=i+sxy;
					  } else 
						  maxx=nx-1;

					  pz=inputData+offsetZ*minz+offsetY*miny+minx;
					  val=(double)0;
					  val2=(double)0;

					  for (w=minz;w<=maxz;++w,pz+=offsetZ) {
						  py=pz;
						  for (v=miny;v<=maxy;++v,py+=offsetY) {
							  px=py;
							  for (u=minx;u<=maxx;++u,++px) {
								  val+=(*px);
								  val2+=(*px)*(*px);
							  }
						  }
					  }
					  size=(maxx-minx+1)*(maxy-miny+1)*(maxz-minz+1);
					  val/=size;
					  val2/=size;
					  *out=(T) sqrt(val2-val*val);

				  }
			  }
		  }



	  }
#define BLOCKVARIANCESCALAR(TYPE,IN,OUT,SXY,SZ) storeBlockStandardDeviationScalar((TYPE)0,IN,OUT,SXY,SZ)

	  void storeBlockStandardDeviation(Inrimage *inputImage,Inrimage *outputImage,
		  unsigned int sxy,unsigned int sz){
		  switch(inputImage->getType()) {
  case Inrimage::WT_UNSIGNED_CHAR:
	  BLOCKVARIANCESCALAR(unsigned char,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_SIGNED_CHAR:
	  BLOCKVARIANCESCALAR(char,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_UNSIGNED_SHORT:
	  BLOCKVARIANCESCALAR(unsigned short,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_SIGNED_SHORT:
	  BLOCKVARIANCESCALAR(short,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_UNSIGNED_INT:
	  BLOCKVARIANCESCALAR(unsigned int,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_SIGNED_INT:
	  BLOCKVARIANCESCALAR(int,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_UNSIGNED_LONG:
	  BLOCKVARIANCESCALAR(unsigned LONGINT,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_SIGNED_LONG:
	  BLOCKVARIANCESCALAR(LONGINT,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_FLOAT:
	  BLOCKVARIANCESCALAR(float,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_DOUBLE:
	  BLOCKVARIANCESCALAR(double,inputImage,outputImage,sxy,sz);
	  break;
 /* case Inrimage::WT_RGB:
	  BLOCKVARIANCESCALAR(unsigned char,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_RGBA:
	  BLOCKVARIANCESCALAR(unsigned char,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_FLOAT_VECTOR:
	  BLOCKVARIANCESCALAR(float,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_DOUBLE_VECTOR:
	  BLOCKVARIANCESCALAR(double,inputImage,outputImage,sxy,sz);
	  break;*/
  default:
	  break;
		  }
	  }

	Inrimage *computeBlockStandardDeviation(Inrimage *inputImage,unsigned int blockSizeXY,unsigned int blockSizeZ){
		Inrimage *outputImage=new Inrimage(*inputImage);
		outputImage->setName(inputImage->getName().baseName()+"-stdDev.inr.gz");

		unsigned int sxy=blockSizeXY/2;
		unsigned int sz=blockSizeZ/2;

		storeBlockStandardDeviation(inputImage,outputImage,sxy,sz);

		return outputImage;
	}

	template <class T>
	  void storeBlockMedianScalar(T ,Inrimage *inputImage,		
		Inrimage *outputImage,unsigned int sxy, unsigned int sz) 
	  {
		  T *inputData= (T *)inputImage->getData();
		  T *outputData= (T *)outputImage->getData();
		  T *pz,*py,*px,*out;
		  unsigned int minx, miny,minz,maxx,maxy,maxz;
		  unsigned int i,j,k,u,v,w;
		  unsigned int nx=inputImage->getX();
		  unsigned int ny=inputImage->getY();
		  unsigned int nz=inputImage->getZ();

		  unsigned int offsetY=inputImage->getX();
		  unsigned int offsetZ=inputImage->getX()*inputImage->getY();

		  out=outputData;
		  for (k=0;k<nz;++k) {
			  if (k<sz)
				  minz=0;
			  else minz=k-sz;
			  if (k<(nz-sz)) {
				  maxz=k+sz;
			  } else 
				  maxz=nz-1;
			  for (j=0;j<ny;++j) {
				  if (j<sxy)
					  miny=0;
				  else miny=j-sxy;
				  if (j<(ny-sxy)) {
					  maxy=j+sxy;
				  } else 
					  maxy=ny-1;

				  for (i=0;i<nx;++i,++out) {
					  if (i<sxy)
						  minx=0;
					  else minx=i-sxy;
					  if (i<(nx-sxy)) {
						  maxx=i+sxy;
					  } else 
						  maxx=nx-1;

					  pz=inputData+offsetZ*minz+offsetY*miny+minx;
					  std::vector<T> val;

					  for (w=minz;w<=maxz;++w,pz+=offsetZ) {
						  py=pz;
						  for (v=miny;v<=maxy;++v,py+=offsetY) {
							  px=py;
							  for (u=minx;u<=maxx;++u,++px) {
								  val.push_back(*px);
							  }
						  }
					  }
					  nth_element(val.begin(),
						  val.begin()+val.size()/2, 
						  val.end());
					  *out= *(val.begin()+val.size()/2);

				  }
			  }
		  }



	  }
#define BLOCKMEDIANSCALAR(TYPE,IN,OUT,SXY,SZ) storeBlockMedianScalar((TYPE)0,IN,OUT,SXY,SZ)

	  void storeBlockMedian(Inrimage *inputImage,Inrimage *outputImage,
		  unsigned int sxy,unsigned int sz){
		  switch(inputImage->getType()) {
  case Inrimage::WT_UNSIGNED_CHAR:
	  BLOCKMEDIANSCALAR(unsigned char,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_SIGNED_CHAR:
	  BLOCKMEDIANSCALAR(char,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_UNSIGNED_SHORT:
	  BLOCKMEDIANSCALAR(unsigned short,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_SIGNED_SHORT:
	  BLOCKMEDIANSCALAR(short,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_UNSIGNED_INT:
	  BLOCKMEDIANSCALAR(unsigned int,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_SIGNED_INT:
	  BLOCKMEDIANSCALAR(int,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_UNSIGNED_LONG:
	  BLOCKMEDIANSCALAR(unsigned LONGINT,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_SIGNED_LONG:
	  BLOCKMEDIANSCALAR(LONGINT,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_FLOAT:
	  BLOCKMEDIANSCALAR(float,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_DOUBLE:
	  BLOCKMEDIANSCALAR(double,inputImage,outputImage,sxy,sz);
	  break;
 /* case Inrimage::WT_RGB:
	  BLOCKMEDIANSCALAR(unsigned char,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_RGBA:
	  BLOCKMEDIANSCALAR(unsigned char,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_FLOAT_VECTOR:
	  BLOCKMEDIANSCALAR(float,inputImage,outputImage,sxy,sz);
	  break;
  case Inrimage::WT_DOUBLE_VECTOR:
	  BLOCKMEDIANSCALAR(double,inputImage,outputImage,sxy,sz);
	  break;*/
  default:
	  break;
		  }
	  }

	Inrimage *computeBlockMedian(Inrimage *inputImage,unsigned int blockSizeXY,unsigned int blockSizeZ){
		Inrimage *outputImage=new Inrimage(*inputImage);
		outputImage->setName(inputImage->getName().baseName()+"-median.inr.gz");

		unsigned int sxy=blockSizeXY/2;
		unsigned int sz=blockSizeZ/2;

		storeBlockMedian(inputImage,outputImage,sxy,sz);

		return outputImage;
	}

	template <class T>
	double computeBlockSSDScalar(T, const Inrimage *image1,
		const Inrimage *image2,
		const unsigned int sizeX,const unsigned int sizeY,const unsigned int sizeZ,
		const unsigned int p1x,unsigned int p1y,const unsigned int p1z,
		const unsigned int p2x,const unsigned int p2y,const unsigned int p2z)
	{
		T *image1Data= (T *)image1->getData();
		T *image2Data= (T *)image2->getData();
		T *pp1z,*pp1y,*pp1x,*pp2z,*pp2y,*pp2x;
		unsigned int u,v,w;
		double val,tmp;

		unsigned int offset1Y=image1->getX();
		unsigned int offset1Z=image1->getX()*image1->getY();
		unsigned int offset2Y=image2->getX();
		unsigned int offset2Z=image2->getX()*image2->getY();

		val=0;

		pp1z=image1Data+offset1Z*p1z+offset1Y*p1y+p1x;
		pp2z=image2Data+offset2Z*p2z+offset2Y*p2y+p2x;

		for (w=0;w<sizeZ;++w,pp1z+=offset1Z,pp2z+=offset2Z) {
			pp1y=pp1z;
			pp2y=pp2z;
			for (v=0;v<sizeY;++v,pp1y+=offset1Y,pp2y+=offset2Y) {
				pp1x=pp1y;
				pp2x=pp2y;
				for (u=0;u<sizeX;++u,++pp1x,++pp2x) {
					tmp=((*pp1x)-(*pp2x));
					val+=tmp*tmp;
				}
			}
		}
		val/=sizeX*sizeY*sizeZ;
		return val;
	}
#define BLOCKSSDCALAR(TYPE,IN,OUT,SX,SY,SZ,P1X,P1Y,P1Z,P2X,P2Y,P2Z) return computeBlockSSDScalar((TYPE)0,IN,OUT,SX,SY,SZ,P1X,P1Y,P1Z,P2X,P2Y,P2Z);

	 double computeBlockSSD(const Inrimage *image1, 
		const Inrimage *image2,
		const unsigned int sizeX,const unsigned int sizeY,const unsigned int sizeZ,
		const unsigned int p1x,unsigned int p1y,const unsigned int p1z,
		const unsigned int p2x,const unsigned int p2y,const unsigned int p2z)
	 {
		   switch(image1->getType()) {
  case Inrimage::WT_UNSIGNED_CHAR:
	  BLOCKSSDCALAR(unsigned char,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_SIGNED_CHAR:
	  BLOCKSSDCALAR(char,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_UNSIGNED_SHORT:
	  BLOCKSSDCALAR(unsigned short,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_SIGNED_SHORT:
	  BLOCKSSDCALAR(short,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_UNSIGNED_INT:
	  BLOCKSSDCALAR(unsigned int,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_SIGNED_INT:
	  BLOCKSSDCALAR(int,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_UNSIGNED_LONG:
	  BLOCKSSDCALAR(unsigned LONGINT,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_SIGNED_LONG:
	  BLOCKSSDCALAR(LONGINT,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_FLOAT:
	  BLOCKSSDCALAR(float,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_DOUBLE:
	  BLOCKSSDCALAR(double,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
 /* case Inrimage::WT_RGB:
	  BLOCKSSDCALAR(unsigned char,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_RGBA:
	  BLOCKSSDCALAR(unsigned char,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_FLOAT_VECTOR:
	  BLOCKSSDCALAR(float,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_DOUBLE_VECTOR:
	  BLOCKSSDCALAR(double,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;*/
  default:
	  return 0;
	  break;
		  }
	 }

	template <class T>
	double computeBlockSADScalar(T, const Inrimage *image1,
		const Inrimage *image2,
		const unsigned int sizeX,const unsigned int sizeY,const unsigned int sizeZ,
		const unsigned int p1x,unsigned int p1y,const unsigned int p1z,
		const unsigned int p2x,const unsigned int p2y,const unsigned int p2z)
	{
		T *image1Data= (T *)image1->getData();
		T *image2Data= (T *)image2->getData();
		T *pp1z,*pp1y,*pp1x,*pp2z,*pp2y,*pp2x;
		unsigned int u,v,w;
		double val,tmp;

		unsigned int offset1Y=image1->getX();
		unsigned int offset1Z=image1->getX()*image1->getY();
		unsigned int offset2Y=image2->getX();
		unsigned int offset2Z=image2->getX()*image2->getY();

		val=0;

		pp1z=image1Data+offset1Z*p1z+offset1Y*p1y+p1x;
		pp2z=image2Data+offset2Z*p2z+offset2Y*p2y+p2x;

		for (w=0;w<sizeZ;++w,pp1z+=offset1Z,pp2z+=offset2Z) {
			pp1y=pp1z;
			pp2y=pp2z;
			for (v=0;v<sizeY;++v,pp1y+=offset1Y,pp2y+=offset2Y) {
				pp1x=pp1y;
				pp2x=pp2y;
				for (u=0;u<sizeX;++u,++pp1x,++pp2x) {
					tmp=((*pp1x)-(*pp2x));
					val+=fabs(tmp);
				}
			}
		}
		val/=sizeX*sizeY*sizeZ;
		return val;
	}
#define BLOCKSADCALAR(TYPE,IN,OUT,SX,SY,SZ,P1X,P1Y,P1Z,P2X,P2Y,P2Z) return computeBlockSADScalar((TYPE)0,IN,OUT,SX,SY,SZ,P1X,P1Y,P1Z,P2X,P2Y,P2Z);

	 double computeBlockSAD(const Inrimage *image1,
		const Inrimage *image2,
		const unsigned int sizeX,const unsigned int sizeY,const unsigned int sizeZ,
		const unsigned int p1x,unsigned int p1y,const unsigned int p1z,
		const unsigned int p2x,const unsigned int p2y,const unsigned int p2z)
	 {
		   switch(image1->getType()) {
  case Inrimage::WT_UNSIGNED_CHAR:
	  BLOCKSADCALAR(unsigned char,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_SIGNED_CHAR:
	  BLOCKSADCALAR(char,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_UNSIGNED_SHORT:
	  BLOCKSADCALAR(unsigned short,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_SIGNED_SHORT:
	  BLOCKSADCALAR(short,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_UNSIGNED_INT:
	  BLOCKSADCALAR(unsigned int,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_SIGNED_INT:
	  BLOCKSADCALAR(int,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_UNSIGNED_LONG:
	  BLOCKSADCALAR(unsigned LONGINT,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_SIGNED_LONG:
	  BLOCKSADCALAR(LONGINT,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_FLOAT:
	  BLOCKSADCALAR(float,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_DOUBLE:
	  BLOCKSADCALAR(double,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
 /* case Inrimage::WT_RGB:
	  BLOCKSADCALAR(unsigned char,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_RGBA:
	  BLOCKSADCALAR(unsigned char,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_FLOAT_VECTOR:
	  BLOCKSADCALAR(float,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_DOUBLE_VECTOR:
	  BLOCKSADCALAR(double,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;*/
  default:
	  return 0;
	  break;
		  }
	 }

	template <class T>
	double computeBlockCorrelationScalar(T, const Inrimage *image1,
		const Inrimage *image2,
		const unsigned int sizeX,const unsigned int sizeY,const unsigned int sizeZ,
		const unsigned int p1x,unsigned int p1y,const unsigned int p1z,
		const unsigned int p2x,const unsigned int p2y,const unsigned int p2z)
	{
		T *image1Data= (T *)image1->getData();
		T *image2Data= (T *)image2->getData();
		T *pp1z,*pp1y,*pp1x,*pp2z,*pp2y,*pp2x;
		unsigned int u,v,w,size;
		double val1,val2,val11,val22,val12;

		unsigned int offset1Y=image1->getX();
		unsigned int offset1Z=image1->getX()*image1->getY();
		unsigned int offset2Y=image2->getX();
		unsigned int offset2Z=image2->getX()*image2->getY();

		val1=val2=val11=val22=val12=0;


		pp1z=image1Data+offset1Z*p1z+offset1Y*p1y+p1x;
		pp2z=image2Data+offset2Z*p2z+offset2Y*p2y+p2x;

		for (w=0;w<sizeZ;++w,pp1z+=offset1Z,pp2z+=offset2Z) {
			pp1y=pp1z;
			pp2y=pp2z;
			for (v=0;v<sizeY;++v,pp1y+=offset1Y,pp2y+=offset2Y) {
				pp1x=pp1y;
				pp2x=pp2y;
				for (u=0;u<sizeX;++u,++pp1x,++pp2x) {
					val1+=(*pp1x);
					val2+=(*pp2x);
					val11+=(*pp1x)*(*pp1x);
					val22+=(*pp2x)*(*pp2x);
					val12+=(*pp1x)*(*pp2x);
				}
			}
		}
		size=sizeX*sizeY*sizeZ;
		val1/=size;
		val2/=size;
		val11/=size;
		val22/=size;
		val12/=size;
		val11-=val1*val1;
		val22-= val2*val2;
		val12-= val1*val2;
		if ((val11!=0)&& (val22!=0))
			return (val12/(sqrt(val11*val22)));
		else 
			return 0;
	}
#define BLOCKCORRELATIONCALAR(TYPE,IN,OUT,SX,SY,SZ,P1X,P1Y,P1Z,P2X,P2Y,P2Z) return computeBlockCorrelationScalar((TYPE)0,IN,OUT,SX,SY,SZ,P1X,P1Y,P1Z,P2X,P2Y,P2Z);

	 double computeBlockCorrelation(const Inrimage *image1,
		const Inrimage *image2,
		const unsigned int sizeX,const unsigned int sizeY,const unsigned int sizeZ,
		const unsigned int p1x,unsigned int p1y,const unsigned int p1z,
		const unsigned int p2x,const unsigned int p2y,const unsigned int p2z)
	 {
		   switch(image1->getType()) {
  case Inrimage::WT_UNSIGNED_CHAR:
	  BLOCKCORRELATIONCALAR(unsigned char,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_SIGNED_CHAR:
	  BLOCKCORRELATIONCALAR(char,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_UNSIGNED_SHORT:
	  BLOCKCORRELATIONCALAR(unsigned short,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_SIGNED_SHORT:
	  BLOCKCORRELATIONCALAR(short,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_UNSIGNED_INT:
	  BLOCKCORRELATIONCALAR(unsigned int,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_SIGNED_INT:
	  BLOCKCORRELATIONCALAR(int,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_UNSIGNED_LONG:
	  BLOCKCORRELATIONCALAR(unsigned LONGINT,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_SIGNED_LONG:
	  BLOCKCORRELATIONCALAR(LONGINT,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_FLOAT:
	  BLOCKCORRELATIONCALAR(float,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_DOUBLE:
	  BLOCKCORRELATIONCALAR(double,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
 /* case Inrimage::WT_RGB:
	  BLOCKCORRELATIONCALAR(unsigned char,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_RGBA:
	  BLOCKCORRELATIONCALAR(unsigned char,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_FLOAT_VECTOR:
	  BLOCKCORRELATIONCALAR(float,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;
  case Inrimage::WT_DOUBLE_VECTOR:
	  BLOCKCORRELATIONCALAR(double,image1,image2,sizeX,sizeY,sizeZ,p1x,p1y,p1z,p2x,p2y,p2z);
	  break;*/
  default:
	  return 0;
	  break;
		  }
	 }

	 Inrimage *computeSSDWithMask(const Inrimage *image, const Inrimage *maskImage){
		 if (image->getType()!=maskImage->getType())
			 return 0;
		 if ((image->getVX()!=maskImage->getVX()) || (image->getVY()!=maskImage->getVY()) ||
			 (image->getVZ()!=maskImage->getVZ()))
			 return 0;
		unsigned int mx=maskImage->getX();
		unsigned int my=maskImage->getY();
		unsigned int mz=maskImage->getZ();

		unsigned int ix=image->getX();
		unsigned int iy=image->getY();
		unsigned int iz=image->getZ();

		 if ((ix<mx) || (iy<my) || (iz<mz)) 
			 return 0;
		 Inrimage *outputImage=new Inrimage(ix,iy,iz,Inrimage::WT_DOUBLE,
			 1,VM_INTERLACED,image->getVX(),image->getVY(),image->getVZ());
		 outputImage->setName(image->getName().baseName()+"-maskSSD.inr.gz");

		 unsigned int i,j,k,sx,sy,sz;
		 sx=mx/2;sy=my/2;sz=mz/2;
		 unsigned int maxx,maxy,maxz,minx,miny,minz;
		 double *outputData=(double *)outputImage->getData();
		 double *out;

		 out=outputData;
		  for (k=0;k<iz;++k) {
			  if (k<sz)
				  minz= k;
			  else minz=sz;
			  if (k<(iz-sz)) {
				  maxz=sz;
			  } else 
				  maxz=iz-1-k;
			  for (j=0;j<iy;++j) {
				  if (j<sy)
					  miny=j;
				  else miny=sy;
				  if (j<(iy-sy)) {
					  maxy=sy;
				  } else 
					  maxy=iy-1-j;

				  for (i=0;i<ix;++i,++out) {
					  if (i<sx)
						  minx=i;
					  else minx=sx;
					  if (i<(ix-sy)) {
						  maxx=sx;
					  } else 
						  maxx=ix-1-i;
					  *out=computeBlockSSD(image,maskImage,(maxx+minx+1),
						  (maxy+miny+1),(maxz+minz+1),i-minx,j-miny,k-minz,
						  sx-minx,sy-miny,sz-minz);

				  }
			  }
		  }

		  return outputImage;
	 }

	 Inrimage *computeSADWithMask(const Inrimage *image, const Inrimage *maskImage){
		 if (image->getType()!=maskImage->getType())
			 return 0;
		 if ((image->getVX()!=maskImage->getVX()) || (image->getVY()!=maskImage->getVY()) ||
			 (image->getVZ()!=maskImage->getVZ()))
			 return 0;
		unsigned int mx=maskImage->getX();
		unsigned int my=maskImage->getY();
		unsigned int mz=maskImage->getZ();

		unsigned int ix=image->getX();
		unsigned int iy=image->getY();
		unsigned int iz=image->getZ();

		 if ((ix<mx) || (iy<my) || (iz<mz)) 
			 return 0;
		 Inrimage *outputImage=new Inrimage(ix,iy,iz,Inrimage::WT_DOUBLE,
			 1,VM_INTERLACED,image->getVX(),image->getVY(),image->getVZ());
		 outputImage->setName(image->getName().baseName()+"-maskSAD.inr.gz");

		 unsigned int i,j,k,sx,sy,sz;
		 sx=mx/2;sy=my/2;sz=mz/2;
		 unsigned int maxx,maxy,maxz,minx,miny,minz;
		 double *outputData=(double *)outputImage->getData();
		 double *out;

		 out=outputData;
		  for (k=0;k<iz;++k) {
			  if (k<sz)
				  minz= k;
			  else minz=sz;
			  if (k<(iz-sz)) {
				  maxz=sz;
			  } else 
				  maxz=iz-1-k;
			  for (j=0;j<iy;++j) {
				  if (j<sy)
					  miny=j;
				  else miny=sy;
				  if (j<(iy-sy)) {
					  maxy=sy;
				  } else 
					  maxy=iy-1-j;

				  for (i=0;i<ix;++i,++out) {
					  if (i<sx)
						  minx=i;
					  else minx=sx;
					  if (i<(ix-sy)) {
						  maxx=sx;
					  } else 
						  maxx=ix-1-i;
					  *out=computeBlockSAD(image,maskImage,(maxx+minx+1),
						  (maxy+miny+1),(maxz+minz+1),i-minx,j-miny,k-minz,
						  sx-minx,sy-miny,sz-minz);

				  }
			  }
		  }

		  return outputImage;
	 }
	 
	 Inrimage *computeCorrelationRatioWithMask(const Inrimage *image, const Inrimage *maskImage){
		 if (image->getType()!=maskImage->getType())
			 return 0;
		 if ((image->getVX()!=maskImage->getVX()) || (image->getVY()!=maskImage->getVY()) ||
			 (image->getVZ()!=maskImage->getVZ()))
			 return 0;
		unsigned int mx=maskImage->getX();
		unsigned int my=maskImage->getY();
		unsigned int mz=maskImage->getZ();

		unsigned int ix=image->getX();
		unsigned int iy=image->getY();
		unsigned int iz=image->getZ();

		 if ((ix<mx) || (iy<my) || (iz<mz)) 
			 return 0;
		 Inrimage *outputImage=new Inrimage(ix,iy,iz,Inrimage::WT_DOUBLE,
			 1,VM_INTERLACED,image->getVX(),image->getVY(),image->getVZ());
		 outputImage->setName(image->getName().baseName()+"-maskCR.inr.gz");

		 unsigned int i,j,k,sx,sy,sz;
		 sx=mx/2;sy=my/2;sz=mz/2;
		 unsigned int maxx,maxy,maxz,minx,miny,minz;
		 double *outputData=(double *)outputImage->getData();
		 double *out;

		 out=outputData;
		  for (k=0;k<iz;++k) {
			  if (k<sz)
				  minz= k;
			  else minz=sz;
			  if (k<(iz-sz)) {
				  maxz=sz;
			  } else 
				  maxz=iz-1-k;
			  for (j=0;j<iy;++j) {
				  if (j<sy)
					  miny=j;
				  else miny=sy;
				  if (j<(iy-sy)) {
					  maxy=sy;
				  } else 
					  maxy=iy-1-j;

				  for (i=0;i<ix;++i,++out) {
					  if (i<sx)
						  minx=i;
					  else minx=sx;
					  if (i<(ix-sy)) {
						  maxx=sx;
					  } else 
						  maxx=ix-1-i;
					  *out=computeBlockCorrelation(image,maskImage,(maxx+minx+1),
						  (maxy+miny+1),(maxz+minz+1),i-minx,j-miny,k-minz,
						  sx-minx,sy-miny,sz-minz);

				  }
			  }
		  }

		  return outputImage;
	 }

	 bool findBlockCorrespondance(const Inrimage *refImage, const Inrimage *targetImage,
 		 const unsigned int negativeBlockSizeX,const unsigned int negativeBlockSizeY,const unsigned int negativeBlockSizeZ,
		 const unsigned int positiveBlockSizeX,const unsigned int positiveBlockSizeY,const unsigned int positiveBlockSizeZ,
		 const unsigned int pix, const unsigned int piy, const unsigned int piz,
		 const unsigned int minptx, const unsigned int minpty, const unsigned int minptz,
		 const unsigned int _searchSizeX,const unsigned int _searchSizeY,const unsigned int _searchSizeZ,
		 const unsigned int searchStepX,const unsigned int searchStepY,const unsigned int searchStepZ,
		 unsigned int &bestptx, unsigned int &bestpty, unsigned int &bestptz,
		 double variance[6]) 
	 {

			 unsigned int ix=targetImage->getX();
			 unsigned int iy=targetImage->getY();
			 unsigned int iz=targetImage->getZ();

			 unsigned int searchSizeX,searchSizeY,searchSizeZ;
			
			 if ((iz-minptz)<_searchSizeZ) 
				 searchSizeZ=iz-minptz; 
			 else 
				 searchSizeZ=_searchSizeZ;
			 if ((iy-minpty)<_searchSizeY) 
				 searchSizeY=iy-minpty;
			 else 
				 searchSizeY=_searchSizeY;
			 if ((ix-minptx)<_searchSizeX) 
				 searchSizeX=ix-minptx;
			 else 
				 searchSizeX=_searchSizeX;

			 unsigned int nsx=(searchSizeX-1)/searchStepX +1;
			 unsigned int nsy=(searchSizeY-1)/searchStepY +1;
			 unsigned int nsz=(searchSizeZ-1)/searchStepZ +1;

			 unsigned int ptx=minptx;
			 unsigned int pty=minpty;
			 unsigned int ptz=minptz;

			 unsigned int maxBsizeX,maxBsizeY,maxBsizeZ;

			 double *correlationContainer= new double [nsx*nsy*nsz];
			 memset((void *)correlationContainer,nsx*nsy*nsz,0);
			 double *c=correlationContainer;
			 double minCorrelation= MAX_FLOAT;
			 unsigned int besti,bestj,bestk,i,j,k;

			 for (k=0;(k<searchSizeZ);k+=searchStepZ,ptz+=searchStepZ) {
				 if ((ptz+positiveBlockSizeZ)>=iz)
					 maxBsizeZ=iz-ptz-1;
				 else 
					 maxBsizeZ=positiveBlockSizeZ;

				 for (j=0;(j<searchSizeY) ;j+=searchStepY,pty+=searchStepY) {
					 if ((pty+positiveBlockSizeY)>=iy)
						 maxBsizeY=iy-pty-1;
					 else 
						 maxBsizeY=positiveBlockSizeY;

					 for (i=0;(i<searchSizeX);i+=searchStepX,ptx+=searchStepX,c++) {
						 if ((ptx+positiveBlockSizeX)>=ix)
							 maxBsizeX=ix-ptx-1;
						 else 
							 maxBsizeX=positiveBlockSizeX;

						  *c=computeBlockSSD(refImage,targetImage,
							  (negativeBlockSizeX+maxBsizeX+1), (negativeBlockSizeY+maxBsizeY+1),
							  (negativeBlockSizeZ+maxBsizeZ+1),
							  pix-negativeBlockSizeX,piy-negativeBlockSizeY,piz-negativeBlockSizeZ,
							  ptx-negativeBlockSizeX,pty-negativeBlockSizeY,ptz-negativeBlockSizeZ);

						  if ((*c)<minCorrelation) {
							  minCorrelation= *c;
							  besti=i; bestj=j; bestk=k;
						  }
					 }
				 }
			 }
			assert(c==(correlationContainer+nsx*nsy*nsz));
			assert(nsx>2);
			assert(nsy>2);
			// compute variance as second order derivative
			besti/=searchStepX; bestj/=searchStepY; bestk/=searchStepZ;
			unsigned int offset=besti+bestj*nsx+bestk*nsx*nsy;
			unsigned int ns=nsx*nsy;
			unsigned int offsetXp,weightX;
			unsigned int offsetYp,weightY;
			unsigned int offsetZp,weightZ;
			int offsetXm,offsetYm,offsetZm;
			if (besti==0) {
				variance[0]=correlationContainer[offset] -
					2 *correlationContainer[offset+1] +correlationContainer[offset+2];
				offsetXp=1; offsetXm=0; weightX=1;
			} else if (besti==(nsx-1)) {
				variance[0]=correlationContainer[offset] -
					2 *correlationContainer[offset-1] +correlationContainer[offset-2];
				offsetXp=0; offsetXm= -1; weightX=1;
			} else {
				variance[0]=correlationContainer[offset-1] -
					2 *correlationContainer[offset] +correlationContainer[offset+1];
				offsetXp=1; offsetXm=-1; weightX=2;
			}
			variance[0]/=(searchStepX*searchStepX);

			if (bestj==0) {
				variance[1]=correlationContainer[offset] -
					2 *correlationContainer[offset+nsx] +correlationContainer[offset+2*nsx];
					offsetYp=nsx; offsetYm=0; weightY=1;
			} else if (bestj==(nsy-1)) {
				variance[1]=correlationContainer[offset] -
					2 *correlationContainer[offset-nsx] +correlationContainer[offset-2*nsx];
					offsetYp=0; offsetYm= -(int)nsx; weightY=1;
			} else {
				variance[1]=correlationContainer[offset-nsx] -
					2 *correlationContainer[offset] +correlationContainer[offset+nsx];
					offsetYp=nsx; offsetYm= -(int)nsx; weightY=2;
			}
			variance[1]/=(searchStepY*searchStepY);
			
			if (nsz>1) {
				if (bestk==0) {
					variance[2]=correlationContainer[offset] -
						2 *correlationContainer[offset+ns] +correlationContainer[offset+2*ns];
					offsetZp=ns; offsetZm=0; weightZ=1;
				} else if (bestk==(nsz-1)) {
					variance[2]=correlationContainer[offset] -
						2 *correlationContainer[offset-ns] +correlationContainer[offset-2*ns];
					offsetZp=0; offsetZm=-(int)ns; weightZ=1;
				} else {
					variance[2]=correlationContainer[offset-ns] -
						2 *correlationContainer[offset] +correlationContainer[offset+ns];
					offsetZp=ns; offsetZm= -(int)ns; weightZ=2;
				}
				variance[2]/=(searchStepZ*searchStepZ);


				variance[4]= correlationContainer[offset+offsetXp +offsetZp] +
					correlationContainer[offset+offsetXm +offsetZm] -
					correlationContainer[offset+offsetXm +offsetZp] -
					correlationContainer[offset+offsetXp +offsetZm];
				variance[4] /=(searchStepX*searchStepZ*weightX*weightZ);

				variance[5]= correlationContainer[offset+offsetZp +offsetYp] +
					correlationContainer[offset+offsetZm +offsetYm] -
					correlationContainer[offset+offsetZm +offsetYp] -
					correlationContainer[offset+offsetZp +offsetYm];
				variance[5] /=(searchStepZ*searchStepY*weightZ*weightY);
			}
			variance[3]= correlationContainer[offset+offsetXp +offsetYp] +
				correlationContainer[offset+offsetXm +offsetYm] -
				correlationContainer[offset+offsetXm +offsetYp] -
				correlationContainer[offset+offsetXp +offsetYm];
			variance[3] /=(searchStepX*searchStepY*weightX*weightY);

			

			 delete [] correlationContainer;
			 return true;
	 }
} // namespace yav



