#include <inrimage/Inrimage.h>


namespace yav {


template <class FROM, class TO>
void convertScaleScalarBuffer(FROM b, TO c, 
			      Inrimage &inr, Voxel *imin, Voxel *imax){
  TO *to;
  FROM *from;
  unsigned LONGINT i, size = inr.getX() * inr.getY() * inr.getZ();
  double a, m1, m2;

  from = (FROM *) (inr.getData());
  to = new TO[size];

  m1 = imax->getValueAsDouble() / 2.0 + imin->getValueAsDouble() / 2.0;
  m2 = (double) maxTypeValue<TO>(0) / 2.0 +
    (double) minTypeValue<TO>(0) / 2.0;
  a = ((double) maxTypeValue<TO>(0) - m2) / (imax->getValueAsDouble() - m1);

  m2 += 0.5;

  for(i = 0; i < size; i++, from++, to++)
    *to = (TO) (a * ((double) *from - m1) + m2);

  to -= size;

  delete[] (unsigned char *) inr.setData(to);
}



//
// scale an image of a given type
//
template <class FROM>
void scaleScalarBuffer(FROM b, Inrimage &inr, Voxel *imin, Voxel *imax){
  FROM *to;
  FROM *from;
  unsigned LONGINT i, size = inr.getX() * inr.getY() * inr.getZ();
  double a, m1, m2;

  from = to = (FROM *) (inr.getData());
  
  m1 = imax->getValueAsDouble() / 2.0 + imin->getValueAsDouble() / 2.0;
  m2 = (double) maxTypeValue<FROM>(0) / 2.0 + (double) minTypeValue<FROM>(0) / 2.0;
  a = ((double) maxTypeValue<FROM>(0) - m2) / (imax->getValueAsDouble() - m1);

  m2 += 0.5;
  
  for(i = 0; i < size; i++, from++, to++)
    *to = (FROM) (a * ((double) *from - m1) + m2);

}




 /// buffer convertion routine
 template <class FROM, class TO>
   void convertCastScalarBuffer(FROM a, TO b, Inrimage &inr){
  TO *to;
  FROM *from;
  unsigned LONGINT i, size = inr.getX() * inr.getY() * inr.getZ();

  from = (FROM *) inr.getData();
  to = new TO[size];
  for(i = 0; i < size; i++, from++, to++) {
    *to = (TO) *from;
  }
  to -= size;

  delete[] (unsigned char *) inr.setData(to);
}
	 	 
 /// buffer convertion routine
 template <class FROM>
   void convertCastToRGBBuffer(FROM a, Inrimage &inr){
  unsigned char *to, uc;
  FROM *from;
  unsigned LONGINT i, size = inr.getX() * inr.getY() * inr.getZ();
  inr.setVdim(3);

  from = (FROM *) inr.getData();
  to = new unsigned char[size * 3];
  for(i = 0; i < size; i++, from++) {
    uc = (unsigned char) *from;
    *to++ = uc;
    *to++ = uc;
    *to++ = uc;
  }
  to -= 3 * size;
  delete[] (unsigned char *) inr.setData(to);
  inr.setVectorialMode(VM_INTERLACED);
}
 /// buffer convertion routine
 template <class FROM>
   void convertCastToRGBABuffer(FROM a, Inrimage &inr) {
   unsigned char *to, uc;
  FROM *from;
  unsigned LONGINT i, size = inr.getX() * inr.getY() * inr.getZ();
  inr.setVdim(4);

  from = (FROM *) inr.getData();
  to = new unsigned char[size * 4];
  for(i = 0; i < size; i++, from++) {
    uc = (unsigned char) *from;
#if (defined (_ALPHA_) || defined (_LINUX_))
    *to++ = 0xFF;
    *to++ = uc;
    *to++ = uc;
    *to++ = uc;
#else
    *to++ = uc;
    *to++ = uc;
    *to++ = uc;
    *to++ = 0xFF;
#endif
  }
  to -= 4 * size;
  delete[] (unsigned char *) inr.setData(to);
  inr.setVectorialMode(VM_INTERLACED);
}

/* From the Color FAQ ( http://www.inforamp.net/~poynton/Poynton-colour.html)
   ...
   8. What weighting of red, green and blue corresponds to luminance?
   
   Direct acquisition of luminance requires use of a very specific spectral
   weighting. However, luminance can also be computed as a weighted sum
   of red, green and blue components.
   ...
   The weights to compute true CIE luminance from
   linear red, green and blue, for the Rec. 709, are these:
   Y = 0.2126 R + 0.7152 * G + 0.0722 B
   ...
   
   10. What is luma?

   It is useful in a video system to convey a component representative of
   luminance and two other components representative of color. It is important
   to convey the component representative of luminance in such a way
   that noise (or quantization) introduced in transmission, processing and
   storage has a perceptually similar effect across the entire tone scale from
   black to white. The ideal way to accomplish these goals would be to form
   a luminance signal by matrixing RGB, then subjecting luminance to a
   nonlinear transfer function similar to the L* function.

   There are practical reasons in video to perform these operations in the
   opposite order. First a nonlinear transfer function - gamma correction - is
   applied to each of the linear R, G and B. Then a weighted sum of the
   nonlinear components is computed to form a signal representative of
   luminance. The resulting component is related to brightness but is not
   CIE luminance. Many video engineers call it luma and give it the symbol Y'.
   It is often carelessly called luminance and given the symbol Y. You
   must be careful to determine whether a particular author assigns a linear
   or nonlinear interpretation to the term luminance and the symbol Y.

   The coefficients that correspond to the "NTSC" red, green and blue CRT
   phosphors of 1953 are standardized in ITU-R Recommendation BT. 601
   (formerly CCIR Rec. 601). I call it Rec. 601. To compute nonlinear video
   luma from nonlinear red, green and blue:
   ...
   Y' = 0.299 * R' + 0.587 * G' + 0.114 * B'
   ...
*/

#if ( defined ( _LINEAR_COLORS_ ) )
static double red_coeff = 0.2126;
static double green_coeff = 0.7152;
static double blue_coeff = 0.0722;
#else
static double red_coeff = 0.299;
static double green_coeff = 0.587;
static double blue_coeff =  0.114;
#endif

/// buffer convertion routine
template <class TO>
void convertCastFromRGBBuffer(TO a, Inrimage &inr) 
{
  TO *to;
  unsigned char *from;
  unsigned LONGINT i, size = inr.getX() * inr.getY() * inr.getZ();
  inr.setVdim(1);
  
  from = (unsigned char *) inr.getData();
  to = new TO[size];

  for(i = 0; i < size; i++, from += 3, to++) {
    /* Instead of the average of the three components
       it's preferable to compute the intensity.
       GM.
     *to = (TO) ((*from + *(from+1) + *(from+2)) / 3);
     */
    *to = (TO) (red_coeff * *from + green_coeff * *(from+1) + blue_coeff * *(from+2));
  }
  to -= size;

  delete[] (unsigned char *) inr.setData(to);
  inr.setVectorialMode(VM_SCALAR);
}

/// buffer convertion routine
template <class TO>
void convertCastFromRGBABuffer(TO a, Inrimage &inr)
{
  TO *to;
  unsigned char *from;
  unsigned LONGINT i, size = inr.getX() * inr.getY() * inr.getZ();
  inr.setVdim(1);
  
  from = (unsigned char *) inr.getData();
  to = new TO[size];
  
  for(i = 0; i < size; i++, from += 4, to++) {
#if (defined (_ALPHA_) || defined (_LINUX_) )
    /* *to = (TO) ((*(from+1) + *(from+2) + *(from+3)) / 3); */
    *to = (TO) (red_coeff * *(from+1) + green_coeff * *(from+2) + blue_coeff * *(from+3));
#else 
    /* *to = (TO) ((*from + *(from+1) + *(from+2)) / 3); */
    *to = (TO) (red_coeff * *from + green_coeff * *(from+1) + blue_coeff * *(from+2));
#endif
  }

  to -= size;
  delete[] (unsigned char *) inr.setData(to);
  inr.setVectorialMode(VM_SCALAR);
}

/// buffer convertion routine
template <class FROM>
    void convertScaleToRGBBuffer(FROM b,Inrimage &inr, Voxel *imin, Voxel *imax){
  unsigned char *to, uc;
  FROM *from;
  unsigned LONGINT i, size = inr.getX() * inr.getY() * inr.getZ();
  double a;
  
  inr.setVdim(3);
  from = (FROM *) inr.getData();
  to = new unsigned char[size * 3];

  a = 255.0 / (imax->getValueAsDouble() - imin->getValueAsDouble());

  for(i = 0; i < size; i++, from++) {
    uc = (unsigned char) (a * ((double) *from - imin->getValueAsDouble())
			  + 0.5);
    *to++ = uc;
    *to++ = uc;
    *to++ = uc;
  }
  to -= 3 * size;

  delete[] (unsigned char *) inr.setData(to);
  inr.setVectorialMode(VM_INTERLACED);
}
  /// buffer convertion routine
  template <class FROM>
    void convertScaleToRGBABuffer(FROM b,Inrimage &inr, Voxel *imin, Voxel *imax){
  unsigned char *to, uc;
  FROM *from;
  unsigned LONGINT i, size = inr.getX() * inr.getY() * inr.getZ();
  double a;
  
  inr.setVdim(4);
  from = (FROM *) inr.getData();
  to = new unsigned char[size * 4];

  a = 255.0 / (imax->getValueAsDouble() - imin->getValueAsDouble());

  for(i = 0; i < size; i++, from++) {
    uc = (unsigned char) (a * ((double) *from - imin->getValueAsDouble())
			  + 0.5);
#if (defined (_ALPHA_) || defined (_LINUX_))
    *to++ = 0xFF;
    *to++ = uc;
    *to++ = uc;
    *to++ = uc;
#else
    *to++ = uc;
    *to++ = uc;
    *to++ = uc;
    *to++ = 0xFF;
#endif
  }
  to -= 4 * size;

  delete[] (unsigned char *) inr.setData(to);  
  inr.setVectorialMode(VM_INTERLACED);
}
  /// buffer convertion routine
  template <class TO>
    void convertScaleFromRGBBuffer(TO c, Inrimage &inr ){
    TO *to;
  unsigned char *from;
  unsigned LONGINT i, size = inr.getX() * inr.getY() * inr.getZ();
  double a, b;
  
  inr.setVdim(1);
  from = (unsigned char *) inr.getData();
  to = new TO[size];

  a = (maxTypeValue<TO>(0) / 255.0 - minTypeValue<TO>(0) / 255.0);

  b = minTypeValue<TO>(0) + 0.5;
  for(i = 0; i < size; i++, to++, from += 3)
    *to = (TO) (a * ((double) (*from + *(from+1) + *(from+2)) / 3.0) + b);
  to -= size;

  delete[] (unsigned char *) inr.setData(to);
  inr.setVectorialMode(VM_SCALAR);
}

   template <class TO>
    void convertScaleFromRGBABuffer(TO c, Inrimage &inr) {
     TO *to;
  unsigned char *from;
  unsigned LONGINT i, size = inr.getX() * inr.getY() * inr.getZ();
  double a, b;
  
  inr.setVdim(1);
  from = (unsigned char *) inr.getData();
  to = new TO[size];

  a = (maxTypeValue<TO>(0) - minTypeValue<TO>(0)) / 255.0;

  b = minTypeValue<TO>(0) + 0.5;
  for(i = 0; i < size; i++, to++, from += 4)
#if (defined (_ALPHA_) || defined (_LINUX_))
    *to = (TO) (a * ((double) (*(from+1) + *(from+2) + *(from+3)) / 3.0) + b);
#else
    *to = (TO) (a * ((double) (*from + *(from+1) + *(from+2)) / 3.0) + b);
#endif
  to -= size;

  delete[] (unsigned char *) inr.setData(to);
  inr.setVectorialMode(VM_SCALAR);
}


  template <class FROM, class TO>
     void convertFromScalarToVectorialBuffer(FROM a, TO b, Inrimage &inr){
     TO *to;
  FROM *from;
  unsigned LONGINT i, size = inr.getX() * inr.getY() * inr.getZ();
  inr.setVdim(1);

  from = (FROM *) inr.getData();
  to = new TO[size];

  for(i = 0; i < size; i++, to++, from++) *to = (TO) *from;

  to -= size;
  delete[] (unsigned char *) inr.setData(to);
  inr.setVectorialMode(VM_INTERLACED);
}

   template <class TO>
    void convertFromRGBToVectorialBuffer(TO a, Inrimage &inr){

  TO *to;
  unsigned char *from;
  unsigned LONGINT i, size = 3 * inr.getX() * inr.getY() * inr.getZ();
  inr.setVdim(3);

  from = (unsigned char *) inr.getData();
  to = new TO[size];

  for(i = 0; i < size; i++, to++, from++) *to = (TO) *from;

  to -= size;
  delete[] (unsigned char *) inr.setData(to); 
  inr.setVectorialMode(VM_INTERLACED);
}


    template <class TO>
    void convertFromRGBAToVectorialBuffer(TO a, Inrimage &inr){
  TO *to;
  unsigned char *from;
  unsigned LONGINT i, size = inr.getX() * inr.getY() * inr.getZ();
  inr.setVdim(4);

  from = (unsigned char *) inr.getData();
  to = new TO[size * 4];

#if (defined (_ALPHA_) || defined (_LINUX_))
  for(i = 0; i < size; i++, to += 4, from += 4) {
    *to = (TO) *(from+3);
    *(to+1) = (TO) *(from+2);
    *(to+2) = (TO) *(from+1);
    *(to+3) = (TO) *from;
  }
#else
  for(i = 0; i < 4 * size; i++, to++, from++)
    *to = (TO) *from;
#endif

  to -= 4 * size;
  delete[] (unsigned char *) inr.setData(to);  
  inr.setVectorialMode(VM_INTERLACED);
}


  template <class FROM, class TO>
    void convertCastFromVectorialToScalarBuffer(FROM a, TO b, Inrimage &inr){
  TO *to;
  FROM *from;
  double val;
  unsigned LONGINT i, j, size = inr.getX() * inr.getY() * inr.getZ();

  to = new TO[size];
  from = (FROM *) inr.getData();

  if(inr.getVectorialMode() == VM_INTERLACED) {
    for(i = 0; i < size; i++, to++) {
      val = 0.0;
      for(j = 0; j < inr.getVdim(); j++, from++) val += *from * *from;
      *to = (TO) sqrt(val);
    }
  }
  else {
    FROM *fv;

    for(i = 0; i < size; i++, to++, from++) {
      val = 0.0;
      for(fv = from, j = 0; j < inr.getVdim(); j++, fv += size) val += *fv * *fv;
      *to = (TO) sqrt(val);
    }
  }
  to -= size;

  inr.setVdim(1);
  delete[] (unsigned char *) inr.setData(to);
  inr.setVectorialMode(VM_SCALAR);
}

   template <class FROM, class TO>
    void convertScaleFromVectorialToScalarBuffer(FROM c, TO b, Inrimage &inr){

  TO *to;
  FROM *from;
  double *tmp, max, a;
  unsigned LONGINT i, j, size = inr.getX() * inr.getY() * inr.getZ();

  from = (FROM *) inr.getData();
  to = new TO[size];
  tmp = new double[size];
  max = minTypeValue<double>(0);

  if(inr.getVectorialMode() == VM_INTERLACED) {
    for(i = 0; i < size; i++, tmp++) {
      *tmp = 0.0;
      for(j = 0; j < inr.getVdim(); j++, from++) *tmp += *from * *from;
      *tmp = sqrt(*tmp);
      if(*tmp > max) max = *tmp;
    }
  }

  else {
    FROM *fv;

    for(i = 0; i < size; i++, tmp++, from++) {
      *tmp = 0.0;
      for(j = 0, fv = from; j < inr.getVdim(); j++, fv += size) *tmp += *fv * *fv;
      *tmp = sqrt(*tmp);
      if(*tmp > max) max = *tmp;
    }
  }
    
  a = (maxTypeValue<TO>(0) / max - minTypeValue<TO>(0) / max);
    
  tmp -= size;
  for(i = 0; i < size; i++, tmp++, to++)
    *to = (TO) (a * *tmp + minTypeValue<TO>(0) + 0.5);

  to -= size;
  tmp -= size;
  delete[] tmp;

  inr.setVdim(1);

  delete[] (unsigned char *) inr.setData(to);
  inr.setVectorialMode(VM_SCALAR);
}


   template <class FROM>
    void convertFromVectorialToRGBBuffer(FROM a, Inrimage &inr) {
  unsigned char *to;
  FROM *from;
  unsigned LONGINT i, size = inr.getX() * inr.getY() * inr.getZ();

  from = (FROM *) inr.getData();
  to = new unsigned char[size * 3];

  if(inr.getVdim() == 3) {
    if(inr.getVectorialMode() == VM_INTERLACED) {
      for(i = 0; i < 3 * size; i++)
	*to++ = (unsigned char) *from++;
    }
    else {
      for(i = 0; i < size; i++) {
	*to++ = (unsigned char) *from;
	*to++ = (unsigned char) *(from + size);
	*to++ = (unsigned char) *(from + 2*size);
      }
    }
  }

  else {
    double *tmp, max, a;
    unsigned LONGINT j;
    unsigned char uc;

    tmp = new double[size];
    max = minTypeValue<double>(0);

    if(inr.getVectorialMode() == VM_INTERLACED) {
      for(i = 0; i < size; i++, tmp++) {
	*tmp = 0.0;
	for(j = 0; j < inr.getVdim(); j++, from++) *tmp += *from * *from;
	*tmp = sqrt(*tmp);
	if(*tmp > max) max = *tmp;
      }
    }
    else {
      FROM *fv;

      for(i = 0; i < size; i++, from++) {
	*tmp = 0.0;
	for(j = 0, fv = from; j < inr.getVdim(); j++, fv += size) *tmp += *fv * *fv;
	*tmp = sqrt(*tmp);
	if(*tmp > max) max = *tmp;
      }     
    }

    a = 255.0 / max;

    tmp -= size;
    for(i = 0; i < size; i++, tmp++) {
      uc = (unsigned char) (a * *tmp + 0.5);
      *to++ = uc;
      *to++ = uc;
      *to++ = uc;
    }
    tmp -= size;
    delete[] tmp;
  }

  to -= 3 * size;
  inr.setVdim(3);
  delete[] (unsigned char *) inr.setData(to);
  inr.setVectorialMode(VM_INTERLACED);
}

   template <class FROM>
    void convertFromVectorialToRGBABuffer(FROM a, Inrimage &inr){
   unsigned char *to;
  FROM *from;
  unsigned LONGINT i, size = inr.getX() * inr.getY() * inr.getZ();

  from = (FROM *) inr.getData();
  to = new unsigned char[size * 4];

  if(inr.getVdim() == 4) {
    if(inr.getVectorialMode() == VM_INTERLACED) {
#if (defined (_ALPHA_) || defined (_LINUX_))
      for(i = 0; i < size; i++, from += 4) {
	*to++ = (unsigned char) *(from+3);
	*to++ = (unsigned char) *(from+2);
	*to++ = (unsigned char) *(from+1);
	*to++ = (unsigned char) *from;
      }
#else
      for(i = 0; i < 4 * size; i++)
	*to++ = (unsigned char) *from++;
#endif
    }
    else {
#if (defined (_ALPHA_) || defined (_LINUX_))
      for(i = 0; i < size; i++, to += 4) {
	*to++ = (unsigned char) *(from + 3 * size);
	*to++ = (unsigned char) *(from + 2 * size);
	*to++ = (unsigned char) *(from + size);
	*to++ = (unsigned char) *from;
      }
#else
      for(i = 0; i < size; i++) {
	*to++ = (unsigned char) *from;
	*to++ = (unsigned char) *(from + size);
	*to++ = (unsigned char) *(from + 2 * size);
	*to++ = (unsigned char) *(from + 3 * size);
      }
#endif
    }
  }

  else {
    double *tmp, max, a;
    unsigned LONGINT j;
    unsigned char uc;

    tmp = new double[size];
    max = minTypeValue<double>(0);

    if(inr.getVectorialMode() == VM_INTERLACED) {
      for(i = 0; i < size; i++, tmp++) {
	*tmp = 0.0;
	for(j = 0; j < inr.getVdim(); j++, from++) *tmp += *from * *from;
	*tmp = sqrt(*tmp);
	if(*tmp > max) max = *tmp;
      }
    }
    else {
      FROM *fv;
      
      for(i = 0; i < size; i++, from++) {
	*tmp = 0.0;
	for(j = 0, fv = from; j < inr.getVdim(); j++, fv += size) *tmp += *fv * *fv;
	*tmp = sqrt(*tmp);
	if(*tmp > max) max = *tmp;
      }     
    }

    a = 255.0 / max;

    tmp -= size;
    for(i = 0; i < size; i++, tmp++) {
      uc = (unsigned char) (a * *tmp + 0.5);
#if (defined (_ALPHA_) || defined (_LINUX_))
      *to++ = 0xFF;
      *to++ = uc;
      *to++ = uc;
      *to++ = uc;
#else
      *to++ = uc;
      *to++ = uc;
      *to++ = uc;
      *to++ = 0xFF;
#endif
    }
    tmp -= size;
    delete[] tmp;
  }

  to -= 4 * size;
  inr.setVdim(4);
  delete[] (unsigned char *) inr.setData(to);
  inr.setVectorialMode(VM_INTERLACED);
}

  template <class FROM, class TO>
    void convertFromVectorialToVectorialBuffer(FROM a, TO b, Inrimage &inr){
  TO *to;
  FROM *from;
  unsigned LONGINT i, size = inr.getVdim() * inr.getX() * inr.getY() * inr.getZ();

  from = (FROM *) inr.getData();
  to = new TO[size];
  for(i = 0; i < size; i++, from++, to++) *to = (TO) *from;
  to -= size;

  delete[] (unsigned char *) inr.setData(to);
}

void Inrimage::convertFromRGBToRGBABuffer( ) {
  unsigned char *rgb, *rgba;
  unsigned LONGINT i, size;
  vdim = 4;

  size = xdim * ydim * zdim;
  rgb = (unsigned char *) data;
  rgba = new unsigned char[size * 4];

  for(i = 0; i < size; i++, rgb += 3, rgba += 4) {
    *rgba = *rgb;
    *(rgba+1) = *(rgb+1);
    *(rgba+2) = *(rgb+2);
    *(rgba+3) = 0xFF;
  }
  rgba -= 4 * size;

  delete[] (unsigned char *) setData(rgba);
}


void Inrimage::convertFromRGBAToRGBBuffer(void) {
  unsigned char *rgb, *rgba;
  unsigned LONGINT i, size;
  vdim = 3;
  
  size = xdim * ydim * zdim;
  rgba = (unsigned char *) data;
  rgb = new unsigned char[size * 3];

  for(i = 0; i < size; i++, rgb += 3, rgba += 4) {
    *rgb = *rgba;
    *(rgb+1) = *(rgba+1);
    *(rgb+2) = *(rgba+2);
  }
  rgb -= 3 * size;

  delete[] (unsigned char *) setData(rgb);
}


#define CONVERTFROMSCALARTOVECTORIALBUFFER(FROM,TO,INR) convertFromScalarToVectorialBuffer((FROM) 0,(TO) 0, INR)

#define CONVERTFROMRGBTOVECTORIALBUFFER(TO,INR) convertFromRGBToVectorialBuffer((TO) 0,INR)

#define CONVERTFROMRGBATOVECTORIALBUFFER(TO,INR) convertFromRGBAToVectorialBuffer((TO) 0,INR)

#define CONVERTCASTFROMVECTORIALTOSCALARBUFFER(FROM,TO,INR) convertCastFromVectorialToScalarBuffer((FROM) 0,(TO) 0,INR)

#define CONVERTSCALEFROMVECTORIALTOSCALARBUFFER(FROM,TO,INR) convertScaleFromVectorialToScalarBuffer((FROM) 0,(TO) 0,INR)

#define CONVERTFROMVECTORIALTORGBBUFFER(FROM,INR) convertFromVectorialToRGBBuffer((FROM) 0,INR)

#define CONVERTFROMVECTORIALTORGBABUFFER(FROM,INR) convertFromVectorialToRGBABuffer((FROM) 0,INR)

#define CONVERTFROMVECTORIALTOVECTORIALBUFFER(FROM,TO,INR) convertFromVectorialToVectorialBuffer((FROM) 0,(TO) 0,INR)


#define CONVERTSCALESCALARBUFFER(FROM,TO,INR,VMIN,VMAX) convertScaleScalarBuffer((FROM) 0,(TO) 0,INR,VMIN,VMAX)

#define SCALESCALARBUFFER(FROM,INR,VMIN,VMAX) scaleScalarBuffer((FROM) 0,INR,VMIN,VMAX)

#define CONVERTCASTSCALARBUFFER(FROM,TO,INR) convertCastScalarBuffer((FROM) 0,(TO) 0,INR)

#define CONVERTSCALETORGBBUFFER(FROM,INR,VMIN,VMAX) convertScaleToRGBBuffer((FROM) 0,INR,VMIN,VMAX)

#define CONVERTSCALETORGBABUFFER(FROM,INR,VMIN,VMAX) convertScaleToRGBABuffer((FROM) 0,INR,VMIN,VMAX)

#define CONVERTCASTTORGBBUFFER(FROM,INR) convertCastToRGBBuffer((FROM) 0,INR)


#define CONVERTCASTTORGBABUFFER(FROM,INR) convertCastToRGBABuffer((FROM) 0,INR)

#define CONVERTSCALEFROMRGBBUFFER(TO,INR) convertScaleFromRGBBuffer((TO) 0,INR)


#define CONVERTSCALEFROMRGBABUFFER(TO,INR) convertScaleFromRGBABuffer((TO) 0,INR)

#define CONVERTCASTFROMRGBBUFFER(TO,INR) convertCastFromRGBBuffer((TO) 0,INR)

#define CONVERTCASTFROMRGBABUFFER(TO,INR) convertCastFromRGBABuffer((TO) 0,INR)


void Inrimage::convertCast(WORD_TYPE ntype) {
  WORD_TYPE orgtype = type;
  type = ntype;

  switch(orgtype) {

  case WT_UNSIGNED_CHAR:
    switch(ntype) {
    case WT_SIGNED_CHAR:
      CONVERTCASTSCALARBUFFER(unsigned char, char, *this);
      break;
    case WT_SIGNED_SHORT:
      CONVERTCASTSCALARBUFFER(unsigned char, short int, *this);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTCASTSCALARBUFFER(unsigned char, unsigned short int, *this);
      break;
    case WT_UNSIGNED_INT:
      CONVERTCASTSCALARBUFFER(unsigned char, unsigned int, *this);
      break;
    case WT_SIGNED_INT:
      CONVERTCASTSCALARBUFFER(unsigned char, int, *this);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTCASTSCALARBUFFER(unsigned char, unsigned LONGINT, *this);
      break;
    case WT_SIGNED_LONG:
      CONVERTCASTSCALARBUFFER(unsigned char, LONGINT, *this);
      break;
    case WT_DOUBLE:
      CONVERTCASTSCALARBUFFER(unsigned char, double, *this);
      break;
    case WT_FLOAT:
      CONVERTCASTSCALARBUFFER(unsigned char, float, *this);
      break;
    case WT_RGB:
      CONVERTCASTTORGBBUFFER(unsigned char, *this);
      break;
    case WT_RGBA:
      CONVERTCASTTORGBABUFFER(unsigned char, *this);
     break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(unsigned char, float, *this);
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(unsigned char, double, *this);
      break;
    default:
      break;
    }
    break;

  case WT_SIGNED_CHAR:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTCASTSCALARBUFFER(char, unsigned char, *this);
      break;
    case WT_SIGNED_SHORT:
      CONVERTCASTSCALARBUFFER(char, short int, *this);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTCASTSCALARBUFFER(char, unsigned short int, *this);
      break;
    case WT_UNSIGNED_INT:
      CONVERTCASTSCALARBUFFER(char, unsigned int, *this);
      break;
    case WT_SIGNED_INT:
      CONVERTCASTSCALARBUFFER(char, int, *this);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTCASTSCALARBUFFER(char, unsigned LONGINT, *this);
      break;
    case WT_SIGNED_LONG:
      CONVERTCASTSCALARBUFFER(char, LONGINT, *this);
      break;
    case WT_DOUBLE:
      CONVERTCASTSCALARBUFFER(char, double, *this);
      break;
    case WT_FLOAT:
      CONVERTCASTSCALARBUFFER(char, float, *this);
      break;
    case WT_RGB:
      CONVERTCASTTORGBBUFFER(char, *this);
      break;
    case WT_RGBA:
      CONVERTCASTTORGBABUFFER(char, *this);
     break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(char, float, *this);
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(char, double, *this);
      break;
    default:
      break;
    }
    break;

  case WT_UNSIGNED_SHORT:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTCASTSCALARBUFFER(unsigned short int, unsigned char, *this);
      break;
    case WT_SIGNED_CHAR:
      CONVERTCASTSCALARBUFFER(unsigned short int, char, *this);
      break;
    case WT_SIGNED_SHORT:
      CONVERTCASTSCALARBUFFER(unsigned short int, short int, *this);
      break;
    case WT_UNSIGNED_INT:
      CONVERTCASTSCALARBUFFER(unsigned short int, unsigned int, *this);
      break;
    case WT_SIGNED_INT:
      CONVERTCASTSCALARBUFFER(unsigned short int, int, *this);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTCASTSCALARBUFFER(unsigned short int, unsigned LONGINT, *this);
      break;
    case WT_SIGNED_LONG:
      CONVERTCASTSCALARBUFFER(unsigned short int, LONGINT, *this);
      break;
    case WT_DOUBLE:
      CONVERTCASTSCALARBUFFER(unsigned short int, double, *this);
      break;
    case WT_FLOAT:
      CONVERTCASTSCALARBUFFER(unsigned short int, float, *this);
      break;
    case WT_RGB:
      CONVERTCASTTORGBBUFFER(unsigned short int, *this);
      break;
    case WT_RGBA:
      CONVERTCASTTORGBABUFFER(unsigned short int, *this);
      break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(unsigned short int, float, *this);
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(unsigned short int, double, *this);
      break;
    default:
      break;
    }
    break;

  case WT_SIGNED_SHORT:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTCASTSCALARBUFFER(short int, unsigned char, *this);
      break;
    case WT_SIGNED_CHAR:
      CONVERTCASTSCALARBUFFER(short int, char, *this);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTCASTSCALARBUFFER(short int, unsigned short int, *this);
      break;
    case WT_UNSIGNED_INT:
      CONVERTCASTSCALARBUFFER(short int, unsigned int, *this);
      break;
    case WT_SIGNED_INT:
      CONVERTCASTSCALARBUFFER(short int, int, *this);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTCASTSCALARBUFFER(short int, unsigned LONGINT, *this);
      break;
    case WT_SIGNED_LONG:
      CONVERTCASTSCALARBUFFER(short int, LONGINT, *this);
      break;
    case WT_DOUBLE:
      CONVERTCASTSCALARBUFFER(short int, double, *this);
      break;
    case WT_FLOAT:
      CONVERTCASTSCALARBUFFER(short int, float, *this);
      break;
    case WT_RGB:
      CONVERTCASTTORGBBUFFER(short int, *this);
      break;
    case WT_RGBA:
      CONVERTCASTTORGBABUFFER(short int, *this);
      break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(short int, float, *this);
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(short int, double, *this);
      break;
    default:
      break;
    }
    break;

  case WT_UNSIGNED_INT:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTCASTSCALARBUFFER(unsigned int, unsigned char, *this);
      break;
    case WT_SIGNED_CHAR:
      CONVERTCASTSCALARBUFFER(unsigned int, char, *this);
      break;
    case WT_SIGNED_SHORT:
      CONVERTCASTSCALARBUFFER(unsigned int, short int, *this);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTCASTSCALARBUFFER(unsigned int, unsigned short int, *this);
      break;
    case WT_SIGNED_INT:
      CONVERTCASTSCALARBUFFER(unsigned int, int, *this);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTCASTSCALARBUFFER(unsigned int, unsigned LONGINT, *this);
      break;
    case WT_SIGNED_LONG:
      CONVERTCASTSCALARBUFFER(unsigned int, LONGINT, *this);
      break;
    case WT_DOUBLE:
      CONVERTCASTSCALARBUFFER(unsigned int, double, *this);
      break;
    case WT_FLOAT:
      CONVERTCASTSCALARBUFFER(unsigned int, float, *this);
      break;
    case WT_RGB:
      CONVERTCASTTORGBBUFFER(unsigned int, *this);
      break;
    case WT_RGBA:
      CONVERTCASTTORGBABUFFER(unsigned int, *this);
      break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(unsigned int, float, *this);
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(unsigned int, double, *this);
      break;
    default:
      break;
    }
    break;

  case WT_SIGNED_INT:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTCASTSCALARBUFFER(int, unsigned char, *this);
      break;
    case WT_SIGNED_CHAR:
      CONVERTCASTSCALARBUFFER(int, char, *this);
      break;
    case WT_SIGNED_SHORT:
      CONVERTCASTSCALARBUFFER(int, short int, *this);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTCASTSCALARBUFFER(int, unsigned short int, *this);
      break;
    case WT_UNSIGNED_INT:
      CONVERTCASTSCALARBUFFER(int, unsigned int, *this);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTCASTSCALARBUFFER(int, unsigned LONGINT, *this);
      break;
    case WT_SIGNED_LONG:
      CONVERTCASTSCALARBUFFER(int, LONGINT, *this);
      break;
    case WT_DOUBLE:
      CONVERTCASTSCALARBUFFER(int, double, *this);
      break;
    case WT_FLOAT:
      CONVERTCASTSCALARBUFFER(int, float, *this);
      break;
    case WT_RGB:
      CONVERTCASTTORGBBUFFER(int, *this);
      break;
    case WT_RGBA:
      CONVERTCASTTORGBABUFFER(int, *this);
       break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(int, float, *this);
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(int, double, *this);
      break;
    default:
      break;
    }
    break;

  case WT_UNSIGNED_LONG:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTCASTSCALARBUFFER(unsigned LONGINT, unsigned char, *this);
      break;
    case WT_SIGNED_CHAR:
      CONVERTCASTSCALARBUFFER(unsigned LONGINT, char, *this);
      break;
    case WT_SIGNED_SHORT:
      CONVERTCASTSCALARBUFFER(unsigned LONGINT, short int, *this);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTCASTSCALARBUFFER(unsigned LONGINT, unsigned short int, *this);
      break;
    case WT_UNSIGNED_INT:
      CONVERTCASTSCALARBUFFER(unsigned LONGINT, unsigned int, *this);
      break;
    case WT_SIGNED_INT:
      CONVERTCASTSCALARBUFFER(unsigned LONGINT, int, *this);
      break;
    case WT_SIGNED_LONG:
      CONVERTCASTSCALARBUFFER(unsigned LONGINT, LONGINT, *this);
      break;
    case WT_DOUBLE:
      CONVERTCASTSCALARBUFFER(unsigned LONGINT, double, *this);
      break;
    case WT_FLOAT:
      CONVERTCASTSCALARBUFFER(unsigned LONGINT, float, *this);
      break;
    case WT_RGB:
      CONVERTCASTTORGBBUFFER(unsigned LONGINT, *this);
      break;
    case WT_RGBA:
      CONVERTCASTTORGBABUFFER(unsigned LONGINT, *this);
      break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(unsigned LONGINT, float, *this);
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(unsigned LONGINT, double, *this);
      break;
    default:
      break;
    }
    break;

  case WT_SIGNED_LONG:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTCASTSCALARBUFFER(LONGINT, unsigned char, *this);
      break;
    case WT_SIGNED_CHAR:
      CONVERTCASTSCALARBUFFER(LONGINT, char, *this);
      break;
    case WT_SIGNED_SHORT:
      CONVERTCASTSCALARBUFFER(LONGINT, short int, *this);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTCASTSCALARBUFFER(LONGINT, unsigned short int, *this);
      break;
    case WT_UNSIGNED_INT:
      CONVERTCASTSCALARBUFFER(LONGINT, unsigned int, *this);
      break;
    case WT_SIGNED_INT:
      CONVERTCASTSCALARBUFFER(LONGINT, int, *this);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTCASTSCALARBUFFER(LONGINT, unsigned LONGINT, *this);
      break;
    case WT_DOUBLE:
      CONVERTCASTSCALARBUFFER(LONGINT, double, *this);
      break;
    case WT_FLOAT:
      CONVERTCASTSCALARBUFFER(LONGINT, float, *this);
      break;
    case WT_RGB:
      CONVERTCASTTORGBBUFFER(LONGINT, *this);
      break;
    case WT_RGBA:
      CONVERTCASTTORGBABUFFER(LONGINT, *this);
      break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(LONGINT, float, *this);
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(LONGINT, double, *this);
      break;
    default:
      break;
    }
    break;

  case WT_FLOAT:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTCASTSCALARBUFFER(float, unsigned char, *this);
      break;
    case WT_SIGNED_CHAR:
      CONVERTCASTSCALARBUFFER(float, char, *this);
      break;
    case WT_SIGNED_SHORT:
      CONVERTCASTSCALARBUFFER(float, short int, *this);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTCASTSCALARBUFFER(float, unsigned short int, *this);
      break;
    case WT_UNSIGNED_INT:
      CONVERTCASTSCALARBUFFER(float, unsigned int, *this);
      break;
    case WT_SIGNED_INT:
      CONVERTCASTSCALARBUFFER(float, int, *this);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTCASTSCALARBUFFER(float, unsigned LONGINT, *this);
      break;
    case WT_SIGNED_LONG:
      CONVERTCASTSCALARBUFFER(float, LONGINT, *this);
      break;
    case WT_DOUBLE:
      CONVERTCASTSCALARBUFFER(float, double, *this);
      break;
    case WT_RGB:
      CONVERTCASTTORGBBUFFER(float, *this);
      break;
    case WT_RGBA:
      CONVERTCASTTORGBABUFFER(float, *this);
      break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(float, float, *this);
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(float, double, *this);
      break;
    default:
      break;
    }
    break;

  case WT_DOUBLE:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTCASTSCALARBUFFER(double, unsigned char, *this);
      break;
    case WT_SIGNED_CHAR:
      CONVERTCASTSCALARBUFFER(double, char, *this);
      break;
    case WT_SIGNED_SHORT:
      CONVERTCASTSCALARBUFFER(double, short int, *this);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTCASTSCALARBUFFER(double, unsigned short int, *this);
      break;
    case WT_UNSIGNED_INT:
      CONVERTCASTSCALARBUFFER(double, unsigned int, *this);
      break;
    case WT_SIGNED_INT:
      CONVERTCASTSCALARBUFFER(double, int, *this);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTCASTSCALARBUFFER(double, unsigned LONGINT, *this);
      break;
    case WT_SIGNED_LONG:
      CONVERTCASTSCALARBUFFER(double, LONGINT, *this);
      break;
    case WT_FLOAT:
      CONVERTCASTSCALARBUFFER(double, float, *this);
      break;
    case WT_RGB:
      CONVERTCASTTORGBBUFFER(double, *this);
      break;
    case WT_RGBA:
      CONVERTCASTTORGBABUFFER(double, *this);
      break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(double, float, *this);
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(double, double, *this);
      break;
    default:
      break;
    }
    break;

  case WT_RGB:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTCASTFROMRGBBUFFER(unsigned char, *this);
      break;
    case WT_SIGNED_CHAR:
      CONVERTCASTFROMRGBBUFFER(char, *this);
      break;
    case WT_SIGNED_SHORT:
      CONVERTCASTFROMRGBBUFFER(short int, *this);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTCASTFROMRGBBUFFER(unsigned short int, *this);
      break;
    case WT_UNSIGNED_INT:
      CONVERTCASTFROMRGBBUFFER(unsigned int, *this);
      break;
    case WT_SIGNED_INT:
      CONVERTCASTFROMRGBBUFFER(int, *this);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTCASTFROMRGBBUFFER(unsigned LONGINT, *this);
      break;
    case WT_SIGNED_LONG:
      CONVERTCASTFROMRGBBUFFER(LONGINT, *this);
      break;
    case WT_FLOAT:
      CONVERTCASTFROMRGBBUFFER(float, *this);
      break;
    case WT_DOUBLE:
      CONVERTCASTFROMRGBBUFFER(double, *this);
      break;
    case WT_RGBA:
      convertFromRGBToRGBABuffer();
      break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMRGBTOVECTORIALBUFFER(float, *this);
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMRGBTOVECTORIALBUFFER(double, *this);
      break;
    default:
      break;
    }
    break;

  case WT_RGBA:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTCASTFROMRGBABUFFER(unsigned char, *this);
      break;
    case WT_SIGNED_CHAR:
      CONVERTCASTFROMRGBABUFFER(char, *this);
      break;
    case WT_SIGNED_SHORT:
      CONVERTCASTFROMRGBABUFFER(short int, *this);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTCASTFROMRGBABUFFER(unsigned short int, *this);
      break;
    case WT_UNSIGNED_INT:
      CONVERTCASTFROMRGBABUFFER(unsigned int, *this);
      break;
    case WT_SIGNED_INT:
      CONVERTCASTFROMRGBABUFFER(int, *this);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTCASTFROMRGBABUFFER(unsigned LONGINT, *this);
      break;
    case WT_SIGNED_LONG:
      CONVERTCASTFROMRGBABUFFER(LONGINT, *this);
      break;
    case WT_FLOAT:
      CONVERTCASTFROMRGBABUFFER(float, *this);
      break;
    case WT_DOUBLE:
      CONVERTCASTFROMRGBABUFFER(double, *this);
      break;
    case WT_RGB:
      convertFromRGBAToRGBBuffer();
      break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMRGBATOVECTORIALBUFFER(float, *this);
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMRGBATOVECTORIALBUFFER(double, *this);
      break;
    default:
      break;
    }
    break;

  case WT_FLOAT_VECTOR:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTCASTFROMVECTORIALTOSCALARBUFFER(float, unsigned char, *this);
      break;
    case WT_SIGNED_CHAR:
      CONVERTCASTFROMVECTORIALTOSCALARBUFFER(float, char, *this);
      break;
    case WT_SIGNED_SHORT:
      CONVERTCASTFROMVECTORIALTOSCALARBUFFER(float, short int, *this);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTCASTFROMVECTORIALTOSCALARBUFFER(float, unsigned short int, *this);
      break;
    case WT_UNSIGNED_INT:
      CONVERTCASTFROMVECTORIALTOSCALARBUFFER(float, unsigned int, *this);
      break;
    case WT_SIGNED_INT:
      CONVERTCASTFROMVECTORIALTOSCALARBUFFER(float, int, *this);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTCASTFROMVECTORIALTOSCALARBUFFER(float, unsigned LONGINT, *this);
      break;
    case WT_SIGNED_LONG:
      CONVERTCASTFROMVECTORIALTOSCALARBUFFER(float, LONGINT, *this);
      break;
    case WT_FLOAT:
      CONVERTCASTFROMVECTORIALTOSCALARBUFFER(float, float, *this);
      break;
    case WT_DOUBLE:
      CONVERTCASTFROMVECTORIALTOSCALARBUFFER(float, double, *this);
      break;
    case WT_RGB:
      CONVERTFROMVECTORIALTORGBBUFFER(float, *this);
      break;
    case WT_RGBA:
      CONVERTFROMVECTORIALTORGBABUFFER(float, *this);
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMVECTORIALTOVECTORIALBUFFER(float, double, *this);
      break;
    default:
      break;
    }
    break;

  case WT_DOUBLE_VECTOR:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTCASTFROMVECTORIALTOSCALARBUFFER(double, unsigned char, *this);
      break;
    case WT_SIGNED_CHAR:
      CONVERTCASTFROMVECTORIALTOSCALARBUFFER(double, char, *this);
      break;
    case WT_SIGNED_SHORT:
      CONVERTCASTFROMVECTORIALTOSCALARBUFFER(double, short int, *this);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTCASTFROMVECTORIALTOSCALARBUFFER(double, unsigned short int, *this);
      break;
    case WT_UNSIGNED_INT:
      CONVERTCASTFROMVECTORIALTOSCALARBUFFER(double, unsigned int, *this);
      break;
    case WT_SIGNED_INT:
      CONVERTCASTFROMVECTORIALTOSCALARBUFFER(double, int, *this);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTCASTFROMVECTORIALTOSCALARBUFFER(double, unsigned LONGINT, *this);
      break;
    case WT_SIGNED_LONG:
      CONVERTCASTFROMVECTORIALTOSCALARBUFFER(double, LONGINT, *this);
      break;
    case WT_FLOAT:
      CONVERTCASTFROMVECTORIALTOSCALARBUFFER(double, float, *this);
      break;
    case WT_DOUBLE:
      CONVERTCASTFROMVECTORIALTOSCALARBUFFER(double, double, *this);
      break;
    case WT_RGB:
      CONVERTFROMVECTORIALTORGBBUFFER(double, *this);
      break;
    case WT_RGBA:
      CONVERTFROMVECTORIALTORGBABUFFER(double, *this);
      break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMVECTORIALTOVECTORIALBUFFER(double, float, *this);
      break;
    default:
      break;
    }
    break;

  default:
    break;
  }

  setType(type, vdim, vectMode, false);
}



void Inrimage::convertScale(WORD_TYPE ntype) {
  Voxel *imin, *imax;
  WORD_TYPE orgtype = type;

  imin = Voxel::voxelByType(type);
  imax = Voxel::voxelByType(type);
  getBounds(imin, imax);



  if ( ntype == orgtype ) {
    
    //
    // input and output types are identical
    // checks if scaling is really necessary
    //

    switch(orgtype) {

    case WT_UNSIGNED_CHAR:
      if ( (dynamic_cast<Voxel8u *>(imin)->getValue() > minTypeValue<unsigned char>(0) ||
	    dynamic_cast<Voxel8u *>(imax)->getValue() < maxTypeValue<unsigned char>(0)) &&
	   dynamic_cast<Voxel8u *>(imin)->getValue() != dynamic_cast<Voxel8u *>(imax)->getValue() )
	SCALESCALARBUFFER(unsigned char, *this, imin, imax);
      break;
    case WT_SIGNED_CHAR:
      if ( (dynamic_cast<Voxel8s *>(imin)->getValue() > minTypeValue<char>(0) ||
	    dynamic_cast<Voxel8s *>(imax)->getValue() < maxTypeValue<char>(0)) &&
	   dynamic_cast<Voxel8s *>(imin)->getValue() != dynamic_cast<Voxel8s *>(imax)->getValue() )
	SCALESCALARBUFFER(char,          *this, imin, imax);
      break;
    case WT_SIGNED_SHORT:
      if ( (dynamic_cast<Voxel16s *>(imin)->getValue() > minTypeValue<short int>(0) ||
	    dynamic_cast<Voxel16s *>(imax)->getValue() < maxTypeValue<short int>(0)) &&
	   dynamic_cast<Voxel16s *>(imin)->getValue() != dynamic_cast<Voxel16s *>(imax)->getValue() )
	SCALESCALARBUFFER(short int,     *this, imin, imax);
      break;
    case WT_UNSIGNED_SHORT:
      if ( (dynamic_cast<Voxel16u *>(imin)->getValue() > minTypeValue<unsigned short int>(0) ||
	    dynamic_cast<Voxel16u *>(imax)->getValue() < maxTypeValue<unsigned short int>(0)) &&
	   dynamic_cast<Voxel16u *>(imin)->getValue() != dynamic_cast<Voxel16u *>(imax)->getValue() )
	SCALESCALARBUFFER(unsigned short int, *this, imin, imax);
      break;
    case WT_UNSIGNED_INT:
      if ( (dynamic_cast<Voxel32u *>(imin)->getValue() > minTypeValue<unsigned int>(0) ||
	    dynamic_cast<Voxel32u *>(imax)->getValue() < maxTypeValue<unsigned int>(0)) &&
	   dynamic_cast<Voxel32u *>(imin)->getValue() != dynamic_cast<Voxel32u *>(imax)->getValue() )
	SCALESCALARBUFFER(unsigned int,  *this, imin, imax);
      break;
    case WT_SIGNED_INT:
      if ( (dynamic_cast<Voxel32s *>(imin)->getValue() > minTypeValue<int>(0) ||
	    dynamic_cast<Voxel32s *>(imax)->getValue() < maxTypeValue<int>(0)) &&
	   dynamic_cast<Voxel32s *>(imin)->getValue() != dynamic_cast<Voxel32s *>(imax)->getValue() )
	SCALESCALARBUFFER(int,           *this, imin, imax);
      break;
    case WT_UNSIGNED_LONG:
      if ( (dynamic_cast<Voxel64u *>(imin)->getValue() > minTypeValue<unsigned LONGINT>(0) ||
	    dynamic_cast<Voxel64u *>(imax)->getValue() < maxTypeValue<unsigned LONGINT>(0)) &&
	   dynamic_cast<Voxel64u *>(imin)->getValue() != dynamic_cast<Voxel64u *>(imax)->getValue() )
	SCALESCALARBUFFER(unsigned LONGINT, *this, imin, imax);
      break;
    case WT_SIGNED_LONG:
      if ( (dynamic_cast<Voxel64s *>(imin)->getValue() > minTypeValue<LONGINT>(0) ||
	    dynamic_cast<Voxel64s *>(imax)->getValue() < maxTypeValue<LONGINT>(0)) &&
	   dynamic_cast<Voxel64s *>(imin)->getValue() != dynamic_cast<Voxel64s *>(imax)->getValue() )
	SCALESCALARBUFFER(LONGINT,       *this, imin, imax);
      break;
    case WT_DOUBLE:
      if ( (dynamic_cast<VoxelDouble *>(imin)->getValue() > minTypeValue<double>(0) ||
	    dynamic_cast<VoxelDouble *>(imax)->getValue() < maxTypeValue<double>(0)) &&
	   dynamic_cast<VoxelDouble *>(imin)->getValue() != dynamic_cast<VoxelDouble *>(imax)->getValue() )
	SCALESCALARBUFFER(double,        *this, imin, imax);
      break;
    case WT_FLOAT:
      if ( (dynamic_cast<VoxelFloat *>(imin)->getValue() > minTypeValue<float>(0) ||
	    dynamic_cast<VoxelFloat *>(imax)->getValue() < maxTypeValue<float>(0))  &&
	   dynamic_cast<VoxelFloat *>(imin)->getValue() != dynamic_cast<VoxelFloat *>(imax)->getValue() )
	SCALESCALARBUFFER(float,         *this, imin, imax);
      break;
    case WT_RGB:
    case WT_RGBA:
    case WT_FLOAT_VECTOR:
    case WT_DOUBLE_VECTOR:
      //
      // we should reject an exception somewhere here
      //
      break;
    default:
      break;
    }
 
    return;
  }




  type = ntype;

  switch(orgtype) {

  case WT_UNSIGNED_CHAR:
    switch(ntype) {
    case WT_SIGNED_CHAR:
      CONVERTSCALESCALARBUFFER(unsigned char, char,*this, imin, imax);
      break;
    case WT_SIGNED_SHORT:
      CONVERTSCALESCALARBUFFER(unsigned char, short int,*this, imin, imax);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTSCALESCALARBUFFER(unsigned char, unsigned short int,*this, imin, imax);
      break;
    case WT_UNSIGNED_INT:
      CONVERTSCALESCALARBUFFER(unsigned char, unsigned int,*this, imin, imax);
      break;
    case WT_SIGNED_INT:
      CONVERTSCALESCALARBUFFER(unsigned char, int,*this, imin, imax);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTSCALESCALARBUFFER(unsigned char, unsigned LONGINT,*this, imin, imax);
      break;
    case WT_SIGNED_LONG:
      CONVERTSCALESCALARBUFFER(unsigned char, LONGINT,*this, imin, imax);
      break;
    case WT_DOUBLE:
      CONVERTSCALESCALARBUFFER(unsigned char, double,*this, imin, imax);
      break;
    case WT_FLOAT:
      CONVERTSCALESCALARBUFFER(unsigned char, float,*this, imin, imax);
      break;
    case WT_RGB:
      CONVERTSCALETORGBBUFFER(unsigned char,*this, imin, imax);
      break;
    case WT_RGBA:
      CONVERTSCALETORGBABUFFER(unsigned char,*this, imin, imax);
     break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(unsigned char, float, *this );
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(unsigned char, double, *this );
      break;
    default:
      break;
    }
    break;

  case WT_SIGNED_CHAR:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTSCALESCALARBUFFER(char, unsigned char,*this, imin, imax);
      break;
    case WT_SIGNED_SHORT:
      CONVERTSCALESCALARBUFFER(char, short int,*this, imin, imax);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTSCALESCALARBUFFER(char, unsigned short int,*this, imin, imax);
      break;
    case WT_UNSIGNED_INT:
      CONVERTSCALESCALARBUFFER(char, unsigned int,*this, imin, imax);
      break;
    case WT_SIGNED_INT:
      CONVERTSCALESCALARBUFFER(char, int,*this, imin, imax);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTSCALESCALARBUFFER(char, unsigned LONGINT,*this, imin, imax);
      break;
    case WT_SIGNED_LONG:
      CONVERTSCALESCALARBUFFER(char, LONGINT,*this, imin, imax);
      break;
    case WT_DOUBLE:
      CONVERTSCALESCALARBUFFER(char, double,*this, imin, imax);
      break;
    case WT_FLOAT:
      CONVERTSCALESCALARBUFFER(char, float,*this, imin, imax);
      break;
    case WT_RGB:
      CONVERTSCALETORGBBUFFER(char,*this, imin, imax);
      break;
    case WT_RGBA:
      CONVERTSCALETORGBABUFFER(char,*this, imin, imax);
     break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(char, float, *this );
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(char, double, *this );
      break;
    default:
      break;
    }
    break;

  case WT_UNSIGNED_SHORT:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTSCALESCALARBUFFER(unsigned short int, unsigned char,*this, imin, imax);
      break;
    case WT_SIGNED_CHAR:
      CONVERTSCALESCALARBUFFER(unsigned short int, char,*this, imin, imax);
      break;
    case WT_SIGNED_SHORT:
      CONVERTSCALESCALARBUFFER(unsigned short int, short int,*this, imin, imax);
      break;
    case WT_UNSIGNED_INT:
      CONVERTSCALESCALARBUFFER(unsigned short int, unsigned int,*this, imin, imax);
      break;
    case WT_SIGNED_INT:
      CONVERTSCALESCALARBUFFER(unsigned short int, int,*this, imin, imax);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTSCALESCALARBUFFER(unsigned short int, unsigned LONGINT,*this, imin, imax);
      break;
    case WT_SIGNED_LONG:
      CONVERTSCALESCALARBUFFER(unsigned short int, LONGINT,*this, imin, imax);
      break;
    case WT_DOUBLE:
      CONVERTSCALESCALARBUFFER(unsigned short int, double,*this, imin, imax);
      break;
    case WT_FLOAT:
      CONVERTSCALESCALARBUFFER(unsigned short int, float,*this, imin, imax);
      break;
    case WT_RGB:
      CONVERTSCALETORGBBUFFER(unsigned short int,*this, imin, imax);
      break;
    case WT_RGBA:
      CONVERTSCALETORGBABUFFER(unsigned short int,*this, imin, imax);
      break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(unsigned short int, float,*this);
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(unsigned short int, double,*this);
      break;
    default:
      break;
    }
    break;

  case WT_SIGNED_SHORT:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTSCALESCALARBUFFER(short int, unsigned char,*this, imin, imax);
      break;
    case WT_SIGNED_CHAR:
      CONVERTSCALESCALARBUFFER(short int, char,*this, imin, imax);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTSCALESCALARBUFFER(short int, unsigned short int,*this, imin, imax);
      break;
    case WT_UNSIGNED_INT:
      CONVERTSCALESCALARBUFFER(short int, unsigned int,*this, imin, imax);
      break;
    case WT_SIGNED_INT:
      CONVERTSCALESCALARBUFFER(short int, int,*this, imin, imax);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTSCALESCALARBUFFER(short int, unsigned LONGINT,*this, imin, imax);
      break;
    case WT_SIGNED_LONG:
      CONVERTSCALESCALARBUFFER(short int, LONGINT,*this, imin, imax);
      break;
    case WT_DOUBLE:
      CONVERTSCALESCALARBUFFER(short int, double,*this, imin, imax);
      break;
    case WT_FLOAT:
      CONVERTSCALESCALARBUFFER(short int, float,*this, imin, imax);
      break;
    case WT_RGB:
      CONVERTSCALETORGBBUFFER(short int,*this,imin, imax);
      break;
    case WT_RGBA:
      CONVERTSCALETORGBABUFFER(short int,*this,imin, imax);
      break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(short int, float,*this);
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(short int, double,*this);
      break;
    default:
      break;
    }
    break;

  case WT_UNSIGNED_INT:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTSCALESCALARBUFFER(unsigned int, unsigned char,*this, imin, imax);
      break;
    case WT_SIGNED_CHAR:
      CONVERTSCALESCALARBUFFER(unsigned int, char,*this, imin, imax);
      break;
    case WT_SIGNED_SHORT:
      CONVERTSCALESCALARBUFFER(unsigned int, short int,*this, imin, imax);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTSCALESCALARBUFFER(unsigned int, unsigned short int,*this, imin, imax);
      break;
    case WT_SIGNED_INT:
      CONVERTSCALESCALARBUFFER(unsigned int, int,*this, imin, imax);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTSCALESCALARBUFFER(unsigned int, unsigned LONGINT,*this, imin, imax);
      break;
    case WT_SIGNED_LONG:
      CONVERTSCALESCALARBUFFER(unsigned int, LONGINT,*this, imin, imax);
      break;
    case WT_DOUBLE:
      CONVERTSCALESCALARBUFFER(unsigned int, double,*this, imin, imax);
      break;
    case WT_FLOAT:
      CONVERTSCALESCALARBUFFER(unsigned int, float,*this, imin, imax);
      break;
    case WT_RGB:
      CONVERTSCALETORGBBUFFER(unsigned int,*this, imin, imax);
      break;
    case WT_RGBA:
      CONVERTSCALETORGBABUFFER(unsigned int,*this,imin, imax);
      break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(unsigned int, float,*this);
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(unsigned int, double,*this);
      break;
    default:
      break;
    }
    break;

  case WT_SIGNED_INT:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTSCALESCALARBUFFER(int, unsigned char,*this, imin, imax);
      break;
    case WT_SIGNED_CHAR:
      CONVERTSCALESCALARBUFFER(int, char,*this, imin, imax);
      break;
    case WT_SIGNED_SHORT:
      CONVERTSCALESCALARBUFFER(int, short int,*this, imin, imax);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTSCALESCALARBUFFER(int, unsigned short int,*this, imin, imax);
      break;
    case WT_UNSIGNED_INT:
      CONVERTSCALESCALARBUFFER(int, unsigned int,*this, imin, imax);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTSCALESCALARBUFFER(int, unsigned LONGINT,*this, imin, imax);
      break;
    case WT_SIGNED_LONG:
      CONVERTSCALESCALARBUFFER(int, LONGINT,*this, imin, imax);
      break;
    case WT_DOUBLE:
      CONVERTSCALESCALARBUFFER(int, double,*this, imin, imax);
      break;
    case WT_FLOAT:
      CONVERTSCALESCALARBUFFER(int, float,*this, imin, imax);
      break;
    case WT_RGB:
      CONVERTSCALETORGBBUFFER(int,*this,imin, imax);
      break;
    case WT_RGBA:
      CONVERTSCALETORGBABUFFER(int,*this,imin, imax);
       break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(int, float,*this);
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(int, double,*this);
      break;
    default:
      break;
    }
    break;

  case WT_UNSIGNED_LONG:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTSCALESCALARBUFFER(unsigned LONGINT, unsigned char,*this, imin, imax);
      break;
    case WT_SIGNED_CHAR:
      CONVERTSCALESCALARBUFFER(unsigned LONGINT, char,*this, imin, imax);
      break;
    case WT_SIGNED_SHORT:
      CONVERTSCALESCALARBUFFER(unsigned LONGINT, short int,*this, imin, imax);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTSCALESCALARBUFFER(unsigned LONGINT, unsigned short int,*this, imin, imax);
      break;
    case WT_UNSIGNED_INT:
      CONVERTSCALESCALARBUFFER(unsigned LONGINT, unsigned int,*this, imin, imax);
      break;
    case WT_SIGNED_INT:
      CONVERTSCALESCALARBUFFER(unsigned LONGINT, int,*this, imin, imax);
      break;
    case WT_SIGNED_LONG:
      CONVERTSCALESCALARBUFFER(unsigned LONGINT, LONGINT,*this, imin, imax);
      break;
    case WT_DOUBLE:
      CONVERTSCALESCALARBUFFER(unsigned LONGINT, double,*this, imin, imax);
      break;
    case WT_FLOAT:
      CONVERTSCALESCALARBUFFER(unsigned LONGINT, float,*this, imin, imax);
      break;
    case WT_RGB:
      CONVERTSCALETORGBBUFFER(unsigned LONGINT,*this,imin, imax);
      break;
    case WT_RGBA:
      CONVERTSCALETORGBABUFFER(unsigned LONGINT,*this,imin, imax);
      break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(unsigned LONGINT, float,*this);
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(unsigned LONGINT, double,*this);
      break;
    default:
      break;
    }
    break;

  case WT_SIGNED_LONG:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTSCALESCALARBUFFER(LONGINT, unsigned char,*this, imin, imax);
      break;
    case WT_SIGNED_CHAR:
      CONVERTSCALESCALARBUFFER(LONGINT, char,*this, imin, imax);
      break;
    case WT_SIGNED_SHORT:
      CONVERTSCALESCALARBUFFER(LONGINT, short int,*this, imin, imax);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTSCALESCALARBUFFER(LONGINT, unsigned short int,*this, imin, imax);
      break;
    case WT_UNSIGNED_INT:
      CONVERTSCALESCALARBUFFER(LONGINT, unsigned int,*this, imin, imax);
      break;
    case WT_SIGNED_INT:
      CONVERTSCALESCALARBUFFER(LONGINT, int,*this, imin, imax);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTSCALESCALARBUFFER(LONGINT, unsigned LONGINT,*this, imin, imax);
      break;
    case WT_DOUBLE:
      CONVERTSCALESCALARBUFFER(LONGINT, double,*this, imin, imax);
      break;
    case WT_FLOAT:
      CONVERTSCALESCALARBUFFER(LONGINT, float,*this, imin, imax);
      break;
    case WT_RGB:
      CONVERTSCALETORGBBUFFER(LONGINT,*this,imin, imax);
      break;
    case WT_RGBA:
      CONVERTSCALETORGBABUFFER(LONGINT,*this,imin, imax);
      break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(LONGINT,float, *this );
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(LONGINT, double,*this);
      break;
    default:
      break;
    }
    break;

  case WT_FLOAT:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTSCALESCALARBUFFER(float, unsigned char,*this, imin, imax);
      break;
    case WT_SIGNED_CHAR:
      CONVERTSCALESCALARBUFFER(float, char,*this, imin, imax);
      break;
    case WT_SIGNED_SHORT:
      CONVERTSCALESCALARBUFFER(float, short int,*this, imin, imax);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTSCALESCALARBUFFER(float, unsigned short int,*this, imin, imax);
      break;
    case WT_UNSIGNED_INT:
      CONVERTSCALESCALARBUFFER(float, unsigned int,*this, imin, imax);
      break;
    case WT_SIGNED_INT:
      CONVERTSCALESCALARBUFFER(float, int,*this, imin, imax);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTSCALESCALARBUFFER(float, unsigned LONGINT,*this, imin, imax);
      break;
    case WT_SIGNED_LONG:
      CONVERTSCALESCALARBUFFER(float, LONGINT,*this, imin, imax);
      break;
    case WT_DOUBLE:
      CONVERTSCALESCALARBUFFER(float, double,*this, imin, imax);
      break;
    case WT_RGB:
      CONVERTSCALETORGBBUFFER(float,*this,imin, imax);
      break;
    case WT_RGBA:
      CONVERTSCALETORGBABUFFER(float,*this,imin, imax);
      break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(float, float,*this);
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(float, double,*this);
      break;
    default:
      break;
    }
    break;

  case WT_DOUBLE:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTSCALESCALARBUFFER(double, unsigned char,*this, imin, imax);
      break;
    case WT_SIGNED_CHAR:
      CONVERTSCALESCALARBUFFER(double, char,*this, imin, imax);
      break;
    case WT_SIGNED_SHORT:
      CONVERTSCALESCALARBUFFER(double, short int,*this, imin, imax);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTSCALESCALARBUFFER(double, unsigned short int,*this, imin, imax);
      break;
    case WT_UNSIGNED_INT:
      CONVERTSCALESCALARBUFFER(double, unsigned int,*this, imin, imax);
      break;
    case WT_SIGNED_INT:
      CONVERTSCALESCALARBUFFER(double, int,*this, imin, imax);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTSCALESCALARBUFFER(double, unsigned LONGINT,*this, imin, imax);
      break;
    case WT_SIGNED_LONG:
      CONVERTSCALESCALARBUFFER(double, LONGINT,*this, imin, imax);
      break;
    case WT_FLOAT:
      CONVERTSCALESCALARBUFFER(double, float,*this, imin, imax);
      break;
    case WT_RGB:
      CONVERTSCALETORGBBUFFER(double,*this,imin, imax);
      break;
    case WT_RGBA:
      CONVERTSCALETORGBABUFFER(double,*this,imin, imax);
      break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(double, float,*this);
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMSCALARTOVECTORIALBUFFER(double, double,*this);
      break;
    default:
      break;
    }
    break;

  case WT_RGB:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTSCALEFROMRGBBUFFER(unsigned char,*this);
      break;
    case WT_SIGNED_CHAR:
      CONVERTSCALEFROMRGBBUFFER(char,*this);
      break;
    case WT_SIGNED_SHORT:
      CONVERTSCALEFROMRGBBUFFER(short int,*this);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTSCALEFROMRGBBUFFER(unsigned short int,*this);
      break;
    case WT_UNSIGNED_INT:
      CONVERTSCALEFROMRGBBUFFER(unsigned int,*this);
      break;
    case WT_SIGNED_INT:
      CONVERTSCALEFROMRGBBUFFER(int,*this);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTSCALEFROMRGBBUFFER(unsigned LONGINT,*this);
      break;
    case WT_SIGNED_LONG:
      CONVERTSCALEFROMRGBBUFFER(LONGINT,*this);
      break;
    case WT_FLOAT:
      CONVERTSCALEFROMRGBBUFFER(float,*this);
      break;
    case WT_DOUBLE:
      CONVERTSCALEFROMRGBBUFFER(double,*this);
      break;
    case WT_RGBA:
      convertFromRGBToRGBABuffer();
      break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMRGBTOVECTORIALBUFFER(float,*this);
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMRGBTOVECTORIALBUFFER(double,*this);
      break;
    default:
      break;
    }
    break;

  case WT_RGBA:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTSCALEFROMRGBABUFFER(unsigned char,*this);
      break;
    case WT_SIGNED_CHAR:
      CONVERTSCALEFROMRGBABUFFER(char,*this);
      break;
    case WT_SIGNED_SHORT:
      CONVERTSCALEFROMRGBABUFFER(short int,*this);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTSCALEFROMRGBABUFFER(unsigned short int,*this);
      break;
    case WT_UNSIGNED_INT:
      CONVERTSCALEFROMRGBABUFFER(unsigned int,*this);
      break;
    case WT_SIGNED_INT:
      CONVERTSCALEFROMRGBABUFFER(int,*this);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTSCALEFROMRGBABUFFER(unsigned LONGINT,*this);
      break;
    case WT_SIGNED_LONG:
      CONVERTSCALEFROMRGBABUFFER(LONGINT,*this);
      break;
    case WT_FLOAT:
      CONVERTSCALEFROMRGBABUFFER(float,*this);
      break;
    case WT_DOUBLE:
      CONVERTSCALEFROMRGBABUFFER(double,*this);
      break;
    case WT_RGB:
      convertFromRGBAToRGBBuffer();
      break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMRGBATOVECTORIALBUFFER(float,*this);
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMRGBATOVECTORIALBUFFER(double,*this);
      break;
    default:
      break;
    }
    break;

  case WT_FLOAT_VECTOR:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTSCALEFROMVECTORIALTOSCALARBUFFER(float, unsigned char,*this);
      break;
    case WT_SIGNED_CHAR:
      CONVERTSCALEFROMVECTORIALTOSCALARBUFFER(float, char,*this);
      break;
    case WT_SIGNED_SHORT:
      CONVERTSCALEFROMVECTORIALTOSCALARBUFFER(float, short int,*this);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTSCALEFROMVECTORIALTOSCALARBUFFER(float, unsigned short int,*this);
      break;
    case WT_UNSIGNED_INT:
      CONVERTSCALEFROMVECTORIALTOSCALARBUFFER(float, unsigned int,*this);
      break;
    case WT_SIGNED_INT:
      CONVERTSCALEFROMVECTORIALTOSCALARBUFFER(float, int,*this);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTSCALEFROMVECTORIALTOSCALARBUFFER(float, unsigned LONGINT,*this);
      break;
    case WT_SIGNED_LONG:
      CONVERTSCALEFROMVECTORIALTOSCALARBUFFER(float, LONGINT,*this);
      break;
    case WT_FLOAT:
      CONVERTSCALEFROMVECTORIALTOSCALARBUFFER(float, float,*this);
      break;
    case WT_DOUBLE:
      CONVERTSCALEFROMVECTORIALTOSCALARBUFFER(float, double,*this);
      break;
    case WT_RGB:
      CONVERTFROMVECTORIALTORGBBUFFER(float,*this);
      break;
    case WT_RGBA:
      CONVERTFROMVECTORIALTORGBABUFFER(float,*this);
      break;
    case WT_DOUBLE_VECTOR:
      CONVERTFROMVECTORIALTOVECTORIALBUFFER(float, double,*this);
      break;
    default:
      break;
    }
    break;

  case WT_DOUBLE_VECTOR:
    switch(ntype) {
    case WT_UNSIGNED_CHAR:
      CONVERTSCALEFROMVECTORIALTOSCALARBUFFER(double, unsigned char,*this);
      break;
    case WT_SIGNED_CHAR:
      CONVERTSCALEFROMVECTORIALTOSCALARBUFFER(double, char,*this);
      break;
    case WT_SIGNED_SHORT:
      CONVERTSCALEFROMVECTORIALTOSCALARBUFFER(double, short int,*this);
      break;
    case WT_UNSIGNED_SHORT:
      CONVERTSCALEFROMVECTORIALTOSCALARBUFFER(double, unsigned short int,*this);
      break;
    case WT_UNSIGNED_INT:
      CONVERTSCALEFROMVECTORIALTOSCALARBUFFER(double, unsigned int,*this);
      break;
    case WT_SIGNED_INT:
      CONVERTSCALEFROMVECTORIALTOSCALARBUFFER(double, int,*this);
      break;
    case WT_UNSIGNED_LONG:
      CONVERTSCALEFROMVECTORIALTOSCALARBUFFER(double, unsigned LONGINT,*this);
      break;
    case WT_SIGNED_LONG:
      CONVERTSCALEFROMVECTORIALTOSCALARBUFFER(double, LONGINT,*this);
      break;
    case WT_FLOAT:
      CONVERTSCALEFROMVECTORIALTOSCALARBUFFER(double, float,*this);
      break;
    case WT_DOUBLE:
      CONVERTSCALEFROMVECTORIALTOSCALARBUFFER(double, double,*this);
      break;
    case WT_RGB:
      CONVERTFROMVECTORIALTORGBBUFFER(double,*this);
      break;
    case WT_RGBA:
      CONVERTFROMVECTORIALTORGBABUFFER(double,*this);
      break;
    case WT_FLOAT_VECTOR:
      CONVERTFROMVECTORIALTOVECTORIALBUFFER(double, float,*this);
      break;
    default:
      break;
    }
    break;

  default:
    break;
  }

  delete imin;
  delete imax;
  setType(type, vdim, vectMode, false);
}



} // namespace yav

