#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#include <inrimage/ImageIO.h>
#include <gipl.h>
#include <string.h>

#define GIPL_MAGIC 719555000
#define GIPL_VERSION 2

struct gipl_header {
  /* Dimensions of image in voxels and time (x,y,z,t). Bytes: 0 to 7. */
  short int dim[4];
  /* Image data type (binary, char, short etc). Bytes: 8 and 9. */
  short int image_type; 	     
  /* Dimensions of individual image voxels (x,y,z,t). Bytes: 10 to 25. */
  float pixdim[4];	     
  /* Patient description string. Bytes: 26 to 105. */
  char patDesc[80];	     
  /* Matrix (transformation?) stored with image. Bytes: 106 to 153. */
  float matrixElements[12]; 
  /* Image identification number. Bytes: 154 to 157. */
  int identifier;	     
  /* Spare storage string. Bytes: 158 to 185. */
  char spare[28];
  /* Orientation of the image (AP, lateral, axial etc). Byte 186. */
  char orientationFlag;    
  /* GIPL version number. Byte 187. */
  char flag1;		     
  /* Minimum data value. Bytes: 188 to 195. */
  double min;		     
  /* Maximum data value. Bytes: 196 to 203. */
  double max;		     
  /* Image origin in mm. Bytes: 204 to 235. */
  double origin[4];	     
  /* Pixel value offset. Bytes: 236 to 239. */
  float pixval_offset;	     
  /* Pixel value calibration. Bytes: 240 to 243. */
  float pixval_cal;	     
  /* User defined float 1. Bytes: 244 to 247. */
  float user_def1;	     
  /* User defined float 2. Bytes: 248 to 251. */
  float user_def2;	     
  /* GIPL magic number. Bytes: 252 to 255. */
  unsigned int magic_number;
}
#ifndef WIN32
 __attribute__ ((packed));
#endif
;
/*
  IT_UNDEFINED	  0
  IT_BINARY	  1
  IT_CHAR	  7
  IT_U_CHAR	  8
  IT_SHORT	 15
  IT_U_SHORT	 16
  IT_U_INT	 31
  IT_INT	 32
  IT_FLOAT	 64
  IT_DOUBLE	 65
  IT_C_SHORT	144
  IT_C_INT      160
  IT_C_FLOAT	192
  IT_C_DOUBLE	193 */

int readGiplHeader(const char* name,_image *im) {
  if(im->openMode != OM_CLOSE) {
    char buffer[256];
    struct gipl_header hdr;
    
    /** as default the header is read on sun machines */
    im->endianness=END_BIG;
    
    ImageIO_seek(im,0,SEEK_SET);
    ImageIO_gets(im,(char*)&buffer,256);
    ImageIO_seek(im,256,SEEK_SET);

    hdr.dim[0]=_readShortInt(im,&buffer[0]);
    hdr.dim[1]=_readShortInt(im,&buffer[2]);
    hdr.dim[2]=_readShortInt(im,&buffer[4]);
    hdr.dim[3]=_readShortInt(im,&buffer[6]);

    hdr.image_type=_readShortInt(im,&buffer[8]);

    hdr.pixdim[0]=_readFloat(im,&buffer[10]);
    hdr.pixdim[1]=_readFloat(im,&buffer[14]);
    hdr.pixdim[2]=_readFloat(im,&buffer[18]);
    hdr.pixdim[3]=_readFloat(im,&buffer[22]);

    hdr.matrixElements[0]=_readFloat(im,&buffer[106]);
    hdr.matrixElements[1]=_readFloat(im,&buffer[110]);
    hdr.matrixElements[2]=_readFloat(im,&buffer[114]);
    hdr.matrixElements[3]=_readFloat(im,&buffer[118]);
    hdr.matrixElements[4]=_readFloat(im,&buffer[122]);
    hdr.matrixElements[5]=_readFloat(im,&buffer[126]);
    hdr.matrixElements[6]=_readFloat(im,&buffer[130]);
    hdr.matrixElements[7]=_readFloat(im,&buffer[134]);
    hdr.matrixElements[8]=_readFloat(im,&buffer[138]);
    hdr.matrixElements[9]=_readFloat(im,&buffer[142]);
    hdr.matrixElements[10]=_readFloat(im,&buffer[146]);
    hdr.matrixElements[11]=_readFloat(im,&buffer[150]);

    hdr.identifier=_readInt(im,&buffer[154]);
    
    hdr.orientationFlag=buffer[186];

    hdr.min=_readDouble(im,&buffer[188]);
    hdr.max=_readDouble(im,&buffer[196]);

    hdr.origin[0]=_readDouble(im,&buffer[204]);
    hdr.origin[1]=_readDouble(im,&buffer[212]);
    hdr.origin[2]=_readDouble(im,&buffer[220]);
    hdr.origin[3]=_readDouble(im,&buffer[228]);

    hdr.pixval_offset=_readFloat(im,&buffer[236]);

    hdr.pixval_cal=_readFloat(im,&buffer[240]);

    hdr.user_def1= _readFloat(im,&buffer[244]);

    hdr.user_def2= _readFloat(im,&buffer[248]);

    hdr.magic_number= _readUnsignedInt(im,&buffer[252]);

  
    im->xdim = hdr.dim[0];
    im->ydim = hdr.dim[1];
    im->zdim = hdr.dim[2];
    fprintf(stderr,"Size= %d %d %d\n",hdr.dim[0],hdr.dim[1],hdr.dim[2]);
    fprintf(stderr,"Image Type = %d\n",hdr.image_type);
    fprintf(stderr,"magic = %d\n",hdr.magic_number);
    im->dataMode = DM_BINARY;
    im->vx = hdr.pixdim[0];
    im->vy = hdr.pixdim[1];
    im->vz = hdr.pixdim[2];
    
    switch(hdr.image_type) {
    case 7:
      im->wordKind = WK_FIXED;
      im->sign = SGN_SIGNED;
      im->vectMode = VM_SCALAR;
      im->wdim = 1;
      break;
    case 8:
      im->wordKind = WK_FIXED;
      im->sign = SGN_UNSIGNED;
      im->vectMode = VM_SCALAR;
      im->wdim = 1;
      break;
    case 15:
      im->wordKind = WK_FIXED;
      im->sign = SGN_SIGNED;
      im->vectMode = VM_SCALAR;
      im->wdim = 2;
      break;
    case 16:
      im->wordKind = WK_FIXED;
      im->sign = SGN_UNSIGNED;
      im->vectMode = VM_SCALAR;
      im->wdim = 2;
      break;
    case 32:
      im->wordKind = WK_FIXED;
      im->sign = SGN_SIGNED;
      im->vectMode = VM_SCALAR;
      im->wdim = 4;
      break;
    case 31:
      im->wordKind = WK_FIXED;
      im->sign = SGN_UNSIGNED;
      im->vectMode = VM_SCALAR;
      im->wdim = 4;
      break;
    case 64:
      im->sign = SGN_UNKNOWN;
      im->vectMode = VM_SCALAR;
      im->wdim = 4;
      break;
    case 65:
      im->sign = SGN_UNKNOWN;
      im->vectMode = VM_SCALAR;
      im->wdim = 8;
      break;
    default:
      return -1;
      break;
  };
  return 0;
}
  return -1;
}

int testGiplHeader(char *magic,const char *name){
  _image *im2;
  char buffer[256];
  unsigned int magicNumber;

  im2 = _initImage(); 
  _openReadImage(im2, name);
  if(im2->fd) {
    ImageIO_seek(im2,0,SEEK_SET);
    ImageIO_read(im2, buffer, 256);
    im2->endianness=END_BIG;
    magicNumber= _readUnsignedInt(im2,&buffer[252]);
    ImageIO_close(im2);
    return 1;
    if (magicNumber!=GIPL_MAGIC)
      return -1;
    else 
      return 1;
  }
  else {
    ImageIO_close(im2);
    return -1;
  }
} 


/* 
   return:
   -1: error
    1: success */
int writeGipl( char *name, _image* im ) {
	 char *proc = "writeGiplHeader";
  struct gipl_header hdr;
  unsigned int i;
  _openWriteImage(im, name);
  if(!im->fd) {
    fprintf(stderr, "writeGipl: error: unable to open file \'%s\'\n", name );
    return -1;
  }
 
  memset(&hdr,0, sizeof(struct gipl_header));

  hdr.dim[0]=im->xdim;
  hdr.dim[1]=im->ydim;
  hdr.dim[2]=im->zdim;
  hdr.dim[3]=1;
  if(im->wordKind == WK_FIXED &&  im->vectMode == VM_SCALAR && im->wdim == 1) {
    if (im->sign == SGN_SIGNED)
      hdr.image_type = 7; 
    else if (im->sign == SGN_UNSIGNED)
      hdr.image_type = 8;
  }
  else if (im->wordKind == WK_FIXED && im->vectMode == VM_SCALAR && im->wdim == 2) {
    if(im->sign == SGN_SIGNED)
      hdr.image_type = 15;
    else if (im->sign == SGN_UNSIGNED)
      hdr.image_type = 16;
  }
  else if (im->wordKind == WK_FIXED && im->vectMode == VM_SCALAR &&  im->wdim == 4) {
    if(im->sign == SGN_SIGNED)
      hdr.image_type = 32;
    else if (im->sign == SGN_UNSIGNED)
      hdr.image_type = 31;
  }
  else if (im->sign == SGN_UNKNOWN && im->vectMode == VM_SCALAR) {
    if(im->wdim == 4)
      hdr.image_type = 64;
    else if(im->wdim == 8)
      hdr.image_type = 65;
  }
  else {
    fprintf( stderr, "%s: unsupported image type\n", proc );
    return -1;
  }
  hdr.pixdim[0]=im->vx;
  hdr.pixdim[1]=im->vy;
  hdr.pixdim[2]=im->vz;
  hdr.pixdim[3]=1.0;

  for(i=0;i<4;++i) hdr.origin[i] = 0.0;
  for(i=0;i<80;++i) hdr.patDesc[i] = 0;
  for(i=0;i<12;++i) {
    if(i%5) hdr.matrixElements[i] = 0.0;
    else hdr.matrixElements[i] = 1.0;
  }
  hdr.identifier = 0;
  for(i=0;i<28;++i) hdr.spare[i] = 0;
  hdr.orientationFlag = 0;
  hdr.min = 0.0;
  hdr.max = 4095.0;
  hdr.pixval_offset = 0.0;
  hdr.pixval_cal = 1.0;
  hdr.user_def1 = 0.0;
  hdr.user_def2 = 0.0;
  hdr.magic_number = GIPL_MAGIC;
  hdr.flag1 = GIPL_VERSION;
  printf(" %d " ,hdr.dim[0]);
  printf(" %d " ,hdr.dim[1]);
  printf(" %d " ,hdr.dim[2]);
  printf(" %d " ,hdr.dim[3]);
  printf(" %d " ,hdr.flag1);
  printf(" %d " ,hdr.magic_number);
  printf(" %d " ,hdr.image_type);

  if(ImageIO_write(im, &hdr, sizeof(struct gipl_header)) != sizeof(struct gipl_header))
    return -1;		 

  ImageIO_write(im, im->data, im->xdim*im->ydim*im->zdim*im->wdim);

  if (name!=NULL) ImageIO_free(name);
  ImageIO_close(im);
  im->fd = NULL;
  im->openMode = OM_CLOSE;

  return 1;
}

 

/** creates an return the file format structure associated with the Gipl file format */
PTRIMAGE_FORMAT createGiplFormat(){
  PTRIMAGE_FORMAT f=(PTRIMAGE_FORMAT) ImageIO_alloc(sizeof(IMAGE_FORMAT));
  
  f->testImageFormat=&testGiplHeader;
  f->readImageHeader=&readGiplHeader;
  f->writeImage=&writeGipl;
  strcpy(f->fileExtension,".gipl,.gipl.gz,.gipl.Z");
  strcpy(f->realName,"Gipl");
  return f;
}

