#ifndef GLSCALE_H
#define GLSCALE_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/GLScale.h"
#include <basics/Pointed.h>
#include <graphics2D/TclScale.h>

namespace yav {


/** \class GLScale
    \ingroup Inrimage
    \brief grey-level scale representation */
class GLScale : public TclScale {
 public:
  /** builds a grey level scale object from an existing TCL scale
      @param module module this GLScale is created in
      @param scaleName TCL scale name 
      @param elt a scalable managed by this scale*/
  GLScale(TclModule *module, const TclString& scaleName,Scalable *elt);

  /** builds a grey level scale object from an existing TCL scale
      manage all scalables given in list
      @param module module this GLScale is created in
      @param scaleName TCL scale name
      @param list managed scalables */
  GLScale(TclModule *module, const TclString& scaleName,
	  std::list<Scalable *>& list);

  /// empty
  virtual ~GLScale();
  
  /** reset bounds and label
      @param sa object to set min and max values from
      @param lab new label (if not EMPTY_STRING) */
  virtual void reset(Scalable *sa, const TclString& lab = EMPTY_STRING);
  
  /** set given value to GL scale
      @param val new scale value */
  virtual void set(double val);

  /// raises ambigous getRef() function call
  unsigned int getRef(void) const { return TclObject::getRef(); }

  /// TCL new operator callback
  static int create(TclModule *m, int argc, char **argv);

  /// returns scale type
  virtual const std::type_info& getType(void) const { return typeid(GLScale); }

 protected:
  /// Activable::activation method
  virtual void activation(void);

 private:
  /// common constructors code
  void init(void);
};



/** \class MinGLScale
    \ingroup Inrimage
    \brief Image min grey level scale */
class MinGLScale : public GLScale, public Pointed<MinGLScale> {
 public:
  /** builds a min grey level scale object from an existing TCL scale
      @param module module this GLScale is created in
      @param scaleName TCL scale name 
      @param elt a scalable managed by this scale*/
  MinGLScale(TclModule *module, const TclString& scaleName,Scalable *elt=0);

  /** builds a min grey level scale object from an existing TCL scale
      manage all given list of scalables
      @param module module this GLScale is created in
      @param scaleName TCL scale name
      @param list managed scalables */
  MinGLScale(TclModule *module, const TclString& scaleName,
	     std::list<Scalable *> list);

  /// empty destructor
  virtual ~MinGLScale();

  /// returns object id by number
  static MinGLScale *getObject(unsigned int n) {
    return dynamic_cast<MinGLScale *>(TclObjectKind::getObject(typeid(GLScale), n));
  }

  /// returns object id by command line arguments
  static MinGLScale *getObject(int argc, char **argv, int i = 1) {
    return
      dynamic_cast<MinGLScale *>(TclObjectKind::getObject(typeid(GLScale), argc, argv, i));
  }

  /** reset bounds and label
      @param sa object to set min and max values from
      @param lab new label (if not EMPTY_STRING) */
  virtual void reset(Scalable *sa, const TclString& lab = EMPTY_STRING);
  
  /// returns scale type
  virtual const std::type_info& getType(void) const { return typeid(MinGLScale); }

};


/** \class MaxGLScale
    \ingroup Inrimage
    \brief Image max grey level scale */
class MaxGLScale : public GLScale, public Pointed<MaxGLScale> {
 public:
  /** builds a min grey level scale object from an existing TCL scale
      @param module module this GLScale is created in
      @param scaleName scale name 
      @param elt a scalable managed by this scale*/
  MaxGLScale(TclModule *module, const TclString& scaleName,Scalable *elt=0);

  /** builds a max grey level scale object from an existing TCL scale
      manage all given list inrimage views
      @param module module this GLScale is created in
      @param scaleName TCL scale name
      @param list managed inrimage views */
  MaxGLScale(TclModule *module, const TclString& scaleName,
	     std::list<Scalable *> list);

  /// empty destructor
  ~MaxGLScale();

  /// returns object id by number
  static MaxGLScale *getObject(unsigned int n) {
    return dynamic_cast<MaxGLScale *>(TclObjectKind::getObject(typeid(GLScale), n));
  }

  /// returns object id by command line arguments
  static MaxGLScale *getObject(int argc, char **argv, int i = 1) {
    return
      dynamic_cast<MaxGLScale *>(TclObjectKind::getObject(typeid(GLScale), argc, argv, i));
  }

  /** reset bounds and label
      @param sa object to set min and max values from
      @param lab new label (if not EMPTY_STRING) */
  virtual void reset( Scalable *sa, const TclString& lab = EMPTY_STRING);
  
  /// returns scale type
  virtual const std::type_info& getType(void) const { return typeid(MaxGLScale); }

};


} // namespace yav

#endif
