#ifndef IMAGELANDMARKSET_H
#define IMAGELANDMARKSET_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 )
#pragma warning ( disable : 4786 )
#pragma warning ( disable : 4081 )
#endif


#include <framework/Transformation.h>



#include <graphics2D/Camera2D.h>
#include <inrimage/LandmarkSet.h>


namespace yav {


/** \class ImageLandmarkSet
    \ingroup Inrimage
    \brief class that defines a set of landmarks */
class ImageLandmarkSet : public LandmarkSet, public Data2D {
 public:
  /// defines the different ways to display a landmark on a slicecamera
  enum DisplayMode {VOXEL,CROSS,HIDE};
    /// the left mouse bouton action
  enum SelectionMode {CREATE_LANDMARK,MOVE_LANDMARK,DELETE_LANDMARK,SELECT_LANDMARK};

 protected:
  /// the current method for display
  DisplayMode displayMode;
  /// color of the landmarks
  float landmarkColor[3];
  /// color of the selected landmark
  float selectedLandmarkColor[3];
  /// the size in pixels of the cross for display
  unsigned int crossSize;
  /// if the landmark name should be displayed
  bool displayName;
  /// the selected landmark
  ImageLandmark *selectedLandmark;
  /// the moving landmark
  ImageLandmark *movingLandmark;
  /// the different left mouse action
  SelectionMode selectionMode;
 
 public:
  /** @name constructor */
  //@{
  /** build an empty set of landmarks 
      @param scene the 2D where the object is located 
  */
  ImageLandmarkSet(Scene2D *scene);
  /** build an set of landmarks  from a file
      @param scene the 2D where the object is located
      @param filename
  */
  ImageLandmarkSet(Scene2D *scene,char *filename);
 /** build from an STL list of  landmarks 
      @param scene the 2D where the object is located 
      @param list the list of landmarks : each landmark pointer should not be null  */
  ImageLandmarkSet(Scene2D *scene, std::list<ImageLandmark *> list);

  //@}

  /** @name destructor */
  virtual ~ImageLandmarkSet();
 protected:
  /// initialize variuous variables;
  void init();
  
 public:

  /// add a landmark in the set and update the listbox if any
  void addLandmark(ImageLandmark *imagelandmark);
  /// remove landmark from list (free landmark)
  void removeLandmark(ImageLandmark *imagelandmark);

 /** Translates the 2D object
      @param t translation vector */
  virtual void translate(const Vec2<double>& t){}

  /** Rotates the 2D object: no rotation available for images
      @param r rotation matrix */
  virtual void rotate(const Rotation2D &r){}

  /** Scales the 2D object
      @param s scale factor */
  virtual void scale(const double s){}

 /** callback called when a button is pressed in a camera
      @param button indicates which button is pressed
      @param modifier indicates if a modifier (shift or control key) is pressed
      @param cam the 2D camera where occured the click button
      @param mx first coordinate of the pixel clicked 
      @param my second coordinate of the pixel clicked */ 
  virtual void mouseButtonPressed(CameraTypes::WHICH_BUTTON button,
				  CameraTypes::KEY_MODIFIER modifier,
				  Camera2D *cam,int mx,int my);
 /** callback called when a button is released in a camera
     @param button indicates which button is released
     @param modifier indicates if a modifier (shift or control key) is pressed
     @param cam the 2D camera where occured the click button
     @param mx first coordinate of the pixel clicked 
     @param my second coordinate of the pixel clicked */ 
  virtual void mouseButtonReleased(CameraTypes::WHICH_BUTTON button,
				   CameraTypes::KEY_MODIFIER modifier,
				     Camera2D *cam,int mx,int my);
 /** callback called when a button is moved after having been pressed
     @param cam the 2D camera where occured the click button
     @param x first coordinate of the pixel clicked 
     @param y second coordinate of the pixel clicked */ 
  virtual void mouseButtonMoved(Camera2D *cam,int x,int y);

  /// Returns the graphic object center and weight.
  virtual void updateCenter(Vec2<double>& center, unsigned int& n)  {
    center = Vec2<double>(0.0, 0.0);
    n = 0;
  }
  /// Returns the graphic object radius given its center
  virtual double updateRadius(const Vec2<double>& center){return 0.0;}

  /** Return MAXDOUBLE, an image is considered as background
      and may not be selected */
  virtual double computeClosestDistance(DataTypes::MouseClick& mc) const{
  return -1.0;}
  /// refresh inrimage in any viewer it is displayed   ??????????
  void redraw(Camera2D *cam);

    /// returns current selection mode
  DisplayMode getDisplayMode() const {
    return displayMode;
  }
  /// sets the current selection mode 
  void setDisplayMode(const DisplayMode d)  {
    displayMode=d;
  }

  /// returns current selection mode
  SelectionMode getSelectionMode() const {
    return selectionMode;
  }
  /// sets the current selection mode 
  void setSelectionMode(const SelectionMode d)  {
    selectionMode=d;
  }
  /// enquire if the landmark names should be displayed
  bool getDisplayName() const {
    return displayName;
  }
  /// set if the landmark names should be displayed
  void setDisplayName(const bool d) {
    displayName=d;
  }
 /// get the default color of the landmarks
  void getColor( float &r, float &g, float &b) const {
    r=landmarkColor[0];g=landmarkColor[1];b=landmarkColor[2];
  }
  /// get the default color of the landmarks
  void setColor( const float r,const  float g, const float b) {
    landmarkColor[0]=r;landmarkColor[1]=g;landmarkColor[2]=b;
  }
 /// get the color of the selected landmark
  void getSelectedColor( float &r, float &g, float &b) const {
    r=selectedLandmarkColor[0];g=selectedLandmarkColor[1];b=selectedLandmarkColor[2];
  }
  /// get the color of the selected landmark
  void setSelectedColor( const float r,const  float g, const float b) {
    selectedLandmarkColor[0]=r;selectedLandmarkColor[1]=g;selectedLandmarkColor[2]=b;
  }
  /// returns the selected landmark
  ImageLandmark *getSelectedLandmark() const {
    return selectedLandmark;
  }
  /// sets the selected landmark
  void setSelectedLandmark(ImageLandmark *l) {
    selectedLandmark=l;
  }
    /// returns the selected landmark
  ImageLandmark *getMovingLandmark() const {
    return movingLandmark;
  }
  /// sets the selected landmark
  void setMovingLandmark(ImageLandmark *l) {
    movingLandmark=l;
  }
  /// indicates the current cross size for each landmark
  unsigned int getCrossSize() const {
    return crossSize;
  }
  /// sets the current cross size for each landmark
  void  setCrossSize(const unsigned int s)  {
    crossSize=s;
  }


};

}


#endif
