#ifndef INRIMAGE4DVIEW_H
#define INRIMAGE4DVIEW_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/Inrimage4DView.h"
#include <graphics2D/TimeSliceCamera.h>
#include <inrimage/Inrimage4D.h>
#include <graphics2D/TimeSliceable.h>

namespace yav {


#ifdef TCL_INTERFACE
class TclInrimage4DView;
#endif

/** \class Inrimage4DBackground
    \ingroup Inrimage

    an interface class that adds specific functions for a background attached
    to an inrimage */
class Inrimage4DBackground : public TimeSliceCameraBackground {
 public:
  /** @name constructors */
  //@{
  /** Dummy constructor since no inrimage is specified
      @param cam camera attached to the background */
  Inrimage4DBackground(TimeSliceCamera *cam) :TimeSliceCameraBackground(cam) {}
  //@}
  /// destruction
  virtual ~Inrimage4DBackground(){}
  
  /// image type changed
  virtual void inrimageChanged(void)=0;
  /** redraw camera image */
  virtual void redraw()=0;
  /// reset pointer to camera
  virtual void refreshCamera(void)=0;

  friend class Inrimage4DView;
  };


/** \class Inrimage4DView
    \ingroup Inrimage
    \brief A 2D object attached to a 4D inrimage that display image slices */
class Inrimage4DView : public TimeSliceable, public Owner<Inrimage4D>
{
 protected:
  /** @name constructors */
  //@{
  /** constructor that reads an image from a file 
      @param filename unused name
      @param scene  scene to build data in */
  Inrimage4DView(const TclString &filename,
	       Scene2D *scene);
  /** Constructor that does not call init() 
      @param inr1 inrimage to draw
      @param i dummy inrimage pointer
      @param scene  scene to build data in 
      @param _createGLScale boolean indicating if we should create  GLScales associated with each background */
  Inrimage4DView(Inrimage4D *inr1, Inrimage4D *i,Scene2D *scene,
	       bool _createGLScale=true);
  //@}

  /// constructors common code
  void init(void);
  
 public:
    
  /** @name constructors */
  //@{
  /** Normal constructor by specifiying the inrimage to display
      @param inr 4D image  to display
      @param scene  scene to build data in 
      @param _createGLScale boolean indicating if we should create  GLScales 
  associated with each background */
  Inrimage4DView(Inrimage4D *inr, Scene2D *scene,bool _createGLScale=true);
  //@}

  /// destruction: do not delete visualized inrimage
  virtual ~Inrimage4DView();
  
  /// die on inrimage deletion
  virtual void dying(const ObjectPointer<Inrimage4D>&);

  /** focus camera on image center
      @param cam camera to focus */
  void focusCamera(SliceCamera& cam);

  /// add a camera in inrimage view management if needed
  virtual void createdCamera(SliceCamera& cam);
  /// add a camera in inrimage view management
  virtual void doCreateCamera(SliceCamera& cam);

  /// remove a camera from inrimage view management
  void deletedCamera(SliceCamera& cam);

  /// return this object scene
  Scene2D *getScene(void) const { return scene; }


  /// returns eventually the TclImage associated with a given slice camera
  Inrimage4DBackground *getBackgroundFromCamera(TimeSliceCamera *cam);
  
  /** Translates the 2D object
      @param t translation vector */
  virtual void translate(const Vec2<double>& t) {}

  /** Rotates the 2D object: no rotation available for images
      @param r rotation matrix */
  virtual void rotate(const Rotation2D &r) {}

  /** Scales the 2D object
      @param s scale factor */
  virtual void scale(const double s) {}

  /// Returns the graphic object center and weight.
  virtual void updateCenter(Vec2<double>& center, unsigned int& n)  {
    center = Vec2<double>(0.0, 0.0);
    n = 0;
  }

  /// nothing done when an image is selected
  virtual void select(Camera2D *) {}


  /// Returns the graphic object radius given its center
  virtual double updateRadius(const Vec2<double>& center)  { return 0.0; }

  /** Return MAXDOUBLE, an image is considered as background
      and may not be selected */
  virtual double computeClosestDistance(DataTypes::MouseClick& mc) const {
    return -1.0;
  }
  /// return iterator on the first background 
  std::list<Inrimage4DBackground *>::iterator firstBackground() {
    return(backgroundList.begin());
  }
  /// return iterator on the last background 
  std::list<Inrimage4DBackground *>::iterator lastBackground() {
    return(backgroundList.end());
  }
  /** returns the pointed inrimage */
  Inrimage4D *getInrimage() const;

  /** Modifies the pointed image and Updates all images */
  virtual void changeInrimage(Inrimage4D *inr);

  /// reset InrimageView::TclImage objects list pointers to cameras
  void refreshCameras(void);

  /// nothing done when an image is unselected
  virtual void unselect(Camera2D *) {}

  /** default redraw function for sliceable objects
      @param cam camera to redraw */
  virtual void redraw(Camera2D *cam);
  

#ifdef TCL_INTERFACE
  TclInrimage4DView *getInterface(void) const;

  friend class TclInrimageView;
#endif

 protected:
  /// the list of backgrounds associated with the inrimage view
  std::list<Inrimage4DBackground *> backgroundList;
  /// whether we should create
  bool createGLScale;

  friend class TclInrimage4DBackground;
  friend class TclInrimage4D;
};


} // namespace yav



#endif

