#ifndef INRIMAGESLICEINTERFACE_H
#define INRIMAGESLICEINTERFACE_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/InrimageSliceInterface.h"

#include <graphics2D/SliceCamera.h>
#include <inrimage/Inrimage.h>

namespace yav {

/** \class InrimageSliceDisplayContainer
    \ingroup Inrimage
    \brief a class that keeps an update on Inrimage slice */
class InrimageSliceDisplayContainer : public ListElement<InrimageSliceDisplayContainer> {

 public:
  /** function called when a TclInrimageBackground is created 
      @param b a slice-camera background*/
   virtual void sliceCreation(SliceCameraBackground *b)=0;
  /** function called when a TclInrimageBackground is deleted
      @param b a slice-camera background*/
   virtual void sliceDestruction(SliceCameraBackground *b)=0;
  /** function called when a TclInrimageBackground is modified
      @param b a slice-camera background*/
   virtual void sliceUpdate(SliceCameraBackground *b)=0;
  /** function called when the inrimage is modified
      @param i an inrimage */
   virtual void imageUpdate(const Inrimage *i)=0;
  /** function called when the inrimage is deleted
      @param i an inrimage */
   virtual void imageDestroyed(const Inrimage *i)=0;
 
};


/** \class InrimageSliceInterface
 *  \ingroup Inrimage
 *
 *  a class associated with an image, used to store information about : 1)
 *  operation about the image itself 2) inrimage background created with each
 *  slicecam (creation, update and destruction of a background)  */
class InrimageSliceInterface : public Owner<Inrimage>, 
			       public AutoLink<InrimageSliceInterface> 
{
 public:
  /** @name Constructors */
  //@{
  /** creation from an image */
  InrimageSliceInterface(Inrimage *inr); 
  //@}
  
  /** empty destructor */
  ~InrimageSliceInterface() {}

  /* @name Exception classes */
  //@{
  /** \class NoImageProvided
      \ingroup Inrimage
      \briefno image  */
  class NoImageProvided : public Exception {
  public:
    NoImageProvided(const std::string& from);
  };
  //@}
  
 protected:
  /// list of slice display container
  ObjectsList<InrimageSliceDisplayContainer> containerList;
 public:
  /// returns the pointed image 
  Inrimage *getInrimage() const {
    return pointer();
  }
  /// gives the first container of the list
  ObjectsList<InrimageSliceDisplayContainer>::iterator getFirstContainer() {
    return containerList.begin();
  }
  /// gives the last container of the list
  ObjectsList<InrimageSliceDisplayContainer>::iterator getLastContainer() {
    return containerList.end();
  }
  /// warn all slice container that a slice has been created
  void addSlice(SliceCameraBackground *b) const;
  /// warn all slice container that a slice has been created
  void deleteSlice(SliceCameraBackground *b) const;
  /// warn all slice container that a slice has been created
  void updateSlice(SliceCameraBackground *b) const;
  /// warn all slice container that the image has been modifed
  void updateImage() const;
  /// warn all slice container that the image is being deleted
  void deleteImage() const;
  /// add a container to the list
  void addContainer(InrimageSliceDisplayContainer *c);
  /// die on inrimage deletion
  virtual void dying(const ObjectPointer < Inrimage > &);
  
};

} // namespace yav

#endif
