#ifndef OPACITYSCALE_H
#define OPACITYSCALE_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/OpacityScale.h"
#include <basics/Pointed.h>
#include <graphics2D/TclScale.h>

namespace yav {


/** \class OpacityScale
    \ingroup Inrimage
    \brief opacity-level scale representation */
class OpacityScale : public TclScale, public Pointed<OpacityScale> {
 public:
  /** builds an opacity level scale object from an existing TCL scale
      @param module module this GLScale is created in
      @param scaleName TCL scale name 
      @param elt a scalable managed by this scale*/
  OpacityScale(TclModule *module, const TclString& scaleName,Scalable *elt);

  /** builds an opacity level scale object from an existing TCL scale
      manage all scalables given in list
      @param module module this GLScale is created in
      @param scaleName TCL scale name
      @param list managed scalables */
  OpacityScale(TclModule *module, const TclString& scaleName,
	       std::list<Scalable *>& list);

  /// empty
  virtual ~OpacityScale();
  
  /** reset bounds and label
      @param sa object to set min and max values from
      @param lab new label (if not EMPTY_STRING) */
  virtual void reset(Scalable *sa, const TclString& lab = EMPTY_STRING);
  
  /** set given value to opacity scale
      @param val new scale value */
  virtual void set(double val);

  /// raises ambigous getRef() function call
  unsigned int getRef(void) const { return TclObject::getRef(); }

  /// TCL new operator callback
  static int create(TclModule *m, int argc, char **argv);

  /// returns scale type
  virtual const std::type_info& getType(void) const { return typeid(OpacityScale); }
  
  /// returns object id by number
  static OpacityScale *getObject(unsigned int n) {
    return dynamic_cast<OpacityScale *>(TclObjectKind::getObject(typeid(OpacityScale), n));
  }
 protected:
  /// Activable::activation method
  virtual void activation(void);

 private:
  /// common constructors code
  void init(void);
};


} // namespace yav


#endif
