#ifndef OVERLAYINRIMAGE4DVIEW_H
#define OVERLAYINRIMAGE4DVIEW_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/OverlayInrimage4DView.h"
#include <inrimage/Inrimage4DView.h>

namespace yav {


#ifdef TCL_INTERFACE
class TclOverlayInrimage4DView;
#endif


/** \class OverlayInrimage4DView
    \ingroup Inrimage
    \brief A viewer for merging 2 inrimage4D of the same size and type */
class OverlayInrimage4DView : public Inrimage4DView
{
 public:
  /** @name Exception classes */
  //@{
  /** \class MismatchedImageType
      \ingroup Inrimage
      \brief mismatched inrimage type */
  class MismatchedImageType : public Exception {
  public:
    MismatchedImageType(const std::string& from);
  };
  /** \class MismatchedImageSize
      \ingroup Inrimage
      \brief mismatched inrimage size */
  class MismatchedImageSize : public Exception {
  public:
    MismatchedImageSize(const std::string& from);
  };
  /** \class MismatchedImageTime
      \ingroup Inrimage
      \brief mismatched inrimage time */
  class MismatchedImageTime : public Exception {
  public:
    MismatchedImageTime(const std::string& from);
  };
  /** \class NoImage
      \ingroup Inrimage
      \brief empty inrimage  */
  class NoImage : public Exception {
  public:
    NoImage(const std::string& from);
  };
  //@}
 public:
    
  /** @name constructors */
  //@{
  /** Normal constructor by specifiying 2 inrimages to merge and display
      @param inr1 first inrimage to draw
      @param inr2 second inrimage to draw
      @param scene  scene to build data in 
      @param _createGLScale boolean indicating if we should create  GLScales 
  associated with each background */
  OverlayInrimage4DView(Inrimage4D *inr1,Inrimage4D *inr2, 
		      Scene2D *scene,bool _createGLScale=true) 
    throw(MismatchedImageType,MismatchedImageSize,MismatchedImageTime,NoImage);
  //@}

  /// destruction: do not delete visualized inrimage
  virtual ~OverlayInrimage4DView() {}
  

  /// add a camera in inrimage view management
  virtual void doCreateCamera(SliceCamera& cam);


  /** returns the first pointed inrimage */
  Inrimage4D *getFirstInrimage() const {
    return getInrimage();
  }
  /** returns the second pointed inrimage */
  Inrimage4D *getSecondInrimage() const {
    return (Inrimage4D *)secondImage();
  }
   /** Modifies the first inrimage and updates all slices */
  virtual void changeInrimage(Inrimage4D *inr);
  /** Modifies the second inrimage and Updates all slices  */
  void changeSecondInrimage(Inrimage4D *inr);
  

#ifdef TCL_INTERFACE
  TclOverlayInrimage4DView *getInterface(void) const;

  friend class TclOverlayInrimage4DView;
#endif

 protected:
  /// second intensity image
  ObjectPointer<Inrimage4D> secondImage; 

  friend class TclOverlayInrimage4DBackground;
  friend class TclInrimage4D;
};


} // namespace yav


#endif
