#ifndef OVERLAYINRIMAGEVIEW_H
#define OVERLAYINRIMAGEVIEW_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/OverlayInrimageView.h"
#include <inrimage/InrimageView.h>

namespace yav {


#ifdef TCL_INTERFACE
class TclOverlayInrimageView;
#endif


/** \class OverlayInrimageView
    \ingroup Inrimage
    \brief A viewer for merging 2 inrimage of the same size and type */
class OverlayInrimageView : public InrimageView
{
 public:
  /** @name Exception classes */
  //@{
  /** \class MismatchedImageType
      \ingroup Inrimage
      \brief mismatched inrimage type */
  class MismatchedImageType : public Exception {
  public:
    MismatchedImageType(const std::string& from);
  };
  /** \class MismatchedImageSize
      \ingroup Inrimage
      \brief mismatched inrimage size */
  class MismatchedImageSize : public Exception {
  public:
    MismatchedImageSize(const std::string& from);
  };
  /** \class NoImage
      \ingroup Inrimage
      \brief empty inrimage  */
  class NoImage : public Exception {
  public:
    NoImage(const std::string& from);
  };
  //@}
 protected:
  /// second intensity image
  ObjectPointer<Inrimage> secondImage; 
 /// the slice interface linked to the second image
  InrimageSliceInterface *secondSliceInterface;
 public:
    
  /** @name constructors */
  //@{
  /** Normal constructor by specifiying 2 inrimages to merge and display
      @param inr1 first inrimage to draw
      @param inr2 second inrimage to draw
      @param scene  scene to build data in 
      @param _createGLScale boolean indicating if we should create  GLScales 
  associated with each background */
  OverlayInrimageView(Inrimage *inr1,Inrimage *inr2, 
		      Scene2D *scene,bool _createGLScale=true) 
    throw(MismatchedImageType,MismatchedImageSize,NoImage);
  //@}

  /// destruction: do not delete visualized inrimage
  virtual ~OverlayInrimageView() {}
  

  /// add a camera in inrimage view management
  virtual void doCreateCamera(SliceCamera& cam);


  /** returns the first pointed inrimage */
  Inrimage *getFirstInrimage() const {
    return getInrimage();
  }
  /** returns the second pointed inrimage */
  Inrimage *getSecondInrimage() const {
    return (Inrimage *)secondImage();
  }
   /** Modifies the first inrimage and updates all slices */
  virtual void changeInrimage(Inrimage *inr);
  /** Modifies the secoond inrimage and Updates all slices  */
  virtual void changeSecondInrimage(Inrimage *inr);
  /// returns the second slice interface associated with the image
  InrimageSliceInterface *getSecondSliceInterface() const {
    return sliceInterface;
  }

#ifdef TCL_INTERFACE
  TclOverlayInrimageView *getInterface(void) const;

  friend class TclOverlayInrimageView;
#endif


  friend class TclOverlayInrimageBackground;
  friend class TclInrimage;
};


} // namespace yav


#endif

