#ifndef TCLCOLORMAP_H
#define TCLCOLORMAP_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/TclColorMap.h"
#include <modules/TclObject.h>
#include <inrimage/Inrimage.h>

namespace yav {

/** \class TclColorMap
    \ingroup Inrimage
    \brief a colormap consisting of 3 red, green, blue array that can be used by a slicecamera background  */
class TclColorMap : public TclObject {
 protected:
  /// red band of the colormap
  unsigned char red[258];
  /// green band of the colormap
  unsigned char green[258];
  /// blue band of the colormap
  unsigned char blue[258];
 public:
  /** @name Exception classes */
  //@{
  /** \class ImageWrongSize
      \ingroup Inrimage
      \brief the input image does not have the right size : 256x256x3 or 257x257x3 */
  class ImageWrongSize : public Exception {
  public:
    ImageWrongSize(const std::string& from) :
      Exception(from,"image  is not a colormap image " ) {}
  };
  /** \class FileNotFound
      \ingroup Inrimage
      \brief the input file cannot be found */
  class FileNotFound : public Exception {
  public:
    FileNotFound(const std::string& from,
		 const std::string& name) : 
      Exception(from,"unable to open colormap " + name) {}
  };
  //@}

  /** @name constructor */
  //@{ 
  /** create a colormap from the 3 RGB arrays 
   @param m  a module to create a TCL Object
   @param r the red array of size 257 
   @param g the green array of size 257 
   @param b the blue array of size 257   */
  TclColorMap(TclModule *m, const  unsigned char *r,
	      const  unsigned char *g,const  unsigned char *b);
  /** create a colormap from an inrimage of size 256x256x3 
   @param m  a module to create a TCL Object
   @param inr an inrimage of size 256x256x3 ou 257x257x3 ou 258x258x3*/
  TclColorMap(TclModule *m, Inrimage *inr) throw(ImageWrongSize);
  /** create a colormap from an inrimage from a file 
   @param m  a module to create a TCL Object
   @param fname the file name associated with a colormap */
  TclColorMap(TclModule *m, char *fname) throw(FileNotFound);
  //@}
  
  /** @name destructor */
  //@{
  /// destroys this colormap
  virtual ~TclColorMap(){}
  //@}
  
 public:
  /// returns a pointer to the red array
  const unsigned char *getRed() const {
    return red; }
  /// returns a pointer to the green array
  const unsigned char *getGreen() const {
    return green; }
  /// returns a pointer to the blue array
  const unsigned char *getBlue() const {
    return blue; }
  /// new TCL command callback
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }
  /// new TCL command callback
  static int create(TclModule *m, int argc, char **argv);
  /// -destroy option
  int destroyCmd(int argc, char **argv);
  /// -getColor option
  int getColorCmd(int argc, char **argv);
  /// -setColor option
  int setColorCmd(int argc, char **argv);
};

} // namespace yav

#endif
