#ifndef TCLDISPLACEMENTFIELD3D_H
#define TCLDISPLACEMENTFIELD3D_H

#pragma interface "inrimage/TclDisplacementField3D.h"

#include <inrimage/DisplacementField3D.h>
#include <graphics/TclSceneObject.h>

namespace yav {


  /** \class TclDisplacementField3D
      \ingroup Inrimage
      \brief Tcl interface to class DisplacementField3D */
  class TclDisplacementField3D : public TclSceneObject {

     public:
  /** @name constructors */
  //@{
  /** Reads 3D displacement field from file whose path is given as argument.
      @param m module this field is created in
      @param name 3D displacement  file name
      @exception ImageNotFound image file not found */
    
    TclDisplacementField3D(TclModule *m, const char *name) 
      throw ( GeneralFormat::FileNotFound, Inrimage::ImageNotFound, DisplacementField3D::InvalidDisplacementField3DImagesNb, DisplacementField3D::DisplacementField3DMismatchType,DisplacementField3D::DisplacementField3DMismatchSize ) ;

    /** create a transformation depending on its mode
	The images of the deformation are needed : 2 in 2D and 3 in 3D 
	In case of VOXEL mode, size of voxel is needed */

    TclDisplacementField3D(TclModule *m, DisplacementField3D::FIELD_MODE mode, Inrimage *imx, Inrimage *imy, Inrimage *imz, float vx=1, float vy=1, float vz=1 ) throw (DisplacementField3D::DisplacementField3DMismatchType, DisplacementField3D::DisplacementField3DMismatchSize, DisplacementField3D::DisplacementField3DMismatchVoxelSize);

    //@}
    
    /** @name destructor */
    //@{
    // desallocates images buffers and destroy object
    ~TclDisplacementField3D();
    //@}
    
    
    /// return the TCL object associated with this interface */
    DisplacementField3D *getClient(void) const {
      return (DisplacementField3D *) client;
    }
    
    
    /// constructors common code
    void init(TclModule *m);  
    
    // return object by ref or 0 if it is not found
    static TclDisplacementField3D *getObject(unsigned int n) {
      return dynamic_cast<TclDisplacementField3D*>(TclObjectKind::getObject(typeid(TclDisplacementField3D),
									    n));
    }
    
    // return object by name (argv[i]) or 0 with an error message
    // in module interpretor if it is not found
    static TclDisplacementField3D *getObject(int argc, char **argv, int i = 1) {
      return
	dynamic_cast<TclDisplacementField3D *>(TclObjectKind::getObject(typeid(TclDisplacementField3D),
									argc, argv, i));
    }
    
    // return object id by command line arguments
    static const char *getTCLObjectName(void) {
      return TclObjectKind::getObjectName(typeid(TclDisplacementField3D));
    }
    
    // return object id by command line arguments
    static int list(TclModule *m, int argc, char **argv) {
      return TclObject::list(m, argc, argv);
    }
    
    /// Tcl command to create the TclDisplacementField3D
    static int create(TclModule *m, int argc, char **argv); 

    /// -resample command option
    int defxCmd(int argc, char **argv);
    /// -resample command option
    int defyCmd(int argc, char **argv);
    /// -resample command option
    int defzCmd(int argc, char **argv);

  };



} // namespace yav

#endif
