#ifndef TCLIMAGE3DBACKGROUNDMANAGER_H
#define TCLIMAGE3DBACKGROUNDMANAGER_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/TclImage3DBackgroundManager.h"

#include <inrimage/Image3DBackgroundManager.h>
#include <graphics2D/Scalable.h>
#include <modules/TclObject.h>
#include <modules/TclSon.h>
#include <inrimage/GLScale.h>
#include <inrimage/OpacityScale.h>

namespace yav {

class TclInrimage;  

/** \class BasicCameraBackground
    \ingroup Graphics2D
    \brief A single image Background on a 2D camera */
class TclImage3DBackgroundManager : public Image3DBackgroundManager, 
				    public Scalable,
				    public TclSon,
				    public TclObject,
				    public Owner<MinGLScale>,
				    public Owner<MaxGLScale>,
				    public Owner<OpacityScale> {
protected:
	 /// the name of the label widget
	 TclString labelWidget;
	 /// store the scale factor between the current camera widget size and the standard one
	 double zoom;
	 /// second Tcl minimim scale for grey-level windowing
	 ObjectPointer<MinGLScale> secondMinScale; 
	 /// second Tcl maximim scale for grey-level windowing
	 ObjectPointer<MaxGLScale> secondMaxScale; 
public:
	/** @name constructor */
  //@{
	 /** constructor requires Tcl window
	 @param m the Tcl Module where the object is defined  
	 @param window the name of the Tcl widget associated with this multiview : when this window is destroyed then the manager is destroyed */ 
  TclImage3DBackgroundManager(TclModule *m,TclInrimage *inr,
			      const TclString& window = EMPTY_STRING);
  //@}

  /** @name destructor */
  //@{
  /** destroyed the backgrounds that it manages at the same time */
  virtual ~TclImage3DBackgroundManager();
  //@}
 
    // return object id by command line arguments
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }

  /** get object string identifier
  @return object name */
  virtual TclString objectName(void) const;
 
  /** scale value changed
      @param scale calling scale object
      @param val new scale value */
  virtual void scaleValue(const TclScale *scale, double val);

  /** scale was required to move without updating scalable objects: warn them anyway
      @param scale calling scale object
      @param val new scale value */
  virtual void scaleTo(const TclScale *scale, double val);

  /** get min possible scale value for this scalable
      @param scale calling scale object
      @return min possible value */
  virtual double getMin(const TclScale *scale) const;

  /** get max possible scale value for this scalable
      @param scale calling scale object
      @return max possible value */
  virtual double getMax(const TclScale *scale) const;

  /** get required scale step for this scalable
      @param scale calling scale object
      @return step value */
  virtual double getStep(const TclScale *scale) const;
  /** callback issued by a background informing that the grey-level of the image
  has changed */
  virtual void imageHasChangedTexture(Inrimage *);
  /** get current object slice
      @param scale calling scale object
      @return slice number */
 // virtual double getSlice(const TclScale *scale) const;
 /** an element owned by this object is being destroyed
      @param obj element before destruction */
  virtual void dying(const ObjectPointer<MinGLScale>& obj){}

  /** an element owned by this object is being destroyed
  @param obj element before destruction */
  virtual void dying(const ObjectPointer<MaxGLScale>& obj){}
   /** an element owned by this object is being destroyed
  @param obj element before destruction */
  virtual void dying(const ObjectPointer<OpacityScale>& obj){}
  /** update the label that describes the position and value of
  the voxel pointed by the current mouse position */
  virtual void updateLabel(const char *label);
  /// update the value of the grey level sliders
  virtual void updateSliders() const;


/** change the size of all managed camera (call the resize function)
@param val it is the current ratio of the current camera sizes compared to the initial size */ 
void changeCameraSize(const double val);

  /// -destroy option
  int destroyCmd(int argc, char **argv);
   /// -colormap option
  int colorMapCmd(int argc, char **argv);
   /// -image option
  int imageCmd(int argc, char **argv);
   /// -colormap2 option
  int colorMap2Cmd(int argc, char **argv);
 /// -showAxis option
  int showAxisCmd(int argc, char **argv);
  /// -GLScale option
  int GLScaleCmd(int argc, char **argv);
    /// -GLScale2 option
  int GLScale2Cmd(int argc, char **argv);
    /// -OpacityScale option
  int opacityScaleCmd(int argc, char **argv);
  /// -synchronize option
  int synchronizeCmd(int argc, char **argv);
     /// -hideAxis option
  int hideAxisCmd(int argc, char **argv);
  /// -background option
  int backgroundCmd(int argc, char **argv);
  /// -window option
  int windowCmd(int argc, char **argv);
    /// -window2 option
  int window2Cmd(int argc, char **argv);
  /// -secondImage option
  int secondImageCmd(int argc, char **argv);
  /// -parentWidget option
  int parentWidgetCmd(int argc, char **argv);
  /// -interpolation option
  int interpolationCmd(int argc, char **argv);
  /// -labelWidget option
  int labelWidgetCmd(int argc, char **argv);
  /// -widgetSize option
  int widgetSizeCmd(int argc, char **argv);
  /// -middleMouseAction option
  int middleMouseActionCmd(int argc, char **argv);
    /// -showVectors option
  int showVectorsCmd(int argc, char **argv);
  /// -showArrows option
  int showArrowsCmd(int argc, char **argv);
   /// -movePointTo option
  int moveToPointCmd(int argc, char **argv);
  /// -arrow option
  int arrowCmd(int argc, char **argv);
  /// -vectorScale option
  int vectorScaleCmd(int argc, char **argv);
  /// -vectorColor option
  int vectorColorCmd(int argc, char **argv);
  /// -vectorStep option
  int vectorStepCmd(int argc, char **argv);
  /// new TCL command callback
  static int create(TclModule *m, int argc, char **argv);
  /// creates the Tcl commands when the module is loaded
  static void createCommands(TclModule *m);
};

}

#endif
