#ifndef TCLIMAGE3DCAMERABACKGROUND_H
#define TCLIMAGE3DCAMERABACKGROUND_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/TclImage3DCameraBackground.h"

#include <inrimage/Image3DCameraBackground.h>
#include <graphics2D/Scalable.h>
#include <modules/TclObject.h>


namespace yav {

	class TclCamera2D;
	class TclInrimage;
	class SliceScale;
	class TclImage3DBackgroundManager;

/** \class TclImage3DCameraBackground
    \ingroup Inrimage
    \brief Tcl Interface around a 3D Image background  */
class TclImage3DCameraBackground : public Image3DCameraBackground, public TclObject, public Scalable {

public:
	/** @name constructor */
  //@{
	 /** constructor based an image filename
	 @param cam the camera associated with this background
	  @param filename the name of an SGI format image */ 
	TclImage3DCameraBackground(TclCamera2D *cam,TclImage3DBackgroundManager *manager,
		SLICE_ORIENTATION geom);
  //@}

  /** @name destructor */
  //@{
  virtual ~TclImage3DCameraBackground();
  //@}
 
    // return object id by command line arguments
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }

   /** get object string identifier
      @return object name */
  virtual TclString objectName(void) const;
 /** get object string identifier : useful to recognize background without knowing its type
  @return object name */
  virtual TclString getName(void) const {
	  return objectName();
  }
  /*** change the slice Number and recomputes the texture */
  virtual void setSliceNumber(unsigned int s);
  /** scale value changed
      @param scale calling scale object
      @param val new scale value */
  virtual void scaleValue(const TclScale *scale, double val);

 
  /** scale was required to move without updating scalable objects: warn them anyway
      @param scale calling scale object
      @param val new scale value */
  virtual void scaleTo(const TclScale *scale, double val);

  /** get min possible scale value for this scalable
      @param scale calling scale object
      @return min possible value */
  virtual double getMin(const TclScale *scale) const;

  /** get max possible scale value for this scalable
      @param scale calling scale object
      @return max possible value */
  virtual double getMax(const TclScale *scale) const;

  /** get required scale step for this scalable
      @param scale calling scale object
      @return step value */
  virtual double getStep(const TclScale *scale) const;

  /** get current object slice
      @param scale calling scale object
      @return slice number */
  virtual double getSlice(const TclScale *scale) const;


  /** warn the background that the image has changed it size */
  virtual void imageHasChangedSize(Inrimage *inr);

   /// -sliceScale option
  int sliceScaleCmd(int argc, char **argv);
  /// -destroy option
  int destroyCmd(int argc, char **argv);
  /// -camera option
  int cameraCmd(int argc, char **argv);
    /// -sliceNumber option
  int sliceNumberCmd(int argc, char **argv);
  /// -requiredSize option
  int requiredSizeCmd(int argc, char **argv);
   /// -sliceName option
  int sliceNameCmd(int argc, char **argv);

  /// new TCL command callback
  static int create(TclModule *m, int argc, char **argv);
  /// creates the Tcl commands when the module is loaded
  static void createCommands(TclModule *m);
};

}

#endif
