#ifndef TCLIMAGELANDMARKSET_H
#define TCLIMAGELANDMARKSET_H


#ifdef _MSC_VER
#pragma warning ( disable : 4068 )
#pragma warning ( disable : 4786 )
#pragma warning ( disable : 4081 )
#endif

#pragma interface "inrimage/TclImageLandmarkSet.h"


#include <graphics2D/TclScene2DObject.h> 
#include <graphics2D/TclScene2D.h>
#include <inrimage/ImageLandmarkSet.h>


namespace yav {

/** \class TclImageLandmarkSet
    \ingroup Inrimage
    \brief a Tcl interface class for a ImageLandmarkSet */
class TclImageLandmarkSet : public TclScene2DObject {

public: 

  /** @name constructor */
  //@{
  /** builds a new TclImageLandmarkSet from an ImageLandmarkSet
      @param m Tcl module 
      @param data the structure that is pointed by this
      @param scene the scene where the set is  located 
  **/
   TclImageLandmarkSet(TclModule *m,   
		       ImageLandmarkSet *data, TclScene2D *scene); 
  /** builds a new TclImageLandmarkSet from an ImageLandmarkSet
      @param m Tcl module 
      @param filename the file name describing a landmark set
      @param scene the scene where the set is  located 
  **/
   TclImageLandmarkSet(TclModule *m,   
		       char *filename, TclScene2D *scene); 
   /** builds a new TclImageLandmarkSet from an inherited class of ImageLandmarkSet
       @param m Tcl module 
       @param info the type of the pointed class (usually ImageLandmarkSet or an inherited class)
       @param data the structure that is pointed by this
       @param scene the scene where the set is  located 
   **/
   TclImageLandmarkSet(TclModule *m, const std::type_info& info,  
		       ImageLandmarkSet *data, TclScene2D *scene); 

   //@}

   /** @name destructor */
   //@{
   /// destroy each object in list
   virtual ~TclImageLandmarkSet();
   //@}
 protected:
  /// routine call during initialization;
   void init();
   /// initialize the selection modes
   void initSelectionModes();
 public:
   /// returns ImageLandmarkSet client
   ImageLandmarkSet *getClient(void) const 
     {
       return dynamic_cast<ImageLandmarkSet *>( client);
     }
  /** callback called when a button is moved in the slicecamera
      @param slicecam the slicecam where is located the mouse
      @param mx first coordinate of the mouse position
      @param my second coordinate of the mouse position */
  virtual void buttonMotionCallback(Camera2D *slicecam, 
				     int  mx, int my); 
  /** callback called when a button is released in the slicecamera
      @param slicecam the slicecam where is located the mouse
      @param mx first coordinate of the mouse position
      @param my second coordinate of the mouse position */
  virtual void buttonReleasedCallback(Camera2D *slicecam, 
				     int  mx, int my); 
 
  /** callback called when a button is pressed in the slicecamera
      @param slicecam the slicecam where occured the click button
      @param mx first coordinate of the pixel clicked 
      @param my second coordinate of the pixel clicked */
  void buttonPressedCallback(Camera2D *slicecam, 
			     int  mx, int my);

  /// Returns the graphic object center and weight.
   /** creates a new ImageLandmarkSet
       @param m module to create camera in
       @param argc command line number of arguments 
       @param argv command line arguments */
   static int create(TclModule *m, int argc, char **argv);
   /** build a list of all available cameras */
   static int list(TclModule *m, int argc, char **argv){
     return TclObject::list(m, argc, argv);
   }

   // return object by ref or 0 if it is not found
   static TclImageLandmarkSet *getObject(unsigned int n) {
     return dynamic_cast<TclImageLandmarkSet *>(TclObjectKind::getObject(typeid(TclImageLandmarkSet), n));
   }
   
   // return object by name (argv[i]) or 0 with an error message
   // in module interpretor if it is not found
   static TclImageLandmarkSet *getObject(int argc, char **argv, int i = 1) {
     return
       dynamic_cast<TclImageLandmarkSet *>(TclObjectKind::getObject(typeid(TclImageLandmarkSet), argc, argv, i));
   }
   /// create the tcl commands option with this object
   static void createCommands(TclModule *m);
   /// nbLandmark  TCL command option
   int nbLandmarksCmd(int argc, char **argv);
   /// add a Landmark to the landmark list
   int addLandmarkCmd(int argc, char **argv);
   /// remove a landmark to the landmark list 
   int removeLandmarkCmd(int argc, char **argv);
   /// landmark TCL command
   int landmarksCmd(int argc, char **argv);
   /// -scene TCL command option
   int sceneCmd(int argc, char **);

   /// print the different point you clicked
   int saveAsCmd(int argc, char **argv);
   /// -displayMode TCL command option
   int displayModeCmd(int argc, char **);
   /// -color TCL command option
   int colorCmd(int argc, char **);
   /// -selectedColor TCL command option
   int selectedColorCmd(int argc, char **);
   /// -destroy TCL command option
   int destroyCmd(int argc, char **);
   /// -crossSize TCL command option
   int crossSizeCmd(int argc, char **);
   /// -displayName TCL command option
   int displayNameCmd(int argc, char **);
   /// -selectedLandmark TCL command option
   int selectedLandmarkCmd(int argc, char **);
   /// -selectionMode TCL command option
   int selectionModeCmd(int argc, char **);
   /// -movingLandmark TCL command option
   int movingLandmarkCmd(int argc, char **);
   /// -buttonPressedCmd  TCL command option
   int buttonPressedCmd(int argc, char **);
   /// -buttonReleasaedCmd  TCL command option
   int buttonReleasedCmd(int argc, char **);
   /// -buttonMotionCmd  TCL command option
   int buttonMotionCmd(int argc, char **);
   /// -registerCmd  TCL command option
   int registerCmd(int argc, char **);
   /// -computeErrorCmd  TCL command option
   int computeErrorCmd(int argc, char **);
   /// -init   TCL command option
   int initCmd(int argc, char **);
   
};
/** Callback function called when the landmark selection mode is selected */
void SelectLandmarkInstallProc(TclScene2DObject *,TclCamera2D *);
/** Callback function called when the landmark creation mode is selected */
void CreateLandmarkInstallProc(TclScene2DObject *,TclCamera2D *);
/** Callback function called when the landmark motion mode is selected */
void MoveLandmarkInstallProc(TclScene2DObject *,TclCamera2D *);
/** Callback function called when the landmark deletion mode is selected */
void DeleteLandmarkInstallProc(TclScene2DObject *,TclCamera2D *);

}

#endif

