#ifndef TCLINRIMAGE4DVIEW_H
#define TCLINRIMAGE4DVIEW_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/TclInrimage4DView.h"
#include <graphics2D/TclScene2DObject.h>
#include <inrimage/Inrimage4DView.h>
#include <inrimage/TclInrimage4D.h>


namespace yav {


/** \class TclInrimage4DView
    \ingroup Inrimage
    \brief InrimageView interface */
class TclInrimage4DView : public TclScene2DObject {
 public:
  /** @name constructors */
  //@{
  /** builds object client from an inrimage then this interface
      @param m module this object is built in
      @param inr displayed inrimage
      @param scene scene to build client in 
      @param _createGLScale boolean indicating if we should create  GLScales 
  associated with each background */
  TclInrimage4DView(TclModule *m, TclInrimage4D *inr, TclScene2D *scene, 
		    bool _createGLScale=true);

 protected:
  /** builds interface for an existing client
      @param scene client scene
      @param id object dynamic type
      @param inrview client
  */
  TclInrimage4DView(TclScene2D *scene,
		    const std::type_info& id, 
		    Inrimage4DView *inrview);
  //@}

  /// constructors common code
  void init(const char *install, const char *uninstall);


 public:
  /// destroys client and interface
  virtual ~TclInrimage4DView();

  /// returns client
  Inrimage4DView *getClient() const {
    return((Inrimage4DView *)client);
  }

  /// raise ambigous list() function call
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }
  /// returns the tcl inrimage interface attached to this object
  TclInrimage4D *getInrimage() const {
    return TclInrimage4D::getInterface(getClient()->getInrimage());  }
  /// returns the tcl inrimage background interface attached to a given slicecam
  TclInrimage4DBackground *getInrimageBackground(TimeSliceCamera *cam) const {
    return  (TclInrimage4DBackground *)
      getClient()->getBackgroundFromCamera(cam); }

  /// TCL object creation function
  static int create(TclModule *m, int argc, char **argv);

  /// -destroy TCL option
  int destroyCmd(int argc, char **argv);
  /// -scene TCL option
  int sceneCmd(int argc, char **argv);
  /// -GLScale TCL option
  int GLScaleCmd(int argc, char **argv);
  /// -background TCL option
  int backgroundCmd(int argc, char **argv);
  /// -inrimage TCL option
  int inrimageCmd(int argc, char **argv);

};


} // namespace sm

#endif
