#ifndef TCLINRIMAGEROI_H
#define TCLINRIMAGEROI_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/TclInrimageROI.h"

#include <inrimage/InrimageROI.h>
#include <graphics2D/TclScene2DObject.h>


namespace yav {


/** \class TclInrimageROI
    \ingroup Inrimage
    \brief InrimageROI interface */
class TclInrimageROI : public TclScene2DObject {
 public:
  /** @name constructors */
  //@{
  /** builds object client from an inrimage then this interface
      @param m module this object is built in
      @param image displayed inrimage
      @param scene scene to build client in
      @param s a pointer to the mesh to be rasterized
      @param param the width of the band
	  @param selectedZone if -1 then raterized the whole mesh otherwise rasterize the selected surface zone
  */
  TclInrimageROI(TclScene2D *scene, Inrimage *image, 
		 SurfaceSliceable *s,
		 InrimageROI::SurfaceRoiMethod m=InrimageROI::SURFACE_BOUNDARY,  
		 double param=1.0,int selectedZone= -1);
  /** builds a region of interest from a grey-level image and a binary image
      @param scene the 2D scene where the object is located 
      @param image a 3D grey-value image where the ROI is defined : the grey-level value of each voxel is used from this original image as well as the image size
      @param binary a 3D binary image where each voxel of non-zero value is considered to belong to the region of interest. It must have the same size as the original image */
  TclInrimageROI(TclScene2D *scene,     
		 Inrimage *image, 
		 const Inrimage *binary) throw(InrimageROI::ImageMissing,
					       InrimageROI::BinaryImageWrongSize);  
  /** builds an empty  region of interest from a binary image
      @param scene the 2D scene where the object is located 
      @param image a 3D grey-value image where the ROI is defined : the grey-level value of each voxel is used from this original image as well as the image size */
  TclInrimageROI(TclScene2D *scene,     
		 Inrimage *image) throw(InrimageROI::ImageMissing);

  /** builds a region of interest from a file
      @param scene the 2D scene where the object is located 
      @param filename file name */
   TclInrimageROI(TclScene2D *scene,     
		  char *filename);
 protected:
  /** builds from an inherited classa region of interest from a file
      @param scene the 2D scene where the object is located
      @param info
      @param ir
  */
   TclInrimageROI(TclScene2D *scene,const std::type_info& info,
		  InrimageROI *ir);
  //@}
 protected:

  /// constructors common code
  void init();

 public:
  /** @name destructors */
  //@{
  /// destroys client and interface
  virtual ~TclInrimageROI();
  //@}

  /// returns client
  InrimageROI *getClient() const {
    return((InrimageROI *)client);
  }

  /// raise ambigous list() function call
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }

  /// add command options for that object
  static void createCommands(TclModule *m);
  /// TCL object creation function
  static int create(TclModule *m, int argc, char **argv);

  /// -destroy TCL option
  int destroyCmd(int argc, char **argv);
  /// -mesh TCL option
  int meshCmd(int argc, char **argv);
  /// -meanVariance TCL option
  int meanVarianceCmd(int argc, char **argv);
  /// -opacity TCL option
  int opacityCmd(int argc, char **argv);
  /// -scene TCL option
  int sceneCmd(int argc, char **argv);
  /// -greyLevelImage TCL option
  int greyLevelImageCmd(int argc, char **argv);
  /// -binaryImage TCL option
  int binaryImageCmd(int argc, char **argv);
  /// -morpho TCL option
  int morphoCmd(int argc, char **argv);
  /// -greyLevel TCL option
  int intensityImageCmd(int argc, char **argv);
  /// -histogram TCL option
  int histogramCmd(int argc, char **argv);
  /// -color TCL option
  int colorCmd(int argc, char **argv);
  /// -centerAndRadius3DCmd Tcl option
  int centerAndRadius3DCmd(int argc, char **argv);
   /// -saveOption  TCL command
  int saveOptionCmd(int argc, char **argv);
  /// -loadOption TCL command
  int loadOptionCmd(int argc, char **argv);
  /// -save TCL command
  int saveCmd(int argc, char **argv);
  /// -visibility TCL command
  int visibilityCmd(int argc, char **argv);
  /// -realName TCL command
  int realNameCmd(int argc, char **argv);
  /** load function
      @param filename Inrimage ROI file name
      @param scn 2D scene to store the ROI */
  static void load(const TclString &filename, Scene2D *scn);   
  /** callback called when a button is moved in the slicecamera
      @param slicecam the slicecam where is located the mouse
      @param mx first coordinate of the mouse position
      @param my second coordinate of the mouse position */
  virtual void buttonMotionCallback(Camera2D *slicecam, 
				      int  mx, int my){}
  /** callback called when a button is released in the slicecamera
      @param slicecam the slicecam where is located the mouse
      @param mx first coordinate of the mouse position
      @param my second coordinate of the mouse position */
  virtual void buttonReleasedCallback(Camera2D *slicecam, 
					int  mx, int my){} 
 
  /** callback called when a button is pressed in the slicecamera
      @param slicecam the slicecam where occured the click button
      @param mx first coordinate of the pixel clicked 
      @param my second coordinate of the pixel clicked */
  virtual void buttonPressedCallback(Camera2D *slicecam, 
			       int  mx, int my){}
};


} // namespace yav

#endif
