#ifndef TCLINRIMAGEVIEW_H
#define TCLINRIMAGEVIEW_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/TclInrimageView.h"
#include <graphics2D/TclScene2DObject.h>
#include <inrimage/InrimageView.h>


namespace yav {


/** \class TclInrimageView
    \ingroup Inrimage
    \brief InrimageView interface */
class TclInrimageView : public TclScene2DObject {

 public:

#ifdef LIBINRIMAGE
  /** \class InvalidDisplayMode
      \ingroup Inrimage
  */
  class InvalidDisplayMode : public std::exception {
  public:
    InvalidDisplayMode();
  };
#else
  /** \class InvalidDisplayMode
      \ingroup Inrimage
      \brief DisplayMode is not valid */
  class InvalidDisplayMode : public Exception {
  public:
    InvalidDisplayMode(const std::string& from, 
		       const std::string& error);
  };
#endif // #ifdef LIBINRIMAGE

  
  /** different DISPLAY_MODE for vectorial image only */
  enum DISPLAY_MODE {
    /// we draw the vectorial image and the vectors
    WITH_VECTOR = 1,
    /// we draw only the vectorial image
    WITHOUT_VECTOR = 2,
    /// we draw only the vectors
    VECTOR_ONLY = 3
  };

  /** different NORM_MODE to compute the norm of the vectors */
  enum NORM_MODE {
    /// the norm is proportionnal with the real norm
    LINEAR,
    /// the norm is limited with the max length depending of the LONG_MAX_MODE
    /// it avoids vectors to cross each other
    EXPONENTIAL,
    /// the norms is uniform egal to the max norm
    NORMALIZED
  };
  
  
  /// displayMode for vectorial image only (default value = WITHOUT_VECTOR)
  DISPLAY_MODE displayMode;
  /// normMode controls the way to compute the norm of the vector
  NORM_MODE normMode;
  /// undersample coefficients to display the field of vectors
  unsigned int under_sample_x;
  unsigned int under_sample_y;
  unsigned int under_sample_z;
  /// vector_width controls the appearance of the vectors
  float vector_width;
  /// vector_scale controls a general scale of the vectors
  float vector_scale; 
  /// coeff pour regler le rapport de norme dans le cas exponentiel
  float alpha; 
  /// coeff pour regler la norme maximale dans le cas exponentiel
  float beta; 
  /// if we want to draw the end of the arrow or not
  bool withArrow;
  /// tab of norme3D for vectorial image. Has to be computed just once
  double ***norme3D;
  /// tab of norme2D projected on YZ plane for vectorial image. Has to be computed just once
  double ***norme2D_X;
  /// tab of norme2D projected on XZ plane for vectorial image. Has to be computed just once
  double ***norme2D_Y;
  /// tab of norme2D projected on XY plane for vectorial image. Has to be computed just once
  double ***norme2D_Z;
  ///
  double norme3D_max;
  double norme3D_min;


  
  DISPLAY_MODE getDisplayMode(void) { return displayMode; }
  void setDisplayMode(DISPLAY_MODE mode) { displayMode = mode; }
  DISPLAY_MODE displayModeByName(const char *name) const throw (InvalidDisplayMode);
  const char *displayModeName(DISPLAY_MODE mode) const;

  NORM_MODE normModeByName(const char *name) const;
  const char *normModeName(NORM_MODE mode) const;
  
  /// raise ambigous getRef() function call
    unsigned int getRef(void) const { 
    return TclObject::getRef();
  }
  /// returns object id by command line arguments
  static const char *getTCLObjectName(void) {
    return TclObjectKind::getObjectName(typeid(TclInrimageView));
  }


 public:
  /** @name constructors */
  //@{
  /** builds object client from an inrimage then this interface
      @param m module this object is built in
      @param inr displayed inrimage
      @param scene scene to build client in 
      @param _createGLScale boolean indicating if we should create  GLScales 
  associated with each background */
  TclInrimageView(TclModule *m, TclInrimage *inr, TclScene2D *scene, 
		  bool _createGLScale=true);

 protected:
  /** builds interface for an existing client
      @param scene client scene
      @param id object dynamic type
      @param inrview client
  */
  TclInrimageView(TclScene2D *scene,
		  const std::type_info& id, InrimageView *inrview);
  //@}

  /// constructors common code
  void init(const char *install, const char *uninstall);


 public:
  /// destroys client and interface
  virtual ~TclInrimageView();

  /// returns client
  InrimageView *getClient() const {
    return((InrimageView *)client);
  }

  /// raise ambigous list() function call
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }
  /// returns the tcl inrimage interface attached to this object
  /*  TclInrimage *getInrimage() const {
      return ((Inrimage *)(getClient()->getInrimage()))->getInterface();  }*/
  /// returns the tcl inrimage background interface attached to a given slicecam
  TclInrimageBackground *getInrimageBackground(SliceCamera *cam) const {
    return  (TclInrimageBackground *)
      getClient()->getBackgroundFromCamera(cam); }

  /// TCL object creation function
  static int create(TclModule *m, int argc, char **argv);

  /// -destroy TCL option
  int destroyCmd(int argc, char **argv);
  /// -scene TCL option
  int sceneCmd(int argc, char **argv);
  /// -GLScale TCL option
  int GLScaleCmd(int argc, char **argv);
  /// -background TCL option
  int backgroundCmd(int argc, char **argv);
  /// -inrimage TCL option
  int inrimageCmd(int argc, char **argv);
  /// -displayMode TCL option
  int displayModeCmd(int argc, char **argv);
  /// -vectorScale TCL option
  int vectorScaleCmd(int argc, char **argv); 
  /// -vectorWidth TCL option
  int vectorWidthCmd(int argc, char **argv); 
  /// -arrow TCL option
  int arrowCmd(int argc, char **argv); 

 protected:
	   /// function which initialize the 4 tab of norms
  template <class T> void computeNorms(T *type){
  T *buffer = (T *) (getClient()->getInrimage()->getData());
  double x=0.0,y=0.0,z=0.0;
  Inrimage *inr = getClient()->getInrimage();
  unsigned int vdim=inr->getVdim();
  if (inr->getVectorialMode()==VM_INTERLACED) {
	  
	  if (vdim==3) {
		  norme3D_min = sqrt((double)(buffer[0]*buffer[0]+buffer[1]*buffer[1]+buffer[2]*buffer[2]));
	  } else {
		  norme3D_min = sqrt((double)(buffer[0]*buffer[0]+buffer[1]*buffer[1]));
	  }
	  norme3D_max = 0.0;

	  /// Allocation des tab
	  norme3D = (double ***) (new double**[inr->getZ()]);
	  norme2D_X = (double ***) new double**[inr->getZ()];
	  norme2D_Y = (double ***) new double**[inr->getZ()];
	  norme2D_Z = (double ***) new double**[inr->getZ()];

	  for (unsigned int k=0; k<inr->getZ(); ++k){
		  norme3D[k] = (double **) new double*[inr->getY()]; 
		  norme2D_X[k] = (double **) new double*[inr->getY()];
		  norme2D_Y[k] = (double **) new double*[inr->getY()];
		  norme2D_Z[k] = (double **) new double*[inr->getY()];
		  for (unsigned int j=0; j<inr->getY(); ++j){
			  norme3D[k][j] = (double *) new double[inr->getX()]; 
			  norme2D_X[k][j] = (double *) new double[inr->getX()];
			  norme2D_Y[k][j] = (double *) new double[inr->getX()];
			  norme2D_Z[k][j] = (double *) new double[inr->getX()];
			  for (unsigned int i=0; i<inr->getX(); ++i){
				  x = buffer[k*inr->getX()*inr->getY()*inr->getVdim() + j*inr->getX()*inr->getVdim() + i*inr->getVdim()];
				  y = buffer[k*inr->getX()*inr->getY()*inr->getVdim() + j*inr->getX()*inr->getVdim() + i*inr->getVdim()+1];
				  if (vdim==3) {
					  z = buffer[k*inr->getX()*inr->getY()*inr->getVdim() + j*inr->getX()*inr->getVdim() + i*inr->getVdim()+2];
					  norme3D[k][j][i] = sqrt(x*x+y*y+z*z);
					  norme2D_X[k][j][i] = sqrt(y*y+z*z);
					  norme2D_Y[k][j][i] = sqrt(x*x+z*z);
					  norme2D_Z[k][j][i] = sqrt(x*x+y*y);
				  } else {
					  norme3D[k][j][i] = sqrt(x*x+y*y);
					  norme2D_X[k][j][i] =  fabs(y);
					  norme2D_Y[k][j][i] = fabs(x);
					  norme2D_Z[k][j][i] = norme3D[k][j][i];
				  }
				  if (norme3D[k][j][i]>norme3D_max)
					  norme3D_max=norme3D[k][j][i];
				  if (norme3D[k][j][i]<norme3D_min)
					  norme3D_min=norme3D[k][j][i];
			  }
		  }
	  }
  } else {
	  unsigned int componentOffset=inr->getX()*inr->getY()*inr->getZ();
	  if (getClient()->getInrimage()->getVdim()==3) {
		  norme3D_min = sqrt((double)(buffer[0]*buffer[0]+buffer[componentOffset]*buffer[componentOffset]+buffer[2*componentOffset]*buffer[2*componentOffset]));
	  } else {
		  norme3D_min = sqrt((double)(buffer[0]*buffer[0]+buffer[componentOffset]*buffer[componentOffset]));
	  }
	  norme3D_max = 0.0;

	  /// Allocation des tab
	  norme3D = (double ***) (new double**[inr->getZ()]);
	  norme2D_X = (double ***) new double**[inr->getZ()];
	  norme2D_Y = (double ***) new double**[inr->getZ()];
	  norme2D_Z = (double ***) new double**[inr->getZ()];

	  for (unsigned int k=0; k<inr->getZ(); ++k){
		  norme3D[k] = (double **) new double*[inr->getY()]; 
		  norme2D_X[k] = (double **) new double*[inr->getY()];
		  norme2D_Y[k] = (double **) new double*[inr->getY()];
		  norme2D_Z[k] = (double **) new double*[inr->getY()];
		  for (unsigned int j=0; j<inr->getY(); ++j){
			  norme3D[k][j] = (double *) new double[inr->getX()]; 
			  norme2D_X[k][j] = (double *) new double[inr->getX()];
			  norme2D_Y[k][j] = (double *) new double[inr->getX()];
			  norme2D_Z[k][j] = (double *) new double[inr->getX()];
			  for (unsigned int i=0; i<inr->getX(); ++i){
				  x = buffer[k*inr->getX()*inr->getY() + j*inr->getX() + i];
				  y = buffer[componentOffset+k*inr->getX()*inr->getY() + j*inr->getX() + i];
				  if (getClient()->getInrimage()->getVdim()==3) {
					  z = buffer[2*componentOffset+k*inr->getX()*inr->getY() + j*inr->getX() + i];
					  norme3D[k][j][i] = sqrt(x*x+y*y+z*z);
					  norme2D_X[k][j][i] = sqrt(y*y+z*z);
					  norme2D_Y[k][j][i] = sqrt(x*x+z*z);
					  norme2D_Z[k][j][i] = sqrt(x*x+y*y);
				  } else {
					  norme3D[k][j][i] = sqrt(x*x+y*y);
					  norme2D_X[k][j][i] =  fabs(y);
					  norme2D_Y[k][j][i] = fabs(x);
					  norme2D_Z[k][j][i] = norme3D[k][j][i];
				  }
				  if (norme3D[k][j][i]>norme3D_max)
					  norme3D_max=norme3D[k][j][i];
				  if (norme3D[k][j][i]<norme3D_min)
					  norme3D_min=norme3D[k][j][i];
			  }
		  }
	  }
  }
 } 
};



} // namespace yav


#endif
