#ifndef TCLOVERLAY4DBACKGROUND_H
#define TCLOVERLAY4DBACKGROUND_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/TclOverlay4DBackground.h"

#include <inrimage/TclOverlayBackground.h>
#include <inrimage/TclInrimage4DBackground.h>
#include <inrimage/OverlayInrimage4DView.h>
#include <inrimage/OpacityScale.h>

namespace yav {

/** \class TclOverlay4DBackground
    \ingroup Inrimage
    \brief The background of a slicecamera that is linked with an OverlayInrimageView*/
class TclOverlay4DBackground : public TclInrimage4DBackground, 
			     public Owner<OpacityScale> {
 protected:
  /// second Tcl minimim scale for grey-level windowing
  ObjectPointer<MinGLScale> secondMinScale; 
  /// second Tcl maximim scale for grey-level windowing
  ObjectPointer<MaxGLScale> secondMaxScale; 
  /// min and max GL values for inrimage
  Voxel *secondGlmin, *secondGlmax;
    /// the red component colormap for the display of the second image
  const unsigned char *redColorMap2;
  /// the green component colormap for the display of the second image
  const unsigned char *greenColorMap2;
  /// the blue component colormap that depends on transfo
  const unsigned char *blueColorMap2;
  /// grey level transformation type for the 2nd image
  GLTRANSFO transfo2;
 public:
  /** builds a Tcl object corresponding to  photo data block attached to 
      an inrimage in a given slice camera
      @param m module to create TCL object
      @param inrv an inrimage view attached to an Inrimage
      @param cam given slice camera
      @param useMinGLScale if a  TclScale controlling the minimum grey-level windowing value should be used
      @param minScale a TclScale controlling the minimum windowing value
      @param useMaxGLScale if a  TclScale controlling the maximum grey-level windowing value should be used
      @param maxScale a TclScale controlling the maximum windowing value 
      @param useSecondMinGLScale if a  TclScale controlling the minimum grey-level windowing of the second image should be created if secondMinScale==0
      @param secondMinScale a TclScale controlling the minimum windowing value of the second image
      @param useSecondMaxGLScale if a  MaxScale controlling the maximum grey-level windowing value of the second image should be created if secondMaxScale==0
      @param secondMaxScale a TclScale controlling the maximum windowing value of the second image 
      @param color color of the widget 
      @param useOpacityScale if a  OpacityScale controlling the blending between the 2 images should be created if opacityScale==0
      @param opacityScale a OpacityScale controlling the blending between the 2 images  */
  TclOverlay4DBackground(TclModule *m,
                         OverlayInrimage4DView *inrv,
                         TimeSliceCamera *cam,
                         bool useMinGLScale = true,
                         MinGLScale *minScale = 0,
                         bool useMaxGLScale = true,
                         MaxGLScale *maxScale = 0, 
                         bool useSecondMinGLScale=true,
                         MinGLScale *secondMinScale = 0,
                         bool useSecondMaxGLScale=true,
                         MaxGLScale *secondMaxScale = 0,
                         bool useOpacityScale=true,
                         OpacityScale *opacityScale=0,
                         const char *color = "#E8E0D2" );
  /// destructor: free buffers
  virtual ~TclOverlay4DBackground();

  /** scale value changed
      @param scale calling scale object
      @param val new scale value */
  virtual void scaleValue(const TclScale *scale, double val);
  /// add a minimum grey-level scale to the second image
  void addSecondMinGLScale(MinGLScale *min);
  /// add a maximum grey-level scale to the second image
  void addSecondMaxGLScale(MaxGLScale *max);
  /// resize GL window and redraw
  virtual void imageGLChanged(void);
    /// image type changed
  void inrimageChanged(void);
  /// return second  inrimage
  Inrimage4D *getSecondInrimage(void) const;
   /// returns grey-level transformation mode
  GLTRANSFO getGLTransfo2(void) const { return transfo2; }
  /** set grey-level transformation mode
      @param t new transformation mode
      @param red, green, blue */
  void setGLTransfo2(GLTRANSFO t,const unsigned char *red=0,
		     const  unsigned char *green=0,const  unsigned char *blue=0);
  /// set opacity value 
  void setOpacity(double val);
    /// set min voxel value to val
  void setSecondMinThreshold(double val) throw(Voxel::InvalidValue);
  
  /// set max voxel value to val
  void setSecondMaxThreshold(double val) throw(Voxel::InvalidValue);
  
  /// set min and max voxel value to min and max
  void setSecondGlThresholds(double min, double max) throw(Voxel::InvalidValue);
  
  /// max grey level scale owner callback
  virtual void dying(const ObjectPointer<OpacityScale>& obj) {}
  
  /// set min and max voxel value to min and max
  void setSecondGlThresholds(const TclString& min, const TclString& max)
    throw(Voxel::InvalidValue); 
  /// return mn voxel value
  virtual double getMin(const TclScale *scale) const;

  /// return max voxel value
  virtual double getMax(const TclScale *scale) const;
    
  /** redraw camera image */
  virtual void redraw();
  /// return the label coresponding to pixel (x,y)
  virtual TclString getLabel(int &x,int &y);
  /// return the second min scale associated with the second image
  MinGLScale *getSecondMinScale() const {
    return secondMinScale();}
  /// return the second max scale associated with the second image
  MaxGLScale *getSecondMaxScale() const {
    return secondMaxScale();}
  /// return the opacity scale associated with the background
  OpacityScale *getOpacityScale() const {
    return Owner<OpacityScale>::pointer;}
  /// return the overlay view that has created this background
  OverlayInrimage4DView *getOverlay4DView() const {
    return ( OverlayInrimage4DView *) inrv;
  } 
  /// return opacity scale value
  double getOpacity() const {
    if (Owner<OpacityScale>::pointer)
      return Owner<OpacityScale>::pointer->get();
    else
      return 0.0;
  }
  /// compute inrimage grey levels bounds
  virtual void glBounds(void);
  /// new TCL command callback
  static int create(TclModule *m, int argc, char **argv);
  /// TCL -opacity command callback
  int opacityCmd(int argc, char **argv);
  /// TCL -scalemode2 command callback
  int scalemode2Cmd(int argc, char **argv);
  /// TCL -greyLevel command callback
  int greyLevelCmd(int argc, char **argv);

};


} // namespace yav

#endif
