#ifndef TCLTRANSFORMEDINRIMAGEBACKGROUND_H
#define TCLTRANSFORMEDINRIMAGEBACKGROUND_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/TclTransformedInrimageBackground.h"
#include <inrimage/TclInrimageBackground.h>
#include <inrimage/TransformedInrimageView.h>

namespace yav {

/** \class TclTransformedInrimageBackground
    \ingroup Inrimage
    \brief The background of a slicecamera that is linked with an TransformedInrimageView*/
class TclTransformedInrimageBackground : public TclInrimageBackground   {
 protected:
  /// the 2D image created after rotating and translating the image
  Inrimage *resampledSlice;
  /// boolean used in redraw() that allows to avoid resample the image when only a windowing is performed
  bool resampleImage;
 public:
  /** builds a Tcl object corresponding to  photo data block attached to 
      an inrimage in a given slice camera
      @param m module to create TCL object
      @param inrv an inrimage view attached to an Inrimage
      @param cam slice camera to display image in 
      @param useMinGLScale if a  TclScale controlling the minimum grey-level windowing value should be used
      @param minScale a TclScale controlling the minimum windowing value
      @param useMaxGLScale if a  TclScale controlling the maximum grey-level windowing value should be used
      @param maxScale a TclScale controlling the maximum windowing value 
      @param color color of the widget 
  */
  TclTransformedInrimageBackground(TclModule *m,
				   TransformedInrimageView *inrv,
                                   SliceCamera *cam,
				   bool useMinGLScale = true,
				   MinGLScale *minScale = 0,
				   bool useMaxGLScale = true,
				   MaxGLScale *maxScale = 0,
				   const char *color = "#E8E0D2" );
  /// destructor: free buffers
  virtual ~TclTransformedInrimageBackground();

  /** scale value changed
      @param scale calling scale object
      @param val new scale value */
  /// \code virtual void scaleValue(const TclScale *scale, double val); \endcode
  /// resize GL window and redraw
  virtual void imageGLChanged(void);
  /// image type changed
  void inrimageChanged(void);
  
  /// create resampled slice based on the 4x4 transformation matrix stored in the TransformedInrimageView
  void createResampledSlice();
    
  /** redraw camera image */
  virtual void redraw();
  /// return the label coresponding to pixel (x,y)
  virtual TclString getLabel(int &x,int &y);
  /// return the transformedInrimage view that has created this background
  TransformedInrimageView *getTransformedInrimageView() const {
    return ( TransformedInrimageView *) inrv;
  } 
  /// new TCL command callback
  static int create(TclModule *m, int argc, char **argv);

};

} // namespace yav

#endif
