#ifndef TRANSFORMEDINRIMAGEVIEW_H
#define TRANSFORMEDINRIMAGEVIEW_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/TransformedInrimageView.h"
#include <inrimage/InrimageView.h>

namespace yav {


#ifdef TCL_INTERFACE
class TclTransformedInrimageView;
#endif


/** \class TransformedInrimageView
    \ingroup Inrimage
    \brief A viewer that allow for interaction rotation and translation */
  class TransformedInrimageView : public InrimageView 
  {
 public:
  /** @name Exception classes */
  //@{
  /** \class NoImage
      \ingroup Inrimage
      \brief empty inrimage  */
  class NoImage : public Exception {
  public:
    NoImage(const std::string& from);
  };
  //@}
 protected:
  /// second intensity image
  Matrix4x4<double> transformation;
  //the previous pixel position when interactively dragging the image
  Vec3<unsigned int> previousVoxelPosition;
  // the previous pixel position for rotating the image
  Vec2<double> previousPixelPosition;
  // the current Slicecam where interaction takes place
  SliceCamera *currentSlicecam;
  // the current Slicecam where rotation takes place
  SliceCamera *rotationCurrentSlicecam;
  // the current Slicecam where scale takes place
  SliceCamera *scaleCurrentSlicecam;
 public:
    
  /** @name constructors */
  //@{
  /** Normal constructor by specifiying 1 inrimage 
      @param inr1 first inrimage to draw
      @param scene  scene to build data in 
      @param _createGLScale boolean indicating if we should create  GLScales 
      associated with each background */
  TransformedInrimageView(Inrimage *inr1, 
			  Scene2D *scene,bool _createGLScale=true) 
    throw(NoImage);
  //@}

  /// destruction: do not delete visualized inrimage
  virtual ~TransformedInrimageView() {}


  virtual void mouseButtonPressed(CameraTypes::WHICH_BUTTON button,
				  CameraTypes::KEY_MODIFIER modifier,
				  Camera2D *cam,int x,int y);
  virtual void mouseButtonReleased(CameraTypes::WHICH_BUTTON button,
				  CameraTypes::KEY_MODIFIER modifier,
				  Camera2D *cam,int x,int y);
  virtual void mouseButtonMoved(Camera2D *cam,int x,int y);

  /// add a camera in inrimage view management
  virtual void doCreateCamera(SliceCamera& cam);

   /** callback called when a button is pressed in the slicecamera
      @param slicecam the slicecam where occured the click button
      @param mx first coordinate of the pixel clicked 
      @param my second coordinate of the pixel clicked */
  virtual void buttonPressedCallback(SliceCamera *slicecam, 
				     int  mx, int my);
  /** callback called when a button is moved in the slicecamera
      @param slicecam the slicecam where is located the mouse
      @param mx first coordinate of the mouse position
      @param my second coordinate of the mouse position */
  virtual void buttonMotionCallback(SliceCamera *slicecam, 
				     int  mx, int my);
  /** callback called when a button is released in the slicecamera
      @param slicecam the slicecam where is located the mouse
      @param mx first coordinate of the mouse position
      @param my second coordinate of the mouse position */
  virtual void buttonReleasedCallback(SliceCamera *slicecam, 
				     int  mx, int my);
 /** callback called when a shift + button is pressed in the slicecamera
      @param slicecam the slicecam where occured the click button
      @param mx first coordinate of the pixel clicked 
      @param my second coordinate of the pixel clicked */
  virtual void shiftButtonPressedCallback(SliceCamera *slicecam, 
				     int  mx, int my);
  /** callback called when a shift + button is moved in the slicecamera
      @param slicecam the slicecam where is located the mouse
      @param mx first coordinate of the mouse position
      @param my second coordinate of the mouse position */
  virtual void shiftButtonMotionCallback(SliceCamera *slicecam, 
				     int  mx, int my);
  /** callback called when a shift + button is released in the slicecamera
      @param slicecam the slicecam where is located the mouse
      @param mx first coordinate of the mouse position
      @param my second coordinate of the mouse position */
  virtual void shiftButtonReleasedCallback(SliceCamera *slicecam, 
					   int  mx, int my);
   /** Modifies the first inrimage and updates all slices */
  virtual void changeInrimage(Inrimage *inr);

  /** returns the current matrix transformation */
  Matrix4x4<double> &getTransformation() {
    return transformation;
  }
  /** set the current matrix transformation */
  void  setTransformation(Matrix4x4<double> &m) {
    transformation=m;
  }

  friend class TclTransformedInrimageView;
  friend class TclTransformedInrimageBackground;
  friend class TclInrimage;
};


} // namespace yav


#endif

