#ifdef _MSC_VER 
#pragma warning ( disable : 4068 4786 4081 ) 
#endif 
 
#ifndef VOXEL_ITERATOR_C
#define VOXEL_ITERATOR_C

/* #define INRIMAGE_DEBUG */

template <class T> VoxelIteratorScalar<T>::VoxelIteratorScalar(Inrimage *im) throw(VoxelIteratorObject::MissingInrimage,VoxelIteratorObject::WrongImageType): VoxelIteratorObject(im), Owner<Inrimage>(im) {
  buffer= (T *) im->getData();
}
template <class T> VoxelIteratorScalar<T>::VoxelIteratorScalar(const VoxelIteratorObject &it) throw (VoxelIteratorObject::OutOfRange,VoxelIteratorObject::WrongImageType):VoxelIteratorObject(it), Owner<Inrimage>(it.getInrimage()){
  buffer= ((T*) getInrimage()->getData())+index;
}

template <class T> void VoxelIteratorScalar<T>::operator++(int) throw (VoxelIteratorObject::OutOfRange) {
#ifdef INRIMAGE_DEBUG
  if (index>=maxIndex) throw(VoxelIteratorObject::OutOfRange("void VoxelIteratorScalar<T>::operator++(int)"));
#endif
  buffer++;
  index++;
}

template <class T> void VoxelIteratorScalar<T>::operator++(void) throw (VoxelIteratorObject::OutOfRange) {
#ifdef INRIMAGE_DEBUG
  if (index>=maxIndex) throw(VoxelIteratorObject::OutOfRange("void VoxelIteratorScalar<T>::operator++(void)"));
#endif
  buffer++;
  index++;
}

template <class T> void VoxelIteratorScalar<T>::operator--(int) throw (VoxelIteratorObject::OutOfRange) {
#ifdef INRIMAGE_DEBUG
  if (index<0) throw(VoxelIteratorObject::OutOfRange("void VoxelIteratorScalar<T>::operator--(int)"));
#endif
  buffer--;
  index--;
}


template <class T> void VoxelIteratorScalar<T>::operator--(void) throw (VoxelIteratorObject::OutOfRange) {
#ifdef INRIMAGE_DEBUG
  if (index<0) throw(VoxelIteratorObject::OutOfRange("void VoxelIteratorScalar<T>::operator--(void)"));
#endif
  buffer--;
  index--;
}
template <class T> double VoxelIteratorScalar<T>::operator*() const{
  return (double) (*buffer);
}

template <class T> LONGINT VoxelIteratorScalar<T>::operator()(void) const{
  return (LONGINT) (*buffer);
}
template <class T> void VoxelIteratorScalar<T>::setValueFromDouble(const double v) {
  (*buffer)=(T) v;
}
template <class T> void VoxelIteratorScalar<T>::setValueFromLong(const long int v)  {
  (*buffer)=(T) v;
}

template <class T> void VoxelIteratorScalar<T>::first(void) {
  index=0;
  buffer= (T*) getInrimage()->getData();
}

template <class T> void VoxelIteratorScalar<T>::last(void) {
  index=maxIndex-1;
  buffer= ((T*) getInrimage()->getData())+index;
}
template <class T> VoxelIteratorObject&  VoxelIteratorScalar<T>::operator=(const VoxelIteratorObject& v) {
  if (v.getPosition()<maxIndex) {
    index=v.getPosition();
    buffer= ((T*) getInrimage()->getData())+index;
  }
  return *this;
}


template <class T> void VoxelIteratorScalar<T>::setPosition(const long int ind) throw(VoxelIteratorObject::OutOfRange) {
  if ((ind>=maxIndex) || (ind<0)) {
    throw VoxelIteratorObject::OutOfRange("VoxelIteratorScalar<T>::setPosition(const long int ind)");
  } else {
    index=ind;
    buffer= ((T*) getInrimage()->getData())+index;
  }
}

template <class T> void VoxelIteratorScalar<T>::setPosition(const unsigned int x,const unsigned int y,const unsigned int z) throw(VoxelIteratorObject::OutOfRange) {
  Inrimage *inr=getInrimage();
  long int ind=z*inr->getX()*inr->getVdim()*inr->getY()+y *inr->getX()*inr->getVdim()+x;
  if ((ind>=maxIndex) || (ind<0)) {
    throw VoxelIteratorObject::OutOfRange("VoxelIteratorScalar<T>::setPosition(const long int ind)");
  } else {
    index=ind;
    buffer= ((T*) getInrimage()->getData())+index;
  }
}



#endif
