#ifndef ARITHMETIC_H
#define ARITHMETIC_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/arithmetic.h"

#include <inrimage/Inrimage.h>

namespace yav {


  /** set an image to a value */
  void setValue(Inrimage *inr, double val);
  /** negate image (negate each voxel). Does not oper on unsigned
      types. The negation of each vector component is computed for
      vectorial images. */
  void inv(Inrimage *inr);
  /** compute image absolute value (the absolute value of each
      vector component is computed for vectorial images) */
  void abs(Inrimage *inr);
  /** adds a constant value to image voxels
      @param inr
      @param val value to add */
  void add(Inrimage *inr, double val);
    /** substractss a constant value from image voxels
        @param inr
        @param val value to substract */
  void sub(Inrimage *inr, double val);
  /** multiplicates image voxels by a constant value
      @param inr
      @param val value to multiuplicate */
  void mul(Inrimage *inr, double val);
  /** divides image voxels by a constant value
      @param inr
      @param val value to divide */
  void div(Inrimage *inr, double val);
  /** voxel by voxel adds inr to this
      @param inr inrimage to add
      @param from 
      @exception InvalidArguments inr size does not match this size */
  void add(Inrimage *inr, const Inrimage *from) throw(Inrimage::InvalidArguments);
  void add(Inrimage *inr, const Vec3<double>& v) throw(Inrimage::InvalidArguments);
  /** voxel by voxel substracts inr to this
      @param inr inrimage to substract
      @param from
      @exception InvalidArguments inr size does not match this size */
  void sub(Inrimage *inr, const Inrimage *from) throw(Inrimage::InvalidArguments);
  /** voxel by voxel multiplicates inr to this
      @param inr inrimage to multiplicate
      @param from
      @exception InvalidArguments inr size does not match this size */
  void mul(Inrimage *inr, const Inrimage *from) throw(Inrimage::InvalidArguments);
  void mul(Inrimage *inr, const Matrix3x3<double>& m) throw(Inrimage::InvalidArguments); 
  /** voxel by voxel divides inr to this
      @param inr inrimage to divide
      @param from
      @exception InvalidArguments inr size does not match this size */
  void div(Inrimage *inr, const Inrimage *from) throw(Inrimage::InvalidArguments);

  /// logical not operator (this = !this)
  void logicalNot(Inrimage *inr);
  /** logical or operator (this |= inr)
      @param inr inrimage to OR
      @param from
      @exception InvalidArguments inr size does not match this size or inr type does not match this type */
  void logicalOr(Inrimage *inr, const Inrimage *from) throw(Inrimage::InvalidArguments);
  /** logical and operator (this &= inr)
      @param inr inrimage to AND
      @param from
      @exception InvalidArguments inr size does not match this size or inr type does not match this type */
  void logicalAnd(Inrimage *inr, const Inrimage *from) throw(Inrimage::InvalidArguments);
  /** logical xor operator (this ^= inr)
      @param inr inrimage to XOR
      @param from 
      @exception InvalidArguments inr size does not match this size or inr type does not match this type */
  void logicalXor(Inrimage *inr, const Inrimage *from) throw(Inrimage::InvalidArguments);








}




#endif
