#ifndef _SLICEIMAGE_H 
#define _SLICEIMAGE_H 
 
#ifdef _MSC_VER 
#pragma warning ( disable : 4068 4786 4081 ) 
#endif 
 
#include <inrimage/Inrimage.h>
#include <framework/SliceOrientation.h>
#include <basics/Probability.h>
namespace yav { 


/** output a 3D image output whose intensity ouput(x) is defined as output(x)=pdf(input(x)) where
pdf(x) is a probability density function. The output image is of type double and has the same
size as the input image
@param inr the input image (any type)
@param pdf a probability density function that returns the probability of a given intensity value 
@return  the probability map of the input image */
Inrimage *createProbabilityImage( Inrimage *inr,const ProbabilityDensityFunction *pdf);

/** output a 2D image corresponding to a slice of a 3D image. If no colormap is
    provided the output image is of type unsigned char  and is RGB otherwise
    @param input the input image (usually 3D)
    @param orientation the orientation of the slice that should be outputed
    @param sliceNumber the index  (starting from 0) of the slice 
    @param scaleFactor the scale factor for the image size : 1.0 corresponds to its natural size (eventually using voxel size)
    @param useVoxelSize whether to consider the voxel size or not
    @param minWindow if windowing applies, corresponds to the minimum grey level value
    @param maxWindow if windowing applies, corresponds to the maximum grey level value. if (minWindow==maxWindow) then the minimum and maximum grey level value of the whole image is considered
    @param interpolate if true then trilinear interpolation is performed. Otherwise nearest neighbor interpolation is performed
    @param colormap should be an array of size unsigned char col[3][257]. Thus col[0][i] is the red component of the grey level i. col[0][256] is the color used to draw pixel whose intensity is greater than maxWindow.If colormap==0 then the output image is a grey-level image and linear windowing is applied*/
Inrimage *createSliceImage(const Inrimage *input,
			   SLICE_ORIENTATION orientation=SLICE_Z,
			   unsigned int sliceNumber=0,
			   double scaleFactor=1.0,
			   bool useVoxelSize=true,
			   double minWindow=0,
			   double maxWindow=0,
			   bool interpolate=false,
			   unsigned char **colormap=0); 

} // namespace yav 
 
#endif 
