#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#include <inrimage/Inrimage.h>
#include <inrimage/Interpolator.h>


namespace yav {


  /// buffer undersampling

  template <class T>
	  void underSampleScalar(T ww, Inrimage &inr, unsigned int sx, unsigned int sy, unsigned int sz) {
  if(inr.getVectorialMode() == VM_NON_INTERLACED) return ; 
  unsigned int x, y, z, nx, ny, nz, i, j, k, v, size, plane; 
  unsigned int vdim = inr.getVdim();
  T *d; 
  if(sx == 0) sx = 1; 
  if(sy == 0) sy = 1; 
  if(sz == 0) sz = 1; 
  nx = inr.getX() / sx; 
  if(nx < 1) nx = 1; 
  ny = inr.getY() / sy; 
  if(ny < 1) ny = 1; 
  nz = inr.getZ() / sz; 
  if(nz < 1) nz = 1; 

  plane = inr.getX() * inr.getY() * vdim; 
  size = nx * ny * nz * vdim; 
  double *nd = new double[size]; 
  double *ptr = nd; 
  memset(nd, 0, size * sizeof(double)); 
  for(z = 0; z < nz; z++) { for(k = 0; k < sz; k++) { 
    for(y = 0; y < ny; y++) { 
      for(j = 0; j < sy; j++) { 
       for(x = 0; x < nx; x++) { 
	    d = (T *) inr.getData() + (sz * z + k) * plane + 
	      (sy * y + j) * inr.getX() * vdim + x * sx * vdim; 
           for(v = 0; v < vdim; v++, ptr++) { 
	      for(i = 0; i < sx; i++, d += vdim) *ptr += (double) *d; 
	      d -= vdim * sx - 1; 
		   }  
	   } 
	  ptr -= nx * vdim; 
	  } 
	ptr += nx * vdim; 
	} 
   ptr -= nx * ny * vdim; 
  } 
  ptr += nx * ny * vdim; 
  } 
 
  T *ndd = new T[size]; 
  double m = sx * sy * sz; 
  for(i = 0; i < size; i++) *ndd++ = (T) (*nd++ / m); 
  ndd -= size;   nd -= size; 

  delete [] nd; 
  inr.setX(nx); 
  inr.setY(ny); 
  inr.setZ(nz); 
  inr.setVoxelSize(inr.getVX() * sx, inr.getVY() * sy, inr.getVZ() * sz); 
  delete[] (unsigned char *) inr.setData(ndd);  
  }





#define UNDERSAMPLESCALAR(TYPE,INR,SX,SY,SZ) underSampleScalar((TYPE)0,INR,SX,SY,SZ)

 
void Inrimage::underSample(unsigned int sx, unsigned int sy, unsigned int sz) {
  switch(type) {
  case WT_UNSIGNED_CHAR:
    UNDERSAMPLESCALAR(unsigned char,*this,sx, sy, sz);
    break;
  case WT_SIGNED_CHAR:
    UNDERSAMPLESCALAR(char,*this,sx, sy, sz);
    break;
  case WT_UNSIGNED_SHORT:
    UNDERSAMPLESCALAR(unsigned short,*this,sx, sy, sz);
    break;
  case WT_SIGNED_SHORT:
    UNDERSAMPLESCALAR(short,*this,sx, sy, sz);
    break;
  case WT_UNSIGNED_INT:
    UNDERSAMPLESCALAR(unsigned int,*this,sx, sy, sz);
    break;
  case WT_SIGNED_INT:
    UNDERSAMPLESCALAR(int,*this,sx, sy, sz);
    break;
  case WT_UNSIGNED_LONG:
    UNDERSAMPLESCALAR(unsigned LONGINT,*this,sx, sy, sz);
    break;
  case WT_SIGNED_LONG:
    UNDERSAMPLESCALAR(LONGINT,*this,sx, sy, sz);
    break;
  case WT_FLOAT:
    UNDERSAMPLESCALAR(float,*this,sx, sy, sz);
    break;
  case WT_DOUBLE:
    UNDERSAMPLESCALAR(double,*this,sx, sy, sz);
    break;
  case WT_RGB:
    UNDERSAMPLESCALAR(unsigned char,*this,sx, sy, sz);
    break;
  case WT_RGBA:
    UNDERSAMPLESCALAR(unsigned char,*this,sx, sy, sz);
    break;
  case WT_FLOAT_VECTOR:
    UNDERSAMPLESCALAR(float,*this,sx, sy, sz);
    break;
  case WT_DOUBLE_VECTOR:
    UNDERSAMPLESCALAR(double,*this,sx, sy, sz);
    break;
  default:
    break;
  }
}



void Inrimage::overSample(unsigned int sx, unsigned int sy, unsigned int sz)
  throw(Inrimage::InvalidSize) {

  if(sx < xdim || sy < ydim || sz < zdim)
    throw InvalidSize("Inrimage::overSample");
  resample(sx, sy, sz);
}

void Inrimage::resample(unsigned int sx, unsigned int sy, unsigned int sz) {
  Inrimage *inr = new Inrimage(sx, sy, sz, type, vdim, VM_INTERLACED,
			 (float)(  vx * (double) xdim / (double) sx),
			 (float)(  vy * (double) ydim / (double) sy),
			 (float)(  vz * (double) zdim / (double) sz));
  Voxel *voxel = Voxel::voxelByType(type);
  double dx = xdim * vx / (double) sx;
  double dy = ydim * vy / (double) sy;
  double dz = zdim * vz / (double) sz;
  double rx, ry, rz;
  unsigned int x, y, z;

//   for(z = 0, rz = -0.5 * dz; z < sz; z++, rz += dz) {
//     for(y = 0, ry = -0.5 * dy; y < sy; y++, ry += dy) {
//       for(x = 0, rx = -0.5 * dx; x < sx; x++, rx += dx) {
// 	inr->setVoxel(x, y, z, (*interpolator)(*this, rx, ry, rz, *voxel));
//       }
//     }
//   }
  for(z = 0, rz = 0; z < sz; z++, rz += dz) {
    for(y = 0, ry = 0; y < sy; y++, ry += dy) {
      for(x = 0, rx = 0; x < sx; x++, rx += dx) {
	inr->setVoxel(x, y, z, (*interpolator)(*this, rx, ry, rz, *voxel));
      }
    }
  }
  delete voxel;

  
  xdim = sx;
  ydim = sy;
  zdim = sz;
  vx = inr->getVX();
  vy = inr->getVY();
  vz = inr->getVZ();
  delete[] (unsigned char *) setData(inr->getData());
  inr->setData(0);
  delete inr;
}


} // namespace yav



