#ifndef PREDEFINEDMASKS_H
#define PREDEFINEDMASKS_H

#include <iostream>

/* This file contains pre-defined coefficients for masks of different sizes and dimensions.
   To cope with the notion of shape-based interpolation ('Shape-based Interpolation' G.T.HERMAN & al.),
   coefficients for initialization masks are given (initialization coefficients are smaller than the other ones)
*/

namespace yav {
  
  class preDefinedMasks
  {
    public:
      preDefinedMasks();
      ~preDefinedMasks();
    
      // coefficients for 2D 3x3 chamfer mask. 
      unsigned char ** chamfer2D3x3;
      // coefficients for 2D 3x3 initialization chamfer mask, corresponding to the previous one. 
      unsigned char ** chamfer2D3x3Init;
      float epsilon2D3x3;
      float maxRelErr2D3x3;
      
      // coefficients for 2D 5x5 chamfer mask. 
      unsigned char ** chamfer2D5x5;
      // coefficients for 2D 5x5 initialization chamfer mask, corresponding to the previous one.
      unsigned char ** chamfer2D5x5Init;
      float epsilon2D5x5;
      float maxRelErr2D5x5;
    
      // coefficients for 2D 7x7 chamfer mask. 
      unsigned char ** chamfer2D7x7;
      unsigned char ** chamfer2D7x7Init;
      float epsilon2D7x7;
      float maxRelErr2D7x7;
      
      
      // coefficients for 3D 3x3x3 chamfer mask. 
      unsigned char *** chamfer3D3x3x3;
      // coefficients for 3D 3x3x3 initialization chamfer mask, corresponding to the previous one.
      unsigned char *** chamfer3D3x3x3Init;
      float epsilon3D3x3x3;
      float maxRelErr3D3x3x3;
      
      // coefficients for 3D 5x5x5 chamfer mask. 
      unsigned char *** chamfer3D5x5x5;
      // coefficients for 3D 5x5x5 initialization chamfer mask, corresponding to the previous one.
      unsigned char *** chamfer3D5x5x5Init;
      float epsilon3D5x5x5;
      float maxRelErr3D5x5x5;
      
      // coefficients for 3D 7x7x7 chamfer mask. 
      unsigned char *** chamfer3D7x7x7;
      // coefficients for 3D 7x7x7 initialization chamfer mask, corresponding to the previous one.
      unsigned char *** chamfer3D7x7x7Init;
      float epsilon3D7x7x7;
      float maxRelErr3D7x7x7;
      
      // allocation and initialization of the pre-defined masks
      void initPredefinedMasks();
      
      // destruction of the pre-defined masks
      void deletePredefinedMasks();
  };

};

#endif // PREDEFINEDMASKS_H
