#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/recursiveFilters.h"

#include <inrimage/recursiveFilters.h>
#include <inrimage/arithmetic.h>
#include <linearAlgebra/Vec3.h>

namespace yav {

/* GHow to compute alpha based on the variance sigma */
#define SIGMA_TO_ALPHA 8.0 

static void _fft(float *buf, int *dimar, int ndim, FFT_DIRECTION isign);

static double *getSobelFilter(int);
static double *getGaussianFilter(int);
static void lineConvolution(double *, double *, int , double *, int);

Uninitialized::Uninitialized(const std::string& from, 
			     const std::string& arg) :
  Exception(from, arg) { }
  
InvalidDericheArgument::InvalidDericheArgument(const std::string& from,
					       const std::string& arg) :
    Exception(from, arg) { }
  
  

   template <class L, class B> 
   void copyStep(L *input, B *out,
		 unsigned int size,
		 unsigned int step)  {
   B *end;
   
   for(end = out + size; out < end; out++, input += step)
     *out = (B) *input;
 }
 
 

 template <class L, class B> 
   void copyIncr(L *input, B *out,
		 unsigned int size)  {
   B *end;
   
   for(end = out + size; out < end; out++, input++)
     *out = (B) *input;
 }
 

 template <class B>
   void copyLineToBuffer(B *b, const Inrimage *inr, Inrimage::WORD_TYPE type,
			 void *input, B *output,
			 unsigned int size, unsigned int step) {
   if(step != 1) {
     switch(type) {
     case Inrimage::WT_UNSIGNED_CHAR:
       copyStep((unsigned char *) input, (B *) output, size, step);
       break;
     case Inrimage::WT_UNSIGNED_SHORT:
       copyStep((unsigned short int *) input, (B *) output, size, step);
       break;
     case Inrimage::WT_SIGNED_SHORT:
       copyStep((short int *) input, (B *) output, size, step);
       break;
     case Inrimage::WT_UNSIGNED_INT:
       copyStep((unsigned int *) input, (B *) output, size, step);
       break;
     case Inrimage::WT_SIGNED_INT:
       copyStep((int *) input, (B *) output, size, step);
       break;
     case Inrimage::WT_UNSIGNED_LONG:
       copyStep((unsigned LONGINT *) input, (B *) output, size, step);
       break;
     case Inrimage::WT_SIGNED_LONG:
       copyStep((LONGINT *) input, (B *) output, size, step);
       break;
     case Inrimage::WT_FLOAT:
     case Inrimage::WT_FLOAT_VECTOR:
       copyStep((float *) input, (B *) output, size, step);
       break;
     case Inrimage::WT_DOUBLE:
     case Inrimage::WT_DOUBLE_VECTOR:
       copyStep((double *) input, (B *) output, size, step);
       break;
     default:
       break;
     }
   }
   
   else {
     switch(type) {
     case Inrimage::WT_UNSIGNED_CHAR:
      copyIncr((unsigned char *) input, (B *) output, size);
      break;
    case Inrimage::WT_UNSIGNED_SHORT:
      copyIncr((unsigned short int*) input, (B *) output,
				      size);
      break;
    case Inrimage::WT_SIGNED_SHORT:
      copyIncr((short int *) input, (B *) output, size);
      break;
    case Inrimage::WT_UNSIGNED_INT:
      copyIncr((unsigned int *) input, (B *) output, size);
      break;
    case Inrimage::WT_SIGNED_INT:
      copyIncr((int *) input, (B *) output, size);
      break;
    case Inrimage::WT_UNSIGNED_LONG:
      copyIncr((unsigned LONGINT *) input, (B *) output, size);
      break;
    case Inrimage::WT_SIGNED_LONG:
      copyIncr((LONGINT *) input, (B *) output, size);
      break;
    case Inrimage::WT_FLOAT:
    case Inrimage::WT_FLOAT_VECTOR:
      copyIncr((float *) input, (B *) output, size);
      break;
    case Inrimage::WT_DOUBLE:
    case Inrimage::WT_DOUBLE_VECTOR:
      copyIncr((double *) input, (B *) output, size);
      break;
    default:
      break;
    }
  }
}





  /// line convolution
  template <class T> 
    void lineConvolute(T ***output, const Inrimage *inr, 
		       double *filter, unsigned int filterSize,
		       SLICE_ORIENTATION direction,
		       unsigned int component = 0) throw (Inrimage::InvalidArguments) {
  unsigned int i,j,k,step;
  unsigned int image_size=0;
  register double *ptrin;
  register T *ptrout;
  double *line_input,*line_output;
  int row_size,column_size,plane_size;
  unsigned char *ptr;
  char from[] = "Inrimage::lineConvolute";
  unsigned int xdim, ydim, zdim, vdim, wdim;
  xdim = inr->getX();  ydim = inr->getY();
  zdim = inr->getZ();  vdim = inr->getVdim();
  wdim = inr->getWdim();
  void *data = inr->getData();

  if (!output) throw Inrimage::InvalidArguments(from,
				      "output buffer must be allocated");
  switch(direction) {
  case SLICE_Y:
    image_size = xdim;
    break;
  case SLICE_X:
    image_size = ydim;
    break;
  case SLICE_Z:
  case SLICE_ALPHA:
    image_size = zdim;
    break;
  default:
    break;
  }

  if(image_size < (filterSize+1))
    throw Inrimage::InvalidArguments(from,
			   "cannot make a convolution with a too small array");
  column_size = inr->sizeofWords();
  row_size = xdim * column_size;
  plane_size = ydim * row_size;

  /* Allocating the buffers */ 
  line_input = new double[image_size];
  line_output = new double[image_size];

  switch(direction) {
    case SLICE_Y:
      for(i = 0; i < zdim;++i) {
	for(j = 0; j < ydim; ++j) {
	  /* Copy to the input buffer */
	  if(vdim == 1) {
	    ptr = (unsigned char *) data + i*plane_size+ j*row_size;
	  }
	  else {
	    ptr = (unsigned char *) data + 
	      component * xdim * ydim * zdim * wdim + i*plane_size +
	      j*row_size;
	  }

	  copyLineToBuffer((double *) 0, inr, inr->getType(), (void *) ptr, line_input, xdim, 1);

	  /* Now process the lines */
	  lineConvolution(line_input, line_output, image_size, filter,
			  filterSize);
	  /* Stores the result into output */
	  for(k=0, ptrin = line_output, ptrout = &output[i][j][0];
	      k < image_size; ++k, ++ptrin, ptrout++)
	    *ptrout = (T) *ptrin;
	}
      }
      break;
  case SLICE_X:
    step = xdim;
    for(i = 0; i < zdim; ++i) {
      for(j = 0; j < xdim; ++j) {
	/* Copy to the input buffer */
	if(vdim == 1) {
	  ptr = (unsigned char *) data + i*plane_size+ j*column_size;
	}
	else {
	  ptr = (unsigned char *) data +
	    component * xdim * ydim * zdim * wdim + i*plane_size +
	    j*column_size;
	}

	copyLineToBuffer((double *) 0, inr, inr->getType(), (void *) ptr, line_input, ydim, step);

	/* Now process the lines */
	lineConvolution(line_input, line_output, image_size, filter,
			filterSize);
	/* Stores the result into output */
	for(k = 0, ptrin= line_output, ptrout = &output[i][0][j];
	    k<image_size; ++k, ++ptrin, ptrout += step)
	  *ptrout = (T) *ptrin;		
      }
    }
    break;
  case SLICE_Z:	
  case SLICE_ALPHA:	
    step = xdim * ydim;
    for(i = 0; i < ydim; ++i) {
      for(j = 0; j < xdim; ++j) {
	/* Copy to the input buffer */
	if(vdim == 1) {
	  ptr = (unsigned char *) data + i*row_size+ j*column_size;
	}
	else {
	  ptr = (unsigned char *) data +
	    component * xdim * ydim * zdim * wdim + i*row_size +
	    j*column_size;
	}

	copyLineToBuffer((double *) 0, inr, inr->getType(), (void *) ptr, line_input, zdim, step);

	/* Now process the lines */
	lineConvolution(line_input, line_output, image_size, filter,
			filterSize);
	/* Stores the result into output */
	for(k=0, ptrin = line_output, ptrout = &output[0][i][j];
	    k<image_size; ++k, ++ptrin, ptrout += step)
	  *ptrout = (T) *ptrin;		
      }
    }
    
    break;
  default:
	break;
  }

  /* free allocated memory */
  delete[] line_input;
  delete[] line_output;  
}



RECURSIVE_FILTER *recursiveFilterAlloc(void) {
  RECURSIVE_FILTER *rf1;

  rf1 = new RECURSIVE_FILTER;
  rf1->sd1 = rf1->sd2 = rf1->sd3 = rf1->sd4 = 0.0;
  rf1->sp0 = rf1->sp1 = rf1->sp2 = rf1->sp3 = 0.0;
  rf1->sn0 = rf1->sn1 = rf1->sn2 = rf1->sn3 = rf1->sn4 = 0.0;
  rf1->alpha = 1.0;
  rf1->margin = 0;
  return(rf1);
}


  /// recursive filtering
  template <class T> 
    void filterRecursiveImage(T *output, const Inrimage *inr, RECURSIVE_FILTER *rf1, SLICE_ORIENTATION direction) throw(Inrimage::InvalidArguments) {

    double first_val,last_val;
    unsigned int h, total_size, i, j, k;
	unsigned int image_size=0;  
    int l;
    register double *ptrin,*fi,*li;
    register T *ptrout;
    double *line_input,*line_output,*line_work,*first_image_pixel;
    int row_size,column_size,plane_size,dim_size,step;
    int orow_size,ocolumn_size,oplane_size,odim_size,ostep;
    unsigned char *ptr;
    
    if(!output) throw Inrimage::InvalidArguments("filterRecursiveImage",
						 "output buffer must be allocated");
    
    unsigned int xmin,ymin,zmin,xmax,ymax,zmax;
    inr->getWorkRegion(xmin,ymin,zmin,xmax,ymax,zmax);
    
    switch(direction) {
    case SLICE_X:
      image_size = xmax-xmin;
      break;
    case SLICE_Y:
      image_size = ymax-ymin;
      break;
    case SLICE_Z:
    case SLICE_ALPHA:
      image_size = zmax-zmin;
      break;
    default:
      break;
    }
    total_size = image_size + 2 * rf1->margin;
    
    /* Allocating the buffers */ 
    line_input = new double[total_size];
    line_output = new double[total_size];
    line_work = new double[total_size];
    
    /* First_image_pixel is the pointer of the first image pixel*/
    VECTORIAL_MODE vectMode = inr->getVectorialMode();
    unsigned int xdim, ydim, zdim, vdim;
    xdim = inr->getX(); ydim = inr->getY(); 
    zdim = inr->getZ(); vdim = inr->getVdim();
    unsigned int sizeofWords = inr->sizeofWords();
    void *data = inr->getData();
    
    first_image_pixel = line_input + rf1->margin;
    step = vectMode==VM_INTERLACED ? vdim : 1;
    column_size = sizeofWords * step;
    row_size = inr->getX() * column_size;
    plane_size = inr->getY() * row_size;
    dim_size = vectMode==VM_NON_INTERLACED ? plane_size * inr->getZ() : sizeofWords;
    
    ostep = vectMode==VM_INTERLACED ? vdim : 1;
    ocolumn_size = ostep;
    orow_size = (xmax-xmin) * ocolumn_size;
    oplane_size = (ymax-ymin) * orow_size;
    odim_size = vectMode==VM_NON_INTERLACED ? oplane_size * (zmax-zmin) : 1;
    
    switch(direction) {
      
    case SLICE_X:
      for(h=0;h<vdim;++h) {
	for(i=zmin;i<zmax;++i) {
	  for(j=ymin;j<ymax;++j) {
	    /* Copy to the input buffer */
	    ptr = (unsigned char *) data + i * plane_size + j * row_size + xmin * column_size + h * dim_size;
	    copyLineToBuffer((double *) 0, inr, inr->getType(), (void *) ptr, 
				  first_image_pixel, image_size, step);
	    /* Now deals with the image extension */
	    // shoul take in account surrounding voxels they are part of the image. AG.
	    if (rf1->margin>0) {
	      /* Stores the first and last items */
	      first_val = *first_image_pixel;
	      last_val = *(first_image_pixel + image_size - 1);
	      for(fi = line_input, li = line_input + total_size - 1, l = 0;
		  l < rf1->margin; ++l, ++fi, --li) {
		*fi = first_val;
		*li = last_val;
	      }
	    }
	    /* Now process the lines */
	    filterRecursiveLine(rf1, line_input, line_output, line_work,
				total_size);
	    /* Stores the result into output */
	    if(ostep==1)
	      for(k = 0, ptrin = line_output + rf1->margin,
		    ptrout = output + (i-zmin) * oplane_size + (j-ymin) * orow_size + h * odim_size;
		  k < image_size; ++k, ++ptrin, ++ptrout)
		*ptrout = (T) *ptrin;
	    else
	      for(k = 0, ptrin = line_output + rf1->margin,
		    ptrout = output + (i-zmin) * oplane_size + (j-ymin) * orow_size + h * odim_size;
		  k < image_size; ++k, ++ptrin, ptrout+=ostep)
		*ptrout = (T) *ptrin;
	  }
	}
      }
      break;
      
    case SLICE_Y:
      step *= xdim;
      ostep *= (xmax-xmin);
      for(h=0;h<vdim;++h) {
	for(i = zmin; i < zmax; ++i) {
	  for(j = xmin; j < xmax; ++j) {
	    /* Copy to the input buffer */
	    ptr = (unsigned char *) data + i*plane_size + ymin * row_size + j*column_size + h * dim_size;
	    
	    copyLineToBuffer((double *) 0, inr, inr->getType(), (void *) ptr, first_image_pixel,
				  image_size, step);
	    
	    /* Now deals with the image extension */
	    if (rf1->margin>0) {
	      /* Stores the first and last items */
	      first_val = *first_image_pixel;
	      last_val = *(first_image_pixel + image_size - 1);
	      for(fi = line_input, li = line_input + total_size - 1, l = 0;
		  l < rf1->margin; ++l, ++fi, --li) {
		*fi = first_val;
		*li = last_val;
	      }
	    }
	    /* Now process the lines */
	    filterRecursiveLine(rf1, line_input, line_output, line_work,
				total_size);
	    /* Stores the result into output */
	    for(k = 0,ptrin = line_output + rf1->margin,
		  ptrout = output + (i-zmin) * oplane_size + (j-xmin) * ocolumn_size + h * odim_size;
		k < image_size; ++k, ++ptrin, ptrout += ostep)
	      *ptrout = (T) *ptrin;
	  }
	}
      }
      break;
      
    case SLICE_Z:	
    case SLICE_ALPHA:	
      if (zdim>1) {
	step *= xdim*ydim;
	ostep *= (xmax-xmin)*(ymax-ymin);
	for(h=0;h<vdim;++h) {
	  for(i = ymin; i < ymax; ++i) {
	    for(j = xmin; j < xmax; ++j) {
	      /* Copy to the input buffer */
	      ptr = (unsigned char *) data + zmin * plane_size + i*row_size + j*column_size + h * dim_size;
	      
	      copyLineToBuffer((double *) 0, inr, inr->getType(),(void *) ptr, first_image_pixel,
				    image_size, step);
	      
	      /* Now deals with the image extension */
	      if (rf1->margin > 0) {
		/* Stores the first and last items */
		first_val = *first_image_pixel;
		last_val = *(first_image_pixel+image_size-1);
		for(fi = line_input, li = line_input + total_size - 1, l = 0;
		    l < rf1->margin; ++l, ++fi, --li) {
		  *fi = first_val;
		  *li = last_val;
		}
	      }
	      /* Now process the lines */
	      filterRecursiveLine(rf1, line_input, line_output, line_work,
				  total_size);
	      /* Stores the result into output */
	      for(k = 0,ptrin = line_output + rf1->margin,
		    ptrout = output + (i-ymin) * orow_size + (j-xmin) * ocolumn_size + h * odim_size;
		  k < image_size; ++k, ++ptrin, ptrout += ostep)
		*ptrout = (T) *ptrin;
	    }
	  }
	}
      } else {
	/* We only have one plane */
	if (rf1->filter_derivative == SMOOTHING_FILTER) {
	  /* Duplicate the slice and convert into double */
	  for(h=0;h<vdim;++h)
	    for(i = ymin; i < ymax; ++i)
	      
	      copyLineToBuffer((T *) 0, inr, inr->getType(),
				    (char *) data + i * row_size + xmin * column_size + h * dim_size,
				    output + (i-ymin) * orow_size + h * odim_size,
				    image_size, ostep);
	  
	} else {
	  /* We use a differential operator, therefore but zero */
	  for(h=0;h<vdim;++h)
	    for(i = ymin; i < ymax; ++i)
	      memset((char *) (output + (i-ymin) * orow_size + h * odim_size), 0, sizeof(T) * image_size);
	}
      }
      break;
    default:
	break;
    }
    
    /* Free allocated memory */
    delete[] line_input;
    delete[] line_output;
    delete[] line_work;
  }
  




void computeSn(RECURSIVE_FILTER *rf1) {
  switch(rf1->filter_derivative) {
  case SMOOTHING_FILTER:
  case SECOND_DERIVATIVE:
    rf1->sn1 = rf1->sp1 - rf1->sd1 * rf1->sp0;
    rf1->sn2 = rf1->sp2 - rf1->sd2 * rf1->sp0;
    rf1->sn3 = rf1->sp3 - rf1->sd3 * rf1->sp0;
    rf1->sn4 = -rf1->sd4 * rf1->sp0;
    break;
  case FIRST_DERIVATIVE:
  case THIRD_DERIVATIVE:
    rf1->sn1 = - rf1->sp1 + rf1->sd1 * rf1->sp0;
    rf1->sn2 = - rf1->sp2 + rf1->sd2 * rf1->sp0;
    rf1->sn3 = - rf1->sp3 + rf1->sd3 * rf1->sp0;
    rf1->sn4 = rf1->sd4 * rf1->sp0;
    break;
  default:
    break;
  }
}




RECURSIVE_FILTER *InitRecursiveFilter(double sigma, RECURSIVE_FILTER_TYPE filterType, FILTER_DERIVATIVE filter_derivative, int margin) throw (InvalidDericheArgument) {

  if(sigma>100) sigma=100;	// sigma shouldn't be > 100.
				// that shouldn't be a problem, being such
				// a high bound.
  if(sigma<=0) sigma=0.001;	// sigma needs to be > 0.
				// actually, anything < 0.5 doesn't really
				// make sense. (according to G. Malandain)
  
  double ex, k1, k2;
  double a0, a1, c0, c1, omega0, omega1, b0, b1;
  double cos0, sin0, cos1, sin1;
  double sumA = 0.0, sumC = 0.0, aux;
  double alpha = SIGMA_TO_ALPHA / (2 * sigma + 1);
  RECURSIVE_FILTER *rf1;
  char from[] = "InitRecursiveFilter";

  if(alpha>=2) alpha-=0.0001;	// alpha needs to be < 2.
  if(alpha<=0) alpha+=0.0001;	// alpha needs to be > 0.
  
  ex = k1 = k2 = 0.0;
  a0 = a1 = c0 = c1 = 0.0;
  b0 = b1 = omega0 = omega1 = 0.0;

  /* Do nothing if no filtering is required */
  if (filter_derivative == NO_FILTERING) return 0;

  rf1= yav::recursiveFilterAlloc();
  rf1->filter_type = filterType;
  rf1->filter_derivative = filter_derivative;
  if (margin >= 0) rf1->margin = margin;
  
  switch ( filterType ) {
  case MARTA_GAUSSIAN:
     
    switch(filter_derivative) {
    case SMOOTHING_FILTER:
      a0 =  0.6570033214 / sigma;
      a1 =  1.978946687  / sigma;
      c0 = -0.2580640608 / sigma;
      c1 = -0.2391206463 / sigma;
      omega0 = 0.6512453378;
      omega1 = 2.05339943;
      b0 = 1.906154352;
      b1 = 1.881305409;
      break;
    case FIRST_DERIVATIVE:
      a0 = -0.1726729496 / sigma;
      a1 = -2.003565572  / sigma;
      c0 =  0.1726730777 / sigma;
      c1 =  0.4440126835 / sigma;
      b0 = 1.560644213;
      b1 = 1.594202256;
      omega0 = 0.6995461735;
      omega1 = 2.144671764;
      break;
    case SECOND_DERIVATIVE:
      a0 = -0.7241334169 / sigma;
      a1 =  1.688628765  / sigma;
      c0 =  0.3251949838 / sigma;
      c1 = -0.7211796018 / sigma;
      b0 = 1.294951143;
      b1 = 1.427007123;
      omega0 = 0.7789803775;
      omega1 = 2.233566862;
      break;
    case THIRD_DERIVATIVE:
      a0 =  1.285774106  / sigma;
      a1 = -0.2896378408 / sigma;
      c0 = -1.28577129   / sigma;
      c1 =  0.26249833   / sigma;
      b0 = 1.01162886;
      b1 = 1.273344739;
      omega0 = 0.9474270928;
      omega1 = 2.337607006;
      break;
    default:
      break;
    }

    omega0 /= sigma;   sin0 = sin( omega0 );   cos0 = cos( omega0 ); 
    omega1 /= sigma;   sin1 = sin( omega1 );   cos1 = cos( omega1 ); 
    b0 /= sigma;
    b1 /= sigma;

    rf1->sp0  = a0 + c0;
    rf1->sp1  = exp( -b1 ) * (c1 * sin1 - (c0 + 2 * a0) * cos1);
    rf1->sp1 += exp( -b0 ) * (a1 * sin0 - (2 * c0 + a0) * cos0);
    rf1->sp2  = 2.0 * exp( -b0 - b1 ) *
      ((a0 + c0) * cos1 * cos0 - cos1 * a1 * sin0 - cos0 * c1 * sin1);
    rf1->sp2 += c0 * exp( -2.0 * b0 ) + a0 * exp( -2.0 * b1 );
    rf1->sp3  = exp( -b1 - 2.0 * b0 ) * (c1 * sin1 - c0 * cos1);
    rf1->sp3 += exp( -b0 - 2.0 * b1 ) * (a1 * sin0 - a0 * cos0);
    
    rf1->sd1  = -2.0 * exp( -b1 ) * cos1 - 2.0 * exp( -b0 ) * cos0;
    rf1->sd2  = 4.0 * cos1 * cos0 * exp( -b0 - b1 ) + exp( -2.0 * b1 ) +
      exp( -2.0 * b0 );
    rf1->sd3 = -2.0 * cos0 * exp( -b0 - 2.0 * b1 ) -
      2.0 * cos1 * exp( -b1 - 2.0 * b0 );
    rf1->sd4 = exp( -2.0 * b0 - 2.0 * b1 );
    
    yav::computeSn(rf1);
    break;

  case DERICHE_GAUSSIAN:
    switch(filter_derivative) {
    case SMOOTHING_FILTER:
      a0     =  1.68;
      omega0 =  0.6318;
      a1     =  3.735;
      b0     =  1.783;
      c0     = -0.6803;
      omega1 =  1.997;
      c1     = -0.2598;
      b1     =  1.723;
      break;
    case FIRST_DERIVATIVE:
      a0     =  -0.6472;
      omega0 =  0.6719;
      a1     =  -4.531;
      b0     =  1.527;
      c0     =  0.6494;
      omega1 =  2.072;
      c1     =  0.9557;
      b1     =  1.516;
      break;
    case SECOND_DERIVATIVE:
      a0     = -1.331;
      omega0 =  0.748;
      a1     =  3.661;
      b0     =  1.24;
      c0     =  0.3225;
      omega1 =  2.166;
      c1     = -1.738;
      b1     =  1.314;
      break;
    case THIRD_DERIVATIVE:
      throw InvalidDericheArgument(from,
		       "No 3rd derivatives available for DERICHE_GAUSSIAN");
      return 0;
    default:
      break;
    }
    
    omega0 /= sigma;   sin0 = sin( omega0 );   cos0 = cos( omega0 ); 
    omega1 /= sigma;   sin1 = sin( omega1 );   cos1 = cos( omega1 ); 
    b0 /= sigma;
    b1 /= sigma;
    /*--- normalisation ---*/
    switch(filter_derivative) {
    case SMOOTHING_FILTER:
      sumA  = 2.0 * a1 * exp( b0 ) * cos0 * cos0 - a0 * sin0 * exp( 2.0 * b0 );
      sumA += a0 * sin0 - 2.0 * a1 * exp( b0 );
      sumA /= ( 2.0 * cos0 * exp( b0 ) - exp( 2.0 * b0 ) - 1 ) * sin0;
      sumC  = 2.0 * c1 * exp( b1 ) * cos1 * cos1 - c0 * sin1 * exp( 2.0 * b1 );
      sumC += c0 * sin1 - 2.0 * c1 * exp( b1 );
      sumC /= ( 2.0 * cos1 * exp( b1 ) - exp( 2.0 * b1 ) - 1 ) * sin1;
      break;
    case FIRST_DERIVATIVE:
      aux   = exp( 4.0 * b0 ) - 4.0 * cos0 * exp( 3.0 * b0 );
      aux  += 2.0 * exp( 2.0 * b0 ) + 4.0 * cos0 * cos0 * exp( 2.0 * b0 );
      aux  += 1.0 - 4.0 * cos0 * exp( b0 );
      sumA  = a0 * cos0 - a1 * sin0 + a1 * sin0 * exp( 2.0 * b0 );
      sumA += a0 * cos0 * exp( 2.0 * b0 ) - 2.0 * a0 * exp( b0 );
      sumA *= exp( b0 ) / aux;
      aux   = exp( 4.0 * b1 ) - 4.0 * cos1 * exp( 3.0 * b1 );
      aux  += 2.0 * exp( 2.0 * b1 ) + 4.0 * cos1 * cos1 * exp( 2.0 * b1 );
      aux  += 1.0 - 4.0 * cos1 * exp( b1 );
      sumC  = c0 * cos1 - c1 * sin1 + c1 * sin1 * exp( 2.0 * b1 );
      sumC += c0 * cos1 * exp( 2.0 * b1 ) - 2.0 * c0 * exp( b1 );
      sumC *= exp( b1 ) / aux;
      sumA *= (-2.0);
      sumC *= (-2.0);
      break;
    case SECOND_DERIVATIVE:
      aux   = 12.0 * cos0 * exp( 3.0 * b0 ) - 3.0 * exp( 2.0 * b0 );
      aux  += 8.0 * cos0 * cos0 * cos0 * exp( 3.0 * b0 ) -
	12.0 * cos0 * cos0 * exp( 4.0 * b0 );
      aux  -= 3.0 * exp( 4.0 * b0 );
      aux  += 6.0 * cos0 * exp( 5.0 * b0 ) -  exp( 6.0 * b0 ) +
	6.0 * cos0 * exp( b0 );
      aux  -= ( 1.0 + 12.0 * cos0 * cos0 * exp( 2.0 * b0 ) );
      sumA  = 4.0 * a0 * sin0 * exp( 3.0 * b0 ) +
	a1 * cos0 * cos0 * exp( 4.0 * b0 );
      sumA -= ( 4.0 * a0 * sin0 * exp( b0 ) +
		6.0 * a1 * cos0 * cos0 * exp( 2.0 * b0 ) );
      sumA += 2.0 * a1 * cos0 * cos0 * cos0 * exp( b0 ) -
	2.0 * a1 * cos0 * exp( b0 );
      sumA += 2.0 * a1 * cos0 * cos0 * cos0 * exp( 3.0 * b0 ) -
	2.0 * a1 * cos0 * exp( 3.0 * b0 );
      sumA += a1 * cos0 * cos0 - a1 * exp( 4.0 * b0 );
      sumA += 2.0 * a0 * sin0 * cos0 * cos0 * exp( b0 ) -
	2.0 * a0 * sin0 * cos0 * cos0 * exp( 3.0 * b0 );
      sumA -= ( a0 * sin0 * cos0 * exp( 4.0 * b0 ) + a1 );
      sumA += 6.0 * a1 * exp( 2.0 * b0 ) + a0 * cos0 * sin0;
      sumA *= 2.0 * exp( b0 ) / ( aux * sin0 );
      aux   = 12.0 * cos1 * exp( 3.0 * b1 ) - 3.0 * exp( 2.0 * b1 );
      aux  += 8.0 * cos1 * cos1 * cos1 * exp( 3.0 * b1 ) -
	12.0 * cos1 * cos1 * exp( 4.0 * b1 );
      aux  -= 3.0 * exp( 4.0 * b1 );
      aux  += 6.0 * cos1 * exp( 5.0 * b1 ) -  exp( 6.0 * b1 ) +
	6.0 * cos1 * exp( b1 );
      aux  -= ( 1.0 + 12.0 * cos1 * cos1 * exp( 2.0 * b1 ) );
      sumC  = 4.0 * c0 * sin1 * exp( 3.0 * b1 ) +
	c1 * cos1 * cos1 * exp( 4.0 * b1 );
      sumC -= ( 4.0 * c0 * sin1 * exp( b1 ) +
		6.0 * c1 * cos1 * cos1 * exp( 2.0 * b1 ) );
      sumC += 2.0 * c1 * cos1 * cos1 * cos1 * exp( b1 ) -
	2.0 * c1 * cos1 * exp( b1 );
      sumC += 2.0 * c1 * cos1 * cos1 * cos1 * exp( 3.0 * b1 ) -
	2.0 * c1 * cos1 * exp( 3.0 * b1 );
      sumC += c1 * cos1 * cos1 - c1 * exp( 4.0 * b1 );
      sumC += 2.0 * c0 * sin1 * cos1 * cos1 * exp( b1 ) -
	2.0 * c0 * sin1 * cos1 * cos1 * exp( 3.0 * b1 );
      sumC -= ( c0 * sin1 * cos1 * exp( 4.0 * b1 ) + c1 );
      sumC += 6.0 * c1 * exp( 2.0 * b1 ) + c0 * cos1 * sin1;
      sumC *= 2.0 * exp( b1 ) / ( aux * sin1 );
      sumA /= 2;
      sumC /= 2;
      break;
      
    default:
      break;
      
    }
    a0 /= ( sumA + sumC );
    a1 /= ( sumA + sumC );
    c0 /= ( sumA + sumC );
    c1 /= ( sumA + sumC );
    
    rf1->sp0  = a0 + c0;
    rf1->sp1  = exp( -b1 ) * (c1 * sin1 - (c0 + 2 * a0) * cos1);
    rf1->sp1 += exp( -b0 ) * (a1 * sin0 - (2 * c0 + a0) * cos0);
    rf1->sp2  = 2.0 * exp( -b0 - b1 ) * ((a0 + c0) * cos1 * cos0 -
					 cos1 * a1 * sin0 - cos0 * c1 * sin1);
    rf1->sp2 += c0 * exp( -2.0 * b0 ) + a0 * exp( -2.0 * b1 );
    rf1->sp3  = exp( -b1 - 2.0 * b0 ) * (c1 * sin1 - c0 * cos1);
    rf1->sp3 += exp( -b0 - 2.0 * b1 ) * (a1 * sin0 - a0 * cos0);
    
    rf1->sd1  = -2.0 * exp( -b1 ) * cos1 - 2.0 * exp( -b0 ) * cos0;
    rf1->sd2  = 4.0 * cos1 * cos0 * exp( -b0 - b1 ) + exp( -2.0 * b1 ) +
      exp( -2.0 * b0 );
    rf1->sd3 = -2.0 * cos0 * exp( -b0 - 2.0 * b1 ) -
      2.0 * cos1 * exp( -b1 - 2.0 * b0 );
    rf1->sd4 = exp( -2.0 * b0 - 2.0 * b1 );

    yav::computeSn(rf1);	
    break;

  case DERICHE_STEP_EDGE:
    ex = exp( (-alpha) );
    switch(filter_derivative) {
    case SMOOTHING_FILTER:
      rf1->sp0 = (1.0 - ex) * (1.0 - ex) / (1.0 + 2.0 * alpha * ex - ex * ex);
      rf1->sp1 = rf1->sp0 * (alpha - 1.0) * ex;
      rf1->sn1 = rf1->sp0 * (alpha + 1.0) * ex;
      rf1->sn2 = (- rf1->sp0) * ex * ex;
      rf1->sd1 = (- 2.0) * ex;
      rf1->sd2 = ex * ex;
      break;
    case FIRST_DERIVATIVE:
      rf1->sp1 = - (1.0 - ex) * (1.0 - ex) * (1.0 - ex) / (2.0 * (1.0 + ex));
      rf1->sn1 = (- rf1->sp1);
      rf1->sd1 = (- 2.0) * ex;
      rf1->sd2 = ex * ex;
      break;
    case SECOND_DERIVATIVE:
      k1 = (- 2.0) * (1.0 - ex) * (1.0 - ex) * (1.0 - ex);
      k1 /= (1.0 + ex) * (1.0 + ex) * (1.0 + ex);
      k2 = (1.0 - ex * ex) / (2.0 * ex);
      rf1->sp0 = k1;
      rf1->sp1 = (- k1) * (1.0 + k2) * ex;
      rf1->sn1 = k1 * (1.0 - k2) * ex;
      rf1->sn2 = (- k1) * ex * ex;
      rf1->sd1 = (- 2.0) * ex;
      rf1->sd2 = ex * ex;
      break;
    case THIRD_DERIVATIVE:
      k1 = (1.0 + alpha) * ex + (alpha - 1.0);
      k2 = (1.0 - ex) / k1;
      k1 *= (1.0 - ex) * (1.0 - ex) * (1.0 - ex) * (1.0 - ex);
      k1 /= 2.0 * alpha * alpha * ex * ex;
      k1 /= ex + 1.0;
      rf1->sp0 = k1 * alpha * (k2 + 1.0);
      rf1->sp1 = (- k1) * alpha * (1.0 + k2 + k2*alpha) * ex;
      rf1->sn0 = (- rf1->sp0);
      rf1->sn1 = (- rf1->sp1);
      rf1->sd1 = (- 2.0) * ex;
      rf1->sd2 = ex * ex;
      break;
    default:
      break;
    }
    break;
  }

  return rf1;    
}




void filterRecursiveLine(RECURSIVE_FILTER *rf1,
				   double *input, double *output,
				   double *work1, unsigned int size) 
  throw (Uninitialized) {

  register double rp0, rp1, rp2, rp3;
  register double rd1, rd2, rd3, rd4;
  register double rn0, rn1, rn2, rn3, rn4;
  double *pw0,*pw1,*pw2,*pw3,*pw4,*pout;
  double *pin0,*pin1,*pin2,*pin3,*pin4;
  register int i;
  register unsigned int j;
 
  if (rf1 == 0) throw Uninitialized("filterRecursiveLine",
				    "RECURSIVE_FILTER *rf1 must be initialized");
  
  switch(rf1->filter_type) {
  case DERICHE_GAUSSIAN:
  case MARTA_GAUSSIAN:
    /*--- Generic filtering of order 4 ---*/
    rp0 = rf1->sp0;   rp1 = rf1->sp1;   rp2 = rf1->sp2;   rp3 = rf1->sp3;
    rd1 = rf1->sd1;   rd2 = rf1->sd2;   rd3 = rf1->sd3;   rd4 = rf1->sd4;
    rn1 = rf1->sn1;   rn2 = rf1->sn2;   rn3 = rf1->sn3;   rn4 = rf1->sn4;
    /*--- computation of y+ ---*/
    pw0=pw1=pw2=pw3=pw4=work1;
    pin0=pin1=pin2=pin3=input;

    *pw0++ = rp0 * *pin0++;
    *pw0++ = rp0 * *pin0++ + rp1 * *pin1++ -
      rd1 * *pw1++;   
    *pw0++ = rp0 * *pin0++ + rp1 * *pin1++ + rp2 * *pin2++ -
      rd1 * *pw1++ - rd2 * *pw2++;
    *pw0++ = rp0 * *pin0++ + rp1 * *pin1++ + rp2 * *pin2++ + rp3 * *pin3++ -
      rd1 * *pw1++ - rd2 * *pw2++ - rd3 * *pw3++;

    for (j=4; j<size; ++j,++pw0,++pin0) {
      *pw0 = rp0 * *pin0 + rp1 * *pin1 +
	rp2 * *pin2 + rp3 * *pin3 -
	rd1 * *pw1 - rd2 * *pw2 -
	rd3 * *pw3 - rd4 * *pw4;
      pw4=pw3;
      pw3=pw2;
      pw2=pw1;
      pw1=pw0;

      pin3=pin2;
      pin2=pin1;
      pin1=pin0;
    }
    /*--- computation of y- ---*/
    pw0=pw1=pw2=pw3=pw4=output+size-1;
    pin0=pin1=pin2=pin3=pin4=input+size-1;

    *pw0-- = 0.0;
    *pw0-- = rn1* *pin1-- -rd1 * *pw1--;
    *pw0-- = rn1* *pin1-- +rn2 * *pin2-- -rd1* *pw1-- -rd2 * *pw2--;
    *pw0-- = rn1* *pin1-- +rn2 * *pin2-- +rn3* *pin3-- - rd1* *pw1-- -rd2 * *pw2-- -rd3 * *pw3--;

    for (i=size-5; i>=0; i--,--pw0,--pin1)
      {
	*pw0= rn1* *pin1 +rn2 * *pin2 +rn3* *pin3 +rn4 * *pin4 
	  - rd1* *pw1 -rd2 * *pw2 -rd3 * *pw3 -rd4 * *pw4;
	pw4=pw3;
	pw3=pw2;
	pw2=pw1;
	pw1=pw0;

	pin4=pin3;
	pin3=pin2;
	pin2=pin1;
      }
    /*--- final computation  ---*/
    for (pw1=work1,pout=output,j=0 ; j<size ; j++,++pw1,++pout)
      *pout += *pw1;
    break;

  case DERICHE_STEP_EDGE:
    switch(rf1->filter_derivative) {
    case SMOOTHING_FILTER:
    case SECOND_DERIVATIVE:
      rp0 = rf1->sp0;   rp1 = rf1->sp1;
      rd1 = rf1->sd1;   rd2 = rf1->sd2;
      rn1 = rf1->sn1;   rn2 = rf1->sn2;

      /*--- computation of y+ ---*/
      pw0=pw1=pw2=work1;
      pin0=pin1=input;
	   
      *pw0++   = rp0 * *(pin0++);
      *pw0++ = rp0 * *pin0++ + rp1 * *pin1++ -
	rd1 * *pw1++;   
	   
      for (j=2; j<size; j++,++pw0,++pin0)
	{
	  *pw0 = rp0 * *pin0 + rp1 * *pin1 -
	    rd1 * *pw1 - rd2 * *pw2;

	  pw2=pw1;
	  pw1=pw0;
	  pin1=pin0;
	}
      /*--- computation of y- ---*/
      pw0=pw1=pw2=output+size-1;
      pin0=pin1=pin2=input+size-1;
	   
      *pw0--=0.0;
      *pw0--= rn1* *pin1-- -rd1 * *pw1--;
	   
      for (i=size-3; i>=0; i--,--pw0,--pin1)
	{
	  *pw0= rn1* *pin1 +rn2 * *pin2 
	    - rd1* *pw1 -rd2 * *pw2;
		 
	  pw2=pw1;
	  pw1=pw0;
	       
	  pin2=pin1;
	}
      /*--- final computation  ---*/
      for (pw1=work1,pout=output,j=0 ; j<size ; j++,++pw1,++pout)
	*pout += *pw1;

      /*--- final computation  ---*/
      /*	   for (pw1=work1,pw2=work2,pout=output,i=0 ; i<size ; i++,++pw1,++pw2,++pout)
       *pout = *pw1 + *pw2;*/
      break;

    case FIRST_DERIVATIVE:
      rp1 = rf1->sp1;
      rn1 = rf1->sn1;
      rd1 = rf1->sd1;   rd2 = rf1->sd2;
      /*--- computation of y+ ---*/
      pw0=pw1=pw2=work1;
      pin1=input;
	   
      *pw0++ = 0.0;
      *pw0++ = rp1 * *pin1++ -
	rd1 * *pw1++;   
	   
      for (j=2; j<size; j++,++pw0,++pin1)
	{
	  *pw0 = rp1 * *pin1 -
	    rd1 * *pw1 - rd2 * *pw2;

	  pw2=pw1;
	  pw1=pw0;
	}
      /*--- computation of y- ---*/
      pw0=pw1=pw2=output+size-1;
      pin1=input+size-1;
	   
      *pw0--=0.0;
      *pw0--= rn1* *pin1-- -rd1 * *pw1--;
	   
      for (i=size-3; i>=0; i--,--pw0,--pin1)
	{
	  *pw0= rn1* *pin1  
	    - rd1* *pw1 -rd2 * *pw2;
		 
	  pw2=pw1;
	  pw1=pw0;
	       
	}
      /*--- final computation  ---*/
      for (pw1=work1,pout=output,j=0 ; j<size ; j++,++pw1,++pout)
	*pout += *pw1;
      break;

    case THIRD_DERIVATIVE:
      rp0 = rf1->sp0;   rp1 = rf1->sp1;
      rd1 = rf1->sd1;   rd2 = rf1->sd2;
      rn0 = rf1->sn0;   rn1 = rf1->sn1;

      /*--- computation of y+ ---*/
      pw0=pw1=pw2=work1;
      pin0=pin1=input;
	  
      *pw0++   = rp0 * *(pin0++);
      *pw0++ = rp0 * *pin0++ + rp1 * *pin1++ -
	rd1 * *pw1++;   
	  
      for (j=2; j<size; j++,++pw0,++pin0)
	{
	  *pw0 = rp0 * *pin0 + rp1 * *pin1 -
	    rd1 * *pw1 - rd2 * *pw2;

	  pw2=pw1;
	  pw1=pw0;
	  pin1=pin0;
	}
      /*--- computation of y- ---*/
      pw0=pw1=pw2=output+size-1;
      pin0=pin1=pin2=input+size-1;
	   
      *pw0--= rn0* *pin0--;
      *pw0--= rn0* *pin0-- + rn1* *pin1-- -rd1 * *pw1--;
	   
      for (i=size-3; i>=0; i--,--pw0,--pin0)
	{
	  *pw0=  rn0* *pin0 + rn1* *pin1  
	    - rd1* *pw1 -rd2 * *pw2;
		 
	  pw2=pw1;
	  pw1=pw0;
	       
	  pin1=pin0;
	}
      /*--- final computation  ---*/
      for (pw1=work1,pout=output,j=0 ; j<size ; j++,++pw1,++pout)
	*pout += *pw1;
      /*	   for (pw1=work1,pw2=work2,pout=output,j=0 ; j<size ; j++,++pw1,++pw2,++pout)
       *pout = *pw1 + *pw2;*/
      break;

    default:
      break;
    }
    break;
  }

}


  void fastGradient2DRecursiveImage(const Inrimage *inr, Inrimage ** dirvect, RECURSIVE_FILTER_TYPE filter, 
				    double sigma, bool proportionalSigma) {


    float vx, vy, vz;
    vx = inr->getVX(); vy = inr->getVY(); vz = inr->getVZ();

    RECURSIVE_FILTER *rf1,*rf2;
    int margin;
    double const sigmax=sigma;
    double const sigmay=proportionalSigma && vx!=vy ? sigma*vx/vy : sigma;
    
    margin = (int) floor(sigma);
    
    rf1 = yav::InitRecursiveFilter(sigmax, filter, FIRST_DERIVATIVE, margin);
    if(sigmay!=sigmax)
      rf2 = yav::InitRecursiveFilter(sigmay, filter, FIRST_DERIVATIVE, margin);
    else rf2=rf1;
    
    unsigned int xmin,ymin,zmin,xmax,ymax,zmax;
    inr->getWorkRegion(xmin,ymin,zmin,xmax,ymax,zmax);
    if(*dirvect==0) {
      *dirvect = new Inrimage(xmax-xmin,ymax-ymin,zmax-zmin, Inrimage::WT_FLOAT_VECTOR, 2, VM_NON_INTERLACED,vx,vy,vz);
      (*dirvect)->copyHeader(inr);
    }

    filterRecursiveImage(((float ****) (*dirvect)->getVarray())[0][0][0], inr, rf1, SLICE_X);

    filterRecursiveImage(((float ****) (*dirvect)->getVarray())[1][0][0], inr, rf2, SLICE_Y);
    
    delete rf1;
    if(sigmay!=sigmax) delete rf2;
  }
  
  
  void gradient2DRecursiveImage(const Inrimage *inr, Inrimage ** dirvect,
				RECURSIVE_FILTER_TYPE filter,
				double sigma,
				bool proportionalSigma) {
    
    float vx, vy, vz;
    vx = inr->getVX(); vy = inr->getVY(); vz = inr->getVZ();
    unsigned int vdim = inr->getVdim();

    RECURSIVE_FILTER *rf1,*rf2,*rf3,*rf4;
    int margin;
    double const sigmax=sigma;
    double const sigmay=proportionalSigma && vx!=vy ? sigma*vx/vy : sigma;
    
    margin = (int) (2*sigma);
    
    rf1 = yav::InitRecursiveFilter(sigmax, filter, FIRST_DERIVATIVE, margin);
    rf3 = yav::InitRecursiveFilter(sigmax, filter, SMOOTHING_FILTER, margin);
    if(sigmay!=sigmax){
      rf2 = yav::InitRecursiveFilter(sigmay, filter, FIRST_DERIVATIVE, margin);
      rf4 = yav::InitRecursiveFilter(sigmay, filter, SMOOTHING_FILTER, margin);
    } else {
      rf2=rf1;
      rf4=rf3;
    }
    
    unsigned int xmin,ymin,zmin,xmax,ymax,zmax;
    inr->getWorkRegion(xmin,ymin,zmin,xmax,ymax,zmax);
    Inrimage tmp(xmax-xmin,ymax-ymin,zmax-zmin, Inrimage::WT_FLOAT,vdim, VM_SCALAR, vx, vy, vz);
    
    if(*dirvect==0) {
      *dirvect = new Inrimage(xmax-xmin,ymax-ymin,zmax-zmin, Inrimage::WT_FLOAT_VECTOR, 2, 
			      VM_NON_INTERLACED,vx,vy,vz);
      (*dirvect)->copyHeader(inr);
    }

  
    filterRecursiveImage((float*)tmp.getData(), inr, rf4, SLICE_Y);
    
    filterRecursiveImage(((float ****) (*dirvect)->getVarray())[0][0][0], &tmp, rf1, SLICE_X);
    
    filterRecursiveImage((float*)tmp.getData(), inr, rf3, SLICE_X);
			 
    filterRecursiveImage(((float ****) (*dirvect)->getVarray())[1][0][0], &tmp, rf2, SLICE_Y);
    
    delete rf1;
    delete rf3;
    if(sigmay!=sigmax){
      delete rf2;
      delete rf4;
    }
  }
  
  
  void fastGradient3DRecursiveImage(const Inrimage *inr, Inrimage ** dirvect,
				    RECURSIVE_FILTER_TYPE filter,
				    double sigma,
				    bool proportionalSigma) {

    float vx, vy, vz;
    vx = inr->getVX(); vy = inr->getVY(); vz = inr->getVZ();
 
   RECURSIVE_FILTER *rf1,*rf2,*rf3;
    int margin;
    double const sigmax=sigma;
    double const sigmay=proportionalSigma && vx!=vy ? sigma*vx/vy : sigma;
    double const sigmaz=proportionalSigma && vx!=vz ? sigma*vx/vz : sigma;
    
    margin = (int) floor(2 * sigma);
    
    unsigned int xmin,ymin,zmin,xmax,ymax,zmax;
    inr->getWorkRegion(xmin,ymin,zmin,xmax,ymax,zmax);

    if(*dirvect==0) {
      *dirvect = new Inrimage(xmax-xmin,ymax-ymin,zmax-zmin, Inrimage::WT_FLOAT_VECTOR, 3, VM_NON_INTERLACED,vx, vy, vz);
      (*dirvect)->copyHeader(inr);
    }

    rf1 = yav::InitRecursiveFilter(sigmax, filter, FIRST_DERIVATIVE, margin);
    if(sigmay!=sigmax)
      rf2 = yav::InitRecursiveFilter(sigmay, filter, FIRST_DERIVATIVE, margin);
    else rf2=rf1;
    if(sigmaz!=sigmax)
      rf3 = yav::InitRecursiveFilter(sigmaz, filter, FIRST_DERIVATIVE, margin);
    else rf3=rf1;
  
    filterRecursiveImage(((float ****) (*dirvect)->getVarray())[0][0][0], inr, rf1, SLICE_X);
    filterRecursiveImage(((float ****) (*dirvect)->getVarray())[1][0][0], inr, rf2, SLICE_Y);
  filterRecursiveImage(((float ****) (*dirvect)->getVarray())[2][0][0], inr, rf3, SLICE_Z);
  
  delete rf1;
  if(sigmay!=sigmax) delete rf2;
  if(sigmaz!=sigmax) delete rf3;
  }
  

void gradient3DRecursiveImage(const Inrimage *inr, Inrimage ** dirvect,
			      RECURSIVE_FILTER_TYPE filter,
			      double sigma,
			      bool proportionalSigma) {

  float vx, vy, vz;
  vx = inr->getVX(); vy = inr->getVY(); vz = inr->getVZ();
  
  RECURSIVE_FILTER *rf1,*rf2,*rf3,*rf4,*rf5,*rf6;
  int margin;
  double const sigmax=sigma;
  double const sigmay=proportionalSigma && vx!=vy ? sigma*vx/vy : sigma;
  double const sigmaz=proportionalSigma && vx!=vz ? sigma*vx/vz : sigma;
  
  /* the magic formulas */
  margin = (int) (2*sigma);
  
  rf1 = yav::InitRecursiveFilter(sigmax, filter, FIRST_DERIVATIVE, margin);
  rf4 = yav::InitRecursiveFilter(sigmax, filter, SMOOTHING_FILTER, margin);
  if(sigmay!=sigmax){
    rf2 = yav::InitRecursiveFilter(sigmay, filter, FIRST_DERIVATIVE, margin);
    rf5 = yav::InitRecursiveFilter(sigmay, filter, SMOOTHING_FILTER, margin);
  } else {
    rf2=rf1;
    rf5=rf4;
  }
  if(sigmaz!=sigmax){
    rf3 = yav::InitRecursiveFilter(sigmaz, filter, FIRST_DERIVATIVE, margin);
    rf6 = yav::InitRecursiveFilter(sigmaz, filter, SMOOTHING_FILTER, margin);
  } else {
    rf3=rf1;
    rf6=rf4;
  }
  
  unsigned int xmin,ymin,zmin,xmax,ymax,zmax;
  inr->getWorkRegion(xmin,ymin,zmin,xmax,ymax,zmax);
  Inrimage tmp(xmax-xmin,ymax-ymin,zmax-zmin,Inrimage::WT_FLOAT,inr->getVdim(), VM_SCALAR, vx, vy, vz);
  Inrimage tmp2(xmax-xmin,ymax-ymin,zmax-zmin,Inrimage::WT_FLOAT,inr->getVdim(), VM_SCALAR, vx, vy, vz);


  if(*dirvect==0) {
    *dirvect = new Inrimage(xmax-xmin,ymax-ymin,zmax-zmin, Inrimage::WT_FLOAT_VECTOR, 3, VM_NON_INTERLACED,vx, vy, vz);
    (*dirvect)->copyHeader(inr);
  }


  filterRecursiveImage((float*)tmp.getData(),inr,rf6, SLICE_Z);
  filterRecursiveImage((float*)tmp2.getData(), &tmp, rf4, SLICE_X);
  filterRecursiveImage(((float ****) (*dirvect)->getVarray())[1][0][0], &tmp2, rf2, SLICE_Y);

  filterRecursiveImage((float*)tmp.getData(), &tmp, rf5, SLICE_Y);
  filterRecursiveImage(((float ****) (*dirvect)->getVarray())[0][0][0], &tmp, rf1, SLICE_X);
  
  filterRecursiveImage((float*)tmp.getData(), inr, rf4, SLICE_X);
  
  filterRecursiveImage((float*)tmp.getData(), &tmp, rf5, SLICE_Y);
  
  filterRecursiveImage(((float ****) (*dirvect)->getVarray())[2][0][0], &tmp, rf3, SLICE_Z);
  
  delete rf1;
  delete rf4;
  if(sigmay!=sigmax){
     delete rf2;
     delete rf5;
  }
  if(sigmaz!=sigmax){
    delete rf3;
     delete rf6;
  }  
}




void recursiveFilter(Inrimage *inr, RECURSIVE_FILTER_TYPE rfilterType,
			       FILTER_TYPE filterType,
			       double sigma,
			       bool twoD,
			       bool proportionalSigma,
			       double hdelingMultFactor) {

  float vx, vy, vz;
  vx = inr->getVX(); vy = inr->getVY(); vz = inr->getVZ();
  unsigned int xdim, ydim, zdim, vdim;
  xdim = inr->getX(); ydim = inr->getY(); 
  zdim = inr->getZ(); vdim = inr->getVdim();

  FILTER_DERIVATIVE fd=NO_FILTERING;
  RECURSIVE_FILTER *rf1, *rf2, *rf3;
  int margin;
  double const sigmax=sigma;
  double const sigmay=proportionalSigma && vx!=vy ? sigma*vx/vy : sigma;
  double const sigmaz=proportionalSigma && vx!=vz ? sigma*vx/vz : sigma;

  Inrimage::WORD_TYPE orgtype = inr->getType();

  switch(filterType) {
  case GAUSSIAN:
  case SMOOTH_X:
  case SMOOTH_Y:
  case SMOOTH_Z:
    fd = SMOOTHING_FILTER;
    break;
  case DERIVATIVE_X:
  case DERIVATIVE_Y:
  case DERIVATIVE_Z:
    fd = FIRST_DERIVATIVE;
    break;
  case SECOND_DERIVATIVE_X:
  case SECOND_DERIVATIVE_Y:
  case SECOND_DERIVATIVE_Z:
    fd = SECOND_DERIVATIVE;
    break;
  }

  /* the magic formula */
  margin = (int) (hdelingMultFactor*2*sigma);
    
  rf1 = yav::InitRecursiveFilter(sigmax, rfilterType, fd, margin);
  if(sigmay!=sigmax)
     rf2 = yav::InitRecursiveFilter(sigmay, rfilterType, fd, margin);
  else rf2=rf1;
  if(sigmaz!=sigmax){
    rf3 = yav::InitRecursiveFilter(sigmaz, rfilterType, fd, margin);
  }
  else rf3=rf1;

  if(vdim==1) inr->convertCast(Inrimage::WT_FLOAT);
  else inr->convertCast(Inrimage::WT_FLOAT_VECTOR);

  unsigned int xmin,ymin,zmin,xmax,ymax,zmax;
  inr->getWorkRegion(xmin,ymin,zmin,xmax,ymax,zmax);
  Inrimage* tmp=inr;
  if(xmin!=0 || xmax!=xdim || ymin!=0 || ymax!=ydim || zmin!=0 || zmax!=zdim)
    {
      tmp=new Inrimage(xmax-xmin,ymax-ymin,zmax-zmin,Inrimage::WT_FLOAT_VECTOR,vdim,    inr->getVectorialMode(),vx,vy,vz);
      tmp->copyHeader(inr);
    }

  switch(filterType) {
  case SMOOTH_X:
    /* Compute the second  derivative in row stored in Laplacian*/
    filterRecursiveImage((float *) tmp->getData(), inr, rf1, SLICE_X);
    if(tmp!=inr) inr->mel(tmp,(int) xmin,(int)ymin,(int)zmin,0,0,0,xmax-xmin,ymax-ymin,zmax-zmin);
    inr->convertCast(orgtype); 
    break;
  case SMOOTH_Y:
    /* Compute the second  derivative in row stored in Laplacian*/
    filterRecursiveImage((float *) tmp->getData(), inr, rf2, SLICE_Y);
    if(tmp!=inr) inr->mel(tmp,(int) xmin,(int)ymin,(int)zmin,0,0,0,xmax-xmin,ymax-ymin,zmax-zmin);
    inr->convertCast(orgtype); 
    break;
  case SMOOTH_Z:
    /* Compute the second  derivative in row stored in Laplacian*/
    filterRecursiveImage((float *) tmp->getData(), inr, rf3, SLICE_Z);
    if(tmp!=inr) inr->mel(tmp,(int) xmin,(int)ymin,(int)zmin,0,0,0,xmax-xmin,ymax-ymin,zmax-zmin);
    inr->convertCast(orgtype); 
    break;
  case GAUSSIAN:{
    /* Compute the second  derivative in row stored in Laplacian*/
    filterRecursiveImage((float *) tmp->getData(), inr, rf1, SLICE_X);
    /* Compute the second  derivative in column stored in Derivative*/
    filterRecursiveImage((float *) tmp->getData(), tmp, rf2, SLICE_Y);
    if(!twoD) {

       filterRecursiveImage((float *) tmp->getData(), tmp, rf3, SLICE_Z);
       // XP 15-04-5: rf3 may also be used by other filters
       // Delete it for everybody, i.e. at the end
       // if(sigmaz!=sigmax) delete rf3; 
    }
    
    if(tmp!=inr) inr->mel(tmp,(int) xmin,(int)ymin,(int)zmin,0,0,0,xmax-xmin,ymax-ymin,zmax-zmin);
    inr->convertCast(orgtype);
    // XP 15-04-5: rf2 may also be used by other filters
    // Delete it for everybody, i.e. at the end
    // if(sigmay!=sigmax) delete rf2;
    break;
  }
  
  case DERIVATIVE_X:
    filterRecursiveImage((float *) tmp->getData(), inr, rf1, SLICE_X);    
    if(tmp!=inr) inr->mel(tmp,(int)xmin,(int)ymin,(int)zmin,0,0,0,xmax-xmin,ymax-ymin,zmax-zmin);
    break;
   
  case DERIVATIVE_Y:
    filterRecursiveImage((float *) tmp->getData(), inr, rf2, SLICE_Y);    
    if(tmp!=inr) inr->mel(tmp,(int)xmin,(int)ymin,(int)zmin,0,0,0,xmax-xmin,ymax-ymin,zmax-zmin);
    break;
   
  case DERIVATIVE_Z:
    filterRecursiveImage((float *) tmp->getData(), inr, rf3, SLICE_Z);    
    if(tmp!=inr) inr->mel(tmp,(int)xmin,(int)ymin,(int)zmin,0,0,0,xmax-xmin,ymax-ymin,zmax-zmin);
    break;
   
  case SECOND_DERIVATIVE_X:
    filterRecursiveImage((float *) tmp->getData(), inr, rf1, SLICE_X);    
    if(tmp!=inr) inr->mel(tmp,(int)xmin,(int)ymin,(int)zmin,0,0,0,xmax-xmin,ymax-ymin,zmax-zmin);
    break;
   
  case SECOND_DERIVATIVE_Y:
    filterRecursiveImage((float *) tmp->getData(), inr, rf2, SLICE_Y);    
    if(tmp!=inr) inr->mel(tmp,(int)xmin,(int)ymin,(int)zmin,0,0,0,xmax-xmin,ymax-ymin,zmax-zmin);
    break;
   
  case SECOND_DERIVATIVE_Z:
    filterRecursiveImage((float *) tmp->getData(), inr, rf3, SLICE_Z);
    if(tmp!=inr) inr->mel(tmp,(int)xmin,(int)ymin,(int)zmin,0,0,0,xmax-xmin,ymax-ymin,zmax-zmin);
    break;   
  }
  if(tmp!=inr) delete tmp;

  if(sigmay!=sigmax) delete rf2;
  if(sigmaz!=sigmax) delete rf3;
  delete rf1;
}


void laplacian(Inrimage *inr, RECURSIVE_FILTER_TYPE filterType, double sigma, bool twoD, bool proportionalSigma) {

  float vx, vy, vz;
  vx = inr->getVX(); vy = inr->getVY(); vz = inr->getVZ();
  unsigned int xdim, ydim, zdim, vdim;
  xdim = inr->getX(); ydim = inr->getY(); 
  zdim = inr->getZ(); vdim = inr->getVdim();

  RECURSIVE_FILTER *rf1,*rf2,*rf3;
  Inrimage *derivative, *derivativeZ;
  int margin;
  double const sigmax=sigma;
  double const sigmay=proportionalSigma && vx!=vy ? sigma*vx/vy : sigma;
  double const sigmaz=proportionalSigma && vx!=vz ? sigma*vx/vz : sigma;

  margin = (int) floor(2*sigma);

  rf1 = yav::InitRecursiveFilter(sigmax, filterType, SECOND_DERIVATIVE, margin);
  if(sigmay!=sigmax)
     rf2 = yav::InitRecursiveFilter(sigmay, filterType, SECOND_DERIVATIVE, margin);
  else rf2=rf1;
  if(sigmaz!=sigmax){
    rf3 = yav::InitRecursiveFilter(sigmaz, filterType, SECOND_DERIVATIVE, margin);
  }
  else rf3=rf1;

  inr->convertCast(Inrimage::WT_FLOAT);

  unsigned int xmin,ymin,zmin,xmax,ymax,zmax;
  inr->getWorkRegion(xmin,ymin,zmin,xmax,ymax,zmax);
  derivative = new Inrimage(xmax-xmin,ymax-ymin,zmax-zmin,Inrimage::WT_FLOAT);
  derivative->copyHeader(inr);

  /* Compute the second  derivative in row stored in Laplacian*/
  filterRecursiveImage((float *) derivative->getData(), inr, rf1, SLICE_X);
  if(!twoD) {
    derivativeZ = new Inrimage(xmax-xmin,ymax-ymin,zmax-zmin, Inrimage::WT_FLOAT);
    derivativeZ->copyHeader(inr);
    filterRecursiveImage((float *) derivativeZ->getData(), inr, rf3, SLICE_Z);
    yav::add(derivative, derivativeZ);
    delete derivativeZ;
  }
  /* Compute the second  derivative in column stored in Derivative*/
  if(xmin!=0 || xmax!=xdim || ymin!=0 || ymax!=ydim || zmin!=0 || zmax!=zdim){
     Inrimage tmp(xmax-xmin,ymax-ymin,zmax-zmin,Inrimage::WT_FLOAT);
     filterRecursiveImage((float *) tmp.getData(), inr, rf2, SLICE_Y);
     yav::add(derivative, &tmp);
     inr->mel(&tmp,xmin,ymin,zmin,0,0,0,xmax-xmin,ymax-ymin,zmax-zmin);
  }else{
     filterRecursiveImage((float *) inr->getData(), inr, rf2, SLICE_Y);
     /* Add d2row+d2col stored in Laplacian*/
     yav::add(inr, derivative);
  }
  delete derivative;

  delete rf1;
  if(sigmay!=sigmax) delete rf2;
  if(sigmaz!=sigmax) delete rf3;
}





static float extrapolatePixel2D(float *ptr, int r, int c, float xd, float yd, int ncols);
static float extrapolatePixel3D(float *ptr, int r, int c, int p, float xd, float yd,
				float zd, int ncols, int size_plane);


void gradient(const Inrimage *inr, Inrimage*& directions, double sigma,
			RECURSIVE_FILTER_TYPE filter,
			bool twoD, bool proportionalSigma) {
  float vx, vy, vz;
  vx = inr->getVX(); vy = inr->getVY(); vz = inr->getVZ();
  if(twoD)
    yav::gradient2DRecursiveImage(inr, &directions, filter, sigma, proportionalSigma);
  else
    yav::gradient3DRecursiveImage(inr, &directions, filter, sigma, proportionalSigma);

  directions->setGeometry(inr->getGeometry());
  directions->setVoxelSize(vx, vy, vz);
}



void gradient(const Inrimage *inr, Inrimage *& out, Inrimage *& dirimg, double sigma, RECURSIVE_FILTER_TYPE filter, GRADIENT_DIRECTION dir, bool max, bool twoD, double lt, bool proportionalSigma) {
  Inrimage *dirvect=0, *norm;
  float vx, vy, vz;
  vx = inr->getVX(); vy = inr->getVY(); vz = inr->getVZ();

  if(twoD)
    yav::gradient2DRecursiveImage(inr, &dirvect, filter, sigma, proportionalSigma);
  else
    yav::gradient3DRecursiveImage(inr, &dirvect, filter, sigma, proportionalSigma);

  norm = new Inrimage(*dirvect);
  norm->copyHeader(inr);

  if((dir != GD_NONE) || max) {
    norm->convertCast(Inrimage::WT_FLOAT);
    yav::div(dirvect, norm);
  }
  if(max) keepGradientExtrema(inr, norm, dirvect, lt);
  switch(dir) {
  case GD_NONE:
    delete dirvect;
    break;
  case GD_VECTORS:
    dirimg = dirvect;
    dirimg->setGeometry(inr->getGeometry());
    dirimg->setVoxelSize(vx, vy, vz);
    dirimg->setInterlacedData();
    break;
  case GD_SAMPLE:
    createImageGradientDirection(inr, dirimg, dirvect, DEFAULT_SPHERE_FREQUENCY);
    delete dirvect;
    dirimg->setVoxelSize(vx, vy, vz);
    break;
  }
  
  norm->convertScale(Inrimage::WT_UNSIGNED_CHAR);
  out = norm;
  out->setGeometry(inr->getGeometry());
  out->setVoxelSize(vx, vy, vz);
}



void sobel(const Inrimage *inr, Inrimage *& out, Inrimage *& dirimg, unsigned int sigma, GRADIENT_DIRECTION dir, bool max, bool twoD, double lt) {
  Inrimage *dirvect, *norm;
  float vx, vy, vz;
  vx = inr->getVX(); vy = inr->getVY(); vz = inr->getVZ();

  if(twoD)
    gradient2DSobelImage(inr, dirvect, sigma);
  else
    gradient3DSobelImage(inr, dirvect, sigma);


  norm = new Inrimage(*dirvect);
  norm->copyHeader(inr);


  if((dir != GD_NONE) || max) {
    norm->convertCast(Inrimage::WT_FLOAT);
    yav::div(dirvect, norm);
  }
  if(max) keepGradientExtrema(inr, norm, dirvect, lt);
  switch(dir) {
  case GD_NONE:
    delete dirvect;
    break;
  case GD_VECTORS:
    dirimg = dirvect;
    dirimg->setGeometry(inr->getGeometry());
    dirimg->setVoxelSize(vx, vy, vz);
    dirimg->setInterlacedData();
    break;
  case GD_SAMPLE:
    createImageGradientDirection(inr, dirimg, dirvect, DEFAULT_SPHERE_FREQUENCY);
    delete dirvect;
    dirimg->setVoxelSize(vx, vy, vz);
    break;
  }

  norm->convertScale(Inrimage::WT_UNSIGNED_CHAR);
  out = norm;
  out->setGeometry(inr->getGeometry());
  out->setVoxelSize(vx, vy, vz);
}



void keepGradientExtrema(const Inrimage *inr, Inrimage *grnorm, Inrimage *dirvect,
				   double minimum_gradient) {
  unsigned long int size, size_plane, size_plane_bytes, i;
  float *ptrc, *ptrr, *ptrp;
  unsigned int r, c, p;
  float xd, yd, zd, norm, *ptr;
  float *out, *buffer1, *buffer2;
  float *previous_plane_buffer, *current_plane_buffer, *temp, *plane;
  float *previous_plane_output =0;
  unsigned int xdim, ydim, zdim, vdim;
  xdim=inr->getX(); ydim=inr->getY(); zdim=inr->getZ(); vdim=inr->getVdim();

  // 2D gradient image
  if(dirvect->getVdim() == 2) {
    size = ydim * xdim;
    size_plane = size * sizeof(float);
    /* Allocate temporary buffer */
    temp = new float[size_plane];

    ptrc = &((float ****) dirvect->getVarray())[0][0][0][0];
    ptrr = &((float ****) dirvect->getVarray())[1][0][0][0];

    for(ptr = (float *) grnorm->getData(), p = 0; p < zdim; ++p) {
      /* save the output of processing into into temp buffer*/
      out = temp;
      /* Record the beginning of the plane */
      plane = ptr;
      
      for(r = 0; r < ydim; ++r) {
	if ((r == 0) || (r == ydim-1)) {
	  for(c = 0; c < xdim; ++c, ++ptr, ++out, ptrc++, ptrr++) *out=0.0;
	}
	else {
	  for(c = 0; c < xdim; ++c, ++ptr, ++out, ptrc++, ptrr++) {
	    if ((c == 0) || (c == xdim-1)) *out=0.0;
	    else {
	      if ((*ptr) < minimum_gradient) *out=0.0;
	      else {
		/* comparaison par rapport au premier point : extrema ? */
		xd = (float) ((double) c - *ptrr);
		yd = (float) ((double) r - *ptrc);
		norm = extrapolatePixel2D(ptr,r,c,xd,yd,xdim);
			  
		if (norm>(*ptr)) (*out)=0.0;
		else {
		  xd = (float) ((double) c + *ptrr);
		  yd = (float) ((double) r + *ptrc);
		  norm = extrapolatePixel2D(ptr,r,c,xd,yd,xdim);
		  if (norm > (*ptr)) (*out)=0.0;
		  else (*out)=(*ptr);
		}
	      }
	    }
	  }
	}
      }
      /* save the temporary  into the slice */
      memcpy((void *) plane, (const void *) temp, size_plane);
    }
    delete[] temp;

    ptr = (float *) grnorm->getData();
    out = ptr + xdim * ydim * zdim;
    for(ptrc = &((float ****) dirvect->getVarray())[0][0][0][0],
	  ptrr = &((float ****) dirvect->getVarray())[1][0][0][0];
	ptr < out;
	ptr++, ptrr++, ptrc++)
      if(*ptr == 0.0) *ptrr = *ptrc = 0.0;
  }
  

  // 3D gradient image
  else {
    size_plane = ydim * xdim;
    size_plane_bytes = size_plane*sizeof(float);

    /* Allocate temporary buffer */
    buffer1 = new float[size_plane_bytes];
    buffer2 = new float[size_plane_bytes];
    current_plane_buffer = buffer1;
    previous_plane_buffer = buffer2;

    ptrc = &((float ****) dirvect->getVarray())[0][0][0][0];
    ptrr = &((float ****) dirvect->getVarray())[1][0][0][0];
    ptrp = &((float ****) dirvect->getVarray())[2][0][0][0];

    for(ptr = (float *) grnorm->getData(), p = 0; p < zdim; ++p) {
      /* updates the pointer towards the previous plane in the
	 resulting image */
      if(p <= 1) previous_plane_output = (float *) grnorm->getData();
      else previous_plane_output += size_plane;
      /* out is initialized as the current plane */
      out = current_plane_buffer;
      if((p == 0) || (p == (zdim - 1))) {
	for(i=0;i<size_plane;++i,++ptr,++out,ptrc++,ptrr++,ptrp++)
	  *out=0.0;
      }
      else {
	for(r = 0; r < ydim; ++r) {
	  if ((r==0) || (r==ydim-1)) {
	    for(c = 0; c < xdim; ++c, ++ptr, ++out, ptrc++, ptrr++, ptrp++)
	      *out=0.0;
	  }
	  else {
	    for(c = 0; c < xdim; ++c, ++ptr, ++out, ptrc++, ptrr++, 
		  ptrp++) {
	      if ((c==0) || (c==xdim-1)) *out=0.0;
	      else {
		if ((*ptr)<minimum_gradient) *out=0.0;
		else {
		  /* comparaison par rapport au premier point : extrema ? */
		  xd = (float) ((double) c- *ptrr);
		  yd = (float) ((double) r- *ptrc);
		  zd = (float) ((double) p- *ptrp);
		  norm = extrapolatePixel3D(ptr, r, c, p, xd, yd, zd,
					    xdim, size_plane);
		  
		  if (norm>(*ptr)) (*out)=0.0;
		  else {
		    xd = (float) ((double) c + *ptrr);
		    yd = (float) ((double) r + *ptrc);
		    zd = (float) ((double) p + *ptrp);
		    
		    norm = extrapolatePixel3D(ptr, r, c, p, xd, yd, zd,
					      xdim, size_plane);
		    if (norm>(*ptr)) (*out)=0.0;
		    else (*out)=(*ptr);
		  }
		}
	      }
	    }
	  }
	}
      }
      if (p>0)
	/* copy the previous plane */
	memcpy((void *) previous_plane_output,
	       (const void *) previous_plane_buffer, size_plane_bytes);
      /* now swaps the two buffer */
      if (current_plane_buffer == buffer1) {
	previous_plane_buffer = buffer1;
	current_plane_buffer = buffer2;
      }
      else {
	previous_plane_buffer = buffer2;
	current_plane_buffer = buffer1;
      }
      
    }
    previous_plane_output += size_plane;
    /* vopy the last plane */
    memcpy((void *) previous_plane_output, 
	   (const void *) previous_plane_buffer, size_plane_bytes);

    delete[] buffer1;
    delete[] buffer2;

    ptr = (float *) grnorm->getData();
    out = ptr + xdim * ydim * zdim;
    for(ptrc = &((float ****) dirvect->getVarray())[0][0][0][0],
	  ptrr = &((float ****) dirvect->getVarray())[1][0][0][0],
	  ptrp = &((float ****) dirvect->getVarray())[2][0][0][0];
	ptr < out;
	ptr++, ptrr++, ptrc++, ptrp++)
      if(*ptr == 0.0) *ptrr = *ptrc = *ptrp = 0.0;
  }
}


void createImageGradientDirection(const Inrimage *inr, Inrimage *& direction,
				  Inrimage *dirvect,
				  unsigned int frequency) {
  Vec3<double> *direction_array, *dir, v;
  int nb_directions, j, k, best;
  unsigned long int i, size;
  float dot_product, max_dot_product, *ptr, *ptc, *ptp;
  unsigned char *outc;
  unsigned short *outs;
  Inrimage::WORD_TYPE type;
  int *yn, *yp;
  double x, y=0.0, sinpsi, cosphi, l, step;
  unsigned int xdim, ydim, zdim, vdim;
  xdim = inr->getX(); ydim = inr->getY();
  zdim = inr->getZ(); vdim = inr->getVdim();

  
  /* Get the array of direction vectors */
  direction_array = inr->createDirectionArray(frequency);
  nb_directions = 20 * frequency * frequency;

  step = 2.0 / 72.0;

  // 2D gradient image
  if(dirvect->getVdim() == 2) {

    /* sampled shpere direction buffers */
    yn = new int[72];
    yp = new int[72];
  
    /* sample circle in 72 areas and compute closest direction to
       each area center */
    v[2] = 0.0;
    for(i = 0, cosphi = -1.0 + 1.0 / 72.0; cosphi < 1.0;
	i++, cosphi += step) {
      /* compute normalized vector v with y > 0 and cos(phi) = cosphi */
      v[0] = cosphi;
      v[1] = sqrt(1.0 - v[0] * v[0]);

      /* Find the direction with the highest dot_product */
      for(max_dot_product = -1.0, k = 0, best = 0, dir = direction_array;
	  k < nb_directions; ++k, ++dir) {
	dot_product = (float)((*dir)[0] * v[0] + (*dir)[1] * v[1] +
	  (*dir)[2] * v[2]);
	if (dot_product > max_dot_product) {
	  max_dot_product = dot_product;
	  best = k;
	}
      }
      yp[i] = best;

      /* compute normalized vector v with y < 0 and cos(phi) = cosphi */
      v[1] = -v[1];

      /* Find the direction with the highest dot_product */
      for(max_dot_product = -1.0, k = 0, best = 0, dir = direction_array;
	  k < nb_directions; ++k, ++dir) {
	dot_product = (float)((*dir)[0] * v[0] + (*dir)[1] * v[1] +
	  (*dir)[2] * v[2]);
	if (dot_product > max_dot_product) {
	  max_dot_product = dot_product;
	  best = k;
	}
      }
      yn[i] = best;
    }
    delete[] direction_array;

    /* Choosing the type depending on the number of directions :
       maximum frequency = 57 */ 
    if(nb_directions > 255) type = Inrimage::WT_UNSIGNED_SHORT;
    else type = Inrimage::WT_UNSIGNED_CHAR;

    /* Create the output buffer */
    direction = new Inrimage(xdim, ydim, zdim, type);
    direction->copyHeader(inr);

    size = xdim * ydim * zdim;

    if (type == Inrimage::WT_UNSIGNED_CHAR) {

      /* scan all voxels */
      for(ptc = &((float ****) dirvect->getVarray())[0][0][0][0],
	  ptr = &((float ****) dirvect->getVarray())[1][0][0][0],
	  outc = (unsigned char *) direction->getData(), i = 0;
	  i < size; ++i, ++ptr, ++ptc, ++outc) {
      
	/* if the gradient is non-zero */
	if (((*ptr)!=0.0) || ((*ptc)!=0.0)) {
	  /* compute cos(phi) of gradient vector */
	  j = (int) ((*ptr + 1.0) * 71.0 / 2.0 + 0.5);
	
	  /*  get close direction vector */
	  if(*ptc >= 0.0) *outc = (unsigned char) yp[j];
	  else *outc = (unsigned char) yn[j];
	}
	/* Store an index out of range */
	else
	  *outc = (unsigned char) nb_directions;
      }
    }
    
    else {
      /* Same thing for unsigned short image */
      for(ptc = &((float ****) dirvect->getVarray())[0][0][0][0],
	  ptr = &((float ****) dirvect->getVarray())[1][0][0][0],
	  outs = (unsigned short int *) direction->getData(), i = 0;
	  i < size; ++i, ++ptr, ++ptc, ++outs) {
	
	/* if the gradient is non-zero */
	if (((*ptr)!=0.0) || ((*ptc)!=0.0)) {
	  /* compute cos(phi) of gradient vector */
	  j = (int) ((*ptr + 1.0) * 71.0 / 2.0 + 0.5);
	  
	  /*  get close direction vector */
	  if(*ptc >= 0.0) *outs = (unsigned short int) yp[j];
	  else *outs = (unsigned short int) yn[j];
	}
	/* Store an index out of range */
	else
	  *outs = (unsigned short int) nb_directions;
	
      }
    }
  }


  // 3D gradient image
  else {
    /* sampled shpere direction buffers */
    yn = new int[72 * 72];
    yp = new int[72 * 72];
  
    /* sample shpere in 72 * 72 areas and compute closest direction to
       each area center */
    for(i = 0, cosphi = -1.0 + 1.0 / 72.0; cosphi < 1.0;
	i++, cosphi += step) {
      for(j = 0, sinpsi = -1.0 + 1.0 / 72.0; sinpsi < 1.0;
	  j++, sinpsi += step) {
	/* compute normalized vector v with y > 0, cos(phi) = cosphi and
	   sin(psi) = sinpsi */
	l = cos(asin(sinpsi));
	v[0] = l * cosphi;
	v[1] = sqrt(l * l - v[0] * v[0]);
	v[2] = sinpsi;

	/* Find the direction with the highest dot_product */
	for(max_dot_product = -1.0, k = 0, best = 0, dir = direction_array;
	    k < nb_directions; ++k, ++dir) {
	  dot_product = (float)((*dir)[0] * v[0] + (*dir)[1] * v[1] +
	    (*dir)[2] * v[2]);
	  if (dot_product > max_dot_product) {
	    max_dot_product = dot_product;
	    best = k;
	  }
	}
	yp[i * 72 + j] = best;

	/* compute normalized vector v with y < 0, cos(phi) = cosphi and
	   sin(psi) = sinpsi; */
	v[1] = -v[1];

	/* Find the direction with the highest dot_product */
	for(max_dot_product = -1.0, k = 0, best = 0, dir = direction_array;
	    k < nb_directions; ++k, ++dir) {
	  dot_product = (float)((*dir)[0] * v[0] + (*dir)[1] * v[1] +
	    (*dir)[2] * v[2]);
	  if (dot_product > max_dot_product) {
	    max_dot_product = dot_product;
	    best = k;
	  }
	}
	yn[i * 72 + j] = best;
      }
    }
    delete[] direction_array;

    /* Choose the type depending on the number of directions :
       maximum frequency = 57 */
    if (nb_directions > 255) type = Inrimage::WT_UNSIGNED_SHORT;
    else type = Inrimage::WT_UNSIGNED_CHAR;

    /* Create the output buffer */
    direction = new Inrimage(xdim, ydim, zdim, type);
    direction->copyHeader(inr);

    size = xdim * ydim * zdim;

    /* few directions: unsigned char direction image */
    if(type == Inrimage::WT_UNSIGNED_CHAR) {
      /* scan all voxels */
      for(ptc = &((float ****) dirvect->getVarray())[0][0][0][0],
	  ptr = &((float ****) dirvect->getVarray())[1][0][0][0],
	  ptp = &((float ****) dirvect->getVarray())[2][0][0][0],
	  outc = (unsigned char *) direction->getData(), i = 0;
	  i < size;
	  ++i,++ptr,++ptc,++ptp,++outc) {
      
	/* if the gradient is non-zero */
	if (((*ptr)!=0.0) || ((*ptc)!=0.0) || ((*ptp)!=0.0)) {
	  /* compute sin(psi) and cos(phi) of gradient vector */
	  double norm = sqrt(*ptr* *ptr + *ptr * *ptr + *ptp * *ptp);
	  sinpsi = (double) *ptp / norm;
	  if(sinpsi < 1.0 && sinpsi > -1.0) {
	    x = (double) *ptr;
	    y = (double) *ptc;
	    cosphi = x / sqrt(x*x + y*y);
	  }
	  else
	    cosphi = 0.0;
	  j = (int) ((cosphi + 1.0) * 71.0 / 2.0 + 0.5);
	  k = (int) ((sinpsi + 1.0) * 71.0 / 2.0 + 0.5);
	  
	  /*  get close direction vector */
	  if(y >= 0.0) *outc = (unsigned char) yp[j*72+k];
	  else *outc = (unsigned char) yn[j*72+k];
	  
	}
	/* Store an index out of range */
	else
	  *outc = (unsigned char) nb_directions;
      }
    }
  
    /* Same thing for unsigned short image */
    else {

      for(ptc = &((float ****) dirvect->getVarray())[0][0][0][0],
	    ptr = &((float ****) dirvect->getVarray())[1][0][0][0],
	    ptp = &((float ****) dirvect->getVarray())[2][0][0][0],
	    outs = (unsigned short int *) direction->getData(), i = 0; 
	  i < size;
	  ++i, ++ptr, ++ptc, ++ptp, ++outs) {
	
	/* if the gradient is non-zero */
	if (((*ptr)!=0.0) || ((*ptc)!=0.0) || ((*ptp)!=0.0)) {
	  /* compute sin(psi) and cos(phi) of gradient vector */
	  sinpsi = (double) *ptp;
	  if(sinpsi < 1.0 && sinpsi > -1.0) {
	    x = (double) *ptr;
	    y = (double) *ptc;
	    cosphi = x / sqrt(x*x + y*y);
	  }
	  else 
	    cosphi = 0.0;
	  j = (int) ((cosphi + 1.0) * 71.0 / 2.0 + 0.5);
	  k = (int) ((sinpsi + 1.0) * 71.0 / 2.0 + 0.5);
	  
	  /*  get close direction vector */
	  if(y >= 0.0) *outs = (unsigned short int) yp[j*72+k];
	  else *outs = (unsigned short int) yn[j*72+k];
	  
	}
	/* Store an index out of range */
	else
	  *outs = (unsigned short int) nb_directions;
      }
    }
  }

  delete[] yn;
  delete[] yp;
}



/*---------------------------------------------------------------------------*/
static float extrapolatePixel2D(float *ptr, int r, int c,
				float xd, float yd, int ncols) {
  float norm;
  int xi, yi, offset;
  float dx, dy;
  float coef[2][2];
  
  xi=(int)xd;
  yi=(int)yd;
  offset = (xi-c)+(yi-r)*ncols;

  dx = xd - (float)xi;   dy = yd - (float)yi; 
  coef[0][0] = (1.0f - dx) * (1.0f - dy); /* xi,   yi   */
  coef[0][1] = dx         * (1.0f - dy); /* xi+1, yi   */
  coef[1][0] = (1.0f- dx) * dy;         /* xi,   yi+1 */
  coef[1][1] = dx         * dy;         /* xi+1, yi+1 */
  

  norm  = *(ptr+offset)      * coef[0][0] + *(ptr+offset+1)       * coef[0][1];
  norm += *(ptr+offset+ncols) * coef[1][0] + *(ptr+offset+1+ncols) * coef[1][1];
  return(norm);
}


/*---------------------------------------------------------------------------*/
static float extrapolatePixel3D(float *ptr, int r, int c, int p,
				float xd, float yd, float zd,
				int ncols, int size_plane) {
  float norm;
  int xi,yi,zi,offset;
  float dx,dy,dz;
  float coef[2][2][2];
  
  xi = (int) xd;
  yi = (int) yd;
  zi = (int) zd;
  dx = xd - (float)xi;   dy = yd - (float)yi; dz = zd - (float)zi; 
  /*--- interpolation lineaire : calcul des coefficients ---*/
  coef[0][0][0] = (1.0f - dx) * (1.0f - dy) * (1.0f - dz); /* xi,   yi,   zi   */
  coef[1][0][0] = dx         * (1.0f - dy) * (1.0f - dz); /* xi+1, yi,   zi   */
  coef[0][1][0] = (1.0f - dx) * dy         * (1.0f - dz); /* xi,   yi+1, zi   */
  coef[1][1][0] = dx         * dy         * (1.0f - dz); /* xi+1, yi+1, zi   */
  coef[0][0][1] = (1.0f - dx) * (1.0f - dy) * dz;         /* xi,   yi,   zi+1 */
  coef[1][0][1] = dx         * (1.0f - dy) * dz;         /* xi+1, yi,   zi+1 */
  coef[0][1][1] = (1.0f - dx) * dy         * dz;         /* xi,   yi+1, zi+1 */
  coef[1][1][1] = dx         * dy         * dz;         /* xi+1, yi+1, zi+1 */
  
  offset = (xi-c)+(yi-r)*ncols +(zi-p)*size_plane;
  norm  = *(ptr+offset)      * coef[0][0][0] + *(ptr+offset+1)       * coef[1][0][0];
  norm += *(ptr+offset+ncols)* coef[0][1][0] + *(ptr+offset+1+ncols) * coef[1][1][0];
  if (dz>0) {
    offset +=size_plane;
    norm  += *(ptr+offset)     * coef[0][0][1] + *(ptr+offset+1)       * coef[1][0][1];
    norm += *(ptr+offset+ncols)* coef[0][1][1] + *(ptr+offset+1+ncols) * coef[1][1][1];
  }
  return(norm);
}



/*===========================================================
          Definition of the Icosahedron
===========================================================*/

static double icosahedron_vertices_pos[12][3]=
{{0.4249358858193742, -0.6234212590752522, -0.901521749427252}, 
 {-0.5999406907417391, 0.05914133178387721, -1.009227188986324}, 
 {0.4980672603913731, 0.6101449500603583, -0.872707378430782}, 
 {1.170388006823068, -0.06734654200022822, -0.0873067158600265}, 
 {0.4878971283264427, -1.037062929459035, 0.2615777778398508}, 
 {-0.606226178027942, -0.958889124295748, -0.3082004094765843}, 
 {-0.4249358858193759, 0.6234212590752533, 0.90152174942725}, 
 {0.5999406907417364, -0.05914133178387813, 1.009227188986327}, 
 {-0.4980672603913764, -0.6101449500603557, 0.87270737843078}, 
 {-1.170388006823072, 0.0673465420002311, 0.0873067158600247}, 
 {-0.4878971283264417, 1.037062929459036, -0.2615777778398516}, 
 {0.606226178027942, 0.958889124295747, 0.3082004094765856}};
static int icosahedron_faces_triangles[20][3]=
{{0, 1, 5}, {0, 2, 1}, {0, 3, 2}, {0, 4, 3}, {0, 5, 4}, 
 {1, 2, 10}, {1, 9, 5}, {1, 10, 9}, {2, 3, 11}, {2, 11, 10}, 
 {3, 4, 7}, {3, 7, 11}, {4, 5, 8}, {4, 8, 7}, {5, 9, 8}, 
 {6, 7, 8}, {6, 8, 9}, {6, 9, 10}, {6, 10, 11}, {6, 11, 7}};
static int icosahedron_faces_neighbor[20][3]=
{{6,4,1},{5,0,2},{8,1,3},{10,2,4},{12,3,0},
 {9,7,1},{14,0,7},{17,6,5},{11,9,2},{18,5,8},
 {13,11,3},{19,8,10},{14,13,4},{15,10,12},{16,12,6},
 {13,16,19},{14,17,15},{7,18,16},{9,19,17},{11,15,18}};
static int icosahedron_edge_vertex[30][2]=
{{0,1},{0,2},{0,3},{0,4},{0,5},
 {1,2},{1,5},{1,9},{1,10},{2,3},
 {2,10},{2,11},{3,4},{3,7},{3,11},
 {4,5},{4,7},{4,8},{5,8},{5,9},
 {6,7},{6,8},{6,9},{6,10},{6,11},
 {7,8},{7,11},{8,9},{9,10},{10,11}};
static int icosahedron_edge_triangle[30][2]=
{{0,1},{1,2},{2,3},{3,4},{0,4},
 {1,5},{0,6},{6,7},{7,5},{2,8},
 {5,9},{8,9},{3,10},{10,11},{8,11},
 {4,12},{10,13},{12,13},{12,14},{6,14},
 {15,19},{15,16},{16,17},{17,18},{18,19},
 {13,15},{11,19},{14,16},{7,17},{9,18}};

/*===========================================================
       Structure of Vertex
===========================================================*/

struct TMPVERTEX {
  Vec3<double> pos; /* Position of the subvertex */
int rank; /* Its rank */
};

/*===========================================================
       Structure of Subtriangle
===========================================================*/

struct SUBTRIANGLE {
  int vertex[3]; /* The three subvertices that make the subtriangle */
  Vec3<double> center; /* coordinates of the centroid of the subtriangle */
  int rank; /* The rank of the subtriangle */
  int neighbor[3]; /* The three subtriangles that share a commun edge */
};

/*===========================================================
       Structure of Triangle
===========================================================*/

struct BIG_TRIANGLE {
  int vertex[3]; /* The three vertices that make the triangle */
  int neighbor[3]; /* The three triangles that share a commun edge */
  TMPVERTEX *subvertex; /* List of subvertices */
  SUBTRIANGLE *subtriangle;  /* List of subtriangles */
};

/*===========================================================
       Structure of Edge
===========================================================*/

struct BIG_EDGE {
  int vertex[2]; /* the 2 vertices that make an edge */
  int triangle[2]; /* the 2 triangles that share this edge */
};

/*===========================================================
       Structure of Platonic Solid
===========================================================*/

typedef struct {
  int p_face; /* Number of vertices connected to each vertex */
  int nvertices; /* Number of vertices */
  TMPVERTEX *vertex; /* List of vertices */
  int ntriangles; /* Number of triangles */
  BIG_TRIANGLE *triangle; /* List of triangle */
  int nedges ; /* Number of edges */
  BIG_EDGE *edge; /* List of edges */
} platonic;


/*-------------------------------------------------------------------*/
Vec3<double> *Inrimage::createDirectionArray(int frequency) {
  int i,j;
  BIG_TRIANGLE *triangle;
  SUBTRIANGLE *subtriangle;
  Vec3<double> pos;
  int l,m,n;
  int rank_triangle,rank_vertex,rank0,rank1,rank2;
  BIG_TRIANGLE *next_triangle;
  int ind,f;
  platonic solid;
  int nb_subtriangles,nb_subvertices;
  Vec3<double> *list_pos;

  solid.nvertices = 12;
  solid.vertex = new TMPVERTEX[solid.nvertices];
  solid.ntriangles = 20;
  solid.triangle = new BIG_TRIANGLE[solid.ntriangles];
  /* Relation d'Euler */
  solid.nedges = solid.nvertices + solid.ntriangles - 2;
  solid.edge= new BIG_EDGE[solid.nedges];

  solid.p_face = 5;

  for (i = 0; i < solid.nvertices; ++i)
    solid.vertex[i].pos.setValue(icosahedron_vertices_pos[i][0],
				 icosahedron_vertices_pos[i][1],
				 icosahedron_vertices_pos[i][2]);
				 
  for (i=0;i<solid.ntriangles;++i) {
    for (j=0;j<3;++j) {
      solid.triangle[i].vertex[j] = icosahedron_faces_triangles[i][j];
      solid.triangle[i].neighbor[j] = icosahedron_faces_neighbor[i][j];
    }
  }

  for (i=0;i<solid.nedges;++i) {
    for (j=0;j<2;++j) {
      solid.edge[i].vertex[j] = icosahedron_edge_vertex[i][j];
      solid.edge[i].triangle[j] = icosahedron_edge_triangle[i][j];
    }
  }

  nb_subtriangles = frequency * frequency; 
  /* Number of subtriangles inside each triangle */
  nb_subvertices=(frequency+1)*(frequency+2)/2;
  /* Number of subverticess inside each triangle */

  /* For each triangle */
  for (i = 0; i < solid.ntriangles; ++i) {
      
    /* Allocates Memory */
    triangle = &(solid.triangle[i]);
    triangle->subvertex = new TMPVERTEX[nb_subvertices];
    triangle->subtriangle = new SUBTRIANGLE[nb_subtriangles];

    rank_triangle = i * nb_subtriangles; 
    /* rank_triangle is the "Global" rank of the current subtriangle*/
    
    /* ===================================
       Initialize the subvertices 
       =================================== */

    for (l = 0; l <= frequency; ++l) {
      for (m=0;m<=l;++m) {
	n = l-m; /*(l,m,n) are the coefficients of a subvertex
		   they are the weights of the 3 vertices of the triangle
		   Especially l+n+m=frequency */
			
	rank_vertex = (l*(l+1)/2) + m;
	/* Rank Vertex is the "Local" rank of the vertex inside
	   the triangle */

	pos = solid.vertex[triangle->vertex[0]].pos * (double) 
	  (frequency - l);
	pos += solid.vertex[triangle->vertex[1]].pos * (double) n;
	pos += solid.vertex[triangle->vertex[2]].pos * (double) m;
	pos /= (double) frequency;

	/* The subvertex (l,m,n) is a weighted sum of the 3 vertices
	   that make the triangle */
	triangle->subvertex[rank_vertex].pos = pos;
	triangle->subvertex[rank_vertex].rank = 
	  i * nb_subvertices+ rank_vertex;
	/* The global rank of the vertex is the local rank  plus the rank
	   of the rank of the first subvertex i*nb_subvertices */
      }
    }
    /* ===================================
       Initialize the subtriangles 
       =================================== */
      
    for (l = 0; l <= frequency; ++l) {
      for (m = 0; m <= l; ++m) {
	n = l-m; /*(l,m,n) are the coefficients of a subvertex
		   they are the weights of the 3 vertices of the triangle
		   Especially l+n+m=frequency */
	      
	rank_vertex = (l*(l+1)/2) + m;
	      
	if (l < frequency) /* Build first type of triangle */ {
		  
	  subtriangle=&(triangle->subtriangle
			[rank_triangle - i*nb_subtriangles]);
	  
	  /* Define the vertices of the subtriangle */
	  rank0 = rank_vertex;
	  rank1 = (l+1)*(l+2)/2 +m;
	  rank2 = (l+1)*(l+2)/2 +m+1;
	  /* rank0,rank1 rank2 are the local rank of the subvertices */
	  
	  subtriangle->vertex[0] = rank0 + nb_subvertices*i;
	  subtriangle->vertex[1] = rank1 + nb_subvertices*i;
	  subtriangle->vertex[2] = rank2 + nb_subvertices*i;
	  
	  /* The Center of the subtriangle is the centroid of its 
	     vertices */
	  subtriangle->center = triangle->subvertex[rank0].pos +
	    triangle->subvertex[rank1].pos; 
	  subtriangle->center = subtriangle->center + 
	    triangle->subvertex[rank2].pos;
	  subtriangle->center /= subtriangle->center.norm();

	  /* Define the neighbors of the subtriangle */
	  if (l<(frequency-1))
	    subtriangle->neighbor[0] = rank_triangle+ 2*(l+1);
	  else 
	     /* Quand un subtriangle se trouve en bord inferieur 
		du triangle*/
	    {
	      next_triangle = &(solid.triangle[triangle->neighbor[0]]);
	      ind = 0;
	      while (next_triangle->neighbor[ind] != i)
		ind++;
		    /* next_triangle est le triangle voisin du triangle
		       courant. ind est l'indice de l'arete commune avec
		       le triangle courant */
	      if (ind==0) {
		subtriangle->neighbor[0] = nb_subtriangles *
		  triangle->neighbor[0] + (l+1)*(l+1)-2*m-1;
	      }
	      if (ind==1) {
		subtriangle->neighbor[0] = nb_subtriangles *
		  triangle->neighbor[0] + (m+1)*(m+1)-1;
	      }
	      if (ind==2) {
		subtriangle->neighbor[0] = nb_subtriangles *
		  triangle->neighbor[0] + (frequency-m-1) *
		  (frequency-m-1)+1-1;
	      }
	    }
	  if (m<l)
	    subtriangle->neighbor[1]=rank_triangle+ 1 ;
	  else
		 /* Quand un subtriangle se trouve en bord droit
		    du triangle*/
	    {
	      next_triangle=&(solid.triangle[triangle->neighbor[1]]);
	      ind=0;
	      while (next_triangle->neighbor[ind]!=i)
		ind++;
	      if (ind==0) {
		subtriangle->neighbor[1] = nb_subtriangles*
		  triangle->neighbor[1]+frequency*frequency-2*
		  (frequency-1-l)-1;
	      }
	      if (ind==1) {
		subtriangle->neighbor[1] = nb_subtriangles*
		  triangle->neighbor[1]+(frequency-l+1)*
		  (frequency-l+1)-1;
	      }
	      if (ind==2) {
		subtriangle->neighbor[1] = nb_subtriangles*
		  triangle->neighbor[1]+l*l;
	      }
					
	    }
	  if (m>0)
	    subtriangle->neighbor[2]=rank_triangle -1 ;
	  else 
	      /*Quand un subtriangle se trouve en bord gauche du triangle*/
	    {
	      next_triangle=&(solid.triangle[triangle->neighbor[2]]);
	      ind=0;
	      while (next_triangle->neighbor[ind]!= i)
		ind++;
	      if (ind==0) {
		subtriangle->neighbor[2] = nb_subtriangles*
		  triangle->neighbor[2]+frequency*frequency-2*l-1;
	      }
	      if (ind==1) {
		subtriangle->neighbor[2] = nb_subtriangles*
		  triangle->neighbor[2]+(l+1)*(l+1)-1;
	      }
	      if (ind==2) {
		subtriangle->neighbor[2] = nb_subtriangles*
		  triangle->neighbor[2]+(frequency-l-1)*
		  (frequency-l-1)+1-1;
	      }  
	    }
	  subtriangle->rank = rank_triangle++;
	  
	}
	if ((m<l) && (l<frequency))
	  /* Deuxieme type de subtriangle */
	  {
	    subtriangle=&(triangle->subtriangle
			  [rank_triangle - i*nb_subtriangles]);
		   /* Define the vertices of the subtriangle */
	    rank0=rank_vertex;
	    rank1=(l+1)*(l+2)/2 +m+1;
	    rank2=rank_vertex+1;
	    subtriangle->vertex[0] = rank0 + nb_subvertices*i;
	    subtriangle->vertex[1] = rank1 + nb_subvertices*i;
	    subtriangle->vertex[2] = rank2 + nb_subvertices*i;
	    
	    /* The Center of the subtriangle is the centroid of 
	       its vertices */
	    subtriangle->center = triangle->subvertex[rank0].pos +
	      triangle->subvertex[rank1].pos; 
	    subtriangle->center += triangle->subvertex[rank2].pos;
	    subtriangle->center /= subtriangle->center.norm();

	    /* Define the neighbors of the subtriangle */
	    subtriangle->neighbor[0] = rank_triangle+1 ;
	    subtriangle->neighbor[1] = rank_triangle-2*l ;
	    subtriangle->neighbor[2] = rank_triangle-1 ;

	    subtriangle->rank=rank_triangle++;
	  }
      }
    }
  }
  
  list_pos = new Vec3<double>[solid.ntriangles * nb_subtriangles];

  for(f=0,i=0;i<solid.ntriangles;++i) {
    triangle=&(solid.triangle[i]);
    for (j=0;j<nb_subtriangles;++j,++f) {
      list_pos[f]=triangle->subtriangle[j].center;
    }
    delete[] triangle->subtriangle;
    delete[] triangle->subvertex;
  }
  delete[] solid.triangle;
  delete[] solid.vertex;
  delete[] solid.edge;
  
  return(list_pos);
}

/*-----------------------------------------*/
static double *getGaussian(int filterSize)
{
  int total_filterSize,i,j;
  double *smooth,y;
  
  total_filterSize=2*filterSize+5;
  smooth = new double[total_filterSize];
  
  memset(smooth, 0, sizeof(double) * total_filterSize);
  smooth[2] = 1.0;
  for(i = 2; i < 2*filterSize+2; i++) {
    for(j = i+1; j > 0; j--)
      smooth[j] = smooth[j-1] + smooth[j];
  }

  for(y=0,i=0; i<total_filterSize; i++) 
    y += smooth[i];
  for(i=0; i<total_filterSize; i++) 
    smooth[i] /= y;

  return(smooth);
}
/*-----------------------------------------*/
static double *getSobelFilter(int filterSize) {
  int total_filterSize,i;
  double *deriv, *gaus, y;

  total_filterSize=2*filterSize+1;
  gaus = getGaussian(filterSize-1);
  deriv = new double[total_filterSize];

  for(i = 0; i < total_filterSize; i++) {
    deriv[i] = gaus[i] - gaus[i+2];
  }
  delete[] gaus;

  y = 0.0;
  for(i = filterSize+1; i < total_filterSize; i++)
    y += deriv[i];
  for(i = 0; i < total_filterSize; i++)
    deriv[i] /= y;

  return(deriv);
}

/*-----------------------------------------*/
static double *getLaplacianFilter(int filterSize) {
  int total_filterSize,i;
  double *deriv, *gaus, yp, yn;

  total_filterSize=2*filterSize+1;
  gaus = getGaussian(filterSize-1);
  deriv = new double[total_filterSize];

  for(i = 0; i < total_filterSize; i++) {
    deriv[i] = gaus[i] -2.0*gaus[i+1] + gaus[i+2];
  }
  delete[] gaus;

  yp = yn = 0.0;
  for(i = 0; i < total_filterSize; i++) {
    if(deriv[i] > 0.0) yp += deriv[i];
    else yn -= deriv[i];
  }
  yp = (yp+yn) / 2.0;
  for(i = 0; i < total_filterSize; i++)
    deriv[i] /= yp;

  return(deriv);
}

/*-----------------------------------------*/
static double *getGaussianFilter(int filterSize)
{
  int total_filterSize,i,j;
  double *smooth,y;
  
  total_filterSize=2*filterSize+1;
  smooth = new double[total_filterSize];
  
  memset(smooth, 0, sizeof(double) * total_filterSize);
  smooth[0] = 1.0;
  for(i = 0; i < 2*filterSize; i++) {
    for(j = i+1; j > 0; j--)
      smooth[j] = smooth[j-1] + smooth[j];
  }

  for(y=0,i=0; i<total_filterSize; i++) 
    y += smooth[i];
  for(i=0; i<total_filterSize; i++) 
    smooth[i] /= y;

  return(smooth);
}


/*--------------------------------------------------------------------------*/
static void lineConvolution(double *input, double *output, int line_size,
			    double *filter, int filterSize) {
  register double *in, *out;
  int total_filter_width, center_size;
  register int i, j;
  
  total_filter_width = 2*filterSize+1;
  center_size = line_size - filterSize;

  /* process the beginning of the line */
  for(i=0,out=output;i<filterSize;++i,++out)
    {
      *out=0.0;
      /* Complete the line with the same value *input */
      for(in=input,j=0;j<(filterSize-i);++j)
	*out+= (*in)*filter[j];
	
      /* Do the rest of the convulution */
      for(j=filterSize-i;j<total_filter_width;++j,++in)
	 *out+= (*in)*filter[j];
    }
  /* process the center of the line */
  for(i=filterSize;i<center_size;++i,++out)
    {
      *out=0.0;
      /* Do the full convolution */
      for(in=input+i-filterSize,j=0;j<total_filter_width;++j,++in)
	*out+= (*in)*filter[j];	
    }
  /* process the center of the line */
  for(i=0;i<filterSize;++i,++out)
    {
      *out=0.0;
      /* Do the full convolution */
      for(in=input+center_size+i-filterSize,j=0;j<(total_filter_width-i-1);++j,++in)
	*out+= (*in)*filter[j];	

      for(in=input+line_size-1,j=total_filter_width-i-1;j<total_filter_width;++j)
	*out+= (*in)*filter[j];	
    }
}



/*----------------------------------------------------------------------*/
void gradient2DSobelImage(const Inrimage *inr, Inrimage *& dirvect,
				    unsigned int filterSize) {
  double *derivative_filter, *smoothing_filter;
  unsigned int xdim, ydim, zdim, vdim;
  xdim = inr->getX(); ydim = inr->getY();
  zdim = inr->getZ(); vdim = inr->getVdim();

  /* Get the two filters */
  derivative_filter = getSobelFilter(filterSize);
  smoothing_filter = getGaussianFilter(filterSize);

  dirvect = new Inrimage(xdim, ydim, zdim, Inrimage::WT_FLOAT_VECTOR, 2, 
			 VM_NON_INTERLACED);
  dirvect->copyHeader(inr);

  /* First process dr */
  /* Gradient in row */
  lineConvolute(((float ****) dirvect->getVarray())[1], inr, 
		       derivative_filter, filterSize, SLICE_Y);

  /* Smoothing in col */
  lineConvolute(((float ****) dirvect->getVarray())[1], dirvect, 
				smoothing_filter, filterSize, SLICE_X, 1);


  /* First process dc */
  /* Gradient in col */
  lineConvolute(((float ****) dirvect->getVarray())[0], inr, 
		       derivative_filter, filterSize, SLICE_X);
  /* Smoothing in row */
  lineConvolute(((float ****) dirvect->getVarray())[0], dirvect, 
				smoothing_filter, filterSize, SLICE_Y, 0);

  /* Free Memory */
  delete[] derivative_filter;
  delete[] smoothing_filter;
}


/*------------------------------------------------------------------------*/
void gradient3DSobelImage(const Inrimage *inr, Inrimage *& dirvect,
				    unsigned int filterSize) {
  double *derivative_filter,*smoothing_filter;
  double *derivative_filter_plane,*smoothing_filter_plane;
  int filterSizePlane;
  float vx, vy, vz;
  unsigned int xdim, ydim, zdim, vdim;
  xdim = inr->getX(); ydim = inr->getY();
  zdim = inr->getZ(); vdim = inr->getVdim();
  vx = inr->getVX(); vy = inr->getVY();
  vz = inr->getVZ();

  /* Get the two filters */
  derivative_filter = getSobelFilter(filterSize);
  smoothing_filter = getGaussianFilter(filterSize);
  /* Get the filter in z */
  filterSizePlane = (int) (filterSize * vx / vz);
  if(filterSizePlane == 0) filterSizePlane = 1;
  /* Get the two filters */
  derivative_filter_plane = getSobelFilter(filterSizePlane);
  smoothing_filter_plane = getGaussianFilter(filterSizePlane);

  dirvect = new Inrimage(xdim, ydim, zdim, Inrimage::WT_FLOAT_VECTOR, 3, 
			 VM_NON_INTERLACED);

  dirvect->copyHeader(inr);

  /* Smoothing in plane */
  lineConvolute(((float ****) dirvect->getVarray())[1], inr, 
		       smoothing_filter_plane, filterSizePlane, SLICE_Z);
  /* Copy the result in dc */
  memcpy((void *) &((float ****) dirvect->getVarray())[0][0][0][0],
	 (const void *) &((float ****) dirvect->getVarray())[1][0][0][0],
	 xdim * ydim * zdim * sizeof(float));
  /* Smoothing in col */
  lineConvolute(((float ****) dirvect->getVarray())[1], dirvect, 
				smoothing_filter, filterSize, SLICE_X, 1);
  /* Gradient in row */
  lineConvolute(((float ****) dirvect->getVarray())[1], dirvect, 
				derivative_filter, filterSize, SLICE_Y, 1);

  /* Then process dc */
  /* Smoothing in row */
  lineConvolute(((float ****) dirvect->getVarray())[0], dirvect, 
				smoothing_filter, filterSize, SLICE_Y, 0);
  /* Gradient in col */
  lineConvolute(((float ****) dirvect->getVarray())[0], dirvect, 
				derivative_filter, filterSize, SLICE_X, 0);

   /* Finally process dp */
  /* Smoothing in row */
  lineConvolute(((float ****) dirvect->getVarray())[2], inr, 
		       smoothing_filter, filterSize, SLICE_Y);
  /* Smoothing in row */
  lineConvolute(((float ****) dirvect->getVarray())[2], dirvect, 
				smoothing_filter, filterSize, SLICE_X, 2);
  /* Gradient in plane */
  lineConvolute(((float ****) dirvect->getVarray())[2], dirvect, 
				derivative_filter_plane, filterSizePlane,
				SLICE_Z, 2);

  /* Free Memory */
  delete[] derivative_filter;
  delete[] smoothing_filter;
  delete[] derivative_filter_plane;
  delete[] smoothing_filter_plane;
}


void convolute(Inrimage *inr, FILTER_TYPE filterType, int filterSize, bool twoD) {
  double *filter=0;
  double *filterPlane=0;
  int filterSizePlane;
  Inrimage::WORD_TYPE orgtype = inr->getType();
  float vx, vy, vz;
  vx = inr->getVX(); vy = inr->getVY();
  vz = inr->getVZ();

  switch(filterType) {
  case GAUSSIAN:
  case SMOOTH_X:
  case SMOOTH_Y:
  case SMOOTH_Z:
    filter = getGaussianFilter(filterSize);
    break;
  case DERIVATIVE_X:
  case DERIVATIVE_Y:
  case DERIVATIVE_Z:
    filter = getSobelFilter(filterSize);
    break;
  case SECOND_DERIVATIVE_X:
  case SECOND_DERIVATIVE_Y:
  case SECOND_DERIVATIVE_Z:
    filter = getLaplacianFilter(filterSize);
    break;
  }

  inr->convertCast(Inrimage::WT_FLOAT);

  switch(filterType) {
  case GAUSSIAN:
    /* Smoothing in row */
    lineConvolute((float ***) inr->getArray(), inr, filter, filterSize,
			 SLICE_Y);
    /* Smoothing in col */
    lineConvolute((float ***) inr->getArray(), inr, filter, filterSize,
			 SLICE_X);
    if(!twoD) {
      /* Get the filter in z */
      filterSizePlane = (int) (filterSize * vx / vz);
      if (filterSizePlane == 0) filterSizePlane = 1;
      filterPlane = getGaussianFilter(filterSizePlane);

      /* Smoothing in plane */
      lineConvolute((float ***) inr->getArray(), inr, 
			   filterPlane, filterSizePlane, SLICE_Z);

      delete[] filterPlane;
    }
    inr->convertCast(orgtype);
    break;

  case SMOOTH_X:
  case DERIVATIVE_X:
  case SECOND_DERIVATIVE_X:
    lineConvolute((float ***) inr->getArray(), inr, filter, filterSize,
			 SLICE_Y);
    break;
  case SMOOTH_Y:
  case DERIVATIVE_Y:
  case SECOND_DERIVATIVE_Y:
    lineConvolute((float ***) inr->getArray(), inr, filter, filterSize,
			 SLICE_X);
    break;
  case SMOOTH_Z:
  case DERIVATIVE_Z:
  case SECOND_DERIVATIVE_Z:
    lineConvolute((float ***) inr->getArray(), inr, filter, filterSize,
			 SLICE_Z);
    break;
  }
    
  delete[] filter;
}



void curvature(Inrimage *inr) {
  unsigned int xdim, ydim, zdim, vdim;
  xdim=inr->getX(); ydim = inr->getY();
  zdim = inr->getZ(); vdim = inr->getVdim();
  double dx, dy, ddxy, ddxx, ddyy, d;
  inr->convertCast(Inrimage::WT_FLOAT);
  Inrimage *tmp = new Inrimage(xdim, ydim, zdim, Inrimage::WT_FLOAT);
  tmp->copyHeader(inr);
  float ***src = (float ***) inr->getArray();
  float ***dst = (float ***) tmp->getArray();
  for(unsigned int z = 0; z < inr->getZ(); z++)
    for(unsigned int y = 1; y < inr->getY()-1; y++)
      for(unsigned int x = 1; x < inr->getX()-1; x++) {
	dx = (-src[z][y][x-1] + src[z][y][x+1]) * 0.5;
	dy = (-src[z][y-1][x] + src[z][y+1][x]) * 0.5;
	d = pow(dx*dx + dy*dy, 1.5);
	if(d > 0.0) {
	  ddxx = src[z][y][x-1] - 2*src[z][y][x] + src[z][y][x+1];
	  ddyy = src[z][y-1][x] - 2*src[z][y][x] + src[z][y+1][x];
	  ddxy = (src[z][y-1][x-1] - src[z][y-1][x+1]
		  - src[z][y+1][x-1] + src[z][y+1][x+1]) * 0.25;
	  dst[z][y][x] = (float)((ddxx*dy*dy - 2.0f*dx*dy*ddxy + ddyy*dx*dx) / d);
	}
	else dst[z][y][x] = 0.0;
      }
  memcpy(inr->getData(), tmp->getData(), inr->getX()*inr->getY()*inr->getZ()*inr->getVdim()*inr->getWdim());
  delete tmp;
}

/*********************************************************************/

#define LOG2(x) (log(x) / log(2))

// fast Fourrier transform

void fft(Inrimage *inr, bool twoD ) throw(Inrimage::InvalidSize) {

  char from[] = "yav::fft";
  float *src, *dest, *end, *ndata;
  int *dimar;
  unsigned long size, i, plane;
  unsigned int x, y, z;
  unsigned int xdim, ydim, zdim, vdim;
  xdim = inr->getX(); ydim = inr->getY();
  zdim = inr->getZ(); vdim = inr->getVdim();
  
  if (Inrimage::vectorial(inr->getType())) {
    
    // compute inverse DFT
    if(Inrimage::floating(inr->getType()) && vdim == 2) {

      // 2D DFT -1
      if(twoD) {

	/* format data */
	dimar = new int[2];
	i = (int) (log((double) xdim)/log(2.0) + (double)0.5);
	dimar[1] = 1 << i;
	if((unsigned int) dimar[1] < xdim) throw Inrimage::InvalidSize(from);
	i = (int) (log((double) ydim)/log(2.0) + 0.5);
	dimar[0] = 1 << i;
	if((unsigned int) dimar[0] < ydim) throw Inrimage::InvalidSize(from);

	/* compute DFT */
	plane = dimar[0] * dimar[1];
	for(z = 0; z < zdim; z++)
	  _fft((float *) inr->getData() + 2 * z * plane, dimar, 2, DFT_INVERSE);

	/* create result */
	size = dimar[0] * dimar[1] * zdim;
	ndata = new float[size];
	
	for(src = (float *) inr->getData(), dest = (float *) ndata, end = dest + size;
	    dest < end; dest++, src += 2)
	  *dest = *src;

	delete[] dimar;
      }

      // 3D DFT -1
      else {
	inr->setInterlacedData();

	/* format data */
	dimar = new int[3];
	i = (unsigned int) (log((double) xdim)/log(2.0) + 0.5);
	dimar[2] = 1 << i;
	if((unsigned int) dimar[2] < xdim) throw Inrimage::InvalidSize(from);
	i = (unsigned int) (log((double) ydim)/log(2.0) + 0.5);
	dimar[1] = 1 << i;
	if((unsigned int) dimar[1] < ydim) throw Inrimage::InvalidSize(from);
	i = (unsigned int) (log((double) zdim)/log(2.0) + 0.5);
	dimar[0] = 1 << i;
	if((unsigned int) dimar[0] < zdim) throw Inrimage::InvalidSize(from);

	/* compute DFT */
	_fft((float *) inr->getData(), dimar, 3, DFT_INVERSE);
	
	size = dimar[0] * dimar[1] * dimar[2];
	ndata = new float[size];

	for(dest = ndata, end = dest + size, src = (float *) inr->getData();
	    dest < end; dest++, src += 2)
	  *dest = *src;

	delete[] dimar;
      }

      inr->setType(Inrimage::WT_FLOAT);
      delete[] (unsigned char *) inr->setData(ndata);
    }
  }

  // compute DFT
  else {
    inr->convertCast(Inrimage::WT_FLOAT);

    // 2D DFT
    if(twoD) {

      /* format data */
      dimar = new int[2];
      i = (unsigned int) (log((double) xdim)/log(2.0) + 0.5);
      dimar[1] = 1 << i;
      if((unsigned int) dimar[1] < xdim) dimar[1] <<= 1;
      i = (unsigned int) (log((double) ydim)/log(2.0) + 0.5);
      dimar[0] = 1 << i;
      if((unsigned int) dimar[0] < ydim) dimar[0] <<= 1;

      size = dimar[1] * dimar[0] * zdim;
      ndata = new float[2 * size];
      memset((char *) ndata, 0, 2 * size * sizeof(float));

      src = (float *) inr->getData();
      plane = dimar[1] * dimar[0];
      for(z = 0; z < zdim; z++) {
	for(y = 0; y <ydim; y++) {
	  i = 2 * (z * plane + y * dimar[1]);
	  for(x = 0; x < xdim; x++, i += 2) {
	    ndata[i] = *src++;
	  }
	}
      }

      /* compute DFTs */
      for(z = 0; z < zdim; z++)
	_fft(ndata + plane * z * 2, dimar, 2, DFT_DIRECT);

      xdim = dimar[1];
      ydim = dimar[0];

      delete[] dimar;
    }

    // DFT 3D
    else {

      /* format data */
      dimar = new int[3];
      i = (unsigned int) (log((double) xdim)/log(2.0) + 0.5);
      dimar[2] = 1 << i;
      if((unsigned int) dimar[2] < xdim) dimar[2] <<= 1;
      i = (unsigned int) (log((double) ydim)/log(2.0) + 0.5);
      dimar[1] = 1 << i;
      if((unsigned int) dimar[1] < ydim) dimar[1] <<= 1;
      i = (unsigned int) (log((double) zdim)/log(2.0) + 0.5);
      dimar[0] = 1 << i;
      if((unsigned int) dimar[0] < zdim) dimar[0] <<= 1;

      size = dimar[2] * dimar[1] * dimar[0];
      ndata = new float[2 * size];
      memset((char *) ndata, 0, 2 * size * sizeof(float));

      src = (float *) inr->getData();
      plane = dimar[2] * dimar[1];
      for(z = 0; z < zdim; z++) {
	for(y = 0; y < ydim; y++) {
	  i = 2 * (z * plane + y * dimar[2]);
	  for(x = 0; x < xdim; x++, i += 2) {
	    ndata[i] = *src++;
	  }
	}
      }
      
      /* compute DFT */
      _fft(ndata, dimar, 3, DFT_DIRECT);
      
      xdim = dimar[2];
      ydim = dimar[1];
      zdim = dimar[0];

      delete[] dimar;
    }
    inr->setType(Inrimage::WT_FLOAT_VECTOR, 2);
    delete[] (unsigned char *) inr->setData(ndata);
  }
}



static void _fft(float *buf, int *dimar, int ndim, FFT_DIRECTION isign) {
#define SWAP_NB(a, b) tempr = (a); (a) = (b); (b) = tempr;

  int idim, *nn;
  unsigned long i1, i2, i3, i2rev, i3rev, ip1, ip2, ip3, ifp1, ifp2;
  unsigned long ibit, k1, k2, n, nprev, nrem, ntot;
  float tempi, tempr, *data;
  double theta, wi, wpi, wpr, wr, wtemp;

  data = buf - 1;
  nn = dimar - 1;

  /* array size */
  for(ntot=1, idim=1; idim <= ndim; idim++) ntot *= nn[idim];

  /* main loop over dimensions */
  nprev = 1;
  for(idim = ndim; idim >= 1; idim--) {

    /* bit reversal section */
    n = nn[idim];
    nrem = ntot / (n*nprev);
    ip1 = nprev << 1;
    ip2 = ip1 * n;
    ip3 = ip2 * nrem;
    i2rev = 1;
    for(i2 = 1; i2 <= ip2; i2 += ip1) {
      if(i2 < i2rev) {
	for(i1 = i2; i1 <= i2 + ip1 - 2; i1 += 2) {
	  for(i3 = i1; i3 <= ip3; i3 += ip2) {
	    i3rev = i2rev + i3 - i2;
	    SWAP_NB(data[i3], data[i3rev]);
	    SWAP_NB(data[i3+1], data[i3rev+1]);
	  }
	}
      }
      ibit = ip2 >> 1;
      while(ibit >= ip1 && i2rev > ibit) {
	i2rev -= ibit;
	ibit >>= 1;
      }
      i2rev += ibit;
    }

    /* Danielson-Lancrzos section */
    ifp1 = ip1;
    while(ifp1 < ip2) {
      ifp2 = ifp1 << 1;
      theta = isign*6.28318530717959 / (ifp2 / ip1);
      wtemp = sin(0.5*theta);
      wpr = -2.0*wtemp*wtemp;
      wpi = sin(theta);
      wr = 1.0;
      wi = 0.0;
      for(i3 = 1; i3 <= ifp1; i3 += ip1) {
	for(i1 = i3; i1 <= i3 + ip1 - 2; i1+=2) {
	  for(i2 = i1; i2 <= ip3; i2 += ifp2) {
	    /* Danielson-Lancrzos formula */
	    k1 = i2;
	    k2 = k1 + ifp1;
	    tempr = (float)wr*data[k2] - (float)wi*data[k2+1];
	    tempi = (float)wr*data[k2+1] + (float)wi*data[k2];
	    data[k2] = data[k1] - tempr;
	    data[k2+1] = data[k1+1] - tempi;
	    data[k1] += tempr;
	    data[k1+1] += tempi;
	  }
	}
	/* trigonometric recurrence */
	wr = (wtemp=wr)*wpr - wi*wpi + wr;
	wi = wi*wpr + wtemp*wpi + wi;
      }
      ifp1 = ifp2;
    }
    nprev *= n;
  }

  if(isign == DFT_INVERSE)
    for(i1 = 1; i1 <= 2*ntot; i1 += 2) data[i1] /= ntot;

#undef SWAP_NB
}

}
