#ifdef _MSC_VER 
#pragma warning ( disable : 4068 4786 4081 ) 
#endif 
 
#include <inrimage/sliceImage.h> 
#include <inrimage/VoxelIterator.h>  

namespace yav { 
 

template <class T>
void createProbabilityBuffer(const T *inputBuffer, float *output, const ProbabilityDensityFunction *pdf, unsigned int size) 
{
	unsigned int i;
	const T *in;
	float *out;
	for (i=0,in=inputBuffer, out=output;i<size;++i,++in,++out){
		(*out)=(float)pdf->getProbability((double)(*in));
	}
}
#define CREATEPDFBUFFER(TYPE,BUFIN,BUFOUT,PDF,SIZE) createProbabilityBuffer<TYPE>((TYPE *)BUFIN,(float *)BUFOUT,PDF,SIZE);
Inrimage *createProbabilityImage( Inrimage *inr,const ProbabilityDensityFunction *pdf)
{
	Inrimage *output= new Inrimage(inr->getX(),inr->getY(),inr->getZ(),
		Inrimage::WT_FLOAT,1,VM_SCALAR,
		inr->getVX(),inr->getVY(),inr->getVZ());
	unsigned int size=inr->getX()*inr->getY()*inr->getZ();

	switch(inr->getType()) {
  case Inrimage::WT_UNSIGNED_CHAR:
    CREATEPDFBUFFER(unsigned char,inr->getData(),output->getData(),pdf,size)
    break;
  case Inrimage::WT_SIGNED_CHAR:
    CREATEPDFBUFFER(signed char,inr->getData(),output->getData(),pdf,size)
    break;
  case Inrimage::WT_UNSIGNED_SHORT:
    CREATEPDFBUFFER(unsigned short,inr->getData(),output->getData(),pdf,size)
    break;
  case Inrimage::WT_SIGNED_SHORT:
    CREATEPDFBUFFER(signed short,inr->getData(),output->getData(),pdf,size)
    break;
  case Inrimage::WT_UNSIGNED_INT:
    CREATEPDFBUFFER(unsigned int,inr->getData(),output->getData(),pdf,size)
    break;
  case Inrimage::WT_SIGNED_INT:
    CREATEPDFBUFFER(int,inr->getData(),output->getData(),pdf,size)
    break;
  case Inrimage::WT_UNSIGNED_LONG:
    CREATEPDFBUFFER(unsigned long,inr->getData(),output->getData(),pdf,size)
    break;
  case Inrimage::WT_SIGNED_LONG:
    CREATEPDFBUFFER(LONGINT,inr->getData(),output->getData(),pdf,size)
    break;
  case Inrimage::WT_FLOAT:
    CREATEPDFBUFFER(float,inr->getData(),output->getData(),pdf,size)
    break;
  case Inrimage::WT_DOUBLE:
    CREATEPDFBUFFER(double,inr->getData(),output->getData(),pdf,size)
    break;
  default:
    break;
  }
	return(output);
}

	

	


Inrimage *createSliceImage(const Inrimage *input,
			   SLICE_ORIENTATION orientation,
			   unsigned int sliceNumber,
			   double scaleFactor,
			   bool useVoxelSize,
			   double minWindow,
			   double maxWindow,
			   bool interpolate,
			   unsigned char **colormap) {
  if (input==0) 
    return 0;
  int sizex=0, sizey=0;
  double vx,vy,vz;
  
  // set dummy voxel size
  if (useVoxelSize) {
    vx=input->getVX()*scaleFactor;
    vy=input->getVY()*scaleFactor;
    vz=input->getVZ()*scaleFactor;
  } else {
    vx=scaleFactor;
    vy=scaleFactor;
    vz=scaleFactor;
  }


  // computes size of output image and check sliceNumber
  switch(orientation) {
  case SLICE_Z:
    sizex=(unsigned int) (input->getX()*vx);
    sizey=(unsigned int)(input->getY()*vy);
    if (sliceNumber>=input->getZ())
      sliceNumber=input->getZ()-1;
    break;
  case SLICE_Y:
  case SLICE_Y_FLIPY:
  case SLICE_Y_FLIPX:
    sizex=(unsigned int) (input->getX()*vx);
    sizey=(unsigned int)(input->getZ()*vz);
    if (sliceNumber>=input->getY())
      sliceNumber=input->getY()-1;
    break;
  case SLICE_X_ROTATED:
    sizey=(unsigned int)(input->getZ()*vz);
    sizex=(unsigned int)(input->getY()*vy);
    if (sliceNumber>=input->getX())
      sliceNumber=input->getX()-1;
    break;
  case SLICE_X:
    sizex=(unsigned int)(input->getZ()*vz);
    sizey=(unsigned int)(input->getY()*vy);
    if (sliceNumber>=input->getX())
      sliceNumber=input->getX()-1;
    break;
  default:
    break;
  }
  // create output image
 Inrimage *output;
 if (colormap) {
   output= new Inrimage(sizex,sizey,1,Inrimage::WT_RGB,1,VM_SCALAR,
			1,1,1);
  }
  else
    // create grey level image
    output= new Inrimage(sizex,sizey,1,Inrimage::WT_UNSIGNED_CHAR,1,VM_SCALAR,
			 1,1,1);

  // get buffer on output image
  unsigned char *outbuf=(unsigned char *)output->getData();

  // get voxel on min and max image value in input image
  Voxel *voxmin=Voxel::voxelByType(input->getType());
  Voxel *voxmax=Voxel::voxelByType(input->getType());
  input->getBounds(voxmin,voxmax);
  
  // check the values of minWindow and maxWindow
  if (minWindow>=maxWindow) {
    minWindow=voxmin->getValueAsDouble();
    maxWindow=voxmax->getValueAsDouble();
  }
  else {
    if (minWindow<voxmin->getValueAsDouble()) 
      minWindow=voxmin->getValueAsDouble();
    if (minWindow>=voxmax->getValueAsDouble())       
      minWindow=voxmin->getValueAsDouble();
    if (maxWindow>voxmax->getValueAsDouble()) 
      maxWindow=voxmax->getValueAsDouble();
    if (maxWindow<voxmin->getValueAsDouble()) 
      maxWindow=voxmax->getValueAsDouble();
  }

  int x,y;
  unsigned char *p,v;

  unsigned int dp=input->getVdim();
  if (colormap) 
    dp=3;
  else dp=1;
  Voxel *vox=Voxel::voxelByType(input->getType());

  double a= 1.0/(maxWindow-minWindow);
  double val;
  unsigned int i,j,k;
  // must differentiate depending on slice orientation
  switch(orientation) {
  case SLICE_Z:
    k=sliceNumber;
    for (y=0,p=outbuf;y<sizey;y++) {
      j=(unsigned int) (y/vy);
      for (x=0;x<sizex;x++,p+=dp) {
	i=(unsigned int) (x/vx);

	if ((interpolate)&&(useVoxelSize))
	  input->getRealVoxel((double)(x/scaleFactor),(double)(y/scaleFactor),(double)(k*input->getVZ()),*vox);
	else {
	  input->getVoxel(i,j,k,*vox);
	}
	val=vox->getValueAsDouble();
	if (!colormap) {
	  if (val<minWindow)
	    (*p)=0;
	  else if  (val>maxWindow)
	    (*p)=255;
	  else {
	    (*p)=(unsigned char )(255*(val-minWindow)*a);
	  }
	}
	else {
	  if (val<minWindow) {
	    p[0]=colormap[0][0];
	    p[1]=colormap[1][0];
	    p[2]=colormap[2][0];
	  }  
	  else if  (val>maxWindow){
	    p[0]=colormap[0][256];
	    p[1]=colormap[1][256];
	    p[2]=colormap[2][256];
	  } 
	  else {
	    v=(unsigned char )(255*(val-minWindow)*a);
	    p[0]=colormap[0][v];
	    p[1]=colormap[1][v];
	    p[2]=colormap[2][v];
	  }
	}
      }
    }
    break;
  case SLICE_X:
    i=sliceNumber;
    for (y=0,p=outbuf;y<sizey;y++) {
      j=(unsigned int) (y/vy);
      for (x=0;x<sizex;x++,p+=dp) {
	k=(unsigned int) (x/vx);

	if ((interpolate)&&(useVoxelSize))
	  input->getRealVoxel(i*input->getVX(),y/scaleFactor,x/scaleFactor,*vox);
	else {
	  input->getVoxel(i,j,k,*vox);
	}

	val=vox->getValueAsDouble();
	if (!colormap) {
	  if (val<minWindow)
	    (*p)=0;
	  else if  (val>maxWindow)
	    (*p)=255;
	  else {
	    (*p)=(unsigned char )(255*(val-minWindow)*a);
	  }
	}
	else {
	  if (val<minWindow) {
	    p[0]=colormap[0][0];
	    p[1]=colormap[1][0];
	    p[2]=colormap[2][0];
	  }  
	  else if  (val>maxWindow){
	    p[0]=colormap[0][256];
	    p[1]=colormap[1][256];
	    p[2]=colormap[2][256];
	  } 
	  else {
	    v=(unsigned char )(255*(val-minWindow)*a);
	    p[0]=colormap[0][v];
	    p[1]=colormap[1][v];
	    p[2]=colormap[2][v];
	  }
	}
      }
    }
    break;
  case SLICE_X_ROTATED:
    i=sliceNumber;
    for (p=outbuf,x=0;x<sizex;x++) {
      k=(unsigned int) ((sizex-1-x)/vx);
      for (y=sizey-1;y>=0;y--,p+=dp) {

	if ((interpolate)&&(useVoxelSize))
	  input->getRealVoxel(i*input->getVX(),y/scaleFactor,(sizex-1-x)/scaleFactor,*vox);
	else {
	  j=(unsigned int) (y/vy);
	  input->getVoxel(i,j,k,*vox);
	}

	val=vox->getValueAsDouble();
	if (!colormap) {
	  if (val<minWindow)
	    (*p)=0;
	  else if  (val>maxWindow)
	    (*p)=255;
	  else {
	    (*p)=(unsigned char )(255*(val-minWindow)*a);
	  }
	}
	else {
	  if (val<minWindow) {
	    p[0]=colormap[0][0];
	    p[1]=colormap[1][0];
	    p[2]=colormap[2][0];
	  }  
	  else if  (val>maxWindow){
	    p[0]=colormap[0][256];
	    p[1]=colormap[1][256];
	    p[2]=colormap[2][256];
	  } 
	  else {
	    v=(unsigned char )(255*(val-minWindow)*a);
	    p[0]=colormap[0][v];
	    p[1]=colormap[1][v];
	    p[2]=colormap[2][v];
	  }
	}
      }
    }
    break;
  case SLICE_Y_FLIPX:
    j=sliceNumber;
    for (y=0,p=outbuf;y<sizey;y++) {
      k=(unsigned int) (y/vy);
      for (x=sizex-1;x>=0;x--,p+=dp) {
	if ((interpolate)&&(useVoxelSize))
	  input->getRealVoxel(x/scaleFactor,j*input->getVY(),y/scaleFactor,*vox);
	else {
	  i=(unsigned int) (x/vx);
	  input->getVoxel(i,j,k,*vox);
	}

	val=vox->getValueAsDouble();
	if (!colormap) {
	  if (val<minWindow)
	    (*p)=0;
	  else if  (val>maxWindow)
	    (*p)=255;
	  else {
	    (*p)=(unsigned char )(255*(val-minWindow)*a);
	  }
	}
	else {
	  if (val<minWindow) {
	    p[0]=colormap[0][0];
	    p[1]=colormap[1][0];
	    p[2]=colormap[2][0];
	  }  
	  else if  (val>maxWindow){
	    p[0]=colormap[0][256];
	    p[1]=colormap[1][256];
	    p[2]=colormap[2][256];
	  } 
	  else {
	    v=(unsigned char )(255*(val-minWindow)*a);
	    p[0]=colormap[0][v];
	    p[1]=colormap[1][v];
	    p[2]=colormap[2][v];
	  }
	}
      }
    }
    break;
  case SLICE_Y_FLIPY:
    j=sliceNumber;
    for (y=0,p=outbuf;y<sizey;y++) {
      k=(unsigned int) ((sizey-1-y)/vy);
      for (x=0;x<sizex;x++,p+=dp) {
	if ((interpolate)&&(useVoxelSize))
	  input->getRealVoxel(x/scaleFactor,j*input->getVY(),(sizey-1-y)/scaleFactor,*vox);
	else {
	  i=(unsigned int) (x/vx);
	  input->getVoxel(i,j,k,*vox);
	}

	val=vox->getValueAsDouble();
	if (!colormap) {
	  if (val<minWindow)
	    (*p)=0;
	  else if  (val>maxWindow)
	    (*p)=255;
	  else {
	    (*p)=(unsigned char )(255*(val-minWindow)*a);
	  }
	}
	else {
	  if (val<minWindow) {
	    p[0]=colormap[0][0];
	    p[1]=colormap[1][0];
	    p[2]=colormap[2][0];
	  }  
	  else if  (val>maxWindow){
	    p[0]=colormap[0][256];
	    p[1]=colormap[1][256];
	    p[2]=colormap[2][256];
	  } 
	  else {
	    v=(unsigned char )(255*(val-minWindow)*a);
	    p[0]=colormap[0][v];
	    p[1]=colormap[1][v];
	    p[2]=colormap[2][v];
	  }
	}
      }
    }
    break;
  case SLICE_Y:
    j=sliceNumber;
    for (y=0,p=outbuf;y<sizey;y++) {
      k=(unsigned int) (y/vy);
      for (x=0;x<sizex;x++,p+=dp) {
	i=(unsigned int) (x/vx);
	if ((interpolate)&&(useVoxelSize))
	  input->getRealVoxel(x/scaleFactor,j*input->getVY(),y/scaleFactor,*vox);
	else {
	  input->getVoxel(i,j,k,*vox);
	}

	val=vox->getValueAsDouble();
	if (!colormap) {
	  if (val<minWindow)
	    (*p)=0;
	  else if  (val>maxWindow)
	    (*p)=255;
	  else {
	    (*p)=(unsigned char )(255*(val-minWindow)*a);
	  }
	}
	else {
	  if (val<minWindow) {
	    p[0]=colormap[0][0];
	    p[1]=colormap[1][0];
	    p[2]=colormap[2][0];
	  }  
	  else if  (val>maxWindow){
	    p[0]=colormap[0][256];
	    p[1]=colormap[1][256];
	    p[2]=colormap[2][256];
	  } 
	  else {
	    v=(unsigned char )(255*(val-minWindow)*a);
	    p[0]=colormap[0][v];
	    p[1]=colormap[1][v];
	    p[2]=colormap[2][v];
	  }
	}
      }
    }
    break;
  default:
    break;
  }
  delete voxmin;
  delete voxmax;
  delete vox;
  return output;
}
 
} // namespace yav 
 
