#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/threshold.h"

#include <inrimage/threshold.h>

namespace yav {





#include "_threshold.C"






/* Threholding procedures
   with one or two thresholds
   one threshold : background is defined by I(x) <  min
                   foreground is defined by I(x) >= min
   two threshold : background is defined by I(x) <  min || I(x) >  max
                   foreground is defined by I(x) >= min && I(x) <= max
 */

void threshold(Inrimage *inr, double min )
  throw( Exception )
{
  Voxel *vmin = Voxel::voxelByType( inr->getType() );
  vmin->setValueFromDouble( min ) ;
  threshold( inr, vmin, (Voxel*)NULL );
  delete vmin;
}

void threshold(Inrimage *inr, double min, double max )
  throw( Exception )
{
  Voxel *vmin = Voxel::voxelByType( inr->getType() );
  Voxel *vmax = Voxel::voxelByType( inr->getType() );
  vmin->setValueFromDouble( min ) ;
  vmax->setValueFromDouble( max ) ;
  threshold( inr, vmin, vmax );
}

void threshold(Inrimage *inr, Voxel *min, Voxel *max ) 
  throw( Exception )
{
  // bg is set to 0
  Voxel *bg = Voxel::voxelByType( inr->getType() );
  // fg is set to max
  // to be verify on WIN
  Voxel *fg = Voxel::voxelByType( inr->getType() );
  fg->setMax();

  threshold( inr, min, max, bg, fg );
  delete bg;
  delete fg;
}

void threshold(Inrimage *inr, Voxel *min, Voxel *bg, Voxel *fg ) 
  throw( Exception )
{
    threshold( inr, min, (Voxel*)NULL, bg, fg );
}

void threshold(Inrimage *inr, Voxel *min, Voxel *max, Voxel *bg, Voxel *fg) 
  throw( ErrTypeMismatch )
{
  char *from = "threshold(Inrimage *, Voxel *, Voxel *, Voxel *, Voxel *)";
  unsigned int size = inr->getX() * inr->getY() * inr->getZ() * inr->getVdim();

  switch( inr->getType() ) {
  default :
    throw ErrTypeMismatch( from );
    return;
  case Inrimage::WT_UNSIGNED_CHAR:
    thresholdBuffer( (unsigned char *)inr->getData(), size,
		     dynamic_cast<Voxel8u *>(min), 
		     dynamic_cast<Voxel8u *>(max), 
		     dynamic_cast<Voxel8u *>(bg), 
		     dynamic_cast<Voxel8u *>(fg) );
    break;
  case Inrimage::WT_SIGNED_CHAR:
    thresholdBuffer( (char *)inr->getData(), size,
		     dynamic_cast<Voxel8s *>(min), 
		     dynamic_cast<Voxel8s *>(max), 
		     dynamic_cast<Voxel8s *>(bg), 
		     dynamic_cast<Voxel8s *>(fg) );
    break;
  case Inrimage::WT_UNSIGNED_SHORT:
    thresholdBuffer( (unsigned short int *)inr->getData(), size,
		     dynamic_cast<Voxel16u *>(min),
		     dynamic_cast<Voxel16u *>(max),
		     dynamic_cast<Voxel16u *>(bg),
		     dynamic_cast<Voxel16u *>(fg) );
    break;
  case Inrimage::WT_SIGNED_SHORT:
    thresholdBuffer( (short int *)inr->getData(), size,
		     dynamic_cast<Voxel16s *>(min),
		     dynamic_cast<Voxel16s *>(max),
		     dynamic_cast<Voxel16s *>(bg),
		     dynamic_cast<Voxel16s *>(fg) );
    break;
  case Inrimage::WT_UNSIGNED_INT:
    thresholdBuffer( (unsigned int *)inr->getData(), size,
		     dynamic_cast<Voxel32u *>(min),
		     dynamic_cast<Voxel32u *>(max),
		     dynamic_cast<Voxel32u *>(bg),
		     dynamic_cast<Voxel32u *>(fg) );
    break;
  case Inrimage::WT_SIGNED_INT:
    thresholdBuffer( (int *)inr->getData(), size,
		     dynamic_cast<Voxel32s *>(min),
		     dynamic_cast<Voxel32s *>(max),
		     dynamic_cast<Voxel32s *>(bg),
		     dynamic_cast<Voxel32s *>(fg) );
    break;
  case Inrimage::WT_UNSIGNED_LONG:
    thresholdBuffer( (unsigned LONGINT *)inr->getData(), size,
		     dynamic_cast<Voxel64u *>(min),
		     dynamic_cast<Voxel64u *>(max),
		     dynamic_cast<Voxel64u *>(bg),
		     dynamic_cast<Voxel64u *>(fg) );
    break;
  case Inrimage::WT_SIGNED_LONG:
    thresholdBuffer( (LONGINT *)inr->getData(), size,
		     dynamic_cast<Voxel64s *>(min),
		     dynamic_cast<Voxel64s *>(max),
		     dynamic_cast<Voxel64s *>(bg),
		     dynamic_cast<Voxel64s *>(fg) );
    break;
  case Inrimage::WT_FLOAT:
    thresholdBuffer( (float *)inr->getData(), size,
		     dynamic_cast<VoxelFloat *>(min),
		     dynamic_cast<VoxelFloat *>(max),
		     dynamic_cast<VoxelFloat *>(bg),
		     dynamic_cast<VoxelFloat *>(fg) );
    break;
  case Inrimage::WT_DOUBLE:
    thresholdBuffer( (double *)inr->getData(), size,
		     dynamic_cast<VoxelDouble *>(min),
		     dynamic_cast<VoxelDouble *>(max),
		     dynamic_cast<VoxelDouble *>(bg),
		     dynamic_cast<VoxelDouble *>(fg) );
    break;
  }
}





void threshold( Inrimage *inr, Inrimage::WORD_TYPE ntype,
	       Voxel *theMin, Voxel *theMax, Voxel *resBg, Voxel *resFg )    
  throw( ErrTypeMismatch )
{
  char *from = "threshold( Inrimage *, Inrimage::WORD_TYPE, Voxel *, Voxel *, Voxel *, Voxel * )";
  
  if ( inr->getType() == ntype ) {
    threshold( inr, theMin, theMax, resBg, resFg );
    return;
  }

  switch( ntype ) {
  default:
    throw ErrTypeMismatch( from );
    return;
  case Inrimage::WT_UNSIGNED_CHAR:
    thresholdImageToBuffer( (unsigned char *) 0, inr, ntype, theMin, theMax,
			   dynamic_cast<Voxel8u *>(resBg), 
			   dynamic_cast<Voxel8u *>(resFg) );
    break;
  case Inrimage::WT_SIGNED_CHAR:
    thresholdImageToBuffer( (char *) 0, inr, ntype, theMin, theMax,
			   dynamic_cast<Voxel8s *>(resBg), 
			   dynamic_cast<Voxel8s *>(resFg) );
    break;
  case Inrimage::WT_UNSIGNED_SHORT:
    thresholdImageToBuffer( (unsigned short int *) 0, inr, ntype, theMin, theMax,
			   dynamic_cast<Voxel16u *>(resBg), 
			   dynamic_cast<Voxel16u *>(resFg) );
    break;
  case Inrimage::WT_SIGNED_SHORT:
    thresholdImageToBuffer( (short int *) 0, inr, ntype, theMin, theMax,
			   dynamic_cast<Voxel16s *>(resBg), 
			   dynamic_cast<Voxel16s *>(resFg) );
    break;
  case Inrimage::WT_UNSIGNED_INT:
    thresholdImageToBuffer( (unsigned int *) 0, inr, ntype, theMin, theMax,
			   dynamic_cast<Voxel32u *>(resBg), 
			   dynamic_cast<Voxel32u *>(resFg) );
    break;
  case Inrimage::WT_SIGNED_INT:
    thresholdImageToBuffer( (int *) 0, inr, ntype, theMin, theMax,
			   dynamic_cast<Voxel32s *>(resBg), 
			   dynamic_cast<Voxel32s *>(resFg) );
    break;
  case Inrimage::WT_UNSIGNED_LONG:
    thresholdImageToBuffer( (unsigned LONGINT *) 0, inr, ntype, theMin, theMax,
			   dynamic_cast<Voxel64u *>(resBg), 
			   dynamic_cast<Voxel64u *>(resFg) );
    break;
  case Inrimage::WT_SIGNED_LONG:
    thresholdImageToBuffer( (LONGINT *) 0, inr, ntype, theMin, theMax,
			   dynamic_cast<Voxel64s *>(resBg), 
			   dynamic_cast<Voxel64s *>(resFg) );
    break;
  case Inrimage::WT_FLOAT:
    thresholdImageToBuffer( (float *) 0, inr, ntype, theMin, theMax,
			   dynamic_cast<VoxelFloat *>(resBg), 
			   dynamic_cast<VoxelFloat *>(resFg) );
    break;
  case Inrimage::WT_DOUBLE:
    thresholdImageToBuffer( (double *) 0, inr, ntype, theMin, theMax,
			   dynamic_cast<VoxelDouble *>(resBg), 
			   dynamic_cast<VoxelDouble *>(resFg) );
    break;
  }
}
  



  








  




/** "binarize" procedures 
 */
void binarize( Inrimage *inr, double bg, double fg )
  throw( Exception )
{
  Voxel *vbg = Voxel::voxelByType( inr->getType() );
  Voxel *vfg = Voxel::voxelByType( inr->getType() );
  vbg->setValueFromDouble( bg ) ;
  vfg->setValueFromDouble( fg ) ;
  binarize( inr, vbg, vbg, vfg );
}

void binarize( Inrimage *inr, Voxel *bg, Voxel *fg )
  throw( Exception )
{
  binarize( inr, bg, bg ,fg );
}

void binarize( Inrimage *inr, Voxel *theBg, Voxel *resBg, Voxel *resFg ) 
    throw( ErrTypeMismatch )
{
  char *from = "binarize( Inrimage *, Voxel *, Voxel *, Voxel * )";
  unsigned int size = inr->getX() * inr->getY() * inr->getZ() * inr->getVdim();

  switch( inr->getType() ) {
  default:
    throw ErrTypeMismatch( from );
    return;
  case Inrimage::WT_UNSIGNED_CHAR:
    binarizeBuffer( (unsigned char *)inr->getData(), size,
		    dynamic_cast<Voxel8u *>(theBg),
		    dynamic_cast<Voxel8u *>(resBg),
		    dynamic_cast<Voxel8u *>(resFg) );
    break;
  case Inrimage::WT_SIGNED_CHAR:
    binarizeBuffer( (char *)inr->getData(), size,
		    dynamic_cast<Voxel8s *>(theBg),
		    dynamic_cast<Voxel8s *>(resBg),
		    dynamic_cast<Voxel8s *>(resFg) );
    break;
  case Inrimage::WT_UNSIGNED_SHORT:
    binarizeBuffer( (unsigned short int *)inr->getData(), size,
		    dynamic_cast<Voxel16u *>(theBg),
		    dynamic_cast<Voxel16u *>(resBg),
		    dynamic_cast<Voxel16u *>(resFg) );
    break;
  case Inrimage::WT_SIGNED_SHORT:
    binarizeBuffer( (short int *)inr->getData(), size,
		    dynamic_cast<Voxel16s *>(theBg),
		    dynamic_cast<Voxel16s *>(resBg),
		    dynamic_cast<Voxel16s *>(resFg) );
    break;
  case Inrimage::WT_UNSIGNED_INT:
    binarizeBuffer( (unsigned int *)inr->getData(), size,
		    dynamic_cast<Voxel32u *>(theBg),
		    dynamic_cast<Voxel32u *>(resBg),
		    dynamic_cast<Voxel32u *>(resFg) );
    break;
  case Inrimage::WT_SIGNED_INT:
    binarizeBuffer( (int *)inr->getData(), size,
		    dynamic_cast<Voxel32s *>(theBg),
		    dynamic_cast<Voxel32s *>(resBg),
		    dynamic_cast<Voxel32s *>(resFg) );
    break;
  case Inrimage::WT_UNSIGNED_LONG:
    binarizeBuffer( (unsigned LONGINT *)inr->getData(), size,
		    dynamic_cast<Voxel64u *>(theBg),
		    dynamic_cast<Voxel64u *>(resBg),
		    dynamic_cast<Voxel64u *>(resFg) );
    break;
  case Inrimage::WT_SIGNED_LONG:
    binarizeBuffer( (LONGINT *)inr->getData(), size,
		    dynamic_cast<Voxel64s *>(theBg),
		    dynamic_cast<Voxel64s *>(resBg),
		    dynamic_cast<Voxel64s *>(resFg) );
    break;
  case Inrimage::WT_FLOAT:
    binarizeBuffer( (float *)inr->getData(), size,
		    dynamic_cast<VoxelFloat *>(theBg),
		    dynamic_cast<VoxelFloat *>(resBg),
		    dynamic_cast<VoxelFloat *>(resFg) );
    break;
  case Inrimage::WT_DOUBLE:
    binarizeBuffer( (double *)inr->getData(), size,
		    dynamic_cast<VoxelDouble *>(theBg),
		    dynamic_cast<VoxelDouble *>(resBg),
		    dynamic_cast<VoxelDouble *>(resFg) );
    break;
  }
}



void binarizeToOtherFormat( Inrimage *inr, Inrimage::WORD_TYPE ntype,
			   Voxel *theBg, Voxel *resBg, Voxel *resFg )    
  throw( Exception )
{
  fprintf( stderr, "WARNING: binarizeToOtherFormat() is deprecated\n" );
  fprintf( stderr, "\t use binarize() instead\n" );
  binarize( inr, ntype, theBg, resBg, resFg );
}



void binarize( Inrimage *inr, Inrimage::WORD_TYPE ntype,
	       Voxel *theBg, Voxel *resBg, Voxel *resFg )    
  throw( ErrTypeMismatch )
{
  char *from = "binarize( Inrimage *, Inrimage::WORD_TYPE, Voxel *, Voxel *, Voxel * )";
  
  if ( inr->getType() == ntype ) {
    binarize( inr, theBg, resBg, resFg );
    return;
  }

  switch( ntype ) {
  default:
    throw ErrTypeMismatch( from );
    return;
  case Inrimage::WT_UNSIGNED_CHAR:
    binarizeImageToBuffer( (unsigned char *) 0, inr, ntype, theBg,
			   dynamic_cast<Voxel8u *>(resBg), 
			   dynamic_cast<Voxel8u *>(resFg) );
    break;
  case Inrimage::WT_SIGNED_CHAR:
    binarizeImageToBuffer( (char *) 0, inr, ntype, theBg,
			   dynamic_cast<Voxel8s *>(resBg), 
			   dynamic_cast<Voxel8s *>(resFg) );
    break;
  case Inrimage::WT_UNSIGNED_SHORT:
    binarizeImageToBuffer( (unsigned short int *) 0, inr, ntype, theBg,
			   dynamic_cast<Voxel16u *>(resBg), 
			   dynamic_cast<Voxel16u *>(resFg) );
    break;
  case Inrimage::WT_SIGNED_SHORT:
    binarizeImageToBuffer( (short int *) 0, inr, ntype, theBg,
			   dynamic_cast<Voxel16s *>(resBg), 
			   dynamic_cast<Voxel16s *>(resFg) );
    break;
  case Inrimage::WT_UNSIGNED_INT:
    binarizeImageToBuffer( (unsigned int *) 0, inr, ntype, theBg,
			   dynamic_cast<Voxel32u *>(resBg), 
			   dynamic_cast<Voxel32u *>(resFg) );
    break;
  case Inrimage::WT_SIGNED_INT:
    binarizeImageToBuffer( (int *) 0, inr, ntype, theBg,
			   dynamic_cast<Voxel32s *>(resBg), 
			   dynamic_cast<Voxel32s *>(resFg) );
    break;
  case Inrimage::WT_UNSIGNED_LONG:
    binarizeImageToBuffer( (unsigned LONGINT *) 0, inr, ntype, theBg,
			   dynamic_cast<Voxel64u *>(resBg), 
			   dynamic_cast<Voxel64u *>(resFg) );
    break;
  case Inrimage::WT_SIGNED_LONG:
    binarizeImageToBuffer( (LONGINT *) 0, inr, ntype, theBg,
			   dynamic_cast<Voxel64s *>(resBg), 
			   dynamic_cast<Voxel64s *>(resFg) );
    break;
  case Inrimage::WT_FLOAT:
    binarizeImageToBuffer( (float *) 0, inr, ntype, theBg,
			   dynamic_cast<VoxelFloat *>(resBg), 
			   dynamic_cast<VoxelFloat *>(resFg) );
    break;
  case Inrimage::WT_DOUBLE:
    binarizeImageToBuffer( (double *) 0, inr, ntype, theBg,
			   dynamic_cast<VoxelDouble *>(resBg), 
			   dynamic_cast<VoxelDouble *>(resFg) );
    break;
  }
}
  



  




/** "ternarize" procedures 
 */
void ternarize(Inrimage *inr, double lt, double ht, 
	       double bg, double lf, double hf)
    throw( Exception )
{
 ternarize( inr, inr->getType(), lt, ht, bg, lf, hf );
}



void ternarize( Inrimage *inr, Voxel *low, Voxel *high, Voxel *val1, Voxel *val2, Voxel *val3 ) 
  throw( ErrTypeMismatch )
{
  char *from = "ternarize( Inrimage*, Voxel*, Voxel*, Voxel*, Voxel*, Voxel* )";
  unsigned int size = inr->getX() * inr->getY() * inr->getZ() * inr->getVdim();

  switch( inr->getType() ) {
  default:
    throw ErrTypeMismatch( from );
    return;
  case Inrimage::WT_UNSIGNED_CHAR:
    ternarizeBuffer( (unsigned char *)inr->getData(), size, 
		     dynamic_cast<Voxel8u *>(low), 
		     dynamic_cast<Voxel8u *>(high), 
		     dynamic_cast<Voxel8u *>(val1), 
		     dynamic_cast<Voxel8u *>(val2), 
		     dynamic_cast<Voxel8u *>(val3) );
    break;
  case Inrimage::WT_SIGNED_CHAR:
    ternarizeBuffer( (char *)inr->getData(), size, 
		     dynamic_cast<Voxel8s *>(low), 
		     dynamic_cast<Voxel8s *>(high), 
		     dynamic_cast<Voxel8s *>(val1), 
		     dynamic_cast<Voxel8s *>(val2), 
		     dynamic_cast<Voxel8s *>(val3) );
    break;
  case Inrimage::WT_UNSIGNED_SHORT:
    ternarizeBuffer( (unsigned short int *)inr->getData(), size, 
		     dynamic_cast<Voxel16u *>(low), 
		     dynamic_cast<Voxel16u *>(high), 
		     dynamic_cast<Voxel16u *>(val1), 
		     dynamic_cast<Voxel16u *>(val2), 
		     dynamic_cast<Voxel16u *>(val3) );
    break;
  case Inrimage::WT_SIGNED_SHORT:
    ternarizeBuffer( (short int *)inr->getData(), size, 
		     dynamic_cast<Voxel16s *>(low), 
		     dynamic_cast<Voxel16s *>(high), 
		     dynamic_cast<Voxel16s *>(val1), 
		     dynamic_cast<Voxel16s *>(val2), 
		     dynamic_cast<Voxel16s *>(val3) );
    break;
  case Inrimage::WT_UNSIGNED_INT:
    ternarizeBuffer( (unsigned int *)inr->getData(), size, 
		     dynamic_cast<Voxel32u *>(low), 
		     dynamic_cast<Voxel32u *>(high), 
		     dynamic_cast<Voxel32u *>(val1), 
		     dynamic_cast<Voxel32u *>(val2), 
		     dynamic_cast<Voxel32u *>(val3) );
    break;
  case Inrimage::WT_SIGNED_INT:
    ternarizeBuffer( (int *)inr->getData(), size, 
		     dynamic_cast<Voxel32s *>(low), 
		     dynamic_cast<Voxel32s *>(high), 
		     dynamic_cast<Voxel32s *>(val1), 
		     dynamic_cast<Voxel32s *>(val2), 
		     dynamic_cast<Voxel32s *>(val3) );
    break;
  case Inrimage::WT_UNSIGNED_LONG:
    ternarizeBuffer( (unsigned LONGINT *)inr->getData(), size, 
		     dynamic_cast<Voxel64u *>(low), 
		     dynamic_cast<Voxel64u *>(high), 
		     dynamic_cast<Voxel64u *>(val1), 
		     dynamic_cast<Voxel64u *>(val2), 
		     dynamic_cast<Voxel64u *>(val3) );
    break;
  case Inrimage::WT_SIGNED_LONG:
    ternarizeBuffer( (LONGINT *)inr->getData(), size, 
		     dynamic_cast<Voxel64s *>(low), 
		     dynamic_cast<Voxel64s *>(high), 
		     dynamic_cast<Voxel64s *>(val1), 
		     dynamic_cast<Voxel64s *>(val2), 
		     dynamic_cast<Voxel64s *>(val3) );
    break;
  case Inrimage::WT_FLOAT:
    ternarizeBuffer( (float *)inr->getData(), size, 
		     dynamic_cast<VoxelFloat *>(low), 
		     dynamic_cast<VoxelFloat *>(high), 
		     dynamic_cast<VoxelFloat *>(val1), 
		     dynamic_cast<VoxelFloat *>(val2), 
		     dynamic_cast<VoxelFloat *>(val3) );
    break;
  case Inrimage::WT_DOUBLE:
    ternarizeBuffer( (double *)inr->getData(), size, 
		     dynamic_cast<VoxelDouble *>(low), 
		     dynamic_cast<VoxelDouble *>(high), 
		     dynamic_cast<VoxelDouble *>(val1), 
		     dynamic_cast<VoxelDouble *>(val2), 
		     dynamic_cast<VoxelDouble *>(val3) );
    break;
  }
}
  


void ternarizeToOtherFormat( Inrimage *inr, Inrimage::WORD_TYPE ntype, 
			    Voxel *min, Voxel *max,
			    Voxel *bg, Voxel *val1, Voxel *val2 ) 
  throw( Exception )
{
  fprintf( stderr, "WARNING: ternarizeToOtherFormat() is deprecated\n" );
  fprintf( stderr, "\t use ternarize() instead\n" );
  ternarize( inr, ntype, min, max, bg, val1, val2 );
}


 
void ternarize(Inrimage *inr, Inrimage::WORD_TYPE ntype, double lt, double ht, 
	       double bg, double lf, double hf)
    throw( Exception )
{
  Voxel *vlow = Voxel::voxelByType( inr->getType() );
  Voxel *vhigh = Voxel::voxelByType( inr->getType() );
  Voxel *vval1 = Voxel::voxelByType( ntype );
  Voxel *vval2 = Voxel::voxelByType( ntype );
  Voxel *vval3 = Voxel::voxelByType( ntype );

  vlow->setValueFromDouble( lt ) ;    
  vhigh->setValueFromDouble( ht ) ;
  vval1->setValueFromDouble( bg ) ;
  vval2->setValueFromDouble( lf ) ;
  vval3->setValueFromDouble( hf ) ;

  ternarize( inr, ntype, vlow, vhigh, vval1,  vval2,  vval3 );

  delete vval3;
  delete vval2;
  delete vval1;
  delete vhigh;
  delete vlow;
}



void ternarize( Inrimage *inr, Inrimage::WORD_TYPE ntype, 
		Voxel *min, Voxel *max,
		Voxel *val1, Voxel *val2, Voxel *val3 ) 
  throw( ErrTypeMismatch )
{
  char *from = "ternarize( Inrimage*, Inrimage::WORD_TYPE, Voxel*, Voxel*, Voxel*, Voxel*, Voxel* )";
  
  if ( inr->getType() == ntype ) {
    ternarize( inr, min, max, val1, val2, val3 );
    return;
  }
  switch( ntype ) {
  default:
    throw ErrTypeMismatch( from );
    return;
  case Inrimage::WT_UNSIGNED_CHAR:
    ternarizeImageToBuffer( (unsigned char *)0, inr, ntype, min, max,
			    dynamic_cast<Voxel8u *>(val1), 
			    dynamic_cast<Voxel8u *>(val2), 
			    dynamic_cast<Voxel8u *>(val3) );
    break;
  case Inrimage::WT_SIGNED_CHAR:
    ternarizeImageToBuffer( (char *)0, inr, ntype, min, max,
			    dynamic_cast<Voxel8s *>(val1), 
			    dynamic_cast<Voxel8s *>(val2), 
			    dynamic_cast<Voxel8s *>(val3) );
    break;
  case Inrimage::WT_UNSIGNED_SHORT:
    ternarizeImageToBuffer( (unsigned short int *)0, inr, ntype, min, max,
			    dynamic_cast<Voxel16u *>(val1), 
			    dynamic_cast<Voxel16u *>(val2), 
			    dynamic_cast<Voxel16u *>(val3) );
    break;
  case Inrimage::WT_SIGNED_SHORT:
    ternarizeImageToBuffer( (short int *)0, inr, ntype, min, max,
			    dynamic_cast<Voxel16s *>(val1), 
			    dynamic_cast<Voxel16s *>(val2), 
			    dynamic_cast<Voxel16s *>(val3) );
    break;
  case Inrimage::WT_UNSIGNED_INT:
    ternarizeImageToBuffer( (unsigned int *)0, inr, ntype, min, max,
			    dynamic_cast<Voxel32u *>(val1), 
			    dynamic_cast<Voxel32u *>(val2), 
			    dynamic_cast<Voxel32u *>(val3) );
    break;
  case Inrimage::WT_SIGNED_INT:
    ternarizeImageToBuffer( (int *)0, inr, ntype, min, max,
			    dynamic_cast<Voxel32s *>(val1), 
			    dynamic_cast<Voxel32s *>(val2), 
			    dynamic_cast<Voxel32s *>(val3) );
    break;
  case Inrimage::WT_UNSIGNED_LONG:
    ternarizeImageToBuffer( (unsigned LONGINT *)0, inr, ntype, min, max,
			    dynamic_cast<Voxel64u *>(val1), 
			    dynamic_cast<Voxel64u *>(val2), 
			    dynamic_cast<Voxel64u *>(val3) );
    break;
  case Inrimage::WT_SIGNED_LONG:
    ternarizeImageToBuffer( (LONGINT *)0, inr, ntype, min, max,
			    dynamic_cast<Voxel64s *>(val1), 
			    dynamic_cast<Voxel64s *>(val2), 
			    dynamic_cast<Voxel64s *>(val3) );
    break;
  case Inrimage::WT_FLOAT:
    ternarizeImageToBuffer( (float *)0, inr, ntype, min, max,
			    dynamic_cast<VoxelFloat *>(val1), 
			    dynamic_cast<VoxelFloat *>(val2), 
			    dynamic_cast<VoxelFloat *>(val3) );
    break;
  case Inrimage::WT_DOUBLE:
    ternarizeImageToBuffer( (double *)0, inr, ntype, min, max,
			    dynamic_cast<VoxelDouble *>(val1), 
			    dynamic_cast<VoxelDouble *>(val2), 
			    dynamic_cast<VoxelDouble *>(val3) );
    break;
  }
}






















  

  template <class TYPE, class V>
    void cropHistogrammBuffer(TYPE *T, V *v, Inrimage *inr, V *vmin, V *vmax, V *vbg) {
    
    void *data=inr->getData();
    TYPE *buf, *end;
    TYPE mn, mx, bbg;
    
    mn = vmin->getValue();
    mx = vmax->getValue();
    bbg = vbg->getValue();
    
    for(buf = (TYPE *) data, end = buf + inr->getX() * inr->getY() * inr->getZ(); buf < end; buf++) {
      if(*buf < mn || *buf > mx) 
	*buf = bbg;
    }
  }
  






template <class TYPE, class V>
 void windowBuffer(TYPE *T, unsigned char *out, Inrimage *inr, V *vmin, V *vmax) {
  
  void *data=inr->getData();
  TYPE *buf, *end;
  TYPE mn, mx;
  unsigned char *o;
  
  double scale=1.0;
  
  
  mn = vmin->getValue();
  mx = vmax->getValue();
  if (mx>mn)
    {
      scale=255.0/(mx-mn);
    }
  
  for(o=out,buf = (TYPE *) data, end = buf + inr->getX() * inr->getY() * inr->getZ(); buf < end; buf++,o++) {
    if (*buf<mn)
      *o=0;
    else if (*buf>mx)
      *o=255;
    else {
      *o=(unsigned char)((*buf-mn)*scale);
    }
  }
} 

Inrimage *windowing(Inrimage *inr, Voxel *min, Voxel *max) {
  Inrimage *out=new Inrimage(inr->getX(),inr->getY(),inr->getZ(),1,1,WK_FIXED,SGN_UNSIGNED,inr->getVX(),inr->getVY(),inr->getVZ());
  switch(inr->getType()) {
  case Inrimage::WT_UNSIGNED_CHAR :
    windowBuffer((unsigned char *) 0,(unsigned char *)out->getData(),inr,
		 dynamic_cast<Voxel8u *>(min), 
		 dynamic_cast<Voxel8u *>(max));
    break; 
  case Inrimage::WT_SIGNED_CHAR :
    windowBuffer((char *) 0,(unsigned char *)out->getData(),inr,
		 dynamic_cast<Voxel8s *>(min), 
		 dynamic_cast<Voxel8s *>(max));
    break; 
  case Inrimage::WT_UNSIGNED_SHORT :
    windowBuffer((unsigned short *) 0,(unsigned char *)out->getData(),inr,
		 dynamic_cast<Voxel16u *>(min), 
		 dynamic_cast<Voxel16u *>(max));
    break;
 case Inrimage::WT_SIGNED_SHORT :
    windowBuffer(( short *) 0,(unsigned char *)out->getData(),inr,
		 dynamic_cast<Voxel16s *>(min), 
		 dynamic_cast<Voxel16s *>(max));
    break;
  case Inrimage::WT_UNSIGNED_INT :
    windowBuffer((unsigned int *) 0,(unsigned char *)out->getData(),inr,
		 dynamic_cast<Voxel32u *>(min), 
		 dynamic_cast<Voxel32u *>(max));
    break;
  case Inrimage::WT_SIGNED_INT :
    windowBuffer(( int *) 0,(unsigned char *)out->getData(),inr,
		 dynamic_cast<Voxel32s *>(min), 
		 dynamic_cast<Voxel32s *>(max));
    break;  
  case Inrimage::WT_UNSIGNED_LONG :
    windowBuffer((unsigned LONGINT *) 0,(unsigned char *)out->getData(),inr,
		 dynamic_cast<Voxel64u *>(min), 
		 dynamic_cast<Voxel64u *>(max));
    break;
  case Inrimage::WT_SIGNED_LONG :
    windowBuffer(( LONGINT *) 0,(unsigned char *)out->getData(),inr,
		 dynamic_cast<Voxel64s *>(min), 
		 dynamic_cast<Voxel64s *>(max));
    break;
  case Inrimage::WT_FLOAT :
    windowBuffer((float *) 0,(unsigned char *)out->getData(),inr,
		 dynamic_cast<VoxelFloat *>(min), 
		 dynamic_cast<VoxelFloat *>(max));
    break;
  case Inrimage::WT_DOUBLE :
    windowBuffer((double *) 0,(unsigned char *)out->getData(),inr,
		 dynamic_cast<VoxelDouble *>(min), 
		 dynamic_cast<VoxelDouble *>(max));
    break;
  default:
    break;
  };
  return out;
}




void cropHistogramm(Inrimage *inr, Voxel *min, Voxel *max, Voxel *bg) {
  switch(inr->getType()) {
  case Inrimage::WT_UNSIGNED_CHAR:
    cropHistogrammBuffer((unsigned char *) 0, (Voxel8u *) 0, inr, dynamic_cast<Voxel8u *>(min), 
			 dynamic_cast<Voxel8u *>(max), 
			 dynamic_cast<Voxel8u *>(bg));
    break;
  case Inrimage::WT_SIGNED_CHAR:
    cropHistogrammBuffer((char *) 0, (Voxel8s *) 0, inr, dynamic_cast<Voxel8s *>(min), 
			 dynamic_cast<Voxel8s *>(max), 
			 dynamic_cast<Voxel8s *>(bg));
    break;
  case Inrimage::WT_UNSIGNED_SHORT:
    cropHistogrammBuffer((unsigned short int *) 0, (Voxel16u *) 0, inr, dynamic_cast<Voxel16u *>(min),
			 dynamic_cast<Voxel16u *>(max),
			 dynamic_cast<Voxel16u *>(bg));
    break;
  case Inrimage::WT_SIGNED_SHORT:
    cropHistogrammBuffer((short int *) 0, (Voxel16s *) 0, inr, dynamic_cast<Voxel16s *>(min),
			 dynamic_cast<Voxel16s *>(max),
			 dynamic_cast<Voxel16s *>(bg));
    break;
  case Inrimage::WT_UNSIGNED_INT:
    cropHistogrammBuffer((unsigned int *) 0, (Voxel32u *) 0, inr, dynamic_cast<Voxel32u *>(min),
			 dynamic_cast<Voxel32u *>(max),
			 dynamic_cast<Voxel32u *>(bg));
    break;
  case Inrimage::WT_SIGNED_INT:
    cropHistogrammBuffer((int *) 0, (Voxel32s *) 0, inr, dynamic_cast<Voxel32s *>(min),
			 dynamic_cast<Voxel32s *>(max),
			 dynamic_cast<Voxel32s *>(bg));
    break;
  case Inrimage::WT_UNSIGNED_LONG:
    cropHistogrammBuffer((unsigned LONGINT *) 0, (Voxel64u *) 0, inr, dynamic_cast<Voxel64u *>(min),
			 dynamic_cast<Voxel64u *>(max),
			 dynamic_cast<Voxel64u *>(bg));
    break;
  case Inrimage::WT_SIGNED_LONG:
    cropHistogrammBuffer((LONGINT *) 0, (Voxel64s *) 0, inr, dynamic_cast<Voxel64s *>(min),
			 dynamic_cast<Voxel64s *>(max),
			 dynamic_cast<Voxel64s *>(bg));
    break;
  case Inrimage::WT_FLOAT:
    cropHistogrammBuffer((float *) 0, (VoxelFloat *) 0, inr, dynamic_cast<VoxelFloat *>(min),
			 dynamic_cast<VoxelFloat *>(max),
			 dynamic_cast<VoxelFloat *>(bg));
    break;
  case Inrimage::WT_DOUBLE:
    cropHistogrammBuffer((double *) 0, (VoxelDouble *) 0, inr, dynamic_cast<VoxelDouble *>(min),
			 dynamic_cast<VoxelDouble *>(max),
			 dynamic_cast<VoxelDouble *>(bg));
    break;
  default:
    break;
  }
}
  
  
  
  
  

  
  


}
