#include <iostream>
#include "itkTestMain.h"
#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"
#include "itkImage.h"
#include "mipsInrimageImageIOFactory.h"

int mipsImageIOReadWriteTest (int argc, char *argv[]) {

  // Verify the number of parameters in the command line
  if( argc < 3 )
  {
    std::cerr << "Usage: " << argc<<std::endl;
    std::cerr << argv[0] << " inputImageFile  outputImageFile " << std::endl;
    return -1;
  }
  
  typedef  float      PixelType;
  const   unsigned int        Dimension = 10;
  typedef itk::Image< PixelType, Dimension >    ImageType;
  typedef itk::ImageFileReader< ImageType >  ReaderType;
  typedef itk::ImageFileWriter< ImageType >  WriterType;

  ReaderType::Pointer reader = ReaderType::New();
  WriterType::Pointer writer = WriterType::New();

  const char * inputFilename  = argv[1];
  const char * outputFilename = argv[2];

  std::cout << "In file: " << inputFilename << std::endl;
  std::cout << "Out file: " << outputFilename << std::endl;
  
  reader->SetFileName( inputFilename  );
  writer->SetFileName( outputFilename );
  writer->SetInput( reader->GetOutput() );

  std::cout << "Reading..." << std::flush;
  try
    {
      reader->Update();
    }
  catch(itk::ExceptionObject &e)
    {
      std::cerr << e;
      return 1;
    }
  std::cout << "Done." << std::endl;
  

  std::cout << "Writing..." << std::flush;
  try 
    { 
    writer->Update(); 
    } 
  catch( itk::ExceptionObject & err ) 
    { 
    std::cerr << "ExceptionObject caught !" << std::endl; 
    std::cerr << err << std::endl; 
    return 1;
    } 
  std::cout << "Done." << std::endl;

  std::cout << "Convert operation completed" << std::endl;
  return 0;
}


void RegisterTests() {
  itk::ObjectFactoryBase::
    RegisterFactory(itk::InrimageImageIOFactory::New() );

  REGISTER_TEST(mipsImageIOReadWriteTest);
}
