/*=========================================================================

Program:   Insight Segmentation & Registration Toolkit
Module:    $RCSfile: itkAnalyzeImageIOFactory.h,v $
Language:  C++
Date:      $Date: 2002/12/20 18:22:43 $
Version:   $Revision: 1.1 $

Copyright (c) 2002 Insight Consortium. All rights reserved.
See ITKCopyright.txt or http://www.itk.org/HTML/Copyright.htm for details.

This software is distributed WITHOUT ANY WARRANTY; without even
the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#ifndef __mipsInrimageImageIOFactory_h
#define __mipsInrimageImageIOFactory_h

#include "itkObjectFactoryBase.h"
#include "itkImageIOBase.h"

namespace itk
{
  /** \class InrimageImageIOFactory
   * \brief Create instances of InrimageImageIO objects using an object factory.
   */
  class ITK_EXPORT InrimageImageIOFactory : public ObjectFactoryBase
  {
    public:
      /** Standard class typedefs. */
      typedef InrimageImageIOFactory   Self;
      typedef ObjectFactoryBase  Superclass;
      typedef SmartPointer<Self>  Pointer;
      typedef SmartPointer<const Self>  ConstPointer;

      /** Class methods used to interface with the registered factories. */
      virtual const char* GetITKSourceVersion(void) const;
      virtual const char* GetDescription(void) const;

      /** Method for class instantiation. */
      itkFactorylessNewMacro(Self);

      /** Run-time type information (and related methods). */
      itkTypeMacro(InrimageImageIOFactory, ObjectFactoryBase);

      /** Register one factory of this type  */
      static void RegisterOneFactory(void)
      {
        InrimageImageIOFactory::Pointer metaFactory = InrimageImageIOFactory::New();
        ObjectFactoryBase::RegisterFactory(metaFactory);
      }

    protected:
      InrimageImageIOFactory();
      ~InrimageImageIOFactory();
      virtual void PrintSelf(std::ostream& os, Indent indent) const;

    private:
      InrimageImageIOFactory(const Self&); //purposely not implemented
      void operator=(const Self&); //purposely not implemented

  };


} // end namespace itk

#endif
