#include <inrimage/Inrimage.h>

using namespace yav;

int usage(char* pName)
{
	CERR << "usage: " << pName << " <-help | input_image1=%s input_image2=%s output_image=%s start_overlap=%i end_overlap=%i" << ENDL;
	return(0);
}

/* Main: let's go */
int main(int argc, char** argv)
{
 try
 {
	// Check input parameters
	for(int l=0; l<argc; l++)
	{
      	if(!strcmp(argv[l], "-help"))
			return usage(argv[0]);
    	}
    	if(argc!=6)
		return usage(argv[0]);

	unsigned int i, j, k, widthImage1, widthImage2, heightImage1, heightImage2, depthImage1, depthImage2, startOverlap, endOverlap, sizeOverlap, midOverlap;
	double vx1, vx2, vy1, vy2, vz1, vz2, partIm1, partIm2, voxelVal;
	Inrimage::WORD_TYPE myType1, myType2;

	Inrimage* inputImage1 = new Inrimage(argv[1]);
	Inrimage* inputImage2 = new Inrimage(argv[2]);
	startOverlap = (unsigned int) atoi(argv[4]);
	endOverlap = (unsigned int) atoi(argv[5]);
	midOverlap = (endOverlap+startOverlap)/2;
	sizeOverlap = endOverlap-startOverlap;
	myType1 = inputImage1->getType();
	myType2 = inputImage2->getType();
	Voxel* voxel1 = Voxel::voxelByType(myType1);
	Voxel* voxel2 = Voxel::voxelByType(myType2);
	Voxel* voxel3 = Voxel::voxelByType(myType1);

	// Check size of both input images
	widthImage1 = inputImage1->ncols();
	widthImage2 = inputImage2->ncols();
	heightImage1 = inputImage1->nrows();
	heightImage2 = inputImage2->nrows();
	depthImage1 = inputImage1->nplanes();
	depthImage2 = inputImage2->nplanes();

	if((widthImage1!=widthImage2)||(heightImage1!=heightImage2)||(depthImage1!=depthImage2))
		CERR << "ERROR: Size of both input images do not match" << ENDL;

	// Voxel size of both input images
	vx1 = inputImage1->getVX();
	vx2 = inputImage2->getVX();
	vy1 = inputImage1->getVY();
	vy2 = inputImage2->getVY();
	vz1 = inputImage1->getVZ();
	vz2 = inputImage2->getVZ();

	// Create output image
	Inrimage* outputImage = new Inrimage(*inputImage1);

	// Place images in arrays
	for (k=0; k<depthImage1; k++)
	{
		for (j=0; j<heightImage1; j++)
		{
			for (i=0; i<widthImage1; i++)
			{
				inputImage1->getVoxel(i,j,k,*voxel1);
				inputImage2->getVoxel(i,j,k,*voxel2);

				if ((k>=startOverlap)&&(k<=endOverlap))
				{
					partIm1 = (double)(k-startOverlap)/(double)sizeOverlap;
					partIm2 = 1.0-partIm1;
					voxelVal = ((partIm1*voxel1->getValueAsDouble())+(partIm2*voxel2->getValueAsDouble()));
				}
				else
					voxelVal = voxel1->getValueAsDouble()+voxel2->getValueAsDouble();

				voxel3->setValueFromDouble(voxelVal);
				outputImage->setVoxel(i,j,k,*voxel3);
			}
		}
	}

	// Write the modified mesh in the output	
	outputImage->write(argv[3]);
 }
 catch(Exception& e)
 {
	CERR << e << ENDL;
 }
 return(0);
}
