#include <inrimage/anisotropicDiffusion.h>

using namespace yav;


static char *usage= "[image-in] [image-out]\n\
[-dt %f] [-tmax %f] [-quantx %f] [-quanty %f] [-quantz %f] [-threshold %f]\n\
[-lisse] [-sigma %f] [-diffusionFunc %df] [-disk]\n\
[-diffusionType %dt] [-numScheme %sc] [-help]";

static char *detail = "Pour comprendre la diffusion et les termes employes : A Review of Nonlinear Diffusion Filtering, Joachim Weickert\n\
Disponible en ligne a http://www.cvgpr.uni-mannheim.de/weickert/Papers/ss97_inv.ps.gz\n\
'image-in' : image entree\n\
'image-out' :image sortie \n\
-dt est le pas de temps (environ 0.1 si on choisit un schema explicite,\n\
 jusqu a 1 si on choisit le schema AOS)\n\
-tmax est le temps de diffusion (tmax doit etre superieur a dt,\n\
 tmax = 1 diffusion moyenne tmax = 5 diffusion forte)\n\
-quantx, -quanty, -quantz : Quantile du gradient dans les differentes directions x, y et z \n\
Valeurs entre 0 et 100. Ce seuil sur le gradient permet de decider a partir de quel gradient\n\
 on considere que l on a un bord. On ne peut choisir un quantile dans chaque direction que \n\
pour  la diffusion scalaire, sinon seul quantx est considere\n\
Si on prefere un seuil constant, il suffit d'attribuer une valeur positive de son choix \n\
a -threshold. \n\
-lisse  si on veut  prealablement le gradient (conseille)\n\
Si oui -sigma donne la variance de ce lissage gaussien\n\
-diffusionFunc est une fonction de diffusion choisie parmi  'aubert', 'perona', 'perona_2', 'weickert' et 'green'\n\
Ces fonctions sont decrites entre autres dans la these de K.Krissian (annexe A).\n\
-disk Si pour la diffusion scalaire on veut economiser de la meoire en sauvegardant sur le disque\n\
-diffusionType doit etre pris parmi 'scalar' (these de K. Krissian p69),\n\
 'matrix' ( A Review of Nonlinear Diffusion Filtering, Joachim Weickert p 14)\n\
,'krissian' ou 'matkrissian' (these de K.Krissian p107 pour la premiere et 108 pour la deuxieme)\n\
'krissian' ne peut fonctionner qu'en 3D, 'matkrissian' fonctionne en 2D et en 3D mais est tres long en 3D\n\
-numScheme a choisir parmi 'AOS' (scema semi implicite dimension par dimension) et 'explicite'\n\
'AOS' n'est implemente que pour la diffusion scalaire";


static char *program = NULL;

static void _errorMessage( char *str, int flag )
{
  CERR<<" Usage: "<<program<<" "<<usage<<"\n"<<ENDL;
  if ( flag == 1 )
    CERR<<detail;
  CERR<<" Error: "<<str<<ENDL;
  exit( -1 );
}

int main(int argc, char *argv[])
{
 
  char *inputImage  = NULL;
  char *outputImage = NULL;
  float dt=0.1f;
  float tmax=1.0;
  float quantx=80, quanty=80, quantz=80, threshold=-1.0;
  bool lisse=0;
  float sigma=1;

  DIFFUSION_FUNCTION df=WEICKERT;
  DIFFUSION_TYPE td = SCALAR;
  NUMERICAL_SCHEME sc = EXPLICIT;

  bool disk=0;
  
  Inrimage *image = NULL; 
  Inrimage *output = NULL;


  program = argv[0];
  if ( argc == 1 ) _errorMessage( "\n", 0 );
    
  //
  // lecture des arguments
  //

  for ( int nimages = 0, i=1; i<argc; i++ ) {
    int status;
    // arguments
    if ( argv[i][0] == '-' ) {
      if ( argv[i][1] == '\0' ) {
	// standard input
	if ( nimages == 0 ) {
	  nimages ++;
	}
	else {
	  _errorMessage( "standard input specified twice\n", 0 );
	}
      }
      else if ( strcmp ( argv[i], "-help" ) == 0 ) {
	_errorMessage( "\n", 1 );
      }
     
      //times
      else if (strcmp ( argv[i], "-dt") == 0){
	i += 1;
	if ( i >= argc)    _errorMessage( "parsing -dt...\n", 0 );
	status = sscanf( argv[i],"%f",&dt );
	if ( status <= 0 ) _errorMessage( "parsing -dt...\n", 0 );
      }
      else if (strcmp ( argv[i], "-tmax") == 0){
	i += 1;
	if ( i >= argc)    _errorMessage( "parsing -tmax...\n", 0 );
	status = sscanf( argv[i],"%f",&tmax );
	if ( status <= 0 ) _errorMessage( "parsing -tmax...\n", 0 );
      }

      //gradient threshold
      else if (strcmp ( argv[i], "-quantx")==0){
	i += 1;
	if ( i >= argc)    _errorMessage( "parsing -quantx...\n", 0 );
	status = sscanf( argv[i],"%f",&quantx );
	if ( status <= 0 ) _errorMessage( "parsing -quantx...\n", 0 );
      }
       else if (strcmp ( argv[i], "-quanty")==0){
	i += 1;
	if ( i >= argc)    _errorMessage( "parsing -quanty...\n", 0 );
	status = sscanf( argv[i],"%f",&quanty );
	if ( status <= 0 ) _errorMessage( "parsing -quanty...\n", 0 );
      }
      else if (strcmp ( argv[i], "-quantz")==0){
	i += 1;
	if ( i >= argc)    _errorMessage( "parsing -quantz...\n", 0 );
	status = sscanf( argv[i],"%f",&quantz );
	if ( status <= 0 ) _errorMessage( "parsing -quantz...\n", 0 );
      }
      else if (strcmp ( argv[i], "-threshold")==0){
	i += 1;
	if ( i >= argc)    _errorMessage( "parsing -threshold...\n", 0 );
	status = sscanf( argv[i],"%f",&threshold );
	if ( status <= 0 ) _errorMessage( "parsing -threshold...\n", 0 );
      }
      
      //regularisation
      else if (strcmp ( argv[i], "-lisse")==0){
	lisse=1;
      }
      else if (strcmp ( argv[i], "-sigma")==0){
	i+=1;
	if ( i >= argc)    _errorMessage( "parsing -sigma...\n", 0 );
	status = sscanf( argv[i],"%f",&sigma );
	if ( status <= 0 ) _errorMessage( "parsing -sigma...\n", 0 );
      }

      // description de la diffusion
      else if (strcmp ( argv[i], "-diffusionFunc")==0){
	i+=1;
	if ( i >= argc)    _errorMessage( "parsing -diffusionFunc...\n", 0 );
	if (!strcmp(argv[i],"perona"))
	  df=PERONA_MALIK;
	else if (!strcmp(argv[i],"aubert"))
	  df=AUBERT;
	else if (!strcmp(argv[i],"weickert"))
	  df=WEICKERT; 
	else if (!strcmp(argv[i],"green"))
	  df=GREEN;
	else if (!strcmp(argv[i],"perona_2"))
	  df=PERONA_MALIK_2;
	else _errorMessage( "parsing -diffusionFunc...\n", 1 );
      }
      else if (strcmp ( argv[i], "-diffusionType")==0){
	i+=1;
	if ( i >= argc)    _errorMessage( "parsing -diffusionType...\n", 0 );
	if (!strcmp(argv[i],"scalar"))
	  td=SCALAR;
	else if (!strcmp(argv[i],"matrix"))
	  td=MATRIX;
	else if (!strcmp(argv[i],"krissian"))
       td=KRISSIAN; 
	else if (!strcmp(argv[i],"matkrissian"))
	  td=MATRIX_KRISSIAN;
	else _errorMessage( "parsing -diffusionType...\n", 1 );
      }
      
      else if (strcmp ( argv[i], "-numScheme")==0){
	i+=1;
	if ( i >= argc)    _errorMessage( "parsing -numScheme...\n", 0 ); 
	if (!strcmp(argv[i],"explicite"))
	  sc=EXPLICIT;
	else if (!strcmp(argv[i],"aos"))
	  sc=AOS_scheme;
	else _errorMessage( "parsing -numSchem...\n", 1 );
      }

      //disque
      else if(strcmp ( argv[i], "-disk")==0){
	disk=1;
      }

      
      /*--- option inconnue ---*/
      else {
	char message[256];
	sprintf( message, "unknown option '%s'\n", argv[i] );
	_errorMessage( message, 0 );
      }
    }
     // image names
    else {
      if ( nimages == 0 ) {
	inputImage = argv[i];
	nimages ++;
      }
      else if ( nimages == 1 ) {
	outputImage = argv[i];
	nimages ++;
      }
      else {
	_errorMessage( "too much file names when parsing\n", 0 );
      }
    }
  }

  //
  // fin de la lecture des arguments
  //
  


  try {

    //
    // lecture de l'image
    //
    image = new Inrimage( inputImage ); 
    output = new Inrimage( inputImage );
#ifdef USE_MTL
    anisotropicDiffusion(image, output,dt,tmax,quantx, quanty,quantz,threshold,lisse,sigma,df,disk,td,sc);
#else
    CERR << "Need MTL Library, recompile with USE_MTL ON" << ENDL;
#endif
  }
  catch( Exception e)  {
    CERR << e << ENDL;
    exit( -1 );
  }
  //
  // ecriture
  // 
  output->write( outputImage );
  delete image;
  return( 0 );
}
