#include <basics/TclString.h>
#include <inrimage/Inrimage.h>

#define MAX_NUMBER_LABELS 50

using namespace yav;

int usage(char* pName)
{
	CERR << "usage: " << pName << " <-help | input_image=%s" << ENDL;
	return(0);
}

/* Main: let's go */
int main(int argc, char** argv)
{
 try
 {
	unsigned int h, i, j, k, l, nbrPlanes, nbrRows, nbrCols, nlabel;
	double label;
	char buffer [50];
	bool dejavu;
	Inrimage::WORD_TYPE wt;

	// Check input parameters
	for(i=0; i<(unsigned int)argc; i++)
	{
      		if(!strcmp(argv[i], "-help"))
			return usage(argv[0]);
    	}
    	if(argc!=2)
		return usage(argv[0]);

	// Take care of the input parameter
	Inrimage* inputImage = new Inrimage(argv[1]);
	wt = inputImage->getType();
	Voxel* voxel = Voxel::voxelByType(wt);

	// Initialization
	double* labels = new double [MAX_NUMBER_LABELS];

	for(i=0; i<MAX_NUMBER_LABELS; i++)
		labels[i] = 0.0;

	// Find the labels
	nbrPlanes = inputImage->nplanes();
	nbrRows   = inputImage->nrows();
	nbrCols   = inputImage->ncols();
	nlabel = 0;

	for(i=0; i<nbrPlanes; i++)
	{
		for(j=0; j<nbrRows; j++)
		{
			for(k=0; k<nbrCols; k++)
			{
				inputImage->getVoxel(k,j,i,*voxel);
				label = voxel->getValueAsDouble();

				if (label!=0.0)
				{
					dejavu = false;

					// Does the label already exist?
					for(l=0; l<MAX_NUMBER_LABELS; l++)
						if (labels[l]==label)
							dejavu = true;

					// If not, add it to the labels
					if (!dejavu)
					{
						labels[nlabel] = label;
						nlabel++;
					
						// Check if the number of labels reaches the maximum allowed
						if (nlabel==MAX_NUMBER_LABELS)
						{
							CERR << "WARNING: Number of labels in the image is greater than the maximum allowed (" << MAX_NUMBER_LABELS << "). Only the " << MAX_NUMBER_LABELS << " first labels will be taken into account..." << ENDL;
							goto exit;
						}
					}
				}
			}
		}
	}

	exit:

	// Print number of labels
	CERR << nlabel << " labels found in the image..." << ENDL;

	// Create inrimages in a loop
	for(h=0; h<nlabel; h++)
	{
		CERR << "Writing label image " << h << " of intensity " << labels[h] << "...";
	
		Inrimage* binaryImage = new Inrimage(nbrCols,nbrRows,nbrPlanes,wt,1,inputImage->getVectorialMode(),inputImage->getVX(),inputImage->getVY(),inputImage->getVZ());

		for(i=0; i<nbrPlanes; i++)
		{
			for(j=0; j<nbrRows; j++)
			{
				for(k=0; k<nbrCols; k++)
				{
					inputImage->getVoxel(k,j,i,*voxel);
					label = voxel->getValueAsDouble();
	
					if (label==labels[h])
					{
						voxel->setValueFromDouble(255.0);
						binaryImage->setVoxel(k,j,i,*voxel);
					}
					else
						voxel->setValueFromDouble(0.0);
				}
			}
		}

		// Write the binary image
		sprintf(buffer,"Label%i_Intensity%3.3f.inr.gz",h,labels[h]);
		binaryImage->write((TclString)buffer);
		CERR << " done" << ENDL;

		delete binaryImage;
	}
 }
 catch(Exception& e)
 {
	CERR << e << ENDL;
 }
 return(0);
}
