/** change Voxel Size
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <inrimage/ImageIO.h>


static char *options="\
input vx vy vz [-h|-help]";

static char *details="\
\n\
Modify the size of each voxel and overwrite the image \n\
\n\
input : name of the input file\n\
vx vy vz : the size in mmm of each voxel in the x, y and z direction\n\
-help : display this message\n";



void error( char *name, 
	    int fullhelp,
	    char *error,
	    int exitcode )
{
  fprintf( stderr, "%s %s\n", name, options );
  if ( fullhelp ) 
    fprintf( stderr, "%s\n", details );
  if ( error )
    fprintf( stderr, "error: %s\n", error );
  exit( exitcode );
}



int main ( int argc, char *argv[] )
{
  int i, n=0;
  char *input = NULL;
  _image *theIm;
  int ret;
  double vx,vy,vz= -1.0;
  vx=vy= -1.0;

  if ( argc == 1 ) {
    error( argv[0], 0, NULL, 0 );
  }

  for ( i=1; i<argc; i++ ) {
    if ( argv[i][0] == '-' ) {
      if ( strlen( argv[i] ) == 1 ) {
	if ( n == 0 ) {
	  input = argv[i];
	}
	else {
	  error( argv[0], 0, "too many filenames or \'-\'", -1 );
	}
	n++;
      }
      else if ( (strncmp( argv[i], "-help", 5 ) == 0 && strlen( argv[i] ) == 5)
		|| (strncmp( argv[i], "-h", 2 ) == 0 && strlen( argv[i] ) == 2) ) {
	error( argv[0], 1, NULL, 0 );
      }
    }
    else {
      if ( n == 0 ) {
	input = argv[i];
      } 
	  else if ((n==1) && (i <=(argc-3))) {
		  vx= (double) atof(argv[i++]);
		  vy= (double) atof(argv[i++]);
		  vz= (double) atof(argv[i++]);
	  }
      else {
	error( argv[0], 0, "too many filenames or \'-\'", -1 );
      }
      n++;
    }
  }

  if ( 0 )
    fprintf( stderr, "... reading '%s'\n", input );
  theIm = _readImage( input );

  if ( theIm == NULL ) {
    error( argv[0], 0, " when reading input", -1 );
  }
  /* set voxel size */

  theIm->vx=vx;
  theIm->vy=vy;
  theIm->vz=vz;

  if ( 0 )
    fprintf( stderr, "... writing %s\n", input );
  ret =  _writeImage( theIm, input );

  if ( ret != 0 ) {
    error( argv[0], 0, " when writing output", -1 );
  }

  exit( 0 );
}
