/***************************************************************************
 * zmorpho.C -
 *
 * $Author: mtraina $
 * $Revision: 1.6 $
 * $Log: zconnexe.C,v $
 * Revision 1.6  2003/06/06 13:58:15  mtraina
 * Correction for win32 build
 *
 * Revision 1.5  2003/06/03 13:09:59  greg
 * split string for windows compilation
 *
 * Revision 1.4  2003/05/28 13:37:32  greg
 * Add an informative online help
 *
 * Revision 1.3  2002/02/19 13:47:46  jdlemare
 * correct a bug with the fillContours function
 *
 * Revision 1.2  2002/02/07 15:42:58  jdlemare
 * use CERR and ENDL instead of cerr and endl
 *
 * Revision 1.1  2002/02/01 13:20:23  jdlemare
 * *** empty log message ***
 *
 *
 *
 * $Id: zconnexe.C,v 1.6 2003/06/06 13:58:15 mtraina Exp $
 ***************************************************************************/

#include <inrimage/ConnectingElement.h>
#include <inrimage/arithmetic.h>

#include <time.h>

using namespace yav;


typedef enum {
  _COMPONENTS_,
  _HYSTERESIS_,
  _FILL_CONTOURS_,
  _RECONSTRUCT_
} enumComputation;


static char *usage = "[image-in] [image-out] [-2D] [-binary|-labels]\n\
   [-components|-comp | -fill-contours|-fill | -hysteresis|-hyst]\n\
   [-connectivity|-connexite|-con|-c %d]\n\
   [-low-threshold|-lt|-sb %lf] [-high-threshold|-ht|-sh %lf]\n\
   [-high-threshold-minimal-size|-high-ms|-hms %d] [-max]\n\
   [-maximal-number-of-connected-components|-max-cc|-mcc|-ncc %d]\n\
   [-minimal-size|-ms|-tcc %d] [-seed %d %d [%d]] [-seeds %s]\n\
   [-work-type|-wtype|-wt %d] [-times]\n\
   [-inverse|-inv] [-help] [-o %d] [-f|-r] [-s]"; 

static char *detail1 = "\
Extract and label connected components in an image\n\
\n\
  if 'image-in' is '-' or absent, read from stdin\n\
  if 'image-out' is '-' or absent, write on stdout\n\
-2D # slice by slice computation (only with -fill)\n\
-binary # binary output\n\
-components # labels the connected components\n\
-connectivity # specifies the connectivity (4, 8, 6, 10, 18, 26)\n\
-fill-contours # consider the binary input images as closed contours\n\
   and try to fill them\n\
-hysteresis # hysteresis thresholding, uses both low and high thresholds.\n\
   Selected components are characterized by\n\
   o all points are above the low threshold\n\
   o at least one (or more if -high-threshold-minimal-size %d is used) point\n\
     is above the high threshold\n\
   o a size greater than the one provided by -minimal-size %d\n\
-inv # invert the input image before processing\n\
-labels # in the output image, each individual component got its own label\n\
   (from 1 to the number of components)\n\
-low-threshold|-lt|-sb %lf  # low threshold to binarize input image\n\
   (default is 1.0)\n\
-help # display this message\n\
-high-threshold|-ht|-sh %lf # high threshold (default is 1.0)\n\
   it implies hysteresis thresholding  and binary output (this can be\n\
   changed by using other options afterwards)\n\
-high-threshold-minimal-size %d # the minimal number of points above the high\n\
   threshold that a component must have to be further considered\n";

static char *detail2 = "\
-max # only the largest (in size) component is retained\n\
-maximal-number-of-connected-components %d #  maximal number of components to\n\
   be retained. Only the largest ones are retained, and, as a side effect, they\n\
   are ordered and labeled from the largest one (label=1) to the smallest one.\n\
-minimal-size|-ms|-tcc %d # minimal size of selected connected components\n\
-seed %d %d [%d] # will retain the components containing the seed points\n\
   This option can be repeated.\n\
-seeds %s # specifies a binary image whose foreground is considered as seed\n\
   points.\n\
-times # print times (reading, processing, writing)\n\
-work-type %d # number of bytes for the auxiliary label image encoding.\n\
   Default is 2. 1 can be useful to process large images, and 4 can be used\n\
   if a (very) large number of components are to be found.\
output type image can be specified with\n\
-s: signed type\n\
-f: floating point type (float or double)\n\
-o %s: number of bytes per point\n\
\n\
   $Revision: 1.6 $ $Date: 2003/06/06 13:58:15 $ $Author: mtraina $\n"; 

static char *program = NULL;

static void _errorMessage( char *str, int flag )
{
  CERR<<" Usage: "<<program<<" "<<usage<<"\n"<<ENDL;
  if ( flag == 1 )
    CERR<<detail1<<detail2;
  CERR<<" Error: "<<str<<ENDL;
  exit( -1 );
}



int main(int argc, char *argv[])
{
  int o=0,s=0,r=0;

  char *inputImage  = NULL;
  char *outputImage = NULL;
  char *seedsImage = NULL;

  Inrimage *image = NULL;
  Inrimage *seeds = NULL;
  Inrimage::WORD_TYPE outputType = Inrimage::WT_UNSIGNED_CHAR;
  bool outputTypeIsSet = false;
  bool thereIsAHighThreshold = false;

  ConnectingElement CE;
  
  double low_threshold = 1.0;
  double high_threshold = 1.0;
  int x, y, z, wt, c, size;
  Inrimage::WORD_TYPE work_type;
  enumComputation typeComputation = _COMPONENTS_;

  bool inverse = false;
  bool threeD = true;
  
  bool print_time = false;
  clock_t t0, t1, t2, t3;

  program = argv[0];
  if ( argc == 1 ) _errorMessage( "\n", 0 );
    
  CE.setOutputType( COMPONENT_LABEL );

  //
  // lecture des arguments
  //

  for ( int nimages = 0, i=1; i<argc; i++ ) {
    int status;
    // arguments
    if ( argv[i][0] == '-' ) {
      if ( argv[i][1] == '\0' ) {
	// standard input
	if ( nimages == 0 ) {
	  nimages ++;
	}
	else {
	  _errorMessage( "standard input specified twice\n", 0 );
	}
      }
      else if ( strcmp ( argv[i], "-help" ) == 0 ) {
	_errorMessage( "\n", 1 );
      }
      
      // thresholds
      else if ( strcmp ( argv[i], "-low-threshold" ) == 0 
		|| strcmp ( argv[i], "-lt" ) == 0 
		|| strcmp ( argv[i], "-sb" ) == 0 ) {
	i += 1;
	if ( i >= argc)    _errorMessage( "parsing -low-threshold...\n", 0 );
	status = sscanf( argv[i],"%lf",&low_threshold );
	if ( status <= 0 ) _errorMessage( "parsing -low-threshold...\n", 0 );
	high_threshold = low_threshold;
      }
      else if ( strcmp ( argv[i], "-high-threshold" ) == 0 
		|| strcmp ( argv[i], "-ht" ) == 0 
		|| strcmp ( argv[i], "-sh" ) == 0 ) {
	i += 1;
	if ( i >= argc)    _errorMessage( "parsing -high-threshold...\n", 0 );
	status = sscanf( argv[i],"%lf",&high_threshold );
	if ( status <= 0 ) _errorMessage( "parsing -high-threshold...\n", 0 );
	typeComputation = _HYSTERESIS_;
	CE.setOutputType( COMPONENT_BINARY );
	thereIsAHighThreshold = true;
      }

      else if ( strcmp ( argv[i], "-minimal-size" ) == 0 
		|| strcmp ( argv[i], "-ms" ) == 0 
		|| strcmp ( argv[i], "-tcc" ) == 0 ) {
	i += 1;
	if ( i >= argc)    _errorMessage( "parsing -minimal-size...\n", 0 );
	status = sscanf( argv[i],"%d",&size );
	if ( status <= 0 ) _errorMessage( "parsing -minimal-size...\n", 0 );
	CE.setMinimalSize( size );
      }

      else if ( strcmp ( argv[i], "-connectivity" ) == 0 
		|| strcmp ( argv[i], "-connexite" ) == 0 
		|| strcmp ( argv[i], "-con" ) == 0 
		|| strcmp ( argv[i], "-c" ) == 0 ) {
	i += 1;
	if ( i >= argc)    _errorMessage( "parsing -connectivity...\n", 0 );
	status = sscanf( argv[i],"%d",&c );
	if ( status <= 0 ) _errorMessage( "parsing -connectivity...\n", 0 );
	CE.setConnectivity( c );
      }

      else if ( strcmp ( argv[i], "-high-threshold-minimal-size" ) == 0 
		|| strcmp ( argv[i], "-high-ms" ) == 0 
		|| strcmp ( argv[i], "-hms" ) == 0 ) {
	i += 1;
	if ( i >= argc)    _errorMessage( "parsing -high-threshold-minimal-size...\n", 0 );
	status = sscanf( argv[i],"%d",&size );
	if ( status <= 0 ) _errorMessage( "parsing -high-threshold-minimal-size...\n", 0 );
	CE.setMinimalSizeHighThreshold( size );
      }

      else if ( strcmp ( argv[i], "-maximal-number-of-connected-components" ) == 0 
		|| strcmp ( argv[i], "-max-cc" ) == 0 
		|| strcmp ( argv[i], "-mcc" ) == 0 
		|| strcmp ( argv[i], "-ncc" ) == 0 ) {
	i += 1;
	if ( i >= argc)    _errorMessage( "parsing -maximal-number-of-connected-components...\n", 0 );
	status = sscanf( argv[i],"%d",&c );
	if ( status <= 0 ) _errorMessage( "parsing -maximal-number-of-connected-components...\n", 0 );
	CE.setMaximalNumberOfComponents( c );
      }
      
      else if ( strcmp ( argv[i], "-seeds" ) == 0 ) {
	i += 1;
	if ( i >= argc)    _errorMessage( "parsing -seeds...\n", 0 );
	seedsImage = argv[i];
	typeComputation = _RECONSTRUCT_;
	CE.setOutputType( COMPONENT_BINARY );
      }

      else if ( strcmp ( argv[i], "-seed" ) == 0 ) {
	i += 1;
	if ( i >= argc)    _errorMessage( "parsing -seed...\n", 0 );
	status = sscanf( argv[i],"%d",&x );
	if ( status <= 0 ) _errorMessage( "parsing -seed...\n", 0 );
	i += 1;
	if ( i >= argc)    _errorMessage( "parsing -seed...\n", 0 );
	status = sscanf( argv[i],"%d",&y );
	if ( status <= 0 ) _errorMessage( "parsing -seed...\n", 0 );
	if ( i+1 < argc && sscanf( argv[i+1],"%d",&z ) == 1 ) {
	  i += 1;
	  CE.addSeedPoint( x, y, z );
	}
	else {
	  CE.addSeedPoint( x, y );
	}
	CE.setOutputType( COMPONENT_BINARY );
      }
      
      else if ( strcmp ( argv[i], "-max" ) == 0 ) {
	CE.setMaximalNumberOfComponents( 1 );
      }
      else if ( strcmp ( argv[i], "-times" ) == 0 ) {
	print_time = true;
      }

      else if ( strcmp ( argv[i], "-binary" ) == 0 || strcmp ( argv[i], "-bin" ) == 0 ) {
	CE.setOutputType( COMPONENT_BINARY );
      }
      else if ( strcmp ( argv[i], "-labels" ) == 0 || strcmp ( argv[i], "-label" ) == 0 ) {
	CE.setOutputType( COMPONENT_LABEL );
      }

      else if ( strcmp ( argv[i], "-work-type" ) == 0 
		||  strcmp ( argv[i], "-wtype" ) == 0 
		||  strcmp ( argv[i], "-wt" ) == 0 ) {
	i += 1;
	if ( i >= argc)    _errorMessage( "parsing -work-type...\n", 0 );
	status = sscanf( argv[i],"%d",&wt );
	if ( status <= 0 ) _errorMessage( "parsing -work-type...\n", 0 );
	switch ( wt ) {
	default :
	  _errorMessage( "parsing -work-type...\n", 0 );
	case 1  : work_type = Inrimage::WT_UNSIGNED_CHAR; break;
	case 2  : work_type = Inrimage::WT_UNSIGNED_SHORT; break;
	case 4  : work_type = Inrimage::WT_UNSIGNED_INT; break;
	}
	CE.setWorkType( work_type );
      }

					     
      // computation type
      else if ( strcmp ( argv[i], "-hysteresis" ) == 0 
		|| strcmp ( argv[i], "-hyst" ) == 0 ) {
	typeComputation = _HYSTERESIS_;
      }

      else if ( strcmp ( argv[i], "-components" ) == 0 
		|| strcmp ( argv[i], "-comp" ) == 0 ) {
	typeComputation = _COMPONENTS_;
      }
      else if ( strcmp ( argv[i], "-fill-contours" ) == 0 
		|| strcmp ( argv[i], "-fill" ) == 0 ) {
	typeComputation = _FILL_CONTOURS_;
      }
      else if ( strcmp ( argv[i], "-reconstruct" ) == 0 
		|| strcmp ( argv[i], "-rec" ) == 0 ) {
	typeComputation = _RECONSTRUCT_;
      }

      else if ( strcmp ( argv[i], "-2D" ) == 0 ) {
	threeD = false;
      }

      // image inversion
      else if ( strcmp ( argv[i], "-inverse" ) == 0 || strcmp ( argv[i], "-inv" ) == 0 ) {
	inverse = true;
      }
      
      /*--- lecture du type de l'image de sortie ---*/
      else if ( strcmp ( argv[i], "-r" ) == 0 || strcmp ( argv[i], "-f" ) == 0  ) {
	r = 1;
      }
      else if ( strcmp ( argv[i], "-s" ) == 0 ) {
	s = 1;
      }
      else if ( strcmp ( argv[i], "-o" ) == 0 ) {
	i += 1;
	if ( i >= argc)    _errorMessage( "parsing -o...\n", 0 );
	status = sscanf( argv[i],"%d",&o );
	if ( status <= 0 ) _errorMessage( "parsing -o...\n", 0 );
      }
      /*--- option inconnue ---*/
      else {
	char message[256];
	sprintf( message, "unknown option '%s'\n", argv[i] );
	_errorMessage( message, 0 );
      }
    }
    // image names
    else {
      if ( nimages == 0 ) {
	inputImage = argv[i];
	nimages ++;
      }
      else if ( nimages == 1 ) {
	outputImage = argv[i];
	nimages ++;
      }
      else {
	_errorMessage( "too much file names when parsing\n", 0 );
      }
    }
  }
  //
  // determination du type
  //
  switch ( r ) {
  case 0 :
    switch ( o ) {
    case 1 :
      outputType = (s == 0) ? Inrimage::WT_UNSIGNED_CHAR : Inrimage::WT_SIGNED_CHAR;
      outputTypeIsSet = true; 
      break;
    case 2 :
      outputType = (s == 0) ? Inrimage::WT_UNSIGNED_SHORT : Inrimage::WT_SIGNED_SHORT;
      outputTypeIsSet = true; 
      break;
    case 4 :
      outputType = (s == 0) ? Inrimage::WT_UNSIGNED_INT : Inrimage::WT_SIGNED_INT;
      outputTypeIsSet = true; 
      break;
    case 8 :
      outputType = (s == 0) ? Inrimage::WT_UNSIGNED_LONG : Inrimage::WT_SIGNED_LONG;
      outputTypeIsSet = true; 
      break;
    }
    break;
  case 1 :
    switch ( o ) {
    case 0 :
    case 4 : outputType = Inrimage::WT_FLOAT;  outputTypeIsSet = true; break;
    case 8 : outputType = Inrimage::WT_DOUBLE; outputTypeIsSet = true; break;
    }
  }
  //
  // fin de la lecture des arguments
  //
  


  try {

    //
    // lecture de l'image
    //
    t0 = clock();
    image = new Inrimage( inputImage );

    t1 = clock();
    if ( inverse == true ) logicalNot( image );
    
    switch( typeComputation ) {
    default :
    case _COMPONENTS_ :
      if ( thereIsAHighThreshold == true )
	connectedComponents( image, outputType, low_threshold, high_threshold, &CE );
      else 
	connectedComponents( image, outputType, low_threshold, &CE );
      break;
    case _HYSTERESIS_ :
      hysteresis( image, outputType, low_threshold, high_threshold, &CE );
      break;
    case _FILL_CONTOURS_ :
      if ( threeD == true ) {
	fillContours( image, outputType, &CE );
      } else {
	fillContours( image, outputType, &CE, NULL, NULL, threeD );
      }
      break;
    case _RECONSTRUCT_ :
      if ( seedsImage != NULL ) seeds = new Inrimage( seedsImage );
      reconstruct( image, outputType, &CE, seeds );
      break;
    }

    if ( inverse == true && CE.getOutputType() == COMPONENT_BINARY ) logicalNot( image );

    //
    // ecriture
    //
    t2 = clock();
    image->write( outputImage );
    t3 = clock();
  }
  catch( Exception e)  {
    CERR << e << ENDL;
    exit( -1 );
  }

  if ( print_time == true ) {
    printf( "%s: reading= %f sec,\n\tprocessing= %f sec,\n\twriting= %f sec\n",
	    argv[0], 
	    (double)(t1-t0)/(double)CLOCKS_PER_SEC,
	    (double)(t2-t1)/(double)CLOCKS_PER_SEC,
	    (double)(t3-t2)/(double)CLOCKS_PER_SEC );
  }

  //
  //
  //
  delete image;
  if ( seeds != NULL ) delete seeds;


  return( 0 );
}
