/** test program for image I/0
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <inrimage/ImageIO.h>


static char *options="\
[input|-] [output|-] [-print] [-h|-help]";

static char *details="\
\n\
Convert images from one format to another\n\
\n\
input : name of the input file\n\
        if none or \'-\' is provided, read from stdin\n\
output: name of the output file\n\
        if none or \'-\' is provided, write on stdout\n\
-print: print available formats\n\
-help : display this message\n";



void error( char *name, 
	    int fullhelp,
	    char *error,
	    int exitcode )
{
  fprintf( stderr, "%s %s\n", name, options );
  if ( fullhelp ) 
    fprintf( stderr, "%s\n", details );
  if ( error )
    fprintf( stderr, "error: %s\n", error );
  exit( exitcode );
}



int main ( int argc, char *argv[] )
{
  int i, n=0;
  char *input = NULL;
  char *output = NULL;
  _image *theIm;
  int ret;

  if ( argc == 1 ) {
    error( argv[0], 0, NULL, 0 );
  }

  for ( i=1; i<argc; i++ ) {
    if ( argv[i][0] == '-' ) {
      if ( strlen( argv[i] ) == 1 ) {
	if ( n == 0 ) {
	  input = argv[i];
	}
	else if ( n == 1 ) {
	  output = argv[i];
	}
	else {
	  error( argv[0], 0, "too many filenames or \'-\'", -1 );
	}
	n++;
      }
      else if ( (strncmp( argv[i], "-help", 5 ) == 0 && strlen( argv[i] ) == 5)
		|| (strncmp( argv[i], "-h", 2 ) == 0 && strlen( argv[i] ) == 2) ) {
	error( argv[0], 1, NULL, 0 );
      }
      else if ( strncmp( argv[i], "-print", 6 ) == 0 && strlen( argv[i] ) == 6 ) {
	printSupportedFileFormat();
	printf( "\n" );
	exit( 0 );
      }
    }
    else {
      if ( n == 0 ) {
	input = argv[i];
      }
      else if ( n == 1 ) {
	output = argv[i];
      }
      else {
	error( argv[0], 0, "too many filenames or \'-\'", -1 );
      }
      n++;
    }
  }

  if ( 0 )
    fprintf( stderr, "... reading '%s'\n", input );
  theIm = _readImage( input );

  if ( theIm == NULL ) {
    error( argv[0], 0, " when reading input", -1 );
  }

  if ( 0 )
    fprintf( stderr, "... writing %s\n", output );
  ret =  _writeImage( theIm, output );

  if ( ret != 0 ) {
    error( argv[0], 0, " when writing output", -1 );
  }

  exit( 0 );
}
