#include <inrimage/Inrimage.h>
#include <inrimage/arithmetic.h>

using namespace yav;


static char *usage = "[image-in] [image-out]\n\
   [-inv] [-norma] [-swap] [-r] [-o %d] [-s]\n\
   [-v] [-help]\n";

static char *detail = "\
 si 'image-in' est '-', on prendra stdin\n\
 si 'image-out' est absent, on prendra stdout\n\
 si les deux sont absents, on prendra stdin et stdout\n\
\n\
  $Revision: 1.6 $\n";


static char *program = NULL;

static void _errorMessage( char *str, int flag )
{
  CERR<<" Usage: "<<program<<" "<<usage<<"\n"<<ENDL;
  if ( flag == 1 )
    CERR<<detail<<ENDL;
  CERR<<" Error: "<<str<<ENDL;
  exit( -1 );
}



int main(int argc, char *argv[])
{
  char *inputImage  = NULL;
  char *outputImage = NULL;

  Inrimage *image = NULL;

  WORD_KIND wkind = WK_UNKNOWN; /* WK_FIXED WK_FLOAT */
  SIGN      sign  = SGN_UNKNOWN; /* SGN_SIGNED SGN_UNKNOWN */
  unsigned int wdim = 0;
  Inrimage::WORD_TYPE wtype = Inrimage::WT_FLOAT;
  
  bool inversion = false;
  bool swapping = false;
  bool normalization = false;
  
  /* entre lacee ou pas , vectoriel... */

  program = argv[0];
  if ( argc == 1 ) _errorMessage( "\n", 0 );
    
  //
  // lecture des arguments
  //

  for ( int nimages = 0, i=1; i<argc; i++ ) {
    int status;
    // arguments
    if ( argv[i][0] == '-' ) {
      if ( argv[i][1] == '\0' ) {
	// standard input
	if ( nimages == 0 ) {
	  nimages ++;
	}
	else {
	  _errorMessage( "standard input specified twice\n", 0 );
	}
      }
      else if ( strcmp ( argv[i], "-help" ) == 0 ) {
	_errorMessage( "\n", 1 );
      }
      


      else if ( strcmp ( argv[i], "-norma" ) == 0 ) {
	normalization = true;
      }

      else if ( strcmp ( argv[i], "-inv" ) == 0 ) {
	inversion = true;
      }

      else if ( strcmp ( argv[i], "-swap" ) == 0 ) {
	swapping = true;
      }


      else if ( strcmp ( argv[i], "-octets"  ) == 0 ||
		strcmp ( argv[i], "-o"  ) == 0 ||
		strcmp ( argv[i], "-bytes"  ) == 0 ||
		strcmp ( argv[i], "-b"  ) == 0 ) {
	i++;
	if ( i >= argc )   _errorMessage( "parsing -bytes ...\n", 0 );
	status = sscanf( argv[i],"%d",&wdim );
	if ( status != 1 ) _errorMessage( "parsing -bytes ...\n", 0 );
      }

      else if ( strcmp ( argv[i], "-reel"  ) == 0 ||
		strcmp ( argv[i], "-real"  ) == 0 ||
		strcmp ( argv[i], "-r"  ) == 0 ) {
	wkind = WK_FLOAT;
      }
      
      else if ( strcmp ( argv[i], "-signed"  ) == 0 ||
		strcmp ( argv[i], "-s"  ) == 0 ) {
	sign  = SGN_SIGNED;
      }
      
      else if ( strcmp ( argv[i], "-float" ) == 0 ) {
	wkind = WK_FLOAT;
	wdim  = 4;
      }
      else if ( strcmp ( argv[i], "-double" ) == 0 ) {
	wkind = WK_FLOAT;
	wdim  = 8;
      }
      else {
	char message[256];
	sprintf( message, "unknown option '%s'\n", argv[i] );
	_errorMessage( message, 0 );
      }
    }
    
    // image names
    else {
      if ( nimages == 0 ) {
	inputImage = argv[i];
	nimages ++;
      }
      else if ( nimages == 1 ) {
	outputImage = argv[i];
	nimages ++;
      }
      else {
	_errorMessage( "too much file names when parsing\n", 0 );
      }
    }
  }
  //
  // fin de la lecture des arguments
  //
  

  //
  // new image type ?
  //
  if ( wdim > 0 || wkind != WK_UNKNOWN ) {
    switch ( wkind ) {
    default :
    case WK_FIXED :
      switch( wdim ) {
      case 1 : wtype = ( sign == SGN_SIGNED ) ? Inrimage::WT_SIGNED_CHAR  : Inrimage::WT_UNSIGNED_CHAR;  break;
      case 2 : wtype = ( sign == SGN_SIGNED ) ? Inrimage::WT_SIGNED_SHORT : Inrimage::WT_UNSIGNED_SHORT; break;
      case 4 : wtype = ( sign == SGN_SIGNED ) ? Inrimage::WT_SIGNED_INT   : Inrimage::WT_UNSIGNED_INT;   break;
      case 8 : wtype = ( sign == SGN_SIGNED ) ? Inrimage::WT_SIGNED_LONG  : Inrimage::WT_UNSIGNED_LONG;  break;
      default : _errorMessage( " unable to deal with such word dim in FIXED images", 0 );
      }
      break;
    case WK_FLOAT :
      switch( wdim ) {
      case 4 : wtype = Inrimage::WT_FLOAT;  break;
      case 8 : wtype = Inrimage::WT_DOUBLE; break;
      default : _errorMessage( " unable to deal with such word dim in FLOAT images", 0 );
      }
      break;
    }
  }




  try {

    //
    // lecture de l'image
    //
    image = new Inrimage( inputImage );


    //
    // inversion
    //
    if ( inversion == true )
      logicalNot( image );

    //
    // swapping 
    //
    if ( swapping == true )
      fprintf( stderr, "%s: swapping not implemented yet\n", argv[0] );

    //
    // normalisation
    //
    if ( normalization == true ) {
      if ( wdim > 0 || wkind != WK_UNKNOWN ) {
	image->convertScale( wtype );
      }
      else {
	image->convertScale( image->getType() );
      }
    }

    //
    // conversion
    //
    if ( wdim > 0 || wkind != WK_UNKNOWN ) {
      if ( image->getType() != wtype ) image->convertCast( wtype );
    }

    //
    // entralacage
    //

    //
    // ecriture
    //
    image->write( outputImage );
  }

  catch( Exception e)  {
    CERR << e << ENDL;
    exit( -1 );
  }

    
  //
  //
  //
  delete image;

  return( 0 );
}
