#include <inrimage/Inrimage.h>
#include <inrimage/recursiveFilters.h>

using namespace yav;
char usage[]="zgradient input_image output_image [-2D | -3D ] [ -rec | -sobel ] [-marta | -gaussian | -deriche] [-sigma s] [-dir] [-max] [-thres threshold] \n";
/*---------------------------*/
static void print_info()
{
  fprintf(stderr,"%s",usage);
  fprintf(stderr,"gradient-inr computes the image gradient \n"); 
}

/*---------------------------*/
int main(int argc,char **argv)
{
  Inrimage  *inr1;
  double sigma;
  bool compute_maximum,use_sobel,compute_2D,compute_direction;
  double minimum_gradient;
  char gradient_name[300],direction_name[300],*prefix;
  int i;

  if (argc<3)
    {
      print_info();
      exit(1);
    }
  
  
  /* Load Image */
  fprintf(stderr,"Loading Image....");
  try {
    inr1=new Inrimage(argv[1]);
  } 
  catch (Exception) {
    CERR << "Cannot load image " << argv[1]<<ENDL;
    exit(1);
  }
  fprintf(stderr,"Done\n");
  /* Initialize variables */
  compute_maximum=false;
  GRADIENT_DIRECTION dir=GD_NONE;
  compute_2D=false;
  use_sobel=true;
  RECURSIVE_FILTER_TYPE filter=DERICHE_GAUSSIAN;
  sigma=3.0;
  minimum_gradient=1e-4;
  /* Parsing parameters */
  for (i=3;i<argc;++i)
    {
      if (strncmp(argv[i],"-2D",3)==0)
	{
	  compute_2D=true;
	}
      else if (strncmp(argv[i],"-3D",3)==0)
	{
	  compute_2D=false;
	}
      else if (strncmp(argv[i],"-gaussian",9)==0)
	{
	  use_sobel=false;
	  filter=DERICHE_GAUSSIAN;
	}
      else if (strncmp(argv[i],"-marta",6)==0)
	{
	  use_sobel=false;
	  filter=MARTA_GAUSSIAN;
	}
      else if (strncmp(argv[i],"-deriche",8)==0)
	{
	  use_sobel=false;
	  filter=DERICHE_STEP_EDGE;
	}
      else if (strncmp(argv[i],"-sobel",6)==0)
	{
	  use_sobel=true;
	}
      else if (strncmp(argv[i],"-rec",4)==0)
	{
	  use_sobel=false;
	}
      else if (strncmp(argv[i],"-dir",4)==0)
	{
	  dir=GD_VECTORS;
	}
      else if (strcmp(argv[i],"-sample")==0)
	{
	  dir=GD_SAMPLE;
	}
      else if (strncmp(argv[i],"-max",4)==0)
	{
	  compute_direction=true;
	  compute_maximum=true;
	}
      else if ((strncmp(argv[i],"-sigma",6)==0) && (i!=(argc-1)))
	{
	  sigma=(double) atof(argv[++i]);
	}
      else if ((strncmp(argv[i],"-thres",6)==0) && (i!=(argc-1)))
	{
	  minimum_gradient=atof(argv[++i]);
	}
      else
	{
	  fprintf(stderr,"Can't parse Option %s\n",argv[i]);
	  exit(2);
	}
    }
  if (strlen(argv[2])>7) 
    {
      prefix=argv[2]+(strlen(argv[2])-7);
      if (strcmp(prefix,".inr.gz")==0)
	*prefix='\0';
    }
  if (strlen(argv[2])>4)
    {
      prefix=argv[2]+(strlen(argv[2])-4);
      if (strcmp(prefix,".inr")==0)
	*prefix='\0';
    }
  sprintf(gradient_name,"%s.gradient.inr.gz",argv[2]);
  sprintf(direction_name,"%s.gradient_direction.inr.gz",argv[2]);
  
  Inrimage *output, *outdir=0;
  fprintf(stderr,"Computing gradient....");
  if (use_sobel==true)
    {
      sobel(inr1, output, outdir, (int) sigma, dir, 
	    compute_maximum, compute_2D, minimum_gradient);
    }
  else
    {
      gradient(inr1, output, outdir, (int) sigma, filter,dir, 
	       compute_maximum, compute_2D, minimum_gradient);
    }
  fprintf(stderr,"Done\n");


  fprintf(stderr,"Saving gradient....");
  output->write(gradient_name);
  if (outdir)
    outdir->write(direction_name);
  fprintf(stderr,"Done\n"); 
  return 0;
}

