#include <inrimage/Inrimage.h>
using namespace yav;

int usage(char *progName) {
  CERR << "usage: " << progName
       << " file_names [-voxel %f %f %f] [-geom <cartesian|cylindrical|spherical>]]"
       << ENDL;
  return -1;
}


void info(Inrimage *inr, char *name) {
  COUT << name;
  COUT << inr->getFileFormat()->realName;
  COUT << ENDL;
  COUT << "   size = " << inr->getX() << "*" << inr->getY() << "*"
       << inr->getZ() << ENDL;
  if(inr->getVdim() > 1) COUT << "   vectorial dim = "
			      << inr->getVdim() << ENDL;
  else COUT << "   scalar" << ENDL;
  COUT << "   " << inr->getWdim() << " byte"
       << ((inr->getWdim() > 1) ? "s" : "") << " per word";
  switch(inr->getWordKind()) {
  case WK_FIXED: COUT << ", fixed"; break;
  case WK_FLOAT: COUT << ", float"; break;
  case WK_UNKNOWN: break;
  }
  switch(inr->getWordSign()) {
  case SGN_SIGNED: COUT << ", signed" << ENDL; break;
  case SGN_UNSIGNED: COUT << ", unsigned" << ENDL; break;
  case SGN_UNKNOWN: break;
  }
  COUT << "   voxel size = " << inr->getVX() << "*" << inr->getVY()
       << "*" << inr->getVZ() << ENDL;
  switch(inr->getGeometry()) {
  case Inrimage::CARTESIAN: COUT << "   cartesian"; break;
  case Inrimage::CYLINDRICAL: COUT << "   cylindrical"; break;
  case Inrimage::SPHERICAL: COUT << "   spherical"; break;
  }
  COUT << " geometry" << ENDL;
}


int main(int argc, char **argv) {
  Inrimage *inr;
  int i, n;
  bool setvs = false, setgeom = false;
  float vx, vy, vz;
  Inrimage::GEOMETRY geom = Inrimage::CARTESIAN;

  for(n = 1; (n < argc) && (argv[n][0] != '-'); n++);
  if(n < 1) return usage(argv[0]);

  if(n < argc) {
    for(i = n; i < argc; i++) {
      if(!strcmp(argv[i], "-voxel")) {
	setvs = true;
	if(i+3 >= argc) return usage(argv[0]);
	if(sscanf(argv[i+1], "%f", &vx) != 1) return usage(argv[0]);
	if(sscanf(argv[i+2], "%f", &vy) != 1) return usage(argv[0]);
	if(sscanf(argv[i+3], "%f", &vz) != 1) return usage(argv[0]);
	i += 3;
      }
      else if(!strcmp(argv[i], "-geom")) {
	setgeom = true;
	if(++i >= argc) return usage(argv[0]);
	if(!strcmp(argv[i], "cartesian")) geom = Inrimage::CARTESIAN;
	else if(!strcmp(argv[i], "cylindrical")) geom = Inrimage::CYLINDRICAL;
	else if(!strcmp(argv[i], "spherical")) geom = Inrimage::SPHERICAL;
	else return usage(argv[0]);
      }
      else return usage(argv[0]);
    }
  }

  for(i = 1; i < n; i++) {
    try {
      inr = new Inrimage(argv[i], !(setvs || setgeom));
      if(setvs) inr->setVoxelSize(vx, vy, vz);
      if(setgeom) inr->setGeometry(geom);
      info(inr, argv[i]);
      if(setvs || setgeom) inr->write(argv[i]);
    }
    catch(Inrimage::UnableToCreateFile) {
      CERR << argv[i] << ": impossible to create file" << ENDL;
    }
    catch(Inrimage::UnableToWriteHeader) {
      CERR << argv[i] << ": header write error" << ENDL;
    }
    catch(Inrimage::UnableToWriteData) {
       CERR << argv[i] << ": data write error" << ENDL;
    }
    catch(...) {
      CERR << argv[i] << ": write error" << ENDL;
    }
  }

  return 0;
}
