#include <stdlib.h>


#ifdef WIN32
/// detect memory leaks
#define CRTDBG_MAP_ALLOC
#include <stdlib.h>
#include <crtdbg.h>

#endif

#include <iostream>
#include <tcl.h>
#include <tk.h>
#include <basics/AutoLink.h>
#include <modules/TclModule.h>
#include <inrimage/TclInrimage.h>

#ifdef STATIC_OBJECT
#include <graphics2D/Graphics2DCmd.h>
#include <inrimage/InrimageCmd.h>
#endif

#ifdef EMBEDED_TCL_SCRIPTS
#include "yav++.tcl.h"
#endif

using namespace yav;

static void readStdin(ClientData, int);
static TclString command, initrc = EMPTY_STRING;
static Tcl_Interp *globinterp;



class AutoExit : public WarnedObject {
protected:
  bool exitPending;
public:
  AutoExit() : WarnedObject(typeid(TclCamera2D)) {
  exitPending=false;}
  static void doExit(ClientData clientData) {
    TclModule::exit();
  }
  void emptyObjectsList(void) { 
    if (exitPending==false)
      {
	exitPending=true;
	Tk_DoWhenIdle(AutoExit::doExit,NULL);
      }
  }
  void createdObject(TclObject *, const std::type_info&) {}
  void deletedObject(TclObject *, const std::type_info&) {}
};



static int letsgo(TclModule *core, int argc, char **argv);
static void usage(char *name);



static char releaseInfo[] = "\
\n\
                              zviewer\n\
\n\
         A script interpretor for volumetric image processing.\n\
         (in fact an extensible C++ front-end for TCL/TK)\n\
\n\
                          Version 1.4 (beta)\n\
\n\
authors:\n\
  Johan Montagnat (Johan.Montagnat@sophia.inria.fr)\n\
  Herv Delingette (Herve.Delingette@sophia.inria.fr)\n\
contributor:\n\
  Alexandre Guimond (Alexandre.Guimond@sophia.inria.fr)\n\
\n\
  INRIA Sophia-Antipolis, Epidaure project\n\
  2004 route des lucioles, BP93, 06902 Sophia-Antipolis, France\n\
  http://www.inria.fr/epidaure/Epidaure-eng.html\n\
\n\
Using TCL/TK 8.1\n\
  http://www.scriptics.com/software/8.3.html\n\
\n";


/* Main: let's go */
int main(int argc, char *argv[]) {
  bool verbose = false, rcopt = false;
  char *rfile = 0;
  unsigned int nbModules;
  TclString *modules;
  /// do not load the file dynamically with dlopen()
  TclModule::loadDynamically=false;
#ifdef WIN32
_CrtSetDbgFlag ( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
#endif
  try {
    if(argc > 1 && !strcmp(argv[1], "-help")) usage(argv[0]);
    for(int i = 1; i < argc; i++) {
      if(!strcmp(argv[i], "-rc")) {
	rcopt = true;
	if(++i >= argc) rfile = 0;
	else {
	  if(argv[i][0] == '-') rfile = 0;
	  else rfile = argv[i];
	}
      }
      else if(!strcmp(argv[i], "-verbose")) verbose = true;
      else if(!strcmp(argv[i], "-version")) {
	COUT << releaseInfo << ENDL;
      }
    }

#ifdef STATIC_OBJECT
    INIT_MODULE_FUNCTION j=Tclgraphics2d_Init;
    TclModule::getModuleNameMap().insert(std::pair<char *,INIT_MODULE_FUNCTION>(((char *)TclGraphics2DInit::tclGraphics2DName),j));
    j=Tclinrimage_Init;
    TclModule::getModuleNameMap().insert(std::pair<char *,INIT_MODULE_FUNCTION>(((char *)TclInrimageInit::tclInrimageName),j));
#endif
    if(rcopt && rfile) TclModule::readrc(rfile, initrc, nbModules, modules);
   TclModule::init(letsgo, verbose, argc, argv, "tclGraphics2D", "tclInrimage", 0);
  }
  catch(...) {
    CERR << argv[0] << ": fatal: Unexpected exception caught" << ENDL;
    exit(-1);
  }

  return 0;
}


void usage(char *name) {

  fprintf(stderr, "usage: %s [[image_file_1 [image_file_2 [...]] | script_file | -command %%s | -text | -graphic]\n\
  image_file_1, image_file_2, ... : display images (\"-\" means read image from stdin)\n\
  script_file                     : evaluate TCL script in file and exit\n\
  -command %%s                    : evaluate given command and exit\n\
  -init %%s                       : evaluate given string before anything else\n\
  -rc [%%s]                       : load given ressource file (none if no ressource file\n\
                                   name is given).\n\
                                   By default, $HOME/.zviewerrc file is read.\n\
  -load %%s...                    : load given modules\n\
  -version                        : print out release informations\n\
  -verbose                        : print out lot of message (libraries search paths...)\n\
  -text|-graphic                  : select either text or graphic (tk) interface\n",
	  name);
  
  exit(0);
}


enum ARGS {
  NONE,
  CMD,
  TEXT,
  GRAPHIC,
  FILES
};

// Starting point called by TK
int letsgo(TclModule *core, int argc, char **argv) {
  Tcl_Channel inChannel;
  TclInrimage *inr;
  int i, j, ninr = 0;
  char *cmdstr = 0;
  ARGS cmd = NONE;
  int ninit = 0;
  char *scriptname=0;

  for(i = 1; i < argc; i++) {

    if(strlen(argv[i]) > 0) {
      if(!strcmp(argv[i], "-") || argv[i][0] != '-') {
	if(cmd != NONE) usage(argv[0]);
	cmd = FILES;
	argv += i;
	argc -= i;
	/*	for(j = 1; j < argc; j++)
		if(argv[j][0] == '-') usage(argv[0]);*/
	break;
      }
      else {
	if(!strcmp(argv[i], "-command")) {
	  if(cmd != NONE) usage(argv[0]);
	  if(++i < argc && argv[i][0] != '-') {
	    cmdstr = argv[i];
	    cmd = CMD;
	  }
	  else usage(argv[0]);
	}
	else if(!strcmp(argv[i], "-text")) {
	  if(cmd != NONE) usage(argv[0]);
	  cmd = TEXT;
	  }
	else if(!strcmp(argv[i], "-graphic")) {
	  if(cmd != NONE) usage(argv[0]);
	  cmd = GRAPHIC;
	}
	else if(!strcmp(argv[i], "-init")) {
	  if(++i < argc && argv[i][0] != '-')
	    ninit = i;
	}	
	else if(!strcmp(argv[i], "-load")) {
	  if(++i < argc) {
	    int j;
	    for(j = i; j < argc && argv[j][0] != '-'; j++) {
	      try {
		(void) core->loadNewModule(core->getInterp(), argv[j]);
	      }
	      catch(TclModule::ModuleAlreadyExists) {
		COUT << ENDL << "yav++: warning: attempt to load module "
		     << argv[j] << " twice" << ENDL;
	      }
	      catch(Exception& e) {
		CERR << ENDL << e << ENDL;
		exit(-1);
	      }
	    }
	    i = j - 1;
	  }
	}
	else if(!strcmp(argv[i], "-version")) ;
	else if(!strcmp(argv[i], "-verbose")) ;
	else if(!strcmp(argv[i], "-rc"))  {
	  if(i + 1 < argc && argv[i+1][0] != '-') i++;
	}
	else {
	  fprintf(stderr, "%s: error: invalid option \'%s\'\n", argv[0], argv[i]);
	  exit(-1);
	}
      }
    }
  }
  
  globinterp = core->getInterp(); 
 
  /* handle initialization command */
  if(ninit > 0) {
    TclString str;
    str.sprintf("modules::evaluate \"%s\"",argv[ninit]);
    core->tickle(argv[ninit]);
  }
  core->tickle("wm withdraw .");

  if(initrc != EMPTY_STRING) {
    if(core->tickle((char *) initrc) == TCL_ERROR) {
      CERR << argv[0] << " warning: error while executing .zviewerrc init command."
	   << ENDL;
    }
  }

  switch(cmd) {
  case NONE:
  case GRAPHIC:

#ifdef EMBEDED_TCL_SCRIPTS
  /* do not load tcl scripts dynamically anymore */
  Tcl_Eval(core->getInterp(), yav___tcl);
#else
    try {
      core->loadSource("yav++.tcl");
    }
    catch(TclModule::ShellNotFound&) {
      CERR << "Could not find yav++.tcl" << ENDL
	   << "Recompile libModules with the right path " << ENDL;
      exit(-1);
    }
#endif

    /* launch graphic interface */
    core->tickle("yav::interface");
    TclModule::run();
#ifdef WIN32
	_CrtDumpMemoryLeaks();
#endif
    break;

  case TEXT:
    inChannel = Tcl_GetStdChannel(TCL_STDIN);
    Tcl_CreateChannelHandler(inChannel, TCL_READABLE, readStdin,
			     (ClientData) inChannel);
    TclModule::Prompt(globinterp, 0);
    Tcl_ResetResult(globinterp);

    TclModule::run();
    break;

  case CMD:
    core->tickle(cmdstr);
    break;

  case FILES:

#ifdef EMBEDED_TCL_SCRIPTS
  /* do not load tcl scripts dynamically anymore */
  Tcl_Eval(core->getInterp(), yav___tcl);
#else
    try {
      core->loadSource("yav++.tcl");
    }
    catch(TclModule::ShellNotFound&) {
      CERR << "Could not find yav++.tcl" << ENDL
	   << "Recompile libModules with the right path " << ENDL;
      exit(-1);
    }
#endif

    // if file is not an image, consider this is a script file
    if(strcmp(argv[0], "-") && imageType(argv[0]) == 0) {
      core->tickle("script %s", argv[0]);
      TclModule::exit(0);
    }

    std::list<TclInrimage *> inrlist;
    // else display all inrimages
    for(scriptname=0,ninit=0,i = 0; i < argc; i++) {
      if(!strcmp(argv[i], "-init")) {
	if(++i < argc && argv[i][0] != '-')
	  ninit = i;
      }
      else if(!strcmp(argv[i], "-script")) {
	if(++i < argc && argv[i][0] != '-') {
	  scriptname=argv[i];
	}
      }
      else {
	try {
	  if(argv[i][0] == '-') inr = new TclInrimage(core, (const char *) 0);
	  else 
	    inr = new TclInrimage(core, argv[i]);
	  ninr++;
	  inrlist.push_back(inr);
	}
	
	catch(Inrimage::UnknownImageType) {
	  fprintf(stderr, "%s: warning: unknown image type \'%s\'\n",
		  argv[0], argv[i]);
	  continue;
	}
	catch(Inrimage::ImageNotFound) {
	  fprintf(stderr, "%s: warning: file \'%s\' not found\n", argv[0], argv[i]);
	  continue;
	}
	catch(Inrimage::InvalidHeader) {
	  fprintf(stderr, "%s: warning: invalid header in \'%s\'\n",
		  argv[0], argv[i]);
	  continue;
	}
	catch(Inrimage::InvalidData) {
	  fprintf(stderr, "%s: warning: invalid data in \'%s\' (truncated file?)\n",
		  argv[0], argv[i]);
	  continue;
	}
	catch(Inrimage::UnknownType) {
	  fprintf(stderr, "%s: warning: unhandled inrimage type \'%s\'\n",
		  argv[0], argv[i]);
	  continue;
	} 
	core->tickle("view %s%u", TclInrimage::getTCLObjectName(), inr->getRef());
      }
    }
    if(ninr == 0) TclModule::exit();
    std::list<TclInrimage *>::iterator it1,it2;
    for(i=1,it1=inrlist.begin();it1!=inrlist.end();it1++,i++)
      {
	for(j=i+1,it2=it1,it2++;it2!=inrlist.end();it2++,j++) {
	  if (((*it2)->getX()==(*it1)->getX()) && 
	      ((*it2)->getY()==(*it1)->getY()) &&
	      ((*it2)->getZ()==(*it1)->getZ()))
	    {
	      core->tickle("multiview%u -synchronize multiview%u",
			   i,j);
	    }
	}
      }
    // if necessary execute a script after loading a tetrahedrisation
    if(scriptname) {
      core->tickle("script %s", scriptname);
      TclModule::exit(0);
    }  
    else {
      if(ninit > 0) {
	TclString str;
	str.sprintf("modules::evaluate \"%s\"",argv[ninit]);
	core->tickle(argv[ninit]);
      }
      AutoExit autoexit;
      TclModule::run();
    }
    break;
  }

  exit(0);
  return TCL_OK;
}



/* read stdin in text interface */
static void readStdin(ClientData clientData, int mask) {
  static int partial = 0;
  Tcl_Channel outChannel, inChannel = (Tcl_Channel) clientData;
  Tcl_DString line;

  Tcl_DStringInit(&line);

  if(Tcl_Gets(inChannel, &line) < 0) {
    outChannel = Tcl_GetChannel(globinterp, "stdout", NULL);
    Tcl_Write(outChannel, "\n", 1);
    Tcl_Exit(0);
    return;
  }

  
  command += Tcl_DStringValue(&line);
  TclString cmd = command + "\n";
  Tcl_DStringFree(&line);
    
  if(!Tcl_CommandComplete((char *) cmd)) partial = 1;
  else {
    partial = 0;
    Tcl_CreateChannelHandler(inChannel, 0, readStdin, (ClientData) inChannel);
    Tcl_DStringInit(&line);
    switch(Tcl_RecordAndEval(globinterp, (char *) cmd, TCL_EVAL_GLOBAL)) {
    case TCL_OK:
      break;
    case TCL_ERROR:
      Tcl_DStringAppend(&line, "*Error* ", -1);
      break;
    case TCL_RETURN:
      Tcl_DStringAppend(&line, "*Return* ", -1);
      break;
    case TCL_BREAK:
      Tcl_DStringAppend(&line, "*Break* ", -1);
      break;
    case TCL_CONTINUE:
      Tcl_DStringAppend(&line, "*Continue* ", -1);
      break;
    }
    Tcl_DStringAppend(&line, globinterp->result, -1);
    if(*Tcl_DStringValue(&line)) puts(Tcl_DStringValue(&line));
    Tcl_DStringFree(&line);
    Tcl_CreateChannelHandler(inChannel, TCL_READABLE, readStdin,
			     (ClientData) inChannel);
    command = "";
  }

  TclModule::Prompt(globinterp, partial);
  Tcl_ResetResult(globinterp);
}

